#!/usr/bin/python

import re
import argparse
import hashlib
import base64
import xml.etree.ElementTree
import xml.dom.minidom
import socket
import struct
import unittest

def get_section_from_debug_file(f, secname):
    """Look for:
       --------------------------------------------------------------------------------
       Contents of "/etc/rinstrum/factory1.rdb":
       <rinstrum>...</rinstrum>
       --------------------------------------------------------------------------------
       and parse out the XML
    """

    sep_line = "--------------------------------------------------------------------------------"
    contents_regex = "^Contents of \"([^\"]+)\":.*$"

    LOOK_FOR_SEP = 0
    LOOK_FOR_CONTENTS_LINE = 1
    LOOK_FOR_END_SEP = 2

    state = LOOK_FOR_SEP
    capture = False
    section = []

    for line in f:
        line = line.rstrip("\n")
        if (state == LOOK_FOR_SEP):
            if (line == sep_line):
                state = LOOK_FOR_CONTENTS_LINE
        elif (state == LOOK_FOR_CONTENTS_LINE):
            m = re.search(contents_regex, line)
            if (m is not None and m.group(1) is not None and m.group(1) == secname):
                capture = True
            state = LOOK_FOR_END_SEP
        elif (state == LOOK_FOR_END_SEP):
            if (line == sep_line):
                state = LOOK_FOR_CONTENTS_LINE
                if (capture):
                    break
            elif (capture):
                section.append(line)
    return section

def factory_file_to_xml(facinfo):
    return xml.etree.ElementTree.fromstring(facinfo)

def get_factory_db_xml(f, copy):
    """Get the factory db from the file and convert to XML"""
    secname = "/etc/rinstrum/factory%d.rdb" % (copy)
    facinfo = get_section_from_debug_file(f=f, secname=secname)
    facstr = "\n".join(facinfo)
    return factory_file_to_xml(facstr)

def parse_uboot_env_lines(ubootinfo):
    """Convert uboot environment lines to list of (name, value) tuples"""
    faclist = []
    for line in ubootinfo:
        nv = line.split("=", 1)
        if (len(nv) == 2):
            name = nv[0]
            value = nv[1]
            faclist.append((name, value))
    return faclist

def get_factory_db_uboot(f):
    """Get the uboot copy of factory db from the debug file and convert to list of name value tuples"""
    ubootinfo = get_section_from_debug_file(f=f, secname="fw_printenv")
    return parse_uboot_env_lines(ubootinfo=ubootinfo)

def to_buf(value):
    """Convert 32bit value to buffer in network endian format"""
    return struct.pack("!L", value & 0xFFFFFFFF)

def factory_db_xmlarray_to_list(facdbxml, name):
    """Convert array elements to list"""

    nelems_path = ".//database/elem[@name='%s']" % (name)
    nelems = int(facdbxml.find(nelems_path).get("numelems"))

    values = []
    for i in range(0, nelems):
        elem_path = ".//database/elem[@name='%s']/arrayelem[@index='%d']" % (name, i)
        value = int(facdbxml.find(elem_path).text)
        values.append(value)
    return values

def factory_values_to_buffer(serno, caltemp, zerooffs, zerodriftpos, zerodriftneg, spancal, spandriftpos, spandriftneg):
    """Convert factory database values into a buffer. This is the "security by obscurity" bit that shouldn't be published"""

    # Convert to buffer - assumes each array has the same number of elements
    buffer = to_buf(serno) + to_buf(caltemp)
    num_elem = len(zerooffs)
    for i in range(0, num_elem):
        buffer += to_buf(zerooffs[i])
        buffer += to_buf(spandriftneg[i])
        buffer += to_buf(zerodriftpos[num_elem - 1 - i])
        buffer += to_buf(spancal[num_elem - 1 - i])
        buffer += to_buf(zerodriftneg[i])
        buffer += to_buf(spandriftpos[i])
    return buffer

def factory_db_xml_to_buffer(facdbxml):
    """Convert factory database XML into a buffer. This is the "security by obscurity" bit that shouldn't be published"""

    serno = int(facdbxml.find(".//database/elem[@name='serialno']").text)
    caltemp = int(facdbxml.find(".//database/elem[@name='adccaltemp']").text)

    zerooffs = factory_db_xmlarray_to_list(facdbxml=facdbxml, name="adczerooffset")
    zerodriftpos = factory_db_xmlarray_to_list(facdbxml=facdbxml, name="adczerodriftpos")
    zerodriftneg = factory_db_xmlarray_to_list(facdbxml=facdbxml, name="adczerodriftneg")

    spancal = factory_db_xmlarray_to_list(facdbxml=facdbxml, name="adcspancal")
    spandriftpos = factory_db_xmlarray_to_list(facdbxml=facdbxml, name="adcspandriftpos")
    spandriftneg = factory_db_xmlarray_to_list(facdbxml=facdbxml, name="adcspandriftneg")

    buffer = factory_values_to_buffer(serno=serno, caltemp=caltemp,
                                      zerooffs=zerooffs, zerodriftpos=zerodriftpos, zerodriftneg=zerodriftneg,
                                      spancal=spancal, spandriftpos=spandriftpos, spandriftneg=spandriftneg)
    return buffer

def factory_db_uboot_get_value(facdb_uboot_list, name):
    """Get a value for a name in the uboot factory database list"""
    filtlist = [ x for x in facdb_uboot_list if x[0] == name ]
    return filtlist[0][1]

def factory_db_uboot_get_int_value_list(facdb_uboot_list, name):
    """Get a list of integer values for a name in the uboot factory database list"""
    values = factory_db_uboot_get_value(facdb_uboot_list, name)
    values_list = values.split(",")
    values_int_list = []
    for value in values_list:
        values_int_list.append(int(value))
    return values_int_list

def factory_db_uboot_to_buffer(facdb_uboot_list):
    """Convert factory database uboot environment into a buffer. This is the "security by obscurity" bit that shouldn't be published"""

    serno = int(factory_db_uboot_get_value(facdb_uboot_list=facdb_uboot_list, name="factorydb_serialno"))
    caltemp = int(factory_db_uboot_get_value(facdb_uboot_list=facdb_uboot_list, name="factorydb_adccaltemp"))

    zerooffs = factory_db_uboot_get_int_value_list(facdb_uboot_list=facdb_uboot_list, name="factorydb_adczerooffset")
    zerodriftpos = factory_db_uboot_get_int_value_list(facdb_uboot_list=facdb_uboot_list, name="factorydb_adczerodriftpos")
    zerodriftneg = factory_db_uboot_get_int_value_list(facdb_uboot_list=facdb_uboot_list, name="factorydb_adczerodriftneg")

    spancal = factory_db_uboot_get_int_value_list(facdb_uboot_list=facdb_uboot_list, name="factorydb_adcspancal")
    spandriftpos = factory_db_uboot_get_int_value_list(facdb_uboot_list=facdb_uboot_list, name="factorydb_adcspandriftpos")
    spandriftneg = factory_db_uboot_get_int_value_list(facdb_uboot_list=facdb_uboot_list, name="factorydb_adcspandriftneg")

    buffer = factory_values_to_buffer(serno=serno, caltemp=caltemp,
                                      zerooffs=zerooffs, zerodriftpos=zerodriftpos, zerodriftneg=zerodriftneg,
                                      spancal=spancal, spandriftpos=spandriftpos, spandriftneg=spandriftneg)
    return buffer

def factory_db_buffer_to_password(buffer):
    """Convert the factory database buffer into an actual password"""
    sha = hashlib.sha256()
    sha.update(buffer)
    hash = sha.digest()

    return base64.b64encode(hash).decode('utf-8')

#def generate_root_from_xml(facdbxml):
#    buff = factory_db_xml_to_buffer(facdbxml=facdbxml)
#    password = factory_db_buffer_to_password(buffer=buff)
#    return (buff, password)

class TestGeneratePassword(unittest.TestCase):
    """Test case for generating root passwords

       Run with: python -m unittest generate_password.TestGeneratePassword
    """

    facdata ="""
--------------------------------------------------------------------------------
Start of debug data for C520 S/N 3477729

--------------------------------------------------------------------------------
Contents of "/etc/rinstrum/factory1.rdb":
<rinstrum><database name="factorydb" version="3" checksum="1203351525"><elem name="adczerooffset" numelems="9"><arrayelem index="0">21179</arrayelem><arrayelem index="1">21687</arrayelem><arrayelem index="2">25097</arrayelem><arrayelem index="3">21956</arrayelem><arrayelem index="4">22486</arrayelem><arrayelem index="5">26029</arrayelem><arrayelem index="6">22492</arrayelem><arrayelem index="7">26030</arrayelem><arrayelem index="8">22504</arrayelem></elem><elem name="adczerodriftpos" numelems="9"><arrayelem index="0">193</arrayelem><arrayelem index="1">198</arrayelem><arrayelem index="2">227</arrayelem><arrayelem index="3">207</arrayelem><arrayelem index="4">204</arrayelem><arrayelem index="5">234</arrayelem><arrayelem index="6">202</arrayelem><arrayelem index="7">243</arrayelem><arrayelem index="8">185</arrayelem></elem><elem name="adczerodriftneg" numelems="9"><arrayelem index="0">314</arrayelem><arrayelem index="1">323</arrayelem><arrayelem index="2">367</arrayelem><arrayelem index="3">314</arrayelem><arrayelem index="4">326</arrayelem><arrayelem index="5">375</arrayelem><arrayelem index="6">322</arrayelem><arrayelem index="7">369</arrayelem><arrayelem index="8">324</arrayelem></elem><elem name="adcspancal" numelems="9"><arrayelem index="0">7815304</arrayelem><arrayelem index="1">7937462</arrayelem><arrayelem index="2">6857933</arrayelem><arrayelem index="3">7637722</arrayelem><arrayelem index="4">7757062</arrayelem><arrayelem index="5">6702129</arrayelem><arrayelem index="6">7757094</arrayelem><arrayelem index="7">6702114</arrayelem><arrayelem index="8">7757133</arrayelem></elem><elem name="adcspandriftpos" numelems="9"><arrayelem index="0">51</arrayelem><arrayelem index="1">41</arrayelem><arrayelem index="2">30</arrayelem><arrayelem index="3">15</arrayelem><arrayelem index="4">19</arrayelem><arrayelem index="5">8</arrayelem><arrayelem index="6">12</arrayelem><arrayelem index="7">12</arrayelem><arrayelem index="8">-1</arrayelem></elem><elem name="adcspandriftneg" numelems="9"><arrayelem index="0">18</arrayelem><arrayelem index="1">31</arrayelem><arrayelem index="2">18</arrayelem><arrayelem index="3">-8</arrayelem><arrayelem index="4">-6</arrayelem><arrayelem index="5">-7</arrayelem><arrayelem index="6">-10</arrayelem><arrayelem index="7">-8</arrayelem><arrayelem index="8">-6</arrayelem></elem><elem name="adccaltemp">9902</elem><elem name="adcadccalflag">165</elem><elem name="adcchksum">9293</elem><elem name="serialno">3477729</elem><elem name="economy">0</elem><elem name="applicid">99</elem><elem name="tradeenable">1</elem><elem name="model">C520</elem><elem name="testresult">160</elem></database></rinstrum>
--------------------------------------------------------------------------------
Contents of "/etc/rinstrum/factory2.rdb":
<rinstrum><database name="factorydb" version="3" checksum="1203351525"><elem name="adczerooffset" numelems="9"><arrayelem index="0">21179</arrayelem><arrayelem index="1">21687</arrayelem><arrayelem index="2">25097</arrayelem><arrayelem index="3">21956</arrayelem><arrayelem index="4">22486</arrayelem><arrayelem index="5">26029</arrayelem><arrayelem index="6">22492</arrayelem><arrayelem index="7">26030</arrayelem><arrayelem index="8">22504</arrayelem></elem><elem name="adczerodriftpos" numelems="9"><arrayelem index="0">193</arrayelem><arrayelem index="1">198</arrayelem><arrayelem index="2">227</arrayelem><arrayelem index="3">207</arrayelem><arrayelem index="4">204</arrayelem><arrayelem index="5">234</arrayelem><arrayelem index="6">202</arrayelem><arrayelem index="7">243</arrayelem><arrayelem index="8">185</arrayelem></elem><elem name="adczerodriftneg" numelems="9"><arrayelem index="0">314</arrayelem><arrayelem index="1">323</arrayelem><arrayelem index="2">367</arrayelem><arrayelem index="3">314</arrayelem><arrayelem index="4">326</arrayelem><arrayelem index="5">375</arrayelem><arrayelem index="6">322</arrayelem><arrayelem index="7">369</arrayelem><arrayelem index="8">324</arrayelem></elem><elem name="adcspancal" numelems="9"><arrayelem index="0">7815304</arrayelem><arrayelem index="1">7937462</arrayelem><arrayelem index="2">6857933</arrayelem><arrayelem index="3">7637722</arrayelem><arrayelem index="4">7757062</arrayelem><arrayelem index="5">6702129</arrayelem><arrayelem index="6">7757094</arrayelem><arrayelem index="7">6702114</arrayelem><arrayelem index="8">7757133</arrayelem></elem><elem name="adcspandriftpos" numelems="9"><arrayelem index="0">51</arrayelem><arrayelem index="1">41</arrayelem><arrayelem index="2">30</arrayelem><arrayelem index="3">15</arrayelem><arrayelem index="4">19</arrayelem><arrayelem index="5">8</arrayelem><arrayelem index="6">12</arrayelem><arrayelem index="7">12</arrayelem><arrayelem index="8">-1</arrayelem></elem><elem name="adcspandriftneg" numelems="9"><arrayelem index="0">18</arrayelem><arrayelem index="1">31</arrayelem><arrayelem index="2">18</arrayelem><arrayelem index="3">-8</arrayelem><arrayelem index="4">-6</arrayelem><arrayelem index="5">-7</arrayelem><arrayelem index="6">-10</arrayelem><arrayelem index="7">-8</arrayelem><arrayelem index="8">-6</arrayelem></elem><elem name="adccaltemp">9902</elem><elem name="adcadccalflag">165</elem><elem name="adcchksum">9293</elem><elem name="serialno">3477729</elem><elem name="economy">0</elem><elem name="applicid">99</elem><elem name="tradeenable">1</elem><elem name="model">C520</elem><elem name="testresult">160</elem></database></rinstrum>
--------------------------------------------------------------------------------
Contents of "fw_printenv":
bootdelay=1
ipaddr=192.162.1.2
serverip=192.162.1.1
gatewayip=192.162.1.1
netmask=255.255.255.0
hostname=WE2111
netdev=eth0
loadaddr=40010000
bootaddr=40500000
bootcmd=fsload ${bootaddr} /boot/uImage; bootm ${bootaddr}
bootargs=root=/dev/mtdblock1 rw,noatime rootfstype=jffs2
ethaddr=70:4a:e4:00:07:6d
recoveryconsole=mcfser1
baudrate=9600
factorydb_factorydb_info=3,1203351525
factorydb_adczerooffset=21179,21687,25097,21956,22486,26029,22492,26030,22504
factorydb_adczerodriftpos=193,198,227,207,204,234,202,243,185
factorydb_adczerodriftneg=314,323,367,314,326,375,322,369,324
factorydb_adcspancal=7815304,7937462,6857933,7637722,7757062,6702129,7757094,6702114,7757133
factorydb_adcspandriftpos=51,41,30,15,19,8,12,12,-1
factorydb_adcspandriftneg=18,31,18,-8,-6,-7,-10,-8,-6
factorydb_adccaltemp=9902
factorydb_adcadccalflag=165
factorydb_adcchksum=9293
factorydb_serialno=3477729
factorydb_economy=0
factorydb_applicid=99
factorydb_tradeenable=1
factorydb_model=QzUyMA==
factorydb_testresult=160

--------------------------------------------------------------------------------
End of debug data
--------------------------------------------------------------------------------

"""

    def test_genpw_fac1(self):
        """Test generating the root password from the first copy of factory database"""
        faclines = self.facdata.splitlines()
        facdbxml = get_factory_db_xml(f=faclines, copy=1)
        buff = factory_db_xml_to_buffer(facdbxml=facdbxml)
        password = factory_db_buffer_to_password(buffer=buff)
        self.assertEqual(first="vnCFxLvZE949MAtQJkUGIy70yqXXETDo287g5rsTeVo=", second=password)

    def test_genpw_fac2(self):
        """Test generating the root password from the second copy of factory database"""
        faclines = self.facdata.splitlines()
        facdbxml = get_factory_db_xml(f=faclines, copy=2)
        buff = factory_db_xml_to_buffer(facdbxml=facdbxml)
        password = factory_db_buffer_to_password(buffer=buff)
        self.assertEqual(first="vnCFxLvZE949MAtQJkUGIy70yqXXETDo287g5rsTeVo=", second=password)

    def test_genpw_facuboot(self):
        """Test generating the root password from the u-boot copy of factory database"""
        faclines = self.facdata.splitlines()
        facdblist = get_factory_db_uboot(f=faclines)
        buff = factory_db_uboot_to_buffer(facdb_uboot_list=facdblist)
        password = factory_db_buffer_to_password(buffer=buff)
        self.assertEqual(first="vnCFxLvZE949MAtQJkUGIy70yqXXETDo287g5rsTeVo=", second=password)

def main(filename, database_copy):
    with open(filename) as f:
        if (database_copy == "facxml1"):
            facdbxml = get_factory_db_xml(f=f, copy=1)
            buff = factory_db_xml_to_buffer(facdbxml=facdbxml)
        elif (database_copy == "facxml2"):
            facdbxml = get_factory_db_xml(f=f, copy=2)
            buff = factory_db_xml_to_buffer(facdbxml=facdbxml)
        elif (database_copy == "facuboot"):
            facdblist = get_factory_db_uboot(f=f)
            buff = factory_db_uboot_to_buffer(facdb_uboot_list=facdblist)
        #xmlstr = xml.etree.ElementTree.tostring(facdbxml, encoding="utf8", method="xml")
        #print(xml.dom.minidom.parseString(xmlstr).toprettyxml())
        password = factory_db_buffer_to_password(buffer=buff)
        print(password)

if __name__ == '__main__':

    # Parse command line options
    parser = argparse.ArgumentParser(description='Generate the root password for the C500 from the debug info file.')
    parser.add_argument('filename', action='store',
                            help='The debug info file to extract the factory data from')
    parser.add_argument('--database-copy', action='store', choices=["facxml1", "facxml2", "facuboot" ], default="facxml1",
                            help='Which database copy to generate the password from (facxml1=First XML facdb file, facxml2=Second XML facdb file, facuboot=Uboot copy of factory db)')
    args = parser.parse_args()

    main(filename=args.filename, database_copy=args.database_copy)
