/**
 * @file CheckArrSize.h
 *
 * @brief Define for checking array size matches a constant at compile time
 */
#ifndef CHECKARRSIZE_DOT_H
#define CHECKARRSIZE_DOT_H 1

/*
 * NOTE : These are compile time checks only, the goal is to have checks that stop compilation
 * but produce no code if they pass.
 * The compiler will throw an error like:
 * "Illegal array size" if the test fails. Declaring the dummy variable as extern and
 * never using it results in no storage and no code used for it.
 *
 * NOTE: This approach WILL NOT WORK inside a C++ class as you can't have a class member
 * with external reference. Not an issue until we port to C++ though :-)
 */


/**
 * Check that the number of elements in an array matches a constant.
 * Compiles a zero sized array if they don't match which will cause a compiler error
 * on an ANSII compliant compiler
 * Using extern for the zero sized array results in no code/ram usage if the test
 * passes
 * USEAGE :
 *    arr =   array to check
 *    size =  the size to match
 */
#define CHECK_ARRAY_SIZE( arr, size ) \
extern unsigned char s_ui8CompileTest ## arr [ ( sizeof( arr ) / sizeof( arr[ 0 ] ) == size ) ? 1 : 0 ]


/**
 * Check that a structure size is less than or equal to a constant.
 * Compiles a zero sized array if they don't match which will cause a compiler error
 * on an ANSII compliant compiler
 * Using extern for the zero sized array results in no code/ram usage if the test
 * passes
 * USEAGE :
 *    st =    struct to check
 *    size =  the size to be less than or equal to
 */
#define CHECK_STRUCT_SIZE_LESS_EQUAL( st, size ) \
extern unsigned char s_ui8CompileTest ## st [ ( sizeof( st ) <= size ) ? 1 : 0 ]


#endif /* #ifndef CHECKARRSIZE_DOT_H */
