#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_startup.h>
#include <rinblock/rb_shutdown.h>
#include <rinblock/rb_dio.h>
#include <rinblock/rb_aout.h>
#include <rinblock/rb_aout_cal.h>
#include <rinblock/rb_c500_lcd.h>
#include <rinblock/rb_we21_lcd.h>
#include <rinblock/rb_we21_adc.h>
#include <rinblock/rb_we21_kbd.h>
#include <rinblock/rb_we21_buzzer.h>
#include <rinblock/rb_serial.h>
#include <rinblock/crc.h>
#include <rinblock/rb_debug.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "database.h"
#include "meas.h"
#include "serial_ports.h"
#include "db_setpoint.h"
#include "kbd_5000.h"
#include "adc_8bit.h"
#include "apperror.h"
#include "devices.h"
#include "iface_analog.h"
#include "we_debug.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "timer.h"

#include "access.h"


static short last_analog_zero_cal;
static short last_analog_span_cal;

static int32_t register_analogue;


/*
 * Card information
 */

typedef enum
{ ACCESS_UNKNOWN = 0, ACCESS_ANALOGUE,
  ACCESS_SETPT, ACCESS_COMBO,
  ACCESS_NONE
} eAccessCard;

typedef struct
{
  eAccessCard CardType;
  bool installed;
  uint8_t rinblock;
} AccessCard_t;


static AccessCard_t AccessCardList[] =
{
  {ACCESS_UNKNOWN, false, 0},
  {ACCESS_ANALOGUE, false, 0},
  {ACCESS_SETPT, false, 0},
  {ACCESS_COMBO, false, 0},
  {ACCESS_NONE, false, 0}
};


static bool access_init_done = false;
static uint8_t access_outputs = 0x00;
static uint8_t access_inputs = 0x00;
static int32_t access_pulses[PULSE_COUNT_NUM] = {0,0};


/** Current analog forced state */
static enum eAnalogForce force_analogue = ANALOGOUT_FORCE_NONE;


/** Set comms analogue out valeu */
void access_set_register_analogue(int32_t val)
{
  register_analogue = val;
}


/** Get comms analogue out value */
int32_t acces_get_register_analogue(void)
{
  return register_analogue;
}


/** Used to force analog out state low or high for adjustment, as per @ref enum eAnalogForce */
void access_force_analogue(enum eAnalogForce force_state)
{
  force_analogue = force_state;
}


/* Get the current forced state of the analog output */
enum eAnalogForce access_get_force_analogue(void)
{
  return force_analogue;
}


/***************************************************
access_get_outputs()
returns the current state of the Accessory Bus outputs
**************************************************/
uint8_t access_get_outputs(void)
{
  return access_outputs;
}

/***************************************************
access_set_outputs()
sets the current state of the Accessory Bus outputs
**************************************************/
void access_set_outputs(uint8_t outputs)
{
  access_outputs = outputs;
}


uint8_t access_get_inputs(void)
{
  return access_inputs;
}


static void access_set_inputs(uint8_t new_inputs)
{
  access_inputs = new_inputs;
}

/***************************************************
access_get_pulses()
returns the current pulse count of the Accessory Bus inputs
**************************************************/
int32_t access_get_pulses(enum ePulseCount pulse)
{
  return (pulse < PULSE_COUNT_NUM) ? access_pulses[pulse] : 0;
}

void access_clear_pulses(enum ePulseCount pulse)
{
  enum ePulseCount i;

  if (pulse < PULSE_COUNT_NUM)
  {
    access_pulses[pulse] = 0;
  }
  else
  {
    for (i = 0; i < PULSE_COUNT_NUM; i++)
    {
      access_pulses[i] = 0;
    }
  }
}

static void access_add_pulses(enum ePulseCount pulse, uint8_t count)
{
  if (pulse < PULSE_COUNT_NUM)
    access_pulses[pulse] += count;
}

/** From the digital inputs, return a mask of any configured remote keys that are pressed */
unsigned long access_kbd_readhikeybits(void *hw_data)
{
  unsigned long inputs;
  unsigned long keybits;
  unsigned long mask;
  size_t i;

  inputs = access_get_inputs();
  keybits = 0;

  if (remote_keys_enabled())
  {
    /* Mask off IO pins not configured as inputs */
    for (i = 0, mask = 1; i < NUM_REMOTE_KEYS; i++, mask <<= 1)
    {
      if (access_get_io_is_key(i) && ((inputs & mask) != 0))
      {
        keybits |= mask;
      }
    }
  }

  return keybits;
}


/** Configure analog out */
static void access_configure_analog(int rb)
{
  struct s_rb_aout_configure aout_cfg;

  /* For rinblock 1 is voltage output, 0 current */
  aout_cfg.ucAnalogOutputType = (ANALOG_TYPE_VOLT == db_app_user.analog_output_type) ? 1 : 0;
  if (0 != db_app_user.analog_output_clip)
  {
    aout_cfg.cMaxLimit = 100;
    aout_cfg.cMinLimit = 0;
  }
  else
  {
    aout_cfg.cMaxLimit = 127;
    aout_cfg.cMinLimit = -128;
  }
  aout_cfg.iUserZero = db_app_user.analog_zero_cal;
  aout_cfg.iUserSpan = db_app_user.analog_span_cal;
  aout_cfg.uiUpdateRate = AdcSyncFrequencies[adcSuperStructure.adcUserDatabase.syncFreqIdx];

  WE_RINBLOCK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "access_configure_analog: analog module %d output configuration: "
                                                       "output type: %u (%s), "
                                                       "user zero: %d, user span: %d, "
                                                       "max limit: %d, min limit: %d, "
                                                       "update rate: %u.%01u Hz\n",
                           rb,
                           (unsigned) aout_cfg.ucAnalogOutputType, (0 == aout_cfg.ucAnalogOutputType) ? "current" : "voltage",
                           (int) aout_cfg.iUserZero, (int) aout_cfg.iUserSpan,
                           (int) aout_cfg.cMaxLimit, (int) aout_cfg.cMinLimit,
                           (unsigned) aout_cfg.uiUpdateRate / 10, (unsigned) aout_cfg.uiUpdateRate % 10);

  /* Write the settings to the rinblock */
  if (!RinBlockAOutWriteConfig(&rinblocks[rb], &aout_cfg))
  {
    rdebug_printf(LOG_ERR, "access_configure_analog: failed to configure analog output on rinblock %d\n", rb);
  }

  /* Update variables used to determine if cal has changed */
  last_analog_zero_cal = db_app_user.analog_zero_cal;
  last_analog_span_cal = db_app_user.analog_span_cal;
}


/** Configure the analog output card with the database settings */
void access_dbconfigure_analog(void)
{
  int rb;

  if (AccessCardList[ACCESS_COMBO].installed)
  {
    rb = AccessCardList[ACCESS_COMBO].rinblock;
    access_configure_analog(rb);
  }
}


/**
 * Create devices for rinblocks
 *
 * @param headless is whether run as a headless device (ie no lcd, buzzer or keyboard)
 * @param adcdev is the adc device to create if a rinblock supporting an ADC is found. If NULL no device will be created
 * @param kbddev is the keyboard device to create if a rinblock supporting a keyboard is found. If NULL no device will be created
 * @param lcddev is the LCD device to create if a rinblock supporting a LCD is found. If NULL no device will be created
 * @param buzdev is the buzzer device to create if a rinblock supporting a buzzer is found. If NULL no device will be created
 * @param ser_rb1a_dev is the serial device to create for a serial rinblock A channel on port 1. If NULL no device will be created
 * @param ser_rb1b_dev is the serial device to create for a serial rinblock B channel on port 1. If NULL no device will be created
 * @param ser_rb2a_dev is the serial device to create for a serial rinblock A channel on port 2. If NULL no device will be created
 * @param ser_rb2b_dev is the serial device to create for a serial rinblock B channel on port 2. If NULL no device will be created
 * @param ser_frontrl_dev is the serial device to create for a rinlink serial on the display. If NULL no device will be created
 */
static void RinBlockAssignDevices(bool headless, struct t_io_dev *adcdev, struct t_io_dev *kbddev, struct t_io_dev *lcddev, struct t_io_dev *buzdev,
                                  struct t_io_dev *ser_rb1a_dev, struct t_io_dev *ser_rb1b_dev, struct t_io_dev *ser_rb2a_dev, struct t_io_dev *ser_rb2b_dev,
                                  struct t_io_dev *ser_frontrl_dev)
{
  int i;
  bool adcdone = false;
  bool lcddone = false;
  bool buzdone = false;
  bool serial1done = false;
  bool serial2done = false;
  bool serial_front_rl_done = false;
  struct t_rinblock *kbd_rbs[MAX_RINBLOCKS + 1];
  int kbd_ctr;

  for (i = 0, kbd_ctr = 0; i < MAX_RINBLOCKS; i++)
  {
    /* ADC Device */
    if (RB_IFACE_TYPE_WE21_CONN == rinblocks[i].tag.interface_type
        && !adcdone)
    {
      /* Create dev_adc_rb as a rinblock one */
      if (NULL != adcdev)
      {
        if (RIO_SUCCESS == RinBlockCreateWE21AdcDevice(adcdev, &rinblocks[i]))
        {
          WE_RINBLOCK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "RinBlockAssignDevices: Created ADC device on rinblock %d\n", i);
          adcdone = true;
        }
        else
        {
          rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create rinblock adc device\n");
        }
      }

      /* This one is a keyboard as well */
      kbd_rbs[kbd_ctr++] = &rinblocks[i];
    }

    /* Serial modules on slots 2 & 3 (labeled on the back as 1 & 2) */
    if ((RB_IFACE_TYPE_SERIAL == rinblocks[i].tag.interface_type &&
         (RB_MODEL_NUMBER_485_485 == rinblocks[i].tag.model_number ||
          RB_MODEL_NUMBER_232_232 == rinblocks[i].tag.model_number ||
          RB_MODEL_NUMBER_232_485 == rinblocks[i].tag.model_number)) ||
        RB_IFACE_TYPE_ETHERNET == rinblocks[i].tag.interface_type)
    {
      if (2 == i && !serial1done)
      {
        rdebug_printf(LOG_INFO, "RinBlockAssignDevices: Detected serial module on slot %d\n", i);
        if (NULL != ser_rb1a_dev)
        {
          if (RIO_SUCCESS != RinBlockCreateSerialDevice(ser_rb1a_dev, &rinblocks[i], rbsc_a))
            rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create rinblock slot 2 (external 1) serial a device\n");
        }
        if (NULL != ser_rb1b_dev)
        {
          if (RIO_SUCCESS != RinBlockCreateSerialDevice(ser_rb1b_dev, &rinblocks[i], rbsc_b))
            rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create rinblock slot 2 (external 1) serial b device\n");
        }
        serial1done = true;
      }

      if (3 == i && !serial2done)
      {
        rdebug_printf(LOG_INFO, "RinBlockAssignDevices: Detected serial module on slot %d\n", i);
        if (NULL != ser_rb2a_dev)
        {
          if (RIO_SUCCESS != RinBlockCreateSerialDevice(ser_rb2a_dev, &rinblocks[i], rbsc_a))
            rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create rinblock slot 3 (external 2) serial a device\n");
        }
        if (NULL != ser_rb2b_dev)
        {
          if (RIO_SUCCESS != RinBlockCreateSerialDevice(ser_rb2b_dev, &rinblocks[i], rbsc_b))
            rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create rinblock slot 3 (external 2) serial b device\n");
        }
        serial2done = true;
      }
    }

    /* WE2111 LCD Device */
    if (!headless && RB_IFACE_TYPE_WE21_DISP == rinblocks[i].tag.interface_type
        && !lcddone)
    {
      /* Create rinblock LCD device */
      if (NULL != lcddev)
      {
        if (RIO_SUCCESS == RinBlockCreateWE21LcdDeviceSegFlash(lcddev, &rinblocks[i]))
        {
          WE_RINBLOCK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "RinBlockAssignDevices: Created display device on rinblock %d\n", i);
          lcddone = true;
        }
        else
        {
          rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create WE2111 rinblock lcd device\n");
        }
      }

      /* Create rinblock buzzer device */
      if (NULL != buzdev && !buzdone)
      {
        if (RIO_SUCCESS == RinBlockCreateWE21BuzzerDevice(buzdev, &rinblocks[i]))
        {
          WE_RINBLOCK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "RinBlockAssignDevices: Created buzzer device on rinblock %d\n", i);
          buzdone = true;
        }
        else
        {
          rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create rinblock buzzer device\n");
        }
      }

      /* This one is a keyboard */
      kbd_rbs[kbd_ctr++] = &rinblocks[i];
    }

    /* C500 LCD Device */
    if (!headless && RB_IFACE_TYPE_C500_DISP == rinblocks[i].tag.interface_type
        && !lcddone)
    {
      rdebug_printf(LOG_INFO, "RinBlockAssignDevices: Detected C500 display module on slot %d\n", i);

      /* Create rinblock LCD device */
      if (NULL != lcddev)
      {
        if (RIO_SUCCESS == RinBlockCreateC500LcdDeviceSegFlash(lcddev, &rinblocks[i]))
        {
          lcddone = true;
        }
        else
        {
          rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create C500 rinblock lcd device\n");
        }
      }

      /* Create rinblock buzzer device */
      if (NULL != buzdev && !buzdone)
      {
        if (RIO_SUCCESS == RinBlockCreateWE21BuzzerDevice(buzdev, &rinblocks[i]))
        {
          buzdone = true;
        }
        else
        {
          rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create rinblock buzzer device\n");
        }
      }

      /* Create rinblock serial device */
      if (NULL != ser_frontrl_dev && !serial_front_rl_done)
      {
        if (RIO_SUCCESS == RinBlockCreateSerialDevice(ser_frontrl_dev, &rinblocks[i], rbsc_a))
          serial_front_rl_done = true;
        else
          rdebug_printf(LOG_ERR, "RinBlockAssignDevices: Failed to create front rinlink serial device on C500 display rinblock\n");
      }

      /* This one is a keyboard */
      kbd_rbs[kbd_ctr++] = &rinblocks[i];
    }
  }

  /* If we found a device that has a keyboard */
  if (!headless && kbd_ctr > 0 && NULL != kbddev)
  {
    /* Create rinblock keyboard device */
    if (RIO_SUCCESS == RinBlockCreateWE21KbdDevice(kbddev, kbd_rbs, kbd_ctr, &dev_buz_we, access_kbd_readhikeybits))
    {
      WE_RINBLOCK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "RinBlockAssignDevices: Replaced keyboard device with WE2111 rinblock keyboard on %d rinblocks\n", kbd_ctr);
    }
  }
}


/**
 * Initialise the rinblocks, and create any necessary devices
 *
 * @param headless is whether run as a headless device (ie no lcd, buzzer or keyboard)
 * @param iicdev is the i2c device to use for the rinblock bus
 * @param adcdev is the adc device to create if a rinblock supporting an ADC is found. If NULL no device will be created
 * @param kbddev is the keyboard device to create if a rinblock supporting a keyboard is found. If NULL no device will be created
 * @param lcddev is the LCD device to create if a rinblock supporting a LCD is found. If NULL no device will be created
 * @param buzdev is the buzzer device to create if a rinblock supporting a buzzer is found. If NULL no device will be created
 * @param ser_rb1a_dev is the serial device to create for a serial rinblock A channel on port 1. If NULL no device will be created
 * @param ser_rb1b_dev is the serial device to create for a serial rinblock B channel on port 1. If NULL no device will be created
 * @param ser_rb2a_dev is the serial device to create for a serial rinblock A channel on port 2. If NULL no device will be created
 * @param ser_rb2b_dev is the serial device to create for a serial rinblock B channel on port 2. If NULL no device will be created
 * @param ser_frontrl_dev is the serial device to create for a rinlink serial on the display. If NULL no device will be created
 */
void init_rinblocks(bool headless, struct t_io_dev *iicdev, struct t_io_dev *adcdev, struct t_io_dev *kbddev, struct t_io_dev *lcddev, struct t_io_dev *buzdev,
                    struct t_io_dev *ser_rb1a_dev, struct t_io_dev *ser_rb1b_dev, struct t_io_dev *ser_rb2a_dev, struct t_io_dev *ser_rb2b_dev,
                    struct t_io_dev *ser_frontrl_dev)
{
#ifdef RINBLOCK_DEBUG
  /* Rinblock library debug to be sent via rdebug_vprintf() */
  rb_set_debug_printf_fn(rdebug_vprintf);
#endif

  RinBlockInit(iicdev);

  RinBlockEnumerate();

  RinBlockStart();

  RinBlockAssignDevices(headless, adcdev, kbddev, lcddev, buzdev, ser_rb1a_dev, ser_rb1b_dev, ser_rb2a_dev, ser_rb2b_dev, ser_frontrl_dev);
}


/** Clean up rinblock devices */
void destroy_rinblocks(void)
{
  rdestroy_io(&dev_adc_rb);
  rdestroy_io(&dev_ser3_rb_rs422);
  rdestroy_io(&dev_ser3_rb_rs422b);
  rdestroy_io(&dev_ser4_rb_rs232);
  rdestroy_io(&dev_ser4_rb_rs232b);
  rdestroy_io(&dev_ser_front_rl);

  rdestroy_io(&dev_lcd_we);
  rdestroy_io(&dev_kbd_we);
  rdestroy_io(&dev_buz_we);

  /* Shut down rinblock library */
  RinBlockDestroy();
}


/** Configure an IO pin as either and output or input */
void access_configure_iopin(uint8_t pin, enum eDIOType iotype)
{
  uint8_t rb;
  bool found = false;

  /* Figure out where the IO is installed */
  if (AccessCardList[ACCESS_SETPT].installed)
  {
    rb = AccessCardList[ACCESS_SETPT].rinblock;
    found = true;
  }
  else if (AccessCardList[ACCESS_COMBO].installed)
  {
    rb = AccessCardList[ACCESS_COMBO].rinblock;
    found = true;
  }

  if (found)
  {
    RinBlockConfigureIO(&rinblocks[rb], pin, iotype, ioNormal, 0);
  }
}


/** Configure all IO from the database settings */
void access_dbconfigure_allio(void)
{
  enum eDIOType diotype;
  size_t i;

  for (i = 0; i < NUM_REMOTE_KEYS; i++)
  {
    switch (access_get_io_type(i))
    {
      default:
      case IOTypeUnused:
        diotype = (is_direct_drive_output(&setpoint_running_info, i)) ? ioOutput : ioInput;
        break;
      case IOTypeInput:
        diotype = ioInput;
        break;
      case IOTypeOutput:
        diotype = ioOutput;
        break;
    }

    /* Configure IO */
    access_configure_iopin(i, diotype);
  }
}


/** Dump the analog module calibration into syslog */
static void analog_dump_calibration(int rb, const char *rb_name)
{
  struct s_rb_aout_cal aout_cal;
  uint16_t calculated_crc;

  if (!RinBlockAOutReadFactoryCal(&rinblocks[rb], &aout_cal))
  {
    rdebug_printf(LOG_ERR, "analog_dump_calibration: read analog module %d (%s) factory calibration failed\n", rb, rb_name);
    return;
  }

  calculated_crc = CRC16_Calc((unsigned char *) &aout_cal, sizeof(aout_cal) - sizeof(aout_cal.uiCRCValue), 0);

  rdebug_printf(LOG_INFO, "analog_dump_calibration: module %d (%s) analog output factory calibration: "
                          "0v: %u, 10v: %u, 4mA: %u, 20mA: %u, "
                          "stored CRC: 0x%04X, calculated CRC: 0x%04X\n",
                rb, rb_name,
                (unsigned) aout_cal.uiFactoryCal0V, (unsigned) aout_cal.uiFactoryCal10V,
                (unsigned) aout_cal.uiFactoryCal4mA, (unsigned) aout_cal.uiFactoryCal20mA,
                (unsigned) aout_cal.uiCRCValue, (unsigned) calculated_crc);
}


/** Check that fitted modules are running, and set an error if not */
static void check_module_running(int rb, bool headless)
{
  const char *const module_names[MAX_RINBLOCKS] = { "display", "connector", "slot 1", "slot 2" };
  const struct s_rb_general_tag *tag;

  /* Don't bother running these checks on the display rinblock if we are running headless */
  if (!headless || 0 != rb)
  {
    /* Applies to all rinblocks (if present, must be programmed and running) */
    if (rinblocks[rb].present)
    {
      if (!rinblocks[rb].programmed)
        set_error(APP_ERR_MODULE, "check_module_running: setting APP_ERR_MODULE because module %d (%s) is present but not programmed\n", rb, module_names[rb]);
      else if (!rinblocks[rb].running)
        set_error(APP_ERR_MODULE, "check_module_running: setting APP_ERR_MODULE because module %d (%s) is present but not running\n", rb, module_names[rb]);
    }

    /*
     * Applies to display and connector (must be present)
     *
     * This is a bit dodgy, in that if a rinblock is not running
     * we can't determine what type it is, so we assume here
     * that rinblock 0 is the display, and
     * rinblock 1 is the connector PCB.
     */
    if ((0 == rb || 1 == rb) && !rinblocks[rb].present)
    {
      set_error(APP_ERR_MODULE, "check_module_running: setting APP_ERR_MODULE because module %d (%s) is not present\n", rb, module_names[rb]);
    }

    /* Log some info about each attached module */
    if (rinblocks[rb].running)
    {
      tag = &rinblocks[rb].tag;

      rdebug_printf(LOG_INFO, "check_module_running: module %d (%s) tag: "
                              "model: %u, serial: %lu, swvers: %u.%.02u, "
                              "ifacevers: %u, ifacetype: 0x%04x, hwvers: %u\n",
                    rb, module_names[rb],
                    (unsigned) tag->model_number, (unsigned long) tag->serial_number,
                    (unsigned) tag->software_version / 100, (unsigned) tag->software_version % 100,
                    (unsigned) tag->interface_version, (unsigned) tag->interface_type, (unsigned) tag->hardware_version);

      if (RB_IFACE_TYPE_AOUT == rinblocks[rb].tag.interface_type)
      {
        analog_dump_calibration(rb, module_names[rb]);
      }
    }
  }
}


/**
 * Initalise IO & analog rinblocks, plus do a check that all rinblocks are running correctly
 *
 * @param headless is whether we are running without a display
 */
void init_access(bool headless)
{
  uint8_t rb;

  for (rb = 0; rb < MAX_RINBLOCKS; rb++)
  {
    if (rinblocks[rb].present &&
        rinblocks[rb].programmed && rinblocks[rb].running)
    {
      /* Card running, what is it? */
      if ((rinblocks[rb].tag.interface_type == RB_IFACE_TYPE_8DIO ||
           rinblocks[rb].tag.interface_type == RB_IFACE_TYPE_16DIO ||
           rinblocks[rb].tag.interface_type == RB_IFACE_TYPE_24DIO ||
           rinblocks[rb].tag.interface_type == RB_IFACE_TYPE_32DIO) &&
          !AccessCardList[ACCESS_SETPT].installed)
      {
        AccessCardList[ACCESS_SETPT].installed = true;
        AccessCardList[ACCESS_SETPT].rinblock = rb;
      }
      else if (rinblocks[rb].tag.interface_type == RB_IFACE_TYPE_AOUT &&
               !AccessCardList[ACCESS_COMBO].installed)
      {
        AccessCardList[ACCESS_COMBO].installed = true;
        AccessCardList[ACCESS_COMBO].rinblock = rb;

        access_configure_analog(rb);
      }
    }

    /* Check that fitted modules are running, and set an error if not */
    check_module_running(rb, headless);
  }

  /* Configure any IO found on the rinblocks */
  access_dbconfigure_allio();

  /* Now initialised and ready to run */
  access_init_done = true;
}


/** Get the real (weight based) analog output value */
static long access_get_unforced_analog(void)
{
  long l;
  long low;
  long high;

  /* Get current weight reading */
  switch (db_app_user.analog_output_source)
  {
    default:
    case ANALOG_SOURCE_DISPLAY:
      l = get_user_rdg();
      break;
    case ANALOG_SOURCE_GROSS:
      l = get_user_gross();
      break;
    case ANALOG_SOURCE_NET:
      l = get_user_net();
      break;
    case ANALOG_SOURCE_COMMS:
      return register_analogue;
  }

  /* If absolute, then negate negative readings */
  if (l < 0L && 0 != db_app_user.analog_output_abs)
  {
    l = -l;
  }

  /* Scale from 0..fullscale to 0..DAC_FULLSCALE */
  low = 0;
  high = true_full_scale();
  l = AdcMultDiv(l, ANALOG_OUTPUT_DAC_FULLSCALE, high - low);
  return l;
}


/** Get the analog output reading, which may be forced */
static long access_get_analog_reading(void)
{
  long l;

  /* Figure out value to use (forced high/low or normal weight) */
  switch (force_analogue)
  {
    default:
    case ANALOGOUT_FORCE_NONE:
      l = access_get_unforced_analog();
      break;

    case ANALOGOUT_FORCE_HIGH:
      l = ANALOG_OUTPUT_DAC_FULLSCALE;
      break;

    case ANALOGOUT_FORCE_LOW:
      l = 0;
      break;
  }

  return l;
}


/** Fill out the output structure for the digital IO card */
static void access_fill_out_dio(union u_rb_output_write *out, uint8_t outputs, struct s_setpoint_running_info *spri, struct timespec *start)
{
  int ro;
  uint8_t mask;
  struct timespec tsWrite = {0,0}, tsDiff = {0,0};

  /* figure out how much time has elapsed since the reading */
  clock_gettime(CLOCK_MONOTONIC, &tsWrite);
  tsDiff = diffTime(*start, tsWrite);

  mask = 0x01;
  for (ro = 0; ro < MAX_DIO_IO; ro++)
  {
    out->rb_dio_output_write.rb_output[ro].output_bits.level = (mask == (outputs & mask));
    if (spri->setpoint_info[ro].changeTime < (tsDiff.tv_nsec / 1000000))
      out->rb_dio_output_write.rb_output[ro].output_bits.holdoff = 0;
    else
      out->rb_dio_output_write.rb_output[ro].output_bits.holdoff = spri->setpoint_info[ro].changeTime - (tsDiff.tv_nsec / 1000000);
    mask <<= 1;
  }
}


/** Fill out the output structure for the analog output card */
static void access_fill_out_aout(union u_rb_output_write *out, long aoutrdg, uint8_t outputs, struct s_setpoint_running_info *spri, struct timespec *start)
{
  int ro;
  uint8_t mask;
  struct timespec tsWrite = {0,0}, tsDiff = {0,0};

  /* Set analog output data */
  out->rb_aout_output_write.aout_value = aoutrdg;

  /* figure out how much time has elapsed since the reading */
  clock_gettime(CLOCK_MONOTONIC, &tsWrite);
  tsDiff = diffTime(*start, tsWrite);

  /* Set all outputs on all cards to count (across 2 bits) */
  mask = 0x01;
  for (ro = 0; ro < MAX_AOUT_IO; ro++)
  {
    out->rb_aout_output_write.rb_output[ro].output_bits.level = (mask == (outputs & mask));
    if (spri->setpoint_info[ro].changeTime < (tsDiff.tv_nsec / 1000000))
      out->rb_dio_output_write.rb_output[ro].output_bits.holdoff = 0;
    else
      out->rb_aout_output_write.rb_output[ro].output_bits.holdoff = spri->setpoint_info[ro].changeTime - (tsDiff.tv_nsec / 1000000);
    mask <<= 1;
  }
}


/** Write to the digital IO card, needs a corresponding call to RinBlockReadInputs() */
static bool access_write_dio(int rb, uint8_t outputs, struct s_setpoint_running_info *spri, struct timespec *start)
{
  union u_rb_output_write out;

  /* Fill out data */
  access_fill_out_dio(&out, outputs, spri, start);

  /* Write data to rinblock */
  return RinBlockWriteOutputs(&rinblocks[rb], &out);
}


/** Write to the analog output card, needs a corresponding call to RinBlockReadInputs() */
static bool access_write_aout(int rb, uint8_t outputs, struct s_setpoint_running_info *spri, struct timespec *start)
{
  union u_rb_output_write out;

  /* Fill out data */
  access_fill_out_aout(&out, access_get_analog_reading(), outputs, spri, start);

  /* Write data to rinblock */
  return RinBlockWriteOutputs(&rinblocks[rb], &out);
}


/** Read from the IO on a analog or DIO card, and save as necessary */
static bool access_read_io(int rb, eAccessCard ct, bool is_setp_present, uint8_t *inputs)
{
  bool result;
  struct s_rb_dio_read input;

  /* Write data to rinblock */
  result = RinBlockReadInputs(&rinblocks[rb], &input);
  if (result)
  {
    /* Setpoint card inputs dominate combo card inputs */
    if (ACCESS_SETPT == ct || (ACCESS_COMBO == ct && !is_setp_present))
    {
      *inputs = input.io_data;
      access_add_pulses(PULSE_COUNT_1, input.pulse_count.pulse_count_bits.pulse_count_1);
      access_add_pulses(PULSE_COUNT_2, input.pulse_count.pulse_count_bits.pulse_count_2);
    }
  }
  return result;
}


/** Write outputs and read inputs to/from IO cards */
void access_io_cards(struct s_setpoint_running_info *spri, struct timespec *start)
{
  int i;
  bool rb_written[MAX_RINBLOCKS];
  uint8_t outputs;
  uint8_t inputs;
  int rb;
  bool setp_present = false;

  if (AccessCardList[ACCESS_COMBO].installed)
  {
    /* Detect change in analog output calibration, and force re-configure of card */
    if (last_analog_zero_cal != db_app_user.analog_zero_cal ||
        last_analog_span_cal != db_app_user.analog_span_cal)
    {
      access_configure_analog(AccessCardList[ACCESS_COMBO].rinblock);
    }
  }

  memset(rb_written, false, MAX_RINBLOCKS);

  /* Get output levels */
  outputs = access_get_outputs();

  /* Write all output data first */
  for (i = 0; AccessCardList[i].CardType != ACCESS_NONE; i++)
  {
    if (AccessCardList[i].installed)
    {
      rb = AccessCardList[i].rinblock;
      switch (AccessCardList[i].CardType)
      {
        case ACCESS_SETPT:
          rb_written[rb] = access_write_dio(rb, outputs, spri, start);
          setp_present = true;
          break;

        case ACCESS_COMBO:
          rb_written[rb] = access_write_aout(rb, outputs, spri, start);
          break;

        default:
          break;
      }
    }
  }

  /* Read all input data */
  inputs = 0;
  for (i = 0; AccessCardList[i].CardType != ACCESS_NONE; i++)
  {
    if (AccessCardList[i].installed)
    {
      rb = AccessCardList[i].rinblock;
      switch (AccessCardList[i].CardType)
      {
        case ACCESS_COMBO:
        case ACCESS_SETPT:
          if (rb_written[rb])
          {
            access_read_io(rb, AccessCardList[i].CardType, setp_present, &inputs);
          }
          break;
        default:
          break;
      }
    }
  }

  /* Set inputs for everyone to access */
  access_set_inputs(inputs);
}


/** Return if IO is present on the WE2111 */
bool access_io_is_present(void)
{
  int i;
  bool has_io;

  for (i = 0, has_io = false; AccessCardList[i].CardType != ACCESS_NONE; i++)
  {
    if (AccessCardList[i].installed &&
        (ACCESS_COMBO == AccessCardList[i].CardType ||
         ACCESS_SETPT == AccessCardList[i].CardType))
    {
      has_io = true;
      break;
    }
  }
  return has_io;
}
