/*
 * app_db.c
 *
 *  Created on: Apr 17, 2012
 *      Author: ian
 */
#include <stdio.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcinterface.h>
#include "app_db_factory.h"
#include "app_db_user.h"
#include "app_db_licence.h"
#include "app_db_runtime.h"
#include "apperror.h"
#include "adcthread.h"
#include "version.h"
#include "we_debug.h"

#include "app_db.h"


/**
 * Check if the software version has changed, and if so, bump the
 * calibration counters, put an entry in the change log
 * and save the user database (containing the current software version)
 *
 * NOTE: This can only be called after the change log, user database and
 * ADC thread/adcSuperStructure have been initialised
 */
void app_db_check_software_version(void)
{

  /*
   * Check whether software version
   * has changed, and if so bump the cal counter (required by OIML)
   * and enter a record in the change log (AdcSaveDatabases() will do this for us)
   */
  if (sw_version_cnt != db_app_user.sw_ver)
  {
    db_app_user.sw_ver = sw_version_cnt;
    AdcBumpCalCounters(&adcSuperStructure, true);
    AdcSaveDatabases(&adcSuperStructure, true, NULL, &adcSuperStructure.adcUserDatabase, NULL, &db_app_user.adc_user_database);
  }
}


/** Initialise and read all the databases. Call only once on startup. */
void app_init_and_read_databases(void)
{
  bool fdb_needs_default;
  bool udb_needs_default;
  bool rdb_needs_default;
  bool ldb_needs_default;

  /* Initialise and read factory database */
  fdb_needs_default = true;
  if (app_init_factory_database())
  {
    if (app_read_factory_database(true))
    {
      fdb_needs_default = false;
    }
  }
  if (fdb_needs_default)
  {
    /* Something went wrong, set to defaults */
    app_default_factory_database();
    set_error(ADC_ERR_FACTORY_DB, "app_init_and_read_databases: setting error ADC_ERR_FACTORY_DB because factory database set to defaults\n");
  }

  /* Initialise and read user database */
  udb_needs_default = true;
  if (app_init_user_database())
  {
    if (app_read_user_database())
    {
      udb_needs_default = false;
    }
  }
  if (udb_needs_default)
  {
    /* Something went wrong, set to defaults, including calibration */
    app_default_user_database(true);
    set_error(ADC_ERR_USER_DB, "app_init_and_read_databases: setting errors ADC_ERR_USER_DB because user database set to defaults\n");
  }

  /* Initialise and read licence database */
  ldb_needs_default = true;
  if (app_init_licence_database())
  {
    if (app_read_licence_database())
    {
      ldb_needs_default = false;
    }
  }
  if (ldb_needs_default)
  {
    /*
     * Something went wrong, set to defaults.
     * If the user database is ok, then copy in any licence codes from there
     */
    if (!udb_needs_default)
    {
      app_default_licence_database(db_app_user.licence_codes_obsolete, db_app_user.num_licence_codes_obsolete);
      /*
       * We probably should set db_app_user.num_licence_codes_obsolete to 0 here,
       * but because AdcInit() hasn't been called, we can call AdcSaveDatabases() here.
      if (db_app_user.num_licence_codes_obsolete > 0)
      {
        db_app_user.num_licence_codes_obsolete = 0;
        AdcSaveDatabases(&adcSuperStructure, true, NULL, &adcSuperStructure.adcUserDatabase, NULL, &db_app_user.adc_user_database);
      }
       */
    }
    else
    {
      app_default_licence_database(NULL, 0);
    }
    app_write_licence_database();
    rdebug_printf(LOG_ERR, "app_init_and_read_databases: setting licensing database to defaults\n");
  }

  /* Initialise and read runtime database */
  rdb_needs_default = true;
  if (app_init_runtime_database())
  {
    if (app_read_runtime_database())
    {
      rdb_needs_default = false;
    }
  }
  if (rdb_needs_default)
  {
    /* Something went wrong, set to defaults, including ADC runtime variables */
    app_default_runtime_database(true);
    set_error(ADC_ERR_RUNTIME_DB, "app_init_and_read_databases: setting ADC_ERR_RUNTIME_DB because runtime database set to defaults\n");
  }

  /*
   * If factory database is ok, and indicates it isn't the default model,
   * and the factory database model doesn't match this software's model, then flag an error
   * IE Someone has put different software in the indicator
   */
  if (!fdb_needs_default)
  {
    if (db_app_factory.application_id != APPLICATION_ID_DEFAULT &&
        db_app_factory.application_id != APPLICATION_ID_WE2111)
    {
      set_error(APP_ERR_ROM, "app_init_and_read_databases: setting APP_ERR_ROM because of invalid application id\n");
    }
  }
}


/** Destroy all the databases */
void app_destroy_databases(void)
{
  app_destroy_runtime_database();
  app_destroy_user_database();
  app_destroy_runtime_database();
  app_destroy_factory_database();
}


/** Save the databases, this function is used as a callback from the ADC library */
bool app_save_databases(AdcSuperStructure *ass, const AdcFactoryDatabase *afd, const AdcUserDatabase *aud, const AdcRuntimeDatabase *ard)
{
  bool resf = true;
  bool resu = true;
  bool resr = true;

  if (NULL != afd)
  {
#if 0 /* App code should never write this */
    rdebug_printf(LOG_INFO, "Writing factory database\n");
    resf = app_write_factory_database(true, false);
#endif
    resf = false;
  }
  if (NULL != aud)
  {
    rdebug_printf(LOG_INFO, "Writing user database\n");
    resu = app_write_user_database(true, false);
  }
  if (NULL != ard)
  {
    /* rdebug_printf(LOG_INFO, "Writing runtime database\n"); */
    resr = app_write_runtime_database(true, true, false);
  }

  return (resf && resu && resr);
}
