/*
 * app_db_factory_funcs.c
 *
 *  Created on: Oct 19, 2017
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <sys/wait.h>
#include <rindb/db_element.h>
#include <rindb/db_database.h>
#include <rindb/db_database_xml.h>
#include <rindb/db_database_uboot.h>
#include <rinexec/el_runcmd.h>
#include <rinexec/el_writepipe_buffer.h>
#include <rinexec/el_readpipe_bufalloc.h>
#include <rinweigh/adcinterface.h>
#include "adcthread.h"
#include "we_config.h"
#include "app_db.h"
#include "app_db_changelog.h"
#include "we_debug.h"

#include "app_db_factory.h"


static struct s_db_database_interp app_factory_ddi_xml1;
static struct s_db_database_interp app_factory_ddi_xml2;
static struct s_db_database_interp app_factory_ddi_uboot;
static struct s_db_buffer app_factory_db_buffer_uboot;


/** Write the factory database to the u-boot environment variables using fw_setenv */
static bool app_write_factory_database_uboot(void)
{
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_buffer wcb;
  const char *const fw_setenv_cmd[] = { "fw_setenv", "-s", "-", NULL };
  int ret_code;
  bool result;
  enum eDBError res1;
  enum e_el_result res2;
  size_t len;

  /* Write factory database to u-boot environment variables */
  res1 = db_database_write(&app_factory_database, &app_factory_ddi_uboot, app_factory_database.db_database_version);
  if (dbeSuccess != res1)
  {
    rdebug_printf(LOG_ERR, "app_write_factory_database_uboot: db_database_write() uboot copy returned %d\n", res1);
    result = false;
  }
  else
  {
    /* Don't send the '\0' from the end of the string (fw_setenv doesn't like them) */
    len = strlen((const char *) app_factory_db_buffer_uboot.buffer);

    /* Configure to write uboot database output to "fw_setenv" via it's stdin */
    el_init_runcmd_pipe_config(&rcpc);
    el_init_write_cmd_pipe_buffer(&rcpc, &wcb, (const char *) app_factory_db_buffer_uboot.buffer, len);

    /* Run "fw_setenv -s -" and pipe in the contents of app_factory_db_buffer_uboot.buffer, and wait for the command to complete */
    res2 = el_run_command(fw_setenv_cmd[0], fw_setenv_cmd, &ret_code, &rcpc, true, true, true);
    if (elr_success != res2)
    {
      rdebug_printf(LOG_ERR, "app_write_factory_database_uboot: el_run_command(fw_setenv) returned %d\n", res2);
      result = false;
    }
    else if (!WIFEXITED(ret_code) || 0 != WEXITSTATUS(ret_code))
    {
      rdebug_printf(LOG_ERR, "app_write_factory_database_uboot: fw_setenv failed, or return code not 0\n");
      result = false;
    }
    else
    {
      result = true;
    }
  }
  return result;
}


/**
 * Write the factory database to file
 *
 * @param write_ass_copy specifies whether the adcSuperStructure copy of the adcFactoryDatabase should be written
 * @param lock specifies whether adcSuperStructure should be locked before writing
 */
bool app_write_factory_database(bool write_ass_copy, bool lock)
{
  enum eDBError res1;
  enum eDBError res2;
  bool res3;

  if (write_ass_copy)
  {
    /* Write copy from adcSuperStructure */
    if (lock)
      AdcMutexLock(&adcSuperStructure);
    db_app_factory.adc_factory_database = adcSuperStructure.adcFactoryDatabase;
    if (lock)
      AdcMutexUnlock(&adcSuperStructure);
  }

  /* Write factory database to first file */
  res1 = db_database_write(&app_factory_database, &app_factory_ddi_xml1, app_factory_database.db_database_version);
  if (dbeSuccess != res1)
  {
    rdebug_printf(LOG_ERR, "app_write_factory_database: db_database_write() first copy returned %d\n", res1);
  }

  /* Write factory database to second file */
  res2 = db_database_write(&app_factory_database, &app_factory_ddi_xml2, app_factory_database.db_database_version);
  if (dbeSuccess != res2)
  {
    rdebug_printf(LOG_ERR, "app_write_factory_database: db_database_write() second copy returned %d\n", res2);
  }

  /* Write the factory database to u-boot environment variables using fw_setenv */
  res3 = app_write_factory_database_uboot();

  /* Update the change log */
  (void) app_changelog_update(&app_factory_database);

  return (dbeSuccess == res1 && dbeSuccess == res2 && res3);
}


/** Read the factory database from the u-boot environment variables using fw_printenv */
static enum eDBError app_read_factory_database_uboot(void)
{
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_bufferalloc rcba;
  const char *const fw_printenv_cmd[] = { "fw_printenv", NULL };
  int ret_code;
  enum e_el_result res1;
  enum eDBError res2;
  enum eDBError result;
  struct s_db_buffer db_buf_uboot;
  struct s_db_database_interp ddi_uboot;
  uint32_t version;

  /* Configure to read uboot database input from "fw_printenv" via it's stdout */
  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_bufferalloc(&rcpc, &rcba, 1024, erp_stdout);

  /* Run "fw_printenv" and read out the results to rcba (grows as necessary) */
  res1 = el_run_command(fw_printenv_cmd[0], fw_printenv_cmd, &ret_code, &rcpc, true, true, true);
  if (elr_success != res1)
  {
    rdebug_printf(LOG_ERR, "app_read_factory_database_uboot: el_run_command(fw_printenv) returned %d\n", res1);
    result = dbeFailure;
  }
  else if (!WIFEXITED(ret_code) || 0 != WEXITSTATUS(ret_code))
  {
    rdebug_printf(LOG_ERR, "app_read_factory_database_uboot: fw_printenv failed, or return code not 0\n");
    result = dbeFailure;
  }
  else
  {
    /* Success, now interpret the data with u-boot database interpreter */
    res2 = db_buffer_create_static(&db_buf_uboot, (uint8_t *) rcba.data, rcba.data_posn);
    if (dbeSuccess != res2)
    {
      rdebug_printf(LOG_ERR, "app_read_factory_database_uboot: db_buffer_create_static returned %d\n", res2);
      result = res2;
    }
    else
    {
      res2 = db_create_database_interp_uboot(&ddi_uboot, &db_buf_uboot);
      if (dbeSuccess != res2)
      {
        rdebug_printf(LOG_ERR, "app_read_factory_database_uboot: db_create_database_interp_uboot returned %d\n", res2);
        result = res2;
      }
      else
      {
        /* Read database from file */
        res2 = db_database_read(&app_factory_database, &ddi_uboot, &version, false);
        if (dbeSuccess != res2)
        {
          rdebug_printf(LOG_ERR, "app_read_factory_database_uboot: db_database_read returned %d\n", res2);
          result = res2;
        }
        else
        {
          result = dbeSuccess;
        }

        /* Clean up */
        db_destroy_database_interp_uboot(&ddi_uboot);
      }
      db_buffer_destroy(&db_buf_uboot);
    }
  }
  el_free_read_cmd_pipe_bufferalloc(&rcba);

  return result;
}


/**
 * Read the factory database from file
 *
 * @param recreate will cause any corrupt/missing copies of the database to be re-created, only if an intact database has been read
 */
bool app_read_factory_database(bool recreate)
{
  enum eDBError res_xml1_read;
  enum eDBError res_xml2_read;
  enum eDBError res_uboot_read;
  enum eDBError res_xml1_write;
  enum eDBError res_xml2_write;
  enum eDBError res_uboot_write;
  enum eDBError res_read = dbeFailure;
  uint32_t version;

  /* First read all databases and note which ones are corrupt */
  res_xml1_read = db_database_read(&app_factory_database, &app_factory_ddi_xml1, &version, false);
  if (dbeSuccess != res_xml1_read)
    rdebug_printf(LOG_ERR, "app_read_factory_database: db_database_read() first copy returned %d\n", res_xml1_read);
  res_xml2_read = db_database_read(&app_factory_database, &app_factory_ddi_xml2, &version, false);
  if (dbeSuccess != res_xml2_read)
    rdebug_printf(LOG_ERR, "app_read_factory_database: db_database_read() second copy returned %d\n", res_xml2_read);
  res_uboot_read = app_read_factory_database_uboot();
  if (dbeSuccess != res_uboot_read)
    rdebug_printf(LOG_ERR, "app_read_factory_database: db_database_read() uboot copy returned %d\n", res_uboot_read);

  /* Now read the first good database (order: xml1, xml2, uboot) */
  if (dbeSuccess == res_xml1_read)
  {
    res_read = db_database_read(&app_factory_database, &app_factory_ddi_xml1, &version, false);
  }
  else if (dbeSuccess == res_xml2_read)
  {
    res_read = db_database_read(&app_factory_database, &app_factory_ddi_xml2, &version, false);
  }
  else if (dbeSuccess == res_uboot_read)
  {
    res_read = app_read_factory_database_uboot();
  }

  /* If we have at least one good copy of the database, then write it to the bad ones if requested */
  if (recreate && dbeSuccess == res_read)
  {
    if (dbeSuccess != res_xml1_read)
    {
      /* Write factory database to corrupt/missing first file */
      rdebug_printf(LOG_WARNING, "app_read_factory_database: recreating first copy of factory database\n");
      res_xml1_write = db_database_write(&app_factory_database, &app_factory_ddi_xml1, app_factory_database.db_database_version);
      if (dbeSuccess != res_xml1_write)
      {
        rdebug_printf(LOG_ERR, "app_read_factory_database: db_database_write() first copy returned %d\n", res_xml1_write);
      }
    }

    if (dbeSuccess != res_xml2_read)
    {
      /* Write factory database to corrupt/missing second file */
      rdebug_printf(LOG_WARNING, "app_read_factory_database: recreating second copy of factory database\n");
      res_xml2_write = db_database_write(&app_factory_database, &app_factory_ddi_xml2, app_factory_database.db_database_version);
      if (dbeSuccess != res_xml2_write)
      {
        rdebug_printf(LOG_ERR, "app_read_factory_database: db_database_write() second copy returned %d\n", res_xml2_write);
      }
    }

    if (dbeSuccess != res_uboot_read)
    {
      /* Write the factory database to corrupt/missing u-boot environment variables using fw_setenv */
      rdebug_printf(LOG_WARNING, "app_read_factory_database: recreating uboot copy of factory database\n");
      res_uboot_write = app_write_factory_database_uboot();
      if (dbeSuccess != res_uboot_write)
      {
        rdebug_printf(LOG_ERR, "app_read_factory_database: db_database_write() uboot copy returned %d\n", res_uboot_write);
      }
    }
  }

  /* Update previous data, so change log will work */
  db_database_update_previous(&app_factory_database);

  return (dbeSuccess == res_read);
}


/** Default the factory database */
bool app_default_factory_database(void)
{
  enum eDBError res;

  res = db_database_default(&app_factory_database);
  if (dbeSuccess != res)
  {
    rdebug_printf(LOG_ERR, "app_default_factory_database: db_database_default() returned %d\n", res);
  }

  /* Use the ADC database defaulter for ADC values */
  AdcDefaultDatabases(&adcSuperStructure, true, &db_app_factory.adc_factory_database, NULL, NULL, false, false);

  return (dbeSuccess == res);
}


/** Initialise the factory database */
bool app_init_factory_database(void)
{
  enum eDBError res1;
  enum eDBError res2;
  enum eDBError res3;
  enum eDBError res4;
  enum eDBError res5;

  /* Initialise the database (creates previous data for some elements) */
  res1 = db_database_init(&app_factory_database);
  if (dbeSuccess != res1)
  {
    rdebug_printf(LOG_ERR, "app_init_factory_database: db_database_init() returned %d\n", res1);
  }

  /* Create database interpreters */
  res2 = db_create_database_interp_xml(&app_factory_ddi_xml1, we_configuration.factorydb1_file);
  if (dbeSuccess != res2)
  {
    rdebug_printf(LOG_ERR, "app_init_factory_database: db_create_database_interp_xml() for first copy returned %d\n", res2);
  }
  res3 = db_create_database_interp_xml(&app_factory_ddi_xml2, we_configuration.factorydb2_file);
  if (dbeSuccess != res3)
  {
    rdebug_printf(LOG_ERR, "app_init_factory_database: db_create_database_interp_xml() for second copy returned %d\n", res3);
  }

  /* Create u-boot environment variable database interpreters */
  res4 = db_buffer_create_malloc(&app_factory_db_buffer_uboot, 1024, 1024);
  if (dbeSuccess != res4)
  {
    rdebug_printf(LOG_ERR, "app_init_factory_database: db_buffer_create_malloc() for u-boot copy returned %d\n", res4);
    res5 = dbeFailure;
  }
  else
  {
    res5 = db_create_database_interp_uboot(&app_factory_ddi_uboot, &app_factory_db_buffer_uboot);
    if (dbeSuccess != res5)
    {
      rdebug_printf(LOG_ERR, "app_init_factory_database: db_create_database_interp_uboot() for u-boot copy returned %d\n", res5);
    }
  }

  return (dbeSuccess == res1 && dbeSuccess == res2 && dbeSuccess == res3 && dbeSuccess == res4 && dbeSuccess == res5);
}


/** Destroy the factory database */
bool app_destroy_factory_database(void)
{
  (void) db_database_destroy(&app_factory_database);
  (void) db_destroy_database_interp_xml(&app_factory_ddi_xml1);
  (void) db_destroy_database_interp_xml(&app_factory_ddi_xml2);
  (void) db_destroy_database_interp_uboot(&app_factory_ddi_uboot);
  (void) db_buffer_destroy(&app_factory_db_buffer_uboot);
  return true;
}


/** Default Indicator Model Number */
void app_factory_default_model_no(void)
{
  strncpy(db_app_factory.model_str, APP_DB_F_MODEL_DEFAULT, APP_DB_F_MODEL_MAXLEN);
}
