/*
 * app_db_licence.c
 *
 *  Created on: 10/07/2017
 *      Author: ian
 */
#include <syslog.h>
#include <stdbool.h>
#include <rindb/db_element.h>
#include <rindb/db_database.h>
#include <rindb/db_database_xml.h>
#include <rindb/db_error.h>
#include "we_config.h"
#include "app_db.h"
#include "we_debug.h"

#include "app_db_licence.h"


/** Licence application database */
struct s_db_app_licence db_app_licence;


struct s_db_element app_licence_elements[] =
{
  /* Element type                                                Element database name            Element menu name               Element data                                     Number of elements             Element minimum               Element maximum               Element default                   First appeared in      Last appeared in       Change log mask      Change log lookup */
  /* Licensing settings */
  DB_DECLARE_UINT32_ARRAY_ELEM_SINGLE(                           APP_DB_L_LICENCECODES_STR,       NULL,                           db_app_licence.licence_codes,                    APP_DB_L_LICENCECODES_NUM,     APP_DB_L_LICENCECODES_MIN,    APP_DB_L_LICENCECODES_MAX,    APP_DB_L_LICENCECODES_DEFAULT,    DB_NO_MINIMUM_VERSION, DB_NO_MAXIMUM_VERSION, APP_DB_CLMASK_LICENCE),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_L_LICENCECODENUM_STR,     NULL,                           &db_app_licence.num_licence_codes,                                              APP_DB_L_LICENCECODENUM_MIN,  APP_DB_L_LICENCECODENUM_MAX,  APP_DB_L_LICENCECODENUM_DEFAULT,  DB_NO_MINIMUM_VERSION, DB_NO_MAXIMUM_VERSION, APP_DB_CLMASK_LICENCE),
};


struct s_db_database app_licence_database =
{
  .db_database_version = APP_DB_L_VERSION1,
  .db_database_chksum_seed = APP_DB_L_CHECKSUM_SEED,
  .db_elements = app_licence_elements,
  .db_num_elements = DB_NUM_ELEMENTS(app_licence_elements),
  .db_name = APP_DB_L_NAME,
  .db_checksum_buffer = NULL
};


static struct s_db_database_interp app_licence_ddi_xml1;
static struct s_db_database_interp app_licence_ddi_xml2;


/** Write the licence database to file */
bool app_write_licence_database(void)
{
  enum eDBError res1;
  enum eDBError res2;

  /* Write licence database to first file */
  res1 = db_database_write(&app_licence_database, &app_licence_ddi_xml1, app_licence_database.db_database_version);
  if (dbeSuccess != res1)
  {
    rdebug_printf(LOG_ERR, "app_write_licence_database: db_database_write() first copy returned %d\n", res1);
  }

  /* Write licence database to second file */
  res2 = db_database_write(&app_licence_database, &app_licence_ddi_xml2, app_licence_database.db_database_version);
  if (dbeSuccess != res2)
  {
    rdebug_printf(LOG_ERR, "app_write_licence_database: db_database_write() second copy returned %d\n", res2);
  }
  return (dbeSuccess == res1 && dbeSuccess == res2);
}


/** Read the licence database from file */
bool app_read_licence_database(void)
{
  enum eDBError res;
  uint32_t version;

  /* Read the database from file */
  res = db_database_read(&app_licence_database, &app_licence_ddi_xml1, &version, false);
  if (dbeSuccess != res)
  {
    rdebug_printf(LOG_ERR, "app_read_licence_database: db_database_read() first copy returned %d, trying second copy\n", res);
    res = db_database_read(&app_licence_database, &app_licence_ddi_xml2, &version, false);
    if (dbeSuccess != res)
    {
      rdebug_printf(LOG_ERR, "app_read_licence_database: db_database_read() second copy returned %d\n", res);
    }
  }
  return (dbeSuccess == res);
}


/**
 * Default the licence database.
 *
 * @param old_userdb_passcodes if not NULL, these codes will be copied into the licence database
 * @param old_userdb_num_passcodes if not 0, this number of codes in old_userdb_passcodes will be copied into the licence database
 *
 * @return true on success
 */
bool app_default_licence_database(const uint32_t *old_userdb_passcodes, uint32_t old_userdb_num_passcodes)
{
  enum eDBError res;
  uint32_t min_codes;
  uint32_t i;

  /* Default the whole database */
  res = db_database_default(&app_licence_database);
  if (dbeSuccess != res)
  {
    rdebug_printf(LOG_ERR, "app_default_licence_database: db_database_default() returned %d\n", res);
  }

  /* Since we had to default the licence database, copy in any codes from the old licence codes in the user database */
  if (NULL != old_userdb_passcodes && old_userdb_num_passcodes > 0)
  {
    min_codes = (old_userdb_num_passcodes < APP_DB_L_LICENCECODES_NUM) ? old_userdb_num_passcodes : APP_DB_L_LICENCECODES_NUM;
    for (i = 0; i < min_codes; i++)
      db_app_licence.licence_codes[i] = old_userdb_passcodes[i];
    db_app_licence.num_licence_codes = min_codes;
    rdebug_printf(LOG_ERR, "app_default_licence_database: copying %lu licence codes from old location in user database\n", (unsigned long) min_codes);
  }

  return (dbeSuccess == res);
}


/** Initialise the licence database */
bool app_init_licence_database(void)
{
  enum eDBError res1;
  enum eDBError res2;
  enum eDBError res3;

  /* Initialise the database (creates previous data for some elements) */
  res1 = db_database_init(&app_licence_database);
  if (dbeSuccess != res1)
  {
    rdebug_printf(LOG_ERR, "app_init_licence_database: db_database_init() returned %d\n", res1);
  }

  /* Create database interpreters */
  res2 = db_create_database_interp_xml(&app_licence_ddi_xml1, we_configuration.licencedb1_file);
  if (dbeSuccess != res2)
  {
    rdebug_printf(LOG_ERR, "app_init_licence_database: db_create_database_interp_xml() for first copy returned %d\n", res2);
  }
  res3 = db_create_database_interp_xml(&app_licence_ddi_xml2, we_configuration.licencedb2_file);
  if (dbeSuccess != res3)
  {
    rdebug_printf(LOG_ERR, "app_init_licence_database: db_create_database_interp_xml() for second copy returned %d\n", res3);
  }

  return (dbeSuccess == res1 && dbeSuccess == res2 && dbeSuccess == res3);
}


/** Destroy the licence database */
bool app_destroy_licence_database(void)
{
  (void) db_database_destroy(&app_licence_database);
  (void) db_destroy_database_interp_xml(&app_licence_ddi_xml1);
  (void) db_destroy_database_interp_xml(&app_licence_ddi_xml2);
  return true;
}
