/*
 * app_db_runtime_vars.c
 *
 *  Created on: Apr 19, 2012
 *      Author: ian
 */
#include <string.h>
#include "app_db_runtime.h"
#include "database.h"

#include "app_db_runtime_vars.h"


/** Holds the last memory plus action reading for current total */
static long last_mem_plus = 0L;


/** Set true if last_mem_plus valid */
static bool last_mem_valid = false;


/** Holds the last memory plus action gross value */
static bool last_mem_gross = false;


/** The type of reading to total */
enum eTotalType { ttDisplay, ttGross, ttNet };


/*
 * Setpoint data functions
 */


enum eRecipeTargetType { rttTarget, rttPreflight, rttHysteresis };


static bool setpoint_internal_target_set(enum eRecipeTargetType target_type, uint8_t num, int32_t data)
{
  int32_t *dataptr = NULL;
  bool success = false;

  /* Get exclusive access */
  app_runtime_lock();

  /* Get pointer to data we want to change */
  switch (target_type)
  {
    case rttTarget:
      dataptr = &db_app_runtime.targets[num];
      break;
    case rttPreflight:
      dataptr = &db_app_runtime.preflight[num];
      break;
    case rttHysteresis:
      dataptr = &db_app_runtime.hys[num];
      break;
  }

  /* If valid target */
  if (NULL != dataptr && num < NUM_SETPOINTS)
  {
    /* And data has changed */
    if (*dataptr != data)
    {
      /* Alter the data in the db_app_runtime structure, and write to NVRAM */
      *dataptr = data;
      success = app_write_runtime_database(true, false, true);
    }
  }

  /* Release the nvram */
  app_runtime_unlock();
  return success;
}


static int32_t setpoint_internal_target_get(enum eRecipeTargetType target_type, uint8_t num)
{
  int32_t data = 0L;

  /* Get exclusive access */
  app_runtime_lock();

  if (num < NUM_SETPOINTS)
  {
    switch (target_type)
    {
      case rttTarget:
        data = db_app_runtime.targets[num];
        break;
      case rttPreflight:
        data = db_app_runtime.preflight[num];
        break;
      case rttHysteresis:
        data = db_app_runtime.hys[num];
        break;
    }
  }

  /* Release the nvram */
  app_runtime_unlock();
  return data;
}


/** Set the hysteresis for a recipe target */
bool setpoint_hysteresis_set(uint8_t num, int32_t data)
{
  return setpoint_internal_target_set(rttHysteresis, num, data);
}


/** Get the hysteresis for a recipe target */
int32_t setpoint_hysteresis_get(uint8_t num)
{
  return setpoint_internal_target_get(rttHysteresis, num);
}


/** Set the preflight for a recipe target */
bool setpoint_preflight_set(uint8_t num, int32_t data)
{
  return setpoint_internal_target_set(rttPreflight, num, data);
}


/** Get the preflight for a recipe target */
int32_t setpoint_preflight_get(uint8_t num)
{
  return setpoint_internal_target_get(rttPreflight, num);
}


/** Set the target for a recipe target */
bool setpoint_target_set(uint8_t num, int32_t data)
{
  return setpoint_internal_target_set(rttTarget, num, data);
}


/** Get the target for a recipe target */
int32_t setpoint_target_get(uint8_t num)
{
  return setpoint_internal_target_get(rttTarget, num);
}


/** Clear the recipe statistics */
static void clear_recipe_statistics(void)
{
  db_app_runtime.stat_total = 0;
  db_app_runtime.stat_num_items = 0;
  db_app_runtime.stat_gross_total = 0;
  db_app_runtime.stat_net_total = 0;
}


/** Return total in final units for current recipe total */
int32_t statistics_internal_get_total(enum eTotalType total_type)
{
  long total;

  /* Get exclusive access */
  app_runtime_lock();

  switch (total_type)
  {
    case ttGross:
      total = db_app_runtime.stat_gross_total;
      break;

    case ttNet:
      total = db_app_runtime.stat_net_total;
      break;

    case ttDisplay:
    default:
      total = db_app_runtime.stat_total;
      break;
  }

  /* Release the NVRAM */
  app_runtime_unlock();
  return total;
}


/** Return total in final units for current recipe total */
int32_t statistics_get_total(void)
{
  return statistics_internal_get_total(ttDisplay);
}


/** Return total in final units for gross weight total */
int32_t statistics_get_gross_total(void)
{
  return statistics_internal_get_total(ttGross);
}


/** Return total in final units for net weight total */
int32_t statistics_get_net_total(void)
{
  return statistics_internal_get_total(ttNet);
}


/** Returns number of items in the current recipe total */
uint32_t statistics_get_num_items(void)
{
  uint32_t items;

  /* Get exclusive access */
  app_runtime_lock();

  items = db_app_runtime.stat_num_items;

  /* Release the NVRAM */
  app_runtime_unlock();
  return items;
}


/** Clear total sum for current recipe total */
void statistics_clear(void)
{
  /* Get exclusive access */
  app_runtime_lock();

  /* Clear statistics and save */
  clear_recipe_statistics();
  app_write_runtime_database(true, false, true);

  /* Last reading is invalid */
  last_mem_valid = false;

  /* Release the NVRAM */
  app_runtime_unlock();
}


/**
 * Memory minus/undo action for totalising for current recipe total.
 *
 * @return true if something was undone
 */
bool statistics_undo_mem_plus(void)
{
  bool result;

  /* Get exclusive access */
  app_runtime_lock();

  if (!last_mem_valid)
  {
    result = false;
  }
  else
  {
    db_app_runtime.stat_total -= last_mem_plus;
    db_app_runtime.stat_num_items--;
    if (last_mem_gross)
      db_app_runtime.stat_gross_total -= last_mem_plus;
    else
      db_app_runtime.stat_net_total -= last_mem_plus;
    app_write_runtime_database(true, false, true);

    last_mem_valid = false;
    result = true;
  }

  /* Release the NVRAM */
  app_runtime_unlock();
  return result;
}


/**
 * Memory plus action for totalising for current recipe total
 *
 * @param rdg is in final units
 * @param gross is boolean value for if the weight is gross
 */
void statistics_mem_plus(long rdg, bool gross)
{
  /* Get exclusive access */
  app_runtime_lock();

  /* Update stats */
  db_app_runtime.stat_total += rdg;
  db_app_runtime.stat_num_items++;
  if (gross)
    db_app_runtime.stat_gross_total += rdg;
  else
    db_app_runtime.stat_net_total += rdg;
  app_write_runtime_database(true, false, true);

  /* Keep last reading, in case of undo */
  last_mem_plus = rdg;
  last_mem_valid = true;
  last_mem_gross = gross;

  /* Release the NVRAM */
  app_runtime_unlock();
}


/*
 * Text data functions
 */


/** Set the text in the NVRAM */
void text_set_text(size_t num, const char *str)
{
  if (num >= APP_DB_PRINT_TXT_NUM)
  {
    return;
  }

  /* Get exclusive access */
  app_runtime_lock();

  /* Copy text to runtime database copy */
  strncpy(db_app_runtime.print_text[num], str, APP_DB_PRINT_TXT_LEN);
  db_app_runtime.print_text[num][APP_DB_PRINT_TXT_LEN - 1] = '\0';

  /* Save the text data to NVRAM */
  app_write_runtime_database(true, false, true);

  /* Release the NVRAM */
  app_runtime_unlock();
}


/** Get the text in the NVRAM */
void text_get_text(char *str, size_t maxlen, size_t num)
{
  if (num >= APP_DB_PRINT_TXT_NUM)
  {
    strncpy(str, "", maxlen);
    return;
  }

  /* Get exclusive access */
  app_runtime_lock();

  /* Copy text to NVRAM copy */
  strncpy(str, db_app_runtime.print_text[num], maxlen);
  str[maxlen - 1] = '\0';

  /* Release the NVRAM */
  app_runtime_unlock();
}
