/*
 * app_db_user.c
 *
 *  Created on: Apr 13, 2012
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <syslog.h>
#include <rindb/db_element.h>
#include <rindb/db_database.h>
#include <rindb/db_database_xml.h>
#include <rindb/db_error.h>
#include <rinweigh/adcinterface.h>
#include "adcthread.h"
#include "we_config.h"
#include "app_db.h"
#include "app_db_changelog.h"
#include "db_option.h"
#include "db_serial.h"
#include "db_serial_ip.h"
#include "iface_sd.h"
#include "iface.h"
#include "iface_cal.h"
#include "iface_build.h"
#include "iface_special.h"
#include "iface_clock.h"
#include "commonstrings.h"
#include "we_debug.h"

#include "app_db_user.h"


/* These are obsolete, so we don't want people using them. See new licence database */
#define APP_DB_U_LICENCECODES_STR "licencecodes"
#define APP_DB_U_LICENCECODES_MIN 0
#define APP_DB_U_LICENCECODES_MAX (UINT32_MAX)
#define APP_DB_U_LICENCECODES_DEFAULT 0

#define APP_DB_U_LICENCECODENUM_STR "numlicencecodes"
#define APP_DB_U_LICENCECODENUM_MIN 0
#define APP_DB_U_LICENCECODENUM_MAX APP_DB_U_LICENCECODES_OBSOLETE_NUM
#define APP_DB_U_LICENCECODENUM_DEFAULT 0


/** At version 6 of the database we increased the number of serial settings to account for the DB9 serial ports on the C520-DB9 */
#define OLD_NUM_DB_SERIAL_SETTINGS 8

static uint8_t old_serial_type[OLD_NUM_DB_SERIAL_SETTINGS];          /*!< Old function of each port (autoout, printer, network, etc) */
static uint8_t old_serial_baud_idx[OLD_NUM_DB_SERIAL_SETTINGS];      /*!< Old baud rate index for each port */
static uint8_t old_serial_bit_settings[OLD_NUM_DB_SERIAL_SETTINGS];  /*!< Old bit packed parity, stop, data bits for each port */


/** Resolution string constants, must match @ref enum eAdcResolutions */
static const char *const adc_resolution_strs[ADC_NUM_RESOLUTION] = ADC_RESOLUTION_STRS;


/** ADC units constants, must match @ref enum eAdcUnits */
static const char *const adc_units_strs[ADC_NUM_UNITS] = ADC_UNITS_STRS;


static const uint8_t adc_linpts_minimums[APP_DB_U_ADCLINPTS_NUM] =
{
  APP_DB_U_ADCLINPTS_MIN, APP_DB_U_ADCLINPTS_MIN, APP_DB_U_ADCLINPTS_MIN, APP_DB_U_ADCLINPTS_MIN, APP_DB_U_ADCLINPTS_MIN,
  APP_DB_U_ADCLINPTS_MIN, APP_DB_U_ADCLINPTS_MIN, APP_DB_U_ADCLINPTS_MIN, APP_DB_U_ADCLINPTS_MIN, APP_DB_U_ADCLINPTS_MIN,
  APP_DB_U_ADCLINPTS_MIN
};
static const uint8_t adc_linpts_maximums[APP_DB_U_ADCLINPTS_NUM] =
{
  APP_DB_U_ADCLINPTS_MAX, APP_DB_U_ADCLINPTS_MAX, APP_DB_U_ADCLINPTS_MAX, APP_DB_U_ADCLINPTS_MAX, APP_DB_U_ADCLINPTS_MAX,
  APP_DB_U_ADCLINPTS_MAX, APP_DB_U_ADCLINPTS_MAX, APP_DB_U_ADCLINPTS_MAX, APP_DB_U_ADCLINPTS_MAX, APP_DB_U_ADCLINPTS_MAX,
  APP_DB_U_ADCLINPTS_MAX
};
static const uint8_t adc_linpts_defaults[APP_DB_U_ADCLINPTS_NUM] =
{
  APP_DB_U_ADCLINPTS_DEFAULT, APP_DB_U_ADCLINPTS_DEFAULT, APP_DB_U_ADCLINPTS_DEFAULT, APP_DB_U_ADCLINPTS_DEFAULT, APP_DB_U_ADCLINPTS_DEFAULT,
  APP_DB_U_ADCLINPTS_DEFAULT, APP_DB_U_ADCLINPTS_DEFAULT, APP_DB_U_ADCLINPTS_DEFAULT, APP_DB_U_ADCLINPTS_DEFAULT, APP_DB_U_ADCLINPTS_DEFAULT,
  APP_DB_U_ADCLINPTSLAST_DEFAULT
};

static const int32_t adc_fs_minimums[APP_DB_U_ADCFS_NUM] = { APP_DB_U_ADCFS0_MIN, APP_DB_U_ADCFS1_MIN, APP_DB_U_ADCFS2_MIN };
static const int32_t adc_fs_maximums[APP_DB_U_ADCFS_NUM] = { APP_DB_U_ADCFS0_MAX, APP_DB_U_ADCFS1_MAX, APP_DB_U_ADCFS2_MAX };
static const int32_t adc_fs_defaults[APP_DB_U_ADCFS_NUM] = { APP_DB_U_ADCFS0_DEFAULT, APP_DB_U_ADCFS0_DEFAULT, APP_DB_U_ADCFS0_DEFAULT };
static const uint8_t adc_res_minimums[APP_DB_U_ADCRES_NUM] = { APP_DB_U_ADCRES0_MIN, APP_DB_U_ADCRES1_MIN, APP_DB_U_ADCRES2_MIN };
static const uint8_t adc_res_maximums[APP_DB_U_ADCRES_NUM] = { APP_DB_U_ADCRES0_MAX, APP_DB_U_ADCRES1_MAX, APP_DB_U_ADCRES2_MAX };
static const uint8_t adc_res_defaults[APP_DB_U_ADCRES_NUM] = { APP_DB_U_ADCRES0_DEFAULT, APP_DB_U_ADCRES1_DEFAULT, APP_DB_U_ADCRES2_DEFAULT };


static const uint8_t frntkey_f_minimums[NUM_FRONT_FUNCTION_KEYS] = { APP_DB_U_FRNTKEYFN_MIN, APP_DB_U_FRNTKEYFN_MIN, APP_DB_U_FRNTKEYFN_MIN };
static const uint8_t frntkey_f_maximums[NUM_FRONT_FUNCTION_KEYS] = { APP_DB_U_FRNTKEYFN_MAX, APP_DB_U_FRNTKEYFN_MAX, APP_DB_U_FRNTKEYFN_MAX };

static const uint8_t remote_f_minimums[NUM_REMOTE_KEYS] = { APP_DB_U_REMKEYFN_MIN, APP_DB_U_REMKEYFN_MIN, APP_DB_U_REMKEYFN_MIN, APP_DB_U_REMKEYFN_MIN, APP_DB_U_REMKEYFN_MIN, APP_DB_U_REMKEYFN_MIN, APP_DB_U_REMKEYFN_MIN, APP_DB_U_REMKEYFN_MIN };
static const uint8_t remote_f_maximums[NUM_REMOTE_KEYS] = { APP_DB_U_REMKEYFN_MAX, APP_DB_U_REMKEYFN_MAX, APP_DB_U_REMKEYFN_MAX, APP_DB_U_REMKEYFN_MAX, APP_DB_U_REMKEYFN_MAX, APP_DB_U_REMKEYFN_MAX, APP_DB_U_REMKEYFN_MAX, APP_DB_U_REMKEYFN_MAX };

/*                                                                    SERPORT_RL_FRONT                 SERPORT_SER1_RL_REAR            SERPORT_RB1A                 SERPORT_RB1B                 SERPORT_RB2A                 SERPORT_RB2B                 SERPORT_USBGADGET            SERPORT_USBHOST              SERPORT_SER2 */
       const uint8_t serial_type_minimums[NUM_DB_SERIAL_SETTINGS] = { APP_DB_U_SERTYPE_MIN,            APP_DB_U_SERTYPE_MIN,           APP_DB_U_SERTYPE_MIN,        APP_DB_U_SERTYPE_MIN,        APP_DB_U_SERTYPE_MIN,        APP_DB_U_SERTYPE_MIN,        APP_DB_U_SERTYPE_MIN,        APP_DB_U_SERTYPE_MIN,        APP_DB_U_SERTYPE_MIN        };
       const uint8_t serial_type_maximums[NUM_DB_SERIAL_SETTINGS] = { APP_DB_U_SERTYPE_BASIC_MAX,      APP_DB_U_SERTYPE_BASIC_MAX,     APP_DB_U_SERTYPE_BASIC_MAX,  APP_DB_U_SERTYPE_TXONLY_MAX, APP_DB_U_SERTYPE_BASIC_MAX,  APP_DB_U_SERTYPE_TXONLY_MAX, APP_DB_U_SERTYPE_BASIC_MAX,  APP_DB_U_SERTYPE_FULL_MAX,   APP_DB_U_SERTYPE_FULL_MAX   };
static const uint8_t serial_type_defaults[NUM_DB_SERIAL_SETTINGS] = { DB_SERIAL_DEFAULT_TYPE_RL_FRONT, DB_SERIAL_DEFAULT_TYPE_RL_REAR, DB_SERIAL_DEFAULT_TYPE_RB1A, DB_SERIAL_DEFAULT_TYPE_RB1B, DB_SERIAL_DEFAULT_TYPE_RB2A, DB_SERIAL_DEFAULT_TYPE_RB2B, DB_SERIAL_DEFAULT_TYPE_USBG, DB_SERIAL_DEFAULT_TYPE_USBH, DB_SERIAL_DEFAULT_TYPE_SER2 };

static const uint8_t serial_baud_minimums[NUM_DB_SERIAL_SETTINGS] = { APP_DB_U_SERBAUD_MIN,            APP_DB_U_SERBAUD_MIN,           APP_DB_U_SERBAUD_MIN,        APP_DB_U_SERBAUD_MIN,        APP_DB_U_SERBAUD_MIN,        APP_DB_U_SERBAUD_MIN,        APP_DB_U_SERBAUD_MIN,        APP_DB_U_SERBAUD_MIN,        APP_DB_U_SERBAUD_MIN        };
static const uint8_t serial_baud_maximums[NUM_DB_SERIAL_SETTINGS] = { APP_DB_U_SERBAUD_MAX,            APP_DB_U_SERBAUD_MAX,           APP_DB_U_SERBAUD_MAX,        APP_DB_U_SERBAUD_MAX,        APP_DB_U_SERBAUD_MAX,        APP_DB_U_SERBAUD_MAX,        APP_DB_U_SERBAUD_MAX,        APP_DB_U_SERBAUD_MAX,        APP_DB_U_SERBAUD_MAX        };
static const uint8_t serial_baud_defaults[NUM_DB_SERIAL_SETTINGS] = { APP_DB_U_SERBAUD_DEFAULT_RL,     APP_DB_U_SERBAUD_DEFAULT_RL,    APP_DB_U_SERBAUD_DEFAULT,    APP_DB_U_SERBAUD_DEFAULT,    APP_DB_U_SERBAUD_DEFAULT,    APP_DB_U_SERBAUD_DEFAULT,    APP_DB_U_SERBAUD_DEFAULT,    APP_DB_U_SERBAUD_DEFAULT,    APP_DB_U_SERBAUD_DEFAULT    };


static const char serial_print_text_default[APP_DB_U_SERPRINTTEXT_NUMELEMS][APP_DB_PRINT_HDR_LEN] =
{
  /*
   *          1         2         3         4
   * 1234567890123456789012345678901234567890
   */
  { "            WEIGHT            " },
  { "            TICKET            " },
  { "                              " },
  { "                              " },
  { "                              " },
  { "                              " }
};


static const char *const serial_print_text_default_ptrs[APP_DB_U_SERPRINTTEXT_NUMELEMS] =
{
  serial_print_text_default[0],
  serial_print_text_default[1],
  serial_print_text_default[2],
  serial_print_text_default[3],
  serial_print_text_default[4],
  serial_print_text_default[5]
};


static char *const serial_print_text_ptrs[APP_DB_U_SERPRINTTEXT_NUMELEMS] =
{
  db_app_user.print_text[0],
  db_app_user.print_text[1],
  db_app_user.print_text[2],
  db_app_user.print_text[3],
  db_app_user.print_text[4],
  db_app_user.print_text[5]
};

static size_t serial_print_text_maxlengths[APP_DB_U_SERPRINTTEXT_NUMELEMS] =
{
  APP_DB_U_SERPRINTTEXT_MAXLEN,
  APP_DB_U_SERPRINTTEXT_MAXLEN,
  APP_DB_U_SERPRINTTEXT_MAXLEN,
  APP_DB_U_SERPRINTTEXT_MAXLEN,
  APP_DB_U_SERPRINTTEXT_MAXLEN,
  APP_DB_U_SERPRINTTEXT_MAXLEN
};


/** User application database */
struct s_db_app_user db_app_user;


struct s_db_element app_user_elements[] =
{
  /* Element type                                                Element database name            Element menu name               Element data                                     Number of elements             Element minimum               Element maximum               Element default                   First appeared in             Last appeared in              Change log mask      Change log lookup */
  /* ADC User Database */
  DB_DECLARE_INT32_ELEM(                                         APP_DB_U_ADCZERORAW_STR,         "CAL:ZERO",                     &db_app_user.adc_user_database.zeroRaw,                                         APP_DB_U_ADCZERORAW_MIN,      APP_DB_U_ADCZERORAW_MAX,      APP_DB_U_ADCZERORAW_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_INT32_ELEM(                                         APP_DB_U_ADCSPANRAW_STR,         "CAL:SPAN:RAW",                 &db_app_user.adc_user_database.spanRaw,                                         APP_DB_U_ADCSPANRAW_MIN,      APP_DB_U_ADCSPANRAW_MAX,      APP_DB_U_ADCSPANRAW_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_INT32_ELEM(                                         APP_DB_U_ADCSPANCNT_STR,         "CAL:SPAN:CNT",                 &db_app_user.adc_user_database.spanCnt,                                         APP_DB_U_ADCSPANCNT_MIN,      APP_DB_U_ADCSPANCNT_MAX,      APP_DB_U_ADCSPANCNT_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL(                        APP_DB_U_ADCLINPTS_STR,          "CAL:LIN:PTS",                  db_app_user.adc_user_database.linPts,            APP_DB_U_ADCLINPTS_NUM,        adc_linpts_minimums,          adc_linpts_maximums,          adc_linpts_defaults,              DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_INT32_ARRAY_ELEM_SINGLE(                            APP_DB_U_ADCLINCOR_STR,          "CAL:LIN:COR",                  db_app_user.adc_user_database.linCor,            APP_DB_U_ADCLINCOR_NUM,        APP_DB_U_ADCLINCOR_MIN,       APP_DB_U_ADCLINCOR_MAX,       APP_DB_U_ADCLINCOR_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCSCALETYPE_STR,       "BUILD:TYPE:1",                 &db_app_user.adc_user_database.scaleType,                                       APP_DB_U_ADCSCALETYPE_MIN,    APP_DB_U_ADCSCALETYPE_MAX,    APP_DB_U_ADCSCALETYPE_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, build_adc_scale_type_strs),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCSYNCFREQ_STR,        "SPEC:SYNC",                    &db_app_user.adc_user_database.syncFreqIdx,                                     APP_DB_U_ADCSYNCFREQ_MIN,     APP_DB_U_ADCSYNCFREQ_MAX,     APP_DB_U_ADCSYNCFREQ_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, sync_cal_setting_strs),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCJITTER_STR,          "OPTION:JITTER",                &db_app_user.adc_user_database.jitter,                                          APP_DB_U_ADCJITTER_MIN,       APP_DB_U_ADCJITTER_MAX,       APP_DB_U_ADCJITTER_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, jitter_setting),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ADCDPPOS_STR,           "BUILD:DP",                     &db_app_user.adc_user_database.dpPos,                                           APP_DB_U_ADCDPPOS_MIN,        APP_DB_U_ADCDPPOS_MAX,        APP_DB_U_ADCDPPOS_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_INT32_ARRAY_ELEM_INDIVIDUAL(                        APP_DB_U_ADCFS_STR,              "BUILD:CAP",                    db_app_user.adc_user_database.fullScale,         APP_DB_U_ADCFS_NUM,            adc_fs_minimums,              adc_fs_maximums,              adc_fs_defaults,                  DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_ARRAY_INDEX_ELEM_INDIVIDUAL(                  APP_DB_U_ADCRES_STR,             "BUILD:RES",                    db_app_user.adc_user_database.resolution,        APP_DB_U_ADCRES_NUM,           adc_res_minimums,             adc_res_maximums,             adc_res_defaults,                 DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, adc_resolution_strs),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCUNITS_STR,           "BUILD:UNITS",                  &db_app_user.adc_user_database.units,                                           APP_DB_U_ADCUNITS_MIN,        APP_DB_U_ADCUNITS_MAX,        APP_DB_U_ADCUNITS_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, adc_units_strs),
  DB_DECLARE_INT32_ELEM(                                         APP_DB_U_ADCCALWEIGHT_STR,       "CAL:SPAN:WEIGHT",              &db_app_user.adc_user_database.calWeight,                                       APP_DB_U_ADCCALWEIGHT_MIN,    APP_DB_U_ADCCALWEIGHT_MAX,    APP_DB_U_ADCCALWEIGHT_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_INT32_ELEM(                                         APP_DB_U_ADCADDTARE_STR,         "BUILD:AD.TARE",                &db_app_user.adc_user_database.addTare,                                         APP_DB_U_ADCADDTARE_MIN,      APP_DB_U_ADCADDTARE_MAX,      APP_DB_U_ADCADDTARE_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_INT32_ELEM(                                         APP_DB_U_ADCZEROBAND_STR,        "OPTION:Z.BAND",                &db_app_user.adc_user_database.zeroBand,                                        APP_DB_U_ADCZEROBAND_MIN,     APP_DB_U_ADCZEROBAND_MAX,     APP_DB_U_ADCZEROBAND_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCMOTION_STR,          "OPTION:MOTION",                &db_app_user.adc_user_database.motion,                                          APP_DB_U_ADCMOTION_MIN,       APP_DB_U_ADCMOTION_MAX,       APP_DB_U_ADCMOTION_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, motion_setting),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCZEROTRACK_STR,       "OPTION:Z.TRAC",                &db_app_user.adc_user_database.zeroTracking,                                    APP_DB_U_ADCZEROTRACK_MIN,    APP_DB_U_ADCZEROTRACK_MAX,    APP_DB_U_ADCZEROTRACK_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, zero_tracking_setting),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_ADCAVERAGE_STR,         "OPTION:FILTER",                &db_app_user.adc_user_database.average,                                         APP_DB_U_ADCAVERAGE_MIN,      APP_DB_U_ADCAVERAGE_MAX,      APP_DB_U_ADCAVERAGE_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ADCRANGEOFZERO_STR,     "OPTION:Z.RANGE",               &db_app_user.adc_user_database.rangeOfZero,                                     APP_DB_U_ADCRANGEOFZERO_MIN,  APP_DB_U_ADCRANGEOFZERO_MAX,  APP_DB_U_ADCRANGEOFZERO_DEFAULT,  DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCAPPROVMODE_STR,      "OPTION:USE",                   &db_app_user.adc_user_database.approvalMode,                                    APP_DB_U_ADCAPPROVMODE_MIN,   APP_DB_U_ADCAPPROVMODE_MAX,   APP_DB_U_ADCAPPROVMODE_DEFAULT,   DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, adc_approval_mode_strings),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCZEROONSTART_STR,     "OPTION:AUTO.Z",                &db_app_user.adc_user_database.zeroOnStart,                                     APP_DB_U_ADCZEROONSTART_MIN,  APP_DB_U_ADCZEROONSTART_MAX,  APP_DB_U_ADCZEROONSTART_DEFAULT,  DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, off_on_setting),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ADC4WIRE_STR,           NULL,                           &db_app_user.adc_user_database.wire4Mode,                                       APP_DB_U_ADC4WIRE_MIN,        APP_DB_U_ADC4WIRE_MAX,        APP_DB_U_ADC4WIRE_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ADCTEMPCALZERO_STR,     NULL,                           &db_app_user.adc_user_database.tempCalZero,                                     APP_DB_U_ADCTEMPCALZERO_MIN,  APP_DB_U_ADCTEMPCALZERO_MAX,  APP_DB_U_ADCTEMPCALZERO_DEFAULT,  DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCHIRES_STR,           "TEST:HI.RES",                  &db_app_user.adc_user_database.hiRes,                                           APP_DB_U_ADCHIRES_MIN,        APP_DB_U_ADCHIRES_MAX,        APP_DB_U_ADCHIRES_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, off_on_setting),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_ADCQAENABLE_STR,        "CLOC:QA.OPT",                  &db_app_user.adc_user_database.qaEnable,                                        APP_DB_U_ADCQAENABLE_MIN,     APP_DB_U_ADCQAENABLE_MAX,     APP_DB_U_ADCQAENABLE_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, off_on_setting),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ADCQADAY_STR,           "CLOC:QA.DATE:DAY",             &db_app_user.adc_user_database.qaDay,                                           APP_DB_U_ADCQADAY_MIN,        APP_DB_U_ADCQADAY_MAX,        APP_DB_U_ADCQADAY_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ADCQAMONTH_STR,         "CLOC:QA.DATE:MTH",             &db_app_user.adc_user_database.qaMonth,                                         APP_DB_U_ADCQAMONTH_MIN,      APP_DB_U_ADCQAMONTH_MAX,      APP_DB_U_ADCQAMONTH_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_ADCQAYEAR_STR,          "CLOC:QA.DATE:YR",              &db_app_user.adc_user_database.qaYear,                                          APP_DB_U_ADCQAYEAR_MIN,       APP_DB_U_ADCQAYEAR_MAX,       APP_DB_U_ADCQAYEAR_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_ADCCCNTEPCAL_STR,       NULL,                           &db_app_user.adc_user_database.calCntNTEPCalibration,                           APP_DB_U_ADCCCNTEPCAL_MIN,    APP_DB_U_ADCCCNTEPCAL_MAX,    APP_DB_U_ADCCCNTEPCAL_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_ADCCCNTEPSET_STR,       NULL,                           &db_app_user.adc_user_database.calCntNTEPSettings,                              APP_DB_U_ADCCCNTEPSET_MIN,    APP_DB_U_ADCCCNTEPSET_MAX,    APP_DB_U_ADCCCNTEPSET_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_ADCCCOIML_STR,          NULL,                           &db_app_user.adc_user_database.calCntOIML,                                      APP_DB_U_ADCCCOIML_MIN,       APP_DB_U_ADCCCOIML_MAX,       APP_DB_U_ADCCCOIML_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ADCCALFLAG_STR,         NULL,                           &db_app_user.adc_user_database.calFlag,                                         APP_DB_U_ADCCALFLAG_MIN,      APP_DB_U_ADCCALFLAG_MAX,      APP_DB_U_ADCCALFLAG_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_ADCCHKSUM_STR,          NULL,                           &db_app_user.adc_user_database.chksum,                                          APP_DB_U_ADCCHKSUM_MIN,       APP_DB_U_ADCCHKSUM_MAX,       APP_DB_U_ADCCHKSUM_DEFAULT,       APP_DB_U_VERSION2,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* General settings */
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SWVER_STR,              "SYSTEM:SW.VER",                &db_app_user.sw_ver,                                                            APP_DB_U_SWVER_MIN,           APP_DB_U_SWVER_MAX,           APP_DB_U_SWVER_DEFAULT,           DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_REARENTRY_STR,          "SPEC:R.ENTRY",                 &db_app_user.rear_entry,                                                        APP_DB_U_REARENTRY_MIN,       APP_DB_U_REARENTRY_MAX,       APP_DB_U_REARENTRY_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, off_on_setting),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SNSCHECK_STR,           NULL,                           &db_app_user.sense_check,                                                       APP_DB_U_SNSCHECK_MIN,        APP_DB_U_SNSCHECK_MAX,        APP_DB_U_SNSCHECK_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_BUZENABLE_STR,          NULL,                           &db_app_user.buzzer_enable,                                                     APP_DB_U_BUZENABLE_MIN,       APP_DB_U_BUZENABLE_MAX,       APP_DB_U_BUZENABLE_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_BRIGHTNESS_STR,         NULL,                           &db_app_user.brightness,                                                        APP_DB_U_BRIGHTNESS_MIN,      APP_DB_U_BRIGHTNESS_MAX,      APP_DB_U_BRIGHTNESS_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* Passcodes */
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SAFEPASS_STR,           "SPEC:SAFE.PC",                 &db_app_user.safe_passcode,                                                     APP_DB_U_SAFEPASS_MIN,        APP_DB_U_SAFEPASS_MAX,        APP_DB_U_SAFEPASS_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_FULLPASS_STR,           "SPEC:FULL.PC",                 &db_app_user.full_passcode,                                                     APP_DB_U_FULLPASS_MIN,        APP_DB_U_FULLPASS_MAX,        APP_DB_U_FULLPASS_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE),

  /* Setpoints */
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SETPACTIVE_STR,         NULL,                           db_app_user.setp_active,                         NUM_SETPOINTS,                 APP_DB_U_SETPACTIVE_MIN,      APP_DB_U_SETPACTIVE_MAX,      APP_DB_U_SETPACTIVE_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SETPALARM_STR,          NULL,                           db_app_user.setp_alarm,                          NUM_SETPOINTS,                 APP_DB_U_SETPALARM_MIN,       APP_DB_U_SETPALARM_MAX,       APP_DB_U_SETPALARM_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SETPLOCK_STR,           NULL,                           db_app_user.setp_lock,                           NUM_SETPOINTS,                 APP_DB_U_SETPLOCK_MIN,        APP_DB_U_SETPLOCK_MAX,        APP_DB_U_SETPLOCK_DEFAULT,        DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SETPSRC_STR,            NULL,                           db_app_user.setp_source,                         NUM_SETPOINTS,                 APP_DB_U_SETPSRC_MIN,         APP_DB_U_SETPSRC_MAX,         APP_DB_U_SETPSRC_DEFAULT,         DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SETPLOGIC_STR,          NULL,                           db_app_user.setp_logic,                          NUM_SETPOINTS,                 APP_DB_U_SETPLOGIC_MIN,       APP_DB_U_SETPLOGIC_MAX,       APP_DB_U_SETPLOGIC_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SETPTIMING_STR,         NULL,                           db_app_user.setp_timing,                         NUM_SETPOINTS,                 APP_DB_U_SETPTIMING_MIN,      APP_DB_U_SETPTIMING_MAX,      APP_DB_U_SETPTIMING_DEFAULT,      APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SETPRESET_STR,          NULL,                           db_app_user.setp_reset,                          NUM_SETPOINTS,                 APP_DB_U_SETPRESET_MIN,       APP_DB_U_SETPRESET_MAX,       APP_DB_U_SETPRESET_DEFAULT,       APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SETPJOGON_STR,          NULL,                           &db_app_user.setp_jog_on,                                                       APP_DB_U_SETPJOGON_MIN,       APP_DB_U_SETPJOGON_MAX,       APP_DB_U_SETPJOGON_DEFAULT,       APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SETPJOGOFF_STR,         NULL,                           &db_app_user.setp_jog_off,                                                      APP_DB_U_SETPJOGOFF_MIN,      APP_DB_U_SETPJOGOFF_MAX,      APP_DB_U_SETPJOGOFF_DEFAULT,      APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SETPJOGSET_STR,         NULL,                           &db_app_user.setp_jog_set,                                                      APP_DB_U_SETPJOGSET_MIN,      APP_DB_U_SETPJOGSET_MAX,      APP_DB_U_SETPJOGSET_DEFAULT,      APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_SETPFEEDER_STR,         NULL,                           &db_app_user.setp_feeder,                                                       APP_DB_U_SETPFEEDER_MIN,      APP_DB_U_SETPFEEDER_MAX,      APP_DB_U_SETPFEEDER_DEFAULT,      APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER, setp_feeder_strings),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SETPDELAYCHECK_STR,     NULL,                           &db_app_user.setp_delay_check,                                                  APP_DB_U_SETPDELAYCHECK_MIN,  APP_DB_U_SETPDELAYCHECK_MAX,  APP_DB_U_SETPDELAYCHECK_DEFAULT,  APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SETPMAXSET_STR,         NULL,                           &db_app_user.setp_max_sets,                                                     APP_DB_U_SETPMAXSET_MIN,      APP_DB_U_SETPMAXSET_MAX,      APP_DB_U_SETPMAXSET_DEFAULT,      APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* Key functions */
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_FRNTKEYPERMIS_STR,      NULL,                           db_app_user.front_key_permissions,               NUM_FRONT_KEYS,                APP_DB_U_FRNTKEYPERMIS_MIN,   APP_DB_U_FRNTKEYPERMIS_MAX,   APP_DB_U_FRNTKEYPERMIS_DEFAULT,   DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL(                        APP_DB_U_FRNTKEYFN_STR,          NULL,                           db_app_user.front_key_function,                  NUM_FRONT_FUNCTION_KEYS,       frntkey_f_minimums,           frntkey_f_maximums,           frontfunc_key_defaults,           DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL(                        APP_DB_U_REMKEYFN_STR,           NULL,                           db_app_user.remote_f,                            NUM_REMOTE_KEYS,               remote_f_minimums,            remote_f_maximums,            remote_key_defaults,              DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* Analog output */
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ANALOUTTYPE_STR,        NULL,                           &db_app_user.analog_output_type,                                                APP_DB_U_ANALOUTTYPE_MIN,     APP_DB_U_ANALOUTTYPE_MAX,     APP_DB_U_ANALOUTTYPE_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ANALOUTSRC_STR,         NULL,                           &db_app_user.analog_output_source,                                              APP_DB_U_ANALOUTSRC_MIN,      APP_DB_U_ANALOUTSRC_MAX,      APP_DB_U_ANALOUTSRC_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_INT32_ELEM(                                         APP_DB_U_ANALOUTSPAN_STR,        NULL,                           &db_app_user.analog_span_cal,                                                   APP_DB_U_ANALOUTSPAN_MIN,     APP_DB_U_ANALOUTSPAN_MAX,     APP_DB_U_ANALOUTSPAN_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_INT32_ELEM(                                         APP_DB_U_ANALOUTZERO_STR,        NULL,                           &db_app_user.analog_zero_cal,                                                   APP_DB_U_ANALOUTZERO_MIN,     APP_DB_U_ANALOUTZERO_MAX,     APP_DB_U_ANALOUTZERO_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ANALOUTABS_STR,         NULL,                           &db_app_user.analog_output_abs,                                                 APP_DB_U_ANALOUTABS_MIN,      APP_DB_U_ANALOUTABS_MAX,      APP_DB_U_ANALOUTABS_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_ANALOUTCLIP_STR,        NULL,                           &db_app_user.analog_output_clip,                                                APP_DB_U_ANALOUTCLIP_MIN,     APP_DB_U_ANALOUTCLIP_MAX,     APP_DB_U_ANALOUTCLIP_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* DSD */
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_DSDENABLE_STR,          "DSD:ENABLE",                   &db_app_user.dsd_enable,                                                        APP_DB_U_DSDENABLE_MIN,       APP_DB_U_DSDENABLE_MAX,       APP_DB_U_DSDENABLE_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, off_on_setting),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_DSDAUTOPURGE_STR,       "DSD:A.PURGE",                  &db_app_user.dsd_auto_purge,                                                    APP_DB_U_DSDAUTOPURGE_MIN,    APP_DB_U_DSDAUTOPURGE_MAX,    APP_DB_U_DSDAUTOPURGE_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, off_on_setting),

  /* Serial settings */
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERADDRESS_STR,         NULL,                           &db_app_user.address,                                                           APP_DB_U_SERADDRESS_MIN,      APP_DB_U_SERADDRESS_MAX,      APP_DB_U_SERADDRESS_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL(                        APP_DB_U_SERTYPE_STR,            NULL,                           old_serial_type,                                OLD_NUM_DB_SERIAL_SETTINGS,     serial_type_minimums,         serial_type_maximums,         serial_type_defaults,             DB_NO_MINIMUM_VERSION,        APP_DB_U_VERSION5,            APP_DB_CLMASK_USER),  /* We added a serial port at version 6, so the shorter array ends at version 5 */
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL(                        APP_DB_U_SERTYPE_STR,            NULL,                           db_app_user.serial_type,                        NUM_DB_SERIAL_SETTINGS,         serial_type_minimums,         serial_type_maximums,         serial_type_defaults,             APP_DB_U_VERSION6_DB9_SERIAL, DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),  /* And the longer array starts at version 6 */
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL(                        APP_DB_U_SERBAUD_STR,            NULL,                           old_serial_baud_idx,                            OLD_NUM_DB_SERIAL_SETTINGS,     serial_baud_minimums,         serial_baud_maximums,         serial_baud_defaults,             DB_NO_MINIMUM_VERSION,        APP_DB_U_VERSION5,            APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL(                        APP_DB_U_SERBAUD_STR,            NULL,                           db_app_user.serial_baud_idx,                    NUM_DB_SERIAL_SETTINGS,         serial_baud_minimums,         serial_baud_maximums,         serial_baud_defaults,             APP_DB_U_VERSION6_DB9_SERIAL, DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SERBITS_STR,            NULL,                           old_serial_bit_settings,                        OLD_NUM_DB_SERIAL_SETTINGS,     APP_DB_U_SERBITS_MIN,         APP_DB_U_SERBITS_MAX,         APP_DB_U_SERBITS_DEFAULT,         DB_NO_MINIMUM_VERSION,        APP_DB_U_VERSION5,            APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(                            APP_DB_U_SERBITS_STR,            NULL,                           db_app_user.serial_bit_settings,                NUM_DB_SERIAL_SETTINGS,         APP_DB_U_SERBITS_MIN,         APP_DB_U_SERBITS_MAX,         APP_DB_U_SERBITS_DEFAULT,         APP_DB_U_VERSION6_DB9_SERIAL, DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* TCP/UDP port settings */
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_SERUDPO1TYPE_STR,       NULL,                           &db_app_user.serial_udpo1_type,                                                 APP_DB_U_SERTYPE_MIN,         APP_DB_U_SERTYPE_BASIC_MAX,   APP_DB_U_SERUDPO1TYPE_DEFAULT,    APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, ser_type_setting),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_SERUDPO2TYPE_STR,       NULL,                           &db_app_user.serial_udpo2_type,                                                 APP_DB_U_SERTYPE_MIN,         APP_DB_U_SERTYPE_BASIC_MAX,   APP_DB_U_SERUDPO2TYPE_DEFAULT,    APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, ser_type_setting),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_SERUDPI1TYPE_STR,       NULL,                           &db_app_user.serial_udpi1_type,                                                 APP_DB_U_SERTYPE_MIN,         APP_DB_U_SERTYPE_BASIC_MAX,   APP_DB_U_SERUDPI1TYPE_DEFAULT,    APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, ser_type_setting),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_SERUDPI2TYPE_STR,       NULL,                           &db_app_user.serial_udpi2_type,                                                 APP_DB_U_SERTYPE_MIN,         APP_DB_U_SERTYPE_BASIC_MAX,   APP_DB_U_SERUDPI2TYPE_DEFAULT,    APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, ser_type_setting),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_SERTCPI2TYPE_STR,       NULL,                           &db_app_user.serial_tcpi2_type,                                                 APP_DB_U_SERTCPI2TYPE_MIN,    APP_DB_U_SERTCPI2TYPE_MAX,    APP_DB_U_SERTCPI2TYPE_DEFAULT,    APP_DB_U_VERSION5,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_TRADE, ser_type_setting),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERUDPO1DEST_STR,       NULL,                           &db_app_user.serial_udpo1_dstaddr,                                              APP_DB_U_SERUDPODEST_MIN,     APP_DB_U_SERUDPODEST_MAX,     APP_DB_U_SERUDPODEST_DEFAULT,     APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERUDPO2DEST_STR,       NULL,                           &db_app_user.serial_udpo2_dstaddr,                                              APP_DB_U_SERUDPODEST_MIN,     APP_DB_U_SERUDPODEST_MAX,     APP_DB_U_SERUDPODEST_DEFAULT,     APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERUDPO1PORT_STR,       NULL,                           &db_app_user.serial_udpo1_port,                                                 APP_DB_U_SERUDPO1PORT_MIN,    APP_DB_U_SERUDPO1PORT_MAX,    APP_DB_U_SERUDPO1PORT_DEFAULT,    APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERUDPO2PORT_STR,       NULL,                           &db_app_user.serial_udpo2_port,                                                 APP_DB_U_SERUDPO2PORT_MIN,    APP_DB_U_SERUDPO2PORT_MAX,    APP_DB_U_SERUDPO2PORT_DEFAULT,    APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERTCPI1PORT_STR,       NULL,                           &db_app_user.serial_tcpi1_port,                                                 APP_DB_U_SERTCPI1PORT_MIN,    APP_DB_U_SERTCPI1PORT_MAX,    APP_DB_U_SERTCPI1PORT_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERTCPI2PORT_STR,       NULL,                           &db_app_user.serial_tcpi2_port,                                                 APP_DB_U_SERTCPI2PORT_MIN,    APP_DB_U_SERTCPI2PORT_MAX,    APP_DB_U_SERTCPI2PORT_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERUDPI1PORT_STR,       NULL,                           &db_app_user.serial_udpi1_port,                                                 APP_DB_U_SERUDPI1PORT_MIN,    APP_DB_U_SERUDPI1PORT_MAX,    APP_DB_U_SERUDPI1PORT_DEFAULT,    APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERUDPI2PORT_STR,       NULL,                           &db_app_user.serial_udpi2_port,                                                 APP_DB_U_SERUDPI2PORT_MIN,    APP_DB_U_SERUDPI2PORT_MAX,    APP_DB_U_SERUDPI2PORT_DEFAULT,    APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERTCPI1TIMEOUT_STR,    NULL,                           &db_app_user.serial_tcpi1_timeout,                                              APP_DB_U_SERTCPI1TIMEOUT_MIN, APP_DB_U_SERTCPI1TIMEOUT_MAX, APP_DB_U_SERTCPI1TIMEOUT_DEFAULT, APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_SERTCPI2TIMEOUT_STR,    NULL,                           &db_app_user.serial_tcpi2_timeout,                                              APP_DB_U_SERTCPI2TIMEOUT_MIN, APP_DB_U_SERTCPI2TIMEOUT_MAX, APP_DB_U_SERTCPI2TIMEOUT_DEFAULT, APP_DB_U_VERSION3,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* Auto out/printer settings */
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERSTARTCHAR_STR,       NULL,                           &db_app_user.start_char,                                                        APP_DB_U_SERSTARTCHAR_MIN,    APP_DB_U_SERSTARTCHAR_MAX,    APP_DB_U_SERSTARTCHAR_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERENDCHAR1_STR,        NULL,                           &db_app_user.end_char1,                                                         APP_DB_U_SERENDCHAR1_MIN,     APP_DB_U_SERENDCHAR1_MAX,     APP_DB_U_SERENDCHAR1_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERENDCHAR2_STR,        NULL,                           &db_app_user.end_char2,                                                         APP_DB_U_SERENDCHAR2_MIN,     APP_DB_U_SERENDCHAR2_MAX,     APP_DB_U_SERENDCHAR2_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERPRINTMODE_STR,       NULL,                           &db_app_user.printer_mode,                                                      APP_DB_U_SERPRINTMODE_MIN,    APP_DB_U_SERPRINTMODE_MAX,    APP_DB_U_SERPRINTMODE_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERTICKETROW_STR,       NULL,                           &db_app_user.tic_row,                                                           APP_DB_U_SERTICKETROW_MIN,    APP_DB_U_SERTICKETROW_MAX,    APP_DB_U_SERTICKETROW_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERTICKETCOL_STR,       NULL,                           &db_app_user.tic_col,                                                           APP_DB_U_SERTICKETCOL_MIN,    APP_DB_U_SERTICKETCOL_MAX,    APP_DB_U_SERTICKETCOL_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_STRING_ELEM(                                        APP_DB_U_SERID_STR,              NULL,                           db_app_user.id,                                 APP_DB_U_SERID_MAXLEN,                                                                      APP_DB_U_SERID_DEFAULT,           DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERCOF_STR,             NULL,                           &db_app_user.cof_setting,                                                       APP_DB_U_SERCOF_MIN,          APP_DB_U_SERCOF_MAX,          APP_DB_U_SERCOF_DEFAULT,          DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERAOFTYPE_STR,         NULL,                           &db_app_user.aof_setting,                                                       APP_DB_U_SERAOFTYPE_MIN,      APP_DB_U_SERAOFTYPE_MAX,      APP_DB_U_SERAOFTYPE_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_SERAOFSRC_STR,          NULL,                           &db_app_user.aof_source,                                                        APP_DB_U_SERAOFSRC_MIN,       APP_DB_U_SERAOFSRC_MAX,       APP_DB_U_SERAOFSRC_DEFAULT,       DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_STRING_ARRAY_ELEM_INDIVIDUAL(                       APP_DB_U_SERPRINTTEXT_STR,       NULL,                           serial_print_text_ptrs,                         APP_DB_U_SERPRINTTEXT_NUMELEMS,                               serial_print_text_maxlengths, serial_print_text_default_ptrs,   DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_STRING_ELEM(                                        APP_DB_U_SERPRINTFMT_STR,        NULL,                           db_app_user.print_fmt,                          APP_DB_U_SERPRINTFMT_MAXLEN,                                                                APP_DB_U_SERPRINTFMT_DEFAULT,     DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_STRING_ELEM(                                        APP_DB_U_SERAUTOFMT_STR,         NULL,                           db_app_user.auto_fmt,                           APP_DB_U_SERAUTOFMT_MAXLEN,                                                                 APP_DB_U_SERAUTOFMT_DEFAULT,      DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_PRINTINTERLOCK_STR,     NULL,                           &db_app_user.print_interlock,                                                   APP_DB_U_PRINTINTERLOCK_MIN,  APP_DB_U_PRINTINTERLOCK_MAX,  APP_DB_U_PRINTINTERLOCK_DEFAULT,  APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* Modbus settings */
  DB_DECLARE_UINT8_ELEM(                                         APP_DB_U_MODBUSENDIAN_STR,       NULL,                           &db_app_user.modbus_endian,                                                     APP_DB_U_MODBUSENDIAN_MIN,    APP_DB_U_MODBUSENDIAN_MAX,    APP_DB_U_MODBUSENDIAN_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_MODBUSTCPPORT_STR,      NULL,                           &db_app_user.modbus_tcpport,                                                    APP_DB_U_MODBUSTCPPORT_MIN,   APP_DB_U_MODBUSTCPPORT_MAX,   APP_DB_U_MODBUSTCPPORT_DEFAULT,   DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* Licensing settings */
  DB_DECLARE_UINT32_ARRAY_ELEM_SINGLE(                           APP_DB_U_LICENCECODES_STR,       NULL,                           db_app_user.licence_codes_obsolete,             APP_DB_U_LICENCECODES_OBSOLETE_NUM, APP_DB_U_LICENCECODES_MIN, APP_DB_U_LICENCECODES_MAX,   APP_DB_U_LICENCECODES_DEFAULT,    DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_LICENCECODENUM_STR,     NULL,                           &db_app_user.num_licence_codes_obsolete,                                        APP_DB_U_LICENCECODENUM_MIN,  APP_DB_U_LICENCECODENUM_MAX,  APP_DB_U_LICENCECODENUM_DEFAULT,  DB_NO_MINIMUM_VERSION,        DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),

  /* Clock */
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_TIME_FMT_STR,           NULL,                           &db_app_user.time_fmt,                                                          APP_DB_U_TIME_FMT_MIN,         APP_DB_U_TIME_FMT_MAX,       APP_DB_U_TIME_FMT_DEFAULT,        APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER, time_format_settings),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_DATE_FMT_STR,           NULL,                           &db_app_user.date_fmt,                                                          APP_DB_U_DATE_FMT_MIN,         APP_DB_U_DATE_FMT_MAX,       APP_DB_U_DATE_FMT_DEFAULT,        APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER, date_format_settings),

  /* Units switching */
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_UNITS_SEC_STR,          NULL,                           &db_app_user.units_sec,                                                         APP_DB_U_UNITS_SEC_MIN,        APP_DB_U_UNITS_SEC_MAX,      APP_DB_U_UNITS_SEC_DEFAULT,       APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER, alt_units_menus),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_UNITS_TER_STR,          NULL,                           &db_app_user.units_ter,                                                         APP_DB_U_UNITS_TER_MIN,        APP_DB_U_UNITS_TER_MAX,      APP_DB_U_UNITS_TER_DEFAULT,       APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER, alt_units_menus),
  DB_DECLARE_UINT32_ELEM(                                        APP_DB_U_UNITS_CONV_STR,         NULL,                           &db_app_user.unit_conv_factor,                                                  APP_DB_U_UNITS_CONV_MIN,       APP_DB_U_UNITS_CONV_MAX,     APP_DB_U_UNITS_CONV_DEFAULT,      APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER),
  DB_DECLARE_UINT8_INDEX_ELEM(                                   APP_DB_U_SHOW_LBOZ_STR,          NULL,                           &db_app_user.show_lboz,                                                         APP_DB_U_SHOW_LBOZ_MIN,        APP_DB_U_SHOW_LBOZ_MAX,      APP_DB_U_SHOW_LBOZ_DEFAULT,       APP_DB_U_VERSION4,            DB_NO_MAXIMUM_VERSION,        APP_DB_CLMASK_USER, off_on_setting),
};


struct s_db_database app_user_database =
{
  .db_database_version = APP_DB_U_VERSION6_DB9_SERIAL,
  .db_database_chksum_seed = APP_DB_U_CHECKSUM_SEED,
  .db_elements = app_user_elements,
  .db_num_elements = DB_NUM_ELEMENTS(app_user_elements),
  .db_name = APP_DB_U_NAME,
  .db_checksum_buffer = NULL
};


static struct s_db_database_interp app_user_ddi_xml1;
static struct s_db_database_interp app_user_ddi_xml2;


/**
 * Write the user database to file
 *
 * @param write_ass_copy specifies whether the adcSuperStructure copy of the adcUserDatabase should be written
 * @param lock specifies whether adcSuperStructure should be locked before writing
 */
bool app_write_user_database(bool write_ass_copy, bool lock)
{
  enum eDBError res1;
  enum eDBError res2;

  /* Check that the change log is not full */
  if (!app_can_save_databases())
  {
    rdebug_printf(LOG_ERR, "app_write_user_database: change log is full, not writing database\n");
    return false;
  }

  if (write_ass_copy)
  {
    /* Write the adcSuperStructure copy */
    if (lock)
      AdcMutexLock(&adcSuperStructure);
    db_app_user.adc_user_database = adcSuperStructure.adcUserDatabase;
    if (lock)
      AdcMutexUnlock(&adcSuperStructure);
  }

  /* Write user database to first file */
  res1 = db_database_write(&app_user_database, &app_user_ddi_xml1, app_user_database.db_database_version);
  if (dbeSuccess != res1)
  {
    rdebug_printf(LOG_ERR, "app_write_user_database: db_database_write() first copy returned %d\n", res1);
  }

  /* Write user database to second file */
  res2 = db_database_write(&app_user_database, &app_user_ddi_xml2, app_user_database.db_database_version);
  if (dbeSuccess != res2)
  {
    rdebug_printf(LOG_ERR, "app_write_user_database: db_database_write() second copy returned %d\n", res2);
  }

  /* Update the change log */
  (void) app_changelog_update(&app_user_database);

  return (dbeSuccess == res1 && dbeSuccess == res2);
}


/** If we read an old database version, then copy across the serial settings */
static void app_adopt_old_serial_settings(uint32_t read_at_version)
{
  if (read_at_version < APP_DB_U_VERSION6_DB9_SERIAL)
  {
    memcpy(db_app_user.serial_type, old_serial_type, OLD_NUM_DB_SERIAL_SETTINGS);
    memcpy(db_app_user.serial_baud_idx, old_serial_baud_idx, OLD_NUM_DB_SERIAL_SETTINGS);
    memcpy(db_app_user.serial_bit_settings, old_serial_bit_settings, OLD_NUM_DB_SERIAL_SETTINGS);
  }
}


/** Read the user database from file */
bool app_read_user_database(void)
{
  enum eDBError res;
  uint32_t version;

  /* Read the database from file */
  res = db_database_read(&app_user_database, &app_user_ddi_xml1, &version, false);
  if (dbeSuccess != res)
  {
    rdebug_printf(LOG_ERR, "app_read_user_database: db_database_read() first copy returned %d, trying second copy\n", res);
    res = db_database_read(&app_user_database, &app_user_ddi_xml2, &version, false);
    if (dbeSuccess != res)
    {
      rdebug_printf(LOG_ERR, "app_read_user_database: db_database_read() second copy returned %d\n", res);
    }
  }
  if (dbeSuccess == res)
    app_adopt_old_serial_settings(version);

  /* Update previous data, so change log will work */
  db_database_update_previous(&app_user_database);

  return (dbeSuccess == res);
}


/**
 * Default the user database
 *
 * @param default_calibration_also if true will also default the calibration (db_app_user.adc_user_database)
 */
bool app_default_user_database(bool default_calibration_also)
{
  enum eDBError res;

  AdcMutexLock(&adcSuperStructure);

  /* Default the whole database */
  res = db_database_default(&app_user_database);
  if (dbeSuccess != res)
  {
    rdebug_printf(LOG_ERR, "app_default_user_database: db_database_default() returned %d\n", res);
  }

  /* Either default the calibration, or leave it unchanged */
  if (default_calibration_also)
  {
    /* Use the ADC database defaulter for all ADC values */
    AdcDefaultDatabases(&adcSuperStructure, false, NULL, &adcSuperStructure.adcUserDatabase, NULL, true, false);
    db_app_user.adc_user_database = adcSuperStructure.adcUserDatabase;
  }
  else
  {
    /* But don't overwrite the calibration */
    db_app_user.adc_user_database = adcSuperStructure.adcUserDatabase;
  }

  AdcMutexUnlock(&adcSuperStructure);

  return (dbeSuccess == res);
}


/** Initialise the user database */
bool app_init_user_database(void)
{
  enum eDBError res1;
  enum eDBError res2;
  enum eDBError res3;

  /* Initialise the database (creates previous data for some elements) */
  res1 = db_database_init(&app_user_database);
  if (dbeSuccess != res1)
  {
    rdebug_printf(LOG_ERR, "app_init_user_database: db_database_init() returned %d\n", res1);
  }

  /* Create database interpreters */
  res2 = db_create_database_interp_xml(&app_user_ddi_xml1, we_configuration.userdb1_file);
  if (dbeSuccess != res2)
  {
    rdebug_printf(LOG_ERR, "app_init_user_database: db_create_database_interp_xml() for first copy returned %d\n", res2);
  }
  res3 = db_create_database_interp_xml(&app_user_ddi_xml2, we_configuration.userdb2_file);
  if (dbeSuccess != res3)
  {
    rdebug_printf(LOG_ERR, "app_init_user_database: db_create_database_interp_xml() for second copy returned %d\n", res3);
  }

  return (dbeSuccess == res1 && dbeSuccess == res2 && dbeSuccess == res3);
}


/** Destroy the user database */
bool app_destroy_user_database(void)
{
  (void) db_database_destroy(&app_user_database);
  (void) db_destroy_database_interp_xml(&app_user_ddi_xml1);
  (void) db_destroy_database_interp_xml(&app_user_ddi_xml2);
  return true;
}
