/*
 * app_db_user.h
 *
 *  Created on: Apr 13, 2012
 *      Author: ian
 */

#ifndef APP_DB_USER_H_
#define APP_DB_USER_H_

#include <stddef.h>
#include <stdint.h>
#include <limits.h>
#include <rinweigh/adcdatatypes.h>
#include <modbus.h>
#include "db_setpoint.h"
#include "db_special.h"
#include "db_remote.h"
#include "db_serial.h"
#include "db_serial_ip.h"
#include "db_build.h"
#include "db_clock.h"
#include "iface_analog.h"
#include "iface_serial_config.h"
#include "iface_serial_ports.h"
#include "comms_util.h"
#include "version.h"
#include "modbus_task.h"


#define DISPLAY_INT_MAX 999999
#define DISPLAY_INT_MIN -99999

#define PERCENT_MAX 100
#define PERCENT_MIN 0

#define BOOLEAN_MAX 1
#define BOOLEAN_MIN 0

#define APP_DB_ID_STR_LEN 16

#define APP_DB_PRINT_FMT_LEN 250
#define APP_DB_AUTO_FMT_LEN 50
#define APP_DB_PRINT_HDR_LEN 31
#define APP_DB_PRINT_HDR_LINES 6

#define APP_DB_U_LICENCECODES_OBSOLETE_NUM 50


/** Application user database definition */
struct s_db_app_user
{
  /* ADC User database */
  AdcUserDatabase adc_user_database;                    /*!< All settings for ADC user calibration */

  /* General settings */
  uint32_t sw_ver;                                      /*!< Software version check to allow cal count to be updated when software changed */
  uint8_t rear_entry;                                   /*!< True if full menu access only permitted from rear key */
  uint8_t sense_check;                                  /*!< True if sense line checking enabled */
  uint8_t buzzer_enable;                                /*!< True if the buzzer is enabled */
  uint8_t brightness;                                   /*!< Display brightness */

  /* Passcodes */
  uint32_t safe_passcode;                               /*!< Safe setup passcode */
  uint32_t full_passcode;                               /*!< Full setup passcode */

  /* Setpoint settings */
  uint8_t setp_active[NUM_SETPOINTS];                   /*!< Setpoint type */
  uint8_t setp_alarm[NUM_SETPOINTS];                    /*!< Setpoint alarm */
  uint8_t setp_lock[NUM_SETPOINTS];                     /*!< Setpoint lock */
  uint8_t setp_source[NUM_SETPOINTS];                   /*!< Setpoint source */
  uint8_t setp_logic[NUM_SETPOINTS];                    /*!< Setpoint logic */
  uint8_t setp_timing[NUM_SETPOINTS];                   /*!< Setpoint timing */
  uint8_t setp_reset[NUM_SETPOINTS];                    /*!< Setpoint reset */
  uint32_t setp_jog_on;                                 /*!< Setpoint jog on time */
  uint32_t setp_jog_off;                                /*!< Setpoint jog off time */
  uint8_t setp_jog_set;                                 /*!< Setpoint jog set */
  uint8_t setp_max_sets;                                /*!< Setpoint max sets */
  uint8_t setp_feeder;                                  /*!< Setpoint feeder */
  uint32_t setp_delay_check;                            /*!< Setpoint delay check */

  /* Key functions */
  uint8_t front_key_permissions[NUM_FRONT_KEYS];        /*!< Enable permission for each of the front keys */
  uint8_t front_key_function[NUM_FRONT_FUNCTION_KEYS];  /*!< What each of the front function keys do */
  uint8_t remote_f[NUM_REMOTE_KEYS];                    /*!< What each of the remote keys do */

  /* Analog output */
  uint8_t analog_output_type;                           /*!< Analog output type */
  uint8_t analog_output_source;                         /*!< Analog output data source */
  int32_t analog_span_cal;                              /*!< Analog output span correction */
  int32_t analog_zero_cal;                              /*!< Analog output zero correction */
  uint8_t analog_output_abs;                            /*!< Analog output use absolute values */
  uint8_t analog_output_clip;                           /*!< Analog output clip output to 0-100% */

  /* DSD */
  uint8_t dsd_enable;                                   /*!< Is the DSD enabled */
  uint8_t dsd_auto_purge;                               /*!< If enabled, automatically purge x% of records when DSD is full */

  /* Serial settings */
  uint8_t address;                                      /*!< Unit address (used with Sxx; commands) */
  uint8_t serial_type[NUM_DB_SERIAL_SETTINGS];          /*!< Function of each port (autoout, printer, network, etc) */
  uint8_t serial_baud_idx[NUM_DB_SERIAL_SETTINGS];      /*!< Baud rate index for each port */
  uint8_t serial_bit_settings[NUM_DB_SERIAL_SETTINGS];  /*!< Bit packed parity, stop, data bits for each port */

  /* TCP/UDP port settings */
  uint8_t serial_udpo1_type;                            /*!< Serial UDP output 1 type (autoout, printer, network, etc) */
  uint8_t serial_udpo2_type;                            /*!< Serial UDP output 2 type (autoout, printer, network, etc) */
  uint8_t serial_udpi1_type;                            /*!< Serial UDP input 1 type (autoout, printer, network, etc) */
  uint8_t serial_udpi2_type;                            /*!< Serial UDP input 2 type (autoout, printer, network, etc) */
  uint8_t serial_tcpi2_type;                            /*!< Serial TCP input 2 type (autolo, autohi) */
  uint32_t serial_udpo1_dstaddr;                        /*!< Serial UDP output 1 destination IP address */
  uint32_t serial_udpo2_dstaddr;                        /*!< Serial UDP output 2 destination IP address */
  uint32_t serial_udpo1_port;                           /*!< Serial UDP output 1 destination UDP port */
  uint32_t serial_udpo2_port;                           /*!< Serial UDP output 2 destination UDP port */
  uint32_t serial_tcpi1_port;                           /*!< Serial TCP input 1 listens on this TCP port */
  uint32_t serial_tcpi2_port;                           /*!< Serial TCP input 2 listens on this TCP port */
  uint32_t serial_udpi1_port;                           /*!< Serial UDP input 1 listens on this TCP port */
  uint32_t serial_udpi2_port;                           /*!< Serial UDP input 2 listens on this TCP port */
  uint32_t serial_tcpi1_timeout;                        /*!< Serial TCP input 1 timeout period in seconds */
  uint32_t serial_tcpi2_timeout;                        /*!< Serial TCP input 2 timeout period in seconds */

  /* Auto out/printer settings */
  uint8_t start_char;                                   /*!< Auto output start char */
  uint8_t end_char1;                                    /*!< Auto output first end char */
  uint8_t end_char2;                                    /*!< Auto output second end char */
  uint8_t printer_mode;                                 /*!< Type of printer output */
  uint8_t tic_row;                                      /*!< Number of blank rows to print before a printout */
  uint8_t tic_col;                                      /*!< Number of blank columns to print before a printout */
  char id[APP_DB_ID_STR_LEN];                           /*!< User writable identifier */
  uint8_t cof_setting;                                  /*!< COF setting (controls MSV command output format) */
  uint8_t aof_setting;                                  /*!< Auto output format */
  uint8_t aof_source;                                   /*!< Auto output source */
  char print_text[APP_DB_PRINT_HDR_LINES][APP_DB_PRINT_HDR_LEN];
  char print_fmt[APP_DB_PRINT_FMT_LEN];                 /*!< Null terminated string for printer output formatting */
  char auto_fmt[APP_DB_AUTO_FMT_LEN];                   /*!< Null terminated string for auto output formatting */
  uint8_t modbus_endian;                                /*!< Endian for 32bit data types in modbus (big/little) */
  uint32_t modbus_tcpport;                              /*!< Modbus listens on this TCP port */
  uint8_t print_interlock;                              /*!< Print interlock setting */

  /* Licensing */
  uint32_t licence_codes_obsolete[APP_DB_U_LICENCECODES_OBSOLETE_NUM]; /*!< Obsolete, see new licence database */
  uint32_t num_licence_codes_obsolete;                  /*!< Obsolete, see new licence database */

  /* Clock */
  uint8_t time_fmt;                                     /*!< Time format */
  uint8_t date_fmt;                                     /*!< Date format */

  /* Unit switching */
  uint8_t units_sec;                                    /*!< Secondary units */
  uint8_t units_ter;                                    /*!< Tertiary units */
  uint32_t unit_conv_factor;                            /*!< Custom unit conversion factor */
  uint8_t show_lboz;                                    /*!< Show oz weights as lb:oz where possible */
};


#define APP_DB_U_NAME "userdb"
#define APP_DB_U_CHECKSUM_SEED 0x12345678


/** Initial user database version */
#define APP_DB_U_VERSION1 1


/** User database version which added checksum AdcUserDatabase.chksum */
#define APP_DB_U_VERSION2 2


/** User database version which added TCP timeout */
#define APP_DB_U_VERSION3 3


/** User database version which added setpoint jogging, print interlock, clock formats and unit switching */
#define APP_DB_U_VERSION4 4


/** User database version which added TCP.IN.2 type (autolo, autohi) */
#define APP_DB_U_VERSION5 5


/** User database version which support for DB9 serial 1 & 2 was added */
#define APP_DB_U_VERSION6_DB9_SERIAL 6


/*
 * ADC database settings
 */


#define APP_DB_U_ADCZERORAW_STR "adczeroraw"
#define APP_DB_U_ADCZERORAW_MIN (ADC_DB_USER_ZERORAW_MINIMUM)
#define APP_DB_U_ADCZERORAW_MAX (ADC_DB_USER_ZERORAW_MAXIMUM)
#define APP_DB_U_ADCZERORAW_DEFAULT (ADC_DB_USER_ZERORAW_DEFAULT)

#define APP_DB_U_ADCSPANRAW_STR "adcspanraw"
#define APP_DB_U_ADCSPANRAW_MIN (ADC_DB_USER_SPANRAW_MINIMUM)
#define APP_DB_U_ADCSPANRAW_MAX (ADC_DB_USER_SPANRAW_MAXIMUM)
#define APP_DB_U_ADCSPANRAW_DEFAULT (ADC_DB_USER_SPANRAW_DEFAULT)

#define APP_DB_U_ADCSPANCNT_STR "adcspancnt"
#define APP_DB_U_ADCSPANCNT_MIN (ADC_DB_USER_SPANCNT_MINIMUM)
#define APP_DB_U_ADCSPANCNT_MAX (ADC_DB_USER_SPANCNT_MAXIMUM)
#define APP_DB_U_ADCSPANCNT_DEFAULT (ADC_DB_USER_SPANCNT_DEFAULT)

#define APP_DB_U_ADCLINPTS_STR "adclinpoints"
#define APP_DB_U_ADCLINPTS_NUM (ADC_MAX_NUM_LIN_PTS + 1)
#define APP_DB_U_ADCLINPTS_MIN (ADC_DB_USER_LINPTS_MINIMUM)
#define APP_DB_U_ADCLINPTS_MAX (ADC_DB_USER_LINPTS_MAXIMUM)
#define APP_DB_U_ADCLINPTS_DEFAULT (ADC_DB_USER_LINPTS_DEFAULT)
#define APP_DB_U_ADCLINPTSLAST_DEFAULT (ADC_DB_USER_LINPTSLAST_DEFAULT)

#define APP_DB_U_ADCLINCOR_STR "adclincorrection"
#define APP_DB_U_ADCLINCOR_NUM (ADC_MAX_NUM_LIN_PTS + 1)
#define APP_DB_U_ADCLINCOR_MIN (ADC_DB_USER_LINCOR_MINIMUM)
#define APP_DB_U_ADCLINCOR_MAX (ADC_DB_USER_LINCOR_MAXIMUM)
#define APP_DB_U_ADCLINCOR_DEFAULT (ADC_DB_USER_LINCOR_DEFAULT)

#define APP_DB_U_ADCSCALETYPE_STR "adcscaletype"
#define APP_DB_U_ADCSCALETYPE_MIN (ADC_DB_USER_SCALETYPE_MINIMUM)
#define APP_DB_U_ADCSCALETYPE_MAX (ADC_DB_USER_SCALETYPE_MAXIMUM)
#define APP_DB_U_ADCSCALETYPE_DEFAULT (ADC_DB_USER_SCALETYPE_DEFAULT)

#define APP_DB_U_ADCSYNCFREQ_STR "adcsyncfreq"
#define APP_DB_U_ADCSYNCFREQ_MIN (ADC_DB_USER_SYNCFREQIDX_MINIMUM)
#define APP_DB_U_ADCSYNCFREQ_MAX (ADC_DB_USER_SYNCFREQIDX_MAXIMUM)
#define APP_DB_U_ADCSYNCFREQ_DEFAULT (ADC_DB_USER_SYNCFREQIDX_DEFAULT)

#define APP_DB_U_ADCJITTER_STR "adcjitter"
#define APP_DB_U_ADCJITTER_MIN (ADC_DB_USER_JITTER_MINIMUM)
#define APP_DB_U_ADCJITTER_MAX (ADC_DB_USER_JITTER_MAXIMUM)
#define APP_DB_U_ADCJITTER_DEFAULT (ADC_JITTER_OFF)

#define APP_DB_U_ADCDPPOS_STR "adcdppos"
#define APP_DB_U_ADCDPPOS_MIN (ADC_DB_USER_DPPOS_MINIMUM)
#define APP_DB_U_ADCDPPOS_MAX (ADC_DB_USER_DPPOS_MAXIMUM)
#define APP_DB_U_ADCDPPOS_DEFAULT (ADC_DB_USER_DPPOS_DEFAULT)

#define APP_DB_U_ADCFS_STR "adcfullscale"
#define APP_DB_U_ADCFS_NUM (ADC_MAX_NUM_INTERVAL)
#define APP_DB_U_ADCFS0_MIN (ADC_DB_USER_FULLSCALE0_MINIMUM)
#define APP_DB_U_ADCFS0_MAX (ADC_DB_USER_FULLSCALE0_MAXIMUM)
#define APP_DB_U_ADCFS0_DEFAULT (ADC_DB_USER_FULLSCALE0_DEFAULT)
#define APP_DB_U_ADCFS1_MIN (ADC_DB_USER_FULLSCALE1_MINIMUM)
#define APP_DB_U_ADCFS1_MAX (ADC_DB_USER_FULLSCALE1_MAXIMUM)
#define APP_DB_U_ADCFS1_DEFAULT (ADC_DB_USER_FULLSCALE1_DEFAULT)
#define APP_DB_U_ADCFS2_MIN (ADC_DB_USER_FULLSCALE2_MINIMUM)
#define APP_DB_U_ADCFS2_MAX (ADC_DB_USER_FULLSCALE2_MAXIMUM)
#define APP_DB_U_ADCFS2_DEFAULT (ADC_DB_USER_FULLSCALE2_DEFAULT)

#define APP_DB_U_ADCRES_STR "adcresolution"
#define APP_DB_U_ADCRES_NUM (ADC_MAX_NUM_INTERVAL)
#define APP_DB_U_ADCRES0_MIN (ADC_DB_USER_RESOLUTION0_MINIMUM)
#define APP_DB_U_ADCRES0_MAX (ADC_DB_USER_RESOLUTION0_MAXIMUM)
#define APP_DB_U_ADCRES0_DEFAULT (ADC_DB_USER_RESOLUTION0_DEFAULT)
#define APP_DB_U_ADCRES1_MIN (ADC_DB_USER_RESOLUTION1_MINIMUM)
#define APP_DB_U_ADCRES1_MAX (ADC_DB_USER_RESOLUTION1_MAXIMUM)
#define APP_DB_U_ADCRES1_DEFAULT (ADC_DB_USER_RESOLUTION1_DEFAULT)
#define APP_DB_U_ADCRES2_MIN (ADC_DB_USER_RESOLUTION2_MINIMUM)
#define APP_DB_U_ADCRES2_MAX (ADC_DB_USER_RESOLUTION2_MAXIMUM)
#define APP_DB_U_ADCRES2_DEFAULT (ADC_DB_USER_RESOLUTION2_DEFAULT)

#define APP_DB_U_ADCUNITS_STR "adcunits"
#define APP_DB_U_ADCUNITS_MIN (ADC_DB_USER_UNITS_MINIMUM)
#define APP_DB_U_ADCUNITS_MAX (ADC_DB_USER_UNITS_MAXIMUM)
#define APP_DB_U_ADCUNITS_DEFAULT (ADC_DB_USER_UNITS_DEFAULT)

#define APP_DB_U_ADCCALWEIGHT_STR "adccalweight"
#define APP_DB_U_ADCCALWEIGHT_MIN (ADC_DB_USER_CALWEIGHT_MINIMUM)
#define APP_DB_U_ADCCALWEIGHT_MAX (ADC_DB_USER_CALWEIGHT_MAXIMUM)
#define APP_DB_U_ADCCALWEIGHT_DEFAULT (ADC_DB_USER_CALWEIGHT_DEFAULT)

#define APP_DB_U_ADCADDTARE_STR "adcaddtare"
#define APP_DB_U_ADCADDTARE_MIN (ADC_DB_USER_ADDTARE_MINIMUM)
#define APP_DB_U_ADCADDTARE_MAX (ADC_DB_USER_ADDTARE_MAXIMUM)
#define APP_DB_U_ADCADDTARE_DEFAULT (ADC_DB_USER_ADDTARE_DEFAULT)

#define APP_DB_U_ADCZEROBAND_STR "adczeroband"
#define APP_DB_U_ADCZEROBAND_MIN (0L)
#define APP_DB_U_ADCZEROBAND_MAX (100000L)
#define APP_DB_U_ADCZEROBAND_DEFAULT (ADC_DB_USER_ZEROBAND_DEFAULT)

#define APP_DB_U_ADCMOTION_STR "adcmotion"
#define APP_DB_U_ADCMOTION_MIN (ADC_DB_USER_MOTION_MINIMUM)
#define APP_DB_U_ADCMOTION_MAX (ADC_DB_USER_MOTION_MAXIMUM)
#define APP_DB_U_ADCMOTION_DEFAULT (ADC_DB_USER_MOTION_DEFAULT)

#define APP_DB_U_ADCZEROTRACK_STR "adczerotracking"
#define APP_DB_U_ADCZEROTRACK_MIN (ADC_DB_USER_ZEROTRACKING_MINIMUM)
#define APP_DB_U_ADCZEROTRACK_MAX (ADC_DB_USER_ZEROTRACKING_MAXIMUM)
#define APP_DB_U_ADCZEROTRACK_DEFAULT (ADC_DB_USER_ZEROTRACKING_DEFAULT)

#define APP_DB_U_ADCAVERAGE_STR "adcaverage"
#define APP_DB_U_ADCAVERAGE_MIN (ADC_DB_USER_AVERAGE_MINIMUM)
#define APP_DB_U_ADCAVERAGE_MAX (ADC_DB_USER_AVERAGE_MAXIMUM)
#define APP_DB_U_ADCAVERAGE_DEFAULT (ADC_DB_USER_AVERAGE_DEFAULT)

#define APP_DB_U_ADCRANGEOFZERO_STR "adcrangeofzero"
#define APP_DB_U_ADCRANGEOFZERO_MIN (ADC_DB_USER_RANGEOFZERO_MINIMUM)
#define APP_DB_U_ADCRANGEOFZERO_MAX (ADC_DB_USER_RANGEOFZERO_MAXIMUM)
#define APP_DB_U_ADCRANGEOFZERO_DEFAULT (ADC_DB_USER_RANGEOFZERO_DEFAULT)

#define APP_DB_U_ADCAPPROVMODE_STR "adcapprovalmode"
#define APP_DB_U_ADCAPPROVMODE_MIN (ADC_DB_USER_APPROVALMODE_MINIMUM)
#define APP_DB_U_ADCAPPROVMODE_MAX (ADC_DB_USER_APPROVALMODE_MAXIMUM)
#define APP_DB_U_ADCAPPROVMODE_DEFAULT (ADC_APPROVAL_MODE_NONE)

#define APP_DB_U_ADCZEROONSTART_STR "adczeroonstartup"
#define APP_DB_U_ADCZEROONSTART_MIN (ADC_DB_USER_ZEROONSTART_MINIMUM)
#define APP_DB_U_ADCZEROONSTART_MAX (ADC_DB_USER_ZEROONSTART_MAXIMUM)
#define APP_DB_U_ADCZEROONSTART_DEFAULT (ADC_DB_USER_ZEROONSTART_DEFAULT)

#define APP_DB_U_ADC4WIRE_STR "adc4wiremode"
#define APP_DB_U_ADC4WIRE_MIN (ADC_DB_USER_WIRE4MODE_MINIMUM)
#define APP_DB_U_ADC4WIRE_MAX (ADC_DB_USER_WIRE4MODE_MAXIMUM)
#define APP_DB_U_ADC4WIRE_DEFAULT (ADC_DB_USER_WIRE4MODE_DEFAULT)

#define APP_DB_U_ADCTEMPCALZERO_STR "adctempcalzero"
#define APP_DB_U_ADCTEMPCALZERO_MIN (ADC_DB_USER_TEMPCALZERO_MINIMUM)
#define APP_DB_U_ADCTEMPCALZERO_MAX (ADC_DB_USER_TEMPCALZERO_MAXIMUM)
#define APP_DB_U_ADCTEMPCALZERO_DEFAULT (ADC_DB_USER_TEMPCALZERO_DEFAULT)

#define APP_DB_U_ADCHIRES_STR "adchires"
#define APP_DB_U_ADCHIRES_MIN (ADC_DB_USER_HIRES_MINIMUM)
#define APP_DB_U_ADCHIRES_MAX (ADC_DB_USER_HIRES_MAXIMUM)
#define APP_DB_U_ADCHIRES_DEFAULT (ADC_DB_USER_HIRES_DEFAULT)

#define APP_DB_U_ADCQAENABLE_STR "adcqaenable"
#define APP_DB_U_ADCQAENABLE_MIN (ADC_DB_USER_QAENABLE_MINIMUM)
#define APP_DB_U_ADCQAENABLE_MAX (ADC_DB_USER_QAENABLE_MAXIMUM)
#define APP_DB_U_ADCQAENABLE_DEFAULT (ADC_DB_USER_QAENABLE_DEFAULT)

#define APP_DB_U_ADCQADAY_STR "adcqaday"
#define APP_DB_U_ADCQADAY_MIN (ADC_DB_USER_QADAY_MINIMUM)
#define APP_DB_U_ADCQADAY_MAX (ADC_DB_USER_QADAY_MAXIMUM)
#define APP_DB_U_ADCQADAY_DEFAULT (ADC_DB_USER_QADAY_DEFAULT)

#define APP_DB_U_ADCQAMONTH_STR "adcqamonth"
#define APP_DB_U_ADCQAMONTH_MIN (ADC_DB_USER_QAMONTH_MINIMUM)
#define APP_DB_U_ADCQAMONTH_MAX (ADC_DB_USER_QAMONTH_MAXIMUM)
#define APP_DB_U_ADCQAMONTH_DEFAULT (ADC_DB_USER_QAMONTH_DEFAULT)

#define APP_DB_U_ADCQAYEAR_STR "adcqayear"
#define APP_DB_U_ADCQAYEAR_MIN (ADC_DB_USER_QAYEAR_MINIMUM)
#define APP_DB_U_ADCQAYEAR_MAX (ADC_DB_USER_QAYEAR_MAXIMUM)
#define APP_DB_U_ADCQAYEAR_DEFAULT (ADC_DB_USER_QAYEAR_DEFAULT)

#define APP_DB_U_ADCCCNTEPCAL_STR "adcntepcalibrationcounter"
#define APP_DB_U_ADCCCNTEPCAL_MIN (ADC_DB_USER_CALCNTNTEPCALIBRATION_MINIMUM)
#define APP_DB_U_ADCCCNTEPCAL_MAX (ADC_DB_USER_CALCNTNTEPCALIBRATION_MAXIMUM)
#define APP_DB_U_ADCCCNTEPCAL_DEFAULT (ADC_DB_USER_CALCNTNTEPCALIBRATION_DEFAULT)

#define APP_DB_U_ADCCCNTEPSET_STR "adcntepsettingscounter"
#define APP_DB_U_ADCCCNTEPSET_MIN (ADC_DB_USER_CALCNTNTEPSETTINGS_MINIMUM)
#define APP_DB_U_ADCCCNTEPSET_MAX (ADC_DB_USER_CALCNTNTEPSETTINGS_MAXIMUM)
#define APP_DB_U_ADCCCNTEPSET_DEFAULT (ADC_DB_USER_CALCNTNTEPSETTINGS_DEFAULT)

#define APP_DB_U_ADCCCOIML_STR "adcoimlcalibrationcounter"
#define APP_DB_U_ADCCCOIML_MIN (ADC_DB_USER_CALCNTOIML_MINIMUM)
#define APP_DB_U_ADCCCOIML_MAX (ADC_DB_USER_CALCNTOIML_MAXIMUM)
#define APP_DB_U_ADCCCOIML_DEFAULT (ADC_DB_USER_CALCNTOIML_DEFAULT)

#define APP_DB_U_ADCCALFLAG_STR "adccalflag"
#define APP_DB_U_ADCCALFLAG_MIN (ADC_DB_USER_CALFLAG_MINIMUM)
#define APP_DB_U_ADCCALFLAG_MAX (ADC_DB_USER_CALFLAG_MAXIMUM)
#define APP_DB_U_ADCCALFLAG_DEFAULT (ADC_DB_USER_CALFLAG_DEFAULT)

#define APP_DB_U_ADCCHKSUM_STR "adcchksum"
#define APP_DB_U_ADCCHKSUM_MIN ADC_DB_USER_CHKSUM_MINIMUM
#define APP_DB_U_ADCCHKSUM_MAX ADC_DB_USER_CHKSUM_MAXIMUM
#define APP_DB_U_ADCCHKSUM_DEFAULT ADC_DB_USER_CHKSUM_DEFAULT

/*
 * Application database settings
 */


#define APP_DB_U_SWVER_STR "softwareversion"
#define APP_DB_U_SWVER_MIN 0
#define APP_DB_U_SWVER_MAX UINT32_MAX
#define APP_DB_U_SWVER_DEFAULT APP_VERSION_INT

#define APP_DB_U_REARENTRY_STR "rearentry"
#define APP_DB_U_REARENTRY_MIN BOOLEAN_MIN
#define APP_DB_U_REARENTRY_MAX BOOLEAN_MAX
#define APP_DB_U_REARENTRY_DEFAULT 0

#define APP_DB_U_SNSCHECK_STR "sensecheck"
#define APP_DB_U_SNSCHECK_MIN BOOLEAN_MIN
#define APP_DB_U_SNSCHECK_MAX BOOLEAN_MAX
#define APP_DB_U_SNSCHECK_DEFAULT 1

#define APP_DB_U_BUZENABLE_STR "buzzerenable"
#define APP_DB_U_BUZENABLE_MIN BOOLEAN_MIN
#define APP_DB_U_BUZENABLE_MAX BOOLEAN_MAX
#define APP_DB_U_BUZENABLE_DEFAULT 1

#define APP_DB_U_BRIGHTNESS_STR "brightness"
#define APP_DB_U_BRIGHTNESS_MIN 0
#define APP_DB_U_BRIGHTNESS_MAX 100
#define APP_DB_U_BRIGHTNESS_DEFAULT 100

#define APP_DB_U_SAFEPASS_STR "safepasscode"
#define APP_DB_U_SAFEPASS_MIN 0
#define APP_DB_U_SAFEPASS_MAX (DISPLAY_INT_MAX)
#define APP_DB_U_SAFEPASS_DEFAULT 0

#define APP_DB_U_FULLPASS_STR "fullpasscode"
#define APP_DB_U_FULLPASS_MIN 0
#define APP_DB_U_FULLPASS_MAX (DISPLAY_INT_MAX)
#define APP_DB_U_FULLPASS_DEFAULT 0

#define APP_DB_U_SETPACTIVE_STR "setpactive"
#define APP_DB_U_SETPACTIVE_MIN 0
#define APP_DB_U_SETPACTIVE_MAX (SETP_TYPE_NUM - 1)
#define APP_DB_U_SETPACTIVE_DEFAULT DB_SETP_DEFAULT_TYPE

#define APP_DB_U_SETPALARM_STR "setpalarm"
#define APP_DB_U_SETPALARM_MIN 0
#define APP_DB_U_SETPALARM_MAX (SETP_ALARM_NUM - 1)
#define APP_DB_U_SETPALARM_DEFAULT DB_SETP_DEFAULT_ALARM

#define APP_DB_U_SETPLOCK_STR "setplock"
#define APP_DB_U_SETPLOCK_MIN 0
#define APP_DB_U_SETPLOCK_MAX (SETP_LOCK_NUM - 1)
#define APP_DB_U_SETPLOCK_DEFAULT DB_SETP_DEFAULT_LOCK

#define APP_DB_U_SETPSRC_STR "setpsource"
#define APP_DB_U_SETPSRC_MIN 0
#define APP_DB_U_SETPSRC_MAX (SETP_SOURCE_NUM - 1)
#define APP_DB_U_SETPSRC_DEFAULT DB_SETP_DEFAULT_SOURCE

#define APP_DB_U_SETPDIRN_STR "setpdirection"
#define APP_DB_U_SETPDIRN_MIN 0
#define APP_DB_U_SETPDIRN_MAX (SETP_DIRECTION_NUM - 1)
#define APP_DB_U_SETPDIRN_DEFAULT DB_SETP_DEFAULT_DIRECTION

#define APP_DB_U_SETPLOGIC_STR "setplogic"
#define APP_DB_U_SETPLOGIC_MIN 0
#define APP_DB_U_SETPLOGIC_MAX (SETP_LOGIC_NUM - 1)
#define APP_DB_U_SETPLOGIC_DEFAULT DB_SETP_DEFAULT_LOGIC

#define APP_DB_U_SETPTIMING_STR "setptiming"
#define APP_DB_U_SETPTIMING_MIN 0
#define APP_DB_U_SETPTIMING_MAX (SETP_TIMING_NUM - 1)
#define APP_DB_U_SETPTIMING_DEFAULT DB_SETP_DEFAULT_TIMING

#define APP_DB_U_SETPRESET_STR "setpreset"
#define APP_DB_U_SETPRESET_MIN 0
#define APP_DB_U_SETPRESET_MAX (SETP_RESET_NUM - 1)
#define APP_DB_U_SETPRESET_DEFAULT DB_SETP_DEFAULT_RESET

#define APP_DB_U_SETPJOGON_STR "setpjogon"
#define APP_DB_U_SETPJOGON_MIN 100
#define APP_DB_U_SETPJOGON_MAX 60000
#define APP_DB_U_SETPJOGON_DEFAULT 100

#define APP_DB_U_SETPJOGOFF_STR "setpjogoff"
#define APP_DB_U_SETPJOGOFF_MIN 100
#define APP_DB_U_SETPJOGOFF_MAX 60000
#define APP_DB_U_SETPJOGOFF_DEFAULT 100

#define APP_DB_U_SETPJOGSET_STR "setpjogset"
#define APP_DB_U_SETPJOGSET_MIN 1
#define APP_DB_U_SETPJOGSET_MAX 20
#define APP_DB_U_SETPJOGSET_DEFAULT 1

#define APP_DB_U_SETPFEEDER_STR "setpfeeder"
#define APP_DB_U_SETPFEEDER_MIN 0
#define APP_DB_U_SETPFEEDER_MAX 1
#define APP_DB_U_SETPFEEDER_DEFAULT 0

#define APP_DB_U_SETPDELAYCHECK_STR "setpdelaycheck"
#define APP_DB_U_SETPDELAYCHECK_MIN 0
#define APP_DB_U_SETPDELAYCHECK_MAX 60000
#define APP_DB_U_SETPDELAYCHECK_DEFAULT 0

#define APP_DB_U_SETPMAXSET_STR "setpmaxset"
#define APP_DB_U_SETPMAXSET_MIN 0
#define APP_DB_U_SETPMAXSET_MAX 20
#define APP_DB_U_SETPMAXSET_DEFAULT 0

#define APP_DB_U_FRNTKEYPERMIS_STR "frontkeypermissions"
#define APP_DB_U_FRNTKEYPERMIS_MIN 0
#define APP_DB_U_FRNTKEYPERMIS_MAX (NUM_FRONTKEY_PERMIS - 1)
#define APP_DB_U_FRNTKEYPERMIS_DEFAULT FRONTKEY_PERMIS_ENABLED

#define APP_DB_U_FRNTKEYFN_STR "frontkeyfunction"
#define APP_DB_U_FRNTKEYFN_MIN 0
#define APP_DB_U_FRNTKEYFN_MAX (NUM_FRONTFUNC - 1)

#define APP_DB_U_REMKEYFN_STR "remotekeyfunction"
#define APP_DB_U_REMKEYFN_MIN 0
#define APP_DB_U_REMKEYFN_MAX (NUM_REMOTE_KEY_FUNCTIONS - 1)

#define APP_DB_U_ANALOUTTYPE_STR "analogoutputtype"
#define APP_DB_U_ANALOUTTYPE_MIN 0
#define APP_DB_U_ANALOUTTYPE_MAX (ANALOG_TYPE_NUM - 1)
#define APP_DB_U_ANALOUTTYPE_DEFAULT ANALOG_TYPE_CURRENT

#define APP_DB_U_ANALOUTSRC_STR "analogoutputsource"
#define APP_DB_U_ANALOUTSRC_MIN 0
#define APP_DB_U_ANALOUTSRC_MAX (ANALOG_SOURCE_NUM - 1)
#define APP_DB_U_ANALOUTSRC_DEFAULT ANALOG_SOURCE_DISPLAY

#define APP_DB_U_ANALOUTSPAN_STR "analogoutputspan"
#define APP_DB_U_ANALOUTSPAN_MIN (-ANALOG_OUTPUT_CAL_LIMIT_COUNTS)
#define APP_DB_U_ANALOUTSPAN_MAX (ANALOG_OUTPUT_CAL_LIMIT_COUNTS)
#define APP_DB_U_ANALOUTSPAN_DEFAULT 0

#define APP_DB_U_ANALOUTZERO_STR "analogoutputzero"
#define APP_DB_U_ANALOUTZERO_MIN (-ANALOG_OUTPUT_CAL_LIMIT_COUNTS)
#define APP_DB_U_ANALOUTZERO_MAX (ANALOG_OUTPUT_CAL_LIMIT_COUNTS)
#define APP_DB_U_ANALOUTZERO_DEFAULT 0

#define APP_DB_U_ANALOUTABS_STR "analogoutputabsolute"
#define APP_DB_U_ANALOUTABS_MIN BOOLEAN_MIN
#define APP_DB_U_ANALOUTABS_MAX BOOLEAN_MAX
#define APP_DB_U_ANALOUTABS_DEFAULT 0

#define APP_DB_U_ANALOUTCLIP_STR "analogoutputclip"
#define APP_DB_U_ANALOUTCLIP_MIN BOOLEAN_MIN
#define APP_DB_U_ANALOUTCLIP_MAX BOOLEAN_MAX
#define APP_DB_U_ANALOUTCLIP_DEFAULT 0

#define APP_DB_U_DSDENABLE_STR "dsdenable"
#define APP_DB_U_DSDENABLE_MIN BOOLEAN_MIN
#define APP_DB_U_DSDENABLE_MAX BOOLEAN_MAX
#define APP_DB_U_DSDENABLE_DEFAULT 0

#define APP_DB_U_DSDAUTOPURGE_STR "dsdautopurge"
#define APP_DB_U_DSDAUTOPURGE_MIN BOOLEAN_MIN
#define APP_DB_U_DSDAUTOPURGE_MAX BOOLEAN_MAX
#define APP_DB_U_DSDAUTOPURGE_DEFAULT 0

#define APP_DB_U_SERADDRESS_STR "serialaddress"
#define APP_DB_U_SERADDRESS_MIN 0
#define APP_DB_U_SERADDRESS_MAX 31
#define APP_DB_U_SERADDRESS_DEFAULT 31

#define APP_DB_U_SERTYPE_STR "serialtype"
#define APP_DB_U_SERTYPE_MIN 0
#define APP_DB_U_SERTYPE_TXONLY_MAX (NUM_DB_SERIAL_TYPES_TXONLY - 1)
#define APP_DB_U_SERTYPE_BASIC_MAX (NUM_DB_SERIAL_TYPES_BASIC - 1)
#define APP_DB_U_SERTYPE_FULL_MAX (NUM_DB_SERIAL_TYPES - 1)

#define APP_DB_U_SERBAUD_STR "serialbaud"
#define APP_DB_U_SERBAUD_MIN 0
#define APP_DB_U_SERBAUD_MAX (NUM_DB_BAUD_SETTINGS - 1)
#define APP_DB_U_SERBAUD_DEFAULT DB_SERIAL_DEFAULT_BAUD_INDEX
#define APP_DB_U_SERBAUD_DEFAULT_RL DB_SERIAL_DEFAULT_BAUD_INDEX_RL

#define APP_DB_U_SERBITS_STR "serialbits"
#define APP_DB_U_SERBITS_MIN 0
#define APP_DB_U_SERBITS_MAX UINT8_MAX
#define APP_DB_U_SERBITS_DEFAULT DB_SERIAL_DEFAULT

#define APP_DB_U_SERUDPO1TYPE_STR "serialudpo1type"
#define APP_DB_U_SERUDPO1TYPE_DEFAULT DB_SERIAL_DEFAULT_TYPE_UDPO1

#define APP_DB_U_SERUDPO2TYPE_STR "serialudpo2type"
#define APP_DB_U_SERUDPO2TYPE_DEFAULT DB_SERIAL_DEFAULT_TYPE_UDPO2

#define APP_DB_U_SERUDPI1TYPE_STR "serialudpi1type"
#define APP_DB_U_SERUDPI1TYPE_DEFAULT DB_SERIAL_DEFAULT_TYPE_UDPI1

#define APP_DB_U_SERUDPI2TYPE_STR "serialudpi2type"
#define APP_DB_U_SERUDPI2TYPE_DEFAULT DB_SERIAL_DEFAULT_TYPE_UDPI2

#define APP_DB_U_SERTCPI2TYPE_STR "serialtcpi2type"
#define APP_DB_U_SERTCPI2TYPE_MIN DB_SERIAL_TYPE_AUTO_LO
#define APP_DB_U_SERTCPI2TYPE_MAX DB_SERIAL_TYPE_AUTO_HI
#define APP_DB_U_SERTCPI2TYPE_DEFAULT DB_SERIAL_DEFAULT_TYPE_TCPI2

#define APP_DB_U_SERUDPO1DEST_STR "serialudpo1dest"
#define APP_DB_U_SERUDPO2DEST_STR "serialudpo2dest"
#define APP_DB_U_SERUDPODEST_MIN 0
#define APP_DB_U_SERUDPODEST_MAX UINT32_MAX
#define APP_DB_U_SERUDPODEST_DEFAULT DB_SERIAL_DEFAULT_UDPO_DEST

#define APP_DB_U_SERUDPO1PORT_STR "serialudpo1port"
#define APP_DB_U_SERUDPO1PORT_MIN DB_SERIAL_UDP_PORT_LOW_LIMIT
#define APP_DB_U_SERUDPO1PORT_MAX DB_SERIAL_UDP_PORT_HIGH_LIMIT
#define APP_DB_U_SERUDPO1PORT_DEFAULT DB_SERIAL_DEFAULT_UDPO_PORT

#define APP_DB_U_SERUDPO2PORT_STR "serialudpo2port"
#define APP_DB_U_SERUDPO2PORT_MIN DB_SERIAL_UDP_PORT_LOW_LIMIT
#define APP_DB_U_SERUDPO2PORT_MAX DB_SERIAL_UDP_PORT_HIGH_LIMIT
#define APP_DB_U_SERUDPO2PORT_DEFAULT DB_SERIAL_DEFAULT_UDPO_PORT

#define APP_DB_U_SERTCPI1PORT_STR "serial5tcpport"
#define APP_DB_U_SERTCPI1PORT_MIN DB_SERIAL_TCPPORT_LOW_LIMIT
#define APP_DB_U_SERTCPI1PORT_MAX DB_SERIAL_TCPPORT_HIGH_LIMIT
#define APP_DB_U_SERTCPI1PORT_DEFAULT DB_SERIAL_DEFAULT_TCPI1_PORT

#define APP_DB_U_SERTCPI2PORT_STR "serial6tcpport"
#define APP_DB_U_SERTCPI2PORT_MIN DB_SERIAL_TCPPORT_LOW_LIMIT
#define APP_DB_U_SERTCPI2PORT_MAX DB_SERIAL_TCPPORT_HIGH_LIMIT
#define APP_DB_U_SERTCPI2PORT_DEFAULT DB_SERIAL_DEFAULT_TCPI2_PORT

#define APP_DB_U_SERUDPI1PORT_STR "serialudpi1port"
#define APP_DB_U_SERUDPI1PORT_MIN DB_SERIAL_UDP_PORT_LOW_LIMIT
#define APP_DB_U_SERUDPI1PORT_MAX DB_SERIAL_UDP_PORT_HIGH_LIMIT
#define APP_DB_U_SERUDPI1PORT_DEFAULT DB_SERIAL_DEFAULT_UDPI_PORT

#define APP_DB_U_SERUDPI2PORT_STR "serialudpi2port"
#define APP_DB_U_SERUDPI2PORT_MIN DB_SERIAL_UDP_PORT_LOW_LIMIT
#define APP_DB_U_SERUDPI2PORT_MAX DB_SERIAL_UDP_PORT_HIGH_LIMIT
#define APP_DB_U_SERUDPI2PORT_DEFAULT DB_SERIAL_DEFAULT_UDPI_PORT

#define APP_DB_U_SERTCPI1TIMEOUT_STR "serialtcpi1timeout"
#define APP_DB_U_SERTCPI1TIMEOUT_MIN DB_SERIAL_TCPTIMEOUT_LOW_LIMIT
#define APP_DB_U_SERTCPI1TIMEOUT_MAX DB_SERIAL_TCPTIMEOUT_HIGH_LIMIT
#define APP_DB_U_SERTCPI1TIMEOUT_DEFAULT DB_SERIAL_DEFAULT_TCPTIMEOUT

#define APP_DB_U_SERTCPI2TIMEOUT_STR "serialtcpi2timeout"
#define APP_DB_U_SERTCPI2TIMEOUT_MIN DB_SERIAL_TCPTIMEOUT_LOW_LIMIT
#define APP_DB_U_SERTCPI2TIMEOUT_MAX DB_SERIAL_TCPTIMEOUT_HIGH_LIMIT
#define APP_DB_U_SERTCPI2TIMEOUT_DEFAULT DB_SERIAL_DEFAULT_TCPTIMEOUT

#define APP_DB_U_SERSTARTCHAR_STR "serialstartchar"
#define APP_DB_U_SERSTARTCHAR_MIN 0
#define APP_DB_U_SERSTARTCHAR_MAX UINT8_MAX
#define APP_DB_U_SERSTARTCHAR_DEFAULT '\x02'

#define APP_DB_U_SERENDCHAR1_STR "serialendchar1"
#define APP_DB_U_SERENDCHAR1_MIN 0
#define APP_DB_U_SERENDCHAR1_MAX UINT8_MAX
#define APP_DB_U_SERENDCHAR1_DEFAULT '\x03'

#define APP_DB_U_SERENDCHAR2_STR "serialendchar2"
#define APP_DB_U_SERENDCHAR2_MIN 0
#define APP_DB_U_SERENDCHAR2_MAX UINT8_MAX
#define APP_DB_U_SERENDCHAR2_DEFAULT '\0'

#define APP_DB_U_SERPRINTMODE_STR "serialprintermode"
#define APP_DB_U_SERPRINTMODE_MIN 0
#define APP_DB_U_SERPRINTMODE_MAX (NUM_PRINTER_MODES - 1)
#define APP_DB_U_SERPRINTMODE_DEFAULT P_SINGLE

#define APP_DB_U_SERTICKETROW_STR "serialticketrows"
#define APP_DB_U_SERTICKETROW_MIN 0
#define APP_DB_U_SERTICKETROW_MAX 10
#define APP_DB_U_SERTICKETROW_DEFAULT 0

#define APP_DB_U_SERTICKETCOL_STR "serialticketcolumns"
#define APP_DB_U_SERTICKETCOL_MIN 0
#define APP_DB_U_SERTICKETCOL_MAX 20
#define APP_DB_U_SERTICKETCOL_DEFAULT 0

#define APP_DB_U_SERID_STR "serialid"
#define APP_DB_U_SERID_MAXLEN APP_DB_ID_STR_LEN
#define APP_DB_U_SERID_DEFAULT "C500"

#define APP_DB_U_SERCOF_STR "serialcof"
#define APP_DB_U_SERCOF_MIN 0
#define APP_DB_U_SERCOF_MAX (COF_SETTING_NUM - 1)
#define APP_DB_U_SERCOF_DEFAULT COF_SETTING3_ASCII_WEIGHT

#define APP_DB_U_SERAOFTYPE_STR "serialaoftype"
#define APP_DB_U_SERAOFTYPE_MIN 0
#define APP_DB_U_SERAOFTYPE_MAX (AUTO_OUT_FMT_NUM - 1)
#define APP_DB_U_SERAOFTYPE_DEFAULT AUTO_OUT_FMT_RINAUTO

#define APP_DB_U_SERAOFSRC_STR "serialaofsource"
#define APP_DB_U_SERAOFSRC_MIN 0
#define APP_DB_U_SERAOFSRC_MAX (AUTO_OUT_SRC_NUM - 1)
#define APP_DB_U_SERAOFSRC_DEFAULT AUTO_OUT_SRC_DISP

#define APP_DB_U_SERPRINTTEXT_STR "serialprinttext"
#define APP_DB_U_SERPRINTTEXT_NUMELEMS (APP_DB_PRINT_HDR_LINES)
#define APP_DB_U_SERPRINTTEXT_MAXLEN (APP_DB_PRINT_HDR_LEN)

#define APP_DB_U_SERPRINTFMT_STR "serialprintformat"
#define APP_DB_U_SERPRINTFMT_MAXLEN APP_DB_PRINT_FMT_LEN
#define APP_DB_U_SERPRINTFMT_DEFAULT ""

#define APP_DB_U_SERAUTOFMT_STR "serialautoformat"
#define APP_DB_U_SERAUTOFMT_MAXLEN APP_DB_AUTO_FMT_LEN
#define APP_DB_U_SERAUTOFMT_DEFAULT ""

#define APP_DB_U_MODBUSENDIAN_STR "modbusendian"
#define APP_DB_U_MODBUSENDIAN_MIN 0
#define APP_DB_U_MODBUSENDIAN_MAX (RINMODBUS_ENDIAN_NUM - 1)
#define APP_DB_U_MODBUSENDIAN_DEFAULT RINMODBUS_ENDIAN_BIG

#define APP_DB_U_MODBUSTCPPORT_STR "modbustcpport"
#define APP_DB_U_MODBUSTCPPORT_MIN MODBUS_TCP_DEFAULT_PORT      /* MODBUS_TCP_DEFAULT_PORT is lower than DB_SERIAL_TCPPORT_LOW_LIMIT */
#define APP_DB_U_MODBUSTCPPORT_MAX DB_SERIAL_TCPPORT_HIGH_LIMIT
#define APP_DB_U_MODBUSTCPPORT_DEFAULT MODBUS_TCP_DEFAULT_PORT

#define APP_DB_U_PRINTINTERLOCK_STR "serialprintinterlock"
#define APP_DB_U_PRINTINTERLOCK_MIN 0
#define APP_DB_U_PRINTINTERLOCK_MAX (NUM_PRINT_INTERLOCK - 1)
#define APP_DB_U_PRINTINTERLOCK_DEFAULT PRINT_INTERLOCK_NONE

#define APP_DB_U_TIME_FMT_STR "timeformat"
#define APP_DB_U_TIME_FMT_MIN 0
#define APP_DB_U_TIME_FMT_MAX (NUM_TIME_FMTS - 1)
#define APP_DB_U_TIME_FMT_DEFAULT 0

#define APP_DB_U_DATE_FMT_STR "dateformat"
#define APP_DB_U_DATE_FMT_MIN 0
#define APP_DB_U_DATE_FMT_MAX (NUM_DATE_FMTS - 1)
#define APP_DB_U_DATE_FMT_DEFAULT 0

#define APP_DB_U_UNITS_SEC_STR "secondaryunits"
#define APP_DB_U_UNITS_SEC_MIN 0
#define APP_DB_U_UNITS_SEC_MAX (ALT_UNITS_CUSTOM)
#define APP_DB_U_UNITS_SEC_DEFAULT 0

#define APP_DB_U_UNITS_TER_STR "tertiaryunits"
#define APP_DB_U_UNITS_TER_MIN 0
#define APP_DB_U_UNITS_TER_MAX (ALT_UNITS_OUNCES)
#define APP_DB_U_UNITS_TER_DEFAULT 0

#define APP_DB_U_UNITS_CONV_STR "unitconversionfactor"
#define APP_DB_U_UNITS_CONV_MIN 1
#define APP_DB_U_UNITS_CONV_MAX 999999
#define APP_DB_U_UNITS_CONV_DEFAULT 1000

#define APP_DB_U_SHOW_LBOZ_STR "showlboz"
#define APP_DB_U_SHOW_LBOZ_MIN BOOLEAN_MIN
#define APP_DB_U_SHOW_LBOZ_MAX BOOLEAN_MAX
#define APP_DB_U_SHOW_LBOZ_DEFAULT 0

/** Minimum limits for the serial type for each port */
extern const uint8_t serial_type_minimums[NUM_DB_SERIAL_SETTINGS];

/** Maximum limits for the serial type for each port */
extern const uint8_t serial_type_maximums[NUM_DB_SERIAL_SETTINGS];


/** User application database structure (container of database variables) */
extern struct s_db_app_user db_app_user;


/** User application database description (used by db library) */
extern struct s_db_database app_user_database;


/**
 * Write the user database to file
 *
 * @param write_ass_copy specifies whether the adcSuperStructure copy of the adcUserDatabase should be written
 * @param lock specifies whether adcSuperStructure should be locked before writing
 */
bool app_write_user_database(bool write_ass_copy, bool lock);


/** Read the user database from file */
bool app_read_user_database(void);


/**
 * Default the user database
 *
 * @param default_calibration_also if true will also default the calibration (db_app_user.adc_user_database)
 */
bool app_default_user_database(bool default_calibration_also);


/** Initialise the user database */
bool app_init_user_database(void);


/** Destroy the user database */
bool app_destroy_user_database(void);


#endif /* APP_DB_USER_H_ */
