/*
 * app_devicemount.c
 *
 *  Created on: Mar 31, 2014
 *      Author: ian
 */
#include <stdbool.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <we_config.h>
#include "devicemounter.h"
#include "dm_event.h"
#include "dm_partition.h"
#include "dm_printer.h"
#include "dm_serial.h"
#include "iface_file.h"
#include "we_debug.h"
#include "serial_ports.h"
#include "devices.h"
#include "iface_usb.h"
#include "serial_usb_host.h"

#include "app_devicemount.h"


/** Device mounter for handling addition/removal of devices */
static struct s_devicemounter *we2111_device_mounter = NULL;

#define DM_NO_HANDLER -1

/** Partition device mounter partition handle */
static int dm_partition_handle = DM_NO_HANDLER;
static int dm_event_handle = DM_NO_HANDLER;
static int dm_printer_handle = DM_NO_HANDLER;
static int dm_serial_handle = DM_NO_HANDLER;


/** Initialise the device mounter for the WE2111 */
void we2111_device_mounter_init(const char *mount_location, const char *device_mounter_notification_pipe)
{
  struct stat statbuf;
  bool have_folder;

  /* Setup the WE2111 device mounter, to monitor the fifo for device appearance/disappearance */
  we2111_device_mounter = devicemounter_init(device_mounter_notification_pipe);
  if (NULL != we2111_device_mounter)
  {
    /* Create folder if it doesn't exist */
    have_folder = false;
    if (stat(mount_location, &statbuf) < 0)
    {
      /* Try making the folder TODO this is not recursive (ie it will only make the bottom level directory) */
      have_folder = (0 == mkdir(mount_location, 0755));
    }
    else
    {
      /* Exists, but is it a dir? */
      have_folder = S_ISDIR(statbuf.st_mode);
    }

    /* Only add the partition mounter if we've got a folder to mount into */
    if (have_folder)
    {
      /*
       * Add a partition (ie hdd/flash drive) mounter.
       * We make any mounted vfat disks writable by everyone, because vfat doesn't support permissions,
       * and we both the C500 main app running as "root" and lua running as "user" to be able to write to the disk
       */
      dm_partition_handle = dm_partition_init(we2111_device_mounter, mount_location, "vfat", "sync,nosuid,noexec,dmask=000,fmask=111", disk_mount_callback, NULL);
      if (dm_partition_handle < 0)
      {
        rdebug_printf(LOG_ERR, "main: add partition mounter failed\n");
      }
    }
    else
    {
      rdebug_printf(LOG_ERR, "main: mount folder (%s) doesn't exist and can't be created, or isn't a folder\n", mount_location);
    }

    /* Add a printer mount handler */
    dm_printer_handle = dm_printer_init(we2111_device_mounter, &dev_ser9_usblp, iface_usb_printer_callback, NULL);
    if (dm_printer_handle < 0)
    {
      rdebug_printf(LOG_ERR, "main: add printer mounter failed\n");
    }

    /* Add an event device mount handler */
    dm_event_handle = dm_event_init(we2111_device_mounter, &dev_kbd_event, iface_usb_event_callback, NULL);
    if (dm_event_handle < 0)
    {
      rdebug_printf(LOG_ERR, "main: add event mounter failed\n");
    }

    /* Add a serial port device mount handler */
    dm_serial_handle = dm_serial_init(we2111_device_mounter, usb_host_serial_device_callback, NULL);
    if (dm_serial_handle < 0)
    {
      rdebug_printf(LOG_ERR, "main: add serial port mounter failed\n");
    }

    /* Start the device mounter, now everything is registered */
    if (devicemounter_run(we2111_device_mounter) < 0)
    {
      rdebug_printf(LOG_ERR, "main: device mounter run failed\n");
    }
  }
  else
  {
    rdebug_printf(LOG_ERR, "main: device mounter initialisation failed\n");
  }
}


/** Shutdown the device mounter for the WE2111 */
void we2111_device_mounter_destroy(void)
{
  (void) devicemounter_destroy(we2111_device_mounter);
}


/** Unmount a mounted USB disk */
int we2111_unmount_usbdisk(void)
{
  return devicemounter_handlercommand(we2111_device_mounter, dm_partition_handle, DM_PARTITION_CMD_UNMOUNT, NULL);
}


/** Dismount printer and event devices */
void we2111_device_mounter_dismount_lua_devices(void)
{
  devicemounter_handlercommand(we2111_device_mounter, dm_event_handle, DM_EVENT_CMD_UNREGISTER, NULL);
  devicemounter_handlercommand(we2111_device_mounter, dm_printer_handle, DM_PRINTER_CMD_REMOVE, NULL);
  devicemounter_handlercommand(we2111_device_mounter, dm_serial_handle, DM_SERIAL_CMD_REMOVE, NULL);
}


/** Remove printer and event device handlers */
void we2111_device_mounter_remove_lua_devices(void)
{
  devicemounter_unregisterhandler(we2111_device_mounter, dm_event_handle);
  devicemounter_unregisterhandler(we2111_device_mounter, dm_printer_handle);
  devicemounter_unregisterhandler(we2111_device_mounter, dm_serial_handle);
  dm_event_handle = DM_NO_HANDLER;
  dm_printer_handle = DM_NO_HANDLER;
  dm_serial_handle = DM_NO_HANDLER;
}


