#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <rinweigh/adcdatabase.h>
#include "system.h"
#include "adc_8bit.h"
#include "temperature.h"
#include "meas.h"
#include "kbd_5000.h"
#include "database.h"
#include "editor.h"
#include "iface_sd.h"
#include "buzz5000.h"
#include "resolution.h"
#include "app_db_factory.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "we_debug.h"
#include "status_led.h"
#include "apperror.h"

#include "backgrnd.h"


static long adc_results_converted[NUM_ADC_CHANNELS];


/* ADC readings limits */
#define ADC_PWR_MAX_MV                30000       /* 30 V */
#define ADC_PWR_MIN_MV                9000        /* 9 V */
#define ADC_SNS_PLUS_MAX_MV           8000        /* 8V */
#define ADC_SNS_PLUS_MISSING_MV       3000        /* Pos sense missing > ADC_SNS_PLUS_MISSING_MV */
#define ADC_SNS_MINUS_MISSING_MV      2000        /* Neg missing > ADC_SNS_MINUS_MISSING_MV */
#define ADC_SNS_MINUS_BOTH_MISSING_MV 2400        /* Neg missing > ADC_SNS_MINUS_BOTH_MISSING_MV if Pos sense already missing */


static uint16_t qa_ack_timer;              /* counts 3600 seconds to implement qa ack */
uint8_t qa_timeout;                /* 0 for ok, 1 for cal due 2 for cal due ack */


static void check_temperature(void)
{
  long kelvin;

  kelvin = temperature_get(tuKelvin);

  if ((kelvin > WECFG_TEMPERATURE_MAX_KELVIN || kelvin < WECFG_TEMPERATURE_MIN_KELVIN) && (get_current_errors() & ADC_ERR_TEMPERATURE) == 0)
  {
    set_error(ADC_ERR_TEMPERATURE, "check_temperature: setting error ADC_ERR_TEMPERATURE because current temperature (%ld K) is outside allowable limits %ld K..%ld K\n",
              kelvin, (long) WECFG_TEMPERATURE_MIN_KELVIN, (long) WECFG_TEMPERATURE_MAX_KELVIN);
  }
  else if ((kelvin <= WECFG_TEMPERATURE_MAX_KELVIN && kelvin >= WECFG_TEMPERATURE_MIN_KELVIN) && (get_current_errors() & ADC_ERR_TEMPERATURE) != 0)
  {
    clr_error(ADC_ERR_TEMPERATURE, "check_temperature: clearing error ADC_ERR_TEMPERATURE\n");
  }
}


long get_12V_reading(void)
{
  return adc_results_converted[ADC_CH_VIN];
}


long get_pos_sense_reading(void)
{
  return adc_results_converted[ADC_CH_SNS_PLUS];
}


long get_neg_sense_reading(void)
{
  return adc_results_converted[ADC_CH_SNS_MINUS];
}


void ack_qa_timeout(void)
{
  qa_timeout = 2;
  qa_ack_timer = 3600;
}


bool check_qa_timeout(void)
{
  t_datetime dt;
  bool result;

  /* Check QA enabled */
  if (!adcSuperStructure.adcUserDatabase.qaEnable)
  {
    result = false;
  }
  else if (GetDateTime(&dt))
  {
    /* Get current date, and check against QA date */
    if (dt.year > adcSuperStructure.adcUserDatabase.qaYear)
      result = true;
    else if (dt.year < adcSuperStructure.adcUserDatabase.qaYear)
      result = false;
    else if (dt.month > adcSuperStructure.adcUserDatabase.qaMonth)
      result = true;
    else if (dt.month < adcSuperStructure.adcUserDatabase.qaMonth)
      result = false;
    else if (dt.date > adcSuperStructure.adcUserDatabase.qaDay)
      result = true;
    else if (dt.date < adcSuperStructure.adcUserDatabase.qaDay)
      result = false;
    else
      result = true;
  }
  else
  {
    /* Date read failed */
    result = false;
  }
  return result;
}


static void check_voltages(void)
{
  long rdg;
  long pos_rdg;
  long neg_rdg;

  /* Check upper supply voltage limits */
  rdg = get_12V_reading();
  if (rdg > ADC_PWR_MAX_MV && (get_current_errors() & APP_ERR_HIGH_12V) == 0)
    set_error(APP_ERR_HIGH_12V, "check_voltages: setting error APP_ERR_HIGH_12V because input voltage %ld is over threshold %ld\n", rdg, (long) ADC_PWR_MAX_MV);
  else if (rdg <= ADC_PWR_MAX_MV && (get_current_errors() & APP_ERR_HIGH_12V) != 0)
    clr_error(APP_ERR_HIGH_12V, "check_voltages: clearing error APP_ERR_HIGH_12V\n");

  /* Check lower supply voltage limits */
  if (rdg < ADC_PWR_MIN_MV && (get_current_errors() & APP_ERR_LOW_12V) == 0)
    set_error(APP_ERR_LOW_12V, "check_voltages: setting error APP_ERR_LOW_12V because input voltage %ld is under threshold %ld\n", rdg, (long) ADC_PWR_MIN_MV);
  else if (rdg >= ADC_PWR_MIN_MV && (get_current_errors() & APP_ERR_LOW_12V) != 0)
    clr_error(APP_ERR_LOW_12V, "check_voltages: clearing error APP_ERR_LOW_12V\n");

  /* Check excitation voltages */
  rdg = get_pos_sense_reading();
  if (rdg > ADC_SNS_PLUS_MAX_MV && (get_current_errors() & APP_ERR_HIGH_EXCITATION) == 0)
    set_error(APP_ERR_HIGH_EXCITATION, "check_voltages: setting error APP_ERR_HIGH_EXCITATION because excitation voltage %ld is over threshold %ld\n", rdg, (long) ADC_SNS_PLUS_MAX_MV);
  else if (rdg <= ADC_SNS_PLUS_MAX_MV && (get_current_errors() & APP_ERR_HIGH_EXCITATION) != 0)
    clr_error(APP_ERR_HIGH_EXCITATION, "check_voltages: clearing error APP_ERR_HIGH_EXCITATION\n");

  if (db_app_user.sense_check)
  {
    pos_rdg = get_pos_sense_reading();
    neg_rdg = get_neg_sense_reading();

    /* Positive sense */
    if (pos_rdg < ADC_SNS_PLUS_MISSING_MV && (get_current_errors() & APP_ERR_SENSE_POS) == 0)
    {
      set_error(APP_ERR_SENSE_POS, "check_voltages: setting error APP_ERR_SENSE_POS because positive sense wire missing (rdg %ld < limit %ld)\n", pos_rdg, (long) ADC_SNS_PLUS_MISSING_MV);
    }
    else if (pos_rdg >= ADC_SNS_PLUS_MISSING_MV && (get_current_errors() & APP_ERR_SENSE_POS) != 0)
    {
      clr_error(APP_ERR_SENSE_POS, "check_voltages: clearing error APP_ERR_SENSE_POS\n");
    }

    /* Negative sense */
    if ((pos_rdg < ADC_SNS_PLUS_MISSING_MV && neg_rdg > ADC_SNS_MINUS_BOTH_MISSING_MV) || (neg_rdg > ADC_SNS_MINUS_MISSING_MV))
    {
      if ((get_current_errors() & APP_ERR_SENSE_NEG) == 0)
        set_error(APP_ERR_SENSE_NEG, "check_voltages: setting error APP_ERR_SENSE_NEG because negative sense wire missing (pos (%ld) < limit (%ld) and neg (%ld) > limit (%ld)) or (neg (%ld) > limit (%ld)\n",
                  pos_rdg, (long) ADC_SNS_PLUS_MISSING_MV, neg_rdg, (long) ADC_SNS_MINUS_BOTH_MISSING_MV, neg_rdg, (long) ADC_SNS_MINUS_MISSING_MV);
    }
    else
    {
      if ((get_current_errors() & APP_ERR_SENSE_NEG) != 0)
        clr_error(APP_ERR_SENSE_NEG, "check_voltages: clearing error APP_ERR_SENSE_NEG\n");
    }
  }
  else
  {
    /* Not checking sense lines, so clear any errors */
    if ((get_current_errors() & APP_ERR_SENSE_BOTH) != 0)
      clr_error(APP_ERR_SENSE_BOTH, "check_voltages: clearing error APP_ERR_SENSE_BOTH\n");
  }
}


/** Set error for outside valid resolutions */
static void check_resolution_error(void)
{
  bool res_err;

  res_err = scale_has_resolution_error();
  if (res_err && (get_current_errors() & APP_ERR_RESOLUTION) == 0)
  {
    set_error(APP_ERR_RESOLUTION, "check_resolution_error: setting APP_ERR_RESOLUTION because scale has resolution error\n");
  }
  else if (!res_err && (get_current_errors() & APP_ERR_RESOLUTION) != 0)
  {
    clr_error(APP_ERR_RESOLUTION, "check_resolution_error: clearing error APP_ERR_RESOLUTION\n");
  }
}


/** Main background routine called on a from idle by bf1000 */
void do_background(void)
{
  uint16_t temperature_raw = 0;

  /* Check analog readings */
  adc_read_all(adc_results_converted, &temperature_raw);
  temperature_update(temperature_raw);
  check_voltages();
  check_temperature();

  /* Check valid serial number, ADC library checks other members of factory database like calFlag */
  if (0 == db_app_factory.serial_no && (get_current_errors() & ADC_ERR_FACTORY_DB) == 0)
    set_error(ADC_ERR_FACTORY_DB, "do_background: setting ADC_ERR_FACTORY_DB because serial number is 0\n");
  else if (0 != db_app_factory.serial_no && (get_current_errors() & ADC_ERR_FACTORY_DB) != 0)
    clr_error(ADC_ERR_FACTORY_DB, "do_background: clearing error ADC_ERR_FACTORY_DB\n");

  /* Check that the resolution is in a sensible range */
  check_resolution_error();

  /* Check QA timeout */
  if (qa_ack_timer > 0)
    qa_ack_timer--;
  if (check_qa_timeout())
  {
    if ((qa_timeout == 0) || (qa_ack_timer == 0))
      qa_timeout = 1;
  }
  else
    qa_timeout = 0;
}


void init_background(void)
{
  uint16_t temperature_raw = 0;

  adc_read_all(adc_results_converted, &temperature_raw);
  temperature_update(temperature_raw);
}
