#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <rinhw/rstdio.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "iface_serial.h"
#include "sysclock.h"
#include "database.h"
#include "meas.h"
#include "iface.h"
#include "editor.h"
#include "comms_auto_trigger.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "app_db_runtime_vars.h"
#include "apperror.h"
#include "comms_send_buffer.h"
#include "iface_clock.h"
#include "unitswitch.h"

#include "comms_auto.h"


#define ERROR_MODE_NONE 0
#define ERROR_MODE_BLANK_SPACE 1
#define ERROR_MODE_BLANK_DASH 2


static void auto_send_char(struct s_comms_auto *comms_auto, char ch)
{
  (void) comms_send_buffer_addch(comms_auto->csb, ch);
}


static void auto_send_str(struct s_comms_auto *comms_auto, const char *s)
{
  while (*s != '\0')
  {
    auto_send_char(comms_auto, *s++);
  }
}


static void auto_send_units(struct s_comms_auto *comms_auto, const AdcRdg *rdg)
{
  enum eBuildUnits weunits;

  auto_send_char(comms_auto, ' ');
  weunits = xlat_adc_library_units_to_we2111(rdg->info.unitdp.bit.units);
  auto_send_str(comms_auto, build_units_print[weunits]);
}


static void auto_send_range(struct s_comms_auto *comms_auto, const AdcRdg *rdg)
{
  if (!rdg->info.status.bit.multiRange)
    auto_send_char(comms_auto, ' ');
  else if (0 == rdg->info.status.bit.range)
    auto_send_char(comms_auto, '1');
  else
    auto_send_char(comms_auto, '2');
}


static void auto_send_space3(struct s_comms_auto *comms_auto)
{
  auto_send_str(comms_auto, "   ");
}


static void auto_send_weight(struct s_comms_auto *comms_auto, const AdcRdg *rdg)
{
  #define WEIGHT_STR_LEN 20
  char weight_str[WEIGHT_STR_LEN];

  (void) comms_convert_num_with_dp_pad(weight_str, WEIGHT_STR_LEN, rdg->rdg, rdg->info.unitdp.bit.dp, 8, true, comms_auto->resolution_length);
  auto_send_str(comms_auto, weight_str);
}


static void auto_send_time(struct s_comms_auto *comms_auto)
{
  #define TIME_STR_LEN 12
  char time_str[TIME_STR_LEN];

  to_string_time(time_str, TIME_STR_LEN, &comms_auto->datetime);
  auto_send_str(comms_auto, time_str);
}


static void auto_send_date(struct s_comms_auto *comms_auto)
{
  #define DATE_STR_LEN 11
  char date_str[DATE_STR_LEN];

  to_string_date(date_str, DATE_STR_LEN, &comms_auto->datetime);
  auto_send_str(comms_auto, date_str);
}


static void auto_send_char_upper(struct s_comms_auto *comms_auto, char ch)
{
  if (!comms_auto->token_upper)
    ch = tolower(ch);
  auto_send_char(comms_auto, ch);
}


static void auto_send_str_upper(struct s_comms_auto *comms_auto, char *p)
{
  while (*p != '\0')
  {
    auto_send_char_upper(comms_auto, *p++);
  }
}


/** Something causing an error */
static bool comms_auto_error(struct s_comms_auto *comms_auto, const AdcRdg *rdg)
{
  return (APP_ERR_NONE != comms_auto->sys_errors ||
          rdg->info.status.bit.oload ||
          rdg->info.status.bit.uload);
}


/** Send weight (in finals) for an auto output */
static void auto_sent_weight_token(struct s_comms_auto *comms_auto, const AdcRdg *rdg)
{
  #define TOKSTR_LEN 20
  char str[TOKSTR_LEN];
  char *s;
  long rdgl;
  size_t cnt;
  size_t dp;
  size_t pad;

  rdgl = rdg->rdg;

  /* Work out sign */
  if (comms_auto->token_plus_sign != '\0')
  {
    if (rdgl < 0)
    {
      /* Negative - always send '-' */
      rdgl = -rdgl;
      auto_send_char(comms_auto, '-');
    }
    else
    {
      /* Positive with sign char specified - send specified char */
      auto_send_char(comms_auto, comms_auto->token_plus_sign);
    }
  }
  else
  {
    if (rdgl < 0)
      rdgl = -rdgl;
  }

  /* Work out decimal point position. == 0 means don't print decimal point */
  dp = (0 == comms_auto->token_dp) ? 0 : rdg->info.unitdp.bit.dp;

  /* Convert reading to string */
  snprintf(str, TOKSTR_LEN, "%lu", rdgl);

  /* Work out width */
  pad = comms_auto->token_digits;
  if (comms_auto->token_plus_sign != 0)
  {
    pad--;
  }

  if (comms_auto_error(comms_auto, rdg) && comms_auto->token_error_mode > ERROR_MODE_NONE)
  {
    strcpy(str, "");
    dp = 0;
    pad_str(str, (comms_auto->token_error_mode == ERROR_MODE_BLANK_SPACE) ? ' ' : '-', pad);
  }
  else
  {
    if (dp != 0)
      pad--;
    if (comms_auto->token_digits != 0)
      pad_str(str, '0', pad);
    if (!comms_auto->token_lz)
      clear_leading_zeros(str, dp);
  }

  cnt = strlen(str);
  s = str;
  while (*s != '\0')
  {
    auto_send_char(comms_auto, *s++);
    if ((--cnt == dp) && (dp > 0))
      auto_send_char(comms_auto, comms_auto->token_dp);
  }
}


static void auto_get_reading_from_source(const struct s_comms_auto *comms_auto, uint8_t src, AdcRdg *rdg)
{
  switch (src)
  {
    default:
    case AUTO_OUT_SRC_DISP:
      *rdg = comms_auto->rdg_displayed;
      break;
    case AUTO_OUT_SRC_GROSS:
      *rdg = comms_auto->rdg_gross;
      break;
    case AUTO_OUT_SRC_NET:
      *rdg = comms_auto->rdg_net;
      break;
    case AUTO_OUT_SRC_TOTAL:
      *rdg = comms_auto->rdg_total;
      break;
    case AUTO_OUT_SRC_TARE:
      *rdg = comms_auto->rdg_tare;
      break;
  }
}


/** Send an automatic output weight from a source src */
static void auto_send_weight_from_src(struct s_comms_auto *comms_auto, uint8_t src, AdcRdg *rdg)
{

  /* Get the reading */
  auto_get_reading_from_source(comms_auto, src, rdg);

  /* Send it */
  auto_sent_weight_token(comms_auto, rdg);
}


static char auto_get_rdg_status_GNT_char(const AdcRdg *rdg)
{
  char status;

  if (rdg->info.status.bit.grossOrNet)
  {
    status = (rdg->info.status.bit.gross) ? 'G' : 'N';
  }
  else if (rdg->info.status.bit.tare || rdg->info.status.bit.preTare)
  {
    status = 'T';
  }
  else
  {
    status = ' ';
  }
  return status;
}


/** Get the status character, based on the latest reading status */
static char auto_get_rdg_status_char(struct s_comms_auto *comms_auto, const AdcRdg *rdg)
{
  char status;

  if (APP_ERR_NONE != comms_auto->sys_errors)
  {
    status = 'E';
  }
  else if (rdg->info.status.bit.oload)
  {
    status = 'O';
  }
  else if (rdg->info.status.bit.uload)
  {
    status = 'U';
  }
  else
  {
    status = auto_get_rdg_status_GNT_char(rdg);
  }
  return status;
}


/** This function handles tokens that control formatting, but don't output any data */
static bool auto_custom_fmt_handle_formatting(struct s_comms_auto *comms_auto, const uint8_t token)
{
  bool handled = true;

  switch (token)
  {
    /* Width of weight */
    case 170:
    case 171:
    case 172:
    case 173:
    case 174:
      comms_auto->token_digits = token - 170 + 5;
      break;
    case 179:
      comms_auto->token_digits = 0;
      break;

    /* Sign character for +ve numbers */
    case 180:
      comms_auto->token_plus_sign = 0;
      break;
    case 181:
      comms_auto->token_plus_sign = ' ';
      break;
    case 182:
      comms_auto->token_plus_sign = '+';
      break;
    case 183:
      comms_auto->token_plus_sign = '0';
      break;

    /* Decimal point character */
    case 184:
      comms_auto->token_dp = 0;
      break;
    case 185:
      comms_auto->token_dp = '.';
      break;
    case 186:
      comms_auto->token_dp = ',';
      break;

    /* ???? */
    case 187:
      comms_auto->token_lz = true;
      break;
    case 188:
      comms_auto->token_lz = false;
      break;

    /* ??? */
    case 189:
      comms_auto->token_error_mode = ERROR_MODE_NONE;
      break;
    case 190:
      comms_auto->token_error_mode = ERROR_MODE_BLANK_SPACE;
      break;
    case 191:
      comms_auto->token_error_mode = ERROR_MODE_BLANK_DASH;
      break;

    /* Status in uppercase or lowercase */
    case 192:
      comms_auto->token_upper = true;
      break;
    case 193:
      comms_auto->token_upper = false;
      break;
    default:
      handled = false;
      break;
  }
  return handled;
}


/** This function handles tokens that send weight information */
static bool auto_custom_fmt_handle_weight(struct s_comms_auto *comms_auto, const uint8_t token, AdcRdg *rdg)
{
  bool handled = true;

  switch (token)
  {
    /* Send weight from various sources */
    case 200:
      auto_send_weight_from_src(comms_auto, comms_auto->source, rdg);
      break;
    case 201:
      auto_send_weight_from_src(comms_auto, AUTO_OUT_SRC_DISP, rdg);
      break;
    case 202:
      auto_send_weight_from_src(comms_auto, AUTO_OUT_SRC_GROSS, rdg);
      break;
    case 203:
      auto_send_weight_from_src(comms_auto, AUTO_OUT_SRC_NET, rdg);
      break;
    case 204:
      auto_send_weight_from_src(comms_auto, AUTO_OUT_SRC_TARE, rdg);
      break;
    case 205:
      auto_send_weight_from_src(comms_auto, AUTO_OUT_SRC_TOTAL, rdg);
      break;
    default:
      handled = false;
      break;
  }
  return handled;
}


/** This function handles miscellaneous tokens that send weight information */
static bool auto_custom_fmt_handle_misc(struct s_comms_auto *comms_auto, const uint8_t token, const AdcRdg *rdg)
{
  bool handled = true;
  char status = auto_get_rdg_status_char(comms_auto, rdg);

  switch (token)
  {
    case 128:
      auto_send_char(comms_auto, '\0');
      break;

    case 210:
      auto_send_units(comms_auto, rdg);
      break;

    case 211:
      auto_send_char_upper(comms_auto,
                           (!comms_auto_error(comms_auto, rdg) && rdg->info.status.bit.motion)
                           ? 'M' : status);
      break;

    case 212:
      auto_send_char_upper(comms_auto, status);
      break;

    case 213:
      auto_send_char_upper(comms_auto, auto_get_rdg_status_GNT_char(rdg));
      break;

    case 214:
      auto_send_char_upper(comms_auto, (rdg->info.status.bit.motion) ? 'M' : ' ');
      break;

    case 215:
      auto_send_char_upper(comms_auto, (rdg->info.status.bit.motion) ? 'M' : 'S');
      break;

    case 216:
      if (rdg->info.status.bit.motion)
        auto_send_space3(comms_auto);
      else
        auto_send_units(comms_auto, rdg);
      break;

    case 217:
      if (comms_auto_error(comms_auto, rdg))
        auto_send_char_upper(comms_auto, 'C');
      else if (rdg->info.status.bit.motion)
        auto_send_char_upper(comms_auto, 'M');
      else
        auto_send_char(comms_auto, ' ');
      break;

    case 218:
      if (APP_ERR_NONE != comms_auto->sys_errors)
        auto_send_char_upper(comms_auto, 'I');
      else if (comms_auto_error(comms_auto, rdg))
        auto_send_char_upper(comms_auto, 'O');
      else if (rdg->info.status.bit.motion)
        auto_send_char_upper(comms_auto, 'M');
      else
        auto_send_char(comms_auto, ' ');
      break;

    case 219:
      if (rdg->info.status.bit.uload)
        auto_send_char_upper(comms_auto, 'U');
      else if (comms_auto_error(comms_auto, rdg))
        auto_send_char_upper(comms_auto, 'O');
      else
        auto_send_char_upper(comms_auto, 'I');
      break;

    case 220:
      auto_send_char_upper(comms_auto, (rdg->info.status.bit.coz) ? 'Z' : ' ');
      break;

    case 221:
      auto_send_range(comms_auto, rdg);
      break;

    case 222:
      if (comms_auto_error(comms_auto, rdg))
        auto_send_str_upper(comms_auto, "OL");
      else if (rdg->info.status.bit.motion)
        auto_send_str_upper(comms_auto, "US");
      else
        auto_send_str_upper(comms_auto, "ST");
      break;

    case 230:
      auto_send_time(comms_auto);
      break;

    case 231:
      auto_send_date(comms_auto);
      break;

    default:
      handled = false;
      break;
  }
  return handled;
}


/** This function handles characters that arn't tokens */
static void auto_custom_fmt_handle_nontoken(struct s_comms_auto *comms_auto, char token)
{
  auto_send_char(comms_auto, token);
}


static void auto_custom_send_fmt(struct s_comms_auto *comms_auto, const char *fmt, AdcRdg *rdg)
{
  const uint8_t *p;

  p = (const uint8_t *) fmt;

  /* Auto output loop through format string */
  while ('\0' != *p)
  {
    if (!auto_custom_fmt_handle_formatting(comms_auto, *p))
    {
      if (!auto_custom_fmt_handle_weight(comms_auto, *p, rdg))
      {
        if (!auto_custom_fmt_handle_misc(comms_auto, *p, rdg))
        {
          auto_custom_fmt_handle_nontoken(comms_auto, *p);
        }
      }
    }
    p++;
  }
}


static void auto_collect_data(struct s_comms_auto *comms_auto)
{
  int32_t total;

  /* Defaults for formatting */
  comms_auto->token_dp = '.';
  comms_auto->token_plus_sign = ' ';
  comms_auto->token_error_mode = ERROR_MODE_NONE;
  comms_auto->token_upper = true;
  comms_auto->token_digits = 8;
  comms_auto->token_lz = false;

  /* Collect weight data in 1 go */
  GetDateTime(&comms_auto->datetime);
  total = statistics_get_total();
  AdcMutexLock(&adcSuperStructure);
  comms_auto->sys_errors = get_current_errors();

  /* Get weights unit switched */
  comms_auto->displayed_units = unitsw_displaymode_to_unitswitch(adcSuperStructure.adcDataWorking.dispMode);
  comms_auto->resolution_length = (ADC_APPROVAL_MODE_NTEP == adcSuperStructure.adcUserDatabase.approvalMode) ? unitsw_get_resolution_length(comms_auto->displayed_units) : 0;

  unitsw_get_adcrdg(&comms_auto->rdg_displayed, comms_auto->displayed_units, wt_displayed);
  unitsw_get_adcrdg(&comms_auto->rdg_gross, comms_auto->displayed_units, wt_gross);
  unitsw_get_adcrdg(&comms_auto->rdg_net, comms_auto->displayed_units, wt_net);
  unitsw_get_adcrdg(&comms_auto->rdg_tare, comms_auto->displayed_units, wt_tare);

  memset(&comms_auto->rdg_total, 0, sizeof(comms_auto->rdg_total));
  comms_auto->rdg_total.rdg = unitsw_convert_final(total, ADC_UNIT_SWITCH_PRIMARY, comms_auto->displayed_units, false, false);
  comms_auto->rdg_total.info.unitdp.bit.units = unitsw_get_units(comms_auto->displayed_units);
  comms_auto->rdg_total.info.unitdp.bit.dp = unitsw_get_decimal_point(comms_auto->displayed_units);
  AdcMutexUnlock(&adcSuperStructure);
}


static void auto_send_start_char(struct s_comms_auto *comms_auto)
{
  /* Add start delimiter */
  if ('\0' != comms_auto->start_char)
  {
    auto_send_char(comms_auto, (char) comms_auto->start_char);
  }
}


static void auto_send_end_chars(struct s_comms_auto *comms_auto)
{
  /* Add trailing delimiters */
  if ('\0' != comms_auto->end_char1)
  {
    auto_send_char(comms_auto, (char) comms_auto->end_char1);
  }
  if ('\0' != comms_auto->end_char2)
  {
    auto_send_char(comms_auto, (char) comms_auto->end_char2);
  }
}


static void auto_internal_send(struct s_comms_auto *comms_auto)
{
  char status_char;
  AdcRdg rdg;
  const void *buffer;
  size_t buflength;

  /* Clear the comms send buffer */
  comms_send_buffer_clear(comms_auto->csb);

  /* Collect data for this auto output string */
  auto_collect_data(comms_auto);

  /* Add start delimiter */
  auto_send_start_char(comms_auto);

  /* Get reading to auto output */
  auto_get_reading_from_source(comms_auto, comms_auto->source, &rdg);

  /* Sort out status */
  status_char = auto_get_rdg_status_char(comms_auto, &rdg);

  /* Print appropriate format */
  switch (comms_auto->format)
  {
    default:
    case AUTO_OUT_FMT_RINAUTO:
      auto_send_weight(comms_auto, &rdg);
      if (!comms_auto_error(comms_auto, &rdg) && rdg.info.status.bit.motion)
        auto_send_char(comms_auto, 'M');
      else
        auto_send_char(comms_auto, status_char);
      break;

    case AUTO_OUT_FMT_HBMAUTO:
      auto_send_char(comms_auto, status_char);
      auto_send_weight(comms_auto, &rdg);
      if (rdg.info.status.bit.motion)
        auto_send_space3(comms_auto);
      else
        auto_send_units(comms_auto, &rdg);
      break;

    case AUTO_OUT_FMT_RINEXTEND:
      auto_send_weight(comms_auto, &rdg);
      auto_send_char(comms_auto, status_char);
      if (rdg.info.status.bit.motion)
        auto_send_char(comms_auto, 'M');
      else
        auto_send_char(comms_auto, ' ');
      if (rdg.info.status.bit.coz)
        auto_send_char(comms_auto, 'Z');
      else
        auto_send_char(comms_auto, ' ');
      auto_send_range(comms_auto, &rdg);
      auto_send_units(comms_auto, &rdg);
      break;

    case AUTO_OUT_FMT_RINSIMPLE:
      auto_send_weight(comms_auto, &rdg);
      break;

    case AUTO_OUT_FMT_CUSTOM:
      auto_custom_send_fmt(comms_auto, comms_auto->custom_format_str, &rdg);
      break;

    case AUTO_OUT_FMT_CONDEC:
      /* Send weight */
      auto_send_weight(comms_auto, &rdg);
      /* Units K/L/G/T */
      switch (rdg.info.unitdp.bit.units)
      {
        case ADC_UNITS_KILOGRAMS:
          auto_send_char(comms_auto, 'K');
          break;

        case ADC_UNITS_GRAMS:
          auto_send_char(comms_auto, 'G');
          break;

        case ADC_UNITS_TONNES:
          auto_send_char(comms_auto, 'T');
          break;

        case ADC_UNITS_POUNDS:
          auto_send_char(comms_auto, 'L');
          break;

        default:
          auto_send_char(comms_auto, ' ');
          break;
      }
      /* Gross/Net status */
      auto_send_char(comms_auto, (rdg.info.status.bit.gross) ? 'G' : 'N');
      /* Underload/Overload status */
      if (rdg.info.status.bit.oload || rdg.info.status.bit.uload)
        auto_send_char(comms_auto, 'O');
      else if (APP_ERR_NONE != comms_auto->sys_errors)
        auto_send_char(comms_auto, 'I');
      else if (rdg.info.status.bit.motion)
        auto_send_char(comms_auto, 'M');
      else
        auto_send_char(comms_auto, ' ');
      break;
  }

  /* Add trailing delimiters */
  auto_send_end_chars(comms_auto);

  /* Send the assembled data out the serial port as a single write */
  if (comms_send_buffer_get_buffer(comms_auto->csb, &buffer, &buflength) && buffer != NULL && buflength > 0)
    (void) rwrite(comms_auto->serdev, buffer, buflength, OS_TICKS_PER_SEC);
}


/** Send an automatic output */
void comms_auto_send(struct s_comms_auto *comms_auto)
{
  auto_internal_send(comms_auto);
}


/** Called when weight reading requested via Kp command -> This sends an automatic output to the port it was requested on, rather than the auto output port. */
void comms_auto_send_request(struct t_io_dev *serdev)
{
  struct s_comms_auto *ca;

  ca = comms_auto_create(serdev, false);
  if (NULL != ca)
  {
    ca->format = db_app_user.aof_setting;
    ca->source = AUTO_OUT_SRC_DISP;
    auto_internal_send(ca);
    comms_auto_destroy(ca);
  }
}
