/*
 * comms_auto_trigger.c
 *
 *  Created on: 02/03/2011
 *      Author: ian
 */
#include <stdlib.h>
#include <string.h>
#include <semaphore.h>
#include <rinhw/timemath.h>
#include <rinhw/rioctl.h>
#include "database.h"
#include "serial_levels.h"
#include "comms_task.h"
#include "app_db_user.h"
#include "we_config.h"

#include "comms_auto_trigger.h"


/** Re-initialise parts of ca from the database settings */
void comms_auto_reinit(struct s_comms_auto *ca)
{
  if (NULL == ca)
    return;

  ca->format = db_app_user.aof_setting;
  ca->source = db_app_user.aof_source;
  ca->start_char = db_app_user.start_char;
  ca->end_char1 = db_app_user.end_char1;
  ca->end_char2 = db_app_user.end_char2;
  ca->custom_format_str = db_app_user.auto_fmt;
}


/** Create and initialise a new s_comms_auto structure */
struct s_comms_auto *comms_auto_create(struct t_io_dev *serdev, bool is_incoming_udp)
{
  struct s_comms_auto *ca = NULL;
  bool sem_inited = false;

  /* Allocate new comms_auto */
  ca = calloc(1, sizeof(struct s_comms_auto));
  if (NULL == ca)
    goto cac_fail;

  /* Initialise it */
  ca->serdev = serdev;
  ca->is_incoming_udp = is_incoming_udp;
  if (!comms_send_buffer_create(&ca->csb, 10) || NULL == ca->csb)
    goto cac_fail;
  comms_auto_reinit(ca);
  if (sem_init(&ca->trigger_sem, 0, 0) != 0)
    goto cac_fail;
  sem_inited = true;
  return ca;

cac_fail:
  if (NULL != ca)
  {
    if (NULL != ca->csb)
    {
      comms_send_buffer_destroy(ca->csb);
      ca->csb = NULL;
    }
    if (sem_inited)
      sem_destroy(&ca->trigger_sem);
    free(ca);
    ca = NULL;
  }
  return NULL;
}


/** Clean up and destroy an existing s_comms_auto structure */
void comms_auto_destroy(struct s_comms_auto *ca)
{
  if (NULL != ca)
  {
    comms_send_buffer_destroy(ca->csb);
    ca->csb = NULL;
    sem_destroy(&ca->trigger_sem);
    free(ca);
  }
}


/** Does the work of sending an automatic output. Waits until signalled. */
void comms_auto_handler(struct s_comms_auto *ca)
{
  size_t buf_free;
  bool buf_has_room;

  /* Check pointer */
  if (NULL == ca)
  {
    goto fail_delay;
  }

  /* Wait until signalled or timeout */
  if (sem_timedwait_ticks(&ca->trigger_sem, OS_TICKS_PER_SEC) < 0)
  {
    goto fail_delay;
  }

  /* Special handling for incoming UDP connections */
  if (ca->is_incoming_udp)
    comms_udp_read_for_txonly(ca->serdev);

  /* Figure out if there is room in the buffer */
  buf_free = get_serial_tx_free(ca->serdev);
  buf_has_room = (buf_free > WECFG_AUTO_BUF_REQUIRED_FREE);
  if (buf_has_room)
  {
    comms_auto_send(ca);
  }
  return;

fail_delay:
  OSTimeDly(OS_TICKS_PER_SEC);
}


/** Trigger an auto output on a specific port */
void auto_output_trigger(struct s_comms_auto *ca)
{
  if (NULL != ca)
  {
    sem_post(&ca->trigger_sem);
  }
}


/** Trigger an auto-output transmission on all configured ports */
static void auto_output_trigger_all(enum eSerialPortType auto_type)
{
  enum eSerialPort port;

  for (port = 0; port < SERPORT_NUM; port++)
  {
    if (auto_type == comms_task_list_get_port_type(port))
    {
      comms_auto_trigger(port);
    }
  }
}


/** Trigger an single output transmission on specified port */
static void single_output_trigger(enum eSerialPort port)
{
  if (DB_SERIAL_TYPE_SINGLE == comms_task_list_get_port_type(port))
  {
    comms_auto_trigger(port);
  }
}


/** Trigger an auto-output transmission on all configured ports */
void auto_lo_output_trigger(void)
{
  auto_output_trigger_all(DB_SERIAL_TYPE_AUTO_LO);
}


/** Trigger an auto-output transmission on all configured ports */
void auto_hi_output_trigger(void)
{
  auto_output_trigger_all(DB_SERIAL_TYPE_AUTO_HI);
}


/** Trigger an single auto-output transmission on specified port */
bool auto_single_output_trigger(enum eSerialPort port)
{
  int8u sertype;
  bool result;

  sertype = comms_task_list_get_port_type(port);

  switch (port)
  {
    case SERPORT_RL_FRONT:
    case SERPORT_SER1_RL_REAR:
    case SERPORT_RB1A:
    case SERPORT_RB1B:
    case SERPORT_RB2A:
    case SERPORT_RB2B:
    case SERPORT_USBGADGET:
    case SERPORT_USBHOST:
    case SERPORT_SER2:
      result = (DB_SERIAL_TYPE_SINGLE == sertype);
      break;

    default:
      result = false;
      break;
  }
  if (result)
  {
    single_output_trigger(port);
  }

  return result;
}
