/*
 * comms_escape.c
 *
 *  Created on: Jun 11, 2012
 *      Author: ian
 */
#include <ctype.h>
#include <stdlib.h>
#include <stdint.h>

#include "comms_escape.h"

#define COMMS_ESCAPE_MAX_CHAR 29


static const char comms_escape_chars[COMMS_ESCAPE_MAX_CHAR + 1] = ".ABCDEFGHIJKLMNOPQRSTUVWXYZ+-";


/**
 * Escape a character as per the strange HBM comms character escape rules:
 * 0x00-0x1F,",0x98..0xFF  -> \nnn
 * 0x80..0x97 -> As per comms_escape_chars (\A, etc)
 * \ -> \\
 * Other printables -> Not escaped
 *
 * @param s is a string of length 5 characters or greater, that ch is escaped into
 * @param ch is the character to be escaped
 *
 * @return s
 */
char *comms_escape_char(char *s, char ch)
{
  char *sp;
  uint8_t b;

  b = (uint8_t) ch;
  sp = s;
  if (b < ' ')
  {
    /* Control characters 0x00-0x1F */
    *sp++ = '\\';
    *sp++ = '0';
    *sp++ = '0' + b / 10;
    *sp++ = '0' + b % 10;
  }
  else if ('\\' == b)
  {
    /* \ is the escape character, so send two to signify a \ */
    *sp++ = '\\';
    *sp++ = '\\';
  }
  else if ('"' == b)
  {
    /* " is escaped as it's ASCII decimal representation */
    *sp++ = '\\';
    *sp++ = '0';
    *sp++ = '3';
    *sp++ = '4';
  }
  else if (b < 128)
  {
    /* Normal printable character */
    *sp++ = b;
  }
  else
  {
    /* Characters above 127 (non-printable) */
    *sp++ = '\\';
    if ((b - 128) < COMMS_ESCAPE_MAX_CHAR)
    {
      /* 128..151 are escaped as \A, etc */
      *sp++ = comms_escape_chars[b - 128];
    }
    else
    {
      /* 152..255 are escaped as the decimal representation, eg \152 */
      *sp++ = '0' + b / 100;
      b = b % 100;
      *sp++ = '0' + b / 10;
      *sp++ = '0' + b % 10;
    }
  }
  *sp = '\0';
  return s;
}


/**
 * Un-escape a character as per the strange HBM comms character escape rules:
 * \nnn -> char value
 * \\ -> \
 * \A, etc -> 129 as per comms_escape_chars
 * \. -> 128
 * \+ -> 155
 * \- -> 156
 * Other printables -> Not escaped
 *
 * @param unescstr is the output un-escaped string (always <= length of escstr, unless max_len gets in the way)
 * @param escstr is the escaped string to be decoded
 * @param max_len specifies the maximum capacity of unescstr including null terminating character
 */
void comm_unescape_str(char *unescstr, const char *escstr, size_t max_len)
{
  size_t i, j;
  char ch;

  for (i = 0; '\0' != *escstr && i < max_len; i++, escstr++)
  {
    if (*escstr == '\\')
    {
      /* Escaped character */
      escstr++;
      switch (*escstr)
      {
        case '\\':
          unescstr[i] = '\\';
          break;
        case '.':
          unescstr[i] = COMMS_ESCAPED_DOT_UNESCAPED;
          break;
        case '+':
          unescstr[i] = COMMS_ESCAPED_PLUS_UNESCAPED;
          break;
        case '-':
          unescstr[i] = COMMS_ESCAPED_MINUS_UNESCAPED;
          break;
        default:
          if (isalpha(*escstr))
          {
            unescstr[i] = COMMS_ESCAPED_CHAR_TO_UNESCAPED(*escstr);
          }
          else
          {
            for (ch = 0, j = 0; j < 3; j++, escstr++)
            {
              if (!isdigit(*escstr))
                break;
              ch = ch * 10 + *escstr - '0';
            }
            if (j == 3)
            {
              if (ch == 0)
                ch = 128;
              unescstr[i] = ch;
            }
            else
              unescstr[i] = ' ';
            escstr--;
          }
          break;
      } /* switch */
    }
    else
    {
      /* Non-escaped character */
      unescstr[i] = *escstr;
    }
  }

  /* Don't let it go over max_len */
  if (i == max_len)
    unescstr[max_len - 1] = '\0';
  else
    unescstr[i] = '\0';
}
