/*
 * comms_hdlr_network.c
 *
 * Since the introduction of libsetnetconfig, everything here is in network endian
 *
 *  Created on: Jan 11, 2018
 *      Author: ian
 */
#include <stdint.h>
#include <stdbool.h>
#include <linux/if_ether.h>
#include <string.h>
#include <stdio.h>
#include <rincmd/rc_public.h>
#include <setnetconfig.h>
#include "comms_util.h"
#include "comms_errors.h"
#include "ipinfo.h"
#include "iface_ethnet.h"
#include "we_config.h"
#include "rc_gen_basic.h"
#include "comms_hdlr_defaults.h"
#include "edit_string.h"

#include "comms_hdlr_network.h"


/** Number of network configuration registers */
#define RINCMD_840X_NUM_REGS (RINREG_BASIC_DNS_SERVER_3 - RINREG_BASIC_DHCP_ENABLE)


/** True if the corresponding network configuration register has been written */
static bool rincmd_840x_written[RINCMD_840X_NUM_REGS];


/** Network configuration for rincmd */
struct s_snc_net_config rincmd_840x_net_config;


/** MAC address as a string for rincmd */
char rincmd_8408_mac_address_string[MAC_ADDRESS_STR_LENGTH];


static void send_ethnet_ip(struct s_comms_control *comms_ctrl, struct in_addr ip)
{
  char ip_str[INET_ADDRSTRLEN];

  (void) net_ip_inet_ntop_network_afinet(&ip, ip_str, INET_ADDRSTRLEN);
  comms_net_rfprintf(comms_ctrl, "\"%s\"", ip_str);
}


/** Send Ethernet settings */
static void send_ETH(struct s_comms_control *comms_ctrl, const struct s_snc_net_config *ipc)
{
  comms_net_rfprintf(comms_ctrl, "%u,", (ipc->use_dhcp) ? 1 : 0);
  send_ethnet_ip(comms_ctrl, ipc->ip_address);
  comms_net_rfprintf(comms_ctrl, ",");
  send_ethnet_ip(comms_ctrl, ipc->ip_netmask);
  comms_net_rfprintf(comms_ctrl, ",");
  send_ethnet_ip(comms_ctrl, ipc->ip_gateway);
  comms_net_rfprintf(comms_ctrl, ",");
  send_ethnet_ip(comms_ctrl, ipc->ip_dns[0]);
  comms_net_rfprintf(comms_ctrl, ",");
  send_ethnet_ip(comms_ctrl, ipc->ip_dns[1]);
  comms_net_rfprintf(comms_ctrl, ",\"%02X:%02X:%02X:%02X:%02X:%02X\",\"%s\",\"%s\",",
                     (uint8_t) ipc->mac_address.sa_data[0],
                     (uint8_t) ipc->mac_address.sa_data[1],
                     (uint8_t) ipc->mac_address.sa_data[2],
                     (uint8_t) ipc->mac_address.sa_data[3],
                     (uint8_t) ipc->mac_address.sa_data[4],
                     (uint8_t) ipc->mac_address.sa_data[5],
                     ipc->hostname,
                     ipc->search_domain
                    );
  send_ethnet_ip(comms_ctrl, ipc->ip_dns[2]);
  comms_net_rfprintf(comms_ctrl, COMMS_TERMINATOR);
}


static enum eCommsError parse_ip_param(struct s_comms_control *comms_ctrl, uint8_t param, struct in_addr *ip, bool required)
{
  enum eCommsError result;

  /* Check parameter string */
  if (!param_str_ok(&CTRL_COMP, param, INET_ADDRSTRLEN, 0))
  {
    /* Parameter not provided, but not required */
    result = (required) ? MESS_ERROR : MESS_OK;
  }
  else
  {
    result = (net_ip_inet_pton_network_afinet(COMPARAMSTR(param), ip)) ? MESS_OK : MESS_ERROR;
  }
  return result;
}


static enum eCommsError get_current_netconfig(struct s_snc_net_config *net_current)
{
  bool result;

  /* Get current active configuration */
  result = net_ip_get_config(false, net_current);
  if (!result)
  {
    /* Failed, so try getting the current settings */
    result = net_ip_get_config(true, net_current);
  }

  return (result) ? MESS_OK : MESS_ERROR;
}


/** Ethernet settings */
uint8_t handle_ETH(struct s_comms_control *comms_ctrl, char *s)
{
  struct s_snc_net_config current_net_config;
  enum eCommsError res;

  res = get_current_netconfig(&current_net_config);
  if (MESS_OK != res)
    return (uint8_t) res;

  if (comms_ctrl->cmd_query)
  {
    send_ETH(comms_ctrl, &current_net_config);
    return REPLY_SENT;
  }

  /* Param 0 - DHCP */
  if (param_ok(&CTRL_COMP, 0, 1, 0))
    current_net_config.use_dhcp = (0 != COMPARAM(0));

  /* Only set other parameters if DHCP is not enabled */
  if (!current_net_config.use_dhcp)
  {
    /* Param 1 - IP Address */
    if (parse_ip_param(comms_ctrl, 1, &current_net_config.ip_address, true) != MESS_OK)
      return MESS_ERROR;

    /* Param 2 - netmask */
    if (parse_ip_param(comms_ctrl, 2, &current_net_config.ip_netmask, true) != MESS_OK)
      return MESS_ERROR;

    /* Param 3 - gateway */
    if (parse_ip_param(comms_ctrl, 3, &current_net_config.ip_gateway, true) != MESS_OK)
      return MESS_ERROR;

    /* Param 4 - primary dns */
    if (parse_ip_param(comms_ctrl, 4, &current_net_config.ip_dns[0], true) != MESS_OK)
      return MESS_ERROR;

    /* Param 5 - secondary dns */
    if (parse_ip_param(comms_ctrl, 5, &current_net_config.ip_dns[1], true) != MESS_OK)
      return MESS_ERROR;
  }

  /* Param 6 - MAC Address is read only, and can't be set */

  /* Param 7 - hostname */
  if (param_str_charset_ok(&CTRL_COMP, 7, SNC_HOST_NAME_MAX, 1, charset_hostname))
    snprintf(current_net_config.hostname, SNC_HOST_NAME_MAX, "%s", COMPARAMSTR(7));

  /* Only set other parameters if DHCP is not enabled */
  if (!current_net_config.use_dhcp)
  {
    /* Param 8 - search domain */
    if (param_str_charset_ok(&CTRL_COMP, 8, SNC_DOMAIN_NAME_MAX, 0, charset_domainname))
      snprintf(current_net_config.search_domain, SNC_DOMAIN_NAME_MAX, "%s", COMPARAMSTR(8));

    /* Param 9 - tertiary dns (optional) */
    if (parse_ip_param(comms_ctrl, 9, &current_net_config.ip_dns[2], false) != MESS_OK)
      return MESS_ERROR;
  }

  /* Validate and set modified IP address configuration */
  return (net_ip_save_config(&current_net_config, NULL) == nesr_success) ? MESS_OK : MESS_ERROR;
}


/** Rincmd callback for network settings */
bool rincmd_840x_network_settings(char *val, size_t len, size_t maxLen, struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, uint8_t opt)
{
  struct s_snc_net_config net_current;
  size_t i;

  /* Only network setting registers, others shouldn't ever occur */
  if (reg->addr < RINREG_BASIC_DHCP_ENABLE || reg->addr > RINREG_BASIC_DNS_SERVER_3)
    return false;

  switch (opt)
  {
    case RC_FUNC_READ_PRE:
      /*
       * Pre-read: Get the current network settings
       * and copy the appropriate setting for the register over to rincmd_840x_net_config
       * only if the rincmd_840x_net_config hasn't already been written
       */
      net_ip_set_defaults(&net_current);
      (void) get_current_netconfig(&net_current);
      for (i = 0; i < RINCMD_840X_NUM_REGS; i++)
      {
        if (!rincmd_840x_written[i])
        {
          switch (i + RINREG_BASIC_DHCP_ENABLE)
          {
            case RINREG_BASIC_DHCP_ENABLE:
              rincmd_840x_net_config.use_dhcp = net_current.use_dhcp;
              break;
            case RINREG_BASIC_IP_ADDRESS:
              rincmd_840x_net_config.ip_address.s_addr = net_current.ip_address.s_addr;
              break;
            case RINREG_BASIC_SUBNET_MASK:
              rincmd_840x_net_config.ip_netmask.s_addr = net_current.ip_netmask.s_addr;
              break;
            case RINREG_BASIC_DEFAULT_GATEWAY:
              rincmd_840x_net_config.ip_gateway.s_addr = net_current.ip_gateway.s_addr;
              break;
            case RINREG_BASIC_DNS_SERVER_1:
              rincmd_840x_net_config.ip_dns[0].s_addr = net_current.ip_dns[0].s_addr;
              break;
            case RINREG_BASIC_DNS_SERVER_2:
              rincmd_840x_net_config.ip_dns[1].s_addr = net_current.ip_dns[1].s_addr;
              break;
            case RINREG_BASIC_NETWORK_DEFAULT:
              /* This one has it's own handler - rincmd_8406_network_defaults() */
              break;
            case RINREG_BASIC_MAC_ADDRESS:
              net_ip_macaddress_to_string(&net_current.mac_address, ':', rincmd_8408_mac_address_string);
              break;
            case RINREG_BASIC_HOSTNAME:
              snprintf(rincmd_840x_net_config.hostname, SNC_HOST_NAME_MAX, "%s", net_current.hostname);
              break;
            case RINREG_BASIC_SEARCH_DOMAIN:
              snprintf(rincmd_840x_net_config.search_domain, SNC_DOMAIN_NAME_MAX, "%s", net_current.search_domain);
              break;
            case RINREG_BASIC_DNS_SERVER_3:
              rincmd_840x_net_config.ip_dns[2].s_addr = net_current.ip_dns[2].s_addr;
              break;
          }
        } /* Not written */
      } /* for */
      break;

    case RC_FUNC_WRITE_POST:
      /*
       * To keep the same behavior as the R400,
       * don't save network settings here,
       * but wait until the "save settings" register (0x0010) is executed
       * (void) net_ip_save_config(&rincmd_840x_net_config, false);
       *
       * We do however note that the setting has been written,
       * so future reads don't overwrite the written value, until it has been saved
       */
      rincmd_840x_written[reg->addr - RINREG_BASIC_DHCP_ENABLE] = true;
      break;

    default:
      break;
  }

  return false;
}


/** Call this when the rincmd network settings have been saved */
void rincmd_840x_saved_settings(void)
{
  size_t i;

  for (i = 0; i < RINCMD_840X_NUM_REGS; i++)
    rincmd_840x_written[i] = false;
}


/** Rincmd callback for defaulting the network settings */
bool rincmd_8406_network_defaults(char *val, size_t len, size_t maxLen, struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, uint8_t opt)
{
  bool result;

  if (opt == RC_FUNC_EXEC)
  {
    result = comms_default_settings(cds_network);
    rincmd_840x_saved_settings();
  }
  else
  {
    result = false;
  }

  return result;
}
