#include <setjmp.h>
#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/timemath.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcliveimpl_simple.h>
#include "serial_levels.h"
#include "asciidef.h"
#include "meas.h"
#include "database.h"
#include "backgrnd.h"
#include "version.h"
#include "iface.h"
#include "system.h"
#include "editor.h"
#include "kbd_5000.h"
#include "setpoint.h"
#include "comms_print_trigger.h"
#include "access.h"
#include "sysclock.h"
#include "buzz5000.h"
#include "modbus_old.h"
#include "comms_util.h"
#include "display.h"
#include "comms_params.h"
#include "comms_hdlr_am.h"
#include "comms_kbd.h"
#include "comms_setup.h"
#include "comms_errors.h"
#include "reboot.h"
#include "comms_task.h"
#include "db_remote.h"
#include "iface_operator.h"
#include "we_config.h"
#include "traceprint.h"
#include "iface_serial.h"
#include "adcthread.h"
#include "app_db_factory.h"
#include "app_db_user.h"
#include "app_db_runtime.h"
#include "app_db_runtime_vars.h"
#include "app_db_changelog.h"
#include "comms_escape.h"
#include "iface_menus.h"
#include "comms_bdxwfp.h"
#include "temperature.h"
#include "app_save.h"

#include "comms_hdlr_nz.h"


/** Printer formats */
uint8_t handle_PFT(struct s_comms_control *comms_ctrl, char *s)
{
  char temp_str[10];
  const char *p;

  if (comms_ctrl->cmd_query)
  {
    comms_net_rfprintf(comms_ctrl, "\"");
    p = db_app_user.print_fmt;
    while (*p)
    {
      comms_net_rfprintf(comms_ctrl, "%s", comms_escape_char(temp_str, *p));
      p++;
    }
    comms_net_rfprintf(comms_ctrl, "\"" COMMS_TERMINATOR);
    return REPLY_SENT;
  }

  if (param_str_ok(&CTRL_COMP, 0, 250, 0))
  {
    comm_unescape_str(db_app_user.print_fmt, COMPARAMSTR(0), APP_DB_PRINT_FMT_LEN);
    return MESS_OK;
  }
  return MESS_ERROR;
}


/** Digital inputs and outputs */
uint8_t handle_POR(struct s_comms_control *comms_ctrl, char *s)
{
  uint8_t temp;
  uint8_t i, mask_on, mask_off, mask;

  if (comms_ctrl->cmd_query)
  {
    temp = access_get_io_states();

    /* Print IO states */
    for (i = 0; i < NUM_SETPOINTS - 1; i++)
    {
      comms_net_rfprintf(comms_ctrl, "%u,", ((temp & (0x01 << i)) != 0) ? 1 : 0);
    }
    comms_net_rfprintf(comms_ctrl, "%u" COMMS_TERMINATOR, ((temp & (0x01 << i)) != 0) ? 1 : 0);
    return REPLY_SENT;
  }

  /* Set outputs */
  mask_on = mask_off = 0;
  for (mask = 0x01, i = 0; i < NUM_SETPOINTS; i++, mask <<= 1)
  {
    if (param_ok(&CTRL_COMP, i, 1, 0))
    {
      /* Only unused IO can be forced as an output */
      if (IOTypeUnused != access_get_io_type(i))
      {
        return MESS_ERROR;
      }
      if (COMPARAM(i) == 1)
        mask_on |= mask;
      else
        mask_off |= mask;
    }
  }

  for (mask = 0x01, i = 0; i < NUM_SETPOINTS; i++, mask <<= 1)
  {
    if ((mask_on & mask) != 0)
      direct_drive_output(&setpoint_running_info, i, true);
    else if ((mask_off & mask) != 0)
      direct_drive_output(&setpoint_running_info, i, false);
  }
  return MESS_OK;
}


/** PRS command serial types, need a special translation to match old PRS command */
enum ePRSSerialType { PRS_SERIAL_TYPE_OFF = 0,
                      PRS_SERIAL_TYPE_AUTO_LO,
                      PRS_SERIAL_TYPE_PRINTING,
                      PRS_SERIAL_TYPE_SINGLE,
                      PRS_SERIAL_TYPE_AUTO_HI,
                      PRS_SERIAL_TYPE_NETWORK,
                      PRS_SERIAL_TYPE_NUM
                    };


/** Convert serial type to old PRS type parameter order */
static uint8_t xlat_serial_type_to_prs(uint8_t serial_type)
{
  uint8_t prs_type;

  switch (serial_type)
  {
    default:
    case DB_SERIAL_TYPE_OFF:
      prs_type = PRS_SERIAL_TYPE_OFF;
      break;
    case DB_SERIAL_TYPE_AUTO_LO:
      prs_type = PRS_SERIAL_TYPE_AUTO_LO;
      break;
    case DB_SERIAL_TYPE_AUTO_HI:
      prs_type = PRS_SERIAL_TYPE_AUTO_HI;
      break;
    case DB_SERIAL_TYPE_PRINTING:
      prs_type = PRS_SERIAL_TYPE_PRINTING;
      break;
    case DB_SERIAL_TYPE_SINGLE:
      prs_type = PRS_SERIAL_TYPE_SINGLE;
      break;
    case DB_SERIAL_TYPE_NETWORK:
      prs_type = PRS_SERIAL_TYPE_NETWORK;
      break;
  }
  return prs_type;
}


/** Convert old PRS type parameter order to serial type */
static uint8_t xlat_prs_to_serial_type(uint8_t prs_type)
{
  uint8_t serial_type;

  switch (prs_type)
  {
    default:
    case PRS_SERIAL_TYPE_OFF:
      serial_type = DB_SERIAL_TYPE_OFF;
      break;
    case PRS_SERIAL_TYPE_AUTO_LO:
      serial_type = DB_SERIAL_TYPE_AUTO_LO;
      break;
    case PRS_SERIAL_TYPE_PRINTING:
      serial_type = DB_SERIAL_TYPE_PRINTING;
      break;
    case PRS_SERIAL_TYPE_SINGLE:
      serial_type = DB_SERIAL_TYPE_SINGLE;
      break;
    case PRS_SERIAL_TYPE_AUTO_HI:
      serial_type = DB_SERIAL_TYPE_AUTO_HI;
      break;
    case PRS_SERIAL_TYPE_NETWORK:
      serial_type = DB_SERIAL_TYPE_NETWORK;
      break;
  }
  return serial_type;
}


/** Send the response to the PRS (printer settings) query */
static void send_PRS(struct s_comms_control *comms_ctrl)
{
  uint8_t prs_type;

  prs_type = xlat_serial_type_to_prs(db_get_serial_type(SERPORT_SER1_RL_REAR));

  comms_net_rfprintf(comms_ctrl, "%u,%u,%u,%u,%u,%u,%u,%u,%u,%u" COMMS_TERMINATOR,
                     (unsigned) prs_type,
                     (unsigned) db_app_user.printer_mode + 1,
                     (unsigned) db_app_user.aof_setting + 1,
                     (unsigned) db_app_user.tic_col,
                     (unsigned) db_app_user.tic_row,
                     (unsigned) db_app_user.aof_source + 1,
                     (unsigned) db_app_user.print_interlock,
                     (unsigned) db_app_user.start_char,
                     (unsigned) db_app_user.end_char1,
                     (unsigned) db_app_user.end_char2
                    );
}


/** Printer settings */
uint8_t handle_PRS(struct s_comms_control *comms_ctrl, char *s)
{
  uint8_t serial_type;

  if (comms_ctrl->cmd_query)
  {
    send_PRS(comms_ctrl);
    return REPLY_SENT;
  }

  /* Param 0, Serial 2 port type */
  if (param_ok(&CTRL_COMP, 0, PRS_SERIAL_TYPE_NUM - 1, 0))
  {
    serial_type = xlat_prs_to_serial_type(COMPARAM(0));
    db_app_user.serial_type[SERPORT_SER1_RL_REAR] = serial_type;
  }

  /* Param 1 - Printer mode */
  if (param_ok(&CTRL_COMP, 1, APP_DB_U_SERPRINTMODE_MAX + 1, APP_DB_U_SERPRINTMODE_MIN + 1))
  {
    if (db_app_user.printer_mode != (uint8_t) (COMPARAM(1) - 1))
    {
      db_app_user.printer_mode = (uint8_t) (COMPARAM(1) - 1);
      if (db_app_user.printer_mode >= P_TOTAL)
      {
        statistics_clear();
      }
    }
  }

  /* Param 2 - auto output format type */
  if (param_ok(&CTRL_COMP, 2, APP_DB_U_SERAOFTYPE_MAX + 1, APP_DB_U_SERAOFTYPE_MIN + 1))
    db_app_user.aof_setting = (uint8_t) (COMPARAM(2) - 1);

  /* Param 3 - printer ticket columns */
  if (param_ok(&CTRL_COMP, 3, APP_DB_U_SERTICKETCOL_MAX, APP_DB_U_SERTICKETCOL_MIN))
    db_app_user.tic_col = (uint8_t) COMPARAM(3);

  /* Param 4 - printer ticket rows */
  if (param_ok(&CTRL_COMP, 4, APP_DB_U_SERTICKETROW_MAX, APP_DB_U_SERTICKETROW_MIN))
    db_app_user.tic_row = (uint8_t) COMPARAM(4);

  /* Param 5 - auto output source */
  if (param_ok(&CTRL_COMP, 5, APP_DB_U_SERAOFSRC_MAX + 1, APP_DB_U_SERAOFSRC_MIN + 1))
    db_app_user.aof_source = (uint8_t) (COMPARAM(5) - 1);

  /* Param 6 - print interlock */
  if (param_ok(&CTRL_COMP, 6, APP_DB_U_PRINTINTERLOCK_MAX, APP_DB_U_PRINTINTERLOCK_MIN))
    db_app_user.print_interlock = (uint8_t) COMPARAM(6);

  /* Param 7 - start character */
  if (param_ok(&CTRL_COMP, 7, APP_DB_U_SERSTARTCHAR_MAX, APP_DB_U_SERSTARTCHAR_MIN))
    db_app_user.start_char = (uint8_t) COMPARAM(7);

  /* Param 8 - end character 1 */
  if (param_ok(&CTRL_COMP, 8, APP_DB_U_SERENDCHAR1_MAX, APP_DB_U_SERENDCHAR1_MIN))
    db_app_user.end_char1 = (uint8_t) COMPARAM(8);

  /* Param 9 - end character 2 */
  if (param_ok(&CTRL_COMP, 9, APP_DB_U_SERENDCHAR2_MAX, APP_DB_U_SERENDCHAR2_MIN))
    db_app_user.end_char2 = (uint8_t) COMPARAM(9);

  return MESS_OK;
}


static enum eCommsError send_last_printing(struct s_comms_control *comms_ctrl)
{
  t_datetime dt;
  int32u printid;
  AdcRdg rdg;
  long l;

  if (sem_timedwait_ticks(&comms_ctrl->print_finished_sem, OS_TICKS_PER_SEC * 5) == 0)
  {
    print_get_last_print_info(&l, &dt, &printid);

    memset(&rdg, 0, sizeof(AdcRdg));
    rdg.info.unitdp.bit.units = adcSuperStructure.adcUserDatabase.units;
    rdg.info.unitdp.bit.dp = adcSuperStructure.adcUserDatabase.dpPos;
    rdg.rdg = l;

    comms_net_rfprintf(comms_ctrl, "%lu,", (unsigned long) printid);
    send_CLK_no_line_termination(comms_ctrl, &dt);
    comms_net_rfprintf(comms_ctrl, ",");
    comms_send_num_with_dp(comms_ctrl, &rdg, 8);
    comms_net_rfprintf(comms_ctrl, COMMS_TERMINATOR);
    return REPLY_SENT;
  }
  else
  {
    return MESS_ERROR;
  }
}


static void send_PRT_buffer(struct s_comms_control *comms_ctrl)
{
  #define MAX_PRT_MESSAGE_LENGTH 100
  char ch;
  size_t cnt;
  bool have_print_data;

  comms_net_rfprintf(comms_ctrl, "\"");
  for (cnt = 0; cnt < MAX_PRT_MESSAGE_LENGTH; cnt++)
  {
    have_print_data = comms_print_buffer_get_char(comms_ctrl->cpb, &ch);
    if (!have_print_data)
      break;

    /* Escape and send data */
    if (ch != '"' && ch != '\\' && ch >= ' ' && ch <= '~')
    {
      /* Covers ' ' .. '~', but not \ or " */
      comms_net_rfprintf(comms_ctrl, "%c", ch);
    }
    else
    {
      /* Covers 0..31 and 127..255 and \ and " */
      comms_net_rfprintf(comms_ctrl, "\\%03d", ch);
      if ('\n' == ch)
        break;
    }
  }
  comms_net_rfprintf(comms_ctrl, "\"" COMMS_TERMINATOR);
}


static uint8_t send_PRT(struct s_comms_control *comms_ctrl, uint8_t typ)
{
  uint8_t result;

  switch (typ)
  {
    case 0:
      comms_net_rfprintf(comms_ctrl, "%lu" COMMS_TERMINATOR, (unsigned long) adcSuperStructure.adcRuntimeDatabase.printId);
      result = REPLY_SENT;
      break;
    case 1:
      send_PRT_buffer(comms_ctrl);
      result = REPLY_SENT;
      break;
    default:
      result = MESS_ERROR;
      break;
  }
  return result;
}


/** See if we can do a traceable print */
uint8_t traceable_can_print(void)
{
  int8u adc_result;
  uint8_t result;

  if (APP_ERR_NONE != get_current_errors())
  {
    result = MESS_ERR_SYS;
  }
  else
  {
    /* Test if operation would succeed */
    adc_result = AdcGetTraceableRdg(&adcSuperStructure, true, false, true, false);
    switch (adc_result)
    {
      case ADC_OP_STATUS_FAILED_MOTION:
        result = MESS_ERR_MOTION;
        break;
      case ADC_OP_STATUS_SUCCESS:
        result = MESS_OK;
        break;
      case ADC_OP_STATUS_FAILED_ERROR:
        result = MESS_ERR_SYS;
        break;
      default:
        result = MESS_ERROR;
        break;
    }
  }
  return result;
}


#define PRT_REQUIRES_REPLY(prttype) (0 != prttype)


/** Perform printout */
uint8_t handle_PRT(struct s_comms_control *comms_ctrl, char *s)
{
  uint8_t typ;
  uint8_t result;
  sem_t *print_finished;

  typ = 0;
  if (param_ok(&CTRL_COMP, 0, 4, 0))
    typ = (uint8_t) COMPARAM(0);

  if (comms_ctrl->cmd_query)
  {
    return send_PRT(comms_ctrl, typ);
  }

  set_last_print_type(typ);

  if (!comms_print_system_has_printer() && db_app_user.printer_mode < P_TOTAL)
  {
    return MESS_ERROR;
  }
  if (!comms_print_clear_to_send())
  {
    return MESS_ERR_DEV;
  }

  print_finished = (PRT_REQUIRES_REPLY(typ)) ? &comms_ctrl->print_finished_sem : NULL;
  if (param_str_ok(&CTRL_COMP, 1, 250, 0))
  {
    if (is_weight_fmt(COMPARAMSTR(1)))    /* TODO is_weight_fmt() I'm not sure what this is doing */
    {
      result = traceable_can_print();
      if (MESS_OK != result)
      {
        return result;
      }
    }
    traceable_printout_initiate_special_remote_comms(print_finished, COMPARAMSTR(1));
  }
  else
  {
    result = traceable_can_print();
    if (MESS_OK != result)
    {
      return result;
    }
    traceable_printout_initiate_standard_comms(print_finished);
  }

  /* PRT 1-4 require a reply */
  if (PRT_REQUIRES_REPLY(typ))
    return send_last_printing(comms_ctrl);
  else
    return MESS_OK;
}


static void send_PST(struct s_comms_control *comms_ctrl, size_t line)
{
  size_t i;

  comms_net_rfprintf(comms_ctrl, "\"");
  for (i = 0; i < APP_DB_PRINT_HDR_LEN - 1; i++)
  {
    comms_net_rfprintf(comms_ctrl, "%c", (char) db_app_user.print_text[line][i]);
  }
  comms_net_rfprintf(comms_ctrl, "\"" COMMS_TERMINATOR);
}


/** Printer text */
uint8_t handle_PST(struct s_comms_control *comms_ctrl, char *s)
{
  size_t line;
  size_t i;
  const char *p;

  line = 0;
  if (param_ok(&CTRL_COMP, 0, APP_DB_PRINT_HDR_LINES, 1))
  {
    line = COMPARAM(0) - 1;
  }

  if (comms_ctrl->cmd_query)
  {
    send_PST(comms_ctrl, line);
    return REPLY_SENT;
  }

  if (param_str_ok(&CTRL_COMP, 1, APP_DB_PRINT_HDR_LEN - 1, 0))
  {
    /* Check string contains valid characters */
    for (p = COMPARAMSTR(1); '\0' != *p; p++)
    {
      if (!isprint(*p))
        return MESS_ERROR;
    }

    /* Copy data over */
    for (i = 0, p = COMPARAMSTR(1); '\0' != *p && i < APP_DB_PRINT_HDR_LEN - 1; p++, i++)
    {
      db_app_user.print_text[line][i] = *p;
    }
    /* Blank rest of line */
    for (; i < APP_DB_PRINT_HDR_LEN - 1; i++)
    {
      db_app_user.print_text[line][i] = ' ';
    }
    /* Null terminate */
    db_app_user.print_text[line][APP_DB_PRINT_HDR_LEN - 1] = '\0';
    return MESS_OK;
  }
  return MESS_ERROR;
}


static void send_QAF(struct s_comms_control *comms_ctrl)
{
  comms_net_rfprintf(comms_ctrl, "%u,%u,%u,%u" COMMS_TERMINATOR,
                     (adcSuperStructure.adcUserDatabase.qaEnable) ? 1 : 0,
                     adcSuperStructure.adcUserDatabase.qaDay,
                     adcSuperStructure.adcUserDatabase.qaMonth,
                     adcSuperStructure.adcUserDatabase.qaYear % 100
                    );
}


/** Quality Assurance settings */
uint8_t handle_QAF(struct s_comms_control *comms_ctrl, char *s)
{
  uint8_t date;
  uint8_t month;
  uint16_t year;

  if (comms_ctrl->cmd_query)
  {
    send_QAF(comms_ctrl);
    return REPLY_SENT;
  }

  if (param_ok(&CTRL_COMP, 0, 1, 0))
    adcSuperStructure.adcUserDatabase.qaEnable = (COMPARAM(0) == 1);

  date = adcSuperStructure.adcUserDatabase.qaDay;
  month = adcSuperStructure.adcUserDatabase.qaMonth;
  year = adcSuperStructure.adcUserDatabase.qaYear;

  if (param_ok(&CTRL_COMP, 1, 31, 1))
    date = (uint8_t) COMPARAM(1);
  if (param_ok(&CTRL_COMP, 2, 12, 1))
    month = (uint8_t) COMPARAM(2);
  if (param_ok(&CTRL_COMP, 3, 2097, 0))
  {
    year = (uint16_t) COMPARAM(3);
    if (year < 98)
      year += 2000;
    else if (year < 100)
      year += 1900;
    else if (year < 1998)
      return MESS_ERROR;
  }
  if (!check_date(date, month, year))
    return MESS_ERROR;

  adcSuperStructure.adcUserDatabase.qaDay = date;
  adcSuperStructure.adcUserDatabase.qaMonth = month;
  adcSuperStructure.adcUserDatabase.qaYear = year;
  return MESS_OK;
}


/** Remote button settings */
uint8_t handle_RBT(struct s_comms_control *comms_ctrl, char *s)
{
  return handle_xBT(comms_ctrl, db_app_user.remote_f, NUM_REMOTE_KEYS, NUM_REMOTE_KEY_FUNCTIONS, remote_button_to_key_str);
}


/** Reset unit */
uint8_t handle_RES(struct s_comms_control *comms_ctrl, char *s)
{
  if (comms_ctrl->cmd_query)
  {
    return MESS_ERROR;
  }

  /* Start system reboot */
  system_reboot();
  return MESS_OK;
}


/** Sxx addressing commands, including STP */
uint8_t handle_S(struct s_comms_control *comms_ctrl, char *s)
{
  uint8_t addr;

  /* STP stops MSV streaming */
  if (s[1] == 'T' && s[2] == 'P')
  {
    comms_ctrl->msv_count = 1;
    return REPLY_SENT;
  }

  /* If streaming, Sxx does nothing */
  if (comms_ctrl->msv_cmd_streaming)
  {
    return REPLY_SENT;
  }

  /* Check address in command, and convert to integer */
  if (!isdigit(s[1]) || !isdigit(s[2]))
  {
    return MESS_ERROR;
  }
  addr = (uint8_t) (10 * (s[1] - '0')) + (s[2] - '0');

  /* Check passcode */
  if (0 == db_app_user.full_passcode)
  {
    comms_ctrl->passcode_ok = true;
  }

  /* Work out what to do based on address in Sxx command */
  switch (addr)
  {
    /* De-select all units */
    case 96:
      comms_ctrl->reply_active = false;
      comms_ctrl->unit_selected = false;
      comms_ctrl->passcode_ok = (0 == db_app_user.full_passcode);
      break;

    /* Select all units, but don't reply to messages */
    case 97:
    case 98:
      comms_ctrl->reply_active = false;
      comms_ctrl->unit_selected = true;
      break;

    /* Select all units, and reply to messages */
    case 99:
      comms_ctrl->reply_active = true;
      comms_ctrl->unit_selected = true;
      break;

    /* Select if our address matches */
    default:
      if (addr == db_app_user.address)
      {
        comms_ctrl->reply_active = true;
        comms_ctrl->unit_selected = true;
      }
      else
      {
        comms_ctrl->reply_active = false;
        comms_ctrl->unit_selected = false;
        comms_ctrl->passcode_ok = (0 == db_app_user.full_passcode);
      }
      break;
  }

  /* No reply to Sxx commands */
  return REPLY_SENT;
}


/** Tare */
uint8_t handle_TAR(struct s_comms_control *comms_ctrl, char *s)
{
  int8u res;

  if (comms_ctrl->cmd_query)
  {
    comms_net_rfprintf(comms_ctrl, "%ld" COMMS_TERMINATOR, adcSuperStructure.adcDataWorking.userTare);
    return REPLY_SENT;
  }

  /* Run some checks first, so it won't block */
  res = AdcSemiautoTare(&adcSuperStructure, true, false, true, false);
  if (ADC_OP_STATUS_SUCCESS == res)
  {
    /* Do the tare */
    res = AdcSemiautoTare(&adcSuperStructure, true, true, false, false);
  }
  return convert_adc_result_to_mess_result(res);
}


/** Gross/net select */
uint8_t handle_TAS(struct s_comms_control *comms_ctrl, char *s)
{
  if (comms_ctrl->cmd_query)
  {
    comms_net_rfprintf(comms_ctrl, "%u" COMMS_TERMINATOR,
                       (is_gross()) ? 1 : 0
                      );
    return REPLY_SENT;
  }

  if (param_ok(&CTRL_COMP, 0, 1, 0))
  {
    set_gross(COMPARAM(0) > 0);
  }
  return MESS_OK;
}


/** Preset tare */
uint8_t handle_TAV(struct s_comms_control *comms_ctrl, char *s)
{
  int32s tr_h, tr_l;

  if (comms_ctrl->cmd_query)
  {
    comms_net_rfprintf(comms_ctrl, "%ld" COMMS_TERMINATOR, adcSuperStructure.adcDataWorking.presetTare);
    return REPLY_SENT;
  }

  /* Work out limits */
  calculate_preset_tare_limits(&tr_l, &tr_h);

  /* Get new preset tare value */
  if (param_ok(&CTRL_COMP, 0, tr_h, tr_l))
  {
    AdcSetPTare(&adcSuperStructure, true, false, false, COMPARAM(0));
    return MESS_OK;
  }
  else
  {
    return MESS_ERR_RANGE;
  }
}


static void send_TDD(struct s_comms_control *comms_ctrl)
{
  comms_net_rfprintf(comms_ctrl, "%u" COMMS_TERMINATOR, adcSuperStructure.adcUserDatabase.calCntOIML);
}


/** Save database */
uint8_t handle_TDD(struct s_comms_control *comms_ctrl, char *s)
{
  uint8_t result;

  if (comms_ctrl->cmd_query)
  {
    send_TDD(comms_ctrl);
    return REPLY_SENT;
  }

  result = MESS_ERR_SYS;
  if (param_ok(&CTRL_COMP, 0, 2, 0))
  {
    switch (COMPARAM(0))
    {

      /* Default databases */
      case 0:
        if (!comms_ctrl->passcode_ok)
        {
          result = MESS_ERROR;
        }
        else if (app_can_save_databases())
        {
          /* Default user database, except calibration, in keeping with WE2110 */
          app_default_user_database(false);

          /* Default runtime database, except ADC runtime variables, in keeping with WE2110 */
          app_default_runtime_database(false);

          /* Save and re-initialise */
          result = app_save_databases_and_reinit(false, NULL) ? MESS_OK : MESS_ERR_SYS;
        }
        break;

      /* Save (changed) settings */
      case 1:
        /* Save and re-initialise */
        result = app_save_databases_and_reinit(true, NULL) ? MESS_OK : MESS_ERR_SYS;
        break;

      /* Read databases and re-initialise with read settings */
      case 2:
        app_read_factory_database(false);
        app_read_user_database();
        AdcReInit(&adcSuperStructure, true, &db_app_factory.adc_factory_database, &db_app_user.adc_user_database, NULL);

        /* Re-initialise everything after save */
        reinitialise_after_save();
        result = MESS_OK;
        break;
    }
  }
  return result;
}


static void send_TST(struct s_comms_control *comms_ctrl)
{
  comms_net_rfprintf(comms_ctrl, "0,%ld,%ld,%ld,%ld" COMMS_TERMINATOR,
                     get_12V_reading(),
                     temperature_get(tuCelcius),
                     get_pos_sense_reading(),
                     get_neg_sense_reading()
                    );
}


/** Test command, enable factory commands */
uint8_t handle_TST(struct s_comms_control *comms_ctrl, char *s)
{
  if (comms_ctrl->cmd_query)
  {
    send_TST(comms_ctrl);
    return REPLY_SENT;
  }

  /* No TST command */
  return MESS_ERROR;
}


static void send_TXT(struct s_comms_control *comms_ctrl, uint8_t sp)
{
  char temp_str[APP_DB_PRINT_TXT_LEN];

  text_get_text(temp_str, APP_DB_PRINT_TXT_LEN, sp);
  comms_net_rfprintf(comms_ctrl, "%u,\"%s\"" COMMS_TERMINATOR,
                     sp + 1,
                     temp_str
                    );
}


/** Text strings */
uint8_t handle_TXT(struct s_comms_control *comms_ctrl, char *s)
{
  uint8_t sp;

  sp = 42;
  if (param_ok(&CTRL_COMP, 0, APP_DB_PRINT_TXT_NUM, 1))
  {
    sp = (uint8_t) (COMPARAM(0) - 1);
  }

  if (comms_ctrl->cmd_query && sp != 42)
  {
    send_TXT(comms_ctrl, sp);
    return REPLY_SENT;
  }

  if (param_str_ok(&CTRL_COMP, 1, APP_DB_PRINT_TXT_LEN - 1, 0) && sp != 42)
  {
    text_set_text(sp, COMPARAMSTR(1));
    return MESS_OK;
  }
  return MESS_ERROR;
}


static void send_UNT(struct s_comms_control *comms_ctrl)
{
  comms_net_rfprintf(comms_ctrl, "%u,%u,%lu,%u" COMMS_TERMINATOR,
                     (unsigned) db_app_user.units_sec,
                     (unsigned) db_app_user.units_ter,
                     (unsigned long) db_app_user.unit_conv_factor,
                     (unsigned) db_app_user.show_lboz
                    );
}


/** Unit switching settings */
uint8_t handle_UNT(struct s_comms_control *comms_ctrl, char *s)
{
  if (comms_ctrl->cmd_query)
  {
    send_UNT(comms_ctrl);
    return REPLY_SENT;
  }

  if (param_ok(&CTRL_COMP, 0, APP_DB_U_UNITS_SEC_MAX, APP_DB_U_UNITS_SEC_MIN))
  {
    db_app_user.units_sec = (uint8_t) COMPARAM(0);
  }

  if (param_ok(&CTRL_COMP, 1, APP_DB_U_UNITS_TER_MAX, APP_DB_U_UNITS_TER_MIN))
  {
    db_app_user.units_ter = (uint8_t) COMPARAM(1);
  }

  if (param_ok(&CTRL_COMP, 2, APP_DB_U_UNITS_CONV_MAX, APP_DB_U_UNITS_CONV_MIN))
  {
    db_app_user.unit_conv_factor = (uint32_t) COMPARAM(2);
  }

  if (param_ok(&CTRL_COMP, 3, APP_DB_U_SHOW_LBOZ_MAX, APP_DB_U_SHOW_LBOZ_MIN))
  {
    db_app_user.show_lboz = (uint8_t) COMPARAM(3);
  }

  return MESS_OK;
}


/** mv/V reading */
uint8_t handle_VAL(struct s_comms_control *comms_ctrl, char *s)
{
  if (comms_ctrl->cmd_query)
  {
    comms_net_rfprintf(comms_ctrl, "%d" COMMS_TERMINATOR,
                       get_abs_mv_v()
                      );
    return REPLY_SENT;
  }

  /* No command for this */
  return MESS_ERROR;
}


/** Which serial port is this? */
uint8_t handle_WFP(struct s_comms_control *comms_ctrl, char *s)
{
  enum eBDXWFPSerialPort wfpport;

  if (comms_ctrl->cmd_query)
  {
    /*
     * comms_task_list_get_port_from_device() returns a fixed number
     * for the fixed ports, and a variable number above SERPORT_NETWORK_START
     * for a network port depending on the number of TCP connections.
     *
     * xlat_serial_port_to_bdxwfp() deals with this.
     */
    wfpport = xlat_serial_port_to_bdxwfp(comms_task_list_get_port_from_device(comms_ctrl->serdev));
    comms_net_rfprintf(comms_ctrl, "%u" COMMS_TERMINATOR,
                       (unsigned) wfpport
                      );
    return REPLY_SENT;
  }
  else
  {
    return MESS_ERROR;
  }
}


/** Weighing mode */
uint8_t handle_WMD(struct s_comms_control *comms_ctrl, char *s)
{
  enum eWMDBuildType wmd_scale_type;
  enum eWeApprovalMode we_approval_mode;

  if (comms_ctrl->cmd_query)
  {
    wmd_scale_type = xlat_adc_library_scale_type_to_wmd(adcSuperStructure.adcUserDatabase.scaleType);
    we_approval_mode = xlat_adc_library_approval_mode_to_we2111(adcSuperStructure.adcUserDatabase.approvalMode);
    comms_net_rfprintf(comms_ctrl, "%u,%u" COMMS_TERMINATOR,
                       wmd_scale_type + 1,
                       we_approval_mode
                      );
    return REPLY_SENT;
  }

  /* Can only set these with full access */
  if (!comms_ctrl->passcode_ok)
  {
    return MESS_ERROR;
  }

  /* Param 0 - Build Type */
  if (param_ok(&CTRL_COMP, 0, NUM_BUILD_TYPE, 1))
  {
    wmd_scale_type = (enum eWMDBuildType) (COMPARAM(0) - 1);
    adcSuperStructure.adcUserDatabase.scaleType = xlat_wmd_scale_type_to_adc_library(wmd_scale_type);
  }

  /* Param 1 - approval/weigh mode */
  if (param_ok(&CTRL_COMP, 1, WE_APPROVAL_MODE_NUM - 1, 0))
  {
    we_approval_mode = (enum eWeApprovalMode) COMPARAM(1);
    adcSuperStructure.adcUserDatabase.approvalMode = xlat_we2111_approval_mode_to_adc_library(we_approval_mode);
  }

  comms_reinit_adc_and_setpoints();
  return MESS_OK;
}


static void send_ZST(struct s_comms_control *comms_ctrl)
{
  uint8_t zr;

  /* Convert from ADC library setting */
  zr = xlat_adc_library_zero_range_to_we2111(adcSuperStructure.adcUserDatabase.rangeOfZero);

  comms_net_rfprintf(comms_ctrl, "%u,%u,%u,%ld" COMMS_TERMINATOR,
                     adcSuperStructure.adcUserDatabase.zeroOnStart,
                     adcSuperStructure.adcUserDatabase.zeroTracking,
                     zr + 1,
                     adcSuperStructure.adcUserDatabase.zeroBand
                    );
}


/** Zero settings */
uint8_t handle_ZST(struct s_comms_control *comms_ctrl, char *s)
{
  bool bump;

  if (comms_ctrl->cmd_query)
  {
    send_ZST(comms_ctrl);
    return REPLY_SENT;
  }

  bump = false;
  if (!comms_ctrl->passcode_ok)
  {
    return MESS_ERROR;
  }

  /* Param 0 - zero on start */
  if (param_ok(&CTRL_COMP, 0, ADC_DB_USER_ZEROONSTART_MAXIMUM, ADC_DB_USER_ZEROONSTART_MINIMUM))
  {
    adcSuperStructure.adcUserDatabase.zeroOnStart = COMPARAM(0);
  }

  /* Param 1 - zero tracking */
  if (param_ok(&CTRL_COMP, 1, APP_DB_U_ADCZEROTRACK_MAX, APP_DB_U_ADCZEROTRACK_MIN))
  {
    adcSuperStructure.adcUserDatabase.zeroTracking = COMPARAM(1);
    bump = true;
  }

  /* Param 2 - range of zero */
  if (param_ok(&CTRL_COMP, 2, WE21_NUM_ZERO_RANGE, 1))
  {
    adcSuperStructure.adcUserDatabase.rangeOfZero = xlat_we2111_zero_range_to_adc_library(COMPARAM(2) - 1);
    bump = true;
  }

  /* Param 3 - zero band */
  if (param_ok(&CTRL_COMP, 3, APP_DB_U_ADCZEROBAND_MAX, APP_DB_U_ADCZEROBAND_MIN))
  {
    adcSuperStructure.adcUserDatabase.zeroBand = COMPARAM(3);
    bump = true;
  }

  if (bump)
  {
    comms_reinit_adc_and_setpoints();
  }
  return MESS_OK;
}
