/*
 * comms_hdlr_passcode.c
 *
 * Rincmd and HBM comms share common passcode handling
 *
 *  Created on: Nov 27, 2017
 *      Author: ian
 */
#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <rincmd/rc_public.h>
#include "comms_util.h"
#include "comms_errors.h"
#include "comms_params.h"
#include "rc_gen_basic.h"
#include "we_config.h"
#include "app_db_user.h"

#include "comms_hdlr_passcode.h"


/**
 * The number of invalid attempts to access DPS, DPF and rincmd RINREG_BASIC_SET_FULL_PERMISSIONS RINREG_BASIC_SET_SAFE_PERMISSIONS.
 *
 * This is global rather than in struct s_comms_control because
 * we only allow MAX_PASSCODE_INVALID_ATTEMPTS in total
 * across all comms connections. The only way to clear
 * this is to reset the unit
 */
static int passcode_invalid_attempts = 0;


/** Rincmd check full passcode for access. TODO this should be on a per connection basis */
uint32_t rincmd_check_fullPC = 0;


/** Rincmd check safe passcode for access. TODO this should be on a per connection basis */
uint32_t rincmd_check_safePC = 0;


/** Full passcode */
uint8_t handle_DPF(struct s_comms_control *comms_ctrl, char *s)
{

  /* Check that there haven't been too many invalid passcode attempts */
  if (passcode_invalid_attempts > WECFG_MAX_PASSCODE_INVALID_ATTEMPTS)
  {
    comms_ctrl->passcode_ok = false;
    return MESS_ERROR;
  }

  /* No restrictions on access if passcode is not set */
  if (0 == db_app_user.full_passcode)
  {
    comms_ctrl->passcode_ok = true;
  }

  /* Query */
  if (comms_ctrl->cmd_query)
  {
    /* Yes it is inverted */
    comms_net_rfprintf(comms_ctrl, "%u" COMMS_TERMINATOR,
                       (comms_ctrl->passcode_ok) ? 0 : 1
                      );
    return REPLY_SENT;
  }

  if (param_ok(&CTRL_COMP, 0, APP_DB_U_FULLPASS_MAX, APP_DB_U_FULLPASS_MIN))
  {
    /* If passcode matches, give access */
    if ((unsigned long) COMPARAM(0) == db_app_user.full_passcode)
    {
      comms_ctrl->passcode_ok = true;
      return MESS_OK;
    }
    else
    {
      /* No match, but if already have access, then change passcode */
      if (comms_ctrl->passcode_ok)
      {
        db_app_user.full_passcode = COMPARAM(0);

        /* Passcode of 0 means unlocked */
        comms_ctrl->passcode_ok = (0 == db_app_user.full_passcode);
        return MESS_OK;
      }
      else
      {
        /* No match and no access already granted */
        if (passcode_invalid_attempts <= WECFG_MAX_PASSCODE_INVALID_ATTEMPTS)
        {
          passcode_invalid_attempts++;
        }
        return MESS_ERROR;
      }
    }
  }
  return MESS_OK;
}


/** Safe passcode */
uint8_t handle_DPS(struct s_comms_control *comms_ctrl, char *s)
{
  if (comms_ctrl->cmd_query)
  {
    /* Yes it is inverted */
    comms_net_rfprintf(comms_ctrl, "%u" COMMS_TERMINATOR,
                       (0 == db_app_user.safe_passcode) ? 0 : 1
                      );
    return REPLY_SENT;
  }

  /* Need full passcode access to have been granted to be able to change safe passcode */
  if (!comms_ctrl->passcode_ok)
    return MESS_ERROR;

  if (param_ok(&CTRL_COMP, 0, APP_DB_U_SAFEPASS_MAX, APP_DB_U_SAFEPASS_MIN))
  {
    db_app_user.safe_passcode = COMPARAM(0);
  }
  return MESS_OK;
}


/** Initialise access to registers if passcodes are 0 */
void rincmd_initPasscode(struct s_rc_localRegisterTable *topLevel)
{
  uint32_t priv;

  priv = (0 == db_app_user.full_passcode) ? RC_PRIVILEGE_FULL :
         (0 == db_app_user.safe_passcode) ? RC_PRIVILEGE_SAFE : RC_PRIVILEGE_GENERAL;
  rc_setPrivilege(topLevel, priv);
}


/** Rincmd handler for enabling access to safe and full passcode protected registers */
bool rincmd_checkPasscode(char *val, size_t len, size_t maxLen, struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, uint8_t opt)
{
  if (RC_FUNC_WRITE_POST == opt)
  {
    if (RINREG_BASIC_SET_FULL_PERMISSIONS == reg->addr)
    {
      /* Check that there haven't been too many invalid passcode attempts */
      if (passcode_invalid_attempts > WECFG_MAX_PASSCODE_INVALID_ATTEMPTS)
      {
        rc_setPrivilege(topLevel, RC_PRIVILEGE_GENERAL);
        return false;
      }

      /* If passcode matches, give access */
      if (rincmd_check_fullPC == db_app_user.full_passcode)
      {
        rc_setPrivilege(topLevel, RC_PRIVILEGE_FULL);
        passcode_invalid_attempts = 0;
      }
      else
      {
        /* No match */
        if (passcode_invalid_attempts <= WECFG_MAX_PASSCODE_INVALID_ATTEMPTS)
        {
          rc_setPrivilege(topLevel, RC_PRIVILEGE_GENERAL);
          passcode_invalid_attempts++;
        }
      }
    }

    if (RINREG_BASIC_SET_SAFE_PERMISSIONS == reg->addr)
    {
      /* If passcode matches, give access */
      if (rincmd_check_safePC == db_app_user.safe_passcode)
      {
        rc_setPrivilege(topLevel, RC_PRIVILEGE_SAFE);
      }
      else
      {
        rc_setPrivilege(topLevel, RC_PRIVILEGE_GENERAL);
      }
    }
  }

  if (RC_FUNC_READ_PRE == opt)
  {
    /* Clear both variables */
    rincmd_check_fullPC = 0;
    rincmd_check_safePC = 0;
  }

  return false;
}
