/*
 * comms_hdlr_rdb.c
 *
 *  Created on: Aug 1, 2014
 *      Author: ian
 *
 * Rinstrum debug command. Give access to syslog and other system files
 * we might be interested in
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <rinexec/el_runcmd.h>
#include <rinexec/el_pipe.h>
#include <rinblock/rinblock.h>
#include "we_debug.h"
#include "comms_errors.h"
#include "we_config.h"
#include "cmdline.h"
#include "version.h"
#include "app_db_factory.h"
#include "comms_send_buffer.h"
#include "comms_util.h"

#include "comms_hdlr_rdb.h"


/** Struct for RDB handling output to comms channel or FILE */
struct s_rdb_control
{
  struct s_comms_control *comms_ctrl;   /*!< Set this if the output is to be sent to a comms channel */
  FILE *output_file;                    /*!< Set this if the output should be written to a file */
};


/** Special RDB printf to send to comms channel or FILE */
static int send_rdb_printf(struct s_rdb_control *rdb, const char *format, ...)
{
  va_list args;
  int length = 0;
  const void *buffer = NULL;
  size_t buflength = 0;

  va_start(args, format);

  if (NULL != rdb->comms_ctrl)
  {
    /* Send to comms channel */
    length = comms_net_rvfprintf(rdb->comms_ctrl, format, args);

    if (comms_send_buffer_get_buffer(rdb->comms_ctrl->csb, &buffer, &buflength) && buffer != NULL && buflength > 100)
    {
      (void) rwrite(rdb->comms_ctrl->serdev, buffer, buflength, 0);
      (void) comms_send_buffer_clear(rdb->comms_ctrl->csb);
    }
  }
  else if (NULL != rdb->output_file)
  {
    /* Send to output file */
    length = vfprintf(rdb->output_file, format, args);
  }

  va_end(args);
  return length;
}


static void send_rdb_break(struct s_rdb_control *rdb)
{
  send_rdb_printf(rdb, "\n--------------------------------------------------------------------------------\n");
}


static void send_rdb_header(struct s_rdb_control *rdb, const char *header)
{
  send_rdb_break(rdb);
  send_rdb_printf(rdb, "Contents of \"%s\":\n", header);
}


/** Dump the contents of a file to the comms channel */
static uint8_t send_rdb_file(struct s_rdb_control *rdb, const char *filename)
{
  #define FILEDUMP_BUFSIZE 1024
  FILE *file = NULL;
  char *buffer = NULL;
  int result;

  send_rdb_header(rdb, filename);

  buffer = malloc(FILEDUMP_BUFSIZE);
  if (NULL == buffer)
  {
    rdebug_printf(LOG_ERR, "send_rdb_file: failed to allocate file dump buffer\n");
    result = MESS_ERROR;
    goto srf_fail;
  }

  file = fopen(filename, "r");
  if (NULL == file)
  {
    rdebug_printf(LOG_ERR, "send_rdb_file: failed to open file %s\n", filename);
    result = MESS_ERROR;
    goto srf_fail;
  }
  if (fcntl(fileno(file), F_SETFD, FD_CLOEXEC) < 0)
  {
    result = MESS_ERROR;
    goto srf_fail;
  }

  while (NULL != fgets(buffer, FILEDUMP_BUFSIZE, file))
  {
    send_rdb_printf(rdb, "%s", buffer);
  }
  result = REPLY_SENT;

srf_fail:
  if (NULL != buffer)
    free(buffer);
  if (NULL != file)
    fclose(file);
  return result;
}


/** Callback function used with send_rdb_process() */
enum e_el_result send_rdb_process_callback(void *usrptr, const char *data, size_t bytes_read)
{
  struct s_rdb_control *rdb = usrptr;

  if (NULL != rdb && NULL != data && bytes_read > 0)
  {
    send_rdb_printf(rdb, "%.*s", bytes_read, data);
  }
  return elr_success;
}


/** Run a process and dump the reply from stdout to the comms channel */
static uint8_t send_rdb_process(struct s_rdb_control *rdb, const char *const cmd_args[])
{
  struct s_runcmd_pipe_config rcpc;
  enum e_el_result elresult;
  int ret_code;
  char *cmdbuf;
  size_t cmdlen;
  const char *const *cmdptr;
  uint8_t result;

  /* Calculate length of all command arguments, plus spaces between */
  for (cmdptr = cmd_args, cmdlen = 0; *cmdptr != NULL; cmdptr++)
  {
    cmdlen += strlen(*cmdptr) + sizeof(char);
  }
  cmdlen += sizeof(char);

  /* Allocate string, and assemble command */
  cmdbuf = calloc(cmdlen, sizeof(char));
  if (NULL != cmdbuf)
  {
    for (cmdptr = cmd_args; *cmdptr != NULL; cmdptr++)
    {
      strncat(cmdbuf, *cmdptr, cmdlen - strlen(cmdbuf));
      strcat(cmdbuf, " ");
    }
    if (cmdbuf[strlen(cmdbuf) - 1] == ' ')
      cmdbuf[strlen(cmdbuf) - 1] = '\0';
  }
  else
  {
    cmdbuf = (char *) cmd_args[0];
  }

  send_rdb_header(rdb, cmdbuf);

  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_callback(&rcpc, erp_stdout, send_rdb_process_callback, rdb);

  /* Run opkg and check the return code */
  elresult = el_run_command(cmd_args[0], cmd_args, &ret_code, &rcpc, true, true, true);
  if (elr_success != elresult || !WIFEXITED(ret_code))
  {
    rdebug_printf(LOG_ERR, "send_rdb_process: \"%s\" failed\n", cmdbuf);
    send_rdb_printf(rdb, "send_rdb_process: \"%s\" failed\n", cmdbuf);
    result = MESS_ERROR;
    goto fail;
  }

  /* Check result code */
  if (0 != WEXITSTATUS(ret_code))
  {
    rdebug_printf(LOG_ERR, "send_rdb_process: \"%s\" return code not 0: %d\n", cmdbuf, ret_code);
    send_rdb_printf(rdb, "send_rdb_process: \"%s\" return code not 0: %d\n", cmdbuf, ret_code);
    result = MESS_ERROR;
    goto fail;
  }

  /* Run successfully */
  result = REPLY_SENT;

  /* Something went wrong */
fail:
  if (NULL != cmdbuf && cmdbuf != cmd_args[0])
    free(cmdbuf);
  return result;
}


/** Print info on a file descriptor */
static void send_rdb_fd_info(struct s_rdb_control *rdb, int fd)
{
  int fd_flags;
  int fl_flags;
  ssize_t s;
  struct flock fl;
  char path[256];
  char buf[256];

  fd_flags = fcntl(fd, F_GETFD);
  if (fd_flags < 0)
  {
    /* This file descriptor not open */
    return;
  }

  /* Get file descriptor flags */
  fl_flags = fcntl(fd, F_GETFL);
  if (fl_flags < 0)
    return;

  snprintf(path, 256, "/proc/self/fd/%d", fd);
  memset(&buf[0], 0, 256);
  s = readlink(path, &buf[0], 256);
  if (s < 0)
  {
    send_rdb_printf(rdb, " (%s): not available", path);
    return;
  }
  send_rdb_printf(rdb, "%d (%s): ", fd, buf);

  /* CLOEXEC */
  if (fd_flags & FD_CLOEXEC)
    send_rdb_printf(rdb, "cloexec ");

  /* File status */
  if (fl_flags & O_APPEND)
    send_rdb_printf(rdb, "append ");
  if (fl_flags & O_NONBLOCK)
    send_rdb_printf(rdb, "nonblock ");

  /* Access mode */
  if (fl_flags & O_RDONLY)
    send_rdb_printf(rdb, "read-only ");
  if (fl_flags & O_RDWR)
    send_rdb_printf(rdb, "read-write ");
  if (fl_flags & O_WRONLY)
    send_rdb_printf(rdb, "write-only ");

  if (fl_flags & O_DSYNC)
    send_rdb_printf(rdb, "dsync ");
  if (fl_flags & O_RSYNC)
    send_rdb_printf(rdb, "rsync ");
  if (fl_flags & O_SYNC)
    send_rdb_printf(rdb, "sync ");

  /* File locking */
  fl.l_type = F_WRLCK;
  fl.l_whence = 0;
  fl.l_start = 0;
  fl.l_len = 0;
  fcntl(fd, F_GETLK, &fl);
  if (F_UNLCK != fl.l_type)
  {
    if (F_WRLCK == fl.l_type)
      send_rdb_printf(rdb, "write-locked");
    else
      send_rdb_printf(rdb, "read-locked");
    send_rdb_printf(rdb, "(pid:%d) ", (int) fl.l_pid);
  }
  send_rdb_printf(rdb, "\n");
}


/** Print info on all file descriptors */
static void send_rdb_all_fd_info(struct s_rdb_control *rdb)
{
  int num_handles;
  int fd;

  send_rdb_header(rdb, "File descriptors");

  num_handles = getdtablesize();
  for (fd = 0; fd < num_handles; fd++)
  {
    send_rdb_fd_info(rdb, fd);
  }
}


/** Print info about rinblocks */
static void send_rdb_rinblock_info(struct s_rdb_control *rdb)
{
  size_t i;

  send_rdb_header(rdb, "Rinblock modules");
  for (i = 0; i < MAX_RINBLOCKS; i++)
  {
    send_rdb_printf(rdb, "Rinblock module %u (addr %u, slot %u):\n", (unsigned) i, (unsigned) rinblocks[i].iic_address, (unsigned) rinblocks[i].bus_slot);
    send_rdb_printf(rdb, "  Present: %s\n", (rinblocks[i].present) ? "Yes" : "No");
    send_rdb_printf(rdb, "  Programmed: %s\n", (rinblocks[i].programmed) ? "Yes" : "No");
    send_rdb_printf(rdb, "  Running: %s\n", (rinblocks[i].running) ? "Yes" : "No");

    if (rinblocks[i].present && rinblocks[i].programmed && rinblocks[i].running)
    {
      send_rdb_printf(rdb, "  Model number: %u\n", (unsigned) rinblocks[i].tag.model_number);
      send_rdb_printf(rdb, "  Serial number: %u\n", (unsigned) rinblocks[i].tag.serial_number);
      send_rdb_printf(rdb, "  Software version: %u.%02u\n", (unsigned) rinblocks[i].tag.software_version / 100, (unsigned) rinblocks[i].tag.software_version % 100);
      send_rdb_printf(rdb, "  Interface version: %u\n", (unsigned) rinblocks[i].tag.interface_version);
      send_rdb_printf(rdb, "  Interface type: %u\n", (unsigned) rinblocks[i].tag.interface_type);
      send_rdb_printf(rdb, "  Hardware version: %u\n", (unsigned) rinblocks[i].tag.hardware_version);
    }
  }
}


/** Dump lots of things that might be interesting when trying to debug problems */
static uint8_t send_RDB(struct s_rdb_control *rdb)
{
  const char *const uname_args[] = { "uname", "-a", NULL };
  const char *const date_args[] = { "date", NULL };
  const char *const uptime_args[] = { "uptime", NULL };
  const char *const df_args[] = { "df", NULL };
  const char *const opkg_installed_args[] = { "opkg-cl", "list-installed", NULL };
  const char *const opkg_info_args[] = { "opkg-cl", "info", NULL };
  const char *const ifconfig_args[] = { "ifconfig", NULL };
  const char *const route_args[] = { "route", "-n", NULL };
  const char *const arp_args[] = { "arp", "-a", "-n", NULL };
  const char *const netstat_args[] = { "netstat", "-an", NULL };
  const char *const mount_args[] = { "mount", NULL };
  const char *const fwprintenv_args[] = { "fw_printenv", NULL };
  #define HEXDUMP_NVRAM_NUM_ARGS 4
  char *hexdump_nvram1_args[HEXDUMP_NVRAM_NUM_ARGS];
  char *hexdump_nvram2_args[HEXDUMP_NVRAM_NUM_ARGS];
  const char *const ps_args[] = { "ps", NULL };
  const char *const lsmod_args[] = { "lsmod", NULL };
  const char *const lsusb_args[] = { "lsusb", NULL };

  /* Setup hexdump for nvram arguments */
  hexdump_nvram1_args[0] = "hexdump";
  hexdump_nvram1_args[1] = "-Cv";
  hexdump_nvram1_args[2] = we_configuration.nvram_device1;
  hexdump_nvram1_args[3] = NULL;
  hexdump_nvram2_args[0] = "hexdump";
  hexdump_nvram2_args[1] = "-Cv";
  hexdump_nvram2_args[2] = we_configuration.nvram_device2;
  hexdump_nvram2_args[3] = NULL;

  /* Start it */
  send_rdb_break(rdb);
  send_rdb_printf(rdb, "Start of debug data for %s S/N %lu\n", db_app_factory.model_str, (unsigned long) db_app_factory.serial_no);

  /* System */
  send_rdb_process(rdb, uname_args);
  send_rdb_process(rdb, date_args);
  send_rdb_process(rdb, uptime_args);
  send_rdb_file(rdb, "/var/log/messages");

  /* Filesystems */
  send_rdb_process(rdb, mount_args);
  send_rdb_process(rdb, df_args);
  send_rdb_process(rdb, opkg_installed_args);
  send_rdb_process(rdb, opkg_info_args);

  /* Networking */
  send_rdb_process(rdb, ifconfig_args);
  send_rdb_process(rdb, route_args);
  send_rdb_process(rdb, arp_args);
  send_rdb_file(rdb, "/etc/resolv.conf");
  send_rdb_file(rdb, "/etc/hosts");
  send_rdb_process(rdb, netstat_args);

  /* Configuration */
  send_rdb_file(rdb, "/etc/rc.d/rc.conf");
  send_rdb_file(rdb, "/etc/mdev.conf");
  send_rdb_file(rdb, cmdline_config_file);
  send_rdb_file(rdb, we_configuration.factorydb1_file);
  send_rdb_file(rdb, we_configuration.factorydb2_file);
  send_rdb_file(rdb, we_configuration.userdb1_file);
  send_rdb_file(rdb, we_configuration.userdb2_file);
  send_rdb_file(rdb, we_configuration.licencedb1_file);
  send_rdb_file(rdb, we_configuration.licencedb2_file);
  send_rdb_file(rdb, "/etc/fw_env.config");
  send_rdb_process(rdb, fwprintenv_args);
  send_rdb_process(rdb, (const char *const *) hexdump_nvram1_args);
  send_rdb_process(rdb, (const char *const *) hexdump_nvram2_args);

  /* Memory/processes/kernel */
  send_rdb_file(rdb, "/proc/meminfo");
  send_rdb_process(rdb, ps_args);
  send_rdb_process(rdb, lsmod_args);
  send_rdb_process(rdb, lsusb_args);
  send_rdb_all_fd_info(rdb);
  send_rdb_file(rdb, "/sys/kernel/rinhw_revision/we2111_board_rev");

  /* Rinblock modules */
  send_rdb_rinblock_info(rdb);

  /* Finish it off */
  send_rdb_break(rdb);
  send_rdb_printf(rdb, "End of debug data");
  send_rdb_break(rdb);

  return REPLY_SENT;
}


/** Rinstrum debug command */
uint8_t handle_RDB(struct s_comms_control *comms_ctrl, char *s)
{
  struct s_rdb_control rdb_control;

  /* This could dump sensitive information, so require full passcode */
  if (!comms_ctrl->passcode_ok)
    return MESS_ERROR;

  /* Query only */
  if (comms_ctrl->cmd_query)
  {
    /* Direct output to comms channel */
    rdb_control.comms_ctrl = comms_ctrl;
    rdb_control.output_file = NULL;

    /*
     * Even though send_rdb_printf() only writes chunks of > 100 bytes to the
     * underlying socket, any data left in the comms send buffer gets written out
     * by handle_comms() in comms_net.c (back up the call chain), after this
     * function returns
     */
    return send_RDB(&rdb_control);
  }

  /* No command */
  return MESS_ERROR;
}


/** Save the debug info into a file */
bool save_debug_info_to_file(FILE *file)
{
  struct s_rdb_control rdb_control;

  /* Direct output to comms channel */
  rdb_control.comms_ctrl = NULL;
  rdb_control.output_file = file;
  return (send_RDB(&rdb_control) == REPLY_SENT);
}
