/*
 * comms_net.c
 *
 *  Created on: 02/03/2011
 *      Author: ian
 */
#define _XOPEN_SOURCE 700
#include <setjmp.h>
#include <ctype.h>
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/select.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/ucos_ii.h>
#include <rinweigh/adcinterface.h>
#include <rincmd/rc_public.h>
#include <rincmd/rc_registers.h>
#include <rincmd/rc_streaming.h>
#include "asciidef.h"
#include "comms_stxetx.h"
#include "comms_dispatch.h"
#include "comms_hdlr_nz.h"
#include "database.h"
#include "iface.h"
#include "modbus_old.h"
#include "meas.h"
#include "comms_params.h"
#include "comms_stxetx.h"
#include "comms_errors.h"
#include "comms_util.h"
#include "comms_setup.h"
#include "comms_hdlr_passcode.h"
#include "serial_levels.h"
#include "adcthread.h"
#include "we_config.h"
#include "comms_regtable.h"
#include "rc_gen_basic.h"
#include "app_db_user.h"

#include "comms_net.h"


/** Responses to HBM commands, match MESS_ constants */
static const char hbm_comms_mess_responses[MAX_COMMS_ERROR] = MESS_RESPONSES;


/** Call when a message has been handled to reset indexes */
static void comms_net_message_handled(struct s_comms_net *cn)
{
  size_t len;

  if (cn->incoming_mesg_length > cn->last_examined_mesg_posn)
  {
    /* There is some data left, copy it back to start of buffer */
    len = cn->incoming_mesg_length - cn->last_examined_mesg_posn;
    memcpy(&cn->incoming_mesg[0], &cn->incoming_mesg[cn->last_examined_mesg_posn], len);
    cn->last_examined_mesg_posn = 0;
    cn->incoming_mesg_length = len;
    cn->start_of_mesg = 0;
  }
  else
  {
    /* All data used, reset to start of incoming data buffer */
    cn->last_examined_mesg_posn = 0;
    cn->incoming_mesg_length = 0;
    cn->start_of_mesg = 0;
  }
}


/**
 * Clear out data from the incoming data buffer if it is full
 *
 * We could do something clever with only dumping data before start_of_mesg,
 * but not all message have a start character (ie HBM, rincmd without CRC),
 * so the safest method is to dump everything we have examined
 */
static void comms_net_clear_full_incoming_buffer(struct s_comms_net *cn)
{
  if (cn->incoming_mesg_length >= COMMS_NET_MAX_MESS_LEN)
  {
    comms_net_message_handled(cn);
  }
}


/** Process Rinstrum test commands (<STX><ETX> style commands) */
static void process_rintest_cmd(struct s_comms_net *cn)
{
  process_stxetx_cmd(&cn->comms_ctrl, &cn->incoming_mesg[cn->start_of_mesg]);
  comms_net_message_handled(cn);
}


/** Process Rinstrum rincmd messages (eg 20050005:;). Return true if handled, false if not */
static bool process_rincmd(struct s_comms_net *cn, bool hascrc)
{
  #define COMM_OUTPUT_BUFFSIZE 200
  rc_errcode_t rincmd_result;
  char outBuf[COMM_OUTPUT_BUFFSIZE] = {0};
  bool result;

  /* Try handling with rincmd */
  rincmd_result = rc_commandHandler(&cn->incoming_mesg[cn->start_of_mesg],
                                    hascrc,
                                    outBuf,
                                    COMM_OUTPUT_BUFFSIZE,   /* TODO rincmd buffer */
                                    cn->comms_ctrl.regTable,
                                    db_app_user.address);

  switch (rincmd_result)
  {
    case RC_HNDL_RET_REPLY:
      cn->comms_ctrl.reply_active = true;
      if (hascrc)
        comms_net_rfprintf(&cn->comms_ctrl, "\x01%s\x04", outBuf);
      else
        comms_net_rfprintf(&cn->comms_ctrl, "%s\r\n", outBuf);
      comms_net_message_handled(cn);
      result = true;
      break;

    case RC_HNDL_RET_NOREPLY:
      cn->comms_ctrl.reply_active = false;
      comms_net_message_handled(cn);
      result = true;
      break;

    case RC_HNDL_RET_ERR:
    default:
      result = false;
      break;
  }
  return result;
}


/** Process HBM commands (IDN; style commands) */
static enum eCommsError process_hbm_cmd(struct s_comms_net *cn)
{
  char *msg;
  enum eCommsError respond;
  uint16_t cmd;
  size_t length;

  msg = &cn->incoming_mesg[cn->start_of_mesg];
  length = strlen(msg);

  if (0 == length)
  {
    respond = MESS_ERROR;
    goto phc_done;
  }

  /* This is where any longjmp will jump back to */
  if (setjmp(cn->comms_ctrl.comms_params.parse_error) != 0)
  {
    comms_net_rfprintf(&cn->comms_ctrl, "?" COMMS_TERMINATOR);
    respond = MESS_ERROR;
  }
  else
  {
    respond = REPLY_SENT;
    if (length < 3)
    {
      respond = MESS_ERROR;
    }
    else
    {
      /* Can handle commands in lower or upper case */
      msg[0] = toupper(msg[0]);
      msg[1] = toupper(msg[1]);
      msg[2] = toupper(msg[2]);

      if ('S' == msg[0])
      {
        /* Handle the special case of Sxx messages here */
        respond = handle_S(&cn->comms_ctrl, msg);
      }
      else if (cn->comms_ctrl.unit_selected && !cn->comms_ctrl.msv_cmd_streaming)
      {
        /* Calculate the key from the 3 character header base 26 */
        cmd = HBM_CMD_CHARS_TO_NUM(msg[0], msg[1], msg[2]);
        msg += 3;

        /* Is it a query (eg IDN?) */
        if ('?' == msg[0])
        {
          cn->comms_ctrl.cmd_query = true;
          msg++;
        }
        else
        {
          cn->comms_ctrl.cmd_query = false;
        }

        /* Handle the command */
        respond = comms_handle_hbm_cmd(cmd, &cn->comms_ctrl, msg);
      }
    }

    /* Send response */
    switch (respond)
    {
      case REPLY_SENT:
        /* No need to send anything if the reply has already been sent */
        break;

      default:
        comms_net_rfprintf(&cn->comms_ctrl, "%c" COMMS_TERMINATOR,
                           hbm_comms_mess_responses[respond]
                          );
        break;
    }
  }

phc_done:
  /* Ready for next message */
  comms_net_message_handled(cn);

  return respond;
}


/** Send rin comms streaming at 10Hz, 5Hz & 1Hz */
static void comms_rin_autosend_10hz(struct s_comms_control *comms_ctrl)
{
  uint32_t current_time;
  uint8_t i;
  rc_errcode_t ret;
  char buf[RC_SIZE_BUFFER];

  /*
   * This may be called faster than 10Hz if we are also streaming at the sync frequency,
   * so we need to check how long since the last call
   */
  current_time = OSTimeGet();
  if (current_time > comms_ctrl->rincomms_last_10hz_time + OS_TICKS_PER_SEC / 10)
  {
    /* Iterate across each stream set and send the data if there is any to send. */
      for (i = 0; i < RC_MAX_STREAM_SETS; i ++)
      {
        ret = rc_streamDataFreq(comms_ctrl->regTable, i, buf, RC_SIZE_BUFFER, 0x05);
        if (ret == RC_STREAM_ERR_NONE)
        {
          comms_ctrl->reply_active = true;
          comms_net_rfprintf(comms_ctrl, "%s\r\n", buf);
        }
      }

      /* Save the time we can count how long it should be until this can run again. */
      comms_ctrl->rincomms_last_10hz_time = current_time;
  }
}


/** Send rin comms streaming at the sync frequency */
static void comms_rin_autosend_sync(struct s_comms_control *comms_ctrl)
{
  uint8_t i;
  rc_errcode_t ret;
  char buf[RC_SIZE_BUFFER];

  /* Iterate across each stream set and send the data if there is any to send. */
  for (i = 0; i < RC_MAX_STREAM_SETS; i ++)
  {
    ret = rc_streamDataSync(comms_ctrl->regTable, i, buf, RC_SIZE_BUFFER, 0x05);

    if (ret == RC_STREAM_ERR_NONE)
    {
      comms_ctrl->reply_active = true;
      comms_net_rfprintf(comms_ctrl, "%s\r\n", buf);
    }
  }
}


static void comms_autosend(struct s_comms_control *comms_ctrl)
{
  size_t buf_free;
  bool buf_has_room;
  int32u rdgid;

  /* Sync frequency streaming */
  rdgid = adcSuperStructure.adcDataWorking.rdgId;
  if (comms_ctrl->last_rdg_id != rdgid)
  {
    /* Rin command streaming at sync frequency */
    comms_rin_autosend_sync(comms_ctrl);

    /* Old skool Kr & MSV streaming */
    if (comms_ctrl->kr_cmd_streaming || comms_ctrl->msv_cmd_streaming)
    {
      /* We have something to send, but do we have room in the transmit buffer. For reference TCP buffer = 16k, serial buffer = 1k. */
      buf_free = get_serial_tx_free(comms_ctrl->serdev);
      buf_has_room = (buf_free > WECFG_NET_BUF_REQUIRED_FREE);

      if (buf_has_room)
      {
        if (comms_ctrl->kr_cmd_streaming)
        {
          /* Kr data streaming */
          SendKrData(comms_ctrl);
        }
        else if (comms_ctrl->msv_cmd_streaming)
        {
          /* MSV data streaming */
          comms_send_oiml(comms_ctrl);
          if (comms_ctrl->msv_count > 0)
          {
            if (--comms_ctrl->msv_count == 0)
            {
              comms_ctrl->msv_cmd_streaming = false;
              if (comms_ctrl->send_extra_line_termination)
              {
                comms_net_rfprintf(comms_ctrl, COMMS_TERMINATOR);
              }
            }
          }
        }
      }
    }
    comms_ctrl->last_rdg_id = rdgid;
  }

  /* Rin comms streaming at 10Hz */
  comms_rin_autosend_10hz(comms_ctrl);
}


/** Return the number of ticks of delay for highest frequency we are currently streaming at */
static void comms_get_delay_ticks_for_highest_streaming_freq(struct s_comms_control *comms_ctrl, uint16_t *ticks, struct timespec *ts)
{
  uint32_t streaming_mask;
  uint32_t sync_freq;
  uint32_t fastest_freq;

  /* Get current sync frequency in Hz * 10 (10Hz = 100) */
  sync_freq = AdcSyncFrequencies[adcSuperStructure.adcUserDatabase.syncFreqIdx];

  /* What frequency streams are currently running */
  streaming_mask = rc_getStreamFreqMask(comms_ctrl->regTable);

  /* Work out what is the fastest stream in Hz * 10 */
  fastest_freq = 0;
  if ((streaming_mask & RC_STREAM_1HZ_MASK) != 0)
    fastest_freq = 1 * 10;
  if ((streaming_mask & RC_STREAM_5HZ_MASK) != 0)
    fastest_freq = 5 * 10;
  if ((streaming_mask & RC_STREAM_10HZ_MASK) != 0)
    fastest_freq = 10 * 10;
  if ((((streaming_mask & RC_STREAM_AUTO_MASK) != 0)
        || ((streaming_mask & RC_STREAM_CHANGE_MASK) != 0))
      && (fastest_freq < sync_freq))
    fastest_freq = sync_freq;

  /* Old skool Kr & MSV streaming are at the sync frequency too */
  if (comms_ctrl->kr_cmd_streaming || comms_ctrl->msv_cmd_streaming)
    fastest_freq = sync_freq;

  /* Work out read delay */
  if (0 == fastest_freq)
  {
    /* No-one is streaming, just time out every now and again (10 times per sec) */
    *ticks = OS_TICKS_PER_SEC / 10;
    ts->tv_sec = 0;
    ts->tv_nsec = 1000000000 / 10;
  }
  else
  {
    /*
     * Calculate timeout ticks based on sync frequency. It would be nice to have finer grain control of the timeout...
     *
     * Because we are looking for reading id change in comms_autosend(), we want to poll at twice the fastest frequency
     */
    fastest_freq *= 2;
    ts->tv_sec = 0;
    ts->tv_nsec = 1000000000 / fastest_freq;
    ts->tv_nsec *= 10;
    *ticks = (OS_TICKS_PER_SEC * 10) / fastest_freq;
    if (0 == *ticks)
      *ticks = 1;
  }
}


enum e_cn_read_result
{
  crr_success,
  crr_timeout,
  crr_failure
};


/**
 * When using linux reads, if an error occurs, the
 * rinstrum device doesn't know anything about it.
 * So here we just do a rread() and that will fail and close
 * the socket.
 *
 * Later the ioctl(IOCTL_SER_GET_CTS_STATE) call will
 * determine that the socket within the rindevice has been closed, and
 * the thread will quit.
 */
static void comms_net_fixup_rindev_state(struct t_io_dev *rindev)
{
  char dummy;

  (void) rread(rindev, &dummy, sizeof(dummy), 1);
}


/** Do the necessary mucking around for a UDP read */
static enum e_cn_read_result comm_net_udp_read(struct s_comms_net *cn, int fd)
{
  ssize_t retval;
  socklen_t sockaddr_size;
  char dummy;
  struct sockaddr_in sender;
  enum e_cn_read_result res;
  size_t bufroom;

  if (0 == cn->udp_info.dest.sin_addr.s_addr ||
      0 == cn->udp_info.dest.sin_port)
  {
    /* This is a incoming UDP port, so just read data from anyone who sends it to us */
    bufroom = COMMS_NET_MAX_MESS_LEN - cn->incoming_mesg_length;
    sockaddr_size = sizeof(sender);
    retval = recvfrom(fd, &cn->incoming_mesg[cn->incoming_mesg_length], bufroom, 0,
                      (struct sockaddr *) &sender, &sockaddr_size);
    if (retval < 0)
    {
      res = crr_failure;
    }
    else
    {
      /* Need to let rindevice know where to send the data back to */
      sender.sin_addr.s_addr = ntohl(sender.sin_addr.s_addr);
      sender.sin_port = ntohs(sender.sin_port);
      rioctl(cn->comms_ctrl.serdev, IOCTL_SER_UDP_SET_DEST_SOCKADDR, &sender);
      cn->incoming_mesg_length += retval;
      res = crr_success;
    }
  }
  else
  {
    /* This is outgoing UDP port, so peek at what is available to get the source */
    sockaddr_size = sizeof(sender);
    retval = recvfrom(fd, &dummy, sizeof(dummy), MSG_PEEK,
                      (struct sockaddr *) &sender, &sockaddr_size);
    if (retval == sizeof(dummy))
    {
      /* There is data available, is it for us? */
      if (ntohl(sender.sin_addr.s_addr) == cn->udp_info.dest.sin_addr.s_addr &&
          ntohs(sender.sin_port) == cn->udp_info.dest.sin_port)
      {
        /* Data for us, read into our buffer */
        bufroom = COMMS_NET_MAX_MESS_LEN - cn->incoming_mesg_length;
        sockaddr_size = sizeof(sender);
        retval = recvfrom(fd, &cn->incoming_mesg[cn->incoming_mesg_length], bufroom, 0,
                          (struct sockaddr *) &sender, &sockaddr_size);
        if (retval < 0)
        {
          res = crr_failure;
        }
        else
        {
          cn->incoming_mesg_length += retval;
          res = crr_success;
        }
      }
      else
      {
        /* From someone else, so just dump it */
        sockaddr_size = sizeof(sender);
        retval = recvfrom(fd, &dummy, sizeof(dummy), 0,
                          (struct sockaddr *) &sender, &sockaddr_size);
        res = (retval < 0) ? crr_failure : crr_success;
      }
    }
    else
    {
      res = crr_failure;
    }
  }

  if (crr_failure == res)
    comms_net_fixup_rindev_state(cn->comms_ctrl.serdev);

  return res;
}


/**
 * Try to read as much as we can using linux read(),
 * but if that is not available, fall back to rinstrum rread()
 */
static enum e_cn_read_result comms_net_linux_read(struct s_comms_net *cn)
{
  int fd;
  size_t bufroom;
  uint8_t ch;
  ssize_t res;
  fd_set readfds;
  uint16_t ticks;
  struct timespec timeout;

  /*
   * This should never occur, because the buffer should have been emptied
   * prior to calling this, but never try to read into a full buffer because:
   *
   *  - With rread(), buffer overflows are bad news
   *  - With read(), a read length of 0 will drop a byte (we never get to read it)
   *    I'm not sure if this is a bug in uclibc/linux and the horrible mix of
   *    pthreads and select() that we use, as I couldn't reproduce it
   *    in a simple program either on the PC or C500. Anyway preventing a read of
   *    0 bytes will stop it from occurring.
   */
  if ( cn->incoming_mesg_length >= COMMS_NET_MAX_MESS_LEN)
    return crr_timeout;

  /*
   * Wait for incoming character or timeout.
   * The timeout here is set 100ms = 10 Hz when not MSV, Ks or rincmd streaming.
   * When streaming, we up frequency to 100Hz so we can send the responses at
   * the sync frequency. The rate limiting is done by checking the reading id in
   * comms_autosend(), and sending whenever it changes.
   *
   * Ideally we'd be triggered by the ADC thread,
   * but there is no easy way to pend on a rio device
   * and a sem_t.
   */
  comms_get_delay_ticks_for_highest_streaming_freq(&cn->comms_ctrl, &ticks, &timeout);

  /* Try to get linux file descriptor */
  fd = -1;
  if (rioctl(cn->comms_ctrl.serdev, IOCTL_GET_LINUX_FD, &fd) < 0 || fd < 0)
  {
    /* Nup, fall back to rinstrum rread() */
    res = rread(cn->comms_ctrl.serdev, &ch, sizeof(ch), ticks);
    if (res != sizeof(ch))
    {
      /* Timeout */
      return crr_timeout;
    }
    cn->incoming_mesg[cn->incoming_mesg_length++] = ch;
  }
  else
  {
    /* Got the linux file descriptor, select() for timeout, then read as much as we can using linux read() */
    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);
    res = pselect(fd + 1, &readfds, NULL, NULL, &timeout, NULL);
    if (res < 0)
    {
      /* Failure */
      comms_net_fixup_rindev_state(cn->comms_ctrl.serdev);
      return crr_failure;
    }
    else if (0 == res)
    {
      /* Timeout */
      return crr_timeout;
    }
    else if (FD_ISSET(fd, &readfds))
    {
      if (cn->is_udp)
      {
        /* UDP is a special case */
        if (comm_net_udp_read(cn, fd) != crr_success)
          return crr_failure;
      }
      else
      {
        /* Data available, read from fd */
        bufroom = COMMS_NET_MAX_MESS_LEN - cn->incoming_mesg_length;
        res = read(fd, &cn->incoming_mesg[cn->incoming_mesg_length], bufroom);
        if (res <= 0)
        {
          /* Failure */
          comms_net_fixup_rindev_state(cn->comms_ctrl.serdev);
          return crr_failure;
        }
        cn->incoming_mesg_length += res;
      }
    }
    else
    {
      /* Shouldn't ever get here */
      return crr_failure;
    }
  }
  return crr_success;
}


/** Return codes from comms_process_incoming() */
enum e_cn_process_incoming_result
{
  cpir_failure,             /*!< Something went horribly wrong, best kill off this task */
  cpir_cmd_none,            /*!< No command was found this time */
  cpir_cmd_rintest,         /*!< Found a Rinstrum <STX><ETX> test command */
  cpir_cmd_modbus,          /*!< Found a legacy Modbus ASCII command */
  cpir_cmd_hbm_rincmd,      /*!< Found a HBM IDN style command, or rincmd (non-crc) */
  cpir_cmd_rincmd_crc,      /*!< Found a rincmd CRC checksummed command */
};


/** Process incoming data, looking for start or end characters */
static enum e_cn_process_incoming_result comms_process_incoming(struct s_comms_net *cn, bool first_pass)
{
  size_t i;
  bool found_msg;
  enum e_cn_read_result crr;
  enum e_cn_process_incoming_result res;

  for (found_msg = false, res = cpir_cmd_none; !found_msg; )
  {
    /*
     * Since there may still be a valid message that we have read, but not processed,
     * look for terminating characters in the incoming data first, before trying to read more data
     */
    for (i = cn->last_examined_mesg_posn, found_msg = false; i < cn->incoming_mesg_length && !found_msg; i++)
    {
      switch (cn->incoming_mesg[i])
      {
        case ETX:
          cn->incoming_mesg[i] = '\0';
          res = cpir_cmd_rintest;
          found_msg = true;
          break;

        case EOT:
          cn->incoming_mesg[i] = '\0';
          res = cpir_cmd_rincmd_crc;
          found_msg = true;
          break;

        case ';':
        case CR:
        case LF:
          cn->incoming_mesg[i] = '\0';

          /* Figure out what kind of command it is */
          res = (cn->incoming_mesg[cn->start_of_mesg] == ':') ? cpir_cmd_modbus : cpir_cmd_hbm_rincmd;
          found_msg = true;
          break;

        case STX:
        case DC2:
        case SOH:
          /* Skip over start character */
          cn->start_of_mesg = i + 1;
          break;

        default:
          break;
      }
    }
    cn->last_examined_mesg_posn = i;

    if (!found_msg)
    {
      /* There was no message found in the incoming data buffer. We may need dump some data to make some room if the buffer is full */
      comms_net_clear_full_incoming_buffer(cn);
      if (first_pass)
      {
        /* On the first pass, try reading from port */
        first_pass = false;
        crr = comms_net_linux_read(cn);
        switch (crr)
        {
          default:
          case crr_failure:
            return cpir_failure;
            break;
          case crr_timeout:
            return cpir_cmd_none;
            break;
          case crr_success:
            break;
        }

        /* Set last time we received some data */
        clock_gettime(CLOCK_MONOTONIC, &cn->ts_last_rx_data);
      }
      else
      {
        /* On subsequent passes, don't do a read */
        return cpir_cmd_none;
      }
    }
  }

  return res;
}


static enum eCommsError handle_comms(struct s_comms_net *cn)
{
  enum e_cn_process_incoming_result result;
  enum eCommsError error = MAX_COMMS_ERROR;
  const void *buffer = NULL;
  size_t buflength = 0;
  size_t passes;
  bool quit;

  /* Clear the comms send buffer */
  comms_send_buffer_clear(cn->comms_ctrl.csb);

  /* Auto send if required */
  comms_autosend(&cn->comms_ctrl);

  /* Process all received incoming messages and bundle the responses into a single write */
  for (passes = 0, quit = false; !quit; passes++)
  {
    /* Retrieve characters from serial port, looking for message delimiters */
    result = comms_process_incoming(cn, (0 == passes));

    /* Try processing the message as rintest/rincmd/hbm/modbus legacy */
    switch (result)
    {
      case cpir_cmd_rintest:
        process_rintest_cmd(cn);
        break;

      case cpir_cmd_modbus:
        handle_modbus(&cn->comms_ctrl, cn->incoming_mesg, cn->incoming_mesg_length, db_app_user.modbus_endian);
        comms_net_message_handled(cn);
        break;

      case cpir_cmd_hbm_rincmd:
        if (!process_rincmd(cn, false))
          error = process_hbm_cmd(cn);
        break;

      case cpir_cmd_rincmd_crc:
        (void) process_rincmd(cn, true);
        break;

      case cpir_failure:
      case cpir_cmd_none:
      default:
        quit = true;
        break;
    }
  }

  /* Send any assembled data out the serial port as a single write */
  if (comms_send_buffer_get_buffer(cn->comms_ctrl.csb, &buffer, &buflength) && buffer != NULL && buflength > 0)
    (void) rwrite(cn->comms_ctrl.serdev, buffer, buflength, OS_TICKS_PER_SEC);

  return error;
}


/** Initialise a s_comms_control structure */
static bool comms_init_comms_control(struct s_comms_control *comms_ctrl, struct t_io_dev *serdev, struct s_comms_print_buffer *cpb)
{
  bool sem_inited = false;

  if (NULL == comms_ctrl || NULL == serdev)
    return false;

  memset(comms_ctrl, 0, sizeof(struct s_comms_control));

  comms_ctrl->serdev = serdev;
  comms_ctrl->cpb = cpb;
  if (!comms_send_buffer_create(&comms_ctrl->csb, 10) || NULL == comms_ctrl->csb)
    goto cicc_fail;

  comms_ctrl->reply_active = false;
  comms_ctrl->unit_selected = false;
  comms_ctrl->kr_cmd_streaming = false;
  comms_ctrl->msv_cmd_streaming = false;
  comms_ctrl->msv_source = MSV_SRC_DISPLAYED;
  comms_ctrl->msv_count = 0;
  comms_ctrl->cmd_query = false;
  comms_ctrl->passcode_ok = false;
  comms_ctrl->send_extra_line_termination = false;
  comms_ctrl->last_rdg_id = 0;
  comms_ctrl->rincomms_last_10hz_time = 0;
  comms_ctrl->comms_params.p_set = 0;
  comms_ctrl->comms_params.ps_set = 0;
  if (sem_init(&comms_ctrl->print_finished_sem, 0, 0) != 0)
    goto cicc_fail;
  sem_inited = true;
  comms_ctrl->regTable = rc_initLocalRegTable(rinCmdTable);
  if (NULL == comms_ctrl->regTable)
    goto cicc_fail;
  rc_gen_local_basic(comms_ctrl->regTable);
  rincmd_initPasscode(comms_ctrl->regTable);

  return true;

cicc_fail:
  if (NULL != comms_ctrl->regTable)
  {
    rc_freeLocalRegTable(comms_ctrl->regTable);
    comms_ctrl->regTable = NULL;
  }
  if (sem_inited)
    sem_destroy(&comms_ctrl->print_finished_sem);
  if (NULL != comms_ctrl->csb)
  {
    comms_send_buffer_destroy(comms_ctrl->csb);
    comms_ctrl->csb = NULL;
  }
  return false;
}


/** Destroy a s_comms_control structure */
static void comms_destroy_comms_control(struct s_comms_control *comms_ctrl)
{
  if (NULL != comms_ctrl)
  {
    comms_ctrl->serdev = NULL;
    comms_ctrl->cpb = NULL;
    comms_send_buffer_destroy(comms_ctrl->csb);
    comms_ctrl->csb = NULL;
    sem_destroy(&comms_ctrl->print_finished_sem);
    rc_freeLocalRegTable(comms_ctrl->regTable);
    comms_ctrl->regTable = NULL;
  }
}


/** Re-initialise parts of cn from the database settings */
void comms_net_reinit(struct s_comms_net *cn)
{
  /* Nothing to be re-initalised from database */
}


/** Create and initialise a new s_comms_net structure. udp_info in host endian. */
struct s_comms_net *comms_net_create(struct t_io_dev *serdev, struct s_comms_print_buffer *cpb, bool is_udp, const struct s_comms_udp_info *udp_info)
{
  struct s_comms_net *cn;

  /* Allocate new comms_net */
  cn = calloc(1, sizeof(struct s_comms_net));
  if (NULL != cn)
  {
    /* Initialise it */
    memset(cn->incoming_mesg, 0, sizeof(cn->incoming_mesg));
    cn->incoming_mesg_length = 0;
    cn->last_examined_mesg_posn = 0;
    cn->start_of_mesg = 0;
    comms_init_comms_control(&cn->comms_ctrl, serdev, cpb);
    clock_gettime(CLOCK_MONOTONIC, &cn->ts_last_rx_data);
    cn->is_udp = is_udp;
    cn->udp_info = *udp_info;
  }
  return cn;
}


/** Clean up and destroy an existing s_comms_net structure */
void comms_net_destroy(struct s_comms_net *cn)
{
  if (NULL != cn)
  {
    comms_destroy_comms_control(&cn->comms_ctrl);
    free(cn);
  }
}


/** Does the work of listening and replying to network commands */
enum eCommsError comms_net_handler(struct s_comms_net *cn)
{
  /* Check pointer */
  if (NULL == cn)
  {
    OSTimeDly(OS_TICKS_PER_SEC);
    return MESS_ERROR;
  }

  return handle_comms(cn);
}
