/*
 * comms_params.c
 *
 *  Created on: 13/10/2010
 *      Author: ian
 */
#include <setjmp.h>
#include <ctype.h>
#include <string.h>

#include "comms_params.h"


/**
 * Break string up into integer and string parameters. It modifies the string by replacing '"'s with '\0'.
 * longjmp()'s back to cp->parse_error if parsing error occurs
 *
 * @param cp is the comms parameters structure that the parameters are stored into
 * @param s is the string to be broken up into parameters. This gets modified, and must not be destroyed until
 *        cp is no longer required (cp points to elements in this string)
 */
void get_params(struct s_comms_params *cp, char *s)
{
  long tmp;
  size_t cnt;
  bool found, minus;
  t_comms_param_mask mask;

  mask = 0x0001;
  cp->p_set = 0;
  cp->ps_set = 0;
  cnt = 0;
  while ('\0' != *s)
  {
    tmp = 0;
    found = false;
    while (isspace (*s))
      s++;

    /* Check for string parameter */
    if (*s == '"')
    {
      s++;
      if (cnt < COMMS_PARAMS_MAX)
        cp->param_str[cnt++] = s;
      while (*s != '"')
      {
        if (*s == '\0')
          longjmp(cp->parse_error, 1);
        s++;
      }

      /* Replace '"' with null to end parameter string */
      *s++ = '\0';
      cp->ps_set |= mask;
    }
    else
    {
      /* Parse the numeric params */
      if (*s == '-')
      {
        minus = true;
        s++;
      }
      else
      {
        minus = false;
      }

      while (isspace (*s))
        s++;
      while (isdigit (*s))
      {
        found = true;
        tmp = tmp * 10 + *s - '0';
        s++;
      }
      if (minus)
        tmp = -tmp;
      while (isspace (*s))
        s++;
      if (found)
        cp->p_set |= mask;
      if (cnt < COMMS_PARAMS_MAX)
      {
        cp->params[cnt++] = tmp;
      }
    } /* if string or number */

    /* Check for end of string or next parameter */
    mask <<= 1;
    if (*s == '\0')
      break;
    if (*s == ',')
      s++;
    else
      longjmp(cp->parse_error, 1);
  } /* while */
}


/** Check string parameter */
bool param_str_ok(struct s_comms_params *cp, size_t p, size_t max_len, size_t min_len)
{
  size_t len;

  if (p >= COMMS_PARAMS_MAX)
    return false;

  /* Check if parameter was passed */
  if ((cp->ps_set & (0x0001 << p)) == 0)
  {
    if ((cp->p_set & (0x0001 << p)) != 0)
    {
      /* It's an integer, not a string, error */
      longjmp(cp->parse_error, 1);
    }

    /* Parameter not set */
    return false;
  }

  /* Check lengths */
  len = strlen(cp->param_str[p]);
  if (len > max_len || len < min_len)
  {
    longjmp(cp->parse_error, 1);
  }
  return true;
}


/** Check string parameter ensuring all characters fall within the character set */
bool param_str_charset_ok(struct s_comms_params *cp, size_t p, size_t max_len, size_t min_len, const char *character_set)
{
  size_t i;
  const char *posn;

  if (!param_str_ok(cp, p, max_len, min_len))
    return false;

  /* Ensure each character is within the character set */
  for (i = 0; '\0' != cp->param_str[p][i]; i++)
  {
    posn = strchr(character_set, cp->param_str[p][i]);
    if (NULL == posn)
      longjmp(cp->parse_error, 1);
  }
  return true;
}


/** Check integer parameter */
bool param_ok(struct s_comms_params *cp, size_t p, long maxl, long minl)
{
  if (p >= COMMS_PARAMS_MAX)
    return false;

  /* Check if parameter was passed */
  if ((cp->p_set & (0x0001 << p)) == 0)
  {
    if ((cp->ps_set & (0x0001 << p)) != 0)
    {
      /* It's a string, not an integer, error */
      longjmp(cp->parse_error, 1);
    }

    /* Parameter not set */
    return false;
  }

  /* Check limits */
  if (cp->params[p] < minl || cp->params[p] > maxl)
  {
    longjmp(cp->parse_error, 1);
  }
  return true;
}
