#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <pthread.h>
#include <rinhw/timemath.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include "system.h"
#include "meas.h"
#include "database.h"
#include "iface.h"
#include "kbd_5000.h"
#include "editor.h"
#include "asciidef.h"
#include "backgrnd.h"
#include "setpoint.h"
#include "comms_util.h"
#include "iface_serial.h"
#include "comms_task.h"
#include "dbprint.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "app_db_runtime_vars.h"
#include "app_db_runtime.h"
#include "comms_escape.h"
#include "iface_clock.h"

#include "comms_print.h"


static const char CRLF[] = "\r\n";
static const char UNDO_STR[] = "Last Entry Cancelled";


/** Returns the last print weight & ID, and the time/date when it occurred */
void print_get_last_print_info(long *rdg, t_datetime *time, int32u *printid)
{
  struct s_comms_print *cp;

  if (NULL == rdg || NULL == time || NULL == printid)
    return;

  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    *printid = cp->last_print_id;
    *rdg = cp->last_print_weight;
    *time = cp->last_print_time;
    comms_print_release_cp(cp);
  }
  else
  {
    *printid = 0;
    *rdg = 0;
    memset(time, 0, sizeof(t_datetime));
  }
}


/** Set the printer text?? */
void print_set_prn_txt(uint8_t prn_txt)
{
  struct s_comms_print *cp;

  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    cp->prn_txt = prn_txt;
    comms_print_release_cp(cp);
  }
}


/** Set which weight reading is to be stored in last_print_weight */
void set_last_print_type(enum eLastPrintWeightType typ)
{
  struct s_comms_print *cp;

  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    cp->last_print_type = typ;
    comms_print_release_cp(cp);
  }
}


/** Store the last printed weight type, based on last_print_type setting */
static void store_last_print(struct s_comms_print *cp)
{
  AdcRdg *lpw_rdg;

  if (cp->traceable_weight.valid)
  {
    cp->last_print_id = cp->traceable_weight.printId;
    cp->last_print_time = cp->traceable_weight.dt;
  }
  else
  {
    cp->last_print_id = 0;
    cp->last_print_time = cp->datetime;
  }

  switch (cp->last_print_type)
  {
    case lpwtGross:
      lpw_rdg = &cp->weights[wt_gross];
      break;
    case lpwtNet:
      lpw_rdg = &cp->weights[wt_net];
      break;
    case lpwtTare:
      lpw_rdg = &cp->weights[wt_tare];
      break;
    case lpwtDisplayed:
    default:
      lpw_rdg = &cp->weights[wt_displayed];
      break;
  }
  cp->last_print_weight = unitsw_convert_final(lpw_rdg->rdg, cp->displayed_units, ADC_UNIT_SWITCH_PRIMARY, lpw_rdg->info.status.bit.hiRes, lpw_rdg->info.status.bit.hiRes);
}


/** Send a character to the printer */
static void print_send_ch(struct s_comms_print *cp, char ch)
{
  /* Write the character */
  (void) comms_send_buffer_addch(cp->csb, ch);

  /* Keep a copy in the global print buffer */
  (void) comms_print_buffer_add_char(cp->cpb, ch);

  /* Keep track of horizontal position */
  cp->print_pos++;
  if (CR == ch || LF == ch)
  {
    cp->print_pos = 0;
  }
}


/** Send a string to the printer */
static void print_send_str(struct s_comms_print *cp, const char *s)
{
  while (*s)
  {
    print_send_ch(cp, *s++);
  }
}


/* Send an indented string to the printer */
static void print_send_tab_str(struct s_comms_print *cp, const char *s)
{
  uint8_t i;

  if (cp->print_pos > 0)
    return;

  /* Indent */
  for (i = 0; i < cp->ticket_columns; i++)
    print_send_ch(cp, ' ');

  /* Send string */
  print_send_str(cp, s);
}


/** Send the system weigh units to the printer */
static void print_send_units(struct s_comms_print *cp, const AdcRdg *rdg)
{
  print_send_ch(cp, ' ');
  print_send_str(cp, build_units_print[xlat_adc_library_units_to_we2111(rdg->info.unitdp.bit.units)]);
}


/** Send gross or net identifier to the printer */
static void print_send_gross_net(struct s_comms_print *cp, bool gross)
{
  print_send_str(cp, (gross) ? " G" : " N");
}


/** Get the status character, based on the latest reading status */
static char print_get_rdg_status_char(const struct s_comms_print *cp, const AdcRdg *rdg)
{
  char status;

  if (APP_ERR_NONE != cp->sys_errors)
  {
    status = 'E';
  }
  else if (rdg->info.status.bit.oload)
  {
    status = 'O';
  }
  else if (rdg->info.status.bit.uload)
  {
    status = 'U';
  }
  else
  {
    status = (rdg->info.status.bit.gross) ? 'G' : 'N';
  }
  return status;
}


/** Print the weight */
static void print_weight_new(struct s_comms_print *cp, const AdcRdg *rdg)
{
  #define WEIGHT_STR_LEN 20
  char weight_str[WEIGHT_STR_LEN];

  (void) comms_convert_num_with_dp_pad(weight_str, WEIGHT_STR_LEN, rdg->rdg, rdg->info.unitdp.bit.dp, 8, true, cp->resolution_length);
  print_send_str(cp, weight_str);
}


/**
 * Print weight and units in primary or alternate units depending on display mode
 *
 * @param cp is the comms print structure
 * @param rdg is the reading to print
 */
static void print_weight_units_disp_mode(struct s_comms_print *cp, const AdcRdg *rdg)
{
  print_weight_new(cp, rdg);
  print_send_units(cp, rdg);
}


/** Send the time to the printer */
static void print_send_time(struct s_comms_print *cp)
{
  #define TIME_STR_LEN 12
  char time_str[TIME_STR_LEN];

  to_string_time(time_str, TIME_STR_LEN, &cp->datetime);
  print_send_str(cp, time_str);
}


/** Send the date to the printer */
static void print_send_date(struct s_comms_print *cp)
{
  #define DATE_STR_LEN 11
  char date_str[DATE_STR_LEN];

  to_string_date(date_str, DATE_STR_LEN, &cp->datetime);
  print_send_str(cp, date_str);
}


/** Send the date and time to the printer */
static void print_send_date_time(struct s_comms_print *cp, bool indented)
{
  if (indented)
  {
    print_send_tab_str(cp, "");
  }
  print_send_date(cp);
  print_send_ch(cp, ' ');
  print_send_time(cp);
}


/** Send line feeds to the printer */
static void print_send_vertical_space(struct s_comms_print *cp)
{
  size_t i;

  for (i = 0; i < cp->ticket_rows; i++)
  {
    print_send_ch(cp, LF);
  }
}


/** Send the printout number */
static void print_send_printout_no(struct s_comms_print *cp)
{
  char str[20];

  snprintf(str, 20, "%06lu", (unsigned long) cp->traceable_weight.printId);
  print_send_str(cp, str);
}


static bool header_valid(size_t line)
{
  size_t i;

  for (i = 0; i < APP_DB_PRINT_HDR_LEN - 1; i++)
  {
    if (db_app_user.print_text[line][i] != ' ')
      return true;
  }
  return false;
}


static void print_header(struct s_comms_print *cp, size_t line)
{
  char temp_str[APP_DB_PRINT_HDR_LEN];
  size_t i;
  size_t end;

  /* Work out last non space character in string */
  for (i = 0, end = 0; i < APP_DB_PRINT_HDR_LEN - 1; i++)
  {
    if (db_app_user.print_text[line][i] != ' ')
      end = i;
  }

  /* Copy data up until last non space character into buffer */
  for (i = 0; i <= end; i++)
  {
    temp_str[i] = db_app_user.print_text[line][i];
  }
  temp_str[i] = '\0';
  print_send_str(cp, temp_str);
}


static void print_txt_info_str(struct s_comms_print *cp)
{
  char temp_str[APP_DB_PRINT_TXT_LEN];

  if (cp->prn_txt)
  {
    text_get_text(temp_str, APP_DB_PRINT_TXT_LEN, cp->prn_txt - 1);
    print_send_str(cp, temp_str);
  }
}


static bool print_use_token_fmt(void)
{
  const char *p;
  bool ret;

  ret = false;
  p = db_app_user.print_fmt;
  while (*p)
  {
    if (*p != ' ')
      ret = true;
    p++;
  }
  return ret;
}


/** Make an AdcRdg for a statistics value */
static void print_make_statistics_adcrdg(AdcRdg *rdg, int32_t value, bool gross, bool hires, enum eAdcUnitSwitch usw_from, enum eAdcUnitSwitch usw_to)
{
  AdcRdg rdgin;

  memset(&rdgin, 0, sizeof(AdcRdg));
  rdgin.rdg = value;
  rdgin.info.status.bit.gross = gross;
  rdgin.info.status.bit.grossOrNet = 1;
  rdgin.info.status.bit.hiRes = hires;
  unitsw_convert_final_adcrdg(rdg, &rdgin, usw_from, usw_to);
}


/** Add to the statistics in primary units, regardless of what units we are printing in */
static void print_statistics_mem_plus_conv_units(struct s_comms_print *cp, const AdcRdg *rdg)
{
  long rdg_primary;

  rdg_primary = unitsw_convert_final(rdg->rdg, cp->displayed_units, ADC_UNIT_SWITCH_PRIMARY, rdg->info.status.bit.hiRes, rdg->info.status.bit.hiRes);
  statistics_mem_plus(rdg_primary, rdg->info.status.bit.gross);
}


/** Print the statistics totals */
static void print_statistics_totals(struct s_comms_print *cp)
{
  AdcRdg statrdg;

  if (ADC_APPROVAL_MODE_NTEP == cp->approval_mode)
  {
    print_make_statistics_adcrdg(&statrdg, statistics_get_gross_total(), true, cp->weights[wt_displayed].info.status.bit.hiRes, ADC_UNIT_SWITCH_PRIMARY, cp->displayed_units);
    print_weight_units_disp_mode(cp, &statrdg);
    print_send_gross_net(cp, true);
    print_send_ch(cp, ' ');
    print_make_statistics_adcrdg(&statrdg, statistics_get_net_total(), false, cp->weights[wt_displayed].info.status.bit.hiRes, ADC_UNIT_SWITCH_PRIMARY, cp->displayed_units);
    print_weight_units_disp_mode(cp, &statrdg);
    print_send_gross_net(cp, false);
  }
  else
  {
    print_make_statistics_adcrdg(&statrdg, statistics_get_total(), true, cp->weights[wt_displayed].info.status.bit.hiRes, ADC_UNIT_SWITCH_PRIMARY, cp->displayed_units);
    print_weight_units_disp_mode(cp, &statrdg);
  }
}


/** Print a token string to the printer */
static void print_token_string(struct s_comms_print *cp, const char *s)
{
  size_t i;

  /* Keep record of this printout */
  store_last_print(cp);

  /* Parse format string */
  while (*s)
  {
    switch ((uint8_t) *s)
    {
      /* \. -> '\0' */
      case COMMS_ESCAPED_DOT_UNESCAPED:
        print_send_ch(cp, 0);
        break;

      /* \+ -> Same as \W and add to total */
      case COMMS_ESCAPED_PLUS_UNESCAPED:
        print_statistics_mem_plus_conv_units(cp, &cp->weights[wt_displayed]);
        print_weight_units_disp_mode(cp, &cp->weights[wt_displayed]);
        print_send_gross_net(cp, cp->weights[wt_displayed].info.status.bit.gross);
        break;

      /* \- -> Undo last print */
      case COMMS_ESCAPED_MINUS_UNESCAPED:
        if (statistics_undo_mem_plus())
        {
          print_send_str(cp, UNDO_STR);
        }
        break;

      /* \T -> Tare weight */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('T'):
        print_weight_units_disp_mode(cp, &cp->weights[wt_tare]);
        print_send_str(cp, (cp->weights[wt_tare].info.status.bit.preTare) ? " PT" : " T");
        break;

      /* \G -> Gross weight */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('G'):
        print_weight_units_disp_mode(cp, &cp->weights[wt_gross]);
        print_send_gross_net(cp, true);
        break;

      /* \N -> Net weight */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('N'):
        print_weight_units_disp_mode(cp, &cp->weights[wt_net]);
        print_send_gross_net(cp, false);
        break;

      /* \W -> Displayed weight */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('W'):
        print_weight_units_disp_mode(cp, &cp->weights[wt_displayed]);
        print_send_gross_net(cp, cp->weights[wt_displayed].info.status.bit.gross);
        break;

      /* \I -> Print ID number */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('I'):
        print_send_printout_no(cp);
        break;

      /* \X -> Print thumbwheel IO selected text line */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('X'):
        print_txt_info_str(cp);
        break;

      /* \H -> Print all header lines */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('H'):
        for (i = 0; i < APP_DB_PRINT_HDR_LINES; i++)
        {
          if (header_valid(i))
          {
            print_send_tab_str(cp, "");
            print_header(cp, i);
            print_send_str(cp, CRLF);
          }
        }
        break;

      /* \A -> Print first half of header lines */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('A'):
        for (i = 0; i < APP_DB_PRINT_HDR_LINES / 2; i++)
        {
          if (header_valid(i))
          {
            print_send_tab_str(cp, "");
            print_header(cp, i);
            print_send_str(cp, CRLF);
          }
        }
        break;

      /* \B -> Print second half of header lines */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('B'):
        for (i = APP_DB_PRINT_HDR_LINES / 2; i < APP_DB_PRINT_HDR_LINES; i++)
        {
          if (header_valid(i))
          {
            print_send_tab_str(cp, "");
            print_header(cp, i);
            print_send_str(cp, CRLF);
          }
        }
        break;

      /* \D -> Date and time */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('D'):
        print_send_date_time(cp, false);
        break;

      /* \U -> Units */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('U'):
        print_send_units(cp, &cp->weights[wt_displayed]);
        break;

      /* \E -> CRLF */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('E'):
        print_send_str(cp, CRLF);
        break;

      /* \O -> Print total weight */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('O'):
        print_statistics_totals(cp);
        print_send_str(cp, "Tot");
        statistics_clear();
        break;

      /* \F -> CRLF & indent next line */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('F'):
        print_send_str(cp, CRLF);
        /* Fall through */

      /* \C -> Indent next line */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('C'):
        print_send_tab_str(cp, "");
        break;

      /* \R -> Print blank rows at top of docket */
      case COMMS_ESCAPED_CHAR_TO_UNESCAPED('R'):
        print_send_vertical_space(cp);
        break;

      /* Normal non-formatting character */
      default:
        print_send_ch(cp, (char) *s);
        break;
    }
    s++;
  }
}


/**
 * Send a ticket printout to the printer of the form:
 *
 *             WEIGHT
 *             TICKET
 * 08/10/2031 21:14:40
 * ID: 000021
 * T:            750  g
 * G:           3000  g
 * N:           2250  g
 *
 */
static void print_send_ticket(struct s_comms_print *cp)
{
  size_t i;

  /* Keep record of this printout */
  store_last_print(cp);

  /* Send gap between printout */
  print_send_vertical_space(cp);

  /* Send header */
  for (i = 0; i < APP_DB_PRINT_HDR_LINES; i++)
  {
    if (header_valid(i))
    {
      print_send_tab_str(cp, "");
      print_header(cp, i);
      print_send_str(cp, CRLF);
    }
  }

  /* Send date and time */
  print_send_date_time(cp, true);
  print_send_str(cp, CRLF);

  /* Send print number on new line */
  print_send_tab_str(cp, "ID: ");
  print_send_printout_no(cp);
  print_send_str(cp, CRLF);

  /* Send preset tare/tare weight */
  print_send_tab_str(cp, (cp->weights[wt_tare].info.status.bit.preTare) ? "PT:      " : "T:       ");
  print_weight_units_disp_mode(cp, &cp->weights[wt_tare]);
  print_send_str(cp, CRLF);

  /* Send gross weight */
  print_send_tab_str(cp, "G:       ");
  print_weight_units_disp_mode(cp, &cp->weights[wt_gross]);
  print_send_str(cp, CRLF);

  /* Send net weight */
  print_send_tab_str(cp, "N:       ");
  print_weight_units_disp_mode(cp, &cp->weights[wt_net]);
  print_send_str(cp, CRLF);
}


/**
 * Creates a printout of the form:
 *    000015 08/10/2031 21:13:09     2250  g N
 * ie ID     Date       Time         Weight Units Gross/Net/Tare
 */
static void print_send_line(struct s_comms_print *cp, bool double_space, bool send_date)
{
  /* Keep record of this printout */
  store_last_print(cp);

  /* Send gap between printouts */
  print_send_vertical_space(cp);

  /* Send indent & printout number */
  print_send_tab_str(cp, "");
  print_send_printout_no(cp);

  /* Send date (if requested) */
  print_send_ch(cp, ' ');
  if (send_date)
  {
    print_send_date_time(cp, false);
    print_send_ch(cp, ' ');
  }

  /* Send double line printout if requested */
  if (double_space)
  {
    print_send_str(cp, CRLF);
    print_send_tab_str(cp, "             ");
  }

  /* Send weight, units and status */
  print_weight_units_disp_mode(cp, &cp->weights[wt_displayed]);
  print_send_ch(cp, ' ');
  print_send_ch(cp, print_get_rdg_status_char(cp, &cp->weights[wt_displayed]));
  print_send_str(cp, CRLF);
}


static void do_printout(struct s_comms_print *cp)
{

  switch (db_app_user.printer_mode)
  {
    /* Single line printout */
    case P_SINGLE:
    case PA_SINGLE:
      print_send_line(cp, false, true);
      break;

    /* Double line printout */
    case P_DOUBLE:
    case PA_DOUBLE:
      print_send_line(cp, true, true);
      break;

    /* Ticket printout */
    case P_TIC:
    case PA_TIC:
      if (print_use_token_fmt())
      {
        print_token_string(cp, db_app_user.print_fmt);
      }
      else
      {
        print_send_ticket(cp);
      }
      break;

    /* Total printout */
    case P_TOTAL:
    case PA_TOTAL:
      print_send_line(cp, false, true);
      print_statistics_mem_plus_conv_units(cp, &cp->weights[wt_displayed]);
      initiate_display_total(&display_total_struct);
      break;
  }
}


/** Print the system total weight */
static void print_total(struct s_comms_print *cp)
{
  char temp_str[20];

  if (statistics_get_num_items() == 0)
    return;

  /* Send gap between printouts */
  print_send_vertical_space(cp);

  /* Print the number of items */
  print_send_tab_str(cp, "ITEMS: ");
  snprintf(temp_str, 20, "%5u", statistics_get_num_items());
  print_send_str(cp, temp_str);

  /* Print the total weight and units */
  print_send_str(cp, "   TOTAL: ");
  print_statistics_totals(cp);
  print_send_str(cp, CRLF);
}


static void print_M_minus(struct s_comms_print *cp)
{
  print_send_vertical_space(cp);
  print_send_tab_str(cp, UNDO_STR);
  print_send_str(cp, CRLF);
}


/** Printout when a batch is aborted */
static void print_abort(struct s_comms_print *cp)
{
  char temp_str[30];

  print_send_vertical_space(cp);
  print_send_tab_str(cp, "Batch ");

  snprintf(temp_str, 30, "%05u", statistics_get_num_items());
  print_send_str(cp, temp_str);

  print_send_str(cp, " Aborted.");
  print_send_str(cp, CRLF);
}


static void print_fill_out_cp(struct s_comms_print *cp)
{
  AdcTraceableWeight *atw;
  AdcRdg disp_rdg;

  /*
   * These are already filled out
   * cp->mutex
   * cp->serdev
   * cp->csb
   * cp->trigger_sem
   * cp->potype
   * cp->data
   * cp->print_pos
   * cp->prn_txt
   * cp->printer_state
   * cp->traceable_weight (if traceable print)
   */

  /* Collect weight data in 1 go */
  cp->ticket_rows = db_app_user.tic_row;
  cp->ticket_columns = db_app_user.tic_col;

  cp->displayed_units = unitsw_displaymode_to_unitswitch(adcSuperStructure.adcDataWorking.dispMode);
  cp->approval_mode = adcSuperStructure.adcUserDatabase.approvalMode;
  cp->resolution_length = (ADC_APPROVAL_MODE_NTEP == cp->approval_mode) ? unitsw_get_resolution_length(cp->displayed_units) : 0;

  /* If we are printing with a traceable weight, use it */
  if (cp->traceable_weight.valid)
  {
    /* Use traceable weight */
    atw = &cp->traceable_weight;
    cp->datetime = atw->dt;
    cp->sys_errors = get_current_errors();

    /* Get the appropriate weight from the traceable weight */
    switch (cp->displayed_units)
    {
      default:
      case ADC_UNIT_SWITCH_PRIMARY:
        cp->weights[wt_displayed] = atw->rdg;
        break;
      case ADC_UNIT_SWITCH_SECONDARY:
        cp->weights[wt_displayed] = atw->altRdg;
        break;
      case ADC_UNIT_SWITCH_TERTIARY:
        cp->weights[wt_displayed] = atw->tertRdg;
        break;
    }

    /* Now fill out the rest */
    cp->weights[wt_tare] = cp->weights[wt_displayed];
    cp->weights[wt_tare].rdg = unitsw_convert_final(atw->tareWgt, ADC_UNIT_SWITCH_PRIMARY, cp->displayed_units, cp->weights[wt_displayed].info.status.bit.hiRes, cp->weights[wt_displayed].info.status.bit.hiRes);
    cp->weights[wt_tare].info.status.bit.gross = 0;
    cp->weights[wt_tare].info.status.bit.grossOrNet = 0;
    cp->weights[wt_tare].info.status.bit.tare = !atw->pTare;
    cp->weights[wt_tare].info.status.bit.preTare = atw->pTare;

    if (cp->weights[wt_displayed].info.status.bit.gross)
    {
      /* Gross weight */
      cp->weights[wt_gross] = cp->weights[wt_displayed];

      cp->weights[wt_net] = cp->weights[wt_displayed];
      cp->weights[wt_net].rdg = cp->weights[wt_displayed].rdg - cp->weights[wt_tare].rdg;
      cp->weights[wt_net].info.status.bit.gross = 0;
      cp->weights[wt_net].info.status.bit.grossOrNet = 1;
      cp->weights[wt_net].info.status.bit.tare = 0;
      cp->weights[wt_net].info.status.bit.preTare = 0;
    }
    else
    {
      /* Net weight */
      cp->weights[wt_net] = cp->weights[wt_displayed];

      cp->weights[wt_gross] = cp->weights[wt_displayed];
      cp->weights[wt_gross].rdg = cp->weights[wt_displayed].rdg + cp->weights[wt_tare].rdg;
      cp->weights[wt_gross].info.status.bit.gross = 1;
      cp->weights[wt_gross].info.status.bit.grossOrNet = 1;
      cp->weights[wt_gross].info.status.bit.tare = 0;
      cp->weights[wt_gross].info.status.bit.preTare = 0;
    }
  }
  else
  {
    /* Else get ADC current weight */
    GetDateTime(&cp->datetime);
    AdcMutexLock(&adcSuperStructure);
    cp->sys_errors = get_current_errors();

    /* Get weights unit switched */
    cp->displayed_units = unitsw_displaymode_to_unitswitch(adcSuperStructure.adcDataWorking.dispMode);
    unitsw_get_adcrdg(&cp->weights[wt_displayed], cp->displayed_units, wt_displayed);
    /* Use get_displayed_adcrdg() and convert units because displayed weight might be held/etc */
    get_displayed_adcrdg(&disp_rdg);
    unitsw_convert_final_adcrdg(&cp->weights[wt_displayed], &disp_rdg, ADC_UNIT_SWITCH_PRIMARY, cp->displayed_units);
    unitsw_get_adcrdg(&cp->weights[wt_gross], cp->displayed_units, wt_gross);
    unitsw_get_adcrdg(&cp->weights[wt_net], cp->displayed_units, wt_net);
    unitsw_get_adcrdg(&cp->weights[wt_tare], cp->displayed_units, wt_tare);

    AdcMutexUnlock(&adcSuperStructure);
  }
}


/** Set the time that the printout occurred, so we can time out later if data is not sent */
static void comms_print_set_print_time(struct s_comms_print *cp)
{
  cp->printout_time = time(NULL);
}


/** Work out if a printout is complete (ie all data sent) */
static bool comms_print_internal_is_complete(struct s_comms_print *cp)
{
  bool complete;
  unsigned short level;

  complete = true;
  if (NULL != cp)
  {
    level = 0;
    if (rioctl(cp->serdev, IOCTL_SER_GET_TX_BUF_LEVEL, &level) < 0)
    {
      complete = true;
    }
    else
    {
      complete = (0 == level);
    }
  }
  return complete;
}


/** Work out if a printout is complete (ie all data sent) */
bool comms_print_is_complete(void)
{
  struct s_comms_print *cp;
  bool complete;

  complete = true;
  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    complete = comms_print_internal_is_complete(cp);
    comms_print_release_cp(cp);
  }
  return complete;
}


/** Work out if a printout has timed out */
bool comms_print_has_timedout(void)
{
  struct s_comms_print *cp;
  bool timedout;

  timedout = false;
  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    /* Check printing complete */
    if (!comms_print_internal_is_complete(cp))
    {
      /* No, data still in buffer, has timeout occurred? */
      timedout = (time(NULL) > cp->printout_time + COMMS_PRINT_TIMEOUT);
    }
    comms_print_release_cp(cp);
  }
  return timedout;
}


/** Work out if the CTS pin is blocking sending (returns true if CTS is not blocking) */
bool comms_print_clear_to_send(void)
{
  struct s_comms_print *cp;
  bool clear_to_send;
  unsigned char cts;

  clear_to_send = true;
  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    cts = 1;
    rioctl(cp->serdev, IOCTL_SER_GET_CTS_STATE, &cts);
    clear_to_send = (0 != cts);
    comms_print_release_cp(cp);
  }
  return clear_to_send;
}


/** Flush all data from the transmit buffer */
void comms_print_flush_transmit(void)
{
  struct s_comms_print *cp;

  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    rioctl(cp->serdev, IOCTL_SER_FLUSH_TX_BUF);
    comms_print_release_cp(cp);
  }
}


/** Does the work of sending a printout. Waits until signalled. */
void comms_print_handler(struct s_comms_print *cp)
{
  char *printstr;
  char *rawprintstr;
  const void *buffer;
  size_t buflength;

  /* Check pointer */
  if (NULL == cp)
  {
    goto fail_delay;
  }

  /* Wait until signaled or timeout */
  if (sem_timedwait_ticks(&cp->trigger_sem, OS_TICKS_PER_SEC) < 0)
  {
    goto fail_delay;
  }

  /* Special handling for incoming UDP connections */
  if (cp->is_incoming_udp)
    comms_udp_read_for_txonly(cp->serdev);

  /* Get exclusive access */
  pthread_mutex_lock(&cp->mutex);

  /* Clear the comms send buffer */
  comms_send_buffer_clear(cp->csb);

  /* Get snapshot of data for printout */
  print_fill_out_cp(cp);

  /* Generate requested printout */
  switch (cp->potype)
  {
    case ptStandard:
    case ptStandardComms:
      do_printout(cp);
      break;

    case ptTotal:
      print_total(cp);
      break;

    case ptMemMinus:
      print_M_minus(cp);
      break;

    case ptAbort:
      print_abort(cp);
      break;

    case ptDatabaseConfig:
      comms_print_db(cp, (0 != cp->data));
      break;

    case ptSpecialRemote:
      printstr = (char *) cp->ptr_data;
      if (NULL != printstr)
      {
        print_token_string(cp, printstr);
        free(printstr);
      }
      break;

    case ptRawDataString:
      rawprintstr = (char *) cp->ptr_data;
      if (NULL != rawprintstr)
      {
        print_send_str(cp, rawprintstr);
        free(rawprintstr);
      }
      break;

    case ptNone:
    default:
      break;
  }

  /* Send the assembled data out the serial port as a single write */
  if (comms_send_buffer_get_buffer(cp->csb, &buffer, &buflength) && buffer != NULL && buflength > 0)
    (void) rwrite(cp->serdev, buffer, buflength, 0);

  /* Set printout timeout, so an error can be generated if tx buffer doesn't empty */
  comms_print_set_print_time(cp);

  /* Clear flags now that all done */
  cp->potype = ptNone;
  cp->data = 0;
  cp->ptr_data = NULL;
  cp->traceable_weight.valid = false;

  /* If the comms is waiting for the print to complete, then signal it */
  if (NULL != cp->print_finished_sem)
  {
    sem_post(cp->print_finished_sem);
    cp->print_finished_sem = NULL;
  }

  /* Release exclusive access */
  pthread_mutex_unlock(&cp->mutex);

  return;

fail_delay:
  OSTimeDly(OS_TICKS_PER_SEC);
}


/** Returns true if a printer is installed on the system */
bool comms_print_system_has_printer(void)
{
  struct s_comms_print *cp;
  bool has_printer = false;

  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    has_printer = true;
    comms_print_release_cp(cp);
  }
  return has_printer;
}
