#ifndef COMMS_PRINT_DOT_H
#define COMMS_PRINT_DOT_H 1

#include <stdint.h>
#include <stdbool.h>
#include <semaphore.h>
#include <rinweigh/adcdatatypes.h>
#include "sysclock.h"
#include "apperror.h"
#include "comms_print_buffer.h"
#include "comms_send_buffer.h"
#include "unitswitch.h"


/** Printout timeout in seconds */
#define COMMS_PRINT_TIMEOUT 15


/** The printout types */
enum ePrintoutType { ptNone = 0,                /*!< No printout in progress. */
                     ptStandard,                /*!< Standard printout configured in the system. Traceable. */
                     ptTotal,                   /*!< Total print. Non-traceable. */
                     ptMemMinus,                /*!< Memory minus print. Non-traceable. */
                     ptAbort,                   /*!< Batch abort print. Non-traceable. */
                     ptDatabaseConfig,          /*!< Print out the contents of the database. Non-traceable. */
                     ptSpecialRemote,           /*!< Special remote printout triggered over comms. Traceable. */
                     ptRawDataString,           /*!< Send the raw data straight to the printer. Non-traceable. */
                     ptStandardComms,           /*!< Same as ptStandard, but initiated by the comms. Traceable. */
                   };


/** The type of weight that was printed last (gross/net/etc) */
enum eLastPrintWeightType { lpwtNone = 0,
                            lpwtDisplayed,
                            lpwtGross,
                            lpwtNet,
                            lpwtTare
                          };


/** Printing communications structure */
struct s_comms_print
{
  /* Printing control */
  pthread_mutex_t mutex;                  /*!< Exclusive access mutex */
  struct t_io_dev *serdev;                /*!< Serial port for the printout */
  bool is_incoming_udp;                   /*!< Port is an incoming UDP connection */
  struct s_comms_send_buffer *csb;        /*!< Send buffer to assemble the print out into, so we can do a single write to serdev */
  struct s_comms_print_buffer *cpb;       /*!< Comms print buffer for storing printer output for PRT command */
  sem_t trigger_sem;                      /*!< Semaphore to trigger a printout */
  enum ePrintoutType potype;              /*!< The requested printout type */
  int data;                               /*!< Data to accompany the printout type */
  void *ptr_data;                         /*!< Pointer data to accompany the printout type */
  time_t printout_time;                   /*!< When the last printout occurred, used for timeouts */

  uint8_t print_pos;
  uint8_t prn_txt;

  /* Formatting */
  uint8_t ticket_rows;
  uint8_t ticket_columns;

  /* Weight information */
  enum eAdcUnitSwitch displayed_units;    /*!< What units are currently being displayed (primary/secondary/tertiary) */
  enum eAdcApprovalMode approval_mode;    /*!< Approval mode is required for statistics prints (NTEP has some special cases) */
  AdcTraceableWeight traceable_weight;    /*!< Traceable weight if this is a traceable printout */
  t_datetime datetime;                    /*!< Timestamp for printout */
  we_error_t sys_errors;                  /*!< Any system errors */
  AdcRdg weights[wt_num];                 /*!< Displayed/gross/net/tare weights in currently selected units */
  int resolution_length;                  /*!< In NTEP this affects the width of zero padding in printed weights */

  /* Last print information */
  sem_t *print_finished_sem;              /*!< Semaphore to signal that the printout has completed */
  int32u last_print_id;                   /*!< Print ID of last printout */
  t_datetime last_print_time;             /*!< Time/date of last printout */
  long last_print_weight;                 /*!< Weight of last printout */
  enum eLastPrintWeightType last_print_type;
};


/** Returns the last print weight & ID, and the time/date when it occurred */
void print_get_last_print_info(long *rdg, t_datetime *time, int32u *printid);


/** Set the printer text?? */
void print_set_prn_txt(uint8_t prn_txt);


/** Set which weight reading is to be stored in last_print_weight */
void set_last_print_type(enum eLastPrintWeightType typ);


/** Work out if a printout is complete (ie all data sent) */
bool comms_print_is_complete(void);


/** Work out if a printout has timed out */
bool comms_print_has_timedout(void);


/** Work out if the CTS pin is blocking sending (returns true if CTS is not blocking) */
bool comms_print_clear_to_send(void);


/** Flush all data from the transmit buffer */
void comms_print_flush_transmit(void);


/** Does the work of sending a printout. Waits until signalled. */
void comms_print_handler(struct s_comms_print *cp);


/** Returns true if a printer is installed on the system */
bool comms_print_system_has_printer(void);


#endif /* #ifndef COMMS_PRINT_DOT_H */
