/*
 * comms_print_buffer.c
 *
 * Buffering of all printed text, so that the PRT command can retrieve it
 */
#include <stdlib.h>
#include <string.h>

#include "comms_print_buffer.h"


/** Create a new comms print buffer */
struct s_comms_print_buffer *comms_print_buffer_create(size_t size)
{
  struct s_comms_print_buffer *cpb = NULL;

  /* Create structure */
  cpb = malloc(sizeof(struct s_comms_print_buffer));
  if (NULL == cpb)
  {
    goto cpbc_fail;
  }
  memset(cpb, 0, sizeof(struct s_comms_print_buffer));

  /* Create exclusive access semaphore */
  if (0 != pthread_mutex_init(&cpb->mutex, NULL))
  {
    goto cpbc_fail;
  }

  /* Allocate buffer */
  cpb->buffer = malloc(size);
  if (NULL == cpb->buffer)
  {
    goto cpbc_fail;
  }

  cpb->buf_in = 0;
  cpb->buf_out = 0;
  cpb->buf_size = size;

  return cpb;

cpbc_fail:
  if (NULL != cpb)
  {
    if (NULL != cpb->buffer)
    {
      free(cpb->buffer);
      cpb->buffer = NULL;
    }

    pthread_mutex_destroy(&cpb->mutex);

    free(cpb);
    cpb = NULL;
  }
  return NULL;
}


/** Destroy a comms print buffer */
void comms_print_buffer_destroy(struct s_comms_print_buffer *cpb)
{
  if (NULL != cpb)
  {
    if (NULL != cpb->buffer)
    {
      free(cpb->buffer);
      cpb->buffer = NULL;
    }

    pthread_mutex_destroy(&cpb->mutex);

    free(cpb);
    cpb = NULL;
  }
}


/** Work out number of bytes free in the buffer */
static size_t comms_print_buffer_bytes_free(struct s_comms_print_buffer *cpb)
{
  ssize_t bytes_free;

  bytes_free = cpb->buf_out - cpb->buf_in - 1;
  if (bytes_free < 0)
  {
    return (size_t) (bytes_free + cpb->buf_size);
  }
  return (size_t) bytes_free;
}


/** Add a character to the print buffer if there is room */
bool comms_print_buffer_add_char(struct s_comms_print_buffer *cpb, char ch)
{
  bool result;

  result = false;
  if (NULL != cpb)
  {
    pthread_mutex_lock(&cpb->mutex);

    if (comms_print_buffer_bytes_free(cpb) > 0)
    {
      cpb->buffer[cpb->buf_in] = ch;
      cpb->buf_in++;
      if (cpb->buf_in >= cpb->buf_size)
        cpb->buf_in = 0;
      result = true;
    }

    pthread_mutex_unlock(&cpb->mutex);
  }
  return result;
}


/** Get a character from the print buffer if there are any */
bool comms_print_buffer_get_char(struct s_comms_print_buffer *cpb, char *ch)
{
  bool result;

  result = false;
  if (NULL != cpb && NULL != ch)
  {
    pthread_mutex_lock(&cpb->mutex);

    if (cpb->buf_in != cpb->buf_out)
    {
      *ch = cpb->buffer[cpb->buf_out];
      cpb->buf_out++;
      if (cpb->buf_out >= cpb->buf_size)
        cpb->buf_out = 0;
      result = true;
    }

    pthread_mutex_unlock(&cpb->mutex);
  }
  return result;
}
