/*
 * comms_print_trigger.c
 *
 *  Created on: Mar 16, 2012
 *      Author: ian
 */
#include <string.h>
#include "comms_print.h"
#include "comms_task.h"
#include "database.h"
#include "comms_escape.h"

#include "comms_print_trigger.h"


/** Initiate a printout (on first serial port configured as a printer). Only called internally */
static bool comms_print_trigger(enum ePrintoutType potype, int data, void *ptr_data, const AdcTraceableWeight *atw, sem_t *print_finished_sem)
{
  struct s_comms_print *cp;
  bool result = false;

  cp = comms_print_get_cp(NULL);
  if (NULL != cp)
  {
    if (ptNone == cp->potype)
    {
      /* Can only print if a printout hasn't already been triggered */
      cp->potype = potype;
      cp->data = data;
      cp->ptr_data = ptr_data;
      if (NULL == atw)
      {
        memset(&cp->traceable_weight, 0, sizeof(AdcTraceableWeight));
        cp->traceable_weight.valid = false;
      }
      else
      {
        cp->traceable_weight = *atw;
      }
      cp->print_finished_sem = print_finished_sem;
      sem_post(&cp->trigger_sem);
      result = true;
    }
    comms_print_release_cp(cp);
  }
  return result;
}


/** Determine if a printout is traceable */
bool comms_print_is_traceable(enum ePrintoutType potype)
{
  return (ptStandard == potype || ptSpecialRemote == potype || ptStandardComms == potype);
}


/** Determine if a printout is initiated by the comms, rather than UI */
bool comms_print_is_comms_initiated(enum ePrintoutType potype)
{
  return (ptSpecialRemote == potype || ptStandardComms == potype);
}


/** Initiate a traceable printout (on first serial port configured as a printer) */
static bool comms_print_trigger_traceable(enum ePrintoutType potype, int data, void *ptr_data, const AdcTraceableWeight *atw, sem_t *print_finished_sem)
{
  /* Only some printouts are traceable */
  if (!comms_print_is_traceable(potype))
  {
    return false;
  }

  /* Traceable weight can't be NULL */
  if (NULL == atw)
  {
    return false;
  }

  /* Trigger the printout */
  return comms_print_trigger(potype, data, ptr_data, atw, print_finished_sem);
}


/** Initiate a non-traceable printout (on first serial port configured as a printer) */
static bool comms_print_trigger_nontraceable(enum ePrintoutType potype, int data, void *ptr_data, sem_t *print_finished_sem)
{
  /* Only some printouts are non-traceable */
  if (comms_print_is_traceable(potype))
  {
    return false;
  }

  /* Trigger the printout */
  return comms_print_trigger(potype, data, ptr_data, NULL, print_finished_sem);
}


/** Initiate a standard printout (traceable) */
bool comms_print_trigger_traceable_standard(const AdcTraceableWeight *atw, sem_t *print_finished_sem)
{
  return comms_print_trigger_traceable(ptStandard, 0, NULL, atw, print_finished_sem);
}


/** Initiate a total printout (non-traceable) */
bool comms_print_trigger_nontraceable_total(sem_t *print_finished_sem)
{
  return comms_print_trigger_nontraceable(ptTotal, 0, NULL, print_finished_sem);
}


/** Initiate a memory minus printout (non-traceable) */
bool comms_print_trigger_nontraceable_memory_minus(void)
{
  return comms_print_trigger_nontraceable(ptMemMinus, 0, NULL, NULL);
}


/** Initiate an abort printout (non-traceable) */
bool comms_print_trigger_nontraceable_abort(void)
{
  return comms_print_trigger_nontraceable(ptAbort, 0, NULL, NULL);
}


/** Initiate a database configuration printout (non-traceable) */
bool comms_print_trigger_nontraceable_database_config(bool full)
{
  return comms_print_trigger_nontraceable(ptDatabaseConfig, (int) full, NULL, NULL);
}


/** Initiate a special remote format printout (traceable). s is a format string eg "\g". */
bool comms_print_trigger_traceable_special_remote(const char *s, const AdcTraceableWeight *atw, sem_t *print_finished_sem)
{
  char *printstr;
  size_t len;

  if (NULL == s)
  {
    return false;
  }

  /* Allocate string, freed in printing task */
  printstr = strdup(s);
  if (NULL == printstr)
  {
    return false;
  }

  /* Convert tokens into characters, result string will same length or shorter than escaped string (+1 for null terminator) */
  len = strlen(s) + 1;
  comm_unescape_str(printstr, s, len);

  /* Trigger printout */
  return comms_print_trigger_traceable(ptSpecialRemote, 0, (void *) printstr, atw, print_finished_sem);
}


/** Initiate a raw data printout (non-traceable). s is null terminated raw data to be sent straight to the printer */
bool comms_print_trigger_nontraceable_raw_data(const char *s)
{
  char *printstr;

  if (NULL == s)
  {
    return false;
  }

  /* Allocate string, freed in printing task */
  printstr = strdup(s);
  if (NULL == printstr)
  {
    return false;
  }

  /* Trigger printout */
  return comms_print_trigger_nontraceable(ptRawDataString, 0, printstr, NULL);
}
