/*
 * comms_send_buffer.c
 *
 *  Created on: Feb 28, 2017
 *      Author: ian
 */
#include <stdbool.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "comms_send_buffer.h"


/** Create a comms send buffer */
bool comms_send_buffer_create(struct s_comms_send_buffer **csb, size_t alloc_increments)
{
  struct s_comms_send_buffer *pcsb;

  /* Allocate structure */
  pcsb = calloc(1, sizeof(struct s_comms_send_buffer));
  if (NULL == pcsb)
    return false;

  /* Initialise members */
  pcsb->buffer = NULL;
  pcsb->alloc_increments = alloc_increments;
  pcsb->allocated_size = 0;
  pcsb->used_size = 0;

  /* Return new buffer and result */
  *csb = pcsb;
  return true;
}


/** Destroy a comms send buffer */
bool comms_send_buffer_destroy(struct s_comms_send_buffer *csb)
{
  if (NULL == csb)
    return false;

  csb->allocated_size = 0;
  csb->alloc_increments = 0;
  csb->used_size = 0;
  if (NULL == csb->buffer)
  {
    free(csb->buffer);
    csb->buffer = NULL;
  }
  free(csb);
  csb = NULL;
  return true;
}


/** Clear the data from the comms send buffer (but doesn't free anything) */
bool comms_send_buffer_clear(struct s_comms_send_buffer *csb)
{
  if (NULL == csb)
    return false;
  csb->used_size = 0;
  return true;
}


/** Get the comms send buffer pointer */
bool comms_send_buffer_get_buffer(struct s_comms_send_buffer *csb, const void **buffer, size_t *length)
{
  if (NULL == csb || NULL == buffer || NULL == length || NULL == csb->buffer)
    return false;
  *buffer = csb->buffer;
  *length = csb->used_size;
  return true;
}


/** Calculate a new buffer size (do rounding up to next multiple of resize_increment) */
static bool comms_send_buffer_calc_bufsize(size_t resize_increment, size_t *length)
{
  size_t len;

  /* Prevent divide by 0 */
  if (0 == resize_increment)
  {
    return false;
  }

  /* Work out new buffer size */
  len = *length;
  if (0 == len)
  {
    /* Zero is a special case -> allocate 1 resize_increment */
    *length = resize_increment;
  }
  else
  {
    if (0 == (len % resize_increment))
    {
      /* On the boundary, no round up required */
      *length = len;
    }
    else
    {
      /* Somewhere in between, round up */
      *length = ((len / resize_increment) + 1) * resize_increment;
    }
  }
  return true;
}


/** Resize a comms send buffer */
static bool comms_send_buffer_resize(struct s_comms_send_buffer *csb, size_t extra_length)
{
  size_t newsize;
  bool result;

  if (csb->allocated_size < csb->used_size + extra_length)
  {
    result = false;
    newsize = csb->used_size + extra_length;
    if (comms_send_buffer_calc_bufsize(csb->alloc_increments, &newsize))
    {
      csb->buffer = realloc(csb->buffer, newsize);
      if (csb->buffer != NULL)
      {
        csb->allocated_size = newsize;
        result = true;
      }
    }
  }
  else
  {
    result = true;
  }
  return result;
}


/** Write a buffer to the comms send buffer */
bool comms_send_buffer_addbuf(struct s_comms_send_buffer *csb, const void *buf, size_t length)
{
  bool result;

  if (comms_send_buffer_resize(csb, length))
  {
    memcpy(&csb->buffer[csb->used_size], buf, length);
    csb->used_size += length;
    result = true;
  }
  else
  {
    result = false;
  }
  return result;
}


/** Write a character to the comms send buffer */
bool comms_send_buffer_addch(struct s_comms_send_buffer *csb, char ch)
{
  return comms_send_buffer_addbuf(csb, &ch, sizeof(ch));
}


/** vprintf into a comms send buffer */
int comms_send_buffer_vprintf(struct s_comms_send_buffer *csb, const char *format, va_list args)
{
  va_list argcopy;
  int length;

  /* Work out length */
  va_copy(argcopy, args);
  length = vsnprintf(NULL, 0, format, args);
  if (length >= 0)
  {
    /* Resize buffer if extra length required */
    if (comms_send_buffer_resize(csb, length + 1))
    {
      /* Print into buffer */
      length = vsnprintf(&csb->buffer[csb->used_size], length + 1, format, argcopy);
      if (length > 0)
        csb->used_size += length;
    }
  }
  va_end(argcopy);
  return length;
}


/* printf into a comms send buffer */
int comms_send_buffer_printf(struct s_comms_send_buffer *csb, const char *format, ...)
{
  va_list args;
  int length;

  va_start(args, format);
  length = comms_send_buffer_vprintf(csb, format, args);
  va_end(args);
  return length;
}
