/*
 * comms_setup.c
 *
 *  Created on: 13/10/2010
 *      Author: ian
 */
#include <stdio.h>
#include <time.h>
#include <rinhw/rio.h>
#include <rinhw/sio.h>
#include "database.h"
#include "comms_task.h"
#include "serial_ports.h"

#include "comms_setup.h"


/** Wait for the transmit buffer to empty before reconfiguring the port to stop data corruption after TDD1; */
static void comms_internal_waittx(struct t_io_dev *port)
{
  t_ioctl_siocfg config;
  struct timespec ts;
  uint32_t bits;
  unsigned short level;
  unsigned char cts;

  /* Get current serial port config */
  if (rioctl(port, IOCTL_SER_GET_CONFIG, &config) < 0)
  {
    /* Failed, set some defaults */
    config.baud = 9600;
    config.data_bits = SIO_DEF_DATA_BITS;
    config.stop_bits = SIO_DEF_STOP_BITS;
    config.parity_bits = SIO_DEF_PARITY_BITS;
  }

  /* Work out how long it takes send a character at the current configuration */
  bits = 1 + config.data_bits + config.stop_bits + (config.parity_bits != SIO_PARITY_NONE) ? 1 : 0;
  ts.tv_nsec = 1000000000 / config.baud;
  ts.tv_nsec *= bits;
  ts.tv_sec = 0;

  /* Wait for the transmit buffer to empty before reconfiguring the port to stop data corruption after TDD1; */
  level = 0;
  do
  {
    /*
     * If CTS is blocking transmission on this port,
     * then there is no point waiting for the transmit buffer to empty
     * before re-configuring this port.
     */
    cts = 1;
    rioctl(port, IOCTL_SER_GET_CTS_STATE, &cts);
    if (0 == cts)
      break;

    /* Get transmit buffer level */
    if (rioctl(port, IOCTL_SER_GET_TX_BUF_LEVEL, &level) < 0)
      break;

    /*
     * We need to do at least one character time sleep here,
     * because even if the buffer is empty,
     * there may be one character still being transmitted by the hardware
     */
    nanosleep(&ts, NULL);
  }
  while (level > 0);
}


/**
 * Configure a serial port with the database settings
 *
 * Assumes port is already open.
 *
 * @param port is the serial port device to reconfigure
 * @param port_index
 * @param wait_transmit causes the function to wait until the transmit buffer is empty before reconfiguring it
 */
void comms_internal_reconfigure(struct t_io_dev *serdev, enum eSerialPort port, bool wait_transmit)
{
  t_ioctl_siocfg config;
  bool use_rs232;

  if (port <= SERPORT_RB2B ||
      SERPORT_USBHOST == port ||
      SERPORT_SER2 == port)
  {
    /* Configured port, so get settings from database */
    db_get_serialconfig(port, &config, &use_rs232);
  }
  else
  {
    /* No configuration for this port, so set defaults */
    db_get_serialconfig_default(&config, &use_rs232);
  }

  if (wait_transmit)
  {
    comms_internal_waittx(serdev);
  }

  /* Configure the serial device */
  rioctl(serdev, IOCTL_SER_SET_CONFIG, &config);
  rioctl(serdev, IOCTL_SER_SET_RS232_NOT_RS422, (use_rs232) ? 1 : 0);
}
