#include <string.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "meas.h"
#include "access.h"
#include "apperror.h"
#include "kbd_5000.h"
#include "version.h"
#include "setpoint.h"
#include "display.h"
#include "comms_print.h"
#include "comms_setup.h"
#include "comms_kbd.h"
#include "comms_util.h"
#include "comms_task.h"
#include "serial_levels.h"
#include "comms_auto.h"
#include "we_config.h"
#include "adcthread.h"
#include "app_db_factory.h"
#include "app_db_user.h"
#include "app_db_runtime.h"
#include "backgrnd.h"
#include "temperature.h"

#include "comms_stxetx.h"


/** Send the "ZA" part of a test command (eg ZA123400) as a response */
void testcommand_sendreply(struct s_comms_control *comms_ctrl, const char *cmd)
{
  comms_net_rfprintf(comms_ctrl, "%c%c" COMMS_TERMINATOR,
                 cmd[0], cmd[1]);
}


void SendKrData(struct s_comms_control *comms_ctrl)
{
  char stat;
  uint8_t outputs;
  AdcRdg rdg;

  /* This is done using gross reading */
  AdcMutexLock(&adcSuperStructure);
  rdg = adcSuperStructure.adcDataWorking.grossRdg;
  AdcMutexUnlock(&adcSuperStructure);

  /* Work out status */
  if (rdg.info.status.bit.gross)
    stat = 'G';
  else
    stat = 'N';
  if (APP_ERR_NONE != get_current_errors())
    stat = 'E';
  else if (rdg.info.status.bit.oload)
    stat = 'O';
  else if (rdg.info.status.bit.uload)
    stat = 'U';
  else if (rdg.info.status.bit.motion)
    stat = 'M';

  /* Work out outputs */
  outputs = access_get_outputs();

  /* Print reply */
  comms_net_rfprintf(comms_ctrl, "%ld,%c,%02X" COMMS_TERMINATOR, rdg.rdg, stat, outputs);
}


/** Handle the <stx>Tx00<etx> commands */
static void send_test_data(struct s_comms_control *comms_ctrl, const char *s)
{
  t_datetime dt;

  switch (s[1])
  {

    /* System readings */
    case 'R':
      comms_net_rfprintf(comms_ctrl, "%ld,%ld,%ld,%ld,%lu,%u,%lu" COMMS_TERMINATOR,
                     get_pos_sense_reading(),
                     get_neg_sense_reading(),
                     get_12V_reading(),
                     temperature_get(tuKelvin),
                     0UL,  /* was front_keys */
                     (get_serial_cts_blocking(&dev_ser_front_rl)) ? 1 : 0,
                     (unsigned long) access_get_inputs()
                    );
      break;

    /* System time */
    case 'T':
      (void) GetDateTime(&dt);
      comms_net_rfprintf(comms_ctrl, "%u,%u,%u,%u,%u,%u" COMMS_TERMINATOR,
                     dt.date,
                     dt.month,
                     dt.year % 100,
                     dt.hour,
                     dt.minute,
                     dt.second
                    );
      break;

    /* Millivolt per volt reading */
    case 'A':
      comms_net_rfprintf(comms_ctrl, "%ld" COMMS_TERMINATOR,
                         adcSuperStructure.adcDataWorking.mvvAbsRdg * 256 /* *256 required so it matches the WE2110 result (which is mV/V * 10000 * 256) */
                        );
      break;

    /* System identification */
    case 'I':
      comms_net_rfprintf(comms_ctrl, "%lu,%s,%s,%02X" COMMS_TERMINATOR,
                         db_app_factory.serial_no,
                         db_app_factory.model_str,
                         app_version_str,
                         db_app_factory.economy
                        );
      break;
  }
}


/** Handle the <stx>Yx00<etx> commands */
static void handle_fac_rd(struct s_comms_control *comms_ctrl, const char *s)
{
  long l = 0;

  if (s[1] == 'A')
    l = db_app_factory.serial_no;
  if (s[1] == 'B')
    l = adcSuperStructure.adcFactoryDatabase.spanCal[adcSuperStructure.adcUserDatabase.syncFreqIdx];
  if (s[1] == 'C')
    l = adcSuperStructure.adcFactoryDatabase.spanDriftPos[adcSuperStructure.adcUserDatabase.syncFreqIdx];
  if (s[1] == 'D')
    l = adcSuperStructure.adcFactoryDatabase.calTemp;
  if (s[1] == 'E')
    l = db_app_factory.economy;
  if (s[1] == 'F')
    l = db_app_factory.application_id;
  if (s[1] == 'J')
    l = adcSuperStructure.adcFactoryDatabase.calFlag;
  if (s[1] == 'K')
    l = adcSuperStructure.adcUserDatabase.spanCnt;
  if (s[1] == 'L')
    l = adcSuperStructure.adcUserDatabase.spanRaw;
  if (s[1] == 'M')
    l = adcSuperStructure.adcUserDatabase.fullScale[0];
  if (s[1] == 'N')
    l = adcSuperStructure.adcUserDatabase.hiRes;
  if (s[1] == 'O')
    l = adcSuperStructure.adcUserDatabase.syncFreqIdx;
  if (s[1] == 'R')
    l = db_app_user.address;
  if (s[1] == 'U')
    l = (db_app_user.serial_bit_settings[0] & DB_SERIAL_USETERM_MASK) ? 1 : 0;
  if (s[1] == 'V')
    l = (db_app_user.serial_bit_settings[0] & DB_SERIAL_USERS232_MASK) ? 1 : 0;

  /* Send the result */
  comms_net_rfprintf(comms_ctrl, "%ld" COMMS_TERMINATOR, l);
}


/** Handle <stx>Kx00<etx> commands */
static void handle_k_cmds(struct s_comms_control *comms_ctrl, const char *msg)
{
  comms_ctrl->kr_cmd_streaming = false;
  switch (msg[1])
  {
    case 'p':
      comms_auto_send_request(comms_ctrl->serdev);
      break;

    case 'r':
      SendKrData(comms_ctrl);
      break;

    case 's':
      comms_ctrl->kr_cmd_streaming = true;
      break;

    default:
      comms_kbd_put_key_char(msg[1]);
      break;
  }
}


/** Process <stx><etx> command stored in msg. Will trash msg */
void process_stxetx_cmd(struct s_comms_control *comms_ctrl, char *msg)
{
  uint8_t in_addr;
  bool temp_active;
/*  long l; */
  size_t len;

  /* Get length of message and check it's ok */
  len = strlen(msg);
  if (len < 2)
  {
    return;
  }

  /* CHECK POLL ADDRESS */
  in_addr = (uint8_t) (((msg[len - 2] - '0') * 10) + (msg[len - 1] - '0'));

  /* ONLY CHECK ADDRESS IF NOT ZERO OR OUR ADDR NOT ZERO */
  if (in_addr != 0 && db_app_user.address != in_addr)
  {
    return;
  }

  /* remove address */
  msg[len - 2] = '\0';
  len -= 2;

  /* Set reply_active */
  temp_active = comms_ctrl->reply_active;
  comms_ctrl->reply_active = true;

  switch (msg[0])
  {
    case 'K':
      handle_k_cmds(comms_ctrl, msg);
      break;

    case 'T':
      send_test_data(comms_ctrl, msg);
      break;

    case 'Y':
      handle_fac_rd(comms_ctrl, msg);
      break;
  }
  comms_ctrl->reply_active = temp_active;
}
