/*
 * comms_task.c
 *
 *  Created on: 13/10/2010
 *      Author: ian
 */
#include <pthread.h>
#include <semaphore.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <rinhw/rinthread.h>
#include <rinhw/rioctl.h>
#include <rinhw/sio_local_established_socket.h>
#include <rinhw/timemath.h>
#include "database.h"
#include "comms_setup.h"
#include "iface.h"
#include "we_debug.h"
#include "we_config.h"
#include "comms_auto_trigger.h"
#include "comms_print_init.h"
#include "comms_print.h"
#include "comms_net.h"
#include "comms_print_buffer.h"
#include "comms_regtable.h"
#include "rc_gen_basic.h"
#include "db_serial.h"
#include "db_serial_ip.h"
#include "iface_serial_ports.h"
#include "comms_task_listen.h"
#include "comms_task_udp.h"

#include "comms_task.h"


/** Maximum number of comms threads */
#define MAX_COMMS_THREADS SERPORT_NUM


/** Structure to keep track of all the comms threads */
struct s_comms_thread
{
  volatile bool forcequit;              /*!< Should the thread quit? */
  bool used;                            /*!< Is this thread currently in use? */
  volatile bool reconfigure;            /*!< Should this thread re-configure itself from database settings? */
  bool exited;                          /*!< Thread has exited and is awaiting cleaning, used for TCP connections */
  bool listenercreated;                 /*!< Created from a listening thread? */
  enum eSerialPort serialport;          /*!< The serial port (ie rinlink front, etc). comms_thread_list[SERPORT_NETWORK_START..SERPORT_NUM].serialport will be one of SERPORT_TCPNET, SERPORT_TCPAUTO, SERPORT_UDSNET, SERPORT_UDSAUTO */
  enum eSerialPortType serialtype;      /*!< What the serial port operates as (ie autoout, network) */
  pthread_t thread;                     /*!< The pthread */
  struct t_io_dev *serdev;              /*!< The rindevice we do the IO over */
  struct s_comms_auto *ca;              /*!< Comms auto structure for running as automatic output */
  struct s_comms_print *cp;             /*!< Comms print structure for running as printer */
  struct s_comms_net *cn;               /*!< Comms net structure for running as network comms */
  struct s_comms_print_buffer *cpb;     /*!< The comms print buffer */
  struct timespec ts_last_tx_buf_empty; /*!< Last time that the transmit buffer was empty */
  uint32_t timeout_seconds;             /*!< Number of seconds for the timeout */
  struct s_comms_udp_info udp_info;     /*!< If this is a UDP port, this port info and who we are communicating with. All members in host endian. */
};


/** List of all comms threads */
static struct s_comms_thread comms_thread_list[MAX_COMMS_THREADS];


/** Get serial type as a string */
static const char *get_serial_type_as_string(enum eSerialPortType serialtype)
{
  if (serialtype > NUM_DB_SERIAL_TYPES)
    return "unknown";
  else if (serialtype == NUM_DB_SERIAL_TYPES)
    return "unconfigured";
  else
    return ser_type_setting[serialtype];
}


/** Serial port names, must match @ref enum eSerialPort */
static const char *const ser_port_strings[SERPORT_NETWORK_START] = SERIAL_PORT_STRINGS;


/** Serial network port names, must match @ref enum eSerialPort */
static const char *const ser_tcp_port_strings[NUM_TCP_LISTENER_PORTS] = SERIAL_TCP_PORT_STRINGS;


/** Serial unix domain socket port names, must match @ref enum eSerialPort */
static const char *const ser_uds_port_strings[NUM_UDS_LISTENER_PORTS] = SERIAL_UDS_PORT_STRINGS;


/** Get serial port name as a string */
const char *get_serial_port_as_string(enum eSerialPort serialport)
{
  if (serialport >= SERPORT_TCP_FIRST && serialport <= SERPORT_TCP_FINAL)
    return ser_tcp_port_strings[serialport - SERPORT_TCP_FIRST];
  if (serialport >= SERPORT_UDS_FIRST && serialport <= SERPORT_UDS_FINAL)
    return ser_uds_port_strings[serialport - SERPORT_UDS_FIRST];
  else if (serialport < SERPORT_NETWORK_START)
    return ser_port_strings[serialport];
  else
    return "unknown";
}


/** Determine if port is a TCP connection */
bool comms_is_tcp_connection(enum eSerialPort serialport)
{
  return (serialport >= SERPORT_TCP_FIRST && serialport < SERPORT_TCP_FINAL);
}


/** Determine if port is a UDS (unix domain socket) connection */
bool comms_is_uds_connection(enum eSerialPort serialport)
{
  return (serialport >= SERPORT_UDS_FIRST && serialport < SERPORT_UDS_FINAL);
}

/** Determine if port is a stream connection (TCP or UDS) */
bool comms_is_stream_connection(enum eSerialPort serialport)
{
  return (comms_is_tcp_connection(serialport) || comms_is_uds_connection(serialport));
}


/** Determine if port is an incoming UDP connection */
bool comms_is_udp_in_connection(enum eSerialPort serialport)
{
  return (serialport >= SERPORT_UDPIN1 && serialport <= SERPORT_UDPIN2);
}


/** Determine if port is an outgoing UDP connection */
bool comms_is_udp_out_connection(enum eSerialPort serialport)
{
  return (serialport >= SERPORT_UDPOUT1 && serialport <= SERPORT_UDPOUT2);
}


/** Determine if port is a UDP connection */
bool comms_is_udp_connection(enum eSerialPort serialport)
{
  return comms_is_udp_in_connection(serialport) || comms_is_udp_out_connection(serialport);
}


/**
 * Reset a comms thread structure
 * Must have the lock before calling this helper
 */
static void comms_reset_thread(struct s_comms_thread *const p)
{
  p->used = false;
  if (p->listenercreated && NULL != p->serdev)
  {
    if (RIO_SUCCESS != rdestroy_io(p->serdev))
      rdebug_printf(LOG_ERR, "comms_reset_thread: destroying socket serial device failed\n");
    free(p->serdev);
  }
  p->serdev = NULL;
  p->forcequit = false;
  p->exited = false;
  p->serialport = SERPORT_UNKNOWN;
  p->listenercreated = false;
  p->serialtype = NUM_DB_SERIAL_TYPES;
  p->ca = NULL;
  p->cp = NULL;
  p->cn = NULL;
  p->cpb = NULL;
  p->udp_info.localport = 0;
  memset(&p->udp_info.dest, 0, sizeof(p->udp_info.dest));
  p->udp_info.dest.sin_addr.s_addr = 0;
  p->udp_info.dest.sin_port = 0;
}


/** Mutex to lock access */
static pthread_mutex_t comms_thread_list_mutex;


/** The comms print buffer - global to all comms threads, but passed around via struct s_comms_thread member cpb */
struct s_comms_print_buffer *comms_print_buffer = NULL;


/** Re-initialise the appropriate data structure for a type of serial output from the database */
static void comms_reinit_by_type(struct s_comms_thread *ct, enum eSerialPortType sertype)
{
  switch (sertype)
  {
    case DB_SERIAL_TYPE_AUTO_LO:
    case DB_SERIAL_TYPE_AUTO_HI:
    case DB_SERIAL_TYPE_SINGLE:
      comms_auto_reinit(ct->ca);
      break;

    case DB_SERIAL_TYPE_PRINTING:
      comms_print_reinit(ct->cp);
      break;

    case DB_SERIAL_TYPE_NETWORK:
    case DB_SERIAL_TYPE_MODBUS_LEGACY:
      comms_net_reinit(ct->cn);
      break;

    case DB_SERIAL_TYPE_OFF:
    default:
      break;
  }
}


/** Create the necessary data structure for a type of serial output */
static void comms_create_by_type(struct s_comms_thread *ct, enum eSerialPortType sertype)
{
  bool is_udp;
  bool is_incoming_udp;

  is_incoming_udp = comms_is_udp_in_connection(ct->serialport);
  switch (sertype)
  {
    case DB_SERIAL_TYPE_AUTO_LO:
    case DB_SERIAL_TYPE_AUTO_HI:
    case DB_SERIAL_TYPE_SINGLE:
      ct->ca = comms_auto_create(ct->serdev, is_incoming_udp);
      break;

    case DB_SERIAL_TYPE_PRINTING:
      ct->cp = comms_print_create(ct->serdev, ct->cpb, is_incoming_udp);
      break;

    case DB_SERIAL_TYPE_NETWORK:
    case DB_SERIAL_TYPE_MODBUS_LEGACY:
      is_udp = comms_is_udp_connection(ct->serialport);
      ct->cn = comms_net_create(ct->serdev, ct->cpb, is_udp, &ct->udp_info);
      break;

    case DB_SERIAL_TYPE_OFF:
    default:
      ct->ca = NULL;
      ct->cp = NULL;
      ct->cn = NULL;
      break;
  }
}


/** Create the necessary data structure for a type of serial output */
static void comms_destroy_by_type(struct s_comms_thread *ct, enum eSerialPortType sertype)
{
  switch (sertype)
  {
    case DB_SERIAL_TYPE_AUTO_LO:
    case DB_SERIAL_TYPE_AUTO_HI:
    case DB_SERIAL_TYPE_SINGLE:
      comms_auto_destroy(ct->ca);
      break;

    case DB_SERIAL_TYPE_PRINTING:
      comms_print_destroy(ct->cp);
      break;

    case DB_SERIAL_TYPE_NETWORK:
    case DB_SERIAL_TYPE_MODBUS_LEGACY:
      comms_net_destroy(ct->cn);
      break;

    case DB_SERIAL_TYPE_OFF:
    default:
      break;
  }
  ct->ca = NULL;
  ct->cp = NULL;
  ct->cn = NULL;
}


/** Print the number of used/available TCP connections */
static void comms_tcp_avail(void)
{
  size_t i;
  const size_t total = (SERPORT_NUM - SERPORT_NETWORK_START);
  size_t used;

  pthread_mutex_lock(&comms_thread_list_mutex);

  for (i = SERPORT_NETWORK_START, used = 0; i < MAX_COMMS_THREADS; i++)
  {
    if (comms_thread_list[i].used)
    {
      used++;
    }
  }
  pthread_mutex_unlock(&comms_thread_list_mutex);

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_tcp_avail: there are %u connections in use of total %u\n",
                             (unsigned) used, (unsigned) total);
}


/**
 * Deal with timeouts on TCP connections
 *
 * Write timeout occurs if there is un-acknowledged data in the TCP
 * transmit buffer for longer than the timeout period.
 *
 * Read timeout occurs only for network connections when no
 * data has been received for longer than the timeout period.
 *
 * Returns true if a timeout has occurred.
 */
static bool comms_check_timeouts(struct s_comms_thread *const ct)
{
  uint16_t tx_buffer_level;
  struct timespec now;
  bool timedout = false;

  if (comms_is_tcp_connection(ct->serialport) && ct->timeout_seconds > 0)
  {
    /* Transmit buffer timeout */
    tx_buffer_level = 0;
    rioctl(ct->serdev, IOCTL_SER_GET_TX_BUF_LEVEL, &tx_buffer_level);
    if (tx_buffer_level > 0)
    {
      /* Un-acked data in output buffer, have we timed out yet */
      clock_gettime(CLOCK_MONOTONIC, &now);
      SubTimespec(&now, &ct->ts_last_tx_buf_empty);
      if (now.tv_sec > (time_t) ct->timeout_seconds)
      {
        WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_check_timeouts: write timeout occurred on port %d(%s), type %d(%s), closing connection\n",
                                   (int) ct->serialport, get_serial_port_as_string(ct->serialport),
                                   (int) ct->serialtype, get_serial_type_as_string(ct->serialtype));
        timedout = true;
      }
    }
    else
    {
      /* Buffer empty, set last time transmit buffer was empty */
      clock_gettime(CLOCK_MONOTONIC, &ct->ts_last_tx_buf_empty);
    }

    /* Read timeout only applies to bi-directional port types */
    if (DB_SERIAL_TYPE_NETWORK == ct->serialtype)
    {
      clock_gettime(CLOCK_MONOTONIC, &now);
      SubTimespec(&now, &ct->cn->ts_last_rx_data);
      if (now.tv_sec > (time_t) ct->timeout_seconds)
      {
        WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_check_timeouts: read timeout occurred on port %d(%s), type %d(%s), closing connection\n",
                                   (int) ct->serialport, get_serial_port_as_string(ct->serialport),
                                   (int) ct->serialtype, get_serial_type_as_string(ct->serialtype));
        timedout = true;
      }
    }
  }
  return timedout;
}


/** Reconfigure the timeout */
static void comms_timeout_reconfigure(struct s_comms_thread *ct)
{
  /* Get timeout value */
  ct->timeout_seconds = db_get_serial_tcp_timeout(ct->serialport);

  /* Preset timeout */
  clock_gettime(CLOCK_MONOTONIC, &ct->ts_last_tx_buf_empty);
  if (DB_SERIAL_TYPE_NETWORK == ct->serialtype)
    clock_gettime(CLOCK_MONOTONIC, &ct->cn->ts_last_rx_data);
}


static void comms_type_reconfigure(struct s_comms_thread *ct, bool first_configure)
{
  const enum eSerialPortType newsertype = db_get_serial_type(ct->serialport);

  /* Re-create comms if first configuration, serial type has changed */
  if (first_configure || newsertype != ct->serialtype)
  {
    /* Yeah, destroy whatever it was */
    comms_destroy_by_type(ct, ct->serialtype);

    /* Make a new one of appropriate type */
    comms_create_by_type(ct, newsertype);
  }
  else
  {
    /* If type unchanged, we want to re-initialise bits from the database, but nothing else */
    comms_reinit_by_type(ct, newsertype);
  }

  /*
   * Normal re-configures only flush data if the serial type has changed (ie auto.lo -> net)
   * First time configures only flush data if it's not a TCP connection, because every new TCP connection is a first time configuration (new thread)
   */
  if ((!first_configure && newsertype != ct->serialtype) || (first_configure && !comms_is_tcp_connection(ct->serialport)))
  {
    rioctl(ct->serdev, IOCTL_SER_FLUSH_RX_BUF);
  }

  /* Re-configure timeouts */
  comms_timeout_reconfigure(ct);

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_type_reconfigure: changed port %d(%s) from %d(%s) to %d(%s)\n",
                             ct->serialport, get_serial_port_as_string(ct->serialport),
                             (int) ct->serialtype, get_serial_type_as_string(ct->serialtype),
                             (int) newsertype, get_serial_type_as_string(newsertype));

  /* Update serial type */
  ct->serialtype = newsertype;
}


/** Cork/uncork a TCP connection. Checks first if the connection is TCP, and doesn't bother if not TCP connection. */
static void comms_cork_tcp(struct s_comms_thread *ct, bool cork)
{
  int fd;
  int state = (cork) ? 1 : 0;

  /* Don't bother with non-TCP connections */
  if (!comms_is_tcp_connection(ct->serialport))
  {
    return;
  }

  /* Find out the underlying Linux file descriptor */
  if (rioctl(ct->serdev, IOCTL_GET_LINUX_FD, &fd) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_cork_tcp: failed to determine Linux fd\n");
    return;
  }

  /* Cork/uncork the connection */
  if (setsockopt(fd, IPPROTO_TCP, TCP_CORK, &state, sizeof(state)) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_cork_tcp: failed to %s port\n", (cork) ? "cork" : "uncork");
    return;
  }
}


/** Set/clear TCP_NODELAY on a TCP connection. Checks first if the connection is TCP, and doesn't bother if not TCP connection. */
static void comms_nodelay_tcp(struct s_comms_thread *ct, bool nodelay)
{
  int fd;
  int state = (nodelay) ? 1 : 0;

  /* Don't bother with non-TCP connections */
  if (!comms_is_tcp_connection(ct->serialport))
  {
    return;
  }

  /* Find out the underlying Linux file descriptor */
  if (rioctl(ct->serdev, IOCTL_GET_LINUX_FD, &fd) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_nodelay_tcp: failed to determine Linux fd\n");
    return;
  }

  /* Cork/uncork the connection */
  if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &state, sizeof(state)) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_nodelay_tcp: failed to %s TCP_NODELAY on port\n", (nodelay) ? "set" : "clear");
    return;
  }
}


/** Handle comms depending on the type */
static void comms_handle_by_type(struct s_comms_thread *ct)
{
  if (NULL == ct)
  {
    OSTimeDly(OS_TICKS_PER_SEC);
    return;
  }

  /* Cork the connection for TCP */
  comms_cork_tcp(ct, true);

  /* Handle based on type */
  switch (ct->serialtype)
  {
    case DB_SERIAL_TYPE_AUTO_LO:
    case DB_SERIAL_TYPE_AUTO_HI:
    case DB_SERIAL_TYPE_SINGLE:
      comms_auto_handler(ct->ca);
      break;

    case DB_SERIAL_TYPE_PRINTING:
      comms_print_handler(ct->cp);
      break;

    case DB_SERIAL_TYPE_NETWORK:
    case DB_SERIAL_TYPE_MODBUS_LEGACY:
      (void) comms_net_handler(ct->cn);
      break;

    case DB_SERIAL_TYPE_OFF:
    default:
      /* Simply yield for a bit, then loop again in case of a change */
      OSTimeDly(OS_TICKS_PER_SEC);
      break;
  }

  /* Uncork the connection for TCP */
  comms_cork_tcp(ct, false);
}


/** Comms handling thread, pass serial port device as data */
static void *comms_thread(void *data)
{
  struct s_comms_thread *const ct = (struct s_comms_thread *const) data;
  bool first_configure;
  unsigned char cts;

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_thread: started for port %d(%s), pthread %lu, linux pid %d\n",
                             (int) ct->serialport, get_serial_port_as_string(ct->serialport),
                             (unsigned long) pthread_self(), (int) getpid());
  if (comms_is_stream_connection(ct->serialport))
    comms_tcp_avail();

  /* Open serial port, if not already open */
  if (ct->serdev->status & O_OPEN)
  {
    ropen(ct->serdev, 0);
  }

  /* Set TCP_NODELAY on all TCP ports */
  comms_nodelay_tcp(ct, true);

  /* First thing to do is configure serial port */
  ct->reconfigure = true;
  first_configure = true;

  /* Main thread loop */
  while (!ct->forcequit)
  {
    /* Reconfigure serial port */
    if (ct->reconfigure)
    {
      /* Reconfigure the port with the settings from the database */
      comms_internal_reconfigure(ct->serdev, ct->serialport, true);
      ct->reconfigure = false;

      /* Change the port type if required */
      comms_type_reconfigure(ct, first_configure);

      if (comms_is_udp_connection(ct->serialport))
      {
        /* Comms UDP reconfigure (port changes, quit task if disabled) */
        if (!comms_task_udp_running_reconfigure(ct->serialport, ct->serdev, &ct->udp_info))
          break;
      }

      first_configure = false;
    }

    if (edit_functions && DB_SERIAL_TYPE_PRINTING != ct->serialtype && DB_SERIAL_TYPE_NETWORK != ct->serialtype)
    {
      /* Do nothing if in menus (except printing to allow the database print to occur, and network comms to allow read of settings and viewer to navigate menus) */
      OSTimeDly(OS_TICKS_PER_SEC);

      /* Flush the serial port of anything that has come in */
      rioctl(ct->serdev, IOCTL_SER_FLUSH_RX_BUF);
    }
    else
    {
      /* Handle the particular serial port type */
      comms_handle_by_type(ct);
    }

    /* Check to see if the socket is still open or not */
    if (ct->listenercreated)
    {
      cts = 1;
      rioctl(ct->serdev, IOCTL_SER_GET_CTS_STATE, &cts);
      if (!cts)
        break;
    }

    /* Check for timeouts */
    if (comms_check_timeouts(ct))
    {
      break;
    }
  }

  /* Clean up and exit */
  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_thread: loop finished for port %d(%s), type %d(%s)\n",
                             (int) ct->serialport, get_serial_port_as_string(ct->serialport),
                             (int) ct->serialtype, get_serial_type_as_string(ct->serialtype));

  comms_destroy_by_type(ct, ct->serialtype);
  rclose(ct->serdev);

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_thread: finished for port %d(%s), type %d(%s)\n",
                             (int) ct->serialport, get_serial_port_as_string(ct->serialport),
                             (int) ct->serialtype, get_serial_type_as_string(ct->serialtype));

  if (!ct->forcequit)
  {
    ct->exited = true;
    comms_task_listen_cleanup();
  }
  return NULL;
}


/** Create a new comms handler thread for a serial port. udp info in host endian. */
int comms_create_thread_for_serial_port(struct t_io_dev *port, enum eSerialPort serialport, const struct s_comms_udp_info *udpinfo)
{
  bool rc;
  int result = RIO_ERROR;
  struct s_comms_thread *c = NULL;
  size_t i;

  pthread_mutex_lock(&comms_thread_list_mutex);

  if (serialport < SERPORT_NETWORK_START)
  {
    /* One of the fixed ports */
    c = &comms_thread_list[serialport];
    if (c->used)
    {
      WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_create_thread_for_serial_port: fixed port already used %d\n", serialport);
      goto cctfsp_done;
    }
  }
  else
  {
    /* A network port, find an empty slot */
    for (i = SERPORT_NETWORK_START; i < MAX_COMMS_THREADS; i++)
    {
      if (!comms_thread_list[i].used)
      {
        c = &comms_thread_list[i];
        break;
      }
    }

    /* No room for any more */
    if (c == NULL)
    {
      WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_create_thread_for_serial_port: no room for additional network ports\n");
      goto cctfsp_done;
    }
  }

  /* Setup the new comms thread */
  c->used = true;
  c->serdev = port;
  c->forcequit = false;
  c->reconfigure = true;
  c->serialport = serialport;
  c->listenercreated = comms_is_stream_connection(serialport);
  c->serialtype = NUM_DB_SERIAL_TYPES;
  c->ca = NULL;
  c->cp = NULL;
  c->cn = NULL;
  c->cpb = comms_print_buffer;
  if (NULL != udpinfo)
  {
    c->udp_info = *udpinfo;
  }
  else
  {
    c->udp_info.localport = 0;
    c->udp_info.dest.sin_addr.s_addr = 0;
    c->udp_info.dest.sin_port = 0;
  }
  rc = rin_thread_create(&c->thread, comms_thread, c, true);
  if (!rc)
  {
    /* Fail */
    WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_create_thread_for_serial_port: create comms thread failed\n");
    comms_reset_thread(c);
  }
  else
  {
    result = RIO_SUCCESS;
  }

cctfsp_done:
  pthread_mutex_unlock(&comms_thread_list_mutex);

  return result;
}


/** Kill a specific serial port task */
void comms_destroy_task(struct t_io_dev *port)
{
  size_t i;

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_destroy_task: called\n");

  pthread_mutex_lock(&comms_thread_list_mutex);

  /* Look for the specific serial port task, and force it to quit */
  for (i = 0; i < MAX_COMMS_THREADS; i++)
  {
    if (comms_thread_list[i].used &&
        comms_thread_list[i].serdev == port)
    {
      WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_destroy_task: Notifying task to quit\n");

      /* Wait for thread to quit */
      comms_thread_list[i].forcequit = true;
      (void) pthread_join(comms_thread_list[i].thread, NULL);

      /* Clear stuff */
      comms_reset_thread(&comms_thread_list[i]);
      break;
    }
  }

  pthread_mutex_unlock(&comms_thread_list_mutex);

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_destroy_task: finished\n");
}


/** Kill all serial port tasks */
void comms_tasks_destroy(void)
{
  size_t i;
  size_t count;

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_tasks_destroy: called\n");

  pthread_mutex_lock(&comms_thread_list_mutex);

  /* Force each used comms thread to quit */
  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_tasks_destroy: Notifying tasks to quit\n");
  for (i = 0, count = 0; i < MAX_COMMS_THREADS; i++)
  {
    if (comms_thread_list[i].used)
    {
      comms_thread_list[i].forcequit = true;
      count++;
    }
  }

  /* Wait for each thread to quit */
  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_tasks_destroy: Waiting for %u tasks to quit\n", (unsigned) count);
  for (i = 0; i < MAX_COMMS_THREADS; i++)
  {
    if (comms_thread_list[i].used)
    {
      /* Wait for each thread to quit */
      WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_tasks_destroy: Waiting for task %d(%s) to quit\n",
                                 (int) comms_thread_list[i].serialport, get_serial_port_as_string(comms_thread_list[i].serialport));
      (void) pthread_join(comms_thread_list[i].thread, NULL);
      WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_tasks_destroy: Task %d(%s) has quit\n",
                                 (int) comms_thread_list[i].serialport, get_serial_port_as_string(comms_thread_list[i].serialport));

      /* Clear stuff */
      comms_reset_thread(&comms_thread_list[i]);
    }
  }

  pthread_mutex_unlock(&comms_thread_list_mutex);

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_tasks_destroy: finished\n");
}


/**
 * Force comms port(s) to reconfigure from database settings
 *
 * @param all set true if you want all ports to be re-configured, false if a specific port
 * @param port is the specific port to configure if all is false
 */
void comms_reconfigure(bool all, enum eSerialPort port)
{
  size_t i;

  if (all || comms_is_tcp_connection(port))
  {
    /* Full reconfigure means TCP ports may need to change */
    comms_task_listen_reconfigure();
  }

  if (all || comms_is_udp_connection(port))
  {
    /* And new UDP tasks may need to start */
    comms_task_udp_reconfigure();
  }

  pthread_mutex_lock(&comms_thread_list_mutex);
  if (all || comms_is_tcp_connection(port))
  {
    /*
     * Force each used comms thread to reconfigure itself.
     * Existing TCP tasks have a timeout value, so may need re-configuring
     */
    for (i = 0; i < MAX_COMMS_THREADS; i++)
    {
      if (comms_thread_list[i].used &&
          (all || comms_thread_list[i].serialport == port))
      {
        comms_thread_list[i].reconfigure = true;
      }
    }
  }
  else if (port < MAX_COMMS_THREADS)
  {
    if (comms_thread_list[port].used)
    {
      comms_thread_list[port].reconfigure = true;
    }
  }
  pthread_mutex_unlock(&comms_thread_list_mutex);
}


/** Force comms port to trigger an automatic output, if it is configured as one */
void comms_auto_trigger(enum eSerialPort port)
{
  pthread_mutex_lock(&comms_thread_list_mutex);

  if (port < SERPORT_NUM &&
      comms_thread_list[port].used &&
      comms_thread_list[port].ca != NULL)
  {
    auto_output_trigger(comms_thread_list[port].ca);
  }

  pthread_mutex_unlock(&comms_thread_list_mutex);
}


/**
 * Get the s_comms_print structure for a port
 * (locks it first - must call comms_print_release_cp() when done)
 *
 * @param port == serial port -> gets cp for that serial port, port == NULL -> first serial port defined as a printer
 */
static struct s_comms_print *comms_print_internal_get_cp(struct t_io_dev *port)
{
  size_t i;
  struct s_comms_print *cp = NULL;

  pthread_mutex_lock(&comms_thread_list_mutex);

  for (i = 0; i < MAX_COMMS_THREADS; i++)
  {
    if (comms_thread_list[i].used &&
        ((NULL == port && NULL != comms_thread_list[i].serdev) ||     /* First printer port found OR */
         (NULL != port && comms_thread_list[i].serdev == port)) &&    /* Specific printer port */
        comms_thread_list[i].cp != NULL)
    {
      cp = comms_thread_list[i].cp;
      pthread_mutex_lock(&cp->mutex);
      break;
    }
  }

  pthread_mutex_unlock(&comms_thread_list_mutex);

  /* Not found */
  return cp;
}


/**
 * Get the s_comms_print structure for a port
 * (locks it first - must call comms_print_release_cp() when done)
 *
 * @param port == serial port -> gets cp for that serial port, port == NULL -> USB printer if attached, else first serial port defined as a printer
 */
struct s_comms_print *comms_print_get_cp(struct t_io_dev *port)
{
  struct s_comms_print *cp = NULL;

  /* If looking for any printer, rather than specific one, then check USB printer first */
  if (NULL == port)
  {
    cp = comms_print_internal_get_cp(&dev_ser9_usblp);
  }

  if (NULL == cp)
  {
    /*
     * No USB printer, or looking for specific port,
     * so try again, this time taking the first printer we come across
     */
    cp = comms_print_internal_get_cp(NULL);
  }

  return cp;
}


/** Release a s_comms_print structure acquired by comms_print_get_cp() */
void comms_print_release_cp(struct s_comms_print *cp)
{
  if (NULL != cp)
  {
    pthread_mutex_unlock(&cp->mutex);
  }
}


/**
 * Create any tasks for the serial ports. Call only once.
 *
 * @param ports is a list of serial ports to create tasks for. Last entry must be NULL
 */
void comms_tasks_create(struct t_io_dev **ports, const enum eSerialPort *types)
{
  /* int res; */
  while (NULL != *ports)
  {
    /* Create comms handler */
    /* res = */ comms_create_thread_for_serial_port(*ports, *types, NULL);
    /* rdebug_printf(LOG_INFO, "Result of thread create for port %p = %d\n", *ports, res); */
    ports++;
    types++;
  }
}


/** Clean up a task if it has exited, only called by comms_task_list_clean() */
static bool comms_task_cleanup(enum eSerialPort port)
{
  bool cleaned = false;

  if (comms_thread_list[port].exited)
  {
    WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_cleaner: clean %u\n", (unsigned) port);
    (void) pthread_join(comms_thread_list[port].thread, NULL);
    comms_reset_thread(&comms_thread_list[port]);
    cleaned = true;
  }
  return cleaned;
}


/**
 * Run through the list of comms threads, and clean any terminated network threads.
 * If we don't do this, we'll end up with lots of zombie threads hanging around.
 */
void comms_task_list_clean(void)
{
  enum eSerialPort i;
  bool stream_cleaned;

  WE_COMMS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_cleaner: cleaner activated\n");

  pthread_mutex_lock(&comms_thread_list_mutex);
  /* Clean up any UDP ports that have quit */
  for (i = SERPORT_UDPOUT1; i <= SERPORT_UDPIN2; i++)
  {
    comms_task_cleanup(i);
  }
  /* And any stream ones (TCP/UDS) */
  for (i = SERPORT_NETWORK_START, stream_cleaned = false; i < SERPORT_NUM; i++)
  {
    if (comms_task_cleanup(i))
      stream_cleaned = true;
  }
  pthread_mutex_unlock(&comms_thread_list_mutex);
  if (stream_cleaned)
    comms_tcp_avail();
}


/** Initialise the comms task list. Call before any other comms_ functions. Returns false if failed */
bool comms_task_list_init(const char *uds_net_file, const char *uds_auto_file)
{
  size_t i;
  bool ctl_mutex_inited = false;
  bool listen_thread_inited = false;

  /* Initialise the rincmd register table */
  rinCmdTable = rc_initRegTable();
  if (NULL == rinCmdTable)
  {
    rdebug_printf(LOG_ERR, "comms_task_list_init: failed to create rincmd register table\n");
    goto ctli_fail;
  }
  rc_gen_basic(rinCmdTable);

  /* Set up the rincmd debug */
  rc_set_debug_printf_fn(rdebug_vprintf);

  /* Create mutex for thread list */
  if (pthread_mutex_init(&comms_thread_list_mutex, NULL) != 0)
  {
    rdebug_printf(LOG_ERR, "comms_task_list_init: failed to create thread list mutex\n");
    goto ctli_fail;
  }
  ctl_mutex_inited = true;

  /* Create comms print buffer, so all threads get to see it */
  comms_print_buffer = comms_print_buffer_create(WECFG_PRINT_BUFFER_SIZE);
  if (NULL == comms_print_buffer)
  {
    rdebug_printf(LOG_ERR, "comms_task_list_init: failed to create comms print buffer\n");
    goto ctli_fail;
  }

  /* Setup thread list variables */
  for (i = 0; i < MAX_COMMS_THREADS; i++)
  {
    comms_thread_list[i].used = false;
    comms_thread_list[i].serdev = NULL;
    comms_thread_list[i].listenercreated = false;
    comms_thread_list[i].cpb = comms_print_buffer;
  }

  /* Create listening thread */
  if (!comms_task_listen_init(uds_net_file, uds_auto_file))
  {
    rdebug_printf(LOG_ERR, "comms_task_list_init: failed to create listening thread\n");
    goto ctli_fail;
  }
  listen_thread_inited = true;

  /* Init the UDP tasks */
  comms_task_udp_init();

  /* Success */
  return true;

ctli_fail:
  /* Something failed, clean up */
  if (listen_thread_inited)
  {
    comms_task_listen_destroy();
  }
  if (NULL != comms_print_buffer)
  {
    comms_print_buffer_destroy(comms_print_buffer);
    comms_print_buffer = NULL;
  }
  if (ctl_mutex_inited)
    pthread_mutex_destroy(&comms_thread_list_mutex);
  if (NULL != rinCmdTable)
  {
    rc_freeRegTable(rinCmdTable);
    rinCmdTable = NULL;
  }

  return false;
}


/** Shutdown the comms task list. Call after everything else is shut down */
void comms_task_list_destroy(void)
{
  /* Kill all serial port tasks, incase they aren't already */
  comms_tasks_destroy();

  /* Get rid of the listening thread/sockets */
  comms_task_listen_destroy();

  /* Destroy print buffer */
  comms_print_buffer_destroy(comms_print_buffer);
  comms_print_buffer = NULL;

  /* Destroy mutex */
  pthread_mutex_destroy(&comms_thread_list_mutex);

  /* Free the register table */
  rc_freeRegTable(rinCmdTable);
  rinCmdTable = NULL;
}


/** Get the serial port type for a running serial port task */
enum eSerialPortType comms_task_list_get_port_type(enum eSerialPort port)
{
  enum eSerialPortType spt;

  switch (port)
  {
    case SERPORT_TCPNET:
    case SERPORT_UDSNET:
      spt = DB_SERIAL_TYPE_NETWORK;
      break;
    case SERPORT_TCPAUTO:
    case SERPORT_UDSAUTO:
      spt = DB_SERIAL_TYPE_AUTO_LO;
      break;
    default:
      spt = (port < MAX_COMMS_THREADS && comms_thread_list[port].used) ? comms_thread_list[port].serialtype : NUM_DB_SERIAL_TYPES;
      break;
  }
  return spt;
}


/** Return the serial port of the specified serial port device */
enum eSerialPort comms_task_list_get_port_from_device(const struct t_io_dev *const dev)
{
  enum eSerialPort i;

  for (i = 0; i < MAX_COMMS_THREADS; i++)
  {
    if (comms_thread_list[i].used && comms_thread_list[i].serdev == dev)
      return i;
  }
  return SERPORT_UNKNOWN;
}
