/*
 * comms_task_listen.c
 *
 *  Created on: Sep 25, 2017
 *      Author: ian
 */
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <errno.h>
#include <stdio.h>
#include <rinhw/rstdio.h>
#include <rinhw/sio_local_established_socket.h>
#include <rinhw/sio_local_established_dgram.h>
#include <rinhw/rinthread.h>
#include "we_debug.h"
#include "comms_task.h"
#include "db_serial_ip.h"
#include "ipinfo.h"

#include "comms_task_listen.h"


/** Number of listening threads and the listen queue length for each */
#define MAX_COMMS_TCP_LISTENERS (SERPORT_TCP_FINAL - SERPORT_TCP_FIRST)
#define MAX_COMMS_UDS_LISTENERS (SERPORT_UDS_FINAL - SERPORT_UDS_FIRST)
#define MAX_COMMS_LISTENERS (MAX_COMMS_TCP_LISTENERS + MAX_COMMS_UDS_LISTENERS)
#define MAX_TCP_CONNECTION_COUNT 5
#define MAX_UDS_CONNECTION_COUNT 5


/** Convert a SERPORT_UDS_* into an index into the comms_listener_ports array */
#define COMMS_UDS_LISTENER_IDX(sp) (MAX_COMMS_TCP_LISTENERS + (sp) - SERPORT_UDS_FIRST)


/** Structure to define the behavior of a listening thread & maintain its global state */
struct s_comms_listen_port
{
  char *uds_file;                 /*!< Pointer to socket file if this listener is a Unix domain socket, NULL otherwise */
  int port;                       /*!< Listen TCP port number if this is a TCP port, -1 otherwise  */
  int fd;                         /*!< Listen socket descriptor */
  enum eSerialPort serialport;    /*!< The serial port that this listener spawns */
};


struct s_comms_listener
{
  pthread_t pthread;
  struct s_comms_listen_port comms_listener_ports[MAX_COMMS_LISTENERS];
  int activate_read_fd;
  int activate_write_fd;
};


static struct s_comms_listener comms_listener;


#define COMMS_LISTENER_CMD_QUIT 'q'
#define COMMS_LISTENER_CMD_CLEAN 'c'
#define COMMS_LISTENER_CMD_RECONFIGURE 'r'


/** Establish a listening TCP socket for a port */
static bool comms_listener_establish_tcp(struct s_comms_listen_port *listen_port)
{
  int flag;
  struct sockaddr_in servaddr;

  listen_port->fd = socket(AF_INET, SOCK_STREAM, 0);
  if (listen_port->fd < 0)
  {
    rdebug_printf(LOG_ERR, "comms_listener_socket_establish: create listen socket for tcp port %d failed (errno %d)\n", listen_port->port, errno);
    goto cle_fail;
  }
  if (fcntl(listen_port->fd, F_SETFD, FD_CLOEXEC) < 0)
    rdebug_printf(LOG_ERR, "comms_listener_socket_establish: setting FD_CLOEXEC on socket for tcp port %d failed (errno %d)\n", listen_port->port, errno);

  /* Set the reuse flag on the socket before binding */
  flag = 1;
  if (setsockopt(listen_port->fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(int)) < 0)
    rdebug_printf(LOG_WARNING, "comms_listener_socket_establish: setting SO_REUSEADDR on socket for tcp port %d failed (errno %d)\n", listen_port->port, errno);

  /* Bind port to socket on any interface */
  memset(&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servaddr.sin_port = htons(listen_port->port);
  if (bind(listen_port->fd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_listener_socket_establish: bind to socket for tcp port %d failed (errno %d)\n", listen_port->port, errno);
    goto cle_fail;
  }

  /* Enter passive mode */
  if (listen(listen_port->fd, MAX_TCP_CONNECTION_COUNT) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_listener_establish: listen to socket for tcp port %d failed (errno %d)\n", listen_port->port, errno);
    goto cle_fail;
  }

  /* Success */
  return true;

cle_fail:
  if (listen_port->fd >= 0)
  {
    close(listen_port->fd);
    listen_port->fd = -1;
  }
  return false;
}


/** Establish a listening UDS socket for a socket file */
static bool comms_listener_establish_uds(struct s_comms_listen_port *listen_port)
{
  struct sockaddr_un servaddr;
  struct stat statbuf;

  listen_port->fd = -1;

  /* First check that we are happy about using this UDS file, and delete if necessary */
  if (stat(listen_port->uds_file, &statbuf) < 0)
  {
    if (ENOENT != errno)
    {
      rdebug_printf(LOG_ERR, "comms_listener_establish_uds: stat of uds file \"%s\" failed (errno %d), not establishing socket\n", listen_port->uds_file, errno);
      goto cleu_fail;
    }
  }
  else
  {
    if (!S_ISSOCK(statbuf.st_mode))
    {
      rdebug_printf(LOG_ERR, "comms_listener_establish_uds: uds file \"%s\" is not a socket, refusing to delete, not establishing socket\n", listen_port->uds_file);
      goto cleu_fail;
    }
    unlink(listen_port->uds_file);
  }

  /* Create the socket */
  listen_port->fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (listen_port->fd < 0)
  {
    rdebug_printf(LOG_ERR, "comms_listener_establish_uds: create listen socket for uds file \"%s\" failed (errno %d)\n", listen_port->uds_file, errno);
    goto cleu_fail;
  }
  if (fcntl(listen_port->fd, F_SETFD, FD_CLOEXEC) < 0)
    rdebug_printf(LOG_ERR, "comms_listener_establish_uds: setting FD_CLOEXEC on socket for uds file \"%s\" failed (errno %d)\n", listen_port->uds_file, errno);

  /* Bind uds socket to file. TODO support abstract ports */
  memset(&servaddr, 0, sizeof(servaddr));
  servaddr.sun_family = AF_UNIX;
  snprintf(servaddr.sun_path, sizeof(servaddr.sun_path), "%s", listen_port->uds_file);
  if (bind(listen_port->fd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_listener_socket_establish: bind to socket for uds file \"%s\" failed (errno %d)\n", listen_port->uds_file, errno);
    goto cleu_fail;
  }

  /* Enter passive mode */
  if (listen(listen_port->fd, MAX_UDS_CONNECTION_COUNT) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_listener_establish: listen to socket for uds file \"%s\" failed (errno %d)\n", listen_port->uds_file, errno);
    goto cleu_fail;
  }

  /* Success */
  return true;

cleu_fail:
  if (listen_port->fd >= 0)
  {
    close(listen_port->fd);
    listen_port->fd = -1;
  }
  return false;
}


/** Called by comms_listener_task to see if the database has changed, and we need to change our listening sockets */
static void comms_listener_reconfigure_tcp(struct s_comms_listener *cl)
{
  enum eSerialPort i;
  size_t j;
  int port;

  for (i = SERPORT_TCP_FIRST, j = 0; i < SERPORT_TCP_FINAL && j < MAX_COMMS_TCP_LISTENERS; i++, j++)
  {
    port = db_get_serial_tcpport(i);
    if (port > 0)
    {
      if (port != cl->comms_listener_ports[j].port)
      {
        /* Port changed, close existing port if open, and establish a new listening socket */
        if (cl->comms_listener_ports[j].fd >= 0)
        {
          close(cl->comms_listener_ports[j].fd);
          cl->comms_listener_ports[j].fd = -1;
        }

        WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_listener_reconfigure_tcp: port %d(%s) has changed tcp port from %d to %d\n",
                                          i, get_serial_port_as_string(i), cl->comms_listener_ports[j].port, port);
        cl->comms_listener_ports[j].port = port;
        cl->comms_listener_ports[j].serialport = i;
        comms_listener_establish_tcp(&cl->comms_listener_ports[j]);
      }
      /* else no port change, nothing to do */
    }
    else
    {
      /* TCP/UDP port number is 0 for this port, make sure it's closed */
      if (cl->comms_listener_ports[j].fd >= 0)
      {
        close(cl->comms_listener_ports[j].fd);
        cl->comms_listener_ports[j].fd = -1;
      }
    }
  }
  for (; j < MAX_COMMS_TCP_LISTENERS; j++)
  {
    if (cl->comms_listener_ports[j].fd >= 0)
    {
      rdebug_printf(LOG_ERR, "comms_listener_reconfigure_tcp: closed one\n");
      close(cl->comms_listener_ports[j].fd);
      cl->comms_listener_ports[j].fd = -1;
    }
    cl->comms_listener_ports[j].port = -1;
    cl->comms_listener_ports[j].serialport = SERPORT_UNKNOWN;
  }
}


/** Called by comms_listener_task to establish our listening unix domain sockets (these don't change) */
static void comms_listener_configure_uds(struct s_comms_listener *cl)
{
  enum eSerialPort i;

  for (i = SERPORT_UDS_FIRST; i < SERPORT_UDS_FINAL; i++)
  {
    if (NULL != cl->comms_listener_ports[COMMS_UDS_LISTENER_IDX(i)].uds_file)
    {
      WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_listener_configure_uds: establish uds listening port %d(%s) on file \"%s\"\n",
                                        i, get_serial_port_as_string(i), cl->comms_listener_ports[COMMS_UDS_LISTENER_IDX(i)].uds_file);
      comms_listener_establish_uds(&cl->comms_listener_ports[COMMS_UDS_LISTENER_IDX(i)]);
      cl->comms_listener_ports[COMMS_UDS_LISTENER_IDX(i)].serialport = i;
    }
  }
}


/** Called by comms_listener_task to accept a new connection from either TCP or UDS, and fire up another comms task to handle it */
static bool comms_listener_tcp_uds_accept(struct s_comms_listen_port *const listen_port)
{
  int fd = -1;
  struct sockaddr addr;
  socklen_t len = sizeof(addr);
  struct t_io_dev *dev = NULL;
  char portstr[40];

  if (NULL != listen_port->uds_file)
    snprintf(portstr, sizeof(portstr), "uds file \"%s\"", listen_port->uds_file);
  else
    snprintf(portstr, sizeof(portstr), "tcp port %d", listen_port->port);

  /* Accept new connection */
  fd = accept(listen_port->fd, &addr, &len);
  if (fd < 0)
  {
    rdebug_printf(LOG_ERR, "comms_listener_tcp_uds_accept: %s, new sock fd %d: error accepting new connection (errno %d)\n",
                  portstr, fd, errno);
    goto cla_fail;
  }
  if (fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_listener_tcp_uds_accept: %s, new sock fd %d: error setting FD_CLOEXEC on new connection\n",
                  portstr, fd, errno);
    goto cla_fail;
  }

  /* Allocate and setup rin device */
  dev = calloc(1, sizeof(struct t_io_dev));
  if (NULL == dev)
  {
    rdebug_printf(LOG_ERR, "comms_listener_tcp_uds_accept: %s, new sock fd %d: failed to allocate rin io device\n",
                  portstr, fd);
    goto cla_fail;
  }
  if (rio_create_device_sio_established_localskt(dev, fd) == RIO_ERROR)
  {
    rdebug_printf(LOG_ERR, "comms_listener_tcp_uds_accept: %s, new sock fd %d: failed to initialise rin io dev\n",
                  portstr, fd);
    goto cla_fail;
  }

  /* Create thread to handle it */
  if (comms_create_thread_for_serial_port(dev, listen_port->serialport, NULL) == RIO_ERROR)
  {
    rdebug_printf(LOG_ERR, "comms_listener_tcp_uds_accept: %s, new sock fd %d: failed to create thread\n",
                  portstr, fd);
    goto cla_fail;
  }

  WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_listener_tcp_uds_accept: %s, new sock fd %d: accepted new connection\n",
                                    listen_port->port);
  return true;

cla_fail:
  if (NULL != dev)
  {
    if (0 != (dev->status & O_INITIALISED))
    {
      rdestroy_io(dev);
    }
    free(dev);
    dev = NULL;
  }
  if (fd >= 0)
  {
    close(fd);
    fd = -1;
  }
  return false;
}


/** Shut down all listening connections */
static void comms_listener_shutdown(struct s_comms_listener *cl)
{
  size_t i;

  for (i = 0; i < MAX_COMMS_LISTENERS; i++)
  {
    if (cl->comms_listener_ports[i].fd >= 0)
    {
      close(cl->comms_listener_ports[i].fd);
      cl->comms_listener_ports[i].fd = -1;
      if (NULL != cl->comms_listener_ports[i].uds_file)
        unlink(cl->comms_listener_ports[i].uds_file);
    }
    cl->comms_listener_ports[i].port = -1;
    cl->comms_listener_ports[i].serialport = SERPORT_UNKNOWN;
  }
}


/** Handle any commands on the control channel */
static bool comms_listener_cmd_handler(struct s_comms_listener *cl, bool *quit)
{
  char cmd;

  if (read(cl->activate_read_fd, &cmd, sizeof(cmd)) != sizeof(cmd))
  {
    rdebug_printf(LOG_ERR, "comms_listener_cmd_handler: read() failed\n");
    return false;
  }

  switch (cmd)
  {
    case COMMS_LISTENER_CMD_QUIT:
      WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_listener_cmd_handler: received quit command\n");
      *quit = true;
      break;
    case COMMS_LISTENER_CMD_RECONFIGURE:
      WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_listener_cmd_handler: reconfiguring\n");
      comms_listener_reconfigure_tcp(cl);
      break;
    case COMMS_LISTENER_CMD_CLEAN:
      WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_listener_cmd_handler: cleaning tasks that have quit\n");
      comms_task_list_clean();
      break;
    default:
      rdebug_printf(LOG_ERR, "comms_listener_cmd_handler: unhandled command %u\n", (unsigned) cmd);
      break;
  }
  return true;
}


/**
 * A listening thread that just sits and spins waiting for connections and responding to
 * them by creating new serial devices with their own threads.
 */
static void *comms_listener_task(void *args)
{
  struct s_comms_listener *const cl = (struct s_comms_listener *) args;
  size_t i;
  fd_set readfds;
  int maxfd;
  bool quit;

  WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_listener_task: started\n");

  /* Establish listening UDS sockets */
  comms_listener_configure_uds(cl);

  /* Establish listening TCP sockets */
  comms_listener_reconfigure_tcp(cl);

  /* Hang around forever waiting for and dispatching connections */
  for (quit = false; !quit; )
  {
    /* Select on our control pipe, and any listening sockets */
    FD_ZERO(&readfds);
    FD_SET(cl->activate_read_fd, &readfds);
    maxfd = cl->activate_read_fd;
    for (i = 0; i < MAX_COMMS_LISTENERS; i++)
    {
      if (cl->comms_listener_ports[i].fd >= 0)
      {
        FD_SET(cl->comms_listener_ports[i].fd, &readfds);
        if (cl->comms_listener_ports[i].fd > maxfd)
          maxfd = cl->comms_listener_ports[i].fd;
      }
    }

    /* Wait for incoming connection, or something from our control channel */
    if (select(maxfd + 1, &readfds, NULL, NULL, NULL) < 0)
    {
      rdebug_printf(LOG_ERR, "comms_listener_task: select() failed\n");
      break;
    }

    /* Deal with incoming connections */
    for (i = 0; i < MAX_COMMS_LISTENERS; i++)
    {
      if (cl->comms_listener_ports[i].fd >= 0 && FD_ISSET(cl->comms_listener_ports[i].fd, &readfds))
      {
        (void) comms_listener_tcp_uds_accept(&cl->comms_listener_ports[i]);
      }
    }

    /* Control channel */
    if (FD_ISSET(cl->activate_read_fd, &readfds))
    {
      if (!comms_listener_cmd_handler(cl, &quit))
        break;
    }
  }

  comms_listener_shutdown(cl);
  WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_listener_task: finished\n");
  return NULL;
}


/** Signal the comms listener, so select() returns and it can either clean up a connection, or quit */
static void signal_comms_listener(const char cmd)
{
  if (comms_listener.activate_write_fd >= 0)
    write(comms_listener.activate_write_fd, &cmd, sizeof(cmd));
}


/** Cleanup after a comms thread has quit (gets rid of zombies) */
void comms_task_listen_cleanup(void)
{
  WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_task_listen_cleanup: cleaning up closing comms thread\n");
  signal_comms_listener(COMMS_LISTENER_CMD_CLEAN);
}


/** Reload configuration for the tcp port listening thread */
void comms_task_listen_reconfigure(void)
{
  WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_task_listen_reconfigure: re-initialising tcp port listeners\n");
  signal_comms_listener(COMMS_LISTENER_CMD_RECONFIGURE);
}


/**
 * Create the comms listen task that accepts incoming socket connections (TCP and UDS) and creates the appropriate handler task
 *
 * Unix domain sockets are specified here, and don't change
 * TCP ports are specified in the database and will change when @ref comms_task_listen_reconfigure() is called
 */
bool comms_task_listen_init(const char *uds_net_file, const char *uds_auto_file)
{
  int pipe_fds[2] = { -1, -1 };
  enum eSerialPort i;
  size_t j;

  memset(&comms_listener, 0, sizeof(comms_listener));
  comms_listener.activate_read_fd = -1;
  comms_listener.activate_write_fd = -1;

  /* Initialise TCP listener port structure, the task fills this out later from database settings */
  for (i = 0; i < MAX_COMMS_TCP_LISTENERS; i++)
  {
    comms_listener.comms_listener_ports[i].uds_file = NULL;
    comms_listener.comms_listener_ports[i].port = -1;
    comms_listener.comms_listener_ports[i].fd = -1;
    comms_listener.comms_listener_ports[i].serialport = SERPORT_UNKNOWN;
  }

  /* Initialise UDS listener port structure, the task fills this out later */
  comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSNET)].uds_file = strdup(uds_net_file);
  if (NULL == comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSNET)].uds_file)
  {
    rdebug_printf(LOG_ERR, "comms_task_listen_init: strdup(uds_net_file) failed\n");
    goto ctli_fail;
  }
  comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSNET)].port = -1;
  comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSNET)].fd = -1;
  comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSNET)].serialport = SERPORT_UNKNOWN;

  comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSAUTO)].uds_file = strdup(uds_auto_file);
  if (NULL == comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSAUTO)].uds_file)
  {
    rdebug_printf(LOG_ERR, "comms_task_listen_init: strdup(uds_auto_file) failed\n");
    goto ctli_fail;
  }
  comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSAUTO)].port = -1;
  comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSAUTO)].fd = -1;
  comms_listener.comms_listener_ports[COMMS_UDS_LISTENER_IDX(SERPORT_UDSAUTO)].serialport = SERPORT_UNKNOWN;

  /* Create a pipe to write to, to shut down the task */
  if (pipe(pipe_fds) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_task_listen_init: create control pipe failed\n");
    goto ctli_fail;
  }

  /* Keep fds for shutting down task */
  comms_listener.activate_read_fd = pipe_fds[0];
  comms_listener.activate_write_fd = pipe_fds[1];
  if (fcntl(comms_listener.activate_read_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_task_listen_init: error setting FD_CLOEXEC on read end of control pipe (errno %d)\n", errno);
    goto ctli_fail;
  }
  if (fcntl(comms_listener.activate_write_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_task_listen_init: error setting FD_CLOEXEC on write end of control pipe (errno %d)\n", errno);
    goto ctli_fail;
  }

  /* Create the thread */
  if (!rin_thread_create(&comms_listener.pthread, &comms_listener_task, (void *) &comms_listener, true))
  {
    rdebug_printf(LOG_ERR, "comms_task_listen_init: failed to create listening thread\n");
    goto ctli_fail;
  }

  /* Success */
  return true;

ctli_fail:
  if (comms_listener.activate_read_fd >= 0)
  {
    close(comms_listener.activate_read_fd);
    comms_listener.activate_read_fd = -1;
  }
  if (comms_listener.activate_write_fd >= 0)
  {
    close(comms_listener.activate_write_fd);
    comms_listener.activate_write_fd = -1;
  }

  /* Free up any uds_file names */
  for (j = 0; j < MAX_COMMS_LISTENERS; j++)
  {
    if (NULL != comms_listener.comms_listener_ports[j].uds_file)
    {
      free(comms_listener.comms_listener_ports[j].uds_file);
      comms_listener.comms_listener_ports[j].uds_file = NULL;
    }
  }

  return false;
}


/** Destroy the listening thread */
void comms_task_listen_destroy(void)
{
  size_t i;

  WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "comms_task_listen_destroy: asking listener thread to quit\n");
  signal_comms_listener(COMMS_LISTENER_CMD_QUIT);
  (void) pthread_join(comms_listener.pthread, NULL);

  /* Close the pipes */
  if (comms_listener.activate_read_fd >= 0)
  {
    close(comms_listener.activate_read_fd);
    comms_listener.activate_read_fd = -1;
  }
  if (comms_listener.activate_write_fd >= 0)
  {
    close(comms_listener.activate_write_fd);
    comms_listener.activate_write_fd = -1;
  }

  /* Free up any uds_file names */
  for (i = 0; i < MAX_COMMS_LISTENERS; i++)
  {
    if (NULL != comms_listener.comms_listener_ports[i].uds_file)
    {
      free(comms_listener.comms_listener_ports[i].uds_file);
      comms_listener.comms_listener_ports[i].uds_file = NULL;
    }
  }

  WE_COMMS_LISTEN_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_task_listen_destroy: listener thread has quit\n");
}
