/*
 * comms_task_udp.c
 *
 *  Created on: Sep 28, 2017
 *      Author: ian
 */
#include <stdbool.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <rinhw/rio.h>
#include <rinhw/sio_local_established_dgram.h>
#include "we_debug.h"
#include "serial_ports.h"
#include "comms_task.h"
#include "db_serial.h"
#include "db_serial_ip.h"

#include "comms_task_udp.h"


/** Fill out a struct sockaddr_in */
static void sockaddr_in_fillout(struct sockaddr_in *sa, sa_family_t family, in_addr_t addr, in_port_t port, bool netendian)
{
  memset(sa, 0, sizeof(*sa));
  sa->sin_family = family;
  sa->sin_addr.s_addr = (netendian) ? htonl(addr) : addr;
  sa->sin_port = (netendian) ? htons(port) : port;
}


/** Establish a listening socket for a udp port. port is host endian */
static bool comms_task_udp_socket_establish(int *fd, int port)
{
  int flag;
  struct sockaddr_in servaddr;

  *fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (*fd < 0)
  {
    rdebug_printf(LOG_ERR, "comms_task_udp_socket_establish: create listen socket for udp port %d failed (errno %d)\n", port, errno);
    goto cle_fail;
  }
  if (fcntl(*fd, F_SETFD, FD_CLOEXEC) < 0)
    rdebug_printf(LOG_ERR, "comms_task_udp_socket_establish: setting FD_CLOEXEC on socket for udp port %d failed (errno %d)\n", port, errno);

  /* Set the reuse flag on the socket before binding */
  flag = 1;
  if (setsockopt(*fd, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(flag)) < 0)
    rdebug_printf(LOG_WARNING, "comms_task_udp_socket_establish: setting SO_REUSEADDR on socket for udp port %d failed (errno %d)\n", port, errno);

  /* Enable broadcast on this socket */
  flag = 1;
  if (setsockopt(*fd, SOL_SOCKET, SO_BROADCAST, &flag, sizeof(flag)) < 0)
    rdebug_printf(LOG_WARNING, "comms_task_udp_socket_establish: setting SO_BROADCAST on socket for udp port %d failed (errno %d)\n", port, errno);

  /* Bind port to socket on any interface */
  sockaddr_in_fillout(&servaddr, AF_INET, INADDR_ANY, port, true);
  if (bind(*fd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0)
  {
    rdebug_printf(LOG_ERR, "comms_task_udp_socket_establish: bind to socket for udp port %d failed (errno %d)\n", port, errno);
    goto cle_fail;
  }

  /* Success */
  return true;

cle_fail:
  if (*fd >= 0)
  {
    close(*fd);
    *fd = -1;
  }
  return false;
}


/** Establish a new UDP socket, and create rindevice to handle it. dev must already be allocated. localport and udpdest is host endian */
static bool comms_task_udp_make_new_rio(struct t_io_dev *dev, enum eSerialPort port, int localport, const struct sockaddr_in *udpdest)
{
  int fd = -1;

  /* Create a new socket, rindevice and task to talk to this port */
  if (!comms_task_udp_socket_establish(&fd, localport))
  {
    rdebug_printf(LOG_ERR, "comms_task_udp_create: error creating socket for %u(%s) local udp port %d\n",
                  port, get_serial_port_as_string(port), localport);
    goto ctumnr_fail;
  }

  /* Create a new rin UDP serial device to handle it */
  if (rio_create_device_sio_established_dgram_unbuffered(dev, fd, udpdest) == RIO_ERROR)
  {
    rdebug_printf(LOG_ERR, "comms_task_udp_create: error initialising %u(%s) rin device\n", port, get_serial_port_as_string(port));
    goto ctumnr_fail;
  }

  /* Success */
  return true;

ctumnr_fail:
  if (0 != (dev->status & O_INITIALISED))
  {
    rdestroy_io(dev);
  }
  if (fd >= 0)
  {
    close(fd);
    fd = -1;
  }
  return false;
}


/** Create new UDP connection, and fire up another comms task to handle it. localport and udpdest in host endian */
static bool comms_task_udp_create(enum eSerialPort port, int localport, const struct sockaddr_in *udpdest)
{
  int fd = -1;
  struct t_io_dev *dev = NULL;
  struct s_comms_udp_info udpinfo;
  char ip_str[INET_ADDRSTRLEN];

  /* Allocate and setup rin device */
  dev = calloc(1, sizeof(struct t_io_dev));
  if (NULL == dev)
  {
    rdebug_printf(LOG_ERR, "comms_task_udp_create: error allocating memory for %u(%s) rin device\n", port, get_serial_port_as_string(port));
    goto cla_fail;
  }

  if (!comms_task_udp_make_new_rio(dev, port, localport, udpdest))
    goto cla_fail;

  /* Create thread to handle it */
  if (comms_is_udp_in_connection(port))
  {
    udpinfo.localport = localport;
    sockaddr_in_fillout(&udpinfo.dest, AF_INET, 0, 0, false);
  }
  else
  {
    udpinfo.localport = 0;
    udpinfo.dest = *udpdest;
  }
  if (comms_create_thread_for_serial_port(dev, port, &udpinfo) == RIO_ERROR)
  {
    rdebug_printf(LOG_ERR, "comms_task_udp_create: error creating thread for %u(%s)\n", port, get_serial_port_as_string(port));
    goto cla_fail;
  }

  if (comms_is_udp_in_connection(port))
  {
    WE_COMMS_UDP_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_task_udp_create: created new incoming udp connection/task %u(%s) on udp port %d\n",
                                   port, get_serial_port_as_string(port), localport);
  }
  else
  {
    (void) net_ip_inet_ntop_host_afinet(&udpinfo.dest.sin_addr, ip_str, INET_ADDRSTRLEN);
    WE_COMMS_UDP_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_task_udp_create: created new outgoing udp connection/task %u(%s) to %s:%d\n",
                                   port, get_serial_port_as_string(port), ip_str, udpinfo.dest.sin_port);
  }

  return true;

cla_fail:
  if (NULL != dev)
  {
    if (0 != (dev->status & O_INITIALISED))
    {
      rdestroy_io(dev);
    }
    free(dev);
    dev = NULL;
  }
  if (fd >= 0)
  {
    close(fd);
    fd = -1;
  }
  return false;
}


/** Reconfigure an already running UDP task. Returns false if the task should quit. udpinfo in host endian. */
bool comms_task_udp_running_reconfigure(enum eSerialPort port, struct t_io_dev *rindev, struct s_comms_udp_info *udpinfo)
{
  enum eSerialPortType db_type;
  int db_udpport;
  struct in_addr db_udpdestip;
  struct sockaddr_in udpdest;
  char oldip_str[INET_ADDRSTRLEN];
  char newip_str[INET_ADDRSTRLEN];

  db_type = db_get_serial_type(port);
  db_udpport = db_get_serial_tcpport(port);
  db_get_serial_dest_ipaddress(port, &db_udpdestip);

  /* Do we need to quit? */
  if (comms_is_udp_in_connection(port) &&
      (DB_SERIAL_TYPE_OFF == db_type ||
       0 == db_udpport))
  {
    /* Incoming UDP must have a type (not off) and valid local UDP port */
    WE_COMMS_UDP_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_task_udp_running_reconfigure: incoming udp task %u(%s) to quit because type is off, or udp port is 0\n",
                                   port, get_serial_port_as_string(port));
    return false;
  }
  else if (comms_is_udp_out_connection(port) &&
           (DB_SERIAL_TYPE_OFF == db_type ||
            0 == db_udpport ||
            0 == db_udpdestip.s_addr))
  {
    /* Outgoing UDP must have a type (not off) and valid destination IP and UDP port */
    WE_COMMS_UDP_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_task_udp_running_reconfigure: outgoing udp task %u(%s) to quit because type is off, or udp dest port is 0, or udp dest ip is 0.0.0.0\n",
                                   port, get_serial_port_as_string(port));
    return false;
  }

  /* Do we need to change ports/destinations? */
  if (comms_is_udp_in_connection(port) && udpinfo->localport != db_udpport)
  {
    /* Source port has changed, re-establish socket */
    rdestroy_io(rindev);
    if (!comms_task_udp_make_new_rio(rindev, port, db_udpport, NULL))
      return false;
    WE_COMMS_UDP_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_task_udp_running_reconfigure: incoming udp task %u(%s) port reconfigure from %d to %d\n",
                                   port, get_serial_port_as_string(port), udpinfo->localport, db_udpport);
    udpinfo->localport = db_udpport;
    sockaddr_in_fillout(&udpinfo->dest, AF_INET, 0, 0, false);
  }
  if (comms_is_udp_out_connection(port) && (udpinfo->dest.sin_port != db_udpport || udpinfo->dest.sin_addr.s_addr != db_udpdestip.s_addr))
  {
    /* Destination has changed, re-establish socket */
    rdestroy_io(rindev);
    sockaddr_in_fillout(&udpdest, AF_INET, db_udpdestip.s_addr, db_udpport, false);
    if (!comms_task_udp_make_new_rio(rindev, port, db_udpport, &udpdest)) /* TODO for the time being, lets bind to the same port number on our end */
      return false;

    (void) net_ip_inet_ntop_host_afinet(&udpinfo->dest.sin_addr, oldip_str, INET_ADDRSTRLEN);
    (void) net_ip_inet_ntop_host_afinet(&udpdest.sin_addr, newip_str, INET_ADDRSTRLEN);
    WE_COMMS_UDP_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "comms_task_udp_running_reconfigure: outgoing udp task %u(%s) destination change from %s:%d to %s:%d\n",
                                   port, get_serial_port_as_string(port), oldip_str, udpinfo->dest.sin_port, newip_str, db_udpport);

    udpinfo->localport = 0;
    sockaddr_in_fillout(&udpinfo->dest, AF_INET, db_udpdestip.s_addr, db_udpport, false);
  }

  /* Open serial port, if not already open */
  if (rindev->status & O_OPEN)
  {
    ropen(rindev, 0);
  }

  /* Don't need to worry about type, because normal comms_type_reconfigure() deals with that */
  return true;
}


/**
 * Reconfigure the UDP tasks from the database settings.
 *
 * This really only starts tasks if they have been enabled.
 * UDP tasks are responsible for exiting if they have been disabled,
 * and reconfiguring their ports if a change has been made
 */
void comms_task_udp_reconfigure(void)
{
  enum eSerialPortType task_type;
  enum eSerialPortType db_type;
  enum eSerialPort port;
  int db_udpport;
  struct in_addr db_udpdestip;
  struct sockaddr_in udpdest;

  for (port = SERPORT_UDP_FIRST; port < SERPORT_UDP_FINAL; port++)
  {
    task_type = comms_task_list_get_port_type(port);
    db_type = db_get_serial_type(port);
    db_udpport = db_get_serial_tcpport(port);
    db_get_serial_dest_ipaddress(port, &db_udpdestip);

    /* Do we need to start a UDP in connection? */
    if (comms_is_udp_in_connection(port) &&
        NUM_DB_SERIAL_TYPES == task_type &&
        DB_SERIAL_TYPE_OFF != db_type &&
        0 != db_udpport)
    {
      /* Yes, UDP in port: create udp task listening on udpport */
      (void) comms_task_udp_create(port, db_udpport, NULL);
    }

    /* Do we need to start a UDP out connection? */
    if (comms_is_udp_out_connection(port) &&
        NUM_DB_SERIAL_TYPES == task_type &&
        DB_SERIAL_TYPE_OFF != db_type &&
        0 != db_udpport &&
        0 != db_udpdestip.s_addr)
    {
      /* Yes, UDP out port: create udp socket and task, with fixed UDP destination */
      sockaddr_in_fillout(&udpdest, AF_INET, db_udpdestip.s_addr, db_udpport, false);
      (void) comms_task_udp_create(port, db_udpport, &udpdest); /* TODO for the time being, lets bind to the same port number on our end */
    }
  }
}


/** Init the UDP tasks */
void comms_task_udp_init(void)
{
  comms_task_udp_reconfigure();
}
