#include <setjmp.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <rinhw/rstdio.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "asciidef.h"
#include "meas.h"
#include "database.h"
#include "apperror.h"
#include "version.h"
#include "iface.h"
#include "system.h"
#include "editor.h"
#include "kbd_5000.h"
#include "setpoint.h"
#include "comms_print.h"
#include "access.h"
#include "sysclock.h"
#include "buzz5000.h"
#include "modbus_old.h"
#include "display.h"
#include "remote_cal.h"
#include "comms_errors.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "app_db_runtime_vars.h"

#include "comms_util.h"


/** MSV weight readings are upto 8 digits */
#define MSV_RDG_MAX_WIDTH 8


/** Returns true if ch is 0..9 A..F */
bool is_hex(char ch)
{
  if (ch >= '0' && ch <= '9')
    return true;
  if (ch >= 'A' && ch <= 'F')
    return true;
  if (ch >= 'a' && ch <= 'f')
    return true;
  return false;
}


/** Convert nibble b to '0..9A..F' */
char to_hex(uint8_t b)
{
  if (b < 10)
    return (char) ('0' + b);
  else
    return (char) ('A' + b - 10);
}


/** Convert single hex character to nibble */
uint8_t to_bin(char c)
{
  if ((c >= '0') && (c <= '9'))
    return (uint8_t) (c - '0');
  else if ((c >= 'A') && (c <= 'F'))
    return (uint8_t) (10 + c - 'A');
  else if ((c >= 'a') && (c <= 'f'))
    return (uint8_t) (10 + c - 'a');
  else
    return (0);
}


/** Special version of vprintf for comms messages */
int comms_net_rvfprintf(struct s_comms_control *comms_ctrl, const char *format, va_list arglist)
{
  int length = 0;

  if (comms_ctrl->reply_active)
  {
    length = comms_send_buffer_vprintf(comms_ctrl->csb, format, arglist);
  }
  return length;
}


/** Special version of printf for comms messages */
int comms_net_rfprintf(struct s_comms_control *comms_ctrl, const char *format, ...)
{
  va_list arglist;
  int length = 0;

  va_start(arglist, format);
  length = comms_net_rvfprintf(comms_ctrl, format, arglist);
  va_end(arglist);
  return length;
}


/**
 * Send a number with decimal point to comms_ctrl, or print to a string
 *
 * comms_ctrl is the comms net port to send the formatted number to. If NULL, printed to str instead
 * str is a string that the number will be printed to if serdev is NULL
 * str_max_len is the maximum length of str. Ignored if serdev is used
 * rdg is the number to be formatted and sent
 * dp is the decimal point position
 * maxlen is the desired length of the string
 * pad_spaces_not_zeros selects leading space padding (true), or leading '0' padding (false)
 * zeropadwidth is the minimum zero pad width. Used for NTEP where a reading of 0 must be padded to the resolution width, otherwise 0
 *
 * @return number of characters sent
 */
static int comms_internal_convert_num_with_dp_pad(struct s_comms_control *comms_ctrl, char *str, size_t str_max_len,
                                                  long rdg, int dp, int maxlen, bool pad_spaces_not_zeros, int zeropadwidth)
{
  int divisor;
  int i;
  long remainder;
  long quotient;
  char sign;
  const char *fmtstr;
  int length;
  int rdgwidth;
  int zeropad_width;
  int spacepad_width;
  long temp;

  /* Work out sign character */
  if (rdg < 0)
  {
    rdg = -rdg;
    sign = '-';
  }
  else
  {
    sign = ' ';
  }

  /* Adjust length for sign character */
  maxlen--;

  /* Work out decimal point divisor */
  for (i = 0, divisor = 1; i < dp; i++)
  {
    divisor *= 10;
  }

  /* Work out number split at decimal point */
  quotient = rdg / divisor;
  remainder = rdg % divisor;

  /* Print the number with or without decimal point */
  if (dp > 0)
  {
    /* With decimal point */
    if (pad_spaces_not_zeros)
    {
      /* Pad with spaces (and maybe some zeros based on zerowidth) */

      /* Work out width of number */
      for (rdgwidth = 0, temp = (quotient != 0) ? quotient : 1; temp > 0; rdgwidth++, temp /= 10)
        ;

      /* Work out split between spaces and zeros */
      zeropad_width = (zeropadwidth > rdgwidth) ? zeropadwidth : rdgwidth;
      spacepad_width = maxlen - dp - 1 - zeropad_width;
    }
    else
    {
      /* Pad with all zeros */
      zeropad_width = maxlen - dp - 1;
      spacepad_width = 0;
    }
    fmtstr = "%c%*s%*ld.%0*lu";

    /* Print it */
    if (NULL != comms_ctrl)
      length = comms_net_rfprintf(comms_ctrl, fmtstr, sign, spacepad_width, "", zeropad_width, quotient, dp, (unsigned long) remainder);
    else if (NULL != str)
      length = snprintf(str, str_max_len, fmtstr, sign, spacepad_width, "", zeropad_width, quotient, dp, (unsigned long) remainder);
    else
      length = 0;
  }
  else
  {
    /* No decimal point */
    if (pad_spaces_not_zeros)
    {
      /* Pad with spaces (and maybe some zeros based on zerowidth) */

      /* Work out width of number */
      for (rdgwidth = 0, temp = (quotient != 0) ? quotient : 1; temp > 0; rdgwidth++, temp /= 10)
        ;

      /* Work out split between spaces and zeros */
      zeropad_width = (zeropadwidth > rdgwidth) ? zeropadwidth : rdgwidth;
      spacepad_width = maxlen - zeropad_width;
    }
    else
    {
      /* Pad with all zeros */
      zeropad_width = maxlen;
      spacepad_width = 0;
    }
    fmtstr = "%c%*s%0*ld";

    /* Print it */
    if (NULL != comms_ctrl)
      length = comms_net_rfprintf(comms_ctrl, fmtstr, sign, spacepad_width, "", zeropad_width, (unsigned long) quotient);
    else if (NULL != str)
      length = snprintf(str, str_max_len, fmtstr, sign, spacepad_width, "", zeropad_width, (unsigned long) quotient);
    else
      length = 0;
  }
  return length;
}


/**
 * Send a number with decimal point over the comms
 *
 * comms_ctrl is the comms net port to send the formatted number to
 * rdg is the number to be formatted and sent
 * dp is the decimal point position
 * maxlen is the desired length of the string
 * pad_spaces_not_zeros selects leading space padding (true), or leading '0' padding (false)
 * zeropadwidth is the minimum zero pad width. Used for NTEP where a reading of 0 must be padded to the resolution width, otherwise 0
 *
 * @return number of characters sent
 */
int comms_send_num_with_dp_pad(struct s_comms_control *comms_ctrl, long rdg, int dp, int maxlen, bool pad_spaces_not_zeros, int zeropadwidth)
{
  return comms_internal_convert_num_with_dp_pad(comms_ctrl, NULL, 0, rdg, dp, maxlen, pad_spaces_not_zeros, zeropadwidth);
}


/**
 * Convert a number with decimal point to a string
 *
 * str is a string that the number will be printed to
 * str_max_len is the maximum length of str
 * rdg is the number to be formatted and sent
 * dp is the decimal point position
 * maxlen is the desired length of the string
 * pad_spaces_not_zeros selects leading space padding (true), or leading '0' padding (false)
 * zeropadwidth is the minimum zero pad width. Used for NTEP where a reading of 0 must be padded to the resolution width, otherwise 0
 *
 * @return number of characters sent
 */
int comms_convert_num_with_dp_pad(char *str, size_t str_max_len, long rdg, int dp, int maxlen, bool pad_spaces_not_zeros, int zeropadwidth)
{
  return comms_internal_convert_num_with_dp_pad(NULL, str, str_max_len, rdg, dp, maxlen, pad_spaces_not_zeros, zeropadwidth);
}


/** Send a number with decimal point over the comms */
void comms_send_num_with_dp(struct s_comms_control *comms_ctrl, const AdcRdg *rdg, int maxlen)
{
  comms_send_num_with_dp_pad(comms_ctrl, rdg->rdg, rdg->info.unitdp.bit.dp, maxlen, false, 0);
}


/** Set various members of adcrdg for a non-weight reading */
static void comms_set_adcrdg_nonweight(AdcRdg *adcrdg, int32_t rdg, uint8_t units, uint8_t dp)
{
  memset(adcrdg, 0, sizeof(AdcRdg));

  adcrdg->rdg = rdg;
  adcrdg->info.unitdp.bit.units = units;
  adcrdg->info.unitdp.bit.dp = dp;
}


/** Get the appropriate reading based on source for the MSV command */
static void comms_get_reading_for_msv(AdcRdg *rdg, enum eMSVSource source, bool *hide)
{
  /* Fill in some other status info */
  *hide = false;

  /* Figure out weight source */
  switch (source)
  {
    case MSV_SRC_DISPLAYED:
      *rdg = adcSuperStructure.adcDataWorking.dispRdg;
      break;
    case MSV_SRC_GROSS:
      *rdg = adcSuperStructure.adcDataWorking.grossRdg;
      break;
    case MSV_SRC_NET:
      *rdg = adcSuperStructure.adcDataWorking.netRdg;
      break;
    case MSV_SRC_TOTAL_NUMITEMS:
      comms_set_adcrdg_nonweight(rdg, statistics_get_num_items(), UNITS_NONE, 0);
      break;
    case MSV_SRC_TOTAL1:
    case MSV_SRC_TOTAL2:
      comms_set_adcrdg_nonweight(rdg, statistics_get_total(),
                                 adcSuperStructure.adcDataWorking.grossRdg.info.unitdp.bit.units,
                                 adcSuperStructure.adcDataWorking.grossRdg.info.unitdp.bit.dp);
      break;
    case MSV_SRC_PEAKMAX:
      *rdg = adcSuperStructure.adcDataWorking.peakMaxDispRdg;
      break;
    case MSV_SRC_LIVEWEIGH:
      *rdg = adcSuperStructure.adcHeldData.dispRdg;
      *hide = !adcSuperStructure.adcHeldData.dataValid;
      break;
    default:
      comms_set_adcrdg_nonweight(rdg, 0, UNITS_NONE, 0);
      break;
  }
}


/** Make normal status for comms_send_oiml (COF 9 & 10). Note this omits outputs 5..8. */
#define CSO_MAKE_NORMAL_STATUS(io, hide, adcrdg) \
  (((io << 4) & 0xF0) | \
   ((hide || adcrdg.info.status.bit.oload || adcrdg.info.status.bit.uload) ? 0x01 : 0x00) | \
   (!adcrdg.info.status.bit.motion ? 0x02 : 0x00) | \
   (adcrdg.info.status.bit.gross ? 0x04 : 0x00) | \
   (adcrdg.info.status.bit.range > 0 ? 0x08 : 0x00))


/** Make extended status for comms_send_oiml (COF 11). Note this omits outputs 5..8. */
#define CSO_MAKE_EXTENDED_STATUS(io, hide, adcrdg) \
    (CSO_MAKE_NORMAL_STATUS(io, hide, adcrdg) | \
     (adcrdg.info.status.bit.coz ? 0x0100 : 0x0000))


/** Send readings as per COF settings */
void comms_send_oiml(struct s_comms_control *comms_ctrl)
{
  uint8_t io_status;
  uint8_t norm_status;
  uint16_t ext_status;
  uint16_t split_status;
  bool hide;
  enum eCOFSetting cof;
  long bin_rdg;
  AdcRdg adcrdg;

  /* Get the reading for the weight source */
  comms_get_reading_for_msv(&adcrdg, comms_ctrl->msv_source, &hide);

  /* Work out status */
  io_status = access_get_io_states();
  norm_status = CSO_MAKE_NORMAL_STATUS(io_status, hide, adcrdg);
  ext_status = CSO_MAKE_EXTENDED_STATUS(io_status, hide, adcrdg);
  split_status = CSO_MAKE_EXTENDED_STATUS(0x00, hide, adcrdg);

  cof = db_app_user.cof_setting;
  hide = (hide ||
          (ADC_APPROVAL_MODE_OIML == adcSuperStructure.adcUserDatabase.approvalMode &&
           (APP_ERR_NONE != get_current_errors() || adcrdg.info.status.bit.oload || adcrdg.info.status.bit.uload)));
  if (hide)
  {
    if (adcrdg.info.status.bit.uload)
      bin_rdg = 0x00800000L;
    else
      bin_rdg = 0x007FFFFFL;

    /* Move down reading to fit in two bytes for 2 byte modes */
    if (COF_SETTING2_BINARY2_MSB_FIRST == cof || COF_SETTING6_BINARY2_LSB_FIRST == cof)
      bin_rdg >>= 8;
  }
  else
  {
    bin_rdg = adcrdg.rdg;
  }

  /* Print in format specified by COF */
  switch (cof)
  {
    /* Binary 4 byte - MSB first, LSB=0 */
    case COF_SETTING0_BINARY4_MSB_FIRST:
      comms_net_rfprintf(comms_ctrl, "%c%c%c%c",
                     (char) ((bin_rdg >> 16) & 0xFF),
                     (char) ((bin_rdg >> 8) & 0xFF),
                     (char) ((bin_rdg >> 0) & 0xFF),
                     '\0'
                    );
      break;

    /* Binary 2 byte - MSB then LSB */
    case COF_SETTING2_BINARY2_MSB_FIRST:
      comms_net_rfprintf(comms_ctrl, "%c%c",
                     (char) ((bin_rdg >> 8) & 0xFF),
                     (char) ((bin_rdg >> 0) & 0xFF)
                    );
      break;

    /* Normal ASCII, weight only */
    case COF_SETTING1_ASCII_WEIGHT:
    case COF_SETTING3_ASCII_WEIGHT:
      comms_send_num_with_dp(comms_ctrl, &adcrdg, MSV_RDG_MAX_WIDTH);
      comms_net_rfprintf(comms_ctrl, COMMS_TERMINATOR);
      break;

    /* Binary 4 byte - LSB=0 first, MSB */
    case COF_SETTING4_BINARY4_LSB_FIRST:
      comms_net_rfprintf(comms_ctrl, "%c%c%c%c",
                     '\0',
                     (char) ((bin_rdg >> 0) & 0xFF),
                     (char) ((bin_rdg >> 8) & 0xFF),
                     (char) ((bin_rdg >> 16) & 0xFF)
                    );
      break;

    case COF_SETTING5_ASCII_WEIGHT_ADDR:
    case COF_SETTING7_ASCII_WEIGHT_ADDR:
      comms_send_num_with_dp(comms_ctrl, &adcrdg, MSV_RDG_MAX_WIDTH);
      comms_net_rfprintf(comms_ctrl, ",%02u" COMMS_TERMINATOR, db_app_user.address);
      break;

    case COF_SETTING9_ASCII_WEIGHT_ADDR_STATUS:
    case COF_SETTING10_ASCII_WEIGHT_ADDR_STATUS:
      comms_send_num_with_dp(comms_ctrl, &adcrdg, MSV_RDG_MAX_WIDTH);
      comms_net_rfprintf(comms_ctrl, ",%02u,%03u" COMMS_TERMINATOR,
                         db_app_user.address,
                         (unsigned) norm_status);
      break;

    case COF_SETTING11_ASCII_WEIGHT_ADDR_EXTSTATUS:
      comms_send_num_with_dp(comms_ctrl, &adcrdg, MSV_RDG_MAX_WIDTH);
      comms_net_rfprintf(comms_ctrl, ",%02u,%03u" COMMS_TERMINATOR,
                         db_app_user.address,
                         (unsigned) ext_status);
      break;

    case COF_SETTING12_ASCII_WEIGHT_ADDR_SPLITSTATUS:
      comms_send_num_with_dp(comms_ctrl, &adcrdg, MSV_RDG_MAX_WIDTH);
      comms_net_rfprintf(comms_ctrl, ",%02u,%03u,%03u" COMMS_TERMINATOR,
                         db_app_user.address,
                         (unsigned) split_status,
                         (unsigned) io_status);
      break;

    /* Binary 2 byte - LSB then MSB */
    case COF_SETTING6_BINARY2_LSB_FIRST:
      comms_net_rfprintf(comms_ctrl, "%c%c",
                     (char) ((bin_rdg >> 0) & 0xFF),
                     (char) ((bin_rdg >> 8) & 0xFF)
                    );
      break;

    /* Binary 4 byte - MSB first, LSB=status */
    case COF_SETTING8_BINARY4_MSB_FIRST_LSB_STATUS:
      comms_net_rfprintf(comms_ctrl, "%c%c%c%c",
                     (char) ((bin_rdg >> 16) & 0xFF),
                     (char) ((bin_rdg >> 8) & 0xFF),
                     (char) ((bin_rdg >> 0) & 0xFF),
                     (char) norm_status
                    );
      break;

    default:
      return;
  }
}


/** Send reading for MSM command (Modbus variant of MSV) */
void comms_send_oiml_msm(struct s_comms_control *comms_ctrl, enum eMSVSource source)
{
  uint32_t status;
  AdcRdg adcrdg;
  bool hide;

  /* Get the reading for the weight source */
  comms_get_reading_for_msv(&adcrdg, source, &hide);

  /* Work out status */
  status = (access_get_outputs() << 16);
  status |= (adcrdg.info.status.bit.uload) ? 0x00000001 : 0x0;
  status |= (adcrdg.info.status.bit.oload) ? 0x00000002 : 0x0;
  status |= (adcrdg.info.status.bit.motion) ? 0x00000004 : 0x0;
  status |= (adcrdg.info.status.bit.gross) ? 0x00000008 : 0x0;
  status |= (adcrdg.info.status.bit.range << 4);
  status |= (adcrdg.info.status.bit.coz) ? 0x00000040 : 0x0;
  status |= (hide || adcrdg.info.status.bit.oload || adcrdg.info.status.bit.uload ||
             APP_ERR_NONE != get_current_errors()) ? 0x00000080 : 0x0;

  /* Because MSM gets parsed by the Modbus interpreter, it can't deal with decimal points */
  adcrdg.info.unitdp.bit.dp = 0;

  /* Send the reading */
  comms_send_num_with_dp(comms_ctrl, &adcrdg, MSV_RDG_MAX_WIDTH);
  comms_net_rfprintf(comms_ctrl, ",%lu" COMMS_TERMINATOR, (unsigned long) status);
}


/** Converts ADC Library ADC_OP_STATUS_x codes to WE2111 comms MESS_x codes */
uint8_t convert_adc_result_to_mess_result(int8u adc_result)
{
  uint8_t mess_result;

  switch (adc_result)
  {
    case ADC_OP_STATUS_SUCCESS:
      mess_result = MESS_OK;
      break;

    case ADC_OP_STATUS_FAILED_MOTION:
      mess_result = MESS_ERR_MOTION;
      break;

    case ADC_OP_STATUS_FAILED_BAND:
      mess_result = MESS_ERR_RANGE;
      break;

    default:
    case ADC_OP_STATUS_FAILED_ERROR:
      mess_result = MESS_ERR_SYS;
      break;
  }
  return mess_result;
}
