#ifndef COMMUTIL_DOT_H
#define COMMUTIL_DOT_H 1

#include <stdarg.h>
#include <semaphore.h>
#include <rinhw/rio.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcdatatypes.h>
#include <rincmd/rc_public.h>
#include "sysclock.h"
#include "comms_params.h"
#include "comms_print_buffer.h"
#include "comms_send_buffer.h"
#include "db_setpoint.h"


/** Sources of data for the MSV response */
enum eMSVSource { MSV_SRC_DISPLAYED = 1,
                  MSV_SRC_GROSS,
                  MSV_SRC_NET,
                  MSV_SRC_TOTAL_NUMITEMS,
                  MSV_SRC_TOTAL1,
                  MSV_SRC_TOTAL2,
                  MSV_SRC_PEAKMAX,
                  MSV_SRC_LIVEWEIGH
                };
#define MSV_SRC_NUM (MSV_SRC_LIVEWEIGH)


/** COF settings */
enum eCOFSetting { COF_SETTING0_BINARY4_MSB_FIRST = 0,
                   COF_SETTING1_ASCII_WEIGHT,
                   COF_SETTING2_BINARY2_MSB_FIRST,
                   COF_SETTING3_ASCII_WEIGHT,
                   COF_SETTING4_BINARY4_LSB_FIRST,
                   COF_SETTING5_ASCII_WEIGHT_ADDR,
                   COF_SETTING6_BINARY2_LSB_FIRST,
                   COF_SETTING7_ASCII_WEIGHT_ADDR,
                   COF_SETTING8_BINARY4_MSB_FIRST_LSB_STATUS,
                   COF_SETTING9_ASCII_WEIGHT_ADDR_STATUS,
                   COF_SETTING10_ASCII_WEIGHT_ADDR_STATUS,
                   COF_SETTING11_ASCII_WEIGHT_ADDR_EXTSTATUS,
                   COF_SETTING12_ASCII_WEIGHT_ADDR_SPLITSTATUS,
                   COF_SETTING_NUM
                 };


/** Number of legacy modbus holding registers: for each setpoint, there are 2 target registers and 2 preflight registers */
#define NUM_HOLDING_REGS (4 * NUM_SETPOINTS)


/** Number of legacy modbus input registers */
#define NUM_INPUT_REGS 18


/** Struct for managing a single network comms connection/device (eg 1 serial port) */
struct s_comms_control
{
  struct t_io_dev *serdev;                         /*!< The serial device for communications */
  struct s_comms_send_buffer *csb;                 /*!< Send buffer to assemble the response into, so we can do a single write to serdev */
  struct s_comms_print_buffer *cpb;                /*!< The comms print buffer to retrieve data for the PRT command */
  bool reply_active;                               /*!< Are replies being sent?. Equivalent of WE2110 reply_active */
  bool unit_selected;                              /*!< Has the unit been selected by Sxx. Equivalent of WE2110 unit_selected */
  bool kr_cmd_streaming;                           /*!< Has Ks command been handled and we are streaming replies. Equivalent of WE2110 streaming */
  bool msv_cmd_streaming;                          /*!< Has MSV command been handled and we are streaming replies. Equivalent of WE2110 auto_send */
  enum eMSVSource msv_source;                      /*!< MSV command data source (displayed, gross, etc). Equivalent of WE2110 auto_sig */
  int msv_count;                                   /*!< How many MSV replies are still to be sent. Equivalent of WE2110 auto_cnt */
  bool cmd_query;                                  /*!< Command is a query (eg IDN?). Equivalent to WE2110 cmd_query */
  bool passcode_ok;                                /*!< Has a valid passcode been sent. Equivalent of WE2110 passcode_ok */
  bool send_extra_line_termination;                /*!< Extra line termination on MSV. Equivalent of WE2110 send_extra_end */
  int32u last_rdg_id;                              /*!< Id of last reading, used with MSV? */
  uint32_t rincomms_last_10hz_time;                /*!< The last time the rincomms 10Hz streaming sender was called */
  struct s_comms_params comms_params;              /*!< Command parameters */
  sem_t print_finished_sem;                        /*!< Semaphore to signal that the printout has completed */
  struct s_rc_localRegisterTable *regTable;        /*!< Local register table for this comms port */
  char *mod_mess;                                  /*!< Pointer to received legacy modbus message */
  size_t mod_len;                                  /*!< Length of received legacy modbus message */
  uint8_t lrc_out;                                 /*!< Checksum for legacy modbus response */
  uint16_t modbus_input_data[NUM_INPUT_REGS];      /*!< Local input register data for legacy modbus */
  uint16_t modbus_holding_data[NUM_HOLDING_REGS];  /*!< Local holding register data for legacy modbus */
};


#define CTRL_COMP (comms_ctrl->comms_params)


/** Simplification to get an integer parameter */
#define COMPARAM(x) (comms_ctrl->comms_params.params[x])


/** Simplification to get a string parameter */
#define COMPARAMSTR(x) (comms_ctrl->comms_params.param_str[x])


/** Returns true if ch is 0..9 A..F */
bool is_hex(char ch);


/** Convert nibble b to '0..9A..F' */
char to_hex(uint8_t b);


/** Convert single hex character to nibble */
uint8_t to_bin(char c);


/** Send readings as per COF settings */
void comms_send_oiml(struct s_comms_control *comms_ctrl);


/** Send reading for MSM command (Modbus variant of MSV) */
void comms_send_oiml_msm(struct s_comms_control *comms_ctrl, enum eMSVSource source);


/** Special version of vprintf for comms messages */
int comms_net_rvfprintf(struct s_comms_control *comms_ctrl, const char *format, va_list arglist);


/** Special version of printf for comms messages */
int comms_net_rfprintf(struct s_comms_control *comms_ctrl, const char *format, ...);


/**
 * Send a number with decimal point over the comms
 *
 * comms_ctrl is the comms net port to send the formatted number to
 * rdg is the number to be formatted and sent
 * dp is the decimal point position
 * maxlen is the desired length of the string
 * pad_spaces_not_zeros selects leading space padding (true), or leading '0' padding (false)
 * zeropadwidth is the minimum zero pad width. Used for NTEP where a reading of 0 must be padded to the resolution width, otherwise 0
 *
 * @return number of characters sent
 */
int comms_send_num_with_dp_pad(struct s_comms_control *comms_ctrl, long rdg, int dp, int maxlen, bool pad_spaces_not_zeros, int zeropadwidth);


/**
 * Convert a number with decimal point to a string
 *
 * str is a string that the number will be printed to
 * str_max_len is the maximum length of str
 * rdg is the number to be formatted and sent
 * dp is the decimal point position
 * maxlen is the desired length of the string
 * pad_spaces_not_zeros selects leading space padding (true), or leading '0' padding (false)
 * zeropadwidth is the minimum zero pad width. Used for NTEP where a reading of 0 must be padded to the resolution width, otherwise 0
 *
 * @return number of characters sent
 */
int comms_convert_num_with_dp_pad(char *str, size_t str_max_len, long rdg, int dp, int maxlen, bool pad_spaces_not_zeros, int zeropadwidth);


/** Send a number with decimal point over the comms */
void comms_send_num_with_dp(struct s_comms_control *comms_ctrl, const AdcRdg *rdg, int maxlen);


#define COMMS_TERMINATOR "\r\n"


/** Converts ADC Library ADC_OP_STATUS_x codes to WE2111 comms MESS_x codes */
uint8_t convert_adc_result_to_mess_result(int8u adc_result);


#endif /* #ifndef COMMUTIL_DOT_H */
