/*
 * daemonise.c
 *
 *  Created on: 16/11/2010
 *      Author: deans
 */
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include "we_debug.h"
#include "cmdline.h"

#include "daemonise.h"


/*
 * Do all needed to turn process into daemon.
 *
 * Separated into two functions. The first stage will
 * call vFork/exec and exit this is because the
 * standard fork() won't work. in no-MMU world.
 *
 * The second stage continues where the first left off.
 * The new exec'ed process can know that it is to continue as
 * the Start process sets env var cont_var to 1.
 *
 * May not return if a system call fails, will send
 * to stderr or logs using the errorlog functions.
 */


/** Start the daemonise process.
 *
 * Passes argv to the exec'ed process with the current environment.
 * The calling process can tell it has been restarted by the presence
 * of cont_var=1 in the environment.
 *
 * Returns 1 on error, 0 on success
 */
int daemoniseStart(int argc, char *argv[], const char *cont_var)
{
  pid_t pid;
  int result;

  WE_DAEMONISE_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "daemoniseStart: argc:%d, argv[0]:%s, cont_var:%s\n", argc, argv[0], cont_var);

  /* Allow u=rwx, g=rx, o=rx, deny g=w, o=w */
  WE_DAEMONISE_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "daemoniseStart: Setting file creation mask: deny g=w, o=w\n");
  umask(S_IWGRP | S_IWOTH);

  /* Now we vFork, let the child exec ourselves again and exit */
  WE_DAEMONISE_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "daemoniseStart: vfork() to get new process group\n");
  pid = vfork();
  if (pid < 0)
  {
    /* Failed */
    rdebug_printf(LOG_ERR, "daemoniseStart: vfork(%s) failed with result %d\n", argv[0], (int) pid);
    result = 1;
  }
  else if (0 != pid)
  {
    /* Parent, will sleep until child exits */
    WE_DAEMONISE_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "daemoniseStart: child started with pid %d, parent exiting\n", (int) pid);
    result = 0;
  }
  else
  {

    /*
     * This is the vforked child now
     * We first add *cont_var = 1
     * to the environment, then exec ourselves again
     *
     * The environment is inherited across the exec, and tested as part of the process startup
     */
    WE_DAEMONISE_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "daemoniseStart: child process, setting env var: %s=1\n", cont_var);
    if (setenv(cont_var, "1", 1) < 0)
    {
      rdebug_printf(LOG_INFO, "daemoniseStart: child started setenv() failed\n");
      exit(1);
    }

    WE_DAEMONISE_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "daemoniseStart: child execvp(%s)\n", argv[0]);
    execvp(argv[0], argv);

    /* execvp() never returns unless it failed */
    rdebug_printf(LOG_INFO, "daemoniseStart: child execvp() failed, errno %d\n", errno);
    exit(1);
  }
  return result;
}


/**
 * Continues where daemoniseStart off, as the child process. It is the responsibility
 * of the main process to call this, basically if the cont_var above
 * is set, then we start here.
 */
int daemoniseContinue(int argc, char *argv[])
{
  struct sigaction sa;

  WE_DAEMONISE_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "daemoniseContinue: argc:%d, argv[0]:%s\n", argc, argv[0]);

  /* Become a session leader to lose controlling TTY */
  if (setsid() < 0)
  {
    rdebug_printf(LOG_ERR, "daemoniseContinue: setsid() failed\n");
    return 1;
  }

  /* Ensure future opens won't allocate controlling TTYs */
  sa.sa_handler = SIG_IGN;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = 0;
  if (sigaction(SIGHUP, &sa, NULL) < 0)
  {
    rdebug_printf(LOG_ERR, "daemoniseContinue: sigaction(SIGHUP) failed\n");
    return 1;
  }

  /* Attach std* files to /dev/null */
  stdin = freopen("/dev/null", "r", stdin);
  if (NULL == stdin)
  {
    rdebug_printf(LOG_ERR, "daemoniseContinue: freopen(stdin) failed\n");
    return 1;
  }
  if (!cmdline_logtostdout)
  {
    stdout = freopen("/dev/null", "w", stdout);
    if (NULL == stdout)
    {
      rdebug_printf(LOG_ERR, "daemoniseContinue: freopen(stdout) failed\n");
      return 1;
    }
  }
  stderr = freopen("/dev/null", "w", stderr);
  if (NULL == stderr)
  {
    rdebug_printf(LOG_ERR, "daemoniseContinue: freopen(stderr) failed\n");
    return 1;
  }

  /*
   * Change the current working directory to the root so
   * we won't prevent file systems from being unmounted.
   */
  if (chdir("/") < 0)
  {
    rdebug_printf(LOG_ERR, "daemoniseContinue: chdir(\"/\") failed\n");
    return 1;
  }

  WE_DAEMONISE_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "daemoniseContinue: finished\n");
  return 0;
}
