#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include "system.h"
#include "meas.h"
#include "kbd_5000.h"
#include "version.h"
#include "resolution.h"
#include "devices.h"
#include "iface_analog.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "we_debug.h"

#include "database.h"


bool is_single_range(void)
{
  return (ADC_SCALE_TYPE_SINGLE == adcSuperStructure.adcUserDatabase.scaleType);
}


/** Check for any OIML errors, run on exiting setup */
enum eOIMLError check_oiml(void)
{
  int32s total_grads;
  size_t i;

  /* If not trade mode, then we don't check */
  if (adcSuperStructure.adcUserDatabase.approvalMode != ADC_APPROVAL_MODE_OIML)
    return OIML_ERROR_NONE;

  /* e1 or e2 > 50 */
  if (AdcResolutions[adcSuperStructure.adcUserDatabase.resolution[0]] > 50)
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_VERIFICATION_INTERVAL because e1 > 50\n");
    return OIML_ERROR_VERIFICATION_INTERVAL;
  }
  if (AdcResolutions[adcSuperStructure.adcUserDatabase.resolution[1]] > 50)
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_VERIFICATION_INTERVAL because e2 > 50\n");
    return OIML_ERROR_VERIFICATION_INTERVAL;
  }

  /* Check number of graduation */
  total_grads = calculate_scale_total_graduations();
  if (total_grads > OIML_SINGLE)
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_GRADUATIONS because single range graduations > %ld\n", OIML_SINGLE);
    return OIML_ERROR_GRADUATIONS;
  }
  if (!is_single_range() && (total_grads > OIML_MULTI))
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_GRADUATIONS because multiple range graduations > %ld\n", OIML_MULTI);
    return OIML_ERROR_GRADUATIONS;
  }

  /* Check units */
  if (UNITS_NONE == adcSuperStructure.adcUserDatabase.units)
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_NO_UNITS because no units assigned\n");
    return OIML_ERROR_NO_UNITS;
  }

  /* Check motion detection */
  if (adcSuperStructure.adcUserDatabase.motion == ADC_MOTION_DETECT_OFF)
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_MOTION_NONE because motion detection is disabled\n");
    return OIML_ERROR_MOTION_NONE;
  }

  /* Allowable settings for zero tracking are off, or 0.5d/s in trade mode */
  if (adcSuperStructure.adcUserDatabase.zeroTracking > ADC_ZERO_TRACKING_HALFD_1S)
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_ZERO_TRACK because zero tracking is out of range\n");
    return OIML_ERROR_ZERO_TRACK;
  }

  /* Allowable ranges are for zero range are +/-2% and -1% to 3% in trade mode */
  if (xlat_adc_library_zero_range_to_we2111(adcSuperStructure.adcUserDatabase.rangeOfZero) < WE21_ZERO_RANGE_NEG2_TO_PLUS2)
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_ZERO_RANGE because zero range is not +-2%% or -1%% to +3%%\n");
    return OIML_ERROR_ZERO_RANGE;
  }

  /* Check zero band */
  if (adcSuperStructure.adcUserDatabase.zeroBand > 0)
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_ZERO_BAND because zero band > 0\n");
    return OIML_ERROR_ZERO_BAND;
  }

  /* Immediate operation on zero, tare, print is an OIML fail */
  if (FRONTKEY_PERMIS_IMMEDIATE == db_app_user.front_key_permissions[FRONTKEY_ZERO])
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_KEY_IMMED because zero key set for immediate operation\n");
    return OIML_ERROR_KEY_IMMED;
  }
  if (FRONTKEY_PERMIS_IMMEDIATE == db_app_user.front_key_permissions[FRONTKEY_TARE])
  {
    rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_KEY_IMMED because tare key set for immediate operation\n");
    return OIML_ERROR_KEY_IMMED;
  }
  for (i = 0; i < NUM_FRONT_FUNCTION_KEYS; i++)
  {
    if (FRONTKEY_PERMIS_IMMEDIATE == db_app_user.front_key_permissions[FRONTKEY_FUNC1 + i] &&
        PRINT_KEY == translate_front_func_key(false, FUNC1_KEY + i))
    {
      rdebug_printf(LOG_WARNING, "check_oiml: reporting OIML_ERROR_KEY_IMMED because print key set for immediate operation\n");
      return OIML_ERROR_KEY_IMMED;
    }
  }

  /* Success, scale configured correctly in trade mode */
  return OIML_ERROR_NONE;
}


long true_full_scale(void)
{
  if (is_single_range())
    return adcSuperStructure.adcUserDatabase.fullScale[0];
  else
    return adcSuperStructure.adcUserDatabase.fullScale[1];
}


long min_span_value (void)
{
  return (true_full_scale () >> 6);     /* approx 2 % */
}


bool is_weight_fmt(const char *s)
{
  bool bslash_last;

  bslash_last = false;
  while (*s)
  {
    if (bslash_last)
    {
      bslash_last = false;
      switch (*s)
      {
        case 'G':
        case 'g':
        case 'N':
        case 'n':
        case 'W':
        case 'w':
        case '+':
          return true;
      }
    }
    else if (*s == '\\')
      bslash_last = true;
    s++;
  }
  return false;
}
