/*
 * db_remote.h
 *
 * Remote inputs
 */

#ifndef DB_REMOTE_H_
#define DB_REMOTE_H_

#include <stdint.h>
#include "kbd_5000.h"


/** Number of remote keys supported by the system */
#define NUM_REMOTE_KEYS 8


/** Remote key functions */
enum eRemoteKeyFn { REMOTE_DISABLED = 0,
                    REMOTE_ZERO,
                    REMOTE_TARE,
                    REMOTE_GROSS,
                    REMOTE_PRINT,
                    REMOTE_BLANK,
                    REMOTE_LOCK,
                    REMOTE_SHOW,
                    REMOTE_CLEAR,
                    REMOTE_UNDO,
                    REMOTE_SINGLE_1,
                    REMOTE_SINGLE_2,
                    REMOTE_MHOLD,
                    REMOTE_PHOLD,
                    REMOTE_LHOLD,
                    REMOTE_TXTPRN,
                    REMOTE_SINGLE_3,
                    REMOTE_SINGLE_4,
                    REMOTE_UNITS,
                    REMOTE_SINGLE_5,
                    REMOTE_SINGLE_6,
                    REMOTE_SP_RESET_A,
                    REMOTE_SP_RESET_B,
                    NUM_REMOTE_KEY_FUNCTIONS
                  };


/** Remote function keys functions string list. Matches @ref enum eRemoteKeyFn */
#define REMOTE_KEYFN_OPT_STRS { "NONE", \
                                "ZERO", \
                                "TARE", \
                                "GROSS", \
                                "PRINT", \
                                "BLANK", \
                                "LOCK", \
                                "SHW.TOT", \
                                "CLR.TOT", \
                                "UNDO", \
                                "SIN.M1A", \
                                "SIN.M1B", \
                                "M.HOLD", \
                                "P.HOLD", \
                                "L.HOLD", \
                                "TXT.PRN", \
                                "SIN.M2A", \
                                "SIN.M2B", \
                                "UNITS", \
                                "SIN.SLV", \
                                "SIN.HST", \
                                "SP.RST.A", \
                                "SP.RST.B" \
                              }


#define REMOTE_KEYFN_XLAT  "-zTGq--SCU12Hef-34n56--"


/** Lookup to translate each of the remote key functions in @ref enum eRemoteKeyFn into a keypress, but no the batching ones */
#define REMOTE_KEYFN_LOOKUP { \
                              NO_KEY,           /* REMOTE_DISABLED */ \
                              ZERO_KEY,         /* REMOTE_ZERO */ \
                              TARE_KEY,         /* REMOTE_TARE */ \
                              GROSS_KEY,        /* REMOTE_GROSS */ \
                              PRINT_KEY,        /* REMOTE_PRINT */ \
                              NO_KEY,           /* REMOTE_BLANK */ \
                              NO_KEY,           /* REMOTE_LOCK */ \
                              SHOW_TOTAL_KEY,   /* REMOTE_SHOW */ \
                              CLR_TOTAL_KEY,    /* REMOTE_CLEAR */ \
                              UNDO_KEY,         /* REMOTE_UNDO */ \
                              SINGLE1_KEY,      /* REMOTE_SINGLE_1 */ \
                              SINGLE2_KEY,      /* REMOTE_SINGLE_2 */ \
                              MHOLD_KEY,        /* REMOTE_MHOLD */ \
                              PHOLD_KEY,        /* REMOTE_PHOLD */ \
                              LHOLD_KEY,        /* REMOTE_LHOLD */ \
                              NO_KEY,           /* REMOTE_TXTPRN */ \
                              SINGLE3_KEY,      /* REMOTE_SINGLE_3 */ \
                              SINGLE4_KEY,      /* REMOTE_SINGLE_4 */ \
                              UNITS_KEY,        /* REMOTE_UNITS */ \
                              SINGLE5_KEY,      /* REMOTE_SINGLE_5 */ \
                              SINGLE6_KEY,      /* REMOTE_SINGLE_6 */ \
                              NO_KEY,           /* REMOTE_SP_RESET_A */ \
                              NO_KEY            /* REMOTE_SP_RESET_B */ \
                            }


/** Remote inputs implemented as key presses, not inputs. Must match @ref enum eRemoteKeyFn */
#define REMOTE_KEYFN_IS_KEY { \
                              false,        /* REMOTE_DISABLED */ \
                              true,         /* REMOTE_ZERO */ \
                              true,         /* REMOTE_TARE */ \
                              true,         /* REMOTE_GROSS */ \
                              true,         /* REMOTE_PRINT */ \
                              false,        /* REMOTE_BLANK */ \
                              false,        /* REMOTE_LOCK */ \
                              true,         /* REMOTE_SHOW */ \
                              true,         /* REMOTE_CLEAR */ \
                              true,         /* REMOTE_UNDO */ \
                              true,         /* REMOTE_SINGLE_1 */ \
                              true,         /* REMOTE_SINGLE_2 */ \
                              true,         /* REMOTE_MHOLD */ \
                              true,         /* REMOTE_PHOLD */ \
                              true,         /* REMOTE_LHOLD */ \
                              false,        /* REMOTE_TXTPRN */ \
                              true,         /* REMOTE_SINGLE_3 */ \
                              true,         /* REMOTE_SINGLE_4 */ \
                              true,         /* REMOTE_UNITS */ \
                              true,         /* REMOTE_SINGLE_5 */ \
                              true,         /* REMOTE_SINGLE_6 */ \
                              false,        /* REMOTE_SP_RESET_A */ \
                              false         /* REMOTE_SP_RESET_B */ \
                            }


/** Remote function keys functions string list. Matches @ref enum eRemoteKeyFn */
extern const char *const remote_keyfn_strings[NUM_REMOTE_KEY_FUNCTIONS];


/** Lookup to translate each of the remote key functions in @ref enum eRemoteKeyFn into a keypress, but not the batching ones */
extern const uint8_t remote_key_lookup[NUM_REMOTE_KEY_FUNCTIONS];


/** Remote inputs implemented as key presses, not inputs. Must match @ref enum eRemoteKeyFn */
extern const bool remote_key_is_key[NUM_REMOTE_KEY_FUNCTIONS];


/** Defaults for each of the remote key inputs */
extern const uint8_t remote_key_defaults[NUM_REMOTE_KEYS];


/** Number of function keys on the front of the indicator */
#define NUM_FRONT_FUNCTION_KEYS 3


/** Front function key functions */
enum eFunctionKeyFn { FRONTFUNC_DISABLED = 0,
                      FRONTFUNC_PRINT,
                      FRONTFUNC_SHOW,
                      FRONTFUNC_CLEAR,
                      FRONTFUNC_UNDO,
                      FRONTFUNC_SINGLE_1,
                      FRONTFUNC_SINGLE_2,
                      FRONTFUNC_MHOLD,
                      FRONTFUNC_PHOLD,
                      FRONTFUNC_LHOLD,
                      FRONTFUNC_SINGLE_3,
                      FRONTFUNC_SINGLE_4,
                      FRONTFUNC_UNITS,
                      FRONTFUNC_SINGLE_5,
                      FRONTFUNC_SINGLE_6,
                      NUM_FRONTFUNC
                    };


/** Front function keys functions string list. Matches @ref enum eFunctionKeyFn */
#define FRONTFUNC_KEYFN_OPT_STRS { "NONE", \
                                   "PRINT", \
                                   "SHW.TOT", \
                                   "CLR.TOT", \
                                   "UNDO", \
                                   "SIN.M1A", \
                                   "SIN.M1B", \
                                   "M.HOLD", \
                                   "P.HOLD", \
                                   "L.HOLD", \
                                   "SIN.M2A", \
                                   "SIN.M2B", \
                                   "UNITS", \
                                   "SIN.SLV", \
                                   "SIN.HST" \
                                 }


/** Convert front functions into keys */
#define FRONTFUNC_KEYFN_XLAT  "-qSCU12Hef34n56"


/** Lookup to translate each of the front function key functions in @ref enum eFunctionKeyFn into a keypress, but no the batching ones */
#define FRONTFUNC_KEYFN_LOOKUP { NO_KEY, \
                                 PRINT_KEY, \
                                 SHOW_TOTAL_KEY, \
                                 CLR_TOTAL_KEY, \
                                 UNDO_KEY, \
                                 SINGLE1_KEY, \
                                 SINGLE2_KEY, \
                                 MHOLD_KEY, \
                                 PHOLD_KEY, \
                                 LHOLD_KEY, \
                                 SINGLE3_KEY, \
                                 SINGLE4_KEY, \
                                 UNITS_KEY, \
                                 SINGLE5_KEY, \
                                 SINGLE6_KEY \
                               }


/** Front function keys functions string list. Matches @ref enum eFunctionKeyFn */
extern const char *const frontfunc_keyfn_strings[NUM_FRONTFUNC];


/** Lookup to translate each of the front function key functions in @ref enum eFunctionKeyFn into a keypress, but not the batching ones */
extern const enum eIndicatorKey frontfunc_key_lookup[NUM_FRONTFUNC];


/** Defaults for each of the front function keys */
extern const uint8_t frontfunc_key_defaults[NUM_FRONT_FUNCTION_KEYS];


/** String to convert a remote button into a key */
extern const char remote_button_to_key_str[NUM_REMOTE_KEY_FUNCTIONS + 1];


/** String to convert a front function key press into a key */
extern const char front_func_to_key_str[NUM_REMOTE_KEY_FUNCTIONS + 1];


#endif /* DB_REMOTE_H_ */
