/*
 * db_serial.c
 *
 *  Created on: 27/10/2010
 *      Author: ian
 */
#include <rinhw/stdrinstrum.h>
#include <rinhw/sio.h>
#include "database.h"
#include "serial_ports.h"
#include "app_db_user.h"
#include "comms_task.h"
#include "modbus_config.h"
#include "indicator_model_features.h"

#include "db_serial.h"


/** Lookup of baud rate from baud rate index */
static const int32u db_baud_setting_ints[NUM_DB_BAUD_SETTINGS] = DB_BAUD_INTEGERS;


/** Convert a database baud rate index to real baud rate */
static int32u db_get_serial_baud_from_index(enum eSerialBaudIndex baud_index)
{
  int32u baud;

  baud = (baud_index < NUM_DB_BAUD_SETTINGS)
          ? db_baud_setting_ints[baud_index]
          : db_get_serial_baud_from_index(DB_SERIAL_DEFAULT_BAUD_INDEX);
  return baud;
}


/** Get the baud rate index for a serial port */
enum eSerialBaudIndex db_get_serial_baud_index(enum eSerialPort port)
{
  return (port < NUM_DB_SERIAL_SETTINGS) ?
          (enum eSerialBaudIndex) db_app_user.serial_baud_idx[port] :
          DB_SERIAL_DEFAULT_BAUD_INDEX;
}


/** Set the baud rate index for a serial port */
void db_set_serial_baud_index(enum eSerialPort port, enum eSerialBaudIndex baud_idx)
{
  if (port < NUM_DB_SERIAL_SETTINGS && baud_idx < NUM_DB_BAUD_SETTINGS)
    db_app_user.serial_baud_idx[port] = baud_idx;
}


/** Get real baud rate for a port from the database */
static int32u db_get_serial_baud(enum eSerialPort port)
{
  enum eSerialBaudIndex baud_idx;

  baud_idx = db_get_serial_baud_index(port);
  return db_get_serial_baud_from_index(baud_idx);
}


/** Convert database packed parity bits into real parity setting */
int8u db_get_serial_paritybits_from_bits(int8u bits)
{
  int8u parity_setting;

  switch (bits & DB_SERIAL_PARITY_MASK)
  {
    case DB_SERIAL_PARITY_NONE:
      parity_setting = SIO_PARITY_NONE;
      break;
    case DB_SERIAL_PARITY_ODD:
      parity_setting = SIO_PARITY_ODD;
      break;
    case DB_SERIAL_PARITY_EVEN:
      parity_setting = SIO_PARITY_EVEN;
      break;
    default:
      parity_setting = db_get_serial_paritybits_from_bits(DB_SERIAL_DEFAULT_PARITY);
      break;
  }
  return parity_setting;
}


/** Convert real parity setting to database packed parity bits */
int8u db_get_bits_from_serial_paritybits(int8u parity_setting)
{
  int8u bits;

  switch (parity_setting)
  {
    case SIO_PARITY_NONE:
      bits = DB_SERIAL_PARITY_NONE;
      break;
    case SIO_PARITY_ODD:
      bits = DB_SERIAL_PARITY_ODD;
      break;
    case SIO_PARITY_EVEN:
      bits = DB_SERIAL_PARITY_EVEN;
      break;
    default:
      bits = DB_SERIAL_DEFAULT_PARITY;
      break;
  }
  return bits;
}


/** Get real parity bits setting for a port from the database */
int8u db_get_serial_paritybits(enum eSerialPort port)
{
  int8u db_parity_bits;

  db_parity_bits = (port < NUM_DB_SERIAL_SETTINGS) ?
                    db_app_user.serial_bit_settings[port] :
                    DB_SERIAL_DEFAULT;

  return db_get_serial_paritybits_from_bits(db_parity_bits);
}


/** Convert database packed data bits into real data bits setting */
int8u db_get_serial_databits_from_bits(int8u bits)
{
  int8u data_bits_setting;

  switch (bits & DB_SERIAL_DATA_MASK)
  {
    case DB_SERIAL_DATA_7:
      data_bits_setting = SIO_DATABITS_SEVEN;
      break;
    case DB_SERIAL_DATA_8:
      data_bits_setting = SIO_DATABITS_EIGHT;
      break;
    default:
      data_bits_setting = db_get_serial_databits_from_bits(DB_SERIAL_DEFAULT_DATA);
      break;
  }
  return data_bits_setting;
}


/** Convert real data bits setting into database packed data bits */
int8u db_get_bits_from_serial_databits(int8u data_bits_setting)
{
  int8u bits;

  switch (data_bits_setting)
  {
    case SIO_DATABITS_SEVEN:
      bits = DB_SERIAL_DATA_7;
      break;
    case SIO_DATABITS_EIGHT:
      bits = DB_SERIAL_DATA_8;
      break;
    default:
      bits = DB_SERIAL_DEFAULT_DATA;
      break;
  }
  return bits;
}


/** Get real data bits setting for a port from the database */
int8u db_get_serial_databits(enum eSerialPort port)
{
  int8u db_data_bits;

  db_data_bits = (port < NUM_DB_SERIAL_SETTINGS) ?
                  db_app_user.serial_bit_settings[port] :
                  DB_SERIAL_DEFAULT;

 return db_get_serial_databits_from_bits(db_data_bits);
}


/** Convert database packed stop bits into real stop bits setting */
int8u db_get_serial_stopbits_from_bits(int8u bits)
{
  int8u stop_bits_setting;

  switch (bits & DB_SERIAL_STOP_MASK)
  {
    case DB_SERIAL_STOP_1:
      stop_bits_setting = SIO_STOPBITS_ONE;
      break;
    case DB_SERIAL_STOP_2:
      stop_bits_setting = SIO_STOPBITS_TWO;
      break;
    default:
      stop_bits_setting = db_get_serial_stopbits_from_bits(DB_SERIAL_DEFAULT_STOP);
      break;
  }
  return stop_bits_setting;
}


/** Convert real stop bits setting into database packed stop bits */
int8u db_get_bits_from_serial_stopbits_from_bits(int8u stop_bits_setting)
{
  int8u bits;

  switch (stop_bits_setting)
  {
    case SIO_STOPBITS_ONE:
      bits = DB_SERIAL_STOP_1;
      break;
    case SIO_STOPBITS_TWO:
      bits = DB_SERIAL_STOP_2;
      break;
    default:
      bits = DB_SERIAL_DEFAULT_STOP;
      break;
  }
  return bits;
}


/** Get stopbits for a port from the database */
int8u db_get_serial_stopbits(enum eSerialPort port)
{
  int8u db_stop_bits;

  db_stop_bits = (port < NUM_DB_SERIAL_SETTINGS) ?
                  db_app_user.serial_bit_settings[port] :
                  DB_SERIAL_DEFAULT;

  return db_get_serial_stopbits_from_bits(db_stop_bits);
}


/** Get real use cts setting for a port from the database */
bool db_get_serial_usects(enum eSerialPort port)
{
  int8u db_bits;

  db_bits = (port < NUM_DB_SERIAL_SETTINGS) ?
             db_app_user.serial_bit_settings[port] :
             DB_SERIAL_DEFAULT;

 return DB_GET_SERIAL_USECTS_FROM_BITS(db_bits);
}


/** Get real use terminations setting for a port from the database */
bool db_get_serial_useterm(enum eSerialPort port)
{
  int8u db_bits;

  db_bits = (port < NUM_DB_SERIAL_SETTINGS) ?
             db_app_user.serial_bit_settings[port] :
             DB_SERIAL_DEFAULT;

 return DB_GET_SERIAL_USETERM_FROM_BITS(db_bits);
}


/** Get real use RS232 or RS422 setting for a port from the database */
bool db_get_serial_users232(enum eSerialPort port)
{
  int8u db_bits;

  db_bits = (port < NUM_DB_SERIAL_SETTINGS) ?
             db_app_user.serial_bit_settings[port] :
             DB_SERIAL_DEFAULT;

  return DB_GET_SERIAL_USERS232_FROM_BITS(db_bits);
}


/** Get the serial port configuration for port from the database */
void db_get_serialconfig(enum eSerialPort port, t_ioctl_siocfg *siocfg, bool *use_rs232)
{
  siocfg->baud = db_get_serial_baud(port);
  siocfg->parity_bits = db_get_serial_paritybits(port);
  siocfg->data_bits = db_get_serial_databits(port);
  siocfg->stop_bits = db_get_serial_stopbits(port);
  siocfg->use_cts = db_get_serial_usects(port);
  siocfg->use_terminations = db_get_serial_useterm(port);

  *use_rs232 = db_get_serial_users232(port);
}


/** Get the default serial port configuration */
void db_get_serialconfig_default(t_ioctl_siocfg *siocfg, bool *use_rs232)
{
  siocfg->baud = db_get_serial_baud_from_index(DB_SERIAL_DEFAULT_BAUD_INDEX);
  siocfg->parity_bits = db_get_serial_paritybits_from_bits(DB_SERIAL_DEFAULT);
  siocfg->data_bits = db_get_serial_databits_from_bits(DB_SERIAL_DEFAULT);
  siocfg->stop_bits = db_get_serial_stopbits_from_bits(DB_SERIAL_DEFAULT);
  siocfg->use_cts = DB_GET_SERIAL_USECTS_FROM_BITS(DB_SERIAL_DEFAULT);
  siocfg->use_terminations = DB_GET_SERIAL_USETERM_FROM_BITS(DB_SERIAL_DEFAULT);

  *use_rs232 = DB_GET_SERIAL_USERS232_FROM_BITS(DB_SERIAL_DEFAULT);
}


/**
 * Get serial port type from database.
 *
 * Only works for port 0..SERPORT_NETWORK_START - 1 and SERPORT_NETWORK_FIRST..SERPORT_NETWORK_FINAL
 * SERPORT_NETWORK_START..SERPORT_NUM are created from the listening TCP sockets, so get their s_comms_thread->devicetype
 * from the thread that created them.
 *
 * See @ref comms_task_list_get_port_type() for the type of a running serial port task
 */
enum eSerialPortType db_get_serial_type(enum eSerialPort port)
{
  enum eSerialPortType db_sertype;

  switch (port)
  {
    case 0 ... NUM_DB_SERIAL_SETTINGS - 1:
      /* Serial 1-8 live in the database (configurable) */
      db_sertype = (enum eSerialPortType) db_app_user.serial_type[port];
      break;

    case SERPORT_USBLP:
      db_sertype = DB_SERIAL_TYPE_PRINTING;
      break;

    case SERPORT_UDPOUT1:
      db_sertype = (enum eSerialPortType) db_app_user.serial_udpo1_type;
      break;

    case SERPORT_UDPOUT2:
      db_sertype = (enum eSerialPortType) db_app_user.serial_udpo2_type;
      break;

    case SERPORT_UDPIN1:
      db_sertype = (enum eSerialPortType) db_app_user.serial_udpi1_type;
      break;

    case SERPORT_UDPIN2:
      db_sertype = (enum eSerialPortType) db_app_user.serial_udpi2_type;
      break;

    case SERPORT_TCPNET:
    case SERPORT_UDSNET:
      db_sertype = DB_SERIAL_TYPE_NETWORK;
      break;

    case SERPORT_TCPAUTO:
      db_sertype = (enum eSerialPortType) db_app_user.serial_tcpi2_type;
      break;

    case SERPORT_UDSAUTO:
      db_sertype = DB_SERIAL_TYPE_AUTO_LO;
      break;

    default:
      /* I dunno, turn it off? */
      db_sertype = DB_SERIAL_TYPE_OFF;
      break;
  }
  return db_sertype;
}


/** Get the lower limit on serial port type for a particular port */
enum eSerialPortType db_get_serial_type_min(enum eSerialPort port)
{
  enum eSerialPortType db_sertype_min;

  switch (port)
  {
    case 0 ... NUM_DB_SERIAL_SETTINGS - 1:
      /* Serial 1-8 live in the database (configurable) */
      db_sertype_min = (enum eSerialPortType) serial_type_minimums[port];
      break;

    case SERPORT_UDPOUT1:
    case SERPORT_UDPOUT2:
    case SERPORT_UDPIN1:
    case SERPORT_UDPIN2:
    case SERPORT_TCPNET:
    case SERPORT_UDSNET:
    case SERPORT_UDSAUTO:
    case SERPORT_USBLP:
    default:
      db_sertype_min = DB_SERIAL_TYPE_OFF;
      break;
    case SERPORT_TCPAUTO:
      db_sertype_min = DB_SERIAL_TYPE_AUTO_LO;
      break;
  }
  return db_sertype_min;
}


/** Get the upper limit on serial port type for a particular port */
enum eSerialPortType db_get_serial_type_max(enum eSerialPort port)
{
  enum eSerialPortType db_sertype_max;

  switch (port)
  {
    case 0 ... NUM_DB_SERIAL_SETTINGS - 1:
      /* Serial 1-8 live in the database (configurable) */
      db_sertype_max = (enum eSerialPortType) serial_type_maximums[port];
      break;

    case SERPORT_UDPOUT1:
    case SERPORT_UDPOUT2:
    case SERPORT_UDPIN1:
    case SERPORT_UDPIN2:
    case SERPORT_TCPNET:
    case SERPORT_UDSNET:
      db_sertype_max = DB_SERIAL_TYPE_NETWORK;
      break;
    case SERPORT_TCPAUTO:
      db_sertype_max = DB_SERIAL_TYPE_AUTO_HI;
      break;
    case SERPORT_UDSAUTO:
      db_sertype_max = DB_SERIAL_TYPE_AUTO_LO;
      break;
    case SERPORT_USBLP:
      db_sertype_max = DB_SERIAL_TYPE_PRINTING;
      break;
    default:
      db_sertype_max = NUM_DB_SERIAL_TYPES_BASIC;
      break;
  }
  return db_sertype_max;
}


/**
 * Set serial port type in the database.
 *
 * Only works for port 0..SERPORT_NETWORK_START - 1 and SERPORT_NETWORK_FIRST..SERPORT_NETWORK_FINAL
 * SERPORT_NETWORK_START..SERPORT_NUM are created from the listening TCP sockets, so get their s_comms_thread->devicetype
 * from the thread that created them.
 */
void db_set_serial_type(enum eSerialPort port, enum eSerialPortType porttype)
{
  switch (port)
  {
    case 0 ... NUM_DB_SERIAL_SETTINGS - 1:
      /* Serial 1-8 live in the database (configurable) */
      db_app_user.serial_type[port] = porttype;
      break;

    case SERPORT_UDPOUT1:
      db_app_user.serial_udpo1_type = porttype;
      break;

    case SERPORT_UDPOUT2:
      db_app_user.serial_udpo2_type = porttype;
      break;

    case SERPORT_UDPIN1:
      db_app_user.serial_udpi1_type = porttype;
      break;

    case SERPORT_UDPIN2:
      db_app_user.serial_udpi2_type = porttype;
      break;

    case SERPORT_TCPAUTO:
      db_app_user.serial_tcpi2_type = porttype;
      break;

    default:
      break;
  }
}


/** Force the serial settings for any attached rinlinks to the rinlink fixed settings */
void db_force_rinlink_serial_settings(void)
{
  const enum e_indicator_model_sub_type imst = get_indicator_model_subtype();

  if (indicator_model_subtype_has_front_rinlink(imst))
  {
    if (DB_SERIAL_DEFAULT_BAUD_INDEX_RL != db_app_user.serial_baud_idx[SERPORT_RL_FRONT])
    {
      db_app_user.serial_baud_idx[SERPORT_RL_FRONT] = DB_SERIAL_DEFAULT_BAUD_INDEX_RL;
    }
    if (DB_SERIAL_DEFAULT != db_app_user.serial_bit_settings[SERPORT_RL_FRONT])
    {
      db_app_user.serial_bit_settings[SERPORT_RL_FRONT] = DB_SERIAL_DEFAULT;
    }
    if (DB_SERIAL_DEFAULT_TYPE_RL_FRONT != db_app_user.serial_type[SERPORT_RL_FRONT])
    {
      db_app_user.serial_type[SERPORT_RL_FRONT] = DB_SERIAL_DEFAULT_TYPE_RL_FRONT;
    }
  }
  if (indicator_model_subtype_has_rear_rinlink(imst))
  {
    if (DB_SERIAL_DEFAULT_BAUD_INDEX_RL != db_app_user.serial_baud_idx[SERPORT_SER1_RL_REAR])
    {
      db_app_user.serial_baud_idx[SERPORT_SER1_RL_REAR] = DB_SERIAL_DEFAULT_BAUD_INDEX_RL;
    }
    if (DB_SERIAL_DEFAULT != db_app_user.serial_bit_settings[SERPORT_SER1_RL_REAR])
    {
      db_app_user.serial_bit_settings[SERPORT_SER1_RL_REAR] = DB_SERIAL_DEFAULT;
    }
    if (DB_SERIAL_DEFAULT_TYPE_RL_REAR != db_app_user.serial_type[SERPORT_SER1_RL_REAR])
    {
      db_app_user.serial_type[SERPORT_SER1_RL_REAR] = DB_SERIAL_DEFAULT_TYPE_RL_REAR;
    }
  }
}
