/*
 * db_serial.h
 *
 *  Created on: 27/10/2010
 *      Author: ian
 */

#ifndef DB_SERIAL_H_
#define DB_SERIAL_H_

#include <stdint.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rstdio.h>
#include <rinhw/sio.h>
#include "serial_ports.h"
#include "modbus_config.h"


/** Available baud rate settings */
enum eSerialBaudIndex
{
  DB_SERIAL_BAUDIDX_300 = 0,
  DB_SERIAL_BAUDIDX_600,
  DB_SERIAL_BAUDIDX_1200,
  DB_SERIAL_BAUDIDX_2400,
  DB_SERIAL_BAUDIDX_4800,
  DB_SERIAL_BAUDIDX_9600,
  DB_SERIAL_BAUDIDX_19200,
  DB_SERIAL_BAUDIDX_38400,
  DB_SERIAL_BAUDIDX_57600,
  DB_SERIAL_BAUDIDX_115200,

  NUM_DB_BAUD_SETTINGS
};


#define DB_BAUD_STRINGS  { "300", "600", "1200", "2400", "4800", "9600", "19200", "38400", "57600", "115200" }
#define DB_BAUD_INTEGERS {  300,   600,   1200,   2400,   4800,   9600,   19200,   38400,   57600,   115200  }


/*      parity options     */
#define DB_SERIAL_PARITY_NONE     0x00
#define DB_SERIAL_PARITY_ODD      0x10
#define DB_SERIAL_PARITY_EVEN     0x20

/*      data bit options     */
#define DB_SERIAL_DATA_7     0x00
#define DB_SERIAL_DATA_8     0x08

/*      stop bit options     */
#define DB_SERIAL_STOP_1     0x00
#define DB_SERIAL_STOP_2     0x04

#define DB_SERIAL_NOCTS      0x00
#define DB_SERIAL_USECTS     0x40

#define DB_SERIAL_NOTERM     0x00
#define DB_SERIAL_USETERM    0x02

#define DB_SERIAL_RS485      0x00
#define DB_SERIAL_RS232      0x01


/* COMMS SETTINGS */
#define DB_SERIAL_PARITY_MASK     0x30
#define DB_SERIAL_DATA_MASK       0x08
#define DB_SERIAL_STOP_MASK       0x04
#define DB_SERIAL_USETERM_MASK    0x02
#define DB_SERIAL_USERS232_MASK   0x01
#define DB_SERIAL_USECTS_MASK     0x40

#define DB_SERIAL_PARITY_SHIFTS 4
#define DB_SERIAL_DATA_SHIFTS 3
#define DB_SERIAL_STOP_SHIFTS 2
#define DB_SERIAL_TERM_SHIFTS 1
#define DB_SERIAL_RS232_SHIFTS 0
#define DB_SERIAL_CTS_SHIFTS 6


#define DB_SERIAL_DEFAULT_BAUD_INDEX    DB_SERIAL_BAUDIDX_9600
#define DB_SERIAL_DEFAULT_BAUD_INDEX_RL DB_SERIAL_BAUDIDX_9600
#define DB_SERIAL_DEFAULT_PARITY       DB_SERIAL_PARITY_NONE
#define DB_SERIAL_DEFAULT_DATA         DB_SERIAL_DATA_8
#define DB_SERIAL_DEFAULT_STOP         DB_SERIAL_STOP_1
#define DB_SERIAL_DEFAULT_USECTS       DB_SERIAL_NOCTS
#define DB_SERIAL_DEFAULT_USETERM      DB_SERIAL_NOTERM
#define DB_SERIAL_DEFAULT_USERS232     DB_SERIAL_RS232


#define DB_SERIAL_DEFAULT (DB_SERIAL_DEFAULT_PARITY | \
                           DB_SERIAL_DEFAULT_DATA | \
                           DB_SERIAL_DEFAULT_STOP | \
                           DB_SERIAL_DEFAULT_USECTS | \
                           DB_SERIAL_DEFAULT_USETERM | \
                           DB_SERIAL_DEFAULT_USERS232 )


#define DB_GET_SERIAL_USECTS_FROM_BITS(bits) (0 != (bits & DB_SERIAL_USECTS_MASK))
#define DB_GET_SERIAL_USETERM_FROM_BITS(bits) (0 != (bits & DB_SERIAL_USETERM_MASK))
#define DB_GET_SERIAL_USERS232_FROM_BITS(bits) (0 != (bits & DB_SERIAL_USERS232_MASK))


/** Serial port type settings */
enum eSerialPortType
{
  DB_SERIAL_TYPE_OFF = 0,
  DB_SERIAL_TYPE_AUTO_LO,
  DB_SERIAL_TYPE_AUTO_HI,
  DB_SERIAL_TYPE_PRINTING,
  DB_SERIAL_TYPE_SINGLE,
  DB_SERIAL_TYPE_NETWORK,
  DB_SERIAL_TYPE_MODBUS_LEGACY,
  DB_SERIAL_TYPE_MODBUS_ASCII,
  DB_SERIAL_TYPE_MODBUS_RTU,

  NUM_DB_SERIAL_TYPES,                                            /*!< For a fully featured port eg USB host serial */
  NUM_DB_SERIAL_TYPES_TXONLY = DB_SERIAL_TYPE_SINGLE + 1,         /*!< For a transmit only port eg rinblock b port */
  NUM_DB_SERIAL_TYPES_BASIC = DB_SERIAL_TYPE_MODBUS_LEGACY + 1,   /*!< For a rinblock tx/rx port eg rinblock a port */
};

#define DB_SERIAL_TYPE_STRINGS { "OFF", "AUTO.LO", "AUTO.HI", "PRINT", "SINGLE", "NET", "MB.LGCY", "MB.ASCI", "MB.RTU" }

#define DB_SERIAL_DEFAULT_TYPE_RL_FRONT DB_SERIAL_TYPE_NETWORK
#define DB_SERIAL_DEFAULT_TYPE_RL_REAR DB_SERIAL_TYPE_NETWORK
#define DB_SERIAL_DEFAULT_TYPE_RB1A DB_SERIAL_TYPE_NETWORK
#define DB_SERIAL_DEFAULT_TYPE_RB1B DB_SERIAL_TYPE_AUTO_LO
#define DB_SERIAL_DEFAULT_TYPE_RB2A DB_SERIAL_TYPE_NETWORK
#define DB_SERIAL_DEFAULT_TYPE_RB2B DB_SERIAL_TYPE_AUTO_LO
#define DB_SERIAL_DEFAULT_TYPE_USBG DB_SERIAL_TYPE_NETWORK
#define DB_SERIAL_DEFAULT_TYPE_USBH DB_SERIAL_TYPE_NETWORK
#define DB_SERIAL_DEFAULT_TYPE_SER2 DB_SERIAL_TYPE_NETWORK
#define DB_SERIAL_DEFAULT_TYPE_UDPO1 DB_SERIAL_TYPE_OFF
#define DB_SERIAL_DEFAULT_TYPE_UDPO2 DB_SERIAL_TYPE_OFF
#define DB_SERIAL_DEFAULT_TYPE_UDPI1 DB_SERIAL_TYPE_OFF
#define DB_SERIAL_DEFAULT_TYPE_UDPI2 DB_SERIAL_TYPE_OFF
#define DB_SERIAL_DEFAULT_TYPE_TCPI2 DB_SERIAL_TYPE_AUTO_LO


/** Get the baud rate index for a serial port */
enum eSerialBaudIndex db_get_serial_baud_index(enum eSerialPort port);


/** Set the baud rate index for a serial port */
void db_set_serial_baud_index(enum eSerialPort port, enum eSerialBaudIndex baud_idx);


/** Convert database packed parity bits into real parity setting */
int8u db_get_serial_paritybits_from_bits(int8u bits);


/** Convert real parity setting to database packed parity bits */
int8u db_get_bits_from_serial_paritybits(int8u parity_setting);


/** Get real parity bits setting for a port from the database */
int8u db_get_serial_paritybits(enum eSerialPort port);


/** Convert database packed data bits into real data bits setting */
int8u db_get_serial_databits_from_bits(int8u bits);


/** Convert real data bits setting into database packed data bits */
int8u db_get_bits_from_serial_databits(int8u data_bits_setting);


/** Get real data bits setting for a port from the database */
int8u db_get_serial_databits(enum eSerialPort port);


/** Convert database packed stop bits into real stop bits setting */
int8u db_get_serial_stopbits_from_bits(int8u bits);


/** Convert real stop bits setting into database packed stop bits */
int8u db_get_bits_from_serial_stopbits_from_bits(int8u stop_bits_setting);


/** Get stopbits for a port from the database */
int8u db_get_serial_stopbits(enum eSerialPort port);


/** Get real use cts setting for a port from the database */
bool db_get_serial_usects(enum eSerialPort port);


/** Get real use terminations setting for a port from the database */
bool db_get_serial_useterm(enum eSerialPort port);


/** Get real use RS232 or RS422 setting for a port from the database */
bool db_get_serial_users232(enum eSerialPort port);


/** Get the serial port configuration for port from the database */
void db_get_serialconfig(enum eSerialPort port, t_ioctl_siocfg *siocfg, bool *use_rs232);


/** Get the default serial port configuration */
void db_get_serialconfig_default(t_ioctl_siocfg *siocfg, bool *use_rs232);


/**
 * Get serial port type from database.
 *
 * Only works for port 0..SERPORT_NETWORK_START - 1 and SERPORT_NETWORK_FIRST..SERPORT_NETWORK_FINAL
 * SERPORT_NETWORK_START..SERPORT_NUM are created from the listening TCP sockets, so get their s_comms_thread->devicetype
 * from the thread that created them.
 *
 * See @ref comms_task_list_get_port_type() for the type of a running serial port task
 */
enum eSerialPortType db_get_serial_type(enum eSerialPort port);


/** Get the lower limit on serial port type for a particular port */
enum eSerialPortType db_get_serial_type_min(enum eSerialPort port);


/** Get the upper limit on serial port type for a particular port */
enum eSerialPortType db_get_serial_type_max(enum eSerialPort port);


/**
 * Set serial port type in the database.
 *
 * Only works for port 0..SERPORT_NETWORK_START - 1 and SERPORT_NETWORK_FIRST..SERPORT_NETWORK_FINAL
 * SERPORT_NETWORK_START..SERPORT_NUM are created from the listening TCP sockets, so get their s_comms_thread->devicetype
 * from the thread that created them.
 */
void db_set_serial_type(enum eSerialPort port, enum eSerialPortType porttype);


/** Force the serial settings for any attached rinlinks to the rinlink fixed settings */
void db_force_rinlink_serial_settings(void);


#endif /* DB_SERIAL_H_ */
