/*
 * dbprint.c
 *
 * Print out contents of the user database in human readable form
 */
#include <string.h>
#include "db_build.h"
#include "menus.h"
#include "app_db_factory.h"
#include "buzz5000.h"
#include "iface_sd.h"
#include "comms_print_trigger.h"
#include "version.h"
#include "iface_menus.h"
#include "commonstrings.h"
#include "comms_escape.h"

#include "dbprint.h"


struct s_config_print
{
  void *client_data;
  void (*config_print_string)(struct s_config_print *cp, const char *str);
};


#define CRLF_STR "\r\n"

static const char crlf_str[] = CRLF_STR;
static const char sep_str[] = ": ";


/** Print a string using the supplied character print function */
static void print_db_str(struct s_config_print *cp, const char *str)
{
  if (NULL != cp && NULL != cp->config_print_string && NULL != str)
    cp->config_print_string(cp, str);
}


static void print_db_str_indented(struct s_config_print *cp, const char *s, int indent_level)
{
  #define INDENT_BUF_LEN 50
  size_t i;
  char indent_buf[INDENT_BUF_LEN + 1];

  /* Indent */
  i = indent_level * 5;
  if (i > INDENT_BUF_LEN)
    i = INDENT_BUF_LEN;
  memset(indent_buf, ' ', i);
  indent_buf[i] = '\0';
  print_db_str(cp, indent_buf);

  print_db_str(cp, s);
}


/** Print a string database item */
static void print_db_item_str(struct s_config_print *cp, const char *item, const char *setting)
{
  print_db_str(cp, item);
  print_db_str(cp, sep_str);
  print_db_str(cp, setting);
  print_db_str(cp, crlf_str);
}


/** DB print for menu item with no setting (like FAC.DFT) */
void to_string_no_setting(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", "No setting");
}


/** DB print for a weight menu item */
void to_string_weight(const AdcRdg *rdg, bool print_units, char *buffer, size_t buflen)
{
  int divisor;
  int i;
  long remainder;
  long quotient;
  char sign;
  const char *fmtstr;
  long num;
  int dp;
  enum eBuildUnits we_units;
  const char *unitsstr;

  num = rdg->rdg;
  dp = rdg->info.unitdp.bit.dp;

  /* Work out sign character */
  if (num < 0)
  {
    num = -num;
    sign = '-';
  }
  else
  {
    sign = ' ';
  }

  /* Work out decimal point divisor */
  for (i = 0, divisor = 1; i < dp; i++)
  {
    divisor *= 10;
  }

  /* Work out number split at decimal point */
  quotient = num / divisor;
  remainder = num % divisor;

  if (print_units)
  {
    if (ADC_UNITS_USER7 == rdg->info.unitdp.bit.units)
    {
      unitsstr = "mV/V";
    }
    else
    {
      we_units = xlat_adc_library_units_to_we2111(rdg->info.unitdp.bit.units);
      unitsstr = build_units_print[we_units];
    }
  }
  else
  {
    unitsstr = "";
  }

  /* Print the number with or without decimal point */
  if (dp > 0)
  {
    fmtstr = "%c%ld.%0*lu %s";
    snprintf(buffer, buflen, fmtstr, sign, quotient, dp, remainder, unitsstr);
  }
  else
  {
    fmtstr = "%c%ld %s";
    snprintf(buffer, buflen, fmtstr, sign, quotient, unitsstr);
  }
}


#define SNPRINTF_CHECK(ret, bp, len) \
  if (ret < 0) \
    return; \
  bp += ret; \
  len -= ret


/** Convert a print/autooutput format to a string */
void to_string_print_fmt(const char *fmt, char *buffer, size_t buflen)
{
  char temp_str[10];
  char *bufptr;
  int ret;
  ssize_t len;
  size_t count;

  if ('\0' == *fmt)
  {
    ret = snprintf(buffer, buflen, "<<empty>>");
  }
  else
  {
    bufptr = buffer;
    len = buflen;

    for (count = 0; *fmt != '\0' && len > 0; fmt++, count++)
    {
      if ((count & 0x0F) == 0)
      {
        ret = snprintf(bufptr, buflen, "" CRLF_STR);
        SNPRINTF_CHECK(ret, bufptr, len);
      }
      ret = snprintf(bufptr, len, "%s", comms_escape_char(temp_str, *fmt));
      SNPRINTF_CHECK(ret, bufptr, len);
    }
  }
}


static void print_menu_level(struct s_config_print *cp, bool full, const struct s_menu_item *menu, int indent_level)
{
  #define PRINT_SETTING_BUF_LEN 500 /* Needs to fit entire ticket formats */
  char print_setting_buf[PRINT_SETTING_BUF_LEN];
  size_t i;
  void *user_data;
  bool can_run = false;

  for (i = 0; menu->sub_menu[i] != NULL; i++)
  {
    print_db_str_indented(cp, menu->sub_menu[i]->title, indent_level);

    user_data = (menu->sub_menu[i]->user_data_from_menu_above) ? menu->user_data : menu->sub_menu[i]->user_data;

    if (NULL != menu->sub_menu[i]->menu_enter_exit)
    {
      can_run = menu->sub_menu[i]->menu_enter_exit(true, full, user_data, false);
    }
    else
    {
      can_run = true;
    }

    if (can_run && NULL != menu->sub_menu[i]->to_string_fn)
    {
      print_db_str(cp, ": ");
      menu->sub_menu[i]->to_string_fn(full, user_data, i, print_setting_buf, PRINT_SETTING_BUF_LEN);
      print_db_str(cp, print_setting_buf);
    }
    print_db_str(cp, CRLF_STR);

    if (can_run && NULL != menu->sub_menu[i]->sub_menu)
    {
      print_menu_level(cp, full, menu->sub_menu[i], indent_level + 1);
    }
  }
}


/** Does the work of printing the database */
static void print_db_internal(struct s_config_print *cp, bool full)
{
  /*
   * Header
   */
  print_db_str(cp, CRLF_STR CRLF_STR);
  print_db_str(cp, db_app_factory.model_str);
  print_db_str(cp, " Digital Setup" CRLF_STR);
  print_db_item_str(cp, "Software Version", app_version_str);
  print_db_str(cp, CRLF_STR);

  print_menu_level(cp, full, &menu_top_level, 0);

  print_db_str(cp, CRLF_STR "--- End of Settings ---" CRLF_STR);
}

/** Used by comms_print_db() to print a string out the printer port */
static void print_db_comms_print_string(struct s_config_print *config_print, const char *str)
{
  struct s_comms_print *cp;

  if (NULL != config_print && NULL != config_print->client_data && NULL != str)
  {
    cp = (struct s_comms_print *) config_print->client_data;
    if (NULL != cp->serdev)
      rwrite(cp->serdev, str, strlen(str), 0);
  }
}


/** Does the work of printing the database. Called from the comms printer thread. Don't call directly */
void comms_print_db(struct s_comms_print *cp, bool full)
{
  struct s_config_print config_print;

  config_print.client_data = cp;
  config_print.config_print_string = print_db_comms_print_string;

  print_db_internal(&config_print, full);
}


/** Print out the entire configuration for the WE2111. This will be sent to the attached printer, or generate an error */
void print_current_config(bool full)
{
  if (!comms_print_system_has_printer())
  {
    beep_long();
    prompt(no_str, "PRINTR");
  }
  else
  {
    comms_print_trigger_nontraceable_database_config(full);
  }
}


/** Used by print_db_to_file() to print a string to a file */
static void print_db_file_print_string(struct s_config_print *config_print, const char *str)
{
  FILE *f;

  if (NULL != config_print && NULL != config_print->client_data && NULL != str)
  {
    f = (FILE *) config_print->client_data;
    fputs(str, f);
  }
}


/** Print the database in human readable form to a file */
void print_db_to_file(FILE *f, bool full)
{
  struct s_config_print config_print;

  config_print.client_data = f;
  config_print.config_print_string = print_db_file_print_string;

  print_db_internal(&config_print, full);
}
