/*
 * device_finder.c
 *
 *  Created on: 27/07/2011
 *      Author: ian
 */
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <ezxml.h>
#include "we_debug.h"

#include "device_finder.h"


/** Create XML containing necessary information for the device finder */
char *create_device_finder_xml(const char *class, const char *name, const char *serialno,
                               const char *version, const char *customid, const char *macaddr,
                               const char *tcpport, const char *hostname)
{
  ezxml_t root_xml;
  ezxml_t dev_xml;
  ezxml_t elem_xml;
  char *xml_string;

  /* Default to error */
  xml_string = NULL;

  /* Create XML root */
  root_xml = ezxml_new(DEVICE_FINDER_XML_ROOT);
  if (NULL == root_xml)
  {
    rdebug_printf(LOG_ERR, "Error creating device finder XML root node\n");
    goto cdfx_done;
  }

  /* Add device element */
  dev_xml = ezxml_add_child(root_xml, DEVICE_FINDER_XML_DEVICE, 0);
  if (NULL == dev_xml)
  {
    rdebug_printf(LOG_ERR, "Error creating device finder XML device node\n");
    goto cdfx_cleanup;
  }

  /* Add class */
  elem_xml = ezxml_add_child(dev_xml, DEVICE_FINDER_XML_CLASS, 0);
  (void) ezxml_set_txt_d(elem_xml, class);

  /* Add name */
  elem_xml = ezxml_add_child(dev_xml, DEVICE_FINDER_XML_NAME, 0);
  (void) ezxml_set_txt_d(elem_xml, name);

  /* Add serial number */
  elem_xml = ezxml_add_child(dev_xml, DEVICE_FINDER_XML_SERIALNUMBER, 0);
  (void) ezxml_set_txt_d(elem_xml, serialno);

  /* Add version */
  elem_xml = ezxml_add_child(dev_xml, DEVICE_FINDER_XML_VERSION, 0);
  (void) ezxml_set_txt_d(elem_xml, version);

  /* Add custom id */
  elem_xml = ezxml_add_child(dev_xml, DEVICE_FINDER_XML_CUSTOMID, 0);
  (void) ezxml_set_txt_d(elem_xml, customid);

  /* Add MAC address */
  elem_xml = ezxml_add_child(dev_xml, DEVICE_FINDER_XML_MACADDRESS, 0);
  (void) ezxml_set_txt_d(elem_xml, macaddr);

  /* Add tcp port for connections */
  elem_xml = ezxml_add_child(dev_xml, DEVICE_FINDER_XML_TCPPORT, 0);
  (void) ezxml_set_txt_d(elem_xml, tcpport);

  /* Add hostname */
  elem_xml = ezxml_add_child(dev_xml, DEVICE_FINDER_XML_HOSTNAME, 0);
  (void) ezxml_set_txt_d(elem_xml, hostname);

  /* Convert to string */
  xml_string = ezxml_toxml(root_xml);
  if (NULL == xml_string)
  {
    rdebug_printf(LOG_ERR, "Error converting XML to string\n");
  }

  /* Success */

cdfx_cleanup:
  /* Clean up */
  ezxml_free(root_xml);

cdfx_done:
  return xml_string;
}


/** Create XML file containing necessary information for the device finder */
int write_device_finder_xml(const char *filename, const char *xml)
{
  struct flock fl;
  int fd;
  size_t len;
  int res = -1;

  /* Open file */
  fd = open(filename, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  if (fd < 0)
  {
    rdebug_printf(LOG_ERR, "write_device_finder_xml: error opening device finder file \"%s\"\n", filename);
    goto wdfx_done;
  }
  if (fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "write_device_finder_xml: error setting FD_CLOEXEC on device finder file \"%s\" (errno %d)\n", filename, errno);
    goto wdfx_close;
  }

  /* Set write lock over entire file (block until granted) */
  fl.l_type = F_WRLCK;
  fl.l_whence = SEEK_SET;
  fl.l_start = 0;
  fl.l_len = 0;
  fl.l_pid = getpid();
  if (fcntl(fd, F_SETLKW, &fl) < 0)
  {
    rdebug_printf(LOG_ERR, "Error locking device finder file \"%s\"\n", filename);
    goto wdfx_close;
  }

  /* Truncate file to 0 bytes */
  if (ftruncate(fd, 0) < 0)
  {
    rdebug_printf(LOG_ERR, "Error truncating device finder file \"%s\"\n", filename);
    goto wdfx_unlock;
  }

  /* Update file */
  if (lseek(fd, 0, SEEK_SET) != 0)
  {
    rdebug_printf(LOG_ERR, "Error seeking start of device finder file \"%s\"\n", filename);
    goto wdfx_unlock;
  }
  len = strlen(xml);
  if ((ssize_t) len != write(fd, xml, len))
  {
    rdebug_printf(LOG_ERR, "Error writing to device finder file \"%s\"\n", filename);
    goto wdfx_unlock;
  }

  /* Success */
  res = 0;

wdfx_unlock:
  /* Clear write lock over entire file */
  fl.l_type = F_UNLCK;
  fl.l_whence = SEEK_SET;
  fl.l_start = 0;
  fl.l_len = 0;
  fl.l_pid = getpid();
  (void) fcntl(fd, F_SETLK, &fl); /* Not much we can do if it fails */

wdfx_close:
  close(fd);

wdfx_done:
  return res;
}
