/*
 * dn.c
 *
 * This is like a mini daemon that mdev starts up when
 * a device notification comes through before the
 * WE application has started. It simply sits in memory
 * and buffers device notifications until the WE
 * application appears, and then sends them on.
 *
 * If the WE application is running, then notifications
 * aren't buffered, but sent straight on
 *
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdbool.h>


/*
 * Debug
 */


#define DEBUG_DEVICE_NOTIFY 0


/**
 * When device notify first starts, syslogd may not have started.
 * If you enable DEBUG_SYSLOG_BUFFERING, debug prints will be buffered
 * until such time as the syslog fifo appears.
 *
 * This adds some extra code, and memory allocation to store the
 * messages until they can be expunged to syslog.
 *
 * Even then it will only log the daemon messages, not the
 * messages from devicenotify running as an application
 *
 * Messy...
 */
#define DEBUG_SYSLOG_BUFFERING 1


#if DEBUG_DEVICE_NOTIFY
  #define DEBUGF(level, format, args...) log_printf(level, format, ##args)
#else
  #define DEBUGF(level, format, args...)
#endif


/*
 * Defines
 */


extern char **environ;


/** Name of program output in debug/error messages */
#define PROGRAM_NAME "devicenotify"


/** Buffer malloc size increments */
#define BUFFER_INCREMENTS 1024


/** Buffer maximum size (don't malloc above this) */
#define BUFFER_MAX_SIZE 65536


/** Maximum number of connected applications */
#define MAX_APPLICATIONS 10


/*
 * Command line options
 */


/** Show help info */
static int show_help = 0;


/** Run as a daemon */
static int daemonise = 0;


/** Application fifo filename (daemon buffers, then writes to this when it appears) */
static char *appfifo_filename[MAX_APPLICATIONS] = { NULL };
static int appfifo_counter = 0;


/** Daemon fifo filename (device notifier writes to this fifo, that daemon reads and buffers until application appears) */
static char *daemonfifo_filename = NULL;


/** Error messages go to syslog */
static bool log_to_syslog = true;


/** Error messages go to stderr */
static bool log_to_stderr = false;


/** Stderr debug level */
static int debug_level = LOG_WARNING;


/*
 * Code
 */


#if DEBUG_DEVICE_NOTIFY

#if DEBUG_SYSLOG_BUFFERING


struct s_syslog_buffer
{
  int level;
  char *message;
  struct s_syslog_buffer *next;
};

static struct s_syslog_buffer *syslog_buffer_first = NULL;
static struct s_syslog_buffer *syslog_buffer_last = NULL;

static bool syslog_fifo_has_appeared = false;


/** Does the syslog fifo exist yet (/dev/log)? */
static bool syslog_fifo_present(void)
{
  struct stat statbuf;
  bool result;

  result = (stat("/dev/log", &statbuf) == 0);
  return result;
}


static void syslog_dump_buffer(void)
{
  struct s_syslog_buffer *sb;
  struct s_syslog_buffer *old_sb;
  unsigned i;

  syslog(LOG_INFO, "syslog has started, dumping buffered messages:\n");
  for (i = 0, sb = syslog_buffer_first; NULL != sb; i++)
  {
    /* Log message */
    syslog(sb->level, "buffered message %u: %s\n", i, sb->message);

    /* Free up as we go along */
    old_sb = sb;
    sb = sb->next;

    free(old_sb->message);
    old_sb->message = NULL;
    old_sb->next = NULL;
    free(old_sb);
    old_sb = NULL;
  }

  syslog_buffer_first = NULL;
  syslog_buffer_last = NULL;
}


static void syslog_put_buffer(int level, const char *format, va_list args)
{
  struct s_syslog_buffer *nsb = NULL;

  /* Allocate new buffer */
  nsb = (struct s_syslog_buffer *) calloc(1, sizeof(struct s_syslog_buffer));
  if (NULL == nsb)
  {
    goto spb_fail;
  }

  /* printf allocate message */
  nsb->level = level;
  nsb->message = NULL;
  nsb->next = NULL;
  if (vasprintf(&nsb->message, format, args) < 0)
  {
    goto spb_fail;
  }

  /* Put into linked list */
  if (NULL != syslog_buffer_last)
    syslog_buffer_last->next = nsb;
  else
    syslog_buffer_last = nsb;
  if (NULL == syslog_buffer_first)
    syslog_buffer_first = nsb;

  return;

spb_fail:
  /* Clean up on failure */
  if (NULL != nsb)
  {
    if (NULL != nsb->message)
    {
      free(nsb->message);
      nsb->message = NULL;
    }
    free(nsb);
    nsb = NULL;
  }
}


static void syslog_buffer_vprintf(int level, const char *format, va_list args)
{
  if (!syslog_fifo_has_appeared)
  {
    if (!syslog_fifo_present())
    {
      /* No syslog fifo yet, buffer syslog printout */
      syslog_put_buffer(level, format, args);
    }
    else
    {
      /* Syslog fifo has just appeared, dump all buffered data to syslog */
      syslog_fifo_has_appeared = true;
      openlog(PROGRAM_NAME, LOG_PID, LOG_USER);
      syslog_dump_buffer();

      /* And log current message normally */
      goto sbv_log_normal;
    }
  }
  else
  {
sbv_log_normal:
    /* Syslog fifo is already present, just log straight to syslog */
    vsyslog(level, format, args);
  }
}


#else /* #if DEBUG_SYSLOG_BUFFERING */


static void syslog_buffer_vprintf(int level, const char *format, va_list args)
{
  /* Syslog buffering is not enabled, just log straight to syslog. Messages logged before syslog appears will be lost */
  vsyslog(level, format, args);
}


#endif /* #else -> #if DEBUG_SYSLOG_BUFFERING */

#else /* #else -> #if DEBUG_DEVICE_NOTIFY */


static void syslog_buffer_vprintf(int level, const char *format, va_list args)
{
  /* Debugging is not enabled, but we need this for error messages */
  vsyslog(level, format, args);
}


#endif /* #if DEBUG_DEVICE_NOTIFY */


/** Print logging information to the appropriate place (syslog or stderr) */
static void log_printf(int level, const char *format, ...)
{
  va_list args;
  va_list argssyslog;

  va_start(args, format);
  va_copy(argssyslog, args);

  if (log_to_syslog)
  {
    syslog_buffer_vprintf(level, format, argssyslog);
  }

  if (log_to_stderr && level >= debug_level)
  {
    (void) vfprintf(stderr, format, args);
  }

  va_end(argssyslog);
  va_end(args);
}


/** Write all environment variables to fifo */
static int write_environment_to_fifo(int fd)
{
  char **eptr;
  const char nullchar = '\0';
  const char lfchar = '\n';
  size_t len;

  /* TODO make single write */

  /* Write '\0' to start with, just to sync things up */
  if (sizeof(nullchar) != write(fd, &nullchar, sizeof(nullchar)))
  {
    return -1;
  }

  /* Write each environment variable (var=value), followed by LF */
  for (eptr = environ; NULL != *eptr; eptr++)
  {
    len = strlen(*eptr);
    if ((ssize_t) len != write(fd, *eptr, len))
    {
      return -1;
    }
    if (sizeof(lfchar) != write(fd, &lfchar, sizeof(lfchar)))
    {
      return -1;
    }
  }

  /* Terminate with '\0' */
  if (sizeof(nullchar) != write(fd, &nullchar, sizeof(nullchar)))
  {
    return -1;
  }

  /* Success */
  return 0;
}


/** Add new data to buffer, and increase buffer size as required */
static int add_to_buffer(char **buffer, size_t *bufferlen, size_t *bufferpos, char data)
{
  char *newbuf;

  if (NULL == buffer)
  {
    return -1;
  }

  if (NULL == *buffer)
  {
    /* First allocation of buffer */
    *buffer = malloc(BUFFER_INCREMENTS);
    if (NULL == *buffer)
    {
      return -1;
    }
    *bufferlen = BUFFER_INCREMENTS;
    *bufferpos = 0;
  }
  else if (*bufferpos >= *bufferlen)
  {
    /* Resize buffer */
    if (*bufferlen >= BUFFER_MAX_SIZE)
    {
      log_printf(LOG_WARNING, PROGRAM_NAME ": buffer maximum size reached (%u)\n", BUFFER_MAX_SIZE);
    }
    else
    {
      newbuf = realloc(*buffer, *bufferlen + BUFFER_INCREMENTS);
      if (NULL == newbuf)
      {
        return -1;
      }
      *buffer = newbuf;
      *bufferlen += BUFFER_INCREMENTS;
    }
  } /* else will fit */

  /* Store data into buffer */
  if (*bufferpos < *bufferlen)
  {
    (*buffer)[*bufferpos] = data;
    (*bufferpos)++;
  }
  return 0;
}


static int set_blocking(int fd)
{
  int flags;

  /* Set back to blocking */
  flags = fcntl(fd, F_GETFL, 0);
  if (-1 == flags)
  {
    return -1;
  }
  flags &= ~O_NONBLOCK;
  if (-1 == fcntl(fd, F_SETFL, flags))
  {
    return -1;
  }
  return 0;
}


/** Buffer incoming data from the daemon fifo until application fifo appears */
static int buffer_daemon_fifo(void)
{
  int readfd = -1;
  int writefd = -1;
  int app_fd = -1;
  fd_set readfds;
  struct timeval tv;
  char *buffer = NULL;
  size_t bufferlen = 0;
  size_t bufferpos = 0;
  char data;
  int result = 1;
  int retval;
  bool appfifo_exists[MAX_APPLICATIONS];
  bool all_appfifos_present;
  int i;

  DEBUGF(LOG_DEBUG, PROGRAM_NAME ": daemon buffering\n");

  /* For each application fifo, see if it already exists, in which case we don't want to write to it */
  for (i = 0, all_appfifos_present = true; i < appfifo_counter; i++)
  {
    /* If the application fifo */
    app_fd = open(appfifo_filename[i], O_WRONLY | O_NONBLOCK);
    appfifo_exists[i] = (app_fd >= 0);
    all_appfifos_present = all_appfifos_present & appfifo_exists[i];
    close(app_fd);
  }

  /* Open the daemon fifo for reading */
  readfd = open(daemonfifo_filename, O_RDONLY | O_NONBLOCK);
  if (readfd < 0)
  {
    log_printf(LOG_ERR, PROGRAM_NAME ": can't open daemon fifo '%s' for read\n", daemonfifo_filename);
    goto bdf_fail;
  }

  /* Set back to blocking */
  if (set_blocking(readfd) < 0)
  {
    log_printf(LOG_ERR, PROGRAM_NAME ": can't set daemon fifo '%s' blocking\n", daemonfifo_filename);
    goto bdf_fail;
  }

  /*
   * Now open the fifo for writing. We never write to it
   * however, unless the write end is open, select will fail.
   * So we open it here, but never write to it, as other
   * processes do this
   */
  writefd = open(daemonfifo_filename, O_WRONLY);
  if (writefd < 0)
  {
    log_printf(LOG_ERR, PROGRAM_NAME ": can't open daemon fifo '%s' for write\n", daemonfifo_filename);
    goto bdf_fail;
  }

  /* Now buffer any incoming data until all application fifos are present */
  for (; !all_appfifos_present; )
  {
    /* Wait 1 second for incoming data or timeout */
    FD_ZERO(&readfds);
    FD_SET(readfd, &readfds);
    tv.tv_sec = 1;
    tv.tv_usec = 0;
    retval = select(readfd + 1, &readfds, NULL, NULL, &tv);
    if (retval < 0)
    {
      /* Fail */
      log_printf(LOG_ERR, PROGRAM_NAME ": buffering select() fail\n");
      break;
    }
    else if (retval != 0)
    {
      /* Data available, read incoming data and buffer */
      if (FD_ISSET(readfd, &readfds))
      {
        if (sizeof(data) != read(readfd, &data, sizeof(data)))
        {
          /* Fail */
          log_printf(LOG_ERR, PROGRAM_NAME ": buffering read() fail\n");
          break;
        }
        if (add_to_buffer(&buffer, &bufferlen, &bufferpos, data) < 0)
        {
          /* Fail */
          log_printf(LOG_ERR, PROGRAM_NAME ": add to buffer fail\n");
          break;
        }
      }
    }
    else
    {

      /* See what application fifos have appeared */
      for (i = 0, all_appfifos_present = true; i < appfifo_counter; i++)
      {
        /* Only handle if the app fifo hasn't arrived yet */
        if (!appfifo_exists[i])
        {
          /* Timeout, see if application fifo has appeared yet */
          app_fd = open(appfifo_filename[i], O_WRONLY | O_NONBLOCK);
          if (app_fd >= 0)
          {
            /* It's here */
            DEBUGF(LOG_DEBUG, PROGRAM_NAME ": application fifo %s has arrived\n", appfifo_filename[i]);

            /* Write all buffered data to application fifo */
            if (app_fd >= 0)
            {
              if ((ssize_t) bufferlen == write(app_fd, buffer, bufferlen))
              {
                result = 0;
                appfifo_exists[i] = true;
              }
              else
              {
                log_printf(LOG_ERR, PROGRAM_NAME ": failed to write data to application fifo %s\n", appfifo_filename[i]);
                goto bdf_fail;
              }
            }
          }
        }

        /* Update all present flag */
        all_appfifos_present = all_appfifos_present & appfifo_exists[i];

      } /* for */
    }
  } /* while */

bdf_fail:
  DEBUGF(LOG_DEBUG, PROGRAM_NAME ": daemon finished buffering, result %d\n", result);
  /* Shut up shop */
  if (NULL != buffer)
    free(buffer);
  if (app_fd >= 0)
    close(app_fd);
  if (writefd >= 0)
    close(writefd);
  if (readfd >= 0)
    close(readfd);

  return result;
}


static int run_daemon(void)
{
  /* Do setup daemon stuff */
  setsid();
  chdir("/");
  umask(0);
  close(STDIN_FILENO);
#if (!DEBUG_DEVICE_NOTIFY)
  close(STDOUT_FILENO);
  close(STDERR_FILENO);
#endif

  /* Buffer data and write to application fifo when it appears */
  return buffer_daemon_fifo();
}


/** Run a new instance of ourselves as a daemon */
static int start_daemon(char *progname)
{
  char *daemon_args[4 + MAX_APPLICATIONS];
  int pid, i;

  /* Create child process */
  pid = vfork();
  if (pid < 0)
  {
    log_printf(LOG_ERR, PROGRAM_NAME ": vfork() failed\n");
    return -1;
  }
  else if (0 == pid)
  {
    /* Child process, setup arguments and execv */
    daemon_args[0] = progname;
    daemon_args[1] = "-d";
    daemon_args[2] = daemonfifo_filename;

    /* Iterate over each application and add it to the daemon_args */
    for (i = 0; i < appfifo_counter; i++)
      daemon_args[i + 3] = appfifo_filename[i];

    daemon_args[i + 3] = NULL;
    execv(daemon_args[0], daemon_args);

    /* Not reached */
    log_printf(LOG_ERR, PROGRAM_NAME ": execv() failed\n");
    _exit(1);
  }
  else
  {
    /* Parent process, nothing to do here */
  }
  return 0;
}


/** Deal with the situation of at least one application fifo not present yet */
static int no_app_fifo(char *progname)
{
  int fd;
  struct stat statbuf;

  DEBUGF(LOG_DEBUG, PROGRAM_NAME ": at least one application fifo isn't present, dealing with daemon\n");

  /*
   * Application fifo doesn't exist. This means WE application is not running
   * and therefore we should first see if the buffering daemon is running.
   * If it is, we write to it via its fifo.
   * If not, then we start the buffering daemon, and then write the data to it
   */

  /* First see if the daemon fifo exists, and create if not (may already exist, which is fine) */
  if (stat(daemonfifo_filename, &statbuf) < 0)
  {
    if (ENOENT == errno)
    {
      /* Doesn't exist, try creating it */
      if (mkfifo(daemonfifo_filename, 0600) < 0)
      {
        if (EEXIST != errno)
        {
          log_printf(LOG_ERR, PROGRAM_NAME ": mkfifo '%s' failed\n", daemonfifo_filename);
          return -1;
        }
      }
    }
    else
    {
      /* Other error, just quit */
      log_printf(LOG_ERR, PROGRAM_NAME ": stat '%s' failed\n", daemonfifo_filename);
      return -1;
    }
  }
  else
  {
    /* Already exists, make sure its a fifo */
    if (!S_ISFIFO(statbuf.st_mode))
    {
      /* Not a fifo */
      log_printf(LOG_ERR, PROGRAM_NAME ": '%s' is not a fifo\n", daemonfifo_filename);
      return -1;
    }
  }

  /*
   * Now try opening the daemon fifo to write to,
   * if the daemon is listening, then we know it's running, and can write to it
   */
  fd = open(daemonfifo_filename, O_WRONLY | O_NONBLOCK);
  if (fd < 0)
  {
    if (ENXIO == errno)
    {
      /* Other end not open, so start daemon */
      DEBUGF(LOG_DEBUG, PROGRAM_NAME ": no daemon, so starting one\n");
      if (start_daemon(progname) < 0)
      {
        return -1;
      }

      /* Now try to open again, this time blocking */
      fd = open(daemonfifo_filename, O_WRONLY);
      if (fd < 0)
      {
        /* Something bad happened */
        log_printf(LOG_ERR, PROGRAM_NAME ": error opening daemon fifo '%s'\n", daemonfifo_filename);
        return -1;
      }
    }
    else
    {
      /* Something bad happened */
      log_printf(LOG_ERR, PROGRAM_NAME ": error opening daemon fifo '%s'\n", daemonfifo_filename);
      return -1;
    }
  }
  else
  {
    /* It opened fine, which means the daemon is already running */
    /* Set back to blocking */
    DEBUGF(LOG_DEBUG, PROGRAM_NAME ": daemon already running\n");
    if (set_blocking(fd) < 0)
    {
      log_printf(LOG_ERR, PROGRAM_NAME ": failed to set blocking on daemon fifo '%s'\n", daemonfifo_filename);
      return -1;
    }
  }

  /* Write environment to waiting daemon */
  DEBUGF(LOG_DEBUG, PROGRAM_NAME ": writing data to daemon fifo\n");
  if (write_environment_to_fifo(fd) < 0)
  {
    log_printf(LOG_ERR, PROGRAM_NAME ": failed to write environment to daemon fifo '%s'\n", daemonfifo_filename);
    return -1;
  }
  close(fd);
  return 0;
}


/** Run in normal (non daemon) mode */
static int run_normal(char *progname)
{
  struct stat statbuf;
  int fd;
  int i;
  bool need_to_buffer;

  /** Iterate over all application fifos */
  for (i = 0, need_to_buffer = false; i < appfifo_counter; i++)
  {
    /* First, does the application fifo exist? If so, write directly to it */
    if (stat(appfifo_filename[i], &statbuf) < 0)
    {
      if (ENOENT == errno)
      {
        /* No application yet, start/continue buffering */
        need_to_buffer = true;
      }
      else
      {
        /* Other error, just quit */
        log_printf(LOG_ERR, PROGRAM_NAME ": stat '%s' failed\n", appfifo_filename[i]);
        return 1;
      }
    }
    else
    {
      /* Fifo does exist, check it's a fifo */
      if (S_ISFIFO(statbuf.st_mode))
      {
        /* Write environment to application fifo */
        DEBUGF(LOG_DEBUG, PROGRAM_NAME ": writing data to application fifo\n");
        fd = open(appfifo_filename[i], O_WRONLY);
        if (fd < 0)
        {
          log_printf(LOG_ERR, PROGRAM_NAME ": error opening application fifo '%s'\n", appfifo_filename[i]);
          return 1;
        }
        if (write_environment_to_fifo(fd) < 0)
        {
          log_printf(LOG_ERR, PROGRAM_NAME ": failed to write environment to application fifo '%s'\n", appfifo_filename[i]);
          close(fd);
          return 1;
        }
        close(fd);
      }
      else
      {
        /* Not a fifo */
        log_printf(LOG_ERR, PROGRAM_NAME ": '%s' is not a fifo\n", appfifo_filename[i]);
        return 1;
      }
    }
  }

  /* Need to buffer for at least one application fifo, so start daemon */
  if (need_to_buffer)
  {
    if (no_app_fifo(progname) < 0)
    {
      return 1;
    }
  }

  return 0;
}


/** Help */
static void print_usage(const char *prog)
{
  printf("Usage: %s [options] daemonfifo appfifos\n", prog);
  printf("  -d --daemon       run as a daemon (don't use this directly)\n"
         "  -h --help         this help\n"
         "  -s --syslog       log to syslog (enabled by default)\n"
         "  -e --stderr       log to stderr (disabled by default)\n"
         "  -l --debug-level  set debug level (as per LOG_ constants in syslog.h, default LOG_ERR)\n"
         "  daemonfifo is the daemon fifo to write device notifications to while the daemon is waiting for the application to start\n"
         "  appfifos are optional application fifos to write device notifications to. Between 1 and 10 may be used.\n"
  );
  printf("Rinstrum device arrival/removal notifier. This program is called by mdev\n"
         "to notify the main application of device arrival/removal. However, because\n"
         "mdev starts long before the main application, notifications sent during this\n"
         "period need to be buffered until the main application starts. This program does\n"
         "that by running as a daemon.\n");
}


/** Parse command line options */
static int parse_opts(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "daemon",      0, 0, 'd' },
    { "help",        0, 0, 'h' },
    { "syslog",      0, 0, 's' },
    { "stderr",      0, 0, 'e' },
    { "debug-level", 1, 0, 'l' },
    { NULL,          0, 0, '\0' },
  };
  int c;

  while (1)
  {
    c = getopt_long(argc, argv, "dhsel:", lopts, NULL);
    if (c == -1)
      break;

    switch (c)
    {
      /* Daemonise */
      case 'd':
        daemonise = 1;
        break;
      /* Show help */
      case 'h':
        show_help = 1;
        return 0;
      /* Log to syslog */
      case 's':
        log_to_syslog = true;
        break;
      /* Log to stderr */
      case 'e':
        log_to_stderr = true;
        break;
      /* Debug level */
      case 'l':
        debug_level = atoi(optarg);
        break;
      /* Anything else */
      default:
        log_printf(LOG_ERR, PROGRAM_NAME ": unknown command line option\n");
        return -1;
    }
  }

  if (optind >= argc)
  {
    log_printf(LOG_ERR, PROGRAM_NAME ": no daemon fifo specified\n");
    return -1;
  }
  daemonfifo_filename = argv[optind++];

  if (optind >= argc)
  {
    log_printf(LOG_ERR, PROGRAM_NAME ": no application fifo specified\n");
    return -1;
  }

  /* Iterate over the remaining arguments and add them to the appfifo list */
  for (appfifo_counter = 0;
       appfifo_counter < MAX_APPLICATIONS && optind < argc;
       appfifo_counter++, optind++)
  {
    appfifo_filename[appfifo_counter] = argv[optind];
  }

  /** If there are leftover arguments, throw an error */
  if (optind < argc)
  {
    log_printf(LOG_ERR, PROGRAM_NAME ": too many arguments\n");
    return -1;
  }

  return 0;
}


/** Entry point */
int main(int argc, char *argv[])
{
  int result;

  /* Parse command line */
  if (parse_opts(argc, argv) < 0)
  {
    return 1;
  }

  /* Show help */
  if (show_help)
  {
    print_usage(argv[0]);
    return 0;
  }

  /* Setup syslog */
  if (log_to_syslog)
  {
    openlog(PROGRAM_NAME, LOG_PID, LOG_USER);
  }

  /* Run as daemon or normal (based on command line options) */
  if (daemonise)
  {
    DEBUGF(LOG_DEBUG, PROGRAM_NAME ": running as daemon\n");
    result = run_daemon();
#if DEBUG_DEVICE_NOTIFY && DEBUG_SYSLOG_BUFFERING
    /* Clean/free up any buffered debug messages */
    syslog_dump_buffer();
#endif
  }
  else
  {
    DEBUGF(LOG_DEBUG, PROGRAM_NAME ": running normally\n");
    result = run_normal(argv[0]);
  }
  return result;
}
