/*
 * devicenotifyclient.c
 *
 * Example client for device notify
 *
 *  Created on: Jul 20, 2016
 *      Author: ian
 */
#define _XOPEN_SOURCE 700
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <getopt.h>
#include <sys/select.h>
#include <signal.h>
#include "../devmount/dm_envlist.h"


enum eDebugLevel
{
  dl_off = 0,
  dl_basic,
  dl_full,

  dl_num
};


static bool cmdline_showhelp = false;
static char *cmdline_fifo_filename = NULL;
static enum eDebugLevel cmdline_debuglevel = dl_off;


/** Maximum number of bytes held in the data buffer */
#define DEVICE_MOUNTER_MAX_BUFFER 2048


/** Device mounter structure */
struct s_devicemounter
{
  char *mdev_fifo_filename;               /*!< Filename of fifo that we read from to be informed of device arrival/removal */
  int mdev_fifo_fd;                       /*!< File descriptor of fifo that we read from to be informed of device arrival/removal */
  int mdev_fifo_write_fd;                 /*!< We also open the fifo for write, to keep select & read happy, but we never write to this */

  size_t mdev_data_buffer_max;            /*!< Maximum length of mdev data buffer */
  size_t mdev_data_buffer_length;         /*!< Current length of mdev data buffer */
  char *mdev_data_buffer;                 /*!< Data from mdev */
};


#define DM_DEBUG_OFF(lvl, format, args...)
#define DM_DEBUG_ON(lvl, format, args...) { if (cmdline_debuglevel >= lvl) fprintf(stdout, format, ##args); }

#define DM_DEBUG DM_DEBUG_ON


/** See if there are any clients waiting on this device event, and dispatch */
static int devicemounter_findclientmatches(struct s_devicemounter *dm)
{
  #define DEVICE_NAME_MAX_LEN 100
  char device_name[DEVICE_NAME_MAX_LEN];
  const char *action;

  /* Check for add or remove */
  if (0 == envlist_findequals(dm->mdev_data_buffer, "ACTION", "add"))
  {
    action = "added";
  }
  else if (0 == envlist_findequals(dm->mdev_data_buffer, "ACTION", "remove"))
  {
    action = "removed";
  }
  else if (0 != envlist_variable_exists(dm->mdev_data_buffer, "ACTION"))
  {
    /* If ACTION not defined, then assume it is add (occurs during startup with 'mdev -s') */
    action = "added";
  }
  else
  {
    /* Fail */
    fprintf(stderr, "devicemounter_findclientmatches: No ACTION found, or invalid ACTION\n");
    return -1;
  }

  /* Work out which device it is */
  if (0 != envlist_find(dm->mdev_data_buffer, "MDEV", device_name, DEVICE_NAME_MAX_LEN - 1))
  {
    /* Fail */
    fprintf(stderr, "devicemounter_findclientmatches: No device found\n");
    return -1;
  }

  printf("Device \"%s\" was %s.\n", device_name, action);
  return 0;
}


/** Parse the data read from the fifo into the buffer */
static int devicemounter_parseread(struct s_devicemounter *dm, const char *buffer, size_t bytes_read)
{
  int result = 0;
  size_t i;

  for (i = 0; i < bytes_read; i++)
  {
    if (dm->mdev_data_buffer_length >= dm->mdev_data_buffer_max)
    {
      /* Buffer full, scrap everything until a '\0' comes through */
      if ('\0' == buffer[i])
      {
        dm->mdev_data_buffer_length = 0;
      }
      result = 0;
    }
    else
    {
      /* Room in buffer */
      dm->mdev_data_buffer[dm->mdev_data_buffer_length++] = buffer[i];
      if ('\0' == buffer[i])
      {
        /* Make sure there is more than just a '\0' in the buffer */
        if (dm->mdev_data_buffer_length > 1)
        {
          DM_DEBUG(dl_basic, "devicemounter_getdata: got device event message of length %u\n", (unsigned) dm->mdev_data_buffer_length);
          DM_DEBUG(dl_basic, "devicemounter_getdata: data was:\n%s\n", dm->mdev_data_buffer);
          /* End of message, check for anyone listening */
          result = devicemounter_findclientmatches(dm);
        }

        /* Wait for next message */
        dm->mdev_data_buffer_length = 0;
      }
    } /* if */
  } /* for */
  return result;
}


/** Get data sent to us from mdev, and store into mdev_data_buffer */
static int devicemounter_getdata(struct s_devicemounter *dm)
{
  #define GET_DATA_BUF_SIZE 100
  int result = -1;
  char buffer[GET_DATA_BUF_SIZE];
  ssize_t bytes_read;

  /* Read chunks of data from the pipe */
  while (1)
  {
    bytes_read = read(dm->mdev_fifo_fd, buffer, GET_DATA_BUF_SIZE);
    if (bytes_read >= 0)
    {
      DM_DEBUG(dl_full, "devicemounter_getdata: read() %ld bytes\n", (long) bytes_read);
      result = devicemounter_parseread(dm, buffer, bytes_read);
    }
    else
    {
      if (EAGAIN == errno || EWOULDBLOCK == errno)
        result = 0;
      else
        fprintf(stderr, "devicemounter_getdata: read failed\n");
      break;
    }
  }

  return result;
}


/** Function to handle sigint */
static void sigint_handler(int sig)
{
  /* Do nothing */
}


/** Worker thread for the device mounter - looks for notifications of device arrival/removal, and dispatches accordingly */
static int devicemounter_listen(struct s_devicemounter *dm)
{
  fd_set readfds;
  int retval;
  int max_fd;
  struct sigaction sa;
  sigset_t emptyset;
  sigset_t blockset;
  int result;
  bool quit;

  if (NULL == dm)
  {
    fprintf(stderr, "devicemounter_listen: dm is NULL\n");
    result = -1;
    goto dml_done;
  }

  DM_DEBUG(dl_basic, "devicemounter_listen: started, dm: %p, fifo fd: %d\n",
           dm,
           dm->mdev_fifo_fd);

  /* Block SIGINT */
  sigemptyset(&blockset);
  sigaddset(&blockset, SIGINT);
  sigprocmask(SIG_BLOCK, &blockset, NULL);

  /* Establish signal handler */
  sa.sa_handler = sigint_handler;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGINT, &sa, NULL);

  /* pselect() loop */
  for (result = 0, quit = false; !quit; )
  {
    /* Setup file descriptors to pend on */
    FD_ZERO(&readfds);
    FD_SET(dm->mdev_fifo_fd, &readfds);
    max_fd = dm->mdev_fifo_fd;

    /* Wait for device appearance/disappearance data or data on quit pipe (which means quit) */
    DM_DEBUG(dl_full, "devicemounter_listen: waiting on data\n");

    /* Unblock signal, then wait for signal or ready file descriptor */
    sigemptyset(&emptyset);
    retval = pselect(max_fd + 1, &readfds, NULL, NULL, NULL, &emptyset);
    if (retval < 0)
    {
      if (EINTR == errno)
      {
        fprintf(stderr, "devicemounter_listen: ctrl-c pressed, quitting\n");
        result = 0;
      }
      else
      {
        /* Serious error (<0) */
        fprintf(stderr, "devicemounter_listen: select() error %d\n", errno);
        result = -1;
      }
      quit = true;
    }
    else if (0 == retval)
    {
      /* Timed out (=0) */
      DM_DEBUG(dl_full, "devicemounter_listen: select() timed out\n");
      continue;
    }
    else
    {

      /* Check data */
      if (FD_ISSET(dm->mdev_fifo_fd, &readfds))
      {
        /* Got data from the FIFO, which means something has appeared or disappeared, so see what's happened */
        DM_DEBUG(dl_full, "devicemounter_listen: device arrival/removal data available\n");
        if (devicemounter_getdata(dm) < 0)
        {
          result = -1;
          quit = true;
        }
      }
    }
  }

  /* Stuff gets closed/freed by the destroy function */
dml_done:
  DM_DEBUG(dl_basic, "devicemounter_listen: finished\n");
  return result;
}


/** Create a new device mounter, monitoring fifo_filename for device appearance/disappearance */
struct s_devicemounter *devicemounter_init(const char *fifo_filename)
{
  struct s_devicemounter *dm = NULL;
  bool madefifo = false;

  DM_DEBUG(dl_basic, "devicemounter_init: called\n");

  if (NULL == fifo_filename)
  {
    fprintf(stderr, "devicemounter_init: filename is NULL pointer\n");
    goto dmi_fail;
  }

  /* Allocate device mounter structure */
  dm = calloc(1, sizeof(struct s_devicemounter));
  if (NULL == dm)
  {
    fprintf(stderr, "devicemounter_init: malloc dm struct failed\n");
    goto dmi_fail;
  }
  dm->mdev_fifo_fd = -1;
  dm->mdev_fifo_write_fd = -1;

  /* Copy the filename */
  dm->mdev_fifo_filename = strdup(fifo_filename);
  if (NULL == dm->mdev_fifo_filename)
  {
    fprintf(stderr, "devicemounter_init: malloc filename failed\n");
    goto dmi_fail;
  }

  /* Create named fifo */
  if (mkfifo(dm->mdev_fifo_filename, 0600) < 0)
  {
    fprintf(stderr, "devicemounter_init: mkfifo failed\n");
    goto dmi_fail;
  }
  madefifo = true;
  dm->mdev_fifo_fd = open(dm->mdev_fifo_filename, O_RDONLY | O_NONBLOCK);
  if (dm->mdev_fifo_fd < 0)
  {
    fprintf(stderr, "devicemounter_init: open fifo for read failed\n");
    goto dmi_fail;
  }
  if (fcntl(dm->mdev_fifo_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    fprintf(stderr, "devicemounter_init: error setting FD_CLOEXEC on read end of fifo \"%s\" (errno %d)\n", dm->mdev_fifo_filename, errno);
    goto dmi_fail;
  }

  /*
   * Now open the fifo for writing. We never write to it
   * however, unless the write end is open, select will fail.
   * So we open it here, but never write to it, as other
   * processes do this
   */
  dm->mdev_fifo_write_fd = open(dm->mdev_fifo_filename, O_WRONLY);
  if (dm->mdev_fifo_write_fd < 0)
  {
    fprintf(stderr, "devicemounter_init: open fifo for write failed\n");
    goto dmi_fail;
  }
  if (fcntl(dm->mdev_fifo_write_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    fprintf(stderr, "devicemounter_init: error setting FD_CLOEXEC on write end of fifo \"%s\" (errno %d)\n", dm->mdev_fifo_filename, errno);
    goto dmi_fail;
  }

  /* Allocate the buffer */
  dm->mdev_data_buffer_max = DEVICE_MOUNTER_MAX_BUFFER;
  dm->mdev_data_buffer_length = 0;
  dm->mdev_data_buffer = malloc(dm->mdev_data_buffer_max);
  if (NULL == dm->mdev_data_buffer)
  {
    fprintf(stderr, "devicemounter_init: create buffer failed\n");
    goto dmi_fail;
  }

  DM_DEBUG(dl_full, "devicemounter_init: data fd: %d\n",
           dm->mdev_fifo_fd);

  /* Success */
  DM_DEBUG(dl_basic, "devicemounter_init: success\n");
  return dm;

dmi_fail:
  /* Failed, clean up */
  if (NULL != dm)
  {
    if (NULL != dm->mdev_fifo_filename)
    {
      if (madefifo)
        unlink(dm->mdev_fifo_filename);

      if (dm->mdev_fifo_fd > 0)
      {
        close(dm->mdev_fifo_fd);
        dm->mdev_fifo_fd = -1;
      }

      if (dm->mdev_fifo_write_fd > 0)
      {
        close(dm->mdev_fifo_write_fd);
        dm->mdev_fifo_write_fd = -1;
      }

      if (NULL != dm->mdev_data_buffer)
      {
        free(dm->mdev_data_buffer);
        dm->mdev_data_buffer = NULL;
      }

      free(dm->mdev_fifo_filename);
      dm->mdev_fifo_filename = NULL;
    }
    free(dm);
    dm = NULL;
  }
  return NULL;
}


/** Clean up device mounter */
int devicemounter_destroy(struct s_devicemounter *dm)
{
  DM_DEBUG(dl_basic, "devicemounter_destroy: called\n");

  if (NULL == dm)
  {
    return -1;
  }

  /* Free the buffer */
  free(dm->mdev_data_buffer);
  dm->mdev_data_buffer = NULL;

  /* Close fifo */
  unlink(dm->mdev_fifo_filename);
  close(dm->mdev_fifo_write_fd);
  dm->mdev_fifo_write_fd = -1;
  close(dm->mdev_fifo_fd);
  dm->mdev_fifo_fd = -1;

  /* De-allocate memory */
  free(dm->mdev_fifo_filename);
  dm->mdev_fifo_filename = NULL;
  free(dm);
  dm = NULL;

  DM_DEBUG(dl_basic, "devicemounter_destroy: finished\n");
  return 0;
}


/** Print out the command line options */
static void print_commandlineopts(const char *program_name)
{
  fprintf(stdout, "Usage: %s [-options] fifo\n"
                  "  -d <lvl>     --debug <lvl>      Set debug level 0..%u\n"
                  "  -h           --help             Print help\n",
          program_name, (unsigned) dl_num - 1);
}


/** Parse the command line options */
static int parse_commandline(int argc, char *argv[])
{
  int c;
  int result = 0;
  int param;
  static const char opts[] = "d:h";
  static const struct option lopts[] =
  {
    { "debug",         1, 0, 'd'  },
    { "help",          0, 0, 'h'  },
    { NULL,            0, 0, '\0' }
  };

  while (1)
  {
    c = getopt_long(argc, argv, opts, lopts, NULL);
    if (-1 == c)
    {
      break;
    }

    switch (c)
    {
      case 'd':
        param = atoi(optarg);
        if (param < 0 || param >= dl_num)
        {
          fprintf(stderr, "Debug level should be 0..%d\n", dl_num - 1);
          return -1;
        }
        cmdline_debuglevel = param;
        break;
      case 'h':
        cmdline_showhelp = true;
        return 0;
        break;
      default:
        fprintf(stderr, "Invalid command line option, try --help\n");
        return -1;
        break;
    }
  }

  /* Fifo to listen to */
  if (optind >= argc)
  {
    fprintf(stderr, "ERROR: no fifo specified, try --help\n");
    return -1;
  }
  cmdline_fifo_filename = argv[optind];
  optind++;

  return result;
}


int main(int argc, char *argv[])
{
  struct s_devicemounter *dm = NULL;
  int result;

  /* Parse command line */
  if (parse_commandline(argc, argv) < 0)
  {
    result = 1;
    goto m_fail;
  }

  /* Show help if requested */
  if (cmdline_showhelp)
  {
    print_commandlineopts(argv[0]);
    result = 0;
    goto m_fail;
  }

  /* Create device mounter */
  dm = devicemounter_init(cmdline_fifo_filename);
  if (NULL == dm)
  {
    result = 1;
    goto m_fail;
  }

  /* Get stuff from fifo */
  if (devicemounter_listen(dm) < 0)
  {
    result = 1;
    goto m_fail;
  }

  result = 0;

m_fail:
  /* Clean up */
  if (NULL != dm)
  {
    (void) devicemounter_destroy(dm);
    dm = NULL;
  }
  return result;
}
