/*
 * devices.c
 *
 *  Created on: Dec 12, 2011
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <rinhw/hwlib_init.h>
#include <rinhw/debug.h>
#include <rinhw/we21hw.h>
#include <rinhw/iic_local_linux.h>
#include <rinhw/kbd_linux_event.h>
#include <rinhw/sio_local_comport.h>
#include <rinhw/rtc_local.h>
#include <rinhw/adc_ads1216.h>
#include <rinhw/adc_ads1216_local.h>
#include <rinhw/adc_ads1216_we21.h>
#include <rinhw/buzzer_local.h>
#include <rinhw/buzzer_dummy.h>
#include <rinhw/kbd_local.h>
#include <rinremote/iic_remote.h>
#include <rinremote/adc_ads1216_remote.h>
#include <rinremote/remoteexec.h>
#include <rinblock/rb_we21_lcd_local_skt.h>
#include <rinblock/rb_we21_lcd_dummy.h>
#include <rinblock/rb_c500_lcd_dummy.h>
#include "we_debug.h"
#include "we_config.h"
#include "apperror.h"
#include "serial_usb.h"
#include "access.h"
#include "serial_ports.h"
#include "adc_8bit.h"
#include "indicator_model.h"
#include "indicator_model_features.h"
#include "cmdline.h"

#include "devices.h"


/** Buzzer device (located on display rinblock) */
struct t_io_dev dev_buz_we;


/** Primary keyboard device, (located on display rinblock) */
struct t_io_dev dev_kbd_we;


/** Second keyboard device, used for linux event devices (usb keyboards and the like) */
struct t_io_dev dev_kbd_event;


/** Real time clock device */
struct t_io_dev dev_rtc_we;


/** I2C device */
struct t_io_dev dev_iic0;


/** 24bit ADC device */
struct t_io_dev dev_adc24bit_we;


/** LCD device */
struct t_io_dev dev_lcd_we;


/** Create dummy devices when running headless */
static void create_dummy_devices(void)
{
  struct s_kbd_hw_definition dummy_kbd_hw_definition;

  /* Create a dummy buzzer (won't beep) */
  if (rio_create_device_buzzer_dummy(&dev_buz_we) < 0)
  {
    rdebug_printf(LOG_ERR, "Create dummy buzzer device failed\n");
  }

  /* Create a dummy LCD */
  if (rio_create_device_lcd_dummy_c500(&dev_lcd_we) < 0)
  {
    rdebug_printf(LOG_ERR, "Create dummy LCD device failed\n");
  }

  /* Create a dummy keyboard */
  memset(&dummy_kbd_hw_definition, 0x00, sizeof(dummy_kbd_hw_definition));
  dummy_kbd_hw_definition.hw_data = NULL;
  dummy_kbd_hw_definition.kbd_hw_init = NULL;
  dummy_kbd_hw_definition.kbd_hw_readkeybits = NULL;
  dummy_kbd_hw_definition.already_debounced = 1;
  dummy_kbd_hw_definition.kbd_hw_destroy = NULL;
  dummy_kbd_hw_definition.kbd_hw_readhikeybits = access_kbd_readhikeybits;
  if (rio_create_device_kbd_local(&dev_kbd_we, &dummy_kbd_hw_definition, &dev_buz_we) < 0)
  {
    rdebug_printf(LOG_ERR, "Create dummy keyboard device failed\n");
  }
}


/**
 * Create devices for the system when running normally that don't required database settings.
 *
 * Returns false on major problem (ie don't bother running).
 */
static bool init_app_devices_pre_db_normal(void)
{
  /* Start the hardware library */
  WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_pre_db_normal: initialise hardware library\n");
  rinhw_set_debug_printf_fn(rdebug_vprintf);
  hwlib_init();
  if (rio_init_we21hw(we_configuration.we2111_hardware_device) < 0)
  {
    rdebug_printf(LOG_ERR, "init_app_devices_pre_db_normal: WE2111 hardware access initialisation failed\n");
  }

  /* Create the devices TODO deal with failures */
  WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_pre_db_normal: create devices\n");

  /* Buzzer, Keyboard, LCD devices are created as part of the rinblocks */

  /* I2C device for rinblocks */
  if (rio_create_device_iic_local(&dev_iic0, we_configuration.rinblock_iic_device, we21hw_set_slot_select, NULL) < 0)
  {
    set_error(APP_ERR_MODULE, "init_app_devices_pre_db_normal: setting ERROR_MODULE because i2c device creation failed\n");
  }

  /* Linux keyboard event device */
  if (rio_create_device_kbd_linux_event(&dev_kbd_event, &dev_buz_we) < 0)
    rdebug_printf(LOG_ERR, "init_app_devices_pre_db_normal: linux event keyboard device creation failed\n");

  /* 24bit ADC device */
  if (rio_create_device_adc24bit_local(&dev_adc24bit_we, we_configuration.adc_spi_device,
                                       adc24bit_user_hw_init_we21,
                                       adc24bit_sync_spi_we21,
                                       NULL,
                                       adc24bit_wait_for_reading_we21) < 0)
  {
    set_error(ADC_ERR_HARDWARE, "init_app_devices_pre_db_normal: setting ADC_ERR_HARDWARE because 24bit adc device creation failed\n");
  }
  if (rio_create_device_rtc_local(&dev_rtc_we) < 0)
  {
    set_error(APP_ERR_RTC, "init_app_devices_pre_db_normal: setting APP_ERR_RTC because rtc device creation failed\n");
  }

  /* Setup local serial port devices */
  if (setup_linux_serial_gadget(&dev_ser7_usb) < 0)
  {
    rdebug_printf(LOG_ERR, "init_app_devices_pre_db_normal: usb gadget serial 7 setup fail\n");
  }

  return true;
}


/**
 * Create devices for the system when running normally that require database settings.
 *
 * Returns false on major problem (ie don't bother running).
 */
static bool init_app_devices_post_db_normal(enum e_indicator_model_sub_type imst)
{
  struct s_sio_local_rs422 sio_ser1_rs422;

  /* Initialise rinblocks, and create devices. This needs to happen early on so display works */
  if (!indicator_model_subtype_has_display(imst))
  {
    WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_post_db_normal: initialise rinblocks (with no display)\n");
    /* Don't create keyboard, LCD, front rinlink or buzzer when running headless */
    init_rinblocks(true, &dev_iic0, &dev_adc_rb, NULL, NULL, NULL, &dev_ser3_rb_rs422, &dev_ser3_rb_rs422b, &dev_ser4_rb_rs232, &dev_ser4_rb_rs232b, NULL);

    /* Create dummy versions of these devices */
    create_dummy_devices();
  }
  else
  {
    WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_post_db_normal: initialise rinblocks (with display)\n");
    /* Run normally with a display attached */
    init_rinblocks(false, &dev_iic0, &dev_adc_rb, &dev_kbd_we, &dev_lcd_we, &dev_buz_we, &dev_ser3_rb_rs422, &dev_ser3_rb_rs422b, &dev_ser4_rb_rs232, &dev_ser4_rb_rs232b, &dev_ser_front_rl);
  }

  /* Initialise onboard serial ports */
  if (indicator_model_subtype_has_rear_rinlink(imst))
  {
    WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_post_db_normal: initialise onboard serial 1 rinlink\n");
    /* Create serial port for rear rinlink */
    if (rio_create_device_sio_localcom(&dev_ser1_local, we_configuration.serial1_device, NULL, false) < 0)
    {
      rdebug_printf(LOG_ERR, "init_app_devices_pre_db_normal: onboard serial 1 (rinlink) setup fail\n");
    }
  }
  else if (indicator_model_subtype_has_db9_serial(imst))
  {
    WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_post_db_normal: initialise onboard serial 1 & 2 db9s\n");
    /* Create serial ports for DB9 ports */
    sio_ser1_rs422.hw_data = NULL;
    sio_ser1_rs422.rs422_mode = slrm_rs485_rts_deasserted;
    sio_ser1_rs422.sio_select_rs232_not_rs422 = we21hw_select_rs232_not_rs422;
    sio_ser1_rs422.sio_set_rs422_termination = we21hw_set_rs422_termination;
    sio_ser1_rs422.sio_set_rs422_transmitenable = NULL;
    if (rio_create_device_sio_localcom(&dev_ser1_local, we_configuration.serial1_device, &sio_ser1_rs422, false) < 0)
    {
      rdebug_printf(LOG_ERR, "init_app_devices_post_db_normal: onboard serial 1 (db9) setup fail\n");
    }
    if (rio_create_device_sio_localcom(&dev_ser2_local, we_configuration.serial2_device, NULL, false) < 0)
    {
      rdebug_printf(LOG_ERR, "init_app_devices_post_db_normal: onboard serial 2 (db9) setup fail\n");
    }
  }

  return true;
}


/** Set up the remote device communication channel */
static bool init_app_remote_device_comms(const char *remote_dev)
{
  char hostname[100];
  unsigned port;

  if (sscanf(remote_dev, "%99[^:]:%u", hostname, &port) == 2)
  {
    /* Initialise remote device comms using TCP host:port */
    WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_remote_device_comms: initialising remote device communications using host %s port %u\n", hostname, port);
    if (vRemoteExecuteInit(NULL, hostname, port) < 0)
    {
      rdebug_printf(LOG_ERR, "init_app_remote_device_comms: failed to initialise remote device communications using host %s port %u\n", hostname, port);
      return false;
    }
  }
  else
  {
    /* Initialise remote device comms using serial port */
    WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_remote_device_comms: initialising remote device communications using serial port %s\n", remote_dev);
    if (vRemoteExecuteInit(remote_dev, NULL, 0) < 0)
    {
      rdebug_printf(LOG_ERR, "init_app_remote_device_comms: failed to initialise remote device communications using serial port %s\n", remote_dev);
      return false;
    }
  }

  /* Success */
  return true;
}


/** Create devices for the system when running remotely. Returns false on major problem (ie don't bother running). */
static bool init_app_devices_remote(const char *remote_dev)
{
  /* struct s_sio_local_rs422 sio_ser1_rs422; */

  /* Start the hardware library */
  WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_remote: initialise hardware library\n");
  hwlib_init();
  rinhw_set_debug_printf_fn(rdebug_vprintf);

  /* Initialise remote device comms */
  if (!init_app_remote_device_comms(remote_dev))
  {
    return false;
  }

  /* Create the devices TODO deal with failures */
  WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_remote: create devices\n");

  /* Remote I2C device for rinblocks */
  if (rio_create_device_iic_remote(&dev_iic0, REMOTE_DEV_IIC0) < 0)
  {
    set_error(APP_ERR_MODULE, "init_app_devices_remote: setting APP_ERR_MODULE because i2c device creation failed\n");
  }

  /* Local Linux keyboard event device */
  if (rio_create_device_kbd_linux_event(&dev_kbd_event, &dev_buz_we) < 0)
    rdebug_printf(LOG_ERR, "init_app_devices_remote: Linux Event Keyboard device creation failed\n");

  /* Remote 24bit ADC device */
  if (rio_create_device_adc24bit_remote(&dev_adc24bit_we) < 0)
  {
    set_error(ADC_ERR_HARDWARE, "init_app_devices_remote: setting ADC_ERR_HARDWARE because 24bit adc device creation failed\n");
  }

  /* Local RTC */
  if (rio_create_device_rtc_local(&dev_rtc_we) < 0)
  {
    set_error(APP_ERR_RTC, "init_app_devices_remote: setting APP_ERR_RTC because rtc device creation failed\n");
  }

  /* Setup local serial port devices */
#if 0
  if (rio_create_device_sio_localcom(&dev_ser1_local, we_configuration.serial1_device, &sio_ser1_rs422) < 0)
  {
    rdebug_printf(LOG_ERR, "Local serial 1 setup fail\n");
  }
#endif
#if 0
  if (setup_linux_serial_gadget(&dev_ser7_usb) < 0)
  {
    rdebug_printf(LOG_ERR, "USB gadget serial 7 setup fail\n");
  }
#endif

  /* Initialise rinblocks, and create devices. This needs to happen early on so display works */
  WE_STARTUP_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "init_app_devices_remote: initialise rinblocks\n");
  init_rinblocks(true, &dev_iic0, &dev_adc_rb, NULL, NULL, NULL, &dev_ser3_rb_rs422, &dev_ser3_rb_rs422b, &dev_ser4_rb_rs232, &dev_ser4_rb_rs232b, NULL);

  /* Create dummy versions of the buzzer, keyboard and lcd */
  create_dummy_devices();

  return true;
}


/**
 * Create devices for the system that don't required any settings from the database.
 *
 * Returns false on major problem (ie don't bother running).
 */
bool init_app_devices_pre_db(void)
{
  bool result;

  if (NULL != cmdline_remote_device)
  {
    result = init_app_devices_remote(cmdline_remote_device);
  }
  else
  {
    result = init_app_devices_pre_db_normal();
  }
  return result;
}


/**
 * Create devices for the system that require settings from the database.
 *
 * Returns false on major problem (ie don't bother running).
 */
bool init_app_devices_post_db(enum e_indicator_model_sub_type imst)
{
  bool result = true;

  if (NULL == cmdline_remote_device)
  {
    result = init_app_devices_post_db_normal(imst);
  }
  return result;
}


/** Clean up devices */
void destroy_app_devices(void)
{
  WE_SHUTDOWN_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "destroy_app_devices called\n");

  /* Clean up rinblock devices */
  destroy_rinblocks();
  WE_SHUTDOWN_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "destroy_app_devices: rinblocks destroyed\n");

  /* Destroy devices */
  rdestroy_io(&dev_ser7_usb);
  rdestroy_io(&dev_ser1_local);
  rdestroy_io(&dev_rtc_we);
  rdestroy_io(&dev_adc24bit_we);
  rdestroy_io(&dev_kbd_event);
  rdestroy_io(&dev_iic0);
  WE_SHUTDOWN_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "destroy_app_devices: devices destroyed\n");

  /* If running remote, then shut that down */
  if (NULL != cmdline_remote_device)
  {
    vRemoteExecuteDestroy();
  }

  /* Shutdown the hardware library */
  (void) rio_destroy_we21hw();
  hwlib_destroy();

  WE_SHUTDOWN_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "destroy_app_devices finished\n");
}
