/*
 * devicemounter.c
 *
 * Handle the appearance and disappearance of linux devices
 */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rinthread.h>
#include "dm_envlist.h"
#include "we_debug.h"

#include "devicemounter.h"

#define DM_COMMAND_PIPE_RUN_CHAR 'r'
#define DM_COMMAND_PIPE_SHUTDOWN_CHAR 's'


/** Get the full device path for a device. It's up to the caller to free the returned string */
char *devicemounter_getfulldevicepath(const char *device)
{
  const char device_prefix[] = "/dev/";
  char *fullpath;

  fullpath = malloc(strlen(device_prefix) + strlen(device) + 1);
  if (NULL != fullpath)
  {
    strcpy(fullpath, device_prefix);
    strcat(fullpath, device);
  }
  return fullpath;
}


/** See if there are any clients waiting on this device event, and dispatch */
static int devicemounter_findclientmatches(struct s_devicemounter *dm)
{
  #define DEVICE_NAME_MAX_LEN 100
  char device_name[DEVICE_NAME_MAX_LEN];
  enum eDeviceMounterAction action;
  int i;

  /* Check for add or remove */
  if (0 == envlist_findequals(dm->mdev_data_buffer, "ACTION", "add"))
  {
    action = dmaAdd;
  }
  else if (0 == envlist_findequals(dm->mdev_data_buffer, "ACTION", "remove"))
  {
    action = dmaRemove;
  }
  else if (0 != envlist_variable_exists(dm->mdev_data_buffer, "ACTION"))
  {
    /* If ACTION not defined, then assume it is add (occurs during startup with 'mdev -s') */
    action = dmaAdd;
  }
  else
  {
    /* Fail */
    rdebug_printf(LOG_ERR, "devicemounter_findclientmatches: No ACTION found, or invalid ACTION\n");
    return -1;
  }

  /* Work out which device it is */
  if (0 != envlist_find(dm->mdev_data_buffer, "MDEV", device_name, DEVICE_NAME_MAX_LEN - 1))
  {
    /* Fail */
    rdebug_printf(LOG_ERR, "devicemounter_findclientmatches: No device found\n");
    return -1;
  }

  /* Now see who wants to know about it */
  for (i = 0; i < DEVICE_MOUNTER_MAX_CLIENTS; i++)
  {
    if (NULL != dm->clients[i])
    {
      if (1 == dm->clients[i]->match_fn(dm->clients[i]->client_data, dm->mdev_data_buffer))
      {
        /* Handle device event by this client */
        WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_findclientmatches: Found client match, notifying client\n");
        dm->clients[i]->callback_fn(dm->clients[i]->client_data, action, device_name, dm->mdev_data_buffer);
      }
    }
  }

  /* TODO Return code */
  return 0;
}


/** Parse the data read from the fifo into the buffer */
static int devicemounter_parseread(struct s_devicemounter *dm, const char *buffer, size_t bytes_read)
{
  int result = 0;
  size_t i;

  for (i = 0; i < bytes_read; i++)
  {
    if (dm->mdev_data_buffer_length >= dm->mdev_data_buffer_max)
    {
      /* Buffer full, scrap everything until a '\0' comes through */
      if ('\0' == buffer[i])
      {
        dm->mdev_data_buffer_length = 0;
        rdebug_printf(LOG_ERR, "devicemounter_parseread: data buffer is full, your system environment variables are larger than %lu\n", (unsigned long) dm->mdev_data_buffer_max);
        rdebug_printf(LOG_ERR, "devicemounter_parseread: try running devicenotify with less environment variables set, perhaps \"env -i devicenotify ...\"\n");
      }
      result = 0;
    }
    else
    {
      /* Room in buffer */
      dm->mdev_data_buffer[dm->mdev_data_buffer_length++] = buffer[i];
      if ('\0' == buffer[i])
      {
        /* Make sure there is more than just a '\0' in the buffer */
        if (dm->mdev_data_buffer_length > 1)
        {
          WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_getdata: got device event message of length %d\n", dm->mdev_data_buffer_length);
          WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_getdata: data was %s\n", dm->mdev_data_buffer);
          /* End of message, check for anyone listening */
          result = devicemounter_findclientmatches(dm);
        }

        /* Wait for next message */
        dm->mdev_data_buffer_length = 0;
      }
    } /* if */
  } /* for */
  return result;
}


/** Get data sent to us from mdev, and store into mdev_data_buffer */
static int devicemounter_getdata(struct s_devicemounter *dm)
{
  #define GET_DATA_BUF_SIZE 100
  int result = -1;
  char buffer[GET_DATA_BUF_SIZE];
  ssize_t bytes_read;

  /* Read chunks of data from the pipe */
  while (1)
  {
    bytes_read = read(dm->mdev_fifo_fd, buffer, GET_DATA_BUF_SIZE);
    if (bytes_read >= 0)
    {
      WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_getdata: read() %ld bytes\n", (long) bytes_read);
      result = devicemounter_parseread(dm, buffer, bytes_read);
    }
    else
    {
      if (EAGAIN == errno || EWOULDBLOCK == errno)
        result = 0;
      else
        rdebug_printf(LOG_ERR, "devicemounter_getdata: read failed\n");
      break;
    }
  }

  return result;
}


/** Deal with data from the command pipe. Returns -1 for error, else received command DM_THREAD_ */
static int command_pipe_handler(int cmdreadfd)
{
  char ch;

  /* Wait here until command comes through */
  if (sizeof(ch) != read(cmdreadfd, &ch, sizeof(ch)))
  {
    return -1;
  }

  /* Valid command? */
  if (DM_COMMAND_PIPE_RUN_CHAR == ch || DM_COMMAND_PIPE_SHUTDOWN_CHAR == ch)
  {
    return ch;
  }

  /* Invalid command */
  return -1;
}


/** Worker thread for the device mounter - looks for notifications of device arrival/removal, and dispatches accordingly */
static void *devicemounter_thread(void *data)
{
  struct s_devicemounter *dm = data;
  int datafd;
  int cmdreadfd;
  fd_set readfds;
  int retval;
  int max_fd;
  int cmd;

  if (NULL == dm)
  {
    goto dmt_done;
  }

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_thread: started, dm: %p, fifo fd: %d, shutdown fd: %d\n",
                                dm,
                                dm->mdev_fifo_fd,
                                dm->command_pipe_read_fd);

  /* Get file descriptors */
  pthread_mutex_lock(&dm->mutex);
  datafd = dm->mdev_fifo_fd;
  cmdreadfd = dm->command_pipe_read_fd;
  pthread_mutex_unlock(&dm->mutex);

  /* Wait here until told to start */
  cmd = command_pipe_handler(cmdreadfd);
  if (cmd < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_thread: command pipe error\n");
    goto dmt_done;
  }
  else if (DM_COMMAND_PIPE_SHUTDOWN_CHAR == cmd)
  {
    goto dmt_done;
  }
  /* else run */

  /* Thread loop */
  while (1)
  {
    /* Setup file descriptors to pend on */
    FD_ZERO(&readfds);
    FD_SET(datafd, &readfds);
    FD_SET(cmdreadfd, &readfds);
    max_fd = (datafd > cmdreadfd) ? datafd : cmdreadfd;

    /* Wait for device appearance/disappearance data or data on quit pipe (which means quit) */
    WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_thread: waiting on data\n");
    retval = select(max_fd + 1, &readfds, NULL, NULL, NULL);
    if (retval < 0)
    {
      /* Serious error (<0) */
      rdebug_printf(LOG_ERR, "devicemounter_thread: select() error %d\n", errno);
      continue;
    }
    else if (0 == retval)
    {
      /* Timed out (=0) */
      WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_thread: select() timed out\n");
      continue;
    }
    else
    {

      /* Check data */
      if (FD_ISSET(datafd, &readfds))
      {
        /* Got data from the FIFO, which means something has appeared or disappeared, so see what's happened */
        WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_thread: device arrival/removal data available\n");
        pthread_mutex_lock(&dm->mutex);
        (void) devicemounter_getdata(dm);
        pthread_mutex_unlock(&dm->mutex);
      }

      /* Check quit */
      if (FD_ISSET(cmdreadfd, &readfds))
      {
        /* Quit notification */
        cmd = command_pipe_handler(cmdreadfd);
        if (cmd < 0)
        {
          rdebug_printf(LOG_ERR, "devicemounter_thread: command pipe error\n");
          goto dmt_done;
        }
        else if (DM_COMMAND_PIPE_SHUTDOWN_CHAR == cmd)
        {
          WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_thread: quit notification\n");
          break;
        }
      }
    }
  }

  /* Stuff gets closed/freed by the destroy function */
dmt_done:
  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_thread: finished\n");
  rin_thread_exit(NULL);
}


/** Register a handler for a device */
int devicemounter_registerhandler(struct s_devicemounter *dm,
                                  t_dm_devevent_match_fn matchfn, t_dm_devevent_callback_fn callbackfn,
                                  t_dm_devevent_command_fn command_fn, t_dm_destroy_fn destroyfn,
                                  void *client_data)
{
  struct s_devicemounter_client *dmc;
  int newidx;
  bool found;
  int result = -1;

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_registerhandler: called\n");
  if (NULL == dm)
  {
    return -1;
  }

  /* Lock the structure */
  pthread_mutex_lock(&dm->mutex);

  /* Look for room in client list */
  for (newidx = 0, found = false; newidx < DEVICE_MOUNTER_MAX_CLIENTS; newidx++)
  {
    if (NULL == dm->clients[newidx])
    {
      found = true;
      break;
    }
  }

  if (found)
  {
    /* Found a spot for the new entry, create new client structure and add */
    dmc = calloc(1, sizeof(struct s_devicemounter_client));
    if (NULL != dmc)
    {
      /* Setup struct */
      dmc->client_data = client_data;
      dmc->match_fn = matchfn;
      dmc->callback_fn = callbackfn;
      dmc->command_fn = command_fn;
      dmc->destroy_fn = destroyfn;

      /* Add to list */
      dm->clients[newidx] = dmc;

      WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_registerhandler: registered new device mounter client\n");
      result = newidx;
    }
    else
    {
      rdebug_printf(LOG_ERR, "devicemounter_registerhandler: allocating device mounter client structure failed\n");
    }
  }
  else
  {
    rdebug_printf(LOG_ERR, "devicemounter_registerhandler: maximum devicemounter clients reached (%u)\n", DEVICE_MOUNTER_MAX_CLIENTS);
  }

  pthread_mutex_unlock(&dm->mutex);
  return result;
}


/** Issue a command to a registered handler */
int devicemounter_handlercommand(struct s_devicemounter *dm, size_t dm_handle, int command, void *command_data)
{
  struct s_devicemounter_client *dmc;
  int result = -1;

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_handlercommand: called: handle %u, command %d\n", (unsigned) dm_handle, command);
  if (NULL == dm)
  {
    return -1;
  }

  /* Lock the structure */
  pthread_mutex_lock(&dm->mutex);

  /* Find the client in the client list */
  if (dm_handle < DEVICE_MOUNTER_MAX_CLIENTS && NULL != dm->clients[dm_handle])
  {
    dmc = dm->clients[dm_handle];
    if (NULL != dmc->command_fn)
    {
      result = dmc->command_fn(dmc->client_data, command, command_data);
    }
  }

  pthread_mutex_unlock(&dm->mutex);
  return result;
}


/** Internal unregister a handler for a device. No locking, do not call directly */
static int devicemounter_unregisterhandler_internal(struct s_devicemounter *dm, size_t dm_handle)
{
  struct s_devicemounter_client *dmc;

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_unregisterhandler_internal: called for handle %u\n", (unsigned) dm_handle);
  if (NULL == dm || dm_handle >= DEVICE_MOUNTER_MAX_CLIENTS)
  {
    return -1;
  }

  /* Check data is valid */
  dmc = dm->clients[dm_handle];
  if (NULL == dmc)
  {
    return -1;
  }

  /* Client clean up */
  if (NULL != dmc->destroy_fn)
    dmc->destroy_fn(dmc->client_data);

  /* Free data */
  dm->clients[dm_handle] = NULL;
  free(dmc);
  dmc = NULL;

  return 0;
}


/** Unregister a handler for a device */
int devicemounter_unregisterhandler(struct s_devicemounter *dm, size_t dm_handle)
{
  int result;

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_unregisterhandler: called for handle %u\n", (unsigned) dm_handle);
  if (NULL == dm)
  {
    return -1;
  }

  pthread_mutex_lock(&dm->mutex);
  result = devicemounter_unregisterhandler_internal(dm, dm_handle);
  pthread_mutex_unlock(&dm->mutex);

  return result;
}


/** Unregister all device handlers, done as part of devicemounter_destroy(). Don't call from anywhere else */
static int devicemounter_unregisterhandlers(struct s_devicemounter *dm)
{
  size_t i;

  pthread_mutex_lock(&dm->mutex);

  for (i = 0; i < DEVICE_MOUNTER_MAX_CLIENTS; i++)
  {
    if (NULL != dm->clients[i])
    {
      (void) devicemounter_unregisterhandler_internal(dm, i);
    }
  }

  pthread_mutex_unlock(&dm->mutex);

  return 0;
}


/** Create a new device mounter, monitoring fifo_filename for device appearance/disappearance */
struct s_devicemounter *devicemounter_init(const char *fifo_filename)
{
  struct s_devicemounter *dm;
  int pipe_fds[2];
  int i;

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_init: called\n");

  if (NULL == fifo_filename)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: filename is NULL pointer\n");
    goto dmi_fail;
  }

  /* Allocate device mounter structure */
  dm = malloc(sizeof(struct s_devicemounter));
  if (NULL == dm)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: malloc dm struct failed\n");
    goto dmi_fail;
  }

  /* Copy the filename */
  dm->mdev_fifo_filename = strdup(fifo_filename);
  if (NULL == dm->mdev_fifo_filename)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: malloc filename failed\n");
    goto dmi_freedm;
  }

  /* Create named fifo */
  if (mkfifo(dm->mdev_fifo_filename, 0600) < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: mkfifo failed\n");
    goto dmi_freestr;
  }
  dm->mdev_fifo_fd = open(dm->mdev_fifo_filename, O_RDONLY | O_NONBLOCK);
  if (dm->mdev_fifo_fd < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: open fifo for read failed\n");
    goto dmi_deletefifo;
  }
  if (fcntl(dm->mdev_fifo_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: error setting FD_CLOEXEC on read end of fifo \"%s\" (errno %d)\n", dm->mdev_fifo_filename, errno);
    goto dmi_closefiford;
  }

  /*
   * Now open the fifo for writing. We never write to it
   * however, unless the write end is open, select will fail.
   * So we open it here, but never write to it, as other
   * processes do this
   */
  dm->mdev_fifo_write_fd = open(dm->mdev_fifo_filename, O_WRONLY);
  if (dm->mdev_fifo_write_fd < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: open fifo for write failed\n");
    goto dmi_closefiford;
  }
  if (fcntl(dm->mdev_fifo_write_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: error setting FD_CLOEXEC on write end of fifo \"%s\" (errno %d)\n", dm->mdev_fifo_filename, errno);
    goto dmi_closefifowr;
  }

  /* Create a pipe to write to, to shut down the task */
  if (pipe(pipe_fds) < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: create pipe failed\n");
    goto dmi_closefifowr;
  }

  /* Keep fds for shutting down task */
  dm->command_pipe_read_fd = pipe_fds[0];
  dm->command_pipe_write_fd = pipe_fds[1];
  if (fcntl(dm->command_pipe_read_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: error setting FD_CLOEXEC on read end of shutdown pipe (errno %d)\n", errno);
    goto dmi_closepipe;
  }
  if (fcntl(dm->command_pipe_write_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: error setting FD_CLOEXEC on write end of shutdown pipe (errno %d)\n", errno);
    goto dmi_closepipe;
  }

  /* Create exclusive access semaphore */
  if (0 != pthread_mutex_init(&dm->mutex, NULL))
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: create mutex failed\n");
    goto dmi_closepipe;
  }

  /* Allocate the buffer */
  dm->mdev_data_buffer_max = DEVICE_MOUNTER_MAX_BUFFER;
  dm->mdev_data_buffer_length = 0;
  dm->mdev_data_buffer = malloc(dm->mdev_data_buffer_max);
  if (NULL == dm->mdev_data_buffer)
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: create buffer failed\n");
    goto dmi_mutex;
  }

  /* Make all clients null */
  for (i = 0; i < DEVICE_MOUNTER_MAX_CLIENTS; i++)
  {
    dm->clients[i] = NULL;
  }

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_init: data fd: %d, shutdown read fd: %d, shutdown write fd: %d\n",
                                dm->mdev_fifo_fd,
                                dm->command_pipe_read_fd,
                                dm->command_pipe_write_fd);

  /* Create device mounter thread */
  if (!rin_thread_create(&dm->dm_pthread, devicemounter_thread, (void *) dm, true))
  {
    rdebug_printf(LOG_ERR, "devicemounter_init: pthread task create failed\n");
    goto dmi_buffer;
  }

  /* Success */
  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_init: success\n");
  return dm;

dmi_buffer:
  free(dm->mdev_data_buffer);

dmi_mutex:
  pthread_mutex_destroy(&dm->mutex);

dmi_closepipe:
  close(dm->command_pipe_read_fd);
  close(dm->command_pipe_write_fd);

dmi_closefifowr:
  close(dm->mdev_fifo_write_fd);

dmi_closefiford:
  close(dm->mdev_fifo_fd);

dmi_deletefifo:
  unlink(dm->mdev_fifo_filename);

dmi_freestr:
  free(dm->mdev_fifo_filename);

dmi_freedm:
  free(dm);

dmi_fail:
  return NULL;
}


/** Start the device mounter running. This should be called after registering the device handlers */
int devicemounter_run(struct s_devicemounter *dm)
{
  char data;

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_run: called\n");

  if (NULL == dm)
  {
    return -1;
  }

  /* Write (any) data to the pipe to force select() to return */
  data = DM_COMMAND_PIPE_RUN_CHAR;
  if (sizeof(data) != write(dm->command_pipe_write_fd, &data, sizeof(data)))
  {
    rdebug_printf(LOG_ERR, "devicemounter_run: write to command pipe failed\n");
    return -1;
  }

  return 0;
}


/** Clean up device mounter */
int devicemounter_destroy(struct s_devicemounter *dm)
{
  char data;

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_destroy: called\n");

  if (NULL == dm)
  {
    return -1;
  }

  /* Lock dm */
  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_destroy: wait on mutex\n");
  pthread_mutex_lock(&dm->mutex);

  /* Write (any) data to the pipe to force select() to return */
  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_destroy: write to shutdown\n");
  data = DM_COMMAND_PIPE_SHUTDOWN_CHAR;
  (void) write(dm->command_pipe_write_fd, &data, sizeof(data));

  /* Unlock dm */
  pthread_mutex_unlock(&dm->mutex);

  /* Wait for thread to quit */
  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_destroy: wait for task quit\n");
  pthread_join(dm->dm_pthread, NULL);

  /* Destroy any clients */
  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "devicemounter_destroy: unregister handlers\n");
  (void) devicemounter_unregisterhandlers(dm);

  /* Free the buffer */
  free(dm->mdev_data_buffer);
  dm->mdev_data_buffer = NULL;

  /* Destroy the mutex */
  pthread_mutex_destroy(&dm->mutex);

  /* Close the pipes */
  close(dm->command_pipe_read_fd);
  dm->command_pipe_read_fd = -1;
  close(dm->command_pipe_write_fd);
  dm->command_pipe_write_fd = -1;

  /* Close fifo */
  unlink(dm->mdev_fifo_filename);
  close(dm->mdev_fifo_write_fd);
  dm->mdev_fifo_write_fd = -1;
  close(dm->mdev_fifo_fd);
  dm->mdev_fifo_fd = -1;

  /* De-allocate memory */
  free(dm->mdev_fifo_filename);
  dm->mdev_fifo_filename = NULL;
  free(dm);
  dm = NULL;

  WE_DEVICEMOUNTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "devicemounter_destroy: finished\n");
  return 0;
}
