/*
 * devicemounter.h
 *
 * Handles the arrival and removal of linux devices.
 *
 * Usage:
 * 1 - Call @ref devicemounter_init() once to create a device mounter
 * 2 - Call @ref devicemounter_registerhandler() for each device notification you want to receive (usually called from dm_XXXXX_init() functions)
 * 3 - Call @ref devicemounter_run() to start the device mounter
 *
 * It is necessary to buffer event arrival/removal notifications generated by mdev before the WE2111 application starts to run.
 * This is done by scripts/devicenotify.c, which is compiled as a separate binary, and called by mdev as configured in /etc/mdev.conf.
 * mdev can be configured to notify us via the scripts/deviceregistration script. See postinst script for examples.
 *
 * The separate init & run calls are required so any buffered arrival/removal do not get (mis)handled before all the device handlers have
 * been registered.
 *
 */
#ifndef DEVICEMOUNTER_H_
#define DEVICEMOUNTER_H_

#include <pthread.h>


/** Device mounter action types */
enum eDeviceMounterAction
{
  dmaAdd,         /*!< Device was added to the system */
  dmaRemove       /*!< Device was removed from the system */
};


/** Typedef of function to determine if the event matches. Return -1 for error, 0 for no match, 1 for match */
typedef int (*t_dm_devevent_match_fn)(void *client_data, const char *environment);


/** Typedef of function to call when the event matches */
typedef int (*t_dm_devevent_callback_fn)(void *client_data, enum eDeviceMounterAction action, const char *devicename, const char *environment);


/** Typedef of function to call from devicemounter_handlercommand() */
typedef int (*t_dm_devevent_command_fn)(void *client_data, int command, void *command_data);


/** Typedef of function to call when cleaning up the device mounter */
typedef int (*t_dm_destroy_fn)(void *client_data);


/** Structure for a device mounter client */
struct s_devicemounter_client
{
  void *client_data;
  t_dm_devevent_match_fn match_fn;
  t_dm_devevent_callback_fn callback_fn;
  t_dm_devevent_command_fn command_fn;
  t_dm_destroy_fn destroy_fn;
};


/** Maximum number of clients (printer, partition, event) */
#define DEVICE_MOUNTER_MAX_CLIENTS 5


/** Maximum number of bytes held in the data buffer */
#define DEVICE_MOUNTER_MAX_BUFFER 2048


/** Device mounter structure */
struct s_devicemounter
{
  char *mdev_fifo_filename;               /*!< Filename of fifo that we read from to be informed of device arrival/removal */
  int mdev_fifo_fd;                       /*!< File descriptor of fifo that we read from to be informed of device arrival/removal */
  int mdev_fifo_write_fd;                 /*!< We also open the fifo for write, to keep select & read happy, but we never write to this */

  char *mdev_preapp_filename;             /*!< Because mdev comes up well before WE application, we store any device arrivals in that time in this file */
  int mdev_preapp_fd;                     /*!< File descriptor of file containing device arrivals prior to WE application starting */

  pthread_t dm_pthread;                   /*!< Device mounter thread */
  pthread_mutex_t mutex;                  /*!< Exclusive access mutex */

  int command_pipe_read_fd;              /*!< Read end of pipe, used by device mounter thread to know when to exit */
  int command_pipe_write_fd;             /*!< Write end of pipe, used by destroy call to notify device mounter thread to quit */

  size_t mdev_data_buffer_max;            /*!< Maximum length of mdev data buffer */
  size_t mdev_data_buffer_length;         /*!< Current length of mdev data buffer */
  char *mdev_data_buffer;                 /*!< Data from mdev */

  /** Array of pointers to client structures */
  struct s_devicemounter_client *clients[DEVICE_MOUNTER_MAX_CLIENTS];
};


/** Get the full device path for a device. It's up to the caller to free the returned string */
char *devicemounter_getfulldevicepath(const char *device);


/** Register a handler for a device */
int devicemounter_registerhandler(struct s_devicemounter *dm,
                                  t_dm_devevent_match_fn matchfn, t_dm_devevent_callback_fn callbackfn,
                                  t_dm_devevent_command_fn command_fn, t_dm_destroy_fn destroyfn,
                                  void *client_data);


/** Issue a command to a registered handler */
int devicemounter_handlercommand(struct s_devicemounter *dm, size_t dm_handle, int command, void *command_data);


/** Unregister a handler for a device */
int devicemounter_unregisterhandler(struct s_devicemounter *dm, size_t dm_handle);


/** Create a new device mounter, monitoring fifo_filename for device appearance/disappearance */
struct s_devicemounter *devicemounter_init(const char *fifo_filename);


/** Start the device mounter running. This should be called after registering the device handlers */
int devicemounter_run(struct s_devicemounter *dm);


/** Clean up device mounter */
int devicemounter_destroy(struct s_devicemounter *dm);


#endif /* DEVICEMOUNTER_H_ */
