/*
 * dm_envlist.c
 *
 * List of environment variables passed over fifo
 */
#include <stddef.h>
#include <string.h>

#include "dm_envlist.h"


/**
 * Determine if an environment variable exists in the list:
 * varname=varvalue\n
 * varname=varvalue\n
 * \0
 *
 * Returns 0 if found, -1 if not
 *
 */
int envlist_variable_exists(const char *buf, const char *varname)
{
  char *pos;
  size_t vnl;
  int result = -1;

  /* Find variable */
  pos = strstr(buf, varname);

  /* Should be first entry, or preceded by '\n' */
  if (NULL != pos && (pos == buf || '\n' == pos[-1]))
  {
    /* Should be followed by '=' */
    vnl = strlen(varname);
    if ('=' == pos[vnl])
    {
      result = 0;
    }
  }
  return result;
}


/**
 * Find a environment variable in a list of the form:
 * varname=varvalue\n
 * varname=varvalue\n
 * \0
 *
 * Returns 0 if found, -1 if not
 *
 */
int envlist_find(const char *buf, const char *varname, char *varvalue, size_t varvalue_max_len)
{
  char *pos;
  size_t vnl;
  size_t vvl;
  int result = -1;

  if (NULL == buf || NULL == varname || NULL == varvalue)
    return -1;

  /* Must have some room */
  if (varvalue_max_len > 0)
  {
    /* Find variable */
    pos = strstr(buf, varname);

    /* Should be first entry, or preceded by '\n' */
    if (NULL != pos && (pos == buf || '\n' == pos[-1]))
    {
      /* Should be followed by '=' */
      vnl = strlen(varname);
      if ('=' == pos[vnl])
      {
        /* Copy the rest up to '\n', or '\0' into varvalue */
        pos = &pos[vnl + 1];
        vvl = 0;
        while ('\0' != *pos && '\n' != *pos)
        {
          /* Check no room */
          if (vvl >= (varvalue_max_len - 1))
          {
            return -1;
          }

          varvalue[vvl++] = *pos++;
        }
        varvalue[vvl] = '\0';
        result = 0;
      }
    }
  }
  return result;
}


/**
 * Find a environment variable in a list of the form:
 * varname=varvalue\n
 * varname=varvalue\n
 * \0
 *
 * and check that the value equals varvalue
 *
 * Returns 0 for match, -1 otherwise
 *
 */
int envlist_findequals(const char *buf, const char *varname, const char *varvalue)
{
  char *pos;
  size_t vnl;
  int result = -1;

  /* Check pointers */
  if (NULL == buf || NULL == varname || NULL == varvalue)
    return -1;

  /* Find variable */
  pos = strstr(buf, varname);

  /* Should be first entry, or preceded by '\n' */
  if (NULL != pos && (pos == buf || '\n' == pos[-1]))
  {
    /* Should be followed by '=' */
    vnl = strlen(varname);
    if ('=' == pos[vnl])
    {
      /* Compare the rest up to '\n', or '\0' to varvalue */
      pos = &pos[vnl + 1];
      while ('\0' != *varvalue)
      {
        if (*varvalue++ != *pos++)
        {
          return -1;
        }
      }

      /* And pos should be terminated by '\0' or '\n' */
      if ('\0' != *pos && '\n' != *pos)
      {
        return -1;
      }

      /* Success */
      result = 0;
    }
  }

  return result;
}
