/*
 * dm_event.c
 *
 *  Created on: Nov 14, 2011
 *      Author: ian
 */
#include <string.h>
#include <stdlib.h>
#include <rinhw/kbd_linux_event.h>
#include "we_debug.h"
#include "devicemounter.h"
#include "dm_envlist.h"

#include "dm_event.h"


/*
 * When an event device (keyboard, mouse) appears, mdev delivers us:
 *
 * ACTION=add
 * DEVPATH=/class/input/input1/event1
 * SUBSYSTEM=input
 * MAJOR=13
 * MINOR=65
 * PHYSDEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.1
 * PHYSDEVBUS=usb
 * PHYSDEVDRIVER=usbhid
 * SEQNUM=718
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=event1
 * HUSH_VERSION=1.18.3
 * PWD=/dev
 *
 * When a event device (keyboard, mouse) disappears, mdev delivers us:
 *
 * ACTION=remove
 * DEVPATH=/class/input/input3/event1
 * SUBSYSTEM=input
 * MAJOR=13
 * MINOR=65
 * PHYSDEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.1
 * PHYSDEVBUS=usb
 * PHYSDEVDRIVER=usbhid
 * SEQNUM=745
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=event1
 * HUSH_VERSION=1.18.3
 * PWD=/dev
 *
 */


/** Information about the event device */
struct dm_event_data
{
  struct t_io_dev *kbd_dev;                  /*!< Rinstrum keyboard device structure to deal with event devices */
  t_dm_event_callback_fn event_callback_fn;  /*!< Function to call when a event device is added/removed */
  void *client_callback_data;                /*!< Pointer to data to pass to callback function */
};


/** Function to match variables for when an event device appears (eg. USB keyboard /dev/event0) */
static int dm_event_match(void *client_data, const char *environment)
{
  /* Match SUBSYSTEM=input */
  return (0 == envlist_findequals(environment, "SUBSYSTEM", "input")) ? 1 : 0;
}


/** Log an error message */
static void dm_event_error(enum eDeviceMounterAction action, const char *const device, int error)
{
  const char *const addremove_str = (dmaAdd == action) ? "adding" : "removing";
  const char *const device_str = (NULL == device) ? "all (null)" : device;

  switch (error)
  {
    /* Leave a more detailed syslog message for these errors */
    case RIO_ERROR_EVKBD_FULL:
      rdebug_printf(LOG_WARNING, "dm_event_error: maximum number of event devices reached when %s event device %s to keyboard driver\n",
                    addremove_str, device_str);
      break;
    case RIO_ERROR_EVKBD_KEYS_UNSUPPORTED:
      rdebug_printf(LOG_WARNING, "dm_event_error: required keys are not supported on %s, so not %s to keyboard driver\n",
                    device_str, addremove_str);
      break;

    /* These errors should not regularly occur, so just report the error code */
    default:
    case RIO_ERROR:
    case RIO_ERROR_EVKBD_INVALID_PARAM:
    case RIO_ERROR_EVKBD_ALREADY_HANDLED:
    case RIO_ERROR_EVKBD_NOT_HANDLED:
    case RIO_ERROR_EVKBD_IO:
    case RIO_ERROR_EVKBD_OUT_OF_MEMORY:
    case RIO_ERROR_EVKBD_NOT_EVENT:
    case RIO_ERROR_EVKBD_MISC:
      rdebug_printf(LOG_ERR, "dm_event_error: error %d reported when %s event device %s to keyboard driver\n",
                    error, addremove_str, device_str);
      break;
    case RIO_SUCCESS:
      rdebug_printf(LOG_INFO, "dm_event_error: success: %s event device %s to keyboard driver\n",
                    addremove_str, device_str);
      break;
  }
}


/** Handle arrival/removal of a event device, by trying to create a rinstrum device to handle it */
static int dm_event_handler(void *client_data, enum eDeviceMounterAction action, const char *devicename, const char *environment)
{
  struct dm_event_data *ded = client_data;
  char *fulldev;
  int result;

  /* Check pointers */
  if (NULL == ded || NULL == ded->kbd_dev ||
      (NULL == devicename && dmaRemove != action))
  {
    return -1;
  }

  result = RIO_ERROR;
  if (dmaAdd == action)
  {
    /* Get the full device path */
    fulldev = devicemounter_getfulldevicepath(devicename);
    if (NULL != fulldev)
    {
      /* Add a new event device (eg /dev/event0) to be monitored */
      result = rio_kbd_linux_event_add_event_device(ded->kbd_dev, fulldev);

      /* Clean up */
      free(fulldev);
      fulldev = NULL;
    }
  }
  else /* action == remove */
  {
    if (NULL != devicename)
    {
      /* Get the full device path */
      fulldev = devicemounter_getfulldevicepath(devicename);
      if (NULL != fulldev)
      {
        /* Remove an event device from a rinstrum linux event keyboard driver */
        result = rio_kbd_linux_event_remove_event_device(ded->kbd_dev, fulldev);

        /* Clean up */
        free(fulldev);
        fulldev = NULL;
      }
    }
    else
    {
      /* Remove all event devices from a rinstrum linux event keyboard driver */
      result = rio_kbd_linux_event_remove_event_device(ded->kbd_dev, NULL);
    }
  }

  /* Call the callback to notify of usb event */
  dm_event_error(action, devicename, result);
  if (NULL != ded->event_callback_fn)
  {
    ded->event_callback_fn(ded->client_callback_data, action, result);
  }

  return 0;
}


/**
 * Command call back, used to unregister an event device
 *
 * @param client_data
 * @param command is one of DM_EVENT_CMD_*
 * @param command_data is:
 *    DM_EVENT_CMD_UNREGISTER: pointer to string of device name (eg "event0") to unregister, or NULL to remove all
 */
static int dm_event_command(void *client_data, int command, void *command_data)
{
  int result = -1;
  struct dm_event_data *ded = client_data;

  if (NULL == ded)
    return -1;

  if (DM_EVENT_CMD_UNREGISTER == command)
  {
    WE_DEVICEMOUNTER_EVENT_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_event_command: unmount all registered event handlers\n");
    result = dm_event_handler(client_data, dmaRemove, NULL, NULL);
    WE_DEVICEMOUNTER_EVENT_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_event_command: result of user unregister: %d\n", result);
  }

  return result;
}


/** Destroy the event device mounter handler */
static int dm_event_destroy(void *client_data)
{
  struct dm_event_data *ded = client_data;

  WE_DEVICEMOUNTER_EVENT_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_event_destroy: called\n");

  if (NULL != ded)
  {
    /*
     * There's nothing to clean up here, because the keyboard
     * device is created externally to this event handler
     * and passed in to dm_event_init(), so the responsibility
     * of destroying it is not here
     */

    /* Free up our structure */
    free(ded);
    ded = NULL;
  }
  return 0;
}


/** Initialise handling of event devices. kbd_io_dev is an existing keyboard device created by rio_create_device_kbd_linux_event() */
int dm_event_init(struct s_devicemounter *dm, struct t_io_dev *kbd_io_dev, t_dm_event_callback_fn event_callback_fn, void *client_callback_data)
{
  struct dm_event_data *ded = NULL;
  int dm_handle;

  WE_DEVICEMOUNTER_EVENT_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_event_init: called\n");

  /* Check pointers */
  if (NULL == dm || NULL == kbd_io_dev)
  {
    goto dei_fail;
  }

  /* Allocate memory */
  ded = calloc(1, sizeof(struct dm_event_data));
  if (NULL == ded)
  {
    goto dei_fail;
  }

  /* Setup members */
  ded->kbd_dev = kbd_io_dev;
  ded->event_callback_fn = event_callback_fn;
  ded->client_callback_data = client_callback_data;

  /* Register */
  dm_handle = devicemounter_registerhandler(dm,
                                            dm_event_match, dm_event_handler,
                                            dm_event_command, dm_event_destroy,
                                            (void *) ded);
  if (dm_handle < 0)
  {
    goto dei_fail;
  }

  /* Success */
  WE_DEVICEMOUNTER_EVENT_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_event_init: success, dm_handle %d\n", dm_handle);
  return dm_handle;

dei_fail:
  if (NULL != ded)
  {
    free(ded);
    ded = NULL;
  }

  rdebug_printf(LOG_ERR, "dm_event_init: failed\n");
  return -1;
}
