/*
 * dm_partition.c
 *
 * Handles mounting of one partition to a
 * location specified in the call to dm_partition_init().
 *
 * In it's current form this only supports one partition being mounted.
 * So the first device DEVTYPE=partition and SUBSYSTEM=block
 * reported by mdev will be mounted RW to the location.
 *
 * Any further partitions appearing mounted, until the mounted
 * partition is unmounted.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <rinexec/el_runcmd.h>
#include "we_debug.h"
#include "devicemounter.h"
#include "dm_envlist.h"

#include "dm_partition.h"


/*
 * When a partition appears, mdev delivers us:
 *
 * ACTION=add
 * DEVPATH=/block/sda/sda1
 * SUBSYSTEM=block
 * MAJOR=8
 * MINOR=1
 * DEVTYPE=partition
 * PHYSDEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.0/host1/target1:0:0/1:0:0:0
 * PHYSDEVBUS=scsi
 * PHYSDEVDRIVER=sd
 * SEQNUM=763
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=sda1
 * HUSH_VERSION=1.18.3
 * PWD=/dev
 *
 * When a partition disappears, mdev delivers us:
 *
 * ACTION=remove
 * DEVPATH=/block/sda/sda1
 * SUBSYSTEM=block
 * MAJOR=8
 * MINOR=1
 * DEVTYPE=partition
 * PHYSDEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.0/host0/target0:0:0/0:0:0:0
 * PHYSDEVBUS=scsi
 * PHYSDEVDRIVER=sd
 * SEQNUM=743
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=sda1
 * HUSH_VERSION=1.18.3
 * PWD=/dev
 *
 */


/** Information about the mounted partition */
struct dm_partition_data
{
  char *location;           /*!< Location to mount the partition to (eg /mnt/rwfs) */
  char *device;             /*!< The device mounted (eg sda1). Is NULL if no partition mounted) */

  char *filesystem_type;    /*!< The file system type to mount. NOTE: this only allows one type of file system to be mounted. Can be NULL if you want mount to try and guess. */
  char *mount_options;      /*!< Options passed to mount with -o (eg sync, nosuid), can be NULL if no options */

  t_dm_partition_callback_fn dm_partition_callback_fn;  /*!< Function to call when a partition is mounted/unmounted */
  void *client_callback_data;                           /*!< Pointer to data to pass to callback function */
};


/** Mount a partition */
static int dm_mount_partition(const char *device, const char *location, const char *fstype, const char *options)
{
  char *fulldev = NULL;
  const char *mount_args[8];
  int ret_code;
  int result = -1;
  const char mount_cmd[] = "/bin/mount";
  const char options_option[] = "-o";
  const char fstype_option[] = "-t";
  int argcount;

  /* Check pointers */
  if (NULL == device || NULL == location)
    goto dmp_done;

  /* Get full device path */
  fulldev = devicemounter_getfulldevicepath(device);
  if (NULL == fulldev)
  {
    /* Fail */
    goto dmp_done;
  }

  /* Setup mount arguments */
  argcount = 0;
  mount_args[argcount++] = mount_cmd;
  if (NULL != options)
  {
    mount_args[argcount++] = options_option;
    mount_args[argcount++] = options;
  }
  if (NULL != fstype)
  {
    mount_args[argcount++] = fstype_option;
    mount_args[argcount++] = fstype;
  }
  mount_args[argcount++] = fulldev;
  mount_args[argcount++] = location;
  mount_args[argcount++] = NULL;

  /* Run mount and check the return code */
  if (el_run_command(mount_args[0], mount_args, &ret_code, NULL, true, false, true) != elr_success ||
      !WIFEXITED(ret_code) || 0 != WEXITSTATUS(ret_code))
  {
    result = -1;
  }
  else
  {
    result = 0;
  }

  /* Clean up */
dmp_done:
  if (NULL != fulldev)
  {
    free(fulldev);
  }

  return result;
}


/**
 * Unmount a partition
 *
 * (use location, rather than device because if the device has
 * already disappeared (by unplugging it), then umount will
 * complain)
 */
static int dm_unmount_partition(const char *location)
{
  int ret_code;
  const char *const umount_args[] = { "/bin/umount", location, NULL };
  int result = -1;

  if (NULL == location)
    goto dup_done;

  /* Run umount and check the return code */
  if (el_run_command(umount_args[0], umount_args, &ret_code, NULL, true, false, true) != elr_success ||
      !WIFEXITED(ret_code) || 0 != WEXITSTATUS(ret_code))
  {
    result = -1;
  }
  else
  {
    result = 0;
  }

  /* Clean up */
dup_done:
  return result;
}


/** Function to match variables for when a block device partition appears (eg. HDD) */
static int dm_partition_match(void *client_data, const char *environment)
{
  /* Match DEVTYPE=partition and SUBSYSTEM=block */
  return (0 == envlist_findequals(environment, "DEVTYPE", "partition") &&
          0 == envlist_findequals(environment, "SUBSYSTEM", "block")) ? 1 : 0;
}


/** Handle arrival of a partition, by trying to mount it */
static int dm_partition_handler(void *client_data, enum eDeviceMounterAction action, const char *devicename, const char *environment)
{
  struct dm_partition_data *dpd = client_data;
  bool result;

  if (NULL == dpd)
    return -1;

  result = false;
  if (dmaAdd == action)
  {
    if (NULL == dpd->device)
    {
      /** Add a partition, and not already present */
      dpd->device = strdup(devicename);
      if (NULL != dpd->device)
      {
        /* Mount the partition into location */
        if (dm_mount_partition(dpd->device, dpd->location, dpd->filesystem_type, dpd->mount_options) < 0)
        {
          /* Fail, so free up memory */
          rdebug_printf(LOG_ERR, "dm_partition_handler: failed to mount partition %s at %s, is it vfat?\n", dpd->device, dpd->location);
          free(dpd->device);
          dpd->device = NULL;
        }
        else
        {
          /* Success */
          rdebug_printf(LOG_INFO, "dm_partition_handler: successfully mounted partition %s at %s\n", dpd->device, dpd->location);
          result = true;
        }
      }
      else
      {
        rdebug_printf(LOG_ERR, "dm_partition_handler: failed to allocate memory when trying to mount new device %s\n", devicename);
      }
    }
    else
    {
      rdebug_printf(LOG_WARNING, "dm_partition_handler: a partition (%s) is already mounted, not mounting new device %s\n", dpd->device, devicename);
    }
  }
  else /* action == remove */
  {
    if (NULL != dpd->device)
    {
      /* Can only remove the partition we know about */
      if (0 == strcmp(dpd->device, devicename))
      {
        /* Unmount device (by location rather than device) */
        if (dm_unmount_partition(dpd->location) < 0)
        {
          rdebug_printf(LOG_ERR, "dm_partition_handler: failed to unmount partition %s at %s\n", dpd->device, dpd->location);
        }
        else
        {
          rdebug_printf(LOG_INFO, "dm_partition_handler: successfully unmounted partition %s at %s\n", dpd->device, dpd->location);
          result = true;
        }

        /*
         * Clean up regardless of unmount result
         * and make device NULL so another mount could take place
         * (there's not much else we can do if it fails)
         */
        free(dpd->device);
        dpd->device = NULL;
      }
      else
      {
        rdebug_printf(LOG_INFO, "dm_partition_handler: %s does not match mounted partition %s, not unmounting\n", devicename, dpd->device);
      }
    }
    else
    {
      rdebug_printf(LOG_INFO, "dm_partition_handler: no partition mounted, so can't unmount %s\n", devicename);
    }
  }

  /* Call notification function */
  if (NULL != dpd->dm_partition_callback_fn)
  {
    dpd->dm_partition_callback_fn(dpd->client_callback_data, action, dpd->location, result);
  }

  return (result) ? 0 : -1;
}


/** Command call back, used to unmount the disk */
static int dm_partition_command(void *client_data, int command, void *command_data)
{
  int result = -1;
  struct dm_partition_data *dpd = client_data;

  if (DM_PARTITION_CMD_UNMOUNT == command)
  {
    result = dm_partition_handler(client_data, dmaRemove, dpd->device, NULL);
    WE_DEVICEMOUNTER_PARTITION_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_partition_command: result of user unmount %s: %d\n", dpd->device, result);
  }


  return result;
}


/** Destroy the partition device mounter handler */
static int dm_partition_destroy(void *client_data)
{
  struct dm_partition_data *dpd = client_data;

  WE_DEVICEMOUNTER_PARTITION_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_partition_destroy: called\n");

  if (NULL != dpd)
  {
    if (NULL != dpd->device)
    {
      /* Unmount device */
      WE_DEVICEMOUNTER_PARTITION_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_partition_destroy: unmounting partition %s at %s\n", dpd->device, dpd->location);
      (void) dm_unmount_partition(dpd->device);

      /* Clean up */
      free(dpd->device);
      dpd->device = NULL;
    }
    if (NULL != dpd->location)
    {
      free(dpd->location);
      dpd->location = NULL;
    }
    if (NULL != dpd->filesystem_type)
    {
      free(dpd->filesystem_type);
      dpd->filesystem_type = NULL;
    }
    if (NULL != dpd->mount_options)
    {
      free(dpd->mount_options);
      dpd->mount_options = NULL;
    }
    free(dpd);
  }
  return 0;
}


/**
 * Initialise handling of partition devices. Currently only handles one device mounted at a time.
 *
 * @param dm is a device mounter
 * @param mount_location is the location to mount the partition to (eg /mnt/rwfs)
 * @param filesystem_type is the file system type to mount. NOTE: this only allows one type of file system to be mounted. Can be NULL if you want mount to try and guess. (eg vfat)
 * @param mount_options are the options passed to mount with -o (eg sync, nosuid), can be NULL if no options
 * @param partition_callback_fn is the function to call when a partition is mounted/unmounted
 * @param client_callback_data is a pointer to data to pass to callback function
 *
 * @return 0 for success, < 0 otherwise
 */
int dm_partition_init(struct s_devicemounter *dm, const char *mount_location,
                      const char *filesystem_type, const char *mount_options,
                      t_dm_partition_callback_fn partition_callback_fn, void *client_callback_data)
{
  struct dm_partition_data *dpd = NULL;
  int dm_handle;

  WE_DEVICEMOUNTER_PARTITION_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_partition_init: called\n");

  /* Quick checks, file system type and mount options can be NULL */
  if (NULL == dm || NULL == mount_location)
  {
    goto dpi_fail;
  }

  dpd = malloc(sizeof(struct dm_partition_data));
  if (NULL == dpd)
  {
    goto dpi_fail;
  }

  /* Initialise elements */
  memset(dpd, 0, sizeof(struct dm_partition_data));
  dpd->device = NULL;
  dpd->location = strdup(mount_location);
  if (NULL == dpd->location)
  {
    goto dpi_fail;
  }
  if (NULL != filesystem_type)
  {
    dpd->filesystem_type = strdup(filesystem_type);
    if (NULL == dpd->filesystem_type)
    {
      goto dpi_fail;
    }
  }
  else
  {
    dpd->filesystem_type = NULL;
  }
  if (NULL != mount_options)
  {
    dpd->mount_options = strdup(mount_options);
    if (NULL == dpd->mount_options)
    {
      goto dpi_fail;
    }
  }
  else
  {
    dpd->mount_options = NULL;
  }
  dpd->dm_partition_callback_fn = partition_callback_fn;
  dpd->client_callback_data = client_callback_data;

  /* Register the partition device handler */
  dm_handle = devicemounter_registerhandler(dm,
                                             dm_partition_match, dm_partition_handler,
                                             dm_partition_command, dm_partition_destroy,
                                             (void *) dpd);
  if (dm_handle < 0)
  {
    /* Fail */
    goto dpi_fail;
  }

  /* Success */
  WE_DEVICEMOUNTER_PARTITION_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_partition_init: success, dm_handle %d\n", dm_handle);
  return dm_handle;

  /* Fails */
dpi_fail:
  if (NULL != dpd)
  {
    if (NULL != dpd->location)
    {
      free(dpd->location);
      dpd->location = NULL;
    }
    if (NULL != dpd->filesystem_type)
    {
      free(dpd->filesystem_type);
      dpd->filesystem_type = NULL;
    }
    if (NULL != dpd->mount_options)
    {
      free(dpd->mount_options);
      dpd->mount_options = NULL;
    }
    dpd->client_callback_data = NULL;
    dpd->dm_partition_callback_fn = NULL;
    free(dpd);
    dpd = NULL;
  }

  rdebug_printf(LOG_ERR, "dm_partition_init: failed\n");
  return -1;
}
