/*
 * dm_printer.c
 *
 *  Created on: Nov 14, 2011
 *      Author: ian
 */
#include <string.h>
#include <stdlib.h>
#include <rinhw/sio_local_printer.h>
#include "devicemounter.h"
#include "dm_envlist.h"
#include "comms_task.h"
#include "we_debug.h"

#include "dm_printer.h"


/*
 * When a printer appears, mdev delivers us:
 *
 * ACTION=add
 * DEVPATH=/class/usb/lp0
 * SUBSYSTEM=usb
 * MAJOR=180
 * MINOR=0
 * PHYSDEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.0
 * PHYSDEVBUS=usb
 * PHYSDEVDRIVER=usblp
 * SEQNUM=783
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=lp0
 * HUSH_VERSION=1.18.3
 * PWD=/dev
 *
 * When a printer disappears, mdev delivers us:
 *
 * ACTION=remove
 * DEVPATH=/class/usb/lp0
 * SUBSYSTEM=usb
 * MAJOR=180
 * MINOR=0
 * PHYSDEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.0
 * PHYSDEVBUS=usb
 * PHYSDEVDRIVER=usblp
 * SEQNUM=785
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=lp0
 * HUSH_VERSION=1.18.3
 * PWD=/dev
 *
 * Actually when "mdev -s" is run on startup, all we really get is
 * MDEV=lpX and SUBSYSTEM=usb, so thats what we use
 */


/** Information about the printer */
struct dm_printer_data
{
  char *printer;                                 /*!< The new printer device (eg lp0) */
  struct t_io_dev *printer_dev;                  /*!< Rinstrum device structure to drive a printer */
  t_dm_printer_callback_fn printer_callback_fn;  /*!< Function to call when a printer device is added/removed */
  void *client_callback_data;                    /*!< Pointer to data to pass to callback function */
};


/** Function to match variables for when a printer device appears (eg. USB printer /dev/lp0) */
static int dm_printer_match(void *client_data, const char *environment)
{
  #define SERIAL_DEV_LEN 10
  char print_device[SERIAL_DEV_LEN];

  /* Check "SUBSYSTEM=usb" */
  if (0 != envlist_findequals(environment, "SUBSYSTEM", "usb"))
  {
    return 0;
  }

  /* Get MDEV variable */
  if (0 != envlist_find(environment, "MDEV", print_device, SERIAL_DEV_LEN - 1))
  {
    return 0;
  }

  /* Is it 'lp[0-9]'? */
  return ('l' == print_device[0] &&
          'p' == print_device[1] &&
          '0' <= print_device[2] &&
          '9' >= print_device[2] &&
          '\0' == print_device[3]) ? 1 : 0;
}


/** Handle arrival/removal of a printer, by trying to create a rinstrum device to handle it */
static int dm_printer_handler(void *client_data, enum eDeviceMounterAction action, const char *devicename, const char *environment)
{
  struct dm_printer_data *dpd = client_data;
  char *fulldev;
  bool result;

  /* Check pointers */
  if (NULL == dpd || NULL == devicename)
    return -1;

  result = false;
  if (dmaAdd == action)
  {
    if (NULL == dpd->printer)
    {
      /* Add a printer, and one not already present */
      dpd->printer = strdup(devicename);
      if (NULL != dpd->printer)
      {
        /* Get the full device path */
        fulldev = devicemounter_getfulldevicepath(dpd->printer);
        if (NULL != fulldev)
        {
          /* Create rinstrum device */
          if (RIO_SUCCESS == rio_create_device_sio_localprinter(dpd->printer_dev, fulldev))
          {
            /* Add new printer comms task to handle it */
            if (RIO_SUCCESS == comms_create_thread_for_serial_port(dpd->printer_dev, SERPORT_USBLP, NULL))
            {
              WE_DEVICEMOUNTER_PRINTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_printer_handler: successfully created rinstrum printer device and task for device %s\n", dpd->printer);
              result = true;
            }
            else
            {
              /* Fail */
              rdebug_printf(LOG_ERR, "dm_printer_handler: failed create thread for printer device %s\n", dpd->printer);
            }
          }
          else
          {
            rdebug_printf(LOG_ERR, "dm_printer_handler: failed to create rinstrum printer for device %s\n", dpd->printer);
          }
          free(fulldev);
          fulldev = NULL;
        }
      }
    }
    else
    {
      rdebug_printf(LOG_WARNING, "dm_printer_handler: a printer is already mounted, not mounting %s\n", devicename);
    }
  }
  else /* action == remove */
  {
    if (NULL != dpd->printer)
    {
      /* Can only remove the printer we know about */
      if (0 == strcmp(dpd->printer, devicename))
      {
        WE_DEVICEMOUNTER_PRINTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_printer_handler: cleaning up registered printer %s\n", dpd->printer);

        /* Tell printer comms task to quit, this blocks until it quits */
        comms_destroy_task(dpd->printer_dev);

        /* Destroy device */
        result = (rdestroy_io(dpd->printer_dev) == RIO_SUCCESS);

        /* Clean up */
        free(dpd->printer);
        dpd->printer = NULL;
      }
    }
    else
    {
      rdebug_printf(LOG_ERR, "dm_printer_handler: no printer registered, so can't clean up %s\n", devicename);
    }
  }

  /* Call the callback to notify of usb event */
  if (NULL != dpd->printer_callback_fn)
  {
    dpd->printer_callback_fn(dpd->client_callback_data, action, result);
  }

  return 0;
}


/**
 * Command call back, used to remove the printer
 *
 * @param client_data
 * @param command is one of DM_PRINTER_CMD_*
 * @param command_data is:
 *    DM_PRINTER_CMD_REMOVE: NULL
 */
static int dm_printer_command(void *client_data, int command, void *command_data)
{
  int result = -1;
  struct dm_printer_data *dpd = client_data;

  if (NULL == dpd)
    return -1;

  if (DM_PRINTER_CMD_REMOVE == command)
  {
    if (NULL != dpd->printer)
    {
      result = dm_printer_handler(client_data, dmaRemove, dpd->printer, NULL);
      WE_DEVICEMOUNTER_PRINTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_printer_command: result of user initiated printer remove %d\n", result);
    }
    else
    {
      rdebug_printf(LOG_ERR, "dm_printer_command: no printer available to remove\n");
    }
  }

  return result;
}


/** Destroy the printer device mounter handler */
static int dm_printer_destroy(void *client_data)
{
  struct dm_printer_data *dpd = client_data;

  WE_DEVICEMOUNTER_PRINTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_printer_destroy: called\n");

  if (NULL != dpd)
  {
    if (NULL != dpd->printer)
    {
      /* There is a printer attached */
      if (NULL != dpd->printer_dev)
      {
        /* There is a rinstrum device for this printer */

        /* Tell printer comms task to quit. Blocks until quit. */
        comms_destroy_task(dpd->printer_dev);

        /* Destroy rin device */
        rdestroy_io(dpd->printer_dev);
        dpd->printer_dev = NULL;
      }

      /* Free up printer string (eg lp0) */
      free(dpd->printer);
      dpd->printer = NULL;
    }

    free(dpd);
  }
  return 0;
}


/** Initialise handling of printer devices. printer_dev must not be destroyed/freed before the device mounter. */
int dm_printer_init(struct s_devicemounter *dm, struct t_io_dev *printer_dev, t_dm_printer_callback_fn printer_callback_fn, void *client_callback_data)
{
  struct dm_printer_data *dpd;
  int dm_handle;

  WE_DEVICEMOUNTER_PRINTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_printer_init: called\n");

  /* Quick checks */
  if (NULL == dm || NULL == printer_dev)
  {
    goto dpi_fail;
  }

  /* Allocate memory */
  dpd = malloc(sizeof(struct dm_printer_data));
  if (NULL == dpd)
  {
    goto dpi_fail;
  }

  /* Setup members */
  memset(dpd, 0, sizeof(struct dm_printer_data));
  dpd->printer = NULL;
  dpd->printer_dev = printer_dev;
  dpd->printer_callback_fn = printer_callback_fn;
  dpd->client_callback_data = client_callback_data;

  /* Register */
  dm_handle = devicemounter_registerhandler(dm,
                                             dm_printer_match, dm_printer_handler,
                                             dm_printer_command, dm_printer_destroy,
                                             (void *) dpd);
  if (dm_handle < 0)
  {
    goto dpi_dpd;
  }

  /* Success */
  WE_DEVICEMOUNTER_PRINTER_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_printer_init: success, dm_handle %d\n", dm_handle);
  return dm_handle;

  /* Fails */
dpi_dpd:
  free(dpd);

dpi_fail:
  rdebug_printf(LOG_ERR, "dm_printer_init: failed\n");
  return -1;
}
