/*
 * dm_serial.c
 *
 *  Created on: Sep 2, 2016
 *      Author: ian
 */
#include <string.h>
#include <stdlib.h>
#include <rinhw/sio_local_printer.h>
#include "devicemounter.h"
#include "dm_envlist.h"
#include "comms_task.h"
#include "we_debug.h"

#include "dm_serial.h"


/*
 * When a serial port appears, mdev delivers us two devices:
 *
 * ACTION=add
 * DEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.0/ttyUSB0
 * SUBSYSTEM=usb-serial
 * PHYSDEVBUS=usb-serial
 * SEQNUM=737
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=ttyUSB0
 * PWD=/dev
 * HUSH_VERSION=1.24.2
 *
 * and
 *
 * ACTION=add
 * DEVPATH=/class/tty/ttyUSB0
 * SUBSYSTEM=tty
 * MAJOR=188
 * MINOR=0
 * PHYSDEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.0/ttyUSB0
 * PHYSDEVBUS=usb-serial
 * PHYSDEVDRIVER=ftdi_sio
 * SEQNUM=738
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=ttyUSB0
 * PWD=/dev
 * HUSH_VERSION=1.24.2
 *
 * When a serial port disappears, mdev delivers us 2 devices:
 *
 * ACTION=remove
 * DEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.0/ttyUSB0
 * SUBSYSTEM=usb-serial
 * PHYSDEVBUS=usb-serial
 * SEQNUM=743
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=ttyUSB0
 * PWD=/dev
 * HUSH_VERSION=1.24.2
 *
 * and
 *
 * ACTION=remove
 * DEVPATH=/class/tty/ttyUSB0
 * SUBSYSTEM=tty
 * MAJOR=188
 * MINOR=0
 * PHYSDEVPATH=/devices/platform/fsl-ehci.0/usb1/1-1/1-1:1.0/ttyUSB0
 * PHYSDEVBUS=usb-serial
 * PHYSDEVDRIVER=ftdi_sio
 * SEQNUM=742
 * HOME=/
 * PATH=/sbin:/bin:/usr/sbin:/usr/bin
 * MDEV=ttyUSB0
 * PWD=/dev
 * HUSH_VERSION=1.24.2
 *
 * Since we have to modprobe ftdi_sio, we don't get anything from "mdev -s"
 * on startup.
 */


/** Information about the serial port */
struct s_dm_serial_data
{
  char *serial_port;                             /*!< The new serial port device (eg ttyUSB0) */
  t_dm_serial_callback_fn serial_callback_fn;    /*!< Function to call when a serial device is added/removed */
  void *client_callback_data;                    /*!< Pointer to data to pass to callback function */
};


/** Function to match variables for when a serial port device appears (eg. USB serial port /dev/ttyUSB0) */
static int dm_serial_match(void *client_data, const char *environment)
{
  #define SERIAL_DEV_LEN 20
  char serial_device[SERIAL_DEV_LEN];

  /* Check "SUBSYSTEM=tty" */
  if (0 != envlist_findequals(environment, "SUBSYSTEM", "tty"))
  {
    return 0;
  }

  /* Get MDEV variable */
  if (0 != envlist_find(environment, "MDEV", serial_device, SERIAL_DEV_LEN - 1))
  {
    return 0;
  }

  /* Is it 'ttyUSB[0-9]'? */
  return (0 == strncmp(serial_device, "ttyUSB", 6) &&
          '0' <= serial_device[6] &&
          '9' >= serial_device[6] &&
          '\0' == serial_device[7]) ? 1 : 0;
}


/** Handle arrival/removal of a serial port */
static int dm_serial_handler(void *client_data, enum eDeviceMounterAction action, const char *devicename, const char *environment)
{
  struct s_dm_serial_data *dsd = client_data;
  char *fulldev = NULL;
  bool result;

  /* Check pointers */
  if (NULL == dsd || NULL == devicename)
    return -1;

  result = true;
  if (dmaAdd == action)
  {
    if (NULL == dsd->serial_port)
    {
      /* Keep a copy of the device name */
      dsd->serial_port = strdup(devicename);
      if (NULL == dsd->serial_port)
      {
        rdebug_printf(LOG_ERR, "dm_serial_handler: failed to allocate memory for device name %s\n", devicename);
        result = false;
      }

      /* Get the full device path */
      fulldev = devicemounter_getfulldevicepath(devicename);
      if (NULL == fulldev)
      {
        rdebug_printf(LOG_ERR, "dm_serial_handler: failed to allocate memory for full device path when mounting device %s\n", devicename);
        result = false;
      }
    }
    else
    {
      rdebug_printf(LOG_WARNING, "dm_serial_handler: a serial port is already mounted at %s, not mounting %s\n", dsd->serial_port, devicename);
      result = false;
    }
  }
  else /* action == remove */
  {
    if (NULL != dsd->serial_port)
    {
      /* Can only remove the serial port we know about */
      if (0 == strcmp(dsd->serial_port, devicename))
      {
        WE_DEVICEMOUNTER_SERIAL_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_serial_handler: cleaning up registered serial port %s\n", dsd->serial_port);

        /* Get the full device path */
        fulldev = devicemounter_getfulldevicepath(dsd->serial_port);
        if (NULL == fulldev)
        {
          rdebug_printf(LOG_ERR, "dm_serial_handler: failed to allocate memory for full device path when removing device %s\n", devicename);
          result = false;
        }

        /* Clean up */
        free(dsd->serial_port);
        dsd->serial_port = NULL;
      }
      else
      {
        rdebug_printf(LOG_ERR, "dm_serial_handler: serial port %s is registered, %s has been disconnected, nothing to do\n", dsd->serial_port, devicename);
        result = false;
      }
    }
    else
    {
      rdebug_printf(LOG_ERR, "dm_serial_handler: no serial port registered, so can't clean up %s\n", devicename);
      result = false;
    }
  }

  /* Call the callback to notify of usb event */
  if (NULL != dsd->serial_callback_fn)
  {
    WE_DEVICEMOUNTER_SERIAL_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_serial_handler: calling client handler with result %u, device %s, action %d\n", (unsigned) result, fulldev, (unsigned) action);
    dsd->serial_callback_fn(dsd->client_callback_data, action, result, fulldev);
  }

  /* Free up full device path if allocated */
  if (NULL != fulldev)
  {
    free(fulldev);
    fulldev = NULL;
  }

  return 0;
}


/**
 * Command call back, used to remove the serial port
 *
 * @param client_data
 * @param command is one of DM_SERIAL_CMD_*
 * @param command_data is:
 *    DM_SERIAL_CMD_REMOVE: NULL
 */
static int dm_serial_command(void *client_data, int command, void *command_data)
{
  int result = -1;
  struct s_dm_serial_data *dsd = client_data;

  WE_DEVICEMOUNTER_SERIAL_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_serial_command: called\n");

  if (NULL == dsd)
    return -1;

  if (DM_SERIAL_CMD_REMOVE == command)
  {
    if (NULL != dsd->serial_port)
    {
      result = dm_serial_handler(client_data, dmaRemove, dsd->serial_port, NULL);
      WE_DEVICEMOUNTER_SERIAL_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_serial_command: result of user initiated serial port remove %d\n", result);
    }
    else
    {
      rdebug_printf(LOG_ERR, "dm_serial_destroy: no serial port available to remove\n");
    }
  }

  return result;
}


/** Destroy the serial device mounter handler */
static int dm_serial_destroy(void *client_data)
{
  struct s_dm_serial_data *dsd = client_data;
  int result = -1;

  WE_DEVICEMOUNTER_SERIAL_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_serial_destroy: called\n");

  if (NULL != dsd)
  {
    if (NULL != dsd->serial_port)
    {
      result = dm_serial_handler(client_data, dmaRemove, dsd->serial_port, NULL);
      WE_DEVICEMOUNTER_SERIAL_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_serial_destroy: result of shutdown to remove port: %d\n", result);

      /* Free up printer string (eg ttyUSB0) */
      free(dsd->serial_port);
      dsd->serial_port = NULL;
    }

    free(dsd);
  }
  return 0;
}


/** Initialise handling of serial devices */
int dm_serial_init(struct s_devicemounter *dm, t_dm_serial_callback_fn serial_callback_fn, void *client_callback_data)
{
  struct s_dm_serial_data *dsd = NULL;
  int dm_handle;

  WE_DEVICEMOUNTER_SERIAL_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_serial_init: called\n");

  /* Quick checks */
  if (NULL == dm)
  {
    goto dsi_fail;
  }

  /* Allocate memory */
  dsd = calloc(1, sizeof(struct s_dm_serial_data));
  if (NULL == dsd)
  {
    goto dsi_fail;
  }

  /* Setup members */
  dsd->serial_port = NULL;
  dsd->serial_callback_fn = serial_callback_fn;
  dsd->client_callback_data = client_callback_data;

  /* Register */
  dm_handle = devicemounter_registerhandler(dm,
                                            dm_serial_match, dm_serial_handler,
                                            dm_serial_command, dm_serial_destroy,
                                            (void *) dsd);
  if (dm_handle < 0)
  {
    goto dsi_fail;
  }

  /* Success */
  WE_DEVICEMOUNTER_SERIAL_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "dm_serial_init: success, dm_handle %d\n", dm_handle);
  return dm_handle;

  /* Fails */
dsi_fail:
  if (NULL != dsd)
  {
    free(dsd);
    dsd = NULL;
  }
  rdebug_printf(LOG_ERR, "dm_serial_init: failed\n");
  return -1;
}
