/*
 * display.c
 *
 *  Created on: 19/07/2010
 *      Author: ian
 */
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/lcd.h>
#include <rinblock/rb_we21_lcd.h>
#include <rinblock/rb_c500_lcd.h>
#include "database.h"
#include "devices.h"
#include "commonstrings.h"
#include "app_db_user.h"

#include "display.h"


static void Dputstr_dp_internal(const char *s, uint8_t dp, bool showdp)
{
  uint8_t i;

  ropen(&dev_lcd_we, 0);
  rioctl(&dev_lcd_we, IOCTL_LCD_ACCESS_CHAR);

  dp = (uint8_t) (NUM_DIGITS - dp - 1);

  /* Write the data */
  for (i = 0; (i < NUM_DIGITS && '\0' != *s); s++)
  {
    rwrite(&dev_lcd_we, s, 1, 0);
    if (i == dp && showdp)
    {
      rwrite(&dev_lcd_we, ".", 1, 0);
    }

    /* Only increment i if it wasn't a decimal point in the string */
    if ('.' != *s)
    {
      i++;
    }
  }

  /* Blank the rest */
  for (; i < NUM_DIGITS; i++)
  {
    rwrite(&dev_lcd_we, " ", 1, 0);
  }

  rclose(&dev_lcd_we);
}


/** Display a string on the display, with a decimal point */
void Dputstr_dp(const char *s, uint8_t dp)
{
  Dputstr_dp_internal(s, dp, true);
}


/**
 * Display a string on the display,
 * without a decimal point, though the string may
 * contain 1 or more, which will be displayed
 */
void Dputstr(const char *s)
{
  Dputstr_dp_internal(s, 0, false);
}


/**
 * Display a string on the display,
 * without a decimal point, though the string may
 * contain 1 or more, which will be displayed
 *
 * Any flashing will be stopped
 */
void Dputstr_nofl(const char *s)
{
  Dclear_flashing();
  Dputstr(s);
}


/** Sets the units on WE2111, but hides them if there is motion if the hide_if_motion flag is set */
static void Dset_units_we2111(enum eAdcUnits adcunits, bool motion, bool hide_if_motion)
{
  tWE21LcdSegments_Units we21_units;
  const int start = offsetof(tWE21LcdUnits, Segments);

  rlseek(&dev_lcd_we, start, SEEK_SET);
  rread(&dev_lcd_we, &we21_units, sizeof(we21_units), 0);

  we21_units.Tonnes = 0;
  we21_units.Kilograms = 0;
  we21_units.Pounds = 0;
  we21_units.Grams = 0;

  /* Only show units if not hide_if_motion, or no motion */
  if (!hide_if_motion || (hide_if_motion && !motion))
  {
    /* Units */
    switch (adcunits)
    {
      case ADC_UNITS_TONNES:
        {
          we21_units.Tonnes = 1;
        }
        break;

      case ADC_UNITS_KILOGRAMS:
        {
          we21_units.Kilograms = 1;
        }
        break;

      case ADC_UNITS_POUNDS:
        {
          we21_units.Pounds = 1;
        }
        break;

      case ADC_UNITS_GRAMS:
        {
          we21_units.Grams = 1;
        }
        break;

      /* This is a special case, if adcunits == ADC_UNITS_NONE && showing hold, then we need to flash the all the units annunciators */
      case APP_UNITS_ALL:
        {
          we21_units.Grams = 1;
          we21_units.Kilograms = 1;
          we21_units.Pounds = 1;
          we21_units.Tonnes = 1;
        }
        break;

      default:
        break;
    }
  }

  rlseek(&dev_lcd_we, start, SEEK_SET);
  rwrite(&dev_lcd_we, &we21_units, sizeof(we21_units), 0);
}


/**
 * Sets the units on C500.
 *
 * Unlike the WE2111 there is a motion annunciator, so there is no
 * need to hide the units during motion
 */
static void Dset_units_c500(enum eAdcUnits adcunits)
{
  struct s_rb_c500_disp_write_14seg segs;

  /* Read current data */
  rlseek(&dev_lcd_we, 0, SEEK_SET);
  rread(&dev_lcd_we, &segs, sizeof(segs), 0);

  /* Set bits as appropriate */
  segs.annunciators.bits.units_lb = 0;
  segs.annunciators.bits.units_kg = 0;
  segs.annunciators.bits.units_g = 0;
  segs.annunciators.bits.units_oz = 0;
  segs.annunciators.bits.units_t = 0;
  segs.annunciators.bits.units_spare = 0;

  /* Units */
  switch (adcunits)
  {
    case ADC_UNITS_TONNES:
      {
        segs.annunciators.bits.units_t = 1;
      }
      break;

    case ADC_UNITS_KILOGRAMS:
      {
        segs.annunciators.bits.units_kg = 1;
      }
      break;

    case ADC_UNITS_OUNCES:
      {
        segs.annunciators.bits.units_oz = 1;
      }
      break;

    case ADC_UNITS_POUNDS:
      {
        segs.annunciators.bits.units_lb = 1;
      }
      break;

    case ADC_UNITS_GRAMS:
      {
        segs.annunciators.bits.units_g = 1;
      }
      break;

    case ADC_UNITS_USER0:
      {
        segs.annunciators.bits.units_spare = 1;
      }
      break;

    case ADC_UNITS_POUNDS_OUNCES:
      {
        segs.annunciators.bits.units_lb = 1;
        segs.annunciators.bits.units_oz = 1;
      }
      break;

    /* This is a special case, if adcunits == ADC_UNITS_NONE && showing hold, then we need to flash the all the units annunciators */
    case APP_UNITS_ALL:
      {
        segs.annunciators.bits.units_lb = 1;
        segs.annunciators.bits.units_kg = 1;
        segs.annunciators.bits.units_g = 1;
        segs.annunciators.bits.units_oz = 1;
        segs.annunciators.bits.units_t = 1;
        segs.annunciators.bits.units_spare = 1;
      }
      break;

    default:
      break;
  }

  rlseek(&dev_lcd_we, 0, SEEK_SET);
  rwrite(&dev_lcd_we, &segs, sizeof(segs), 0);
}


/** Sets the units, but hides them if there is motion if the hide_if_motion flag is set */
void Dset_units(enum eAdcUnits adcunits, bool motion, bool hide_if_motion)
{
  ropen(&dev_lcd_we, 0);
  rioctl(&dev_lcd_we, IOCTL_LCD_ACCESS_DATA);
  if (rfilelength(&dev_lcd_we) == sizeof(struct s_rb_c500_disp_write_14seg))
    Dset_units_c500(adcunits);
  else
    Dset_units_we2111(adcunits, motion, hide_if_motion);
  rclose(&dev_lcd_we);
}


/** Set the units regardless of motion */
void Dset_units_ignore_motion(enum eAdcUnits adcunits)
{
  Dset_units(adcunits, false, false);
}


/** Set the annunciators on a WE2111 display. seg_bits contains ANNUN_* masks */
static void Dset_annunciators_we2111(uint32_t seg_bits)
{
  tWE21LcdSegments_Annunciators we21_annunciators;
  const int start = offsetof(tWE21LcdAnnunciators, Segments);

  /* Read current data */
  rlseek(&dev_lcd_we, start, SEEK_SET);
  rread(&dev_lcd_we, &we21_annunciators, sizeof(we21_annunciators), 0);

  /* Set bits as appropriate */
  we21_annunciators.Coz = (ANNUN_ZERO == (seg_bits & ANNUN_ZERO));
  we21_annunciators.Gross = (ANNUN_GROSS == (seg_bits & ANNUN_GROSS));
  we21_annunciators.Net = (ANNUN_NET == (seg_bits & ANNUN_NET));
/*
 * No motion annunciator on WE2111, units are hidden on motion:
 * we21_annunciators.Motion = (ANNUN_MOTION == (seg_bits & ANNUN_MOTION));
 */

  /* Range annunciator */
  we21_annunciators.Range1 = (ANNUN_RANGE1 == (seg_bits & ANNUN_RANGE1));
  we21_annunciators.Range2 = (ANNUN_RANGE2 == (seg_bits & ANNUN_RANGE2));

  /* IO status */
  we21_annunciators.Setpoint0 = (ANNUN_IO1 == (seg_bits & ANNUN_IO1));
  we21_annunciators.Setpoint1 = (ANNUN_IO2 == (seg_bits & ANNUN_IO2));
  we21_annunciators.Setpoint2 = (ANNUN_IO3 == (seg_bits & ANNUN_IO3));
  we21_annunciators.Setpoint3 = (ANNUN_IO4 == (seg_bits & ANNUN_IO4));
  we21_annunciators.Setpoint4 = (ANNUN_IO5 == (seg_bits & ANNUN_IO5));
  we21_annunciators.Setpoint5 = (ANNUN_IO6 == (seg_bits & ANNUN_IO6));
  we21_annunciators.Setpoint6 = (ANNUN_IO7 == (seg_bits & ANNUN_IO7));
  we21_annunciators.Setpoint7 = (ANNUN_IO8 == (seg_bits & ANNUN_IO8));

  /* Write back modified data */
  rlseek(&dev_lcd_we, start, SEEK_SET);
  rwrite(&dev_lcd_we, &we21_annunciators, sizeof(we21_annunciators), 0);
}


/** Set the annunciators on a C500 display. seg_bits contains ANNUN_* masks */
static void Dset_annunciators_c500(uint32_t seg_bits)
{
  struct s_rb_c500_disp_write_14seg segs;

  /* Read current data */
  rlseek(&dev_lcd_we, 0, SEEK_SET);
  rread(&dev_lcd_we, &segs, sizeof(segs), 0);

  /* Set bits as appropriate */
  segs.annunciators.bits.status_coz = (ANNUN_ZERO == (seg_bits & ANNUN_ZERO));
  segs.annunciators.bits.status_motion = (ANNUN_MOTION == (seg_bits & ANNUN_MOTION));
  /* There is no gross annunciator (ANNUN_GROSS == (seg_bits & ANNUN_GROSS)); */
  segs.annunciators.bits.status_net = (ANNUN_NET == (seg_bits & ANNUN_NET));
  segs.annunciators.bits.status_zeroband = (ANNUN_ZEROBAND == (seg_bits & ANNUN_ZEROBAND));
  segs.annunciators.bits.status_held = (ANNUN_HELD == (seg_bits & ANNUN_HELD));
  segs.annunciators.bits.status_spare = (ANNUN_SPARE == (seg_bits & ANNUN_SPARE));

  /* Range annunciator */
  segs.annunciators.bits.range1 = (ANNUN_RANGE1 == (seg_bits & ANNUN_RANGE1));
  segs.annunciators.bits.range2 = (ANNUN_RANGE2 == (seg_bits & ANNUN_RANGE2));
  segs.annunciators.bits.range3 = (ANNUN_RANGE3 == (seg_bits & ANNUN_RANGE3));

  /* Check weigh */
  segs.annunciators.bits.check_under2 = (ANNUN_CHECK_UNDER2 == (seg_bits & ANNUN_CHECK_UNDER2));
  segs.annunciators.bits.check_under1 = (ANNUN_CHECK_UNDER1 == (seg_bits & ANNUN_CHECK_UNDER1));
  segs.annunciators.bits.check_pass = (ANNUN_CHECK_PASS == (seg_bits & ANNUN_CHECK_PASS));
  segs.annunciators.bits.check_over1 = (ANNUN_CHECK_OVER1 == (seg_bits & ANNUN_CHECK_OVER1));
  segs.annunciators.bits.check_over2 = (ANNUN_CHECK_OVER2 == (seg_bits & ANNUN_CHECK_OVER2));

  /* IO status */
  segs.annunciators.bits.setpoint0 = (ANNUN_IO1 == (seg_bits & ANNUN_IO1));
  segs.annunciators.bits.setpoint1 = (ANNUN_IO2 == (seg_bits & ANNUN_IO2));
  segs.annunciators.bits.setpoint2 = (ANNUN_IO3 == (seg_bits & ANNUN_IO3));
  segs.annunciators.bits.setpoint3 = (ANNUN_IO4 == (seg_bits & ANNUN_IO4));
  segs.annunciators.bits.setpoint4 = (ANNUN_IO5 == (seg_bits & ANNUN_IO5));
  segs.annunciators.bits.setpoint5 = (ANNUN_IO6 == (seg_bits & ANNUN_IO6));
  segs.annunciators.bits.setpoint6 = (ANNUN_IO7 == (seg_bits & ANNUN_IO7));
  segs.annunciators.bits.setpoint7 = (ANNUN_IO8 == (seg_bits & ANNUN_IO8));

  /* Digger annunciators */
  segs.annunciators.bits.hobbin = (ANNUN_SPARE2 == (seg_bits & ANNUN_SPARE2));
  segs.annunciators.bits.nobbin = (ANNUN_SPARE3 == (seg_bits & ANNUN_SPARE3));

  /* Write back modified data */
  rlseek(&dev_lcd_we, 0, SEEK_SET);
  rwrite(&dev_lcd_we, &segs, sizeof(segs), 0);
}


/** Set the annunciators on the display. seg_bits contains ANNUN_* masks */
static void Dset_annunciators_internal(int func, uint32_t seg_bits)
{
  ropen(&dev_lcd_we, 0);
  rioctl(&dev_lcd_we, func);
  if (rfilelength(&dev_lcd_we) == sizeof(struct s_rb_c500_disp_write_14seg))
    Dset_annunciators_c500(seg_bits);
  else
    Dset_annunciators_we2111(seg_bits);
  rclose(&dev_lcd_we);
}


/** Set the annunciators on the display. seg_bits contains ANNUN_* masks */
void Dset_annunciators(uint32_t seg_bits)
{
  Dset_annunciators_internal(IOCTL_LCD_ACCESS_DATA, seg_bits);
}


/** Set annunciators on the display to flash. seg_bits contains ANNUN_* masks */
void Dflash_annunciators(uint32_t seg_bits)
{
  Dset_annunciators_internal(IOCTL_LCD_ACCESS_FLASH, seg_bits);
}


/** Convert the io bits into an annunciator bitmap */
uint32_t map_io_to_annunciators(uint8_t io)
{
  uint32_t annun;
  int i;
  const uint32_t io_lookup[NUM_SETPOINTS] = { ANNUN_IO1, ANNUN_IO2, ANNUN_IO3, ANNUN_IO4, ANNUN_IO5, ANNUN_IO6, ANNUN_IO7, ANNUN_IO8 };

  annun = 0;

  /* Put io status into annun */
  for (i = 0; i < NUM_SETPOINTS; i++)
  {
    if ((io & (0x01 << i)) != 0)
    {
      annun |= io_lookup[i];
    }
  }

  return annun;
}


/** Convert info & io_states into a bitmask of annunciators for display */
uint32_t get_annunciators(const AdcInfo *info, uint8_t io_states)
{
  uint32_t annun;

  annun = ANNUN_NONE;

  /* Gross/net annunciator */
  if (info->status.bit.grossOrNet)
  {
    annun |= (info->status.bit.gross) ? ANNUN_GROSS : ANNUN_NET;
  }

  /* Motion annunciator */
  if (info->status.bit.motion)
  {
    annun |= ANNUN_MOTION;
  }

  /* Centre of zero annunciator */
  if (info->status.bit.coz)
  {
    annun |= ANNUN_ZERO;
  }

  if (info->status.bit.zeroBand)
  {
    annun |= ANNUN_ZEROBAND;
  }

  if (info->status.bit.held)
  {
    annun |= ANNUN_HELD;
  }

  /* Range annunciator */
  if (info->status.bit.multiRange)
  {
    switch (info->status.bit.range)
    {
      case 0:
        annun |= ANNUN_RANGE1;
        break;
      case 1:
        annun |= ANNUN_RANGE2;
        break;
      case 2:
        annun |= ANNUN_RANGE3;
        break;
      default:
        break;
    }
  }

  /* Put io status into annun */
  annun |= map_io_to_annunciators(io_states);

  return annun;
}


/** Display a string will all associate information (annunciators, dp, units, io states) */
void Dshow_str_info(const char *str, const AdcInfo *info, uint8_t io_status, bool hide_units_if_motion)
{
  Dputstr(str);
  Dset_annunciators(get_annunciators(info, io_status));
  Dset_units(info->unitdp.bit.units, info->status.bit.motion, hide_units_if_motion);
}


/** Display a number on the display */
void Dshow_number(int32s weight, size_t dp, size_t min_digits)
{
  char weight_str[NUM_DIGITS + 1];
  bool minus;
  size_t i;
  int32s display_max;
  int32s display_min;
  int32s temp;
  size_t weight_width;
  size_t zeropad_width;
  size_t spacepad_width;

  /* Work out biggest and smallest number we can display */
  for (i = 0, display_max = 1; i < NUM_DIGITS; i++, display_max *= 10)
    ;
  display_max--;
  for (i = 0, display_min = -1; i < NUM_DIGITS - 1; i++, display_min *= 10)
    ;
  display_min++;

  /* Display blank/overload/underload if it won't fit, otherwise display number */
  if (dp >= NUM_DIGITS - 1 && weight < 0)
  {
    strncpy(weight_str, blank_str, NUM_DIGITS);
    weight_str[NUM_DIGITS] = '\0';
    dp = 0;
  }
  else if (weight > display_max)
  {
    strncpy(weight_str, overload_str, NUM_DIGITS);
    weight_str[NUM_DIGITS] = '\0';
    dp = 0;
  }
  else if (weight < display_min)
  {
    strncpy(weight_str, underload_str, NUM_DIGITS);
    weight_str[NUM_DIGITS] = '\0';
    dp = 0;
  }
  else
  {
    /* Deal with minus */
    minus = (weight < 0);
    if (minus)
    {
      weight = -weight;
    }

    /* Work out width of number */
    for (weight_width = 0, temp = (weight != 0) ? weight : 1; temp > 0; weight_width++, temp /= 10)
      ;

    /* Zero pad to the largest of min_digits, dp and weight_width */
    zeropad_width = (min_digits > dp + 1) ? min_digits : dp + 1;
    zeropad_width = (zeropad_width > weight_width) ? zeropad_width : weight_width;
    if (minus)
    {
      /* Limit zero padding to NUM_DIGITS - 1, so we have room for sign char: we get -.00001 */
      if (zeropad_width >= NUM_DIGITS)
        zeropad_width = NUM_DIGITS - 1;
      /* Space pad the rest up to the minus sign */
      spacepad_width = NUM_DIGITS - 1 - zeropad_width;
      snprintf(weight_str, NUM_DIGITS + 1, "-%*s%0*ld", (int) spacepad_width, "", (int) zeropad_width, (long) weight);
    }
    else
    {
      /* Space pad the rest */
      spacepad_width = NUM_DIGITS - zeropad_width;
      snprintf(weight_str, NUM_DIGITS + 1, "%*s%0*ld", (int) spacepad_width, "", (int) zeropad_width, (long) weight);
    }
  }

  /* Put weight on display */
  Dputstr_dp(weight_str, dp);
}


static const int16_t lb_div[NUM_DIGITS] = {   16, 160, 1600,  1,  1,  1 };
static const int16_t lb_max[NUM_DIGITS] = { 1000, 100,   10,  1,  1,  1 };
static const int16_t lb_min[NUM_DIGITS] = { -100, -10,   -1, -1, -1, -1 };
static const uint8_t lb_len[NUM_DIGITS] = {    3,   2,    1,  0,  0,  0 };
static const uint8_t oz_len[NUM_DIGITS] = {    2,   3,    4,  6,  6,  6 };


/** Display an oz weight as lb:oz on the display */
void Dshow_lb_oz_weight(int32s weight, int8u dp, uint8_t min_digits)
{
  char weight_str[NUM_DIGITS + 1];

  /* See if the resulting string will fit on the display */
  if (lb_len[dp] != 0 && (weight / lb_div[dp]) < lb_max[dp] && (weight / lb_div[dp]) > lb_min[dp])
  {
    /* Deal with minus */
    if (weight < 0)
    {
      weight = -weight;
      snprintf(weight_str, NUM_DIGITS + 1, "-%*ld %*.*ld",
               (int) lb_len[dp + 1], (long) (weight / lb_div[dp]),
               (int) oz_len[dp], dp + 1, (long) (weight % lb_div[dp]));
    }
    else
    {
      snprintf(weight_str, NUM_DIGITS + 1, "%*ld %*.*ld",
               (int) lb_len[dp], (long) (weight / lb_div[dp]),
               (int) oz_len[dp], dp + 1, (long) (weight % lb_div[dp]));
    }

    /* Put weight and units on display */
    Dputstr_dp(weight_str, dp);
    Dset_units_ignore_motion(ADC_UNITS_POUNDS_OUNCES);
  }
  else
  {
    /* Just show the ounce weight instead */
    Dshow_number(weight, dp, min_digits);
    Dset_units_ignore_motion(ADC_UNITS_OUNCES);
  }
}


/** Display a reading will all associate information (annunciators, dp, units, io states) */
void Dshow_rdg_info(const AdcRdg *rdg, uint8_t io_status, bool hide_units_if_motion, uint8_t min_digits)
{
  Dshow_number(rdg->rdg, rdg->info.unitdp.bit.dp, min_digits);
  Dset_annunciators(get_annunciators(&rdg->info, io_status));
  Dset_units(rdg->info.unitdp.bit.units, rdg->info.status.bit.motion, hide_units_if_motion);
}


/** Set all segments in digit to flash on the WE2111 display */
static void Dflash_segments_we2111(size_t digit, bool flashon)
{
  const int start = offsetof(tWE21LcdSegments_Digits, Digits[digit]);
  tWE21LcdSegments_Digits we21_digits;
  const size_t digitsize = sizeof(we21_digits.Digits[digit]);

  if (digit < WE21_DISPLAY_NUM_DIGITS)
  {
    memset(&we21_digits.Digits[digit], (flashon) ? 0xFF : 0x00, digitsize);
    rlseek(&dev_lcd_we, start, SEEK_SET);
    rwrite(&dev_lcd_we, &we21_digits.Digits[digit], digitsize, 0);
  }
}


/** Set all segments in digit to flash on the C500 display */
static void Dflash_segments_c500(size_t digit, bool flashon)
{
  const int start = offsetof(struct s_rb_c500_disp_write_14seg, digits[digit]);
  union s_c500_disp_digit data;

  if (digit < C500_DISPLAY_NUM_DIGITS)
  {
    memset(&data, (flashon) ? 0xFF : 0x00, sizeof(data));
    rlseek(&dev_lcd_we, start, SEEK_SET);
    rwrite(&dev_lcd_we, &data, sizeof(data), 0);
  }
}


/** Set all segments in digit to flash */
void Dflash_segments(size_t digit, bool flashon)
{
  ropen(&dev_lcd_we, 0);
  rioctl(&dev_lcd_we, IOCTL_LCD_ACCESS_FLASH);
  if (rfilelength(&dev_lcd_we) == sizeof(struct s_rb_c500_disp_write_14seg))
    Dflash_segments_c500(digit, flashon);
  else
    Dflash_segments_we2111(digit, flashon);
  rclose(&dev_lcd_we);
}


/** Enable/disable units flashing on the WE2111 */
static void Dflash_units_we2111(bool on)
{
  tWE21LcdSegments_Units we21_units;
  const int start = offsetof(tWE21LcdUnits, Segments);
  uint8_t state = (on) ? 1 : 0;

  /* Read current data */
  rlseek(&dev_lcd_we, start, SEEK_SET);
  rread(&dev_lcd_we, &we21_units, sizeof(we21_units), 0);

  /* Any of the units will flash (or not) */
  we21_units.Tonnes = state;
  we21_units.Kilograms = state;
  we21_units.Pounds = state;
  we21_units.Grams = state;

  /* Write data back */
  rlseek(&dev_lcd_we, start, SEEK_SET);
  rwrite(&dev_lcd_we, &we21_units, sizeof (we21_units), 0);
}


/** Enable/disable units flashing on the C500 */
static void Dflash_units_c500(bool on)
{
  struct s_rb_c500_disp_write_14seg segs;

  /* Read current data */
  rlseek(&dev_lcd_we, 0, SEEK_SET);
  rread(&dev_lcd_we, &segs, sizeof(segs), 0);

  /* Set bits as appropriate */
  segs.annunciators.bits.units_lb = on;
  segs.annunciators.bits.units_kg = on;
  segs.annunciators.bits.units_g = on;
  segs.annunciators.bits.units_oz = on;
  segs.annunciators.bits.units_t = on;
  segs.annunciators.bits.units_spare = on;

  /* Write back modified data */
  rlseek(&dev_lcd_we, 0, SEEK_SET);
  rwrite(&dev_lcd_we, &segs, sizeof(segs), 0);
}


/** Enable/disable units flashing */
void Dflash_units(bool on)
{
  /* Read current data */
  ropen(&dev_lcd_we, 0);
  rioctl(&dev_lcd_we, IOCTL_LCD_ACCESS_FLASH);
  if (rfilelength(&dev_lcd_we) == sizeof(struct s_rb_c500_disp_write_14seg))
    Dflash_units_c500(on);
  else
    Dflash_units_we2111(on);
  rclose(&dev_lcd_we);
}


/** Write all segments on the display */
static void Dwrite_all(int ioctl, uint8_t data)
{
  int i, len;

  ropen(&dev_lcd_we, 0);
  rioctl(&dev_lcd_we, ioctl);
  len = rfilelength(&dev_lcd_we);

  for (i = 0; i < len; i++)
  {
    rwrite(&dev_lcd_we, &data, 1, 0);
  }
  rclose(&dev_lcd_we);
}


/** Set digits specified in digit_mask to flash, and those not set to clear flashing */
void Dflash_digits(uint8_t digit_mask)
{
  uint8_t mask;
  size_t digit;
  bool on;

  for (digit = 0, mask = DFLASHMASK_D0 << (NUM_DIGITS - 1); digit < NUM_DIGITS; digit++, mask >>= 1)
  {
    on = (0 != (digit_mask & mask));
    Dflash_segments(digit, on);
  }
}


/** Turn on all segments on the display */
void Dshow_all(void)
{
  Dwrite_all(IOCTL_LCD_ACCESS_DATA, 0xFF);
}


/** Turn off all segments on the display */
void Dclear_all(void)
{
  Dwrite_all(IOCTL_LCD_ACCESS_DATA, 0x00);
}


/** Turn off all flashing on the display */
void Dclear_flashing(void)
{
  Dwrite_all(IOCTL_LCD_ACCESS_FLASH, 0x00);
}


/** Set the display brightness from the database setting */
void Dset_brightness(void)
{
  ropen(&dev_lcd_we, 0);
  rioctl(&dev_lcd_we, IOCTL_LCD_SET_BACKLIGHT, (int) db_app_user.brightness);
  rclose(&dev_lcd_we);
}


/** Setup the display */
void init_display(void)
{
  Dset_brightness();
  Dclear_all();
}
