/*
 * dsd.c
 *
 *  Created on: Mar 15, 2012
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <rinweigh/adcdsd.h>
#include <rinweigh/adcmuteximpl_pthreads.h>
#include "database.h"
#include "we_config.h"
#include "sysclock.h"
#include "app_db_factory.h"
#include "app_db_user.h"
#include "filebackup.h"
#include "dsd_export.h"
#include "we_debug.h"

#include "dsd.h"


struct s_adc_dsd *wedsd = NULL;
static AdcMutexImplementer *wedsdmutex = NULL;


/** Returns true if there is a working DSD */
bool dsd_is_running(void)
{
  return (db_app_user.dsd_enable &&
          NULL != wedsd &&
          ADC_DSD_SUCCESS == AdcDSDIsRunning(wedsd, true));
}


/** Check if the DSD is full, will return false if DSD is not present */
bool dsd_is_full(void)
{
  int32u num_records;
  int32u capacity;

  /* Get size and number of records in the DSD */
  if (NULL == wedsd ||
      ADC_DSD_SUCCESS != AdcDSDGetNumberRecords(wedsd, true, &num_records) ||
      ADC_DSD_SUCCESS != AdcDSDGetCapacity(wedsd, true, &capacity))
  {
    return false;
  }

  return (num_records >= capacity);
}


/** Purge records from the DSD */
bool dsd_purge(void)
{
  int32u num_records;
  int32u num_purge;
  int32u dsd_capacity;
  bool result;

  if (NULL == wedsd)
  {
    rdebug_printf(LOG_ERR, "dsd_purge: dsd not assigned\n");
    return false;
  }

  AdcDSDMutexLock(wedsd);

  /* Get the capacity of the DSD */
  if (ADC_DSD_SUCCESS != AdcDSDGetCapacity(wedsd, false, &dsd_capacity))
  {
    rdebug_printf(LOG_ERR, "dsd_purge: AdcDSDGetCapacity() failed\n");
    result = false;
    goto dp_fail;
  }

  /* Get number of records in the DSD */
  if (ADC_DSD_SUCCESS != AdcDSDGetNumberRecords(wedsd, false, &num_records))
  {
    rdebug_printf(LOG_ERR, "dsd_purge: AdcDSDGetNumberRecords() failed\n");
    result = false;
    goto dp_fail;
  }

  /* Work out percent of capacity */
  num_purge = (dsd_capacity * WECFG_DSD_PURGE_PERCENT) / 100;
  if (num_purge > num_records)
    num_purge = num_records;

  /* Purge them */
  if (ADC_DSD_SUCCESS != AdcDSDPurge(wedsd, false, num_purge))
  {
    rdebug_printf(LOG_ERR, "dsd_purge: AdcDSDPurge() failed. cap=%lu, recs=%lu, purge=%lu\n",
                  (unsigned long) dsd_capacity,
                  (unsigned long) num_records,
                  (unsigned long) num_purge);
    result = false;
    goto dp_fail;
  }

  rdebug_printf(LOG_INFO, "dsd_purge: purged %lu records from dsd containing %lu records and capacity of %lu\n",
                (unsigned long) num_purge,
                (unsigned long) num_records,
                (unsigned long) dsd_capacity);
  result = true;

dp_fail:
  AdcDSDMutexUnlock(wedsd);
  return result;
}


/** Write a record to the DSD */
bool dsd_write(const AdcTraceableWeight *atw)
{
  struct s_we_dsd_record dsd_record;

  /* Check DSD is running */
  if (!dsd_is_running())
  {
    return false;
  }

  /* Fill in data */
  AdcTraceablePackToNetwork(atw, &dsd_record.atwp);

  /* Write to DSD */
  return (ADC_DSD_SUCCESS == AdcDSDWriteRecord(wedsd, true, &dsd_record, sizeof(dsd_record)));
}


static bool dsd_structure_create(void)
{
  if (NULL == wedsdmutex)
  {
    /* Create the DSD mutex */
    wedsdmutex = AdcMutexImplementerPthreadsCreate();
    if (NULL == wedsdmutex)
    {
      return false;
    }
  }

  /* Create the DSD structure */
  if (NULL == wedsd)
  {
    if (ADC_DSD_SUCCESS != AdcDSDInit(&wedsd, wedsdmutex))
    {
      return false;
    }
  }

  return true;
}


/** Start up the DSD, will open existing DSD file, but won't create a new one. */
bool dsd_init(void)
{
  int32u recordsize;
  bool result;

  if (!db_app_user.dsd_enable)
  {
    return false;
  }

  /* Create the DSD mutex & structure */
  if (!dsd_structure_create())
  {
    return false;
  }

  AdcDSDMutexLock(wedsd);

  /* See if there is a DSD file already */
  if (ADC_DSD_SUCCESS != AdcDSDOpenExisting(wedsd, false, we_configuration.dsd_file))
  {
    result = false;
    goto di_fail;
  }

  /* Check some stuff */
  if (ADC_DSD_SUCCESS != AdcDSDGetRecordSize(wedsd, false, &recordsize))
  {
    result = false;
    goto di_fail;
  }
  if (WE_DSD_RECORDSIZE != recordsize ||
      wedsd->file_index.dsd_serialnumber != db_app_factory.serial_no ||
      WE_DSD_CRCSEED != wedsd->file_index.dsd_recordcrcseed)
  {
    /* TODO DSD actually do something with this problem */
    result = false;
    goto di_fail;
  }

  /* Success */
  result = true;

di_fail:
  AdcDSDMutexUnlock(wedsd);
  return result;
}


/** Clean up the DSD structure. File is closed, but not deleted */
void dsd_destroy(void)
{
  (void) AdcDSDClose(wedsd, true);

  (void) AdcDSDDestroy(wedsd);
  wedsd = NULL;

  (void) AdcMutexImplementerPthreadsDestroy(wedsdmutex);
  wedsdmutex = NULL;
}


/** Re-create a DSD file. All data will be lost. */
bool dsd_recreate_file(void)
{
  enum eAdcDsdError res;

  if (!db_app_user.dsd_enable)
  {
    return false;
  }

  if (NULL == wedsd)
  {
    /* Create the DSD structure */
    if (!dsd_structure_create())
    {
      return false;
    }
  }
  else
  {
    /* Close the/any existing one */
    (void) AdcDSDClose(wedsd, true);
  }

  /* Create a new DSD file */
  res = AdcDSDCreateNew(wedsd, true, we_configuration.dsd_file,
                        WE_DSD_RECORDSIZE, WE_DSD_MAXRECORDS,
                        WE_DSD_CRCSEED, db_app_factory.serial_no);
  return (ADC_DSD_SUCCESS == res);
}


/** Back up the DSD file to a file called "folder/WE2111_serno_date_time_dsd.csv" */
bool dsd_backup_file(const char *folder)
{
  char str[255];

  /* Create the filename */
  create_backup_filename(str, 255, folder, "dsd", ".csv");

  /* Export contents of DSD as a CSV file */
  return dsd_export(wedsd, str);
}
