/*
 * dsd_export.c
 *
 *  Created on: Jun 19, 2012
 *      Author: ian
 */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcdsd.h>
#include <rinweigh/adcpacked.h>
#include "db_build.h"

#include "dsd_export.h"


static void dsd_export_traceable_csv_headerrow(FILE *outfile)
{
  fprintf(outfile, "\"Print ID\","
                   "\"Reading\","
                   "\"Units\","
                   "\"Status\","
                   "\"Tare Weight\","
                   "\"Units\","
                   "\"Status\","
                   "\"Valid\","
                   "\"Timestamp\""
                   "\r\n");
}


/** Convert an ADC status structure into a string */
static void dsd_export_reading_status(FILE *outfile, const AdcStatus *s)
{
  if (s->bit.grossOrNet)
    fprintf(outfile, "%s", (s->bit.gross) ? "G" : "N");
  else if (s->bit.tare)
    fprintf(outfile, "T");
  else if (s->bit.preTare)
    fprintf(outfile, "PT");
}


static void dsd_export_print_weight(FILE *outfile, const AdcRdg *rdg)
{
  int divisor;
  int i;
  long remainder;
  long quotient;
  char sign;
  const char *fmtstr;
  long num;
  int dp;

  num = rdg->rdg;
  dp = rdg->info.unitdp.bit.dp;

  /* Work out sign character */
  if (num < 0)
  {
    num = -num;
    sign = '-';
  }
  else
  {
    sign = ' ';
  }

  /* Work out decimal point divisor */
  for (i = 0, divisor = 1; i < dp; i++)
  {
    divisor *= 10;
  }

  /* Work out number split at decimal point */
  quotient = num / divisor;
  remainder = num % divisor;

  /* Print the number with or without decimal point */
  if (dp > 0)
  {
    fmtstr = "%c%ld.%0*lu";
    fprintf(outfile, fmtstr, sign, quotient, dp, remainder);
  }
  else
  {
    fmtstr = "%c%ld";
    fprintf(outfile, fmtstr, sign, quotient);
  }
}


static void dsd_export_print_units(FILE *outfile, const AdcUnitDp *ud)
{
  enum eAdcUnits adc_units;
  enum eBuildUnits we_units;

  adc_units = ud->bit.units;
  we_units = xlat_adc_library_units_to_we2111(adc_units);

  fprintf(outfile, "\"%s\"", build_units_print[we_units]);
}


static void dsd_export_print_rdg(FILE *outfile, const AdcRdg *rdg)
{
  dsd_export_print_weight(outfile, rdg);
  fprintf(outfile, ",");
  dsd_export_print_units(outfile, &rdg->info.unitdp);
  fprintf(outfile, ",\"");
  dsd_export_reading_status(outfile, &rdg->info.status);
  fprintf(outfile, "\"");
}


static void dsd_export_print_timestamp(FILE *outfile, const t_datetime *dt)
{
  fprintf(outfile, "\"%04u-%02u-%02u %02u:%02u:%02u\"",
          dt->year, dt->month, dt->date,
          dt->hour, dt->minute, dt->second);
}


static void dsd_export_traceable_csv_record(FILE *outfile, const AdcTraceableWeight *atw)
{
  AdcRdg rdg;

  fprintf(outfile, "%lu,", (unsigned long) atw->printId);

  dsd_export_print_rdg(outfile, &atw->rdg); fprintf(outfile, ",");

  memset(&rdg, 0, sizeof(rdg));
  rdg.rdg = atw->tareWgt;
  rdg.info.unitdp = atw->rdg.info.unitdp;
  rdg.info.status.bit.preTare = atw->pTare;
  rdg.info.status.bit.tare = !atw->pTare;
  dsd_export_print_rdg(outfile, &rdg); fprintf(outfile, ",");

  fprintf(outfile, "\"%s\",", (atw->valid) ? "Yes" : "No");
  dsd_export_print_timestamp(outfile, &atw->dt);
  fprintf(outfile, "\r\n");
}


static bool dsd_export_record_csv(FILE *outfile, struct s_adc_dsd *dsd)
{
  int32u num_records;
  int32u i;
  AdcTraceableWeight atw;
  AdcTraceableWeightPacked atwp;

  if (ADC_DSD_SUCCESS != AdcDSDGetNumberRecords(dsd, false, &num_records))
  {
    fprintf(stderr, "ERROR: failed to get number of records in DSD\n");
    return false;
  }

  dsd_export_traceable_csv_headerrow(outfile);
  if (0 == num_records)
  {
    fprintf(outfile, "\"No records in DSD\"\r\n");
  }
  else
  {
    for (i = 0; i < num_records; i++)
    {
      if (ADC_DSD_SUCCESS == AdcDSDReadRecord(dsd, false, i, &atwp, sizeof(atwp)))
      {
        AdcTraceableUnpackFromNetwork(&atwp, &atw);
        dsd_export_traceable_csv_record(outfile, &atw);
      }
      else
      {
        fprintf(outfile, "\"Invalid record\"\r\n");
      }
    }
  }
  return true;
}


/** Export the DSD as a CSV file */
bool dsd_export(struct s_adc_dsd *dsd, const char *filename)
{
  FILE *outfile = NULL;
  bool result;

  if (NULL == dsd || NULL == filename)
  {
    return false;
  }

  AdcDSDMutexLock(dsd);

  if (ADC_DSD_SUCCESS != AdcDSDIsRunning(dsd, false))
  {
    result = false;
    goto de_fail;
  }

  outfile = fopen(filename, "w");
  if (NULL == outfile)
  {
    result = false;
    goto de_fail;
  }
  if (fcntl(fileno(outfile), F_SETFD, FD_CLOEXEC) < 0)
  {
    result = false;
    goto de_fail;
  }

  result = dsd_export_record_csv(outfile, dsd);

de_fail:
  if (NULL != outfile)
  {
    fclose(outfile);
    outfile = NULL;
  }
  AdcDSDMutexUnlock(dsd);
  return result;
}
