/*
 *  edit_ip.c
 *
 *  Created on: Sep 19, 2017
 *      Author: ian
 */
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#include <netinet/in.h>
#include "ipinfo.h"
#include "editor.h"
#include "iface.h"
#include "display.h"
#include "kbd_5000.h"

#include "edit_ip.h"


#define OCTETS_PER_V4 4


/** Increment/decrement and wrap decimal digit (digit) within an ip octet */
static void cycle_ip_digit(unsigned *octet, size_t digit, bool up_not_down)
{
  unsigned newoctet;
  unsigned editdigit;
  unsigned multiplier;
  unsigned oldoctet;
  unsigned upper;
  size_t i;

  oldoctet = *octet;

  /* Get the digit we want to edit out of the octet */
  for (i = 0, editdigit = oldoctet; i < 2 - digit; i++)
    editdigit /= 10;
  editdigit %= 10;

  /* Increment/decrement digit and wrap. Digit 0 has range 0..2, other digits 0..9 */
  upper = (0 == digit) ? 2 : 9;
  if (editdigit >= upper && up_not_down)
    editdigit = 0;
  else if (editdigit == 0 && !up_not_down)
    editdigit = upper;
  else
    editdigit += (up_not_down) ? 1 : -1;

  /* Put digit back into data */
  for (i = 0, newoctet = 0, multiplier = 1; i < 3; i++, multiplier *= 10, oldoctet /= 10)
  {
    if (i == 2 - digit)
      newoctet += editdigit * multiplier;
    else
      newoctet += (oldoctet % 10) * multiplier;
  }
  *octet = newoctet;
}


/** Put the IP address on the display, and flash the currently selected digit */
static void display_ip(const unsigned octets[OCTETS_PER_V4], size_t octet, size_t digit)
{
  const char *ptr;
  char ipstr[INET_ADDRSTRLEN];
  size_t flash_digit;

  snprintf(ipstr, INET_ADDRSTRLEN, "%03u.%03u.%03u.%03u",
           octets[0], octets[1], octets[2], octets[3]);

  ptr = &ipstr[octet / 2 * 8];
  Dputstr(ptr);

  flash_digit = (octet % 2) * 3 + digit;
  Dflash_digits(DFLASHMASK_D0 << (NUM_DIGITS - 1 - flash_digit));
}


/** Verify and convert the entered ipv4 address */
static bool verify_convert_ipv4(const unsigned octets[OCTETS_PER_V4], struct in_addr *ip_address)
{
  size_t i;
  struct in_addr ipconv;

  for (i = 0, ipconv.s_addr = 0; i < OCTETS_PER_V4; i++)
  {
    if (octets[i] <= UINT8_MAX)
      SET_IN_ADDR_OCTET(ipconv, OCTETS_PER_V4 - 1 - i, octets[i]);
    else
      return false;
  }

  *ip_address = ipconv;
  return true;
}


/** A menu editor that allows ipv4 addresses to be edited */
static enum eEditResult internal_edit_ipv4_address(struct in_addr *ip_address, bool readonly, enum e_edit_ip_endian endian)
{
  uint8_t key;
  bool quit = false;
  enum eEditResult result = EDIT_ABORT;
  size_t octet;
  size_t digit;
  size_t i;
  unsigned octets[OCTETS_PER_V4];
  struct in_addr origaddr;
  struct in_addr editaddr;

  /* Everything is done internally in host endian */
  origaddr.s_addr = (eie_network == endian) ? ntohl(ip_address->s_addr) : ip_address->s_addr;
  editaddr = origaddr;

  /*
   * Moving right we count octet=0, digit=0,1,2, octet=1, digit=0,1,2, as in:
   * Octet  0   1   2   3
   * Digit 012 012 012 012
   *       172.017.001.002
   */
  for (i = 0; i < OCTETS_PER_V4; i++)
    octets[i] = GET_IN_ADDR_OCTET(editaddr, OCTETS_PER_V4 - 1 - i);

  for (octet = 0, digit = 0; !quit_setup_menus() && !quit; )
  {
    display_ip(octets, octet, digit);

    key = get_key();
    switch (key)
    {
      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = EDIT_MENU_QUIT;
        quit = true;
        break;

      /* Change current digit */
      case KEY_EDIT_UP:
        if (!readonly)
          cycle_ip_digit(&octets[octet], digit, true);
        break;
      case KEY_EDIT_DOWN:
        if (!readonly)
          cycle_ip_digit(&octets[octet], digit, false);
        break;

      /* Move right */
      case KEY_EDIT_RIGHT:
        if (octet < 3)
        {
          if (digit < 2)
            digit++;
          else
          {
            digit = 0;
            octet++;
          }
        }
        else
        {
          if (digit < 2)
            digit++;
        }
        break;

      /* Move left */
      case KEY_EDIT_LEFT:
        if (octet > 0)
        {
          if (digit > 0)
            digit--;
          else
          {
            digit = 2;
            octet--;
          }
        }
        else
        {
          if (digit > 0)
            digit--;
        }
        break;

      /* Enter/accept */
      case KEY_EDIT_OK:
        if (readonly)
        {
          result = EDIT_OK;
          quit = true;
        }
        else if (verify_convert_ipv4(octets, &editaddr))
        {
          if (editaddr.s_addr != origaddr.s_addr)
          {
            ip_address->s_addr = (eie_network == endian) ? htonl(editaddr.s_addr) : editaddr.s_addr;
            result = EDIT_CHG;
          }
          else
          {
            result = EDIT_OK;
          }
          quit = true;
        }
        else
        {
          block_entry();
        }
        break;

      /* Cancel */
      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
        result = EDIT_ABORT;
        quit = true;
        break;
    }
  }
  Dflash_digits(DFLASHMASK_NONE);
  return result;
}


/** A menu editor that allows IPv4 addresses to be edited in the specified endian */
enum eEditResult edit_ipv4_address(struct in_addr *ip_address, enum e_edit_ip_endian endian)
{
  return internal_edit_ipv4_address(ip_address, false, endian);
}


/** A menu editor that allows IPv4 addresses to be viewed but not edited in the specified endian */
enum eEditResult view_ipv4_address(const struct in_addr *ip_address, enum e_edit_ip_endian endian)
{
  struct in_addr ip;

  ip = *ip_address;
  return internal_edit_ipv4_address(&ip, true, endian);
}
