/*
 * edit_string.c
 *
 *  Created on: Apr 15, 2013
 *      Author: ian
 */
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>
#include "editor.h"
#include "iface.h"
#include "display.h"
#include "kbd_5000.h"

#include "edit_string.h"


/** Character set of all printable characters */
const char charset_printable[] = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~";


/** Character set of all alpha numeric characters */
const char charset_alpha_numeric[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";


/** Character set of valid characters for a hostname */
const char charset_hostname[] = "abcdefghijklmnopqrstuvwxyz0123456789-";


/** Character set of valid characters for a domain name */
const char charset_domainname[] = "abcdefghijklmnopqrstuvwxyz0123456789-.";


#define WINDOW_SIZE NUM_DIGITS * 2 + 1

//static void fill_window(char window[NUM_DIGITS * 2 + 1], const char *str, size_t pos)
static void fill_window(char *window, const char *str, size_t strposn, size_t windposn)
{
  size_t i;
  size_t wpos;
  char ch;
  size_t showposn;

  memset(window, '\0', WINDOW_SIZE);

  showposn = strposn - windposn;
  wpos = 0;
  for (i = 0; i < NUM_DIGITS; i++)
  {
    ch = str[showposn];
    if ('\0' == ch)
    {
      if (i == windposn)
      {
        /* If cursor is on a '\0' null termination, then all segs on so we can see something */
        window[wpos++] = '\x0B';
      }
      break;
    }

    /* Replace "." with " ." so it displays on WE display */
    if ('.' == ch)
    {
      window[wpos++] = ' ';
    }
    if (' ' == ch)
      window[wpos++] = '_';
    else
      window[wpos++] = ch;
    showposn++;
  }
}


static void cycle_character(char *ch, const char *character_set, bool fwd_not_back)
{
  const char *cs_posn;
  size_t cs_length;

  if (NULL == character_set)
  {
    *ch = '\0';
    return;
  }
  cs_length = strlen(character_set);
  if (0 == cs_length)
  {
    *ch = '\0';
    return;
  }

  if ('\0' == *ch)
  {
    /* Cycle from '\0' termination to character set in the appropriate direction */
    *ch = (fwd_not_back) ? character_set[0] : character_set[cs_length - 1];
  }
  else
  {
    /* Cycle though the character set */
    cs_posn = strchr(character_set, *ch);
    if (NULL == cs_posn)
    {
      /* Not found so, first char in character set */
      *ch = character_set[0];
    }
    else
    {
      if (fwd_not_back)
      {
        /* Forward through character set */
        if (cs_posn == &character_set[cs_length - 1])
        {
          /* Cycle from last character in character set to '\0' null termination */
          *ch = '\0';
        }
        else
        {
          /* Next character in character set */
          *ch = cs_posn[1];
        }
      }
      else
      {
        /* Backward through character set */
        if (cs_posn == character_set)
        {
          /* Cycle from first character in character set to '\0' null termination */
          *ch = '\0';
        }
        else
        {
          /* Previous character in character set */
          *ch = cs_posn[-1];
        }
      }
    }
  }
}


/**
 * A menu editor that allows printable strings to be edited or viewed, only allowing characters from character_set
 *
 * @param str is the string to be edited or viewed
 * @param max_len is the maximum length of str (including a null termination character).
 *        ie max_len = 10 permits up to 9 printable characters and 1 null termination
 * @param character_set is the set of valid characters that can be used in the string
 * @param readonly set true if altering the string should be prevented
 *
 * @return enum eEditResult
 */
static enum eEditResult internal_edit_printable_string(char *str, size_t max_len, const char *character_set, bool readonly)
{
  uint8_t key;
  bool quit = false;
  enum eEditResult result = EDIT_ABORT;
  size_t posn;
  size_t wposn;
  size_t len;
  char window[WINDOW_SIZE];

  /* 0 out remainder of string */
  len = strlen(str);
  memset(&str[len], '\0', max_len - len);

  posn = 0;
  wposn = 0;
  while (!quit_setup_menus() && !quit)
  {
    fill_window(window, str, posn, wposn);
    Dflash_digits(DFLASHMASK_D0 << (NUM_DIGITS - 1 - wposn));
    Dputstr(window);

    key = get_key();
    switch (key)
    {
      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = EDIT_MENU_QUIT;
        quit = true;
        break;

      /* Change current digit */
      case KEY_EDIT_UP:
        if (!readonly)
          cycle_character(&str[posn], character_set, true);
        break;
      case KEY_EDIT_DOWN:
        if (!readonly)
          cycle_character(&str[posn], character_set, false);
        break;

      /* Insert a character */
      case DEKEY + KEY_EDIT_UP:
        if (!readonly)
        {
          memmove(&str[posn + 1], &str[posn], max_len - posn - 1);
          str[posn] = character_set[0];
          str[max_len - 1] = '\0';
        }
        break;

      /* Delete a character */
      case DEKEY + KEY_EDIT_DOWN:
        if (!readonly)
        {
          memmove(&str[posn], &str[posn + 1], max_len - posn - 1);
          str[max_len - 1] = '\0';
        }
        break;

      /* Move right, stopping at the null termination, and a hard limit of 1 character before the null termination at the maximum length of the string */
      case KEY_EDIT_RIGHT:
        if (posn < strlen(str) && posn < max_len - 1 - 1)
        {
          posn++;
          if (wposn < NUM_DIGITS - 1)
            wposn++;
        }
        break;

      /* Move left */
      case KEY_EDIT_LEFT:
        if (posn > 0)
        {
          posn--;
          if (wposn > 0)
            wposn--;
        }
        break;

      /* Move all the way right */
      case DEKEY + KEY_EDIT_RIGHT:
        posn = strlen(str);
        posn = (posn > 0) ? posn - 1 : posn;
        wposn = (posn < NUM_DIGITS) ? posn : NUM_DIGITS - 1;
        break;

      /* Move all the way left */
      case DEKEY + KEY_EDIT_LEFT:
        posn = 0;
        wposn = 0;
        break;

      /* Enter/accept */
      case KEY_EDIT_OK:
        result = EDIT_OK;
        quit = true;
        break;

      /* Cancel */
      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
        result = EDIT_ABORT;
        quit = true;
        break;
    }
  }
  Dflash_digits(DFLASHMASK_NONE);
  return result;
}


/**
 * A menu editor that allows printable strings to be edited, only allowing characters from character_set
 *
 * @param str is the string to be edited or viewed
 * @param max_len is the maximum length of str (including a null termination character).
 *        ie max_len = 10 permits up to 9 printable characters and 1 null termination
 * @param character_set is the set of valid characters that can be used in the string
 *
 * @return enum eEditResult
 */
enum eEditResult edit_printable_string(char *str, size_t max_len, const char *character_set)
{
  return internal_edit_printable_string(str, max_len, character_set, false);
}


/**
 * A menu editor that allows a printable string to be viewed but not edited
 *
 * @param str is the string to be edited or viewed
 * @param max_len is the maximum length of str (including a null termination character).
 *        ie max_len = 10 permits up to 9 printable characters and 1 null termination
 * @param character_set is the set of valid characters that can be used in the string
 *
 * @return enum eEditResult
 */
enum eEditResult view_printable_string(const char *str)
{
  char *mutable_str;
  enum eEditResult er;

  mutable_str = strdup(str);
  if (NULL == mutable_str)
    return EDIT_ABORT;

  er = internal_edit_printable_string(mutable_str, strlen(mutable_str), NULL, true);
  free(mutable_str);
  mutable_str = NULL;
  return er;
}
