#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "kbd_5000.h"
#include "buzz5000.h"
#include "iface.h"
#include "database.h"
#include "display.h"
#include "iface_sd.h"
#include "commonstrings.h"
#include "we_config.h"

#include "editor.h"


/** Display dashes on the display for WECFG_DELAY_FULL_OPERATING_MILLISECONDS when an incorrect value is entered */
void block_entry(void)
{
  Dclear_flashing();
  Dputstr(blank_str);
  delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
}


/**
 * Edit a list of items
 *
 * @param data is a pointer to the setting value. Not NULL
 * @param min_setting is the minimum value of the setting
 * @param max_setting is the maximum value of the setting
 * @param opts is an array of pointers to strings containing the names of each setting. Must have at least max_settings entries.
 * @param allow_edit true if the setting can be edited, false if the setting can be viewed only
 *
 * @return EDIT_CHG on value changed, EDIT_OK on no change, EDIT_ABORT on cancel, EDIT_MENU_QUIT to move up a menu level
 */
enum eEditResult edit_register_data(uint8_t *data, uint8_t min_setting, uint8_t max_setting, const char *const *opts, bool allow_edit)
{
  uint8_t temp = *data;
  uint8_t kbd;
  bool quit = false;
  enum eEditResult result = EDIT_ABORT;

  /* Can't edit in this case */
  if (min_setting > max_setting || max_setting < min_setting)
    return EDIT_ABORT;

  /* Make it safe */
  if (temp > max_setting)
    temp = max_setting;
  if (temp < min_setting)
    temp = min_setting;

  /* Run editor loop */
  while (!quit_setup_menus() && !quit)
  {
    Dputstr(opts[temp]);

    kbd = get_key();
    switch (kbd)
    {
      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = EDIT_MENU_QUIT;
        quit = true;
        break;

      /* Enter/accept */
      case KEY_EDIT_OK:
        if (allow_edit)
        {
          if (*data != temp)
          {
            *data = temp;
            result = EDIT_CHG;
          }
          else
          {
            result = EDIT_OK;
          }
        }
        else
        {
          result = EDIT_ABORT;
        }
        quit = true;
        break;

      /* Cancel */
      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
        result = EDIT_ABORT;
        quit = true;
        break;

      /* Forwards through list */
      case KEY_EDIT_UP:
        if (allow_edit)
        {
          if (temp < max_setting)
            temp++;
          else
            temp = min_setting;
        }
        else
        {
          access_denied();
        }
        break;

      /* Backwards through list */
      case KEY_EDIT_DOWN:
        if (allow_edit)
        {
          if (temp > min_setting)
            temp--;
          else
            temp = max_setting;
        }
        else
        {
          access_denied();
        }
        break;
    }
  }
  return result;
}


/** Bit list editor */
enum eEditResult edit_bits(uint8_t *data, uint8_t sets, const char *const opts[], bool full)
{
  uint8_t kbd;
  char temp_str[10];
  uint8_t data_max[NUM_DIGITS];
  uint8_t temp_data[NUM_DIGITS];
  bool quit = false;
  int i, pos;
  enum eEditResult result = EDIT_ABORT;

  for (i = 0; i < NUM_DIGITS; i++)
  {
    if (i < sets)
    {
      temp_data[i] = data[i];
      data_max[i] = (uint8_t) strlen(opts[i]);
    }
    temp_str[i] = ' ';
  }
  temp_str[i] = '\0';
  pos = sets - 1;

  Dflash_digits(DFLASHMASK_D0 << pos);

  while (!quit_setup_menus() && !quit)
  {
    for (i = 0; i < sets; i++)
      temp_str[NUM_DIGITS - 1 - i] = opts[i][temp_data[i]];
    Dputstr(temp_str);

    kbd = get_key();
    switch (kbd)
    {
      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = EDIT_MENU_QUIT;
        quit = true;
        break;

      /* Enter/accept */
      case KEY_EDIT_OK:
        if (full)
        {
          for (i = 0; i < sets; i++)
            data[i] = temp_data[i];
          result = EDIT_CHG;
        }
        else
        {
          result = EDIT_ABORT;
        }
        quit = true;
        break;

      /* Cancel */
      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
        result = EDIT_ABORT;
        quit = true;
        break;

      /* Forwards through list of options for this position */
      case KEY_EDIT_UP:
        if (full)
        {
          temp_data[pos]++;
          if (temp_data[pos] >= data_max[pos])
          {
            temp_data[pos] = 0;
          }
        }
        else
        {
          access_denied();
          Dflash_digits(DFLASHMASK_D0 << pos);
        }
        break;

      /* Backwards through list of options for this position */
      case KEY_EDIT_DOWN:
        if (full)
        {
          if (0 == temp_data[pos] || temp_data[pos] >= data_max[pos])
            temp_data[pos] = data_max[pos] - 1;
          else
            temp_data[pos]--;
        }
        else
        {
          access_denied();
          Dflash_digits(DFLASHMASK_D0 << pos);
        }
        break;

      /* Move right to next position */
      case KEY_EDIT_RIGHT:
        if (pos > 0)
        {
            pos--;
        }
        else
        {
          pos = sets - 1;
        }
        Dflash_digits(DFLASHMASK_D0 << pos);
        break;

      /* Move left to previous position */
      case KEY_EDIT_LEFT:
        if (pos >= sets - 1)
        {
            pos = 0;
        }
        else
        {
          pos++;
        }
        Dflash_digits(DFLASHMASK_D0 << pos);
        break;
    }
  }

  /* Stop digits flashing */
  Dflash_digits(DFLASHMASK_NONE);
  return result;
}


static void show_digits(uint8_t *digits, uint8_t dp, uint8_t dp2)
{
  char temp_str[10];
  uint8_t i, pos, dig;

  for (pos = 0, dig = NUM_DIGITS - 1, i = 0; i < NUM_DIGITS; i++, dig--)
  {
    if (digits[dig] > 9)
      temp_str[pos] = '-';
    else
      temp_str[pos] = (char) ('0' + digits[dig]);
    pos++;
    if ((dp2 == dig) && (dp2 > 0))
    {
      temp_str[pos] = '.';
      pos++;
    }
  }
  temp_str[pos] = '\0';
  Dputstr_dp (temp_str, dp);
}


/** Edit a 32 bit signed number */
enum eEditResult edit_long(long *val, long max_val, long min_val, uint8_t dp, uint8_t num_digits, uint8_t dp2, bool full)
{
  uint8_t kbd;
  uint8_t pos, i;
  uint8_t digits[NUM_DIGITS];
  long temp;
  bool use_minus, minus;
  bool quit = false;
  enum eEditResult result = EDIT_ABORT;

  use_minus = (min_val < 0L);
  pos = num_digits - 1;

  /* Number to string */
  temp = *val;
  if (temp < 0)
  {
    minus = true;
    temp = -temp;
  }
  else
  {
    minus = false;
  }

  for (i = 0; i < NUM_DIGITS - 1; i++)
  {
    digits[i] = (uint8_t) (temp % 10);
    temp /= 10;
  }
  if (use_minus && minus)
  {
    digits[NUM_DIGITS - 1] = 10;
  }
  else
  {
    digits[NUM_DIGITS - 1] = (uint8_t) (temp % 10);
  }
  temp = *val;

  /* Show number and flash edit position */
  show_digits(digits, dp, dp2);
  Dflash_digits(DFLASHMASK_D0 << pos);

  while (!quit_setup_menus() && !quit)
  {
    kbd = get_key();
    switch (kbd)
    {
      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = EDIT_MENU_QUIT;
        quit = true;
        break;

      /* Enter/accept */
      case KEY_EDIT_OK:
        if (full)
        {
          if (temp >= min_val && temp <= max_val)
          {
            if (*val != temp)
            {
              *val = temp;
              result = EDIT_CHG;
            }
            else
            {
              result = EDIT_OK;
            }
            quit = true;
          }
          else
          {
            block_entry();
            Dflash_digits(DFLASHMASK_D0 << pos);
          }
        }
        else
        {
          result = EDIT_ABORT;
          quit = true;
        }
        break;

      /* Cancel */
      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
        result = EDIT_ABORT;
        quit = true;
        break;

      /* Forwards through digit value (0->9->0) */
      case KEY_EDIT_UP:
        if (full)
        {
          if ((pos == NUM_DIGITS - 1) && use_minus)
          {
            if (++digits[pos] > 10)
              digits[pos] = 0;
          }
          else if (++digits[pos] > 9)
          {
            digits[pos] = 0;
          }
        }
        else
        {
          access_denied();
          Dflash_digits(DFLASHMASK_D0 << pos);
        }
        break;

      /* Backwards through digit value (9->0->9) */
      case KEY_EDIT_DOWN:
        if (full)
        {
          if ((pos == NUM_DIGITS - 1) && use_minus)
          {
            /* Wrap 0->10 (minus) or decrement */
            if (0 == digits[pos] || digits[pos] > 10)
              digits[pos] = 10;
            else
              digits[pos]--;
          }
          else
          {
            /* Wrap 0->9 or decrement */
            if (0 == digits[pos] || digits[pos] > 9)
              digits[pos] = 9;
            else
              digits[pos]--;
          }
        }
        else
        {
          access_denied();
          Dflash_digits(DFLASHMASK_D0 << pos);
        }
        break;

      /* Move cursor to the right */
      case KEY_EDIT_RIGHT:
        if (pos > 0)
        {
          pos--;
        }
        else
        {
          pos = num_digits - 1;
        }
        Dflash_digits(DFLASHMASK_D0 << pos);
        break;

      /* Move cursor to the left */
      case KEY_EDIT_LEFT:
        if (pos >= num_digits - 1)
        {
            pos = 0;
        }
        else
        {
          pos++;
        }
        Dflash_digits(DFLASHMASK_D0 << pos);
        break;
    }

    /* Put digit back into number */
    minus = false;
    temp = digits[NUM_DIGITS - 1];
    if (temp > 9)
    {
      temp = 0;
      minus = true;
    }
    for (i = 1; i < NUM_DIGITS; i++)
    {
      /*
       * new compiler (V2.10 [21 Mar 97]) didn't like:
       * temp = temp * 10 + digits[NUM_DIGITS-1-i];
       */
      temp *= 10;
      temp += digits[NUM_DIGITS - 1 - i];
    }
    if (minus)
    {
      temp = -temp;
    }
    show_digits(digits, dp, dp2);
  }

  Dflash_digits(DFLASHMASK_NONE);
  return result;
}


/** Edit a 8 bit unsigned number */
enum eEditResult edit_byte(uint8_t *val, uint8_t max_val, uint8_t min_val, uint8_t dp, uint8_t num_digits, bool full)
{
  long l;
  enum eEditResult result;

  l = (long) *val;
  result = edit_long(&l, (long) max_val, (long) min_val, dp, num_digits, 0, full);
  *val = (uint8_t) l;
  return result;
}


char *clear_leading_zeros(char *s, uint8_t dp)
{
  uint8_t cnt;
  char *p;

  /* if (strlen(s) > NUM_DIGITS) return(s); */
  p = s;
  cnt = (uint8_t) (strlen (s) - 1);
  if (*p == '-')
  {
    p++;
    cnt--;
  }

  /*
   * GET RID OF ZEROS TO THE LEFT OF THE DECIMAL
   */
  while (*p == '0')
  {
    if (cnt <= dp)
      break;
    *p++ = ' ';
    cnt--;
  }
  return s;
}


char *pad_str(char *s, char pad, uint8_t len)
{
  char *p, *n;
  uint8_t cnt;

  if (strlen(s) >= len)
    return s;

  p = s;
  cnt = 1;
  while (*p)
  {
    p++;
    cnt++;
  }
  n = &s[len];
  while (cnt > 0)
  {
    *n-- = *p--;
    cnt--;
  }
  while (n >= s)
    *n-- = pad;

  return s;
}
