/*
 * facdbtool.c
 *
 *  Created on: Mar 15, 2017
 *      Author: ian
 */
#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>
#include <endian.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <rindb/db_error.h>
#include <rindb/db_buffer.h>
#include <rindb/db_database.h>
#include <rindb/db_database_xml.h>
#include <rindb/db_database_uboot.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcdatabase.h>
#include "app_db_factory.h"
#include "indicator_model.h"
#include "apperror.h"
#include "facdb_endian.h"
#include "config.h"


#define GUESS_ENDIAN 0


static bool cmdline_display_help = false;
static bool cmdline_display_version = false;
static const char *cmdline_dbfilename = NULL;
static const char *cmdline_action = NULL;
static int cmdline_endian = GUESS_ENDIAN;
static int cmdline_output_endian = __BYTE_ORDER;
static const char *cmdline_outputfilename = NULL;
static const char *cmdline_input_type = "guess";
static const char *cmdline_output_type = "unchanged";
static int cmdline_output_version = DB_NO_MAXIMUM_VERSION;


static const int32s adc_sync_freqs_int32s[ADC_NUM_SYNC_FREQ] = ADC_SYNC_FREQUENCIES;


static const char *actual_input_type = "";


static void print_usage(const char *prog)
{
  printf("Usage: %s [OPTIONS] facdb action\n", prog);
  printf("Tool for managing a C500 factory database.\n"
         "Parameters:\n"
         "  facdb             The factory database file\n"
         "  action            What to do {print, fixendian}\n"
         "Options:\n"
         "  -e --in-endian    Import the database with this endianness (default unchanged)\n"
         "  -n --out-endian   Write the database with the endianness (default host)\n"
         "  -o --output       Write to output file\n"
         "  -t --in-type      The input factory database file type (default guess)\n"
         "  -y --out-type     The output factory database file type (default unchanged)\n"
         "  -u --out-version  Write the database at this version (default latest)\n"
         "  -v --version      Display application version\n"
         "  -h --help         Display this help\n"
         "Supported endian modes are:\n"
         "  big, little, guess, host\n"
         "\n"
         "Supported factory database types are:\n"
         "  uboot, xml, guess\n"
         "\n"
         "This tool is required when migrating a factory database between\n"
         "machines of differing endianness. This is because the ADC library\n"
         "simply calculates the checksum of the factory database without\n"
         "swapping the members to a known endian first.\n");
}


static bool parse_endian(const char *endianstr, int *endian)
{
  bool result = true;

  if (strcasecmp("big", optarg) == 0)
    *endian = __BIG_ENDIAN;
  else if (strcasecmp("little", optarg) == 0)
    *endian = __LITTLE_ENDIAN;
  else if (strcasecmp("guess", optarg) == 0)
    *endian = GUESS_ENDIAN;
  else if (strcasecmp("host", optarg) == 0)
    *endian = __BYTE_ORDER;
  else
    result = false;
  return result;
}


static bool parse_opts(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "in-endian",   1, 0, 'e' },
    { "out-endian",  1, 0, 'n' },
    { "output",      1, 0, 'o' },
    { "in-type",     1, 0, 't' },
    { "out-type",    1, 0, 'y' },
    { "out-version", 1, 0, 'u' },
    { "version",     0, 0, 'v' },
    { "help",        0, 0, 'h' },
    { NULL, 0, 0, 0 },
  };
  int c;

  while (1)
  {
    c = getopt_long(argc, argv, "e:n:o:t:y:u:vh", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      /* in-endian */
      case 'e':
        if (!parse_endian(optarg, &cmdline_endian))
        {
          fprintf(stderr, "Error: Invalid input endian option '%s'. Try --help.\n", optarg);
          return false;
        }
        break;
      /* out-endian */
      case 'n':
        if (!parse_endian(optarg, &cmdline_output_endian))
        {
          fprintf(stderr, "Error: Invalid output endian option '%s'. Try --help.\n", optarg);
          return false;
        }
        break;
      /* output file */
      case 'o':
        cmdline_outputfilename = optarg;
        break;
      /* in-type */
      case 't':
        if (strcasecmp("uboot", optarg) != 0 &&
            strcasecmp("xml", optarg) != 0 &&
            strcasecmp("guess", optarg) != 0)
        {
          fprintf(stderr, "Error: Invalid input type '%s'. Try --help.\n", optarg);
          return false;
        }
        cmdline_input_type = optarg;
        break;
      /* out-type */
      case 'y':
        if (strcasecmp("uboot", optarg) != 0 &&
            strcasecmp("xml", optarg) != 0 &&
            strcasecmp("unchanged", optarg) != 0)
        {
          fprintf(stderr, "Error: Invalid output type '%s'. Try --help.\n", optarg);
          return false;
        }
        cmdline_output_type = optarg;
        break;
      /* out-version */
      case 'u':
        cmdline_output_version = atoi(optarg);
        if (cmdline_output_version < APP_DB_F_VERSION1 ||
            cmdline_output_version > APP_DB_F_VERSION4_HW_VARIANT)
        {
          fprintf(stderr, "Error: Invalid factory database output version '%s', should be %u..%u.\n",
                  optarg, APP_DB_F_VERSION1, APP_DB_F_VERSION4_HW_VARIANT);
          return false;
        }
        break;
      /* version */
      case 'v':
        cmdline_display_version = true;
        return true;
        break;
      /* help */
      case 'h':
        cmdline_display_help = true;
        return true;
        break;
      default:
        fprintf(stderr, "Error: Unknown option. Try --help.\n");
        return false;
        break;
    }
  }

  /* Database file */
  if (optind >= argc)
  {
    fprintf(stderr, "Error: No database file specified\n");
    return false;
  }
  cmdline_dbfilename = argv[optind++];

  /* Action */
  if (optind >= argc)
  {
    fprintf(stderr, "Error: No action specified\n");
    return false;
  }
  cmdline_action = argv[optind++];

  return true;
}


static void print_version(void)
{
  printf(PACKAGE_VERSION "\n");
}


/** Implement rdebug_printf() which gets called in indicator_model.c, but not by functions that we use */
void rdebug_printf(int level, const char *format, ...)
{
  /* Do nothing */
}


/** Implement set_error() which gets called in indicator_model.c, but not by functions that we use */
void set_error(we_error_t err, const char *format, ...)
{
  /* Do nothing */
}


static bool read_uboot_factory_database(const char *filename, struct s_db_database *facdb, uint32_t *read_at_version)
{
  FILE *f = NULL;
  struct s_db_buffer buffer;
  struct s_db_database_interp app_factory_ddi_uboot;
  bool buf_inited = false;
  bool ddi_inited = false;
  enum eDBError res;

  /* Open file and create buffer */
  f = fopen(filename, "r");
  if (NULL == f)
  {
    fprintf(stderr, "read_uboot_database: open %s failed\n", filename);
    res = dbeFileFail;
    goto rufd_fail;
  }
  res = db_buffer_create_file(&buffer, f, 1024);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "read_uboot_database: db_buffer_create_file() returned %d (%s)\n", res, db_error_to_string(res));
    goto rufd_fail;
  }
  buf_inited = true;

  /* Create u-boot database interpreter */
  res = db_create_database_interp_uboot(&app_factory_ddi_uboot, &buffer);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "read_uboot_database: db_create_database_interp_uboot() returned %d (%s)\n", res, db_error_to_string(res));
    goto rufd_fail;
  }
  ddi_inited = true;

  /* Read the database */
  res = db_database_read(&app_factory_database, &app_factory_ddi_uboot, read_at_version, false);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "read_uboot_database: db_database_read() returned %d (%s)\n", res, db_error_to_string(res));
    goto rufd_fail;
  }

  /* Success */
  res = dbeSuccess;

rufd_fail:
  if (ddi_inited)
    (void) db_destroy_database_interp_uboot(&app_factory_ddi_uboot);
  if (buf_inited)
    (void) db_buffer_destroy(&buffer);
  if (NULL != f)
  {
    fclose(f);
    f = NULL;
  }

  return (dbeSuccess == res);
}


static bool read_xml_factory_database(const char *filename, struct s_db_database *facdb, uint32_t *read_at_version)
{
  struct s_db_database_interp app_factory_ddi_xml;
  enum eDBError res;

  /* Create database interpreter */
  res = db_create_database_interp_xml(&app_factory_ddi_xml, filename);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "read_xml_factory_database: db_create_database_interp_xml() returned %d (%s)\n", res, db_error_to_string(res));
    goto rxfd_fail;
  }

  /* Read the database */
  res = db_database_read(facdb, &app_factory_ddi_xml, read_at_version, false);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "read_xml_factory_database: db_database_read() returned %d (%s)\n", res, db_error_to_string(res));
    goto rxfd_fail;
  }

  /* Success */
rxfd_fail:
  (void) db_destroy_database_interp_xml(&app_factory_ddi_xml);
  return (dbeSuccess == res);
}


static bool create_read_database(const char *filename, uint32_t *read_at_version)
{
  enum eDBError res;
  bool read_uboot_result = false;
  bool read_xml_result = false;

  /* Initialise the database (creates previous data for some elements) */
  res = db_database_init(&app_factory_database);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "create_read_database: db_database_init() returned %d (%s)\n", res, db_error_to_string(res));
    goto crd_fail;
  }

  if (strcasecmp("uboot", cmdline_input_type) == 0 ||
      strcasecmp("guess", cmdline_input_type) == 0)
  {
    /* Try to read the database as uboot */
    read_uboot_result = read_uboot_factory_database(filename, &app_factory_database, read_at_version);
    if (read_uboot_result)
      actual_input_type = "uboot";
  }
  if (strcasecmp("xml", cmdline_input_type) == 0 ||
      (strcasecmp("guess", cmdline_input_type) == 0 && !read_uboot_result))
  {
    /* Try to read the database as xml */
    read_xml_result = read_xml_factory_database(filename, &app_factory_database, read_at_version);
    if (read_xml_result)
      actual_input_type = "xml";
  }
  if (!read_uboot_result && !read_xml_result)
  {
    goto crd_fail;
  }

  /* Success */
  return true;

crd_fail:
  (void) db_database_destroy(&app_factory_database);
  return false;
}


static bool write_uboot_factory_database(const char *filename, const struct s_db_database *facdb, uint32_t write_at_version)
{
  FILE *f = NULL;
  struct s_db_buffer buffer;
  struct s_db_database_interp app_factory_ddi_uboot;
  bool buf_inited = false;
  bool ddi_inited = false;
  enum eDBError res;

  /* Open file to write to */
  f = fopen(filename, "w");
  if (NULL == f)
  {
    fprintf(stderr, "write_uboot_factory_database: open %s failed\n", filename);
    res = dbeFileFail;
    goto wufd_fail;
  }

  /* Create buffer to write u-boot database into */
  res = db_buffer_create_malloc(&buffer, 1024, 1024);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "write_uboot_factory_database: db_buffer_create_file() returned %d (%s)\n", res, db_error_to_string(res));
    goto wufd_fail;
  }
  buf_inited = true;

  /* Create u-boot database interpreter */
  res = db_create_database_interp_uboot(&app_factory_ddi_uboot, &buffer);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "write_uboot_factory_database: db_create_database_interp_uboot() returned %d (%s)\n", res, db_error_to_string(res));
    goto wufd_fail;
  }
  ddi_inited = true;

  /* Write out the database into the buffer */
  res = db_database_write(facdb, &app_factory_ddi_uboot, write_at_version);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "write_uboot_factory_database: db_database_write() returned %d (%s)\n", res, db_error_to_string(res));
    goto wufd_fail;
  }

  /* Write the contents of the buffer to the file */
  if (fwrite(buffer.buffer, sizeof(uint8_t), buffer.posn, f) != buffer.posn)
  {
    fprintf(stderr, "write_uboot_factory_database: writing database to file failed\n");
    res = dbeFileFail;
    goto wufd_fail;
  }

  /* Success */
  res = dbeSuccess;

wufd_fail:
  if (ddi_inited)
    (void) db_destroy_database_interp_uboot(&app_factory_ddi_uboot);
  if (buf_inited)
    (void) db_buffer_destroy(&buffer);
  if (NULL != f)
  {
    fclose(f);
    f = NULL;
  }

  return (dbeSuccess == res);
}


static bool write_xml_factory_database(const char *filename, const struct s_db_database *facdb, uint32_t write_at_version)
{
  struct s_db_database_interp app_factory_ddi_xml;
  enum eDBError res;

  /* Create database interpreter */
  res = db_create_database_interp_xml(&app_factory_ddi_xml, filename);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "write_xml_factory_database: db_create_database_interp_xml() returned %d (%s)\n", res, db_error_to_string(res));
    goto wxfd_fail;
  }

  /* Write out the database */
  res = db_database_write(facdb, &app_factory_ddi_xml, write_at_version);
  if (dbeSuccess != res)
  {
    fprintf(stderr, "write_xml_factory_database: db_database_write() returned %d (%s)\n", res, db_error_to_string(res));
    goto wxfd_fail;
  }

  /* Success */
  res = dbeSuccess;

wxfd_fail:
  (void) db_destroy_database_interp_xml(&app_factory_ddi_xml);
  return (dbeSuccess == res);
}


static bool write_database(const struct s_db_database *facdb, const char *filename, uint32_t write_at_version)
{
  bool result;

  if (write_at_version == DB_NO_MAXIMUM_VERSION)
    write_at_version = facdb->db_database_version;

  if (strcasecmp("xml", cmdline_output_type) == 0 ||
      (strcasecmp("unchanged", cmdline_output_type) == 0 &&
       strcasecmp("xml", actual_input_type) == 0))
  {
    result = write_xml_factory_database(filename, facdb, write_at_version);

  }
  else if (strcasecmp("uboot", cmdline_output_type) == 0 ||
           (strcasecmp("unchanged", cmdline_output_type) == 0 &&
            strcasecmp("uboot", actual_input_type) == 0))
  {
    result = write_uboot_factory_database(filename, facdb, write_at_version);
  }
  else
  {
    result = false;
  }
  return result;
}


static const char *get_endian_string(int endian)
{
  const char *str;

  switch (endian)
  {
    case __BIG_ENDIAN:
      str = "big";
      break;
    case __LITTLE_ENDIAN:
      str = "little";
      break;
    default:
      str = "unknown";
      break;
  }
  return str;
}


/** Print out a factory database array */
static void print_facdb_calarray(const int32s *array, size_t len, const char *name)
{
  size_t i;

  printf("%13s: ", name);
  for (i = 0; i < len; i++)
  {
    printf("%8ld, ", (long) array[i]);
  }
  printf("\n");
}


/** Print out the ADC portion of the factory database */
static void print_adc_facdb(const AdcFactoryDatabase *facdb, int endian_for_checksum)
{
  unsigned long calc_chksum;

  printf("Host is %s endian\n", get_endian_string(__BYTE_ORDER));

  printf("AdcFactoryDatabase size: %lu\n", (unsigned long) sizeof(AdcFactoryDatabase));
  printf("AdcFactoryDatabase checksum data size: %lu\n", (unsigned long) offsetof(AdcFactoryDatabase, chksum));
  print_facdb_calarray(adc_sync_freqs_int32s, ADC_NUM_SYNC_FREQ, "syncFreqs");
  print_facdb_calarray(facdb->zeroOffset, ADC_NUM_SYNC_FREQ, "zeroOffset");
  print_facdb_calarray(facdb->zeroDriftPos, ADC_NUM_SYNC_FREQ, "zeroDriftPos");
  print_facdb_calarray(facdb->zeroDriftNeg, ADC_NUM_SYNC_FREQ, "zeroDriftNeg");
  print_facdb_calarray(facdb->spanCal, ADC_NUM_SYNC_FREQ, "spanCal");
  print_facdb_calarray(facdb->spanDriftPos, ADC_NUM_SYNC_FREQ, "spanDriftPos");
  print_facdb_calarray(facdb->spanDriftNeg, ADC_NUM_SYNC_FREQ, "spanDriftNeg");
  printf("calTemp: %lu\n", (unsigned long) facdb->calTemp);
  printf("calFlag: %02X\n", facdb->calFlag);

  printf("chksum (stored): %04lX\n", (unsigned long) facdb->chksum);
  calc_chksum = facdb_calc_checksum(facdb, endian_for_checksum);
  printf("chksum (calculated): %04lX\n", calc_chksum);
  printf("File checksum calculated with %s endian\n", get_endian_string(endian_for_checksum));
}


#define MODULE_TEST_FIRST_MASK 0x01
#define MODULE_TEST_SECOND_MASK 0x02
#define MODULE_TEST_WRITTEN_FLAG 0xA0
#define MODULE_TEST_WRITTEN_MASK 0xF0


static void print_facdb_test_result(uint8_t test_result)
{
  char test_result_str[100];
  size_t len;
  char *ptr;

  len = 100;
  ptr = test_result_str;
  if ((test_result & MODULE_TEST_WRITTEN_MASK) != MODULE_TEST_WRITTEN_FLAG)
  {
    snprintf(ptr, len, "untested");
  }
  else
  {
    /* Check result of first test */
    if (0 != (test_result & 0x04))
    {
      snprintf(ptr, len, "no initial test");
    }
    else if (0 != (test_result & 0x01))
    {
      snprintf(ptr, len, "failed initial test");
    }
    else
    {
      snprintf(ptr, len, "passed initial test");
    }
    len -= strlen(test_result_str);
    ptr = &test_result_str[100 - len];

    /* Check result of second test */
    if (0 != (test_result & 0x08))
    {
      snprintf(ptr, len, ", no second test");
    }
    else if (0 != (test_result & 0x02))
    {
      snprintf(ptr, len, ", failed second test");
    }
    else
    {
      snprintf(ptr, len, ", passed second test");
    }
  }

  printf("Test Result: 0x%02X (%s)\n", (unsigned) test_result, test_result_str);

}


/** Print out the factory database */
static void print_facdb(const struct s_db_app_factory *appfacdb, int endian_for_checksum, uint32_t db_version)
{
  printf("Factory database read at version: %lu\n", (unsigned long) db_version);

  print_adc_facdb(&appfacdb->adc_factory_database, endian_for_checksum);

  printf("Serial number: %lu\n", (unsigned long) appfacdb->serial_no);
  printf("Economy (obsolete): 0x%02X\n", (unsigned) appfacdb->economy);
  printf("Application ID: %lu\n", (unsigned long) appfacdb->application_id);
  printf("Trade Enable (obsolete): %u\n", (unsigned) appfacdb->trade_enable);
  printf("Model: %s\n", appfacdb->model_str);
  print_facdb_test_result(appfacdb->test_result);
  printf("Model Subtype: 0x%08lX (%s)\n", (unsigned long) appfacdb->model_subtype,
                                          get_indicator_model_subtype_string(appfacdb->model_subtype, imdt_long));
}


/** Write a factory database with checksum corrected for endian */
static bool fixup_write_facdb(struct s_db_app_factory *facdb_struct,
                              struct s_db_database *facdb,
                              int in_endian, int out_endian, int out_version, const char *outfilename)
{
  unsigned long in_calc_chksum;
  unsigned long out_calc_chksum;

  if (NULL == outfilename)
  {
    fprintf(stderr, "No output filename specified for action fixendian. Try --help.\n");
    return false;
  }

  printf("Host is %s endian\n", get_endian_string(__BYTE_ORDER));

  printf("input chksum (stored): %04lX\n", (unsigned long) facdb_struct->adc_factory_database.chksum);
  in_calc_chksum = facdb_calc_checksum(&facdb_struct->adc_factory_database, in_endian);
  printf("input chksum (calculated with %s endian): %04lX\n", get_endian_string(in_endian), in_calc_chksum);

  /* Set the checksum with the appropriate endian */
  out_calc_chksum = facdb_calc_checksum(&facdb_struct->adc_factory_database, out_endian);
  facdb_struct->adc_factory_database.chksum = out_calc_chksum;
  printf("output chksum (calculated with %s endian): %04lX\n", get_endian_string(out_endian), out_calc_chksum);

  /* Write to file */
  return write_database(facdb, outfilename, out_version);
}


/** Destroy the factory database */
bool destroy_factory_database(void)
{
  (void) db_database_destroy(&app_factory_database);
  return true;
}


int main(int argc, char *argv[])
{
  int result = 1;
  int in_endian;
  int out_endian;
  uint32_t read_at_version;

  /* Parse command line*/
  if (!parse_opts(argc, argv))
  {
    goto m_done;
  }

  /* Print help if requested */
  if (cmdline_display_help)
  {
    print_usage(argv[0]);
    result = 0;
    goto m_done;
  }

  /* Print version if requested */
  if (cmdline_display_version)
  {
    print_version();
    result = 0;
    goto m_done;
  }

  memset(&db_app_factory, 0, sizeof(db_app_factory));

  /* Read factory db */
  if (!create_read_database(cmdline_dbfilename, &read_at_version))
  {
    result = 1;
    goto m_done;
  }

  /* Guess endian for input */
  if (GUESS_ENDIAN == cmdline_endian)
  {
    in_endian = facdb_guess_endian(&db_app_factory.adc_factory_database);
    if (0 == in_endian)
    {
      fprintf(stderr, "Failed to guess the endianness of the factory database. Does it have a valid checksum?\n");
      result = 1;
      goto m_cleanup;
    }
  }
  else
  {
    in_endian = cmdline_endian;
  }

  /* I suppose guess -> host endian for output? */
  out_endian = (GUESS_ENDIAN == cmdline_output_endian) ? __BYTE_ORDER : cmdline_output_endian;

  /* Perform action */
  if (strcasecmp(cmdline_action, "print") == 0)
  {
    print_facdb(&db_app_factory, in_endian, read_at_version);
  }
  else if (strcasecmp(cmdline_action, "fixendian") == 0)
  {
    if (!fixup_write_facdb(&db_app_factory, &app_factory_database, in_endian, out_endian, cmdline_output_version, cmdline_outputfilename))
    {
      result = 1;
      goto m_cleanup;
    }
  }

  /* Success */
  result = 0;

m_cleanup:
  /* Clean up */
  destroy_factory_database();

m_done:
  return result;
}
