#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <pthread.h>
#include <rinhw/lcd.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include "system.h"
#include "kbd_5000.h"
#include "database.h"
#include "meas.h"
#include "backgrnd.h"
#include "adc_8bit.h"
#include "asciidef.h"
#include "version.h"
#include "editor.h"
#include "iface_sd.h"
#include "buzz5000.h"
#include "key_drvr.h"
#include "comms_print_trigger.h"
#include "comms_print_auto.h"
#include "setpoint.h"
#include "access.h"
#include "iface_ethnet.h"
#include "display.h"
#include "comms_setup.h"
#include "comms_util.h"
#include "comms_task.h"
#include "comms_auto_trigger.h"
#include "iface_serial.h"
#include "iface_factory.h"
#include "iface_special.h"
#include "iface_test.h"
#include "iface_setpoint.h"
#include "iface_analog.h"
#include "iface_clock.h"
#include "iface_build.h"
#include "iface_dsd.h"
#include "iface_cal.h"
#include "iface_option.h"
#include "remote_cal.h"
#include "iface_operator.h"
#include "iface_file.h"
#include "iface_alibi.h"
#include "device_finder_we2111.h"
#include "we_debug.h"
#include "debug_loop.h"
#include "traceprint.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "app_db_runtime_vars.h"
#include "app_db_runtime.h"
#include "iface_changelog.h"
#include "commonstrings.h"
#include "iface_menus.h"
#include "iface_usb.h"
#include "devices.h"
#include "licensing.h"
#include "model_licensing.h"

#include "iface.h"


static pthread_mutex_t display_mutex;
static bool freeze_disp = true;       /* false for free, true locks */
static bool freeze_annun = true;      /* true to stop background update of annunciators */

static bool rincmd_disp = false;       /* false for free, true locks */
static bool rincmd_annun = false;      /* true to stop background update of annunciators */

static char rincmd_buf[DISP_BUF_SIZE] = {0};

static enum eAdcUnits rincmd_units = ADC_UNITS_NONE;
static bool rincmd_auto_annun = false;
static uint32_t rincmd_disp_annun = 0;

static bool *app_quit_ptr = NULL;
bool edit_functions;            /* non-zero if calibration in progress via keyboard */
bool rear_key_entry;            /* true if the rear setup key has been used to access setup */
bool setpt_edit;                /* true if full calibration allowed, false otherwise */
bool blank_display = false;     /* true if display blanked due to tilt inputs */
bool kbd_locked = false;        /* true if key functions are locked */


struct s_display_total display_total_struct;


/** What kind of hold the display is currently showing */
enum eDisplayHold show_hold = dhNone;


/** Current ADC operation */
int8u adc_operation_counter;


/** Used to manage auto printing */
static struct s_comms_print_auto comms_print_auto;


/** Used to manage print interlock */
struct s_comms_print_interlock comms_print_interlock;


/** Get whatever weight reading is on the display, based on the hold state */
void get_displayed_adcrdg(AdcRdg *rdg)
{
  switch (show_hold)
  {
    case dhManual:
    case dhLivestock:
      *rdg = adcSuperStructure.adcHeldData.dispRdg;
      break;
    case dhPeak:
      *rdg = adcSuperStructure.adcDataWorking.peakMaxDispRdg;
      break;
    case dhNone:
    default:
      *rdg = adcSuperStructure.adcDataWorking.dispRdg;
      break;
  }
}


/** Update operation counter so errors with previous operation will get displayed */
void update_adc_operation_counter(void)
{
  adc_operation_counter = AdcGetCurrentCommandCnt(&adcSuperStructure, true);
}


/** Return true if the application has been asked to shutdown */
bool application_quit_forced(void)
{
  return (NULL != app_quit_ptr && *app_quit_ptr);
}


/** Force the application to shutdown */
void application_quit_force(void)
{
  if (NULL != app_quit_ptr)
    *app_quit_ptr = true;
}


/** Return true if the setup menus should be exited */
bool quit_setup_menus(void)
{
  return (NULL != app_quit_ptr && *app_quit_ptr);
}


/** Display "PRE.TAR" and delay */
void display_pretar_and_delay(void)
{
  /* Display "PRE.TAR" */
  Dset_annunciators(ANNUN_NONE);
  Dset_units_ignore_motion(ADC_UNITS_NONE);
  Dputstr(preset_tare_str);
  delay(1000);
}


/** Display the preset tare on the display briefly */
static void show_preset_tare(void)
{
  AdcRdg preset_tare;
  uint8_t min_digits;

  /* Display "Pre.Tar" */
  display_pretar_and_delay();

  if (adcSuperStructure.adcUserDatabase.approvalMode == ADC_APPROVAL_MODE_NTEP)
    min_digits = AdcResolutionLengths[adcSuperStructure.adcUserDatabase.resolution[0]];
  else
    min_digits = 0;

  /* Display preset tare weight, but no annunciators (since we don't have a preset tare annunciator) */
  unitsw_get_preset_tare(&preset_tare, unitsw_displaymode_to_unitswitch(adcSuperStructure.adcDataWorking.dispMode));
  Dshow_rdg_info(&preset_tare, 0, false, min_digits);
  delay(1000);
}


/* Display PT value if non-zero */
void AdcGrossNetErrorDisplay(void)
{
  if ((adcSuperStructure.adcDataWorking.presetTare.rdg != 0) &&
      (adcSuperStructure.adcRuntimeDatabase.netMode != 0))
  {
    show_preset_tare();
  }
}


/** Display any problems with tare, zero, etc */
static void AdcTareErrorDisplay(void)
{
  switch (AdcGetCurrentOpStatus(&adcSuperStructure, true))
  {
    case ADC_OP_STATUS_SUCCESS:
      switch (AdcGetCurrentCommand(&adcSuperStructure, true))
      {
        case ADC_OP_CMD_SET_PRESET_TARE:
          show_preset_tare();
          if (0 == adcSuperStructure.adcDataWorking.presetTare.rdg && ADC_APPROVAL_MODE_NONE != adcSuperStructure.adcUserDatabase.approvalMode)
            AdcToGross(&adcSuperStructure, true, false);
          else
            AdcToNet(&adcSuperStructure, true, false);
          break;
      }
      break;

    case ADC_OP_STATUS_FAILED_MOTION:
      error_strings(motion_str, err_str);
      break;

    case ADC_OP_STATUS_FAILED_BAND:
      error_strings(range_str, err_str);
      break;

    case ADC_OP_STATUS_FAILED_TRADE:
      error_strings(trade_str, err_str);
      break;

    default:
    case ADC_OP_STATUS_FAILED_ERROR:
      error_strings(adc_str, err_str);
      break;
  }
}


static void AdcTraceableReadingErrorDisplay(void)
{
  int8u cmd_status;
  int8u cmd;

  /* Get command and status */
  AdcMutexLock(&adcSuperStructure);
  cmd = AdcGetCurrentCommand(&adcSuperStructure, false);
  cmd_status = AdcGetCurrentOpStatus(&adcSuperStructure, false);
  AdcMutexUnlock(&adcSuperStructure);

  /* Display errors */
  switch (cmd_status)
  {
    case ADC_OP_STATUS_SUCCESS:
      break;
    case ADC_OP_STATUS_FAILED_MOTION:
      error_strings(motion_str, err_str);
      break;
    case ADC_OP_STATUS_FAILED_BAND:
      error_strings(range_str, err_str);
      break;
    case ADC_OP_STATUS_FAILED_TRADE:
      error_strings(trade_str, err_str);
      break;
    case ADC_OP_STATUS_FAILED_HI_RES:
      error_strings(hi_res_str, err_str);
      break;
    default:
      error_strings(adc_str, err_str);
      break;
  }

  /* If requesting a traceable reading, finish it up by entering in DSD & printing */
  if (ADC_OP_CMD_TRACEABLE_RDG == cmd)
  {
    traceable_printout_finish_ui(&adcSuperStructure.adcTraceableWeight, (ADC_OP_STATUS_SUCCESS == cmd_status));
  }
}


static void AdcManualHoldErrorDisplay(void)
{
  switch (AdcGetCurrentOpStatus(&adcSuperStructure, true))
  {
    case ADC_OP_STATUS_SUCCESS:
      show_hold = dhManual;
      break;
    default:
      error_strings(hold_str, err_str);
      break;
  }
}


/** Display weighing errors, and perform post-operation processing */
static void weigh_errors_post_processing(void)
{
  int8u cmd_cnt;
  int8u cmd_status;
  int8u cmd;

  AdcMutexLock(&adcSuperStructure);
  cmd_cnt = AdcGetCurrentCommandCnt(&adcSuperStructure, false);
  cmd_status = AdcGetCurrentOpStatus(&adcSuperStructure, false);
  cmd = AdcGetCurrentCommand(&adcSuperStructure, false);
  AdcMutexUnlock(&adcSuperStructure);

  /* Operation error messages and post-operation processing */
  if (cmd_cnt == adc_operation_counter && ADC_OP_STATUS_IN_PROGRESS != cmd_status && !blank_display)
  {

    /* Make sure that same error display cannot happen twice */
    adc_operation_counter--;

    /* Setup to display stuff */
    freeze_display_and_annunciators(true);
    Dclear_flashing();

    /* Do the error displays */
    switch (cmd)
    {
      case ADC_OP_CMD_TRACEABLE_RDG:
        AdcTraceableReadingErrorDisplay();
        break;
      case ADC_OP_CMD_SEMIAUTO_ZERO:
      case ADC_OP_CMD_SEMIAUTO_TARE:
      case ADC_OP_CMD_SET_PRESET_TARE:
      case ADC_OP_CMD_CLEAR_SEMIAUTO_ZERO:
        AdcTareErrorDisplay();
        break;
      case ADC_OP_CMD_GROSSNET_TOGGLE:
        AdcGrossNetErrorDisplay();
        break;
      case ADC_OP_CMD_MANUAL_HOLD:
        AdcManualHoldErrorDisplay();
        break;
    }

    /* Setup to display stuff */
    freeze_display_and_annunciators(false);
  }
}


/***************************************************************************/
static bool alarm_flash;


static void setp_buzzer(void)
{
  uint8_t typ;
  uint8_t on;
  static uint8_t cnt = 0;
  bool flash;
  int i, mask;
  uint8_t outputs;

  if (cnt > 0)
    cnt--;

  outputs = access_get_outputs();
  on = SETP_ALARM_NONE;
  flash = false;

  /* Work out status for each setpoint */
  for (i = 0; i < NUM_SETPOINTS; i++)
  {
    mask = 1 << i;
    typ = db_app_user.setp_alarm[i];
    if (typ == SETP_ALARM_FLASH)
    {
      if ((outputs & mask) != 0)
      {
        flash = true;
      }
    }
    else if ((on < typ) && (outputs & mask) != 0)
    {
      on = typ;
    }
  }

  switch (on)
  {
    case SETP_ALARM_SINGLE:
    case SETP_ALARM_DOUBLE:
      if (cnt == 0)
      {
        cnt = 10;
        if (on == SETP_ALARM_SINGLE)
          beep();
        else
          beep_twice();
      }
      break;

    case SETP_ALARM_CONTINUOUS:
      buzz(200, BUZZ_NORMAL);
      break;
  }

  alarm_flash = flash;
}


/** Shows current weight units, flashing if weight held on WE2111 */
static void show_units(const AdcInfo *info)
{
  enum eAdcUnits adcunits;
  bool is_c500;

  adcunits = info->unitdp.bit.units;

  ropen(&dev_lcd_we, 0);
  rioctl(&dev_lcd_we, IOCTL_LCD_ACCESS_DATA);
  is_c500 = (rfilelength(&dev_lcd_we) == sizeof(struct s_rb_c500_disp_write_14seg));
  rclose(&dev_lcd_we);

  if (!is_c500 && show_hold > dhNone)
  {
    /* Flash units if held on WE2111 */
    if (ADC_UNITS_NONE == adcunits)
      adcunits = APP_UNITS_ALL;
    Dset_units_ignore_motion(adcunits);
    Dflash_units(true);
  }
  else
  {
    /*
     * C500 has hold annunciator, so no need to flash units,
     * OR weight not held, so just display units normally
     */
    Dset_units(adcunits, info->status.bit.motion, true);
    Dflash_units(false);
  }
}


static void put_weight_on_display(AdcRdg *rdg)
{
  bool normal, ok, weight;
  uint8_t min_digits = 0;

  AdcMutexLock(&adcSuperStructure);
  normal = false;
  ok = true;
  weight = true;

  if (dhManual == show_hold)
  {
    /* Held weight */
    if (adcSuperStructure.adcDataWorking.dispMode == ADC_DISPLAY_MODE_ALT_W)
      *rdg = adcSuperStructure.adcHeldData.altDispRdg;
    else if (adcSuperStructure.adcDataWorking.dispMode == ADC_DISPLAY_MODE_TERT_W)
      *rdg = adcSuperStructure.adcHeldData.tertDispRdg;
    else
      *rdg = adcSuperStructure.adcHeldData.dispRdg;
  }
  else if (dhPeak == show_hold)
  {
    /* Peak hold weight */
    if (adcSuperStructure.adcDataWorking.dispMode == ADC_DISPLAY_MODE_ALT_W)
    {
      *rdg = adcSuperStructure.adcDataWorking.peakMaxDispRdg;
      rdg->rdg = app_final_to_final_alt(rdg->rdg);
    }
    else if (adcSuperStructure.adcDataWorking.dispMode == ADC_DISPLAY_MODE_TERT_W)
    {
      *rdg = adcSuperStructure.adcDataWorking.peakMaxDispRdg;
      rdg->rdg = app_final_to_final_tert(rdg->rdg);
    }
    else
      *rdg = adcSuperStructure.adcDataWorking.peakMaxDispRdg;
  }
  else if (dhLivestock == show_hold)
  {
    /* Live weight */
    if (adcSuperStructure.adcDataWorking.dispMode == ADC_DISPLAY_MODE_ALT_W)
      *rdg = adcSuperStructure.adcHeldData.altDispRdg;
    else if (adcSuperStructure.adcDataWorking.dispMode == ADC_DISPLAY_MODE_TERT_W)
      *rdg = adcSuperStructure.adcHeldData.tertDispRdg;
    else
      *rdg = adcSuperStructure.adcHeldData.dispRdg;
    ok = adcSuperStructure.adcHeldData.dataValid;
  }
  else
  {
    /* Normal display */
    switch (adcSuperStructure.adcDataWorking.dispMode)
    {
      case ADC_DISPLAY_MODE_PIECES:
        *rdg = adcSuperStructure.adcDataWorking.dispPieceRdg;
        break;

      case ADC_DISPLAY_MODE_ALT_W:
        *rdg = adcSuperStructure.adcDataWorking.altDispRdg;
        break;

      case ADC_DISPLAY_MODE_TERT_W:
        *rdg = adcSuperStructure.adcDataWorking.tertDispRdg;
        break;

      case ADC_DISPLAY_MODE_WEIGHT:
      default:
        *rdg = adcSuperStructure.adcDataWorking.dispRdg;
        break;
    }
    normal = true;
  }

  AdcMutexUnlock(&adcSuperStructure);

  /* Make sure the reading will fit on the display */
  if (rdg->rdg > 999999L)
    ok = false;

  if (!ok)
  {
    Dputstr(blank_str);
    Dset_units_ignore_motion(ADC_UNITS_NONE);
    Dflash_units(false);
  }
  else
  {
    if ((adcSuperStructure.adcUserDatabase.approvalMode == ADC_APPROVAL_MODE_NTEP) && weight && (min_digits == 0))
    {
      if (adcSuperStructure.adcDataWorking.dispMode == ADC_DISPLAY_MODE_ALT_W)
        min_digits = adcSuperStructure.adcInternalData.altResLen[0];
      else if (adcSuperStructure.adcDataWorking.dispMode == ADC_DISPLAY_MODE_TERT_W)
        min_digits = adcSuperStructure.adcInternalData.tertResLen[0];
      else
        min_digits = AdcResolutionLengths[adcSuperStructure.adcUserDatabase.resolution[0]];
    }

    /* Handle lb:oz */
    if ((rdg->info.unitdp.bit.units == ADC_UNITS_OUNCES) && db_app_user.show_lboz)
    {
      /* Show lb:oz string */
      Dshow_lb_oz_weight(rdg->rdg, rdg->info.unitdp.bit.dp, min_digits);
    }
    else
    {
      /* Show the weight */
      Dshow_number(rdg->rdg, rdg->info.unitdp.bit.dp, min_digits);
      /* Display the units */
      show_units(&rdg->info);
    }

    /* Flash the display */
    if (!edit_functions)
    {
      if (normal && alarm_flash)
        Dflash_digits(DFLASHMASK_ALL);
      else
        Dflash_digits(DFLASHMASK_NONE);
    }
  }
}


void do_disp(void)
{
  uint32_t annun;
  char s[10];
  AdcRdg rdg;

  pthread_mutex_lock(&display_mutex);

  memset(&rdg, 0, sizeof(rdg));

  if (!freeze_disp)
  {
    /* Handle normal operation of screen */
    if (!rincmd_disp)
    {
      setp_buzzer();
      if (blank_display)
      {
        Dputstr_nofl(blank_str);
        Dset_annunciators(ANNUN_NONE);
        Dset_units_ignore_motion(ADC_UNITS_NONE);
      }
      else if (iface_usb_show_notification())
      {
        /* Do nothing, iface_usb_show_notification() does it */
      }
      else
      {
        if (get_tenths() < 5 && APP_ERR_NONE != get_current_errors())
        {
          get_error_str(s, 10);
          Dputstr(s);
          Dset_annunciators(ANNUN_NONE);
          Dset_units_ignore_motion(ADC_UNITS_NONE);
        }
        else if (adcSuperStructure.adcDataWorking.dispRdg.info.status.bit.oload)
        {
          Dputstr(overload_str);
          if (!edit_functions)
            Dclear_flashing();

          Dset_annunciators(ANNUN_NONE);
          Dset_units_ignore_motion(ADC_UNITS_NONE);
        }
        else if (adcSuperStructure.adcDataWorking.dispRdg.info.status.bit.uload)
        {
          Dputstr(underload_str);
          if (!edit_functions)
            Dclear_flashing();

          Dset_annunciators(ANNUN_NONE);
          Dset_units_ignore_motion(ADC_UNITS_NONE);
        }
        else
        {
          if ((qa_timeout == 1) && ((get_seconds() % 2) == 0))
          {
            if (get_tenths() < 5)
              Dputstr_nofl(" CAL ");
            else
              Dputstr_nofl(" DUE ");
            Dset_annunciators(ANNUN_NONE);
            Dset_units_ignore_motion(ADC_UNITS_NONE);
          }
          else
          {
            put_weight_on_display(&rdg);
          }
        }
      }
    }
    /* Handle rincmd control of screen */
    else
    {
      Dputstr(rincmd_buf);
      Dset_units_ignore_motion(rincmd_units);
    }
  }

  if (!freeze_annun)
  {
    /* Handle normal control of annun */
    if (!rincmd_annun)
    {
      annun = get_annunciators(&rdg.info, access_get_io_states());
      Dset_annunciators(annun);
    }
    /* Handle rincmd control of annun */
    else
    {
      if (rincmd_auto_annun)
      {
        annun = get_annunciators(&adcSuperStructure.adcDataWorking.dispRdg.info, access_get_io_states());
        Dset_annunciators(annun);
      }
      else
        Dset_annunciators(rincmd_disp_annun);
    }
  }

  pthread_mutex_unlock(&display_mutex);
}


/** Freeze the display (no display updates performed) */
void freeze_display(bool freeze)
{
  pthread_mutex_lock(&display_mutex);
  freeze_disp = freeze;
  pthread_mutex_unlock(&display_mutex);
}


/** Freeze the annunciators (no annunciator updates performed) */
void freeze_annunciators(bool freeze)
{
  pthread_mutex_lock(&display_mutex);
  freeze_annun = freeze;
  pthread_mutex_unlock(&display_mutex);
}


/** Freeze both the display and annunciators (no display or annunciator updates performed) */
void freeze_display_and_annunciators(bool freeze)
{
  pthread_mutex_lock(&display_mutex);
  freeze_disp = freeze;
  freeze_annun = freeze;
  pthread_mutex_unlock(&display_mutex);
}

/** Give registers control of display and annunciators */
void register_control_display_and_annunciators(bool control)
{
  pthread_mutex_lock(&display_mutex);
  rincmd_disp = control;
  rincmd_annun = control;
  pthread_mutex_unlock(&display_mutex);
}

/** Update the register display buffer */
void write_register_display_buf(const char *buf)
{
  pthread_mutex_lock(&display_mutex);
  snprintf(rincmd_buf, DISP_BUF_SIZE, "%s", buf);
  pthread_mutex_unlock(&display_mutex);
}

/** Update the register display units */
void write_register_display_units(enum eAdcUnits units)
{
  pthread_mutex_lock(&display_mutex);
  rincmd_units = units;
  pthread_mutex_unlock(&display_mutex);
}

/** Update the register display annuns */
void write_register_display_annuns(uint32_t annun, bool auto_annun)
{
  pthread_mutex_lock(&display_mutex);
  rincmd_disp_annun = annun;
  rincmd_auto_annun = auto_annun;
  pthread_mutex_unlock(&display_mutex);
}

/** Limit cal count to WECFG_CAL_COUNT_MAXIMUM, and convert to string to display */
static void get_cal_cnt_str(char *str, int maxlen, uint32_t cal_count, char prefix)
{
  /* Limit cal count */
  if (cal_count > WECFG_CAL_COUNT_MAXIMUM)
    cal_count = WECFG_CAL_COUNT_MAXIMUM;

  /* Convert cal count to string */
  snprintf(str, maxlen, "%c%0*lu", prefix, NUM_DIGITS - 1, (unsigned long) cal_count);
}


/** Show the version and calibration count, delaying delayms after each */
static void show_version_and_cal_cnt_num(int delayms)
{
  char str[NUM_DIGITS + 1];

  Dclear_flashing();
  Dputstr(app_version_str);
  delay(delayms);
  if(adcSuperStructure.adcUserDatabase.approvalMode == ADC_APPROVAL_MODE_NTEP)
  {
    get_cal_cnt_str(str, NUM_DIGITS + 1, adcSuperStructure.adcUserDatabase.calCntNTEPCalibration, 'C');
    Dputstr(str);
    delay(delayms);
    get_cal_cnt_str(str, NUM_DIGITS + 1, adcSuperStructure.adcUserDatabase.calCntNTEPSettings, 'F');
  }
  else
  {
    get_cal_cnt_str(str, NUM_DIGITS + 1, adcSuperStructure.adcUserDatabase.calCntOIML, 'C');
  }
  Dputstr(str);
  delay(delayms);
}


static void prompt_prototype_software(bool *quit)
{
#if (IS_PROTOTYPE_SOFTWARE())
  prompt_timeout("PROTO", "ONLY", 1500, 0, quit);
#endif
}


/** Show the version and calibration count, delaying WECFG_DELAY_FULL_OPERATING_MILLISECONDS after each */
void show_version_and_cal_cnt(void)
{
  show_version_and_cal_cnt_num(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
}


/** Run the display startup sequence on first start */
static void run_startup_display_sequence(bool *quit)
{
  int i;
  char str[NUM_DIGITS + 1];
#if (WECFG_ELONGATED_STARTUP)
  const int delay_on_off_ms = WECFG_DELAY_FULL_STARTUP_MILLISECONDS;
  const int delay_count_ms = WECFG_DELAY_FULL_STARTUP_MILLISECONDS;
#else
  const int delay_on_off_ms = WECFG_DELAY_FULL_OPERATING_MILLISECONDS;
  const int delay_count_ms = WECFG_DELAY_DISPLAY_TEST_MILLISECONDS;
#endif

  /* All on */
  Dshow_all();
  delay(delay_on_off_ms);
  if (*quit)
    return;

  /* All off */
  Dclear_all();
  delay(delay_on_off_ms);
  if (*quit)
    return;

  /* Count to 9 */
  for (i = 0; i < 10; i++)
  {
    memset(str, i + '0', NUM_DIGITS);
    str[NUM_DIGITS] = '\0';
    Dputstr(str);
    delay(delay_count_ms);
    if (*quit)
      return;
  }

  /* Show version and cal count */
  show_version_and_cal_cnt();

  /* Pause for operator input for prototype software */
  prompt_prototype_software(quit);
}


void check_cal_cnt(void)
{
  char str[NUM_DIGITS + 1];

  if (adcSuperStructure.adcUserDatabase.calCntOIML >= WECFG_CAL_COUNT_MAXIMUM)
  {
    freeze_display(true);

    get_cal_cnt_str(str, NUM_DIGITS + 1, adcSuperStructure.adcUserDatabase.calCntOIML, 'C');
    Dputstr(str);

    while (1)
    {
      /* deliberately lock up the system */
      OSTimeDly(OS_TICKS_PER_SEC);
    }
  }
}


/** Display "Print" "Error" on display */
static void handle_print_error(bool aborted)
{
  freeze_display_and_annunciators(true);
  error_strings("PRINT", (aborted) ? "ABORT" : err_str);
  comms_print_flush_transmit();
  freeze_display_and_annunciators(false);
}


void check_lock(void)
{
  uint8_t lock_bits, i;

  lock_bits = 0;
  for (i = 0; i < NUM_REMOTE_KEYS; i++)
  {
    if (db_app_user.remote_f[i] == REMOTE_LOCK)
      lock_bits |= (uint8_t) (0x01 << i);
  }
  if (access_get_inputs() & lock_bits)
    kbd_locked = true;
  else
    kbd_locked = false;
}


void check_blank(void)
{
  uint8_t blank_bits, i;

  blank_bits = 0;
  for (i = 0; i < NUM_REMOTE_KEYS; i++)
  {
    if (db_app_user.remote_f[i] == REMOTE_BLANK)
      blank_bits |= (uint8_t) (0x01 << i);
  }
  if (access_get_inputs() & blank_bits)
    blank_display = true;
  else
    blank_display = false;
}


/** Display "Print" and wait until serial transmit buffer has emptied */
static void wait_print(void)
{
  uint8_t kbd;
  bool quit;

  for (quit = false; !quit && !comms_print_is_complete(); )
  {
    /* Display "Print" alternating with weight */
    if ((get_seconds() % 2) < 1)
    {
      freeze_display_and_annunciators(true);
      Dset_annunciators(ANNUN_NONE);
      Dset_units_ignore_motion(ADC_UNITS_NONE);
      Dputstr_nofl("PRINT");
    }
    else
    {
      freeze_display_and_annunciators(false);
    }

    /* Wait for keypress */
    kbd = get_key();
    switch (kbd)
    {
      case NO_KEY:
      case IDLE_CHG:
/*      case PRINT_KEY: */
        break;

      case DEFUNC1_KEY:
      case DEFUNC2_KEY:
      case DEFUNC3_KEY:
        /* Check assigned key */
        if (DE_PRINT_KEY == translate_front_func_key(true, kbd - DEKEY))
        {

          /*
           * Abort print, purge buffer, stop timeout alarm appearing
           */
          handle_print_error(true);
          quit = true;
        }
        else
        {
          beep_long();
        }
        break;

      default:
        beep_long();
        break;
    }

    /*
     * If timeout occurs, exit
     * calling routine will bring up "PRINT ERROR"
     */
    if (!quit && comms_print_has_timedout())
    {
      handle_print_error(false);
      quit = true;
    }
  }

  /* Normal display */
  freeze_display_and_annunciators(false);
}


void check_prn_sel(void)
{
  uint8_t tmp, prn_bits, mask, i, f, weight;

  tmp = 0;
  mask = 0x01;
  weight = 1;
  prn_bits = access_get_inputs ();
  for (i = 0; i < NUM_REMOTE_KEYS; i++)
  {
    f = db_app_user.remote_f[i];
    if (f == REMOTE_TXTPRN)
    {
      if (prn_bits & mask)
      {
        tmp += weight;
      }
      weight <<= 1;
    }
    mask <<= 1;
  }
  if (tmp <= APP_DB_PRINT_TXT_NUM)
  {
    print_set_prn_txt(tmp);
  }
}


void do_user_bgnd(void)
{
  if (!edit_functions)
  {
    check_remote_cal();
    comms_print_auto_check(&comms_print_auto);
    wait_print();
    check_lock();
    check_blank();
    check_prn_sel();
    comms_print_interlock_update(&comms_print_interlock);
  }
}


/** Handle a keypress within the main user interface */
void user_iface_key_dispatcher(enum eIndicatorKey kbd, bool immediate)
{
  switch (kbd)
  {
    case CAL_KEY:
      rear_cal_key_driver();
      break;

    case ZERO_KEY:
      zero_key_driver();
      break;

    case DEZERO_KEY:
      dezero_key_driver();
      break;

    case TARE_KEY:
      tare_key_driver();
      break;

    case DETARE_KEY:
      detare_key_driver();
      break;

    case GROSS_KEY:
      gross_key_driver();
      break;

    case KEY_CAL_FRONT:
      front_cal_key_driver();
      break;

    case FUNC1_KEY:
    case FUNC2_KEY:
    case FUNC3_KEY:
      function_key_driver(kbd);
      break;

    case DEFUNC1_KEY:
    case DEFUNC2_KEY:
    case DEFUNC3_KEY:
      defunction_key_driver(kbd);
      break;

    case PRINT_KEY:
      print_key_driver(immediate);
      break;

    case DE_PRINT_KEY:
      deprint_key_driver();
      break;

    case UNDO_KEY:
      if (statistics_undo_mem_plus())
      {
        comms_print_trigger_nontraceable_memory_minus();
        initiate_display_total(&display_total_struct);
      }
      break;

    case CLR_TOTAL_KEY:
      clear_total_key_driver();
      break;

    case SHOW_TOTAL_KEY:
      initiate_display_total(&display_total_struct);
      break;

    case MHOLD_KEY:
      if (!adcSuperStructure.adcHeldData.dataValid)
      {
        AdcManualHold(&adcSuperStructure, true, false);
        /* When operation completes show_hold will be set */
      }
      else
      {
        AdcClearManualHold(&adcSuperStructure, true, false);
        show_hold = dhNone;
      }
      update_adc_operation_counter();
      break;

    case PHOLD_KEY:
      if (adcSuperStructure.adcUserDatabase.approvalMode != ADC_APPROVAL_MODE_OIML && dhPeak != show_hold)
      {
        show_hold = dhPeak;
      }
      else
      {
        show_hold = dhNone;
      }
      break;

    case LHOLD_KEY:
      if (adcSuperStructure.adcUserDatabase.approvalMode != ADC_APPROVAL_MODE_OIML)
      {
        AdcLivestockWeigh(&adcSuperStructure, true, false, false);
        update_adc_operation_counter();
        show_hold = dhLivestock;
      }
      else
      {
        show_hold = dhNone;
      }
      break;

    case DE_PHOLD_KEY:
      AdcClearPeakRdg(&adcSuperStructure, true, false, true);
      update_adc_operation_counter();
      break;

    case DE_LHOLD_KEY:
#if 0 /* Not required, just change the show_hold */
      AdcClearLivestockHold(&adcSuperStructure, false);
      update_adc_operation_counter();
#endif
      show_hold = dhNone;
      break;

    case SINGLE1_KEY:
      if (!auto_single_output_trigger(SERPORT_RB1A))
      {
        beep_long();
      }
      break;

    case SINGLE2_KEY:
      if (!auto_single_output_trigger(SERPORT_RB1B))
      {
        beep_long();
      }
      break;

    case SINGLE3_KEY:
      if (!auto_single_output_trigger(SERPORT_RB2A))
      {
        beep_long();
      }
      break;

    case SINGLE4_KEY:
      if (!auto_single_output_trigger(SERPORT_RB2B))
      {
        beep_long();
      }
      break;

    case SINGLE5_KEY:
      if (!auto_single_output_trigger(SERPORT_USBGADGET))
      {
        beep_long();
      }
      break;

    case SINGLE6_KEY:
      if (!auto_single_output_trigger(SERPORT_USBHOST))
      {
        beep_long();
      }
      break;

    case UNITS_KEY:
      switch (adcSuperStructure.adcRuntimeDatabase.dispMode)
      {
        case ADC_DISPLAY_MODE_ALT_W:
          if (db_app_user.units_ter != ALT_UNITS_OFF)
            AdcSetDisplayMode(&adcSuperStructure, true, false, ADC_DISPLAY_MODE_TERT_W);
          else
            AdcSetDisplayMode(&adcSuperStructure, true, false, ADC_DISPLAY_MODE_WEIGHT);
          break;

        case ADC_DISPLAY_MODE_WEIGHT:
          if (db_app_user.units_sec != ALT_UNITS_OFF)
            AdcSetDisplayMode(&adcSuperStructure, true, false, ADC_DISPLAY_MODE_ALT_W);
          else if  (db_app_user.units_ter != ALT_UNITS_OFF)
            AdcSetDisplayMode(&adcSuperStructure, true, false, ADC_DISPLAY_MODE_TERT_W);
          break;

        case ADC_DISPLAY_MODE_PIECES:
        case ADC_DISPLAY_MODE_TERT_W:
        default:
          AdcSetDisplayMode(&adcSuperStructure, true, false, ADC_DISPLAY_MODE_WEIGHT);
          break;
      }
      break;

    default:
      break;
  }
}


/** Main user interface loop */
void do_user_iface(bool *quit)
{
  uint8_t kbd;

#if RDEBUG_ENABLED(WE_IFACE_TIMING_DEBUG_LVL, RDEBUG_LVL_FULL)
  struct s_loop_timing lt;

  loop_timing_init(&lt, 1);
#endif

  /* Run the display startup sequence */
  run_startup_display_sequence(quit);
  buzz_times(BEEP_TIME, BEEP_WAIT, BUZZ_NORMAL, 4);

  comms_print_auto_init(&comms_print_auto);
  comms_print_interlock_init(&comms_print_interlock);
  setup_for_normal_weighing();
  check_and_display_startup_zero_error();

  app_quit_ptr = quit;

  while (!*quit)
  {
    /* Check for licensed packages */
    model_licensing_check(quit);
    app_licensing_check(quit);
    weigh_errors_post_processing();
    check_cal_cnt();
    do_user_bgnd();
    if (comms_print_has_timedout())
    {
      handle_print_error(false);
    }
    run_total_display(&display_total_struct);
    /* TODO new adc library display zero tracking error */
#if 0
    if (tracking_error)
    {
      tracking_error = false;
      freeze_display_and_annunciators(true);
      error_strings(zero_str, err_str);
      freeze_display_and_annunciators(false);
    }
#endif

    /* Get key pressed */
    kbd = get_key();

    /* Deal with locked keyboard */
    if (kbd_locked)
    {
      if (NO_KEY != kbd)
        beep_long();
      kbd = NO_KEY;
    }

    /* Handle keys */
    if (NO_KEY != kbd)
    {
      /* freeze_disp = false; */
      if (qa_timeout == 1)
      {
        ack_qa_timeout();
        kbd = NO_KEY;
      }

      /* Handle keypresses */
      user_iface_key_dispatcher(kbd, false);
    }

    /* Debug readings per second */
#if RDEBUG_ENABLED(WE_IFACE_TIMING_DEBUG_LVL, RDEBUG_LVL_FULL)
    loop_timing_check(&lt, "do_user_iface");
#endif
  }

  /* On quit, display halted, no units or annunciators */
  freeze_display_and_annunciators(true);
  Dputstr_nofl("HALTED");
  Dset_annunciators(ANNUN_NONE);
  Dset_units_ignore_motion(ADC_UNITS_NONE);
}


/** Initialise the user interface */
void init_iface(void)
{
  (void) pthread_mutex_init(&display_mutex, NULL);
}


/** Shutdown the user interface */
void destroy_iface(void)
{
  (void) pthread_mutex_destroy(&display_mutex);
}
