/*
 * iface_analog.c
 *
 *  Created on: 24/03/2011
 *      Author: ian
 */
#include <stdio.h>
#include <rinhw/ucos_ii.h>
#include "display.h"
#include "access.h"
#include "kbd_5000.h"
#include "system.h"
#include "iface.h"
#include "editor.h"
#include "database.h"
#include "iface_sd.h"
#include "we_debug.h"
#include "we_config.h"
#include "app_db_user.h"
#include "commonstrings.h"
#include "dbprint.h"

#include "iface_analog.h"


/** Analog source option strings */
const char *const anal_src_setting[ANALOG_SOURCE_NUM] = ANALOG_SOURCE_STRS;


/** Analog type option strings */
const char *const anal_type_setting[ANALOG_TYPE_NUM] = ANALOG_TYPE_STRS;


static enum eMenuLevel test_analog(bool full_edit, void *user_data, size_t index)
{
  bool finished = false;
  uint8_t opt = 0;
  uint8_t kbd;
  bool changed = true;
  enum eMenuLevel result = mlNum;

  while (!quit_setup_menus() && !finished)
  {
    if (changed)
    {
      if (opt > 0)
      {
        Dputstr(hi_str);
        access_force_analogue(ANALOGOUT_FORCE_HIGH);
      }
      else
      {
        Dputstr(lo_str);
        access_force_analogue(ANALOGOUT_FORCE_LOW);
      }
      changed = false;
    }
    kbd = get_key();
    switch (kbd)
    {
      case KEY_EDIT_UP:
        if (++opt > 1)
          opt = 0;
        changed = true;
        break;
      case KEY_EDIT_DOWN:
        if (opt == 1)
          opt--;
        else
          opt = 1;
        changed = true;
        break;
      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = mlQuit;
        finished = true;
        break;
      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
      case KEY_MENU_OK:
        finished = true;
        break;
    }
  }
  access_force_analogue(ANALOGOUT_FORCE_NONE);
  return result;
}


static void adjust_and_limit_value(int32_t *pvalue, int32_t inc)
{
  int32_t value;

  value = *pvalue;
  value += inc;
  if (value > ANALOG_OUTPUT_CAL_LIMIT_COUNTS)
    value = (short) ANALOG_OUTPUT_CAL_LIMIT_COUNTS;
  if (value < -ANALOG_OUTPUT_CAL_LIMIT_COUNTS)
    value = -(short) ANALOG_OUTPUT_CAL_LIMIT_COUNTS;
  *pvalue = value;
}


static void cal_analog_display(short cal_val, bool up)
{
  int pct;
  int pctdp;
  char minus;
  char dpystr[NUM_DIGITS + 1 + 1];    /* Null terminator + decimal point */

  pct = cal_val;
  pct *= 100 * 10;      /* Give 1 decimal place */
  pct /= ANALOG_OUTPUT_DAC_FULLSCALE;
  minus = (pct < 0) ? '-' : ' ';
  if (pct < 0)
  {
    pct = -pct;
  }
  pctdp = pct % 10;
  pct /= 10;
  snprintf(dpystr, NUM_DIGITS + 1 + 1, "%s %c%1d.%1d", (up) ? "UP" : "DN", minus, pct, pctdp);
  Dputstr(dpystr);
}


static enum eMenuLevel cal_analog(bool full_edit, void *user_data, size_t index)
{
  bool anal_hi = (bool) user_data;
  bool finished = false;
  short adj;
  uint8_t kbd;
  bool dpy_changed = true;
  enum eMenuLevel result = mlNum;

  setup_kbd(KBD_CAL_ANALOG_REPEAT, KBD_CAL_ANALOG_HOLD);
  adj = ANALOG_OUTPUT_CAL_INCREMENT_COUNTS;
  access_force_analogue((anal_hi) ? ANALOGOUT_FORCE_HIGH : ANALOGOUT_FORCE_LOW);

  while (!quit_setup_menus() && !finished)
  {
    if (dpy_changed)
    {
      cal_analog_display((anal_hi) ? db_app_user.analog_span_cal : db_app_user.analog_zero_cal, (adj > 0));
      dpy_changed = false;
    }

    kbd = get_key();
    switch (kbd)
    {
      case KEY_EDIT_UP:
        if (anal_hi)
        {
          WE_IFACE_ANALOG_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "cal_analog: adjust span (%d) by %d\n", db_app_user.analog_span_cal, adj);
          adjust_and_limit_value(&db_app_user.analog_span_cal, adj);
        }
        else
        {
          WE_IFACE_ANALOG_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "cal_analog: adjust zero (%d) by %d\n", db_app_user.analog_zero_cal, adj);
          adjust_and_limit_value(&db_app_user.analog_zero_cal, adj);
        }
        dpy_changed = true;
        break;

      case KEY_EDIT_RIGHT:
        adj = -adj;
        dpy_changed = true;
        break;

      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = mlQuit;
        finished = true;
        break;
      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
      case KEY_MENU_OK:
        finished = true;
        break;
    }
  }
  access_force_analogue(ANALOGOUT_FORCE_NONE);
  setup_kbd(KBD_MENUS_REPEAT, KBD_MENUS_HOLD);
  return result;
}


/** Convert the analog calibration setting to a string */
static void to_string_analog_cal(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  bool anal_hi = (bool) user_data;
  int32_t cal_value;

  cal_value = (anal_hi) ? db_app_user.analog_span_cal : db_app_user.analog_zero_cal;
  snprintf(buffer, buflen, "%ld", (long) cal_value);
}


static enum eMenuLevel analog_edit_source(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_register_data(&db_app_user.analog_output_source,
                                                      0, ANALOG_SOURCE_NUM  - 1, anal_src_setting, ALLOW_SAFE));
}


/** Convert the analog source to a string */
static void to_string_analog_source(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", anal_src_setting[db_app_user.analog_output_source]);
}


static enum eMenuLevel analog_edit_type(bool full_edit, void *user_data, size_t index)
{
  uint8_t oldtype = db_app_user.analog_output_type;
  enum eEditResult result;

  result = edit_register_data(&db_app_user.analog_output_type,
                              0, ANALOG_TYPE_NUM - 1, anal_type_setting,
                              ALLOW_SAFE);

  /* On change, reconfigure hardware */
  if (oldtype != db_app_user.analog_output_type)
  {
    access_dbconfigure_analog();
  }
  return edit_result_to_menu_level(result);
}


/** Convert the analog type to a string */
static void to_string_analog_type(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", anal_type_setting[db_app_user.analog_output_type]);
}


static enum eMenuLevel analog_edit_abs(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_register_data(&db_app_user.analog_output_abs,
                                                      0, 1, off_on_setting, ALLOW_SAFE));
}


/** Convert the analog absolute setting to a string */
static void to_string_analog_abs(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", off_on_setting[db_app_user.analog_output_abs]);
}


static enum eMenuLevel analog_edit_clip(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_register_data(&db_app_user.analog_output_clip,
                                                      0, 1, off_on_setting, ALLOW_SAFE));
}


/** Convert the analog clip setting to a string */
static void to_string_analog_clip(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", off_on_setting[db_app_user.analog_output_clip]);
}


/** Analog menu item list */
static const struct s_menu_item sub_menu_analog_source = { .title = "SOURCE",    .sub_menu = NULL, .edit_fn = analog_edit_source, .to_string_fn = to_string_analog_source };
static const struct s_menu_item sub_menu_analog_type =   { .title = "TYPE",      .sub_menu = NULL, .edit_fn = analog_edit_type,   .to_string_fn = to_string_analog_type };
static const struct s_menu_item sub_menu_analog_abs =    { .title = "ABS",       .sub_menu = NULL, .edit_fn = analog_edit_abs,    .to_string_fn = to_string_analog_abs };
static const struct s_menu_item sub_menu_analog_clip =   { .title = "CLIP",      .sub_menu = NULL, .edit_fn = analog_edit_clip,   .to_string_fn = to_string_analog_clip };
static const struct s_menu_item sub_menu_analog_callo =  { .title = "CAL.LO",    .sub_menu = NULL, .edit_fn = cal_analog,         .to_string_fn = to_string_analog_cal, .user_data = (void *) false };
static const struct s_menu_item sub_menu_analog_calhi =  { .title = "CAL.HI",    .sub_menu = NULL, .edit_fn = cal_analog,         .to_string_fn = to_string_analog_cal, .user_data = (void *) true };
static const struct s_menu_item sub_menu_analog_force =  { .title = "FRC.ANL",   .sub_menu = NULL, .edit_fn = test_analog,        .to_string_fn = to_string_no_setting };


static const struct s_menu_item *const sub_menu_analog[] =
{
  &sub_menu_analog_source,
  &sub_menu_analog_type,
  &sub_menu_analog_abs,
  &sub_menu_analog_clip,
  &sub_menu_analog_callo,
  &sub_menu_analog_calhi,
  &sub_menu_analog_force,
  NULL
};


/** Analog menu item */
const struct s_menu_item menu_analog =
{
  .title = "ANALOG",
  .sub_menu = sub_menu_analog
};
