/*
 * iface_cal.c
 *
 *  Created on: 17/04/2011
 *      Author: ian
 */
#include <stdio.h>
#include <string.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "meas.h"
#include "iface.h"
#include "kbd_5000.h"
#include "display.h"
#include "editor.h"
#include "buzz5000.h"
#include "remote_cal.h"
#include "iface_sd.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "commonstrings.h"
#include "dbprint.h"
#include "database.h"

#include "iface_cal.h"


/** Length of linearisation menu strings */
#define LINEARISATION_MENU_STR_LEN 20


/** Typedef of array of strings for linearisation menu names */
typedef char t_sub_menu_linearisation_names[ADC_MAX_NUM_LIN_PTS][LINEARISATION_MENU_STR_LEN];


#if 0
static void show_cal_error(byte error)
{
  switch (error)
  {
    case ZERO_TOO_HI:
      error_strings(zero_str, hi_str);
      break;
    case ZERO_TOO_LO:
      error_strings(zero_str, lo_str);
      break;
    case SPAN_TOO_LO:
      error_strings(span_str, lo_str);
      break;
    case SPAN_TOO_HI:
      error_strings(span_str, hi_str);
      break;
    case INVALID_ZERO:
      error_strings(no_str, zero_str);
      break;
    default:
      error_strings("CAL", err_str);
      break;
  }
}
#endif


// returns when adc operation is finished
// parameter: timeout in os ticks
// return: adc operation status.  Note: IN_PROGRESS indicates timeout.
int8u WaitForAdcOperation(int16s ticks)
{
  int32u rdgNum;
  int8u rdgCnt = 0;

  // wait for operation to complete
  while ((ticks > 0) && (AdcGetCurrentOpStatus(&adcSuperStructure, true) == ADC_OP_STATUS_IN_PROGRESS))
  {
    OSTimeDly(5);
    ticks -= 5;
  }

  // wait for new (valid) adc reading
  rdgNum = adcSuperStructure.adcDataWorking.rdgId;
  while (ticks > 0)
  {
    if (rdgNum != adcSuperStructure.adcDataWorking.rdgId)
      rdgCnt++;
    if (rdgCnt >= 3)
      break;
    OSTimeDly(5);
    ticks -= 5;
  }

  return AdcGetCurrentOpStatus(&adcSuperStructure, true);
}


static void wait_and_display_cal_result(const char *firststr)
{
  switch(WaitForAdcOperation(OS_TICKS_PER_SEC * 30))
  {
    case ADC_OP_STATUS_SUCCESS:
      beep_thrice();
      break;
    case ADC_OP_STATUS_FAILED_RES:
      error_strings(firststr, band_str);  /* TODO NEW ADC Library - band ok, or need high and low like before? */
      break;
    case ADC_OP_STATUS_FAILED_BAND:
      error_strings(firststr, band_str);  /* TODO NEW ADC Library - band ok, or need high and low like before? */
      break;
    case ADC_OP_STATUS_IN_PROGRESS:
      error_strings(firststr, err_str);   /* TODO NEW ADC Library - timeout ? */
      break;
    case ADC_OP_STATUS_FAILED_PNT_DUPL:
      error_strings(firststr, err_str);
      break;
    default:
      error_strings(firststr, err_str);
      break;
  }
}


static enum eEditResult perform_zero_cal(bool full_edit)
{
  /* Setup display */
  Dset_annunciators(ANNUN_NONE);
  Dset_units_ignore_motion(ADC_UNITS_NONE);
  Dputstr(zero_in_progress_str);

  /* Do zero calibration */
  AdcZeroCalibration(&adcSuperStructure, true, false);
  wait_and_display_cal_result(zero_str);
  finish_calibration();

  return EDIT_CHG;
}


static enum eEditResult perform_span_cal(bool full_edit)
{
  long span_val;
  long fs;
  enum eEditResult result;

  span_val = adcSuperStructure.adcUserDatabase.spanCnt / 16;
  fs = true_full_scale();

  Dset_annunciators(ANNUN_NONE);
  Dset_units_ignore_motion(adcSuperStructure.adcUserDatabase.units);
  result = edit_long(&span_val, fs, min_span_value(), adcSuperStructure.adcUserDatabase.dpPos,
                     NUM_DIGITS, 0, full_edit);
  if (EDIT_CHG == result || EDIT_OK == result)
  {
    Dputstr(span_in_progress_str);
    Dset_units_ignore_motion(ADC_UNITS_NONE);

    (void) AdcSpanCalibration(&adcSuperStructure, true, false, span_val);
    wait_and_display_cal_result(span_str);
  }

  finish_calibration();
  return result;
}


static enum eEditResult perform_direct_zero_cal(bool full_edit)
{
  long mvv;
  enum eEditResult result;

  Dset_units_ignore_motion(ADC_UNITS_NONE);
  Dset_annunciators(ANNUN_NONE);
  mvv = ADC_RAW_TO_DUVV_ROUNDED(adcSuperStructure.adcUserDatabase.zeroRaw);
  result = edit_long(&mvv, ADC_RAW_TO_DUVV(APP_DB_U_ADCZERORAW_MAX), ADC_RAW_TO_DUVV(APP_DB_U_ADCZERORAW_MIN), 4, NUM_DIGITS, 0, full_edit);
  if (EDIT_CHG == result || EDIT_OK == result)
  {
    AdcDirectZeroCalibration(&adcSuperStructure, true, false, mvv);
    wait_and_display_cal_result(zero_str);
  }
  return result;
}


static enum eEditResult perform_direct_span_cal(bool full_edit)
{
  long mvv;
  long rdg;
  bool finished;
  enum eEditResult result = EDIT_ABORT;

  Dset_units_ignore_motion(ADC_UNITS_NONE);
  Dset_annunciators(ANNUN_NONE);

  for (finished = false; !finished; )
  {
    rdg = true_full_scale();
    mvv = ADC_RAW_TO_DUVV_ROUNDED(adcSuperStructure.adcUserDatabase.spanRaw);
    result = edit_long(&mvv, ADC_RAW_TO_DUVV(APP_DB_U_ADCSPANRAW_MAX), ADC_RAW_TO_DUVV(APP_DB_U_ADCSPANRAW_MIN), 4, NUM_DIGITS, 0, full_edit);
    if (EDIT_OK == result || EDIT_CHG == result)
    {
      AdcDirectSpanCalibration(&adcSuperStructure, true, false, rdg, mvv);
      wait_and_display_cal_result(span_str);
      finished = true;
    }
    else //if (EDIT_MENU_QUIT == result)
    {
      finished = true;
    }
  }

  return result;
}


/** Function typedef for all calibration functions */
typedef enum eEditResult (*t_cal_function_ptr)(bool full_edit);


static enum eEditResult do_calibration(t_cal_function_ptr cal_fn, bool full_edit)
{
  bool quit;
  uint8_t kbd;
  enum eEditResult result = EDIT_OK;

  /* Setup ADC for calibration */
  AdcSetCompValues(&adcSuperStructure, true, 0, ADC_RAW_MULT_UNITY);
  AdcReInit(&adcSuperStructure, true, NULL, NULL, NULL);
  set_gross(true);

  /* Show weight on display */
  freeze_display_and_annunciators(false);

  /* Run calibration */
  Dflash_digits(DFLASHMASK_ALL);
  for (quit = false; !quit_setup_menus() && !quit; )
  {
    kbd = get_key();
    switch (kbd)
    {

      /* Do the calibration */
      case KEY_EDIT_OK:
        freeze_display_and_annunciators(true);
        Dflash_digits(DFLASHMASK_NONE);

        result = cal_fn(full_edit);
        if (EDIT_ABORT == result || EDIT_MENU_QUIT == result)
        {
          /* Calibration aborted */
          quit = true;
        }

        /* Put weight on display */
        freeze_display_and_annunciators(false);
        Dflash_digits(DFLASHMASK_ALL);
        break;

      /* Exit menus and save */
      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = EDIT_MENU_QUIT;
        quit = true;
        break;

      /* Exit calibration */
      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
        result = EDIT_ABORT;
        quit = true;
        break;

      /* These keys quit the calibration */
      case KEY_EDIT_DOWN:
      case KEY_EDIT_LEFT:
      case KEY_EDIT_RIGHT:
      case KEY_EDIT_UP:
        result = EDIT_OK;
        quit = true;
        break;
    }
  }

  /* Freeze display for menus */
  freeze_display_and_annunciators(true);
  Dflash_digits(DFLASHMASK_NONE);
  Dset_annunciators(ANNUN_NONE);
  Dset_units_ignore_motion(ADC_UNITS_NONE);
  return result;
}


static enum eMenuLevel do_normal_zero_cal(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(do_calibration(perform_zero_cal, full_edit));
}


/** Convert a raw adc number into a mv/v string */
static void to_string_mvv(int32_t raw, char *buffer, size_t buflen)
{
  AdcRdg rdg;

  /* zeroRaw = mvv is millivolts per volt * 256 * 10000 */
  memset(&rdg, 0, sizeof(rdg));
  rdg.rdg = ADC_RAW_TO_DUVV_ROUNDED(raw);
  rdg.info.unitdp.bit.dp = 4;
  rdg.info.unitdp.bit.units = ADC_UNITS_USER7;
  to_string_weight(&rdg, true, buffer, buflen);
}


#define MVV_BUF_LEN 20


/** Convert the zero calibration to a string */
static void to_string_normal_zero_cal(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  char tmp_mvv_buf[MVV_BUF_LEN];

  to_string_mvv(adcSuperStructure.adcUserDatabase.zeroRaw, tmp_mvv_buf, MVV_BUF_LEN);
  snprintf(buffer, buflen, "RAW %s", tmp_mvv_buf);
}


static enum eMenuLevel do_normal_span_cal(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(do_calibration(perform_span_cal, full_edit));
}


/** Convert the span calibration to a string */
static void to_string_normal_span_cal(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  char tmp_mvv_buf[MVV_BUF_LEN];
  char tmp_weight_buf[MVV_BUF_LEN];
  AdcRdg rdg;

  to_string_mvv(adcSuperStructure.adcUserDatabase.spanRaw, tmp_mvv_buf, MVV_BUF_LEN);

  memset(&rdg, 0, sizeof(rdg));
  rdg.rdg = adcSuperStructure.adcUserDatabase.calWeight;
  rdg.info.unitdp.bit.dp = adcSuperStructure.adcUserDatabase.dpPos;
  rdg.info.unitdp.bit.units = adcSuperStructure.adcUserDatabase.units;
  to_string_weight(&rdg, true, tmp_weight_buf, MVV_BUF_LEN);

  snprintf(buffer, buflen, "RAW %s WEIGHT %s", tmp_mvv_buf, tmp_weight_buf);
}


static enum eMenuLevel do_direct_zero_cal(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(do_calibration(perform_direct_zero_cal, full_edit));
}


static enum eMenuLevel do_direct_span_cal(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(do_calibration(perform_direct_span_cal, full_edit));
}


#if 0
static enum eMenuLevel do_zero_cal(bool full_edit, void *user_data, size_t index)
{
  enum eEditResult result;

  if (CAL_TYPE_DIRECT == db_app_user.cal_type)
    result = do_direct_zero_cal(full_edit);
  else
    result = do_normal_zero_cal(full_edit);
  return edit_result_to_menu_level(result);
}


static enum eMenuLevel do_span_cal(bool full_edit, void *user_data, size_t index)
{
  enum eEditResult result;

  if (CAL_TYPE_DIRECT == db_app_user.cal_type)
    result = do_direct_span_cal(full_edit);
  else
    result = do_normal_span_cal(full_edit);
  return edit_result_to_menu_level(result);
}
#endif


/** Create name for linearisation point */
static void update_linpt_name(char *name, size_t linpt)
{
  long full_scale_point;
  long pt_val;
  long pt_pct;

  full_scale_point = AdcGetLinPt(&adcSuperStructure, true, AdcFinalToCounts(&adcSuperStructure, true, adcSuperStructure.adcUserDatabase.fullScale[(adcSuperStructure.adcUserDatabase.scaleType + 1) / 2], false));
  full_scale_point -= 128;

  /* Get linearisation point value */
  pt_val = adcSuperStructure.adcUserDatabase.linPts[linpt];

  /* Convert to percent of full scale linearisation point */
  pt_pct = ((200 * (pt_val - 128) / full_scale_point) + 1) / 2;

  /* Set up menu string for each point */
  if (0 == pt_pct)
  {
    /* No point set */
    snprintf(name, LINEARISATION_MENU_STR_LEN, "L%u. ---", (unsigned int) linpt + 1);
  }
  else
  {
    /* Point set */
    snprintf(name, LINEARISATION_MENU_STR_LEN, "L%u.%4ld", (unsigned int) linpt + 1, pt_pct);
  }
}


/** Delete a linearisation point, specified by index */
static enum eMenuLevel clear_linearisation_point(bool full_edit, void *user_data, size_t index)
{
  enum eEditResult result;
  char *name = (char *) user_data;

  AdcReInit(&adcSuperStructure, true, NULL, NULL, NULL);

  result = EDIT_ABORT;
  if (ask(cont_str, false))
  {
    /* Clear the point */
    AdcClearLinPnt(&adcSuperStructure, true, false, index);
    result = EDIT_CHG;
  }

  /* Update linearisation point name in the menus */
  update_linpt_name(name, index);

  return edit_result_to_menu_level(result);
}


/** Edit a specific linearisation point, specified by index */
static enum eMenuLevel edit_linearisation_point(bool full_edit, void *user_data, size_t index)
{
  bool finished;
  uint8_t kbd;
  long upper_limit;
  long lower_limit;
  long rdg;
  enum eEditResult result;
  char *name = (char *) user_data;

  AdcReInit(&adcSuperStructure, true, NULL, NULL, NULL);

  /* Calculate limits */
  upper_limit = adcSuperStructure.adcUserDatabase.fullScale[(adcSuperStructure.adcUserDatabase.scaleType+1)/2];
  lower_limit = 100;

  /* Display the weight flashing */
  set_gross(true);
  freeze_display_and_annunciators(false);
  Dflash_digits(DFLASHMASK_ALL);

  for (finished = false, result = EDIT_ABORT; !quit_setup_menus() && !finished; )
  {
    kbd = get_key();
    switch (kbd)
    {
      case KEY_EDIT_OK:

        /* Editing of linearisation point */
        freeze_display_and_annunciators(true);
        Dflash_digits(DFLASHMASK_NONE);

        rdg = adcSuperStructure.adcDataWorking.grossRdg.rdg;
        Dset_units_ignore_motion(adcSuperStructure.adcUserDatabase.units);
        result = edit_long(&rdg, upper_limit, lower_limit, adcSuperStructure.adcUserDatabase.dpPos, NUM_DIGITS, 0, full_edit);
        if (EDIT_OK == result || EDIT_CHG == result)
        {
          Dputstr(linearisation_in_progress_str);
          (void) AdcAddLinPnt(&adcSuperStructure, true, false, index, rdg);
          wait_and_display_cal_result("LIN");
        }
        else
        {
          finished = true;
        }

        freeze_display_and_annunciators(false);
        Dflash_digits(DFLASHMASK_ALL);
        break;

      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        result = EDIT_MENU_QUIT;
        finished = true;
        break;

      case KEY_EDIT_CANCEL1:
      case KEY_EDIT_CANCEL2:
        result = EDIT_ABORT;
        finished = true;
        break;

      /* These keys quit the calibration */
      case KEY_EDIT_DOWN:
      case KEY_EDIT_LEFT:
      case KEY_EDIT_RIGHT:
      case KEY_EDIT_UP:
        result = EDIT_OK;
        finished = true;
        break;
    }
  }

  freeze_display_and_annunciators(true);
  Dflash_digits(DFLASHMASK_NONE);

  /* Update linearisation point name in the menus */
  update_linpt_name(name, index);

  return edit_result_to_menu_level(result);
}


/** Convert linearisation point to string */
static void to_string_linearisation_point(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  AdcRdg rdg;

  memset(&rdg, 0, sizeof(rdg));
  rdg.rdg = adcSuperStructure.adcUserDatabase.linCor[index] * 10 / 16; /* *10, +1 for extra decimal place */
  rdg.info.unitdp.bit.dp = adcSuperStructure.adcUserDatabase.dpPos + 1;
  rdg.info.unitdp.bit.units = adcSuperStructure.adcUserDatabase.units;
  to_string_weight(&rdg, true, buffer, buflen);
}


/** Construct menu items for the linearisation menu */
static void construct_linearisation_menu(struct s_menu_item **items_list, struct s_menu_item *items, t_sub_menu_linearisation_names *names, bool clear_not_edit)
{
  size_t linpt;

  /* Create list of points */
  for (linpt = 0; linpt < ADC_MAX_NUM_LIN_PTS; linpt++)
  {
    update_linpt_name((*names)[linpt], linpt);

    memset(&items[linpt], 0, sizeof(items[linpt]));
    items[linpt].title = (*names)[linpt];
    items[linpt].edit_fn = (clear_not_edit) ? clear_linearisation_point : edit_linearisation_point;
    items[linpt].to_string_fn = (clear_not_edit) ? NULL : to_string_linearisation_point;
    items[linpt].sub_menu = NULL;
    items[linpt].menu_enter_exit = NULL;
    items[linpt].user_data_from_menu_above = false;
    items[linpt].user_data = (*names)[linpt];

    items_list[linpt] = &items[linpt];
  }

  /* The remainder don't show */
  for ( ; linpt < ADC_MAX_NUM_LIN_PTS + 1; linpt++)
  {
    items_list[linpt] = NULL;
  }
}


static enum eMenuLevel do_calibration_defaults(bool full_edit, void *user_data, size_t index)
{
  if (ask(cont_str, false))
  {
    AdcDoDefaultCal(&adcSuperStructure, true, false);
  }
  return mlNum;
}


/** Run checks before the calibration menu runs */
bool cal_menu_enter_exit(bool entry_not_exit, bool full_edit, void *user_data, bool edit_not_tostring)
{
  bool result = true;

  if (edit_not_tostring && entry_not_exit)
  {
    if (!full_edit)
    {
      access_denied();
      result = false;
    }
  }
  return result;
}


/** Names for each of the linearisation points */
static t_sub_menu_linearisation_names sub_menu_linearisation_names;


/** Linearisation sub menu items, filled out by @ref linearisation_menu_enter_exit() */
static struct s_menu_item sub_menu_linearisation[ADC_MAX_NUM_LIN_PTS];


/** Setpoint sub menu list, constructed by @ref operator_target_menu_enter_exit() */
static struct s_menu_item *sub_menu_linearisation_list[ADC_MAX_NUM_LIN_PTS + 1];


/** Construct the linearisation submenu */
static bool linearisation_menu_enter_exit(bool entry_not_exit, bool full_edit, void *user_data, bool edit_not_tostring)
{
  bool clear_not_edit = (bool) user_data;

  if (entry_not_exit)
  {
    /* Don't print the clear linearisation menu */
    if (!edit_not_tostring && clear_not_edit)
      return false;

    construct_linearisation_menu(sub_menu_linearisation_list, sub_menu_linearisation, &sub_menu_linearisation_names, clear_not_edit);
  }
  return true;
}


static const struct s_menu_item sub_menu_cal_normal_zero = { .title = "ZERO",    .sub_menu = NULL, .edit_fn = do_normal_zero_cal,      .to_string_fn = to_string_normal_zero_cal };
static const struct s_menu_item sub_menu_cal_normal_span = { .title = "SPAN",    .sub_menu = NULL, .edit_fn = do_normal_span_cal,      .to_string_fn = to_string_normal_span_cal };
static const struct s_menu_item sub_menu_cal_editlin =     { .title = "ED.LIN",  .sub_menu = sub_menu_linearisation_list, .menu_enter_exit = linearisation_menu_enter_exit, .user_data = (void *) false };
static const struct s_menu_item sub_menu_cal_clearlin =    { .title = "CLR.LIN", .sub_menu = sub_menu_linearisation_list, .menu_enter_exit = linearisation_menu_enter_exit, .user_data = (void *) true };
static const struct s_menu_item sub_menu_cal_direct_zero = { .title = "DIR.ZER", .sub_menu = NULL, .edit_fn = do_direct_zero_cal,      .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_cal_direct_span = { .title = "DIR.SPN", .sub_menu = NULL, .edit_fn = do_direct_span_cal,      .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_cal_defaults =    { .title = "FAC.CAL", .sub_menu = NULL, .edit_fn = do_calibration_defaults, .to_string_fn = to_string_no_setting };


static const struct s_menu_item *const sub_menu_cal[] =
{
  &sub_menu_cal_normal_zero,
  &sub_menu_cal_normal_span,
  &sub_menu_cal_editlin,
  &sub_menu_cal_clearlin,
  &sub_menu_cal_direct_zero,
  &sub_menu_cal_direct_span,
  &sub_menu_cal_defaults,
  NULL
};


/** Calibration menu item */
const struct s_menu_item menu_cal =
{
  .title = "CAL",
  .sub_menu = sub_menu_cal,
  .menu_enter_exit = cal_menu_enter_exit
};
