/*
 * iface_display_total.c
 *
 *  Created on: Apr 16, 2018
 *      Author: ian
 */
#include <stdbool.h>
#include <string.h>
#include <rinweigh/adcdatatypes.h>
#include "display.h"
#include "system.h"
#include "iface.h"
#include "adcthread.h"
#include "unitswitch.h"
#include "app_db_runtime_vars.h"

#include "iface_display_total.h"


#define PROMPT_DELAY_MS 500
#define WEIGHT_DELAY_MS 1500


/** Display a prompt, delay */
static void prompt_delay(const char *prompt, unsigned delay_prompt_ms)
{
  Dputstr(prompt);
  Dset_units_ignore_motion(ADC_UNITS_NONE);
  Dset_annunciators(ANNUN_NONE);
  delay(delay_prompt_ms);
}


/** Display a prompt, delay, display a weight, delay */
static void prompt_delay_weight_delay(const char *prompt, unsigned delay_prompt_ms, const AdcRdg *rdg, unsigned delay_rdg_ms, size_t min_digits)
{
  prompt_delay(prompt, delay_prompt_ms);

  /* Show the weight, units, annunciators */
  Dshow_number(rdg->rdg, rdg->info.unitdp.bit.dp, min_digits);
  Dset_units_ignore_motion(rdg->info.unitdp.bit.units);
  Dset_annunciators(get_annunciators(&rdg->info, 0));
  delay(delay_rdg_ms);
}


/** Run the display sequence for a weight >= 1000000, which is split into two halves */
static void run_total_display_hilo_weight_plus(const char *prompt_hi, const char *prompt_lo, unsigned delay_prompt_ms, const AdcRdg *rdg, unsigned delay_rdg_ms, size_t min_digits)
{
  AdcRdg hirdg;
  AdcRdg lordg;

  hirdg = *rdg;
  hirdg.rdg /= 1000000;
  hirdg.info.unitdp.bit.dp = 0;
  prompt_delay_weight_delay(prompt_hi, delay_prompt_ms, &hirdg, delay_rdg_ms, min_digits);

  lordg = *rdg;
  lordg.rdg %= 1000000;
  prompt_delay_weight_delay(prompt_lo, delay_prompt_ms, &lordg, delay_rdg_ms, NUM_DIGITS);
}


/** Run the display sequence for a weight <= -100000, which is split into two halves */
static void run_total_display_hilo_weight_minus(const char *prompt_hi, const char *prompt_lo, unsigned delay_prompt_ms, const AdcRdg *rdg, unsigned delay_rdg_ms, size_t min_digits)
{
  AdcRdg hirdg;
  AdcRdg lordg;

  if (rdg->rdg >= -1000000 && rdg->rdg <= -100000)
  {
    prompt_delay(prompt_hi, delay_prompt_ms);
    prompt_delay("     -", delay_rdg_ms);

    lordg = *rdg;
    lordg.rdg = -lordg.rdg;
    prompt_delay_weight_delay(prompt_lo, delay_prompt_ms, &lordg, delay_rdg_ms, NUM_DIGITS);
  }
  else
  {
    hirdg = *rdg;
    hirdg.rdg /= 1000000;
    hirdg.info.unitdp.bit.dp = 0;
    prompt_delay_weight_delay(prompt_hi, delay_prompt_ms, &hirdg, delay_rdg_ms, min_digits);

    lordg = *rdg;
    lordg.rdg = -lordg.rdg;
    lordg.rdg %= 1000000;
    prompt_delay_weight_delay(prompt_lo, delay_prompt_ms, &lordg, delay_rdg_ms, NUM_DIGITS);
  }
}


/** Run the display sequence for a weight, taking into consideration if it will fit on the display */
static void run_total_display_weight(const char *prompt_single, const char *prompt_hi, const char *prompt_lo, unsigned prompt_delay_ms,
                                     const AdcRdg *rdg, unsigned rdg_delay_ms, size_t min_digits)
{
  if (rdg->rdg >= 1000000)
  {
    run_total_display_hilo_weight_plus(prompt_hi, prompt_lo, prompt_delay_ms, rdg, rdg_delay_ms, min_digits);
  }
  else if (rdg->rdg <= -100000)
  {
    run_total_display_hilo_weight_minus(prompt_hi, prompt_lo, prompt_delay_ms, rdg, rdg_delay_ms, min_digits);
  }
  else
  {
    prompt_delay_weight_delay(prompt_single, prompt_delay_ms, rdg, rdg_delay_ms, min_digits);
  }
}


/** Run the display sequence for a NTEP total */
static void run_total_display_ntep(const struct s_display_total *dt)
{
  run_total_display_weight("TOTAL.G", "TOT.G.HI", "TOT.G.LO", PROMPT_DELAY_MS, &dt->ntep_gross_total, WEIGHT_DELAY_MS, dt->ntep_min_digits);
  run_total_display_weight("TOTAL.N", "TOT.N.HI", "TOT.N.LO", PROMPT_DELAY_MS, &dt->ntep_net_total, WEIGHT_DELAY_MS, dt->ntep_min_digits);
}


/** Run the display sequence for a non-NTEP total */
static void run_total_display_non_ntep(const struct s_display_total *dt)
{
  run_total_display_weight("TOTAL", "TOT.HI", "TOT.LO", PROMPT_DELAY_MS, &dt->non_ntep_total, WEIGHT_DELAY_MS, 0);
}


/** Run the total display sequence */
void run_total_display(struct s_display_total *dt)
{
  AdcRdg rdg;

  if (dt->show_total)
  {
    freeze_display_and_annunciators(true);
    Dclear_flashing();

    memset(&rdg, 0, sizeof(rdg));
    rdg.rdg = dt->item_count;
    prompt_delay_weight_delay("COUNT", PROMPT_DELAY_MS, &rdg, WEIGHT_DELAY_MS, 0);

    if (ADC_APPROVAL_MODE_NTEP == dt->approval_mode)
    {
      run_total_display_ntep(dt);
    }
    else
    {
      run_total_display_non_ntep(dt);
    }
    freeze_display_and_annunciators(false);
    dt->show_total = false;
  }
}


/** Initiate the display of the statistics totals */
void initiate_display_total(struct s_display_total *dt)
{
  enum eAdcUnitSwitch usw_to;
  uint8_t dp;
  enum eAdcUnits units;

  dt->approval_mode = adcSuperStructure.adcUserDatabase.approvalMode;
  usw_to = unitsw_displaymode_to_unitswitch(adcSuperStructure.adcDataWorking.dispMode);

  dp = unitsw_get_decimal_point(usw_to);
  units = unitsw_get_units(usw_to);
  dt->ntep_min_digits = unitsw_get_resolution_length(usw_to);

  dt->item_count = statistics_get_num_items();

  memset(&dt->ntep_gross_total, 0, sizeof(dt->ntep_gross_total));
  dt->ntep_gross_total.rdg = unitsw_convert_final(statistics_get_gross_total(), ADC_UNIT_SWITCH_PRIMARY, usw_to, false, false);
  dt->ntep_gross_total.info.unitdp.bit.dp = dp;
  dt->ntep_gross_total.info.unitdp.bit.units = units;
  dt->ntep_gross_total.info.status.bit.grossOrNet = 1;
  dt->ntep_gross_total.info.status.bit.gross = 1;
  memset(&dt->ntep_net_total, 0, sizeof(dt->ntep_net_total));
  dt->ntep_net_total.rdg = unitsw_convert_final(statistics_get_net_total(), ADC_UNIT_SWITCH_PRIMARY, usw_to, false, false);
  dt->ntep_net_total.info.unitdp.bit.dp = dp;
  dt->ntep_net_total.info.unitdp.bit.units = units;
  dt->ntep_net_total.info.status.bit.grossOrNet = 1;
  dt->ntep_net_total.info.status.bit.gross = 0;
  memset(&dt->non_ntep_total, 0, sizeof(dt->non_ntep_total));
  dt->non_ntep_total.rdg = unitsw_convert_final(statistics_get_total(), ADC_UNIT_SWITCH_PRIMARY, usw_to, false, false);
  dt->non_ntep_total.info.unitdp.bit.dp = dp;
  dt->non_ntep_total.info.unitdp.bit.units = units;

  dt->show_total = true;
}
