/*
 * iface_ethnet.c
 *
 * Since the introduction of libsetnetconfig, everything here is in network endian
 *
 *  Created on: 07/06/2010
 *      Author: ian
 */
#include <arpa/inet.h>
#include <sys/socket.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <setnetconfig.h>
#include "editor.h"
#include "edit_ip.h"
#include "ipinfo.h"
#include "we_debug.h"
#include "display.h"
#include "iface_sd.h"
#include "commonstrings.h"
#include "edit_string.h"
#include "app_db_factory.h"
#include "version.h"
#include "dbprint.h"
#include "we_config.h"
#include "iface.h"

#include "iface_ethnet.h"


/** Index of items in the Ethernet sub menu. Must match order in @ref sub_menu_ethernet */
enum e_sub_menu_ethernet_index
{
  mei_dhcp = 0,
  mei_ip,
  mei_nm,
  mei_gw,
  mei_ns0,
  mei_ns1,
  mei_ns2,
  mei_hostname,
  /* mei_domain, */
  mei_search,
  mei_mac,
  mei_ethdef
};


/** Info for running the Ethernet menus */
struct s_eth_menu_info
{
  struct s_snc_net_config net_config;         /*!< Current network configuration */
  struct s_snc_net_config net_current;        /*!< Current running network information */
  bool dhcp_allow_edit;                       /*!< Allow edit of ip/netmask/gateway/dns, because we are not running DHCP */
  struct s_snc_net_config *ipconfig;          /*!< Pointer to the configuration we should display/edit for ip/netmask/gateway/dns depending on DHCP setting */
  bool read_only;                             /*!< All Ethernet menu items cannot be edited, overrides allow_edit */
};


/** Called when the setup menus are entered. Retrieves IP configuration from the system */
static void ethnet_start_setup(struct s_eth_menu_info *eth_menu_info)
{
  /* Set defaults, in case something fails */
  net_ip_set_defaults(&eth_menu_info->net_config);
  net_ip_set_defaults(&eth_menu_info->net_current);

  /* Get system ip configuration */
  if (!net_ip_get_config(true, &eth_menu_info->net_config))
    rdebug_printf(LOG_ERR, "ethnet_start_setup: net_ip_get_config(config) failed\n");

  /* If all the rc.conf ip values are 0.0.0.0, then use defaults */
  if (htonl(0) == eth_menu_info->net_config.ip_address.s_addr &&
      htonl(0) == eth_menu_info->net_config.ip_netmask.s_addr &&
      htonl(0) == eth_menu_info->net_config.ip_broadcast.s_addr &&
      htonl(0) == eth_menu_info->net_config.ip_gateway.s_addr &&
      htonl(0) == eth_menu_info->net_config.ip_dns[0].s_addr &&
      htonl(0) == eth_menu_info->net_config.ip_dns[1].s_addr &&
      htonl(0) == eth_menu_info->net_config.ip_dns[2].s_addr)
  {
    WE_IPCONF_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "ethnet_start_setup: all network config values were 0.0.0.0, setting to defaults\n");
    eth_menu_info->net_config.ip_address.s_addr = DEFAULT_ETH_IPADDRESS;
    eth_menu_info->net_config.ip_netmask.s_addr = DEFAULT_ETH_NETMASK;
    eth_menu_info->net_config.ip_broadcast.s_addr = DEFAULT_ETH_BROADCAST;
    eth_menu_info->net_config.ip_gateway.s_addr = DEFAULT_ETH_GATEWAY;
    eth_menu_info->net_config.ip_dns[0].s_addr = DEFAULT_ETH_DNS1;
    eth_menu_info->net_config.ip_dns[1].s_addr = DEFAULT_ETH_DNS2;
    eth_menu_info->net_config.ip_dns[2].s_addr = DEFAULT_ETH_DNS3;
  }

  /* Get the current (running) IP configuration from the system */
  if (!net_ip_get_config(false, &eth_menu_info->net_current))
    rdebug_printf(LOG_ERR, "ethnet_start_setup: net_ip_get_config(current) failed\n");

#if RDEBUG_ENABLED(WE_IPCONF_DEBUG_LVL, RDEBUG_LVL_FULL)
  /* Print out debugging information */
  char *net_config_str = NULL;
  char *net_current_str = NULL;

  (void) snc_net_config_to_allocstr(&eth_menu_info->net_config, true, &net_config_str);
  (void) snc_net_config_to_allocstr(&eth_menu_info->net_current, true, &net_current_str);

  rdebug_printf(LOG_INFO, "ethnet_start_setup: network config: %s\n",
                net_config_str);
  rdebug_printf(LOG_INFO, "ethnet_start_setup: network current: %s\n",
                net_current_str);
  free(net_current_str);
  free(net_config_str);
#endif /* #if WE_DEBUG(WE_IPCONF_DEBUG_LVL, WE_DEBUG_FULL) */
}


/** Callback from net_ip_save_config() to display errors when network configuration settings are invalid */
static void ethnet_save_setup_ui_callback(enum e_net_ip_save_result save_result)
{
  /* Don't run error_strings() if we are bailing out */
  if (quit_setup_menus())
    return;

  switch (save_result)
  {
    default:
    case nesr_success:
      break;
    case nesr_failure:
      error_strings("ETH.SAV", err_str);
      break;
    case nesr_about_to_save:
      Dputstr("ETH.SAV");
      break;
    case nesr_bad_hostname:
      error_strings(menu_ethernet.sub_menu[mei_hostname]->title, err_str);
      break;
    case nesr_bad_searchdomain:
      error_strings(menu_ethernet.sub_menu[mei_search]->title, err_str);
      break;
    case nesr_bad_ipaddress:
      error_strings(menu_ethernet.sub_menu[mei_ip]->title, err_str);
      break;
    case nesr_bad_netmask:
      error_strings(menu_ethernet.sub_menu[mei_nm]->title, err_str);
      break;
    case nesr_bad_gateway:
      error_strings(menu_ethernet.sub_menu[mei_gw]->title, err_str);
      break;
    case nesr_bad_broadcast:
      error_strings("BCAST", err_str);
      break;
    case nesr_bad_dns0:
      error_strings(menu_ethernet.sub_menu[mei_ns0]->title, err_str);
      break;
    case nesr_bad_dns1:
      error_strings(menu_ethernet.sub_menu[mei_ns1]->title, err_str);
      break;
    case nesr_bad_dns2:
      error_strings(menu_ethernet.sub_menu[mei_ns2]->title, err_str);
      break;
  }
}


/** Called when the Ethernet setup menu is exited. Saves any changes to the system. Returns true if the menus can exit, false otherwise. */
static bool ethnet_save_setup(struct s_snc_net_config *net_config)
{
  enum e_net_ip_save_result save_result;

  save_result = net_ip_save_config(net_config, ethnet_save_setup_ui_callback);

  /* Any of the verification failed results prevent the menus from exiting */
  return (nesr_success == save_result || nesr_failure == save_result);
}


/** Edit the DHCP setting */
static enum eMenuLevel edit_use_dhcp(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  uint8_t new_usedhcp;

  new_usedhcp = (eth_menu_info->net_config.use_dhcp) ? 1 : 0;
  edit_register_data(&new_usedhcp, 0, 1, off_on_setting, !eth_menu_info->read_only);

  if (eth_menu_info->net_config.use_dhcp != (bool) new_usedhcp)
  {
    /* On change, re-configure eth_menu_info */
    eth_menu_info->net_config.use_dhcp = (bool) new_usedhcp;

    eth_menu_info->dhcp_allow_edit = !eth_menu_info->net_config.use_dhcp;
    eth_menu_info->ipconfig = (eth_menu_info->net_config.use_dhcp) ? &eth_menu_info->net_current : &eth_menu_info->net_config;

    /* If we are turning DHCP off,
     * then lets adopt the current (dhcp supplied) settings as our static settings,
     * as a good starting point for the user to edit
     */
    if (!eth_menu_info->net_config.use_dhcp)
    {
      eth_menu_info->net_config = eth_menu_info->net_current;
      eth_menu_info->net_config.use_dhcp = false;
    }
  }

  return mlNum;
}


/** Print the use dhcp setting */
static void to_string_use_dhcp(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  snprintf(buffer, buflen, "%s", off_on_setting[eth_menu_info->net_config.use_dhcp]);
}


/** Edit/view an IPv4 address */
static enum eMenuLevel edit_ip(struct in_addr *ip_address, bool allow_edit)
{
  enum eEditResult er;

  if (allow_edit)
  {
    er = edit_ipv4_address(ip_address, eie_network);
  }
  else
  {
    er = view_ipv4_address(ip_address, eie_network);
  }
  return edit_result_to_menu_level(er);
}


/** Convert an IP address to a string. ip is host endian. */
static void to_string_ip(const struct in_addr *ip, char *buffer, size_t buflen)
{
  char ip_str[INET_ADDRSTRLEN];

  (void) net_ip_inet_ntop_host_afinet(ip, ip_str, INET_ADDRSTRLEN);
  snprintf(buffer, buflen, "%s", ip_str);
}


/** Edit the IP address */
static enum eMenuLevel edit_ip_address(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  return edit_ip(&eth_menu_info->ipconfig->ip_address, eth_menu_info->dhcp_allow_edit && !eth_menu_info->read_only);
}


/** Convert the IP address to string */
static void to_string_ip_address(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  to_string_ip(&eth_menu_info->ipconfig->ip_address, buffer, buflen);
}


/** Edit the netmask address */
static enum eMenuLevel edit_netmask_address(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  return edit_ip(&eth_menu_info->ipconfig->ip_netmask, eth_menu_info->dhcp_allow_edit && !eth_menu_info->read_only);
}


/** Convert the netmask address to string */
static void to_string_netmask_address(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  to_string_ip(&eth_menu_info->ipconfig->ip_netmask, buffer, buflen);
}


/** Edit the default gateway address */
static enum eMenuLevel edit_gateway_address(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  return edit_ip(&eth_menu_info->ipconfig->ip_gateway, eth_menu_info->dhcp_allow_edit && !eth_menu_info->read_only);
}


/** Convert the default gateway address to string */
static void to_string_gateway_address(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  to_string_ip(&eth_menu_info->ipconfig->ip_gateway, buffer, buflen);
}


/** Edit the DNS server addresses */
static enum eMenuLevel edit_dnsserver_address(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  struct in_addr *dns;

  switch (index)
  {
    case mei_ns0:
      dns = &eth_menu_info->ipconfig->ip_dns[0];
      break;
    case mei_ns1:
      dns = &eth_menu_info->ipconfig->ip_dns[1];
      break;
    case mei_ns2:
      dns = &eth_menu_info->ipconfig->ip_dns[2];
      break;
    default:
      return mlNum;
      break;
  }

  return edit_ip(dns, eth_menu_info->dhcp_allow_edit && !eth_menu_info->read_only);
}


/** Convert the DNS server to string */
static void to_string_dnsserver_address(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  struct in_addr *dns;

  switch (index)
  {
    case mei_ns0:
      dns = &eth_menu_info->ipconfig->ip_dns[0];
      break;
    case mei_ns1:
      dns = &eth_menu_info->ipconfig->ip_dns[1];
      break;
    case mei_ns2:
      dns = &eth_menu_info->ipconfig->ip_dns[2];
      break;
    default:
      snprintf(buffer, buflen, "%s", "");
      return;
      break;
  }

  to_string_ip(dns, buffer, buflen);
}


/** Edit the hostname setting */
static enum eMenuLevel edit_hostname(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  enum eEditResult er;

  if (!eth_menu_info->read_only)
  {
    er = edit_printable_string(eth_menu_info->net_config.hostname, SNC_HOST_NAME_MAX, charset_hostname);
  }
  else
  {
    er = view_printable_string(eth_menu_info->net_config.hostname);
  }
  return edit_result_to_menu_level(er);
}


/** Convert the hostname to string */
static void to_string_hostname(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  snprintf(buffer, buflen, "%s", eth_menu_info->net_config.hostname);
}


#if 0
/** Edit the domain name setting */
static enum eMenuLevel edit_domain(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  enum eEditResult er;

  if (!eth_menu_info->read_only)
  {
    er = edit_printable_string(eth_menu_info->net_config.domain, SNC_DOMAIN_NAME_MAX, charset_domainname);
  }
  else
  {
    er = view_printable_string(eth_menu_info->net_config.domain);
  }
  return edit_result_to_menu_level(er);
}


/** Convert the domain name to string */
static void to_string_domain(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  snprintf(buffer, buflen, "%s", eth_menu_info->net_config.domain);
}
#endif


/** Edit the search domain name setting */
static enum eMenuLevel edit_searchdomain(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  enum eEditResult er;

  if (eth_menu_info->dhcp_allow_edit && !eth_menu_info->read_only)
  {
    er = edit_printable_string(eth_menu_info->ipconfig->search_domain, SNC_DOMAIN_NAME_MAX, charset_domainname);
  }
  else
  {
    er = view_printable_string(eth_menu_info->ipconfig->search_domain);
  }
  return edit_result_to_menu_level(er);
}


/** Convert the search domain name to string */
static void to_string_searchdomain(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  snprintf(buffer, buflen, "%s", eth_menu_info->ipconfig->search_domain);
}


/** View a MAC address */
static enum eMenuLevel view_macaddress(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  char str[MAC_ADDRESS_STR_LENGTH];

  net_ip_macaddress_to_string(&eth_menu_info->net_current.mac_address, '.', str);
  return edit_result_to_menu_level(view_printable_string(str));
}


/** Print the MAC address */
static void to_string_macaddress(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  char str[MAC_ADDRESS_STR_LENGTH];

  net_ip_macaddress_to_string(&eth_menu_info->net_current.mac_address, '.', str);
  snprintf(buffer, buflen, "%s", str);
}


/** Prompt, then set the Ethernet configuration to defaults */
static enum eMenuLevel do_ethernet_defaults(bool full_edit, void *user_data, size_t index)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  if (ask(cont_str, false))
  {
    net_ip_set_defaults(&eth_menu_info->net_config);
  }
  return mlNum;
}


/** Run checks before the ethernet menu runs */
bool ethernet_menu_enter_exit(bool entry_not_exit, bool full_edit, void *user_data, bool edit_not_tostring)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;
  bool result;

  if (entry_not_exit)
  {
    /* Get Ethernet setup */
    ethnet_start_setup(eth_menu_info);

    /*
     * If DHCP is enabled, then editing is read-only, and we display the current
     * IP configuration retrieved from DHCP
     */
    eth_menu_info->dhcp_allow_edit = !eth_menu_info->net_config.use_dhcp;
    eth_menu_info->ipconfig = (eth_menu_info->net_config.use_dhcp) ? &eth_menu_info->net_current : &eth_menu_info->net_config;
    eth_menu_info->read_only = false;
    result = true;
  }
  else
  {
    /* If something has changed, save setting. We don't exit the menu level unless all settings were validated */
    result = ethnet_save_setup(&eth_menu_info->net_config);
  }
  return result;
}


/** Run checks before the operator ip info menu runs */
bool oper_ip_info_menu_enter_exit(bool entry_not_exit, bool full_edit, void *user_data, bool edit_not_tostring)
{
  struct s_eth_menu_info *eth_menu_info = (struct s_eth_menu_info *) user_data;

  if (entry_not_exit)
  {
    /* Get Ethernet setup */
    ethnet_start_setup(eth_menu_info);

    /* For the operator menu, nothing can be edited */
    eth_menu_info->dhcp_allow_edit = false;
    eth_menu_info->ipconfig = &eth_menu_info->net_current;
    eth_menu_info->read_only = true;
  }
  return true;
}


/*
 * Ethernet menu
 */


static const struct s_menu_item sub_menu_ethernet_dhcp =     { .title = "DHCP",    .sub_menu = NULL, .edit_fn = edit_use_dhcp,          .to_string_fn = to_string_use_dhcp,          .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_ip =       { .title = "IP",      .sub_menu = NULL, .edit_fn = edit_ip_address,        .to_string_fn = to_string_ip_address,        .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_netmask =  { .title = "MASK",    .sub_menu = NULL, .edit_fn = edit_netmask_address,   .to_string_fn = to_string_netmask_address,   .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_gateway =  { .title = "GATE.W",  .sub_menu = NULL, .edit_fn = edit_gateway_address,   .to_string_fn = to_string_gateway_address,   .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_dns0 =     { .title = "DNS.1",   .sub_menu = NULL, .edit_fn = edit_dnsserver_address, .to_string_fn = to_string_dnsserver_address, .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_dns1 =     { .title = "DNS.2",   .sub_menu = NULL, .edit_fn = edit_dnsserver_address, .to_string_fn = to_string_dnsserver_address, .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_dns2 =     { .title = "DNS.3",   .sub_menu = NULL, .edit_fn = edit_dnsserver_address, .to_string_fn = to_string_dnsserver_address, .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_hostname = { .title = "HST.NAM", .sub_menu = NULL, .edit_fn = edit_hostname,          .to_string_fn = to_string_hostname,          .user_data_from_menu_above = true };
/* static const struct s_menu_item sub_menu_ethernet_domain =   { .title = "DOMAIN",  .sub_menu = NULL, .edit_fn = edit_domain,            .to_string_fn = to_string_domain,            .user_data_from_menu_above = true }; */
static const struct s_menu_item sub_menu_ethernet_search =   { .title = "SEARCH",  .sub_menu = NULL, .edit_fn = edit_searchdomain,      .to_string_fn = to_string_searchdomain,      .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_mac =      { .title = "MAC",     .sub_menu = NULL, .edit_fn = view_macaddress,        .to_string_fn = to_string_macaddress,        .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_ethernet_defaults = { .title = "ETH.DEF", .sub_menu = NULL, .edit_fn = do_ethernet_defaults,   .to_string_fn = to_string_no_setting,        .user_data_from_menu_above = true };


static const struct s_menu_item *const sub_menu_ethernet[] =
{
  &sub_menu_ethernet_dhcp,
  &sub_menu_ethernet_ip,
  &sub_menu_ethernet_netmask,
  &sub_menu_ethernet_gateway,
  &sub_menu_ethernet_dns0,
  &sub_menu_ethernet_dns1,
  &sub_menu_ethernet_dns2,
  &sub_menu_ethernet_hostname,
  &sub_menu_ethernet_search,
  &sub_menu_ethernet_mac,
  &sub_menu_ethernet_defaults,
  NULL
};


static struct s_eth_menu_info eth_menu_info;


/** Ethernet menu item */
const struct s_menu_item menu_ethernet =
{
  .title = "ETH.NET",
  .sub_menu = sub_menu_ethernet,
  .menu_enter_exit = ethernet_menu_enter_exit,
  .user_data = &eth_menu_info
};


/*
 * Operator IP info menu
 */


static const struct s_menu_item *const sub_menu_oper_ip_list[] =
{
    &sub_menu_ethernet_dhcp,
    &sub_menu_ethernet_ip,
    &sub_menu_ethernet_netmask,
    &sub_menu_ethernet_gateway,
    &sub_menu_ethernet_dns0,
    &sub_menu_ethernet_dns1,
    &sub_menu_ethernet_dns2,
    &sub_menu_ethernet_hostname,
    &sub_menu_ethernet_search,
    &sub_menu_ethernet_mac,
    NULL
};


/** Operator IP info menu item */
const struct s_menu_item sub_menu_oper_ip =
{
  .title = "IP.INFO",
  .sub_menu = sub_menu_oper_ip_list,
  .menu_enter_exit = oper_ip_info_menu_enter_exit,
  .user_data = &eth_menu_info
};
