/*
 * iface_factory.c
 *
 *  Created on: 16/12/2010
 *      Author: ian
 */
#include <wait.h>
#include <rinexec/el_runcmd.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "database.h"
#include "display.h"
#include "system.h"
#include "editor.h"
#include "iface_sd.h"
#include "iface.h"
#include "meas.h"
#include "dbprint.h"
#include "setpoint.h"
#include "we_config.h"
#include "app_db_factory.h"
#include "app_db_runtime.h"
#include "app_db_user.h"
#include "adcthread.h"
#include "commonstrings.h"
#include "comms_task.h"
#include "comms_hdlr_am.h"

#include "iface_factory.h"


/** Set indicator to factory defaults */
static enum eMenuLevel do_factory_default(bool full_edit, void *user_data, size_t index)
{
  if (!full_edit)
  {
    access_denied();
    return mlNum;
  }
  if (ask(cont_str, false))
  {
    /* Set user database to defaults, except calibration */
    app_default_user_database(false);

    /* Set runtime database to defaults, except ADC runtime variables */
    app_default_runtime_database(false);
    comms_reinit_setpoints();

    /* re-init buzzer to keep Dane happy */
    setup_kbd_buzz();
  }
  return mlNum;
}


/** Print out the current configuration settings to the configured printer */
static enum eMenuLevel print_config(bool full_edit, void *user_data, size_t index)
{
  if (!ask(cont_str, false))
  {
    return mlNum;
  }

  print_current_config(full_edit);
  return mlNum;
}


#if 0
/** Start inetd so we can telnet to the indicator */
static enum eMenuLevel factory_service_start_inetd(bool full_edit, void *user_data, size_t index)
{
  int ret_code;
  const char *const inetd_args[] = { "/etc/rc.d/init.d/inetd", "start", NULL };

  /* Start inetd and check the return code */
  if (el_run_command(inetd_args[0], inetd_args, &ret_code, NULL, true, false, true) != elr_success ||
      !WIFEXITED(ret_code) || 0 != WEXITSTATUS(ret_code))
  {
    goto fail;
  }

  /* Started successfully */
  Dputstr(done_str);
  delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
  return mlNum;

  /* Something went wrong */
fail:
  error_strings("INETD", "FAIL");
  return mlNum;
}


/** Start a serial console on specified port */
static enum eMenuLevel factory_service_start_serial_console(bool full_edit, void *user_data, size_t index)
{
  intptr_t port = (intptr_t) user_data;
  const char *getty_args[6];
  struct t_io_dev *dev_serial;

  /* Setup arguments to getty */
  getty_args[0] = "/sbin/getty";
  getty_args[1] = "-L";
  getty_args[2] = (1 == port) ? we_configuration.serial1_device : we_configuration.serial2_device;
  getty_args[3] = "115200";
  getty_args[4] = "screen";
  getty_args[5] = NULL;

  /* Which serial port are we getty'ing on */
  dev_serial = (1 == port) ? &dev_ser1_local : &dev_ser_front_rl;

  /* Clean up our access to port */
  comms_destroy_task(dev_serial);
  rdestroy_io(dev_serial);

  /* Start getty and let it run */
  if (el_run_command(getty_args[0], getty_args, NULL, NULL, false, false, true) != elr_success)
  {
    goto fail;
  }

  /* Started successfully */
  Dputstr(done_str);
  delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
  return mlNum;

  /* Something went wrong */
fail:
  error_strings("CON.SER", "FAIL");
  return mlNum;
}


/** Quit the WE2111 application */
static enum eMenuLevel factory_service_quit(bool full_edit, void *user_data, size_t index)
{
  if (ask(cont_str, false))
  {
    application_quit_force();
  }
  return mlNum;
}


/** Run checks before the factory service menu runs */
static bool factory_service_menu_enter_exit(bool entry_not_exit, bool full_edit, void *user_data, bool edit_not_tostring)
{
  long l;

  /* Hide the submenu when printing the database */
  if (!edit_not_tostring)
    return false;

  if (entry_not_exit)
  {
    /* Ask user for password to enter factory service menu */
    prompt("ENTER", "PASS");
    l = 0;
    if (edit_long(&l, 999999L, 0L, 0, NUM_DIGITS, 0, ALLOW_SAFE) == EDIT_ABORT)
    {
      return false;
    }
    if (WECFG_FACTORY_ACCESS_CODE != l)
    {
      access_denied();
      return false;
    }
  }
  return true;
}


static const struct s_menu_item sub_menu_factory_service_inetd =       { .title = "INETD",  .sub_menu = NULL, .edit_fn = factory_service_start_inetd };
static const struct s_menu_item sub_menu_factory_service_serconsole1 = { .title = "CON.S1", .sub_menu = NULL, .edit_fn = factory_service_start_serial_console, .user_data = (void *) 1 };
static const struct s_menu_item sub_menu_factory_service_serconsole2 = { .title = "CON.S2", .sub_menu = NULL, .edit_fn = factory_service_start_serial_console, .user_data = (void *) 2 };
static const struct s_menu_item sub_menu_factory_service_quit =        { .title = "QUIT",   .sub_menu = NULL, .edit_fn = factory_service_quit };


/** Factory service sub menu item list */
static const struct s_menu_item *const sub_menu_factory_service_list[] =
{
  &sub_menu_factory_service_inetd,
  &sub_menu_factory_service_serconsole1,
  &sub_menu_factory_service_serconsole2,
  &sub_menu_factory_service_quit,
  NULL
};
#endif


static const struct s_menu_item sub_menu_factory_default =     { .title = "DEFLT",   .sub_menu = NULL, .edit_fn = do_factory_default, .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_factory_printconfig = { .title = "PRN.CFG", .sub_menu = NULL, .edit_fn = print_config,       .to_string_fn = to_string_no_setting };
/* static const struct s_menu_item sub_menu_factory_service =     { .title = "FAC.SVC", .sub_menu = sub_menu_factory_service_list,       .menu_enter_exit = factory_service_menu_enter_exit }; */


const struct s_menu_item *const menu_factory_list[] =
{
  &sub_menu_factory_default,
  &sub_menu_factory_printconfig,
/* &sub_menu_factory_service, */
  NULL
};


/** Factory menu item */
const struct s_menu_item menu_factory =
{
  .title = "FACTRY",
  .sub_menu = menu_factory_list,
};
