#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "iface_sd.h"
#include "dm_partition.h"
#include "dsd.h"
#include "iface.h"
#include "display.h"
#include "system.h"
#include "we_config.h"
#include "fwinstall.h"
#include "app_db_changelog.h"
#include "commonstrings.h"
#include "iface_usb.h"
#include "app_devicemount.h"
#include "dbprint.h"
#include "filebackup.h"
#include "watchdog.h"
#include "comms_hdlr_rdb.h"

#include "iface_file.h"


/** Max length for filenames */
#define FILENAME_LEN 256


/** Is a usb disk mounted */
static bool usb_disk_mounted = false;


/* USB Partition Notification. This gets set high when a USB event happens */
bool usb_partition_notification = false;


/* USB Partition Text. This gets written with the event and the mount path. */
char usb_partition_notification_text[USB_PARTITION_NOTIFICATION_TEXT_SIZE] = "";


/** Function to call when a USB disk is mounted/unmounted */
void disk_mount_callback(void *client_callback_data, enum eDeviceMounterAction action, const char *mount_directory, bool result)
{
  usb_disk_mounted = (result && dmaAdd == action);

  /* Handle usb notifications to lua */
  if (usb_disk_mounted == true)
  {
    sprintf(usb_partition_notification_text, "added: %s", mount_directory);
    usb_partition_notification = true;
  }
  else if (action == dmaRemove)
  {
    sprintf(usb_partition_notification_text, "removed: %s", mount_directory);
    usb_partition_notification = true;
  }

  iface_usb_partition_callback(action, result);
}


/** Run a firmware update from a file on the USB disk. Only from full setup. */
static enum eMenuLevel run_firmware_update(bool full_edit, void *user_data, size_t index)
{
#define FULLPATH_LEN 1024
  char *filename = NULL;
  char *fullpath = NULL;
  bool res;
  int rpks;

  if (!full_edit)
  {
    access_denied();
    return mlNum;
  }

  filename = malloc(FILENAME_LEN);
  if (NULL == filename)
  {
    goto rfu_fail;
  }
  fullpath = malloc(FULLPATH_LEN);
  if (NULL == fullpath)
  {
    goto rfu_fail;
  }

  /*
   * Since there is no sensible way to list the directory contents
   * on the 6 character display, we make sure there is only
   * 1 .rpk file in the root of the disk before we allow installing
   */
  rpks = find_rpk(we_configuration.partition_mount_location, filename, FILENAME_LEN);
  if (1 == rpks)
  {
    /* Assemble full path to filename */
    strncpy(fullpath, we_configuration.partition_mount_location, FULLPATH_LEN);
    fullpath[FULLPATH_LEN - 1] = '\0';
    if ('/' != we_configuration.partition_mount_location[strlen(we_configuration.partition_mount_location) - 1])
      strncat(fullpath, "/", FULLPATH_LEN);
    strncat(fullpath, filename, FULLPATH_LEN);

    if (ask(cont_str, false))
    {

      /*
       * Shut down and disable the watchdog before calling install_rpk().
       * Calling install_rpk() for some as yet undetermined reason causes and call afterwards to
       * close(watchdog_fd) to not reach the kernel DS1511 watchdog driver .release function,
       * and so the hardware watchdog doesn't get disabled, and will cause a hard reset midway
       * through a firmware update.
       *
       * This is a workaround to http://bugs.rinstrumau.local/show_bug.cgi?id=76
       */
      watchdog_destroy();

      Dputstr(busy_str);
      res = install_rpk(fullpath);
      if (!res)
      {
        Dputstr(fail_str);
        delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
      }
/*
 * Not quite sure what to do here with success since WE2111 application will have to quit
      Dputstr((res) ? done_str : fail_str);
      delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
*/
    }
  }
  else if (0 == rpks)
  {
    error_strings(no_str, "RPK");
  }
  else
  {
    error_strings("MANY", "RPK");
  }

rfu_fail:
  if (NULL != fullpath)
  {
    free(fullpath);
    fullpath = NULL;
  }
  if (NULL != filename)
  {
    free(filename);
    filename = NULL;
  }
  return mlNum;
}


/** Backup configuration to USB disk */
static enum eMenuLevel run_config_file_save(bool full_edit, void *user_data, size_t index)
{
  char filename[FILENAME_LEN];
  const char *const folder = we_configuration.partition_mount_location;
  bool res;
  FILE *outfile;

  res = false;
  outfile = NULL;

  if (ask(cont_str, false))
  {
    Dputstr(busy_str);

    /* Create the filename */
    create_backup_filename(filename, FILENAME_LEN, folder, "configuration", ".txt");

    /* Open output file, truncate to 0 bytes ready for writing */
    outfile = fopen(filename, "w");
    if (NULL == outfile)
    {
      res = false;
      goto rcfs_done;
    }
    if (fcntl(fileno(outfile), F_SETFD, FD_CLOEXEC) < 0)
    {
      res = false;
      goto rcfs_done;
    }

    /* Write the database out to the file in human readable form */
    print_db_to_file(outfile, full_edit);
    res = true;

rcfs_done:
    if (NULL != outfile)
      fclose(outfile);

    Dputstr((res) ? done_str : fail_str);
    delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
  }
  return mlNum;
}


/** Save changelog to USB disk. Allowed in safe setup. */
static enum eMenuLevel run_changelog_file_save(bool full_edit, void *user_data, size_t index)
{
  bool res;

  if (ask(cont_str, false))
  {
    Dputstr(busy_str);
    res = app_changelog_backup_file(we_configuration.partition_mount_location);
    Dputstr((res) ? done_str : fail_str);
    delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
  }
  return mlNum;
}


/** Save the DSD file to the USB disk. Allowed in safe setup. */
static enum eMenuLevel run_dsd_file_save(bool full_edit, void *user_data, size_t index)
{
  bool res;

  if (!dsd_is_running())
  {
    error_strings(no_str, dsd_str);
  }
  else if (ask(cont_str, false))
  {
    Dputstr(busy_str);
    res = dsd_backup_file(we_configuration.partition_mount_location);
    Dputstr((res) ? done_str : fail_str);
    delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
  }
  return mlNum;
}


/** Save the debug information to a file to the USB disk. Only from full setup. */
static enum eMenuLevel run_debug_info_file_save(bool full_edit, void *user_data, size_t index)
{
  char filename[FILENAME_LEN];
  const char *const folder = we_configuration.partition_mount_location;
  FILE *outfile = NULL;
  bool res;

  if (!full_edit)
  {
    access_denied();
    return mlNum;
  }

  if (!ask(cont_str, false))
  {
    return mlNum;
  }

  Dputstr(busy_str);

  /* Create the filename */
  create_backup_filename(filename, FILENAME_LEN, folder, "debug_info", ".txt");

  /* Open output file, truncate to 0 bytes ready for writing */
  outfile = fopen(filename, "w");
  if (NULL == outfile)
  {
    res = false;
    goto rdifs_done;
  }
  if (fcntl(fileno(outfile), F_SETFD, FD_CLOEXEC) < 0)
  {
    res = false;
    goto rdifs_done;
  }

  /* Write the database out to the file in human readable form */
  res = save_debug_info_to_file(outfile);

rdifs_done:
  if (NULL != outfile)
  {
    fclose(outfile);
    outfile = NULL;
  }

  Dputstr((res) ? done_str : fail_str);
  delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
  return mlNum;
}


/** Eject the USB disk */
static enum eMenuLevel run_usb_disk_eject(bool full_edit, void *user_data, size_t index)
{
  int res;

  if (ask(cont_str, false))
  {
    Dputstr(busy_str);
    res = we2111_unmount_usbdisk();
    Dputstr((res >= 0) ? done_str : fail_str);
    delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
  }
  return mlZero;
}


/** Run checks before the file menu runs */
static bool file_menu_enter_exit(bool entry_not_exit, bool full_edit, void *user_data, bool edit_not_tostring)
{
  bool result = true;

  if (entry_not_exit && edit_not_tostring && !usb_disk_mounted)
  {
    error_strings(no_str, "DISK");
    result = false;
  }
  return result;
}


static const struct s_menu_item sub_menu_file_firmwareupdate = { .title = "FW.UPD",    .sub_menu = NULL, .edit_fn = run_firmware_update,      .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_file_savesettings =   { .title = "SAV.SET",   .sub_menu = NULL, .edit_fn = run_config_file_save,     .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_file_savechangelog =  { .title = "SAV.CL",    .sub_menu = NULL, .edit_fn = run_changelog_file_save,  .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_file_savedsd =        { .title = "SAV.DSD",   .sub_menu = NULL, .edit_fn = run_dsd_file_save,        .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_file_savedebug =      { .title = "SAV.DBG",   .sub_menu = NULL, .edit_fn = run_debug_info_file_save, .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_file_eject =          { .title = "EJECT",     .sub_menu = NULL, .edit_fn = run_usb_disk_eject,       .to_string_fn = to_string_no_setting };


static const struct s_menu_item *const menu_file_list[] =
{
  &sub_menu_file_firmwareupdate,
  &sub_menu_file_savesettings,
  &sub_menu_file_savechangelog,
  &sub_menu_file_savedsd,
  &sub_menu_file_savedebug,
  &sub_menu_file_eject,
  NULL
};


/** File menu item */
const struct s_menu_item menu_file =
{
  .title = "FILE",
  .sub_menu = menu_file_list,
  .menu_enter_exit = file_menu_enter_exit
};

static const struct s_menu_item *const sub_menu_usb_list[] =
{
  &sub_menu_file_eject,
  NULL
};

/** USB menu item */
const struct s_menu_item sub_menu_usb =
{
  .title = "USB",
  .sub_menu = sub_menu_usb_list,
  .menu_enter_exit = file_menu_enter_exit
};

