/*
 * iface_menus.c
 *
 *  Created on: Jun 15, 2012
 *      Author: ian
 */
#include <string.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcliveimpl_simple.h>
#include "display.h"
#include "comms_task.h"
#include "kbd_5000.h"
#include "access.h"
#include "setpoint.h"
#include "app_db_user.h"
#include "adcthread.h"
#include "device_finder_we2111.h"
#include "iface.h"
#include "editor.h"
#include "system.h"
#include "iface_sd.h"
#include "database.h"
#include "app_db_changelog.h"
#include "buzz5000.h"
#include "serial_usb_host.h"
#include "licensing.h"
#include "unitswitch.h"
#include "app_save.h"

#include "iface_alibi.h"
#include "iface_analog.h"
#include "iface_build.h"
#include "iface_cal.h"
#include "iface_changelog.h"
#include "iface_clock.h"
#include "iface_dsd.h"
#include "iface_ethnet.h"
#include "iface_factory.h"
#include "iface_file.h"
#include "iface_operator.h"
#include "iface_option.h"
#include "iface_serial.h"
#include "iface_setpoint.h"
#include "iface_special.h"
#include "iface_test.h"

#include "iface_menus.h"


static const struct s_menu_item menu_end = { .title = "-END-", .sub_menu = NULL };


static const struct s_menu_item *const menu_top_level_list[] =
{
  &menu_build,
  &menu_option,
  &menu_cal,
  &menu_serial,
  &menu_ethernet,
  &menu_special,
  &menu_test,
  &menu_setpoint,
  &menu_analog,
  &menu_clock,
  &menu_file,
  &menu_dsd,
  &menu_changelog,
  &menu_factory,
  &menu_end,
  NULL
};


/** The top level menu item for the setup menus */
const struct s_menu_item menu_top_level = { .title = "TL", .sub_menu = menu_top_level_list };


/** Returns true if passcode not needed or correct one entered */
static bool check_passcode(bool full_setup)
{
  bool passcode_required;
  long l;
  bool allow_access;
  uint32_t bypass_code;
  bool valid_bypass_code;

  allow_access = false;
  passcode_required = ((full_setup && db_app_user.full_passcode != 0) ||
                       (!full_setup && db_app_user.safe_passcode != 0));
  if (passcode_required)
  {
    valid_bypass_code = app_licensing_generate_setup_bypass_code(&bypass_code);

    prompt("ENTER", "PASS");
    l = 0;
    if (edit_long(&l, 999999L, 0L, 0, NUM_DIGITS, 0, ALLOW_SAFE) != EDIT_ABORT)
    {
      if (full_setup)
      {
        /* Check full setup passcode */
        allow_access = (l == (long) db_app_user.full_passcode ||
                        (valid_bypass_code && l == (long) bypass_code));
      }
      else
      {
        /* Check safe setup passcode */
        allow_access = (l == (long) db_app_user.safe_passcode);
      }
    }

    /* Incorrect passcode entered */
    if (!allow_access)
    {
      access_denied();
    }
  }
  else
  {
    /* Passcode is zero */
    allow_access = true;
  }
  return allow_access;
}


/** Attempt to save the user database. Return true on change log not full (db saved) */
static bool save_user_db_and_reconfigure(void)
{
  bool changelog_not_full;

  changelog_not_full = app_can_save_databases();

  if (changelog_not_full)
  {
    /* Changelog not full, ok to save */
    Dputstr("SAVING");
    (void) app_save_databases_and_reinit(false, NULL);
    clr_error(ADC_ERR_USER_DB, "save_user_db_and_reconfigure: clearing error ADC_ERR_USER_DB\n");
  }
  else
  {
    /* Changelog full, alert user and remain in menus */
    beep_long();
    error_strings("NO.SAVE", "CL.FULL");
  }

  return changelog_not_full;
}


/** Setup display and keyboard for running menus */
static void setup_for_menus(void)
{
  /* Setup keyboard for all long presses */
  setup_kbd(KBD_MENUS_REPEAT, KBD_MENUS_HOLD);

  show_hold = dhNone;
  blank_display = false;

  freeze_display_and_annunciators(true);

  Dclear_flashing();
  Dset_units_ignore_motion(ADC_UNITS_NONE);
  Dset_annunciators(ANNUN_NONE);

  edit_functions = true;
}


/** Setup for normal weighing */
void setup_for_normal_weighing(void)
{
  edit_functions = false;
  setup_kbd(KBD_DEFAULT_REPEAT, KBD_DEFAULT_HOLD);
  setup_remote_kbd();
  Dclear_flashing();
  Dset_brightness();
  freeze_display_and_annunciators(false);
  update_adc_operation_counter();
}


static void go_setup(bool full_edit, bool rear_key)
{
  #define OIML_STR_LEN 10
  char oiml_str[OIML_STR_LEN];
  bool saved;
  enum eOIMLError oiml;

  rear_key_entry = rear_key;

  /* Safe or full setup, display menu stuff */
  Dputstr("SETUP");

  init_setpoints(&setpoint_running_info, &adcSuperStructure);
  delay(WECFG_DELAY_FULL_OPERATING_MILLISECONDS);
  show_version_and_cal_cnt();

  /* Check passcodes */
  if (check_passcode(full_edit))
  {
    for (saved = false; !saved && !application_quit_forced(); )
    {
      /* Run the top level menu loop */
      run_menu_editor(&menu_top_level, full_edit, mlZero);

      /* Save settings, if change log not full */
      saved = save_user_db_and_reconfigure();
    }

    /* Show calibration count on menu exit */
    if (!application_quit_forced())
      show_version_and_cal_cnt();
  }

  if (!application_quit_forced())
  {
    /* Check trade settings */
    oiml = check_oiml();
    if (oiml > OIML_ERROR_NONE)
    {
      snprintf(oiml_str, OIML_STR_LEN, "TRADE.%01u", (unsigned) oiml);
      error_strings("CHECK", oiml_str);
    }
  }
}


static enum eMenuLevel menus_from_front_key(bool dummy_full_edit, void *user_data, size_t index)
{
  bool full_setup = (bool) user_data;
  bool full_edit;

  if (!full_setup)
  {
    /* Safe setup option selected */
    full_edit = false;
  }
  else
  {
    /* Full setup option selected */
    if (db_app_user.rear_entry)
    {
      /* No full access if rear_entry setting set */
      access_denied();
      return mlQuit;
    }
    else
    {
      /* Full edit allowed */
      full_edit = true;
    }
  }

  go_setup(full_edit, false);
  return mlQuit;
}


static const struct s_menu_item menu_front_setup_safe =  { .title = "SAFE",  .sub_menu = NULL, .edit_fn = menus_from_front_key, .user_data = (void *) false };
static const struct s_menu_item menu_front_setup_full =  { .title = "FULL",  .sub_menu = NULL, .edit_fn = menus_from_front_key, .user_data = (void *) true };
static const struct s_menu_item menu_front_setup_alibi = { .title = "ALIBI", .sub_menu = NULL, .edit_fn = do_alibi };


static const struct s_menu_item *const sub_menu_front_setup[] =
{
  &menu_operator,
  &menu_front_setup_safe,
  &menu_front_setup_full,
  &menu_front_setup_alibi,
  NULL
};


/** Build menu item */
static const struct s_menu_item menu_front_setup =
{
  .title = "F.SETUP",
  .sub_menu = sub_menu_front_setup,
};


/** Run the setup menus */
void run_setup_menus(bool rear_key)
{
  setup_for_menus();

  /* Re-configure setpoints to disable last outputs while in menus */
  init_setpoints(&setpoint_running_info, &adcSuperStructure);

  /* Display SAFE/FULL/Alibi menu if front key pressed */
  if (!rear_key)
  {
    run_menu_editor(&menu_front_setup, false, mlTwo);
  }
  else
  {
    /* Rear key, so always full setup */
    go_setup(true, true);
  }

  /* Re-configure setpoints in case targets have changed */
  init_setpoints(&setpoint_running_info, &adcSuperStructure);

  setup_for_normal_weighing();
  app_licensing_reset_check_time();
}
