/*
 * iface_operator.c
 *
 *  Created on: 08/09/2011
 *      Author: ian
 */
#include <stdio.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "db_setpoint.h"
#include "iface.h"
#include "database.h"
#include "editor.h"
#include "display.h"
#include "iface_sd.h"
#include "meas.h"
#include "iface_setpoint.h"
#include "iface_ethnet.h"
#include "iface_rinblocks.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "app_db_runtime_vars.h"
#include "app_db_runtime.h"
#include "commonstrings.h"
#include "iface_file.h"
#include "unitswitch.h"

#include "iface_operator.h"


/** Length of operator setpoint menu strings */
#define SETPOINT_MENU_STR_LEN 20


/** Typedef of array of strings for setpoint names */
typedef char t_sub_menu_operator_setpoint_names[NUM_SETPOINTS][SETPOINT_MENU_STR_LEN];


/** Work out the upper and lower limits for a preset tare TODO new adc library checks this, so we shouldn't do it here */
void calculate_preset_tare_limits(int32s *lower, int32s *upper)
{
  int32s tr_l, tr_h;

  /* Work out upper limit */
  tr_h = true_full_scale();
  if (adcSuperStructure.adcUserDatabase.scaleType == ADC_SCALE_TYPE_DUAL_INTERVAL)
    tr_h = adcSuperStructure.adcUserDatabase.fullScale[0];
  *upper = tr_h;

  /* Work out lower limit */
  tr_l = (ADC_APPROVAL_MODE_NONE != adcSuperStructure.adcUserDatabase.approvalMode) ? 0L : -tr_h;
  *lower = tr_l;
}


static struct s_menu_item sub_menu_operator_setpoints[NUM_SETPOINTS];


/** Target sub menu list, constructed by @ref operator_target_menu_enter_exit() */
static struct s_menu_item *sub_menu_operator_setpoints_list[NUM_SETPOINTS + 1];


typedef int32_t (*t_setpoint_getter_fn)(uint8_t num);
typedef bool (*t_setpoint_setter_fn)(uint8_t num, int32_t data);


/** Edit a setpoint value (target, inflight, hysteresis) */
static enum eMenuLevel edit_setpoint_value(size_t setpoint, t_setpoint_getter_fn setpoint_getter, t_setpoint_setter_fn setpoint_setter, long max, long min)
{
  enum eEditResult result;
  long l;
  uint8_t dp;
  enum eAdcUnits adcunits;
  enum eAdcUnitSwitch usw;

  /* Get value, units and decimal point */
  l = setpoint_getter(setpoint);
  usw = (ADC_APPROVAL_MODE_NTEP == adcSuperStructure.adcUserDatabase.approvalMode) ?
        unitsw_displaymode_to_unitswitch(adcSuperStructure.adcDataWorking.dispMode) :
        ADC_UNIT_SWITCH_PRIMARY;
  l = unitsw_convert_final(l, ADC_UNIT_SWITCH_PRIMARY, usw, false, false);
  adcunits = unitsw_get_units(usw);
  dp = unitsw_get_decimal_point(usw);

  /* Edit the value */
  Dset_units_ignore_motion(adcunits);
  result = edit_long(&l, max, min, dp, NUM_DIGITS, 0, ALLOW_SAFE);

  /* Save result */
  if (EDIT_CHG == result)
  {
    l = unitsw_convert_final(l, usw, ADC_UNIT_SWITCH_PRIMARY, false, false);
    setpoint_setter(setpoint, l);
  }
  return edit_result_to_menu_level(result);
}


/** Edit a setpoint target */
static enum eMenuLevel edit_setpoint_target(bool full_edit, void *user_data, size_t index)
{
  intptr_t setpoint = (intptr_t) user_data;

  return edit_setpoint_value(setpoint, setpoint_target_get, setpoint_target_set, APP_DB_R_TARGETS_MAX, APP_DB_R_TARGETS_MIN);
}


/** Edit a setpoint inflight */
static enum eMenuLevel edit_setpoint_inflight(bool full_edit, void *user_data, size_t index)
{
  intptr_t setpoint = (intptr_t) user_data;

  return edit_setpoint_value(setpoint, setpoint_preflight_get, setpoint_preflight_set, APP_DB_R_PREFLIGHT_MAX, APP_DB_R_PREFLIGHT_MIN);
}


/** Edit a setpoint hysteresis */
static enum eMenuLevel edit_setpoint_hysteresis(bool full_edit, void *user_data, size_t index)
{
  intptr_t setpoint = (intptr_t) user_data;

  return edit_setpoint_value(setpoint, setpoint_hysteresis_get, setpoint_hysteresis_set, APP_DB_R_HYSTERESIS_MAX, APP_DB_R_HYSTERESIS_MIN);
}


/**
 * Construct menu items for the setpoint parameters from the operator menu
 *
 * @param param is which parameter to edit (target/inflight/hysteresis)
 */
static int construct_setpoint_menu(struct s_menu_item **items_list, struct s_menu_item *items, t_sub_menu_operator_setpoint_names *names, enum eSetpointParameter param)
{
  size_t sp;
  size_t active_sp;
  size_t num_active_sp;

  /* Make a list of unlocked active setpoints */
  for (sp = 0, active_sp = 0; sp < NUM_SETPOINTS; sp++)
  {
    if (0 == db_app_user.setp_lock[sp] && (SETP_OVER == db_app_user.setp_active[sp] || SETP_UNDER == db_app_user.setp_active[sp] ||
                                           SETP_W_OUT == db_app_user.setp_active[sp] || SETP_W_IN == db_app_user.setp_active[sp]))
    {
      switch (param)
      {
        case spTarget:
          snprintf((*names)[active_sp], SETPOINT_MENU_STR_LEN, "TARG%u", (unsigned int) sp + 1);
          items[active_sp].title = (*names)[active_sp];
          items[active_sp].edit_fn = edit_setpoint_target;
          break;
        case spInFlight:
          snprintf((*names)[active_sp], SETPOINT_MENU_STR_LEN, "FLT %u",  (unsigned int) sp + 1);
          items[active_sp].title = (*names)[active_sp];
          items[active_sp].edit_fn = edit_setpoint_inflight;
          break;
        case spHysteresis:
          snprintf((*names)[active_sp], SETPOINT_MENU_STR_LEN, "HYS %u",  (unsigned int) sp + 1);
          items[active_sp].title = (*names)[active_sp];
          items[active_sp].edit_fn = edit_setpoint_hysteresis;
          break;
      }
      items[active_sp].sub_menu = NULL;
      items[active_sp].menu_enter_exit = NULL;
      items[active_sp].user_data_from_menu_above = false;
      items[active_sp].user_data = (void *) sp;

      items_list[active_sp] = &items[active_sp];
      active_sp++;
    }
  }
  num_active_sp = active_sp;

  /* The remainder don't show */
  for ( ; active_sp < NUM_SETPOINTS + 1; active_sp++)
  {
    items_list[active_sp] = NULL;
  }

  return num_active_sp;
}


/** Edit the preset tare value */
enum eMenuLevel edit_preset_tare(bool full_edit, void *user_data, size_t index)
{
  long l;
  int32s tr_h, tr_l;
  AdcRdg rdg;
  enum eEditResult result;
  enum eAdcUnitSwitch usw;

  /* Work out limits */
  calculate_preset_tare_limits(&tr_l, &tr_h);

  /* Get the preset tare */
  usw = (ADC_APPROVAL_MODE_NTEP == adcSuperStructure.adcUserDatabase.approvalMode) ?
         unitsw_displaymode_to_unitswitch(adcSuperStructure.adcDataWorking.dispMode) : ADC_UNIT_SWITCH_PRIMARY;
  unitsw_get_preset_tare(&rdg, usw);

  /* Edit the preset tare */
  l = rdg.rdg;
  Dset_units_ignore_motion(rdg.info.unitdp.bit.units);
  result = edit_long(&l, tr_h, tr_l, rdg.info.unitdp.bit.dp, NUM_DIGITS, 0, ALLOW_SAFE);
  if (EDIT_CHG == result || EDIT_OK == result)
  {
    l = unitsw_convert_final(l, usw, ADC_UNIT_SWITCH_PRIMARY, rdg.info.status.bit.hiRes, false);

    /* Set new preset tare (always in non-hires) */
    AdcSetPTare(&adcSuperStructure, true, false, false, l);
    update_adc_operation_counter();

    /* Exit menus */
    result = EDIT_MENU_QUIT;
  }
  return edit_result_to_menu_level(result);
}


/** Names for each of the setpoints */
static t_sub_menu_operator_setpoint_names sub_menu_operator_setpoint_names;


/** Setpoint sub menu items, filled out by @ref operator_target_menu_enter_exit() */
static struct s_menu_item sub_menu_operator_setpoints[NUM_SETPOINTS];


/** Setpoint sub menu list, constructed by @ref operator_target_menu_enter_exit() */
static struct s_menu_item *sub_menu_operator_setpoints_list[NUM_SETPOINTS + 1];


/** Construct the operator target submenu */
static bool operator_setpoints_menu_enter_exit(bool entry_not_exit, bool full_edit, void *user_data, bool edit_not_tostring)
{
  enum eSetpointParameter param = (enum eSetpointParameter) user_data;
  bool can_run;

  if (entry_not_exit)
  {
    can_run = (construct_setpoint_menu(sub_menu_operator_setpoints_list, sub_menu_operator_setpoints, &sub_menu_operator_setpoint_names, param) > 0);
    if (!can_run && edit_not_tostring)
    {
      error_strings(no_str, "SETP");
    }
    return can_run;
  }
  else
  {
    /* Can exit */
    return true;
  }
}


/*
 * Remove preset tare because this is available from long press tare key
 * static const struct s_menu_item sub_menu_operator_presettare = { .title = preset_tare_str, .sub_menu = NULL, .edit_fn = edit_preset_tare };
 */
static const struct s_menu_item sub_menu_operator_target =   { .title = "TARGET", .sub_menu = sub_menu_operator_setpoints_list, .menu_enter_exit = operator_setpoints_menu_enter_exit, .user_data = (void *) spTarget };
static const struct s_menu_item sub_menu_operator_inflight = { .title = "FLIGHT", .sub_menu = sub_menu_operator_setpoints_list, .menu_enter_exit = operator_setpoints_menu_enter_exit, .user_data = (void *) spInFlight };
static const struct s_menu_item sub_menu_operator_end =      { .title = "-END-",  .sub_menu = NULL };


static const struct s_menu_item *const menu_operator_list[] =
{
  /* &sub_menu_operator_presettare, */
  &sub_menu_operator_target,
  &sub_menu_operator_inflight,
  &sub_menu_usb,
  &sub_menu_oper_ip,
  &menu_operator_modules,
  &sub_menu_operator_end,
  NULL
};


/** Operator top level menu item */
static const struct s_menu_item menu_operator_top_level =
{
  .title = "OPER.TL",
  .sub_menu = menu_operator_list,
};


/** Run the operator menu from the top level */
static enum eMenuLevel run_operator_menu(bool dummy_full_edit, void *user_data, size_t index)
{
  /* Run the operator top level menu loop */
  run_menu_editor(&menu_operator_top_level, false, mlZero);
  return mlQuit;
}


/** Operator menu item */
const struct s_menu_item menu_operator =
{
  .title = "OPER",
  .edit_fn = run_operator_menu,
};
