#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "iface.h"
#include "system.h"
#include "kbd_5000.h"
#include "database.h"
#include "meas.h"
#include "backgrnd.h"
#include "adc_8bit.h"
#include "asciidef.h"
#include "version.h"
#include "buzz5000.h"
#include "comms_print.h"
#include "editor.h"
#include "meas.h"
#include "sysclock.h"
#include "access.h"
#include "setpoint.h"
#include "dbprint.h"
#include "display.h"
#include "iface_cal.h"
#include "we_config.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "commonstrings.h"

#include "iface_sd.h"


/** Get permissions for the front key */
enum eFrontKeyPermissions get_frontkey_permissions(enum eFrontKeys key)
{
  if (key < NUM_FRONT_KEYS)
    return db_app_user.front_key_permissions[key];
  else
    return FRONTKEY_PERMIS_DISABLED;
}


/**
 * Waits while there is an error or motion for a max of 15 seconds.
 *
 * @return true if the reading settles, or displays an error and returns false
 */
bool wait_till_no_motion(void)
{
  uint32_t finish_time;
  bool failed;

  /* Wait for 15 seconds */
  finish_time = OSTimeGet() + OS_TICKS_PER_SEC * 15;
  failed = false;
  while (motion())
  {
    if (APP_ERR_NONE != get_current_errors() || OSTimeGet() >= finish_time)
    {
      failed = true;
      break;
    }
    OSTimeDly(OS_TICKS_PER_SEC / 10);
  }

  if (failed)
  {
    freeze_display(true);
    error_strings("STABLE", err_str);
    freeze_display(false);
    return false;
  }

  return true;
}


/** Check for error with initial zero, and if so display error */
void check_and_display_startup_zero_error(void)
{
  if (ADC_OP_CMD_INIT_ZERO == AdcGetCurrentCommand(&adcSuperStructure, true))
  {
    if (ADC_OP_STATUS_IN_PROGRESS == AdcGetCurrentOpStatus(&adcSuperStructure, true))
    {
      /* Display zero in progress message */
      freeze_display_and_annunciators(true);
      Dset_annunciators(ANNUN_NONE);
      Dset_units_ignore_motion(ADC_UNITS_NONE);
      Dputstr(zero_in_progress_str);

      /* Wait till complete */
      while (ADC_OP_STATUS_IN_PROGRESS == AdcGetCurrentOpStatus(&adcSuperStructure, true))
      {
        delay(100);
      }
    }

    /* Display result */
    if (ADC_OP_STATUS_SUCCESS != AdcGetCurrentOpStatus(&adcSuperStructure, true))
    {
      freeze_display_and_annunciators(true);
      error_strings(zero_str, err_str);
    }
    freeze_display_and_annunciators(false);
  }
}


/**
 * Prompt user for 'Y'/'n' response to prompt
 *
 * @param s is the prompt
 * @param def is the default response
 *
 * @return true for Y, false for n
 *
 */
bool ask(const char *s, bool def)
{
  uint8_t kbd;
  bool ans;
  char p[NUM_DIGITS + 1];
  int i;

  /* Make question string */
  for (i = 0; i < NUM_DIGITS; i++)
  {
    if (*s)
    {
      p[i] = *s++;
    }
    else
      p[i] = ' ';
  }
  p[i] = '\0';

  ans = def;

  /* Tack on Y/n at the end */
  p[NUM_DIGITS - 1] = (ans) ? 'Y' : 'N';
  Dputstr(p);

  /* Wait for user response */
  while (1)
  {
    kbd = get_key();
    if (KEY_EDIT_UP == kbd || KEY_EDIT_DOWN == kbd)
    {
      /* Invert answer and re-display */
      ans = !ans;
      p[NUM_DIGITS - 1] = (ans) ? 'Y' : 'N';
      Dputstr(p);
    }
    /* else if (NO_KEY != kbd && IDLE_CHG != kbd) */
    else if (KEY_EDIT_OK == kbd)
    {
      break;
    }
  }

  /* Return result */
  return ans;
}


static uint32_t calculate_finish_time_ticks(uint32_t timeout_ms)
{
  uint32_t finish_time;

  finish_time = timeout_ms * OS_TICKS_PER_SEC;
  finish_time /= 1000;
  finish_time += OSTimeGet();

  return finish_time;
}


/**
 * Wait timeout_ms milliseconds for a key press.
 *
 * @param timeout_ms is the timeout in milliseconds, or 0 for indefinite timeout
 * @param quit is a bool pointer to check to signal the application should quit. Can be NULL
 *
 * @return true if key pressed.
 */
bool wait_for_press(uint32_t timeout_ms, bool *quit)
{
  uint32_t finish_time = calculate_finish_time_ticks(timeout_ms);

  while ((timeout_ms == 0 || OSTimeGet() < finish_time) && (NULL == quit || !*quit))
  {
    switch (get_key())
    {
      case IDLE_CHG:
      case NO_KEY:
        break;
      default:
        return true;
    }
  }
  return false;
}


/**
 * Prompt user with multiple strings.
 *
 * @param rate_ms is the delay between alternating strings in milliseconds.
 * @param timeout_ms is the timeout in milliseconds, or 0 for indefinite timeout
 * @param quit is a bool pointer to check to signal the application should quit. Can be NULL
 * @param ... is a number of strings terminated by NULL: "hello", "there", "bob", NULL
 *
 * @return true if key pressed, false if timed out or quit
 */
bool prompt_multi(uint32_t rate_ms, uint32_t timeout_ms, bool *quit, ...)
{
  va_list arglist;
  const char *prompt;
  uint32_t finish_time = calculate_finish_time_ticks(timeout_ms);
  bool quitfn;
  bool result = false;

  Dset_annunciators(ANNUN_NONE);
  Dset_units_ignore_motion(ADC_UNITS_NONE);

  for (quitfn = false; !quitfn; )
  {
    va_start(arglist, quit);

    while (1)
    {
      prompt = va_arg(arglist, const char *);
      if (NULL == prompt)
        break;

      Dputstr(prompt);
      if (wait_for_press(rate_ms, quit))
      {
        quitfn = true;
        result = true;
        break;
      }

      if ((0 != timeout_ms && OSTimeGet() >= finish_time) || (NULL != quit && *quit))
      {
        quitfn = true;
        result = false;
        break;
      }
    }

    va_end(arglist);
  }
  return result;
}


/**
 * Prompt user with 2 strings. Second string can be "".
 *
 * @param s1 is the first prompt string. Not NULL
 * @param s2 is the second prompt string. Can be NULL or "" if not required.
 * @param rate_ms is the delay between alternating strings in milliseconds.
 * @param timeout_ms is the timeout in milliseconds, or 0 for indefinite timeout
 * @param quit is a bool pointer to check to signal the application should quit. Can be NULL
 *
 * @return true if key pressed, false if timed out or quit
 */
bool prompt_timeout(const char *s1, const char *s2, uint32_t rate_ms, uint32_t timeout_ms, bool *quit)
{
  return prompt_multi(rate_ms, timeout_ms, quit, s1, s2, NULL);
}


/** Prompt user with 2 strings. Second string can be "". */
bool prompt(const char *s1, const char *s2)
{
  return prompt_timeout(s1, s2, 1250, 1250 * 2, NULL);
}


/** Display an error message on the display */
void error_strings(const char *s1, const char *s2)
{
  Dclear_flashing();
  beep_error();

  (void) prompt_timeout(s1, s2, 1250, 1250 * 6, NULL);
}


/** "NO" "ENTRY" error message on display */
void access_denied(void)
{
  error_strings(no_str, "ENTRY");
}
