/*
 * iface_serial_config.c
 *
 * Serial configuration menus (auto opts, printer opts, net opts)
 *
 *  Created on: 03/09/2016
 *      Author: ian
 */
#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <ctype.h>
#include "app_db_user.h"
#include "app_db_runtime_vars.h"
#include "editor.h"
#include "iface_sd.h"
#include "modbus_config.h"
#include "dbprint.h"
#include "edit_string.h"

#include "iface_serial_config.h"


/** Automatic output formats */
const char *const auto_setting[AUTO_OUT_FMT_NUM] = AUTO_OUT_FMT_STRS;


/** Automatic output sources */
const char *const auto_output_src[AUTO_OUT_SRC_NUM] = AUTO_OUTPUT_SRC_STRS;


/** Printer mode setting strings */
const char *const printer_mode_setting[NUM_PRINTER_MODES] = PRINTER_MODE_STRS;


/** Modbus endian setting strings */
const char *const rinmodbus_endian_strs[RINMODBUS_ENDIAN_NUM] = RINMODBUS_ENDIAN_STRS;


/** Printer interlock setting strings */
const char *const print_interlock_setting[NUM_PRINT_INTERLOCK] = PRINT_INTERLOCK_STRS;


static enum eEditResult edit_fmt(char *bp, size_t len, bool full)
{
  size_t i;
  long l;
  enum eEditResult result = mlNum;

  for (i = 0; i < len; )
  {
    /* Edit number ppp.ccc (p-position, c-char) */
    l = (uint8_t) bp[i];
    l += (i + 1) * 1000;
    result = edit_long(&l, 999999L, 0L, 0, 3, 3, full);
    if (EDIT_OK == result || EDIT_CHG == result)
    {
      /* What's the new character value */
      l -= (i + 1) * 1000;
      if (l > 255L)
      {
        block_entry();
      }
      else
      {
        bp[i] = (uint8_t) l;
        if (0 == l)
        {
          for (; i <= len; i++)
            bp[i] = '\0';
          break;
        }
        i++;
      }
    }
    else
    {
      /* Exit loop on abort key */
      break;
    }
  } /* for */
  return result;
}


/** Edit the serial network address */
static enum eMenuLevel edit_network_address(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_byte(&db_app_user.address, 31, 0, 0, 2, ALLOW_SAFE));
}


/** Convert the serial network address to a string */
static void to_string_network_address(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%u", (unsigned) db_app_user.address);
}


/** Edit the serial network custom identifier */
static enum eMenuLevel edit_network_custom_id(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_printable_string(db_app_user.id, APP_DB_ID_STR_LEN, charset_alpha_numeric));
}


/** Convert the serial network custom identifier string */
static void to_string_network_custom_id(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", db_app_user.id);
}


static enum eMenuLevel edit_autoout_type(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_register_data(&db_app_user.aof_setting, 0, AUTO_OUT_FMT_NUM - 1, auto_setting, ALLOW_SAFE));
}


/** Convert the auto output type to a string */
static void to_string_autoout_type(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", auto_setting[db_app_user.aof_setting]);
}


static enum eMenuLevel edit_autoout_source(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_register_data(&db_app_user.aof_source, 0, AUTO_OUT_SRC_NUM - 1, auto_output_src, ALLOW_SAFE));
}


/** Convert the auto output source to a string */
static void to_string_autoout_source(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", auto_output_src[db_app_user.aof_source]);
}


static enum eMenuLevel edit_autoout_format(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_fmt(db_app_user.auto_fmt, APP_DB_AUTO_FMT_LEN, ALLOW_SAFE));
}


/** Convert the auto output format to a string */
static void to_string_autoout_format(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  to_string_print_fmt(db_app_user.auto_fmt, buffer, buflen);
}


static enum eMenuLevel edit_autoout_start_char(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_byte(&db_app_user.start_char, 255, 0, 0, 3, ALLOW_SAFE));
}

/** Convert the auto output start char to a string */
static void to_string_autoout_start_char(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%u", (unsigned) db_app_user.start_char);
}


static enum eMenuLevel edit_autoout_end_char1(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_byte(&db_app_user.end_char1, 255, 0, 0, 3, ALLOW_SAFE));
}


/** Convert the auto output end char 1 to a string */
static void to_string_autoout_end_char1(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%u", (unsigned) db_app_user.end_char1);
}


static enum eMenuLevel edit_autoout_end_char2(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_byte(&db_app_user.end_char2, 255, 0, 0, 3, ALLOW_SAFE));
}


/** Convert the auto output end char 2 to a string */
static void to_string_autoout_end_char2(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%u", (unsigned) db_app_user.end_char2);
}


static enum eMenuLevel edit_print_type(bool full_edit, void *user_data, size_t index)
{
  enum eEditResult result;

  result = edit_register_data(&db_app_user.printer_mode, 0, NUM_PRINTER_MODES - 1, printer_mode_setting, ALLOW_SAFE);
  if (EDIT_CHG == result)
  {
    if (db_app_user.printer_mode >= P_TOTAL)
    {
      statistics_clear();
    }
  }
  return result;
}


/** Convert the print type to a string */
static void to_string_print_type(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", printer_mode_setting[db_app_user.printer_mode]);
}


static enum eMenuLevel edit_print_header(bool full_edit, void *user_data, size_t index)
{
  size_t pos;
  size_t line;
  long l;
  enum eEditResult result = EDIT_OK;

  for (line = 0; line < APP_DB_PRINT_HDR_LINES; line++)
  {
    for (pos = 0; pos < APP_DB_PRINT_HDR_LEN - 1; )
    {
      /* We display line.pos.char as l.pp.ccc, so mangle the numbers into place */
      l = (long) db_app_user.print_text[line][pos];
      l += (line + 1) * 100000;
      l += (pos + 1) * 1000;

      /* Allow edit of last 3 digits (ccc) */
      result = edit_long(&l, 999999L, 0L, 5, 3, 3, ALLOW_SAFE);
      if (EDIT_OK == result || EDIT_CHG == result)
      {
        /* Strip off line & pos */
        l = l % 1000;
        if (l > UCHAR_MAX || !isprint((char) l))
        {
          error_strings("BAD", "CHAR");
        }
        else
        {
          db_app_user.print_text[line][pos] = (char) l;
          pos++;
        }
      }
      else
      {
        return edit_result_to_menu_level(result);
      }
    }
  }
  return edit_result_to_menu_level(result);
}


static enum eMenuLevel edit_print_format(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_fmt(db_app_user.print_fmt, APP_DB_PRINT_FMT_LEN, ALLOW_SAFE));
}


/** Convert the auto output format to a string */
static void to_string_print_format(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  to_string_print_fmt(db_app_user.print_fmt, buffer, buflen);
}


static enum eMenuLevel edit_print_space(bool full_edit, void *user_data, size_t index)
{
  long l;
  uint16_t w;
  uint8_t r, c;
  enum eEditResult result;

  w = ((uint16_t) (db_app_user.tic_col) * 100) + db_app_user.tic_row;
  l = w;
  for (result = EDIT_OK; ; )
  {
    result = edit_long(&l, 9999L, 0L, 2, 4, 0, ALLOW_SAFE);
    if (EDIT_ABORT == result || EDIT_MENU_QUIT == result)
    {
      break;
    }
    w = (uint16_t) (l);
    c = (uint8_t) (w / 100);
    r = (uint8_t) (w % 100);
    if ((r > APP_DB_U_SERTICKETROW_MAX) || (c > APP_DB_U_SERTICKETCOL_MAX))
    {
      block_entry();
    }
    else
    {
      db_app_user.tic_col = c;
      db_app_user.tic_row = r;
      break;
    }
  }
  return edit_result_to_menu_level(result);
}


/** Convert the print space to a string */
static void to_string_print_space(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%02u.%02u", (unsigned) db_app_user.tic_col, (unsigned) db_app_user.tic_row);
}


/** Edit the modbus endian setting */
static enum eMenuLevel edit_modbus_endian(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_register_data(&db_app_user.modbus_endian, 0, RINMODBUS_ENDIAN_NUM - 1, rinmodbus_endian_strs, ALLOW_SAFE));
}


/** Convert the modbus endian setting to a string */
static void to_string_modbus_endian(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", rinmodbus_endian_strs[db_app_user.modbus_endian]);
}


/** Edit TCP socket port */
static enum eMenuLevel edit_modbus_tcpport(bool full_edit, void *user_data, size_t index)
{
  enum eEditResult result;
  long tcpport;

  tcpport = db_app_user.modbus_tcpport;
  result = edit_long(&tcpport, APP_DB_U_MODBUSTCPPORT_MAX, APP_DB_U_MODBUSTCPPORT_MIN, 0, 5, 0, true);
  if (EDIT_CHG == result)
  {
    db_app_user.modbus_tcpport = tcpport;
  }
  return edit_result_to_menu_level(result);
}


/** Convert the modbus tcp port to a string */
static void to_string_modbus_tcpport(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%lu", (unsigned long) db_app_user.modbus_tcpport);
}


static enum eMenuLevel edit_print_interlock(bool full_edit, void *user_data, size_t index)
{
  return edit_result_to_menu_level(edit_register_data(&db_app_user.print_interlock, 0, NUM_PRINT_INTERLOCK - 1, print_interlock_setting, ALLOW_SAFE));
}


/** Convert the auto output type to a string */
static void to_string_print_interlock(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%s", print_interlock_setting[db_app_user.print_interlock]);
}


/*
 * Network options menu
 */


static const struct s_menu_item menu_serial_netopt_address =       { .title = "ADDRES", .sub_menu = NULL, .edit_fn = edit_network_address, .to_string_fn = to_string_network_address };
static const struct s_menu_item menu_serial_netopt_custom_id =     { .title = "ID",     .sub_menu = NULL, .edit_fn = edit_network_custom_id,    .to_string_fn = to_string_network_custom_id };


static const struct s_menu_item *const menu_serial_netopt_list[] =
{
  &menu_serial_netopt_address,
  &menu_serial_netopt_custom_id,
  NULL
};


/*
 * Auto output options menu
 */


static const struct s_menu_item sub_menu_serial_autoopt_type =    { .title = "TYPE",    .sub_menu = NULL, .edit_fn = edit_autoout_type,       .to_string_fn = to_string_autoout_type };
static const struct s_menu_item sub_menu_serial_autoopt_source =  { .title = "SOURCE",  .sub_menu = NULL, .edit_fn = edit_autoout_source,     .to_string_fn = to_string_autoout_source };
static const struct s_menu_item sub_menu_serial_autoopt_format =  { .title = "AUT.FMT", .sub_menu = NULL, .edit_fn = edit_autoout_format,     .to_string_fn = to_string_autoout_format };
static const struct s_menu_item sub_menu_serial_autoopt_startch = { .title = "ST.CHR",  .sub_menu = NULL, .edit_fn = edit_autoout_start_char, .to_string_fn = to_string_autoout_start_char };
static const struct s_menu_item sub_menu_serial_autoopt_endch1 =  { .title = "END.CH1", .sub_menu = NULL, .edit_fn = edit_autoout_end_char1,  .to_string_fn = to_string_autoout_end_char1 };
static const struct s_menu_item sub_menu_serial_autoopt_endch2 =  { .title = "END.CH2", .sub_menu = NULL, .edit_fn = edit_autoout_end_char2,  .to_string_fn = to_string_autoout_end_char2 };


static const struct s_menu_item *const menu_serial_autoopt_list[] =
{
  &sub_menu_serial_autoopt_type,
  &sub_menu_serial_autoopt_source,
  &sub_menu_serial_autoopt_format,
  &sub_menu_serial_autoopt_startch,
  &sub_menu_serial_autoopt_endch1,
  &sub_menu_serial_autoopt_endch2,
  NULL
};


/*
 * Print options menu
 */


static const struct s_menu_item sub_menu_serial_prnopt_printtype = { .title = "PRNT.TP", .sub_menu = NULL, .edit_fn = edit_print_type,      .to_string_fn = to_string_print_type };
static const struct s_menu_item sub_menu_serial_prnopt_header =    { .title = "HEADER",  .sub_menu = NULL, .edit_fn = edit_print_header };
static const struct s_menu_item sub_menu_serial_prnopt_ticfmt =    { .title = "TIC.FMT", .sub_menu = NULL, .edit_fn = edit_print_format,    .to_string_fn = to_string_print_format };
static const struct s_menu_item sub_menu_serial_prnopt_space =     { .title = "SPACE",   .sub_menu = NULL, .edit_fn = edit_print_space,     .to_string_fn = to_string_print_space };
static const struct s_menu_item sub_menu_serial_prnopt_interlock = { .title = "I.LOCK",  .sub_menu = NULL, .edit_fn = edit_print_interlock, .to_string_fn = to_string_print_interlock };


static const struct s_menu_item *const menu_serial_prnopt_list[] =
{
  &sub_menu_serial_prnopt_printtype,
  &sub_menu_serial_prnopt_header,
  &sub_menu_serial_prnopt_ticfmt,
  &sub_menu_serial_prnopt_space,
  &sub_menu_serial_prnopt_interlock,
  NULL
};


/*
 * Modbus options menu
 */


static const struct s_menu_item sub_menu_serial_modbus_tcpport = { .title = "PORT",   .sub_menu = NULL, .edit_fn = edit_modbus_tcpport,  .to_string_fn = to_string_modbus_tcpport };
static const struct s_menu_item sub_menu_serial_modbus_endian =  { .title = "ENDIAN", .sub_menu = NULL, .edit_fn = edit_modbus_endian,   .to_string_fn = to_string_modbus_endian };


static const struct s_menu_item *const menu_serial_modbusopt_list[] =
{
  &sub_menu_serial_modbus_tcpport,
  &sub_menu_serial_modbus_endian,
  NULL
};


const struct s_menu_item sub_menu_serial_netopt =    { .title = "NET.OPT",  .sub_menu = menu_serial_netopt_list };
const struct s_menu_item sub_menu_serial_autoopt =   { .title = "AUT.OPT",  .sub_menu = menu_serial_autoopt_list };
const struct s_menu_item sub_menu_serial_printopt =  { .title = "PRN.OPT",  .sub_menu = menu_serial_prnopt_list };
const struct s_menu_item sub_menu_serial_modbusopt = { .title = "MOD.OPT",  .sub_menu = menu_serial_modbusopt_list };
