/*
 * iface_serial_ports.c
 *
 *  Created on: 03/09/2016
 *      Author: ian
 */
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <rinblock/rinblock.h>
#include "app_db_user.h"
#include "indicator_model.h"
#include "iface_sd.h"
#include "editor.h"
#include "indicator_model_features.h"

#include "iface_serial_ports.h"


const char *const baud_setting[NUM_DB_BAUD_SETTINGS] = DB_BAUD_STRINGS;

static const char cs_parity[] = "NOE";
static const char cs_data[] = "78";
static const char cs_stop[] = "12";
static const char cs_term[] = "-T";
static const char cs_rs24[] = "42";
static const char cs_dtr[] = "-D";
static const char cs_disabled[] = "-";


enum e_comms_bits_settings
{
  cbs_dtr,
  cbs_rs24,
  cbs_terms,
  cbs_stop,
  cbs_data,
  cbs_parity,

  cbs_num
};


/*                                                                cbs_dtr,     cbs_rs232_rs422, cbs_terms,   cbs_stop,    cbs_data,    cbs_parity */
/** Front rinlink has no settings */
static const char *const comms_settings_ser_rl_front[cbs_num] = { cs_disabled, cs_disabled,     cs_disabled, cs_disabled, cs_disabled, cs_disabled };
static const bool comms_settings_ser_rl_front_has[cbs_num] =    { false,       false,           false,       false,       false,       false       };

/** Serial 1 rear rinlink has no settings */
static const char *const comms_settings_ser_rl_rear[cbs_num] =  { cs_disabled, cs_disabled,     cs_disabled, cs_disabled, cs_disabled, cs_disabled };
static const bool comms_settings_ser_rl_rear_has[cbs_num] =     { false,       false,           false,       false,       false,       false       };

/** Serial 1 DB9 has all settings, except DTR */
static const char *const comms_settings_ser1_db9[cbs_num] =     { cs_disabled, cs_rs24,         cs_term,     cs_stop,     cs_data,     cs_parity };
static const bool comms_settings_ser1_db9_has[cbs_num] =        { false,       true,            true,        true,        true,        true      };

/** Serial 2 DB9 has all settings except 232/422 switch and terminations */
static const char *const comms_settings_ser2_db9[cbs_num] =     { cs_dtr,      cs_disabled,     cs_disabled, cs_stop,     cs_data,     cs_parity };
static const bool comms_settings_ser2_db9_has[cbs_num] =        { true,        false,           false,       true,        true,        true      };

/** An RS232 module channel A DTR and data bits */
static const char *const comms_settings_ser_m_rs232a[cbs_num] = { cs_dtr,      cs_disabled,     cs_disabled, cs_stop,     cs_data,     cs_parity };
static const bool comms_settings_ser_m_rs232a_has[cbs_num] =    { true,        false,           false,       true,        true,        true      };

/** An RS232 module channel B has nothing, because tied to A channel */
static const char *const comms_settings_ser_m_rs232b[cbs_num] = { cs_disabled, cs_disabled,     cs_disabled, cs_disabled, cs_disabled, cs_disabled };
static const bool comms_settings_ser_m_rs232b_has[cbs_num] =    { false,       false,           false,       false,       false,       false       };

/** An RS485 module channel A with termination resistors */
static const char *const comms_settings_ser_m_rs485a[cbs_num] = { cs_disabled, cs_disabled,     cs_term,     cs_stop,     cs_data,     cs_parity };
static const bool comms_settings_ser_m_rs485a_has[cbs_num] =    { false,       false,           true,        true,        true,        true      };

/** An RS485 module channel B has only termination resistors, because tied to A channel */
static const char *const comms_settings_ser_m_rs485b[cbs_num] = { cs_disabled, cs_disabled,     cs_term,     cs_disabled, cs_disabled, cs_disabled };
static const bool comms_settings_ser_m_rs485b_has[cbs_num] =    { false,       false,           true,        false,       false,       false       };

/** A USB host serial port has most settings, except 232/422 switch and terminations */
static const char *const comms_settings_ser_usb_host[cbs_num] = { cs_dtr,      cs_disabled,     cs_disabled, cs_stop,     cs_data,     cs_parity };
static const bool comms_settings_ser_usb_host_has[cbs_num] =    { true,        false,           false,       true,        true,        true      };


static const uint8_t comms_settings_mask[cbs_num] =   { DB_SERIAL_USECTS_MASK, DB_SERIAL_USERS232_MASK, DB_SERIAL_USETERM_MASK, DB_SERIAL_STOP_MASK,   DB_SERIAL_DATA_MASK,   DB_SERIAL_PARITY_MASK };
static const uint8_t comms_settings_shifts[cbs_num] = { DB_SERIAL_CTS_SHIFTS,  DB_SERIAL_RS232_SHIFTS,  DB_SERIAL_TERM_SHIFTS,  DB_SERIAL_STOP_SHIFTS, DB_SERIAL_DATA_SHIFTS, DB_SERIAL_PARITY_SHIFTS };

struct s_comms_set
{
  uint8_t *baudrate_idx;
  uint8_t *porttype_idx;
  const uint8_t *porttype_max;
  uint8_t *bits;
  const char *const *bits_sets;
  const bool *bits_has;
};


/** Front rinlink has these settings */
static const struct s_comms_set comms_set_rl_front =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_RL_FRONT],
  .porttype_idx = &db_app_user.serial_type[SERPORT_RL_FRONT],
  .porttype_max = &serial_type_maximums[SERPORT_RL_FRONT],
  .bits = &db_app_user.serial_bit_settings[SERPORT_RL_FRONT],
  .bits_sets = comms_settings_ser_rl_front,
  .bits_has = comms_settings_ser_rl_front_has,
};


/** Rear rinlink has these settings */
static const struct s_comms_set comms_set_rl_rear =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_SER1_RL_REAR],
  .porttype_idx = &db_app_user.serial_type[SERPORT_SER1_RL_REAR],
  .porttype_max = &serial_type_maximums[SERPORT_SER1_RL_REAR],
  .bits = &db_app_user.serial_bit_settings[SERPORT_SER1_RL_REAR],
  .bits_sets = comms_settings_ser_rl_rear,
  .bits_has = comms_settings_ser_rl_rear_has,
};


/** Serial 1 DB9 has these settings */
static const struct s_comms_set comms_set_serial1 =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_SER1_RL_REAR],
  .porttype_idx = &db_app_user.serial_type[SERPORT_SER1_RL_REAR],
  .porttype_max = &serial_type_maximums[SERPORT_SER1_RL_REAR],
  .bits = &db_app_user.serial_bit_settings[SERPORT_SER1_RL_REAR],
  .bits_sets = comms_settings_ser1_db9,
  .bits_has = comms_settings_ser1_db9_has,
};


/** Serial 2 DB9 has these settings */
static const struct s_comms_set comms_set_serial2 =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_SER2],
  .porttype_idx = &db_app_user.serial_type[SERPORT_SER2],
  .porttype_max = &serial_type_maximums[SERPORT_SER2],
  .bits = &db_app_user.serial_bit_settings[SERPORT_SER2],
  .bits_sets = comms_settings_ser2_db9,
  .bits_has = comms_settings_ser2_db9_has,
};


/** Module 1, channel a, RS232 */
static const struct s_comms_set comms_set_m_1a_rs232 =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_RB1A],
  .porttype_idx = &db_app_user.serial_type[SERPORT_RB1A],
  .porttype_max = &serial_type_maximums[SERPORT_RB1A],
  .bits = &db_app_user.serial_bit_settings[SERPORT_RB1A],
  .bits_sets = comms_settings_ser_m_rs232a,
  .bits_has = comms_settings_ser_m_rs232a_has,
};


/** Module 1, channel a, RS485 */
static const struct s_comms_set comms_set_m_1a_rs485 =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_RB1A],
  .porttype_idx = &db_app_user.serial_type[SERPORT_RB1A],
  .porttype_max = &serial_type_maximums[SERPORT_RB1A],
  .bits = &db_app_user.serial_bit_settings[SERPORT_RB1A],
  .bits_sets = comms_settings_ser_m_rs485a,
  .bits_has = comms_settings_ser_m_rs485a_has,
};


/** Module 1, channel b, RS232 */
static const struct s_comms_set comms_set_m_1b_rs232 =
{
  .baudrate_idx = NULL,
  .porttype_idx = &db_app_user.serial_type[SERPORT_RB1B],
  .porttype_max = &serial_type_maximums[SERPORT_RB1B],
  .bits = NULL,
  .bits_sets = NULL,
  .bits_has = NULL,
};


/** Module 1, channel b, RS485 */
static const struct s_comms_set comms_set_m_1b_rs485 =
{
  .baudrate_idx = NULL,
  .porttype_idx = &db_app_user.serial_type[SERPORT_RB1B],
  .porttype_max = &serial_type_maximums[SERPORT_RB1B],
  .bits = &db_app_user.serial_bit_settings[SERPORT_RB1B],
  .bits_sets = comms_settings_ser_m_rs485b,
  .bits_has = comms_settings_ser_m_rs485b_has,
};

/** Module 2, channel a, RS232 */
static const struct s_comms_set comms_set_m_2a_rs232 =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_RB2A],
  .porttype_idx = &db_app_user.serial_type[SERPORT_RB2A],
  .porttype_max = &serial_type_maximums[SERPORT_RB2A],
  .bits = &db_app_user.serial_bit_settings[SERPORT_RB2A],
  .bits_sets = comms_settings_ser_m_rs232a,
  .bits_has = comms_settings_ser_m_rs232a_has,
};


/** Module 2, channel a, RS485 */
static const struct s_comms_set comms_set_m_2a_rs485 =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_RB2A],
  .porttype_idx = &db_app_user.serial_type[SERPORT_RB2A],
  .porttype_max = &serial_type_maximums[SERPORT_RB2A],
  .bits = &db_app_user.serial_bit_settings[SERPORT_RB2A],
  .bits_sets = comms_settings_ser_m_rs485a,
  .bits_has = comms_settings_ser_m_rs485a_has,
};


/** Module 2, channel b, RS232 */
static const struct s_comms_set comms_set_m_2b_rs232 =
{
  .baudrate_idx = NULL,
  .porttype_idx = &db_app_user.serial_type[SERPORT_RB2B],
  .porttype_max = &serial_type_maximums[SERPORT_RB2B],
  .bits = NULL,
  .bits_sets = NULL,
  .bits_has = NULL,
};


/** Module 2, channel b, RS485 */
static const struct s_comms_set comms_set_m_2b_rs485 =
{
  .baudrate_idx = NULL,
  .porttype_idx = &db_app_user.serial_type[SERPORT_RB2B],
  .porttype_max = &serial_type_maximums[SERPORT_RB2B],
  .bits = &db_app_user.serial_bit_settings[SERPORT_RB2B],
  .bits_sets = comms_settings_ser_m_rs485b,
  .bits_has = comms_settings_ser_m_rs485b_has,
};


/** USB gadget serial port */
static const struct s_comms_set comms_set_usb_gadget =
{
  .baudrate_idx = NULL,
  .porttype_idx = &db_app_user.serial_type[SERPORT_USBGADGET],
  .porttype_max = &serial_type_maximums[SERPORT_USBGADGET],
  .bits = NULL,
  .bits_sets = NULL,
  .bits_has = NULL,
};


/** USB host serial port */
static const struct s_comms_set comms_set_usb_host =
{
  .baudrate_idx = &db_app_user.serial_baud_idx[SERPORT_USBHOST],
  .porttype_idx = &db_app_user.serial_type[SERPORT_USBHOST],
  .porttype_max = &serial_type_maximums[SERPORT_USBHOST],
  .bits = &db_app_user.serial_bit_settings[SERPORT_USBHOST],
  .bits_sets = comms_settings_ser_usb_host,
  .bits_has = comms_settings_ser_usb_host_has,
};


/** Serial port type setting (auto out, net, etc) */
const char *const ser_type_setting[NUM_DB_SERIAL_TYPES] = DB_SERIAL_TYPE_STRINGS;


/** Fill out the comms settings structure based on what hardware is attached */
static void comms_configure_structure(const struct s_comms_set *comms_settings[NUM_DB_SERIAL_SETTINGS])
{
  size_t i;
  const enum e_indicator_model_sub_type imst = get_indicator_model_subtype();

  /* Set all to NULL */
  for (i = 0; i < NUM_DB_SERIAL_SETTINGS; i++)
  {
    comms_settings[i] = NULL;
  }

  /* Fill out based on what's attached */
  comms_settings[SERPORT_RL_FRONT] = (indicator_model_subtype_has_front_rinlink(imst)) ? &comms_set_rl_front : NULL;
  if (indicator_model_subtype_has_db9_serial(imst))
  {
    comms_settings[SERPORT_SER1_RL_REAR] = &comms_set_serial1;
    comms_settings[SERPORT_SER2] = &comms_set_serial2;
  }
  else
  {
    comms_settings[SERPORT_SER1_RL_REAR] = (indicator_model_subtype_has_rear_rinlink(imst)) ? &comms_set_rl_rear : NULL;
    comms_settings[SERPORT_SER2] = NULL;
  }

  if (rinblocks[2].running && RB_IFACE_TYPE_SERIAL == rinblocks[2].tag.interface_type)
  {
    /* Serial rinblock in slot 1, what is it */
    switch (rinblocks[2].tag.model_number)
    {
      case RB_MODEL_NUMBER_232_232:
        comms_settings[SERPORT_RB1A] = &comms_set_m_1a_rs232;
        comms_settings[SERPORT_RB1B] = &comms_set_m_1b_rs232;
        break;
      case RB_MODEL_NUMBER_232_485:
        comms_settings[SERPORT_RB1A] = &comms_set_m_1a_rs232;
        comms_settings[SERPORT_RB1B] = &comms_set_m_1b_rs485;
        break;
      case RB_MODEL_NUMBER_485_485:
        comms_settings[SERPORT_RB1A] = &comms_set_m_1a_rs485;
        comms_settings[SERPORT_RB1B] = &comms_set_m_1b_rs485;
        break;
      default:
        comms_settings[SERPORT_RB1A] = NULL;
        comms_settings[SERPORT_RB1B] = NULL;
        break;
    }
  }

  if (rinblocks[3].running && RB_IFACE_TYPE_SERIAL == rinblocks[3].tag.interface_type)
  {
    /* Serial rinblock in slot 1, what is it */
    switch (rinblocks[3].tag.model_number)
    {
      case RB_MODEL_NUMBER_232_232:
        comms_settings[SERPORT_RB2A] = &comms_set_m_2a_rs232;
        comms_settings[SERPORT_RB2B] = &comms_set_m_2b_rs232;
        break;
      case RB_MODEL_NUMBER_232_485:
        comms_settings[SERPORT_RB2A] = &comms_set_m_2a_rs232;
        comms_settings[SERPORT_RB2B] = &comms_set_m_2b_rs485;
        break;
      case RB_MODEL_NUMBER_485_485:
        comms_settings[SERPORT_RB2A] = &comms_set_m_2a_rs485;
        comms_settings[SERPORT_RB2B] = &comms_set_m_2b_rs485;
        break;
      default:
        comms_settings[SERPORT_RB2A] = NULL;
        comms_settings[SERPORT_RB2B] = NULL;
        break;
    }
  }

  comms_settings[SERPORT_USBGADGET] = &comms_set_usb_gadget;
  comms_settings[SERPORT_USBHOST] = &comms_set_usb_host; /* TODO hide if serial port not attached */
}


/** Convert serial port bits to a string */
static void serport_bits_to_string(enum eSerialPort port, char bits_str[cbs_num + 1], bool reverse_order)
{
  uint8_t bits;
  size_t i;
  size_t value;
  const bool *has;
  const struct s_comms_set *comms_settings[NUM_DB_SERIAL_SETTINGS];
  const struct s_comms_set *comms_setting_for_port;

  /* Work out what ports are where */
  comms_configure_structure(comms_settings);

  /* */
  if (port >= NUM_DB_SERIAL_SETTINGS || NULL == comms_settings[port])
  {
    memset(bits_str, '\0', cbs_num);
    return;
  }

  comms_setting_for_port = comms_settings[port];
  has = comms_setting_for_port->bits_has;

  /* Expand packed bits into array for editing */
  bits = *comms_setting_for_port->bits;
  for (i = 0; i < cbs_num; i++)
  {
    value = (has[i]) ? ((bits & comms_settings_mask[i]) >> comms_settings_shifts[i]) : 0;
    if (!reverse_order)
      bits_str[i] = comms_setting_for_port->bits_sets[i][value];
    else
      bits_str[cbs_num - i - 1] = comms_setting_for_port->bits_sets[i][value];
  }
  bits_str[cbs_num] = '\0';
}


/** Display error if serial port is not available */
void comms_error_setting_not_avail(void)
{
  error_strings("NOT", "AVAIL");
}


static enum eMenuLevel edit_serport_bits(bool full_edit, void *user_data, size_t index)
{
  enum eSerialPort port = (enum eSerialPort) user_data;
  const struct s_comms_set *comms_settings[NUM_DB_SERIAL_SETTINGS];
  enum eEditResult result;
  uint8_t bits;
  size_t i;
  uint8_t unpacked_bits[cbs_num];
  const bool *has;

  comms_configure_structure(comms_settings);

  if (NULL == comms_settings[port] || NULL == comms_settings[port]->bits ||
      NULL == comms_settings[port]->bits_has || NULL == comms_settings[port]->bits_sets)
  {
    comms_error_setting_not_avail();
    result = EDIT_ABORT;
  }
  else
  {
    /* Expand packed bits into array for editing */
    has = comms_settings[port]->bits_has;
    bits = *comms_settings[port]->bits;
    for (i = 0; i < cbs_num; i++)
    {
      unpacked_bits[i] = (has[i]) ? ((bits & comms_settings_mask[i]) >> comms_settings_shifts[i]) : 0;
    }

    /* Edit bits */
    result = edit_bits(unpacked_bits, cbs_num, comms_settings[port]->bits_sets, ALLOW_SAFE);
    if (EDIT_CHG == result || EDIT_OK == result)
    {

      /* Pack back into byte */
      bits = 0;
      for (i = 0; i < cbs_num; i++)
      {
        if (has[i])
        {
          bits |= unpacked_bits[i] << comms_settings_shifts[i];
        }
      }
      *comms_settings[port]->bits = bits;
    }
  }
  return edit_result_to_menu_level(result);
}


/** Convert the serial port bits settings to a string */
static void to_string_serport_bits(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  enum eSerialPort port = (enum eSerialPort) user_data;
  char bits_str[cbs_num + 1];

  serport_bits_to_string(port, bits_str, true);
  snprintf(buffer, buflen, "%s", bits_str);
}


/** Edit serial port type - this version allows all ports to be whatever they want */
static enum eMenuLevel edit_serport_type(bool full_edit, void *user_data, size_t index)
{
  enum eSerialPort port = (enum eSerialPort) user_data;
  const struct s_comms_set *comms_settings[NUM_DB_SERIAL_SETTINGS];
  enum eEditResult result;

  comms_configure_structure(comms_settings);

  if (NULL == comms_settings[port] || NULL == comms_settings[port]->porttype_idx)
  {
    comms_error_setting_not_avail();
    result = EDIT_ABORT;
  }
  else
  {
    result = edit_register_data(comms_settings[port]->porttype_idx, 0, *comms_settings[port]->porttype_max, ser_type_setting, ALLOW_SAFE);
  }
  return edit_result_to_menu_level(result);
}


/** Convert the serial port type to a string */
static void to_string_serport_type(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  enum eSerialPort port = (enum eSerialPort) user_data;
  const struct s_comms_set *comms_settings[NUM_DB_SERIAL_SETTINGS];

  comms_configure_structure(comms_settings);
  if (NULL == comms_settings[port] || NULL == comms_settings[port]->porttype_idx)
  {
    snprintf(buffer, buflen, "Port not available");
  }
  else
  {
    snprintf(buffer, buflen, "%s", ser_type_setting[*comms_settings[port]->porttype_idx]);
  }
}


/** Edit serial port baudrate */
static enum eMenuLevel edit_serport_baudrate(bool full_edit, void *user_data, size_t index)
{
  enum eSerialPort port = (enum eSerialPort) user_data;
  const struct s_comms_set *comms_settings[NUM_DB_SERIAL_SETTINGS];
  enum eEditResult result;

  comms_configure_structure(comms_settings);

  if (NULL == comms_settings[port] || NULL == comms_settings[port]->baudrate_idx)
  {
    comms_error_setting_not_avail();
    result = EDIT_ABORT;
  }
  else
  {
    result = edit_register_data(comms_settings[port]->baudrate_idx, 0, NUM_DB_BAUD_SETTINGS - 1, baud_setting, ALLOW_SAFE);
  }
  return edit_result_to_menu_level(result);
}


/** Convert the serial port baudrate to a string */
static void to_string_serport_baudrate(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  enum eSerialPort port = (enum eSerialPort) user_data;
  const struct s_comms_set *comms_settings[NUM_DB_SERIAL_SETTINGS];

  comms_configure_structure(comms_settings);
  if (NULL == comms_settings[port] || NULL == comms_settings[port]->baudrate_idx)
  {
    snprintf(buffer, buflen, "Port not available");
  }
  else
  {
    snprintf(buffer, buflen, "%s", baud_setting[*comms_settings[port]->baudrate_idx]);
  }
}


/** Check if a port is present */
bool serial_menu_check_port_present(enum eSerialPort port)
{
  bool is_present;
  const enum e_indicator_model_sub_type imst = get_indicator_model_subtype();

  switch (port)
  {
    /* No settings for front rinlink */
    case SERPORT_RL_FRONT:
      /* is_present = indicator_model_subtype_has_front_rinlink(imst); */
      is_present = false;
      break;

    /* Present if the rinblock is present & running & serial */
    case SERPORT_RB1A:
    case SERPORT_RB1B:
      is_present = (rinblocks[2].running && RB_IFACE_TYPE_SERIAL == rinblocks[2].tag.interface_type);
      break;
    case SERPORT_RB2A:
    case SERPORT_RB2B:
      is_present = (rinblocks[3].running && RB_IFACE_TYPE_SERIAL == rinblocks[3].tag.interface_type);
      break;

    /* Always present */
    case SERPORT_USBGADGET:
      is_present = true;
      break;

    /* Present based on indicator model features */
    case SERPORT_SER1_RL_REAR:
    case SERPORT_SER2:
      is_present = indicator_model_subtype_has_db9_serial(imst);
      break;

    /* TODO Has this been connected */
    case SERPORT_USBHOST:
      is_present = true;
      break;

    default:
      is_present = false;
      break;
  }

  return is_present;
}


/*
 * Serial port configuration
 */


static const struct s_menu_item sub_menu_serial_port_type = { .title = "TYPE", .sub_menu = NULL, .edit_fn = edit_serport_type,     .to_string_fn = to_string_serport_type,     .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_serial_port_baud = { .title = "BAUD", .sub_menu = NULL, .edit_fn = edit_serport_baudrate, .to_string_fn = to_string_serport_baudrate, .user_data_from_menu_above = true };
static const struct s_menu_item sub_menu_serial_port_bits = { .title = "BITS", .sub_menu = NULL, .edit_fn = edit_serport_bits,     .to_string_fn = to_string_serport_bits,     .user_data_from_menu_above = true };


static const struct s_menu_item *const menu_serial_portoptions_list[] =
{
  &sub_menu_serial_port_type,
  &sub_menu_serial_port_baud,
  &sub_menu_serial_port_bits,
  NULL
};


/*
 * Rinlink serial ports are fixed type/baud/bits
 */
/*
const struct s_menu_item sub_menu_serial_rlfront =   { .title = "SER.RL.F", .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_SER1_RL_REAR };
const struct s_menu_item sub_menu_serial_rlrear =    { .title = "SER.RL.R", .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_RL_REAR      };
*/
const struct s_menu_item sub_menu_serial_ser1db9 =   { .title = "SER.1",    .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_SER1_RL_REAR };
const struct s_menu_item sub_menu_serial_ser2db9 =   { .title = "SER.2",    .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_SER2         };
const struct s_menu_item sub_menu_serial_mod1a =     { .title = "SER.M.1A", .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_RB1A         };
const struct s_menu_item sub_menu_serial_mod1b =     { .title = "SER.M.1B", .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_RB1B         };
const struct s_menu_item sub_menu_serial_mod2a =     { .title = "SER.M.2A", .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_RB2A         };
const struct s_menu_item sub_menu_serial_mod2b =     { .title = "SER.M.2B", .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_RB2B         };
const struct s_menu_item sub_menu_serial_usbslave =  { .title = "SER.SLV",  .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_USBGADGET    };
const struct s_menu_item sub_menu_serial_usbhost =   { .title = "SER.HST",  .sub_menu = menu_serial_portoptions_list, .menu_enter_exit = NULL, .user_data = (void *) SERPORT_USBHOST      };
