/*
 * iface_test.c
 *
 * Test menu
 */
#include <stdio.h>
#include <string.h>
#include <rinhw/ucos_ii.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcinterface.h>
#include "database.h"
#include "meas.h"
#include "display.h"
#include "kbd_5000.h"
#include "editor.h"
#include "system.h"
#include "iface.h"
#include "access.h"
#include "buzz5000.h"
#include "iface_sd.h"
#include "adcthread.h"
#include "commonstrings.h"
#include "dbprint.h"

#include "iface_test.h"


/** Number of IO pins that can be tested at once */
#define MAX_IO_TEST 4


/** Display the millivolt per volt reading */
static enum eMenuLevel test_display_mvv(bool full_edit, void *user_data, size_t index)
{
  uint8_t kbd;
  AdcRdg rdg;
  uint32_t finish_time;
  bool timeout;
  enum eMenuLevel result;

  /* Exit after 5 seconds in trade mode */
  finish_time = OSTimeGet() + 5 * OS_TICKS_PER_SEC;

  /* Display mV/V and wait for timeout, or keypress */
  memset(&rdg, 0, sizeof(rdg));
  rdg.info.unitdp.bit.dp = 4;
  do
  {
    rdg.rdg = get_abs_mv_v();
    Dshow_rdg_info(&rdg, 0, false, 0);

    kbd = get_key();
    timeout = (ADC_APPROVAL_MODE_NONE != adcSuperStructure.adcUserDatabase.approvalMode && OSTimeGet() >= finish_time);
  }
  while (!timeout && !quit_setup_menus() && (kbd == IDLE_CHG || kbd == NO_KEY));

  /* Exit menus? */
  switch (kbd)
  {
    case KEY_MENU_EXIT1:
    case KEY_MENU_EXIT2:
      result = mlQuit;
      break;
    default:
      result = mlNum;
      break;
  }
  return result;
}


static void test_io_show_string(uint8_t data, int first, int forced)
{
  int i;
  uint8_t mask;
  char str[MAX_IO_TEST + 2]; /* Null terminator and . when forced */
  char *ptr;

  ptr = str;
  mask = 0x01 << first;
  for (i = 0; i < MAX_IO_TEST; i++, mask <<= 1)
  {
    *ptr++ = (data & mask) ? (char) ('1' + first + i) : '-';
    if (i == forced)
    {
      /* Show a dot on a IO pin that is a forced output */
      *ptr++ = '.';
    }
  }
  *ptr = '\0';
  Dputstr(str);
}


static enum eMenuLevel test_io(bool full_edit, void *user_data, size_t index)
{
  intptr_t first = (intptr_t) user_data;
  bool finished;
  uint8_t i, mask;
  uint8_t data = access_get_inputs();
  uint8_t olddata = ~data;
  uint8_t kbd;
  uint8_t temp;
  bool active;
  int pos;
  int forced;
  int oldforced;
  enum eMenuLevel result;

  result = mlNum;
  active = remote_keys_enabled();
  enable_remote_keys(false);

  /* Is there IO present on the device */
  if (!access_io_is_present())
  {
    error_strings(no_str, "I.O.");
  }
  else
  {
    pos = 0;
    forced = -1;
    oldforced = ~forced;
    Dflash_digits(DFLASHMASK_D0 << (NUM_DIGITS - 1 - pos));
    finished = false;
    while (!quit_setup_menus() && !finished)
    {
      /* Display on change */
      if (data != olddata || forced != oldforced)
      {
        test_io_show_string(data, first, forced);
        Dset_annunciators(map_io_to_annunciators(data));
        olddata = data;
        oldforced = forced;
      }

      kbd = get_key();
      switch (kbd)
      {
        case KEY_MENU_EXIT1:
        case KEY_MENU_EXIT2:
          result = mlQuit;
          finished = true;
          break;

        case KEY_EDIT_OK:
        case KEY_EDIT_CANCEL1:
        case KEY_EDIT_CANCEL2:
          finished = true;
          break;

        /* Select position right */
        case KEY_EDIT_RIGHT:
          forced = -1;
          access_configure_iopin(pos + first, ioInput);
          pos++;
          if (pos >= MAX_IO_TEST)
          {
            pos = 0;
          }
          Dflash_digits(DFLASHMASK_D0 << (NUM_DIGITS - 1 - pos));
          break;

        /* Select position left */
        case KEY_EDIT_LEFT:
          forced = -1;
          access_configure_iopin(pos + first, ioInput);
          if (pos > 0)
            pos--;
          else
            pos = MAX_IO_TEST - 1;
          Dflash_digits(DFLASHMASK_D0 << (NUM_DIGITS - 1 - pos));
          break;

        /* Force output */
        case KEY_EDIT_UP:
        case KEY_EDIT_DOWN:
          access_configure_iopin(pos + first, ioOutput);
          mask = 0x01 << (pos + first);
          i = access_get_outputs();
          temp = (i & ~mask) | (~i & mask);
          access_set_outputs(temp);
          forced = pos;
          break;
      }
      data = access_get_inputs();
    }

    /* Configure IO back to how it is in the database */
    access_dbconfigure_allio();
  }

  if (active)
    enable_remote_keys(true);
  Dflash_digits(DFLASHMASK_NONE);
  return result;
}


static enum eMenuLevel show_oload(bool full_edit, void *user_data, size_t index)
{
  uint8_t kbd;
  char str[NUM_DIGITS + 1];
  enum eMenuLevel result;

  do
  {
    snprintf(str, NUM_DIGITS + 1, "%*ld", NUM_DIGITS, (long) AdcGetOverloadCounter(&adcSuperStructure, true));
    Dputstr(str);

    kbd = get_key();
    if (KEY_EDIT_OK == kbd)
    {
      kbd = NO_KEY;
      if (!full_edit)
      {
        beep_long();
        access_denied();
      }
      else if (ask("CLR ", false))
      {
        AdcClearOverloadCounter(&adcSuperStructure, true);
      }
    }
  }
  while (!quit_setup_menus() && (kbd == IDLE_CHG || kbd == NO_KEY));

  switch (kbd)
  {
    case KEY_MENU_EXIT1:
    case KEY_MENU_EXIT2:
      result = mlQuit;
      break;
    default:
      result = mlNum;
      break;
  }
  return result;
}


/** Convert the hires setting to a string */
static void to_string_oload(bool full, void *user_data, size_t index, char *buffer, size_t buflen)
{
  snprintf(buffer, buflen, "%lu", (unsigned long) AdcGetOverloadCounter(&adcSuperStructure, true));
}


static const struct s_menu_item sub_menu_test_scale =    { .title = "SCALE",  .sub_menu = NULL, .edit_fn = test_display_mvv, .to_string_fn = to_string_no_setting };
static const struct s_menu_item sub_menu_test_io14 =     { .title = "IO-1.4", .sub_menu = NULL, .edit_fn = test_io,          .to_string_fn = to_string_no_setting, .user_data = (void *) 0 };
static const struct s_menu_item sub_menu_test_io58 =     { .title = "IO-5.8", .sub_menu = NULL, .edit_fn = test_io,          .to_string_fn = to_string_no_setting, .user_data = (void *) 4 };
static const struct s_menu_item sub_menu_test_overload = { .title = "O-LOAD", .sub_menu = NULL, .edit_fn = show_oload,       .to_string_fn = to_string_oload };


static const struct s_menu_item *const menu_test_list[] =
{
  &sub_menu_test_scale,
  &sub_menu_test_io14,
  &sub_menu_test_io58,
  &sub_menu_test_overload,
  NULL
};


/** Test menu item */
const struct s_menu_item menu_test =
{
  .title = "TEST",
  .sub_menu = menu_test_list
};
