/*
 * iface_usb.c
 *
 *  Created on: Mar 31, 2014
 *      Author: ian
 */
#include <rinhw/kbd_linux_event.h>
#include "display.h"

#include "iface_usb.h"


static pthread_mutex_t iface_usb_mutex = PTHREAD_MUTEX_INITIALIZER;


/** USB add & remove messages */
enum eDisplayUSBDevice {                                dudNone, dudKbdAdd, dudKbdRemove, dudKbdError, dudDiskAdd, dudDiskRemove, dudDiskError, dudPrintAdd, dudPrintRemove, dudPrintError, dudSerialAdd, dudSerialRemove, dudSerialError, dudNum };
static const char *const show_usb_messages1[dudNum] = { "",      "USB.KBD", "USB.KBD",    "USB.KBD",   "USB.DSK",  "USB.DSK",     "USB.DSK",    "USB.PRT",   "USB.PRT",      "USB.PRT",     "USB.SER",    "USB.SER",       "USB.SER" };
static const char *const show_usb_messages2[dudNum] = { "",      "ADDED",   "REMOVE",     "ERROR",     "ADDED",    "REMOVE",      "ERROR",      "ADDED",     "REMOVE",       "ERROR",       "ADDED",      "REMOVE",        "ERROR"   };


#define USB_NOTIFICATION_QUEUE_SIZE 10

/** List of USB notification events */
static enum eDisplayUSBDevice usb_notification_queue[USB_NOTIFICATION_QUEUE_SIZE];
static size_t usb_notification_queue_head = 0;
static size_t usb_notification_queue_tail = 0;
static size_t usb_notification_count = 0;

static bool usb_notification_enable = true;

#define USB_SHOW_COUNT_PRESET 20

static int show_usb_count = 0;
static enum eDisplayUSBDevice current_usb_notification = dudNone;


static enum eDisplayUSBDevice iface_usb_get_notification(void)
{
  enum eDisplayUSBDevice result;

  pthread_mutex_lock(&iface_usb_mutex);
  if (0 == usb_notification_count)
  {
    result = dudNone;
  }
  else
  {
    result = usb_notification_queue[usb_notification_queue_tail];
    usb_notification_queue_tail++;
    if (usb_notification_queue_tail >= USB_NOTIFICATION_QUEUE_SIZE)
      usb_notification_queue_tail = 0;
    usb_notification_count--;
  }
  pthread_mutex_unlock(&iface_usb_mutex);

  return result;
}


static void iface_usb_add_notification(enum eDisplayUSBDevice event)
{
  pthread_mutex_lock(&iface_usb_mutex);
  if (usb_notification_count < USB_SHOW_COUNT_PRESET && usb_notification_enable == true)
  {
    usb_notification_queue[usb_notification_queue_head] = event;
    usb_notification_queue_head++;
    if (usb_notification_queue_head >= USB_NOTIFICATION_QUEUE_SIZE)
      usb_notification_queue_head = 0;
    usb_notification_count++;
  }
  pthread_mutex_unlock(&iface_usb_mutex);
}


bool iface_usb_show_notification(void)
{
  bool is_showing;

  if (dudNone == current_usb_notification)
  {
    current_usb_notification = iface_usb_get_notification();
    if (dudNone != current_usb_notification)
      show_usb_count = USB_SHOW_COUNT_PRESET;
  }

  if (show_usb_count > USB_SHOW_COUNT_PRESET / 2)
  {
    Dputstr_nofl(show_usb_messages1[current_usb_notification]);
    Dset_annunciators(ANNUN_NONE);
    Dset_units_ignore_motion(ADC_UNITS_NONE);
  }
  else if (show_usb_count > 0)
  {
    Dputstr_nofl(show_usb_messages2[current_usb_notification]);
    Dset_annunciators(ANNUN_NONE);
    Dset_units_ignore_motion(ADC_UNITS_NONE);
  }

  if (show_usb_count > 0)
  {
    show_usb_count--;
    is_showing = true;
  }
  else
  {
    current_usb_notification = dudNone;
    is_showing = false;
  }

  return is_showing;
}


void iface_usb_event_callback(void *client_callback_data, enum eDeviceMounterAction action, int result)
{
  enum eDisplayUSBDevice showdev;

  /*
   * When we get a keyboard without the keys we need, we get a RIO_ERROR_EVKBD_KEYS_UNSUPPORTED
   * error. Don't display an error in this case, 'cause it just causes confusion.
   *
   * Usually a keyboard will be split into a media keyboard and a normal keyboard. These
   * appear as two independent eventX devices under Linux. We used to get
   * USB.KBD Added (event0) followed by USB.KBD Error, which was just confusing.
   *
   * It will end up in syslog, so someone can find out that way if they want
   * to know why their media keyboard doesn't work
   */
  if (RIO_ERROR_EVKBD_KEYS_UNSUPPORTED != result)
  {
    showdev = (RIO_SUCCESS != result) ? dudKbdError : (dmaAdd == action) ? dudKbdAdd : dudKbdRemove;
    iface_usb_add_notification(showdev);
  }
}


void iface_usb_partition_callback(enum eDeviceMounterAction action, bool result)
{
  enum eDisplayUSBDevice showdev;

  showdev = (!result) ? dudDiskError : (dmaAdd == action) ? dudDiskAdd : dudDiskRemove;
  iface_usb_add_notification(showdev);
}


void iface_usb_printer_callback(void *client_callback_data, enum eDeviceMounterAction action, bool result)
{
  enum eDisplayUSBDevice showdev;

  showdev = (!result) ? dudPrintError : (dmaAdd == action) ? dudPrintAdd : dudPrintRemove;
  iface_usb_add_notification(showdev);
}


void iface_usb_serial_callback(void *client_callback_data, enum eDeviceMounterAction action, bool result)
{
  enum eDisplayUSBDevice showdev;

  showdev = (!result) ? dudSerialError : (dmaAdd == action) ? dudSerialAdd : dudSerialRemove;
  iface_usb_add_notification(showdev);
}


void iface_usb_set_notifications_status(bool enable)
{
  usb_notification_enable = enable;
}
