/*
 * indicator_model.c
 *
 *  Created on: Feb 11, 2015
 *      Author: ian
 */
#include <stdbool.h>
#include <string.h>
#include "app_db_factory.h"
#include "we_debug.h"
#include "apperror.h"

#include "indicator_model.h"


/** The indicator model subtype */
static enum e_indicator_model indicator_model = im_unknown;


/** The indicator model subtype */
static enum e_indicator_model_sub_type indicator_model_subtype = imst_unknown;


static const char *const indicator_model_strs[im_num] = INDICATOR_MODEL_STRS;


static const char *const indicator_model_short_upper_strs[im_num] = INDICATOR_MODEL_SHORT_UPPER_STRS;


static const char *const indicator_model_short_lower_strs[im_num] = INDICATOR_MODEL_SHORT_LOWER_STRS;


/**
 * Set the indicator model and subtype from the factory database
 *
 * The logic is as follows:
 *  - If valid indicator model matches valid indicator model subtype, then run
 *  - If indicator model is unknown, but valid indicator model subtype, then use subtype to set model and run
 *  - If valid indicator model doesn't match valid indicator model subtype, then don't run (APP_ERR_ROM)
 *  - For indicator models where there is only one subtype (C510, C530, WE2111), and indicator model is valid, but indicator subtype is invalid or unset, then use model to set subtype and run
 *  - For indicator models where there are multiple subtypes (C520), and indicator model is valid, but indicator subtype unset, then this is a upgraded original C520 (rinlink), use model to set subtype and run
 *  - For indicator models where there are multiple subtypes (C520), and indicator model is valid, but indicator subtype unknown, then we can't discern what hardware is fitted, don't run (APP_ERR_ROM)
 *  - If indicator model is unknown, and indicator model subtype is unknown, then don't run (APP_ERR_ROM)
 */
void init_indicator_model_and_subtype(const struct s_db_app_factory *facdb)
{
  enum e_indicator_model im;
  enum e_indicator_model_sub_type imst;
  bool mismatch;

  /* Convert model string into enum e_indicator_model */
  im = string_to_indicator_model(facdb->model_str);

  /* Now check that model_str and hardware variant match */
  imst = (enum e_indicator_model_sub_type) facdb->model_subtype;
  switch (imst)
  {
    case imst_we2111_original:
    case imst_c510_original:
    case imst_c520_rinlink:
    case imst_c520_db9serial:
    case imst_c530_original:
      mismatch = (im != indicator_model_subtype_to_indicator_model(imst));
      if (mismatch)
      {
        rdebug_printf(LOG_ERR, "init_indicator_model_and_subtype: factory database model string (%s) doesn't match indicator model subtype (%s)\n",
                      facdb->model_str, get_indicator_model_subtype_string(imst, imdt_long));
        if (im_unknown == im)
        {
          /*
           * Invalid indicator model, valid indicator model subtype.
           * Use indicator model subtype to set indicator model
           */
          im = indicator_model_subtype_to_indicator_model(imst);
        }
        else
        {
          /* Valid but differing indicator model and indicator model subtype. Don't run */
          set_error(APP_ERR_ROM, "init_indicator_model_and_subtype: setting APP_ERR_ROM because factory database model string doesn't match indicator model subtype\n");
        }
      }
      break;
    case imst_unset:
    case imst_unknown:
    default:
      /* Indicator model subtype is unset or invalid, so try to use indicator model if it is valid */
      switch (im)
      {
        case im_c510:
          imst = imst_c510_original;
          break;
        case im_c520:
          if (imst_unset == imst)
          {
            /* If the model subtype is unset, then this is a old rinlink C520 that has been upgraded */
            imst = imst_c520_rinlink;
            rdebug_printf(LOG_INFO, "init_indicator_model_and_subtype: unset model subtype, but model string is C520, so assuming original rinlink hardware\n");
          }
          else
          {
            /* If the model subtype is invalid, we can't discern what type of C520 this is */
            set_error(APP_ERR_ROM, "init_indicator_model_and_subtype: setting APP_ERR_ROM because cannot discern C520 subtype as indicator model subtype is invalid\n");
          }
          break;
        case im_c530:
          imst = imst_c530_original;
          break;
        case im_we2111:
          imst = imst_we2111_original;
          break;
        case im_unknown:
        default:
          set_error(APP_ERR_ROM, "init_indicator_model_and_subtype: setting APP_ERR_ROM because both indicator model subtype and indicator model are invalid\n");
          break;
      }
      break;
  }
  indicator_model = im;
  indicator_model_subtype = imst;
}


/** Get the indicator model */
enum e_indicator_model get_indicator_model(void)
{
  return indicator_model;
}


/** Get the indicator model subtype */
enum e_indicator_model_sub_type get_indicator_model_subtype(void)
{
  return indicator_model_subtype;
}


/** Convert the indicator model sub type into an indicator model */
enum e_indicator_model indicator_model_subtype_to_indicator_model(enum e_indicator_model_sub_type imst)
{
  enum e_indicator_model im;

  switch (imst)
  {
    case imst_we2111_original:
      im = im_we2111;
      break;

    case imst_c510_original:
      im = im_c510;
      break;

    case imst_c520_rinlink:
    case imst_c520_db9serial:
      im = im_c520;
      break;

    case imst_c530_original:
      im = im_c530;
      break;

    case imst_unknown:
    default:
      im = im_unknown;
      break;
  }
  return im;
}


/** Get a model string for an indicator model */
const char *get_indicator_model_string(enum e_indicator_model im, enum e_indicator_model_description_type imdt)
{
  const char *const *model_str;

  switch (imdt)
  {
    case imdt_short_lower:
      model_str = indicator_model_short_lower_strs;
      break;
    case imdt_short_upper:
      model_str = indicator_model_short_upper_strs;
      break;
    case imdt_long:
    default:
      model_str = indicator_model_strs;
      break;
  }

  return (im < im_num) ? model_str[im] : model_str[im_unknown];
}


/** Get the indicator model for an indicator model string, im_unknown if no match */
enum e_indicator_model string_to_indicator_model(const char *model_str)
{
  enum e_indicator_model im;
  enum e_indicator_model_description_type imdt;

  for (imdt = 0; imdt < imdt_num; imdt++)
  {
    for (im = im_c510; im < im_num; im++)
    {
      if (strcmp(model_str, get_indicator_model_string(im, imdt)) == 0)
      {
        return im;
      }
    }
  }
  return im_unknown;
}


/** Get a model subtype string for an indicator model subtype */
const char *get_indicator_model_subtype_string(enum e_indicator_model_sub_type imst, enum e_indicator_model_description_type imdt)
{
  const char *model_str;
  const enum e_indicator_model im = indicator_model_subtype_to_indicator_model(imst);

  switch (imst)
  {
    case imst_c520_rinlink:
      switch (imdt)
      {
        case imdt_long:
        default:
          model_str = "C520 Original (w/Rinlink) Indicator";
          break;
        case imdt_short_lower:
          model_str = "c520.rl";
          break;
        case imdt_short_upper:
          model_str = "C520.RL";
          break;
      }
      break;

    case imst_c520_db9serial:
      switch (imdt)
      {
        case imdt_long:
        default:
          model_str = "C520 New (w/DB9 serial) Indicator";
          break;
        case imdt_short_lower:
          model_str = "c520.db9";
          break;
        case imdt_short_upper:
          model_str = "C520.DB9";
          break;
      }
      break;

    default:
      model_str = get_indicator_model_string(im, imdt);
      break;
  }
  return model_str;
}


/** Get the indicator model subtype for an indicator model subtype string, imst_unknown if no match */
enum e_indicator_model_sub_type string_to_indicator_model_subtype(const char *model_subtype_str)
{
  #define NUM_IMSTS 5
  enum e_indicator_model_description_type imdt;
  size_t i;
  const enum e_indicator_model_sub_type imsts[NUM_IMSTS] = { imst_we2111_original, imst_c510_original, imst_c520_rinlink, imst_c520_db9serial, imst_c530_original };

  for (imdt = 0; imdt < imdt_num; imdt++)
  {
    for (i = 0; i < NUM_IMSTS; i++)
    {
      if (strcmp(model_subtype_str, get_indicator_model_subtype_string(imsts[i], imdt)) == 0)
      {
        return imsts[i];
      }
    }
  }
  return imst_unknown;
}
