/*
 * indicator_model.h
 *
 *  Created on: Feb 11, 2015
 *      Author: ian
 */

#ifndef INDICATOR_MODEL_H_
#define INDICATOR_MODEL_H_

#include "app_db_factory.h"


/** Supported indicator models */
enum e_indicator_model
{
  im_unknown,   /*!< Could not detect indicator model */
  im_c510,      /*!< Rinstrum C510 (Rinstrum version of the WE2111) */
  im_c520,      /*!< Rinstrum C520 (indicator with display) */
  im_c530,      /*!< Rinstrum C530 (weight transmitter without display) */
  im_we2111,    /*!< HBM WE2111 */
  im_num        /*!< Number of indicator models */
};


/** Indicator model subtype */
enum e_indicator_model_sub_type
{
  imst_unknown = 0,                        /*!< Unknown/invalid indicator subtype */
  imst_unset = 1,                          /*!< Indicator subtype has not been set. Used when upgrading from old factory database */

  /* WE2111 model subtypes */
  imst_we2111_original = 2111 << 16,       /*!< The one and only WE2111 subtype */

  /* C510 model subtypes */
  imst_c510_original = (510 << 16) | 0,    /*!< The one and only C510 subtype */

  /* C520 model subtypes */
  imst_c520_rinlink = (520 << 16) | 0,     /*!< Original C520 with a rinlink on the rear board */
  imst_c520_db9serial = (520 << 16) | 1,   /*!< New C520 with the DB9 serial ports on the rear board */

  /* C530 model subtypes */
  imst_c530_original = (530 << 16) | 0,    /*!< The one and only C530 subtype */
};


enum e_indicator_model_description_type
{
  imdt_long,
  imdt_short_upper,
  imdt_short_lower,
  imdt_num
};


/** Must match @ref enum e_indicator_model */
#define INDICATOR_MODEL_STRS \
  { "Unknown", "C510 Indicator", "C520 Indicator", "C530 Weight Transmitter", "WE2111 Indicator" }


/** Must match @ref enum e_indicator_model */
#define INDICATOR_MODEL_SHORT_UPPER_STRS \
  { "DUNNO", "C510", "C520", "C530", "WE2111" }


/** Must match @ref enum e_indicator_model */
#define INDICATOR_MODEL_SHORT_LOWER_STRS \
  { "dunno", "c510", "c520", "c530", "we2111" }


/**
 * Set the indicator model and subtype from the factory database
 *
 * The logic is as follows:
 *  - If valid indicator model matches valid indicator model subtype, then run
 *  - If indicator model is unknown, but valid indicator model subtype, then use subtype to set model and run
 *  - If valid indicator model doesn't match valid indicator model subtype, then don't run (APP_ERR_ROM)
 *  - For indicator models where there is only one subtype (C510, C530, WE2111), and indicator model is valid, but indicator subtype is invalid or unset, then use model to set subtype and run
 *  - For indicator models where there are multiple subtypes (C520), and indicator model is valid, but indicator subtype unset, then this is a upgraded original C520 (rinlink), use model to set subtype and run
 *  - For indicator models where there are multiple subtypes (C520), and indicator model is valid, but indicator subtype unknown, then we can't discern what hardware is fitted, don't run (APP_ERR_ROM)
 *  - If indicator model is unknown, and indicator model subtype is unknown, then don't run (APP_ERR_ROM)
 */
void init_indicator_model_and_subtype(const struct s_db_app_factory *facdb);


/** Get the indicator model */
enum e_indicator_model get_indicator_model(void);


/** Get the indicator model subtype */
enum e_indicator_model_sub_type get_indicator_model_subtype(void);


/** Convert the indicator model sub type into an indicator model */
enum e_indicator_model indicator_model_subtype_to_indicator_model(enum e_indicator_model_sub_type imst);


/** Get a model string for an indicator model */
const char *get_indicator_model_string(enum e_indicator_model im, enum e_indicator_model_description_type imdt);


/** Get the indicator model for an indicator model string, im_unknown if no match */
enum e_indicator_model string_to_indicator_model(const char *model_str);


/** Get a model subtype string for an indicator model subtype */
const char *get_indicator_model_subtype_string(enum e_indicator_model_sub_type imst, enum e_indicator_model_description_type imdt);


/** Get the indicator model subtype for an indicator model subtype string, imst_unknown if no match */
enum e_indicator_model_sub_type string_to_indicator_model_subtype(const char *model_subtype_str);


#endif /* INDICATOR_MODEL_H_ */
