#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdbool.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <setnetconfig.h>
#include "app_db_factory.h"
#include "we_debug.h"
#include "we_config.h"

#include "ipinfo.h"


/** Setnetconfig library configuration */
struct s_snc_configuration *libsnc_config = NULL;


/** Initialise libsetnetconfig to enable us to get/set the network configuration */
bool net_ip_setnetconfig_init(void)
{
  enum e_snc_result result;

  /* Try to load the library configuration */
  result = snc_config_init(&libsnc_config);
  if (sncr_success != result)
    return false;
  (void) snc_config_load(SNC_CONFIG_FILE, libsnc_config);
  return true;
}


/** Shutdown libsetnetconfig */
void net_ip_setnetconfig_destroy(void)
{
  if (NULL != libsnc_config)
  {
    snc_config_destroy(libsnc_config);
    libsnc_config = NULL;
  }
}


/** Convert an AF_INET (IPV4) address in network endian into a string */
const char *net_ip_inet_ntop_network_afinet(const struct in_addr *ipv4_net_endian, char *buf, socklen_t len)
{
  return inet_ntop(AF_INET, ipv4_net_endian, buf, len);
}


/** Convert an AF_INET (IPV4) address in host endian into a string */
const char *net_ip_inet_ntop_host_afinet(const struct in_addr *ipv4_host_endian, char *buf, socklen_t len)
{
  struct in_addr ipv4_net_endian;

  ipv4_net_endian.s_addr = htonl(ipv4_host_endian->s_addr);
  return net_ip_inet_ntop_network_afinet(&ipv4_net_endian, buf, len);
}


/** Convert string into an AF_INET (IPV4) address in network endian */
bool net_ip_inet_pton_network_afinet(const char *str, struct in_addr *ipv4_net_endian)
{
  return (inet_pton(AF_INET, str, ipv4_net_endian) == 1);
}


/** Convert string into an AF_INET (IPV4) address in host endian */
bool net_ip_inet_pton_host_afinet(const char *str, struct in_addr *ipv4_host_endian)
{
  bool result;
  struct in_addr ipv4_net_endian;

  result = net_ip_inet_pton_network_afinet(str, &ipv4_net_endian);
  if (result)
    ipv4_host_endian->s_addr = ntohl(ipv4_net_endian.s_addr);
  return result;
}


/** Convert a MAC address to a string (xx:xx:xx:xx:xx:xx) */
void net_ip_macaddress_to_string(const struct sockaddr *mac_address, char separator, char str[MAC_ADDRESS_STR_LENGTH])
{
  snprintf(str, MAC_ADDRESS_STR_LENGTH, "%02X%c%02X%c%02X%c%02X%c%02X%c%02X",
           (uint8_t) mac_address->sa_data[0], separator,
           (uint8_t) mac_address->sa_data[1], separator,
           (uint8_t) mac_address->sa_data[2], separator,
           (uint8_t) mac_address->sa_data[3], separator,
           (uint8_t) mac_address->sa_data[4], separator,
           (uint8_t) mac_address->sa_data[5]
          );
}


/** Set the C500 preferred network defaults in a s_snc_net_config */
void net_ip_set_defaults(struct s_snc_net_config *net_config)
{
  #define MODEL_LOCASE_LEN 20
  char model_locase[MODEL_LOCASE_LEN];
  size_t i;
  size_t model_len;

  net_config->use_dhcp = DEFAULT_ETH_USEDHCP;
  net_config->ip_address.s_addr = DEFAULT_ETH_IPADDRESS;
  net_config->ip_netmask.s_addr = DEFAULT_ETH_NETMASK;
  net_config->ip_broadcast.s_addr = DEFAULT_ETH_BROADCAST;
  net_config->ip_gateway.s_addr = DEFAULT_ETH_GATEWAY;
  net_config->ip_dns[0].s_addr = DEFAULT_ETH_DNS1;
  net_config->ip_dns[1].s_addr = DEFAULT_ETH_DNS2;
  net_config->ip_dns[2].s_addr = DEFAULT_ETH_DNS3;
  snprintf(net_config->search_domain, SNC_DOMAIN_NAME_MAX, "%s", DEFAULT_ETH_SEARCH_DOMAIN);

  /* Model to lower case */
  model_len = strlen(db_app_factory.model_str);
  if (model_len > MODEL_LOCASE_LEN - 1)
    model_len = MODEL_LOCASE_LEN - 1;
  for (i = 0; i < model_len; i++)
    model_locase[i] = tolower(db_app_factory.model_str[i]);
  model_locase[model_len] = '\0';

  /* Default hostname is model_locase-serial_number */
  snprintf(net_config->hostname, SNC_HOST_NAME_MAX, "%s-%lu", model_locase, (unsigned long) db_app_factory.serial_no);
  snprintf(net_config->domain, SNC_DOMAIN_NAME_MAX, "%s", DEFAULT_ETH_DOMAIN);
}


/** Get the network configuration or network current settings */
bool net_ip_get_config(bool config_not_current, struct s_snc_net_config *net_config)
{
  enum e_snc_result result;

  snc_net_config_struct_init(net_config);

  if (config_not_current)
  {
    result = snc_get_net_config(libsnc_config, SNC_METHOD_LIB_CONFIG, we_configuration.ethernet_interface, net_config);
    if (sncr_success != result)
      rdebug_printf(LOG_ERR, "net_ip_get_config: snc_get_net_config() failed with result %s\n", snc_result_to_string(result));
  }
  else
  {
    result = snc_get_net_current(libsnc_config, SNC_METHOD_LIB_CONFIG, we_configuration.ethernet_interface, net_config);
    if (sncr_success != result)
      rdebug_printf(LOG_ERR, "net_ip_get_config: snc_get_net_current() failed with result %s\n", snc_result_to_string(result));
  }
  return (sncr_success == result);
}


#define NET_IP_SAVE_CONFIG_DO_CALLBACK(nesr) \
do \
{ \
  if (NULL != callback) \
    callback(nesr); \
} \
while (0)


#define NET_IP_SAVE_CONFIG_VALIDATE(mask, what, nesr) \
do \
{ \
  if ((valid_mask & mask) != 0) \
  { \
    NET_IP_SAVE_CONFIG_DO_CALLBACK(nesr); \
    rdebug_printf(LOG_ERR, "net_ip_save_changes: snc_net_config_validate() indicated invalid " what "\n"); \
    save_result = nesr; \
  } \
} \
while (0)


/** Saves any network configuration changes to the system */
enum e_net_ip_save_result net_ip_save_config(struct s_snc_net_config *net_config, t_net_ip_save_ui_callback_fn callback)
{
  struct s_snc_net_config existing_net_config;
  enum e_snc_result result;
  bool ifcfg_differs;
  bool hostname_differs;
  bool domain_differs;
  uint32_t valid_mask;
  enum e_net_ip_save_result save_result;

  /* Work out the broadcast address from the netmask and IP address */
  (void) snc_ipv4_broadcast_from_ip_netmask(&net_config->ip_address, &net_config->ip_netmask, &net_config->ip_broadcast);

  /* Get current system IP configuration */
  result = snc_get_net_config(libsnc_config, SNC_METHOD_LIB_CONFIG, we_configuration.ethernet_interface, &existing_net_config);
  if (sncr_success != result)
  {
    rdebug_printf(LOG_ERR, "net_ip_save_changes: snc_get_net_config() failed with error %s\n", snc_result_to_string(result));
    goto ess_fail;
  }

  /* Compare it against our modified one */
  result = snc_net_config_diff(net_config, &existing_net_config, &ifcfg_differs, &hostname_differs, &domain_differs, NULL);
  if (sncr_success != result)
  {
    rdebug_printf(LOG_ERR, "net_ip_save_changes: snc_net_config_diff() failed with error %s\n", snc_result_to_string(result));
    goto ess_fail;
  }

  /* Validate it */
  save_result = nesr_success;
  (void) snc_net_config_validate(net_config, &valid_mask, false);
  if (hostname_differs)
  {
    NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_HOSTNAME_INVALID, "hostname", nesr_bad_hostname);
  }
  if (ifcfg_differs)
  {
    NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_IPV4_NETMASK_INVALID, "netmask", nesr_bad_netmask);
    if ((valid_mask & SNC_VALIDATE_IPV4_NETMASK_INVALID) == 0)
    {
      /* An invalid netmask will make these invalid, so lets just flag that */
      NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_IPV4_ADDRESS_INVALID, "ip address", nesr_bad_ipaddress);
      NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_IPV4_BROADCAST_INVALID, "broadcast", nesr_bad_broadcast);
      NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_IPV4_GATEWAY_INVALID, "gateway", nesr_bad_gateway);
    }
    NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_IPV4_DNS0_INVALID, "name server 0", nesr_bad_dns0);
    NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_IPV4_DNS1_INVALID, "name server 1", nesr_bad_dns1);
    NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_IPV4_DNS2_INVALID, "name server 2", nesr_bad_dns1);
    NET_IP_SAVE_CONFIG_VALIDATE(SNC_VALIDATE_SEARCHDOMAIN_INVALID, "search domain", nesr_bad_searchdomain);
  }
  if (nesr_success != save_result)
  {
#if RDEBUG_ENABLED(WE_IPCONF_DEBUG_LVL, RDEBUG_LVL_FULL)
    /* Print out invalid configuration */
    char *new_net_config_str = NULL;

    snc_net_config_to_allocstr(net_config, true, &new_net_config_str);
    rdebug_printf(LOG_INFO, "net_ip_save_changes: invalid config was: %s\n",
                  new_net_config_str);
    free(new_net_config_str);
#endif /* #if WE_DEBUG(WE_IPCONF_DEBUG_LVL, WE_DEBUG_FULL) */
    return save_result;
  }

  if (ifcfg_differs || hostname_differs || domain_differs)
  {
    WE_IPCONF_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "net_ip_save_changes: network configuration has changed\n");
#if RDEBUG_ENABLED(WE_IPCONF_DEBUG_LVL, RDEBUG_LVL_FULL)
    /* Print out debugging information */
    char *old_net_config_str = NULL;
    char *new_net_config_str = NULL;

    snc_net_config_to_allocstr(&existing_net_config, true, &old_net_config_str);
    snc_net_config_to_allocstr(net_config, true, &new_net_config_str);

    rdebug_printf(LOG_INFO, "net_ip_save_changes: existing config: %s\n",
                  old_net_config_str);
    rdebug_printf(LOG_INFO, "net_ip_save_changes: new config: %s\n",
                  new_net_config_str);
    free(old_net_config_str);
    free(new_net_config_str);
#endif /* #if WE_DEBUG(WE_IPCONF_DEBUG_LVL, WE_DEBUG_FULL) */

    /* It's changed, so save it */
    NET_IP_SAVE_CONFIG_DO_CALLBACK(nesr_about_to_save);
    result = snc_set_net_config(libsnc_config, SNC_METHOD_LIB_CONFIG, we_configuration.ethernet_interface, net_config, &ifcfg_differs, &hostname_differs, &domain_differs);
    if (sncr_success != result)
    {
      rdebug_printf(LOG_ERR, "ethnet_save_setup: snc_set_net_config() failed with error %s\n", snc_result_to_string(result));
      goto ess_fail;
    }
  }
  else
  {
    /* Print out debugging information */
    WE_IPCONF_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "net_ip_save_changes: no change in network configuration\n");
  }
  NET_IP_SAVE_CONFIG_DO_CALLBACK(nesr_success);
  return nesr_success;

ess_fail:
  NET_IP_SAVE_CONFIG_DO_CALLBACK(nesr_failure);
  return nesr_failure;
}
