#include <stddef.h>
#include <string.h>
#include <rinhw/kbd.h>
#include "system.h"
#include "access.h"
#include "iface.h"
#include "database.h"
#include "buzz5000.h"
#include "modbus_old.h"
#include "comms_kbd.h"
#include "devices.h"
#include "app_db_user.h"

#include "kbd_5000.h"


/** Are remote keys enabled */
static bool internal_remote_keys_enabled = false;


/** Who is handling keys (C500 app or lua)? */
static enum e_key_handler internal_rincmd_key_handler = kh_application;


/** Rincmd remote key pressed */
static union un_key internal_rincmd_key = { KEY_PHYSICAL_NOKEY };


/** Get the current key handler (app or lua) */
enum e_key_handler get_rincmd_key_handler(void)
{
  return internal_rincmd_key_handler;
}


/** Get the current key handler (app or lua), overridden to kh_application when the menus are running */
static enum e_key_handler get_rincmd_key_handler_menu_override(void)
{
  return (edit_functions) ? kh_application : get_rincmd_key_handler();
}


/** Set the current key handler (app or lua) */
void set_rincmd_key_handler(enum e_key_handler kh)
{
  internal_rincmd_key_handler = kh;
}


/**
 * Set the current rincmd key from the R400 KEY_PHYSICAL_ constants
 *
 * @param key is the key from the R400 KEY_PHYSICAL_ constants
 * @param longpress is true if the key is a long press, false for short press
 * @param from_lua is true if the key is from the lua key register (0x324)
 *        in which case it will only be accepted if lua is the key handling application,
 *        otherwise false for key register (0x0008), which is always accepted
 */
void set_rincmd_comms_key(uint8_t key, bool longpress, bool from_lua)
{
  if (!from_lua || get_rincmd_key_handler() == kh_lua)
  {
    internal_rincmd_key.BIT.EVENT = (longpress) ? keHeld : keDown;
    internal_rincmd_key.BIT.KEY = key;
  }
}


/** Test if remote keys are enabled */
bool remote_keys_enabled(void)
{
  return internal_remote_keys_enabled;
}


/** Enable/disable remote keys */
void enable_remote_keys(bool on)
{
  internal_remote_keys_enabled = on;
}


/** Translate an input into a remote function keypress */
static uint8_t translate_remote_key(uint8_t input, bool held)
{
  uint8_t keyfnidx;
  uint8_t key;
  if (input < NUM_REMOTE_KEYS)
  {
    keyfnidx = db_app_user.remote_f[input];
    key = remote_key_lookup[keyfnidx];
    if (held)
    {
      key += DEKEY;
    }
    if (DEKEY == key)
    {
      key = NO_KEY;
    }
  }
  else
  {
    key = NO_KEY;
  }
  return key;
}


/** Translate front function key into a keycode */
enum eIndicatorKey translate_front_func_key(bool held, enum eIndicatorKey funckey)
{
  size_t keyfnidx;
  uint8_t key;

  keyfnidx = db_app_user.front_key_function[funckey - FUNC1_KEY];
  keyfnidx = (keyfnidx < NUM_FRONTFUNC) ? keyfnidx : FRONTFUNC_DISABLED;
  key = frontfunc_key_lookup[keyfnidx];
  if (held)
  {
    key += DEKEY;
  }
  if (DEKEY == key)
  {
    key = NO_KEY;
  }
  return key;
}


/** Read a key from a keyboard device */
static bool device_read_key(struct t_io_dev *dev, union un_key *hwlib_key, bool use_timeout, uint16_t timeout)
{
  bool result = false;
  bool do_read;

  if (ropen(dev, 0) >= 0)
  {
    do_read = use_timeout || (rfilelength(dev) > 0);

    if (do_read)
    {
      if (!use_timeout)
        timeout = 0;

      if (sizeof(union un_key) == rread(dev, hwlib_key, sizeof(union un_key), timeout))
      {
        result = true;
      }
    }
    rclose(dev);
  }
  return result;
}


/** Returns key press according to _KEY constants, with a timeout */
uint8_t get_key_timeout(int16u timeout)
{
  uint8_t key;
  union un_key r400_key;
  bool fail;
  int16u timeout1;
  int16u timeout2;
  bool is_indicator_key;

  /* Try to get a key from normal keyboard, then any USB keyboards */
  fail = false;
  if (timeout == 1)
  {
    timeout1 = 1;
    timeout2 = 1;
  }
  else
  {
    timeout1 = timeout / 2;
    timeout2 = timeout - timeout1;
  }

  is_indicator_key = false;
  fail = true;

  /* Only read keys from the C500 keyboard if Lua hadn't taken over, since lua will retrieve them via get_rincmd_r400_key() */
  if (get_rincmd_key_handler_menu_override() == kh_application &&
      device_read_key(&dev_kbd_we, &r400_key, true, timeout1))
  {
    is_indicator_key = true;
    fail = false;
  }

  /* If we didn't get a key from the keyboard, then always read USB keyboard keys */
  if (fail && device_read_key(&dev_kbd_event, &r400_key, true, timeout2))
  {
    fail = false;
  }

  /* Maybe there is a rincmd simulated keypress waiting */
  if (fail && internal_rincmd_key.BIT.KEY != KEY_PHYSICAL_NOKEY)
  {
    r400_key = internal_rincmd_key;
    internal_rincmd_key.BIT.KEY = KEY_PHYSICAL_NOKEY;
    fail = false;
  }

  key = NO_KEY;
  if (!fail)
  {
    /* convert kbd bits to keys */
    if (keDown == r400_key.BIT.EVENT)
    {
      switch(r400_key.BIT.KEY)
      {
        case KEY_PHYSICAL_FUNC0:
          key = ZERO_KEY;
          break;
        case KEY_PHYSICAL_FUNC1:
          key = TARE_KEY;
          break;
        case KEY_PHYSICAL_FUNC2:
          key = GROSS_KEY;
          break;
        case KEY_PHYSICAL_FUNC3:
          key = FUNC1_KEY;
          break;
        case KEY_PHYSICAL_FUNC4:
          key = FUNC2_KEY;
          break;
        case KEY_PHYSICAL_FUNC5:
          key = FUNC3_KEY;
          break;
        case KEY_PHYSICAL_SETUP:
          /* Setup key only works from indicator, not USB keyboards */
          if (is_indicator_key)
            key = CAL_KEY;
          break;

        /* External inputs */
        case KEY_PHYSICAL_EXTERNAL0:
        case KEY_PHYSICAL_EXTERNAL1:
        case KEY_PHYSICAL_EXTERNAL2:
        case KEY_PHYSICAL_EXTERNAL3:
        case KEY_PHYSICAL_EXTERNAL4:
        case KEY_PHYSICAL_EXTERNAL5:
        case KEY_PHYSICAL_EXTERNAL6:
        case KEY_PHYSICAL_EXTERNAL7:
          key = translate_remote_key(r400_key.BIT.KEY - KEY_PHYSICAL_EXTERNAL0, false);
          break;
      }
    }
    else if (keHeld == r400_key.BIT.EVENT)
    {
      switch(r400_key.BIT.KEY)
      {
        case KEY_PHYSICAL_FUNC0:
          key = DEZERO_KEY;
          break;
        case KEY_PHYSICAL_FUNC1:
          key = DETARE_KEY;
          break;
        case KEY_PHYSICAL_FUNC2:
          key = DEGROSS_KEY;
          break;
        case KEY_PHYSICAL_FUNC3:
          key = DEFUNC1_KEY;
          break;
        case KEY_PHYSICAL_FUNC4:
          key = DEFUNC2_KEY;
          break;
        case KEY_PHYSICAL_FUNC5:
          key = DEFUNC3_KEY;
          break;
        case KEY_PHYSICAL_SETUP:
          /* Setup key only works from indicator, not USB keyboards */
          if (is_indicator_key)
            key = DECAL_KEY;
          break;

        /* External inputs */
        case KEY_PHYSICAL_EXTERNAL0:
        case KEY_PHYSICAL_EXTERNAL1:
        case KEY_PHYSICAL_EXTERNAL2:
        case KEY_PHYSICAL_EXTERNAL3:
        case KEY_PHYSICAL_EXTERNAL4:
        case KEY_PHYSICAL_EXTERNAL5:
        case KEY_PHYSICAL_EXTERNAL6:
        case KEY_PHYSICAL_EXTERNAL7:
          key = translate_remote_key(r400_key.BIT.KEY - KEY_PHYSICAL_EXTERNAL0, true);
          break;
      }
    }
    else if (keUp == r400_key.BIT.EVENT)
    {
      key = IDLE_CHG;
    }
  }

  if (NO_KEY == key)
  {
    /* Check for HBM comms keys */
    key = comms_kbd_get_cmd();
  }
  if (NO_KEY == key)
  {
    key = get_modbus_key();
  }
  return key;
}


/** Returns key press according to _KEY constants */
uint8_t get_key(void)
{
  return get_key_timeout(OS_TICKS_PER_SEC / 10);
}


/**
 * Returns key press according to R400 _KEY constants, if lua is the key handling application.
 * Checks for keys, but does not block.
 */
uint8_t get_rincmd_r400_key(void)
{
  union un_key r400_key;
  uint8_t key = KEY_PHYSICAL_NOKEY;

  if (get_rincmd_key_handler_menu_override() == kh_lua)
  {
    if (device_read_key(&dev_kbd_we, &r400_key, false, 0))
      key = r400_key.BYTE;
  }

  return key;
}


/**
 * Setup keys for repeat and hold keys
 *
 * Use a mask of *_KBD constants
 */
void setup_kbd(uint16_t repeat_keys, uint16_t hold_keys)
{
  const int8u normalkey[] =
    { KEY_PHYSICAL_FUNC0, KEY_PHYSICAL_FUNC1, KEY_PHYSICAL_FUNC2, KEY_PHYSICAL_FUNC3,
      KEY_PHYSICAL_FUNC4, KEY_PHYSICAL_FUNC5, KEY_PHYSICAL_SETUP };
  int8u repeatkey[16];
  int8u holdkey[16];
  int rptctr;
  int holdctr;

  rptctr = 0;
  if (repeat_keys & ZERO_KBD)
  {
    repeatkey[rptctr++] = KEY_PHYSICAL_FUNC0;
  }
  if (repeat_keys & TARE_KBD)
  {
    repeatkey[rptctr++] = KEY_PHYSICAL_FUNC1;
  }
  if (repeat_keys & GROSS_KBD)
  {
    repeatkey[rptctr++] = KEY_PHYSICAL_FUNC2;
  }
  if (repeat_keys & FUNC1_KBD)
  {
    repeatkey[rptctr++] = KEY_PHYSICAL_FUNC3;
  }
  if (repeat_keys & FUNC2_KBD)
  {
    repeatkey[rptctr++] = KEY_PHYSICAL_FUNC4;
  }
  if (repeat_keys & FUNC3_KBD)
  {
    repeatkey[rptctr++] = KEY_PHYSICAL_FUNC5;
  }
  if (repeat_keys & CAL_KBD)
  {
    repeatkey[rptctr++] = KEY_PHYSICAL_SETUP;
  }

  holdctr = 0;
  if (hold_keys & ZERO_KBD)
  {
    holdkey[holdctr++] = KEY_PHYSICAL_FUNC0;
  }
  if (hold_keys & TARE_KBD)
  {
    holdkey[holdctr++] = KEY_PHYSICAL_FUNC1;
  }
  if (hold_keys & GROSS_KBD)
  {
    holdkey[holdctr++] = KEY_PHYSICAL_FUNC2;
  }
  if (hold_keys & FUNC1_KBD)
  {
    holdkey[holdctr++] = KEY_PHYSICAL_FUNC3;
  }
  if (hold_keys & FUNC2_KBD)
  {
    holdkey[holdctr++] = KEY_PHYSICAL_FUNC4;
  }
  if (hold_keys & FUNC3_KBD)
  {
    holdkey[holdctr++] = KEY_PHYSICAL_FUNC5;
  }
  if (hold_keys & CAL_KBD)
  {
    holdkey[holdctr++] = KEY_PHYSICAL_SETUP;
  }

  if (RIO_SUCCESS == ropen(&dev_kbd_we, 0))
  {
    rioctl(&dev_kbd_we, IOCTL_KBD_SET_BASIC, sizeof(normalkey), normalkey);
    rioctl(&dev_kbd_we, IOCTL_KBD_SET_REPEAT, rptctr, repeatkey);
    rioctl(&dev_kbd_we, IOCTL_KBD_SET_HOLD, holdctr, holdkey);
    rclose(&dev_kbd_we);
  }

  if (RIO_SUCCESS == ropen(&dev_kbd_event, 0))
  {
    rioctl(&dev_kbd_event, IOCTL_KBD_SET_BASIC, sizeof(normalkey), normalkey);
    rioctl(&dev_kbd_event, IOCTL_KBD_SET_REPEAT, rptctr, repeatkey);
    rioctl(&dev_kbd_event, IOCTL_KBD_SET_HOLD, holdctr, holdkey);
    rclose(&dev_kbd_event);
  }
}


/** Setup any remote inputs as held keys */
void setup_remote_kbd(void)
{
  size_t i;
  size_t holdctr;
  uint8_t normalkey[NUM_EXTERNAL_KEYS];
  uint8_t holdkey[NUM_REMOTE_KEYS];

  /* Set all external keys as normal first */
  for (i = 0; i < NUM_EXTERNAL_KEYS; i++)
  {
    normalkey[i] = KEY_PHYSICAL_EXTERNAL0 + i;
  }

  /* Then set any inputs as hold keys */
  for (i = 0, holdctr = 0; i < NUM_REMOTE_KEYS; i++)
  {
    if (access_get_io_is_key(i))
    {
      holdkey[holdctr++] = KEY_PHYSICAL_EXTERNAL0 + i;
    }
  }

  if (ropen(&dev_kbd_we, 0) == RIO_SUCCESS)
  {
    rioctl(&dev_kbd_we, IOCTL_KBD_SET_BASIC, sizeof(normalkey), normalkey);
    rioctl(&dev_kbd_we, IOCTL_KBD_SET_HOLD, holdctr, holdkey);
    rclose(&dev_kbd_we);
  }
}


/** Enable the keypress buzzer, based on database setting */
void setup_kbd_buzz(void)
{
  if (RIO_SUCCESS == ropen(&dev_kbd_we, 0))
  {
    rioctl(&dev_kbd_we, IOCTL_KBD_ENABLE_BUZZER, db_app_user.buzzer_enable);
    rclose(&dev_kbd_we);
  }

  if (RIO_SUCCESS == ropen(&dev_kbd_event, 0))
  {
    rioctl(&dev_kbd_event, IOCTL_KBD_ENABLE_BUZZER, db_app_user.buzzer_enable);
    rclose(&dev_kbd_event);
  }
}


/** Initialise the WE2111 keyboard */
void init_kbd(void)
{
  /* Setup repeat and held keys */
  setup_kbd(KBD_DEFAULT_REPEAT, KBD_DEFAULT_HOLD);
}
