#ifndef KBD_5000_DOT_H
#define KBD_5000_DOT_H 1

#include <stdbool.h>
#include <stdint.h>
#include <rinhw/rio.h>


/* Indicator key presses */
enum eIndicatorKey
{
  /* Front keys */
  NO_KEY = 0,
  ZERO_KEY,
  TARE_KEY,
  GROSS_KEY,
  FUNC1_KEY,
  FUNC2_KEY,
  FUNC3_KEY,
  CAL_KEY,

  /* Extended functions */
  PRINT_KEY,
  SHOW_TOTAL_KEY,
  UNDO_KEY,
  CLR_TOTAL_KEY,
  SINGLE1_KEY,
  SINGLE2_KEY,
  SINGLE3_KEY,
  SINGLE4_KEY,
  SINGLE5_KEY,
  SINGLE6_KEY,
  MHOLD_KEY,
  PHOLD_KEY,
  LHOLD_KEY,
  UNITS_KEY,

  /* Long press keys */
  DEKEY = 100,
  DEZERO_KEY,
  DETARE_KEY,
  DEGROSS_KEY,
  DEFUNC1_KEY,
  DEFUNC2_KEY,
  DEFUNC3_KEY,
  DECAL_KEY,

  /* Extended functions long press */
  DE_PRINT_KEY,
  DE_SHOW_TOTAL_KEY,
  DE_UNDO_KEY,
  DE_CLR_TOTAL_KEY,
  DE_SINGLE1_KEY,
  DE_SINGLE2_KEY,
  DE_SINGLE3_KEY,
  DE_SINGLE4_KEY,
  DE_SINGLE5_KEY,
  DE_SINGLE6_KEY,
  DE_MHOLD_KEY,
  DE_PHOLD_KEY,
  DE_LHOLD_KEY,
  DE_UNITS_KEY,
};


/** Number of physical keys on the indicator */
#define WE_NUM_PHYSICAL_KEYS (FUNC3_KEY + 1)


#define IDLE_CHG        200     /* used to tell that key released so keys with long presses can be handled properly */


/* Front cal key */
#define KEY_CAL_FRONT        DEGROSS_KEY

/* Menu keys */
#define KEY_MENU_LEVEL0      ZERO_KEY
#define KEY_MENU_LEVEL1      TARE_KEY
#define KEY_MENU_LEVEL2      GROSS_KEY
#define KEY_MENU_LEVEL3      FUNC1_KEY
#define KEY_MENU_LEVEL4      FUNC2_KEY
#define KEY_MENU_LEVEL5      FUNC3_KEY
#define KEY_MENU_OK          FUNC2_KEY
#define KEY_MENU_CANCEL      FUNC3_KEY
#define KEY_MENU_EXIT1       CAL_KEY
#define KEY_MENU_EXIT2       (DEKEY + FUNC2_KEY)

/* Editor keys */
#define KEY_EDIT_DOWN        ZERO_KEY
#define KEY_EDIT_LEFT        TARE_KEY
#define KEY_EDIT_RIGHT       GROSS_KEY
#define KEY_EDIT_UP          FUNC1_KEY
#define KEY_EDIT_OK          FUNC2_KEY
#define KEY_EDIT_CANCEL1     FUNC3_KEY
#define KEY_EDIT_CANCEL2     (DEKEY + FUNC3_KEY)


/* Keyboard bit mapping for held/repeat keys */
#define NONE_KBD      0x0000
#define ZERO_KBD      (1 << ZERO_KEY)
#define TARE_KBD      (1 << TARE_KEY)
#define GROSS_KBD     (1 << GROSS_KEY)
#define FUNC1_KBD     (1 << FUNC1_KEY)
#define FUNC2_KBD     (1 << FUNC2_KEY)
#define FUNC3_KBD     (1 << FUNC3_KEY)
#define CAL_KBD       (1 << CAL_KEY)


/* Default key repeats/held */
#define KBD_DEFAULT_REPEAT NONE_KBD
#define KBD_DEFAULT_HOLD (ZERO_KBD | TARE_KBD | GROSS_KBD | FUNC1_KBD | FUNC2_KBD | FUNC3_KBD)

/* Key repeats/held for accessing the menus */
#define KBD_MENUS_REPEAT NONE_KBD
#define KBD_MENUS_HOLD (ZERO_KBD | TARE_KBD | GROSS_KBD | FUNC1_KBD | FUNC2_KBD | FUNC3_KBD)

/* Key repeats/held for calibrating the analog output */
#define KBD_CAL_ANALOG_REPEAT (FUNC1_KBD)
#define KBD_CAL_ANALOG_HOLD (ZERO_KBD | TARE_KBD | GROSS_KBD | FUNC1_KBD)

/** Who is responsible for handling keys */
enum e_key_handler
{
  kh_application = 0,  /*!< The main application is handling keys */
  kh_lua,              /*!< Lua via rincmd is handling keys */
  kh_num               /*!< The number of key handlers */
};


/** Get the current key handler (app or lua) */
enum e_key_handler get_rincmd_key_handler(void);


/** Set the current key handler (app or lua) */
void set_rincmd_key_handler(enum e_key_handler kh);


/**
 * Set the current rincmd key from the R400 KEY_PHYSICAL_ constants
 *
 * @param key is the key from the R400 KEY_PHYSICAL_ constants
 * @param longpress is true if the key is a long press, false for short press
 * @param from_lua is true if the key is from the lua key register (0x324)
 *        in which case it will only be accepted if lua is the key handling application,
 *        otherwise false for key register (0x0008), which is always accepted
 */
void set_rincmd_comms_key(uint8_t key, bool longpress, bool from_lua);


/** Test if remote keys are enabled */
bool remote_keys_enabled(void);


/** Enable/disable remote keys */
void enable_remote_keys(bool on);


/** Translate front function key into a keycode */
enum eIndicatorKey translate_front_func_key(bool held, enum eIndicatorKey funckey);


/** Returns key press according to _KEY constants, with a timeout */
uint8_t get_key_timeout(uint16_t timeout);


/** Returns key press according to _KEY constants */
uint8_t get_key(void);


/**
 * Returns key press according to R400 _KEY constants, if lua is the key handling application.
 * Checks for keys, but does not block.
 */
uint8_t get_rincmd_r400_key(void);


/**
 * Setup keys for repeat and hold keys
 *
 * Use a mask of *_KBD constants
 */
void setup_kbd(uint16_t repeat_kbd, uint16_t hold_kbd);


/** Setup any remote inputs as held keys */
void setup_remote_kbd(void);


/** Enable the keypress buzzer, based on database setting */
void setup_kbd_buzz(void);


/** Initialise the WE2111 keyboard */
void init_kbd(void);


#endif /* #ifndef KBD_5000_DOT_H */
