#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <semaphore.h>
#include <rinhw/timemath.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include "iface.h"
#include "system.h"
#include "kbd_5000.h"
#include "database.h"
#include "meas.h"
#include "backgrnd.h"
#include "adc_8bit.h"
#include "asciidef.h"
#include "version.h"
#include "buzz5000.h"
#include "comms_print_trigger.h"
#include "comms_print.h"
#include "editor.h"
#include "iface_sd.h"
#include "setpoint.h"
#include "traceprint.h"
#include "adcthread.h"
#include "iface_serial.h"
#include "app_db_user.h"
#include "app_db_runtime_vars.h"
#include "iface_menus.h"
#include "iface_operator.h"
#include "commonstrings.h"

#include "key_drvr.h"


static sem_t key_drvr_print_finished_sem;


/** See if it's ok to handle the key */
static void check_key_permission(enum eFrontKeys key, bool *ok)
{
  enum eFrontKeyPermissions permis;

  permis = get_frontkey_permissions(key);
  if (FRONTKEY_PERMIS_DISABLED == permis || blank_display)
  {
    /* Key disabled */
    beep_long();
    *ok = false;
  }
  else if (FRONTKEY_PERMIS_ENABLED == permis)
  {
    /* Normal key operation */
    *ok = true;
  }
  else
  {
    /* Immediate key not available on this key */
    beep_long();
    *ok = false;
  }
}


/** See if it's ok to handle the key */
static void check_immed_key_permission(enum eFrontKeys key, bool *ok, bool *immediate)
{
  enum eFrontKeyPermissions permis;

  permis = get_frontkey_permissions(key);
  if (FRONTKEY_PERMIS_DISABLED == permis || blank_display)
  {
    /* Key disabled */
    beep_long();
    *ok = false;
    *immediate = false;
  }
  else if (FRONTKEY_PERMIS_ENABLED == permis)
  {
    /* Normal key operation */
    *ok = true;
    *immediate = false;
  }
  else if (FRONTKEY_PERMIS_IMMEDIATE == permis && ADC_APPROVAL_MODE_NONE == adcSuperStructure.adcUserDatabase.approvalMode)
  {
    /* Immediate key operation only in non-trade mode */
    *ok = true;
    *immediate = true;
  }
  else
  {
    /* Immediate key in trade mode - disabled */
    beep_long();
    *ok = false;
    *immediate = false;
  }
}


/** Handle zero key press */
void zero_key_driver(void)
{
  bool ok;
  bool immediate;

  check_immed_key_permission(FRONTKEY_ZERO, &ok, &immediate);
  if (ok)
  {
    /* Perform zero */
    AdcSemiautoZero(&adcSuperStructure, true, false, false, immediate);
    update_adc_operation_counter();
  }
}


/** Handle de-zero (long press of zero key). Only works in non-trade mode */
void dezero_key_driver(void)
{
  bool ok;

  check_key_permission(FRONTKEY_ZERO, &ok);
  if (ok)
  {
    /* Key enabled */
    AdcClearSemiautoZero(&adcSuperStructure, true, false);
    update_adc_operation_counter();
  }
}


/** Handle a press of the tare key */
void tare_key_driver(void)
{
  bool ok;
  bool immediate;

  check_immed_key_permission(FRONTKEY_TARE, &ok, &immediate);
  if (ok)
  {
    /* Perform tare */
    AdcSemiautoTare(&adcSuperStructure, true, false, false, immediate);
    update_adc_operation_counter();
  }
}


/** Handle longpress of tare key (operator menu) */
void detare_key_driver(void)
{
  freeze_display_and_annunciators(true);
  display_pretar_and_delay();
  (void) edit_preset_tare(false, NULL, 0);
  freeze_display_and_annunciators(false);
}


/** Handle gross/net key press */
void gross_key_driver(void)
{
  bool ok;

  check_key_permission(FRONTKEY_GROSSNET, &ok);
  if (ok)
  {
    AdcGrossNetToggle(&adcSuperStructure, true, false);
    freeze_display_and_annunciators(true);
    AdcGrossNetErrorDisplay();
    freeze_display_and_annunciators(false);
  }
}


/** Handle setup from front keys (long press of gross/net key) */
void front_cal_key_driver(void)
{
  enable_remote_keys(false);
  run_setup_menus(false);
  enable_remote_keys(true);
}


static void print_error_strings(const char *s1, const char *s2)
{
  freeze_display_and_annunciators(true);
  error_strings(s1, s2);
  freeze_display_and_annunciators(false);
}


/** Handle a print key press */
void print_key_driver(bool immediate)
{
  if (!comms_print_system_has_printer())
  {
    print_error_strings(no_str, printr_str);
    return;
  }
  if (dhLivestock == show_hold && !livest_ok())
  {
    print_error_strings("LIVE.ST", active_str);
    return;
  }
  if (!immediate && !wait_till_no_motion())
  {
    return;
  }
  if (ADC_APPROVAL_MODE_NTEP == adcSuperStructure.adcUserDatabase.approvalMode &&
      dhNone != show_hold)
  {
    print_error_strings("HOLD", active_str);
    return;
  }
  /* Check print interlock */
  if (!comms_print_interlock_ok(&comms_print_interlock))
  {
    print_error_strings("I.LOCK", active_str);
    return;
  }

  traceable_printout_initiate_standard_ui(true, immediate);
}


/** Display print and clear the total */
static void display_print_and_clear_total(void)
{
  initiate_display_total(&display_total_struct);
  if (comms_print_trigger_nontraceable_total(&key_drvr_print_finished_sem))
  {
    if (sem_timedwait_ticks(&key_drvr_print_finished_sem, OS_TICKS_PER_SEC * 5) == 0)
    {
      statistics_clear();
    }
  }
}


/** Handle a long print key press */
void deprint_key_driver(void)
{
  if (!comms_print_system_has_printer())
  {
    print_error_strings(no_str, printr_str);
    return;
  }

  display_print_and_clear_total();
}


/** Handle a function key press */
void internal_function_key_driver(bool held, enum eIndicatorKey funckey)
{
  enum eIndicatorKey key;
  bool ok, immediate;

  check_immed_key_permission(xlat_indicator_key_to_front_key(funckey), &ok, &immediate);
  if (ok)
  {
    key = translate_front_func_key(held, funckey);

    /* Prevent re-entrant loop */
    if (funckey != key)
    {
      user_iface_key_dispatcher(key, immediate);
    }
  }
}


/** Handle a function key press */
void function_key_driver(enum eIndicatorKey key)
{
  internal_function_key_driver(false, key);
}


/** Handle a function key long press */
void defunction_key_driver(enum eIndicatorKey key)
{
  internal_function_key_driver(true, key - DEKEY);
}


/** Handle setup from rear setup button */
void rear_cal_key_driver(void)
{
  enable_remote_keys(false);
  run_setup_menus(true);
  enable_remote_keys(true);
}


/** Handle the clear total key */
void clear_total_key_driver(void)
{
  display_print_and_clear_total();
}


/** Initialise the key driver */
void init_key_driver(void)
{
  (void) sem_init(&key_drvr_print_finished_sem, 0, 0);
}


/** Shutdown the key driver */
void destroy_key_driver(void)
{
  (void) sem_destroy(&key_drvr_print_finished_sem);
}
