/*
 * licensing.c
 *
 *  Created on: Nov 21, 2016
 *      Author: ian
 */
#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include <wait.h>
#include <stdio.h>
#include <rinweigh/adcinterface.h>
#include <rinlicence/rl_licence.h>
#include <rinexec/el_runcmd.h>
#include "app_db_factory.h"
#include "app_db_licence.h"
#include "we_config.h"
#include "we_debug.h"
#include "display.h"
#include "iface_sd.h"
#include "edit_string.h"
#include "adcthread.h"
#include "reboot.h"
#include "iface.h"
#include "we_config.h"

#include "licensing.h"


/** The licensing structure */
static struct s_rl_licence *c500_licence = NULL;


/** The next time a licensing check should be performed in ticks */
static uint32_t licence_check_time = 0;


/** Number of times a licensing check has been performed */
static int licence_check_count = 0;


/** True if all packages are licensed */
static bool licence_check_all_licensed = false;


/** True if all the packages should be re-checked for licences. Used to trigger the UI when the LRP command licenses a package */
static bool licence_check_force_recheck = false;


/** Initalise licensing for the application. Requires configuration and databases to be initialised */
void app_licensing_init(void)
{
  enum e_rl_result result;
  size_t num_packages;

  result = rl_init_c500(&c500_licence,
                        db_app_factory.serial_no,
                        db_app_licence.licence_codes,
                        &db_app_licence.num_licence_codes,
                        APP_DB_L_LICENCECODENUM_MAX,
                        we_configuration.opkg_executable);
  if (rlr_success != result)
  {
    rdebug_printf(LOG_ERR, "app_licensing_init: failed to create licensing: result %s\n", rl_result_to_str(result));
    return;
  }
  (void) rl_get_num_installed_packages(c500_licence, &num_packages);
  rdebug_printf(LOG_INFO, "app_licensing_init: there are %u licensable packages installed\n", (unsigned) num_packages);

  licence_check_time = OSTimeGet();
}


/** Shut down licensing for the application */
void app_licensing_destroy(void)
{
  enum e_rl_result result;

  result = rl_destroy(c500_licence);
  if (rlr_success != result)
  {
    rdebug_printf(LOG_ERR, "app_licensing_destroy: failed to destroy licensing: result %s\n", rl_result_to_str(result));
  }
}


/** Call when a package is unlicensed to force the UI to re-check the licensing */
void app_licensing_force_recheck(void)
{
  licence_check_all_licensed = false;
  licence_check_time = OSTimeGet();
}


/** Split a package name at the "-" into two halves for putting on the display. Ie "C500" "K501" */
void app_licensing_split_package_name_for_display(const char *packagename, char split_pn[2][NUM_DIGITS + 1])
{
  char *posn;
  ptrdiff_t len;

  posn = strchr(packagename, '-');
  if (posn != NULL)
  {
    len = posn - packagename;
    snprintf(split_pn[0], NUM_DIGITS + 1, "%.*s", (int) len, packagename);
    snprintf(split_pn[1], NUM_DIGITS + 1, "%.*s", (int) NUM_DIGITS, &posn[1]);
  }
  else
  {
    snprintf(split_pn[0], NUM_DIGITS + 1, "%s", packagename);
    snprintf(split_pn[1], NUM_DIGITS + 1, "%s", "");
  }
}


enum plp_result { plpr_quit, plpr_licensed, plpr_recheck };


/** Prompt user to license a package. Returns true if licensed, false otherwise */
static enum plp_result prompt_licence_package(const char *package_name, bool *quit)
{
  char split_pn[2][NUM_DIGITS + 1];
  char licence_code[LICENCECODE_LEN_C500 + 1];
  const char licence_code_digits[] = LICENCECODE_DIGIT_COMBINATIONS_C500;
  enum e_rl_result licence_result;
  enum plp_result result;
  bool keypressed;

  /* Freeze the display */
  freeze_display_and_annunciators(true);
  Dclear_flashing();

  /* Prompt user that package is not licensed */
  app_licensing_split_package_name_for_display(package_name, split_pn);
  do
  {
    if (strcmp(split_pn[1], "") == 0)
      keypressed = prompt_multi(1250, 1250 * 3, quit, split_pn[0], "NOT", "LCNSED", NULL);
    else
      keypressed = prompt_multi(1250, 1250 * 4, quit, split_pn[0], split_pn[1], "NOT", "LCNSED", NULL);
    if (NULL != quit && *quit)
    {
      rdebug_printf(LOG_ERR, "prompt_licence_package: quit set, exiting\n");
      result = plpr_quit;
      goto plp_done;
    }
    if (licence_check_force_recheck)
    {
      licence_check_force_recheck = false;
      rdebug_printf(LOG_ERR, "prompt_licence_package: licence recheck forced\n");
      result = plpr_recheck;
      goto plp_done;
    }
  } while (!keypressed);

  do
  {
    /* Get the user to enter the string */
    if (rl_licence_code_to_string(c500_licence, 0, licence_code, LICENCECODE_LEN_C500 + 1) != rlr_success)
    {
      rdebug_printf(LOG_ERR, "prompt_licence_package: converting licence code to string failed\n");
      result = plpr_quit;
      goto plp_done;
    }
    if (edit_printable_string(licence_code, LICENCECODE_LEN_C500 + 1, licence_code_digits) != EDIT_OK)
    {
      rdebug_printf(LOG_ERR, "prompt_licence_package: editing licence code string failed/aborted\n");
      result = plpr_quit;
      goto plp_done;
    }

    /* License the package */
    licence_result = rl_license_package_string(c500_licence, package_name, licence_code);
    if (rlr_success != licence_result)
    {
      rdebug_printf(LOG_ERR, "prompt_licence_package: failed to license %s, result %s\n", package_name, rl_result_to_str(licence_result));
      error_strings("BAD", "CODE");
    }
    else
    {
      (void) prompt_multi(1250, 1250 * 2, quit, "LCNSE", "OK", NULL);
    }
  }
  while (rlr_success != licence_result);
  result = plpr_licensed;

plp_done:
  /* Unfreeze the display */
  freeze_display_and_annunciators(false);

  return result;
}


/** Start inetd (L900-500) package */
static void app_licensing_l900_500(void)
{
  int ret_code;
  const char *const inetd_args[] = { "/etc/rc.d/init.d/inetd", "start", NULL };

  /* Start inetd and check the return code */
  if (el_run_command(inetd_args[0], inetd_args, &ret_code, NULL, true, false, true) != elr_success ||
      !WIFEXITED(ret_code) || 0 != WEXITSTATUS(ret_code))
  {
    rdebug_printf(LOG_ERR, "app_licensing_l900_500: failed to start inetd\n");
  }
}


/** Start internal packages */
static void app_licensing_start_internal_packages(void)
{
  enum e_rl_result result;
  bool is_licensed;

  result = rl_package_is_licensed(c500_licence, "L900-500", &is_licensed);
  if (rlr_success == result && is_licensed)
  {
    app_licensing_l900_500();
  }
}


/** Check for unlicensed packages, and prompt for passcodes for them */
void app_licensing_check(bool *quit)
{
  enum e_rl_result result;
  size_t num_packages;
  size_t i;
  const char *package_name;
  bool is_licensed;
  bool changed;
  bool all_licensed;
  enum plp_result plp;

  /* Check every 5 minutes */
  if (licence_check_all_licensed || OSTimeGet() < licence_check_time)
    return;

  result = rl_get_num_installed_packages(c500_licence, &num_packages);
  if (rlr_success != result)
  {
    rdebug_printf(LOG_ERR, "app_licensing_check: failed to get number of packages\n");
    return;
  }

  /* No changes so far */
  changed = false;

  if (0 == licence_check_count && num_packages >= APP_DB_L_LICENCECODENUM_MAX)
  {
    /* First time this runs, if the licence code list is full, look for codes to remove */
    rdebug_printf(LOG_INFO, "app_licensing_check: first run and licence code is is full, looking for codes to remove\n");
    result = rl_flush_licence_codes_for_uninstalled_packages(c500_licence);
    if (rlr_success != result && rlr_licence_not_removed != result)
    {
      rdebug_printf(LOG_ERR, "app_licensing_check: failed to flush licence code list\n");
      return;
    }
    changed = (rlr_success == result);
  }

  /* Check that each package is licensed, and prompt to license any that aren't */
alc_recheck:
  for (i = 0, all_licensed = true; i < num_packages; i++)
  {
    result = rl_get_installed_package(c500_licence, i, &package_name);
    if (rlr_success != result)
    {
      rdebug_printf(LOG_ERR, "app_licensing_check: get installed package failed\n");
      break;
    }

    result = rl_package_is_licensed(c500_licence, package_name, &is_licensed);
    if (rlr_success != result)
    {
      rdebug_printf(LOG_ERR, "app_licensing_check: check package licensed failed\n");
      break;
    }

    if (!is_licensed)
    {
      plp = prompt_licence_package(package_name, quit);
      switch (plp)
      {
        case plpr_recheck:
          goto alc_recheck;
          break;
        case plpr_licensed:
          changed = true;
          break;
        case plpr_quit:
          break;
      }

      /* Check again, and keep track of all_licensed */
      result = rl_package_is_licensed(c500_licence, package_name, &is_licensed);
      if (rlr_success != result)
      {
        rdebug_printf(LOG_ERR, "app_licensing_check: check package licensed failed\n");
        break;
      }
      if (!is_licensed)
        all_licensed = false;
    }
  }

  /* If there has been a change of the licence code list, then save the licencing database */
  if (changed)
  {
    (void) app_write_licence_database();
  }

  /* And give another 5 minutes */
  licence_check_time = OSTimeGet() + WECFG_LICENSING_CHECK_INTERVAL;
  licence_check_count++;

  /* Check all packages are licensed */
  if (!all_licensed)
  {
    if (licence_check_count >= WECFG_LICENSING_CHECK_LIMIT)
    {
      rdebug_printf(LOG_INFO, "app_licensing_check: license limit reached with one or more unlicensed packages, rebooting\n");

      /*
       * Only do the reboot part here if we haven't been instructed to quit,
       * because if we are shutting down, it is probably because
       * a firmware update is taking place, and doing a reboot
       * in the middle of a firmware update really isn't the best idea.
       */
      if (NULL == quit || (NULL != quit && !*quit))
      {
        /* Notify user */
        freeze_display_and_annunciators(true);
        Dclear_flashing();
        (void) prompt_multi(1250, 1250 * 3, quit, "LCNSE", "LIMIT", "REBOOT", NULL);
        freeze_display_and_annunciators(false);
        system_reboot();
      }
    }
  }
  else
  {
    rdebug_printf(LOG_INFO, "app_licensing_check: all packages licensed, full functionality\n");
    licence_check_all_licensed = true;
    app_licensing_start_internal_packages();
  }
}


/**
 * Reset the licensing check time after exiting the menus so
 * you get another 5 minutes before prompting for unlicensed packages again
 */
void app_licensing_reset_check_time(void)
{
  licence_check_time = OSTimeGet() + WECFG_LICENSING_CHECK_INTERVAL;
  licence_check_count = 0;
}


/** Is a package licensed? */
bool app_licensing_is_installed(const char *package_name)
{
  enum e_rl_result result;
  bool is_installed;

  result = rl_package_is_installed(c500_licence, package_name, &is_installed);
  return (rlr_success == result && is_installed);
}


/**
 * Are there any addon packages installed?
 *
 * package_name is set to the first addon package found, otherwise not set
 */
bool app_licensing_are_addon_packages_installed(const char **package_name)
{
  enum e_rl_result result;
  size_t i;
  size_t num_packages;
  enum e_rl_package_type package_type;

  result = rl_get_num_installed_packages(c500_licence, &num_packages);
  if (rlr_success != result)
    return false;

  for (i = 0; i < num_packages; i++)
  {
    result = rl_get_installed_package(c500_licence, i, package_name);
    if (rlr_success != result)
      return false;
    result = rl_get_package_type(c500_licence, *package_name, &package_type);
    if (rlr_success != result)
      return false;

    if (rlpt_addon == package_type)
      return true;
  }

  return false;
}


/** Used by LRP comms command to get the data it needs */
bool app_licensing_comms_query(size_t package_index, size_t *num_packages,
                               const char **package_name, bool *is_licensed,
                               char licence_code[LICENCECODE_LEN_C500 + 1])
{
  uint32_t licence_code_int;
  enum e_rl_result rlr;

  rlr = rl_get_num_installed_packages(c500_licence, num_packages);
  if (rlr_success != rlr)
    return false;

  rlr = rl_get_installed_package(c500_licence, package_index, package_name);
  if (rlr_success != rlr)
    return false;

  rlr = rl_package_is_licensed(c500_licence, *package_name, is_licensed);
  if (rlr_success != rlr)
    return false;

  if (*is_licensed)
  {
    rlr = rl_generate_licence_code(c500_licence, *package_name, &licence_code_int);
    if (rlr_success != rlr)
      return false;

    rlr = rl_licence_code_to_string(c500_licence, licence_code_int, licence_code, LICENCECODE_LEN_C500 + 1);
    if (rlr_success != rlr)
      return false;
  }
  else
  {
    memset(licence_code, '\0', LICENCECODE_LEN_C500 + 1);
  }
  return true;
}


/** Used by LRP comms command to license/unlicense a package */
bool app_licensing_comms_cmd(const char *package_name,
                             bool licence_not_unlicence,
                             const char *licence_code)
{
  enum e_rl_result rlr;
  bool is_licensed;

  if (licence_not_unlicence)
  {
    is_licensed = false;
    rlr = rl_package_is_licensed(c500_licence, package_name, &is_licensed);
    if (rlr_success == rlr)
    {
      if (!is_licensed)
      {
        /* Only license if not already */
        rlr = rl_license_package_string(c500_licence, package_name, licence_code);
        if (rlr_success == rlr)
          licence_check_force_recheck = true;
      }
      else
      {
        /* Already licensed, success */
        rlr = rlr_success;
      }
    }
  }
  else
  {
    rlr = rl_unlicense_package(c500_licence, package_name);
    if (rlr_success == rlr)
    {
      /* Package has been un-licensed, so force the UI to run the licence check/prompt */
      app_licensing_force_recheck();
    }
  }
  return (rlr_success == rlr);
}


/** Generate the setup bypass code for those fools who have forgotten their full setup passcode */
bool app_licensing_generate_setup_bypass_code(uint32_t *bypass_code)
{
  uint32_t licence_code_int;
  uint32_t divisor;
  size_t i;

  /*
   * To generate a bypass code, use the rinlicence tool from librinhw/rinlicence.
   * Generate a numeric code, and take the last 6 digits.
   *
   * ./rinlicence -n -a C500 -s <serial num> -p forgetful generate
   *
   * For example:
   * ./rinlicence -n -a C500 -s 3545530 -p forgetful generate
   * 245641383
   * Therefore, the bypass code is: 641383
   */

  /* Generate a passcode */
  if (rl_generate_licence_code(c500_licence, "forgetful", &licence_code_int) != rlr_success)
    return false;

  /* Work out a divisor to limit code to 6 decimal digits */
  for (i = 0, divisor = 1; i < NUM_DIGITS; i++, divisor *= 10)
    ;

  /* Limit bypass code to 6 digits */
  *bypass_code = licence_code_int % divisor;
  return true;
}
