/*
 * adcthread.c
 *
 *  Created on: 08/03/2011
 *      Author: ian
 */
#include <pthread.h>
#include <stdint.h>
#include <unistd.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rio.h>
#include <rinhw/adc_ads1216.h>
#include <rinhw/we21hw.h>
#include <rinhw/rinthread.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcmuteximpl_pthreads.h>
#include <rinweigh/adcmuteximpl_pthreads_timed.h>
#include <rinweigh/adcliveimpl_simple.h>
#include <rinweigh/adcdbdefimpl_std.h>
#include "we_debug.h"
#include "iface.h"
#include "comms_auto_trigger.h"
#include "setpoint.h"
#include "access.h"
#include "debug_loop.h"
#include "backgrnd.h"
#include "adc_8bit.h"
#include "devices.h"
#include "app_db_factory.h"
#include "app_db_user.h"
#include "app_db_runtime.h"
#include "app_db.h"
#include "app_db_defaults.h"
#include "remote_cal.h"
#include "app_db_user_calcnt.h"
#include "temperature.h"
#include "watchdog.h"

#include "adcthread.h"


/** Used to force the ADC thread to quit */
static bool adc_thread_quit = false;

/** Used to wait until the ADC thread is up and running */
static bool adc_thread_running = false;

/** The ADC thread */
static pthread_t adc_pthread;

/** The ADC mutex */
static AdcMutexImplementer *adc_mutex = NULL;

/** The ADC live weigh implementation */
static AdcLiveWeighImplementer *adc_liveweigh = NULL;

/** Used to check whether the ADC super structure has been initialised */
static bool adc_super_structure_initialised = false;

/** Structure containing all ADC data */
AdcSuperStructure adcSuperStructure;

/** Structure containing all setpoint data */
struct s_setpoint_running_info setpoint_running_info;


/*
 * Thread performance debugging
 */

#define adc_thread_performance(x) ((void) we21hw_set_rs422_termination(NULL, x))

#if RDEBUG_ENABLED(WE_ADC_THREAD_DEBUG_PERFORMANCE_ADCUPDATE, RDEBUG_LVL_BASIC)
#  warning "Do not leave WE_ADC_THREAD_DEBUG_PERFORMANCE_ADCUPDATE enabled in production code"
#  define adc_thread_performance_adcupdate(x) (adc_thread_performance(x))
#else
#  define adc_thread_performance_adcupdate(x)
#endif
#if RDEBUG_ENABLED(WE_ADC_THREAD_DEBUG_PERFORMANCE_SYNCPULSE, RDEBUG_LVL_BASIC)
#  warning "Do not leave WE_ADC_THREAD_DEBUG_PERFORMANCE_SYNCPULSE enabled in production code"
#  define adc_thread_performance_syncpulse(x) (adc_thread_performance(x))
#else
#  define adc_thread_performance_syncpulse(x)
#endif
#if RDEBUG_ENABLED(WE_ADC_THREAD_DEBUG_PERFORMANCE_IO, RDEBUG_LVL_BASIC)
#  warning "Do not leave WE_ADC_THREAD_DEBUG_PERFORMANCE_IO enabled in production code"
#  define adc_thread_performance_io(x) (adc_thread_performance(x))
#else
#  define adc_thread_performance_io(x)
#endif


static bool AdcGetCurrentDateTime(AdcSuperStructure *ass, t_datetime *dt)
{
  return GetDateTime(dt);
}


/** Get temperature for ADC library */
static bool AdcThreadGetTemperature(AdcSuperStructure *ass, int32u *temperature)
{
  if (NULL == temperature)
    return false;
  *temperature = temperature_get(tuRawAveraged);
  return true;
}


#define REPORT_NEW_ERROR(err, new, old) \
  if ((err & new) != 0 && (err & old) == 0) \
    set_error(err, "report_adc_errors: setting " #err " reported by weigh library\n")

#define CLEAR_ERROR(err, new, old) \
  if ((err & new) == 0 && (err & old) != 0) \
    clr_error(err, "report_adc_errors: clearing " #err " no longer reported by weigh library\n")


/** Report any new errors reported by the ADC library */
static void report_adc_errors(AdcErrors adc_errors)
{
  const AdcErrors app_errors = get_current_errors();

  /* ADC library will report these errors */
  REPORT_NEW_ERROR(ADC_ERR_TEMPERATURE, adc_errors, app_errors);
  REPORT_NEW_ERROR(ADC_ERR_FACTORY_DB, adc_errors, app_errors);
  REPORT_NEW_ERROR(ADC_ERR_USER_DB, adc_errors, app_errors);
  REPORT_NEW_ERROR(ADC_ERR_ADC_MAXED, adc_errors, app_errors);
  REPORT_NEW_ERROR(ADC_ERR_RUNTIME_DB, adc_errors, app_errors);
  REPORT_NEW_ERROR(ADC_ERR_HARDWARE, adc_errors, app_errors);

  /* These errors from the ADC library can be cleared if no longer reported */
  CLEAR_ERROR(ADC_ERR_ADC_MAXED, adc_errors, app_errors);
}


static void *adc_thread(void *data)
{
  bool validRdg;
  AdcErrors app_errors;
  bool data_wait_failed;
  int16u ticks;
  int16u freq;

#if RDEBUG_ENABLED(WE_ADC_THREAD_DEBUG_LVL, RDEBUG_LVL_FULL)
  struct s_loop_timing lt;

  loop_timing_init(&lt, 1);
#endif

  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_thread: started as pthread %lu, linux pid %d\n", (unsigned long) pthread_self(), (int) getpid());

  /* Now configure live weigh, since ADC library is up */
  AdcLiveWeighImplementerSimpleConfigure(&adcSuperStructure, adcSuperStructure.adcLiveWeigh, WECFG_LIVE_WEIGH_AVERAGING_MILLISECONDS);

  adc_thread_running = true;

  /* Thread loop */
  while (!adc_thread_quit)
  {
    /* Report application errors to the ADC library */
    app_errors = get_current_errors() & APP_ERR_MASK;

    /* Run ADC hander */
    adc_thread_performance_adcupdate(1);
    validRdg = AdcUpdate(&adcSuperStructure, true, app_errors, &data_wait_failed);
    adc_thread_performance_adcupdate(0);

    /* Performance timing */
    adc_thread_performance_syncpulse(1);
    adc_thread_performance_syncpulse(0);

    if (data_wait_failed)
    {
      /*
       * Something didn't work, delay so we don't use 100% CPU
       * AdcSyncFrequencies: 100 == 10Hz
       */
      freq = (adcSuperStructure.adcUserDatabase.syncFreqIdx < ADC_NUM_SYNC_FREQ) ?
             AdcSyncFrequencies[adcSuperStructure.adcUserDatabase.syncFreqIdx] : AdcSyncFrequencies[ADC_SYNC_FREQ_50_HZ];
      ticks = (OS_TICKS_PER_SEC * 10) / freq;
      if (0 == ticks)
        ticks = 1;
      OSTimeDly(ticks);
      /* TODO Set ADC error */
    }

    /* Report errors from ADC library */
    report_adc_errors(adcSuperStructure.adcDataWorking.error);

    /* Auto outputs and setpoints */
    if (!edit_functions && get_remote_cal_mode() == REMOTE_CAL_MODE_IDLE)
    {
      auto_hi_output_trigger();
      update_setpoints(&setpoint_running_info, &adcSuperStructure);
    }

    /* Write the outputs */
    adc_thread_performance_io(1);
    access_io_cards(&setpoint_running_info, &adcSuperStructure.adcInternalData.tsReading);
    adc_thread_performance_io(0);

    /* Kick the watchdog */
    watchdog_kick();

    /* Debug readings per second */
#if RDEBUG_ENABLED(WE_ADC_THREAD_DEBUG_LVL, RDEBUG_LVL_FULL)
    loop_timing_check(&lt, "adc_thread");
#endif
  }

  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_thread: finished\n");
  rin_thread_exit(NULL);
}


/** Initialise the ADC super structure */
bool adc_super_structure_init(void)
{
  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_super_structure_init: called\n");

  /* Initialise the ADC library */
  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_super_structure_init: initialising mutex\n");
  /* adc_mutex = AdcMutexImplementerPthreadsTimedCreate(); Enable this to monitor how much time the ADC mutex is spending locked */
  adc_mutex = AdcMutexImplementerPthreadsCreate();
  if (NULL == adc_mutex)
  {
    goto assi_fail;
  }
  
  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_super_structure_init: initialising liveweigh\n");
  adc_liveweigh = AdcLiveWeighImplementerSimpleCreate();
  if (NULL == adc_liveweigh)
  {
    goto assi_fail;
  }
  
  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_super_structure_init: initialising adc super structure\n");
  if (!AdcPrepare(&adcSuperStructure,
                  &dev_adc24bit_we,
                  AdcThreadGetTemperature,
                  AdcGetCurrentDateTime,
                  AppDatabaseDefaultImplementer,
                  app_save_databases,
                  NULL,
                  AppUserDBCalCountCheckImplementer,
                  adc_mutex,
                  adc_liveweigh,
                  rdebug_vprintf))
  {
    goto assi_fail;
  }

  adc_super_structure_initialised = true;

assi_fail:

  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_super_structure_init: result %u\n", adc_super_structure_initialised);
  return adc_super_structure_initialised;
}


/** Destroy the ADC super structure */
void adc_super_structure_destroy(void)
{
  if (adc_super_structure_initialised)
  {
    AdcDestroy(&adcSuperStructure);
    adc_super_structure_initialised = false;
  }
  if (NULL != adc_liveweigh)
  {
    AdcLiveWeighDestroy(adc_liveweigh);
  }
  AdcMutexImplementerPthreadsDestroy(adc_mutex);
  /* AdcMutexImplementerPthreadsTimedDestroy(adc_mutex); */
  adc_mutex = NULL;
}


/** Start the ADC thread, to be called after @ref adc_super_structure_init() */
bool adc_thread_init(void)
{
  bool res;
  uint32_t start_time;

  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_thread_init: called\n");

  /* Check that ADC super structure has been initialised */
  if (!adc_super_structure_initialised)
    return false;
    
  if (!AdcInit(&adcSuperStructure,
               &db_app_factory.adc_factory_database,
               &db_app_user.adc_user_database,
               &db_app_runtime.adc_runtime_database))
    return false;

  /* Now configure live weigh, since ADC library is up */
  AdcLiveWeighImplementerSimpleConfigure(&adcSuperStructure, adcSuperStructure.adcLiveWeigh, WECFG_LIVE_WEIGH_AVERAGING_MILLISECONDS);

  adc_thread_running = false;

  res = rin_thread_create(&adc_pthread, adc_thread, NULL, true);

  /* Wait until the ADC thread starts */
  if (res)
  {
    start_time = OSTimeGet();
    while (!adc_thread_running && OSTimeGet() < start_time + OS_TICKS_PER_SEC)
    {
      OSTimeDly(OS_TICKS_PER_SEC/10);
    }
  }

  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_thread_init: finished, res %d, thread running %d\n", res, adc_thread_running);
  return (res && adc_thread_running);
}


/** Shut down the ADC thread, to be called before @ref adc_super_structure_destroy()  */
void adc_thread_destroy(void)
{
  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_thread_destroy: called\n");

  if (adc_thread_running)
  {
    /* Notify thread to quit */
    adc_thread_quit = true;

    /* Wait for thread to quit */
    pthread_join(adc_pthread, NULL);
  }

  WE_ADC_THREAD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "adc_thread_destroy: finished\n");
}
