/*
 * menus.c
 *
 *  Created on: Feb 13, 2015
 *      Author: ian
 */
#include <stdint.h>
#include <stdbool.h>
#include "display.h"
#include "kbd_5000.h"
#include "iface.h"
#include "db_build.h"

#include "menus.h"


static const uint32_t menu_level_annunciators[mlNum] = { ANNUN_NONE, ANNUN_RANGE1, ANNUN_RANGE2, ANNUN_RANGE3, ANNUN_NONE, ANNUN_NONE, ANNUN_NONE };


static const uint32_t menu_level_keys[mlNum] = { NO_KEY, KEY_MENU_LEVEL0, KEY_MENU_LEVEL1, KEY_MENU_LEVEL2, KEY_MENU_LEVEL3, KEY_MENU_LEVEL4, KEY_MENU_LEVEL5 };


static void internal_run_menu_editor(const struct s_menu_item *item, bool full_edit, enum eMenuLevel *level, void *above_user_data);


/** Lookup the menu level for a key */
enum eMenuLevel find_level_for_key(enum eMenuLevel to_level, bool longpress, uint8_t key)
{
  enum eMenuLevel ml;
  enum eMenuLevel result;

  for (ml = mlZero, result = mlNum; ml <= to_level; ml++)
  {
    if (longpress)
    {
      if (menu_level_keys[ml] + DEKEY == key)
      {
        result = ml;
        break;
      }
    }
    else
    {
      if (menu_level_keys[ml] == key)
      {
        result = ml;
        break;
      }
    }
  }
  return result;
}


static void run_menu_item(const struct s_menu_item *item, bool full_edit, enum eMenuLevel *level, size_t index, void *user_data)
{
  bool can_run;
  bool can_exit;
  enum eMenuLevel edit_result;
  enum eMenuLevel my_level;

  /* Run menu entry/exit function */
  if (NULL != item->menu_enter_exit)
  {
    can_run = item->menu_enter_exit(true, full_edit, user_data, true);
  }
  else
  {
    can_run = true;
  }

  for (can_exit = false, my_level = *level; can_run && !can_exit && !quit_setup_menus(); )
  {
    /* Run submenu or edit function */
    if (can_run)
    {
      if (NULL != item->sub_menu)
      {
        internal_run_menu_editor(item, full_edit, level, user_data);
      }
      else if (NULL != item->edit_fn)
      {
        Dset_units_ignore_motion(UNITS_NONE);
        Dset_annunciators(ANNUN_NONE);
        Dflash_annunciators(ANNUN_NONE);
        edit_result = item->edit_fn(full_edit, user_data, index);
        if (mlNum > edit_result)
          *level = edit_result;
      }
      else
      {
        /* Both NULL means exit menus (like -END-) */
        *level = mlQuit;
      }
    }

    /* Run menu entry/exit function */
    if (NULL != item->menu_enter_exit)
    {
      /* Don't change level if prevented */
      can_exit = item->menu_enter_exit(false, full_edit, user_data, true);
      if (!can_exit)
      {
        *level = my_level;
      }
    }
    else
    {
      can_exit = true;
    }
    Dflash_annunciators(ANNUN_RANGE1 | ANNUN_RANGE2 | ANNUN_RANGE3);
  } /* while */
}


#define RUN_NEXT_LEVEL_MENUS() \
{ \
  (*level)++; \
  user_data = (item->sub_menu[index]->user_data_from_menu_above) ? above_user_data : item->sub_menu[index]->user_data; \
  run_menu_item(item->sub_menu[index], full_edit, level, index, user_data); \
  if (*level > mylevel) \
  { \
    *level = mylevel; \
  } \
}


/**
 * Run the menu editor
 *
 * @param
 */
static void internal_run_menu_editor(const struct s_menu_item *item, bool full_edit, enum eMenuLevel *level, void *above_user_data)
{
  size_t index;
  size_t numitems;
  enum eMenuLevel mylevel;
  enum eMenuLevel ml;
  uint8_t key;
  void *user_data;

  /* Check pointers */
  if (NULL == item || NULL == level)
    return;

  /* Check 1 or more items in list */
  for (numitems = 0; NULL != item->sub_menu[numitems]; numitems++)
    ;
  if (0 == numitems)
    return;

  mylevel = *level;

  Dflash_annunciators(ANNUN_RANGE1 | ANNUN_RANGE2 | ANNUN_RANGE3);
  for (index = 0; !quit_setup_menus() && mylevel <= *level; )
  {
    Dputstr(item->sub_menu[index]->title);
    Dset_units_ignore_motion(UNITS_NONE);
    Dset_annunciators(menu_level_annunciators[*level]);

    /* Get key */
    key = get_key();
    switch (key)
    {
      /* Down to next level menus/editors */
      case KEY_MENU_OK:
        if (mylevel == find_level_for_key(mylevel + 1, false, KEY_MENU_OK))
          goto navigate_this_level;
        else
          RUN_NEXT_LEVEL_MENUS();
        break;

      /* Up to previous level menus */
      case KEY_MENU_CANCEL:
        if (mylevel + 1 == find_level_for_key(mylevel + 1, false, KEY_MENU_CANCEL))
        {
          RUN_NEXT_LEVEL_MENUS();
        }
        else if (*level > mlZero)
        {
          /* Back up as far as top level */
          (*level)--;
        }
        break;

      /* Quit all menus */
      case KEY_MENU_EXIT1:
      case KEY_MENU_EXIT2:
        *level = mlQuit;
        break;

      default:
navigate_this_level:
        /* Short press */
        ml = find_level_for_key(mylevel + 1, false, key);
        if (ml < mylevel)
        {
          /* Heading back up tree */
          *level = ml;
        }
        else if (ml == mylevel)
        {
          /* Forwards through list */
          index++;
          if (NULL == item->sub_menu[index])
            index = 0;
        }
        else if (mlNum != ml && ml == mylevel + 1)
        {
          /* Down to next menu level */
          RUN_NEXT_LEVEL_MENUS();
        }

        /* Long press */
        ml = find_level_for_key(mylevel, true, key);
        if (ml == mylevel)
        {
          /* Backwards through list */
          if (0 == index || index >= numitems)
            index = numitems - 1;
          else
            index--;
        }
        break;
    }
  }
  Dflash_annunciators(ANNUN_NONE);
}


void run_menu_editor(const struct s_menu_item *item, bool full_edit, enum eMenuLevel start_level)
{
  enum eMenuLevel level = start_level;

  internal_run_menu_editor(item, full_edit, &level, NULL);
}


/** Convert an edit result into menu level */
enum eMenuLevel edit_result_to_menu_level(enum eEditResult edit_result)
{
  enum eMenuLevel menu_level;

  switch (edit_result)
  {
    case EDIT_MENU_QUIT:
      menu_level = mlQuit;
      break;
    default:
      menu_level = mlNum;
      break;
  }
  return menu_level;
}
