/*
 * modbus_config.c
 *
 *  Created on: 03/09/2016
 *      Author: ian
 */
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <pthread.h>
#include <netinet/in.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include "we_debug.h"

#include "modbus_config.h"


/** Handle commands from the command channel, called from modbus task */
bool rinmodbus_configure_read_message(struct s_rinmodbus_configure *configure, struct s_rinmodbus_configuration *configuration,
                                      bool *quit, bool *config_changed)
{
  char data;

  /* Read data */
  if (read(configure->cmd_to_thread_pipe_read_fd, &data, sizeof(data)) != sizeof(data))
  {
    rdebug_printf(LOG_ERR, "rinmodbus_configure_read_message: read from command channel failed\n");
    goto rcrm_fail;
  }

  switch(data)
  {
    case RINMODBUS_CMD_QUIT:
      *quit = true;
      break;
    case RINMODBUS_CMD_RECONFIGURE:
      *config_changed = true;
      break;
    default:
      break;
  }

  return true;

rcrm_fail:
  return false;
}


/**
 * Configure the serial port for the Modbus task
 *
 * @param configure is the configuration
 * @param port is the serial port ie (/dev/ttyUSB0). Set to "", or NULL to disconnect
 */
void rinmodbus_configure_send_cmd(struct s_rinmodbus_configure *configure, char command)
{
  bool wait_complete = configure->thread != pthread_self();

  /* Clear out the command complete semaphore */
  while (sem_trywait(&configure->cmd_complete_sem) == 0)
    ;

  configure->cmd_notify_complete = wait_complete;

  /* Send command to modbus thread adopt settings */
  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_configure_send_cmd: sending command %c to modbus task\n", command);
  write(configure->cmd_to_thread_pipe_write_fd, &command, sizeof(command));

  if (wait_complete)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_configure_send_cmd: waiting for modbus task to complete command %c\n", command);
    sem_wait(&configure->cmd_complete_sem);
  }
}


/** Tell the Modbus task to quit */
void rinmodbus_send_quit(struct s_rinmodbus_configure *configure)
{
  rinmodbus_configure_send_cmd(configure, RINMODBUS_CMD_QUIT);
}


/**
 * Configure the serial port for the Modbus task
 *
 * @param configure is the configuration
 * @param port is the serial port ie (/dev/ttyUSB0). Set to "", or NULL to disconnect
 */
void rinmodbus_configure_serial_port(struct s_rinmodbus_configure *configure,
                                     const char *port)
{
  if (NULL == configure || NULL == configure->config)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_configure_serial_port: null pointer\n");
    return;
  }

  /*
   * Because the HBM comms command called by the modbus task (eg TDD)
   * might call reinitialise_after_save(), and therefore this function
   * we can get locked up, so we test the pthread id, and don't do
   * any lock
   */

  /* Get exclusive lock on configuration structure */
  pthread_mutex_lock(&configure->cmd_lock_mutex);

  /* Free up old port */
  if (NULL != configure->config->mb_serial_port)
  {
    free(configure->config->mb_serial_port);
    configure->config->mb_serial_port = NULL;
  }

  /* Adopt new configuration */
  if (NULL != port)
    configure->config->mb_serial_port = strdup(port);
  else
    configure->config->mb_serial_port = NULL;

  rinmodbus_configure_send_cmd(configure, RINMODBUS_CMD_RECONFIGURE);

  /* Release lock */
  pthread_mutex_unlock(&configure->cmd_lock_mutex);
}


/** Send the full configuration information to the Modbus task */
void rinmodbus_configure_full_config(struct s_rinmodbus_configure *configure,
                                     enum eRinModbusEndian mb_endian,
                                     unsigned tcp_port,
                                     enum eRinModbusSerialMode mb_serial_mode,
                                     unsigned mb_serial_address,
                                     unsigned long mb_serial_baud,
                                     char mb_serial_parity_bits,
                                     unsigned mb_serial_data_bits,
                                     unsigned mb_serial_stop_bits)
{
  if (NULL == configure || NULL == configure->config)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_configure_full_config: null pointer\n");
    return;
  }

  /* Get exclusive lock on configuration structure */
  pthread_mutex_lock(&configure->cmd_lock_mutex);

  /* Adopt new configuration */
  configure->config->mb_endian = mb_endian;
  configure->config->tcp_port = tcp_port;
  configure->config->mb_serial_mode = mb_serial_mode;
  configure->config->mb_serial_address = mb_serial_address;
  configure->config->mb_serial_baud = mb_serial_baud;
  configure->config->mb_serial_parity_bits = mb_serial_parity_bits;
  configure->config->mb_serial_data_bits = mb_serial_data_bits;
  configure->config->mb_serial_stop_bits = mb_serial_stop_bits;

  rinmodbus_configure_send_cmd(configure, RINMODBUS_CMD_RECONFIGURE);

  /* Release lock */
  pthread_mutex_unlock(&configure->cmd_lock_mutex);
}


bool rinmodbus_configure_init(struct s_rinmodbus_configure *configure, size_t num_socket_connections)
{
  int cmd_pipe_fds[2] = { -1, -1 };
  bool mutex_inited = false;
  bool sem_inited = false;

  if (NULL == configure)
    return false;

  /* Zero structure */
  memset(configure, 0, sizeof(struct s_rinmodbus_configure));

  /* Create mutex */
  if (pthread_mutex_init(&configure->cmd_lock_mutex, NULL) != 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_configure_init: mutex creation failed\n");
    goto rci_fail;
  }
  mutex_inited = true;

  /* Create semaphore */
  if (sem_init(&configure->cmd_complete_sem, 0, 0) != 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_configure_init: semaphore creation failed\n");
    goto rci_fail;
  }
  sem_inited = true;

  /* Create a pipe to write commands to the task */
  if (pipe(cmd_pipe_fds) < 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_configure_init: create command write failed\n");
    goto rci_fail;
  }

  /* Keep fds for write command to task */
  configure->cmd_to_thread_pipe_read_fd = cmd_pipe_fds[0];
  configure->cmd_to_thread_pipe_write_fd = cmd_pipe_fds[1];
  if (fcntl(configure->cmd_to_thread_pipe_read_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_configure_init: error setting FD_CLOEXEC on read end of command write pipe (errno %d)\n", errno);
    goto rci_fail;
  }
  if (fcntl(configure->cmd_to_thread_pipe_write_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_configure_init: error setting FD_CLOEXEC on write end of command write pipe (errno %d)\n", errno);
    goto rci_fail;
  }

  configure->num_socket_connections = num_socket_connections;

  /* These gets set later by the modbus task */
  configure->config = NULL;
  configure->thread_running = false;
  configure->thread_has_quit = false;

  return true;

rci_fail:
  if (cmd_pipe_fds[0] >= 0)
    close(cmd_pipe_fds[0]);
  if (cmd_pipe_fds[1] >= 0)
    close(cmd_pipe_fds[1]);
  if (sem_inited)
    sem_destroy(&configure->cmd_complete_sem);
  if (mutex_inited)
    pthread_mutex_destroy(&configure->cmd_lock_mutex);
  return false;
}


/** Clean up rinmodbus configure */
void rinmodbus_configure_destroy(struct s_rinmodbus_configure *configure)
{
  /* Clean up write to thread pipe */
  close(configure->cmd_to_thread_pipe_read_fd);
  configure->cmd_to_thread_pipe_read_fd = -1;
  close(configure->cmd_to_thread_pipe_write_fd);
  configure->cmd_to_thread_pipe_write_fd = -1;

  /* Clean up semaphore and mutex */
  sem_destroy(&configure->cmd_complete_sem);
  pthread_mutex_destroy(&configure->cmd_lock_mutex);

  /* configure->config points to data cleaned up by modbus task */
}
