/*
 * modbus_config.h
 *
 *  Created on: 03/09/2016
 *      Author: ian
 */

#ifndef MODBUS_CONFIG_H_
#define MODBUS_CONFIG_H_


#include <stdbool.h>
#include <stddef.h>
#include <semaphore.h>
#include <pthread.h>


/** Rinmodbus endian setting */
enum eRinModbusEndian { RINMODBUS_ENDIAN_BIG = 0, RINMODBUS_ENDIAN_LITTLE,
                        RINMODBUS_ENDIAN_NUM
                      };


/** Rinmodbus serial mode setting */
enum eRinModbusSerialMode { RINMODBUS_SERIAL_MODE_ASCII = 0, RINMODBUS_SERIAL_MODE_RTU,
                            RINMODBUS_SERIAL_MODE_NUM
                          };


#define RINMODBUS_CMD_QUIT 'Q'
#define RINMODBUS_CMD_RECONFIGURE 'R'


/** Modbus internal configuration information */
struct s_rinmodbus_configuration
{
  enum eRinModbusEndian mb_endian;
  unsigned tcp_port;
  enum eRinModbusSerialMode mb_serial_mode;
  unsigned mb_serial_address;
  unsigned long mb_serial_baud;
  char mb_serial_parity_bits;
  unsigned mb_serial_data_bits;
  unsigned mb_serial_stop_bits;
  char *mb_serial_port;
};


/** Structure for configuring modbus */
struct s_rinmodbus_configure
{
  struct s_rinmodbus_configuration *config;         /*!< Pointer to the modbus configuration info */
  int cmd_to_thread_pipe_read_fd;                   /*!< Read by modbus task for configuration/quit instructions */
  int cmd_to_thread_pipe_write_fd;                  /*!< Written by modbus configuration function called by other threads to notify modbus task of configuration changes, etc */
  pthread_mutex_t cmd_lock_mutex;                   /*!< Used to lock the configuration command stuff while waiting for the modbus task to adopt the new configuration */
  bool cmd_notify_complete;                         /*!< Flag to indicate if the modbus task should notify the caller of completion */
  sem_t cmd_complete_sem;                           /*!< Semaphore that the modbus task uses to notify the caller of completion */
  pthread_t thread;                                 /*!< The modbus pthread */
  size_t num_socket_connections;                    /*!< The number of socket connections */
  volatile bool thread_running;                     /*!< Used to detect that the modbus thread has started */
  volatile bool thread_has_quit;                    /*!< Used to detect that the modbus thread has quit prematurely */
};


/** Handle commands from the command channel, called from modbus task */
bool rinmodbus_configure_read_message(struct s_rinmodbus_configure *configure, struct s_rinmodbus_configuration *configuration,
                                      bool *quit, bool *config_changed);


/** Tell the Modbus task to quit */
void rinmodbus_send_quit(struct s_rinmodbus_configure *configure);


/**
 * Configure the serial port for the Modbus task
 *
 * @param configure is the configuration
 * @param port is the serial port ie (/dev/ttyUSB0). Set to "", or NULL to disconnect
 */
void rinmodbus_configure_serial_port(struct s_rinmodbus_configure *configure,
                                     const char *port);


/** Send the full configuration information to the Modbus task */
void rinmodbus_configure_full_config(struct s_rinmodbus_configure *configure,
                                     enum eRinModbusEndian mb_endian,
                                     unsigned tcp_port,
                                     enum eRinModbusSerialMode mb_serial_mode,
                                     unsigned mb_serial_address,
                                     unsigned long mb_serial_baud,
                                     char mb_serial_parity_bits,
                                     unsigned mb_serial_data_bits,
                                     unsigned mb_serial_stop_bits);


bool rinmodbus_configure_init(struct s_rinmodbus_configure *configure, size_t num_socket_connections);


/** Clean up rinmodbus configure */
void rinmodbus_configure_destroy(struct s_rinmodbus_configure *configure);


#endif /* MODBUS_CONFIG_H_ */
