/*
 * modbus_task.c
 *
 * Implements:
 * Read Holding Registers (0x03)
 * Write Single Register (0x06)
 * Write Multiple Register (0x10)
 *
 */
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/select.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <rinhw/rinthread.h>
#include <rinhw/sio_local_pipe.h>
#include <netinet/in.h>
#include <modbus.h>
#include "we_debug.h"
#include "comms_net.h"
#include "modbus_xlat.h"
#include "c500_modbus_holding_registers.h"
#include "we2111_modbus_input_registers.h"
#include "modbus_config.h"

#include "modbus_task.h"


#define RINMODBUS_TOTAL_CONTEXTS(mconns) (mconns->num_client_serials + mconns->num_client_sockets)


struct s_rinmodbus_connections
{
  modbus_t *modbus_server_ctx;
  int server_socket;
  modbus_t **modbus_client_ctxs;
  size_t num_client_serials;
  size_t num_client_sockets;
  struct t_io_dev pipe_to_comms_net[2];
  struct s_rinmodbus_interp interp;
  size_t num_connections;
  struct s_rinmodbus_configuration current_config;
  struct s_rinmodbus_configuration old_config;
};


static struct s_rinmodbus_configure rinmodbus_configure;


#define RINMODBUS_CMD_READ_HOLDING_REGISTERS 0x03
#define RINMODBUS_CMD_READ_INPUT_REGISTERS 0x04
#define RINMODBUS_CMD_WRITE_SINGLE_REGISTER 0x06
#define RINMODBUS_CMD_WRITE_MULTIPLE_REGISTERS 0x10


static void rinmodbus_function_dispatch(struct s_rinmodbus_connections *m_conns, int query_length, int client_idx)
{
  uint8_t function_code;
  int header_length;

  /* Read holding registers */
  header_length = modbus_get_header_length(m_conns->modbus_client_ctxs[client_idx]);
  function_code = m_conns->interp.query[header_length];

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_function_dispatch: got request type %d, query length %d, header length %d\n", (int) function_code, query_length, header_length);

  switch (function_code)
  {
    case RINMODBUS_CMD_READ_HOLDING_REGISTERS:
      WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_function_dispatch: read holding registers\n");
      rinmodbus_read_registers(c500_modbus_holding_registers, m_conns->modbus_client_ctxs[client_idx], &m_conns->interp, query_length, m_conns->current_config.mb_endian);
      break;

    case RINMODBUS_CMD_READ_INPUT_REGISTERS:
      WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_function_dispatch: read input registers\n");
      rinmodbus_read_registers(we2111_modbus_input_registers, m_conns->modbus_client_ctxs[client_idx], &m_conns->interp, query_length, m_conns->current_config.mb_endian);
      break;

    case RINMODBUS_CMD_WRITE_SINGLE_REGISTER:
      WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_function_dispatch: write single register\n");
      rinmodbus_write_single_register(c500_modbus_holding_registers, m_conns->modbus_client_ctxs[client_idx], &m_conns->interp, query_length, m_conns->current_config.mb_endian);
      break;

    case RINMODBUS_CMD_WRITE_MULTIPLE_REGISTERS:
      WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_function_dispatch: write multiple registers\n");
      rinmodbus_write_multiple_registers(c500_modbus_holding_registers, m_conns->modbus_client_ctxs[client_idx], &m_conns->interp, query_length, m_conns->current_config.mb_endian);
      break;

    default:
      modbus_reply_exception(m_conns->modbus_client_ctxs[client_idx], m_conns->interp.query, MODBUS_EXCEPTION_ILLEGAL_FUNCTION);
      break;
  }
}


static modbus_t *rinmodbus_get_client_socket(struct s_rinmodbus_connections *m_conns)
{
  size_t i;

  for (i = 0; i < m_conns->num_client_sockets; i++)
  {
    if (modbus_get_socket(m_conns->modbus_client_ctxs[i]) < 0)
    {
      return m_conns->modbus_client_ctxs[i];
    }
  }
  return NULL;
}


static bool rinmodbus_accept_new_connection(struct s_rinmodbus_connections *m_conns)
{
  modbus_t *client;

  /* Get a free client */
  client = rinmodbus_get_client_socket(m_conns);
  if (NULL == client)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_accept_new_connection: no unused connections available\n");
    return false;
  }

  /* Accept the new connection */
  if (modbus_tcp_accept(client, &m_conns->server_socket) < 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_accept_new_connection: modbus_tcp_accept() failed\n");
    modbus_close(client);
    return false;
  }

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_accept_new_connection: accepted new connection\n");

  return true;
}


static void rinmodbus_receive(struct s_rinmodbus_connections *conns, const fd_set *readset)
{
  int req_length;
  int clientfd;
  modbus_t *ctx;
  size_t i;

  for (i = 0; i < RINMODBUS_TOTAL_CONTEXTS(conns); i++)
  {
    ctx = conns->modbus_client_ctxs[i];
    if (NULL != ctx)
    {
      clientfd = modbus_get_socket(ctx);
      if (clientfd >= 0 && FD_ISSET(clientfd, readset))
      {
        WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_receive: receiving data for client %u\n", (unsigned) i);
        req_length = modbus_receive(ctx, conns->interp.query);
        if (req_length < 0)
        {
          /* Error */
          modbus_close(ctx);
          rdebug_printf(LOG_ERR, "rinmodbus_receive: modbus_receive() error\n");
        }
        else if (0 == req_length)
        {
          /* Ignored request, not an error, but nothing to do */
        }
        else
        {
          /* Request for us to process */
          rinmodbus_function_dispatch(conns, req_length, i);
        }
      }
    }
  }
}


static modbus_t *rinmodbus_create_context(in_port_t tcp_port)
{
  modbus_t *ctx;

  /* Create new context for listening on specified TCP port. ip_address == NULL -> Binds to INADDR_ANY */
  ctx = modbus_new_tcp(NULL, tcp_port);
  if (NULL == ctx)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_create_context: modbus_new_tcp() failed with error %d: %s\n", errno, modbus_strerror(errno));
    return NULL;
  }

#if 0
  /* No point turning on debug, because it uses fprintf(stderr) which we don't support when running as daemon */
  if (modbus_set_debug(ctx, true) < 0)
  {
    rdebug_printf(LOG_WARNING, "rinmodbus_create_context: modbus_set_debug() failed\n");
  }
#endif

  return ctx;
}


static bool rinmodbus_reconfigure_tcpport(struct s_rinmodbus_connections *m_conns)
{
  struct s_rinmodbus_configuration *old_cfg = &m_conns->old_config;
  struct s_rinmodbus_configuration *new_cfg = &m_conns->current_config;

  /* Change of TCP port means some work to do */
  if (old_cfg->tcp_port != new_cfg->tcp_port)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure_tcpport: tcp port has changed, re-configuring\n");

    /* Close and clean up existing settings */
    if (m_conns->server_socket >= 0)
    {
      WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure_tcpport: closing existing socket %d\n", m_conns->server_socket);
      close(m_conns->server_socket);
      m_conns->server_socket = -1;
    }
    if (NULL != m_conns->modbus_server_ctx)
    {
      WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure_tcpport: freeing existing server context\n");
      modbus_free(m_conns->modbus_server_ctx);
      m_conns->modbus_server_ctx = NULL;
    }

    /* Create new server context */
    m_conns->modbus_server_ctx = rinmodbus_create_context(new_cfg->tcp_port);
    if (NULL == m_conns->modbus_server_ctx)
    {
      rdebug_printf(LOG_ERR, "rinmodbus_reconfigure_tcpport: rinmodbus_create_context()\n");
      goto rrt_fail;
    }

    /*
     * Handle a number of established connections.
     *
     * We allow this to fail if something else is bound to our TCP port, so that modbus serial will continue to work.
     */
    m_conns->server_socket = modbus_tcp_listen(m_conns->modbus_server_ctx, m_conns->num_connections);
    if (m_conns->server_socket < 0)
    {
      switch (errno)
      {
        case EADDRINUSE:
          rdebug_printf(LOG_ERR, "rinmodbus_reconfigure_tcpport: failed to establish listening socket as tcp port %u is already in use. check system setup for conflicts.\n", new_cfg->tcp_port);
          rdebug_printf(LOG_ERR, "rinmodbus_reconfigure_tcpport: modbus tcp will not work until you resolve this conflict\n");
          break;
        case EACCES:
          rdebug_printf(LOG_ERR, "rinmodbus_reconfigure_tcpport: access denied while trying to establish listening socket on tcp port %u.\n", new_cfg->tcp_port);
          rdebug_printf(LOG_ERR, "rinmodbus_reconfigure_tcpport: if you are not running as root, configure modbus tcp to use a tcp port above 1024.\n");
          rdebug_printf(LOG_ERR, "rinmodbus_reconfigure_tcpport: modbus tcp will not work until you resolve this.\n");
          break;
        default:
          rdebug_printf(LOG_ERR, "rinmodbus_reconfigure_tcpport: failed establish listening socket on tcp port %u with error %d: %s\n", new_cfg->tcp_port, errno, modbus_strerror(errno));
          goto rrt_fail;
          break;
      }
    }
    else
    {
      rdebug_printf(LOG_INFO, "rinmodbus_reconfigure_tcpport: modbus tcp listening on tcp port %u\n", new_cfg->tcp_port);
    }
  }
  return true;

rrt_fail:
  /* Close and clean up existing settings */
  if (m_conns->server_socket >= 0)
  {
    close(m_conns->server_socket);
    m_conns->server_socket = -1;
  }
  if (NULL != m_conns->modbus_server_ctx)
  {
    modbus_free(m_conns->modbus_server_ctx);
    m_conns->modbus_server_ctx = NULL;
  }
  return false;
}


/** Has the serial port changed */
static bool rinmodbus_serial_port_changed(const char *old_port, const char *new_port)
{
  bool result;

  if (NULL == old_port && NULL != new_port)
    result = true;
  else if (NULL != old_port && NULL == new_port)
    result = true;
  else if (NULL == old_port && NULL == new_port)
    result = false;
  else
    result = strcmp(old_port, new_port) != 0;
  return result;
}


/** Re-configure the serial context */
static bool rinmodbus_reconfigure_serial(struct s_rinmodbus_connections *m_conns)
{
  struct s_rinmodbus_configuration *old_cfg = &m_conns->old_config;
  struct s_rinmodbus_configuration *new_cfg = &m_conns->current_config;
  bool recreated = false;
  size_t seridx;

  /* Next one after the client sockets is the serial socket. There is only one modbus serial */
  seridx = m_conns->num_client_sockets;

  if (old_cfg->mb_serial_baud != new_cfg->mb_serial_baud ||
      old_cfg->mb_serial_parity_bits != new_cfg->mb_serial_parity_bits ||
      old_cfg->mb_serial_data_bits != new_cfg->mb_serial_data_bits ||
      old_cfg->mb_serial_stop_bits != new_cfg->mb_serial_stop_bits ||
      old_cfg->mb_serial_mode != new_cfg->mb_serial_mode ||
      rinmodbus_serial_port_changed(old_cfg->mb_serial_port, new_cfg->mb_serial_port))
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure_serial: serial configuration has changed\n");

    /* Free existing one */
    if (NULL != m_conns->modbus_client_ctxs[seridx])
    {
      WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure_serial: freeing existing modbus serial context\n");
      modbus_free(m_conns->modbus_client_ctxs[seridx]);
      m_conns->modbus_client_ctxs[seridx] = NULL;
    }

    /* If we have a serial port specified */
    if (NULL != new_cfg->mb_serial_port)
    {
      /* Serial configuration changed, re-create serial context */
      if (RINMODBUS_SERIAL_MODE_ASCII == new_cfg->mb_serial_mode)
      {
        WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure_serial: creating modbus serial ascii context for %s\n", new_cfg->mb_serial_port);
        m_conns->modbus_client_ctxs[seridx] = modbus_new_ascii(new_cfg->mb_serial_port,
                                                               new_cfg->mb_serial_baud,
                                                               new_cfg->mb_serial_parity_bits,
                                                               new_cfg->mb_serial_data_bits,
                                                               new_cfg->mb_serial_stop_bits);
      }
      else
      {
        WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure_serial: creating modbus serial rtu context for %s\n", new_cfg->mb_serial_port);
        m_conns->modbus_client_ctxs[seridx] = modbus_new_rtu(new_cfg->mb_serial_port,
                                                             new_cfg->mb_serial_baud,
                                                             new_cfg->mb_serial_parity_bits,
                                                             new_cfg->mb_serial_data_bits,
                                                             new_cfg->mb_serial_stop_bits);
      }
      if (NULL == m_conns->modbus_client_ctxs[seridx])
      {
        rdebug_printf(LOG_INFO, "rinmodbus_reconfigure_serial: failed to create modbus serial context for port %s, errno %d\n", new_cfg->mb_serial_port, errno);
        goto rrs_fail;
      }

      /* Connect it */
      if (modbus_connect(m_conns->modbus_client_ctxs[seridx]) != 0)
      {
        rdebug_printf(LOG_INFO, "rinmodbus_reconfigure_serial: failed to connect modbus serial for port %s, errno %d\n", new_cfg->mb_serial_port, errno);
        goto rrs_fail;
      }

      /* We've re-created a context */
      rdebug_printf(LOG_INFO, "rinmodbus_reconfigure_serial: modbus %s serial available on port %s at %lu%c%u%u\n",
                    (RINMODBUS_SERIAL_MODE_ASCII == new_cfg->mb_serial_mode) ? "ascii" : "rtu",
                    new_cfg->mb_serial_port,
                    new_cfg->mb_serial_baud,
                    new_cfg->mb_serial_parity_bits,
                    new_cfg->mb_serial_data_bits,
                    new_cfg->mb_serial_stop_bits);
      recreated = true;
    }
    else
    {
      rdebug_printf(LOG_INFO, "rinmodbus_reconfigure_serial: modbus serial is not available because no serial port has been attached and configured\n");
    }
  }

  /* Set modbus serial address */
  if (NULL != m_conns->modbus_client_ctxs[seridx] &&
      (recreated || old_cfg->mb_serial_address != new_cfg->mb_serial_address))
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure_serial: serial address has changed, setting for serial context\n");
    if (modbus_set_slave(m_conns->modbus_client_ctxs[seridx], new_cfg->mb_serial_address) != 0)
    {
      rdebug_printf(LOG_INFO, "rinmodbus_reconfigure_serial: failed to set serial slave address to %u, errno %d\n", new_cfg->mb_serial_address, errno);
      goto rrs_fail;
    }
  }
  return true;

rrs_fail:
  if (NULL != m_conns->modbus_client_ctxs[seridx])
  {
    modbus_free(m_conns->modbus_client_ctxs[seridx]);
    m_conns->modbus_client_ctxs[seridx] = NULL;
  }
  return false;
}


/** Copy the modbus configuration from src to dst */
static void rinmodbus_configuration_copy(struct s_rinmodbus_configuration *dst, const struct s_rinmodbus_configuration *src)
{
  /* Free old serial port */
  if (NULL != dst->mb_serial_port)
  {
    free(dst->mb_serial_port);
    dst->mb_serial_port = NULL;
  }

  /* Copy all members */
  *dst = *src;

  /* Duplicate string */
  if (NULL != src->mb_serial_port)
    dst->mb_serial_port = strdup(src->mb_serial_port);
}


static bool rinmodbus_reconfigure(struct s_rinmodbus_connections *m_conns)
{
  struct s_rinmodbus_configuration *old_cfg = &m_conns->old_config;
  struct s_rinmodbus_configuration *new_cfg = &m_conns->current_config;

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure: called\n");

  /* Nothing to be done with endian setting (new->mb_endian) */

  /* Reconfigure tcp */
  if (!rinmodbus_reconfigure_tcpport(m_conns))
  {
    goto rmr_fail;
  }

  /* Reconfigure serial */
  if (!rinmodbus_reconfigure_serial(m_conns))
  {
    goto rmr_fail;
  }

  /* Update old copy */
  rinmodbus_configuration_copy(old_cfg, new_cfg);

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reconfigure: finished\n");
  return true;

rmr_fail:
  return false;
}


static bool rinmodbus_handle_reconfigure(struct s_rinmodbus_connections *m_conns, struct s_rinmodbus_configure *configure, bool *quit)
{
  bool config_changed;
  bool result;

  config_changed = false;
  result = rinmodbus_configure_read_message(configure, &m_conns->current_config, quit, &config_changed);
  if (result)
  {
    if (config_changed)
    {
      result = rinmodbus_reconfigure(m_conns);
    }

    if (configure->cmd_notify_complete)
      sem_post(&configure->cmd_complete_sem);
  }
  return result;
}


/** Select on all sockets */
static bool rinmodbus_select(struct s_rinmodbus_connections *m_conns, struct s_rinmodbus_configure *configure, bool *quit)
{
  fd_set readset;
  int maxfd;
  size_t i;
  int client_socket;

  /* Clear the reference set of sockets */
  maxfd = -1;
  FD_ZERO(&readset);

  /* Add the server socket */
  if (m_conns->server_socket >= 0)
  {
    FD_SET(m_conns->server_socket, &readset);
    if (maxfd < m_conns->server_socket)
      maxfd = m_conns->server_socket;
  }

  /* Add the command channel pipe */
  FD_SET(configure->cmd_to_thread_pipe_read_fd, &readset);
  if (maxfd < configure->cmd_to_thread_pipe_read_fd)
    maxfd = configure->cmd_to_thread_pipe_read_fd;

  /* Add any clients (serial + socket) */
  for (i = 0; i < RINMODBUS_TOTAL_CONTEXTS(m_conns); i++)
  {
    if (NULL != m_conns->modbus_client_ctxs[i])
    {
      client_socket = modbus_get_socket(m_conns->modbus_client_ctxs[i]);
      if (client_socket >= 0)
      {
        FD_SET(client_socket, &readset);
        if (maxfd < client_socket)
          maxfd = client_socket;
      }
    }
  }

  /* Wait for any data */
  if (select(maxfd + 1, &readset, NULL, NULL, NULL) < 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_select: select() failed\n");
    return false;
  }
  else
  {
    /* Shutdown issued */
    if (FD_ISSET(configure->cmd_to_thread_pipe_read_fd, &readset))
    {
      if (!rinmodbus_handle_reconfigure(m_conns, configure, quit))
        return false;
    }

    /* Someone has data */
    if (m_conns->server_socket >= 0 && FD_ISSET(m_conns->server_socket, &readset))
    {
      /* New connection */
      rinmodbus_accept_new_connection(m_conns);
    }

    /* Handle data on an established connection */
    rinmodbus_receive(m_conns, &readset);
  }
  return true;
}


static bool rinmodbus_create_client_contexts(struct s_rinmodbus_connections *m_conns, size_t num_socket_connections, size_t num_serial_connections)
{
  size_t i;

  m_conns->num_client_sockets = num_socket_connections;
  m_conns->num_client_serials = num_serial_connections;

  /* Create an array of pointers */
  m_conns->modbus_client_ctxs = calloc(RINMODBUS_TOTAL_CONTEXTS(m_conns), sizeof(modbus_t *));
  if (NULL == m_conns->modbus_client_ctxs)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_create_client_contexts: failed to allocate connections array\n");
    goto rccc_fail;
  }

  /* NULL it */
  for (i = 0; i < RINMODBUS_TOTAL_CONTEXTS(m_conns); i++)
  {
    m_conns->modbus_client_ctxs[i] = NULL;
  }

  /* Create socket contexts */
  for (i = 0; i < num_socket_connections; i++)
  {
    m_conns->modbus_client_ctxs[i] = rinmodbus_create_context(0);
    if (NULL == m_conns->modbus_client_ctxs[i])
    {
      rdebug_printf(LOG_ERR, "rinmodbus_create_client_contexts: failed to allocate contexts\n");
      goto rccc_fail;
    }
  }

  /* Set serial contexts to NULL, this is created later when serial port is registered */
  for (i = 0; i < num_serial_connections; i++)
  {
    m_conns->modbus_client_ctxs[num_socket_connections + i] = NULL;
  }

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_create_client_contexts: there are %u tcp contexts and %u serial contexts, total %u\n",
                              (unsigned) m_conns->num_client_sockets,
                              (unsigned) m_conns->num_client_serials,
                              (unsigned) RINMODBUS_TOTAL_CONTEXTS(m_conns));
  return true;

rccc_fail:
  if (NULL != m_conns->modbus_client_ctxs)
  {
    for (i = 0; i < RINMODBUS_TOTAL_CONTEXTS(m_conns); i++)
    {
      if (NULL != m_conns->modbus_client_ctxs[i])
      {
        modbus_free(m_conns->modbus_client_ctxs[i]);
        m_conns->modbus_client_ctxs[i] = NULL;
      }
    }
    free(m_conns->modbus_client_ctxs);
    m_conns->modbus_client_ctxs = NULL;
  }
  m_conns->num_client_sockets = 0;
  m_conns->num_client_serials = 0;

  return false;
}


static void rinmodbus_free_client_contexts(struct s_rinmodbus_connections *m_conns)
{
  size_t i;

  if (NULL != m_conns->modbus_client_ctxs)
  {
    for (i = 0; i < RINMODBUS_TOTAL_CONTEXTS(m_conns); i++)
    {
      if (NULL != m_conns->modbus_client_ctxs[i])
      {
        modbus_free(m_conns->modbus_client_ctxs[i]);
        m_conns->modbus_client_ctxs[i] = NULL;
      }
    }
    free(m_conns->modbus_client_ctxs);
    m_conns->modbus_client_ctxs = NULL;
  }
  m_conns->num_client_sockets = 0;
  m_conns->num_client_serials = 0;
}


/** Modbus handling thread */
static void *rinmodbus_thread(void *data)
{
  struct s_rinmodbus_configure *m_cfg = (struct s_rinmodbus_configure *) data;
  struct s_rinmodbus_connections *m_conns = NULL;
  bool quit;

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_thread: started\n");

  if (NULL == m_cfg)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_thread: configure structure passed as null pointer\n");
    goto mt_fail;
  }

  /* Create connection structure */
  m_conns = calloc(1, sizeof(struct s_rinmodbus_connections));
  if (NULL == m_conns)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_thread: failed to allocate modbus_connections\n");
    goto mt_fail;
  }
  m_conns->num_client_sockets = 0;
  m_conns->num_client_serials = 0;
  m_conns->server_socket = -1;
  m_conns->num_connections = m_cfg->num_socket_connections;

  /* Current config, some sensible defaults */
  m_conns->current_config.mb_endian = RINMODBUS_ENDIAN_BIG;
  m_conns->current_config.tcp_port = MODBUS_TCP_DEFAULT_PORT;
  m_conns->current_config.mb_serial_mode = RINMODBUS_SERIAL_MODE_ASCII;
  m_conns->current_config.mb_serial_address = 1;
  m_conns->current_config.mb_serial_baud = 9600;
  m_conns->current_config.mb_serial_parity_bits = 'N';
  m_conns->current_config.mb_serial_data_bits = '8';
  m_conns->current_config.mb_serial_stop_bits = '1';
  m_conns->current_config.mb_serial_port = NULL;

  /* Old config with some different ones to force reconfigure */
  m_conns->old_config.mb_endian = RINMODBUS_ENDIAN_BIG;
  m_conns->old_config.tcp_port = 0;
  m_conns->old_config.mb_serial_mode = RINMODBUS_SERIAL_MODE_RTU;
  m_conns->old_config.mb_serial_address = 0;
  m_conns->old_config.mb_serial_baud = 9600;
  m_conns->old_config.mb_serial_parity_bits = 'E';
  m_conns->old_config.mb_serial_data_bits = '7';
  m_conns->old_config.mb_serial_stop_bits = '2';
  m_conns->old_config.mb_serial_port = NULL;

  m_cfg->config = &m_conns->current_config;

  /* Create client contexts */
  if (!rinmodbus_create_client_contexts(m_conns, m_cfg->num_socket_connections, 1))
  {
    goto mt_fail;
  }

  /* Re-configure, which re-creates context */
  if (!rinmodbus_reconfigure(m_conns))
  {
    goto mt_fail;
  }

  /* Create pipe */
  if (rio_create_device_sio_pipe(&m_conns->pipe_to_comms_net[0], &m_conns->pipe_to_comms_net[1]) < 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_thread: rio_create_device_sio_pipe() failed\n");
    goto mt_fail;
  }
  if (ropen(&m_conns->pipe_to_comms_net[0], 0) < 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_thread: open of pipe 1 failed\n");
    goto mt_fail;
  }
  if (ropen(&m_conns->pipe_to_comms_net[1], 0) < 0)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_thread: open of pipe 2 failed\n");
    goto mt_fail;
  }

  /* Create a comms handler used to convert modbus registers to commands */
  m_conns->interp.cn = comms_net_create(&m_conns->pipe_to_comms_net[0], NULL, false, NULL);
  if (NULL == m_conns->interp.cn)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_thread: comms_net_create() failed\n");
    goto mt_fail;
  }
  m_conns->interp.cn_pipe = &m_conns->pipe_to_comms_net[1];
  m_conns->interp.cn->comms_ctrl.reply_active = true;
  m_conns->interp.cn->comms_ctrl.unit_selected = true;
  m_conns->interp.cn->comms_ctrl.passcode_ok = false;

  /* Run S99 in the comms handler to get things ready */
  if (rinmodbus_run_hbm_cmd(m_conns->interp.cn, m_conns->interp.cn_pipe, "S99;") != MODBUS_EXCEPTION_MAX)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_thread: configure comms net with S99; failed\n");
    goto mt_fail;
  }

  /* Main thread loop */
  m_cfg->thread_running = true;
  for (quit = false; !quit; )
  {
    if (!rinmodbus_select(m_conns, m_cfg, &quit))
      break;
  }

mt_fail:
  if (NULL != m_cfg)
  {
    m_cfg->thread_has_quit = true;
    m_cfg->config = NULL;
  }
  if (NULL != m_conns)
  {
    if (NULL != m_conns->modbus_server_ctx)
    {
      modbus_free(m_conns->modbus_server_ctx);
      m_conns->modbus_server_ctx = NULL;
    }
    if (m_conns->server_socket >= 0)
    {
      close(m_conns->server_socket);
      m_conns->server_socket = -1;
    }
    rinmodbus_free_client_contexts(m_conns);
    if (NULL != m_conns->current_config.mb_serial_port)
    {
      free(m_conns->current_config.mb_serial_port);
      m_conns->current_config.mb_serial_port = NULL;
    }
    if (NULL != m_conns->old_config.mb_serial_port)
    {
      free(m_conns->old_config.mb_serial_port);
      m_conns->old_config.mb_serial_port = NULL;
    }
    if (NULL != m_conns->interp.cn)
    {
      comms_net_destroy(m_conns->interp.cn);
      m_conns->interp.cn = NULL;
    }
    m_conns->interp.cn_pipe = NULL;
    rdestroy_io(&m_conns->pipe_to_comms_net[0]);
    rdestroy_io(&m_conns->pipe_to_comms_net[1]);
    free(m_conns);
    m_conns = NULL;
  }

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_thread: finished\n");
  return NULL;
}


/** Initialise modbus */
bool rinmodbus_init(size_t num_connections)
{
  bool rc;
  uint32_t start_time;

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_init: called, starting modbus thread\n");

  if (!rinmodbus_configure_init(&rinmodbus_configure, num_connections))
  {
    goto mi_fail;
  }

  rc = rin_thread_create(&rinmodbus_configure.thread, rinmodbus_thread, &rinmodbus_configure, true);
  if (!rc)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_init: failed to create rinmodbus thread\n");
    goto mi_fail;
  }

  /*
   * Since we rely on the thread to set up the configure->config pointer, we want to wait here
   * until this has been done.
   */
  start_time = OSTimeGet();
  while (!rinmodbus_configure.thread_running && OSTimeGet() < start_time + OS_TICKS_PER_SEC * 2)
  {
    OSTimeDly(OS_TICKS_PER_SEC / 10);
  }
  if (!rinmodbus_configure.thread_running)
  {
    rdebug_printf(LOG_ERR, "rinmodbus_init: rinmodbus thread failed to start\n");
    goto mi_fail;
  }

  return true;

mi_fail:
  return false;
}


/** Force rin modbus to re-initalise settings from database */
void rinmodbus_reinit(enum eRinModbusEndian mb_endian,
                      unsigned tcp_port,
                      enum eRinModbusSerialMode mb_serial_mode,
                      unsigned mb_serial_address,
                      unsigned long mb_serial_baud,
                      char mb_serial_parity_bits,
                      unsigned mb_serial_data_bits,
                      unsigned mb_serial_stop_bits)
{
  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reinit: called\n");

  rinmodbus_configure_full_config(&rinmodbus_configure,
                                  mb_endian,
                                  tcp_port,
                                  mb_serial_mode,
                                  mb_serial_address,
                                  mb_serial_baud,
                                  mb_serial_parity_bits,
                                  mb_serial_data_bits,
                                  mb_serial_stop_bits);

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_reinit: finished\n");
}


/** Change the serial port used by Modbus */
void rinmodbus_reinit_serial_port(const char *port)
{
  rinmodbus_configure_serial_port(&rinmodbus_configure, port);
}


/** Shutdown modbus */
void rinmodbus_destroy(void)
{
  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_destroy: called\n");

  if (!rinmodbus_configure.thread_has_quit)
  {
    /* Write shutdown char to the pipe to force select() to return */
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "rinmodbus_destroy: write shutdown command to command pipe\n");
    rinmodbus_send_quit(&rinmodbus_configure);

    /* Wait for thread to quit */
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "rinmodbus_destroy: wait for task quit\n");
    pthread_join(rinmodbus_configure.thread, NULL);
  }

  /* Clean up */
  rinmodbus_configure_destroy(&rinmodbus_configure);

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_destroy: finished\n");
}
