/*
 * modbus_xlat.c
 *
 *  Created on: Sep 5, 2014
 *      Author: ian
 */
#include <stdint.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <stdbool.h>
#include <modbus.h>
#include <rinhw/sio.h>
#include "comms_net.h"
#include "we_debug.h"
#include "modbus_regs.h"

#include "modbus_xlat.h"


static const struct s_rinmodbus_translate *rinmodbus_lookup_register(const struct s_rinmodbus_translate *xlat_table, uint16_t mb_register)
{
  size_t i;

  /* Lookup register */
  for (i = 0; rmsUNKNOWN != xlat_table[i].regtype; i++)
  {
    if (mb_register == xlat_table[i].regnum)
    {
      return &xlat_table[i];
    }
  }
  return NULL;
}


static void rinmodbus_reg_to_query(const struct s_rinmodbus_translate *const xlat, char *buffer, size_t buf_len)
{
  char idx_str[20];
  int i;

  strncpy(buffer, xlat->cmd, buf_len);
  buffer[buf_len - 1] = '\0';

  strncat(buffer, "?", buf_len);
  if (xlat->index_param >= 0)
  {
    for (i = 0; i < xlat->index_param - 1; i++)
      strncat(buffer, ",", buf_len);
    snprintf(idx_str, 20, "%u", (unsigned) xlat->index);
    strncat(buffer, idx_str, buf_len);
  }
  strncat(buffer, ";", buf_len);
}


/** Calculate the number of bytes used by this register */
static int rinmodbus_reg_bytes_used(const struct s_rinmodbus_translate *const xlat)
{
  int bytes_used;

  switch (xlat->regtype)
  {
    case rmsSINT16:
    case rmsUINT16:
      bytes_used = 2;
      break;
    case rmsSINT32:
    case rmsUINT32:
      bytes_used = 4;
      break;
    case rmsSTRING:
      bytes_used = xlat->stringlen;

      /*
       * Since all modbus registers are 2 bytes, if the string length is odd then we
       * need an extra byte to round out to a register boundary
       */
      if ((xlat->stringlen & 0x01) != 0)
        bytes_used = bytes_used + 1;
      break;
    default:
      bytes_used = -1;
      break;
  }
  return bytes_used;
}


/** Convert integer modbus register to string */
static void rinmodbus_regint_to_string(char *buffer, size_t buf_len, uint32_t data, bool is_signed)
{
  if (is_signed)
    snprintf(buffer, buf_len, "%ld", (long) data);
  else
    snprintf(buffer, buf_len, "%lu", (unsigned long) data);
}


/** Convert string modbus register to string */
static void rinmodbus_regstr_to_string(char *buffer, size_t buf_len, const uint8_t *data, size_t data_len, int stringlen)
{
  int i;

  if (buf_len > 0)
  {
    buffer[0] = '"';
    for (i = 0; i < stringlen && i < buf_len - 3 && i < data_len; i++)
    {
      if ('\0' == data[i])
        break;
      buffer[i + 1] = data[i];
    }
    buffer[i + 1] = '"';
    buffer[i + 2] = '\0';
  }
}


static void rinmodbus_reg_to_param32(const uint8_t *data, uint32_t *data32, enum eRinModbusEndian mb_endian)
{
  if (RINMODBUS_ENDIAN_BIG == mb_endian)
  {
    /* Big endian: Modbus defacto standard */
    *data32 = (MODBUS_GET_INT16_FROM_INT8(data, 0) << 16) | MODBUS_GET_INT16_FROM_INT8(data, 2);
  }
  else
  {
    /* Little endian: HBM standard, unpack 32bit numbers as per Open_ModbusTCP_Standard.pdf Section B.2.2 IEC-1131 data types */
    *data32 = MODBUS_GET_INT16_FROM_INT8(data, 0) | (MODBUS_GET_INT16_FROM_INT8(data, 2) << 16);
  }
}


static bool rinmodbus_reg_to_string(const struct s_rinmodbus_translate *const xlat, const uint8_t *data, size_t data_len, char *buffer, size_t buf_len, enum eRinModbusEndian mb_endian)
{
  bool result;
  uint32_t data32 = 0;

  result = true;
  switch (xlat->regtype)
  {
    case rmsSINT16:
      rinmodbus_regint_to_string(buffer, buf_len, MODBUS_GET_INT16_FROM_INT8(data, 0), true);
      break;
    case rmsUINT16:
      rinmodbus_regint_to_string(buffer, buf_len, MODBUS_GET_INT16_FROM_INT8(data, 0), false);
      break;
    case rmsSINT32:
      /* Unpack 32bit numbers as per Open_ModbusTCP_Standard.pdf Section B.2.2 IEC-1131 data types */
      rinmodbus_reg_to_param32(data, &data32, mb_endian);
      rinmodbus_regint_to_string(buffer, buf_len, data32, true);
      break;
    case rmsUINT32:
      /* Unpack 32bit numbers as per Open_ModbusTCP_Standard.pdf Section B.2.2 IEC-1131 data types */
      rinmodbus_reg_to_param32(data, &data32, mb_endian);
      rinmodbus_regint_to_string(buffer, buf_len, data32, false);
      break;
    case rmsSTRING:
      rinmodbus_regstr_to_string(buffer, buf_len, data, data_len, xlat->stringlen);
      break;
    default:
      result = false;
      break;
  }
  return result;
}


static void rinmodbus_paramstr_to_reg(const char *str, uint8_t *outdata, size_t reglength)
{
  size_t i;
  size_t strlength;

  /*
   * Since all modbus registers are 2 bytes, if the string length is odd then we
   * need an extra byte to round out to a register boundary
   */
  if ((reglength & 0x01) != 0)
    reglength = reglength + 1;

  /* Copy string across and zero out additional data */
  strlength = strlen(str);
  for (i = 0; i < reglength; i++)
  {
    if (i < strlength)
      outdata[i] = str[i];
    else
      outdata[i] = '\0';
  }
}


static void rinmodbus_param32_to_reg(uint8_t *outdata, uint16_t data16_lo, uint16_t data16_hi, enum eRinModbusEndian mb_endian)
{
  if (RINMODBUS_ENDIAN_BIG == mb_endian)
  {
    /* Big endian: Modbus defacto standard */
    MODBUS_SET_INT16_TO_INT8(outdata, 0, data16_hi);
    MODBUS_SET_INT16_TO_INT8(outdata, 2, data16_lo);
  }
  else
  {
    /* Little endian: HBM standard, pack 32bit numbers as per Open_ModbusTCP_Standard.pdf Section B.2.2 IEC-1131 data types */
    MODBUS_SET_INT16_TO_INT8(outdata, 0, data16_lo);
    MODBUS_SET_INT16_TO_INT8(outdata, 2, data16_hi);
  }
}


/** Convert parsed response to HBM query into modbus registers */
static bool rinmodbus_param_to_reg(const struct s_rinmodbus_translate *const xlat, struct s_comms_params *cp, uint8_t *outdata, size_t outdata_len, enum eRinModbusEndian mb_endian)
{
  bool result;
  uint16_t data16_lo;
  uint16_t data16_hi;
  int8_t param;

  /* Work out which is the read parameter */
  param = (xlat->readparam >= 0) ? xlat->readparam : xlat->param;

  result = true;
  switch (xlat->regtype)
  {
    case rmsSINT16:
      if (!param_ok(cp, param, INT16_MAX, INT16_MIN))
          result = false;
      else
        MODBUS_SET_INT16_TO_INT8(outdata, 0, (int16_t) cp->params[param]);
      break;
    case rmsUINT16:
      if (!param_ok(cp, param, UINT16_MAX, 0))
          result = false;
      else
        MODBUS_SET_INT16_TO_INT8(outdata, 0, (uint16_t) cp->params[param]);
      break;
    case rmsSINT32:
      if (!param_ok(cp, param, INT32_MAX, INT32_MIN))
          result = false;
      else
      {
        data16_lo = cp->params[param] & 0x0000FFFF;
        data16_hi = (cp->params[param] >> 16) & 0x0000FFFF;
        rinmodbus_param32_to_reg(outdata, data16_lo, data16_hi, mb_endian);
      }
      break;
    case rmsUINT32:
      /* While we really should use UINT32_MAX here, the signed logic of param_ok() & get_params() means it doesn't work correctly */
      if (!param_ok(cp, param, INT32_MAX, 0))
          result = false;
      else
      {
        /* Pack 32bit numbers as per Open_ModbusTCP_Standard.pdf Section B.2.2 IEC-1131 data types */
        data16_lo = cp->params[param] & 0x0000FFFF;
        data16_hi = (cp->params[param] >> 16) & 0x0000FFFF;
        rinmodbus_param32_to_reg(outdata, data16_lo, data16_hi, mb_endian);
      }
      break;
    case rmsSTRING:
      if (!param_str_ok(cp, param, xlat->stringlen, 0))
        result = false;
      else
        rinmodbus_paramstr_to_reg(cp->param_str[param], outdata, xlat->stringlen);
      break;
    default:
      result = false;
      break;
  }
  return result;
}


/** Converts a modbus register into a HBM command string. Returns the number of input data bytes used */
static int rinmodbus_reg_to_cmd(const struct s_rinmodbus_translate *const xlat, const uint8_t *data, size_t data_len, char *buffer, size_t buf_len, enum eRinModbusEndian mb_endian)
{
  int max;
  int i;
  int bytes_used = -1;
  size_t posn;
  ssize_t space_left;

  /* Work out how many bytes used by this register */
  bytes_used = rinmodbus_reg_bytes_used(xlat);
  if (bytes_used < 0)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_reg_to_cmd: invalid register type (%d) for register %d\n", (int) xlat->regtype, xlat->regnum);
    return -1;
  }
  if ((size_t) bytes_used > data_len)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_reg_to_cmd: insufficient data supplied for register %d (requires %d, supplied %u)\n", xlat->regnum, bytes_used, (unsigned) data_len);
    return -1;
  }

  /* Copy command to buffer */
  strncpy(buffer, xlat->cmd, buf_len);
  buffer[buf_len - 1] = '\0';

  /* Add index (if any) and parameter */
  max = (xlat->index_param > xlat->param) ? xlat->index_param : xlat->param;
  for (i = 0; i < max + 1; i++)
  {
    if (xlat->param == i)
    {
      posn = strlen(buffer);
      space_left = buf_len - posn - 1;
      if (!rinmodbus_reg_to_string(xlat, data, data_len, &buffer[posn], space_left, mb_endian))
      {
        WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_reg_to_cmd: rinmodbus_reg_to_string() failed for register %d\n", xlat->regnum);
        bytes_used = -1;
        break;
      }
    }
    else if (xlat->index_param == i)
    {
      posn = strlen(buffer);
      space_left = buf_len - posn - 1;
      snprintf(&buffer[posn], space_left, "%u", (unsigned) xlat->index);
    }

    if (i < max)
      strncat(buffer, ",", buf_len);
  }

  /* Add terminator */
  strncat(buffer, ";", buf_len);

  return bytes_used;
}


/** Converts HBM query response into a modbus register (or registers). Returns the number of output data bytes used */
static int rinmodbus_query_to_reg(const struct s_rinmodbus_translate *const xlat, struct s_rinmodbus_interp *interp, uint8_t *outdata, size_t outdata_len, enum eRinModbusEndian mb_endian)
{
  int bytes_used;

  /* Work out how many bytes used by this register */
  bytes_used = rinmodbus_reg_bytes_used(xlat);
  if (bytes_used < 0)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_query_to_reg: invalid register type (%d) for register %d\n", (int) xlat->regtype, xlat->regnum);
    return -1;
  }
  if ((size_t) bytes_used > outdata_len)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_query_to_reg: insufficient room in response for register %d (requires %d, supplied %u)\n", xlat->regnum, bytes_used, (unsigned) outdata_len);
    return -1;
  }

  /* This is where any parse error longjmp will jump back to */
  if (setjmp(interp->comms_params.parse_error) != 0)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_query_to_reg: response parsing error\n");
    return -1;
  }

  /* Break apart the response (in buffer) */
  get_params(&interp->comms_params, interp->buffer);

  /* Put appropriate parameter from response into modbus reply */
  if (!rinmodbus_param_to_reg(xlat, &interp->comms_params, outdata, outdata_len, mb_endian))
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_query_to_reg: response parse error\n");
    return -1;
  }

  return bytes_used;
}


/** Read reply back from comms into interp->buffer */
static bool rinmodbus_read_reply(struct s_rinmodbus_interp *interp)
{
  long bytes_avail;
  long i;

  /* Find out how much data is available */
  bytes_avail = rfilelength(interp->cn_pipe);
  if (bytes_avail > (long) RINMODBUS_COMMS_BUFFER_LEN - 1 || bytes_avail <= 0)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_run_hbm_cmd: too many bytes to read back, no data or error (bytes available = %ld)\n", bytes_avail);
    return false;
  }

  /* Read it back */
  if (rread(interp->cn_pipe, interp->buffer, bytes_avail, 0) != bytes_avail)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_run_hbm_cmd: pipe read failed\n");
    return false;
  }

  /* Null terminate it */
  interp->buffer[bytes_avail] = '\0';

  /* Strip off CRLFs */
  for (i = bytes_avail - 1; i >= 0; i--)
  {
    if ('\r' == interp->buffer[i] || '\n' == interp->buffer[i])
      interp->buffer[i] = '\0';
    else
      break;
  }

  return true;
}


/** Convert a eCommsError to a modbus exception */
static int rinmodbus_comms_error_to_modbus_exception(enum eCommsError error)
{
  int result;

  switch (error)
  {
    /* No error */
    case MESS_OK:
    case REPLY_SENT:
      result = MODBUS_EXCEPTION_MAX;
      break;
    default:
    /*
    case MESS_ERR_MOTION:
    case MESS_ERR_RANGE:
    case MESS_ERR_SYS:
    case MESS_ERR_DEV:
    case MESS_ERROR:
    */
      result = MODBUS_EXCEPTION_ILLEGAL_DATA_VALUE;
      break;
  }
  return result;
}


/** Run a HBM command, returns a MODBUS_EXCEPTION_*, MODBUS_EXCEPTION_MAX on success */
int rinmodbus_run_hbm_cmd(struct s_comms_net *cn, struct t_io_dev *cn_write_pipe, const char *cmd)
{
  size_t len;
  enum eCommsError result;

  len = strlen(cmd);
  if (rwrite(cn_write_pipe, cmd, len, 0) != (int) len)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_run_hbm_cmd: pipe write failed\n");
    return MODBUS_EXCEPTION_SLAVE_OR_SERVER_FAILURE;
  }

  /* Run the command to process it */
  result = comms_net_handler(cn);
  return rinmodbus_comms_error_to_modbus_exception(result);
}


/** Read a register by converting mb_register to a HBM comms query and executing it, then parsing the output for what we need */
static int rinmodbus_register_read(const struct s_rinmodbus_translate *xlat_table, struct s_rinmodbus_interp *interp, uint8_t *outdata, size_t *data_pos, size_t *data_len, uint16_t mb_register, enum eRinModbusEndian mb_endian)
{
  const struct s_rinmodbus_translate *xlat;
  int mb_result;
  int bytes_used;

  /* Lookup translation in the table */
  xlat = rinmodbus_lookup_register(xlat_table, mb_register);
  if (NULL == xlat)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_register_read: lookup of modbus register %u did not find a matching entry\n", (unsigned) mb_register);
    return MODBUS_EXCEPTION_ILLEGAL_DATA_ADDRESS;
  }

  /* Check if we have read permission */
  if ((xlat->permis & RMPM_READ) == 0)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_register_read: no read permission for register %u\n", (unsigned) mb_register);
    return MODBUS_EXCEPTION_ILLEGAL_FUNCTION;
  }

  /* Convert modbus register into HBM command */
  rinmodbus_reg_to_query(xlat, interp->buffer, RINMODBUS_COMMS_BUFFER_LEN);

  /* Clear out pipe from cn receive buffer */
  if (rioctl(interp->cn_pipe, IOCTL_SER_FLUSH_RX_BUF) < 0)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_register_read: clear of pipe receive buffer failed\n");
    return MODBUS_EXCEPTION_SLAVE_OR_SERVER_FAILURE;
  }

  /* Run the query to retrieve the data */
  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_register_read: read of modbus register %u translated to query '%s'\n", (unsigned) mb_register, interp->buffer);
  mb_result = rinmodbus_run_hbm_cmd(interp->cn, interp->cn_pipe, interp->buffer);
  if (mb_result != MODBUS_EXCEPTION_MAX)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_register_read: running HBM query failed\n");
    return mb_result;
  }

  /* Collect the data into buffer */
  if (!rinmodbus_read_reply(interp))
  {
    return MODBUS_EXCEPTION_SLAVE_OR_SERVER_FAILURE;
  }

  /* Parse the response */
  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_register_read: HBM query response was '%s'\n", interp->buffer);
  bytes_used = rinmodbus_query_to_reg(xlat, interp, &outdata[*data_pos], *data_len, mb_endian);
  if (bytes_used < 0)
  {
    return MODBUS_EXCEPTION_SLAVE_OR_SERVER_FAILURE;
  }

  *data_pos += bytes_used;
  *data_len -= bytes_used;

  return MODBUS_EXCEPTION_MAX;
}


/** Write a register by converting mb_register to a HBM comms command and executing it */
static int rinmodbus_register_write(const struct s_rinmodbus_translate *xlat_table, struct s_rinmodbus_interp *interp, const uint8_t *data, size_t *data_pos, size_t *data_len, uint16_t mb_register, enum eRinModbusEndian mb_endian)
{
  const struct s_rinmodbus_translate *xlat;
  int mb_result;
  int bytes_used;

  /* Lookup translation in the table */
  xlat = rinmodbus_lookup_register(xlat_table, mb_register);
  if (NULL == xlat)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_register_write: lookup of modbus register %u did not find a matching entry\n", (unsigned) mb_register);
    return MODBUS_EXCEPTION_ILLEGAL_DATA_ADDRESS;
  }

  /* Check if we have write permission */
  if ((xlat->permis & RMPM_WRITE) == 0)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_register_write: no write permission for register %u\n", (unsigned) mb_register);
    return MODBUS_EXCEPTION_ILLEGAL_FUNCTION;
  }

  /* Convert modbus register into HBM command */
  bytes_used = rinmodbus_reg_to_cmd(xlat, &data[*data_pos], *data_len, interp->buffer, RINMODBUS_COMMS_BUFFER_LEN, mb_endian);
  if (bytes_used < 0)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "rinmodbus_register_write: translation of modbus register %u to HBM command failed\n", (unsigned) mb_register);
    return MODBUS_EXCEPTION_SLAVE_OR_SERVER_FAILURE;
  }
  *data_pos += bytes_used;
  *data_len -= bytes_used;

  /* Run the command to process it */
  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_register_write: write of modbus register %u translated to '%s'\n", (unsigned) mb_register, interp->buffer);
  mb_result = rinmodbus_run_hbm_cmd(interp->cn, interp->cn_pipe, interp->buffer);
  return mb_result;
}


/** Implement the Modbus read registers command */
void rinmodbus_read_registers(const struct s_rinmodbus_translate *xlat_table, modbus_t *mb_ctx, struct s_rinmodbus_interp *interp, size_t query_length, enum eRinModbusEndian mb_endian)
{
  uint16_t mb_first_register;
  uint16_t mb_num_registers;
  int header_length;
  size_t data_pos_old;
  size_t data_pos;
  size_t data_len;
  int result;
  int i;

  /* Work out what we are reading */
  header_length = modbus_get_header_length(mb_ctx);
  mb_first_register = MODBUS_GET_INT16_FROM_INT8(interp->query, header_length + 1);
  mb_num_registers = MODBUS_GET_INT16_FROM_INT8(interp->query, header_length + 3);
  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_read_registers: read of registers from %u for %u\n", (unsigned) mb_first_register, (unsigned) mb_num_registers);

  /* Setup reply header, we only fill in data, function filled in by modbus_send_raw_reply() */
  interp->response[0] = (uint8_t) (mb_num_registers * sizeof(uint16_t));

  data_pos = 1;
  data_len = MODBUS_TCP_MAX_ADU_LENGTH - data_pos;
  result = MODBUS_EXCEPTION_MAX;
  for (i = mb_first_register; i < mb_first_register + mb_num_registers; )
  {
    data_pos_old = data_pos;
    result = rinmodbus_register_read(xlat_table, interp, interp->response, &data_pos, &data_len, i, mb_endian);
    if (MODBUS_EXCEPTION_MAX != result)
    {
      break;
    }
    i += (data_pos - data_pos_old) / sizeof(uint16_t);
  }

  if (MODBUS_EXCEPTION_MAX == result)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_read_registers: all rinmodbus_register_read()s successful, replying\n");
    modbus_send_raw_reply(mb_ctx, interp->query, query_length, interp->response, data_pos);
  }
  else
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_read_registers: rinmodbus_register_read() failed, replying with exception\n");
    modbus_reply_exception(mb_ctx, interp->query, result);
  }
}


/** Implement the Modbus read single register command */
void rinmodbus_write_single_register(const struct s_rinmodbus_translate *xlat_table, modbus_t *mb_ctx, struct s_rinmodbus_interp *interp, size_t query_length, enum eRinModbusEndian mb_endian)
{
  uint16_t mb_register;
  int header_length;
  int result;
  size_t data_pos;
  size_t data_len;

  header_length = modbus_get_header_length(mb_ctx);
  mb_register = MODBUS_GET_INT16_FROM_INT8(interp->query, header_length + 1);

  data_pos = header_length + 3;
  data_len = query_length - data_pos;
  result = rinmodbus_register_write(xlat_table, interp, interp->query, &data_pos, &data_len, mb_register, mb_endian);
  if (MODBUS_EXCEPTION_MAX == result)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_write_single_register: rinmodbus_register_write() success, replying with data\n");

    /* We only fill in data, function filled in by modbus_send_raw_reply() */
    MODBUS_SET_INT16_TO_INT8(interp->response, 0, mb_register);
    MODBUS_SET_INT16_TO_INT8(interp->response, 2, MODBUS_GET_INT16_FROM_INT8(interp->query, header_length + 3));
    modbus_send_raw_reply(mb_ctx, interp->query, query_length, interp->response, 4);
  }
  else
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_write_single_register: rinmodbus_register_write() failed, replying with exception\n");
    modbus_reply_exception(mb_ctx, interp->query, result);
  }
}


/** Implement the Modbus write multiple registers command */
void rinmodbus_write_multiple_registers(const struct s_rinmodbus_translate *xlat_table, modbus_t *mb_ctx, struct s_rinmodbus_interp *interp, size_t query_length, enum eRinModbusEndian mb_endian)
{
  uint16_t mb_first_register;
  uint16_t mb_num_registers;
  int header_length;
  size_t data_pos_old;
  size_t data_pos;
  size_t data_len;
  int result;
  int i;

  header_length = modbus_get_header_length(mb_ctx);
  mb_first_register = MODBUS_GET_INT16_FROM_INT8(interp->query, header_length + 1);
  mb_num_registers = MODBUS_GET_INT16_FROM_INT8(interp->query, header_length + 3);
  /* Byte count at header_length + 5 */

  WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_write_multiple_registers: write of registers from %u for %u\n", (unsigned) mb_first_register, (unsigned) mb_num_registers);

  data_pos = header_length + 6;
  data_len = query_length - data_pos;
  result = MODBUS_EXCEPTION_MAX;
  for (i = mb_first_register; i < mb_first_register + mb_num_registers; )
  {
    data_pos_old = data_pos;
    result = rinmodbus_register_write(xlat_table, interp, interp->query, &data_pos, &data_len, i, mb_endian);
    if (MODBUS_EXCEPTION_MAX != result)
    {
      break;
    }
    i += (data_pos - data_pos_old) / sizeof(uint16_t);
  }

  if (MODBUS_EXCEPTION_MAX == result)
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_write_multiple_registers: all rinmodbus_register_write()s successful, replying\n");

    /* We only fill in data, function filled in by modbus_send_raw_reply() */
    MODBUS_SET_INT16_TO_INT8(interp->response, 0, mb_first_register);
    MODBUS_SET_INT16_TO_INT8(interp->response, 2, mb_num_registers);
    modbus_send_raw_reply(mb_ctx, interp->query, query_length, interp->response, 4);
  }
  else
  {
    WE_MODBUS_TASK_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "rinmodbus_write_multiple_registers: rinmodbus_register_write() failed, replying with exception\n");
    modbus_reply_exception(mb_ctx, interp->query, result);
  }
}
