#!/usr/bin/python
'''
modregtoc - Convert CSV Modbus register definitions to .c/.h/.py/.tex

Created on 09/09/2014

@author: IanB
'''

import argparse
import csv
import sys
import re
import os


def permis_to_str(permis):
    string = ""
    if ("r" in permis):
        string = string + "RMPM_READ"
    if ("w" in permis):
        if (len(string) > 0):
            string = string + " | "
        string = string + "RMPM_WRITE"
    if (len(string) == 0):
        string = "0x00"
    return string


def register_to_c_struct(reg):

    # Construct the structure element
    cstruct = "  {"
    cstruct = cstruct + " .regtype = rms{},".format(reg["type"])
    cstruct = cstruct + " .regnum = {},".format(reg["register"])
    if (reg["cmd"] is not None):
        cstruct = cstruct + " .cmd = \"{}\",".format(reg["cmd"])
    else:
        cstruct = cstruct + " .cmd = NULL,"
    cstruct = cstruct + " .index_param = {},".format(reg["index_param"])
    cstruct = cstruct + " .index = {},".format(reg["index"])
    cstruct = cstruct + " .param = {},".format(reg["param"])
    cstruct = cstruct + " .readparam = {},".format(reg["readparam"])
    cstruct = cstruct + " .permis = {}".format(permis_to_str(reg["permis"]))

    if (reg["type"] == "STRING"):
        cstruct = cstruct + ", .stringlen = {} ".format(reg["size"])
    else:
        cstruct = cstruct + " "

    cstruct = cstruct + "},"

    cstruct = cstruct.ljust(130)

    # Construct the comment
    if (reg["index"] != -1):
        index_str = " index {}".format(reg["index"])
    else:
        index_str = ""
    if (reg["param"] != -1):
        param_str = " param {}".format(reg["param"])
    else:
        param_str = ""
    cstruct = cstruct + "     /* {}{}{} - {} */".format(reg["cmd"], index_str, param_str, reg["description"])

    return cstruct

def type_to_size(typestr):
    """ Convert type from CSV file into a size in bytes """
    if (("UINT8" == typestr) or ("SINT8" == typestr) or ("UINT16" == typestr) or ("SINT16" == typestr)):
        size = 2
    elif (("UINT32" == typestr) or ("SINT32" == typestr)):
        size = 4
    else:
        m = re.match(pattern="(STRING)\(([0-9]+)\)", string=typestr)
        if (m is not None and m.group(1) is not None and m.group(2) is not None):
            typestr = m.group(1)
            size = int(m.group(2))
        else:
            raise Exception("Unknown type \"{}\"".format(typestr))
    return typestr, size

def csvrow_to_int(row, default=-1):
    if ("" == row):
        rowint = default
    else:
        rowint = int(row)
    return rowint

def csvrow_to_dict(csvrow):
    """ Convert a CSV row into a dict for more understandable handling """
    mbreg = int(csvrow[0])
    typestr, size = type_to_size(csvrow[1])
    hbmcmd = csvrow[2]
    param = csvrow_to_int(csvrow[3])
    index_param = csvrow_to_int(csvrow[4])
    index = csvrow_to_int(csvrow[5])
    readparam = csvrow_to_int(csvrow[6])
    description = csvrow[7]
    permis = csvrow[8]
    minimum = csvrow_to_int(csvrow[9], 0)
    maximum = csvrow_to_int(csvrow[10], 0)
    regex = csvrow[11]
    simple = csvrow[12]
    regname = csvrow[13]

    mbreg = { "register": mbreg, "type": typestr, "size": size, "cmd": hbmcmd, "param": param, "index_param": index_param, \
              "index": index, "readparam": readparam, "description": description, "permis": permis, "minimum": minimum, "maximum": maximum, \
              "regex": regex, "simple": simple, "regname": regname }
    return mbreg


def termination_reg():
    mbreg = { "register": -1, "type": "UNKNOWN", "size": -1, "cmd": None, "param": -1, "index_param": -1, \
              "index": -1, "readparam": -1, "description": "Last entry", "permis": "", "minimum": 0, "maximum": 0, \
              "regex": "", "simple": "", "regname": "" }
    return mbreg


def check_register_overlap(mb_registers):
    """ Ensure that none of the registers overlap """
    for reg in mb_registers:
        reg_low = reg["register"]
        reg_high = reg["register"] + (reg["size"] // 2) - 1
        for reg2 in mb_registers:
            reg2_low = reg2["register"]
            reg2_high = reg2["register"] + (reg2["size"] // 2) - 1
            if (reg != reg2) and ((reg2_low in range(reg_low, reg_high + 1)) or (reg2_high in range(reg_low, reg_high + 1))):
                raise Exception("Registers {} ({}) and {} ({}) overlap".format(reg_low, reg["description"], reg2_low, reg2["description"]))


def import_registers_from_csv(csvfilename):
    """ Import the Modbus register definitions from a CSV file """
    with open(csvfilename, 'r') as csvfile:
        reader = csv.reader(csvfile)

        # Load registers into list
        mb_registers = []
        for i, row in enumerate(reader):
            if (i != 0):
                try:
                    mbreg = csvrow_to_dict(row)
                    mb_registers.append(mbreg)
                except:
                    e = sys.exc_info()[0]
                    print("Exception parsing line {}: {}: {}".format(i + 1, row, repr(e)))
        mb_registers.append(termination_reg())

    # Check that none overlap
    check_register_overlap(mb_registers=mb_registers)
    return mb_registers

def apply_register_offset(mb_registers, reg_offset):
    """The modbus specification has a PDU (Protocol Data Unit) and MDM (Modbus Data Model).
       Basically a register in the MDM is 1+ the register in the PDU. When generating
       documentation use the MDM, when writing code use PDU"""
    if (reg_offset == 'pdu'):
        offset = 0
    else:
        offset = 1
    for reg in mb_registers:
        if (reg["register"] >= 0):
            reg["register"] = reg["register"] + offset

def write_c_file(outfile, tablename, hfilename, mb_registers, infilename):
    """Write registers out to C file"""
    outfile.write("/* This file is automatically generated by modregtoc.py from {}. DO NOT EDIT. */\n".format(infilename))
    outfile.write("#include <stdio.h>\n")
    outfile.write("#include \"modbus_regs.h\"\n\n")
    hfile = os.path.basename(hfilename)
    outfile.write("#include \"{}\"\n\n".format(hfile))
    outfile.write("const struct s_rinmodbus_translate {}[] =\n".format(tablename))
    outfile.write("{\n")
    for reg in mb_registers:
        cstruct = register_to_c_struct(reg)
        outfile.write("{}\n".format(cstruct))
    outfile.write("};\n")

def reg_to_hash_define(mb_reg):
    """ Convert register to #define """
    regstr = "#define MB_REG_{}_{} {} ".format(mb_reg["cmd"], mb_reg["regname"], mb_reg["register"])
    regstr = regstr.ljust(60)

    typestr = mb_reg["type"]
    if (mb_reg["type"] == "STRING"):
        typestr = typestr + "({})".format(mb_reg["size"])
    regstr = regstr + "/*!< {} - {} */\n".format(typestr, mb_reg["description"])

    return regstr

def reg_to_py_const(mb_reg):
    """ Convert register to python constant """
    regstr = "MB_REG_{}_{} = {} ".format(mb_reg["cmd"], mb_reg["regname"], mb_reg["register"])
    regstr = regstr.ljust(60)

    typestr = mb_reg["type"]
    if (mb_reg["type"] == "STRING"):
        typestr = typestr + "({})".format(mb_reg["size"])
    regstr = regstr + "# {} - {}\n".format(typestr, mb_reg["description"])

    return regstr

def write_h_file(outfile, tablename, hfilename, mb_registers, infilename):
    """Write registers out to H file"""
    outfile.write("/* This file is automatically generated by modregtoc.py from {}. DO NOT EDIT. */\n".format(infilename))
    hname = "{}_H_".format(os.path.basename(os.path.splitext(hfilename)[0]).upper())
    outfile.write("#ifndef {}\n".format(hname))
    outfile.write("#define {} 1\n\n".format(hname))
    outfile.write("#include \"modbus_regs.h\"\n\n")

    for reg in mb_registers:
        if (reg["type"] != "UNKNOWN"):
            outfile.write(reg_to_hash_define(reg))

    outfile.write("\nextern const struct s_rinmodbus_translate {}[];\n\n".format(tablename))
    outfile.write("#endif /* {} */\n".format(hname))

def write_py_file(outfile, mb_registers, infilename):
    """Write registers out to python file"""
    outfile.write("\"\"\" This file is automatically generated by modregtoc.py from {}. DO NOT EDIT. \"\"\"\n\n".format(infilename))

    for reg in mb_registers:
        if (reg["type"] != "UNKNOWN"):
            outfile.write(reg_to_py_const(reg))

def reg_to_hbm_cmd(mb_reg):
    maxparam = max(mb_reg["param"], mb_reg["index_param"])
    paramstr = ""
    for param in range(0, maxparam + 1):
        if (param == mb_reg["param"]):
            paramstr = paramstr + "$<$value$>$"
        if (param == mb_reg["index_param"]):
            paramstr = paramstr + "{}".format(mb_reg["index"])
        paramstr = paramstr + ","
    hbm_cmd = "{}{};".format(mb_reg["cmd"], paramstr.rstrip(","))
    return hbm_cmd

def reg_to_hbm_query_cmd(mb_reg):
    paramstr = ""
    for param in range(0, mb_reg["index_param"] + 1):
        if (param == mb_reg["index_param"]):
            paramstr = paramstr + "{}".format(mb_reg["index"])
            break
        paramstr = paramstr + ","
    hbm_cmd = "{}?{};".format(mb_reg["cmd"], paramstr.rstrip(","))
    return hbm_cmd

def reg_to_hbm_query_resp(mb_reg):
    if (mb_reg["readparam"] != -1):
        paramstr = ""
        for param in range(0, mb_reg["readparam"] + 1):
            if (param == mb_reg["readparam"]):
                paramstr = paramstr + "$<$value$>$"
                break
            paramstr = paramstr + ","
        hbm_cmd = "{}".format(paramstr.rstrip(","))
    else:
        maxparam = max(mb_reg["param"], mb_reg["index_param"])
        paramstr = ""
        for param in range(0, maxparam + 1):
            if (param == mb_reg["param"]):
                paramstr = paramstr + "$<$value$>$"
            if (param == mb_reg["index_param"]):
                paramstr = paramstr + "{}".format(mb_reg["index"])
            paramstr = paramstr + ","
        hbm_cmd = "{}".format(paramstr.rstrip(","))
    return hbm_cmd

def write_tex_header(outfile, infilename):
    """Write the latex header"""
    outfile.write("%% This file is automatically generated by modregtoc.py from {}. DO NOT EDIT.\n\n".format(infilename))
    outfile.write("\\begin{longtabu}{ | l | l | l | l | l | l | X | }\n")
    outfile.write("\\hline\n\n")
    outfile.write("Register & Name & Type & Write & Read & Read Response & Description\\\\ \\hline\n")
    outfile.write("\\endhead\n\n")

def reg_to_tex_table_row(mb_reg):
    """ Convert register to latex table row """
    if ("w" in mb_reg["permis"]):
        hbm_cmd = reg_to_hbm_cmd(mb_reg=mb_reg)
    else:
        hbm_cmd = ""
    if ("r" in mb_reg["permis"]):
        hbm_query_cmd = reg_to_hbm_query_cmd(mb_reg=mb_reg)
        hbm_query_resp = reg_to_hbm_query_resp(mb_reg=mb_reg)
    else:
        hbm_query_cmd = ""
        hbm_query_resp = ""
    if (mb_reg["type"] == "STRING"):
        regtype = "STR({})".format(mb_reg["size"])
    else:
        regtype = mb_reg["type"]

    regname = mb_reg["regname"].replace("_", "\_")
    regstr = "{} & {} & {} & {} & {} & {} & \\raggedright {}\\\\ \\hline\n".format(mb_reg["register"], regname, regtype,
                                                                 hbm_cmd, hbm_query_cmd, hbm_query_resp, mb_reg["description"])
    return regstr, int(mb_reg["register"])

def write_tex_footer(outfile):
    """Write the latex footer"""
    outfile.write("\\end{longtabu}\n\n")

def write_tex_file(outfile, mb_registers, infilename):
    """Write registers out to latex file"""
    write_tex_header(outfile=outfile, infilename=infilename)

    dataDict = {}

    for reg in mb_registers:
        if (reg["type"] != "UNKNOWN"):
            # Save the data into a table so it can be read later (in order)
            str, id = reg_to_tex_table_row(mb_reg=reg)
            dataDict[id] = str

    # Print the data ordered by id.
    for key, str in sorted(dataDict.items()):
        outfile.write(str)

    write_tex_footer(outfile)

def main(infilename, outputtype, outputfilename, tablename, reg_offset):
    mb_registers = import_registers_from_csv(csvfilename=infilename)
    apply_register_offset(mb_registers=mb_registers, reg_offset=reg_offset)

    hfilename = os.path.splitext(infilename)[0] + ".h"
    with open(outputfilename, 'w') as outfile:
        if (outputtype == 'c'):
            write_c_file(outfile=outfile, tablename=tablename, hfilename=hfilename, mb_registers=mb_registers, infilename=infilename)
        elif (outputtype == 'h'):
            write_h_file(outfile=outfile, tablename=tablename, hfilename=hfilename, mb_registers=mb_registers, infilename=infilename)
        elif (outputtype == "py"):
            write_py_file(outfile=outfile, mb_registers=mb_registers, infilename=infilename)
        elif (outputtype == "tex"):
            write_tex_file(outfile=outfile, mb_registers=mb_registers, infilename=infilename)


if (__name__ == '__main__'):
    # Parse arguments
    parser = argparse.ArgumentParser(description='modregtoc - Convert CSV Modbus register definitions to .c/.h/.py/.tex')
    parser.add_argument('infile', action='store',
                        help='The input CSV file')
    parser.add_argument('--output-type', action='store', choices=['c', 'h', 'py', 'tex'], required=True,
                        help='The type of output file to generate')
    parser.add_argument('--output-file', action='store',
                        help='The file to write to. If not specified, will be <infile>.<output-type>.')
    parser.add_argument('--table-name', action='store', default=None,
                        help='The .c/.h table name to create. If not specified, will be <infile>.')
    parser.add_argument('--register-offset', action='store', choices=['pdu', 'mdm'], default='pdu',
                        help='The register offset to apply. PDU is the Protocol Data Unit (+0). MDM is the Modbus Data Model (+1). Default is PDU.')
    args = parser.parse_args()

    # Generate output file name
    infilename = args.infile

    if (args.output_file is not None):
        outfilename = args.output_file
    else:
        outfilename = os.path.splitext(infilename)[0] + "." + args.output_type

    if (args.table_name is not None):
        tablename = args.table_name
    else:
        tablename = os.path.basename(os.path.splitext(infilename)[0])

    main(infilename=infilename, outputtype=args.output_type, outputfilename=outfilename, tablename=tablename, reg_offset=args.register_offset)
