#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stddef.h>
#include <pthread.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include "meas.h"
#include "we_config.h"
#include "we_debug.h"

#include "nv_ram.h"


/** File descriptor to access hardware */
static int nvram_dev_fd = -1;


/** Read data from the NVRAM hardware device */
bool nvram_hw_read(void *buffer, off_t offset, size_t length)
{
  uint8_t *ptr;
  size_t bytes_read;
  int result;

  /* Check device is open */
  if (nvram_dev_fd < 0)
  {
    return false;
  }

  /* Seek to the position we want to read */
  result = lseek(nvram_dev_fd, offset, SEEK_SET);
  if (result < 0)
  {
    return false;
  }

  /* Read the data */
  for (bytes_read = 0, ptr = buffer; bytes_read < length; )
  {
    /* Do the read */
    result = read(nvram_dev_fd, ptr, length - bytes_read);
    if (result < 0)
    {
      return false;
    }
    bytes_read += result;
    ptr += result;
  }

  /* All done successfully */
  return true;
}


/** Write data to the NVRAM hardware device */
bool nvram_hw_write(const void *buffer, off_t offset, size_t length)
{
  const uint8_t *ptr;
  size_t bytes_written;
  int result;

  /* Check device is open */
  if (nvram_dev_fd < 0)
  {
    return false;
  }

  /* Seek to the position we want to write */
  result = lseek(nvram_dev_fd, offset, SEEK_SET);
  if (result < 0)
  {
    return false;
  }

  /* Write the data */
  for (bytes_written = 0, ptr = buffer; bytes_written < length; )
  {
    /* Do the write */
    result = write(nvram_dev_fd, ptr, length - bytes_written );
    if (result < 0)
    {
      return false;
    }
    bytes_written += result;
    ptr += result;
  }

  /* All done successfully */
  return true;
}


/**
 * Initialise access to the NVRAM, and check the contents for valid data
 * Returns false if any part of the NVRAM had to be defaulted
 */
bool nvram_init(void)
{
  bool result;
  int fd;
  const char *device;
  int errno_dev1;

  /* Try to open the first NVRAM device */
  device = we_configuration.nvram_device1;
  fd = open(device, O_RDWR, S_IRUSR | S_IWUSR);
  if (fd < 0)
  {
    errno_dev1 = errno;

    /* That failed, try the second device (which should work if we are running on new hardware) */
    device = we_configuration.nvram_device2;
    fd = open(device, O_RDWR, S_IRUSR | S_IWUSR);
    if (fd < 0)
    {
      rdebug_printf(LOG_ERR, "nvram_init: attempt to open both nvram devices (\"%s\", \"%s\") failed (errno1: %d, errno2: %d)\n",
                    we_configuration.nvram_device1, we_configuration.nvram_device2, errno_dev1, errno);
      result = false;
      goto ni_fail;
    }
  }
  if (fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "nvram_init: error setting FD_CLOEXEC on nvram device \"%s\" (errno %d)\n", device, errno);
    result = false;
    goto ni_fail;
  }
  nvram_dev_fd = fd;
  result = true;

ni_fail:
  if (!result && fd >= 0)
  {
    close(fd);
    fd = -1;
    nvram_dev_fd = -1;
  }

  return result;
}


/** Clean up the nvram */
void nvram_destroy(void)
{
  /* Close file */
  if (nvram_dev_fd >= 0)
  {
    close(nvram_dev_fd);
    nvram_dev_fd = -1;
  }
}
