/*
 * remote_cal.c
 *
 *  Created on: 19/05/2011
 *      Author: ian
 */
#include <rinweigh/adcinterface.h>
#include "meas.h"
#include "iface_cal.h"
#include "display.h"
#include "iface.h"
#include "adcthread.h"
#include "commonstrings.h"
#include "database.h"

#include "remote_cal.h"


static enum eRemoteCalMode calibrating = REMOTE_CAL_MODE_IDLE;
static enum eRemoteCalResult last_cal_result = REMOTE_CAL_RESULT_NO_ERROR;
static int32s remote_cal_param1;
static uint8_t last_cal_output = ADC_OP_STATUS_SUCCESS;


/** Sets the result of the last calibration, see ADC_OP_STATUS_* constants */
void set_last_cal_output(uint8_t res)
{
  last_cal_output = res;
}


/** Gets the result of the last calibration, see ADC_OP_STATUS* constants  */
uint8_t get_last_cal_output(void)
{
  return last_cal_output;
}


/** Sets the result of the last calibration, see REMOTE_CAL_* constants */
void set_last_cal_result(enum eRemoteCalResult res)
{
  last_cal_result = res;
}


/** Gets the result of the last calibration, see REMOTE_CAL_* constants  */
enum eRemoteCalResult get_last_cal_result(void)
{
  return last_cal_result;
}


/** Returns the current calibration mode */
enum eRemoteCalMode get_remote_cal_mode(void)
{
  return calibrating;
}


/** Sets the current calibration mode */
void set_remote_cal_mode(enum eRemoteCalMode mode)
{
  calibrating = mode;
}


/** Call when calibration is complete */
void finish_calibration(void)
{
  set_remote_cal_mode(REMOTE_CAL_MODE_IDLE);
}


/** Called by the comms thread to kick off a zero calibration */
void setup_for_zero_cal(void)
{
  set_remote_cal_mode(REMOTE_CAL_MODE_ZERO_NORMAL);
  set_last_cal_result(REMOTE_CAL_RESULT_NO_ERROR);
}


/** Called by the comms thread to initiate a direct zero calibration. mvv is milli-volts per volt * 10000 (1mvv = 10000) */
void setup_for_zero_direct_cal(int32s mvv)
{
  set_remote_cal_mode(REMOTE_CAL_MODE_ZERO_DIRECT);
  remote_cal_param1 = mvv;
  set_last_cal_result(REMOTE_CAL_RESULT_NO_ERROR);
}


/** Called by the comms thread to kick off a span calibration */
void setup_for_span_cal(void)
{
  set_remote_cal_mode(REMOTE_CAL_MODE_SPAN_NORMAL);
  set_last_cal_result(REMOTE_CAL_RESULT_NO_ERROR);
}


/** Called by the comms thread to initiate a direct span calibration. mvv is milli-volts per volt * 10000 (1mvv = 10000) */
void setup_for_span_direct_cal(int32s mvv)
{
  set_remote_cal_mode(REMOTE_CAL_MODE_SPAN_DIRECT);
  remote_cal_param1 = mvv;
  set_last_cal_result(REMOTE_CAL_RESULT_NO_ERROR);
}


/** Wait for the calibration to complete */
static uint8_t wait_for_cal_result(bool span_not_zero)
{
  uint8_t res, adcRes;

  adcRes = WaitForAdcOperation(OS_TICKS_PER_SEC * 30);
  set_last_cal_output(adcRes);
  switch (adcRes)
  {
    case ADC_OP_STATUS_SUCCESS:
      res = REMOTE_CAL_RESULT_NO_ERROR;
      break;
    case ADC_OP_STATUS_FAILED_RES:
    case ADC_OP_STATUS_FAILED_BAND:
      res = (span_not_zero) ? REMOTE_CAL_RESULT_SPAN : REMOTE_CAL_RESULT_ZERO;
      break;
    case ADC_OP_STATUS_IN_PROGRESS:
    case ADC_OP_STATUS_FAILED_PNT_DUPL:
    default:
      res = REMOTE_CAL_RESULT_ERROR;
      break;
  }
  return res;
}


/** Perform the remote zero calibration */
static uint8_t calibrate_zero_normal(void)
{
  (void) AdcZeroCalibration(&adcSuperStructure, true, false);
  return wait_for_cal_result(false);
}


/** Perform the remote direct zero calibration */
static uint8_t calibrate_zero_direct(int32s mvv)
{
  (void) AdcDirectZeroCalibration(&adcSuperStructure, true, false, mvv);
  return wait_for_cal_result(false);
}


static uint8_t calibrate_span_normal(void)
{
  (void) AdcSpanCalibration(&adcSuperStructure, true, false, adcSuperStructure.adcUserDatabase.calWeight);
  return wait_for_cal_result(true);
}


static uint8_t calibrate_span_direct(int32s mvv)
{
  int32s calwt;

  calwt = true_full_scale();
  (void) AdcDirectSpanCalibration(&adcSuperStructure, true, false, calwt, mvv);
  return wait_for_cal_result(true);
}


/** Start the remote calibration */
static void start_remote_cal(const char *display_msg)
{
  Dclear_flashing();
  freeze_display(true);
  Dputstr(display_msg);
}


/** Finish the remote calibration */
static void finish_remote_cal(void)
{
  finish_calibration();
  freeze_display(false);
}


/**
 * Poll to check if a remote calibration has been requested
 * by the comms thread, and perform the calibration
 */
void check_remote_cal(void)
{
  switch (get_remote_cal_mode())
  {
    case REMOTE_CAL_MODE_IDLE:
      /* Nothing to do */
      break;

    case REMOTE_CAL_MODE_ZERO_NORMAL:
      start_remote_cal(zero_in_progress_str);
      set_last_cal_result(calibrate_zero_normal());
      finish_remote_cal();
      break;

    case REMOTE_CAL_MODE_SPAN_NORMAL:
      start_remote_cal(span_in_progress_str);
      set_last_cal_result(calibrate_span_normal());
      finish_remote_cal();
      break;

    case REMOTE_CAL_MODE_ZERO_DIRECT:
      start_remote_cal(zero_in_progress_str);
      set_last_cal_result(calibrate_zero_direct(remote_cal_param1));
      finish_remote_cal();
      break;

    case REMOTE_CAL_MODE_SPAN_DIRECT:
      start_remote_cal(span_in_progress_str);
      set_last_cal_result(calibrate_span_direct(remote_cal_param1));
      finish_remote_cal();
      break;

    default:
      /* Something weird going on */
      set_last_cal_result(REMOTE_CAL_RESULT_ERROR);
      finish_remote_cal();
      break;
  }
}
