/*
 * root_passwd.c
 *
 *  Created on: Aug 3, 2016
 *      Author: ian
 */
#include <stdbool.h>
#include <regex.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <ctype.h>
#define LTC_SHA256
#include <tomcrypt.h>
#include <rinexec/el_runcmd.h>
#include <rinexec/el_readpipe_bufalloc.h>
#include <rinexec/el_writepipe_buffer.h>
#include "we_debug.h"

#include "root_passwd.h"


#define SHADOW_LINE_LEN 200


#define SHADOW_REGEX_GROUPS 4
#define SHADOW_REGEX_POSTFIX ":\\$([0-9]+)\\$([^$]+)\\$([^:]+):.*$"
#define SHADOW_REGEX_USER_STR "^user" SHADOW_REGEX_POSTFIX
#define SHADOW_REGEX_ROOT_STR "^root" SHADOW_REGEX_POSTFIX


enum e_password_hash_type
{
  pht_des,
  pht_md5,
  pht_blowfish,
  pht_unused1,
  pht_unused2,
  pht_sha256,
  pht_sha512,

  pht_num_hash_types
};


static const char *const password_hash_types[pht_num_hash_types] =
{
  NULL,         /* "des", DES not supported */
  "md5",
  NULL,         /* Blowfish not supported */
  NULL,
  NULL,
  "sha256",     /* Yuck - on ubuntu this is sha-256, on busybox s[^ ]* */
  "sha512"      /* Yuck - on ubuntu this is sha-512, on busybox sha512 */
};


#define BUFFER_ADD_VAL32(buf, val) \
{ \
  *((uint32_t *) buf) = htonl(val); \
  buf += sizeof(uint32_t); \
}


static bool generate_root_password(const struct s_db_app_factory *facdb, char **generated_pw)
{
  #define RPW_BUFFER_SIZE (sizeof(uint32_t) * (2 + ADC_NUM_SYNC_FREQ * 6))
  uint8_t rpw_buffer[RPW_BUFFER_SIZE];
  uint8_t *ptr;
  size_t i;
  hash_state md;
  uint8_t *rpw_hash = NULL;
  char *rpw_base64 = NULL;
  unsigned long rpw_base64_max_len;

  /*
   * Stuff parts of factory db into a buffer:
   *
   * Serial number
   * zeroOffset
   * zeroDriftPos
   * zeroDriftNeg
   * spanCal
   * spanDriftPos
   * spanDriftNeg
   * calTemp
   */
  memset(rpw_buffer, 0, RPW_BUFFER_SIZE);
  ptr = rpw_buffer;
  BUFFER_ADD_VAL32(ptr, facdb->serial_no);
  BUFFER_ADD_VAL32(ptr, facdb->adc_factory_database.calTemp);
  for (i = 0; i < ADC_NUM_SYNC_FREQ; i++)
  {
    /* Mix things up a bit so it's not straight forward to figure out */
    BUFFER_ADD_VAL32(ptr, facdb->adc_factory_database.zeroOffset[i]);
    BUFFER_ADD_VAL32(ptr, facdb->adc_factory_database.spanDriftNeg[i]);
    BUFFER_ADD_VAL32(ptr, facdb->adc_factory_database.zeroDriftPos[ADC_NUM_SYNC_FREQ - 1 - i]);
    BUFFER_ADD_VAL32(ptr, facdb->adc_factory_database.spanCal[ADC_NUM_SYNC_FREQ - 1 - i]);
    BUFFER_ADD_VAL32(ptr, facdb->adc_factory_database.zeroDriftNeg[i]);
    BUFFER_ADD_VAL32(ptr, facdb->adc_factory_database.spanDriftPos[i]);
  }

  /* Use libtomcrypt to generate a SHA256 hash of the buffer */
  rpw_hash = malloc(sha256_desc.hashsize);
  if (NULL == rpw_hash)
  {
    goto grp_fail;
  }
  if (sha256_init(&md) != CRYPT_OK)
  {
    goto grp_fail;
  }
  if (sha256_process(&md, rpw_buffer, RPW_BUFFER_SIZE) != CRYPT_OK)
  {
    goto grp_fail;
  }
  if (sha256_done(&md, rpw_hash) != CRYPT_OK)
  {
    goto grp_fail;
  }

  /* Convert hash to base64 string that we will use as a password */
  rpw_base64_max_len = sha256_desc.hashsize + (sha256_desc.hashsize / 3) + 16;
  rpw_base64 = malloc(rpw_base64_max_len);
  if (NULL == rpw_base64)
  {
    goto grp_fail;
  }
  if (base64_encode(rpw_hash, sha256_desc.hashsize, (unsigned char *) rpw_base64, &rpw_base64_max_len) != CRYPT_OK)
  {
    goto grp_fail;
  }
  free(rpw_hash);
  rpw_hash = NULL;

  /* Return generated password */
  *generated_pw = rpw_base64;
  return true;

grp_fail:
  WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "generate_root_password: Failed to generate SHA hash\n");
  if (NULL != rpw_hash)
  {
    free(rpw_hash);
    rpw_hash = NULL;
  }
  if (NULL != rpw_base64)
  {
    free(rpw_base64);
    rpw_base64 = NULL;
  }
  return false;
}


/** Set a user password using chpasswd: echo 'username:newpassword' | chpasswd */
static bool set_password_for_username(const char *username, const char *newpasswd, enum e_password_hash_type hash_id)
{
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_buffer wcb;
  #define NUM_CHPASSWD_ARGS 4
  const char *chpasswd_cmd[NUM_CHPASSWD_ARGS];
  int ret_code;
  bool result;
  enum e_el_result res;
  size_t len;
  #define USER_PASS_STR_LEN 200
  char user_pass_str[USER_PASS_STR_LEN + 1];
  const char *hash_type;

  /* Work out hash type */
  hash_type = (hash_id < pht_num_hash_types) ? password_hash_types[hash_id] : NULL;
  if (NULL == hash_type)
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "set_password_for_username: unknown/unsupported hash id %d\n", hash_id);
    return false;
  }

  /* Generate chpasswd command like: chpasswd -c sha256 */
  chpasswd_cmd[0] = "chpasswd";
  chpasswd_cmd[1] = "-c";
  chpasswd_cmd[2] = hash_type;
  chpasswd_cmd[3] = NULL;

  /* Create string of user:password to send to chpasswd */
  snprintf(user_pass_str, USER_PASS_STR_LEN, "%s:%s", username, newpasswd);

  /* Don't send the '\0' from the end of the string (fw_setenv doesn't like them) */
  len = strlen(user_pass_str);

  /* Configure to set root passwd using "chpasswd" via it's stdin */
  el_init_runcmd_pipe_config(&rcpc);
  el_init_write_cmd_pipe_buffer(&rcpc, &wcb, user_pass_str, len);

  /* Run "chpasswd", pipe in the contents ofuser_pass_str, and wait for the command to complete */
  res = el_run_command(chpasswd_cmd[0], chpasswd_cmd, &ret_code, &rcpc, true, true, true);
  if (elr_success != res)
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "set_password_for_username: el_run_command(chpasswd) returned %d\n", res);
    result = false;
  }
  else if (!WIFEXITED(ret_code) || 0 != WEXITSTATUS(ret_code))
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "set_password_for_username: chpasswd failed, or return code not 0\n");
    result = false;
  }
  else
  {
    result = true;
  }

  memset(user_pass_str, 0, USER_PASS_STR_LEN);
  return result;
}


static bool parse_shadow_entry(regex_t *regex, const char *shadow_entry, int *hash_id, char salt[SHADOW_LINE_LEN], char pw_hash[SHADOW_LINE_LEN])
{
  #define CFG_STR_LEN 100
  regmatch_t regex_groups[SHADOW_REGEX_GROUPS];
  size_t group;
  char str[CFG_STR_LEN];
  size_t so;
  size_t eo;
  size_t len;
  bool result;

  /* Parse the shadow password entry against the regex, and pick it apart */
  if (regexec(regex, shadow_entry, SHADOW_REGEX_GROUPS, regex_groups, 0) == 0)
  {
    result = true;
    for (group = 0; group < SHADOW_REGEX_GROUPS; group++)
    {
      if (regex_groups[group].rm_so == -1)
      {
        WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "parse_shadow_entry: Not enough groups.\n");
        result = false;
        goto psc_done;
      }
      so = regex_groups[group].rm_so;
      eo = regex_groups[group].rm_eo;
      len = eo - so;
      if (len >= CFG_STR_LEN)
      {
        WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "parse_shadow_entry: Group %d too long.\n", group);
        result = false;
        goto psc_done;
      }
      memcpy(str, &shadow_entry[so], len);
      str[len] = '\0';

      switch (group)
      {
        case 0:
          break;
        case 1:
          *hash_id = atol(str);
          break;
        case 2:
          strncpy(salt, str, SHADOW_LINE_LEN - 1);
          salt[SHADOW_LINE_LEN - 1] = '\0';
          break;
        case 3:
          strncpy(pw_hash, str, SHADOW_LINE_LEN - 1);
          pw_hash[SHADOW_LINE_LEN - 1] = '\0';
          break;
      }
    }
  }
  else
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "parse_shadow_entry: Could not execute regex.\n");
    result = false;
  }

psc_done:
  return result;
}


/** Compare the existing hashed password with the one generated by mkpasswd */
static bool compare_password_hash(int existing_hash_id, const char *existing_salt, const char *existing_pw_hash, const char *generated_pw_entry)
{
  char existing_pw_entry[SHADOW_LINE_LEN];
  bool match;

  /*
   * Convert existing hashid/salt/hash into a string to compare against the generated one:
   * $<hashid>$<salt>$<hashedpw>
   * $5$ix5rdVA58xRQlJ.z$Bn5L/Y78T1j1V7NCZo8XA4QsjTKfk2zsz.nO0u4Hcl/
   */
  snprintf(existing_pw_entry, SHADOW_LINE_LEN, "$%d$%s$%s", existing_hash_id, existing_salt, existing_pw_hash);

  WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "compare_password_hash: Existing password entry \"%s\", generated password entry \"%s\"\n",
                                existing_pw_entry, generated_pw_entry);
  match = (strcmp(existing_pw_entry, generated_pw_entry) == 0);

  /* Zero out data */
  memset(existing_pw_entry, 0, sizeof(SHADOW_LINE_LEN));
  return match;
}


/** Use mkpasswd to hash our generated password, and then compare against the current password hash */
static bool mkpasswd_compare_password_hash(int hash_id, const char *salt, const char *existing_pw_hash, const char *generated_pw, bool *pw_match)
{
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_bufferalloc rcba;
  #define NUM_MKPASSWD_ARGS 6
  const char *mkpasswd_cmd[NUM_MKPASSWD_ARGS];
  int ret_code;
  bool result;
  enum e_el_result res;
  const char *hash_type;
  char *ptr;

  /* Work out hash type */
  hash_type = (hash_id < pht_num_hash_types) ? password_hash_types[hash_id] : NULL;
  if (NULL == hash_type)
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "compare_password_hash: unknown/unsupported hash id %d\n", hash_id);
    return false;
  }

  /* Generate mkpasswd command like: mkpasswd -m sha256 <password> <salt> */
  mkpasswd_cmd[0] = "mkpasswd";
  mkpasswd_cmd[1] = "-m";
  mkpasswd_cmd[2] = hash_type;
  mkpasswd_cmd[3] = generated_pw;
  mkpasswd_cmd[4] = salt;
  mkpasswd_cmd[5] = NULL;

  /* Configure to read back hashed password from mkpasswd */
  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_stringalloc(&rcpc, &rcba, 1024, erp_stdout);

  /* Run "mkpasswd", wait for the command to complete, and read back the hashed password */
  res = el_run_command(mkpasswd_cmd[0], mkpasswd_cmd, &ret_code, &rcpc, true, true, true);
  if (elr_success != res || !WIFEXITED(ret_code) || 0 != WEXITSTATUS(ret_code) || NULL == rcba.data || 0 == rcba.data_posn)
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "compare_password_hash: running mkpasswd failed in some way\n");
    result = false;
    goto inigc_freedata;
  }

  /* Strip off trailing whitespace */
  for (ptr = &rcba.data[strlen(rcba.data) - 1]; ptr > rcba.data && isspace(*ptr) ; ptr--)
    ;
  ptr[1] = '\0';

  /* Now pick it apart so we can compare the existing password hash with the new password hash */
  *pw_match = compare_password_hash(hash_id, salt, existing_pw_hash, rcba.data);
  result = true;

inigc_freedata:
  if (NULL != rcba.data && 0 != rcba.data_posn)
    memset(rcba.data, 0, rcba.data_posn);
  el_free_read_cmd_pipe_stringalloc(&rcba);
  return result;
}


/**
 * Check and set the password for username based on:
 *
 * If set_if_match and match_pw == stored_pw, then set password to set_pw
 * If set_if_match and match_pw != stored_pw, do nothing
 * If !set_if_match and match_pw == stored_pw, then do nothing
 * If !set_if_match and match_pw != stored_pw, then set password to set_pw
 */
static void check_set_password(const char *username, const char *match_pw, const char *set_pw, const char *shadow_regexstr, bool set_if_match)
{
  /*
   * To prevent writing the shadow password file every time this runs,
   * we need to check if the shadow root password is different to the
   * one we calculated from the factory data.
   *
   * There is no simple tool to do this, so we need to read the shadow
   * password file, and parse out the root password field, then hash our
   * calculated password combined with the salt from the root password field
   * and compare with the password file one. If they differ then, the root
   * password needs to be set.
   *
   * Shadow password entry looks like:
   *   username:$<hash id>$<salt>$<hashed password>:11883:0:99999:7:::
   *
   * mkpasswd -m sha-256 <password> <salt>
   *
   * $5$ix5rdVA58xRQlJ.z$Bn5L/Y78T1j1V7NCZo8XA4QsjTKfk2zsz.nO0u4Hcl/
   */
  FILE *shadow_file = NULL;
  #define SHADOW_LINE_LEN 200
  char shadow_line[SHADOW_LINE_LEN];
  char salt[SHADOW_LINE_LEN];
  char pw_hash[SHADOW_LINE_LEN];
  int hash_id;
  regex_t regex;
  bool regex_inited = false;
  bool pw_match;

  /* Read shadow file */
  shadow_file = fopen("/etc/shadow", "r");
  if (NULL == shadow_file)
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "check_set_password: Could not open shadow file\n");
    goto crp_fail;
  }

  if (regcomp(&regex, shadow_regexstr, REG_EXTENDED))
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "check_set_password: Could not compile regex string \"%s\"\n", shadow_regexstr);
    goto crp_fail;
  }
  regex_inited = true;

  do
  {
    /* Read line from shadow file */
    if (fgets(shadow_line, SHADOW_LINE_LEN, shadow_file) == NULL)
    {
      WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "check_set_password: Could not read line from shadow password file\n");
      break;
    }

    /* Is it the shadow password entry (root) we are after? */
    hash_id = 0;
    if (parse_shadow_entry(&regex, shadow_line, &hash_id, salt, pw_hash))
    {
      /* Yes, then see if it matches */
      WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "check_set_password: Parsed shadow entry: hash id: %d, salt: \"%s\", hash: \"%s\"\n", hash_id, salt, pw_hash);
      pw_match = true;
      if (!mkpasswd_compare_password_hash(hash_id, salt, pw_hash, match_pw, &pw_match))
      {
        WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "check_set_password: Failed to compare generated and current passwords\n");
        goto crp_fail;
      }
      if ((set_if_match && pw_match) || (!set_if_match && !pw_match))
      {
        WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "check_set_password: Passwords %smatch, setting password to \"%s\" using sha256 for user \"%s\"\n", (set_if_match) ? "" : "don't ", set_pw, username);
        if (!set_password_for_username(username, set_pw, pht_sha256))
        {
          WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "check_set_password: Failed to set password\n");
          goto crp_fail;
        }
      }
      else
      {
        WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "check_set_password: Existing and generated passwords %smatch, not setting for user \"%s\"\n", (set_if_match) ? "don't " : "", username);
      }
      break;
    }
  }
  while (1);

crp_fail:
  memset(shadow_line, 0, SHADOW_LINE_LEN);
  memset(salt, 0, SHADOW_LINE_LEN);
  memset(pw_hash, 0, SHADOW_LINE_LEN);

  if (regex_inited)
    regfree(&regex);
  if (NULL != shadow_file)
  {
    fclose(shadow_file);
    shadow_file = NULL;
  }
}


/** Check the root password has been set, and if not set it */
void check_set_root_password(const struct s_db_app_factory *facdb)
{
  char *generated_pw = NULL;

  /* Generate our root password from factory data */
  if (!generate_root_password(facdb, &generated_pw))
  {
    WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_ERR, "check_set_root_password: Could not generate root password\n");
    goto csrp_fail;
  }
  WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "check_set_root_password: The generated password is \"%s\"\n", generated_pw);

  check_set_password("root", generated_pw, generated_pw, SHADOW_REGEX_ROOT_STR, false);

csrp_fail:
  if (NULL != generated_pw)
  {
    memset(generated_pw, 0, strlen(generated_pw));
    free(generated_pw);
    generated_pw = NULL;
  }
}


/** Check if the user password is "user" and if it is then set it to "<serialnumber>", otherwise do nothing */
void check_set_user_password(uint32_t serialnumber)
{
  char generated_pw[SHADOW_LINE_LEN];

  /* Our user password is the serial number */
  snprintf(generated_pw, SHADOW_LINE_LEN, "%lu", (unsigned long) serialnumber);
  WE_ROOT_PASSWORD_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "check_set_user_password: The generated password is \"%s\"\n", generated_pw);

  /* If the user password is "user", then set the password to "<serialnumber>" */
  check_set_password("user", "user", generated_pw, SHADOW_REGEX_USER_STR, true);

  /* Clear out password */
  memset(generated_pw, 0, SHADOW_LINE_LEN);
}
