#include <stdlib.h>
#include <string.h>
#include "access.h"
#include "system.h"
#include "meas.h"
#include "comms_print.h"
#include "database.h"
#include "asciidef.h"
#include "buzz5000.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "app_db_runtime_vars.h"
#include "timer.h"

#include "setpoint.h"


/** Setpoint check function */
typedef void (*t_chk_f)(struct s_setpoint_running_info *spri, size_t setpoint, int32_t rdg);


/** Initialise the active setpoint readings and reading period */
static void init_setp_rdgs(struct s_setpoint_running_info *spri, const AdcSuperStructure *ass)
{
  /* Set current readings to both */
  spri->rdg_1[SETP_SOURCE_GROSS] = spri->rdg_2[SETP_SOURCE_GROSS] = ass->adcInternalData.cntGross;
  spri->rdg_1[SETP_SOURCE_NET] = spri->rdg_2[SETP_SOURCE_NET] = ass->adcInternalData.cntNet;
  /* Get reading period in ms  TODO magic number */
  spri->rdg_period = 10000 / AdcSyncFrequencies[ass->adcUserDatabase.syncFreqIdx];
}


/** Update the active setpoints readings */
static void update_setp_rdgs(struct s_setpoint_running_info *spri, const AdcSuperStructure *ass)
{
  int32_t change;

  /* Get how much the weight has changed since the last update */
  change = ass->adcInternalData.cntGross - spri->rdg_1[SETP_SOURCE_GROSS];
  /* Get the current weight */
  spri->rdg_1[SETP_SOURCE_GROSS] = ass->adcInternalData.cntGross;
  spri->rdg_1[SETP_SOURCE_NET] = ass->adcInternalData.cntNet;
  /* predict the next weight */
  spri->rdg_2[SETP_SOURCE_GROSS] = spri->rdg_1[SETP_SOURCE_GROSS] + change;
  spri->rdg_2[SETP_SOURCE_NET] = spri->rdg_1[SETP_SOURCE_NET] + change;
}


/** Turn on setpoint */
static void turn_on(struct s_setpoint_running_info *spri, size_t setpoint)
{
  setpoint_data_t setp_mask = 0x01 << setpoint;

  if ((setp_mask & spri->setp_active) == 0)
  {
    spri->setp_active |= setp_mask;
    spri->outp_active |= setp_mask;
    spri->setpoint_info[setpoint].changeTime = 0;
  }
}


/** Turn off setpoint */
static void turn_off(struct s_setpoint_running_info *spri, size_t setpoint)
{
  setpoint_data_t setp_mask = 0x01 << setpoint;

  if ((setp_mask & spri->setp_active) != 0)
  {
    spri->setp_active &= ~setp_mask;
    spri->outp_active &= ~setp_mask;
    spri->setpoint_info[setpoint].changeTime = 0;
  }
}

/** Invert setpoint */
static void invert(struct s_setpoint_running_info *spri, size_t setpoint)
{
  setpoint_data_t setp_mask = 0x01 << setpoint;

  if ((setp_mask & spri->setp_active) != 0)
  {
    spri->setp_active &= ~setp_mask;
    spri->outp_active &= ~setp_mask;
  }
  else
  {
    spri->setp_active |= setp_mask;
    spri->outp_active |= setp_mask;
  }
  spri->setpoint_info[setpoint].changeTime = 0;
}


/** Activate or de-activate a setpoint */
static void activate(struct s_setpoint_running_info *spri, size_t setpoint, bool on)
{
  bool active_high;

  active_high = (spri->setpoint_info[setpoint].splogic == SETP_LOGIC_HIGH);
  if (on)
  {
    if (active_high)
      turn_on(spri, setpoint);
    else
      turn_off(spri, setpoint);
  }
  else
  {
    if (active_high)
      turn_off(spri, setpoint);
    else
      turn_on(spri, setpoint);
  }
}


/** Calculate the predicted delay for activating or de-activating a setpoint */
static void set_change_time(struct s_setpoint_running_info *spri, size_t setpoint, bool high)
{
  int32_t time, target, weight;

  /* Get the weight change */
  weight  = spri->rdg_2[spri->setpoint_info[setpoint].spsource] - spri->rdg_1[spri->setpoint_info[setpoint].spsource];
  /* Get remaining weight to target */
  if (high)
  {
    target = spri->setpoint_info[setpoint].splimit_high - spri->rdg_1[spri->setpoint_info[setpoint].spsource];
  }
  else
  {
    target = spri->setpoint_info[setpoint].splimit_low - spri->rdg_1[spri->setpoint_info[setpoint].spsource];
  }
  /* Calculate the time */
  time = target * spri->rdg_period / weight;
  /* See if we should wait for next time instead */
  if (time > spri->rdg_period)
  {
    invert(spri, setpoint);
  }
  /* Not sure how this would happen but we check for this in the R420 so there might have been a reason for that */
  if (time < 0)
  {
    time = 0;
  }
  spri->setpoint_info[setpoint].changeTime = (uint8_t)time;
}


/** Active setpoint check function for over and weigh in setpoints */
static void chk_over(struct s_setpoint_running_info *spri, size_t setpoint, bool error, bool batch)
{
  if (error)
  {
    spri->setpoint_info[setpoint].onLogical = false;
    return;
  }
  /* Check if we are already over the target */
  if (spri->rdg_1[spri->setpoint_info[setpoint].spsource] > spri->setpoint_info[setpoint].splimit_high)
  {
    spri->setpoint_info[setpoint].onLogical = !batch;
  }
  /* Check if we are likely to be over the target before the next reading */
  else if (spri->rdg_2[spri->setpoint_info[setpoint].spsource] > spri->setpoint_info[setpoint].splimit_high)
  {
    spri->setpoint_info[setpoint].onLogical = !batch;
    /* Figure out when that is likely to happen */
    set_change_time(spri, setpoint, true);
  }
  /* Check if we are already below hysteresis value */
  if (spri->rdg_1[spri->setpoint_info[setpoint].spsource] < spri->setpoint_info[setpoint].splimit_low)
  {
    spri->setpoint_info[setpoint].onLogical = batch;
  }
  /* Check if we are likely to be under the hysteresis value before the next reading */
  else if (spri->rdg_2[spri->setpoint_info[setpoint].spsource] < spri->setpoint_info[setpoint].splimit_low)
  {
    spri->setpoint_info[setpoint].onLogical = batch;
    /* Figure out when that is likely to happen */
    set_change_time(spri, setpoint, false);
  }
}


/** Active setpoint check function for under and weigh out setpoints */
static void chk_under(struct s_setpoint_running_info *spri, size_t setpoint, bool error, bool batch)
{
  if (error)
  {
    spri->setpoint_info[setpoint].onLogical = false;
    return;
  }
  /* Check if we are already below the target */
  if (spri->rdg_1[spri->setpoint_info[setpoint].spsource] < spri->setpoint_info[setpoint].splimit_low)
  {
    spri->setpoint_info[setpoint].onLogical = !batch;
  }
  /* Check if we are likely to be below the target by the next reading */
  else if (spri->rdg_2[spri->setpoint_info[setpoint].spsource] < spri->setpoint_info[setpoint].splimit_low)
  {
    spri->setpoint_info[setpoint].onLogical = !batch;
    /* Figure out when that is likely to happen */
    set_change_time(spri, setpoint, false);
  }
  /* Check if we are already above the hysteresis value */
  if (spri->rdg_1[spri->setpoint_info[setpoint].spsource] > spri->setpoint_info[setpoint].splimit_high)
  {
    spri->setpoint_info[setpoint].onLogical = batch;
  }
  /* Check if we are likely to be above the hysteresis value by the next reading */
  else if (spri->rdg_2[spri->setpoint_info[setpoint].spsource] > spri->setpoint_info[setpoint].splimit_high)
  {
    spri->setpoint_info[setpoint].onLogical = batch;
    /* Figure out when that is likely to happen */
    set_change_time(spri, setpoint, true);
  }
}


/** Internal function to set/clear any outputs driven directly via comms commands (FOP/POR) */
static void do_direct_drive(struct s_setpoint_running_info *spri)
{
  setpoint_data_t out_mask;

  out_mask = spri->setp_active;
  out_mask |= spri->direct_set;
  out_mask &= ~spri->direct_clear;
  spri->outp_active = out_mask;
}


/** Drive an output directly (for use with comms commands like FOP/POR) */
bool direct_drive_output(struct s_setpoint_running_info *spri, size_t setpoint, bool on)
{
  setpoint_data_t setp_mask;

  if (setpoint >= NUM_SETPOINTS)
    return false;

  /* Can only FOP outputs or unused IO */
  if (access_sptype_remkeytype_to_io_type(spri->setpoint_info[setpoint].sptype,
                                          spri->setpoint_info[setpoint].inputtype) == IOTypeInput)
    return false;

  setp_mask = 0x01 << setpoint;
  if ((setp_mask & spri->direct_is_driven) == 0)
  {
    /* Hasn't been driven before, so configure IO as output */
    access_configure_iopin(setpoint, ioOutput);
    spri->direct_is_driven |= setp_mask;
  }

  /* Set state */
  if (on)
  {
    spri->direct_set |= setp_mask;
    spri->direct_clear &= ~setp_mask;
  }
  else
  {
    spri->direct_clear |= setp_mask;
    spri->direct_set &= ~setp_mask;
  }
  return true;
}


/** Is setpoint a direct drive output */
bool is_direct_drive_output(const struct s_setpoint_running_info *spri, size_t setpoint)
{
  setpoint_data_t setp_mask;

  if (setpoint >= NUM_SETPOINTS)
    return false;

  setp_mask = 0x01 << setpoint;

  return ((spri->direct_is_driven & setp_mask) != 0);
}


/** Is setpoint output currently driven */
bool is_setpoint_output_active(const struct s_setpoint_running_info *spri, size_t setpoint)
{
  setpoint_data_t setp_mask = 0x01 << setpoint;

  return ((setp_mask & spri->setp_active) != 0);
}


/** Stop direct drive of setpoint */
void clear_direct_drive_output(struct s_setpoint_running_info *spri, size_t setpoint)
{
  setpoint_data_t setp_mask;

  setp_mask = 0x01 << setpoint;

  spri->direct_is_driven &= ~setp_mask;
  spri->direct_set &= ~setp_mask;
  spri->direct_clear &= ~setp_mask;
}


/** If setpoint used (as an input) then clear out any direct drive settings */
static void reinit_direct_drive(struct s_setpoint_running_info *spri)
{
  size_t i;

  for (i = 0; i < NUM_SETPOINTS; i++)
  {
    if (access_sptype_remkeytype_to_io_type(spri->setpoint_info[i].sptype,
                                            spri->setpoint_info[i].inputtype) == IOTypeInput)
    {
      clear_direct_drive_output(spri, i);
    }
  }
}


/** Is the scale in error, used for error setpoints */
static bool is_setp_error(const AdcSuperStructure *ass)
{
  return (APP_ERR_NONE != get_current_errors() ||
          ass->adcDataWorking.grossRdg.info.status.bit.oload ||
          ass->adcDataWorking.grossRdg.info.status.bit.uload);
}


static bool check_reset(enum eSetpointReset spReset)
{
  uint8_t reset_bits;
  enum eRemoteKeyFn func;
  size_t i;

  switch (spReset)
  {
    case SETP_RESET_A:
      func = REMOTE_SP_RESET_A;
      break;
    case SETP_RESET_B:
      func = REMOTE_SP_RESET_B;
      break;
    default:
      func = REMOTE_DISABLED;
      break;
  }

  if (func == REMOTE_DISABLED)
    return false;

  reset_bits = 0;
  for (i = 0; i < NUM_REMOTE_KEYS; i++)
  {
    if (db_app_user.remote_f[i] == func)
      reset_bits |= 0x01 << i;
  }
  if (access_get_inputs() & reset_bits)
    return true;

  return false;
}


/** Called to update setpoint relays that are controlled by time instead of weight */
static void internal_update_setpoints(struct s_setpoint_running_info *spri, const AdcSuperStructure *ass)
{
  size_t setpoint;
  bool nextSet = false;

  for (setpoint = 0; setpoint < NUM_SETPOINTS; setpoint++)
  {
    switch(spri->setpoint_info[setpoint].sptype)
    {
      case SETP_OVER:
        chk_over(spri, setpoint, is_setp_error(ass), false);
        break;
      case SETP_UNDER:
        chk_under(spri, setpoint, is_setp_error(ass), false);
        break;
      case SETP_MOTION:
        spri->setpoint_info[setpoint].onLogical = ass->adcDataWorking.grossRdg.info.status.bit.motion;
        break;
      case SETP_ZERO:
        spri->setpoint_info[setpoint].onLogical = ass->adcDataWorking.grossRdg.info.status.bit.zeroBand;
        break;
      case SETP_ERROR:
        spri->setpoint_info[setpoint].onLogical = is_setp_error(ass);
        break;
      case SETP_NET:
        spri->setpoint_info[setpoint].onLogical = !ass->adcDataWorking.dispRdg.info.status.bit.gross;
        break;
      case SETP_ON:
        spri->setpoint_info[setpoint].onLogical = true;
        break;
      case SETP_COZ:
        spri->setpoint_info[setpoint].onLogical = ass->adcDataWorking.grossRdg.info.status.bit.coz;
        break;
      case SETP_BUZZ:
        spri->setpoint_info[setpoint].onLogical = is_beeping();
        break;
      case SETP_W_OUT:
        /* See if there is already a weigh in/weigh out setpoint active */
        if ((spri->feeder == SETP_FEEDER_SINGLE) && spri->feed_active)
          spri->setpoint_info[setpoint].onLogical = false;
        else
        {
          /* Check if we are delaying making decisions */
          if (msDiffTime(ass->adcInternalData.tsReading, spri->active_delay) > 0)
            break;
          chk_under(spri, setpoint, is_setp_error(ass), true);
          /* Set feed active */
          spri->feed_active |= spri->setpoint_info[setpoint].onLogical == true;
          /* Set delay time if we have just turned on */
          if (spri->setpoint_info[setpoint].onLogical && !spri->setpoint_info[setpoint].prevLogical)
            spri->active_delay = addTime(ass->adcInternalData.tsReading, spri->delay_check);
          /* Save onLogical value for next time */
          spri->setpoint_info[setpoint].prevLogical = spri->setpoint_info[setpoint].onLogical;
        }
        break;
      case SETP_W_IN:
        /* See if there is already a weigh in/weigh out setpoint active */
        if ((spri->feeder == SETP_FEEDER_SINGLE) && spri->feed_active)
          spri->setpoint_info[setpoint].onLogical = false;
        else
        {
          /* Check if we are delaying making decisions */
          if (msDiffTime(ass->adcInternalData.tsReading, spri->active_delay) > 0)
            break;
          chk_over(spri, setpoint, is_setp_error(ass), true);
          /* Set feed active */
          spri->feed_active |= spri->setpoint_info[setpoint].onLogical == true;
          /* Set delay time if we have just turned on */
          if (spri->setpoint_info[setpoint].onLogical && !spri->setpoint_info[setpoint].prevLogical)
            spri->active_delay = addTime(ass->adcInternalData.tsReading, spri->delay_check);
          /* Save onLogical value for next time */
          spri->setpoint_info[setpoint].prevLogical = spri->setpoint_info[setpoint].onLogical;
        }
        break;
      default:
        break;
    }
    /* Check reset input */
    spri->setpoint_info[setpoint].reset = check_reset(spri->setpoint_info[setpoint].spReset);
    /* Latch the reset value */
    spri->setpoint_info[setpoint].latchReset |= spri->setpoint_info[setpoint].reset;

    /* Handle timing */
    switch(spri->setpoint_info[setpoint].spTiming)
    {
      case SETP_TIMING_LEVEL:
        /* Check reset input (if one is set) */
        if (spri->setpoint_info[setpoint].spReset != SETP_RESET_NONE)
        {
          /* Clear onLogical if reset is true */
          if (spri->setpoint_info[setpoint].reset)
            spri->setpoint_info[setpoint].onLogical = false;
        }
        break;
      case SETP_TIMING_EDGE:
        if (spri->setpoint_info[setpoint].spReset != SETP_RESET_NONE)
        {
          /* Clear latched reset if onLogical is false */
          if (!spri->setpoint_info[setpoint].onLogical)
            spri->setpoint_info[setpoint].latchReset = false;
          /* Clear onLogical if reset or latched reset are true */
          if ((spri->setpoint_info[setpoint].reset) || (spri->setpoint_info[setpoint].latchReset))
            spri->setpoint_info[setpoint].onLogical = false;
        }
        break;
      case SETP_TIMING_LATCH:
        if (spri->setpoint_info[setpoint].spReset != SETP_RESET_NONE)
        {
          /* Clear latched reset if onLogical is false */
          if (!spri->setpoint_info[setpoint].onLogical)
            spri->setpoint_info[setpoint].latchReset = false;
          /* Clear onLogical if reset or latched reset are true */
          if ((spri->setpoint_info[setpoint].reset) || (spri->setpoint_info[setpoint].latchReset))
            spri->setpoint_info[setpoint].onLogical = false;
          else /* use lastLogical as a latch */
            spri->setpoint_info[setpoint].onLogical |= spri->setpoint_info[setpoint].lastLogical;
          /* Save onLogical for next time */
          spri->setpoint_info[setpoint].lastLogical = spri->setpoint_info[setpoint].onLogical;
        }
        break;
      case SETP_TIMING_PULSE:
        /* Handle waiting for no motion between jog sets */
        if (spri->setpoint_info[setpoint].waitMotion)
        {
          if (!ass->adcDataWorking.grossRdg.info.status.bit.motion)
          {
            spri->setpoint_info[setpoint].waitMotion = false;
            if (spri->setpoint_info[setpoint].onLogical)
              nextSet = true;
            else
              spri->setpoint_info[setpoint].latchReset = true;
          }
          else
          {
            spri->setpoint_info[setpoint].onLogical = false;
            break;
          }
        }
        /* Latch onLogical and start on time */
        if (spri->setpoint_info[setpoint].onLogical && !spri->setpoint_info[setpoint].latchLogical)
        {
          spri->setpoint_info[setpoint].latchLogical = spri->setpoint_info[setpoint].onLogical;
          /* Add jog on time to reading time */
          spri->setpoint_info[setpoint].endTime = addTime(ass->adcInternalData.tsReading, spri->jog_on);
          spri->setpoint_info[setpoint].plsNum = 0;
          if (!nextSet)
            spri->setpoint_info[setpoint].setNum = 0;
        }
        /* Clear latched reset if onLogical is false */
        if (spri->setpoint_info[setpoint].latchReset && !spri->setpoint_info[setpoint].onLogical)
        {
          spri->setpoint_info[setpoint].latchReset = false;
          spri->setpoint_info[setpoint].latchLogical = false;
        }
        /* Run the pulse then delay */
        if (spri->setpoint_info[setpoint].latchLogical && !spri->setpoint_info[setpoint].latchReset)
        {
          if (spri->setpoint_info[setpoint].lastLogical)
          {
            /* See if we need to end before the next reading */
            if (msDiffTime(ass->adcInternalData.tsReading, spri->setpoint_info[setpoint].endTime) > spri->rdg_period)
            {
              /* Check end time is less than reading time plus sample time */
              spri->setpoint_info[setpoint].onLogical = true;
            }
            else
            {
              /* Check if it is less than current reading time */
              if (msDiffTime(ass->adcInternalData.tsReading, spri->setpoint_info[setpoint].endTime) <= 0)
              {
                /* Change immediately */
                spri->setpoint_info[setpoint].changeTime = 0;
              }
              else
              {
                /* Figure out how long until end time since last reading */
                spri->setpoint_info[setpoint].changeTime = msDiffTime(ass->adcInternalData.tsReading, spri->setpoint_info[setpoint].endTime);
              }
              spri->setpoint_info[setpoint].onLogical = false;
              /* Set to jog off time */
              spri->setpoint_info[setpoint].endTime = addTime(ass->adcInternalData.tsReading, spri->jog_off);
            }
          }
          else
          {
            /* Compare end time to next sample time */
            if (msDiffTime(ass->adcInternalData.tsReading, spri->setpoint_info[setpoint].endTime) > spri->rdg_period)
            {
              spri->setpoint_info[setpoint].onLogical = false;
            }
            else
            {
              /* Compare end time to current sample time */
              if (msDiffTime(ass->adcInternalData.tsReading, spri->setpoint_info[setpoint].endTime) <= 0)
              {
                spri->setpoint_info[setpoint].changeTime = 0;
              }
              else
              {
                /* Set to difference between current sample time and end time */
                spri->setpoint_info[setpoint].changeTime = msDiffTime(ass->adcInternalData.tsReading, spri->setpoint_info[setpoint].endTime);
              }
              spri->setpoint_info[setpoint].plsNum++;
              if (spri->setpoint_info[setpoint].plsNum > spri->jog_set)
              {
                spri->setpoint_info[setpoint].onLogical = false;
                spri->setpoint_info[setpoint].setNum++;
                if ((spri->max_sets == 0) || (spri->setpoint_info[setpoint].setNum < spri->max_sets))
                {
                  spri->setpoint_info[setpoint].waitMotion = true;
                  spri->setpoint_info[setpoint].latchLogical = false;
                }
                else
                {
                  /* Pretend we have seen the reset input to stop more pulses from happening */
                  spri->setpoint_info[setpoint].latchReset = true;
                }
              }
              else
              {
                spri->setpoint_info[setpoint].onLogical = true;
                /* Set to jog on time */
                spri->setpoint_info[setpoint].endTime = addTime(ass->adcInternalData.tsReading, spri->jog_on);
              }
            }
          }
        }
        /* Clear onLogical if reset or latched reset are true */
        if ((spri->setpoint_info[setpoint].reset) || (spri->setpoint_info[setpoint].latchReset))
          spri->setpoint_info[setpoint].onLogical = false;
        /* Keep the state for next time */
        spri->setpoint_info[setpoint].lastLogical = spri->setpoint_info[setpoint].onLogical;
        break;
      default:
        break;
    }
    /* Drive outputs */
    activate(spri, setpoint, spri->setpoint_info[setpoint].onLogical);
  }
}


/** Poll this regularly to update setpoints */
void update_setpoints(struct s_setpoint_running_info *spri, const AdcSuperStructure *ass)
{
  /* Only update on new reading */
  if (spri->last_reading_id == ass->adcDataWorking.rdgId)
    return;
  spri->last_reading_id = ass->adcDataWorking.rdgId;
  spri->feed_active = false;

  /* Update readings */
  update_setp_rdgs(spri, ass);

  /* Update setpoint states */
  internal_update_setpoints(spri, ass);

  /* Mash in direct drive outputs */
  do_direct_drive(spri);

  /* Drive outputs */
  access_set_outputs(spri->outp_active);
}


/** (Re)initialise setpoints with database settings */
void init_setpoints(struct s_setpoint_running_info *spri, const AdcSuperStructure *ass)
{
  size_t i;
  int32_t sp_limit_low;
  int32_t sp_limit_high;

  for (i = 0; i < NUM_SETPOINTS; i++)
  {
    /* Keep a copy of DB values */
    spri->setpoint_info[i].sptype = db_app_user.setp_active[i];
    spri->setpoint_info[i].spsource = db_app_user.setp_source[i];
    spri->setpoint_info[i].splogic = db_app_user.setp_logic[i];
    spri->setpoint_info[i].inputtype = db_app_user.remote_f[i];
    spri->setpoint_info[i].spTiming = db_app_user.setp_timing[i];
    spri->setpoint_info[i].spReset = db_app_user.setp_reset[i];

    /* Work out active setpoint modes/limits */
    if (spri->setpoint_info[i].sptype == SETP_UNDER || spri->setpoint_info[i].sptype == SETP_W_OUT)
    {
      sp_limit_low = setpoint_target_get(i) + setpoint_preflight_get(i);
      sp_limit_high = sp_limit_low +  setpoint_hysteresis_get(i);
    }
    else
    {
      sp_limit_high = setpoint_target_get(i) - setpoint_preflight_get(i);
      sp_limit_low = sp_limit_high - setpoint_hysteresis_get(i);
    }
    /* Limits are now in counts */
    spri->setpoint_info[i].splimit_low = sp_limit_low << 4;
    spri->setpoint_info[i].splimit_high = sp_limit_high << 4;

    /* Clear hold off times */
    spri->setpoint_info[i].changeTime = 0;

    /* Clear all timing variables */
    spri->setpoint_info[i].endTime = ass->adcInternalData.tsReading;
    spri->setpoint_info[i].lastLogical = false;
    spri->setpoint_info[i].latchLogical = false;
    spri->setpoint_info[i].latchReset = false;
    spri->setpoint_info[i].onLogical = false;
    spri->setpoint_info[i].plsNum = 0;
    spri->setpoint_info[i].reset = false;
    spri->setpoint_info[i].waitMotion = false;
    spri->setpoint_info[i].setNum = 0;
    spri->setpoint_info[i].prevLogical = false;
  }

  /* Keep a copy of DB values */
  spri->jog_off = db_app_user.setp_jog_off;
  spri->jog_on = db_app_user.setp_jog_on;
  spri->jog_set = db_app_user.setp_jog_set;
  spri->max_sets = db_app_user.setp_max_sets;
  spri->feeder = db_app_user.setp_feeder;
  spri->delay_check = db_app_user.setp_delay_check;

  /* Init batching setpoint variables */
  spri->active_delay = ass->adcInternalData.tsReading;
  spri->feed_active = false;

  /* Init readings */
  init_setp_rdgs(spri, ass);

  /* Disable any direct drives that are now inputs or outputs */
  reinit_direct_drive(spri);

  /* Update setpoint states */
  spri->setp_active = 0;
  spri->outp_active = 0;

  /* Mash in direct drive outputs */
  do_direct_drive(spri);

  /* Drive outputs */
  access_set_outputs(spri->outp_active);
}
