/*
 * signals.c
 *
 *  Created on: Jun 6, 2012
 *      Author: ian
 */
#include <signal.h>
#include <string.h>
#include <pthread.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rinthread.h>
#include "we_debug.h"

#include "signals.h"


/** Structure passed to the signal handling thread */
struct s_signal_thread_config
{
  bool *application_shutdown_flag;
  bool running_as_daemon;
};


static pthread_t signal_processing_thread;
static struct s_signal_thread_config signal_thread_config;


/** Dummy signal handler, because we use sigwait() for synchronous handling of signals */
static void dummy_signal_handler(int sig)
{
  /* Nothing to do here */
}


/** Install the default signal handler for a signal */
static void install_signal_handler(int sig)
{
  struct sigaction new_action, old_action;

  /* Set up the structure to specify the new action. */
  memset(&new_action, 0, sizeof(new_action));
  new_action.sa_handler = dummy_signal_handler;
  sigemptyset(&new_action.sa_mask);
  new_action.sa_flags = 0;

  /* Get current */
  sigaction(sig, NULL, &old_action);
}


/** Signal handling thread */
static void *signal_processor(void *arg)
{
  struct s_signal_thread_config *thread_config = arg;
  bool shutdown;
  sigset_t set;
  int sig;

  WE_SIGNALS_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "signal_processing_thread: started\n");

  /*
   * Install a dummy signal handler for these signals,
   * so we can use sigwait() below to handle them
   * synchronously
   */
  install_signal_handler(SIGINT);
  install_signal_handler(SIGTERM);
  install_signal_handler(SIGHUP);
  install_signal_handler(SIGUSR1);
  install_signal_handler(SIGUSR2);
  install_signal_handler(SIGALRM);

  for (shutdown = false; !shutdown; )
  {
    /* Wait for all signals */
    WE_SIGNALS_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "signal_processing_thread: waiting for signal\n");
    sigfillset(&set);
    if (sigwait(&set, &sig) != 0)
    {
      rdebug_printf(LOG_ERR, "signal_processing_thread: error waiting for signal\n");
      continue;
    }
    WE_SIGNALS_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "signal_processing_thread: received signal %d\n", sig);

    /* Handle the signal synchronously here, rather than in a signal handler */
    switch (sig)
    {
      case SIGINT:
      case SIGTERM:
      case SIGKILL:
        shutdown = true;
        break;

      case SIGHUP:
        if (NULL != thread_config && !thread_config->running_as_daemon)
          shutdown = true;
        break;

      default:
        WE_SIGNALS_DEBUG_PRINTF(RDEBUG_LVL_FULL, LOG_INFO, "signal_processing_thread: unhandled signal %d\n", sig);
        break;
    }
  }

  /* Done */
  WE_SIGNALS_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "signal_processing_thread: finished\n");
  if (NULL != thread_config && NULL != thread_config->application_shutdown_flag)
    *thread_config->application_shutdown_flag = true;
  rin_thread_exit(NULL);
}


/** Initialise the signal handling code */
bool init_signals(bool *application_quit_flag, bool as_daemon)
{
  sigset_t set;
  bool res;
  bool blockres;

  signal_thread_config.application_shutdown_flag = application_quit_flag;
  signal_thread_config.running_as_daemon = as_daemon;

  /* Create signal processing thread, with no signals blocked (it will inherit what the parent had) */
  res = rin_thread_create(&signal_processing_thread, signal_processor, &signal_thread_config, true);

  /* Now block pretty much all signals, so other child threads won't receive any signals */
  blockres = (0 == sigemptyset(&set) &&
              0 == sigaddset(&set, SIGINT) &&
              0 == sigaddset(&set, SIGTERM) &&
              0 == sigaddset(&set, SIGHUP) &&
              0 == sigaddset(&set, SIGUSR1) &&
              0 == sigaddset(&set, SIGUSR2) &&
              0 == sigaddset(&set, SIGALRM) &&
              0 == sigprocmask(SIG_BLOCK, &set, NULL));

  return blockres && res;
}
