/*
 * temperature.c
 *
 *  Created on: Aug 22, 2013
 *      Author: ian
 */
#include <stdint.h>
#include <stddef.h>
#include <pthread.h>
#include "adc_8bit.h"

#include "temperature.h"


#define TEMPERATURE_NUM_AVERAGING 16


static pthread_mutex_t temperature_mutex = PTHREAD_MUTEX_INITIALIZER;
static uint16_t temperature_readings[TEMPERATURE_NUM_AVERAGING];
static size_t temperature_readings_index = 0;
static uint32_t temperature_rdg_average;
static bool first_temperature_reading = true;


/** Average the raw temperature readings, will be TEMPERATURE_NUM_AVERAGING times larger than rawrdg */
static void temperature_average(uint16_t rawrdg)
{
  size_t i;

  if (first_temperature_reading)
  {
    /* Prefill readings */
    first_temperature_reading = false;
    for (i = 0; i < TEMPERATURE_NUM_AVERAGING; i++)
      temperature_readings[i] = rawrdg;
    temperature_readings_index = 0;
  }
  else
  {
    /* Store the new reading */
    temperature_readings_index++;
    if (temperature_readings_index >= TEMPERATURE_NUM_AVERAGING)
      temperature_readings_index = 0;
    temperature_readings[temperature_readings_index] = rawrdg;
  }

  /* Average all readings */
  for (i = 0, temperature_rdg_average = 0; i < TEMPERATURE_NUM_AVERAGING; i++)
    temperature_rdg_average += temperature_readings[i];
}


/** Call to update the temperature with a new reading */
void temperature_update(uint16_t rawrdg)
{
  pthread_mutex_lock(&temperature_mutex);
  temperature_average(rawrdg);
  pthread_mutex_unlock(&temperature_mutex);
}


/** Get the current temperature in requested units */
long temperature_get(enum eTemperatureUnits units)
{
  long result;

  pthread_mutex_lock(&temperature_mutex);
  switch (units)
  {
    case tuRaw:
      /* Return latest reading */
      result = temperature_readings[temperature_readings_index];
      break;
    case tuRawAveraged:
      result = temperature_rdg_average;
      break;
    case tuKelvin:
      /* Return latest reading converted to Kelvin */
      result = WE21_TEMP_KELVIN_REAL_FROM_CNT(temperature_readings[temperature_readings_index]);
      break;
    case tuCelcius:
      /* Return latest reading converted to Celcius */
      result = WE21_TEMP_KELVIN_REAL_FROM_CNT(temperature_readings[temperature_readings_index]) - 273;
      break;
    default:
      result = 0;
      break;
  }
  pthread_mutex_unlock(&temperature_mutex);
  return result;
}
