/*
 * test_comms_perf.c
 *
 * Test the performance of the comms
 *
 *  Created on: May 12, 2017
 *      Author: ian
 */
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <unistd.h>


static bool cmdline_showhelp = false;
static unsigned cmdline_port = 2222;
static unsigned cmdline_count = 0;
static const char *cmdline_hostname = NULL;
static const char *cmdline_command = NULL;
static const char *cmdline_resp_termination = "\r\n";


/** Print out the command line options */
static void print_commandlineopts(const char *program_name)
{
  fprintf(stdout, "Usage: %s [-options] host\n"
                  "Arguments:\n"
                  "  host                         The host to connect to\n"
                  "Options:\n"
                  "  -h          --help           Print help\n"
                  "  -p <port>   --port <port>    The TCP port to connect to (default 2222)\n"
                  "  -n <count>  --count <count>  The number of times to poll the host.\n"
                  "                               0 for indefinite\n"
                  "  -c <cmd>    --command <cmd>  The command to send. Escape\n"
                  "                               non-printables with \\xhh.\n"
                  "  -t <cmd>    --termination <term>  The termination character(s) used in the response.\n"
                  "                                    Escape non-printables with \\xhh. Default \\x0d\\x0a.\n",
          program_name);
}


/** Parse the command line options */
static bool parse_commandline(int argc, char *argv[])
{
  int c;
  static const char opts[] = "hp:n:c:t:";
  static const struct option lopts[] =
  {
    { "help",          0, 0, 'h'  },
    { "port",          1, 0, 'p'  },
    { "count",         1, 0, 'n'  },
    { "command",       1, 0, 'c'  },
    { "termination",   1, 0, 't'  },
    { NULL,            0, 0, '\0' }
  };

  while (1)
  {
    c = getopt_long(argc, argv, opts, lopts, NULL);
    if (-1 == c)
    {
      break;
    }

    switch (c)
    {
      case 'h':
        cmdline_showhelp = true;
        return true;
        break;
      case 'p':
        cmdline_port = strtoul(optarg, NULL, 10);
        break;
      case 'n':
        cmdline_count = strtoul(optarg, NULL, 10);
        break;
      case 'c':
        cmdline_command = optarg;
        break;
      case 't':
        cmdline_resp_termination = optarg;
        break;
      default:
        return false;
        break;
    }
  }

  /* Hostname */
  if (optind >= argc)
  {
    fprintf(stderr, "No host specified, try --help\n");
    return false;
  }
  cmdline_hostname = argv[optind];
  optind++;

  return true;
}


#define NANO_SEC_PER_SEC 1000000000


static void sub_timespec(struct timespec *pts1, const struct timespec *pts2)
{
  if ( pts1->tv_sec < pts2->tv_sec ||
       ( pts1->tv_sec == pts2->tv_sec &&
         pts1->tv_nsec <= pts2->tv_nsec ) )
  {
    /* pts1 <= pts2, clamp to zero TODO calculate this correctly */
    pts1->tv_sec = 0;
    pts1->tv_nsec = 0;
  }
  else
  {
    /* pts1 > pts2 */
    pts1->tv_sec -= pts2->tv_sec;
    if (pts1->tv_nsec < pts2->tv_nsec)
    {
      /* Borrow a second */
      pts1->tv_nsec = pts1->tv_nsec + NANO_SEC_PER_SEC - pts2->tv_nsec;
      pts1->tv_sec--;
    }
    else
    {
      pts1->tv_nsec -= pts2->tv_nsec;
    }
  }
}


static bool ishexchar(char ch)
{
  return ((ch >= '0' && ch <= '9') ||
          (ch >= 'a' && ch <= 'f') ||
          (ch >= 'A' && ch <= 'F'));
}

static uint8_t hexchartonibble(char ch)
{
  uint8_t nibble;

  if (ch >= '0' && ch <= '9')
    nibble = ch - '0';
  else if (ch >= 'a' && ch <= 'f')
    nibble = ch - 'a' + 0x0A;
  else if (ch >= 'A' && ch <= 'F')
    nibble = ch - 'A' + 0x0A;
  else
    nibble = 0;
  return nibble;
}


static void unescape_string(uint8_t **unesc, size_t *unesc_len, const char *str)
{
  uint8_t *unesc_str = NULL;
  const char *posn;
  const char *next;
  ptrdiff_t copylen;
  uint8_t conv;
  size_t unesc_posn;

  unesc_str = calloc(1, strlen(str) + sizeof(char));
  if (NULL == unesc_str)
  {
    *unesc = NULL;
    *unesc_len = 0;
  }

  for (posn = str, unesc_posn = 0; ; )
  {
    next = strstr(posn, "\\x");
    if (NULL != next)
    {
      /* Copy up to \x into output string */
      copylen = next - posn;
      memcpy(&unesc_str[unesc_posn], posn, copylen);
      unesc_posn += copylen;

      /* Found \x, check for 2 hex chars */
      if (ishexchar(next[2]) && ishexchar(next[3]))
      {
        /* Convert 2 hex chars */
        conv = (hexchartonibble(next[2]) << 4) | hexchartonibble(next[3]);
        unesc_str[unesc_posn++] = conv;
        posn = &next[4];
      }
      else if (ishexchar(next[2]))
      {
        /* Convert 1 hex chars */
        conv = hexchartonibble(next[2]);
        unesc_str[unesc_posn++] = conv;
        posn = &next[3];
      }
      else
      {
        /* \x is missing hex characters, just skip it */
        posn = &next[2];
      }
    }
    else
    {
      /* No more escaping, just copy what is left, and finish */
      copylen = strlen(posn);
      memcpy(&unesc_str[unesc_posn], posn, copylen);
      unesc_posn += copylen;
      posn = &posn[copylen];
      break;
    }
  }

  *unesc = unesc_str;
  *unesc_len = unesc_posn;
}


#if 0
/** Filter the list of installed packages down to those that match the regex */
static enum e_rl_result rl_opkg_filter_installed_packages(struct s_rl_licence *rl_licence, regex_t *regex, const char *installed_package)
{
  int regex_ret;
  enum e_rl_result result;
  #define REGEX_GROUPS 2
  #define PKG_NAME_GROUP 1
  regmatch_t regex_groups[REGEX_GROUPS];
  size_t so;
  size_t eo;
  size_t len;
  char *package_name;

  /* Compile regex, ready for comparison against installed packages */
  if (regcomp(&regex, rl_licence->licensable_package_regex, REG_EXTENDED) != REG_NOERROR)
  {
    result = rlr_regexfail;
    goto rogip_fail;
  }
  regex_inited = true;


  /* Parse the shadow password entry against the regex, and pick it apart */
  regex_ret = regexec(regex, installed_package, REGEX_GROUPS, regex_groups, 0);
  if (REG_NOERROR == regex_ret)
  {
    if (regex_groups[PKG_NAME_GROUP].rm_so == -1)
      return rlr_regexfail;

    /* Matches, add this package to the installed_licensable_packages list */
    so = regex_groups[PKG_NAME_GROUP].rm_so;
    eo = regex_groups[PKG_NAME_GROUP].rm_eo;
    len = eo - so;
    package_name = calloc(1, len + 1);
    if (NULL == package_name)
      return rlr_allocfail;
    memcpy(package_name, &installed_package[so], len);
    package_name[len] = '\0';

    utarray_push_back(rl_licence->installed_licensable_packages, &package_name);
    free(package_name);
    result = rlr_success;
  }
  else if (REG_NOMATCH == regex_ret)
  {
    /* Doesn't match, nothing to do */
    result = rlr_success;
  }
  else
  {
    result = rlr_regexfail;
  }
  return result;
}
#endif


static bool read_response(int sockfd, uint8_t *buffer, size_t buffer_maxlen, const uint8_t *resp_term, size_t resp_term_length)
{
  size_t bufend;
  uint8_t ch;

  for (bufend = 0; bufend < buffer_maxlen; )
  {
    if (read(sockfd, &ch, sizeof(ch)) != sizeof(ch))
    {
      break;
    }
    buffer[bufend++] = ch;

    if (bufend >= resp_term_length && memcmp(&buffer[bufend - resp_term_length], resp_term, resp_term_length) == 0)
    {
      return true;
    }
  }
  return false;
}


static void test_poll_response(int sockfd, unsigned count,
                               const uint8_t *command, size_t command_length,
                               const uint8_t *resp_term, size_t resp_term_length)
{
  unsigned i;
  struct timespec ts_start;
  struct timespec ts_end;
  uint8_t *buffer;

  buffer = calloc(1, 2048);
  if (NULL == buffer)
  {
    fprintf(stderr, "Failed to allocate buffer\n");
    return;
  }

  for (i = 0; i < count; i++)
  {
    clock_gettime(CLOCK_MONOTONIC, &ts_start);
    if (write(sockfd, command, command_length) != (ssize_t) command_length)
    {
      fprintf(stderr, "Failed writing command to host\n");
      return;
    }

    if (!read_response(sockfd, buffer, 2048, resp_term, resp_term_length))
    {
      fprintf(stderr, "Failed reading response to host\n");
      return;
    }

    clock_gettime(CLOCK_MONOTONIC, &ts_end);
    sub_timespec(&ts_end, &ts_start);
    printf("%lu.%09lu\n", (unsigned long) ts_end.tv_sec, (unsigned long) ts_end.tv_nsec);
  }
}


static int connect_host(const char *hostname, unsigned port)
{
  int sockfd;
  struct addrinfo hints, *servinfo, *p;
  int rv;
  char portstr[20];
  int flag;

  snprintf(portstr, 20, "%u", port);

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;

  rv = getaddrinfo(hostname, portstr, &hints, &servinfo);
  if (rv != 0)
  {
    fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
    return -1;
  }

  /* Try to connect on all the returned addresses */
  for (p = servinfo; p != NULL; p = p->ai_next)
  {
    sockfd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
    if (sockfd >= 0)
    {
      if (connect(sockfd, p->ai_addr, p->ai_addrlen) == 0)
      {
        /* Connected to host */
        flag = 1;
        rv = setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
        if (rv < 0)
        {
          close(sockfd);
          return -1;
        }
        return sockfd;
      }
      else
      {
        close(sockfd);
      }
    }
  }

  /* Failed */
  return -1;
}


int main(int argc, char *argv[])
{
  int sockfd = -1;
  int result;
  uint8_t *unesc_command = NULL;
  size_t unesc_command_length;
  uint8_t *unesc_resp_term = NULL;
  size_t unesc_resp_term_length;

  /* Parse command line */
  if (!parse_commandline(argc, argv))
  {
    fprintf(stderr, "Invalid command line, try --help\n");
    result = 1;
    goto m_fail;
  }

  /* Show help if requested */
  if (cmdline_showhelp)
  {
    print_commandlineopts(argv[0]);
    result = 0;
    goto m_fail;
  }

  /* Connect to host */
  sockfd = connect_host(cmdline_hostname, cmdline_port);
  if (sockfd < 0)
  {
    fprintf(stderr, "Failed to connect to host\n");
    result = 1;
    goto m_fail;
  }

  /* Un-escape command line string */
  if (NULL != cmdline_command)
  {
    unescape_string(&unesc_command, &unesc_command_length, cmdline_command);
    if (NULL == unesc_command)
    {
      fprintf(stderr, "Un-escaping command string failed\n");
      result = 1;
      goto m_fail;
    }
  }

  /* Un-escape response termination */
  if (NULL != cmdline_resp_termination)
  {
    unescape_string(&unesc_resp_term, &unesc_resp_term_length, cmdline_resp_termination);
    if (NULL == unesc_resp_term)
    {
      fprintf(stderr, "Un-escaping response termination string failed\n");
      result = 1;
      goto m_fail;
    }
  }

  /* Send command */
  test_poll_response(sockfd, cmdline_count,
                     unesc_command, unesc_command_length,
                     unesc_resp_term, unesc_resp_term_length);

  /* Success */
  result = 0;

m_fail:
  if (NULL != unesc_resp_term)
  {
    free(unesc_resp_term);
    unesc_resp_term = NULL;
  }
  if (NULL != unesc_command)
  {
    free(unesc_command);
    unesc_command = NULL;
  }
  if (sockfd >= 0)
  {
    close(sockfd);
    sockfd = -1;
  }
  return result;
}
