/*
 * test_comms_print_buffer.c
 *
 *  Created on: May 11, 2015
 *      Author: ian
 */
#include <stdio.h>
#include <rinexec/el_runcmd.h>
#include <cutest/CuTest.h>
#include "comms_print_buffer.h"

#include "test_comms_print_buffer.h"


#define TEST_BUF_SIZE 10


/** Check buffer wrap in a comms print buffer */
static void test_comms_print_buffer_print_struct(const struct s_comms_print_buffer *cpb)
{
  size_t i;

  printf("s_comms_print_buffer.buf_in=%u\n"
         "s_comms_print_buffer.buf_out=%u\n"
         "s_comms_print_buffer.buf_size=%u\n",
         (unsigned) cpb->buf_in, (unsigned) cpb->buf_out, (unsigned) cpb->buf_size);

  for (i = 0; i < cpb->buf_size; i++)
  {
    printf("%04u: %02x", (unsigned) i, cpb->buffer[i]);
    if (i == cpb->buf_in)
      printf(" in");
    if (i == cpb->buf_out)
      printf(" out");
    printf("\n");
  }
}


/** Check that we can create a comms print buffer */
static void test_comms_print_buffer_create(CuTest* tc)
{
  struct s_comms_print_buffer *cpb;

  /* Create buffer */
  cpb = comms_print_buffer_create(TEST_BUF_SIZE);
  CuAssertPtrNotNull(tc, cpb);

  /* Check contents */
  CuAssertIntEquals(tc, 0, cpb->buf_in);
  CuAssertIntEquals(tc, 0, cpb->buf_out);
  CuAssertIntEquals(tc, TEST_BUF_SIZE, cpb->buf_size);
  CuAssertPtrNotNull(tc, cpb->buffer);

  /* Destroy it, can't really test this */
  comms_print_buffer_destroy(cpb);
}


/** Check bad parameters */
static void test_comms_print_buffer_bad_param(CuTest* tc)
{
  struct s_comms_print_buffer *cpb;
  bool result;
  char ch;

  /* Create buffer */
  cpb = comms_print_buffer_create(TEST_BUF_SIZE);
  CuAssertPtrNotNull(tc, cpb);

  /* Check some bad parameters */
  result = comms_print_buffer_add_char(NULL, 'a');
  CuAssertTrue(tc, !result);

  (void) comms_print_buffer_add_char(cpb, 'a');
  result = comms_print_buffer_get_char(cpb, NULL);
  CuAssertTrue(tc, !result);

  result = comms_print_buffer_get_char(NULL, &ch);
  CuAssertTrue(tc, !result);

  /* Destroy it, can't really test this */
  comms_print_buffer_destroy(cpb);
}


/** Check reading and writing to a comms print buffer */
static void test_comms_print_buffer_read_write(CuTest* tc)
{
  struct s_comms_print_buffer *cpb;
  size_t i;
  bool result;
  char ch;
  #define STRLEN 100
  char str[STRLEN];

  /* Create buffer */
  cpb = comms_print_buffer_create(TEST_BUF_SIZE);
  CuAssertPtrNotNull(tc, cpb);

  /* Read empty should fail */
  result = comms_print_buffer_get_char(cpb, &ch);
  CuAssertTrue(tc, !result);

  /* Fill it up */
  for (i = 0; i < TEST_BUF_SIZE - 1; i++)
  {
    result = comms_print_buffer_add_char(cpb, 'a' + i);
    snprintf(str, STRLEN, "comms_print_buffer_add_char() failed adding char %u of %u\n", (unsigned) i + 1, TEST_BUF_SIZE);
    CuAssert(tc, str, result);
  }

  /* Overfill should fail */
  result = comms_print_buffer_add_char(cpb, 'a' + i);
  CuAssertTrue(tc, !result);

  /* Read it back */
  for (i = 0; i < TEST_BUF_SIZE - 1; i++)
  {
    result = comms_print_buffer_get_char(cpb, &ch);
    snprintf(str, STRLEN, "comms_print_buffer_get_char() failed getting char %u of %u\n", (unsigned) i + 1, TEST_BUF_SIZE);
    CuAssert(tc, str, result);
    CuAssertIntEquals(tc, 'a' + i, ch);
  }

  /* Read extra should fail */
  result = comms_print_buffer_get_char(cpb, &ch);
  CuAssertTrue(tc, !result);

  /* Destroy it, can't really test this */
  comms_print_buffer_destroy(cpb);
}


/** Check buffer wrap in a comms print buffer */
static void test_comms_print_buffer_wrap(CuTest* tc)
{
  struct s_comms_print_buffer *cpb;
  size_t i;
  bool result;
  char ch;
  #define STRLEN 100
  char str[STRLEN];

  /* Create buffer */
  cpb = comms_print_buffer_create(TEST_BUF_SIZE);
  CuAssertPtrNotNull(tc, cpb);

  /* Add and get characters */
  for (i = 0; i < TEST_BUF_SIZE * 4; i++)
  {
    result = comms_print_buffer_add_char(cpb, 'a' + (i % 26));
    snprintf(str, STRLEN, "comms_print_buffer_add_char() failed adding char %u of %u\n", (unsigned) i + 1, TEST_BUF_SIZE * 4);
    CuAssert(tc, str, result);

    test_comms_print_buffer_print_struct(cpb);

    result = comms_print_buffer_get_char(cpb, &ch);
    snprintf(str, STRLEN, "comms_print_buffer_get_char() failed getting char %u of %u\n", (unsigned) i + 1, TEST_BUF_SIZE * 4);
    CuAssert(tc, str, result);
    CuAssertIntEquals(tc, 'a' + (i % 26), ch);
  }

  /* Destroy it, can't really test this */
  comms_print_buffer_destroy(cpb);
}



CuSuite *CuGetSuite_comms_print_buffer(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_comms_print_buffer_create);
  SUITE_ADD_TEST(suite, test_comms_print_buffer_bad_param);
  SUITE_ADD_TEST(suite, test_comms_print_buffer_read_write);
  SUITE_ADD_TEST(suite, test_comms_print_buffer_wrap);

  return suite;
}
