/*
 * traceprint.c
 *
 *  Created on: Mar 16, 2012
 *      Author: ian
 */
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcinterface.h>
#include "comms_print.h"
#include "comms_print_trigger.h"
#include "dsd.h"
#include "iface.h"
#include "iface_sd.h"
#include "iface_cal.h"
#include "database.h"
#include "adcthread.h"
#include "app_db_user.h"
#include "commonstrings.h"

#include "traceprint.h"


static enum ePrintoutType current_printout_type = ptNone;


/**
 * Initiate a traceable print by requesting a traceable reading
 * from the ADC library
 */
static bool request_traceable_reading(enum ePrintoutType print_type, bool immediate)
{
  bool result = false;

  if (ptNone == current_printout_type)
  {
    if (AdcGetTraceableRdg(&adcSuperStructure, true, false, false, immediate) == ADC_OP_STATUS_IN_PROGRESS)
    {
      adc_operation_counter = AdcGetCurrentCommandCnt(&adcSuperStructure, true);
      current_printout_type = print_type;
      result = true;
    }
  }
  return result;
}


/**
 * Initiate a traceable print by requesting a traceable reading
 * from the ADC library. Generate errors on failure.
 */
static bool request_traceable_reading_ui(enum ePrintoutType print_type, bool display_errors, bool immediate)
{
  bool result;

  if (ptNone != current_printout_type)
  {
    if (display_errors)
    {
      /* Traceable weight already requested */
      freeze_display_and_annunciators(true);
      error_strings("TRACE", "BUSY");
      freeze_display_and_annunciators(false);
    }
    return false;
  }

  /* Request a traceable reading */
  result = request_traceable_reading(print_type, immediate);
  if (!result && display_errors)
  {
    freeze_display_and_annunciators(true);
    error_strings(adc_str, "TRACE");
    freeze_display_and_annunciators(false);
  }
  return result;
}


/** Purge the DSD, and display errors */
static void traceable_printout_purge_dsd(bool user_interaction)
{
  /* Automatic purging */
  if (!dsd_purge() && user_interaction)
  {
    freeze_display_and_annunciators(true);
    error_strings(dsd_err_str, "PURGE");
    freeze_display_and_annunciators(false);
  }
}


/**
 * Called to complete a traceable print after the
 * traceable reading has been created by the ADC library
 */
static void traceable_printout_finish(const char *s, const AdcTraceableWeight *atw, bool user_interaction, bool success, sem_t *print_finish_sem)
{
  if (!comms_print_is_traceable(current_printout_type))
  {
    return;
  }

  if (success)
  {
    /* Write record to DSD module on success */
    if (dsd_is_running())
    {

      /* Purge records if DSD is full */
      if (dsd_is_full())
      {
        if (db_app_user.dsd_auto_purge)
        {
          /* Automatic purging */
          traceable_printout_purge_dsd(user_interaction);
        }
        else if (user_interaction)
        {
          /* Ask user before purging DSD */
          freeze_display_and_annunciators(true);
          prompt(dsd_str, "FULL");
          if (ask("CLR", true))
          {
            traceable_printout_purge_dsd(user_interaction);
          }
          freeze_display_and_annunciators(false);
        }
        else
        {
          /* Do nothing */
        }
      }

      /* Now write to DSD */
      if (!dsd_write(atw) && user_interaction)
      {
        freeze_display_and_annunciators(true);
        error_strings(dsd_err_str, "WRITE");
        freeze_display_and_annunciators(false);
      }
    }

    /* Do the printout on success */
    switch (current_printout_type)
    {
      case ptStandard:
      case ptStandardComms:
        (void) comms_print_trigger_traceable_standard(atw, print_finish_sem);
        break;
      case ptSpecialRemote:
        (void) comms_print_trigger_traceable_special_remote(s, atw, print_finish_sem);
        break;
      default:
        break;
    }

    /* Clear interlock */
    comms_print_interlock_clear(&comms_print_interlock);
  }

  /* Always clear the printout type, regardless of success */
  current_printout_type = ptNone;
}


static void is_traceable_ready(bool *failed, bool *ready)
{
  int8u cmd_status;
  int8u cmd;

  AdcMutexLock(&adcSuperStructure);
  cmd = AdcGetCurrentCommand(&adcSuperStructure, false);
  cmd_status = AdcGetCurrentOpStatus(&adcSuperStructure, false);
  AdcMutexUnlock(&adcSuperStructure);

  if (ADC_OP_CMD_TRACEABLE_RDG == cmd)
  {
    if (ADC_OP_STATUS_IN_PROGRESS == cmd_status)
    {
      *ready = false;
      *failed = false;
    }
    else if (ADC_OP_STATUS_SUCCESS == cmd_status)
    {
      *ready = true;
      *failed = false;
    }
  }
  else
  {
    *ready = false;
    *failed = true;
  }
}


static bool wait_for_traceable(int16u timeout)
{
  bool failed;
  bool ready;
  int32u start_time;

  for (failed = false, ready = false, start_time = OSTimeGet();
       !failed && !ready && (OSTimeGet() < start_time + timeout); )
  {
    is_traceable_ready(&failed, &ready);
    if (!failed && !ready)
    {
      OSTimeDly(1);
    }
  }

  return (!failed && ready);
}


static bool request_and_wait_for_traceable(enum ePrintoutType print_type, AdcTraceableWeight *atw)
{
  bool result = false;

  if (request_traceable_reading(print_type, false))
  {
    if (wait_for_traceable(OS_TICKS_PER_SEC))
    {
      AdcMutexLock(&adcSuperStructure);
      *atw = adcSuperStructure.adcTraceableWeight;
      AdcMutexUnlock(&adcSuperStructure);
      result = true;
    }
  }
  return result;
}


/**
 * Initiate a (traceable) standard printout.
 *
 * For use with the comms, will wait for a traceable weight
 * and then print using it.
 */
void traceable_printout_initiate_standard_comms(sem_t *print_finished_sem)
{
  AdcTraceableWeight atw;
  bool success;

  success = request_and_wait_for_traceable(ptStandardComms, &atw);
  traceable_printout_finish(NULL, &atw, false, success, print_finished_sem);
}


/**
 * Initiate a (traceable) special remote printout.
 *
 * For use with the comms, will wait for a traceable weight
 * and then print using it.
 */
void traceable_printout_initiate_special_remote_comms(sem_t *print_finished_sem, const char *s)
{
  AdcTraceableWeight atw;
  bool success;

  success = request_and_wait_for_traceable(ptSpecialRemote, &atw);
  traceable_printout_finish(s, &atw, false, success, print_finished_sem);
}


/**
 * Finish up (traceable) standard printout, now that we have the traceable weight. For use with the UI.
 *
 * Must be called after a call to @ref traceable_printout_initiate_standard_ui(), regardless of the
 * success of the ADC library request traceable weight operation.
 */
void traceable_printout_finish_ui(const AdcTraceableWeight *atw, bool success)
{
  if (!comms_print_is_comms_initiated(current_printout_type))
  {
    traceable_printout_finish(NULL, atw, true, success, NULL);
  }
}


/**
 * Initiate a (traceable) standard printout.
 *
 * Will display errors if requested. For use with the UI, will return immediately.
 *
 * Must be matched with a call to @ref traceable_printout_finish_ui() after the ADC
 * library has returned with success and traceable weight, or failure
 */
void traceable_printout_initiate_standard_ui(bool display_errors, bool immediate)
{
  /* Requires a traceable weight with a standard printout */
  request_traceable_reading_ui(ptStandard, display_errors, immediate);
}
