/*
 * unitswitch.c
 *
 *  Created on: Apr 10, 2018
 *      Author: ian
 */
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <rinweigh/adcdatatypes.h>
#include "adcthread.h"
#include "db_special.h"
#include "app_db_user.h"

#include "unitswitch.h"


/* Unit switching magic numbers */
static const uint8_t  altToAdcUnits[NUM_ALT_UNITS]     = {ADC_UNITS_NONE, ADC_UNITS_GRAMS, ADC_UNITS_KILOGRAMS, ADC_UNITS_POUNDS, ADC_UNITS_OUNCES, ADC_UNITS_NONE};
static const uint32_t gramsConvFact[NUM_ALT_UNITS]     = {1000000L,       1000000L,        1000L,               2204L,            35274L,           1000000L};
static const uint32_t kilogramsConvFact[NUM_ALT_UNITS] = {1000000L,       1000000000L,     1000000L,            2204623L,         35273960L,        1000000L};
static const uint32_t poundsConvFact[NUM_ALT_UNITS]    = {1000000L,       453592000L,      453592L,             1000000L,         16000000L,        1000000L};
static const uint32_t ouncesConvFact[NUM_ALT_UNITS]    = {1000000L,       28349500L,       28350L,              62500L,           1000000L,         1000000L};


/** A table to allow us to lookup unit switched and disp/gross/net/tare weights */
static AdcRdg *const usw_adcrdgs_lookup[ADC_NUM_UNIT_SWITCH][wt_num] =
{
  /* ADC_UNIT_SWITCH_PRIMARY */
  {
    &adcSuperStructure.adcDataWorking.dispRdg,      /* uswt_displayed */
    &adcSuperStructure.adcDataWorking.grossRdg,     /* uswt_gross */
    &adcSuperStructure.adcDataWorking.netRdg,       /* uswt_net */
    &adcSuperStructure.adcDataWorking.anyTare,      /* uswt_tare */
  },
  /* ADC_UNIT_SWITCH_SECONDARY */
  {
    &adcSuperStructure.adcDataWorking.altDispRdg,   /* uswt_displayed */
    &adcSuperStructure.adcDataWorking.altGrossRdg,  /* uswt_gross */
    &adcSuperStructure.adcDataWorking.altNetRdg,    /* uswt_net */
    &adcSuperStructure.adcDataWorking.altAnyTare,   /* uswt_tare */
  },
  /* ADC_UNIT_SWITCH_TERTIARY */
  {
    &adcSuperStructure.adcDataWorking.tertDispRdg,  /* uswt_displayed */
    &adcSuperStructure.adcDataWorking.tertGrossRdg, /* uswt_gross */
    &adcSuperStructure.adcDataWorking.tertNetRdg,   /* uswt_net */
    &adcSuperStructure.adcDataWorking.tertAnyTare,  /* uswt_tare */
  }
};


/** Get a weight from the adcSuperStructure in an AdcRdg */
bool unitsw_get_adcrdg(AdcRdg *rdgout, enum eAdcUnitSwitch usw, enum e_weight_type uwt)
{
  if (NULL == rdgout || usw >= ADC_NUM_UNIT_SWITCH || uwt >= wt_num)
    return false;

  *rdgout = *usw_adcrdgs_lookup[usw][uwt];
  return true;
}


/** Get the currently configured units for primary/secondary/tertiary units */
enum eAdcUnits unitsw_get_units(enum eAdcUnitSwitch usw)
{
  enum eAdcUnits units;

  switch (usw)
  {
    default:
    case ADC_UNIT_SWITCH_PRIMARY:
      units = adcSuperStructure.adcUserDatabase.units;
      break;
    case ADC_UNIT_SWITCH_SECONDARY:
      units = adcSuperStructure.adcInternalData.altUnits;
      break;
    case ADC_UNIT_SWITCH_TERTIARY:
      units = adcSuperStructure.adcInternalData.tertUnits;
      break;
  }
  return units;
}


/** Get the currently configured decimal point position for primary/secondary/tertiary units */
uint8_t unitsw_get_decimal_point(enum eAdcUnitSwitch usw)
{
  uint8_t dp;

  switch (usw)
  {
    default:
    case ADC_UNIT_SWITCH_PRIMARY:
      dp = adcSuperStructure.adcUserDatabase.dpPos;
      break;
    case ADC_UNIT_SWITCH_SECONDARY:
      dp = adcSuperStructure.adcUserDatabase.dpPos + adcSuperStructure.adcInternalData.altDpInc;
      break;
    case ADC_UNIT_SWITCH_TERTIARY:
      dp = adcSuperStructure.adcUserDatabase.dpPos + adcSuperStructure.adcInternalData.tertDpInc;
      break;
  }
  return dp;
}


/** Get the resolution length for primary/secondary/tertiary units */
uint8_t unitsw_get_resolution_length(enum eAdcUnitSwitch usw)
{
  uint8_t reslen;

  switch (usw)
  {
    default:
    case ADC_UNIT_SWITCH_PRIMARY:
      reslen = AdcResolutionLengths[adcSuperStructure.adcUserDatabase.resolution[0]];
      break;
    case ADC_UNIT_SWITCH_SECONDARY:
      reslen = adcSuperStructure.adcInternalData.altResLen[0];
      break;
    case ADC_UNIT_SWITCH_TERTIARY:
      reslen = adcSuperStructure.adcInternalData.tertResLen[0];
      break;
  }
  return reslen;
}


/** Get the current preset tare value for primary/secondary/tertiary units */
void unitsw_get_preset_tare(AdcRdg *preset_tare, enum eAdcUnitSwitch usw)
{
  /* Get the tare value based on unit switch */
  switch (usw)
  {
    default:
    case ADC_UNIT_SWITCH_PRIMARY:
      *preset_tare = adcSuperStructure.adcDataWorking.presetTare;
      break;
    case ADC_UNIT_SWITCH_SECONDARY:
      *preset_tare = adcSuperStructure.adcDataWorking.altAnyTare;
      break;
    case ADC_UNIT_SWITCH_TERTIARY:
      *preset_tare = adcSuperStructure.adcDataWorking.tertAnyTare;
      break;
  }

  /* If the tare is not a preset tare, then preset tare value is 0 */
  if (!preset_tare->info.status.bit.preTare)
  {
    preset_tare->rdg = 0;
    preset_tare->info.status.bit.preTare = 1;
    preset_tare->info.status.bit.tare = 0;
  }
}


/** Convert ADC display mode into ADC unit switch */
enum eAdcUnitSwitch unitsw_displaymode_to_unitswitch(enum eAdcDisplayMode dispmode)
{
  enum eAdcUnitSwitch usw;

  switch (dispmode)
  {
    case ADC_DISPLAY_MODE_WEIGHT:
    case ADC_DISPLAY_MODE_PIECES: /* I guess */
    default:
      usw = ADC_UNIT_SWITCH_PRIMARY;
      break;
    case ADC_DISPLAY_MODE_ALT_W:
      usw = ADC_UNIT_SWITCH_SECONDARY;
      break;
    case ADC_DISPLAY_MODE_TERT_W:
      usw = ADC_UNIT_SWITCH_TERTIARY;
      break;
  }
  return usw;
}


/** Convert ADC unit switch into ADC display mode */
enum eAdcDisplayMode unitsw_unitswitch_to_displaymode(enum eAdcUnitSwitch usw)
{
  enum eAdcDisplayMode dispmode;

  switch (usw)
  {
    default:
    case ADC_UNIT_SWITCH_PRIMARY:
      dispmode = ADC_DISPLAY_MODE_WEIGHT;
      break;
    case ADC_UNIT_SWITCH_SECONDARY:
      dispmode = ADC_DISPLAY_MODE_ALT_W;
      break;
    case ADC_UNIT_SWITCH_TERTIARY:
      dispmode = ADC_DISPLAY_MODE_TERT_W;
      break;
  }
  return dispmode;
}


long app_final_to_final_alt(long rdg)
{
  int32s altRdg, counts;
  int8u range, hiRes;

  hiRes = adcSuperStructure.adcRuntimeDatabase.hiRes;
  range = AdcGetRangeForFinalWeight(&adcSuperStructure, false, rdg, hiRes);
  counts = AdcFinalToCounts(&adcSuperStructure, false, rdg, hiRes);
  altRdg = AdcCountsToFinalAlt(&adcSuperStructure, false, counts, range, hiRes);

  return altRdg;
}


long app_final_to_final_tert(long rdg)
{
  int32s tertRdg, counts;
  int8u range, hiRes;

  hiRes = adcSuperStructure.adcRuntimeDatabase.hiRes;
  range = AdcGetRangeForFinalWeight(&adcSuperStructure, false, rdg, hiRes);
  counts = AdcFinalToCounts(&adcSuperStructure, false, rdg, hiRes);
  tertRdg = AdcCountsToFinalTert(&adcSuperStructure, false, counts, range, hiRes);

  return tertRdg;
}


/** Convert final units of usw_from to usw_to units */
long unitsw_convert_final(long rdg, enum eAdcUnitSwitch usw_from, enum eAdcUnitSwitch usw_to, bool hires_from, bool hires_to)
{
  int32s counts;
  int32s final;
  int8u range;

  /* Convert rdg in usw_from units to counts */
  switch (usw_from)
  {
    case ADC_UNIT_SWITCH_PRIMARY:
    default:
      counts = AdcFinalToCounts(&adcSuperStructure, false, rdg, hires_from);
      break;
    case ADC_UNIT_SWITCH_SECONDARY:
      counts = AdcFinalToCountsAlt(&adcSuperStructure, false, rdg, hires_from);
      break;
    case ADC_UNIT_SWITCH_TERTIARY:
      counts = AdcFinalToCountsTert(&adcSuperStructure, false, rdg, hires_from);
      break;
  }

  /* Get range */
  range = AdcGetRangeForCounts(&adcSuperStructure, false, counts);

  /* Convert counts into rdg in usw_to units to counts */
  switch (usw_to)
  {
    case ADC_UNIT_SWITCH_PRIMARY:
    default:
      final = AdcCountsToFinal(&adcSuperStructure, false, counts, range, hires_to);
      break;
    case ADC_UNIT_SWITCH_SECONDARY:
      final = AdcCountsToFinalAlt(&adcSuperStructure, false, counts, range, hires_to);
      break;
    case ADC_UNIT_SWITCH_TERTIARY:
      final = AdcCountsToFinalTert(&adcSuperStructure, false, counts, range, hires_to);
      break;
  }
  return final;
}


/** Convert final units of usw_from to usw_to units in an AdcRdg */
void unitsw_convert_final_adcrdg(AdcRdg *rdgout, const AdcRdg *rdgin, enum eAdcUnitSwitch usw_from, enum eAdcUnitSwitch usw_to)
{
  memset(rdgout, 0, sizeof(*rdgout));

  /* Copy status and id */
  rdgout->info.id = rdgin->info.id;
  rdgout->info.status = rdgin->info.status;

  /* Convert reading value */
  rdgout->rdg = unitsw_convert_final(rdgin->rdg, usw_from, usw_to, rdgin->info.status.bit.hiRes, rdgin->info.status.bit.hiRes);

  /* Set units and decimal point in output units */
  rdgout->info.unitdp.bit.units = unitsw_get_units(usw_to);
  rdgout->info.unitdp.bit.dp = unitsw_get_decimal_point(usw_to);
}


/** Initialise alternate units */
static void init_alt_units(enum eAltUnits alt_unit, enum eAdcUnitSwitch usw)
{
  int32s conv;
  int8u units;

  /* Check the alternate units are sane */
  if (alt_unit >= NUM_ALT_UNITS)
    alt_unit = ALT_UNITS_OFF;

  /* Set alternate units if possible */
  switch (alt_unit)
  {
    case ALT_UNITS_OFF:
      conv = 1000000L;
      units = ADC_UNITS_NONE;
      break;

    case ALT_UNITS_CUSTOM:
      conv = db_app_user.unit_conv_factor * 1000L;
      units = ADC_UNITS_NONE;
      break;

    default:
      switch (adcSuperStructure.adcUserDatabase.units)
      {
        case ADC_UNITS_GRAMS:
          conv = gramsConvFact[alt_unit];
          units = altToAdcUnits[alt_unit];
          break;

        case ADC_UNITS_KILOGRAMS:
          conv = kilogramsConvFact[alt_unit];
          units = altToAdcUnits[alt_unit];
          break;

        case ADC_UNITS_POUNDS:
          conv = poundsConvFact[alt_unit];
          units = altToAdcUnits[alt_unit];
          break;

        case ADC_UNITS_OUNCES:
          conv = ouncesConvFact[alt_unit];
          units = altToAdcUnits[alt_unit];
          break;

        default:
          conv = 1000000L;
          units = ADC_UNITS_NONE;
          break;
      }
      break;
  }
  AdcSetAltUnits(&adcSuperStructure, true, true, conv, units, usw);
}


/** Initialise unit switching */
void init_unit_switch(void)
{
  /* init secondary units */
  init_alt_units(db_app_user.units_sec, ADC_UNIT_SWITCH_SECONDARY);
  /* init tertiary units */
  init_alt_units(db_app_user.units_ter, ADC_UNIT_SWITCH_TERTIARY);
}
