/*
 * verify_tested.c
 *
 * Used to verify that each board has been tested (at some point)
 * Vario will use this to confirm a sample of each batch has been tested before it leaves the factory
 *
 */
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_msg.h>
#include <rinblock/rb_scratch.h>
#include "app_db_factory.h"
#include "apperror.h"
#include "we_debug.h"

#include "verify_tested.h"


enum PcbTypes {PcbAll=0, PcbDisplay, PcbProcessor, PcbRear};


#define MODULE_TESTED_OFFSET 0x00
#define MODULE_TEST_UNTESTED_FLAG 0xFF
#define MODULE_TEST_FIRST_MASK 0x01
#define MODULE_TEST_SECOND_MASK 0x02
#define MODULE_TEST_WRITTEN_FLAG 0xA0
#define MODULE_TEST_WRITTEN_MASK 0xF0


/** Find rinblock index for the test type, -1 for not found */
static int FindRinblockForTest(enum PcbTypes testType)
{
  int i;
  int rbidx;

  for (i = 0, rbidx = -1; i < MAX_RINBLOCKS; i++)
  {
    if ((PcbRear == testType && RB_IFACE_TYPE_WE21_CONN == rinblocks[i].tag.interface_type) ||
        (PcbDisplay == testType && RB_IFACE_TYPE_WE21_DISP == rinblocks[i].tag.interface_type) ||
        (PcbDisplay == testType && RB_IFACE_TYPE_C500_DISP == rinblocks[i].tag.interface_type))
    {
      rbidx = i;
      break;
    }
  }
  return rbidx;
}


static void VerifyModule(uint8_t test_result, const char *modname)
{
  if ((test_result & MODULE_TEST_WRITTEN_MASK) != MODULE_TEST_WRITTEN_FLAG)
  {
    set_error(APP_ERR_UNTESTED, "VerifyModule: setting APP_ERR_UNTESTED because %s board has not been tested\n", modname);
  }
  else
  {
    /* Check result of first test */
    if (0 != (test_result & 0x04))
    {
      set_error(APP_ERR_UNTESTED, "VerifyModule: setting APP_ERR_UNTESTED because %s board has not been through first test\n", modname);
    }
    else if (0 != (test_result & 0x01))
    {
      set_error(APP_ERR_UNTESTED, "VerifyModule: setting APP_ERR_UNTESTED because %s board has failed first test\n", modname);
    }

    /* Check result of second test */
    if (0 != (test_result & 0x08))
    {
      set_error(APP_ERR_UNTESTED, "VerifyModule: setting APP_ERR_UNTESTED because %s board has not been through second test\n", modname);
    }
    else if (0 != (test_result & 0x02))
    {
      set_error(APP_ERR_UNTESTED, "VerifyModule: setting APP_ERR_UNTESTED because %s board has failed second test\n", modname);
    }
  }
}


/** Confirm that a rinblock module has been tested, by a variable stored in the scratchpad/eeprom */
static void VerifyModuleTested(enum PcbTypes testType)
{
  int rbidx;
  uint8_t data;
  const char *modname;

  modname = (PcbDisplay == testType) ? "display" : (PcbRear == testType) ? "connector" : "unknown";

  rbidx = FindRinblockForTest(testType);
  if (rbidx < 0)
  {
    /* Hopefully some other check will generate an error if this module is missing */
    rdebug_printf(LOG_ERR, "VerifyModuleTested: can't find rinblock index for %s\n", modname);
    return;
  }
  else
  {

    /* Read back previously written test result */
    if (!RinBlockReadScratchPad(&rinblocks[rbidx], &data, MODULE_TESTED_OFFSET, sizeof(data)))
    {
      /* Hopefully some other check will generate an error if this module is not readable */
      rdebug_printf(LOG_ERR, "VerifyModuleTested: can't read scratchpad for rinblock %s\n", modname);
      return;
    }
    else
    {
      VerifyModule(data, modname);
    }
  }
}


static void VerifyProcessorBoardTested(void)
{
  VerifyModule(db_app_factory.test_result, "processor");
}


/** Verify that all boards have been tested (first & second), and set error if not. Requires rinblocks to be up running. */
void VerifyHardwareTested(void)
{
  VerifyModuleTested(PcbRear);
  VerifyModuleTested(PcbDisplay);
  VerifyProcessorBoardTested();
}
