#
# Convert autoconf config.h PACKAGE_VERSION into APP_VERSION_*
#
import getopt
import sys
import os
import re
import subprocess


def getpackageversion(file):
    """"Get the package version from config.h"""
    for line in file:
        m = re.match(pattern="^.define *PACKAGE_VERSION *\"([0-9]+)\.([0-9]+)\.([0-9]+)\".*$", string=line)
        if (m is not None and m.group(1) is not None and m.group(2) is not None and m.group(3) is not None):
            version_major = int(m.group(1))
            version_minor = int(m.group(2))
            version_revision = int(m.group(3))
            return (version_major, version_minor, version_revision)
    raise Exception("PACKAGE_VERSION not found in input file")


def getgitversion():
    try:
	# Get latest tag plus commits and last commit hash
        output = subprocess.Popen(["git", "describe"], stdout=subprocess.PIPE).communicate()
        result = output[0].rstrip('\n')

        # Are there uncommitted changes
        subprocess.Popen(["git", "update-index", "--refresh", "--unmerged"], stdout=subprocess.PIPE).communicate()
        output = subprocess.Popen(["git", "diff-index", "--name-only", "HEAD"], stdout=subprocess.PIPE).communicate()
        if (output[0] != ""):
            result = result + "-dirty"
    except:
        result = "Git version unavailable"
    return result


header_file_template = \
"\
/* This file is automatically generated by appversion.py. DO NOT EDIT. */\n\
#ifndef {}\n\
#define {} 1\n\
\n\
\n\
/** The major part of the version (0-999) */\n\
#define APP_VERSION_MAJOR {}\n\
\n\
\n\
/** The minor part of the version (0-999) */\n\
#define APP_VERSION_MINOR {}\n\
\n\
\n\
/** The revision part of the version (0-999) */\n\
#define APP_VERSION_REVISION {}\n\
\n\
\n\
/** The version reported by git */\n\
#define APP_GIT_VERSION \"{}\"\n\
\n\
\n\
#endif /* {} */\n\
"


def write_h_file(outfile, hfilename, version_major, version_minor, version_revision, gitversion):
    """Write registers out to H file"""
    hname = "{}_H_".format(os.path.basename(os.path.splitext(hfilename)[0]).upper())

    header_file_text = header_file_template.format(hname, hname, version_major, version_minor, version_revision, gitversion, hname)
    outfile.write(header_file_text)


def main(infilename, outhfilename):
    with open(infilename, 'r') as infile:
        version_major, version_minor, version_revision = getpackageversion(file=infile)

    gitversion = getgitversion()

    if (outhfilename is not None):
        with open(outhfilename, 'w') as outfile:
            write_h_file(outfile=outfile, hfilename=outhfilename,
                         version_major=version_major, version_minor=version_minor, version_revision=version_revision, gitversion=gitversion)
    else:
        write_h_file(outfile=sys.stdout, hfilename="APP_VERSION",
                     version_major=version_major, version_minor=version_minor, version_revision=version_revision, gitversion=gitversion)


def useage():
    print("appversion.py [OPTIONS] config.h [outfile.h]")
    print("  -h --help    Print this help")
    print("")
    print("Convert the PACKAGE_VERSION define in autoconf config.h into C500 version header file.")
    print("outfile.h is optional. If not specified, output will be written to stdout.")


if (__name__ == '__main__'):
    try:
        opts, args = getopt.getopt(sys.argv[1:], "h", ["help"])
    except getopt.GetoptError as err:
        print(str(err))
        sys.exit(2)
    output = None
    verbose = False
    for o, a in opts:
        if o in ("-h", "--help"):
            useage()
            sys.exit(0)
        else:
            assert False, "Unhandled option"

    if (len(args) < 1):
        print("No input file specified. Try --help")
        sys.exit(1)
    infilename = args[0]
    temp = os.path.splitext(infilename)[0]
    tablename = os.path.basename(temp)
    if (len(args) < 2):
        outhfilename = None
    else:
        outhfilename = args[1]

    main(infilename=infilename, outhfilename=outhfilename)
