/*
 * watchdog.c
 *
 *  Created on: Apr 2, 2014
 *      Author: ian
 */
#include <linux/watchdog.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>
#include <stdbool.h>
#include "we_debug.h"

#include "watchdog.h"


#define WATCHDOG_MAGIC_CLOSE_CHAR 'V'


static int watchdog_fd = -1;


/** Kick the watchdog, this should be called at regular intervals */
void watchdog_kick(void)
{
  int dummy;

  if (watchdog_fd >= 0)
  {
    (void) ioctl(watchdog_fd, WDIOC_KEEPALIVE, &dummy);
    /* alternative way to kick the watchdog: res = write(watchdog_fd, "data", 4); */
  }
}


/** Initialise the watchdog using specified device (eg /dev/watchdog) */
void watchdog_init(const char *device)
{
  bool result;

  if (NULL == device || 0 == strlen(device))
  {
    rdebug_printf(LOG_INFO, "watchdog_init: watchdog is disabled because device not specified\n");
    result = false;
    goto wi_fail;
  }

  /* If only... O_CLOEXEC */
  watchdog_fd = open(device, O_WRONLY);
  if (watchdog_fd < 0)
  {
    rdebug_printf(LOG_ERR, "watchdog_init: error opening watchdog device \"%s\" (errno %d)\n", device, errno);
    result = false;
    goto wi_fail;
  }
  if (fcntl(watchdog_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    rdebug_printf(LOG_ERR, "watchdog_init: error setting FD_CLOEXEC on watchdog device \"%s\" (errno %d)\n", device, errno);
    result = false;
    goto wi_fail;
  }

  /* Success */
  watchdog_kick();
  result = true;

wi_fail:
  if (!result && watchdog_fd >= 0)
  {
    close(watchdog_fd);
    watchdog_fd = -1;
  }
}


/**
 * Shut down the watchdog.
 *
 * This sends the magic close char before closing the device, to prevent reboots once the
 * application has quit. This allows us to do firmware upgrades without having a reboot
 * in the middle of it
 */
void watchdog_destroy(void)
{
  const char ch = WATCHDOG_MAGIC_CLOSE_CHAR;
  int res;

  if (watchdog_fd >= 0)
  {
    /*
     * Disable watchdog with magic close character.
     * This is important, because if we are doing
     * a firmware upgrade, then we want the application
     * to quit, which will stop the watchdog being kicked.
     * The last thing we want is the system hard rebooting in
     * the middle of a firmware upgrade.
     */
    res = write(watchdog_fd, &ch, sizeof(ch));
    if (res != sizeof(ch))
    {
      rdebug_printf(LOG_ERR, "watchdog_destroy: error disabling watchdog, write result %d, errno %d\n", res, errno);
    }
    res = close(watchdog_fd);
    if (res < 0)
      rdebug_printf(LOG_ERR, "watchdog_destroy: close() failed, result %d, errno %d\n", res, errno);
    watchdog_fd = -1;
  }
}
