/*
 * weconfig.c
 *
 *  Created on: 22/03/2012
 *      Author: ian
 */
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ezxml.h>
#include "we_debug.h"

#include "we_config.h"


/** Struct of configuration read from the config file */
struct s_hw_config we_configuration;


static ezxml_t get_config_hardware_device(ezxml_t ezxml_config, const char *device_name)
{
  ezxml_t hardware = ezxml_get(ezxml_config, "config", 0, "hardware", -1);
  ezxml_t device;
  const char *xml_device_name;

  WE_CONFIG_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "get_config_hardware_device: found hardware node: %s\n", (NULL != hardware) ? "yes" : "no");

  for (device = ezxml_child(hardware, "device" ); (NULL != device); device = device->next)
  {
    xml_device_name = ezxml_attr(device, "name");
    if (0 == strcmp(xml_device_name, device_name))
    {
      WE_CONFIG_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "get_config_hardware_device: found node for device \"%s\"\n", device_name);
      return device;
    }
  }

  rdebug_printf(LOG_ERR, "get_config_hardware_device: no node found for device \"%s\"\n", device_name);
  return (ezxml_t) NULL;
}


bool get_config_hardware_device_text(ezxml_t ezxml_config, const char *device_name, char **device_setting)
{
  bool result = false;
  ezxml_t device;
  char *dev_setting;

  device = get_config_hardware_device(ezxml_config, device_name);
  if (NULL != device && NULL != device->txt)
  {
    dev_setting = strdup(device->txt);
    if (NULL != dev_setting)
    {
      WE_CONFIG_DEBUG_PRINTF(RDEBUG_LVL_BASIC, LOG_INFO, "get_config_hardware_device: setting for \"%s\": \"%s\"\n", device_name, dev_setting);
      *device_setting = dev_setting;
      result = true;
    }
  }
  return result;
}


static bool get_config_str_to_int(const char *name, const char *value, int *converted)
{
  char *endptr;

  errno = 0;
  endptr = (char *) value;
  *converted = strtol(value, &endptr, 0);
  if (0 != errno || '\0' != *endptr)
  {
    rdebug_printf(LOG_ERR, "get_config_str_to_int: error converting %s \"%s\" to integer, errno %d\n", name, value, errno);
    return false;
  }
  return true;
}


static bool get_config_usb_gadget_settings(ezxml_t ezxml_config, const char *device_name, struct s_hw_config *config)
{
  ezxml_t gadget_settings;
  ezxml_t setting;

  gadget_settings = get_config_hardware_device(ezxml_config, device_name);
  if (NULL == gadget_settings)
    return false;

  /* Device entry */
  setting = ezxml_get(gadget_settings, "deventry", -1);
  if (NULL == setting || NULL == setting->txt)
  {
    rdebug_printf(LOG_ERR, "get_config_usb_gadget_settings: no entry found for deventry\n");
    return false;
  }
  config->usb_gadget_device_name = strdup(setting->txt);
  if (NULL == config->usb_gadget_device_name)
  {
    rdebug_printf(LOG_ERR, "get_config_usb_gadget_settings: strdup(\"%s\") for deventry failed\n", setting->txt);
    return false;
  }

  /* Device major number */
  setting = ezxml_get(gadget_settings, "devmajor", -1);
  if (NULL == setting || NULL == setting->txt)
  {
    rdebug_printf(LOG_ERR, "get_config_usb_gadget_settings: no entry found for devmajor\n");
    return false;
  }
  if (!get_config_str_to_int("devmajor", setting->txt, &config->usb_gadget_device_major))
  {
    return false;
  }

  /* Device minor number */
  setting = ezxml_get(gadget_settings, "devminor", -1);
  if (NULL == setting || NULL == setting->txt)
  {
    rdebug_printf(LOG_ERR, "get_config_usb_gadget_settings: no entry found for devminor\n");
    return false;
  }
  if (!get_config_str_to_int("devminor", setting->txt, &config->usb_gadget_device_minor))
  {
    return false;
  }

  /* USB Vendor ID */
  setting = ezxml_get(gadget_settings, "usbvid", -1);
  if (NULL == setting || NULL == setting->txt)
  {
    rdebug_printf(LOG_ERR, "get_config_usb_gadget_settings: no entry found for usbvid\n");
    return false;
  }
  if (!get_config_str_to_int("usbvid", setting->txt, &config->usb_gadget_vid))
  {
    return false;
  }

  /* USB Product ID */
  setting = ezxml_get(gadget_settings, "usbpid", -1);
  if (NULL == setting || NULL == setting->txt)
  {
    rdebug_printf(LOG_ERR, "get_config_usb_gadget_settings: no entry found for usbpid\n");
    return false;
  }
  if (!get_config_str_to_int("usbpid", setting->txt, &config->usb_gadget_pid))
  {
    return false;
  }

  return true;
}


/** Load the configuration from the configuration file */
bool load_configuration(const char *filename, struct s_hw_config *config)
{
  ezxml_t ezxml_config = NULL;
  bool result = false;

  memset(config, 0, sizeof(struct s_hw_config));
  config->usb_gadget_device_major = -1;
  config->usb_gadget_device_minor = -1;
  config->usb_gadget_vid = -1;
  config->usb_gadget_pid = -1;

  ezxml_config = ezxml_parse_file(filename);
  if (NULL == ezxml_config)
  {
    rdebug_printf(LOG_ERR, "load_configuration: xml parsing of file %s failed\n", filename);
    goto lc_fail;
  }

  if (!get_config_hardware_device_text(ezxml_config, "we21hw", &config->we2111_hardware_device))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "rinblock_iic", &config->rinblock_iic_device))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "adc_spi", &config->adc_spi_device))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "serial1", &config->serial1_device))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "serial2", &config->serial2_device))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "nvram", &config->nvram_device1))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "nvram2", &config->nvram_device2))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "factorydb1", &config->factorydb1_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "factorydb2", &config->factorydb2_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "userdb1", &config->userdb1_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "userdb2", &config->userdb2_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "licencedb1", &config->licencedb1_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "licencedb2", &config->licencedb2_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "dsd", &config->dsd_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "changelog", &config->changelog_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "partition_mount_location", &config->partition_mount_location))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "watchdog", &config->watchdog_device))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "opkg", &config->opkg_executable))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "ethiface", &config->ethernet_interface))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "devmnt_notification_pipe", &config->device_mounter_notification_pipe))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "uds_network_file", &config->uds_network_file))
    goto lc_fail;
  if (!get_config_hardware_device_text(ezxml_config, "uds_autoout_file", &config->uds_autoout_file))
    goto lc_fail;
  if (!get_config_usb_gadget_settings(ezxml_config, "gadget", config))
    goto lc_fail;

  /* Success */
  result = true;

lc_fail:
  if (NULL != ezxml_config)
  {
    ezxml_free(ezxml_config);
    ezxml_config = NULL;
  }
  if (!result)
  {
    free_configuration(config);
  }
  return result;
}


/** Free up the contents of the configuration structure */
void free_configuration(struct s_hw_config *config)
{
  if (NULL != config->we2111_hardware_device)
  {
    free(config->we2111_hardware_device);
    config->we2111_hardware_device = NULL;
  }
  if (NULL != config->rinblock_iic_device)
  {
    free(config->rinblock_iic_device);
    config->rinblock_iic_device = NULL;
  }
  if (NULL != config->adc_spi_device)
  {
    free(config->adc_spi_device);
    config->adc_spi_device = NULL;
  }
  if (NULL != config->serial1_device)
  {
    free(config->serial1_device);
    config->serial1_device = NULL;
  }
  if (NULL != config->serial2_device)
  {
    free(config->serial2_device);
    config->serial2_device = NULL;
  }
  if (NULL != config->nvram_device1)
  {
    free(config->nvram_device1);
    config->nvram_device1 = NULL;
  }
  if (NULL != config->nvram_device2)
  {
    free(config->nvram_device2);
    config->nvram_device2 = NULL;
  }
  if (NULL != config->factorydb1_file)
  {
    free(config->factorydb1_file);
    config->factorydb1_file = NULL;
  }
  if (NULL != config->factorydb2_file)
  {
    free(config->factorydb2_file);
    config->factorydb2_file = NULL;
  }
  if (NULL != config->userdb1_file)
  {
    free(config->userdb1_file);
    config->userdb1_file = NULL;
  }
  if (NULL != config->userdb2_file)
  {
    free(config->userdb2_file);
    config->userdb2_file = NULL;
  }
  if (NULL != config->licencedb1_file)
  {
    free(config->licencedb1_file);
    config->licencedb1_file = NULL;
  }
  if (NULL != config->licencedb2_file)
  {
    free(config->licencedb2_file);
    config->licencedb2_file = NULL;
  }
  if (NULL != config->dsd_file)
  {
    free(config->dsd_file);
    config->dsd_file = NULL;
  }
  if (NULL != config->changelog_file)
  {
    free(config->changelog_file);
    config->changelog_file = NULL;
  }
  if (NULL != config->partition_mount_location)
  {
    free(config->partition_mount_location);
    config->partition_mount_location = NULL;
  }
  if (NULL != config->watchdog_device)
  {
    free(config->watchdog_device);
    config->watchdog_device = NULL;
  }
  if (NULL != config->opkg_executable)
  {
    free(config->opkg_executable);
    config->opkg_executable = NULL;
  }
  if (NULL != config->ethernet_interface)
  {
    free(config->ethernet_interface);
    config->ethernet_interface = NULL;
  }
  if (NULL != config->device_mounter_notification_pipe)
  {
    free(config->device_mounter_notification_pipe);
    config->device_mounter_notification_pipe = NULL;
  }
  if (NULL != config->uds_network_file)
  {
    free(config->uds_network_file);
    config->uds_network_file = NULL;
  }
  if (NULL != config->uds_autoout_file)
  {
    free(config->uds_autoout_file);
    config->uds_autoout_file = NULL;
  }
  if (NULL != config->usb_gadget_device_name)
  {
    free(config->usb_gadget_device_name);
    config->usb_gadget_device_name = NULL;
  }
}
