#include <string.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/debug.h>
#include <rinhw/adc_ads1216.h>

#include "_adc_ads1216_local.h"
#include <rinhw/adc_ads1216_local.h>


#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
static void adc24bit_dump_registers(int fd);
#endif


static int adc24bit_linux_transfer(int fd, const int8u *cmd_data, int8u cmd_len, int8u *ret_data, int8u ret_len)
{
  #define NUM_SPI_TRANSFERS 2
  struct spi_ioc_transfer transfer[NUM_SPI_TRANSFERS];
  int result;

  /* Setup write, need different cast for 32/64bit */
#if __LP64__
  transfer[0].tx_buf = (uint64_t) cmd_data;
#else
  transfer[0].tx_buf = (uint32_t) cmd_data;
#endif
  transfer[0].rx_buf = 0;
  transfer[0].len = cmd_len;
  transfer[0].speed_hz = ADS1216_SCLK_T1_FREQ_HZ;
  transfer[0].delay_usecs = ADS1216_RREG_T6_MICROSEC;
  transfer[0].bits_per_word = 8;
  transfer[0].cs_change = 0;
  transfer[0].pad = 0;

  /* Test to see if ADC will return data */
  if (ret_len > 0 && NULL != ret_data)
  {
    /* Setup read, need different cast for 32/64bit */
    transfer[1].tx_buf = 0;
#if __LP64__
    transfer[1].rx_buf = (uint64_t) ret_data;
#else
    transfer[1].rx_buf = (uint32_t) ret_data;
#endif
    transfer[1].len = ret_len;
    transfer[1].speed_hz = ADS1216_SCLK_T1_FREQ_HZ;
    transfer[1].delay_usecs = ADS1216_RREG_T11_MICROSEC;
    transfer[1].bits_per_word = 8;
    transfer[1].cs_change = 0;
    transfer[1].pad = 0;

    /*
     * There are two alternate calls to ioctl here because
     * when compiling for M68K, passing in a variable to
     * SPI_IOC_MESSAGE will cause the link to fail with
     * undefined reference to __invalid_size_argument_for_IOC.
     * Passing in a compile time constant works...
     *
     * This is a bug in the (older) M68k uclibc headers, or
     * maybe the kernel headers are old, or maybe both.
     * The problem exists in <asm-generic/ioctl.h> where
     * _IOC_TYPECHECK is incorrectly defined. This has been
     * fixed in newer kernel headers.
     */
    result = ioctl(fd, SPI_IOC_MESSAGE(NUM_SPI_TRANSFERS), transfer);
  }
  else
  {
    result = ioctl(fd, SPI_IOC_MESSAGE(1), transfer);
  }
  return (result < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/* Execute a command on the ADC */
static int adc24bit_execute_command(int fd, const int8u *cmd_data, int8u cmd_len, int8u *ret_data, int8u ret_len)
{
#if 0
  /* Write the command */
  adc_ads1216_spi_write_bytes(cmd_data, cmd_len);

  /* Test to see if ADC will return data */
  if (ret_len > 0 && NULL != ret_data)
  {
    /* Make the DIO pin an input ready to clock the data in */
    adc_spi_configure_data_pin(FALSE);

    /* Delay until the ADC is read to send us data, then clock it in */
    DELAY_MICROSECOND(ADS1216_RREG_T6_MICROSEC);
    adc_ads1216_spi_read_bytes(ret_data, ret_len);
  }
  DELAY_MICROSECOND(ADS1216_RREG_T11_MICROSEC);
  return RIO_SUCCESS;
#endif
  return adc24bit_linux_transfer(fd, cmd_data, cmd_len, ret_data, ret_len);
}


/** Get a reading from the ADC */
static int adc24bit_get_reading(int fd, int32s *data)
{
  int8u aui8Buffer[sizeof(int32u)];
  int8u i;
  int32u ui32Data;

  /* Send the RDATA command, and read back 24bits of ADC reading */
  aui8Buffer[0] = ADS1216_CMD_RDATA;
  if (adc24bit_execute_command(fd, aui8Buffer, 1, aui8Buffer, 3) < 0)
  {
    return RIO_ERROR;
  }

  /* Transfer data from the buffer into the long in the order that the ADC spits it out (MSB first) */
  for (i = 0, ui32Data = 0; i < 3; i++)
  {
    ui32Data |= ((int32u) (aui8Buffer[2 - i])) << (i * 8);
  }

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("adc24bit_get_reading: data bytes are 0:0x%02x, 1:0x%02x, 2:0x%02x\n", aui8Buffer[0], aui8Buffer[1], aui8Buffer[2]);
  DEBUG_PRINTF("adc24bit_get_reading: assembled result: 0x%lx\n", ui32Data);
#endif

  /* Sign extend from 24 bits to 32bits */
  if (0x00800000 == (ui32Data & 0x00800000))
  {
    ui32Data |= 0xFF000000;
  }
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("adc24bit_get_reading: sign extended result: %ld\n", (int32s) ui32Data);
#endif
  *data = (int32s) ui32Data;
  return RIO_SUCCESS;
}


/** Write a 1, 2 or 3 byte register on the ADC */
static int adc24bit_write_register(int fd, int8u reg, int32u data, int8u len)
{
  int8u aui8Buffer[2 + sizeof(data)];
  int8u i;

  /* Biggest register on the ADC is 24 bits */
  if (len > 3)
  {
    len = 3;
  }

  /* Setup write command with register to read, and setup length byte of number of registers to read */
  aui8Buffer[0] = (int8u) (reg | ADS1216_CMD_WREG);
  aui8Buffer[1] = (int8u) (len - 1);

  /* Put data into buffer in the order that the ADC expects (low order first - yes this is different order to RDATA command) */
  for (i = 0; i < len; i++)
  {
    aui8Buffer[i + 2] = (int8u) (data & 0x000000FF);
    data >>= 8;
  }

  /* Write the register write command and data to the ADC */
  return adc24bit_execute_command(fd, aui8Buffer, (int8u) ( 2 + len ), NULL, 0);
}


/** Read a 1, 2 or 3 byte register on the ADC */
static int adc24bit_read_register(int fd, int8u reg, int32u *pdata, int8u len)
{
  int8u aui8CmdBuffer[2];
  int8u aui8DataBuffer[sizeof(*pdata)];
  int8u i;
  int32u ui32Data;

  /* Biggest register on the ADC is 24 bits */
  if (len > 3)
  {
    len = 3;
  }

  /* Setup read command with register to read, and setup length byte of number of registers to read */
  aui8CmdBuffer[0] = (int8u) (reg | ADS1216_CMD_RREG);
  aui8CmdBuffer[1] = (int8u) (len - 1);

  /* Read the register */
  if (adc24bit_execute_command(fd, aui8CmdBuffer, 2, aui8DataBuffer, len) < 0)
  {
    return RIO_ERROR;
  }

  /* Transfer data from the buffer into the long in the order that the ADC spits it out (low order first - yes this is different order to RDATA command) */
  ui32Data = 0;
  for (i = 0; i < len; i++)
  {
    ui32Data |= ((int32u) (aui8DataBuffer[i])) << (i * 8);
  }
  *pdata = ui32Data;
  return RIO_SUCCESS;
}


int adc24bit_read_register8(int fd, int8u ui8Reg, int8u *pui8Data)
{
  int res;
  int32u ui32Data;

  res = adc24bit_read_register(fd, ui8Reg, &ui32Data, sizeof(*pui8Data));
  *pui8Data = (int8u) ui32Data;
  return res;
}


int adc24bit_write_register8(int fd, int8u ui8Reg, int8u ui8Data)
{
  return adc24bit_write_register(fd, ui8Reg, (int32u) ui8Data, sizeof(int8u));
}


static int adc24bit_read_register24(int fd, int8u ui8Reg, int32u *pui32Data)
{
  return adc24bit_read_register(fd, ui8Reg, pui32Data, sizeof(*pui32Data) - 1);
}


static int adc24bit_write_register24(int fd, int8u ui8Reg, int32u ui32Data)
{
  return adc24bit_write_register(fd, ui8Reg, ui32Data, sizeof(ui32Data) - 1);
}


/** Set the gain in the programmable gain amplifier (PGA) */
static int adc24bit_set_pga(int fd,  int level)
{
  int8u data, acr_value;

  /* Figure out the setting, and check for invalid ones */
  switch (level)
  {
    case IOCTL_ADC24BIT_PGA_GAIN1:
      data = REG_ACR_PGA1;
      break;
    case IOCTL_ADC24BIT_PGA_GAIN2:
      data = REG_ACR_PGA2;
      break;
    case IOCTL_ADC24BIT_PGA_GAIN4:
      data = REG_ACR_PGA4;
      break;
    case IOCTL_ADC24BIT_PGA_GAIN8:
      data = REG_ACR_PGA8;
      break;
    case IOCTL_ADC24BIT_PGA_GAIN16:
      data = REG_ACR_PGA16;
      break;
    case IOCTL_ADC24BIT_PGA_GAIN32:
      data = REG_ACR_PGA32;
      break;
    case IOCTL_ADC24BIT_PGA_GAIN64:
      data = REG_ACR_PGA64;
      break;
    case IOCTL_ADC24BIT_PGA_GAIN128:
      data = REG_ACR_PGA128;
      break;
    default:
      return RIO_ERROR;
  }

  /* Read/modify/write the setting */
  adc24bit_read_register8(fd, ADS1216_REG_ACR, &acr_value);
  acr_value &= ~REG_ACR_PGA_MASK;
  acr_value |= (int8u) REG_ACR_PGA(data);
  return adc24bit_write_register8(fd, ADS1216_REG_ACR, acr_value);
}


static int adc24bit_set_burnout_current_source(int fd, int enable)
{
  int8u acr_value;

  adc24bit_read_register8(fd, ADS1216_REG_ACR, &acr_value);
  if (IOCTL_ADC24BIT_BURNOUT_OFF == enable)
  {
    acr_value &= (int8u) ~REG_ACR_BOCS_MASK;
  }
  else
  {
    acr_value |= REG_ACR_BOCS_ON;
  }
  return adc24bit_write_register8(fd, ADS1216_REG_ACR, acr_value);
}


/** Set the decimation ratio of the ADC, frequency in 0.1Hz (125 = 12.5Hz) */
static int adc24bit_set_decimation_ratio(int fd, int frequency)
{
  int32u decimation;
  int8u decimation_lo;
  int8u mdec;
  int8u speed;
  int8u setup;

  /* First work out decimation value using SPEED_DIV128 */
  speed = REG_SETUP_SPEED_DIV128;
  decimation = ADS1216_FMOD_SPEED128 * 10 / frequency;
  if (decimation > 2047)
  {
    /* Can't do that slow using DIV128, but what about DIV256? */
    decimation = ADS1216_FMOD_SPEED256 * 10 / frequency;
    speed = REG_SETUP_SPEED_DIV256;
  }

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("adc24bit_set_decimation_ratio: @Fosc %ld, SPEED %d, freq %d (0.1Hz) gives dec ratio of %d\n",
               ADS1216_FOSC, (speed == REG_SETUP_SPEED_DIV256) ? 1 : 0, frequency, decimation);
#endif

  /* Make sure it's in range */
  if (decimation > 2047 || decimation < 20)
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_set_decimation_ratio: decimation ratio out of range\n");
#endif
    return RIO_ERROR;
  }

  /* Set the SPEED bit in the setup register */
  adc24bit_read_register8(fd, ADS1216_REG_SETUP, &setup);
  setup &= ~REG_SETUP_SPEED_MASK;
  setup |= speed;
  adc24bit_write_register8(fd, ADS1216_REG_SETUP, setup);

  /* Split into two parts for DEC and MDEC registers */
  decimation_lo = (int8u) (decimation & 0x00FF);
  mdec = (int8u) MDEC_VALUE_WITH_DEC((int8u) (int16u) ((decimation >> 8) & 0x00FF));

  /* Write the two registers */
  adc24bit_write_register8(fd, ADS1216_REG_DEC0, decimation_lo);
  adc24bit_write_register8(fd, ADS1216_REG_MDEC1, mdec);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF( "adc24bit_set_decimation_ratio: decimation ratio set\n" );
#endif
  return RIO_SUCCESS;
}


/** Set the decimation ratio of the ADC */
static int adc24bit_perform_self_zero_calibration(int fd)
{
  const int8u aui8CmdBuffer[1] = { ADS1216_CMD_SELFOCAL };

  /* Execute the self zero calibration command */
  adc24bit_execute_command(fd, aui8CmdBuffer, 1, NULL, 0);

  return 7;
}


#if 0
/** Check the ADC is responding by writing 2 bytes in the RAM and reading it back */
bool adc24bit_check_adc_present(void)
{
  int8u aui8CmdBuffer[4];
  int8u aui8DataBuffer[2];
  bool bPresent;

  /* Setup write RAM command to bank 0 with some dummy data */
  aui8CmdBuffer[0] = (int8u) (0 | ADS1216_CMD_WRAM);
  aui8CmdBuffer[1] = (int8u) (2 - 1);
  aui8CmdBuffer[2] = 0xA5;
  aui8CmdBuffer[3] = 0x5A;

  /* Execute the write RAM command */
  adc24bit_execute_command(aui8CmdBuffer, 4, NULL, 0);

  /* Setup read RAM command to bank 0 to read back the data we just wrote */
  aui8CmdBuffer[0] = (int8u) (0 | ADS1216_CMD_RRAM);
  aui8CmdBuffer[1] = (int8u) (2 - 1);

  /* Execute the read RAM command */
  adc24bit_execute_command(aui8CmdBuffer, 2, aui8DataBuffer, 2);

  /* Did we read back what we wrote, in which case the ADC is present */
  bPresent = (aui8DataBuffer[0] == 0xA5 && aui8DataBuffer[1] == 0x5A);

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_check_adc_present: read back data: 0x%02x, 0x%02x, present %u\n",
               aui8DataBuffer[0], aui8DataBuffer[1], bPresent);
#endif
  return bPresent;
}
#endif


/** Check the ADC is responding by reading the SETUP register and checking the ID bits */
bool adc24bit_check_adc_present(int fd)
{
  int8u data;
  bool bPresent;

  /* Read the setup register */
  if (adc24bit_read_register8(fd, ADS1216_REG_SETUP, &data) < 0)
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_check_adc_present: adc24bit_read_register8(ADS1216_REG_SETUP) failed\n");
#endif
    return false;
  }
  bPresent = (REG_SETUP_ID_VALUE1 == (data & REG_SETUP_ID_MASK) ||
              REG_SETUP_ID_VALUE2 == (data & REG_SETUP_ID_MASK));
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_check_adc_present: SETUP ID bits: expected 0x%02x or 0x%02x, actual 0x%02x, present %u\n",
               REG_SETUP_ID_VALUE1, REG_SETUP_ID_VALUE2, data, bPresent);
#endif
  return bPresent;
}


/** Get the ID from the setup register */
static bool adc24bit_get_chip_id(int fd, int *chip_id)
{
  int8u mask;
  int8u data;

  /* Read the setup register */
  if (adc24bit_read_register8(fd, ADS1216_REG_SETUP, &data) < 0)
  {
    return false;
  }

  /* Shuffle the setup bits right */
  data &= REG_SETUP_ID_MASK;
  for (mask = REG_SETUP_ID_MASK; 0 == (mask & 0x01); mask >>= 1)
  {
    data >>= 1;
  }

  *chip_id = data;
  return true;
}


#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
static void adc24bit_dump_registers(int fd)
{
  int8u data;
  int32u uldata;

  DEBUG_PRINTF("adc24bit_dump_registers: registers:\n");

  adc24bit_read_register8(fd, ADS1216_REG_SETUP, &data);
  DEBUG_PRINTF("SETUP: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_MUX, &data);
  DEBUG_PRINTF("MUX: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_ACR, &data);
  DEBUG_PRINTF("ACR: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_DIR, &data);
  DEBUG_PRINTF("DIR: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_DIO, &data);
  DEBUG_PRINTF("DIO: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_DEC0, &data);
  DEBUG_PRINTF("DEC0: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_MDEC1, &data);
  DEBUG_PRINTF("MDEC: 0x%02x\n", data);
  adc24bit_read_register24(fd, ADS1216_REG_OCR0, &uldata);
  DEBUG_PRINTF("OCR: 0x%06lx\n", uldata);
  adc24bit_read_register24(fd, ADS1216_REG_FSR0, &uldata);
  DEBUG_PRINTF("FSR: 0x%06lx\n", uldata);
}
#endif


static int adc24bit_init_hardware(struct sAdcSpiDevice *dev_info)
{
  int fd;

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_init_hardware: syncing spi\n");
#endif
  fd = dev_info->fd;
  if (NULL != dev_info->adc24bit_sync_spi)
  {
    if (dev_info->adc24bit_sync_spi(dev_info) < 0)
    {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("adc24bit_init_hardware: sync spi failed\n");
#endif
      return RIO_ERROR;
    }
  }

  /* Configure registers with defaults */
  adc24bit_write_register8(fd, ADS1216_REG_SETUP, SETUP_VALUE);
  adc24bit_write_register8(fd, ADS1216_REG_MUX, MUX_VALUE);
  adc24bit_write_register8(fd, ADS1216_REG_ACR, (int8u) ACR_VALUE_DEFAULT);
  /* Don't care about IDAC1, IDAC2, ODAC */

  /* Set IOs as all inputs as default, adc24bit_user_hw_init() can change this as necessary */
  adc24bit_write_register8(fd, ADS1216_REG_DIR, 0xFF);
  adc24bit_set_decimation_ratio(fd, DEFAULT_FREQ); /* Sets DEC0 and MDEC1 */
  adc24bit_write_register24(fd, ADS1216_REG_OCR0, OCR_VALUE_DEFAULT);
  adc24bit_write_register24(fd, ADS1216_REG_FSR0, FSR_VALUE_DEFAULT);

  if (NULL != dev_info->adc24bit_user_hw_init)
  {
    dev_info->adc24bit_user_hw_init(dev_info);
  }

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  adc24bit_dump_registers(fd);
#endif
  return RIO_SUCCESS;
}


/** Initialise linux device driver */
static int adc24bit_init_linux_spi_device(struct sAdcSpiDevice *dev_info)
{
  int fd;
  int result;
  int8u mode;
  int8u bits;
  int32u speed;

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_init_linux_device() called\n");
#endif

  fd = open(dev_info->filename, O_RDWR);
  if (0 > fd)
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_init_linux_device: failed opening %s, errno:%d, %s\n",
                 dev_info->filename, errno, strerror(errno));
#endif
    return RIO_ERROR;
  }
  if (fcntl(fd, F_SETFD, FD_CLOEXEC) < 0)
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_init_linux_device: error setting FD_CLOEXEC on device %s, errno %d, %s\n",
                 dev_info->filename, errno, strerror(errno));
#endif
    close(fd);
    return RIO_ERROR;
  }
  dev_info->fd = fd;

  /* Set SPI mode */
  mode = 0; /* ( ~SPI_LOOP & ~SPI_CS_HIGH & ~SPI_LSB_FIRST & ~SPI_CPHA & ~SPI_CPOL ); */
  mode |= (SPI_3WIRE | SPI_MODE_3);
  result = ioctl(fd, SPI_IOC_WR_MODE, &mode);
  if (result < 0)
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_init_linux_device: can't set SPI write mode, errno:%d, %s\n",
                 errno, strerror(errno));
#endif
    return RIO_ERROR;
  }

  /* Set bits per word */
  bits = 8;
  result = ioctl(fd, SPI_IOC_WR_BITS_PER_WORD, &bits);
  if (result < 0)
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_init_linux_device: can't set SPI bits per word, errno:%d, %s\n",
                 errno, strerror(errno));
#endif
    return RIO_ERROR;
  }

  /* Set max speed in Hz */
  speed = 500000; /* TODO Set from constant */
  result = ioctl(fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed);
  if (result < 0)
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_init_linux_device: can't set SPI speed, errno:%d, %s\n",
                 errno, strerror(errno));
#endif
    return RIO_ERROR;
  }

  /* All done, return success */
  return RIO_SUCCESS;
}


/** Initialise ADC hardware */
static int adc24bit_local_init(struct t_io_dev *io_dev)
{
  struct sAdcSpiDevice *dev_info = (struct sAdcSpiDevice *) io_dev->dev_data;
  int result;

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_local_init called\n");
#endif

  /* Setup SPI */
  result = adc24bit_init_linux_spi_device(dev_info);
  if (result < 0)
  {
    return result;
  }

  /* Setup ADS1216 */
  result = adc24bit_init_hardware(dev_info);
  if (result < 0)
  {
    return result;
  }

  /* Its a block device */
  io_dev->status &= ~O_DEVCHAR;

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_local_init finished\n");
#endif
  return RIO_SUCCESS;
}


/** Destroy ADC hardware */
static int adc24bit_local_destroy(struct t_io_dev *io_dev)
{
  struct sAdcSpiDevice *dev_info = (struct sAdcSpiDevice *) io_dev->dev_data;

  io_dev->dev_data = NULL;

  if (NULL != dev_info)
  {
    /* Close device file handle */
    if (dev_info->fd >= 0)
    {
      close(dev_info->fd);
      dev_info->fd = -1;
    }

    if (NULL != dev_info->filename)
    {
      /* Free device filename */
      free(dev_info->filename);
      dev_info->filename = NULL;
    }
    free(dev_info);
    dev_info = NULL;
  }

  return RIO_SUCCESS;
}


static int adc24bit_local_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct sAdcSpiDevice *dev_info = (struct sAdcSpiDevice *) io_dev->dev_data;
  int32s *pi32Data = (int32s *) buf;

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("adc24bit_local_read called\n");
#endif

  /* Check request is reasonable */
  if ((io_dev->posn + len) > sizeof(int32s))
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_local_read: position or length bad\n");
#endif
    return RIO_ERROR;
  }

  /*
   * We don't adc24bit_wait_for_reading() here, because
   * there are cases where you want to wait for
   * DRDY interrupts, but not necessarily read from the ADC.
   * An example of this is the self zero calibration -
   * you need to wait 7 DRDY cycles, but accessing the ADC
   * during this period can cause undefined behavior
   * (I've seen readings that don't match the expected for the
   * calibration, and won't come good until the ADC gets re-initialised)
   *
   * Instead use IOCTL_ADC24BIT_WAIT_FOR_DATA_READY, followed by a rread()
   */

  /* Get data from ADC */
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("adc24bit_local_read: reading data from ADC\n");
#endif
  if (adc24bit_get_reading(dev_info->fd, pi32Data) < 0)
  {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("adc24bit_local_read: reading from ADC failed\n");
#endif
    return RIO_ERROR;
  }

  /* Update position and return result */
  io_dev->posn += len;
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("adc24bit_local_read: success, result %d\n", len);
#endif
  return (int) len;
}


static int adc24bit_local_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_local_write called\n");
#endif

  return RIO_ERROR;
}


static long adc24bit_local_filelength(struct t_io_dev *io_dev)
{
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_local_filelength called\n");
#endif

  return (long) sizeof(long);
}


/*
 * adc24bit_ioctl -
 *
 * ADC specific configuration
 */
static int adc24bit_local_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct sAdcSpiDevice *dev_info = (struct sAdcSpiDevice *) io_dev->dev_data;
  int result = RIO_SUCCESS;
  unsigned long ldata;
  int32u i32data;
  int ndata;

#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_local_ioctl called, func %d\n", func);
#endif

  /* Function dispatch */
  switch (func)
  {

    /* Set zero register */
    case IOCTL_ADC24BIT_SET_ZERO_REGISTER:
      ldata = va_arg(*arglist, unsigned long);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_SET_ZERO_REGISTER: level %lu\n", ldata);
#endif
      result = adc24bit_write_register24(dev_info->fd, ADS1216_REG_OCR0, ldata);
      break;

    /* Get zero register */
    case IOCTL_ADC24BIT_GET_ZERO_REGISTER:
      result = adc24bit_read_register24(dev_info->fd, ADS1216_REG_OCR0, &i32data);
      *(va_arg(*arglist, unsigned long *)) = i32data;
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_GET_ZERO_REGISTER: level %lu\n", (unsigned long) i32data);
#endif
      break;

    /* Set span register */
    case IOCTL_ADC24BIT_SET_SPAN_REGISTER:
      ldata = va_arg(*arglist, unsigned long);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_SET_SPAN_REGISTER: level %lu\n", ldata);
#endif
      result = adc24bit_write_register24(dev_info->fd, ADS1216_REG_FSR0, ldata);
      break;

    /* Get zero register */
    case IOCTL_ADC24BIT_GET_SPAN_REGISTER:
      result = adc24bit_read_register24(dev_info->fd, ADS1216_REG_FSR0, &i32data);
      *(va_arg(*arglist, unsigned long *)) = i32data;
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_GET_SPAN_REGISTER: level %lu\n", (unsigned long) i32data);
#endif
      break;

    /* Set the PGA gain register */
    case IOCTL_ADC24BIT_SET_PGA_GAIN:
      ndata = va_arg(*arglist, int);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_SET_PGA_GAIN: level %d\n", ndata);
#endif
      result = adc24bit_set_pga(dev_info->fd, ndata);
      break;

    /* Set the decimation ratio register */
    case IOCTL_ADC24BIT_SET_DECIMATION_RATIO:
      ndata = va_arg(*arglist, int);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_SET_DECIMATION_RATIO: freq %d (0.1Hz)\n", ndata);
#endif
      result = adc24bit_set_decimation_ratio(dev_info->fd, ndata);
      break;

    /* Set 4 or 6 wire mode */
    case IOCTL_ADC24BIT_SET_LCWIRE_MODE:
      ndata = va_arg(*arglist, int);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_SET_LCWIRE_MODE: level %u\n", ndata);
#endif
      if (NULL != dev_info->adc24bit_set_loadcell_wire_mode)
      {
        result = dev_info->adc24bit_set_loadcell_wire_mode(dev_info, ndata);
      }
      break;

    /* Force reset of ADC */
    case IOCTL_ADC24BIT_FORCE_RESET:
      result = adc24bit_init_hardware(dev_info);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_FORCE_RESET: reset completed\n");
#endif
      break;

    /* Figure out if the ADC is still responding */
    case IOCTL_ADC24BIT_IS_RESPONDING:
      ndata = (int) adc24bit_check_adc_present(dev_info->fd);
      *(va_arg(*arglist, int *)) = ndata;
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_IS_RESPONDING: responding %u\n", ndata);
#endif
      break;

    /* Enable/disable the burnout current source */
    case IOCTL_ADC24BIT_SET_BURNOUTCURRENTSOURCE:
      ndata = va_arg(*arglist, int);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_SET_BURNOUTCURRENTSOURCE: enabled %u\n", ndata);
#endif
      result = adc24bit_set_burnout_current_source(dev_info->fd, ndata);
      break;

    /* Perform a self zero calibration */
    case IOCTL_ADC24BIT_PERFORM_SELF_ZERO_CALIBRATION:
      result = adc24bit_perform_self_zero_calibration(dev_info->fd);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_PERFORM_SELF_ZERO_CALIBRATION: self zero calibration, result %d\n", result);
#endif
      break;

    /* Get the chip id */
    case IOCTL_ADC24BIT_GET_ID:
      ndata = 0;
      result = adc24bit_get_chip_id(dev_info->fd, &ndata);
      *(va_arg(*arglist, int *)) = ndata;
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_GET_ID: result %d, chip_id: %d\n", result, ndata);
#endif
      break;

    /* Wait for data ready interrupt, or timeout */
    case IOCTL_ADC24BIT_WAIT_FOR_DATA_READY:
      if (NULL == dev_info->adc24bit_wait_for_reading)
      {
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("adc24bit_local_read: no wait for reading function provided\n");
#endif
        result = RIO_ERROR;
      }
      else
      {
        ldata = va_arg(*arglist, unsigned long);
        result = dev_info->adc24bit_wait_for_reading(dev_info, ldata);
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
        if (result < 0)
          DEBUG_PRINTF("adc24bit_local_read: timeout waiting for data\n");
#endif
      }
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_ADC24BIT_WAIT_FOR_DATA_READY: result %d\n", result);
#endif
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = dev_info->fd;
      result = RIO_SUCCESS;
      break;

    /* Error */
    default:
      result = RIO_ERROR;
      break;
  }

  /* Done, return result */
#if HW_DEBUG(ADC_ADS1216_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc24bit_local_ioctl finished with result %d\n", result);
#endif
  return result;
}


/**
 * Function to create and initialise a local 24bit ADC device
 *
 * @param io_dev is the IO device to be created. Not NULL.
 * @param device_filename is the SPI device that the ADC is connected to. Not NULL.
 * @param adc24bit_user_hw_init is a user specific hardware initialisation called as part of the ADC hardware initialisation. May be called multiple times. Can be NULL.
 * @param adc24bit_sync_spi is a user implemented function to synchronise ADC SPI, only required when a chip select is not used. Not NULL.
 * @param adc24bit_set_loadcell_wire_mode is a device specific loadcell wire mode function pointer, can be NULL.
 * @param adc24bit_wait_for_reading is a user implemented function to wait for an ADC reading (either polling status register, or wait on hardware interrupt)
 */
int rio_create_device_adc24bit_local(struct t_io_dev *io_dev, const char *device_filename,
                                     t_adc24bit_user_hw_init adc24bit_user_hw_init,
                                     t_adc24bit_sync_spi adc24bit_sync_spi,
                                     t_adc24bit_set_loadcell_wire_mode adc24bit_set_loadcell_wire_mode,
                                     t_adc24bit_wait_for_reading adc24bit_wait_for_reading)
{
  struct sAdcSpiDevice *adc_spi_device_info = NULL;

  /* Some checks */
  if (NULL == io_dev || NULL == device_filename || NULL == adc24bit_sync_spi)
  {
    goto rcdal_fail;
  }

  /* Create device structure */
  adc_spi_device_info = malloc(sizeof(struct sAdcSpiDevice));
  if (NULL == adc_spi_device_info)
  {
    goto rcdal_fail;
  }

  /* Allocate copy of filename */
  adc_spi_device_info->filename = strdup(device_filename);
  if (NULL == adc_spi_device_info->filename)
  {
    goto rcdal_fail;
  }
  adc_spi_device_info->fd = -1;

  /* Set function pointers */
  adc_spi_device_info->adc24bit_user_hw_init = adc24bit_user_hw_init;
  adc_spi_device_info->adc24bit_sync_spi = adc24bit_sync_spi;
  adc_spi_device_info->adc24bit_set_loadcell_wire_mode = adc24bit_set_loadcell_wire_mode;
  adc_spi_device_info->adc24bit_wait_for_reading = adc24bit_wait_for_reading;

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) adc_spi_device_info;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = adc24bit_local_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = adc24bit_local_read;
  io_dev->dev_write = adc24bit_local_write;
  io_dev->dev_filelength = adc24bit_local_filelength;
  io_dev->dev_ioctl = adc24bit_local_ioctl;
  io_dev->dev_destroy = adc24bit_local_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
    goto rcdal_fail;
  }
  return RIO_SUCCESS;

rcdal_fail:
  if (NULL != adc_spi_device_info)
  {
    if (adc_spi_device_info->fd >= 0)
    {
      close(adc_spi_device_info->fd);
      adc_spi_device_info->fd = -1;
    }
    if (NULL != adc_spi_device_info->filename)
    {
      free(adc_spi_device_info->filename);
      adc_spi_device_info->filename = NULL;
    }
    free(adc_spi_device_info);
    adc_spi_device_info = NULL;
  }
  io_dev->dev_data = NULL;
  return RIO_ERROR;
}
