/*
 * adc_up_local.c
 *
 *  Created on: 16/02/2009
 *      Author: ianb
 */
#include <string.h>
#include <stdint.h>
#include <rinhw/rstdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/debug.h>
#include <rinhw/adc_up.h>

#include <rinhw/adc_up_local.h>


static uint16_t adc_results[NUM_ADC_CHANNELS];


/* Conversion factor from millivolts to ADC counts */
#define ADC_MAX_MV 5000
#define ADC_MAX_COUNT 1024

/* Resistor divider for measuring VIN */
#define ADC_CH_VIN_R_UPPER 33000
#define ADC_CH_VIN_R_LOWER 4700

/* At 25degrees celcius, the LM335 puts out 2.98 volts */
#define ADC_CH_TEMP_MV_REF 2980
#define ADC_CH_TEMP_KELVIN_REF ( 273 + 25 )

/* Kelvin to millivolts conversion factor */
#define ADC_CH_TEMP_MV_PER_KELVIN 10

/* Resistor divider for measuring negative sense (5v-0v when connected) */
#define ADC_CH_SNS_MINUS_R_UPPER 100000
#define ADC_CH_SNS_MINUS_R_LOWER 100000
#define ADC_CH_SNS_MINUS_MV_UPPER_REF 5000

/* Resistor divider for measuring positive sense (7.4v-0v when connected) */
#define ADC_CH_SNS_PLUS_R_UPPER 100000
#define ADC_CH_SNS_PLUS_R_LOWER 100000


/** Convert millivolts at the ADC pin into ADC counts */
static uint16_t adc_mv_to_counts(uint32_t mv)
{
  uint32_t temp;

  temp = mv * ADC_MAX_COUNT;
  temp /= ADC_MAX_MV;
  return (uint16_t) temp;
}


/** Convert the VIN voltage in millivolts into ADC counts */
uint16_t adc_ch_vin_realmv_to_counts(uint32_t mv)
{
  uint32_t temp;

  /* First convert the reading at the 24v point back to a 5v level as per the resistor divider on the R400 (R42, R44) */
  temp = mv * ADC_CH_VIN_R_LOWER;
  temp /= ( ADC_CH_VIN_R_UPPER + ADC_CH_VIN_R_LOWER );

  /* Now convert this into ADC counts */
  return adc_mv_to_counts( temp );
}


/** Convert the temperature into ADC counts */
uint16_t adc_ch_temp_realkelvin_to_counts(uint32_t kelvin)
{
  int32_t temp;
  int32_t mv;

  /* Use gradient and known position to solve for mv */
  temp = ADC_CH_TEMP_MV_PER_KELVIN * ( kelvin - ADC_CH_TEMP_KELVIN_REF );
  mv = temp + ADC_CH_TEMP_MV_REF;

  /* Now convert this into ADC counts */
  return adc_mv_to_counts( mv );
}


/** Convert the negative sense voltage into ADC counts */
uint16_t adc_ch_sns_minus_realmv_to_counts(uint32_t mv)
{
  uint32_t temp;

  /* First convert the reading at the AN_SEN- point back to a 5v level as per the resistor divider on the R400 (R3, R4) */
  temp = ADC_CH_SNS_MINUS_MV_UPPER_REF * ADC_CH_SNS_MINUS_R_LOWER + mv * ADC_CH_SNS_MINUS_R_UPPER;
  temp /= (ADC_CH_SNS_MINUS_R_LOWER + ADC_CH_SNS_MINUS_R_UPPER);

  /* Now convert this into ADC counts */
  return adc_mv_to_counts( temp );
}


/** Convert the positive sense voltage into ADC counts */
uint16_t adc_ch_sns_plus_realmv_to_counts(uint32_t mv)
{
  uint32_t temp;

  /* First convert the reading at the AN_SEN+ point back to a 5v level as per the resistor divider on the R400 (R1, R2) */
  temp = mv * ADC_CH_SNS_PLUS_R_LOWER;
  temp /= ( ADC_CH_SNS_PLUS_R_UPPER + ADC_CH_SNS_PLUS_R_LOWER );

  /* Now convert this into ADC counts */
  return adc_mv_to_counts( temp );
}


static int adc_local_init(struct t_io_dev *io_dev)
{
#if HW_DEBUG(ADC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc_local_init called\n");
#endif

  /*
   * Its a block device
   * Wipe the results
   */
  io_dev->status &= ~O_DEVCHAR;
  memset(adc_results, 0, sizeof(adc_results));

  /* TODO Get readings from XML config file */
  adc_results[ ADC_CH_VIN ] = adc_ch_vin_realmv_to_counts( 12000 );
  adc_results[ ADC_CH_TEMP ] = adc_ch_temp_realkelvin_to_counts( 273 + 30 );
  adc_results[ ADC_CH_SNS_MINUS ] = adc_ch_sns_minus_realmv_to_counts( 0 );
  adc_results[ ADC_CH_SNS_PLUS ] = adc_ch_sns_plus_realmv_to_counts( 7400 );

#if HW_DEBUG(ADC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc_local_init finished\n");
#endif

  return RIO_SUCCESS;
}


/** Destroy ADC hardware */
static int adc_local_destroy(struct t_io_dev *io_dev)
{
  return RIO_SUCCESS;
}


#ifndef min
#define min(x, y) ((x < y) ? x : y)
#endif


static int adc_local_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  unsigned i;
  unsigned char *rd_ptr = (unsigned char *) &adc_results[0];
  unsigned char *wr_ptr = (unsigned char *) buf;

  if ((io_dev->posn + len) > sizeof(adc_results))
    return RIO_ERROR;

  rd_ptr = &rd_ptr[io_dev->posn];
  for (i = 0; i < len; i++)
  {
    *wr_ptr++ = *rd_ptr++;
  }
  io_dev->posn += len;
  return (int) len;
}


static int adc_local_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  return RIO_ERROR;
}


static long adc_local_filelength(struct t_io_dev *io_dev)
{
  return (long) sizeof(adc_results);
}


/** Device specific configuration */
static int adc_local_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  int *int_ptr_param;
  int result = RIO_SUCCESS;

#if HW_DEBUG(ADC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("adc_local_ioctl called, function %d\n", func);
#endif

  /* Function dispatch */
  switch (func)
  {

    /* Get the number of channels on this ADC */
    case IOCTL_ADC_GET_NUM_CHANNELS:
#if HW_DEBUG(ADC_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("adc_local_ioctl(): IOCTL_ADC_GET_NUM_CHANNELS called\n");
#endif
      int_ptr_param = va_arg(*arglist, int *);
      *int_ptr_param = NUM_ADC_CHANNELS;
      break;

    /* Unknown option */
    default:
#if HW_DEBUG(ADC_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("adc_local_ioctl() error: Unknown IOCTL\n");
#endif
      result = RIO_ERROR;
      break;
  }

  return result;
}


/** Function to create and initialise a local auxiliary ADC device */
int rio_create_device_adc_local(struct t_io_dev *io_dev)
{

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) NULL;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = adc_local_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = adc_local_read;
  io_dev->dev_write = adc_local_write;
  io_dev->dev_filelength = adc_local_filelength;
  io_dev->dev_ioctl = adc_local_ioctl;
  io_dev->dev_destroy = adc_local_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);
}
