/*
 * buzzer_generic.c
 *
 *  Created on: May 1, 2012
 *      Author: ian
 */
#include <string.h>
#include <stdlib.h>
#include <rinhw/debug.h>
#include <rinhw/timemath.h>
#include <rinhw/rinthread.h>

#include <rinhw/buzzer_generic.h>


/** Buzzer task */
static void *buz_generic_task(void *data)
{
  struct s_generic_buzzer *gb;

#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_task: started\n");
#endif

  /* Check pointers */
  if (NULL == data)
    goto bgt_fail;
  gb = (struct s_generic_buzzer *) data;

  while (1)
  {
    /* Wait indefinitely for buzzer writes */
    if (sem_wait(&gb->buz_signal_sem) < 0)
    {
#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("buz_generic_task: buzzer semaphore wait error\n");
#endif
      /* Thread finished */
      break;
    }

    /* Quit if signaled */
    if (gb->buz_task_quit)
    {
      break;
    }

    /* Now wait until we have access, indefinitely for buzzer writes */
    if (sem_wait(&gb->buz_busy_sem) < 0)
    {
#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("buz_generic_task: buzzer mutex lock error\n");
#endif
      /* Thread finished */
      break;
    }

    /* Do the buzzing */
    if (NULL != gb->buz_hw_run_seq_fn)
    {
      gb->buz_hw_run_seq_fn(gb);
    }

    /* All done, let in new writes */
    if (sem_post(&gb->buz_busy_sem) < 0)
    {
#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("buz_generic_task: buzzer mutex unlock error\n");
#endif
      /* Thread finished */
      break;
    }

    /* Quit if signaled */
    if (gb->buz_task_quit)
    {
      break;
    }
  }

bgt_fail:
#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_task: quitting\n");
#endif
  rin_thread_exit(NULL);
}


/** Buzzer device initialisation */
static int buz_generic_init(struct t_io_dev *io_dev)
{
#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_init called\n");
#endif
  /* Nothing to do here */

  return RIO_SUCCESS;
}


/** Destroy buzzer device */
static int buz_generic_destroy(struct t_io_dev *io_dev)
{
  struct s_generic_buzzer *gb;

#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_destroy called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
    return RIO_ERROR;
  gb = (struct s_generic_buzzer *) io_dev->dev_data;

#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_destroy: waiting for task to quit\n");
#endif

  /* Notify thread to quit */
  gb->buz_task_quit = true;
  sem_post(&gb->buz_signal_sem);

  /* Wait for thread to quit */
  pthread_join(gb->buz_thread, NULL);

#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_destroy: task has quit, cleaning up\n");
#endif

  /* Pointers to NULL */
  io_dev->dev_data = NULL;

  /* Call hardware destroy function */
  if (NULL != gb->buz_hw_destroy_fn)
  {
    gb->buz_hw_destroy_fn(gb);
  }

  /* Destroy semaphores */
  sem_destroy(&gb->buz_signal_sem);
  sem_destroy(&gb->buz_busy_sem);

  /* Function pointers to NULL */
  gb->buz_hw_init_fn = NULL;
  gb->buz_hw_destroy_fn = NULL;
  gb->buz_hw_run_seq_fn = NULL;

  free(gb);
  gb = NULL;

#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_destroy done\n");
#endif
  return RIO_SUCCESS;
}


/** Buzzer read */
static int buz_generic_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_generic_buzzer *gb = (struct s_generic_buzzer *) io_dev->dev_data;
  const tBuzzer NoBeep = { 0,  0 };

#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_read called\n");
#endif

  /*
   * Make sure right size
   */
  if (len != sizeof(tBuzzer))
  {
#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("buz_generic_read: read size problem\n");
#endif
    return RIO_ERROR;
  }

  /*
   * If current_buz_seq is larger than num_buz_seq then there are no more steps to run
   * so return the NoBeep tBuzzer instead
   */
  if (gb->current_buz_seq > gb->num_buz_seq)
  {
    memcpy(buf, &NoBeep, len);
  }
  /*
   * buz_seq will usually be sitting on the next step to run rather than the step currently running
   * so return the previous (currently running) step if available
   */
  else if (gb->current_buz_seq > 0)
  {
    memcpy(buf, &gb->buz_seq[gb->current_buz_seq - 1], len);
  }
  else
  {
    memcpy(buf, &gb->buz_seq[gb->current_buz_seq], len);
  }

  return (int) len;
}


/** Buzzer write */
static int buz_generic_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_generic_buzzer *gb;
  unsigned short num_buz = len / sizeof(tBuzzer);

#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_write called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
    return RIO_ERROR;
  gb = (struct s_generic_buzzer *) io_dev->dev_data;

  /* Make sure right size */
  if (((len % sizeof(tBuzzer)) != 0) || (num_buz > MAX_BUZ_SEQ) || (num_buz == 0))
  {
#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("write size problem\n");
#endif
    return RIO_ERROR;
  }

  /* Pend on buzzer completing current work */
  if (sem_timedwait_ticks(&gb->buz_busy_sem, timeout) < 0)
  {
#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("buzzer timeout or other error\n");
#endif
    return RIO_ERROR;
  }

  /* Copy the data into the buzzer buffer and kick it off */
  memcpy(gb->buz_seq, buf, len);
  gb->num_buz_seq = (unsigned char) num_buz;

  /* Free up the lock on the data */
  sem_post(&gb->buz_busy_sem);

  /* Signal that there is new data available */
  sem_post(&gb->buz_signal_sem);

  return (int) len;
}


/** Buzzer file length */
static long buz_generic_filelength(struct t_io_dev *io_dev)
{
  struct s_generic_buzzer *gb;

#if HW_DEBUG(BUZ_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("buz_generic_filelength called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
    return RIO_ERROR;
  gb = (struct s_generic_buzzer *) io_dev->dev_data;

  return (long) (sizeof(gb->buz_seq));
}


/** Function to create and initialise a local buzzer device */
int rio_create_device_buzzer_generic(struct t_io_dev *io_dev,
                                     void *buz_hw_data,
                                     t_buz_hw_init_fn buz_hw_init_fn,
                                     t_buz_hw_destroy_fn buz_hw_destroy_fn,
                                     t_buz_hw_run_seq_fn buz_hw_run_seq_fn)
{
  struct s_generic_buzzer *gb = NULL;
  bool bbs_inited = false;
  bool bss_inited = false;
  bool thread_created = false;
  bool hw_inited = false;

  /* Allocate structure */
  gb = malloc(sizeof(struct s_generic_buzzer));
  if (NULL == gb)
  {
    goto rcdbg_fail;
  }

  /* Fill out members */
  memset(gb, 0, sizeof(struct s_generic_buzzer));
  gb->buz_hw_data = buz_hw_data;
  gb->buz_hw_init_fn = buz_hw_init_fn;
  gb->buz_hw_destroy_fn = buz_hw_destroy_fn;
  gb->buz_hw_run_seq_fn = buz_hw_run_seq_fn;
  gb->num_buz_seq = 0;
  gb->current_buz_seq = 0;
  gb->buz_task_quit = false;

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = gb;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = buz_generic_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = buz_generic_read;
  io_dev->dev_write = buz_generic_write;
  io_dev->dev_filelength = buz_generic_filelength;
  io_dev->dev_ioctl = default_ioctl;
  io_dev->dev_destroy = buz_generic_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Its a block device */
  io_dev->status &= ~O_DEVCHAR;

  /* Create semaphores */
  if (sem_init(&gb->buz_busy_sem, 0, 1) < 0)
  {
    goto rcdbg_fail;
  }
  bbs_inited = true;
  if (sem_init(&gb->buz_signal_sem, 0, 0) < 0)
  {
    goto rcdbg_fail;
  }
  bss_inited = true;

  /* Call hardware initialisation function */
  if (NULL != gb->buz_hw_init_fn)
  {
    gb->buz_hw_init_fn(gb);
  }
  hw_inited = true;

  /* Create buzzer task */
  if (!rin_thread_create(&gb->buz_thread, buz_generic_task, gb, true))
  {
    goto rcdbg_fail;
  }
  thread_created = true;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
    goto rcdbg_fail;
  }

  /* Success */
  return RIO_SUCCESS;

rcdbg_fail:
  io_dev->dev_data = NULL;
  if (NULL != gb)
  {
    if (thread_created)
    {
      /* Notify thread to quit */
      gb->buz_task_quit = true;
      sem_post(&gb->buz_signal_sem);

      /* Wait for thread to quit */
      pthread_join(gb->buz_thread, NULL);
    }
    if (hw_inited && NULL != gb->buz_hw_destroy_fn)
    {
      gb->buz_hw_destroy_fn(gb);
    }
    if (bbs_inited)
    {
      sem_destroy(&gb->buz_busy_sem);
    }
    if (bss_inited)
    {
      sem_destroy(&gb->buz_signal_sem);
    }

    free(gb);
    gb = NULL;
  }
  return RIO_ERROR;
}
