#define _GNU_SOURCE
#include <semaphore.h>
#include <errno.h>
#include <pthread.h>
#include <rinhw/timemath.h>

#include <rinhw/byteq.h>


/**
 * Read a byte from a byte queue
 *
 * Parameters:
 *   q is a pointer to an existing initialised byte queue struct
 *   to is the timeout period (use NULL for indefinite wait)
 *   err is a pointer to a byte that will be set to the error value from the operation, if any
 *
 * Returns:
 *   Next character from the buffer on success (*err == bqrSuccess), 0 otherwise
 */
uint8_t ByteQRead(struct s_byteq *q, const struct timespec *to, enum eBQResult *err)
{
  uint8_t c;
  struct timespec ts_end;
  int res;

  /*
   * If pend on read, then wait for a character to be put into the buffer
   * if an error has occurred, then return now with error code
   */
  if (q->PendDirn == bqPendOnRead)
  {
    if (NULL != to)
    {
      clock_gettime(CLOCK_REALTIME, &ts_end);
      AddTimespec(&ts_end, to);
      res = sem_timedwait(&q->Sem, &ts_end);
    }
    else
    {
      res = sem_wait(&q->Sem);
    }
    if (res != 0)
    {
      if (ETIMEDOUT == errno)
        *err = bqrTimeout;
      else
        *err = bqrError;
      return '\0';
    }
  }

  /*
   * Lock this section
   * Since we may not have pended (bqPendOnWrite case), so we need to check if there is
   * data in the buffer
   */
  if (pthread_mutex_lock(&q->lock_mutex) != 0)
  {
    *err = bqrError;
    return '\0';
  }

  if (q->Level > 0)
  {
    /*
     * Yep there's some data, read it out
     * and update the level and read pointers, wrapping at the end of the buffer
     */
    q->Level--;
    c = *q->pRead++;
    if (q->pRead >= &q->pBuffer[q->Size])
      q->pRead = &q->pBuffer[0];

    /* If this is a bqPendOnWrite queue, indicate that a character has been removed from the buffer */
    if (q->PendDirn == bqPendOnWrite)
      sem_post(&q->Sem);

    /* Keep a count of how many bytes have been sent through this queue */
    q->Count++;

    /* Its all worked ok so return no error, and the character we extracted */
    *err = bqrSuccess;
  }
  else
  {
    /* Nope there are no characters in the buffer, return queue empty error */
    *err = bqrEmpty;
    c = '\0';
  }

  /* Unlock and return data */
  pthread_mutex_unlock(&q->lock_mutex);
  return c;
}
