#define _GNU_SOURCE
#include <semaphore.h>
#include <errno.h>
#include <pthread.h>
#include <rinhw/timemath.h>

#include <rinhw/byteq.h>


/**
 * Write a byte to a byte queue
 *
 * Parameters:
 *   q is a pointer to an existing initialised byte queue struct
 *   c is the byte to write to the queue
 *   to is the timeout period (use NULL for indefinite wait)
 *
 * Returns:
 *   Error value from the operation, if any (bqrSuccess on success)
 */
enum eBQResult ByteQWrite(struct s_byteq *q, uint8_t c, const struct timespec *to)
{
  enum eBQResult err;
  struct timespec ts_end;
  int res;

  /*
   * If pend on write, then wait for room in the buffer
   * if an error has occurred, then return now with error code
   */
  if (q->PendDirn == bqPendOnWrite)
  {
    if (NULL != to)
    {
      clock_gettime(CLOCK_REALTIME, &ts_end);
      AddTimespec(&ts_end, to);
      res = sem_timedwait(&q->Sem, &ts_end);
    }
    else
    {
      res = sem_wait(&q->Sem);
    }
    if (res != 0)
    {
      err = (ETIMEDOUT == errno) ? bqrTimeout : bqrError;
      return err;
    }
  }

  /*
   * Lock this section
   * Since we may not have pended (bqPendOnRead case), so we need to check if there is
   * room in the buffer
   */
  if (pthread_mutex_lock(&q->lock_mutex) != 0)
  {
    return bqrError;
  }

  if (q->Level < q->Size)
  {

    /*
     * Yep there's some room left, write in the new data
     * and update the level and write pointers, wrapping at the end of the buffer
     */
    q->Level++;
    *q->pWrite++ = c;
    if (q->pWrite == &q->pBuffer[q->Size])
      q->pWrite = &q->pBuffer[0];

    /* If this is a bqPendOnRead queue, indicate that a character has been added to the buffer */
    if (q->PendDirn == bqPendOnRead)
      sem_post(&q->Sem);

    /* Its all worked ok so return no error */
    err = bqrSuccess;
  }
  else
  {
    /* Nope there is no room in the buffer, return queue empty error */
    err = bqrFull;
  }

  /* Unlock and return result */
  pthread_mutex_unlock(&q->lock_mutex);
  return err;
}
