#ifndef KDEVICE_DOT_H
#define KDEVICE_DOT_H 1

#include <rinhw/os_cpu.h>
#include "device.h"

/*
 * Accurate timing for flash programming -
 *
 * This generates accurate timing suitable for the delays required in flash
 * programming. It is calibrated to the SYS_FREQ, however relies on the compiler
 * to generate tight loops in a certain way. You should check the compiler output
 * matches what is specified below, otherwise delays will be incorrect.
 *
 * Note also that delays MUST NOT be stored in variables, only #defines that
 * are evaluated at compile time, since the maths involves floats.
 */

#define DECLARE_DELAY_FUNCTION() \
__inline static void inline_delay(unsigned short iterations) \
{ \
    while (--iterations > 0) \
        ; \
}

/*
 * All delay counters are implemented as this type
 */
#define DELAY_TYPE unsigned short

/*
 * Number of picoseconds for one CPU state
 */
#define PICOSECONDS_PER_STATE ((float) (1000000000.0 / (float) SYS_FREQ))

/*
 * Number of CPU states to run a tight loop:
 *
 * ?L1:
 *     DEC.W #1, R0  ; 1 state
 *     BNE ?L1:8     ; 2 states
 *
 * NOTE This assumes the compiler generates the loop like this
 * You should check the ASM output of the compiler
 */
#define ASM_STATES_PER_LOOP 3

/*
 * Number of picoseconds it takes to complete one iteration of the loop
 */
#define PICOSECONDS_PER_LOOP ((float) (PICOSECONDS_PER_STATE * ASM_STATES_PER_LOOP))

/*
 * Number of CPU states required to setup for the loop. These
 * are subtracted from loop time to make the loop more exact.
 * Includes:
 * 2 - for the immediate load into R0 (MOV.W #x, R0)
 * 4 - for the bitset address (BSET.B #x, @a:8)
 */
#define ASM_SETUP_STATES (2 + 4)

/*
 * Calculate the number of loop iterations required
 * to generate a delay of us microseconds
 */
#define CALC_DELAY_COUNTS(us) ((float) ((float) (us) * 1000.0 / PICOSECONDS_PER_LOOP) - (float) (ASM_SETUP_STATES / ASM_STATES_PER_LOOP))

/*
 * Make sure iterations > 0, so the loop won't underflow and run for ages
 */
#define SAFE_DELAY_COUNTS(us) ((CALC_DELAY_COUNTS(us) > 0) ? CALC_DELAY_COUNTS(us) : 1)

/*
 * Convert microseconds into iterations
 */
#define MICROSECONDS_TO_ITERATIONS(us) ((DELAY_TYPE) (SAFE_DELAY_COUNTS(us)))

/*
 * Generate a delay in number of iterations
 */
#define DELAY_ITERATIONS(iter) (inline_delay(iter))

/*
 * Generate a delay in microseconds
 */
#define DELAY_MICROSECOND(us) (DELAY_ITERATIONS(MICROSECONDS_TO_ITERATIONS(us)))

/*
 * Do a bit toggle with a delay in number of iterations
 */
#define ITERATION_BIT_TOGGLE(iter, biton, bitoff) { biton; DELAY_ITERATIONS(iter); bitoff; }

/*
 * Do a bit toggle with a delay in microseconds
 */
#define MICROSECOND_BIT_TOGGLE(delay, biton, bitoff) { biton; DELAY_MICROSECOND(delay); bitoff; }

#endif
