#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/debug.h>
#include "KDevice.h"
#include <rinhw/flash2238.h>
#include <rinhw/memorydev_local_file.h>

#include <rinhw/flash2238_local.h>


/*
 * List of flash block start addresses
 */
static const unsigned long erase_block_starts[NUM_FLASH_BLOCKS + 1] =
{
    FLASH_BLOCK_STARTS
};

#if 0
/*
 * List of flash block erase block register settings
 */
static const unsigned short erase_block_ebrs[NUM_FLASH_BLOCKS] =
{
    FLASH_BLOCK_EBRS
};
#endif


/*
 * CalcFlashBlock -
 *
 * From posn, work out which flash erase block we are in
 */
static unsigned char CalcFlashBlock(unsigned long posn, unsigned char *block)
{
    unsigned char i;

    for (i = 0; i < NUM_FLASH_BLOCKS; i++)
    {
        if ((posn >= erase_block_starts[i]) && (posn < erase_block_starts[i + 1]))
        {
            *block = i;
            return 1;
        }
    }
    return 0;
}


/*
 * EraseFlashBlock -
 *
 * Erases a specified flash block by calling the EraseFunc, which it loads into RAM
 */
static unsigned char EraseFlashBlock(struct t_io_dev *io_dev)
{
  struct sMemoryDeviceFileInfo *dev_info = (struct sMemoryDeviceFileInfo *) io_dev->dev_data;
  unsigned long flash_posn = FLASH_START_ADDRESS + io_dev->posn;
  unsigned char result, block;

#if HW_DEBUG(FLASH_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("EraseFlashBlock called\n");
#endif

  /*
   * Work out which flash block we are in (based on IO device position), and
   * quit if invalid
   */
  if (CalcFlashBlock(flash_posn, &block) == 0)
  {
#if HW_DEBUG(FLASH_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("Erase block requested (%d) > NUM_FLASH_BLOCKS (%d)\n", block, NUM_FLASH_BLOCKS);
#endif
    return 0;
  }

  /*
   * Erase the requested flash block
   */
#if HW_DEBUG(FLASH_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("Erasing flash block now, start = 0x%lx, end = 0x%ld\n",
               erase_block_starts[block], erase_block_starts[block + 1]);
#endif

  /* Fill the region with the blankchar */
  rlseek( io_dev, erase_block_starts[ block ], SEEK_SET );
  while ( rtell( io_dev ) < (long) erase_block_starts[ block + 1 ] )
  {
    rwrite( io_dev, &dev_info->blankfillchar, sizeof( dev_info->blankfillchar ), 0 );
  }
  /* TODO flash2238 -> set result from rwrite return value */
  result = 1;

  /* Return result */
#if HW_DEBUG(FLASH_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("Erasing flash block complete, result %d\n", result);
#endif
  return result;
}


/*
 * flash2238_local_init -
 *
 * Driver init entry point, setup the hardware and device
 */
static int flash2238_local_init(struct t_io_dev *io_dev)
{
  /* struct sMemoryDeviceFileInfo *dev_info = (struct sMemoryDeviceFileInfo *) io_dev->dev_data; */
  int result;

#if HW_DEBUG(FLASH_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("flash2238_local_init called\n");
#endif

  /* Call the memory device initialiser */
  result = memorydev_local_file_init( io_dev );

#if HW_DEBUG(FLASH_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("flash2238_local_init finished\n");
#endif
  return result;
}


/** Destroy flash device */
static int flash2238_local_destroy(struct t_io_dev *io_dev)
{
  struct sMemoryDeviceFileInfo *dev_info = (struct sMemoryDeviceFileInfo *) io_dev->dev_data;

#if HW_DEBUG(FLASH_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("flash2238_local_destroy called\n");
#endif

  /* Close file */
  memorydev_local_file_close(io_dev);

  /* Free memory */
  io_dev->dev_data = NULL;
  if (NULL != dev_info)
  {
    if (NULL != dev_info->filename)
    {
      free(dev_info->filename);
      dev_info->filename = NULL;
    }
    free(dev_info);
    dev_info = NULL;
  }

#if HW_DEBUG(FLASH_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("flash2238_local_destroy finished\n");
#endif
  return RIO_SUCCESS;
}


/*
 * flash2238_ioctl -
 *
 * Device specific configuration
 */
static int flash2238_local_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
    int result = RIO_SUCCESS;

    /*
     * NOTE: Since character parameters are pushed as words,
     * we'll deal with them as words.
     *
     * Function dispatch
     */
    switch (func)
    {

        /*
         * Set the address of the slave we want to talk to
         */
        case IOCTL_FLASH_ERASE_BLOCK:
            if (EraseFlashBlock(io_dev) == 0)
                result = RIO_ERROR;
            break;

        /*
         * Unknown option
         */
        default:
            result = RIO_ERROR;
            break;
    }
    return result;
}


/** Function to create and initialise a local flash2238 device */
int rio_create_device_flash2238_local(struct t_io_dev *io_dev, const char *filename, size_t length, uint8_t blankfillchar)
{
  struct sMemoryDeviceFileInfo *dev_flash2238_info = NULL;

  /* Allocate memory device information struct */
  dev_flash2238_info = malloc(sizeof(struct sMemoryDeviceFileInfo));
  if (NULL == dev_flash2238_info)
  {
    goto rcdfl_fail;
  }

  /* Fill out struct */
  dev_flash2238_info->fd = -1;
  dev_flash2238_info->maxsize = length;
  dev_flash2238_info->blankfillchar = blankfillchar;

  dev_flash2238_info->filename = strdup(filename);
  if (NULL == dev_flash2238_info->filename)
  {
    goto rcdfl_fail;
  }

  /*
   * Setup device data pointer to point to the memory device struct
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) dev_flash2238_info;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = flash2238_local_init;
  io_dev->dev_open = memorydev_local_file_open;
  io_dev->dev_close = memorydev_local_file_close;
  io_dev->dev_read = memorydev_local_file_read;
  io_dev->dev_write = memorydev_local_file_write;
  io_dev->dev_filelength = memorydev_local_file_filelength;
  io_dev->dev_ioctl = flash2238_local_ioctl;
  io_dev->dev_destroy = flash2238_local_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
    goto rcdfl_fail;
  }

  /* Success */
  return RIO_SUCCESS;

rcdfl_fail:
  io_dev->dev_data = NULL;
  if (NULL != dev_flash2238_info)
  {
    if (dev_flash2238_info->fd >= 0)
    {
      close(dev_flash2238_info->fd);
      dev_flash2238_info->fd = -1;
    }
    if (NULL != dev_flash2238_info->filename)
    {
      free(dev_flash2238_info->filename);
      dev_flash2238_info->filename = NULL;
    }
    free(dev_flash2238_info);
    dev_flash2238_info = NULL;
  }
  return RIO_ERROR;
}
