/*
 * framdev_local.c
 *
 *  Created on: 03/02/2009
 *      Author: ianb
 */
#include <fcntl.h>
#include <stddef.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <rinhw/rio.h>
#include <rinhw/debug.h>

#include <rinhw/memorydev_local_file.h>


/** Memory device driver init entry point, setup the hardware and device */
int memorydev_local_file_init( struct t_io_dev *io_dev )
{
#if HW_DEBUG(MEMORYDEV_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("memorydev_local_file_init called\n");
#endif

  /* Its a block device */
  io_dev->status &= ~O_DEVCHAR;

#if HW_DEBUG(MEMORYDEV_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("memorydev_local_file_init finished\n");
#endif
  return RIO_SUCCESS;
}


/** Memory device open device call, opens a local file */
int memorydev_local_file_open( struct t_io_dev *io_dev, unsigned short timeout )
{
  struct sMemoryDeviceFileInfo *dev_info = (struct sMemoryDeviceFileInfo *) io_dev->dev_data;
  int handle;

#if HW_DEBUG(MEMORYDEV_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("memorydev_local_file_open called\n");
#endif

  handle = open( dev_info->filename, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR );
  if ( 0 > handle )
  {
#if HW_DEBUG(MEMORYDEV_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("memorydev_local_file_open: failed opening %s\n", dev_info->filename);
#endif
    return RIO_ERROR;
  }
  if (fcntl(handle, F_SETFD, FD_CLOEXEC) < 0)
  {
#if HW_DEBUG(MEMORYDEV_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("memorydev_local_file_open: error setting FD_CLOEXEC on %s (errno %d)\n", dev_info->filename, errno);
#endif
    close(handle);
    return RIO_ERROR;
  }

  dev_info->fd = handle;

#if HW_DEBUG(MEMORYDEV_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("memorydev_local_file_open finished\n");
#endif
  return RIO_SUCCESS;
}


/** Memory device close device call, closes a local file */
int memorydev_local_file_close( struct t_io_dev *io_dev )
{
  struct sMemoryDeviceFileInfo *dev_info = (struct sMemoryDeviceFileInfo *) io_dev->dev_data;
  const int handle = dev_info->fd;
  int result;

#if HW_DEBUG(MEMORYDEV_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("memorydev_local_file_close called\n");
#endif

  dev_info->fd = -1;
  if (handle >= 0)
  {
    result = (close(handle) < 0) ? RIO_ERROR : RIO_SUCCESS;
  }
  else
  {
    result = RIO_ERROR;
  }
  return result;
}


/** Get the length of a file */
static long get_filelength( int fd )
{
  struct stat filestat;
  long result = RIO_ERROR;

  if ( 0 <= fstat( fd, &filestat ) )
  {
    result = filestat.st_size;
  }
  return result;
}


/** Expand or shrink the file to the size of the memory device if it isn't already */
static int expand_to_memory_device_size( int fd, size_t expected_size, uint8_t fillchar )
{
  const long filelen = get_filelength( fd );
  int result = RIO_SUCCESS;
  #define EXPAND_BUF_MAX_LEN 1024
  char *buffer;
  size_t i;
  size_t to_write;

  if ( filelen < 0 )
  {
    return RIO_ERROR;
  }

  if ( filelen < (long) expected_size )
  {
    /*
     * File is smaller than expected size, so expand the file, filling with fillchar.
     * We do this with a large buffer, rather than individual byte writes to make it
     * faster
     */

    /* Allocate buffer, and set bytes */
    buffer = malloc(EXPAND_BUF_MAX_LEN);
    if (NULL == buffer)
    {
      return RIO_ERROR;
    }
    memset(buffer, fillchar, EXPAND_BUF_MAX_LEN);

    /* Seek end of file */
    if ( lseek( fd, 0, SEEK_END ) != filelen )
    {
      result = RIO_ERROR;
      goto buffree;
    }

    /* Expand to new size, with block writes */
    for ( i = filelen; i < expected_size; i += to_write )
    {
      to_write = (expected_size - i > EXPAND_BUF_MAX_LEN) ? EXPAND_BUF_MAX_LEN : (expected_size - i);
      if ( (ssize_t) to_write != write( fd, buffer, to_write ) )
      {
        result = RIO_ERROR;
        goto buffree;
      }
    }

    /* Success */
    result = RIO_SUCCESS;

buffree:
    free(buffer);
  }
  else if ( filelen > (long) expected_size )
  {
    /* File is larger than expected size, so truncate it */
    result = ( ftruncate( fd, expected_size ) < 0 ) ? RIO_ERROR : RIO_SUCCESS;
  }
  return result;
}


/** Memory device read device call, reads data from a local file */
int memorydev_local_file_read( struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout )
{
  struct sMemoryDeviceFileInfo *dev_info = (struct sMemoryDeviceFileInfo *) io_dev->dev_data;
  size_t end = io_dev->posn + len;
  int result;

  /* Does the read fit in fram */
  if ( dev_info->maxsize < end )
  {
    return RIO_ERROR;
  }

  /* Expand the memory device file to expected size */
  if ( expand_to_memory_device_size( dev_info->fd, dev_info->maxsize, dev_info->blankfillchar ) != RIO_SUCCESS  )
  {
    return RIO_ERROR;
  }

  /* Seek to the position we want to read */
  result = lseek( dev_info->fd, io_dev->posn, SEEK_SET );
  if ( result < 0 )
  {
    return RIO_ERROR;
  }

  /* Do the read */
  result = read( dev_info->fd, buf, len );

  /* Update io device position */
  if ( result == len )
  {
    io_dev->posn += len;
  }

  /* Return the result */
  return ( result != len ) ? RIO_ERROR : len;
}


/** Memory device write device call, writes data to a local file */
int memorydev_local_file_write( struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout )
{
  struct sMemoryDeviceFileInfo *dev_info = (struct sMemoryDeviceFileInfo *) io_dev->dev_data;
  size_t end = io_dev->posn + len;
  int result;

  /* Does the write fit in fram */
  if ( dev_info->maxsize < end )
  {
    return RIO_ERROR;
  }

  /* Expand the memory device file to expected size */
  if ( expand_to_memory_device_size( dev_info->fd, dev_info->maxsize, dev_info->blankfillchar ) != RIO_SUCCESS  )
  {
    return RIO_ERROR;
  }

  /* Seek to the position we want to write */
  result = lseek( dev_info->fd, io_dev->posn, SEEK_SET );
  if ( result < 0 )
  {
    return RIO_ERROR;
  }

  /* Do the write */
  result = write( dev_info->fd, buf, len );

  /* Update io device position */
  if ( result == len )
  {
    io_dev->posn += len;
  }

  /* Return the result */
  return ( result != len ) ? RIO_ERROR : len;
}


/** Memory device filelength device call, returns size of memory device */
long memorydev_local_file_filelength( struct t_io_dev *io_dev )
{
  struct sMemoryDeviceFileInfo *dev_info = (struct sMemoryDeviceFileInfo *) io_dev->dev_data;

  return dev_info->maxsize;
}
