/**
 * @file iic_ftdi.c
 *
 * @brief I2C driver using FTDI chips in MPSSE mode
 */
#include <stdbool.h>
#include <stdint.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <rinhw/rio.h>
#include <rinhw/iic.h>
#include <rinhw/debug.h>
#include <ftdi.h>
#include <mpsse.h>

#include <rinhw/iic_ftdi.h>


/** Default tristate register: SK/DO/CS and GPIOs are outputs, DI is an input */
#define FTDI_DEFAULT_TRISTATE (FTDI_PIN_SK | FTDI_PIN_DO | FTDI_PIN_CS | FTDI_PIN_GPIO0 | FTDI_PIN_GPIO1 | FTDI_PIN_GPIO2 | FTDI_PIN_GPIO3)

/** Default port state: SK and CS are high, all others low */
#define FTDI_DEFAULT_PORT     (FTDI_PIN_SK | FTDI_PIN_CS)



#if 0
#include <stdio.h>
#include <stdlib.h>
#include <mpsse.h>

#define SIZE    0x8000          // Size of EEPROM chip (32KB)
#define WCMD    "\xA0\x00\x00"  // Write start address command
#define RCMD    "\xA1"          // Read command
#define FOUT    "eeprom.bin"    // Output file

int main(void)
{
        FILE *fp = NULL;
        char *data = NULL;
        int retval = EXIT_FAILURE;
        struct mpsse_context *eeprom = NULL;

        if((eeprom = MPSSE(I2C, FOUR_HUNDRED_KHZ, MSB)) != NULL && eeprom->open)
        {
                printf("%s initialized at %dHz (I2C)\n", GetDescription(eeprom), GetClock(eeprom));

                /* Write the EEPROM start address */
                Start(eeprom);
                Write(eeprom, WCMD, sizeof(WCMD) - 1);

                if(GetAck(eeprom) == ACK)
                {
                        /* Send the EEPROM read command */
                        Start(eeprom);
                        Write(eeprom, RCMD, sizeof(RCMD) - 1);

                        if(GetAck(eeprom) == ACK)
                        {
                                /* Read in SIZE bytes from the EEPROM chip */
                                data = Read(eeprom, SIZE);
                                if(data)
                                {
                                        fp = fopen(FOUT, "wb");
                                        if(fp)
                                        {
                                                fwrite(data, 1, SIZE, fp);
                                                fclose(fp);

                                                printf("Dumped %d bytes to %s\n", SIZE, FOUT);
                                                retval = EXIT_SUCCESS;
                                        }

                                        free(data);
                                }

                                /* Tell libmpsse to send NACKs after reading data */
                                SendNacks(eeprom);

                                /* Read in one dummy byte, with a NACK */
                                Read(eeprom, 1);
                        }
                }

                Stop(eeprom);
        }
        else
        {
                printf("Failed to initialize MPSSE: %s\n", ErrorString(eeprom));
        }

        Close(eeprom);

        return retval;
}

#endif



/** i2c FTDI device struct */
struct s_i2c_ftdi_port
{
  struct mpsse_context *mpsse_iic;
  uint8_t slave_address;                      /*!< Currently selected i2c slave address */
  uint8_t slot_mask;                          /*!< Selected slot mask */

#if 0
  char *device_name;                                /*!< Name of linux i2c device (eg /dev/i2c-0) */
  int iic_fd;                                       /*!< File descriptor for device */
  unsigned char slave_address;                      /*!< Currently selected i2c slave address */
  unsigned char slot_mask;                          /*!< Selected slot mask */
  t_iic_set_slot_select_fn iic_set_slot_select_fn;  /*!< Function to select an i2c slot */
  void *iic_set_slot_select_fn_data;                /*!< Data to pass to the slot select function */
#endif
};


/** Set the I2C slave address that we are addressing */
static int iic_ftdi_set_slave_address(struct s_i2c_ftdi_port *i2c_ftdi, unsigned char address)
{
  i2c_ftdi->slave_address = address;
  return RIO_SUCCESS;
}


/*
 * set_slot_enable_mask -
 *
 * Enable each module slot as defined by mask -
 * on the R400, this drives the IRQ line low
 */
static int iic_ftdi_set_slot_enable_mask(struct s_i2c_ftdi_port *i2c_ftdi, unsigned char mask)
{
#if 0
  int slot;
  int found;
#endif
  int ret = RIO_SUCCESS;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_set_slot_enable_mask called\n");
#endif

#if 0

  /* Figure out which slot to enable from the mask */
  for (found = 0, slot = 0; slot < 5; slot++)
  {
    if ((1 << slot) == mask)
    {
      found = 1;
      break;
    }
  }

  if (!found)
  {
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("set_slot_enable_mask: invalid slot\n");
#endif
    return ret;
  }

  /* Configure slot */
  if (NULL != iic_port->iic_set_slot_select_fn)
  {
    ret = iic_port->iic_set_slot_select_fn(iic_port->iic_set_slot_select_fn_data, slot);
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
    if (ret < 0)
    {
      DEBUG_PRINTF("set_slot_enable_mask: i2c slot select function fail\n");
    }
#endif
  }
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  else
  {
    DEBUG_PRINTF("set_slot_enable_mask: no i2c slot select function defined\n");
  }
#endif
#endif

  i2c_ftdi->slot_mask = mask;
  return ret;
}


/*
 * default_slot_enable -
 *
 * Set the slot enable back to defaults for the R400 -
 * as all 4 IRQ lines set to inputs
 */
static int iic_ftdi_default_slot_enable(struct s_i2c_ftdi_port *i2c_ftdi)
{
  int ret = RIO_SUCCESS;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_default_slot_enable called\n");
#endif

#if 0 /* TODO */
  /* De-select all slots */
  if (NULL != iic_port->iic_set_slot_select_fn)
  {
    ret = iic_port->iic_set_slot_select_fn(iic_port->iic_set_slot_select_fn_data, IIC_SLOT_DESELECT);
  }
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  if (ret < 0)
  {
    DEBUG_PRINTF("iic_ftdi_default_slot_enable: we21hw_set_slot_select() fail\n");
  }
#endif
#endif

  i2c_ftdi->slot_mask = 0;
  return ret;
}


/*
 * get_slot_enable_mask -
 *
 * Get the current value of the slot enable mask
 */
static unsigned char iic_ftdi_get_slot_enable_mask(const struct s_i2c_ftdi_port *i2c_ftdi)
{
  return i2c_ftdi->slot_mask;
}


/*
 * iic_setspeed -
 *
 * Set the bus speed for the IIC port (0..15) as defined in the header file
 * by IIC_SPEED_CLK_DIV_xxx
 */
static int iic_ftdi_setspeed(const struct s_i2c_ftdi_port *i2c_ftdi, unsigned char speed)
{
    /* TODO implement bus speed */
  return RIO_SUCCESS;
}


/** Driver init entry point, setup the hardware and device */
static int iic_ftdi_init(struct t_io_dev *io_dev)
{
  /* struct s_i2c_ftdi_port *i2c_ftdi = io_dev->dev_data; */

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_init called\n");
#endif

  /* Its a block device */
  io_dev->status &= ~O_DEVCHAR;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_init finished\n");
#endif
  return RIO_SUCCESS;
}


/** Destroy iic device */
static int iic_ftdi_destroy(struct t_io_dev *io_dev)
{
  struct s_i2c_ftdi_port *i2c_ftdi = io_dev->dev_data;
  int result;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* Free up stuff we've allocated */
  if (NULL != i2c_ftdi)
  {
    if (NULL != i2c_ftdi->mpsse_iic)
    {
      /* Shutdown mpsse */
      Close(i2c_ftdi->mpsse_iic);
      i2c_ftdi->mpsse_iic = NULL;
    }
    free(i2c_ftdi);
    io_dev->dev_data = NULL;
  }

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_destroy: result %d\n", result);
#endif
  return result;
}


/** FTDI I2C device write */
static int iic_ftdi_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_i2c_ftdi_port *i2c_ftdi;
  int result;
  uint8_t address;
  size_t i;
  const uint8_t *byte_buf;
  int mpsse_res;
  char data;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_write called\n");
#endif

  if (NULL == io_dev || NULL == io_dev->dev_data || NULL == buf)
	return RIO_ERROR;
  i2c_ftdi = io_dev->dev_data;

  /* Send start condition and address */
  address = i2c_ftdi->slave_address & 0xFE;
  mpsse_res = Start(i2c_ftdi->mpsse_iic);
  if (MPSSE_OK != mpsse_res)
  {
    ERROR_PRINTF("iic_ftdi_write: mpsse Start() failed with result %d, libftdi error %s\n", mpsse_res, ErrorString(i2c_ftdi->mpsse_iic));
    result = RIO_ERROR;
    goto ifw_done;
  }
  mpsse_res = Write(i2c_ftdi->mpsse_iic, (char *) &address, sizeof(address));
  if (MPSSE_OK != mpsse_res)
  {
    ERROR_PRINTF("iic_ftdi_write: mpsse Write(address) failed with result %d, libftdi error %s\n", mpsse_res, ErrorString(i2c_ftdi->mpsse_iic));
    result = RIO_ERROR;
    goto ifw_done;
  }
  if (GetAck(i2c_ftdi->mpsse_iic) != ACK)
  {
	ERROR_PRINTF("iic_ftdi_write: mpsse GetAck() returned NACK\n");
	result = RIO_ERROR;
	goto ifw_done;
  }

  /* Write data */
  for (i = 0, byte_buf = buf, result = len; i < len; i++)
  {
    data = byte_buf[i];
    mpsse_res = Write(i2c_ftdi->mpsse_iic, &data, sizeof(uint8_t));
    if (MPSSE_OK != mpsse_res)
    {
      ERROR_PRINTF("iic_ftdi_write: mpsse Write(data) failed with result %d, libftdi error %s\n", mpsse_res, ErrorString(i2c_ftdi->mpsse_iic));
      result = RIO_ERROR;
      goto ifw_done;
    }

	if (GetAck(i2c_ftdi->mpsse_iic) != ACK)
	{
      ERROR_PRINTF("iic_ftdi_write: mpsse GetAck() returned NACK\n");
      result = RIO_ERROR;
      goto ifw_done;
    }
  }

ifw_done:
  /* Send stop condition */
  mpsse_res = Stop(i2c_ftdi->mpsse_iic);
  if (MPSSE_OK != mpsse_res)
  {
    ERROR_PRINTF("iic_ftdi_write: mpsse Stop() failed with result %d, libftdi error %s\n", mpsse_res, ErrorString(i2c_ftdi->mpsse_iic));
    result = RIO_ERROR;
  }

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_write finished, result %d\n", result);
#endif
  return result;
}


/** FTDI I2C device read */
static int iic_ftdi_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_i2c_ftdi_port *i2c_ftdi;
  int result;
  uint8_t address;
  size_t i;
  uint8_t *byte_buf;
  uint8_t *read_buf;
  int mpsse_res;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_read called\n");
#endif

  if (NULL == io_dev || NULL == io_dev->dev_data || NULL == buf)
	return RIO_ERROR;
  i2c_ftdi = io_dev->dev_data;

  /* Send start condition and address */
  address = (i2c_ftdi->slave_address & 0xFE) | 0x01;
  mpsse_res = Start(i2c_ftdi->mpsse_iic);
  if (MPSSE_OK != mpsse_res)
  {
    ERROR_PRINTF("iic_ftdi_read: mpsse Start() failed with result %d, libftdi error %s\n", mpsse_res, ErrorString(i2c_ftdi->mpsse_iic));
    result = RIO_ERROR;
    goto ifr_done;
  }
  mpsse_res = Write(i2c_ftdi->mpsse_iic, (char *) &address, sizeof(address));
  if (MPSSE_OK != mpsse_res)
  {
    ERROR_PRINTF("iic_ftdi_read: mpsse Write(address) failed with result %d, libftdi error %s\n", mpsse_res, ErrorString(i2c_ftdi->mpsse_iic));
    result = RIO_ERROR;
    goto ifr_done;
  }
  if (GetAck(i2c_ftdi->mpsse_iic) != ACK)
  {
    ERROR_PRINTF("iic_ftdi_read: mpsse GetAck() returned NACK\n");
    result = RIO_ERROR;
    goto ifr_done;
  }

  /* Read data */
  for (i = 0, byte_buf = buf, result = len; i < len; i++)
  {
    read_buf = (uint8_t *) Read(i2c_ftdi->mpsse_iic, sizeof(uint8_t));
    if (NULL == read_buf)
    {
      ERROR_PRINTF("iic_ftdi_read: mpsse Read(data) returned NULL, libftdi error %s\n", ErrorString(i2c_ftdi->mpsse_iic));
      result = RIO_ERROR;
      break;
    }
    else
    {
      byte_buf[i] = read_buf[0];
      free(read_buf);
    }

    if (GetAck(i2c_ftdi->mpsse_iic) != ACK)
    {
      ERROR_PRINTF("iic_ftdi_read: mpsse GetAck() returned NACK\n");
      result = RIO_ERROR;
      goto ifr_done;
    }
  }

ifr_done:
  /* Send stop condition */
  mpsse_res = Stop(i2c_ftdi->mpsse_iic);
  if (MPSSE_OK != mpsse_res)
  {
    ERROR_PRINTF("iic_ftdi_read: mpsse Stop() failed with result %d, libftdi error %s\n", mpsse_res, ErrorString(i2c_ftdi->mpsse_iic));
    result = RIO_ERROR;
  }

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_read finished, result %d\n", result);
#endif
  return result;
}


/** Don't know about this, so return 0 */
static long iic_ftdi_filelength(struct t_io_dev *io_dev)
{
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_filelength called\n");
#endif

  return 0L;
}


/** Device specific configuration */
static int iic_ftdi_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_i2c_ftdi_port *i2c_ftdi = io_dev->dev_data;
  unsigned char **byte_ptr_param1;
  int result = RIO_SUCCESS;
  int temp;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_ftdi_ioctl called, function %d\n", func);
#endif

  /*
   * NOTE: Since character parameters are pushed as words,
   * we'll deal with them as words.
   *
   * Function dispatch
   */
  switch (func)
  {

    /*
     * Set the address of the slave we want to talk to
     */
    case IOCTL_IIC_SET_SLAVE_ADDR:
      temp = va_arg(*arglist, int);
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_SET_SLAVE_ADDR: new slave address 0x%02x\n", temp);
#endif
      result = iic_ftdi_set_slave_address(i2c_ftdi, (unsigned char) temp);
      break;

    /*
     * Get the address of the slave we are talking to
     */
    case IOCTL_IIC_GET_SLAVE_ADDR:
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_GET_SLAVE_ADDR called\n");
#endif
      byte_ptr_param1 = va_arg(*arglist, unsigned char **);
      **byte_ptr_param1 = i2c_ftdi->slave_address;
      break;

    /*
     * Set the slot enable mask for the port
     */
    case IOCTL_IIC_SET_SLOT_MASK:
      temp = (unsigned char) va_arg(*arglist, int);
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_SET_SLOT_MASK called, mask 0x%02x\n", temp);
#endif
      result = iic_ftdi_set_slot_enable_mask(i2c_ftdi, temp);
      break;

    /*
     * Get the slot enable mask for the port
     */
    case IOCTL_IIC_GET_SLOT_MASK:
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_GET_SLOT_MASK called\n");
#endif
      byte_ptr_param1 = va_arg(*arglist, unsigned char **);
      **byte_ptr_param1 = iic_ftdi_get_slot_enable_mask(i2c_ftdi);
      break;

    /*
     * Set the slot enable mask to defaults
     */
    case IOCTL_IIC_DEFAULT_SLOT_MASK:
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_DEFAULT_SLOT_MASK called\n");
#endif
      result = iic_ftdi_default_slot_enable(i2c_ftdi);
      break;

    /*
     * Set the IIC bus speed
     */
    case IOCTL_IIC_SET_SPEED:
      temp = (unsigned char) va_arg(*arglist, int);
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_SET_SPEED called, speed %d\n", temp);
#endif
      result = iic_ftdi_setspeed(i2c_ftdi, temp);
      break;

    /*
     * Unknown option
     */
    default:
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("Unknown IOCTL for IIC\n");
#endif
      result = RIO_ERROR;
      break;
  }

  return result;
}


/**
 * Function to create and initialise a FTDI I2C device.
 *
 * @param io_dev is the IO device to be initialised
 */
int rio_create_device_iic_ftdi(struct t_io_dev *io_dev)
{
  struct s_i2c_ftdi_port *i2c_ftdi = NULL;

  /* Allocate i2c info here, iic_local_destroy() frees it */
  i2c_ftdi = malloc(sizeof(struct s_i2c_ftdi_port));
  if (NULL == i2c_ftdi)
  {
    ERROR_PRINTF("rio_create_device_iic_ftdi: allocate iic_port failed\n");
    goto rcdif_fail;
  }
  memset(i2c_ftdi, 0, sizeof(struct s_i2c_ftdi_port));

  /* Setup the iic port */
  i2c_ftdi->slave_address = 0;
  i2c_ftdi->slot_mask = 0;

  //i2c_ftdi->mpsse_iic = MPSSE(I2C, FOUR_HUNDRED_KHZ, MSB);
  i2c_ftdi->mpsse_iic = MPSSE(I2C, ONE_HUNDRED_KHZ, MSB);
  if (NULL == i2c_ftdi->mpsse_iic)
  {
    ERROR_PRINTF("rio_create_device_iic_ftdi: open MPSSE failed\n");
    goto rcdif_fail;
  }

  /* Set the slot enable mask to defaults */
  if (iic_ftdi_default_slot_enable(i2c_ftdi) < 0)
  {
    ERROR_PRINTF("rio_create_device_iic_ftdi: failed defaulting slot enable\n");
    goto rcdif_fail;
  }

  /* Initialise at the "safe" IIC speed of 100khz */
  if (iic_ftdi_setspeed(i2c_ftdi, IIC_SPEED_DEFAULT) < 0)
  {
    ERROR_PRINTF("rio_create_device_iic_ftdi: failed setting bus speed\n");
    goto rcdif_fail;
  }

  /* Set the slave address */
  if (iic_ftdi_set_slave_address(i2c_ftdi, 0) < 0)
  {
    ERROR_PRINTF("rio_create_device_iic_ftdi: failed setting slave address\n");
    goto rcdif_fail;
  }

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) i2c_ftdi;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = iic_ftdi_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = iic_ftdi_read;
  io_dev->dev_write = iic_ftdi_write;
  io_dev->dev_filelength = iic_ftdi_filelength;
  io_dev->dev_ioctl = iic_ftdi_ioctl;
  io_dev->dev_destroy = iic_ftdi_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
    goto rcdif_fail;
  }

  /* Success */
  return RIO_SUCCESS;

  /* Clean up on failure */
rcdif_fail:
  if (NULL != i2c_ftdi)
  {
    if (NULL != i2c_ftdi->mpsse_iic)
    {
      Close(i2c_ftdi->mpsse_iic);
      i2c_ftdi->mpsse_iic = NULL;
    }
    free(i2c_ftdi);
    i2c_ftdi = NULL;
  }
  return RIO_ERROR;
}
