/**
 * @file iic_local_linux.c
 *
 * @brief Driver using Linux kernel drivers (i2c-parport and so on)
 */
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/i2c-dev.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <rinhw/rstdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/iic.h>
#include <rinhw/debug.h>
#include <rinhw/we21hw.h>

#include <rinhw/iic_local_linux.h>

/*
 * Getting the linux I2C drivers to work:
 *
 * I used the i2c-parport-light module. i2c-parport didn't seem to work.
 * i2c-parport-light bashes the parallel port hardware, and doesn't play nice
 * with the parport module.
 *
 * Here are the steps you should go through:
 * 1 - Set parallel port setup in BIOS to ECP (not sure if this is needed), but ECP worked for me
 *
 * 2 - Prevent the following modules from being loaded: lp, ppdev, parport_pc, parport
 *     This is done by creating a file called /etc/modprobe.d/blacklist-custom.conf
 *     which contains the following lines:
 *       blacklist lp
 *       blacklist ppdev
 *       blacklist parport_pc
 *       blacklist parport
 *     NOTE: I found that rmmod is not good enough because (I assume) the drivers have already played
 *           around with the parallel port hardware.
 *     NOTE: If you find that the lp module is still being loaded, rename the module under
 *           /lib/modules/2.6.28-15-generic/kernel/drivers/char/lp.ko to something else, so the kernel
 *           can't find it.
 *     NOTE: Turning off the CUPS daemon might help. Under System->Administration->Services in Ubuntu GUI,
 *           called "Printer service (cups)". Obviously then you can't print...
 *
 * 2 - Reboot, and lsmod to check that the modules aren't loaded
 *
 * 3 - sudo modprobe i2c-parport-light type=3
 *
 * 4 - sudo modprobe i2c-dev
 *
 * 5 - ls /dev | grep i2c should list i2c-0, if not look in dmesg
 *
 * 6 - sudo chmod a+rw /dev/i2c-0 so your process can access it
 *
 * 7 - i2cdetect 0 to see what's out there
 *
 * 8 - Setup IIC settings in driverconfig.xml to point to /dev/i2c-0
 *
 */


/** i2c device struct */
struct s_i2c_linux_local_port
{
  char *device_name;                                /*!< Name of linux i2c device (eg /dev/i2c-0) */
  int iic_fd;                                       /*!< File descriptor for device */
  unsigned char slave_address;                      /*!< Currently selected i2c slave address */
  unsigned char slot_mask;                          /*!< Selected slot mask */
  t_iic_set_slot_select_fn iic_set_slot_select_fn;  /*!< Function to select an i2c slot */
  void *iic_set_slot_select_fn_data;                /*!< Data to pass to the slot select function */
};


/** Set the I2C slave address that we are addressing */
static int set_iic_slave_address(struct s_i2c_linux_local_port *iic_port, unsigned char address)
{
  int temp;

  iic_port->slave_address = address;

  /* Compensate for the shift left done in the Linux driver when adding in the R/W bit */
  temp = address;
  temp >>= 1;
  return ( ioctl( iic_port->iic_fd, I2C_SLAVE, temp ) < 0 ) ? RIO_ERROR : RIO_SUCCESS;
}


/*
 * set_slot_enable_mask -
 *
 * Enable each module slot as defined by mask -
 * on the R400, this drives the IRQ line low
 */
static int set_slot_enable_mask(struct s_i2c_linux_local_port *iic_port, unsigned char mask)
{
  int slot;
  int found;
  int ret = RIO_ERROR;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("set_slot_enable_mask called\n");
#endif

  /* Figure out which slot to enable from the mask */
  for (found = 0, slot = 0; slot < 5; slot++)
  {
    if ((1 << slot) == mask)
    {
      found = 1;
      break;
    }
  }

  if (!found)
  {
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("set_slot_enable_mask: invalid slot\n");
#endif
    return ret;
  }

  /* Configure slot */
  if (NULL != iic_port->iic_set_slot_select_fn)
  {
    ret = iic_port->iic_set_slot_select_fn(iic_port->iic_set_slot_select_fn_data, slot);
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
    if (ret < 0)
    {
      DEBUG_PRINTF("set_slot_enable_mask: i2c slot select function fail\n");
    }
#endif
  }
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  else
  {
    DEBUG_PRINTF("set_slot_enable_mask: no i2c slot select function defined\n");
  }
#endif

  iic_port->slot_mask = mask;

  return ret;
}


/*
 * default_slot_enable -
 *
 * Set the slot enable back to defaults for the R400 -
 * as all 4 IRQ lines set to inputs
 */
static int default_slot_enable(struct s_i2c_linux_local_port *iic_port)
{
  int ret = RIO_ERROR;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("default_slot_enable called\n");
#endif

  /* De-select all slots */
  if (NULL != iic_port->iic_set_slot_select_fn)
  {
    ret = iic_port->iic_set_slot_select_fn(iic_port->iic_set_slot_select_fn_data, IIC_SLOT_DESELECT);
  }
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  if (ret < 0)
  {
    DEBUG_PRINTF("default_slot_enable: we21hw_set_slot_select() fail\n");
  }
#endif
  iic_port->slot_mask = 0;
  return ret;
}


/*
 * get_slot_enable_mask -
 *
 * Get the current value of the slot enable mask
 */
static unsigned char get_slot_enable_mask(const struct s_i2c_linux_local_port *iic_port)
{
  return iic_port->slot_mask;
}


/*
 * iic_setspeed -
 *
 * Set the bus speed for the IIC port (0..15) as defined in the header file
 * by IIC_SPEED_CLK_DIV_xxx
 */
static int iic_setspeed(struct s_i2c_linux_local_port *iic_port, unsigned char speed)
{
    /* TODO implement bus speed */
  return RIO_SUCCESS;
}


/*
 * iic_local_init -
 *
 * Driver init entry point, setup the hardware and device
 */
static int iic_local_init(struct t_io_dev *io_dev)
{
  struct s_i2c_linux_local_port *iic_port = (struct s_i2c_linux_local_port *) io_dev->dev_data;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_init called\n");
#endif

  /* Open device */
  iic_port->iic_fd = open( iic_port->device_name, O_RDWR, 0 );
  if (iic_port->iic_fd < 0)
  {
    ERROR_PRINTF("iic_local_init: failed opening %s\n", iic_port->device_name);
    goto ili_fail;
  }
  if (fcntl(iic_port->iic_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    ERROR_PRINTF("iic_local_init: error setting FD_CLOEXEC on iic device %s (errno %d)\n", iic_port->device_name, errno);
    goto ili_fail;
  }

  /*
   * First we need to release the SDA line
   * if it is being held low by a DS1337 RTC chip
   */
/*  ReleaseSDALine(iic_port); */

  /* Set 7 bit addressing */
  if (ioctl(iic_port->iic_fd, I2C_TENBIT, 0) < 0)
  {
    ERROR_PRINTF("iic_local_init: failed setting 7 bit\n");
    goto ili_fail;
  }

  /* Set the slot enable mask to defaults */
  if (default_slot_enable(iic_port) < 0)
  {
    ERROR_PRINTF("iic_local_init: failed defaulting slot enable\n");
    goto ili_fail;
  }

  /* Initialise at the "safe" IIC speed of 100khz */
  if (iic_setspeed(iic_port, IIC_SPEED_DEFAULT) < 0)
  {
    ERROR_PRINTF("iic_local_init: failed setting bus speed\n");
    goto ili_fail;
  }

  /* Set the slave address */
  if (set_iic_slave_address(iic_port, 0) < 0)
  {
    ERROR_PRINTF("iic_local_init: failed setting slave address\n");
    goto ili_fail;
  }

  /* Its a block device */
  io_dev->status &= ~O_DEVCHAR;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_init finished\n");
#endif
  return RIO_SUCCESS;

ili_fail:
  if (iic_port->iic_fd >= 0)
  {
    close(iic_port->iic_fd);
    iic_port->iic_fd = -1;
  }

  return RIO_ERROR;
}


/** Destroy iic device */
static int iic_local_destroy(struct t_io_dev *io_dev)
{
  struct s_i2c_linux_local_port *iic_port = (struct s_i2c_linux_local_port *) io_dev->dev_data;
  int result;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* Free up stuff we've allocated */
  if (NULL != iic_port)
  {
    if (iic_port->iic_fd >= 0)
    {
      close(iic_port->iic_fd);
      iic_port->iic_fd = -1;
    }
    if (NULL != iic_port->device_name)
    {
      free(iic_port->device_name);
      iic_port->device_name = NULL;
    }
    free(iic_port);
    io_dev->dev_data = NULL;
  }

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_destroy: result %d\n", result);
#endif
  return result;
}


static int iic_local_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_i2c_linux_local_port *iic_port = (struct s_i2c_linux_local_port *) io_dev->dev_data;
  int result;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_write called\n");
#endif

  /*
   * TODO timeout - This can't really be done on linux easily,
   * as there isn't a way to set the timeout for an I2C device
   * from the userland driver
   */

  /* Write data to i2c bus */
  result = write( iic_port->iic_fd, buf, len );
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
  if ( result < 0 )
  {
    DEBUG_PRINTF("iic_local_write: write failed, result %d, errno %d, error %s\n",
                 result, errno, strerror( errno ) );
  }
#endif

  if ( result != len )
  {
    result = RIO_ERROR;
  }

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_write finished, result %d\n", result);
#endif
  return result;
}


static int iic_local_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_i2c_linux_local_port *iic_port = (struct s_i2c_linux_local_port *) io_dev->dev_data;
  int result;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_read called\n");
#endif

  /*
   * TODO timeout - This can't really be done on linux easily,
   * as there isn't a way to set the timeout for an I2C device
   * from the userland driver
   */

  /* Read data from i2c bus */
  result = read( iic_port->iic_fd, buf, len );
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
  if ( result < 0 )
  {
    DEBUG_PRINTF("iic_local_read: read failed, result %d, errno %d, error %s\n",
                 result, errno, strerror( errno ) );
  }
#endif

  if ( result != len )
  {
    result = RIO_ERROR;
  }

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_read finished, result %d\n", result);
#endif
  return result;
}


/*
 * iic_local_filelength -
 *
 * Don't know about this, so return 0
 */
static long iic_local_filelength(struct t_io_dev *io_dev)
{
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("iic_local_filelength called\n");
#endif

    return 0L;
}


/*
 * iic_local_ioctl -
 *
 * Device specific configuration
 */
static int iic_local_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_i2c_linux_local_port *iic_port = (struct s_i2c_linux_local_port *) io_dev->dev_data;
  unsigned char **byte_ptr_param1;
  int result = RIO_SUCCESS;
  int temp;

#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("iic_local_ioctl called, function %d\n", func);
#endif

  /*
   * NOTE: Since character parameters are pushed as words,
   * we'll deal with them as words.
   *
   * Function dispatch
   */
  switch (func)
  {

    /*
     * Set the address of the slave we want to talk to
     */
    case IOCTL_IIC_SET_SLAVE_ADDR:
      temp = va_arg(*arglist, int);
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_SET_SLAVE_ADDR: new slave address 0x%02x\n", temp);
#endif
      result = set_iic_slave_address(iic_port, (unsigned char) temp);
      break;

    /*
     * Get the address of the slave we are talking to
     */
    case IOCTL_IIC_GET_SLAVE_ADDR:
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_GET_SLAVE_ADDR called\n");
#endif
      byte_ptr_param1 = va_arg(*arglist, unsigned char **);
      **byte_ptr_param1 = iic_port->slave_address;
      break;

    /*
     * Set the slot enable mask for the port
     */
    case IOCTL_IIC_SET_SLOT_MASK:
      temp = (unsigned char) va_arg(*arglist, int);
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_SET_SLOT_MASK called, mask 0x%02x\n", temp);
#endif
      result = set_slot_enable_mask(iic_port, temp);
      break;

    /*
     * Get the slot enable mask for the port
     */
    case IOCTL_IIC_GET_SLOT_MASK:
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_GET_SLOT_MASK called\n");
#endif
      byte_ptr_param1 = va_arg(*arglist, unsigned char **);
      **byte_ptr_param1 = get_slot_enable_mask(iic_port);
      break;

    /*
     * Set the slot enable mask to defaults
     */
    case IOCTL_IIC_DEFAULT_SLOT_MASK:
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_DEFAULT_SLOT_MASK called\n");
#endif
      result = default_slot_enable(iic_port);
      break;

    /*
     * Set the IIC bus speed
     */
    case IOCTL_IIC_SET_SPEED:
      temp = (unsigned char) va_arg(*arglist, int);
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_IIC_SET_SPEED called, speed %d\n", temp);
#endif
      result = iic_setspeed(iic_port, temp);
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = iic_port->iic_fd;
      result = RIO_SUCCESS;
      break;

    /*
     * Unknown option
     */
    default:
#if HW_DEBUG(IIC_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("Unknown IOCTL for IIC\n");
#endif
      result = RIO_ERROR;
      break;
  }

  return result;
}


/**
 * Function to create and initialise a local iic device. Driver takes a local copy of device_name.
 *
 * @param io_dev is the IO device to be initialised
 * @param is the linux i2c device to map the IO device to (eg /dev/i2c-0)
 * @param iic_set_slot_select_fn is a function pointer to a function that sets the slot mask (like we21hw_set_slot_select())
 * @param iic_set_slot_select_fn_data is a pointer to data used by iic_set_slot_select_fn(). Untouched/unused by the i2c driver itself
 */
int rio_create_device_iic_local(struct t_io_dev *io_dev,
                                const char *device_name,
                                t_iic_set_slot_select_fn iic_set_slot_select_fn,
                                void *iic_set_slot_select_fn_data)
{
  struct s_i2c_linux_local_port *iic_port = NULL;

  /* Allocate i2c info here, iic_local_destroy() frees it */
  iic_port = malloc(sizeof(struct s_i2c_linux_local_port));
  if (NULL == iic_port)
  {
    ERROR_PRINTF("rio_create_device_iic_local: allocate iic_port failed\n");
    goto rcdil_fail;
  }

  /* Setup the iic port */
  iic_port->iic_fd = -1;
  iic_port->iic_set_slot_select_fn = iic_set_slot_select_fn;
  iic_port->iic_set_slot_select_fn_data = iic_set_slot_select_fn_data;
  iic_port->slave_address = 0;
  iic_port->slot_mask = 0;

  /* Create our own internal copy of the device name */
  iic_port->device_name = strdup(device_name);
  if (NULL == iic_port->device_name)
  {
    ERROR_PRINTF("rio_create_device_iic_local: allocate device_name failed\n");
    goto rcdil_fail;
  }

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) iic_port;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = iic_local_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = iic_local_read;
  io_dev->dev_write = iic_local_write;
  io_dev->dev_filelength = iic_local_filelength;
  io_dev->dev_ioctl = iic_local_ioctl;
  io_dev->dev_destroy = iic_local_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
    goto rcdil_fail;
  }

  /* Success */
  return RIO_SUCCESS;

  /* Clean up on failure */
rcdil_fail:
  if (NULL != iic_port)
  {
    if (NULL != iic_port->device_name)
    {
      free(iic_port->device_name);
      iic_port->device_name = NULL;
    }
    free(iic_port);
    iic_port = NULL;
  }
  return RIO_ERROR;
}
