/*
 * kbd_linux_event.c
 *
 * Driver for linux event devices
 *
 *  Created on: Dec 7, 2011
 *      Author: ian
 */

#include <stdarg.h>
#include <string.h>
#include <linux/input.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <rinhw/kbd.h>
#include "_kbd_hw.h"
#include <rinhw/debug.h>
#include <rinhw/timemath.h>
#include "kbd_buzzer.h"

#include <rinhw/kbd_linux_event.h>


/*
 * This driver translates events from upto 10 linux event devices (USB HID)
 * into Rinstrum hardware library keys (KEY_PHYSICAL_x)
 *
 * This is an interim measure, to be replaced by a wrapper around linux events, which will
 * allow many more keys, so we can do numeric and alphanumeric editors
 */


#define KEYS_PER_KB_DB 32


/** Can handle up to 10 event devices */
#define KBD_LINUX_EVENT_MAX_DEVICES 10


/** Represents a event device (/dev/eventX) */
struct s_kbd_linux_event_device
{
  char *device_name;
  int device_fd;
  int ctrl_left_pressed;
  int ctrl_right_pressed;
};


/** Key information */
struct s_keybit_information
{
  unsigned long keybits;              /*!< Bit mask of currently pressed keys */
  unsigned long disabled_mask;        /*!< Bit mask of disabled keys (if pressed these keys will generate keDisabled) */
  unsigned long hold_mask;            /*!< Bit mask of held keys (if long pressed these keys will generate keHeld) */
  unsigned long repeat_mask;          /*!< Bit mask of repeat keys (if long pressed these keys will generate multiple keDown events) */
};


/** Linux event keyboard handler data */
struct s_kbd_linux_event
{
  pthread_mutex_t mutex;                  /*!< Exclusive access mutex */
  struct s_kbd_linux_event_device event_devs[KBD_LINUX_EVENT_MAX_DEVICES];
  struct t_io_dev *buzzer_dev;
  int key_buzz_enabled;
  struct s_keybit_information kb_info;
};


/** Keyboard device initialisation */
static int kbd_linux_event_init(struct t_io_dev *io_dev)
{
/*  struct s_kbd_linux_event *kbd_le = io_dev->dev_data; */

#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_init called\n");
#endif
  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
    return RIO_ERROR;

  /* Nothing much to do here cause the create call did it all */

  /* All done */
#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_init finished\n");
#endif
  return RIO_SUCCESS;
}


/** Destroy keyboard device */
static int kbd_linux_event_destroy(struct t_io_dev *io_dev)
{
  struct s_kbd_linux_event *kbd_le;
  int i;

#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_destroy called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
    return RIO_ERROR;
  kbd_le = io_dev->dev_data;

  /* Lock */
  pthread_mutex_lock(&kbd_le->mutex);

  /* Clean up attached Linux devices */
  for (i = 0; i < KBD_LINUX_EVENT_MAX_DEVICES; i++)
  {
    if (NULL != kbd_le->event_devs[i].device_name)
    {
      free(kbd_le->event_devs[i].device_name);
      kbd_le->event_devs[i].device_name = NULL;
      if (kbd_le->event_devs[i].device_fd > 0)
      {
        close(kbd_le->event_devs[i].device_fd);
        kbd_le->event_devs[i].device_fd = -1;
      }
    }
  }

  kbd_le->buzzer_dev = NULL;

  /* Unlock */
  pthread_mutex_unlock(&kbd_le->mutex);

  /* Clean up */
  io_dev->dev_data = NULL;
  pthread_mutex_destroy(&kbd_le->mutex);
  free(kbd_le);
  kbd_le = NULL;

#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_destroy finished\n");
#endif

  return RIO_SUCCESS;
}


/** Read characters from the keyboard device */
static int kbd_linux_event_internal_wait_timeout(struct s_kbd_linux_event *kbd_le, unsigned short timeout, fd_set *readfds)
{
  int i;
  int maxfd;
  struct timeval tv;
  struct timeval *ptv;
  int retval;

#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_wait_timeout called\n");
#endif

  /* Lock */
  pthread_mutex_lock(&kbd_le->mutex);

  /* Setup file descriptors to pend on */
  FD_ZERO(readfds);
  for (i = 0, maxfd = -1; i < KBD_LINUX_EVENT_MAX_DEVICES; i++)
  {
    if (NULL != kbd_le->event_devs[i].device_name &&
        kbd_le->event_devs[i].device_fd >= 0)
    {
      FD_SET(kbd_le->event_devs[i].device_fd, readfds);
      maxfd = (kbd_le->event_devs[i].device_fd > maxfd) ? kbd_le->event_devs[i].device_fd : maxfd;
    }
  }

  /* Unlock */
  pthread_mutex_unlock(&kbd_le->mutex);

  /* Work out time out, 0 means indefinite */
  if (0 == timeout)
  {
    ptv = (struct timeval *) NULL;
  }
  else
  {
    OSTicksToTimeval(&tv, timeout);
    ptv = &tv;
  }

  /* Wait for data on any of these fds */
  retval = select(maxfd + 1, readfds, NULL, NULL, ptv);
  if (retval <= 0)
  {
    /* Timed out (=0) or more serious error (<0) */
#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
    if (retval < 0)
      DEBUG_PRINTF("kbd_linux_event_wait_timeout: select() error\n");
#endif
    return RIO_ERROR;
  }

  return RIO_SUCCESS;
}


#define LINUX_EVENT_KEY_UP 0
#define LINUX_EVENT_KEY_DOWN 1
#define LINUX_EVENT_KEY_REPEAT 2


static int kbd_linux_event_internal_read_translate(struct s_kbd_linux_event *kbd_le, struct s_kbd_linux_event_device *kled, union un_key *key)
{
  struct input_event ev;
  ssize_t rd;
  unsigned char idx;
  unsigned long mask;

  /* Get key from event device */
  rd = read(kled->device_fd, &ev, sizeof(struct input_event));
  if (rd != sizeof(struct input_event))
  {
    return RIO_ERROR;
  }

  if (EV_KEY == ev.type)
  {
    /* Translate code to rin */
    switch (ev.code)
    {
      case KEY_F1:
        key->BIT.KEY = KEY_PHYSICAL_FUNC0;
        break;
      case KEY_F2:
        key->BIT.KEY = KEY_PHYSICAL_FUNC1;
        break;
      case KEY_F3:
        key->BIT.KEY = KEY_PHYSICAL_FUNC2;
        break;
      case KEY_F4:
        key->BIT.KEY = KEY_PHYSICAL_FUNC3;
        break;
      case KEY_F5:
        key->BIT.KEY = KEY_PHYSICAL_FUNC4;
        break;
      case KEY_F6:
        key->BIT.KEY = KEY_PHYSICAL_FUNC5;
        break;
      case KEY_F12:
        key->BIT.KEY = KEY_PHYSICAL_SETUP;
        break;

      case KEY_LEFTCTRL:
        if (LINUX_EVENT_KEY_DOWN == ev.value)
          kled->ctrl_left_pressed = 1;
        else if (LINUX_EVENT_KEY_UP == ev.value)
          kled->ctrl_left_pressed = 0;
        return RIO_ERROR;

      case KEY_RIGHTCTRL:
        if (LINUX_EVENT_KEY_DOWN == ev.value)
          kled->ctrl_right_pressed = 1;
        else if (LINUX_EVENT_KEY_UP == ev.value)
          kled->ctrl_right_pressed = 0;
        return RIO_ERROR;

      default:
        return RIO_ERROR;
    }

    /* Check masks */
    idx = KeyToIndex(key->BIT.KEY);
    if (KEY_LOOKUP_FAIL == idx || idx >= KEYS_PER_KB_DB)
    {
      return RIO_ERROR;
    }
    mask = 1UL << idx;
    if (mask & kbd_le->kb_info.disabled_mask)
    {
      if (LINUX_EVENT_KEY_DOWN == ev.value)
      {
        key->BIT.EVENT = keDisabled;
      }
      else
      {
        return RIO_ERROR;
      }
    }
    else if (mask & kbd_le->kb_info.repeat_mask)
    {
      /* Repeat keys */
      if (LINUX_EVENT_KEY_DOWN == ev.value || LINUX_EVENT_KEY_REPEAT == ev.value)
      {
        key->BIT.EVENT = keDown;
      }
      else if (LINUX_EVENT_KEY_UP == ev.value)
      {
        key->BIT.EVENT = keUp;
      }
      else
      {
        return RIO_ERROR;
      }
    }
    else if (mask & kbd_le->kb_info.hold_mask)
    {
      /* Held keys */
      if (kled->ctrl_right_pressed || kled->ctrl_left_pressed)
      {
        if (LINUX_EVENT_KEY_DOWN == ev.value)
        {
          key->BIT.EVENT = keHeld;
        }
        else if (LINUX_EVENT_KEY_UP == ev.value)
        {
          key->BIT.EVENT = keUp;
        }
        else
        {
          return RIO_ERROR;
        }
      }
      else
      {
        if (LINUX_EVENT_KEY_DOWN == ev.value)
        {
          key->BIT.EVENT = keDown;
        }
        else if (LINUX_EVENT_KEY_UP == ev.value)
        {
          key->BIT.EVENT = keUp;
        }
        else
        {
          return RIO_ERROR;
        }
      }
    }
    else
    {
      /* Normal (basic) keys */
      if (LINUX_EVENT_KEY_DOWN == ev.value)
      {
        key->BIT.EVENT = keDown;
      }
      else if (LINUX_EVENT_KEY_UP == ev.value)
      {
        key->BIT.EVENT = keUp;
      }
      else
      {
        return RIO_ERROR;
      }
    }

    /* Fill out keybits mask */
    if (keUp == key->BIT.EVENT)
    {
      kbd_le->kb_info.keybits &= ~mask;
    }
    else
    {
      kbd_le->kb_info.keybits |= mask;
    }
  }
  else
  {
    return RIO_ERROR;
  }
  kbd_buzzer(kbd_le->buzzer_dev, kbd_le->key_buzz_enabled, key->BIT.EVENT);
  return RIO_SUCCESS;
}


/** Read characters from the keyboard device */
static int kbd_linux_event_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_kbd_linux_event *kbd_le;
  fd_set readfds;
  unsigned short count;
  union un_key *keybuf;
  bool found;
  int i;

#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_read called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data || NULL == buf)
    return RIO_ERROR;
  kbd_le = io_dev->dev_data;
  keybuf = (union un_key *) buf;

  /* Read keys */
  for (count = 0; count < len; )
  {
    /* Wait for timeout or data */
    if (kbd_linux_event_internal_wait_timeout(kbd_le, timeout, &readfds) < 0)
    {
      return RIO_ERROR;
    }

    /* Lock */
    pthread_mutex_lock(&kbd_le->mutex);

    /* Find the first fd set */
    for (i = 0, found = false; i < KBD_LINUX_EVENT_MAX_DEVICES; i++)
    {
      if (NULL != kbd_le->event_devs[i].device_name &&
          FD_ISSET(kbd_le->event_devs[i].device_fd, &readfds))
      {
        /* Read keypress data */
        found = true;
        if (kbd_linux_event_internal_read_translate(kbd_le, &kbd_le->event_devs[i], &keybuf[count]) >= 0)
        {
          count++;
        }
        break;
      }
    }

    /* Unlock */
    pthread_mutex_unlock(&kbd_le->mutex);

    if (!found)
    {
      return RIO_ERROR;
    }
  }

  /* Done, success */
  return (int) len;
}


/** Write characters to the keyboard device, not implemented */
static int kbd_linux_event_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_write called\n");
#endif
  /* Not implemented */
  return RIO_ERROR;
}


/*
 * kbd_local_filelength -
 *
 * Get the number of characters in the keyboard queue
 */
static long kbd_linux_event_filelength(struct t_io_dev *io_dev)
{
#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_filelength called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
    return RIO_ERROR;

/*  return (long) ByteQGetLevel(&kbd_q); TODO implement this */
  return 0;
}


/** Setup a number of keys */
static unsigned char kbd_linux_event_configure_keys(struct s_kbd_linux_event *kbd_le, int func, unsigned char numkeys, unsigned char *keys)
{
  unsigned long mask;
  unsigned char i, idx;

  /* For the number of keys to configure... */
  for (i = 0; i < numkeys; i++, keys++)
  {
#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("kbd_linux_event_configure_keys: setting key %d\n", *keys);
#endif

    /* Convert the key to a bitmask */
    if ((idx = KeyToIndex(*keys)) == KEY_LOOKUP_FAIL)
      return 0;

    /* In this implementation, there are no external keys */
    if (idx < KEYS_PER_KB_DB)
    {
      mask = 1UL << idx;

      /* Turn all bitmasks off to start with */
      kbd_le->kb_info.disabled_mask &= ~mask;
      kbd_le->kb_info.repeat_mask &= ~mask;
      kbd_le->kb_info.hold_mask &= ~mask;

      /* Configure as requested */
      switch (func)
      {
        case IOCTL_KBD_SET_BASIC:
          break;
        case IOCTL_KBD_SET_REPEAT:
          kbd_le->kb_info.repeat_mask |= mask;
          break;
        case IOCTL_KBD_SET_HOLD:
          kbd_le->kb_info.hold_mask |= mask;
          break;
        case IOCTL_KBD_SET_DISABLED:
          kbd_le->kb_info.disabled_mask |= mask;
          break;
        default:
          return 0;
      }
    }
  }
  return 1;
}


static int kbd_linux_event_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_kbd_linux_event *kbd_le;
  unsigned char num_keys;
  unsigned char *keys;
  int result = RIO_ERROR;
  unsigned char *pressed;

#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_linux_event_ioctl called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
    return RIO_ERROR;
  kbd_le = io_dev->dev_data;

  /* Function dispatch */
  switch (func)
  {
    /* Configure some keys */
    case IOCTL_KBD_SET_BASIC:
    case IOCTL_KBD_SET_REPEAT:
    case IOCTL_KBD_SET_HOLD:
    case IOCTL_KBD_SET_DISABLED:
      num_keys = (unsigned char) va_arg(*arglist, int);
      keys = va_arg(*arglist, unsigned char *);
#if HW_DEBUG(KBD_EVENT_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("kbd_linux_event_ioctl: config keys, func=%d, num_keys=%d, keys_ptr=%p\n", func, num_keys, keys);
#endif
      if (kbd_linux_event_configure_keys(kbd_le, func, num_keys, keys))
        result = RIO_SUCCESS;
      break;

    /* Set the key bits for the external 32keys */
    case IOCTL_KBD_WRITE_HI_KEY_BITS:
      /* external_keybits = va_arg(*arglist, unsigned long); NOT supported */
      result = RIO_SUCCESS;
      break;

    /* Returns true or false depending on whether the key is currently pressed */
    case IOCTL_KBD_IS_KEY_PRESSED:
      pressed = va_arg(*arglist, unsigned char *);
      *pressed = ((kbd_le->kb_info.keybits & (0x00000001 << KeyToIndex(*pressed))) != 0);
      result = RIO_SUCCESS;
      break;

    /* Returns internal key status as a maskable unsigned long */
    case IOCTL_KBD_INTERNAL_BITS:
      *(va_arg(*arglist, unsigned long *)) = kbd_le->kb_info.keybits;
      result = RIO_SUCCESS;
      break;

    /* Enable or disable the buzzer when keys are pressed */
    case IOCTL_KBD_ENABLE_BUZZER:
      kbd_le->key_buzz_enabled = va_arg(*arglist, int);
      result = RIO_SUCCESS;
      break;
  }

  /* Done, return result */
  return result;
}


/** Function to create and initialise a linux event keyboard device */
int rio_create_device_kbd_linux_event(struct t_io_dev *io_dev, struct t_io_dev *buzzer_dev)
{
  struct s_kbd_linux_event *kbd_le;
  int i;
  int result;

  if (NULL == io_dev)
  {
    result = RIO_ERROR_EVKBD_INVALID_PARAM;
    goto rcdkle_fail;
  }

  /* Make a structure */
  kbd_le = malloc(sizeof(struct s_kbd_linux_event));
  if (NULL == kbd_le)
  {
    result = RIO_ERROR_EVKBD_OUT_OF_MEMORY;
    goto rcdkle_fail;
  }

  /* Make the mutex */
  if (0 != pthread_mutex_init(&kbd_le->mutex, NULL))
  {
    result = RIO_ERROR_EVKBD_MISC;
    goto rcdkle_free;
  }

  /* Init struct */
  memset(kbd_le, 0, sizeof(*kbd_le));
  for (i = 0; i < KBD_LINUX_EVENT_MAX_DEVICES; i++)
  {
    memset(&kbd_le->event_devs[i], 0, sizeof(kbd_le->event_devs[i]));
    kbd_le->event_devs[i].device_name = NULL;
    kbd_le->event_devs[i].device_fd = -1;
    kbd_le->event_devs[i].ctrl_left_pressed = 0;
    kbd_le->event_devs[i].ctrl_right_pressed = 0;
  }
  kbd_le->buzzer_dev = buzzer_dev;
  kbd_le->key_buzz_enabled = 1;
  kbd_le->kb_info.keybits = 0;
  kbd_le->kb_info.disabled_mask = 0;
  kbd_le->kb_info.hold_mask = 0;
  kbd_le->kb_info.repeat_mask = 0;

  /*
   * Setup device data pointer
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) kbd_le;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = kbd_linux_event_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = kbd_linux_event_read;
  io_dev->dev_write = kbd_linux_event_write;
  io_dev->dev_filelength = kbd_linux_event_filelength;
  io_dev->dev_ioctl = kbd_linux_event_ioctl;
  io_dev->dev_destroy = kbd_linux_event_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

  /* Register and initialise the device */
  return rregister_io(io_dev);

rcdkle_free:
  free(kbd_le);

rcdkle_fail:
  return result;
}


#define BITS_PER_LONG (sizeof(long) * 8)
#define NBITS(x) ((((x)-1)/BITS_PER_LONG)+1)
#define OFF(x)  ((x)%BITS_PER_LONG)
#define BIT(x)  (1UL<<OFF(x))
#define LONG(x) ((x)/BITS_PER_LONG)
#define test_bit(bit, array)  ((array[LONG(bit)] >> OFF(bit)) & 1)


/** Add a new event device (eg /dev/event0) to monitor to a rinstrum linux event keyboard driver (created by rio_create_device_kbd_linux_event())  */
int rio_kbd_linux_event_add_event_device(struct t_io_dev *io_dev, const char *new_device_name)
{
  struct s_kbd_linux_event *kbd_le;
  int i;
  int idx;
  struct s_kbd_linux_event_device *ed;
  int version;
  unsigned long evtypebits[NBITS(EV_MAX)];
  unsigned long evkeybits[NBITS(KEY_MAX)];
  int result;

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data || NULL == new_device_name)
    return RIO_ERROR_EVKBD_INVALID_PARAM;
  kbd_le = io_dev->dev_data;

  /* Lock */
  pthread_mutex_lock(&kbd_le->mutex);

  /* Are we already monitoring it */
  for (i = 0; i < KBD_LINUX_EVENT_MAX_DEVICES; i++)
  {
    if (NULL != kbd_le->event_devs[i].device_name &&
        0 == strcmp(new_device_name, kbd_le->event_devs[i].device_name))
    {
      /* Here already */
      result = RIO_ERROR_EVKBD_ALREADY_HANDLED;
      goto kleaed_fail_unlock;
    }
  }

  /* Find a spare one */
  for (i = 0, idx = -1; i < KBD_LINUX_EVENT_MAX_DEVICES; i++)
  {
    if (NULL == kbd_le->event_devs[i].device_name)
    {
      /* Found one */
      idx = i;
      break;
    }
  }

  if (idx < 0)
  {
    /* No room for a new one */
    result = RIO_ERROR_EVKBD_FULL;
    goto kleaed_fail_unlock;
  }

  /* Have room for one */
  ed = &kbd_le->event_devs[idx];
  memset(ed, 0, sizeof(*ed));
  ed->ctrl_left_pressed = 0;
  ed->ctrl_right_pressed = 0;
  ed->device_name = strdup(new_device_name);
  if (NULL == ed->device_name)
  {
    result = RIO_ERROR_EVKBD_OUT_OF_MEMORY;
    goto kleaed_fail_unlock;
  }

  /* Open event device */
  ed->device_fd = open(ed->device_name, O_RDONLY);
  if (ed->device_fd < 0)
  {
    result = RIO_ERROR_EVKBD_IO;
    goto kleaed_fail_devname;
  }
  if (fcntl(ed->device_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    result = RIO_ERROR_EVKBD_IO;
    goto kleaed_fail_devopen;
  }

  /* Quick check it is an event device */
  if (ioctl(ed->device_fd, EVIOCGVERSION, &version))
  {
    result = RIO_ERROR_EVKBD_NOT_EVENT;
    goto kleaed_fail_devopen;
  }

  /* Are EV_KEY (key events) supported */
  memset(evtypebits, 0, sizeof(evtypebits));
  if (ioctl(ed->device_fd, EVIOCGBIT(0, EV_MAX), evtypebits) < 0)
  {
    result = RIO_ERROR_EVKBD_IO;
    goto kleaed_fail_devopen;
  }
  if (!test_bit(EV_KEY, evtypebits))
  {
    /* No, they're not, there's no point looking at this device */
    result = RIO_ERROR_EVKBD_KEYS_UNSUPPORTED;
    goto kleaed_fail_devopen;
  }

  /* Are the keys we're interested in supported? */
  memset(evkeybits, 0, sizeof(evkeybits));
  if (ioctl(ed->device_fd, EVIOCGBIT(EV_KEY, KEY_MAX), evkeybits) < 0)
  {
    result = RIO_ERROR_EVKBD_IO;
    goto kleaed_fail_devopen;
  }
  if (!test_bit(KEY_F1, evkeybits) ||
      !test_bit(KEY_F2, evkeybits) ||
      !test_bit(KEY_F3, evkeybits) ||
      !test_bit(KEY_F4, evkeybits) ||
      !test_bit(KEY_F5, evkeybits) ||
      !test_bit(KEY_F6, evkeybits) ||
      !test_bit(KEY_F12, evkeybits))
  {
    /* No, there's no point looking at this device */
    result = RIO_ERROR_EVKBD_KEYS_UNSUPPORTED;
    goto kleaed_fail_devopen;
  }

  /* Unlock */
  pthread_mutex_unlock(&kbd_le->mutex);

  /* We're good to go */
  return RIO_SUCCESS;

kleaed_fail_devopen:
  close(ed->device_fd);
  ed->device_fd = -1;

kleaed_fail_devname:
  free(ed->device_name);
  ed->device_name = NULL;

kleaed_fail_unlock:
  /* Unlock */
  pthread_mutex_unlock(&kbd_le->mutex);

  return result;
}


/**
 * Remove an event device (eg /dev/event0) from a rinstrum linux event keyboard driver (created by rio_create_device_kbd_linux_event())
 *
 * @param io_dev is the linux keyboard event device
 * @param device_name is the name of the event device to remove (eg /dev/event0) from the driver, or NULL to remove all event devices
 */
int rio_kbd_linux_event_remove_event_device(struct t_io_dev *io_dev, const char *device_name)
{
  struct s_kbd_linux_event *kbd_le;
  int i;
  int idx;

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
    return RIO_ERROR_EVKBD_INVALID_PARAM;
  kbd_le = io_dev->dev_data;

  /* Lock */
  pthread_mutex_lock(&kbd_le->mutex);

  /* Find it in the list */
  for (i = 0, idx = -1; i < KBD_LINUX_EVENT_MAX_DEVICES; i++)
  {
    if (NULL != kbd_le->event_devs[i].device_name &&
        (NULL == device_name || 0 == strcmp(device_name, kbd_le->event_devs[i].device_name)))
    {
      /* Found it, remove it */
      free(kbd_le->event_devs[i].device_name);
      kbd_le->event_devs[i].device_name = NULL;
      close(kbd_le->event_devs[i].device_fd);
      kbd_le->event_devs[i].device_fd = -1;
      idx = i;
      break;
    }
  }

  /* Unlock list */
  pthread_mutex_unlock(&kbd_le->mutex);

  return (idx < 0) ? RIO_ERROR_EVKBD_NOT_HANDLED : RIO_SUCCESS;
}
