#include <stdarg.h>
#include <string.h>
#include <semaphore.h>
#include <stdlib.h>
#include <stdio.h>
#include <rinhw/rstdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/byteq.h>
#include <rinhw/rio.h>
#include <rinhw/gpio.h>
#include <rinhw/kbd.h>
#include <rinhw/bkgrnd.h>
#include "_kbd_hw.h"
#include <rinhw/debug.h>
#include <rinhw/buzzer.h>
#include "kbd_buzzer.h"
#include <rinhw/timemath.h>

#include <rinhw/kbd_local.h>


/*
 * Keyboard timing constants
 */
#define KBD_POLL_FREQ 50
#define KBD_DEBOUNCE_PRESET 5
#define KBD_REPEAT_HOLDOFF_TIME (OS_TICKS_PER_SEC / 2)
#define KBD_REPEAT_TIME (OS_TICKS_PER_SEC / 10)
#define KBD_HELD_HOLDOFF_TIME (OS_TICKS_PER_SEC / 2)

/*
 * Keyboard counter preset calculations
 */
#define KBD_POLL_PRESET (OS_TICKS_PER_SEC / KBD_POLL_FREQ)
#define KBD_REPEAT_HOLDOFF_PRESET ((KBD_REPEAT_HOLDOFF_TIME * KBD_POLL_FREQ) / OS_TICKS_PER_SEC)
#define KBD_REPEAT_PRESET ((KBD_REPEAT_TIME * KBD_POLL_FREQ) / OS_TICKS_PER_SEC)
#define KBD_HELD_HOLDOFF_PRESET ((KBD_HELD_HOLDOFF_TIME * KBD_POLL_FREQ) / OS_TICKS_PER_SEC)

/*
 * Keyboard character queue
 */
#define KBD_BUF_SIZE 20

/*
 * Number of keys handled
 */
#define NUM_KEYS 64
#define KEYS_PER_KB_DB 32
#define NUM_KB_DB (NUM_KEYS / KEYS_PER_KB_DB)

/*
 * Key handling struct, deals with 32 keys at a time
 * doing debounce, disable, short, long, repeat presses
 *
 */
struct s_keybit_debouncing
{
    unsigned char offset;
    unsigned long keybits;
    unsigned long last_keybits;         /* Stores raw key states from last time */
    unsigned long repeating_mask;       /* Stores whether a key is currently repeating */
    unsigned long holding_mask;         /* Stores whether a key is currently being held */
    unsigned long held_mask;            /* Stores whether a key has been held until the long press occurred */
    unsigned long debouncing_mask;      /* Stores whether a key is currently being debounced */
    unsigned long debounced_keybits;    /* Stores debounced key states */

    unsigned long disabled_mask;
    unsigned long hold_mask;
    unsigned long repeat_mask;

    unsigned char key_ctr[KEYS_PER_KB_DB];
    unsigned char already_debounced;	/* If set the debouncing has already been done */
};


/** Data required for a keyboard */
struct s_kbd_data
{
  sem_t kbd_busy_sem;
  struct t_tick_fn kbd_bkgrnd;
  int key_buzz_enabled;
  unsigned long external_keybits;
  struct s_keybit_debouncing kb_db[NUM_KB_DB];
  unsigned char kbd_buf[KBD_BUF_SIZE];
  struct s_byteq kbd_q;
  struct s_kbd_hw_definition kbd_hw_definition;
  struct t_io_dev *dev_buzzer;
};


/*
 * PostKeyQuiet -
 *
 * Post a key press/release to the kbd queue, with no beep
 *
 * idx is the key index (0..NUM_KEYS - 1), released if the key has been released
 */
static void PostKeyQuiet(struct s_kbd_data *kd, unsigned char idx, enum tKeyEvent event)
{
  union un_key key;

  key.BIT.KEY = IndexToKey(idx);
  key.BIT.EVENT = (unsigned char) event;

  (void) ByteQWrite(&kd->kbd_q, key.BYTE, NULL);
}


/*
 * PostKey -
 *
 * Post a key press/release to the kbd queue, with a beep
 *
 * idx is the key index (0..NUM_KEYS - 1), released if the key has been released
 */
static void PostKey(struct s_kbd_data *kd, unsigned char idx, enum tKeyEvent event)
{
  PostKeyQuiet(kd, idx, event);
  kbd_buzzer(kd->dev_buzzer, kd->key_buzz_enabled, event);
}


/** Initialise the keyboard struct, offset is the key offset 0, 32, etc */
static void init_kb_struct(struct s_keybit_debouncing *kb, unsigned char offset, unsigned char external_debounce)
{
  memset((void *) kb, 0, sizeof(struct s_keybit_debouncing));
  kb->offset = offset;
  kb->already_debounced = external_debounce;
}


/** De-bounce a set of keys */
static void debounce_keybits(struct s_kbd_data *kd, struct s_keybit_debouncing *kb)
{
  unsigned long bit_changes; /* Which keys changed this time */
  unsigned long mask = 1UL;
  unsigned char i;

  /*
   * Work out which bits have changed
   */
  bit_changes = kb->last_keybits ^ kb->keybits;

  for (i = 0; i < KEYS_PER_KB_DB; i++, mask <<= 1)
  {
    if ((bit_changes & mask) != 0UL)
    {

      /*
       * This bit has changed, re-initialise its counter
       */
      if (kb->already_debounced)
        kb->key_ctr[i] = 1;
      else
        kb->key_ctr[i] = KBD_DEBOUNCE_PRESET;
      kb->debouncing_mask |= mask;
    }
    else if (kb->already_debounced || ((bit_changes & mask) == 0UL))
    {

      /*
       * No change this time or key externally debounced.
       * If debouncing the key (or first change), decrement the debounce counter
       */
      if ((kb->debouncing_mask & mask) != 0UL)
      {
        if (--kb->key_ctr[i] == 0)
        {

          /*
           * Counter has got to zero, the key is debounced, turn it on/off
           */
          kb->debouncing_mask &= ~mask;
          if ((kb->keybits & mask) != 0UL)
          {

            /*
             * Key down, keep track of this key down
             */
            kb->debounced_keybits |= mask;

            /*
             * If the key is disabled, post a key disabled event
             */
            if ((kb->disabled_mask & mask) != 0UL)
            {
              PostKey(kd, (unsigned char) (i + kb->offset), keDisabled);
            }
            else
            {

              /*
               * Key is held, so don't send the key down now, but do beep
               */
              if ((kb->hold_mask & mask) != 0UL)
              {
                kb->holding_mask |= mask;
                kb->key_ctr[i] = KBD_HELD_HOLDOFF_PRESET;
                kbd_buzzer(kd->dev_buzzer, kd->key_buzz_enabled, keDown);
              }
              else
              {
                PostKey(kd, (unsigned char) (i + kb->offset), keDown);
                if ((kb->repeat_mask & mask) != 0UL)
                {
                  kb->repeating_mask |= mask;
                  kb->key_ctr[i] = KBD_REPEAT_HOLDOFF_PRESET;
                }
              }
            }
          }
          else
          {

            /*
             * Key released
             *
             * If the key is configured as a held key, but not held long enough to hold
             * Then its just a basic keydown/keyup key press
             */
            if ((kb->debounced_keybits & mask) != 0UL)
            {
              if (((kb->holding_mask & mask) != 0UL) && ((kb->held_mask & mask) == 0UL))
              {
                PostKeyQuiet(kd, (unsigned char) (i + kb->offset), keDown);
              }

              /*
               * If the key is enabled, post the key up event
               */
              if ((kb->disabled_mask & mask) == 0UL)
                PostKey(kd, (unsigned char) (i + kb->offset), keUp);
            }

            /*
             * Remove the keys from the mask
             */
            kb->debounced_keybits &= ~mask;
            kb->repeating_mask &= ~mask;
            kb->holding_mask &= ~mask;
            kb->held_mask &= ~mask;
          }
        }
      }
      else if ((kb->repeating_mask & mask) != 0UL)
      {
        if (--kb->key_ctr[i] == 0)
        {

          /*
           * Counter has got to zero, time to do a key repeat
           */
          kb->key_ctr[i] = KBD_REPEAT_PRESET;
          PostKey(kd, (unsigned char) (i + kb->offset), keDown);
        }
      }
      else if ((kb->holding_mask & mask) != 0UL)
      {
        if (kb->key_ctr[i] > 0)
        {
          kb->key_ctr[i]--;
          if (kb->key_ctr[i] == 0)
          {

            /*
             * Counter has got to zero, time to signal the key held,
             * then do nothing.
             */
            PostKey(kd, (unsigned char) (i + kb->offset), keHeld);
            kb->held_mask |= mask;
          }
        }
      }
    }
  }
  kb->last_keybits = kb->keybits;
}


/** The poll function which samples the keyboard and generates keystrokes */
static void KeyboardPoll(void *data)
{
  int i;
  struct s_kbd_data *kd = (struct s_kbd_data *) data;

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("KeyboardPoll called\n");
#endif

  /* Check we have a keyboard data */
  if (NULL == kd)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("KeyboardPoll: no keyboard data\n");
#endif
    return;
  }

  /* Lock the semaphore */
  if (sem_wait(&kd->kbd_busy_sem) < 0)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("KeyboardPoll: keyboard semaphore wait error\n");
#endif
    return;
  }

  /* Get low 32bits from keyboard */
  if (NULL != kd->kbd_hw_definition.kbd_hw_readkeybits)
    kd->kb_db[0].keybits = kd->kbd_hw_definition.kbd_hw_readkeybits(kd->kbd_hw_definition.hw_data);

  /* Get high 32bits from the function call or external IOCTL write */
  if (NULL != kd->kbd_hw_definition.kbd_hw_readhikeybits)
    kd->kb_db[1].keybits = kd->kbd_hw_definition.kbd_hw_readhikeybits(kd->kbd_hw_definition.hw_data);
  else
    kd->kb_db[1].keybits = kd->external_keybits;

  /* Transfer over the already de-bounced flag */
  kd->kb_db[0].already_debounced = kd->kbd_hw_definition.already_debounced;

  /* Convert keybits to keys in each of the key structs */
  for (i = 0; i < NUM_KB_DB; i++)
  {
    debounce_keybits(kd, &kd->kb_db[i]);
  }

  /* Release the semaphore */
  if (sem_post(&kd->kbd_busy_sem) < 0)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("KeyboardPoll: keyboard semaphore post error\n");
#endif
    return;
  }

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("KeyboardPoll finished\n");
#endif
}


/** Keyboard device initialisation */
static int kbd_local_init(struct t_io_dev *io_dev)
{
  /* struct s_kbd_data *kd; */

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_local_init called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
  {
    return RIO_ERROR;
  }
  /* kd = (struct s_kbd_data *) io_dev->dev_data; */

  /* Nothing to do here, since it's all done in the device create function */

  /* All done */
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_local_init done\n");
#endif
  return RIO_SUCCESS;
}


/** Destroy keyboard device */
static int kbd_local_destroy(struct t_io_dev *io_dev)
{
  struct s_kbd_data *kd;

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_local_destroy called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == io_dev->dev_data)
  {
    return RIO_ERROR;
  }
  kd = (struct s_kbd_data *) io_dev->dev_data;

  /* Remove the background poll function */
  BackgroundRemoveFunction(&kd->kbd_bkgrnd);

  /* If we have a hardware destroy function, then call it */
  if (NULL != kd->kbd_hw_definition.kbd_hw_destroy)
  {
    kd->kbd_hw_definition.kbd_hw_destroy(kd->kbd_hw_definition.hw_data);
  }

  /* Destroy byte queue */
  (void) ByteQDestroy(&kd->kbd_q);

  /* Delete busy semaphore */
  sem_destroy(&kd->kbd_busy_sem);

  /* Free memory */
  io_dev->dev_data = NULL;
  free(kd);
  kd = NULL;

  return RIO_SUCCESS;
}


/** Read characters from the keyboard device */
static int kbd_local_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_kbd_data *kd;
  unsigned char ch;
  unsigned char *ptr;
  unsigned count = len;
  struct timespec to;
  enum eBQResult err;

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("kbd_local_read called\n");
#endif

  if (NULL == io_dev || NULL == io_dev->dev_data || NULL == buf)
  {
    return RIO_ERROR;
  }
  kd = (struct s_kbd_data *) io_dev->dev_data;
  ptr = (unsigned char *) buf;

  /* Retrieve the data, byte by byte */
  while (count--)
  {

    /* Get char from byte queue, and handle errors */
    OSTicksToTimespec(&to, timeout);
    ch = ByteQRead(&kd->kbd_q, &to, &err);
    if (err != bqrSuccess)
    {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("kbd_local_read: kbd queue timeout or read error (code: %d)\n", err);
#endif
      return RIO_ERROR;
    }

    /* Put character into buffer */
    *ptr++ = ch;
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("kbd_local_read: kbd queue got char %c (%d)\n", ch, ch);
#endif
  }

  /* Done, success */
  return (int) len;
}


/** Write characters to the keyboard device */
static int kbd_local_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_kbd_data *kd;
  unsigned long mask;
  struct s_keybit_debouncing *kb;
  unsigned char idx;
  unsigned char *ptr;
  unsigned count = len;
  union un_key key;

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("kbd_local_write called\n");
#endif

  if (NULL == io_dev || NULL == io_dev->dev_data || NULL == buf)
  {
    return RIO_ERROR;
  }
  kd = (struct s_kbd_data *) io_dev->dev_data;
  ptr = (unsigned char *) buf;

  /* Lock other hippies out */
  if (sem_timedwait_ticks(&kd->kbd_busy_sem, timeout) < 0)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("kbd_local_write: keyboard semaphore wait error\n");
#endif
    return RIO_ERROR;
  }

  /* For each character */
  while (count-- > 0)
  {

    /* Get the key from the buffer */
    key.BYTE = *ptr++;

    /* Make sure the key is a real key */
    if ((idx = KeyToIndex(key.BIT.KEY)) == KEY_LOOKUP_FAIL)
    {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("kbd_local_write: not a real key\n");
#endif
      goto KWFail;
    }

    /*
     * Figure out which kb struct we're talking about,
     * and create a mask for this key
     */
    kb = &kd->kb_db[idx / KEYS_PER_KB_DB];
    mask = 1UL << (idx % KEYS_PER_KB_DB);

    /* Make sure the (physical) key is not pressed */
    if ((kb->debounced_keybits & mask) != 0UL)
    {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("kbd_local_write: the real version of the key is already pressed\n");
#endif
      goto KWFail;
    }

    /* Allow only a key down or a key held event */
    if ((key.BIT.EVENT != keDown) && (key.BIT.EVENT != keHeld))
    {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("kbd_local_write: only key down and key held write is allowed\n");
#endif
      goto KWFail;
    }

    /* Deal with the case of a disabled key */
    if ((kb->disabled_mask & mask) != 0UL)
    {
      PostKeyQuiet(kd, idx, keDisabled);
    }
    else
    {

      /*
       * Post the key to the buffer
       * If it is a held key, and key is defined as a held key, then insert held keycode into buffer
       */
      if (key.BIT.EVENT == keDown)
      {
        PostKeyQuiet(kd, idx, keDown);
        PostKeyQuiet(kd, idx, keUp);
      }
      else if ((kb->hold_mask & mask) != 0UL)
      {
        PostKeyQuiet(kd, idx, keHeld);
        PostKeyQuiet(kd, idx, keUp);
      }
      else
      {
        goto KWFail;
      }
    }
  }

  /* Success, let the hippies back in, and return success */
  if (sem_post(&kd->kbd_busy_sem) < 0)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("kbd_local_write: keyboard semaphore post error\n");
#endif
    return RIO_ERROR;
  }

  return (int) len;

  /* It failed, let the hippies back in, and return error :( */
KWFail:
  (void) sem_post(&kd->kbd_busy_sem);
  return RIO_ERROR;
}


/** Get the number of characters in the keyboard queue */
static long kbd_local_filelength(struct t_io_dev *io_dev)
{
  struct s_kbd_data *kd = (struct s_kbd_data *) io_dev->dev_data;

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_local_filelength called\n");
#endif

  if (NULL == io_dev || NULL == io_dev->dev_data)
  {
    return RIO_ERROR;
  }
  kd = (struct s_kbd_data *) io_dev->dev_data;

  return (long) ByteQGetLevel(&kd->kbd_q);
}


/** Setup a number of keys */
static unsigned char configure_keys(struct s_kbd_data *kd, int func, unsigned char numkeys, unsigned char *keys)
{
  unsigned long mask;
  unsigned char i, idx;
  struct s_keybit_debouncing *kb;

  /* For the number of keys to configure... */
  for (i = 0; i < numkeys; i++, keys++)
  {

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("Setting key %d\n", *keys);
#endif

    /*
     * Convert the key to a bitmask
     */
    if ((idx = KeyToIndex(*keys)) == KEY_LOOKUP_FAIL)
      return 0;

    /*
     * Figure out which kb struct we're talking about,
     * and create a mask for this key
     */
    kb = &kd->kb_db[idx / KEYS_PER_KB_DB];
    mask = 1UL << (idx % KEYS_PER_KB_DB);

    /* Turn all bitmasks off to start with */
    kb->disabled_mask &= ~mask;
    kb->repeat_mask &= ~mask;
    kb->hold_mask &= ~mask;

    /* Configure as requested */
    switch (func)
    {
      case IOCTL_KBD_SET_BASIC:
        break;
      case IOCTL_KBD_SET_REPEAT:
        kb->repeat_mask |= mask;
        break;
      case IOCTL_KBD_SET_HOLD:
        kb->hold_mask |= mask;
        break;
      case IOCTL_KBD_SET_DISABLED:
        kb->disabled_mask |= mask;
        break;
      default:
        return 0;
    }
  }
  return 1;
}


/** Keyboard IOCTL functions */
static int kbd_local_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_kbd_data *kd;
  unsigned char num_keys;
  unsigned char *keys;
  int result = RIO_ERROR;
  unsigned char *pressed;

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("kbd_local_ioctl called\n");
#endif

  if (NULL == io_dev || NULL == io_dev->dev_data)
  {
    return RIO_ERROR;
  }
  kd = (struct s_kbd_data *) io_dev->dev_data;

  /* Function dispatch */
  switch (func)
  {

    /* Configure some keys */
    case IOCTL_KBD_SET_BASIC:
    case IOCTL_KBD_SET_REPEAT:
    case IOCTL_KBD_SET_HOLD:
    case IOCTL_KBD_SET_DISABLED:
      num_keys = (unsigned char) va_arg(*arglist, int);
      keys = va_arg(*arglist, unsigned char *);
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("Config keys, func=%d, num_keys=%d, keys_ptr=%p\n", func, num_keys, keys);
#endif
      if (configure_keys(kd, func, num_keys, keys))
        result = RIO_SUCCESS;
      break;

    /* Set the key bits for the external 32keys */
    case IOCTL_KBD_WRITE_HI_KEY_BITS:
      if (NULL != kd->kbd_hw_definition.kbd_hw_readhikeybits)
        kd->external_keybits = va_arg(*arglist, unsigned long);
      result = RIO_SUCCESS;
      break;

    /* Returns true or false depending on whether the key is currently pressed */
    case IOCTL_KBD_IS_KEY_PRESSED:
      pressed = va_arg(*arglist, unsigned char *); // get address of answer
      *pressed = ((kd->kb_db[0].keybits & (0x00000001 << KeyToIndex(*pressed))) != 0); // work out if key is pressed
      result = RIO_SUCCESS;
      break;

    /* Returns internal key status as a maskable unsigned long */
    case IOCTL_KBD_INTERNAL_BITS:
      *(va_arg(*arglist, unsigned long *)) = kd->kb_db[0].debounced_keybits;
      result = RIO_SUCCESS;
      break;

    /* Enable or disable the buzzer when keys are pressed */
    case IOCTL_KBD_ENABLE_BUZZER:
      kd->key_buzz_enabled = va_arg(*arglist, int);
      result = RIO_SUCCESS;
      break;
  }

  /* Done, return result */
  return result;
}


/**
 * Function to create and initialise a local keyboard device
 *
 * @param io_dev is the device structure to be initialised as a keyboard
 * @param kbd_hw_definition is the keyboard hardware device structure, gets copied into driver data, so doesn't have to be allocated for the duration of the device. Can be NULL (in which case driver will run, but no keys will ever appear).
 * @param dev_buzzer is the buzzer device to use for key beeps. Can be NULL (in which case there will be no key beeps).
 */
int rio_create_device_kbd_local(struct t_io_dev *io_dev, struct s_kbd_hw_definition *kbd_hw_definition, struct t_io_dev *dev_buzzer)
{
  struct s_kbd_data *kd;

#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("rio_create_device_kbd_local called\n");
#endif

  /* Check pointers, kbd_hw_definition & dev_buzzer can be NULL */
  if (NULL == io_dev)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_kbd_local: io_dev cannot be NULL\n");
#endif
    goto rcdkl_fail;
  }

  /* Allocate memory, and initialise */
  kd = malloc(sizeof(struct s_kbd_data));
  if (NULL == kd)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_kbd_local: failed to malloc keyboard data\n");
#endif
    goto rcdkl_fail;
  }
  memset(kd, 0, sizeof(struct s_kbd_data));

  /* Create semaphore */
  if (sem_init(&kd->kbd_busy_sem, 0, 1) < 0)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_kbd_local: error creating busy semaphore\n");
#endif
    goto rcdkl_free;
  }

  /* Create the byte queue */
  if (ByteQCreate(&kd->kbd_q, bqPendOnRead, &kd->kbd_buf[0], KBD_BUF_SIZE) != bqrSuccess)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_kbd_local: error creating kbd buffer\n");
#endif
    goto rcdkl_sem;
  }

  /* Setup the structs for the keys */
  init_kb_struct(&kd->kb_db[0], 0, FALSE);
  init_kb_struct(&kd->kb_db[1], KEYS_PER_KB_DB, TRUE);

  /* Initialise keyboard hardware */
  if (NULL != kbd_hw_definition)
  {
    /* Take a copy */
    kd->kbd_hw_definition = *kbd_hw_definition;
  }
  else
  {
    /* No keyboard hardware */
    kd->kbd_hw_definition.already_debounced = true;
    kd->kbd_hw_definition.hw_data = NULL;
    kd->kbd_hw_definition.kbd_hw_destroy = NULL;
    kd->kbd_hw_definition.kbd_hw_init = NULL;
    kd->kbd_hw_definition.kbd_hw_readkeybits = NULL;
    kd->kbd_hw_definition.kbd_hw_readhikeybits = NULL;
  }

  /* Setup keyboard buzzer device */
  kd->dev_buzzer = dev_buzzer;

  /* Init other members */
  kd->key_buzz_enabled = true;
  kd->external_keybits = 0UL;

  /* Initialise the keyboard hardware before the poll function starts */
  if (NULL != kd->kbd_hw_definition.kbd_hw_init)
  {
    kd->kbd_hw_definition.kbd_hw_init(kd->kbd_hw_definition.hw_data);
  }

  /* Create the background polling function */
  kd->kbd_bkgrnd.preset = KBD_POLL_PRESET;
  kd->kbd_bkgrnd.count = 0;
  kd->kbd_bkgrnd.fn = KeyboardPoll;
  kd->kbd_bkgrnd.data = kd;
  kd->kbd_bkgrnd.next = NULL;
  BackgroundAddFunction(&kd->kbd_bkgrnd);

  /*
   * Setup character device
   */
  io_dev->posn = 0UL;
  io_dev->status = O_DEVCHAR;
  io_dev->dev_data = (void *) kd;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = kbd_local_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = kbd_local_read;
  io_dev->dev_write = kbd_local_write;
  io_dev->dev_filelength = kbd_local_filelength;
  io_dev->dev_ioctl = kbd_local_ioctl;
  io_dev->dev_destroy = kbd_local_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_kbd_local: device registration failed\n");
#endif
    goto rcdkl_bq;
  }

  /* Success */
  return RIO_SUCCESS;

  /* Failure */
rcdkl_bq:
  BackgroundRemoveFunction(&kd->kbd_bkgrnd);
  ByteQDestroy(&kd->kbd_q);
rcdkl_sem:
  sem_destroy(&kd->kbd_busy_sem);
rcdkl_free:
  free(kd);
rcdkl_fail:
  return RIO_ERROR;
}
