/*
 * lcd_char.c
 *
 *  Created on: 29/05/2012
 *      Author: ian
 */
#include <string.h>
#include <rinhw/debug.h>
#include <rinhw/rio.h>
#include <rinhw/lcd_generic.h>
#include "lcd_generic_common.h"

#include "lcd_generic_char.h"


/* Macro to work out smaller of two numbers */
#ifndef min
  #define min(x, y) ((x < y) ? x : y)
#endif


/** Device read from the LCD character buffer */
int lcd_generic_char_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;
  unsigned short count;

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* More pointer checks */
  if (NULL == buf || NULL == lcd_definition->lcd_characters)
  {
    return RIO_ERROR;
  }

  /* Work out how many bytes to read, with limit */
  if (io_dev->posn >= lcd_definition->lcd_num_characters)
  {
    return RIO_ERROR;
  }
  count = (unsigned short) min(len, lcd_definition->lcd_num_characters - io_dev->posn);

  /* Copy the data */
  memcpy(buf, (void *) (&lcd_definition->lcd_characters[io_dev->posn]), (size_t) count);

  /* Update device position, return count */
  io_dev->posn += count;
  return (int) count;
}


/** Device write of characters to the LCD (converts to digit bitmap before writing to LCD) */
int lcd_generic_char_write(struct t_io_dev *io_dev, const void *buf, unsigned short len)
{
  struct s_lcd_definition *lcd_definition;
  unsigned short i, real_count = 0;
  unsigned char *rd_ptr;
  unsigned char wr_pos = (unsigned char) io_dev->posn;
  unsigned char ch;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_char_write: called to write %d bytes\n", len);
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* More pointer checks */
  if (NULL == buf || NULL == lcd_definition->lcd_characters)
  {
    return RIO_ERROR;
  }
  rd_ptr = (unsigned char *) buf;

  /* Loopy */
  for (i = 0; (i < len && wr_pos < lcd_definition->lcd_num_characters); i++)
  {

    /* Get each character to be written */
    ch = *rd_ptr++;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("lcd_generic_char_write: Writing char %d of %d, char %c (0x%02x), write buf posn %d, real character count %d\n",
                 i, len, ch, ch, wr_pos, real_count);
#endif

    /* Deal with decimal points */
    if (ch == '.')
    {

      /*
       * Special handling for the decimal point -
       * If the previous digit has a decimal point, turn it on
       */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("lcd_generic_char_write: lcd_does_digit_have_dp() at posn %d = %d\n", wr_pos - 1, lcd_definition->lcd_does_digit_have_dp(lcd_definition, (unsigned char) (wr_pos - 1)));
#endif
      if (NULL != lcd_definition->lcd_does_digit_have_dp && lcd_definition->lcd_does_digit_have_dp(lcd_definition, (unsigned char) (wr_pos - 1)))
      {

        /* Turn the decimal point on the previous digit */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
        DEBUG_PRINTF("Setting decimal point on previous digit\n");
#endif
        if (NULL != lcd_definition->lcd_put_char)
          lcd_definition->lcd_put_char(lcd_definition, (unsigned char) (wr_pos - 1), '.', cwmOR);
      }
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
      else
      {

        /*
         * Otherwise turn the decimal point on for this digit.
         */
        DEBUG_PRINTF("lcd_generic_char_write: Previous digit (%d) does not have DP, ignoring\n", wr_pos - 1);
      }
#endif
    }
    else
    {

      /*
       * Normal character
       */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("lcd_generic_char_write: Writing normal character %c (0x%02x) to posn %d\n", ch, ch, wr_pos);
#endif
      /* Keep a copy of the character (decimal points not recorded) */
      lcd_definition->lcd_characters[wr_pos] = ch;

      /* Write character to segment buffer */
      if (NULL != lcd_definition->lcd_put_char)
        lcd_definition->lcd_put_char(lcd_definition, wr_pos, ch, cwmSet);
      wr_pos++;
      real_count++;
    }
  }

  /*
   * All done, update device position and return
   * count of characters written
   */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("lcd_generic_char_write: finished, wrote %d characters, %d of which were not decimal points\n", i, real_count);
#endif
  io_dev->posn += real_count;
  return (int) i;
}
