/*
 * lcd_generic_nosegflash.c
 *
 *  Created on: 29/05/2012
 *      Author: ian
 */
#include <stdarg.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <stdlib.h>
#include <rinhw/rio.h>
#include <rinhw/lcd.h>
#include <rinhw/debug.h>
#include <rinhw/timemath.h>
#include "lcd_generic_char.h"
#include "lcd_generic_common.h"

#include <rinhw/lcd_generic.h>


/** Function to force an LCD update */
static int lcd_generic_nosegflash_update(struct s_lcd_definition *lcd_definition)
{
  /* Update the LCD directly, no update task */
  if (NULL != lcd_definition->lcd_hw_write)
  {
    return lcd_definition->lcd_hw_write(lcd_definition);
  }
  else
  {
    return RIO_SUCCESS;
  }
}


/** Device initialisation function */
static int lcd_generic_nosegflash_init(struct t_io_dev *io_dev)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_nosegflash_init called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* More pointer checks */
  if (NULL == lcd_definition->lcd_segments ||
      NULL == lcd_definition->lcd_characters)
  {
    return RIO_ERROR;
  }

  /* Call hardware initialisation function if there is one */
  if (NULL != lcd_definition->lcd_hw_init)
  {
    lcd_definition->lcd_hw_init(lcd_definition);
  }

  /* Zero out segments */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("lcd_generic_nosegflash_init: clearing segments\n");
#endif
  memset(lcd_definition->lcd_segments, 0, lcd_definition->lcd_num_segment_bytes);
  memset(lcd_definition->lcd_characters, 0, lcd_definition->lcd_num_characters);

  /* Write data to LCD */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("lcd_generic_nosegflash_init: updating lcd\n");
#endif
  lcd_generic_nosegflash_update(lcd_definition);

  /* Its a block device */
  io_dev->status &= ~O_DEVCHAR;

  return RIO_SUCCESS;
}


/** Device shutdown function */
static int lcd_generic_nosegflash_destroy(struct t_io_dev *io_dev)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_nosegflash_destroy called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Free up stuff */
  io_dev->dev_data = NULL;

  /* Shutdown the hardware */
  if (NULL != lcd_definition->lcd_hw_destroy)
  {
    lcd_definition->lcd_hw_destroy(lcd_definition);
  }

  /* Pointers to NULL */
  lcd_definition->lcd_current_access = NULL;
  lcd_definition->hw_data = NULL;

  /* Function pointers to NULL */
  lcd_definition->lcd_hw_init = NULL;
  lcd_definition->lcd_hw_destroy = NULL;
  lcd_definition->lcd_hw_write = NULL;
  lcd_definition->lcd_put_char = NULL;
  lcd_definition->lcd_does_digit_have_dp = NULL;
  lcd_definition->lcd_hw_setbacklight_fn = NULL;

  /* Free buffers */
  if (NULL != lcd_definition->lcd_segments)
  {
    free(lcd_definition->lcd_segments);
    lcd_definition->lcd_segments = NULL;
  }
  if (NULL != lcd_definition->lcd_characters)
  {
    free(lcd_definition->lcd_characters);
    lcd_definition->lcd_characters = NULL;
  }

  /* These don't get allocated with no segment flashing driver */
  /* lcd_definition->lcd_output_segments */
  /* lcd_definition->lcd_flash_segments */

  /* Free LCD definition */
  free(lcd_definition);
  lcd_definition = NULL;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_nosegflash_destroy finished\n");
#endif
  return RIO_SUCCESS;
}


/** Open the generic LCD device */
static int lcd_generic_nosegflash_open(struct t_io_dev *io_dev, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;
  UNREFERENCED_PARAMETER(timeout); /* Stop compiler warnings on unused parameter */

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_nosegflash_open called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Set default state on open */
  lcd_definition->lcd_access_segs_not_chars = TRUE;

  return RIO_SUCCESS;
}


/** Close the generic LCD device, and force an update */
static int lcd_generic_nosegflash_close(struct t_io_dev *io_dev)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_nosegflash_close called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* The LCD close call forces an update */
  return lcd_generic_nosegflash_update(lcd_definition);
}


/*
 * Macro to work out smaller of two numbers
 */
#ifndef min
#define min(x, y) ((x < y) ? x : y)
#endif


/** Read from the LCD data segment buffer */
static int lcd_generic_nosegflash_segs_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;
  const unsigned char *rd_ptr;
  unsigned short count;

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* More pointer checks */
  if (NULL == buf || NULL == lcd_definition->lcd_segments)
  {
    return RIO_ERROR;
  }
  rd_ptr = &lcd_definition->lcd_segments[io_dev->posn];

  /* Work out how many bytes to read, with limit */
  if (io_dev->posn >= lcd_definition->lcd_num_segment_bytes)
  {
    return RIO_ERROR;
  }
  count = min(len, (unsigned short) (lcd_definition->lcd_num_segment_bytes - io_dev->posn));

  /* Copy data to LCD buffer, and update position */
  memcpy(buf, rd_ptr, (size_t) count);
  io_dev->posn += count;

  /* Return number of bytes copied */
  return (int) count;
}


/** LCD device read */
static int lcd_generic_nosegflash_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_nosegflash_read called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Read data from the buffer we are accessing */
  if (lcd_definition->lcd_access_segs_not_chars)
  {
    return lcd_generic_nosegflash_segs_read(io_dev, buf, len, timeout);
  }
  else
  {
    return lcd_generic_char_read(io_dev, buf, len, timeout);
  }
}


/**
 * Write to the LCD data segment buffer
 *
 * NOTE: This doesn't force an update
 */
static int lcd_generic_nosegflash_segs_write(struct t_io_dev *io_dev, const void *buf, unsigned short len)
{
  struct s_lcd_definition *lcd_definition;
  unsigned char *wr_ptr;
  unsigned short count;

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* More pointer checks */
  if (NULL == buf || NULL == lcd_definition->lcd_segments)
  {
    return RIO_ERROR;
  }
  wr_ptr = &lcd_definition->lcd_segments[io_dev->posn];

  /* Work out how many bytes to write, with limit */
  if (io_dev->posn >= lcd_definition->lcd_num_segment_bytes)
  {
    return RIO_ERROR;
  }
  count = min(len, (unsigned short) (lcd_definition->lcd_num_segment_bytes - io_dev->posn));

  /*
   * Copy data to LCD buffer, and update position
   * Don't update LCD, since this is done by lcd_close
   */
  memcpy(wr_ptr, buf, (size_t) count);
  io_dev->posn += count;

  /* Return number of bytes copied */
  return (int) count;
}


/** LCD device write */
static int lcd_generic_nosegflash_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_write called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Write data to the buffer we are accessing */
  if (lcd_definition->lcd_access_segs_not_chars)
  {
    return lcd_generic_nosegflash_segs_write(io_dev, buf, len);
  }
  else
  {
    return lcd_generic_char_write(io_dev, buf, len);
  }
}


/** LCD device specific configuration */
static int lcd_generic_nosegflash_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_lcd_definition *lcd_definition;
  unsigned char level;
  int result = RIO_SUCCESS;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_nosegflash_ioctl called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Function dispatch */
  switch (func)
  {

    /* Set the backlight level */
    case IOCTL_LCD_SET_BACKLIGHT:
      level = (unsigned char) va_arg(*arglist, int);
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_LCD_SET_BACKLIGHT: level %d\n", level);
#endif
      if (NULL != lcd_definition->lcd_hw_setbacklight_fn)
      {
        lcd_definition->lcd_hw_setbacklight_fn(lcd_definition, level);
      }
      break;

    /* Change the read and write buffer access */
    case IOCTL_LCD_ACCESS_OUTPUT:
    case IOCTL_LCD_ACCESS_DATA:
      lcd_definition->lcd_access_segs_not_chars = TRUE;
      io_dev->posn = 0UL;
      break;

    case IOCTL_LCD_ACCESS_CHAR:
      lcd_definition->lcd_access_segs_not_chars = FALSE;
      io_dev->posn = 0UL;
      break;

    /* Error */
    default:
      result = RIO_ERROR;
      break;
  }

  /*
   * Done, return result
   */
  return result;
}


/**
 * Create a generic LCD device without segment flashing (doesn't require a task to run)
 *
 * @param io_dev is the uninitialised IO device to be filled out
 * @param lcd_hw_data is a pointer to your LCD hardware implementation data
 * @param lcd_num_segment_bytes is the number of segment bytes in the LCD
 * @param lcd_num_chars is the number of characters on the LCD
 * @param lcd_hw_init is a function pointer called to initialise the LCD. Can be NULL, in which case it won't be called
 * @param lcd_hw_destroy is a function pointer called to destroy the LCD. Can be NULL, in which case it won't be called
 * @param lcd_hw_write is a function pointer called to write data to the LCD. Can be NULL, in which case it won't be called
 * @param lcd_put_char is a function pointer called to put a character on the LCD. Not NULL.
 * @param lcd_does_digit_have_dp is a function pointer called to determine if a character position has a decimal point. Not NULL
 * @param lcd_hw_setbacklight_fn is a function pointer called to set the back light level 0..100%. Can be NULL, in which case it won't be called
 *
 * @return RIO_SUCCESS on success, RIO_ERROR otherwise
 */
int lcd_generic_nosegflash_create_device(struct t_io_dev *io_dev,
                                       void *lcd_hw_data,
                                       size_t lcd_num_segment_bytes,
                                       size_t lcd_num_characters,
                                       t_lcd_hw_init_fn lcd_hw_init,
                                       t_lcd_hw_destroy_fn lcd_hw_destroy,
                                       t_lcd_hw_write_fn lcd_hw_write,
                                       t_lcd_put_char_fn lcd_put_char,
                                       t_lcd_does_digit_have_dp_fn lcd_does_digit_have_dp,
                                       t_lcd_hw_setbacklight_fn lcd_hw_setbacklight_fn
                                      )
{
  struct s_lcd_definition *lcd_definition = NULL;

  /* Check pointers */
  if (NULL == io_dev || NULL == lcd_put_char || NULL == lcd_does_digit_have_dp)
  {
    goto lgcd_fail;
  }

  /* Create LCD definition */
  lcd_definition = malloc(sizeof(struct s_lcd_definition));
  if (NULL == lcd_definition)
  {
    goto lgcd_fail;
  }
  memset(lcd_definition, 0, sizeof(struct s_lcd_definition));

  lcd_definition->hw_data = lcd_hw_data;

  /* Allocate segments */
  lcd_definition->lcd_num_segment_bytes = lcd_num_segment_bytes;
  lcd_definition->lcd_segments = malloc(lcd_definition->lcd_num_segment_bytes);
  if (NULL == lcd_definition->lcd_segments)
  {
    goto lgcd_fail;
  }

  /* Segment flash, not supported by this driver */
  lcd_definition->lcd_output_segments = NULL;
  lcd_definition->lcd_flash_segments = NULL;
  lcd_definition->lcd_current_access = NULL;
  /* lcd_definition->sem_lcd_update = 0; */
  /* lcd_definition->lcd_flash_thread = 0; */
  /* lcd_definition->lcd_thread_quit = 0; */

  /* Allocate characters */
  lcd_definition->lcd_num_characters = lcd_num_characters;
  lcd_definition->lcd_characters = malloc(lcd_definition->lcd_num_characters);
  if (NULL == lcd_definition->lcd_characters)
  {
    goto lgcd_fail;
  }

  /* Functions */
  lcd_definition->lcd_hw_init = lcd_hw_init;
  lcd_definition->lcd_hw_destroy = lcd_hw_destroy;
  lcd_definition->lcd_hw_write = lcd_hw_write;
  lcd_definition->lcd_put_char = lcd_put_char;
  lcd_definition->lcd_does_digit_have_dp = lcd_does_digit_have_dp;
  lcd_definition->lcd_hw_setbacklight_fn = lcd_hw_setbacklight_fn;

  /* Other members */
  lcd_definition->lcd_access_segs_not_chars = true;

  /*
   * Setup device data pointer to point to the LCD definition
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) lcd_definition;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = lcd_generic_nosegflash_init;
  io_dev->dev_open = lcd_generic_nosegflash_open;
  io_dev->dev_close = lcd_generic_nosegflash_close;
  io_dev->dev_read = lcd_generic_nosegflash_read;
  io_dev->dev_write = lcd_generic_nosegflash_write;
  io_dev->dev_filelength = lcd_generic_common_filelength;
  io_dev->dev_ioctl = lcd_generic_nosegflash_ioctl;
  io_dev->dev_destroy = lcd_generic_nosegflash_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
    goto lgcd_fail;
  }

  /* Success */
  return RIO_SUCCESS;

  /* Clean up after fail */
lgcd_fail:
  io_dev->dev_data = NULL;
  if (NULL != lcd_definition)
  {
    if (NULL != lcd_definition->lcd_segments)
    {
      free(lcd_definition->lcd_segments);
      lcd_definition->lcd_segments = NULL;
    }
    if (NULL != lcd_definition->lcd_characters)
    {
      free(lcd_definition->lcd_characters);
      lcd_definition->lcd_characters = NULL;
    }

    /* No segment flash, so these haven't been allocated */
    /* lcd_definition->lcd_output_segments */
    /* lcd_definition->lcd_flash_segments */

    lcd_definition->lcd_current_access = NULL;
    lcd_definition->hw_data = NULL;

    free(lcd_definition);
    lcd_definition = NULL;
  }
  return RIO_ERROR;
}
