/*
 * lcd_generic_segflash.c
 *
 *  Created on: 29/05/2012
 *      Author: ian
 */
#include <stdarg.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <stdlib.h>
#include <rinhw/rio.h>
#include <rinhw/lcd.h>
#include <rinhw/debug.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/timemath.h>
#include <rinhw/bkgrnd.h>
#include <rinhw/rinthread.h>
#include "lcd_generic_char.h"
#include "lcd_generic_common.h"

#include <rinhw/lcd_generic.h>


/*
 * LCD flash timing settings
 */
#define LCD_FLASH_CYCLE_TIME (OS_TICKS_PER_SEC / 2)
#define LCD_FLASH_OFF_DELAY (LCD_FLASH_CYCLE_TIME / 4)
#define LCD_FLASH_ON_DELAY (LCD_FLASH_CYCLE_TIME - LCD_FLASH_OFF_DELAY)


/** Force an LCD update */
static int lcd_generic_segflash_update(struct s_lcd_definition *lcd_definition)
{
  /* Let other LCD update task send new data to the LCD */
  sem_post(&lcd_definition->sem_lcd_update);
  return RIO_SUCCESS;
}


/*
 * FlashLcdData -
 *
 * Modify the output data to flash
 * based on the flash data buffer
 * state is the state of the flash (0 off, 1 on)
 *
 * If a flash bit = 1, and state = 1, the output = lcd_data
 * If a flash bit = 1, and state = 0, the output = 0
 * If a flash bit = 0, the output = lcd_data, regardless of state
 * As described below:
 *
 *   F S D | O
 *   ---------
 *   0 0 0 | 0
 *   0 0 1 | 1
 *   0 1 0 | 0
 *   0 1 1 | 1
 *   1 0 0 | 0
 *   1 0 1 | 0
 *   1 1 0 | 0
 *   1 1 1 | 1
 */
static unsigned char lcd_generic_segflash_flash_lcd_data(struct s_lcd_definition *lcd_definition, unsigned char state)
{
  unsigned char data, flash, i, new;
  unsigned char *output_ptr;
  unsigned char changed = false;

  /* Check pointers */
  if (NULL == lcd_definition ||
      NULL == lcd_definition->lcd_output_segments ||
      NULL == lcd_definition->lcd_segments ||
      NULL == lcd_definition->lcd_flash_segments)
  {
    return false;
  }
  output_ptr = lcd_definition->lcd_output_segments;

  /* We work on whole bytes */
  state = (state != 0) ? 0xFF : 0x00;

  /* For each LCD data byte */
  for (i = 0; i < lcd_definition->lcd_num_segment_bytes; i++)
  {

    /* Work out the new output byte based on the table above */
    data = lcd_definition->lcd_segments[i];
    flash = lcd_definition->lcd_flash_segments[i];
    new = (unsigned char) ((~flash & data) | (state & data));

    /* Figure out if this byte has changed */
    if (new != *output_ptr)
    {
      *output_ptr = new;
      changed = true;
    }
    output_ptr++;
  }

  /* Return if anything has changed */
  return changed;
}


/** Task to update the LCD. Handles all flashing, and data updates */
static void *lcd_generic_segflash_flash_update_thread(void *data)
{
  struct t_io_dev *lcd_io_dev;
  struct s_lcd_definition *lcd_definition;
  unsigned char timeout, err, flash_state = 0;
  unsigned short ticks = LCD_FLASH_OFF_DELAY;
  unsigned long start_time, delta;
  unsigned char update, changed;
  unsigned short waittime;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_flash_update_thread: started\n");
#endif

  /* Check pointers */
  lcd_io_dev = data;
  if (NULL == lcd_io_dev || NULL == lcd_io_dev->dev_data)
  {
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("lcd_generic_segflash_flash_update_thread: quitting because of null pointers\n");
#endif
    goto lgsfut_fail;
  }
  lcd_definition = lcd_io_dev->dev_data;

  /* Task loop, quit when flag set */
  while (!lcd_definition->lcd_thread_quit)
  {

    /* Wait for an update, or a timeout to update the flashing */
    start_time = OSTimeGet();
    update = 0;
    if (sem_timedwait_ticks(&lcd_definition->sem_lcd_update, ticks) < 0)
    {
      /* Timeout is a valid error, anything else is bad */
      if (ETIMEDOUT != errno)
      {
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("lcd_generic_segflash_flash_update_thread: semaphore pend failed\n");
#endif
        break;
      }
    }
    else
    {
      /* Semaphore signaled, so it's an update */
      update = 1;
    }

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
    if (update)
    {
      DEBUG_PRINTF("lcd_generic_segflash_flash_update_thread: update signaled\n");
    }
#endif

    /*
     * Wait for the LCD buffer to become free
     * This is normally indefinite, but if quitting, then we use a timeout so the
     * shutdown process doesn't lock up. This allows for one last update before shutting
     * down
     */
    waittime = (lcd_definition->lcd_thread_quit) ? OS_TICKS_PER_SEC / 2 : 0;
    OSSemPend(lcd_io_dev->sem_open_excl, waittime, &err);

    /*
     * This is a bit of an optimisation,
     * if there are multiple updates pending,
     * then get rid of them, since it will all get updated
     * in one hit now.
     */
    while (sem_trywait(&lcd_definition->sem_lcd_update) == 0)
      ;

    /* Work out if a timeout has expired */
    delta = OSTimeGet() - start_time;
    timeout = (delta >= (unsigned long) ticks);

    /*
     * If timed out, its time to change flash state
     * and reset the ticks for the next timeout
     * otherwise shorten the delay by the amount
     * of time that has expired
     */
    if (timeout)
    {
      flash_state = !flash_state;
      ticks = (flash_state) ? LCD_FLASH_ON_DELAY : LCD_FLASH_OFF_DELAY;
    }
    else
    {
      ticks -= (unsigned short) delta;
    }

    /*
     * It's an update or a timeout, so
     * fix up the output data based on the
     * flash_state
     */
    if (update || timeout)
    {
/*
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("lcd_generic_segflash_flash_update_thread: upd %d, to %d, fs %d\n",
                   update, timeout, flash_state);
#endif
*/
      changed = lcd_generic_segflash_flash_lcd_data(lcd_definition, flash_state);
    }
    else
    {
      changed = false;
    }

    /* Write the data to the LCD only on change */
    if (changed && NULL != lcd_definition->lcd_hw_write)
    {
      (void) lcd_definition->lcd_hw_write(lcd_definition);
    }

    /* Let other tasks send new data to the LCD */
    OSSemPost(lcd_io_dev->sem_open_excl);
  }

lgsfut_fail:
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_flash_update_thread: quitting\n");
#endif

  rin_thread_exit(NULL);
}


/** Device initialisation function */
static int lcd_generic_segflash_init(struct t_io_dev *io_dev)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_init called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* More pointer checks */
  if (NULL == lcd_definition->lcd_segments ||
      NULL == lcd_definition->lcd_flash_segments ||
      NULL == lcd_definition->lcd_output_segments ||
      NULL == lcd_definition->lcd_characters)
  {
    return RIO_ERROR;
  }

  /* Call hardware initialisation function if there is one */
  if (NULL != lcd_definition->lcd_hw_init)
  {
    lcd_definition->lcd_hw_init(lcd_definition);
  }

  /*
   * Zero out segments, except output_segments which are set to 0xFF,
   * so the first write will see that the data is changed, and write
   * the data to the hardware
   */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("lcd_generic_segflash_init: clearing segments\n");
#endif
  memset(lcd_definition->lcd_segments, 0, lcd_definition->lcd_num_segment_bytes);
  memset(lcd_definition->lcd_flash_segments, 0, lcd_definition->lcd_num_segment_bytes);
  memset(lcd_definition->lcd_output_segments, 0xFF, lcd_definition->lcd_num_segment_bytes);
  memset(lcd_definition->lcd_characters, 0, lcd_definition->lcd_num_characters);

  /* Create the update semaphore */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("lcd_generic_segflash_init: creating update semaphore\n");
#endif
  if (sem_init(&lcd_definition->sem_lcd_update, 0, 0) < 0)
  {
    return RIO_ERROR;
  }

  /* Create update task. TODO check create result */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("lcd_generic_segflash_init: creating lcd update task\n");
#endif /* #if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL) */
  lcd_definition->lcd_thread_quit = false;
  (void) rin_thread_create(&lcd_definition->lcd_flash_thread, lcd_generic_segflash_flash_update_thread, io_dev, true);

  /* Write data to LCD */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("lcd_generic_segflash_init: updating lcd\n");
#endif
  lcd_generic_segflash_update(lcd_definition);

  /* Its a block device */
  io_dev->status &= ~O_DEVCHAR;

  return RIO_SUCCESS;
}


/** Device shutdown function */
static int lcd_generic_segflash_destroy(struct t_io_dev *io_dev)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_destroy called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Free up stuff */
  io_dev->dev_data = NULL;

  /* Notify thread to quit */
  lcd_definition->lcd_thread_quit = true;
  sem_post(&lcd_definition->sem_lcd_update);

  /* Wait for thread to quit */
  pthread_join(lcd_definition->lcd_flash_thread, NULL);

  /* Destroy semaphore */
  sem_destroy(&lcd_definition->sem_lcd_update);

  /* Shutdown the hardware */
  if (NULL != lcd_definition->lcd_hw_destroy)
  {
    lcd_definition->lcd_hw_destroy(lcd_definition);
  }

  /* Pointers to NULL */
  lcd_definition->lcd_current_access = NULL;
  lcd_definition->hw_data = NULL;

  /* Function pointers to NULL */
  lcd_definition->lcd_hw_init = NULL;
  lcd_definition->lcd_hw_destroy = NULL;
  lcd_definition->lcd_hw_write = NULL;
  lcd_definition->lcd_put_char = NULL;
  lcd_definition->lcd_does_digit_have_dp = NULL;
  lcd_definition->lcd_hw_setbacklight_fn = NULL;

  /* Free buffers */
  if (NULL != lcd_definition->lcd_segments)
  {
    free(lcd_definition->lcd_segments);
    lcd_definition->lcd_segments = NULL;
  }
  if (NULL != lcd_definition->lcd_characters)
  {
    free(lcd_definition->lcd_characters);
    lcd_definition->lcd_characters = NULL;
  }
  if (NULL != lcd_definition->lcd_output_segments)
  {
    free(lcd_definition->lcd_output_segments);
    lcd_definition->lcd_output_segments = NULL;
  }
  if (NULL != lcd_definition->lcd_flash_segments)
  {
    free(lcd_definition->lcd_flash_segments);
    lcd_definition->lcd_flash_segments = NULL;
  }

  /* Free LCD definition */
  free(lcd_definition);
  lcd_definition = NULL;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_destroy finished\n");
#endif
  return RIO_SUCCESS;
}


/** Open the generic LCD device */
static int lcd_generic_segflash_open(struct t_io_dev *io_dev, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;
  UNREFERENCED_PARAMETER(timeout); /* Stop compiler warnings on unused parameter */

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_open called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Set default state on open */
  lcd_definition->lcd_access_segs_not_chars = TRUE;
  lcd_definition->lcd_current_access = lcd_definition->lcd_segments;

  return RIO_SUCCESS;
}


/** Close the generic LCD device, and force an update */
static int lcd_generic_segflash_close(struct t_io_dev *io_dev)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_close called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* The LCD close call forces an update */
  return lcd_generic_segflash_update(lcd_definition);
}


/*
 * Macro to work out smaller of two numbers
 */
#ifndef min
#define min(x, y) ((x < y) ? x : y)
#endif


/** Read from the LCD data segment buffer */
static int lcd_generic_segflash_segs_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;
  const unsigned char *rd_ptr;
  unsigned short count;

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* More pointer checks */
  if (NULL == buf || NULL == lcd_definition->lcd_current_access)
  {
    return RIO_ERROR;
  }
  rd_ptr = &lcd_definition->lcd_current_access[io_dev->posn];

  /* Work out how many bytes to read, with limit */
  if (io_dev->posn >= lcd_definition->lcd_num_segment_bytes)
  {
    return RIO_ERROR;
  }
  count = min(len, (unsigned short) (lcd_definition->lcd_num_segment_bytes - io_dev->posn));

  /* Copy data to LCD buffer, and update position */
  memcpy(buf, rd_ptr, (size_t) count);
  io_dev->posn += count;

  /* Return number of bytes copied */
  return (int) count;
}


/** LCD device read */
static int lcd_generic_segflash_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_read called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Read data from the buffer we are accessing */
  if (lcd_definition->lcd_access_segs_not_chars)
  {
    return lcd_generic_segflash_segs_read(io_dev, buf, len, timeout);
  }
  else
  {
    return lcd_generic_char_read(io_dev, buf, len, timeout);
  }
}


/**
 * Write to the LCD data segment buffer
 *
 * NOTE: This doesn't force an update
 */
static int lcd_generic_segflash_segs_write(struct t_io_dev *io_dev, const void *buf, unsigned short len)
{
  struct s_lcd_definition *lcd_definition;
  unsigned char *wr_ptr;
  unsigned short count;

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* More pointer checks */
  if (NULL == buf || NULL == lcd_definition->lcd_current_access)
  {
    return RIO_ERROR;
  }
  wr_ptr = &lcd_definition->lcd_current_access[io_dev->posn];

  /* Work out how many bytes to write, with limit */
  if (io_dev->posn >= lcd_definition->lcd_num_segment_bytes)
  {
    return RIO_ERROR;
  }
  count = min(len, (unsigned short) (lcd_definition->lcd_num_segment_bytes - io_dev->posn));

  /*
   * Copy data to LCD buffer, and update position
   * Don't update LCD, since this is done by lcd_close
   */
  memcpy(wr_ptr, buf, (size_t) count);
  io_dev->posn += count;

  /* Return number of bytes copied */
  return (int) count;
}


/** LCD device write */
static int lcd_generic_segflash_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_lcd_definition *lcd_definition;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_write called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Write data to the buffer we are accessing */
  if (lcd_definition->lcd_access_segs_not_chars)
  {
    return lcd_generic_segflash_segs_write(io_dev, buf, len);
  }
  else
  {
    return lcd_generic_char_write(io_dev, buf, len);
  }
}


/** LCD device specific configuration */
static int lcd_generic_segflash_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_lcd_definition *lcd_definition;
  unsigned char level;
  int result = RIO_SUCCESS;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_generic_segflash_ioctl called\n");
#endif

  /* Check pointers */
  lcd_generic_pointer_check();
  lcd_definition = io_dev->dev_data;

  /* Function dispatch */
  switch (func)
  {

    /* Set the backlight level */
    case IOCTL_LCD_SET_BACKLIGHT:
      level = (unsigned char) va_arg(*arglist, int);
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("IOCTL_LCD_SET_BACKLIGHT: level %d\n", level);
#endif
      if (NULL != lcd_definition->lcd_hw_setbacklight_fn)
      {
        lcd_definition->lcd_hw_setbacklight_fn(lcd_definition, level);
      }
      break;

    /* Change the read and write buffer access */
    case IOCTL_LCD_ACCESS_DATA:
      lcd_definition->lcd_access_segs_not_chars = TRUE;
      lcd_definition->lcd_current_access = lcd_definition->lcd_segments;
      io_dev->posn = 0UL;
      break;

    case IOCTL_LCD_ACCESS_FLASH:
      lcd_definition->lcd_access_segs_not_chars = TRUE;
      lcd_definition->lcd_current_access = lcd_definition->lcd_flash_segments;
      io_dev->posn = 0UL;
      break;
    case IOCTL_LCD_ACCESS_OUTPUT:
      lcd_definition->lcd_access_segs_not_chars = TRUE;
      lcd_definition->lcd_current_access = lcd_definition->lcd_output_segments;
      io_dev->posn = 0UL;
      break;

    case IOCTL_LCD_ACCESS_CHAR:
      lcd_definition->lcd_access_segs_not_chars = FALSE;
      io_dev->posn = 0UL;
      break;

    /* Error */
    default:
      result = RIO_ERROR;
      break;
  }

  /*
   * Done, return result
   */
  return result;
}


/**
 * Create a generic LCD device with segment flashing (requires an extra task)
 *
 * @param io_dev is the uninitialised IO device to be filled out
 * @param lcd_hw_data is a pointer to your LCD hardware implementation data
 * @param lcd_num_segment_bytes is the number of segment bytes in the LCD
 * @param lcd_num_chars is the number of characters on the LCD
 * @param lcd_hw_init is a function pointer called to initialise the LCD. Can be NULL, in which case it won't be called
 * @param lcd_hw_destroy is a function pointer called to destroy the LCD. Can be NULL, in which case it won't be called
 * @param lcd_hw_write is a function pointer called to write data to the LCD. Can be NULL, in which case it won't be called
 * @param lcd_put_char is a function pointer called to put a character on the LCD. Not NULL.
 * @param lcd_does_digit_have_dp is a function pointer called to determine if a character position has a decimal point. Not NULL
 * @param lcd_hw_setbacklight_fn is a function pointer called to set the back light level 0..100%. Can be NULL, in which case it won't be called
 *
 * @return RIO_SUCCESS on success, RIO_ERROR otherwise
 */
int lcd_generic_segflash_create_device(struct t_io_dev *io_dev,
                                       void *lcd_hw_data,
                                       size_t lcd_num_segment_bytes,
                                       size_t lcd_num_characters,
                                       t_lcd_hw_init_fn lcd_hw_init,
                                       t_lcd_hw_destroy_fn lcd_hw_destroy,
                                       t_lcd_hw_write_fn lcd_hw_write,
                                       t_lcd_put_char_fn lcd_put_char,
                                       t_lcd_does_digit_have_dp_fn lcd_does_digit_have_dp,
                                       t_lcd_hw_setbacklight_fn lcd_hw_setbacklight_fn
                                      )
{
  struct s_lcd_definition *lcd_definition = NULL;

  /* Check pointers */
  if (NULL == io_dev || NULL == lcd_put_char || NULL == lcd_does_digit_have_dp)
  {
    goto lgcd_fail;
  }

  /* Create LCD definition */
  lcd_definition = malloc(sizeof(struct s_lcd_definition));
  if (NULL == lcd_definition)
  {
    goto lgcd_fail;
  }
  memset(lcd_definition, 0, sizeof(struct s_lcd_definition));

  lcd_definition->hw_data = lcd_hw_data;

  /* Allocate segments */
  lcd_definition->lcd_num_segment_bytes = lcd_num_segment_bytes;
  lcd_definition->lcd_segments = malloc(lcd_definition->lcd_num_segment_bytes);
  if (NULL == lcd_definition->lcd_segments)
  {
    goto lgcd_fail;
  }
  lcd_definition->lcd_output_segments = malloc(lcd_definition->lcd_num_segment_bytes);
  lcd_definition->lcd_flash_segments = malloc(lcd_definition->lcd_num_segment_bytes);
  if (NULL == lcd_definition->lcd_output_segments || NULL == lcd_definition->lcd_flash_segments)
  {
    goto lgcd_fail;
  }
  lcd_definition->lcd_current_access = lcd_definition->lcd_segments;
  /* lcd_definition->sem_lcd_update = 0; */
  /* lcd_definition->lcd_flash_thread = 0; */
  lcd_definition->lcd_thread_quit = 0;

  /* Allocate characters */
  lcd_definition->lcd_num_characters = lcd_num_characters;
  lcd_definition->lcd_characters = malloc(lcd_definition->lcd_num_characters);
  if (NULL == lcd_definition->lcd_characters)
  {
    goto lgcd_fail;
  }

  /* Functions */
  lcd_definition->lcd_hw_init = lcd_hw_init;
  lcd_definition->lcd_hw_destroy = lcd_hw_destroy;
  lcd_definition->lcd_hw_write = lcd_hw_write;
  lcd_definition->lcd_put_char = lcd_put_char;
  lcd_definition->lcd_does_digit_have_dp = lcd_does_digit_have_dp;
  lcd_definition->lcd_hw_setbacklight_fn = lcd_hw_setbacklight_fn;

  /* Other members */
  lcd_definition->lcd_access_segs_not_chars = true;

  /*
   * Setup device data pointer to point to the LCD definition
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) lcd_definition;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = lcd_generic_segflash_init;
  io_dev->dev_open = lcd_generic_segflash_open;
  io_dev->dev_close = lcd_generic_segflash_close;
  io_dev->dev_read = lcd_generic_segflash_read;
  io_dev->dev_write = lcd_generic_segflash_write;
  io_dev->dev_filelength = lcd_generic_common_filelength;
  io_dev->dev_ioctl = lcd_generic_segflash_ioctl;
  io_dev->dev_destroy = lcd_generic_segflash_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
    goto lgcd_fail;
  }

  /* Success */
  return RIO_SUCCESS;

  /* Clean up after fail */
lgcd_fail:
  io_dev->dev_data = NULL;
  if (NULL != lcd_definition)
  {
    if (NULL != lcd_definition->lcd_segments)
    {
      free(lcd_definition->lcd_segments);
      lcd_definition->lcd_segments = NULL;
    }
    if (NULL != lcd_definition->lcd_characters)
    {
      free(lcd_definition->lcd_characters);
      lcd_definition->lcd_characters = NULL;
    }
    if (NULL != lcd_definition->lcd_output_segments)
    {
      free(lcd_definition->lcd_output_segments);
      lcd_definition->lcd_output_segments = NULL;
    }
    if (NULL != lcd_definition->lcd_flash_segments)
    {
      free(lcd_definition->lcd_flash_segments);
      lcd_definition->lcd_flash_segments = NULL;
    }
    lcd_definition->lcd_current_access = NULL;
    lcd_definition->hw_data = NULL;

    free(lcd_definition);
    lcd_definition = NULL;
  }
  return RIO_ERROR;
}
