/*
 * lcd_local_skt.c
 *
 *  Created on: 15/02/2009
 *      Author: ianb
 */
#if defined( WIN32 )
  #include <winsock.h>
#elif defined(__linux__)
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <strings.h>
  #include <string.h>
  #include <unistd.h>
  #include <fcntl.h>
  #include <errno.h>
#endif
#include <pthread.h>
#include <stdio.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rstdio.h>
#include <rinhw/debug.h>
#include <rinhw/kbd.h>
#include <rinhw/lcd_generic.h>
#include <rinhw/rinthread.h>

#include <rinhw/lcd_local_skt.h>


/* Why couldn't Windows do things like everyone else? */
#if defined( WIN32 )
  #define sockclose closesocket
#elif defined(__linux__)
  #define sockclose close
#endif


/* code to establish a socket */
static int establish(unsigned short portnum)
{
  int sockfd = -1;
  struct sockaddr_in servaddr;

  /* Setup socket */
  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0)
  {
    ERROR_PRINTF("lcd_local_skt establish: socket() error for port %d (errno %d)\n", portnum, errno);
    goto e_fail;
  }
  if (fcntl(sockfd, F_SETFD, FD_CLOEXEC) < 0)
  {
    ERROR_PRINTF("lcd_local_skt establish: error setting FD_CLOEXEC for port %d (errno %d)\n", portnum, errno);
    goto e_fail;
  }

  memset(&servaddr, 0, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servaddr.sin_port = htons(portnum);
  if (bind(sockfd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0)
  {
    ERROR_PRINTF("lcd_local_skt establish: bind() error for port %d (errno %d)\n", portnum, errno);
    goto e_fail;
  }
  if (listen(sockfd, 1) < 0)
  {
    ERROR_PRINTF("lcd_local_skt establish: listen() error for port %d (errno %d)\n", portnum, errno);
    goto e_fail;
  }
  return sockfd;

e_fail:
  if (sockfd >= 0)
    sockclose(sockfd);
  return -1;
}


static bool PretendViewerGetChar(int sockfd, unsigned char *message, size_t maxlen, size_t *msgposn)
{
  char ch;
  int flags;
  bool result;

/* TODO check if MSG_WAITALL is actually required. It shouldn't be, since recv() is setup
 * to block until more data is available, so the minimum amount that can be available is 1 byte
 * therefore it will block until all (1 bytes) of data have arrived.
 */
#ifdef MSG_WAITALL
  flags = MSG_WAITALL;
#else
  flags = 0;
#endif

  /* Get character */
  if (recv(sockfd, &ch, 1, flags) != 1)
  {
    result = false;
  }
  else
  {
    if (*msgposn >= maxlen)
      *msgposn = 0;
    message[*msgposn] = ch;
    (*msgposn)++;
    result = true;
  }
  return result;
}


static void PretendViewerWriteKeypress(struct s_lcd_hw_socket *sock_hw, union un_key key)
{
  /* Because the keyboard has it's own locking internally, just call the device write, rather than ropen/rwrite/rclose */
  if (NULL != sock_hw->kbddev && NULL != sock_hw->kbddev->dev_write)
  {
    sock_hw->kbddev->dev_write(sock_hw->kbddev, &key, sizeof(key), 0);
  }

  /*
  ropen( &dev_kbd, 0 );
  rwrite( &dev_kbd, &ui8Key, sizeof( ui8Key ), 0 );
  rclose( &dev_kbd );
  */
}


static bool lcd_socket_listen_parser(int sockfd, struct s_lcd_definition *lcd_definition, unsigned char *buffer, size_t buffer_maxlen, size_t *buffer_posn)
{
  struct s_lcd_hw_socket *sock_hw = lcd_definition->hw_data;
  union un_key key;
  bool gotkey;
  size_t lcddatalen;
  const unsigned char *lcddata;
  unsigned char err;

  /* Get data from socket */
  if (!PretendViewerGetChar(sockfd, buffer, buffer_maxlen, buffer_posn))
    return false;

  /* Parse it */
  if (NULL != sock_hw->data_parser)
  {

    /* Wait for the LCD buffer to become free */
    if (NULL != sock_hw->sem_exclusive)
    {
      OSSemPend(sock_hw->sem_exclusive, 0, &err);
    }

    lcddatalen = lcd_definition->lcd_num_segment_bytes;
    if (NULL != lcd_definition->lcd_output_segments)
      lcddata = lcd_definition->lcd_output_segments;
    else
      lcddata = lcd_definition->lcd_segments;

    gotkey = false;
    if (sock_hw->data_parser(sockfd, buffer, *buffer_posn, lcddata, lcddatalen, &gotkey, &key))
    {
      *buffer_posn = 0;

      /* Key press has been returned */
      if (gotkey)
      {
        PretendViewerWriteKeypress(sock_hw, key);
      }
    }

    /* Let other tasks send new data to the LCD */
    if (NULL != sock_hw->sem_exclusive)
    {
      OSSemPost(sock_hw->sem_exclusive);
    }

  }
  return true;
}


/*
 * TaskLcdUpdate -
 *
 * Sends lcd data to a socket on change
 */
static void *lcd_socket_listen_task(void *data)
{
  struct s_lcd_definition *lcd_definition = data;
  struct s_lcd_hw_socket *sock_hw = lcd_definition->hw_data;
  int listen_sockfd;
  int connection_sockfd;
  fd_set readfds;
  int max_fd;
  int retval;
  #define LCD_LISTEN_BUFFER_LENGTH 200
  unsigned char buffer[LCD_LISTEN_BUFFER_LENGTH];
  size_t buffer_posn;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_socket_listen_task() started, listen on port %d\n", sock_hw->port);
#endif

  listen_sockfd = establish(sock_hw->port);
  if (listen_sockfd < 0)
  {
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("lcd_socket_listen_task: could not establish a listening socket on port %d\n", sock_hw->port);
#endif
    rin_thread_exit(NULL);
  }

  /* Task loop */
  connection_sockfd = -1;
  buffer_posn = 0;
  while (1)
  {

    /* Setup file descriptors to pend on */
    FD_ZERO(&readfds);
#if defined(__linux__)
    FD_SET(sock_hw->shutdown_pipe_read_fd, &readfds);
    /* TODO create pipe() equivalent to shutdown task under windows, pipe() not available */
#endif
    if (connection_sockfd < 0)
    {
      /* No existing connection, so wait for one on listening socket */
      FD_SET(listen_sockfd, &readfds);
#if defined(__linux__)
    max_fd = (listen_sockfd > sock_hw->shutdown_pipe_read_fd) ? listen_sockfd : sock_hw->shutdown_pipe_read_fd;
#elif defined(WIN32) /* #if defined(__linux__) */
    max_fd = listen_sockfd;
#endif /* #else -> #elif defined( WIN32 ) -> #if defined(__linux__) */
    }
    else
    {
      /* Existing connection, so wait for data on it */
      FD_SET(connection_sockfd, &readfds);
#if defined(__linux__)
      max_fd = (connection_sockfd > sock_hw->shutdown_pipe_read_fd) ? connection_sockfd : sock_hw->shutdown_pipe_read_fd;
#elif defined(WIN32) /* #if defined(__linux__) */
      max_fd = connection_sockfd;
#endif /* #else -> #elif defined( WIN32 ) -> #if defined(__linux__) */
    }

    /* Wait for connection or data on pipe (which means quit) */
    retval = select(max_fd + 1, &readfds, NULL, NULL, NULL);
    if (retval <= 0)
    {
      /* Timed out (=0) or more serious error (<0) */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("lcd_socket_listen_task: select() error\n");
#endif
    }
    else if (FD_ISSET(listen_sockfd, &readfds) && connection_sockfd < 0)
    {
      /* All good, so accept connection */
      connection_sockfd = accept(listen_sockfd, NULL, NULL);
      if (connection_sockfd < 0)
      {
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("lcd_socket_listen_task: error accepting for new connection\n", sock_hw->port);
#endif
      }
      if (fcntl(connection_sockfd, F_SETFD, FD_CLOEXEC) < 0)
      {
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("lcd_socket_listen_task: error setting FD_CLOEXEC for new connection (errno %d)\n", errno);
#endif
      }

      buffer_posn = 0;
    }
    else if (FD_ISSET(connection_sockfd, &readfds) && connection_sockfd >= 0)
    {
      /* Incoming data on connection socket */
      if (!lcd_socket_listen_parser(connection_sockfd, lcd_definition, buffer, LCD_LISTEN_BUFFER_LENGTH, &buffer_posn))
      {
        sockclose(connection_sockfd);
        connection_sockfd = -1;
        buffer_posn = 0;
      }
    }
#if defined(__linux__)
    else if (FD_ISSET(sock_hw->shutdown_pipe_read_fd, &readfds))
    {
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("lcd_socket_listen_task: quit notification\n");
#endif
      break;
    }
#endif /* #if defined(__linux__) */
  }

  /* Close connection socket if open */
  if (connection_sockfd >= 0)
  {
    sockclose(connection_sockfd);
  }

  /* Close listening socket */
  sockclose(listen_sockfd);

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("lcd_socket_listen_task: finished\n");
#endif
  rin_thread_exit(NULL);
}


/** LCD hardware is a local socket */
int lcd_socket_hardware_write(struct s_lcd_definition *lcd_definition)
{
  /* Nothing to do here */
  return RIO_SUCCESS;
}


/** LCD hardware is a local socket */
int lcd_socket_hardware_init(struct s_lcd_definition *lcd_definition)
{
  struct s_lcd_hw_socket *sock_hw = lcd_definition->hw_data;
  bool res;
  int pipe_fds[2];

  /* Create socket task */
#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("lcd_socket_hardware_init: Creating Lcd Update Task.\n");
#endif /* #if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL) */

#if defined(__linux__)
  /* Create a pipe to write to, to shut down the task */
  if (pipe(pipe_fds) < 0)
  {
    ERROR_PRINTF("lcd_socket_hardware_init: create pipe failed\n");
    return RIO_ERROR;
  }
#elif defined(WIN32) /* #if defined(__linux__) */
  /* TODO create pipe() equivalent to shutdown task under windows, pipe() not available */
#else /* #elif defined( WIN32 ) -> #if defined(__linux__) */
  #error "lcd_socket_hardware_init() not defined for this platform"
#endif /* #else -> #elif defined( WIN32 ) -> #if defined(__linux__) */

  /* Keep fds for shutting down task */
  sock_hw->shutdown_pipe_read_fd = pipe_fds[0];
  sock_hw->shutdown_pipe_write_fd = pipe_fds[1];
  if (fcntl(sock_hw->shutdown_pipe_read_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    ERROR_PRINTF("lcd_socket_hardware_init: error setting FD_CLOEXEC on read end of shutdown pipe (errno %d)\n", errno);
  }
  if (fcntl(sock_hw->shutdown_pipe_write_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    ERROR_PRINTF("lcd_socket_hardware_init: error setting FD_CLOEXEC on write end of shutdown pipe (errno %d)\n", errno);
  }

  /* Create thread */
  res = rin_thread_create(&sock_hw->thread, lcd_socket_listen_task, lcd_definition, true);

  return (res) ? RIO_SUCCESS : RIO_ERROR;
}


/** LCD hardware is a local socket */
void lcd_socket_hardware_destroy(struct s_lcd_definition *lcd_definition)
{
#if defined(__linux__)
  struct s_lcd_hw_socket *sock_hw = lcd_definition->hw_data;

  /* Write (any) data to the pipe to force select() to return */
  write(sock_hw->shutdown_pipe_write_fd, "shutdown", 8);

  /* Wait for task to quit */
  pthread_join(sock_hw->thread, NULL);

  /* Close the pipes */
  close(sock_hw->shutdown_pipe_read_fd);
  close(sock_hw->shutdown_pipe_write_fd);
#elif defined(WIN32) /* #if defined(__linux__) */
  /* TODO shutdown task under windows, pipe() not available */
#else /* #elif defined( WIN32 ) -> #if defined(__linux__) */
  #error "lcd_socket_hardware_destroy() not defined for this platform"
#endif /* #else -> #elif defined( WIN32 ) -> #if defined(__linux__) */
}
