/*
 * lcd_r400.c
 *
 *  Created on: 01/06/2010
 *      Author: ian
 */

#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rio.h>
#include <rinhw/lcd.h>
#include <rinhw/debug.h>
#include <rinhw/lcd_generic.h>
#include <rinhw/rinendian.h>
#include <rinhw/lcd_local_skt.h>
#include "_lcdtbl.h"
#include "_lcdxlat.h"
#include <rinhw/kbd.h>

#include <rinhw/lcd_local_skt_r400.h>


#if ( BYTE_ORDER == BIG_ENDIAN )
/*
 * ByteSwapLong -
 *
 * Swap the endian on a 32bit number
 */
static unsigned long ByteSwapLong(unsigned long l)
{
    union { unsigned long l; unsigned char b[4]; } u;
    unsigned char t;

    u.l = l;

    t = u.b[0];
    u.b[0] = u.b[3];
    u.b[3] = t;

    t = u.b[1];
    u.b[1] = u.b[2];
    u.b[2] = t;

    return u.l;
}
#endif /* #if ( BYTE_ORDER == BIG_ENDIAN ) */


/*
 * SegXlat -
 *
 * Perform the necessary segment translation on a digit
 * This is necessary because different digits on the LCD have different
 * mappings. This also re-arranges the segment mask in the same
 * way.
 */
static unsigned long SegXlat(const signed short *xlat_tbl, unsigned short segment, unsigned long *segmask)
{
    unsigned short mask = 0x0001;
    unsigned long temp, result = 0L;
    unsigned long sm_temp = 0L;
    signed short shifts;
    unsigned char i;

    /*
     * For each segment in a digit
     */
    for (i = 0; i < SEGS_PER_DIGIT; i++, mask <<= 1, xlat_tbl++)
    {

        /*
         * If this segment is set, perform the translation
         */
        if ((temp = segment & mask) != 0)
        {
            shifts = *xlat_tbl;
            if (shifts > 0)
            {
                temp <<= shifts;
            }
            else if (shifts < 0)
            {
                temp >>= -shifts;
            }
            result |= temp;
/*             DEBUG_PRINTF("XLAT: i=%d, mask=%x, shifts=%d, seg=%x, temp=%lx\n", i, mask, shifts, segment, temp); */
        }

        /*
         * If this segment of the segment mask is set, perform the translation
         */
        if ((temp = *segmask & mask) != 0)
        {
            shifts = *xlat_tbl;
            if (shifts > 0)
            {
                temp <<= shifts;
            }
            else if (shifts < 0)
            {
                temp >>= -shifts;
            }
            sm_temp |= temp;
/*             DEBUG_PRINTF("XLAT: i=%d, mask=%x, shifts=%d, seg=%x, temp=%lx\n", i, mask, shifts, segment, temp); */
        }
    }

    /*
     * Set the segmask and return the translated result
     */
    *segmask = sm_temp;
    return result;
}


/*
 * CharToBitmap -
 *
 * Convert the character to be written into a bitmap (segs)
 * that can be written to LCD data buffer. seg_mask is also
 * set with the appropriate mask for the character. Index is the
 * character number to be written.
 */
static unsigned short CharToBitmap(unsigned char index, unsigned char ch, unsigned long *segs, unsigned long *seg_mask)
{
    const tLcdDigit *digit;
    unsigned short bitnum, devpos;
    unsigned char shifts;

    /*
     * Wrap the index at the last digit
     */
    if (index >= NUM_LCD_DIGITS)
        index = 0;

    /*
     * Get the bitmap for the character from the segment table
     */
    *segs = (unsigned long) seg_table[ch];
#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("CharToBitmap: untranslated digit segments 0x%08lx\n", *segs);
#endif

    /*
     * Work out what translation is required from the translation table
     * and do the translation
     */
    digit = &lcd_digits[index];
    *seg_mask = digit->mask;
#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("CharToBitmap: untranslated digit mask 0x%08lx\n", *seg_mask);
#endif
    *segs = SegXlat(digit->xlat_tbl, (unsigned short) *segs, seg_mask);

#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("CharToBitmap: translated digit mask 0x%08lx\n", *seg_mask);
    DEBUG_PRINTF("CharToBitmap: translated digit segments 0x%08lx\n", *segs);
#endif

    /*
     * Work out the byte and bit number of the lowest order segment in the digit
     * then calculate the number of shifts required to get the digit into place
     */
    devpos = digit->bit_num >> 3;
    bitnum = digit->bit_num & 0x07;
    shifts = (unsigned char) (bitnum);
#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("CharToBitmap: devpos %d, shifts %d\n", devpos, shifts);
#endif

    /*
     * Shift the new segments, and the segment mask into place
     */
    *segs <<= shifts;
    *seg_mask <<= shifts;
#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("CharToBitmap: translated and shifted digit mask 0x%08lx\n", *seg_mask);
    DEBUG_PRINTF("CharToBitmap: translated and shifted digit segments 0x%08lx\n", *segs);
#endif

    return devpos;
}


/*
 * rb_we21_lcd_put_char -
 *
 * Write a character ch to digit position index on the LCD.
 * cwm is the write mode - either set or or the data
 */
static void r400_lcd_put_char(struct s_lcd_definition *lcd_definition, unsigned char index, unsigned char ch, enum eCharWriteMode cwm)
{
  unsigned long segments, seg_mask, olddata;
  unsigned short devpos;

#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("r400_lcd_put_char called to write char 0x%02x to lcd digit %d\n", ch, index);
#endif

  /* Is the character in the table, if not display it with all segments on */
  if (ch >= NUM_SEG_TABLE)
  {
    ch = 11;
#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("Char modified to 0x%02x because outside table range (max 0x%02x)\n", ch, NUM_SEG_TABLE);
#endif
  }

  /* Convert the segments into a bitmap */
  devpos = CharToBitmap(index, ch, &segments, &seg_mask);
#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("Character 0x%02x translates to bitmap 0x%08lx, with segmask 0x%08lx, at byte location %d\n", ch, segments, seg_mask, devpos);
#endif

  /*
   * Read the current data out of the LCD buffer
   */
  olddata = *((unsigned long *) &lcd_definition->lcd_segments[devpos]);

  /*
   * Modify the data, as dictated by the write mode
   */
#if ( BYTE_ORDER == BIG_ENDIAN )
  olddata = ByteSwapLong(olddata);
#endif /* #if ( BYTE_ORDER == BIG_ENDIAN ) */

#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("Old data in lcd buf = %08lx\n", olddata);
#endif
  if (cwm == cwmSet)
      olddata &= ~seg_mask;
  olddata |= (segments & seg_mask);
#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("Old data in lcd buf = %08lx\n", olddata);
#endif

#if ( BYTE_ORDER == BIG_ENDIAN )
  olddata = ByteSwapLong(olddata);
#endif /* #if ( BYTE_ORDER == BIG_ENDIAN ) */


#if HW_DEBUG(LCD_CHAR_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("Character 0x%02x translates to bitmap 0x%02x\n", ch, segments);
#endif

  /*
   * Seek the start index of this character in the LCD,
   * and write the modified data back
   */
  *((unsigned long *) &lcd_definition->lcd_segments[devpos]) = olddata;
}


/*
 * rb_does_digit_have_dp -
 *
 * Work out if the digit index has a decimal point
 */
static bool r400_lcd_does_digit_have_dp(struct s_lcd_definition *lcd_definition, unsigned char index)
{
  /* Make sure its in range */
  if (index >= lcd_definition->lcd_num_characters)
    return false;

  /* Digits table contains information on which digits have a decimal point */
  return (bool) lcd_digits[index].has_dp;
}


static void r400_lcd_local_skt_parser_send_reply(int sockfd, const char *reply)
{
  char ch;
  const int len = strlen(reply);

  ch = '\1';
  send(sockfd, &ch, 1, 0);
  send(sockfd, reply, len, 0);
  ch = '\4';
  send(sockfd, &ch, 1, 0);
}


/**
 * Calculate a 16bit CCITT CRC checksum for a block of data
 *
 * @param byt is an array of bytes to calculate the checksum for
 * @param size is the number of bytes to calculate the checksum over
 * @param crc is the checksum seed
 * @return the checksum of the given data
 */
static int16u CalcCRC16( const int8u *byt, int16u size, int16u crc )
{
  /*
   * Calculate CRC-16 value; uses The CCITT-16 Polynomial,
   * expressed as X^16 + X^12 + X^5 + 1
   */
  int16u index;
  int8u b;

  for(index=0; index<size; ++index )
  {
    crc ^= (((int16u) byt[index]) << 8);
    for( b=0; b<8; ++b )
    {
      if( crc & (int16u) 0x8000 )
        crc = (crc << 1) ^ (int16u) 0x1021;
      else
        crc = (crc << 1);
    }
  }
  return crc;
}


static void r400_lcd_local_skt_parser_send_lcddata(int outsockfd, const unsigned char *lcd_data, size_t lcd_data_length)
{
  char message[400];
  char str[ 10 ];
  size_t i;
  int16u checksum;

  /* Assemble LCD data into a string */
  strncpy(message, "81110009:00", 400); /* 00 on the end is the magic "id" byte in the blob reply */
  for (i = 0; i < lcd_data_length; i++)
  {
    snprintf(str, 10, "%02X", lcd_data[i]);
    strncat(message, str, 400);
  }

  /* Generate checksum */
  checksum = CalcCRC16((int8u *) message, strlen(message), 0xFFFF);
  snprintf(str, 10, "%02X%02X", checksum >> 8, checksum & 0x00FF);
  strncat(message, str, 400);

  /* Send message to to socket */
  r400_lcd_local_skt_parser_send_reply(outsockfd, message);
}


static bool r400_lcd_local_skt_parser_get_message(const unsigned char *sock_data, size_t sock_data_length, char *outbuf, size_t outbuflen)
{
  size_t i;
  size_t ocnt;

  if (0 == sock_data_length || 0 == outbuflen)
    return false;

  /* Discard everything upto '\1' */
  for (i = 0; '\1' != sock_data[i] && i < sock_data_length; i++)
    ;
  if ('\1' != sock_data[i])
    return false;

  /* Copy everything upto '\4' into outbuf */
  for (ocnt = 0; '\4' != sock_data[i] && i < sock_data_length && ocnt < outbuflen; i++, ocnt++)
  {
    outbuf[ocnt] = sock_data[i];
  }
  if ('\4' != sock_data[i])
    return false;

  /* Finished */
  outbuf[ocnt] = '\0';
  return true;
}


static bool r400_lcd_local_skt_parser_disassemble_message(const char *message, unsigned char *pui8Addr, unsigned char *pui8Cmd, unsigned short *pui16Reg, char *pszData, int maxlen)
{
  int res;
  unsigned int addr, cmd, reg;
  BOOLEAN success;

  if (0 != pszData)
  {
    /* TODO maxlen */
    res = sscanf(message, "%02x%02x%04x:%s%*04x", &addr, &cmd, &reg, pszData);
    success = (4 == res);
  }
  else
  {
    res = sscanf(message, "%02x%02x%04x:", &addr, &cmd, &reg);
    success = (3 == res);
  }
  if (success)
  {
    *pui8Addr = addr;
    *pui8Cmd = cmd;
    *pui16Reg = reg;
  }
  return success;
}


static void r400_lcd_local_skt_parser_parse_keypress(const char *data, bool *gotkey, union un_key *key)
{
  int intkey;

  if (1 != sscanf(data, "%04x", &intkey))
  {
    *gotkey = false;
  }
  else
  {
    *gotkey = true;
    key->BYTE = intkey;
  }
}


static void r400_lcd_local_skt_parser_parse_message(int outsockfd, const char *message, const unsigned char *lcd_data, size_t lcd_data_length, bool *gotkey, union un_key *key)
{
  unsigned char ui8Addr, ui8Cmd;
  unsigned short ui16Reg;
  char data[200];

  if (r400_lcd_local_skt_parser_disassemble_message(message, &ui8Addr, &ui8Cmd, &ui16Reg, data, 200))
  {
    if (0x11 == ui8Cmd && 0x0005 == ui16Reg)
    {
      r400_lcd_local_skt_parser_send_reply(outsockfd, "81110005:003066C5AB2B");
    }
    else if (0x11 == ui8Cmd && 0x0003 == ui16Reg)
    {
      r400_lcd_local_skt_parser_send_reply(outsockfd, "81110003:K4812408");
    }
    else if (0x11 == ui8Cmd && 0x0004 == ui16Reg)
    {
      r400_lcd_local_skt_parser_send_reply(outsockfd, "81110004:V1.131B34");
    }
    else if (0x11 == ui8Cmd && 0x0001 == ui16Reg)
    {
      r400_lcd_local_skt_parser_send_reply(outsockfd, "81110001:004426");
    }
    else if (0x11 == ui8Cmd && 0x0009 == ui16Reg)
    {
      /* r400_lcd_local_skt_parser_send_reply(sockfd, "81110009:0099000039B3431E23C33C303C33E000000000000002000000C008AC0000000000000000000000000C3C88A02000000000000000004574"); */
      r400_lcd_local_skt_parser_send_lcddata(outsockfd, lcd_data, lcd_data_length);
    }
    else if (0x12 == ui8Cmd && 0x0008 == ui16Reg)
    {
      r400_lcd_local_skt_parser_parse_keypress(data, gotkey, key);
      r400_lcd_local_skt_parser_send_reply(outsockfd, "81120008:00004F47");
    }
  }
}


/** Parse the data off the socket, and reply as required */
bool r400_lcd_local_skt_parser(int outsockfd, const unsigned char *sock_data, size_t sock_data_length, const unsigned char *lcd_data, size_t lcd_data_length, bool *gotkey, union un_key *key)
{
  #define R400_LCD_BUFFER_LEN 200
  char buffer[R400_LCD_BUFFER_LEN];

  /* See if there is a message in the data */
  if (!r400_lcd_local_skt_parser_get_message(sock_data, sock_data_length, buffer, R400_LCD_BUFFER_LEN))
    return false;

  /* Got a message, parse it */
  r400_lcd_local_skt_parser_parse_message(outsockfd, buffer, lcd_data, lcd_data_length, gotkey, key);
  return true;
}


/** Function to create and initialise a local socket R400 LCD device (use the viewer to attach to) */
int rio_create_device_lcd_localskt_r400(struct t_io_dev *io_dev, int sockport, struct t_io_dev *kbddev)
{
  struct s_lcd_hw_socket *r400_lcd_socket;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("rio_create_device_lcd_localskt_r400 called\n");
#endif

  /* Setup socket */
  r400_lcd_socket = malloc(sizeof(struct s_lcd_hw_socket));
  if (NULL == r400_lcd_socket)
  {
    goto rcdllr_fail;
  }
  r400_lcd_socket->port = sockport;
  r400_lcd_socket->shutdown_pipe_write_fd = -1;
  r400_lcd_socket->shutdown_pipe_read_fd = -1;
  /* r400_lcd_socket->thread = 0; */
  r400_lcd_socket->sem_exclusive = NULL;
  r400_lcd_socket->kbddev = kbddev;
  r400_lcd_socket->data_parser = r400_lcd_local_skt_parser;

  /* Create device */
  if (lcd_generic_segflash_create_device(io_dev,
                                         (void *) r400_lcd_socket,
                                         sizeof(tLcdSegments),
                                         NUM_LCD_DIGITS,
                                         lcd_socket_hardware_init,
                                         lcd_socket_hardware_destroy,
                                         NULL,
                                         r400_lcd_put_char,
                                         r400_lcd_does_digit_have_dp,
                                         NULL) < 0)
  {
    goto rcdllr_fail;
  }
  r400_lcd_socket->sem_exclusive = io_dev->sem_open_excl;

#if HW_DEBUG(LCD_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("rio_create_device_lcd_localskt_r400: lcd created successfully\n");
#endif

  /* Success */
  return RIO_SUCCESS;

  /* Clean up after fail */
rcdllr_fail:
  if (NULL != r400_lcd_socket)
  {
    free(r400_lcd_socket);
    r400_lcd_socket = NULL;
  }
  return RIO_ERROR;
}
