/*
 * adc_ads1216_local.h
 */

#ifndef ADC_ADS1216_LOCAL_H_
#define ADC_ADS1216_LOCAL_H_ 1

#include <rinhw/rio.h>

/* Forward declaration */
struct sAdcSpiDevice;


/** User specific hardware initialisation called as part of the ADC hardware initialisation. May be called multiple times */
typedef int (*t_adc24bit_user_hw_init)(struct sAdcSpiDevice *adcdev);


/** Function pointer for device specific ADC SPI syncronisation, only required when a chip select is not used */
typedef int (*t_adc24bit_sync_spi)(struct sAdcSpiDevice *adcdev);


/** Function pointer to set 4/6 wire mode, level is IOCTL_ADC24BIT_LCWIRE_4WIRE or IOCTL_ADC24BIT_LCWIRE_6WIRE */
typedef int (*t_adc24bit_set_loadcell_wire_mode)(struct sAdcSpiDevice *adcdev, int level);


/** Function pointer to wait for a reading to become available */
typedef int (*t_adc24bit_wait_for_reading)(struct sAdcSpiDevice *adcdev, unsigned short timeout);


/** Struct for describing the ADC SPI device under linux. Your struct t_io_dev->dev_data must point to this. */
struct sAdcSpiDevice
{
  char *filename;                                                          /*!< SPI device filename */
  int fd;                                                                  /*!< SPI device file descriptor */
  void *hw_data;                                                           /*!< Pointer to hardware specific data, can be used by the functions in this struct for user defined data */
  t_adc24bit_user_hw_init adc24bit_user_hw_init;                           /*!< User specific hardware initialisation called as part of the ADC hardware initialisation. May be called multiple times */
  t_adc24bit_sync_spi adc24bit_sync_spi;                                   /*!< User implemented function to synchronise ADC SPI, only required when a chip select is not used */
  t_adc24bit_set_loadcell_wire_mode adc24bit_set_loadcell_wire_mode;       /*!< User implemented function to set 4 or 6 wire mode */
  t_adc24bit_wait_for_reading adc24bit_wait_for_reading;                   /*!< User implemented function to wait for an ADC reading (either polling status register, or wait on hardware interrupt) */
};


/**
 * Function to create and initialise a local 24bit ADC device
 *
 * @param io_dev is the IO device to be created. Not NULL.
 * @param device_filename is the SPI device that the ADC is connected to. Not NULL.
 * @param adc24bit_user_hw_init is a user specific hardware initialisation called as part of the ADC hardware initialisation. May be called multiple times. Can be NULL.
 * @param adc24bit_sync_spi is a user implemented function to synchronise ADC SPI, only required when a chip select is not used. Not NULL.
 * @param adc24bit_set_loadcell_wire_mode is a device specific loadcell wire mode function pointer, can be NULL.
 * @param adc24bit_wait_for_reading is a user implemented function to wait for an ADC reading (either polling status register, or wait on hardware interrupt)
 */
int rio_create_device_adc24bit_local(struct t_io_dev *io_dev, const char *device_filename,
                                     t_adc24bit_user_hw_init adc24bit_user_hw_init,
                                     t_adc24bit_sync_spi adc24bit_sync_spi,
                                     t_adc24bit_set_loadcell_wire_mode adc24bit_set_loadcell_wire_mode,
                                     t_adc24bit_wait_for_reading adc24bit_wait_for_reading);


#endif /* ADC_ADS1216_LOCAL_H_ */
