#ifndef BYTEQ_DOT_H
#define BYTEQ_DOT_H 1

/*
 * byteq.h
 *
 * Implements a general purpose byte queue using pthread services to wait for data/room in the buffer
 *
 * Functions available are create, read, write, get level, get space
 *
 * Copyright (C) Rinstrum 2003
 *
 * Written by Ian Burns 27/8/2003
 */
#include <stddef.h>
#include <stdint.h>
#include <pthread.h>
#include <semaphore.h>
#include <time.h>


/** Determines where the pend for data/space occurs (either when read is called, or write is called) */
enum eBQPendDirn { bqPendOnRead = 0, bqPendOnWrite, bqNoPend };


/** Byte queue function results */
enum eBQResult { bqrSuccess = 0, bqrTimeout, bqrEmpty, bqrFull, bqrError };


/** The queue structure used in all functions in this module */
struct s_byteq
{
    enum eBQPendDirn PendDirn;          /*!< Do we pend for read or write? */
    size_t Size;                        /*!< Size of the ring buffer */
    size_t Level;                       /*!< Number of characters currently in the ring buffer */
    sem_t Sem;                          /*!< Read/write semaphore */
    uint8_t *pWrite;                    /*!< Pointer to where next character will be inserted */
    uint8_t *pRead;                     /*!< Pointer from where next character will be extracted */
    uint8_t *pBuffer;                   /*!< Pointer to ring buffer character storage */
    size_t Count;                       /*!< Number of bytes that have passed through the buffer (incremented on read) */
    pthread_mutex_t lock_mutex;         /*!< Exclusive access mutex */
};


/**
 * Create a new byte queue
 *
 * Parameters:
 *   q is a pointer to an existing un-initialised byte queue struct
 *   pend_dirn is the direction in which the queue should pend (either pend for room to write (bqPendOnWrite), or pend for
 *     data to read (bqPendOnRead))
 *   buf is a pointer to a existing unused buffer to put the queue data into
 *   size is the size in bytes of the buffer and hence the queue
 *
 * Returns:
 *   bqrSuccess on success, bqrError otherwise
 */
enum eBQResult ByteQCreate(struct s_byteq *q, enum eBQPendDirn pend_dirn, uint8_t *buf, size_t size);


/**
 * Destroy an existing byte queue
 *
 * Parameters:
 *   q is a pointer to an existing initialised byte queue struct
 *
 * Returns:
 *   bqrSuccess on success, bqrError otherwise
 */
enum eBQResult ByteQDestroy(struct s_byteq *q);


/**
 * Read a byte from a byte queue
 *
 * Parameters:
 *   q is a pointer to an existing initialised byte queue struct
 *   to is the timeout period (use NULL for indefinite wait)
 *   err is a pointer to a byte that will be set to the error value from the operation, if any
 *
 * Returns:
 *   Next character from the buffer on success (*err == bqrSuccess), 0 otherwise
 */
uint8_t ByteQRead(struct s_byteq *q, const struct timespec *to, enum eBQResult *err);


/**
 * Write a byte to a byte queue
 *
 * Parameters:
 *   q is a pointer to an existing initialised byte queue struct
 *   c is the byte to write to the queue
 *   to is the timeout period (use NULL for indefinite wait)
 *
 * Returns:
 *   Error value from the operation, if any (bqrSuccess on success)
 */
enum eBQResult ByteQWrite(struct s_byteq *q, uint8_t c, const struct timespec *to);


/**
 * Get the number of bytes in the queue
 *
 * Parameters:
 *   q is a pointer to an existing initialised byte queue struct
 *
 * Returns:
 *   The number of bytes in the queue
 */
size_t ByteQGetLevel(struct s_byteq *q);


/**
 * Get the amount of space left in the queue
 *
 * Parameters:
 *   q is a pointer to an existing initialised byte queue struct
 *
 * Returns:
 *   The amount of space (in bytes) left in the queue
 */
size_t ByteQGetSpace(struct s_byteq *q);


#endif /* #ifndef BYTEQ_DOT_H */
