#ifndef	LCD_DOT_H
#define	LCD_DOT_H 1

#include <rinhw/rinendian.h>
#include <rinhw/rio.h>


/*
 * Different LCDs that are supported
 */
#define LCD_EWIRELESS 1
#define LCD_DISPLAYTECH 2	/* Not supported */
#define LCD_SHENZHEN 3


/*
 * This is the LCD that is compiled in to the library
 */
#define THE_LCD LCD_EWIRELESS


/*
 * Work out which LCD to use, and pull in the appropriate header file
 */
#if (THE_LCD == LCD_EWIRELESS)
#if (!defined(BYTE_ORDER) || !defined(BIG_ENDIAN))
  #error "Endian not defined"
#endif
  #if (BYTE_ORDER == BIG_ENDIAN)
  	/* Hitachi - big endian */
    #include <rinhw/lcd_ew.h>
  #else
  	/* i386 - little endian */
    #include <rinhw/lcd_ew_le.h>
  #endif
#elif (THE_LCD == LCD_DISPLAYTECH)
    #error DisplayTech LCD Not Supported
#elif (THE_LCD == LCD_SHENZHEN)
    #include <rinhw/lcd_sz.h>
#else
    #error LCD Not Defined
#endif


/*
 * Union of all segments (defined in Specific LCD header file)
 */
typedef union
{
    tAnnunciators Annunciators;
    tTopRowDigits TopRowDigits;
    tTopRowUnits TopRowUnits;
    tTopRightRowDigits TopRightRowDigits;
    tBottomRowUnits BottomRowUnits;
    tBottomRowDigits BottomRowDigits;
    tBottomRightRowDigits BottomRightRowDigits;
    tMiscSegments MiscSegments;
} tNamedSegments;


/*
 * Union of raw/named/byte access for the lcd
 */
typedef union
{
    unsigned char Bytes[NUM_LCD_BYTES];
    tRawSegments Raw;
    tNamedSegments Named;
} tLcdSegments;


/*
 * The lcd data buffer, really shouldn't be available,
 * as it is accessable via device reads/writes, but is
 * required for a link time assigned pointer in the application.
 * Do not access the buffer unless you have the device open
 */
extern tLcdSegments lcd_segments;


/*
 * IO Control functions for the LCD device -
 *
 * rioctl(&dev_lcd, IOCTL_LCD_SET_BACKLIGHT, int level);
 *  Set backlight level 0..100%
 *
 * rioctl(&dev_lcd, IOCTL_LCD_GET_DATAPTR, tLcdSegments **lcd_segs);
 *  Get a pointer to the LCD data buffer. Only access this buffer when
 *  you have the device open. Dodgy. R400 only.
 *
 * rioctl(&dev_lcd, IOCTL_LCD_GET_FLASHPTR, tLcdSegments **lcd_segs);
 *  Get a pointer to the LCD flash buffer. Write this buffer to make
 *  segments flash. Only access this buffer when you have the device open. Dodgy. R400 only.
 *
 * rioctl(&dev_lcd, IOCTL_LCD_ACCESS_DATA);
 *   Set LCD Device reads/writes from/to the data buffer. Available in all implementations.
 *
 * rioctl(&dev_lcd, IOCTL_LCD_ACCESS_FLASH);
 *   Set LCD Device reads/writes from/to the flash buffer (make segments flash if data set). Available in segflash implementations only.
 *
 * rioctl(&dev_lcd, IOCTL_LCD_ACCESS_OUTPUT);
 *   Set LCD Device reads from the output buffer. Do not write. Available in segflash implementations only.
 *
 * rioctl(&dev_lcd, IOCTL_LCD_ACCESS_CHAR);
 *   Set LCD Device reads/writes from/to the character buffer (write characters to the LCD digits). Available in all implmentations.
 *
 */
#define IOCTL_LCD_SET_BACKLIGHT 500

/* These are dodges that are only available on the R400, and really shouldn't even be there */
/* #define IOCTL_LCD_GET_DATAPTR 501 */
/* #define IOCTL_LCD_GET_FLASHPTR 502 */

#define IOCTL_LCD_ACCESS_DATA 503
#define IOCTL_LCD_ACCESS_FLASH 504
#define IOCTL_LCD_ACCESS_OUTPUT 505
#define IOCTL_LCD_ACCESS_CHAR 506


/*
 * Notes on writing to the LCD character buffer (specifically decimal points):
 *
 * When you write a decimal point to the character buffer, the previous digit is
 * checked to see if if has a decimal point, and if so that decimal point is lit,
 * if not nothing is done. In either case, the write position is not incremented,
 * so things like writing a line of "..." for instance, will not work (the LCD will remain unchanged).
 * You would have to space them thus: " . . . .", the spaces being used to increment the write position.
 *
 */

#endif /* #ifndef LCD_DOT_H */
