/*
 * cd_generic.h
 *
 *  Created on: 17/05/2010
 *      Author: ianb
 */

#ifndef LCD_GENERIC_H_
#define LCD_GENERIC_H_

#include <pthread.h>
#include <semaphore.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/stdrinstrum.h>


/*
 * The character write mode -
 * Mainly used for writing decimal points
 *
 *  cwmSet - Clear what is there already before writing
 *  cwmOR - OR the new data with what is there already
 */
enum eCharWriteMode
{
  cwmSet = 0,
  cwmOR
};


/* Forward declaration */
struct s_lcd_definition;


typedef int (*t_lcd_hw_init_fn)(struct s_lcd_definition *lcd_definition);
typedef void (*t_lcd_hw_destroy_fn)(struct s_lcd_definition *lcd_definition);
typedef int (*t_lcd_hw_write_fn)(struct s_lcd_definition *lcd_definition);
typedef void (*t_lcd_put_char_fn)(struct s_lcd_definition *lcd_definition, unsigned char index, unsigned char ch, enum eCharWriteMode cwm);
typedef bool (*t_lcd_does_digit_have_dp_fn)(struct s_lcd_definition *lcd_definition, unsigned char index);
typedef int (*t_lcd_hw_setbacklight_fn)(struct s_lcd_definition *lcd_definition, int lvl_pct);


/** Structure to describe a LCD */
struct s_lcd_definition
{
  void *hw_data;
  unsigned char *lcd_segments;
  size_t lcd_num_segment_bytes;
  unsigned char *lcd_characters;
  size_t lcd_num_characters;
  bool lcd_access_segs_not_chars;
  t_lcd_hw_init_fn lcd_hw_init;
  t_lcd_hw_destroy_fn lcd_hw_destroy;
  t_lcd_hw_write_fn lcd_hw_write;
  t_lcd_put_char_fn lcd_put_char;
  t_lcd_does_digit_have_dp_fn lcd_does_digit_have_dp;
  t_lcd_hw_setbacklight_fn lcd_hw_setbacklight_fn;
  unsigned char *lcd_output_segments;
  unsigned char *lcd_flash_segments;
  unsigned char *lcd_current_access;
  sem_t sem_lcd_update;
  pthread_t lcd_flash_thread;
  int lcd_thread_quit;
};


/**
 * Create a generic LCD device with segment flashing (requires an extra task)
 *
 * @param io_dev is the uninitialised IO device to be filled out
 * @param lcd_hw_data is a pointer to your LCD hardware implementation data
 * @param lcd_num_segment_bytes is the number of segment bytes in the LCD
 * @param lcd_num_chars is the number of characters on the LCD
 * @param lcd_hw_init is a function pointer called to initialise the LCD. Can be NULL, in which case it won't be called
 * @param lcd_hw_destroy is a function pointer called to destroy the LCD. Can be NULL, in which case it won't be called
 * @param lcd_hw_write is a function pointer called to write data to the LCD. Can be NULL, in which case it won't be called
 * @param lcd_put_char is a function pointer called to put a character on the LCD. Not NULL.
 * @param lcd_does_digit_have_dp is a function pointer called to determine if a character position has a decimal point. Not NULL
 * @param lcd_hw_setbacklight_fn is a function pointer called to set the back light level 0..100%. Can be NULL, in which case it won't be called
 *
 * @return RIO_SUCCESS on success, RIO_ERROR otherwise
 */
int lcd_generic_segflash_create_device(struct t_io_dev *io_dev,
                                       void *lcd_hw_data,
                                       size_t lcd_num_segment_bytes,
                                       size_t lcd_num_characters,
                                       t_lcd_hw_init_fn lcd_hw_init,
                                       t_lcd_hw_destroy_fn lcd_hw_destroy,
                                       t_lcd_hw_write_fn lcd_hw_write,
                                       t_lcd_put_char_fn lcd_put_char,
                                       t_lcd_does_digit_have_dp_fn lcd_does_digit_have_dp,
                                       t_lcd_hw_setbacklight_fn lcd_hw_setbacklight_fn
                                      );


/**
 * Create a generic LCD device without segment flashing (doesn't require a task to run)
 *
 * @param io_dev is the uninitialised IO device to be filled out
 * @param lcd_hw_data is a pointer to your LCD hardware implementation data
 * @param lcd_num_segment_bytes is the number of segment bytes in the LCD
 * @param lcd_num_chars is the number of characters on the LCD
 * @param lcd_hw_init is a function pointer called to initialise the LCD. Can be NULL, in which case it won't be called
 * @param lcd_hw_destroy is a function pointer called to destroy the LCD. Can be NULL, in which case it won't be called
 * @param lcd_hw_write is a function pointer called to write data to the LCD. Can be NULL, in which case it won't be called
 * @param lcd_put_char is a function pointer called to put a character on the LCD. Not NULL.
 * @param lcd_does_digit_have_dp is a function pointer called to determine if a character position has a decimal point. Not NULL
 * @param lcd_hw_setbacklight_fn is a function pointer called to set the back light level 0..100%. Can be NULL, in which case it won't be called
 *
 * @return RIO_SUCCESS on success, RIO_ERROR otherwise
 */
int lcd_generic_nosegflash_create_device(struct t_io_dev *io_dev,
                                       void *lcd_hw_data,
                                       size_t lcd_num_segment_bytes,
                                       size_t lcd_num_characters,
                                       t_lcd_hw_init_fn lcd_hw_init,
                                       t_lcd_hw_destroy_fn lcd_hw_destroy,
                                       t_lcd_hw_write_fn lcd_hw_write,
                                       t_lcd_put_char_fn lcd_put_char,
                                       t_lcd_does_digit_have_dp_fn lcd_does_digit_have_dp,
                                       t_lcd_hw_setbacklight_fn lcd_hw_setbacklight_fn
                                      );


#endif /* LCD_GENERIC_H_ */
