#ifndef RIO_DOT_H
#define RIO_DOT_H 1

#include <stdarg.h>
#include <rinhw/ucos_ii.h>


/*
 * NOTE: This is not a full implementation of the C io functions
 * We don't support any of the following IO functions:
 * access, chmod, _chmod, chsize, _close, creat, _creat, creatnew, creattemp,
 * dup, dup2, getftime, lock, locking, mktemp, _open, _read, remove, rename,
 * setftime, setmode, sopen, umask, unlink, unlock, _write
 */


/*
 * IO Device struct
 */
struct t_io_dev
{
    void *dev_data;
    unsigned long status;
    unsigned long posn;

    OS_EVENT *sem_open_excl;

    int (*dev_init)(struct t_io_dev *io_dev);

    int (*dev_open)(struct t_io_dev *io_dev, unsigned short timeout);
    int (*dev_close)(struct t_io_dev *io_dev);

    int (*dev_read)(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout);
    int (*dev_write)(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout);

    long (*dev_filelength)(struct t_io_dev *io_dev);
    int (*dev_ioctl)(struct t_io_dev *io_dev, int func, va_list *arglist);

    int (*dev_destroy)(struct t_io_dev *io_dev);

    int (*dev_isatty)(struct t_io_dev *io_dev);
    long (*dev_tell)(struct t_io_dev *io_dev);
    long (*dev_lseek)(struct t_io_dev *dev, long offset, int fromwhere);
};


/*
 * Internal flags used by devices
 *
 * O_OPEN - Clear when the device is OPEN
 * O_DEVCHAR - Set when the device is a character device
 * O_INITIALISED - Set when device has been initialised
 */
#define O_OPEN        0x00000001
#define O_DEVCHAR     0x00010000
#define O_INITIALISED 0x10000000


/*
 * Seek modes used with rlseek
 *
 * SEEK_SET - Offset is from 0
 * SEEK_CUR - Offset is from current position
 * SEEK_END - Offset is from end of file
 */
#define SEEK_SET    0
#define SEEK_CUR    1
#define SEEK_END    2


/*
 * Result codes
 *
 * Functions that fail return RIO_ERROR
 * Most functions (ropen, rclose, rioctl) will return RIO_SUCCESS on success
 * Others like rread, rwrite return the number of bytes read/written
 */
#define RIO_ERROR (-1)
#define RIO_SUCCESS (1)


/*
 * rregister_io -
 *
 * Registers a new io device, and calls the init function for that device
 */
int rregister_io(struct t_io_dev *io_dev);


/**
 * rdestroy_io -
 *
 * Shutdown the device and free any resources
 */
int rdestroy_io(struct t_io_dev *io_dev);


/*
 * ropen -
 *
 * Open a device for access, with a specified timeout in OS ticks
 * Timeout of 0 is indefinate
 *
 * Locking of a device is done via ropen()
 * and releasing it is done via rclose()
 *
 * Returns -1 for error, 1 for success
 */
int ropen(struct t_io_dev *dev, unsigned short timeout);


/*
 * rclose -
 *
 * Closes/release a device/file
 *
 * Returns -1 for error, 1 for success
 */
int rclose(struct t_io_dev *dev);


/*
 * reof -
 *
 * Returns whether the current file position is the end of the file
 *
 * Returns -1 for error, 1 for eof reached, 0 for not eof
 */
int reof(struct t_io_dev *dev);


/*
 * rfilelength -
 *
 * Returns the size of the file/device in bytes
 */
long rfilelength(struct t_io_dev *dev);


/*
 * rioctl -
 *
 * Device specific io control functions
 *
 * See the individual device headers for available functions
 */
int rioctl(struct t_io_dev *dev, int func, ...);


/*
 * risatty -
 *
 * Returns whether the device is a character device (serialport, etc)
 */
int risatty(struct t_io_dev *dev);


/*
 * rlseek -
 *
 * Seek to a position within a device.
 *    fromwhere is one of SEEK_SET, SEEK_CUR, SEEK_END
 *    offset is a relative offset (in bytes) from the fromwhere position
 *
 * Result is -1 on error, new position in file otherwise
 *
 * Only available on block devices
 */
long rlseek(struct t_io_dev *dev, long offset, int fromwhere);


/*
 * rread -
 *
 * Read len bytes from the device into buf.
 * Will timeout after timeout OS ticks, or 0 for indefinate timeout
 *
 * Returns -1 for error/timeout, len on success
 */
int rread(struct t_io_dev *dev, void *buf, unsigned short len, unsigned short timeout);


/*
 * rwrite -
 *
 * Write len bytes to the device from buf.
 * Will timeout after timeout OS ticks, or 0 for indefinate timeout
 *
 * Returns -1 for error/timeout, len on success
 */
int rwrite(struct t_io_dev *dev, const void *buf, unsigned short len, unsigned short timeout);


/*
 * rtell -
 *
 * Current position in device, returns 0 for character devices
 */
long rtell(struct t_io_dev *dev);


/*
 * default_open -
 *
 * Standard do nothing open device, use this when declaring
 * a device that doesn't need to do anything in a device open
 */
int default_open(struct t_io_dev *io_dev, unsigned short timeout);


/*
 * default_close -
 *
 * Standard do nothing close device, use this when declaring
 * a device that doesn't need to do anything in a device open
 */
int default_close(struct t_io_dev *io_dev);


/*
 * default_ioctl -
 *
 * IOCTL function that returns RIO_ERROR for devices with
 * nothing to configure
 */
int default_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist);


/** Returns whether the device is a character device (serialport, etc) */
int default_isatty(struct t_io_dev *io_dev);


/** Current position in device, returns 0 for character devices */
long default_tell(struct t_io_dev *io_dev);


/*
 * rlseek -
 *
 * Seek to a position within a device.
 *    fromwhere is one of SEEK_SET, SEEK_CUR, SEEK_END
 *    offset is a relative offset (in bytes) from the fromwhere position
 *
 * Result is -1 on error, new position in file otherwise
 *
 * Only available on block devices
 */
long default_lseek(struct t_io_dev *io_dev, long offset, int fromwhere);


#endif /* #ifndef RIO_DOT_H */
