#include <string.h>
#include <stdlib.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include "_rio.h"

#include <rinhw/debug.h>


/*
 * init_io -
 *
 * Initialises the io system. Must be called before any calls to the io functions
 */
void init_io(void *data)
{

    /*
     * I don't think it has to do anything anymore
     */
}


/** Shut down the IO system */
void destroy_io(void)
{
  /* Nothing to do here */
}


/*
 * rregister_io -
 *
 * Create any resources for the device, and call the init function for the device
 */
int rregister_io(struct t_io_dev *io_dev)
{
  int nResult = RIO_SUCCESS;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("rregister_io called\n");
#endif

  /* Only initialise device if it hasn't been already */
  if ( O_INITIALISED == ( io_dev->status & O_INITIALISED ) )
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rregister_io: Device already initialised\n");
#endif
    return RIO_ERROR;
  }

  /* Initialise the device, and make sure its closed */
  io_dev->status |= O_OPEN;

  /*
   * It's possible that io_dev->dev_init()
   * may call this function (rregister_io()) as it creates
   * the device. In this case we don't want to create
   * 2 semaphores (1 would be lost), so we set it to NULL here, and
   * then check before creating it after the io_dev->dev_init()
   * call.
   */
  io_dev->sem_open_excl = 0;

  /* Do the device specific init */
  if ( 0 != io_dev->dev_init )
  {
    nResult = io_dev->dev_init( io_dev );
  }

  /* Handle error */
  if ( nResult < 0 )
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rregister_io: Device specific init failed\n");
#endif
    return nResult;
  }

  /* Create the open exclusive semaphore, if it hasn't already been created */
  if (0 == io_dev->sem_open_excl)
  {
    io_dev->sem_open_excl = OSSemCreate( 1 );
    if ( 0 == io_dev->sem_open_excl )
    {
  #if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("rregister_io: Can't create device open excl sem\n");
  #endif
      return RIO_ERROR;
    }
  }

  /* All successfully initialised */
  io_dev->status |= O_INITIALISED;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("rregister_io finished, result\n", nResult);
#endif
  return nResult;
}


/**
 * rdestroy_io -
 *
 * Shutdown the device and free any resources
 */
int rdestroy_io(struct t_io_dev *io_dev)
{
  int nResult = RIO_SUCCESS;
  unsigned char err;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("rdestroy_io called\n");
#endif

  /* Only destroy device if it has been initialised */
  if ( 0 == ( io_dev->status & O_INITIALISED ) )
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rdestroy_io: Device not initialised\n");
#endif
    return RIO_ERROR;
  }

  /* TODO if something is pending, stop it pending */

  /* Do the device specific destroy */
  if ( 0 != io_dev->dev_destroy )
  {
    nResult = io_dev->dev_destroy( io_dev );
  }

  /* Handle error */
  if ( nResult < 0 )
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rregister_io: Device specific destroy failed\n");
#endif
    return nResult;
  }

  /* Destroy the open exclusive semaphore */
  OSSemDel(io_dev->sem_open_excl, OS_DEL_ALWAYS, &err);
  if (OS_NO_ERR != err)
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rregister_io: Can't create device open excl sem\n");
#endif
  }
  io_dev->sem_open_excl = NULL;

  /* Device destroyed, device closed */
  io_dev->status &= ~O_INITIALISED;
  io_dev->status |= O_OPEN;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("rdestroy_io finished, result\n", nResult);
#endif
  return nResult;
}


/*
 * ropen -
 *
 * Open a device for access, with a specified timeout in OS ticks
 * Timeout of 0 is indefinite
 *
 * Locking of a device is done via ropen()
 * and releasing it is done via rclose()
 *
 * Returns -1 for error, 1 for success
 */
int ropen(struct t_io_dev *io_dev, unsigned short timeout)
{
    unsigned char err;
    int result = RIO_SUCCESS;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("ropen called\n");
#endif

    /*
     * First, wait until the device becomes available
     */
    OSSemPend( io_dev->sem_open_excl, timeout, &err );
    if (err != OS_NO_ERR)
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("Timeout or error waiting for device\n");
#endif
        return RIO_ERROR;
    }

    /*
     * Now open the device, if it fails, release the device and return the error
     */
    if ( 0 != io_dev->dev_open )
    {
      result = io_dev->dev_open( io_dev, timeout );
    }

    if ( result < 0 )
    {
        OSSemPost( io_dev->sem_open_excl );
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("Device specific open failed\n");
#endif
        return result;
    }

    /*
     * Mark it as open, and set the position to the start
     */
    io_dev->status &= ~O_OPEN;
    io_dev->posn = 0UL;

    /*
     * Success
     */
    return RIO_SUCCESS;
}


/*
 * rclose -
 *
 * Closes/release a device/file
 *
 * Returns -1 for error, 1 for success
 */
int rclose(struct t_io_dev *io_dev)
{
    int result = RIO_SUCCESS;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rclose called\n");
#endif

    /*
     *  Check its actually open
     */
    if ( io_dev->status & O_OPEN )
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("Device not open\n");
#endif
        return RIO_ERROR;
    }

    /*
     * Call the device close function
     */
    if ( 0 != io_dev->dev_close )
    {
      result = io_dev->dev_close( io_dev );
    }

    if ( result < 0 )
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("Device specific close failed\n");
#endif
        return result;
    }

    /*
     * Mark the device as closed
     */
    io_dev->status |= O_OPEN;

    /*
     * Signal that other tasks can now access the device
     */
    OSSemPost( io_dev->sem_open_excl );

    /*
     * Success
     */
    return RIO_SUCCESS;
}


/*
 * rread -
 *
 * Read len bytes from the device into buf.
 * Will timeout after timeout OS ticks, or 0 for indefinate timeout
 *
 * Returns -1 for error/timeout, len on success
 */
int rread(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
    int nResult = RIO_ERROR;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rread called\n");
#endif

    /*
     *  Check its actually open
     */
    if ( io_dev->status & O_OPEN )
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("Device not open\n");
#endif
        return RIO_ERROR;
    }

    /*
     * Do the read
     */
    if ( 0 != io_dev->dev_read )
    {
      nResult = io_dev->dev_read( io_dev, buf, len, timeout );
    }

    return nResult;
}


/*
 * rwrite -
 *
 * Write len bytes to the device from buf.
 * Will timeout after timeout OS ticks, or 0 for indefinite timeout
 *
 * Returns -1 for error/timeout, len on success
 */
int rwrite(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  int nResult = RIO_SUCCESS;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("rwrite() called\n");
#endif

  /*
   *  Check its actually open
   */
  if ( io_dev->status & O_OPEN )
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rwrite: Device not open\n");
#endif
    return RIO_ERROR;
  }

  if ( 0 != io_dev->dev_write )
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("rwrite: doing local device write\n");
#endif
    nResult = io_dev->dev_write( io_dev, buf, len, timeout );
  }

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("rwrite() finished with result %d\n", nResult);
#endif
  return nResult;
}


/*
 * risatty -
 *
 * Returns whether the device is a character device (serialport, etc)
 */
int risatty(struct t_io_dev *io_dev)
{
  int result;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("risatty called\n");
#endif

  if (0 != io_dev->dev_isatty)
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("risatty: doing device isatty\n");
#endif
    result = io_dev->dev_isatty(io_dev);
  }
  else
  {
    result = RIO_ERROR;
  }

  return result;
}


/*
 * rtell -
 *
 * Current position in device, returns 0 for character devices
 */
long rtell(struct t_io_dev *io_dev)
{
  long result;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("rtell called\n");
#endif

  /* Check its actually open */
  if (io_dev->status & O_OPEN)
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("Device not open\n");
#endif
    return RIO_ERROR;
  }

  /* For a character device, return a position of 0, otherwise get the position from the iob */
  if (0 != io_dev->dev_tell)
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("rtell: doing device tell\n");
#endif
    result = io_dev->dev_tell(io_dev);
  }
  else
  {
    result = RIO_ERROR;
  }

  return result;
}


/*
 * rfilelength -
 *
 * Returns the size of the file/device in bytes
 */
long rfilelength(struct t_io_dev *io_dev)
{
    int nResult = RIO_ERROR;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rfilelength called\n");
#endif

    /*
     *  Check its actually open
     */
    if ( io_dev->status & O_OPEN )
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("Device not open\n");
#endif
        return RIO_ERROR;
    }

    /*
     * Get the file length
     */
    if ( 0 != io_dev->dev_filelength )
    {
      nResult = io_dev->dev_filelength( io_dev );
    }
    return nResult;
}


/*
 * rioctl -
 *
 * Device specific io control functions
 *
 * See the individual device headers for available functions
 */
int rioctl(struct t_io_dev *io_dev, int func, ...)
{
    va_list arglist;
    int result;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rioctl called\n");
#endif

    /*
     *  Check its actually open
     */
    if ( io_dev->status & O_OPEN )
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("Device not open\n");
#endif
        return RIO_ERROR;
    }

    /*
     *  Check there is a ioctl function
     */
    if ( 0 == io_dev->dev_ioctl )
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("No rioctl function\n");
#endif
        return RIO_ERROR;
    }

    /*
     * Start the variable argument list
     */
    va_start(arglist, func);

    /*
     * Call the device's ioctl handler
     */
    result = io_dev->dev_ioctl( io_dev, func, &arglist );

    /*
     * Return the result
     */
    va_end(arglist);

    return result;
}


/*
 * rlseek -
 *
 * Seek to a position within a device.
 *    fromwhere is one of SEEK_SET, SEEK_CUR, SEEK_END
 *    offset is a relative offset (in bytes) from the fromwhere position
 *
 * Result is -1 on error, new position in file otherwise
 *
 * Only available on block devices
 */
long rlseek(struct t_io_dev *io_dev, long offset, int fromwhere)
{
  long result;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("rlseek called\n");
#endif

  /* Check its actually open */
  if (io_dev->status & O_OPEN)
  {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("Device not open\n");
#endif
    return (long) RIO_ERROR;
  }

  if (0 != io_dev->dev_lseek)
  {
    result = io_dev->dev_lseek(io_dev, offset, fromwhere);
  }
  else
  {
    result = RIO_ERROR;
  }

  return result;
}


/*
 * reof -
 *
 * Returns whether the current file position is the end of the file
 *
 * Returns -1 for error, 1 for eof reached, 0 for not eof
 */
int reof(struct t_io_dev *io_dev)
{
    int result, tty;

#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("reof called\n");
#endif

    /*
     *  Check its actually open
     */
    if (io_dev->status & O_OPEN)
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("Device not open\n");
#endif
        return RIO_ERROR;
    }

    /*
     * For a character device
     */
    if ((tty = risatty(io_dev)) < 0)
    {
#if HW_DEBUG(RIO_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("risatty call failed\n");
#endif
        return RIO_ERROR;
    }

    /*
     * If a character device, EOF if no characters in buffer
     */
    if (tty)
    {
        result = (rfilelength(io_dev) == 0L);
    }
    else
    {

        /*
         * Otherwise block device, EOF if position at filelength
         */
        result = (rtell(io_dev) >= rfilelength(io_dev));
    }

    return result;
}
