/*
 * rtc_local.c
 *
 *  Created on: 10/02/2009
 *      Author: ianb
 */
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <rinhw/rio.h>
#include <rinhw/rinendian.h>
#include <rinhw/rtc.h>
#include <rinhw/debug.h>

#include <rinhw/rtc_local.h>


/** Initialise the real time clock */
static int rtc_local_init(struct t_io_dev *io_dev)
{
#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rtc_local_init called\n");
#endif

    /* Nothing to be done here */
#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rtc_local_init: rtc initialisation success\n");
#endif
    return RIO_SUCCESS;
}


/** Write to the real time clock. buf must point to a t_datetime */
static int rtc_local_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
    t_datetime *dt = (t_datetime *) buf;
    struct tm newtime;
    struct timeval tv;
    INT32U oldticks;
    int result;

#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rtc_local_write called\n");
#endif

    /*
     * Check the length will fit
     */
    if (len != sizeof(t_datetime))
    {
#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("rtc_local_write: size error\n");
#endif
        return RIO_ERROR;
    }

    /* Fill out tm structure */
    memset(&newtime, 0, sizeof(newtime));
    newtime.tm_sec = dt->second;
    newtime.tm_min = dt->minute;
    newtime.tm_hour = dt->hour;
    newtime.tm_mday = dt->date;
    newtime.tm_mon = dt->month - 1;
    newtime.tm_year = dt->year - 1900;

    /* Convert to timeval */
    tv.tv_usec = 0;
    tv.tv_sec = mktime(&newtime);
    if ((time_t) -1 == tv.tv_sec)
    {
#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("rtc_local_write: mktime() failed\n");
#endif
      return RIO_ERROR;
    }

    /* Lock */
    oldticks = OSPreFixTimeGet();

    /* Now set the time, need root type permissions */
    result = settimeofday(&tv, NULL);

    /* Adjust ucos time, and unlock */
    OSPostFixTimeGet(oldticks);
    if (result < 0)
    {
#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("rtc_local_write: settimeofday() failed\n");
#endif
      return RIO_ERROR;
    }

    /* And write it to the hardware clock, need root type permissions */
    if (0 != system("hwclock -w"))
    {
#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("rtc_local_write: hwclock -w failed\n");
#endif
      return RIO_ERROR;
    }

#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rtc_local_write finished\n");
#endif

    /* Success, return size of struct written */
    return (int) sizeof(t_datetime);
}


/** Destroy real time clock device */
static int rtc_local_destroy(struct t_io_dev *io_dev)
{
#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rtc_local_destroy called\n");
#endif

    /* Nothing to be done here */
    return RIO_SUCCESS;
}


/** Read time from the real time clock. buf must point to a t_datetime */
static int rtc_local_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
    t_datetime *dt = (t_datetime *) buf;
    const time_t the_time = time ( ( time_t * ) 0 );
    struct tm *current_time = localtime ( &the_time );

#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rtc_local_read called\n");
#endif

    /*
     * Check the length will fit
     */
    if (len != sizeof(t_datetime))
    {
#if HW_DEBUG(RTC_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("rtc_local_read: size error\n");
#endif
        return RIO_ERROR;
    }

    /*
     * Convert it to the standard RTC format
     * and fill out the struct to return
     */
    dt->second = current_time->tm_sec;
    dt->minute = current_time->tm_min;
    dt->hour = current_time->tm_hour;
    dt->day = current_time->tm_wday + 1;
    dt->date = current_time->tm_mday;
    dt->month = current_time->tm_mon + 1;
    dt->year = current_time->tm_year + 1900;

    /* Success */
    return (int) sizeof(t_datetime);
}


/** Return the size of the datetime struct */
static long rtc_local_filelength(struct t_io_dev *io_dev)
{
    return (long) sizeof(t_datetime);
}


/** Function to create and initialise a local RTC device */
int rio_create_device_rtc_local(struct t_io_dev *io_dev)
{

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) 0;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = rtc_local_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = rtc_local_read;
  io_dev->dev_write = rtc_local_write;
  io_dev->dev_filelength = rtc_local_filelength;
  io_dev->dev_ioctl = default_ioctl;
  io_dev->dev_destroy = rtc_local_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);
}
