/*
 * linuxserial.c
 *
 *  Created on: 23/10/2008
 *      Author: ianb
 */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <linux/serial.h>
#include <rinhw/sio.h>

#include "linuxserial.h"


/*
 * This RS485 stuff is all quite horrible.
 * On the WE2111 we are running 2.6.26 kernel
 * Our version of GCC is pre-compiled from Code Sourcery against 2.6.35 kernel headers
 *
 * struct serial_rs485 in <linux/serial.h> has changed a bit since then, and
 * I've backported the necessary functionality from newer kernels.
 *
 * However because we don't re-build GCC, our kernel headers are stuck at 2.6.35.
 *
 * So I define struct serial_rs485_dodgy here to match the newer kernel one.
 *
 * The long term solution is to build our own GCC.
 */

/* RS485 driver-specific ioctls: */
#ifndef TIOCGRS485
#define TIOCGRS485      0x542E
#endif
#ifndef TIOCSRS485
#define TIOCSRS485      0x542F
#endif

#ifndef SER_RS485_RX_DURING_TX
struct serial_rs485_dodgy {
  __u32 flags;      /* RS485 feature flags */
#if 0
#define SER_RS485_ENABLED   (1 << 0)  /* If enabled */
#define SER_RS485_RTS_ON_SEND   (1 << 1)  /* Logical level for
                 RTS pin when
                 sending */
#define SER_RS485_RTS_AFTER_SEND  (1 << 2)  /* Logical level for
                 RTS pin after sent*/
#endif
#define SER_RS485_RX_DURING_TX    (1 << 4)
  __u32 delay_rts_before_send;  /* Delay before send (milliseconds) */
  __u32 delay_rts_after_send; /* Delay after send (milliseconds) */
  __u32 padding[5];   /* Memory is cheap, new structs
             are a royal PITA .. */
};
#else
#define serial_rs485_dodgy serial_rs485
#endif


/** Initialise the serial port structure with the port name. Doesn't open port. */
bool LinuxSerialInit( struct sLinuxSerialPort *pLinuxSerialPort, const char *pszPortName )
{
  pLinuxSerialPort->m_nSerFd = -1;

  pLinuxSerialPort->m_pszPortName = strdup(pszPortName);
  if (NULL == pLinuxSerialPort->m_pszPortName)
  {
    return false;
  }
  pLinuxSerialPort->m_nWriteTimeoutMilliseconds = 0;
  pLinuxSerialPort->m_nReadTimeoutMilliseconds = 0;
  return true;
}


/** Close port if open and clean up allocated memory in struct */
void LinuxSerialShutdown( struct sLinuxSerialPort *pLinuxSerialPort )
{
  LinuxSerialClose( pLinuxSerialPort );
  if ( pLinuxSerialPort->m_pszPortName )
  {
    free( pLinuxSerialPort->m_pszPortName );
    pLinuxSerialPort->m_pszPortName = NULL;
  }
}


/** Open the serial port */
bool LinuxSerialOpen( struct sLinuxSerialPort *pLinuxSerialPort )
{
  int flags;
  bool result;

  /* Use O_NDELAY to open the port without blocking for CTS? or something */
  pLinuxSerialPort->m_nSerFd = open(pLinuxSerialPort->m_pszPortName, O_RDWR | O_NOCTTY | O_NDELAY);
  if (pLinuxSerialPort->m_nSerFd < 0)
  {
    result = false;
    goto lso_done;
  }
  if (fcntl(pLinuxSerialPort->m_nSerFd, F_SETFD, FD_CLOEXEC) < 0)
  {
    result = false;
    goto lso_done;
  }

  /* Now setup blocking reads and writes */
  flags = fcntl(pLinuxSerialPort->m_nSerFd, F_GETFL);
  if (flags < 0)
  {
    result = false;
    goto lso_done;
  }
  flags &= ~O_NONBLOCK;
  if (fcntl(pLinuxSerialPort->m_nSerFd, F_SETFL, flags) < 0)
  {
    result = false;
    goto lso_done;
  }

  /* Success */
  result = true;

lso_done:
  return result;
}


/** Close the serial port */
bool LinuxSerialClose( struct sLinuxSerialPort *pLinuxSerialPort )
{
  if ( pLinuxSerialPort->m_nSerFd >= 0 )
  {
    close( pLinuxSerialPort->m_nSerFd );
    pLinuxSerialPort->m_nSerFd = -1;
  }
  return true;
}


/** Check if the serial port is open */
bool LinuxSerialIsOpen(struct sLinuxSerialPort *pLinuxSerialPort, bool *pbIsOpen)
{
  if (NULL == pLinuxSerialPort || NULL == pbIsOpen)
    return false;
  *pbIsOpen = (pLinuxSerialPort->m_nSerFd >= 0);
  return true;
}


/** Set the write timeout, 0 for indefinite */
bool LinuxSerialSetWriteTimeout( struct sLinuxSerialPort *pSerialPort, int32u ui32MilliSeconds )
{
  pSerialPort->m_nWriteTimeoutMilliseconds = ui32MilliSeconds;
  return true;
}


/** Set the read timeout, 0 for indefinite */
bool LinuxSerialSetReadTimeout( struct sLinuxSerialPort *pSerialPort, int32u ui32MilliSeconds )
{
  pSerialPort->m_nReadTimeoutMilliseconds = ui32MilliSeconds;
  return true;
}


struct sBaudConversion
{
  speed_t LinuxBaud;
  unsigned int ActualBaud;
};

/* #define NUM_BAUD_CONVERSIONS 30 */
#define NUM_BAUD_CONVERSIONS 17
static const struct sBaudConversion sC_BaudConversions[ NUM_BAUD_CONVERSIONS ] =
{
  { B50, 50 },
  { B75, 75 },
  { B110, 110 },
  { B134, 134 },
  { B150, 150 },
  { B200, 200 },
  { B300, 300 },
  { B600, 600 },
  { B1200, 1200 },
  { B1800, 1800 },
  { B2400, 2400 },
  { B4800, 4800 },
  { B9600, 9600 },
  { B19200, 19200 },
  { B38400, 38400 },
  { B57600, 57600 },
  { B115200, 115200 },
#if 0 /* These are #if'd out in uCLibc, or at least the version we have. We don't need them so far, so I've just commented them out for now. */
  { B230400, 230400 },
  { B460800, 460800 },
  { B500000, 500000 },
  { B576000, 576000 },
  { B921600, 921600 },
  { B1000000, 1000000 },
  { B1152000, 1152000 },
  { B1500000, 1500000 },
  { B2000000, 2000000 },
  { B2500000, 2500000 },
  { B3000000, 3000000 },
  { B3500000, 3500000 },
  { B4000000, 4000000 }
#endif
};


static speed_t LookupLinuxBaudRate( int32u ui32Baud )
{
  int i;
  for ( i = 0; i < NUM_BAUD_CONVERSIONS; i++ )
  {
    if ( sC_BaudConversions[i].ActualBaud == ui32Baud )
    {
      return sC_BaudConversions[i].LinuxBaud;
    }
  }
  return B9600;
}


static int32u LookupRealBaudRate( speed_t LinuxBaud )
{
  int i;
  for ( i = 0; i < NUM_BAUD_CONVERSIONS; i++ )
  {
    if ( ( sC_BaudConversions[i].LinuxBaud & CBAUD ) == LinuxBaud )
    {
      return sC_BaudConversions[i].ActualBaud;
    }
  }
  return 9600;
}


/** Get the configuration for the serial port */
bool LinuxSerialGetConfig( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32BaudRate,
                           int8u *pDataBits, int8u *pParity, int8u *pStopBits, bool *pUseCts )
{
  struct termios options;

  /* Get the current options for the port */
  if ( tcgetattr( pLinuxSerialPort->m_nSerFd, &options ) < 0 )
  {
    return false;
  }

  /* Get the current baud rate */
  *pui32BaudRate = LookupRealBaudRate( cfgetispeed( &options ) );

  /* Set the data bits */
  if ( CS5 == ( options.c_cflag & CSIZE ) )
    *pDataBits = 5;
  else if ( CS6 == ( options.c_cflag & CSIZE ) )
    *pDataBits = 6;
  else if ( CS7 == ( options.c_cflag & CSIZE ) )
    *pDataBits = SIO_DATABITS_SEVEN;
  else /* ( CS8 == ( options.c_cflag & CSIZE ) ) */
    *pDataBits = SIO_DATABITS_EIGHT;

  /* Set the stop bits */
  *pStopBits = ( CSTOPB == ( options.c_cflag & CSTOPB ) ) ? SIO_STOPBITS_TWO : SIO_STOPBITS_ONE;

  /* Set the parity bits */
  if ( PARENB == ( options.c_cflag & PARENB ) )
  {
    /* 1 or 2 parity bits */
    *pParity = ( PARODD == ( options.c_cflag & PARODD ) ) ? SIO_PARITY_ODD : SIO_PARITY_EVEN;
  }
  else
  {
    /* No parity */
    *pParity = SIO_PARITY_NONE;
  }

  /* CTS handshaking */
  *pUseCts = ( CRTSCTS == ( options.c_cflag & CRTSCTS ));

  return true;
}


/** Set the configuration for the serial port */
bool LinuxSerialSetConfig( struct sLinuxSerialPort *pLinuxSerialPort, int32u ui32BaudRate,
                           int8u DataBits, int8u Parity, int8u StopBits, bool UseCts )
{
  struct termios options;
  speed_t linuxbaud;

  /* Zero out settings */
  bzero( &options, sizeof(options) );

  /* Set the baud rates */
  linuxbaud = LookupLinuxBaudRate( ui32BaudRate );
  cfsetispeed( &options, linuxbaud );
  cfsetospeed( &options, linuxbaud );

  /* Set the data bits */
  options.c_cflag &= ~CSIZE;
  if ( 5 == DataBits )
    options.c_cflag |= CS5;
  else if ( 6 == DataBits )
    options.c_cflag |= CS6;
  else if ( SIO_DATABITS_SEVEN == DataBits )
    options.c_cflag |= CS7;
  else /* if ( SIO_DATABITS_EIGHT == DataBits ) */
    options.c_cflag |= CS8;

  /* Set the stop bits */
  if ( SIO_STOPBITS_ONE != StopBits )
    options.c_cflag |= CSTOPB;  /* 2 Stop bits */
  else
    options.c_cflag &= ~CSTOPB; /* 1 Stop bits */

  /* Set the parity bits */
  if ( SIO_PARITY_ODD == Parity )  /* Odd */
  {
    options.c_cflag |= PARENB;
    options.c_cflag |= PARODD;
  }
  else if ( SIO_PARITY_EVEN == Parity ) /* Even */
  {
    options.c_cflag |= PARENB;
    options.c_cflag &= ~PARODD;
  }
  else /* No parity SIO_PARITY_NONE */
  {
    options.c_cflag &= ~PARENB;
    options.c_cflag &= ~PARODD;
  }

  /* Enable the receiver and set local mode */
  options.c_cflag |= (CLOCAL | CREAD);

  /* Ignore parity ??? */
  options.c_iflag = IGNPAR;
  options.c_oflag = 0;

  options.c_cc[VTIME] = 0; /* No Time delay */
  options.c_cc[VMIN] = 1; /* Blocking read until 1 character arrives */


  /* CTS handshaking */
  if (UseCts)
  {
    options.c_cflag |= CRTSCTS;
  }

  /*
   * Set the new options for the port...
   */
  tcsetattr( pLinuxSerialPort->m_nSerFd, TCSANOW, &options );
  return true;
}


/** Read data from the serial port */
enum eAbstractSerialResult LinuxSerialRead( struct sLinuxSerialPort *pLinuxSerialPort, int8u *pui8Buf, int32u ui32Size )
{
  int32u ui32Read = 0;
  int res;
  struct timeval timeoutsrc;
  struct timeval timeout;
  fd_set readfds;
  int maxfd;

  /* If not open, return error */
  if (pLinuxSerialPort->m_nSerFd < 0)
  {
    return asrError;
  }

  /* Work out timeout */
  timeoutsrc.tv_usec = (pLinuxSerialPort->m_nReadTimeoutMilliseconds % 1000) * 1000;
  timeoutsrc.tv_sec = pLinuxSerialPort->m_nReadTimeoutMilliseconds / 1000;

  while ( ui32Read < ui32Size )
  {
    if (0 != pLinuxSerialPort->m_nReadTimeoutMilliseconds)
    {
      /* Wait for data or timeout, we use two copies of timeout because select changes it */
      timeout = timeoutsrc;
      FD_ZERO(&readfds);
      FD_SET(pLinuxSerialPort->m_nSerFd, &readfds);
      maxfd = pLinuxSerialPort->m_nSerFd + 1;
      res = select(maxfd, &readfds, NULL, NULL, &timeout);
      if (res < 0)
      {
        /* Error */
        return asrError;
      }
      else if (0 == res)
      {
        /* Timeout */
        return asrTimeout;
      }
    }

    /* Got data, do a read */
    res = read( pLinuxSerialPort->m_nSerFd, &pui8Buf[ ui32Read ], ui32Size - ui32Read );
    if ( res < 0 )
    {
      /* <0 means error */
      return asrError;
    }
    if (0 == res)
    {
      /* 0 means EOF */
      return asrEndOfFile;
    }
    ui32Read += res;
  }
  return asrSuccess;
}


/** Write data to the serial port */
enum eAbstractSerialResult LinuxSerialWrite( struct sLinuxSerialPort *pLinuxSerialPort, const int8u *pui8Buf, int32u ui32Size )
{
  int32u ui32Sent = 0;
  int res;
  struct timeval timeoutsrc;
  struct timeval timeout;
  fd_set writefds;
  int maxfd;

  /* If not open, return error */
  if (pLinuxSerialPort->m_nSerFd < 0)
  {
    return asrError;
  }

  /* Work out timeout */
  timeoutsrc.tv_usec = (pLinuxSerialPort->m_nWriteTimeoutMilliseconds % 1000) * 1000;
  timeoutsrc.tv_sec = pLinuxSerialPort->m_nWriteTimeoutMilliseconds / 1000;

  while (ui32Sent < ui32Size)
  {
    if (0 != pLinuxSerialPort->m_nWriteTimeoutMilliseconds)
    {
      /* Wait for room in write buffer or timeout, we use two copies of timeout because select changes it */
      timeout = timeoutsrc;
      FD_ZERO(&writefds);
      FD_SET(pLinuxSerialPort->m_nSerFd, &writefds);
      maxfd = pLinuxSerialPort->m_nSerFd + 1;
      res = select(maxfd, NULL, &writefds, NULL, &timeout);
      if (res < 0)
      {
        /* Error */
        return asrError;
      }
      else if (0 == res)
      {
        /* Timeout */
        return asrTimeout;
      }
    }

    /* Ok to write */
    res = write(pLinuxSerialPort->m_nSerFd, &pui8Buf[ui32Sent], ui32Size - ui32Sent);
    if (res < 0)
    {
      return asrError;
    }
    ui32Sent += res;
  }
  return asrSuccess;
}


/** Flush the serial port buffers. queue_selector is one of TCIFLUSH, TCOFLUSH, TCIOFLUSH */
bool LinuxSerialFlush( struct sLinuxSerialPort *pLinuxSerialPort, int queue_selector )
{
  return ( tcflush( pLinuxSerialPort->m_nSerFd, queue_selector ) >= 0 );
}


/** Get the number of bytes available in the transmit buffer */
bool LinuxSerialGetTxBufferLevel( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32Level )
{
  int out_buf_level;
  int res;

  res = ioctl( pLinuxSerialPort->m_nSerFd, TIOCOUTQ, &out_buf_level);
  *pui32Level = (res >= 0 ) ? out_buf_level : 0;
  return ( res >= 0 );
}


/** Get the size in bytes of the transmit buffer */
bool LinuxSerialGetTxBufferSize( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32Level )
{
  *pui32Level = 1024; /* TODO sort out buffer sizes */
  return true;
}


/** Get the number of bytes available in the receive buffer */
bool LinuxSerialGetRxBufferLevel( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32Level )
{
  int bytes_avail;
  int res;

  res = ioctl( pLinuxSerialPort->m_nSerFd, FIONREAD, &bytes_avail);
  *pui32Level = (res >= 0 ) ? bytes_avail : 0;
  return ( res >= 0 );
}


/** Get the size in bytes of the receive buffer */
bool LinuxSerialGetRxBufferSize( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32Level )
{
  *pui32Level = 1024; /* TODO sort out buffer sizes */
  return true;
}


/**
 * Get the line state for a mask
 *
 * @param pbDtr set to true if line specified by mask is asserted (RS232 +ve, TTL 0v), false otherwise
 * @param mask is the line mask (TIOCM_RTS, TIOCM_DTR, etc)
 */
static bool LinuxSerialGetLineState(struct sLinuxSerialPort *pLinuxSerialPort, bool *pbLine, int mask)
{
  int status = 0;

  if (NULL == pLinuxSerialPort || NULL == pbLine)
    return false;

  /* Get the serial port status */
  if (ioctl(pLinuxSerialPort->m_nSerFd, TIOCMGET, &status) < 0)
    return false;

  /* Get the line state */
  *pbLine = (0 != (status & mask));
  return true;
}


/**
 * Get the state of the CTS pin (regardless of whether CTS blocking is enabled).
 * Returns 1 == Clear to send data, 0 == blocking
 */
bool LinuxSerialGetCtsState( struct sLinuxSerialPort *pLinuxSerialPort, bool *pbCts )
{
  bool state;
  bool result;

  if (NULL == pbCts)
    return false;

  result = LinuxSerialGetLineState(pLinuxSerialPort, &state, TIOCM_CTS);
  *pbCts = !state;
  return result;
}


/**
 * Get the state of the RTS pin
 *
 * @param pbRts set to true if RTS asserted (RS232 +ve, TTL 0v), false otherwise
 */
bool LinuxSerialGetRtsState( struct sLinuxSerialPort *pLinuxSerialPort, bool *pbRts )
{
  return LinuxSerialGetLineState(pLinuxSerialPort, pbRts, TIOCM_RTS);
}


/**
 * Get the state of the DTR pin
 *
 * @param pbDtr set to true if DTR asserted (RS232 +ve, TTL 0v), false otherwise
 */
bool LinuxSerialGetDtrState( struct sLinuxSerialPort *pLinuxSerialPort, bool *pbDtr )
{
  return LinuxSerialGetLineState(pLinuxSerialPort, pbDtr, TIOCM_DTR);
}


/**
 * Set the state of a pin
 *
 * @param bLine is the state of the line (true = asserted = RS232 +ve = TTL 0v)
 * @param mask is the line mask (TIOCM_RTS, TIOCM_DTR, etc)
 */
static bool LinuxSerialSetLineState( struct sLinuxSerialPort *pLinuxSerialPort, bool bLine, int mask )
{
  int status = 0;
  int result;

  if (NULL == pLinuxSerialPort)
  {
    return false;
  }

  /* Get the serial port status */
  result = ioctl(pLinuxSerialPort->m_nSerFd, TIOCMGET, &status);
  if (result < 0)
  {
    return false;
  }

  /* Set the line state, true asserts the line */
  if (bLine)
  {
    status |= mask;
  }
  else
  {
    status &= ~mask;
  }

  /* Set the serial port status */
  result = ioctl(pLinuxSerialPort->m_nSerFd, TIOCMSET, &status);
  return (result >= 0);
}


/**
 * Set the state of the RTS pin
 *
 * @param bRts true to assert RTS (RS232 +ve, TTL 0v), false otherwise
 */
bool LinuxSerialSetRtsState( struct sLinuxSerialPort *pLinuxSerialPort, bool bRts )
{
  return LinuxSerialSetLineState( pLinuxSerialPort, bRts, TIOCM_RTS );
}


/**
 * Set the state of the DTR pin
 *
 * @param bDts true to assert DTR (RS232 +ve, TTL 0v), false otherwise
 */
bool LinuxSerialSetDtrState( struct sLinuxSerialPort *pLinuxSerialPort, bool bDtr )
{
  return LinuxSerialSetLineState( pLinuxSerialPort, bDtr, TIOCM_DTR );
}


#define set_flag(flag, mask, level) if (level) flag |= mask; else flag &= ~mask


/**
 * Enable the use of the kernel RS485 transmitter enable using RTS support via TIOCSRS485
 *
 * @param pLinuxSerialPort is the serial port
 * @param enable is used to enable or disable RS485 transmitter enable using RTS
 * @param assert_on_transmit will assert RTS while transmitting if set, otherwise will de-assert RTS while transmitting
 */
bool LinuxSerialSetRS485Rts(struct sLinuxSerialPort *pLinuxSerialPort, bool enable, bool assert_on_transmit)
{
  struct serial_rs485_dodgy rs485_config;

  memset(&rs485_config, 0, sizeof(rs485_config));
  set_flag(rs485_config.flags, SER_RS485_ENABLED, enable);
  set_flag(rs485_config.flags, SER_RS485_RTS_ON_SEND, assert_on_transmit);
  set_flag(rs485_config.flags, SER_RS485_RTS_AFTER_SEND, !assert_on_transmit);
  set_flag(rs485_config.flags, SER_RS485_RX_DURING_TX, true);
  rs485_config.delay_rts_before_send = 0;
  rs485_config.delay_rts_after_send = 0;

  if (ioctl(pLinuxSerialPort->m_nSerFd, TIOCSRS485, &rs485_config) < 0)
  {
    return false;
  }
  return true;
}
