/*
 * linuxserial.h
 *
 *  Created on: 23/10/2008
 *      Author: ianb
 */
#ifndef LINUXSERIAL_H_
#define LINUXSERIAL_H_

#include <sys/types.h>
#include <linux/types.h>
#include <rinhw/stdrinstrum.h>


/** Result of a serial operation */
enum eAbstractSerialResult { asrSuccess = 0, asrTimeout, asrError, asrEndOfFile };


struct sLinuxSerialPort
{
  char *m_pszPortName;
  int m_nSerFd;
  int m_nReadTimeoutMilliseconds;
  int m_nWriteTimeoutMilliseconds;
};


/** Initialise the serial port structure with the port name. Doesn't open port. */
bool LinuxSerialInit( struct sLinuxSerialPort *pLinuxSerialPort, const char *pszPortName );


/** Close port if open and clean up allocated memory in struct */
void LinuxSerialShutdown( struct sLinuxSerialPort *pLinuxSerialPort );


/** Open the serial port */
bool LinuxSerialOpen( struct sLinuxSerialPort *pLinuxSerialPort );


/** Close the serial port */
bool LinuxSerialClose( struct sLinuxSerialPort *pLinuxSerialPort );


/** Check if the serial port is open */
bool LinuxSerialIsOpen(struct sLinuxSerialPort *pLinuxSerialPort, bool *pbIsOpen);


/** Set the write timeout, 0 for indefinate */
bool LinuxSerialSetWriteTimeout( struct sLinuxSerialPort *pSerialPort, int32u ui32MilliSeconds );


/** Set the read timeout, 0 for indefinate */
bool LinuxSerialSetReadTimeout( struct sLinuxSerialPort *pSerialPort, int32u ui32MilliSeconds );


/** Get the configuration for the serial port */
bool LinuxSerialGetConfig( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32BaudRate,
                           int8u *pDataBits, int8u *pParity, int8u *pStopBits, bool *pUseCts );


/** Set the configuration for the serial port */
bool LinuxSerialSetConfig( struct sLinuxSerialPort *pLinuxSerialPort, int32u ui32BaudRate,
                           int8u DataBits, int8u Parity, int8u StopBits, bool UseCts );


/** Read data from the serial port */
enum eAbstractSerialResult LinuxSerialRead( struct sLinuxSerialPort *pLinuxSerialPort, int8u *pui8Buf, int32u ui32Size );


/** Write data to the serial port */
enum eAbstractSerialResult LinuxSerialWrite( struct sLinuxSerialPort *pLinuxSerialPort, const int8u *pui8Buf, int32u ui32Size );


/** Flush the serial port buffers. queue_selector is one of TCIFLUSH, TCOFLUSH, TCIOFLUSH */
bool LinuxSerialFlush( struct sLinuxSerialPort *pLinuxSerialPort, int queue_selector );


/** Get the number of bytes available in the transmit buffer */
bool LinuxSerialGetTxBufferLevel( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32Level );


/** Get the size in bytes of the transmit buffer */
bool LinuxSerialGetTxBufferSize( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32Level );


/** Get the number of bytes available in the receive buffer */
bool LinuxSerialGetRxBufferLevel( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32Level );


/** Get the size in bytes of the receive buffer */
bool LinuxSerialGetRxBufferSize( struct sLinuxSerialPort *pLinuxSerialPort, int32u *pui32Level );


/**
 * Get the state of the CTS pin (regardless of whether CTS blocking is enabled).
 * Returns 1 == Clear to send data, 0 == blocking
 */
bool LinuxSerialGetCtsState( struct sLinuxSerialPort *pLinuxSerialPort, bool *pbCts );


/**
 * Get the state of the RTS pin
 *
 * @param pbRts set to true if RTS asserted (RS232 +ve, TTL 0v), false otherwise
 */
bool LinuxSerialGetRtsState( struct sLinuxSerialPort *pLinuxSerialPort, bool *pbRts );


/**
 * Get the state of the DTR pin
 *
 * @param pbDtr set to true if DTR asserted (RS232 +ve, TTL 0v), false otherwise
 */
bool LinuxSerialGetDtrState( struct sLinuxSerialPort *pLinuxSerialPort, bool *pbDtr );


/**
 * Set the state of the RTS pin
 *
 * @param bRts true to assert RTS (RS232 +ve, TTL 0v), false otherwise
 */
bool LinuxSerialSetRtsState( struct sLinuxSerialPort *pLinuxSerialPort, bool bRts );


/**
 * Set the state of the DTR pin
 *
 * @param bDts true to assert DTR (RS232 +ve, TTL 0v), false otherwise
 */
bool LinuxSerialSetDtrState( struct sLinuxSerialPort *pLinuxSerialPort, bool bDtr );


/**
 * Enable the use of the kernel RS485 transmitter enable using RTS support via TIOCSRS485
 *
 * @param pLinuxSerialPort is the serial port
 * @param enable is used to enable or disable RS485 transmitter enable using RTS
 * @param assert_on_transmit will assert RTS while transmitting if set, otherwise will de-assert RTS while transmitting
 */
bool LinuxSerialSetRS485Rts(struct sLinuxSerialPort *pLinuxSerialPort, bool enable, bool assert_on_transmit);


#endif /* LINUXSERIAL_H_ */
