/*
 * sio_helper.c
 *
 *  Created on: Sep 5, 2014
 *      Author: ian
 */
#include <stdbool.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <linux/sockios.h>
#include <sys/ioctl.h>
#include <rinhw/rio.h>
#include <rinhw/timemath.h>

#include "sio_helper.h"


/** Close the socket */
void sio_helper_close(int *sockfd)
{
  if (*sockfd >= 0)
  {
    close(*sockfd);
    *sockfd = -1;
  }
}


/** Set blocking on a socket */
int sio_helper_setblocking(int sockfd, bool blocking)
{
  int flags;

  flags = fcntl(sockfd, F_GETFL, 0);
  if (flags < 0)
  {
    return RIO_ERROR;
  }

  if (blocking)
  {
    flags &= ~O_NONBLOCK;
  }
  else
  {
    flags |= O_NONBLOCK;
  }
  return (fcntl(sockfd, F_SETFL, flags) < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/** Set a timeout on a socket. 0 for indefinite timeout */
int sio_helper_settimeout(int sockfd, bool send_not_recv, unsigned short timeoutticks)
{
  struct timeval tv;
  int result;
  const int optname = (send_not_recv) ? SO_SNDTIMEO : SO_RCVTIMEO;

  OSTicksToTimeval(&tv, timeoutticks);
  result = setsockopt(sockfd, SOL_SOCKET, optname, &tv, sizeof(tv));

  return (result < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/**
 * Function to receive len bytes (recv doesn't guarantee to return len bytes, unless used
 * with MSG_WAITALL which isn't defined on Windows < Windows server 2003)
 */
int sio_helper_read_all(int sockfd, void *buf, size_t len)
{
  int retval;
  size_t numrecvd;
  char *pBuf;

  pBuf = (char *) buf;
  for (numrecvd = 0; numrecvd < len; )
  {
    retval = read(sockfd, &pBuf[numrecvd], len - numrecvd);
    if (retval <= 0)
    {
      return retval;
    }
    else
    {
      numrecvd += retval;
    }
  }
  return numrecvd;
}


/** Function to send len bytes (send doesn't guarantee to send len bytes) */
int sio_helper_write_all(int sockfd, const void *buf, size_t len)
{
  int retval;
  size_t numsent;
  const char *pBuf;

  pBuf = (char *) buf;
  for (numsent = 0; numsent < len; )
  {
    retval = write(sockfd, &pBuf[numsent], len - numsent);
    if (retval <= 0)
    {
      return retval;
    }
    else
    {
      numsent += retval;
    }
  }
  return numsent;
}


/** Get the buffer size of a socket */
int sio_helper_getbuffersize(int sockfd, bool send_not_recv)
{
  int bufsize;
  socklen_t len;
  int res;
  const int optname = (send_not_recv) ? SO_SNDBUF : SO_RCVBUF;

  if (sockfd < 0)
  {
    return 0;
  }

  len = sizeof(bufsize);
  res = getsockopt(sockfd, SOL_SOCKET, optname, (void *) &bufsize, &len);
  if (res < 0)
  {
    bufsize = 0;
  }
  return bufsize;
}


/** Return the number of bytes available in either the receive or transmit buffers for a socket */
int sio_helper_getbufferlevel(int sockfd, bool send_not_recv)
{
  int buflevel;
  int res;
  int ioctl_type = (send_not_recv) ? SIOCOUTQ : SIOCINQ;

  if (sockfd < 0)
  {
    return 0;
  }
  res = ioctl(sockfd, ioctl_type, &buflevel);
  if (res < 0)
  {
    buflevel = 0;
  }
  return buflevel;
}
