/*
 * sio_local.c
 *
 *  Created on: 11/02/2009
 *      Author: ianb
 */
#include <stdarg.h>
#include <stdlib.h>
#include <rinhw/rstdio.h>
#include <rinhw/rioctl.h>
#include "abstractserial.h"
#include <rinhw/rio.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>

#include <rinhw/sio_local_comport.h>


/** Local serial port support information */
struct s_sio_local
{
  struct sAbstractSerialPort serial_port;       /*!< Serial port */
  struct s_sio_local_rs422 serial_rs422;        /*!< RS422 control information */
  bool enable_rs232_not_rs422;                  /*!< Which one is enabled */
  bool rs422_termination_state;                 /*!< State of the rs422 terminations */
  bool is_gadget_serial;                        /*!< Set to true if this is a Linux USB gadget serial port (needs some funny handling) */
  t_ioctl_siocfg siocfg;                        /*!< Keep a copy of the serial configuration for gadget serial, because we need to re-apply it when we open the device */
};


/** Set RTS/GPIO to appropriate state, or enable RS485 RTS mode in the kernel as required by mode */
static bool sio_set_rs422_transmit_enable(struct s_sio_local *sio_local, bool enable)
{
  switch (sio_local->serial_rs422.rs422_mode)
  {
    case slrm_rs422_gpio:
      if (NULL != sio_local->serial_rs422.sio_set_rs422_transmitenable)
      {
        if (sio_local->serial_rs422.sio_set_rs422_transmitenable(sio_local->serial_rs422.hw_data, enable) < 0)
          return false;
      }
      break;

    case slrm_rs422_rts_asserted:
      if (!AbstractSerialSetRtsState(&sio_local->serial_port, enable))
        return false;
      break;

    case slrm_rs422_rts_deasserted:
      if (!AbstractSerialSetRtsState(&sio_local->serial_port, !enable))
        return false;
      break;

    case slrm_rs485_rts_asserted:
      /* Turn on/off RS485 RTS mode in the kernel */
      if (!AbstractSerialSetRS485Rts(&sio_local->serial_port, enable, enable))
      {
        /* If this fails, because not supported, then just assert RTS so at least RS422 will work */
        if (!AbstractSerialSetRtsState(&sio_local->serial_port, enable))
          return false;
      }
      else
      {
        /* Set RTS deasserted so transmit enable is disabled as an initial state/leave it disabled if disabling */
        if (!AbstractSerialSetRtsState(&sio_local->serial_port, false))
          return false;
      }
      break;

    case slrm_rs485_rts_deasserted:
      /* Turn on/off RS485 RTS mode in the kernel */
      if (!AbstractSerialSetRS485Rts(&sio_local->serial_port, enable, !enable))
      {
        /* If this fails, because not supported, then just assert RTS so at least RS422 will work */
        if (!AbstractSerialSetRtsState(&sio_local->serial_port, !enable))
          return false;
      }
      else
      {
        /* Set RTS asserted so transmit enable is disabled as an initial state/leave it disabled if disabling */
        if (!AbstractSerialSetRtsState(&sio_local->serial_port, true))
          return false;
      }
      break;

    case slrm_no_rs422:
    default:
      /* Nothing to do */
      break;
  }
  return true;
}


/** Configure RS232 & RS422 on the serial port, also controls RS422 transmit enable and terminations */
static bool sio_set_rs232_not_rs422(struct s_sio_local *sio_local)
{
  bool rs422_terms;

  /* Switch between RS232 & RS422 if enabled and as required */
  if (NULL != sio_local->serial_rs422.sio_select_rs232_not_rs422)
  {
    if (sio_local->serial_rs422.sio_select_rs232_not_rs422(sio_local->serial_rs422.hw_data, sio_local->enable_rs232_not_rs422) < 0)
      return false;
  }

  /* Figure out the state of the RS422 terminations */
  rs422_terms = !sio_local->enable_rs232_not_rs422 && sio_local->rs422_termination_state;

  /* Enable RS422 terminations if present */
  if (NULL != sio_local->serial_rs422.sio_set_rs422_termination)
  {
    if (sio_local->serial_rs422.sio_set_rs422_termination(sio_local->serial_rs422.hw_data, rs422_terms) < 0)
      return false;
  }

  /* Enable RS485/RS422 transmitter if necessary */
  if (!sio_set_rs422_transmit_enable(sio_local, !sio_local->enable_rs232_not_rs422))
    return false;

  return true;
}


/** Configure the serial port */
static bool sio_set_config(struct s_sio_local *sio_local, t_ioctl_siocfg *siocfg_ptr)
{
  /* Update local copy */
  if (siocfg_ptr != &sio_local->siocfg)
    sio_local->siocfg = *siocfg_ptr;

  /* Configure serial port */
  if (!AbstractSerialSetConfig(&sio_local->serial_port, siocfg_ptr->baud, siocfg_ptr->data_bits,
                               siocfg_ptr->parity_bits, siocfg_ptr->stop_bits, siocfg_ptr->use_cts))
  {
    return false;
  }

  /* Keep local copy of termination state */
  sio_local->rs422_termination_state = siocfg_ptr->use_terminations;

  /* Configure RS232 or RS422 */
  if (!sio_set_rs232_not_rs422(sio_local))
    return false;

  /* Success */
  return true;
}


static bool sio_get_config(struct s_sio_local *sio_local, t_ioctl_siocfg *siocfg_ptr)
{
  int32u ui32BaudRate;
  int8u ui8DataBits;
  int8u ui8ParityBits;
  int8u ui8StopBits;
  bool bUseCts;

  if (!AbstractSerialGetConfig(&sio_local->serial_port, &ui32BaudRate, &ui8DataBits, &ui8ParityBits, &ui8StopBits, &bUseCts))
    return false;

  siocfg_ptr->baud = ui32BaudRate;
  siocfg_ptr->data_bits = ui8DataBits;
  siocfg_ptr->parity_bits = ui8ParityBits;
  siocfg_ptr->stop_bits = ui8StopBits;

  siocfg_ptr->use_cts = bUseCts;
  siocfg_ptr->use_terminations = sio_local->rs422_termination_state;

  /* Update local copy */
  if (siocfg_ptr != &sio_local->siocfg)
    sio_local->siocfg = *siocfg_ptr;

  return true;
}


/** Try to open the serial port if it's a gadget serial port, returns true if open */
static bool sio_gadget_establish(struct s_sio_local *sio_local, bool blocking, unsigned short timeout)
{
  uint32_t start_time;
  bool result;
  bool isopen;

  if (sio_local->is_gadget_serial)
  {
    if (AbstractSerialIsOpen(&sio_local->serial_port, &isopen))
    {
      if (!isopen)
      {
        if (!blocking)
        {
          /* Not blocking, try to open once */
          result = AbstractSerialOpen(&sio_local->serial_port);
        }
        else
        {
          /* Blocking, try until timeout, with a time delay to stop CPU usage */
          for (start_time = OSTimeGet(), result = false; OSTimeGet() < start_time + timeout; )
          {
            if (AbstractSerialOpen(&sio_local->serial_port))
            {
              result = true;
              break;
            }
            OSTimeDly(OS_TICKS_PER_SEC / 100);
          }
        }

        if (result)
          (void) sio_set_config(sio_local, &sio_local->siocfg);
      }
      else
      {
        /* Already open */
        result = true;
      }
    }
    else
    {
      /* Fail of AbstractSerialIsOpen() */
      result = false;
    }
  }
  else
  {
    /* Not a gadget serial */
    result = true;
  }
  return result;
}


/** Close a gadget serial port because it's been unplugged. Next operation will try to re-establish */
static void sio_gadget_close(struct s_sio_local *sio_local)
{
  bool isopen;

  if (sio_local->is_gadget_serial)
  {
    if (AbstractSerialIsOpen(&sio_local->serial_port, &isopen) && isopen)
    {
      (void) AbstractSerialClose(&sio_local->serial_port);
    }
  }
}


/** Initialise a serial device */
static int sio_localcom_init(struct t_io_dev *io_dev)
{
  struct s_sio_local *sio_local = (struct s_sio_local *) io_dev->dev_data;
  t_ioctl_siocfg siocfg;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("sio_localcom_init called\n");
#endif

  /* Configure the serial port to defaults */
  siocfg.baud = SIO_DEF_BAUD;
  siocfg.data_bits = SIO_DEF_DATA_BITS;
  siocfg.parity_bits = SIO_DEF_PARITY_BITS;
  siocfg.stop_bits = SIO_DEF_STOP_BITS;
  siocfg.use_cts = false;
  siocfg.use_terminations = SIO_DEF_TERMINATIONS;
  sio_local->enable_rs232_not_rs422 = SIO_DEF_RS232_NOT_RS422;
  if (!sio_set_config(sio_local, &siocfg) && !sio_local->is_gadget_serial)
    return RIO_ERROR;

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

  return RIO_SUCCESS;
}


/** Destroy serial device */
static int sio_localcom_destroy(struct t_io_dev *io_dev)
{
  struct s_sio_local *sio_local = (struct s_sio_local *) io_dev->dev_data;

  /* Make sure RS485 transmitter is disabled */
  (void) sio_set_rs422_transmit_enable(sio_local, false);

  /* Close the serial port */
  (void) AbstractSerialClose(&sio_local->serial_port);

  /* Shut down the serial port */
  AbstractSerialShutdown(&sio_local->serial_port);

  /* Free up memory */
  free(sio_local);
  io_dev->dev_data = NULL;

  return RIO_SUCCESS;
}


/** Read from the serial port */
static int sio_localcom_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_sio_local *sio_local = (struct s_sio_local *) io_dev->dev_data;
  unsigned char *ptr = (unsigned char *) buf;
  enum eAbstractSerialResult res;
  int result;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localcom_read called\n");
#endif

  /* Try to establish a gadget serial connection */
  if (!sio_gadget_establish(sio_local, true, timeout))
  {
    result = RIO_ERROR;
  }
  else
  {
    /* Set the read timeout */
    if (!AbstractSerialSetReadTimeout( &sio_local->serial_port, timeout * 1000 / OS_TICKS_PER_SEC))
    {
      sio_gadget_close(sio_local);
      result = RIO_ERROR;
    }
    else
    {
      /* Do the read */
      res = AbstractSerialRead(&sio_local->serial_port, ptr, len);
      switch (res)
      {
        case asrSuccess:
          result = (int) len;
          break;
        case asrTimeout:
          result = RIO_ERROR;
          break;

        /*
         * This bit of code is aimed toward the USB gadget serial port.
         * When unplugged from the host, the read() call above will return EOF.
         * The only way to recover from this is to close the port and re-open it.
         *
         * On 2.6.26: Open will fail until the host connects
         * On 2.6.29: Open will succeed straight away
         */
        default:
        case asrEndOfFile:
        case asrError:
          sio_gadget_close(sio_local);
          result = RIO_ERROR;
          break;
      }
    }
  }
  return result;
}


/** Write to the serial port */
static int sio_localcom_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_sio_local *sio_local = (struct s_sio_local *) io_dev->dev_data;
  unsigned char *ptr = (unsigned char *) buf;
  enum eAbstractSerialResult res;
  int result;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localcom_write called\n");
#endif

  /* Try to establish a gadget serial connection */
  if (!sio_gadget_establish(sio_local, true, timeout))
  {
    /* Disregard result, and pretend that the write still occurred */
    result = (int) len;
  }
  else
  {
    if (!AbstractSerialSetWriteTimeout(&sio_local->serial_port, timeout * 1000 / OS_TICKS_PER_SEC))
    {
      sio_gadget_close(sio_local);
      result = RIO_ERROR;
    }
    else
    {
      /* Do the write */
      res = AbstractSerialWrite(&sio_local->serial_port, ptr, len);
      switch (res)
      {
        case asrSuccess:
          result = (int) len;
          break;
        case asrTimeout:
          result = RIO_ERROR;
          break;

        /*
         * This bit of code is aimed toward the USB gadget serial port.
         * When unplugged from the host, the write() call above will return error.
         * The only way to recover from this is to close the port and re-open it.
         *
         * On 2.6.26: Open will fail until the host connects
         * On 2.6.29: Open will succeed straight away
         */
        default:
        case asrEndOfFile:
        case asrError:
          sio_gadget_close(sio_local);
          result = RIO_ERROR;
          break;
      }
    }
  }

  return result;
}


/** Get the number of characters in the receive buffer */
static long sio_localcom_filelength(struct t_io_dev *io_dev)
{
  struct s_sio_local *sio_local = (struct s_sio_local *) io_dev->dev_data;
  int32u ui32Level;
  int result;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localcom_filelength called\n");
#endif

  /* Try to establish a gadget serial connection, but don't block waiting for one */
  if (!sio_gadget_establish(sio_local, false, 0))
  {
    result = RIO_ERROR;
  }
  else if (AbstractSerialGetRxBufferLevel(&sio_local->serial_port, &ui32Level))
  {
    result = (long) ui32Level;
  }
  else
  {
    sio_gadget_close(sio_local);
    result = RIO_ERROR;
  }
  return result;
}


/*
 * sio_ioctl -
 *
 * Configure etc the serial port
 */
static int sio_localcom_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_sio_local *sio_local = (struct s_sio_local *) io_dev->dev_data;
  t_ioctl_siocfg *siocfg_ptr;
  int32u ui32Level;
  bool bLevel;
  unsigned short *word_ptr;
  int result = RIO_SUCCESS;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localcom_ioctl called\n");
#endif

  /* Try to establish a gadget serial connection, but don't block waiting for one */
  if (!sio_gadget_establish(sio_local, false, 0))
  {
    return RIO_ERROR;
  }

  /* IOCTL function dispatch */
  switch (func)
  {
    /* Get/set configuration */
    case IOCTL_SER_SET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_SET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      result = (sio_set_config( sio_local, siocfg_ptr )) ? RIO_SUCCESS : RIO_ERROR;
      break;
    case IOCTL_SER_GET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_GET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      result = (sio_get_config(sio_local, siocfg_ptr)) ? RIO_SUCCESS : RIO_ERROR;
      break;

    /* Get buffer levels and sizes */
    case IOCTL_SER_GET_RX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      result = (AbstractSerialGetRxBufferLevel(&sio_local->serial_port, &ui32Level)) ? RIO_SUCCESS : RIO_ERROR;
      *word_ptr = ui32Level;
      break;
    case IOCTL_SER_GET_TX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      result = (AbstractSerialGetTxBufferLevel(&sio_local->serial_port, &ui32Level)) ? RIO_SUCCESS : RIO_ERROR;
      *word_ptr = ui32Level;
      break;
    case IOCTL_SER_GET_RX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      result = (AbstractSerialGetRxBufferSize(&sio_local->serial_port, &ui32Level)) ? RIO_SUCCESS : RIO_ERROR;
      *word_ptr = ui32Level;
      break;
    case IOCTL_SER_GET_TX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      result = (AbstractSerialGetTxBufferSize(&sio_local->serial_port, &ui32Level)) ? RIO_SUCCESS : RIO_ERROR;
      *word_ptr = ui32Level;
      break;

    /* Get the level of the CTS pin */
    case IOCTL_SER_GET_CTS_STATE:
      result = (AbstractSerialGetCtsState(&sio_local->serial_port, &bLevel)) ? RIO_SUCCESS : RIO_ERROR;
      *va_arg(*arglist, unsigned char *) = bLevel;
      break;

    /* Get the number of receive errors on this port */
    case IOCTL_SER_GET_RX_ERR_COUNT:
      *va_arg(*arglist, unsigned long *) = 0; /* TODO */
      break;

    /* Switch between RS232 & RS422 */
    case IOCTL_SER_SET_RS232_NOT_RS422:
      sio_local->enable_rs232_not_rs422 = (0 != va_arg(*arglist, int));
      result = (sio_set_rs232_not_rs422(sio_local)) ? RIO_SUCCESS : RIO_ERROR;
      break;

    /* Flush the transmit buffer */
    case IOCTL_SER_FLUSH_TX_BUF:
      result = (AbstractSerialFlush(&sio_local->serial_port, FlushOutput)) ? RIO_SUCCESS : RIO_ERROR;
      break;

    /* Flush the receive buffer */
    case IOCTL_SER_FLUSH_RX_BUF:
      result = (AbstractSerialFlush(&sio_local->serial_port, FlushInput)) ? RIO_SUCCESS : RIO_ERROR;
      break;

    /* Get the RTS pin state */
    case IOCTL_SER_GET_RTS_STATE:
      result = (AbstractSerialGetRtsState(&sio_local->serial_port, &bLevel)) ? RIO_SUCCESS : RIO_ERROR;
      *va_arg(*arglist, int *) = bLevel;
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_GET_RTS_STATE result %d, level %u\n", result, (unsigned int) bLevel);
#endif
      break;

    /* Set the RTS pin state */
    case IOCTL_SER_SET_RTS_STATE:
      bLevel = (0 != va_arg(*arglist, int));
      result = (AbstractSerialSetRtsState(&sio_local->serial_port, bLevel)) ? RIO_SUCCESS : RIO_ERROR;
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_SET_RTS_STATE result %d, level %u\n", result, (unsigned int) bLevel);
#endif
      break;

    /* Get the DTR pin state */
    case IOCTL_SER_GET_DTR_STATE:
      result = (AbstractSerialGetDtrState(&sio_local->serial_port, &bLevel)) ? RIO_SUCCESS : RIO_ERROR;
      *va_arg(*arglist, int *) = bLevel;
      break;

    /* Set the DTR pin state */
    case IOCTL_SER_SET_DTR_STATE:
      bLevel = (0 != va_arg(*arglist, int));
      result = (AbstractSerialSetDtrState(&sio_local->serial_port, bLevel)) ? RIO_SUCCESS : RIO_ERROR;
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = sio_local->serial_port.m_nSerFd;
      result = RIO_SUCCESS;
      break;

      /* Unknown function */
/*    case IOCTL_SER_SET_RING_NETWORK: This is unsupported on the onboard serial ports */
    default:
      result = RIO_ERROR;
      break;
  }

  /* Close gadget connection on failure */
  if (RIO_SUCCESS != result)
    sio_gadget_close(sio_local);

  return result;
}


/**
 * Create a serial port connected to local serial port hardware.
 *
 * @param io_dev is an uninitialised io_dev structure to be fill out as a serial port
 * @param port_name is the serial port device name (eg /dev/ttyS0, COM1)
 * @param rs422_info is control information if this is a RS422 port. Driver takes a local copy of this struct. Can be NULL if not RS422 port
 * @param is_gadget_serial should be set to true if this is a gadget serial port, because we have to do some funny things to deal with these ports
 */
int rio_create_device_sio_localcom(struct t_io_dev *io_dev, const char *port_name, struct s_sio_local_rs422 *rs422_info, bool is_gadget_serial)
{
  struct s_sio_local *sio_local;
  int result;

  /* Allocate serial port here, sio_localcom_destroy() frees it */
  sio_local = malloc(sizeof(struct s_sio_local));
  if (NULL == sio_local)
  {
    result = RIO_ERROR;
    goto rcdsl_fail;
  }

  /* Initialise the serial port. AbstractSerialInit() takes it's own copy of pSerialPort->m_pszPortName */
  sio_local->is_gadget_serial = is_gadget_serial;
  sio_local->serial_port.m_nSerFd = -1;
  sio_local->serial_port.m_pszPortName = NULL;
  if (!AbstractSerialInit(&sio_local->serial_port, port_name))
  {
    result = RIO_ERROR;
    goto rcdsl_fail;
  }

  /* Try to open, on USB gadget serials this may fail, but that's ok because unless the other end is connected, it will fail. We'll try again later */
  if (!sio_local->is_gadget_serial && !AbstractSerialOpen(&sio_local->serial_port))
  {
    result = RIO_ERROR;
    goto rcdsl_fail;
  }

  /* Setup RS422 if requested */
  if (NULL != rs422_info)
  {
    sio_local->serial_rs422 = *rs422_info;
  }
  else
  {
    sio_local->serial_rs422.hw_data = NULL;
    sio_local->serial_rs422.rs422_mode = slrm_no_rs422;
    sio_local->serial_rs422.sio_select_rs232_not_rs422 = NULL;
    sio_local->serial_rs422.sio_set_rs422_termination = NULL;
    sio_local->serial_rs422.sio_set_rs422_transmitenable = NULL;
  }

  sio_local->enable_rs232_not_rs422 = SIO_DEF_RS232_NOT_RS422;
  sio_local->rs422_termination_state = SIO_DEF_TERMINATIONS;

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) sio_local;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_localcom_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = sio_localcom_read;
  io_dev->dev_write = sio_localcom_write;
  io_dev->dev_filelength = sio_localcom_filelength;
  io_dev->dev_ioctl = sio_localcom_ioctl;
  io_dev->dev_destroy = sio_localcom_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  result = rregister_io(io_dev);
  if (result < 0)
  {
    result = RIO_ERROR;
    goto rcdsl_fail;
  }

  /* Success */
  return RIO_SUCCESS;

rcdsl_fail:
  if (NULL != sio_local)
  {
    AbstractSerialShutdown(&sio_local->serial_port);
    free(sio_local);
    sio_local = NULL;
  }
  return result;
}
