/*
 * sio_console.c
 *
 *  Created on: Feb 17, 2012
 *      Author: ian
 */
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <termios.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/rinendian.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/timemath.h>

#include <rinhw/sio_local_console.h>


struct s_local_console
{
  struct termios oldtermios;
  t_ioctl_siocfg pretend_serial_cfg;
};


static void sio_local_console_restore_terminal_settings(struct s_local_console *lc)
{
  /* Apply saved settings */
  tcsetattr(STDIN_FILENO, TCSANOW, &lc->oldtermios);
}


static void sio_local_console_disable_waiting_for_enter(struct s_local_console *lc)
{
  struct termios newt;

  /* Make terminal read 1 char at a time */
  tcgetattr(STDIN_FILENO, &lc->oldtermios);  /* Save terminal settings */
  newt = lc->oldtermios;  /* Init new settings */
  newt.c_lflag &= ~(ICANON | ECHO);  /* Change settings */
  tcsetattr(STDIN_FILENO, TCSANOW, &newt);  /* Apply settings */
}


/** Get the number of bytes available in the receive buffer */
static int sio_local_console_get_rx_buf_level(struct s_local_console *lc, int *level)
{
  int bytes_avail;
  int res;

  res = ioctl(STDIN_FILENO, FIONREAD, &bytes_avail);
  *level = (res >= 0) ? bytes_avail : 0;
  return (res >= 0);
}

#if 0
int sio_local_console_is_char_avail(int16u timeout)
{
  fd_set readfds;
  struct timeval tv;
  int select_retval;
  bool res = false;

  FD_ZERO(&readfds);
  FD_SET(STDIN_FILENO, &readfds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;

  select_retval = select(STDIN_FILENO + 1, &readfds, NULL, NULL, &tv);
  if (select_retval < 0)pui8Buf
  {
    //perror("select");
    res = false;
  }
  else if (select_retval > 0)
  {
    if (FD_ISSET(STDIN_FILENO, &readfds))
    {
      res = true;
    }
  }
  return res;
}


bool get_stdin_char(char *ch)
{
  bool result;

  if (is_char_avail())
  {
    read(STDIN_FILENO, ch, sizeof(*ch));
    result = true;
  }
  else
  {
    result = false;
  }
  return result;
}


bool get_stdin_char_block(char *ch)
{
  bool result;

  read(STDIN_FILENO, ch, sizeof(*ch));
  result = true;
  return result;
}
#endif


/** Initialise serial device */
static int sio_local_console_init(struct t_io_dev *io_dev)
{
  struct s_local_console *lc = io_dev->dev_data;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_console_init called\n");
#endif

  /* Disable waiting for enter */
  sio_local_console_disable_waiting_for_enter(lc);

  /* Set defaults for serial port */
  lc->pretend_serial_cfg.baud = SIO_DEF_BAUD;
  lc->pretend_serial_cfg.data_bits = SIO_DEF_DATA_BITS;
  lc->pretend_serial_cfg.parity_bits = SIO_DEF_PARITY_BITS;
  lc->pretend_serial_cfg.stop_bits = SIO_DEF_STOP_BITS;
  lc->pretend_serial_cfg.use_cts = SIO_DEF_USECTS;
  lc->pretend_serial_cfg.use_terminations = SIO_DEF_TERMINATIONS;

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

  return RIO_SUCCESS;
}


/** Destroy serial device */
static int sio_local_console_destroy(struct t_io_dev *io_dev)
{
  struct s_local_console *lc = io_dev->dev_data;
  int result;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_console_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  io_dev->dev_data = NULL;
  if (NULL != lc)
  {
    /* Restore terminal settings */
    sio_local_console_restore_terminal_settings(lc);

    /* Free up stuff we've allocated */
    free(lc);
  }

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_console_destroy: result %d\n", result);
#endif
  return result;
}


/** Read from the serial port */
static int sio_local_console_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  /* struct s_local_console *lc = io_dev->dev_data; */

  size_t bytesread;
  int res;
  struct timeval timeoutsrc;
  struct timeval timeoutuse;
  fd_set readfds;
  int maxfd;
  char *charbuf = buf;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_console_read called\n");
#endif

  /* Work out timeout */
  OSTicksToTimeval(&timeoutsrc, timeout);

  bytesread = 0;
  res = 0;
  while (bytesread < len)
  {
    if (0 != timeout)
    {
      /* Wait for data or timeout, we use two copies of timeout because select changes it */
      timeoutuse = timeoutsrc;
      FD_ZERO(&readfds);
      FD_SET(STDIN_FILENO, &readfds);
      maxfd = STDIN_FILENO + 1;
      res = select(maxfd, &readfds, NULL, NULL, &timeoutuse);
      if (res < 0)
      {
        /* Error */
        break;
      }
      if (0 == res)
      {
        /* Timeout */
        break;
      }
    }

    /* Got data, do a read */
    res = read(STDIN_FILENO, &charbuf[bytesread], len - bytesread);
    if (res < 0)
    {
      /* <0 means error */
      break;
    }
    if (0 == res)
    {
      /* 0 means EOF */
      break;
    }
    bytesread += res;
  }

  /* Return the result */
  return (res >= 0 && bytesread == len) ? len : RIO_ERROR;
}


/** Write to the serial port */
static int sio_local_console_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  /* struct s_local_console *lc = io_dev->dev_data; */
  size_t sent;
  ssize_t res;
  struct timeval timeoutsrc;
  struct timeval timeoutuse;
  fd_set writefds;
  int maxfd;
  const char *charbuf = buf;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_console_write called\n");
#endif

  /* Work out timeout */
  OSTicksToTimeval(&timeoutsrc, timeout);

  sent = 0;
  res = 0;
  while (sent < len)
  {
    if (0 != timeout)
    {
      /* Wait for room in write buffer or timeout, we use two copies of timeout because select changes it */
      timeoutuse = timeoutsrc;
      FD_ZERO(&writefds);
      FD_SET(STDOUT_FILENO, &writefds);
      maxfd = STDOUT_FILENO + 1;
      res = select(maxfd, NULL, &writefds, NULL, &timeoutuse);
      if (res <= 0)
      {
        /* Error or timeout */
        break;
      }
    }

    /* Ok to write */
    res = write(STDOUT_FILENO, &charbuf[sent], len - sent );
    if ( res < 0 )
    {
      break;
    }
    sent += res;
  }

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_console_write: wrote stdout, sent %u, len %d\n", sent, len);
#endif

  /* Return the result */
  return (res >= 0 && sent == len) ? len : RIO_ERROR;
}


/** Get the number of characters in the receive buffer */
static long sio_local_console_filelength(struct t_io_dev *io_dev)
{
  struct s_local_console *lc = io_dev->dev_data;
  long result;
  int level;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_console_filelength called\n");
#endif

  if (sio_local_console_get_rx_buf_level(lc, &level) < 0)
  {
    result = RIO_ERROR;
  }
  else
  {
    result = level;
  }
  return result;
}


/** Flush (discard) any incoming data from the console */
static int sio_local_console_flush_incoming(struct s_local_console *lc)
{
  int level;
  char buffer;

  if (sio_local_console_get_rx_buf_level(lc, &level) < 0)
    level = -1;
  while (level > 0)
  {
    if (read(STDIN_FILENO, &buffer, sizeof(buffer)) == sizeof(buffer))
      level--;
    else
      level = -1;
  }
  return (level < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/** Configure etc the serial port */
static int sio_local_console_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_local_console *lc = io_dev->dev_data;
  t_ioctl_siocfg *siocfg_ptr;
  unsigned short *word_ptr;
  int result = RIO_SUCCESS;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_console_ioctl called\n");
#endif

  /* IOCTL function dispatch */
  switch (func)
  {
    case IOCTL_SER_SET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_SET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      lc->pretend_serial_cfg = *siocfg_ptr;
      break;

    case IOCTL_SER_GET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_GET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      *siocfg_ptr = lc->pretend_serial_cfg;
      break;

    /* Get buffer levels and sizes */
    case IOCTL_SER_GET_RX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = 0;
      break;
    case IOCTL_SER_GET_TX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = 0;
      break;
    case IOCTL_SER_GET_RX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = 100;
      break;
    case IOCTL_SER_GET_TX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = 100;
      break;

    /* Get the level of the cts pin */
    case IOCTL_SER_GET_CTS_STATE:
      *va_arg(*arglist, unsigned char *) = 1;
      break;

    /* Get the number of receive errors on this port */
    case IOCTL_SER_GET_RX_ERR_COUNT:
      *va_arg(*arglist, unsigned long *) = 0;
      break;

    /* Flush the transmit buffer */
    case IOCTL_SER_FLUSH_TX_BUF:
      result = RIO_SUCCESS; /* Not really much we can do here, so let's just say it worked */
      break;

    /* Flush the receive buffer */
    case IOCTL_SER_FLUSH_RX_BUF:
      result = sio_local_console_flush_incoming(lc);
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      /* Not perfect, what about writes STDOUT_FILENO */
      *va_arg(*arglist, int *) = STDIN_FILENO;
      result = RIO_SUCCESS;
      break;

    /* Unknown function */
/*    case IOCTL_SER_SET_RING_NETWORK: This is unsupported on the onboard serial ports */
    default:
      result = RIO_ERROR;
      break;
  }
  return result;
}


/** Create a serial port connected to the local console */
int rio_create_device_sio_localconsole(struct t_io_dev *io_dev)
{
  struct s_local_console *lc;

  /* Allocate serial port file info here, sio_localcom_destroy() frees it */
  lc = malloc(sizeof(struct s_local_console));
  if (NULL == lc)
  {
    return RIO_ERROR;
  }

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) lc;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_local_console_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = sio_local_console_read;
  io_dev->dev_write = sio_local_console_write;
  io_dev->dev_filelength = sio_local_console_filelength;
  io_dev->dev_ioctl = sio_local_console_ioctl;
  io_dev->dev_destroy = sio_local_console_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);
}
