/*
 * sio_local_dgram.c
 *
 *  Created on: Apr 4, 2012
 *      Author: ian
 */
#if defined( WIN32 )
  #include <winsock.h>
#elif defined(__linux__)
  #include <sys/types.h>
  #include <errno.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <linux/sockios.h>
  #include <fcntl.h>
  #include <sys/ioctl.h>
  #include <string.h>
#endif
#include <strings.h>
#include <stdarg.h>
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinhw/timemath.h>

#include <rinhw/sio_local_dgram.h>


/** Device data for a serial port mapped to a socket */
struct sDatagramSerialPort
{
  int nSocketPort;                /* The socket port to listen on */
  int nSocketFd;                  /* The UDP socket file descriptor */
  t_ioctl_siocfg siocfg;          /* The serial port configuration */
  char *outgoing_terminating_characters;
  int num_outgoing_terminating_characters;
  char *outgoing_buffer;
  int outgoing_buffer_size;
  int outgoing_buffer_posn;
  int response_addr_valid;
  struct sockaddr_in response_addr;
};


static void sio_localdgram_set_config(struct sDatagramSerialPort *pDgramSerial, t_ioctl_siocfg *siocfg_ptr)
{
  /* Take a copy of the serial configuration */
  pDgramSerial->siocfg = *siocfg_ptr;
}


static void sio_localdgram_get_config(struct sDatagramSerialPort *pDgramSerial, t_ioctl_siocfg *siocfg_ptr)
{
  *siocfg_ptr = pDgramSerial->siocfg;
}


/** Close the connection socket, but leave the listen socket open */
static void sio_localdgram_connection_close(struct sDatagramSerialPort *pDgramSerial)
{
  if (pDgramSerial->nSocketFd >= 0)
  {
    close(pDgramSerial->nSocketFd);
    pDgramSerial->nSocketFd = -1;
  }
}


/** Change the UDP port this serial port is bound to */
static int sio_localdgram_change_bound_port(struct sDatagramSerialPort *pDgramSerial, int newport)
{
  int sktfd;
  struct sockaddr_in servaddr;
  int result;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_change_bound_port called, new port:%d\n", newport);
#endif

  /* Shutdown current sockets */
  sio_localdgram_connection_close(pDgramSerial);

  /* This is our new listen port */
  pDgramSerial->nSocketPort = newport;

  /* Setup new socket */
  sktfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (sktfd >= 0)
  {
    if (fcntl(sktfd, F_SETFD, FD_CLOEXEC) < 0)
    {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_localdgram_change_bound_port: error setting FD_CLOEXEC for port %d (errno %d)\n", pSktSerial->nSocketNumber, errno);
#endif
    }

    /* Bind port to socket on any interface */
    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(pDgramSerial->nSocketPort);
    if (0 == bind(sktfd, (struct sockaddr *) &servaddr, sizeof(servaddr)))
    {
      pDgramSerial->nSocketFd = sktfd;
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_localdgram_change_bound_port: socket configured correctly on port %d\n", pDgramSerial->nSocketPort);
#endif
    }
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
    else
    {
      DEBUG_PRINTF("sio_localdgram_change_bound_port: bind() fail, errno: %d\n", errno);
    }
#endif
  }
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  else
  {
    DEBUG_PRINTF("sio_localdgram_change_bound_port: socket() fail, errno: %d\n", errno);
  }
#endif

  result = (pDgramSerial->nSocketFd >= 0) ? RIO_SUCCESS : RIO_ERROR;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_change_bound_port: result %d, listen sockfd %d\n", result, pDgramSerial->nSocketFd);
#endif
  return result;
}


/** Initialise a datagram serial device */
static int sio_localdgram_init(struct t_io_dev *io_dev)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  t_ioctl_siocfg siocfg;
  int result;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_init called\n");
#endif

  /* Default configuration for the serial port */
  siocfg.baud = SIO_DEF_BAUD;
  siocfg.data_bits = SIO_DEF_DATA_BITS;
  siocfg.parity_bits = SIO_DEF_PARITY_BITS;
  siocfg.stop_bits = SIO_DEF_STOP_BITS;
  siocfg.use_cts = SIO_DEF_USECTS;
  siocfg.use_terminations = SIO_DEF_TERMINATIONS;
  sio_localdgram_set_config(pDgramSerial, &siocfg);

  /* Bind to listen port */
  result = sio_localdgram_change_bound_port(pDgramSerial, pDgramSerial->nSocketPort);

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_init: result %d\n", result);
#endif
  return result;
}


/** Destroy datagram serial device */
static int sio_localdgram_destroy(struct t_io_dev *io_dev)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  int result;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* TODO force anything pending to stop */

  /* Shutdown socket */
  sio_localdgram_connection_close(pDgramSerial);

  /* Free up stuff we've allocated */
  io_dev->dev_data = NULL;
  if (NULL != pDgramSerial)
  {
    if (NULL != pDgramSerial->outgoing_terminating_characters)
    {
      free(pDgramSerial->outgoing_terminating_characters);
      pDgramSerial->outgoing_terminating_characters = NULL;
    }
    if (NULL != pDgramSerial->outgoing_buffer)
    {
      free(pDgramSerial->outgoing_buffer);
      pDgramSerial->outgoing_buffer = NULL;
    }
    free(pDgramSerial);
    pDgramSerial = NULL;
  }

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_destroy: result %d\n", result);
#endif
  return result;
}


#if defined (__WIN32__)
static int ucos_time_to_millisec( unsigned short timeout )
{
  int nTimeoutMillisec;

  nTimeoutMillisec = timeout * 1000;
  nTimeoutMillisec /= OS_TICKS_PER_SEC;
  return nTimeoutMillisec;
}
#endif /* #if defined (__WIN32__) */


#if defined(__linux__)
static void sio_localdgram_setblocking(int sockfd, bool bBlocking)
{
  int flags;

  flags = fcntl(sockfd, F_GETFL, 0);
  if (flags < 0)
  {
    return;
  }

  if (bBlocking)
  {
    flags &= ~O_NONBLOCK;
  }
  else
  {
    flags |= O_NONBLOCK;
  }
  fcntl(sockfd, F_SETFL, flags);
}
#elif defined (__WIN32__) /* #if defined(__linux__) */
static void sio_localdgram_setblocking(int sockfd, bool bBlocking)
{
  u_long arg;

  arg = (bBlocking) ? 0 : 1;
  ioctlsocket(sockfd, FIONBIO, &arg);
}
#else /* #elif defined (__WIN32__) -> #if defined(__linux__) */
  #error sio_localskt_setblocking() not defined on this platform
#endif /* #else -> #elif defined (__WIN32__) -> #if defined(__linux__) */


#if defined(__linux__)
static int sio_localdgram_setrecvtimeout(int sockfd, unsigned short timeout)
{
  struct timeval tv;
  int result;

  OSTicksToTimeval(&tv, timeout);
  result = setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, (char *) &tv,  sizeof(tv));

  return (result < 0) ? RIO_ERROR : RIO_SUCCESS;
}
#elif defined (__WIN32__) /* #if defined(__linux__) */
static int sio_localdgram_setrecvtimeout( int sockfd, unsigned short timeout )
{
  int nMilliSeconds = ucos_time_to_millisec(timeout);
  int result;

  result = setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, (char *) &nMilliSeconds,  sizeof(nMilliSeconds));

  return (result < 0) ? RIO_ERROR : RIO_SUCCESS;
}
#else /* #elif defined (__WIN32__) -> #if defined(__linux__) */
  #error sio_localskt_setblocking() not defined on this platform
#endif /* #else -> #elif defined (__WIN32__) -> #if defined(__linux__) */


#if 0
/** Get the last error */
static int sio_get_last_error(void)
{
#if defined(__linux__)
  return errno;
#elif defined(__WIN32__)
  return WSAGetLastError();
#endif
}
#endif


#if defined(__WIN32__)
#define EWOULDBLOCK WSAETIMEDOUT
#endif


/**
 * Function to receive len bytes (recv doesn't guarantee to return len bytes, unless used
 * with MSG_WAITALL which isn't defined on Windows < Windows server 2003)
 */
static int sio_localdgram_recv_all(struct sDatagramSerialPort *pDgramSerial, void *buf, size_t len)
{
  int retval;
  size_t numrecvd;
  char *pBuf;
  socklen_t sockaddr_size;

  pBuf = (char *) buf;
  for (numrecvd = 0; numrecvd < len; )
  {
    sockaddr_size = sizeof(pDgramSerial->response_addr);
    retval = recvfrom(pDgramSerial->nSocketFd,
                      &pBuf[numrecvd], len - numrecvd, 0,
                      (struct sockaddr *)&pDgramSerial->response_addr, &sockaddr_size);
    if (retval <= 0)
    {
      return retval;
    }
    else
    {
      pDgramSerial->response_addr_valid = 1;
      numrecvd += retval;
    }
  }
  return numrecvd;
}


/** Read from the datagram serial port */
static int sio_localdgram_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  int retval;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_read called\n");
#endif

  /* Check socket has been setup correctly */
  if (pDgramSerial->nSocketFd < 0)
  {
    return RIO_ERROR;
  }

  sio_localdgram_setblocking(pDgramSerial->nSocketFd, true);
  if (RIO_SUCCESS != sio_localdgram_setrecvtimeout(pDgramSerial->nSocketFd, timeout))
  {
    return RIO_ERROR;
  }
  retval = sio_localdgram_recv_all(pDgramSerial, buf, len);
#if 0
  if ( retval < 0 )
  {
    error = sio_get_last_error();
  }
  if ( ( retval < 0 && EWOULDBLOCK != error ) || 0 == retval )
  {
    sio_localdgram_connection_close( pDgramSerial );
  }
#endif
  return retval;
}


/** Is ch one of the nominated terminating characters */
static int sio_localdgram_isterminator(struct sDatagramSerialPort *pDgramSerial, char ch)
{
  int i;
  int result;

  for (i = 0, result = 0; i < pDgramSerial->num_outgoing_terminating_characters; i++)
  {
    if (ch == pDgramSerial->outgoing_terminating_characters[i])
    {
      result = 1;
      break;
    }
  }
  return result;
}


/** Send a buffer */
static int sio_localdgram_sendbuffer(struct sDatagramSerialPort *pDgramSerial)
{
  int result;

  if (0 != pDgramSerial->response_addr_valid)
  {

    result = sendto(pDgramSerial->nSocketFd,
                    pDgramSerial->outgoing_buffer, pDgramSerial->outgoing_buffer_posn,
                    0, (struct sockaddr *) &pDgramSerial->response_addr, sizeof(pDgramSerial->response_addr));
    result = (result != pDgramSerial->outgoing_buffer_posn) ? RIO_ERROR : RIO_SUCCESS;
  }
  else
  {
    /* We have no socket connected, so just discard the data and tell the caller that it was all written */
    result = RIO_SUCCESS;
  }
  pDgramSerial->outgoing_buffer_posn = 0;
  return result;
}


/** Write to the UDP datagram serial port */
static int sio_localdgram_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  int retval;
  char ch;
  const char *inbuf;
  int i;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_write called\n");
#endif

  /* Check socket has been setup correctly */
  if (pDgramSerial->nSocketFd < 0)
  {
    return RIO_ERROR;
  }

  /* Setup timeout */
  sio_localdgram_setblocking(pDgramSerial->nSocketFd, true);
  if (RIO_SUCCESS != sio_localdgram_setrecvtimeout(pDgramSerial->nSocketFd, timeout))
  {
    return RIO_ERROR;
  }

  /* Inspect each character, copy to out buffer, and write when terminating character found or buffer full */
  for (i = 0, inbuf = buf; i < len; i++, inbuf++)
  {
    ch = *inbuf;
    pDgramSerial->outgoing_buffer[pDgramSerial->outgoing_buffer_posn++] = ch;
    if (sio_localdgram_isterminator(pDgramSerial, ch) > 0 || pDgramSerial->outgoing_buffer_posn >= pDgramSerial->outgoing_buffer_size)
    {
      /* Time to send */
      if (RIO_SUCCESS != sio_localdgram_sendbuffer(pDgramSerial))
      {
        return RIO_ERROR;
      }
    }
  }

  retval = i;
  return retval;
}


#if defined(__linux__)
  #define bufsize_cast void *
#elif defined(__WIN32__)
  /* Define socklen_t on windows */
  #define socklen_t int
  #define bufsize_cast char *
#endif


static int sio_localskt_getbuffersize(int sockfd, bool bTxNotRx)
{
  int bufsize;
  socklen_t len;
  int res;
  int optname = (bTxNotRx) ? SO_SNDBUF : SO_RCVBUF;

  if ( sockfd < 0 )
  {
    return 0;
  }

  len = sizeof( bufsize );
  res = getsockopt( sockfd, SOL_SOCKET, optname, (bufsize_cast) &bufsize, &len );
  if ( res < 0 )
  {
    bufsize = 0;
  }
  return bufsize;
}


#if defined(__linux__)
/** Return the number of bytes available in either the receive or transmit buffers for a socket */
static int sio_localdgram_getbufferlevel(int sockfd, bool bTxNotRx)
{
  int buflevel;
  int res;
  int ioctl_type = (bTxNotRx) ? SIOCOUTQ : SIOCINQ;

  if (sockfd < 0)
  {
    return 0;
  }
  res = ioctl(sockfd, ioctl_type, &buflevel);
  if (res < 0)
  {
    buflevel = 0;
  }
  return buflevel;
}
#elif defined(__WIN32__) /* #if defined(__linux__) */
/** Return the number of bytes available in either the receive or transmit buffers for a socket */
static int sio_localdgram_getbufferlevel( int sockfd, bool bTxNotRx )
{
  int buflevel;
  int res;
  char *pszBuf;

  /* Check socket is legitimate */
  if (sockfd < 0)
  {
    return 0;
  }

  if (bTxNotRx)
  {
    /* Transmit buffer, have no way of telling how much data is waiting to be sent */
    buflevel = 0;
  }
  else
  {
    /* Receive buffer, use MSG_PEEK to see what's available. That's the only option we have on windows. */
    pszBuf = malloc(1000);
    if ((char *) 0 != pszBuf)
    {
      sio_localdgram_setblocking(sockfd, false);
      res = recv(sockfd, pszBuf, 1000, MSG_PEEK);
      free(pszBuf);
      buflevel = (res > 0) ? res : 0;
    }
  }
  return buflevel;
}
#endif /* #elif defined(__WIN32__) -> #if defined(__linux__) */


/** Get the number of characters in the receive buffer. We can't know this */
static long sio_localdgram_filelength(struct t_io_dev *io_dev)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_filelength called\n");
#endif

  /* Check socket has been setup correctly */
  if (pDgramSerial->nSocketFd < 0)
  {
    return RIO_ERROR;
  }

  return (long) sio_localdgram_getbufferlevel(pDgramSerial->nSocketFd, false);
}


/** Use CTS to indicate there is a valid reply to address */
static void sio_localdgram_ioctl_cts(struct sDatagramSerialPort *pSktSerial, unsigned char *cts_state)
{
  *cts_state = pSktSerial->response_addr_valid;
}


/** Configure etc the serial port */
static int sio_localdgram_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct sDatagramSerialPort *pSktSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  t_ioctl_siocfg *siocfg_ptr;
  unsigned short *word_ptr;
  int result = RIO_SUCCESS;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_ioctl called, func %d\n", func);
#endif

  /* Check socket has been setup correctly */
  if (pSktSerial->nSocketFd < 0)
  {
    return RIO_ERROR;
  }

  /* IOCTL function dispatch */
  switch (func)
  {
    case IOCTL_SER_SET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_SET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      sio_localdgram_set_config( pSktSerial, siocfg_ptr );
      break;

    case IOCTL_SER_GET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_GET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      sio_localdgram_get_config( pSktSerial, siocfg_ptr );
      break;

    /*
     * Get buffer levels and sizes
     */
    case IOCTL_SER_GET_RX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_localdgram_getbufferlevel(pSktSerial->nSocketFd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_localdgram_getbufferlevel(pSktSerial->nSocketFd, true);
      break;
    case IOCTL_SER_GET_RX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_localskt_getbuffersize(pSktSerial->nSocketFd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_localskt_getbuffersize(pSktSerial->nSocketFd, true);
      break;

    /* Get the level of the cts pin */
    case IOCTL_SER_GET_CTS_STATE:
      sio_localdgram_ioctl_cts(pSktSerial, va_arg(*arglist, unsigned char *));
      break;

    /* Get the number of receive errors on this port */
    case IOCTL_SER_GET_RX_ERR_COUNT:
      *va_arg(*arglist, unsigned long *) = 0; /* TODO */
      break;

    /* Change listen TCP port for this serial port */
    case IOCTL_SER_SET_SOCKET_PORT:
      result = sio_localdgram_change_bound_port(pSktSerial, va_arg(*arglist, int));
      break;

    /* Get the listen TCP port for this serial port */
    case IOCTL_SER_GET_SOCKET_PORT:
      *va_arg(*arglist, int *) = pSktSerial->nSocketPort;
      break;

    /* Flush the transmit buffer */
    case IOCTL_SER_FLUSH_TX_BUF:
      result = RIO_SUCCESS; /* Not really much we can do here, so let's just say it worked */
      break;

    /* Flush the receive buffer */
    case IOCTL_SER_FLUSH_RX_BUF:
      result = RIO_SUCCESS; /* Not really much we can do here, so let's just say it worked */
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = pSktSerial->nSocketFd;
      result = RIO_SUCCESS;
      break;

    /* Unknown function */
/*    case IOCTL_SER_SET_RING_NETWORK: This is unsupported on the onboard serial ports */
    default:
      result = RIO_ERROR;
      break;
  }
  return result;
}


/**
 * Create a local datagram serial port, listening on port.
 *
 * This port works by buffering writes until an outgoing_termination_char is found in the write stream or
 * the maximum buffer size is reached, and then all data since the last write (including the termination character)
 * is packaged up into a UDP datagram and sent to the last host that contacted the port
 *
 * @param io_dev is the IO device to be initialised as a datagram serial port
 * @param port is the UDP port to listen for data on
 * @param outgoing_termination_chars is a list of characters that will terminate a write and force the data to be transmitted (not '\0' terminated). Can be NULL, in which case outgoing_buffer_size is used to determine when to send the data.
 * @param num_outgoing_termination_chars is the number of characters in outgoing_termination_chars. Can be 0, in which case outgoing_buffer_size is used to determine when to send the data.
 * @param outgoing_buffer_size is the maximum number of characters to buffer before sending a UDP datagram. Must be > 0.
 */
int rio_create_device_sio_localdgram(struct t_io_dev *io_dev, int port,
                                     const char *outgoing_termination_chars,
                                     int num_outgoing_termination_chars,
                                     int outgoing_buffer_size)
{
  struct sDatagramSerialPort *dgramserial = NULL;

  /* Allocate serial port datagram info here, sio_localdgram_destroy() frees it */
  dgramserial = malloc(sizeof(struct sDatagramSerialPort));
  if (NULL == dgramserial)
  {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_sio_localdgram: failed to allocate memory for dgramserial\n");
#endif
    goto rcdsl_fail;
  }

  /*
   * Set up the datagram serial port
   */
  memset(dgramserial, 0, sizeof(struct sDatagramSerialPort));

  /* Copy outgoing terminating characters */
  if (0 != num_outgoing_termination_chars && NULL != outgoing_termination_chars)
  {
    dgramserial->outgoing_terminating_characters = malloc(num_outgoing_termination_chars);
    if (NULL == dgramserial->outgoing_terminating_characters)
    {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("rio_create_device_sio_localdgram: failed to allocate memory for outgoing terminating characters\n");
#endif
      goto rcdsl_fail;
    }
    memcpy(dgramserial->outgoing_terminating_characters, outgoing_termination_chars, num_outgoing_termination_chars);
  }
  else
  {
    dgramserial->outgoing_terminating_characters = NULL;
  }
  dgramserial->num_outgoing_terminating_characters = num_outgoing_termination_chars;

  /* Create outgoing data buffer */
  if (0 == outgoing_buffer_size)
  {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("rio_create_device_sio_localdgram: outgoing_buffer_size cannot be 0\n");
#endif
      goto rcdsl_fail;
  }
  dgramserial->outgoing_buffer = malloc(outgoing_buffer_size);
  if (NULL == dgramserial->outgoing_buffer)
  {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_sio_localdgram: failed to allocate memory for outgoing buffer\n");
#endif
    goto rcdsl_fail;
  }
  dgramserial->outgoing_buffer_size = outgoing_buffer_size;

  dgramserial->nSocketPort = port;
  dgramserial->nSocketFd = -1;
  dgramserial->response_addr_valid = 0;

  /* Set up the IO device */
  io_dev->dev_data = (void *) dgramserial;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_localdgram_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = sio_localdgram_read;
  io_dev->dev_write = sio_localdgram_write;
  io_dev->dev_filelength = sio_localdgram_filelength;
  io_dev->dev_ioctl = sio_localdgram_ioctl;
  io_dev->dev_destroy = sio_localdgram_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_sio_localdgram: failed to register device\n");
#endif
    goto rcdsl_fail;
  }

  /* Success */
  return RIO_SUCCESS;

rcdsl_fail:
  io_dev->dev_data = NULL;
  if (NULL != dgramserial)
  {
    if (NULL != dgramserial->outgoing_terminating_characters)
    {
      free(dgramserial->outgoing_terminating_characters);
      dgramserial->outgoing_terminating_characters = NULL;
    }
    if (NULL != dgramserial->outgoing_buffer)
    {
      free(dgramserial->outgoing_buffer);
      dgramserial->outgoing_buffer = NULL;
    }
    free(dgramserial);
  }
  return RIO_ERROR;
}
