/*
 * sio_local_established_dgram.c
 *
 *  Created on: Apr 4, 2012
 *      Author: ian
 */
#include <sys/types.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/sockios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinhw/timemath.h>
#include "sio_helper.h"

#include <rinhw/sio_local_established_dgram.h>


/** Device data for a serial port mapped to a socket */
struct sDatagramSerialPort
{
  int nSocketFd;                  /* The UDP socket file descriptor */
  t_ioctl_siocfg siocfg;          /* The serial port configuration */
  char *outgoing_terminating_characters;
  size_t num_outgoing_terminating_characters;
  char *outgoing_buffer;
  size_t outgoing_buffer_size;
  size_t outgoing_buffer_posn;
  struct sockaddr_in fixed_response_addr;   /*!< If set, rwrite() will always be sent to this address/port, and data will only be rread() from this address/port. In host endian. */
  struct sockaddr_in last_source_addr;      /*!< If fixed_response_addr is not set, this contains the address/port of the last rread(), and rwrite()s will sendto this address/port. In host endian. */
};


/** Initialise an established datagram serial device */
static int sio_local_established_dgram_init(struct t_io_dev *io_dev)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_init called\n");
#endif

  /* Default configuration for the serial port */
  pDgramSerial->siocfg.baud = SIO_DEF_BAUD;
  pDgramSerial->siocfg.data_bits = SIO_DEF_DATA_BITS;
  pDgramSerial->siocfg.parity_bits = SIO_DEF_PARITY_BITS;
  pDgramSerial->siocfg.stop_bits = SIO_DEF_STOP_BITS;
  pDgramSerial->siocfg.use_cts = SIO_DEF_USECTS;
  pDgramSerial->siocfg.use_terminations = SIO_DEF_TERMINATIONS;

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

  /* Set blocking on socket */
  (void) sio_helper_setblocking(pDgramSerial->nSocketFd, true);

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_init: result %d\n", result);
#endif
  return RIO_SUCCESS;
}


/** Destroy established datagram serial device */
static int sio_local_established_dgram_destroy(struct t_io_dev *io_dev)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  int result;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* Shutdown socket */
  sio_helper_close(&pDgramSerial->nSocketFd);

  /* Free up stuff we've allocated */
  io_dev->dev_data = NULL;
  if (NULL != pDgramSerial)
  {
    if (NULL != pDgramSerial->outgoing_terminating_characters)
    {
      free(pDgramSerial->outgoing_terminating_characters);
      pDgramSerial->outgoing_terminating_characters = NULL;
    }
    if (NULL != pDgramSerial->outgoing_buffer)
    {
      free(pDgramSerial->outgoing_buffer);
      pDgramSerial->outgoing_buffer = NULL;
    }
    free(pDgramSerial);
    pDgramSerial = NULL;
  }

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_destroy: result %d\n", result);
#endif
  return result;
}


/** Get the destination ip/port, in host endian */
static bool sio_local_established_dgram_get_dest(const struct sDatagramSerialPort *dgram_serial, struct sockaddr_in *dest)
{
  bool res = false;

  if (0 != dgram_serial->fixed_response_addr.sin_addr.s_addr &&
      0 != dgram_serial->fixed_response_addr.sin_port)
  {
    *dest = dgram_serial->fixed_response_addr;
    res = true;
  }
  else if (0 != dgram_serial->last_source_addr.sin_addr.s_addr &&
           0 != dgram_serial->last_source_addr.sin_port)
  {
    *dest = dgram_serial->last_source_addr;
    res = true;
  }
  return res;
}


/** Does this connection only support a fixed destination */
static bool sio_local_established_dgram_is_fixed_dest(struct sDatagramSerialPort *dgram_serial)
{
  return (0 != dgram_serial->fixed_response_addr.sin_addr.s_addr &&
          0 != dgram_serial->fixed_response_addr.sin_port);
}


/** Receive a udp datagram from a fixed source, data from other sources is dropped */
static int sio_local_established_dgram_recv_fixed_source(struct sDatagramSerialPort *dgram_serial, void *buffer, size_t length)
{
  ssize_t retval;
  socklen_t sockaddr_size;
  char dummy;
  struct sockaddr_in sender;
  int numrecvd;

  /* First peek at what is available */
  sockaddr_size = sizeof(sender);
  retval = recvfrom(dgram_serial->nSocketFd,
                    &dummy, sizeof(dummy), MSG_PEEK,
                    (struct sockaddr *) &sender, &sockaddr_size);
  if (retval < 0)
  {
    if (EWOULDBLOCK != errno)
      sio_helper_close(&dgram_serial->nSocketFd);
    return RIO_ERROR;
  }

  /* There is data available, is it for us? */
  if (sender.sin_addr.s_addr == htonl(dgram_serial->fixed_response_addr.sin_addr.s_addr) &&
      sender.sin_port == htons(dgram_serial->fixed_response_addr.sin_port))
  {
    /* Data for us */
    sockaddr_size = sizeof(sender);
    retval = recvfrom(dgram_serial->nSocketFd,
                      buffer, length, 0,
                      (struct sockaddr *) &sender, &sockaddr_size);
    if (retval < 0)
    {
      if (EWOULDBLOCK != errno)
        sio_helper_close(&dgram_serial->nSocketFd);
      return RIO_ERROR;
    }
    numrecvd += retval;
  }
  else
  {
    /* From someone else, so just dump it */
    sockaddr_size = sizeof(sender);
    retval = recvfrom(dgram_serial->nSocketFd,
                      &dummy, sizeof(dummy), 0,
                      (struct sockaddr *) &sender, &sockaddr_size);
    if (retval < 0)
    {
      if (EWOULDBLOCK != errno)
        sio_helper_close(&dgram_serial->nSocketFd);
      return RIO_ERROR;
    }
    /* Otherwise try again */
    numrecvd = 0;
  }
  return numrecvd;
}


/** Receive a udp datagram from a any source, and the source ip/port becomes our new send destination */
static int sio_local_established_dgram_recv_any_source(struct sDatagramSerialPort *dgram_serial, void *buffer, size_t length)
{
  ssize_t retval;
  socklen_t sockaddr_size;
  struct sockaddr_in sender;
  int numrecvd = 0;

  /* Read the data */
  sockaddr_size = sizeof(sender);
  retval = recvfrom(dgram_serial->nSocketFd,
                    buffer, length, 0,
                    (struct sockaddr *) &sender, &sockaddr_size);
  if (retval < 0)
  {
    if (EWOULDBLOCK != errno)
      sio_helper_close(&dgram_serial->nSocketFd);
    return RIO_ERROR;
  }
  numrecvd += retval;
  dgram_serial->last_source_addr.sin_addr.s_addr = ntohl(sender.sin_addr.s_addr);
  dgram_serial->last_source_addr.sin_port = ntohs(sender.sin_port);

  return numrecvd;
}


/**
 * Function to receive len bytes (recv doesn't guarantee to return len bytes, unless used
 * with MSG_WAITALL which isn't defined on Windows < Windows server 2003)
 */
static int sio_local_established_dgram_recv_all(struct sDatagramSerialPort *dgram_serial, void *buf, size_t len)
{
  int retval;
  size_t numrecvd;
  char *charbuf;
  const bool fixed_dest = sio_local_established_dgram_is_fixed_dest(dgram_serial);

  for (numrecvd = 0, charbuf = (char *) buf; numrecvd < len; )
  {
    /* Receive the data */
    if (fixed_dest)
      retval = sio_local_established_dgram_recv_fixed_source(dgram_serial, &charbuf[numrecvd], len - numrecvd);
    else
      retval = sio_local_established_dgram_recv_any_source(dgram_serial, &charbuf[numrecvd], len - numrecvd);

    /* Deal with error */
    if (retval < 0)
      return retval;

    numrecvd += retval;
  }
  return numrecvd;
}


/** Read from the established datagram serial port */
static int sio_local_established_dgram_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  int retval;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_read called\n");
#endif

  /* Check socket has been setup correctly */
  if (pDgramSerial->nSocketFd < 0)
    return RIO_ERROR;

  if (sio_helper_settimeout(pDgramSerial->nSocketFd, false, timeout) < 0)
  {
    return RIO_ERROR;
  }
  retval = sio_local_established_dgram_recv_all(pDgramSerial, buf, len);
  return retval;
}


/** Is ch one of the nominated terminating characters */
static bool sio_local_established_dgram_isterminator(struct sDatagramSerialPort *pDgramSerial, char ch)
{
  size_t i;
  bool result;

  for (i = 0, result = false; i < pDgramSerial->num_outgoing_terminating_characters; i++)
  {
    if (ch == pDgramSerial->outgoing_terminating_characters[i])
    {
      result = true;
      break;
    }
  }
  return result;
}


/** Send a buffer */
static int sio_local_established_dgram_sendbuffer(struct sDatagramSerialPort *pDgramSerial, const struct sockaddr_in *dest, const void *buf, size_t length)
{
  ssize_t result;
  struct sockaddr_in dest_net_endian;

  memset(&dest_net_endian, 0, sizeof(dest_net_endian));
  dest_net_endian.sin_family = dest->sin_family;
  dest_net_endian.sin_addr.s_addr = htonl(dest->sin_addr.s_addr);
  dest_net_endian.sin_port = htons(dest->sin_port);

  result = sendto(pDgramSerial->nSocketFd,
                  buf, length,
                  0, (struct sockaddr *) &dest_net_endian, sizeof(dest_net_endian));
  if (result < 0 && EWOULDBLOCK != errno)
  {
    sio_helper_close(&pDgramSerial->nSocketFd);
  }
  result = ((size_t) result != length) ? RIO_ERROR : RIO_SUCCESS;
  return result;
}


/** Write to the UDP datagram serial port */
static int sio_local_established_dgram_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  int retval;
  char ch;
  const char *inbuf;
  size_t i;
  struct sockaddr_in dest;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_write called\n");
#endif

  /* Check socket has been setup correctly */
  if (pDgramSerial->nSocketFd < 0)
    return RIO_ERROR;

  /* Who are we sending to */
  if (!sio_local_established_dgram_get_dest(pDgramSerial, &dest))
  {
    /* We don't have a valid destination, so just dump the data, and pretend we sent it */
    return (int) len;
  }

  /* Setup timeout */
  if (sio_helper_settimeout(pDgramSerial->nSocketFd, true, timeout) < 0)
    return RIO_ERROR;

  if (NULL != pDgramSerial->outgoing_buffer)
  {
    /* Buffered port, inspect each character, copy to out buffer, and write when terminating character found or buffer full */
    for (i = 0, inbuf = buf; i < len; i++, inbuf++)
    {
      ch = *inbuf;
      pDgramSerial->outgoing_buffer[pDgramSerial->outgoing_buffer_posn++] = ch;
      if (sio_local_established_dgram_isterminator(pDgramSerial, ch) || pDgramSerial->outgoing_buffer_posn >= pDgramSerial->outgoing_buffer_size)
      {
        /* Time to send */
        if (sio_local_established_dgram_sendbuffer(pDgramSerial, &dest, pDgramSerial->outgoing_buffer, pDgramSerial->outgoing_buffer_posn) < 0)
        {
          return RIO_ERROR;
        }
        pDgramSerial->outgoing_buffer_posn = 0;
      }
    }
    retval = i;
  }
  else
  {
    /* Unbuffered port, so just write what data we have */
    retval = (sio_local_established_dgram_sendbuffer(pDgramSerial, &dest, buf, len) == RIO_SUCCESS) ? len : RIO_ERROR;
  }

  return retval;
}


/** Get the number of characters in the receive buffer. We can't know this */
static long sio_local_established_dgram_filelength(struct t_io_dev *io_dev)
{
  struct sDatagramSerialPort *pDgramSerial = (struct sDatagramSerialPort *) io_dev->dev_data;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_filelength called\n");
#endif

  /* Check socket has been setup correctly */
  if (pDgramSerial->nSocketFd < 0)
  {
    return RIO_ERROR;
  }

  return (long) sio_helper_getbufferlevel(pDgramSerial->nSocketFd, false);
}


/** Use CTS to indicate we are always connected */
static void sio_local_established_dgram_ioctl_cts(struct sDatagramSerialPort *dgram_serial, unsigned char *cts_state)
{
  *cts_state = (dgram_serial->nSocketFd >= 0);
}


/** Configure etc the serial port */
static int sio_local_established_dgram_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct sDatagramSerialPort *pSktSerial = (struct sDatagramSerialPort *) io_dev->dev_data;
  t_ioctl_siocfg *siocfg_ptr;
  unsigned short *word_ptr;
  int result = RIO_SUCCESS;
  struct sockaddr_in *sockaddr;

#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localdgram_ioctl called, func %d\n", func);
#endif

  /* Check socket has been setup correctly */
  if (pSktSerial->nSocketFd < 0)
  {
    return RIO_ERROR;
  }

  /* IOCTL function dispatch */
  switch (func)
  {
    case IOCTL_SER_SET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_SET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      pSktSerial->siocfg = *siocfg_ptr;
      break;

    case IOCTL_SER_GET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_GET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      *siocfg_ptr = pSktSerial->siocfg;
      break;

    /*
     * Get buffer levels and sizes
     */
    case IOCTL_SER_GET_RX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_helper_getbufferlevel(pSktSerial->nSocketFd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_helper_getbufferlevel(pSktSerial->nSocketFd, true);
      break;
    case IOCTL_SER_GET_RX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_helper_getbuffersize(pSktSerial->nSocketFd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_helper_getbuffersize(pSktSerial->nSocketFd, true);
      break;

    /* Get the level of the cts pin */
    case IOCTL_SER_GET_CTS_STATE:
      sio_local_established_dgram_ioctl_cts(pSktSerial, va_arg(*arglist, unsigned char *));
      break;

    /* Get the number of receive errors on this port */
    case IOCTL_SER_GET_RX_ERR_COUNT:
      *va_arg(*arglist, unsigned long *) = 0; /* TODO */
      break;

    /* Flush the transmit buffer */
    case IOCTL_SER_FLUSH_TX_BUF:
      result = RIO_SUCCESS; /* Not really much we can do here, so let's just say it worked */
      break;

    /* Flush the receive buffer */
    case IOCTL_SER_FLUSH_RX_BUF:
      result = RIO_SUCCESS; /* Not really much we can do here, so let's just say it worked */
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = pSktSerial->nSocketFd;
      result = RIO_SUCCESS;
      break;

    /* Get the destination */
    case IOCTL_SER_UDP_GET_DEST_SOCKADDR:
      sockaddr = va_arg(*arglist, struct sockaddr_in *);
      if (!sio_local_established_dgram_get_dest(pSktSerial, sockaddr))
      {
        sockaddr->sin_addr.s_addr = 0;
        sockaddr->sin_port = 0;
      }
      result = RIO_SUCCESS;
      break;

    /* Set the destination ip:port for fixed response address */
    case IOCTL_SER_UDP_SET_DEST_SOCKADDR:
      sockaddr = va_arg(*arglist, struct sockaddr_in *);
      if (0 != pSktSerial->fixed_response_addr.sin_port &&
          0 != pSktSerial->fixed_response_addr.sin_addr.s_addr)
      {
        pSktSerial->fixed_response_addr.sin_port = sockaddr->sin_port;
        pSktSerial->fixed_response_addr.sin_addr.s_addr = sockaddr->sin_addr.s_addr;
      }
      else
      {
        pSktSerial->last_source_addr.sin_port = sockaddr->sin_port;
        pSktSerial->last_source_addr.sin_addr.s_addr = sockaddr->sin_addr.s_addr;
      }
      result = RIO_SUCCESS;
      break;

    /* Unknown function */
/*    case IOCTL_SER_SET_RING_NETWORK: This is unsupported on the onboard serial ports */
/*    case IOCTL_SER_SET_SOCKET_PORT: This is unsupported for established socket serial ports */
/*    case IOCTL_SER_GET_SOCKET_PORT: This is unsupported for established socket serial ports */
    default:
      result = RIO_ERROR;
      break;
  }
  return result;
}


/**
 * Create a local established datagram serial port, already bound to sockfd
 *
 * This port works by buffering writes until an outgoing_termination_char is found in the write stream or
 * the maximum buffer size is reached, and then all data since the last write (including the termination character)
 * is packaged up into a UDP datagram and sent to the last host that contacted the port
 *
 * @param io_dev is the IO device to be initialised as a datagram serial port
 * @param sockfd is the already bound datagram socket
 * @param dest is the UDP IP/port that data from this device will be read/written to. In host endian.
 * @param outgoing_termination_chars is a list of characters that will terminate a write and force the data to be transmitted (not '\0' terminated). Can be NULL, in which case outgoing_buffer_size is used to determine when to send the data.
 * @param num_outgoing_termination_chars is the number of characters in outgoing_termination_chars. Can be 0, in which case outgoing_buffer_size is used to determine when to send the data.
 * @param outgoing_buffer_size is the maximum number of characters to buffer before sending a UDP datagram. Must be > 0.
 */
static int rio_internal_create_device_sio_established_dgram(struct t_io_dev *io_dev,
                                                            int sockfd,
                                                            const struct sockaddr_in *dest,
                                                            const char *outgoing_termination_chars,
                                                            int num_outgoing_termination_chars,
                                                            int outgoing_buffer_size)
{
  struct sDatagramSerialPort *dgramserial = NULL;

  /* Allocate serial port datagram info here, sio_localdgram_destroy() frees it */
  dgramserial = calloc(1, sizeof(struct sDatagramSerialPort));
  if (NULL == dgramserial)
  {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_sio_localdgram: failed to allocate memory for dgramserial\n");
#endif
    goto rcdsl_fail;
  }

  /* Copy outgoing terminating characters */
  if (0 != num_outgoing_termination_chars && NULL != outgoing_termination_chars)
  {
    dgramserial->outgoing_terminating_characters = malloc(num_outgoing_termination_chars);
    if (NULL == dgramserial->outgoing_terminating_characters)
    {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("rio_create_device_sio_localdgram: failed to allocate memory for outgoing terminating characters\n");
#endif
      goto rcdsl_fail;
    }
    memcpy(dgramserial->outgoing_terminating_characters, outgoing_termination_chars, num_outgoing_termination_chars);
  }
  else
  {
    dgramserial->outgoing_terminating_characters = NULL;
  }
  dgramserial->num_outgoing_terminating_characters = num_outgoing_termination_chars;

  /* Create outgoing data buffer, if buffering */
  dgramserial->outgoing_buffer_size = outgoing_buffer_size;
  if (0 == outgoing_buffer_size)
  {
    dgramserial->outgoing_buffer = NULL;
  }
  else
  {
    dgramserial->outgoing_buffer = malloc(outgoing_buffer_size);
    if (NULL == dgramserial->outgoing_buffer)
    {
  #if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("rio_create_device_sio_localdgram: failed to allocate memory for outgoing buffer\n");
  #endif
      goto rcdsl_fail;
    }
  }

  dgramserial->nSocketFd = sockfd;
  if (NULL == dest)
  {
    dgramserial->fixed_response_addr.sin_addr.s_addr = 0;
    dgramserial->fixed_response_addr.sin_port = 0;
  }
  else
  {
    dgramserial->fixed_response_addr = *dest;
  }
  dgramserial->last_source_addr.sin_addr.s_addr = 0;
  dgramserial->last_source_addr.sin_port = 0;

  /* Set up the IO device */
  io_dev->dev_data = (void *) dgramserial;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_local_established_dgram_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = sio_local_established_dgram_read;
  io_dev->dev_write = sio_local_established_dgram_write;
  io_dev->dev_filelength = sio_local_established_dgram_filelength;
  io_dev->dev_ioctl = sio_local_established_dgram_ioctl;
  io_dev->dev_destroy = sio_local_established_dgram_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
  {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_sio_localdgram: failed to register device\n");
#endif
    goto rcdsl_fail;
  }

  /* Success */
  return RIO_SUCCESS;

rcdsl_fail:
  io_dev->dev_data = NULL;
  if (NULL != dgramserial)
  {
    if (NULL != dgramserial->outgoing_terminating_characters)
    {
      free(dgramserial->outgoing_terminating_characters);
      dgramserial->outgoing_terminating_characters = NULL;
    }
    if (NULL != dgramserial->outgoing_buffer)
    {
      free(dgramserial->outgoing_buffer);
      dgramserial->outgoing_buffer = NULL;
    }
    free(dgramserial);
  }
  return RIO_ERROR;
}


/**
 * Create a unbuffered established datagram (UDP) serial port, already bound to sockfd
 *
 * This port works by writing data passed by rwrite() as a UDP datagram and sent to dest
 *
 * @param io_dev is the IO device to be initialised as a datagram serial port
 * @param sockfd is the already bound datagram socket
 * @param dest is the UDP IP/port that data from this device will be read/written from/to.
 *          In host endian.
 *          If NULL, then the device will reply to the last IP/port it received data from.
 *
 */
int rio_create_device_sio_established_dgram_unbuffered(struct t_io_dev *io_dev,
                                                       int sockfd,
                                                       const struct sockaddr_in *dest)
{
  return rio_internal_create_device_sio_established_dgram(io_dev, sockfd, dest, NULL, 0, 0);
}


/**
 * Create a buffered established datagram serial port, already bound to sockfd
 *
 * This port works by buffering writes until an outgoing_termination_char is found in the write stream or
 * the maximum buffer size is reached, and then all data since the last write (including the termination character)
 * is packaged up into a UDP datagram and sent to dest
 *
 * @param io_dev is the IO device to be initialised as a datagram serial port
 * @param sockfd is the already bound datagram socket
 * @param dest is the UDP IP/port that data from this device will be read/written from/to.
 *          In host endian.
 *          If NULL, then the device will reply to the last IP/port it received data from.
 * @param outgoing_termination_chars is a list of characters that will terminate a write and force the data to be transmitted (not '\0' terminated). Can be NULL, in which case outgoing_buffer_size is used to determine when to send the data.
 * @param num_outgoing_termination_chars is the number of characters in outgoing_termination_chars. Can be 0, in which case outgoing_buffer_size is used to determine when to send the data.
 * @param outgoing_buffer_size is the maximum number of characters to buffer before sending a UDP datagram. Must be > 0.
 */
int rio_create_device_sio_established_dgram_buffered(struct t_io_dev *io_dev,
                                                     int sockfd,
                                                     const struct sockaddr_in *dest,
                                                     const char *outgoing_termination_chars,
                                                     int num_outgoing_termination_chars,
                                                     int outgoing_buffer_size)
{
  /* Create outgoing data buffer */
  if (0 == outgoing_buffer_size)
  {
#if HW_DEBUG(SIO_DATAGRAM_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("rio_create_device_sio_localdgram: outgoing_buffer_size cannot be 0\n");
#endif
    return RIO_ERROR;
  }

  return rio_internal_create_device_sio_established_dgram(io_dev, sockfd, dest, outgoing_termination_chars, num_outgoing_termination_chars, outgoing_buffer_size);
}
