/*
 * sio_local_established_socket.c -
 *
 * This type of rin serial device takes an established socket
 * file descriptor and simply communicates over. Unlike
 * sio_local_socket, it does not deal with the listen socket
 * and accepting a new connection. You should implement that
 * part in your application
 */
#include <sys/types.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/sockios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinhw/timemath.h>
#include "sio_helper.h"

#include <rinhw/sio_local_established_socket.h>


/** Device data for a serial port mapped to a socket */
struct sSocketSerialPort
{
  int nConnectionSocketFd;        /*!< The actual connection socket file descriptor */
  t_ioctl_siocfg siocfg;          /*!< The serial port configuration */
};


/** Initialise established local socket serial device */
static int sio_local_establishedskt_init(struct t_io_dev *io_dev)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  int result = 0;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_establishedskt_init called\n");
#endif

  /* Default configuration for the serial port */
  pSktSerial->siocfg.baud = SIO_DEF_BAUD;
  pSktSerial->siocfg.data_bits = SIO_DEF_DATA_BITS;
  pSktSerial->siocfg.parity_bits = SIO_DEF_PARITY_BITS;
  pSktSerial->siocfg.stop_bits = SIO_DEF_STOP_BITS;
  pSktSerial->siocfg.use_cts = SIO_DEF_USECTS;
  pSktSerial->siocfg.use_terminations = SIO_DEF_TERMINATIONS;

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

  /* Set blocking on socket */
  (void) sio_helper_setblocking(pSktSerial->nConnectionSocketFd, true);

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_establishedskt_init: result %d\n", result);
#endif
  return result;
}


/** Destroy serial device */
static int sio_local_establishedskt_destroy(struct t_io_dev *io_dev)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  int result;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_establishedskt_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* Shutdown socket */
  sio_helper_close(&pSktSerial->nConnectionSocketFd);

  /* Free up stuff we've allocated */
  free(io_dev->dev_data);
  io_dev->dev_data = NULL;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_establishedskt_destroy: result %d\n", result);
#endif
  return result;
}


/** Read from the socket */
static int sio_local_establishedskt_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  int retval;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_establishedskt_read called\n");
#endif

  /* If we are connected, then wait for data to read */
  if (pSktSerial->nConnectionSocketFd < 0)
    return RIO_ERROR;
  if (sio_helper_settimeout(pSktSerial->nConnectionSocketFd, false, timeout) < 0)
  {
    return RIO_ERROR;
  }
  retval = sio_helper_read_all(pSktSerial->nConnectionSocketFd, buf, len);
  if ((retval < 0 && EWOULDBLOCK != errno) || 0 == retval)
  {
    sio_helper_close(&pSktSerial->nConnectionSocketFd);
  }
  return retval;
}


/** Write to the socket */
static int sio_local_establishedskt_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  int retval;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_establishedskt_write called\n");
#endif

  /* If we are connected, then try to write data */
  if (pSktSerial->nConnectionSocketFd >= 0)
  {
    if (sio_helper_settimeout(pSktSerial->nConnectionSocketFd, true, timeout) < 0)
      return RIO_ERROR;
    retval = sio_helper_write_all(pSktSerial->nConnectionSocketFd, buf, len);
    if (retval < 0 && EWOULDBLOCK != errno)
    {
      sio_helper_close(&pSktSerial->nConnectionSocketFd);
    }
  }
  else
  {
    /* We have no socket connected, so just discard the data and tell the caller that it was all written */
    retval = (int) len;
  }
  return retval;
}


/** Get the number of characters in the receive buffer */
static long sio_local_establishedskt_filelength(struct t_io_dev *io_dev)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_establishedskt_filelength called\n");
#endif

  return (long) sio_helper_getbufferlevel(pSktSerial->nConnectionSocketFd, false);
}


/** Use CTS to determine if there is a socket connection */
static void sio_local_establishedskt_ioctl_cts(struct sSocketSerialPort *pSktSerial, unsigned char *cts_state)
{
  *cts_state = (pSktSerial->nConnectionSocketFd >= 0);
}


/** Flush (discard) any incoming data from the socket */
static int sio_local_establishedskt_flush_incoming(struct sSocketSerialPort *pSktSerial)
{
  int level;
  char buffer;

  level = sio_helper_getbufferlevel(pSktSerial->nConnectionSocketFd, false);
  while (level > 0)
  {
    if (sio_helper_read_all(pSktSerial->nConnectionSocketFd, &buffer, sizeof(buffer)) < 0)
      level = -1;
    else
      level--;
  }
  return (level < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/** Configure the socket serial port */
static int sio_local_establishedskt_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  t_ioctl_siocfg *siocfg_ptr;
  unsigned short *word_ptr;
  int result = RIO_SUCCESS;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_establishedskt_ioctl called, function %d\n", func);
#endif

  /* IOCTL function dispatch */
  switch (func)
  {
    case IOCTL_SER_SET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_SET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      pSktSerial->siocfg = *siocfg_ptr;
      break;

    case IOCTL_SER_GET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_GET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      *siocfg_ptr = pSktSerial->siocfg;
      break;

    /* Get buffer levels and sizes */
    case IOCTL_SER_GET_RX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_helper_getbufferlevel(pSktSerial->nConnectionSocketFd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_helper_getbufferlevel(pSktSerial->nConnectionSocketFd, true);
      break;
    case IOCTL_SER_GET_RX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_helper_getbuffersize(pSktSerial->nConnectionSocketFd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_helper_getbuffersize(pSktSerial->nConnectionSocketFd, true);
      break;

    /* Get the level of the CTS pin */
    case IOCTL_SER_GET_CTS_STATE:
      sio_local_establishedskt_ioctl_cts(pSktSerial, va_arg(*arglist, unsigned char *));
      break;

    /* Get the number of receive errors on this port */
    case IOCTL_SER_GET_RX_ERR_COUNT:
      *va_arg(*arglist, unsigned long *) = 0;
      break;

    /* Flush the transmit buffer */
    case IOCTL_SER_FLUSH_TX_BUF:
      result = RIO_SUCCESS; /* Not really much we can do here, so let's just say it worked */
      break;

    /* Flush the receive buffer */
    case IOCTL_SER_FLUSH_RX_BUF:
      result = sio_local_establishedskt_flush_incoming(pSktSerial);
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = pSktSerial->nConnectionSocketFd;
      result = RIO_SUCCESS;
      break;

    /* Unknown function */
/*    case IOCTL_SER_SET_RING_NETWORK: This is unsupported for socket serial ports */
/*    case IOCTL_SER_SET_SOCKET_PORT: This is unsupported for established socket serial ports */
/*    case IOCTL_SER_GET_SOCKET_PORT: This is unsupported for established socket serial ports */
    default:
      result = RIO_ERROR;
      break;
  }
  return result;
}


/** Close our socket */
static int sio_local_establishedskt_close(struct t_io_dev *io_dev)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;

  sio_helper_close(&pSktSerial->nConnectionSocketFd);
  return RIO_SUCCESS;
}


/**
 * Create a rin serial device for an established socket
 *
 * This type of rin serial device takes an established socket
 * file descriptor and simply communicates over. Unlike
 * sio_local_socket, it does not deal with the listen socket
 * and accepting a new connection. You should implement that
 * part in your application
 *
 * @param io_dev is the un-initialised rin io device structure
 * @param sockfd is the established socket file descriptor to communicate over
 */
int rio_create_device_sio_established_localskt(struct t_io_dev *io_dev, int sockfd)
{
  struct sSocketSerialPort *sockserial;

  /* Allocate serial port socket info here, sio_localskt_destroy() frees it */
  sockserial = malloc(sizeof(struct sSocketSerialPort));
  if (NULL == sockserial)
  {
    return RIO_ERROR;
  }

  /* Set up the sock serial */
  sockserial->nConnectionSocketFd = sockfd;

  /* Set up the IO device */
  io_dev->dev_data = (void *) sockserial;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_local_establishedskt_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = sio_local_establishedskt_close;
  io_dev->dev_read = sio_local_establishedskt_read;
  io_dev->dev_write = sio_local_establishedskt_write;
  io_dev->dev_filelength = sio_local_establishedskt_filelength;
  io_dev->dev_ioctl = sio_local_establishedskt_ioctl;
  io_dev->dev_destroy = sio_local_establishedskt_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);
}
