/*
 * sio_local.c
 *
 *  Created on: 11/02/2009
 *      Author: ianb
 */
#include <stdarg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stddef.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/rinendian.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinhw/stdrinstrum.h>

#include <rinhw/sio_local_file.h>


struct sLocalSIO_SerialFileInfo
{
  int file_handle;
  t_ioctl_siocfg pretend_serial_cfg;
  char *file_name;
};


/*
 * sio_init -
 *
 * Initialise a serial device
 */
static int sio_local_file_init(struct t_io_dev *io_dev)
{
  struct sLocalSIO_SerialFileInfo *pSerFile = (struct sLocalSIO_SerialFileInfo *) io_dev->dev_data;
  const char *filename = pSerFile->file_name; /* Port name passed into rio_create_device_sio_localfile() */

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_init called\n");
#endif

  /* Create our own internal copy of the file name */
  pSerFile->file_name = malloc( strlen( filename ) + 1 );
  if ( (const char *) 0 == pSerFile->file_name )
  {
    return RIO_ERROR;
  }
  strcpy( pSerFile->file_name, filename );
  pSerFile->file_handle = -1;

  /* Set defaults for serial port */
  pSerFile->pretend_serial_cfg.baud = SIO_DEF_BAUD;
  pSerFile->pretend_serial_cfg.data_bits = SIO_DEF_DATA_BITS;
  pSerFile->pretend_serial_cfg.parity_bits = SIO_DEF_PARITY_BITS;
  pSerFile->pretend_serial_cfg.stop_bits = SIO_DEF_STOP_BITS;
  pSerFile->pretend_serial_cfg.use_cts = SIO_DEF_USECTS;
  pSerFile->pretend_serial_cfg.use_terminations = SIO_DEF_TERMINATIONS;

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

  return RIO_SUCCESS;
}


/** Destroy serial device */
static int sio_local_file_destroy(struct t_io_dev *io_dev)
{
  struct sLocalSIO_SerialFileInfo *pSerFile = (struct sLocalSIO_SerialFileInfo *) io_dev->dev_data;
  int result;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* Free up stuff we've allocated */
  free(pSerFile->file_name);
  pSerFile->file_name = NULL;
  free(io_dev->dev_data);
  io_dev->dev_data = NULL;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_destroy: result %d\n", result);
#endif
  return result;
}


/** Serial open device call, opens a local file */
static int sio_local_file_open( struct t_io_dev *io_dev, unsigned short timeout )
{
  struct sLocalSIO_SerialFileInfo *pSerFile = (struct sLocalSIO_SerialFileInfo *) io_dev->dev_data;
  int handle;
  int result;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_open called\n");
#endif

  handle = open( pSerFile->file_name, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR );
  if ( 0 > handle )
  {
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("sio_local_file_open: open of %s failed: result %d, errno %d\n", pSerFile->file_name, handle, errno);
#endif
    result = RIO_ERROR;
  }
  else
  {
    if (fcntl(handle, F_SETFD, FD_CLOEXEC) < 0)
    {
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_local_file_open: error setting FD_CLOEXEC on file %s: errno %d\n", pSerFile->file_name, errno);
#endif
      result = RIO_ERROR;
    }
    else
    {
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_local_file_open: opened %s, handle %d\n", pSerFile->file_name, handle);
#endif
      pSerFile->file_handle = handle;
      result = RIO_SUCCESS;
    }
  }

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_open: result %d\n", result);
#endif
  return result;
}


/** Serial close device call, closes a local file */
static int sio_local_file_close( struct t_io_dev *io_dev )
{
  struct sLocalSIO_SerialFileInfo *pSerFile = (struct sLocalSIO_SerialFileInfo *) io_dev->dev_data;
  const int handle = pSerFile->file_handle;
  int result;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_close called\n");
#endif

  pSerFile->file_handle = -1;
  result = close( handle );

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_close: closed %s, handle %d, result %d\n", pSerFile->file_name, handle, result);
#endif
  return ( result < 0 ) ? RIO_ERROR : RIO_SUCCESS;
}


/*
 * sio_read -
 *
 * Read from the serial port
 */
static int sio_local_file_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_read called, function not implemented\n");
#endif

  return (int) RIO_ERROR;
}


/*
 * sio_write -
 *
 * Write to the serial port
 */
static int sio_local_file_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct sLocalSIO_SerialFileInfo *pSerFile = (struct sLocalSIO_SerialFileInfo *) io_dev->dev_data;
  int result;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_write called\n");
#endif

  /* Do the write */
  result = write( pSerFile->file_handle, buf, len );

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_write: wrote file %s, handle %d, result %d, len %d\n", pSerFile->file_name, pSerFile->file_handle, result, len);
#endif

  /* Return the result */
  return ( result != len ) ? RIO_ERROR : len;
}


/*
 * sio_filelength -
 *
 * Get the number of characters in the receive buffer
 */
static long sio_local_file_filelength(struct t_io_dev *io_dev)
{
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_file_filelength called, function not implemented\n");
#endif

  return 0;
}


/*
 * sio_ioctl -
 *
 * Configure etc the serial port
 */
static int sio_local_file_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
    struct sLocalSIO_SerialFileInfo *pSerFile = (struct sLocalSIO_SerialFileInfo *) io_dev->dev_data;
    t_ioctl_siocfg *siocfg_ptr;
    unsigned short *word_ptr;
    int result = RIO_SUCCESS;

#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("sio_local_file_ioctl called\n");
#endif

    /*
     * IOCTL function dispatch
     */
    switch (func)
    {
        case IOCTL_SER_SET_CONFIG:
            siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
            DEBUG_PRINTF("IOCTL_SER_SET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
            pSerFile->pretend_serial_cfg = *siocfg_ptr;
            break;

        case IOCTL_SER_GET_CONFIG:
            siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_DEBUG, HW_DEBUG_BASIC)
            DEBUG_PRINTF("IOCTL_SER_GET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
            *siocfg_ptr = pSerFile->pretend_serial_cfg;
            break;

        /*
         * Get buffer levels and sizes
         */
        case IOCTL_SER_GET_RX_BUF_LEVEL:
            word_ptr = va_arg(*arglist, unsigned short *);
            *word_ptr = 0;
            break;
        case IOCTL_SER_GET_TX_BUF_LEVEL:
            word_ptr = va_arg(*arglist, unsigned short *);
            *word_ptr = 0;
            break;
        case IOCTL_SER_GET_RX_BUF_SIZE:
            word_ptr = va_arg(*arglist, unsigned short *);
            *word_ptr = 100;
            break;
        case IOCTL_SER_GET_TX_BUF_SIZE:
            word_ptr = va_arg(*arglist, unsigned short *);
            *word_ptr = 100;
            break;

        /*
         * Get the level of the cts pin
         */
        case IOCTL_SER_GET_CTS_STATE:
            *va_arg(*arglist, unsigned char *) = 1;
            break;

        /*
         * Get the number of receive errors on this port
         */
        case IOCTL_SER_GET_RX_ERR_COUNT:
            *va_arg(*arglist, unsigned long *) = 0;
            break;

        /*
         * Flush the transmit buffer
         */
        case IOCTL_SER_FLUSH_TX_BUF:
            result = RIO_SUCCESS; /* Not really much we can do here, so let's just say it worked */
            break;

        /*
         * Flush the receive buffer
         */
        case IOCTL_SER_FLUSH_RX_BUF:
            result = RIO_SUCCESS; /* No receive buffer for local file, so let's just say it worked */
            break;

        /* Get Linux file descriptor */
        case IOCTL_GET_LINUX_FD:
          *va_arg(*arglist, int *) = pSerFile->file_handle;
          result = RIO_SUCCESS;
          break;

        /*
         * Unknown function
         */
/*      case IOCTL_SER_SET_RING_NETWORK: This is unsupported on the onboard serial ports */
        default:
            result = RIO_ERROR;
            break;
    }
    return result;
}


/** Create a serial port connected to a local file. */
int rio_create_device_sio_localfile(struct t_io_dev *io_dev, const char *filename)
{
  struct sLocalSIO_SerialFileInfo *pSerFile;

  /* Allocate serial port file info here, sio_localcom_destroy() frees it */
  pSerFile = malloc(sizeof(struct sLocalSIO_SerialFileInfo));
  if (0 == pSerFile)
  {
    return RIO_ERROR;
  }

  /* Set up the file serial info */
  pSerFile->file_handle = -1;
  pSerFile->file_name = (char *) filename;

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) pSerFile;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_local_file_init;
  io_dev->dev_open = sio_local_file_open;
  io_dev->dev_close = sio_local_file_close;
  io_dev->dev_read = sio_local_file_read;
  io_dev->dev_write = sio_local_file_write;
  io_dev->dev_filelength = sio_local_file_filelength;
  io_dev->dev_ioctl = sio_local_file_ioctl;
  io_dev->dev_destroy = sio_local_file_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);
}
