/*
 * sio_local_pipe.c
 *
 * This type of rin serial device links two serial devices via a pipe()
 */
#include <sys/types.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/sockios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinhw/timemath.h>
#include "sio_helper.h"

#include <rinhw/sio_local_pipe.h>


/** Device data for pipe */
struct s_serial_port_pipe
{
  int pipe_read_fd;        /*!< Read end of pipe */
  int pipe_write_fd;       /*!< Write end of pipe */
  t_ioctl_siocfg siocfg;   /*!< The serial port configuration */
};


static void sio_local_pipe_set_config(struct s_serial_port_pipe *ser_pipe, t_ioctl_siocfg *siocfg_ptr)
{
  /* Take a copy of the serial configuration */
  ser_pipe->siocfg = *siocfg_ptr;
}


static void sio_local_pipe_get_config(struct s_serial_port_pipe *ser_pipe, t_ioctl_siocfg *siocfg_ptr)
{
  *siocfg_ptr = ser_pipe->siocfg;
}


/** Close the pipes */
static void sio_local_pipe_close(struct s_serial_port_pipe *ser_pipe)
{
  if (ser_pipe->pipe_read_fd >= 0)
  {
    close(ser_pipe->pipe_read_fd);
    ser_pipe->pipe_read_fd = -1;
  }
  if (ser_pipe->pipe_write_fd >= 0)
  {
    close(ser_pipe->pipe_write_fd);
    ser_pipe->pipe_write_fd = -1;
  }
}


/** Initialise pipe serial device */
static int sio_local_pipe_init(struct t_io_dev *io_dev)
{
  struct s_serial_port_pipe *ser_pipe = (struct s_serial_port_pipe *) io_dev->dev_data;
  t_ioctl_siocfg siocfg;
  int result = 0;

#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_pipe_init called\n");
#endif

  /* Default configuration for the serial port */
  siocfg.baud = SIO_DEF_BAUD;
  siocfg.data_bits = SIO_DEF_DATA_BITS;
  siocfg.parity_bits = SIO_DEF_PARITY_BITS;
  siocfg.stop_bits = SIO_DEF_STOP_BITS;
  siocfg.use_cts = SIO_DEF_USECTS;
  siocfg.use_terminations = SIO_DEF_TERMINATIONS;
  sio_local_pipe_set_config(ser_pipe, &siocfg);

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_pipe_init: result %d\n", result);
#endif
  return result;
}


/** Destroy pipe serial device */
static int sio_local_pipe_destroy(struct t_io_dev *io_dev)
{
  struct s_serial_port_pipe *ser_pipe = (struct s_serial_port_pipe *) io_dev->dev_data;
  int result;

#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_pipe_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* Shutdown sockets */
  sio_local_pipe_close(ser_pipe);

  /* Free up stuff we've allocated */
  free(io_dev->dev_data);
  io_dev->dev_data = NULL;

#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_pipe_destroy: result %d\n", result);
#endif
  return result;
}


#if 0 /* TODO timeout */
static int sio_local_establishedskt_setrecvtimeout(int sockfd, unsigned short timeout)
{
  struct timeval tv;
  int result;

  OSTicksToTimeval(&tv, timeout);
  result = setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, (char *) &tv,  sizeof(tv));

  return (result < 0) ? RIO_ERROR : RIO_SUCCESS;
}
#endif


/** Read from the pipe */
static int sio_local_pipe_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_serial_port_pipe *ser_pipe = (struct s_serial_port_pipe *) io_dev->dev_data;
  int retval;

#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_pipe_read called\n");
#endif

  /* If we are connected, then wait for data to read */
  if (ser_pipe->pipe_read_fd < 0)
    return RIO_ERROR;
  sio_helper_setblocking(ser_pipe->pipe_read_fd, true);
#if 0 /* TODO timeout */
  if (RIO_SUCCESS != sio_local_establishedskt_setrecvtimeout(ser_pipe->pipe_read_fd, timeout))
  {
    return RIO_ERROR;
  }
#endif
  retval = sio_helper_read_all(ser_pipe->pipe_read_fd, buf, len);
  if ((retval < 0 && EWOULDBLOCK != errno) || 0 == retval)
  {
    sio_local_pipe_close(ser_pipe);
  }
  return retval;
}


/** Write to the pipe */
static int sio_local_pipe_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_serial_port_pipe *ser_pipe = (struct s_serial_port_pipe *) io_dev->dev_data;
  int retval;

#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_pipe_write called\n");
#endif

  sio_helper_setblocking(ser_pipe->pipe_write_fd, true);
#if 0 /* TODO timeout */
  if (RIO_SUCCESS != sio_local_establishedskt_setrecvtimeout(ser_pipe->pipe_write_fd, timeout))
  {
    return RIO_ERROR;
  }
#endif
  retval = sio_helper_write_all(ser_pipe->pipe_write_fd, buf, len);
  if (retval < 0 && EWOULDBLOCK != errno)
  {
    sio_local_pipe_close(ser_pipe);
  }
  return retval;
}


static int sio_local_pipe_getbuffersize(int sockfd, bool bTxNotRx)
{
  int bufsize;

  if (sockfd < 0)
  {
    return 0;
  }

  /*
   * According to man 7 pipe, capacity is 63356 in Linux 2.6.11+.
   * F_GET_PIPE_SZ was only introduced in 2.6.35
   */
#ifdef F_GET_PIPE_SZ
  bufsize = fcntl(sockfd, F_GET_PIPE_SZ);
  if (bufsize < 0)
    bufsize = 0;
#else
  bufsize = 65536;
#endif

  return bufsize;
}


/** Return the number of bytes available in either the receive or transmit buffers for a socket */
static int sio_local_pipe_getbufferlevel(int sockfd, bool bTxNotRx)
{
  int buflevel;
  int res;
  int ioctl_type = (bTxNotRx) ? SIOCOUTQ : SIOCINQ;

  if (sockfd < 0)
  {
    return 0;
  }
  res = ioctl(sockfd, ioctl_type, &buflevel);
  if (res < 0)
  {
    buflevel = 0;
  }
  return buflevel;
}


/** Get the number of characters in the receive buffer */
static long sio_local_pipe_filelength(struct t_io_dev *io_dev)
{
  struct s_serial_port_pipe *ser_pipe = (struct s_serial_port_pipe *) io_dev->dev_data;

#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_pipe_filelength called\n");
#endif

  return (long) sio_local_pipe_getbufferlevel(ser_pipe->pipe_read_fd, false);
}


/** Use CTS to determine if there is a pipe connection */
static void sio_local_pipe_ioctl_cts(struct s_serial_port_pipe *ser_pipe, unsigned char *cts_state)
{
  *cts_state = (ser_pipe->pipe_read_fd >= 0 && ser_pipe->pipe_write_fd >= 0);
}


/** Flush (discard) any incoming data from the socket */
static int sio_local_pipe_flush_incoming(struct s_serial_port_pipe *ser_pipe, bool write_not_read)
{
  int level;
  char buffer;
  int fd;

  fd = (write_not_read) ? ser_pipe->pipe_write_fd : ser_pipe->pipe_read_fd;
  level = sio_local_pipe_getbufferlevel(fd, false);
  while (level > 0)
  {
    if (sio_helper_read_all(fd, &buffer, sizeof(buffer)) < 0)
      level = -1;
    else
      level--;
  }
  return (level < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/** Configure the pipe serial port */
static int sio_local_pipe_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_serial_port_pipe *ser_pipe = (struct s_serial_port_pipe *) io_dev->dev_data;
  t_ioctl_siocfg *siocfg_ptr;
  unsigned short *word_ptr;
  int result = RIO_SUCCESS;

#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_local_pipe_ioctl called, function %d\n", func);
#endif

  /* IOCTL function dispatch */
  switch (func)
  {
    case IOCTL_SER_SET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_SET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      sio_local_pipe_set_config(ser_pipe, siocfg_ptr);
      break;

    case IOCTL_SER_GET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_PIPE_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_GET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      sio_local_pipe_get_config(ser_pipe, siocfg_ptr);
      break;

    /* Get buffer levels and sizes */
    case IOCTL_SER_GET_RX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_local_pipe_getbufferlevel(ser_pipe->pipe_read_fd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_local_pipe_getbufferlevel(ser_pipe->pipe_write_fd, true);
      break;
    case IOCTL_SER_GET_RX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_local_pipe_getbuffersize(ser_pipe->pipe_read_fd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_local_pipe_getbuffersize(ser_pipe->pipe_write_fd, true);
      break;

    /* Get the level of the CTS pin */
    case IOCTL_SER_GET_CTS_STATE:
      sio_local_pipe_ioctl_cts(ser_pipe, va_arg(*arglist, unsigned char *));
      break;

    /* Get the number of receive errors on this port */
    case IOCTL_SER_GET_RX_ERR_COUNT:
      *va_arg(*arglist, unsigned long *) = 0;
      break;

    /* Flush the transmit buffer */
    case IOCTL_SER_FLUSH_TX_BUF:
      result = sio_local_pipe_flush_incoming(ser_pipe, true);
      break;

    /* Flush the receive buffer */
    case IOCTL_SER_FLUSH_RX_BUF:
      result = sio_local_pipe_flush_incoming(ser_pipe, false);
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = ser_pipe->pipe_read_fd;
      result = RIO_SUCCESS;
      break;

    /* Unknown function */
/*    case IOCTL_SER_SET_RING_NETWORK: This is unsupported for socket serial ports */
/*    case IOCTL_SER_SET_SOCKET_PORT: This is unsupported for established socket serial ports */
/*    case IOCTL_SER_GET_SOCKET_PORT: This is unsupported for established socket serial ports */
    default:
      result = RIO_ERROR;
      break;
  }
  return result;
}


static int rio_local_pipe_setup_io_dev(struct t_io_dev *io_dev, struct s_serial_port_pipe *pipeserial)
{
  /* Set up the IO device */
  io_dev->dev_data = (void *) pipeserial;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_local_pipe_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = sio_local_pipe_read;
  io_dev->dev_write = sio_local_pipe_write;
  io_dev->dev_filelength = sio_local_pipe_filelength;
  io_dev->dev_ioctl = sio_local_pipe_ioctl;
  io_dev->dev_destroy = sio_local_pipe_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);
}


/**
 * Create rin serial devices linked via a pipe()
 *
 * @param io_dev1 is the first un-initialised rin io device structure
 * @param io_dev2 is the second un-initialised rin io device structure
 */
int rio_create_device_sio_pipe(struct t_io_dev *io_dev1, struct t_io_dev *io_dev2)
{
  struct s_serial_port_pipe *pipeserial1 = NULL;
  struct s_serial_port_pipe *pipeserial2 = NULL;
  int one_to_two_pipefd[2] = { -1, -1 };
  int two_to_one_pipefd[2] = { -1, -1 };
  bool one_freed;
  bool two_freed;

  #define ONE_TO_TWO_READFD one_to_two_pipefd[0]
  #define ONE_TO_TWO_WRITEFD one_to_two_pipefd[1]
  #define TWO_TO_ONE_READFD two_to_one_pipefd[0]
  #define TWO_TO_ONE_WRITEFD two_to_one_pipefd[1]

  /* Allocate serial port socket info here, sio_localskt_destroy() frees it */
  pipeserial1 = malloc(sizeof(struct s_serial_port_pipe));
  if (NULL == pipeserial1)
  {
    ERROR_PRINTF("rio_create_device_sio_pipe: failed to allocate pipe serial 1\n");
    goto rcdsp_fail;
  }

  /* Allocate serial port socket info here, sio_localskt_destroy() frees it */
  pipeserial2 = malloc(sizeof(struct s_serial_port_pipe));
  if (NULL == pipeserial2)
  {
    ERROR_PRINTF("rio_create_device_sio_pipe: failed to allocate pipe serial 2\n");
    goto rcdsp_fail;
  }

  /* Create pipes to link the two devices */
  if (pipe(one_to_two_pipefd) < 0)
  {
    ERROR_PRINTF("rio_create_device_sio_pipe: pipe() one to two failed\n");
    goto rcdsp_fail;
  }
  pipeserial1->pipe_write_fd = ONE_TO_TWO_WRITEFD;
  pipeserial2->pipe_read_fd = ONE_TO_TWO_READFD;
  if (pipe(two_to_one_pipefd) < 0)
  {
    ERROR_PRINTF("rio_create_device_sio_pipe: pipe() two to one failed\n");
    goto rcdsp_fail;
  }
  pipeserial1->pipe_read_fd = TWO_TO_ONE_READFD;
  pipeserial2->pipe_write_fd = TWO_TO_ONE_WRITEFD;

  if (rio_local_pipe_setup_io_dev(io_dev1, pipeserial1) < 0)
  {
    ERROR_PRINTF("rio_create_device_sio_pipe: create rindevice 1 failed\n");
    goto rcdsp_fail;
  }
  if (rio_local_pipe_setup_io_dev(io_dev2, pipeserial2) < 0)
  {
    ERROR_PRINTF("rio_create_device_sio_pipe: create rindevice 2 failed\n");
    goto rcdsp_fail;
  }

  return RIO_SUCCESS;

rcdsp_fail:
  one_freed = false;
  two_freed = false;
  if (0 != (io_dev2->status & O_INITIALISED))
  {
    rdestroy_io(io_dev2);
    two_freed = true;
  }
  if (0 != (io_dev1->status & O_INITIALISED))
  {
    rdestroy_io(io_dev1);
    one_freed = true;
  }

  if (ONE_TO_TWO_WRITEFD >= 0)
    close(ONE_TO_TWO_WRITEFD);
  if (ONE_TO_TWO_READFD >= 0)
    close(ONE_TO_TWO_READFD);
  if (TWO_TO_ONE_READFD >= 0)
    close(TWO_TO_ONE_READFD);
  if (TWO_TO_ONE_WRITEFD >= 0)
    close(TWO_TO_ONE_WRITEFD);

  if (!two_freed && NULL != pipeserial2)
    free(pipeserial2);

  if (!one_freed && NULL != pipeserial1)
    free(pipeserial1);

  return RIO_ERROR;
}
