/*
 * sio_local_printer.c
 *
 *  Created on: 11/02/2009
 *      Author: ianb
 */
#include <stdarg.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <linux/lp.h>
#include <errno.h>
#include <rinhw/rstdio.h>
#include "abstractserial.h"
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>

#include <rinhw/sio_local_printer.h>


/** Local serial port support information */
struct s_sio_localprinter
{
  t_ioctl_siocfg fake_serial_config;            /*!< We keep a fake serial configuration for this port */
  char *port_name;
  int port_fd;
};


/** Configure the serial port */
static void sio_local_printer_set_config(struct s_sio_localprinter *sio_localprinter, t_ioctl_siocfg *siocfg_ptr)
{
  /* Save configuration */
  sio_localprinter->fake_serial_config = *siocfg_ptr;
}


/** Get the "current" serial port configuration */
static void sio_local_printer_get_config(struct s_sio_localprinter *sio_localprinter, t_ioctl_siocfg *siocfg_ptr)
{
  *siocfg_ptr = sio_localprinter->fake_serial_config;
}


/** Initialise a printer serial device */
static int sio_localprinter_init(struct t_io_dev *io_dev)
{
  struct s_sio_localprinter *sio_localprinter = (struct s_sio_localprinter *) io_dev->dev_data;
  t_ioctl_siocfg siocfg;
  int res;

#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_init called\n");
#endif

  /* Open device */
  sio_localprinter->port_fd = open(sio_localprinter->port_name, O_RDWR);
  if (sio_localprinter->port_fd < 0)
  {
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("sio_localprinter_init: open device %s failed\n", sio_localprinter->port_name);
#endif
    /* Clean up */
    goto sli_fail;
  }
  if (fcntl(sio_localprinter->port_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("sio_localprinter_init: error setting FD_CLOEXEC on device %s: errno %d\n", sio_localprinter->port_name, errno);
#endif
    /* Clean up */
    goto sli_fail;
  }

  /* Set LPABORT, so writes will time out */
  res = ioctl(sio_localprinter->port_fd, LPABORT, 1);
  if (res < 0)
  {
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("sio_localprinter_init: ioctl(LPABORT) failed, result %d, errno %d\n", res, errno);
#endif
    /* Clean up */
    goto sli_fail;
  }

  /* Configure the serial port to defaults */
  siocfg.baud = SIO_DEF_BAUD;
  siocfg.data_bits = SIO_DEF_DATA_BITS;
  siocfg.parity_bits = SIO_DEF_PARITY_BITS;
  siocfg.stop_bits = SIO_DEF_STOP_BITS;
  siocfg.use_cts = SIO_DEF_USECTS;
  siocfg.use_terminations = SIO_DEF_TERMINATIONS;
  sio_local_printer_set_config(sio_localprinter, &siocfg);

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_init finished\n");
#endif

  return RIO_SUCCESS;

sli_fail:
  /* Clean up */
  if (NULL != sio_localprinter->port_name)
  {
    free(sio_localprinter->port_name);
    sio_localprinter->port_name = NULL;
  }
  if (sio_localprinter->port_fd >= 0)
  {
    close(sio_localprinter->port_fd);
    sio_localprinter->port_fd = -1;
  }
  io_dev->dev_data = NULL;
  if (NULL != sio_localprinter)
  {
    free(sio_localprinter);
    sio_localprinter = NULL;
  }
  return RIO_ERROR;
}


/** Destroy serial device */
static int sio_localprinter_destroy(struct t_io_dev *io_dev)
{
  struct s_sio_localprinter *sio_localprinter = (struct s_sio_localprinter *) io_dev->dev_data;

  /* Close device */
  if (NULL != sio_localprinter)
  {
    if (sio_localprinter->port_fd >= 0)
    {
      close(sio_localprinter->port_fd);
      sio_localprinter->port_fd = -1;
    }

    /* Free string */
    if (NULL != sio_localprinter->port_name)
    {
      free(sio_localprinter->port_name);
      sio_localprinter->port_name = NULL;
    }

    /* Free device data */
    io_dev->dev_data = NULL;
    free(sio_localprinter);
  }

  return RIO_SUCCESS;
}


/** Read from the printer "serial port", not implemented */
static int sio_localprinter_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_read called, function not implemented\n");
#endif

  return (int) RIO_ERROR;
}


/** Delay while printer status indicates error */
static int sio_localprinter_waitstatus(struct s_sio_localprinter *sio_localprinter, unsigned short timeout)
{
  int status;
  int error;
  int32u finishtime;

  finishtime = OSTimeGet() + timeout;
  error = 1;

#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_waitstatus called, timeout %u, finishtime %lu\n", timeout, finishtime);
#endif

  while (error)
  {

    /* Query Linux line printer device status */
    if (ioctl(sio_localprinter->port_fd, LPGETSTATUS, &status) < 0)
    {
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_localprinter_waitstatus: ioctl(LPGETSTATUS) failed\n");
#endif
      return RIO_ERROR;
    }

    /* Check printer */
    error = (~status & LP_PERRORP) ||  /* Error - printer on fire */
            (status & LP_POUTPA) ||    /* Paper out */
            (~status & LP_PSELECD);    /* Printer not selected */
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("sio_localprinter_waitstatus: error: %d, status: %s, %s, %s, time: %lu\n",
                 error,
                 (~status & LP_PERRORP) ? "error" : "",
                 (status & LP_POUTPA) ? "paper-out" : "",
                 (~status & LP_PSELECD) ? "offline" : "",
                 OSTimeGet()
                );
#endif

    /* Check for timeout */
    if (0 != timeout && (OSTimeGet() >= finishtime) && error)
    {
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_localprinter_waitstatus: timeout\n");
#endif
      return RIO_ERROR;
    }

    /* Yield on error */
    if (error)
    {
      OSTimeDly(1);
    }
  }

#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_waitstatus: success (nothing blocking write)\n");
#endif

  /* Nothing blocking our write */
  return RIO_SUCCESS;
}


/** Write to the printer "serial port" */
static int sio_localprinter_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_sio_localprinter *sio_localprinter = (struct s_sio_localprinter *) io_dev->dev_data;
  const uint8_t *ptr = buf;
  size_t size = len;
  size_t sent = 0;
  int res;

#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_write called\n");
#endif

  /* If not open, return error */
  if (sio_localprinter->port_fd < 0)
  {
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("sio_localprinter_write: invalid file descriptor\n");
#endif
    return RIO_ERROR;
  }

  /* Send */
  while (sent < size)
  {
    /* Check if any printer errors are blocking our write */
    if (sio_localprinter_waitstatus(sio_localprinter, timeout) < 0)
    {
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_localprinter_write: wait for printer status failed\n");
#endif
      return RIO_ERROR;
    }

    /* Ok to write */
    res = write(sio_localprinter->port_fd, &ptr[sent], size - sent);
    if (res < 0)
    {
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_localprinter_write: write error, res:%d, errno:%d\n", res, errno);
#endif
      return RIO_ERROR;
    }
    sent += res;
  }

#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_write: success, %d bytes written\n", sent);
#endif
  return sent;
}


/** Get the number of characters in the receive buffer, not implemented */
static long sio_localprinter_filelength(struct t_io_dev *io_dev)
{
#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_filelength called, function not implemented\n");
#endif

  return 0;
}


#if 0 /* Doesn't work on line printers */
/** Flush local printer buffers. queue_selector is one of TCIFLUSH, TCOFLUSH, TCIOFLUSH */
static int sio_localprinter_flush(struct s_sio_localprinter *sio_localprinter, int queue_selector)
{
  return ( tcflush(sio_localprinter->port_fd, queue_selector ) >= 0 ? RIO_SUCCESS : RIO_ERROR );
}
#endif


/** Find out the paper out status, which we map to CTS (paper out -> CTS=0) */
static int sio_localprinter_get_cts_state(struct s_sio_localprinter *sio_localprinter, bool *level)
{
  int status;
  int error;

  /* Query Linux line printer device status */
  if (ioctl(sio_localprinter->port_fd, LPGETSTATUS, &status) < 0)
  {
    return RIO_ERROR;
  }

  /* Work out if any errors */
  error = (~status & LP_PERRORP) ||  /* Error - printer on fire */
          (status & LP_POUTPA) ||    /* Paper out */
          (~status & LP_PSELECD);    /* Printer not selected */

  /* CTS (clear to send) = no errors */
  *level = !error;
  return RIO_SUCCESS;
}


/** Configure etc the printer serial port */
static int sio_localprinter_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_sio_localprinter *sio_localprinter = (struct s_sio_localprinter *) io_dev->dev_data;
  t_ioctl_siocfg *siocfg_ptr;
  bool bLevel;
  unsigned short *word_ptr;
  int result = RIO_SUCCESS;

#if HW_DEBUG(SIO_PRINTER_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localprinter_ioctl called\n");
#endif

  /* IOCTL function dispatch */
  switch (func)
  {
    case IOCTL_SER_SET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
      sio_local_printer_set_config(sio_localprinter, siocfg_ptr);
      break;

    case IOCTL_SER_GET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
      sio_local_printer_get_config(sio_localprinter, siocfg_ptr);
      break;

    /*
     * Get buffer levels and sizes
     */
#if 0
    case IOCTL_SER_GET_RX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      AbstractSerialGetRxBufferLevel(&sio_local->serial_port, &ui32Level);
      *word_ptr = ui32Level;
      break;
#endif
    case IOCTL_SER_GET_TX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = 0;
      break;
#if 0
    case IOCTL_SER_GET_RX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      AbstractSerialGetRxBufferSize(&sio_local->serial_port, &ui32Level);
      *word_ptr = ui32Level;
      break;
#endif
    case IOCTL_SER_GET_TX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = 1024;
      break;

    /*
     * Get the level of the CTS pin
     */
    case IOCTL_SER_GET_CTS_STATE:
      bLevel = false;
      result = sio_localprinter_get_cts_state(sio_localprinter, &bLevel);
      *va_arg(*arglist, unsigned char *) = bLevel;
      break;

    /*
     * Get the number of receive errors on this port
     */
    case IOCTL_SER_GET_RX_ERR_COUNT:
      *va_arg(*arglist, unsigned long *) = 0; /* TODO */
      break;

#if 0
    /*
     * Flush the transmit buffer
     */
    case IOCTL_SER_FLUSH_TX_BUF:
      result = sio_localprinter_flush(sio_localprinter, TCOFLUSH);
      break;

    /*
     * Flush the receive buffer
     */
    case IOCTL_SER_FLUSH_RX_BUF:
      result = sio_localprinter_flush(sio_localprinter, TCIFLUSH);
      break;
#endif

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = sio_localprinter->port_fd;
      result = RIO_SUCCESS;
      break;

    /*
     * Unknown function
     */
/*    case IOCTL_SER_SET_RING_NETWORK: This is unsupported on the onboard serial ports */
/*    case IOCTL_SER_SET_RS232_NOT_RS422: */
/*    case IOCTL_SER_SET_SOCKET_PORT: */
/*    case IOCTL_SER_GET_SOCKET_PORT: */
    default:
      result = RIO_ERROR;
      break;
  }
  return result;
}


/**
 * Create a serial port connected to local serial port hardware.
 *
 * @param io_dev is an uninitialised io_dev structure to be fill out as a serial port
 * @param port_name is the serial port device name (eg /dev/ttyS0, COM1)
 * @param rs422_info is control information if this is a RS422 port. Driver takes a local copy of this struct. Can be NULL is not RS422 port
 */
int rio_create_device_sio_localprinter(struct t_io_dev *io_dev, const char *port_name)
{
  struct s_sio_localprinter *sio_localprinter;

  /* Allocate printer device data here, sio_localprinter_destroy() frees it */
  sio_localprinter = malloc(sizeof(struct s_sio_localprinter));
  if (NULL == sio_localprinter)
  {
    return RIO_ERROR;
  }

  /* Set up the struct */
  sio_localprinter->port_fd = -1;
  sio_localprinter->port_name = strdup(port_name);
  if (NULL == sio_localprinter->port_name)
  {
    free(sio_localprinter);
    return RIO_ERROR;
  }

  /*
   * Setup device data pointer to point to the serial port
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) sio_localprinter;
  io_dev->status = io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_localprinter_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = sio_localprinter_read;
  io_dev->dev_write = sio_localprinter_write;
  io_dev->dev_filelength = sio_localprinter_filelength;
  io_dev->dev_ioctl = sio_localprinter_ioctl;
  io_dev->dev_destroy = sio_localprinter_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);
}
