/*
 * sio_local_socket.c -
 *
 * This type of rin serial device takes a TCP port number
 * and opens a listen socket, and accepts a single incoming
 * connection. You should really use sio_local_established_socket
 * and maintain the listen part in your application.
 * This is really only here for legacy, and should be removed
 * once nothing uses it any more.
 */
#include <sys/types.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/sockios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/rioctl.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinhw/timemath.h>

#include <rinhw/sio_local_socket.h>


/** Device data for a serial port mapped to a socket */
struct sSocketSerialPort
{
  int nSocketNumber;              /*!< The socket number to listen on. <=0 to get this data from configuration file based on dev->deviceid, >0 fixed to nSocketNumber */
  int nListenSocketFd;            /*!< The listen socket file descriptor */
  int nConnectionSocketFd;        /*!< The actual connection socket file descriptor */
  t_ioctl_siocfg siocfg;          /*!< The serial port configuration */
};


static void sio_localskt_set_config(struct sSocketSerialPort *pSktSerial, t_ioctl_siocfg *siocfg_ptr)
{
  /* Take a copy of the serial configuration */
  pSktSerial->siocfg = *siocfg_ptr;
}


static void sio_localskt_get_config(struct sSocketSerialPort *pSktSerial, t_ioctl_siocfg *siocfg_ptr)
{
  *siocfg_ptr = pSktSerial->siocfg;
}


/** Close the connection socket, but leave the listen socket open */
static void sio_localskt_connection_close(struct sSocketSerialPort *pSktSerial)
{
  close(pSktSerial->nConnectionSocketFd);
  pSktSerial->nConnectionSocketFd = -1;
}


/** Close both the connection socket and listen socket */
static void sio_localskt_all_close(struct sSocketSerialPort *pSktSerial)
{
  if (pSktSerial->nConnectionSocketFd >= 0)
  {
    sio_localskt_connection_close(pSktSerial);
  }
  if (pSktSerial->nListenSocketFd >= 0)
  {
    close(pSktSerial->nListenSocketFd);
    pSktSerial->nListenSocketFd = -1;
  }
}


/** Change the TCP port this serial port is bound to */
static int sio_localskt_change_bound_port(struct sSocketSerialPort *pSktSerial, int newport)
{
  int sktfd;
  struct sockaddr_in servaddr;
  int result;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_change_bound_port called, new port:%d\n", newport);
#endif

  /* Shutdown sockets */
  sio_localskt_all_close(pSktSerial);

  /* This is our new listen port */
  pSktSerial->nSocketNumber = newport;

  /* Setup new socket */
  sktfd = socket(AF_INET, SOCK_STREAM, 0);
  if (sktfd >= 0)
  {
    if (fcntl(sktfd, F_SETFD, FD_CLOEXEC) < 0)
    {
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("sio_localskt_change_bound_port: error setting FD_CLOEXEC for port %d (errno %d)\n", pSktSerial->nSocketNumber, errno);
#endif
    }

    /* Bind port to socket on any interface */
    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(pSktSerial->nSocketNumber);
    if (0 == bind(sktfd, (struct sockaddr *) &servaddr, sizeof(servaddr)))
    {
      if (0 == listen(sktfd, 1))
      {
        pSktSerial->nListenSocketFd = sktfd;
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("sio_localskt_change_bound_port: listen socket configured correctly on port %d\n", pSktSerial->nSocketNumber);
#endif
      }
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
      else
      {
        DEBUG_PRINTF("sio_localskt_change_bound_port: listen() fail, errno: %d\n", errno);
      }
#endif
    }
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
    else
    {
      DEBUG_PRINTF("sio_localskt_change_bound_port: bind() fail, errno: %d\n", errno);
    }
#endif
  }
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  else
  {
    DEBUG_PRINTF("sio_localskt_change_bound_port: socket() fail, errno: %d\n", errno);
  }
#endif

  result = (pSktSerial->nListenSocketFd >= 0) ? RIO_SUCCESS : RIO_ERROR;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_change_bound_port: result %d, listen sockfd %d\n", result, pSktSerial->nListenSocketFd);
#endif
  return result;
}


/** Initialise socket serial device */
static int sio_localskt_init(struct t_io_dev *io_dev)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  t_ioctl_siocfg siocfg;
  int result;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_init called\n");
#endif

  pSktSerial->nListenSocketFd = -1;
  pSktSerial->nConnectionSocketFd = -1;

  /* Default configuration for the serial port */
  siocfg.baud = SIO_DEF_BAUD;
  siocfg.data_bits = SIO_DEF_DATA_BITS;
  siocfg.parity_bits = SIO_DEF_PARITY_BITS;
  siocfg.stop_bits = SIO_DEF_STOP_BITS;
  siocfg.use_cts = SIO_DEF_USECTS;
  siocfg.use_terminations = SIO_DEF_TERMINATIONS;
  sio_localskt_set_config(pSktSerial, &siocfg);

  /* Bind to listen port */
  result = sio_localskt_change_bound_port(pSktSerial, pSktSerial->nSocketNumber);

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_init: result %d\n", result);
#endif
  return result;
}


/** Destroy serial device */
static int sio_localskt_destroy(struct t_io_dev *io_dev)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  int result;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* Shutdown sockets */
  sio_localskt_all_close(pSktSerial);

  /* Free up stuff we've allocated */
  free(io_dev->dev_data);
  io_dev->dev_data = NULL;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_destroy: result %d\n", result);
#endif
  return result;
}


static void sio_localskt_setblocking(int sockfd, bool bBlocking)
{
  int flags;

  flags = fcntl(sockfd, F_GETFL, 0);
  if (flags < 0)
  {
    return;
  }

  if (bBlocking)
  {
    flags &= ~O_NONBLOCK;
  }
  else
  {
    flags |= O_NONBLOCK;
  }
  fcntl(sockfd, F_SETFL, flags);
}


static int sio_localskt_setrecvtimeout(int sockfd, unsigned short timeout)
{
  struct timeval tv;
  int result;

  OSTicksToTimeval(&tv, timeout);
  result = setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, (char *) &tv,  sizeof(tv));

  return (result < 0) ? RIO_ERROR : RIO_SUCCESS;
}


static int sio_localskt_acceptconnection(struct sSocketSerialPort *pSktSerial, bool bBlocking, unsigned short timeout)
{
  fd_set readfds;
  struct timeval tv;
  struct timeval *ptv;
  int retval;
  int result;
  int newsockfd;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("sio_localskt_acceptconnection called\n");
#endif

  if (bBlocking)
  {
    /* We are not connected, so first wait for a connection, upto the timeout period */
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("sio_localskt_acceptconnection: block waiting for connection or timeout\n");
#endif
    sio_localskt_setblocking(pSktSerial->nListenSocketFd, true);

    /* Work out time out, 0 means indefinite */
    if (0 == timeout)
    {
      ptv = (struct timeval *) 0;
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("sio_localskt_acceptconnection: indefinite wait\n");
#endif
    }
    else
    {
      OSTicksToTimeval(&tv, timeout);
      ptv = &tv;
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("sio_localskt_acceptconnection: wait for tv.tv_sec=%ld, tv.tv_usec=%ld\n", ptv->tv_sec, ptv->tv_usec);
#endif
    }
  }
  else
  {
    /* Non-blocking, so use timeout of 0 with select to check for incoming connection */
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    ptv = &tv;
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("sio_localskt_acceptconnection: check for connection with immediate timeout: tv.tv_sec=%ld, tv.tv_usec=%ld\n", ptv->tv_sec, ptv->tv_usec);
#endif
  }

  /* Setup file descriptors to pend on */
  FD_ZERO(&readfds);
  FD_SET(pSktSerial->nListenSocketFd, &readfds);

  /* Wait for connection or (optional) timeout */
  retval = select(pSktSerial->nListenSocketFd + 1, &readfds, NULL, NULL, ptv);
  if (retval <= 0)
  {
    /* Timed out (=0) or more serious error (<0) */
    result = RIO_ERROR;
  }
  else if (!FD_ISSET(pSktSerial->nListenSocketFd, &readfds))
  {
    /* Not sure what is happening, select() indicated new connection, but FD_ISSET() doesn't agree */
    result = RIO_ERROR;
  }
  else
  {
    /* All good, so accept connection */
    newsockfd = accept(pSktSerial->nListenSocketFd, NULL, NULL);
    if (newsockfd >= 0)
    {
      if (fcntl(newsockfd, F_SETFD, FD_CLOEXEC) < 0)
      {
  #if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("sio_localskt_acceptconnection: error setting FD_CLOEXEC for new connection (errno %d)\n", errno);
  #endif
      }

      pSktSerial->nConnectionSocketFd = newsockfd;
      result = RIO_SUCCESS;
    }
    else
    {
      result = RIO_ERROR;
    }
  }

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("sio_localskt_acceptconnection: result %d, socket connection %d\n", result, pSktSerial->nConnectionSocketFd);
#endif
  return result;
}


/**
 * Function to receive len bytes (recv doesn't guarantee to return len bytes, unless used
 * with MSG_WAITALL which isn't defined on Windows < Windows server 2003)
 */
static int sio_localskt_recv_all(int sockfd, void *buf, size_t len)
{
  int retval;
  size_t numrecvd;
  char *pBuf;

  pBuf = (char *) buf;
  for (numrecvd = 0; numrecvd < len; )
  {
    retval = recv(sockfd, &pBuf[numrecvd], len - numrecvd, 0);
    if (retval <= 0)
    {
      return retval;
    }
    else
    {
      numrecvd += retval;
    }
  }
  return numrecvd;
}


/** Function to send len bytes (send doesn't guarantee to send len bytes) */
static int sio_localskt_send_all(int sockfd, const void *buf, size_t len)
{
  int retval;
  size_t numsent;
  const char *pBuf;

  pBuf = (char *) buf;
  for (numsent = 0; numsent < len; )
  {
    retval = send(sockfd, &pBuf[numsent], len - numsent, 0);
    if (retval <= 0)
    {
      return retval;
    }
    else
    {
      numsent += retval;
    }
  }
  return numsent;
}


/** Get the last error */
static int sio_localskt_get_last_error(void)
{
  return errno;
}


/** Read from the socket serial port */
static int sio_localskt_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  int retval = RIO_ERROR;
  int error;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_read called\n");
#endif

  /* Check socket has been setup correctly */
  if (pSktSerial->nListenSocketFd < 0)
  {
    return RIO_ERROR;
  }

  if (pSktSerial->nConnectionSocketFd < 0)
  {
    /* We are not connected, so first wait for a connection, upto the timeout period */
    retval = sio_localskt_acceptconnection(pSktSerial, true, timeout);
    if (retval <= 0)
    {
      /* Timed out (=0) or more serious error (<0) */
      return RIO_ERROR;
    }
  }

  /* If we are connected, then wait for data to read */
  if (pSktSerial->nConnectionSocketFd >= 0)
  {
    sio_localskt_setblocking(pSktSerial->nConnectionSocketFd, true);
    if (RIO_SUCCESS != sio_localskt_setrecvtimeout(pSktSerial->nConnectionSocketFd, timeout))
    {
      return RIO_ERROR;
    }
    retval = sio_localskt_recv_all(pSktSerial->nConnectionSocketFd, buf, len);
    if (retval < 0)
    {
      error = sio_localskt_get_last_error();
    }
    if ((retval < 0 && EWOULDBLOCK != error) || 0 == retval)
    {
      sio_localskt_connection_close(pSktSerial);
    }
  }
  return retval;
}


/** Write to the socket serial port */
static int sio_localskt_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  int retval;
  int error;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_write called\n");
#endif

  /* Check socket has been setup correctly */
  if (pSktSerial->nListenSocketFd < 0)
  {
    return RIO_ERROR;
  }

  if (pSktSerial->nConnectionSocketFd < 0)
  {
    /* We are not connected, so first see if there is a connection now, but don't wait */
    retval = sio_localskt_acceptconnection(pSktSerial, false, 0);
#if 0 /* Disregard result, so that if we are not connected then the write still occurs */
    if (retval <= 0)
    {
      /* Timed out (=0) or more serious error (<0) */
      return RIO_ERROR;
    }
#endif
  }

  /* If we are connected, then try to write data */
  if (pSktSerial->nConnectionSocketFd >= 0)
  {
    sio_localskt_setblocking(pSktSerial->nConnectionSocketFd, true);
    if (RIO_SUCCESS != sio_localskt_setrecvtimeout(pSktSerial->nConnectionSocketFd, timeout))
    {
      return RIO_ERROR;
    }
    retval = sio_localskt_send_all(pSktSerial->nConnectionSocketFd, buf, len);
    if (retval < 0)
    {
      error = sio_localskt_get_last_error();
    }
    if (retval < 0 && EWOULDBLOCK != error)
    {
      sio_localskt_connection_close(pSktSerial);
    }
  }
  else
  {
    /* We have no socket connected, so just discard the data and tell the caller that it was all written */
    retval = (int) len;
  }
  return retval;
}


static int sio_localskt_getbuffersize(int sockfd, bool bTxNotRx)
{
  int bufsize;
  socklen_t len;
  int res;
  int optname = (bTxNotRx) ? SO_SNDBUF : SO_RCVBUF;

  if (sockfd < 0)
  {
    return 0;
  }

  len = sizeof(bufsize);
  res = getsockopt(sockfd, SOL_SOCKET, optname, (void *) &bufsize, &len);
  if (res < 0)
  {
    bufsize = 0;
  }
  return bufsize;
}


/** Return the number of bytes available in either the receive or transmit buffers for a socket */
static int sio_localskt_getbufferlevel(int sockfd, bool bTxNotRx)
{
  int buflevel;
  int res;
  int ioctl_type = (bTxNotRx) ? SIOCOUTQ : SIOCINQ;

  if (sockfd < 0)
  {
    return 0;
  }
  res = ioctl(sockfd, ioctl_type, &buflevel);
  if (res < 0)
  {
    buflevel = 0;
  }
  return buflevel;
}


/** Get the number of characters in the receive buffer */
static long sio_localskt_filelength(struct t_io_dev *io_dev)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_filelength called\n");
#endif

  /* Check socket has been setup correctly */
  if (pSktSerial->nListenSocketFd < 0)
  {
    return RIO_ERROR;
  }

  if (pSktSerial->nConnectionSocketFd < 0)
  {
    /* We are not connected, so first see if there is a connection now, but don't wait */
    (void) sio_localskt_acceptconnection(pSktSerial, false, 0);
    /* Disregard result, because sio_localskt_getbufferlevel() can deal with no connected socket */
  }

  return (long) sio_localskt_getbufferlevel(pSktSerial->nConnectionSocketFd, false);
}


/** Use CTS to determine if there is a socket connection */
static void sio_localskt_ioctl_cts(struct sSocketSerialPort *pSktSerial, unsigned char *cts_state)
{
  *cts_state = (pSktSerial->nConnectionSocketFd >= 0);
}


/** Flush (discard) any incoming data from the socket */
static int sio_localskt_flush_incoming(struct sSocketSerialPort *pSktSerial)
{
  int level;
  char buffer;

  level = sio_localskt_getbufferlevel(pSktSerial->nConnectionSocketFd, false);
  while (level > 0)
  {
    if (sio_localskt_recv_all(pSktSerial->nConnectionSocketFd, &buffer, sizeof(buffer)) < 0)
      level = -1;
    else
      level--;
  }
  return (level < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/** Configure the socket serial port */
static int sio_localskt_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct sSocketSerialPort *pSktSerial = (struct sSocketSerialPort *) io_dev->dev_data;
  t_ioctl_siocfg *siocfg_ptr;
  unsigned short *word_ptr;
  int result = RIO_SUCCESS;

#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("sio_localskt_ioctl called, function %d\n", func);
#endif

  /* Check socket has been setup correctly */
  if (pSktSerial->nListenSocketFd < 0)
  {
    return RIO_ERROR;
  }

  if (pSktSerial->nConnectionSocketFd < 0)
  {
    /* We are not connected, so first see if there is a connection now, but don't wait */
    (void) sio_localskt_acceptconnection(pSktSerial, false, 0);
    /* Disregard result, because rioctl functions can deal with no connected socket */
  }

  /* IOCTL function dispatch */
  switch (func)
  {
    case IOCTL_SER_SET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_SET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      sio_localskt_set_config(pSktSerial, siocfg_ptr);
      break;

    case IOCTL_SER_GET_CONFIG:
      siocfg_ptr = va_arg(*arglist, t_ioctl_siocfg *);
#if HW_DEBUG(SIO_SOCKET_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("IOCTL_SER_GET_CONFIG, siocfg_ptr = %p\n", siocfg_ptr);
#endif
      sio_localskt_get_config(pSktSerial, siocfg_ptr);
      break;

    /* Get buffer levels and sizes */
    case IOCTL_SER_GET_RX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_localskt_getbufferlevel(pSktSerial->nConnectionSocketFd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_LEVEL:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_localskt_getbufferlevel(pSktSerial->nConnectionSocketFd, true);
      break;
    case IOCTL_SER_GET_RX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_localskt_getbuffersize(pSktSerial->nConnectionSocketFd, false);
      break;
    case IOCTL_SER_GET_TX_BUF_SIZE:
      word_ptr = va_arg(*arglist, unsigned short *);
      *word_ptr = (unsigned short) sio_localskt_getbuffersize(pSktSerial->nConnectionSocketFd, true);
      break;

    /* Get the level of the CTS pin */
    case IOCTL_SER_GET_CTS_STATE:
      sio_localskt_ioctl_cts(pSktSerial, va_arg(*arglist, unsigned char *));
      break;

    /* Get the number of receive errors on this port */
    case IOCTL_SER_GET_RX_ERR_COUNT:
      *va_arg(*arglist, unsigned long *) = 0;
      break;

    /* Change listen TCP port for this serial port */
    case IOCTL_SER_SET_SOCKET_PORT:
      result = sio_localskt_change_bound_port(pSktSerial, va_arg(*arglist, int));
      break;

    /* Get the listen TCP port for this serial port */
    case IOCTL_SER_GET_SOCKET_PORT:
      *va_arg(*arglist, int *) = pSktSerial->nSocketNumber;
      break;

    /* Flush the transmit buffer */
    case IOCTL_SER_FLUSH_TX_BUF:
      result = RIO_SUCCESS; /* Not really much we can do here, so let's just say it worked */
      break;

    /* Flush the receive buffer */
    case IOCTL_SER_FLUSH_RX_BUF:
      result = sio_localskt_flush_incoming(pSktSerial);
      break;

    /* Get Linux file descriptor */
    case IOCTL_GET_LINUX_FD:
      *va_arg(*arglist, int *) = pSktSerial->nConnectionSocketFd;
      result = RIO_SUCCESS;
      break;

    /* Unknown function */
/*    case IOCTL_SER_SET_RING_NETWORK: This is unsupported for socket serial ports */
    default:
      result = RIO_ERROR;
      break;
  }
  return result;
}


/** 
 * Create a local socket serial port, listening on a TCP port
 * 
 * This type of rin serial device takes a TCP port number
 * and opens a listen socket, and accepts a single incoming
 * connection. You should really use sio_local_established_socket
 * and maintain the listen part in your application.
 * This is really only here for legacy, and should be removed
 * once nothing uses it any more.
 * 
 * @param io_dev is the un-initialised rin io device structure
 * @param tcp_port is the TCP port to listen on
 */
int rio_create_device_sio_localskt(struct t_io_dev *io_dev, int tcp_port)
{
  struct sSocketSerialPort *sockserial;

  /* Allocate serial port socket info here, sio_localskt_destroy() frees it */
  sockserial = malloc(sizeof(struct sSocketSerialPort));
  if (NULL == sockserial)
  {
    return RIO_ERROR;
  }

  /* Set up the sock serial */
  sockserial->nSocketNumber = tcp_port;
  sockserial->nListenSocketFd = -1;
  sockserial->nConnectionSocketFd = -1;

  /* Set up the IO device */
  io_dev->dev_data = (void *) sockserial;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = sio_localskt_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = sio_localskt_read;
  io_dev->dev_write = sio_localskt_write;
  io_dev->dev_filelength = sio_localskt_filelength;
  io_dev->dev_ioctl = sio_localskt_ioctl;
  io_dev->dev_destroy = sio_localskt_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);
}
