/*
 * winserial.c
 *
 *  Created on: 23/10/2008
 *      Author: ianb
 */
#include <string.h>
#include "winserial.h"
#include <rinhw/sio.h>


#define TX_BUF_SIZE 1024
#define RX_BUF_SIZE 1024


/** Initialise the serial port structure with the port name. Doesn't open port. */
bool WinSerialInit( struct sWinSerialPort *pWinSerialPort, const char *pszPortName )
{
  pWinSerialPort->m_hSerialComm = INVALID_HANDLE_VALUE;

  pWinSerialPort->m_pszPortName = malloc( strlen(pszPortName) + 1 );
  if (0 == pWinSerialPort->m_pszPortName)
  {
    return false;
  }
  strcpy( pWinSerialPort->m_pszPortName, pszPortName );
  return true;
}


/** Close port if open and clean up allocated memory in struct */
void WinSerialShutdown( struct sWinSerialPort *pWinSerialPort )
{
  WinSerialClose( pWinSerialPort );

  if ( pWinSerialPort->m_pszPortName )
  {
    free( pWinSerialPort->m_pszPortName );
  }
}


/** Open the serial port */
bool WinSerialOpen( struct sWinSerialPort *pWinSerialPort )
{
  pWinSerialPort->m_hSerialComm = CreateFile( pWinSerialPort->m_pszPortName, /* Port Name */
                                              GENERIC_READ | GENERIC_WRITE, /* Desired Access */
                                              0, /* Shared Mode */
                                              NULL, /* Security */
                                              OPEN_EXISTING, /* Creation Disposition */
                                              FILE_FLAG_OVERLAPPED, /* Overlapped */
                                              NULL);

  return ( pWinSerialPort->m_hSerialComm != INVALID_HANDLE_VALUE );
}


/** Close the serial port */
bool WinSerialClose( struct sWinSerialPort *pWinSerialPort )
{
  if ( pWinSerialPort->m_hSerialComm != INVALID_HANDLE_VALUE )
  {
    CloseHandle( pWinSerialPort->m_hSerialComm );
    pWinSerialPort->m_hSerialComm = INVALID_HANDLE_VALUE;
  }
  return true;
}


/** Set the write timeout, 0 for indefinite */
bool WinSerialSetWriteTimeout( struct sWinSerialPort *pWinSerialPort, int32u ui32MilliSeconds )
{
  COMMTIMEOUTS commTimeout;

  /* Configuring Read & Write Time Outs */
  if ( GetCommTimeouts( pWinSerialPort->m_hSerialComm, &commTimeout ) )
  {
    commTimeout.WriteTotalTimeoutMultiplier = 0;
    commTimeout.WriteTotalTimeoutConstant = ui32MilliSeconds;
  }
  else
  {
    return false;
  }

  return (bool) SetCommTimeouts( pWinSerialPort->m_hSerialComm, &commTimeout );
}


/** Set the read timeout, 0 for indefinite */
bool WinSerialSetReadTimeout( struct sWinSerialPort *pWinSerialPort, int32u ui32MilliSeconds )
{
  COMMTIMEOUTS commTimeout;

  /* Configuring Read & Write Time Outs */
  if ( GetCommTimeouts( pWinSerialPort->m_hSerialComm, &commTimeout ) )
  {
    commTimeout.ReadIntervalTimeout = 0;
    commTimeout.ReadTotalTimeoutMultiplier = 0;
    commTimeout.ReadTotalTimeoutConstant = ui32MilliSeconds;
  }
  else
  {
    return false;
  }

  return SetCommTimeouts( pWinSerialPort->m_hSerialComm, &commTimeout );
}


/** Get the configuration for the serial port */
bool WinSerialGetConfig( struct sWinSerialPort *pWinSerialPort, int32u *pui32BaudRate,
                           int8u *pDataBits, int8u *pParity, int8u *pStopBits, bool *pUseCts )
{
  DCB dcbConfig;

  /* Configuring Serial Port Settings */
  if ( !GetCommState( pWinSerialPort->m_hSerialComm, &dcbConfig ) )
  {
    return false;
  }

  /* Baud rate */
  *pui32BaudRate = dcbConfig.BaudRate;

  /* Parity */
  switch (dcbConfig.Parity)
  {
    case EVENPARITY:
      *pParity = SIO_PARITY_EVEN;
      break;
    case ODDPARITY:
      *pParity = SIO_PARITY_ODD;
      break;
    default:
/*    case MARKPARITY:
    case SPACEPARITY:
    case NOPARITY: */
      *pParity = SIO_PARITY_NONE;
      break;
  }

  /* Stop bits */
  switch (dcbConfig.StopBits)
  {
    case TWOSTOPBITS:
      *pStopBits = SIO_STOPBITS_TWO;
      break;
    default:
/*    case ONESTOPBIT: */
      *pStopBits = SIO_STOPBITS_ONE;
      break;
  }

  /* Data bits */
  *pDataBits = dcbConfig.ByteSize;

  /* CTS control */
  *pUseCts = dcbConfig.fOutxCtsFlow;

  return true;
}


/** Set the configuration for the serial port */
bool WinSerialSetConfig( struct sWinSerialPort *pWinSerialPort, int32u ui32BaudRate,
                           int8u DataBits, int8u Parity, int8u StopBits, bool UseCts )
{
  DCB dcbConfig;

  /* Setup buffers */
  if ( !SetupComm( pWinSerialPort->m_hSerialComm, RX_BUF_SIZE, TX_BUF_SIZE ) )
  {
    return false;
  }

  /* Get current settings */
  if ( !GetCommState( pWinSerialPort->m_hSerialComm, &dcbConfig ) )
  {
    return false;
  }

  /* Configure Serial Port Settings */
  dcbConfig.BaudRate = ui32BaudRate;
  dcbConfig.fBinary = TRUE;
  dcbConfig.fParity = TRUE;
  dcbConfig.fOutxCtsFlow = UseCts;
  dcbConfig.fOutxDsrFlow = FALSE;
  dcbConfig.fDtrControl = DTR_CONTROL_DISABLE;
  dcbConfig.fDsrSensitivity = FALSE;
  dcbConfig.fTXContinueOnXoff = FALSE;
  dcbConfig.fOutX = FALSE;
  dcbConfig.fInX = FALSE;
  dcbConfig.fNull = FALSE;
  dcbConfig.fRtsControl = RTS_CONTROL_DISABLE;
  dcbConfig.fAbortOnError = FALSE;

  dcbConfig.ByteSize = DataBits;

  switch( Parity )
  {
    default:
    case SIO_PARITY_NONE:
      dcbConfig.Parity = NOPARITY;
      break;
    case SIO_PARITY_ODD:
      dcbConfig.Parity = ODDPARITY;
      break;
    case SIO_PARITY_EVEN:
      dcbConfig.Parity = EVENPARITY;
      break;
  }

  switch( StopBits )
  {
    default:
    case SIO_STOPBITS_ONE:
      dcbConfig.StopBits = ONESTOPBIT;
      break;
    case SIO_STOPBITS_TWO:
      dcbConfig.StopBits = TWOSTOPBITS;
      break;
  }

  /* Set new comport settings */
  if ( !SetCommState( pWinSerialPort->m_hSerialComm, &dcbConfig ) )
  {
    return false;
  }

  /* Set indefinite timeouts */
  if ( !WinSerialSetWriteTimeout( pWinSerialPort, 0 ) )
  {
    return false;
  }
  if ( !WinSerialSetReadTimeout( pWinSerialPort, 0 ) )
  {
    return false;
  }

  return true;
}

#if 0
/** Read data from the serial port */
bool WinSerialRead( struct sWinSerialPort *pWinSerialPort, int8u *pui8Buf, int32u ui32Size )
{
  DWORD dwRead = 0;
  char szBuf;
  DWORD dwIncommingReadSize;

  do
  {
    if ( ReadFile( pWinSerialPort->m_hSerialComm, &szBuf, 1, &dwIncommingReadSize, NULL ) != 0 )
    {
      if ( dwIncommingReadSize > 0 )
      {
        dwRead += dwIncommingReadSize;
        *pui8Buf++ = (int8u) szBuf;
      }
    }
    else
    {
      return false;
    }
  }
  while ( dwIncommingReadSize > 0 && dwRead < ui32Size );

  return true;
}
#endif


/** Read data from the serial port */
enum eAbstractSerialResult WinSerialRead( struct sWinSerialPort *pWinSerialPort, int8u *pui8Buf, int32u ui32Size )
{
  DWORD dwRead;
  OVERLAPPED osReader = { 0 };
  DWORD dwTotalRead = 0;
  DWORD dwTryToReadThisTime;
  bool bResult = true;

  while ( dwTotalRead < ui32Size && bResult )
  {

    /* Create the overlapped event. */
    osReader.hEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    if ( osReader.hEvent == NULL )
    {
      /* Error creating overlapped event; abort. */
      return asrrError;
    }

    /* Issue read operation. */
    dwTryToReadThisTime = ui32Size - dwTotalRead;
    if ( !ReadFile( pWinSerialPort->m_hSerialComm, &pui8Buf[ dwTotalRead ],
                    dwTryToReadThisTime, &dwRead, &osReader ) )
    {
      if ( GetLastError() != ERROR_IO_PENDING )
      {
        /* Error in communications; report it. */
        bResult = false;
      }
      else
      {
        /* Waiting on read to complete */
        if ( !GetOverlappedResult( pWinSerialPort->m_hSerialComm, &osReader, &dwRead, TRUE ) )
        {
          /* Error in communications; report it. */
          bResult = false;
        }
        else
        {
          /* Read completed successfully. */
          dwTotalRead += dwRead;
        }
      }
    }
    else
    {
      /* Read completed immediately */
      dwTotalRead += dwRead;
    }
    CloseHandle( osReader.hEvent );
  }
  return bResult ? asrSuccess : asrError; /* TODO return timeout asrTimeout */
}


#if 0
/** Write data to the serial port */
bool WinSerialWrite( struct sWinSerialPort *pWinSerialPort, const int8u *pui8Buf, int32u ui32Size )
{
  unsigned long dwNumberOfBytesSent = 0;
  DWORD dwNumberOfBytesWritten;

  while( dwNumberOfBytesSent < ui32Size )
  {

    if ( WriteFile( pWinSerialPort->m_hSerialComm, &pui8Buf[dwNumberOfBytesSent], 1, &dwNumberOfBytesWritten, NULL ) != 0 )
    {
      if( dwNumberOfBytesWritten > 0 )
      {
        ++dwNumberOfBytesSent;
      }
      else
      {
        return false;
      }
    }
    else
    {
      return false;
    }
  }
  return true;
}
#endif


/** Write data to the serial port */
enum eAbstractSerialResult WinSerialWrite( struct sWinSerialPort *pWinSerialPort, const int8u *pui8Buf, int32u ui32Size )
{
  OVERLAPPED osWrite = { 0 };
  DWORD dwNumberOfBytesSent = 0;
  DWORD dwNumberOfBytesWritten;
  DWORD dwNumberOfBytesToWriteThisTime;
  bool bResult = true;

  while ( dwNumberOfBytesSent < ui32Size && bResult )
  {

    /* Create this write's OVERLAPPED structure hEvent. */
    osWrite.hEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    if ( osWrite.hEvent == NULL )
    {
      /* Error creating overlapped event handle. */
      return asrError;
    }

    /* Do the write */
    dwNumberOfBytesToWriteThisTime = ui32Size - dwNumberOfBytesSent;
    if ( !WriteFile( pWinSerialPort->m_hSerialComm, &pui8Buf[ dwNumberOfBytesSent ],
                     dwNumberOfBytesToWriteThisTime, &dwNumberOfBytesWritten, &osWrite ) )
    {
      if ( GetLastError() != ERROR_IO_PENDING )
      {
        /* WriteFile failed, but it isn't delayed. Return error. */
        bResult = false;
      }
      else
      {
        /* Write is pending */
        if ( !GetOverlappedResult( pWinSerialPort->m_hSerialComm, &osWrite, &dwNumberOfBytesWritten, TRUE ) )
        {
          /* Something failed, return error. */
          bResult = false;
        }
        else
        {
          /* Write operation completed successfully. */
          dwNumberOfBytesSent += dwNumberOfBytesWritten;
        }
      }
    }
    else
    {
      /* WriteFile completed immediately. */
      dwNumberOfBytesSent += dwNumberOfBytesWritten;
    }
    CloseHandle( osWrite.hEvent );
  }
  return bResult ? asrSuccess : asrError; /* TODO return timeout asrTimeout */
}


/** Flush the serial port buffers. queue_selector is one of WINSERIAL_FLUSH_INPUT, WINSERIAL_FLUSH_OUTPUT or WINSERIAL_FLUSH_INOUT */
bool WinSerialFlush( struct sWinSerialPort *pWinSerialPort, int nWhichBuffers )
{
  DWORD flags;

  switch (nWhichBuffers)
  {
    case WINSERIAL_FLUSH_INPUT:
      flags = PURGE_RXCLEAR;
      break;
    case WINSERIAL_FLUSH_OUTPUT:
      flags = PURGE_TXCLEAR;
      break;
    case WINSERIAL_FLUSH_INOUT:
      flags = PURGE_RXCLEAR | PURGE_TXCLEAR;
      break;
    default:
      return false;
      break;
  }
  return ( PurgeComm( pWinSerialPort->m_hSerialComm, flags ) );
}


/** Get the number of bytes available in the transmit buffer */
bool WinSerialGetTxBufferLevel( struct sWinSerialPort *pWinSerialPort, int32u *pui32Level )
{
  COMSTAT comstat;
  DWORD dwErrors;

  if ( ClearCommError( pWinSerialPort->m_hSerialComm, &dwErrors, &comstat ) )
  {
    *pui32Level = comstat.cbOutQue;
    return true;
  }
  return false;
}


/** Get the size in bytes of the transmit buffer */
bool WinSerialGetTxBufferSize( struct sWinSerialPort *pWinSerialPort, int32u *pui32Level )
{
  *pui32Level = TX_BUF_SIZE;
  return true;
}


/** Get the number of bytes available in the receive buffer */
bool WinSerialGetRxBufferLevel( struct sWinSerialPort *pWinSerialPort, int32u *pdwLevel )
{
  COMSTAT comstat;
  DWORD dwErrors;

  if ( ClearCommError( pWinSerialPort->m_hSerialComm, &dwErrors, &comstat ) )
  {
    *pdwLevel = comstat.cbInQue;
    return true;
  }
  return false;
}


/** Get the size in bytes of the receive buffer */
bool WinSerialGetRxBufferSize( struct sWinSerialPort *pWinSerialPort, int32u *pui32Level )
{
  *pui32Level = RX_BUF_SIZE;
  return true;
}


/**
 * Get the state of the CTS pin (regardless of whether CTS blocking is enabled).
 * Returns 1 == Clear to send data, 0 == blocking
 */
bool WinSerialGetCtsState( struct sWinSerialPort *pWinSerialPort, bool *pbCts )
{
  DWORD dwModemStatus;

  if ( GetCommModemStatus( pWinSerialPort->m_hSerialComm, &dwModemStatus ) )
  {
    *pbCts = (0 != ( MS_CTS_ON & dwModemStatus ) );
    return true;
  }
  return false;
}
