#include <stdio.h>
#include <stdint.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdlib.h>
#include <rinhw/debug.h>

#include <rinhw/rinthread.h>


/**
 * Uncomment the following symbol to enable a canary based stack overflow check.
 * This will only detect one stack running down into another and isn't guaranteed
 * to even do that properly. Sprinkle calls to check_pthread_canary() throughout
 * the suspect code.
 */
/* #define THREAD_CANARY_DEBUG */


/* Parameters for the canary based checking */
#define CANARY_SIZE     64
#define CANARY_VALUE    0xdeadbeef
#define CANARY_TYPE     uint32_t


#ifdef THREAD_CANARY_DEBUG
#warning slow pthread stack overflow collision checking enabled.


static pthread_mutex_t pthread_canary_mutex = PTHREAD_MUTEX_INITIALIZER;


/** Linked list of thread canaries */
struct s_rin_thread_canary
{
  struct s_rin_thread_canary *next, *prev;
  CANARY_TYPE *mine;
};


/** List of thread canaries */
static struct s_rin_thread_canary *rin_canary_list = NULL;


struct s_rin_thread_args
{
  void *(*f)();
  void *a;
};


/**
 * Check for problems with thread stacks.
 *
 * Sprinkle calls to rin_thread_check_canary() throughout the suspect code.
 */
void rin_thread_check_canary(void)
{
  struct s_rin_thread_canary *p;
  size_t i;

  if (pthread_mutex_trylock(&pthread_canary_mutex) == 0)
  {
    for (p = rin_canary_list; p != NULL; p = p->next)
    {
      for (i = 0; i < CANARY_SIZE; i++)
      {
        if (p->mine[i] != CANARY_VALUE)
        {
          rinhw_debug_printf("rin_thread_check_canary: thread %p corrupt\n", p);
          break;
        }
      }
    }
    pthread_mutex_unlock(&pthread_canary_mutex);
  }
}


/** When a thread finishes, remove canary from linked list and free */
static void cleanup_canary(void *arg)
{
  struct s_rin_thread_canary *canary;

  if (NULL == arg)
    return;
  canary = (struct s_rin_thread_canary *) arg;

  if (pthread_mutex_lock(&pthread_canary_mutex) == 0)
  {
    if (canary->prev != NULL)
      canary->prev->next = canary->next;
    if (canary->next != NULL)
      canary->next->prev = canary->prev;
    if (rin_canary_list == canary)
      rin_canary_list = canary->next;
    pthread_mutex_unlock(&pthread_canary_mutex);

    free(canary);
    canary = NULL;
  }
}


/** Wrapper to setup thread canary */
static void *wrapper_start(void *args)
{
  struct s_rin_thread_args *arg;
  void *(*f)();
  void *a;
  CANARY_TYPE canary_mem[CANARY_SIZE];
  size_t i;
  struct s_rin_thread_canary *canary;

  /* Copy and free arguments */
  if (NULL == args)
    return NULL;
  arg = (struct s_rin_thread_args *) args;
  f = arg->f;
  a = arg->a;
  free(args);
  args = NULL;
  arg = NULL;

  /* Allocate canary list entry */
  canary = calloc(1, sizeof(struct s_rin_thread_canary));
  pthread_cleanup_push(&cleanup_canary, canary);
  for (i = 0; i < CANARY_SIZE; i++)
  {
    canary_mem[i] = CANARY_VALUE;
  }
  canary->mine = canary_mem;

  /* Add to canary list */
  if (pthread_mutex_lock(&pthread_canary_mutex) == 0)
  {
    canary->next = rin_canary_list;
    canary->prev = NULL;
    if (rin_canary_list != NULL)
      rin_canary_list->prev = canary;
    rin_canary_list = canary;
    pthread_mutex_unlock(&pthread_canary_mutex);
  }

  /* Call real thread function */
  a = f(a);

  /* Check canary, and clean up */
  rin_thread_check_canary();
  pthread_cleanup_pop(1);
  return a;
}


#else /* #ifdef THREAD_CANARY_DEBUG */


void rin_thread_check_canary(void)
{

}


#endif /* #else -> #ifdef THREAD_CANARY_DEBUG */


/** Exit a thread, rinstrum wrapper for pthread_exit() */
void rin_thread_exit(void *retval)
{
  pthread_exit(retval);
}


/**
 * Create a new thread.
 *
 * Rinstrum wrapper for pthread_create().
 * Permits checking of thread stack overflows when THREAD_CANARY_DEBUG is enabled
 */
bool rin_thread_create(pthread_t *thread, void *(*start_routine)(void *), void *arg, bool joinable)
{
  int res;
  pthread_attr_t attr;
  bool attr_inited = false;

  if (NULL == thread || NULL == start_routine)
    return false;

  /* Create pthread attribute */
  res = pthread_attr_init(&attr);
  if (0 != res)
  {
    goto rtc_fail;
  }
  attr_inited = true;

  /* Set joinable if requested */
  if (joinable)
  {
    res = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    if (0 != res)
    {
      goto rtc_fail;
    }
  }

#ifdef THREAD_CANARY_DEBUG
  {
    struct s_rin_thread_args *args;

    args = calloc(1, sizeof(struct s_rin_thread_args));
    if (NULL == args)
    {
      goto rtc_fail;
    }
    args->f = start_routine;
    args->a = arg;

    /* Create thread, wrapper_start() frees args */
    res = pthread_create(thread, &attr, wrapper_start, args);
    if (0 != res)
    {
      /* Failed, so we should free args */
      free(args);
      args = NULL;
    }
  }
#else
  res = pthread_create(thread, &attr, start_routine, arg);
#endif

  /* Run the canary on thread creation success */
  if (0 == res)
    rin_thread_check_canary();

rtc_fail:
  if (attr_inited)
    pthread_attr_destroy(&attr);
  return (0 == res);
}
