#include <limits.h>
#include <pthread.h>
#include <semaphore.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rstdio.h>
#include <rinhw/tick.h>
#include <rinhw/debug.h>
#include <rinhw/timemath.h>
#include <rinhw/rinthread.h>

#include <rinhw/bkgrnd.h>
#include "_bkgrnd.h"


/** This is a pointer that will maintain the linked list of functions to call */
static struct t_tick_fn *bg_fn_list = (struct t_tick_fn *) NULL;


/** Semaphore used to wake up background task to quit or after a new tick function has been added */
static sem_t bg_wakeup_sem;


/** Exclusive access mutex */
static pthread_mutex_t bg_lock_mutex;


/** Flag to signal that the thread should quit */
static bool bg_quit = false;


/** Background thread */
static pthread_t bg_thread;


static unsigned short RunBGFunctions(struct t_tick_fn *tick, unsigned short ticks_passed)
{
  unsigned short next_timeout = USHRT_MAX;

  while (tick != NULL)
  {

    /* If the tick function is enabled, check if its ready to run */
    if (tick->preset > 0)
    {

      /* Has enough time passed yet to run this function */
      if (tick->count <= ticks_passed)
      {

        /* Yes, call the function attached, and set the count back to the preset */
        if (tick->fn != NULL)
          tick->fn(tick->data);
        tick->count = tick->preset;
      }
      else
      {
        tick->count -= ticks_passed;
      }

      /* Find the next function to run */
      if (tick->count < next_timeout)
        next_timeout = tick->count;
    }
    tick = tick->next;
  }
  return next_timeout;
}


#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_FULL)
static void background_print_list(void)
{
  int num_fns;
  struct t_tick_fn *tick_fn;
  struct t_tick_fn *prev_fn;

  /* Find it in the list, first is a special case */
  for (num_fns = 0, prev_fn = NULL, tick_fn = bg_fn_list; NULL != tick_fn; prev_fn = tick_fn, tick_fn = tick_fn->next, num_fns++)
  {
    DEBUG_PRINTF("background_print_list: function %d, struct %p, preset %d, fn %p\n", num_fns, tick_fn, tick_fn->preset, tick_fn->fn);
  }
  if (0 == num_fns)
  {
    DEBUG_PRINTF("background_print_list: no background functions in list\n");
  }
}
#endif /* #if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_FULL) */



static void *background_task(void *pdata)
{
  unsigned short delay_time, next_timeout;
  unsigned long start_time, run_time;

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("background_task: started\n");
#endif

  delay_time = 0;
  while (1)
  {

    /*
     * Wait for a wake up by semaphore post, or a timeout,
     * and calculate the time taken
     */
    start_time = OSTimeGet();
    if (sem_timedwait_ticks(&bg_wakeup_sem, delay_time) < 0)
    {
      /* Timeout is a valid error, anything else is bad */
      if (ETIMEDOUT != errno)
      {
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("background_task: wake up semaphore pend failed\n");
#endif
        break;
      }
    }

    /* Quit if signaled */
    if (bg_quit)
    {
      break;
    }

    /* Get exclusive access */
    if (0 != pthread_mutex_lock(&bg_lock_mutex))
    {
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("background_task: exclusive access mutex lock failed\n");
#endif
      break;
    }

    /* Calculate how long we have been waiting for */
    run_time = OSTimeGet() - start_time;

    /* Keep the run time inside something we can handle */
    if (run_time > 65535)
      run_time = 65535;

    /* If there is something to run */
    if (bg_fn_list != NULL)
    {

      /*
       * Run any functions that have come due,
       * and calculate the time taken
       */
      start_time = OSTimeGet();
      next_timeout = RunBGFunctions(bg_fn_list, (unsigned short) run_time);
      run_time = OSTimeGet() - start_time;

      /* Keep the run time inside something we can handle */
      if (run_time > 65535)
        run_time = 65535;

      /*
       * Calculate how long to delay for based on how long it took to run the functions
       * Must do a delay of > 1, since 0 = delay indefinitely
       */
      if (run_time >= next_timeout)
        delay_time = 1;
      else
        delay_time = next_timeout - (unsigned short) run_time;
    }
    else
    {

      /* Nothing to run, delay indefinitely */
      delay_time = 0;
    }

    /* Release exclusive access */
    pthread_mutex_unlock(&bg_lock_mutex);
  }

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("background_task: quitting\n");
#endif
  rin_thread_exit(NULL);
}


/*
 * BackgroundAddFunction -
 *
 * Add a new function to call to the list of functions
 * Parameter new_tick_fn must have preset and function filled out
 */
void BackgroundAddFunction(struct t_tick_fn *new_tick_fn)
{
  struct t_tick_fn *tick_fn;

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundAddFunction(new_tick_fn=%p) called\n", new_tick_fn);
#endif

  /* Don't want to get interrupted while stuffing round with the list */
  if (0 != pthread_mutex_lock(&bg_lock_mutex))
  {
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("BackgroundAddFunction: exclusive access mutex lock failed\n");
#endif
    return;
  }

  /* Add new function to list */
  if (bg_fn_list == NULL)
  {
    /* Its the first one ever, just point the list pointer to it */
    bg_fn_list = new_tick_fn;
  }
  else
  {

    /* There is 1 or more already, navigate the list to find the end */
    tick_fn = bg_fn_list;
    while (tick_fn->next != NULL)
    {
      tick_fn = tick_fn->next;
    }

    /* Attach it to the end */
    tick_fn->next = new_tick_fn;
  }

  /*
   * Preset the new ones counter,
   * and make its next pointer null to mark the new end of the list
   */
  new_tick_fn->count = new_tick_fn->preset;
  new_tick_fn->next = (struct t_tick_fn *) NULL;

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_FULL)
  background_print_list();
#endif

  /* Out of critical section, and signal background task to run */
  pthread_mutex_unlock(&bg_lock_mutex);
  sem_post(&bg_wakeup_sem);

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundAddFunction finished\n");
#endif
}


/*
 * BackgroundRemoveFunction -
 *
 * Remove an existing function from the list of functions
 */
void BackgroundRemoveFunction(struct t_tick_fn *remove_tick_fn)
{
  struct t_tick_fn *tick_fn;
  struct t_tick_fn *prev_fn;

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundRemoveFunction(remove_tick_fn=%p) called\n", remove_tick_fn);
#endif

  /* Don't want to get interrupted while stuffing round with the list */
  if (0 != pthread_mutex_lock(&bg_lock_mutex))
  {
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("BackgroundRemoveFunction: exclusive access mutex lock failed\n");
#endif
    return;
  }

  /* Find it in the list, first is a special case */
  for (prev_fn = NULL, tick_fn = bg_fn_list; NULL != tick_fn; prev_fn = tick_fn, tick_fn = tick_fn->next)
  {
    if (tick_fn == remove_tick_fn)
    {
      /* Found it, remove it */
      if (NULL == prev_fn)
      {
        bg_fn_list = tick_fn->next;
      }
      else
      {
        prev_fn->next = tick_fn->next;
      }
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("BackgroundRemoveFunction: found and removed function\n");
#endif
    }
  }

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_FULL)
  background_print_list();
#endif

  /* Out of critical section */
  pthread_mutex_unlock(&bg_lock_mutex);

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundRemoveFunction finished\n");
#endif
}


/*
 * BackgroundChangePreset -
 *
 * Change the preset time of the background function. Count is reset when this is called,
 * so the preset interval will expire before the next function call.
 */
void BackgroundChangePreset(struct t_tick_fn *tick_fn, unsigned short preset)
{
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundChangePreset(tick_fn=%p, preset=%u) called\n", tick_fn, preset);
#endif

  /* Don't want to get interrupted while stuffing round with the list */
  if (0 != pthread_mutex_lock(&bg_lock_mutex))
  {
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("BackgroundChangePreset: exclusive access mutex lock failed\n");
#endif
    return;
  }

  /* Change preset */
  tick_fn->count = tick_fn->preset = preset;

  /* Release lock, and signal change to background task */
  pthread_mutex_unlock(&bg_lock_mutex);
  sem_post(&bg_wakeup_sem);
}


/*
 * BackgroundInit -
 *
 * Initialise the background task. Maintains a list of functions that
 * get called at preset intervals.
 * Parameter data is unused, and just for compatibility with init
 */
void BackgroundInit(void *data)
{
  bool wakeup_inited = false;
  bool lock_inited = false;
  bool task_inited = false;

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundInit called\n");
#endif

  /* Create a semaphore to wake the task, when a new background function is added */
  if (sem_init(&bg_wakeup_sem, 0, 0) < 0)
  {
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("BackgroundInit: error creating background task wake up semaphore\n");
#endif
    goto bi_fail;
  }
  wakeup_inited = true;

  /* Create a mutex for locking when stuffing round with the background variables */
  if (0 != pthread_mutex_init(&bg_lock_mutex, NULL))
  {
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("BackgroundInit: error creating background task mutex\n");
#endif
    goto bi_fail;
  }
  lock_inited = true;

  /* Create the background task */
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundInit: creating background task\n");
#endif
  bg_quit = false;
  bg_fn_list = NULL;
  if (!rin_thread_create(&bg_thread, background_task, NULL, true))
  {
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("BackgroundInit: create task failed\n");
#endif
    goto bi_fail;
  }
  task_inited = true;

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundInit finished\n");
#endif
  return;

bi_fail:
  if (task_inited)
  {
    bg_quit = true;
    sem_post(&bg_wakeup_sem);
    pthread_join(bg_thread, NULL);
  }
  if (lock_inited)
    pthread_mutex_destroy(&bg_lock_mutex);
  if (wakeup_inited)
    sem_destroy(&bg_wakeup_sem);
}


/** Shut down the background task */
void BackgroundDestroy(void)
{
#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("BackgroundDestroy called\n");
#endif

  /* Set task to quit, and wake task */
  bg_quit = true;
  sem_post(&bg_wakeup_sem);

  /* Wait for thread to quit */
  pthread_join(bg_thread, NULL);

  /* Destroy mutex & semaphore */
  pthread_mutex_destroy(&bg_lock_mutex);
  sem_destroy(&bg_wakeup_sem);

#if HW_DEBUG(BKGRND_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("BackgroundDestroy done\n");
#endif
}
