/*
 * timemath.c
 *
 *  Created on: 03/03/2011
 *      Author: ian
 */
#if defined(WIN32)
  #include <sys/timeb.h>
#endif
#include <rinhw/ucos_ii.h>

#include <rinhw/timemath.h>


#define MILLI_SEC_PER_SEC 1000
#define MICRO_SEC_PER_SEC 1000000
#define MICRO_SEC_PER_MILLI_SEC 1000
#define NANO_SEC_PER_SEC 1000000000
#define NANO_SEC_PER_MILLI_SEC 1000000


bool bGetCurrentTimeAsTimeSpec( struct timespec *pts )
{
#if defined(__linux__)
  return ( clock_gettime( CLOCK_REALTIME, pts ) >= 0 );
#elif defined(WIN32)

  /* This uses the same method to calculate time as pthreads-win32, so the absolute time in sem_timedwait should work */
  struct _timeb currSysTime;

  _ftime(&currSysTime);

  pts->tv_sec = currSysTime.time;
  pts->tv_nsec = currSysTime.millitm * NANO_SEC_PER_MILLI_SEC;
  return TRUE;
#endif
}


void OSTicksToTimespec( struct timespec *pts, uint16_t ticks )
{
  long nsec;

  pts->tv_sec = ticks / OS_TICKS_PER_SEC;
  nsec = ticks % OS_TICKS_PER_SEC;
  nsec *= NANO_SEC_PER_SEC / OS_TICKS_PER_SEC;
  pts->tv_nsec = nsec;
}


void OSTicksToTimeval( struct timeval *ptv, uint16_t ticks )
{
  long usec;

  ptv->tv_sec = ticks / OS_TICKS_PER_SEC;
  usec = ticks % OS_TICKS_PER_SEC;
  usec *= MICRO_SEC_PER_SEC / OS_TICKS_PER_SEC;
  ptv->tv_usec = usec;
}


void OSTimespecToTicks( uint32_t *pticks, const struct timespec *pts )
{
  uint32_t ticks;
  unsigned long long ns_ticks;

  ticks = pts->tv_sec * OS_TICKS_PER_SEC;
  ns_ticks = pts->tv_nsec;
  ns_ticks *= OS_TICKS_PER_SEC;
  ns_ticks /= NANO_SEC_PER_SEC;
  *pticks = ticks + ns_ticks;
}


/** pts1' = pts1 + pts2 */
void AddTimespec( struct timespec *pts1, const struct timespec *pts2 )
{
  const unsigned long int ns = pts1->tv_nsec + (unsigned long int)pts2->tv_nsec;
  pts1->tv_sec += ns / NANO_SEC_PER_SEC;
  pts1->tv_nsec = ns % NANO_SEC_PER_SEC;
  pts1->tv_sec += pts2->tv_sec;
}


/** pts1' = pts1 - pts2 */
void SubTimespec( struct timespec *pts1, const struct timespec *pts2 )
{
  if ( pts1->tv_sec < pts2->tv_sec ||
       ( pts1->tv_sec == pts2->tv_sec &&
         pts1->tv_nsec <= pts2->tv_nsec ) )
  {
    /* pts1 <= pts2, clamp to zero TODO calculate this correctly */
    pts1->tv_sec = 0;
    pts1->tv_nsec = 0;
  }
  else
  {
    /* pts1 > pts2 */
    pts1->tv_sec -= pts2->tv_sec;
    if (pts1->tv_nsec < pts2->tv_nsec)
    {
      /* Borrow a second */
      pts1->tv_nsec = pts1->tv_nsec + NANO_SEC_PER_SEC - pts2->tv_nsec;
      pts1->tv_sec--;
    }
    else
    {
      pts1->tv_nsec -= pts2->tv_nsec;
    }
  }
}


/** Wait on a pthreads semaphore for a number of uCOS ticks */
int sem_timedwait_ticks(sem_t *sem, uint16_t timeout)
{
  struct timespec ts, ts2;
  int res;

  if (0 == timeout)
  {
    /* Pend on semaphore indefinitely */
    res = sem_wait(sem);
  }
  else
  {
    /* Pend on semaphore with timeout */
    if (!bGetCurrentTimeAsTimeSpec(&ts))
    {
      res = -1;
    }
    else
    {
      OSTicksToTimespec(&ts2, timeout);
      AddTimespec(&ts, &ts2);
      res = sem_timedwait(sem, &ts);
    }
  }
  return res;
}
