/*
 * ucos_ii_linux.c
 *
 *  Created on: 11/08/2009
 *      Author: ian
 */
#define _POSIX_C_SOURCE 200112L
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <errno.h>
#include <string.h>
#include <rinhw/timemath.h>
#include <rinhw/bkgrnd.h>

#include <rinhw/ucos_ii.h>


struct sUcosPthreadXlat
{
  BOOLEAN bInUse;
  pthread_t Thread;
};

#define MILLI_SEC_PER_SEC 1000
#define NANO_SEC_PER_SEC 1000000000
#define NANO_SEC_PER_MILLI_SEC 1000000

OS_EVENT OSEventTbl[OS_MAX_EVENTS];/* Table of EVENT control blocks                   */


static struct timespec ts_start;
static pthread_mutex_t ts_start_mutex;


static void OSInitTime(void)
{
  (void) pthread_mutex_init(&ts_start_mutex, NULL);

  /* Work out the timeout */
  (void) pthread_mutex_lock(&ts_start_mutex);
  if (!bGetCurrentTimeAsTimeSpec(&ts_start))
  {
    ts_start.tv_sec = 0;
    ts_start.tv_nsec = 0;
  }
  (void) pthread_mutex_unlock(&ts_start_mutex);
}


static void OSInitEvents(void)
{
  int i;

  for ( i = 0; i < OS_MAX_EVENTS; i++ )
  {
    OSEventTbl[ i ].OSEventType = OS_EVENT_TYPE_UNUSED;
  }
}


/*
*********************************************************************************************************
*                                             INITIALIZATION
*
* Description: This function is used to initialize the internals of uC/OS-II and MUST be called prior to
*              creating any uC/OS-II object and, prior to calling OSStart().
*
* Arguments  : none
*
* Returns    : none
*********************************************************************************************************
*/

void  OSInit (void)
{
  OSInitTime();
  OSInitEvents();
}


void OSDestroy(void)
{
  /* TODO Implement this */
  pthread_mutex_destroy(&ts_start_mutex);
}


/*
*********************************************************************************************************
*                                           ACCEPT SEMAPHORE
*
* Description: This function checks the semaphore to see if a resource is available or, if an event
*              occurred.  Unlike OSSemPend(), OSSemAccept() does not suspend the calling task if the
*              resource is not available or the event did not occur.
*
* Arguments  : pevent     is a pointer to the event control block
*
* Returns    : >  0       if the resource is available or the event did not occur the semaphore is
*                         decremented to obtain the resource.
*              == 0       if the resource is not available or the event did not occur or,
*                         if 'pevent' is a NULL pointer or,
*                         if you didn't pass a pointer to a semaphore
*********************************************************************************************************
*/

INT16U  OSSemAccept (OS_EVENT *pevent)
{
  sem_t *sem;
  int res;
  int semval;

#if OS_ARG_CHK_EN > 0
  /* Validate 'pevent' */
  if (pevent == (OS_EVENT *)0)
  {
    return (0);
  }
#endif

  /* Validate event block type */
  if (pevent->OSEventType != OS_EVENT_TYPE_SEM)
  {
    return (0);
  }

  /* See if resource is available */
  sem = (sem_t *) pevent->OSEventPtr;
  res = sem_trywait( sem );
  if ( 0 == res )
  {
    return 0;
  }
  else
  {
    res = sem_getvalue( sem, &semval );
    return ( 0 == res ) ? semval : 0;
  }
}


/*
*********************************************************************************************************
*                                           CREATE A SEMAPHORE
*
* Description: This function creates a semaphore.
*
* Arguments  : cnt           is the initial value for the semaphore.  If the value is 0, no resource is
*                            available (or no event has occurred).  You initialize the semaphore to a
*                            non-zero value to specify how many resources are available (e.g. if you have
*                            10 resources, you would initialize the semaphore to 10).
*
* Returns    : != (void *)0  is a pointer to the event control clock (OS_EVENT) associated with the
*                            created semaphore
*              == (void *)0  if no event control blocks were available
*********************************************************************************************************
*/

OS_EVENT  *OSSemCreate (INT16U cnt)
{
  OS_EVENT *pevent = (OS_EVENT *) 0;
  sem_t *sem;
  int res;
  int i;

  for ( i = 0; i < OS_MAX_EVENTS; i++ )
  {
    if ( OS_EVENT_TYPE_UNUSED == OSEventTbl[ i ].OSEventType )
    {
      pevent = &OSEventTbl[ i ];
      break;
    }
  }
  if ( (OS_EVENT *) 0 == pevent )
  {
    return (OS_EVENT *) 0;
  }

  /* Set the event type */
  pevent->OSEventType = OS_EVENT_TYPE_SEM;

  /* Allocate memory for the semaphore */
  sem = (sem_t *) malloc( sizeof( sem_t ) );
  pevent->OSEventPtr = (void *) sem;

  /* Initialise the semaphore */
  res = sem_init( sem, 0, cnt );
  if ( res < 0 )
  {
    pevent->OSEventType = OS_EVENT_TYPE_UNUSED;
    free( sem );
  }
  return ( 0 == res ) ? pevent : (OS_EVENT *) 0;
}


/*
*********************************************************************************************************
*                                           PEND ON SEMAPHORE
*
* Description: This function waits for a semaphore.
*
* Arguments  : pevent        is a pointer to the event control block associated with the desired
*                            semaphore.
*
*              timeout       is an optional timeout period (in clock ticks).  If non-zero, your task will
*                            wait for the resource up to the amount of time specified by this argument.
*                            If you specify 0, however, your task will wait forever at the specified
*                            semaphore or, until the resource becomes available (or the event occurs).
*
*              err           is a pointer to where an error message will be deposited.  Possible error
*                            messages are:
*
*                            OS_NO_ERR           The call was successful and your task owns the resource
*                                                or, the event you are waiting for occurred.
*                            OS_TIMEOUT          The semaphore was not received within the specified
*                                                timeout.
*                            OS_ERR_EVENT_TYPE   If you didn't pass a pointer to a semaphore.
*                            OS_ERR_PEND_ISR     If you called this function from an ISR and the result
*                                                would lead to a suspension.
*                            OS_ERR_PEVENT_NULL  If 'pevent' is a NULL pointer.
*
* Returns    : none
*********************************************************************************************************
*/

void  OSSemPend (OS_EVENT *pevent, INT16U timeout, INT8U *err)
{
  sem_t *sem;
  int res;
  struct timespec ts, ts2;

#if OS_ARG_CHK_EN > 0
  /* Validate 'pevent' */
  if (pevent == (OS_EVENT *)0)
  {
    *err = OS_ERR_PEVENT_NULL;
    return;
  }
#endif
  /* Validate event block type */
  if (pevent->OSEventType != OS_EVENT_TYPE_SEM)
  {
    *err = OS_ERR_EVENT_TYPE;
    return;
  }

  sem = (sem_t *) pevent->OSEventPtr;
  if ( 0 == timeout )
  {
    /* Pend on semaphore indefinitely */
    res = sem_wait( sem );
  }
  else
  {
    /* Pend on semaphore with timeout */
    if ( !bGetCurrentTimeAsTimeSpec( &ts ) )
    {
      *err = OS_ERR_PEVENT_NULL;
      return;
    }
    OSTicksToTimespec( &ts2, timeout );
    AddTimespec( &ts, &ts2 );

    res = sem_timedwait( sem, &ts );
  }

  /* Sort out errors */
  if ( res < 0 )
  {
    *err = ( ETIMEDOUT == errno) ? OS_TIMEOUT : OS_ERR_PEVENT_NULL;
  }
  else
  {
    *err = OS_NO_ERR;
  }
}


/*
*********************************************************************************************************
*                                         POST TO A SEMAPHORE
*
* Description: This function signals a semaphore
*
* Arguments  : pevent        is a pointer to the event control block associated with the desired
*                            semaphore.
*
* Returns    : OS_NO_ERR           The call was successful and the semaphore was signaled.
*              OS_SEM_OVF          If the semaphore count exceeded its limit.  In other words, you have
*                                  signalled the semaphore more often than you waited on it with either
*                                  OSSemAccept() or OSSemPend().
*              OS_ERR_EVENT_TYPE   If you didn't pass a pointer to a semaphore
*              OS_ERR_PEVENT_NULL  If 'pevent' is a NULL pointer.
*********************************************************************************************************
*/

INT8U  OSSemPost (OS_EVENT *pevent)
{
  sem_t *sem;
  int res;

#if OS_ARG_CHK_EN > 0
  /* Validate 'pevent' */
  if (pevent == (OS_EVENT *)0)
  {
    return (OS_ERR_PEVENT_NULL);
  }
#endif

  /* Validate event block type */
  if (pevent->OSEventType != OS_EVENT_TYPE_SEM)
  {
    return (OS_ERR_EVENT_TYPE);
  }

  sem = (sem_t *) pevent->OSEventPtr;

  /* Post to the semaphore */
  res = sem_post( sem );
  return ( 0 == res ) ? OS_NO_ERR : OS_ERR_PEVENT_NULL;
}


/*
*********************************************************************************************************
*                                          QUERY A SEMAPHORE
*
* Description: This function obtains information about a semaphore
*
* Arguments  : pevent        is a pointer to the event control block associated with the desired
*                            semaphore
*
*              p_sem_data    is a pointer to a structure that will contain information about the
*                            semaphore.
*
* Returns    : OS_NO_ERR           The call was successful and the message was sent
*              OS_ERR_EVENT_TYPE   If you are attempting to obtain data from a non semaphore.
*              OS_ERR_PEVENT_NULL  If 'pevent' is a NULL pointer.
*********************************************************************************************************
*/

INT8U  OSSemQuery (OS_EVENT *pevent, OS_SEM_DATA *p_sem_data)
{
  sem_t *sem;
  int res;
  int semval;

#if OS_ARG_CHK_EN > 0
  /* Validate 'pevent' */
  if (pevent == (OS_EVENT *)0)
  {
    return OS_ERR_PEVENT_NULL;
  }
#endif

  /* Validate event block type */
  if (pevent->OSEventType != OS_EVENT_TYPE_SEM)
  {
    return OS_ERR_EVENT_TYPE;
  }

  /* Get semaphore count */
  sem = (sem_t *) pevent->OSEventPtr;
  res = sem_getvalue( sem, &semval );
  if ( 0 != res )
  {
    semval = 0;
  }
  if (semval > 65535)
  {
    semval = 65535;
  }
  p_sem_data->OSCnt = (INT16U) semval;

  /* Don't have valid data for anything except OSCnt  */
  memset( p_sem_data->OSEventTbl, 0, sizeof( p_sem_data->OSEventTbl ) );
  p_sem_data->OSEventGrp = 0;

  return OS_NO_ERR;
}


/*$PAGE*/
/*
*********************************************************************************************************
*                                         DELETE A SEMAPHORE
*
* Description: This function deletes a semaphore and readies all tasks pending on the semaphore.
*
* Arguments  : pevent        is a pointer to the event control block associated with the desired
*                            semaphore.
*
*              opt           determines delete options as follows:
*                            opt == OS_DEL_NO_PEND   Delete semaphore ONLY if no task pending
*                            opt == OS_DEL_ALWAYS    Deletes the semaphore even if tasks are waiting.
*                                                    In this case, all the tasks pending will be readied.
*
*              err           is a pointer to an error code that can contain one of the following values:
*                            OS_NO_ERR               The call was successful and the semaphore was deleted
*                            OS_ERR_DEL_ISR          If you attempted to delete the semaphore from an ISR
*                            OS_ERR_INVALID_OPT      An invalid option was specified
*                            OS_ERR_TASK_WAITING     One or more tasks were waiting on the semaphore
*                            OS_ERR_EVENT_TYPE       If you didn't pass a pointer to a semaphore
*                            OS_ERR_PEVENT_NULL      If 'pevent' is a NULL pointer.
*
* Returns    : pevent        upon error
*              (OS_EVENT *)0 if the semaphore was successfully deleted.
*
* Note(s)    : 1) This function must be used with care.  Tasks that would normally expect the presence of
*                 the semaphore MUST check the return code of OSSemPend().
*              2) OSSemAccept() callers will not know that the intended semaphore has been deleted unless
*                 they check 'pevent' to see that it's a NULL pointer.
*              3) This call can potentially disable interrupts for a long time.  The interrupt disable
*                 time is directly proportional to the number of tasks waiting on the semaphore.
*              4) Because ALL tasks pending on the semaphore will be readied, you MUST be careful in
*                 applications where the semaphore is used for mutual exclusion because the resource(s)
*                 will no longer be guarded by the semaphore.
*********************************************************************************************************
*/
OS_EVENT  *OSSemDel (OS_EVENT *pevent, INT8U opt, INT8U *err)
{
  sem_t *sem;
  int res;

#if OS_ARG_CHK_EN > 0
  /* Validate 'pevent' */
  if (pevent == (OS_EVENT *)0)
  {
    *err = OS_ERR_PEVENT_NULL;
    return pevent;
  }
#endif

  /* Validate event block type */
  if (pevent->OSEventType != OS_EVENT_TYPE_SEM)
  {
    *err = OS_ERR_EVENT_TYPE;
    return pevent;
  }

  sem = (sem_t *) pevent->OSEventPtr;

  /* Destroy the semaphore */
  res = sem_destroy( sem );
  if ( res < 0 )
  {
    *err = OS_ERR_EVENT_TYPE;
    return pevent;
  }

  /* Free up resources */
  free( sem );
  pevent->OSEventType = OS_EVENT_TYPE_UNUSED;
  pevent->OSEventPtr = 0;

  *err = OS_NO_ERR;
  return (OS_EVENT *) 0;
}


/*
*********************************************************************************************************
*                                DELAY TASK 'n' TICKS   (n from 0 to 65535)
*
* Description: This function is called to delay execution of the currently running task until the
*              specified number of system ticks expires.  This, of course, directly equates to delaying
*              the current task for some time to expire.  No delay will result If the specified delay is
*              0.  If the specified delay is greater than 0 then, a context switch will result.
*
* Arguments  : ticks     is the time delay that the task will be suspended in number of clock 'ticks'.
*                        Note that by specifying 0, the task will not be delayed.
*
* Returns    : none
*********************************************************************************************************
*/
void  OSTimeDly (INT16U ticks)
{
#if defined(__linux__)

  struct timespec req, rem;

  OSTicksToTimespec( &req, ticks );
  nanosleep(&req, &rem);
#elif defined(WIN32)
  INT32U milliseconds;

  milliseconds = ticks * MILLI_SEC_PER_SEC / OS_TICKS_PER_SEC;
  Sleep( milliseconds );
#endif
}


/** Internal version of OSTimeGet() with no locking */
static INT32U OSInternalTimeGet(void)
{
  struct timespec ts;
  uint32_t ticks;

  /* Get current time */
  if (!bGetCurrentTimeAsTimeSpec(&ts))
  {
    return 0;
  }

  /* Work out difference to start time */
  SubTimespec(&ts, &ts_start);
  OSTimespecToTicks(&ticks, &ts);
  return ticks;
}


/*
*********************************************************************************************************
*                                         GET CURRENT SYSTEM TIME
*
* Description: This function is used by your application to obtain the current value of the 32-bit
*              counter which keeps track of the number of clock ticks.
*
* Arguments  : none
*
* Returns    : The current value of OSTime
*********************************************************************************************************
*/
INT32U OSTimeGet(void)
{
  uint32_t ticks;

  (void) pthread_mutex_lock(&ts_start_mutex);
  ticks = OSInternalTimeGet();
  (void) pthread_mutex_unlock(&ts_start_mutex);
  return ticks;
}


/**
 * If you change the system time (ie settimeofday()),
 * you need to call this before changing the time, change the time, then call OSPostFixTimeGet()
 *
 * This will fix OSTimeGet() which relies on the start time of the application
 *
 * Example:
 * INT32U oldticks = OSPreFixTimeGet();
 * settimeofday();
 * OSPostFixTimeGet(oldticks);
 */
INT32U OSPreFixTimeGet(void)
{
  (void) pthread_mutex_lock(&ts_start_mutex);
  return OSInternalTimeGet();
}


/**
 * If you change the system time (ie settimeofday()),
 * you need to call this after changing the time
 *
 * This will fix OSTimeGet() which relies on the start time of the application
 *
 * Example:
 * INT32U oldticks = OSPreFixTimeGet();
 * settimeofday();
 * OSPostFixTimeGet(oldticks);
 */
void OSPostFixTimeGet(INT32U oldticks)
{
  struct timespec current;
  struct timespec delta;

  /* Adjust ts_start to compensate for changed time */
  if (!bGetCurrentTimeAsTimeSpec(&current))
  {
    goto opftg_done;
  }
  OSTicksToTimespec(&delta, oldticks);
  SubTimespec(&current, &delta);
  ts_start = current;

opftg_done:
  (void) pthread_mutex_unlock(&ts_start_mutex);
}
