/*
 * we21hw.c
 *
 *  Created on: 24/05/2010
 *      Author: ian
 */
#if defined(__linux__)
  #include <sys/ioctl.h>
  #include <fcntl.h>
  #include <unistd.h>
  #include <errno.h>
  #include <string.h>
  #include "we21_hw_access.h"
#endif /* #if defined(__linux__) */
#include <rinhw/rio.h>
#include <rinhw/debug.h>
#include <rinhw/iic_local_linux.h>

#include <rinhw/we21hw.h>


#if defined(__linux__)


/** File descriptor for hardware device */
static int we21hw_fd = -1;


/** Initialise access to the WE2111 hardware, via device node at devicename */
int rio_init_we21hw(const char *devicename)
{
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("we21hw_init called\n");
#endif

  /* Open WE2111 hardware device */
  we21hw_fd = -1;
  we21hw_fd = open(devicename, O_RDWR);
  if (we21hw_fd < 0)
  {
    ERROR_PRINTF("we21hw_init: couldn't open WE2111 hardware device '%s'\n", devicename);
    goto riw_fail;
  }
  if (fcntl(we21hw_fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    ERROR_PRINTF("we21hw_init: error setting FD_CLOEXEC on WE2111 hardware device '%s' (errno %d)\n", devicename, errno);
    goto riw_fail;
  }

  /* Set some defaults for the hardware */
  if (we21hw_set_rs422_termination(NULL, 0) < 0)
  {
    goto riw_fail;
  }
/*  we21hw_set_rs422_transmitenable(NULL, 0); */
  if (we21hw_select_rs232_not_rs422(NULL, 1) < 0)
  {
    goto riw_fail;
  }
  if (we21hw_set_slot_select(NULL, -1) < 0)
  {
    goto riw_fail;
  }

  /* Success */
  return RIO_SUCCESS;

riw_fail:
  if (we21hw_fd >= 0)
  {
    close(we21hw_fd);
  }
  we21hw_fd = -1;
  return RIO_ERROR;
}


/** Clean up after we21 hardware */
int rio_destroy_we21hw(void)
{
  if (we21hw_fd >= 0)
  {
    close(we21hw_fd);
  }
  we21hw_fd = -1;
  return RIO_SUCCESS;
}


/**
 * we21hw_set_rs422_termination -
 *
 * Call the Linux IOCTL to control the RS422 terminations
 * 0 = disable, 1 = enable
 */
int we21hw_set_rs422_termination(void *dummy, int enable)
{
  int ret;

#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("we21hw_set_rs422_termination called\n");
#endif

  /* Can't do anything if device is not open */
  if (we21hw_fd < 0)
  {
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("we21hw_set_rs422_termination: linux device not open\n");
#endif
    return RIO_ERROR;
  }

  /* Configure terminations */
  ret = ioctl(we21hw_fd, IOCTL_WE21HW_RS422_TERM, enable);
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  if (ret < 0)
  {
    DEBUG_PRINTF("we21hw_set_rs422_termination: IOCTL_WE21HW_RS422_TERM failed %d\n", ret);
  }
#endif
  return (ret < 0) ? RIO_ERROR : RIO_SUCCESS;
}


#if 0
/**
 *
 * As of WE21-103-120 RTS line controls transmit enable
 *
 * we21hw_set_rs422_transmitenable -
 *
 * Call the Linux IOCTL to control the RS422 transmit enable
 * 0 = disable, 1 = enable
 */
int we21hw_set_rs422_transmitenable(void *dummy, int enable)
{
  int ret;

#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("we21hw_set_rs422_transmitenable called\n");
#endif

  /* Can't do anything if device is not open */
  if (we21hw_fd < 0)
  {
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("we21hw_set_rs422_transmitenable: linux device not open\n");
#endif
    return RIO_ERROR;
  }

  /* Configure terminations */
  ret = ioctl(we21hw_fd, IOCTL_WE21HW_RS422_TXEN, enable);
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  if (ret < 0)
  {
    DEBUG_PRINTF("we21hw_set_rs422_transmitenable: IOCTL_WE21HW_RS422_TXEN failed %d\n", ret);
  }
#endif
  return (ret < 0) ? RIO_ERROR : RIO_SUCCESS;
}
#endif


/**
 * we21hw_select_rs232_not_rs422 -
 *
 * Call the Linux IOCTL to control the RS232/RS422 selection
 * 0 = RS422, 1 = RS232
 */
int we21hw_select_rs232_not_rs422(void *dummy, int select)
{
  int ret;

#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("we21hw_select_rs232_not_rs422 called\n");
#endif

  /* Can't do anything if device is not open */
  if (we21hw_fd < 0)
  {
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("we21hw_select_rs232_not_rs422: linux device not open\n");
#endif
    return RIO_ERROR;
  }

  /* Configure terminations */
  ret = ioctl(we21hw_fd, IOCTL_WE21HW_SELECT_RS232_NOT_RS422, select);
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  if (ret < 0)
  {
    DEBUG_PRINTF("we21hw_select_rs232_not_rs422: IOCTL_WE21HW_SELECT_RS232_NOT_RS422 failed %d\n", ret);
  }
#endif
  return (ret < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/**
 * we21hw_set_slot_select -
 *
 * Call the Linux IOCTL to set the slot select for a rinblock
 * -1 will clear all
 */
int we21hw_set_slot_select(void *dummy, int slot)
{
  int ret;

#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("we21hw_set_slot_select called\n");
#endif

  /* Can't do anything if device is not open */
  if (we21hw_fd < 0)
  {
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("we21hw_set_slot_select: linux device not open\n");
#endif
    return RIO_ERROR;
  }

  /* Configure slot */
  ret = ioctl(we21hw_fd, IOCTL_WE21HW_I2C_SLOT_SELECT, slot);
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  if (ret < 0)
  {
    DEBUG_PRINTF("we21hw_set_slot_select: IOCTL_WE21HW_I2C_SLOT_SELECT failed %d\n", ret);
  }
#endif
  return (ret < 0) ? RIO_ERROR : RIO_SUCCESS;
}


#if 1

/** Wait for ADC data or timeout (in uCOS ticks), using the IOCTL_WE21HW_ADC_WAIT_DATA */
int we21hw_adc_wait_data(int32u timeout)
{
  int msecs;
  int ret;

  /* Can't do anything if device is not open */
  if (we21hw_fd < 0)
  {
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("we21hw_adc_wait_data: linux device not open\n");
#endif
    return RIO_ERROR;
  }

  if (0 == timeout)
  {
    /* uCOS ticks == 0 means indefinite */
    msecs = -1;
  }
  else
  {
    msecs = timeout * 1000;
    msecs /= OS_TICKS_PER_SEC;
  }

  /* Wait for ADC data */
  ret = ioctl(we21hw_fd, IOCTL_WE21HW_ADC_WAIT_DATA, msecs);
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  if (ret < 0)
  {
    DEBUG_PRINTF("we21hw_adc_wait_data: IOCTL_WE21HW_ADC_WAIT_DATA failed %d\n", ret);
  }
#endif
  return (ret < 0) ? RIO_ERROR : RIO_SUCCESS;
}

#else

/** Wait for ADC data or timeout (in uCOS ticks), using select() */
int we21hw_adc_wait_data(int32u timeout)
{
  fd_set rfds;
  struct timeval tv;
  struct timeval *ptv;
  int msecs;
  int ret;

  /* Can't do anything if device is not open */
  if (we21hw_fd < 0)
  {
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("we21hw_adc_wait_data: linux device not open\n");
#endif
    return RIO_ERROR;
  }

  /* Setup file descriptors */
  FD_ZERO(&rfds);
  FD_SET(we21hw_fd, &rfds);

  /* uCOS ticks->milliseconds */
  msecs = timeout * 1000;
  msecs /= OS_TICKS_PER_SEC;

  /* Setup timeout (in milliseconds) */
  tv.tv_sec = msecs / 1000;
  tv.tv_usec = (msecs % 1000) * 1000;

  /* Indefinite wait for timeout == 0 */
  ptv = (0 == timeout) ? NULL : &tv;

  /* Wait for ADC data */
  ret = select(we21hw_fd + 1, &rfds, NULL, NULL, ptv);
#if HW_DEBUG(WE21HW_DEBUG, HW_DEBUG_BASIC)
  if (ret < 0)
  {
    DEBUG_PRINTF("we21hw_adc_wait_data: IOCTL_WE21HW_ADC_WAIT_DATA failed %d, errno:%d,%s\n", ret, errno, strerror(errno));
  }
#endif
  return (ret < 0) ? RIO_ERROR : RIO_SUCCESS;
}

#endif


#else /* #if defined(__linux__) */
/* Dunno how to implement these under windows, but at least stubs will let us compile */

/**
 * init_we21hw -
 *
 * Initialise access to the WE2111 hardware,
 * device name is retrieved from XML config file
 * under hwlib_config/we21hw/device
 */
void init_we21hw(void *data)
{
}


/** Clean up after we21 hardware */
void destroy_we21hw(void)
{
}


/**
 * we21hw_set_rs422_termination -
 *
 * Call the Linux IOCTL to control the RS422 terminations
 * 0 = disable, 1 = enable
 */
int we21hw_set_rs422_termination(int enable)
{
  return RIO_SUCCESS;
}


/**
 * we21hw_set_rs422_transmitenable -
 *
 * Call the Linux IOCTL to control the RS422 transmit enable
 * 0 = disable, 1 = enable
 */
int we21hw_set_rs422_transmitenable(int enable)
{
  return RIO_SUCCESS;
}


/**
 * we21hw_select_rs232_not_rs422 -
 *
 * Call the Linux IOCTL to control the RS232/RS422 selection
 * 0 = RS422, 1 = RS232
 */
int we21hw_select_rs232_not_rs422(int select)
{
  return RIO_SUCCESS;
}


/**
 * we21hw_set_slot_select -
 *
 * Call the Linux IOCTL to set the slot select for a rinblock
 * -1 will clear all
 */
int we21hw_set_slot_select(int slot)
{
  return RIO_SUCCESS;
}


/** Wait for ADC data or timeout (in uCOS ticks) */
int we21hw_adc_wait_data(int32u timeout)
{
  return RIO_SUCCESS;
}

#endif /* #else -> #if defined(__linux__) */
