#include <ctype.h>
#include <rinhw/flash2238.h>
#include <rinhw/rstdio.h>
#include "dumpdev.h"

#include "flashtest.h"




static void EraseFlashBlock(struct t_io_dev *dev, struct t_io_dev *flash_dev)
{
    long block_address;
    int res;

    rfprintf(dev, "Enter address of block to erase: ");
    block_address = (long) GetLong(dev);

    if (rlseek(flash_dev, block_address, SEEK_SET) < 0)
    {
        rfprintf(dev, "Seek error, abort\n");
        return;
    }
    res = rioctl(flash_dev, IOCTL_FLASH_ERASE_BLOCK);

    rfprintf(dev, "Result of flash erase block address 0x%lx = %d\n", block_address, res);
}


static unsigned char GetChar(struct t_io_dev *dev)
{
    unsigned char ch;

    rread(dev, &ch, 1, 0);
    return ch;
}


unsigned char CharToNib(unsigned char ch, unsigned char *data)
{
    if ((ch >= '0') && (ch <= '9'))
        *data = (unsigned char) (ch - '0');
    else if ((ch >= 'A') && (ch <= 'F'))
        *data = (unsigned char) (ch - 'A' + 10);
    else if ((ch >= 'a') && (ch <= 'f'))
        *data = (unsigned char) (ch - 'a' + 10);
    else
        return 0;
    return 1;
}


unsigned char GetHexByte(struct t_io_dev *dev, unsigned char *data)
{
    unsigned char ch, d1, d2;

    ch = GetChar(dev);
    if (CharToNib(ch, &d1) == 0)
        return 0;

    ch = GetChar(dev);
    if (CharToNib(ch, &d2) == 0)
        return 0;

    *data = (unsigned char)  ((d1 << 4) | (d2 & 0x0F));
    return 1;
}


static int LoadS2Line(struct t_io_dev *dev, unsigned char *buf, unsigned short max_len, unsigned long *prog_addr)
{
    unsigned char line_len, i, temp, chksum;
    unsigned long address;

    /*
     * Now two chars (1 byte) for s line length
     */
    if (GetHexByte(dev, &line_len) == 0)
    {
        rfprintf(dev, "Bad S Line length\n");
        return -1;
    }
    if (line_len < 4)
    {
        rfprintf(dev, "S Line length too short (<4)\n");
        return -1;
    }
    chksum = line_len;

    line_len -= 4;
    if (line_len > max_len)
    {
        rfprintf(dev, "S Line length too long for this parser (>%d)\n", max_len);
        return -1;
    }

    /*
     * Now 6 chars (3 bytes) for address
     */
    if (GetHexByte(dev, &temp) == 0)
    {
        rfprintf(dev, "S Line bad address byte 1\n");
        return -1;
    }
    address = (unsigned long) (temp) << 16;
    chksum += temp;
    if (GetHexByte(dev, &temp) == 0)
    {
        rfprintf(dev, "S Line bad address byte 2\n");
        return -1;
    }
    address += (unsigned long) (temp) << 8;
    chksum += temp;
    if (GetHexByte(dev, &temp) == 0)
    {
        rfprintf(dev, "S Line bad address byte 3\n");
        return -1;
    }
    address += temp;
    chksum += temp;

    /*
     * Now get line_len * 2 bytes of data
     */
    for (i = 0; i < line_len; i++)
    {
        if (GetHexByte(dev, &temp) == 0)
        {
            rfprintf(dev, "S Line bad data byte %d\n", i);
            return -1;
        }
        buf[i] = temp;
        chksum += temp;
    }

    /*
     * Calc local checksum
     */
    chksum = (unsigned char) (0xFF - chksum);

    /*
     * Now two chars (1 byte) for checksum
     */
    if (GetHexByte(dev, &temp) == 0)
    {
        rfprintf(dev, "S Line bad checksum byte %d\n", i);
        return -1;
    }
    if (temp != chksum)
    {
        rfprintf(dev, "S Line checksums dont match (calcd %d, actual %d)\n", chksum, temp);
        return -1;
    }

    /*
     * Return address and length
     */
    *prog_addr = address;
    return (int) line_len;
}


static unsigned char ProgramSLine(struct t_io_dev *dev, struct t_io_dev *flash_dev, unsigned long prog_addr, unsigned char *data, unsigned char length)
{
    int res;

    /*
     * Seek the posn in flash
     */
#if (FLASH_START_OFFSET != 0)
    prog_addr -= FLASH_START_OFFSET;
#endif
    if (rlseek(flash_dev, (long) prog_addr, SEEK_SET) < 0)
    {
        rfprintf(dev, "Flash seek error, abort\n");
        return 0;
    }

    res = rwrite(flash_dev, data, (unsigned short) length, 0);
    return (res == length);
}


#define MAX_LINE_BUF_LEN 16


static void S2ProgramFlash(struct t_io_dev *dev, struct t_io_dev *flash_dev)
{
    unsigned char ch, s_type;
    unsigned char line_buf[MAX_LINE_BUF_LEN];
    unsigned long prog_addr;
    int len;

    while (1)
    {

        /*
         * Wait for S at start of line
         */
        do
        {
            ch = GetChar(dev);
            if (ch == 0x1B)
            {
                rfprintf(dev, "Escape from S Loader\n");
                return;
            }
        }
        while (ch != 'S');

        /*
         * Next one should be line type
         */
        s_type = GetChar(dev);
        switch (s_type)
        {
            case '2':
                len = LoadS2Line(dev, line_buf, MAX_LINE_BUF_LEN, &prog_addr);
                if (len < 0)
                {
                    rfprintf(dev, "Load S Line failed\n");
                    return;
                }
                break;
            default:
                rfprintf(dev, "Bad S Line Type\n");
                return;
        }

        /*
         * Program the data
         */
        if (len > 0)
        {
            if (ProgramSLine(dev, flash_dev, prog_addr, line_buf, (unsigned char) len) == 0)
            {
                rfprintf(dev, "Progamming failed\n");
                return;
            }
        }
    }
}


#define PROG_LINE_SIZE 128
#define PROG_DATA 0xA5
static unsigned char prog_buf[PROG_LINE_SIZE];


static void ProgramFlashLine(struct t_io_dev *dev, struct t_io_dev *flash_dev)
{
    unsigned short i;
    long address;
    int res;

    rfprintf(dev, "Flash Program 1 Line (%d bytes with 0x%02x)\nEnter start programming address: ",
             PROG_LINE_SIZE, PROG_DATA);
    address = GetLong(dev);
    if (rlseek(flash_dev, address, SEEK_SET) < 0)
    {
        rfprintf(dev, "Seek error, abort\n");
        return;
    }

    for (i = 0; i < PROG_LINE_SIZE; i++)
    {
        prog_buf[i] = PROG_DATA;
    }

    res = rwrite(flash_dev, prog_buf, PROG_LINE_SIZE, 0);
    rfprintf(dev, "Result of flash device write line to address 0x%lx = %d\n", address, res);
}


void TestFlash2238(struct t_io_dev *dev, struct t_io_dev *flashdev)
{
    unsigned char ch;

    ropen(flashdev, 0);

    do
    {
        rfprintf(dev, "Flash Test\nQ - Quit\nE - Erase Flash Block\nP - Program Flash Line\nD - Dump Flash\nS - S Record Dump Flash\nL - Load S Records into Flash\n");

        rread(dev, &ch, 1, 0);
        switch (ch)
        {
            case 'E':
                EraseFlashBlock(dev, flashdev);
                break;
            case 'P':
                ProgramFlashLine(dev, flashdev);
                break;
            case 'D':
                DumpBlockDevice(dev, flashdev);
                break;
            case 'S':
                S2DumpBlockDevice(dev, flashdev);
                break;
            case 'L':
                S2ProgramFlash(dev, flashdev);
                break;

        }
    }
    while (ch != 'Q');

    rclose(flashdev);
}


#if 0
void InitFlash2238(struct t_io_dev *dev)
{
    int result;

    result = rregister_io(&dev_flash2238);

    rfprintf(dev, "Result of registering flash device %d\n", result);
}
#endif
