#include <string.h>
#include <rinhw/rio.h>
#include <rinhw/rstdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/iic.h>

#include "iictest.h"


#define DEF_ADDR 0x16
#define RUN_ADDR 2

struct t_io_dev *sdev = NULL;



unsigned char WriteIIC(struct t_io_dev *iic_dev, unsigned char address, unsigned char *buf, unsigned char length)
{
    int res;

    if (rioctl(iic_dev, IOCTL_IIC_SET_SLAVE_ADDR, address) < 0)
        return 0;
    res = rwrite(iic_dev, buf, (unsigned short) length, OS_TICKS_PER_SEC);
    if (res != length)
        rfprintf(sdev, "WriteIIC: write returned %d, length %d\r\n", res, length);
    return (res == length);
}


unsigned char ReadIIC(struct t_io_dev *iic_dev, unsigned char address, unsigned char *buf, unsigned char length)
{
    int res;

    if (rioctl(iic_dev, IOCTL_IIC_SET_SLAVE_ADDR, address) < 0)
        return 0;
    res = rread(iic_dev, buf, (unsigned short) length, OS_TICKS_PER_SEC);
//    if (res != length)
//        rfprintf(sdev, "ReadIIC: read returned %d, length %d\r\n", res, length);
    return (res == length);
}


unsigned char WriteReadIIC(struct t_io_dev *iic_dev, unsigned char address, unsigned char *buf, unsigned char wlen, unsigned char rlen)
{
    unsigned long time;

    if (!WriteIIC(iic_dev, address, buf, wlen))
        return 0;

    time = OSTimeGet();
    do
    {
        if (OSTimeGet() >= (time + OS_TICKS_PER_SEC))
            return 0;
    }
    while (!ReadIIC(iic_dev, address, buf, rlen));
    return 1;
}


unsigned char ModuleStartup(struct t_io_dev *iic_dev)
{
    unsigned char buf[2];

    /*
     * Set address
     */
    rfprintf(sdev, "Set module address: ");
    buf[0] = 0x41;
    buf[1] = RUN_ADDR;
    if (!WriteReadIIC(iic_dev, DEF_ADDR, buf, 2, 1))
    {
        rfprintf(sdev, "no reply\r\n");
        return 0;
    }
    if ((buf[0] & 0x80) != 0)
    {
        rfprintf(sdev, "reply with error\r\n");
        return 0;
    }
    rfprintf(sdev, "success\r\n");

    /*
     * Set run mode
     */
    rfprintf(sdev, "Set module mode to run: ");
    buf[0] = 0x42;
    buf[1] = 0x03;  // Mode run
    if (!WriteReadIIC(iic_dev, RUN_ADDR, buf, 2, 1))
    {
        rfprintf(sdev, "no reply\r\n");
        return 0;
    }
    if ((buf[0] & 0x80) != 0)
    {
        rfprintf(sdev, "reply with error\r\n");
        return 0;
    }
    rfprintf(sdev, "success\r\n");
    return 1;
}


void RepeatedGetAddress(struct t_io_dev *iic_dev)
{
    unsigned char buf[2];

//    while (reof(sdev))
    while (1)
    {

        /*
         * Get address
         */
//        rfprintf(sdev, "Get address from module: ");
        buf[0] = 0x01;
        if (!WriteReadIIC(iic_dev, RUN_ADDR, buf, 2, 2))
        {
            rfprintf(sdev, "no reply\r\n");
            return;
        }
        if ((buf[0] & 0x80) != 0)
        {
            rfprintf(sdev, "reply with error\r\n");
            return;
        }
//        rfprintf(sdev, "success, address %d\r\n", buf[1]);
    }
    rread(sdev, buf, 1, 0);
}


void GetAddress(struct t_io_dev *iic_dev)
{
    unsigned char buf[2];

    /*
     * Get address
     */
    buf[0] = 0x01;
    if (!WriteReadIIC(iic_dev, RUN_ADDR, buf, 1, 2))
    {
        rfprintf(sdev, "Get address from module: no reply\r\n");
        return;
    }
    if ((buf[0] & 0x80) != 0)
    {
        rfprintf(sdev, "Get address from module: reply with error\r\n");
        return;
    }
    rfprintf(sdev, "Get address from module: success, address %d\r\n", buf[1]);
}


void GetTag(struct t_io_dev *iic_dev)
{
    unsigned char buf[32], i;

    /*
     * Get tag
     */
    buf[0] = 0x04;
    if (!WriteReadIIC(iic_dev, RUN_ADDR, buf, 1, 17))
    {
        rfprintf(sdev, "Get tag from module: no reply\r\n");
        return;
    }
    if ((buf[0] & 0x80) != 0)
    {
        rfprintf(sdev, "Get tag from module: reply with error\r\n");
        return;
    }
    rfprintf(sdev, "Get tag from module: success\r\n");
    for (i = 0; i < 16; i++)
    {
        rfprintf(sdev, "%02x ", buf[i]);
    }
    rfprintf(sdev, "\r\n");
}


void RepeatedGetTag(struct t_io_dev *iic_dev)
{
    unsigned char buf[32];

    while (reof(sdev))
    {

        /*
         * Get tag
         */
        buf[0] = 0x04;
        if (!WriteReadIIC(iic_dev, RUN_ADDR, buf, 1, 17))
        {
            rfprintf(sdev, "Get tag from module: no reply\r\n");
            return;
        }
        if ((buf[0] & 0x80) != 0)
        {
            rfprintf(sdev, "Get tag from module: reply with error\r\n");
            return;
        }
        rfprintf(sdev, "Get tag from module: success\r\n");
    }
    rread(sdev, buf, 1, 0);
}


void IICSetSpeed(struct t_io_dev *iic_dev)
{
    unsigned char ch;

    rfprintf(sdev, "Enter new IIC speed (0..9): ");
    rread(sdev, &ch, 1, 0);
    rwrite(sdev, &ch, 1, 0);

    if ((ch >= '0') && (ch <= '9'))
    {
        rioctl(iic_dev, IOCTL_IIC_SET_SPEED, ch - '0');
        rfprintf(sdev, "\r\nNew speed set\r\n");
    }
    else
    {
        rfprintf(sdev, "\r\nInvalid Speed\r\n");
    }
}


#pragma pack (1)
struct t_rb_dio_configure
{
  unsigned char data_direction;
  unsigned char data_invert;
  unsigned char aucDebounceDelay[8];
};
#pragma pack ()


static unsigned char write_rb_io_config(struct t_io_dev *iic_dev)
{
    unsigned char buf[1 + sizeof(struct t_rb_dio_configure)];
    struct t_rb_dio_configure *dio_cfg = (struct t_rb_dio_configure *) &buf[1];
    unsigned char i;

    buf[0] = 0x4B;
    dio_cfg->data_direction = 0xFF;
    dio_cfg->data_invert = 0x00;
    for (i = 0; i < 8; i++)
        dio_cfg->aucDebounceDelay[i] = 0;

    if (!WriteReadIIC(iic_dev, RUN_ADDR, buf, 1 + sizeof(struct t_rb_dio_configure), 1))
        return 0;
    if ((buf[0] & 0x80) != 0)
        return 0;
    return 1;
}


static unsigned char write_rb_io_hw(struct t_io_dev *iic_dev, unsigned char data)
{
    unsigned char buf[2];

    buf[0] = 0x4C;
    buf[1] = data;

    if (!WriteReadIIC(iic_dev, RUN_ADDR, buf, 2, 2))
        return 0;
    if ((buf[0] & 0x80) != 0)
        return 0;
    return 1;
}


static void dio_direct_drive_output_test(struct t_io_dev *iic_dev)
{
    unsigned char i;

    rfprintf(sdev, "DIO Direct Drive Output Test\r\n");
    if (!write_rb_io_config(iic_dev))
    {
        rfprintf(sdev, "DIO Configure Failed\r\n");
        return;
    }

    rfprintf(sdev, "Cycling outputs, any key to quit\r\n");
    i = 0;
//    while (rfilelength(debug_dev) == 0)
    while (1)
    {
        if (!write_rb_io_hw(iic_dev, i))
            return;
        i++;
    }
    rread(sdev, &i, 1, 0);
}


#if 0
void DumpIICRegisters(volatile struct st_iic *iic_hw)
{
    unsigned char t;
    OS_DECLARE_CPU_SR();        /* Must be last variable */


    rfprintf(sdev, "Dump of IIC Registers for port %c\r\n", ((iic_hw == &P_IIC0) ? '0' : '1'));

    /*
     * ICMR
     */
    t = iic_hw->ICMR.BYTE;
    rfprintf(sdev, "ICMR: 0x%02x MLS:%1d WAIT:%1d CKS:%02d BC:%02d\r\n",
             t, ((t & 0x80) != 0), ((t & 0x40) != 0), ((t >> 3) & 0x07), (t & 0x07));

    /*
     * ICCR
     */
    t = iic_hw->ICCR.BYTE;
    rfprintf(sdev, "ICCR: 0x%02x ICE:%1d IEIC:%1d MST:%1d TRS:%1d ACKE:%1d BBSY:%1d IRIC:%1d SCP:%1d\r\n",
             t, ((t & 0x80) != 0), ((t & 0x40) != 0), ((t & 0x20) != 0), ((t & 0x10) != 0),
             ((t & 0x08) != 0), ((t & 0x04) != 0), ((t & 0x02) != 0), ((t & 0x01) != 0));

    /*
     * ICSR
     */
    t = iic_hw->ICSR.BYTE;
    rfprintf(sdev, "ICSR: 0x%02x ESTP:%1d STOP:%1d IRTR:%1d AASX:%1d AL:%1d AAS:%1d ADZ:%1d ACKB:%1d\r\n",
             t, ((t & 0x80) != 0), ((t & 0x40) != 0), ((t & 0x20) != 0), ((t & 0x10) != 0),
             ((t & 0x08) != 0), ((t & 0x04) != 0), ((t & 0x02) != 0), ((t & 0x01) != 0));

    /*
     * ICDR
     */
    rfprintf(sdev, "ICDR: 0x%02x\r\n", iic_hw->ICDR);

    /*
     * SAR0
     */
    OS_ENTER_CRITICAL();
    iic_hw->ICCR.BIT.ICE = 0;
    t = iic_hw->EQU.ICE0.UN_SAR.BYTE;
    iic_hw->ICCR.BIT.ICE = 1;
    OS_EXIT_CRITICAL();
    rfprintf(sdev, "SAR0: 0x%02x SVA:%02d FS:%1d\r\n",
             t, (t & 0xFE) >> 1, ((t & 0x01) != 0));

    /*
     * SARX
     */
    OS_ENTER_CRITICAL();
    iic_hw->ICCR.BIT.ICE = 0;
    t = iic_hw->EQU.ICE0.UN_SARX.BYTE;
    iic_hw->ICCR.BIT.ICE = 1;
    OS_EXIT_CRITICAL();
    rfprintf(sdev, "SARX: 0x%02x SVAX:%02d FSX:%1d\r\n",
             t, (t & 0xFE) >> 1, ((t & 0x01) != 0));

    /*
     * SCRX
     */
    t = P_SCRX.BYTE;
    rfprintf(sdev, "SCRX: 0x%02x IICX1:%1d IICX0:%1d IICE:%1d FLSHE:%1d\r\n",
             t, ((t & 0x40) != 0), ((t & 0x20) != 0), ((t & 0x10) != 0), ((t & 0x08) != 0));

    /*
     * MSTPCRA
     */
    t = P_MSTP.CRA.BYTE;
    rfprintf(sdev, "MSTPCRA: 0x%02x DMA:%1d DTC:%1d TPU:%1d TMR01:%1d ADC:%1d TMR23:%1d\r\n",
             t, ((t & 0x80) != 0), ((t & 0x40) != 0), ((t & 0x20) != 0), ((t & 0x10) != 0),
             ((t & 0x02) != 0), ((t & 0x01) != 0));

    /*
     * MSTPCRB
     */
    t = P_MSTP.CRB.BYTE;
    rfprintf(sdev, "MSTPCRB: 0x%02x SCI0:%1d SCI1:%1d SCI2:%1d IIC0:%1d IIC1:%1d\r\n",
             t, ((t & 0x80) != 0), ((t & 0x40) != 0), ((t & 0x20) != 0), ((t & 0x10) != 0),
             ((t & 0x08) != 0));

    /*
     * MSTPCRC
     */
    t = P_MSTP.CRC.BYTE;
    rfprintf(sdev, "MSTPCRB: 0x%02x SCI3:%1d DAC:%1d PBC:%1d\r\n",
             t, ((t & 0x80) != 0), ((t & 0x20) != 0), ((t & 0x10) != 0));

#if 0
    DDCSWR
#endif
}
#endif

#if 0
void DumpIICStates(void)
{
    unsigned short i;
    extern unsigned char iic_states[];
    extern unsigned short state_index;

    rfprintf(sdev, "IIC States from last transaction (%d)\r\n", state_index);
    for (i = 0; i < state_index; i++)
    {
        rfprintf(sdev, "0x%02x,", iic_states[i]);
        if ((i & 0x07) == 0x07)
            rfprintf(sdev, "\r\n");
    }
}
#endif


void TestIICModule(struct t_io_dev *dev, struct t_io_dev *iic_dev)
{
    unsigned char ch;

    sdev = dev;

    rfprintf(sdev, "IIC Module Test\r\nYou will need a comms module connected to the bus. Press any key to end.\r\n");

    rfprintf(sdev, "Open IIC device: ");
    if (ropen(iic_dev, 0) < 0)
    {
        rfprintf(sdev, "failed\r\n");
        return;
    }

    rfprintf(sdev, "\r\nStartup module\r\n");
    if (!ModuleStartup(iic_dev))
    {
        return;
    }

    ch = 0;
    while (ch != 'Q')
    {
        rfprintf(sdev, "IIC Module Test\r\nQ - Quit\r\nS - Set IIC Speed\r\nA - Get Address\r\nG - Repeated Get Address\r\nT - Get Tag\r\nR - Repeated Get Tag\r\nD - DIO Output Test\r\nH - Dump IIC Hardware Registers\r\n");
        //I - Dump IIC States\r\n");

        rread(sdev, &ch, 1, 0);
        switch (ch)
        {
            case 'S':
                IICSetSpeed(iic_dev);
                break;
            case 'A':
                GetAddress(iic_dev);
                break;
            case 'G':
                RepeatedGetAddress(iic_dev);
                break;
            case 'T':
                GetTag(iic_dev);
                break;
            case 'R':
                RepeatedGetTag(iic_dev);
                break;
            case 'D':
                dio_direct_drive_output_test(iic_dev);
                break;
#if 0
            case 'H':
                DumpIICRegisters(&P_IIC0);
                break;
#endif
//            case 'I':
//                DumpIICStates();
//                break;

        }
    }
    rclose(iic_dev);
}


void TestIIC(struct t_io_dev *dev, struct t_io_dev *iic_dev)
{
//    char *buf;
//    unsigned short len;
    int res;
    char ch;
//    char rxbuf[20];

    rfprintf(dev, "IIC Test on channel 0.\r\nYou will need a comms module connected to the bus\r\n. Press any key to end.");

    if (ropen(iic_dev, 0) < 0)
    {
        rfprintf(dev, "Could not open IIC channel 0\r\n");
        return;
    }

    res = rioctl(iic_dev, IOCTL_IIC_SET_SLAVE_ADDR, DEF_ADDR);
    rfprintf(dev, "Result of IIC0 ioctl set address: %d\r\n", res);

    res = rioctl(iic_dev, IOCTL_IIC_GET_SLAVE_ADDR, &ch);
    rfprintf(dev, "Result of IIC0 ioctl get address: %d, address %d\r\n", res, ch);

#if 0
//    buf = "This is a test string for the IIC";
    buf = "IIC Test String";
    len = (unsigned short) strlen(buf);
    while (reof(dev))
    {
        res = rwrite(iic_dev, buf, len, 0);
        rfprintf(dev, "Result of write to IIC0: %d\r\n", res);

        res = rread(iic_dev, rxbuf, 20, 0);
        if (res >= 0)
            rxbuf[res] = '\0';
        else
            rxbuf[0] = '\0';

        rfprintf(dev, "Result of read from IIC0: %d, data was \"%s\"\r\n", res, rxbuf);
        OSTimeDly(OS_TICKS_PER_SEC / 2);
    }
    rread(dev, &ch, 1, 0);
#endif
    ch = 0;
    while (ch != 0x1B)
    {
        rread(dev, &ch, 1, 0);
        res = rwrite(iic_dev, &ch, 1, 0);
        rfprintf(dev, "Result of write to IIC0: %d\r\n", res);

        OSTimeDly(OS_TICKS_PER_SEC / 10);

        res = rread(iic_dev, &ch, 1, 0);
        rfprintf(dev, "Result of read from IIC0: %d, data was %c\r\n", res, ch);
        OSTimeDly(OS_TICKS_PER_SEC / 2);
    }

    rclose(iic_dev);
}
