#include <rinhw/rio.h>
#include <rinhw/rstdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/stdrinstrum.h>

#include "ostest.h"


static OS_EVENT *test_sem;


static void TestSemaphore( struct t_io_dev *dev )
{
  int8u err;

  rfprintf(dev, "Running semaphore tests\r\n");

  rfprintf(dev, "Create Semaphore\r\n");
  test_sem = OSSemCreate( 0 );
  if ( 0 == test_sem )
  {
    rfprintf(dev, "Create Semaphore failed\r\n");
    return;
  }

  rfprintf(dev, "Post to semaphore\r\n");
  if ( OS_NO_ERR != OSSemPost( test_sem ) )
  {
    rfprintf(dev, "Post to semaphore failed\r\n");
    return;
  }

  rfprintf(dev, "Pend on semaphore\r\n");
  OSSemPend( test_sem, 0, &err );
  if ( OS_NO_ERR != err )
  {
    rfprintf(dev, "Pend on semaphore failed\r\n");
  }

  rfprintf(dev, "Pend on semaphore with timeout\r\n");
  OSSemPend( test_sem, OS_TICKS_PER_SEC, &err );
  if ( OS_TIMEOUT != err )
  {
    rfprintf(dev, "Pend on semaphore failed\r\n");
  }

  rfprintf(dev, "Semaphore tests complete\r\n");
}


static void TestTiming( struct t_io_dev *dev )
{
  int8u err, ch;
  int32u start_time, end_time, delta, seconds;
  int i;

  rfprintf(dev, "Running timing tests\r\n");

  /* 5 second delay */
  rfprintf(dev, "Any key to start 5 second delay\r\n");
  rread(dev, &ch, 1, 0);
  rfprintf(dev, "Start 5 second delay\r\n");
  start_time = OSTimeGet();
  OSTimeDly(OS_TICKS_PER_SEC * 5);
  end_time = OSTimeGet();
  delta = end_time - start_time;
  seconds = delta / OS_TICKS_PER_SEC;
  rfprintf(dev, "Finished 5 second delay, start @ %lu, end @ %lu ticks, delta = %lu ticks (%lu seconds)\r\n\r\n",
           start_time, end_time, delta, seconds);

  /* Lots of short delays */
  rfprintf(dev, "Any key to start 50 iterations of 100ms delay\r\n");
  rread(dev, &ch, 1, 0);
  rfprintf(dev, "Start 50 iterations of 100ms delay\r\n");
  start_time = OSTimeGet();
  for (i = 0; i < 5 * 10; i++)
  {
    OSTimeDly(OS_TICKS_PER_SEC / 10);
  }
  end_time = OSTimeGet();
  delta = end_time - start_time;
  seconds = delta / OS_TICKS_PER_SEC;
  rfprintf(dev, "Finished 50 iterations of 100ms delay, start @ %lu, end @ %lu ticks, delta = %lu ticks (%lu seconds)\r\n",
           start_time, end_time, delta, seconds);

  /* Test timeout on semaphore */
  test_sem = OSSemCreate( 0 );
  rfprintf(dev, "Any key to start 5 second semaphore timeout\r\n");
  rread(dev, &ch, 1, 0);
  rfprintf(dev, "Start 5 second semaphore timeout\r\n");
  start_time = OSTimeGet();
  OSSemPend( test_sem, OS_TICKS_PER_SEC * 5, &err );
  end_time = OSTimeGet();
  if ( OS_TIMEOUT != err )
  {
    rfprintf(dev, "Semaphore timeout\r\n");
  }
  delta = end_time - start_time;
  seconds = delta / OS_TICKS_PER_SEC;
  rfprintf(dev, "Finished 5 semaphore timeout, start @ %lu, end @ %lu ticks, delta = %lu ticks (%lu seconds)\r\n\r\n",
           start_time, end_time, delta, seconds);

  /* Lots of short semaphore timeouts */
  rfprintf(dev, "Any key to start 50 iterations of 100ms semaphore timeouts\r\n");
  rread(dev, &ch, 1, 0);
  rfprintf(dev, "Start 50 iterations of 100ms semaphore timeouts\r\n");
  start_time = OSTimeGet();
  for (i = 0; i < 5 * 10; i++)
  {
    OSSemPend( test_sem, OS_TICKS_PER_SEC / 10, &err );
    if ( OS_TIMEOUT != err )
    {
      rfprintf(dev, "Semaphore error\r\n");
      break;
    }
  }
  end_time = OSTimeGet();
  delta = end_time - start_time;
  seconds = delta / OS_TICKS_PER_SEC;
  rfprintf(dev, "Finished 50 iterations of 100ms semaphore timeouts, start @ %lu, end @ %lu ticks, delta = %lu ticks (%lu seconds)\r\n",
           start_time, end_time, delta, seconds);

  rfprintf(dev, "Timing tests complete\r\n");
}



static void MultipleSemaphoreTimeouts( struct t_io_dev *dev )
{
  int8u err, ch;
  int32u start_time, end_time, delta, seconds;

  test_sem = OSSemCreate( 0 );

  rfprintf(dev, "Multiple semaphore timeouts. Any key to quit.\r\n");

  /* Lots of short semaphore timeouts */
  while (reof(dev))
  {
    start_time = OSTimeGet();
    OSSemPend( test_sem, OS_TICKS_PER_SEC / 10, &err );
    if ( OS_TIMEOUT != err )
    {
      rfprintf(dev, "Semaphore error\r\n");
      return;
    }
    end_time = OSTimeGet();
    delta = end_time - start_time;
    seconds = delta / OS_TICKS_PER_SEC;
    rfprintf(dev, "100ms timeout, start @ %8lu, end @ %8lu ticks, delta = %8lu ticks (%8lu seconds)\r\n",
             start_time, end_time, delta, seconds);
  }
  rread(dev, &ch, 1, 0);
}


void TestOS(struct t_io_dev *dev, struct t_io_dev *dummy)
{
  char ch = '\0';

  while (ch != 'Q')
  {
    rfprintf(dev, "Operating System Test\r\n"
                  "  s - Test Semaphore\r\n"
                  "  t - Test Timing\r\n"
                  "  m - Multiple Semaphore Timeouts\r\n"
                  "  Q - Quit\r\n\r\n" );

    rread(dev, &ch, 1, 0);
    switch (ch)
    {
      case 's':
        rfprintf( dev, "Test Semaphore\r\n" );
        TestSemaphore( dev );
        break;
      case 't':
        rfprintf( dev, "Test Timing\r\n" );
        TestTiming( dev );
        break;
      case 'm':
        rfprintf( dev, "Multiple Semaphore Timeouts\r\n" );
        MultipleSemaphoreTimeouts(dev);
        break;
    }
  }
}

