#include <stdint.h>
#include <rinhw/rio.h>
#include <rinhw/rstdio.h>
#include <rinhw/sio.h>
#include "splitsiotest.h"
#include "dumpdev.h"
#include "devices.h"

#include "siotest.h"


void SerialReadPort(struct t_io_dev *console, struct t_io_dev *serial)
{
  unsigned long num_chars;
  unsigned long timeout;
  int res;

  #define BUF_LEN 100
  char buffer[BUF_LEN];

  rfprintf(console, "Enter number of characters to receive followed by LF (max %d): ", BUF_LEN);
  num_chars = GetLong(console);
  if (num_chars > BUF_LEN)
  {
    rfprintf(console, "Too many (max %d)\n", BUF_LEN);
    return;
  }

  rfprintf(console, "Enter timeout in OS_TICK_PER_SEC (%d) increments, or 0 for indefinite, followed by LF: ", OS_TICKS_PER_SEC);
  timeout = GetLong(console);
  if (num_chars > UINT16_MAX)
  {
    rfprintf(console, "Too big (max %u)\n", UINT16_MAX);
    return;
  }

  res = rread(serial, buffer, num_chars, timeout);
  if (res != (int) num_chars)
  {
    rfprintf(console, "Error reading from serial port (result %d)\n", res);
    return;
  }

  rfprintf(console, "Data received from serial port:\n");
  DumpMemory(console, buffer, num_chars, false);
}


void SerialWritePort(struct t_io_dev *console, struct t_io_dev *serial)
{
  unsigned long num_chars;
  unsigned long timeout;
  int res;

  #define BUF_LEN 100
  char buffer[BUF_LEN];

  rfprintf(console, "Enter number of characters to send followed by LF (max %d): ", BUF_LEN);
  num_chars = GetLong(console);
  if (num_chars > BUF_LEN)
  {
    rfprintf(console, "Too many (max %d)\n", BUF_LEN);
    return;
  }

  rfprintf(console, "Enter timeout in OS_TICK_PER_SEC (%d) increments, or 0 for indefinite, followed by LF: ", OS_TICKS_PER_SEC);
  timeout = GetLong(console);
  if (num_chars > UINT16_MAX)
  {
    rfprintf(console, "Too big (max %u)\n", UINT16_MAX);
    return;
  }

  rfprintf(console, "Enter %lu characters to send\n", num_chars);
  res = rread(console, buffer, num_chars, 0);
  if (res != (int) num_chars)
  {
    rfprintf(console, "Error reading from console (result %d)\n", res);
    return;
  }

  res = rwrite(serial, buffer, num_chars, timeout);
  if (res != (int) num_chars)
  {
    rfprintf(console, "Error writing to serial port (result %d)\n", res);
    return;
  }

  rfprintf(console, "Data written to serial port successfully\n");
}


void SerialDumpBufferLevels(struct t_io_dev *console, struct t_io_dev *serial)
{
  unsigned short rx_lvl;
  unsigned short tx_lvl;
  unsigned short rx_size;
  unsigned short tx_size;

  if (rioctl(serial, IOCTL_SER_GET_RX_BUF_LEVEL, &rx_lvl) < 0)
  {
    rfprintf(console, "IOCTL_SER_GET_RX_BUF_LEVEL failed\n");
    return;
  }
  if (rioctl(serial, IOCTL_SER_GET_TX_BUF_LEVEL, &tx_lvl) < 0)
  {
    rfprintf(console, "IOCTL_SER_GET_RX_BUF_LEVEL failed\n");
    return;
  }
  if (rioctl(serial, IOCTL_SER_GET_RX_BUF_SIZE, &rx_size) < 0)
  {
    rfprintf(console, "IOCTL_SER_GET_RX_BUF_LEVEL failed\n");
    return;
  }
  if (rioctl(serial, IOCTL_SER_GET_TX_BUF_SIZE, &tx_size) < 0)
  {
    rfprintf(console, "IOCTL_SER_GET_RX_BUF_LEVEL failed\n");
    return;
  }

  rfprintf(console,
           "IOCTL_SER_GET_RX_BUF_LEVEL %u, IOCTL_SER_GET_TX_BUF_LEVEL %u\n"
           "IOCTL_SER_GET_RX_BUF_SIZE %u, IOCTL_SER_GET_TX_BUF_SIZE %u\n",
           rx_lvl, tx_lvl,
           rx_size, tx_size
          );
}


static void SerialLineStateTest(struct t_io_dev *console, struct t_io_dev *serial)
{
  unsigned char cts_lvl = 0;
  int rts_lvl = 0;
  int dtr_lvl = 0;
  unsigned char ch = '\0';
  const char asserted_str[] = "asserted (RS232 +ve, TTL 0v)";
  const char deasserted_str[] = "deasserted (RS232 -ve, TTL 5v)";

  while ('Q' != ch)
  {
    rfprintf(console, "Line state testing\n"
                      "  Current State: ");
    if (rioctl(serial, IOCTL_SER_GET_CTS_STATE, &cts_lvl) < 0)
      rfprintf(console, "IOCTL_SER_GET_CTS_STATE failed ");
    else
      rfprintf(console, "IOCTL_SER_GET_CTS_STATE %u ", (unsigned int) cts_lvl);
#ifdef IOCTL_SER_GET_RTS_STATE
    if (rioctl(serial, IOCTL_SER_GET_RTS_STATE, &rts_lvl) < 0)
      rfprintf(console, "IOCTL_SER_GET_RTS_STATE failed ");
    else
      rfprintf(console, "IOCTL_SER_GET_RTS_STATE %s ", (rts_lvl) ? asserted_str : deasserted_str);
#else
    rfprintf(console, "IOCTL_SER_GET_RTS_STATE not supported ");
#endif
#ifdef IOCTL_SER_GET_DTR_STATE
    if (rioctl(serial, IOCTL_SER_GET_DTR_STATE, &dtr_lvl) < 0)
      rfprintf(console, "IOCTL_SER_GET_DTR_STATE failed ");
    else
      rfprintf(console, "IOCTL_SER_GET_DTR_STATE %s ", (dtr_lvl) ? asserted_str : deasserted_str);
#else
    rfprintf(console, "IOCTL_SER_GET_DTR_STATE not supported ");
#endif

    rfprintf(console, "\n"
#ifdef IOCTL_SER_SET_RTS_STATE
                      "  r - Deassert RTS (RS232 -ve, TTL 5v)\n"
                      "  R - Assert RTS (RS232 +ve, TTL 0v)\n"
#endif
#ifdef IOCTL_SER_SET_DTR_STATE
                      "  d - Deassert DTR (RS232 -ve, TTL 5v)\n"
                      "  D - Assert DTR (RS232 +ve, TTL 0v)\n"
#endif
                      "  Q - Quit\n");

    rread(console, &ch, 1, 0);
    switch (ch)
    {
#ifdef IOCTL_SER_SET_RTS_STATE
      case 'r':
        rfprintf(console, "Deassert RTS (RS232 -ve, TTL 5v)\n");
        if (rioctl(serial, IOCTL_SER_SET_RTS_STATE, 0) < 0)
        {
          rfprintf(console, "IOCTL_SER_SET_RTS_STATE failed\n");
        }
        break;
      case 'R':
        rfprintf(console, "Assert RTS (RS232 +ve, TTL 0v)\n");
        if (rioctl(serial, IOCTL_SER_SET_RTS_STATE, 1) < 0)
        {
          rfprintf(console, "IOCTL_SER_SET_RTS_STATE failed\n");
        }
        break;
#endif /* #ifdef IOCTL_SER_SET_RTS_STATE */
#ifdef IOCTL_SER_SET_DTR_STATE
      case 'd':
        rfprintf(console, "Deassert DTR (RS232 -ve, TTL 5v)\n");
        if (rioctl(serial, IOCTL_SER_SET_DTR_STATE, 0) < 0)
        {
          rfprintf(console, "IOCTL_SER_SET_DTR_STATE failed\n");
        }
        break;
      case 'D':
        rfprintf(console, "Assert DTR (RS232 +ve, TTL 0v)\n");
        if (rioctl(serial, IOCTL_SER_SET_DTR_STATE, 1) < 0)
        {
          rfprintf(console, "IOCTL_SER_SET_DTR_STATE failed\n");
        }
        break;
#endif /* #ifdef IOCTL_SER_SET_RTS_STATE */
    }
  }
}


void TestSio(struct t_io_dev *console, struct t_io_dev *serial)
{
  unsigned char ch = '\0';

  while (ch != 'Q')
  {
    rfprintf(console, "Serial Port Test\r\n"
                      "  O - Open port\r\n"
                      "  C - Close port\r\n"
                      "  S - Setup port\r\n"
                      "  D - Dump port config\r\n"
                      "  W - Write data to port\r\n"
                      "  R - Read data to port\r\n"
                      "  B - Print buffer levels for port\r\n"
                      "  T - Print line states for port (CTS, RTS, DTR)\r\n"
                      "  Q - Quit\r\n"
            );
    rread(console, &ch, 1, 0);

    switch (ch)
    {
      case 'O':
        if (serial->status & O_OPEN)
          rfprintf(console, "Result of ropen() %d\n", ropen(serial, 0));
        else
          rfprintf(console, "Device already open\n");
        break;
      case 'C':
        if (serial->status & O_OPEN)
          rfprintf(console, "Device already closed\n");
        else
          rfprintf(console, "Result of rclose() %d\n", rclose(serial));
        break;
      case 'S':
        ConfigureSerial(console, serial);
        break;
      case 'D':
        SerialDumpConfiguration(console, serial);
        break;
      case 'W':
        SerialWritePort(console, serial);
        break;
      case 'R':
        SerialReadPort(console, serial);
        break;
      case 'B':
        SerialDumpBufferLevels(console, serial);
        break;
      case 'T':
        SerialLineStateTest(console, serial);
        break;
    }
  }
}


void TestUsbSio(struct t_io_dev *console, struct t_io_dev *serial)
{
  unsigned char ch = '\0';
  int i;

  rfprintf(console, "Select USB serial port to test:\n");
  for (i = 0; i < MAX_DEV_USB_SERIAL; i++)
  {
    if (0 != (dev_usb_serial[i].status & O_INITIALISED))
    {
      rfprintf(console, "  %u - /dev/ttyUSB%u\n", i, i);
    }
  }

  rread(console, &ch, 1, 0);
  ch -= '0';
  if (ch < MAX_DEV_USB_SERIAL)
  {
    if (0 != (dev_usb_serial[ch].status & O_INITIALISED))
    {
      TestSio(console, &dev_usb_serial[ch]);
    }
    else
    {
      rfprintf(console, "Serial port not present\n");
    }
  }
  else
  {
    rfprintf(console, "Invalid selection\n");
  }
}
