/*
 * byteq_tests.c
 *
 *  Created on: Feb 23, 2017
 *      Author: ian
 */
#include <stdint.h>
#include <stdlib.h>
#include <cutest/CuTest.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/byteq.h>

#include "byteq_tests.h"


#define BQ_BUF_LEN 10


/** Test creating byte queues */
static void test_byteq_create(CuTest *tc)
{
  enum eBQResult res;
  struct s_byteq q1;
  struct s_byteq q2;
  uint8_t buffer1[BQ_BUF_LEN];
  uint8_t buffer2[BQ_BUF_LEN];

  /* Create a byte queue pending on read */
  res = ByteQCreate(&q1, bqPendOnRead, buffer1, BQ_BUF_LEN);
  CuAssertIntEquals(tc, bqrSuccess, res);

  /* Create a byte queue pending on write */
  res = ByteQCreate(&q2, bqPendOnWrite, buffer2, BQ_BUF_LEN);
  CuAssertIntEquals(tc, bqrSuccess, res);

  /* Clean up */
  res = ByteQDestroy(&q1);
  CuAssertIntEquals(tc, bqrSuccess, res);
  res = ByteQDestroy(&q2);
  CuAssertIntEquals(tc, bqrSuccess, res);
}


static void fill_queue(CuTest *tc, struct s_byteq *q)
{
  size_t i;
  enum eBQResult res;
  size_t space;
  size_t level;

  /* Fill the queue */
  for (i = 0; i < BQ_BUF_LEN; i++)
  {
    res = ByteQWrite(q, (uint8_t) 'a' + i, NULL);
    CuAssertIntEquals(tc, bqrSuccess, res);

    /* Should be i + 1 bytes in the queue */
    level = ByteQGetLevel(q);
    CuAssertIntEquals(tc, i + 1, level);

    /* Should be BQ_BUF_LEN - i - 1 bytes of space in the queue */
    space = ByteQGetSpace(q);
    CuAssertIntEquals(tc, BQ_BUF_LEN - i - 1, space);
  }
}


/** Test a queue that pends on write */
static void test_byteq_pend_write(CuTest *tc)
{
  enum eBQResult res;
  struct s_byteq q;
  uint8_t buffer[BQ_BUF_LEN];
  struct timespec to;

  /* Create a byte queue pending on write */
  res = ByteQCreate(&q, bqPendOnWrite, buffer, BQ_BUF_LEN);
  CuAssertIntEquals(tc, bqrSuccess, res);

  /* A read from an empty pend on write queue should just fail with bqrEmpty */
  to.tv_nsec = 1000;
  to.tv_sec = 0;
  (void) ByteQRead(&q, &to, &res);
  CuAssertIntEquals(tc, bqrEmpty, res);

  /* Fill the queue */
  fill_queue(tc, &q);

  /* Queue full, another write should block and time out */
  to.tv_nsec = 1000;
  to.tv_sec = 0;
  res = ByteQWrite(&q, '0', &to);
  CuAssertIntEquals(tc, bqrTimeout, res);

  /* Clean up */
  res = ByteQDestroy(&q);
  CuAssertIntEquals(tc, bqrSuccess, res);
}


/** Test a queue that pends on read */
static void test_byteq_pend_read(CuTest *tc)
{
  enum eBQResult res;
  struct s_byteq q;
  uint8_t buffer[BQ_BUF_LEN];
  struct timespec to;

  /* Create a byte queue pending on read */
  res = ByteQCreate(&q, bqPendOnRead, buffer, BQ_BUF_LEN);
  CuAssertIntEquals(tc, bqrSuccess, res);

  /* A read from an empty pend on read queue should block and timeout */
  to.tv_nsec = 1000;
  to.tv_sec = 0;
  (void) ByteQRead(&q, &to, &res);
  CuAssertIntEquals(tc, bqrTimeout, res);

  /* Fill the queue */
  fill_queue(tc, &q);

  /* Queue full, another write should should just fail with bqrFull */
  to.tv_nsec = 1000;
  to.tv_sec = 0;
  res = ByteQWrite(&q, '0', &to);
  CuAssertIntEquals(tc, bqrFull, res);

  /* Clean up */
  res = ByteQDestroy(&q);
  CuAssertIntEquals(tc, bqrSuccess, res);
}


/** Test a queue that doesn't pend */
static void test_byteq_no_pend(CuTest *tc)
{
  enum eBQResult res;
  struct s_byteq q;
  uint8_t buffer[BQ_BUF_LEN];
  struct timespec to;

  /* Create a byte queue that doesn't pend */
  res = ByteQCreate(&q, bqNoPend, buffer, BQ_BUF_LEN);
  CuAssertIntEquals(tc, bqrSuccess, res);

  /* A read from an empty no pend queue just fail with bqrEmpty */
  to.tv_nsec = 1000;
  to.tv_sec = 0;
  (void) ByteQRead(&q, &to, &res);
  CuAssertIntEquals(tc, bqrEmpty, res);

  /* Fill the queue */
  fill_queue(tc, &q);

  /* Queue full, another write should should just fail with bqrFull */
  to.tv_nsec = 1000;
  to.tv_sec = 0;
  res = ByteQWrite(&q, '0', &to);
  CuAssertIntEquals(tc, bqrFull, res);

  /* Clean up */
  res = ByteQDestroy(&q);
  CuAssertIntEquals(tc, bqrSuccess, res);
}


/** Test writing and reading to a byte queue */
static void test_byteq_write_read(CuTest *tc)
{
  enum eBQResult res;
  struct s_byteq q;
  uint8_t buffer[BQ_BUF_LEN];
  size_t i;
  uint8_t data;

  /* Create a byte queue pending on read */
  res = ByteQCreate(&q, bqPendOnRead, buffer, BQ_BUF_LEN);
  CuAssertIntEquals(tc, bqrSuccess, res);

  /* Fill the queue */
  fill_queue(tc, &q);

  /* Read everything out of the queue */
  for (i = 0; i < BQ_BUF_LEN; i++)
  {
    /* A read from an empty no pend queue just fail with bqrEmpty */
    data = ByteQRead(&q, NULL, &res);
    CuAssertIntEquals(tc, bqrSuccess, res);
    CuAssertIntEquals(tc, 'a' + i, data);
  }

  /* Clean up */
  res = ByteQDestroy(&q);
  CuAssertIntEquals(tc, bqrSuccess, res);
}


CuSuite *CuGetSuite_byteq(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_byteq_create);
  SUITE_ADD_TEST(suite, test_byteq_pend_write);
  SUITE_ADD_TEST(suite, test_byteq_pend_read);
  SUITE_ADD_TEST(suite, test_byteq_no_pend);
  SUITE_ADD_TEST(suite, test_byteq_write_read);
  return suite;
}
