/*
 * os_time_tests.c
 *
 *  Created on: Jun 9, 2015
 *      Author: ian
 *      Author: pauli
 */
#include <stdlib.h>
#include <time.h>
#include <cutest/CuTest.h>
#include <rinhw/timemath.h>
#include <stdint.h>
#include <stdio.h>
#include <linux/param.h>
#include "os_time_tests.h"


#define MILLI_SEC_PER_SEC 1000
#define MICRO_SEC_PER_SEC 1000000
#define MICRO_SEC_PER_MILLI_SEC 1000
#define NANO_SEC_PER_SEC 1000000000
#define NANO_SEC_PER_MILLI_SEC 1000000


static const int stochastic_count = 1000;


static int64_t rantime(void) {
    int64_t r;

    r = random() & 0xfffff;
    r = (r << 20) | (random() & 0xfffff);
    r = (r << 20) | (random() & 0xfffff);
    return r;
}

static struct timespec rants(int large) {
    struct timespec r;
    r.tv_sec = random() & 0x3fff;
    r.tv_sec = (r.tv_sec << 16) | (random() & 0xffff);
    r.tv_nsec = random() & (large ? 0xffff : 0x7fff);
    r.tv_nsec = (r.tv_nsec << 16) | (random() & 0xffff);
    return r;
}

static struct timespec int2ts(int64_t x) {
    struct timespec r;
    r.tv_sec  = x / NANO_SEC_PER_SEC;
    r.tv_nsec = x % NANO_SEC_PER_SEC;
    return r;
}

/** Test the AddTimespec function */
static void test_AddTimespec_simple(CuTest* tc)
{
  struct timespec pts1;
  struct timespec pts2;

  /* Check add nanos */
  pts1.tv_nsec = 100;
  pts1.tv_sec = 0;
  pts2.tv_nsec = 100;
  pts2.tv_sec = 0;
  AddTimespec(&pts1, &pts2);
  CuAssertIntEquals(tc, 200, pts1.tv_nsec);
  CuAssertIntEquals(tc, 0, pts1.tv_sec);

  /* Check add seconds */
  pts1.tv_nsec = 0;
  pts1.tv_sec = 100;
  pts2.tv_nsec = 0;
  pts2.tv_sec = 100;
  AddTimespec(&pts1, &pts2);
  CuAssertIntEquals(tc, 0, pts1.tv_nsec);
  CuAssertIntEquals(tc, 200, pts1.tv_sec);
}


/** Test the AddTimespec function when nano seconds overflow into seconds */
static void test_AddTimespec_nsec_overflow(CuTest* tc)
{
  struct timespec pts1;
  struct timespec pts2;

  /* Check add nanos */
  pts1.tv_nsec = NANO_SEC_PER_SEC - 100;
  pts1.tv_sec = 1;
  pts2.tv_nsec = 200;
  pts2.tv_sec = 3;
  AddTimespec(&pts1, &pts2);
  CuAssertIntEquals(tc, 100, pts1.tv_nsec);
  CuAssertIntEquals(tc, 5, pts1.tv_sec);
}

static void test_AddTimespec(CuTest *tc) {
    static const struct {
        struct timespec pts1;
        struct timespec pts2;
        struct timespec res;
    } tests[] = {
        { { 123, 0 },                   { 100, 0 },             { 223, 0 } },
        { { 0, 101 },                   { 0, 122 },             { 0, 223 } },
        { { 1433989681, 12345678 },     { 3, 87654321 },        { 1433989684, 99999999 } },
        { { 3, 999999999 },             { 123, 123456789 },     { 127, 123456788 } },
        { { 3, 1999999999 },            { 123, 1000000002 },    { 129, 1 } },
    };
    size_t i;
    char m[20];

    for (i=0; i<sizeof(tests)/sizeof(*tests); i++) {
        struct timespec ts = tests[i].pts1;
        sprintf(m, "case %u", (unsigned) i);
        AddTimespec(&ts, &tests[i].pts2);
        CuAssertIntEquals_Msg(tc, m, tests[i].res.tv_sec, ts.tv_sec);
        CuAssertIntEquals_Msg(tc, m, tests[i].res.tv_nsec, ts.tv_nsec);
    }
}

static void test_AddTimespec_Stochastic(CuTest *tc) {
    struct timespec pts1, pts2, r;
    int64_t v1, v2;
    int i;
    char m[80];

    for (i=0; i<stochastic_count; i++) {
        pts1 = int2ts(v1 = rantime());
        pts2 = int2ts(v2 = rantime());
        r = int2ts(v1 + v2);
        AddTimespec(&pts1, &pts2);
        sprintf(m, "case %llu %llu", (unsigned long long) v1, (unsigned long long) v2);
        CuAssertIntEquals_Msg(tc, m, r.tv_sec, pts1.tv_sec);
        CuAssertIntEquals_Msg(tc, m, r.tv_nsec, pts1.tv_nsec);
    }
}

static void test_AddTimespec_Stochastic_Large(CuTest *tc) {
    struct timespec pts1, pts2, r;
    int64_t v1, v2;
    int i;
    char m[80];

    for (i=0; i<stochastic_count; i++) {
        pts1 = rants(0);
        pts2 = rants(0);
        v1 = pts1.tv_sec * (int64_t)NANO_SEC_PER_SEC + pts1.tv_nsec;
        v2 = pts2.tv_sec * (int64_t)NANO_SEC_PER_SEC + pts2.tv_nsec;
        r = int2ts(v1 + v2);
        AddTimespec(&pts1, &pts2);
        sprintf(m, "case %ld.%09ld %ld.%09ld", (long) pts1.tv_sec, pts1.tv_nsec, (long) pts2.tv_sec, pts2.tv_nsec);
        CuAssertIntEquals_Msg(tc, m, r.tv_sec, pts1.tv_sec);
        CuAssertIntEquals_Msg(tc, m, r.tv_nsec, pts1.tv_nsec);
    }
}

static void test_SubTimespec(CuTest *tc) {
    static const struct {
        struct timespec pts1;
        struct timespec pts2;
        struct timespec res;
    } tests[] = {
        { { 123, 0 },                   { 100, 0 },             { 23, 0 } },
        { { 0, 201 },                   { 0, 122 },             { 0, 79 } },
        { { 123, 0 },                   { 150, 0 },             { 0, 0 } },
        { { 0, 101 },                   { 0, 122 },             { 0, 0 } },
        { { 1433989681, 12345678 },     { 3, 87654321 },        { 1433989677, 924691357 } },
        { { 3, 999999999 },             { 123, 123456789 },     { 0, 0 } },
        { { 3, 999999999 },             { 1, 123456789 },       { 2, 876543210 } },
        { { 123, 1999999999 },          { 3, 1 },               { 120, 1999999998 } },
    };
    size_t i;
    char m[20];

    for (i=0; i<sizeof(tests)/sizeof(*tests); i++) {
        struct timespec ts = tests[i].pts1;
        sprintf(m, "case %u", (unsigned) i);
        SubTimespec(&ts, &tests[i].pts2);
        CuAssertIntEquals_Msg(tc, m, tests[i].res.tv_sec, ts.tv_sec);
        CuAssertIntEquals_Msg(tc, m, tests[i].res.tv_nsec, ts.tv_nsec);
    }
}

static void test_SubTimespec_Stochastic(CuTest *tc) {
    struct timespec pts1, pts2, r;
    int64_t v1, v2;
    int i;
    char m[80];

    for (i=0; i<stochastic_count; i++) {
        pts1 = int2ts(v1 = rantime());
        pts2 = int2ts(v2 = rantime());
        if (v1 > v2)
            r = int2ts(v1 - v2);
        else {
            r.tv_sec = 0;
            r.tv_nsec = 0;
        }
        SubTimespec(&pts1, &pts2);
        sprintf(m, "case %llu %llu", (unsigned long long) v1, (unsigned long long) v2);
        CuAssertIntEquals_Msg(tc, m, r.tv_sec, pts1.tv_sec);
        CuAssertIntEquals_Msg(tc, m, r.tv_nsec, pts1.tv_nsec);
    }
}

static void test_OSTicksToTimespec(CuTest *tc) {
    uint32_t i;
    int64_t res = 0;
    char m[20];
    struct timespec ts, r;

    // Just test all possible values
    for (i=0; i<65536; i++) {
        const uint16_t n = i & 0xffff;
        r = int2ts(res);
        sprintf(m, "case %u", n);
        OSTicksToTimespec(&ts, n);
        CuAssertIntEquals_Msg(tc, m, r.tv_sec, ts.tv_sec);
        CuAssertIntEquals_Msg(tc, m, r.tv_nsec, ts.tv_nsec);
        res += NANO_SEC_PER_SEC / HZ;
    }
}

static void test_OSTicksToTimeval(CuTest *tc) {
    uint32_t i;
    int64_t res = 0;
    char m[20];
    struct timeval tv;
    struct timespec r;

    // Just test all possible values
    for (i=0; i<65536; i++) {
        const uint16_t n = i & 0xffff;
        r = int2ts(res);
        sprintf(m, "case %u", n);
        OSTicksToTimeval(&tv, n);
        CuAssertIntEquals_Msg(tc, m, r.tv_sec, tv.tv_sec);
        CuAssertIntEquals_Msg(tc, m, r.tv_nsec / 1000, tv.tv_usec);
        res += NANO_SEC_PER_SEC / HZ;
    }
}

static void test_OSTimespecToTicks(CuTest *tc) {
    uint32_t ticks, r;
    struct timespec ts;
    char m[20];
    int i;

    for (i=0; i<stochastic_count; i++) {
        r = random() & 0xffff;
        r = (r << 16) | (random() & 0xffff);
        sprintf(m, "case %u", r);
        ts.tv_sec = r / HZ;
        ts.tv_nsec = (r % HZ) * (NANO_SEC_PER_SEC / HZ);
        OSTimespecToTicks(&ticks, &ts);
        CuAssertIntEquals_Msg(tc, m, r, ticks);
    }
}

CuSuite *CuGetSuite_os_time(void)
{
  CuSuite *suite = CuSuiteNew();

  srandom(time(NULL));

  SUITE_ADD_TEST(suite, test_AddTimespec_simple);
  SUITE_ADD_TEST(suite, test_AddTimespec_nsec_overflow);
  SUITE_ADD_TEST(suite, test_AddTimespec);
  SUITE_ADD_TEST(suite, test_AddTimespec_Stochastic);
  SUITE_ADD_TEST(suite, test_AddTimespec_Stochastic_Large);
  SUITE_ADD_TEST(suite, test_SubTimespec);
  SUITE_ADD_TEST(suite, test_SubTimespec_Stochastic);
  SUITE_ADD_TEST(suite, test_OSTicksToTimespec);
  SUITE_ADD_TEST(suite, test_OSTicksToTimeval);
  SUITE_ADD_TEST(suite, test_OSTimespecToTicks);
  return suite;
}
