/*
 * r400_config.h
 *
 * Configuration file for the R400/R500
 *
 * Was config.h in the hardware library
 *
 *  Created on: May 20, 2013
 *      Author: ian
 */

#ifndef R400_CONFIG_H_
#define R400_CONFIG_H_


#define R400 1
#define R500 2


#ifndef TARGET
    #error "No TARGET is defined"
#endif

#if ((TARGET != R400) && (TARGET != R500))
    #error "TARGET is incorrectly defined"
#endif


/*
 * Defines to decide which serial ports to compile in for
 *
 * This must match the assembler version of this file: config.inc
 */
#if (TARGET == R500)

#define INCLUDE_SERIAL0 0
#define INCLUDE_SERIAL1 0
#define INCLUDE_SERIAL2 0
#define INCLUDE_SERIAL3 1
#define INCLUDE_SERIAL4 0

#define INCLUDE_IIC0 1
#define INCLUDE_IIC1 1

#define INCLUDE_ADCCPU 0
#define INCLUDE_ADC24BIT 1

#define INCLUDE_SSTFLASH 0

#define INCLUDE_RTC_DS1337 0

#define INCLUDE_FRAM 0

#define INCLUDE_FLASH2238 1

#define INCLUDE_LCD_SEG_FLASH 0

#define INCLUDE_TICK_FUNCTION 1

#define INCLUDE_BUZZER 1

#define INCLUDE_KBD 1

#define INCLUDE_LCD 1


#elif (TARGET == R400)

#define INCLUDE_SERIAL0 0
#define INCLUDE_SERIAL1 0
#define INCLUDE_SERIAL2 1
#define INCLUDE_SERIAL3 1
#define INCLUDE_SERIAL4 1

#define INCLUDE_IIC0 1
#define INCLUDE_IIC1 1

#define INCLUDE_ADCCPU 1
#define INCLUDE_ADC24BIT 1

#define INCLUDE_SSTFLASH 0

#define INCLUDE_RTC_DS1337 1

#define INCLUDE_FRAM 1

#define INCLUDE_FLASH2238 1

/**
 * This flag controls whether the segment flashing code is included.
 *
 * Segment flashing was never used in the R400, so I've added this option.
 * Removing segment flashing will remove a task, 2 LCD buffers and several functions,
 * so will save a significant amount of FLASH and RAM.
 */
#define INCLUDE_LCD_SEG_FLASH 0

#define INCLUDE_TICK_FUNCTION 1

#define INCLUDE_BUZZER 1

#define INCLUDE_KBD 1

#define INCLUDE_LCD 1


#endif


/*
 * Generate an error if trying to include IIC and SCI0/1 since the hardware doesn't support the
 * operation of both together.
 */
#if ((INCLUDE_IIC0 || INCLUDE_IIC1) && (INCLUDE_SERIAL0 || INCLUDE_SERIAL1))
#error Hardware does not support IIC and SCI0/1 simulaneously
#endif


/*
 * Number of characters in serial Rx and TX ring buffers
 */
#define SIO_RX_BUF_SIZE 140
#define SIO_TX_BUF_SIZE 64


/*
 * Serial port defaults
 */
#define SIO_DEF_MODE SIO_FDPLX_MODE
#define SIO_DEF_BAUD 57600UL
#define SIO_DEF_DATA_BITS SIO_DATABITS_EIGHT
#define SIO_DEF_PARITY_BITS SIO_PARITY_NONE
#define SIO_DEF_STOP_BITS SIO_STOPBITS_ONE
#define SIO_DEF_USECTS false
#define SIO_DEF_TERMINATIONS false
#define SIO_DEF_RS232_NOT_RS422 true


/*
 * Debugging setup
 */

/* On Windows and Linux debugging comes out on the console */
#if defined(__WIN32__) || defined(__linux__)
  #define DEBUG_CONSOLE 1
#else
  #define DEBUG_CONSOLE 0
#endif

#if (!DEBUG_CONSOLE)
  #if (TARGET == R500)
    #define DEBUG_PORT SioBuf3
    #define DEBUG_DEV dev_com3
  #elif (TARGET == R400)
    #define DEBUG_PORT SioBuf3
    #define DEBUG_DEV dev_com3
  #endif
#endif


/*
 * Task Priorities
 *
 * These are split up, so there is room for user task to be interspersed between them
 */
#define TASK_LCD_UPDATE_PRIO 3
#define TASK_BACKGROUND_PRIO 6
#define TASK_BUZZER_PRIO 7
#define TASK_REMOTE_PRIO 8
#define TASK_MAIN_PRIO 9


/*
 * Reuse initial stack -
 *
 * This is a dodge to get a bit more RAM, which
 * reuses the inital stack for the background task
 *
 * This relies on the fact that the main stack
 * will no longer be being used by the time the
 * background task is started
 */
#define REUSE_INITIAL_STACK 0

/*
 * Task Stack Sizes (in lots of 16 bits)
 */
#define TASK_LCD_UPDATE_STK_SIZE 2048
#define TASK_MAIN_STK_SIZE 2048
#define TASK_BACKGROUND_STK_SIZE_BYTES 4096
#define TASK_BACKGROUND_STK_SIZE (TASK_BACKGROUND_STK_SIZE_BYTES / 2)
#define TASK_BUZZER_STK_SIZE 2048
#define TASK_REMOTE_STK_SIZE 2048


/*
 * System stack initially used to bring up
 * ucos and the system (size in bytes)
 */
#if REUSE_INITIAL_STACK
    #define INITIAL_STK_SIZE TASK_BACKGROUND_STK_SIZE_BYTES
#else
    #define INITIAL_STK_SIZE 2048
#endif


/*
 * System Timer - Must be one of the 8 bit timers (TMR0, TMR1, TMR2, TMR3)
 */
#define SYS_TMR P_TMR3


/*
 * Backlight - Two operating modes
 *
 * - Simple - just write value to pwm
 * - Ramping - adjust the value slowly
 */
#define BL_SIMPLE 0
#define BL_RAMPING 1

#define BKLIGHT_MODE BL_RAMPING


/*
 * Byte Q - Two methods of locking
 *
 * - BQ_LOCK_INT_DISABLE - Disable interrupts to work on byte queue
 * - BQ_LOCK_SEMAPHORE - Use a semaphore to work on byte queue
 */
#define BQ_LOCK_INT_DISABLE 0
#define BQ_LOCK_SEMAPHORE 1

#if ( defined(__linux__) || defined(WIN32) )
  #define BQ_LOCK_MODE BQ_LOCK_SEMAPHORE
#elif ( defined(__HITACHI__) )
  #define BQ_LOCK_MODE BQ_LOCK_INT_DISABLE
#else
  #error "Byte queue locking not defined"
#endif


#ifndef INCLUDE_TICK_FUNCTION
  #error "INCLUDE_TICK_FUNCTION not defined"
#endif


/** Location of the default configuration file, only applies to windows and linux ports */
#define HW_LIB_CONFIG_FILE "./driverconfig.xml"


#endif /* R400_CONFIG_H_ */
