#include "rstdio.h"
#include "config.h"
#include "ucos_ii.h"
#include "tick.h"
#include "_tick.h"

#include "debug.h"


#if INCLUDE_TICK_FUNCTION

/*
 * This is a pointer that will maintain the linked list of functions to call
 */
struct t_tick_fn *tick_fn_list = (struct t_tick_fn *) NULL;


/*
 * TickCallFunctions -
 *
 * Call the functions in the tick_fn_list if they have come due.
 * Updates the counters for each of the functions
 */
static void TickCallFunctions(void)
{
    struct t_tick_fn *tick_fn = tick_fn_list;

    /*
     * For each element in the list
     */
    while (tick_fn != NULL)
    {

        /*
         * Preset == 0 is stopped
         */
        if (tick_fn->preset > 0)
        {
            /*
             * Decrement its counter to zero, but not below
             */
            if (tick_fn->count > 0)
                tick_fn->count--;

            /*
             * Its got to zero, call the function attached, and set the count back to the preset
             */
            if (tick_fn->count == 0)
            {
                if (tick_fn->fn != NULL)
                    tick_fn->fn(tick_fn->data);
                tick_fn->count = tick_fn->preset;
            }
        }

        /*
         * Onto the next one in the list
         */
        tick_fn = tick_fn->next;
    }
}


/*
 * TickAddFunction -
 *
 * Add a new function to call to the list of functions
 * Parameter new_tick_fn must have preset and function filled out
 */
void TickAddFunction(struct t_tick_fn *new_tick_fn)
{
    struct t_tick_fn *tick_fn, *tick_fn_prev;
    OS_DECLARE_CPU_SR();     /* Must be last variable */

    /*
     * Don't want to get interrupted while stuffing round with the list
     */
    OS_ENTER_CRITICAL();

    /*
     * Its the first one ever, just point the list pointer to it
     */
    if (tick_fn_list == NULL)
    {
        tick_fn_list = new_tick_fn;
    } else
    {

        /*
         * There is 1 or more already, navigate the list to find the end
         */
        tick_fn_prev = tick_fn = tick_fn_list;
        while (tick_fn != NULL)
        {
            tick_fn_prev = tick_fn;
            tick_fn = tick_fn->next;
        }

        /*
         * Attach it to the end
         */
        tick_fn_prev->next = new_tick_fn;
    }

    /*
     * Preset the new ones counter,
     * and make its next pointer null to mark the new end of the list
     * Out of critical section
     */
    new_tick_fn->count = new_tick_fn->preset;
    new_tick_fn->next = (struct t_tick_fn *) NULL;
    OS_EXIT_CRITICAL();
}


/*
 * TickChangePreset -
 *
 * Change the preset time of the tick function. Count is reset when this is called,
 * so the preset inteval will expire before the next function call.
 */
void TickChangePreset(struct t_tick_fn *new_tick_fn, unsigned short preset)
{
    OS_DECLARE_CPU_SR();     /* Must be last variable */

    /*
     * Don't want to get interrupted while stuffing round with the list
     */
    OS_ENTER_CRITICAL();

    /*
     * Preset the new ones counter,
     * and make its next pointer null to mark the new end of the list
     * Out of critical section
     */
    new_tick_fn->count = new_tick_fn->preset = preset;
    OS_EXIT_CRITICAL();
}


/*
 * init_tick -
 *
 * Start up the system tick at OS_TICKS_PER_SEC. Data parameter is not used, but just for init compatibility.
 */
void init_tick(void *data)
{

    /*
     * Start our timer at OS_TICKS_PER_SEC Hz, with interrupts enabled, and output toggling disabled
     */
//    init_tmr8_basic((struct st_tmr0 *) &SYS_TMR, OS_TICKS_PER_SEC, TRUE, FALSE);
}


/*
 * TickIntClear -
 *
 * Called by the timer ISR to run any timer functions, and clear the timer interrupt
 * Do not call from anywhere else
 */
void TickIntClear(void)
{
    TickCallFunctions();
/*    TMR8_CLR_CMA_INT(SYS_TMR); */
}


#endif /* #if INCLUDE_TICK_FUNCTION */
