#include "rstdio.h"
#include "config.h"
#include "ucos_ii.h"
#include "debug.h"

#include "_tick.h"
#include "wdt.h"
#include "_rio.h"
#include "rio.h"
#include "sio.h"
#include "lcd.h"
#include "kbd.h"
#include "iic.h"
#include "buzzer.h"
#include "adc_up.h"
#include "adc_ads1216.h"
#include "rtc.h"
#include "_bkgrnd.h"
#include "framdev.h"
#include "remoteexec.h"
#include "winsocket.h"
#include "driverconfig.h"
#include "we21hw.h"
#include "cmdline.h"

#include "init.h"


static void init_io_dev(void *data)
{
#if HW_DEBUG(INIT_DEBUG, HW_DEBUG_BASIC)
    if (rregister_io((struct t_io_dev *) data) < 0)
    {
        DEBUG_PRINTF("Registering device failed\n");
    }
#else
    rregister_io((struct t_io_dev *) data);
#endif
}


static void init_we21hw(void *data)
{
  (void) rio_init_we21hw("/dev/we21hw"); /* TODO set filename */
}


typedef struct
{
    void (*fn)(void *data);
    void *data;
#ifdef DEBUG
    char *fn_name;
    char *data_name;
#endif
} t_init_fns;


#ifdef DEBUG
    #define TBL_ENTRY(func, data) { func, (void *) data, # func, # data }
#else
    #define TBL_ENTRY(func, data) { func, (void *) data }
#endif


static const t_init_fns init_tbl[] =
{
    /* Initialise XML config file */
    TBL_ENTRY( vHwLibXMLConfig_Init, &hwlib_config_file ),

#ifdef __WIN32__
    /* Initialise windows sockets only on windows platform */
    TBL_ENTRY( WindowsSocketsInit, NULL ),
#endif /* #ifdef __WIN32__ */

    /* Give access to the WE2111 miscellaneous hardware */
    TBL_ENTRY(init_we21hw, NULL),

    /* Timer First */
#if INCLUDE_TICK_FUNCTION
    TBL_ENTRY(init_tick, NULL),
#endif

    /* Watchdog */
    TBL_ENTRY(WatchdogInit, NULL),

    /* Background task */
    TBL_ENTRY(BackgroundInit, NULL),

    /* IO subsystem */
    TBL_ENTRY(init_io, NULL),

    /* Remote IO mapper */
    TBL_ENTRY(vRemoteExecuteInit, NULL),

    /* Now any devices, in the order they are required (ie IICs before RTC/LCD) */
#if (INCLUDE_BUZZER != 0)
    TBL_ENTRY(init_io_dev, &dev_buz),
#endif

#if (INCLUDE_SERIAL0 != 0)
    TBL_ENTRY(init_io_dev, &dev_com0),
#endif
#if (INCLUDE_SERIAL1 != 0)
    TBL_ENTRY(init_io_dev, &dev_com1),
#endif
#if (INCLUDE_SERIAL2 != 0)
    TBL_ENTRY(init_io_dev, &dev_com2),
#endif
#if (INCLUDE_SERIAL3 != 0)
    TBL_ENTRY(init_io_dev, &dev_com3),
#endif
#if (INCLUDE_SERIAL4 != 0)
    TBL_ENTRY(init_io_dev, &dev_com4),
#endif

#if (INCLUDE_IIC0 != 0)
    TBL_ENTRY(init_io_dev, &dev_iic0),
#endif
#if (INCLUDE_IIC1 != 0)
    TBL_ENTRY(init_io_dev, &dev_iic1),
#endif

#if (INCLUDE_KBD != 0)
    TBL_ENTRY(init_io_dev, &dev_kbd),
#endif
#if (INCLUDE_LCD != 0)
//    TBL_ENTRY(InitBacklight, NULL),
    TBL_ENTRY(init_io_dev, &dev_lcd),
#endif

#if (INCLUDE_ADCCPU != 0)
    TBL_ENTRY(init_io_dev, &dev_adc),
#endif

#if (INCLUDE_ADC24BIT != 0)
    TBL_ENTRY(init_io_dev, &dev_adc24bit),
#endif

#if (INCLUDE_RTC_DS1337 != 0)
    TBL_ENTRY(init_io_dev, &dev_rtc),
#endif

#if (INCLUDE_FRAM != 0)
    TBL_ENTRY(init_io_dev, &dev_fram),
#endif
};


/*
 * system_init -
 *
 * Initialises all the system devices
 * and other hardware
 */
void system_init(void)
{
    unsigned short i, num_devs = sizeof(init_tbl) / sizeof(t_init_fns);

#if HW_DEBUG(INIT_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("System init:\n");
#endif

    for (i = 0; i < num_devs; i++)
    {
#if HW_DEBUG(INIT_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("  Initialising: %s, param %s...", init_tbl[i].fn_name, init_tbl[i].data_name);
#endif
        init_tbl[i].fn(init_tbl[i].data);
#if HW_DEBUG(INIT_DEBUG, HW_DEBUG_BASIC)
        DEBUG_PRINTF("done\n");
#endif
    }

#if HW_DEBUG(INIT_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("System init complete\n");
#endif
}
