#include "rstdio.h"
#include "ucos_ii.h"
#include "_sio.h"
#include "sio.h"
#include "tick.h"
#include "init.h"
#include "lcd.h"
#include "rio.h"
#include "config.h"
#include "_power.h"
#include "power.h"
#include "_kbd.h"

#include "debug.h"


/*
 * Stack for main task
 */
OS_STK MainTaskStk[TASK_MAIN_STK_SIZE];


/*
 * main -
 *
 * The application main()
 */
extern void appmain(unsigned char key);


/*
 * MainTask -
 *
 * The first task started when the OS is started
 * Used to bring up devices and hardware
 * before calling the users main()
 */
static void MainTask(void *data)
{
    unsigned char bootkey = (unsigned char) (int) data;

#ifdef DEBUG
    DEBUG_PRINTF("Main Task now running.\n");
#endif

    /*
     * Bring up system devices
     */
#ifdef DEBUG
    DEBUG_PRINTF("Bringing up system devices\n");
#endif
    system_init();

#ifdef DEBUG

    /*
     * Interrupt driven serial devices should now be running,
     * Open the debug channel and switch debug over to it
     */
#if DEBUG_CONSOLE
    InitDebug(NULL);
#else /* #if DEBUG_CONSOLE */
    if (ropen(&DEBUG_DEV, 0) < 0)
    {
        DEBUG_PRINTF("Could not open debug device.\n");
    }
    InitDebug(&DEBUG_DEV);
#endif /* #else -> #if DEBUG_CONSOLE */
    DEBUG_PRINTF("Debugging info now on device.\n");
#endif /* #ifdef DEBUG */

#if 0
    /*
     * Initialize uC/OS-II's statistics
     */
#ifdef DEBUG
    DEBUG_PRINTF("Determining  CPU's capacity...\n");
#endif
    OSStatInit();
#endif

#ifdef DEBUG
    DEBUG_PRINTF("All Done, Running main().\n");
#endif
    appmain(bootkey);

    /*
     * If main() returns, then turn power off
     */
#ifdef DEBUG
    DEBUG_PRINTF("main() returned, turn power off\n");
#endif
    PowerTurnOff();
}


/*
 * sys_main -
 *
 * This is where it all starts.
 *
 * PowerOn_Keys is passed through from the bootloader to indicate which
 * keys were pressed at startup
 */
void sys_main(unsigned long PowerOn_Keys, void (*UserFlashFn)(unsigned char sci))
{
    unsigned char bootkey;

#ifdef DEBUG

#if (!DEBUG_CONSOLE)
    /*
     * Bring up the serial port in polled mode for debugging purposes
     * and initialise the debug channel to use direct Sio writes for the moment
     */
    SioBasicInit(&DEBUG_PORT);
    SioCfgPort(&DEBUG_PORT);
    InitDebug(NULL);
#endif /* #if (!DEBUG_CONSOLE) */

    DEBUG_PRINTF("Rinstrum R400/R500 Driver Libraries\nby Ian Burns\n\n");
    DEBUG_PRINTF("Power on key status: %lx, User Flash Function @ %p\n", PowerOn_Keys, UserFlashFn);
#endif /* #ifdef DEBUG */

    /*
     * Note user flash function and boot keys
     */
    UserFlashFn = 0; /* Don't have a function for this */
    PowerOn_Keys = 1; /* != 0 means power key was pressed */
    set_user_flash_fn(UserFlashFn);
    bootkey = BootKeyTranslate(PowerOn_Keys);

    /*
     * Initialise the operating system
     */
#ifdef DEBUG
    DEBUG_PRINTF("Initialising OS\n");
#endif
    OSInit();

    /*
     * Create our main task,
     * and add stack checking for the main task and the idle if required
     */
#ifdef DEBUG
    DEBUG_PRINTF("Creating Main Task\n");
#endif
    OSTaskCreate(MainTask, (void *) (int) bootkey, &MainTaskStk[TASK_MAIN_STK_SIZE - 1], TASK_MAIN_PRIO);

    /*
     * Start the OS
     */
#ifdef DEBUG
    DEBUG_PRINTF("Starting OS (main task should run)\n");
#endif
    OSStart();

#ifdef DEBUG
    DEBUG_PRINTF("OS Has Exited: Bad Bad Stuff. This should never occur.\n");
#endif
}
