#ifndef RB_COMMAND_DOT_H
#define RB_COMMAND_DOT_H 1

#include <rinblock/rb_msg.h>
#include <rinblock/rinblock.h>


/*
 * Timeouts
 */

/** Default number of OS ticks for timeout */
#define RB_TIMEOUT_DEFAULT (OS_TICKS_PER_SEC)

/** Indefinite timeout */
#define RB_TIMEOUT_INDEFINITE 0

/** Timeout used when getting Rinblocks to start */
#define RB_TIMEOUT_STARTUP RB_TIMEOUT_DEFAULT

/** DIO Bus access timeout */
#define RB_TIMEOUT_DIO RB_TIMEOUT_DEFAULT

/** Timeout for rinblock tag operations */
#define RB_TIMEOUT_TAG RB_TIMEOUT_DEFAULT

/** Rinblock flash programming timeout */
#define RB_TIMEOUT_FLASH_LONG (OS_TICKS_PER_SEC * 3)

/** Rinblock watchdog timeout period */
#define RB_TIMEOUT_FLASH_WDT_LONG (RB_TIMEOUT_FLASH_LONG * 3)

/** Analog Output Calibration Timeout - Need to allow time for EEPROM write */
#define RB_TIMEOUT_AOUT_CAL RB_TIMEOUT_DEFAULT

/** DSD status timeout */
#define RB_TIMEOUT_DSD_STATUS (OS_TICKS_PER_SEC * 2   ) /* 2 seconds */

/** DSD write timeout */
#define RB_TIMEOUT_DSD_WRITE (OS_TICKS_PER_SEC * 2   ) /* 2 seconds */

/** DSD read timeout */
#define RB_TIMEOUT_DSD_READ (OS_TICKS_PER_SEC * 10  ) /* 10 second -- longer, especially for binary search in DSD */


/*
 * Defines for creating rinblock command and response buffers
 */


#define MAX(a,b) ((a)>(b)?(a):(b))


/** Define to determine the size of the rinblock command header */
#define RB_GET_SIZE_CMD_HDR() ( sizeof( struct s_rb_cmd_hdr ) )

/** Define to determine the size of a rinblock command (header + data) */
#define RB_GET_SIZE_COMMAND(structure) ( RB_GET_SIZE_CMD_HDR() + sizeof( structure ) )

/** Define to determine the size of a rinblock command and response (header + data) */
#define RB_GET_SIZE_COMMAND_RESPONSE(cmd_struct, reply_struct) ( RB_GET_SIZE_CMD_HDR() + MAX( sizeof( cmd_struct ), sizeof( reply_struct ) ) )

/** Define to determine the size of a scratchpad write command */
#define RB_GET_SIZE_CMD_SCRATCH_WRITE(len) ( RB_GET_SIZE_CMD_HDR() + \
    sizeof( struct s_rb_general_scratchpad_write ) - \
    MAX_RB_SCRATCHPAD_DATA_LENGTH + len )

/* Define to determine the size of a scratchpad read command */
#define RB_GET_SIZE_CMD_SCRATCH_READ(len) ( RB_GET_SIZE_CMD_HDR() + len )

/** Define to declare a command buffer (header + structure), and create a pointer to it (msg_ptr_name) */
#define RB_DECLARE_CMD_BUF(msg_ptr_name, size) \
    unsigned char buffer[ size ]; \
    struct s_rb_message *msg_ptr_name = (struct s_rb_message *) &buffer[ 0 ]

/** Define to declare a command buffer (header + structure), and create a pointer to it (msg_ptr_name) */
#define RB_DECLARE_CMD_BUFFER(msg_ptr_name, structure) \
    RB_DECLARE_CMD_BUF( msg_ptr_name, RB_GET_SIZE_COMMAND( structure ) )

/** Define to declare a command buffer (header + structure), and create a pointer to it (msg_ptr_name) */
#define RB_DECLARE_CMD_RESP_BUFFER(msg_ptr_name, cmd_struct, resp_struct) \
    RB_DECLARE_CMD_BUF( msg_ptr_name, RB_GET_SIZE_COMMAND_RESPONSE( cmd_struct, resp_struct ) )

/** Define to declare a command buffer with no structure (header only), and create a pointer to it (msg_ptr_name) */
#define RB_DECLARE_CMD_BUFFER_NO_PARAMS(msg_ptr_name) \
    RB_DECLARE_CMD_BUF( msg_ptr_name, RB_GET_SIZE_CMD_HDR() )


/*
 * Function prototypes
 */


/*
 * RinBlockInternalExecuteCommand -
 *
 * Send a command to a rinblock, and wait for a response if requested
 * Performs all locking and unlocking.
 */
unsigned char RinBlockInternalExecuteCommand(struct t_rinblock *rb, struct s_rb_message *rb_msg,
                                             unsigned short cmd_len, unsigned short resp_len,
                                             unsigned short timeout, unsigned char no_reply);

/*
 * RinBlockExecuteCommand -
 *
 * Send a command to a rinblock, and wait for a response
 * Performs all locking and unlocking.
 * Also checks that the rinblock is running
 */
unsigned char RinBlockExecuteCommand( struct t_rinblock *rb, struct s_rb_message *rb_msg,
                                      unsigned short cmd_len, unsigned short resp_len,
                                      unsigned short timeout, unsigned char cmd );


/*
 * RinBlockLock -
 *
 * Lock the rinblock for exclusive access.
 * Must have a matching call the RinBlockUnlock()
 *
 * Must call this function before RinBlockSendCommand() and RinBlockReceiveResponse()
 */
unsigned char RinBlockLock(struct t_rinblock *rb, unsigned short timeout);

/*
 * RinBlockUnlock -
 *
 * Unlock the rinblock so other tasks can access it
 * Must be used in a matching pair with RinBlockLock().
 * Don't call this function if RinBlockLock() call failed.
 */
void RinBlockUnlock(struct t_rinblock *rb);

/*
 * RinBlockIsLocked -
 *
 * Check whether the rinblock is currently locked
 */
unsigned char RinBlockIsLocked(struct t_rinblock *rb);

/*
 * RinBlockSendCommand -
 *
 * Send a command rb_msg of length cmd_len to rinblock rb, but don't wait for a response.
 * Use RinBlockReceiveResponse to get the answer back
 */
unsigned char RinBlockSendCommand(struct t_rinblock *rb, const struct s_rb_message *rb_msg,
                                  unsigned short cmd_len, unsigned short timeout);

/*
 * RinBlockReceiveResponse -
 *
 * Get the response rb_msg of length resp_len from rinblock rb, in answer to a command
 * sent by a call to RinBlockSendCommand(). Only call this function if RinBlockSendCommand()
 * call was successful.
 */
unsigned char RinBlockReceiveResponse(struct t_rinblock *rb, struct s_rb_message *rb_msg,
                                      unsigned short resp_len, unsigned short timeout);


#endif /* #ifndef RB_COMMAND_DOT_H */
