/***************************************************************************************************
MODULE:     DSD Module Header
VERSION:    0.01
CONTAINS:   Definitions for DSD (Data Storage Device) Module control.
COPYRIGHT:  (c) 2007 Rinstrum Pty Ltd
NOTES:      Uses 'standard' interface to DSD Module.

            See statuslist.h for a list of possible statuses returned if function return
            is 0(==failure).

18Jul2007 agl Original Version.
16Oct2006 agl Added RinBlockFreeUpDSD().
***************************************************************************************************/

#ifndef RB_DSD_DOT_H
#define RB_DSD_DOT_H 1

#include <rinhw/stdrinstrum.h>    /* for int8u, etc */
#include <rinblock/rinblock.h>       /* #includes rio.h & rb_command.h, which #includes rb_msg.h */

/* diagnostic: sets the last fail line of code. */
extern int16u   rbDsdLastFail;

/***************************************************************************************************
NAME:       RinBlockInitDSDDirectory()  << Public interface: preserves the DSD on every call (if same controller used).
            RinBlockFormatDSDPrivate()  << Not normally called by production code: typically only used by test code.

PARAMETERS: struct t_rinblock  *rb            Pointer to discovered Rinblock structure element .
            rb_dsd_directory_t *dsdDirectory  Pointer to DSD directory structure.
            <bool format>                     RinBlockFormatDSDPrivate() only.

NOTES:      RinBlockFormatDSDPrivate() is typically only called from test code.

            Used to initialise the DSD.

            RinBlockInitDSDDirectory() checks the dsdDirectory information, and if identical,
            returns 1 for success. If DSD detects a different Instrument Serial Number or Record
            Size, the DSD is erased, reconfigured & returns 1 for success.

            Formatting uses the DSD on rinblock rb with the contents of dsdDirectory which contains:
                Record Size;
                Instrument (controller) Serial Number;
            The following are updated by the DSD before returning:
                Record Capacity = ((Flash size minus 2 sectors)/recordLength)
                Records Used    = number of record slots in use (does not exclude 'empty' records).

RETURNS:    Function returns 1 for success, 0 otherwise: check dsdDirectory->status: returns
            STATUS_OK(==0) on success, a failure status otherwise. Normally, errors should not
            occur writing to a working DSD, but for the newbie,
                STATUS_INVALID_COMMAND will be returned if a parameter is out of range, etc.
            Testing DSDs in production, may result in errors such as:
                STATUS_HARDWARE_NOT_FOUND  i.e., no Flash and/or FRAM found
                STATUS_DSD_DIR_WRITE_FAIL  writing directory to FRAM failed

SIDE EFFECT:
*/
#define RinBlockInitDSDDirectory(rb,dsdDirectory)  RinBlockFormatDSDPrivate((rb), (dsdDirectory), FALSE);
int8u   RinBlockFormatDSDPrivate(struct t_rinblock *rb, rb_dsd_directory_t *dsdDirectory, bool format);

/***************************************************************************************************
NAME:       RinBlockReadDSDDirectory()

PARAMETERS: struct t_rinblock  *rb            Pointer to discovered Rinblock structure element .
            rb_dsd_directory_t *dsdDirectory  Pointer to DSD directory structure.

NOTES:      ***TYPICALLY*** only used by RinBlockInitDSDDirectory(), or by a special application
            for R&D or Service to read DSDs returned for service, repair, etc. This structure can
            then be used to spoof the DSD with a call to RinBlockInitDSDDirectory(), if returned
            for service/repair, etc. See RinBlockInitDSDDirectory() for a description of
            "dsdDirectory".

RETURNS:    Function returns 1 for success, 0 otherwise: check dsdDirectory->status returns
            STATUS_OK(==0) on success, a failure status otherwise. Normally, errors should not
            occur reading a working DSD, but for the newbie, a typical status for a new,
            unitialised DSD may be:
                STATUS_UNINITIALISED

SIDE EFFECT:
*/
/* int8u RinBlockReadDSDDirectory(struct t_rinblock *rb, rb_dsd_directory_t *dsdDirectory); */
#define RinBlockReadDSDDirectory(rb,dsdDirectory)  \
            RinBlockDSDDirectoryFunction(rb, dsdDirectory, RB_CMD_DSD_DIRECTORY_QUERY)

/***************************************************************************************************
NAME:       RinBlockClearDsdStatus()

PARAMETERS: struct t_rinblock  *rb            Pointer to discovered Rinblock structure element .
            rb_dsd_directory_t *dsdDirectory  Pointer to DSD directory structure.

NOTES:      ***TYPICALLY*** only used if the DSD reports a dsdStatus != STATUS_OK.

RETURNS:    Function returns 1 for success, 0 otherwise: check dsdDirectory->status returns
            STATUS_OK(==0) on success, a failure status otherwise. Normally, errors should not
            occur reading a working DSD, but for the newbie, a typical status for a new,
            unitialised DSD may be:
                STATUS_UNINITIALISED

SIDE EFFECT:
*/
/* int8u RinBlockClearDsdStatus(struct t_rinblock *rb, rb_dsd_directory_t *dsdDirectory); */
#define RinBlockClearDsdStatus(rb,dsdDirectory)  \
            RinBlockDSDDirectoryFunction(rb, dsdDirectory, RB_CMD_DSD_CLEAR_STATUS)

/**************************************************************************************************
NAME:       RinBlockDSDDirectoryFunction() -- see RinBlockReadDSDDirectory() &
                                              RinBlockClearDsdStatus() above.
*/
int8u RinBlockDSDDirectoryFunction(struct t_rinblock *rb, rb_dsd_directory_t *dsdDirectory,
                        int8u command);

/***************************************************************************************************
NAME:       RinBlockWriteDSD()

PARAMETERS: struct t_rinblock *rb           Pointer to discovered Rinblock structure element .
            rb_dsd_record_t *dsdRecord      Pointer to DSD Record to be written.

NOTES:      Writes data to the rinblock rb. The recordNumber and the data are in dsdRecord. The
            recordNumber must be higher than the last record.

RETURNS:    Function returns 1 for success, 0 otherwise: check dsdRecord->status: returns
            STATUS_OK(==0) on success, a failure status otherwise.  typical failure status is:
                STATUS_DSD_DIR_FULL: the device has the auto/manual free-up set to manual, but the DSD is full.
            Other statuses indicate more severe problems.

SIDE EFFECT:
*/
int8u RinBlockWriteDSD(struct t_rinblock *rb, rb_dsd_record_t *dsdRecord);

/***************************************************************************************************
NAME:       RinBlockReadDSD()

PARAMETERS: struct t_rinblock *rb           Pointer to discovered Rinblock structure element .
            DsdRead_t dsdReadType           Enumerated Read-type: see below.
            rb_dsd_read_cmd_t *dsdReadCmd   Pointer to structure containing Record ID, etc.
            rb_dsd_record_t *dsdRecord      Pointer to returned record buffer.
*/
typedef enum
{
    dsdRdRecord = 0,    /* read requested Record Number */
    dsdRdNext,          /* given Record Number, return next record */
    dsdRdPrevious,      /* given Record Number, return previous record */
    dsdRdOldest,        /* return oldest record. Does *not* use rb_dsd_read_cmd_t */
    dsdRdNewest,        /* return newest record. Does *not* use rb_dsd_read_cmd_t */
    dsdRdLastEntry      /* Not used by interface - used for range-checking only */
}   DsdRead_t;
/*
NOTES:      Reads data from the rinblock rb. recordNumber is in dsdReadCmd. (not used for
            dsdReadType==dsdRdOldest or dsdRdNewest)

RETURNS:    Results contained in dsdRecord.

            Function returns 1 for success, 0 otherwise: check dsdRecord->status: returns
            STATUS_OK(==0) on success, a failure status otherwise. Another typical status, eg, for
            non-existing record, may be:
                STATUS_DSD_REC_NOT_FOUND: record was not found, or next/prev is trying to read
                                          beyond the newest or oldest record.

SIDE EFFECT:In the DSD, when a record is read, it is placed in the DSD's 'Read-Cursor'.

            For example, if a DSD has 150,000 records with RecordIDs ranging from, say 1,000,000 to
            2,000,000, and you ask for record 1,777,777, then a relatively long binary search in
            Flash memory will be performed by the DSD Module.  If you then ask for the same record
            again, the ReadCursor will still point to record 1,777,777: no binary search is
            required, and the record is read from Flash and returned immediately.

            Similarly, the ReadCursor is used when 'Next' and 'Previous' style searching is
            performed, so if you ask for Next, the DSD checks if the RecordID you have passed-in is
            the same as the ReadCursor: if so, no binary search is required, and the next slot from
            Flash memory is reqd quickly, and returned.
*/
int8u RinBlockReadDSD(struct t_rinblock *rb, DsdRead_t dsdReadType,
                      rb_dsd_read_cmd_t *dsdReadCmd, rb_dsd_record_t *dsdRecord);

/***************************************************************************************************
NAME:       RinBlockReadDSD()

PARAMETERS: struct t_rinblock *rb           Pointer to discovered Rinblock structure element .
            rb_dsd_free_up_t  *freeUp       Pointer to 'free-up' structure.

NOTES:      Frees up space on the DSD.  This is a configuration command, and modifies the 'free-up'
            flag in the directory in the DSD. Input parameter is "percentFreeUp" in the
            "rb_dsd_free_up_t *freeUp" structure.

               0: set to automatic deletion from now on, as new records are added.
               1...100: set to manual deletion from now one, plus free-up this % of space immediately.

            Output parameters (apart from status and time) are

               recordCapacity: capacity reported by DSD module ((Flash size minus 2 sectors)/recordLength)
               recordsUsed   : number of record slots in use (does not exclude 'empty' records).

RETURNS:    Function returns 1 for success, 0 otherwise: check freeUp->status: returns
            STATUS_OK(==0) on success, a failure status otherwise.

SIDE EFFECT:If percentFreeUp is passed in as 0, will result in DSD perfoming automatic deletions of
            old records.
*/
int8u RinBlockFreeUpDSD(struct t_rinblock *rb, rb_dsd_free_up_t *freeUp);

#endif
