#ifndef RB_FLASH_DOT_H
#define RB_FLASH_DOT_H 1

#include <rinhw/rio.h>
#include <rinhw/wdt.h>
#include <rinblock/rinblock.h>


/*
 * Rin block programming errors
 *
 * RB_ERR_NO_ERR - success
 * RB_ERR_NOT_PRESENT - No rinblock present to program
 * RB_ERR_PROGRAM_MODE - Rinblock did not enter programming mode
 * RB_ERR_READ_FLASH_CFG - Rinblock did not return the flash configuration data
 * RB_ERR_FLASH_CFG_INCOMPATIBLE - The flash image (buf) does not match the configuration data
 * RB_ERR_FLASH_ERASE_FAIL - The flash erase command failed
 * RB_ERR_FLASH_PROGRAM_FAIL - Programming buf into the rinblock flash failed
 * RB_ERR_FLASH_VERIFY_FAIL - Verifying buf to the rinblock flash failed
 * RB_ERR_PROGRAM_DISCONNECT - Programming disconnect of rinblock failed
 * RB_ERR_REBAUD - Failed to negotiate a new baud rate
 * RB_ERR_FLASH_LOCALBUF_READONLY - RinBlockTestProgram() only, local buffer is readonly, most probably in flash, but
 *   needs to be in RAM, so it can be encrypted for transmission.
 */
#define RB_ERR_NO_ERR 0
#define RB_ERR_NOT_PRESENT 1
#define RB_ERR_PROGRAM_MODE 2
#define RB_ERR_READ_FLASH_CFG 3
#define RB_ERR_FLASH_CFG_INCOMPATIBLE 4
#define RB_ERR_FLASH_ERASE_FAIL 5
#define RB_ERR_FLASH_PROGRAM_FAIL 6
#define RB_ERR_FLASH_VERIFY_FAIL 7
#define RB_ERR_PROGRAM_DISCONNECT 8
#define RB_ERR_REBAUD 9
#define RB_ERR_FLASH_LOCALBUF_READONLY 10

/** Number of errors */
#define RB_NUM_ERR 11


/** Abbreviated error strings to match RB_ERR constants above */
#define RB_PROG_ERR_STRS { "Success", \
                           "No RB", \
                           "Prog.Mode", \
                           "Read Cfg", \
                           "Incomp", \
                           "Eras Fail", \
                           "Prog Fail", \
                           "Vrfy Fail", \
                           "DisConect", \
                           "Re-Baud", \
                           "BufRdOnly" \
                         }


/** More verbose error strings to match RB_ERR constants above */
#define RB_VERBOSE_PROG_ERR_STRS \
{ \
  "Success", \
  "No rinblock present to program", \
  "Rinblock did not enter programming mode", \
  "Rinblock did not return the flash configuration data", \
  "The flash image (buf) does not match the configuration data", \
  "The flash erase command failed", \
  "Programming buf into the rinblock flash failed", \
  "Verifying buf to the rinblock flash failed", \
  "Programming disconnect of rinblock failed", \
  "Failed to negotiate a new baud rate", \
  "Local buffer is readonly, needs to be in writeable, so it can be encrypted for transmission" \
}


/*
 * RinBlockProgram -
 *
 * Program a rin block by echoing characters from the serial port serial_dev to the
 * rin block rb, and visa versa, also toggles the watchdog wdt.
 * If no watchdog needs to be kicked, set wdt to NULL
 * Errors are returned in err
 *
 */
unsigned char RinBlockProgram(struct t_rinblock *rb, struct t_io_dev *serial_dev, struct t_wdt *wdt, unsigned char *err);

#endif
