/***************************************************************************************************
MODULE:     Rin Block Message Interface Header
VERSION:    0.01
CONTAINS:   Definitions for all Rin Block messages.
COPYRIGHT:  (c) 2004-2007 Rinstrum Pty Ltd
NOTES:      This file is common between the module code, and the rinblock library code on the R400.
            If you change one, you MUST change the other, otherwise it will break.

WARNING!    Commands must be less than RB_CMD_ERROR (0x80), else an error will be generated.
            An Error in RinBlock seems to set this top bit in ".resp".

Modifications
DPM 12Jan2004 - added BYTE_SWAP_WORD & MODEL_NUMBER's
SJ   1Feb2006 - Added generic memory access
DJA  7Jun2006 - Added RB_MODEL_NUMBER for M4311, M4321, M4331
DJA  5Feb2007 - Added tilt compensation module
AGL 12Sep2007 - Replaced DSD (Data Storage Device) with latest required functionality.
AGL 20Sep2007 - Replaced BYTE_SWAP_WORD() with BSD version with in-built casting (for safety);
                added BYTE_SWAP_LONG().
AGL 12Oct2007 - Added rb_dsd_free_up_t for DSD Module.
AGL 03Dec2007 - Added typedefs to all structures.

***************************************************************************************************/

#ifndef RB_MSG_DOT_H
#define RB_MSG_DOT_H 1

#include <rinhw/stdrinstrum.h>
#include <rinhw/rinendian.h>


/*
 * MS Bit set in response from module if error,
 * flag an inability to process the message
 */
#define RB_CMD_ERROR 0x80
#define RB_RESP_ERROR_MASK (RB_CMD_ERROR)


/*
 * All modules power up to this address
 */
#define RB_DEFAULT_ADDRESS 0x16


/*
 * Define maximum lengths for iic reads and writes
 * NOTE: Arrays are created on the stack of this many bytes + 1,
 * so be wary about making them too big.
 */
#define MAX_RB_SERIAL_DATA_LENGTH 10
#define MAX_RB_SCRATCHPAD_DATA_LENGTH 10

/* MAX_RB_DSD_DATA_LENGTH must be even. */
#ifndef MAX_RB_DSD_DATA_LENGTH
#define MAX_RB_DSD_DATA_LENGTH 50       /* 2007 July? AGL: Was 10. */
#endif

/*
 * Number of bytes in the scratchpad
 */
#define RB_SCRATCHPAD_SIZE 128


/*
 * Pack all structs since the rinblocks
 * expect the data this way
 */
#ifdef __HITACHI__
#pragma pack 1
#elif (defined(__GNUC__) && !defined(__AVR__))
#pragma pack(1)
#elif (defined(__GNUC__) && defined(__AVR__))
/* No struct packing required for AVR */
#else
#error Rinblock structs not packed
#endif


/*
 * Rin block set and get operating modes
 * and allowable modes
 */
#define RB_CMD_GET_MODE 0x02
#define RB_CMD_SET_MODE 0x42

#define RB_MODE_IDLE 0x00
#define RB_MODE_LISTEN_IRQ 0x01
#define RB_MODE_RESET 0x02
#define RB_MODE_RUN 0x03
#define RB_MODE_PROGRAM 0x04
#define RB_MODE_PROG_DONE 0x05

struct s_rb_general_mode
{
  int8u mode;
};


/*
 * Read and write rinblock iic address
 */
#define RB_CMD_READ_ADDRESS 0x01
#define RB_CMD_WRITE_ADDRESS 0x41

struct s_rb_general_address
{
  int8u address;
};


/*
 * Read and write boot stream,
 * used for re-programming modules
 */
#define RB_CMD_READ_BOOT_STREAM 0x03
#define RB_CMD_WRITE_BOOT_STREAM 0x43

struct s_rb_general_boot_stream
{
  int8u data;
};


/*
 * Read Tag Command
 * and rinblock models (part of the tag)
 */
#define RB_CMD_READ_TAG 0x04

#define RB_IFACE_VERS_NONE 0
#define RB_IFACE_TYPE_NONE 0
#define RB_MODEL_NUMBER_NONE 4000

#define RB_IFACE_TYPE_8DIO 1
#define RB_IFACE_VERS_8DIO 1
#define RB_MODEL_NUMBER_8DIO 4301
#define RB_HW_VERS_8DIO 1                 /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_4BUTTON 2
#define RB_IFACE_VERS_4BUTTON 1
#define RB_MODEL_NUMBER_4BUTTON 4302
#define RB_HW_VERS_4BUTTON 2              /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_AOUT 3
#define RB_IFACE_VERS_AOUT 1
#define RB_MODEL_NUMBER_AOUT 4401
#define RB_HW_VERS_AOUT 3                 /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_SERIAL 4
#define RB_IFACE_VERS_SERIAL 2            /*!< Current serial rinblock interface version */
#define RB_IFACE_VERS_SERIAL_ORIGINAL 1   /*!< Original serial rinblock interface version */
#define RB_IFACE_VERS_SERIAL_RING 2       /*!< Serial rinblock interface version when ring networking was introduced */
#define RB_MODEL_NUMBER_232_232 4201
#define RB_MODEL_NUMBER_232_485 4202
#define RB_MODEL_NUMBER_485_485 4203
#define RB_HW_VERS_SERIAL 4               /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_DSD 5
#define RB_IFACE_VERS_DSD 1
#define RB_MODEL_NUMBER_DSD 4501
#define RB_HW_VERS_DSD 5                 /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_16DIO 6
#define RB_IFACE_VERS_16DIO 1
#define RB_MODEL_NUMBER_16DIO 4311

#define RB_IFACE_TYPE_24DIO 7
#define RB_IFACE_VERS_24DIO 1
#define RB_MODEL_NUMBER_24DIO 4321

#define RB_IFACE_TYPE_32DIO 8
#define RB_IFACE_VERS_32DIO 1
#define RB_MODEL_NUMBER_32DIO 4331

#define RB_IFACE_TYPE_TILT 9
#define RB_IFACE_VERS_TILT 2              /*!< Current tilt rinblock interface version */
#define RB_IFACE_VERS_TILT_ORIGINAL 1     /*!< Original tilt rinblock interface version */
#define RB_IFACE_VERS_TILT_IMPROVED 2     /*!< Tilt rinblock interface version where improved calibration (RB_CMD_CALIBRATEB) and read error messages (RB_CMD_READ_TILT_B) were introduced */
#define RB_MODEL_NUMBER_TILT 4211
#define RB_HW_VERS_TILT 6                 /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_ETHERNET 0x0A
#define RB_IFACE_VERS_ETHERNET 2
#define RB_MODEL_NUMBER_ETHERNET 4221
#define RB_MODEL_NUMBER_WIFI 4222
#define RB_MODEL_NUMBER_ADVANCED_ETHERNET 4223
#define RB_HW_VERS_ETHERNET 7             /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_WE21_DISP 0x0B
#define RB_IFACE_VERS_WE21_DISP 1
#define RB_MODEL_NUMBER_WE21_DISP 2111
#define RB_HW_VERS_WE21_DISP 8            /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_WE21_CONN 0x0C
#define RB_IFACE_VERS_WE21_CONN 1
#define RB_MODEL_NUMBER_WE21_CONN 2111
#define RB_HW_VERS_WE21_CONN 9            /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_AIN 0x0D
#define RB_IFACE_VERS_AIN 1
#define RB_MODEL_NUMBER_AIN 4451
#define RB_HW_VERS_AIN 10                 /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_C500_DISP 0x0E
#define RB_IFACE_VERS_C500_DISP 1
#define RB_MODEL_NUMBER_C500_DISP 500
#define RB_HW_VERS_C500_DISP 11           /*!< Used to discriminate between module types when programming from the R400 */

#define RB_IFACE_TYPE_2AOUT 0x0F
#define RB_IFACE_VERS_2AOUT 1
#define RB_MODEL_NUMBER_2AOUT 4411

#define RB_IFACE_TYPE_3AOUT 0x10
#define RB_IFACE_VERS_3AOUT 1
#define RB_MODEL_NUMBER_3AOUT 4421

#define RB_IFACE_TYPE_4AOUT 0x11
#define RB_IFACE_VERS_4AOUT 1
#define RB_MODEL_NUMBER_4AOUT 4431

struct s_rb_general_tag
{
  int16u model_number;
  int32u serial_number;
  int16u software_version;
  int16u interface_version;
  int16u interface_type;
  int16u hardware_version;
  int8u  unused[2];
};


/*
 * Write tag shite (serial number & hardware version)
 */
#define RB_CMD_WRITE_TAG_SHITE 0x44

struct s_rb_general_tag_shite
{
  int32u serial_number;
  int16u hardware_version;
};

/*
 * Scratchpad read command and response
 */
#define RB_CMD_READ_SCRATCHPAD 0x05

struct s_rb_general_scratchpad_read_cmd
{
  int8u start;
  int8u length;
};

struct s_rb_general_scratchpad_read_reply
{
  int8u data[MAX_RB_SCRATCHPAD_DATA_LENGTH];
};


/*
 * Scratchpad write command
 */
#define RB_CMD_WRITE_SCRATCHPAD 0x45

struct s_rb_general_scratchpad_write
{
  int8u start;
  int8u length;
  int8u data[MAX_RB_SCRATCHPAD_DATA_LENGTH];
};


/*
 * Status short read command, and error bits
 */
#define RB_CMD_READ_STATUS_SHORT 0x06

#define RB_STATUS_ROM_ERROR 0x8000            /*!< Rinblock has wrong code programmed (ie M4201 code in M4302) */
#define RB_STATUS_EEPROM_ERROR 0x4000         /*!< Rinblock EEPROM error */
#define RB_STATUS_RAM_ERROR 0x2000            /*!< Rinblock RAM error (not used) */
#define RB_STATUS_TEMPERATURE_ERROR 0x1000    /*!< Rinblock outside temperature limits (not used) */
#define RB_STATUS_RUNNING 0x0800              /*!< Rinblock is running */
#define RB_STATUS_VOLTAGE_ERROR 0x0400        /*!< Rinblock outside voltage limits (brownout) */
#define RB_STATUS_MISC_ERROR 0x0200           /*!< Rinblock miscellaneous error (not used) */
#define RB_STATUS_BUS_ERROR 0x0100            /*!< Rinblock bus error (not used) */
#define RB_STATUS_FACTORY_ERROR 0x0080        /*!< Rinblock factory error (not used) */
#define RB_STATUS_RESET 0x0040                /*!< Rinblock has restarted */
#define RB_STATUS_MATCHPORT_ERROR 0x0020      /*!< Rinblock is not communicating with the matchport (M4221/M4222 only) */

struct s_rb_status_short_read
{
  int16u status;
};


/*
 * Status long read command
 */
#define RB_CMD_READ_STATUS_LONG 0x07

struct s_rb_status_long_read
{
  int32u long_status;
  int16u short_status;
};


/*
 * Serial Rinblock commands/types
 */
#define RB_CMD_READ_STREAM1 0x22
#define RB_CMD_WRITE_STREAM1 0x62
#define RB_CMD_WRITE_STREAM2 0x63

struct s_rb_serial_stream_read
{
  int8u length;
};

struct s_rb_serial_stream
{
  int8u data[MAX_RB_SERIAL_DATA_LENGTH];
};


/*
 * Configure rinblock serial ports
 */
#define RB_CMD_READ_CONFIG_SERIAL1 0x20
#define RB_CMD_READ_CONFIG_SERIAL2 0x21
#define RB_CMD_WRITE_CONFIG_SERIAL1 0x60
#define RB_CMD_WRITE_CONFIG_SERIAL2 0x61

struct s_rb_serial_configure
{
  int32u baud;
  int8u data_bits;
  int8u parity_bits;
  int8u stop_bits;
  int8u use_cts;
  int8u use_terminations;
};


/*
 * Serial stream usage/buffer sizes
 */
#define RB_CMD_READ_RX_USED 0x24
#define RB_CMD_READ_TX1_FREE 0x25
#define RB_CMD_READ_TX2_FREE 0x26
#define RB_CMD_READ_RX_BUF_SIZE 0x27
#define RB_CMD_READ_TX1_BUF_SIZE 0x28
#define RB_CMD_READ_TX2_BUF_SIZE 0x29

struct s_rb_serial_stream_usage
{
  int8u bytes_used;
};


/*
 * Serial CTS pin state
 */
#define RB_CMD_READ_TX1_CTS_STATE 0x2A

struct s_rb_serial_cts_state
{
  int8u state;
};


/*
 * Serial set ring network enable
 */
#define RB_CMD_SIO_WRITE_RING_ENABLE 0x2B

struct s_rb_serial_ring_enable
{
  int8u state;
};


/*
 * Serial Tx buffer flush
 */
#define RB_CMD_TX1_BUF_FLUSH 0x2C
#define RB_CMD_TX2_BUF_FLUSH 0x2D


/*
 * Tilt Compensation Rinblock commands/types
 */
#define RB_CMD_READ_TILT 0x32
#define RB_CMD_READ_TILT_IMPROVED 0x34
#define RB_CMD_READ_FACTORS 0x33
#define RB_CMD_CALIBRATE 0x66
#define RB_CMD_CALIBRATE_IMPROVED 0x70
#define RB_CMD_FACTORY_ZERO_SENSOR 0x67
#define RB_CMD_ZERO_SENSOR 0x68
#define RB_CMD_WRITE_FACTORS 0x69


struct s_rb_tilt_data
{
  int16s xAngle;    /* raw angle data from sensor */
  int16s yAngle;
  int16u cosX;      /* calculated in tilt compensation module */
  int16u cosY;
  int16u comp;      /* calculated compensation number */
  int8u errFlag;    /* bit error flag */
  /* spare | spare | spare | spare | ERR_MAX_OVER | ERR_TILT_SENSOR | spare | spare */
};


struct s_rb_tilt_data_improved
{
  int16s xAngle;      /* raw angle data from sensor */
  int16s yAngle;
  int16u cosX;        /* calculated in tilt compensation module */
  int16u cosY;
  int16u comp;        /* calculated compensation number */
  int8u errFlag;      /* bit error flag */
  /* spare | spare | spare | spare | ERR_MAX_OVER | ERR_TILT_SENSOR | spare | spare */
  int8u errFlagCalX;  /* bit error flag for calibration results in X direction */
  int8u errFlagCalY;  /* bit error flag for calibration results in Y direction */
  /* spare | spare | spare | NO_CONV | CD_LO | CD_HI | AB_LO | AB_HI */
};


/* These apply to errFlag */
#define RB_TILT_ERR_TILT_SENSOR 0x04
#define RB_TILT_ERR_MAX_OVER 0x08


/* These apply to errFlagCalX, errFlagCalY */
#define RB_TILT_CAL_PASS             0x00
#define RB_TILT_CAL_FAIL_AB_HIGH     0x01
#define RB_TILT_CAL_FAIL_AB_LOW      0x02
#define RB_TILT_CAL_FAIL_CD_HIGH     0x04
#define RB_TILT_CAL_FAIL_CD_LOW      0x08
#define RB_TILT_CAL_FAIL_NO_CONVERGE 0x10


struct s_rb_tilt_factors
{
  int16u aFactor;   /* factors used in calculating compensation */
  int16u bFactor;
  int16s cFactor;
  int16s dFactor;
  int8u maxTilt;    /* maximum tilt from R400 menu setting */
  /*
   * note this struct is used for both CMD_WRITE_FACTORS and CMD_READ_FACTORS
   * during a read, maxTilt is used as a calibration flag to let the R400 know
   * when the module has finished calculating the factors.
   * 0 = FAIL, 1 = IN PROGRESS, 2 = OK
   */
};


struct s_rb_tilt_calibrate /* variables used for automatic tilt calibration */
{
  int16s xPosAngle;
  int16s xNegAngle;
  int16u xPosComp;
  int16u xNegComp;
  int16s yPosAngle;
  int16s yNegAngle;
  int16u yPosComp;
  int16u yNegComp;
};


struct s_rb_tilt_calibrate_improved /* variables used for automatic tilt calibration */
{
  int16s xPosAngle;
  int16s xNegAngle;
  int16u xPosComp;
  int16u xNegComp;
  int16s yPosAngle;
  int16s yNegAngle;
  int16u yPosComp;
  int16u yNegComp;
  int16u maxTilt;
  int16u numDiv;
};


/*
 * Max Number of IO per card
 */
#define MAX_DIO_IO 8
#define MAX_AOUT_IO 2
#define MAX_IO_PULSE_COUNTERS 2


/*
 * Rinblock DIO configuration
 */
#define RB_CMD_READ_DIO_CONFIG 0x18
#define RB_CMD_WRITE_DIO_CONFIG 0x58

struct s_rb_dio_configure
{
  int8u data_direction;
  int8u data_invert;
  int8u aucDebounceDelay[MAX_DIO_IO];
};


/*
 * Structs for writing DIO/AOUT modules
 */
#if ( BYTE_ORDER == BIG_ENDIAN )
union u_rb_output
{
  struct
  {
    int8u level : 1;
    int8u holdoff : 7;
  } output_bits;
  int8u output_byte;
};

#elif ( BYTE_ORDER == LITTLE_ENDIAN )
union u_rb_output
{
  struct
  {
    int8u holdoff : 7;
    int8u level : 1;
  } output_bits;
  int8u output_byte;
};

#else
  #error "Unknown endian while generating type rb_output_t"
#endif


struct s_rb_dio_output_write
{
  union u_rb_output rb_output[MAX_DIO_IO];
};

struct s_rb_aout_output_write
{
  union u_rb_output rb_output[MAX_AOUT_IO];
  int32s aout_value;
};


/*
 * DIO and AOUT write commands
 */
#define RB_CMD_WRITE_AOUT_COMPOSITE 0x53
#define RB_CMD_WRITE_OUT_DELAYED 0x59

union u_rb_output_write
{
  struct s_rb_dio_output_write rb_dio_output_write;
  struct s_rb_aout_output_write rb_aout_output_write;
};

#if ( BYTE_ORDER == BIG_ENDIAN )
union u_rb_pulse_count_read
{
  struct
  {
    int8u pulse_count_1 : 4;
    int8u pulse_count_2 : 4;
  } pulse_count_bits;
  int8u pulse_count_byte;
};

#elif ( BYTE_ORDER == LITTLE_ENDIAN )
union u_rb_pulse_count_read
{
  struct
  {
    int8u pulse_count_2 : 4;
    int8u pulse_count_1 : 4;
  } pulse_count_bits;
  int8u pulse_count_byte;
};

#else
  #error "Unknown endian while generating type rb_pulse_count_read_t"
#endif


/*
 * DIO input read
 */
#define RB_CMD_READ_IN_DEBOUNCED 0x1A

struct s_rb_dio_read
{
  int8u io_data;
  union u_rb_pulse_count_read pulse_count;
};


/*
 * Rin block analog output factory cal (only used in test code)
 */
#define RB_CMD_READ_AO_FACTORY 0x10
#define RB_CMD_WRITE_AO_FACTORY 0x50

struct s_rb_aout_cal
{
  int16u uiFactoryCal0V;
  int16u uiFactoryCal10V;
  int16u uiFactoryCal4mA;
  int16u uiFactoryCal20mA;
  int16u uiCRCValue;
};


/*
 * Rin block analog output configuration
 */
#define RB_CMD_WRITE_AO_CONFIG 0x11
#define RB_CMD_READ_AO_CONFIG 0x51

struct s_rb_aout_configure
{
  int8u ucAnalogOutputType;   /* CURRENT = 0, VOLTAGE = 1 */
  int16s iUserZero;
  int16s iUserSpan;
  int8s cMaxLimit;            /* % of FSD that is maximum (from zero point) */
  int8s cMinLimit;            /* % of FSD that is minimum (from zero point) */
  int16u uiUpdateRate;        /* 0.1Hz per count on this - used in interpolation */
};


/*
 * Rin block analog output (also available from writing DIO output)
 */
#define RB_CMD_WRITE_ANALOG_OUT 0x52

struct s_rb_aout_access
{
  int32s data;
};

/***************************************************************************************************
 * Rin-block DSD (Data Storage Device: contains lots of nonvolatile memory).
 * 20070718 alf: Replaced with simpler interface
 */
#define RB_CMD_DSD_DIRECTORY_FORMAT   0x7E /* Format the directory on the DSD. */
#define RB_CMD_DSD_CLEAR_STATUS       0x7D /* Re-read directory from FRAM & clear status */
#define RB_CMD_DSD_DIRECTORY_QUERY    0x7C /* Read the directory from the DSD. */
#define RB_CMD_DSD_FREE_UP            0x7B /* Free Up requested percentage of memory. */
#define RB_CMD_DSD_WRITE_RECORD       0x7A /* Adds record to DSD. Rec Num must be > prev record */
/*                                            COMMANDS FOR DIFFERENT READS: */
#define RB_CMD_DSD_READ_RECORD        0x79 /* Read requested Record Number. */
#define RB_CMD_DSD_READ_RECORD_NEXT   0x78 /* Given Record Number, return next record. */
#define RB_CMD_DSD_READ_RECORD_PREV   0x77 /* Given Record Number, return previous record. */
#define RB_CMD_DSD_READ_RECORD_OLDEST 0x76 /* Return oldest (head) record. recordNumber is ignored. */
#define RB_CMD_DSD_READ_RECORD_NEWEST 0x75 /* Return newest (tail) record. recordNumber is ignored. */

#define RB_CMD_DSD_COMMENT_SIZE       20   /* maximum comment/description size (+1 will be added to this) */
#define DSD_COMMS_CRC_SEED            (int16u)(0xA1f)     /* Don't use 0. */

/*
 * Header at front of every DSD message.  We could use a structure for this, but this causes an
 * increase in code size since the offsets incur extra calculations by the compiler.
 */
#define RB_CMD_DSD_HEADER \
  int16u  status;         /* Ignored when sent controller=>Module, status when Module=>Controller. */ \
  int16u  time;           /* Length from controller=>Module, processing time in milliseconds for Module=>Controller. */ \
  int16u  baggage;        /* CRC from controller=>Module, else 'capture point' or other debug info for Module=>Controller. */

typedef struct s_rb_dsd_status
{
  RB_CMD_DSD_HEADER
  int32u  lastSectorErased;
}
rb_dsd_status_t;

typedef struct s_rb_dsd_directory
{
  RB_CMD_DSD_HEADER
  int32u  instrSerialNo;
  int32u  recordCapacity; /* Reported capacity returned by DSD (ignored during configuration) */
  int32u  recordsUsed;    /* Reported usage returned by DSD (ignored during configuration) */
  int8u   recordSize;     /* 2 to 254, rounded up if not even. */
  int8u   percentFreeUp;  /* Free up this %. 0: do automatic deletion; 1...100: manual deletion required */
  int8u   flashChipCount; /* Count of number of Flash Chips found */
  char    comment[RB_CMD_DSD_COMMENT_SIZE+1];
}
rb_dsd_directory_t;

typedef struct s_rb_dsd_free_up
{
  RB_CMD_DSD_HEADER
  int32u  recordCapacity; /* Reported capacity returned by DSD (ignored during clearing) */
  int32u  recordsUsed;    /* Reported usage returned by DSD (ignored during clearing) */
  int8u   percentFreeUp;  /* Free up this %. 0: do automatic deletion; 1...100: manual deletion required */
}
rb_dsd_free_up_t;

typedef struct s_rb_dsd_read_cmd
{
  RB_CMD_DSD_HEADER
  int32u  recordNumber;   /* Ignored for RB_CMD_DSD_READ_RECORD_OLDEST & RB_CMD_DSD_READ_RECORD_NEWEST */
}
rb_dsd_read_cmd_t;

typedef struct s_rb_dsd_record      /* used for both writing to, & returning from, DSD */
{
  RB_CMD_DSD_HEADER
  int32u  recordNumber;   /* Check record number on return. */
  int8u   data[MAX_RB_DSD_DATA_LENGTH];
}
rb_dsd_record_t;


/*
 * Ethernet module
 */

/** Command ID to read the IP configuration from the Ethernet module */
#define RB_CMD_ETH_READ_IPCONFIG 0x3A

/** Command ID to write the IP configuration to the Ethernet module */
#define RB_CMD_ETH_WRITE_IPCONFIG 0x6A

/** Structure containing the IP configuration */
struct s_rb_eth_ipconfigure
{
  int32u ui32IpAddress;            /* IP Address to use */
  int32u ui32IpNetMask;            /* Subnet mask */
  int32u ui32IpGateway;            /* Default gateway */
  int32u ui32IpDns1;               /* First DNS server */
  int32u ui32IpDns2;               /* Second DNS server */
  int8u bUseDHCP;                  /* Use DHCP to obtain IP settings */
};


/** Command to force a reset of the Matchport module (but no the Atmel) */
#define RB_CMD_ETH_MATCHPORT_RESET 0x6B

/** Command for forcing the Matchport to Rinstrum defaults */
#define RB_CMD_ETH_MATCHPORT_DEFAULTS 0x6C

/** Command to enable and disable the activity watchdog */
#define RB_CMD_ETH_WATCHDOG 0x6D

struct s_rb_eth_watchdog
{
  int8u enable;
};


/** Command for configuring the WIFI - not implemented, since we don't have a wireless module */
#define RB_CMD_WIFI_WRITE_WIFICONFIG 0x5B /* TODO Ethernet */


/** Structure containing WIFI configuration information */
struct s_rb_wifi_configure
{
  int8u szSSID[ 32 ];
  int8u ui8Encryption;
  int8u szPreSharedKey[ 64 ];
};


/*
 * WE2111 Display
 */


/** Command to write to the WE2111 display */
#define RB_CMD_WE21_DISP_WRITE_7SEG 0x5C


/** WE2111 Display Digit Segment Bitmap */
union s_we21_disp_digit
{
  struct
  {
    int8u seg_A : 1;     /*       A A A       */
    int8u seg_B : 1;     /*      F     B      */
    int8u seg_C : 1;     /*      F     B      */
    int8u seg_D : 1;     /*       G G G       */
    int8u seg_E : 1;     /*      E     C      */
    int8u seg_F : 1;     /*      E     C      */
    int8u seg_G : 1;     /*       D D D  H    */
    int8u seg_H : 1;
  } bits;
  int8u byte;
};


/** WE2111 Display Annunciators Segment Bitmap */
union s_we21_disp_annunciators
{
  struct
  {
    int8u coz          : 1;                /*!< Status centre-of-zero annunciator */
    int8u gross        : 1;                /*!< Status gross annunciator */
    int8u net          : 1;                /*!< Status net annunciator */
    int8u motion       : 1;                /*!< Status motion annunciator */
    int8u units_g      : 1;                /*!< Units g annunciator */
    int8u units_kg     : 1;                /*!< Units kg annunciator */
    int8u units_lb     : 1;                /*!< Units lb annunciator */
    int8u units_t      : 1;                /*!< Units t annunciator */
    int8u spStatus     : 8;                /*!< Setpoint annunciators. SP1 = 0x01, SP4 = 0x08, etc. */
    int8u range1       : 1;                /*!< Range 1 annunciator */
    int8u range2       : 1;                /*!< Range 2 annunciator */
    int8u range3       : 1;                /*!< Range 3 annunciator */
    int8u unused_byte2 : 5;
    int8u unused_byte3 : 8;
  } bits;

  int8u bytes[4];
};

/** Number of digits on the WE2111 display */
#define WE21_DISPLAY_NUM_DIGITS  6


/** Bitmap of the WE2111 display (segments and annunciators) */
struct s_rb_we21_disp_write_7seg
{
  union s_we21_disp_digit        digits[WE21_DISPLAY_NUM_DIGITS];
  union s_we21_disp_annunciators annunicators;
};


/** Command for reading WE2111 keys */
#define RB_CMD_WE21_DISP_READ_KEYBITS 0x1C


/*
#define RB_ERRORS_ALL                 0xFF
#define RB_WE21_DISP_ERR_QT           0x01
#define RB_WE21_DISP_ERR_QT_IIC       0x02
*/


/** Read the keybits from the WE2111 display */
struct s_rb_we21_disp_keybits
{
  int32u key_bits;
/*  int8u error_code; */
};


/** WE21 Buzzer write command */
#define RB_CMD_WE21_DISP_BUZZER         0x5E


/** Maximum number of buz events in 1 message */
#define RB_WE21_DISP_MAX_BUZZES_TO_SEND 5


/** Update frequency of the WE21 buzzer */
#define WE21BUZZ_UPDATE_FREQ_HZ         31


/** WE21 Buzzer operation enum */
typedef enum
{
  BUZZ_NONE = 0,
  BUZZ_CLEAR,       /*!< BUZZ_CLEAR: Stops buzzer and clears buzz buffer */
  BUZZ_ON,          /*!< BUZZ_ON: Switches buzzer on indefinately */
  BUZZ_STEP,        /*!< BUZZ_STEP: Single buzz, deleted when finished */
  BUZZ_CYCLE        /*!< BUZZ_CYCLE: Buzz not deleted when finished (until BUZZ_CLEAR). Will repeat after other items in the buffer have been completed. */
} BuzzerOperation;


/** Buzzer event structure */
struct s_rb_we21_disp_buzz
{
  int8u oper;            /*!< Buzzer operation as per BuzzerOperation enum. Can't use enum here as it will compile to different sizes on different architectures */
  int8u timeout;         /*!< Buzz duration (in units of WE21BUZZ_UPDATE_FREQ_HZ). Not used with BUZZ_CLEAR or BUZZ_ON. */
  int8u kHz;             /*!< Buzz frequency in kHz. 0 = OFF. Not used with BUZZ_CLEAR. */
  int8u unused;
};


/** Rinblock buzzer message */
struct s_rb_we21_disp_buzzer
{
  struct s_rb_we21_disp_buzz buzz[RB_WE21_DISP_MAX_BUZZES_TO_SEND];
};


/*
 * WE2111 Connector PCB
 */


/** Rinblock command to read ADCs */
#define RB_CMD_WE21_CONN_READ_ADC 0x1D

/* ADC setup */
#define RB_WE21_CONN_NUM_ADC_CHANNELS 4
#define RB_WE21_CONN_ADC_CH_VIN 0
#define RB_WE21_CONN_ADC_CH_TEMP 1
#define RB_WE21_CONN_ADC_CH_SNS_MINUS 2
#define RB_WE21_CONN_ADC_CH_SNS_PLUS 3

/** Structure of ADC readings */
struct s_rb_we21_conn_adc
{
  int16u adc_reading[RB_WE21_CONN_NUM_ADC_CHANNELS];
};


/*
 * Analog input module
 */

#define AIN_NUM_AIN 5
#define AIN_NUM_COMP 3
#define AIN_NUM_TOTAL (AIN_NUM_AIN+AIN_NUM_COMP)

#define RB_CMD_READ_AIN 0x10
struct s_rb_ain_data
{
  int16s ainVolts[AIN_NUM_TOTAL];  // x0.01V (or as calibrated)
};


/*
 * Rin block analog input configuration
 */
#define RB_CMD_READ_AIN_FACTORY 0x10
#define RB_CMD_WRITE_AIN_FACTORY 0x50
struct s_rb_ain_configure
{
  int16u rawPerVolt[AIN_NUM_TOTAL];   // x0.01 raw/volt
};


/*
 * C500 display
 */


/** Command to write to the C500 display */
#define RB_CMD_C500_DISP_WRITE_14SEG 0x5F


/** C500 Display Digit Segment Bitmap */
union s_c500_disp_digit
{
  struct
  {
    int16u seg_A : 1;     /*   AAAAAAAAA      */
    int16u seg_B : 1;     /*  FG   H   JB     */
    int16u seg_C : 1;     /*  F G  H  J B     */
    int16u seg_D : 1;     /*  F  G H J  B     */
    int16u seg_E : 1;     /*  F   GHJ   B     */
    int16u seg_F : 1;     /*   PPPP KKKK      */
    int16u seg_G : 1;     /*  E   NML   C     */
    int16u seg_H : 1;     /*  E  N M L  C     */
    int16u seg_J : 1;     /*  E N  M  L C     */
    int16u seg_K : 1;     /*  EN   M   LC     */
    int16u seg_L : 1;     /*   DDDDDDDDD   Q  */
    int16u seg_M : 1;     /*                  */
    int16u seg_N : 1;     /*                  */
    int16u seg_P : 1;     /*                  */
    int16u seg_Q : 1;     /*                  */
    int16u seg_unused : 1;
  } bits;
  int16u word;
};


/** C500 Display Annunciators Segment Bitmap */
union s_c500_disp_annunciators
{
#if (BYTE_ORDER == BIG_ENDIAN)
  /* Coldfire - big endian */
  struct
  {
    int32u unused_byte1    : 2;        /*!< Make it up to 32 bits */
    int32u nobbin          : 1;        /*!< Spare annunciator right below OK key */
    int32u hobbin          : 1;        /*!< Spare annunciator left below OK key */
    int32u check_under2    : 1;        /*!< Check weigh high under */
    int32u check_under1    : 1;        /*!< Check weigh low under */
    int32u check_pass      : 1;        /*!< Check weigh in range */
    int32u check_over1     : 1;        /*!< Check weigh low over */
    int32u check_over2     : 1;        /*!< Check weigh high over */
    int32u range3          : 1;        /*!< Range 3 annunciator */
    int32u range2          : 1;        /*!< Range 2 annunciator */
    int32u range1          : 1;        /*!< Range 1 annunciator */
    int32u setpoint7       : 1;        /*!< Setpoint 7 annunciator */
    int32u setpoint6       : 1;        /*!< Setpoint 6 annunciator */
    int32u setpoint5       : 1;        /*!< Setpoint 5 annunciator */
    int32u setpoint4       : 1;        /*!< Setpoint 4 annunciator */
    int32u setpoint3       : 1;        /*!< Setpoint 3 annunciator */
    int32u setpoint2       : 1;        /*!< Setpoint 2 annunciator */
    int32u setpoint1       : 1;        /*!< Setpoint 1 annunciator */
    int32u setpoint0       : 1;        /*!< Setpoint 0 annunciator */
    int32u units_spare     : 1;        /*!< Units spare annunciator */
    int32u units_t         : 1;        /*!< Units t annunciator */
    int32u units_oz        : 1;        /*!< Units oz annunciator */
    int32u units_g         : 1;        /*!< Units g annunciator */
    int32u units_kg        : 1;        /*!< Units kg annunciator */
    int32u units_lb        : 1;        /*!< Units lb annunciator */
    int32u status_spare    : 1;        /*!< Status spare annunciator (over f3 key) */
    int32u status_held     : 1;        /*!< Status held annunciator */
    int32u status_zeroband : 1;        /*!< Status zero band annunciator */
    int32u status_net      : 1;        /*!< Status net annunciator */
    int32u status_motion   : 1;        /*!< Status motion annunciator */
    int32u status_coz      : 1;        /*!< Status centre-of-zero annunciator */
  } bits;
#else
  /* i386/AVR - little endian */
  struct
  {
    int32u status_coz      : 1;        /*!< Status centre-of-zero annunciator */
    int32u status_motion   : 1;        /*!< Status motion annunciator */
    int32u status_net      : 1;        /*!< Status net annunciator */
    int32u status_zeroband : 1;        /*!< Status zero band annunciator */
    int32u status_held     : 1;        /*!< Status held annunciator */
    int32u status_spare    : 1;        /*!< Status spare annunciator (over f3 key) */
    int32u units_lb        : 1;        /*!< Units lb annunciator */
    int32u units_kg        : 1;        /*!< Units kg annunciator */
    int32u units_g         : 1;        /*!< Units g annunciator */
    int32u units_oz        : 1;        /*!< Units oz annunciator */
    int32u units_t         : 1;        /*!< Units t annunciator */
    int32u units_spare     : 1;        /*!< Units spare annunciator */
    int32u setpoint0       : 1;        /*!< Setpoint 0 annunciator */
    int32u setpoint1       : 1;        /*!< Setpoint 1 annunciator */
    int32u setpoint2       : 1;        /*!< Setpoint 2 annunciator */
    int32u setpoint3       : 1;        /*!< Setpoint 3 annunciator */
    int32u setpoint4       : 1;        /*!< Setpoint 4 annunciator */
    int32u setpoint5       : 1;        /*!< Setpoint 5 annunciator */
    int32u setpoint6       : 1;        /*!< Setpoint 6 annunciator */
    int32u setpoint7       : 1;        /*!< Setpoint 7 annunciator */
    int32u range1          : 1;        /*!< Range 1 annunciator */
    int32u range2          : 1;        /*!< Range 2 annunciator */
    int32u range3          : 1;        /*!< Range 3 annunciator */
    int32u check_over2     : 1;        /*!< Check weigh high over */
    int32u check_over1     : 1;        /*!< Check weigh low over */
    int32u check_pass      : 1;        /*!< Check weigh in range */
    int32u check_under1    : 1;        /*!< Check weigh low under */
    int32u check_under2    : 1;        /*!< Check weigh high under */
    int32u hobbin          : 1;        /*!< Spare annunciator left below OK key */
    int32u nobbin          : 1;        /*!< Spare annunciator right below OK key */
    int32u unused_byte1    : 2;        /*!< Make it up to 32 bits */
  } bits;
#endif
  int32u lword;
};


/** Number of digits on the C500 display */
#define C500_DISPLAY_NUM_DIGITS  6


/** Bitmap of the C500 display (segments and annunciators) */
struct s_rb_c500_disp_write_14seg
{
  union s_c500_disp_digit digits[C500_DISPLAY_NUM_DIGITS];
  union s_c500_disp_annunciators annunciators;
};


/** C500 display intensity */
#define RB_CMD_C500_DISP_BACKLIGHT 0x5D


/** C500 display/backlight intensity */
struct s_rb_c500_disp_backlight
{
  int8u backlight_percent;
};


/***************************************************************************************************/

/** Structure for the rinblock command header */
struct s_rb_cmd_hdr
{
  int8u cmd;
};


/** Structure for the rinblock response header */
struct s_rb_resp_hdr
{
  int8u resp;
};



/***************************************************************************************************
 * Structure containing rinblock command byte and
 * union of command structs
 */
struct s_rb_command
{
  struct s_rb_cmd_hdr header;
  union
  {
    struct s_rb_general_address     rb_general_address;

    struct s_rb_general_tag_shite   rb_general_tag_shite;
    struct s_rb_general_scratchpad_read_cmd rb_general_scratchpad_read_cmd;
    struct s_rb_general_scratchpad_write rb_general_scratchpad_write;
    struct s_rb_general_mode        rb_general_mode;
    struct s_rb_general_boot_stream rb_general_boot_stream;

    struct s_rb_serial_stream       rb_serial_stream;
    struct s_rb_serial_stream_read  rb_serial_stream_read;
    struct s_rb_serial_configure    rb_serial_configure;
    struct s_rb_serial_ring_enable  rb_serial_ring_enable;

    struct s_rb_dio_configure       rb_dio_configure;
    struct s_rb_dio_output_write    rb_dio_output_write;

    struct s_rb_aout_access         rb_aout_access;
    struct s_rb_aout_configure      rb_aout_configure;
    struct s_rb_aout_output_write   rb_aout_output_write;
    struct s_rb_aout_cal            rb_aout_cal;

    rb_dsd_status_t                 rb_dsd_status;
    rb_dsd_directory_t              rb_dsd_directory;
    rb_dsd_free_up_t                rb_dsd_free_up;
    rb_dsd_read_cmd_t               rb_dsd_read_cmd;
    rb_dsd_record_t                 rb_dsd_record;

    struct s_rb_tilt_factors        rb_tilt_factors;
    struct s_rb_tilt_calibrate      rb_tilt_calibrate;
    struct s_rb_tilt_calibrate_improved rb_tilt_calibrate_improved;

    struct s_rb_eth_ipconfigure     rb_eth_ipconfigure;
    /*struct s_rb_wifi_configure      rb_wifi_configure; This is a huge message and we need the RAM */
    struct s_rb_eth_watchdog        rb_eth_watchdog;

    struct s_rb_we21_disp_write_7seg rb_we21_disp_7seg;
    struct s_rb_we21_disp_buzzer     rb_we21_buzzer;

    struct s_rb_ain_configure       rb_ain_configure;

    struct s_rb_c500_disp_write_14seg rb_c500_disp_14seg;
    struct s_rb_c500_disp_backlight rb_c500_disp_backlight;
  }
  params;
};


/*
 * Structure containing rinblock response byte and
 * union of response structs
 */
struct s_rb_response
{
  struct s_rb_resp_hdr header;
  union
  {
    struct s_rb_general_address     rb_general_address;
    struct s_rb_general_tag         rb_general_tag;
    struct s_rb_general_scratchpad_read_reply rb_general_scratchpad_read_reply;
    struct s_rb_general_mode        rb_general_mode;
    struct s_rb_general_boot_stream rb_general_boot_stream;
    struct s_rb_status_short_read   rb_status_short_read;
    struct s_rb_status_long_read    rb_status_long_read;

    struct s_rb_serial_stream       rb_serial_stream;
    struct s_rb_serial_stream_usage rb_serial_stream_usage;
    struct s_rb_serial_configure    rb_serial_configure;
    struct s_rb_serial_cts_state    rb_serial_cts_state;

    struct s_rb_dio_read            rb_dio_read;
    struct s_rb_dio_configure       rb_dio_configure;

    struct s_rb_aout_configure      rb_aout_configure;
    struct s_rb_aout_cal            rb_aout_cal;

    rb_dsd_status_t                 rb_dsd_status;
    rb_dsd_directory_t              rb_dsd_directory;
    rb_dsd_free_up_t                rb_dsd_free_up;
    rb_dsd_record_t                 rb_dsd_record;

    struct s_rb_tilt_data           rb_tilt_data;
    struct s_rb_tilt_data_improved  rb_tilt_data_improved;
    struct s_rb_tilt_factors        rb_tilt_factors;

    struct s_rb_eth_ipconfigure     rb_eth_ipconfigure;

    struct s_rb_we21_disp_keybits   rb_we21_disp_keybits;
    struct s_rb_we21_conn_adc       rb_we21_conn_adc;

    struct s_rb_ain_configure       rb_ain_configure;
    struct s_rb_ain_data            rb_ain_read;
  }
  params;
};


/*
 * Back to normal struct packing
 */
#ifdef __HITACHI__
#pragma pack 2
#elif (defined(__GNUC__) && !defined(__AVR__))
#pragma pack()
#elif (defined(__GNUC__) && defined(__AVR__))
/* No struct packing required for AVR */
#else
#endif


/*
 * Rinblock message (contains command and response)
 */
struct s_rb_message
{
  union
  {
    struct s_rb_command rb_command;
    struct s_rb_response rb_response;
  }
  data;
};


#endif /* #ifndef RB_MSG_DOT_H */
