/*
 * rb_we21_adc.h
 *
 *  Created on: 17/05/2010
 *      Author: ianb
 */
#ifndef RB_WE21_ADC_H_
#define RB_WE21_ADC_H_

#include <rinhw/adc_up.h>
#include <rinblock/rinblock.h>


/* All calculations in millivolts, degrees kelvin, ohms */


/** Convert ADC Vin to real voltage using the divider values. All in millivolts. */
#define DIVIDER_MV_REAL_FROM_VIN( mv_vin, div_hi, div_lo ) ( mv_vin * ( div_hi + div_lo ) / div_lo )

/** Convert real voltage to ADC Vin using the divider values. All in millivolts. */
#define DIVIDER_VIN_FROM_MV_REAL( mv_real, div_hi, div_lo ) ( mv_real * div_lo / ( div_hi + div_lo ) )

/** Convert adc counts to real voltage. All in millivolts. */
#define DIVIDER_MV_REAL_FROM_CNT( adc_cnt, div_hi, div_lo, conv ) ( DIVIDER_MV_REAL_FROM_VIN( conv( adc_cnt ), div_hi, div_lo ) )

/** Convert real voltage to adc counts. All in millivolts. */
#define DIVIDER_CNT_FROM_MV_REAL( mv_real, div_hi, div_lo, conv ) ( conv( DIVIDER_VIN_FROM_MV_REAL( mv_real, div_hi, div_lo ) ) )


/*
 * WE21 Conversions (adc counts to real millivolts)
 */

/** 10 bit ADC on WE2111 */
#define WE21_ADC_BITS 10
#define WE21_ADC_MAX_CNT (1 << WE21_ADC_BITS)

/* Reference voltage on WE2111 in mV */
#define WE21_ADC_MAX_MV 5000

/* Resistor values in the WE2111 positive sense divider circuit */
#define WE21_SEN_POS_DIV_HI 100
#define WE21_SEN_POS_DIV_LO 100

/* Resistor divider for measuring negative sense (5v-0v when connected) */
#define WE21_SEN_NEG_DIV_UPPER 100
#define WE21_SEN_NEG_DIV_LOWER 100
#define WE21_SEN_NEG_UPPER_REF_MV 5000

/* Resistor values in the WE2111 input voltage sense divider circuit */
#define WE21_PWR_DIV_HI 330
#define WE21_PWR_DIV_LO 47
#define WE21_PWR_DIODE_OFFSET 300

/* At 25degrees celcius, the LM335 puts out 2.98 volts */
#define LM335_TEMP_MV_REF 2980
#define LM335_TEMP_KELVIN_REF ( 273 + 25 )

/* Kelvin to millivolts conversion factor */
#define LM335_TEMP_MV_PER_KELVIN 10

/** Convert adc counts to millivolts @ ADC input pin */
#define WE21_VIN_FROM_CNT( adc_cnt ) ( adc_cnt * WE21_ADC_MAX_MV / WE21_ADC_MAX_CNT )

/*
 * Macros below can be used in code for conversions
 */

/** Convert adc counts to real positive sense voltage (in millivolts) */
#define WE21_SEN_POS_MV_REAL_FROM_CNT( adc_cnt ) ( DIVIDER_MV_REAL_FROM_CNT( adc_cnt, WE21_SEN_POS_DIV_HI, WE21_SEN_POS_DIV_LO, WE21_VIN_FROM_CNT ) )

/** Convert adc counts to real input voltage (in millivolts) */
#define WE21_PWR_MV_REAL_FROM_CNT( adc_cnt ) ( DIVIDER_MV_REAL_FROM_CNT( adc_cnt, WE21_PWR_DIV_HI, WE21_PWR_DIV_LO, WE21_VIN_FROM_CNT ) + WE21_PWR_DIODE_OFFSET)

/** Convert adc counts to real temperature (in kelvin) */
#define WE21_TEMP_KELVIN_REAL_FROM_CNT( adc_cnt ) ( ( ( WE21_VIN_FROM_CNT( adc_cnt ) - LM335_TEMP_MV_REF ) / LM335_TEMP_MV_PER_KELVIN ) + LM335_TEMP_KELVIN_REF )

/** Convert adc counts to real negative sense voltage (in millivolts) */
#define WE21_SEN_NEG_MV_REAL_FROM_CNT( adc_cnt ) ( ( WE21_VIN_FROM_CNT( adc_cnt ) * ( WE21_SEN_NEG_DIV_UPPER + WE21_SEN_NEG_DIV_LOWER ) - WE21_SEN_NEG_DIV_LOWER * WE21_SEN_NEG_UPPER_REF_MV ) / WE21_SEN_NEG_DIV_UPPER )


/**
 * Create a WE2111 rinblock adc device
 *
 * @param io_dev is an uninitialised io device
 * @param rb is a pointer to a rinblock with an adc
 */
int RinBlockCreateWE21AdcDevice(struct t_io_dev *io_dev, struct t_rinblock *rb);


/**
 * Create an analog input rinblock adc device
 *
 * @param io_dev is an uninitialised io device
 * @param rb is a pointer to a rinblock with an adc
 */
int RinBlockCreateAnalogInAdcDevice(struct t_io_dev *io_dev, struct t_rinblock *rb);


#endif /* RB_WE21_ADC_H_ */
