#include <rinhw/rio.h>
#include <rinhw/iic.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rinblock.h>

#include "_iic_access.h"


/*
 * RinIICWrite -
 *
 * Write length bytes of the message to the rinblock rb, with a timeout.
 *
 * The sequence is: open rinblock iic device, set slot mask if rb not present,
 * write the data, clear mask if set, close the iic device, and return result
 */
int RinIICWrite(struct t_rinblock *rb, const struct s_rb_message *message, unsigned short length, unsigned short timeout)
{
    int res;
    unsigned char clear_mask = 0;

    /*
     * Open the IIC device
     */
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICWrite: Open IIC device\n");
#endif
    res = ropen(rb->iicdev, timeout);
    if (res < 0)
    {
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICWrite: Error/timeout opening device\n");
#endif
        return res;
    }

    /*
     * If this rinblock is not (yet) present, then set the slot enable mask
     * so only this one gets addressed
     */
    if (!rb->present)
    {
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Rinblock not yet present, set slot mask\n");
#endif
        rioctl(rb->iicdev, IOCTL_IIC_SET_SLOT_MASK, 1 << rb->bus_slot);
        clear_mask = 1;
    }

    /*
     * Set the IIC address to write to
     */
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICWrite: Set address\n");
#endif
    rioctl(rb->iicdev, IOCTL_IIC_SET_SLAVE_ADDR, rb->iic_address);

    /*
     * Do the IIC write
     */
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICWrite: Perform write of %d bytes, timeout %d, msg ptr %p, data ptr %p\n", length, timeout, message, &message->data.rb_command);
#endif
    res = rwrite(rb->iicdev, (void *) &message->data.rb_command, length, timeout);

#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICWrite: Write done, result %d\n", res);
#endif

    /*
     * Clear the slot mask if set afterwards,
     * giving the slave time to check the IRQ line on the R400
     */
    if (clear_mask != 0)
    {
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Clear slot mask\n");
#endif
        OSTimeDly(OS_TICKS_PER_SEC / 100);
        rioctl(rb->iicdev, IOCTL_IIC_DEFAULT_SLOT_MASK);
    }

    /*
     * Close the IIC device
     */
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICWrite: Close IIC device\n");
#endif
    rclose(rb->iicdev);
    return res;
}


/*
 * RinIICRead -
 *
 * Read length bytes from the rinblock rb into the message, with a timeout.
 *
 * The sequence is: open rinblock iic device, read the data,
 * close the iic device, and return result
 */
int RinIICRead(struct t_rinblock *rb, struct s_rb_message *message, unsigned short length, unsigned short timeout)
{
    int res;

    /*
     * Open the IIC device
     */
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICRead: Open IIC device\n");
#endif
    res = ropen(rb->iicdev, timeout);
    if (res < 0)
    {
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICRead: Error/timeout opening device\n");
#endif
        return res;
    }

    /*
     * Set the IIC address to read from
     */
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICRead: Set address\n");
#endif
    rioctl(rb->iicdev, IOCTL_IIC_SET_SLAVE_ADDR, rb->iic_address);

    /*
     * Do the IIC read
     */
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICRead: Perform read of %d bytes, timeout %d\n", length, timeout);
#endif
    res = rread(rb->iicdev, (void *) &message->data.rb_response, length, timeout);

#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICRead: Read done, result %d\n", res);
#endif

    /*
     * Close the IIC device
     */
#if RB_DEBUG(RB_IIC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinIICRead: Close IIC device\n");
#endif
    rclose(rb->iicdev);
    return res;
}
