#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <rinhw/lcd_generic.h>
#include <rinhw/rio.h>
#include <rinhw/lcd.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_msg.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_command.h>
#include "_rb_c500_lcdsegs.h"

#include "_rb_c500_lcd_common.h"


static unsigned char RinBlockC500WriteDisplay(struct t_rinblock *rb, const struct s_rb_c500_disp_write_14seg *c500_disp)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_c500_disp_write_14seg);
  size_t i;

  /* Put the data in network byte order */
  for (i = 0; i < C500_DISPLAY_NUM_DIGITS; i++)
    rb_msg->data.rb_command.params.rb_c500_disp_14seg.digits[i].word = htons(c500_disp->digits[i].word);
  rb_msg->data.rb_command.params.rb_c500_disp_14seg.annunciators.lword = htonl(c500_disp->annunciators.lword);

  return RinBlockExecuteCommand(rb, rb_msg,
                                RB_GET_SIZE_COMMAND(struct s_rb_c500_disp_write_14seg),
                                RB_GET_SIZE_CMD_HDR(),
                                RB_TIMEOUT_DEFAULT, RB_CMD_C500_DISP_WRITE_14SEG);
}


static unsigned char RinBlockC500SetBacklight(struct t_rinblock *rb, int backlight_pct)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_c500_disp_backlight);

  if (backlight_pct > 100)
    backlight_pct = 100;
  if (backlight_pct < 0)
    backlight_pct = 0;
  rb_msg->data.rb_command.params.rb_c500_disp_backlight.backlight_percent = backlight_pct;

  return RinBlockExecuteCommand(rb, rb_msg,
                                RB_GET_SIZE_COMMAND(struct s_rb_c500_disp_backlight),
                                RB_GET_SIZE_CMD_HDR(),
                                RB_TIMEOUT_DEFAULT, RB_CMD_C500_DISP_BACKLIGHT);
}


/** Write data to the LCD on the rinblock */
int rb_c500_lcd_write_data(struct s_lcd_definition *lcd_definition)
{
  struct t_rinblock *rb = lcd_definition->hw_data;
  unsigned char result;
  struct s_rb_c500_disp_write_14seg *segs;

#if RB_DEBUG(RB_LCD_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_c500_lcd_write_data called\n");
#endif

  if (NULL != lcd_definition->lcd_output_segments)
    segs = (struct s_rb_c500_disp_write_14seg *) lcd_definition->lcd_output_segments;
  else
    segs = (struct s_rb_c500_disp_write_14seg *) lcd_definition->lcd_segments;
  result = RinBlockC500WriteDisplay(rb, segs);

#if RB_DEBUG(RB_LCD_DEBUG_LVL, RB_DEBUG_FULL)
  if (0 == result)
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockC500WriteDisplay failed\n");
  }
#endif
  return (result) ? RIO_SUCCESS : RIO_ERROR;
}


/**
 * Write a character ch to digit position index on the LCD.
 *
 * cwm is the write mode - either set or or the data
 */
void rb_c500_lcd_put_char(struct s_lcd_definition *lcd_definition, unsigned char index, unsigned char ch, enum eCharWriteMode cwm)
{
  uint16_t segments;
  uint16_t *buffer;

#if RB_DEBUG(RB_LCD_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_c500_lcd_put_char called to write char 0x%02x to lcd digit %d\n", ch, index);
#endif

  /* Is the character in the table, if not display it with all segments on */
  if (ch >= NUM_SEG_TABLE)
  {
    ch = 11;
#if RB_DEBUG(RB_LCD_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Char modified to 0x%02x because outside table range (max 0x%02x)\n", ch, NUM_SEG_TABLE);
#endif
  }

  /* Convert the segments into a bitmap */
  segments = rb_c500_seg_table[ch];
#if RB_DEBUG(RB_LCD_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Character 0x%02x translates to bitmap 0x%02x\n", ch, segments);
#endif

  /* Modify the data, as dictated by the write mode */
  buffer = (uint16_t *) lcd_definition->lcd_segments;
  if (cwm == cwmOR)
  {
    /* Read the current data out of the LCD buffer */
    segments |= buffer[index];
  }
  buffer[index] = segments;
}


/** Work out if the digit index has a decimal point */
bool rb_c500_lcd_does_digit_have_dp(struct s_lcd_definition *lcd_definition, unsigned char index)
{
  /* Make sure its in range */
  if (index >= lcd_definition->lcd_num_characters)
    return false;

  /* All digits have decimal points on the C500 */
  return true;
}


/** Set the backlight (or intensity in the case of the C500) */
int rb_c500_lcd_setbacklight(struct s_lcd_definition *lcd_definition, int lvl_pct)
{
  unsigned char res;
  struct t_rinblock *rb = lcd_definition->hw_data;

  res = RinBlockC500SetBacklight(rb, lvl_pct);
  return (0 == res) ? RIO_ERROR : RIO_SUCCESS;
}
