#include <rinhw/ucos_ii.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_command.h>
#include <rinblock/rb_debug.h>
#include "_iic_access.h"


/*
 * RinBlockLock -
 *
 * Lock the rinblock for exclusive access.
 * Must have a matching call the RinBlockUnlock()
 *
 * Must call this function before RinBlockSendCommand() and RinBlockReceiveResponse()
 */
unsigned char RinBlockLock(struct t_rinblock *rb, unsigned short timeout)
{
    unsigned char err;

    /*
     * Wait till rin block available
     */
    OSSemPend(rb->msg_handling.busy_sem, timeout, &err);
    return (err == OS_NO_ERR);
}


/*
 * RinBlockUnlock -
 *
 * Unlock the rinblock so other tasks can access it
 * Must be used in a matching pair with RinBlockLock().
 * Don't call this function if RinBlockLock() call failed.
 */
void RinBlockUnlock(struct t_rinblock *rb)
{

    /*
     * Release rin block
     */
    OSSemPost(rb->msg_handling.busy_sem);
}


/*
 * RinBlockIsLocked -
 *
 * Check whether the rinblock is currently locked
 */
unsigned char RinBlockIsLocked(struct t_rinblock *rb)
{
    OS_SEM_DATA semdata;

    /*
     * Query the semaphore for its data
     */
    if (OSSemQuery(rb->msg_handling.busy_sem, &semdata) != OS_NO_ERR)
    {
        return 0;
    }

    /*
     * Check the sem count to determine if locked
     */
    return (semdata.OSCnt == 0);
}


/*
 * RinBlockSendCommand -
 *
 * Send a command rb_msg of length cmd_len to rinblock rb, but don't wait for a response.
 * Use RinBlockReceiveResponse to get the answer back
 */
unsigned char RinBlockSendCommand(struct t_rinblock *rb, const struct s_rb_message *rb_msg,
                                  unsigned short cmd_len, unsigned short timeout)
{
    int res;

    /*
     * Send the command to the rinblock
     */
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockSendCommand: Sending command to device %d, slot %d, rb_msg ptr %p\n", rb->iic_address, rb->bus_slot, rb_msg);
#endif
    res = RinIICWrite(rb, rb_msg, cmd_len, timeout);

    /*
     * Debug error/success print
     */
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_FULL)
    if (res == cmd_len)
    {
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "IIC Write Success, result %d\n", res);
    }
#endif
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    if (res != cmd_len)
    {
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "IIC Write Failed, result %d\n", res);
    }
#endif

    /*
     * Return result
     */
    return (res == cmd_len);
}


/*
 * RinBlockReceiveResponse -
 *
 * Get the response rb_msg of length resp_len from rinblock rb, in answer to a command
 * sent by a call to RinBlockSendCommand(). Only call this function if RinBlockSendCommand()
 * call was successful.
 */
unsigned char RinBlockReceiveResponse(struct t_rinblock *rb, struct s_rb_message *rb_msg,
                                      unsigned short resp_len, unsigned short timeout)
{
    int res;
    unsigned long start_time = OSTimeGet();

    /*
     * Read the response from the rinblock
     */
    do
    {
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_FULL)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockReceiveResponse: Reading response from device %d, slot %d\n", rb->iic_address, rb->bus_slot);
#endif
        res = RinIICRead(rb, rb_msg, resp_len, timeout);

#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_FULL)
        if (res == resp_len)
        {
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "IIC Read Success, result %d\n", res);
        }
#endif
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
        if (res != resp_len)
        {
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "IIC Read Failed, result %d\n", res);
        }
#endif
        if (res == resp_len)
        {
            return 1;
        }
    }
    while ((timeout == 0) || (OSTimeGet() < (start_time + timeout)));

#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "IIC Read Failed, timeout\n");
#endif

    /*
     * Return result
     */
    return 0;
}


/*
 * RinBlockInternalExecuteCommand -
 *
 * Send a command to a rinblock, and wait for a response if requested
 * Performs all locking and unlocking.
 */
unsigned char RinBlockInternalExecuteCommand(struct t_rinblock *rb, struct s_rb_message *rb_msg,
                                             unsigned short cmd_len, unsigned short resp_len,
                                             unsigned short timeout, unsigned char no_reply)
{
    unsigned char res;

#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInternalExecuteCommand() called, rb_msg pointer %p\n", rb_msg);
#endif

    if (!RinBlockLock(rb, timeout))
    {
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Timeout locking rinblock\n");
#endif
        return 0;
    }

    res = RinBlockSendCommand(rb, rb_msg, cmd_len, timeout);
    if ((res) && (!no_reply))
    {
        res = RinBlockReceiveResponse(rb, rb_msg, resp_len, timeout);
    }

    RinBlockUnlock(rb);
    return res;
}


/*
 * RinBlockExecuteCommand -
 *
 * Send a command to a rinblock, and wait for a response
 * Performs all locking and unlocking.
 * Also checks that the rinblock is running
 */
unsigned char RinBlockExecuteCommand( struct t_rinblock *rb, struct s_rb_message *rb_msg,
                                      unsigned short cmd_len, unsigned short resp_len,
                                      unsigned short timeout, unsigned char cmd )
{
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockExecuteCommand called\n");
#endif

    /* Stop if rinblock is not running */
    if ( 0 == rb->running )
    {
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockExecuteCommand: rinblock not running\n");
#endif
        return 0;
    }

    /* Setup command */
    rb_msg->data.rb_command.header.cmd = cmd;

    /*
     * Execute command
     */
    if ( !RinBlockInternalExecuteCommand( rb, rb_msg, cmd_len, resp_len, timeout, 0 ) )
    {
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockExecuteCommand: no reply from rinblock\n");
#endif
        return 0;
    }
    if ( 0 != ( rb_msg->data.rb_response.header.resp & RB_RESP_ERROR_MASK ) )
    {
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockExecuteCommand: rinblock returned error\n");
#endif
        return 0;
    }

#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockExecuteCommand: success\n");
#endif
    return 1;
}
