#include <stdarg.h>
#include <string.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_command.h>

#include "_rb_dio.h"
#include <rinblock/rb_dio.h>


/*
 * write_rb_io_config -
 *
 * Write the io configuration to the rinblock, return 1 for success, 0 for failure
 * Only applicable for modules with configurable IO.
 */
static unsigned char write_rb_io_config(struct t_rinblock *rb, struct s_rb_dio_configure *dio_cfg)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, struct s_rb_dio_configure );

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	RB_DEBUG_PRINTF(RB_DEBUG_DEV, "write_rb_io_config called\n");
#endif

    /*
     * Setup to write the configuration
     */
    rb_msg->data.rb_command.params.rb_dio_configure = *dio_cfg; /* No endian swap required */

    /*
     * Do the configuration write, and get response and handle rinblock not responding
     */
    return RinBlockExecuteCommand( rb, rb_msg,
                                   RB_GET_SIZE_COMMAND( struct s_rb_dio_configure ),
                                   RB_GET_SIZE_CMD_HDR(),
                                   RB_TIMEOUT_DEFAULT, RB_CMD_WRITE_DIO_CONFIG );
}


/*
 * read_rb_io_config -
 *
 * Read the io configuration from the rinblock, return 1 for success, 0 for failure
 * Only applicable for modules with configurable IO.
 */
static unsigned char read_rb_io_config(struct t_rinblock *rb, struct s_rb_dio_configure *dio_cfg)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, struct s_rb_dio_configure );
    unsigned char result;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	RB_DEBUG_PRINTF(RB_DEBUG_DEV, "read_rb_io_config: ");
#endif

    /*
     * Setup command to read the configuration
     * Do the configuration read, and get response and handle rinblock not responding
     */
    result = RinBlockExecuteCommand( rb, rb_msg,
                                     RB_GET_SIZE_CMD_HDR(),
                                     RB_GET_SIZE_COMMAND( struct s_rb_dio_configure ),
                                     RB_TIMEOUT_DEFAULT, RB_CMD_READ_DIO_CONFIG );
    if (result)
    {

        /*
         * Read config success, copy the data to return
         */
        *dio_cfg = rb_msg->data.rb_response.params.rb_dio_configure;  /* No endian swap required */
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success\n");
#endif
    }
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    else
    {
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "failure\n");
    }
#endif
    return result;
}


/*
 * CheckIOAgainstHardware -
 *
 * Check that this hardware io pin can be configured as requested
 */
static unsigned char CheckIOAgainstHardware(struct t_rinblock *rb, unsigned char rb_io_bit, enum eDIOType io_type)
{
    unsigned char result;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	RB_DEBUG_PRINTF(RB_DEBUG_DEV, "CheckIOAgainstHardware: ");
#endif

    /*
     * If there is no IO on this module, then return
     */
    if (rb->RBDioHardware == (enum eRBDioHardware *) NULL)
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no io on this module\n");
#endif
        return 0;
    }

    /*
     * Check IO type against hardware
     */
    switch (rb->RBDioHardware[rb_io_bit])
    {
        case dioInput:
            result = (io_type == ioInput);
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "hardware is input, configuring as io_type %d, result %d\n", io_type, result);
#endif
            break;

        case dioOutput:
            result = (io_type == ioOutput);
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "hardware is output, configuring as io_type %d, result %d\n", io_type, result);
#endif
            break;

        case dioInOut:
            result = (io_type == ioInput) || (io_type == ioOutput);
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "hardware is io, configuring as io_type %d, result %d\n", io_type, result);
#endif
            break;

/*        case dioNotPresent: */
        default:
            result = 0; /* (io_type == ioUnused); */
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "hardware is either not present or error, result %d\n", result);
#endif
            break;
    }
    return result;
}


/*
 * CheckIONeedsConfiguring -
 *
 * Check whether this IO needs configuring in hardware
 */
static unsigned char CheckIONeedsConfiguring(struct t_rinblock *rb, unsigned char rb_io_bit)
{

    /*
     * If there is no IO on this module,
     * then no there is nothing to configure
     */
    if (rb->RBDioHardware == (enum eRBDioHardware *) NULL)
    {
        return 0;
    }

    /*
     * Only in/out pins need configuring
     */
    return (rb->RBDioHardware[rb_io_bit] == dioInOut);
}


/*
 * ConfigureIOHardware -
 *
 * Configure the hardware as requested io type, with inversion and debouncing (only relavent for inputs)
 */
static unsigned char ConfigureIOHardware(struct t_rinblock *rb, unsigned char io_bit, enum eDIOType io_type, enum eDIOInversion io_invert, unsigned char debounce_time)
{
    struct s_rb_dio_configure dio_cfg;

    /*
     * Read its current configuration
     */
    if (!read_rb_io_config(rb, &dio_cfg))
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Module config read failed\n");
#endif
        return 0;
    }

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Module DDR = 0x%02x\n", dio_cfg.data_direction);
#endif

    /*
     * Set/clear the data direction register bit, depending on the io type
     */
    if (io_type == ioOutput)
    {
        dio_cfg.data_direction |= (unsigned char) (0x01 << io_bit);
    }
    else
    {
        dio_cfg.data_direction &= (unsigned char) ~(0x01 << io_bit);
    }

    /*
     * Set/clear the data inversion register bit, depending on the io inversion
     */
    if (io_invert == ioNormal)
    {
        dio_cfg.data_invert &= (unsigned char) ~(0x01 << io_bit);
    }
    else
    {
        dio_cfg.data_invert |= (unsigned char) (0x01 << io_bit);
    }

    /*
     * Set the appropriate debounce time
     */
    dio_cfg.aucDebounceDelay[io_bit] = debounce_time;

    /*
     * Write the DIO configuration back
     */
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
  	RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Modified Module DDR = 0x%02x\n", dio_cfg.data_direction);
#endif
    return write_rb_io_config(rb, &dio_cfg);
}


/*
 * RinBlockConfigureIO -
 *
 * Configure the IO on a rinblock, as input/output, inversion,
 * and debouce time (for inputs only). Will work for all rinblocks
 * with IO (input only included).
 *
 * Returns 1 for success, 0 otherwise
 */
unsigned char RinBlockConfigureIO(struct t_rinblock *rb, unsigned char io_bit, enum eDIOType io_type, enum eDIOInversion io_invert, unsigned char debounce_time)
{
    unsigned char result;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockConfigureIO: rb (ptr) %p, io_bit %d, io_type %d, io_invert %d, debounce_time %d\n", rb, io_bit, io_type, io_invert, debounce_time);
#endif

    /*
     * Check that hardware IO is of the correct type
     */
    if (!CheckIOAgainstHardware(rb, io_bit, io_type))
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "IO Type does not match rinblock hardware, not configuring\n");
#endif
        return 0;
    }

    /*
     * Its all ok, so configure it
     */
    if (CheckIONeedsConfiguring(rb, io_bit))
    {
        result = ConfigureIOHardware(rb, io_bit, io_type, io_invert, debounce_time);
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Configuring hardware result %d\n", result);
#endif
        return result;
    }
    else
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "No hardware configuration required\n");
#endif
        return 1;
    }
}


/*
 * 4 Button input dio hardware configuration
 */
static const enum eRBDioHardware RBDioHw_4Button[MAX_DIO_PER_RINBLOCK] =
    { dioInput, dioInput, dioInput, dioInput, dioNotPresent, dioNotPresent, dioNotPresent, dioNotPresent };

/*
 * 8IO dio hardware configuration
 */
static const enum eRBDioHardware RBDioHw_8IO[MAX_DIO_PER_RINBLOCK] =
    { dioInOut, dioInOut, dioInOut, dioInOut, dioInOut, dioInOut, dioInOut, dioInOut };

/*
 * Analog Ouput dio hardware configuration
 */
static const enum eRBDioHardware RBDioHw_AOut[MAX_DIO_PER_RINBLOCK] =
    { dioInOut, dioInOut, dioNotPresent, dioNotPresent, dioNotPresent, dioNotPresent, dioNotPresent, dioNotPresent };


/*
 * RinBlockSetupDioHardware -
 *
 * Setup the DIO hardware for a rinblock based on the model number.
 * This sets up the rinblock struct to match the type of io
 * hardware on the rinblock. The default is to
 */
void RinBlockSetupDioHardware(struct t_rinblock *rb)
{
    switch (rb->tag.interface_type)
    {
        case RB_IFACE_TYPE_8DIO:
            rb->RBDioHardware = RBDioHw_8IO;
            break;

        case RB_IFACE_TYPE_16DIO:
            rb->RBDioHardware = RBDioHw_8IO;
            break;

        case RB_IFACE_TYPE_24DIO:
            rb->RBDioHardware = RBDioHw_8IO;
            break;

        case RB_IFACE_TYPE_32DIO:
            rb->RBDioHardware = RBDioHw_8IO;
            break;

        case RB_IFACE_TYPE_4BUTTON:
            rb->RBDioHardware = RBDioHw_4Button;
            break;

        case RB_IFACE_TYPE_AOUT:
            rb->RBDioHardware = RBDioHw_AOut;
            break;

        case RB_IFACE_TYPE_2AOUT:
            rb->RBDioHardware = RBDioHw_AOut;
            break;

        case RB_IFACE_TYPE_3AOUT:
            rb->RBDioHardware = RBDioHw_AOut;
            break;

        case RB_IFACE_TYPE_4AOUT:
            rb->RBDioHardware = RBDioHw_AOut;
            break;

/*        case RB_MODEL_SERIAL: */
/*        case RB_MODEL_DSD: */
        default:
            rb->RBDioHardware = (enum eRBDioHardware *) NULL;
            break;
    }
}


/*
 * RinBlockIOCreateCommand -
 *
 * Setup the struct, command and length ready for writing to a
 * rinblock to set the io
 */
static unsigned char RinBlockIOCreateCommand(struct t_rinblock *rb, struct s_rb_message *rb_msg, union u_rb_output_write *out_data, unsigned char *cmd_len)
{
    unsigned char data_len;
    unsigned char cmd;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockIOCreateCommand: ");
#endif

    /*
     * Based on module type, setup cmd, data length,
     * and copy output over (except 4 btn input module)
     */
    switch (rb->tag.interface_type)
    {

        case RB_IFACE_TYPE_8DIO:
        case RB_IFACE_TYPE_16DIO:
        case RB_IFACE_TYPE_24DIO:
        case RB_IFACE_TYPE_32DIO:
            cmd = RB_CMD_WRITE_OUT_DELAYED;
            data_len = RB_GET_SIZE_COMMAND(struct s_rb_dio_output_write);
            rb_msg->data.rb_command.params.rb_dio_output_write = out_data->rb_dio_output_write;
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "DIO Module, cmd %d, write len %d\n", cmd, data_len);
#endif
            break;

        case RB_IFACE_TYPE_4BUTTON:
            cmd = RB_CMD_READ_IN_DEBOUNCED;
            data_len = RB_GET_SIZE_CMD_HDR();
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "4 Btn Input Module, cmd %d, write len %d\n", cmd, data_len);
#endif
            break;

        case RB_IFACE_TYPE_AOUT:
        case RB_IFACE_TYPE_2AOUT:
        case RB_IFACE_TYPE_3AOUT:
        case RB_IFACE_TYPE_4AOUT:
            cmd = RB_CMD_WRITE_AOUT_COMPOSITE;
            data_len = RB_GET_SIZE_COMMAND(struct s_rb_aout_output_write);

#if (BYTE_ORDER == NETWORK_ENDIAN)
            rb_msg->data.rb_command.params.rb_aout_output_write = out_data->rb_aout_output_write;
#else /* #if (BYTE_ORDER == NETWORK_ENDIAN) */
            rb_msg->data.rb_command.params.rb_aout_output_write = out_data->rb_aout_output_write;
            rb_msg->data.rb_command.params.rb_aout_output_write.aout_value = htonl( out_data->rb_aout_output_write.aout_value );
#endif /* #else -> #if (BYTE_ORDER == NETWORK_ENDIAN) */
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Analog Out Module, cmd %d, write len %d\n", cmd, data_len);
#endif
            break;

/*        case RB_MODEL_SERIAL: */
/*        case RB_MODEL_DSD: */
        default:
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Other module type\n");
#endif
            return 0;
    }

    /*
     * Set the command in the struct,
     * and set the command length
     */
    rb_msg->data.rb_command.header.cmd = cmd;
    *cmd_len = data_len;
    return 1;
}


/*
 * RinBlockIOCreateReadCommand -
 *
 * Setup the struct, command and length ready for reading the inputs from a rinblock
 */
static unsigned char RinBlockIOCreateReadCommand(struct t_rinblock *rb, struct s_rb_message *rb_msg, unsigned char *cmd_len)
{
    unsigned char data_len;
    unsigned char cmd;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockIOCreateReadCommand: ");
#endif

    /*
     * Based on module type, setup cmd, data length,
     * and copy output over (except 4 btn input module)
     */
    switch (rb->tag.interface_type)
    {
        case RB_IFACE_TYPE_8DIO:
        case RB_IFACE_TYPE_16DIO:
        case RB_IFACE_TYPE_24DIO:
        case RB_IFACE_TYPE_32DIO:
        case RB_IFACE_TYPE_4BUTTON:
        case RB_IFACE_TYPE_AOUT:
        case RB_IFACE_TYPE_2AOUT:
        case RB_IFACE_TYPE_3AOUT:
        case RB_IFACE_TYPE_4AOUT:
            cmd = RB_CMD_READ_IN_DEBOUNCED;
            data_len = RB_GET_SIZE_CMD_HDR();
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "IO Module, cmd %d, write len %d\n", cmd, data_len);
#endif
            break;
/*        case RB_MODEL_SERIAL: */
/*        case RB_MODEL_DSD: */
        default:
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Other module type\n");
#endif
            return 0;
    }

    /*
     * Set the command in the struct,
     * and set the command length (+ 1 for the command byte)
     */
    rb_msg->data.rb_command.header.cmd = cmd;
    *cmd_len = data_len;
    return 1;
}


/*
 * RinBlockIOGetRespLength -
 *
 * Figure out the number of bytes to expect back from an IO module,
 * based on the module type
 *
 * Returns 1 for success, 0 if the module has no IO
 */
static unsigned char RinBlockIOGetRespLength(struct t_rinblock *rb, unsigned char *resp_len)
{
    switch (rb->tag.interface_type)
    {
        case RB_IFACE_TYPE_8DIO:
        case RB_IFACE_TYPE_16DIO:
        case RB_IFACE_TYPE_24DIO:
        case RB_IFACE_TYPE_32DIO:
        case RB_IFACE_TYPE_4BUTTON:
        case RB_IFACE_TYPE_AOUT:
        case RB_IFACE_TYPE_2AOUT:
        case RB_IFACE_TYPE_3AOUT:
        case RB_IFACE_TYPE_4AOUT:
            *resp_len = RB_GET_SIZE_COMMAND( struct s_rb_dio_read );
            break;

/*        case RB_MODEL_SERIAL: */
/*        case RB_MODEL_DSD: */
        default:
            return 0;
    }
    return 1;
}


/*
 * InternalRinBlockWriteOutputs -
 *
 * Internal function to write the output data or input read command to the rinblock
 * Used by RinBlockWriteOutputs() and RinBlockPrimeForRead()
 */
static unsigned char InternalRinBlockWriteOutputs(struct t_rinblock *rb, const struct s_rb_message *rb_msg, unsigned char cmd_len)
{
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    unsigned char i;
#endif

	/*
	 * Debug dump command
	 */
	#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_FULL)
	RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Cmd byte and output data: ");
	for (i = 0; i < cmd_len; i++)
	{
		RB_DEBUG_PRINTF(RB_DEBUG_DEV, "0x%02x ", buf[i]);
	}
	RB_DEBUG_PRINTF(RB_DEBUG_DEV, "\n");
	#endif

	/*
	 * Lock the rinblock
	 */
	if (!RinBlockLock(rb, RB_TIMEOUT_DIO))
	{
	#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
		RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockLock() failed\n");
	#endif
		return 0;
	}

	/*
	 * Send the command to the rinblock
	 */
	if (!RinBlockSendCommand(rb, rb_msg, (unsigned short) cmd_len, RB_TIMEOUT_DIO))
	{
		RinBlockUnlock(rb);
	#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
		RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockSendCommand() failed, unlocking rinblock and returning\n");
	#endif
		return 0;
	}

	/*
	 * Write success
	 */
	#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
	RB_DEBUG_PRINTF(RB_DEBUG_DEV, "InternalRinBlockWriteOutputs success\n");
	#endif
	return 1;
}


/*
 * RinBlockWriteOutputs -
 *
 * Writes the output data to the rinblock rb. As part of this process
 * it locks the rinblock, and must be unlocked with a call to RinBlockReadInputs()
 * otherwise the system will lock up.
 *
 * Returns 1 for success, 0 otherwise.
 * NOTE only call RinBlockReadInputs() if RinBlockWriteOutputs() has returned 1.
 *
 * It is ok to access other rinblocks while another is locked
 */
unsigned char RinBlockWriteOutputs(struct t_rinblock *rb, union u_rb_output_write *o_data)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, union u_rb_output_write );
    unsigned char cmd_len;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockWriteOutputs called\n");
#endif

    /*
     * Create the command appropriate for this rinblock
     */
    if (!RinBlockIOCreateCommand(rb, rb_msg, o_data, &cmd_len))
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockIOCreateCommand() failed\n");
#endif
        return 0;
    }

    /* Write the command and optional data to the rinblock */
    return ( InternalRinBlockWriteOutputs(rb, rb_msg, cmd_len) );
}


/*
 * RinBlockPrimeForRead -
 *
 * Writes the read inputs command to the rinblock rb. As part of this process
 * it locks the rinblock, and must be unlocked with a call to RinBlockReadInputs()
 * otherwise the system will lock up.
 *
 * Returns 1 for success, 0 otherwise.
 * NOTE only call RinBlockReadInputs() if RinBlockPrimeForRead() has returned 1.
 *
 * It is ok to access other rinblocks while another is locked
 */
unsigned char RinBlockPrimeForRead(struct t_rinblock *rb)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, union u_rb_output_write );
    unsigned char cmd_len;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockWriteOutputs called\n");
#endif

    /*
     * Create the command appropriate for this rinblock
     */
    if (!RinBlockIOCreateReadCommand(rb, rb_msg, &cmd_len))
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockIOCreateReadCommand() failed\n");
#endif
        return 0;
    }

    /* Write the command and optional data to the rinblock */
    return ( InternalRinBlockWriteOutputs(rb, rb_msg, cmd_len) );
}


/*
 * RinBlockReadInputs -
 *
 * Reads the input data from the rinblock rb. As part of this process
 * it locks the rinblock, and must be unlocked with a call to RinBlockReadInputs()
 * otherwise the system will lock up.
 *
 * Returns 1 for success, 0 otherwise.
 * NOTE only call RinBlockReadInputs() if RinBlockWriteOutputs() has returned 1.
 *
 * It is ok to access other rinblocks while another is locked
 */
unsigned char RinBlockReadInputs(struct t_rinblock *rb, struct s_rb_dio_read *data)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, struct s_rb_dio_read );
    unsigned char resp_len;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockReadInputs called\n");
#endif

    /*
     * Figure out how much data to expect back,
     * if any (ie Serial rinblock)
     */
    if (!RinBlockIOGetRespLength(rb, &resp_len))
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockIOGetRespLength() failed\n");
#endif
        return 0;
    }

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_FULL)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "resp_len = %i\n",resp_len);
#endif

    /*
     * Check that the rinblock is locked.
     * It should be from the RinBlockWriteOutputs call
     * and if not something is not right.
     */
    if (!RinBlockIsLocked(rb))
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockIsLocked() reports rinblock is not locked, returning\n");
#endif
        return 0;
    }

    /*
     * Get the response from the rinblock,
     * and unlock it and return 0 in the case of error
     */
    if (!RinBlockReceiveResponse(rb, rb_msg, (unsigned short) resp_len, RB_TIMEOUT_DIO))
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockReceiveResponse() failed\n");
#endif
        RinBlockUnlock(rb);
        return 0;
    }

    /*
     * Check that there wasn't a module error,
     * unlock and return 0 if so.
     */
    if ((rb_msg->data.rb_response.header.resp & RB_RESP_ERROR_MASK) != 0)
    {
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Module error, failed\n");
#endif
        RinBlockUnlock(rb);
        return 0;
    }

    /*
     * Read response success, unlock the module
     */
#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockReadInputs success\n");
#endif
    RinBlockUnlock(rb);
    *data = rb_msg->data.rb_response.params.rb_dio_read;

#if RB_DEBUG(RB_DIO_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "cmd=0x%02x data=", rb_msg->data.rb_response.resp);
    rwrite(RB_DEBUG_DEV, (char *)data, resp_len-1, 0);
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "\n");
#endif

    return 1;
}
