/***************************************************************************************************
MODULE:     DSD Module
VERSION:    0.01
CONTAINS:   Controller code to access DSD (Data Storage Device) Module.
COPYRIGHT:  (c) 2007 Rinstrum Pty Ltd
NOTES:      Uses 'standard' interface to DSD Module.

            >>>>> Refer to rb_dsd.h for a description of the functions and parameters. <<<<<

18Jul2007 agl Original Version.
16Oct2006 agl Added RinBlockFreeUpDSD().
***************************************************************************************************/

#include <stdarg.h>
#include <string.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rstdio.h>
#include <rinhw/rio.h>
#include <rinhw/kbd.h>
#include <rinhw/power.h>
#include <rinhw/lcd.h>
#include <rinhw/sio.h>
#include <rinblock/status.h>
#include <rinhw/rtc.h>
#include <rinblock/crc.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_command.h>

#include <rinblock/rb_dsd.h>


/*
 * Defines from alflb.h
 */
#define CAST_PTR(type_cast,new_ptr,old_address)     type_cast *new_ptr = (type_cast *)(old_address)


int16u   rbDsdLastFail;


/***************************************************************************************************
NAME:       RinBlockDsdCrcCalc()

PARAMETERS: void *pData     Pointer to DSD message.
            int16u length   Length in bytes of DSD message.

NOTES:      Since the CRC is embedded in the CRC structure itself, the CRC footprint has to be
            set to some default known value: this was chosen to be 0.  Note: this is not the CRC
            seed value.

            Also, the 'time' parameter is set to the message length, again prior to CRC calculation.

RETURNS:    Void

SIDE EFFECT:Updates the length and the CRC in the DSD message/structure.
*/
static void RinBlockDsdCrcCalc(void *pData, int16u length)
{
  CAST_PTR(rb_dsd_status_t, genericHeaderRx, pData);

  genericHeaderRx->baggage = 0; /* clear the CRC 'footprint'. */
  genericHeaderRx->time    = length;
  genericHeaderRx->baggage = CRC16_Calc(pData, length, DSD_COMMS_CRC_SEED);
}

/*
 * NOTES:   Typical production code does not call this function directly: see rb_dsd.h for the
 *          public function RinBlockInitDSDDirectory().
 */
int8u RinBlockFormatDSDPrivate(struct t_rinblock *rb,
                               rb_dsd_directory_t *dsdDirectory, bool format)
{
  RB_DECLARE_CMD_BUFFER( rb_msg, rb_dsd_directory_t );
  rb_dsd_directory_t localDsdDirectory;

/*  ezi_dbg_raw(dbuf, dsdDirectory, sizeof(rb_dsd_directory_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'B', dbuf->str); */
  RinBlockDsdCrcCalc(dsdDirectory, sizeof(rb_dsd_directory_t));
/*  ezi_dbg_raw(dbuf, dsdDirectory, sizeof(rb_dsd_directory_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'A', dbuf->str); */

  /*
   * Get a local copy for comparisons, because RinBlockExecuteCommand() reuses the send buffer
   * as the receive buffer.
   */
  localDsdDirectory = *dsdDirectory;

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RB Init Dir: ");
  #endif

  if ( format == FALSE )
  {
    /* Read the DSD Directory (control block), and if identical, returns 1 for success. */
    if ( ! RinBlockReadDSDDirectory(rb, dsdDirectory) ) /* if not successful... */
    {
      /*
       * No need for #if RB_DEBUG(...), since this is done in RinBlockReadDSDDirectory().
       * Also, sets rbDsdLastFail if there was an error.
       */
      rbDsdLastFail = __LINE__;
      return 0;
    }
  }

  /* Check for identical instrument parameters, or a forced format: */
  if ( localDsdDirectory.instrSerialNo != dsdDirectory->instrSerialNo
    || localDsdDirectory.recordSize    != dsdDirectory->recordSize
    || format == TRUE )
  {
    /*
     * We have detected a different Instrument Serial Number or Record Size, or a forced format.
     * Therefore, reconfigure the DSD (which effectively 'erases' it) & return 1 for success.
     */
    rb_msg->data.rb_command.params.rb_dsd_directory = localDsdDirectory; /* TODO Endian swap */

    /* Do the setup, write, and get response and handle rinblock not responding */
    if (!RinBlockExecuteCommand( rb, rb_msg,
                                 RB_GET_SIZE_COMMAND( rb_dsd_directory_t ),
                                 RB_GET_SIZE_COMMAND( rb_dsd_directory_t ),
                                 RB_TIMEOUT_DSD_WRITE, RB_CMD_DSD_DIRECTORY_FORMAT ) )
    {
      #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "fail:no reply\n");
      #endif
      rbDsdLastFail = __LINE__;
      return 0;
    }

    /* Copy the DSD Control Block back to the application. */
    *dsdDirectory = rb_msg->data.rb_response.params.rb_dsd_directory; /* TODO Endian swap */
    #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)                /* Read data success */
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success\n");
    #endif
    return 1;
  }
  else  /* instrument parameters are identical: return success. */
  {
    return 1;
  }
}


int8u RinBlockDSDDirectoryFunction(struct t_rinblock *rb, rb_dsd_directory_t *dsdDirectory,
                                   int8u command)
{
  RB_DECLARE_CMD_BUFFER( rb_msg, rb_dsd_directory_t );

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockReadDSDDirectory: ");
  #endif

/*  ezi_dbg_raw(dbuf, dsdDirectory, sizeof(rb_dsd_directory_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'B', dbuf->str); */
  RinBlockDsdCrcCalc(dsdDirectory, sizeof(rb_dsd_directory_t));
/*  ezi_dbg_raw(dbuf, dsdDirectory, sizeof(rb_dsd_directory_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'A', dbuf->str); */
  rb_msg->data.rb_command.params.rb_dsd_directory = *dsdDirectory; /* TODO Endian swap */

  /* Do the setup, write, and get response and handle rinblock not responding */
  if (!RinBlockExecuteCommand( rb, rb_msg,
                               RB_GET_SIZE_COMMAND( rb_dsd_directory_t ),
                               RB_GET_SIZE_COMMAND( rb_dsd_directory_t ),
                               RB_TIMEOUT_DSD_STATUS, command ) )
  {
    #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no reply, failed\n");
    #endif
    rbDsdLastFail = __LINE__;
    return 0;
  }

  /* Copy the DSD Control Block back to the application. */
  *dsdDirectory = rb_msg->data.rb_response.params.rb_dsd_directory; /* TODO Endian swap */

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)                /* Read data success */
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success\n");
  #endif
  return 1;
}


int8u RinBlockWriteDSD(struct t_rinblock *rb, rb_dsd_record_t *dsdRecord)
{
  RB_DECLARE_CMD_RESP_BUFFER( rb_msg, rb_dsd_record_t, rb_dsd_status_t );

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockWriteDSD: ");
  #endif

  /* Setup to write the record */
/*  ezi_dbg_raw(dbuf, dsdRecord, sizeof(rb_dsd_record_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'B', dbuf->str); */
  RinBlockDsdCrcCalc(dsdRecord, sizeof(rb_dsd_record_t));
/*  ezi_dbg_raw(dbuf, dsdRecord, sizeof(rb_dsd_record_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'A', dbuf->str); */
  rb_msg->data.rb_command.params.rb_dsd_record = *dsdRecord;  /* TODO endian swap */

  /* Do the write, and get response and handle rinblock not responding */
  if (!RinBlockExecuteCommand( rb, rb_msg,
                               RB_GET_SIZE_COMMAND( rb_dsd_record_t ),
                               RB_GET_SIZE_COMMAND( rb_dsd_status_t ),
                               RB_TIMEOUT_DSD_WRITE, RB_CMD_DSD_WRITE_RECORD ) )
  {
    #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no reply, failed\n");
    #endif
    rbDsdLastFail = __LINE__;
    return 0;
  }

  /* TODO Endian swap */
  *dsdRecord = rb_msg->data.rb_response.params.rb_dsd_record; /* may only be rb_dsd_status, so check the 'status' value */

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)                /* Read data success */
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success\n");
  #endif
  return 1;
}


int8u RinBlockReadDSD(struct t_rinblock *rb, DsdRead_t dsdReadType,
                      rb_dsd_read_cmd_t *dsdReadCmd, rb_dsd_record_t *dsdRecord)
{
  RB_DECLARE_CMD_RESP_BUFFER( rb_msg, rb_dsd_read_cmd_t, rb_dsd_record_t );
  int8u cmd;

/*  ezi_dbg_raw(dbuf, dsdReadCmd, sizeof(rb_dsd_read_cmd_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'B', dbuf->str); */
  RinBlockDsdCrcCalc(dsdReadCmd, sizeof(rb_dsd_read_cmd_t));
/*  ezi_dbg_raw(dbuf, dsdReadCmd, sizeof(rb_dsd_read_cmd_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'A', dbuf->str); */

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockReadDSD: ");
  #endif

  /* Setup to perform the selected type of read command. dsdRdOldest & dsdRdNewest don't need dsdReadCmd. */
  switch ( dsdReadType )
  {
    case dsdRdRecord:   cmd = RB_CMD_DSD_READ_RECORD;         break;
    case dsdRdNext:     cmd = RB_CMD_DSD_READ_RECORD_NEXT;    break;
    case dsdRdPrevious: cmd = RB_CMD_DSD_READ_RECORD_PREV;    break;
    case dsdRdOldest:   cmd = RB_CMD_DSD_READ_RECORD_OLDEST;  break;
    case dsdRdNewest:   cmd = RB_CMD_DSD_READ_RECORD_NEWEST;  break;
    default:
      rbDsdLastFail = __LINE__;
      return 0;
      break;
  }
  rb_msg->data.rb_command.params.rb_dsd_read_cmd = *dsdReadCmd; /* TODO swap endian */

  /* Do the read, get response, and handle rinblock not responding */
  if (!RinBlockExecuteCommand( rb, rb_msg,
                               RB_GET_SIZE_COMMAND( rb_dsd_read_cmd_t ),
                               RB_GET_SIZE_COMMAND( rb_dsd_record_t ),
                               RB_TIMEOUT_DSD_READ, cmd ) )
  {
    #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no reply, failed\n");
    #endif
    rbDsdLastFail = __LINE__;
    return 0;
  }

  *dsdRecord = rb_msg->data.rb_response.params.rb_dsd_record; /* TODO Swap endian */

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)                /* Read data success */
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success\n");
  #endif
  return 1;
}


int8u RinBlockFreeUpDSD(struct t_rinblock *rb, rb_dsd_free_up_t *freeUp)
{
  RB_DECLARE_CMD_BUFFER( rb_msg, rb_dsd_free_up_t );

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockFreeUpDSD: ");
  #endif

/*  ezi_dbg_raw(dbuf, freeUp, sizeof(rb_dsd_free_up_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'B', dbuf->str); */
  RinBlockDsdCrcCalc(freeUp, sizeof(rb_dsd_free_up_t));
/*  ezi_dbg_raw(dbuf, freeUp, sizeof(rb_dsd_free_up_t));  rfprintf(&dev_com3, "%c:%s\r\n", 'A', dbuf->str); */
  rb_msg->data.rb_command.params.rb_dsd_free_up = *freeUp; /* copies whole structure */ /* TODO swap endian */

  /* Do the command, get response, and handle rinblock not responding */
  if (!RinBlockExecuteCommand( rb, rb_msg,
                               RB_GET_SIZE_COMMAND( rb_dsd_free_up_t ),
                               RB_GET_SIZE_COMMAND( rb_dsd_free_up_t ),
                               RB_TIMEOUT_DSD_READ, RB_CMD_DSD_FREE_UP ) )

  {
    #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no reply, failed\n");
    #endif
    rbDsdLastFail = __LINE__;
    return 0;
  }

  *freeUp = rb_msg->data.rb_response.params.rb_dsd_free_up; /* copies whole structure */ /* TODO swap endian */

  #if RB_DEBUG(RB_DSD_DEBUG_LVL, RB_DEBUG_BASIC)                /* Read data success */
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success\n");
  #endif
  return 1;
}
