#include <rinhw/rio.h>
#include <rinhw/sio.h>
#include <rinhw/rstdio.h>
#include <rinhw/wdt.h>

#include <rinblock/rb_startup.h>
#include <rinblock/rb_flash.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_command.h>
#include <rinblock/rb_debug.h>


/*
 * The string that gets echoed to the master as part of the connection process
 */
static const unsigned char run_flash_str[] = "FlAsHRuN";


/*
 * RinBlockWatchdogKick -
 *
 * Kick a watchdog while programming, only if wdt is not null
 */
static void RinBlockWatchdogKick(struct t_wdt *wdt)
{
    if (wdt != (struct t_wdt *) NULL)
        WatchdogPreset(wdt, RB_TIMEOUT_FLASH_WDT_LONG);
}


/*
 * RinBlockProgrammingWrite -
 *
 * Write a byte to the rin block programming stream
 */
static unsigned char RinBlockProgrammingWrite(struct t_rinblock *rb, unsigned char dat)
{
    RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_general_boot_stream );
    unsigned char res;

    rb_msg->data.rb_command.header.cmd = RB_CMD_WRITE_BOOT_STREAM;
    rb_msg->data.rb_command.params.rb_general_boot_stream.data = dat;
    res = RinBlockInternalExecuteCommand(rb, rb_msg,
                                         RB_GET_SIZE_COMMAND( struct s_rb_general_boot_stream ),
                                         RB_GET_SIZE_CMD_HDR(),
                                         RB_TIMEOUT_FLASH_LONG, 0);

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RB_CMD_WRITE_BOOT_STREAM (data 0x%02x) returned with res %d, response 0x%02x\n",
                    dat, res, rb_msg->data.rb_response.header.resp);
#endif
    return ((res) && ((rb_msg->data.rb_response.header.resp & RB_RESP_ERROR_MASK) == 0));
}


/*
 * RinBlockProgrammingRead -
 *
 * Read a byte from the rin block programming stream
 */
static unsigned char RinBlockProgrammingRead(struct t_rinblock *rb, unsigned char *dat)
{
    RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_general_boot_stream );
    unsigned char result, res;

    rb_msg->data.rb_command.header.cmd = RB_CMD_READ_BOOT_STREAM;
    res = RinBlockInternalExecuteCommand(rb, rb_msg,
                                         RB_GET_SIZE_CMD_HDR(),
                                         RB_GET_SIZE_COMMAND( struct s_rb_general_boot_stream ),
                                         RB_TIMEOUT_FLASH_LONG, 0);

    result = ((res) && ((rb_msg->data.rb_response.header.resp & RB_RESP_ERROR_MASK) == 0));
    if (result)
        *dat = rb_msg->data.rb_command.params.rb_general_boot_stream.data;

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RB_CMD_READ_BOOT_STREAM returned with result %d, response 0x%02x, data 0x%02x\n",
                    res, rb_msg->data.rb_response.header.resp, *dat);
#endif

    return result;
}


/*
 * RinBlockProgrammingFinished -
 *
 * Check whether the rin block has finished programming
 */
static unsigned char RinBlockProgrammingFinished(struct t_rinblock *rb, unsigned char *finished)
{
    unsigned char mode;

    if (RinBlockGetMode(rb, &mode) == 0)
        return 0;
    *finished = (unsigned char) (mode == RB_MODE_PROG_DONE);
    return 1;
}


#if 0
static unsigned char SerialGetChar(struct t_io_dev *serial_dev)
{
    unsigned char ch;

    rread(serial_dev, &ch, 1, RB_FLASH_TIMEOUT_SHORT);
    return ch;
}
#endif


static void SerialPutChar(struct t_io_dev *serial_dev, unsigned char ch)
{
    rwrite(serial_dev, &ch, 1, 0);
}


/*
 * GetBaudAsByte -
 *
 * Return the current baud rate of this serial_dev scaled to fit in a byte
 */
static unsigned char GetBaudAsByte(struct t_io_dev *serial_dev)
{
    t_ioctl_siocfg siocfg;
    unsigned char brb;

    /*
     * Get the serial configuration
     */
    rioctl(serial_dev, IOCTL_SER_GET_CONFIG, &siocfg);

    /*
     * Scale the baud rate so it fits in a byte
     */
    brb = (unsigned char) (unsigned long) (115200UL / siocfg.baud);

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Baud rate is %ld, and scaled to fit byte 0x%02x\n", siocfg.baud, brb);
#endif
    return brb;
}


/*
 * SetBaudFromByte -
 *
 * Set the baud rate of this serial_dev by scaling up the brb
 */
static void SetBaudFromByte(struct t_io_dev *serial_dev, unsigned char brb)
{
    t_ioctl_siocfg siocfg;

    /*
     * Get current serial configuration
     */
    rioctl(serial_dev, IOCTL_SER_GET_CONFIG, &siocfg);

    /*
     * Scale the baud rate up
     */
    siocfg.baud = (unsigned long) (115200UL / brb);

    /*
     * Set the serial port to this new configuration
     */
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "New baud rate %ld, based on baud rate byte of 0x%02x\n", siocfg.baud, brb);
#endif
    rioctl(serial_dev, IOCTL_SER_SET_CONFIG, &siocfg);
}


#define MAX_ACKLEN 10

/*
 * ReBaudComm -
 *
 * Adjust the baud rate on the serial port to what the PC asks for
 */
static unsigned char ReBaudComm(struct t_io_dev *serial_dev)
{
    unsigned char ack[MAX_ACKLEN];
    unsigned char acklen, ch;

    /*
     * Get 1 char and discard (just part of protocol)
     */
    if (rread(serial_dev, &ch, 1, RB_TIMEOUT_FLASH_LONG) != 1)
        return 0;

    /*
     * Get the acknowldegment length now,
     * and fail if too big for our buffer
     */
    if (rread(serial_dev, &acklen, 1, RB_TIMEOUT_FLASH_LONG) != 1)
        return 0;
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Acknowledge length = %d\n", acklen);
#endif
    if (acklen > MAX_ACKLEN)
        return 0;

    /*
     * Convert the baudrate to a byte and send it
     */
    SerialPutChar(serial_dev, GetBaudAsByte(serial_dev));

    /*
     * Read the new baudrate byte from the serial port
     * and set it
     */
    if (rread(serial_dev, &ch, 1, RB_TIMEOUT_FLASH_LONG) != 1)
        return 0;
    SetBaudFromByte(serial_dev, ch);

    /*
     * Read the acknowledgement data, and echo it
     * to check the re-bauding has worked
     */
    if (rread(serial_dev, ack, (unsigned short) acklen, RB_TIMEOUT_FLASH_LONG) != acklen)
    {
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Read ackdata failed\n");
#endif
        return 0;
    }
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Read ackdata: ");
    rwrite(RB_DEBUG_DEV, ack, (unsigned short) acklen, 0);
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "\n");
#endif
    rwrite(serial_dev, ack, (unsigned short) acklen, RB_TIMEOUT_FLASH_LONG);
    return 1;
}


#if 0
/*
 * RinBlockSerialRelay -
 *
 * Echo characters between the serial port and the Rin block boot stream
 *
 * This implementation will only relay 1 character before checking the alternate end for a
 * character, which results in longer programming time, however will allow fully asynchronous
 * communications, which is not really necessary for progamming the bootloader, as it uses
 * a command/response type system.
 */
static void RinBlockSerialRelay(struct t_rinblock *rb, struct t_io_dev *serial_dev)
{
    unsigned char ch;

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Entering character relay mode\n");
#endif

    /*
     * TODO some escape from this when done?
     */
    while (1)
    {

        /*
         * If there are characters in the serial buffer,
         * then read 1 of them and write it to the rin block boot stream
         */
        if (rfilelength(serial_dev) > 0)
        {
            rread(serial_dev, &ch, 1, 0);

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Serial->Rinblock 0x%02x\n", ch);
#endif

            RinBlockProgrammingWrite(rb, ch);
        }

        /*
         * If the rin block boot stream has a read a character
         * then write it to the serial port
         */
        if (RinBlockProgrammingRead(rb, &ch))
        {

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Rinblock->Serial 0x%02x\n", ch);
#endif
            rwrite(serial_dev, &ch, 1, 0);
        }
    }
}
#endif


/*
 * RinBlockSerialRelay -
 *
 * Echo characters between the serial port and the Rin block boot stream
 *
 * This implementation will relay all buffered characters from one end before checking the alternate
 * end for characters, which speeds up programming, as less checking is required, however will only
 * work with half duplex communications which is suitable for the bootloader, as it uses a
 * command/response type system.
 */
static unsigned char RinBlockSerialRelay(struct t_rinblock *rb, struct t_io_dev *serial_dev, struct t_wdt *wdt)
{
    unsigned char ch, finished;
    unsigned long time = OSTimeGet();

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Entering character relay mode\n");
#endif

    while (1)
    {

        /*
         * Kick the watchdog
         */
        RinBlockWatchdogKick(wdt);

        /*
         * If there are characters in the serial buffer,
         * then read 1 of them and write it to the rin block boot stream
         */
        while (rfilelength(serial_dev) > 0)
        {
            rread(serial_dev, &ch, 1, 0);
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Serial->Rinblock 0x%02x\n", ch);
#endif
            if (!RinBlockProgrammingWrite(rb, ch))
            {
                return 0;
            }
        }

        /*
         * If the rin block boot stream has a read a character
         * then write it to the serial port
         */
        while (RinBlockProgrammingRead(rb, &ch))
        {

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Rinblock->Serial 0x%02x\n", ch);
#endif
            rwrite(serial_dev, &ch, 1, 0);
        }

        /*
         * Once a second, so as not to slow it down too much,
         * ask the module if programming is finished
         */
        if (OSTimeGet() >= (time + OS_TICKS_PER_SEC))
        {
            time = OSTimeGet();
            if (RinBlockProgrammingFinished(rb, &finished) == 0)
                return 0;
            if (finished)
                return 1;
        }
    }
}


/*
 * RinBlockProgram -
 *
 * Program a rin block by echoing characters from the serial port serial_dev to the
 * rin block rb, and visa versa, also toggles the watchdog wdt
 * Errors are returned in err
 *
 */
unsigned char RinBlockProgram(struct t_rinblock *rb, struct t_io_dev *serial_dev, struct t_wdt *wdt, unsigned char *err)
{
    t_ioctl_siocfg siocfg;

#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Program rinblock at address %d\n", rb->iic_address);
#endif

    /*
     * Check the rinblock is actually present on the bus
     */
    if (rb->present == 0)
    {
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "No rinblock present at address %d\n", rb->iic_address);
#endif
        *err = RB_ERR_NOT_PRESENT;
        return 0;
    }

    /*
     * Set the mode of the rin block to programming
     */
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Setting rinblock mode for programming\n");
#endif
    RinBlockWatchdogKick(wdt);
    if (RinBlockSetMode(rb, RB_MODE_PROGRAM) == 0)
    {
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Set mode failed\n");
#endif
        *err = RB_ERR_PROGRAM_MODE;
        return 0;
    }

    /*
     * Save the serial configuration,
     * any failures from here on should goto RBPFail, so
     * the serial config can be restored
     */
    rioctl(serial_dev, IOCTL_SER_GET_CONFIG, &siocfg);

    /*
     * Echo the connect string to the master,
     * and re-baud the serial port
     */
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Echoing connect string, and re-bauding the serial port\n");
#endif
    RinBlockWatchdogKick(wdt);
    rfprintf(serial_dev, (char *) run_flash_str);
    if (ReBaudComm(serial_dev) == 0)
    {
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Re-bauding the serial port failed\n");
#endif
        *err = RB_ERR_REBAUD;
        goto RBPFail;
    }

    /*
     * Relay characters between the serial port and the rinblock
     */
	if (RinBlockSerialRelay(rb, serial_dev, wdt) == 0)
    {
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Programming failed\n");
#endif
        *err = RB_ERR_FLASH_PROGRAM_FAIL;
        goto RBPFail;
    }

    /*
     * Restart the rinblock
     */
#if RB_DEBUG(RB_FLASH_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Restarting the rinblock");
#endif
    RinBlockWatchdogKick(wdt);
    RinBlockRestart(rb);
    *err = RB_ERR_NO_ERR;
    rioctl(serial_dev, IOCTL_SER_SET_CONFIG, &siocfg);
    return 1;

RBPFail:
    rioctl(serial_dev, IOCTL_SER_SET_CONFIG, &siocfg);
    return 0;
}
