#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <rinhw/rio.h>
#include <rinhw/sio.h>
#include <rinhw/byteq.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_command.h>

#include <rinblock/rb_serial.h>


#define LOCAL_RX_BUFFER_SIZE 256


/** Device data for rinblock serial port channel */
struct s_rb_serial_channel
{
  enum e_rb_serial_channel channel;   /*!< Rinblock channel (A or B) */
  struct t_rinblock *rinblock;        /*!< Rinblock */
  uint32_t rxtx_byte_time_ns;         /*!< Nanoseconds to send a byte at current baud/bit settings */
  struct s_byteq rx_queue;            /*!< Local queue of received data, used to reduce the number of bus transactions */
  uint8_t *rx_buffer;                 /*!< Buffer used by rx_queue, don't access directly */
};


/** Get the size/data level of one of the buffers */
static unsigned char rb_sio_get_buf_level(struct t_rinblock *rb, unsigned char cmd, unsigned short *len)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_serial_stream_usage);
  unsigned char result;

  /* Execute command */
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_get_buf_level called, checking buffer level with command 0x%02x\n", cmd);
#endif
  result = RinBlockExecuteCommand(rb, rb_msg,
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_GET_SIZE_COMMAND(struct s_rb_serial_stream_usage),
                                  RB_TIMEOUT_DEFAULT, cmd);
  if (result)
  {
    /* Success, set the bufsize, return success */
    *len = (unsigned short) rb_msg->data.rb_response.params.rb_serial_stream_usage.bytes_used;
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Checking level success, buf level = %d\n", *len);
#endif
  }
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "failure\n");
  }
#endif

  return result;
}


/** Calculate receive/transmit byte delay time in nanoseconds */
static uint32_t rb_sio_calc_rxtx_byte_time(const t_ioctl_siocfg *siocfg)
{
  uint32_t bit_ns;
  unsigned int bits;
  uint32_t byte_ns;

  bit_ns = 1000000000 / siocfg->baud;
  bits = siocfg->data_bits + siocfg->stop_bits + 1; /* +1 = Start bit */
  bits += (siocfg->parity_bits == SIO_PARITY_NONE) ? 0 : 1;
  byte_ns = bit_ns * bits;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_calc_rxtx_byte_time: byte time is %luns\n", (unsigned long) byte_ns);
#endif
  return byte_ns;
}


/**
 * Wait until there are at least 1 character in the module rx buffer /
 * 1 space in the module tx buffer (specified by command), or a timeout
 */
static unsigned char WaitBufSpace(struct s_rb_serial_channel *rbsc, unsigned char cmd, unsigned short *space, unsigned short timeout)
{
  unsigned long finish_time = OSTimeGet() + timeout;
  struct timespec delay;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "WaitBufSpace called, waiting for space/characters in module buffer\n");
#endif

  /* Loop until space/characters in buffer or timeout */
  do
  {
    /* Check how many characters/space is left in the buffer */
    if (rb_sio_get_buf_level(rbsc->rinblock, cmd, space) == 0)
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_get_buf_level failed\n");
#endif
      return 0;
    }

    /*
     * Yield if no characters,
     * and check for timeout
     */
    if (*space == 0)
    {
      delay.tv_nsec = rbsc->rxtx_byte_time_ns;
      delay.tv_sec = 0;
      nanosleep(&delay, NULL);

      /* Have we timed out yet */
      if ((timeout > 0) && (OSTimeGet() > finish_time))
      {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "WaitBufSpace timed out waiting for space/characters\n");
#endif
        return 0;
      }
    }
  }
  while (*space == 0);

  /* Success */
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "WaitBufSpace success, found %u space/characters in module buffer\n", *space);
#endif
  return 1;
}


/** Read or write data to a rinblock */
static unsigned char ReadWriteSerialRB(struct t_rinblock *rb, unsigned char cmd, unsigned char *buf, unsigned short len, unsigned short timeout)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_serial_stream);
  bool reading = (cmd == RB_CMD_READ_STREAM1);
  unsigned short read_len, write_len;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "ReadWriteSerialRB: called\n");
#endif

  /* Setup the command, read or write */
  if (reading)
  {
    rb_msg->data.rb_command.params.rb_serial_stream_read.length = (unsigned char) len;
    read_len = (unsigned short) RB_GET_SIZE_CMD_HDR() + len;
    write_len = RB_GET_SIZE_COMMAND(struct s_rb_serial_stream_read);
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "ReadWriteSerialRB: module read, read_len = %d, write_len = %d\n", read_len, write_len);
#endif
  }
  else
  {
    memcpy(rb_msg->data.rb_command.params.rb_serial_stream.data, buf, (size_t) len);
    read_len = RB_GET_SIZE_CMD_HDR();
    write_len = (unsigned short) RB_GET_SIZE_CMD_HDR() + len;
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "ReadWriteSerialRB: module read, read_len = %d, write_len = %d. Write data dump:\n", read_len, write_len);
    RBDump(rb_msg->data.rb_response.params.rb_serial_stream.data, len);
#endif
  }

  /* Run the command */
  if (!RinBlockExecuteCommand(rb, rb_msg, write_len, read_len, timeout, cmd))
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "ReadWriteSerialRB: module didn't reply\n");
#endif
    return 0;
  }

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "ReadWriteSerialRB: success\n");
#endif

  /* Copy data into the callers receive buffer */
  if (reading)
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "ReadWriteSerialRB: read data dump:\n");
    RBDump(rb_msg->data.rb_response.params.rb_serial_stream.data, len);
#endif
    memcpy(buf, rb_msg->data.rb_response.params.rb_serial_stream.data, (size_t) len);
  }
  return 1;
}


/** Check that is rinblock is a serial module of some kind */
static bool rb_sio_isserial(struct t_rinblock *rb)
{
  /* Check that the rinblock is actually a serial one */
  return (RB_IFACE_TYPE_SERIAL == rb->tag.interface_type ||
          RB_IFACE_TYPE_ETHERNET == rb->tag.interface_type ||
          RB_IFACE_TYPE_C500_DISP == rb->tag.interface_type);
}


static int rb_sio_config_set(struct s_rb_serial_channel *rbsc, const t_ioctl_siocfg *siocfg_ptr)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_serial_configure);
  unsigned char result;
  unsigned char cmd;

  rbsc->rxtx_byte_time_ns = rb_sio_calc_rxtx_byte_time(siocfg_ptr);

  /* Setup command */
  cmd = (rbsc->channel == rbsc_a) ? RB_CMD_WRITE_CONFIG_SERIAL1 : RB_CMD_WRITE_CONFIG_SERIAL2;

  /* Because of struct packing for the IIC, we must transfer all the parameters over */
  rb_msg->data.rb_command.params.rb_serial_configure.baud = htonl(siocfg_ptr->baud);
  rb_msg->data.rb_command.params.rb_serial_configure.data_bits = siocfg_ptr->data_bits;
  rb_msg->data.rb_command.params.rb_serial_configure.parity_bits = siocfg_ptr->parity_bits;
  rb_msg->data.rb_command.params.rb_serial_configure.stop_bits = siocfg_ptr->stop_bits;
  rb_msg->data.rb_command.params.rb_serial_configure.use_cts = siocfg_ptr->use_cts;
  rb_msg->data.rb_command.params.rb_serial_configure.use_terminations = siocfg_ptr->use_terminations;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "cfg baud %ld, msg baud %ld\n", siocfg_ptr->baud, rb_msg->data.rb_command.params.rb_serial_configure.baud);
#endif

  /* Write config to rinblock */
  result = RinBlockExecuteCommand(rbsc->rinblock, rb_msg,
                                  RB_GET_SIZE_COMMAND(struct s_rb_serial_configure),
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_TIMEOUT_INDEFINITE, cmd);
  return (result) ? RIO_SUCCESS : RIO_ERROR;
}


static int rb_sio_config_get(struct s_rb_serial_channel *rbsc, t_ioctl_siocfg *siocfg_ptr)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_serial_configure);
  unsigned char result;
  unsigned char cmd;

  cmd = (rbsc_a == rbsc->channel) ? RB_CMD_READ_CONFIG_SERIAL1 : RB_CMD_READ_CONFIG_SERIAL2;
  result = RinBlockExecuteCommand(rbsc->rinblock, rb_msg,
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_GET_SIZE_COMMAND(struct s_rb_serial_configure),
                                  RB_TIMEOUT_INDEFINITE, cmd);
  if (result)
  {
    /* Because of struct packing for the IIC, we must transfer all the parameters over */
    siocfg_ptr->baud = ntohl(rb_msg->data.rb_response.params.rb_serial_configure.baud);
    siocfg_ptr->data_bits = rb_msg->data.rb_response.params.rb_serial_configure.data_bits;
    siocfg_ptr->parity_bits = rb_msg->data.rb_response.params.rb_serial_configure.parity_bits;
    siocfg_ptr->stop_bits = rb_msg->data.rb_response.params.rb_serial_configure.stop_bits;
    siocfg_ptr->use_cts = rb_msg->data.rb_response.params.rb_serial_configure.use_cts;
    siocfg_ptr->use_terminations = rb_msg->data.rb_response.params.rb_serial_configure.use_terminations;

    rbsc->rxtx_byte_time_ns = rb_sio_calc_rxtx_byte_time(siocfg_ptr);

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_config_get: read config success\n");
#endif
  }
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_config_get: read config failure\n");
  }
#endif

  return (result) ? RIO_SUCCESS : RIO_ERROR;
}


/** Default sio config to be loaded into modules when initialised */
static const t_ioctl_siocfg sio_def_cfg =
{
  SIO_DEF_BAUD,           /* unsigned long baud; */
  SIO_DEF_DATA_BITS,      /* unsigned char data_bits; */
  SIO_DEF_PARITY_BITS,    /* unsigned char parity_bits; */
  SIO_DEF_STOP_BITS,      /* unsigned char stop_bits; */
  SIO_DEF_USECTS,         /* unsigned char use_cts; */
  SIO_DEF_TERMINATIONS    /* unsigned char use_terminations; */
};


/** Device initialisation function */
static int rb_sio_init(struct t_io_dev *io_dev)
{
  struct s_rb_serial_channel *rbsc = (struct s_rb_serial_channel *) io_dev->dev_data;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_init called\n");
#endif

  /* Setup the serial port with a default configuration */
  rbsc->rxtx_byte_time_ns = rb_sio_calc_rxtx_byte_time(&sio_def_cfg);
  if (rb_sio_config_set(rbsc, &sio_def_cfg) < 0)
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_init: setting default configuration failed\n");
#endif
    return RIO_ERROR;
  }

  /* Its a character device */
  io_dev->status |= O_DEVCHAR;

  return RIO_SUCCESS;
}


/** Device destroy function */
static int rb_sio_destroy(struct t_io_dev *io_dev)
{
  struct s_rb_serial_channel *rbsc = (struct s_rb_serial_channel *) io_dev->dev_data;
  int result;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_destroy called\n");
#endif

  /* If open, then close */
  if (0 == (io_dev->status & O_OPEN))
  {
    result = io_dev->dev_close(io_dev);
  }
  else
  {
    result = RIO_SUCCESS;
  }

  /* Free up stuff we've allocated */
  io_dev->dev_data = NULL;
  if (NULL != rbsc)
  {
    if (NULL != rbsc->rx_buffer)
    {
      ByteQDestroy(&rbsc->rx_queue);
      free(rbsc->rx_buffer);
    }
    free(rbsc);
    rbsc = NULL;
  }
  return result;
}


/** Open the rinblock serial device */
static int rb_sio_open(struct t_io_dev *io_dev, unsigned short timeout)
{
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_open called\n");
#endif
  return RIO_SUCCESS;
}


/** Close the rinblock serial device */
static int rb_sio_close(struct t_io_dev *io_dev)
{
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_close called\n");
#endif
  return RIO_SUCCESS;
}


/**
 * Calculate how much time is left before timeout.
 *
 * Returns 0 if timed out, 1 otherwise.
 * time_left is only set if some time remains before timeout
 */
static unsigned char CalculateTimeout(unsigned long finish_time, unsigned short timeout, unsigned short *time_left)
{
  unsigned long time_now;

  /* Timeout of 0 is a special case (never time out) */
  if (timeout > 0)
  {
    /*
     * Still not timed out,
     * so calculate how much time is left,
     *
     * Otherwise return 0
     */
    time_now = OSTimeGet();
    if (time_now < finish_time)
    {
      *time_left = (unsigned short) (finish_time - time_now);
    }
    else
    {
      return 0;
    }
  }
  else
  {
    *time_left = 0;
  }
  return 1;
}


/** Do the actual read from the module, into the local receive queue */
static bool rb_sio_read_module(struct t_io_dev *io_dev, size_t data_avail, size_t time_left)
{
  struct s_rb_serial_channel *rbsc = (struct s_rb_serial_channel *) io_dev->dev_data;
  uint8_t data[MAX_RB_SERIAL_DATA_LENGTH];
  size_t read_size;
  size_t data_read;
  size_t i;
  size_t rx_queue_space;

  /* Limit the size of the read to the amount of space we have in the queue */
  rx_queue_space = ByteQGetSpace(&rbsc->rx_queue);
  if (data_avail > rx_queue_space)
    data_avail = rx_queue_space;

  /*
   * Optimisation: Read as much as we can here.
   * We know the number of bytes available, and how many we still need to read.
   * So loop reading blocks of MAX_RB_SERIAL_DATA_LENGTH until
   * we've read all we need, or we need to wait for more data
   *
   * This is so we don't have to check the buffer level for every read
   */
  for (data_read = 0, read_size = 0; data_read < data_avail; data_read += read_size)
  {
    /* Figure out how many bytes to receive, up to a max of MAX_RB_SERIAL_DATA_LENGTH */
    read_size = (data_avail - data_read < MAX_RB_SERIAL_DATA_LENGTH) ? data_avail - data_read : MAX_RB_SERIAL_DATA_LENGTH;
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read_module(io_dev=%p, ...): avail: %u, read: %u, reading: %u\n", io_dev, data_avail, data_read, read_size);
#endif

    /* Read the data from the rinblock */
    if (!ReadWriteSerialRB(rbsc->rinblock, RB_CMD_READ_STREAM1, data, read_size, time_left))
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read_module(io_dev=%p, ...): error reading data from rinblock\n", io_dev);
#endif
      return false;
    }

    /* Put the data we read from the module into the receive queue */
    for (i = 0; i < read_size; i++)
      ByteQWrite(&rbsc->rx_queue, data[i], NULL);
  }
  return true;
}


/** Read len characters from the rinblock serial device into buffer buf, with timeout */
static int rb_sio_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_rb_serial_channel *rbsc = (struct s_rb_serial_channel *) io_dev->dev_data;
  unsigned char *ptr = (unsigned char *) buf;
  unsigned short module_avail;
  unsigned short time_left;
  unsigned short count = len;
  unsigned long finish_time = OSTimeGet() + timeout;
  uint8_t data;
  enum eBQResult bqerr;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read(io_dev=%p, buf=%p, len=%u, timeout=%u) called\n", io_dev, buf, len, timeout);
#endif

  /* Check that the mapped rinblock is actually a serial one */
  if (!rb_sio_isserial(rbsc->rinblock))
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read(io_dev=%p, ...): configuration error: attempted read from non-serial rinblock\n", io_dev);
#endif
    return RIO_ERROR;
  }

  /* No receiver on serial channel 1, so return error */
  if (rbsc->channel != rbsc_a)
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read(io_dev=%p, ...): error: attempted read from transmit only serial channel\n", io_dev);
#endif
    return RIO_ERROR;
  }

  /* Retrieve the required number of bytes of data */
  while (count > 0)
  {
    /* Use up as much of the local receive buffer as we can to satisfy the request first */
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read(io_dev=%p, ...): local buffer contains %u bytes, need %u bytes\n", io_dev, (unsigned) ByteQGetLevel(&rbsc->rx_queue), (unsigned) count);
#endif
    while (count > 0 && ByteQGetLevel(&rbsc->rx_queue) > 0)
    {
      data = ByteQRead(&rbsc->rx_queue, NULL, &bqerr);
      if (bqrEmpty == bqerr)
        break;
      else if (bqrError == bqerr)
        return RIO_ERROR;
      *ptr++ = data;
      count--;
    }
    if (count <= 0)
      goto rsr_done;

    /* Calculate how much time is left before timeout */
    if (CalculateTimeout(finish_time, timeout, &time_left) == 0)
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read(io_dev=%p, ...): error: timeout\n", io_dev);
#endif
      return RIO_ERROR;
    }

    /* Wait until there are characters in the module rx buffer */
    module_avail = 0;
    if (!WaitBufSpace(rbsc, RB_CMD_READ_RX_USED, &module_avail, time_left))
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read(io_dev=%p, ...): error: timeout/rinblock error waiting for characters\n", io_dev);
#endif
      return RIO_ERROR;
    }

    if (!rb_sio_read_module(io_dev, module_avail, time_left))
    {
      return RIO_ERROR;
    }
  }

  /* All done, success */
rsr_done:
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_read(io_dev=%p, ...): success, read %d bytes\n", io_dev, (int) len);
#endif
  return (int) len;
}


/** Serial Rinblock comms channel device write */
static int rb_sio_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
  struct s_rb_serial_channel *rbsc = (struct s_rb_serial_channel *) io_dev->dev_data;
  unsigned char *ptr = (unsigned char *) buf;
  unsigned short send_count;
  unsigned short buf_free;
  unsigned short time_left;
  unsigned short count = len;
  unsigned short buf_sent;
  unsigned long finish_time = OSTimeGet() + timeout;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_write called\n");
#endif

  /* Check that the mapped rinblock is actually a serial one */
  if (!rb_sio_isserial(rbsc->rinblock))
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_write: configuration error: attempted write to non-serial rinblock\n");
#endif
    return RIO_ERROR;
  }

  /* Write the required number of bytes of data */
  while (count > 0)
  {
    /* Calculate how much time is left before timeout */
    if (CalculateTimeout(finish_time, timeout, &time_left) == 0)
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_write: error: timeout\n");
#endif
      return RIO_ERROR;
    }

    /* Wait until there is space in the module tx buffer */
    if (!WaitBufSpace(rbsc, ((rbsc->channel == rbsc_a) ? RB_CMD_READ_TX1_FREE : RB_CMD_READ_TX2_FREE), &buf_free, timeout))
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_write: error: timeout/rinblock error waiting for buffer space\n");
#endif
      return RIO_ERROR;
    }

    /*
     * Optimisation: Write as much as we can here.
     * We know the number of bytes free in the output buffer, and how many we still need to write.
     * So loop writing blocks of MAX_RB_SERIAL_DATA_LENGTH until
     * we've written all we need, or we need to wait for more space in the output buffer
     *
     * This is so we don't have to check the buffer level for every write
     */
    buf_sent = 0;
    do
    {
      /* Write as many bytes as we can, upto a max of MAX_RB_SERIAL_DATA_LENGTH */
      send_count = (buf_free - buf_sent < MAX_RB_SERIAL_DATA_LENGTH) ? buf_free - buf_sent : MAX_RB_SERIAL_DATA_LENGTH;
      if (send_count > count)
        send_count = count;
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_write: free: %u, written: %u, writing: %u\n", buf_free, buf_sent, send_count);
#endif

      /* Write the data to the rinblock */
      if (!ReadWriteSerialRB(rbsc->rinblock, ((rbsc->channel == rbsc_a) ? RB_CMD_WRITE_STREAM1 : RB_CMD_WRITE_STREAM2), ptr, send_count, time_left))
      {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_write: error writing data from rinblock\n");
#endif
        return RIO_ERROR;
      }

      /* Update total number of bytes sent */
      ptr += send_count;
      count -= send_count;
      buf_sent += send_count;
    }
    while (count > 0 && buf_free - buf_sent > 0);
  }

  /* All done, success */
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_write: success, wrote %d bytes\n", len);
#endif
  return (int) len;
}


static long rb_sio_filelength(struct t_io_dev *io_dev)
{
  struct s_rb_serial_channel *rbsc = (struct s_rb_serial_channel *) io_dev->dev_data;
  unsigned short buf_used;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_filelength called\n");
#endif

  /* No receiver on serial channel 1, so return error */
  if (rbsc_a != rbsc->channel)
    return RIO_ERROR;

  /* Request how many bytes available in the receive buffer */
  if (rb_sio_get_buf_level(rbsc->rinblock, RB_CMD_READ_RX_USED, &buf_used) == 0)
    return RIO_ERROR;

  return (long) buf_used;
}


static unsigned char rb_sio_get_tx_buf_size(struct s_rb_serial_channel *rbsc, unsigned short *size)
{
  /* Request the size of the transmit buffer */
  return rb_sio_get_buf_level(rbsc->rinblock, (rbsc->channel == rbsc_a) ? RB_CMD_READ_TX1_BUF_SIZE : RB_CMD_READ_TX2_BUF_SIZE, size);
}


static int rb_sio_get_tx_buf_lvl(struct s_rb_serial_channel *rbsc, unsigned short *lvl)
{
  unsigned short buf_sz, buf_free;
  unsigned char cmd;

  /* Request the size of the transmit buffer */
  if (rb_sio_get_tx_buf_size(rbsc, &buf_sz) == 0)
    return RIO_ERROR;

  /* Request how many bytes available in the transmit buffer */
  cmd = (rbsc->channel == rbsc_a) ? RB_CMD_READ_TX1_FREE : RB_CMD_READ_TX2_FREE;
  if (rb_sio_get_buf_level(rbsc->rinblock, cmd, &buf_free) == 0)
    return RIO_ERROR;

  /* Calculate buffer level */
  *lvl = buf_sz - buf_free;
  return RIO_SUCCESS;
}


static unsigned char rb_sio_get_cts_state(struct s_rb_serial_channel *rbsc, unsigned char *state)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_serial_cts_state);

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_get_cts_state called: ");
#endif

  /*
   * Since we can't discern the type of serial module,
   * we must ask the rinblock for the cts state,
   * and if it doesn't have one, then it returns 1.
   *
   * This code used to test the model, and not
   * talk to the module if it had no cts pin
   * but since the interface_type has been added
   * its no longer valid.
   */
#if 0
  /*
   * Deal with modules that don't have a cts pin
   */
  if ((rb->tag.model_number != RB_MODEL_RS232_232) &&
      (rb->tag.model_number != RB_MODEL_RS232_485))
  {
    /*
     * If its not a serial module at all then error,
     * otherwise, report that cts state = clear to send
     */
    if (rb->tag.model_number != RB_MODEL_RS485_485)
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "error: not a serial module\n");
#endif
      return 0;
    }
    else
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no cts pin on module, return 1 (clear to send)\n");
#endif
      *state = 1;
      return 1;
    }
  }
#endif

  /*
   * Check which port on the serial module,
   * cts pin only on channel 0
   */
  if (rbsc->channel != rbsc_a)
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no cts pin on this port, return 1 (clear to send)\n");
#endif
    *state = 1;
    return 1;
  }

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "this module/port has a cts pin, getting state from rinblock: ");
#endif

  /* Send command, get response */
  if (!RinBlockExecuteCommand(rbsc->rinblock, rb_msg,
                              RB_GET_SIZE_CMD_HDR(),
                              RB_GET_SIZE_COMMAND(struct s_rb_serial_cts_state),
                              RB_TIMEOUT_DEFAULT, RB_CMD_READ_TX1_CTS_STATE))
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no reply from rinblock\n");
#endif
    return 0;
  }

  /* Success */
  *state = rb_msg->data.rb_response.params.rb_serial_cts_state.state;
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success, cts state = %d\n", *state);
#endif
  return 1;
}


static int rb_sio_set_ring_network(struct s_rb_serial_channel *rbsc, unsigned char state)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_serial_ring_enable);

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_set_ring_network called: ");
#endif

  /* Check whether ring network is available on this port */
  if (rbsc->channel != rbsc_a)
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no ring network available on this port\n");
#endif
    return RIO_ERROR;
  }

  /* Setup the data in the struct */
  rb_msg->data.rb_command.params.rb_serial_ring_enable.state = state;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "enable ring network %d\n", state);
#endif

  /* Write config to rinblock */
  if (!RinBlockExecuteCommand(rbsc->rinblock, rb_msg,
                              RB_GET_SIZE_COMMAND(struct s_rb_serial_ring_enable),
                              RB_GET_SIZE_CMD_HDR(),
                              RB_TIMEOUT_DEFAULT, RB_CMD_SIO_WRITE_RING_ENABLE))
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_set_ring_network: write ring enable failed\n");
#endif
    return RIO_ERROR;
  }

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_set_ring_network: write ring enable success\n");
#endif
  return RIO_SUCCESS;
}


static int rb_sio_rx_flush(struct t_io_dev *io_dev)
{
  char buf[MAX_RB_SERIAL_DATA_LENGTH];
  struct s_rb_serial_channel *rbsc = (struct s_rb_serial_channel *) io_dev->dev_data;
  unsigned short level;
  unsigned short total_read;
  unsigned short to_read;
  int bytes_read;

  /* Get buffer level */
  if (rb_sio_get_buf_level(rbsc->rinblock, RB_CMD_READ_RX_USED, &level) == 0)
  {
    return RIO_ERROR;
  }

  /* Read that many bytes */
  for (total_read = 0; total_read < level; total_read += bytes_read)
  {
    to_read = (level - total_read > MAX_RB_SERIAL_DATA_LENGTH) ? MAX_RB_SERIAL_DATA_LENGTH : level - total_read;
    bytes_read = rb_sio_read(io_dev, buf, to_read, 0);
    if (bytes_read < 0)
    {
      return RIO_ERROR;
    }
  }
  return RIO_SUCCESS;
}


static int rb_sio_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_rb_serial_channel *rbsc = (struct s_rb_serial_channel *) io_dev->dev_data;
  unsigned short *word_ptr;
  int result = RIO_ERROR;
  unsigned char state;

#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_sio_ioctl called\n");
#endif

  switch (func)
  {
    case IOCTL_SER_SET_CONFIG:
      result = rb_sio_config_set(rbsc, va_arg(*arglist, t_ioctl_siocfg *));
      break;

    case IOCTL_SER_GET_CONFIG:
      result = rb_sio_config_get(rbsc, va_arg(*arglist, t_ioctl_siocfg *));
      break;

    case IOCTL_SER_GET_RX_BUF_LEVEL:
      /* No receiver on serial channel 1, buffer_level = 0, and return error */
      word_ptr = va_arg(*arglist, unsigned short *);
      if (rbsc->channel != rbsc_a)
      {
        *word_ptr = 0;
        goto rb_i_end;
      }

      /* Request how many bytes available in the receive buffer */
      if (rb_sio_get_buf_level(rbsc->rinblock, RB_CMD_READ_RX_USED, word_ptr) != 0)
        result = RIO_SUCCESS;
      break;

    case IOCTL_SER_GET_TX_BUF_LEVEL:
      /* Request how many bytes available in the transmit buffer */
      result = rb_sio_get_tx_buf_lvl(rbsc, va_arg(*arglist, unsigned short *));
      break;

    case IOCTL_SER_GET_RX_BUF_SIZE:
      /* No receiver on serial channel 1, buffer size = 0, and return error */
      word_ptr = va_arg(*arglist, unsigned short *);
      if (rbsc->channel != rbsc_a)
      {
        *word_ptr = 0;
        goto rb_i_end;
      }

      /* Request size of receive buffer */
      if (rb_sio_get_buf_level(rbsc->rinblock, RB_CMD_READ_RX_BUF_SIZE, word_ptr) != 0)
        result = RIO_SUCCESS;
      break;

    case IOCTL_SER_GET_TX_BUF_SIZE:
      /* Request size of transmit buffer */
      if (rb_sio_get_tx_buf_size(rbsc, va_arg(*arglist, unsigned short *)) != 0)
        result = RIO_SUCCESS;
      break;

    case IOCTL_SER_GET_CTS_STATE:
      /* Find out the state of the CTS pin */
      if (rb_sio_get_cts_state(rbsc, va_arg(*arglist, unsigned char *)) != 0)
        result = RIO_SUCCESS;
      break;

    case IOCTL_SER_SET_RING_NETWORK:
      /* Configure the rinblock ring network option */
      state = (unsigned char) va_arg(*arglist, int);
      if (rb_sio_set_ring_network(rbsc, state) != 0)
        result = RIO_SUCCESS;
      break;

    case IOCTL_SER_FLUSH_RX_BUF:
      result = rb_sio_rx_flush(io_dev);
      break;

    default:
      break;
  }

rb_i_end:
  return result;
}


/** Create a serial device for a rinblock & channel */
int RinBlockCreateSerialDevice(struct t_io_dev *io_dev, struct t_rinblock *rb, enum e_rb_serial_channel channel)
{
  struct s_rb_serial_channel *rbsc = NULL;
  bool byteq_created = false;

  if (NULL == io_dev || NULL == rb)
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateSerialDevice: parameter failure\n");
#endif
    return RIO_ERROR;
  }

  /* Check that the mapped rinblock is actually a serial one */
  if (!rb_sio_isserial(rb))
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateSerialDevice: failed to map serial channel to non-serial rinblock\n");
#endif
    return RIO_ERROR;
  }

  /* Allocate a structure */
  rbsc = (struct s_rb_serial_channel *) calloc(1, sizeof(struct s_rb_serial_channel));
  if (NULL == rbsc)
  {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateSerialDevice: failed to allocate device structure\n");
#endif
    goto rbcsd_fail;
  }

  /* If this is the A channel, we setup a local queue to hold data so we can do bigger */
  if (channel == rbsc_a)
  {
    rbsc->rx_buffer = calloc(LOCAL_RX_BUFFER_SIZE, sizeof(uint8_t));
    if (NULL == rbsc->rx_buffer)
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateSerialDevice: failed to allocate local rx buffer\n");
#endif
      goto rbcsd_fail;
    }
    if (ByteQCreate(&rbsc->rx_queue, bqNoPend, rbsc->rx_buffer, LOCAL_RX_BUFFER_SIZE) != bqrSuccess)
    {
#if RB_DEBUG(RB_SERIAL_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateSerialDevice: failed to create local rx byte queue\n");
#endif
      goto rbcsd_fail;
    }
    byteq_created = true;
  }

  /* Fill out members */
  rbsc->rinblock = rb;
  rbsc->channel = channel;

  /*
   * Setup device data pointer to point to the rinblock structure
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) rbsc;
  io_dev->status = 0UL;
  io_dev->posn = 0UL;
  io_dev->sem_open_excl = NULL;

  /* Setup function pointers */
  io_dev->dev_init = rb_sio_init;
  io_dev->dev_open = rb_sio_open;
  io_dev->dev_close = rb_sio_close;
  io_dev->dev_read = rb_sio_read;
  io_dev->dev_write = rb_sio_write;
  io_dev->dev_filelength = rb_sio_filelength;
  io_dev->dev_ioctl = rb_sio_ioctl;
  io_dev->dev_destroy = rb_sio_destroy;
  io_dev->dev_isatty = NULL;
  io_dev->dev_tell = NULL;
  io_dev->dev_lseek = NULL;

  /* Register and initialise the device */
  if (rregister_io(io_dev) < 0)
    goto rbcsd_fail;

  /* Success */
  return RIO_SUCCESS;

rbcsd_fail:
  io_dev->dev_data = NULL;
  if (NULL != rbsc)
  {
    if (byteq_created)
      ByteQDestroy(&rbsc->rx_queue);
    if (NULL != rbsc->rx_buffer)
      free(rbsc->rx_buffer);
    free(rbsc);
    rbsc = NULL;
  }
  return RIO_ERROR;
}
