#include <rinhw/iic.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/rstdio.h>
#include <rinblock/rb_command.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_debug.h>
#include "_iic_access.h"
#include <rinblock/rb_tag.h>
#include "_rb_tag.h"
#include "_rb_dio.h"

#include <rinblock/rb_startup.h>


/*
 * Array of rinblocks
 */
struct t_rinblock rinblocks[MAX_RINBLOCKS];


/*
 * RinBlockSetupStruct -
 *
 * Initialise all the elements within the rinblock struct.
 */
static void RinBlockSetupStruct(struct t_rinblock *rinblock, struct t_io_dev *iic_dev, unsigned char slot)
{
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockSetupStruct() called, slot %d\n", slot);
#endif

    /*
     * Setup general rin block parameters
     */
    rinblock->iicdev = iic_dev;
    rinblock->iic_address = RB_DEFAULT_ADDRESS;
    rinblock->bus_slot = slot;
    rinblock->present = rinblock->programmed = rinblock->running = 0;

    /*
     * Setup rin block message handling parameters
     */
    rinblock->msg_handling.busy_sem = OSSemCreate(1);
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    if ( (OS_EVENT *) 0 == rinblock->msg_handling.busy_sem )
    {
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Error creating rinblock semaphore\n");
    }
#endif

    /*
     * Setup Rinblock DIO to have no IO at this stage
     */
    rinblock->RBDioHardware = (enum eRBDioHardware *) NULL;

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockSetupStruct() finished\n");
#endif
}


/*
 * RinBlockSetAddressInitial -
 *
 * Assign an address to a rinblock based on its bus_slot parameter
 * Assumes current iic_address is set to RB_DEFAULT_ADDRESS
 */
static void RinBlockSetAddressInitial(struct t_rinblock *rb)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, struct s_rb_general_address );
    unsigned char res, result, new_addr = SlotToAddress(rb->bus_slot);

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockSetAddressInitial: Set rinblock address at bus slot %d to %d\n", rb->bus_slot, new_addr);
#endif

    /*
     * Set the address for the module on each slot
     */
    rb_msg->data.rb_command.header.cmd = RB_CMD_WRITE_ADDRESS;
    rb_msg->data.rb_command.params.rb_general_address.address = new_addr;
    res = RinBlockInternalExecuteCommand(rb, rb_msg,
                                         RB_GET_SIZE_COMMAND( struct s_rb_general_address ),
                                         RB_GET_SIZE_CMD_HDR(),
                                         RB_TIMEOUT_STARTUP, 0);

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockSetAddressInitial: RB_CMD_WRITE_ADDRESS returned with result %d, response 0x%02x\n", res, rb_msg->data.rb_response.header.resp);
#endif

    result = ((res) && ((rb_msg->data.rb_response.header.resp & RB_RESP_ERROR_MASK) == 0));
    if (result)
    {

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockSetAddressInitial: Rinblock address has been set\n");
#endif
        rb->iic_address = new_addr;
        rb->present = TRUE;
    }
	else
    {

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockSetAddressInitial: No rinblock at slot\n");
#endif
        rb->present = FALSE;
    }
}


/*
 * RinBlockSetMode -
 *
 * Set the running mode for the rinblock rb.
 * Mode defined by the defines RB_MODE_xxxx
 */
unsigned char RinBlockSetMode(struct t_rinblock *rb, unsigned char mode)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, struct s_rb_general_mode );
    unsigned char res;

    /*
     * Setup for rinblock command
     */
    rb_msg->data.rb_command.header.cmd = RB_CMD_SET_MODE;
    rb_msg->data.rb_command.params.rb_general_mode.mode = mode;

    /*
     * Execute the rinblock command
     */
    res = RinBlockInternalExecuteCommand(rb, rb_msg,
                                         RB_GET_SIZE_COMMAND( struct s_rb_general_mode ),
                                         RB_GET_SIZE_CMD_HDR(),
                                         RB_TIMEOUT_STARTUP, 0);

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RB_CMD_SET_MODE returned with result %d, ", res);
    if (res)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "response 0x%02x\n", rb_msg->data.rb_response.header.resp);
    else
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "no response\n");
#endif

    return ((res) && ((rb_msg->data.rb_response.header.resp & RB_RESP_ERROR_MASK) == 0));
}


/*
 * RinBlockGetMode -
 *
 * Find out the current mode of a rinblock.
 * Mode defined by the defines RB_MODE_xxxx
 */
unsigned char RinBlockGetMode(struct t_rinblock *rb, unsigned char *mode)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, struct s_rb_general_mode );
    unsigned char res;

    rb_msg->data.rb_command.header.cmd = RB_CMD_GET_MODE;
    res = RinBlockInternalExecuteCommand(rb, rb_msg,
                                         RB_GET_SIZE_CMD_HDR(),
                                         RB_GET_SIZE_COMMAND( struct s_rb_general_mode ),
                                         RB_TIMEOUT_STARTUP, 0);

    if ((res) && ((rb_msg->data.rb_response.header.resp & RB_RESP_ERROR_MASK) == 0))
    {
        *mode = rb_msg->data.rb_response.params.rb_general_mode.mode;
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RB_CMD_GET_MODE success, rinblock mode: %d\n", *mode);
#endif
        return 1;
    }

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RB_CMD_GET_MODE failed\n");
#endif
    return 0;
}


/*
 * RinBlockInitialAllReset -
 *
 * On startup we send all rinblocks a global call RESET command.
 * This deals with the case when the main board has been restarted, but
 * not powercycled, hence the modules are still running, but
 * we don't know anything about them.
 *
 * DON'T CALL THIS FUNCTION EXCEPT AS PART OF THE STARTUP
 */
static void RinBlockInitialAllReset(void)
{
    RB_DECLARE_CMD_BUFFER( rb_msg, struct s_rb_general_mode );
    unsigned char res, addr_save;

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInitialAllReset() called\n");
#endif

    /*
     * Steal rin block 0 to do the dirty work
     * since none are being used at this point.
     *
     * We need to trick it into working.
     */
    addr_save = rinblocks[0].iic_address;
    rinblocks[0].iic_address = 0;
    rinblocks[0].present = 1;

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInitialAllReset: rb_msg %p\n", rb_msg);
#endif

    /*
     * Setup and run a set mode command
     */
    rb_msg->data.rb_command.header.cmd = RB_CMD_SET_MODE;
    rb_msg->data.rb_command.params.rb_general_mode.mode = RB_MODE_RESET;
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInitialAllReset: rb_msg %p\n", rb_msg);
#endif
    res = RinBlockInternalExecuteCommand(&rinblocks[0], rb_msg,
                                         RB_GET_SIZE_COMMAND( struct s_rb_general_mode ),
                                         RB_GET_SIZE_CMD_HDR(),
                                         RB_TIMEOUT_STARTUP, 1);

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Rin block global call reset returned with result %d\n", res);
#endif

    /*
     * Delay while rin blocks start
     */
    OSTimeDly(OS_TICKS_PER_SEC / 4);

    /*
     * Now, only for non-R500 tell all module bootloaders to
     * listen for IRQ level when setting address
     */
    rb_msg->data.rb_command.header.cmd = RB_CMD_SET_MODE;
    rb_msg->data.rb_command.params.rb_general_mode.mode = RB_MODE_LISTEN_IRQ;
    res = RinBlockInternalExecuteCommand(&rinblocks[0], rb_msg,
                                         RB_GET_SIZE_COMMAND( struct s_rb_general_mode ),
                                         RB_GET_SIZE_CMD_HDR(),
                                         RB_TIMEOUT_STARTUP, 1);

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "Rin block global set mode listen returned with result %d\n", res);
#endif

    /*
     * Set things back to normal
     */
    rinblocks[0].present = 0;
    rinblocks[0].iic_address = addr_save;

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInitialAllReset() finished\n");
#endif
}


void RinBlockStartup(struct t_rinblock *rb)
{
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockStartup() called\n");
#endif

    if (rb->present)
    {
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockStartup: starting rinblock at slot %d, address %d\n", rb->bus_slot, rb->iic_address);
#endif

        /*
         * First try to start the rin block running
         */
        if (!RinBlockSetMode(rb, RB_MODE_RUN))
        {
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockStartup: start rinblock failed\n");
#endif
            rb->programmed = rb->running = FALSE;
            return;
        }
        rb->programmed = rb->running = TRUE;
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockStartup: rinblock now running, reading tag\n");
#endif

        /*
         * Rin block is running,
         * now read back the tag to work out what it is
         */
        if (!RinBlockReadTag(rb, &rb->tag))
        {
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
            RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockStartup: read tag failed\n");
#endif
            rb->running = FALSE;
            return;
        }

#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockStartup: read tag successful\n");
#endif

        /*
         * Dump tag
         */
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_FULL)
        RinBlockDumpTag(&rb->tag, rb->iic_address);
#endif

        /*
         * Setup the DIO hardware for a rinblock based on the model number.
         */
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockStartup: configuring DIO\n");
#endif
        RinBlockSetupDioHardware(rb);
    }
#if RB_DEBUG(RB_STARTUP_DEBUG_LVL, RB_DEBUG_BASIC)
    else
    {
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockStartup: Not starting rinblock at slot %d, address %d as it is not present\n", rb->bus_slot, rb->iic_address);
    }
#endif
}


/*
 * RinBlockEnumerate -
 *
 * Assign an address to each rinblock on the bus
 * Must be called first after RinBlockInit()
 */
void RinBlockEnumerate(void)
{
    struct t_rinblock *rbptr = rinblocks;
    unsigned char i;

#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockEnumerate() called\n");
#endif

    /*
     * Send a broadcast reset to all modules on the bus.
     * This deals with the case when the main board has reset,
     * and the modules are still running.
     */
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockEnumerate: Reset all rinblocks\n");
#endif
    RinBlockInitialAllReset();

    /*
     * Now set the IIC address of each rin block
     */
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockEnumerate: Assign addresses to rinblocks\n");
#endif
    for (i = 0; i < MAX_RINBLOCKS; i++, rbptr++)
    {
    	RinBlockSetAddressInitial(rbptr);
    }

#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockEnumerate() finished\n");
#endif
}


/*
 * RinBlockStart -
 *
 * Start all programmed rin blocks on the bus running,
 * and find out what they are
 */
void RinBlockStart(void)
{
    struct t_rinblock *rbptr = rinblocks;
    unsigned char i;

    /*
     * Now start up all rinblocks
     */
    for (i = 0; i < MAX_RINBLOCKS; i++, rbptr++)
    {
        RinBlockStartup(rbptr);
    }
}


/*
 * RinBlockRestart -
 *
 * Restart a rinblock. Sends a reset command, followed by the startup sequence
 */
void RinBlockRestart(struct t_rinblock *rb)
{

    /*
     * Tell the rinblock to reset,
     * and wait a bit to allow it to restart, before
     * resuming communications with it
     */
    RinBlockSetMode(rb, RB_MODE_RESET);
    OSTimeDly(OS_TICKS_PER_SEC / 2);

    /*
     * Now make the rinblock struct uninitialised
     */
    rb->iic_address = RB_DEFAULT_ADDRESS;
    rb->present = rb->programmed = rb->running = 0;

    /*
     * Set the address for the first time
     */
    RinBlockSetAddressInitial(rb);

    /*
     * Start the rinblock
     */
    RinBlockStartup(rb);
}


/*
 * RinBlockInit -
 *
 * Setup the IIC device, open the debugging device, if debugging
 * is enabled. Initialise the rin block structs.
 */
void RinBlockInit(struct t_io_dev *iicdev)
{
    struct t_rinblock *rbptr;
    unsigned char i;
    int res;

    /* Open the debugging device if debugging is enabled */
#if (defined(RINBLOCK_DEBUG) && !RB_DEBUG_APP_CALLBACK)
    if ((RB_DEBUG_DEV)->status & O_OPEN)
        ropen(RB_DEBUG_DEV, 0);
#endif

#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInit() called\n");
#endif

    /* First set the IIC address of each rin block */
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInit: Setup rinblock structures\n");
#endif
    rbptr = &rinblocks[0];
    for (i = 0; i < MAX_RINBLOCKS; i++, rbptr++)
    {
        RinBlockSetupStruct(rbptr, iicdev, i);
    }

    /* Configure the IIC bus */
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInit: Configure IIC bus\n");
#endif
    res = ropen(iicdev, 0);
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    if (res < 0)
    {
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInit: Could not open IIC device\n");
    }
#endif
    res = rioctl(iicdev, IOCTL_IIC_SET_SPEED, IIC_SPEED_CLK_DIV_40);
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    if (res < 0)
    {
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInit: IIC set bus speed failed\n");
    }
#endif
    res = rioctl(iicdev, IOCTL_IIC_DEFAULT_SLOT_MASK);
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    if (res < 0)
    {
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInit: IIC set slot mask failed\n");
    }
#endif
    res = rclose(iicdev);
#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    if (res < 0)
    {
        RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInit: Could not close IIC device\n");
    }
#endif

#if RB_DEBUG(RB_MESSAGE_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockInit() finished\n");
#endif
}
