/*
 * rb_tilt_basic.c
 *
 * Contains specific commands supported by modules with RB_IFACE_VERS_TILT_ORIGINAL and above
 */
#include <stdarg.h>
#include <string.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_command.h>
#include <rinblock/rb_msg.h>

#include "_rb_tilt.h"
#include <rinblock/rb_tilt.h>


/** Basic read tilt data */
unsigned char RinBlockTiltReadTilt(struct t_rinblock *rb, struct s_rb_tilt_data *tilt_data)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_tilt_data);
  unsigned char result;

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockAOutReadConfig called: ");
#endif

  /*
   * Read the tilt data
   */
  result = RinBlockExecuteCommand(rb, rb_msg,
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_GET_SIZE_COMMAND(struct s_rb_tilt_data),
                                  RB_TIMEOUT_DEFAULT, RB_CMD_READ_TILT);
  if (result)
  {
#if (BYTE_ORDER == NETWORK_ENDIAN)
    *tilt_data = rb_msg->data.rb_response.params.rb_tilt_data;
#else /* #if (BYTE_ORDER == NETWORK_ENDIAN) */
    tilt_data->xAngle = ntohs(rb_msg->data.rb_response.params.rb_tilt_data.xAngle);
    tilt_data->yAngle = ntohs(rb_msg->data.rb_response.params.rb_tilt_data.yAngle);
    tilt_data->cosX = ntohs(rb_msg->data.rb_response.params.rb_tilt_data.cosX);
    tilt_data->cosY = ntohs(rb_msg->data.rb_response.params.rb_tilt_data.cosY);
    tilt_data->comp = ntohs(rb_msg->data.rb_response.params.rb_tilt_data.comp);
    tilt_data->errFlag = rb_msg->data.rb_response.params.rb_tilt_data.errFlag;
#endif /* #else -> #if (BYTE_ORDER == NETWORK_ENDIAN) */

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success, xangle: %04x, yangle: %04x\n",
                    tilt_data->xAngle,
                    tilt_data->yAngle);
#endif
  }
#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "failure\n");
  }
#endif
  return result;
}


/** Perform a tilt calibration */
unsigned char RinBlockTiltCalibrate(struct t_rinblock *rb, const struct s_rb_tilt_calibrate *tilt_calibrate)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_tilt_calibrate);
  unsigned char res;

  /* Check that the rinblock has a tilt interface, and interface version is supported */
  if (!rb_check_has_tilt(rb, RB_IFACE_VERS_TILT_ORIGINAL))
  {
#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltCalibrate(): rinblock does not support tilt\n");
#endif
    return 0;
  }

  /* Perform the calibration */
#if (BYTE_ORDER == NETWORK_ENDIAN)
  rb_msg->data.rb_command.params.rb_tilt_calibrate = *tilt_calibrate;
#else /* #if (BYTE_ORDER == NETWORK_ENDIAN) */
  rb_msg->data.rb_command.params.rb_tilt_calibrate.xPosAngle = htons(tilt_calibrate->xPosAngle);
  rb_msg->data.rb_command.params.rb_tilt_calibrate.xNegAngle = htons(tilt_calibrate->xNegAngle);
  rb_msg->data.rb_command.params.rb_tilt_calibrate.xPosComp = htons(tilt_calibrate->xPosComp);
  rb_msg->data.rb_command.params.rb_tilt_calibrate.xNegComp = htons(tilt_calibrate->xNegComp);
  rb_msg->data.rb_command.params.rb_tilt_calibrate.yPosAngle = htons(tilt_calibrate->yPosAngle);
  rb_msg->data.rb_command.params.rb_tilt_calibrate.yNegAngle = htons(tilt_calibrate->yNegAngle);
  rb_msg->data.rb_command.params.rb_tilt_calibrate.yPosComp = htons(tilt_calibrate->yPosComp);
  rb_msg->data.rb_command.params.rb_tilt_calibrate.yNegComp = htons(tilt_calibrate->yNegComp);
#endif /* #else -> #if (BYTE_ORDER == NETWORK_ENDIAN) */

  res = RinBlockExecuteCommand(rb, rb_msg,
                               RB_GET_SIZE_COMMAND(struct s_rb_tilt_calibrate),
                               RB_GET_SIZE_CMD_HDR(),
                               RB_TIMEOUT_DEFAULT, RB_CMD_CALIBRATE);

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
  if (res)
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltCalibrate(): success, cool!\n");
  }
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltCalibrate(): failed to perform tilt calibration\n");
  }
#endif
  return res;
}
