#include <stdarg.h>
#include <string.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_command.h>

#include "_rb_tilt.h"
#include <rinblock/rb_tilt.h>


/** Perform a tilt zero calibration */
unsigned char RinBlockTiltZero(struct t_rinblock *rb)
{
  RB_DECLARE_CMD_BUFFER_NO_PARAMS(rb_msg);

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltZero called\n");
#endif

  /*
   * Write the rinblock
   */
  return RinBlockExecuteCommand(rb, rb_msg,
                                RB_GET_SIZE_CMD_HDR(),
                                RB_GET_SIZE_CMD_HDR(),
                                RB_TIMEOUT_DEFAULT, RB_CMD_ZERO_SENSOR);
}


/** Perform a tilt factory zero calibration */
unsigned char RinBlockTiltFactoryZero(struct t_rinblock *rb)
{
  RB_DECLARE_CMD_BUFFER_NO_PARAMS(rb_msg);

  #if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltFactoryZero called\n");
  #endif

  /*
   * Write the rinblock
   */
  return RinBlockExecuteCommand(rb, rb_msg,
                                RB_GET_SIZE_CMD_HDR(),
                                RB_GET_SIZE_CMD_HDR(),
                                RB_TIMEOUT_DEFAULT, RB_CMD_FACTORY_ZERO_SENSOR);
}


/** Read the configuration from the tilt module */
unsigned char RinBlockTiltReadConfig(struct t_rinblock *rb, struct s_rb_tilt_factors *tilt_factors)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_tilt_factors);
  unsigned char result;

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltReadConfig called: ");
#endif

  /*
   * Read the tilt factors
   */
  result = RinBlockExecuteCommand(rb, rb_msg,
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_GET_SIZE_COMMAND(struct s_rb_tilt_factors),
                                  RB_TIMEOUT_DEFAULT, RB_CMD_READ_FACTORS);
  if (result)
  {
#if (BYTE_ORDER == NETWORK_ENDIAN)
    *tilt_factors = rb_msg->data.rb_response.params.rb_tilt_factors;
#else /* #if (BYTE_ORDER == NETWORK_ENDIAN) */
    tilt_factors->aFactor = ntohs( rb_msg->data.rb_response.params.rb_tilt_factors.aFactor );
    tilt_factors->bFactor = ntohs( rb_msg->data.rb_response.params.rb_tilt_factors.bFactor );
    tilt_factors->cFactor = ntohs( rb_msg->data.rb_response.params.rb_tilt_factors.cFactor );
    tilt_factors->dFactor = ntohs( rb_msg->data.rb_response.params.rb_tilt_factors.dFactor );
    tilt_factors->maxTilt = rb_msg->data.rb_response.params.rb_tilt_factors.maxTilt;
#endif /* #else -> #if (BYTE_ORDER == NETWORK_ENDIAN) */

  #if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success afact:%04x bfact:%04x cfact:%04x dfact:%04x\n",
             tilt_factors->aFactor,
             tilt_factors->bFactor,
             tilt_factors->cFactor,
             tilt_factors->dFactor);
  #endif
  }
#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "failure\n");
  }
#endif
  return result;
}


/** Write the tilt configuration to a tilt module */
unsigned char RinBlockTiltWriteConfig(struct t_rinblock *rb, const struct s_rb_tilt_factors *tilt_factors)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_tilt_factors);
  unsigned char res;

  /* Check that the rinblock has a tilt interface, and interface version is supported */
  if (!rb_check_has_tilt(rb, RB_IFACE_VERS_TILT_ORIGINAL))
  {
#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltWriteConfig(): rinblock does not support tilt\n");
#endif
    return 0;
  }

  /* Write the config */
#if (BYTE_ORDER == NETWORK_ENDIAN)
  rb_msg->data.rb_command.params.rb_tilt_factors = *tilt_factors;
#else /* #if (BYTE_ORDER == NETWORK_ENDIAN) */
  rb_msg->data.rb_command.params.rb_tilt_factors.aFactor = htons(tilt_factors->aFactor);
  rb_msg->data.rb_command.params.rb_tilt_factors.bFactor = htons(tilt_factors->bFactor);
  rb_msg->data.rb_command.params.rb_tilt_factors.cFactor = htons(tilt_factors->cFactor);
  rb_msg->data.rb_command.params.rb_tilt_factors.dFactor = htons(tilt_factors->dFactor);
  rb_msg->data.rb_command.params.rb_tilt_factors.maxTilt = tilt_factors->maxTilt;
#endif /* #else -> #if (BYTE_ORDER == NETWORK_ENDIAN) */
  res = RinBlockExecuteCommand(rb, rb_msg,
                               RB_GET_SIZE_COMMAND(struct s_rb_tilt_factors),
                               RB_GET_SIZE_CMD_HDR(),
                               RB_TIMEOUT_DEFAULT, RB_CMD_WRITE_FACTORS);

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
  if (res)
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltWriteConfig(): success, cool!\n");
  }
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltWriteConfig(): failed to write tilt configuration\n");
  }
#endif
  return res;
}


/** Check whether a rinblock has a tilt interface type */
unsigned char rb_check_has_tilt(const struct t_rinblock *rb, unsigned char iface_vers)
{
    return (RB_IFACE_TYPE_TILT == rb->tag.interface_type &&
            rb->tag.interface_version >= iface_vers);
}
