/*
 * rb_tilt_improved.c
 *
 * Contains specific commands supported by modules with RB_IFACE_VERS_TILT_IMPROVED and above
 */
#include <stdarg.h>
#include <string.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_command.h>
#include <rinblock/rb_msg.h>

#include "_rb_tilt.h"
#include <rinblock/rb_tilt.h>


/** Improved read tilt data */
unsigned char RinBlockTiltReadTiltImproved(struct t_rinblock *rb, struct s_rb_tilt_data_improved *tilt_data_improved)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_tilt_data_improved);
  unsigned char result;

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltReadTiltImproved() called: ");
#endif

  /*
   * Read the tilt data
   */
  result = RinBlockExecuteCommand(rb, rb_msg,
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_GET_SIZE_COMMAND(struct s_rb_tilt_data_improved),
                                  RB_TIMEOUT_DEFAULT, RB_CMD_READ_TILT_IMPROVED);
  if (result)
  {
#if (BYTE_ORDER == NETWORK_ENDIAN)
    *tilt_data_improved = rb_msg->data.rb_response.params.rb_tilt_data_improved;
#else /* #if (BYTE_ORDER == NETWORK_ENDIAN) */
    tilt_data_improved->xAngle = ntohs(rb_msg->data.rb_response.params.rb_tilt_data_improved.xAngle);
    tilt_data_improved->yAngle = ntohs(rb_msg->data.rb_response.params.rb_tilt_data_improved.yAngle);
    tilt_data_improved->cosX = ntohs(rb_msg->data.rb_response.params.rb_tilt_data_improved.cosX);
    tilt_data_improved->cosY = ntohs(rb_msg->data.rb_response.params.rb_tilt_data_improved.cosY);
    tilt_data_improved->comp = ntohs(rb_msg->data.rb_response.params.rb_tilt_data_improved.comp);
    tilt_data_improved->errFlag = rb_msg->data.rb_response.params.rb_tilt_data_improved.errFlag;
    tilt_data_improved->errFlagCalX = rb_msg->data.rb_response.params.rb_tilt_data_improved.errFlagCalX;
    tilt_data_improved->errFlagCalY = rb_msg->data.rb_response.params.rb_tilt_data_improved.errFlagCalY;
#endif /* #else -> #if (BYTE_ORDER == NETWORK_ENDIAN) */

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "success, xangle: %04x, yangle: %04x\n",
                    tilt_data_improved->xAngle,
                    tilt_data_improved->yAngle);
#endif
  }
#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "failure\n");
  }
#endif
  return result;
}


/** Perform an improved tilt calibration */
unsigned char RinBlockTiltCalibrateImproved(struct t_rinblock *rb, const struct s_rb_tilt_calibrate_improved *tilt_calibrate_improved)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_tilt_calibrate_improved);
  unsigned char res;

  /* Check that the rinblock has a tilt interface, and interface version is supported */
  if (!rb_check_has_tilt(rb, RB_IFACE_VERS_TILT_IMPROVED))
  {
#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltCalibrateImproved(): rinblock does not support tilt\n");
#endif
    return 0;
  }

  /* Perform the calibration */
#if (BYTE_ORDER == NETWORK_ENDIAN)
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved = *tilt_calibrate_improved;
#else /* #if (BYTE_ORDER == NETWORK_ENDIAN) */
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.xPosAngle = htons(tilt_calibrate_improved->xPosAngle);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.xNegAngle = htons(tilt_calibrate_improved->xNegAngle);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.xPosComp = htons(tilt_calibrate_improved->xPosComp);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.xNegComp = htons(tilt_calibrate_improved->xNegComp);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.yPosAngle = htons(tilt_calibrate_improved->yPosAngle);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.yNegAngle = htons(tilt_calibrate_improved->yNegAngle);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.yPosComp = htons(tilt_calibrate_improved->yPosComp);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.yNegComp = htons(tilt_calibrate_improved->yNegComp);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.maxTilt = htons(tilt_calibrate_improved->maxTilt);
  rb_msg->data.rb_command.params.rb_tilt_calibrate_improved.numDiv = htons(tilt_calibrate_improved->numDiv);
#endif /* #else -> #if (BYTE_ORDER == NETWORK_ENDIAN) */

  res = RinBlockExecuteCommand(rb, rb_msg,
                               RB_GET_SIZE_COMMAND(struct s_rb_tilt_calibrate_improved),
                               RB_GET_SIZE_CMD_HDR(),
                               RB_TIMEOUT_DEFAULT, RB_CMD_CALIBRATE_IMPROVED);

#if RB_DEBUG(RB_TILT_DEBUG_LVL, RB_DEBUG_BASIC)
  if (res)
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltCalibrateImproved(): success, cool!\n");
  }
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockTiltCalibrateImproved(): failed to perform tilt calibration\n");
  }
#endif
  return res;
}
