#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <rinhw/rio.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_msg.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_command.h>

#include <rinblock/rb_we21_adc.h>


/** Structure describing an ADC rinblock */
struct s_rb_adc_info
{
  struct t_rinblock *rb;
  int num_channels;
  int16u *channels;
};


/** Get ADC readings from WE2111 connector rinblock */
static unsigned char RinBlockWe21GetAdc(struct s_rb_adc_info *rb_adc_info)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_we21_conn_adc);
  unsigned char result;
  int i;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockWe21GetAdc called\n");
#endif

  /* Read the ADC data */
  result = RinBlockExecuteCommand(rb_adc_info->rb, rb_msg,
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_GET_SIZE_COMMAND(struct s_rb_we21_conn_adc),
                                  RB_TIMEOUT_DEFAULT, RB_CMD_WE21_CONN_READ_ADC);
  if (result)
  {
    for (i = 0; i < rb_adc_info->num_channels; i++)
    {
      rb_adc_info->channels[i] = ntohs(rb_msg->data.rb_response.params.rb_we21_conn_adc.adc_reading[i]);
    }
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockWe21GetAdc: Got ADC readings from rinblock: Vin:%u temp:%u Vsns-:%u Vsns+:%u\n",
                    rb_adc_info->channels[RB_WE21_CONN_ADC_CH_VIN], rb_adc_info->channels[RB_WE21_CONN_ADC_CH_TEMP],
                    rb_adc_info->channels[RB_WE21_CONN_ADC_CH_SNS_MINUS], rb_adc_info->channels[RB_WE21_CONN_ADC_CH_SNS_PLUS]);
#endif
  }
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_FULL)
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockWe21GetAdc: Get ADC readings failed\n");
  }
#endif

  return result;
}


/** Get ADC readings from analog input rinblock */
static unsigned char RinBlockAnalogInGetAdc(struct s_rb_adc_info *rb_adc_info)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_ain_data);
  unsigned char result;
  int i;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockAnalogInGetAdc called\n");
#endif

  /* Read the config */
  result = RinBlockExecuteCommand(rb_adc_info->rb, rb_msg,
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_GET_SIZE_COMMAND(struct s_rb_ain_data),
                                  RB_TIMEOUT_DEFAULT, RB_CMD_READ_AIN);
  if (result)
  {
    for (i = 0; i < AIN_NUM_TOTAL; i++)
    {
      rb_adc_info->channels[i] = ntohs(rb_msg->data.rb_response.params.rb_ain_read.ainVolts[i]);
    }
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockAnalogInGetAdc: Got ADC readings from rinblock\n");
#endif
  }
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_FULL)
  else
  {
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockAnalogInGetAdc: Get ADC readings failed\n");
  }
#endif

  return result;
}


/** Calibrate the ADC readings from analog input rinblock */
static unsigned char RinBlockAnalogInWriteCalibration(struct s_rb_adc_info *rb_adc_info, int32u *raw_per_millivolt)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_ain_configure);
  int i;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockAnalogInWriteCalibration called\n");
#endif

  /* Only the analog input card supports calibration */
  if (RB_IFACE_TYPE_AIN != rb_adc_info->rb->tag.interface_type)
  {
    return 0;
  }

  /* Setup calibration values to transfer to indicator (in raw per hundredth of a volt) */
  for (i = 0; i < rb_adc_info->num_channels; i++)
  {
    rb_msg->data.rb_command.params.rb_ain_configure.rawPerVolt[i] = htons(raw_per_millivolt[i] / 10);
  }

  /* Calibrate the module */
  return RinBlockExecuteCommand(rb_adc_info->rb, rb_msg,
                                RB_GET_SIZE_COMMAND(struct s_rb_ain_configure),
                                RB_GET_SIZE_CMD_HDR(),
                                RB_TIMEOUT_DEFAULT, RB_CMD_WRITE_AIN_FACTORY);
}


/** Read calibration values from analog input rinblock */
static unsigned char RinBlockAnalogInReadCalibration(struct s_rb_adc_info *rb_adc_info, int32u *raw_per_millivolt)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_ain_configure);
  unsigned char result;
  int i;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockAnalogInReadCalibration called\n");
#endif

  /* Only the M5451 supports calibration */
  if (RB_IFACE_TYPE_AIN != rb_adc_info->rb->tag.interface_type)
  {
    return 0;
  }

  /* Read the calibration from the module */
  result = RinBlockExecuteCommand(rb_adc_info->rb, rb_msg,
                                RB_GET_SIZE_CMD_HDR(),
                                RB_GET_SIZE_COMMAND(struct s_rb_ain_configure),
                                RB_TIMEOUT_DEFAULT, RB_CMD_READ_AIN_FACTORY);
  if (result)
  {
    /* Get calibration values from the indicator indicator */
    for (i = 0; i < rb_adc_info->num_channels; i++)
    {
      raw_per_millivolt[i] = ntohs(rb_msg->data.rb_command.params.rb_ain_configure.rawPerVolt[i]) * 10;
    }
  }
  return result;
}


/** Rinblock ADC device init */
static int rb_adc_init(struct t_io_dev *io_dev)
{
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_init called\n");
#endif

  /* Its a block device */
  io_dev->status &= ~O_DEVCHAR;

  return RIO_SUCCESS;
}


/** Destroy Rinblock ADC device */
static int rb_adc_destroy(struct t_io_dev *io_dev)
{
  struct s_rb_adc_info *rb_adc_info;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_destroy called\n");
#endif

  /* Free up memory */
  if (NULL != io_dev)
  {
    if (NULL != io_dev->dev_data)
    {
      rb_adc_info = (struct s_rb_adc_info *) io_dev->dev_data;
      if (NULL != rb_adc_info->channels)
      {
        free(rb_adc_info->channels);
        rb_adc_info->channels = NULL;
        rb_adc_info->rb = NULL;
      }
      io_dev->dev_data = NULL;
      free(rb_adc_info);
    }
  }

  return RIO_SUCCESS;
}


/** Rinblock ADC device read, read data from rinblock and present in expected format */
static int rb_adc_read(struct t_io_dev *io_dev, void *buf, unsigned short len, unsigned short timeout)
{
  struct s_rb_adc_info *rb_adc_info;
  unsigned char *rd_ptr;
  unsigned char *wr_ptr = (unsigned char *) buf;
  unsigned char res;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_read called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == buf || NULL == io_dev->dev_data)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_read: NULL pointers\n");
#endif
    return RIO_ERROR;
  }
  rb_adc_info = (struct s_rb_adc_info *) io_dev->dev_data;
  rd_ptr = (unsigned char *) rb_adc_info->channels;

  /* Check read is sensible */
  if (io_dev->posn + len > rb_adc_info->num_channels * sizeof(int16u))
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_read: read outside of range\n");
#endif
    return RIO_ERROR;
  }

  /* Get readings from rinblock */
  switch (rb_adc_info->rb->tag.interface_type)
  {
    case RB_IFACE_TYPE_WE21_CONN:
      res = RinBlockWe21GetAdc(rb_adc_info);
      break;

    case RB_IFACE_TYPE_AIN:
      res = RinBlockAnalogInGetAdc(rb_adc_info);
      break;

    default:
      res = 0;
      break;
  }
  if (0 == res)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_read: RinBlockxxxxGetAdc() failed\n");
#endif
    return RIO_ERROR;
  }

  /* Copy out data */
  rd_ptr = &rd_ptr[io_dev->posn];
  memcpy(wr_ptr, rd_ptr, (size_t) len);
  io_dev->posn += len;
  return (int) len;
}


/** Rinblock ADC Device write not implemented */
static int rb_adc_write(struct t_io_dev *io_dev, const void *buf, unsigned short len, unsigned short timeout)
{
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_write called, not implemented\n");
#endif

  return RIO_ERROR;
}


/** Rinblock ADC device filelength */
static long rb_adc_filelength(struct t_io_dev *io_dev)
{
  struct s_rb_adc_info *rb_adc_info;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_filelength called\n");
#endif

  /* Check parameters */
  if (NULL == io_dev || NULL == io_dev->dev_data)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_ioctl() error: NULL pointers passed\n");
#endif
    return RIO_ERROR;
  }
  rb_adc_info = (struct s_rb_adc_info *) io_dev->dev_data;

  return (long) rb_adc_info->num_channels * sizeof(int16u);
}


/** Device specific configuration */
static int rb_adc_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
  struct s_rb_adc_info *rb_adc_info;
  int *int_ptr_param;
  int32u *raw_per_millivolt;
  int result = RIO_SUCCESS;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_ioctl called, function %d\n", func);
#endif

  /* Check parameters */
  if (NULL == io_dev || NULL == io_dev->dev_data)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_ioctl() error: NULL pointers passed\n");
#endif
    return RIO_ERROR;
  }
  rb_adc_info = (struct s_rb_adc_info *) io_dev->dev_data;

  /* Function dispatch */
  switch (func)
  {
    /* Get the number of channels on this ADC */
    case IOCTL_ADC_GET_NUM_CHANNELS:
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_ioctl(): IOCTL_ADC_GET_NUM_CHANNELS called\n");
#endif
      int_ptr_param = va_arg(*arglist, int *);
      *int_ptr_param = rb_adc_info->num_channels;
      break;

    /* Calibrate the ADC channels */
    case IOCTL_ADC_WRITE_CALIBRATION:
      raw_per_millivolt = va_arg(*arglist, int32u *);
      if (0 == RinBlockAnalogInWriteCalibration(rb_adc_info, raw_per_millivolt))
      {
        result = RIO_ERROR;
      }
      break;

    /* Read the calibration for the ADC channels */
    case IOCTL_ADC_READ_CALIBRATION:
      raw_per_millivolt = va_arg(*arglist, int32u *);
      if (0 == RinBlockAnalogInReadCalibration(rb_adc_info, raw_per_millivolt))
      {
        result = RIO_ERROR;
      }
      break;

    /* Unknown option */
    default:
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_adc_ioctl() error: Unknown IOCTL\n");
#endif
      result = RIO_ERROR;
      break;
  }

  return result;
}


/**
 * Create a WE2111 rinblock adc device
 *
 * @param io_dev is an uninitialised io device
 * @param rb is a pointer to a rinblock with an adc
 */
static int RinBlockCreateAdcDevice(struct t_io_dev *io_dev, struct t_rinblock *rb, int num_channels)
{
  struct s_rb_adc_info *rb_adc_info = NULL;

#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateAdcDevice called\n");
#endif

  /* Check parameters */
  if (NULL == io_dev || NULL == rb)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateAdcDevice() error: NULL pointers passed\n");
#endif
    goto rbcad_fail;
  }
  if (num_channels <= 0)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateAdcDevice() error: invalid number of channels\n");
#endif
    goto rbcad_fail;
  }

  /* Create structure */
  rb_adc_info = calloc(1, sizeof(struct s_rb_adc_info));
  if (NULL == rb_adc_info)
  {
    goto rbcad_fail;
  }

  /* Fill out members */
  rb_adc_info->rb = rb;
  rb_adc_info->num_channels = num_channels;
  rb_adc_info->channels = malloc(sizeof(int16u) * num_channels);
  if (NULL == rb_adc_info->channels)
  {
    goto rbcad_fail;
  }
  memset(rb_adc_info->channels, 0, sizeof(int16u) * num_channels);

  /*
   * Setup device data pointer to point to the rinblock structure
   * and initialise the status and position
   */
  io_dev->dev_data = (void *) rb_adc_info;
  io_dev->status = io_dev->posn = 0UL;

  /* Setup function pointers */
  io_dev->dev_init = rb_adc_init;
  io_dev->dev_open = default_open;
  io_dev->dev_close = default_close;
  io_dev->dev_read = rb_adc_read;
  io_dev->dev_write = rb_adc_write;
  io_dev->dev_filelength = rb_adc_filelength;
  io_dev->dev_ioctl = rb_adc_ioctl;
  io_dev->dev_destroy = rb_adc_destroy;
  io_dev->dev_isatty = default_isatty;
  io_dev->dev_tell = default_tell;
  io_dev->dev_lseek = default_lseek;

  /* Register and initialise the device */
  return rregister_io(io_dev);

rbcad_fail:

  if (NULL != rb_adc_info)
  {
    if (NULL != rb_adc_info->channels)
    {
      free(rb_adc_info->channels);
      rb_adc_info->channels = NULL;
    }
    free(rb_adc_info);
  }
  if (NULL != io_dev->dev_data)
  {
    io_dev->dev_data = NULL;
  }
  return RIO_ERROR;
}


/**
 * Create a WE2111 rinblock adc device
 *
 * @param io_dev is an uninitialised io device
 * @param rb is a pointer to a rinblock with an adc
 */
int RinBlockCreateWE21AdcDevice(struct t_io_dev *io_dev, struct t_rinblock *rb)
{
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21AdcDevice called\n");
#endif

  if (NULL == io_dev || NULL == rb)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21AdcDevice() error: NULL pointers passed\n");
#endif
    return RIO_ERROR;
  }

  /* Check that the mapped rinblock is actually a WE2111 connector board one */
  if (RB_IFACE_TYPE_WE21_CONN != rb->tag.interface_type)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21AdcDevice() error: attempting to map ADC to non-WE2111 connector board\n");
#endif
    return RIO_ERROR;
  }

  /* Create the device */
  return RinBlockCreateAdcDevice(io_dev, rb, RB_WE21_CONN_NUM_ADC_CHANNELS);
}


/**
 * Create an analog input rinblock adc device
 *
 * @param io_dev is an uninitialised io device
 * @param rb is a pointer to a rinblock with an adc
 */
int RinBlockCreateAnalogInAdcDevice(struct t_io_dev *io_dev, struct t_rinblock *rb)
{
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateAnalogInAdcDevice called\n");
#endif

  if (NULL == io_dev || NULL == rb)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateAnalogInAdcDevice() error: NULL pointers passed\n");
#endif
    return RIO_ERROR;
  }

  /* Check that the mapped rinblock is actually a M5451 rinblock */
  if (RB_IFACE_TYPE_AIN != rb->tag.interface_type)
  {
#if RB_DEBUG(RB_ADC_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateAnalogInAdcDevice() error: attempting to map ADC to non-M5451 rinblock\n");
#endif
    return RIO_ERROR;
  }

  /* Create the device */
  return RinBlockCreateAdcDevice(io_dev, rb, AIN_NUM_TOTAL);
}
