/*
 * buzzer_local.c
 *
 *  Created on: 10/02/2009
 *      Author: ianb
 */

#include <string.h>
#include <time.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/buzzer_generic.h>
#include <rinhw/timemath.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_command.h>

#include <rinblock/rb_we21_buzzer.h>


static unsigned char RinBlockWe21WriteBuzzer(struct t_rinblock *rb, const struct s_rb_we21_disp_buzzer *we21_buzzer)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_we21_disp_buzzer);

#if RB_DEBUG(RB_BUZ_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockWe21WriteBuzzer called\n");
#endif

  /*
   * Write the config
   */
  rb_msg->data.rb_command.params.rb_we21_buzzer = *we21_buzzer;

  return RinBlockExecuteCommand(rb, rb_msg,
                                RB_GET_SIZE_COMMAND(struct s_rb_we21_disp_buzzer),
                                RB_GET_SIZE_CMD_HDR(),
                                RB_TIMEOUT_DEFAULT, RB_CMD_WE21_DISP_BUZZER);
}


static void ConvertBuzzer(struct s_rb_we21_disp_buzzer *rb_we_buzz, const tBuzzer *buzz)
{
  memset(rb_we_buzz, 0, sizeof(struct s_rb_we21_disp_buzzer));
  if (0 == buzz->freq)
  {
    rb_we_buzz->buzz[0].oper = BUZZ_CLEAR;
    rb_we_buzz->buzz[0].kHz = 0;
  }
  else
  {
    rb_we_buzz->buzz[0].oper = BUZZ_ON;
    rb_we_buzz->buzz[0].kHz = buzz->freq / 1000;
  }
}


static int rb_buz_run_seq(struct s_generic_buzzer *gb)
{
  struct t_rinblock *rb;
  struct timespec ts_req, ts_rem;
  unsigned char result;
  struct s_rb_we21_disp_buzzer rb_we_buzz;
  tBuzzer silence_buzz;

  /* Check pointers */
  if (NULL == gb || NULL == gb->buz_hw_data)
    return RIO_ERROR;
  rb = (struct t_rinblock *) gb->buz_hw_data;

  /* Do beeping */
  for (gb->current_buz_seq = 0; gb->current_buz_seq < gb->num_buz_seq && !gb->buz_task_quit; gb->current_buz_seq++)
  {
#if RB_DEBUG(RB_BUZ_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_buz_run_seq: buz %d of %d, ticks %d, freq %d\n",
                    gb->current_buz_seq, gb->num_buz_seq,
                    gb->buz_seq[gb->current_buz_seq].time, gb->buz_seq[gb->current_buz_seq].freq);
#endif

    ConvertBuzzer(&rb_we_buzz, &gb->buz_seq[gb->current_buz_seq]);
    result = RinBlockWe21WriteBuzzer(rb, &rb_we_buzz);
#if RB_DEBUG(RB_BUZ_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_buz_run_seq: result of RinBlockWe21WriteBuzzer: %u\n", result);
#endif
    if (!result)
    {
      return RIO_ERROR;
    }

    /* Delay for the desired time */
    OSTicksToTimespec(&ts_req, gb->buz_seq[gb->current_buz_seq].time);
    nanosleep(&ts_req, &ts_rem);
  }

  /* Make sure current_buz_seq is larger than num_buz_seq for the buz_read func */
  gb->current_buz_seq = gb->num_buz_seq + 1;

  /* Shut the buzzer up */
  silence_buzz.time = 0;
  silence_buzz.freq = 0;
  ConvertBuzzer(&rb_we_buzz, &silence_buzz);
  result = RinBlockWe21WriteBuzzer(rb, &rb_we_buzz);
#if RB_DEBUG(RB_BUZ_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_buz_run_seq: result of RinBlockWe21WriteBuzzer: %u\n", result);
#endif
  return (result) ? RIO_SUCCESS : RIO_ERROR;
}


/**
 * Create a WE2111 rinblock buzzer device
 *
 * @param io_dev is an uninitialised io device
 * @param rb is a pointer to a rinblock with a buzzer
 */
int RinBlockCreateWE21BuzzerDevice(struct t_io_dev *io_dev, struct t_rinblock *rb)
{
  int res;
#if RB_DEBUG(RB_BUZ_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21BuzzerDevice called\n");
#endif

  /* Check pointers */
  if (NULL == io_dev || NULL == rb)
  {
#if RB_DEBUG(RB_BUZ_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21BuzzerDevice() error: NULL pointers passed\n");
#endif
    return RIO_ERROR;
  }

  /* Check that the mapped rinblock is actually a WE2111 or C500 display board */
  if (RB_IFACE_TYPE_WE21_DISP != rb->tag.interface_type &&
      RB_IFACE_TYPE_C500_DISP != rb->tag.interface_type)
  {
#if RB_DEBUG(RB_BUZ_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21BuzzerDevice() error: attempting to map buzzer to non-WE2111 display board\n");
#endif
    return RIO_ERROR;
  }

  /** Function to create and initialise a local buzzer device */
  res = rio_create_device_buzzer_generic(io_dev,
                                         rb,
                                         NULL,
                                         NULL,
                                         rb_buz_run_seq);
  return res;
}
