#include <string.h>
#include <stdlib.h>
#include <rinhw/kbd_local.h>
#include <rinblock/rb_debug.h>
#include <rinblock/rb_msg.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_command.h>

#include <rinblock/rb_we21_kbd.h>


static unsigned char RinBlockWe21GetKeyBits(struct t_rinblock *rb, struct s_rb_we21_disp_keybits *keybits)
{
  RB_DECLARE_CMD_BUFFER(rb_msg, struct s_rb_we21_disp_keybits);
  unsigned char result;

  /* Read the mask of pressed keys */
  result = RinBlockExecuteCommand(rb, rb_msg,
                                  RB_GET_SIZE_CMD_HDR(),
                                  RB_GET_SIZE_COMMAND(struct s_rb_we21_disp_keybits),
                                  RB_TIMEOUT_DEFAULT, RB_CMD_WE21_DISP_READ_KEYBITS);
  if (result)
  {
    keybits->key_bits = ntohl(rb_msg->data.rb_response.params.rb_we21_disp_keybits.key_bits);
  }
  return result;
}


static unsigned long rb_we21_kbd_readkeybits(void *hw_data)
{
  struct t_rinblock **rbs = hw_data;
  struct s_rb_we21_disp_keybits keybits;
  unsigned long result_keybits = 0UL;

#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_we21_kbd_readkeybits called\n");
#endif

  if (NULL == rbs)
  {
    return 0UL;
  }

  while (NULL != *rbs)
  {
    if (RinBlockWe21GetKeyBits(*rbs, &keybits))
    {
#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_FULL)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_we21_kbd_readkeybits: RinBlockWe21GetKeys(), rb %d success, keybits: 0x%08x\n",
                      (*rbs)->iic_address, keybits.key_bits);
#endif
      result_keybits |= keybits.key_bits;
    }
#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_FULL)
    else
    {
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_we21_kbd_readkeybits: RinBlockWe21GetKeys() rb %d fail\n", (*rbs)->iic_address);
    }
#endif
    rbs++;
  }

#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_we21_kbd_readkeybits: result 0x%08x\n", result_keybits);
#endif
  return result_keybits;
}


/** Destroy the we21 rinblock keyboard */
static void rb_we21_kbd_destroy(void *hw_data)
{
  struct t_rinblock **rbs = hw_data;

#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rb_we21_kbd_destroy called\n");
#endif

  /* Free array of rinblocks */
  if (NULL != rbs)
  {
    free(rbs);
  }
}


/**
 * Create a rinblock keyboard device
 *
 * @param io_dev is an uninitialised io device
 * @param rbs is a pointer to an array of pointers to rinblocks with keyboards
 *            each of which will polled for keyboard data, and ORed together.
 *            The driver keeps it's only copy of the array.
 * @param num_rbs is the number of rinblocks in the array
 * @param dev_buzzer is a buzzer device to produce key beeps. Can be NULL.
 * @param kbd_hw_readhikeybits is a function pointer for reading the high 32 key bits. Can be NULL, in which case IOCTL_KBD_WRITE_HI_KEY_BITS can be used
 */
int RinBlockCreateWE21KbdDevice(struct t_io_dev *io_dev, struct t_rinblock *rbs[], int num_rbs, struct t_io_dev *dev_buzzer, t_kbd_hw_readhikeybits kbd_hw_readhikeybits)
{
  struct s_kbd_hw_definition rb_we21_kbd_hw_definition;
  struct t_rinblock **rbs_copy;
  int i;

#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21KbdDevice called\n");
#endif
  if (NULL == rbs)
  {
#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21KbdDevice: rbs is NULL\n");
#endif
    goto rbcwkd_fail;
  }

  /* Check that the mapped rinblocks are actually a WE2111/C500 display board or connector board */
  for (i = 0; i < num_rbs; i++)
  {
    if ( RB_IFACE_TYPE_WE21_DISP != rbs[i]->tag.interface_type &&
         RB_IFACE_TYPE_WE21_CONN != rbs[i]->tag.interface_type &&
         RB_IFACE_TYPE_C500_DISP != rbs[i]->tag.interface_type)
    {
#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_BASIC)
      RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21KbdDevice: attempting to map keyboard to non-WE2111/C500 display or connector board\n");
#endif
      goto rbcwkd_fail;
    }
  }

  /* Take an internal copy of the array of pointers to keyboard rinblocks */
  rbs_copy = malloc(sizeof(struct t_rinblock *) * (num_rbs + 1));
  if (NULL == rbs_copy)
  {
#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21KbdDevice: rbs malloc failed\n");
#endif
    goto rbcwkd_fail;
  }
  for (i = 0; i < num_rbs; i++)
  {
    rbs_copy[i] = rbs[i];
  }
  rbs_copy[i] = NULL;

  /* Configure the keyboard hardware definition for the rinblock */
  memset(&rb_we21_kbd_hw_definition, 0x00, sizeof(rb_we21_kbd_hw_definition));
  rb_we21_kbd_hw_definition.hw_data = rbs_copy;
  rb_we21_kbd_hw_definition.kbd_hw_init = 0;
  rb_we21_kbd_hw_definition.kbd_hw_readkeybits = rb_we21_kbd_readkeybits;
  rb_we21_kbd_hw_definition.already_debounced = 1;
  rb_we21_kbd_hw_definition.kbd_hw_destroy = rb_we21_kbd_destroy;
  rb_we21_kbd_hw_definition.kbd_hw_readhikeybits = kbd_hw_readhikeybits;

  /* Create keyboard device */
  if (rio_create_device_kbd_local(io_dev, &rb_we21_kbd_hw_definition, dev_buzzer) < 0)
  {
#if RB_DEBUG(RB_KBD_DEBUG_LVL, RB_DEBUG_BASIC)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "RinBlockCreateWE21KbdDevice: device creation failed\n");
#endif
    goto rbcwkd_free;
  }

  /* Success */
  return RIO_SUCCESS;

rbcwkd_free:
  free(rbs_copy);
rbcwkd_fail:
  return RIO_ERROR;
}
