/*
 * lcd_r400.c
 *
 *  Created on: 01/06/2010
 *      Author: ian
 */

#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rio.h>
#include <rinhw/lcd.h>
#include <rinhw/debug.h>
#include <rinhw/lcd_generic.h>
#include <rinhw/rinendian.h>
#include <rinhw/lcd_local_skt.h>
#include <rinhw/kbd.h>
#include <rinblock/rb_msg.h>
#include "_rb_we21_lcd_common.h"
#include <rinblock/rb_debug.h>

#include <rinblock/rb_we21_lcd_local_skt.h>


/*
 * Needs to answer to the following commands
 *
 * S98;STP;S99;COF11; -> 0
 * ENU?; -> 1
 * FPL?; -> "9EFCFCE0FCFC00000000"
 * S99;ESR?; -> 00700
 *
 * Commands:
 * FPL1 - Zero key
 * FPL2 - Tare key
 * FPL3 - Gross/net key
 * FPL4 - Print key
 * FPL5 - Cal key
 * FPL6 - Function key
 *
 * Add ,1 for longpress eg FPL1,1
 *
 */


static void we21_lcd_local_skt_parser_send_reply(int sockfd, const char *reply)
{
  const int len = strlen(reply);

  send(sockfd, reply, len, 0);
  send(sockfd, "\r\n", 2, 0);
}


static void we21_lcd_local_skt_parser_send_lcddata(int outsockfd, const unsigned char *lcd_data, size_t lcd_data_length)
{
  char message[400];
  char str[10];
  size_t i;

  /* Assemble LCD data into a string */
  strncpy(message, "\"", 400);
  for (i = 0; i < lcd_data_length; i++)
  {
    snprintf(str, 10, "%02X", lcd_data[i]);
    strncat(message, str, 400);
  }
  strncat(message, "\"", 400);

  /* Send message to to socket */
  we21_lcd_local_skt_parser_send_reply(outsockfd, message);
}


static bool we21_lcd_local_skt_parser_get_message(const unsigned char *sock_data, size_t sock_data_length, char *outbuf, size_t outbuflen)
{
  size_t i;
  size_t ocnt;
  const unsigned char *p;

#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "we21_lcd_local_skt_parser_get_message called, sock_data_length %lu, outbuflen %lu\n", (unsigned long) sock_data_length, (unsigned long) outbuflen);
#endif

  if (0 == sock_data_length || 0 == outbuflen)
  {
#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "we21_lcd_local_skt_parser_get_message: a length is zero, returning false\n");
#endif
    return false;
  }

  /* Discard any STXs, and copy everything else up CR, LF, ;, ETX into output data buffer */
  for (i = 0, ocnt = 0, p = sock_data; '\x03' != *p && '\r' != *p && '\n' != *p && ';' != *p && i < sock_data_length && ocnt < outbuflen; i++, p++)
  {
    if ('\x02' != *p)
      outbuf[ocnt++] = *p;
  }

  if (i >= sock_data_length || ocnt >= outbuflen)
  {
    /* Didn't find a terminating character */
#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "we21_lcd_local_skt_parser_get_message: didn't find a terminating character, returning false\n");
#endif
    return false;
  }

  /* Finished */
  outbuf[ocnt] = '\0';
#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "we21_lcd_local_skt_parser_get_message: success\n");
#endif
  return true;
}


static bool we21_lcd_local_skt_parser_disassemble_message(const char *message, char *cmd, char *params, size_t params_maxlen, bool *is_query)
{
  int res;
  bool success;
  char ch;

  /* TODO max length */

  /* Is it a query with no params eg COF? */
  res = sscanf(message, "%3s%c", cmd, &ch);
  if (2 == res && '?' == ch)
  {
    success = true;
    *is_query = true;
    strncpy(params, "", params_maxlen);
  }
  else
  {
    /* Is it a query with params eg COF?0,0 */
    res = sscanf(message, "%3s%c%s", cmd, &ch, params);
    if (3 == res && '?' == ch)
    {
      success = true;
      *is_query = true;
    }
    else
    {
      /* Is it a command eg COFxxx? */
      *is_query = false;
      res = sscanf(message, "%3s%s", cmd, params);
      success = (2 == res);
    }
  }
  return success;
}


static void we21_lcd_local_skt_parser_parse_keypress(const char *params, bool *gotkey, union un_key *key)
{
  int intkey;
  int longpress;
  int res;
  char ch;

  /* Try with both params */
  res = sscanf(params, "%d%c%d", &intkey, &ch, &longpress);
  if (3 != res || ',' != ch)
  {
    /* Try with first param */
    longpress = 0;
    res = sscanf(params, "%d", &intkey);
    if (1 != res)
    {
      return;
    }
  }

  *gotkey = true;
  switch (intkey)
  {
    case 1:
      key->BIT.KEY = KEY_PHYSICAL_FUNC0;
      break;
    case 2:
      key->BIT.KEY = KEY_PHYSICAL_FUNC1;
      break;
    case 3:
      key->BIT.KEY = KEY_PHYSICAL_FUNC2;
      break;
    case 4:
      key->BIT.KEY = KEY_PHYSICAL_FUNC3;
      break;
    case 5:
      key->BIT.KEY = KEY_PHYSICAL_SETUP;
      break;
    case 6:
      key->BIT.KEY = KEY_PHYSICAL_FUNC4;
      break;
    default:
      *gotkey = false;
      break;
  }

  if (longpress)
  {
    key->BIT.EVENT = keHeld;
  }
  else
  {
    key->BIT.EVENT = keDown;
  }
}


static void we21_lcd_local_skt_parser_parse_message(int outsockfd, const char *message, const unsigned char *lcd_data, size_t lcd_data_length, bool *gotkey, union un_key *key)
{
  char cmd[10];
  char params[100];
  bool is_query;

  /* Try 3 letter commands */
  if (we21_lcd_local_skt_parser_disassemble_message(message, cmd, params, 100, &is_query))
  {
    if (0 == strcmp(cmd, "COF") && !is_query)
    {
      we21_lcd_local_skt_parser_send_reply(outsockfd, "0");
    }
    else if (0 == strcmp(cmd, "ENU") && is_query)
    {
      we21_lcd_local_skt_parser_send_reply(outsockfd, "1");
    }
    else if (0 == strcmp(cmd, "ESR") && is_query)
    {
      we21_lcd_local_skt_parser_send_reply(outsockfd, "00000");
    }
    else if (0 == strcmp(cmd, "FPL") && is_query)
    {
      /* FPL request lcd data */
      /* we21_lcd_local_skt_parser_send_reply(sockfd, "9EFCFCE0FCFC00000000"); */
      we21_lcd_local_skt_parser_send_lcddata(outsockfd, lcd_data, lcd_data_length);
    }
    else if (0 == strcmp(cmd, "FPL") && !is_query)
    {
      /* FPL keystroke */
      we21_lcd_local_skt_parser_parse_keypress(params, gotkey, key);
    }
  }
}


/** Parse the data off the socket, and reply as required */
static bool we21_lcd_local_skt_parser(int outsockfd, const unsigned char *sock_data, size_t sock_data_length, const unsigned char *lcd_data, size_t lcd_data_length, bool *gotkey, union un_key *key)
{
  #define R400_LCD_BUFFER_LEN 200
  char buffer[R400_LCD_BUFFER_LEN];

#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "we21_lcd_local_skt_parser called\n");
#endif

  /* See if there is a message in the data */
  if (!we21_lcd_local_skt_parser_get_message(sock_data, sock_data_length, buffer, R400_LCD_BUFFER_LEN))
  {
#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_FULL)
    RB_DEBUG_PRINTF(RB_DEBUG_DEV, "we21_lcd_local_skt_parser: no message\n");
#endif
    return false;
  }

  /* Got a message, parse it */
#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "we21_lcd_local_skt_parser: got message, parsing '%s'\n", buffer);
#endif
  we21_lcd_local_skt_parser_parse_message(outsockfd, buffer, lcd_data, lcd_data_length, gotkey, key);
  return true;
}


/** Function to create and initialise a local socket WE2111 LCD device (use the viewer to attach to) */
int rio_create_device_lcd_localskt_we21(struct t_io_dev *io_dev, int sockport, struct t_io_dev *kbddev)
{
  struct s_lcd_hw_socket *we21_lcd_socket;

#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_FULL)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rio_create_device_lcd_localskt_r400 called\n");
#endif

  /* Setup socket */
  we21_lcd_socket = malloc(sizeof(struct s_lcd_hw_socket));
  if (NULL == we21_lcd_socket)
  {
    goto rcdllr_fail;
  }
  we21_lcd_socket->port = sockport;
  we21_lcd_socket->shutdown_pipe_write_fd = -1;
  we21_lcd_socket->shutdown_pipe_read_fd = -1;
  /* r400_lcd_socket->thread = 0; */
  we21_lcd_socket->sem_exclusive = NULL;
  we21_lcd_socket->kbddev = kbddev;
  we21_lcd_socket->data_parser = we21_lcd_local_skt_parser;

  /* Create device */
  if (lcd_generic_segflash_create_device(io_dev,
                                         (void *) we21_lcd_socket,
                                         sizeof(struct s_rb_we21_disp_write_7seg),
                                         WE21_DISPLAY_NUM_DIGITS,
                                         lcd_socket_hardware_init,
                                         lcd_socket_hardware_destroy,
                                         NULL,
                                         rb_we21_lcd_put_char,
                                         rb_we21_lcd_does_digit_have_dp,
                                         NULL) < 0)
  {
    goto rcdllr_fail;
  }
  we21_lcd_socket->sem_exclusive = io_dev->sem_open_excl;

  /* Success */
#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rio_create_device_lcd_localskt_r400 success\n");
#endif
  return RIO_SUCCESS;

  /* Clean up after fail */
rcdllr_fail:
#if RB_DEBUG(RB_LCD_SKT_DEBUG_LVL, RB_DEBUG_BASIC)
  RB_DEBUG_PRINTF(RB_DEBUG_DEV, "rio_create_device_lcd_localskt_r400 failed\n");
#endif
  if (NULL != we21_lcd_socket)
  {
    free(we21_lcd_socket);
    we21_lcd_socket = NULL;
  }
  return RIO_ERROR;
}
