#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <rinhw/rstdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinhw/hw_vers.h>
#include <rinhw/power.h>
#include <rinhw/iic.h>

#include "rb_getnum.h"

#include <rinblock/rinblock.h>
#include <rinblock/rb_serial.h>
#include <rinblock/rb_flash.h>
#include <rinblock/rb_dio.h>
#include <rinblock/rb_startup.h>
#include <rinblock/rb_shutdown.h>
#include <rinblock/rb_aout.h>
#include <rinblock/rb_tag.h>
#include <rinblock/rb_address.h>
#include <rinblock/rb_serialno.h>
#include <rinblock/rb_we21_adc.h>
#include <rinblock/rb_we21_buzzer.h>
#include <rinblock/rb_we21_kbd.h>
#include <rinblock/rb_we21_lcd.h>
#include <rinblock/rb_we21_lcd_local_skt.h>
#include <rinblock/rb_c500_lcd.h>

#include "rb_serial_test.h"
#include "rb_dio_test.h"
#include "rb_aout_test.h"
#include "rb_aout_calibrate.h"
#include "rb_scratch_test.h"
#include "rb_program_test.h"
#include "rb_status_test.h"
#include "rb_dsd_test.h"
#include "rb_ethernet_test.h"
#include "rb_tilt_test.h"

#include "rb_we21_adc_test.h"
#include "rb_we21_buzzer_test.h"
#include "rb_we21_kbd_test.h"
#include "rb_we21_lcd_test.h"

#include "config.h"

#include "devices.h"


/** Display command line help */
static bool cmdline_showhelp = false;


/** Serial port device to use for remote devices, if NULL remote devices are not used */
static const char *cmdline_remote_device = NULL;


/** I2C device to use */
static const char *cmdline_i2c_device = "/dev/i2c-0";


/* Define TEST_DEV */
#define TEST_DEV dev_console


/*
 * Serial ports on rinblocks
 */
struct t_io_dev dev_rb_serials[MAX_RINBLOCKS][2];


/*
 * Test rinblock dsd device
 */
struct t_rinblock dev_rb_dsd1;

static int display_rinblock_index = 0;
static bool display_rinblock_lcd_device_is_c500 = false;


unsigned char GetBlockChar(struct t_io_dev *dev)
{
  unsigned char ch;

  rread(dev, &ch, 1, 0);
  return ch;
}


static unsigned char GetCurrentRb(struct t_io_dev *dev)
{
  unsigned char ch;

  do
  {
    rfprintf(dev, "Enter new rinblock (0..%d): ", MAX_RINBLOCKS - 1);
    rread(dev, &ch, 1, 0);
    rfprintf(dev, "%c\r\n", ch);
    ch -= '0';
  }
  while (ch >= MAX_RINBLOCKS);
  return ch;
}


#if 0
static unsigned long GetLong(struct t_io_dev *dev)
{
  unsigned char ch;
  unsigned long n = 0;

  rfprintf(dev, "Enter number: ");
  do
  {
    rread(dev, &ch, 1, 0);
    if ((ch >= '0') && (ch <= '9'))
    {
      rwrite(dev, &ch, 1, 0);
      n *= 10;
      n += (ch - '0');
    }
  }
  while (ch != 0x0A);

  rfprintf(dev, "\r\n");
  return n;
}
#endif


static void SetSerialNumber(struct t_rinblock *rb, struct t_io_dev *dev)
{
  long num;
  unsigned long serial;
  unsigned short hw_version;
  unsigned char res;

  rfprintf(dev, "\r\nSet Serial Number and HW Version Number\r\n");

  rfprintf(dev, "Enter new serial number:");
  get_long(dev, &num);
  serial = num;

  rfprintf(dev, "\r\nEnter new hardware version number:");
  get_long(dev, &num);
  hw_version = (unsigned short) num;
  rfprintf(dev, "\r\n");

  res = RinBlockSetTagShite(rb, serial, hw_version);

  rfprintf(dev, "Set serial number to %ld, hardware version to %d %s\r\n", serial, hw_version, ((res) ? "success" : "failed"));
}


static const unsigned char rinblock_prog_str[] = "FLASHBLOCK";

#define TIMEOUT_PRESET 10

static void RinBlockProgramAtStartup(struct t_io_dev *serial_dev)
{
  unsigned char ch;
  const unsigned char *ptr = rinblock_prog_str;
  unsigned char timeout = TIMEOUT_PRESET;
  unsigned short lvl;
  t_ioctl_siocfg siocfg_old, siocfg;

  /*
   * Wait for the TX buf to empty
   */
  do
  {
    rioctl(serial_dev, IOCTL_SER_GET_TX_BUF_LEVEL, &lvl);
    OSTimeDly(1);
  }
  while (lvl > 0);

  rioctl(serial_dev, IOCTL_SER_GET_CONFIG, &siocfg_old);

  siocfg.baud = 9600UL;
  siocfg.parity_bits = SIO_PARITY_NONE;
  siocfg.data_bits = 8;
  siocfg.stop_bits = 1;
  siocfg.use_cts = 0;
  siocfg.use_terminations = 0;
  rioctl(serial_dev, IOCTL_SER_SET_CONFIG, &siocfg);

  while (timeout > 0)
  {
    if (rread(serial_dev, &ch, 1, OS_TICKS_PER_SEC / TIMEOUT_PRESET) == 1)
    {
      if (*ptr == '\0')
      {
        if ((ch >= '0') && (ch < (MAX_RINBLOCKS + '0')))
        {
          RinBlockProgram(&rinblocks[ch - '0'], serial_dev, (struct t_wdt *) NULL, &ch);
          break;
        }
        else
          ptr = rinblock_prog_str;
      }
      else if (*ptr == ch)
      {
        ptr++;
      }
      else
        ptr = rinblock_prog_str;
    }
    else
    {
      timeout--;
    }
  }

  rioctl(serial_dev, IOCTL_SER_SET_CONFIG, &siocfg_old);
}


static void RinBlockStartup(void)
{

  /*
   * Initialise the rin block handling code
   */
  rfprintf(&TEST_DEV, "Starting Rinblock Task\r\n");
  RinBlockInit(&dev_iic0);

  /*
   * Assign addresses to each module on the bus
   */
  rfprintf(&TEST_DEV, "Assign Rinblock Addresses\r\n");
  RinBlockEnumerate();

  /*
   * Look for someone wanting to program a module here
   */
  rfprintf(&TEST_DEV, "Check for programming\r\n");
  RinBlockProgramAtStartup(&TEST_DEV);

  /*
   * Start rin blocks
   */
  rfprintf(&TEST_DEV, "Starting Rinblocks\r\n");
  RinBlockStart();
}


void RinBlockStatus(struct t_rinblock *rb)
{
  rfprintf(&TEST_DEV, "RinBlock Status: iic_dev: 0x%p, addr %d, slot %d, present %d, programmed %d, running %d\r\n",
           rb->iicdev, rb->iic_address, rb->bus_slot, rb->present, rb->programmed, rb->running);
}


static void RinBlockPrintTag(struct s_rb_general_tag *tag)
{
  rfprintf(&TEST_DEV, "Rinblock TAG: Model %04x, Serial number: %ld, Software Version: %04x, ",
           tag->model_number, tag->serial_number, tag->software_version);
  rfprintf(&TEST_DEV, "Interface Version: %04x, Interface Type: %04x, Hardware Version: %04x\r\n",
           tag->interface_version, tag->interface_type, tag->hardware_version);
}


static void RinBlockAssignDevices(void)
{
  int i, res;
  int8u adcdone = 0, lcddone = 0;
  struct t_rinblock *kbd_rbs[MAX_RINBLOCKS + 1];
  int kbd_ctr;

  for (i = 0, kbd_ctr = 0; i < MAX_RINBLOCKS; i++)
  {
    /* ADC Device */
    if (RB_IFACE_TYPE_WE21_CONN == rinblocks[i].tag.interface_type && !adcdone)
    {
      /* Replace dev_adc with the rinblock one */
      if (RIO_SUCCESS == RinBlockCreateWE21AdcDevice(&dev_adc, &rinblocks[i]))
      {
        rfprintf(&TEST_DEV, "Replaced ADC device with WE2111 rinblock %d ADC\r\n", i);
        adcdone = 1;
      }

      /* This one is a keyboard */
      kbd_rbs[kbd_ctr++] = &rinblocks[i];
    }
    /* Serial Device */
    if (RB_IFACE_TYPE_SERIAL == rinblocks[i].tag.interface_type ||
         RB_IFACE_TYPE_ETHERNET == rinblocks[i].tag.interface_type)
    {
      res = (RIO_SUCCESS == RinBlockCreateSerialDevice(&dev_rb_serials[i][0], &rinblocks[i], rbsc_a) &&
             RIO_SUCCESS == RinBlockCreateSerialDevice(&dev_rb_serials[i][1], &rinblocks[i], rbsc_b));
      if (res)
      {
        rfprintf(&TEST_DEV, "Created Rin Block serial device with result %d on rinblock %d\r\n", res, i);
      }
    }

    /* WE2111 LCD Device */
    if (RB_IFACE_TYPE_WE21_DISP == rinblocks[i].tag.interface_type && !lcddone)
    {
      /* Create dev_lcd with the WE2111 display rinblock */
      if (RIO_SUCCESS == RinBlockCreateWE21LcdDeviceSegFlash(&dev_lcd, &rinblocks[i]))
      {
        rfprintf(&TEST_DEV, "Created LCD device for WE2111 rinblock %d LCD\r\n", i);
        lcddone = 1;
      }
      /* Create dev_buz with the WE2111 display rinblock */
      if (RIO_SUCCESS == RinBlockCreateWE21BuzzerDevice(&dev_buz, &rinblocks[i]))
      {
        rfprintf(&TEST_DEV, "Created buzzer device for WE2111 rinblock %d buzzer\r\n", i);
      }

      /* This one is a keyboard */
      kbd_rbs[kbd_ctr++] = &rinblocks[i];
      display_rinblock_index = i;
    }

    /* C500 LCD Device */
    if (RB_IFACE_TYPE_C500_DISP == rinblocks[i].tag.interface_type && !lcddone)
    {
      /* Create dev_lcd with the C500 display rinblock */
      if (RIO_SUCCESS == RinBlockCreateC500LcdDeviceSegFlash(&dev_lcd, &rinblocks[i]))
      {
        rfprintf(&TEST_DEV, "Created LCD device for C500 rinblock %d LCD\r\n", i);
        lcddone = 1;
        display_rinblock_lcd_device_is_c500 = true;
      }
      /* Create dev_buz with the C500 display rinblock */
      if (RIO_SUCCESS == RinBlockCreateWE21BuzzerDevice(&dev_buz, &rinblocks[i]))
      {
        rfprintf(&TEST_DEV, "Created buzzer device for C500 rinblock %d buzzer\r\n", i);
      }
      /* Create rinblock serial for front rinlink */
      res = (RIO_SUCCESS == RinBlockCreateSerialDevice(&dev_rb_serials[i][0], &rinblocks[i], rbsc_a));
      if (res)
      {
        rfprintf(&TEST_DEV, "Created Rin Block serial device with result %d on rinblock %d\r\n", res, i);
      }

      /* This one is a keyboard */
      kbd_rbs[kbd_ctr++] = &rinblocks[i];
      display_rinblock_index = i;
    }
  }

  /* If we found a device that has a keyboard */
  if (kbd_ctr > 0)
  {
    /* Destroy existing keyboard */
    /* rdestroy_io(&dev_kbd); */

    /* Replace dev_kbd with the rinblock one */
    if (RIO_SUCCESS == RinBlockCreateWE21KbdDevice(&dev_kbd, kbd_rbs, kbd_ctr, &dev_buz, NULL))
    {
      rfprintf(&TEST_DEV, "Created WE2111/C500 keyboard device for %d rinblocks\r\n", kbd_ctr);
    }
  }

  /* Create a local socket WE2111 LCD device (use the viewer to attach to) */
  if (rio_create_device_lcd_localskt_we21(&dev_lcd_local_skt, 1234, &dev_kbd) < 0)
  {
    rfprintf(&TEST_DEV, "Failed to create local socket LCD\r\n");
  }
}

static void RinBlockDestroyDevices(void)
{
  size_t i;

  rdestroy_io(&dev_lcd_local_skt);
  rdestroy_io(&dev_kbd);
  rdestroy_io(&dev_buz);

  for (i = 0; i < MAX_RINBLOCKS; i++)
  {
    if (0 != (dev_rb_serials[i][0].status & O_INITIALISED))
      rdestroy_io(&dev_rb_serials[i][0]);
    if (0 != (dev_rb_serials[i][1].status & O_INITIALISED))
      rdestroy_io(&dev_rb_serials[i][1]);
  }

  rdestroy_io(&dev_adc);
}


/**
 * If a C500 display is attached, we can talk to it as either a C500 LCD or a WE2111 LCD.
 * This function switches between the two
 */
static void switch_c500_lcd_devices(void)
{
  if (RB_IFACE_TYPE_C500_DISP == rinblocks[display_rinblock_index].tag.interface_type)
  {
    rdestroy_io(&dev_lcd);
    if (display_rinblock_lcd_device_is_c500)
    {
      /* Was a C500, make it a WE2111 */
      if (RIO_SUCCESS == RinBlockCreateWE21LcdDeviceSegFlash(&dev_lcd, &rinblocks[display_rinblock_index]))
        rfprintf(&TEST_DEV, "Switched to WE2111 LCD device for C500 rinblock %d\n", display_rinblock_index);
      else
        rfprintf(&TEST_DEV, "Failed to create WE2111 LCD device for C500 rinblock %d\n", display_rinblock_index);
      display_rinblock_lcd_device_is_c500 = false;
    }
    else
    {
      /* Was a WE2111, make it a C500 */
      if (RIO_SUCCESS == RinBlockCreateC500LcdDeviceSegFlash(&dev_lcd, &rinblocks[display_rinblock_index]))
        rfprintf(&TEST_DEV, "Switched to C500 LCD device for C500 rinblock %d\n", display_rinblock_index);
      else
        rfprintf(&TEST_DEV, "Failed to create C500 LCD device for C500 rinblock %d\n", display_rinblock_index);
      display_rinblock_lcd_device_is_c500 = true;
    }
  }
  else
  {
    rfprintf(&TEST_DEV, "Rinblock %d is not a C500 display. Not switching LCD type.", display_rinblock_index);
  }
}


void test_debug_printf_fn(int priority, const char *format, va_list args)
{
  rvfprintf(&TEST_DEV, format, args);
}


static void menu(void)
{
  unsigned char ch;
  unsigned char current_rb;
  int16u ifacetype;

  if (TEST_DEV.status & O_OPEN)
    ropen(&TEST_DEV, 0);

  rfprintf(&TEST_DEV, "\r\n\r\nRin Block Testing.  Build:" __DATE__ " " __TIME__ "\r\n");
  rfprintf(&TEST_DEV, "librinhw Version: %s\r\n", HwLibDrawingNumberStr);

  /* Startup rin blocks */
  RinBlockStartup();

#if (TARGET == WE21 && defined(RINBLOCK_DEBUG) && RB_DEBUG_APP_CALLBACK)
  rb_set_debug_printf_fn(test_debug_printf_fn);
#endif

  /* Assign devices to rinblocks */
  RinBlockAssignDevices();

  ch = '\0';
  current_rb = 0;
  while ('Q' != ch)
  {
    rfprintf(&TEST_DEV, "\r\nCurrent Rinblock: %d\r\n", current_rb);
    rfprintf(&TEST_DEV,
             "A - Get Address\r\n"
             "N - Set Serial Number\r\n"
             "C - Change current rb\r\n"
             "R - Restart Rinblock\r\n"
             "r - Restart R400\r\n"
             "s - Rinblock Status\r\n"
             "G - Print Tag\r\n"
             "f - Test Program Rinblock\r\n"
/*             "P - Program Rinblock\r\n" Was R400 */
             "D - DIO Test\r\n"
             "S - SIO Test\r\n"
             "T - Scratchpad Test\r\n"
             "O - Analog Output Test\r\n"
             "o - Analog Output Manual Calibration\r\n"
             "t - Rinblock Status\r\n"
             "d - DSD Test\r\n"
             "E - Ethernet Test\r\n"
             "a - WE2111/C500 Rinblock ADC Test\r\n"
             "b - WE2111/C500 Rinblock Buzzer Test\r\n"
             "k - WE2111/C500 Rinblock Keyboard Test\r\n"
             "l - WE2111/C500 Rinblock LCD Test\r\n"
             "m - Switch between WE2111 & C500 LCD for C500 Rinblock display only\r\n"
             "L - WE2111 Local Socket LCD Test\r\n"
             "i - Tilt Module Test\r\n"
             "Q - Quit\r\n"
            );

    ch = GetBlockChar(&TEST_DEV);
    switch (ch)
    {
      case 'A':
        RinBlockGetAddress(&rinblocks[current_rb], &ch);
        rfprintf(&TEST_DEV, "Address is %d\r\n", ch);
        break;
      case 'N':
        SetSerialNumber(&rinblocks[current_rb], &TEST_DEV);
        break;
      case 'C':
        current_rb = GetCurrentRb(&TEST_DEV);
        break;
#if 0 /* Was R400 */
      case 'P':
        rfprintf(&TEST_DEV, "Program Running Rinblock\r\n");
        RinBlockSerialProgram(&TEST_DEV, &rinblocks[current_rb]);
        break;
#endif
      case 'R':
        rfprintf(&TEST_DEV, "Restart rinblock\r\n");
        RinBlockRestart(&rinblocks[current_rb]);
        break;
      case 'r':
        rfprintf(&TEST_DEV, "Restart R400\r\n");
        SystemReset();
        break;
      case 'D':
        rfprintf(&TEST_DEV, "DIO Test\r\n");
        dio_test(&TEST_DEV);
        break;
      case 'S':
        serial_test(&TEST_DEV, &dev_rb_serials[current_rb][0], &dev_rb_serials[current_rb][1]);
        break;
      case 's':
        RinBlockStatus(&rinblocks[current_rb]);
        break;
      case 'T':
        scratchpad_test(&TEST_DEV, &rinblocks[current_rb]);
        break;
      case 'O':
        rfprintf(&TEST_DEV, "Analog Output Test\r\n");
        aout_test(&TEST_DEV);
        break;
      case 'o':
        aout_manual_calibrate(&TEST_DEV);
        break;
      case 'f':
        rfprintf(&TEST_DEV, "Test Code Program Rin Block\r\n");
        testprogram_test(&TEST_DEV, &rinblocks[current_rb]);
        break;
      case 't':
        status_test(&TEST_DEV, &rinblocks[current_rb]);
        break;
      case 'G':
        RinBlockPrintTag(&rinblocks[current_rb].tag);
        break;
      case 'd':
        dsd_test(&TEST_DEV, &dev_rb_dsd1);
        break;
      case 'E':
        ethernet_test(&TEST_DEV, &rinblocks[current_rb]);
        break;
      case 'a':
        rb_adc_test(&TEST_DEV, &dev_adc);
        break;
      case 'b':
        rb_buzzer_test(&TEST_DEV, &dev_buz);
        break;
      case 'k':
        rb_keyboard_test(&TEST_DEV, &dev_kbd);
        break;
      case 'l':
        ifacetype = (display_rinblock_lcd_device_is_c500) ? RB_IFACE_TYPE_C500_DISP : RB_IFACE_TYPE_WE21_DISP;
        rb_lcd_test(&TEST_DEV, &dev_lcd, ifacetype);
        break;
      case 'm':
        switch_c500_lcd_devices();
        break;
      case 'L':
        rb_lcd_test(&TEST_DEV, &dev_lcd_local_skt, rinblocks[display_rinblock_index].tag.interface_type);
        break;
      case 'i':
        tilt_test(&TEST_DEV, &rinblocks[current_rb]);
        break;
    }
  }
  rclose(&TEST_DEV);
  RinBlockDestroyDevices();
  RinBlockDestroy();
}


/** Print out the command line options */
static void print_commandlineopts(const char *program_name)
{
  fprintf(stdout, "Usage: %s [-options]\n"
                  "  -h           --help             Print help\n"
                  "  -r <serial>  --remote <serial>  Use remote devices over <serial> port\n"
                  "                                  (/dev/ttySx)\n"
                  "  -i <i2c>     --i2c <i2c>        Use <i2c> as local I2C device\n"
                  "                                  (/dev/i2c-n or ftdi to use libmpsse/libftdi/libusb)\n",
          program_name);
}


/** Parse the command line options */
static int parse_commandline(int argc, char *argv[])
{
  int c;
  int result = 0;
  static const char opts[] = "hr:i:";
  static const struct option lopts[] =
  {
    { "help",          0, 0, 'h'  },
    { "remote",        1, 0, 'r'  },
    { "i2c",           1, 0, 'i'  },
    { NULL,            0, 0, '\0' }
  };

  while (1)
  {
    c = getopt_long(argc, argv, opts, lopts, NULL);
    if (-1 == c)
    {
      break;
    }

    switch (c)
    {
      case 'h':
        cmdline_showhelp = true;
        break;
      case 'r':
        cmdline_remote_device = optarg;
        break;
      case 'i':
        cmdline_i2c_device = optarg;
#if (!HAVE_LIBMPSSE)
        if (strcmp(cmdline_i2c_device, LIBMPSSE_FTDI_DEVICE_NAME) == 0)
        {
          fprintf(stderr, "librinhw has not been built with libmpsse support. Run ./configure --help\n");
          return -1;
        }
#endif
        break;
      default:
        return -1;
        break;
    }
  }
  return result;
}


/** Entry point */
int main(int argc, char *argv[])
{
  int result;

  /* Parse command line */
  if (parse_commandline(argc, argv) < 0)
  {
    fprintf(stderr, "Invalid command line, try --help\n");
    result = 1;
    goto m_fail;
  }

  /* Show help if requested */
  if (cmdline_showhelp)
  {
    print_commandlineopts(argv[0]);
    result = 0;
    goto m_fail;
  }

  /* Create devices */
  if (!init_test_app_devices(cmdline_remote_device, cmdline_i2c_device))
  {
    fprintf(stderr, "Create devices failed\n");
    result = 1;
    goto m_fail;
  }

  menu();

  destroy_test_app_devices();
  result = 0;

m_fail:
  return result;
}
