/*
 * rb_aout_cal.c
 *
 *  Created on: 02/10/2014
 *      Author: IanB
 */
#include <rinhw/stdrinstrum.h>
#include <rinhw/rio.h>
#include <rinhw/sio.h>
#include <rinhw/rstdio.h>
#include <rinblock/rb_aout.h>
#include <rinblock/rb_dio.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_aout_cal.h>
#include <rinblock/crc.h>

#include "rb_aout_calibrate.h"


#define AOUT_UPDATE_RATE_HZ 5
#define AOUT_TIMEOUT_PERIOD (OS_TICKS_PER_SEC / (AOUT_UPDATE_RATE_HZ * 2))
#define AOUT_RINBLOCK 3
#define AOUT_FORCE_LOW 0
#define AOUT_FORCE_HIGH 50000

#define AOUT_CAL_INCREMENT_SMALL 1
#define AOUT_CAL_INCREMENT_LARGE 100


/* Stolen from M4401 module code */
#define AOUT_DEFAULT_FACTORY_CAL_0V    6827UL
#define AOUT_DEFAULT_FACTORY_CAL_10V  34133UL
#define AOUT_DEFAULT_FACTORY_CAL_4mA  10922UL
#define AOUT_DEFAULT_FACTORY_CAL_20mA 54613UL


enum eAOCalStage { aocsZeroVolt = 0, aocsTenVolt, aocsFourMA, aocsTwentyMA, aocsNum };

static const char *const aout_cal_stages[aocsNum + 1] = { "0 Volt", "10 Volt", "4 Milliamp", "20 Milliamp", "Finished" };
static const char *const aout_cal_multimeter[aocsNum + 1] = { "voltage", "voltage", "current", "current", NULL };
static const long aout_cal_forcevalue[aocsNum + 1] = { AOUT_FORCE_LOW, AOUT_FORCE_HIGH, AOUT_FORCE_LOW, AOUT_FORCE_HIGH, AOUT_FORCE_LOW };


static struct t_io_dev *aout_debug_dev = NULL;


static int16u calculate_analog_cal_checksum(const struct s_rb_aout_cal *cal)
{
  return CRC16_Calc((unsigned char *) cal, sizeof(struct s_rb_aout_cal) - sizeof(cal->uiCRCValue), 0);
}


/** Write the calibration to a rinblock */
static void rinblock_aout_set_calibration(struct t_rinblock *rb, const struct s_rb_aout_cal *aout_cal)
{
  struct s_rb_aout_cal aout_cal_chksummed;

  aout_cal_chksummed = *aout_cal;
  aout_cal_chksummed.uiCRCValue = calculate_analog_cal_checksum(&aout_cal_chksummed);

  if (!RinBlockAOutWriteFactoryCal(rb, &aout_cal_chksummed))
  {
    rfprintf(aout_debug_dev, "ERROR: write analog module calibration failed\n");
  }
}


/** Write the configuration to a rinblock */
static void rinblock_aout_set_configuration(struct t_rinblock *rb, const struct s_rb_aout_configure *aout_cfg)
{
  if (!RinBlockAOutWriteConfig(rb, aout_cfg))
  {
    rfprintf(aout_debug_dev, "ERROR: write analog module configuration failed\n");
  }
}


static void rinblock_aout_configure_for_calibrating(struct t_rinblock *rb, enum eAOCalStage stage)
{
  struct s_rb_aout_configure aout_cfg;

  aout_cfg.ucAnalogOutputType = (aocsFourMA == stage || aocsTwentyMA == stage) ? 0 : 1;   /* CURRENT = 0, VOLTAGE = 1 */
  aout_cfg.iUserZero = 0;
  aout_cfg.iUserSpan = 0;
  aout_cfg.cMaxLimit = INT8S_MAX;
  aout_cfg.cMinLimit = INT8S_MIN;
  aout_cfg.uiUpdateRate = AOUT_UPDATE_RATE_HZ * 10;

  rinblock_aout_set_configuration(rb, &aout_cfg);
}


static void rinblock_aout_print_calibration(const struct s_rb_aout_cal *aout_cal)
{
  unsigned short calculated_crc;

  rfprintf(aout_debug_dev, "Factory Cal 0V: %u\n", aout_cal->uiFactoryCal0V);
  rfprintf(aout_debug_dev, "Factory Cal 10V: %u\n", aout_cal->uiFactoryCal10V);
  rfprintf(aout_debug_dev, "Factory Cal 4mA: %u\n", aout_cal->uiFactoryCal4mA);
  rfprintf(aout_debug_dev, "Factory Cal 20mA: %u\n", aout_cal->uiFactoryCal20mA);
  rfprintf(aout_debug_dev, "Factory Cal CRC: 0x%04x\n", aout_cal->uiCRCValue);

  calculated_crc = calculate_analog_cal_checksum(aout_cal);
  rfprintf(aout_debug_dev, "Calculated Cal CRC (should match Factory Cal CRC): 0x%04x\n", calculated_crc);
}


static void aout_cal_adjust(enum eAOCalStage stage, unsigned char adjust, struct t_rinblock *rb)
{
  long increment;
  struct s_rb_aout_cal aout_cal;
  int16u *ptr;

  /* Work out increment */
  switch (adjust)
  {
    case 'u':
      increment = AOUT_CAL_INCREMENT_SMALL;
      break;
    case 'd':
      increment = -AOUT_CAL_INCREMENT_SMALL;
      break;
    case 'U':
      increment = AOUT_CAL_INCREMENT_LARGE;
      break;
    case 'D':
      increment = -AOUT_CAL_INCREMENT_LARGE;
      break;
    default:
      return;
  }

  /* Read current calibration */
  if (RinBlockAOutReadFactoryCal(rb, &aout_cal) == 0)
  {
    rfprintf(aout_debug_dev, "Read current calibration failed\n");
    return;
  }

  /* Get the value to adjust */
  switch (stage)
  {
    case aocsZeroVolt:
      ptr = &aout_cal.uiFactoryCal0V;
      break;
    case aocsTenVolt:
      ptr = &aout_cal.uiFactoryCal10V;
      break;
    case aocsFourMA:
      ptr = &aout_cal.uiFactoryCal4mA;
      break;
    case aocsTwentyMA:
      ptr = &aout_cal.uiFactoryCal20mA;
      break;
    default:
      return;
  }

  /* Adjust it, keeping within limits */
  if (increment >= 0 && *ptr > INT16U_MAX - increment)
    increment = INT16U_MAX - *ptr;
  else if (increment < 0 && *ptr < -increment)
    increment = -*ptr;
  *ptr += increment;

  /* Write it back to module */
  rinblock_aout_set_calibration(rb, &aout_cal);
}


/** Set module to default calibration */
static void aout_default_calibration(struct t_rinblock *rb)
{
  struct s_rb_aout_cal aout_cal;

  rfprintf(aout_debug_dev, "Setting analog calibration to defaults\n");

  aout_cal.uiFactoryCal0V = AOUT_DEFAULT_FACTORY_CAL_0V;
  aout_cal.uiFactoryCal10V = AOUT_DEFAULT_FACTORY_CAL_10V;
  aout_cal.uiFactoryCal4mA = AOUT_DEFAULT_FACTORY_CAL_4mA;
  aout_cal.uiFactoryCal20mA = AOUT_DEFAULT_FACTORY_CAL_20mA;

  rinblock_aout_set_calibration(rb, &aout_cal);
}


static void aout_write_value(struct t_rinblock *rb, enum eAOCalStage stage)
{
  if (!RinBlockAOutWrite(rb, aout_cal_forcevalue[stage]))
    rfprintf(aout_debug_dev, "Write to AOUT rinblock with value %ld failed\n", aout_cal_forcevalue[stage]);
}


static void aout_get_number(struct t_io_dev *debug_dev, char *numstr, int numstrlen, struct t_rinblock *rb, enum eAOCalStage stage)
{
  unsigned char ch;
  int posn = 0;

  for (posn = 0; posn < numstrlen - 1; )
  {
    if (rread(debug_dev, &ch, 1, AOUT_TIMEOUT_PERIOD) > 0)
    {
      if ((ch >= '0' && ch <= '9') || '-' == ch || '.' == ch)
      {
        numstr[posn++] = ch;
        rwrite(debug_dev, &ch, 1, 0);
      }
      else if ('\x08' == ch || '\x7F' == ch)
      {
        if (posn > 0)
        {
          posn--;
          ch = '\x08';
          rwrite(debug_dev, &ch, 1, 0);
        }
      }
      else if ('\r' == ch || '\n' == ch)
        break;
    }

    /* Keep the analog output value going */
    aout_write_value(rb, stage);
  }
  numstr[posn] = '\0';
  rfprintf(debug_dev, "\n");
}


#define MULTIMETER_NUM_LEN 20


/** Run a manual calibration on an analog rinblock */
void aout_manual_calibrate(struct t_io_dev *debug_dev)
{
  unsigned char ch = 0;
  enum eAOCalStage stage;
  bool print_stage;
  struct t_rinblock *rb = &rinblocks[AOUT_RINBLOCK];
  struct s_rb_aout_cal aout_cal_initial;
  struct s_rb_aout_cal aout_cal_final;
  char multimeter_initial_nums[aocsNum][MULTIMETER_NUM_LEN];
  char multimeter_final_nums[aocsNum][MULTIMETER_NUM_LEN];

  aout_debug_dev = debug_dev;

  rfprintf(debug_dev, "Analog Output Manual Calibration of Rinblock %d\n", AOUT_RINBLOCK);
  if (!rb->running || !(RB_IFACE_TYPE_AOUT == rb->tag.interface_type) ||(RB_IFACE_TYPE_2AOUT == rb->tag.interface_type) ||
      (RB_IFACE_TYPE_3AOUT == rb->tag.interface_type) || (RB_IFACE_TYPE_4AOUT == rb->tag.interface_type))
  {
    rfprintf(debug_dev, "Rinblock %d is not running, or is not an analog rinblock\n", AOUT_RINBLOCK);
    return;
  }

  if (!RinBlockAOutReadFactoryCal(rb, &aout_cal_initial))
  {
    rfprintf(aout_debug_dev, "ERROR: read analog module calibration failed\n");
    return;
  }

  rfprintf(debug_dev, "Analog Rinblock Manual Calibration\n"
                      "Commands:\n"
                      "  u - Adjust Calibration Up Slow\n"
                      "  d - Adjust Calibration Down Slow\n"
                      "  U - Adjust Calibration Up Fast\n"
                      "  D - Adjust Calibration Down Fast\n"
                      "  N - Next Calibration Item\n"
                      "  P - Previous Calibration Item\n"
                      "  F - Set Calibration To Defaults\n\n");

  for (stage = aocsZeroVolt, print_stage = true; stage < aocsNum; )
  {
    if (print_stage)
    {
      rinblock_aout_configure_for_calibrating(rb, stage);
      rfprintf(debug_dev, "\n\nCurrent calibration stage is %s\n", aout_cal_stages[stage]);
      if (NULL != aout_cal_multimeter[stage])
      {
        rfprintf(debug_dev, "Connect multimeter to module %s connections\n", aout_cal_multimeter[stage]);
        rfprintf(debug_dev, "Enter Initial Multimeter Reading: ");
        aout_get_number(debug_dev, multimeter_initial_nums[stage], MULTIMETER_NUM_LEN, rb, stage);
      }
      rfprintf(debug_dev, "Adjust calibration now\n");
      print_stage = false;
    }

    if (rread(debug_dev, &ch, 1, AOUT_TIMEOUT_PERIOD) >= 0)
    {
      switch (ch)
      {
        case 'u':
        case 'd':
        case 'U':
        case 'D':
          aout_cal_adjust(stage, ch, rb);
          break;
        case 'N':
          rfprintf(debug_dev, "Enter Final Multimeter Reading: ");
          aout_get_number(debug_dev, multimeter_final_nums[stage], MULTIMETER_NUM_LEN, rb, stage);
          stage++;
          print_stage = true;
          break;
        case 'P':
          if (stage > 0)
            stage--;
          print_stage = true;
          break;
        case 'F':
          aout_default_calibration(rb);
          break;
      }
    }
    aout_write_value(rb, stage);
  }

  if (!RinBlockAOutReadFactoryCal(rb, &aout_cal_final))
  {
    rfprintf(aout_debug_dev, "ERROR: read analog module calibration failed\n");
    return;
  }

  rfprintf(debug_dev, "\n\n\nRinblock Serial Number %lu:\n", rb->tag.serial_number);
  rfprintf(debug_dev, "Analog Output Initial Calibration:\n");
  rinblock_aout_print_calibration(&aout_cal_initial);
  for (stage = 0; stage < aocsNum; stage++)
    rfprintf(debug_dev, "Rinblock Measured Initial %s Reading: %s\n", aout_cal_stages[stage], multimeter_initial_nums[stage]);

  rfprintf(debug_dev, "Analog Output Final Calibration:\n");
  rinblock_aout_print_calibration(&aout_cal_final);
  for (stage = 0; stage < aocsNum; stage++)
    rfprintf(debug_dev, "Rinblock Measured Final %s Reading: %s\n", aout_cal_stages[stage], multimeter_final_nums[stage]);
}
