#include <string.h>
#include <rinhw/rio.h>
#include <rinhw/rstdio.h>
#include <rinblock/rb_msg.h>
#include <rinblock/rb_dsd.h>

#include "rb_dsd_test.h"


static unsigned char get_char(struct t_io_dev *debug_dev)
{
    unsigned char ch;

    rread(debug_dev, &ch, 1, 0);
    return ch;
}


static unsigned char get_1_digit_num(struct t_io_dev *debug_dev, unsigned char *num)
{
    unsigned char ch;

    ch = get_char(debug_dev);
    if ((ch < '0') || (ch > '9'))
        return 0;
    *num = (unsigned char) (ch - '0');
    return 1;
}

static unsigned char get_u_short(struct t_io_dev *debug_dev, unsigned short *num)
{
    unsigned char ch;

    *num = 0;
    do
    {
        ch = get_char(debug_dev);
        if ((ch >= '0') && (ch <= '9'))
        {
            *num *= 10;
            *num += (unsigned char) (ch - '0');
            rwrite(debug_dev, &ch, 1, 0);
        }
    }
    while (ch != 0x0A);
    return 1;
}

static unsigned char get_u_long(struct t_io_dev *debug_dev, int32u *num)
{
    unsigned char ch;

    *num = 0;
    do
    {
        ch = get_char(debug_dev);
        if ((ch >= '0') && (ch <= '9'))
        {
            *num *= 10;
            *num += (unsigned char) (ch - '0');
            rwrite(debug_dev, &ch, 1, 0);
        }
    }
    while (ch != 0x0A);
    return 1;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
/*
 * A read-control line looks like
 *
 * !2    // !2=rinblock 2
 */
static void dsd_read_control(struct t_io_dev *debug_dev, struct t_rinblock *rb)
{
    unsigned char ch, rb_num;
    unsigned char result;
    rb_dsd_directory_t dsdDirectory;

    memset(&dsdDirectory, 0, sizeof(dsdDirectory));
    rfprintf(debug_dev, "Read DSD Dir: Enter:  !rbNum ");

    do
    {
        ch = get_char(debug_dev);
        if (ch == '#')
            return;
    }
    while (ch != '!');

    /*
     * Read the rin block number
     */
    if (!get_1_digit_num(debug_dev, &rb_num))
        return;

    rfprintf(debug_dev, "Read DSD Dir @ rbNum:%u ... ", rb_num);
    result = RinBlockReadDSDDirectory(&rinblocks[rb_num], &dsdDirectory);
    rfprintf(debug_dev, "res=%u st=%u t=%ums rb:%u recSize:%u SN:0x%lX Capacity:0x%lX Used:0x%lX freeUp:%u {%20.20s}\n",
             result,
             (unsigned) dsdDirectory.status, (unsigned) dsdDirectory.time, rb_num,
             (unsigned) dsdDirectory.recordSize, (unsigned long) dsdDirectory.instrSerialNo,
             (unsigned long) dsdDirectory.recordCapacity, (unsigned long) dsdDirectory.recordsUsed,
             (unsigned) dsdDirectory.percentFreeUp, dsdDirectory.comment);
}

/*
 * A control line looks like
 *
 * !2 16 123456 // !2=Rinblock2 16=recordSize 123456=instrSerialNo
 */
static void dsd_reinit_control(struct t_io_dev *debug_dev, struct t_rinblock *rb)
{
    unsigned char       ch, rb_num;
    unsigned char       result;
    rb_dsd_directory_t  dsdDirectory;
    int32u              instrSerialNo;
    int16u              recordSize;     // converted when copied to dsdDirectory from 16 to 8
    int16u              percentFreeUp;  // converted when copied to dsdDirectory from 16 to 8

    memset(&dsdDirectory, 0, sizeof(dsdDirectory));
    rfprintf(debug_dev, "Format DSD Dir: Enter:  !rbNum  RecordSize  SN ");

    do                                              // Look for a ! to start a line
    {
        ch = get_char(debug_dev);
        if (ch == '#')
            return;
    }
    while (ch != '!');

    if (!get_1_digit_num(debug_dev, &rb_num))       // Read the rin block number
        return;

    if (get_char(debug_dev) != ' ')                 // Read a space
        return;

    if (!get_u_short(debug_dev, &recordSize))       // Read the record size: unsigned short converted to int8u
        return;

    if (get_char(debug_dev) != ' ')                 // Read a space
        return;

    if (!get_u_long(debug_dev, &instrSerialNo))     // Read the serial number:unsigned long?
        return;

    if (get_char(debug_dev) != ' ')                 // Read a space
        return;

    if (!get_u_short(debug_dev, &percentFreeUp))    // Read the free-up flag: 0(auto), 1 to 100%(manual): converted to int8u
        return;

    memcpy(dsdDirectory.comment, __DATE__" "__TIME__, 20); // "Sep 21 2007 13:33:39" '20' does not include trailing '\0'

    dsdDirectory.recordSize    = (int8u) recordSize;
    dsdDirectory.instrSerialNo = instrSerialNo;
    dsdDirectory.percentFreeUp = (int8u) percentFreeUp;
    // Configure the dsd -- returns capacity & records used (should be 0)
    rfprintf(debug_dev, "Format DSD Dirl:recSz:%u SN:0x%lX free:%u {%20.20s}... ",              (unsigned) recordSize,              (unsigned long) instrSerialNo,              (unsigned) percentFreeUp, dsdDirectory.comment);
    rfprintf(debug_dev, "Format DSD Dirl:recSz:%u SN:0x%lX free:%u {%20.20s}... ", (unsigned) dsdDirectory.recordSize, (unsigned long) dsdDirectory.instrSerialNo, (unsigned) dsdDirectory.percentFreeUp, dsdDirectory.comment);
    result = RinBlockInitDSDDirectory(&rinblocks[rb_num], &dsdDirectory);
    rfprintf(debug_dev, "res=%u st=%u t=%ums rb:%u recSize:%u SN:0x%lX Capacity:0x%lX Used:0x%lX free:%u {%20.20s}\n",
             result,
             (unsigned) dsdDirectory.status, (unsigned) dsdDirectory.time, rb_num,
             (unsigned) dsdDirectory.recordSize, (unsigned long) dsdDirectory.instrSerialNo,
             (unsigned long) dsdDirectory.recordCapacity, (unsigned long) dsdDirectory.recordsUsed,
             (unsigned) dsdDirectory.percentFreeUp, dsdDirectory.comment);
}

/*
 * A write record line looks like
 *
 * !2 1 abc_123_def_ghi // !2=Rinblock2, 1=recordNumber *MUST INCREMENT!*, abc_123_def_ghi=string without spaces
 */
static void dsd_write_record(struct t_io_dev *debug_dev, struct t_rinblock *rb)
{
    unsigned char   rb_num;
    unsigned char   result;
    unsigned char   ch;
    rb_dsd_record_t record;
    int8u           ii;

    memset(&record, 0, sizeof(record));
    rfprintf(debug_dev, "Write Rec: Enter:  !rbNum  RecNum  strings(!=abort,#=end) ");

    do                                              // Look for a ! to start a line
    {
        ch = get_char(debug_dev);
        if (ch == '#')
            return;
    }
    while (ch != '!');

    if (!get_1_digit_num(debug_dev, &rb_num))       // Read the rin block number
        return;

    if (get_char(debug_dev) != ' ')                 // Read a space
        return;

    if (!get_u_long(debug_dev, &record.recordNumber))   // record number: must be at least 1 greater than previous
        return;

    if (get_char(debug_dev) != ' ')                 // Read a space
        return;

    for ( ii = 0; ii < (MAX_RB_DSD_DATA_LENGTH-1); ii++ )   // Look for a ' ' to end the string, or stop at Max-2...
    {
        if ( '#' == (ch = get_char(debug_dev)) )
            break;
        if ( '!' == ch )
            return;
        record.data[ii]   = ch;
        record.data[ii+1] = '\0';
    }

    rfprintf(debug_dev, "Write To DSD record#:0x%lX{%20.20s}... ", (unsigned long) record.recordNumber, (char *) record.data);
    result = RinBlockWriteDSD(rb, &record);
    rfprintf(debug_dev, "res=%u st=%u t=%ums comms:%s\n", result, (unsigned) record.status, (unsigned) record.time, (result) ? "OK" : "Fail");
}

/*
 * A read record 'N' line looks like
 *
 * !2 1 // !2=Rinblock2, 1=recordNumber
 */
static void dsd_read_record_n(struct t_io_dev *debug_dev, struct t_rinblock *rb)
{
    unsigned char     ch;
    rb_dsd_read_cmd_t command;
    rb_dsd_record_t   record;
    unsigned char     result;
    unsigned char     rb_num;
    int32u            recordNumber;

    memset(&command, 0, sizeof(command));
    memset(&record , 0, sizeof(record));

    rfprintf(debug_dev, "Read rec N: Enter: !rbNum RecNum ", command.recordNumber);
    do                                              // Look for a ! to start a line
    {
        ch = get_char(debug_dev);
        if (ch == '#')
            return;
    }
    while (ch != '!');

    if (!get_1_digit_num(debug_dev, &rb_num))       // Read the rin block number
        return;

    if (get_char(debug_dev) != ' ')                 // Read a space
        return;

    if (!get_u_long(debug_dev, &recordNumber))   // record number
        return;
    command.recordNumber = recordNumber;

    rfprintf(debug_dev, "Read rec N # 0x%lX ", (unsigned long) command.recordNumber);
    result = RinBlockReadDSD(rb, dsdRdRecord, &command, &record);
    rfprintf(debug_dev, "res=%u st=%u t=%ums comms:%s 0x%lX{%20.20s}\n",
             result, (unsigned) record.status, (unsigned) record.time, (result) ? "OK" : "Fail",
             (unsigned long) record.recordNumber, (char *) record.data);
}

/*
 * A read record 'next' or 'prev' line looks like
 *
 * !2 1 // !2=Rinblock2, 1=recordNumber(does not have to exist)
 */
static void dsd_read_next_prev(struct t_io_dev *debug_dev, struct t_rinblock *rb, DsdRead_t dsdRdNextOrPrev)
{
    rb_dsd_read_cmd_t command;
    rb_dsd_record_t   record;
    unsigned char     result;
    unsigned char     ch;
    unsigned char     rb_num;
    int32u            recordNumber;

    memset(&command, 0, sizeof(command));
    memset(&record , 0, sizeof(record ));
    rfprintf(debug_dev, "Read %s rec: Enter: !rbNum RecNum ", (dsdRdNextOrPrev == dsdRdNext) ? "next" : "prev");

    do                                              // Look for a ! to start a line
    {
        ch = get_char(debug_dev);
        if (ch == '#')
            return;
    }
    while (ch != '!');

    if (!get_1_digit_num(debug_dev, &rb_num))       // Read the rin block number
        return;

    if (get_char(debug_dev) != ' ')                 // Read a space
        return;

    if (!get_u_long(debug_dev, &recordNumber))   // record number
        return;

    command.recordNumber = recordNumber;

    rfprintf(debug_dev, "Read %s rec # 0x%lX ", (dsdRdNextOrPrev == dsdRdNext) ? "next" : "prev", (unsigned long) command.recordNumber);
    result = RinBlockReadDSD(rb, dsdRdNextOrPrev, &command, &record);
    rfprintf(debug_dev, "res=%u st=%u t=%ums comms:%s 0x%lX{%20.20s}\n",
             result,
             (unsigned) record.status, (unsigned) record.time, (result) ? "OK" : "Fail",
             (unsigned long) record.recordNumber, (char *) record.data);
}

/*
 * A read record 'oldest/first' or 'newest/last' line looks like
 *
 * !2 1 // !2=Rinblock2, 1=recordNumber(does not have to exist)
 */
static void dsd_read_old_new(struct t_io_dev *debug_dev, struct t_rinblock *rb, DsdRead_t dsdRdOldestOrNewest)
{
    rb_dsd_read_cmd_t command;
    rb_dsd_record_t   record;
    unsigned char     result;
    unsigned char     ch;
    unsigned char     rb_num;

    memset(&command, 0, sizeof(command));
    memset(&record , 0, sizeof(record ));
    rfprintf(debug_dev, "Read %s rec: Enter: !rbNum ", (dsdRdOldestOrNewest == dsdRdOldest) ? "oldest" : "newest");

    do                                              // Look for a ! to start a line
    {
        ch = get_char(debug_dev);
        if (ch == '#')
            return;
    }
    while (ch != '!');

    if (!get_1_digit_num(debug_dev, &rb_num))       // Read the rin block number
        return;

    rfprintf(debug_dev, "Read %s rec # 0x%lX ", (dsdRdOldestOrNewest == dsdRdOldest) ? "oldest" : "newest", (unsigned long) command.recordNumber);
    result = RinBlockReadDSD(rb, dsdRdOldestOrNewest, &command, &record);
    rfprintf(debug_dev, "res=%u st=%u t=%ums comms:%s 0x%lX{%20.20s}\n",
             result,
             (unsigned) record.status, (unsigned) record.time, (result) ? "OK" : "Fail",
             (unsigned long) record.recordNumber, (char *) record.data);
}

/*
 * Set the percentFreeUp flag:
 *    0: set to automatic deletion from now on, as new records are added.
 *    1...100: set to manual deletion from now one, plus free-up this % of space immediately.
 *
 * !2 0 // !2=Rinblock2, 0: do automatic deletion, 1...100: manual deletion required
 */
static void dsd_free_up(struct t_io_dev *debug_dev, struct t_rinblock *rb)
{
    rb_dsd_free_up_t  message;
    unsigned char     result;
    unsigned char     ch;
    unsigned char     rb_num;
    int16u            percentFreeUp;  // converted when copied to dsdDirectory from 16 to 8

    memset(&message , 0, sizeof(message));

    do                                              // Look for a ! to start a line
    {
        ch = get_char(debug_dev);
        if (ch == '#')
            return;
    }
    while (ch != '!');

    if (!get_1_digit_num(debug_dev, &rb_num))       // Read the rin block number
        return;

    if (get_char(debug_dev) != ' ')                 // Read a space
        return;

    if (!get_u_short(debug_dev, &percentFreeUp))    // Read the free-up flag: 0(auto), 1 to 100%(manual): converted to int8u
        return;

    message.percentFreeUp = (int8u)percentFreeUp;

    rfprintf(debug_dev, "Free up %u: ", (unsigned) message.percentFreeUp);
    result = RinBlockFreeUpDSD(rb, &message);
    rfprintf(debug_dev, "res=%u st=%u t=%ums fu=%u%% cap:%lu used:%lu\n", result, (unsigned) message.status, (unsigned) message.time,
             (unsigned) message.percentFreeUp, (unsigned long) message.recordCapacity, (unsigned long) message.recordsUsed);
}

/*
 * dsd_test -
 *
 * Test commands for the DSD module
 */
void dsd_test(struct t_io_dev *debug_dev, struct t_rinblock *rb)
{
    unsigned char ch = 0;

    while (ch != 'q')
    {
        rfprintf(debug_dev,  "\nRin Block DSD Test:\n"
                             "  q - Quit\n"
                             "  d - Read dir\n"
                             "  F - Format dir\n"
                             "  f - Free up flag: 0(auto), 1 to 100%%(manual)\n"
                             "  w - Write #N\n"
                             "  r - Read #N\n"
                             "  n - read Next(newer) after #N\n"
                             "  p - read Previous(older) before #N\n"
                             "  t - read Tail(newest)\n"
                             "  h - read Head(oldest)\n");

        rread(debug_dev, &ch, 1, 0);
        switch (ch)
        {
           case 'd':  dsd_read_control  (debug_dev, rb);    break;
           case 'F':  dsd_reinit_control(debug_dev, rb);    break;
           case 'f':  dsd_free_up       (debug_dev, rb);    break;
           case 'w':  dsd_write_record  (debug_dev, rb);    break;
           case 'r':  dsd_read_record_n (debug_dev, rb);    break;
           case 'n':  dsd_read_next_prev(debug_dev, rb, dsdRdNext);         break;
           case 'p':  dsd_read_next_prev(debug_dev, rb, dsdRdPrevious);     break;
           case 't':  dsd_read_old_new  (debug_dev, rb, dsdRdNewest);       break;
           case 'h':  dsd_read_old_new  (debug_dev, rb, dsdRdOldest);       break;
           default:   break;
        }
    }
}
