#include <rinhw/rio.h>
#include <rinhw/rstdio.h>
#include <rinhw/sio.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_program.h>

#include "rb_program_test.h"

#ifdef __HITACHI__
  extern const unsigned char prog_buf[ RB_FLASH_MEGA8_LENGTH ];
  const unsigned char *program_buf_ptr = prog_buf;
#elif ( defined(__GNUC__) && defined(__WIN32__) )
  extern unsigned char binary___bin_main_chk_bin_start[ RB_FLASH_MEGA8_LENGTH ];
  const unsigned char *program_buf_ptr = binary___bin_main_chk_bin_start;
#elif ( defined(__GNUC__) && defined(__linux__) )
  extern unsigned char _binary_main_chk_bin_start[ RB_FLASH_MEGA8_LENGTH ];
  const unsigned char *program_buf_ptr = _binary_main_chk_bin_start;
#endif


unsigned char ram_work_buf[RB_FLASH_SERIAL_BLOCK_SIZE];


void testprogram_test(struct t_io_dev *debug_dev, struct t_rinblock *rb)
{
    unsigned char res, err;
    rfprintf(debug_dev, "Rin Block Program From RAM Buffer, ");

    res = RinBlockTestProgram(rb, program_buf_ptr, &err, RB_FLASH_START, RB_FLASH_MEGA8_LENGTH, ram_work_buf);
    rfprintf(debug_dev, "result %d, error code %d.\n", res, err);
}


/**
 * Program rinblock rb via serial port serial_dev.
 * Looks for "PFxxxx;", or "ABORT" from the serial port to initiate programming, or abort
 * xxxx is 4 digits of the RB_HW_VERSION_ constants (ie M4301 8 DIO is 0001)
 */
unsigned char RinBlockProgramFromRunning(struct t_io_dev *serial_dev, struct t_rinblock *rb)
{
  const char abort_str[] = "ABORT";
  unsigned char ch;
  unsigned char result = 0;
  char programming_string[20];
  const char *prog_ptr;
  const char *abort_ptr;

  if (rb->running)
  {
    rsprintf(programming_string, "PF%04d;", rb->tag.hardware_version);
  }
  else
  {
    rsprintf(programming_string, "PFxxxx;");
  }

  for (prog_ptr = programming_string, abort_ptr = abort_str; ;)
  {
    /* Get a character */
    rread(serial_dev, &ch, 1, 0);

    /* Look for match of programming_string, or PFxxxx; for non-running rinblocks, where x is any character */
    if (*prog_ptr == ch ||
        ('x' == *prog_ptr && !rb->running))
    {
      prog_ptr++;
      if (*prog_ptr == '\0')
      {
        /* Match, so program rinblock */
        RinBlockProgram(rb, serial_dev, (struct t_wdt *) NULL, &ch);
        result = 1;
        break;
      }
    }
    else
    {
      prog_ptr = programming_string;
    }

    /* Look for match of abort_str */
    if (*abort_ptr == ch)
    {
      abort_ptr++;
      if (*abort_ptr == '\0')
      {
        /* Match, so abort */
        result = 0;
        break;
      }
    }
    else
    {
      abort_ptr = abort_str;
    }
  }
  return result;
}


extern struct t_io_dev dev_rb_serial1;


#if 0 /* Was R400 */
/**
 * Interactive version of RinBlockProgramFromRunning
 * Asks user on serial_dev which serial port to program the rinblock via
 * Looks for "PFxxxx;", or "ABORT" from the chosen serial port to initiate programming, or abort
 * xxxx is 4 digits of the RB_HW_VERSION_ constants (ie M4301 8 DIO is 0001)
 */
void RinBlockSerialProgram(struct t_io_dev *serial_dev, struct t_rinblock *rb)
{
  unsigned char ch, result;
  struct t_io_dev *prog_serial;
  unsigned char alreadyopen;
  t_ioctl_siocfg siocfg, oldsiocfg;


  rfprintf(serial_dev, "About to program rinblock at slot %d via serial port\r\n", rb->bus_slot);
  rfprintf(serial_dev, "Select Programming Port:\r\n"
                       " 2 - dev_com2\r\n"
                       " 3 - dev_com3\r\n"
                       " r - dev_rb_serial1 (must have serial module attached)\r\n"
                       " Anything else - abort\r\n");

  rread(serial_dev, &ch, 1, 0);
  switch (ch)
  {
    case '2':
      prog_serial = &dev_com2;
      rfprintf(serial_dev, "Beginning programming on dev_com2\r\n");
      break;
    case '3':
      prog_serial = &dev_com3;
      rfprintf(serial_dev, "Beginning programming on dev_com3\r\n");
      break;
    case 'r':
      prog_serial = &dev_rb_serial1;
      rfprintf(serial_dev, "Beginning programming on dev_rb_serial1\r\n");
      break;
    default:
      rfprintf(serial_dev, "Abort programming\r\n");
      return;
  }

  alreadyopen = (0 == (prog_serial->status & O_OPEN));
  if (!alreadyopen)
  {
    if (ropen(prog_serial, 0) < 0)
    {
      rfprintf(serial_dev, "Could not open programming port\r\n");
      return;
    }
  }

  rioctl(prog_serial, IOCTL_SER_GET_CONFIG, &oldsiocfg);
  siocfg.baud = 9600;
  siocfg.data_bits = SIO_DATABITS_EIGHT;
  siocfg.parity_bits = SIO_PARITY_NONE;
  siocfg.stop_bits = SIO_STOPBITS_ONE;
  siocfg.use_cts = 0;
  siocfg.use_terminations = 0;
  rioctl(prog_serial, IOCTL_SER_SET_CONFIG, &siocfg);

  result = RinBlockProgramFromRunning(prog_serial, rb);
  rfprintf(serial_dev, "Result of programming %d\r\n", result);

  rioctl(prog_serial, IOCTL_SER_SET_CONFIG, &oldsiocfg);

  if (!alreadyopen)
  {
    rclose(prog_serial);
  }
}
#endif /* #if 0 */
