#include <stdarg.h>
#include <string.h>
#include <rinhw/rstdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/debug.h>
#include <rinhw/rinendian.h>
#include <rinhw/kbd_local.h>
#include <rinhw/kbd.h>
#include <rinremote/remote.h>
#include <rinremote/remoteexec.h>
#include "remote_device.h"
#include "remotehelper.h"

#include <rinremote/kbd_remote.h>


static int remote_dev_ioctl_kbd_configure( struct t_io_dev *io_dev, int nIoCtl, int8u ui8NumKeys, const unsigned char *Keys )
{
  int8u txbuf[ GET_SIZE_REMOTE_CMD_IOCTL_KBD_CONFIGURE( MAX_REMOTE_BUFFER ) ];
  int8u rxbuf[ GET_SIZE_REMOTE_RESP_IOCTL_DEFAULT() ];
  struct s_rem_device_command *prdCmd = (struct s_rem_device_command *) txbuf;
  struct s_rem_device_response *prdResponse = (struct s_rem_device_response *) rxbuf;
  int nResult;
  const struct s_remote_device_info *remote_info = (struct s_remote_device_info *) io_dev->dev_data;
  const int8u ui8Device = remote_info->device_id;

  vInitRemoteIoCtlCmdHdr( prdCmd, ui8Device, nIoCtl );
  prdCmd->rem_device_command.rem_dev_cmd_ioctl.remote_ioctl_command.rem_dev_cmd_ioctl_kbd_configure.ui8NumKeys = ui8NumKeys;
  memcpy( prdCmd->rem_device_command.rem_dev_cmd_ioctl.remote_ioctl_command.rem_dev_cmd_ioctl_kbd_configure.ui8Buffer,
          Keys, ui8NumKeys );

  if ( !bRemoteExecuteCommand( prdCmd, GET_SIZE_REMOTE_CMD_IOCTL_KBD_CONFIGURE( ui8NumKeys ),
                               prdResponse, GET_SIZE_REMOTE_RESP_IOCTL_DEFAULT() ) )
  {
    nResult = RIO_ERROR;
  }
  else
  {
    nResult = nCheckRemoteIoCtlResp_Default( prdResponse, ui8Device, nIoCtl );
  }
  return nResult;
}


static int remote_dev_ioctl_kbd_external_keybits( struct t_io_dev *io_dev, int32u ui32ExtKeybits )
{
  int8u txbuf[ GET_SIZE_REMOTE_CMD_IOCTL_KBD_EXT_KEYBITS() ];
  int8u rxbuf[ GET_SIZE_REMOTE_RESP_IOCTL_DEFAULT() ];
  struct s_rem_device_command *prdCmd = (struct s_rem_device_command *) txbuf;
  struct s_rem_device_response *prdResponse = (struct s_rem_device_response *) rxbuf;
  int nResult;
  const struct s_remote_device_info *remote_info = (struct s_remote_device_info *) io_dev->dev_data;
  const int8u ui8Device = remote_info->device_id;

  vInitRemoteIoCtlCmdHdr( prdCmd, ui8Device, IOCTL_KBD_WRITE_HI_KEY_BITS );
  prdCmd->rem_device_command.rem_dev_cmd_ioctl.remote_ioctl_command.rem_dev_cmd_ioctl_kbd_external_keybits.ui32ExternalKeybits = htonl( ui32ExtKeybits );

  if ( !bRemoteExecuteCommand( prdCmd, GET_SIZE_REMOTE_CMD_IOCTL_KBD_EXT_KEYBITS( ),
                               prdResponse, GET_SIZE_REMOTE_RESP_IOCTL_DEFAULT() ) )
  {
    nResult = RIO_ERROR;
  }
  else
  {
    nResult = nCheckRemoteIoCtlResp_Default( prdResponse, ui8Device, IOCTL_KBD_WRITE_HI_KEY_BITS );
  }
  return nResult;
}


static int remote_dev_ioctl_kbd_internal_keybits( struct t_io_dev *io_dev, int32u *pui32IntKeybits )
{
  int8u txbuf[ GET_SIZE_REMOTE_CMD_IOCTL_KBD_INT_KEYBITS() ];
  int8u rxbuf[ GET_SIZE_REMOTE_RESP_IOCTL_KBD_INT_KEYBITS() ];
  struct s_rem_device_command *prdCmd = (struct s_rem_device_command *) txbuf;
  struct s_rem_device_response *prdResponse = (struct s_rem_device_response *) rxbuf;
  int nResult;
  const struct s_remote_device_info *remote_info = (struct s_remote_device_info *) io_dev->dev_data;
  const int8u ui8Device = remote_info->device_id;

  vInitRemoteIoCtlCmdHdr( prdCmd, ui8Device, IOCTL_KBD_INTERNAL_BITS );

  if ( !bRemoteExecuteCommand( prdCmd, GET_SIZE_REMOTE_CMD_IOCTL_KBD_INT_KEYBITS(),
                               prdResponse, GET_SIZE_REMOTE_RESP_IOCTL_KBD_INT_KEYBITS() ) )
  {
    nResult = RIO_ERROR;
  }
  else if ( !bCheckRemoteIoCtlRespHdr( prdResponse, ui8Device, IOCTL_KBD_INTERNAL_BITS ) )
  {
    nResult = RIO_ERROR;
  }
  else
  {
    nResult = ntohl( prdResponse->rem_device_response.rem_dev_resp_ioctl.remote_ioctl_response.rem_dev_resp_ioctl_kbd_internal_keybits.i32Result );
    *pui32IntKeybits = ntohl( prdResponse->rem_device_response.rem_dev_resp_ioctl.remote_ioctl_response.rem_dev_resp_ioctl_kbd_internal_keybits.ui32InternalKeybits );
  }
  return nResult;
}


static int kbd_remote_ioctl(struct t_io_dev *io_dev, int func, va_list *arglist)
{
    unsigned char num_keys;
    unsigned char *keys;
    int32u external_keybits;
    int32u internal_keybits;
    int result;

    result = RIO_ERROR;
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("kbd_ioctl called\n");
#endif

    /*
     * Function dispatch
     */
    switch (func)
    {

        /*
         * Configure some keys
         */
        case IOCTL_KBD_SET_BASIC:
        case IOCTL_KBD_SET_REPEAT:
        case IOCTL_KBD_SET_HOLD:
        case IOCTL_KBD_SET_DISABLED:
            num_keys = (unsigned char) va_arg(*arglist, int);
            keys = va_arg(*arglist, unsigned char *);
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
            DEBUG_PRINTF("Config keys, func=%d, num_keys=%d, keys_ptr=%p\n", func, num_keys, keys);
#endif
            result = remote_dev_ioctl_kbd_configure( io_dev, func, num_keys, keys );
            break;

        /*
         * Set the key bits for the external 32keys
         */
        case IOCTL_KBD_WRITE_HI_KEY_BITS:
            external_keybits = va_arg(*arglist, unsigned long);
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
            DEBUG_PRINTF("Write external keybits %lX\n", external_keybits );
#endif
            result = remote_dev_ioctl_kbd_external_keybits( io_dev, external_keybits );
            break;

        /*
         * Returns true or false depending on whether the key is currently pressed
         */
        case IOCTL_KBD_IS_KEY_PRESSED:
          /* Not used anywhere so haven't implemented */
//            pressed = va_arg(arglist, unsigned char *);   // get address of answer
//            *pressed = ((kb_db[0].keybits & (0x00000001 << KeyToIndex(*pressed))) != 0); // work out if key is pressed
//            result = RIO_SUCCESS;
            break;

        /*
         * Returns internal key status as a maskable unsigned long
         */
        case IOCTL_KBD_INTERNAL_BITS:
#if HW_DEBUG(KBD_DEBUG, HW_DEBUG_FULL)
            DEBUG_PRINTF( "Read internal keybits\n" );
#endif
            internal_keybits = 0;
            result = remote_dev_ioctl_kbd_internal_keybits( io_dev, &internal_keybits );
            *(va_arg(*arglist, unsigned long *)) = internal_keybits;
            result = RIO_SUCCESS;
            break;
    }

    /*
     * Done, return result
     */
    return result;
}


/** Function to create and initialise a remote keyboard device */
int rio_create_device_kbd_remote(struct t_io_dev *io_dev)
{
  int result;

  /* Create and initialise a default remote device (standard IO functions, no IOCTL function) */
  result = rio_create_device_default_remote(io_dev, REMOTE_DEV_KBD);
  io_dev->dev_ioctl = kbd_remote_ioctl;
  return result;
}
