/*
 * remoteexec_parallel.c
 *
 * Same functionality as remoteexec.c, except it allows multiple devices to execute commands simultaneously
 * Multiple simultaneous commands on one device are prevented by semaphores
 *  Created on: 23/10/2008
 *      Author: ianb
 */
#include <stddef.h>
#include <string.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdint.h>
#include <stdbool.h>
#include <rinhw/debug.h>
#include <rinhw/rinthread.h>
#include <rinremote/remote.h>
#include <rinremote/slip.h>
#include <rinremote/slip_io.h>

#include <rinremote/remoteexec.h>


static pthread_mutex_t mutexRemoteCommandLock;
static pthread_mutex_t mutexRemotePendInfoLock;
static bool bRemoteCommandsEnabled = false;
static struct s_slipio remote_slip_io;
static pthread_t remote_thread;


struct sRemoteRespPendInfo
{
  bool bInUse;
  sem_t semRemoteResponseWait;
  struct s_rem_dev_cmd_header CmdHeader;
  struct s_rem_device_response *pResponse;
  int nRespMaxLength;
};

#define MAX_SIMULTANEOUS_REQUESTS 5


static struct sRemoteRespPendInfo RemRespPendInfos[MAX_SIMULTANEOUS_REQUESTS];


static struct sRemoteRespPendInfo *pGetFreeRemRespPendInfo(void)
{
  size_t i;
  struct sRemoteRespPendInfo *pResult = (struct sRemoteRespPendInfo *) NULL;

  pthread_mutex_lock(&mutexRemotePendInfoLock);
  for (i = 0; i < MAX_SIMULTANEOUS_REQUESTS; i++)
  {
    if (!RemRespPendInfos[i].bInUse)
    {
      RemRespPendInfos[i].bInUse = true;
      pResult = &RemRespPendInfos[i];
      break;
    }
  }
  pthread_mutex_unlock(&mutexRemotePendInfoLock);
  return pResult;
}


static void vReleaseRemRespPendInfo(struct sRemoteRespPendInfo *pRespPendInfo)
{
  pthread_mutex_lock(&mutexRemotePendInfoLock);
  pRespPendInfo->bInUse = false;
  pRespPendInfo->nRespMaxLength = 0;
  pRespPendInfo->pResponse = NULL;
  pRespPendInfo->CmdHeader.ui8Command = 0;
  pRespPendInfo->CmdHeader.ui8Device = 0;
  pthread_mutex_unlock(&mutexRemotePendInfoLock);
}


static struct sRemoteRespPendInfo *pFindWaitingRemRespPendInfos(const struct s_rem_dev_cmd_header *pRespHdr)
{
  size_t i;
  struct sRemoteRespPendInfo *pResult = (struct sRemoteRespPendInfo *) NULL;

  pthread_mutex_lock(&mutexRemotePendInfoLock);
  for (i = 0; i < MAX_SIMULTANEOUS_REQUESTS; i++)
  {
    if (RemRespPendInfos[i].bInUse &&
        RemRespPendInfos[i].CmdHeader.ui8Command == pRespHdr->ui8Command &&
        RemRespPendInfos[i].CmdHeader.ui8Device == pRespHdr->ui8Device)
    {
      pResult = &RemRespPendInfos[i];
      break;
    }
  }
  pthread_mutex_unlock(&mutexRemotePendInfoLock);
  return pResult;
}


static void vInitRemRespPendInfo(void)
{
  size_t i;

  pthread_mutex_init(&mutexRemotePendInfoLock, NULL);
  for (i = 0; i < MAX_SIMULTANEOUS_REQUESTS; i++)
  {
    RemRespPendInfos[i].bInUse = false;
    if (sem_init(&RemRespPendInfos[i].semRemoteResponseWait, 0, 0) != 0)
    {
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("Can't create remote response wait sem\n");
#endif
    }
    RemRespPendInfos[i].nRespMaxLength = 0;
    RemRespPendInfos[i].pResponse = 0;
    RemRespPendInfos[i].CmdHeader.ui8Command = 0;
    RemRespPendInfos[i].CmdHeader.ui8Device = 0;
  }
}


static void vDestroyRemRespPendInfo(void)
{
  size_t i;

  for (i = 0; i < MAX_SIMULTANEOUS_REQUESTS; i++)
  {
    sem_destroy(&RemRespPendInfos[i].semRemoteResponseWait);
  }
  pthread_mutex_destroy(&mutexRemotePendInfoLock);
}


static void *vRemoteExecuteTask(void *pData)
{
  uint8_t aui8RespBuffer[MAX_RESPONSE_BUFFER];
  struct s_rem_device_response *pResponse = (struct s_rem_device_response *) aui8RespBuffer;
  int nRespLength;
  struct sRemoteRespPendInfo *pRespPendInfo;
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
  uint8_t ui8Cmd;
  uint8_t ui8Dev;
#endif

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("vRemoteExecuteTask() start\n");
#endif

  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  for (;;)
  {

    /* Wait for slip packet from remote device */
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("vRemoteExecuteTask: Wait for slip packet\n");
#endif
    nRespLength = slip_recv_packet(&remote_slip_io, aui8RespBuffer, sizeof(aui8RespBuffer));

    if (nRespLength < (int) sizeof(struct s_rem_dev_cmd_header))
    {
      /* Not enough data to process */
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("vRemoteExecuteTask: Packet too short\n");
#endif
      continue;
    }

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
    ui8Cmd = pResponse->rem_dev_cmd_header.ui8Command;
    ui8Dev = pResponse->rem_dev_cmd_header.ui8Device;
#endif
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("vRemoteExecuteTask: cmd %d, dev %d: received reply packet\n", ui8Cmd, ui8Dev);
#endif

    /* See if we can find a task waiting on the command/device combination  */
    pRespPendInfo = pFindWaitingRemRespPendInfos(&pResponse->rem_dev_cmd_header);
    if (NULL == pRespPendInfo)
    {
      /* No one waiting, this is probably quite bad */
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
      DEBUG_PRINTF("vRemoteExecuteTask: cmd %d, dev %d: ERROR no request waiting for this response\n", ui8Cmd, ui8Dev);
#endif
      continue;
    }

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("vRemoteExecuteTask: cmd %d, dev %d: Matching request found\n", ui8Cmd, ui8Dev);
#endif

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
    if (pRespPendInfo->nRespMaxLength < nRespLength)
    {
      DEBUG_PRINTF("vRemoteExecuteTask: cmd %d, dev %d: ERROR response length %d > buffer size %d\n",
                   ui8Cmd, ui8Dev, nRespLength, pRespPendInfo->nRespMaxLength );
    }
#endif

    /* Yep someone is waiting for data, copy the data over, and wake the task */
    nRespLength = (pRespPendInfo->nRespMaxLength < nRespLength) ? pRespPendInfo->nRespMaxLength : nRespLength;
    memcpy(pRespPendInfo->pResponse, pResponse, nRespLength);
    sem_post(&pRespPendInfo->semRemoteResponseWait);
  }
  rin_thread_exit(NULL);
}


/** Restart the remote hardware */
static int nRemoteHardwareRestart(void)
{
  uint8_t txbuf[GET_SIZE_REMOTE_CMD_INIT()];
  uint8_t rxbuf[GET_SIZE_REMOTE_RESP_INIT()];
  struct s_rem_device_command *prdCmd = (struct s_rem_device_command *) txbuf;
  struct s_rem_device_response *prdResponse = (struct s_rem_device_response *) rxbuf;
  int nResult;

  /* Any command sent to device REMOTE_DEV_RESET will cause a reset */
  vInitRemoteCmdHdr(prdCmd, REMOTE_CMD_INIT, REMOTE_DEV_RESET);
  prdCmd->rem_device_command.rem_dev_cmd_init.ui8Unused = 0;
  if (!bRemoteExecuteCommand(prdCmd, GET_SIZE_REMOTE_CMD_INIT(), prdResponse, GET_SIZE_REMOTE_RESP_INIT()))
  {
    nResult = RIO_ERROR;
  }
  else
  {
    nResult = nCheckRemoteResp_Default(prdResponse, REMOTE_CMD_INIT, REMOTE_DEV_RESET);
  }
  return nResult;
}


/**
 * Initialise remote execution.
 *
 * @param serial_port is the serial port to use, if NULL server/port is used instead
 * @param server is the server to connect to if serial_port is NULL
 * @param port is the TCP socket on server to if serial_port is NULL
 */
int vRemoteExecuteInit(const char *serial_port, const char *server, int port)
{
  int result;

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("vRemoteExecuteInit() called\n");
#endif

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("vRemoteExecuteInit: Open serial port/socket\n");
#endif
  if (NULL != serial_port)
  {
    /* Serial */
    if (slip_serial_init(&remote_slip_io, serial_port) < 0)
    {
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("vRemoteExecuteInit: slip_serial_init failed\n");
#endif
      result = RIO_ERROR;
      goto rei_fail;
    }
  }
  else
  {
    /* TCP Socket */
    if (slip_socket_connect_init(&remote_slip_io, server, port) < 0)
    {
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
      DEBUG_PRINTF("vRemoteExecuteInit: slip_socket_init to %s:%d failed\n", server, port);
#endif
      result = RIO_ERROR;
      goto rei_fail;
    }
  }

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("vRemoteExecuteInit: Create sem\n");
#endif

  /* Create the remote command lock mutex */
  if (pthread_mutex_init(&mutexRemoteCommandLock, NULL) != 0)
  {
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("vRemoteExecuteInit: Can't create remote serial port excl sem\n");
#endif
    result = RIO_ERROR;
    goto rei_fail;
  }

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("vRemoteExecuteInit: Initalise response pending info\n");
#endif
  vInitRemRespPendInfo();

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("vRemoteExecuteInit: Create response pending task\n");
#endif
  (void) rin_thread_create(&remote_thread, vRemoteExecuteTask, NULL, 0);
  bRemoteCommandsEnabled = true;

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("vRemoteExecuteInit: Restart remote R400 hardware\n");
#endif
  if (nRemoteHardwareRestart() < 0)
  {
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("vRemoteExecuteInit: Restart remote R400 hardware failed\n");
#endif
    result = RIO_ERROR;
    goto rei_fail;
  }

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("vRemoteExecuteInit() complete\n");
#endif
  return RIO_SUCCESS;

rei_fail:
  /* TODO clean up */
  return result;
}


void vRemoteExecuteDestroy(void)
{
  /* Kill task */
  (void) pthread_cancel(remote_thread);
  (void) pthread_join(remote_thread, NULL);

  /* Destroy vInitRemRespPendInfo */
  vDestroyRemRespPendInfo();

  /* Destroy the open exclusive semaphore */
  pthread_mutex_destroy(&mutexRemoteCommandLock);

  slip_destroy(&remote_slip_io);
}


bool bRemoteExecuteCommand(const struct s_rem_device_command *pCommand, int nCmdLength,
                           struct s_rem_device_response *pResponse, int nRespLength)
{
  const char *pCmdBuf = (const char *) pCommand;
  struct sRemoteRespPendInfo *pRespPendInfo;
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
  const uint8_t ui8Cmd = pCommand->rem_dev_cmd_header.ui8Command;
  const uint8_t ui8Dev = pCommand->rem_dev_cmd_header.ui8Device;
#endif
  int res;

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("bRemoteExecuteCommand() called, cmd %d, dev %d\n", ui8Cmd, ui8Dev);
#endif

  /* Check that remote commands are enabled */
  if (!bRemoteCommandsEnabled)
  {
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
    DEBUG_PRINTF("bRemoteExecuteCommand: Remote commands not enabled, fail\n");
#endif
    return false;
  }

  /* TODO make sure device isn't already being accessed */

  /* Get a free response pend info */
  pRespPendInfo = pGetFreeRemRespPendInfo();
  if (NULL == pRespPendInfo)
  {
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("bRemoteExecuteCommand: cmd %d, dev %d: Could not get free response pend info\n", ui8Cmd, ui8Dev);
#endif
    return false;
  }

  /* Setup info in response pend info */
  pRespPendInfo->CmdHeader = pCommand->rem_dev_cmd_header;
  pRespPendInfo->pResponse = pResponse;
  pRespPendInfo->nRespMaxLength = nRespLength;

  /* Get exclusive access for sending the slip packet to the serial port */
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("bRemoteExecuteCommand: cmd %d, dev %d: Lock serial port for slip send\n", ui8Cmd, ui8Dev);
#endif
  if (pthread_mutex_lock(&mutexRemoteCommandLock) != 0)
  {
    vReleaseRemRespPendInfo(pRespPendInfo);
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
    DEBUG_PRINTF("bRemoteExecuteCommand: cmd %d, dev %d: ERROR pending on serial port exclusive semaphore\n", ui8Cmd, ui8Dev);
#endif
    return false;
  }

  /* Send the packet */
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("bRemoteExecuteCommand: cmd %d, dev %d: Send slip packet over serial port\n", ui8Cmd, ui8Dev);
#endif
  slip_send_packet(&remote_slip_io, (const int8u *) pCmdBuf, nCmdLength );

  /* Release the exclusive access semaphore */
  pthread_mutex_unlock(&mutexRemoteCommandLock);

  /* Wait for the demux task to get back to us that the response is ready */
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("bRemoteExecuteCommand: cmd %d, dev %d: Wait for response\n", ui8Cmd, ui8Dev);
#endif
  res = sem_wait(&pRespPendInfo->semRemoteResponseWait);

  /* Release the pend info struct */
  vReleaseRemRespPendInfo(pRespPendInfo);

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("bRemoteExecuteCommand() finished, cmd %d, dev %d, result %d\n", ui8Cmd, ui8Dev, (0 == res));
#endif
  return (0 == res);
}
