/*
 * slip_io_socket_connect.c
 *
 *  Created on: 31/05/2010
 *      Author: ian
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <errno.h>
#include <rinhw/rio.h>
#include <rinhw/sio.h>
#include <rinhw/debug.h>
#include <rinremote/slip.h>

#include <rinremote/slip_io.h>


/** Cork/uncork a TCP connection */
static void slip_socket_connect_cork_tcp(int fd, bool cork)
{
  int state = (cork) ? 1 : 0;

  /* Cork/uncork the connection */
  if (setsockopt(fd, IPPROTO_TCP, TCP_CORK, &state, sizeof(state)) < 0)
  {
    rinhw_error_printf("slip_socket_connect_cork_tcp: failed to %s port\n", (cork) ? "cork" : "uncork");
    return;
  }
}


static void slip_socket_connect_pre_send(void *data)
{
  intptr_t sockfd = (intptr_t) data;

  /* Cork the port before the slip packet is written */
  if (sockfd >= 0)
    slip_socket_connect_cork_tcp(sockfd, true);
}


static void slip_socket_connect_post_send(void *data)
{
  intptr_t sockfd = (intptr_t) data;

  /* Uncork the port after slip packet is written so it gets sent as one big chunk */
  if (sockfd >= 0)
    slip_socket_connect_cork_tcp(sockfd, false);
}


static void slip_send_char_socket_connect(void *data, uint8_t ch)
{
  intptr_t sockfd = (intptr_t) data;

  if (sockfd >= 0)
    write(sockfd, &ch, 1);
}


static uint8_t slip_recv_char_socket_connect(void *data)
{
  intptr_t sockfd = (intptr_t) data;
  uint8_t ch = '\0';

  if (sockfd >= 0)
    read(sockfd, &ch, 1);
  return ch;
}


/** Clean up the slip socket (connecting to a listening socket) */
static void slip_socket_connect_destroy(void *data)
{
  intptr_t sockfd = (intptr_t) data;

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("slip_socket_connect_destroy() called\n");
#endif
  if (sockfd >= 0)
    close(sockfd);
}


/** Initialise slip using a TCP socket (connecting to a listening socket) */
int slip_socket_connect_init(struct s_slipio *slipio, const char *server, int port)
{
  intptr_t sockfd;
  struct addrinfo hints;
  struct addrinfo *result, *rp;
  int s;
  char portstr[20];
  int state;

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("slip_socket_connect_init() called\n");
#endif

  /* Obtain address(es) matching host/port */
  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_family = AF_UNSPEC;    /* Allow IPv4 or IPv6 */
  hints.ai_socktype = SOCK_STREAM; /* Datagram socket */
  hints.ai_flags = 0;
  hints.ai_protocol = 0;          /* Any protocol */
  snprintf(portstr, 20, "%d", port);
  s = getaddrinfo(server, portstr, &hints, &result);
  if (s != 0)
  {
    rinhw_error_printf("slip_socket_connect_init: getaddrinfo fail: %s\n", gai_strerror(s));
    return RIO_ERROR;
  }

  /*
   * getaddrinfo() returns a list of address structures.
   * Try each address until we successfully connect(2).
   * If socket(2) (or connect(2)) fails, we (close the socket
   * and) try the next address.
   */
  for (rp = result; rp != NULL; rp = rp->ai_next)
  {
    sockfd = socket(rp->ai_family, rp->ai_socktype, rp->ai_protocol);
    if (sockfd == -1)
    {
      continue;
    }
    if (fcntl(sockfd, F_SETFD, FD_CLOEXEC) < 0)
    {
      rinhw_error_printf("slip_socket_connect_init: error setting FD_CLOEXEC for socket connection to %s:%d (errno %d)\n", server, port, errno);
    }

    if (connect(sockfd, rp->ai_addr, rp->ai_addrlen) == 0)
    {
      /* Success */
      break;
    }

    close(sockfd);
  }

  if (rp == NULL)
  {
    /* No address succeeded */
    rinhw_error_printf("slip_socket_connect_init: could not connect to %s:%d\n", server, port);
    return RIO_ERROR;
  }

  /* getaddrinfo data no longer needed */
  freeaddrinfo(result);

  /* Set TCP no delay on the socket, so data is sent immediately */
  state = 1;
  if (setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, &state, sizeof(state)) < 0)
  {
    rinhw_error_printf("slip_socket_connect_init: failed to set TCP_NODELAY on port\n");
    return RIO_ERROR;
  }

  /* Setup slipio struct */
  memset(slipio, 0, sizeof(struct s_slipio));
  slipio->data = (void *) sockfd;
  slipio->slip_send_char = slip_send_char_socket_connect;
  slipio->slip_pre_send = slip_socket_connect_pre_send;
  slipio->slip_post_send = slip_socket_connect_post_send;
  slipio->slip_recv_char = slip_recv_char_socket_connect;
  slipio->slip_destroy = slip_socket_connect_destroy;

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("slip_socket_connect_init: finished\n");
#endif
  return RIO_SUCCESS;
}
