/*
 * slip_io_socket_listen.c
 *
 *  Created on: 23/10/2008
 *      Author: ianb
 */
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <rinhw/debug.h>
#include <rinhw/sio_local_socket.h>
#include <rinhw/rioctl.h>
#include <rinremote/slip.h>

#include <rinremote/slip_io.h>


/** Cork/uncork a TCP connection */
static void slip_socket_cork_tcp(struct t_io_dev *io_dev, bool cork)
{
  int fd;
  int state = (cork) ? 1 : 0;

  /* Find out the underlying Linux file descriptor */
  if (rioctl(io_dev, IOCTL_GET_LINUX_FD, &fd) < 0)
  {
    rinhw_error_printf("slip_socket_cork_tcp: failed to determine linux fd\n");
    return;
  }

  /* Cork/uncork the connection */
  if (setsockopt(fd, IPPROTO_TCP, TCP_CORK, &state, sizeof(state)) < 0)
  {
    rinhw_error_printf("slip_socket_cork_tcp: failed to %s port\n", (cork) ? "cork" : "uncork");
    return;
  }
}


/** Set TCP no delay once the connection is open */
static void slip_socket_tcp_nodelay(struct t_io_dev *io_dev)
{
  int fd;
  int state;

  /* Find out the underlying Linux file descriptor */
  if (rioctl(io_dev, IOCTL_GET_LINUX_FD, &fd) < 0)
  {
    rinhw_error_printf("slip_socket_tcp_nodelay: failed to determine linux fd\n");
    goto sstn_fail;
  }

  if (fd >= 0)
  {
    if (getsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &state, sizeof(state) < 0))
    {
      rinhw_error_printf("slip_socket_tcp_nodelay: get current state of TCP_NODELAY failed\n", errno);
      goto sstn_fail;
    }

    if (0 == state)
    {
      /* Set TCP no delay on the socket, so data is sent immediately */
      state = 1;
      if (setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &state, sizeof(state)) < 0)
      {
        rinhw_error_printf("slip_socket_init: failed to set TCP_NODELAY on port, errno %d\n", errno);
        goto sstn_fail;
      }
    }
  }

sstn_fail:
  ;
}


/** Things to do before sending a slip packet */
static void slip_socket_pre_send(void *data)
{
  struct t_io_dev *io_dev = data;

  /* Cork the port before the slip packet is written */
  if (NULL != io_dev)
  {
    slip_socket_tcp_nodelay(io_dev);
    slip_socket_cork_tcp(io_dev, true);
  }
}


/** Things to do after sending a slip packet */
static void slip_socket_post_send(void *data)
{
  struct t_io_dev *io_dev = data;

  /* Uncork the port after slip packet is written so it gets sent as one big chunk */
  if (NULL != io_dev)
    slip_socket_cork_tcp(io_dev, false);
}


static void slip_socket_send_char(void *data, uint8_t ch)
{
  struct t_io_dev *io_dev = data;

  if (NULL != io_dev)
    rwrite(io_dev, &ch, 1, 0);
}


static uint8_t slip_socket_recv_char(void *data)
{
  uint8_t ch = '\0';
  struct t_io_dev *io_dev = data;

  if (NULL != io_dev)
    rread(io_dev, &ch, 1, 0);
  return ch;
}


/** Clean up the slip socket (listening for connections) */
static void slip_socket_listen_destroy(void *data)
{
  struct t_io_dev *io_dev = data;

  if (NULL != io_dev)
  {
    rclose(io_dev);
    rdestroy_io(io_dev);
    free(io_dev);
  }
}


/** Initialise slip using a TCP socket (listening for connections) */
int slip_socket_listen_init(struct s_slipio *slipio, int socket)
{
  struct t_io_dev *sockiodev;

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("slip_socket_init() called\n");
#endif

  /* Allocate the io device */
  sockiodev = calloc(1, sizeof(struct t_io_dev));
  if (NULL == sockiodev)
  {
    rinhw_error_printf("slip_socket_init: couldn't allocate io device\n");
    goto ssi_fail1;
  }

  /* Configure and open the com port */
#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_FULL)
  DEBUG_PRINTF("slip_socket_init: setup TCP socket %d\n", socket);
#endif
  if (rio_create_device_sio_localskt(sockiodev, socket) < 0)
  {
    rinhw_error_printf("slip_socket_init: rio_create_sio_localskt failed\n");
    goto ssi_fail2;
  }

  /* Open device */
  if (ropen(sockiodev, 0) < 0)
  {
    rinhw_error_printf("slip_socket_init: ropen failed\n");
    goto ssi_fail3;
  }

  /* Setup slip io device */
  memset(slipio, 0, sizeof(struct s_slipio));
  slipio->data = sockiodev;
  slipio->slip_send_char = slip_socket_send_char;
  slipio->slip_pre_send = slip_socket_pre_send;
  slipio->slip_post_send = slip_socket_post_send;
  slipio->slip_recv_char = slip_socket_recv_char;
  slipio->slip_destroy = slip_socket_listen_destroy;

#if HW_DEBUG(REMOTE_IO_DEBUG, HW_DEBUG_BASIC)
  DEBUG_PRINTF("slip_socket_init() finished\n");
#endif
  return RIO_SUCCESS;

ssi_fail3:
  rdestroy_io(sockiodev);

ssi_fail2:
  free(sockiodev);

ssi_fail1:
  return RIO_ERROR;
}
