/*
 * remotedemux.c
 *
 *  Created on: 23/10/2008
 *      Author: ianb
 */

#include <rinremote/remote.h>
#include <rinhw/rio.h>
#include <rinhw/rinendian.h>
#include <rinhw/rstdio.h>
#include "remotehelper.h"
#include "remotedemuxioctl.h"
#include "devices.h"

#include "remotedemux.h"


struct t_io_dev *slip_device = NULL;


static int command_packet_lengths[ REMOTE_MAX_CMD + 1 ] =
{
  GET_SIZE_REMOTE_CMD_INIT(),
  GET_SIZE_REMOTE_CMD_OPEN(),
  GET_SIZE_REMOTE_CMD_CLOSE(),
  GET_SIZE_REMOTE_CMD_READ(),
  GET_SIZE_REMOTE_CMD_WRITE(0),
  GET_SIZE_REMOTE_CMD_ISATTY(),
  GET_SIZE_REMOTE_CMD_FILELENGTH(),
  GET_SIZE_REMOTE_CMD_LSEEK(),
  GET_SIZE_REMOTE_CMD_TELL(),
  IOCTL_CMD_HDR_SIZE,  /* This value is never used, as we need to manually inspect the packet in some cases */
  GET_SIZE_REMOTE_CMD_DESTROY(),
};


bool bCheckPacketSizeBasedOnCommand( const struct s_rem_device_command *prdCmd, int nLength )
{
  int8u ui8Command = prdCmd->rem_dev_cmd_header.ui8Command;
  int nBasePacketLength;

  if ( REMOTE_MAX_CMD < ui8Command )
  {
    return false;
  }

  nBasePacketLength = command_packet_lengths[ ui8Command ];
  if ( nLength < nBasePacketLength )
  {
    return false;
  }

  if ( REMOTE_CMD_WRITE == ui8Command &&
       nLength < command_packet_lengths[ ui8Command ] + prdCmd->rem_device_command.rem_dev_cmd_write.ui16Length )
  {
    return false;
  }

  /* TODO Check IOCTL length better */

  return true;
}


/** Get the size of the command, assuming the packet is good */
int nGetPacketSizeBasedOnCommand( const struct s_rem_device_command *prdCmd )
{
  int8u ui8Command = prdCmd->rem_dev_cmd_header.ui8Command;
  int nPacketLength;

  if ( REMOTE_CMD_IOCTL != ui8Command )
  {
    /* Get the length of the packet */
    nPacketLength = command_packet_lengths[ ui8Command ];

    /* Write is a special case because data follows the command */
    if ( REMOTE_CMD_WRITE == ui8Command )
    {
      nPacketLength += prdCmd->rem_device_command.rem_dev_cmd_write.ui16Length;
    }
  }
  else
  {
    /* IOCTL is a special case, and we may need to inspect the packet in some cases */
    nPacketLength = nGetCmdPacketSizeIoCtl( prdCmd );
  }

  return nPacketLength;
}


/** Check the basics of the command packet are ok */
bool bCheckPacketCommand( const struct s_rem_device_command *prdCmd, int nLength )
{
  if ( nLength < (int) sizeof( struct s_rem_dev_cmd_header ) )
  {
//    rfprintf( &dev_com2, "Header too small\n" );
    return false;
  }

  if ( !bCheckPacketSizeBasedOnCommand( prdCmd, nLength ) )
  {
//    rfprintf( &dev_com2, "Packet overall too small\n" );
    return false;
  }

//  rfprintf( &dev_com2, "Packet size ok\n" );
  return true;
}


typedef void (*cmd_handle_fn)( const struct s_rem_device_command *prdCmd,
                               struct s_rem_device_response *prdResponse, int *pnLength );


static void vHandleCommandInit( const struct s_rem_device_command *prdCmd,
                                struct s_rem_device_response *prdResponse, int *pnLength )
{
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );
  int32s nResult = ( 0 != dev ) ? RIO_SUCCESS : RIO_ERROR;

  /* Don't actually do anything */
  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_INIT, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_INIT();
}


static void vHandleCommandDestroy( const struct s_rem_device_command *prdCmd,
                                   struct s_rem_device_response *prdResponse, int *pnLength )
{
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );
  int32s nResult = ( 0 != dev ) ? RIO_SUCCESS : RIO_ERROR;

  /* Don't actually do anything */
  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_DESTROY, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_DESTROY();
}


static void vHandleCommandOpen( const struct s_rem_device_command *prdCmd,
                                struct s_rem_device_response *prdResponse, int *pnLength )
{
  int32s nResult;
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );

//  rfprintf( &dev_com2, "Open(), device %d,0x%lX, timeout %d\n",
//            prdCmd->rem_device_command.rem_dev_cmd_open.ui8Device, dev,
//            ntohs( prdCmd->rem_device_command.rem_dev_cmd_open.ui16Timeout ) );

  if ( 0 != dev && slip_device != dev )
  {
    nResult = ropen( dev, ntohs( prdCmd->rem_device_command.rem_dev_cmd_open.ui16Timeout ) );
  }
  else
  {
    nResult = RIO_ERROR;
  }

  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_OPEN, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_OPEN();
//  rfprintf( &dev_com2, "Result = %d, size = %d\n", nResult, *pnLength );
}


static void vHandleCommandClose( const struct s_rem_device_command *prdCmd,
                                 struct s_rem_device_response *prdResponse, int *pnLength )
{
  int32s nResult;
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );

//  rfprintf( &dev_com2, "Close(), device %d,0x%lX\n",
//            prdCmd->rem_device_command.rem_dev_cmd_open.ui8Device, dev );

  if ( 0 != dev && slip_device != dev )
  {
    nResult = rclose( dev );
  }
  else
  {
    nResult = RIO_ERROR;
  }

  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_CLOSE, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_CLOSE();
//  rfprintf( &dev_com2, "Result = %d, size = %d\n", nResult, *pnLength );
}


static void vHandleCommandRead( const struct s_rem_device_command *prdCmd,
                                struct s_rem_device_response *prdResponse, int *pnLength )
{
  int nResult;
  int nLength;
  const int8u ui8Device = prdCmd->rem_dev_cmd_header.ui8Device;
  const int16u ui16ReadLen = ntohs( prdCmd->rem_device_command.rem_dev_cmd_read.ui16Length );
  const int16u ui16ReadTimeout = ntohs( prdCmd->rem_device_command.rem_dev_cmd_read.ui16Timeout );
  struct t_io_dev *dev = GetDevice( ui8Device );

//  rfprintf( &dev_com2, "Read(), device %d,0x%lX, len=%d, timeout=%d\n",
//            ui8Device, dev, ui16ReadLen, ui16ReadTimeout );

  if ( 0 != dev && slip_device != dev )
  {
    nResult = rread( dev,
                     prdResponse->rem_device_response.rem_dev_resp_read.ui8Buffer,
                     ui16ReadLen, ui16ReadTimeout );
  }
  else
  {
    nResult = RIO_ERROR;
  }

  vInitRemoteRespHdr( prdResponse, REMOTE_CMD_READ, ui8Device );
  prdResponse->rem_device_response.rem_dev_resp_read.i32Result = htonl( nResult );
  nLength = (nResult == RIO_ERROR) ? 0 : nResult;
  *pnLength = (int) GET_SIZE_REMOTE_RESP_READ( nLength );

//  rfprintf( &dev_com2, "Read() complete: result %d, len %d\n",
//            nResult, nLength );
}


static void vHandleCommandWrite( const struct s_rem_device_command *prdCmd,
                                 struct s_rem_device_response *prdResponse, int *pnLength )
{
  int32s nResult;
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );

//  rfprintf( &dev_com2, "Close(), device %d,0x%lX, len=%d, timeout=%d\n",
//            prdCmd->rem_device_command.rem_dev_cmd_open.ui8Device, dev,
//            ntohs( prdCmd->rem_device_command.rem_dev_cmd_write.ui16Length ),
//            ntohs( prdCmd->rem_device_command.rem_dev_cmd_write.ui16Timeout ) );

  if ( 0 != dev && slip_device != dev )
  {
    nResult = rwrite( dev,
                      prdCmd->rem_device_command.rem_dev_cmd_write.ui8Buffer,
                      ntohs( prdCmd->rem_device_command.rem_dev_cmd_write.ui16Length ),
                      ntohs( prdCmd->rem_device_command.rem_dev_cmd_write.ui16Timeout ) );
  }
  else
  {
    nResult = RIO_ERROR;
  }

  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_WRITE, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_WRITE();
//  rfprintf( &dev_com2, "Result = %d, size = %d\n", nResult, *pnLength );
}


static void vHandleCommandIsATty( const struct s_rem_device_command *prdCmd,
                                  struct s_rem_device_response *prdResponse, int *pnLength )
{
  int32s nResult;
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );

  if ( 0 != dev && slip_device != dev )
  {
    nResult = risatty( dev );
  }
  else
  {
    nResult = RIO_ERROR;
  }

  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_ISATTY, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_ISATTY();
}


static void vHandleCommandFileLength( const struct s_rem_device_command *prdCmd,
                                      struct s_rem_device_response *prdResponse, int *pnLength )
{
  long nResult;
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );

  if ( 0 != dev && slip_device != dev )
  {
    nResult = rfilelength( dev );
  }
  else
  {
    nResult = RIO_ERROR;
  }

  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_FILELENGTH, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_FILELENGTH();
}


static void vHandleCommandLSeek( const struct s_rem_device_command *prdCmd,
                                 struct s_rem_device_response *prdResponse, int *pnLength )
{
  long nResult;
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );

  if ( 0 != dev && slip_device != dev )
  {
    nResult = rlseek( dev,
                      ntohl( prdCmd->rem_device_command.rem_dev_cmd_lseek.i32Offset ),
                      (int) ntohl( prdCmd->rem_device_command.rem_dev_cmd_lseek.i32Fromwhere ) );
  }
  else
  {
    nResult = RIO_ERROR;
  }

  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_LSEEK, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_LSEEK();
}


static void vHandleCommandTell( const struct s_rem_device_command *prdCmd,
                                struct s_rem_device_response *prdResponse, int *pnLength )
{
  long nResult;
  struct t_io_dev *dev = GetDevice( prdCmd->rem_dev_cmd_header.ui8Device );

  if ( 0 != dev && slip_device != dev )
  {
    nResult = rtell( dev );
  }
  else
  {
    nResult = RIO_ERROR;
  }

  vInitRemoteResp_Default( prdResponse, REMOTE_CMD_TELL, prdCmd->rem_dev_cmd_header.ui8Device, nResult );
  *pnLength = GET_SIZE_REMOTE_RESP_TELL();
}


static const cmd_handle_fn cmd_fns[ REMOTE_MAX_CMD + 1 ] =
{
  vHandleCommandInit,
  vHandleCommandOpen,
  vHandleCommandClose,
  vHandleCommandRead,
  vHandleCommandWrite,
  vHandleCommandIsATty,
  vHandleCommandFileLength,
  vHandleCommandLSeek,
  vHandleCommandTell,
  vHandleCommandIoCtl,
  vHandleCommandDestroy,
};


/*
#define REMOTE_CMD_INIT 0
#define REMOTE_CMD_OPEN 1
#define REMOTE_CMD_CLOSE 2
#define REMOTE_CMD_READ 3
#define REMOTE_CMD_WRITE 4
#define REMOTE_CMD_ISATTY 5
#define REMOTE_CMD_FILELENGTH 6
#define REMOTE_CMD_LSEEK 7
#define REMOTE_CMD_TELL 8
#define REMOTE_CMD_IOCTL 9
#define REMOTE_CMD_DESTROY 10
#define REMOTE_MAX_CMD 10
*/

void vHandleGoodPacket( const struct s_rem_device_command *prdCmd,
                        struct s_rem_device_response *prdResponse, int *pnLength )
{
//  rfprintf( &dev_com2, "Calling command function %d\n", prdCmd->rem_dev_cmd_header.ui8Command );
  cmd_fns[ prdCmd->rem_dev_cmd_header.ui8Command ]( prdCmd, prdResponse, pnLength );
}

