/*
 * remotedemuxhandler.c
 *
 *  Created on: 21/08/2009
 *      Author: ianb
 */
#include <string.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdbool.h>
#include <stdint.h>
#include <rinhw/rstdio.h>
#include <rinhw/power.h>
#include <rinhw/rinthread.h>
#include <rinhw/sio_local_comport.h>
#include <rinremote/slip.h>
#include <rinremote/slip_io.h>
#include "remotedemux.h"
#include "remoteserial.h"
#include "remotehelper.h"
#include "remote_debug.h"

#include "remotedemuxhandler.h"


/* Each one needs a task */
#define MAX_CONCURRENT_COMMANDS 5


/* Structure to */
struct sRemoteCommandHandlerInfo
{
  bool bInUse;
  sem_t pSignalSem;
  uint8_t rxbuffer[MAX_COMMAND_BUFFER];
  struct s_rem_device_command *prdCmd; /* Set to point to rxbuffer */
  int nLength;
  uint8_t CurrentDevice;
  pthread_t pthread_id;
};

static pthread_mutex_t s_RemCmdHandlers_mutex;
static struct sRemoteCommandHandlerInfo s_RemCmdHandlers[MAX_CONCURRENT_COMMANDS];
static struct t_io_dev remote_serial_io;
static bool remote_serial_io_needs_destroy = false;
static struct s_slipio remote_slipio;
static pthread_t remote_rx_thread;


/** Task to run each command, and send the response */
static void *vTaskHandleRemoteCommand(void *data)
{
  struct sRemoteCommandHandlerInfo *pRemoteCmd = (struct sRemoteCommandHandlerInfo *) data;
  uint8_t txbuffer[MAX_RESPONSE_BUFFER];
  struct s_rem_device_response *prdResponse = (struct s_rem_device_response *) txbuffer;
  int nTxLength;

  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  while (1)
  {
    /* Wait for something to deal with */
    sem_wait(&pRemoteCmd->pSignalSem);

    /* Handle it, and send the reply */
    vHandleGoodPacket(pRemoteCmd->prdCmd, prdResponse, &nTxLength);
    vExclusiveSlipSendPacket(&remote_slipio, txbuffer, nTxLength);

    /* Not busy any more */
    pthread_mutex_lock(&s_RemCmdHandlers_mutex);
    pRemoteCmd->bInUse = false;
    pRemoteCmd->CurrentDevice = REMOTE_DEV_INVALID;
    pthread_mutex_unlock(&s_RemCmdHandlers_mutex);
  }
  rin_thread_exit(NULL);
}


static struct sRemoteCommandHandlerInfo *pGetNextAvailableCommandHandler(void)
{
  size_t i;
  struct sRemoteCommandHandlerInfo *pRemCmdHdler = (struct sRemoteCommandHandlerInfo *) NULL;

  /* TODO implement with a semaphore so it will block till available */

  pthread_mutex_lock(&s_RemCmdHandlers_mutex);
  for (i = 0; i < MAX_CONCURRENT_COMMANDS; i++)
  {
    if (!s_RemCmdHandlers[i].bInUse)
    {
      s_RemCmdHandlers[i].bInUse = true;
      pRemCmdHdler = &s_RemCmdHandlers[i];
      break;
    }
  }
  pthread_mutex_unlock(&s_RemCmdHandlers_mutex);

  return pRemCmdHdler;
}


/** Check if a device is already handling a command */
static bool bIsDeviceAlreadyHandlingCommand(int8u ui8Device)
{
  size_t i;
  bool bResult = false;

  pthread_mutex_lock(&s_RemCmdHandlers_mutex);
  for (i = 0; i < MAX_CONCURRENT_COMMANDS; i++)
  {
    if (s_RemCmdHandlers[i].bInUse &&
        s_RemCmdHandlers[i].CurrentDevice == ui8Device)
    {
      bResult = true;
      break;
    }
  }
  pthread_mutex_unlock(&s_RemCmdHandlers_mutex);

  return bResult;
}


static void vSendErrorResponse(const struct s_rem_device_command *prdCmd)
{
  uint8_t aui8TxBuffer[GET_SIZE_REMOTE_RESP_DEFAULT()];
  struct s_rem_device_response *prdResponse = (struct s_rem_device_response *) aui8TxBuffer;
  int nTxLength;

  vDefaultErrorResponse(prdCmd, prdResponse, &nTxLength);
  vExclusiveSlipSendPacket(&remote_slipio, aui8TxBuffer, nTxLength);
}


/** Notify the master that the R400 has reset */
static void vSendResetDeviceResponse(void)
{
  int nTxLength;
  uint8_t aui8RxBuffer[GET_SIZE_REMOTE_CMD_INIT()];
  struct s_rem_device_command *prdCmd = (struct s_rem_device_command *) aui8RxBuffer;
  uint8_t aui8TxBuffer[GET_SIZE_REMOTE_RESP_DEFAULT()];
  struct s_rem_device_response *prdResponse = (struct s_rem_device_response *) aui8TxBuffer;

  prdCmd->rem_dev_cmd_header.ui8Command = REMOTE_CMD_INIT;
  prdCmd->rem_dev_cmd_header.ui8Device = REMOTE_DEV_RESET;

  /* Send a success response */
  vDefaultSuccessResponse(prdCmd, prdResponse, &nTxLength);
  vExclusiveSlipSendPacket(&remote_slipio, aui8TxBuffer, nTxLength);
}


/** Reset the R400 */
static void vHandleResetDevice(struct s_rem_device_command *prdCmd)
{
#ifdef __HITACHI__
  /* Just reset the R400. When it starts up, it will send the reset successful response. */
  SystemReset();
#else
  vSendResetDeviceResponse();
#endif
}


static void vHandleRemoteCommand(const struct s_rem_device_command *prdCmd)
{
  struct sRemoteCommandHandlerInfo *pRemCmdHdler;

  /* Get next available command handler */
  pRemCmdHdler = pGetNextAvailableCommandHandler();
  if (NULL != pRemCmdHdler)
  {
    /* Copy the command into the command handler */
    /* *pRemCmdHdler->prdCmd = *prdCmd; This won't work for WRITES */
    memcpy(pRemCmdHdler->prdCmd, prdCmd, (size_t) nGetPacketSizeBasedOnCommand(prdCmd));
    pRemCmdHdler->CurrentDevice = prdCmd->rem_dev_cmd_header.ui8Device;

    /* Wake the task */
    sem_post(&pRemCmdHdler->pSignalSem);
  }
  else
  {
    /* Not enough handlers available, so send a default error response */
    rdebug_printf(LOG_ERR, "No handlers available\n");
    vSendErrorResponse(prdCmd);
  }
}


/** Setup the command handler structures */
static void vInitCommandHandlers(void)
{
  size_t i;

  pthread_mutex_init(&s_RemCmdHandlers_mutex, NULL);

  for (i = 0; i < MAX_CONCURRENT_COMMANDS; i++)
  {
    memset(&s_RemCmdHandlers[i], 0, sizeof(s_RemCmdHandlers[i]));

    sem_init(&s_RemCmdHandlers[i].pSignalSem, 0, 0);
    s_RemCmdHandlers[i].bInUse = false;
    s_RemCmdHandlers[i].nLength = 0;
    s_RemCmdHandlers[i].prdCmd = (struct s_rem_device_command *) s_RemCmdHandlers[i].rxbuffer;
    s_RemCmdHandlers[i].CurrentDevice = REMOTE_DEV_INVALID;

    rin_thread_create(&s_RemCmdHandlers[i].pthread_id, vTaskHandleRemoteCommand, (void *) &s_RemCmdHandlers[i], true);
  }
}


/** Shutdown the command handler structures */
static void vDestroyCommandHandlers(void)
{
  size_t i;

  for (i = 0; i < MAX_CONCURRENT_COMMANDS; i++)
  {
    (void) pthread_cancel(s_RemCmdHandlers[i].pthread_id);
    (void) pthread_join(s_RemCmdHandlers[i].pthread_id, NULL);
    sem_destroy(&s_RemCmdHandlers[i].pSignalSem);
  }

  pthread_mutex_destroy(&s_RemCmdHandlers_mutex);
}


/* Incoming commands get put in here before being copied to the dispatch buffers */
static uint8_t rxbuffer_primary[MAX_COMMAND_BUFFER];


/** Task for the remote device command listener */
static void *RemoteListenTask(void *data)
{
  struct s_rem_device_command *prdCmd = (struct s_rem_device_command *) rxbuffer_primary;
  int nRxLength;

  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

#ifdef __HITACHI__
  /** Notify the master that the R400 has reset */
  vSendResetDeviceResponse();
#endif

  /* Listen for commands, and dispatch them to the command handlers */
  while (1)
  {
    /* Get a packet from the serial port */
    nRxLength = slip_recv_packet(&remote_slipio, rxbuffer_primary, MAX_COMMAND_BUFFER);

    /* First check that the basics are ok */
    if (!bCheckPacketCommand(prdCmd, nRxLength))
    {
      rdebug_printf(LOG_ERR, "Packet command bad\n");
      continue;
    }

    /* Special case, the "reset" device - any command executed on this will cause a reset of the R400 */
    if (REMOTE_DEV_RESET == prdCmd->rem_dev_cmd_header.ui8Device)
    {
      vHandleResetDevice(prdCmd);
      continue;
    }

    /* Check that the device isn't already handling a command */
    if (bIsDeviceAlreadyHandlingCommand(prdCmd->rem_dev_cmd_header.ui8Device))
    {
      rdebug_printf(LOG_ERR, "Command already being handled\n");
      vSendErrorResponse(prdCmd);
      continue;
    }

    /* Things seem in order, so dispatch to handler */
    vHandleRemoteCommand(prdCmd);
  }

  /* Clean up */
  return NULL;
}


static bool create_slip_io(struct s_slipio *slipio, const char *serial_device, int tcpport)
{
  t_ioctl_siocfg sercfg;

  if (NULL != serial_device)
  {
    if (rio_create_device_sio_localcom(&remote_serial_io, serial_device, NULL, false) < 0)
    {
      rdebug_printf(LOG_ERR, "create_slip_io: failed to create serial device for %s\n", serial_device);
      return false;
    }

    ropen(&remote_serial_io, 0);

    /* Configure serial port */
    sercfg.baud = 57600;
    sercfg.data_bits = SIO_DATABITS_EIGHT;
    sercfg.parity_bits = SIO_PARITY_NONE;
    sercfg.stop_bits = SIO_STOPBITS_ONE;
    sercfg.use_cts = 0;
    sercfg.use_terminations = 0;
    if (rioctl(&remote_serial_io, IOCTL_SER_SET_CONFIG, &sercfg) != RIO_SUCCESS)
    {
      rdebug_printf(LOG_ERR, "create_slip_io: failed configure serial port %s\n", serial_device);
      return false;
    }

    if (slip_init_rio_dev(&remote_slipio, &remote_serial_io) != RIO_SUCCESS)
    {
      rdebug_printf(LOG_ERR, "create_slip_io: failed to create slip serial device for %s\n", serial_device);
      return false;
    }
    remote_serial_io_needs_destroy = true;
  }
  else if (tcpport >= 0)
  {
    if (slip_socket_listen_init(&remote_slipio, tcpport) != RIO_SUCCESS)
    {
      rdebug_printf(LOG_ERR, "create_slip_io: failed to create slip listening tcp on port %d\n", tcpport);
      return false;
    }
    remote_serial_io_needs_destroy = false;
  }
  else
  {
    rdebug_printf(LOG_ERR, "create_slip_io: no serial port, or tcp port specified\n");
    return false;
  }

  /* Success */
  return true;
}


/** Start the remote device command listener */
bool remote_listen_init(const char *serial_device, int tcpport)
{
  /* Initialise slip */
  if (!create_slip_io(&remote_slipio, serial_device, tcpport))
  {
    return false;
  }

  /* Initialise exclusive access to sending stuff on the serial port */
  vInitExclusiveSlipSendPacket();

  /* Initialise the command handlers */
  vInitCommandHandlers();

  /* Create thread */
  return rin_thread_create(&remote_rx_thread, RemoteListenTask, NULL, true);
}


/** Shutdown the command handler structures */
void remote_listen_destroy(void)
{
  /* Shut down listener thread */
  (void) pthread_cancel(remote_rx_thread);
  (void) pthread_join(remote_rx_thread, NULL);

  /* Clean up */
  vDestroyCommandHandlers();
  vDestroyExclusiveSlipSendPacket();
  slip_destroy(&remote_slipio);
  if (remote_serial_io_needs_destroy)
    rdestroy_io(&remote_serial_io);
}
