/*
 * alibi_print.c
 *
 * Printouts look like this:
 *
 * 00000009 2003\07\31 07:26:43
 *   2204.0 kg G
 * - 2878.0 kg N
 *   5082.0 kg T
 *
 */
#include <stdio.h>
#include <string.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcalibi.h>
#include "alibi_display.h"
#include "alibi_weight.h"

#include "alibi_print.h"


/** "  2204.0" 6 digits + 1 decimal point + 1 sign + null terminator */
#define WEIGHT_STR_LEN (6 + 1 + 1 + 1)


/** "  5082.0 kg PT" (Weight + dp) + space + units + space + status + null terminator */
#define WEIGHT_FMT_STR_LEN (WEIGHT_STR_LEN + 1 + 2 + 1 + 2 + 1)


/** Printout length: id + date + time + "\r\n" + 3 weights + null terminator */
#define PRINTOUT_STR_LEN (28 + (7 * 2) + WEIGHT_FMT_STR_LEN * 3 + 1)


/** Get an abbreviated units string for a unitsidx */
const char *AdcIntAlibiGetAbbreviatedUnitsString(struct s_adc_alibi_info *adc_alibi_info, int8u unitsidx)
{
  const char *ptr = NULL;

  adc_alibi_info->adc_alibi_get_abbreviated_units_string(adc_alibi_info->implementer_data, unitsidx, &ptr);
  if (NULL == ptr)
  {
    return " ";
  }
  return ptr;
}


static void AdcIntAlibiWeightToPrintStr(struct s_adc_alibi_info *adc_alibi_info, char *str, int strlength, const AdcRdg *rdg)
{
  char *status;
  char weight_str[WEIGHT_STR_LEN];
  const char *units_str;

  /* Status */
  if (rdg->info.status.bit.oload)
    status = "O";
  else if (rdg->info.status.bit.uload)
    status = "U";
  else if (rdg->info.status.bit.motion)
    status = "M";
  else if (rdg->info.status.bit.grossOrNet)
  {
    if (rdg->info.status.bit.gross)
      status = "G";
    else
      status = "N";
  }
  else if (rdg->info.status.bit.tare)
  {
    status = "T";
  }
  else if (rdg->info.status.bit.preTare)
  {
    status = "PT";
  }
  else
  {
    status = " ";
  }

  /* Format weight as a string */
  if (ADC_APPROVAL_MODE_NTEP == adc_alibi_info->ass->adcUserDatabase.approvalMode)
  {
/* TODO ntep
    FormatNumDP (wgtStr, rdg.rdg, 8, ' ', ' ', '-', rdg.info.unitdp.bit.dp, '.',
      AdcResLen[adcUserDatabase.resolution[rdg.info.status.bit.range]]); */

    /* num_digits = WEIGHT_STR_LEN - 1 (for dp) - 1 (for null terminator) */
    AdcIntAlibiWeightToStr(adc_alibi_info, rdg, WEIGHT_STR_LEN - 1 - 1, weight_str, WEIGHT_STR_LEN);
  }
  else
  {
    /* num_digits = WEIGHT_STR_LEN - 1 (for dp) - 1 (for null terminator) */
    AdcIntAlibiWeightToStr(adc_alibi_info, rdg, WEIGHT_STR_LEN - 1 - 1, weight_str, WEIGHT_STR_LEN);
  }

  /* Get units */
  units_str = AdcIntAlibiGetAbbreviatedUnitsString(adc_alibi_info, rdg->info.unitdp.bit.units);

  /* Make a string */
  snprintf(str, strlength, "%s %s %s", weight_str, units_str, status);
}


/** Generate a traceable printout */
void AdcIntAlibiTraceablePrint(struct s_adc_alibi_info *adc_alibi_info)
{
  char w1[WEIGHT_FMT_STR_LEN];
  char w2[WEIGHT_FMT_STR_LEN];
  char w3[WEIGHT_FMT_STR_LEN];
  char printstr[PRINTOUT_STR_LEN];
  AdcRdg rdg;
  AdcRdg tarerdg;
  AdcTraceableWeight atw;

  atw = adc_alibi_info->ass->adcTraceableWeight;

  /* Get weights */
  rdg = atw.rdg;
  if (rdg.info.status.bit.gross)
  {
    rdg.rdg = rdg.rdg - atw.tareWgt;
    rdg.info.status.bit.gross = false;
  }
  else
  {
    rdg.rdg = rdg.rdg + atw.tareWgt;
    rdg.info.status.bit.gross = true;
  }

  /* Convert gross and net weights to string: "weight units status" */
  if (rdg.info.status.bit.gross)
  {
    AdcIntAlibiWeightToPrintStr(adc_alibi_info, w1, WEIGHT_FMT_STR_LEN, &rdg);
    AdcIntAlibiWeightToPrintStr(adc_alibi_info, w2, WEIGHT_FMT_STR_LEN, &atw.rdg);
  }
  else
  {
    AdcIntAlibiWeightToPrintStr(adc_alibi_info, w1, WEIGHT_FMT_STR_LEN, &atw.rdg);
    AdcIntAlibiWeightToPrintStr(adc_alibi_info, w2, WEIGHT_FMT_STR_LEN, &rdg);
  }

  /* Convert tare/preset tare weight to string: "weight units status" */
  tarerdg.rdg = atw.tareWgt;
  tarerdg.info.unitdp = rdg.info.unitdp;
  memset(&tarerdg.info.status, 0, sizeof(tarerdg.info.status));
  tarerdg.info.status.bit.grossOrNet = 0;
  tarerdg.info.status.bit.preTare = (0 != atw.pTare);
  tarerdg.info.status.bit.tare = !tarerdg.info.status.bit.preTare;
  AdcIntAlibiWeightToPrintStr(adc_alibi_info, w3, WEIGHT_FMT_STR_LEN, &tarerdg);

  /* Convert all to string and print */
  snprintf(printstr, PRINTOUT_STR_LEN, "\r\n%08lu %04d\\%02d\\%02d %02d:%02d:%02d\r\n"
                                       "%s\r\n"
                                       "%s\r\n"
                                       "%s\r\n"
                                       "\r\n",
           (unsigned long) atw.printId,
           atw.dt.year,
           atw.dt.month,
           atw.dt.date,
           atw.dt.hour,
           atw.dt.minute,
           atw.dt.second,
           w1,
           w2,
           w3);

  /* Print the string */
  adc_alibi_info->adc_alibi_print_string(adc_alibi_info->implementer_data, printstr);
}
