# RinCmd

## Overview

RinCmd is a protocol for reading/writing data out of scales that is used extensively by the Lua module.
The format should be fairly well documented in the R400 and R300 manuals, which served as the basis for this application.
This application should be well documented with in-code comments, and usage should be extensively tested in librinhw/rincmd_test.

## XML Overview

RinCmd contains a python script (`rincmd/rc_parseXML.py`) to parse an XML file into C for the application and Latex for the documentation.
The input XML should be titled of the form `registers_TYPE.xml`, where `TYPE` is the block of registers this file contains (e.g. streaming, lua, weight, etc.).
An example XML file can be found in `rincmd/registers_demo.xml`.

Some notes:
* The min and max characters of a string are characters in the ascii table.
* The max length of a string is the maximum number of characters NOT INCLUDING the terminator. Ensure the buffer is maxlen + 1.
* As the XML file is converted to C, `#define`'d symbols may be used if the header is included.
* The value of a register should be an constant pointer to that variable. For strings, a pointer to the string is expected.
* XML's use ampersand as a control character. Use `&amp;` instead.
* Bools can be stored as a `uint8` register. Be sure to cast the pointer to `uint8_t *` though, e.g. `<value>(uint8_t *) &amp;val</value>`
* Functions associated with a read/write register are called pre and post read/write. Functions associated with an exec register are called once.
* Function pointers do not require an ampersand, just input the function name, e.g. `<function>doStuff</function>`
* Functions should match the prototype defined in `rc_register`. At the time of writing, this is `bool  (*regFunc)(char *val,  size_t len, struct s_rc_localRegisterTable *topLevel, struct rc_Register *reg, uint8_t opt);`. Most of these arguments are self-explanatory, aside from `opt` which represents one of the `Register function defines` in `rincmd/rc_public.h`.

## Features and Implementation

### Generating Register Sets

Register sets (C files) are generated using rc_registerBuilder.py and an XML file.
This enforces separation between the definitions of the registers and the register handling code, ensuring that the code remains portable.
The generated registers should then be included by the application code, e.g.

    /* Initialise the global register table */
    rinCmdTable = rc_initRegTable();
    rc_gen_base(rinCmdTable);
    
    /* For each comm port */
    commport->regTable = rc_initLocalRegTable(rinCmdTable);
    rc_gen_local_basic(commport->regTable);
    
Upon application termination the data for each comm port should be freed, and the global table should be freed once.
	
	/* For each comm port */
	rc_freeLocalRegTable(commport->regTable);
	
	/* Free the register table */
	rc_freeRegTable(rinCmdTable);
	
An example register set can be found in `librinhw/rincmd_test`.
Multiple register sets can be joined into the global and local register tables, allowing for XML's to be broken up into sensible groups (e.g. Traceable weight register set, Lua register set).

### Read, write, execute registers

Data read from a comms buffer is passed into `rc_register.c:rc_commandHandler()`, along with the register table, the privilege setting and the indicator address (storing this data allows for the application to have independent comm port privileges and indicator addresses).
The data is then parsed into a struct where it can be decided if the message is a read, write, or execute type message, and is then passed to one of these handlers.
The data is interpretted by these handlers based on the register type (e.g. `UINT16`), and if it is a write message the data buffer is updated.
Finally, the struct is rebuilt into a string and returned with a flag, allowing the application to send the reply back to the relevant comm port if the data is valid.

When using an **execute** register, the 'maxLen' parameter determines how much data can be written back into the buffer.
When data is written back into the buffer from an **execute** register, the function should return **true** to show that this buffer has been modified, and return **false** otherwise.

### Local registers

Local registers are registers with comm-port independent data (e.g. streaming, permissions).
This means that streaming registers are independent of comm ports used (e.g. stream register 0x0040 can be streaming gross weight at 5Hz on comm port 1, while simultaneously streaming at 10Hz on comm port 2).
These should be specified under the `<local></local>` group in the register set, see `librinhw/rincmd_test/registers.xml` for an example.

