'''
Rinstrum automatic output client

Created on Oct 11, 2017

@author: ian
'''
import argparse
import time
import rc_comms
import rc_frame


class PyRinCmdAutoOutputClient(object):
    """Rinstrum automatic output client"""

    def __init__(self, *args, **kwargs):
        '''
        Create a PyRinCmdAutoOutputClient using TCP:
          @param tcphost is the host to connect to (ip address or hostname)
          @param tcpport is the TCP port on the host to connect to (default 2222)

        Create a PyRinCmdAutoOutputClient using UDP:
          @param udphost is the host to connect to (ip address or hostname)
          @param udpport is the UDP port on the host to connect to
          @param udplocalport Specify this if you want a the client to bind to a specific local UDP port. 0 means the kernel will choose one for you.

        Create a PyRinCmdAutoOutputClient using serial:
          @param serialport is the serial port to connect to the host with (eg COMx, /dev/ttySx)
          @param settings is the serial settings in a string like "9600n81"

        General parameters:
          @param timeout is the timeout period to wait for a response
        '''

        timeout = kwargs.get('timeout', rc_comms.PyRinCmdCommsBase.DEFAULT_TIMEOUT)

        tcphost = kwargs.get('tcphost')
        if (tcphost is not None):
            tcpport = kwargs.get('tcpport', 2223)
            self.comms = rc_comms.PyRinCmdCommsTcp(host=tcphost, port=tcpport, timeout=timeout)
        else:
            udphost = kwargs.get('udphost')
            udpport = kwargs.get('udpport')
            if (udphost is not None and udpport is not None):
                udplocalport = kwargs.get('udplocalport', 0)
                self.comms = rc_comms.PyRinCmdCommsUdp(host=udphost, port=udpport, localudpport=udplocalport, timeout=timeout)
            else:
                serialport = kwargs.get('serialport')
                if (serialport is not None):
                    settings = kwargs.get('settings', "9600n81")
                    self.comms = rc_comms.PyRinCmdCommsSerial(port=serialport, settings=settings, timeout=timeout)
                else:
                    raise Exception("unknown parameters")

        # Other parameters
        frame_prefix = kwargs.get('frame_prefix', "\x02")
        frame_postfix = kwargs.get('frame_postfix', "\x03")
        self.framer = rc_frame.PyRinCmdFrameAutoOut(resp_prefix=frame_prefix, resp_postfix=frame_postfix)
        self.indata = ""

    def connect(self):
        self.comms.connect()

    def disconnect(self):
        self.comms.disconnect()

    def find_message(self, reply):
        """Look for a validly framed message in the reply"""
        before, message, after = self.framer.deframe_response(response=reply)
        if (before is None and message is None):
            return False, "", after
        else:
            #print("Got reply {}".format(message))
            return True, message, after

    def senddata(self, data):
        """Send data to the host"""
        framed_data = self.framer.frame_command(cmd=data)
        self.comms.write(data=framed_data)

    def recvdata(self):
        """Receive data from the host"""
        return self.comms.read()

    def receive_message(self):
        """Receive an auto output message"""
        if (not self.comms.is_connected()):
            raise Exception("PyRinCmdClient is not connected")
        replystr = ""
        foundmsg = False
        while (not foundmsg):
            self.indata += self.recvdata()
            foundmsg, replystr, self.indata = self.find_message(reply=self.indata)
        return replystr


def main(argv=None):
    """Simple command line auto output client"""

    # Parse command line options
    parser = argparse.ArgumentParser(description='Run the pyrincmd automatic output client')
    parser.add_argument('host', action='store',
                        help='The host to connect to, or serial port if --protocol=serial')
    parser.add_argument('--port', action='store', default=2223,
                        help='The TCP/UDP port to connect to the host on')
    parser.add_argument('--protocol', action='store', choices=["tcp", "udp", "serial"], default="tcp",
                        help='The protocol to communicate using (tcp, udp, serial), default tcp')
    parser.add_argument('--serial-settings', action='store', default="9600n81",
                        help='The serial settings to use when --protocol=serial is used, default 9600n81')
    parser.add_argument('--udp-local-port', action='store', type=int, default=0,
                        help='Set the local udp port to bind to. Default 0 (kernel choose unused).')
    parser.add_argument('--count', action='store', type=int, default=0,
                        help='The number of messages to receive, 0 for indefinate.')
    parser.add_argument('--frame-prefix', action='store', default="\\x02",
                        help='The message framing prefix. Default: \\x02 (STX). Escape non-printables with python/c escaping \\xXX.')
    parser.add_argument('--frame-postfix', action='store', default="\\x03",
                        help='The message framing prefix. Default: \\x03 (ETX). Escape non-printables with python/c escaping \\xXX.')
    parser.add_argument('--show-framing', action='store_true',
                        help='Display the framing when printing messages.')
    parser.add_argument('--udp-wakeup', action='store', type=float, default=0.0,
                        help='On UDP ports, send wakeup data.')
    args = parser.parse_args(args=argv)

    # De-escape framing
    frame_prefix = args.frame_prefix.decode('string-escape')
    frame_postfix = args.frame_postfix.decode('string-escape')

    # Connect to host
    if (args.protocol == 'tcp'):
        client = PyRinCmdAutoOutputClient(tcphost=args.host, tcpport=int(args.port),
                                          frame_prefix=frame_prefix, frame_postfix=frame_postfix)
    elif (args.protocol == 'udp'):
        client = PyRinCmdAutoOutputClient(udphost=args.host, udpport=int(args.port), udplocalport=args.udp_local_port,
                                          frame_prefix=frame_prefix, frame_postfix=frame_postfix)
    elif (args.protocol == 'serial'):
        client = PyRinCmdAutoOutputClient(serialport=args.host, settings=args.serial_settings,
                                          frame_prefix=frame_prefix, frame_postfix=frame_postfix)

    client.connect()
    try:
        count = 0
        last_time = 0.0
        while (args.count == 0 or count < args.count):
            # Send wakeup
            if (args.udp_wakeup > 0):
                now = time.time()
                if (now > last_time + args.udp_wakeup):
                    client.senddata(data=" ")
                    last_time = now

            # Read message
            message = client.receive_message()
            if (args.show_framing):
                message = client.framer.frame_response(resp=message)
            fmtmessage = message.encode('string-escape')
            print(fmtmessage)
            count += 1

    finally:
        client.disconnect()


if (__name__ == '__main__'):
    main()
