'''
Created on Aug 28, 2017

@author: ian
'''
import re
import argparse
import rc_message
import rc_error
import rc_frame
import rc_comms
import rc_register


class PyRinCmdClient(object):
    """Rinstrum rincmd client"""

    def __init__(self, *args, **kwargs):
        '''
        Create a PyRinCmdClient using TCP:
          @param tcphost is the host to connect to (ip address or hostname)
          @param tcpport is the TCP port on the host to connect to (default 2222)

        Create a PyRinCmdClient using UDP:
          @param udphost is the host to connect to (ip address or hostname)
          @param udpport is the UDP port on the host to connect to
          @param udplocalport Specify this if you want a the client to bind to a specific local UDP port. 0 means the kernel will choose one for you.

        Create a PyRinCmdClient using UDS (Unix Domain Socket):
          @param udsfile is the server's unix domain socket file to connect to

        Create a PyRinCmdClient using serial:
          @param serialport is the serial port to connect to the host with (eg COMx, /dev/ttySx)
          @param settings is the serial settings in a string like "9600n81"

        General parameters:
          @param register_list Not used for now, leave as None
          @param rc_addr is the rin cmd device address (0..31). 0 is broadcast and generally the best option
          @param rc_channel is the channel data to use in the rincmd header
          @param usecrc requires the use of CRC on messages
          @param timeout is the timeout period to wait for a response
          @param logfile is a file like object or filename to write the comms log to
        '''

        timeout = kwargs.get('timeout', rc_comms.PyRinCmdCommsBase.DEFAULT_TIMEOUT)
        logfile = kwargs.get('logfile', None)

        tcphost = kwargs.get('tcphost')
        if (tcphost is not None):
            tcpport = kwargs.get('tcpport', 2222)
            self.comms = rc_comms.PyRinCmdCommsTcp(host=tcphost, port=tcpport, timeout=timeout, logfile=logfile)
        else:
            udphost = kwargs.get('udphost')
            udpport = kwargs.get('udpport')
            if (udphost is not None and udpport is not None):
                udplocalport = kwargs.get('udplocalport', 0)
                self.comms = rc_comms.PyRinCmdCommsUdp(host=udphost, port=udpport, localudpport=udplocalport, timeout=timeout, logfile=logfile)
            else:
                udsfile = kwargs.get('udsfile')
                if (udsfile is not None):
                    self.comms = rc_comms.PyRinCmdCommsUds(udsfile=udsfile, timeout=timeout, logfile=logfile)
                else:
                    serialport = kwargs.get('serialport')
                    if (serialport is not None):
                        settings = kwargs.get('settings', "9600n81")
                        self.comms = rc_comms.PyRinCmdCommsSerial(port=serialport, settings=settings, timeout=timeout, logfile=logfile)
                    else:
                        raise Exception("unknown parameters")

        # Other parameters
        self.register_list = kwargs.get('register_list')
        self.rc_addr = kwargs.get('rc_addr', rc_message.PyRinCmdMessage.ADDR_BROADCAST)
        self.rc_channel = kwargs.get('rc_channel')
        usecrc = kwargs.get('usecrc', False)
        if (usecrc):
            self.framer = rc_frame.PyRinCmdFrameCrc()
        else:
            self.framer = rc_frame.PyRinCmdFrameBasic()
        self.indata = ""

    def __del__(self):
        self.disconnect()
        #object.__del__(self)

    def construct_reply(self, rcc, is_error, data):
        """Construct a rincomms reply message to match rcc"""
        addr = self.rc_addr & rc_message.PyRinCmdMessage.ADDR_MASK
        addr |= rc_message.PyRinCmdMessage.ADDR_FLAG_RESPONSE
        if (is_error):
            addr |= rc_message.PyRinCmdMessage.ADDR_FLAG_ERROR
        rcr = rc_message.PyRinCmdMessage(address=addr, command=rcc.command, register=rcc.register, channel=rcc.channel, data=data)
        return rcr

    def find_message(self, reply):
        """Look for a validly framed message in the reply"""
        before, message, after = self.framer.deframe_response(response=reply)
        if (before is None and message is None):
            return False, "", after
        else:
            #print("Got reply {}".format(message))
            return True, message, after

    def senddata(self, data):
        """Send data to the host"""
        framed_data = self.framer.frame_command(cmd=data)
        self.comms.write(data=framed_data)

    def recvdata(self):
        """Receive data from the host"""
        return self.comms.read()

    def exec_message(self, msg, datatype):
        """Send a RincommsMessage to the indicator, and get the reply back"""
        if (not self.comms.is_connected()):
            raise Exception("PyRinCmdClient is not connected")
        msgstr = msg.to_string()
        self.senddata(data=msgstr)

        replystr = ""
        foundmsg = False
        while (not foundmsg):
            self.indata += self.recvdata()
            foundmsg, replystr, self.indata = self.find_message(reply=self.indata)
        self.comms.log_comms_message(in_not_out=True, data=replystr)

        parser = rc_message.PyRinCmdParser()
        reply = parser.str_to_message(message_str=replystr, datatype=datatype)
        parser.check_reply(cmd=msg, reply=reply)
        return reply

    def ask_register_type(self, register):
        """Ask the server for the register type"""
        regtype = self.read_register(command=rc_message.PyRinCmdMessage.CMD_READ_TYPE, register=register, datatype=int)
        return rc_register.PyRinCmdRegister.get_python_type_for_regtype(regtype=regtype.data)

    def read_register(self, command, register, datatype=str):
        """Read a register. Returns a PyRinCmdMessage object

           datatype can be:
             int
             str
             float
             function(x): eg lambda x: int(x, 16)
             'guess': Try to guess the data type
             'fromcmd': Try to guess the data type from the command
             'ask': Ask the indicator what the data type is. Requires extra traffic
        """
        if (datatype == 'ask'):
            datatype = self.ask_register_type(register=register)
        addr = self.rc_addr | rc_message.PyRinCmdMessage.ADDR_FLAG_REPLY
        # TODO check command
        msg = rc_message.PyRinCmdMessage(address=addr, command=command, register=register, channel=self.rc_channel, data=None)
        reply = self.exec_message(msg=msg, datatype=datatype)
        return reply

    def write_register(self, command, register, data):
        """Write a register. Returns a PyRinCmdMessage object"""
        addr = self.rc_addr | rc_message.PyRinCmdMessage.ADDR_FLAG_REPLY
        # TODO check command
        msg = rc_message.PyRinCmdMessage(address=addr, command=command, register=register, channel=self.rc_channel, data=data)
        reply = self.exec_message(msg=msg, datatype=int)
        return reply

    def execute_register(self, register, data, replydatatype=str):
        """Execute a register. Returns a PyRinCmdMessage object

           datatype can be:
             int
             str
             float
             function(x): eg lambda x: int(x, 16)
             'guess': Try to guess the data type
        """
        addr = self.rc_addr | rc_message.PyRinCmdMessage.ADDR_FLAG_REPLY
        msg = rc_message.PyRinCmdMessage(address=addr, command=rc_message.PyRinCmdMessage.CMD_EXECUTE,
                                         register=register, channel=self.rc_channel, data=data)
        reply = self.exec_message(msg=msg, datatype=replydatatype)
        return reply

    def connect(self):
        self.comms.connect()

    def disconnect(self):
        self.comms.disconnect()

    reply_data_types = \
    {
        "int":     { "pytype": int,                  "description": "Parse the result as an decimal integer" },
        "hexint":  { "pytype": lambda x: int(x, 16), "description": "Parse the result as an hexadecimal integer" },
        "str":     { "pytype": str,                  "description": "Parse the result as a string" },
        "float":   { "pytype": float,                "description": "Parse the result as a float" },
        "guess":   { "pytype": "guess",              "description": "Guess the type from the result data" },
        "fromcmd": { "pytype": "fromcmd",            "description": "Guess the type from the command" },
        "ask":     { "pytype": "ask",                "description": "Ask the indicator what the data type is. Requires extra traffic." },
    }

    @classmethod
    def get_reply_data_type(cls, typestr):
        """From a string containing type (eg "int"), get a type suitable for passing to read_register()"""
        return cls.reply_data_types[typestr]["pytype"]

    @classmethod
    def get_supported_reply_data_types(cls):
        """Get a list of all supported data types"""
        return cls.reply_data_types.keys()

    @classmethod
    def get_supported_reply_data_type_descriptions(cls):
        """Get a list of tuples of (typestr, description) of supported data types"""
        return [ (key, val["description"]) for key, val in cls.reply_data_types.items() ]

def parse_actions(actions):
    """Parse the action list into a list of (cmd, reg, data)"""

    i = 0
    pa = []
    while (i < len(actions)):
        cmd = actions[i]
        i += 1
        cmdint = rc_message.PyRinCmdMessage.get_command_value(cmd)
        if (rc_message.PyRinCmdMessage.is_write_command(cmdint) or
            rc_message.PyRinCmdMessage.is_execute_command(cmdint)):
            # Write and execute command need two parameters
            if (i + 1 >= len(actions)):
                raise Exception("Need two parameters for write and execute commands (register & value)")
            register = int(actions[i], 0)
            data = actions[i + 1]
            i += 2
            parsed_action = (cmdint, register, data)
            pa.append(parsed_action)
        elif (rc_message.PyRinCmdMessage.is_read_command(cmdint)):
            # Read command, need one parameters
            if (i >= len(actions)):
                raise Exception("Need one parameter for read commands (register)")
            register = int(actions[i], 0)
            i += 1
            parsed_action = (cmdint, register, None)
            pa.append(parsed_action)
        else:
            raise Exception("Unknown command")
    return pa

def perform_actions(client, actions, outfmt, datatype):
    """Perform the actions from the command line"""
    for action in actions:
        cmd = action[0]
        register = action[1]
        data = action[2]

        if (rc_message.PyRinCmdMessage.is_write_command(command=cmd)):
            reg = client.write_register(command=cmd, register=register, data=data)
        elif (rc_message.PyRinCmdMessage.is_read_command(command=cmd)):
            reg = client.read_register(command=cmd, register=register, datatype=datatype)
        elif (rc_message.PyRinCmdMessage.is_execute_command(command=cmd)):
            reg = client.execute_register(register=register, data=data, replydatatype=datatype)
        else:
            raise Exception("Unknown command")

        if (outfmt == "simple"):
            print("{}".format(reg.data))
        elif (outfmt == "rincmd"):
            print("{}".format(reg))
        elif (outfmt == "python"):
            print("{}".format(repr(reg)))
        elif (outfmt == "pretty"):
            print("{}".format(reg.to_pretty_string()))

def list_known_commands():
    """List the commands"""
    known_actions = rc_message.PyRinCmdMessage.get_supported_commands()
    print("Supported commands:")
    for act in known_actions:
        print("{}: {}".format(act, rc_message.PyRinCmdMessage.get_command_value(act)))

def list_known_datatypes():
    """List the known datatypes"""
    known_datatypes = PyRinCmdClient.get_supported_reply_data_type_descriptions()
    print("Supported datatypes:")
    for dt in known_datatypes:
        print("{}: {}".format(dt[0], dt[1]))

def main(argv=None):
    """Simple command line pyrincmd client"""

    # Parse command line options
    parser = argparse.ArgumentParser(description='Run the pyrincmd client to read/write register(s)')
    parser.add_argument('host', action='store',
                        help='The host to connect to, or server socket file if --protocol=uds, or serial port if --protocol=serial')
    parser.add_argument('--port', action='store', default=2222,
                        help='The TCP/UDP port to connect to the host on. Default 2222')
    parser.add_argument('--protocol', action='store', choices=["tcp", "udp", "uds", "serial"], default="tcp",
                        help='The protocol to communicate using (tcp, udp, uds, serial), default tcp')
    parser.add_argument('--serial-settings', action='store', default="9600n81",
                        help='The serial settings to use when --protocol=serial is used, default 9600n81')
    parser.add_argument('--udp-local-port', action='store', type=int, default=0,
                        help='Set the local udp port to bind to. Default 0 (kernel choose unused).')
    parser.add_argument('--usecrc', action='store_true', default=False,
                        help='Require CRC checksum on messages')
    parser.add_argument('--output', action='store', choices=["simple", "rincmd", "python", "pretty"], default="simple",
                        help='How to format the replies (simple is data value only, rincmd is as a rincmd reply, python is a PyRinCmdMessage() constructor, pretty is fully described).')
    parser.add_argument('--list-commands', action='store_true',
                        help='List the commands that the protocol supports')
    parser.add_argument('--data-type', action='store', choices=PyRinCmdClient.get_supported_reply_data_types(), default="guess",
                        help='Parse the returned data as this type. Only applies to read requests.')
    parser.add_argument('--list-data-types', action='store_true',
                        help='List the datatypes that --data-type supports')
    parser.add_argument('--comms-logfile', type=argparse.FileType('a'),
                        help='Log the communications to this file')
    parser.add_argument('actions', metavar='action', nargs='+',
                        help='The actions to be performed (CMD_READ_DECIMAL, etc). See --list-commands. Eg CMD_READ_DECIMAL 0x0005 ')
    args = parser.parse_args(args=argv)

    if (args.list_commands):
        list_known_commands()
        return
    if (args.list_data_types):
        list_known_datatypes()
        return

    actions = parse_actions(actions=args.actions)

    # Connect to host
    if (args.protocol == 'tcp'):
        client = PyRinCmdClient(tcphost=args.host, tcpport=int(args.port), usecrc=args.usecrc, logfile=args.comms_logfile)
    elif (args.protocol == 'udp'):
        client = PyRinCmdClient(udphost=args.host, udpport=int(args.port), udplocalport=args.udp_local_port, usecrc=args.usecrc, logfile=args.comms_logfile)
    elif (args.protocol == 'uds'):
        client = PyRinCmdClient(udsfile=args.host, usecrc=args.usecrc, logfile=args.comms_logfile)
    elif (args.protocol == 'serial'):
        client = PyRinCmdClient(serialport=args.host, settings=args.serial_settings, usecrc=args.usecrc, logfile=args.comms_logfile)

    client.connect()

    datatype = PyRinCmdClient.get_reply_data_type(args.data_type)
    perform_actions(client=client, actions=actions, outfmt=args.output, datatype=datatype)

    client.disconnect()


if (__name__ == '__main__'):
    main()
