'''
Communication interfaces for pyrincmd (serial, TCP, UDP)

Created on Oct 11, 2017

@author: ian
'''
import socket
import serial
import re
import time
import string


class PyRinCmdCommsBase(object):
    '''Base class for pyrincmd communications'''

    DEFAULT_READ_SIZE = 2048
    DEFAULT_TIMEOUT = 5.0

    def __init__(self, timeout=DEFAULT_TIMEOUT, logfile=None):
        self.timeout = timeout
        if isinstance(logfile, basestring):
            self.logfile = open(logfile, "a")
        elif (hasattr(logfile, 'write') or logfile is None):
            self.logfile = logfile
        else:
            raise Exception("logfile isn't filelike, or a filename")

    def __del__(self):
        self.disconnect()

    def log_escape(self, data):
        """Escape nonprintable data for the logfile, same as the C client format"""

        def escape_char(ch):
            if (not ch in string.printable):
                s = "\\x{:02X}".format(ord(ch))
            elif (ch == "\\"):
                s = "\\\\"
            elif (ch == '"'):
                s = '\\"'
            else:
                s = ch
            return s

        escaped = map(escape_char, data)
        return "".join(escaped)

    def log_message(self, message):
        """Write a timestamped message to the logfile"""
        tsmsg = "{:.9f} {}\n".format(time.time(), message)
        if self.logfile is not None:
            self.logfile.write(tsmsg)

    def log_comms_message(self, in_not_out, data):
        """Log a comms message into the logfile with timestamp and direction"""
        # 1526016996.286492537 >>> "20110306:;"
        # 1526016996.287361606 <<< "81110306:10221\x0D\x0A"
        dirnstr = "<<<" if in_not_out else ">>>"
        escapeddata = self.log_escape(data=data)
        msg = "{} \"{}\"".format(dirnstr, escapeddata)
        self.log_message(message=msg)

    def connect(self):
        self.log_message("connected")

    def disconnect(self):
        self.log_message("disconnected")

    def write(self, data):
        self.log_comms_message(in_not_out=False, data=data)

    def read(self, bufsize=DEFAULT_READ_SIZE):
        pass

    def is_connected(self):
        return False


class PyRinCmdCommsTcp(PyRinCmdCommsBase):
    '''TCP communications class'''

    def __init__(self, host, port, timeout=PyRinCmdCommsBase.DEFAULT_TIMEOUT, logfile=None):
        super(PyRinCmdCommsTcp, self).__init__(timeout=timeout, logfile=logfile)
        self.host = host
        self.port = port
        self.sock = None

    def connect(self):
        """Connect to the host"""
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.host, self.port))
        self.sock.settimeout(self.timeout)
        super(PyRinCmdCommsTcp, self).connect()

    def disconnect(self):
        if (self.sock is not None):
            self.sock.close()
            self.sock = None
            super(PyRinCmdCommsTcp, self).disconnect()

    def write(self, data):
        super(PyRinCmdCommsTcp, self).write(data=data)
        self.sock.sendall(data)

    def read(self, bufsize=PyRinCmdCommsBase.DEFAULT_READ_SIZE):
        return self.sock.recv(bufsize)

    def is_connected(self):
        return self.sock is not None


class PyRinCmdCommsUdp(PyRinCmdCommsBase):
    '''UDP communications class'''

    def __init__(self, host, port, localip="0.0.0.0", localudpport=0, timeout=PyRinCmdCommsBase.DEFAULT_TIMEOUT, logfile=None):
        super(PyRinCmdCommsUdp, self).__init__(timeout=timeout, logfile=logfile)
        self.host = host
        self.port = port
        self.localip = localip
        self.localudpport = localudpport
        self.sock = None

    def connect(self):
        """Connect to the host"""

        # Bind to the local interface
        s = None
        for liface in socket.getaddrinfo(self.localip, self.localudpport, socket.AF_UNSPEC,
                                         socket.SOCK_DGRAM, 0, socket.AI_PASSIVE):
            af, socktype, proto, canonname, sa = liface
            try:
                s = socket.socket(af, socktype, proto)
            except socket.error:
                s = None
                continue
            try:
                s.bind(sa)

                # Connect to the remote host
                for riface in socket.getaddrinfo(self.host, self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM):
                    af, socktype, proto, canonname, sa = riface
                    s.connect(sa)
                    break

            except socket.error:
                s.close()
                s = None
                continue
            break
        if s is None:
            raise Exception("Failed to open UDP connection to {}:{}".format(self.host, self.port))
        self.sock = s

    def disconnect(self):
        if (self.sock is not None):
            self.sock.close()
            self.sock = None
            super(PyRinCmdCommsUdp, self).disconnect()

    def write(self, data):
        super(PyRinCmdCommsUdp, self).write(data=data)
        self.sock.send(data)

    def read(self, bufsize=PyRinCmdCommsBase.DEFAULT_READ_SIZE):
        data = self.sock.recv(bufsize)
        return data

    def is_connected(self):
        return self.sock is not None


class PyRinCmdCommsUds(PyRinCmdCommsBase):
    '''Unix domain socket communications class'''

    def __init__(self, udsfile, timeout=PyRinCmdCommsBase.DEFAULT_TIMEOUT, logfile=None):
        super(PyRinCmdCommsUds, self).__init__(timeout=timeout, logfile=logfile)
        self.udsfile = udsfile
        self.sock = None

    def connect(self):
        """Connect to the host"""
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(self.udsfile)
        self.sock.settimeout(self.timeout)
        super(PyRinCmdCommsUds, self).connect()

    def disconnect(self):
        if (self.sock is not None):
            self.sock.close()
            self.sock = None
            super(PyRinCmdCommsUds, self).disconnect()

    def write(self, data):
        super(PyRinCmdCommsUds, self).write(data=data)
        self.sock.sendall(data)

    def read(self, bufsize=PyRinCmdCommsBase.DEFAULT_READ_SIZE):
        return self.sock.recv(bufsize)

    def is_connected(self):
        return self.sock is not None


class PyRinCmdCommsSerial(PyRinCmdCommsBase):
    '''Rinstrum pyrincmd serial port communications class'''

    def __init__(self, port, **kwargs):
        """
        Parameters are:
          @param port is the serial port to use

          @param settings is a string like "9600n81" OR

          @param baudrate is the baudrate
          @param bytesize is the number of databits
          @param parity is the parity setting "N", "O", "E"
          @param stopbits is the number of stopbits

          @param timeout is the timeout in seconds (float)
          @param logfile is a filelike object or filename to write a comms log to

        Examples:
          PyRinCmdCommsSerial(port='/dev/ttyS0', settings="9600n81")
          PyRinCmdCommsSerial(port='/dev/ttyS0', baudrate=9600, bytesize=8, parity="N", stopbits=1)
        """
        settings = kwargs.get('settings')
        if settings is not None:
            baudrate, parity, bytesize, stopbits = self.parse_serial_settings(settings)
        else:
            baudrate = kwargs.get('baudrate', 9600)
            bytesize = kwargs.get('bytesize', serial.EIGHTBITS)
            parity = kwargs.get('parity', serial.PARITY_NONE)
            stopbits = kwargs.get('stopbits', serial.STOPBITS_ONE)
        timeout = kwargs.get("timeout", PyRinCmdCommsBase.DEFAULT_TIMEOUT)
        logfile = kwargs.get("logfile", None)

        super(PyRinCmdCommsSerial, self).__init__(timeout=timeout, logfile=logfile)
        self.port = port
        self.serial = serial.Serial(port=None, baudrate=baudrate,
                                    bytesize=bytesize, parity=parity,
                                    stopbits=stopbits,
                                    timeout=self.timeout,
                                    xonxoff=False,
                                    rtscts=False,
                                    write_timeout=None,
                                    dsrdtr=False,
                                    inter_byte_timeout=None)

    def parse_serial_settings(self, settings):
        pattern = "([0-9]+)([noeNOE])([5678])([12])"
        m = re.match(pattern, settings)
        if m is None:
            raise Exception("Could not parse serial settings from '{}'".format(settings))
        sets = list(m.groups())
        sets[0] = int(sets[0])
        sets[1] = sets[1].upper()
        sets[2] = int(sets[2])
        sets[3] = int(sets[3])
        return sets

    def connect(self):
        """Connect to the serial port"""
        self.serial.port = self.port
        self.serial.open()
        super(PyRinCmdCommsSerial, self).connect()

    def disconnect(self):
        self.serial.close()
        super(PyRinCmdCommsSerial, self).disconnect()

    def write(self, data):
        self.serial.write(data)
        super(PyRinCmdCommsSerial, self).write(data=data)

    def read(self, bufsize=PyRinCmdCommsBase.DEFAULT_READ_SIZE):
        toread = min(bufsize, max(self.serial.in_waiting, 1))
        data = self.serial.read(toread)
        if (len(data) == 0):
            raise socket.timeout("timed out")
        return data

    def is_connected(self):
        return self.serial.is_open
