'''
Created on Sep 11, 2017

@author: ian
'''

class PyRinCmdError(object):
    """Class for rincmd error"""

    # Major errors
    RC_HNDL_ERR_NONE = 0
    RC_HNDL_ERR_PARSE = 1
    RC_HNDL_ERR_CRC = 2
    RC_HNDL_ERR_REG = 3
    RC_HNDL_ERR_READ = 4
    RC_HNDL_ERR_WRITE = 5
    RC_HNDL_ERR_EXEC = 6
    RC_HNDL_ERR_NOTIMP = 7

    # Success (RC_HNDL_ERR_NONE) minor errors
    RC_NONE_ERR_NONE =        0

    # Parser minor errors
    RC_PARSE_ERR_NONE =       0
    RC_PARSE_ERR_ADDR =       1
    RC_PARSE_ERR_CMD =        2
    RC_PARSE_ERR_REG =        3
    RC_PARSE_ERR_CHAN =       4
    RC_PARSE_ERR_DELIM =      5
    RC_PARSE_ERR_DATA =       6
    RC_PARSE_ERR_FRAMING =    8

    # CRC minor errors
    RC_CRC_ERR_NONE =         0
    RC_CRC_ERR_PARSE =        1
    RC_CRC_ERR_BAD =          2

    # Register minor errors
    RC_REG_ERR_NONE =         0

    # Read minor errors
    RC_READ_ERR_NONE =        0
    RC_READ_ERR_PERMISSION =  1
    RC_READ_ERR_UNKNOWNTYPE = 2
    RC_READ_ERR_NOTYPDAT =    3
    RC_READ_ERR_COMMAND =     4
    RC_READ_ERR_BADDATA =     5
    RC_READ_ERR_REGNULL =     6
    RC_READ_ERR_BUFSIZE =     7
    RC_READ_ERR_SNPRINTF =    8

    # Write minor errors
    RC_WRITE_ERR_NONE =       0
    RC_WRITE_ERR_PERMISSION = 1
    RC_WRITE_ERR_UNKNOWNTYPE = 2
    RC_WRITE_ERR_NOTYPDAT =   3
    RC_WRITE_ERR_COMMAND =    4
    RC_WRITE_ERR_PARSE =      5
    RC_WRITE_ERR_RANGE_LOW =  6
    RC_WRITE_ERR_RANGE_HIGH = 7
    RC_WRITE_ERR_LENGTH_LOW = 8
    RC_WRITE_ERR_LENGTH_HIGH = 9
    RC_WRITE_ERR_REGNULL =    10
    RC_WRITE_ERR_BUFSIZE =   11
    RC_WRITE_ERR_SNPRINTF =  12
    RC_WRITE_ERR_STRTOLL =   13
    RC_WRITE_ERR_TYPE_LEN =  14

    # Execute minor errors
    RC_EXEC_ERR_NONE =        0
    RC_EXEC_ERR_PERMISSION =  1
    RC_EXEC_ERR_UNKNOWNTYPE = 2
    RC_EXEC_ERR_TYPDAT =      3
    RC_EXEC_ERR_COMMAND =     4
    RC_EXEC_ERR_REGNULL =     5
    RC_EXEC_ERR_WRITEBACK =   6

    RC_CNSTR_ERR_NONE =       0
    RC_CNSTR_ERR_BUFSIZE =    1
    RC_CNSTR_ERR_SNPRINTF =   2

    def __init__(self,  *args, **kwargs):
        """Construct with major=N, minor=N, OR error=N where error will be broken up into major and minor"""
        if ("major" in kwargs.keys() and "minor" in kwargs.keys()):
            self.major = kwargs.get("major")
            self.minor = kwargs.get("minor")
        elif "error" in kwargs.keys():
            self.major, self.minor = PyRinCmdError.major_minor_from_error(kwargs.get("error"))
        else:
            raise Exception("Invalid kwargs to PyRinCmdError()")

    def __str__(self):
        return self.get_error_str(split=True)

    def __repr__(self, *args, **kwargs):
        return 'PyRinCmdError(major={}, minor={})'.format(self.major, self.minor)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.major == other.major and
                    self.minor == other.minor)
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def major_minor_from_error(error):
        major = (error & 0xFF00) >> 8
        minor = error & 0x00FF
        return (major, minor)

    @staticmethod
    def make_error(major, minor):
        return major << 8 | minor

    def get_error(self):
        return PyRinCmdError.make_error(major=self.major, minor=self.minor)

    def get_major_string(self):
        """Get the major error name (as a string)"""
        return PyRinCmdError.get_major_error_name(major=self.major)

    def get_minor_string(self):
        """Get the minor error name (as a string)"""
        return PyRinCmdError.get_minor_error_name(major=self.major, minor=self.minor)

    def get_error_str(self, split=False):
        """Return the error code as a string.
           Split indicates if it should split the major:minor with a colon"""
        return "0x{:02X}:0x{:02X}".format(self.major, self.minor) if split else "0x{:04x}".format(self.get_error())

    def get_pretty_string(self):
        """Convert the error string into a pretty string"""
        return "{}:{} (0x{:02X}:0x{:02X})".format(self.get_major_string(), self.get_minor_string(), self.major, self.minor)

    @classmethod
    def get_major_errors(cls):
        """Get a dict of supported major errors"""
        return dict((key, value) for key, value in cls.__dict__.iteritems() if key.startswith("RC_HNDL_ERR_"))

    @classmethod
    def get_major_error_name(cls, major):
        """Get the name of a major error given a value (ie 1 -> 'RC_HNDL_ERR_PARSE')"""
        errs = cls.get_major_errors()
        errdict = dict((key, value) for key, value in errs.iteritems() if value == major)
        return errdict.keys()[0]

    @classmethod
    def get_minor_errors(cls, major):
        """Get a dict of supported minor errors for given major error"""
        name = cls.get_major_error_name(major=major)
        _, minortype = name.split("RC_HNDL_ERR_")
        minor_prefix = "RC_{}_ERR_".format(minortype)
        return dict((key, value) for key, value in cls.__dict__.iteritems() if key.startswith(minor_prefix))

    @classmethod
    def get_minor_error_name(cls, major, minor):
        """Get the name of a minor error given a major and minor value (ie 1 -> 'RC_HNDL_ERR_PARSE')"""
        errs = cls.get_minor_errors(major=major)
        errdict = dict((key, value) for key, value in errs.iteritems() if value == minor)
        return errdict.keys()[0]

class PyRinCmdException(Exception):
    """Exception generated by pyrincmd"""

    def __init__(self, pyrincmd_error=None, major=None, minor=None):
        if (pyrincmd_error is None):
            pyrincmd_error = PyRinCmdError(major=major, minor=minor)
        self.pyrincmd_error = pyrincmd_error

        errmsg = pyrincmd_error.get_error_str(split=True)
        super(PyRinCmdException, self).__init__(errmsg)

    def get_error_tuple(self):
        """Return the PyRinCmdError as a (major, minor) tuple"""
        return self.pyrincmd_error.major, self.pyrincmd_error.minor
