'''
Utility for locating all rinstrum devices on the network

Created on Mar 13, 2012

@author: IanB
'''
import sys
import socket
import xml.etree.ElementTree
import argparse
import ipaddress

def str_to_unicode_py3(s):
    """Under python 2, str and unicode are different, under python 3 str is unicode"""
    if sys.version_info < (3, 0, 0):
        if isinstance(s, str):
            s = s.decode()
    return s

class PyRinCmdDeviceFinder(object):
    """Class for finding rinstrum devices on the network"""

    REQUEST_METHOD_BROADCAST='broadcast'
    REQUEST_METHOD_ITERATIVE='iterative'

    def __init__(self, send_udp_port=2244, num_requests=3, request_timeout=1.0, method=REQUEST_METHOD_BROADCAST):
        """Constructor

        @param send_udp_port is the UDP port to send the device finder message on. Default 2244. Do not change, unless you know what you are doing
        @param num_requests is the number of times to repeat the request
        @param request_timeout is the time in seconds to wait for responses for each request
        @param method is 'broadcast' to use UDP broadcasts, or 'iterative' to iterate though each of the hosts on the network (may get slow for anything larger than /24)
        """
        self.send_udp_port = send_udp_port
        self.num_requests = num_requests
        self.request_timeout = request_timeout
        if (method not in self.get_request_methods()):
            raise ValueError("{} is an invalid request method".format(method))
        self.request_method = method

    tag_and_name = \
    [
        ("ipaddress", "IP Address"),
        ("class", "Class"),
        ("name", "Name"),
        ("serialnumber", "Serial Number"),
        ("version", "Version"),
        ("customid", "Custom ID"),
        ("macaddress", "MAC Address"),
        ("tcpport", "TCP Port"),
        ("hostname", "Hostname")
    ]

    request_message = 'BlahBlah'.encode('utf-8')

    @classmethod
    def get_sortby_list(cls):
        return [ x[0] for x in cls.tag_and_name ]

    @classmethod
    def get_request_methods(cls):
        return [ v for k,v in cls.__dict__.items() if k.startswith("REQUEST_METHOD_") ]

    def sort_list(self, indicator_list, sortby):
        """Sort the indicator list by sortby"""
        sorted_indicator_list = sorted(indicator_list, key=lambda indicator: indicator[sortby])
        return sorted_indicator_list

    def parse_reply_to_dict(self, addr, reply):
        """Parse the reply from addr into a dict"""

        elem = xml.etree.ElementTree.fromstring(reply)
        iterator = elem.findall("device//*")

        ipelem = xml.etree.ElementTree.Element("ipaddress")
        ipelem.text = addr
        iterator.append(ipelem)

        item = { }
        for ttp in self.tag_and_name:
            t = ttp[0]
            lst = [ i for i in iterator if i.tag == t ]
            if (lst is not None and len(lst) > 0):
                val = lst[0]
                if (val is not None and val.text is not None):
                    item[t] = val.text
                else:
                    item[t] = "None"
            else:
                    item[t] = "None"
        return item

    def find_devices_broadcast(self, address, sort_by=None):
        """Look for indicators on the network at address using a UDP broadcast

           address can be hostname (str), ip (str), ipaddress.IPv4Address, ipaddress.IPv4Network
        """

        if isinstance(address, ipaddress.IPv4Address):
            sendto_addr = str(address)
        elif isinstance(address, ipaddress.IPv4Network):
            sendto_addr = str(address.broadcast_address)
        else:
            sendto_addr = address

        # Setup listening UDP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.bind(("", 0))

        # Wait for replies
        indicator_list = []
        sock.settimeout(self.request_timeout)
        for _ in range(0, self.num_requests):

            # Setup UDP send socket
            sock.sendto(self.request_message, (sendto_addr, self.send_udp_port))

            # Wait for and process replies
            while True:
                # Get response
                try:
                    data, addr = sock.recvfrom(2048)
                except socket.timeout:
                    break

                # Parse response
                try:
                    new_indicator = self.parse_reply_to_dict(addr=addr[0], reply=data)

                    # Only add if we haven't seen it before
                    matches = [ x for x in indicator_list if x["ipaddress"] == new_indicator["ipaddress"]]
                    if (len(matches) == 0):
                        indicator_list.append(new_indicator)
                except:
                    e = sys.exc_info()[0]
                    print("Error parsing response from {}: {}".format(addr[0], e))

        if (sort_by is not None):
            indicator_list = self.sort_list(indicator_list=indicator_list, sortby=sort_by)
        return indicator_list

    def find_devices_iterative(self, address, sort_by=None):
        """Look for indicators on the network by iterating the hosts on sendto_network (an ipaddress.IPv4Network)"""

        if isinstance(address, ipaddress.IPv4Address):
            no_reply_net_hosts = [ address ]
        elif isinstance(address, ipaddress.IPv4Network):
            no_reply_net_hosts = [ x for x in address.hosts() ]
        else:
            no_reply_net_hosts = [ address ]

        # Setup listening UDP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind(("", 0))

        # Blast the request to every host in the network
        indicator_list = []

        for _ in range(0, self.num_requests):

            sock.setblocking(1)
            sock.settimeout(None)
            for host in no_reply_net_hosts:
                sock.sendto(self.request_message, (str(host), self.send_udp_port))

            # Wait for and process replies
            sock.settimeout(self.request_timeout)
            while True:
                # Get response
                try:
                    data, addr = sock.recvfrom(2048)
                except socket.timeout:
                    break

                # Parse response
                try:
                    new_indicator = self.parse_reply_to_dict(addr=addr[0], reply=data)
                    new_ind_ip = ipaddress.IPv4Address(str_to_unicode_py3(new_indicator["ipaddress"]))
                    try:
                        no_reply_net_hosts.remove(new_ind_ip)
                    except ValueError:
                        pass

                    # Only add if we haven't seen it before
                    matches = [ x for x in indicator_list if x["ipaddress"] == new_indicator["ipaddress"]]
                    if (len(matches) == 0):
                        indicator_list.append(new_indicator)
                except:
                    e = sys.exc_info()[0]
                    print("Error parsing response from {}: {}".format(addr[0], e))

        if (sort_by is not None):
            indicator_list = self.sort_list(indicator_list=indicator_list, sortby=sort_by)
        return indicator_list

    def find_devices(self, address, sort_by=None):
        """Look for indicators on the network"""
        if self.request_method == self.REQUEST_METHOD_BROADCAST:
            indicator_list = self.find_devices_broadcast(address=address, sort_by=sort_by)
        else:
            indicator_list = self.find_devices_iterative(address=address, sort_by=sort_by)
        return indicator_list

    def pretty_print_list(self, indicator_list, tags_to_print):
        """Print the list with a nice header"""

        colwidths = []
        for t in tags_to_print:
            colwidths.append(len(t[1]) + 2)

        for indicator in indicator_list:
            for i, ttp in enumerate(tags_to_print):
                if (colwidths[i] < len(indicator[ttp[0]]) + 2):
                    colwidths[i] = len(indicator[ttp[0]]) + 2

        header = ""
        for i, t in enumerate(tags_to_print):
            header += "{:<{width}} ".format(t[1] + ',', width=colwidths[i])
        print(header.rstrip(", "))

        for indicator in indicator_list:
            line = ""
            for i, ttp in enumerate(tags_to_print):
                line += "{:<{width}} ".format(indicator[ttp[0]] + ',', width=colwidths[i])
            print(line.rstrip(", "))

def address_arg_conversion(address):
    """Convert the address argument into a hostname (str), ipaddress.IPv4Address or ipaddress.IPv4Network"""
    try:
        conv = ipaddress.IPv4Address(address=address)
    except:
        try:
            conv = ipaddress.IPv4Network(address=address)
        except:
            conv = address
    return conv

def main(argv=None):
    # Parse command line options
    parser = argparse.ArgumentParser(description='Look for Rinstrum indicators on the network using PyRinCmdDeviceFinder')
    parser.add_argument('address', action='store',
                        help='The address to transmit to. Can be a single ip (eg 192.168.0.1), hostname (eg c520-3454623.rinstrum.com), broadcast address (eg 192.168.0.255), or network (eg 192.168.0.1/24).')
    parser.add_argument('--sort-by', action='store', default=None, choices=PyRinCmdDeviceFinder.get_sortby_list(),
                        help='The field to sort the list by.')
    parser.add_argument('--method', action='store', choices=PyRinCmdDeviceFinder.get_request_methods(), default=PyRinCmdDeviceFinder.REQUEST_METHOD_BROADCAST,
                        help='The method by which to contact the hosts. "broadcast" (default) uses UDP broadcast datagrams, "iterative" sends a UDP datagram to each host on the network. Iterative may get slow for anything larger than /24')
    parser.add_argument('--repeats', action='store', type=int, default=3,
                        help='The number of times to send the request (default 3)')
    parser.add_argument('--timeout', action='store', type=float, default=1.0,
                        help='The number seconds to wait for replies (default 1.0)')
    args = parser.parse_args(args=argv)

    unicode_address = str_to_unicode_py3(args.address)
    address_converted = address_arg_conversion(address=unicode_address)

    # Do the device finding
    print("Searching for Rinstrum indicators on {}".format(address_converted))
    devfinder = PyRinCmdDeviceFinder(num_requests=args.repeats, request_timeout=args.timeout, method=args.method)
    indicators = devfinder.find_devices(address=address_converted, sort_by=args.sort_by)

    print("Search results ({} indicators found):".format(len(indicators)))
    devfinder.pretty_print_list(indicator_list=indicators, tags_to_print=devfinder.tag_and_name)

if (__name__ == '__main__'):
    main()
