'''
Framing of rincmd messages

Created on Oct 1, 2017

@author: ian
'''
import unittest
import rc_crc16
import rc_error


class PyRinCmdFrameBase(object):
    '''Base class for framing rincmd messages'''

    def __init__(self, cmd_prefix, cmd_postfix, resp_prefix, resp_postfix):
        '''Constructor for base class'''
        self.cmd_prefix = cmd_prefix
        self.cmd_postfix = cmd_postfix
        self.resp_prefix = resp_prefix
        self.resp_postfix = resp_postfix

    def frame(self, prefix, postfix, data):
        """Frame some data"""
        framed = "{}{}{}".format(prefix, data, postfix)
        return framed

    def frame_command(self, cmd):
        """Frame a command, ready for transmission (client end)"""
        return self.frame(prefix=self.cmd_prefix, postfix=self.cmd_postfix, data=cmd)

    def frame_response(self, resp):
        """Frame a response, ready for transmission (server end)"""
        return self.frame(prefix=self.resp_prefix, postfix=self.resp_postfix, data=resp)

    def deframe(self, prefix, postfix, reply):
        """Deframe a reply.
           Returns tuple (before_prefix, msg_data, after_postfix) if found,
           otherwise (None, None, after_postfix
           """
        spl = reply.split(postfix, 1)
        if (len(spl) != 2):
            return None, None, reply
        before = ""
        message = spl[0]
        after = spl[1]
        if (len(prefix) > 0):
            spl = message.split(prefix, 1)
            if (len(spl) < 2):
                raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_PARSE, minor=rc_error.PyRinCmdError.RC_PARSE_ERR_FRAMING)
            before = spl[0]
            message = spl[1]
        # TODO? if (leftover in [ "\r", "\n", "\r\n" ]):
        #    leftover = ""
        return before, message, after

    def deframe_command(self, command):
        return self.deframe(prefix=self.cmd_prefix, postfix=self.cmd_postfix, reply=command)

    def deframe_response(self, response):
        return self.deframe(prefix=self.resp_prefix, postfix=self.resp_postfix, reply=response)


class PyRinCmdFrameBasic(PyRinCmdFrameBase):
    '''Basic framing of a rincmd message'''

    def __init__(self):
        '''Framing with ; & \r\n'''
        super(PyRinCmdFrameBasic, self).__init__(cmd_prefix="", cmd_postfix=";", resp_prefix="", resp_postfix="\r\n")

    def __repr__(self):
        return "{}()".format(type(self).__name__)


class PyRinCmdFrameAutoOut(PyRinCmdFrameBase):
    '''Framing for an automatic output'''

    def __init__(self, resp_prefix="\x02", resp_postfix="\x03"):
        '''Responses are framed with STX, ETX generally, there are no commands'''
        super(PyRinCmdFrameAutoOut, self).__init__(cmd_prefix="", cmd_postfix="", resp_prefix=resp_prefix, resp_postfix=resp_postfix)

    def __repr__(self):
        return "{}(resp_prefix=\"{}\", resp_postfix=\"{}\")".format(type(self).__name__,
                                                                    self.resp_prefix.encode("string-escape"),
                                                                    self.resp_postfix.encode("string-escape"))


class PyRinCmdFrameCrc(PyRinCmdFrameBase):
    '''Framing of a rincmd message with CRC'''

    def __init__(self):
        '''Framing commands and responses with SOH(01) and EOT(04)'''
        super(PyRinCmdFrameCrc, self).__init__(cmd_prefix="\x01", cmd_postfix="\x04", resp_prefix="\x01", resp_postfix="\x04")

    def __repr__(self):
        return "{}()".format(type(self).__name__)

    def frame(self, prefix, postfix, data):
        """Frame some data with a CRC"""

        # Calculate and append CRC
        crc = rc_crc16.pyrincmd_crc16(data=data)
        crcd_data = "{}{:4X}".format(data, crc)

        # Frame normally
        return super(PyRinCmdFrameCrc, self).frame(prefix=prefix, postfix=postfix, data=crcd_data)

    def deframe(self, prefix, postfix, reply):
        """Deframe some data with a CRC"""

        # Deframe normally
        before, message, after = super(PyRinCmdFrameCrc, self).deframe(prefix=prefix, postfix=postfix, reply=reply)

        # Go looking for the CRC, and check it
        if (before is not None and message is not None):
            msgdata = message[:-4]
            try:
                msgcrc = int(message[-4:], 16)
            except ValueError:
                raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_CRC, minor=rc_error.PyRinCmdError.RC_CRC_ERR_PARSE)
            calccrc = rc_crc16.pyrincmd_crc16(data=msgdata)
            if (calccrc != msgcrc):
                raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_CRC, minor=rc_error.PyRinCmdError.RC_CRC_ERR_BAD)
            message = msgdata

        return before, message, after

class TestPyRinCmdFrameBasic(unittest.TestCase):
    """Test basic framing/deframing of rincmd data"""

    def test_frame_basic_cmd(self):
        """Test framing a basic command"""
        framer = PyRinCmdFrameBasic()
        framed = framer.frame_command(cmd="20110005:")
        self.assertEqual("20110005:;", framed)

    def test_deframe_basic_cmd(self):
        """Test deframing a basic command"""
        framer = PyRinCmdFrameBasic()

        # Single message
        before, message, after = framer.deframe_command(command="20110005:;")
        self.assertEqual("", before)
        self.assertEqual("20110005:", message)
        self.assertEqual("", after)

        # Data following message
        before, message, after = framer.deframe_command(command="20110005:;20120006:;")
        self.assertEqual("", before)
        self.assertEqual("20110005:", message)
        self.assertEqual("20120006:;", after)

        # No framing
        before, message, after = framer.deframe_command(command="20110005:")
        self.assertIsNone(before)
        self.assertIsNone(message)
        self.assertEqual("20110005:", after)

    def test_frame_basic_response(self):
        """Test framing a basic respnse"""
        framer = PyRinCmdFrameBasic()
        framed = framer.frame_response(resp="81110005:00327F78")
        self.assertEqual("81110005:00327F78\r\n", framed)

    def test_deframe_basic_response(self):
        """Test deframing a basic response"""
        framer = PyRinCmdFrameBasic()

        # Single message
        before, message, after = framer.deframe_response(response="81110005:00327F78\r\n")
        self.assertEqual("", before)
        self.assertEqual("81110005:00327F78", message)
        self.assertEqual("", after)

        # Data following message
        before, message, after = framer.deframe_response(response="81110005:00327F78\r\n81110006:hello there\r\n")
        self.assertEqual("", before)
        self.assertEqual("81110005:00327F78", message)
        self.assertEqual("81110006:hello there\r\n", after)

        # No framing
        before, message, after = framer.deframe_response(response="81110005:00327F78")
        self.assertIsNone(before)
        self.assertIsNone(message)
        self.assertEqual("81110005:00327F78", after)


class TestPyRinCmdFrameCrc(unittest.TestCase):
    """Test framing/deframing of rincmd data with a CRC"""

    def test_frame_crc_cmd(self):
        """Test framing a crc command"""
        framer = PyRinCmdFrameCrc()
        framed = framer.frame_command(cmd="20110005:")
        self.assertEqual("\x0120110005:6FD0\x04", framed)

    def test_deframe_crc_cmd(self):
        """Test deframing a crc command"""
        framer = PyRinCmdFrameCrc()

        # Single message
        before, message, after = framer.deframe_command(command="\x0120110005:6FD0\x04")
        self.assertEqual("", before)
        self.assertEqual("20110005:", message)
        self.assertEqual("", after)

        # Bad checksum
        with self.assertRaises(rc_error.PyRinCmdException):
            before, message, after = framer.deframe_command(command="\x0120110005:6FD1\x04")

        # No checksum
        with self.assertRaises(rc_error.PyRinCmdException):
            before, message, after = framer.deframe_command(command="\x0120110005:\x04")

        # Data following message
        before, message, after = framer.deframe_command(command="\x0120110005:6FD0\x04\x0120110004:5CE1\x04")
        self.assertEqual("", before)
        self.assertEqual("20110005:", message)
        self.assertEqual("\x0120110004:5CE1\x04", after)

        # Data before and after message
        before, message, after = framer.deframe_command(command="asdf\x0120110005:6FD0\x04\x0120110004:5CE1\x04")
        self.assertEqual("asdf", before)
        self.assertEqual("20110005:", message)
        self.assertEqual("\x0120110004:5CE1\x04", after)

        # No framing
        before, message, after = framer.deframe_command(command="20110005:")
        self.assertIsNone(before)
        self.assertIsNone(message)
        self.assertEqual("20110005:", after)

        # No postfix yet
        before, message, after = framer.deframe_command(command="\x0120110005:6FD0")
        self.assertIsNone(before)
        self.assertIsNone(message)
        self.assertEqual("\x0120110005:6FD0", after)

        # Missing prefix
        with self.assertRaises(rc_error.PyRinCmdException) as ctx:
            before, message, after = framer.deframe_command(command="20110005:\x04")
        major, minor = ctx.exception.get_error_tuple()
        self.assertEqual(major, rc_error.PyRinCmdError.RC_HNDL_ERR_PARSE)
        self.assertEqual(minor, rc_error.PyRinCmdError.RC_PARSE_ERR_FRAMING)

    def test_frame_crc_response(self):
        """Test framing a crc response"""

        framer = PyRinCmdFrameCrc()
        framed = framer.frame_response(resp="81110005:00327F78")
        self.assertEqual("\x0181110005:00327F785424\x04", framed)

    def test_deframe_crc_response(self):
        """Test deframing a crc response"""
        framer = PyRinCmdFrameCrc()

        # Single message
        before, message, after = framer.deframe_command(command="\x0181110005:00327F785424\x04")
        self.assertEqual("", before)
        self.assertEqual("81110005:00327F78", message)
        self.assertEqual("", after)

        # Bad checksum
        with self.assertRaises(rc_error.PyRinCmdException) as ctx:
            before, message, after = framer.deframe_command(command="\x0181110005:00327F785425\x04")
        major, minor = ctx.exception.get_error_tuple()
        self.assertEqual(major, rc_error.PyRinCmdError.RC_HNDL_ERR_CRC)
        self.assertEqual(minor, rc_error.PyRinCmdError.RC_CRC_ERR_BAD)

        # No checksum
        with self.assertRaises(rc_error.PyRinCmdException) as ctx:
            before, message, after = framer.deframe_command(command="\x01\x0181110005:K401\x04")
        major, minor = ctx.exception.get_error_tuple()
        self.assertEqual(major, rc_error.PyRinCmdError.RC_HNDL_ERR_CRC)
        self.assertEqual(minor, rc_error.PyRinCmdError.RC_CRC_ERR_PARSE)

        # Data following message
        before, message, after = framer.deframe_command(command="\x0181110005:00327F785424\x04\x0181110003:K401ADA1\x04")
        self.assertEqual("", before)
        self.assertEqual("81110005:00327F78", message)
        self.assertEqual("\x0181110003:K401ADA1\x04", after)

        # Data before and after message
        before, message, after = framer.deframe_command(command="asdf\x0181110005:00327F785424\x04\x0181110003:K401ADA1\x04")
        self.assertEqual("asdf", before)
        self.assertEqual("81110005:00327F78", message)
        self.assertEqual("\x0181110003:K401ADA1\x04", after)

        # No framing
        before, message, after = framer.deframe_command(command="81110005:00327F785424")
        self.assertIsNone(before)
        self.assertIsNone(message)
        self.assertEqual("81110005:00327F785424", after)

        # No postfix yet
        before, message, after = framer.deframe_command(command="\x0181110005:00327F785424")
        self.assertIsNone(before)
        self.assertIsNone(message)
        self.assertEqual("\x0181110005:00327F785424", after)

if __name__ == '__main__':
    unittest.main()
