'''
Created on Feb 8, 2018

@author: ian
'''
import argparse
import rc_message
import rc_client
import rc_frame
import rc_error


def generate_message(rc_addr, command, register, data, channel, framing, message_type, output_format):
    """Convert command arguments into a rincmd message and print"""

    if (message_type == 'command'):
        addr = rc_addr | rc_message.PyRinCmdMessage.ADDR_FLAG_REPLY
    else:
        addr = rc_addr | rc_message.PyRinCmdMessage.ADDR_FLAG_RESPONSE

    # Generate rincmd message, and convert to string
    msg = rc_message.PyRinCmdMessage(address=addr, command=command, register=register, channel=channel, data=data)
    msgstr = msg.to_string()

    # Apply any framing
    if (framing == "basic"):
        framer = rc_frame.PyRinCmdFrameBasic()
    elif (framing == "crc"):
        framer = rc_frame.PyRinCmdFrameCrc()
    else:
        framer = None

    if (framer is not None):
        if (message_type == 'command'):
            framed_msg = framer.frame_command(cmd=msgstr)
        else:
            framed_msg = framer.frame_response(cmd=msgstr)
    else:
        framed_msg = msgstr

    if (output_format == "raw"):
        outdata = framed_msg
    else:
        outdata = framed_msg.encode('string_escape')

    print(outdata)

def try_deframing(indata):
    """Try many ways of deframing"""

    framers = [ (rc_frame.PyRinCmdFrameBasic(), "basic"), (rc_frame.PyRinCmdFrameCrc(), "crc") ]

    deframed_msg = None
    how = None
    for framer in framers:
        try:
            before, deframed_msg, after = framer[0].deframe_response(response=indata)
            if (deframed_msg is not None):
                how = (framer[1], "response")
                break
            before, deframed_msg, after = framer[0].deframe_command(command=indata)
            if (deframed_msg is not None):
                how = (framer[1], "command")
                break
        except rc_error.PyRinCmdException as e:
            print("Deframing with {} failed with: {}".format(framer[1], e.pyrincmd_error.get_pretty_string()))

    # Perhaps it is just an un-framed message
    if (deframed_msg is None):
        deframed_msg = indata
        how = "No framing found"
    else:
        how = "Deframed as {} framed {}".format(how[0], how[1])
    return deframed_msg, how

def parse_rincmd(message, input_format, output_format, datatype):
    """Parse a rincmd message and print it consituent parts"""

    # Convert input data format
    if (input_format == "raw"):
        indata = message
    else:
        indata = message.decode('string_escape')

    # Deframe the message
    deframed_msg, how_deframed = try_deframing(indata=indata)

    # Parse data
    parser = rc_message.PyRinCmdParser()
    parsed_msg = parser.str_to_message(message_str=deframed_msg, datatype=datatype)

    print(how_deframed)
    if (output_format == "simple"):
        print("{}".format(parsed_msg.data))
    elif (output_format == "rincmd"):
        print("{}".format(parsed_msg))
    elif (output_format == "python"):
        print("{}".format(repr(parsed_msg)))
    elif (output_format == "pretty"):
        print("{}".format(parsed_msg.to_pretty_string()))


def main(argv=None):
    """Command line utility to generate and parse rincmd messages"""

    # Parse command line options
    parser = argparse.ArgumentParser(description='Command line utility to generate and parse rincmd messages')

    subparsers = parser.add_subparsers(dest='operation', help='sub-command help')

    parser_gen = subparsers.add_parser('generate', help='Generate will create a rincmd message from constiuent parts')
    parser_gen.add_argument('--address', action='store', type=lambda x : int(x, 0), default=rc_message.PyRinCmdMessage.ADDR_BROADCAST,
                            help='The rincmd source/destination address as a number (dec or 0xHEX). Default: {}'.format(rc_message.PyRinCmdMessage.ADDR_BROADCAST))
    parser_gen.add_argument('--list-commands', action='store_true',
                            help='List the rincmd commands supported by the generate command')
    parser_gen.add_argument('actions', metavar='action', nargs='*',
                            help='The actions to be performed (CMD_READ_DECIMAL, etc). See --list-commands. Eg CMD_READ_DECIMAL 0x0005')
    parser_gen.add_argument('--framing', action='store', choices=["none", "basic", "crc"], default='basic',
                            help='The type of framing to place around the rincmd message. Default: basic')
    parser_gen.add_argument('--message-type', action='store', choices=["command", "response"], default='command',
                            help='The type of message we are sending')
    parser_gen.add_argument('--channel', action='store', type=lambda x : int(x, 0), default=None,
                            help='The channel field in the rincmd header as a number (dec or 0xHEX). Default: None')
    parser_gen.add_argument('--output-format', action='store', choices=["raw", "c-escape"], default='c-escape',
                            help='How to escape the non-printable characters in the output. Default: c-escape')

    parser_par = subparsers.add_parser('parse', help='Parse will pull apart a rincmd message')
    parser_par.add_argument('message', action='store',
                            help='The rincmd message to be parsed')
    parser_par.add_argument('--input-format', action='store', choices=["raw", "c-escape"], default='c-escape',
                            help='How to de-escape the non-printable characters in the output. Default: c-escape')
    parser_par.add_argument('--output-format', action='store', choices=["simple", "rincmd", "python", "pretty"], default="pretty",
                            help='How to format the replies (simple is data value only, rincmd is as a rincmd reply, python is a PyRinCmdMessage() constructor, pretty is fully described).')

    args = parser.parse_args(args=argv)

    if (args.operation == 'generate'):
        if (args.list_commands):
            rc_client.list_known_commands()
            return

        actions = rc_client.parse_actions(actions=args.actions)
        for a in actions:
            cmd, reg, data = a
            generate_message(rc_addr=args.address, command=cmd, register=reg, data=data, channel=args.channel, framing=args.framing, message_type=args.message_type, output_format=args.output_format)
    else:
        parse_rincmd(message=args.message, input_format=args.input_format, output_format=args.output_format, datatype=str)

if (__name__ == '__main__'):
    main()
