'''
Created on Sep 11, 2017

@author: ian
'''
import re
import types
import rc_error
import rc_escape


class PyRinCmdMessage(object):
    """Encapsulation of a rincmd message"""

    ADDR_BROADCAST = 0x00
    ADDR_MASK = 0x1F
    ADDR_FLAG_REPLY = 0x20
    ADDR_FLAG_ERROR = 0x40
    ADDR_FLAG_RESPONSE = 0x80

    CMD_READ_TYPE = 0x01
    CMD_READ_RANGE_MIN = 0x02
    CMD_READ_RANGE_MAX = 0x03
    # RC_MESG_READ_RAW        0x04    // Not implemented.
    CMD_READ_LITERAL = 0x05
    # RC_MESG_WRITE_RAW       0x06    // Not implemented.
    # RC_MESG_READ_DEFAULT    0x07    // Not implemented.
    # RC_MESG_READ_MENU_TEXT  0x09    // Not implemented.
    # RC_MESG_READ_FULL_TEXT  0x0A    // Not implemented.
    # RC_MESG_READ_ITEM       0x0D    // Not implemented.
    CMD_READ_PERMISSION = 0x0F
    CMD_EXECUTE = 0x10
    CMD_READ_HEXADEC = 0x11
    CMD_WRITE_HEXADEC = 0x12
    # RC_MESG_READ_WEIGHT     0x13    // Not implemented.
    # RC_MESG_READ_MENU_ITEMS 0x14    // Not implemented.
    # RC_MESG_READ_MENU_DATA  0x05    // Not implemented.
    CMD_READ_DECIMAL = 0x16
    CMD_WRITE_DECIMAL = 0x17

    def __init__(self, address, command, register, channel, data):
        self.address = address
        self.command = command
        self.register = register
        self.channel = channel
        self.data = data

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        if (isinstance(self.data, str)):
            data = "\"{}\"".format(self.data)
        else:
            data = self.data
        res = "PyRinCmdMessage(address={}, command={}, register={}, channel={}, data={})".format(self.address, self.command, self.register, self.channel, data)
        return res

    @classmethod
    def get_supported_commands(cls):
        """Get a list of supported pyrincmd commands"""
        return dict((key, value) for key, value in cls.__dict__.iteritems() if key.startswith("CMD_"))

    @classmethod
    def get_command_name(cls, cmd):
        """Get the name of a command given a value (ie 0x01 -> 'CMD_READ_TYPE')"""
        cmds = cls.get_supported_commands()
        cmddict = dict((key, value) for key, value in cmds.iteritems() if value == cmd)
        return cmddict.keys()[0]

    @classmethod
    def get_command_value(cls, cmd):
        """Get a command value given a key (ie 'CMD_READ_TYPE' -> 0x01)"""
        return cls.__dict__[cmd]

    def is_response(self):
        """Is the message a response"""
        return ((self.address & self.ADDR_FLAG_RESPONSE) != 0)

    def is_reply_requested(self):
        """Is the message a response"""
        return ((self.address & self.ADDR_FLAG_REPLY) != 0)

    def is_error(self):
        """Is the message an error reponse"""
        return ((self.address & self.ADDR_FLAG_ERROR) != 0)

    def get_address(self):
        """Return the address part of the address byte"""
        return self.address & self.ADDR_MASK

    def is_addressed_to_me(self, my_addr):
        """Is the message an error reponse"""
        msg_addr = self.get_address()
        return (msg_addr == self.ADDR_BROADCAST) or (msg_addr == my_addr)

    def does_address_match_command(self, cmd_addr):
        """Does the reply address match the command address"""
        msg_addr = self.get_address()
        cmd_addr &= self.ADDR_MASK
        return (cmd_addr == self.ADDR_BROADCAST) or (msg_addr == cmd_addr)

    def data_to_string(self):
        """Convert data to a string as appropriate for the command"""
        if self.data is None:
            res = ""
        elif isinstance(self.data, str):
            res = rc_escape.pyrincmd_escape(data=self.data)
        elif isinstance(self.data, int):
            if (PyRinCmdMessage.command_has_hex_data(command=self.command) or self.is_error()):
                data = (self.data & 0xFFFFFFFF) if (self.data < 0) else self.data
                res = "{:X}".format(data)
            elif (PyRinCmdMessage.command_has_dec_data(command=self.command)):
                res = "{:d}".format(self.data)
            else:
                pass # TODO error out
        else:
            pass # TODO error out
        return res

    def to_string(self):
        """Convert the rincomms message into a string suitable for transmission"""
        data_str = self.data_to_string()
        if (self.channel is not None):
            rcm_str = "{addr:02X}{cmd:02X}{reg:04X}{chan:x}:{data}".format(addr=self.address, cmd=self.command, reg=self.register, chan=self.channel, data=data_str)
        else:
            rcm_str = "{addr:02X}{cmd:02X}{reg:04X}:{data}".format(addr=self.address, cmd=self.command, reg=self.register, data=data_str)
        return rcm_str

    def to_pretty_string(self):
        """Convert the rincmd message into a pretty string with full message details"""

        # Address field
        respstr = "ADDR_FLAG_RESPONSE " if self.is_response() else ""
        errstr = "ADDR_FLAG_ERROR " if self.is_error() else ""
        replystr = "ADDR_FLAG_REPLY " if self.is_reply_requested() else ""
        indaddr = self.get_address()
        pretty = "Address: 0x{addr:2X} addr:{indaddr} {resp}{err}{reply}\n".format(addr=self.address,
                                                                                 indaddr=indaddr,
                                                                                 resp=respstr, err=errstr, reply=replystr)

        # Command field
        cmdstr = PyRinCmdMessage.get_command_name(cmd=self.command)
        pretty += "Command: 0x{cmd:02X} {cmdstr}\n".format(cmd=self.command,
                                                           cmdstr=cmdstr)

        # Register field
        pretty += "Register: 0x{reg:04X}\n".format(reg=self.register)

        # Channel field
        pretty += "Channel: {chan}\n".format(chan=self.channel)

        # Data field
        if (self.is_response() and self.is_write_command(command=self.command) and not self.is_error()):
            err = rc_error.PyRinCmdError(major=(self.data >> 8) & 0xFF, minor=self.data & 0xFF)
            datastr = "(response) " + err.get_pretty_string()
        elif (self.is_response() and self.is_error()):
            err = rc_error.PyRinCmdError(major=(self.data >> 8) & 0xFF, minor=self.data & 0xFF)
            datastr = "(error) " + err.get_pretty_string()
        else:
            datastr = self.data
        pretty += "Data: {data}".format(data=datastr)
        return pretty

    def convert_data_to_type(self, datatype):
        """Convert the object data member to datatype"""
        self.data = PyRinCmdMessage.data_to_datatype(data=self.data, datatype=datatype, command=self.command)

    @classmethod
    def is_write_command(cls, command):
        return command in [ cls.CMD_WRITE_HEXADEC, cls.CMD_WRITE_DECIMAL ]

    @classmethod
    def is_execute_command(cls, command):
        return command in [ cls.CMD_EXECUTE ]

    @classmethod
    def is_read_command(cls, command):
        return command in [ cls.CMD_READ_TYPE, cls.CMD_READ_RANGE_MIN, cls.CMD_READ_RANGE_MAX,
                            cls.CMD_READ_LITERAL, cls.CMD_READ_PERMISSION, cls.CMD_READ_HEXADEC,
                            cls.CMD_READ_DECIMAL ]

    @classmethod
    def command_has_hex_data(cls, command):
        """Does command have hexadecimal data"""
        return command in [ cls.CMD_READ_TYPE,
                            cls.CMD_READ_PERMISSION, cls.CMD_EXECUTE, cls.CMD_READ_HEXADEC,
                            cls.CMD_WRITE_HEXADEC ]

    @classmethod
    def command_has_dec_data(cls, command):
        """Does command have decimal data"""
        return command in [ cls.CMD_READ_RANGE_MIN, cls.CMD_READ_RANGE_MAX,
                            cls.CMD_READ_DECIMAL, cls.CMD_WRITE_DECIMAL, cls.CMD_READ_LITERAL ]

    @staticmethod
    def hexdata_to_int32(data):
        """Rincmd has a limit of 32bits, so we need to do some special handling of negative hexadecimal numbers

        ie: ABCD->0xABCD
            FFFF222B -> -DDD5
        """
        intdata = int(data, 16)
        if (intdata > (1 << 31) - 1):
            intdata = -((~intdata) & 0xFFFFFFFF) - 1
        return intdata

    @classmethod
    def data_to_datatype(cls, data, datatype, command):
        if (datatype is str):
            ret = rc_escape.pyrincmd_unescape(data=data)
        elif (datatype is int):
            ret = cls.hexdata_to_int32(data) if (cls.command_has_hex_data(command)) else int(data)
        elif (datatype is float):
            ret = float(data)
        elif (callable(datatype)):
            ret = datatype(data)
        elif (datatype == 'guess'):
            try:
                ret = int(data)
            except ValueError:
                try:
                    ret = cls.hexdata_to_int32(data)
                except ValueError:
                    try:
                        ret = float(data)
                    except ValueError:
                        ret = rc_escape.pyrincmd_unescape(data=data)
        elif (datatype == 'fromcmd'):
            if (command in [ cls.CMD_READ_HEXADEC, cls.CMD_WRITE_HEXADEC ]):
                # Might be hex number or string
                try:
                    ret = cls.hexdata_to_int32(data)
                except ValueError:
                    # Or a string
                    ret = rc_escape.pyrincmd_unescape(data=data)
            elif (command in [ cls.CMD_READ_DECIMAL, cls.CMD_WRITE_DECIMAL ]):
                # Might be decimal number or string
                try:
                    ret = int(data)
                except ValueError:
                    ret = rc_escape.pyrincmd_unescape(data=data)
            elif (cls.command_has_hex_data(command)):
                # Always hex
                ret = cls.hexdata_to_int32(data)
            elif (cls.command_has_dec_data(command)):
                # Always decimal
                ret = int(data)
            else:
                # TODO self.CMD_READ_LITERAL is probably a special case
                raise Exception("Could not convert register data by type 'fromcmd'")
        elif (datatype == 'ask'):
            # TODO
            raise Exception("TODO convert register data by type 'ask' not implemented")

        return ret


class PyRinCmdParser(object):
    """Parser to pull apart a rincomms string into a PyRinCmdMessage"""

    def __init__(self):
        pass

    def deal_with_data(self, address, command, data, datatype):
        if ((address & PyRinCmdMessage.ADDR_FLAG_ERROR) != 0):
            # On error, reply is a 16bit hex error code
            dataconv = int(data, 16)
        elif ((address & PyRinCmdMessage.ADDR_FLAG_RESPONSE) != 0 and PyRinCmdMessage.is_write_command(command)):
            # Write commands respond with a hex result code
            dataconv = int(data, 16)
        else:
            # Otherwise try to convert data
            dataconv = PyRinCmdMessage.data_to_datatype(data=data, datatype=datatype, command=command)
        return dataconv

    def str_to_message(self, message_str, datatype):
        """Convert message_str into PyRinCmdMessage

           datatype can be:
             int (for both decimal and hexadecimal)
             str
             float
             function(x): eg lambda x: int(x, 16)
             'guess': Try to guess the data type
             'fromcmd': Try to guess the data type from the command
             'ask': Ask the indicator what the data type is. Requires extra traffic (not handled here, implemented at a higher level)
        """

        pattern = '([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{4})([^:]*):(.*)'
        m = re.match(pattern=pattern, string=message_str)
        if m is not None:
            address = int(m.group(1), 16)
            command = int(m.group(2), 16)
            register = int(m.group(3), 16)
            channel = None if (m.group(4) == "") else m.group(4)
            data = m.group(5)

            dataconv = self.deal_with_data(address=address, command=command, data=data, datatype=datatype)
            rcm = PyRinCmdMessage(address=address, command=command, register=register, channel=channel, data=dataconv)
        else:
            raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_PARSE, minor=0)
        return rcm

    def check_reply(self, cmd, reply):
        """Check that rincmd message reply matches the command cmd that was sent"""
        if (not reply.is_response()):
            raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_PARSE, minor=0)
        if (not reply.does_address_match_command(cmd.address)):
            raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_PARSE, minor=0)
        if (reply.command != cmd.command):
            raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_PARSE, minor=0)
        if (reply.register != cmd.register):
            raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_PARSE, minor=0)
        if (cmd.channel is not None):
            if (reply.channel != cmd.channel):
                raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_PARSE, minor=0)
