import xml.etree.ElementTree
import datetime
import os
import rc_register
import rc_reg_translate


class PyRinCmdRegTranslate_C(rc_reg_translate.PyRinCmdRegTranslate):
    """Class for translating rincmd registers XML into C"""

    def get_h_ifndef_define(self, hfilename):
        """Generate FILE_DOT_H from file.h. Not particularly bulletproof: doesn't deal with strange characters in hfilename (eg -)"""
        filename, file_extension = os.path.splitext(os.path.basename(hfilename))
        file_extension = file_extension.replace('.', '')
        define = "{}_DOT_{}".format(filename.upper(), file_extension.upper())
        return define

    def append_name_comment_to_declaration(self, dec, name, maxlen):
        """Append a name comment to a declaration. Eg { asdf }, /* Name */ """
        newdec = "{dec:{width}s}  /* {name:s} */".format(dec=dec, width=maxlen, name=name)
        return newdec

    def append_name_comment_to_declarations(self, decs):
        """Append a name comment to each line in a declaration. Eg { asdf }, /* Name */
           decs is list of tuples (dec, comment)
           Requires len(decs) == len(names)"""
        if (len(decs) == 0):
            return []
        maxlen = len(max(decs, key=lambda x: len(x[0]))[0])
        new_decs = []
        for dec in decs:
            newdec = self.append_name_comment_to_declaration(dec=dec[0], name=dec[1], maxlen=maxlen)
            new_decs.append(newdec)
        return new_decs

    def create_c_array_for_cdefs(self, fmtstr, cdefs, regname):
        """Create a c array for cdefs"""
        if (len(cdefs) > 0):
            carray = fmtstr.format(regname=regname, defs='\n    '.join(cdefs))
        else:
            carray = ""
        return carray

    def register_to_c_defs(self, reg, type_table_idx, dispsprt_table_idx):
        """Convert a register into C definitions"""

        reg_id = reg.register_id
        reg_comment = "0x{:04X} - {}".format(reg_id, reg.name)
        ctype = reg.get_register_type_cdef()
        cmin = reg.get_register_minimum_cexp()
        cmax = reg.get_register_maximum_cexp()
        cvalue = reg.value
        cread_p = reg.get_register_read_permissions_cdef()
        cwrite_p = reg.get_register_write_permissions_cdef()

        if (reg.function is not None):
            func = '&' + reg.function if (reg.function != "NULL") else "NULL"
        else:
            func = "NULL"

        dispsprt_c_def = None
        type_c_def = None

        if (reg.type == 'register'):
            type_c_def = "{{ .{}Data = {{.min={}, .max={}, .value={} }} }},".format(reg.type, cmin, cmax, cvalue)
        elif (reg.type == 'ipaddress'):
            type_c_def = "{{ .ipData = {{ .value={} }} }},".format(cvalue)
        elif (reg.is_integer_type()):
            if (reg.dispsprt_pointer is not None):
                cdispptr = reg.dispsprt_pointer
            elif (reg.dispsprt_dp is not None):

                units = rc_register.PyRinCmdRegister.reg_units[reg.dispsprt_units]
                unitsother = rc_register.PyRinCmdRegister.reg_units_other[reg.dispsprt_unitsother]

                dispsprt_c_def = "{{ .dp={}, .units={}, .unitOther={} }},".format(reg.dispsprt_dp, units, unitsother)
                cdispptr = "&dsp[{}]".format(dispsprt_table_idx)
            else:
                cdispptr = "NULL"

            type_c_def = "{{ .{}Data = {{.min={}, .max={}, .value={}, .dispSprt={}}} }},".format(reg.type, cmin, cmax, cvalue, cdispptr)

        elif (reg.type == 'string'):
            type_c_def = "{{ .{}Data = {{.min={}, .max={}, .value={}, .minCh={}, .maxCh={} }} }},".format(reg.type, cmin, cmax, cvalue, reg.min_ch, reg.max_ch)

        elif (reg.type == 'binbuffer'):
            type_c_def = "{{ .{}Data = {{.min={}, .max={}, .value={}, .length={}, .minCh={}, .maxCh={} }} }},".format(reg.type, cmin, cmax, cvalue, reg.length, reg.min_ch, reg.max_ch)

        elif (reg.type == 'weight'):
            type_c_def = "{{ .{}Data = {{.min={}, .max={}, .value={} }} }},".format(reg.type, cmin, cmax, cvalue)

        if (reg.type == 'execute'):
            reg_c_def = "{{ .addr=0x{:04X}, .type={}, .perm={{.PByte=({}|{})}}, .regFunc={}, .typDat=NULL }},".format(reg_id, ctype, cread_p, cwrite_p, func)
        else:
            reg_c_def = "{{ .addr=0x{:04X}, .type={}, .perm={{.PByte=({}|{})}}, .regFunc={}, .typDat=&typ[{}] }},".format(reg_id, ctype, cread_p, cwrite_p, func, type_table_idx)
        return reg_comment, reg_c_def, type_c_def, dispsprt_c_def

    def registers_to_c_defs(self):
        """Convert all registers in input XML into C definitions"""

        reg_c_defs = []
        type_c_defs = []
        dispsprt_c_defs = []
        reg_idx_hash_defs = []

        for reg in self.registers:
            type_table_idx = len(type_c_defs)
            dispsprt_table_idx = len(dispsprt_c_defs)

            reg_comment, reg_c_def, type_c_def, dispsprt_c_def = self.register_to_c_defs(reg=reg, type_table_idx=type_table_idx, dispsprt_table_idx=dispsprt_table_idx)

            reg_c_defs.append((reg_c_def, reg_comment))
            if (type_c_def is not None):
                type_c_defs.append((type_c_def, reg_comment))
            if (dispsprt_c_def is not None):
                dispsprt_c_defs.append((dispsprt_c_def, reg_comment))

            prefix = self.get_c_python_define_prefix()
            reg_idx_hash_defs.append(reg.get_register_id_c_define(prefix=prefix))

        type_c_defs = self.append_name_comment_to_declarations(decs=type_c_defs)
        reg_c_defs = self.append_name_comment_to_declarations(decs=reg_c_defs)
        dispsprt_c_defs = self.append_name_comment_to_declarations(decs=dispsprt_c_defs)

        return reg_c_defs, type_c_defs, dispsprt_c_defs, reg_idx_hash_defs

    def headers_to_c_defs(self):
        """Convert all headers in input XML into C definitions"""

        hash_includes = []
        for header in self.headers:
            htype = header[0]
            hfile = header[1]
            quotel = "\"" if (htype == "local") else "<"
            quoter = "\"" if (htype == "local") else ">"
            hinclude = '#include {ql}{hfile}{qr}'.format(ql=quotel, hfile=hfile, qr=quoter)
            hash_includes.append(hinclude)
        return hash_includes

    def local_stream_regs_to_c_defs(self):

        idx = 0

        typedata = []
        regdata = []
        streamlist = []
        freeabledata = []
        freeonfaildata = []
        for idx, reg in enumerate(self.local_stream_registers):

            reg_id = reg.register_id
            reg_comment = "/* 0x{:04X} - {} */".format(reg_id, reg.name)
            ctype = reg.get_register_type_cdef()
            cmin = reg.get_register_minimum_cexp()
            cmax = reg.get_register_maximum_cexp()
            cread_p = reg.get_register_read_permissions_cdef()
            cwrite_p = reg.get_register_write_permissions_cdef()

            regdata.append(reg_comment)
            regdata.append("arr[{idx}].addr = 0x{rid:04X};".format(idx=idx, rid=reg_id))
            regdata.append("arr[{idx}].type = {type};".format(idx=idx, type=ctype))
            regdata.append("arr[{idx}].perm.PByte = ({rp} | {wp});".format(idx=idx, rp=cread_p, wp=cwrite_p))

            cfunc = reg.function if (reg.function is not None) else "NULL"
            regdata.append("arr[{idx}].regFunc = {func};".format(idx=idx, func=cfunc))
            regdata.append("arr[{idx}].typDat = &typ[{idx}];\n".format(idx=idx))

            typedata.append(reg_comment)
            if (reg.type == "stream"):
                typedata.append("typ[{idx}].streamData.data = rc_initStreamData();".format(idx=idx))
                typedata.append("if (NULL == typ[{idx}].streamData.data) goto rgl_fail;\n".format(idx=idx))
                streamlist.append("&arr[{idx}],".format(idx=idx))
                freeabledata.append("rc_addFreeableData(topLevel, typ[{idx}].streamData.data);".format(idx=idx))
                freeonfaildata.append("if (NULL != typ[{idx}].streamData.data) {{ free(typ[{idx}].streamData.data); typ[{idx}].streamData.data = NULL; }}".format(idx=idx))
            elif (reg.is_integer_type()):
                typedata.append("typ[{idx}].{type}Data.min = {min};".format(idx=idx, type=reg.type, min=cmin))
                typedata.append("typ[{idx}].{type}Data.max = {max};".format(idx=idx, type=reg.type, max=cmax))
                typedata.append("typ[{idx}].{type}Data.value = {val};".format(idx=idx, type=reg.type, val=reg.value))
                if (reg.type != "register"):
                    typedata.append("typ[{idx}].{type}Data.dispSprt = NULL;\n".format(idx=idx, type=reg.type))
            elif (reg.type == "string"):
                # TODO string
                pass
            elif (reg.type == "weight"):
                # TODO weight
                pass
        return typedata, regdata, streamlist, freeabledata, freeonfaildata

    def xml_registers_to_c(self, outcfilename=None, outhfilename=None):
        """Generate C & H files containing all global and local registers
           Returns tuple of (cfilename, cfilestr, hfilename, hfilestr)"""

        reg_c_defs, type_c_defs, dispsprt_c_defs, reg_idx_hash_defs = self.registers_to_c_defs()
        hash_includes = self.headers_to_c_defs()
        local_type_c_defs, local_reg_c_defs, local_streamlist, local_freeabledata, local_freeonfaildata = self.local_stream_regs_to_c_defs()

        if (outhfilename is None):
            outhfilename = self.get_output_filename(ext="h")
        if (outcfilename is None):
            outcfilename = self.get_output_filename(ext="c")
        hdefine = self.get_h_ifndef_define(outhfilename)

        root = self.xmltree.getroot()
        localTypeData = []
        localArrData = []
        localStreamData = []
        localFreeableData = []
        localFreeIfFail = []

        numLocalTyp = 0
        numLocalArr = 0

        for child in root:
            if child.tag == "include":
                # This is all done by headers_to_c_defs() now
                pass

            if child.tag == "registers":
                # This is all done by registers_to_c_defs() now
                pass

            if child.tag == "local":
                for local in child:
                    if local.tag == "stream":
                        rid = local.get("addr")
                        localRegName = local.find('name').text
                        intId = int(rid, 16)
                        readP = rc_register.PyRinCmdRegister.reg_perm_read[local.find('permission').find('read').text]["cdef"]
                        writeP = rc_register.PyRinCmdRegister.reg_perm_write[local.find('permission').find('write').text]["cdef"]

                        startTypIndex = numLocalTyp

                        localTypeData.append("/* 0x{:04X} - {} - data */".format(intId, localRegName))
                        localTypeData.append("typ[{}].streamData.data = rc_initStreamData();".format(numLocalTyp))
                        localTypeData.append("if (NULL == typ[{}].streamData.data) goto rgl_fail;\n".format(numLocalTyp))
                        localArrData.append("/* 0x{:04X} - {} - data */".format(intId, localRegName))
                        localArrData.append("arr[{}].addr = {};".format(numLocalArr, rid))
                        localArrData.append("arr[{}].type = RC_REG_TYPE_STREAM;".format(numLocalArr))
                        localArrData.append("arr[{}].perm.PByte = ({} | {});".format(numLocalArr, readP, writeP))
                        localArrData.append("arr[{}].regFunc = &rc_setStreaming;".format(numLocalArr))
                        localArrData.append("arr[{}].typDat = &typ[{}];\n".format(numLocalArr, startTypIndex))
                        localStreamData.append("&arr[{}],".format(numLocalArr))
                        numLocalTyp += 1
                        numLocalArr += 1

                        localTypeData.append("/* 0x{:04X} - {} - mode */".format(intId + 1, localRegName))
                        localTypeData.append("typ[{}].uint8Data.min = 0;".format(numLocalTyp))
                        localTypeData.append("typ[{}].uint8Data.max = RC_STREAM_MODE_NUM - 1;".format(numLocalTyp))
                        localTypeData.append("typ[{}].uint8Data.value = &typ[{}].streamData.data->freq;".format(numLocalTyp, startTypIndex))
                        localTypeData.append("typ[{}].uint8Data.dispSprt = NULL;\n".format(numLocalTyp))
                        localArrData.append("/* 0x{:04X} - {} - mode */".format(intId + 1, localRegName))
                        localArrData.append("arr[{0}].addr = {1:#06x};".format(numLocalArr, intId+1))
                        localArrData.append("arr[{}].type = RC_REG_TYPE_UINT8;".format(numLocalArr))
                        localArrData.append("arr[{}].perm.PByte = ({} | {});".format(numLocalArr, readP, writeP))
                        localArrData.append("arr[{}].regFunc = NULL;".format(numLocalArr))
                        localArrData.append("arr[{}].typDat = &typ[{}];\n".format(numLocalArr, numLocalTyp))
                        numLocalTyp += 1
                        numLocalArr += 1

                        for i in range(0, self.NUM_STREAM_REGS_PER_SET):
                            localTypeData.append("/* 0x{:04X} - {} - stream register {} */".format(intId + i + 2, localRegName, i))
                            localTypeData.append("typ[{}].regData.min = 0;".format(numLocalTyp))
                            localTypeData.append("typ[{}].regData.max = 0xFFFF;".format(numLocalTyp))
                            localTypeData.append("typ[{}].regData.value = &typ[{}].streamData.data->regAddrs[{}];\n".format(numLocalTyp, startTypIndex, i))
                            localArrData.append("/* 0x{:04X} - {} - stream register {} */".format(intId + i + 2, localRegName, i))
                            localArrData.append("arr[{0}].addr = {1:#06x};".format(numLocalArr, intId+2+i))
                            localArrData.append("arr[{}].type = RC_REG_TYPE_REGISTER;".format(numLocalArr))
                            localArrData.append("arr[{}].perm.PByte = ({} | {});".format(numLocalArr, readP, writeP))
                            localArrData.append("arr[{}].regFunc = &rc_handleAddStreamDataReg;".format(numLocalArr))
                            localArrData.append("arr[{}].typDat = &typ[{}];\n".format(numLocalArr, numLocalTyp))
                            numLocalTyp += 1
                            numLocalArr += 1

                        localFreeableData.append("rc_addFreeableData(topLevel, typ[{}].streamData.data);".format(startTypIndex))
                        localFreeIfFail.append("if (NULL != typ[{idx}].streamData.data) {{ free(typ[{idx}].streamData.data); typ[{idx}].streamData.data = NULL; }}".format(idx=startTypIndex))

                    elif local.tag == "reg":
                        pass

        dispsprt_template = \
"""static const struct s_rc_DispSprt dsp[DISPSPRT_TABLE_{regname}_NELMS] =
  {{
    {defs}
  }};
"""

        typedata_template = \
"""static const union u_rc_TypData typ[TYPEDATA_TABLE_{regname}_NELMS] =
  {{
    {defs}
  }};
"""

        regdata_template = \
"""static const struct s_rc_Register arr[REG_TABLE_{regname}_NELMS] =
  {{
    {defs}
  }};
"""

        dispsprt_decl = self.create_c_array_for_cdefs(fmtstr=dispsprt_template, cdefs=dispsprt_c_defs, regname=self.register_group_name.upper())
        typedata_decl = self.create_c_array_for_cdefs(fmtstr=typedata_template, cdefs=type_c_defs, regname=self.register_group_name.upper())
        regdata_decl = self.create_c_array_for_cdefs(fmtstr=regdata_template, cdefs=reg_c_defs, regname=self.register_group_name.upper())

        hTemplate = \
"""/*
 * {outfile}
 *
 * Generated by rc_registerBuilder.py
 *
 * AUTOMATICALLY GENERATED FILE. DO NOT EDIT!
 *
 * Generated from: {infile}
 * Generated at: {date}
 */

#ifndef {hdefine}
#define {hdefine}

#include <stdbool.h>
#include <rincmd/rc_public.h>


{reg_idx_hash_defs}


/** Define all global registers and add them to table */
void rc_gen_{name}(struct s_rc_registerTable *table);


/** Allocate and configure registers, and then add to local register table topLevel. Returns false if allocation failed */
bool rc_gen_local_{name}(struct s_rc_localRegisterTable *topLevel);


#endif /* {hdefine} */
"""    .format(infile=self.xmlfilename,
                   outfile=outhfilename,
                   hdefine=hdefine,
                   reg_idx_hash_defs='\n'.join(reg_idx_hash_defs),
                   name=self.register_group_name,
                   date=str(datetime.datetime.now())
                   )

        cTemplate = \
"""/*
 * {outfile}
 *
 * Generated by rc_registerBuilder.py
 *
 * AUTOMATICALLY GENERATED FILE. DO NOT EDIT!
 *
 * Generated from: {infile}
 * Generated at: {date}
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include <rincmd/rc_public.h>
#include <rincmd/rc_registers.h>
#include <rincmd/rc_streaming.h>

{hash_includes}

#include "rc_gen_{name}.h"


#define DISPSPRT_TABLE_{upname}_NELMS {numDispSprt}
#define TYPEDATA_TABLE_{upname}_NELMS {numTypeData}
#define REG_TABLE_{upname}_NELMS {numRegisters}
#define LOCAL_REG_TABLE_{upname}_NELMS {numLocalArrs}


/** Define all global registers and add them to table */
void rc_gen_{name}(struct s_rc_registerTable *table)
{{
  {dispsprt}

  {typedata}

  {regdata}

  rc_addRegs(table, arr, REG_TABLE_{upname}_NELMS);
}}


/** Allocate and configure registers, and then add to local register table topLevel. Returns false if allocation failed */
bool rc_gen_local_{name}(struct s_rc_localRegisterTable *topLevel)
{{
  union u_rc_TypData *typ = NULL;
  struct s_rc_Register *arr = NULL;

  if (NULL == topLevel)
    return false;

  typ = calloc({numLocalTyps}, sizeof(union u_rc_TypData));
  if (NULL == typ)
    goto rgl_fail;

  {localTypeData}

  arr = calloc(LOCAL_REG_TABLE_{upname}_NELMS, sizeof(struct s_rc_Register));
  if (NULL == arr)
    goto rgl_fail;

  {localArrData}

  /* List of streams to add, rc_addStreams() takes a copy internally */
  struct s_rc_Register *stream[] =
  {{
    {localStreamData}
    NULL,
  }};

  rc_addLocalRegs(topLevel, arr, LOCAL_REG_TABLE_{upname}_NELMS);
  rc_addStreams(topLevel, stream);
  {localFreeableData}
  rc_addFreeableData(topLevel, typ);
  rc_addFreeableData(topLevel, arr);

  return true;

rgl_fail:
  if (NULL != typ)
  {{
    {localFreeIfFail}
    free(typ);
  }}
  if (NULL != arr)
  {{
    free(arr);
  }}
  return false;
}}
"""    .format(infile = self.xmlfilename,
                   outfile = outcfilename,
                   hdefine = hdefine,
                   name = self.register_group_name,
                   upname = self.register_group_name.upper(),
                   date = str(datetime.datetime.now()),
                   hash_includes='\n'.join(hash_includes),
                   dispsprt=dispsprt_decl,
                   typedata=typedata_decl,
                   regdata=regdata_decl,
                   numDispSprt=len(dispsprt_c_defs),
                   numTypeData=len(type_c_defs),
                   numRegisters=len(reg_c_defs),
                   numLocalTyps = numLocalTyp,
                   numLocalArrs = numLocalArr,
                   localTypeData = '\n  '.join(local_type_c_defs),
                   localArrData = '\n  '.join(local_reg_c_defs),
                   localStreamData = '\n    '.join(local_streamlist),
                   localFreeableData = '\n  '.join(local_freeabledata),
                   localFreeIfFail = '\n    '.join(local_freeonfaildata)
                   )

        return outcfilename, cTemplate, outhfilename, hTemplate

    def xml_registers_to_c_variable_declarations(self, extern=False):
        """Convert the registers into a string of C variable declarations.
           Used for testing.
           Example: int8 -> int8_t variable = 0;
        """
        c_var_decls = []
        for reg in self.registers:
            c_var_decl = reg.get_register_c_variable_declaration(extern=extern)
            c_var_decls.append(c_var_decl)
        return "\n".join(c_var_decls)

    def xml_register_to_ch_file(self, outputfolder=None, outputcfilename=None, outputhfilename=None):
        """Convert XML registers into a .c and .h file in outputfolder

           If outputfolder is not None, then it is prepended to outputcfilename and outputhfilename
           If outputcfilename is None, then it will be 'rc_gen_<xmlname>.c'
           If outputhfilename is None, then it will be 'rc_gen_<xmlname>.h'
        """
        if (outputcfilename is None):
            outputcfilename = self.get_output_filename(ext="c")
        if (outputhfilename is None):
            outputhfilename = self.get_output_filename(ext="h")
        if (outputfolder is not None):
            outputcfilename = os.path.join(outputfolder, outputcfilename)
            outputhfilename = os.path.join(outputfolder, outputhfilename)

        _, outcstr, _, outhstr = self.xml_registers_to_c(outcfilename=outputcfilename, outhfilename=outputhfilename)

        f = open(outputcfilename, 'w')
        f.write(outcstr)
        f.close()

        f = open(outputhfilename, 'w')
        f.write(outhstr)
        f.close()
