'''
Created on Sep 13, 2017

@author: ian
'''
import sys
import xml.etree.ElementTree
import rc_register


class PyRinCmdRegTranslate(object):
    """Base class for translating rincmd registers XML into other things (C/latex)"""

    NUM_STREAM_REGS_PER_SET = 5

    def __init__(self, xmltree):
        """Create a list of registers by importing a register XML file
           xml can be filename, file object or existing XML tree"""
        strtypes = (str, unicode) if sys.version_info < (3, 0, 0) else (bytes, str)
        if (isinstance(xmltree, strtypes)):
            self.xmlfilename = xmltree
            self.xmltree = xml.etree.ElementTree.parse(xmltree)
        elif (hasattr(xmltree, 'read')):
            self.xmlfilename = xmltree.name if (hasattr(xmltree, 'name')) else "unknown"
            self.xmltree = xml.etree.ElementTree.parse(xmltree)
        else:
            self.xmlfilename = "unknown"
            self.xmltree = xmltree

        self.register_group_name = self.xmltree.getroot().get('name')

        self.import_registers()
        self.import_headers()
        self.import_local_registers()

    def import_registers(self):
        """Import the registers from the XML"""
        self.registers = rc_register.PyRinCmdRegisterListImport(xmltree=self.xmltree, constants=None)

    def import_headers(self):
        """Import the headers from the XML"""
        root = self.xmltree.getroot()

        self.headers = []
        for child in root:
            if child.tag == "include":
                self.headers.append((child.get("type"), child.text))

    def make_local_stream_registers(self, xmlelt):
        rid = int(xmlelt.get("addr"), 0)
        name = xmlelt.find('name').text
        perm_read = xmlelt.find('permission').find('read').text
        perm_write = xmlelt.find('permission').find('write').text

        # Make stream data register
        dname = name + " Data"
        ddesc = ("Returns a block of data which is selected in Stream Register 1-5. \\newline "
                 "Use a read command to read a single set of data. \\newline "
                 "Use an execute command (with a parameter of 1) to switch on automatic transmition at the mode speed.")
        params = { "addr": rid, "name": dname, "description": ddesc,
                   "type": "stream",
                   "permread": perm_read, "permwrite": perm_write,
                   "function": "&rc_setStreaming" }
        reg = rc_register.PyRinCmdRegister(params=params)
        self.local_stream_registers.append(reg)
        mystreamregidx = len(self.local_stream_registers) - 1

        # Make stream mode register
        mname = name + " Mode"
        mdesc = ("0: Manual - read `Stream Data' register \\newline "
                 "1: Auto Sync - Data is sent at sync freqency \\newline "
                 "2: Auto 10Hz - Data is sent at 10Hz \\newline "
                 "3: Auto 5Hz - Data is sent at 5Hz \\newline "
                 "4: Auto 1Hz - Data is sent at 1Hz ")
        mvalue = "&typ[{}].streamData.data->freq".format(mystreamregidx)
        params = { "addr": rid + 1, "name": mname, "description": mdesc,
                   "type": "uint8",
                   "min": 0, "max": "RC_STREAM_MODE_NUM - 1",
                   "value": mvalue,
                   "permread": perm_read, "permwrite": perm_write }
        reg = rc_register.PyRinCmdRegister(params=params)
        self.local_stream_registers.append(reg)

        for i in range(0, self.NUM_STREAM_REGS_PER_SET):
            # Make stream mode register
            srname = name + ": Register {}".format(i)
            srdesc = "Register to stream"
            srvalue = "&typ[{}].streamData.data->regAddrs[{}]".format(mystreamregidx, i)
            params = { "addr": rid + 2 + i, "name": srname, "description": srdesc,
                       "type": "register",
                       "min": 0, "max": 0xFFFF,
                       "permread": perm_read, "permwrite": perm_write,
                       "value": srvalue,
                       "function": "&rc_handleAddStreamDataReg" }
            reg = rc_register.PyRinCmdRegister(params=params)
            self.local_stream_registers.append(reg)

    def import_local_registers(self):
        """Import the local registers from the XML"""

        root = self.xmltree.getroot()
        self.local_registers = []
        self.local_stream_registers = []

        for child in root:
            if child.tag == "local":
                for local in child:
                    if local.tag == "stream":
                        self.make_local_stream_registers(xmlelt=local)
                    elif local.tag == "reg":
                        # TODO local registers
                        pass

    def get_output_filename(self, ext):
        """Get an output filename for this register table"""
        return 'rc_gen_{name}.{ext}'.format(name=self.register_group_name, ext=ext)

    def get_c_python_define_prefix(self, prefix="RINREG"):
        """Get a prefix for for this register table (ie RINREG_<tablename>_"""
        return "{}_{}".format(prefix, self.register_group_name.upper())
