'''
Created on Sep 4, 2017

@author: ian
'''
import sys
import unittest
import xml.etree.ElementTree
import rc_error


class PyRinCmdRegister(object):
    '''Class to represent a rincmd register'''

    RC_REG_TYPE_INT8 = 0            # Register Type 0: Simple character value
    RC_REG_TYPE_UINT8 = 1           # Register Type 1: Simple unsigned 8-bit value
    RC_REG_TYPE_INT16 = 2           # Register Type 2: Signed 16-bit value
    RC_REG_TYPE_UINT16 = 3          # Register Type 3: Unsigned 16-bit value
    RC_REG_TYPE_INT32 = 4           # Register Type 4: Signed 32-bit value
    RC_REG_TYPE_UINT32 = 5          # Register Type 5: Unsigned 32-bit value
    RC_REG_TYPE_STRING = 6          # Register Type 6: Null terminated string
    # RC_REG_TYPE_OPTION = 7          # Register Type 7: Selectable Option
    # RC_REG_TYPE_MENU = 8            # Register Type 8: Menu type
    RC_REG_TYPE_WEIGHT = 9          # Register Type 9: Weight type
    # RC_REG_TYPE_BLOB = 10           # Register Type 10: Binary Large OBject
    RC_REG_TYPE_EXECUTE = 11        # Register Type 11: Executable
    # RC_REG_TYPE_BITFIELD = 12       # Register Type 12: Bit Field
    RC_REG_TYPE_STREAM = 13         # Register Type 13: register for streamed data
    # RC_REG_TYPE_EXECUTE_STR = 14    # Register Type 14: Executable with string param
    # RC_REG_TYPE_MENU_END = 15       # Register Type 15: Menu end type
    # RC_REG_TYPE_STRING_ARRAY = 16   # Register Type 16: String array
    # RC_REG_TYPE_OPTION16 = 17       # Register Type 17: Selectable Option, stored in a 16bit variable
    RC_REG_TYPE_IP = 18             # Register Type 18: IP Address
    RC_REG_TYPE_REGISTER = 19       # Register Type 19: used for streaming
    RC_REG_TYPE_BINBUFFER = 20      # Register Type 20: Binary buffer (specified by length, not null termination)

    reg_types = \
    {
        'int8':      { "cdef":"RC_REG_TYPE_INT8",      "value":RC_REG_TYPE_INT8,      "isint":True,  "ctype":"int8_t",      "pytype": int },
        'uint8':     { "cdef":"RC_REG_TYPE_UINT8",     "value":RC_REG_TYPE_UINT8,     "isint":True,  "ctype":"uint8_t",     "pytype": int },
        'int16':     { "cdef":"RC_REG_TYPE_INT16",     "value":RC_REG_TYPE_INT16,     "isint":True,  "ctype":"int16_t",     "pytype": int },
        'uint16':    { "cdef":"RC_REG_TYPE_UINT16",    "value":RC_REG_TYPE_UINT16,    "isint":True,  "ctype":"uint16_t",    "pytype": int },
        'int32':     { "cdef":"RC_REG_TYPE_INT32",     "value":RC_REG_TYPE_INT32,     "isint":True,  "ctype":"int32_t",     "pytype": int },
        'uint32':    { "cdef":"RC_REG_TYPE_UINT32",    "value":RC_REG_TYPE_UINT32,    "isint":True,  "ctype":"uint32_t",    "pytype": int },
        'string':    { "cdef":"RC_REG_TYPE_STRING",    "value":RC_REG_TYPE_STRING,    "isint":False,  "ctype":"char",       "pytype": str },
        'weight':    { "cdef":"RC_REG_TYPE_WEIGHT",    "value":RC_REG_TYPE_WEIGHT,    "isint":False,  "ctype":"",           "pytype": None },
        'execute':   { "cdef":"RC_REG_TYPE_EXECUTE",   "value":RC_REG_TYPE_EXECUTE,   "isint":False,  "ctype":"",           "pytype": None },
        'stream':    { "cdef":"RC_REG_TYPE_STREAM",    "value":RC_REG_TYPE_STREAM,    "isint":False,  "ctype":"",           "pytype": None },
        'ipaddress': { "cdef":"RC_REG_TYPE_IP",        "value":RC_REG_TYPE_IP,        "isint":True,   "ctype":"uint32_t",   "pytype": int },
        'register':  { "cdef":"RC_REG_TYPE_REGISTER",  "value":RC_REG_TYPE_REGISTER,  "isint":True,   "ctype":"rc_RegId_t", "pytype": int },
        'binbuffer': { "cdef":"RC_REG_TYPE_BINBUFFER", "value":RC_REG_TYPE_BINBUFFER, "isint":False,  "ctype":"uint8_t",    "pytype": str },
    }

    reg_perm_read = \
    {
        'none': { "cdef":"RC_PERM_RD_N", "value":0x00, "latex":"N" },
        'safe': { "cdef":"RC_PERM_RD_S", "value":0x01, "latex":"S" },
        'full': { "cdef":"RC_PERM_RD_F", "value":0x02, "latex":"F" },
        'all':  { "cdef":"RC_PERM_RD_A", "value":0x03, "latex":"A" },
    }

    reg_perm_write = \
    {
        'none': { "cdef":"RC_PERM_WR_N", "value":0x00, "latex":"N" },
        'safe': { "cdef":"RC_PERM_WR_S", "value":0x04, "latex":"S" },
        'full': { "cdef":"RC_PERM_WR_F", "value":0x08, "latex":"F" },
        'all':  { "cdef":"RC_PERM_WR_A", "value":0x0C, "latex":"A" },
    }

    reg_units = { '' : 'RC_DISP_UNITS_NONE',
             'none': 'RC_DISP_UNITS_NONE',
             'kg': 'RC_DISP_UNITS_KG',
             'lb': 'RC_DISP_UNITS_LB',
              't': 'RC_DISP_UNITS_T',
              'g': 'RC_DISP_UNITS_G',
             'oz': 'RC_DISP_UNITS_OZ',
              'N': 'RC_DISP_UNITS_N',
              'p': 'RC_DISP_UNITS_P',
              'L': 'RC_DISP_UNITS_L',
             }

    reg_units_other = {'':   'RC_DISP_UNITS_OTHER_NONE',
                  'none':   'RC_DISP_UNITS_OTHER_NONE',
                  'm':  'RC_DISP_UNITS_OTHER_M',
                  's':  'RC_DISP_UNITS_OTHER_S',
                  'h':  'RC_DISP_UNITS_OTHER_H',
                  '/':  'RC_DISP_UNITS_OTHER_SLASH',
                  '/m': 'RC_DISP_UNITS_OTHER_SLASH_M',
                  '/s': 'RC_DISP_UNITS_OTHER_SLASH_S',
                  '/h': 'RC_DISP_UNITS_OTHER_SLASH_H',
                  '%':   'RC_DISP_UNITS_OTHER_PERCENT',
                  }

    def __init__(self, *args, **kwargs):
        '''Construct with xmlelt=xml, constants=constant_namespace
        '''
        xmlelt = kwargs.get('xmlelt')
        params = kwargs.get('params')

        if (xmlelt is not None):
            self.init_from_xml(xmlelt=xmlelt)
        elif (params is not None):
            self.init_from_dict(params=params)

        self.constants = kwargs.get('constants')

    @classmethod
    def get_entry_for_regtype(cls, regtype):
        """Get the entry regtype. regtype can be integer, or "int8" or "RC_REG_TYPE_*" string """
        if (isinstance(regtype, str)):
            if (regtype in cls.reg_types.keys()):
                return cls.reg_types[regtype]
            else:
                elem = {k: v for k, v in cls.reg_types.items() if (v["cdef"] == regtype)}
                return elem.values()[0]
        else:
            elem = {k: v for k, v in cls.reg_types.items() if (v["value"] == regtype)}
            return elem.values()[0]

    @classmethod
    def get_python_type_for_regtype(cls, regtype):
        """Get the python type for a regtype. regtype can be integer, or "RC_REG_TYPE_*" string """
        elem = cls.get_entry_for_regtype(regtype=regtype)
        return elem["pytype"]

    def init_from_xml(self, xmlelt):
        self.register_id = int(xmlelt.get("addr"), 0)
        self.name = xmlelt.find("name").text
        self.description = xmlelt.find("description").text
        self.read_perms = xmlelt.find('permission').find('read').text
        self.write_perms = xmlelt.find('permission').find('write').text

        self.type = xmlelt.find("type").text
        if (self.type not in [ 'execute', 'ipaddress' ]):
            self.maximum = xmlelt.find('typeData').find("max").text
            self.minimum = xmlelt.find('typeData').find("min").text
        else:
            self.maximum = None
            self.minimum = None
        if (self.type == 'string' or self.type == 'binbuffer'):
            self.min_ch = xmlelt.find('typeData').find('minCh').text
            self.max_ch = xmlelt.find('typeData').find('maxCh').text
            if (self.type == 'binbuffer'):
                self.length = xmlelt.find('typeData').find('length').text
            self.emulation_value = ""
        else:
            self.emulation_value = 0

        if (xmlelt.find('typeData') is not None):
            self.value = xmlelt.find('typeData').find("value").text
        else:
            self.value = None

        if (xmlelt.find('function') is not None):
            self.function = xmlelt.find('function').text
        else:
            self.function = None

        dispsprt = xmlelt.find('displaySupport')
        if (dispsprt is not None):
            if (dispsprt.get('type') == "pointer"):
                self.dispsprt_pointer = dispsprt.find('value').text
                self.dispsprt_dp = None
                self.dispsprt_units = None
                self.dispsprt_unitsother = None
            else:
                self.dispsprt_pointer = None
                self.dispsprt_dp = dispsprt.find('dp').text
                self.dispsprt_units = dispsprt.findtext('units', "none")
                self.dispsprt_unitsother = dispsprt.findtext('otherunits', "none")
        else:
            self.dispsprt_pointer = None
            self.dispsprt_dp = None
            self.dispsprt_units = None
            self.dispsprt_unitsother = None

    def init_from_dict(self, params):
        self.register_id = params["addr"]
        self.name = params["name"]
        self.description = params["description"]
        self.read_perms = params["permread"]
        self.write_perms = params["permwrite"]

        self.type = params["type"]
        if (self.type not in [ 'execute', 'stream' ]):
            self.maximum = params["max"]
            self.minimum = params["min"]
        else:
            self.maximum = None
            self.minimum = None

        if (self.type == 'string'):
            self.min_ch = params['minCh']
            self.max_ch = params['maxCh']
            self.emulation_value = ""
        else:
            self.emulation_value = 0

        if ('value' in params):
            self.value = params['value']
        else:
            self.value = None

        if ('function' in params):
            self.function = params['function']
        else:
            self.function = None

    def get_nice_name(self):
        """Return the register name with no nasty characters, suitable for c or python define"""
        name = self.name.upper().replace(" ", "_")
        for badch in [ "\\", "/", ":", "(", ")", ".", "," ]:
            name = name.replace(badch, "")
        return name

    def get_register_id_c_define(self, prefix):
        """Return the register c define"""
        cdef = "#define {}_{} 0x{:04X}".format(prefix.upper(), self.get_nice_name(), self.register_id)
        return cdef

    def get_register_id_py_const(self, prefix):
        """Return the register as a python constant"""
        pydef = "{}_{}".format(prefix.upper(), self.get_nice_name(), self.register_id)
        return pydef, self.register_id

    def is_integer_type(self):
        """Is the register type one of the integer ones"""
        return (PyRinCmdRegister.get_python_type_for_regtype(regtype=self.type) is int)
        #inttypes = {k: v for k, v in self.reg_types.items() if v["isint"]}
        #return self.type in inttypes

    def evaluate_min_max_constants(self, expression):
        """Try to evaluate an integer value for min/max
           For example:
            register 0x0003 has <max>APP_DB_F_MODEL_MAXLEN + 1</max>
            self.constants.APP_DB_F_MODEL_MAXLEN = 10
            result would be 11
        """
        value = eval(expression, {}, self.constants.__dict__)
        return value

    def get_register_type_cdef(self):
        """Read the register type as a C define (eg RC_REG_TYPE_INT8)"""
        return self.reg_types[self.type]["cdef"]

    def get_register_type_python(self):
        """Read the register type as a python type (eg int, str)"""
        return self.reg_types[self.type]["pytype"]

    def get_register_type_int(self):
        """Read the register type as an integer (eg 'int8' = 0)"""
        return self.reg_types[self.type]["value"]

    def get_register_minimum_cexp(self):
        """Get the register minimum as a C expression"""
        return self.minimum

    def get_register_minimum_int(self):
        """Get the register minimum as an integer"""
        return self.evaluate_min_max_constants(self.minimum)

    def get_register_maximum_cexp(self):
        """Get the register maximum as a C expression"""
        return self.maximum

    def get_register_maximum_int(self):
        """Get the register maximum as an integer"""
        return self.evaluate_min_max_constants(self.maximum)

    def get_register_read_permissions_cdef(self):
        """Get the register read permissions as an C define"""
        return self.reg_perm_read[self.read_perms]["cdef"]

    def get_register_read_permissions_latex(self):
        """Get the register read permissions for use in latex documentation"""
        return self.reg_perm_read[self.read_perms]["latex"]

    def get_register_read_permissions_int(self):
        """Get the register read permissions as an integer"""
        return self.reg_perm_read[self.read_perms]["value"]

    def get_register_write_permissions_cdef(self):
        """Get the register write permissions as an C define"""
        return self.reg_perm_write[self.write_perms]["cdef"]

    def get_register_write_permissions_latex(self):
        """Get the register write permissions for use in latex documentation"""
        return self.reg_perm_write[self.write_perms]["latex"]

    def get_register_write_permissions_int(self):
        """Get the register write permissions as an integer"""
        return self.reg_perm_write[self.write_perms]["value"]

    def get_register_all_permissions_int(self):
        """Get the register combined reand and write permissions as an integer"""
        return self.get_register_read_permissions_int() | self.get_register_write_permissions_int()

    def get_register_minch_python(self):
        """Get the register minimum allowable character"""
        return self.min_ch.strip("'")

    def get_register_maxch_python(self):
        """Get the register maximum allowable character"""
        return self.max_ch.strip("'")

    def register_read_emulated_value(self):
        return self.emulation_value

    def register_write_emulated_value(self, value):
        """Write the emulated value, after checking limits"""
        try:
            minimum = self.get_register_minimum_int()
        except:
            minimum = None
        try:
            maximum = self.get_register_maximum_int()
        except:
            maximum = None

        # Do checks
        if (self.is_integer_type()):
            if (minimum is not None and value < minimum):
                raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=rc_error.PyRinCmdError.RC_WRITE_ERR_RANGE_LOW)
            if (maximum is not None and value > maximum):
                raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=rc_error.PyRinCmdError.RC_WRITE_ERR_RANGE_HIGH)
        elif (self.type == 'string' or self.type == 'binbuffer'):
            if (minimum is not None and len(value) < minimum):
                raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=rc_error.PyRinCmdError.RC_WRITE_ERR_LENGTH_LOW)
            if (maximum is not None and len(value) > maximum):
                raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=rc_error.PyRinCmdError.RC_WRITE_ERR_LENGTH_HIGH)
            minch = ord(self.get_register_minch_python())
            maxch = ord(self.get_register_maxch_python())
            for ch in value:
                if (ord(ch) < minch):
                    raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=rc_error.PyRinCmdError.RC_WRITE_ERR_RANGE_LOW)
                if (ord(ch) > maxch):
                    raise rc_error.PyRinCmdException(major=rc_error.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=rc_error.PyRinCmdError.RC_WRITE_ERR_RANGE_HIGH)
        self.emulation_value = value

    def register_execute(self, value):
        pass

    def get_register_c_variable_declaration(self, extern=False):
        """Convert the register into a C variable declaration.
           Used for testing.
           Example: int8 -> int8_t variable = 0;
        """
        prefix = "extern " if extern else ""
        ctype = self.reg_types[self.type]["ctype"]
        vname = self.value.strip("&()")
        if (self.is_integer_type()):
            postfix = "" if extern else " = 0"
            vdef = "{}{} {}{};".format(prefix, ctype, vname, postfix)
        elif (self.type == "string"):
            postfix = "" if extern else ' = ""'
            vlen = self.maximum
            vdef = "{}{} {}[{}]{};".format(prefix, ctype, vname, vlen, postfix)
        elif (self.type == "binbuffer"):
            postfix = "" if extern else ' = ""'
            vlen = self.maximum
            vdef = "{}{} {}[{}]{};".format(prefix, ctype, vname, vlen, postfix)
            lenpf = "" if extern else ' = 0'
            length = self.length.strip("&()");
            vdef += "\n{}uint32_t {}{};".format(prefix, length, lenpf)
        elif (self.type == "weight"):
            vdef = ""
        return vdef


def PyRinCmdRegisterListImport(xmltree, constants):
    """Create a list of registers by importing a register XML file
       xml can be filename, file object or existing XML tree"""
    strtypes = (str, unicode) if sys.version_info < (3, 0, 0) else (bytes, str)
    if (isinstance(xmltree, strtypes) or hasattr(xmltree, 'read')):
        xmltree = xml.etree.ElementTree.parse(xmltree)

    root = xmltree.getroot()

    registers = []

    for child in root:
        if child.tag == "registers":
            for reg in child:
                newreg = PyRinCmdRegister(xmlelt=reg, constants=constants)
                registers.append(newreg)
    return registers


class TestPyRinCmdRegister(unittest.TestCase):
    """Test the PyRinCmdRegister class"""

    TEST_INT8_MIN = 10
    TEST_INT8_MAX = 20

    def test_constants(self):
        """Test evaluating constants"""

        params = { "addr": 1, "name": "test", "description": "test",
                   "type": "uint8",
                   "min": "TEST_INT8_MIN", "max": "TEST_INT8_MAX",
                   "value": "test_var",
                   "permread": "all", "permwrite": "all" }
        reg = PyRinCmdRegister(params=params, constants=TestPyRinCmdRegister)

        value = reg.evaluate_min_max_constants(expression="TEST_INT8_MAX")
        self.assertEqual(value, self.TEST_INT8_MAX)

        value = reg.get_register_minimum_int()
        self.assertEqual(value, self.TEST_INT8_MIN)

        value = reg.get_register_maximum_int()
        self.assertEqual(value, self.TEST_INT8_MAX)

    def test_regtypes(self):
        """Test register types"""
        elem = PyRinCmdRegister.get_entry_for_regtype(regtype=PyRinCmdRegister.RC_REG_TYPE_INT8)
        self.assertEqual(elem['cdef'], 'RC_REG_TYPE_INT8')
        elem = PyRinCmdRegister.get_entry_for_regtype(regtype='RC_REG_TYPE_UINT32')
        self.assertEqual(elem['cdef'], 'RC_REG_TYPE_UINT32')
        elem = PyRinCmdRegister.get_entry_for_regtype(regtype='int16')
        self.assertEqual(elem['cdef'], 'RC_REG_TYPE_INT16')

        pytype = PyRinCmdRegister.get_python_type_for_regtype(regtype='int16')
        self.assertIs(pytype, int)
        pytype = PyRinCmdRegister.get_python_type_for_regtype(regtype='RC_REG_TYPE_STRING')
        self.assertIs(pytype, str)

    def test_nicename(self):
        """Test generating a nice name for the register that won't annoy C or PYTHON"""
        testname = "test\\register/name:with()bad.characters_and_underscore"
        params = { "addr": 1, "name": testname, "description": "test",
                   "type": "uint8",
                   "min": "TEST_INT8_MIN", "max": "TEST_INT8_MAX",
                   "value": "test_var",
                   "permread": "all", "permwrite": "all" }
        reg = PyRinCmdRegister(params=params, constants=TestPyRinCmdRegister)
        nicename = reg.get_nice_name()
        self.assertEqual("TESTREGISTERNAMEWITHBADCHARACTERS_AND_UNDERSCORE", nicename)

if __name__ == '__main__':
    unittest.main()
