'''
Created on Aug 28, 2017

@author: ian
'''
import os
import socket
import select
import threading
import logging
import argparse
import stat
import errno
import rc_register
import rc_message
import rc_error
import rc_frame

class PyRinCmdServer(object):
    """Rinstrum rincmd server"""

    def __init__(self, tcpbindaddr='0.0.0.0', tcpbindport=2222, udpbindaddr='0.0.0.0', udpbindport=2222, udsbindfile=None, register_list=None, rc_addr=0x01, loglevel=logging.WARNING):
        '''
        Constructor
        '''
        self.tcpbindaddr = tcpbindaddr
        self.tcpbindport = tcpbindport
        self.udpbindaddr = udpbindaddr
        self.udpbindport = udpbindport
        self.udsbindfile = udsbindfile
        self.register_list = register_list
        self.rc_addr = rc_addr
        self.tcp_listen_sock = None
        self.udp_sock = None
        self.uds_listen_sock = None
        self.data_socks = []
        self.quit = False
        self.framers = [ rc_frame.PyRinCmdFrameBasic(), rc_frame.PyRinCmdFrameCrc() ]
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(loglevel)

    def construct_reply(self, rcc, is_error, data):
        """Construct a rincomms reply message to match rcc"""
        addr = self.rc_addr & rc_message.PyRinCmdMessage.ADDR_MASK
        addr |= rc_message.PyRinCmdMessage.ADDR_FLAG_RESPONSE
        if (is_error):
            addr |= rc_message.PyRinCmdMessage.ADDR_FLAG_ERROR
        rcr = rc_message.PyRinCmdMessage(address=addr, command=rcc.command, register=rcc.register, channel=rcc.channel, data=data)
        return rcr

    def get_register(self, regid):
        """Lookup register in register list"""
        register = filter(lambda x: x.register_id == regid, self.register_list)
        return register[0] if (len(register) > 0) else None

    def reg_fail(self, rcc, major, minor):
        """Construct a failure response"""
        err = rc_error.PyRinCmdError(major=major, minor=minor).get_error()
        return self.construct_reply(rcc=rcc, is_error=True, data=err)

    def reg_read(self, rcc, register, regfunc, errmaj):
        """Read register something (type, value, min, max)"""
        value = regfunc(register)
        if (value is None):
            err = rc_error.PyRinCmdError(major=errmaj, minor=rc_error.PyRinCmdError.RC_READ_ERR_REGNULL)
            self.logger.warning("No value available for register {:04X}, returning {}".format(rcc.register), err)
            rcr = self.construct_reply(rcc=rcc, is_error=True, data=err.get_error())
        else:
            self.logger.debug("Retrived value {} for register {:04X}".format(value, rcc.register))
            rcr = self.construct_reply(rcc=rcc, is_error=False, data=value)
        return rcr

    def set_msg_datatype_from_register(self, register, rcc):
        """For writes we need to convert the PyRinCmdMessage data to the appropriate type for the register"""
        dtype = register.get_register_type_python()
        rcc.convert_data_to_type(datatype=dtype)

    def reg_write(self, rcc, register, regfunc, errmaj):
        """Write register (dec or hexadec)"""

        # Convert data to appropriate data type
        try:
            self.set_msg_datatype_from_register(register=register, rcc=rcc)

            # Write the register
            try:
                regfunc(register, rcc.data)
                rcr = self.construct_reply(rcc=rcc, is_error=False, data=0)
            except rc_error.PyRinCmdException as e:
                errdata = e.pyrincmd_error.get_error()
                rcr = self.construct_reply(rcc=rcc, is_error=True, data=errdata)
        except:
            err = rc_error.PyRinCmdError(major=errmaj, minor=rc_error.PyRinCmdError.RC_READ_ERR_BADDATA).get_error()
            rcr = self.construct_reply(rcc=rcc, is_error=True, data=err)
        return rcr

    dispatch_read_type = {    'cmd':     rc_message.PyRinCmdMessage.CMD_READ_TYPE,
                              'func':    reg_read,
                              'regfunc': rc_register.PyRinCmdRegister.get_register_type_int,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_READ }
    dispatch_read_min = {     'cmd':     rc_message.PyRinCmdMessage.CMD_READ_RANGE_MIN,
                              'func':    reg_read,
                              'regfunc': rc_register.PyRinCmdRegister.get_register_minimum_int,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_READ }
    dispatch_read_max = {     'cmd':     rc_message.PyRinCmdMessage.CMD_READ_RANGE_MAX,
                              'func':    reg_read,
                              'regfunc': rc_register.PyRinCmdRegister.get_register_maximum_int,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_READ }
    dispatch_read_literal = { 'cmd':     rc_message.PyRinCmdMessage.CMD_READ_LITERAL,
                              'func':    reg_read,
                              'regfunc': rc_register.PyRinCmdRegister.register_read_emulated_value,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_READ }
    dispatch_read_perms = {   'cmd':     rc_message.PyRinCmdMessage.CMD_READ_PERMISSION,
                              'func':    reg_read,
                              'regfunc': rc_register.PyRinCmdRegister.get_register_all_permissions_int,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_READ }
    dispatch_execute = {      'cmd':     rc_message.PyRinCmdMessage.CMD_EXECUTE,
                              'func':    None, # TODO execute registers
                              'regfunc': None,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_EXEC }
    dispatch_read_hex = {     'cmd':     rc_message.PyRinCmdMessage.CMD_READ_HEXADEC,
                              'func':    reg_read,
                              'regfunc': rc_register.PyRinCmdRegister.register_read_emulated_value,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_READ }
    dispatch_write_hex = {    'cmd':     rc_message.PyRinCmdMessage.CMD_WRITE_HEXADEC,
                              'func':    reg_write,
                              'regfunc': rc_register.PyRinCmdRegister.register_write_emulated_value,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_WRITE }
    dispatch_read_dec = {     'cmd':     rc_message.PyRinCmdMessage.CMD_READ_DECIMAL,
                              'func':    reg_read,
                              'regfunc': rc_register.PyRinCmdRegister.register_read_emulated_value,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_READ }
    dispatch_write_dec = {    'cmd':     rc_message.PyRinCmdMessage.CMD_WRITE_DECIMAL,
                              'func':    reg_write,
                              'regfunc': rc_register.PyRinCmdRegister.register_write_emulated_value,
                              'errmaj':  rc_error.PyRinCmdError.RC_HNDL_ERR_WRITE }

    dispatch_cmds = \
    [
        dispatch_read_type, dispatch_read_min, dispatch_read_max, dispatch_read_literal, dispatch_read_perms,
        dispatch_execute, dispatch_read_hex, dispatch_write_hex, dispatch_read_dec, dispatch_write_dec
    ]

    def dispatch_request(self, request, framer, conn):
        """Dispatch a request from the socket"""

        # Convert request into PyRinCmdMessage
        try:
            rcc = rc_message.PyRinCmdParser().str_to_message(message_str=request, datatype=str)
        except rc_error.PyRinCmdException as e:
            self.logger.debug("Failed to convert request \"{}\" to PyRinCmdMessage with error {}".format(request, e.pyrincmd_error.get_pretty_string()))
            return

        # Check address fields
        if (not rcc.is_addressed_to_me(my_addr=self.rc_addr)):
            self.logger.debug("Request was not addressed to me")
            return
        if (rcc.is_response()):
            self.logger.debug("Request is response (should be request)")
            return

        # Lookup register
        register = self.get_register(regid=rcc.register)
        if (register is None):
            # TODO regnull different between read and write
            self.logger.warning("Register {:04X} is not in our register list".format(rcc.register))
            rcr = self.reg_fail(rcc=rcc, major=rc_error.PyRinCmdError.RC_HNDL_ERR_REG, minor=rc_error.PyRinCmdError.RC_READ_ERR_REGNULL)
        else:
            # Dispatch by command
            cmd = filter(lambda x : x['cmd'] == rcc.command, self.dispatch_cmds)
            if (len(cmd) > 0):
                self.logger.debug("Dispatching command {:02X}, register {:04X}".format(rcc.command, rcc.register))
                rcr = cmd[0]["func"](self, rcc=rcc, register=register, regfunc=cmd[0]["regfunc"], errmaj=cmd[0]["errmaj"])
            else:
                self.logger.warning("Command {:02X} is not supported".format(rcc.command))
                rcr = self.reg_fail(rcc=rcc, major=rc_error.PyRinCmdError.RC_HNDL_ERR_NOTIMP, minor=0)

        # Send reply to socket
        reply = rcr.to_string()
        framedreply = framer.frame_response(resp=reply)
        self.logger.info("Sending reply \"{}\" to {} {}".format(reply, conn["type"], conn["addr"]))
        if (conn["type"] != "UDP"):
            conn["sock"].sendall(framedreply)
        else:
            conn["sock"].sendto(framedreply, conn["addr"])

    def find_message(self, conn):
        """Look for a message framed by one of our framers, and handle it. Updates conn["data"] with leftover data"""
        data = conn["data"]
        while (True):
            valid_frame = False
            for framer in self.framers:
                try:
                    before, message, after = framer.deframe_command(command=data)
                    data = after
                except:
                    data = after
                if (before is not None and message is not None):
                    # Match
                    self.logger.info("Received message \"{}\" from {} {}, using framer {}".format(message, conn["type"], conn["addr"], repr(framer)))
                    self.dispatch_request(request=message, framer=framer, conn=conn)
                    valid_frame = True
                    break
            if (not valid_frame):
                break
        conn["data"] = data

    def accept_new_connection(self, listen_sock, socktype):
        """Accept a new connection on listen_sock"""
        conn, addr = listen_sock.accept()
        new_conn = { "type": socktype, "sock": conn, "addr": addr, "data": "" }
        self.data_socks.append(new_conn)
        self.logger.info("New {} connection from {}".format(socktype, addr))

    def handle_data_sock(self, data_sock):
        """Handle data on a socket"""
        try:
            data, addr = data_sock["sock"].recvfrom(2048)
            if (len(data) == 0):
                self.logger.info("Client {} closed {} connection".format(data_sock["addr"], data_sock["type"]))
                data_sock["sock"].close()
                data_sock["sock"] = None
            else:
                data_sock["data"] += data
                self.find_message(conn=data_sock)
        except socket.error:
            self.logger.info("Connection error, closing {} connection from {}".format(data_sock["type"], data_sock["addr"]))
            data_sock["sock"].close()
            data_sock["sock"] = None

    def handle_udp_sock(self):
        """Handle data on the udp socket"""
        try:
            data, addr = self.udp_sock.recvfrom(2048)
            if (len(data) == 0):
                self.logger.info("Client {} closed UDP connection".format(addr))
                self.udp_sock.close()
                self.udp_sock = None
            else:
                udp_conn = { "type": "UDP", "sock": self.udp_sock, "addr": addr, "data": data }
                self.find_message(conn=udp_conn)
        except socket.error:
            self.logger.info("Connection error, closing UDP connection from {}".format(addr))
            self.udp_sock.close()
            self.udp_sock = None

    def wait_data(self):
        """Wait for data/new connections and dispatch"""
        if self.tcpbindaddr is not None:
            self.logger.info("PyRinCmdServer listening on TCP {}:{}".format(self.tcpbindaddr, self.tcpbindport))
        if self.udpbindaddr is not None:
            self.logger.info("PyRinCmdServer listening on UDP {}:{}".format(self.udpbindaddr, self.udpbindport))
        if self.udsbindfile is not None:
            self.logger.info("PyRinCmdServer listening on UDS {}".format(self.udsbindfile))

        while (not self.quit):
            # Which sockets do we want to select on
            read_socks = filter(lambda x: x is not None, [ self.tcp_listen_sock, self.udp_sock, self.uds_listen_sock ])
            for s in self.data_socks:
                read_socks.append(s["sock"])

            # Select for data/new connections
            ready_socks, _, _ = select.select(read_socks, [], [], 1.0)

            # Accept new connections
            if (self.tcp_listen_sock in ready_socks):
                self.accept_new_connection(listen_sock=self.tcp_listen_sock, socktype="TCP")
            if (self.uds_listen_sock in ready_socks):
                self.accept_new_connection(listen_sock=self.uds_listen_sock, socktype="UDS")

            # List of ready data sockets
            ready_data_socks = [ ds for ds in self.data_socks if ds["sock"] in ready_socks ]

            # Handle data on established sockets
            for ds in ready_data_socks:
                self.handle_data_sock(data_sock=ds)
            if self.udp_sock in ready_socks:
                self.handle_udp_sock()

            # Purge closed sockets
            pruned_data_socks = [ ds for ds in self.data_socks if ds["sock"] is not None ]
            self.data_socks = pruned_data_socks

        self.logger.info("PyRinCmdServer shutting down")

    def open_listen_sockets(self):
        """Open the listening sockets (TCP, UDP, UDS)"""
        if self.tcpbindaddr is not None:
            self.tcp_listen_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.tcp_listen_sock.bind((self.tcpbindaddr, self.tcpbindport))
            self.tcp_listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.tcp_listen_sock.listen(10)

        if self.udpbindaddr is not None:
            self.udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.udp_sock.bind((self.udpbindaddr, self.udpbindport))
            self.udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        if self.udsbindfile is not None:
            try:
                st = os.stat(self.udsbindfile)
            except OSError as err:
                if err.errno != errno.ENOENT:
                    raise
            else:
                if stat.S_ISSOCK(st.st_mode):
                    os.unlink(self.udsbindfile)
                else:
                    raise ValueError("File {} exists and is not a socket".format(self.udsbindfile))
            self.uds_listen_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.uds_listen_sock.bind(self.udsbindfile)
            self.uds_listen_sock.listen(10)

    def run(self):
        """Run the pyrincmd server"""
        self.open_listen_sockets()
        try:
            self.wait_data()
        finally:
            if self.tcp_listen_sock is not None:
                self.tcp_listen_sock.close()
            if self.udp_sock is not None:
                self.udp_sock.close()
            if self.uds_listen_sock is not None:
                self.uds_listen_sock.close()
                os.unlink(self.udsbindfile)
            for s in self.data_socks:
                s["sock"].close()

    def run_thread(self):
        """Run the pyrincmd server in a thread"""
        self.thread = threading.Thread(target=self.run, args=())
        self.thread.start()

    def stop_thread(self):
        """Stop the thread running if run_thread() has been called"""
        self.quit = True
        self.thread.join()

def main(argv=None):
    """Simple command line pyrincmd server"""

    # Parse command line options
    parser = argparse.ArgumentParser(description='Run the pyrincmd server')
    parser.add_argument('register_xml', action='store',
                        help='The registers XML definition file')
    parser.add_argument('--tcpbindaddr', action='store', default='localhost',
                        help='The local IP address to bind the TCP listen socket to. Default localhost')
    parser.add_argument('--tcpbindport', action='store', type=int, default=2222,
                        help='The local TCP port to bind to. Default 2222')
    parser.add_argument('--udpbindaddr', action='store', default='localhost',
                        help='The local IP address to bind the UDP listen socket to. Default localhost')
    parser.add_argument('--udpbindport', action='store', type=int, default=2222,
                        help='The local UDP port to bind to. Default 2222')
    parser.add_argument('--udsbindfile', action='store', default=None,
                        help='The local UDS file to bind to. Default is none.')
    parser.add_argument('--log-level', action='store', default=logging.ERROR,
                        help='Debug log level, string or int (FATAL(50), ERROR(40), WARNING(30), INFO(20), DEBUG(10), NOTSET(0)')
    parser.add_argument('--log-file', action='store', default=None,
                        help='Store debug info into file')
    args = parser.parse_args(args=argv)

    logging.basicConfig(level=args.log_level, filename=args.log_file)

    reglist = rc_register.PyRinCmdRegisterListImport(xmltree=args.register_xml, constants=None)
    server = PyRinCmdServer(tcpbindaddr=args.tcpbindaddr, tcpbindport=args.tcpbindport,
                            udpbindaddr=args.udpbindaddr, udpbindport=args.udpbindport,
                            udsbindfile=args.udsbindfile,
                            register_list=reglist, loglevel=args.log_level)
    args = parser.parse_args(args=argv)

    server.run()

if (__name__ == '__main__'):
    main()
