'''
Created on Sep 12, 2017

@author: ian
'''
import unittest
import logging
import random
import pyrincmd
import regconst


server = None
client_comms_logfile = None

class RegisterNamespace(object):
    pass


def start_server(loglevel):
    global server

    reglist = pyrincmd.PyRinCmdRegisterListImport(xmltree=regconst.xmlfilename, constants=regconst.TestRegistersConstants)
    server = pyrincmd.PyRinCmdServer(bindaddr='localhost', register_list=reglist, loglevel=loglevel)
    server.run_thread()


def stop_server():
    global server

    if (server is not None):
        server.stop_thread()


class TestPyRinCmdClient(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestPyRinCmdClient, self).__init__(*args, **kwargs)
        regxlatpy = pyrincmd.PyRinCmdRegTranslate_PY(xmltree=regconst.xmlfilename)
        self.regns = RegisterNamespace()
        regxlatpy.xml_register_ids_to_namespace(ns=self.regns)

    def setUp(self):
        self.pyrincmdclient = pyrincmd.PyRinCmdClient(tcphost='localhost', logfile=client_comms_logfile)
        self.pyrincmdclient.connect()

    def tearDown(self):
        self.pyrincmdclient.disconnect()

    def test_simple_read_write_register(self):
        """Test a simple read and write to a register"""

        reg = filter(lambda x: x.register_id == self.regns.RINREG_TESTTABLE_TEST_INT8, server.register_list)
        reg[0].register_write_emulated_value(12)

        reply = self.pyrincmdclient.read_register(command=pyrincmd.PyRinCmdMessage.CMD_READ_DECIMAL, register=self.regns.RINREG_TESTTABLE_TEST_INT8, datatype=int)
        self.assertEqual(first=reply.data, second=12)

        reply = self.pyrincmdclient.read_register(command=pyrincmd.PyRinCmdMessage.CMD_READ_HEXADEC, register=self.regns.RINREG_TESTTABLE_TEST_INT8, datatype=lambda x: int(x, 16))
        self.assertEqual(first=reply.data, second=12)

        reply = self.pyrincmdclient.write_register(command=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC, register=self.regns.RINREG_TESTTABLE_TEST_INT8, data=61)
        self.assertEqual(first=reg[0].register_read_emulated_value(), second=61)

    def test_negative_read_write_register(self):
        """Test reading and writing negative numbers from/to a register"""

        testreg = self.regns.RINREG_TESTTABLE_TEST_INT32

        reg = filter(lambda x: x.register_id == testreg, server.register_list)
        reg[0].register_write_emulated_value(-1234)

        reply = self.pyrincmdclient.read_register(command=pyrincmd.PyRinCmdMessage.CMD_READ_DECIMAL, register=testreg, datatype=int)
        self.assertEqual(first=reply.data, second=-1234)

        reply = self.pyrincmdclient.read_register(command=pyrincmd.PyRinCmdMessage.CMD_READ_HEXADEC, register=testreg, datatype=int)
        self.assertEqual(first=reply.data, second=-1234)

        reply = self.pyrincmdclient.write_register(command=pyrincmd.PyRinCmdMessage.CMD_WRITE_DECIMAL, register=testreg, data=-5678)
        self.assertEqual(first=reg[0].register_read_emulated_value(), second=-5678)

        reply = self.pyrincmdclient.write_register(command=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC, register=testreg, data=-3456)
        self.assertEqual(first=reg[0].register_read_emulated_value(), second=-3456)

    def test_crc_read_write_register(self):
        """Test read and write to a register using message CRCs"""

        pyrincmdclient = pyrincmd.PyRinCmdClient(tcphost='localhost', usecrc=True)
        pyrincmdclient.connect()

        reg = filter(lambda x: x.register_id == self.regns.RINREG_TESTTABLE_TEST_INT8, server.register_list)
        reg[0].register_write_emulated_value(99)

        reply = pyrincmdclient.read_register(command=pyrincmd.PyRinCmdMessage.CMD_READ_DECIMAL, register=self.regns.RINREG_TESTTABLE_TEST_INT8, datatype=int)
        self.assertEqual(first=reply.data, second=99)

        reply = pyrincmdclient.write_register(command=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC, register=self.regns.RINREG_TESTTABLE_TEST_INT8, data=100)
        self.assertEqual(first=reg[0].register_read_emulated_value(), second=100)

    def check_int_register_limits(self, regid, writecmd):
        """Test the upper and lower limits of an integer register"""

        reg = filter(lambda x: x.register_id == regid, server.register_list)
        lowerlimit = reg[0].get_register_minimum_int()
        upperlimit = reg[0].get_register_maximum_int()
        validvalue = random.randint(lowerlimit, upperlimit)

        # Set the register
        reply = self.pyrincmdclient.write_register(command=writecmd, register=regid, data=validvalue)
        self.assertTrue(not reply.is_error())
        self.assertEqual(reply.data, 0)

        expected = pyrincmd.PyRinCmdError(major=pyrincmd.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=pyrincmd.PyRinCmdError.RC_WRITE_ERR_RANGE_HIGH)
        reply = self.pyrincmdclient.write_register(command=writecmd, register=regid, data=upperlimit + 1)
        self.assertTrue(reply.is_error())
        actual = pyrincmd.PyRinCmdError(error=reply.data)
        self.assertEqual(actual, expected)

        expected = pyrincmd.PyRinCmdError(major=pyrincmd.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=pyrincmd.PyRinCmdError.RC_WRITE_ERR_RANGE_LOW)
        reply = self.pyrincmdclient.write_register(command=writecmd, register=regid, data=lowerlimit - 1)
        self.assertTrue(reply.is_error())
        actual = pyrincmd.PyRinCmdError(error=reply.data)
        self.assertEqual(actual, expected)

        reply = self.pyrincmdclient.read_register(command=pyrincmd.PyRinCmdMessage.CMD_READ_DECIMAL, register=regid, datatype=int)
        self.assertEqual(first=reply.data, second=validvalue)

    def test_write_int8_register_limits(self):
        """Test the upper and lower limits of a signed 8bit integer register"""

        self.check_int_register_limits(regid=self.regns.RINREG_TESTTABLE_TEST_INT8, writecmd=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC)
        self.check_int_register_limits(regid=self.regns.RINREG_TESTTABLE_TEST_INT8, writecmd=pyrincmd.PyRinCmdMessage.CMD_WRITE_DECIMAL)

    def test_write_uint8_register_limits(self):
        """Test the upper and lower limits of a unsigned 8bit integer register"""

        self.check_int_register_limits(regid=self.regns.RINREG_TESTTABLE_TEST_UINT8, writecmd=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC)
        self.check_int_register_limits(regid=self.regns.RINREG_TESTTABLE_TEST_UINT8, writecmd=pyrincmd.PyRinCmdMessage.CMD_WRITE_DECIMAL)

    def test_write_int32_register_limits(self):
        """Test the upper and lower limits of a signed 32bit integer register"""

        self.check_int_register_limits(regid=self.regns.RINREG_TESTTABLE_TEST_INT32_LOWLIM, writecmd=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC)
        self.check_int_register_limits(regid=self.regns.RINREG_TESTTABLE_TEST_INT32_LOWLIM, writecmd=pyrincmd.PyRinCmdMessage.CMD_WRITE_DECIMAL)

    def test_write_uint32_register_limits(self):
        """Test the upper and lower limits of a unsigned 32bit integer register"""

        self.check_int_register_limits(regid=self.regns.RINREG_TESTTABLE_TEST_UINT32, writecmd=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC)
        self.check_int_register_limits(regid=self.regns.RINREG_TESTTABLE_TEST_UINT32, writecmd=pyrincmd.PyRinCmdMessage.CMD_WRITE_DECIMAL)

    def test_write_str_register_limits(self):
        """Test the upper and lower limits of a string register"""

        test_regid = self.regns.RINREG_TESTTABLE_TEST_STRING
        reg = filter(lambda x: x.register_id == test_regid, server.register_list)

        # Set the register
        setvalue = "testing"
        expected = pyrincmd.PyRinCmdError(major=pyrincmd.PyRinCmdError.RC_HNDL_ERR_NONE, minor=pyrincmd.PyRinCmdError.RC_NONE_ERR_NONE)
        reply = self.pyrincmdclient.write_register(command=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC, register=test_regid, data=setvalue)
        self.assertTrue(not reply.is_error())
        actual = pyrincmd.PyRinCmdError(error=reply.data)
        self.assertEqual(actual, expected)

        upperlimit = reg[0].constants.TEST_STRING_LENGTH
        data = "a" * (upperlimit + 1)
        expected = pyrincmd.PyRinCmdError(major=pyrincmd.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=pyrincmd.PyRinCmdError.RC_WRITE_ERR_RANGE_HIGH)
        reply = self.pyrincmdclient.write_register(command=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC, register=test_regid, data=data)
        self.assertTrue(reply.is_error())
        actual = pyrincmd.PyRinCmdError(error=reply.data)
        self.assertEqual(actual, expected)

        lowerlimit = 5
        data = "b" * (lowerlimit - 1)
        expected = pyrincmd.PyRinCmdError(major=pyrincmd.PyRinCmdError.RC_HNDL_ERR_WRITE, minor=pyrincmd.PyRinCmdError.RC_WRITE_ERR_RANGE_LOW)
        reply = self.pyrincmdclient.write_register(command=pyrincmd.PyRinCmdMessage.CMD_WRITE_HEXADEC, register=test_regid, data=data)
        self.assertTrue(reply.is_error())
        actual = pyrincmd.PyRinCmdError(error=reply.data)
        self.assertEqual(actual, expected)

        reply = self.pyrincmdclient.read_register(command=pyrincmd.PyRinCmdMessage.CMD_READ_DECIMAL, register=test_regid, datatype=str)
        self.assertEqual(first=reply.data, second=setvalue)
