/*
 * rc_cli_async.c
 *
 *  Created on: Nov 20, 2017
 *      Author: ian
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rincmd/rc_frame.h>
#include <rincmd/rc_timespec.h>
#include <rincmd/rc_cli_parse.h>
#include <rincmd/rc_cli_gettime.h>

#include <rincmd/rc_cli_async.h>


/** Allocate a new callback */
static enum e_rc_cli_result rc_async_client_new_callback(struct s_rc_async_callback **new_cb)
{
  struct s_rc_async_callback *cb;

  if (NULL == new_cb)
    return rcr_paramfail;

  cb = calloc(1, sizeof(*cb));
  if (NULL == cb)
  {
    *new_cb = NULL;
    return rcr_allocfail;
  }

  cb->persistent_not_transient = false;
  cb->userdata = NULL;
  cb->cmd = RC_MESG_NONE;
  cb->reg = 0;
  cb->callback_fn = NULL;
  cb->hastimeout = false;
  cb->abstimeout.tv_sec = 0;
  cb->abstimeout.tv_nsec = 0;
  cb->persistent_timeout.tv_sec = 0;
  cb->persistent_timeout.tv_nsec = 0;
  cb->next = NULL;

  *new_cb = cb;
  return rcr_success;
}


/** Pre-allocate the callbacks */
static enum e_rc_cli_result rc_async_client_create_callbacks(struct s_rc_async_client *rac, size_t num_callbacks)
{
  struct s_rc_async_callback *cb;
  struct s_rc_async_callback *prev_cb;
  struct s_rc_async_callback *next_cb;
  size_t i;
  enum e_rc_cli_result result;

  rac->unused_callbacks = NULL;
  rac->new_active_callbacks = NULL;
  rac->active_callbacks = NULL;

  /* Allocate unused callbacks, and insert them into the linked list */
  for (i = 0, prev_cb = NULL; i < num_callbacks; i++, prev_cb =  cb)
  {
    result = rc_async_client_new_callback(&cb);
    if (rcr_success != result)
      goto raccc_fail;
    cb->next = prev_cb;
  }
  rac->unused_callbacks = prev_cb;

  return rcr_success;

raccc_fail:
  for (cb = rac->unused_callbacks; cb != NULL; )
  {
    next_cb = cb->next;
    free(cb);
    cb = next_cb;
  }
  return result;
}


/**
 * Create an asynchronous client
 *
 * @param rac is set to the created async client on success
 * @param cli_comms is the already existing comms interface. You are responsible for destroying this once rac has been destroyed
 * @param rc_address is the rincmd address to address messages to. RC_MESG_BROADCAST is a good choice
 * @param use_crc is a flag to indicate the use of CRC framed messages, or normal framing
 * @param num_callbacks is the total number of callbacks supported by this client (persistent and transient)
 * @param connect_fn is a callback function called when comms (socket,etc) has connected or disconnected (via rc_async_client_connect() call)
 * @param connect_userdata is a pointer to user data passed to the connect callback function
 * @param bad_message_fn is the function called when received data cannot be parsed
 * @param bad_message_userdata is a pointer to user data passed to the bad message callback function
 */
enum e_rc_cli_result rc_async_client_init(struct s_rc_async_client **rac,
                                          struct s_rc_cli_comms *cli_comms,
                                          uint8_t rc_address, bool use_crc,
                                          size_t num_callbacks,
                                          t_rc_async_connect_callback_fn connect_fn,
                                          void *connect_userdata,
                                          t_rc_async_bad_message_callback_fn bad_message_fn,
                                          void *bad_message_userdata)
{
  struct s_rc_async_client *irac = NULL;
  struct s_rc_frame *frame = NULL;
  enum e_rc_cli_result result;

  if (NULL == rac || NULL == cli_comms)
    return rcr_paramfail;

  /* Allocate the async client structure */
  irac = (struct s_rc_async_client *) calloc(1, sizeof(*irac));
  if (NULL == irac)
  {
    result = rcr_allocfail;
    goto raci_fail;
  }

  /* Setup comms */
  irac->cli_comms = cli_comms;

  /* Setup framing */
  result = (use_crc) ? rc_frame_init_crc(&frame) : rc_frame_init_normal(&frame);
  if (rcr_success != result)
    goto raci_fail;

  /* Setup parser */
  result = rc_cli_parse_init(&irac->cli_parser, frame, rc_address, false, 0);
  if (rcr_success != result)
    goto raci_fail;

  /* Setup callbacks */
  result = rc_async_client_create_callbacks(irac, num_callbacks);
  if (rcr_success != result)
    goto raci_fail;

  /* Create buffers */
  irac->inbuffer = calloc(RC_CLI_MSG_MAXLEN, sizeof(char));
  if (NULL == irac->inbuffer)
  {
    result = rcr_allocfail;
    goto raci_fail;
  }
  irac->inbufferlen = RC_CLI_MSG_MAXLEN;
  irac->outbuffer = calloc(RC_CLI_MSG_MAXLEN, sizeof(char));
  if (NULL == irac->outbuffer)
  {
    result = rcr_allocfail;
    goto raci_fail;
  }
  irac->outbufferlen = RC_CLI_MSG_MAXLEN;

  /* Keep the connect/disconnect callback */
  irac->connect_callback_fn = connect_fn;
  irac->connect_callback_userdata = connect_userdata;
  irac->connect_hastimeout = false;
  irac->connect_callback_abstimeout.tv_sec = 0;
  irac->connect_callback_abstimeout.tv_nsec = 0;

  /* Keep the bad message callback */
  irac->bad_message_callback_fn = bad_message_fn;
  irac->bad_message_callback_userdata = bad_message_userdata;

  /* Success */
  *rac = irac;
  return rcr_success;

raci_fail:
  (void) rc_async_client_destroy(irac);
  irac = NULL;
  return result;
}


/** Destroy an asynchronous client */
enum e_rc_cli_result rc_async_client_destroy(struct s_rc_async_client *rac)
{
  struct s_rc_async_callback *cb;
  struct s_rc_async_callback *next_cb;

  if (NULL != rac)
  {
    if (NULL != rac->cli_comms)
    {
      /* Up to the caller to destroy this */
      /* rc_cli_comms_destroy(rac->cli_comms); */
      rac->cli_comms = NULL;
    }
    if (NULL != rac->cli_parser)
    {
      if (NULL != rac->cli_parser->framing)
      {
        rc_frame_destroy(rac->cli_parser->framing);
        rac->cli_parser->framing = NULL;
      }
      rc_cli_parse_destroy(rac->cli_parser);
      rac->cli_parser = NULL;
    }
    for (cb = rac->unused_callbacks; cb != NULL; )
    {
      next_cb = cb->next;
      free(cb);
      cb = next_cb;
    }
    rac->unused_callbacks = NULL;
    for (cb = rac->new_active_callbacks; cb != NULL; )
    {
      next_cb = cb->next;
      free(cb);
      cb = next_cb;
    }
    rac->new_active_callbacks = NULL;
    for (cb = rac->active_callbacks; cb != NULL; )
    {
      next_cb = cb->next;
      free(cb);
      cb = next_cb;
    }
    rac->active_callbacks = NULL;
    if (NULL != rac->inbuffer)
    {
      free(rac->inbuffer);
      rac->inbuffer = NULL;
    }
    if (NULL != rac->outbuffer)
    {
      free(rac->outbuffer);
      rac->outbuffer = NULL;
    }
    free(rac);
    rac = NULL;
  }
  return rcr_success;
}


/**
 * Inititate the async connection to the server
 *
 * @param rac is the rincmd async client
 * @param connect_timeout is the timeout to use when connecting. NULL for indefinite timeout
 */
enum e_rc_cli_result rc_async_client_connect(struct s_rc_async_client *rac, const struct timespec *timeout)
{
  struct timespec now;

  if (NULL != timeout)
  {
    /* Set absolute timeout for connect */
    (void) rc_cli_external_gettime(&now);
    rc_timespec_add(&rac->connect_callback_abstimeout, &now, timeout);
    rac->connect_hastimeout = true;
  }
  else
  {
    rac->connect_hastimeout = false;
    rac->connect_callback_abstimeout.tv_sec = 0;
    rac->connect_callback_abstimeout.tv_nsec = 0;
  }

  /* Initiate the connect */
  return rc_cli_comms_connect_async(rac->cli_comms, rac->connect_callback_fn, rac->connect_callback_userdata);
}


/** Disconnect from the server */
enum e_rc_cli_result rc_async_client_disconnect(struct s_rc_async_client *rac)
{
  return rc_cli_comms_disconnect(rac->cli_comms);
}


/** Perform a callback */
static enum e_rc_cli_result rc_async_client_do_callback(struct s_rc_async_callback *cb, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct timespec now;

  /* Yes, call the callback if set */
  if (NULL != cb->callback_fn)
    cb->callback_fn(cb->userdata, msg, result);

  /* Set new timeout for persistent callbacks */
  if (cb->persistent_not_transient && cb->hastimeout)
  {
    (void) rc_cli_external_gettime(&now);
    rc_timespec_add(&cb->abstimeout, &now, &cb->persistent_timeout);
  }

  return rcr_success;
}


/**
 * Get the read and connect file descriptors
 *
 * Can be used to determine if the async client is connecting, disconnected, or connected
 *
 * @param rac is the rincmd async client
 * @param readfd is set to the read file descriptor if the comms is connected, < 0 otherwise. Used in the readfds param to select(). Not NULL.
 * @param connectfd is set to the connect file descriptor if the comms is connecting, < 0 otherwise. Used in the writefds param to select(). Not NULL.
 */
enum e_rc_cli_result rc_async_client_get_fd(const struct s_rc_async_client *rac, int *readfd, int *connectfd)
{
  return rc_cli_comms_getfds(rac->cli_comms, readfd, connectfd);
}


/** Create a rincmd message structure and pass to the callback with the result rcr_timeout */
static enum e_rc_cli_result rc_async_client_do_timeout_callback(const struct s_rc_async_client *rac, struct s_rc_async_callback *cb)
{
  struct s_rc_Message timeout_resp;

  (void) rc_cli_message_fillout(rac->cli_parser, &timeout_resp, false, cb->cmd, cb->reg, NULL, 0);
  timeout_resp.addr |= RC_MESG_ERROR;
  return rc_async_client_do_callback(cb, &timeout_resp, rcr_timeout);
}


/** Move any new active callbacks into the active callback list */
void rc_async_client_new_active_to_active(struct s_rc_async_client *rac)
{
  struct s_rc_async_callback *first_active_cb;
  struct s_rc_async_callback *last_new_cb;
  struct s_rc_async_callback *new_cb;

  if (NULL != rac->new_active_callbacks)
  {
    /* Only if there are new active callbacks */
    if (NULL == rac->active_callbacks)
    {
      /* No already active callbacks, so just move the new active callbacks to the active callbacks */
      rac->active_callbacks = rac->new_active_callbacks;
      rac->new_active_callbacks = NULL;
    }
    else
    {
      /* Find the last new active callback */
      for (new_cb = rac->new_active_callbacks, last_new_cb = NULL; new_cb != NULL; last_new_cb = new_cb, new_cb = new_cb->next)
        ;

      /* Put all new active callbacks at the start of the active callbacks list */
      first_active_cb = rac->active_callbacks;
      rac->active_callbacks = rac->new_active_callbacks;
      last_new_cb->next = first_active_cb;
      rac->new_active_callbacks= NULL;
    }
  }
}


/**
 * Call prior to select() to handle any timeouts that may have occurred and
 * return the read and connect file descriptors and the next timeout if there is one
 *
 * @param rac is the rincmd async client
 * @param readfd is set to the read file descriptor if the comms is connected, < 0 otherwise. Used in the readfds param to select(). Not NULL.
 * @param connectfd is set to the connect file descriptor if the comms is connecting, < 0 otherwise. Used in the writefds param to select(). Not NULL.
 * @param hastimeout is set if the there is a timeout. Not NULL
 * @param timeout is set to the timeout period. Not NULL
 */
enum e_rc_cli_result rc_async_client_pre_select(struct s_rc_async_client *rac,
                                                int *readfd, int *connectfd,
                                                bool *hastimeout, struct timespec *timeout)
{
  struct s_rc_async_callback *cb;
  struct s_rc_async_callback *prev_cb;
  struct timespec now;
  struct timespec shortest;
  bool first_timeout;
  struct timespec left;
  const struct timespec tvzero = { .tv_sec=0, .tv_nsec=0 };
  enum e_rc_cli_result result;
  bool remove_cb;

  /* Check params */
  if (NULL == rac || NULL == readfd || NULL == connectfd || NULL == hastimeout || NULL == timeout)
  {
    return rcr_paramfail;
  }

  /* Current time for timeouts */
  (void) rc_cli_external_gettime(&now);
  shortest.tv_sec = 0;
  shortest.tv_nsec = 0;
  first_timeout = true;

  /* Get file descriptors */
  result = rc_cli_comms_getfds(rac->cli_comms, readfd, connectfd);
  if (rcr_success != result)
    return result;

  /* Work out next timeout */
  if (*readfd >= 0)
  {
    /* Move any new active callbacks into the active callbacks list */
    rc_async_client_new_active_to_active(rac);

    /* Connected, so work out timeouts for register callbacks */
    for (cb = rac->active_callbacks, prev_cb = NULL; cb != NULL; )
    {
      remove_cb = false;
      if (cb->hastimeout)
      {
        /* Work out how much time is left for this callback */
        rc_timespec_subtract(&left, &cb->abstimeout, &now);

        /* Has timed out already? */
        if (rc_timespec_compare(&left, &tvzero) <= 0)
        {
          /* Yes, call the callback if set */
          (void) rc_async_client_do_timeout_callback(rac, cb);

          /* Remove callback if transient */
          remove_cb = !cb->persistent_not_transient;
        }
        else
        {
          if (first_timeout || rc_timespec_compare(&left, &shortest) < 0)
          {
            /* This callback has the least time to go */
            shortest = left;
            first_timeout = false;
          }
        }
      }

      if (remove_cb)
      {
        /* Remove from active list */
        if (NULL != prev_cb)
          prev_cb->next = cb->next;
        else
          rac->active_callbacks = cb->next;

        /* Add to unused list */
        cb->next = rac->unused_callbacks;
        rac->unused_callbacks = cb;

        /* Update loop pointers */
        cb = (NULL != prev_cb) ? prev_cb->next : rac->active_callbacks;
      }
      else
      {
        /* Just move onto next element */
        prev_cb = cb;
        cb = cb->next;
      }

    } /* for */
  } /* if */

  if (*connectfd >= 0 && rac->connect_hastimeout)
  {
    /* Connecting, so get connect timeout */

    /* Work out how much time is left for the connect timeout */
    rc_timespec_subtract(&left, &rac->connect_callback_abstimeout, &now);

    /* Has timed out already? */
    if (rc_timespec_compare(&left, &tvzero) <= 0)
    {
      /* Yes, handle connecting timed out */
      (void) rc_cli_comms_complete_connect_async(rac->cli_comms, false);
    }
    else
    {
      if (first_timeout || rc_timespec_compare(&left, &shortest) < 0)
      {
        /* This callback has the least time to go */
        shortest = left;
        first_timeout = false;
      }
    }
  }

  *hastimeout = !first_timeout;
  *timeout = shortest;
  return result;
}


/** Ignore any errors that might result from parsing a message, and call the bad message callback */
static enum e_rc_cli_result rc_async_client_ignore_msg_check_errors(struct s_rc_async_client *rac,
                                                                    const char *msgstr, size_t msgstrlen,
                                                                    const struct s_rc_Message *msg,
                                                                    enum e_rc_cli_result result)
{
  enum e_rc_cli_result ignored;

  switch (result)
  {
    case rcr_framingfail:
    case rcr_parsefail:
    case rcr_reply_badaddress:
    case rcr_reply_notresponse:
    case rcr_reply_badcommand:
    case rcr_reply_badregister:
    case rcr_reply_error:
      if (NULL != rac->bad_message_callback_fn)
        rac->bad_message_callback_fn(rac->bad_message_callback_userdata, msgstr, msgstrlen, msg, result);

      ignored = rcr_success;
      break;
    default:
      ignored = result;
      break;
  }
  return ignored;
}


/** Handle a framed message for the asynchronous client */
static enum e_rc_cli_result rc_async_client_handle_framed_message(struct s_rc_async_client *rac, size_t msg_start, size_t msg_length)
{
  struct s_rc_async_callback *cb;
  struct s_rc_async_callback *prev_cb;
  uint32_t returncode;
  bool remove_cb;
  struct s_rc_Message resp;
  enum e_rc_cli_result result;

  /* Valid message or error, log it first */
  rc_cli_comms_log(rac->cli_comms, true, &rac->inbuffer[msg_start], msg_length);

  /* Parse a response from the host device into msg */
  result = rc_cli_parse_response(rac->cli_parser, &resp, &rac->inbuffer[msg_start]);
  if (rcr_success != result)
    return rc_async_client_ignore_msg_check_errors(rac, &rac->inbuffer[msg_start], msg_length, NULL, result);

  /* Move any new active callbacks into the active callbacks list */
  rc_async_client_new_active_to_active(rac);

  /* Check address field in the response header, but not command and register */
  result = rc_cli_check_response_header_address(rac->cli_parser, &resp);
  if (rcr_success != result)
    return rc_async_client_ignore_msg_check_errors(rac, &rac->inbuffer[msg_start], msg_length, &resp, result);

  /* Look for a callback that matches command and register id, and has not just been added by one of the callbacks */
  for (cb = rac->active_callbacks, prev_cb = NULL; cb != NULL; )
  {
    remove_cb = false;
    if (rc_cli_check_response_header_cmd_reg(&resp, cb->cmd, cb->reg) == rcr_success)
    {
      result = rc_cli_check_response_header_returncode(&resp, &returncode);

      /* Call the callback */
      (void) rc_async_client_do_callback(cb, &resp, result);

      /* Remove the callback if transient */
      remove_cb = !cb->persistent_not_transient;
    }

    if (remove_cb)
    {
      /* Remove from active list */
      if (NULL != prev_cb)
        prev_cb->next = cb->next;
      else
        rac->active_callbacks = cb->next;

      /* Add to unused list */
      cb->next = rac->unused_callbacks;
      rac->unused_callbacks = cb;

      /* Update loop pointers */
      cb = (NULL != prev_cb) ? prev_cb->next : rac->active_callbacks;
    }
    else
    {
      /* Just move onto next element */
      prev_cb = cb;
      cb = cb->next;
    }
  } /* for active callbacks */
  return rcr_success;
}


/** Handle traffic for the asynchronous client that is connected */
static enum e_rc_cli_result rc_async_client_handle_connected(struct s_rc_async_client *rac)
{
  size_t msgend;
  size_t handlepos;
  ssize_t bytes_read;
  enum e_rc_cli_result result;
  bool quit;

  /* Read any data available */
  bytes_read = rac->inbufferlen - rac->inbufferposn;
  result = rc_cli_comms_read(rac->cli_comms, &rac->inbuffer[rac->inbufferposn], &bytes_read);
  if (rcr_success != result)
    return result;

  /* Update buffer position and null terminate */
  rac->inbufferposn += bytes_read;
  rac->inbuffer[rac->inbufferposn] = '\0';

  /* Try to handle the data we received */
  for (handlepos = 0, quit = false; !quit; handlepos += msgend)
  {
    msgend = 0;
    /* Find a framed message */
    result = rc_frame_response_check(rac->cli_parser->framing, false, &rac->inbuffer[handlepos], NULL, NULL, &msgend);
    switch (result)
    {
      case rcr_framingfail:
        /* No full message this time around, so stop processing, but this is not an error */
        result = rcr_success;
        quit = true;
        break;
      case rcr_success:
        /*
         * Got a framed message, now check and handle the contents.
         * rc_async_client_handle_framed_message() will always return rcr_success for parsing/bad messages,
         * so we can just keep looking for messages
         * If something really bad happens, the result will be != rcr_success and we should stop
         */
        result = rc_async_client_handle_framed_message(rac, handlepos, msgend);
        quit = (rcr_success != result);
        break;
      default:
        /* All other errors are bad, and we should stop now */
        quit = true;
        break;
    }
  } /* for */

  /* Now move up any unhandled data */
  if (handlepos > 0)
  {
    memmove(rac->inbuffer, &rac->inbuffer[handlepos], rac->inbufferposn - handlepos);
    rac->inbufferposn -= handlepos;
  }
  return result;
}


/**
 * Handle connection, timeout and traffic for the asynchronous client
 *
 * @param rac is the async connection
 * @param read_avail set true if your select() call has indicated there is data to be read on the read file descriptor
 * @param connect set true if your select() call has indicated there is a connect to be handled
 */
enum e_rc_cli_result rc_async_client_post_select(struct s_rc_async_client *rac, bool read_avail, bool connect)
{
  enum e_rc_cli_result result;

  if (connect)
  {
    /* Handle connecting */
    result = rc_cli_comms_complete_connect_async(rac->cli_comms, true);
  }
  else if (read_avail)
  {
    /* Handle data available */
    result = rc_async_client_handle_connected(rac);
  }
  else
  {
    result = rcr_paramfail;
  }

  return result;
}


/**
 * Add a callback to the list of callbacks. Do not call directly.
 *
 * @param rac is the rincmd async client
 * @param persistent_not_transient is whether this callback is persistent, or should be removed once it has completed
 * @param cmd is the rincmd command to expect a response for
 * @param regid is the rincmd register to expect a response for
 * @param callback is the function to call when a response is received or timeout. Can be NULL.
 * @param userdata is a user data pointer to pass to the callback functions. Can be NULL.
 * @param timeout is a timeout value
 */
enum e_rc_cli_result rc_async_client_add_callback(struct s_rc_async_client *rac, bool persistent_not_transient,
                                                  enum e_rc_reg_command cmd, rc_RegId_t regid,
                                                  t_rc_async_callback_fn callback, void *userdata,
                                                  const struct timespec *timeout)
{
  struct s_rc_async_callback *cb;
  struct timespec now;

  if (NULL == rac)
    return rcr_paramfail;

  if (NULL == rac->unused_callbacks)
    return rcr_listempty;

  /* Remove unused callback from the unused list */
  cb = rac->unused_callbacks;
  rac->unused_callbacks = cb->next;

  /* Fill it out */
  cb->persistent_not_transient = persistent_not_transient;
  cb->userdata = userdata;
  cb->cmd = cmd;
  cb->reg = regid;
  cb->callback_fn = callback;

  if (NULL != timeout)
  {
    cb->persistent_timeout = *timeout;
    (void) rc_cli_external_gettime(&now);
    rc_timespec_add(&cb->abstimeout, &now, timeout);
    cb->hastimeout = true;
  }
  else
  {
    cb->hastimeout = false;
  }

  /* Add it into the new active callbacks list. It will get transferred to the active_callbacks list at an appropriate time */
  cb->next = rac->new_active_callbacks;
  rac->new_active_callbacks = cb;
  return rcr_success;
}


/** Remove a persistent callback from the list of active callbacks. Do not call directly. */
enum e_rc_cli_result rc_async_client_remove_persistent_callback(struct s_rc_async_client *rac,
                                                                enum e_rc_reg_command cmd, rc_RegId_t regid)
{
  struct s_rc_async_callback *cb;
  struct s_rc_async_callback *prev_cb;

  /* First check the active list */
  for (cb = rac->active_callbacks, prev_cb = NULL; cb != NULL; prev_cb = cb, cb = cb->next)
  {
    if (cb->persistent_not_transient &&
        cmd == cb->cmd && regid == cb->reg)
    {
      /* Remove from active list */
      if (NULL != prev_cb)
        prev_cb = cb->next;
      else
        rac->active_callbacks = cb->next;

      /* Insert into unused list */
      cb->next = rac->unused_callbacks;
      rac->unused_callbacks = cb;
      return rcr_success;
    }
  }

  /* And make sure it isn't in the new active callbacks */
  for (cb = rac->new_active_callbacks, prev_cb = NULL; cb != NULL; prev_cb = cb, cb = cb->next)
  {
    if (cb->persistent_not_transient &&
        cmd == cb->cmd && regid == cb->reg)
    {
      /* Remove from new active list */
      if (NULL != prev_cb)
        prev_cb = cb->next;
      else
        rac->new_active_callbacks = cb->next;

      /* Insert into persistent list */
      cb->next = rac->unused_callbacks;
      rac->unused_callbacks = cb;
      return rcr_success;
    }
  }

  return rcr_listempty;
}


/**
 * Add a new command (in rac->outbuffer)
 *
 * @param rac is the rincmd asynchronous client
 * @param cmd is the command to execute
 * @param regid is the register to read
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
static enum e_rc_cli_result rc_async_client_register_initiate(struct s_rc_async_client *rac,
                                                              enum e_rc_reg_command cmd, rc_RegId_t regid,
                                                              t_rc_async_callback_fn callback, void *userdata,
                                                              const struct timespec *timeout)
{
  enum e_rc_cli_result result;
  ssize_t writelen;

  /* Add a callback for the register read */
  result = rc_async_client_add_callback(rac, false, cmd, regid, callback, userdata, timeout);
  if (rcr_success != result)
    return result;

  /* Send the command */
  writelen = strlen(rac->outbuffer);
  return rc_cli_comms_write(rac->cli_comms, rac->outbuffer, &writelen);
}


/**
 * Read a register via an asynchronous client
 *
 * @param rac is the rincmd asynchronous client
 * @param read_cmd is the command to execute
 * @param regid is the register to read
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
enum e_rc_cli_result rc_async_client_register_read(struct s_rc_async_client *rac,
                                                   enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                   t_rc_async_callback_fn callback, void *userdata,
                                                   const struct timespec *timeout)
{
  enum e_rc_cli_result result;

  /* Create the read command */
  result = rc_cli_assemble_register_read(rac->cli_parser, read_cmd, regid, rac->outbuffer, rac->outbufferlen);
  if (rcr_success != result)
    return result;

  /* Initiate it */
  return rc_async_client_register_initiate(rac, read_cmd, regid, callback, userdata, timeout);
}


/**
 * Write buffer register regid
 *
 * @param rac is the rincmd asynchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the buffer to write to the register
 * @param regvalue_len is the length of the buffer to write to the register
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
enum e_rc_cli_result rc_async_client_buf_register_write(struct s_rc_async_client *rac,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                        const void *regvalue, size_t regvalue_len,
                                                        t_rc_async_callback_fn callback, void *userdata,
                                                        const struct timespec *timeout)
{
  enum e_rc_cli_result result;

  /* Create the write command */
  result = rc_cli_assemble_buf_register_write(rac->cli_parser, write_cmd, regid, regvalue, regvalue_len, rac->outbuffer, rac->outbufferlen);
  if (rcr_success != result)
    return result;

  /* Initiate it */
  return rc_async_client_register_initiate(rac, write_cmd, regid, callback, userdata, timeout);
}


/**
 * Write string register regid
 *
 * @param rac is the rincmd asynchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the string to write to the register
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
enum e_rc_cli_result rc_async_client_str_register_write(struct s_rc_async_client *rac,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                        const char *regvalue,
                                                        t_rc_async_callback_fn callback, void *userdata,
                                                        const struct timespec *timeout)
{
  enum e_rc_cli_result result;

  /* Create the write command */
  result = rc_cli_assemble_str_register_write(rac->cli_parser, write_cmd, regid, regvalue, rac->outbuffer, rac->outbufferlen);
  if (rcr_success != result)
    return result;

  /* Initiate it */
  return rc_async_client_register_initiate(rac, write_cmd, regid, callback, userdata, timeout);
}


/**
 * Write integer register regid
 *
 * @param rac is the rincmd asynchronous client
 * @param write_cmd is the command to execute (RC_MESG_EXECUTE, RC_MESG_WRITE_HEXADEC, RC_MESG_WRITE_DECIMAL)
 * @param regid is the register to read
 * @param regvalue is the integer to write to the register
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
enum e_rc_cli_result rc_async_client_int_register_write(struct s_rc_async_client *rac,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                        int64_t regvalue,
                                                        t_rc_async_callback_fn callback, void *userdata,
                                                        const struct timespec *timeout)
{
  enum e_rc_cli_result result;

  /* Create the write command */
  result = rc_cli_assemble_int_register_write(rac->cli_parser, write_cmd, regid, regvalue, rac->outbuffer, rac->outbufferlen);
  if (rcr_success != result)
    return result;

  /* Initiate it */
  return rc_async_client_register_initiate(rac, write_cmd, regid, callback, userdata, timeout);
}


/**
 * Parse the result of an asynchronous integer register read from the response message
 *
 * Call this in your callback handler
 */
enum e_rc_cli_result rc_async_client_parse_int_register_read(struct s_rc_async_client *rac, enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                             const struct s_rc_Message *resp, int64_t *regvalue)
{
  return rc_cli_parse_int_register_message_read(rac->cli_parser, read_cmd, regid, resp, regvalue);
}


/**
 * Parse the result of an asynchronous buffer register read from the response message
 *
 * Call this in your callback handler
 */
enum e_rc_cli_result rc_async_client_parse_buf_register_read(struct s_rc_async_client *rac, enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                             const struct s_rc_Message *resp, void *regvalue, size_t *regvalue_len)
{
  return rc_cli_parse_buf_register_message_read(rac->cli_parser, read_cmd, regid, resp, regvalue, regvalue_len);
}


/**
 * Parse the result of an asynchronous string register read from the response message
 *
 * Call this in your callback handler
 */
enum e_rc_cli_result rc_async_client_parse_str_register_read(struct s_rc_async_client *rac, enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                             const struct s_rc_Message *resp, char *regvalue, size_t regvalue_len)
{
  return rc_cli_parse_str_register_message_read(rac->cli_parser, read_cmd, regid, resp, regvalue, regvalue_len);
}


/**
 * Parse the result of an asynchronous streaming register read (1..RC_STREAM_DATA_REGISTERS integer registers) from the response message
 *
 * Call this in your callback handler
 */
enum e_rc_cli_result rc_async_client_parse_stream_register_read(struct s_rc_async_client *rac, rc_RegId_t stream_reg_base,
                                                                const struct s_rc_Message *resp, int64_t regvalues[RC_STREAM_DATA_REGISTERS])
{
  return rc_cli_parse_stream_register_message_read(rac->cli_parser, stream_reg_base, resp, regvalues);
}
