/*
 * rc_cli_async.h
 *
 *  Created on: Nov 28, 2017
 *      Author: ian
 */

#ifndef RC_CLI_ASYNC_H_
#define RC_CLI_ASYNC_H_

#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>
#include <time.h>
#include <rincmd/rc_cli_comms.h>
#include <rincmd/rc_parser.h>
#include <rincmd/rc_public.h>


/** Callback for a response to a async message */
typedef void (*t_rc_async_callback_fn)(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result);


/**
 * Callback for badly formed or badly parsed received data
 *
 * @param userdata is for some user data to be passed back when the call is made
 * @param msgstr is the string that could not be parsed. Not NULL
 * @param msgstrlen is the length of msgstr
 * @param msg is the parsed version of msgstr. May be NULL if nothing could be parsed
 * @param result is the reason why this message could not be parsed
 */
typedef void (*t_rc_async_bad_message_callback_fn)(void *userdata, const char *msgstr, size_t msgstrlen, const struct s_rc_Message *msg, enum e_rc_cli_result result);


/** Asynchronous callback */
struct s_rc_async_callback
{
  bool persistent_not_transient;        /*!< Is this a persistent callback, or a transient one */
  void *userdata;                       /*!< User specific data */
  enum e_rc_reg_command cmd;            /*!< The expected command */
  rc_RegId_t reg;                       /*!< The expected register */
  t_rc_async_callback_fn callback_fn;   /*!< The function called when the response is received, or the timeout has occurred */
  bool hastimeout;                      /*!< This callback has a timeout */
  struct timespec abstimeout;           /*!< The absolute time to timeout in CLOCK_MONOTONIC */
  struct timespec persistent_timeout;   /*!< The (relative) persistent callback timeout that gets set into abstimeout after every response */
  struct s_rc_async_callback *next;     /*!< Next one in the list */
};


/** Asynchronous client */
struct s_rc_async_client
{
  struct s_rc_async_callback *unused_callbacks;               /*!< List of callbacks that are not in use */
  struct s_rc_async_callback *new_active_callbacks;           /*!< List of just added active callbacks, that get transferred to the active_callbacks at an appropriate time */
  struct s_rc_async_callback *active_callbacks;               /*!< List of active callbacks (persistent and transient) */
  struct s_rc_Message *msg;                                   /*!< A rincmd message */
  char *inbuffer;                                             /*!< The incoming data buffer */
  size_t inbufferlen;                                         /*!< The maximum length of the incoming data buffer */
  size_t inbufferposn;                                        /*!< The current position in the incoming data buffer */
  char *outbuffer;                                            /*!< The output data buffer */
  size_t outbufferlen;                                        /*!< The maximum length of the outgoing data buffer */
  struct s_rc_cli_comms *cli_comms;                           /*!< Communication with the host */
  struct s_rc_cli_parser *cli_parser;                         /*!< Parser for pulling apart and creating messages */

  /* Connect stuff */
  t_rc_async_connect_callback_fn connect_callback_fn;         /*!< The function called when the socket has connected or disconnected */
  void *connect_callback_userdata;                            /*!< User data pointer for the connect callback */
  bool connect_hastimeout;                                    /*!< Does the connect have a timeout */
  struct timespec connect_callback_timeout;                   /*!< The timeout period for connecting */
  struct timespec connect_callback_abstimeout;                /*!< The absolute time to timeout for connecting */

  /* Bad message stuff */
  t_rc_async_bad_message_callback_fn bad_message_callback_fn; /* The function called when received data cannot be parsed */
  void *bad_message_callback_userdata;                        /* User data pointer for the bad message callback */
};


/**
 * Create an asynchronous client
 *
 * @param rac is set to the created async client on success
 * @param cli_comms is the already existing comms interface. You are responsible for destroying this once rac has been destroyed
 * @param rc_address is the rincmd address to address messages to. RC_MESG_BROADCAST is a good choice
 * @param use_crc is a flag to indicate the use of CRC framed messages, or normal framing
 * @param num_callbacks is the total number of callbacks supported by this client (persistent and transient)
 * @param connect_fn is a callback function called when comms (socket,etc) has connected or disconnected (via rc_async_client_connect() call)
 * @param connect_userdata is a pointer to user data passed to the connect callback function
 * @param bad_message_fn is the function called when received data cannot be parsed
 * @param bad_message_userdata is a pointer to user data passed to the bad message callback function
 */
enum e_rc_cli_result rc_async_client_init(struct s_rc_async_client **rac,
                                          struct s_rc_cli_comms *cli_comms,
                                          uint8_t rc_address, bool use_crc,
                                          size_t num_callbacks,
                                          t_rc_async_connect_callback_fn connect_fn,
                                          void *connect_userdata,
                                          t_rc_async_bad_message_callback_fn bad_message_fn,
                                          void *bad_message_userdata);


/** Destroy an asynchronous client */
enum e_rc_cli_result rc_async_client_destroy(struct s_rc_async_client *rac);


/**
 * Inititate the async connection to the server
 *
 * @param rac is the rincmd async client
 * @param connect_timeout is the timeout to use when connecting. NULL for indefinite timeout
 */
enum e_rc_cli_result rc_async_client_connect(struct s_rc_async_client *rac, const struct timespec *timeout);


/** Disconnect from the server */
enum e_rc_cli_result rc_async_client_disconnect(struct s_rc_async_client *rac);


/**
 * Get the read and connect file descriptors
 *
 * Can be used to determine if the async client is connecting, disconnected, or connected
 *
 * @param rac is the rincmd async client
 * @param readfd is set to the read file descriptor if the comms is connected, < 0 otherwise. Used in the readfds param to select(). Not NULL.
 * @param connectfd is set to the connect file descriptor if the comms is connecting, < 0 otherwise. Used in the writefds param to select(). Not NULL.
 */
enum e_rc_cli_result rc_async_client_get_fd(const struct s_rc_async_client *rac, int *readfd, int *connectfd);


/**
 * Call prior to select() to handle any timeouts that may have occurred and
 * return the read and connect file descriptors and the next timeout if there is one
 *
 * @param rac is the rincmd async client
 * @param readfd is set to the read file descriptor if the comms is connected, < 0 otherwise. Used in the readfds param to select(). Not NULL.
 * @param connectfd is set to the connect file descriptor if the comms is connecting, < 0 otherwise. Used in the writefds param to select(). Not NULL.
 * @param hastimeout is set if the there is a timeout. Not NULL
 * @param timeout is set to the timeout period. Not NULL
 */
enum e_rc_cli_result rc_async_client_pre_select(struct s_rc_async_client *rac,
                                                int *readfd, int *connectfd,
                                                bool *hastimeout, struct timespec *timeout);


/**
 * Handle connection, timeout and traffic for the asynchronous client
 *
 * @param rac is the async connection
 * @param read_avail set true if your select() call has indicated there is data to be read on the read file descriptor
 * @param connect set true if your select() call has indicated there is a connect to be handled
 */
enum e_rc_cli_result rc_async_client_post_select(struct s_rc_async_client *rac, bool read_avail, bool connect);


/** Add a callback to the list of callbacks. Do not call directly. */
enum e_rc_cli_result rc_async_client_add_callback(struct s_rc_async_client *rac, bool persistent_not_transient,
                                                  enum e_rc_reg_command cmd, rc_RegId_t regid,
                                                  t_rc_async_callback_fn callback, void *userdata,
                                                  const struct timespec *timeout);


/** Remove a persistent callback from the list of active callbacks. Do not call directly. */
enum e_rc_cli_result rc_async_client_remove_persistent_callback(struct s_rc_async_client *rac,
                                                                enum e_rc_reg_command cmd, rc_RegId_t regid);


/**
 * Read a register via an asynchronous client
 *
 * @param rac is the rincmd asynchronous client
 * @param read_cmd is the command to execute
 * @param regid is the register to read
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
enum e_rc_cli_result rc_async_client_register_read(struct s_rc_async_client *rac,
                                                   enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                   t_rc_async_callback_fn callback, void *userdata,
                                                   const struct timespec *timeout);


/**
 * Write buffer register regid
 *
 * @param rac is the rincmd asynchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the buffer to write to the register
 * @param regvalue_len is the length of the buffer to write to the register
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
enum e_rc_cli_result rc_async_client_buf_register_write(struct s_rc_async_client *rac,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                        const void *regvalue, size_t regvalue_len,
                                                        t_rc_async_callback_fn callback, void *userdata,
                                                        const struct timespec *timeout);


/**
 * Write string register regid
 *
 * @param rac is the rincmd asynchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the string to write to the register
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
enum e_rc_cli_result rc_async_client_str_register_write(struct s_rc_async_client *rac,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                        const char *regvalue,
                                                        t_rc_async_callback_fn callback, void *userdata,
                                                        const struct timespec *timeout);


/**
 * Write integer register regid
 *
 * @param rac is the rincmd asynchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the integer to write to the register
 * @param callback is the function to call back when we get the response. Can be NULL, in which case response is ignored
 * @param userdata is a pointer passed to the callback. Can be NULL
 * @param timeout is the timeout period. Can be NULL, in which case there is no timeout
 */
enum e_rc_cli_result rc_async_client_int_register_write(struct s_rc_async_client *rac,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                        int64_t regvalue,
                                                        t_rc_async_callback_fn callback, void *userdata,
                                                        const struct timespec *timeout);


/**
 * Parse the result of an asynchronous streaming register read (1..RC_STREAM_DATA_REGISTERS integer registers) from the response message
 *
 * Call this in your callback handler
 */
enum e_rc_cli_result rc_async_client_parse_stream_register_read(struct s_rc_async_client *rac, rc_RegId_t stream_reg_base,
                                                                const struct s_rc_Message *resp, int64_t regvalues[RC_STREAM_DATA_REGISTERS]);


/**
 * Parse the result of an asynchronous integer register read from the response message
 *
 * Call this in your callback handler
 */
enum e_rc_cli_result rc_async_client_parse_int_register_read(struct s_rc_async_client *rac, enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                             const struct s_rc_Message *resp, int64_t *regvalue);


/**
 * Parse the result of an asynchronous buffer register read from the response message
 *
 * Call this in your callback handler
 */
enum e_rc_cli_result rc_async_client_parse_buf_register_read(struct s_rc_async_client *rac, enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                             const struct s_rc_Message *resp, void *regvalue, size_t *regvalue_len);


/**
 * Parse the result of an asynchronous string register read from the response message
 *
 * Call this in your callback handler
 */
enum e_rc_cli_result rc_async_client_parse_str_register_read(struct s_rc_async_client *rac, enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                             const struct s_rc_Message *resp, char *regvalue, size_t regvalue_len);


#endif /* RC_CLI_ASYNC_H_ */
