/*
 * rc_cli_async_bulk.c
 *
 *  Created on: 28 May 2018
 *      Author: ian
 */
#include <stdio.h>
#include <string.h>
#include <rincmd/rc_parser.h>
#include <rincmd/rc_cli_async.h>
#include <rincmd/rc_cli_parse.h>

#include <rincmd/rc_cli_async_bulk.h>


static enum e_rc_cli_result rc_async_client_request_fillout(struct s_rc_async_client_request *acr,
                                                            enum e_rc_reg_command command, rc_RegId_t regid,
                                                            const int64_t *regvalue_int,
                                                            const char *regvalue_buf, size_t regvalue_buf_len,
                                                            t_rc_async_callback_fn callback, void *userdata,
                                                            const struct timespec *timeout)
{
  if (NULL == acr)
    return rcr_paramfail;

  acr->command = command;
  acr->regid = regid;
  acr->regvalue_int = regvalue_int;
  acr->regvalue_buf = regvalue_buf;
  acr->regvalue_buf_len = regvalue_buf_len;
  acr->callback = callback;
  acr->userdata = userdata;
  if (NULL == timeout)
  {
    acr->has_timeout = false;
    acr->timeout.tv_sec = 0;
    acr->timeout.tv_nsec = 0;
  }
  else
  {
    acr->has_timeout = true;
    acr->timeout = *timeout;
  }
  return rcr_success;
}


/** Fill out a single bulk request structure for a register read */
enum e_rc_cli_result rc_async_client_read_request_fillout(struct s_rc_async_client_request *acr,
                                                          enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                          t_rc_async_callback_fn callback, void *userdata,
                                                          const struct timespec *timeout)
{
  return rc_async_client_request_fillout(acr, read_cmd, regid,
                                         0, NULL, 0,
                                         callback, userdata, timeout);
}


/** Fill out a single bulk request structure for an integer register write */
enum e_rc_cli_result rc_async_client_int_write_request_fillout(struct s_rc_async_client_request *acr,
                                                               enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                               const int64_t *regvalue,
                                                               t_rc_async_callback_fn callback, void *userdata,
                                                               const struct timespec *timeout)
{
  return rc_async_client_request_fillout(acr, write_cmd, regid,
                                         regvalue, NULL, 0,
                                         callback, userdata, timeout);
}


/** Fill out a single bulk request structure for a buffer register write */
enum e_rc_cli_result rc_async_client_buf_write_request_fillout(struct s_rc_async_client_request *acr,
                                                               enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                               const void *regvalue, size_t regvalue_len,
                                                               t_rc_async_callback_fn callback, void *userdata,
                                                               const struct timespec *timeout)
{
  return rc_async_client_request_fillout(acr, write_cmd, regid,
                                         NULL, regvalue, regvalue_len,
                                         callback, userdata, timeout);
}


/** Fill out a single bulk request structure for a string register write */
enum e_rc_cli_result rc_async_client_str_write_request_fillout(struct s_rc_async_client_request *acr,
                                                               enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                               const char *regvalue,
                                                               t_rc_async_callback_fn callback, void *userdata,
                                                               const struct timespec *timeout)
{
  return rc_async_client_request_fillout(acr, write_cmd, regid,
                                         NULL, regvalue, strlen(regvalue),
                                         callback, userdata, timeout);
}


/**
 * Read or write a number of registers.
 *
 * The resulting rincmds are assembled into a buffer and written with a single write
 * to minimise network overheads.
 *
 * @param rac is the rincmd asynchronous client
 * @param requests is an array of filled out request structures
 * @param num_requests is the number of requests in the requests
 */
enum e_rc_cli_result rc_async_client_bulk_request(struct s_rc_async_client *rac,
                                                  const struct s_rc_async_client_request *requests,
                                                  size_t num_requests)
{
  size_t i;
  enum e_rc_cli_result result;
  size_t outbuf_posn;
  size_t outbuf_left;
  ssize_t writelen;

  if (NULL == rac || NULL == requests)
    return rcr_paramfail;

  for (i = 0, outbuf_posn = 0; i < num_requests; i++)
  {
    outbuf_left = rac->outbufferlen - outbuf_posn;
    if (RC_IS_WRITE_COMMAND(requests[i].command) ||
        RC_IS_EXEC_COMMAND(requests[i].command))
    {
      if (NULL != requests[i].regvalue_int)
      {
        result = rc_cli_assemble_int_register_write(rac->cli_parser, requests[i].command, requests[i].regid,
                                                    *requests[i].regvalue_int,
                                                    &rac->outbuffer[outbuf_posn], outbuf_left);
      }
      else if (NULL != requests[i].regvalue_buf)
      {
        result = rc_cli_assemble_buf_register_write(rac->cli_parser, requests[i].command, requests[i].regid,
                                                    requests[i].regvalue_buf, requests[i].regvalue_buf_len,
                                                    &rac->outbuffer[outbuf_posn], outbuf_left);
      }
    }
    else if (RC_IS_READ_COMMAND(requests[i].command))
    {
      result = rc_cli_assemble_register_read(rac->cli_parser, requests[i].command, requests[i].regid,
                                             &rac->outbuffer[outbuf_posn], outbuf_left);
    }
    else
    {
      result = rcr_paramfail;
    }
    if (rcr_success != result)
      return result;

    /* Add a callback for the register command */
    result = rc_async_client_add_callback(rac, false,
                                          requests[i].command, requests[i].regid,
                                          requests[i].callback, requests[i].userdata,
                                          (requests[i].has_timeout) ? &requests[i].timeout : NULL);
    if (rcr_success != result)
      return result;

    outbuf_posn += strlen(&rac->outbuffer[outbuf_posn]);
  }

  /* Send the commands */
  writelen = strlen(rac->outbuffer);
  return rc_cli_comms_write(rac->cli_comms, rac->outbuffer, &writelen);
}
