/*
 * rc_cli_async_stream.c
 *
 *  Created on: Dec 7, 2017
 *      Author: ian
 */
#include <stdlib.h>
#include <string.h>
#include <rincmd/rc_public.h>
#include <rincmd/rc_streaming.h>
#include <rincmd/rc_cli_async.h>

#include <rincmd/rc_cli_async_stream.h>


/** Structure for setting up streaming */
struct s_rc_async_streaming_setup
{
  struct s_rc_async_client *rac;              /*!< The async client to setup streaming on */
  t_rc_async_callback_fn callback;            /*!< The callback called on streaming setup success/failure, and on each stream data */
  void *userdata;                             /*!< User data passed to the callback */
  rc_RegId_t stream_reg_base;                 /*!< The base register address of the streaming set */
  enum e_rc_stream_mode stream_mode;          /*!< The mode for the streaming set (rate of streaming) */
  rc_RegId_t regs[RC_STREAM_DATA_REGISTERS];  /*!< The registers to be streamed */
  size_t num_regs;                            /*!< The number of registers to be streamed 1..RC_STREAM_DATA_REGISTERS */
  size_t current_reg;                         /*!< The current register being setup to be streamed */
  bool has_setup_timeout;                     /*!< Is there a setup timeout */
  struct timespec setup_timeout;              /*!< The setup timeout if has_setup_timeout */
  bool has_stream_timeout;                    /*!< Is there a streaming timeout */
  struct timespec stream_timeout;             /*!< The stream timeout if has_stream_timeout */
  bool setup_not_teardown;                    /*!< Is this stream being setup or torn down */
};


/** Called when streaming setup fails or is complete. Calls the streaming callback function and frees the streaming setup function. */
static enum e_rc_cli_result rc_async_client_setup_streaming_result(struct s_rc_async_streaming_setup *ass, enum e_rc_cli_result result)
{
  if (NULL == ass && NULL == ass->callback)
  {
    result = rcr_paramfail;
  }
  else
  {
    ass->callback(ass->userdata, NULL, result);
    result = rcr_success;
  }

  free(ass);
  return result;
}


/** Called once the streaming data register has been set */
static void rc_async_client_setup_streaming_datareg(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct s_rc_async_streaming_setup *ass = (struct s_rc_async_streaming_setup *) userdata;

  /* Simply notify of success/failure */
  result = (rcr_success == result) ? rcr_streamsetupsuccess : rcr_streamsetupfail;
  rc_async_client_setup_streaming_result(ass, result);
}


/** Called once the streaming mode register has been set. Writes the streaming data register */
static void rc_async_client_setup_streaming_modereg(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct s_rc_async_streaming_setup *ass = (struct s_rc_async_streaming_setup *) userdata;
  struct timespec *timeout;

  if (rcr_success != result)
    goto racssm_fail;

  if (ass->setup_not_teardown)
  {
    /* Add a persistent callback to handle the streaming of the data register */
    timeout = (ass->has_stream_timeout) ? &ass->stream_timeout : NULL;
    result = rc_async_client_add_callback(ass->rac, true,
                                          RC_MESG_READ_HEXADEC, ass->stream_reg_base,
                                          ass->callback, ass->userdata,
                                          timeout);
    if (rcr_success != result)
      goto racssm_fail;

    /* Write the data register to enable streaming */
    timeout = (ass->has_setup_timeout) ? &ass->setup_timeout : NULL;
    result = rc_async_client_int_register_write(ass->rac,
                                                RC_MESG_EXECUTE, ass->stream_reg_base,
                                                1,
                                                rc_async_client_setup_streaming_datareg, ass,
                                                timeout);
    if (rcr_success != result)
      goto racssm_fail;
  }
  else
  {
    /* Tearing down, all done */
    rc_async_client_setup_streaming_result(ass, rcr_streamsetupsuccess);
  }
  return;

racssm_fail:
  rc_async_client_setup_streaming_result(ass, rcr_streamsetupfail);
}


/** Called once each of the streaming registers have been set. Then writes the next one, or the mode register once all have been set. */
static void rc_async_client_setup_streaming_streamregs(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct s_rc_async_streaming_setup *ass = (struct s_rc_async_streaming_setup *) userdata;
  struct timespec *timeout;

  if (rcr_success != result)
    goto racssd_fail;

  timeout = (ass->has_setup_timeout) ? &ass->setup_timeout : NULL;
  if (ass->current_reg < ass->num_regs)
  {
    /* Write each of the streaming data registers */
    result = rc_async_client_int_register_write(ass->rac,
                                                RC_MESG_WRITE_HEXADEC, ass->stream_reg_base + RC_STREAM_REG_DATA_OFFSET + ass->current_reg,
                                                ass->regs[ass->current_reg],
                                                rc_async_client_setup_streaming_streamregs, ass,
                                                timeout);
    if (rcr_success != result)
      goto racssd_fail;

    ass->current_reg++;
  }
  else
  {
    /* Write the mode register */
    result = rc_async_client_int_register_write(ass->rac,
                                                RC_MESG_WRITE_HEXADEC, ass->stream_reg_base + RC_STREAM_REG_MODE_OFFSET,
                                                (int64_t) ass->stream_mode,
                                                rc_async_client_setup_streaming_modereg, ass,
                                                timeout);
    if (rcr_success != result)
      goto racssd_fail;
  }
  return;

racssd_fail:
  rc_async_client_setup_streaming_result(ass, rcr_streamsetupfail);
}


/** Writes the streaming data register to disable streaming */
static enum e_rc_cli_result rc_async_client_teardown_streaming_datareg(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct s_rc_async_streaming_setup *ass = (struct s_rc_async_streaming_setup *) userdata;
  struct timespec *timeout;

  if (rcr_success != result)
    goto racssm_fail;

  rc_async_client_remove_persistent_callback(ass->rac, RC_MESG_READ_HEXADEC, ass->stream_reg_base);

  /* Write the data register to disable streaming */
  timeout = (ass->has_setup_timeout) ? &ass->setup_timeout : NULL;
  result = rc_async_client_int_register_write(ass->rac,
                                              RC_MESG_EXECUTE, ass->stream_reg_base,
                                              0,
                                              rc_async_client_setup_streaming_streamregs, ass,
                                              timeout);
  if (rcr_success != result)
    goto racssm_fail;
  return rcr_success;

racssm_fail:
  rc_async_client_setup_streaming_result(ass, rcr_streamsetupfail);
  return result;
}


/** Setup/teardown a streaming set for an asynchronous client */
static enum e_rc_cli_result rc_async_client_streaming_setup_internal(struct s_rc_async_client *rac,
                                                                     rc_RegId_t stream_reg_base,
                                                                     t_rc_async_callback_fn callback, void *userdata,
                                                                     enum e_rc_stream_mode stream_mode,
                                                                     const int regs[RC_STREAM_DATA_REGISTERS],
                                                                     const struct timespec *setup_timeout,
                                                                     const struct timespec *stream_timeout,
                                                                     bool setup_not_teardown)
{
  enum e_rc_cli_result result;
  struct s_rc_async_streaming_setup *ass = NULL;

  if (NULL == rac || NULL == callback)
    return rcr_paramfail;

  ass = calloc(1, sizeof(*ass));
  if (NULL == ass)
  {
    result = rcr_allocfail;
    goto racss_fail;
  }

  ass->rac = rac;
  ass->callback = callback;
  ass->userdata = userdata;
  ass->stream_reg_base = stream_reg_base;
  ass->stream_mode = stream_mode;
  ass->current_reg = 0;

  memcpy(ass->regs, regs, sizeof(int) * RC_STREAM_DATA_REGISTERS);
  ass->num_regs = RC_STREAM_DATA_REGISTERS;

  if (NULL == setup_timeout)
  {
    ass->has_setup_timeout = false;
  }
  else
  {
    ass->has_setup_timeout = true;
    ass->setup_timeout = *setup_timeout;
  }
  if (NULL == stream_timeout)
  {
    ass->has_stream_timeout = false;
  }
  else
  {
    ass->has_stream_timeout = true;
    ass->stream_timeout = *stream_timeout;
  }
  ass->setup_not_teardown = setup_not_teardown;

  if (setup_not_teardown)
    rc_async_client_setup_streaming_streamregs(ass, NULL, rcr_success);
  else
    rc_async_client_teardown_streaming_datareg(ass, NULL, rcr_success);
  return rcr_success;

racss_fail:
  if (NULL != ass)
  {
    free(ass);
  }
  return result;
}


/**
 * Setup a streaming set for an asynchronous client
 *
 * Sequence is write steam registers, write mode register, execute data register
 */
enum e_rc_cli_result rc_async_client_streaming_setup(struct s_rc_async_client *rac,
                                                     rc_RegId_t stream_reg_base,
                                                     t_rc_async_callback_fn callback, void *userdata,
                                                     enum e_rc_stream_mode stream_mode,
                                                     const int regs[RC_STREAM_DATA_REGISTERS],
                                                     const struct timespec *setup_timeout,
                                                     const struct timespec *stream_timeout)
{
  return rc_async_client_streaming_setup_internal(rac, stream_reg_base, callback, userdata,
                                                  stream_mode, regs, setup_timeout, stream_timeout, true);
}


/**
 * Teardown a streaming set for an asynchronous client
 *
 * Sequence is execute data register, write steam registers, write mode register
 */
enum e_rc_cli_result rc_async_client_streaming_teardown(struct s_rc_async_client *rac,
                                                        rc_RegId_t stream_reg_base,
                                                        t_rc_async_callback_fn callback, void *userdata,
                                                        const struct timespec *setup_timeout)
{
  const int regs_off[RC_STREAM_DATA_REGISTERS] = { 0, 0, 0, 0, 0 };

  return rc_async_client_streaming_setup_internal(rac, stream_reg_base, callback, userdata,
                                                  RC_STREAM_MODE_OFF, regs_off, setup_timeout, NULL, false);
}
