/*
 * rc_cli_async_timer.c
 *
 *  Created on: 24 May 2018
 *      Author: ian
 */
#include <stdlib.h>
#include <rincmd/rc_timespec.h>
#include <rincmd/rc_cli_gettime.h>

#include <rincmd/rc_cli_async_timer.h>


/** Allocate a new callback */
static enum e_rc_cli_result rc_async_timer_new_callback(struct s_rc_async_timer_callback **new_cb)
{
  struct s_rc_async_timer_callback *cb;

  if (NULL == new_cb)
    return rcr_paramfail;

  cb = calloc(1, sizeof(*cb));
  if (NULL == cb)
  {
    *new_cb = NULL;
    return rcr_allocfail;
  }

  cb->persistent_not_transient = false;
  cb->callback_fn = NULL;
  cb->userdata = NULL;
  cb->abstimeout.tv_sec = 0;
  cb->abstimeout.tv_nsec = 0;
  cb->persistent_timeout.tv_sec = 0;
  cb->persistent_timeout.tv_nsec = 0;
  cb->next = NULL;
  cb->remove_next = NULL;

  *new_cb = cb;
  return rcr_success;
}


/** Pre-allocate the callbacks */
static enum e_rc_cli_result rc_async_timer_create_callbacks(struct s_rc_async_timer *rat, size_t num_callbacks)
{
  struct s_rc_async_timer_callback *cb;
  struct s_rc_async_timer_callback *prev_cb;
  struct s_rc_async_timer_callback *next_cb;
  size_t i;
  enum e_rc_cli_result result;

  rat->unused_callbacks = NULL;
  rat->new_active_callbacks = NULL;
  rat->active_callbacks = NULL;
  rat->remove_callbacks = NULL;

  /* Allocate unused persistent callbacks, and insert them into the linked list */
  for (i = 0, prev_cb = NULL; i < num_callbacks; i++, prev_cb =  cb)
  {
    result = rc_async_timer_new_callback(&cb);
    if (rcr_success != result)
      goto ratcc_fail;
    cb->next = prev_cb;
  }
  rat->unused_callbacks = prev_cb;

  return rcr_success;

ratcc_fail:
  for (cb = rat->unused_callbacks; cb != NULL; )
  {
    next_cb = cb->next;
    free(cb);
    cb = next_cb;
  }
  rat->unused_callbacks = NULL;
  return result;
}


/**
 * Create an asynchronous timer
 *
 * @param rat is set to the created async timer on success
 * @param num_callbacks is the total number of callbacks supported by this timer (persistent and transient)
 */
enum e_rc_cli_result rc_async_timer_init(struct s_rc_async_timer **rat, size_t num_callbacks)
{
  struct s_rc_async_timer *irat = NULL;
  enum e_rc_cli_result result;

  if (NULL == rat)
    return rcr_paramfail;

  /* Allocate the async client structure */
  irat = (struct s_rc_async_timer *) calloc(1, sizeof(*irat));
  if (NULL == irat)
  {
    result = rcr_allocfail;
    goto raci_fail;
  }

  /* Setup callbacks */
  result = rc_async_timer_create_callbacks(irat, num_callbacks);
  if (rcr_success != result)
    goto raci_fail;

  /* Success */
  *rat = irat;
  return rcr_success;

raci_fail:
  (void) rc_async_timer_destroy(irat);
  irat = NULL;
  return result;
}


/** Destroy an asynchronous timer */
enum e_rc_cli_result rc_async_timer_destroy(struct s_rc_async_timer *rat)
{
  struct s_rc_async_timer_callback *cb;
  struct s_rc_async_timer_callback *next_cb;

  if (NULL != rat)
  {
    for (cb = rat->unused_callbacks; cb != NULL; )
    {
      next_cb = cb->next;
      free(cb);
      cb = next_cb;
    }
    rat->unused_callbacks = NULL;
    for (cb = rat->new_active_callbacks; cb != NULL; )
    {
      next_cb = cb->next;
      free(cb);
      cb = next_cb;
    }
    rat->new_active_callbacks = NULL;
    for (cb = rat->active_callbacks; cb != NULL; )
    {
      next_cb = cb->next;
      free(cb);
      cb = next_cb;
    }
    rat->active_callbacks = NULL;
    free(rat);
    rat = NULL;
  }
  return rcr_success;
}


/** Move any new active callbacks into the active callback list */
static void rc_async_timer_new_active_to_active(struct s_rc_async_timer *rat)
{
  struct s_rc_async_timer_callback *first_active_cb;
  struct s_rc_async_timer_callback *last_new_cb;
  struct s_rc_async_timer_callback *new_cb;

  if (NULL != rat->new_active_callbacks)
  {
    /* Only if there are new active callbacks */
    if (NULL == rat->active_callbacks)
    {
      /* No already active callbacks, so just move the new active callbacks to the active callbacks */
      rat->active_callbacks = rat->new_active_callbacks;
      rat->new_active_callbacks = NULL;
    }
    else
    {
      /* Find the last new active callback */
      for (new_cb = rat->new_active_callbacks, last_new_cb = NULL; new_cb != NULL; last_new_cb = new_cb, new_cb = new_cb->next)
        ;

      /* Put all new active callbacks at the start of the active callbacks list */
      first_active_cb = rat->active_callbacks;
      rat->active_callbacks = rat->new_active_callbacks;
      last_new_cb->next = first_active_cb;
      rat->new_active_callbacks= NULL;
    }
  }
}


/** Remove any active or new active callbacks marked for removal (and put back into the appropriate unused list) */
static void rc_async_timer_remove_active(struct s_rc_async_timer *rat)
{
  struct s_rc_async_timer_callback *remove_cb;
  struct s_rc_async_timer_callback *active_cb;
  struct s_rc_async_timer_callback *prev_active_cb;

  /* Iterate the remove callback list */
  for (remove_cb = rat->remove_callbacks; remove_cb != NULL; remove_cb = remove_cb->remove_next)
  {
    /* Is it in the active callbacks */
    for (active_cb = rat->active_callbacks, prev_active_cb = NULL; active_cb != NULL; prev_active_cb = active_cb, active_cb = active_cb->next)
    {
      if (remove_cb == active_cb)
      {
        /* Yep, remove from active list */
        if (NULL == prev_active_cb)
          rat->active_callbacks = active_cb->next;
        else
          prev_active_cb->next = active_cb->next;

        /* And put it back in the unused list */
        active_cb->next = rat->unused_callbacks;
        rat->unused_callbacks = active_cb;
        break;
      }
    } /* for active */

    /* Is it in the new active callbacks */
    for (active_cb = rat->new_active_callbacks, prev_active_cb = NULL; active_cb != NULL; prev_active_cb = active_cb, active_cb = active_cb->next)
    {
      if (remove_cb == active_cb)
      {
        /* Yep, remove from active list */
        if (NULL == prev_active_cb)
          rat->new_active_callbacks = active_cb->next;
        else
          prev_active_cb->next = active_cb->next;

        /* And put it back in the unused list */
        active_cb->next = rat->unused_callbacks;
        rat->unused_callbacks = active_cb;
        break;
      }
    } /* for new active */

  } /* for removal */

  /* And now the removal list is empty */
  rat->remove_callbacks = NULL;
}


/**
 * Call prior to select() to handle any timer timeouts that may have occurred and
 * return the next timeout period if there is one
 *
 * @param rat is the rincmd async timer
 * @param hastimeout is set if the there is a timeout. Not NULL
 * @param timeout is set to the timeout period. Not NULL
 */
enum e_rc_cli_result rc_async_timer_run(struct s_rc_async_timer *rat,
                                        bool *hastimeout, struct timespec *timeout)
{
  struct s_rc_async_timer_callback *cb;
  struct timespec now;
  struct timespec shortest;
  bool first_timeout;
  struct timespec left;
  const struct timespec tvzero = { .tv_sec=0, .tv_nsec=0 };

  /* Check params */
  if (NULL == rat || NULL == hastimeout || NULL == timeout)
    return rcr_paramfail;

  /* Current time for timeouts */
  (void) rc_cli_external_gettime(&now);
  shortest.tv_sec = 0;
  shortest.tv_nsec = 0;
  first_timeout = true;

  /* Move any new active callbacks into the active callbacks list, and remove any removed timers */
  rc_async_timer_new_active_to_active(rat);
  rc_async_timer_remove_active(rat);

  /* Work out timeouts for register callbacks */
  for (cb = rat->active_callbacks; cb != NULL; cb = cb->next)
  {
    /* Work out how much time is left for this callback */
    rc_timespec_subtract(&left, &cb->abstimeout, &now);

    /* Has timed out already? */
    if (rc_timespec_compare(&left, &tvzero) <= 0)
    {
      /* Yes, call the callback if set */
      if (NULL != cb->callback_fn)
        cb->callback_fn(cb->userdata, (intptr_t) cb);

      if (cb->persistent_not_transient)
      {
        /* Re-calculate next timeout */
        rc_timespec_add(&cb->abstimeout, &now, &cb->persistent_timeout);

        /* Work out how much time is left for the next callback */
        rc_timespec_subtract(&left, &cb->abstimeout, &now);
        if (first_timeout || rc_timespec_compare(&left, &shortest) < 0)
        {
          /* This callback has the least time to go */
          shortest = left;
          first_timeout = false;
        }
      }
      else
      {
        /* Mark the callback for removal if transient */
        cb->remove_next = rat->remove_callbacks;
        rat->remove_callbacks = cb;
      }
    }
    else
    {
      if (first_timeout || rc_timespec_compare(&left, &shortest) < 0)
      {
        /* This callback has the least time to go */
        shortest = left;
        first_timeout = false;
      }
    }
  } /* for */

  *hastimeout = !first_timeout;
  *timeout = shortest;
  return rcr_success;
}


/**
 * Add a callback to the list of callbacks. Do not call directly.
 *
 * @param rat is the rincmd async timer
 * @param persistent_not_transient is whether this callback is persistent, or should be removed once it has completed
 * @param callback is the function to call when the timer has timed out. Can be NULL.
 * @param userdata is a user data pointer to pass to the callback function. Can be NULL.
 * @param timeout_is_absolute is true if the timeout is in absolute time (only used with transient timers).
 * @param timeout is an timeout value. Not NULL
 * @param timer_id is set to a unique identifier for each timer (if pointer is not NULL)
 */
static enum e_rc_cli_result rc_async_timer_add_callback(struct s_rc_async_timer *rat, bool persistent_not_transient,
                                                        t_rc_async_timer_callback_fn callback, void *userdata,
                                                        bool timeout_is_absolute, const struct timespec *timeout,
                                                        intptr_t *timer_id)
{
  struct s_rc_async_timer_callback *cb;
  struct timespec now;

  if (NULL == rat || NULL == timeout || (persistent_not_transient && timeout_is_absolute))
    return rcr_paramfail;

  if (NULL == rat->unused_callbacks)
    return rcr_listempty;

  /* Remove unused callback from the unused list */
  cb = rat->unused_callbacks;
  rat->unused_callbacks = cb->next;

  /* Fill it out */
  cb->persistent_not_transient = persistent_not_transient;
  cb->callback_fn = callback;
  cb->userdata = userdata;
  if (timeout_is_absolute)
  {
    cb->abstimeout = *timeout;
  }
  else
  {
    cb->persistent_timeout = *timeout;
    (void) rc_cli_external_gettime(&now);
    rc_timespec_add(&cb->abstimeout, &now, timeout);
  }

  /* Add it into the new active callbacks list. It will get transferred to the active_callbacks list at an appropriate time */
  cb->next = rat->new_active_callbacks;
  rat->new_active_callbacks = cb;

  if (NULL != timer_id)
    *timer_id = (intptr_t) cb;

  return rcr_success;
}


/** Add a one shot timer to the async timer with a absolute timeout in CLOCK_MONOTONIC */
enum e_rc_cli_result rc_async_timer_add_oneshot_absolute(struct s_rc_async_timer *rat,
                                                         t_rc_async_timer_callback_fn callback, void *userdata,
                                                         const struct timespec *abstimeout, intptr_t *timer_id)
{
  return rc_async_timer_add_callback(rat, false, callback, userdata, true, abstimeout, timer_id);
}


/** Add a one shot timer to the async timer with a relative timeout (from now) */
enum e_rc_cli_result rc_async_timer_add_oneshot_relative(struct s_rc_async_timer *rat,
                                                         t_rc_async_timer_callback_fn callback, void *userdata,
                                                         const struct timespec *timeout, intptr_t *timer_id)
{
  return rc_async_timer_add_callback(rat, false, callback, userdata, false, timeout, timer_id);
}


/** Add a persistent timer to the async timer with a relative timeout */
enum e_rc_cli_result rc_async_timer_add_persistent(struct s_rc_async_timer *rat,
                                                   t_rc_async_timer_callback_fn callback, void *userdata,
                                                   const struct timespec *timeout, intptr_t *timer_id)
{
  return rc_async_timer_add_callback(rat, true, callback, userdata, false, timeout, timer_id);
}


/** Remove a persistent timer (just put it in the removal list, to be removed by rc_async_timer_remove_active() at an appropriate time) */
enum e_rc_cli_result rc_async_timer_remove_callback(struct s_rc_async_timer *rat, intptr_t timer_id)
{
  struct s_rc_async_timer_callback *cb;

  /* Is it already marked for removal */
  for (cb = rat->remove_callbacks; cb != NULL; cb = cb->remove_next)
  {
    if ((intptr_t) cb == timer_id)
      return rcr_success;
  }

  /* First check the active list */
  for (cb = rat->active_callbacks; cb != NULL; cb = cb->next)
  {
    if ((intptr_t) cb == timer_id)
    {
      cb->remove_next = rat->remove_callbacks;
      rat->remove_callbacks = cb;
      return rcr_success;
    }
  }

  /* And make sure it isn't in the new active callbacks */
  for (cb = rat->new_active_callbacks; cb != NULL; cb = cb->next)
  {
    if ((intptr_t) cb == timer_id)
    {
      cb->remove_next = rat->remove_callbacks;
      rat->remove_callbacks = cb;
      return rcr_success;
    }
  }

  /* Not one of our timers */
  return rcr_listempty;
}


/**
 * Get timer (function pointer, user data pointer and timeout) referenced by
 * timer_id from the list of running timers
 *
 * @param rat is the async timer
 * @param timer_id is the timer identifier from rc_async_timer_add_*() functions
 * @param callback is a pointer to a async timer callback function pointer, set by the function when the matching timer_id is found
 * @param userdata is a pointer to a the user data pointer for the async timer, set by the function when the matching timer_id is found
 * @param abstimeout is a pointer to a timespec, set by the function to the next absolute time when the function will run
 *
 * @return a rincmd client result
 */
enum e_rc_cli_result rc_async_timer_get(struct s_rc_async_timer *rat, intptr_t timer_id,
                                        t_rc_async_timer_callback_fn *callback, void **userdata,
                                        struct timespec *abstimeout)
{
  struct s_rc_async_timer_callback *cb;

  /* Is it already marked for removal */
  for (cb = rat->remove_callbacks; cb != NULL; cb = cb->remove_next)
  {
    if ((intptr_t) cb == timer_id)
      return rcr_listempty;
  }

  /* First check the active list */
  for (cb = rat->active_callbacks; cb != NULL; cb = cb->next)
  {
    if ((intptr_t) cb == timer_id)
    {
      if (NULL != callback)
        *callback = cb->callback_fn;
      if (NULL != userdata)
        *userdata = cb->userdata;
      if (NULL != abstimeout)
        *abstimeout = cb->abstimeout;
      return rcr_success;
    }
  }

  /* And make sure it isn't in the new active callbacks */
  for (cb = rat->new_active_callbacks; cb != NULL; cb = cb->next)
  {
    if ((intptr_t) cb == timer_id)
    {
      if (NULL != callback)
        *callback = cb->callback_fn;
      if (NULL != userdata)
        *userdata = cb->userdata;
      if (NULL != abstimeout)
        *abstimeout = cb->abstimeout;
      return rcr_success;
    }
  }

  /* Not one of our timers */
  return rcr_listempty;
}
