/*
 * rc_cli_async_timer.h
 *
 *  Created on: 24 May 2018
 *      Author: ian
 */

#ifndef RC_CLI_ASYNC_TIMER_H_
#define RC_CLI_ASYNC_TIMER_H_

#include <stdbool.h>
#include <time.h>
#include <unistd.h>
#include <rincmd/rc_parser.h>
#include <rincmd/rc_public.h>
#include <rincmd/rc_cli_comms.h>


/** Callback for a async timer */
typedef void (*t_rc_async_timer_callback_fn)(void *userdata, intptr_t timer_id);


struct s_rc_async_timer_callback
{
  bool persistent_not_transient;                  /*!< Is this a persistent callback, or a transient one */
  t_rc_async_timer_callback_fn callback_fn;       /*!< The function called when the timeout has occurred */
  void *userdata;                                 /*!< User specific data */

  struct timespec abstimeout;                     /*!< The absolute time to timeout in CLOCK_MONOTONIC */
  struct timespec persistent_timeout;             /*!< The (relative) persistent callback timeout that gets set into abstimeout after every response */
  struct s_rc_async_timer_callback *next;         /*!< Next one in the list */
  struct s_rc_async_timer_callback *remove_next;  /*!< Next one in the removal list (can be in the active list and the removal list) */
};


struct s_rc_async_timer
{
  struct s_rc_async_timer_callback *unused_callbacks;      /*!< List of callbacks that are not in use */
  struct s_rc_async_timer_callback *new_active_callbacks;  /*!< List of just added active callbacks, that get transferred to the active_callbacks at an appropriate time */
  struct s_rc_async_timer_callback *active_callbacks;      /*!< List of active callbacks */
  struct s_rc_async_timer_callback *remove_callbacks;      /*!< List of active callbacks to be removed at a suitable time */
};


/**
 * Create an asynchronous timer
 *
 * @param rat is set to the created async timer on success
 * @param num_callbacks is the total number of callbacks supported by this timer (persistent and transient)
 */
enum e_rc_cli_result rc_async_timer_init(struct s_rc_async_timer **rat, size_t num_callbacks);


/** Destroy an asynchronous timer */
enum e_rc_cli_result rc_async_timer_destroy(struct s_rc_async_timer *rat);


/**
 * Call prior to select() to handle any timer timeouts that may have occurred and
 * return the next timeout period if there is one
 *
 * @param rat is the rincmd async timer
 * @param hastimeout is set if the there is a timeout. Not NULL
 * @param timeout is set to the timeout period. Not NULL
 */
enum e_rc_cli_result rc_async_timer_run(struct s_rc_async_timer *rat,
                                        bool *hastimeout, struct timespec *timeout);


/** Add a one shot timer to the async timer with a absolute timeout in CLOCK_MONOTONIC */
enum e_rc_cli_result rc_async_timer_add_oneshot_absolute(struct s_rc_async_timer *rat,
                                                         t_rc_async_timer_callback_fn callback, void *userdata,
                                                         const struct timespec *abstimeout, intptr_t *timer_id);


/** Add a one shot timer to the async timer with a relative timeout (from now) */
enum e_rc_cli_result rc_async_timer_add_oneshot_relative(struct s_rc_async_timer *rat,
                                                         t_rc_async_timer_callback_fn callback, void *userdata,
                                                         const struct timespec *timeout, intptr_t *timer_id);


/** Add a persistent timer to the async timer with a relative timeout */
enum e_rc_cli_result rc_async_timer_add_persistent(struct s_rc_async_timer *rat,
                                                   t_rc_async_timer_callback_fn callback, void *userdata,
                                                   const struct timespec *timeout, intptr_t *timer_id);


/** Remove a persistent timer (just put it in the removal list, to be removed by rc_async_timer_remove_active() at an appropriate time) */
enum e_rc_cli_result rc_async_timer_remove_callback(struct s_rc_async_timer *rat, intptr_t timer_id);


/**
 * Get timer (function pointer, user data pointer and timeout) referenced by
 * timer_id from the list of running timers
 *
 * @param rat is the async timer
 * @param timer_id is the timer identifier from rc_async_timer_add_*() functions
 * @param callback is a pointer to a async timer callback function pointer, set by the function when the matching timer_id is found
 * @param userdata is a pointer to a the user data pointer for the async timer, set by the function when the matching timer_id is found
 * @param abstimeout is a pointer to a timespec, set by the function to the next absolute time when the function will run
 *
 * @return a rincmd client result
 */
enum e_rc_cli_result rc_async_timer_get(struct s_rc_async_timer *rat, intptr_t timer_id,
                                        t_rc_async_timer_callback_fn *callback, void **userdata,
                                        struct timespec *abstimeout);


#endif /* RC_CLI_ASYNC_TIMER_H_ */
