/*
 * rc_cli_comms.h
 *
 *  Created on: Nov 16, 2017
 *      Author: ian
 */
#ifndef RC_CLI_COMMS_H_
#define RC_CLI_COMMS_H_

#include <unistd.h>
#include <stdio.h>
#include <stdbool.h>
#include <time.h>
#include <rincmd/rc_cli_error.h>


#define RC_CLI_MSG_MAXLEN 4096


/* Forward declaration */
struct s_rc_cli_comms;


/** Callback for a response to a async connect/disconnect event */
typedef void (*t_rc_async_connect_callback_fn)(void *userdata, bool connect_not_disconnect, enum e_rc_cli_result result);


/** Function pointer typedef to connect the communication channel using synchronous connect */
typedef enum e_rc_cli_result (*t_rc_comms_connect_sync_fn)(struct s_rc_cli_comms *cli_comms, const struct timespec *timeout);


/** Function pointer typedef to connect the communication channel using asynchronous connect */
typedef enum e_rc_cli_result (*t_rc_comms_connect_async_fn)(struct s_rc_cli_comms *cli_comms, t_rc_async_connect_callback_fn connect_callback, void *userdata);


/** Function pointer typedef to complete the connection of the communication channel using asynchronous connect */
typedef enum e_rc_cli_result (*t_rc_comms_complete_connect_async_fn)(struct s_rc_cli_comms *cli_comms, bool connected_not_timedout);


/** Function pointer typedef to disconnect the communication channel */
typedef enum e_rc_cli_result (*t_rc_comms_disconnect_fn)(struct s_rc_cli_comms *cli_comms);


/** Function pointer typedef to write to the communication channel */
typedef enum e_rc_cli_result (*t_rc_comms_write_fn)(struct s_rc_cli_comms *cli_comms, const void *data, ssize_t *len);


/** Function pointer typedef to read from the communication channel */
typedef enum e_rc_cli_result (*t_rc_comms_read_fn)(struct s_rc_cli_comms *rc_cli_comms, void *data, ssize_t *len);


/** Function pointer typedef to get the file descriptors for the communication channel */
typedef enum e_rc_cli_result (*t_rc_comms_getfds_fn)(struct s_rc_cli_comms *rc_cli_comms, int *readfd, int *connectfd);


/** Function pointer typedef to destroy the communication channel */
typedef enum e_rc_cli_result (*t_rc_comms_destroy_fn)(struct s_rc_cli_comms *cli_comms);


/** Client communications structure */
struct s_rc_cli_comms
{
  void *impl_data;                                                    /*!< Implementation specific data, passed to the function pointers */
  FILE *comms_log;                                                    /*!< Log file for writing all communications to */
  t_rc_comms_connect_sync_fn comms_connect_sync;                      /*!< Function pointer called to connect communications channel for sync client */
  t_rc_comms_connect_async_fn comms_connect_async;                    /*!< Function pointer called to connect communications channel for async client */
  t_rc_comms_complete_connect_async_fn comms_complete_connect_async;  /*!< Function pointer called to complete the connection for async client */
  t_rc_comms_disconnect_fn comms_disconnect;                          /*!< Function pointer called to disconnect the communications channel */
  t_rc_comms_write_fn comms_write;                                    /*!< Function pointer called to write to the communications channel */
  t_rc_comms_read_fn comms_read;                                      /*!< Function pointer called to read from the communications channel */
  t_rc_comms_getfds_fn comms_getfds;                                  /*!< Function pointer called to get the file descriptors for the communication channel */
  t_rc_comms_destroy_fn comms_destroy;                                /*!< Function pointer called to shutdown and destroy the communications channel */
};


/** Put an entry in the comms log if one is configured */
void rc_cli_comms_log(struct s_rc_cli_comms *rc_cli_comms, bool in_not_out, const void *data, ssize_t len);


/** Connect to the rincmd host to retrieve data from using synchronous connect */
enum e_rc_cli_result rc_cli_comms_connect_sync(struct s_rc_cli_comms *rc_cli_comms, const struct timespec *timeout);


/** Connect to the rincmd host to retrieve data from using asynchronous connect */
enum e_rc_cli_result rc_cli_comms_connect_async(struct s_rc_cli_comms *rc_cli_comms, t_rc_async_connect_callback_fn connect_callback, void *userdata);


/** Complete connection to the rincmd host to retrieve data from using asynchronous connect */
enum e_rc_cli_result rc_cli_comms_complete_connect_async(struct s_rc_cli_comms *rc_cli_comms, bool connected_not_timedout);


/** Disconnect from the rincmd host */
enum e_rc_cli_result rc_cli_comms_disconnect(struct s_rc_cli_comms *rc_cli_comms);


/** Write data to the rincmd host */
enum e_rc_cli_result rc_cli_comms_write(struct s_rc_cli_comms *rc_cli_comms, const void *data, ssize_t *len);


/** Read data from the rincmd host */
enum e_rc_cli_result rc_cli_comms_read(struct s_rc_cli_comms *rc_cli_comms, void *data, ssize_t *len);


/** Get the file descriptors for the communication channel */
enum e_rc_cli_result rc_cli_comms_getfds(struct s_rc_cli_comms *rc_cli_comms, int *readfd, int *connectfd);


/** Create a rincmd client structure. Don't call directly, called by rc_comms_init_tcpudp() etc */
enum e_rc_cli_result rc_cli_comms_init(struct s_rc_cli_comms **rc_cli_comms, FILE *comms_log);


/** Create a rincmd client TCP connection for use with hostname:port */
enum e_rc_cli_result rc_comms_init_tcp(struct s_rc_cli_comms **rcc,
                                       const char *hostname, int port,
                                       FILE *comms_log);


/** Create a rincmd client UDP connection for use with hostname:port */
enum e_rc_cli_result rc_comms_init_udp(struct s_rc_cli_comms **rcc,
                                       const char *hostname, int port, int localudpport,
                                       FILE *comms_log);


/** Create a rincmd client UDS connection for use with udsfile */
enum e_rc_cli_result rc_comms_init_uds(struct s_rc_cli_comms **rcc,
                                       const char *udsfile,
                                       FILE *comms_log);


/** Create a rincmd client serial connection for use with device (eg /dev/ttyUSB0) */
enum e_rc_cli_result rc_comms_init_serial(struct s_rc_cli_comms **rcc,
                                          const char *device, unsigned baud_rate, char parity, unsigned data_bits, unsigned stop_bits,
                                          FILE *comms_log);


/** Clean up a rincmd client comms */
enum e_rc_cli_result rc_cli_comms_destroy(struct s_rc_cli_comms *cli_comms);


#endif /* RC_CLI_COMMS_H_ */
