/*
 * rc_cli_comms_serial.c
 *
 *  Created on: 8 Jan. 2019
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <stdbool.h>
#include <fcntl.h>
#include <termios.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <rincmd/rc_cli_comms.h>


/** Special state information required for doing async connects */
struct s_rc_cli_comms_serial_async
{
  t_rc_async_connect_callback_fn connect_callback;            /*!< Callback for when we connect/timeout/error */
  void *userdata;                                             /*!< User data pointer for connect_callback */
};


/** Client communications structure for serial ports */
struct s_rc_cli_comms_serial
{
  char *device;                                               /*!< Serial port device (eg /dev/ttyUSB0) */
  unsigned baud_rate;                                         /*!< Baud rate to connect at */
  char parity;                                                /*!< Parity setting (N, O, E) */
  unsigned data_bits;                                         /*!< Number of data bits */
  unsigned stop_bits;                                         /*!< Number of stop bits */
  int serialfd;                                               /*!< Serial port file descriptor */
  struct s_rc_cli_comms_serial_async async_connect;           /*!< Special state for handling async connection */
};


struct s_baud_conversion
{
  speed_t linux_baud;
  unsigned int actual_baud;
};


/* #define NUM_BAUD_CONVERSIONS 30 */
#define NUM_BAUD_CONVERSIONS 17
static const struct s_baud_conversion baud_conversions[NUM_BAUD_CONVERSIONS] =
{
  { B50, 50 },
  { B75, 75 },
  { B110, 110 },
  { B134, 134 },
  { B150, 150 },
  { B200, 200 },
  { B300, 300 },
  { B600, 600 },
  { B1200, 1200 },
  { B1800, 1800 },
  { B2400, 2400 },
  { B4800, 4800 },
  { B9600, 9600 },
  { B19200, 19200 },
  { B38400, 38400 },
  { B57600, 57600 },
  { B115200, 115200 },
#if 0 /* These are #if'd out in uCLibc, or at least the version we have. We don't need them so far, so I've just commented them out for now. */
  { B230400, 230400 },
  { B460800, 460800 },
  { B500000, 500000 },
  { B576000, 576000 },
  { B921600, 921600 },
  { B1000000, 1000000 },
  { B1152000, 1152000 },
  { B1500000, 1500000 },
  { B2000000, 2000000 },
  { B2500000, 2500000 },
  { B3000000, 3000000 },
  { B3500000, 3500000 },
  { B4000000, 4000000 }
#endif
};


static speed_t lookup_linux_baud_rate(unsigned baud_rate)
{
  size_t i;

  for (i = 0; i < NUM_BAUD_CONVERSIONS; i++)
  {
    if (baud_conversions[i].actual_baud == baud_rate)
    {
      return baud_conversions[i].linux_baud;
    }
  }
  return B9600;
}


/** Call the callback to notify of connect/disconnect */
static void rc_comms_async_connect_serial_callback(const struct s_rc_cli_comms_serial *serialcomms, bool connect_not_disconnect, enum e_rc_cli_result result)
{
  if (NULL != serialcomms->async_connect.connect_callback)
    serialcomms->async_connect.connect_callback(serialcomms->async_connect.userdata, connect_not_disconnect, result);
}


/** Initiate a connect to the rincomms serial port to retrieve data from using asynchronous connect */
static enum e_rc_cli_result rc_comms_connect_serial_async(struct s_rc_cli_comms *cli_comms,
                                                          t_rc_async_connect_callback_fn connect_callback,
                                                          void *userdata)
{
  struct s_rc_cli_comms_serial *serialcomms = cli_comms->impl_data;
  struct termios options;
  speed_t linuxbaud;
  int flags;

  serialcomms->async_connect.connect_callback = connect_callback;
  serialcomms->async_connect.userdata = userdata;

  /* Use O_NDELAY to open the port without blocking for CTS? or something */
  serialcomms->serialfd = open(serialcomms->device, O_RDWR | O_NOCTTY | O_NDELAY);
  if (serialcomms->serialfd < 0)
    goto rcacsic_fail;
  if (fcntl(serialcomms->serialfd, F_SETFD, FD_CLOEXEC) < 0)
    goto rcacsic_fail;

  /* Now setup blocking reads and writes */
  flags = fcntl(serialcomms->serialfd, F_GETFL);
  if (flags < 0)
    goto rcacsic_fail;
  flags &= ~O_NONBLOCK;
  if (fcntl(serialcomms->serialfd, F_SETFL, flags) < 0)
    goto rcacsic_fail;

  /* Now configure port */

  /* Zero out settings */
  memset(&options, 0, sizeof(options));

  /* Set the baud rates */
  linuxbaud = lookup_linux_baud_rate(serialcomms->baud_rate);
  cfsetispeed(&options, linuxbaud);
  cfsetospeed(&options, linuxbaud);

  /* Set the data bits */
  options.c_cflag &= ~CSIZE;
  if (5 == serialcomms->data_bits)
    options.c_cflag |= CS5;
  else if (6 == serialcomms->data_bits)
    options.c_cflag |= CS6;
  else if (7 == serialcomms->data_bits)
    options.c_cflag |= CS7;
  else if (8 == serialcomms->data_bits)
    options.c_cflag |= CS8;
  else
    goto rcacsic_fail;

  /* Set the stop bits */
  if (2 == serialcomms->stop_bits)
    options.c_cflag |= CSTOPB;  /* 2 Stop bits */
  else if (1 == serialcomms->stop_bits)
    options.c_cflag &= ~CSTOPB; /* 1 Stop bits */
  else
    goto rcacsic_fail;

  /* Set the parity bits */
  if ('O' == toupper(serialcomms->parity))  /* Odd */
  {
    options.c_cflag |= PARENB;
    options.c_cflag |= PARODD;
  }
  else if ('E' == toupper(serialcomms->parity)) /* Even */
  {
    options.c_cflag |= PARENB;
    options.c_cflag &= ~PARODD;
  }
  else if ('N' == toupper(serialcomms->parity)) /* No parity */
  {
    options.c_cflag &= ~PARENB;
    options.c_cflag &= ~PARODD;
  }
  else
  {
    goto rcacsic_fail;
  }

  /* Enable the receiver and set local mode */
  options.c_cflag |= (CLOCAL | CREAD);

  /* Ignore parity ??? */
  options.c_iflag = IGNPAR;
  options.c_oflag = 0;

  options.c_cc[VTIME] = 0; /* No Time delay */
  options.c_cc[VMIN] = 1;  /* Blocking read until 1 character arrives */

  /* CTS handshaking */
  /* options.c_cflag |= CRTSCTS; */

  /* Set the new options for the port */
  if (tcsetattr(serialcomms->serialfd, TCSANOW, &options) < 0)
    goto rcacsic_fail;

  rc_comms_async_connect_serial_callback(serialcomms, true, rcr_success);
  return rcr_success;

rcacsic_fail:
  if (serialcomms->serialfd >= 0)
  {
    close(serialcomms->serialfd);
    serialcomms->serialfd = -1;
  }
  rc_comms_async_connect_serial_callback(serialcomms, true, rcr_connectfail);
  return rcr_connectfail;
}


/** Complete a connection to the rincomms serial device to retrieve data from using asynchronous connect */
static enum e_rc_cli_result rc_comms_complete_connect_serial_async(struct s_rc_cli_comms *cli_comms,
                                                                   bool connected_not_timedout)
{
  /* Nothing to do here, device already open */
  return rcr_success;
}


/** Connect to the rincomms host to retrieve data from */
static enum e_rc_cli_result rc_comms_connect_serial_sync(struct s_rc_cli_comms *cli_comms, const struct timespec *timeout)
{
  enum e_rc_cli_result result;

  /* Open the serial port */
  result = rc_comms_connect_serial_async(cli_comms, NULL, NULL);

  (void) rc_comms_complete_connect_serial_async(cli_comms, true);

  /* Return result */
  return result;
}


/** Disconnect a serial communication channel */
static enum e_rc_cli_result rc_comms_disconnect_serial(struct s_rc_cli_comms *cli_comms)
{
  struct s_rc_cli_comms_serial *serialcomms = cli_comms->impl_data;
  bool wasopen = false;

  if (serialcomms->serialfd >= 0)
  {
    close(serialcomms->serialfd);
    serialcomms->serialfd = -1;
    wasopen = true;
  }
  if (wasopen)
    rc_comms_async_connect_serial_callback(serialcomms, false, rcr_success);
  return rcr_success;
}


/** Write to the serial communications channel */
static enum e_rc_cli_result rc_comms_write_serial(struct s_rc_cli_comms *cli_comms, const void *data, ssize_t *len)
{
  ssize_t written;
  struct s_rc_cli_comms_serial *serialcomms = cli_comms->impl_data;

  written = write(serialcomms->serialfd, data, *len);
  if (written < 0)
    return rcr_writefail;
  *len = written;
  return rcr_success;
}


/** Read from the serial communications channel */
static enum e_rc_cli_result rc_comms_read_serial(struct s_rc_cli_comms *cli_comms, void *data, ssize_t *len)
{
  ssize_t bytesread;
  struct s_rc_cli_comms_serial *serialcomms = cli_comms->impl_data;

  bytesread = read(serialcomms->serialfd, data, *len);
  if (bytesread <= 0)
    return rcr_readfail;
  *len = bytesread;
  return rcr_success;
}


/** Get the file descriptors for a serial communication channel */
static enum e_rc_cli_result rc_comms_getfds_serial(struct s_rc_cli_comms *cli_comms, int *readfd, int *connectfd)
{
  struct s_rc_cli_comms_serial *serialcomms = cli_comms->impl_data;

  *readfd = serialcomms->serialfd;
  *connectfd = -1;
  return rcr_success;
}


/** Destroy the serial communication channel */
static enum e_rc_cli_result rc_comms_destroy_serial(struct s_rc_cli_comms *cli_comms)
{
  struct s_rc_cli_comms_serial *serialcomms;

  if (NULL != cli_comms->impl_data)
  {
    serialcomms = cli_comms->impl_data;
    if (NULL != serialcomms)
    {
      free(serialcomms->device);
      serialcomms->device = NULL;

      if (serialcomms->serialfd >= 0)
      {
        close(serialcomms->serialfd);
        serialcomms->serialfd = -1;
      }

      free(serialcomms);
      cli_comms->impl_data = NULL;
    }
  }
  return rcr_success;
}


/** Create a rincmd client serial connection for use with device (eg /dev/ttyUSB0) */
enum e_rc_cli_result rc_comms_init_serial(struct s_rc_cli_comms **rcc,
                                          const char *device, unsigned baud_rate, char parity, unsigned data_bits, unsigned stop_bits,
                                          FILE *comms_log)
{
  struct s_rc_cli_comms *ircc = NULL;
  struct s_rc_cli_comms_serial *serialcomms = NULL;
  enum e_rc_cli_result result;

  if (NULL == rcc || NULL == device)
  {
    result = rcr_paramfail;
    goto rcis_fail;
  }

  /* Allocate comms structure */
  result = rc_cli_comms_init(&ircc, comms_log);
  if (rcr_success != result)
  {
    goto rcis_fail;
  }

  /* Allocate and fill out serial structure */
  serialcomms = calloc(1, sizeof(*serialcomms));
  if (NULL == serialcomms)
  {
    result = rcr_allocfail;
    goto rcis_fail;
  }
  serialcomms->device = strdup(device);
  if (NULL == serialcomms->device)
  {
    result = rcr_allocfail;
    goto rcis_fail;
  }
  serialcomms->baud_rate = baud_rate;
  serialcomms->parity = parity;
  serialcomms->data_bits = data_bits;
  serialcomms->stop_bits = stop_bits;
  serialcomms->serialfd = -1;
  serialcomms->async_connect.connect_callback = NULL;
  serialcomms->async_connect.userdata = NULL;

  /* Fill out rcc data */
  ircc->impl_data = serialcomms;
  ircc->comms_connect_sync = rc_comms_connect_serial_sync;
  ircc->comms_connect_async = rc_comms_connect_serial_async;
  ircc->comms_complete_connect_async = rc_comms_complete_connect_serial_async;
  ircc->comms_disconnect = rc_comms_disconnect_serial;
  ircc->comms_write = rc_comms_write_serial;
  ircc->comms_read = rc_comms_read_serial;
  ircc->comms_getfds = rc_comms_getfds_serial;
  ircc->comms_destroy = rc_comms_destroy_serial;

  *rcc = ircc;
  return rcr_success;

rcis_fail:
  if (NULL != ircc)
  {
    free(ircc);
  }
  if (NULL != serialcomms)
  {
    if (NULL != serialcomms->device)
    {
      free(serialcomms->device);
      serialcomms->device = NULL;
    }
    free(serialcomms);
  }

  return result;
}
