/*
 * rc_cli_comms_tcp.c
 *
 *  Created on: Nov 16, 2017
 *      Author: ian
 */
#define _XOPEN_SOURCE 700
#include <stdio.h>
#include <stdbool.h>
#include <netdb.h>
#include <netinet/tcp.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/un.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <rincmd/rc_strncpy.h>

#include <rincmd/rc_cli_comms.h>


/** States within the asynchronous connect state machine */
enum e_rc_comms_async_connect_state
{
  rcacs_init,
  rcacs_initiate_connect,
  rcacs_await_connect,
  rcacs_connected,
  rcacs_failed
};


/** Special state information required for doing async connects */
struct s_rc_cli_comms_tcpudp_async
{
  enum e_rc_comms_async_connect_state async_connect_state;    /*!< Current async connection establishment state */
  struct addrinfo *dstinfo;                                   /*!< Destination info (used for asynchronous connects) */
  struct addrinfo *current_dstinfo;                           /*!< Current destination info */
  int connect_sockfd;                                         /*!< Socket to connect with */
  t_rc_async_connect_callback_fn connect_callback;            /*!< Callback for when we connect/timeout/error */
  void *userdata;                                             /*!< User data pointer for connect_callback */
  bool timedout;                                              /*!< Select call timed out */
  unsigned uds_connects;                                      /*!< Number of times the UDS has tried to connect (0 or 1) */
};


/** Client communications structure for TCP and UDP */
struct s_rc_cli_comms_tcpudp
{
  char *hostname;                                             /*!< Remote hostname to connect to for AF_INET, or pathname for AF_UNIX */
  int port;                                                   /*!< Remote port to connect to (AF_INET), not used for AF_UNIX */
  int sockfd;                                                 /*!< Socket file descriptor */
  int sockdomain;                                             /*!< Socket domain (AF_UNIX for unix domain socket, AF_INET for IPv4) */
  int socktype;                                               /*!< Socket type (SOCK_STREAM for TCP, SOCK_DGRAM for UDP) */
  int localudpport;                                           /*!< Local UDP port to bind to for AF_INET:SOCK_DGRAM */
  struct s_rc_cli_comms_tcpudp_async async_connect;           /*!< Special state for handling async connection */
};


/** Call the callback to notify of connect/disconnect */
static void rc_comms_async_connect_tcpudp_callback(struct s_rc_cli_comms_tcpudp *tcpcomms, bool connect_not_disconnect, enum e_rc_cli_result result)
{
  if (NULL != tcpcomms->async_connect.connect_callback)
    tcpcomms->async_connect.connect_callback(tcpcomms->async_connect.userdata, connect_not_disconnect, result);
}


/** Call the callback and then clean up the dstinfo */
static void rc_comms_async_connect_tcpudp_callback_cleanup(struct s_rc_cli_comms_tcpudp *tcpcomms,
                                                           enum e_rc_comms_async_connect_state connect_state,
                                                           bool connect_not_disconnect, enum e_rc_cli_result result)
{
  tcpcomms->async_connect.async_connect_state = connect_state;
  rc_comms_async_connect_tcpudp_callback(tcpcomms, connect_not_disconnect, result);
  if (NULL != tcpcomms->async_connect.dstinfo)
  {
    freeaddrinfo(tcpcomms->async_connect.dstinfo);
    tcpcomms->async_connect.dstinfo = NULL;
    tcpcomms->async_connect.current_dstinfo = NULL;
  }
}


static void rc_comms_async_connect_tcpudp_init(struct s_rc_cli_comms_tcpudp *tcpcomms)
{
  struct addrinfo hints;
  struct addrinfo *dstinfo = NULL;
  int rv;
  char portstr[20];

  /* Use getaddrinfo() for the destination address */
  snprintf(portstr, 20, "%u", tcpcomms->port);
  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;             /* Allow IPv4 or IPv6 */
  hints.ai_socktype = tcpcomms->socktype;  /* UDP or TCP as specified */
  hints.ai_flags = 0;
  hints.ai_protocol = 0;                   /* Any protocol */
  rv = getaddrinfo(tcpcomms->hostname, portstr, &hints, &dstinfo);
  if (rv != 0)
  {
    /* fprintf(stderr, "rincomms_connect: getaddrinfo: %s\n", gai_strerror(rv)); */
    tcpcomms->async_connect.async_connect_state = rcacs_failed;
    rc_comms_async_connect_tcpudp_callback(tcpcomms, true, rcr_connectfail);
    return;
  }
  tcpcomms->async_connect.dstinfo = dstinfo;
  tcpcomms->async_connect.current_dstinfo = tcpcomms->async_connect.dstinfo;
  tcpcomms->async_connect.async_connect_state = rcacs_initiate_connect;
}


static void rc_comms_async_connect_uds_init(struct s_rc_cli_comms_tcpudp *tcpcomms)
{
  tcpcomms->async_connect.dstinfo = NULL;
  tcpcomms->async_connect.current_dstinfo = NULL;
  tcpcomms->async_connect.async_connect_state = rcacs_initiate_connect;
  tcpcomms->async_connect.uds_connects = 0;
}


static void rc_comms_async_connect_tcpudp_init_connect(struct s_rc_cli_comms_tcpudp *tcpcomms)
{
  struct addrinfo hints;
  struct addrinfo *srcinfo = NULL;
  int rv;
  char portstr[20];
  int sockfd;

  if (NULL == tcpcomms->async_connect.current_dstinfo)
  {
    /* Last one, return error */
    rc_comms_async_connect_tcpudp_callback_cleanup(tcpcomms, rcacs_failed, true, rcr_connectfail);
    return;
  }

  /* Establish socket */
  sockfd = socket(tcpcomms->async_connect.current_dstinfo->ai_family,
                  tcpcomms->async_connect.current_dstinfo->ai_socktype,
                  tcpcomms->async_connect.current_dstinfo->ai_protocol);
  if (sockfd < 0)
    goto rcactic_next;

  /* If we are using UDP with a specified local port, we need to bind to that local port */
  if (SOCK_DGRAM == tcpcomms->socktype && 0 != tcpcomms->localudpport)
  {
    /* Use getaddrinfo() for the source address */
    snprintf(portstr, 20, "%u", tcpcomms->localudpport);
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;        /* Allow IPv4 or IPv6 */
    hints.ai_socktype = SOCK_DGRAM;     /* UDP communication */
    hints.ai_flags = AI_PASSIVE;        /* Fill in my IP for me */
    rv = getaddrinfo(NULL, portstr, &hints, &srcinfo);
    if (0 != rv)
    {
      /* fprintf(stderr, "getaddrinfo for src address: %s\n", gai_strerror(rv)); */
      close(sockfd);
      goto rcactic_freesrcinfo;
    }

    /* Bind this datagram socket to source address info */
    rv = bind(sockfd, srcinfo->ai_addr, srcinfo->ai_addrlen);
    if (rv != 0)
    {
      /* fprintf(stderr, "bind: %s\n", gai_strerror(rv)); */
      close(sockfd);
      goto rcactic_freesrcinfo;
    }
  }

  /*
   * Now connect to the destination with timeout.
   * We do this by setting the socket to non-blocking
   * and use select on the socket
   */
  if (fcntl(sockfd, F_SETFL, O_NONBLOCK) != 0)
  {
    close(sockfd);
    goto rcactic_freesrcinfo;
  }
  rv = connect(sockfd, tcpcomms->async_connect.current_dstinfo->ai_addr, tcpcomms->async_connect.current_dstinfo->ai_addrlen);
  if (rv != 0 && EINPROGRESS != errno)
  {
    close(sockfd);
    goto rcactic_freesrcinfo;
  }

  tcpcomms->async_connect.connect_sockfd = sockfd;
  tcpcomms->async_connect.async_connect_state = rcacs_await_connect;

rcactic_freesrcinfo:
  freeaddrinfo(srcinfo);

rcactic_next:
  tcpcomms->async_connect.current_dstinfo = tcpcomms->async_connect.current_dstinfo->ai_next;
}


static void rc_comms_async_connect_uds_init_connect(struct s_rc_cli_comms_tcpudp *tcpcomms)
{
  int rv;
  int sockfd;
  struct sockaddr_un uds_addr;

  if (tcpcomms->async_connect.uds_connects >= 1)
  {
    /* Last one, return error */
    rc_comms_async_connect_tcpudp_callback_cleanup(tcpcomms, rcacs_failed, true, rcr_connectfail);
    return;
  }

  /* Establish socket */
  sockfd = socket(tcpcomms->sockdomain, tcpcomms->socktype, 0);
  if (sockfd < 0)
    goto rcacuic_fail;

  /*
   * Now connect to the destination with timeout.
   * We do this by setting the socket to non-blocking
   * and use select on the socket
   */
  if (fcntl(sockfd, F_SETFL, O_NONBLOCK) != 0)
  {
    close(sockfd);
    goto rcacuic_fail;
  }

  /* Setup UDS address to connect to (handling Linux abstract paths) */
  memset(&uds_addr, 0, sizeof(uds_addr));
  uds_addr.sun_family = AF_UNIX;
  if ('\0' == tcpcomms->hostname[0])
  {
    /* Linux abstract path starts with '\0' */
    uds_addr.sun_path[0] = '\0';
    rc_safe_strncpy(&uds_addr.sun_path[1], &tcpcomms->hostname[1], sizeof(uds_addr.sun_path) - 2);
  }
  else
  {
    rc_safe_strncpy(uds_addr.sun_path, tcpcomms->hostname, sizeof(uds_addr.sun_path) - 1);
  }

  rv = connect(sockfd, (struct sockaddr *) &uds_addr, sizeof(uds_addr));
  if (rv != 0 && EINPROGRESS != errno)
  {
    close(sockfd);
    goto rcacuic_fail;
  }

  tcpcomms->async_connect.connect_sockfd = sockfd;
  tcpcomms->async_connect.async_connect_state = rcacs_await_connect;
  tcpcomms->async_connect.uds_connects++;
  return;

rcacuic_fail:
  rc_comms_async_connect_tcpudp_callback_cleanup(tcpcomms, rcacs_failed, true, rcr_connectfail);
}


/** Run connection after select() completes */
static void rc_comms_async_connect_tcpudp_await_connect(struct s_rc_cli_comms_tcpudp *tcpcomms)
{
  int flag;
  socklen_t flaglen;
  int rv;

  if (tcpcomms->async_connect.timedout)
  {
    /* Last connect timed out */
    close(tcpcomms->async_connect.connect_sockfd);
    tcpcomms->async_connect.connect_sockfd = -1;
    tcpcomms->async_connect.async_connect_state = rcacs_initiate_connect;
    return;
  }
  else
  {
    /* Now check we are connected */
    flaglen = sizeof(flag);
    rv = getsockopt(tcpcomms->async_connect.connect_sockfd, SOL_SOCKET, SO_ERROR, &flag, &flaglen);
    if (rv < 0 || flag != 0)
    {
      close(tcpcomms->async_connect.connect_sockfd);
      tcpcomms->async_connect.connect_sockfd = -1;
      tcpcomms->async_connect.async_connect_state = rcacs_initiate_connect;
      return;
    }

    /* Connected to host */
    if (AF_INET == tcpcomms->sockdomain && SOCK_STREAM == tcpcomms->socktype)
    {
      /* Set TCP no delay on TCP streams only */
      flag = 1;
      rv = setsockopt(tcpcomms->async_connect.connect_sockfd, IPPROTO_TCP, TCP_NODELAY, &flag, sizeof(int));
      if (rv < 0)
      {
        close(tcpcomms->async_connect.connect_sockfd);
        tcpcomms->async_connect.connect_sockfd = -1;
        tcpcomms->async_connect.async_connect_state = rcacs_initiate_connect;
        return;
      }
    }
    tcpcomms->sockfd = tcpcomms->async_connect.connect_sockfd;
    tcpcomms->async_connect.connect_sockfd = -1;
    rc_comms_async_connect_tcpudp_callback_cleanup(tcpcomms, rcacs_connected, true, rcr_success);
  }
}


/** Run the state machine to connect to the rincomms host to retrieve data from */
static enum e_rc_cli_result rc_comms_connect_tcpudp_async_state_machine(struct s_rc_cli_comms *cli_comms)
{
  struct s_rc_cli_comms_tcpudp *tcpcomms = cli_comms->impl_data;

  switch (tcpcomms->async_connect.async_connect_state)
  {
    /* Get the list of destination addresses */
    case rcacs_init:
      if (AF_UNIX != tcpcomms->sockdomain)
        rc_comms_async_connect_tcpudp_init(tcpcomms);
      else
        rc_comms_async_connect_uds_init(tcpcomms);
      goto rcact_initiate_connect;
      break;

    /* Try to connect on all the returned addresses */
    case rcacs_initiate_connect:
rcact_initiate_connect:
      while (rcacs_initiate_connect == tcpcomms->async_connect.async_connect_state)
      {
        if (AF_UNIX != tcpcomms->sockdomain)
          rc_comms_async_connect_tcpudp_init_connect(tcpcomms);
        else
          rc_comms_async_connect_uds_init_connect(tcpcomms);
      }
      break;

    /* After select() returns */
    case rcacs_await_connect:
      rc_comms_async_connect_tcpudp_await_connect(tcpcomms);
      if (rcacs_initiate_connect == tcpcomms->async_connect.async_connect_state)
        goto rcact_initiate_connect;
      break;

    /* Connected, nothing to do */
    case rcacs_connected:
      break;

    /* Failed, nothing to do */
    case rcacs_failed:
      break;

    default:
      break;
  }

  return rcr_success;
}


/** Initiate a connect to the rincomms host to retrieve data from using asynchronous connect */
static enum e_rc_cli_result rc_comms_connect_tcpudp_async(struct s_rc_cli_comms *cli_comms,
                                                          t_rc_async_connect_callback_fn connect_callback,
                                                          void *userdata)
{
  struct s_rc_cli_comms_tcpudp *tcpcomms = cli_comms->impl_data;

  tcpcomms->async_connect.async_connect_state = rcacs_init;
  tcpcomms->async_connect.dstinfo = NULL;
  tcpcomms->async_connect.current_dstinfo = NULL;
  tcpcomms->async_connect.connect_sockfd = -1;
  tcpcomms->async_connect.connect_callback = connect_callback;
  tcpcomms->async_connect.userdata = userdata;
  tcpcomms->async_connect.timedout = false;
  tcpcomms->async_connect.uds_connects = 0;

  return rc_comms_connect_tcpudp_async_state_machine(cli_comms);
}


/** Complete a connection to the rincomms host to retrieve data from using asynchronous connect */
static enum e_rc_cli_result rc_comms_complete_connect_tcpudp_async(struct s_rc_cli_comms *cli_comms,
                                                                   bool connected_not_timedout)
{
  struct s_rc_cli_comms_tcpudp *tcpcomms = cli_comms->impl_data;

  tcpcomms->async_connect.timedout = !connected_not_timedout;
  return rc_comms_connect_tcpudp_async_state_machine(cli_comms);
}


/** Connect to the rincomms host to retrieve data from */
static enum e_rc_cli_result rc_comms_connect_tcpudp_sync(struct s_rc_cli_comms *cli_comms, const struct timespec *timeout)
{
  struct s_rc_cli_comms_tcpudp *tcpcomms = cli_comms->impl_data;
  fd_set writefds;
  struct timespec to;
  struct timespec *toptr;
  int sockfd;
  int select_res;

  /* Initiate the connection */
  (void) rc_comms_connect_tcpudp_async(cli_comms, NULL, NULL);

  /* Go through any addresses for this host */
  while (rcacs_connected != tcpcomms->async_connect.async_connect_state &&
         rcacs_failed != tcpcomms->async_connect.async_connect_state)
  {
    /* Wait for non-blocking connect to succeed or timeout */
    sockfd = tcpcomms->async_connect.connect_sockfd;
    FD_ZERO(&writefds);
    FD_SET(sockfd, &writefds);
    if (NULL != timeout)
    {
      to = *timeout;
      toptr = &to;
    }
    else
    {
      toptr = NULL;
    }
    select_res = pselect(sockfd + 1, NULL, &writefds, NULL, toptr, NULL);

    /* Update state machine with result */
    (void) rc_comms_complete_connect_tcpudp_async(cli_comms, (1 == select_res));
  }

  /* Return result */
  return (rcacs_connected == tcpcomms->async_connect.async_connect_state) ? rcr_success : rcr_connectfail;
}


/** Disconnect a TCP/UDP communication channel */
static enum e_rc_cli_result rc_comms_disconnect_tcpudp(struct s_rc_cli_comms *cli_comms)
{
  struct s_rc_cli_comms_tcpudp *tcpcomms = cli_comms->impl_data;
  bool wasopen = false;

  if (tcpcomms->sockfd >= 0)
  {
    close(tcpcomms->sockfd);
    tcpcomms->sockfd = -1;
    wasopen = true;
  }
  if (tcpcomms->async_connect.connect_sockfd >= 0)
  {
    close(tcpcomms->async_connect.connect_sockfd);
    tcpcomms->async_connect.connect_sockfd = -1;
    wasopen = true;
  }
  tcpcomms->async_connect.async_connect_state = rcacs_init;
  if (wasopen)
    rc_comms_async_connect_tcpudp_callback(tcpcomms, false, rcr_success);
  return rcr_success;
}


/** Write to the TCP communications channel */
static enum e_rc_cli_result rc_comms_write_tcpudp(struct s_rc_cli_comms *cli_comms, const void *data, ssize_t *len)
{
  ssize_t written;
  struct s_rc_cli_comms_tcpudp *tcpcomms = cli_comms->impl_data;

  written = send(tcpcomms->sockfd, data, *len, 0);
  if (written < 0)
    return rcr_writefail;
  *len = written;
  return rcr_success;
}


/** Read from the TCP communications channel */
static enum e_rc_cli_result rc_comms_read_tcpudp(struct s_rc_cli_comms *cli_comms, void *data, ssize_t *len)
{
  ssize_t bytesread;
  struct s_rc_cli_comms_tcpudp *tcpcomms = cli_comms->impl_data;

  bytesread = recv(tcpcomms->sockfd, data, *len, 0);
  if (bytesread <= 0)
    return rcr_readfail;
  *len = bytesread;
  return rcr_success;
}


/** Get the file descriptors for a TCP/UDP communication channel */
static enum e_rc_cli_result rc_comms_getfds_tcpudp(struct s_rc_cli_comms *cli_comms, int *readfd, int *connectfd)
{
  struct s_rc_cli_comms_tcpudp *tcpcomms = cli_comms->impl_data;

  *readfd = tcpcomms->sockfd;
  *connectfd = tcpcomms->async_connect.connect_sockfd;
  return rcr_success;
}


/** Destroy the TCP communication channel */
static enum e_rc_cli_result rc_comms_destroy_tcpudp(struct s_rc_cli_comms *cli_comms)
{
  struct s_rc_cli_comms_tcpudp *tcpcomms;

  if (NULL != cli_comms->impl_data)
  {
    tcpcomms = cli_comms->impl_data;
    if (NULL != tcpcomms)
    {
      free(tcpcomms->hostname);
      tcpcomms->hostname = NULL;

      if (tcpcomms->sockfd >= 0)
      {
        close(tcpcomms->sockfd);
        tcpcomms->sockfd = -1;
      }

      free(tcpcomms);
      cli_comms->impl_data = NULL;
    }
  }
  return rcr_success;
}


/** Create a rincmd client TCP or UDP connection for use with hostname:port */
static enum e_rc_cli_result rc_comms_init_tcpudp(struct s_rc_cli_comms **rcc,
                                                 const char *hostname, int port, int sockdomain, int socktype, int localudpport,
                                                 FILE *comms_log)
{
  struct s_rc_cli_comms *ircc = NULL;
  struct s_rc_cli_comms_tcpudp *tcpudp = NULL;
  enum e_rc_cli_result result;

  if (NULL == rcc || NULL == hostname)
  {
    result = rcr_paramfail;
    goto rcit_fail;
  }

  /* Allocate comms structure */
  result = rc_cli_comms_init(&ircc, comms_log);
  if (rcr_success != result)
  {
    goto rcit_fail;
  }

  /* Allocate and fill out TCP structure */
  tcpudp = calloc(1, sizeof(*tcpudp));
  if (NULL == tcpudp)
  {
    result = rcr_allocfail;
    goto rcit_fail;
  }
  tcpudp->hostname = strdup(hostname);
  if (NULL == tcpudp->hostname)
  {
    result = rcr_allocfail;
    goto rcit_fail;
  }
  tcpudp->port = port;
  tcpudp->sockfd = -1;
  tcpudp->sockdomain = sockdomain;
  tcpudp->socktype = socktype;
  tcpudp->localudpport = localudpport;
  tcpudp->async_connect.async_connect_state = rcacs_init;
  tcpudp->async_connect.dstinfo = NULL;
  tcpudp->async_connect.current_dstinfo = NULL;
  tcpudp->async_connect.connect_sockfd = -1;
  tcpudp->async_connect.connect_callback = NULL;
  tcpudp->async_connect.userdata = NULL;
  tcpudp->async_connect.timedout = false;
  tcpudp->async_connect.uds_connects = 0;

  /* Fill out rcc data */
  ircc->impl_data = tcpudp;
  ircc->comms_connect_sync = rc_comms_connect_tcpudp_sync;
  ircc->comms_connect_async = rc_comms_connect_tcpudp_async;
  ircc->comms_complete_connect_async = rc_comms_complete_connect_tcpudp_async;
  ircc->comms_disconnect = rc_comms_disconnect_tcpudp;
  ircc->comms_write = rc_comms_write_tcpudp;
  ircc->comms_read = rc_comms_read_tcpudp;
  ircc->comms_getfds = rc_comms_getfds_tcpudp;
  ircc->comms_destroy = rc_comms_destroy_tcpudp;

  *rcc = ircc;
  return rcr_success;

rcit_fail:
  if (NULL != ircc)
  {
    free(ircc);
  }
  if (NULL != tcpudp)
  {
    if (NULL != tcpudp->hostname)
    {
      free(tcpudp->hostname);
      tcpudp->hostname = NULL;
    }
    free(tcpudp);
  }

  return result;
}


/** Create a rincmd client TCP connection for use with hostname:port */
enum e_rc_cli_result rc_comms_init_tcp(struct s_rc_cli_comms **rcc,
                                       const char *hostname, int port,
                                       FILE *comms_log)
{
  return rc_comms_init_tcpudp(rcc, hostname, port, AF_INET, SOCK_STREAM, 0, comms_log);
}


/** Create a rincmd client UDP connection for use with hostname:port */
enum e_rc_cli_result rc_comms_init_udp(struct s_rc_cli_comms **rcc,
                                       const char *hostname, int port, int localudpport,
                                       FILE *comms_log)
{
  return rc_comms_init_tcpudp(rcc, hostname, port, AF_INET, SOCK_DGRAM, localudpport, comms_log);
}


/** Create a rincmd client UDS connection for use with udsfile */
enum e_rc_cli_result rc_comms_init_uds(struct s_rc_cli_comms **rcc,
                                       const char *udsfile,
                                       FILE *comms_log)
{
  return rc_comms_init_tcpudp(rcc, udsfile, 0, AF_UNIX, SOCK_STREAM, 0, comms_log);
}
