/*
 * rc_cli_exec.c
 *
 *  Created on: Nov 20, 2017
 *      Author: ian
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rincmd/rc_public.h>
#include <rincmd/rc_registers.h>
#include <rincmd/rc_escape.h>
#include <rincmd/rc_strncpy.h>

#include <rincmd/rc_cli_parse.h>


/** Create a rincmd client parser. The framer must not be destroyed before the parser. */
enum e_rc_cli_result rc_cli_parse_init(struct s_rc_cli_parser **rcp, struct s_rc_frame *framer, uint8_t rc_address, bool use_channel, uint32_t rc_channel)
{
  struct s_rc_cli_parser *parser;

  if (NULL == rcp || NULL == framer)
    return rcr_paramfail;

  parser = (struct s_rc_cli_parser *) calloc(1, sizeof(*parser));
  if (NULL == parser)
    return rcr_allocfail;

  parser->framing = framer;
  parser->rc_address = rc_address;
  parser->use_channel = use_channel;
  parser->rc_channel = rc_channel;
  *rcp = parser;
  return rcr_success;
}


/** Destroy a rincmd client parser */
enum e_rc_cli_result rc_cli_parse_destroy(struct s_rc_cli_parser *rcp)
{
  free(rcp);
  rcp = NULL;
  return rcr_success;
}


/** Assemble a command ready to be sent to the host device */
enum e_rc_cli_result rc_cli_assemble_command(const struct s_rc_cli_parser *rcp, const struct s_rc_Message *cmd, char *buffer, size_t maxlen)
{
  rc_errcode_t res;
  size_t endpos;
  enum e_rc_cli_result result;

  if (NULL == rcp || NULL == cmd || NULL == buffer)
    return rcr_paramfail;

  /* Put pre-framing into buffer */
  result = rc_preframe_message(rcp->framing, true, buffer, maxlen, &endpos);
  if (rcr_success != result)
    return result;

  /* Construct a string to send */
  res = rc_constructor(&buffer[endpos], maxlen - endpos, cmd, rcp->framing->has_crc);
  if (RC_CNSTR_ERR_NONE != res)
    return rcr_parsefail;

  /* Append post framing to message */
  return rc_postframe_message(rcp->framing, true, buffer, maxlen);
}


/** Parse a response from the host device into msg. response will be modified to remove the framing postfix characters */
enum e_rc_cli_result rc_cli_parse_response(const struct s_rc_cli_parser *rcp, struct s_rc_Message *msg, const char *response)
{
  enum e_rc_cli_result result;
  rc_errcode_t parse_result;
  size_t startpos;
  size_t endpos;

  /* Find the start and end position of the encapsulated message */
  result = rc_frame_response_check(rcp->framing, false, response, &startpos, &endpos, NULL);
  if (rcr_success != result)
    return result;

  /* Parse the encapsulated message */
  parse_result = rc_parser_buflen(&response[startpos], endpos - startpos, msg, rcp->framing->has_crc);
  return (parse_result == RC_PARSE_ERR_NONE) ? rcr_success : rcr_parsefail;
}


/** Check that everything in the header address field matches up */
enum e_rc_cli_result rc_cli_check_response_header_address(const struct s_rc_cli_parser *rcp, const struct s_rc_Message *resp)
{
  /* If not a broadcast, check that the address matches */
  if (((rcp->rc_address & RC_MESG_MASK_ID) != RC_MESG_BROADCAST) &&
      (resp->addr & RC_MESG_MASK_ID) != (rcp->rc_address & RC_MESG_MASK_ID))
    return rcr_reply_badaddress;

  /* Check the response bit in the address field */
  if ((resp->addr & RC_MESG_RESPONSE) == 0)
    return rcr_reply_notresponse;

  return rcr_success;
}


/** Check that the header command and register id fields match up */
enum e_rc_cli_result rc_cli_check_response_header_cmd_reg(const struct s_rc_Message *resp,
                                                          enum e_rc_reg_command cmd, rc_RegId_t regid)
{
  /* Check the the command field */
  if ((enum e_rc_reg_command) resp->cmd != cmd)
    return rcr_reply_badcommand;

  /* Check the the register id field */
  if (resp->reg != regid)
    return rcr_reply_badregister;

  return rcr_success;
}


/** Check/parse the return code in the header */
enum e_rc_cli_result rc_cli_check_response_header_returncode(const struct s_rc_Message *resp, uint32_t *returncode)
{
  unsigned long err;
  enum e_rc_cli_result result;
  const bool iswrite = RC_IS_WRITE_COMMAND(resp->cmd);

  /* If error, or for all writes, convert the response errorcode */
  if ((resp->addr & RC_MESG_ERROR) != 0 || iswrite)
  {
    if (sscanf(resp->data, "%lx", &err) != 1)
      return rcr_parsefail;
    if (NULL != returncode)
      *returncode = err;
    result = ((resp->addr & RC_MESG_ERROR) != 0) ? rcr_reply_error : rcr_success;
  }
  else
  {
    if (NULL != returncode)
      *returncode = 0;
    result = rcr_success;
  }
  return result;
}


/** Check that everything in the header matches up */
enum e_rc_cli_result rc_cli_check_response_header(const struct s_rc_cli_parser *rcp, const struct s_rc_Message *resp,
                                                  enum e_rc_reg_command cmd, rc_RegId_t regid, uint32_t *returncode)
{
  enum e_rc_cli_result result;

  /* Check address field */
  result = rc_cli_check_response_header_address(rcp, resp);
  if (rcr_success != result)
    return result;

  /* Check command and register id field */
  result = rc_cli_check_response_header_cmd_reg(resp, cmd, regid);
  if (rcr_success != result)
    return result;

  /* Check/parse the return code */
  return rc_cli_check_response_header_returncode(resp, returncode);
}


/** Fill out a rincomms message */
enum e_rc_cli_result rc_cli_message_fillout(const struct s_rc_cli_parser *rcp, struct s_rc_Message *msg,
                                            bool need_reply, enum e_rc_reg_command cmd, rc_RegId_t regid,
                                            const void *data, size_t datalen)
{
  msg->addr = (RC_MESG_MASK_ID & rcp->rc_address);
  if (need_reply)
    msg->addr |= RC_MESG_REPLY;
  msg->cmd = cmd;
  msg->reg = regid;
  if (rcp->use_channel)
  {
    snprintf(msg->chan, sizeof(msg->chan), "%X", rcp->rc_channel);
  }
  else
  {
    rc_safe_strncpy(msg->chan, "", sizeof(msg->chan));
  }
  if (NULL == data || 0 == datalen)
  {
    rc_safe_strncpy(msg->data, "", RC_SIZE_BUFFER);
  }
  else
  {
    if (rc_escape_data((const uint8_t *) data, datalen, msg->data, RC_SIZE_BUFFER) != rer_success)
      return rcr_msgassemblefail;
  }
  return rcr_success;
}


/** Convert a rincomms read command into a suitable scanf format */
static const char *rincomms_read_cmd_to_scanfformat(enum e_rc_reg_command read_cmd)
{
  const char *result;

  switch (read_cmd)
  {
    case RC_MESG_READ_TYPE:
    case RC_MESG_READ_PERMISSION:
    case RC_MESG_READ_HEXADEC:
      result = "%llx";
      break;
    case RC_MESG_READ_DECIMAL:
    case RC_MESG_READ_RANGE_MIN:
    case RC_MESG_READ_RANGE_MAX:
      result = "%lld";
      break;
    /* case RC_MESG_READ_LITERAL: */
    default:
      result = NULL;
      break;
  }
  return result;
}


/**
 * Convert a rincomms write command into a suitable snprintf format
 *
 * @param write_cmd is the write command to get the snprintf format for
 * @param mask is a mask to be applied to the value so we only get the appropriate 8 digit hexadecimal numbers, since we only support 32bits
 *
 * @return printf format string suitable for the write command
 */
static const char *rincomms_write_cmd_to_snprintfformat(enum e_rc_reg_command write_cmd, uint64_t *mask)
{
  const char *result;

  switch (write_cmd)
  {
    case RC_MESG_EXECUTE:
    case RC_MESG_WRITE_HEXADEC:
      result = "%llX";
      *mask = 0x00000000FFFFFFFF;
      break;
    case RC_MESG_WRITE_DECIMAL:
      result = "%lld";
      *mask = 0xFFFFFFFFFFFFFFFF;
      break;
    default:
      result = NULL;
      break;
  }
  return result;
}


/*
 * Register read
 */


/** Prepare a command for a register read */
enum e_rc_cli_result rc_cli_assemble_register_read(struct s_rc_cli_parser *rcp,
                                                   enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                   char *buffer, size_t maxlen)
{
  enum e_rc_cli_result result;
  struct s_rc_Message cmd;

  /* Fill out request details */
  result = rc_cli_message_fillout(rcp, &cmd, true, read_cmd, regid, NULL, 0);
  if (rcr_success != result)
    return result;

  /* Assemble into the command */
  return rc_cli_assemble_command(rcp, &cmd, buffer, maxlen);
}


/** Parse the result of an integer register read from the message response */
enum e_rc_cli_result rc_cli_parse_int_register_message_read(struct s_rc_cli_parser *rcp,
                                                            enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                            const struct s_rc_Message *resp, int64_t *regvalue)
{
  long long value;
  const char *sscanf_fmt;
  enum e_rc_cli_result result;
  uint32_t returncode;

  sscanf_fmt = rincomms_read_cmd_to_scanfformat(read_cmd);
  if (NULL == sscanf_fmt)
    return rcr_paramfail;

  /* Check everything matches */
  result = rc_cli_check_response_header(rcp, resp, read_cmd, regid, &returncode);
  if (rcr_success != result)
    return result;

  /* Convert response data into an integer */
  if (sscanf(resp->data, sscanf_fmt, &value) != 1)
    return rcr_parsefail;

  /* Success */
  *regvalue = value;
  return rcr_success;
}


/** Parse the result of an integer register read */
enum e_rc_cli_result rc_cli_parse_int_register_read(struct s_rc_cli_parser *rcp,
                                                    enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                    const char *response, int64_t *regvalue)
{
  struct s_rc_Message resp;
  enum e_rc_cli_result result;

  /* Parse a response from the host device into msg */
  result = rc_cli_parse_response(rcp, &resp, response);
  if (rcr_success != result)
    return result;

  /* Parse the data from the resp message */
  return rc_cli_parse_int_register_message_read(rcp, read_cmd, regid, &resp, regvalue);
}


/**
 * Parse the result of an buffer (string with length, can contain '\x00') register read from the message response.
 *
 * @param regvalue_len should be set to the maximum buffer length on call, and will return actual buffer length on return
 */
enum e_rc_cli_result rc_cli_parse_buf_register_message_read(struct s_rc_cli_parser *rcp,
                                                            enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                            const struct s_rc_Message *resp, void *regvalue, size_t *regvalue_len)
{
  enum e_rc_cli_result result;
  uint32_t returncode;

  /* Check everything matches */
  result = rc_cli_check_response_header(rcp, resp, read_cmd, regid, &returncode);
  if (rcr_success != result)
    return result;

  /* Un-escape response data into an string */
  if (rc_unescape_data(resp->data, regvalue, regvalue_len, '\x00', '\xFF', NULL) != rer_success)
    return rcr_parsefail;

  /* Success */
  return rcr_success;
}


/**
 * Parse the result of an buffer (string with length, can contain '\x00') register read.
 *
 * @param regvalue_len should be set to the maximum buffer length on call, and will return actual buffer length on return
 */
enum e_rc_cli_result rc_cli_parse_buf_register_read(struct s_rc_cli_parser *rcp,
                                                    enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                    const char *response, void *regvalue, size_t *regvalue_len)
{
  struct s_rc_Message resp;
  enum e_rc_cli_result result;

  /* Parse a response from the host device into msg */
  result = rc_cli_parse_response(rcp, &resp, response);
  if (rcr_success != result)
    return result;

  /* Parse the data from the resp message */
  return rc_cli_parse_buf_register_message_read(rcp, read_cmd, regid, &resp, regvalue, regvalue_len);
}


/** Parse the result of an string register read from the message response */
enum e_rc_cli_result rc_cli_parse_str_register_message_read(struct s_rc_cli_parser *rcp,
                                                            enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                            const struct s_rc_Message *resp, char *regvalue, size_t regvalue_maxlen)
{
  enum e_rc_cli_result result;
  size_t regvalue_len;

  regvalue_len = regvalue_maxlen - 1;
  result = rc_cli_parse_buf_register_message_read(rcp, read_cmd, regid, resp, regvalue, &regvalue_len);
  if (rcr_success != result)
    return result;
  regvalue[regvalue_len] = '\0';

  /* Success */
  return rcr_success;
}


/** Parse the result of an string register read */
enum e_rc_cli_result rc_cli_parse_str_register_read(struct s_rc_cli_parser *rcp,
                                                    enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                    const char *response, char *regvalue, size_t regvalue_maxlen)
{
  enum e_rc_cli_result result;
  size_t regvalue_len;

  regvalue_len = regvalue_maxlen - 1;
  result = rc_cli_parse_buf_register_read(rcp, read_cmd, regid, response, regvalue, &regvalue_len);
  if (rcr_success != result)
    return result;
  regvalue[regvalue_len] = '\0';

  /* Success */
  return rcr_success;
}


/** Parse the result of an streaming register read (1..RC_STREAM_DATA_REGISTERS integer registers) from the message response */
enum e_rc_cli_result rc_cli_parse_stream_register_message_read(struct s_rc_cli_parser *rcp, rc_RegId_t stream_reg_base,
                                                               const struct s_rc_Message *resp, int64_t regvalues[RC_STREAM_DATA_REGISTERS])
{
  enum e_rc_cli_result result;
  uint32_t returncode;
  #define REGSTR_LEN (RC_STREAM_DATA_REGISTERS * 5 + 1)
  char regstr[REGSTR_LEN] = "";
  size_t i;
  int64_t mask;

  /* Setup scanf format */
  for (i = 0; i < RC_STREAM_DATA_REGISTERS; i++)
  {
    strcat(regstr, "%8llx");
  }

  /* Check everything matches */
  result = rc_cli_check_response_header(rcp, resp, RC_MESG_READ_HEXADEC, stream_reg_base, &returncode);
  if (rcr_success != result)
    return result;

  /* Convert response data into a integers */
  if (sscanf(resp->data, regstr, &regvalues[0], &regvalues[1], &regvalues[2], &regvalues[3], &regvalues[4]) < 1)
    return rcr_parsefail;

  /* Extend sign */
  mask = 0x000000007FFFFFFF;
  for (i = 0; i < RC_STREAM_DATA_REGISTERS; i++)
  {
    /* Sign extend hexadecimal writes if required: eg 8bit: 0x80 -> 0xFFFFFFFFFFFFFF80(-128d) */
    if (regvalues[i] >= 0 && (regvalues[i] & ~mask) == mask + 1)
    {
      regvalues[i] |= ~mask;
    }
  }

  /* Success */
  return rcr_success;
}


/** Parse the result of an streaming register read (1..RC_STREAM_DATA_REGISTERS integer registers) */
enum e_rc_cli_result rc_cli_parse_stream_register_read(struct s_rc_cli_parser *rcp, rc_RegId_t stream_reg_base,
                                                       const char *response, int64_t regvalues[RC_STREAM_DATA_REGISTERS])
{
  struct s_rc_Message resp;
  enum e_rc_cli_result result;

  /* Parse a response from the host device into msg */
  result = rc_cli_parse_response(rcp, &resp, response);
  if (rcr_success != result)
    return result;

  /* Parse the data from the resp message */
  return rc_cli_parse_stream_register_message_read(rcp, stream_reg_base, &resp, regvalues);
}


/*
 * Register write
 */


/** Prepare a command for a integer register write */
enum e_rc_cli_result rc_cli_assemble_int_register_write(struct s_rc_cli_parser *rcp,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid, int64_t regvalue,
                                                        char *buffer, size_t maxlen)
{
  enum e_rc_cli_result result;
  struct s_rc_Message cmd;
  const char *snprintf_fmt;
  uint64_t mask = 0;

  snprintf_fmt = rincomms_write_cmd_to_snprintfformat(write_cmd, &mask);
  if (NULL == snprintf_fmt)
    return rcr_paramfail;

  /* Fill out request details */
  result = rc_cli_message_fillout(rcp, &cmd, true, write_cmd, regid, NULL, 0);
  if (rcr_success != result)
    return result;
  snprintf(cmd.data, sizeof(cmd.data), snprintf_fmt, (long long) (regvalue & mask));

  /* Assemble into the command */
  return rc_cli_assemble_command(rcp, &cmd, buffer, maxlen);
}


/** Prepare a command for a buffer (string with length, can contain '\x00') register write */
enum e_rc_cli_result rc_cli_assemble_buf_register_write(struct s_rc_cli_parser *rcp,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                        const void *regvalue, size_t regvalue_len,
                                                        char *buffer, size_t maxlen)
{
  enum e_rc_cli_result result;
  struct s_rc_Message cmd;

  /* Fill out request details */
  result = rc_cli_message_fillout(rcp, &cmd, true, write_cmd, regid, regvalue, regvalue_len);
  if (rcr_success != result)
    return result;

  /* Assemble into the command */
  return rc_cli_assemble_command(rcp, &cmd, buffer, maxlen);
}


/** Prepare a command for a string register write */
enum e_rc_cli_result rc_cli_assemble_str_register_write(struct s_rc_cli_parser *rcp,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid, const char *regvalue,
                                                        char *buffer, size_t maxlen)
{
  return rc_cli_assemble_buf_register_write(rcp, write_cmd, regid, regvalue, strlen(regvalue), buffer, maxlen);
}


/** Parse the result of a register write */
enum e_rc_cli_result rc_cli_parse_register_write(struct s_rc_cli_parser *rcp,
                                                 enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                 const char *response)
{
  struct s_rc_Message resp;
  enum e_rc_cli_result result;
  uint32_t returncode;

  /* Parse a response from the host device into msg */
  result = rc_cli_parse_response(rcp, &resp, response);
  if (rcr_success != result)
    return result;

  /* Check everything matches */
  result = rc_cli_check_response_header(rcp, &resp, write_cmd, regid, &returncode);
  if (rcr_success != result)
    return result;

  /* Check return code */
  return (RC_BUILD_ERR(RC_HNDL_ERR_NONE, RC_NONE_ERR_NONE) == returncode) ? rcr_success : rcr_reply_error;
}
