/*
 * rc_cli_parse.h
 *
 *  Created on: Nov 22, 2017
 *      Author: ian
 */

#ifndef RC_CLI_PARSE_H_
#define RC_CLI_PARSE_H_

#include <stdbool.h>
#include <stdint.h>
#include <rincmd/rc_cli_comms.h>
#include <rincmd/rc_frame.h>
#include <rincmd/rc_parser.h>


/** Command executer structure */
struct s_rc_cli_parser
{
  struct s_rc_frame *framing;         /*!< Framing of messages */

  uint8_t rc_address;                 /*!< The rincomms address to use, usually 0 (RC_MESG_BROADCAST) */
  bool use_channel;                   /*!< Whether the channel info should be used */
  uint32_t rc_channel;                /*!< The rincomms channel */
};


/** Create a rincmd client parser. The framer must not be destroyed before the parser. */
enum e_rc_cli_result rc_cli_parse_init(struct s_rc_cli_parser **rcp, struct s_rc_frame *framer, uint8_t rc_address, bool use_channel, uint32_t rc_channel);


/** Destroy a rincmd client parser */
enum e_rc_cli_result rc_cli_parse_destroy(struct s_rc_cli_parser *rcp);


/** Assemble a command ready to be sent to the host device */
enum e_rc_cli_result rc_cli_assemble_command(const struct s_rc_cli_parser *rcp, const struct s_rc_Message *cmd, char *buffer, size_t maxlen);


/** Parse a response from the host device into msg. response will be modified to remove the framing postfix characters */
enum e_rc_cli_result rc_cli_parse_response(const struct s_rc_cli_parser *rcp, struct s_rc_Message *msg, const char *response);


/** Check that the header command and register id fields match up */
enum e_rc_cli_result rc_cli_check_response_header_cmd_reg(const struct s_rc_Message *resp,
                                                          enum e_rc_reg_command cmd, rc_RegId_t regid);


/** Check that everything in the header address field matches up */
enum e_rc_cli_result rc_cli_check_response_header_address(const struct s_rc_cli_parser *rcp, const struct s_rc_Message *resp);


/** Check/parse the return code in the header */
enum e_rc_cli_result rc_cli_check_response_header_returncode(const struct s_rc_Message *resp, uint32_t *returncode);


/** Check that everything in the header matches up */
enum e_rc_cli_result rc_cli_check_response_header(const struct s_rc_cli_parser *rcp, const struct s_rc_Message *resp,
                                                  enum e_rc_reg_command cmd, rc_RegId_t regid, uint32_t *returncode);


/** Fill out a rincomms message */
enum e_rc_cli_result rc_cli_message_fillout(const struct s_rc_cli_parser *rcp, struct s_rc_Message *msg,
                                            bool need_reply, enum e_rc_reg_command cmd, rc_RegId_t regid,
                                            const void *data, size_t datalen);


/*
 * Register read
 */


/** Prepare a command for a register read */
enum e_rc_cli_result rc_cli_assemble_register_read(struct s_rc_cli_parser *rcp,
                                                   enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                   char *buffer, size_t maxlen);


/** Parse the result of an integer register read from the message response */
enum e_rc_cli_result rc_cli_parse_int_register_message_read(struct s_rc_cli_parser *rcp,
                                                            enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                            const struct s_rc_Message *resp, int64_t *regvalue);


/** Parse the result of an integer register read */
enum e_rc_cli_result rc_cli_parse_int_register_read(struct s_rc_cli_parser *rcp,
                                                    enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                    const char *response, int64_t *regvalue);


/**
 * Parse the result of an buffer (string with length, can contain '\x00') register read from the message response.
 *
 * @param regvalue_len should be set to the maximum buffer length on call, and will return actual buffer length on return
 */
enum e_rc_cli_result rc_cli_parse_buf_register_message_read(struct s_rc_cli_parser *rcp,
                                                            enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                            const struct s_rc_Message *resp, void *regvalue, size_t *regvalue_len);


/**
 * Parse the result of an buffer (string with length, can contain '\x00') register read.
 *
 * @param regvalue_len should be set to the maximum buffer length on call, and will return actual buffer length on return
 */
enum e_rc_cli_result rc_cli_parse_buf_register_read(struct s_rc_cli_parser *rcp,
                                                    enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                    const char *response, void *regvalue, size_t *regvalue_len);


/** Parse the result of an string register read from the message response */
enum e_rc_cli_result rc_cli_parse_str_register_message_read(struct s_rc_cli_parser *rcp,
                                                            enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                            const struct s_rc_Message *resp, char *regvalue, size_t regvalue_maxlen);


/** Parse the result of an string register read */
enum e_rc_cli_result rc_cli_parse_str_register_read(struct s_rc_cli_parser *rcp,
                                                    enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                    const char *response, char *regvalue, size_t regvalue_maxlen);


/** Parse the result of an streaming register read (1..RC_STREAM_DATA_REGISTERS integer registers) from the message response */
enum e_rc_cli_result rc_cli_parse_stream_register_message_read(struct s_rc_cli_parser *rcp, rc_RegId_t stream_reg_base,
                                                               const struct s_rc_Message *resp, int64_t regvalues[RC_STREAM_DATA_REGISTERS]);


/** Parse the result of an streaming register read (1..RC_STREAM_DATA_REGISTERS integer registers) */
enum e_rc_cli_result rc_cli_parse_stream_register_read(struct s_rc_cli_parser *rcp, rc_RegId_t stream_reg_base,
                                                       const char *response, int64_t regvalues[RC_STREAM_DATA_REGISTERS]);


/*
 * Register write
 */


/** Prepare a command for a integer register write */
enum e_rc_cli_result rc_cli_assemble_int_register_write(struct s_rc_cli_parser *rcp,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid, int64_t regvalue,
                                                        char *buffer, size_t maxlen);


/** Prepare a command for a buffer (string with length, can contain '\x00') register write */
enum e_rc_cli_result rc_cli_assemble_buf_register_write(struct s_rc_cli_parser *rcp,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                        const void *regvalue, size_t regvalue_len,
                                                        char *buffer, size_t maxlen);


/** Prepare a command for a string register write */
enum e_rc_cli_result rc_cli_assemble_str_register_write(struct s_rc_cli_parser *rcp,
                                                        enum e_rc_reg_command write_cmd, rc_RegId_t regid, const char *regvalue,
                                                        char *buffer, size_t maxlen);


/** Parse the result of a register write */
enum e_rc_cli_result rc_cli_parse_register_write(struct s_rc_cli_parser *rcp,
                                                 enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                 const char *response);


#endif /* RC_CLI_PARSE_H_ */
