/*
 * rc_cli_sync.c
 *
 *  Created on: Nov 20, 2017
 *      Author: ian
 */
#define _XOPEN_SOURCE 700
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <rincmd/rc_cli_parse.h>

#include <rincmd/rc_cli_sync.h>


/**
 * Create an synchronous client
 *
 * @param rsc is set to the created sync client on success
 * @param cli_comms is the already existing comms interface. You are responsible for destroying this once rsc has been destroyed
 * @param rc_address is the rincmd address to address messages to. RC_MESG_BROADCAST is a good choice
 * @param use_crc is a flag to indicate the use of CRC framed messages, or normal framing
 */
enum e_rc_cli_result rc_sync_client_init(struct s_rc_sync_client **rsc,
                                         struct s_rc_cli_comms *cli_comms,
                                         uint8_t rc_address, bool use_crc)
{
  struct s_rc_sync_client *irsc = NULL;
  struct s_rc_frame *frame = NULL;
  enum e_rc_cli_result result;

  if (NULL == rsc || NULL == cli_comms)
    return rcr_paramfail;

  /* Allocate the sync client structure */
  irsc = (struct s_rc_sync_client *) calloc(1, sizeof(*irsc));
  if (NULL == irsc)
  {
    result = rcr_allocfail;
    goto rsci_fail;
  }

  /* Setup comms */
  irsc->cli_comms = cli_comms;

  /* Setup framing */
  result = (use_crc) ? rc_frame_init_crc(&frame) : rc_frame_init_normal(&frame);
  if (rcr_success != result)
    goto rsci_fail;

  /* Setup parser */
  result = rc_cli_parse_init(&irsc->cli_parser, frame, rc_address, false, 0);
  if (rcr_success != result)
    goto rsci_fail;

  /* Create buffer */
  irsc->buffer = calloc(RC_CLI_MSG_MAXLEN, sizeof(char));
  if (NULL == irsc->buffer)
    goto rsci_fail;
  irsc->bufferlen = RC_CLI_MSG_MAXLEN;

  /* Success */
  *rsc = irsc;
  return rcr_success;

rsci_fail:
  rc_sync_client_destroy(irsc);
  irsc = NULL;
  return result;
}


/** Destroy a synchronous client */
enum e_rc_cli_result rc_sync_client_destroy(struct s_rc_sync_client *rsc)
{
  if (NULL != rsc)
  {
    if (NULL != rsc->cli_comms)
    {
      /* Up to the caller to destroy this */
      /* rc_cli_comms_destroy(rsc->cli_comms); */
      rsc->cli_comms = NULL;
    }
    if (NULL != rsc->cli_parser)
    {
      if (NULL != rsc->cli_parser->framing)
      {
        rc_frame_destroy(rsc->cli_parser->framing);
        rsc->cli_parser->framing = NULL;
      }
      rc_cli_parse_destroy(rsc->cli_parser);
      rsc->cli_parser = NULL;
    }
    if (NULL != rsc->buffer)
    {
      free(rsc->buffer);
      rsc->buffer = NULL;
    }
    rsc = NULL;
  }
  return rcr_success;
}


/** Connect to the server */
enum e_rc_cli_result rc_sync_client_connect(struct s_rc_sync_client *rsc, const struct timespec *timeout)
{
  return rc_cli_comms_connect_sync(rsc->cli_comms, timeout);
}


/** Disconnect from the server */
enum e_rc_cli_result rc_sync_client_disconnect(struct s_rc_sync_client *rsc)
{
  return rc_cli_comms_disconnect(rsc->cli_comms);
}


/**
 * Get the file descriptor
 *
 * @param rsc is the rincmd sync client
 * @param readfd is set to the read file descriptor if the comms is connected, < 0 otherwise. Not NULL.
 */
enum e_rc_cli_result rc_sync_client_get_fd(struct s_rc_sync_client *rsc, int *readfd)
{
  int connectfd;

  /* Check params */
  if (NULL == rsc || NULL == readfd)
  {
    return rcr_paramfail;
  }

  /* Get file descriptors */
  return rc_cli_comms_getfds(rsc->cli_comms, readfd, &connectfd);
}


/** Read a response to a command. Don't call directly */
enum e_rc_cli_result rc_sync_client_read_response(struct s_rc_sync_client *rsc, const struct timespec *timeout)
{
  size_t posn;
  size_t startpos;
  size_t endpos;
  size_t msgend;
  ssize_t bytes_read;
  enum e_rc_cli_result result;
  fd_set readfds;
  int readfd;
  int connectfd;
  int sres;

  for (posn = 0; posn < rsc->bufferlen; )
  {
    FD_ZERO(&readfds);
    result = rc_cli_comms_getfds(rsc->cli_comms, &readfd, &connectfd);
    if (rcr_success != result)
      return result;
    if (readfd >= 0)
      FD_SET(readfd, &readfds);
    sres = pselect(readfd + 1, &readfds, NULL, NULL, timeout, NULL);
    if (0 == sres)
    {
      /* Timeout, keep looping */
      return rcr_timeout;
    }
    else if (sres < 0)
    {
      return rcr_readfail;
    }
    else if (readfd >= 0 && FD_ISSET(readfd, &readfds))
    {
      bytes_read = rsc->bufferlen - posn;
      result = rc_cli_comms_read(rsc->cli_comms, &rsc->buffer[posn], &bytes_read);
      if (rcr_success != result)
        return result;
      posn += bytes_read;
      rsc->buffer[posn] = '\0';

      result = rc_frame_response_check(rsc->cli_parser->framing, false, rsc->buffer, &startpos, &endpos, &msgend);
      if (rcr_framingfail != result)
      {
        /* Valid message or error, log it first */
        rc_cli_comms_log(rsc->cli_comms, true, rsc->buffer, msgend);
        return result;
      }
    }
  }
  return rcr_bufferoverflow;
}


/**
 * Read buffer register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param read_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the buffer written with the register data read from the register
 * @param regvalue_len should be set to the maximum buffer length on call, and will return actual buffer length on return
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_buf_register_read(struct s_rc_sync_client *rsc,
                                                     enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                     void *regvalue, size_t *regvalue_len,
                                                     const struct timespec *timeout)
{
  enum e_rc_cli_result result;
  ssize_t writelen;

  result = rc_cli_assemble_register_read(rsc->cli_parser, read_cmd, regid, rsc->buffer, rsc->bufferlen);
  if (rcr_success != result)
    return result;

  writelen = strlen(rsc->buffer);
  result = rc_cli_comms_write(rsc->cli_comms, rsc->buffer, &writelen);
  if (rcr_success != result)
    return result;

  result = rc_sync_client_read_response(rsc, timeout);
  if (rcr_success != result)
    return result;

  return rc_cli_parse_buf_register_read(rsc->cli_parser, read_cmd, regid, rsc->buffer, regvalue, regvalue_len);
}


/**
 * Read string register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param read_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the buffer written with the register data read from the register
 * @param regvalue_maxlen is the maximum length of regvalue
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_str_register_read(struct s_rc_sync_client *rsc,
                                                      enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                      char *regvalue, size_t regvalue_maxlen,
                                                      const struct timespec *timeout)
{
  enum e_rc_cli_result result;
  ssize_t writelen;

  result = rc_cli_assemble_register_read(rsc->cli_parser, read_cmd, regid, rsc->buffer, rsc->bufferlen);
  if (rcr_success != result)
    return result;

  writelen = strlen(rsc->buffer);
  result = rc_cli_comms_write(rsc->cli_comms, rsc->buffer, &writelen);
  if (rcr_success != result)
    return result;

  result = rc_sync_client_read_response(rsc, timeout);
  if (rcr_success != result)
    return result;

  return rc_cli_parse_str_register_read(rsc->cli_parser, read_cmd, regid, rsc->buffer, regvalue, regvalue_maxlen);
}


/**
 * Read register regid and return a 64bit integer in regvalue
 *
 * @param rsc is the rincmd synchronous client
 * @param read_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the 64bit integer pointer set to the read register value
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_int_register_read(struct s_rc_sync_client *rsc,
                                                      enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                      int64_t *regvalue,
                                                      const struct timespec *timeout)
{
  enum e_rc_cli_result result;
  ssize_t writelen;

  result = rc_cli_assemble_register_read(rsc->cli_parser, read_cmd, regid, rsc->buffer, rsc->bufferlen);
  if (rcr_success != result)
    return result;

  writelen = strlen(rsc->buffer);
  result = rc_cli_comms_write(rsc->cli_comms, rsc->buffer, &writelen);
  if (rcr_success != result)
    return result;

  result = rc_sync_client_read_response(rsc, timeout);
  if (rcr_success != result)
    return result;

  return rc_cli_parse_int_register_read(rsc->cli_parser, read_cmd, regid, rsc->buffer, regvalue);
}


/**
 * Write buffer register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the buffer to write to the register
 * @param regvalue_len is the length of the buffer to write to the register
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_buf_register_write(struct s_rc_sync_client *rsc,
                                                       enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                       const void *regvalue, size_t regvalue_len,
                                                       const struct timespec *timeout)
{
  enum e_rc_cli_result result;
  ssize_t writelen;

  result = rc_cli_assemble_buf_register_write(rsc->cli_parser, write_cmd, regid, regvalue, regvalue_len, rsc->buffer, rsc->bufferlen);
  if (rcr_success != result)
    return result;

  writelen = strlen(rsc->buffer);
  result = rc_cli_comms_write(rsc->cli_comms, rsc->buffer, &writelen);
  if (rcr_success != result)
    return result;

  result = rc_sync_client_read_response(rsc, timeout);
  if (rcr_success != result)
    return result;

  return rc_cli_parse_register_write(rsc->cli_parser, write_cmd, regid, rsc->buffer);
}


/**
 * Write string register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the string to write to the register
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_str_register_write(struct s_rc_sync_client *rsc,
                                                       enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                       const char *regvalue,
                                                       const struct timespec *timeout)
{
  enum e_rc_cli_result result;
  ssize_t writelen;

  result = rc_cli_assemble_str_register_write(rsc->cli_parser, write_cmd, regid, regvalue, rsc->buffer, rsc->bufferlen);
  if (rcr_success != result)
    return result;

  writelen = strlen(rsc->buffer);
  result = rc_cli_comms_write(rsc->cli_comms, rsc->buffer, &writelen);
  if (rcr_success != result)
    return result;

  result = rc_sync_client_read_response(rsc, timeout);
  if (rcr_success != result)
    return result;

  return rc_cli_parse_register_write(rsc->cli_parser, write_cmd, regid, rsc->buffer);
}


/**
 * Write integer register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the integer to write to the register
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_int_register_write(struct s_rc_sync_client *rsc,
                                                       enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                       int64_t regvalue,
                                                       const struct timespec *timeout)
{
  enum e_rc_cli_result result;
  ssize_t writelen;

  result = rc_cli_assemble_int_register_write(rsc->cli_parser, write_cmd, regid, regvalue, rsc->buffer, rsc->bufferlen);
  if (rcr_success != result)
    return result;

  writelen = strlen(rsc->buffer);
  result = rc_cli_comms_write(rsc->cli_comms, rsc->buffer, &writelen);
  if (rcr_success != result)
    return result;

  result = rc_sync_client_read_response(rsc, timeout);
  if (rcr_success != result)
    return result;

  return rc_cli_parse_register_write(rsc->cli_parser, write_cmd, regid, rsc->buffer);
}
