/*
 * rc_cli_sync.h
 *
 *  Created on: Nov 23, 2017
 *      Author: ian
 */

#ifndef RC_CLI_SYNC_H_
#define RC_CLI_SYNC_H_

#include <stddef.h>
#include <stdint.h>
#include <stdbool.h>
#include <rincmd/rc_cli_parse.h>
#include <rincmd/rc_cli_comms.h>


/** Structure for a running a synchronous rincmd client */
struct s_rc_sync_client
{
  struct s_rc_cli_comms *cli_comms;     /*!< Communication with the host */
  struct s_rc_cli_parser *cli_parser;   /*!< Parser for pulling apart and creating messages */
  char *buffer;                         /*!< Buffer for data */
  size_t bufferlen;                     /*!< Length of the data buffer */
};


/**
 * Create an synchronous client
 *
 * @param rsc is set to the created sync client on success
 * @param cli_comms is the already existing comms interface. You are responsible for destroying this once rsc has been destroyed
 * @param rc_address is the rincmd address to address messages to. RC_MESG_BROADCAST is a good choice
 * @param use_crc is a flag to indicate the use of CRC framed messages, or normal framing
 */
enum e_rc_cli_result rc_sync_client_init(struct s_rc_sync_client **rsc,
                                         struct s_rc_cli_comms *cli_comms,
                                         uint8_t rc_address, bool use_crc);


/** Destroy a synchronous client */
enum e_rc_cli_result rc_sync_client_destroy(struct s_rc_sync_client *rsc);


/** Connect to the server */
enum e_rc_cli_result rc_sync_client_connect(struct s_rc_sync_client *rsc, const struct timespec *timeout);


/** Disconnect from the server */
enum e_rc_cli_result rc_sync_client_disconnect(struct s_rc_sync_client *rsc);


/**
 * Get the file descriptor
 *
 * @param rsc is the rincmd sync client
 * @param readfd is set to the read file descriptor if the comms is connected, < 0 otherwise. Not NULL.
 */
enum e_rc_cli_result rc_sync_client_get_fd(struct s_rc_sync_client *rsc, int *readfd);


/** Read a response to a command. Don't call directly */
enum e_rc_cli_result rc_sync_client_read_response(struct s_rc_sync_client *rsc, const struct timespec *timeout);


/**
 * Read buffer register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param read_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the buffer written with the register data read from the register
 * @param regvalue_len should be set to the maximum buffer length on call, and will return actual buffer length on return
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_buf_register_read(struct s_rc_sync_client *rsc,
                                                     enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                     void *regvalue, size_t *regvalue_len,
                                                     const struct timespec *timeout);


/**
 * Read string register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param read_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the buffer written with the register data read from the register
 * @param regvalue_maxlen is the maximum length of regvalue
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_str_register_read(struct s_rc_sync_client *rsc,
                                                      enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                      char *regvalue, size_t regvalue_maxlen,
                                                      const struct timespec *timeout);


/**
 * Read register regid and return a 64bit integer in regvalue
 *
 * @param rsc is the rincmd synchronous client
 * @param read_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the 64bit integer pointer set to the read register value
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_int_register_read(struct s_rc_sync_client *rsc,
                                                      enum e_rc_reg_command read_cmd, rc_RegId_t regid,
                                                      int64_t *regvalue,
                                                      const struct timespec *timeout);


/**
 * Write buffer register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the buffer to write to the register
 * @param regvalue_len is the length of the buffer to write to the register
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_buf_register_write(struct s_rc_sync_client *rsc,
                                                       enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                       const void *regvalue, size_t regvalue_len,
                                                       const struct timespec *timeout);


/**
 * Write string register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the string to write to the register
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_str_register_write(struct s_rc_sync_client *rsc,
                                                       enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                       const char *regvalue,
                                                       const struct timespec *timeout);


/**
 * Write integer register regid
 *
 * @param rsc is the rincmd synchronous client
 * @param write_cmd is the command to execute
 * @param regid is the register to read
 * @param regvalue is the integer to write to the register
 * @param timeout is the timeout period or NULL for none
 */
enum e_rc_cli_result rc_sync_client_int_register_write(struct s_rc_sync_client *rsc,
                                                       enum e_rc_reg_command write_cmd, rc_RegId_t regid,
                                                       int64_t regvalue,
                                                       const struct timespec *timeout);


#endif /* RC_CLI_SYNC_H_ */
