/*
 * rc_escape.c
 *
 * Escape/de-escape a buffer suitable for sending over rincmd
 *
 *  Created on: Sep 21, 2017
 *      Author: ian
 */
#include <stdint.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "rc_escape.h"


/**
 * Escape a buffer suitable for sending over rincmd
 *
 * The following characters are escaped:
 *   '\x00'..'0x1F', ':', ';', '\x7F'..'\xFF'
 * in the following way:
 *   '\xXX' -> "\XX"
 * '\' is escaped to "\\"
 * All other characters (printables), and not escaped and copied as is
 *
 * @param inbuf is the data to be escaped. It can contain any characters '\0'..'\xFF'
 * @param inbuflen is the length of inbuf
 * @param outbuf is where the escaped data will be stored. It will be null terminated, if there is room, otherwise the function will return rer_buffer_size
 * @param outbuflen needs to be 3 times the length of inbuflen to fit a worst case escaped string, plus one null termination character
 *
 * @return rur_success if converted and fits within outbuflen
 */
enum e_rc_escape_result rc_escape_data(const uint8_t *inbuf, const size_t inbuflen, char *outbuf, size_t outbuflen)
{
  size_t inpos;
  size_t outpos;
  size_t outleft;
  int snres;
  uint8_t ch;

  for (inpos = 0, outpos = 0; inpos < inbuflen; inpos++)
  {
    ch = inbuf[inpos];

    outleft = outbuflen - outpos;
    if (ch < ' ' || ch > '~' || ':' == ch || ';' == ch)
    {
      snres = snprintf(&outbuf[outpos], outleft, "\\%02X", (unsigned) ch);
      if (snres < 0 || (size_t) snres > outleft)
        return rer_buffer_size;
      outpos += snres;
    }
    else if ('\\' == ch)
    {
      if (outleft < 3)
        return rer_buffer_size;
      outbuf[outpos++] = '\\';
      outbuf[outpos++] = '\\';
      outbuf[outpos] = '\0';
    }
    else
    {
      if (outleft < 2)
        return rer_buffer_size;
      outbuf[outpos++] = ch;
      outbuf[outpos] = '\0';
    }
  }
  return rer_success;
}


static enum e_rc_escape_result rc_unescape_store(uint8_t *outbuf, size_t outpos, uint8_t data, uint8_t minch, uint8_t maxch, const uint8_t *character_set)
{
  if (data < minch)
    return rer_data_low_range;
  if (data > maxch)
    return rer_data_high_range;
  if (NULL != character_set && strchr((const char *) character_set, data) == NULL)
    return rer_data_low_range;
  if (NULL != outbuf)
    outbuf[outpos] = data;
  return rer_success;
}


/**
 * Un-escape a string received over rincmd into a buffer
 *
 * The following characters are escaped:
 *   '\x00'..'0x31', ':', ';', '\x7f'..'\xFF'
 * in the following way:
 *   '\xXX' -> "\XX"
 * '\' is escaped to "\\"
 * All other characters (printables), and not escaped and copied as is
 *
 * @param inbuf is the escaped data to be un-escaped. Must be null terminated. Not NULL.
 * @param outbuf is where the unescaped data will be stored. If NULL, will perform a check of the data in inbuf.
 * @param outbuflen is a pointer to the output buffer length. Set it to the allocated size of outbuf before calling
 *        (it needs to be the same length as inbuf to fit a worst case escaped string (no escaped characters in inbuf))
 *        On success, it is set to the length of the converted data.
 * @param minch is the lower limit of characters from inbuf
 * @param maxch is the upper limit of characters from inbuf
 * @param character_set is a set of characters allowable from inbuf. Not used if NULL.
 *
 * @return rur_success if converted and fits within outbuflen
 */
enum e_rc_escape_result rc_unescape_data(const char *inbuf, uint8_t *outbuf, size_t *outbuflen,
                                         uint8_t minch, uint8_t maxch, const uint8_t *character_set)
{
  size_t outpos;
  unsigned conv;
  enum e_rc_escape_result res;

  for (outpos = 0; '\0' != *inbuf; outpos++)
  {
    if (outpos >= *outbuflen)
      return rer_buffer_size;
    if ('\\' == inbuf[0])
    {
      if ('\\' == inbuf[1])
      {
        res = rc_unescape_store(outbuf, outpos, '\\', minch, maxch, character_set);
        if (rer_success != res)
          return res;
        inbuf += 2;
      }
      else if (isxdigit(inbuf[1]) && isxdigit(inbuf[2]))
      {
        if (sscanf(&inbuf[1], "%02x", &conv) != 1)
          return rer_data_conversion;
        inbuf += 3;
        res = rc_unescape_store(outbuf, outpos, (uint8_t) conv, minch, maxch, character_set);
        if (rer_success != res)
          return res;
      }
      else
      {
        /* Something is poorly escaped, just copy it over */
        res = rc_unescape_store(outbuf, outpos, inbuf[0], minch, maxch, character_set);
        if (rer_success != res)
          return res;
        inbuf++;
      }
    }
    else
    {
      res = rc_unescape_store(outbuf, outpos, inbuf[0], minch, maxch, character_set);
      if (rer_success != res)
        return res;
      inbuf++;
    }
  }

  *outbuflen = outpos;
  return rer_success;
}
