/*
 * rc_escape.h
 *
 *  Created on: Sep 21, 2017
 *      Author: ian
 */

#ifndef RC_ESCAPE_H_
#define RC_ESCAPE_H_

#include <stddef.h>
#include <stdint.h>


/** Result of string escaping/unescaping functions */
enum e_rc_escape_result
{
  rer_success,          /*!< Success */
  rer_buffer_size,      /*!< Result string doesn't fit in the output buffer */
  rer_data_conversion,  /*!< Converting hex string into integer failed */
  rer_data_low_range,   /*!< A byte is less than the lower limit, or not in the character set */
  rer_data_high_range,  /*!< A byte is greater than the upper limit */

  num_rer               /*!< Number of result codes */
};


/**
 * Escape a buffer suitable for sending over rincmd
 *
 * The following characters are escaped:
 *   '\x00'..'0x1F', ':', ';', '\x7F'..'\xFF'
 * in the following way:
 *   '\xXX' -> "\XX"
 * '\' is escaped to "\\"
 * All other characters (printables), and not escaped and copied as is
 *
 * @param inbuf is the data to be escaped. It can contain any characters '\0'..'\xFF'
 * @param inbuflen is the length of inbuf
 * @param outbuf is where the escaped data will be stored. It will be null terminated, if there is room, otherwise the function will return rer_buffer_size
 * @param outbuflen needs to be 3 times the length of inbuflen to fit a worst case escaped string, plus one null termination character
 *
 * @return rur_success if converted and fits within outbuflen
 */
enum e_rc_escape_result rc_escape_data(const uint8_t *inbuf, const size_t inbuflen, char *outbuf, size_t outbuflen);


/**
 * Un-escape a string received over rincmd into a buffer
 *
 * The following characters are escaped:
 *   '\x00'..'0x31', ':', ';', '\x7f'..'\xFF'
 * in the following way:
 *   '\xXX' -> "\XX"
 * '\' is escaped to "\\"
 * All other characters (printables), and not escaped and copied as is
 *
 * @param inbuf is the escaped data to be un-escaped. Must be null terminated. Not NULL.
 * @param outbuf is where the unescaped data will be stored. If NULL, will perform a check of the data in inbuf.
 * @param outbuflen is a pointer to the output buffer length. Set it to the allocated size of outbuf before calling
 *        (it needs to be the same length as inbuf to fit a worst case escaped string (no escaped characters in inbuf))
 *        On success, it is set to the length of the converted data.
 * @param minch is the lower limit of characters from inbuf
 * @param maxch is the upper limit of characters from inbuf
 * @param character_set is a set of characters allowable from inbuf. Not used if NULL.
 *
 * @return rur_success if converted and fits within outbuflen
 */
enum e_rc_escape_result rc_unescape_data(const char *inbuf, uint8_t *outbuf, size_t *outbuflen,
                                         uint8_t minch, uint8_t maxch, const uint8_t *character_set);


#endif /* RC_ESCAPE_H_ */
