/*
 * rc_frame.c
 *
 *  Created on: Nov 16, 2017
 *      Author: ian
 */
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>

#include <rincmd/rc_frame.h>


static enum e_rc_cli_result rc_frame_init(struct s_rc_frame **rc_frame,
                                          const char *cmd_prefix, const char *cmd_postfix,
                                          const char *resp_prefix, const char *resp_postfix,
                                          bool has_crc)
{
  struct s_rc_frame *frame;
  enum e_rc_cli_result result;

  frame = (struct s_rc_frame *) calloc(1, sizeof(*frame));
  if (NULL == frame)
  {
    result = rcr_allocfail;
    goto rfi_fail;
  }

  frame->cmd_prefix = strdup(cmd_prefix);
  if (NULL == frame->cmd_prefix)
  {
    result = rcr_allocfail;
    goto rfi_fail;
  }
  frame->cmd_postfix = strdup(cmd_postfix);
  if (NULL == frame->cmd_postfix)
  {
    result = rcr_allocfail;
    goto rfi_fail;
  }
  frame->resp_prefix = strdup(resp_prefix);
  if (NULL == frame->resp_prefix)
  {
    result = rcr_allocfail;
    goto rfi_fail;
  }
  frame->resp_postfix = strdup(resp_postfix);
  if (NULL == frame->resp_postfix)
  {
    result = rcr_allocfail;
    goto rfi_fail;
  }

  frame->has_crc = has_crc;

  *rc_frame = frame;
  return rcr_success;

rfi_fail:
  (void) rc_frame_destroy(frame);
  return result;
}


/** Setup framing for CRC messages */
enum e_rc_cli_result rc_frame_init_crc(struct s_rc_frame **rc_frame)
{
  return rc_frame_init(rc_frame, "\x01", "\x04", "\x01", "\x04", true);
}


/**
 * Setup framing for normal messages with command terminated by semicolon
 *
 * Command: No prefix, ; postfix
 * Response: No prefix, \r\n postfix
 */
enum e_rc_cli_result rc_frame_init_normal(struct s_rc_frame **rc_frame)
{
  return rc_frame_init(rc_frame, "", ";", "", "\r\n", false);
}


/**
 * Setup framing for normal messages with commands terminated by carriage return
 *
 * Command: No prefix, \r postfix
 * Response: No prefix, \r\n postfix
 */
enum e_rc_cli_result rc_frame_init_normal_cr(struct s_rc_frame **rc_frame)
{
  return rc_frame_init(rc_frame, "", "\r", "", "\r\n", false);
}


/**
 * Setup framing for normal messages with commands terminated by line feed
 *
 * Command: No prefix, \r postfix
 * Response: No prefix, \r\n postfix
 */
enum e_rc_cli_result rc_frame_init_normal_lf(struct s_rc_frame **rc_frame)
{
  return rc_frame_init(rc_frame, "", "\n", "", "\r\n", false);
}


/** Destroy the framing */
enum e_rc_cli_result rc_frame_destroy(struct s_rc_frame *rc_frame)
{
  if (NULL != rc_frame)
  {
    if (NULL != rc_frame->cmd_prefix)
    {
      free(rc_frame->cmd_prefix);
      rc_frame->cmd_prefix = NULL;
    }
    if (NULL != rc_frame->cmd_postfix)
    {
      free(rc_frame->cmd_postfix);
      rc_frame->cmd_postfix = NULL;
    }
    if (NULL != rc_frame->resp_prefix)
    {
      free(rc_frame->resp_prefix);
      rc_frame->resp_prefix = NULL;
    }
    if (NULL != rc_frame->resp_postfix)
    {
      free(rc_frame->resp_postfix);
      rc_frame->resp_postfix = NULL;
    }

    free(rc_frame);
    rc_frame = NULL;
  }
  return rcr_success;
}


/** Insert the pre-message framing into buffer, setting endpos as the starting position for following message */
enum e_rc_cli_result rc_preframe_message(const struct s_rc_frame *rc_frame, bool cmd_not_resp, char *message, size_t maxlen, size_t *endpos)
{
  const char *prefix;
  int snres;
  enum e_rc_cli_result result;

  if (NULL == rc_frame || NULL == message)
    return rcr_paramfail;

  prefix = (cmd_not_resp) ? rc_frame->cmd_prefix : rc_frame->resp_prefix;
  snres = snprintf(message, maxlen, "%s", prefix);
  if (snres < 0 || snres >= (int) maxlen)
  {
    result = rcr_framingfail;
  }
  else
  {
    if (NULL != endpos)
      *endpos = snres;
    result = rcr_success;
  }
  return result;
}


/** Append the post message framing to the message */
enum e_rc_cli_result rc_postframe_message(const struct s_rc_frame *rc_frame, bool cmd_not_resp, char *message, size_t maxlen)
{
  const char *postfix;
  size_t end;
  int snres;
  enum e_rc_cli_result result;

  if (NULL == rc_frame || NULL == message)
    return rcr_paramfail;

  postfix = (cmd_not_resp) ? rc_frame->cmd_postfix : rc_frame->resp_postfix;
  end = strlen(message);
  if (end < maxlen)
  {
    snres = snprintf(&message[end], maxlen - end, "%s", postfix);
    if (snres < 0 || snres >= (int) maxlen)
    {
      result = rcr_framingfail;
    }
    else
    {
      result = rcr_success;
    }
  }
  else
  {
    result = rcr_framingfail;
  }
  return rcr_success;
}


/**
 * Check the message pre and post framing against what we expect.
 *
 * @param rc_frame is the rincmd framer. Not NULL
 * @param cmd_not_resp is whether to expect command or response framing
 * @param message is the null terminated string to look for the framed message in. Not NULL
 * @param startpos is the position within message that the response begins at (after pre-framing). Can be NULL
 * @param endpos is the position within message that the response ends at (before post-framing). Can be NULL
 * @param msgend is the position within message that the post-framing ends. Can be NULL
 *
 * @return a rincmd client result code
 */
enum e_rc_cli_result rc_frame_response_check(const struct s_rc_frame *rc_frame, bool cmd_not_resp, const char *message, size_t *startpos, size_t *endpos, size_t *msgend)
{
  const char *prefix;
  const char *postfix;
  const char *prefix_start;
  const char *postfix_start;
  ptrdiff_t msgstart_offs;
  ptrdiff_t msgend_offs;

  if (NULL == rc_frame || NULL == message)
    return rcr_paramfail;

  prefix = (cmd_not_resp) ? rc_frame->cmd_prefix : rc_frame->resp_prefix;
  postfix = (cmd_not_resp) ? rc_frame->cmd_postfix : rc_frame->resp_postfix;

  /* Check framing prefix */
  prefix_start = strstr(message, prefix);
  if (NULL == prefix_start)
    return rcr_framingfail;
  msgstart_offs = prefix_start - message + strlen(prefix);

  /* Check framing postfix */
  postfix_start = strstr(&message[msgstart_offs], postfix);
  if (NULL == postfix_start)
    return rcr_framingfail;
  msgend_offs = postfix_start - message;

  if (NULL != startpos)
    *startpos = msgstart_offs;
  if (NULL != endpos)
    *endpos = msgend_offs;
  if (NULL != msgend)
    *msgend = msgend_offs + strlen(postfix);
  return rcr_success;
}
