/*
 * rc_parser.c
 *
 *  Created on: 19 Mar 2015
 *      Author: merrick
 */
#define _GNU_SOURCE
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include "rc_registers.h"
#include "rc_strncpy.h"

#include "rc_parser.h"


#define PARSE_CHAN_OFFSET 8


/**
 * Calculate an unsigned long value from the hexadecimal ASCII input string
 *
 * @param str String to convert
 * @param len Number of characters to convert. Must be <= 8 to fit in 32bits
 * @param value is a pointer to an unsigned long to set with the value of the conversion
 *
 * @return true if conversion succeeded, false otherwise
 */
bool rc_hexstrtoul(const char *str, size_t len, unsigned long *value)
{
  size_t i;
  char c;
  unsigned long num;

  for (i = 0, num = 0; i < len; i++)
  {
    c = str[i];
    num <<= 4;

    if (c >= '0' && c <= '9')
      num |= c - '0';
    else if (c >= 'A' && c <= 'F')
      num |= c - 'A' + 0xA;
    else if (c >= 'a' && c <= 'f')
      num |= c - 'a' + 0xA;
    else
      return false;
  }

  *value = num;
  return true;
}


/**
 * Calculate a 16bit CCITT CRC checksum for a block of data
 *
 * @param byt is an array of bytes to calculate the checksum for
 * @param size is the number of bytes to calculate the checksum over
 * @param crc is the checksum seed
 * @return the checksum of the given data
 */
uint32_t rc_calcCRC16(const char *byt, size_t size, uint32_t crc)
{
  /*
   * Calculate CRC-16 value; uses The CCITT-16 Polynomial,
   * expressed as X^16 + X^12 + X^5 + 1
   */
  size_t index;
  size_t b;

  for (index = 0; index < size; index++)
  {
    crc ^= (((uint32_t) byt[index]) << 8);
    for (b = 0; b < 8; b++)
    {
      if (crc & 0x8000)
        crc = (crc << 1) ^ 0x1021;
      else
        crc <<= 1;
    }
  }
  return crc & 0x0000FFFF;
}


/**
 * Check if the CRC for a buffer is valid
 *
 * @param buf String to computer the CRC of
 * @param crc is the 4 character CRC string
 *
 * @return the checksum of the given data
 */
bool rc_checkCRC(const char *buf, size_t size, const char *crc)
{
  unsigned long msgcrc;

  if (!rc_hexstrtoul(crc, 4, &msgcrc))
    return false;
  return (msgcrc == rc_calcCRC16(buf, size, 0xFFFF));
}


/**
 * Parse rincmd message string from buf into msg
 *
 * @param buf is the incoming message buffer
 * @param buflen is the length of the incoming message buffer
 * @param msg is the rincmd message to fill out from buf
 * @param crc if the message has a CRC that needs to be checked
 *
 * @return fully formed rincmd error code (major + minor)
 */
rc_errcode_t rc_parser_buflen(const char *buf, size_t buflen, struct s_rc_Message *msg, bool crc)
{
  size_t i;
  size_t j;
  unsigned long l;
  size_t left;
  size_t crcposn;

  // Parse the address and check if it's within the limit
  if (buflen < 2 || !rc_hexstrtoul(buf, 2, &l))
    return RC_BUILD_ERR(RC_HNDL_ERR_PARSE, RC_PARSE_ERR_ADDR);
  if (l >= 0xA0)
    return RC_BUILD_ERR(RC_HNDL_ERR_PARSE, RC_PARSE_ERR_ADDR);
  msg->addr = l;

  // Parse the command
  if (buflen < 4 || !rc_hexstrtoul(&buf[2], 2, &l))
    return RC_BUILD_ERR(RC_HNDL_ERR_PARSE, RC_PARSE_ERR_CMD);
  msg->cmd = l;

  // Parse the reg into a signed int
  if (buflen < 8 || !rc_hexstrtoul(&buf[4], 4, &l))
    return RC_BUILD_ERR(RC_HNDL_ERR_PARSE, RC_PARSE_ERR_REG);
  msg->reg = l;

  // Get the number of channel characters
  for (i = PARSE_CHAN_OFFSET, j = 0; buf[i] != ':' && i < buflen && j < RC_SIZE_CHAN - 1; i++, j++)
  {
    msg->chan[j] = buf[i];
  }
  msg->chan[j] = '\0';

  // Check that the buffer was terminated by the colon
  if (buf[i++] != ':')
    return RC_BUILD_ERR(RC_HNDL_ERR_PARSE, RC_PARSE_ERR_DELIM);

  /* Everything following the : is data and optionally a 4 hex digit crc */
  left = buflen - i;
  if (crc)
  {
    /* Check the checksum */
    if (left < RC_SIZE_CRC)
      return RC_BUILD_ERR(RC_HNDL_ERR_CRC, RC_CRC_ERR_PARSE);
    left -= RC_SIZE_CRC;

    crcposn = i + left;
    if (!rc_checkCRC(buf, crcposn, &buf[crcposn]))
      return RC_BUILD_ERR(RC_HNDL_ERR_CRC, RC_CRC_ERR_BAD);
  }

  /* Copy everything following the : upto the optional crc into data buffer */
  if (left > RC_SIZE_BUFFER - 1)
    return RC_BUILD_ERR(RC_HNDL_ERR_PARSE, RC_PARSE_ERR_DATA);
  rc_safe_strncpy(msg->data, &buf[i], left + 1);
  return RC_BUILD_ERR(RC_HNDL_ERR_NONE, RC_PARSE_ERR_NONE);
}


/**
 * Parse rincmd message string from buf into msg
 *
 * @param buf is the incoming message buffer (null terminated)
 * @param msg is the rincmd message to fill out from buf
 * @param crc if the message has a CRC that needs to be checked
 *
 * @return fully formed rincmd error code (major + minor)
 */
rc_errcode_t rc_parser(const char *buf, struct s_rc_Message *msg, bool crc)
{
  return rc_parser_buflen(buf, strlen(buf), msg, crc);
}


/** Convert msg into a rincmd string, suitable to transmit */
rc_errcode_t rc_constructor(char *buf, size_t bufSize, const struct s_rc_Message *msg, bool crc)
{
  size_t snsize;
  size_t bufleft;
  int snres;
  uint16_t crcvalue;

  /* Work out available buffer size with/without crc */
  if (crc)
  {
    if (bufSize < RC_SIZE_CRC)
      return RC_CNSTR_ERR_BUFSIZE;
    snsize = bufSize - RC_SIZE_CRC;
  }
  else
  {
    snsize = bufSize;
  }

  /* Convert msg to string */
  snres = snprintf(buf, snsize, "%02X%02X%04X%s:%s",
                   msg->addr, msg->cmd,
                   msg->reg, msg->chan,
                   msg->data);

  /* Convert snprintf result into rincmd result */
  if (snres < 0)
    return RC_CNSTR_ERR_SNPRINTF;
  else if ((size_t) snres >= snsize)
    return RC_CNSTR_ERR_BUFSIZE;

  if (crc)
  {
    /* Calculate CRC of message buffer and tack it on the end */
    crcvalue = rc_calcCRC16(buf, snres, 0xFFFF);
    bufleft = bufSize - snres;
    snres = snprintf(&buf[snres], bufleft, "%0*X", RC_SIZE_CRC, crcvalue);

    /* Convert snprintf result into rincmd result */
    if (snres < 0)
      return RC_CNSTR_ERR_SNPRINTF;
    else if ((size_t) snres >= snsize)
      return RC_CNSTR_ERR_BUFSIZE;
  }
  return RC_CNSTR_ERR_NONE;
}
