/*
 * rc_parser.h
 *
 *  Created on: 19 Mar 2015
 *      Author: merrick
 */

#ifndef RC_PARSER_H_
#define RC_PARSER_H_

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <rincmd/rc_public.h>


/** Register command enumeration */
enum e_rc_reg_command
{
  RC_MESG_NONE = 0x00,                  /*!< No command */
  RC_MESG_READ_TYPE = 0x01,             /*!< Read the register type (ie RC_REG_TYPE_*, implemented by all registers */
  RC_MESG_READ_RANGE_MIN = 0x02,        /*!< Read the register minimum value */
  RC_MESG_READ_RANGE_MAX = 0x03,        /*!< Read the register maximum value */
  /* RC_MESG_READ_RAW = 0x04, */        /*!< Not implemented at this point */
  RC_MESG_READ_LITERAL = 0x05,          /*!< Read the literal value (ie weight + units) */
  /* RC_MESG_WRITE_RAW = 0x06, */       /*!< Not implemented at this point */
  /* RC_MESG_READ_DEFAULT = 0x07, */    /*!< Not implemented at this point */
  /* RC_MESG_READ_MENU_TEXT = 0x09, */  /*!< Not implemented at this point */
  /* RC_MESG_READ_FULL_TEXT = 0x0A, */  /*!< Not implemented at this point */
  /* RC_MESG_READ_ITEM = 0x0D, */       /*!< Not implemented at this point */
  RC_MESG_READ_PERMISSION = 0x0F,       /*!< Read the register permissions */
  RC_MESG_EXECUTE = 0x10,               /*!< Execute an execute type register (ie RC_REG_TYPE_EXECUTE) */
  RC_MESG_READ_HEXADEC = 0x11,          /*!< Read the register value (as hex for integer, or string) */
  RC_MESG_WRITE_HEXADEC = 0x12,         /*!< Write the register value (as hex for integer, or string) */
  /* RC_MESG_READ_WEIGHT = 0x13, */     /*!< Not implemented at this point */
  /* RC_MESG_READ_MENU_ITEMS = 0x14, */ /*!< Not implemented at this point */
  /* RC_MESG_READ_MENU_DATA = 0x05, */  /*!< Not implemented at this point */
  RC_MESG_READ_DECIMAL = 0x16,          /*!< Read the register value as decimal */
  RC_MESG_WRITE_DECIMAL = 0x17,         /*!< Write the register value as decimal */
  /* RC_MESG_READ_BINARY = 0x18, */     /*!< Not implemented at this point */
  /* RC_MESG_WRITE_BINARY = 0x19, */    /*!< Not implemented at this point */
};


#define RC_IS_READ_COMMAND(cmd) (cmd == RC_MESG_READ_TYPE || \
                                 cmd == RC_MESG_READ_RANGE_MIN || \
                                 cmd == RC_MESG_READ_RANGE_MAX || \
                                 cmd == RC_MESG_READ_LITERAL || \
                                 cmd == RC_MESG_READ_PERMISSION || \
                                 cmd == RC_MESG_READ_HEXADEC || \
                                 cmd == RC_MESG_READ_DECIMAL)

#define RC_IS_WRITE_COMMAND(cmd) (cmd == RC_MESG_WRITE_HEXADEC || \
                                  cmd == RC_MESG_WRITE_DECIMAL)

#define RC_IS_EXEC_COMMAND(cmd) (cmd == RC_MESG_EXECUTE)


//////////////////////////////////////////////////////////////////////////
// Address Defines
//////////////////////////////////////////////////////////////////////////

#define RC_MESG_BROADCAST       0x00

#define RC_MESG_MASK_ID         0x1F
#define RC_MESG_MASK_NOTID      0xE0

#define RC_MESG_SILENT          0x00
#define RC_MESG_REPLY           0x20
#define RC_MESG_ERROR           0x40
#define RC_MESG_RESPONSE        0x80

#define RC_MESG_BAD             (RC_MESG_RESPONSE | RC_MESG_ERROR)

#define RC_ADDR_GET_ID(addr)       (addr & RC_MESG_MASK_ID)
#define RC_ADDR_IS_SILENT(addr)    ((addr & RC_MESG_MASK_NOTID) == RC_MESG_SILENT)
#define RC_ADDR_IS_REPLYREQ(addr)  ((addr & RC_MESG_MASK_NOTID) == RC_MESG_REPLY)
#define RC_ADDR_IS_BADMSG(addr)    (addr & RC_MESG_BAD)
#define RC_ADDR_IS_ERROR(addr)     (addr & RC_MESG_ERROR)
#define RC_ADDR_IS_RESPONSE(addr)  (addr & RC_MESG_RESPONSE)


struct s_rc_Message
{
  int32_t addr;
  int32_t cmd;
  rc_RegId_t reg;
  char chan[RC_SIZE_CHAN];
  char data[RC_SIZE_BUFFER];
};


/**
 * Calculate an unsigned long value from the hexadecimal ASCII input string
 *
 * @param str String to convert
 * @param len Number of characters to convert. Must be <= 8 to fit in 32bits
 * @param value is a pointer to an unsigned long to set with the value of the conversion
 *
 * @return true if conversion succeeded, false otherwise
 */
bool rc_hexstrtoul(const char *str, size_t len, unsigned long *value);


/**
 * Calculate a 16bit CCITT CRC checksum for a block of data
 *
 * @param byt is an array of bytes to calculate the checksum for
 * @param size is the number of bytes to calculate the checksum over
 * @param crc is the checksum seed
 * @return the checksum of the given data
 */
uint32_t rc_calcCRC16(const char *byt, size_t size, uint32_t crc);


/**
 * Check if the CRC for a buffer is valid
 *
 * @param buf String to computer the CRC of
 * @param crc is the 4 character CRC string
 *
 * @return the checksum of the given data
 */
bool rc_checkCRC(const char *buf, size_t size, const char *crc);


/**
 * Parse rincmd message string from buf into msg
 *
 * @param buf is the incoming message buffer
 * @param buflen is the length of the incoming message buffer
 * @param msg is the rincmd message to fill out from buf
 * @param crc if the message has a CRC that needs to be checked
 *
 * @return fully formed rincmd error code (major + minor)
 */
rc_errcode_t rc_parser_buflen(const char *buf, size_t buflen, struct s_rc_Message *msg, bool crc);


/**
 * Parse rincmd message string from buf into msg
 *
 * @param buf is the incoming message buffer (null terminated)
 * @param msg is the rincmd message to fill out from buf
 * @param crc if the message has a CRC that needs to be checked
 *
 * @return fully formed rincmd error code (major + minor)
 */
rc_errcode_t rc_parser(const char *buf, struct s_rc_Message *msg, bool crc);


rc_errcode_t rc_constructor(char *buf, size_t bufSize, const struct s_rc_Message *msg, bool crc);


#endif /* RC_PARSER_H_ */
