/*
 * rc_public.h
 *
 *  Created on: 23 Mar 2016
 *      Author: merrick
 */

#ifndef RC_PUBLIC_H_
#define RC_PUBLIC_H_

#include <stdint.h>
#include <stdbool.h>
#include <rinweigh/adcdatatypes.h>


/** Max stream sets. Make larger than necessary to prevent issues later. */
#define RC_MAX_STREAM_SETS 10

/** Freeable data pointers, to be freed when localRegisterTable is cleaned up. */
#define RC_MAX_LOCAL_DATA_SETS (3 * RC_MAX_STREAM_SETS)

/** Number of data registers that are concatenated and streamed. */
#define RC_STREAM_DATA_REGISTERS  5

#define RC_SIZE_CHAN          9
#define RC_SIZE_BUFFER        200
#define RC_SIZE_CRC           4

#define RC_NULL_REG           {0}
#define RC_NULL_REG_ADDR      0

/** Width of a UINT32 when printed as ASCII hex (eg 0123ABCD). */
#define RC_UINT32_HEX_WIDTH  8

/** A lot of the code relies on this #define, do not change it. */
#define RC_REG_PAGE_SIZE 256


/** Register type enumeration. Do not change existing registers, it will break everything. */
enum e_rc_reg_type
{
  RC_REG_TYPE_INT8 = 0,                     ///< Register Type 0: Simple character value
  RC_REG_TYPE_UINT8,                        ///< Register Type 1: Simple unsigned 8-bit value
  RC_REG_TYPE_INT16,                        ///< Register Type 2: Signed 16-bit value
  RC_REG_TYPE_UINT16,                       ///< Register Type 3: Unsigned 16-bit value
  RC_REG_TYPE_INT32,                        ///< Register Type 4: Signed 32-bit value
  RC_REG_TYPE_UINT32,                       ///< Register Type 5: Unsigned 32-bit value
  RC_REG_TYPE_STRING,                       ///< Register Type 6: Null terminated string
//  RC_REG_TYPE_OPTION = 7,                   ///< Register Type 7: Selectable Option
//  RC_REG_TYPE_MENU = 8,                     ///< Register Type 8: Menu type
  RC_REG_TYPE_WEIGHT = 9,                   ///< Register Type 9: Weight type
//  RC_REG_TYPE_BLOB = 10,                    ///< Register Type 10: Binary Large OBject
  RC_REG_TYPE_EXECUTE = 11,                 ///< Register Type 11: Executable
//  RC_REG_TYPE_BITFIELD = 12,                ///< Register Type 12: Bit Field
  RC_REG_TYPE_STREAM = 13,                  ///< Register Type 13: register for streamed data
//  RC_REG_TYPE_EXECUTE_STR = 14,             ///< Register Type 14: Executable with string param
//  RC_REG_TYPE_MENU_END = 15,                ///< Register Type 15: Menu end type
//  RC_REG_TYPE_STRING_ARRAY = 16,            ///< Register Type 16: String array
//  RC_REG_TYPE_OPTION16 = 17,                ///< Register Type 17: Selectable Option, stored in a 16bit variable
  RC_REG_TYPE_IP = 18,                      ///< Register Type 18: IP Address
  RC_REG_TYPE_REGISTER = 19,                ///< Register Type 19: used for streaming
  RC_REG_TYPE_BINBUFFER = 20,               ///< Register Type 20: Binary buffer (specified by length, not null termination)
};


/** Is the register type an integer */
#define RC_REG_TYPE_IS_INTEGER(regtype) \
  (RC_REG_TYPE_INT8 == regtype || \
   RC_REG_TYPE_UINT8 == regtype || \
   RC_REG_TYPE_INT16 == regtype || \
   RC_REG_TYPE_UINT16 == regtype || \
   RC_REG_TYPE_INT32 == regtype || \
   RC_REG_TYPE_UINT32 == regtype || \
   RC_REG_TYPE_IP == regtype || \
   RC_REG_TYPE_REGISTER == regtype)


/**
 * Function typedef for calling back to your application for debug printfs
 *
 * @param priority is one of the LOG_xxx constants from syslog.h
 * @param format is a printf format string
 * @param args is an variable argument list to match format
 */
typedef void (*t_rc_debug_printf_fn)(int priority, const char *format, va_list args);


/** Set the debug printf function for the rinblock library, call from you application */
void rc_set_debug_printf_fn(t_rc_debug_printf_fn fn);


/** Rincmd register ID type */
typedef int32_t rc_RegId_t;


/** Rincmd functions return this type */
typedef uint32_t rc_errcode_t;


struct s_rc_registerTable
{
  struct s_rc_Register **table[RC_REG_PAGE_SIZE]; // rcRegister pointer within an array (page) within an array (topLevel).
};


struct s_rc_localRegisterTable
{
  struct s_rc_registerTable *global;
  struct s_rc_registerTable *local;
  struct s_rc_Register *streamRegs[RC_MAX_STREAM_SETS];
  uint32_t priv;
  void *localDatasets[RC_MAX_LOCAL_DATA_SETS];
};


//////////////////////////////////////////////////////////////////////////
// DP and Unit options for DispSprtWeightT
//////////////////////////////////////////////////////////////////////////

//#define DISPSPRT_WT_BASIC          0x00     ///< Display the weight with no status. See @ref DispSprtWeightT.
//#define DISPSPRT_WT_SIMPLE         0x01     ///< Display the weight with simple status (ie 1 char: GNEOU). See @ref DispSprtWeightT.
//#define DISPSPRT_WT_FULL           0x02     ///< Display the weight with full status (ie 4 chars). See @ref DispSprtWeightT.

/** Display support structure - helps with decimal point and units */
struct s_rc_DispSprt
{
  uint8_t dp;
  uint8_t units;
  uint8_t unitOther;
};

/** Weight register support structure - helps with decimal point and units, and weight status */
//typedef struct
//{
//  /** Display support information (units and decimal point */
//  rc_DispSprt_t dispSprt;
//
//  /** Weight status formatting */
//  struct
//  {
//    /* The status formatting of the weight */
//    uint8_t statusFmt : 2;
//  } weightFormatting;
//} rc_DispSprtWeight_t;

//////////////////////////////////////////////////////////////////////////
// Register type specific structure definitions
//////////////////////////////////////////////////////////////////////////

struct s_rc_uint8
{
  uint8_t min;
  uint8_t max;
  uint8_t *value;
  struct s_rc_DispSprt *dispSprt;
};

struct s_rc_int8
{
  int8_t min;
  int8_t max;
  int8_t *value;
  struct s_rc_DispSprt *dispSprt;
};

struct s_rc_uint16
{
  uint16_t min;
  uint16_t max;
  uint16_t *value;
  struct s_rc_DispSprt *dispSprt;
};

struct s_rc_int16
{
  int16_t min;
  int16_t max;
  int16_t *value;
  struct s_rc_DispSprt *dispSprt;
};

struct s_rc_uint32
{
  uint32_t min;
  uint32_t max;
  uint32_t *value;
  struct s_rc_DispSprt *dispSprt;
};

struct s_rc_int32
{
  int32_t min;
  int32_t max;
  int32_t *value;
  struct s_rc_DispSprt *dispSprt;
};

/** @ref RC_REG_TYPE_STRING Null terminated string register type */
struct s_rc_StringData
{
  size_t min;
  size_t max;
  char *value;
  char minCh;
  char maxCh;
};

/** @ref RC_REG_TYPE_BINBUFFER Binary buffer (specified by length, not null termination) */
struct s_rc_BinBufferData
{
  size_t min;
  size_t max;
  uint8_t *value;
  uint32_t *length;
  uint8_t minCh;
  uint8_t maxCh;
};

struct s_rc_WeightData
{
  int32_t min;
  int32_t max;
  AdcRdg *value;
};

struct s_rc_StreamData
{
  struct s_rc_stream *data;
};

struct s_rc_RegisterData
{
  rc_RegId_t min;
  rc_RegId_t max;
  rc_RegId_t *value;
};

/** @ref RC_REG_TYPE_IP IP address */
struct s_rc_IPData
{
  uint32_t *value;
};

//////////////////////////////////////////////////////////////////////////
// Register function defines
//////////////////////////////////////////////////////////////////////////

#define RC_FUNC_READ_PRE    1
#define RC_FUNC_READ_POST   2
#define RC_FUNC_WRITE_PRE   3
#define RC_FUNC_WRITE_POST  4
#define RC_FUNC_EXEC        5

//////////////////////////////////////////////////////////////////////////
// Register permissions
//////////////////////////////////////////////////////////////////////////

#define RC_PERM_RD_N       0x00               ///< Register is not readable
#define RC_PERM_RD_S       0x01               ///< Register is readable in safe setup or higher
#define RC_PERM_RD_F       0x02               ///< Register is readable in full setup only
#define RC_PERM_RD_A       0x03               ///< Register is readable by anyone
#define RC_PERM_WR_N       0x00               ///< Register is not writeable
#define RC_PERM_WR_S       0x04               ///< Register is writeable in safe setup or higher
#define RC_PERM_WR_F       0x08               ///< Register is writeable in full setup only
#define RC_PERM_WR_A       0x0C               ///< Register is writeable by anyone
#define RC_PERM_RW_A       (RC_PERM_RD_A|RC_PERM_WR_A)        ///< Register is readable and writeable by anyone
#define RC_PERM_RW_S       (RC_PERM_RD_S|RC_PERM_WR_S)        ///< Register is readable and writeable in safe setup or higher
#define RC_PERM_RW_F       (RC_PERM_RD_F|RC_PERM_WR_F)        ///< Register is readable and writeable in full setup only
#define RC_PERM_RW_N       (RC_PERM_RD_N|RC_PERM_WR_N)        ///< Register is readable and writeable by nobody
#define RC_PERM_SE_NO      0x80               ///< Register has no side effects (doesn't make other registers show/hide, etc)
#define RC_PERM_SE_YES     0x00               ///< Register has side effects (makes other registers show/hide, etc)

#define RC_PERM_INVALID    0x0F
#define RC_PERM_READ_MASK  RC_PERM_RD_A
#define RC_PERM_WRITE_MASK RC_PERM_WR_A
#define RC_PERM_SE_MASK    RC_PERM_SE_NO

#define RC_PRIVILEGE_GENERAL   0
#define RC_PRIVILEGE_SAFE      RC_PERM_RW_S
#define RC_PRIVILEGE_FULL      (RC_PERM_RW_F | RC_PERM_RW_S)

#define RC_HASREADACCESS(perm, priv) ((perm & RC_PERM_READ_MASK) == RC_PERM_RD_A || (priv & perm & RC_PERM_READ_MASK))
#define RC_HASWRITEACCESS(perm, priv) ((perm & RC_PERM_WRITE_MASK) == RC_PERM_WR_A || (priv & perm & RC_PERM_WRITE_MASK))
#define RC_HASEXECACCESS(perm, priv) RC_HASWRITEACCESS(perm, priv)


bool rc_setPrivilege(struct s_rc_localRegisterTable *topLevel, uint32_t priv);


uint32_t rc_getPrivilege(const struct s_rc_localRegisterTable *topLevel);


//////////////////////////////////////////////////////////////////////////
// Top level register structure definitions
//////////////////////////////////////////////////////////////////////////

union u_rc_TypData
{
  struct s_rc_uint8 uint8Data;
  struct s_rc_int8 int8Data;
  struct s_rc_uint16 uint16Data;
  struct s_rc_int16 int16Data;
  struct s_rc_uint32 uint32Data;
  struct s_rc_int32 int32Data;
  struct s_rc_StringData stringData;
  struct s_rc_BinBufferData binbufferData;
  struct s_rc_WeightData weightData;
  struct s_rc_StreamData streamData;
  struct s_rc_RegisterData registerData;
  struct s_rc_IPData ipData;
};


/** Permission structure, see RC_PERM_* constants */
struct s_rc_Perm_t
{
   uint8_t PByte;
};


/** Function typedef for register general purpose function */
typedef bool (*reg_func_ptr_t)(char *val,  size_t len, size_t maxLen,
                               struct s_rc_localRegisterTable *topLevel,
                               struct s_rc_Register *reg, uint8_t opt);


/** The definition of a register */
struct s_rc_Register
{
  rc_RegId_t addr;                   /*!< Address of register 0..0xFFFF */
  enum e_rc_reg_type type;           /*!< Type of register. See RC_REG_TYPE_* */
  struct s_rc_Perm_t perm;           /*!< Register permissions */
  reg_func_ptr_t regFunc;            /*!< General purpose function */
  const union u_rc_TypData *typDat;  /*!< Type specific register data (data ptr, max, min, etc) */
};


/**
 * Structure for encapsulating a the concept of a stream
 * Only automatically generated code and test suite should access this directly
 */
struct s_rc_stream
{
  bool streamEn;                                        /*!< Is streaming enabled? */
  uint8_t counter;                                      /*!< Counter for periodic streaming */
  char chan[RC_SIZE_CHAN];                              /*!< Store the channel that activated the streaming */
  uint8_t freq;                                         /*!< Frequency to stream at */
  rc_RegId_t regAddrs[RC_STREAM_DATA_REGISTERS];        /*!< Address of Registers to stream */
  uint8_t highestStream;                                /*!< Highest streaming register in use */
  struct s_rc_Register *regs[RC_STREAM_DATA_REGISTERS]; /*!< Pointers to register to stream */
  uint32_t streamValues[RC_STREAM_DATA_REGISTERS];      /*!< Old data for stream on change. Stored as uint32 for quick comparison. */
  char streamStrings[RC_STREAM_DATA_REGISTERS][RC_UINT32_HEX_WIDTH + 1]; /*!< Old data for stream on change of string */
};


#endif /* RC_PUBLIC_H_ */
