/*
 * rc_registers.c
 *
 *  Created on: 24 Mar 2015
 *      Author: merrick
 */
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <stddef.h>
#include <rinweigh/adcdatatypes.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "rc_private.h"
#include "rc_streaming.h"
#include "rc_escape.h"

#include "rc_registers.h"


#define BASE_10 10
#define BASE_16 16

#define HEXWIDTH_8BIT 2
#define HEXWIDTH_16BIT 4
#define HEXWIDTH_32BIT 8


/* Converts units code (ie eAdcUnits) to a string */
static const char *const UnitsArr[] = ADC_UNITS_SPACED_STRS;


static const char *const OtherUnitsArr[] =
{
  "",     /* None - RC_DISP_UNITS_OTHER_NONE */
  "m",    /* Minutes - RC_DISP_UNITS_OTHER_M */
  "s",    /* Seconds - RC_DISP_UNITS_OTHER_S */
  "h",    /* Hours - RC_DISP_UNITS_OTHER_H */
  "/",    /* Slash - RC_DISP_UNITS_OTHER_SLASH */
  "/m",   /* Per Minute - RC_DISP_UNITS_OTHER_SLASH_M */
  "/s",   /* Per Second - RC_DISP_UNITS_OTHER_SLASH_S */
  "/h",   /* Per Hour - RC_DISP_UNITS_OTHER_SLASH_H */
  "%"     /* Percent - RC_DISP_UNITS_OTHER_PERCENT */
};


struct s_rc_registerTable *rc_initRegTable(void)
{
  return (struct s_rc_registerTable *) calloc(1, sizeof(struct s_rc_registerTable));
}


void rc_freeRegTable(struct s_rc_registerTable *regTable)
{
  size_t page;

  // Iterate over each page and free it. Attempting to free a pointer to null
  // causes no action to occur.
  for (page = 0; page < RC_REG_PAGE_SIZE; page++)
  {
    free(regTable->table[page]);
    regTable->table[page] = NULL;
  }

  // Free the top level register table
  free(regTable);
}


struct s_rc_localRegisterTable *rc_initLocalRegTable(struct s_rc_registerTable *global)
{
  struct s_rc_localRegisterTable *temp;

  // Allocate a local register table.
  temp = (struct s_rc_localRegisterTable *) calloc(1, sizeof(struct s_rc_localRegisterTable));
  temp->global = global;
  temp->local = rc_initRegTable();
  temp->priv = RC_PRIVILEGE_GENERAL;

  return temp;
}


void rc_freeLocalRegTable(struct s_rc_localRegisterTable *topLevel)
{
  size_t i;

  // Free the local register table
  rc_freeRegTable(topLevel->local);

  // Free the freeable data
  for (i = 0; i < RC_MAX_LOCAL_DATA_SETS; i++)
  {
    free(topLevel->localDatasets[i]);
    topLevel->localDatasets[i] = NULL;
  }

  // Free the top level
  free(topLevel);
}


bool rc_addRegs(struct s_rc_registerTable *regTable, struct s_rc_Register *regArray, size_t size)
{
  size_t i;
  uint32_t page;
  uint32_t loc;

  // Add all registers in the list
  for (i = 0; i < size; i++)
  {
    // Find the page and location
    page = (regArray[i].addr & 0xFF00) >> 8;
    loc  = (regArray[i].addr & 0x00FF);

    // Allocate the page if it doesn't exist, and clear the page
    if (regTable->table[page] == NULL)
      regTable->table[page] = calloc(RC_REG_PAGE_SIZE, sizeof(struct s_rc_Register *));

    // Put a pointer to the register in the table
    regTable->table[page][loc] = &(regArray[i]);
  }

  return true;
}


bool rc_addLocalRegs(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *regArray, size_t size)
{
  return rc_addRegs(topLevel->local, regArray, size);
}


struct s_rc_Register *rc_findReg(struct s_rc_registerTable *regTable, rc_RegId_t addr)
{
  uint32_t page = (addr & 0xFF00) >> 8;
  uint32_t loc  = (addr & 0x00FF);

  // Check if the page and location is valid
  if (regTable->table[page] == NULL || regTable->table[page][loc] == NULL)
    return NULL;

  return regTable->table[page][loc];
}


bool rc_setPrivilege(struct s_rc_localRegisterTable *topLevel, uint32_t priv)
{
  if (priv != RC_PRIVILEGE_GENERAL
      && priv != RC_PRIVILEGE_SAFE
      && priv != RC_PRIVILEGE_FULL)
    return false;

  topLevel->priv = priv;
  return true;
}


uint32_t rc_getPrivilege(const struct s_rc_localRegisterTable *topLevel)
{
  return topLevel->priv;
}


/**
 * Insert a decimal point into a number already in buf.
 * Note that numbers larger than 999999 will be mutilated by this function
 * Note that numbers smaller than -99999 will be mutilated by this function
 *
 * @param buf is a buffer containing a 7 digit zero padded number
 * @param dp is the decimal point position within the buffer (from the right). Zero means no decimal point
 * @param isNegative if set will insert a minus sign in the first character
 */
void rc_formatLit(char buf[RC_LITERAL_WEIGHT_WIDTH + 1], uint8_t dp, bool isNegative)
{
  size_t i;

  if (0 != dp)
  {
    /* There is a DP, shift the characters to the left of the decimal point left one character to make room for the decimal point */
    for (i = 1; i < RC_LITERAL_WEIGHT_WIDTH - dp; i++)
      buf[i - 1] = buf[i];

    /* Add in the decimal point */
    buf[RC_LITERAL_WEIGHT_WIDTH - dp - 1] = '.';

    /* Replace leading 0's with spaces */
    for (i = 0; i < RC_LITERAL_WEIGHT_WIDTH - dp - 2 && '0' == buf[i]; i++)
        buf[i] = ' ';
  }
  else
  {
    /* No DP, replace leading 0's with spaces */
    for (i = 0; i < RC_LITERAL_WEIGHT_WIDTH - 1 && (buf[i] == '0' || buf[i] == '-'); i++)
      buf[i] = ' ';
  }

  /* Move the negative sign to the front */
  if (isNegative)
    buf[0] = '-';
}


/** Convert a signed integer register to a string */
static int rc_int_to_hex_string(char *buf, size_t buflen, uint32_t value, size_t hexwidth)
{
  uint32_t mask;
  size_t i;

  for (i = 0, mask = 0; i < hexwidth; i++)
  {
    mask <<= 4;
    mask |= 0xF;
  }

  return snprintf(buf, buflen, "%0*X", (int) hexwidth, (unsigned int) mask & value);
}


/** Convert a signed integer register to a string (for register types: RC_REG_TYPE_INT8, RC_REG_TYPE_INT16, RC_REG_TYPE_INT32, RC_REG_TYPE_REGISTER */
static rc_errcode_t rc_int_reg_to_string(char *buf, size_t buflen, enum e_rc_reg_command cmd, int32_t min, int32_t max, int32_t value, const struct s_rc_DispSprt *dispsprt, size_t hexwidth)
{
  rc_errcode_t res = RC_READ_ERR_NONE;
  int snres = 0;

  switch (cmd)
  {
    case RC_MESG_READ_RANGE_MIN:
      snres = snprintf(buf, buflen, "%d", (int) min);
      break;

    case RC_MESG_READ_RANGE_MAX:
      snres = snprintf(buf, buflen, "%d", (int) max);
      break;

    case RC_MESG_READ_LITERAL:
      if (NULL == dispsprt)
      {
        /* No display support, just print as a decimal integer */
        goto rirts_print_dec;
      }
      else
      {
        snres = snprintf(buf, buflen - 1, "%07d%s%s", (int) value,
                         UnitsArr[dispsprt->units],
                         OtherUnitsArr[dispsprt->unitOther]);
        rc_formatLit(buf, dispsprt->dp, (value < 0));
      }
      break;

    case RC_MESG_READ_HEXADEC:
      snres = rc_int_to_hex_string(buf, buflen, value, hexwidth);
      break;

    case RC_MESG_READ_DECIMAL:
rirts_print_dec:
      snres = snprintf(buf, buflen, "%d", (int) value);
      break;

    default:
      res = RC_READ_ERR_COMMAND;
      break;
  }

  /* Convert snprintf result into rincmd result */
  if (snres < 0)
    res = RC_READ_ERR_SNPRINTF;
  else if ((size_t) snres >= buflen)
    res = RC_READ_ERR_BUFSIZE;

  return res;
}


/** Convert an unsigned integer register to a string (for register types: RC_REG_TYPE_UINT8, RC_REG_TYPE_UINT16, RC_REG_TYPE_UINT32, RC_REG_TYPE_IP */
static rc_errcode_t rc_uint_reg_to_string(char *buf, size_t buflen, enum e_rc_reg_command cmd, enum e_rc_reg_type regtype,
                                          uint32_t min, uint32_t max, uint32_t value, const struct s_rc_DispSprt *dispsprt, size_t hexwidth)
{
  rc_errcode_t res = RC_READ_ERR_NONE;
  int snres = 0;
  struct in_addr sin_addr;

  switch (cmd)
  {
    case RC_MESG_READ_RANGE_MIN:
      snres = snprintf(buf, buflen, "%u", (unsigned int) min);
      break;

    case RC_MESG_READ_RANGE_MAX:
      snres = snprintf(buf, buflen, "%u", (unsigned int) max);
      break;

    case RC_MESG_READ_LITERAL:
      if (RC_REG_TYPE_IP == regtype)
      {
        /* IP address is a special case */
        sin_addr.s_addr = htonl(value);
        res = (inet_ntop(AF_INET, &sin_addr, buf, buflen) != NULL) ? RC_READ_ERR_NONE : RC_READ_ERR_BUFSIZE;
      }
      else
      {
        /* Other unsigned integer types */
        if (NULL == dispsprt)
        {
          /* No display support, just print as a decimal integer */
          goto rurts_print_dec;
        }
        else
        {
          snres = snprintf(buf, buflen - 1, "%07u%s%s", (unsigned int) value,
                           UnitsArr[dispsprt->units],
                           OtherUnitsArr[dispsprt->unitOther]);
          rc_formatLit(buf, dispsprt->dp, false);
        }
      }
      break;

    case RC_MESG_READ_HEXADEC:
      snres = rc_int_to_hex_string(buf, buflen, value, hexwidth);
      break;

    case RC_MESG_READ_DECIMAL:
rurts_print_dec:
      snres = snprintf(buf, buflen, "%u", (unsigned int) value);
      break;

    default:
      res = RC_READ_ERR_COMMAND;
      break;
  }

  /* Convert snprintf result into rincmd result */
  if (snres < 0)
    res = RC_READ_ERR_SNPRINTF;
  else if ((size_t) snres >= buflen)
    res = RC_READ_ERR_BUFSIZE;

  return res;
}


/** Convert an string register to a string (for register types: RC_REG_TYPE_STRING */
static rc_errcode_t rc_string_reg_to_string(char *buf, size_t buflen, enum e_rc_reg_command cmd, const struct s_rc_StringData *sd)
{
  rc_errcode_t res;

  switch (cmd)
  {
    case RC_MESG_READ_RANGE_MIN:
    case RC_MESG_READ_RANGE_MAX:
      res = rc_uint_reg_to_string(buf, buflen, cmd, RC_REG_TYPE_STRING, sd->min, sd->max, 0, NULL, 0);
      break;

    case RC_MESG_READ_DECIMAL:
    case RC_MESG_READ_HEXADEC:
    case RC_MESG_READ_LITERAL:
      res = (rc_escape_data((uint8_t *) sd->value, strlen(sd->value), buf, buflen) == rer_success) ? RC_READ_ERR_NONE : RC_READ_ERR_BUFSIZE;
      break;

    default:
      res = RC_READ_ERR_COMMAND;
      break;
  }
  return res;
}


/** Convert a weight register to a string (for register types: RC_REG_TYPE_WEIGHT */
static rc_errcode_t rc_weight_reg_to_string(char *buf, size_t buflen, enum e_rc_reg_command cmd, const struct s_rc_WeightData *wd)
{
  rc_errcode_t res = RC_READ_ERR_NONE;
  int snres = 0;

  switch (cmd)
  {
    case RC_MESG_READ_RANGE_MIN:
    case RC_MESG_READ_RANGE_MAX:
    case RC_MESG_READ_DECIMAL:
    case RC_MESG_READ_HEXADEC:
      res = rc_int_reg_to_string(buf, buflen, cmd, wd->min, wd->max, wd->value->rdg, NULL, 8);
      break;

    case RC_MESG_READ_LITERAL:
      snres = snprintf(buf, buflen - 1, "%07"PRId32"%s%s",
                       wd->value->rdg,
                       UnitsArr[wd->value->info.unitdp.bit.units],
                       wd->value->info.status.bit.grossOrNet ? (wd->value->info.status.bit.gross ? " G" : " N") : "");
      rc_formatLit(buf, wd->value->info.unitdp.bit.dp, (wd->value->rdg < 0));
      break;

    default:
      res = RC_READ_ERR_COMMAND;
      break;
  }

  /* Convert snprintf result into rincmd result */
  if (snres < 0)
    res = RC_READ_ERR_SNPRINTF;
  else if ((size_t) snres >= buflen)
    res = RC_READ_ERR_BUFSIZE;

  return res;
}


/** Convert an binary buffer register to a string (for register types: RC_REG_TYPE_BINBUFFER */
static rc_errcode_t rc_binbuffer_reg_to_string(char *buf, size_t buflen, enum e_rc_reg_command cmd, const struct s_rc_BinBufferData *bb)
{
  rc_errcode_t res;

  switch (cmd)
  {
    case RC_MESG_READ_RANGE_MIN:
    case RC_MESG_READ_RANGE_MAX:
      res = rc_uint_reg_to_string(buf, buflen, cmd, RC_REG_TYPE_BINBUFFER, bb->min, bb->max, 0, NULL, 0);
      break;

    case RC_MESG_READ_DECIMAL:
    case RC_MESG_READ_HEXADEC:
    case RC_MESG_READ_LITERAL:
      res = (rc_escape_data(bb->value, *bb->length, buf, buflen) == rer_success) ? RC_READ_ERR_NONE : RC_READ_ERR_BUFSIZE;
      break;

    default:
      res = RC_READ_ERR_COMMAND;
      break;
  }
  return res;
}


rc_errcode_t rc_readReg(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, char *buf, size_t bufSize, enum e_rc_reg_command command)
{
  bool ret;
  rc_errcode_t res = RC_READ_ERR_NONE;

  // Squash ret == false regi
  ret = false;

  // Check register is non-null
  if (reg == NULL)
    return RC_READ_ERR_REGNULL;

  // Check the permissions on the register. Are we allowed to read it?
  if (!RC_HASREADACCESS(reg->perm.PByte, topLevel->priv))
    return RC_READ_ERR_PERMISSION;

  // Check the type data actually exists if this is not a stream reg.
  if (reg->typDat == NULL && reg->type != RC_REG_TYPE_STREAM)
    return RC_READ_ERR_NOTYPDAT;

  // Check the readType is valid
  if (!RC_IS_READ_COMMAND(command))
    return RC_READ_ERR_COMMAND;

  // Call function with pre-read parameter
  if (NULL != reg->regFunc)
    ret = (*reg->regFunc)(buf, bufSize, RC_SIZE_BUFFER, topLevel, reg, RC_FUNC_READ_PRE);

  // Get the data from the register and write it to the data variable
  if (command == RC_MESG_READ_TYPE)
    snprintf(buf, bufSize, "%02X", reg->type);
  else if (command == RC_MESG_READ_PERMISSION)
    snprintf(buf, bufSize, "%02X", reg->perm.PByte);
  else
  {
    switch (reg->type)
    {
      case RC_REG_TYPE_INT8:
        res = rc_int_reg_to_string(buf, bufSize, command,
                                   reg->typDat->int8Data.min, reg->typDat->int8Data.max,
                                   *reg->typDat->int8Data.value, reg->typDat->int8Data.dispSprt,
                                   HEXWIDTH_8BIT);
        break;

      case RC_REG_TYPE_UINT8:
        res = rc_uint_reg_to_string(buf, bufSize, command, reg->type,
                                    reg->typDat->uint8Data.min, reg->typDat->uint8Data.max,
                                    *reg->typDat->uint8Data.value, reg->typDat->uint8Data.dispSprt,
                                    HEXWIDTH_8BIT);
        break;

      case RC_REG_TYPE_INT16:
        res = rc_int_reg_to_string(buf, bufSize, command,
                                   reg->typDat->int16Data.min, reg->typDat->int16Data.max,
                                   *reg->typDat->int16Data.value, reg->typDat->int16Data.dispSprt,
                                   HEXWIDTH_16BIT);
        break;

      case RC_REG_TYPE_UINT16:
        res = rc_uint_reg_to_string(buf, bufSize, command, reg->type,
                                    reg->typDat->uint16Data.min, reg->typDat->uint16Data.max,
                                    *reg->typDat->uint16Data.value, reg->typDat->uint16Data.dispSprt,
                                    HEXWIDTH_16BIT);
        break;

      case RC_REG_TYPE_INT32:
        res = rc_int_reg_to_string(buf, bufSize, command,
                                   reg->typDat->int32Data.min, reg->typDat->int32Data.max,
                                   *reg->typDat->int32Data.value, reg->typDat->int32Data.dispSprt,
                                   HEXWIDTH_32BIT);
        break;

      case RC_REG_TYPE_UINT32:
        res = rc_uint_reg_to_string(buf, bufSize, command, reg->type,
                                    reg->typDat->uint32Data.min, reg->typDat->uint32Data.max,
                                    *reg->typDat->uint32Data.value, reg->typDat->uint32Data.dispSprt,
                                    HEXWIDTH_32BIT);
        break;

      case RC_REG_TYPE_STRING:
        res = rc_string_reg_to_string(buf, bufSize, command,
                                      &reg->typDat->stringData);
        break;

      case RC_REG_TYPE_WEIGHT:
        res = rc_weight_reg_to_string(buf, bufSize, command,
                                      &reg->typDat->weightData);
        break;

      // Handle manual streaming.
      case RC_REG_TYPE_STREAM:
        // Only handle HEXADEC commands. If the pre-read function call returned
        // false, then there was bad data.
        if (command == RC_MESG_READ_HEXADEC)
          res = (!ret) ? RC_READ_ERR_BADDATA : RC_READ_ERR_NONE;
        else
          res = RC_READ_ERR_COMMAND;
        break;

      case RC_REG_TYPE_REGISTER:
        res = rc_int_reg_to_string(buf, bufSize, command,
                                   reg->typDat->registerData.min, reg->typDat->registerData.max,
                                   *reg->typDat->registerData.value, NULL,
                                   HEXWIDTH_32BIT);
        break;

      case RC_REG_TYPE_BINBUFFER:
        res = rc_binbuffer_reg_to_string(buf, bufSize, command,
                                         &reg->typDat->binbufferData);
        break;

      case RC_REG_TYPE_IP:
        res = rc_uint_reg_to_string(buf, bufSize, command, reg->type,
                                    0, UINT32_MAX,
                                    *reg->typDat->ipData.value, NULL,
                                    HEXWIDTH_32BIT);
        break;

      /* case RC_REG_TYPE_EXECUTE: */
      default:
        res = RC_READ_ERR_UNKNOWNTYPE;
        break;
    } /* switch */
  } /* else */

  if (RC_READ_ERR_NONE == res)
  {
    // Call function with post-read parameter
    if (NULL != reg->regFunc)
      (*reg->regFunc)(buf, bufSize, RC_SIZE_BUFFER, topLevel, reg, RC_FUNC_READ_POST);
  }

  return res;
}


/** Convert a string into an integer */
static rc_errcode_t rc_string_to_int(const char *buf, enum e_rc_reg_command cmd, long long int *value_ptr)
{
  size_t i;
  char *endptr;
  int base;

  // Check the buffer is valid, and the readType is valid
  // Convert the data to a long long (needed for range-checking)

  switch (cmd)
  {
    case RC_MESG_WRITE_DECIMAL:
      // Allow for leading '-' character
      if (!isdigit(buf[0]) && buf[0] != '-')
        return RC_WRITE_ERR_PARSE;

      // Check for decimal digits
      for (i = 1; '\0' != buf[i]; i++)
        if (!isdigit(buf[i]))
          return RC_WRITE_ERR_PARSE;

      base = BASE_10;
      break;

    case RC_MESG_WRITE_HEXADEC:
      // Check for hexadecimal characters
      for (i = 0; '\0' != buf[i]; i++)
        if (!isxdigit(buf[i]))
          return RC_WRITE_ERR_PARSE;

      base = BASE_16;
      break;

    default:
      return RC_WRITE_ERR_COMMAND;
      break;
  }

  /* Do the conversion */
  errno = 0;
  *value_ptr = strtoll(buf, &endptr, base);

  if (endptr == buf || *endptr != '\0' || ERANGE == errno)
  {
    /* Something went wrong with strtoll() */
    return RC_WRITE_ERR_STRTOLL;
  }

  return RC_WRITE_ERR_NONE;
}


/** Convert a string into signed integer value (for register types: RC_REG_TYPE_INT8, RC_REG_TYPE_INT16, RC_REG_TYPE_INT32, RC_REG_TYPE_REGISTER */
static rc_errcode_t rc_string_to_int_reg(const char *buf, enum e_rc_reg_command cmd, int32_t min, int32_t max, size_t hexwidth, long long int *value_ptr)
{
  long long mask;
  size_t i;
  long long int value = 0;
  rc_errcode_t res = RC_WRITE_ERR_NONE;

  /* Do the conversion and checks */
  res = rc_string_to_int(buf, cmd, &value);
  if (RC_WRITE_ERR_NONE != res)
    return res;

  /* Check fits in register type size */
  for (i = 0, mask = 0; i < hexwidth; i++)
  {
    mask <<= 4;
    mask |= 0xF;
  }
  mask >>= 1;

  /* Sign extend hexadecimal writes if required: eg 8bit: 0x80 -> 0xFFFFFFFFFFFFFF80(-128d) */
  if (RC_MESG_WRITE_HEXADEC == cmd && value >= 0 && (value & ~mask) == mask + 1)
  {
    value |= ~mask;
  }

  /* Fail if number outside of it's register type size, or outside min..max */
  if (value > mask || value < -mask - 1)
    res = RC_WRITE_ERR_TYPE_LEN;
  else if (value > max)
    res = RC_WRITE_ERR_RANGE_HIGH;
  else if (value < min)
    res = RC_WRITE_ERR_RANGE_LOW;

  *value_ptr = value;
  return res;
}


/** Convert a string into unsigned integer value (for register types: RC_REG_TYPE_UINT8, RC_REG_TYPE_UINT16, RC_REG_TYPE_UINT32 */
static rc_errcode_t rc_string_to_uint_reg(const char *buf, enum e_rc_reg_command cmd, uint32_t min, uint32_t max, size_t hexwidth, long long int *value_ptr)
{
  unsigned long long mask;
  size_t i;
  long long int value = 0;
  rc_errcode_t res = RC_WRITE_ERR_NONE;

  /* Do the conversion and checks */
  res = rc_string_to_int(buf, cmd, &value);
  if (RC_WRITE_ERR_NONE != res)
    return res;

  /* Check fits in register type size */
  for (i = 0, mask = 0; i < hexwidth; i++)
  {
    mask <<= 4;
    mask |= 0xF;
  }

  /* Fail if number outside of it's register type size, or outside min..max */
  if ((value & ~mask) != 0)
    res = RC_WRITE_ERR_TYPE_LEN;
  else if (value > max)
    res = RC_WRITE_ERR_RANGE_HIGH;
  else if (value < min)
    res = RC_WRITE_ERR_RANGE_LOW;

  *value_ptr = value;
  return res;
}


static rc_errcode_t rc_escape_error_to_write_errcode(enum e_rc_escape_result rurres)
{
  const rc_errcode_t codes[num_rer] = { RC_WRITE_ERR_NONE, RC_WRITE_ERR_LENGTH_HIGH, RC_WRITE_ERR_PARSE, RC_WRITE_ERR_RANGE_LOW, RC_WRITE_ERR_RANGE_HIGH };

  return (rurres >= num_rer) ? RC_WRITE_ERR_PARSE : codes[rurres];
}


/** Copy a string into a string register (for register types: RC_REG_TYPE_STRING */
static rc_errcode_t rc_string_to_string_reg(const char *buf, enum e_rc_reg_command cmd, const struct s_rc_StringData *sd)
{
  size_t outbuflen;
  enum e_rc_escape_result rurres;

  if (RC_MESG_WRITE_DECIMAL != cmd && RC_MESG_WRITE_HEXADEC != cmd)
    return RC_WRITE_ERR_COMMAND;

  if (0 == sd->max)
    return RC_WRITE_ERR_BUFSIZE;

  /* First check the un-escaped data, without copying into register data */
  outbuflen = sd->max - 1;
  rurres = rc_unescape_data(buf, NULL, &outbuflen, sd->minCh, sd->maxCh, NULL);
  if (rer_success != rurres)
    return rc_escape_error_to_write_errcode(rurres);

  /* Check string length */
  if (outbuflen > sd->max)
    return RC_WRITE_ERR_LENGTH_HIGH;
  else if (outbuflen < sd->min)
    return RC_WRITE_ERR_LENGTH_LOW;

  /* Everything is ok, now copy into register data */
  outbuflen = sd->max - 1;
  rurres = rc_unescape_data(buf, (uint8_t *) sd->value, &outbuflen, sd->minCh, sd->maxCh, NULL);
  if (rer_success == rurres)
    sd->value[outbuflen] = '\0';
  return rc_escape_error_to_write_errcode(rurres);
}


/** Copy a string into a binary buffer register (for register types: RC_REG_TYPE_BINBUFFER */
static rc_errcode_t rc_string_to_binbuffer_reg(const char *buf, enum e_rc_reg_command cmd, const struct s_rc_BinBufferData *bb)
{
  size_t outbuflen;
  enum e_rc_escape_result rurres;

  if (RC_MESG_WRITE_DECIMAL != cmd && RC_MESG_WRITE_HEXADEC != cmd)
    return RC_WRITE_ERR_COMMAND;

  /* First check the un-escaped data, without copying into register data */
  outbuflen = bb->max;
  rurres = rc_unescape_data(buf, NULL, &outbuflen, bb->minCh, bb->maxCh, NULL);
  if (rer_success != rurres)
    return rc_escape_error_to_write_errcode(rurres);

  /* Check string length */
  if (outbuflen > bb->max)
    return RC_WRITE_ERR_LENGTH_HIGH;
  else if (outbuflen < bb->min)
    return RC_WRITE_ERR_LENGTH_LOW;

  /* Everything is ok, now copy into register data, and keep length */
  outbuflen = bb->max;
  rurres = rc_unescape_data(buf, bb->value, &outbuflen, bb->minCh, bb->maxCh, NULL);
  if (rer_success == rurres)
    *bb->length = outbuflen;
  return rc_escape_error_to_write_errcode(rurres);
}


rc_errcode_t rc_writeReg(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, char *buf, enum e_rc_reg_command command)
{
  long long int value = 0;
  rc_errcode_t res;
  size_t buflen;

  // Check register is non-null
  if (reg == NULL)
    return RC_WRITE_ERR_REGNULL;

  // Check the permissions on the register. Are we allowed to write it?
  if (!RC_HASWRITEACCESS(reg->perm.PByte, topLevel->priv))
    return RC_WRITE_ERR_PERMISSION;

  // Check the type data actually exists
  if (reg->typDat == NULL)
    return RC_READ_ERR_NOTYPDAT;

  // Call function with pre-write parameter
  if (NULL != reg->regFunc)
  {
    buflen = strlen(buf);
    (*reg->regFunc)(buf, buflen, RC_SIZE_BUFFER, topLevel, reg, RC_FUNC_WRITE_PRE);
  }

  // Get the data from the register and write it to the data variable
  switch (reg->type)
  {
    case RC_REG_TYPE_INT8:
      res = rc_string_to_int_reg(buf, command,
                                 reg->typDat->int8Data.min,
                                 reg->typDat->int8Data.max,
                                 HEXWIDTH_8BIT,
                                 &value);
      if (RC_WRITE_ERR_NONE == res)
        *reg->typDat->int8Data.value = value;
      break;

    case RC_REG_TYPE_UINT8:
      res = rc_string_to_uint_reg(buf, command,
                                  reg->typDat->uint8Data.min,
                                  reg->typDat->uint8Data.max,
                                  HEXWIDTH_8BIT,
                                  &value);
      if (RC_WRITE_ERR_NONE == res)
        *reg->typDat->uint8Data.value = value;
      break;

    case RC_REG_TYPE_INT16:
      res = rc_string_to_int_reg(buf, command,
                                 reg->typDat->int16Data.min,
                                 reg->typDat->int16Data.max,
                                 HEXWIDTH_16BIT,
                                 &value);
      if (RC_WRITE_ERR_NONE == res)
        *reg->typDat->int16Data.value = value;
      break;

    case RC_REG_TYPE_UINT16:
      res = rc_string_to_uint_reg(buf, command,
                                  reg->typDat->uint16Data.min,
                                  reg->typDat->uint16Data.max,
                                  HEXWIDTH_16BIT,
                                  &value);
      if (RC_WRITE_ERR_NONE == res)
        *reg->typDat->uint16Data.value = value;
      break;

    case RC_REG_TYPE_INT32:
      res = rc_string_to_int_reg(buf, command,
                                 reg->typDat->int32Data.min,
                                 reg->typDat->int32Data.max,
                                 HEXWIDTH_32BIT,
                                 &value);
      if (RC_WRITE_ERR_NONE == res)
        *reg->typDat->int32Data.value = value;
      break;

    case RC_REG_TYPE_UINT32:
      res = rc_string_to_uint_reg(buf, command,
                                  reg->typDat->uint32Data.min,
                                  reg->typDat->uint32Data.max,
                                  HEXWIDTH_32BIT,
                                  &value);
      if (RC_WRITE_ERR_NONE == res)
        *reg->typDat->uint32Data.value = value;
      break;

    case RC_REG_TYPE_STRING:
      res = rc_string_to_string_reg(buf, command, &reg->typDat->stringData);
      break;

    case RC_REG_TYPE_REGISTER:
      res = rc_string_to_int_reg(buf, command,
                                 reg->typDat->registerData.min,
                                 reg->typDat->registerData.max,
                                 HEXWIDTH_32BIT,
                                 &value);
      if (RC_WRITE_ERR_NONE == res)
        *reg->typDat->registerData.value = value;
      break;

    case RC_REG_TYPE_BINBUFFER:
      res = rc_string_to_binbuffer_reg(buf, command, &reg->typDat->binbufferData);
      break;

    case RC_REG_TYPE_IP:
      res = rc_string_to_uint_reg(buf, command,
                                  0,
                                  UINT32_MAX,
                                  HEXWIDTH_32BIT,
                                  &value);
      if (RC_WRITE_ERR_NONE == res)
        *reg->typDat->ipData.value = value;
      break;

    default:
      res = RC_WRITE_ERR_UNKNOWNTYPE;
      break;
  }

  if (RC_READ_ERR_NONE == res)
  {
    // Call function with post-read parameter
    if (NULL != reg->regFunc)
    {
      buflen = strlen(buf);
      (*reg->regFunc)(buf, buflen, RC_SIZE_BUFFER, topLevel, reg, RC_FUNC_WRITE_POST);
    }
  }
  return res;
}


rc_errcode_t rc_execReg(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, char *buf, size_t bufLen, size_t maxLen, uint32_t command, const char *chan)
{
  rc_errcode_t ret = RC_EXEC_ERR_NONE;

  // Check register is non-null
  if (reg == NULL)
    return RC_EXEC_ERR_REGNULL;

  // Check the permissions on the register. Are we allowed to exec it?
  if (!RC_HASEXECACCESS(reg->perm.PByte, topLevel->priv))
    return RC_EXEC_ERR_PERMISSION;

  // Check the register is an execute or stream register. If it is an execute,
  // typDat should be a NULL pointer.
  if (reg->type == RC_REG_TYPE_EXECUTE)
  {
    if (reg->typDat != NULL)
      return RC_EXEC_ERR_TYPDAT;
  }
  else if (reg->type != RC_REG_TYPE_STREAM)
    return RC_EXEC_ERR_UNKNOWNTYPE;

  // Check the command
  if (command != RC_MESG_EXECUTE)
    return RC_EXEC_ERR_COMMAND;

  // Call function with exec parameter
  if (reg->regFunc && (*reg->regFunc)(buf, bufLen, maxLen, topLevel, reg, RC_FUNC_EXEC))
    ret = RC_EXEC_ERR_WRITEBACK;

  // If there's no errors at this point and this is a stream register, save
  // the channel with the register so it can be streamed out.
  // We can ignore the return of rc_setChan as we've already checked the size
  // of the channel compared to the buffer.
  if (reg->type == RC_REG_TYPE_STREAM)
    rc_setChan(reg->typDat->streamData.data, chan, RC_SIZE_CHAN);

  return ret;
}


/**
 * Rincmd command handler
 *
 * @param inputBuf is a null terminated rincmd message string
 * @param crc if the message has a CRC checksum attached
 * @param outputBuf is the buffer into which the reply is placed
 * @param topLevel is the register table
 * @param indAddr is the address of the indicator
 *
 * Return RC_HNDL_RET_REPLY if outputBuf must be sent as a reply
 *        RC_HNDL_RET_NOREPLY if no action to be taken
 */
rc_errcode_t rc_commandHandler(const char *inputBuf, bool crc,
                               char *outputBuf, size_t outputBufSize,
                               struct s_rc_localRegisterTable *topLevel, uint32_t indAddr)
{
  struct s_rc_Message msg;
  struct s_rc_Register *reg = NULL;
  rc_errcode_t err = RC_NO_ERROR;
  rc_errcode_t minorErr = 0;
  rc_errcode_t cnstrerr;

  memset(&msg, 0, sizeof(msg));

  // Extract the message. If an error occurs, we've got no idea if this is
  // even for us, so do nothing.
  err = rc_parser(inputBuf, &msg, crc);
  if (RC_NO_ERROR != err)
  {
    if (RC_ERROR_MAJOR(err) == RC_HNDL_ERR_CRC)
    {
        goto RC_REPLY_HANDLER;
    }
    else
    {
      return RC_HNDL_RET_ERR;
    }
  }

  // Check indicator address. Don't reply if the message isn't for us.
  if ((RC_ADDR_GET_ID(msg.addr) != indAddr) && (RC_ADDR_GET_ID(msg.addr) != RC_MESG_BROADCAST))
    return RC_HNDL_RET_NOREPLY;

  // Check if message is something we shouldn't handle.
  if (RC_ADDR_IS_BADMSG(msg.addr))
    return RC_HNDL_RET_NOREPLY;

  // Look up the register. Error if the register doesn't exist.
  reg = rc_findReg(topLevel->global, msg.reg);
  if (reg == NULL)
  {
    reg = rc_findReg(topLevel->local, msg.reg);
    if (reg == NULL)
    {
      err = RC_BUILD_ERR(RC_HNDL_ERR_REG, RC_REG_ERR_NONE);
      goto RC_REPLY_HANDLER;
    }
  }

  // Call the relevant command
  if (RC_IS_READ_COMMAND(msg.cmd))
  {
    // Read a register into the message data. Handle an error if it exists.
    minorErr = rc_readReg(topLevel, reg, msg.data, RC_SIZE_BUFFER, msg.cmd);
    if (minorErr != RC_READ_ERR_NONE)
    {
      err = RC_BUILD_ERR(RC_HNDL_ERR_READ, minorErr);
      goto RC_REPLY_HANDLER;
    }
  }
  else if (RC_IS_WRITE_COMMAND(msg.cmd))
  {
    // Write the input buffer to a register. Handle an error if it exists.
    minorErr = rc_writeReg(topLevel, reg, msg.data, msg.cmd);
    if (minorErr != RC_WRITE_ERR_NONE)
    {
      err = RC_BUILD_ERR(RC_HNDL_ERR_WRITE, minorErr);
      goto RC_REPLY_HANDLER;
    }
    snprintf(msg.data, RC_SIZE_BUFFER, "%04X", RC_NO_ERROR);
  }
  else if (RC_IS_EXEC_COMMAND(msg.cmd))
  {
    // Execute the input buffer on a register. Handle an error if it exists.
    minorErr = rc_execReg(topLevel, reg,
                          msg.data, strlen(msg.data), RC_SIZE_BUFFER,
                          msg.cmd, msg.chan);
    // Handle returning data
    if (minorErr == RC_EXEC_ERR_WRITEBACK)
    {
      // Set the error to NO_ERROR and do nothing here.
      minorErr = RC_NO_ERROR;
    }
    // Otherwise if this isn't NONE something is wrong
    else if (minorErr != RC_EXEC_ERR_NONE)
    {
      err = RC_BUILD_ERR(RC_HNDL_ERR_EXEC, minorErr);
      goto RC_REPLY_HANDLER;
    }
    // There's no error, so write this into the data buffer.
    else
    {
      snprintf(msg.data, RC_SIZE_BUFFER, "%04X", RC_NO_ERROR);
    }
  }
  else
  {
    // Looks like a non-implemented command was given.
    err = RC_BUILD_ERR(RC_HNDL_ERR_NOTIMP, RC_NONE_ERR_NONE);
    goto RC_REPLY_HANDLER;
  }

  // Reply handling code begins here.
RC_REPLY_HANDLER:

  // If this is a 'silent' message. Return without doing anything else.
  if (RC_ADDR_IS_SILENT(msg.addr))
    return RC_HNDL_RET_NOREPLY;

  // Construct the new address
  if (err)
  {
    // On an error, we set the error bit and clear the data
    msg.addr = RC_MESG_RESPONSE | RC_MESG_ERROR | indAddr;
    snprintf(msg.data, RC_SIZE_BUFFER, "%04X", err);
  }
  else
  {
    msg.addr = RC_MESG_RESPONSE | indAddr;
  }

  /* Construct the reply */
  cnstrerr = rc_constructor(outputBuf, outputBufSize, &msg, crc);
  return (RC_CNSTR_ERR_NONE == cnstrerr) ? RC_HNDL_RET_REPLY : RC_HNDL_RET_ERR;
}
