/*
 * rc_streaming.c
 *
 *  Created on: 21 Jan 2016
 *      Author: merrick
 */
#include <inttypes.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "rc_private.h"
#include "rc_streaming.h"


static bool rc_streamDataReader(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *streamReg,
                                char *buf, size_t bufSize, bool onChange, int32_t *regReadError);


struct s_rc_stream *rc_initStreamData(void)
{
  size_t i;
  struct s_rc_stream *tmp;

  tmp = (struct s_rc_stream *) calloc(1, sizeof(struct s_rc_stream));
  if (NULL == tmp)
    return NULL;

  tmp->streamEn = false;
  tmp->freq = RC_STREAM_MODE_OFF;

  for (i = 0; i < RC_STREAM_DATA_REGISTERS; i++)
    strcpy(tmp->streamStrings[i], "00000000");

  return tmp;
}


void rc_freeStreamData(struct s_rc_stream *data)
{
  free(data);
}


bool rc_addFreeableData(struct s_rc_localRegisterTable *topLevel, void *freeable)
{
  size_t i;

  // Find the end of the data set
  for (i = 0; i < RC_MAX_LOCAL_DATA_SETS && topLevel->localDatasets[i] != NULL; i++)
    ;

  if (i == RC_MAX_LOCAL_DATA_SETS)
    return false;

  topLevel->localDatasets[i] = freeable;

  return true;
}


void rc_setChan(struct s_rc_stream *data, const char *chan, size_t size)
{
  snprintf(data->chan, (size > RC_SIZE_CHAN) ? RC_SIZE_CHAN : size, "%s", chan);
}


uint32_t rc_getStreamFreqMask(const struct s_rc_localRegisterTable *topLevel)
{
  uint32_t mask;
  size_t i;

  // Iterate across the streams, create the mask.
  for (i = 0, mask = 0; (i < RC_MAX_STREAM_SETS) && (topLevel->streamRegs[i] != NULL); i++)
    mask |= RC_STREAM_MAKEMASK(topLevel->streamRegs[i]->typDat->streamData.data->freq);

  return mask;
}


bool rc_addStreams(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *regPtrArray[])
{
  size_t i;
  size_t j;

  // Find the end of the stream array.
  for (j = 0; topLevel->streamRegs[j] != NULL; j++)
    ;

  // Add each register.
  i = 0;
  while (regPtrArray[i] != NULL)
  {
    // Check that the limit of the topLevel hasn't been reached.
    if (j >= RC_MAX_STREAM_SETS)
      return false;

    // Copy the pointer to the streamReg
    topLevel->streamRegs[j] = regPtrArray[i];

    // Increment the counters.
    i++; j++;
  }

  return true;
}


/** Enable streaming if buf is "1", disable if "0". Ignore all other cases. */
bool rc_setStreaming(char *buf, size_t bufSize, size_t maxLen,
                     struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, uint8_t opt)
{
  int32_t regReadError = 0;

  if (opt == RC_FUNC_EXEC)
  {
    // If 0, turn off the stream enable data and clear the stream counter.
    if ((bufSize == 1 && buf[0] == '0') || (bufSize >= 2 && buf[0] == '0' && buf[1] == 0))
    {
      reg->typDat->streamData.data->streamEn = false;
      reg->typDat->streamData.data->counter = 0;
    }

    // If 1, turn on the stream enable data.
    if ((bufSize == 1 && buf[0] == '1') || (bufSize >= 2 && buf[0] == '1' && buf[1] == 0))
    {
      reg->typDat->streamData.data->streamEn = true;
      reg->typDat->streamData.data->counter = 0;
    }
  }
  else if (opt == RC_FUNC_READ_PRE)
  {
    // Read the stream register.

    // Check the bufsize.
    if (bufSize < (RC_STREAM_DATA_REGISTERS * RC_UINT32_HEX_WIDTH + 1))
      return false;

    rc_streamDataReader(topLevel, reg, buf, maxLen, false, &regReadError);

    // There was some error reading a stream register.
    if (regReadError > 0)
      return false;

    return true;
  }

  return false;
}


/** Check if the register can be streamed */
static bool rc_isRegisterStreamable(const struct s_rc_localRegisterTable *topLevel, const struct s_rc_Register *reg)
{
  // Check register is non-null
  if (reg == NULL)
    return false;

  // Check the permissions on the register. Are we allowed to read it?
  if (!RC_HASREADACCESS(reg->perm.PByte, topLevel->priv))
    return false;

  // Check the type data actually exists if this is not a stream reg.
  if (reg->typDat == NULL && reg->type != RC_REG_TYPE_STREAM)
    return false;

  // Check the register type is streamable.
  if (reg->type > RC_REG_TYPE_UINT32 && reg->type != RC_REG_TYPE_WEIGHT)
    return false;

  return true;
}


/** Process a register after it has been added to be streamed */
bool rc_handleAddStreamDataReg(char *buf, size_t bufSize, size_t maxLen,
                               struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, uint8_t opt)
{
  size_t i;
  struct s_rc_stream *streamData = NULL;
  rc_RegId_t offset = 0;
  struct s_rc_Register *tempReg = {0};

  if (opt == RC_FUNC_WRITE_POST)
  {
    // Find streamData for this register. The address should be within the
    // offset of the data register.
    for (i = 0; i < RC_MAX_STREAM_SETS; i++)
    {
      offset = reg->addr - (topLevel->streamRegs[i]->addr + RC_STREAM_REG_DATA_OFFSET);
      if (offset >= 0 && offset <= RC_STREAM_DATA_REGISTERS)
      {
        streamData = topLevel->streamRegs[i]->typDat->streamData.data;
        break;
      }
    }

    // This will only occur if this function is called on a non-stream data reg.
    if (streamData == NULL)
      return true;

    // Look up the data register from the addr.
    tempReg = rc_findReg(topLevel->global, streamData->regAddrs[offset]);
    if (tempReg == NULL)
    {
      tempReg = rc_findReg(topLevel->local, streamData->regAddrs[offset]);
    }

    // Regardless of whether this is valid or not, attempting to set the reg
    // will clear the values
    streamData->regs[offset] = NULL;
    streamData->streamValues[offset] = 0;
    strcpy(streamData->streamStrings[offset], "00000000");

    // Check that this register is streamable. If it is, set the pointer.
    if (rc_isRegisterStreamable(topLevel, tempReg))
      streamData->regs[offset] = tempReg;

    // Look up the data registers and find the highest data register actually
    // set.
    for (i = 0; i < RC_STREAM_DATA_REGISTERS; i++)
    {
      // Check if the regAddr is valid
      if (streamData->regAddrs[i] != RC_NULL_REG_ADDR && streamData->regs[i] != NULL)
        streamData->highestStream = i + 1;
    }
  }

  return true;
}


/** Reset all stream data to default for this local register table */
void rc_resetStreaming(struct s_rc_localRegisterTable *topLevel)
{
  size_t i;

  // For each stream register, set the streamData to 0
  for (i = 0; (topLevel->streamRegs[i] != NULL) && (i < RC_MAX_STREAM_SETS); i++)
    memset(topLevel->streamRegs[i]->typDat->streamData.data, 0, sizeof(struct s_rc_stream));
}


/** Read a register value into a uint32_t */
bool fastReadReg(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *reg, uint32_t *val)
{
  if (reg == NULL)
    return false;

  // Call function with pre-read parameter
  if (reg->regFunc)
    (*reg->regFunc)(NULL, 0, RC_SIZE_BUFFER, topLevel, reg, RC_FUNC_READ_PRE);

  if (reg->type == RC_REG_TYPE_INT8)
    *val = (uint32_t) *(reg->typDat->int8Data.value);
  else if (reg->type == RC_REG_TYPE_UINT8)
    *val = (uint32_t) *(reg->typDat->uint8Data.value);
  else if (reg->type == RC_REG_TYPE_INT16)
    *val = (uint32_t) *(reg->typDat->int16Data.value);
  else if (reg->type == RC_REG_TYPE_UINT16)
    *val = (uint32_t) *(reg->typDat->uint16Data.value);
  else if (reg->type == RC_REG_TYPE_INT32)
    *val = (uint32_t) *(reg->typDat->int32Data.value);
  else if (reg->type == RC_REG_TYPE_UINT32)
    *val = (uint32_t) *(reg->typDat->uint32Data.value);
  else if (reg->type == RC_REG_TYPE_WEIGHT)
    *val = (uint32_t) reg->typDat->weightData.value->rdg;
  else
    return false;

  // Call function with post-read parameter
  if (reg->regFunc)
    (*reg->regFunc)(NULL, 0, RC_SIZE_BUFFER, topLevel, reg, RC_FUNC_READ_POST);

  return true;
}


static bool rc_streamDataReader(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *streamReg,
                         char *buf, size_t bufSize, bool onChange, int32_t *regReadError)
{
  uint16_t i;
  struct s_rc_stream *streamData = streamReg->typDat->streamData.data;
  bool sendStream = !onChange;  // Don't send the stream by default if onChange is true.
  uint32_t streamValue = 0;

  // Clear the buffer
  buf[0] = '\0';

  // Look up the data registers and read their data into the output buffer.
  for (i = 0; i < streamData->highestStream; i++)
  {
    // Read the register. Set the error flag if there was a problem.
    if (fastReadReg(topLevel, streamData->regs[i], &streamValue) == false)
      *regReadError = RC_MESG_BAD;
    // If there wasn't a problem, check if the string needs to be updated.
    else if (streamValue != streamData->streamValues[i])
    {
      snprintf(streamData->streamStrings[i], RC_UINT32_HEX_WIDTH + 1, "%08" PRIX32, streamValue);
      streamData->streamValues[i] = streamValue;
      sendStream = true;
    }
    strcat(buf, streamData->streamStrings[i]);
  }

  return sendStream;
}


static bool rc_streamBuilder(struct s_rc_localRegisterTable *topLevel, struct s_rc_Register *streamReg,
                             char *buf, size_t bufSize, int32_t addr, bool onChange)
{
  struct s_rc_Message msg;
  int32_t regReadError = RC_MESG_RESPONSE;

  memset(&msg, 0, sizeof(msg));

  // If onChange is true, send the data.
  if (rc_streamDataReader(topLevel, streamReg, msg.data, RC_SIZE_BUFFER,
                          onChange, &regReadError))
  {
    // Build the message
    msg.addr = regReadError | addr;
    msg.cmd = RC_MESG_READ_HEXADEC;
    strcpy(msg.chan, streamReg->typDat->streamData.data->chan);
    msg.reg = streamReg->addr;

    rc_constructor(buf, bufSize, &msg, false);

    return true;
  }
  else
    return false;
}


/**
 * This function should be called at 10Hz with each of the stream registers.
 * It will fill the buffer, and if RC_STREAM_ERR_NONE is returned, this buffer
 * should be framed (TERMINATE WITH \r\n or ;) and sent to the comm port.
 */
rc_errcode_t rc_streamDataFreq(struct s_rc_localRegisterTable *topLevel, uint8_t streamNo,
                               char *buf, size_t bufSize, int32_t addr)
{
  struct s_rc_stream *streamData;

  // Check if the streamNo register exists
  if (topLevel->streamRegs[streamNo] == NULL)
    return RC_STREAM_ERR_NULLREG;

  streamData = topLevel->streamRegs[streamNo]->typDat->streamData.data;

  // Do nothing if streaming is disabled.
  if (streamData->streamEn == false)
    return RC_STREAM_ERR_DISABLED;

  // Increment the counter on the streaming register.
  // Use modulo 10 here as that is the fastest streaming frequency.
  streamData->counter = (streamData->counter + 1) % 10;

  // Do we need to stream? Stream if this is a 10Hz stream, or it is a
  // 5Hz or 1Hz stream that aligns with the number of counts.
  if ((streamData->freq == RC_STREAM_MODE_10HZ)
      || ((streamData->freq == RC_STREAM_MODE_5HZ) && (streamData->counter % 5 == 0))
      || ((streamData->freq == RC_STREAM_MODE_1HZ) && (streamData->counter % 10 == 0)))
  {
    rc_streamBuilder(topLevel, topLevel->streamRegs[streamNo], buf, bufSize, addr, false);
  }
  else
  {
    // We have no data to stream.
    return RC_STREAM_ERR_NODATA;
  }

  return RC_STREAM_ERR_NONE;
}

/**
 * This function should be called at sync freq with each of the stream registers.
 * It will fill the buffer, and if RC_STREAM_ERR_NONE is returned, this buffer
 * should be framed (TERMINATE WITH \r\n or ;) and sent to the comm port.
 */
rc_errcode_t rc_streamDataSync(struct s_rc_localRegisterTable *topLevel, uint8_t streamNo,
                               char *buf, size_t bufSize, int32_t addr)
{
  struct s_rc_stream *streamData;

  // Check if the streamNo register exists
  if (topLevel->streamRegs[streamNo] == NULL)
    return RC_STREAM_ERR_NULLREG;

  streamData = topLevel->streamRegs[streamNo]->typDat->streamData.data;

  // Do nothing if streaming is disabled.
  if (streamData->streamEn == false)
    return RC_STREAM_ERR_DISABLED;

  // Do we need to stream?
  if (streamData->freq == RC_STREAM_MODE_AUTO)
  {
    rc_streamBuilder(topLevel, topLevel->streamRegs[streamNo], buf, bufSize, addr, false);
  }
  else if (streamData->freq == RC_STREAM_MODE_CHANGE)
  {
    if (rc_streamBuilder(topLevel, topLevel->streamRegs[streamNo], buf,
          bufSize, addr, true) == false)
      return RC_STREAM_ERR_NODATA;
  }
  else
  {
    // We have no data to stream.
    return RC_STREAM_ERR_NODATA;
  }

  return RC_STREAM_ERR_NONE;
}
