/*
 * rc_timespec.c
 *
 *  Created on: Nov 29, 2017
 *      Author: ian
 */
#include <time.h>

#include <rincmd/rc_timespec.h>

#if 0
/*
 * Returns true if the timespec is norm, false if denorm:
 */
#define timespec_valid(ts) \
  (((ts)->tv_sec >= 0) && (((unsigned long) (ts)->tv_nsec) < NSEC_PER_SEC))
#endif


/**
 * Add the 'struct timespec' values x and y, storing the result in result.
 * ie: result = x+y
 *
 * Negative results are always normalised as per the kernel (kernel/time.c set_normalized_timespec()):
 *   Note: The tv_nsec part is always in the range of
 *    0 <= tv_nsec < NSEC_PER_SEC
 *   For negative values only the tv_sec field is negative !
 * So -1.-000000001 would be returned as -1.000000001
 * However if tv_sec is 0, then:
 *  0.-000000001 would be returned unchanged
 */
void rc_timespec_add(struct timespec *result, const struct timespec *x, const struct timespec *y)
{
  struct timespec x1 = *x;
  struct timespec y1 = *y;

  /*
   * It isn't clear what defines a negative timespec:
   *  Is tv_sec < 0 && tv_nsec < 0 equal to tv_sec < 0 && tv_nsec > 0
   *  ie: -1.-000000001 equal to -1.000000001
   * What about when tv_sec = 0:
   *  ie 0.-000000001
   */

  /* Normalise sign of tv_nsec to negative */
  if (x1.tv_sec < 0 && x1.tv_nsec > 0)
    x1.tv_nsec = -x1.tv_nsec;

  if (y1.tv_sec < 0 && y1.tv_nsec > 0)
    y1.tv_nsec = -y1.tv_nsec;

  /* Add tv_nsecs and carry any overflow nanosecond to second */
  result->tv_nsec = x1.tv_nsec + y1.tv_nsec;
  result->tv_sec = x1.tv_sec + y1.tv_sec;
  if (result->tv_nsec >= RC_NANO_SEC_PER_SEC)
  {
    result->tv_nsec -= RC_NANO_SEC_PER_SEC;
    result->tv_sec += 1;
  }
  else if (result->tv_nsec <= -RC_NANO_SEC_PER_SEC)
  {
    result->tv_nsec += RC_NANO_SEC_PER_SEC;
    result->tv_sec -= 1;
  }
  else if (result->tv_nsec < 0 && result->tv_sec > 0)
  {
    result->tv_nsec += RC_NANO_SEC_PER_SEC;
    result->tv_sec -= 1;
  }

  /* Normalise sign of tv_nsec to to +ve if tv_sec is negative */
  if (result->tv_sec < 0 && result->tv_nsec < 0)
    result->tv_nsec = -result->tv_nsec;
}


/**
 * Subtract the 'struct timespec' values X and Y, storing the result in RESULT.
 * ie: result = x-y
 */
void rc_timespec_subtract(struct timespec *result, const struct timespec *x, const struct timespec *y)
{
  struct timespec x1 = *x;
  struct timespec y1 = *y;

  /* Normalise sign of tv_nsec to negative */
  if (x1.tv_sec < 0 && x1.tv_nsec > 0)
    x1.tv_nsec = -x1.tv_nsec;

  if (y1.tv_sec < 0 && y1.tv_nsec > 0)
    y1.tv_nsec = -y1.tv_nsec;

  /*
   * Cases:
   *   +ve x_ns > +ve y_ns: 5 - 2 = 3,    r_ns = 3, r_s unchanged
   *   +ve x_ns < +ve y_ns: 2 - 5 = -3,   r_ns = -3, r_s unchanged
   *   -ve x_ns > -ve y_ns: -2 - -5 = 3,  r_ns = 3, r_s unchanged
   *   -ve x_ns < -ve y_ns: -5 - -2 = -3, r_ns = -3, r_s unchanged
   *
   *   +ve x_ns > -ve y_ns: 3 - -2 = 5, r_ns = 5, r_s unchanged
   *   -ve x_ns < +ve y_ns: -2 - 3 = -5, r_ns = -5, r_s unchanged
   *
   *   +ve x_ns = +ve y_ns: 3 - 3 = 0, r_ns = 0, r_s unchanged
   *   -ve x_ns = -ve y_ns: -3 - -3 = 0, r_ns = 0, r_s unchanged
   *
   */

  /*
   * 2's compliment subtraction:
   *
   * minuend - subtrahend = difference
   *
   * Overflow occurs if difference has same sign as subtrahend
   * Ie: (+A) - (-B) = -C
   *     (-A) - (+B) = +C
   */

  result->tv_nsec = x1.tv_nsec - y1.tv_nsec;
  result->tv_sec = x1.tv_sec - y1.tv_sec;
  if (y1.tv_nsec < 0 && result->tv_nsec >= RC_NANO_SEC_PER_SEC)
  {
    /* Correct */
    result->tv_nsec -= RC_NANO_SEC_PER_SEC;
    result->tv_sec += 1;
  }
  else if (y1.tv_nsec > 0 && result->tv_nsec <= -RC_NANO_SEC_PER_SEC)
  {
    result->tv_nsec += RC_NANO_SEC_PER_SEC;
    result->tv_sec -= 1;
  }
  else if (result->tv_sec > 0 && result->tv_nsec < 0)
  {
    result->tv_nsec += RC_NANO_SEC_PER_SEC;
    result->tv_sec -= 1;
  }

  /* Normalise sign of tv_nsec to to +ve if tv_sec is negative */
  if (result->tv_sec < 0 && result->tv_nsec < 0)
    result->tv_nsec = -result->tv_nsec;
}


/**
 * Compare two 'struct timespec' values x and y
 *
 * Return -1 if the difference is negative (x < y), 1 if positive (x > y), 0 if equal (x = y)
 */
int rc_timespec_compare(const struct timespec *x, const struct timespec *y)
{
  int result;

  /* Work out result */
  if (x->tv_sec < y->tv_sec)
  {
    result = -1;
  }
  else if (x->tv_sec > y->tv_sec)
  {
    result = 1;
  }
  else
  {
    if (x->tv_nsec < y->tv_nsec)
      result = -1;
    else if (x->tv_nsec > y->tv_nsec)
      result = 1;
    else
      result = 0;
  }
  return result;
}
