/*
 * rc_client_async.c
 *
 *  Created on: Dec 15, 2017
 *      Author: ian
 */
#define _XOPEN_SOURCE 700
#include <time.h>
#include <stdbool.h>
#include <stdio.h>
#include <sys/select.h>
#include <rincmd/rc_public.h>
#include <rincmd/rc_cli_async.h>
#include <rincmd/rc_cli_async_stream.h>
#include <rincmd/rc_cli_error.h>
#include <rincmd/rc_cli_async_bulk.h>
#include "rc_client_sync.h"

#include "rc_client_async.h"

#define NUM_BULK_READS 3


struct s_int_reg_callback_data
{
  rc_RegId_t register_id;
  bool streaming;
  bool bulk;
  struct s_rc_async_client *rac;
  unsigned count;
  struct timespec *timeout;
  bool *quit;
  rc_RegId_t stream_reg_base;
  unsigned repeats;
  rc_RegId_t bulk_registers[NUM_BULK_READS];
};


static void stream_register_teardown_callback(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct s_int_reg_callback_data *cbd = userdata;

  switch (result)
  {
    case rcr_streamsetupfail:
      printf("Streaming teardown failed\n");
      *cbd->quit = true;
      break;
    case rcr_streamsetupsuccess:
      printf("Streaming teardown success\n");
      break;
    case rcr_timeout:
      printf("Timeout waiting for teardown\n");
      break;
    default:
      printf("Unexpected result for teardown, result %d\n", (int) result);
      break;
  }
  *cbd->quit = true;
}


static void stream_register_read_callback(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct s_int_reg_callback_data *cbd = userdata;
  int64_t regvalues[RC_STREAM_DATA_REGISTERS] = { 0, 0, 0, 0, 0 };

  switch (result)
  {
    case rcr_streamsetupfail:
      printf("Streaming setup failed\n");
      *cbd->quit = true;
      break;
    case rcr_streamsetupsuccess:
      printf("Streaming setup success, waiting for streaming results\n");
      break;
    case rcr_timeout:
      printf("Response %u: Timeout waiting for register read\n", cbd->count);
      break;
    case rcr_success:
      if (NULL != msg)
      {
        result = rc_async_client_parse_stream_register_read(cbd->rac, cbd->stream_reg_base, msg, regvalues);
        if (rcr_success != result)
          printf("Error parsing response %u: %s\n", cbd->count, msg->data);
        else
          printf("Streaming result %u: %lld, %lld, %lld, %lld, %lld\n", (unsigned) cbd->count,
                 (long long) regvalues[0], (long long) regvalues[1], (long long) regvalues[2], (long long) regvalues[3], (long long) regvalues[4]);
      }
      else
      {
        printf("Response %u: Unexpected callback\n", cbd->count);
      }

      cbd->count++;
      if (cbd->repeats > 0 && cbd->count >= cbd->repeats)
      {
        /* All done teardown the stream */
        result = rc_async_client_streaming_teardown(cbd->rac, 0x0040, stream_register_teardown_callback, cbd, cbd->timeout);
        if (rcr_success != result)
          printf("Error initiating teardown %d\n", (int) result);
      }
      break;
    default:
      printf("Streaming error %d\n", (int) result);
      *cbd->quit = true;
      break;
  }
}


static enum e_rc_cli_result setup_async_stream(struct s_int_reg_callback_data *cbd)
{
  /*                                           Disp    Gross   Net     Status  Error */
  const int regs[RC_STREAM_DATA_REGISTERS] = { 0x0025, 0x0026, 0x0027, 0x0021, 0x0023 };

  return rc_async_client_streaming_setup(cbd->rac, cbd->stream_reg_base,
                                         stream_register_read_callback, cbd,
                                         RC_STREAM_MODE_AUTO,
                                         regs,
                                         cbd->timeout,
                                         cbd->timeout);
}


static enum e_rc_cli_result setup_async_bulk_read(struct s_int_reg_callback_data *cbd, bool first);


static void bulk_int_register_read_callback(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct s_int_reg_callback_data *cbd = userdata;
  int64_t regvalue;
  unsigned regidx;

  if (NULL == msg)
  {
    printf("Response %u: Unexpected msg == NULL\n", cbd->count / NUM_BULK_READS);
    *cbd->quit = true;
  }
  else
  {
    for (regidx = 0; regidx < NUM_BULK_READS; regidx++)
    {
      if (msg->reg == cbd->bulk_registers[regidx])
        break;
    }
    if (regidx >= NUM_BULK_READS)
    {
      printf("Response %u: Unexpected register id 0x%04X\n", cbd->count / NUM_BULK_READS, (unsigned) msg->reg);
      *cbd->quit = true;
      return;
    }

    switch (result)
    {
      case rcr_success:
        result = rc_async_client_parse_int_register_read(cbd->rac, msg->cmd, msg->reg, msg, &regvalue);
        if (rcr_success != result)
          printf("Error parsing response %u, register 0x%04X: %s\n", cbd->count / NUM_BULK_READS, (unsigned) msg->reg, msg->data);
        else
          printf("Response %u, register 0x%04X: %lld\n", (unsigned) cbd->count / NUM_BULK_READS, (unsigned) msg->reg, (long long) regvalue);

        cbd->count++;
        if (cbd->repeats > 0 && cbd->count >= cbd->repeats)
        {
          *cbd->quit = true;
        }
        else if ((cbd->count % NUM_BULK_READS) == 0)
        {
          result = setup_async_bulk_read(cbd, false);
          if (rcr_success != result)
          {
            printf("Bulk register request failed with error %s\n", rc_cli_error_strings(result));
            *cbd->quit = true;
          }
        }
        break;
      case rcr_timeout:
        printf("Response %u: Timeout waiting for register read\n", cbd->count / NUM_BULK_READS);
        *cbd->quit = true;
        break;
      default:
        printf("Response %u: Error %s\n", cbd->count / NUM_BULK_READS, rc_cli_error_strings(result));
        *cbd->quit = true;
        break;
    }
  }
}


static enum e_rc_cli_result setup_async_bulk_read(struct s_int_reg_callback_data *cbd, bool first)
{
  struct s_rc_async_client_request requests[NUM_BULK_READS];

  enum e_rc_cli_result result;

  if (first)
  {
    cbd->bulk_registers[0] = 0x0025;
    cbd->bulk_registers[1] = 0x0026;
    cbd->bulk_registers[2] = 0x0027;
    cbd->repeats *= NUM_BULK_READS;
  }

  result = rc_async_client_read_request_fillout(&requests[0], RC_MESG_READ_DECIMAL, 0x0025, bulk_int_register_read_callback, cbd, cbd->timeout);
  if (rcr_success != result)
    return result;
  result = rc_async_client_read_request_fillout(&requests[1], RC_MESG_READ_HEXADEC, 0x0026, bulk_int_register_read_callback, cbd, cbd->timeout);
  if (rcr_success != result)
    return result;
  result = rc_async_client_read_request_fillout(&requests[2], RC_MESG_READ_DECIMAL, 0x0027, bulk_int_register_read_callback, cbd, cbd->timeout);
  if (rcr_success != result)
    return result;

  return rc_async_client_bulk_request(cbd->rac, requests, NUM_BULK_READS);
}


static void int_register_read_callback(void *userdata, const struct s_rc_Message *msg, enum e_rc_cli_result result)
{
  struct s_int_reg_callback_data *cbd = userdata;
  int64_t regvalue;

  if (NULL == msg)
  {
    printf("Response %u: Unexpected msg == NULL\n", cbd->count);
    *cbd->quit = true;
  }
  else
  {
    switch (result)
    {
      case rcr_success:
        result = rc_async_client_parse_int_register_read(cbd->rac, RC_MESG_READ_DECIMAL, cbd->register_id, msg, &regvalue);
        if (rcr_success != result)
          printf("Error parsing response %u: %s\n", cbd->count, msg->data);
        else
          printf("Response %u: %lld\n", (unsigned) cbd->count, (long long) regvalue);

        cbd->count++;
        if (cbd->repeats > 0 && cbd->count >= cbd->repeats)
        {
          *cbd->quit = true;
        }
        else
        {
          result = rc_async_client_register_read(cbd->rac, RC_MESG_READ_DECIMAL, cbd->register_id, int_register_read_callback, userdata, cbd->timeout);
          if (rcr_success != result)
          {
            printf("Requesting register failed with error %s\n", rc_cli_error_strings(result));
            *cbd->quit = true;
          }
        }
        break;
      case rcr_timeout:
        printf("Response %u: Timeout waiting for register read\n", cbd->count);
        *cbd->quit = true;
        break;
      default:
        printf("Response %u: Error %s\n", cbd->count, rc_cli_error_strings(result));
        *cbd->quit = true;
        break;
    }
  }
}


/** Called when the socket connects */
static void comms_connected_callback(void *userdata, bool connect_not_disconnect, enum e_rc_cli_result result)
{
  struct s_int_reg_callback_data *cbd = userdata;

  if (connect_not_disconnect)
  {
    printf("Connect callback with result %s\n", rc_cli_error_strings(result));
    if (rcr_success == result)
    {
      /* Streaming or poll/response */
      if (cbd->streaming)
        result = setup_async_stream(cbd);
      else if (cbd->bulk)
        result = setup_async_bulk_read(cbd, true);
      else
        result = rc_async_client_register_read(cbd->rac, RC_MESG_READ_DECIMAL, cbd->register_id, int_register_read_callback, cbd, cbd->timeout);
      if (rcr_success != result)
      {
        printf("Initiate stream/register read failed with result %s\n", rc_cli_error_strings(result));
        *cbd->quit = true;
      }
    }
    else
    {
      *cbd->quit = true;
    }
  }
  else
  {
    printf("Disconnect callback with result %s\n", rc_cli_error_strings(result));
    result = rcr_success;
  }
}


/** Asynchronous client example */
enum e_rc_cli_result async_client_example(struct s_rc_cli_comms *rcc, rc_RegId_t regid,
                                          bool usecrc, bool streaming, bool bulk,
                                          unsigned repeats, int timeoutms, bool *quit)
{
  enum e_rc_cli_result result;
  struct s_rc_async_client *rac = NULL;
  fd_set readfds;
  fd_set writefds;
  struct timespec to;
  struct timespec *toptr;
  bool hastimeout;
  int readfd;
  int connectfd;
  int maxfd;
  int sres;
  struct s_int_reg_callback_data cbd;

  toptr = get_timeout_ptr(&to, timeoutms);

  /* Create client and initiate connect */
  result = rc_async_client_init(&rac, rcc, RC_MESG_BROADCAST, usecrc, 40, comms_connected_callback, &cbd, NULL, NULL);
  if (rcr_success != result)
  {
    goto ac_fail;
  }

  /* Initiate a read */
  cbd.register_id = regid;
  cbd.streaming = streaming;
  cbd.bulk = bulk;
  cbd.rac = rac;
  cbd.count = 0;
  cbd.timeout = toptr;
  cbd.quit = quit;
  cbd.stream_reg_base = 0x0040;
  cbd.repeats = repeats;

  result = rc_async_client_connect(rac, toptr);
  if (rcr_success != result)
  {
    goto ac_fail;
  }

  /* Run the handler until we get connected, then setup read/streaming and get a response */
  for (*quit = false; !*quit; )
  {
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    result = rc_async_client_pre_select(rac, &readfd, &connectfd, &hastimeout, &to);
    if (rcr_success != result)
    {
      goto ac_fail;
    }
    maxfd = -1;
    if (readfd >= 0)
    {
      FD_SET(readfd, &readfds);
      if (readfd > maxfd)
        maxfd = readfd;
    }
    if (connectfd >= 0)
    {
      FD_SET(connectfd, &writefds);
      if (connectfd > maxfd)
        maxfd = connectfd;
    }
    toptr = (hastimeout) ? &to : NULL;
    sres = pselect(maxfd + 1, &readfds, &writefds, NULL, toptr, NULL);
    if (0 == sres)
    {
      /* Timeout, keep looping */
    }
    else if (sres < 0)
    {
      result = rcr_reply_error;
      goto ac_fail;
    }
    else if (readfd >= 0 && FD_ISSET(readfd, &readfds))
    {
      result = rc_async_client_post_select(rac, true, false);
      if (rcr_success != result)
        goto ac_fail;
    }
    else if (connectfd >= 0 && FD_ISSET(connectfd, &writefds))
    {
      result = rc_async_client_post_select(rac, false, true);
      if (rcr_success != result)
        goto ac_fail;
    }
  }

  /* Disconnect from the server */
  (void) rc_async_client_disconnect(rac);
  result = rcr_success;

ac_fail:
  rc_async_client_destroy(rac);
  return result;
}
