/*
 * test_rc_cli_async_timer.c
 *
 *  Created on: 24 May 2018
 *      Author: ian
 */
#include <stdio.h>
#include <sys/select.h>
#include <cutest/CuTest.h>
#include <rincmd/rc_cli_async_timer.h>
#include "test_rc_timespec.h"

#include "test_rc_cli_async_timer.h"


static struct timespec fake_time;
static bool use_fake_time = false;


#if 0
static void set_fake_time_ts(const struct timespec *ft)
{
  use_fake_time = true;
  fake_time = *ft;
}
#endif


static void set_fake_time_sec_nsec(time_t seconds, signed long nano_seconds)
{
  use_fake_time = true;
  fake_time.tv_sec = seconds;
  fake_time.tv_nsec = nano_seconds;
}


enum e_rc_cli_result rc_cli_external_gettime(struct timespec *ts)
{
  if (NULL == ts)
  {
    return rcr_paramfail;
  }
  else if (use_fake_time)
  {
    *ts = fake_time;
    return rcr_success;
  }
  else
  {
    return (clock_gettime(CLOCK_MONOTONIC, ts) == 0) ? rcr_success : rcr_failure;
  }
}


static struct s_rc_async_timer *make_async_timer(CuTest *tc)
{
  struct s_rc_async_timer *rat = NULL;
  enum e_rc_cli_result result;

  result = rc_async_timer_init(&rat, 20);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertPtrNotNull(tc, rat);

  return rat;
}


static void destroy_async_timer(CuTest *tc, struct s_rc_async_timer *rat)
{
  enum e_rc_cli_result result;

  result = rc_async_timer_destroy(rat);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_cli_async_timer_init(CuTest *tc)
{
  struct s_rc_async_timer *rat;

  rat = make_async_timer(tc);
  destroy_async_timer(tc, rat);
}


static void test_rc_cli_async_no_timers(CuTest *tc)
{
  struct s_rc_async_timer *rat;
  enum e_rc_cli_result result;
  bool hastimeout;
  struct timespec next_timeout;

  rat = make_async_timer(tc);

  /* No timers have been added, so there should not be a timeout */
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, !hastimeout);

  result = rc_async_timer_destroy(rat);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void async_timer_complete_callback(void *userdata, intptr_t timer_id)
{
  bool *completed = userdata;

  *completed = true;
}


static void check_rc_cli_async_timer_one_shot(CuTest *tc, struct s_rc_async_timer *rat, bool *timer_completed)
{
  enum e_rc_cli_result result;
  bool hastimeout;
  struct timespec next_timeout;
  struct timespec expected;

  /* Run the timer @ 100.000000000 seconds */
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, !*timer_completed);

  expected.tv_sec = 5;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* 2.5 seconds have passed */
  set_fake_time_sec_nsec(102, 500000000);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, !*timer_completed);

  expected.tv_sec = 2;
  expected.tv_nsec = 500000000;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* 5.1 seconds have passed, one shot timer will be removed, and there will be no new timeout */
  set_fake_time_sec_nsec(105, 100000000);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, !hastimeout);
  CuAssertTrue(tc, *timer_completed);
}


static void test_rc_cli_async_timer_one_shot_relative(CuTest *tc)
{
  struct s_rc_async_timer *rat;
  enum e_rc_cli_result result;
  struct timespec timeout;
  bool timer_completed;

  rat = make_async_timer(tc);

  /* Add the timer */
  set_fake_time_sec_nsec(100, 0);
  timeout.tv_sec = 5;
  timeout.tv_nsec = 0;
  timer_completed = false;
  result = rc_async_timer_add_oneshot_relative(rat, async_timer_complete_callback, &timer_completed, &timeout, NULL);

  /* Run the test */
  check_rc_cli_async_timer_one_shot(tc, rat, &timer_completed);

  /* Clean up */
  result = rc_async_timer_destroy(rat);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_cli_async_timer_one_shot_absolute(CuTest *tc)
{
  struct s_rc_async_timer *rat;
  enum e_rc_cli_result result;
  struct timespec timeout;
  bool timer_completed;

  rat = make_async_timer(tc);

  /* Add the timer */
  set_fake_time_sec_nsec(100, 0);
  timeout.tv_sec = 105;
  timeout.tv_nsec = 0;
  timer_completed = false;
  result = rc_async_timer_add_oneshot_absolute(rat, async_timer_complete_callback, &timer_completed, &timeout, NULL);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Run the test */
  check_rc_cli_async_timer_one_shot(tc, rat, &timer_completed);

  /* Clean up */
  result = rc_async_timer_destroy(rat);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_cli_async_timer_persistent(CuTest *tc)
{
  struct s_rc_async_timer *rat;
  enum e_rc_cli_result result;
  struct timespec timeout;
  bool timer_completed;
  bool hastimeout;
  struct timespec next_timeout;
  struct timespec expected;

  rat = make_async_timer(tc);

  /* Add the timer every 5 seconds */
  set_fake_time_sec_nsec(100, 0);
  timeout.tv_sec = 5;
  timeout.tv_nsec = 0;
  timer_completed = false;
  result = rc_async_timer_add_persistent(rat, async_timer_complete_callback, &timer_completed, &timeout, NULL);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Run the timer @ 100.000000000 seconds */
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, !timer_completed);

  expected.tv_sec = 5;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* 2.5 seconds have passed */
  set_fake_time_sec_nsec(102, 500000000);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, !timer_completed);

  expected.tv_sec = 2;
  expected.tv_nsec = 500000000;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* 5.1 seconds have passed, persistent timer still exists, and will run again in 5.000000000 seconds */
  set_fake_time_sec_nsec(105, 100000000);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, timer_completed);

  expected.tv_sec = 5;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* Clean up */
  result = rc_async_timer_destroy(rat);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_cli_async_timer_multiple(CuTest *tc)
{
  struct s_rc_async_timer *rat;
  enum e_rc_cli_result result;
  struct timespec timeout;
  bool timer_persistent_completed;
  bool timer_one_shot_rel_completed;
  bool timer_one_shot_abs_completed;
  bool hastimeout;
  struct timespec next_timeout;
  struct timespec expected;

  rat = make_async_timer(tc);

  /* Add a persistent timer every 2 seconds */
  set_fake_time_sec_nsec(100, 0);
  timeout.tv_sec = 2;
  timeout.tv_nsec = 0;
  timer_persistent_completed = false;
  result = rc_async_timer_add_persistent(rat, async_timer_complete_callback, &timer_persistent_completed, &timeout, NULL);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Add a relative one shot timer for 4 seconds */
  timeout.tv_sec = 4;
  timeout.tv_nsec = 0;
  timer_one_shot_rel_completed = false;
  result = rc_async_timer_add_oneshot_relative(rat, async_timer_complete_callback, &timer_one_shot_rel_completed, &timeout, NULL);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Add a absolute one shot timer for 105 seconds */
  timeout.tv_sec = 105;
  timeout.tv_nsec = 0;
  timer_one_shot_abs_completed = false;
  result = rc_async_timer_add_oneshot_absolute(rat, async_timer_complete_callback, &timer_one_shot_abs_completed, &timeout, NULL);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Run the timer @ 100.000000000 seconds */
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, !timer_persistent_completed);
  CuAssertTrue(tc, !timer_one_shot_rel_completed);
  CuAssertTrue(tc, !timer_one_shot_abs_completed);

  expected.tv_sec = 2;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* 2 seconds have passed, persistent timer runs */
  set_fake_time_sec_nsec(102, 0);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, timer_persistent_completed);
  timer_persistent_completed = false;
  CuAssertTrue(tc, !timer_one_shot_rel_completed);
  CuAssertTrue(tc, !timer_one_shot_abs_completed);

  expected.tv_sec = 2;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* 4 seconds have passed, persistent timer and relative one shot timers run */
  set_fake_time_sec_nsec(104, 0);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, timer_persistent_completed);
  timer_persistent_completed = false;
  CuAssertTrue(tc, timer_one_shot_rel_completed);
  CuAssertTrue(tc, !timer_one_shot_abs_completed);

  expected.tv_sec = 1;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* 5 seconds have passed, absolute one shot timer runs, one second until next persistent timer runs */
  set_fake_time_sec_nsec(105, 0);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, !timer_persistent_completed);
  CuAssertTrue(tc, timer_one_shot_rel_completed);
  CuAssertTrue(tc, timer_one_shot_abs_completed);

  expected.tv_sec = 1;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* 6 seconds have passed, persistent timer runs */
  set_fake_time_sec_nsec(106, 0);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, timer_persistent_completed);
  timer_persistent_completed = false;
  CuAssertTrue(tc, timer_one_shot_rel_completed);
  CuAssertTrue(tc, timer_one_shot_abs_completed);

  expected.tv_sec = 2;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* Clean up */
  result = rc_async_timer_destroy(rat);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_cli_async_timer_remove(CuTest *tc)
{
  struct s_rc_async_timer *rat;
  enum e_rc_cli_result result;
  struct timespec timeout;
  bool hastimeout;
  struct timespec next_timeout;
  struct timespec expected;
  intptr_t persistent_timer_id;
  bool persistent_timer_completed;
  intptr_t oneshot_timer_id;
  bool oneshot_timer_completed;

  rat = make_async_timer(tc);

  /* Add persistent timer every 2 seconds */
  set_fake_time_sec_nsec(100, 0);
  timeout.tv_sec = 2;
  timeout.tv_nsec = 0;
  persistent_timer_completed = false;
  result = rc_async_timer_add_persistent(rat, async_timer_complete_callback, &persistent_timer_completed, &timeout, &persistent_timer_id);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Add one shot timer for 2 seconds time */
  timeout.tv_sec = 2;
  timeout.tv_nsec = 0;
  oneshot_timer_completed = false;
  result = rc_async_timer_add_oneshot_relative(rat, async_timer_complete_callback, &oneshot_timer_completed, &timeout, &oneshot_timer_id);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Now remove one shot before it even gets into the active list */
  result = rc_async_timer_remove_callback(rat, oneshot_timer_id);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Run the timer @ 102.000000000 seconds, only the persistent timer should run */
  set_fake_time_sec_nsec(102, 0);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, persistent_timer_completed);
  persistent_timer_completed = false;
  CuAssertTrue(tc, !oneshot_timer_completed);

  expected.tv_sec = 2;
  expected.tv_nsec = 0;
  CuAssertTimeSpecEquals(tc, expected, next_timeout);

  /* Now remove persistent */
  result = rc_async_timer_remove_callback(rat, persistent_timer_id);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Run the timer @ 104.000000000 seconds, no timers should run */
  set_fake_time_sec_nsec(104, 0);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, !hastimeout);
  CuAssertTrue(tc, !persistent_timer_completed);
  CuAssertTrue(tc, !oneshot_timer_completed);

  /* Clean up */
  result = rc_async_timer_destroy(rat);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_cli_async_timer_get(CuTest *tc)
{
  struct s_rc_async_timer *rat;
  enum e_rc_cli_result result;
  struct timespec timeout;
  bool hastimeout;
  struct timespec next_timeout;
  intptr_t persistent_timer_id;
  bool persistent_timer_completed;
  intptr_t oneshot_timer_id;
  bool oneshot_timer_completed;
  t_rc_async_timer_callback_fn callback;
  void *userdata;

  rat = make_async_timer(tc);

  /* Add persistent timer every 2 seconds */
  set_fake_time_sec_nsec(100, 0);
  timeout.tv_sec = 2;
  timeout.tv_nsec = 0;
  persistent_timer_completed = false;
  result = rc_async_timer_add_persistent(rat, async_timer_complete_callback, &persistent_timer_completed, &timeout, &persistent_timer_id);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Add one shot timer for 1 seconds time */
  timeout.tv_sec = 1;
  timeout.tv_nsec = 0;
  oneshot_timer_completed = false;
  result = rc_async_timer_add_oneshot_relative(rat, async_timer_complete_callback, &oneshot_timer_completed, &timeout, &oneshot_timer_id);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Get the persistent timer */
  callback = NULL;
  userdata = NULL;
  timeout.tv_sec = 0, timeout.tv_nsec = 0;
  result = rc_async_timer_get(rat, persistent_timer_id, &callback, &userdata, &timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertPtrEquals(tc, callback, async_timer_complete_callback);
  CuAssertPtrEquals(tc, userdata, &persistent_timer_completed);
  CuAssertIntEquals(tc, 102, timeout.tv_sec);
  CuAssertIntEquals(tc, 0, timeout.tv_nsec);

  /* Get the one shot timer */
  callback = NULL;
  userdata = NULL;
  timeout.tv_sec = 0, timeout.tv_nsec = 0;
  result = rc_async_timer_get(rat, oneshot_timer_id, &callback, &userdata, &timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertPtrEquals(tc, callback, async_timer_complete_callback);
  CuAssertPtrEquals(tc, userdata, &oneshot_timer_completed);
  CuAssertIntEquals(tc, 101, timeout.tv_sec);
  CuAssertIntEquals(tc, 0, timeout.tv_nsec);

  /* Run the timer @ 101.000000000 seconds, only the one shot timer should run */
  set_fake_time_sec_nsec(101, 0);
  result = rc_async_timer_run(rat, &hastimeout, &next_timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, hastimeout);
  CuAssertTrue(tc, !persistent_timer_completed);
  persistent_timer_completed = false;
  CuAssertTrue(tc, oneshot_timer_completed);

  /* Get the persistent timer (should still be available) */
  callback = NULL;
  userdata = NULL;
  timeout.tv_sec = 0, timeout.tv_nsec = 0;
  result = rc_async_timer_get(rat, persistent_timer_id, &callback, &userdata, &timeout);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertPtrEquals(tc, callback, async_timer_complete_callback);
  CuAssertPtrEquals(tc, userdata, &persistent_timer_completed);
  CuAssertIntEquals(tc, 102, timeout.tv_sec);
  CuAssertIntEquals(tc, 0, timeout.tv_nsec);

  /* Get the one shot timer (should no longer be available) */
  callback = NULL;
  userdata = NULL;
  timeout.tv_sec = 0, timeout.tv_nsec = 0;
  result = rc_async_timer_get(rat, oneshot_timer_id, &callback, &userdata, &timeout);
  CuAssertIntEquals(tc, rcr_listempty, result);

  /* Clean up */
  result = rc_async_timer_destroy(rat);
  CuAssertIntEquals(tc, rcr_success, result);
}


CuSuite *CuGetSuite_rc_cli_async_timer(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rc_cli_async_timer_init);
  SUITE_ADD_TEST(suite, test_rc_cli_async_no_timers);
  SUITE_ADD_TEST(suite, test_rc_cli_async_timer_one_shot_relative);
  SUITE_ADD_TEST(suite, test_rc_cli_async_timer_one_shot_absolute);
  SUITE_ADD_TEST(suite, test_rc_cli_async_timer_persistent);
  SUITE_ADD_TEST(suite, test_rc_cli_async_timer_multiple);
  SUITE_ADD_TEST(suite, test_rc_cli_async_timer_remove);
  SUITE_ADD_TEST(suite, test_rc_cli_async_timer_get);

  return suite;
}
