/*
 * test_rc_cli_parse.c
 *
 *  Created on: Nov 22, 2017
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <cutest/CuTest.h>
#include <rincmd/rc_cli_parse.h>
#include <rincmd/rc_frame.h>
#include <rincmd/rc_strncpy.h>

#include "test_rc_cli_parse.h"

#define TEST_BUFFER_LEN 100
#define TEST_RC_ADDR 0x0A


static struct s_rc_cli_parser *make_cli_parse(CuTest *tc, bool use_crc, bool use_channel, uint32_t channel)
{
  struct s_rc_frame *frame = NULL;
  struct s_rc_cli_parser *rcp = NULL;
  enum e_rc_cli_result result;

  result = (use_crc) ? rc_frame_init_crc(&frame) : rc_frame_init_normal(&frame);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertPtrNotNull(tc, frame);

  result = rc_cli_parse_init(&rcp, frame, TEST_RC_ADDR, use_channel, channel);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertPtrNotNull(tc, rcp);

  return rcp;
}


static void destroy_cli_parse(CuTest *tc, struct s_rc_cli_parser *rcp)
{
  rc_frame_destroy(rcp->framing);
  rc_cli_parse_destroy(rcp);
}


static void test_rc_cli_assemble_command_crc(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  char buffer[TEST_BUFFER_LEN];
  const struct s_rc_Message cmd = { .addr=0x2A, .cmd=RC_MESG_READ_DECIMAL, .reg=0x1234, .chan="", .data="testdata" };

  rcp = make_cli_parse(tc, true, false, 0);

  result = rc_cli_assemble_command(rcp, &cmd, buffer, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "\x01" "2A161234:testdata9D74\x04", buffer);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_assemble_command_normal(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  char buffer[TEST_BUFFER_LEN];
  const struct s_rc_Message cmd = { .addr=0x2A, .cmd=RC_MESG_READ_DECIMAL, .reg=0x1234, .chan="", .data="testdata" };

  rcp = make_cli_parse(tc, false, false, 0);

  result = rc_cli_assemble_command(rcp, &cmd, buffer, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "2A161234:testdata;", buffer);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_assemble_command_channel(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  char buffer[TEST_BUFFER_LEN];
  const struct s_rc_Message cmd = { .addr=0x2A, .cmd=RC_MESG_READ_DECIMAL, .reg=0x1234, .chan="ABCDEF01", .data="testdata" };

  rcp = make_cli_parse(tc, false, false, 0);

  result = rc_cli_assemble_command(rcp, &cmd, buffer, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "2A161234ABCDEF01:testdata;", buffer);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_parse_response_crc(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  char buffer[TEST_BUFFER_LEN] = "\x01" "2A161234:testdata9D74\x04";
  struct s_rc_Message resp;

  rcp = make_cli_parse(tc, true, false, 0);

  result = rc_cli_parse_response(rcp, &resp, buffer);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 0x2A, resp.addr);
  CuAssertIntEquals(tc, 0x16, resp.cmd);
  CuAssertIntEquals(tc, 0x1234, resp.reg);
  CuAssertStrEquals(tc, "", resp.chan);
  CuAssertStrEquals(tc, "testdata", resp.data);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_check_response_header(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  struct s_rc_Message resp;
  uint32_t returncode;

  rcp = make_cli_parse(tc, true, false, 0);

  /* Passing test */
  resp.addr = TEST_RC_ADDR | RC_MESG_RESPONSE;
  resp.cmd = RC_MESG_READ_DECIMAL;
  resp.reg = 0x1234;
  rc_safe_strncpy(resp.chan, "", RC_SIZE_CHAN);
  rc_safe_strncpy(resp.data, "89AB", RC_SIZE_BUFFER);
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_READ_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Not a reply */
  resp.addr = TEST_RC_ADDR;
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_READ_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_reply_notresponse, result);

  /* Address doesn't match */
  resp.addr = (TEST_RC_ADDR + 1) | RC_MESG_RESPONSE;
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_READ_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_reply_badaddress, result);

  /* Broadcast matches any address */
  rcp->rc_address = RC_MESG_BROADCAST;
  resp.addr = TEST_RC_ADDR | RC_MESG_RESPONSE;
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_READ_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_success, result);

  /* Command doesn't match */
  rcp->rc_address = TEST_RC_ADDR;
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_WRITE_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_reply_badcommand, result);

  /* Register doesn't match */
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_READ_DECIMAL, 0x1235, &returncode);
  CuAssertIntEquals(tc, rcr_reply_badregister, result);

  /* Read result is in error, return code will be set to int(resp.data) */
  returncode = 0;
  resp.addr = TEST_RC_ADDR | RC_MESG_RESPONSE | RC_MESG_ERROR;
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_READ_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_reply_error, result);
  CuAssertIntEquals(tc, 0x89AB, returncode);

  /* Read result no error, return code will be set 0 */
  returncode = 1234;
  resp.addr = TEST_RC_ADDR | RC_MESG_RESPONSE;
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_READ_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 0, returncode);

  /* Write result always sets return code to int(resp.data) */
  returncode = 0;
  resp.addr = TEST_RC_ADDR | RC_MESG_RESPONSE;
  resp.cmd = RC_MESG_WRITE_DECIMAL;
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_WRITE_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 0x89AB, returncode);

  /* Bad result code */
  resp.addr = TEST_RC_ADDR | RC_MESG_RESPONSE | RC_MESG_ERROR;
  resp.cmd = RC_MESG_READ_DECIMAL;
  rc_safe_strncpy(resp.data, "GH12", RC_SIZE_BUFFER);
  result = rc_cli_check_response_header(rcp, &resp, RC_MESG_READ_DECIMAL, 0x1234, &returncode);
  CuAssertIntEquals(tc, rcr_parsefail, result);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_message_fillout(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  struct s_rc_Message msg;
  const char data[] = "testdata";
  const uint8_t rc_escape_data[] = "test\x01\x00\xFF" "data";

  rcp = make_cli_parse(tc, true, false, 0);

  /* Check successful fillout */
  result = rc_cli_message_fillout(rcp, &msg, true, RC_MESG_READ_DECIMAL, 0x1234, (const uint8_t *) data, strlen(data));
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, rcp->rc_address | RC_MESG_REPLY, msg.addr);
  CuAssertIntEquals(tc, RC_MESG_READ_DECIMAL, msg.cmd);
  CuAssertIntEquals(tc, 0x1234, msg.reg);
  CuAssertStrEquals(tc, "", msg.chan);
  CuAssertStrEquals(tc, data, msg.data);

  /* No data */
  result = rc_cli_message_fillout(rcp, &msg, true, RC_MESG_READ_DECIMAL, 0x1234, NULL, 0);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "", msg.data);

  /* Rincmd escaped data */
  result = rc_cli_message_fillout(rcp, &msg, true, RC_MESG_READ_DECIMAL, 0x1234, rc_escape_data, sizeof(rc_escape_data) - 1);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "test\\01\\00\\FFdata", msg.data);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_message_fillout_channel(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  struct s_rc_Message msg;
  const char data[] = "testdata";

  /* With channel */
  rcp = make_cli_parse(tc, true, true, 0x5678);
  result = rc_cli_message_fillout(rcp, &msg, true, RC_MESG_READ_DECIMAL, 0x1234, (const uint8_t *) data, strlen(data));
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "5678", msg.chan);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_assemble_register_read(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  char buffer[TEST_BUFFER_LEN];

  rcp = make_cli_parse(tc, true, false, 0);

  /* Create a simple read request */
  result = rc_cli_assemble_register_read(rcp, RC_MESG_READ_DECIMAL, 0x1234, buffer, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "\x01" "2A161234:7D16\x04", buffer);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_assemble_register_read_with_channel(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  char buffer[TEST_BUFFER_LEN];

  rcp = make_cli_parse(tc, false, true, 0xABCDEF01);

  /* Create a simple read request */
  result = rc_cli_assemble_register_read(rcp, RC_MESG_READ_DECIMAL, 0x1234, buffer, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "2A161234ABCDEF01:;", buffer);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_parse_int_register_read(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  const char response_dec[] = "8A161234:5678\r\n";
  const char response_hex[] = "8A111234:ABCD\r\n";
  int64_t regvalue;

  rcp = make_cli_parse(tc, false, false, 0);

  /* Check a simple read reply */
  result = rc_cli_parse_int_register_read(rcp, RC_MESG_READ_DECIMAL, 0x1234, response_dec, &regvalue);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 5678, regvalue);

  /* Check a simple read reply */
  result = rc_cli_parse_int_register_read(rcp, RC_MESG_READ_HEXADEC, 0x1234, response_hex, &regvalue);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 0xABCD, regvalue);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_parse_buf_register_read(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  const char response[] = "8A161234:test\\00\\FF\r\n";
  uint8_t regvalue[TEST_BUFFER_LEN];
  size_t regvalue_len;

  rcp = make_cli_parse(tc, false, false, 0);

  /* Check a simple read reply */
  regvalue_len = TEST_BUFFER_LEN;
  result = rc_cli_parse_buf_register_read(rcp, RC_MESG_READ_DECIMAL, 0x1234, response, regvalue, &regvalue_len);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 6, regvalue_len);
  CuAssertIntEquals(tc, 0, memcmp("test\x00\xFF", regvalue, 6));

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_parse_str_register_read(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  const char response[] = "8A161234:test\\FF\\01\r\n";
  char regvalue[TEST_BUFFER_LEN];

  rcp = make_cli_parse(tc, false, false, 0);

  /* Check a simple read reply */
  result = rc_cli_parse_str_register_read(rcp, RC_MESG_READ_DECIMAL, 0x1234, response, regvalue, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, "test\xFF\x01", regvalue);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_assemble_int_register_write(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  const char *expected_msg;
  char actual_msg[TEST_BUFFER_LEN];

  rcp = make_cli_parse(tc, false, false, 0);

  /* Check a positive decimal integer register write generates the correct rincmd message */
  expected_msg = "2A171234:2843947;";
  result = rc_cli_assemble_int_register_write(rcp, RC_MESG_WRITE_DECIMAL, 0x1234, 2843947, actual_msg, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, expected_msg, actual_msg);

  /* Check a negative decimal integer register write generates the correct rincmd message */
  expected_msg = "2A1712AB:-58373;";
  result = rc_cli_assemble_int_register_write(rcp, RC_MESG_WRITE_DECIMAL, 0x12AB, -58373, actual_msg, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, expected_msg, actual_msg);

  /* Check a positive hexadecimal integer register write generates the correct rincmd message */
  expected_msg = "2A129ABC:1234ABCD;";
  result = rc_cli_assemble_int_register_write(rcp, RC_MESG_WRITE_HEXADEC, 0x9ABC, 0x1234ABCD, actual_msg, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, expected_msg, actual_msg);

  /* Check a negative hexadecimal integer register write generates the correct rincmd message */
  expected_msg = "2A12F123:FFFFFFF6;";
  result = rc_cli_assemble_int_register_write(rcp, RC_MESG_WRITE_HEXADEC, 0xF123, -10, actual_msg, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, expected_msg, actual_msg);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_assemble_buf_register_write(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  const char *expected_msg;
  char actual_msg[TEST_BUFFER_LEN];

  rcp = make_cli_parse(tc, false, false, 0);

  /* Check a buffer register write generates the correct rincmd message */
  expected_msg = "2A1298AB:test\\00\\FF;";
  result = rc_cli_assemble_buf_register_write(rcp, RC_MESG_WRITE_HEXADEC, 0x98AB, "test\x00\xFF", 6, actual_msg, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, expected_msg, actual_msg);

  destroy_cli_parse(tc, rcp);
}


static void test_rc_cli_assemble_str_register_write(CuTest *tc)
{
  struct s_rc_cli_parser *rcp;
  enum e_rc_cli_result result;
  const char *expected_msg;
  char actual_msg[TEST_BUFFER_LEN];

  rcp = make_cli_parse(tc, false, false, 0);

  /* Check a string register write generates the correct rincmd message */
  expected_msg = "2A1245FC:test\\01\\FF;";
  result = rc_cli_assemble_str_register_write(rcp, RC_MESG_WRITE_HEXADEC, 0x45FC, "test\x01\xFF", actual_msg, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertStrEquals(tc, expected_msg, actual_msg);

  destroy_cli_parse(tc, rcp);
}


CuSuite *CuGetSuite_rc_cli_parse(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rc_cli_assemble_command_crc);
  SUITE_ADD_TEST(suite, test_rc_cli_assemble_command_normal);
  SUITE_ADD_TEST(suite, test_rc_cli_assemble_command_channel);
  SUITE_ADD_TEST(suite, test_rc_cli_parse_response_crc);
  SUITE_ADD_TEST(suite, test_rc_cli_check_response_header);
  SUITE_ADD_TEST(suite, test_rc_cli_message_fillout);
  SUITE_ADD_TEST(suite, test_rc_cli_message_fillout_channel);
  SUITE_ADD_TEST(suite, test_rc_cli_assemble_register_read);
  SUITE_ADD_TEST(suite, test_rc_cli_assemble_register_read_with_channel);
  SUITE_ADD_TEST(suite, test_rc_cli_parse_int_register_read);
  SUITE_ADD_TEST(suite, test_rc_cli_parse_buf_register_read);
  SUITE_ADD_TEST(suite, test_rc_cli_parse_str_register_read);
  SUITE_ADD_TEST(suite, test_rc_cli_assemble_int_register_write);
  SUITE_ADD_TEST(suite, test_rc_cli_assemble_buf_register_write);
  SUITE_ADD_TEST(suite, test_rc_cli_assemble_str_register_write);

  return suite;
}
