/*
 * test_rc_escape.c
 *
 *  Created on: Sep 21, 2017
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <cutest/CuTest.h>
#include <rincmd/rc_escape.h>

#include "test_rc_escape.h"


/** Test escaping data */
static void test_rc_escape(CuTest *tc)
{
  const uint8_t buffer[] = "This is some test data. \\:;\x00\x1F\x7F\xFF";
                          /*This is some test data. \   :   ;   \x00\x1F\x7F\xFF */
  const char expected[]  = "This is some test data. \\\\\\3A\\3B\\00\\1F\\7F\\FF";
  char escaped[sizeof(buffer) * 3 + 1];
  enum e_rc_escape_result res;

  res = rc_escape_data(buffer, sizeof(buffer) - 1, escaped, sizeof(escaped));
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, expected, escaped);
}


/** Test escaping buffer sizes */
static void test_rc_escape_length(CuTest *tc)
{
  const uint8_t buffer[] = "Test \x02 data";
  const char expected[]  = "Test \\02 data";
  char escaped[strlen(expected) + 1];
  enum e_rc_escape_result res;

  /* Should just fit */
  res = rc_escape_data(buffer, sizeof(buffer) - 1, escaped, sizeof(escaped));
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, expected, escaped);

  /* Shouldn't fit */
  res = rc_escape_data(buffer, sizeof(buffer) - 1, escaped, sizeof(escaped) - 1);
  CuAssertIntEquals(tc, rer_buffer_size, res);
}


/** Test general use of unescaping data */
static void test_rc_unescape(CuTest *tc)
{
                            /*This is some test data. \   :   ;   \x00\x1F\x7F\xFF */
  const char escaped[]     = "This is some test data. \\\\\\3A\\3B\\00\\1F\\7F\\FF";
  const uint8_t expected[] = "This is some test data. \\:;\x00\x1F\x7F\xFF";
  uint8_t unescaped[sizeof(escaped)];
  enum e_rc_escape_result res;
  const size_t expected_size = sizeof(expected) - 1;
  size_t unescaped_size;

  unescaped_size = sizeof(unescaped);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertIntEquals(tc, 0, memcmp(expected, unescaped, expected_size));
  CuAssertIntEquals(tc, expected_size, unescaped_size);
}


/** Test some poorly escaped data */
static void test_rc_unescape_dodgy(CuTest *tc)
{
  const char escaped[]     = "Test \\not hex";
  const uint8_t expected[] = "Test \\not hex";
  uint8_t unescaped[sizeof(escaped)];
  enum e_rc_escape_result res;
  const size_t expected_size = sizeof(expected) - 1;
  size_t unescaped_size;

  unescaped_size = sizeof(unescaped);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertIntEquals(tc, 0, memcmp(expected, unescaped, expected_size));
  CuAssertIntEquals(tc, expected_size, unescaped_size);
}


/** Test unescaping buffer sizes */
static void test_rc_unescape_length(CuTest *tc)
{
                           /* 01234 5  67890 */
  const char escaped[]     = "Test \\02 data";
  const uint8_t expected[] = "Test \x02 data";
  uint8_t unescaped[sizeof(expected) - 1];
  enum e_rc_escape_result res;
  const size_t expected_size = sizeof(expected) - 1;
  size_t unescaped_size;

  /* This one should just fit */
  unescaped_size = sizeof(unescaped);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertIntEquals(tc, 0, memcmp(expected, unescaped, expected_size));
  CuAssertIntEquals(tc, expected_size, unescaped_size);

  /* This one should fail */
  unescaped_size = sizeof(unescaped) - 1;
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_buffer_size, res);
}


/** Test unescaping with outbuf = NULL, used to check the inbuf */
static void test_rc_unescape_null_outbuf(CuTest *tc)
{
                           /* 01234 5  67890 */
  const char escaped[]     = "Test \\02 data";
  const uint8_t expected[] = "Test \x02 data";
  uint8_t unescaped[sizeof(expected) - 1];
  enum e_rc_escape_result res;
  const size_t expected_size = sizeof(expected) - 1;
  size_t unescaped_size;

  /* This one should just fit */
  unescaped_size = sizeof(unescaped);
  res = rc_unescape_data(escaped, NULL, &unescaped_size, 0, UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertIntEquals(tc, expected_size, unescaped_size);

  /* This one should fail */
  unescaped_size = sizeof(unescaped) - 1;
  res = rc_unescape_data(escaped, NULL, &unescaped_size, 0, UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_buffer_size, res);
}


/** Test unescaping character limits */
static void test_rc_unescape_char_limits(CuTest *tc)
{
                           /* 01234 5   6  78901 */
  const char escaped[]     = "Test \\02\\F0 data";
  const uint8_t expected[] = "Test \x02\xF0 data";
  uint8_t unescaped[sizeof(expected) - 1];
  size_t unescaped_size = sizeof(unescaped);
  enum e_rc_escape_result res;

  /* Lower limit */
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, '\x01', UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_success, res);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, '\x02', UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_success, res);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, '\x03', UINT8_MAX, NULL);
  CuAssertIntEquals(tc, rer_data_low_range, res);

  /* Upper limit */
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, '\xF1', NULL);
  CuAssertIntEquals(tc, rer_success, res);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, '\xF0', NULL);
  CuAssertIntEquals(tc, rer_success, res);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, '\xEF', NULL);
  CuAssertIntEquals(tc, rer_data_high_range, res);

  /* Character set */
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, UINT8_MAX, (const uint8_t *) "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ \x02\xF0");
  CuAssertIntEquals(tc, rer_success, res);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, UINT8_MAX, (const uint8_t *) "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ \xF0");
  CuAssertIntEquals(tc, rer_data_low_range, res);
  res = rc_unescape_data(escaped, unescaped, &unescaped_size, 0, UINT8_MAX, (const uint8_t *) "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ \x02");
  CuAssertIntEquals(tc, rer_data_low_range, res);
}


CuSuite *CuGetSuite_rc_escape(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rc_escape);
  SUITE_ADD_TEST(suite, test_rc_escape_length);
  SUITE_ADD_TEST(suite, test_rc_unescape);
  SUITE_ADD_TEST(suite, test_rc_unescape_dodgy);
  SUITE_ADD_TEST(suite, test_rc_unescape_length);
  SUITE_ADD_TEST(suite, test_rc_unescape_null_outbuf);
  SUITE_ADD_TEST(suite, test_rc_unescape_char_limits);

  return suite;
}
