/*
 * test_rc_frame.c
 *
 *  Created on: Nov 21, 2017
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <cutest/CuTest.h>
#include <rincmd/rc_frame.h>

#include "test_rc_frame.h"

#define TEST_BUFFER_LEN 100


static void check_framing(CuTest *tc, struct s_rc_frame *rc_frame,
                          bool command_not_response, size_t expected_prefix_len,
                          const char *message, const char *expected_framed_message, bool expected_crc)
{
  char buffer[TEST_BUFFER_LEN];
  size_t endpos;
  enum e_rc_cli_result result;

  result = rc_preframe_message(rc_frame, command_not_response, buffer, TEST_BUFFER_LEN, &endpos);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, expected_prefix_len, endpos);

  snprintf(&buffer[endpos], TEST_BUFFER_LEN - endpos, "%s", message);

  result = rc_postframe_message(rc_frame, command_not_response, buffer, TEST_BUFFER_LEN);
  CuAssertIntEquals(tc, rcr_success, result);

  CuAssertStrEquals(tc, expected_framed_message, buffer);

  CuAssertTrue(tc, expected_crc == rc_frame->has_crc);
}


static void test_rc_crc_cmd_framing(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;

  result = rc_frame_init_crc(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  check_framing(tc, rc_frame, true, 1, "framed message", "\x01" "framed message\x04", true);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_crc_resp_framing(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;

  result = rc_frame_init_crc(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  check_framing(tc, rc_frame, false, 1, "framed message", "\x01" "framed message\x04", true);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_normal_cmd_framing(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;

  result = rc_frame_init_normal(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  check_framing(tc, rc_frame, true, 0, "framed message", "framed message;", false);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_normal_resp_framing(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;

  result = rc_frame_init_normal(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  check_framing(tc, rc_frame, false, 0, "framed message", "framed message\r\n", false);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_normal_cr_cmd_framing(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;

  result = rc_frame_init_normal_cr(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  check_framing(tc, rc_frame, true, 0, "framed message", "framed message\r", false);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_normal_cr_resp_framing(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;

  result = rc_frame_init_normal_cr(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  check_framing(tc, rc_frame, false, 0, "framed message", "framed message\r\n", false);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_normal_lf_cmd_framing(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;

  result = rc_frame_init_normal_lf(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  check_framing(tc, rc_frame, true, 0, "framed message", "framed message\n", false);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_normal_lf_resp_framing(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;

  result = rc_frame_init_normal_lf(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  check_framing(tc, rc_frame, false, 0, "framed message", "framed message\r\n", false);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_frame_response_check_crc(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;
  size_t startpos;
  size_t endpos;
  size_t msgend;

  result = rc_frame_init_crc(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  result = rc_frame_response_check(rc_frame, false, "\x01" "framed message\x04", &startpos, &endpos, &msgend);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 1, startpos);
  CuAssertIntEquals(tc, 15, endpos);
  CuAssertIntEquals(tc, 16, msgend);

  result = rc_frame_response_check(rc_frame, false, "framed message\x04", &startpos, &endpos, &msgend);
  CuAssertIntEquals(tc, rcr_framingfail, result);

  result = rc_frame_response_check(rc_frame, false, "\x01" "framed message", &startpos, &endpos, &msgend);
  CuAssertIntEquals(tc, rcr_framingfail, result);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


static void test_rc_frame_response_check_crc_predata(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;
  size_t startpos;
  size_t endpos;
  size_t msgend;

  result = rc_frame_init_crc(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  result = rc_frame_response_check(rc_frame, false, "data before message\x01" "framed message\x04", &startpos, &endpos, &msgend);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 20, startpos);
  CuAssertIntEquals(tc, 20 + 14, endpos);
  CuAssertIntEquals(tc, 20 + 14 + 1, msgend);
}


static void test_rc_frame_response_check_normal(CuTest *tc)
{
  struct s_rc_frame *rc_frame;
  enum e_rc_cli_result result;
  size_t startpos;
  size_t endpos;
  size_t msgend;

  result = rc_frame_init_normal(&rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);

  result = rc_frame_response_check(rc_frame, false, "framed message\r\n", &startpos, &endpos, &msgend);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 0, startpos);
  CuAssertIntEquals(tc, 14, endpos);
  CuAssertIntEquals(tc, 16, msgend);

  result = rc_frame_response_check(rc_frame, false, "framed message\r", &startpos, &endpos, &msgend);
  CuAssertIntEquals(tc, rcr_framingfail, result);

  result = rc_frame_destroy(rc_frame);
  CuAssertIntEquals(tc, rcr_success, result);
}


CuSuite *CuGetSuite_rc_frame(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rc_crc_cmd_framing);
  SUITE_ADD_TEST(suite, test_rc_crc_resp_framing);
  SUITE_ADD_TEST(suite, test_rc_normal_cmd_framing);
  SUITE_ADD_TEST(suite, test_rc_normal_resp_framing);
  SUITE_ADD_TEST(suite, test_rc_normal_cr_cmd_framing);
  SUITE_ADD_TEST(suite, test_rc_normal_cr_resp_framing);
  SUITE_ADD_TEST(suite, test_rc_normal_lf_cmd_framing);
  SUITE_ADD_TEST(suite, test_rc_normal_lf_resp_framing);
  SUITE_ADD_TEST(suite, test_rc_frame_response_check_crc);
  SUITE_ADD_TEST(suite, test_rc_frame_response_check_crc_predata);
  SUITE_ADD_TEST(suite, test_rc_frame_response_check_normal);

  return suite;
}
