/*
 * test_rc_timespec.c
 *
 *  Created on: Nov 29, 2017
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <cutest/CuTest.h>
#include <rincmd/rc_timespec.h>

#include "test_rc_timespec.h"


void CuAssertTimeSpecEquals_LineMsg(CuTest* tc, const char* file, int line, const char* message,
                                    struct timespec expected, struct timespec actual)
{
  char buf[STRING_MAX];
  if (expected.tv_sec == actual.tv_sec && expected.tv_nsec == actual.tv_nsec) return;

  snprintf(buf, STRING_MAX, "expected <%d.%09d> but was <%d.%09d>",
           (int) expected.tv_sec, (int) expected.tv_nsec,
           (int) actual.tv_sec, (int) actual.tv_nsec);
  CuFail_Line(tc, file, line, message, buf);
}

static void test_rc_timespec_SET_TIMESPEC(CuTest *tc)
{
  struct timespec ac;
  const struct timespec ex1 = { .tv_sec=347727, .tv_nsec=481563848 };
  const struct timespec ex2 = { .tv_sec=347727, .tv_nsec=481563847 };
  const struct timespec ex3 = { .tv_sec=347727, .tv_nsec=000000001 };
  const struct timespec ex4 = { .tv_sec=-47727, .tv_nsec=123456789 };
  const struct timespec ex5 = { .tv_sec=0,      .tv_nsec=123456789 };
  const struct timespec ex6 = { .tv_sec=0,      .tv_nsec=-123456789 };
  const struct timespec ex7 = { .tv_sec=0,      .tv_nsec=0 };
  const struct timespec ex8 = { .tv_sec=-47727, .tv_nsec=0 };

  SET_TIMESPEC(ac, 347727.481563848);
  CuAssertTimeSpecEquals(tc, ex1, ac);

  SET_TIMESPEC(ac, 347727.481563847);
  CuAssertTimeSpecEquals(tc, ex2, ac);

  SET_TIMESPEC(ac, 347727.000000001);
  CuAssertTimeSpecEquals(tc, ex3, ac);

  SET_TIMESPEC(ac, -47727.123456789);
  CuAssertTimeSpecEquals(tc, ex4, ac);

  SET_TIMESPEC(ac, 0.123456789);
  CuAssertTimeSpecEquals(tc, ex5, ac);

  SET_TIMESPEC(ac, -0.123456789);
  CuAssertTimeSpecEquals(tc, ex6, ac);

  SET_TIMESPEC(ac, 0.0);
  CuAssertTimeSpecEquals(tc, ex7, ac);

  SET_TIMESPEC(ac, -47727);
  CuAssertTimeSpecEquals(tc, ex8, ac);
}


static void test_rc_timespec_DECLARE_TIMESPEC(CuTest *tc)
{
  DECLARE_TIMESPEC(const, ac1, 347727.481563848);
  const struct timespec ex1 = { .tv_sec=347727, .tv_nsec=481563848 };

  DECLARE_TIMESPEC(tsnoprefix, ac2, 347727.481563848);
  const struct timespec ex2 = { .tv_sec=347727, .tv_nsec=481563848 };

  CuAssertTimeSpecEquals(tc, ex1, ac1);
  CuAssertTimeSpecEquals(tc, ex2, ac2);
}


static void test_rc_timespec_add_sec(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec ac;

  /* +ve x.sec & y.sec */
  x.tv_sec = 1;
  x.tv_nsec = 0;
  y.tv_sec = 2;
  y.tv_nsec = 0;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, 3, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* -ve x.sec & y.sec */
  x.tv_sec = -3;
  x.tv_nsec = 0;
  y.tv_sec = -2;
  y.tv_nsec = 0;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, -5, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* +ve x.sec, -ve y.sec */
  x.tv_sec = 3;
  x.tv_nsec = 0;
  y.tv_sec = -4;
  y.tv_nsec = 0;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, -1, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* -ve x.sec, +ve y.sec */
  x.tv_sec = -2;
  x.tv_nsec = 0;
  y.tv_sec = 4;
  y.tv_nsec = 0;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, 2, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);
}


static void test_rc_timespec_add_nsec(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec ac;

  /* +ve x.nsec & y.nsec 0.000000001 + 0.000000002 = 0.000000003 */
  x.tv_sec = 0;
  x.tv_nsec = 1;
  y.tv_sec = 0;
  y.tv_nsec = 2;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, 3, ac.tv_nsec);

  /* -ve x.sec & y.sec -0.000000001 + -0.000000002 = -0.000000003 */
  x.tv_sec = 0;
  x.tv_nsec = -1;
  y.tv_sec = 0;
  y.tv_nsec = -2;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, -3, ac.tv_nsec);

  /* +ve overflow 0.999999999 + 0.000000003 = 1.000000002 */
  x.tv_sec = 0;
  x.tv_nsec = RC_NANO_SEC_PER_SEC - 1;
  y.tv_sec = 0;
  y.tv_nsec = 3;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, 1, ac.tv_sec);
  CuAssertIntEquals(tc, 2, ac.tv_nsec);

  /* -ve overflow, tv_nsec gets normalised: -0.999999999 + -0.000000003 = -1.000000002 */
  x.tv_sec = 0;
  x.tv_nsec = -(RC_NANO_SEC_PER_SEC - 1);
  y.tv_sec = 0;
  y.tv_nsec = -3;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, -1, ac.tv_sec);
  CuAssertIntEquals(tc, 2, ac.tv_nsec);

  /* -ve tv_sec and tv_nsec with overflow, tv_nsec gets normalised: -1.-999999999 + -1.-000000003 = -3.000000002 */
  x.tv_sec = -1;
  x.tv_nsec = -(RC_NANO_SEC_PER_SEC - 1);
  y.tv_sec = -1;
  y.tv_nsec = -3;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, -3, ac.tv_sec);
  CuAssertIntEquals(tc, 2, ac.tv_nsec);

  /* -ve tv_sec with +ve tv_nsec with overflow, tv_nsec gets normalised -1.999999999 + -1.000000003 = -3.000000002 */
  x.tv_sec = -1;
  x.tv_nsec = RC_NANO_SEC_PER_SEC - 1;
  y.tv_sec = -1;
  y.tv_nsec = 3;
  rc_timespec_add(&ac, &x, &y);
  CuAssertIntEquals(tc, -3, ac.tv_sec);
  CuAssertIntEquals(tc, 2, ac.tv_nsec);
}


static void test_rc_timespec_add_bug(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec ex;
  struct timespec ac;

  /* 1.040996274 + -0.934163948 = 0.106832326 */
  SET_TIMESPEC(x,  1.040996274);
  SET_TIMESPEC(y, -0.934163948);
  SET_TIMESPEC(ex, 0.106832326);
  rc_timespec_add(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);
}


static void test_rc_timespec_subtract_carry(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec ex;
  struct timespec ac;

  /* 0.000000002 - -0.999999999 = 1.000000001 */
  SET_TIMESPEC(x,  0.000000002);
  SET_TIMESPEC(y, -0.999999999);
  SET_TIMESPEC(ex, 1.000000001);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* 4.000000002 - -3.999999999 = 8.000000001 */
  SET_TIMESPEC(x,  4.000000002);
  SET_TIMESPEC(y, -3.999999999);
  SET_TIMESPEC(ex, 8.000000001);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* 3.000000002 - 3.999999999 = 0.-999999997 */
  SET_TIMESPEC(x,   3.000000002);
  SET_TIMESPEC(y,   3.999999999);
  SET_TIMESPEC(ex, -0.999999997);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);
}


static void test_rc_timespec_subtract_sec(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec ac;

  /* +ve x.sec < y.sec */
  x.tv_sec = 1;
  x.tv_nsec = 0;
  y.tv_sec = 2;
  y.tv_nsec = 0;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, -1, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* -ve x.sec < y.sec */
  x.tv_sec = -3;
  x.tv_nsec = 0;
  y.tv_sec = -2;
  y.tv_nsec = 0;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, -1, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* +ve x.sec > y.sec */
  x.tv_sec = 2;
  x.tv_nsec = 0;
  y.tv_sec = 1;
  y.tv_nsec = 0;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 1, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* -ve x.sec > y.sec */
  x.tv_sec = -2;
  x.tv_nsec = 0;
  y.tv_sec = -3;
  y.tv_nsec = 0;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 1, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* +ve x.sec = y.sec */
  x.tv_sec = 2;
  x.tv_nsec = 0;
  y.tv_sec = 2;
  y.tv_nsec = 0;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* -ve x.sec = y.sec */
  x.tv_sec = -2;
  x.tv_nsec = 0;
  y.tv_sec = -2;
  y.tv_nsec = 0;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);
}


static void test_rc_timespec_subtract_nsec(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec ac;

  /* +ve x.nsec < y.nsec: 0.000000001 - 0.000000002 = 0.-000000001 */
  x.tv_sec = 0;
  x.tv_nsec = 1;
  y.tv_sec = 0;
  y.tv_nsec = 2;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, -1, ac.tv_nsec);

  /* -ve x.nsec < y.nsec: 0.-000000003 - 0.-000000002 = 0.-000000001 */
  x.tv_sec = 0;
  x.tv_nsec = -3;
  y.tv_sec = 0;
  y.tv_nsec = -2;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, -1, ac.tv_nsec);

  /* +ve x.nsec > y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = 2;
  y.tv_sec = 0;
  y.tv_nsec = 1;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, 1, ac.tv_nsec);

  /* -ve x.sec > y.sec */
  x.tv_sec = 0;
  x.tv_nsec = -2;
  y.tv_sec = 0;
  y.tv_nsec = -3;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, 1, ac.tv_nsec);

  /* +ve x.sec = y.sec */
  x.tv_sec = 0;
  x.tv_nsec = 2;
  y.tv_sec = 0;
  y.tv_nsec = 2;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);

  /* -ve x.sec = y.sec */
  x.tv_sec = 0;
  x.tv_nsec = -2;
  y.tv_sec = 0;
  y.tv_nsec = -2;
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertIntEquals(tc, 0, ac.tv_sec);
  CuAssertIntEquals(tc, 0, ac.tv_nsec);
}


static void test_rc_timespec_subtract_nsec_conclusive(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec ac;
  struct timespec ex;

  /* +ve x.nsec > +ve y.nsec: 0.000000005 - 0.000000002 = 0.000000003 */
  SET_TIMESPEC(x,  0.000000005);
  SET_TIMESPEC(y,  0.000000002);
  SET_TIMESPEC(ex, 0.000000003);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* +ve x.nsec < +ve y.nsec: 0.000000002 - 0.000000005 = 0.-000000003 */
  SET_TIMESPEC(x,   0.000000002);
  SET_TIMESPEC(y,   0.000000005);
  SET_TIMESPEC(ex, -0.000000003);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* +ve x.nsec < +ve y.nsec: 1.000000002 - 2.000000005 = -1.000000003 */
  SET_TIMESPEC(x,   1.000000002);
  SET_TIMESPEC(y,   2.000000005);
  SET_TIMESPEC(ex, -1.000000003);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* -ve x.nsec > -ve y.nsec: -0.000000002 - -0.000000005 = 0.000000003 */
  SET_TIMESPEC(x, -0.000000002);
  SET_TIMESPEC(y, -0.000000005);
  SET_TIMESPEC(ex, 0.000000003);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* -ve x.nsec < -ve y.nsec: -0.000000005 - -0.000000002 = -0.000000003 */
  SET_TIMESPEC(x,  -0.000000005);
  SET_TIMESPEC(y,  -0.000000002);
  SET_TIMESPEC(ex, -0.000000003);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* +ve x.nsec > -ve y.nsec: 0.000000003 - -0.000000002 = 0.000000005 */
  SET_TIMESPEC(x,  0.000000003);
  SET_TIMESPEC(y, -0.000000002);
  SET_TIMESPEC(ex, 0.000000005);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* -ve x.nsec < +ve y.nsec: -0.000000002 - 0.000000003 = -0.000000005 */
  SET_TIMESPEC(x,  -0.000000002);
  SET_TIMESPEC(y,   0.000000003);
  SET_TIMESPEC(ex, -0.000000005);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* +ve x.nsec = +ve y.nsec: 0.000000003 - 0.000000003 = 0.000000000 */
  SET_TIMESPEC(x,  0.000000003);
  SET_TIMESPEC(y,  0.000000003);
  SET_TIMESPEC(ex, 0.000000000);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  /* -ve x.nsec = -ve y.nsec: -0.000000003 - -0.000000003 = 0.000000000 */
  SET_TIMESPEC(x, -0.000000003);
  SET_TIMESPEC(y, -0.000000003);
  SET_TIMESPEC(ex, 0.000000000);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);
}


static void test_rc_timespec_subtract_bug(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec ex;
  struct timespec ac;

  SET_TIMESPEC(x,   0.481563848);
  SET_TIMESPEC(y,   0.481894426);
  SET_TIMESPEC(ex, -0.000330578);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  SET_TIMESPEC(x,  1.481563848);
  SET_TIMESPEC(y,  0.481894426);
  SET_TIMESPEC(ex, 0.999669422);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  SET_TIMESPEC(x, 347727.481563848);
  SET_TIMESPEC(y, 347722.481894426);
  SET_TIMESPEC(ex,     4.999669422);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  SET_TIMESPEC(x,  -0.779039256);
  SET_TIMESPEC(y,   0.653214606);
  SET_TIMESPEC(ex, -1.432253862);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);

  SET_TIMESPEC(x, -0.935934961);
  SET_TIMESPEC(y, -1.030742653);
  SET_TIMESPEC(ex, 0.094807692);
  rc_timespec_subtract(&ac, &x, &y);
  CuAssertTimeSpecEquals(tc, ex, ac);
}


static void test_rc_timespec_subtract_random(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  struct timespec sub;
  struct timespec add;
  int i;
  long long n;

  for (i = 0; i < 1000000; i++)
  {
    n = random();
    n <<= 31;
    n |= random();
    n -= (long long) 2 << 63;
    x.tv_nsec = n % RC_NANO_SEC_PER_SEC;
    x.tv_sec = n / RC_NANO_SEC_PER_SEC;
    if (x.tv_sec < 0 && x.tv_nsec < 0)
      x.tv_nsec = -x.tv_nsec;

    n = random();
    n <<= 31;
    n |= random();
    n -= (long long) 2 << 63;
    y.tv_nsec = n % RC_NANO_SEC_PER_SEC;
    y.tv_sec = n / RC_NANO_SEC_PER_SEC;
    if (y.tv_sec < 0 && y.tv_nsec < 0)
      y.tv_nsec = -y.tv_nsec;

    rc_timespec_subtract(&sub, &x, &y);

#if 0
    printf("%5d.%09d - %5d.%09d = %5d.%09d\n",
           (int) x.tv_sec, (int) x.tv_nsec,
           (int) y.tv_sec, (int) y.tv_nsec,
           (int) sub.tv_sec, (int) sub.tv_nsec);
#endif

    if (0 == y.tv_sec)
    {
      y.tv_nsec = -y.tv_nsec;
    }
    else
    {
      y.tv_sec = -y.tv_sec;
      if (y.tv_sec < 0 && y.tv_nsec < 0)
        y.tv_nsec = -y.tv_nsec;
    }
    rc_timespec_add(&add, &x, &y);

#if 0
    printf("%5d.%09d + %5d.%09d = %5d.%09d\n",
           (int) x.tv_sec, (int) x.tv_nsec,
           (int) y.tv_sec, (int) y.tv_nsec,
           (int) add.tv_sec, (int) add.tv_nsec);
#endif

    CuAssertTimeSpecEquals(tc, sub, add);
  }
}


static void test_rc_timespec_compare_sec(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  int res;

  /* +ve x.sec < y.sec */
  x.tv_sec = 1;
  x.tv_nsec = 0;
  y.tv_sec = 2;
  y.tv_nsec = 0;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, -1, res);

  /* -ve x.sec < y.sec */
  x.tv_sec = -3;
  x.tv_nsec = 0;
  y.tv_sec = -2;
  y.tv_nsec = 0;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, -1, res);

  /* +ve x.sec > y.sec */
  x.tv_sec = 2;
  x.tv_nsec = 0;
  y.tv_sec = 1;
  y.tv_nsec = 0;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 1, res);

  /* -ve x.sec > y.sec */
  x.tv_sec = -2;
  x.tv_nsec = 0;
  y.tv_sec = -3;
  y.tv_nsec = 0;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 1, res);

  /* +ve x.sec = y.sec */
  x.tv_sec = 2;
  x.tv_nsec = 0;
  y.tv_sec = 2;
  y.tv_nsec = 0;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 0, res);

  /* -ve x.sec = y.sec */
  x.tv_sec = -2;
  x.tv_nsec = 0;
  y.tv_sec = -2;
  y.tv_nsec = 0;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 0, res);
}


static void test_rc_timespec_compare_nsec(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  int res;

  /* +ve x.nsec < y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = 1;
  y.tv_sec = 0;
  y.tv_nsec = 2;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, -1, res);

  /* -ve x.nsec < y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = -3;
  y.tv_sec = 0;
  y.tv_nsec = -2;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, -1, res);

  /* +ve x.nsec > y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = 2;
  y.tv_sec = 0;
  y.tv_nsec = 1;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 1, res);

  /* -ve x.nsec > y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = -2;
  y.tv_sec = 0;
  y.tv_nsec = -3;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 1, res);

  /* +ve x.nsec = y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = 2;
  y.tv_sec = 0;
  y.tv_nsec = 2;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 0, res);

  /* -ve x.nsec = y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = -2;
  y.tv_sec = 0;
  y.tv_nsec = -2;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 0, res);
}


static void test_rc_timespec_compare_both(CuTest *tc)
{
  struct timespec x;
  struct timespec y;
  int res;

  /* +ve x.nsec < y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = 1;
  y.tv_sec = 0;
  y.tv_nsec = 2;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, -1, res);

  /* -ve x.nsec < y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = -3;
  y.tv_sec = 0;
  y.tv_nsec = -2;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, -1, res);

  /* +ve x.nsec > y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = 2;
  y.tv_sec = 0;
  y.tv_nsec = 1;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 1, res);

  /* -ve x.nsec > y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = -2;
  y.tv_sec = 0;
  y.tv_nsec = -3;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 1, res);

  /* +ve x.nsec = y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = 2;
  y.tv_sec = 0;
  y.tv_nsec = 2;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 0, res);

  /* -ve x.nsec = y.nsec */
  x.tv_sec = 0;
  x.tv_nsec = -2;
  y.tv_sec = 0;
  y.tv_nsec = -2;
  res = rc_timespec_compare(&x, &y);
  CuAssertIntEquals(tc, 0, res);
}


CuSuite *CuGetSuite_rc_timespec(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rc_timespec_SET_TIMESPEC);
  SUITE_ADD_TEST(suite, test_rc_timespec_DECLARE_TIMESPEC);
  SUITE_ADD_TEST(suite, test_rc_timespec_add_sec);
  SUITE_ADD_TEST(suite, test_rc_timespec_add_nsec);
  SUITE_ADD_TEST(suite, test_rc_timespec_add_bug);
  SUITE_ADD_TEST(suite, test_rc_timespec_subtract_carry);
  SUITE_ADD_TEST(suite, test_rc_timespec_subtract_sec);
  SUITE_ADD_TEST(suite, test_rc_timespec_subtract_nsec);
  SUITE_ADD_TEST(suite, test_rc_timespec_subtract_nsec_conclusive);
  SUITE_ADD_TEST(suite, test_rc_timespec_subtract_bug);
  SUITE_ADD_TEST(suite, test_rc_timespec_subtract_random);
  SUITE_ADD_TEST(suite, test_rc_timespec_compare_sec);
  SUITE_ADD_TEST(suite, test_rc_timespec_compare_nsec);
  SUITE_ADD_TEST(suite, test_rc_timespec_compare_both);

  return suite;
}
