# rindb - Rinstrum database library

## Overview

The Rinstrum database library takes care of saving and loading settings from a database suitable for trade approved applications.

Features:

 * Read and write a database to XML, binary file, binary buffer, u-boot environment variables
 * Exensable architecture to allow writing to new formats
 * Permit addition and removal of settings as the user application evolves
 * Note database setting changes, and write the differences into a change log (for trade applications).
 * CRC32 checksum verification of settings (for trade applications).


## Getting started

Declare the variables to be stored in the database (a struct is not strictly necessary):

```c

#define DB_EXAMPLE_VERSION1 1

#define NUMBER_ARRAY_LEN 2
#define STRING_LEN 10

struct s_db_example
{
  uint32_t number1;
  uint8_t number2;
  int32_t number_array[NUMBER_ARRAY_LEN];
  char string[STRING_LEN];
};

struct s_db_example db_example;
```

Declare the elements in the database:

```c

/* Changelog masks */
#define DB_EXAMPLE_CLMASK1 0x00000001
#define DB_EXAMPLE_CLMASK2 0x00000002

struct s_db_element example_elements[] =
{
  /*                                 Database name, Menu name,      Data,                    Num elems,        Minimum, Maximum, Default, Min Version,           Max Version,           Changelog Mask */
  DB_DECLARE_UINT32_ELEM(            "number1",     "Number 1",     &db_example.number1,                       0,       100,     20,      DB_NO_MINIMUM_VERSION, DB_NO_MAXIMUM_VERSION, DB_EXAMPLE_CLMASK1),
  DB_DECLARE_UINT8_ELEM(             "number2",     "Number 2",     &db_example.number2,                       10,      40,      11,      DB_NO_MINIMUM_VERSION, DB_NO_MAXIMUM_VERSION, DB_EXAMPLE_CLMASK1),
  DB_DECLARE_INT32_ARRAY_ELEM_SINGLE("numberarray", "Number Array", db_example.number_array, NUMBER_ARRAY_LEN, 0,       100,     0,       DB_NO_MINIMUM_VERSION, DB_NO_MAXIMUM_VERSION, DB_EXAMPLE_CLMASK1),
  DB_DECLARE_STRING_ELEM(            "string",      "String"        db_example.string,       STRING_LEN,                         "str",   DB_NO_MINIMUM_VERSION, DB_NO_MAXIMUM_VERSION, DB_EXAMPLE_CLMASK1),
};
```

Declare the database structure:

```c
struct s_db_database example_database =
{
  .db_database_version = DB_EXAMPLE_VERSION1,
  .db_database_chksum_seed = 0x1234,
  .db_elements = example_elements,
  .db_num_elements = DB_NUM_ELEMENTS(example_elements),
  .db_name = "ExampleDatabase"
};
```

Then in code initialise the database, and a database interpreter to read/write it (in this case XML):

```c
  struct s_db_database_interp ddi_xml;
  enum eDBError res;

  /* Initialise the database (creates previous data for some elements) */
  res = db_database_init(&example_database);
  /* Handle errors here */

  /* Create the XML database interpreter */
  res = db_create_database_interp_xml(&ddi_xml, "example_database.xml");
  /* Handle errors here */
```

At this point the database can be read and written:

```c
  uint32_t version;
  enum eDBError res;

  /* Read database, on success version will be set to the version of database that was read */
  res = db_database_read(&example_database, &ddi_xml, &version, false);
  /* Handle errors here */

  /* Make changes to database variables here */
  db_example.number2 = 100;

  /* Write database at the current database version */
  res = db_database_write(&example_database, &ddi_xml, example_database.db_database_version);
  /* Handle errors here */
```

Cleaning up:

```c
  /* Destroy the database */
  res = db_database_destroy(&example_database);
  /* Handle errors here */

  /* Destroy the XML database interpreter */
  res = db_destroy_database_interp_xml(&ddi_xml);
  /* Handle errors here */
```


## Obsoleting a setting in the database

Settings can be removed from the database if they are no longer required.

Declare a new database version:

```c
#define DB_EXAMPLE_VERSION1 1
#define DB_EXAMPLE_VERSION2 2
```

Set the maximum version of the element you wish to remove (in this case `"number2"`) to the last version in which it occurs (in this case `DB_EXAMPLE_VERSION1`):
If you still require the setting to upgrade from, then leave the element's data pointer pointing to the variable (in this case `&db_example.number2`).

```c
struct s_db_element example_elements[] =
{
  /*                                 Database name, Menu name,      Data,                    Num elems,        Minimum, Maximum, Default, Min Version,           Max Version,           Changelog Mask */
  ...
  DB_DECLARE_UINT8_ELEM(             "number2",     "Number 2",     &db_example.number2,                       10,      40,      11,      DB_NO_MINIMUM_VERSION, DB_EXAMPLE_VERSION1,   DB_EXAMPLE_CLMASK1),
  ...
};
```

Update the database version in the database structure:

```c
struct s_db_database example_database =
{
  .db_database_version = DB_EXAMPLE_VERSION2,
  ...
};
```

Now when the database is written at version 2, "number2" will no longer be contained in it.
However, databases written at older versions (in this case only version 1) can still be read and will pass checksum validation.

If you no longer need to use the setting, the variable can be removed, and the element's data pointer can be set to NULL.
The database library will allocate it internally for checksum verification.
This allows the application code to be kept cleaner.

```c
struct s_db_element example_elements[] =
{
  /*                                 Database name, Menu name,      Data,                    Num elems,        Minimum, Maximum, Default, Min Version,           Max Version,           Changelog Mask */
  ...
  DB_DECLARE_UINT8_ELEM(             "number2",     "Number 2",     NULL,                                      10,      40,      11,      DB_NO_MINIMUM_VERSION, DB_EXAMPLE_VERSION1,   DB_EXAMPLE_CLMASK1),
  ...
};
```


## Introducing new settings into the database

New settings can be added to the database.

Declare the new database version, add the new setting to the database structure:

```c

#define DB_EXAMPLE_VERSION1 1
#define DB_EXAMPLE_VERSION2 2
#define DB_EXAMPLE_VERSION3 3

...

struct s_db_example
{
  ...
  uint32_t new_number;
};
```

Add the new element to the database elements, setting it's minimum version to the new database version:

```c
struct s_db_element example_elements[] =
{
  /*                                 Database name, Menu name,      Data,                    Num elems,        Minimum, Maximum, Default, Min Version,           Max Version,           Changelog Mask */
  ...
  DB_DECLARE_UINT32_ELEM(            "newnumber",   "New Number",   &db_example.new_number,                    0,       200,     5,       DB_EXAMPLE_VERSION3,   DB_NO_MAXIMUM_VERSION, DB_EXAMPLE_CLMASK1),
};
```

Update the database version in the database structure:

```c
struct s_db_database example_database =
{
  .db_database_version = DB_EXAMPLE_VERSION3,
  ...
};
```

Now when the database is written at version 3, "newnumber" will be stored into the database.
However, if an older version database (1 or 2) is read, it will pass checksum validation, and db_example.new_number will be set to its default value (5).
If a version 3 or above database is read, "newnumber" will be read from that database.


## Shrinking or growing an array size in a database

It is possible to shrink or grow an array in a database.
This is done by duplicating the array definition,
and setting the maximum version on the old definition,
and setting the minimum version on the new definition.

```c
static struct s_db_element array_size_change_v2_elements[] =
{
  /*                                      DB name,       Menu name,     Data,                                 Num elems,               Minimum,                    Maximum,                    Default,                    Min Version,             Max Version,             Changelog Mask */

  DB_DECLARE_UINT32_ARRAY_ELEM_INDIVIDUAL("GrowArray",   "GrowArray",   NULL,                                 DB_V1_GROW_ARRAY_SIZE,   db_grow_array_v1_minimum,   db_grow_array_v1_maximum,   db_grow_array_v1_default,   DB_NO_MINIMUM_VERSION,   DB_ARRAY_SIZE_CHANGE_V1, 0x0),
  DB_DECLARE_UINT32_ARRAY_ELEM_INDIVIDUAL("GrowArray",   "GrowArray",   db_array_size_change_v2.grow_array,   DB_V2_GROW_ARRAY_SIZE,   db_grow_array_v2_minimum,   db_grow_array_v2_maximum,   db_grow_array_v2_default,   DB_ARRAY_SIZE_CHANGE_V2, DB_NO_MAXIMUM_VERSION,   0x0),

  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL( "ShrinkArray", "ShrinkArray", NULL,                                 DB_V1_SHRINK_ARRAY_SIZE, db_shrink_array_v1_minimum, db_shrink_array_v1_maximum, db_shrink_array_v1_default, DB_NO_MINIMUM_VERSION,   DB_ARRAY_SIZE_CHANGE_V1, 0x0),
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL( "ShrinkArray", "ShrinkArray", db_array_size_change_v2.shrink_array, DB_V2_SHRINK_ARRAY_SIZE, db_shrink_array_v2_minimum, db_shrink_array_v2_maximum, db_shrink_array_v2_default, DB_ARRAY_SIZE_CHANGE_V2, DB_NO_MAXIMUM_VERSION,   0x0),
};
```
