/*********************************************************************\

MODULE NAME:    b64.c

AUTHOR:         Bob Trower 08/04/01

PROJECT:        Crypt Data Packaging

COPYRIGHT:      Copyright (c) Trantor Standard Systems Inc., 2001

NOTE:           This source code may be used as you wish, subject to
                the MIT license.  See the LICENCE section below.

DESCRIPTION:
                This little utility implements the Base64
                Content-Transfer-Encoding standard described in
                RFC1113 (http://www.faqs.org/rfcs/rfc1113.html).

                This is the coding scheme used by MIME to allow
                binary data to be transferred by SMTP mail.

                Groups of 3 bytes from a binary stream are coded as
                groups of 4 bytes in a text stream.

                The input stream is 'padded' with zeros to create
                an input that is an even multiple of 3.

                A special character ('=') is used to denote padding so
                that the stream can be decoded back to its exact size.

                Encoded output is formatted in lines which should
                be a maximum of 72 characters to conform to the
                specification.  This program defaults to 72 characters,
                but will allow more or less through the use of a
                switch.  The program enforces a minimum line size
                of 4 characters.

                Example encoding:

                The stream 'ABCD' is 32 bits long.  It is mapped as
                follows:

                ABCD

                 A (65)     B (66)     C (67)     D (68)   (None) (None)
                01000001   01000010   01000011   01000100

                16 (Q)  20 (U)  9 (J)   3 (D)    17 (R) 0 (A)  NA (=) NA (=)
                010000  010100  001001  000011   010001 000000 000000 000000


                QUJDRA==

                Decoding is the process in reverse.  A 'decode' lookup
                table has been created to avoid string scans.

DESIGN GOALS:	Specifically:
		Code is a stand-alone utility to perform base64 
		encoding/decoding. It should be genuinely useful 
		when the need arises and it meets a need that is 
		likely to occur for some users.  
		Code acts as sample code to show the author's 
		design and coding style.  

		Generally: 
		This program is designed to survive:
		Everything you need is in a single source file.
		It compiles cleanly using a vanilla ANSI C compiler.
		It does its job correctly with a minimum of fuss.  
		The code is not overly clever, not overly simplistic 
		and not overly verbose. 
		Access is 'cut and paste' from a web page.  
		Terms of use are reasonable.  

VALIDATION:     Non-trivial code is never without errors.  This
                file likely has some problems, since it has only
                been tested by the author.  It is expected with most
                source code that there is a period of 'burn-in' when
                problems are identified and corrected.  That being
                said, it is possible to have 'reasonably correct'
                code by following a regime of unit test that covers
                the most likely cases and regression testing prior
                to release.  This has been done with this code and
                it has a good probability of performing as expected.

                Unit Test Cases:

                case 0:empty file:
                    CASE0.DAT  ->  ->
                    (Zero length target file created
                    on both encode and decode.)

                case 1:One input character:
                    CASE1.DAT A -> QQ== -> A

                case 2:Two input characters:
                    CASE2.DAT AB -> QUJD -> AB

                case 3:Three input characters:
                    CASE3.DAT ABC -> QUJD -> ABC

                case 4:Four input characters:
                    case4.dat ABCD -> QUJDRA== -> ABCD

                case 5:All chars from 0 to ff, linesize set to 50:

                    AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIj
                    JCUmJygpKissLS4vMDEyMzQ1Njc4OTo7PD0+P0BBQkNERUZH
                    SElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWpr
                    bG1ub3BxcnN0dXZ3eHl6e3x9fn+AgYKDhIWGh4iJiouMjY6P
                    kJGSk5SVlpeYmZqbnJ2en6ChoqOkpaanqKmqq6ytrq+wsbKz
                    tLW2t7i5uru8vb6/wMHCw8TFxsfIycrLzM3Oz9DR0tPU1dbX
                    2Nna29zd3t/g4eLj5OXm5+jp6uvs7e7v8PHy8/T19vf4+fr7
                    /P3+/w==

                case 6:Mime Block from e-mail:
                    (Data same as test case 5)

                case 7: Large files:
                    Tested 28 MB file in/out.

                case 8: Random Binary Integrity:
                    This binary program (b64.exe) was encoded to base64,
                    back to binary and then executed.

                case 9 Stress:
                    All files in a working directory encoded/decoded
                    and compared with file comparison utility to
                    ensure that multiple runs do not cause problems
                    such as exhausting file handles, tmp storage, etc.

                -------------

                Syntax, operation and failure:
                    All options/switches tested.  Performs as
                    expected.

                case 10:
                    No Args -- Shows Usage Screen
                    Return Code 1 (Invalid Syntax)
                case 11:
                    One Arg (invalid) -- Shows Usage Screen
                    Return Code 1 (Invalid Syntax)
                case 12:
                    One Arg Help (-?) -- Shows detailed Usage Screen.
                    Return Code 0 (Success -- help request is valid).
                case 13:
                    One Arg Help (-h) -- Shows detailed Usage Screen.
                    Return Code 0 (Success -- help request is valid).
                case 14:
                    One Arg (valid) -- Uses stdin/stdout (filter)
                    Return Code 0 (Sucess)
                case 15:
                    Two Args (invalid file) -- shows system error.
                    Return Code 2 (File Error)
                case 16:
                    Encode non-existent file -- shows system error.
                    Return Code 2 (File Error)
                case 17:
                    Out of disk space -- shows system error.
                    Return Code 3 (File I/O Error)

                -------------

                Compile/Regression test:
                    gcc compiled binary under Cygwin
                    Microsoft Visual Studio under Windows 2000
                    Microsoft Version 6.0 C under Windows 2000

DEPENDENCIES:   None

LICENCE:        Copyright (c) 2001 Bob Trower, Trantor Standard Systems Inc.

                Permission is hereby granted, free of charge, to any person
                obtaining a copy of this software and associated
                documentation files (the "Software"), to deal in the
                Software without restriction, including without limitation
                the rights to use, copy, modify, merge, publish, distribute,
                sublicense, and/or sell copies of the Software, and to
                permit persons to whom the Software is furnished to do so,
                subject to the following conditions:

                The above copyright notice and this permission notice shall
                be included in all copies or substantial portions of the
                Software.

                THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
                KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
                WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
                PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
                OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
                OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
                OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
                SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

VERSION HISTORY:
                Bob Trower 08/04/01 -- Create Version 0.00.00B

\******************************************************************* */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <limits.h>
#include <stdbool.h>
#include <rindb/db_error.h>

#include <rindb/db_base64.h>


#define DB_BASE64_MIME 0
#define DB_BASE64_XML 1


#define DB_BASE64_ENCODING DB_BASE64_XML


/*
 * Translation Table as described in RFC1113
 */
#if (DB_BASE64_ENCODING == DB_BASE64_MIME)
  /* MIME style encoding table */
  static const char encoding_table[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

#elif (DB_BASE64_ENCODING == DB_BASE64_XML)
  /** XML style encoding table */
  static const char encoding_table[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789.-";
#endif


/** Decoding table, populated by first call to db_base64_decode */
static char decoding_table[UCHAR_MAX];

/** Has the decoding table been initialised yet */
static bool decoding_table_initialised = false;

#if 0
/*
** Translation Table to decode (created by author)
*/
static const char cd64[]="|$$$}rstuvwxyz{$$$$$$$>?@ABCDEFGHIJKLMNOPQRSTUVW$$$$$$XYZ[\\]^_`abcdefghijklmnopq";
#endif


/*
** encodeblock
**
** encode 3 8-bit binary bytes as 4 '6-bit' characters
*/
static void encodeblock( unsigned char in[3], unsigned char out[4], int len )
{
    out[0] = encoding_table[ in[0] >> 2 ];
    out[1] = encoding_table[ ((in[0] & 0x03) << 4) | ((in[1] & 0xf0) >> 4) ];
    out[2] = (unsigned char) (len > 1 ? encoding_table[ ((in[1] & 0x0f) << 2) | ((in[2] & 0xc0) >> 6) ] : '=');
    out[3] = (unsigned char) (len > 2 ? encoding_table[ in[2] & 0x3f ] : '=');
    out[4] = '\0';
}


/** base64 encode a buffer */
enum eDBError db_base64_encode(const uint8_t *instr, size_t inlen, char *outstr, size_t outmaxlen)
{
  unsigned char in[3], out[5];
  int i, len;
  size_t bytesout;

  for (bytesout = 0; inlen > 0; )
  {
    len = 0;
    for (i = 0; i < 3; i++)
    {
      if (inlen > 0)
      {
        in[i] = *instr++;
        len++;
        inlen--;
      }
      else
      {
        in[i] = 0;
      }
    }

    if (len)
    {
      encodeblock(in, out, len);
      if (bytesout + 4 > outmaxlen)
      {
        /* Error bad length */
        return dbeBufferFull;
      }
      memcpy(outstr, out, 4);
      outstr += 4;
      bytesout += 4;
    }
  }
  *outstr = '\0';
  return dbeSuccess;
}


#if 0

/*
** decodeblock
**
** decode 4 '6-bit' characters into 3 8-bit binary bytes
*/
static void decodeblock( unsigned char in[4], unsigned char out[3] )
{   
    out[ 0 ] = (unsigned char ) (in[0] << 2 | in[1] >> 4);
    out[ 1 ] = (unsigned char ) (in[1] << 4 | in[2] >> 2);
    out[ 2 ] = (unsigned char ) (((in[2] << 6) & 0xc0) | in[3]);
}

/*
** decode
**
** decode a base64 encoded stream discarding padding, line breaks and noise
*/
int db_base64_decode(const char *instr, size_t inlen, uint8_t *outstr, size_t outmaxlen)
{
  unsigned char in[4], out[3], v;
  int i, len;
  size_t bytesout = 0;

  while (inlen > 0)
  {
    for (len = 0, i = 0; i < 4 && inlen > 0; i++)
    {
      v = 0;
      while (inlen > 0 && v == 0 )
      {
        v = *instr++;
        v = (unsigned char) ((v < 43 || v > 122) ? 0 : cd64[ v - 43 ]);
//        v = (unsigned char) ((v < '+' || v > 'z') ? 0 : cd64[ v - '+' ]);
//        v = (unsigned char) ((v < '.' || v > 'z') ? 0 : cd64[ v - '.' ]);
        if (0 != v)
        {
//          v = (unsigned char) ((v == '$') ? 0 : v - '=');
          v = (unsigned char) ((v == '$') ? 0 : v - 61);
        }
        inlen--;
      }
      if (inlen > 0)
      {
        len++;
        if (0 != v)
        {
          in[i] = (unsigned char) (v - 1);
        }
      }
      else
      {
        in[i] = 0;
      }
    } /* for */

    if (len > 0)
    {
      decodeblock(in, out);
      for (i = 0; i < len - 1; i++)
      {
        if (bytesout > outmaxlen)
        {
          return -1;
        }
        bytesout++;
        *outstr++ = out[i];
      }
    }
  } /* while */

  return 0;
}
#endif


#if 0
char *base64_encode(const char *data,
                    size_t input_length,
                    size_t *output_length) {

    *output_length = (size_t) (4.0 * ceil((double) input_length / 3.0));

    char *encoded_data = malloc(*output_length);
    if (encoded_data == NULL) return NULL;

    for (int i = 0, j = 0; i < input_length;) {

        uint32_t octet_a = i < input_length ? data[i++] : 0;
        uint32_t octet_b = i < input_length ? data[i++] : 0;
        uint32_t octet_c = i < input_length ? data[i++] : 0;

        uint32_t triple = (octet_a << 0x10) + (octet_b << 0x08) + octet_c;

        encoded_data[j++] = encoding_table[(triple >> 3 * 6) & 0x3F];
        encoded_data[j++] = encoding_table[(triple >> 2 * 6) & 0x3F];
        encoded_data[j++] = encoding_table[(triple >> 1 * 6) & 0x3F];
        encoded_data[j++] = encoding_table[(triple >> 0 * 6) & 0x3F];
    }

    for (int i = 0; i < mod_table[input_length % 3]; i++)
        encoded_data[*output_length - 1 - i] = '=';

    return encoded_data;
}
#endif



static void build_decoding_table(void)
{
  int i;

  memset(decoding_table, 0, sizeof(decoding_table));
  for (i = 0; i < 0x40; i++)
  {
    decoding_table[(int) encoding_table[i]] = i;
  }
}


/** Decode a base64 encoded buffer */
enum eDBError db_base64_decode(const char *instr, size_t inlen, uint8_t *outstr, size_t outmaxlen, size_t *outlen)
{
  size_t i_outlen;
  size_t i;
  size_t j;
  uint32_t sextet_a;
  uint32_t sextet_b;
  uint32_t sextet_c;
  uint32_t sextet_d;
  uint32_t triple;

  if (!decoding_table_initialised)
  {
    build_decoding_table();
    decoding_table_initialised = true;
  }

  /* Check length */
  if (0 != (inlen % 4))
  {
    return dbeParseFail;
  }
  if (0 == inlen)
  {
    *outlen = 0;
    return dbeSuccess;
  }

  /* Calculate output length */
  i_outlen = inlen / 4 * 3;
  if (instr[inlen - 1] == '=')
  {
    i_outlen--;
  }
  if (instr[inlen - 2] == '=')
  {
    i_outlen--;
  }

  /* Check room in buffer */
  if (outmaxlen < i_outlen)
  {
    return dbeBufferFull;
  }

  /* Decode */
  for (i = 0, j = 0; i < inlen; )
  {
    sextet_a = (instr[i] == '=') ? 0 : decoding_table[(int) instr[i]];
    i++;
    sextet_b = (instr[i] == '=') ? 0 : decoding_table[(int) instr[i]];
    i++;
    sextet_c = (instr[i] == '=') ? 0 : decoding_table[(int) instr[i]];
    i++;
    sextet_d = (instr[i] == '=') ? 0 : decoding_table[(int) instr[i]];
    i++;

    triple = (sextet_a << 3 * 6)
           + (sextet_b << 2 * 6)
           + (sextet_c << 1 * 6)
           + (sextet_d << 0 * 6);

    if (j < i_outlen) outstr[j++] = (triple >> 2 * 8) & 0xFF;
    if (j < i_outlen) outstr[j++] = (triple >> 1 * 8) & 0xFF;
    if (j < i_outlen) outstr[j++] = (triple >> 0 * 8) & 0xFF;
  }
  *outlen = i_outlen;
  return dbeSuccess;
}
