/*
 * db_buffer.c
 *
 *  Created on: 12/11/2010
 *      Author: ian
 */
#include <stdlib.h>
#include <string.h>

#include <rindb/db_buffer.h>


/** Create a buffer with static (supplied) buffer */
enum eDBError db_buffer_create_static(struct s_db_buffer *db_buffer, uint8_t *buf, size_t length)
{
  db_buffer->buffer = buf;
  db_buffer->length = length;
  db_buffer->posn = 0;
  db_buffer->mallocd = false;
  db_buffer->can_resize = false;
  db_buffer->resize_increment = 0;
  return dbeSuccess;
}


/** Calculate a new buffer size (do rounding up to next multiple of resize_increment) */
static enum eDBError db_buffer_calc_bufsize(size_t resize_increment, size_t *length)
{
  size_t len;

  /* Prevent divide by 0 */
  if (0 == resize_increment)
  {
    return dbeParamFail;
  }

  /* Work out new buffer size */
  len = *length;
  if (0 == len)
  {
    /* Zero is a special case -> allocate 1 resize_increment */
    *length = resize_increment;
  }
  else
  {
    if (0 == (len % resize_increment))
    {
      /* On the boundary, no round up required */
      *length = len;
    }
    else
    {
      /* Somewhere in between, round up */
      *length = ((len / resize_increment) + 1) * resize_increment;
    }
  }
  return dbeSuccess;
}


/** Create a buffer by malloc'ing a buffer of specified length */
enum eDBError db_buffer_create_malloc(struct s_db_buffer *db_buffer, size_t length, size_t resize_increment)
{
  uint8_t *buf;
  enum eDBError res;

  /* Round up to resize_increment if required */
  res = db_buffer_calc_bufsize(resize_increment, &length);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Allocate buffer */
  buf = malloc(length);
  if (NULL == buf)
  {
    return dbeMemAllocFail;
  }

  /* Fill out db_buffer structure */
  db_buffer->buffer = buf;
  db_buffer->length = length;
  db_buffer->posn = 0;
  db_buffer->mallocd = true;
  db_buffer->can_resize = true;
  db_buffer->resize_increment = resize_increment;
  return dbeSuccess;
}


/** Create a buffer from a file stream */
enum eDBError db_buffer_create_file(struct s_db_buffer *db_buffer, FILE *f, size_t resize_increment)
{
  long length;
  enum eDBError res;

  /* Seek to end of file, to work out length */
  if (0 != fseek(f, 0, SEEK_END))
  {
    return dbeFileFail;
  }
  length = ftell(f);
  if (length < 0)
  {
    return dbeFileFail;
  }

  /* Seek back to start of file */
  if (0 != fseek(f, 0, SEEK_SET))
  {
    return dbeFileFail;
  }

  /* Create the buffer of the appropriate size */
  res = db_buffer_create_malloc(db_buffer, length, resize_increment);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Load the file into the buffer */
  return db_buffer_load_file(db_buffer, f);
}


/** Resize a buffer, will round up newlength to next resize_increment */
enum eDBError db_buffer_resize(struct s_db_buffer *db_buffer, size_t newlength)
{
  enum eDBError res;
  uint8_t *newbuf;

  if (!db_buffer->mallocd || !db_buffer->can_resize)
  {
    /* Can't resize a static buffer */
    res = dbeMemAllocFail;
  }
  else
  {
    /* Round up to resize_increment if required */
    res = db_buffer_calc_bufsize(db_buffer->resize_increment, &newlength);
    if (dbeSuccess != res)
    {
      return res;
    }

    /* Resize buffer */
    newbuf = realloc(db_buffer->buffer, newlength);
    if (NULL == newbuf)
    {
      /* realloc() failed, will still have old buffer */
      res = dbeMemAllocFail;
    }
    else
    {
      /* Resized */
      db_buffer->length = newlength;
      db_buffer->buffer = newbuf;
      res = dbeSuccess;
    }
  }
  return res;
}


/** Free the data buffer, if originally malloc'd */
static enum eDBError db_buffer_free(struct s_db_buffer *db_buffer)
{
  enum eDBError res;

  if (NULL == db_buffer)
    return dbeNullPointer;

  if (db_buffer->mallocd)
  {
    if (NULL != db_buffer->buffer)
    {
      free(db_buffer->buffer);
    }
    db_buffer->buffer = NULL;
    res = dbeSuccess;
  }
  else
  {
    res = dbeSuccess;
  }
  return res;
}


/** Destroy the buffer, free if originally malloc'd */
enum eDBError db_buffer_destroy(struct s_db_buffer *db_buffer)
{
  if (NULL == db_buffer)
    return dbeNullPointer;

  if (db_buffer->mallocd)
  {
    db_buffer_free(db_buffer);
  }

  db_buffer->buffer = NULL;
  db_buffer->can_resize = false;
  db_buffer->length = 0;
  db_buffer->mallocd = false;
  db_buffer->posn = 0;
  db_buffer->resize_increment = 0;
  return dbeSuccess;
}


/** Load as much data as possible (all if buffer resizing enabled), from current position in file into buffer */
enum eDBError db_buffer_load_file(struct s_db_buffer *db_buffer, FILE *f)
{
  size_t bytes_read;
  ssize_t current_f_posn;
  ssize_t f_length;
  size_t f_toread;
  enum eDBError res;

  /* Where are we in the file? */
  current_f_posn = ftell(f);
  if (current_f_posn < 0)
  {
    return dbeFileFail;
  }

  /* Seek to end of file, to work out how much data we have to read */
  if (0 != fseek(f, 0, SEEK_END))
  {
    return dbeFileFail;
  }
  f_length = ftell(f);
  if (f_length < 0)
  {
    return dbeFileFail;
  }

  /* Seek back to start position */
  if (0 != fseek(f, current_f_posn, SEEK_SET))
  {
    return dbeFileFail;
  }

  /* Work out how many bytes to read from file */
  f_toread = f_length - current_f_posn;

  /* Will it fit? */
  if (db_buffer->length < f_toread)
  {
    /* No, resize buffer */
    res = db_buffer_resize(db_buffer, f_toread);
    if (dbeSuccess != res)
    {
      return res;
    }
  }

  /* Back to start of buffer */
  db_buffer->posn = 0;

  /* Do the read, may not read read_len bytes if file is smaller */
  bytes_read = fread(db_buffer->buffer, sizeof(uint8_t), f_toread, f);
  if (bytes_read != f_toread)
  {
    return dbeFileFail;
  }

  /* Ran into read error */
  if (ferror(f))
  {
    return dbeFileFail;
  }

  return dbeSuccess;
}


/** Move buffer position back to start of buffer */
enum eDBError db_buffer_rewind(struct s_db_buffer *db_buffer)
{
  db_buffer->posn = 0;
  return dbeSuccess;
}


/** Re-allocate a buffer if required */
static enum eDBError db_buffer_checksize(struct s_db_buffer *db_buffer, size_t elemsize)
{
  enum eDBError res = dbeSuccess;

  if (db_buffer->posn + elemsize > db_buffer->length)
  {
    /* Won't fit, try to resize */
    if (db_buffer->can_resize)
    {
      res = db_buffer_resize(db_buffer, db_buffer->length + db_buffer->resize_increment);
      if (dbeSuccess == res)
      {
        /* Check will it fit */
        res = (db_buffer->posn + elemsize > db_buffer->length) ? dbeBufferFull : dbeSuccess;
      }
    }
    else
    {
      res = dbeBufferFull;
    }
  }
  return res;
}


/** Add a byte to the buffer and advance the position */
enum eDBError db_buffer_addbyte(struct s_db_buffer *db_buffer, uint8_t data)
{
  enum eDBError result;

  /* Do we need to resize? */
  result = db_buffer_checksize(db_buffer, sizeof(data));

  /* Add data to buffer */
  if (dbeSuccess == result)
  {
    db_buffer->buffer[db_buffer->posn] = data;
    db_buffer->posn += sizeof(data);
    result = dbeSuccess;
  }
  return result;
}


/** Add a array of bytes to the buffer and advance the position */
enum eDBError db_buffer_addbuffer(struct s_db_buffer *db_buffer, const void *data, size_t length)
{
  enum eDBError result;

  /* Do we need to resize? */
  result = db_buffer_checksize(db_buffer, length);

  /* Add data to buffer */
  if (dbeSuccess == result)
  {
    memcpy(&db_buffer->buffer[db_buffer->posn], data, length);
    db_buffer->posn += length;
    result = dbeSuccess;
  }
  return result;
}


/** Get a byte from the buffer and advance the position */
enum eDBError db_buffer_getbyte(struct s_db_buffer *db_buffer, uint8_t *data)
{
  enum eDBError result = dbeBufferFull;

  if (db_buffer->posn + sizeof(*data) <= db_buffer->length)
  {
    *data = db_buffer->buffer[db_buffer->posn];
    db_buffer->posn += sizeof(*data);
    result = dbeSuccess;
  }
  return result;
}


/** Get an array of bytes from the buffer and advance the position */
enum eDBError db_buffer_getbuffer(struct s_db_buffer *db_buffer, void *data, size_t length)
{
  enum eDBError result = dbeBufferFull;

  if (db_buffer->posn + length <= db_buffer->length)
  {
    memcpy(data, &db_buffer->buffer[db_buffer->posn], length);
    db_buffer->posn += length;
    result = dbeSuccess;
  }
  return result;
}
