/*
 * db_changelog.c
 *
 *  Created on: 19/06/2011
 *      Author: ian
 */
#include <time.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <rindb/db_error.h>
#include <rindb/db_element.h>
#include <rindb/db_database.h>

#include <rindb/db_changelog.h>


static enum eDBError db_changelog_output_readable_string(FILE *file, const char *str)
{
  size_t i;
  size_t length;

  /* NULL pointers are an empty string */
  if (NULL == str)
  {
    str = "";
  }

  /* Quote it for CSV */
  if (EOF == fputc('"', file))
    return dbeFileFail;

  length = strlen(str);
  for (i = 0; i < length; i++)
  {
    if ('"' == str[i])
    {
      /* Because we are making a CSV file, we need to escape quotes */
      if (EOF != fputs("\"\"", file))
        return dbeFileFail;
    }
    else if ('\\' == str[i])
    {
      /* Backslash \ is escaped as \\, because we use \ to escape unprintable characters */
      if (fprintf(file, "\\\\") < 0)
        return dbeFileFail;
    }
    else if (!isprint(str[i]))
    {
      /* Unprintables get \xNN escaped, like a C string */
      if (fprintf(file, "\\x%02x", str[i]) < 0)
        return dbeFileFail;
    }
    else
    {
      /* Printable characters go out unchanged */
      if (EOF == fputc(str[i], file))
        return dbeFileFail;
    }
  }

  /* Quote it for CSV */
  if (EOF == fputc('"', file))
    return dbeFileFail;

  /* Success */
  return dbeSuccess;
}


static enum eDBError db_changelog_output_readable_uint32(FILE *file, uint32_t value)
{
  int res;

  res = fprintf(file, "%lu", (unsigned long) value);
  return (res < 0) ? dbeFileFail : dbeSuccess;
}


static enum eDBError db_changelog_output_readable_uint8(FILE *file, uint8_t value, const char *const *lookup)
{
  enum eDBError res;

  if (NULL != lookup && NULL != lookup[value])
  {
    /* Index into an array, so look it up */
    res = db_changelog_output_readable_string(file, lookup[value]);
  }
  else
  {
    /* Just a number */
    res = db_changelog_output_readable_uint32(file, value);
  }
  return res;
}


static enum eDBError db_changelog_output_readable_int32(FILE *file, int32_t value)
{
  int res;

  res = fprintf(file, "%ld", (long) value);
  return (res < 0) ? dbeFileFail : dbeSuccess;
}


static enum eDBError db_changelog_output_readable_buffer(FILE *file, const uint8_t *buffer, size_t length)
{
  int res;
  size_t i;

  for (i = 0, res = -1; i < length; i++)
  {
    res = fprintf(file, "%02X", buffer[i]);
    if (res < 0)
      break;
  }

  return (res < 0) ? dbeFileFail : dbeSuccess;
}


/** Output element to FILE */
static enum eDBError db_changelog_output_readable(FILE *file, const struct s_db_element *elem, size_t index, int current_not_previous)
{
  enum eDBError res;
  uint8_t val_uint8;
  uint32_t val_uint32;
  int32_t val_int32;
  const char *val_str;
  const uint8_t *val_buf;

  switch (elem->db_elem_type)
  {
    case dbetUInt8:
      val_uint8 = (current_not_previous) ? *elem->elem_data.elem_uint8.data : elem->elem_data.elem_uint8.previous_data;
      res = db_changelog_output_readable_uint8(file, val_uint8, elem->elem_data.elem_uint8.lookup);
      break;

    case dbetUInt32:
      val_uint32 = (current_not_previous) ? *elem->elem_data.elem_uint32.data : elem->elem_data.elem_uint32.previous_data;
      res = db_changelog_output_readable_uint32(file, val_uint32);
      break;

    case dbetInt32:
      val_int32 = (current_not_previous) ? *elem->elem_data.elem_int32.data : elem->elem_data.elem_int32.previous_data;
      res = db_changelog_output_readable_int32(file, val_int32);
      break;

    case dbetString:
      val_str = (current_not_previous) ? elem->elem_data.elem_string.data : elem->elem_data.elem_string.previous_data;
      res = db_changelog_output_readable_string(file, val_str);
      break;

    case dbetBuffer:
      val_buf = (current_not_previous) ? elem->elem_data.elem_buffer.data : elem->elem_data.elem_buffer.previous_data;
      res = db_changelog_output_readable_buffer(file, val_buf, elem->elem_data.elem_buffer.length);
      break;

    case dbetUInt8Array:
      val_uint8 = (current_not_previous) ? elem->elem_data.elem_uint8_array.data[index] : elem->elem_data.elem_uint8_array.previous_data[index];
      res = db_changelog_output_readable_uint8(file, val_uint8, elem->elem_data.elem_uint8_array.lookup);
      break;

    case dbetUInt32Array:
      val_uint32 = (current_not_previous) ? elem->elem_data.elem_uint32_array.data[index] : elem->elem_data.elem_uint32_array.previous_data[index];
      res = db_changelog_output_readable_uint32(file, val_uint32);
      break;

    case dbetInt32Array:
      val_int32 = (current_not_previous) ? elem->elem_data.elem_int32_array.data[index] : elem->elem_data.elem_int32_array.previous_data[index];
      res = db_changelog_output_readable_int32(file, val_int32);
      break;

    case dbetStringArray:
      val_str = (current_not_previous) ? elem->elem_data.elem_string_array.data[index] : elem->elem_data.elem_string_array.previous_data[index];
      res = db_changelog_output_readable_string(file, val_str);
      break;

    case dbetNone:
    default:
      res = dbeFailure;
      break;
  }

  return res;
}


/** Output index to FILE, for array types */
static enum eDBError db_changelog_output_index(FILE *file, const struct s_db_element *elem, size_t index)
{
  enum eDBError res;

  switch (elem->db_elem_type)
  {
    /* Only array types have an index */
    case dbetUInt8Array:
    case dbetUInt32Array:
    case dbetInt32Array:
    case dbetStringArray:
      res = db_changelog_output_readable_uint32(file, (uint32_t) index);
      break;

    default:
      res = dbeSuccess;
      break;
  }

  return res;
}


/** Convert a log mask into a string description */
static const char *db_changelog_logmask_to_string(struct s_db_changelog *cl, uint32_t logmask)
{
  int i;
  int bits;
  const char *res;

  /* Check for NULL pointers */
  if (NULL == cl->logmask_desc)
  {
    res = NULL;
  }
  else
  {

    /* Work out number of bits to work through */
    bits = sizeof(logmask) * 8;

    /* Find first set log mask bit */
    for (i = 0, res = NULL; i < bits; i++)
    {
      if (db_changelog_visible(logmask, (1 << i)))
      {
        res = cl->logmask_desc[i];
        break;
      }
    }
  }
  return (NULL == res) ? "" : res;
}


static enum eDBError db_changelog_output_readable_timestamp(FILE *file, time_t timestamp)
{
  char time_buf[30];
  struct tm tm_time;

  (void) localtime_r(&timestamp, &tm_time);
  (void) strftime(time_buf, 30, DB_CHANGELOG_TIMESTAMP_FORMAT, &tm_time);
  return db_changelog_output_readable_string(file, time_buf);
}


#define DB_FAIL_IF(fn, resvar, test, lbl) { resvar = fn; if (test) goto lbl; }
#define SIO_FAIL_IF(fn, test, failexpr, lbl) { if (fn test) { failexpr; goto lbl; } }


static enum eDBError db_changelog_add_entry_pre_common_internal(struct s_db_changelog *cl, time_t timestamp,
                                                                const char *dbname, const char *elem_db_name,
                                                                const char *elem_menu_name)
{
  enum eDBError dbres;

  /* Seek end of file for new entries */
  SIO_FAIL_IF(fseek(cl->file, 0, SEEK_END), != 0, dbres = dbeFileFail, dcaepci_fail);

  /*
   * Format is:
   * calcounter, configcounter, timestamp, dbname, db_elem_name, menu_elem_name, index, oldvalue, newvalue, logmask_desc
   * eg 10,10,"2012-04-21 21:49:08,userdb,span,Span,,1234,1235,trade"
   * or 10,10,"2012-04-21 21:49:08,userdb,lincor,Lin.Cor,1,1234,1235,trade"
   */

  /* Print change log entry (calcounter, configcounter, time, dbname, db_elem_name, menu_elem_name) */
  if (NULL != cl->calibration_counter)
    DB_FAIL_IF(db_changelog_output_readable_uint32(cl->file, *(cl->calibration_counter)), dbres, dbeSuccess != dbres, dcaepci_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaepci_fail);
  if (NULL != cl->configuration_counter)
    DB_FAIL_IF(db_changelog_output_readable_uint32(cl->file, *(cl->configuration_counter)), dbres, dbeSuccess != dbres, dcaepci_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaepci_fail);
  DB_FAIL_IF(db_changelog_output_readable_timestamp(cl->file, timestamp), dbres, dbeSuccess != dbres, dcaepci_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaepci_fail);
  DB_FAIL_IF(db_changelog_output_readable_string(cl->file, dbname), dbres, dbeSuccess != dbres, dcaepci_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaepci_fail);
  DB_FAIL_IF(db_changelog_output_readable_string(cl->file, elem_db_name), dbres, dbeSuccess != dbres, dcaepci_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaepci_fail);
  DB_FAIL_IF(db_changelog_output_readable_string(cl->file, elem_menu_name), dbres, dbeSuccess != dbres, dcaepci_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaepci_fail);

  dbres = dbeSuccess;

dcaepci_fail:
  return dbres;
}


static enum eDBError db_changelog_add_entry_post_common_internal(struct s_db_changelog *cl, uint32_t changelog_mask)
{
  const char *logmask_desc;
  enum eDBError dbres;

  /* And finally log mask */
  logmask_desc = db_changelog_logmask_to_string(cl, changelog_mask);
  DB_FAIL_IF(db_changelog_output_readable_string(cl->file, logmask_desc), dbres, dbeSuccess != dbres, dcaepci_fail);

  /* And CRLF to finish off as per CSV specification */
  SIO_FAIL_IF(fprintf(cl->file, "\r\n"), < 0, dbres = dbeFileFail, dcaepci_fail);

  dbres = dbeSuccess;

dcaepci_fail:
  return dbres;
}


static enum eDBError db_changelog_add_entry_internal(struct s_db_changelog *cl, time_t timestamp,
                                                     const char *dbname, int index,
                                                     const struct s_db_element *elem)
{
  enum eDBError dbres;

  /*
   * Format is:
   * calcounter, configcounter, timestamp, dbname, db_elem_name, menu_elem_name, index, oldvalue, newvalue, logmask_desc
   * eg 10,10,"2012-04-21 21:49:08,userdb,span,Span,,1234,1235,trade"
   * or 10,10,"2012-04-21 21:49:08,userdb,lincor,Lin.Cor,1,1234,1235,trade"
   */

  /* Print change log entry (calcounter, configcounter, time, dbname, db_elem_name, menu_elem_name) */
  DB_FAIL_IF(db_changelog_add_entry_pre_common_internal(cl, timestamp, dbname, elem->elem_db_name, elem->elem_menu_name), dbres, dbeSuccess != dbres, dcaei_fail);

  /* Print index if there is one */
  DB_FAIL_IF(db_changelog_output_index(cl->file, elem, index), dbres, dbeSuccess != dbres, dcaei_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaei_fail);

  /* Add old value, then new value */
  DB_FAIL_IF(db_changelog_output_readable(cl->file, elem, index, 0), dbres, dbeSuccess != dbres, dcaei_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaei_fail);
  DB_FAIL_IF(db_changelog_output_readable(cl->file, elem, index, 1), dbres, dbeSuccess != dbres, dcaei_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcaei_fail);

  /* And finally log mask and CRLF to finish off as per CSV specification */
  DB_FAIL_IF(db_changelog_add_entry_post_common_internal(cl, elem->changelog_mask), dbres, dbeSuccess != dbres, dcaei_fail);

  dbres = dbeSuccess;

dcaei_fail:
  fflush(cl->file);
  return dbres;
}


static enum eDBError db_changelog_add_misc_entry_internal(struct s_db_changelog *cl, time_t timestamp,
                                                          const char *db_name,
                                                          const char *entry_db_name, const char *entry_menu_name,
                                                          const char *entry_oldvalue, const char *entry_newvalue,
                                                          uint32_t entry_logmask)
{
  enum eDBError dbres;

  /*
   * Format is:
   * calcounter, configcounter, timestamp, dbname, db_elem_name, menu_elem_name, index, oldvalue, newvalue, logmask_desc
   * eg 10,10,"2012-04-21 21:49:08,userdb,span,Span,,1234,1235,trade"
   * or 10,10,"2012-04-21 21:49:08,userdb,lincor,Lin.Cor,1,1234,1235,trade"
   */

  /* Print change log entry (calcounter, configcounter, time, dbname, db_elem_name, menu_elem_name) */
  DB_FAIL_IF(db_changelog_add_entry_pre_common_internal(cl, timestamp, db_name, entry_db_name, entry_menu_name), dbres, dbeSuccess != dbres, dcamei_fail);

  /* No index */
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcamei_fail);

  /* Add old value, then new value */
  DB_FAIL_IF(db_changelog_output_readable_string(cl->file, entry_oldvalue), dbres, dbeSuccess != dbres, dcamei_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcamei_fail);
  DB_FAIL_IF(db_changelog_output_readable_string(cl->file, entry_newvalue), dbres, dbeSuccess != dbres, dcamei_fail);
  SIO_FAIL_IF(fprintf(cl->file, ","), < 0, dbres = dbeFileFail, dcamei_fail);

  /* And finally log mask and CRLF to finish off as per CSV specification */
  DB_FAIL_IF(db_changelog_add_entry_post_common_internal(cl, entry_logmask), dbres, dbeSuccess != dbres, dcamei_fail);

  dbres = dbeSuccess;

dcamei_fail:
  fflush(cl->file);
  return dbres;
}


/** Add an entry for each array element that is different in a uint8 array */
static enum eDBError db_changelog_add_entry_array(struct s_db_changelog *cl, time_t timestamp,
                                                  const char *dbname,
                                                  const struct s_db_element *elem)
{
  size_t index;
  enum eDBError res;
  size_t num_elems;

  /* Work out number of elements */
  switch (elem->db_elem_type)
  {
    case dbetUInt8Array:
      num_elems = elem->elem_data.elem_uint8_array.num_elements;
      break;

    case dbetUInt32Array:
      num_elems = elem->elem_data.elem_uint32_array.num_elements;
      break;

    case dbetInt32Array:
      num_elems = elem->elem_data.elem_int32_array.num_elements;
      break;

    case dbetNone:
    default:
      return dbeFailure;
      break;
  }

  for (index = 0; index < num_elems; index++)
  {
    if ((dbetUInt8Array == elem->db_elem_type && elem->elem_data.elem_uint8_array.previous_data[index] != elem->elem_data.elem_uint8_array.data[index]) ||
        (dbetUInt32Array == elem->db_elem_type && elem->elem_data.elem_uint32_array.previous_data[index] != elem->elem_data.elem_uint32_array.data[index]) ||
        (dbetInt32Array == elem->db_elem_type && elem->elem_data.elem_int32_array.previous_data[index] != elem->elem_data.elem_int32_array.data[index]) ||
        (dbetStringArray == elem->db_elem_type && 0 != strcmp(elem->elem_data.elem_string_array.previous_data[index], elem->elem_data.elem_string_array.data[index])))
    {
      /* Print element into change log */
      res = db_changelog_add_entry_internal(cl, timestamp, dbname, index, elem);
      if (dbeSuccess != res)
      {
        break;
      }
    }
  }
  return res;
}


static enum eDBError db_changelog_add_entry(struct s_db_changelog *cl, time_t timestamp, const char *dbname, const struct s_db_element *elem, uint32_t at_version)
{
  enum eDBError res;

  /* Is element not valid at this version, or masked out */
  if (!db_element_valid_for_version(elem, at_version) || !db_element_changelog_visible(elem, cl->logmask))
  {
    /* This is not a failure, we simply don't write this value */
    return dbeSuccess;
  }

  /* Convert each element type to string */
  switch (elem->db_elem_type)
  {

    /* Print element into change log */
    case dbetUInt8:
    case dbetUInt32:
    case dbetInt32:
    case dbetString:
    case dbetBuffer:
      res = db_changelog_add_entry_internal(cl, timestamp, dbname, 0, elem);
      break;

    /* Print changed array elements into change log */
    case dbetUInt8Array:
    case dbetUInt32Array:
    case dbetInt32Array:
    case dbetStringArray:
      res = db_changelog_add_entry_array(cl, timestamp, dbname, elem);
      break;

    case dbetNone:
    default:
      res = dbeFailure;
      break;
  }
  return res;
}


/** Write any changes in a database to the change log. */
enum eDBError db_changelog_update_database(struct s_db_changelog *cl, time_t timestamp, const struct s_db_database *database, uint32_t at_version)
{
  size_t i;
  enum eDBError res;
  const struct s_db_element *elem;

  /* Check parameters */
  if (NULL == cl || NULL == database)
  {
    return dbeNullPointer;
  }

  /* Is there room in the change log? */
  res = db_changelog_is_full(cl);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Write each section into the change log */
  for (i = 0, res = dbeSuccess; (i < database->db_num_elements && dbeSuccess == res); i++)
  {
    elem = &database->db_elements[i];
    if (db_element_changed(elem))
    {
      res = db_changelog_add_entry(cl, timestamp, database->db_name, elem, at_version);
      if (dbeSuccess != res)
      {
        return res;
      }
    }
  }

  return res;
}


/**
 * Add a header row to the change log, call after creating or clearing the changelog
 *
 * @param cl is the change log to add the misc entry to. Not NULL
 *
 * @return the result of the operation
 */
static enum eDBError db_changelog_add_header_row(struct s_db_changelog *cl)
{
  enum eDBChangeLogElement elem;
  enum eDBError res;

  if (NULL != cl->column_desc)
  {
    /* Add header row */
    elem = 0;
    res = db_changelog_output_readable_string(cl->file, cl->column_desc[elem++]);
    if (dbeSuccess != res)
    {
      return res;
    }
    for ( ; elem < dbcleNum; elem++)
    {
      if (fprintf(cl->file, ",") < 0)
      {
        return dbeFileFail;
      }
      res = db_changelog_output_readable_string(cl->file, cl->column_desc[elem]);
      {
        if (dbeSuccess != res)
        {
          return res;
        }
      }
    }
  }

  /* And CRLF to finish off as per CSV spec, will be blank  */
  res = (fprintf(cl->file, "\r\n") < 0) ? dbeFileFail : dbeSuccess;
  return res;
}


/**
 * Add a miscellaneous entry in the change log
 *
 * @param cl is the change log to add the misc entry to. Not NULL
 * @param timestamp is the time at which the misc entry occurred
 * @param db_name is the database to which the misc entry is for. Can be NULL, in which case the field will be blank
 * @param entry_db_name is the database name of the entry (may be saved into the database). Can be NULL, in which case the field will be blank
 * @param entry_menu_name is the menu name of the entry, used for easy identification. Can be NULL, in which case the field will be blank
 * @param entry_oldvalue is the previous value of the entry. Can be NULL, in which case the field will be blank
 * @param entry_newvalue is the new value of the entry. Can be NULL, in which case the field will be blank
 * @param entry_logmask is the mask for the entry to determine if it should appear in the change log.
 *
 * @return the result of the operation
 */
enum eDBError db_changelog_add_misc_entry(struct s_db_changelog *cl, time_t timestamp,
                                          const char *db_name,
                                          const char *entry_db_name, const char *entry_menu_name,
                                          const char *entry_oldvalue, const char *entry_newvalue,
                                          uint32_t entry_logmask)
{
  enum eDBError res;

  /* Check parameters */
  if (NULL == cl)
  {
    return dbeNullPointer;
  }

  /* Is element masked out? */
  if (!db_changelog_visible(entry_logmask, cl->logmask))
  {
    /* This is not a failure, we simply don't write this value */
    return dbeSuccess;
  }

  /* Is there room in the change log? */
  res = db_changelog_is_full(cl);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Add the miscellaneous entry */
  res = db_changelog_add_misc_entry_internal(cl, timestamp,
                                             db_name,
                                             entry_db_name, entry_menu_name,
                                             entry_oldvalue, entry_newvalue,
                                             entry_logmask);
  return res;
}


/** Empty the change log, and log an event to signal this with event_logmask */
enum eDBError db_changelog_clear(struct s_db_changelog *cl, time_t timestamp, const char *dbname, uint32_t event_logmask)
{
  enum eDBError res = dbeSuccess;

  /* Check parameters */
  if (NULL == cl)
  {
    return dbeNullPointer;
  }

  /* Truncate file (to 0 bytes) */
  cl->file = freopen(cl->filename, "w+", cl->file);
  if (NULL == cl->file)
  {
    return dbeFileFail;
  }
  /* freopen() loses FD_CLOEXEC, so need to set again */
  if (fcntl(fileno(cl->file), F_SETFD, FD_CLOEXEC) < 0)
  {
    return dbeFileFail;
  }

  /* Back to start of file */
  if (0 != fseek(cl->file, 0, SEEK_SET))
  {
    return dbeFileFail;
  }

  /* Add the header row */
  res = db_changelog_add_header_row(cl);
  if (dbeSuccess != res)
  {
    return res;
  }

  if (NULL != cl->cl_event_db_name && NULL != cl->cleared_entry_value)
  {
    /* Add a record to indicate the change log has been emptied */
    res = db_changelog_add_misc_entry(cl, timestamp,
                                      dbname,
                                      cl->cl_event_db_name, cl->cl_event_menu_name,
                                      "", cl->cleared_entry_value,
                                      event_logmask);
  }
  return res;
}


/**
 * Get the current size of the change log (in bytes)
 *
 * @param cl is the change log. Not NULL
 * @param cl_size is the current size of the change log. Not NULL
 *
 * @return dbeSuccess if change log size retrieved successfully, other errors if something went wrong
 */
enum eDBError db_changelog_get_size(struct s_db_changelog *cl, size_t *cl_size)
{
  long posn;

  /* Check parameters */
  if (NULL == cl || NULL == cl->file || NULL == cl_size)
  {
    return dbeNullPointer;
  }

  /* Seek end of file */
  if (0 != fseek(cl->file, 0, SEEK_END))
  {
    return dbeFileFail;
  }

  /* Get current position */
  posn = ftell(cl->file);
  if (posn < 0)
  {
    return dbeFileFail;
  }

  /* Return size */
  *cl_size = posn;
  return dbeSuccess;
}


/**
 * Get the maximum size of the change log (in bytes)
 *
 * @param cl is the change log. Not NULL
 * @param cl_size is the maximum size of the change log. Not NULL
 *
 * @return dbeSuccess if change log maximum size retrieved successfully, other errors if something went wrong
 */
enum eDBError db_changelog_get_maximum_size(struct s_db_changelog *cl, size_t *cl_size)
{
  if (NULL == cl || NULL == cl_size)
  {
    return dbeNullPointer;
  }

  *cl_size = cl->soft_maxfilesize;
  return dbeSuccess;
}


/**
 * Determine if the change log is full
 *
 * @param cl is the change log. Not NULL
 *
 * @return dbeSuccess if change log not full, dbeChangelogFull if change log full, other errors if something went wrong
 */
enum eDBError db_changelog_is_full(struct s_db_changelog *cl)
{
  enum eDBError res;
  size_t current_size = 0;

  /* Check parameters */
  if (NULL == cl || NULL == cl->file)
  {
    return dbeNullPointer;
  }

  /* Get current size of change log */
  res = db_changelog_get_size(cl, &current_size);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Can any more records fit */
  res = (current_size >= cl->soft_maxfilesize) ? dbeChangelogFull : dbeSuccess;
  return res;
}


/** Create a slave copy of the change log, written to outfile */
enum eDBError db_changelog_slave_copy_file(struct s_db_changelog *cl, FILE *outfile)
{
  #define COPY_BUF_SIZE 1024
  uint8_t *buf = NULL;
  enum eDBError res;
  size_t toread;
  size_t bytesread;
  size_t byteswritten;

  /* Check parameters */
  if (NULL == cl || NULL == cl->file || NULL == outfile)
  {
    res = dbeNullPointer;
    goto dcscf_fail;
  }

  /* Allocate copy buffer */
  buf = malloc(COPY_BUF_SIZE);
  if (NULL == buf)
  {
    res = dbeMemAllocFail;
    goto dcscf_fail;
  }

  /* Seek start of change log */
  if (0 != fseek(cl->file, 0, SEEK_SET))
  {
    res = dbeFileFail;
    goto dcscf_fail;
  }

  /* Copy from one to the other */
  while (1)
  {
    toread = COPY_BUF_SIZE;

    /* Read bytes */
    bytesread = fread(buf, sizeof(uint8_t), toread, cl->file);
    if (0 == bytesread && ferror(cl->file))
    {
      res = dbeFileFail;
      goto dcscf_fail;
    }
    byteswritten = fwrite(buf, sizeof(uint8_t), bytesread, outfile);
    if (ferror(outfile) || bytesread != byteswritten)
    {
      res = dbeFileFail;
      goto dcscf_fail;
    }
    if (feof(cl->file))
    {
      /* Success */
      res = dbeSuccess;
      break;
    }
  }

dcscf_fail:

  /* Seek end of change log, just to leave it ready for next entry, ignoring the result */
  (void) fseek(cl->file, 0, SEEK_END);

  /* Clean up */
  if (NULL != buf)
  {
    free(buf);
    buf = NULL;
  }
  return res;
}


/** Create a slave copy of the change log, written to filename */
enum eDBError db_changelog_slave_copy(struct s_db_changelog *cl, const char *filename)
{
  FILE *outfile = NULL;
  enum eDBError res;

  /* Check parameters */
  if (NULL == cl || NULL == cl->file || NULL == filename)
  {
    res = dbeNullPointer;
    goto dcsc_fail;
  }

  /* Open output file, truncate to 0 bytes ready for writing */
  outfile = fopen(filename, "w");
  if (NULL == outfile)
  {
    res = dbeFileFail;
    goto dcsc_fail;
  }
  if (fcntl(fileno(outfile), F_SETFD, FD_CLOEXEC) < 0)
  {
    res = dbeFileFail;
    goto dcsc_fail;
  }

  /* Write slave copy of changelog */
  res = db_changelog_slave_copy_file(cl, outfile);

dcsc_fail:
  /* Clean up */
  if (NULL != outfile)
  {
    fclose(outfile);
    outfile = NULL;
  }
  return res;
}


/**
 * Create a change log structure, open the change log file, add an initial entry if doesn't exist.
 *
 * @param cl is a pointer to a pointer to a change log structure allocated and filled out by this function. Not NULL
 * @param filename is the name of the file to be used as the change log. Not NULL
 * @param soft_maxfilesize is the approximate maximum size of the change log. Once the filesize goes over this limit, no more records can be added, however the last record written might go over this limit
 * @param logmask is the mask of what elements should be written to the change log
 * @param logmask_desc is an array of pointers to strings that describe the logmask bits (ie logmask_desc[1] = logmask bit 0). Must be valid for the life of cl (no internal copy). Can be NULL if these are not required.
 * @param cl_event_db_name is used when change log events are added to the change log (created, cleared). Must be valid for the life of cl (no internal copy). Can be NULL in which case these events won't be added to change log.
 * @param cl_event_menu_name is used when change log events are added to the change log (created, cleared). Must be valid for the life of cl (no internal copy). Can be NULL in which case these events won't be added to change log.
 * @param initial_entry_value is added to the change log when the change log is created. Must be valid for the life of cl (no internal copy). Can be NULL in which case these events won't be added to change log.
 * @param cleared_entry_value is added to the change log when the change log is cleared. Must be valid for the life of cl (no internal copy). Can be NULL in which case these events won't be added to change log.
 * @param column_desc is a description of each column in the change log, used to add first row of change log. Can be NULL, in which case the first row will be empty. If not NULL, must have dbcleNum entries in it
 * @param calibration_counter is a pointer to the system calibration counter, recorded in change log records if not NULL. NTEP & OIML
 * @param configuration_counter is a pointer to the system configuration counter, recorded in change log records if not NULL. NTEP
 * @param timestamp is only used if the change log has to be created to add an initial entry
 * @param dbname is only used if the change log has to be created to add an initial entry
 * @param event_logmask is only used if the change log has to be created to add an initial entry
 */
enum eDBError db_changelog_create(struct s_db_changelog **cl,
                                  const char *filename,
                                  size_t soft_maxfilesize,
                                  uint32_t logmask,
                                  const char *const *logmask_desc,
                                  const char *cl_event_db_name,
                                  const char *cl_event_menu_name,
                                  const char *initial_entry_value,
                                  const char *cleared_entry_value,
                                  const char *const *column_desc,
                                  const uint32_t *calibration_counter,
                                  const uint32_t *configuration_counter,
                                  time_t timestamp,
                                  const char *dbname,
                                  uint32_t event_logmask)
{
  int needs_initial_entry;
  enum eDBError res;
  struct s_db_changelog *dcl = NULL;

  /* Check parameters */
  if (NULL == cl || NULL == filename)
  {
    res = dbeNullPointer;
    goto dcc_fail;
  }

  /* Create change log structure */
  dcl = malloc(sizeof(struct s_db_changelog));
  if (NULL == dcl)
  {
    res = dbeMemAllocFail;
    goto dcc_fail;
  }
  memset(dcl, 0, sizeof(struct s_db_changelog));

  /* Allocate filename */
  dcl->filename = strdup(filename);
  if (NULL == dcl->filename)
  {
    res = dbeMemAllocFail;
    goto dcc_fail;
  }

  /* Open change log file */
  needs_initial_entry = 0;
  dcl->file = fopen(dcl->filename, "r+");
  if (NULL == dcl->file)
  {
    /* File does not exist, try creating it and add a initial entry */
    needs_initial_entry = 1;
    dcl->file = fopen(dcl->filename, "w+");
    if (NULL == dcl->file)
    {
      res = dbeFileFail;
      goto dcc_fail;
    }
  }
  if (fcntl(fileno(dcl->file), F_SETFD, FD_CLOEXEC) < 0)
  {
    res = dbeFileFail;
    goto dcc_fail;
  }

  /* Fill out change log structure */
  dcl->soft_maxfilesize = soft_maxfilesize;
  dcl->logmask = logmask;
  dcl->logmask_desc = logmask_desc;
  dcl->cl_event_db_name = cl_event_db_name;
  dcl->cl_event_menu_name = cl_event_menu_name;
  dcl->initial_entry_value = initial_entry_value;
  dcl->cleared_entry_value = cleared_entry_value;
  dcl->column_desc = column_desc;
  dcl->calibration_counter = calibration_counter;
  dcl->configuration_counter = configuration_counter;
  *cl = dcl;

  if (needs_initial_entry)
  {
    /* Add a header row */
    res = db_changelog_add_header_row(dcl);
    if (dbeSuccess != res)
    {
      goto dcc_fail;
    }

    if (NULL != dcl->cl_event_db_name && NULL != dcl->initial_entry_value)
    {
      /* Add a record to indicate the change log has been created */
      res = db_changelog_add_misc_entry(dcl, timestamp,
                                        dbname,
                                        dcl->cl_event_db_name, dcl->cl_event_menu_name,
                                        "", dcl->initial_entry_value,
                                        event_logmask);
      if (dbeSuccess != res)
      {
        goto dcc_fail;
      }
    }
  }

  /* Success */
  return dbeSuccess;

dcc_fail:
  *cl = NULL;
  if (NULL != dcl)
  {
    if (NULL != dcl->filename)
    {
      free(dcl->filename);
      dcl->filename = NULL;
    }
    if (NULL != dcl->file)
    {
      fclose(dcl->file);
      dcl->file = NULL;
    }
    memset(dcl, 0, sizeof(struct s_db_changelog));
    free(dcl);
  }
  return res;
}


/** Clean up a change log structure */
enum eDBError db_changelog_destroy(struct s_db_changelog *cl)
{
  if (NULL != cl)
  {
    if (NULL != cl->filename)
    {
      free(cl->filename);
      cl->filename = NULL;
    }
    if (NULL != cl->file)
    {
      fclose(cl->file);
      cl->file = NULL;
    }
    memset(cl, 0, sizeof(struct s_db_changelog));
    free(cl);
    cl = NULL;
  }
  return dbeSuccess;
}
