/*
 * db_changelog.h
 *
 *  Created on: 19/06/2011
 *      Author: ian
 */

#ifndef DB_CHANGELOG_H_
#define DB_CHANGELOG_H_

#include <stdio.h>
#include <stddef.h>
#include <stdint.h>
#include <time.h>
#include <rindb/db_database.h>
#include <rindb/db_error.h>


/** How the time/date should be printed using strftime/strptime */
#define DB_CHANGELOG_TIMESTAMP_FORMAT "%Y-%m-%d %H:%M:%S"


/** Each column/entry member in the change log */
enum eDBChangeLogElement
{
  dbcleCalibrationCounter,        /*!< Calibration counter value (OIML or NTEP) */
  dbcleConfigurationCounter,      /*!< Configuration counter value (NTEP) */
  dbcleTimestamp,                 /*!< Time stamp of entry */
  dbcleDatabaseName,              /*!< Name of the database for the entry */
  dbcleElementDatabaseName,       /*!< The element name with in the database */
  dbcleElementMenuName,           /*!< The element name in the application menus */
  dbcleIndex,                     /*!< Index if it is an array element */
  dbcleOldValue,                  /*!< The previous value of the entry */
  dbcleNewValue,                  /*!< The current value of the entry */
  dbcleLogMask,                   /*!< Log mask description (trade, user, etc */

  dbcleNum                        /*!< Number of columns/entry members in a change log entry */
};


/** Structure to describe a change log */
struct s_db_changelog
{
  char *filename;                         /*!< The change log file name, used to re-open it when clearing the change log */
  FILE *file;                             /*!< The change log file stream */
  size_t soft_maxfilesize;                /*!< The maximum file size for the change log in bytes. This can be exceeded by an entire database write, assuming it was below the limit before the write. Any further attempts to write the change log will fail. */

  uint32_t logmask;                       /*!< The mask ANDed with each elements change log mask to determine if it should be written to the change log */
  const char *const *logmask_desc;        /*!< Descriptions of each log mask (eg mask 0x01 (bit 0) = logmask_desc[0]) */
  const char *cl_event_db_name;           /*!< Used when change log events are added to the change log (database column) (created, cleared) */
  const char *cl_event_menu_name;         /*!< Used when change log events are added to the change log (more human readable column) (created, cleared) */
  const char *initial_entry_value;        /*!< What to add to the change log when the change log is created */
  const char *cleared_entry_value;        /*!< What to add to the change log when the change log is cleared */
  const char *const *column_desc;         /*!< Descriptions of each column in the change log, used to add first row of change log. Has dbcleNum members */

  const uint32_t *calibration_counter;    /*!< The system calibration counter, recorded in change log records if not NULL. NTEP & OIML */
  const uint32_t *configuration_counter;  /*!< The system configuration counter, recorded in change log records if not NULL. NTEP */
};


/** Use element and change log masks to determine if an element should appear in a change log */
#define db_changelog_visible(elem_logmask, changelog_mask) (0 != (elem_logmask & changelog_mask))


/** Write any changes in a database to the change log. */
enum eDBError db_changelog_update_database(struct s_db_changelog *cl, time_t timestamp, const struct s_db_database *database, uint32_t at_version);


/**
 * Add a miscellaneous entry in the change log
 *
 * @param cl is the change log to add the misc entry to. Not NULL
 * @param timestamp is the time at which the misc entry occurred
 * @param db_name is the database to which the misc entry is for. Can be NULL, in which case the field will be blank
 * @param entry_db_name is the database name of the entry (may be saved into the database). Can be NULL, in which case the field will be blank
 * @param entry_menu_name is the menu name of the entry, used for easy identification. Can be NULL, in which case the field will be blank
 * @param entry_oldvalue is the previous value of the entry. Can be NULL, in which case the field will be blank
 * @param entry_newvalue is the new value of the entry. Can be NULL, in which case the field will be blank
 * @param entry_logmask is the mask for the entry to determine if it should appear in the change log.
 *
 * @return the result of the operation
 */
enum eDBError db_changelog_add_misc_entry(struct s_db_changelog *cl, time_t timestamp,
                                          const char *db_name,
                                          const char *entry_db_name, const char *entry_menu_name,
                                          const char *entry_oldvalue, const char *entry_newvalue,
                                          uint32_t entry_logmask);


/** Empty the change log, and log an event to signal this with event_logmask */
enum eDBError db_changelog_clear(struct s_db_changelog *cl, time_t timestamp, const char *section_name, uint32_t event_logmask);


/**
 * Get the current size of the change log (in bytes)
 *
 * @param cl is the change log. Not NULL
 * @param cl_size is the current size of the change log. Not NULL
 *
 * @return dbeSuccess if change log size retrieved successfully, other errors if something went wrong
 */
enum eDBError db_changelog_get_size(struct s_db_changelog *cl, size_t *cl_size);


/**
 * Get the maximum size of the change log (in bytes)
 *
 * @param cl is the change log. Not NULL
 * @param cl_size is the maximum size of the change log. Not NULL
 *
 * @return dbeSuccess if change log maximum size retrieved successfully, other errors if something went wrong
 */
enum eDBError db_changelog_get_maximum_size(struct s_db_changelog *cl, size_t *cl_size);


/**
 * Determine if the change log is full
 *
 * @param cl is the change log. Not NULL
 *
 * @return dbeSuccess if change log not full, dbeChangelogFull if change log full, other errors if something went wrong
 */
enum eDBError db_changelog_is_full(struct s_db_changelog *cl);


/** Create a slave copy of the change log, written to outfile */
enum eDBError db_changelog_slave_copy_file(struct s_db_changelog *cl, FILE *outfile);


/** Create a slave copy of the change log, written to filename */
enum eDBError db_changelog_slave_copy(struct s_db_changelog *cl, const char *filename);


/**
 * Create a change log structure, open the change log file, add an initial entry if doesn't exist.
 *
 * @param cl is a pointer to a pointer to a change log structure allocated and filled out by this function. Not NULL
 * @param filename is the name of the file to be used as the change log. Not NULL
 * @param soft_maxfilesize is the approximate maximum size of the change log. Once the filesize goes over this limit, no more records can be added, however the last record written might go over this limit
 * @param logmask is the mask of what elements should be written to the change log
 * @param logmask_desc is an array of pointers to strings that describe the logmask bits (ie logmask_desc[1] = logmask bit 0). Must be valid for the life of cl (no internal copy). Can be NULL if these are not required.
 * @param cl_event_db_name is used when change log events are added to the change log (created, cleared). Must be valid for the life of cl (no internal copy). Can be NULL in which case these events won't be added to change log.
 * @param cl_event_menu_name is used when change log events are added to the change log (created, cleared). Must be valid for the life of cl (no internal copy). Can be NULL in which case these events won't be added to change log.
 * @param initial_entry_value is added to the change log when the change log is created. Must be valid for the life of cl (no internal copy). Can be NULL in which case these events won't be added to change log.
 * @param cleared_entry_value is added to the change log when the change log is cleared. Must be valid for the life of cl (no internal copy). Can be NULL in which case these events won't be added to change log.
 * @param column_desc is a description of each column in the change log, used to add first row of change log. Can be NULL, in which case the first row will be empty. If not NULL, must have dbcleNum entries in it
 * @param calibration_counter is a pointer to the system calibration counter, recorded in change log records if not NULL. NTEP & OIML
 * @param configuration_counter is a pointer to the system configuration counter, recorded in change log records if not NULL. NTEP
 * @param timestamp is only used if the change log has to be created to add an initial entry
 * @param dbname is only used if the change log has to be created to add an initial entry
 * @param event_logmask is only used if the change log has to be created to add an initial entry
 */
enum eDBError db_changelog_create(struct s_db_changelog **cl,
                                  const char *filename,
                                  size_t soft_maxfilesize,
                                  uint32_t logmask,
                                  const char *const *logmask_desc,
                                  const char *cl_event_db_name,
                                  const char *cl_event_menu_name,
                                  const char *initial_entry_value,
                                  const char *cleared_entry_value,
                                  const char *const *column_desc,
                                  const uint32_t *calibration_counter,
                                  const uint32_t *configuration_counter,
                                  time_t timestamp,
                                  const char *dbname,
                                  uint32_t event_logmask);


/** Clean up a change log structure */
enum eDBError db_changelog_destroy(struct s_db_changelog *cl);


#endif /* DB_CHANGELOG_H_ */
