/*
 * db_changelog_reader.c
 *
 *  Created on: Apr 21, 2012
 *      Author: ian
 */
#define _XOPEN_SOURCE 500
#include <string.h>
#include <time.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <rindb/db_error.h>
#include <rindb/db_changelog.h>
#include <libcsv/csv.h>

#include <rindb/db_changelog_reader.h>


/** Create by mallocing a new change log entry structure. Members are set to NULL/0. */
enum eDBError db_changelog_reader_entry_create(struct s_db_changelog_entry **cl_entry)
{
  enum eDBError res;
  struct s_db_changelog_entry *cle = NULL;

  /* Check parameters */
  if (NULL == cl_entry)
  {
    return dbeNullPointer;
  }

  /* Allocate changelog entry structure */
  cle = calloc(1, sizeof(struct s_db_changelog_entry));
  if (NULL == cle)
  {
    res = dbeMemAllocFail;
    goto dcrec_fail;
  }

  /* Success */
  *cl_entry = cle;
  return dbeSuccess;

dcrec_fail:
  *cl_entry = NULL;
  if (NULL != cle)
  {
    free(cle);
  }
  return res;
}


/** Destroy a changelog reader entry, frees all allocated memory */
enum eDBError db_changelog_reader_entry_destroy(struct s_db_changelog_entry *cl_entry)
{
  if (NULL != cl_entry)
  {
    if (NULL != cl_entry->cal_counter)
    {
      free(cl_entry->cal_counter);
      cl_entry->cal_counter = NULL;
    }
    if (NULL != cl_entry->config_counter)
    {
      free(cl_entry->config_counter);
      cl_entry->config_counter = NULL;
    }
    if (NULL != cl_entry->db_name)
    {
      free(cl_entry->db_name);
      cl_entry->db_name = NULL;
    }
    if (NULL != cl_entry->db_elem_name)
    {
      free(cl_entry->db_elem_name);
      cl_entry->db_elem_name = NULL;
    }
    if (NULL != cl_entry->menu_elem_name)
    {
      free(cl_entry->menu_elem_name);
      cl_entry->menu_elem_name = NULL;
    }
    if (NULL != cl_entry->old_value)
    {
      free(cl_entry->old_value);
      cl_entry->old_value = NULL;
    }
    if (NULL != cl_entry->new_value)
    {
      free(cl_entry->new_value);
      cl_entry->new_value = NULL;
    }
    if (NULL != cl_entry->log_mask)
    {
      free(cl_entry->log_mask);
      cl_entry->log_mask = NULL;
    }
    free(cl_entry);
    cl_entry = NULL;
  }

  return dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_calcounter_set(struct s_db_changelog_entry *cl_entry, const char *cal_counter)
{
  if (NULL == cl_entry || NULL == cal_counter)
    return dbeNullPointer;

  /* Free existing */
  if (NULL != cl_entry->cal_counter)
    free(cl_entry->cal_counter);

  /* Duplicate name */
  cl_entry->cal_counter = strdup(cal_counter);
  return (NULL == cl_entry->cal_counter) ? dbeMemAllocFail : dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_configcounter_set(struct s_db_changelog_entry *cl_entry, const char *config_counter)
{
  if (NULL == cl_entry || NULL == config_counter)
    return dbeNullPointer;

  /* Free existing */
  if (NULL != cl_entry->config_counter)
    free(cl_entry->config_counter);

  /* Duplicate name */
  cl_entry->config_counter = strdup(config_counter);
  return (NULL == cl_entry->config_counter) ? dbeMemAllocFail : dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_time_set(struct s_db_changelog_entry *cl_entry, const char *time)
{
  if (NULL == cl_entry || NULL == time)
    return dbeNullPointer;

  if (NULL == strptime(time, DB_CHANGELOG_TIMESTAMP_FORMAT, &cl_entry->timestamp))
  {
    return dbeParseFail;
  }
  return dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_dbname_set(struct s_db_changelog_entry *cl_entry, const char *db_name)
{
  if (NULL == cl_entry || NULL == db_name)
    return dbeNullPointer;

  /* Free existing */
  if (NULL != cl_entry->db_name)
    free(cl_entry->db_name);

  /* Duplicate name */
  cl_entry->db_name = strdup(db_name);
  return (NULL == cl_entry->db_name) ? dbeMemAllocFail : dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_dbelemname_set(struct s_db_changelog_entry *cl_entry, const char *db_elem_name)
{
  if (NULL == cl_entry || NULL == db_elem_name)
    return dbeNullPointer;

  /* Free existing */
  if (NULL != cl_entry->db_elem_name)
    free(cl_entry->db_elem_name);

  /* Duplicate name */
  cl_entry->db_elem_name = strdup(db_elem_name);
  return (NULL == cl_entry->db_elem_name) ? dbeMemAllocFail : dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_menuelemname_set(struct s_db_changelog_entry *cl_entry, const char *menu_elem_name)
{
  if (NULL == cl_entry || NULL == menu_elem_name)
    return dbeNullPointer;

  /* Free existing */
  if (NULL != cl_entry->menu_elem_name)
    free(cl_entry->menu_elem_name);

  /* Duplicate name */
  cl_entry->menu_elem_name = strdup(menu_elem_name);
  return (NULL == cl_entry->menu_elem_name) ? dbeMemAllocFail : dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_index_set(struct s_db_changelog_entry *cl_entry, const char *index)
{
  unsigned long l;
  enum eDBError res;

  if (NULL == cl_entry || NULL == index)
  {
    res = dbeNullPointer;
  }
  else if (0 == strcmp("", index))
  {
    /* No index */
    cl_entry->has_index = false;
    cl_entry->index = 0;
    res = dbeSuccess;
  }
  else
  {
    /* Try to parse it */
    res = (1 == sscanf(index, "%lu", &l)) ? dbeSuccess : dbeParseFail;
    if (dbeSuccess == res)
    {
      cl_entry->has_index = true;
      cl_entry->index = l;
    }
    else
    {
      cl_entry->has_index = false;
      cl_entry->index = 0;
    }
  }
  return res;
}


static enum eDBError db_changelog_reader_entry_oldvalue_set(struct s_db_changelog_entry *cl_entry, const char *old_value)
{
  if (NULL == cl_entry || NULL == old_value)
    return dbeNullPointer;

  /* Free existing */
  if (NULL != cl_entry->old_value)
    free(cl_entry->old_value);

  /* Duplicate name */
  cl_entry->old_value = strdup(old_value);
  return (NULL == cl_entry->old_value) ? dbeMemAllocFail : dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_newvalue_set(struct s_db_changelog_entry *cl_entry, const char *new_value)
{
  if (NULL == cl_entry || NULL == new_value)
    return dbeNullPointer;

  /* Free existing */
  if (NULL != cl_entry->new_value)
    free(cl_entry->new_value);

  /* Duplicate name */
  cl_entry->new_value = strdup(new_value);
  return (NULL == cl_entry->new_value) ? dbeMemAllocFail : dbeSuccess;
}


static enum eDBError db_changelog_reader_entry_logmask_set(struct s_db_changelog_entry *cl_entry, const char *log_mask)
{
  if (NULL == cl_entry || NULL == log_mask)
    return dbeNullPointer;

  /* Free existing */
  if (NULL != cl_entry->log_mask)
    free(cl_entry->log_mask);

  /* Duplicate name */
  cl_entry->log_mask = strdup(log_mask);
  return (NULL == cl_entry->log_mask) ? dbeMemAllocFail : dbeSuccess;
}


/** Seek back to the start of the previous entry */
static enum eDBError db_changelog_reader_seekback(struct s_db_changelog_reader *clr)
{
  int ch;
  long posn;

  if (NULL == clr || NULL == clr->file)
    return dbeNullPointer;

  /* Are we already at the start of the file? */
  posn = ftell(clr->file);
  if (posn < 0)
  {
    return dbeFileFail;
  }
  else if (0 == posn)
  {
    return dbeChangelogNoMoreRecords;
  }

  /* Somewhere other than the start of the file, so seek back to start of next record */

  /*
   * If things are working correctly, we are positioned just after the \n of record last read,
   * so we need to skip that \n, then seek back until we find the next one
   */
  if (fseek(clr->file, -1, SEEK_CUR) < 0)
  {
    return dbeFileFail;
  }
  /* Read it */
  ch = fgetc(clr->file);
  if (EOF == ch)
  {
    return dbeFileFail;
  }
  if ('\n' == ch)
  {
    /* It was \n, so we need to skip it */
    if (fseek(clr->file, -1, SEEK_CUR) < 0)
    {
      return dbeFileFail;
    }
  }

  /* Now keep working back until we reach the next \n */
  for (ch = '\0'; ; )
  {
    /* Keep a look out for the start of the file */
    posn = ftell(clr->file);
    if (posn < 0)
    {
      return dbeFileFail;
    }
    else if (0 == posn)
    {
      /* At start of file, this is also the start of a record */
      return dbeSuccess;
    }

    /* Back 1 character */
    if (fseek(clr->file, -1, SEEK_CUR) < 0)
    {
      return dbeFileFail;
    }

    /* Read it */
    ch = fgetc(clr->file);
    if (EOF == ch)
    {
      return dbeFileFail;
    }
    if ('\n' == ch)
    {
      break;
    }

    /* And back again 1 character */
    if (fseek(clr->file, -1, SEEK_CUR) < 0)
    {
      return dbeFileFail;
    }
  }

  return dbeSuccess;
}


/** Read an entry from the file */
static enum eDBError db_changelog_reader_entry_getline(struct s_db_changelog_reader *clr)
{
  enum eDBError res;

  if (NULL == clr || NULL == clr->csv_line || NULL == clr->file)
    return dbeNullPointer;

  /* Read a line from the file */
  if (NULL == fgets(clr->csv_line, clr->max_csvlinesize, clr->file))
  {
    if (ferror(clr->file))
    {
      res = dbeFileFail;
    }
    else
    {
      res = dbeChangelogNoMoreRecords;
    }
  }
  else
  {
    res = dbeSuccess;
  }
  return res;
}


static void csv_field_callback(void *buf, size_t length, void *user_data)
{
  struct s_db_changelog_reader *clr = (struct s_db_changelog_reader *) user_data;
  const char *str = buf;
  enum eDBError res;

  if (NULL == clr)
  {
    clr->csvparse_last_result = dbeNullPointer;
    return;
  }

  /* Decode it */
  switch (clr->current_elem)
  {
    case dbcleCalibrationCounter:
      res = db_changelog_reader_entry_calcounter_set(clr->cl_entry, str);
      break;

    case dbcleConfigurationCounter:
      res = db_changelog_reader_entry_configcounter_set(clr->cl_entry, str);
      break;

    case dbcleTimestamp:
      res = db_changelog_reader_entry_time_set(clr->cl_entry, str);
      break;

    case dbcleDatabaseName:
      res = db_changelog_reader_entry_dbname_set(clr->cl_entry, str);
      break;

    case dbcleElementDatabaseName:
      res = db_changelog_reader_entry_dbelemname_set(clr->cl_entry, str);
      break;

    case dbcleElementMenuName:
      res = db_changelog_reader_entry_menuelemname_set(clr->cl_entry, str);
      break;

    case dbcleIndex:
      res = db_changelog_reader_entry_index_set(clr->cl_entry, str);
      break;

    case dbcleOldValue:
      res = db_changelog_reader_entry_oldvalue_set(clr->cl_entry, str);
      break;

    case dbcleNewValue:
      res = db_changelog_reader_entry_newvalue_set(clr->cl_entry, str);
      break;

    case dbcleLogMask:
      res = db_changelog_reader_entry_logmask_set(clr->cl_entry, str);
      break;

    default:
      /* Extra fields, ignore */
      res = dbeSuccess;
      break;

  }

  /* Always for first element, or if a element fails, set the result */
  if (0 == clr->current_elem || dbeSuccess != res)
    clr->csvparse_last_result = res;

  /* Next element */
  clr->current_elem++;
}


/** Read an entry from the file, starting at the current file position */
static enum eDBError db_changelog_reader_entry_get_current(struct s_db_changelog_reader *clr, struct s_db_changelog_entry *cl_entry)
{
  enum eDBError res;
  size_t length;

  /* Read a single line from the file, into clr->csv_line */
  res = db_changelog_reader_entry_getline(clr);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Setup for parser */
  clr->current_elem = 0;
  clr->cl_entry = cl_entry;
  clr->csvparse_last_result = dbeParseFail;

  /* Parse it as CSV */
  length = strlen(clr->csv_line);
  if (csv_parse(&clr->csvparse, clr->csv_line, length, csv_field_callback, NULL, clr) != length)
  {
    switch (csv_error(&clr->csvparse))
    {
      case CSV_EPARSE:
        res = dbeParseFail;
        break;
      case CSV_ENOMEM:
        res = dbeMemAllocFail;
        break;
      case CSV_ETOOBIG:
        res = dbeOutsideLimits;
        break;
      default:
      case CSV_EINVALID:
        res = dbeFailure;
        break;
    /* #define CSV_SUCCESS 0 */
    }
  }
  else
  {
    res = clr->csvparse_last_result;
  }
  return res;
}


/** Read an entry from the file */
enum eDBError db_changelog_reader_entry_get(struct s_db_changelog_reader *clr, enum eDBChangeLogReaderLocator locator, struct s_db_changelog_entry *entry)
{
  enum eDBError res;

  if (NULL == clr || NULL == entry || NULL == clr->file)
    return dbeNullPointer;

  /* Refresh buffers in case file has changed */
  if (0 != fflush(clr->file))
  {
    return dbeFileFail;
  }

  /* Get appropriate record */
  switch (locator)
  {
    /* Seek start of file, skip header row, and read entry */
    case dbclrlFirst:
      res = (fseek(clr->file, 0, SEEK_SET) < 0) ? dbeFileFail : dbeSuccess;
      if (dbeSuccess == res)
      {
        res = db_changelog_reader_entry_getline(clr);
      }
      break;

    /* Seek end of file, and work back to start of last record, then read entry */
    case dbclrlLast:
      res = (fseek(clr->file, 0, SEEK_END) < 0) ? dbeFileFail : dbeSuccess;
      if (dbeSuccess == res)
        res = db_changelog_reader_seekback(clr);
      break;

    /* Just read from current position */
    case dbclrlNext:
      res = dbeSuccess;
      break;

    /*
     * Work back from current position to start of record, then read entry.
     * This needs to be done twice because 1 seek back will take us to the
     * start of the last record read
     */
    case dbclrlPrevious:
      res = db_changelog_reader_seekback(clr);
      if (dbeSuccess == res)
        res = db_changelog_reader_seekback(clr);

      if (dbeSuccess == res)
      {
        /*
         * If we are at the start of the file
         * skip over the header row, and return
         * dbeChangelogNoMoreRecords
         */
        if (ftell(clr->file) == 0)
        {
          res = db_changelog_reader_entry_getline(clr);
          if (dbeSuccess == res)
            res = dbeChangelogNoMoreRecords;
        }
      }
      break;

    default:
      res = dbeFailure;
      break;
  }

  /* Now get the entry */
  if (dbeSuccess == res)
    res = db_changelog_reader_entry_get_current(clr, entry);

  return res;
}


/** Create a change log reader from a FILE */
static enum eDBError db_changelog_reader_internal_create(struct s_db_changelog_reader **clr, FILE *file, size_t max_csvlinesize, bool ourfile)
{
  struct s_db_changelog_reader *cr = NULL;
  enum eDBError res;

  if (NULL == clr || NULL == file)
    return dbeNullPointer;

  /* Allocate reader structure */
  cr = malloc(sizeof(struct s_db_changelog_reader));
  if (NULL == cr)
  {
    res = dbeMemAllocFail;
    goto dcrc_fail;
  }

  /* Fill out members */
  memset(cr, 0, sizeof(struct s_db_changelog_reader));
  cr->current_elem = 0;
  cr->file = file;
  cr->ourfile = ourfile;
  if (csv_init(&cr->csvparse, CSV_APPEND_NULL) < 0)
  {
    res = dbeFailure;
    goto dcrc_fail;
  }
  cr->max_csvlinesize = max_csvlinesize;
  cr->csv_line = malloc(cr->max_csvlinesize);
  if (NULL == cr->csv_line)
  {
    res = dbeMemAllocFail;
    goto dcrc_fail;
  }

  /* Success */
  *clr = cr;
  return dbeSuccess;

dcrc_fail:
  *clr = NULL;
  if (NULL != cr)
  {
    csv_free(&cr->csvparse);
    if (NULL != cr->csv_line)
    {
      free(cr->csv_line);
      cr->csv_line = NULL;
    }
    free(cr);
  }
  return res;
}


/** Create a change log reader from a FILE. Caller is responsible for closing file after db_changelog_reader_destroy() call. */
enum eDBError db_changelog_reader_create_file(struct s_db_changelog_reader **clr, FILE *file, size_t max_csvlinesize)
{
  return db_changelog_reader_internal_create(clr, file, max_csvlinesize, false);
}


/** Create a change log reader from a filename (file is opened read only). db_changelog_reader_destroy() will close the file. */
enum eDBError db_changelog_reader_create_filename(struct s_db_changelog_reader **clr, const char *filename, size_t max_csvlinesize)
{
  enum eDBError res;
  FILE *file;

  if (NULL == clr || NULL == filename)
    return dbeNullPointer;

  /* Open file for reading */
  file = fopen(filename, "r");
  if (NULL == file)
  {
    res = dbeFileFail;
    goto dcrcf_fail;
  }
  if (fcntl(fileno(file), F_SETFD, FD_CLOEXEC) < 0)
  {
    res = dbeFileFail;
    goto dcrcf_fail;
  }

  /* Create change log reader */
  res = db_changelog_reader_internal_create(clr, file, max_csvlinesize, true);
  if (dbeSuccess != res)
  {
    goto dcrcf_fail;
  }

  /* Success */
  return dbeSuccess;

dcrcf_fail:
  if (NULL != file)
  {
    fclose(file);
    file = NULL;
  }
  return res;
}


/** Destroy a change log reader */
enum eDBError db_changelog_reader_destroy(struct s_db_changelog_reader *clr)
{
  if (NULL != clr)
  {
    if (clr->ourfile)
    {
      fclose(clr->file);
      clr->file = NULL;
    }
    csv_free(&clr->csvparse);
    if (NULL != clr->csv_line)
    {
      free(clr->csv_line);
      clr->csv_line = NULL;
    }
    free(clr);
  }
  return dbeSuccess;
}
