/*
 * db_database.c
 *
 *  Created on: 08/07/2011
 *      Author: ian
 */
#include <stdlib.h>
#include <rindb/db_error.h>
#include <rindb/db_buffer.h>
#include <rindb/db_interp_bin.h>
#include <rindb/db_crc.h>

#include <rindb/db_database.h>


/** Write the data for a section to the interpreter */
static enum eDBError db_database_write_element_data(const struct s_db_database *database, struct s_db_interp *interp, uint32_t at_version)
{
  size_t i;
  enum eDBError res;

  /* Write each element of the database out */
  for (i = 0; i < database->db_num_elements; i++)
  {
    res = db_element_write(database->db_name, &database->db_elements[i], interp, at_version);
    if (dbeSuccess != res)
    {
      break;
    }
  }
  return res;
}


/** Calculate the checksum for a database */
static enum eDBError db_database_calculate_checksum(const struct s_db_database *database, crc_t *crc, uint32_t at_version)
{
  struct s_db_interp interp_bin;
  enum eDBError res;

  /* Create a binary interpretation of the database, then checksum that */

  /* Back to start of checksum buffer */
  res = db_buffer_rewind(database->db_checksum_buffer);
  if (dbeSuccess != res)
    goto dscc_done;

  /* Now create a binary interpreter */
  res = db_create_interp_binary(&interp_bin, database->db_checksum_buffer);
  if (dbeSuccess != res)
    goto dscc_done;

  /* Now save the data in this section to buffer via the the binary interpreter */
  res = db_database_write_element_data(database, &interp_bin, at_version);
  if (dbeSuccess != res)
    goto dscc_done;

  /* Now calculate a checksum over the binary data in the buffer */
  *crc = db_crc32(database->db_checksum_buffer->buffer, database->db_checksum_buffer->posn, database->db_database_chksum_seed);
  res = dbeSuccess;

dscc_done:
  return res;
}


/** Write an entire database */
enum eDBError db_database_write(const struct s_db_database *database, struct s_db_database_interp *ddi, uint32_t at_version)
{
  enum eDBError res;
  crc_t crc;

  /* First calculate the checksum of the data */
  res = db_database_calculate_checksum(database, &crc, at_version);
  if (dbeSuccess == res)
  {
    /* Begin export */
    res = ddi->database_export_start_fn(ddi, database, at_version, crc);
    if (dbeSuccess == res)
    {
      /* Export all elements */
      res = db_database_write_element_data(database, &ddi->interp, at_version);
      if (dbeSuccess == res)
      {
        /* Finish up export */
        res = ddi->database_export_end_fn(ddi, database, at_version, crc);
      }
    }
  }

  return res;
}


/** Read an entire database */
enum eDBError db_database_read(struct s_db_database *database, struct s_db_database_interp *ddi, uint32_t *version, bool ignore_checksum)
{
  size_t i;
  enum eDBError res;
  crc_t read_crc;
  crc_t calc_crc;

  /* Begin import, read version and checksum */
  res = ddi->database_import_start_fn(ddi, database, version, &read_crc);
  if (dbeSuccess == res)
  {
    /* Import all elements */
    for (i = 0; i < database->db_num_elements; i++)
    {
      res = db_element_read(database->db_name, &database->db_elements[i], &ddi->interp, *version);
      if (dbeSuccess != res)
      {
        break;
      }
    }
    if (dbeSuccess == res)
    {
      /* Finish up import (version & checksum) */
      res = ddi->database_import_end_fn(ddi, database, version, &read_crc);
      if (dbeSuccess == res && !ignore_checksum)
      {
        /* Check checksum if required, calculate what it should be */
        res = db_database_calculate_checksum(database, &calc_crc, *version);
        if (dbeSuccess == res)
        {
          /* Verify checksum */
          res = (calc_crc == read_crc) ? dbeSuccess : dbeChecksumFail;
        }
      }
    }
  }

  return res;
}


/** Set a database to defaults */
enum eDBError db_database_default(struct s_db_database *database)
{
  size_t i;
  enum eDBError res;

  /* Default all elements */
  for (i = 0, res = dbeSuccess; i < database->db_num_elements; i++)
  {
    res = db_element_default(&database->db_elements[i]);
    if (dbeSuccess != res)
    {
      break;
    }
  }

  return res;
}


/** Update the previous versions for all elements in all sections of the database */
enum eDBError db_database_update_previous(struct s_db_database *database)
{
  size_t i;
  enum eDBError res;

  /* Update previous values for each element */
  for (i = 0, res = dbeSuccess; i < database->db_num_elements; i++)
  {
    res = db_element_update_previous(&database->db_elements[i]);
    if (dbeSuccess != res)
    {
      break;
    }
  }

  return res;
}


/** Revert a database to the values stored in each element's previous value */
enum eDBError db_database_revert(struct s_db_database *database)
{
  size_t i;
  enum eDBError res;

  /* Revert to previous values for each element */
  for (i = 0, res = dbeSuccess; i < database->db_num_elements; i++)
  {
    res = db_element_revert(&database->db_elements[i]);
    if (dbeSuccess != res)
    {
      break;
    }
  }

  return res;
}


/** Initialise all elements in the database */
enum eDBError db_database_init(struct s_db_database *database)
{
  size_t i;
  enum eDBError res;
  size_t elem_max_size;
  size_t db_max_size;

  /* Initialise all elements */
  for (i = 0, res = dbeSuccess, db_max_size = 0; i < database->db_num_elements; i++)
  {
    /* Initialise the element */
    res = db_element_init(&database->db_elements[i]);
    if (dbeSuccess != res)
      goto ddi_fail;

    /* Get the maximum size of the element, so we can calculate the size of the buffer needed for calculating the checksum */
    res = db_element_get_maximum_size(&database->db_elements[i], &elem_max_size);
    if (dbeSuccess != res)
      goto ddi_fail;
    db_max_size += elem_max_size;
  }

  /* Create buffer for calculating checksum */
  database->db_checksum_buffer = calloc(1, sizeof(struct s_db_buffer));
  if (NULL == database->db_checksum_buffer)
  {
    res = dbeMemAllocFail;
    goto ddi_fail;
  }
  res = db_buffer_create_malloc(database->db_checksum_buffer, db_max_size, db_max_size);
  if (dbeSuccess != res)
    goto ddi_fail;

  /* Success */
  return res;

ddi_fail:
  (void) db_database_destroy(database);
  return res;
}


/** Destroy all elements in the database */
enum eDBError db_database_destroy(struct s_db_database *database)
{
  size_t i;
  enum eDBError res;

  if (NULL == database)
    return dbeNullPointer;

  /* Destroy checksum buffer */
  if (NULL != database->db_checksum_buffer)
  {
    (void) db_buffer_destroy(database->db_checksum_buffer);
    free(database->db_checksum_buffer);
    database->db_checksum_buffer = NULL;
  }

  /* Destroy all sections */
  for (i = 0, res = dbeSuccess; i < database->db_num_elements; i++)
  {
    (void) db_element_destroy(&database->db_elements[i]);
  }

  return res;
}
