/*
 * db_database_binbuf.c
 *
 *  Created on: Apr 19, 2012
 *      Author: ian
 */
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <ezxml/ezxml.h>
#include <rindb/db_database.h>
#include <rindb/db_error.h>
#include <rindb/db_interp_bin.h>

#include <rindb/db_database_binbuf.h>


struct s_db_database_binbuffer
{
  struct s_db_buffer *db_buffer;
};


#define DB_DATABASE_VERSION_STR "version"
#define DB_DATABASE_CHECKSUM_STR "checksum"


/** Export database start function for binary */
static enum eDBError db_database_binbuffer_export_start(struct s_db_database_interp *ddi, const struct s_db_database *db, uint32_t at_version, uint32_t checksum)
{
  enum eDBError res;
  struct s_db_database_binbuffer *db_binbuf = (struct s_db_database_binbuffer *) ddi->database_data;
  struct s_db_interp *interp;
  struct s_db_element_uint32 elem;

  /* Position at start of buffer */
  db_binbuf->db_buffer->posn = 0;

  /* Setup interpreter */
  interp = &ddi->interp;
  interp->interpreter = db_binbuf->db_buffer;

  /* Write version to buffer */
  memset(&elem, 0, sizeof(elem));
  elem.data = &at_version;
  elem.default_value = at_version;
  elem.minimum = DB_NO_MINIMUM_VERSION;
  elem.maximum = DB_NO_MAXIMUM_VERSION;
  elem.previous_data = 0;
  res = db_interp_uint32_export(db->db_name, interp, DB_DATABASE_VERSION_STR, &elem);

  /* Write checksum to buffer */
  if (dbeSuccess == res)
  {
    memset(&elem, 0, sizeof(elem));
    elem.data = &checksum;
    elem.default_value = checksum;
    elem.minimum = 0;
    elem.maximum = UINT32_MAX;
    elem.previous_data = 0;
    res = db_interp_uint32_export(db->db_name, interp, DB_DATABASE_CHECKSUM_STR, &elem);
  }

  return res;
}


/** Export database end function for binary */
static enum eDBError db_database_binbuffer_export_end(struct s_db_database_interp *ddi, const struct s_db_database *db, uint32_t at_version, uint32_t version)
{
  /* Nothing to do here */
  return dbeSuccess;
}


/** Import database start function for binary */
static enum eDBError db_database_binbuffer_import_start(struct s_db_database_interp *ddi, struct s_db_database *db, uint32_t *version, uint32_t *checksum)
{
  enum eDBError res;
  struct s_db_database_binbuffer *db_binbuf = (struct s_db_database_binbuffer *) ddi->database_data;
  struct s_db_interp *interp;
  struct s_db_element_uint32 elem;

  /* Position at start of buffer (buffer must have been pre-filled) */
  db_binbuf->db_buffer->posn = 0;

  /* Setup interpreter */
  interp = &ddi->interp;
  interp->interpreter = db_binbuf->db_buffer;

  /* Read version from buffer */
  memset(&elem, 0, sizeof(elem));
  elem.data = version;
  elem.default_value = *version;
  elem.minimum = DB_NO_MINIMUM_VERSION;
  elem.maximum = DB_NO_MAXIMUM_VERSION;
  elem.previous_data = 0;
  res = db_interp_uint32_import(db->db_name, interp, DB_DATABASE_VERSION_STR, &elem);

  /* Read checksum from buffer */
  if (dbeSuccess == res)
  {
    memset(&elem, 0, sizeof(elem));
    elem.data = checksum;
    elem.default_value = *checksum;
    elem.minimum = 0;
    elem.maximum = UINT32_MAX;
    elem.previous_data = 0;
    res = db_interp_uint32_import(db->db_name, interp, DB_DATABASE_CHECKSUM_STR, &elem);
  }

  return res;
}


/** Import database end function for binary */
static enum eDBError db_database_binbuffer_import_end(struct s_db_database_interp *dsi, struct s_db_database *db, uint32_t *version, uint32_t *checksum)
{
  /* Nothing to do here */
  return dbeSuccess;
}


/** Create a binary database interpreter for a buffer */
enum eDBError db_create_database_interp_binbuffer(struct s_db_database_interp *ddi, struct s_db_buffer *buffer)
{
  enum eDBError res;
  struct s_db_database_binbuffer *db_binbuf = NULL;

  /* Check pointers */
  if (NULL == ddi || NULL == buffer)
  {
    res = dbeNullPointer;
    goto dcdib_fail;
  }

  /* Allocate structure */
  db_binbuf = malloc(sizeof(struct s_db_database_binbuffer));
  if (NULL == db_binbuf)
  {
    res = dbeMemAllocFail;
    goto dcdib_fail;
  }

  /* Fill out members */
  db_binbuf->db_buffer = buffer;

  /* Keep db_binbuf for future operations */
  ddi->database_data = db_binbuf;

  /* Create interpreter */
  db_create_interp_binary(&ddi->interp, db_binbuf->db_buffer);

  /* Setup import and export functions */
  ddi->database_export_start_fn = db_database_binbuffer_export_start;
  ddi->database_export_end_fn = db_database_binbuffer_export_end;
  ddi->database_import_start_fn = db_database_binbuffer_import_start;
  ddi->database_import_end_fn = db_database_binbuffer_import_end;

  /* Success */
  return dbeSuccess;

dcdib_fail:
  ddi->database_data = NULL;
  if (NULL != db_binbuf)
  {
    db_binbuf->db_buffer = NULL;
    free(db_binbuf);
  }
  return res;
}


/** Destroy a binary buffer database interpreter. It's up to you to destroy the buffer passed to db_create_database_interp_binbuffer() */
enum eDBError db_destroy_database_interp_binbuffer(struct s_db_database_interp *ddi)
{
  struct s_db_database_binbuffer *db_binbuf;

  /* Check pointers */
  if (NULL == ddi || NULL == ddi->database_data)
    return dbeNullPointer;
  db_binbuf = (struct s_db_database_binbuffer *) ddi->database_data;

  /* Free the db_bin structure */
  ddi->database_data = NULL;
  if (NULL != db_binbuf)
  {
    /* This was passed to db_create_database_interp_binbuffer(), so it's up to the caller to destroy it */
    db_binbuf->db_buffer = NULL;
    free(db_binbuf);
  }

  return dbeSuccess;
}
