/*
 * db_database_bin.c
 *
 *  Created on: 16/08/2011
 *      Author: ian
 */
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <ezxml/ezxml.h>
#include <rindb/db_database.h>
#include <rindb/db_error.h>
#include <rindb/db_interp_bin.h>

#include <rindb/db_database_binfile.h>


struct s_db_database_bin
{
  char *filename;
  FILE *filestream;
  struct s_db_buffer db_buffer;
};


/** Buffer memory allocation increment size */
#define DB_BINARY_BUFFER_ALLOC_INCREMENT 1024


#define DB_DATABASE_VERSION_STR "version"
#define DB_DATABASE_CHECKSUM_STR "checksum"


/** Open and parse the binary file, if not already open/parsed. db_xml->filename should be set before calling */
static enum eDBError db_database_open_bin_file(struct s_db_database_bin *db_bin)
{
  enum eDBError res = dbeSuccess;

  /* Check if file open yet */
  if (NULL == db_bin->filestream)
  {
    /* File not open yet, open for reading and writing, create if doesn't exist */
    db_bin->filestream = fopen(db_bin->filename, "a+");
    if (NULL == db_bin->filestream)
    {
      return dbeFileFail;
    }
    if (fcntl(fileno(db_bin->filestream), F_SETFD, FD_CLOEXEC) < 0)
    {
      return dbeFileFail;
    }
  }

  /* Load into buffer */
  if (NULL == db_bin->db_buffer.buffer)   /* TODO better way to detect */
  {
    /* Read file into buffer */
    res = db_buffer_create_file(&db_bin->db_buffer, db_bin->filestream, DB_BINARY_BUFFER_ALLOC_INCREMENT);
  }
  else
  {
    if (0 != fseek(db_bin->filestream, 0, SEEK_SET))
    {
      return dbeFileFail;
    }
    res = db_buffer_load_file(&db_bin->db_buffer, db_bin->filestream);
  }
  return res;
}


/** Write the current binary buffer to it's file */
static enum eDBError db_database_write_bin_file(struct s_db_database_bin *db_bin)
{
  size_t length;

  /* Truncate file (to 0 bytes) */
  if (ftruncate(fileno(db_bin->filestream), 0) != 0)
  {
    goto ddwbf_fail;
  }

  /* Back to start of file */
  if (0 != fseek(db_bin->filestream, 0, SEEK_SET))
  {
    goto ddwbf_fail;
  }

  /* Work out length */
  length = db_bin->db_buffer.posn;

  /* Write out data */
  if (length != fwrite(db_bin->db_buffer.buffer, sizeof(char), length, db_bin->filestream))
  {
    goto ddwbf_fail;
  }

  /* Force to be written out */
  if (0 != fflush(db_bin->filestream))
  {
    goto ddwbf_fail;
  }

  /* Success */
  return dbeSuccess;

ddwbf_fail:
  return dbeFileFail;
}


/** Close files and free memory */
static enum eDBError db_database_bin_cleanup(struct s_db_database_bin *db_bin)
{
  /* Close file */
  if (NULL != db_bin->filestream)
  {
    fclose(db_bin->filestream);
  }
  db_bin->filestream = NULL;

  /* Free up buffer */
  return db_buffer_destroy(&db_bin->db_buffer);
}


/** Export database start function for binary */
static enum eDBError db_database_bin_export_start(struct s_db_database_interp *ddi, const struct s_db_database *db, uint32_t at_version, uint32_t checksum)
{
  enum eDBError res;
  struct s_db_database_bin *db_bin = (struct s_db_database_bin *) ddi->database_data;
  struct s_db_interp *interp;
  struct s_db_element_uint32 elem;

  /* Open file and load binary file, if not already done so */
  res = db_database_open_bin_file(db_bin);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Position at start of buffer */
  db_bin->db_buffer.posn = 0;

  /* Setup interpreter */
  interp = &ddi->interp;
  interp->interpreter = &db_bin->db_buffer;

  /* Write version to buffer */
  memset(&elem, 0, sizeof(elem));
  elem.data = &at_version;
  elem.default_value = at_version;
  elem.minimum = DB_NO_MINIMUM_VERSION;
  elem.maximum = DB_NO_MAXIMUM_VERSION;
  elem.previous_data = 0;
  res = db_interp_uint32_export(db->db_name, interp, DB_DATABASE_VERSION_STR, &elem);

  /* Write checksum to buffer */
  if (dbeSuccess == res)
  {
    memset(&elem, 0, sizeof(elem));
    elem.data = &checksum;
    elem.default_value = checksum;
    elem.minimum = 0;
    elem.maximum = UINT32_MAX;
    elem.previous_data = 0;
    res = db_interp_uint32_export(db->db_name, interp, DB_DATABASE_CHECKSUM_STR, &elem);
  }

  return res;
}


/** Export database end function for binary */
static enum eDBError db_database_bin_export_end(struct s_db_database_interp *ddi, const struct s_db_database *db, uint32_t at_version, uint32_t version)
{
  /* Write the database out to binary file */
  return db_database_write_bin_file(ddi->database_data);
}


/** Import database start function for binary */
static enum eDBError db_database_bin_import_start(struct s_db_database_interp *ddi, struct s_db_database *db, uint32_t *version, uint32_t *checksum)
{
  enum eDBError res;
  struct s_db_database_bin *db_bin = (struct s_db_database_bin *) ddi->database_data;
  struct s_db_interp *interp;
  struct s_db_element_uint32 elem;

  /* Open file and load binary file, if not already done so */
  res = db_database_open_bin_file(db_bin);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Position at start of buffer */
  db_bin->db_buffer.posn = 0;

  /* Setup interpreter */
  interp = &ddi->interp;
  interp->interpreter = &db_bin->db_buffer;

  /* Read version from buffer */
  memset(&elem, 0, sizeof(elem));
  elem.data = version;
  elem.default_value = *version;
  elem.minimum = DB_NO_MINIMUM_VERSION;
  elem.maximum = DB_NO_MAXIMUM_VERSION;
  elem.previous_data = 0;
  res = db_interp_uint32_import(db->db_name, interp, DB_DATABASE_VERSION_STR, &elem);

  /* Read checksum from buffer */
  if (dbeSuccess == res)
  {
    memset(&elem, 0, sizeof(elem));
    elem.data = checksum;
    elem.default_value = *checksum;
    elem.minimum = 0;
    elem.maximum = UINT32_MAX;
    elem.previous_data = 0;
    res = db_interp_uint32_import(db->db_name, interp, DB_DATABASE_CHECKSUM_STR, &elem);
  }

  return res;
}


/** Import database end function for binary */
static enum eDBError db_database_bin_import_end(struct s_db_database_interp *dsi, struct s_db_database *db, uint32_t *version, uint32_t *checksum)
{
  /* Nothing to do here */
  return dbeSuccess;
}


/** Create a binary database interpreter */
enum eDBError db_create_database_interp_bin(struct s_db_database_interp *ddi, const char *filename)
{
  enum eDBError res;
  struct s_db_database_bin *db_bin;

  /* Allocate structure */
  db_bin = malloc(sizeof(struct s_db_database_bin));
  if (NULL == db_bin)
  {
    res = dbeMemAllocFail;
    goto dcdib_fail;
  }

  /* Allocate filename string */
  db_bin->filename = strdup(filename);
  if (NULL == db_bin->filename)
  {
    res = dbeMemAllocFail;
    goto dcdib_free_bin;
  }

  /* Setup elements */
  db_bin->filestream = NULL;

  /* Create a buffer to start with, this will get re-sized as required when data is loaded from file, or written */
  res = db_buffer_create_malloc(&db_bin->db_buffer, DB_BINARY_BUFFER_ALLOC_INCREMENT, DB_BINARY_BUFFER_ALLOC_INCREMENT);
  if (dbeSuccess != res)
  {
    goto dcdib_free_file;
  }

  /* Keep db_xml for future operations */
  ddi->database_data = db_bin;

  /* Create interpreter */
  db_create_interp_binary(&ddi->interp, &db_bin->db_buffer);

  /* Setup import and export functions */
  ddi->database_export_start_fn = db_database_bin_export_start;
  ddi->database_export_end_fn = db_database_bin_export_end;
  ddi->database_import_start_fn = db_database_bin_import_start;
  ddi->database_import_end_fn = db_database_bin_import_end;

  /* Success */
  return dbeSuccess;

dcdib_free_file:
  free(db_bin->filename);
  db_bin->filename = NULL;

dcdib_free_bin:
  free(db_bin);
  ddi->database_data = NULL;

dcdib_fail:
  return res;
}


/** Destroy a binary database interpreter */
enum eDBError db_destroy_database_interp_bin(struct s_db_database_interp *ddi)
{
  struct s_db_database_bin *db_bin = (struct s_db_database_bin *) ddi->database_data;

  /* Free things within the db_bin structure */
  (void) db_database_bin_cleanup(db_bin);

  /* Free the filename string */
  free(db_bin->filename);
  db_bin->filename = NULL;

  /* Free the db_bin structure */
  ddi->database_data = NULL;
  free(db_bin);

  return dbeSuccess;
}
