/*
 * db_database_uboot.c
 *
 *  Created on: Sep 13, 2012
 *      Author: ian
 */
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <ezxml/ezxml.h>
#include <rindb/db_database.h>
#include <rindb/db_error.h>
#include <rindb/db_interp_uboot.h>

#include <rindb/db_database_uboot.h>


struct s_db_database_uboot
{
  struct s_db_buffer *db_buffer;
};


#define DB_DATABASE_HEADER_STR "_info"
#define DB_UBOOT_VERSION_IDX 0
#define DB_UBOOT_CHECKSUM_IDX 1


static enum eDBError db_database_uboot_allocate_namestr(char **name, const char *db_name)
{
  size_t namestr_len;

  /* Setup name variable <dbname>_info */
  namestr_len = strlen(db_name) + strlen(DB_DATABASE_HEADER_STR) + 1;
  *name = malloc(namestr_len);
  if (NULL == *name)
  {
    return dbeMemAllocFail;
  }
  snprintf(*name, namestr_len, "%s%s", db_name, DB_DATABASE_HEADER_STR);
  return dbeSuccess;
}


/** Export database start function for u-boot environment variable database */
static enum eDBError db_database_uboot_export_start(struct s_db_database_interp *ddi, const struct s_db_database *db, uint32_t at_version, uint32_t checksum)
{
  enum eDBError res;
  struct s_db_database_uboot *db_uboot = (struct s_db_database_uboot *) ddi->database_data;
  struct s_db_interp *interp;
  char *namestr = NULL;
  uint32_t vers_chksum_elem_values[2];
  uint32_t vers_chksum_prev_values[2];
  uint32_t elem_mins[2] = { DB_NO_MINIMUM_VERSION, 0 };
  uint32_t elem_maxs[2] = { DB_NO_MAXIMUM_VERSION, UINT_MAX };
  uint32_t elem_defaults[2] = { 0, 0 };
  struct s_db_element_uint32_array vers_chksum_elem =
  {
    .data = vers_chksum_elem_values,
    .num_elements = 2,
    .minimum = elem_mins,
    .maximum = elem_maxs,
    .default_value = elem_defaults,
    .single_minimum = 0,
    .single_maximum = 0,
    .single_default = 0,
    .previous_data = vers_chksum_prev_values
  };

  /* Position at start of buffer */
  res = db_buffer_rewind(db_uboot->db_buffer);
  if (dbeSuccess != res)
  {
    goto ddues_fail;
  }

  /* Setup interpreter */
  interp = &ddi->interp;
  interp->interpreter = db_uboot->db_buffer;

  /* Setup name variable <dbname>_info */
  res = db_database_uboot_allocate_namestr(&namestr, db->db_name);
  if (dbeSuccess != res)
  {
    goto ddues_fail;
  }

  /* Setup array containing version and checksum */
  vers_chksum_elem_values[DB_UBOOT_VERSION_IDX] = at_version;
  vers_chksum_elem_values[DB_UBOOT_CHECKSUM_IDX] = checksum;

  /* Write array containing version and checksum to buffer */
  res = db_interp_uint32_array_export(db->db_name, interp, namestr, &vers_chksum_elem);

ddues_fail:
  /* Clean up */
  if (NULL != namestr)
    free(namestr);

  return res;
}


/** Export database end function for u-boot environment variable database */
static enum eDBError db_database_uboot_export_end(struct s_db_database_interp *ddi, const struct s_db_database *db, uint32_t at_version, uint32_t version)
{
  struct s_db_database_uboot *db_uboot = (struct s_db_database_uboot *) ddi->database_data;

  /* Null terminate the buffer */
  return db_buffer_addbyte(db_uboot->db_buffer, '\0');
}


/** Import database start function for u-boot environment variable database */
static enum eDBError db_database_uboot_import_start(struct s_db_database_interp *ddi, struct s_db_database *db, uint32_t *version, uint32_t *checksum)
{
  enum eDBError res;
  struct s_db_database_uboot *db_uboot = (struct s_db_database_uboot *) ddi->database_data;
  struct s_db_interp *interp;
  char *namestr = NULL;

  uint32_t vers_chksum_elem_values[2];
  uint32_t vers_chksum_prev_values[2];
  uint32_t elem_mins[2] = { DB_NO_MINIMUM_VERSION, 0 };
  uint32_t elem_maxs[2] = { DB_NO_MAXIMUM_VERSION, UINT_MAX };
  uint32_t elem_defaults[2] = { 0, 0 };
  struct s_db_element_uint32_array vers_chksum_elem =
  {
    .data = vers_chksum_elem_values,
    .num_elements = 2,
    .minimum = elem_mins,
    .maximum = elem_maxs,
    .default_value = elem_defaults,
    .single_minimum = 0,
    .single_maximum = 0,
    .single_default = 0,
    .previous_data = vers_chksum_prev_values
  };

  /* TODO Check for null termination at end of buffer, and add if not??? */

  /* Position at start of buffer (buffer must have been pre-filled) */
  res = db_buffer_rewind(db_uboot->db_buffer);
  if (dbeSuccess != res)
  {
    goto dduis_fail;
  }

  /* Setup name variable <dbname>_info */
  res = db_database_uboot_allocate_namestr(&namestr, db->db_name);
  if (dbeSuccess != res)
  {
    goto dduis_fail;
  }

  /* Setup interpreter */
  interp = &ddi->interp;
  interp->interpreter = db_uboot->db_buffer;

  /* Read version and checksum array */
  res = db_interp_uint32_array_import(db->db_name, interp, namestr, &vers_chksum_elem);
  if (dbeSuccess != res)
  {
    goto dduis_fail;
  }

  /* Return array containing version and checksum */
  *version = vers_chksum_elem_values[DB_UBOOT_VERSION_IDX];
  *checksum = vers_chksum_elem_values[DB_UBOOT_CHECKSUM_IDX];

  res = dbeSuccess;

dduis_fail:
  /* Clean up */
  if (NULL != namestr)
    free(namestr);

  return res;
}


/** Import database end function for u-boot environment variable database */
static enum eDBError db_database_uboot_import_end(struct s_db_database_interp *dsi, struct s_db_database *db, uint32_t *version, uint32_t *checksum)
{
  /* Nothing to do here */
  return dbeSuccess;
}


/** Create a u-boot environment variable database interpreter for a buffer */
enum eDBError db_create_database_interp_uboot(struct s_db_database_interp *ddi, struct s_db_buffer *buffer)
{
  enum eDBError res;
  struct s_db_database_uboot *db_uboot = NULL;

  /* Check pointers */
  if (NULL == ddi || NULL == buffer)
  {
    res = dbeNullPointer;
    goto dcdib_fail;
  }

  /* Allocate structure */
  db_uboot = malloc(sizeof(struct s_db_database_uboot));
  if (NULL == db_uboot)
  {
    res = dbeMemAllocFail;
    goto dcdib_fail;
  }

  /* Fill out members */
  db_uboot->db_buffer = buffer;

  /* Keep db_binbuf for future operations */
  ddi->database_data = db_uboot;

  /* Create interpreter */
  db_create_interp_uboot(&ddi->interp, db_uboot->db_buffer);

  /* Setup import and export functions */
  ddi->database_export_start_fn = db_database_uboot_export_start;
  ddi->database_export_end_fn = db_database_uboot_export_end;
  ddi->database_import_start_fn = db_database_uboot_import_start;
  ddi->database_import_end_fn = db_database_uboot_import_end;

  /* Success */
  return dbeSuccess;

dcdib_fail:
  ddi->database_data = NULL;
  if (NULL != db_uboot)
  {
    db_uboot->db_buffer = NULL;
    free(db_uboot);
  }
  return res;
}


/** Destroy a u-boot environment variable database interpreter. It's up to you to destroy the buffer passed to db_create_database_interp_uboot() */
enum eDBError db_destroy_database_interp_uboot(struct s_db_database_interp *ddi)
{
  struct s_db_database_uboot *db_uboot;

  /* Check pointers */
  if (NULL == ddi || NULL == ddi->database_data)
    return dbeNullPointer;
  db_uboot = (struct s_db_database_uboot *) ddi->database_data;

  /* Free the db_bin structure */
  ddi->database_data = NULL;
  if (NULL != db_uboot)
  {
    /* This was passed to db_create_database_interp_uboot(), so it's up to the caller to destroy it */
    db_uboot->db_buffer = NULL;
    free(db_uboot);
  }

  return dbeSuccess;
}
