/*
 * db_database_xml.c
 *
 *  Created on: 16/08/2011
 *      Author: ian
 */
#include <string.h>
#include <unistd.h>
#include <ezxml/ezxml.h>
#include <rindb/db_database.h>
#include <rindb/db_error.h>
#include <rindb/db_interp_xml.h>

#include <rindb/db_database_xml.h>


struct s_db_database_xml
{
  char *filename;             /*!< XML file name */
  FILE *filestream;           /*!< File stream for XML file */
  ezxml_t xmlroot;            /*!< Root node in the XML */
  ezxml_t dbroot;             /*!< Database root node */
};


#define DB_DATABASE_ROOTNODENAME "rinstrum"
#define DB_DATABASE_NODENAME "database"
#define DB_DATABASE_ATTRNAME "name"
#define DB_DATABASE_ATTRVERSION "version"
#define DB_DATABASE_ATTRCHECKSUM "checksum"


/** Get a database with name */
static enum eDBError db_interp_xml_get_database(ezxml_t xmlroot, const char *name, ezxml_t *database)
{
  ezxml_t db_iter;
  const char *current_db_name;

  for (db_iter = ezxml_child(xmlroot, DB_DATABASE_NODENAME); ((ezxml_t) NULL != db_iter); db_iter = db_iter->next)
  {
    current_db_name = ezxml_attr(db_iter, DB_DATABASE_ATTRNAME);
    if (NULL != current_db_name && 0 == strcmp(current_db_name, name))
    {
      *database = db_iter;
      return dbeSuccess;
    }
  }
  return dbeParseFail;
}


/** Create a new database root node if required */
static enum eDBError db_interp_xml_create_root(ezxml_t *xmlroot)
{
  ezxml_t xrn;
  bool recreate;

  xrn = *xmlroot;
  recreate = (NULL == xrn ||
              0 != strcmp(ezxml_error(xrn), "") ||
              NULL == xrn->name ||
              0 != strcmp(xrn->name, DB_DATABASE_ROOTNODENAME));

  if (recreate)
  {
    *xmlroot = NULL;
    if (NULL != xrn)
    {
      ezxml_free(xrn);
    }

    /* Create root node if not already existing */
    xrn = ezxml_new_d(DB_DATABASE_ROOTNODENAME);
    if (NULL == xrn)
    {
      return dbeMemAllocFail;
    }
    *xmlroot = xrn;
  }

  return dbeSuccess;
}


/** Add a database with name */
static enum eDBError db_interp_xml_add_database(ezxml_t xmlroot, const char *name, ezxml_t *database)
{
  ezxml_t db;
  ezxml_t attr;

  /* Create new database */
  db = ezxml_add_child_d(xmlroot, DB_DATABASE_NODENAME, 0);
  if (NULL == db)
  {
    return dbeMemAllocFail;
  }

  /* Set section name attribute */
  attr = ezxml_set_attr_d(db, DB_DATABASE_ATTRNAME, name);
  if (NULL == attr)
  {
    return dbeMemAllocFail;
  }

  /* Return the new section */
  *database = db;
  return dbeSuccess;
}


/** Set the database version and checksum */
static enum eDBError db_interp_xml_set_database_version_and_checksum_attr(ezxml_t database, uint32_t db_version, uint32_t db_checksum)
{
  char str[40];
  ezxml_t attr;

  /* Set the version attribute */
  snprintf(str, 40, "%lu", (unsigned long) db_version);
  attr = ezxml_set_attr_d(database, DB_DATABASE_ATTRVERSION, str);
  if (NULL == attr)
  {
    return dbeMemAllocFail;
  }

  /* Set the checksum attribute */
  snprintf(str, 40, "%lu", (unsigned long) db_checksum);
  attr = ezxml_set_attr_d(database, DB_DATABASE_ATTRCHECKSUM, str);
  if (NULL == attr)
  {
    return dbeMemAllocFail;
  }

  return dbeSuccess;
}


/** Get the database version and checksum */
static enum eDBError db_interp_xml_get_database_version_and_checksum_attr(ezxml_t database, uint32_t *db_version, uint32_t *db_checksum)
{
  const char *temp;

  /* Get version and try to convert */
  temp = ezxml_attr(database, DB_DATABASE_ATTRVERSION);
  if (NULL == temp)
  {
    *db_version = DB_NO_MINIMUM_VERSION;
  }
  else
  {
    *db_version = strtoul(temp, (char **) NULL, 10);
  }

  /* Get checksum and try to convert */
  temp = ezxml_attr(database, DB_DATABASE_ATTRCHECKSUM);
  if (NULL == temp)
  {
    *db_checksum = 0;
  }
  else
  {
    *db_checksum = strtoul(temp, (char **) NULL, 10);
  }

  return dbeSuccess;
}


/** Open and parse the XML file, if not already open/parsed. db_xml->filename should be set before calling */
static enum eDBError db_database_open_and_parse_xml_file(struct s_db_database_xml *db_xml)
{
  enum eDBError res;

  if (NULL == db_xml->xmlroot)
  {
    /* Hasn't been parsed yet */
    if (NULL == db_xml->filestream)
    {
      /* File not open yet, open for reading and writing */
      db_xml->filestream = fopen(db_xml->filename, "a+");
      if (NULL != db_xml->filestream)
      {
        if (fcntl(fileno(db_xml->filestream), F_SETFD, FD_CLOEXEC) < 0)
        {
          return dbeFileFail;
        }
      }
    }
    if (NULL == db_xml->filestream)
    {
      return dbeFileFail;
    }

    /* Seek to start of file */
    if (0 != fseek(db_xml->filestream, 0, SEEK_SET))
    {
      return dbeFileFail;
    }

    /* Create XML structure from file */
    db_xml->xmlroot = ezxml_parse_fp(db_xml->filestream);
  }

  /* Create a new database root node if required */
  res = db_interp_xml_create_root(&db_xml->xmlroot);
  return res;
}


/** Write the current XML structure to it's file */
static enum eDBError db_database_write_xml_file(struct s_db_database_xml *db_xml)
{
  char *xmlstr;
  size_t length;
  enum eDBError res;

  /* Convert XML to string */
  xmlstr = ezxml_toxml(db_xml->xmlroot);
  if (NULL == xmlstr)
  {
    res = dbeParseFail;
    goto ddwxf_fail;
  }

  /* Work out length */
  length = strlen(xmlstr);

  /* Truncate file (to 0 bytes) */
  if (ftruncate(fileno(db_xml->filestream), 0) != 0)
  {
    res = dbeFileFail;
    goto ddwxf_free;
  }

  /* Back to start of file */
  if (0 != fseek(db_xml->filestream, 0, SEEK_SET))
  {
    res = dbeFileFail;
    goto ddwxf_free;
  }

  /* Write out data */
  if (length != fwrite(xmlstr, sizeof(char), length, db_xml->filestream))
  {
    res = dbeFileFail;
    goto ddwxf_free;
  }

  /* Force to be written out */
  if (0 != fflush(db_xml->filestream))
  {
    res = dbeFileFail;
    goto ddwxf_free;
  }

  /* Success */
  res = dbeSuccess;

ddwxf_free:
  /* Free the string */
  free(xmlstr);

ddwxf_fail:
  return res;
}


/** Close files and free memory */
static enum eDBError db_database_interp_xml_cleanup(struct s_db_database_xml *db_xml)
{
  /* Free filename */
  if (NULL != db_xml->filename)
  {
    free(db_xml->filename);
  }
  db_xml->filename = NULL;

  /* Close file */
  if (NULL != db_xml->filestream)
  {
    fclose(db_xml->filestream);
  }
  db_xml->filestream = NULL;

  /* Free up XML structure */
  ezxml_free(db_xml->xmlroot);
  db_xml->xmlroot = NULL;
  db_xml->dbroot = NULL;

  return dbeSuccess;
}


/** Export database start function for XML */
static enum eDBError db_database_xml_export_start(struct s_db_database_interp *ddi, const struct s_db_database *db, uint32_t at_version, uint32_t checksum)
{
  enum eDBError res;
  struct s_db_database_xml *db_xml;

  /* Check pointers */
  if (NULL == ddi || NULL == db || NULL == ddi->database_data)
    return dbeNullPointer;
  db_xml = (struct s_db_database_xml *) ddi->database_data;

  /* Open file and parse XML, if not already done so */
  res = db_database_open_and_parse_xml_file(db_xml);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Find the database node */
  res = db_interp_xml_get_database(db_xml->xmlroot, db->db_name, &db_xml->dbroot);
  if (dbeSuccess != res)
  {
    /* Not found, create */
    res = db_interp_xml_add_database(db_xml->xmlroot, db->db_name, &db_xml->dbroot);
    if (dbeSuccess != res)
    {
      return res;
    }
  }

  /* Set database version and checksum */
  res = db_interp_xml_set_database_version_and_checksum_attr(db_xml->dbroot, at_version, checksum);

  /* Make database root node available to element interpreter */
  ddi->interp.interpreter = db_xml->dbroot;

  return res;
}


/** Export database end function for XML */
static enum eDBError db_database_xml_export_end(struct s_db_database_interp *ddi, const struct s_db_database *db, uint32_t at_version, uint32_t checksum)
{
  struct s_db_database_xml *db_xml;

  /* Check pointers */
  if (NULL == ddi || NULL == db || NULL == ddi->database_data)
    return dbeNullPointer;
  db_xml = (struct s_db_database_xml *) ddi->database_data;

  /* Write the database out to XML file */
  return db_database_write_xml_file(db_xml);
}


/** Import database start function for XML */
static enum eDBError db_database_xml_import_start(struct s_db_database_interp *ddi, struct s_db_database *db, uint32_t *version, uint32_t *checksum)
{
  enum eDBError res;
  struct s_db_database_xml *db_xml;

  /* Check pointers */
  if (NULL == ddi || NULL == db || NULL == version || NULL == checksum || NULL == ddi->database_data)
    return dbeNullPointer;
  db_xml = (struct s_db_database_xml *) ddi->database_data;

  /* Open file and parse XML, if not already done so */
  res = db_database_open_and_parse_xml_file(db_xml);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Find the database node */
  res = db_interp_xml_get_database(db_xml->xmlroot, db->db_name, &db_xml->dbroot);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Get the database version from attributes */
  res = db_interp_xml_get_database_version_and_checksum_attr(db_xml->dbroot, version, checksum);
  if (dbeSuccess != res)
  {
    /* Not found, fail */
    return res;
  }

  /* Make database root node available to element interpreter */
  ddi->interp.interpreter = db_xml->dbroot;

  return dbeSuccess;
}


/** Import database end function for XML */
static enum eDBError db_database_xml_import_end(struct s_db_database_interp *ddi, struct s_db_database *db, uint32_t *version, uint32_t *checksum)
{
  /* Nothing to do here */
  return dbeSuccess;
}


/** Create an XML database interpreter */
enum eDBError db_create_database_interp_xml(struct s_db_database_interp *ddi, const char *filename)
{
  enum eDBError res;
  struct s_db_database_xml *db_xml = NULL;

  /* Check pointers */
  if (NULL == ddi || NULL == filename)
    return dbeNullPointer;

  /* Allocate structure */
  db_xml = malloc(sizeof(struct s_db_database_xml));
  if (NULL == db_xml)
  {
    res = dbeMemAllocFail;
    goto dcdix_fail;
  }
  memset(db_xml, 0, sizeof(*db_xml));

  /* Allocate filename string */
  db_xml->filename = strdup(filename);
  if (NULL == db_xml->filename)
  {
    res = dbeMemAllocFail;
    goto dcdix_fail;
  }

  /* Setup elements */
  db_xml->filestream = NULL;
  db_xml->xmlroot = NULL;
  db_xml->dbroot = NULL;

  /* Keep db_xml for future operations */
  ddi->database_data = db_xml;

  /* Create interpreter */
  res = db_create_interp_xml(&ddi->interp, NULL);
  if (dbeSuccess != res)
  {
    goto dcdix_fail;
  }

  /* Setup import and export functions */
  ddi->database_export_start_fn = db_database_xml_export_start;
  ddi->database_export_end_fn = db_database_xml_export_end;
  ddi->database_import_start_fn = db_database_xml_import_start;
  ddi->database_import_end_fn = db_database_xml_import_end;

  /* Success */
  return dbeSuccess;

dcdix_fail:
  ddi->database_data = NULL;
  if (NULL != db_xml)
  {
    if (NULL != db_xml->filename)
    {
      free(db_xml->filename);
      db_xml->filename = NULL;
    }
    free(db_xml);
  }
  return res;
}


/** Destroy an XML database interpreter */
enum eDBError db_destroy_database_interp_xml(struct s_db_database_interp *ddi)
{
  struct s_db_database_xml *db_xml;

  /* Check pointers */
  if (NULL == ddi || NULL == ddi->database_data)
    return dbeNullPointer;
  db_xml = (struct s_db_database_xml *) ddi->database_data;

  /* Free things within the db_xml structure */
  (void) db_database_interp_xml_cleanup(db_xml);

  /* Free the db_xml structure */
  ddi->database_data = NULL;
  free(db_xml);

  return dbeSuccess;
}
