/*
 * db_element.c
 *
 *  Created on: 08/07/2011
 *      Author: ian
 */
#include <string.h>
#include <stdlib.h>
#include <rindb/db_changelog.h>
#include <rindb/db_strncpy.h>

#include <rindb/db_element.h>


/** Get the default for a string array element */
#define str_array_get_default(elem, idx) \
  ((NULL != elem->elem_data.elem_string_array.default_value && \
    NULL != elem->elem_data.elem_string_array.default_value[idx]) ? \
   elem->elem_data.elem_string_array.default_value[idx] : \
   elem->elem_data.elem_string_array.single_default)


/** Get the maximum length for a string array element */
#define str_array_get_max_length(elem, idx) \
  ((NULL != elem->elem_data.elem_string_array.max_length) ? \
   elem->elem_data.elem_string_array.max_length[idx] : \
   elem->elem_data.elem_string_array.single_max_length)


/** Determine if a database element is valid at for a particular version of the database */
bool db_element_valid_for_version(const struct s_db_element *elem, uint32_t at_version)
{
  return (at_version >= elem->minimum_version && at_version <= elem->maximum_version);
}


/** Determine if a database element should appear in the change log, based on changelog masks */
bool db_element_changelog_visible(const struct s_db_element *elem, uint32_t logmask)
{
  return db_changelog_visible(elem->changelog_mask, logmask);
}


/** Determine if an database element has changed */
bool db_element_changed(const struct s_db_element *elem)
{
  bool changed = false;
  size_t i;
  size_t maxlen;

  switch (elem->db_elem_type)
  {
    case dbetUInt8:
      if (NULL != elem->elem_data.elem_uint8.data)
      {
        changed = (elem->elem_data.elem_uint8.previous_data != *elem->elem_data.elem_uint8.data);
      }
      break;

    case dbetUInt32:
      if (NULL != elem->elem_data.elem_uint32.data)
      {
        changed = (elem->elem_data.elem_uint32.previous_data != *elem->elem_data.elem_uint32.data);
      }
      break;

    case dbetInt32:
      if (NULL != elem->elem_data.elem_int32.data)
      {
        changed = (elem->elem_data.elem_int32.previous_data != *elem->elem_data.elem_int32.data);
      }
      break;

    case dbetString:
      if (NULL != elem->elem_data.elem_string.data && NULL != elem->elem_data.elem_string.previous_data)
      {
        changed = (0 != strncmp(elem->elem_data.elem_string.previous_data, elem->elem_data.elem_string.data, elem->elem_data.elem_string.max_length));
      }
      break;

    case dbetBuffer:
      if (NULL != elem->elem_data.elem_buffer.data && NULL != elem->elem_data.elem_buffer.previous_data)
      {
        changed = (0 != memcmp(elem->elem_data.elem_buffer.previous_data, elem->elem_data.elem_buffer.data, elem->elem_data.elem_buffer.length));
      }
      break;

    case dbetUInt8Array:
      if (NULL != elem->elem_data.elem_uint8_array.data && NULL != elem->elem_data.elem_uint8_array.previous_data)
      {
        changed = (0 != memcmp(elem->elem_data.elem_uint8_array.previous_data, elem->elem_data.elem_uint8_array.data, elem->elem_data.elem_uint8_array.num_elements * sizeof(uint8_t)));
      }
      break;

    case dbetUInt32Array:
      if (NULL != elem->elem_data.elem_uint32_array.data && NULL != elem->elem_data.elem_uint32_array.previous_data)
      {
        changed = (0 != memcmp(elem->elem_data.elem_uint32_array.previous_data, elem->elem_data.elem_uint32_array.data, elem->elem_data.elem_uint32_array.num_elements * sizeof(uint32_t)));
      }
      break;

    case dbetInt32Array:
      if (NULL != elem->elem_data.elem_int32_array.data && NULL != elem->elem_data.elem_int32_array.previous_data)
      {
        changed = (0 != memcmp(elem->elem_data.elem_int32_array.previous_data, elem->elem_data.elem_int32_array.data, elem->elem_data.elem_int32_array.num_elements * sizeof(int32_t)));
      }
      break;

    case dbetStringArray:
      if (NULL != elem->elem_data.elem_string_array.data && NULL != elem->elem_data.elem_string_array.previous_data)
      {
        for (i = 0, changed = false; i < elem->elem_data.elem_string_array.num_elements; i++)
        {
          maxlen = str_array_get_max_length(elem, i);
          if (NULL != elem->elem_data.elem_string_array.data[i] && NULL != elem->elem_data.elem_string_array.previous_data[i] &&
              0 != strncmp(elem->elem_data.elem_string_array.previous_data[i], elem->elem_data.elem_string_array.data[i], maxlen))
          {
            changed = true;
          }
        }
      }
      break;

    case dbetNone:
    default:
      changed = false;
      break;
  }
  return changed;
}


/** Write out the database element */
enum eDBError db_element_write(const char *db_name, const struct s_db_element *elem, struct s_db_interp *db_interp, uint32_t db_version)
{
  enum eDBError res = dbeSuccess;

  /*
   * If element is not present in this database version,
   * or element is obsolete in this database version,
   * then don't write it
   */
  if (db_element_valid_for_version(elem, db_version))
  {
    switch (elem->db_elem_type)
    {
      case dbetUInt8:
        if (NULL != elem->elem_data.elem_uint8.data)
        {
          res = db_interp_uint8_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_uint8);
        }
        break;

      case dbetUInt32:
        if (NULL != elem->elem_data.elem_uint32.data)
        {
          res = db_interp_uint32_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_uint32);
        }
        break;

      case dbetInt32:
        if (NULL != elem->elem_data.elem_int32.data)
        {
          res = db_interp_int32_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_int32);
        }
        break;

      case dbetString:
        if (NULL != elem->elem_data.elem_string.data)
        {
          res = db_interp_str_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_string);
        }
        break;

      case dbetBuffer:
        if (NULL != elem->elem_data.elem_buffer.data)
        {
          res = db_interp_buf_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_buffer);
        }
        break;

      case dbetUInt8Array:
        if (NULL != elem->elem_data.elem_uint8_array.data)
        {
          res = db_interp_uint8_array_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_uint8_array);
        }
        break;

      case dbetUInt32Array:
        if (NULL != elem->elem_data.elem_uint32_array.data)
        {
          res = db_interp_uint32_array_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_uint32_array);
        }
        break;

      case dbetInt32Array:
        if (NULL != elem->elem_data.elem_int32_array.data)
        {
          res = db_interp_int32_array_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_int32_array);
        }
        break;

      case dbetStringArray:
        if (NULL != elem->elem_data.elem_string_array.data)
        {
          res = db_interp_string_array_export(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_string_array);
        }
        break;

      case dbetNone:
      default:
        /* Nothing to do here */
        res = dbeOutsideLimits;
        break;
    }
  }
  return res;
}


/** Read in the database element */
enum eDBError db_element_read(const char *db_name, struct s_db_element *elem, struct s_db_interp *db_interp, uint32_t db_version)
{
  enum eDBError res = dbeSuccess;

  /*
   * If element was not present in this database version,
   * or element is obsolete in this database version,
   * then just set it to defaults
   */
  if (!db_element_valid_for_version(elem, db_version))
  {
    res = db_element_default(elem);
  }
  else
  {
    /* Present in this database, so do normal read */
    switch (elem->db_elem_type)
    {
      case dbetUInt8:
        if (NULL != elem->elem_data.elem_uint8.data)
        {
          res = db_interp_uint8_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_uint8);
        }
        break;

      case dbetUInt32:
        if (NULL != elem->elem_data.elem_uint32.data)
        {
          res = db_interp_uint32_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_uint32);
        }
        break;

      case dbetInt32:
        if (NULL != elem->elem_data.elem_int32.data)
        {
          res = db_interp_int32_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_int32);
        }
        break;

      case dbetString:
        if (NULL != elem->elem_data.elem_string.data)
        {
          res = db_interp_str_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_string);
        }
        break;

      case dbetBuffer:
        if (NULL != elem->elem_data.elem_buffer.data)
        {
          res = db_interp_buf_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_buffer);
        }
        break;

      case dbetUInt8Array:
        if (NULL != elem->elem_data.elem_uint8_array.data)
        {
          res = db_interp_uint8_array_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_uint8_array);
        }
        break;

      case dbetUInt32Array:
        if (NULL != elem->elem_data.elem_uint32_array.data)
        {
          res = db_interp_uint32_array_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_uint32_array);
        }
        break;

      case dbetInt32Array:
        if (NULL != elem->elem_data.elem_int32_array.data)
        {
          res = db_interp_int32_array_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_int32_array);
        }
        break;

      case dbetStringArray:
        if (NULL != elem->elem_data.elem_string_array.data)
        {
          res = db_interp_string_array_import(db_name, db_interp, elem->elem_db_name, &elem->elem_data.elem_string_array);
        }
        break;

      case dbetNone:
      default:
        /* Nothing to do here */
        res = dbeOutsideLimits;
        break;
    }
  }
  return res;
}


/** Set a database element to defaults */
enum eDBError db_element_default(struct s_db_element *elem)
{
  enum eDBError res = dbeSuccess;
  size_t i;
  const char *default_str;
  size_t maxlen;

  switch (elem->db_elem_type)
  {
    case dbetUInt8:
      if (NULL != elem->elem_data.elem_uint8.data)
      {
        *elem->elem_data.elem_uint8.data = elem->elem_data.elem_uint8.default_value;
      }
      break;

    case dbetUInt32:
      if (NULL != elem->elem_data.elem_uint32.data)
      {
        *elem->elem_data.elem_uint32.data = elem->elem_data.elem_uint32.default_value;
      }
      break;

    case dbetInt32:
      if (NULL != elem->elem_data.elem_int32.data)
      {
        *elem->elem_data.elem_int32.data = elem->elem_data.elem_int32.default_value;
      }
      break;

    case dbetString:
      if (NULL != elem->elem_data.elem_string.data && NULL != elem->elem_data.elem_string.default_value)
      {
        db_safe_strncpy(elem->elem_data.elem_string.data, elem->elem_data.elem_string.default_value, elem->elem_data.elem_string.max_length);
      }
      break;

    case dbetBuffer:
      if (NULL != elem->elem_data.elem_buffer.data && NULL != elem->elem_data.elem_buffer.default_value)
      {
        memcpy(elem->elem_data.elem_buffer.data, elem->elem_data.elem_buffer.default_value, elem->elem_data.elem_buffer.length);
      }
      break;

    case dbetUInt8Array:
      if (NULL != elem->elem_data.elem_uint8_array.data)
      {
        for (i = 0; i < elem->elem_data.elem_uint8_array.num_elements; i++)
        {
          elem->elem_data.elem_uint8_array.data[i] = (NULL != elem->elem_data.elem_uint8_array.default_value) ?
                                                     elem->elem_data.elem_uint8_array.default_value[i] : elem->elem_data.elem_uint8_array.single_default;
        }
      }
      break;

    case dbetUInt32Array:
      if (NULL != elem->elem_data.elem_uint32_array.data)
      {
        for (i = 0; i < elem->elem_data.elem_uint32_array.num_elements; i++)
        {
          elem->elem_data.elem_uint32_array.data[i] = (NULL != elem->elem_data.elem_uint32_array.default_value) ?
                                                      elem->elem_data.elem_uint32_array.default_value[i] : elem->elem_data.elem_uint32_array.single_default;
        }
      }
      break;

    case dbetInt32Array:
      if (NULL != elem->elem_data.elem_int32_array.data)
      {
        for (i = 0; i < elem->elem_data.elem_int32_array.num_elements; i++)
        {
          elem->elem_data.elem_int32_array.data[i] = (NULL != elem->elem_data.elem_int32_array.default_value) ?
                                                     elem->elem_data.elem_int32_array.default_value[i] : elem->elem_data.elem_int32_array.single_default;
        }
      }
      break;

    case dbetStringArray:
      if (NULL != elem->elem_data.elem_string_array.data)
      {
        for (i = 0; i < elem->elem_data.elem_string_array.num_elements; i++)
        {
          default_str = str_array_get_default(elem, i);
          maxlen = str_array_get_max_length(elem, i);
          db_safe_strncpy(elem->elem_data.elem_string_array.data[i], default_str, maxlen);
        }
      }
      break;

    case dbetNone:
    default:
      /* Nothing to do here */
      res = dbeOutsideLimits;
      break;
  }
  return res;
}


/** Update element so previous data matches current data */
enum eDBError db_element_update_previous(struct s_db_element *elem)
{
  enum eDBError res = dbeSuccess;
  size_t i;
  size_t maxlen;

  switch (elem->db_elem_type)
  {
    case dbetUInt8:
      if (NULL != elem->elem_data.elem_uint8.data)
      {
        elem->elem_data.elem_uint8.previous_data = *elem->elem_data.elem_uint8.data;
      }
      break;

    case dbetUInt32:
      if (NULL != elem->elem_data.elem_uint32.data)
      {
        elem->elem_data.elem_uint32.previous_data = *elem->elem_data.elem_uint32.data;
      }
      break;

    case dbetInt32:
      if (NULL != elem->elem_data.elem_int32.data)
      {
        elem->elem_data.elem_int32.previous_data = *elem->elem_data.elem_int32.data;
      }
      break;

    case dbetString:
      if (NULL != elem->elem_data.elem_string.data && NULL != elem->elem_data.elem_string.previous_data)
      {
        db_safe_strncpy(elem->elem_data.elem_string.previous_data, elem->elem_data.elem_string.data, elem->elem_data.elem_string.max_length);
      }
      break;

    case dbetBuffer:
      if (NULL != elem->elem_data.elem_buffer.data && NULL != elem->elem_data.elem_buffer.previous_data)
      {
        memcpy(elem->elem_data.elem_buffer.previous_data, elem->elem_data.elem_buffer.data, elem->elem_data.elem_buffer.length);
      }
      break;

    case dbetUInt8Array:
      if (NULL != elem->elem_data.elem_uint8_array.data && NULL != elem->elem_data.elem_uint8_array.previous_data)
      {
        memcpy(elem->elem_data.elem_uint8_array.previous_data, elem->elem_data.elem_uint8_array.data, elem->elem_data.elem_uint8_array.num_elements * sizeof(uint8_t));
      }
      break;

    case dbetUInt32Array:
      if (NULL != elem->elem_data.elem_uint32_array.data && NULL != elem->elem_data.elem_uint32_array.previous_data)
      {
        memcpy(elem->elem_data.elem_uint32_array.previous_data, elem->elem_data.elem_uint32_array.data, elem->elem_data.elem_uint32_array.num_elements * sizeof(uint32_t));
      }
      break;

    case dbetInt32Array:
      if (NULL != elem->elem_data.elem_int32_array.data && NULL != elem->elem_data.elem_int32_array.previous_data)
      {
        memcpy(elem->elem_data.elem_int32_array.previous_data, elem->elem_data.elem_int32_array.data, elem->elem_data.elem_int32_array.num_elements * sizeof(int32_t));
      }
      break;

    case dbetStringArray:
      if (NULL != elem->elem_data.elem_string_array.data && NULL != elem->elem_data.elem_string_array.previous_data)
      {
        for (i = 0; i < elem->elem_data.elem_string_array.num_elements; i++)
        {
          maxlen = str_array_get_max_length(elem, i);
          if (NULL != elem->elem_data.elem_string_array.data[i] && NULL != elem->elem_data.elem_string_array.previous_data[i])
          {
            db_safe_strncpy(elem->elem_data.elem_string_array.previous_data[i], elem->elem_data.elem_string_array.data[i], maxlen);
          }
        }
      }
      break;

    case dbetNone:
    default:
      /* Nothing to do here */
      res = dbeOutsideLimits;
      break;
  }
  return res;
}


/** Revert element data to previous value */
enum eDBError db_element_revert(struct s_db_element *elem)
{
  enum eDBError res = dbeSuccess;
  size_t i;
  size_t maxlen;

  switch (elem->db_elem_type)
  {
    case dbetUInt8:
      if (NULL != elem->elem_data.elem_uint8.data)
      {
        *elem->elem_data.elem_uint8.data = elem->elem_data.elem_uint8.previous_data;
      }
      break;

    case dbetUInt32:
      if (NULL != elem->elem_data.elem_uint32.data)
      {
        *elem->elem_data.elem_uint32.data = elem->elem_data.elem_uint32.previous_data;
      }
      break;

    case dbetInt32:
      if (NULL != elem->elem_data.elem_int32.data)
      {
        *elem->elem_data.elem_int32.data = elem->elem_data.elem_int32.previous_data;
      }
      break;

    case dbetString:
      if (NULL != elem->elem_data.elem_string.data && NULL != elem->elem_data.elem_string.previous_data)
      {
        db_safe_strncpy(elem->elem_data.elem_string.data, elem->elem_data.elem_string.previous_data, elem->elem_data.elem_string.max_length);
      }
      break;

    case dbetBuffer:
      if (NULL != elem->elem_data.elem_buffer.data && NULL != elem->elem_data.elem_buffer.previous_data)
      {
        memcpy(elem->elem_data.elem_buffer.data, elem->elem_data.elem_buffer.previous_data, elem->elem_data.elem_buffer.length);
      }
      break;

    case dbetUInt8Array:
      if (NULL != elem->elem_data.elem_uint8_array.data && NULL != elem->elem_data.elem_uint8_array.previous_data)
      {
        memcpy(elem->elem_data.elem_uint8_array.data, elem->elem_data.elem_uint8_array.previous_data, elem->elem_data.elem_uint8_array.num_elements * sizeof(uint8_t));
      }
      break;

    case dbetUInt32Array:
      if (NULL != elem->elem_data.elem_uint32_array.data && NULL != elem->elem_data.elem_uint32_array.previous_data)
      {
        memcpy(elem->elem_data.elem_uint32_array.data, elem->elem_data.elem_uint32_array.previous_data, elem->elem_data.elem_uint32_array.num_elements * sizeof(uint32_t));
      }
      break;

    case dbetInt32Array:
      if (NULL != elem->elem_data.elem_int32_array.data && NULL != elem->elem_data.elem_int32_array.previous_data)
      {
        memcpy(elem->elem_data.elem_int32_array.data, elem->elem_data.elem_int32_array.previous_data, elem->elem_data.elem_int32_array.num_elements * sizeof(int32_t));
      }
      break;

    case dbetStringArray:
      if (NULL != elem->elem_data.elem_string_array.data && NULL != elem->elem_data.elem_string_array.previous_data)
      {
        for (i = 0; i < elem->elem_data.elem_string_array.num_elements; i++)
        {
          maxlen = str_array_get_max_length(elem, i);
          if (NULL != elem->elem_data.elem_string_array.data[i] && NULL != elem->elem_data.elem_string_array.previous_data[i])
          {
            db_safe_strncpy(elem->elem_data.elem_string_array.data[i], elem->elem_data.elem_string_array.previous_data[i], maxlen);
          }
        }
      }
      break;

    case dbetNone:
    default:
      /* Nothing to do here */
      res = dbeOutsideLimits;
      break;
  }
  return res;
}


/** Work out the maximum size of an element */
enum eDBError db_element_get_maximum_size(struct s_db_element *elem, size_t *elem_max_size)
{
  enum eDBError res = dbeSuccess;
  size_t i;
  size_t maxlen;

  switch (elem->db_elem_type)
  {
    case dbetUInt8:
      *elem_max_size = sizeof(*elem->elem_data.elem_uint8.data);
      break;

    case dbetUInt32:
      *elem_max_size = sizeof(*elem->elem_data.elem_uint32.data);
      break;

    case dbetInt32:
      *elem_max_size = sizeof(*elem->elem_data.elem_int32.data);
      break;

    case dbetString:
      *elem_max_size = sizeof(char) * elem->elem_data.elem_string.max_length;
      break;

    case dbetBuffer:
      *elem_max_size = sizeof(uint8_t) * elem->elem_data.elem_buffer.length;
      break;

    case dbetUInt8Array:
      *elem_max_size = sizeof(uint8_t) * elem->elem_data.elem_uint8_array.num_elements;
      break;

    case dbetUInt32Array:
      *elem_max_size = sizeof(uint32_t) * elem->elem_data.elem_uint32_array.num_elements;
      break;

    case dbetInt32Array:
      *elem_max_size = sizeof(int32_t) * elem->elem_data.elem_int32_array.num_elements;
      break;

    case dbetStringArray:
      for (i = 0, maxlen = 0; i < elem->elem_data.elem_string_array.num_elements; i++)
        maxlen += str_array_get_max_length(elem, i);
      *elem_max_size = maxlen;
      break;

    case dbetNone:
    default:
      /* Nothing to do here */
      res = dbeOutsideLimits;
      *elem_max_size = 0;
      break;
  }
  return res;
}


#define DB_CALLOC_AND_CHECK(elem, elem_type, elem_member, array_length, set_mallocd) \
do \
{ \
  if (NULL == elem->elem_data.elem_type.elem_member) \
  { \
    elem->elem_data.elem_type.elem_member = calloc(sizeof(*elem->elem_data.elem_type.elem_member), array_length); \
    if (NULL == elem->elem_data.elem_type.elem_member) \
    { \
      res = dbeMemAllocFail; \
      goto dei_fail; \
    } \
    if (set_mallocd) \
      elem->data_is_mallocd = true; \
  } \
} \
while (0)

#define DB_CALLOC_AND_CHECK_DATA(elem, elem_type, array_length) DB_CALLOC_AND_CHECK(elem, elem_type, data, array_length, true)
#define DB_CALLOC_AND_CHECK_PREVDATA(elem, elem_type, array_length) DB_CALLOC_AND_CHECK(elem, elem_type, previous_data, array_length, false)


#define DB_CALLOC_AND_CHECK_STRING_ARRAY(elem, elem_member, set_mallocd) \
do \
{ \
  if (NULL == elem->elem_data.elem_string_array.elem_member) \
  { \
    if (set_mallocd) \
      elem->data_is_mallocd = true; \
    data = calloc(elem->elem_data.elem_string_array.num_elements, sizeof(char *)); \
    if (NULL == data) \
    { \
      res = dbeMemAllocFail; \
      goto dei_fail; \
    } \
    elem->elem_data.elem_string_array.elem_member = data; \
    for (i = 0; i < elem->elem_data.elem_string_array.num_elements; i++) \
    { \
      maxlen = str_array_get_max_length(elem, i); \
      data[i] = calloc(maxlen, sizeof(char)); \
      if (NULL == data[i]) \
      { \
        res = dbeMemAllocFail; \
        goto dei_fail; \
      } \
    } \
  } \
} \
while (0)

#define DB_CALLOC_AND_CHECK_STRING_ARRAY_DATA(elem) DB_CALLOC_AND_CHECK_STRING_ARRAY(elem, data, true)
#define DB_CALLOC_AND_CHECK_STRING_ARRAY_PREVDATA(elem) DB_CALLOC_AND_CHECK_STRING_ARRAY(elem, previous_data, false)


/**
 * Initialise a database element
 *
 * If a data element pointer is NULL (indicating it is no longer used in the current database version),
 * then it is allocated here, as we still need it for validating the checksum when reading database versions that is was contained in.
 * This also sets a flag so we know to free it when the element is destroyed.
 *
 * This also allocates memory for the previous value for some element types, and sets it to current value
 */
enum eDBError db_element_init(struct s_db_element *elem)
{
  enum eDBError res = dbeSuccess;
  size_t i;
  size_t maxlen;
  char **data;

  switch (elem->db_elem_type)
  {
    case dbetUInt8:
      DB_CALLOC_AND_CHECK_DATA(elem, elem_uint8, 1);
      break;
    case dbetUInt32:
      DB_CALLOC_AND_CHECK_DATA(elem, elem_uint32, 1);
      break;
    case dbetInt32:
      DB_CALLOC_AND_CHECK_DATA(elem, elem_int32, 1);
      break;

    case dbetString:
      DB_CALLOC_AND_CHECK_DATA(elem, elem_string, elem->elem_data.elem_string.max_length);
      DB_CALLOC_AND_CHECK_PREVDATA(elem, elem_string, elem->elem_data.elem_string.max_length);
      break;

    case dbetBuffer:
      DB_CALLOC_AND_CHECK_DATA(elem, elem_buffer, elem->elem_data.elem_buffer.length);
      DB_CALLOC_AND_CHECK_PREVDATA(elem, elem_buffer, elem->elem_data.elem_buffer.length);
      break;

    case dbetUInt8Array:
      DB_CALLOC_AND_CHECK_DATA(elem, elem_uint8_array, elem->elem_data.elem_uint8_array.num_elements);
      DB_CALLOC_AND_CHECK_PREVDATA(elem, elem_uint8_array, elem->elem_data.elem_uint8_array.num_elements);
      break;

    case dbetUInt32Array:
      DB_CALLOC_AND_CHECK_DATA(elem, elem_uint32_array, elem->elem_data.elem_uint32_array.num_elements);
      DB_CALLOC_AND_CHECK_PREVDATA(elem, elem_uint32_array, elem->elem_data.elem_uint32_array.num_elements);
      break;

    case dbetInt32Array:
      DB_CALLOC_AND_CHECK_DATA(elem, elem_int32_array, elem->elem_data.elem_int32_array.num_elements);
      DB_CALLOC_AND_CHECK_PREVDATA(elem, elem_int32_array, elem->elem_data.elem_int32_array.num_elements);
      break;

    case dbetStringArray:
      DB_CALLOC_AND_CHECK_STRING_ARRAY_DATA(elem);
      DB_CALLOC_AND_CHECK_STRING_ARRAY_PREVDATA(elem);
      break;

    case dbetNone:
    default:
      /* Nothing to do here */
      res = dbeOutsideLimits;
      break;
  }

dei_fail:
  /* Now set to current data value */
  if (dbeSuccess == res)
  {
    res = db_element_update_previous(elem);
  }
  return res;
}


#define DB_FREE_MALLOCED_ELEM_DATA(elem, elem_type) \
do \
{ \
  if (elem->data_is_mallocd && NULL != elem->elem_data.elem_type.data) \
  { \
    free(elem->elem_data.elem_type.data); \
    elem->elem_data.elem_type.data = NULL; \
    elem->data_is_mallocd = false; \
  } \
} \
while (0)


#define DB_FREE_MALLOCED_ELEM_PREVDATA(elem, elem_type) \
do \
{ \
  if (NULL != elem->elem_data.elem_type.previous_data) \
  { \
    free(elem->elem_data.elem_type.previous_data); \
    elem->elem_data.elem_type.previous_data = NULL; \
  } \
} \
while (0)


/** Destroy string element data */
static void db_element_string_destroy_data(struct s_db_element *elem)
{
  size_t i;
  char **data;

  if (elem->data_is_mallocd && NULL != elem->elem_data.elem_string_array.data)
  {
    data = (char **) elem->elem_data.elem_string_array.data;
    for (i = 0; i < elem->elem_data.elem_string_array.num_elements; i++)
    {
      if (NULL != data[i])
      {
        free(data[i]);
        data[i] = NULL;
      }
    }
    free(data);
    elem->elem_data.elem_string_array.data = NULL;
  }
}


/** Destroy string element previous data */
static void db_element_string_destroy_prevdata(struct s_db_element *elem)
{
  size_t i;

  if (NULL != elem->elem_data.elem_string_array.previous_data)
  {
    for (i = 0; i < elem->elem_data.elem_string_array.num_elements; i++)
    {
      if (NULL != elem->elem_data.elem_string_array.previous_data[i])
      {
        free(elem->elem_data.elem_string_array.previous_data[i]);
        elem->elem_data.elem_string_array.previous_data[i] = NULL;
      }
    }
    free(elem->elem_data.elem_string_array.previous_data);
    elem->elem_data.elem_string_array.previous_data = NULL;
  }
}


/** Destroy element (free memory previous value for some element types) */
enum eDBError db_element_destroy(struct s_db_element *elem)
{
  enum eDBError res = dbeSuccess;

  switch (elem->db_elem_type)
  {
    /* No previous value allocation freeing here */
    case dbetUInt8:
      DB_FREE_MALLOCED_ELEM_DATA(elem, elem_uint8);
      break;

    case dbetUInt32:
      DB_FREE_MALLOCED_ELEM_DATA(elem, elem_uint32);
      break;

    case dbetInt32:
      DB_FREE_MALLOCED_ELEM_DATA(elem, elem_int32);
      break;

    /* Previous value allocation freeing here */
    case dbetString:
      DB_FREE_MALLOCED_ELEM_DATA(elem, elem_string);
      DB_FREE_MALLOCED_ELEM_PREVDATA(elem, elem_string);
      break;

    case dbetBuffer:
      DB_FREE_MALLOCED_ELEM_DATA(elem, elem_buffer);
      DB_FREE_MALLOCED_ELEM_PREVDATA(elem, elem_buffer);
      break;

    case dbetUInt8Array:
      DB_FREE_MALLOCED_ELEM_DATA(elem, elem_uint8_array);
      DB_FREE_MALLOCED_ELEM_PREVDATA(elem, elem_uint8_array);
      break;

    case dbetUInt32Array:
      DB_FREE_MALLOCED_ELEM_DATA(elem, elem_uint32_array);
      DB_FREE_MALLOCED_ELEM_PREVDATA(elem, elem_uint32_array);
      break;

    case dbetInt32Array:
      DB_FREE_MALLOCED_ELEM_DATA(elem, elem_int32_array);
      DB_FREE_MALLOCED_ELEM_PREVDATA(elem, elem_int32_array);
      break;

    case dbetStringArray:
      db_element_string_destroy_data(elem);
      db_element_string_destroy_prevdata(elem);
      break;

    case dbetNone:
    default:
      /* Nothing to do here */
      res = dbeOutsideLimits;
      break;
  }
  return res;
}
