/*
 * db_element.h
 *
 *  Created on: 05/07/2011
 *      Author: ian
 */

#ifndef DB_ELEMENT_H_
#define DB_ELEMENT_H_

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <rindb/db_interp.h>


/* Forward declaration */
struct s_db_interp;


/** Types of database elements */
enum eDBElementType { dbetNone, dbetUInt8, dbetUInt32, dbetInt32, dbetString, dbetBuffer, dbetUInt8Array, dbetUInt32Array, dbetInt32Array, dbetStringArray };


/** Unsigned 8 bit element */
struct s_db_element_uint8
{
  uint8_t *data;              /*!< Pointer to data, can be NULL if obsolete (in which case library will allocate and free it to allow checksum validation of old database versions) */
  uint8_t minimum;            /*!< Minimum value */
  uint8_t maximum;            /*!< Maximum value */
  uint8_t default_value;      /*!< Default value */
  uint8_t previous_data;      /*!< Used to determine if *data has changed */
  const char *const *lookup;  /*!< Used when this value is an array lookup (eg units->"g", "kg"). Can be NULL, in which case the numeric value of *data is used */
};


/** Unsigned 32bit element */
struct s_db_element_uint32
{
  uint32_t *data;             /*!< Pointer to data, can be NULL if obsolete (in which case library will allocate and free it to allow checksum validation of old database versions) */
  uint32_t minimum;           /*!< Minimum value */
  uint32_t maximum;           /*!< Maximum value */
  uint32_t default_value;     /*!< Default value */
  uint32_t previous_data;     /*!< Used to determine if *data has changed */
};


/** Signed 32bit element */
struct s_db_element_int32
{
  int32_t *data;              /*!< Pointer to data, can be NULL if obsolete (in which case library will allocate and free it to allow checksum validation of old database versions) */
  int32_t minimum;            /*!< Minimum value */
  int32_t maximum;            /*!< Maximum value */
  int32_t default_value;      /*!< Default value */
  int32_t previous_data;      /*!< Used to determine if *data has changed */
};


/** String element (NULL terminated) */
struct s_db_element_string
{
  char *data;                 /*!< Pointer to data, can be NULL if obsolete (in which case library will allocate and free it to allow checksum validation of old database versions) */
  size_t max_length;          /*!< The maximum length of the string */
  const char *default_value;  /*!< Default value */
  char *previous_data;        /*!< Used to determine if *data has changed */
};


/** Buffer element */
struct s_db_element_buffer
{
  uint8_t *data;              /*!< Pointer to data, can be NULL if obsolete (in which case library will allocate and free it to allow checksum validation of old database versions) */
  size_t length;              /*!< The length of the buffer */
  const uint8_t *default_value; /*!< Default value */
  uint8_t *previous_data;     /*!< Used to determine if *data has changed */
};


/** Unsigned 8 bit element array */
struct s_db_element_uint8_array
{
  uint8_t *data;                 /*!< Pointer to array data, can be NULL if obsolete (in which case library will allocate and free it to allow checksum validation of old database versions) */
  size_t num_elements;           /*!< Number of elements in data, minimum, maximum, default_values, previous_data */
  const uint8_t *minimum;        /*!< Minimum values, can be NULL, in which case single_minimum is used for all elements in data */
  const uint8_t *maximum;        /*!< Maximum values, can be NULL, in which case single_maximum is used for all elements in data */
  const uint8_t *default_value;  /*!< Default values, can be NULL, in which case single_default is used for all elements in data */
  uint8_t single_minimum;        /*!< Single minimum value, used when minimum is NULL for all elements in data */
  uint8_t single_maximum;        /*!< Single maximum value, used when maximum is NULL for all elements in data */
  uint8_t single_default;        /*!< Single default value, used when default_value is NULL for all elements in data */
  uint8_t *previous_data;        /*!< Used to determine if *data has changed */
  const char *const *lookup;     /*!< Used when this value is an array lookup (eg units->"g", "kg"). The same lookup is used for each element in data. Can be NULL, in which case the numeric value of *data is used */
};


/** Unsigned 32bit element array */
struct s_db_element_uint32_array
{
  uint32_t *data;                /*!< Pointer to array data, can be NULL if obsolete (in which case library will allocate and free it to allow checksum validation of old database versions) */
  size_t num_elements;           /*!< Number of elements in data, minimum, maximum, default_values, previous_data */
  const uint32_t *minimum;       /*!< Minimum values, can be NULL, in which case single_minimum is used for all elements in data */
  const uint32_t *maximum;       /*!< Maximum values, can be NULL, in which case single_maximum is used for all elements in data */
  const uint32_t *default_value; /*!< Default values, can be NULL, in which case single_default is used for all elements in data */
  uint32_t single_minimum;       /*!< Single minimum value, used when minimum is NULL for all elements in data */
  uint32_t single_maximum;       /*!< Single maximum value, used when maximum is NULL for all elements in data */
  uint32_t single_default;       /*!< Single default value, used when default_value is NULL for all elements in data */
  uint32_t *previous_data;       /*!< Used to determine if *data has changed */
};


/** Signed 32bit element array */
struct s_db_element_int32_array
{
  int32_t *data;                 /*!< Pointer to data, can be NULL if obsolete (in which case library will allocate and free it to allow checksum validation of old database versions) */
  size_t num_elements;           /*!< Number of elements in data, minimum, maximum, default_values, previous_data */
  const int32_t *minimum;        /*!< Minimum values, can be NULL, in which case single_minimum is used for all elements in data */
  const int32_t *maximum;        /*!< Maximum values, can be NULL, in which case single_maximum is used for all elements in data */
  const int32_t *default_value;  /*!< Default values, can be NULL, in which case single_default is used for all elements in data */
  int32_t single_minimum;        /*!< Single minimum value, used when minimum is NULL for all elements in data */
  int32_t single_maximum;        /*!< Single maximum value, used when maximum is NULL for all elements in data */
  int32_t single_default;        /*!< Single default value, used when default_value is NULL for all elements in data */
  int32_t *previous_data;        /*!< Used to determine if *data has changed */
};


/** String element (NULL terminated) array (array of strings) */
struct s_db_element_string_array
{
  char *const *data;             /*!< Pointer to array of pointers to string data */
  size_t num_elements;           /*!< Number of elements in data, default_values, previous_data */
  size_t *max_length;            /*!< The maximum lengths of each string (each string can be a different length) */
  const char *const *default_value;  /*!< Default values, can be NULL, in which case single_default is used for all elements in data */
  size_t single_max_length;      /*!< Single maximum length, used when max_length value is NULL for all elements in data */
  const char *single_default;    /*!< Single default value, used when default_value is NULL for all elements in data */
  char **previous_data;          /*!< Used to determine if data has changed */
};


/** Database element has no minimum version (ie present in all databases < maximum_version) */
#define DB_NO_MINIMUM_VERSION 0


/** Database element has no maximum version (ie not obsolete) */
#define DB_NO_MAXIMUM_VERSION UINT_MAX


/** Database element */
struct s_db_element
{
  enum eDBElementType db_elem_type;                 /*!< Element type, describes what's in the union elem_data */
  const char *elem_db_name;                         /*!< Element name, saved into some database file formats (eg XML). Also appears in the change log. Not NULL */
  const char *elem_menu_name;                       /*!< Element menu name, saved into the change log for easy identification of parameters. Can be NULL, in which case "" will appear in changelog */
  uint32_t minimum_version;                         /*!< The lowest version database section this element appears in, used to add new database settings */
  uint32_t maximum_version;                         /*!< The highest version database section this element appears in, used to obsolete database settings */
  uint32_t changelog_mask;                          /*!< This is ANDed with the mask passed when writing the change log to determine if this setting appears in the changelog */
  bool data_is_mallocd;                             /*!< The data field within elem_data has been malloc'd and needs freeing when the element is destroyed */
  union
  {
    struct s_db_element_uint8 elem_uint8;
    struct s_db_element_uint32 elem_uint32;
    struct s_db_element_int32 elem_int32;
    struct s_db_element_string elem_string;
    struct s_db_element_buffer elem_buffer;
    struct s_db_element_uint8_array elem_uint8_array;
    struct s_db_element_uint32_array elem_uint32_array;
    struct s_db_element_int32_array elem_int32_array;
    struct s_db_element_string_array elem_string_array;
  } elem_data;
};


/** Declare an 8bit unsigned integer database element */
#define DB_DECLARE_UINT8_ELEM(dbname, mname, ptr, min, max, def, ver_min, ver_max, cl_mask) \
 { .db_elem_type = dbetUInt8, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_uint8 = { .data = ptr, .minimum = min, .maximum = max, .default_value = def, .previous_data = 0, .lookup = NULL } } }


/** Declare an 8bit unsigned integer index database element (used to look up an array eg units->"kg, "g") */
#define DB_DECLARE_UINT8_INDEX_ELEM(dbname, mname, ptr, min, max, def, ver_min, ver_max, cl_mask, look) \
 { .db_elem_type = dbetUInt8, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_uint8 = { .data = ptr, .minimum = min, .maximum = max, .default_value = def, .previous_data = 0, .lookup = look } } }


/** Declare a 32bit unsigned integer database element */
#define DB_DECLARE_UINT32_ELEM(dbname, mname, ptr, min, max, def, ver_min, ver_max, cl_mask) \
 { .db_elem_type = dbetUInt32, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_uint32 = { .data = ptr, .minimum = min, .maximum = max, .default_value = def, .previous_data = 0 } } }


/** Declare a 32bit signed integer database element */
#define DB_DECLARE_INT32_ELEM(dbname, mname, ptr, min, max, def, ver_min, ver_max, cl_mask) \
 { .db_elem_type = dbetInt32, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_int32 = { .data = ptr, .minimum = min, .maximum = max, .default_value = def, .previous_data = 0 } } }


/** Declare a string database element */
#define DB_DECLARE_STRING_ELEM(dbname, mname, ptr, maxlen, def, ver_min, ver_max, cl_mask) \
 { .db_elem_type = dbetString, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_string = { .data = ptr, .max_length = maxlen, .default_value = def, .previous_data = NULL } } }


/** Declare a buffer database element */
#define DB_DECLARE_BUFFER_ELEM(dbname, mname, ptr, len, def, ver_min, ver_max, cl_mask) \
 { .db_elem_type = dbetBuffer, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_buffer = { .data = ptr, .length = len, .default_value = def, .previous_data = NULL } } }


/** Declare an 8bit unsigned integer database element array, used by defines below */
#define DB_INT_DECLARE_UINT8_ARRAY_ELEM(dbname, mname, ptr, elems, mins, maxs, defs, min, max, def, ver_min, ver_max, cl_mask, look) \
 { .db_elem_type = dbetUInt8Array, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_uint8_array = { .data = ptr, .num_elements = elems, .minimum = mins, .maximum = maxs, .default_value = defs, .single_minimum = min, .single_maximum = max, .single_default = def, .previous_data = NULL, .lookup = look } } }


/** Declare an 8bit unsigned integer database element array, using individual minimums, maximums, defaults */
#define DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL(dbname, mname, ptr, elems, mins, maxs, defs, ver_min, ver_max, cl_mask) \
  DB_INT_DECLARE_UINT8_ARRAY_ELEM(dbname, mname, ptr, elems, mins, maxs, defs, 0, 0, 0, ver_min, ver_max, cl_mask, NULL)


/** Declare an 8bit unsigned integer database element array, using single minimum, maximum, default */
#define DB_DECLARE_UINT8_ARRAY_ELEM_SINGLE(dbname, mname, ptr, elems, min, max, def, ver_min, ver_max, cl_mask) \
  DB_INT_DECLARE_UINT8_ARRAY_ELEM(dbname, mname, ptr, elems, NULL, NULL, NULL, min, max, def, ver_min, ver_max, cl_mask, NULL)


/** Declare an 8bit unsigned integer index database element array (used to look up an array eg units->"kg, "g"), using individual minimums, maximums, defaults */
#define DB_DECLARE_UINT8_ARRAY_INDEX_ELEM_INDIVIDUAL(dbname, mname, ptr, elems, mins, maxs, defs, ver_min, ver_max, cl_mask, look) \
  DB_INT_DECLARE_UINT8_ARRAY_ELEM(dbname, mname, ptr, elems, mins, maxs, defs, 0, 0, 0, ver_min, ver_max, cl_mask, look)


/** Declare an 8bit unsigned integer index database element array (used to look up an array eg units->"kg, "g"), using single minimum, maximum, default */
#define DB_DECLARE_UINT8_ARRAY_INDEX_ELEM_SINGLE(dbname, mname, ptr, elems, min, max, def, ver_min, ver_max, cl_mask, look) \
  DB_INT_DECLARE_UINT8_ARRAY_ELEM(dbname, mname, ptr, elems, NULL, NULL, NULL, min, max, def, ver_min, ver_max, cl_mask, look)


/** Declare a database element array, used by defines below */
#define DB_INT_DECLARE_ARRAY_ELEM(etype, unmemb, dbname, mname, ptr, elems, mins, maxs, defs, min, max, def, ver_min, ver_max, cl_mask) \
  { .db_elem_type = etype, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
    .elem_data = { .unmemb = { .data = ptr, .num_elements = elems, .minimum = mins, .maximum = maxs, .default_value = defs, .single_minimum = min, .single_maximum = max, .single_default = def, .previous_data = NULL } } }


/** Declare a 32bit unsigned integer database element array, using individual minimums, maximums, defaults */
#define DB_DECLARE_UINT32_ARRAY_ELEM_INDIVIDUAL(dbname, mname, ptr, elems, mins, maxs, defs, ver_min, ver_max, cl_mask) \
  DB_INT_DECLARE_ARRAY_ELEM(dbetUInt32Array, elem_uint32_array, dbname, mname, ptr, elems, mins, maxs, defs, 0, 0, 0, ver_min, ver_max, cl_mask)


/** Declare a 32bit unsigned integer database element array, using single minimum, maximum, default */
#define DB_DECLARE_UINT32_ARRAY_ELEM_SINGLE(dbname, mname, ptr, elems, min, max, def, ver_min, ver_max, cl_mask) \
  DB_INT_DECLARE_ARRAY_ELEM(dbetUInt32Array, elem_uint32_array, dbname, mname, ptr, elems, NULL, NULL, NULL, min, max, def, ver_min, ver_max, cl_mask)


/** Declare a 32bit signed integer database element array, using individual minimums, maximums, defaults */
#define DB_DECLARE_INT32_ARRAY_ELEM_INDIVIDUAL(dbname, mname, ptr, elems, mins, maxs, defs, ver_min, ver_max, cl_mask) \
  DB_INT_DECLARE_ARRAY_ELEM(dbetInt32Array, elem_int32_array, dbname, mname, ptr, elems, mins, maxs, defs, 0, 0, 0, ver_min, ver_max, cl_mask)


/** Declare a 32bit signed integer database element array, using single minimum, maximum, default */
#define DB_DECLARE_INT32_ARRAY_ELEM_SINGLE(dbname, mname, ptr, elems, min, max, def, ver_min, ver_max, cl_mask) \
  DB_INT_DECLARE_ARRAY_ELEM(dbetInt32Array, elem_int32_array, dbname, mname, ptr, elems, NULL, NULL, NULL, min, max, def, ver_min, ver_max, cl_mask)


/** Declare a string database element array, using individual maximum lengths and defaults */
#define DB_DECLARE_STRING_ARRAY_ELEM_INDIVIDUAL(dbname, mname, ptr, elems, maxlens, defs, ver_min, ver_max, cl_mask) \
 { .db_elem_type = dbetStringArray, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_string_array = { .data = ptr, .num_elements = elems, .max_length = maxlens, .default_value = defs, .single_max_length = 0, .single_default = NULL, .previous_data = NULL } } }


/** Declare a string database element array, using single maximum length and default */
#define DB_DECLARE_STRING_ARRAY_ELEM_SINGLE(dbname, mname, ptr, elems, maxlen, def, ver_min, ver_max, cl_mask) \
 { .db_elem_type = dbetStringArray, .elem_db_name = dbname, .elem_menu_name = mname, .minimum_version = ver_min, .maximum_version = ver_max, .changelog_mask = cl_mask, .data_is_mallocd = false, \
   .elem_data = { .elem_string_array = { .data = ptr, .num_elements = elems, .max_length = NULL, .default_value = NULL, .single_max_length = maxlen, .single_default = def, .previous_data = NULL } } }


/** Determine if a database element is valid at for a particular version of the database */
bool db_element_valid_for_version(const struct s_db_element *elem, uint32_t at_version);


/** Determine if a database element should appear in the change log, based on changelog masks */
bool db_element_changelog_visible(const struct s_db_element *elem, uint32_t logmask);


/** Determine if an database element has changed */
bool db_element_changed(const struct s_db_element *elem);


/** Write out the database element */
enum eDBError db_element_write(const char *db_name, const struct s_db_element *elem, struct s_db_interp *db_interp, uint32_t db_version);


/** Read in the database element */
enum eDBError db_element_read(const char *db_name, struct s_db_element *elem, struct s_db_interp *db_interp, uint32_t db_version);


/** Set a database element to defaults */
enum eDBError db_element_default(struct s_db_element *elem);


/** Update element so previous data matches current data */
enum eDBError db_element_update_previous(struct s_db_element *elem);


/** Revert element data to previous value */
enum eDBError db_element_revert(struct s_db_element *elem);


/** Work out the maximum size of an element */
enum eDBError db_element_get_maximum_size(struct s_db_element *elem, size_t *elem_max_size);


/**
 * Initialise a database element
 *
 * If a data element pointer is NULL (indicating it is no longer used in the current database version),
 * then it is allocated here, as we still need it for validating the checksum when reading database versions that is was contained in.
 * This also sets a flag so we know to free it when the element is destroyed.
 *
 * This also allocates memory for the previous value for some element types, and sets it to current value
 */
enum eDBError db_element_init(struct s_db_element *elem);


/** Destroy element (free memory previous value for some element types) */
enum eDBError db_element_destroy(struct s_db_element *elem);


#endif /* DB_ELEMENT_H_ */
