/*
 * db_interp_bin.c
 *
 *  Created on: 12/11/2010
 *      Author: ian
 */
#include <string.h>
#include <arpa/inet.h>
#include <rindb/db_buffer.h>
#include <rindb/db_interp.h>
#include <rindb/db_strncpy.h>

#include <rindb/db_interp_bin.h>


/** Function to export a unsigned 8 bit setting to a binary buffer */
static enum eDBError db_interp_bin_uint8_export(const char *db_name, const char *name, void *interpreter, uint8_t data)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;

  /* Store in buffer */
  return db_buffer_addbuffer(databuf, &data, sizeof(data));
}


/** Function to export an unsigned 32 bit integer setting to a binary buffer */
static enum eDBError db_interp_bin_uint32_export(const char *db_name, const char *name, void *interpreter, uint32_t data)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;

  /* Store in buffer, in network byte order so endian independent */
  data = htonl(data);
  return db_buffer_addbuffer(databuf, &data, sizeof(data));
}


/** Function to export a signed 32 bit integer setting to a binary buffer */
static enum eDBError db_interp_bin_int32_export(const char *db_name, const char *name, void *interpreter, int32_t data)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;

  /* Store in buffer, in network byte order so endian independent */
  data = htonl(data);
  return db_buffer_addbuffer(databuf, &data, sizeof(data));
}


/** Function to export a string (null terminated) setting in to a binary buffer */
static enum eDBError db_interp_bin_str_export(const char *db_name, const char *name, void *interpreter, const char *data)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  const size_t buflen = strlen(data) + 1;

  /* Store null terminated string in buffer, including the null terminator */
  return db_buffer_addbuffer(databuf, data, buflen);
}


/** Function to export a buffer setting in to a binary buffer */
static enum eDBError db_interp_bin_buf_export(const char *db_name, const char *name, void *interpreter, const uint8_t *data, size_t datalen)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t netlen;
  enum eDBError res;

  /* Store length in buffer, in network byte order so endian independent */
  netlen = htonl(datalen);
  res = db_buffer_addbuffer(databuf, &netlen, sizeof(netlen));

  /* If successful, write the binary buffer */
  if (dbeSuccess == res)
  {
    res = db_buffer_addbuffer(databuf, data, datalen);
  }
  return res;
}


/** Function to export a unsigned 8bit integer array setting in to a binary buffer */
static enum eDBError db_interp_bin_uint8_array_export(const char *db_name, const char *name, void *interpreter, const uint8_t *data, size_t num_elems)
{
  return db_interp_bin_buf_export(db_name, name, interpreter, data, num_elems * sizeof(uint8_t));
}


/** Function to export an unsigned 32bit integer array setting in to a binary buffer */
static enum eDBError db_interp_bin_uint32_array_export(const char *db_name, const char *name, void *interpreter, const uint32_t *data, size_t num_elems)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t netelems;
  uint32_t netdata;
  size_t i;
  enum eDBError res;

  /* Store number of elements in buffer, in network byte order so endian independent */
  netelems = htonl(num_elems);
  res = db_buffer_addbuffer(databuf, &netelems, sizeof(netelems));

  /* If successful, write each element in network byte order */
  if (dbeSuccess == res)
  {
    for (i = 0; i < num_elems; i++)
    {
      netdata = htonl(data[i]);
      res = db_buffer_addbuffer(databuf, &netdata, sizeof(netdata));
      if (dbeSuccess != res)
      {
        break;
      }
    }
  }
  return res;
}


/** Function to export an signed 32bit integer array setting in to a binary buffer */
static enum eDBError db_interp_bin_int32_array_export(const char *db_name, const char *name, void *interpreter, const int32_t *data, size_t num_elems)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t netelems;
  int32_t netdata;
  size_t i;
  enum eDBError res;

  /* Store number of elements in buffer, in network byte order so endian independent */
  netelems = htonl(num_elems);
  res = db_buffer_addbuffer(databuf, &netelems, sizeof(netelems));

  /* If successful, write each element in network byte order */
  if (dbeSuccess == res)
  {
    for (i = 0; i < num_elems; i++)
    {
      netdata = htonl(data[i]);
      res = db_buffer_addbuffer(databuf, &netdata, sizeof(netdata));
      if (dbeSuccess != res)
      {
        break;
      }
    }
  }
  return res;
}


/** Function to export a string (null terminated) array setting in to a binary buffer */
static enum eDBError db_interp_bin_string_array_export(const char *db_name, const char *name, void *interpreter, const char *const *data, size_t num_elems)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t netelems;
  size_t i;
  enum eDBError res;

  /* Store number of elements in buffer, in network byte order so endian independent */
  netelems = htonl(num_elems);
  res = db_buffer_addbuffer(databuf, &netelems, sizeof(netelems));

  /* If successful, write each string to the buffer */
  if (dbeSuccess == res)
  {
    for (i = 0; i < num_elems; i++)
    {
      res = db_interp_bin_str_export(db_name, name, interpreter, data[i]);
      if (dbeSuccess != res)
      {
        break;
      }
    }
  }
  return res;
}


/** Function to import an unsigned 8 bit setting from a buffer */
static enum eDBError db_interp_bin_uint8_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, uint8_t lowlim, uint8_t hilim, uint8_t default_value)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint8_t bsdata;
  enum eDBError res;

  res = db_buffer_getbuffer(databuf, &bsdata, sizeof(bsdata));
  if (dbeSuccess != res)
  {
    *data = default_value;
  }
  else
  {
    if (bsdata < lowlim || bsdata > hilim)
    {
      bsdata = default_value;
      res = dbeOutsideLimits;
    }
    *data = bsdata;
  }

  return res;
}


/** Function to import an unsigned 32 bit integer setting from a buffer */
static enum eDBError db_interp_bin_uint32_import(const char *db_name, const char *name, void *interpreter, uint32_t *data, uint32_t lowlim, uint32_t hilim, uint32_t default_value)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t bsdata;
  enum eDBError res;

  res = db_buffer_getbuffer(databuf, &bsdata, sizeof(bsdata));
  if (dbeSuccess != res)
  {
    *data = default_value;
  }
  else
  {
    bsdata = ntohl(bsdata);

    if (bsdata < lowlim || bsdata > hilim)
    {
      bsdata = default_value;
      res = dbeOutsideLimits;
    }
    *data = bsdata;
  }

  return res;
}


/** Function to import a signed 32 bit integer setting from a buffer */
static enum eDBError db_interp_bin_int32_import(const char *db_name, const char *name, void *interpreter, int32_t *data, int32_t lowlim, int32_t hilim, int32_t default_value)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  int32_t bsdata;
  enum eDBError res;

  res = db_buffer_getbuffer(databuf, &bsdata, sizeof(bsdata));
  if (dbeSuccess != res)
  {
    *data = default_value;
  }
  else
  {
    bsdata = ntohl(bsdata);

    if (bsdata < lowlim || bsdata > hilim)
    {
      bsdata = default_value;
      res = dbeOutsideLimits;
    }
    *data = bsdata;
  }

  return res;
}


/** Function to import a string (null terminated) setting from a buffer */
static enum eDBError db_interp_bin_str_import(const char *db_name, const char *name, void *interpreter, char *data, size_t maxlen, const char *default_value)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  enum eDBError res;
  size_t i;

  for (res = dbeOutsideLimits, i = 0; i < maxlen; i++)
  {
    res = db_buffer_getbyte(databuf, (uint8_t *) &data[i]);
    if (dbeSuccess != res)
    {
      break;
    }
    if ('\0' == data[i])
    {
      res = dbeSuccess;
      break;
    }
  }

  if (dbeSuccess != res)
  {
    db_safe_strncpy(data, default_value, maxlen);
  }

  return res;
}


/** Function to import a buffer setting from a buffer */
static enum eDBError db_interp_bin_buf_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, size_t maxlen, const uint8_t *default_value)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t buflen;
  enum eDBError res;
  size_t i;

  res = db_buffer_getbuffer(databuf, &buflen, sizeof(buflen));
  if (dbeSuccess == res)
  {
    buflen = ntohl(buflen);
    if (buflen <= maxlen)
    {
      /* Read 0..buflen from buffer */
      res = db_buffer_getbuffer(databuf, data, buflen);

      /* Default the rest */
      for (i = buflen; i < maxlen; i++)
      {
        data[i] = default_value[i];
      }
    }
    else
    {
      memcpy(data, default_value, maxlen);
      buflen = maxlen;
      res = dbeOutsideLimits;
    }
  }
  return res;
}


/** Function to import a unsigned 8bit integer array setting from a binary buffer */
enum eDBError db_interp_bin_uint8_array_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, size_t num_elems, const uint8_t *lowlims, const uint8_t *hilims, const uint8_t *default_values, uint8_t single_lowlim, uint8_t single_hilim, uint8_t single_default)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t netelems;
  size_t i;
  enum eDBError res;
  uint8_t min;
  uint8_t max;

  /* Read number of element from file */
  res = db_buffer_getbuffer(databuf, &netelems, sizeof(netelems));
  if (dbeSuccess == res)
  {
    netelems = ntohl(netelems);
    if (netelems <= num_elems)
    {
      for (i = 0; i < netelems; i++)
      {
        res = db_buffer_getbuffer(databuf, &data[i], sizeof(uint8_t));
        if (dbeSuccess != res)
          break;

        min = (NULL != lowlims) ? lowlims[i] : single_lowlim;
        max = (NULL != hilims) ? hilims[i] : single_hilim;
        if (data[i] < min || data[i] > max)
        {
          res = dbeOutsideLimits;
          break;
        }
      }
      for ( ; i < num_elems; i++)
      {
        data[i] = (NULL != default_values) ? default_values[i] : single_default;
      }
    }
    else
    {
      res = dbeOutsideLimits;
    }
  }

  if (dbeSuccess != res)
  {
    for (i = 0; i < num_elems; i++)
    {
      data[i] = (NULL != default_values) ? default_values[i] : single_default;
    }
  }
  return res;
}


/** Function to import an unsigned 32bit integer array setting from a binary buffer */
enum eDBError db_interp_bin_uint32_array_import(const char *db_name, const char *name, void *interpreter, uint32_t *data, size_t num_elems, const uint32_t *lowlims, const uint32_t *hilims, const uint32_t *default_values, uint32_t single_lowlim, uint32_t single_hilim, uint32_t single_default)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t netdata;
  uint32_t netelems;
  size_t i;
  enum eDBError res;
  uint32_t min;
  uint32_t max;

  /* Read number of element from file */
  res = db_buffer_getbuffer(databuf, &netelems, sizeof(netelems));
  if (dbeSuccess == res)
  {
    netelems = ntohl(netelems);
    if (netelems <= num_elems)
    {
      for (i = 0; i < netelems; i++)
      {
        res = db_buffer_getbuffer(databuf, &netdata, sizeof(netdata));
        if (dbeSuccess != res)
          break;
        data[i] = ntohl(netdata);

        min = (NULL != lowlims) ? lowlims[i] : single_lowlim;
        max = (NULL != hilims) ? hilims[i] : single_hilim;
        if (data[i] < min || data[i] > max)
        {
          res = dbeOutsideLimits;
          break;
        }
      }
      for ( ; i < num_elems; i++)
      {
        data[i] = (NULL != default_values) ? default_values[i] : single_default;
      }
    }
    else
    {
      res = dbeOutsideLimits;
    }
  }

  if (dbeSuccess != res)
  {
    for (i = 0; i < num_elems; i++)
    {
      data[i] = (NULL != default_values) ? default_values[i] : single_default;
    }
  }
  return res;
}


/** Function to import a signed 32bit integer array setting from a binary buffer */
enum eDBError db_interp_bin_int32_array_import(const char *db_name, const char *name, void *interpreter, int32_t *data, size_t num_elems, const int32_t *lowlims, const int32_t *hilims, const int32_t *default_values, int32_t single_lowlim, int32_t single_hilim, int32_t single_default)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  int32_t netdata;
  uint32_t netelems;
  size_t i;
  enum eDBError res;
  int32_t min;
  int32_t max;

  /* Read number of element from file */
  res = db_buffer_getbuffer(databuf, &netelems, sizeof(netelems));
  if (dbeSuccess == res)
  {
    netelems = ntohl(netelems);
    if (netelems <= num_elems)
    {
      for (i = 0; i < netelems; i++)
      {
        res = db_buffer_getbuffer(databuf, &netdata, sizeof(netdata));
        if (dbeSuccess != res)
          break;
        data[i] = ntohl(netdata);

        min = (NULL != lowlims) ? lowlims[i] : single_lowlim;
        max = (NULL != hilims) ? hilims[i] : single_hilim;
        if (data[i] < min || data[i] > max)
        {
          res = dbeOutsideLimits;
          break;
        }
      }
      for ( ; i < num_elems; i++)
      {
        data[i] = (NULL != default_values) ? default_values[i] : single_default;
      }
    }
    else
    {
      res = dbeOutsideLimits;
    }
  }

  if (dbeSuccess != res)
  {
    for (i = 0; i < num_elems; i++)
    {
      data[i] = (NULL != default_values) ? default_values[i] : single_default;
    }
  }
  return res;
}


static void db_interp_bin_string_array_default(char *const *data, size_t first_elem, size_t num_elems, const char *const *default_values, const char *single_default, const size_t *max_lengths, size_t single_max_length)
{
  size_t i;
  size_t maxlen;
  const char *def;

  for (i = first_elem; i < num_elems; i++)
  {
    maxlen = (NULL != max_lengths) ? max_lengths[i] : single_max_length;
    def = (NULL != default_values && NULL != default_values[i]) ? default_values[i] : single_default;
    db_safe_strncpy(data[i], def, maxlen);
  }
}


/** Function to import a string array setting from a binary buffer */
enum eDBError db_interp_bin_string_array_import(const char *db_name, const char *name, void *interpreter, char *const *data, size_t num_elems, const size_t *max_lengths, const char *const *default_values, size_t single_max_length, const char *single_default)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  uint32_t netelems;
  size_t i;
  enum eDBError res;
  size_t maxlen;
  const char *def;

  /* Read number of element from file */
  res = db_buffer_getbuffer(databuf, &netelems, sizeof(netelems));
  if (dbeSuccess == res)
  {
    netelems = ntohl(netelems);
    if (netelems <= num_elems)
    {
      /* Read number of elements from file */
      for (i = 0; i < netelems; i++)
      {
        maxlen = (NULL != max_lengths) ? max_lengths[i] : single_max_length;
        def = (NULL != default_values && NULL != default_values[i]) ? default_values[i] : single_default;

        res = db_interp_bin_str_import(db_name, name, interpreter, data[i], maxlen, def);
        if (dbeSuccess != res)
          break;
      }

      /* Default the rest */
      db_interp_bin_string_array_default(data, i, num_elems, default_values, single_default, max_lengths, single_max_length);
    }
    else
    {
      res = dbeOutsideLimits;
    }
  }

  if (dbeSuccess != res)
  {
    db_interp_bin_string_array_default(data, 0, num_elems, default_values, single_default, max_lengths, single_max_length);
  }
  return res;
}


/** Create an interpreter for a binary output buffer */
enum eDBError db_create_interp_binary(struct s_db_interp *db_interp, struct s_db_buffer *db_buffer)
{
  db_interp->interpreter = (void *) db_buffer;

  /* Export functions */
  db_interp->uint8_export = db_interp_bin_uint8_export;
  db_interp->uint32_export = db_interp_bin_uint32_export;
  db_interp->int32_export = db_interp_bin_int32_export;
  db_interp->str_export = db_interp_bin_str_export;
  db_interp->buf_export = db_interp_bin_buf_export;
  db_interp->uint8_array_export = db_interp_bin_uint8_array_export;
  db_interp->uint32_array_export = db_interp_bin_uint32_array_export;
  db_interp->int32_array_export = db_interp_bin_int32_array_export;
  db_interp->string_array_export = db_interp_bin_string_array_export;

  /* Import functions */
  db_interp->uint8_import = db_interp_bin_uint8_import;
  db_interp->uint32_import = db_interp_bin_uint32_import;
  db_interp->int32_import = db_interp_bin_int32_import;
  db_interp->str_import = db_interp_bin_str_import;
  db_interp->buf_import = db_interp_bin_buf_import;
  db_interp->uint8_array_import = db_interp_bin_uint8_array_import;
  db_interp->uint32_array_import = db_interp_bin_uint32_array_import;
  db_interp->int32_array_import = db_interp_bin_int32_array_import;
  db_interp->string_array_import = db_interp_bin_string_array_import;

  return dbeSuccess;
}
