/*
 * db_interp_uboot.c
 *
 *  Created on: Sep 13, 2012
 *      Author: ian
 */
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <rindb/db_interp.h>
#include <rindb/db_base64.h>
#include <rindb/db_strncpy.h>
#include <rindb/db_strescape.h>
#include <rindb/db_buffer.h>

#include <rindb/db_interp_uboot.h>


/*
 * Export
 */


/** Function to export the "<facdb>_<varname> " part of a u-boot variable */
static enum eDBError db_interp_uboot_str_export_varname(const char *db_name, const char *name, void *interpreter)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  enum eDBError res;
  char ch;

  res = db_buffer_addbuffer(databuf, db_name, strlen(db_name));
  if (dbeSuccess != res)
    goto diusev_fail;
  ch = '_';
  res = db_buffer_addbuffer(databuf, &ch, sizeof(ch));
  if (dbeSuccess != res)
    goto diusev_fail;
  res = db_buffer_addbuffer(databuf, name, strlen(name));
  if (dbeSuccess != res)
    goto diusev_fail;
  ch = ' ';
  res = db_buffer_addbuffer(databuf, &ch, sizeof(ch));
  if (dbeSuccess != res)
    goto diusev_fail;

diusev_fail:
  return res;
}


/** Function to export a string (null terminated) setting to u-boot environment variables */
static enum eDBError db_interp_uboot_str_export_internal(const char *db_name, const char *data, void *interpreter, const char *name)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  enum eDBError res;
  char ch;

  res = db_interp_uboot_str_export_varname(db_name, name, interpreter);
  if (dbeSuccess != res)
    goto diussv_fail;
  res = db_buffer_addbuffer(databuf, data, strlen(data));
  if (dbeSuccess != res)
    goto diussv_fail;
  ch = '\n';
  res = db_buffer_addbuffer(databuf, &ch, sizeof(ch));
  if (dbeSuccess != res)
    goto diussv_fail;

diussv_fail:
  return res;
}


/** Function to export an unsigned 32bit integer setting to u-boot environment variables */
static enum eDBError db_interp_uboot_uint32_export(const char *db_name, const char *name, void *interpreter, uint32_t data)
{
  char str[30];

  snprintf(str, 30, "%lu", (unsigned long) data);
  return db_interp_uboot_str_export_internal(db_name, str, interpreter, name);
}


/** Function to export a unsigned 8bit integer setting to u-boot environment variables */
static enum eDBError db_interp_uboot_uint8_export(const char *db_name, const char *name, void *interpreter, uint8_t data)
{
  return db_interp_uboot_uint32_export(db_name, name, interpreter, data);
}


/** Function to export a signed 32bit integer setting to u-boot environment variables */
static enum eDBError db_interp_uboot_int32_export(const char *db_name, const char *name, void *interpreter, int32_t data)
{
  char str[30];

  snprintf(str, 30, "%ld", (long) data);
  return db_interp_uboot_str_export_internal(db_name, str, interpreter, name);
}


/** Function to translate a buffer to base64 into an allocated buffer. Caller is expected to free if result is dbeSuccess */
static enum eDBError db_interp_uboot_buf_to_base64_alloc(const uint8_t *data, size_t datalen, char **encdata)
{
  char *buf = NULL;
  int buflen;
  enum eDBError res;

  /* Allocate a buffer for the base64 encoded data */
  buflen = datalen * 3;
  buf = malloc(buflen);
  if (NULL == buf)
  {
    res = dbeMemAllocFail;
    goto dbubtba_fail;
  }

  /* Encode data in base64 */
  res = db_base64_encode(data, datalen, buf, buflen);
  if (dbeSuccess != res)
  {
    goto dbubtba_fail;
  }

  /* Success */
  res = dbeSuccess;

dbubtba_fail:
  if (dbeSuccess != res)
  {
    free(buf);
    buf = NULL;
  }
  *encdata = buf;

  return res;
}


/** Function to export a buffer setting to u-boot environment variables */
static enum eDBError db_interp_uboot_buf_export(const char *db_name, const char *name, void *interpreter, const uint8_t *data, size_t datalen)
{
  char *buf = NULL;
  int buflen;
  enum eDBError res;

  /* Allocate a buffer for the base64 encoded data */
  buflen = datalen * 3;
  buf = malloc(buflen);
  if (NULL == buf)
  {
    res = dbeMemAllocFail;
    goto dixbe_done;
  }

  /* Encode data in base64 */
  res = db_base64_encode(data, datalen, buf, buflen);
  if (dbeSuccess != res)
  {
    goto dixbe_done;
  }

  /* Export the base63 encoded string */
  res = db_interp_uboot_str_export_internal(db_name, buf, interpreter, name);
  if (dbeSuccess != res)
  {
    goto dixbe_done;
  }

  /* Success */
  res = dbeSuccess;

dixbe_done:
  if (NULL != buf)
    free(buf);
  return res;
}


/** Function to export a string (null terminated) setting to u-boot environment variables */
static enum eDBError db_interp_uboot_str_export(const char *db_name, const char *name, void *interpreter, const char *data)
{
  size_t length;

  length = strlen(data);
  return db_interp_uboot_buf_export(db_name, name, interpreter, (const uint8_t *) data, length);
}


/** Function to begin the export of an array to u-boot environment variables */
static enum eDBError db_interp_uboot_array_begin_export(const char *db_name, void *interpreter, const char *name)
{
  enum eDBError res;

  res = db_interp_uboot_str_export_varname(db_name, name, interpreter);
  return res;
}


/** Function to export an element of an array to u-boot environment variables */
static enum eDBError db_interp_uboot_array_export(const char *data, void *interpreter, bool islast)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  enum eDBError res;
  char ch;

  res = db_buffer_addbuffer(databuf, data, strlen(data));
  if (dbeSuccess != res)
    goto diuafe_fail;
  if (!islast)
  {
    ch = ',';
    res = db_buffer_addbuffer(databuf, &ch, sizeof(ch));
    if (dbeSuccess != res)
      goto diuafe_fail;
  }

diuafe_fail:
  return res;
}


/** Function to begin the export of an array to u-boot environment variables */
static enum eDBError db_interp_uboot_array_end_export(void *interpreter)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  char ch;

  ch = '\n';
  return db_buffer_addbuffer(databuf, &ch, sizeof(ch));
}


/** Function to export a unsigned 8bit integer array setting to u-boot environment variables */
static enum eDBError db_interp_uboot_uint8_array_export(const char *db_name, const char *name, void *interpreter, const uint8_t *data, size_t num_elems)
{
  size_t i;
  enum eDBError res;
  char str[40];

  /* Add name to export */
  res = db_interp_uboot_array_begin_export(db_name, interpreter, name);
  if (dbeSuccess == res)
  {
    for (i = 0; i < num_elems; i++)
    {
      snprintf(str, 40, "%u", data[i]);
      res = db_interp_uboot_array_export(str, interpreter, (i >= num_elems - 1));
      if (dbeSuccess != res)
      {
        break;
      }
    }
    if (dbeSuccess == res)
    {
      res = db_interp_uboot_array_end_export(interpreter);
    }
  }
  return res;
}


/** Function to export an unsigned 32bit integer array setting to u-boot environment variables */
static enum eDBError db_interp_uboot_uint32_array_export(const char *db_name, const char *name, void *interpreter, const uint32_t *data, size_t num_elems)
{
  size_t i;
  enum eDBError res;
  char str[40];

  /* Create array root node */
  res = db_interp_uboot_array_begin_export(db_name, interpreter, name);
  if (dbeSuccess == res)
  {
    for (i = 0; i < num_elems; i++)
    {
      snprintf(str, 40, "%lu", (unsigned long) data[i]);
      res = db_interp_uboot_array_export(str, interpreter, (i >= num_elems - 1));
      if (dbeSuccess != res)
      {
        break;
      }
    }
    if (dbeSuccess == res)
    {
      res = db_interp_uboot_array_end_export(interpreter);
    }
  }
  return res;
}


/** Function to export a signed 32bit integer array setting to u-boot environment variables */
static enum eDBError db_interp_uboot_int32_array_export(const char *db_name, const char *name, void *interpreter, const int32_t *data, size_t num_elems)
{
  size_t i;
  enum eDBError res;
  char str[40];

  /* Create array root node */
  res = db_interp_uboot_array_begin_export(db_name, interpreter, name);
  if (dbeSuccess == res)
  {
    for (i = 0; i < num_elems; i++)
    {
      snprintf(str, 40, "%ld", (long) data[i]);
      res = db_interp_uboot_array_export(str, interpreter, (i >= num_elems - 1));
      if (dbeSuccess != res)
      {
        break;
      }
    }
    if (dbeSuccess == res)
    {
      res = db_interp_uboot_array_end_export(interpreter);
    }
  }
  return res;
}


/** Function to export a string array setting to u-boot environment variables */
static enum eDBError db_interp_uboot_string_array_export(const char *db_name, const char *name, void *interpreter, const char *const *data, size_t num_elems)
{
  enum eDBError res;
  char *encdata;
  size_t i;

  /* Output name */
  res = db_interp_uboot_array_begin_export(db_name, interpreter, name);
  if (dbeSuccess == res)
  {
    for (i = 0, encdata = NULL; i < num_elems; i++)
    {
      /* Base64 encode value */
      res = db_interp_uboot_buf_to_base64_alloc((uint8_t *) data[i], strlen(data[i]), &encdata);
      if (dbeSuccess == res && NULL != encdata)
      {
        /* Export value */
        res = db_interp_uboot_array_export(encdata, interpreter, (i >= num_elems - 1));
        free(encdata);
        encdata = NULL;
        if (dbeSuccess != res)
        {
          break;
        }
      }
    } /* for */
    if (dbeSuccess == res)
    {
      res = db_interp_uboot_array_end_export(interpreter);
    }
  }
  return res;
}


/*
 * Import
 */


enum eParseStatus { psFindNewline, psInDbName, psUnderscore, psInName, psEquals, psInData, psDone, psDoneNotFound };


/** Find a u-boot environment name (name=value), and copy the data value to outbuf */
static enum eDBError db_interp_uboot_get_value_for_name(const char *db_name, void *interpreter, const char *name, struct s_db_buffer *outbuf)
{
  struct s_db_buffer *databuf = (struct s_db_buffer *) interpreter;
  enum eParseStatus ps;
  enum eDBError res;
  size_t name_pos;
  uint8_t ch;

  /* Back to start of buffer */
  res = db_buffer_rewind(databuf);
  if (dbeSuccess != res)
    goto diugvfn_fail;

  for (ps = psInDbName, name_pos = 0, res = dbeSuccess; (psDone != ps && psDoneNotFound != ps); )
  {
    res = db_buffer_getbyte(databuf, &ch);
    if (dbeSuccess != res)
      ps = psDoneNotFound;

    switch (ps)
    {
      case psFindNewline:
        /* Look for the next newline, <db_name>_<name> follows it immediately */
        if ('\n' == ch)
        {
          ps = psInDbName;
          name_pos = 0;
        }
        break;

      case psInDbName:
        /* Check against db_name */
        if (db_name[name_pos] == ch)
        {
          name_pos++;
          if ('\0' == db_name[name_pos])
          {
            ps = psUnderscore;
          }
        }
        else
        {
          ps = psFindNewline;
        }
        break;

      case psUnderscore:
        /* Check for '_', if no match, then this isn't our variable */
        if ('_' == ch)
        {
          ps = psInName;
          name_pos = 0;
        }
        else
          ps = psFindNewline;
        break;

      case psInName:
        /* Check against name */
        if (name[name_pos] == ch)
        {
          name_pos++;
          if ('\0' == name[name_pos])
          {
            ps = psEquals;
          }
        }
        else
        {
          ps = psFindNewline;
        }
        break;

      case psEquals:
        /* Check for = sign, if no match, then this isn't our variable */
        if ('=' == ch)
        {
          ps = psInData;
        }
        else
          ps = psFindNewline;
        break;

      case psInData:
        /* Copy data to output */
        if ('\n' == ch || '\0' == ch)
          ps = psDone;
        else
        {
          res = db_buffer_addbyte(outbuf, ch);
          if (dbeSuccess != res)
          {
            ps = psDone;
          }
        }
        break;

      default:
        break;
    }
  }

  if (psDone == ps)
  {
    if (dbeSuccess == res)
    {
      res = db_buffer_addbyte(outbuf, '\0');
    }
  }
  else
  {
    res = dbeParseFail;
  }

diugvfn_fail:
  return res;
}


static enum eDBError db_interp_uboot_uint32_get_value(const char *db_name, void *interpreter, const char *name, uint32_t *data, uint32_t default_value, uint32_t minumum, uint32_t maximum)
{
  enum eDBError res;
  unsigned long value;
  char str[40];
  struct s_db_buffer buffer;

  res = db_buffer_create_static(&buffer, (uint8_t *) str, 40);
  if (dbeSuccess != res)
  {
    *data = default_value;
    return res;
  }

  res = db_interp_uboot_get_value_for_name(db_name, interpreter, name, &buffer);
  if (dbeSuccess != res)
  {
    *data = default_value;
    return res;
  }

  if (1 != sscanf((char *) buffer.buffer, "%lu", &value))
  {
    *data = default_value;
    return dbeParseFail;
  }

  if (value < minumum || value > maximum)
  {
    *data = default_value;
    return dbeOutsideLimits;
  }

  /* Success */
  *data = value;
  return dbeSuccess;
}


static enum eDBError db_interp_uboot_int32_get_value(const char *db_name, void *interpreter, const char *name, int32_t *data, int32_t default_value, int32_t minumum, int32_t maximum)
{
  enum eDBError res;
  long value;
  char str[40];
  struct s_db_buffer buffer;

  res = db_buffer_create_static(&buffer, (uint8_t *) str, 40);
  if (dbeSuccess != res)
  {
    *data = default_value;
    return res;
  }

  res = db_interp_uboot_get_value_for_name(db_name, interpreter, name, &buffer);
  if (dbeSuccess != res)
  {
    *data = default_value;
    return res;
  }

  if (1 != sscanf((char *) buffer.buffer, "%ld", &value))
  {
    *data = default_value;
    res = dbeParseFail;
  }

  if (value < minumum || value > maximum)
  {
    *data = default_value;
    res = dbeOutsideLimits;
  }

  /* Success */
  *data = value;
  return dbeSuccess;
}


/** Function to import an unsigned 32bit integer setting from u-boot environment variables */
static enum eDBError db_interp_uboot_uint32_import(const char *db_name, const char *name, void *interpreter, uint32_t *data, uint32_t lowlim, uint32_t hilim, uint32_t default_value)
{
  return db_interp_uboot_uint32_get_value(db_name, interpreter, name, data, default_value, lowlim, hilim);
}


/** Function to import an unsigned 8bit integer setting from u-boot environment variables */
static enum eDBError db_interp_uboot_uint8_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, uint8_t lowlim, uint8_t hilim, uint8_t default_value)
{
  uint32_t intdata;
  enum eDBError result;

  result = db_interp_uboot_uint32_get_value(db_name, interpreter, name, &intdata, default_value, lowlim, hilim);
  *data = (uint8_t) intdata;
  return result;
}


/** Function to import an integer setting from u-boot environment variables */
static enum eDBError db_interp_uboot_int32_import(const char *db_name, const char *name, void *interpreter, int32_t *data, int32_t lowlim, int32_t hilim, int32_t default_value)
{
  return db_interp_uboot_int32_get_value(db_name, interpreter, name, data, default_value, lowlim, hilim);
}


/** Function to import a buffer setting from u-boot environment variables */
static enum eDBError db_interp_uboot_buf_import_internal(const char *db_name, uint8_t *data, void *interpreter, const char *name, size_t maxlen, const uint8_t *default_value, size_t *outlen)
{
  enum eDBError res;
  size_t length;
  size_t buflen;
  struct s_db_buffer buffer;

  buffer.buffer = NULL;

  /* Allocate a buffer for the base64 encoded data */
  buflen = maxlen * 3;
  res = db_buffer_create_malloc(&buffer, buflen, buflen);
  if (dbeSuccess != res)
  {
    goto diubi_fail;
  }

  /* Get base64 encoded data */
  res = db_interp_uboot_get_value_for_name(db_name, interpreter, name, &buffer);
  if (dbeSuccess != res)
  {
    /* Fail */
    goto diubi_fail;
  }

  /* Decode base64 data */
  res = db_base64_decode((char *) buffer.buffer, strlen((char *) buffer.buffer), data, maxlen, &length);
  if (dbeSuccess != res)
  {
    /* Fail */
    goto diubi_fail;
  }

diubi_fail:
  if (dbeSuccess != res)
  {
    /* Set to default */
    memcpy(data, default_value, maxlen);
    if (NULL != outlen)
      *outlen = maxlen;
  }
  else
  {
    if (NULL != outlen)
      *outlen = length;
  }

  if (NULL != buffer.buffer)
  {
    db_buffer_destroy(&buffer);
  }

  return res;
}

/** Function to import a buffer setting from u-boot environment variables */
static enum eDBError db_interp_uboot_buf_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, size_t maxlen, const uint8_t *default_value)
{
  return db_interp_uboot_buf_import_internal(db_name, data, interpreter, name, maxlen, default_value, NULL);
}


/** Function to import a string (null terminated) setting from u-boot environment variables */
static enum eDBError db_interp_uboot_str_import(const char *db_name, const char *name, void *interpreter, char *data, size_t maxlen, const char *default_value)
{
  size_t len;
  enum eDBError res;

  res = db_interp_uboot_buf_import_internal(db_name, (uint8_t *) data, interpreter, name, maxlen, (const uint8_t *) default_value, &len);
  if (dbeSuccess == res && len < maxlen)
  {
    data[len] = '\0';
  }
  else
  {
    data[maxlen - 1] = '\0';
    res = dbeParseFail;
  }
  return res;
}


static void db_interp_uboot_uint8_array_default(uint8_t *data, size_t first_elem, size_t num_elems, const uint8_t *default_values, uint8_t single_default)
{
  size_t i;

  for (i = first_elem; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
}


/** Function to import a unsigned 8bit integer array setting from u-boot environment variables */
static enum eDBError db_interp_uboot_uint8_array_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, size_t num_elems, const uint8_t *lowlims, const uint8_t *hilims, const uint8_t *default_values, uint8_t single_lowlim, uint8_t single_hilim, uint8_t single_default)
{
  enum eDBError res;
  struct s_db_buffer buffer;
  size_t idx;
  uint8_t min;
  uint8_t max;
  char *ptr;
  int chars_parsed;
  unsigned long num;

  buffer.buffer = NULL;

  /* Allocate a buffer for the array of comma separated numbers */
  res = db_buffer_create_malloc(&buffer, 1024, 1024);
  if (dbeSuccess != res)
  {
    goto diuuai_fail;
  }

  /* Get comma separated numbers for name */
  res = db_interp_uboot_get_value_for_name(db_name, interpreter, name, &buffer);
  if (dbeSuccess != res)
  {
    /* Fail */
    goto diuuai_fail;
  }

  /* Parse each number */
  ptr = (char *) buffer.buffer;
  idx = 0;
  res = dbeSuccess;
  while (1 == sscanf(ptr, "%lu%n", &num, &chars_parsed) && idx < num_elems)
  {
    min = (NULL != lowlims) ? lowlims[idx] : single_lowlim;
    max = (NULL != hilims) ? hilims[idx] : single_hilim;

    if (num < min || num > max)
    {
      data[idx] = (NULL != default_values) ? default_values[idx] : single_default;
      res = dbeOutsideLimits;
    }
    else
    {
      data[idx] = num;
    }

    ptr += chars_parsed;
    if (',' != *ptr && '\n' != *ptr && '\0' != *ptr)
      break;
    ptr++;
    idx++;
  }

  if (dbeSuccess == res && idx != num_elems)
  {
    res = dbeParseFail;
  }

  /* Default any others */
  db_interp_uboot_uint8_array_default(data, idx, num_elems, default_values, single_default);

  /* Clean up */
  if (NULL != buffer.buffer)
    db_buffer_destroy(&buffer);

  return res;

diuuai_fail:
  /* Default everything, and clean up */
  db_interp_uboot_uint8_array_default(data, 0, num_elems, default_values, single_default);
  if (NULL != buffer.buffer)
    db_buffer_destroy(&buffer);

  return res;
}


static void db_interp_uboot_uint32_array_default(uint32_t *data, size_t first_elem, size_t num_elems, const uint32_t *default_values, uint32_t single_default)
{
  size_t i;

  for (i = first_elem; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
}


/** Function to import an unsigned 32bit integer array setting from u-boot environment variables */
static enum eDBError db_interp_uboot_uint32_array_import(const char *db_name, const char *name, void *interpreter, uint32_t *data, size_t num_elems, const uint32_t *lowlims, const uint32_t *hilims, const uint32_t *default_values, uint32_t single_lowlim, uint32_t single_hilim, uint32_t single_default)
{
  enum eDBError res;
  struct s_db_buffer buffer;
  size_t idx;
  uint32_t min;
  uint32_t max;
  char *ptr;
  int chars_parsed;
  unsigned long num;

  buffer.buffer = NULL;

  /* Allocate a buffer for the array of comma separated numbers */
  res = db_buffer_create_malloc(&buffer, 1024, 1024);
  if (dbeSuccess != res)
  {
    goto diuuai_fail;
  }

  /* Get comma separated numbers for name */
  res = db_interp_uboot_get_value_for_name(db_name, interpreter, name, &buffer);
  if (dbeSuccess != res)
  {
    /* Fail */
    goto diuuai_fail;
  }

  /* Parse each number */
  ptr = (char *) buffer.buffer;
  idx = 0;
  res = dbeSuccess;
  while (1 == sscanf(ptr, "%lu%n", &num, &chars_parsed) && idx < num_elems)
  {
    min = (NULL != lowlims) ? lowlims[idx] : single_lowlim;
    max = (NULL != hilims) ? hilims[idx] : single_hilim;

    if (num < min || num > max)
    {
      data[idx] = (NULL != default_values) ? default_values[idx] : single_default;
      res = dbeOutsideLimits;
    }
    else
    {
      data[idx] = num;
    }

    ptr += chars_parsed;
    if (',' != *ptr && '\n' != *ptr && '\0' != *ptr)
      break;
    ptr++;
    idx++;
  }

  if (dbeSuccess == res && idx != num_elems)
  {
    res = dbeParseFail;
  }

  /* Default any others */
  db_interp_uboot_uint32_array_default(data, idx, num_elems, default_values, single_default);

  /* Clean up */
  if (NULL != buffer.buffer)
    db_buffer_destroy(&buffer);

  return res;

diuuai_fail:
  /* Default everything, and clean up */
  db_interp_uboot_uint32_array_default(data, 0, num_elems, default_values, single_default);
  if (NULL != buffer.buffer)
    db_buffer_destroy(&buffer);

  return res;
}


static void db_interp_uboot_int32_array_default(int32_t *data, size_t first_elem, size_t num_elems, const int32_t *default_values, int32_t single_default)
{
  size_t i;

  for (i = first_elem; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
}


/** Function to import a signed 32bit integer array setting from u-boot environment variables */
static enum eDBError db_interp_uboot_int32_array_import(const char *db_name, const char *name, void *interpreter, int32_t *data, size_t num_elems, const int32_t *lowlims, const int32_t *hilims, const int32_t *default_values, int32_t single_lowlim, int32_t single_hilim, int32_t single_default)
{
  enum eDBError res;
  struct s_db_buffer buffer;
  size_t idx;
  int32_t min;
  int32_t max;
  char *ptr;
  int chars_parsed;
  long num;

  buffer.buffer = NULL;

  /* Allocate a buffer for the array of comma separated numbers */
  res = db_buffer_create_malloc(&buffer, 1024, 1024);
  if (dbeSuccess != res)
  {
    goto diuuai_fail;
  }

  /* Get comma separated numbers for name */
  res = db_interp_uboot_get_value_for_name(db_name, interpreter, name, &buffer);
  if (dbeSuccess != res)
  {
    /* Fail */
    goto diuuai_fail;
  }

  /* Parse each number */
  ptr = (char *) buffer.buffer;
  idx = 0;
  res = dbeSuccess;
  while (1 == sscanf(ptr, "%ld%n", &num, &chars_parsed) && idx < num_elems)
  {
    min = (NULL != lowlims) ? lowlims[idx] : single_lowlim;
    max = (NULL != hilims) ? hilims[idx] : single_hilim;

    if (num < min || num > max)
    {
      data[idx] = (NULL != default_values) ? default_values[idx] : single_default;
      res = dbeOutsideLimits;
    }
    else
    {
      data[idx] = num;
    }

    ptr += chars_parsed;
    if (',' != *ptr && '\n' != *ptr && '\0' != *ptr)
      break;
    ptr++;
    idx++;
  }

  if (dbeSuccess == res && idx != num_elems)
  {
    res = dbeParseFail;
  }

  /* Default any others */
  db_interp_uboot_int32_array_default(data, idx, num_elems, default_values, single_default);

  /* Clean up */
  if (NULL != buffer.buffer)
    db_buffer_destroy(&buffer);

  return res;

diuuai_fail:
  /* Default everything, and clean up */
  db_interp_uboot_int32_array_default(data, 0, num_elems, default_values, single_default);
  if (NULL != buffer.buffer)
    db_buffer_destroy(&buffer);

  return res;
}


static void db_interp_uboot_string_array_default(char *const *data, size_t first_elem, size_t num_elems, const char *const *default_values, const char *single_default, const size_t *max_lengths, size_t single_max_length)
{
  size_t i;
  size_t maxlen;
  const char *def;

  for (i = first_elem; i < num_elems; i++)
  {
    maxlen = (NULL != max_lengths) ? max_lengths[i] : single_max_length;
    def = (NULL != default_values && NULL != default_values[i]) ? default_values[i] : single_default;
    db_safe_strncpy(data[i], def, maxlen);
  }
}


/** Function to import a string array setting from u-boot environment variables */
static enum eDBError db_interp_uboot_string_array_import(const char *db_name, const char *name, void *interpreter, char *const *data, size_t num_elems, const size_t *max_lengths, const char *const *default_values, size_t single_max_length, const char *single_default)
{
  enum eDBError res;
  struct s_db_buffer buffer;
  size_t idx;
  char *ptr;
  char *ptr_next;
  ptrdiff_t inlen;
  size_t outmaxlen;
  size_t outlen;

  buffer.buffer = NULL;

  /* Allocate a buffer for the array of comma separated base64 encoded strings */
  res = db_buffer_create_malloc(&buffer, 1024, 1024);
  if (dbeSuccess != res)
  {
    goto diuuai_fail;
  }

  /* Get comma separated base64 encoded strings for name */
  res = db_interp_uboot_get_value_for_name(db_name, interpreter, name, &buffer);
  if (dbeSuccess != res)
  {
    /* Fail */
    goto diuuai_fail;
  }

  /* Parse each string */
  for (idx = 0, ptr = (char *) buffer.buffer; idx < num_elems; idx++)
  {
    /* Look for comma separation */
    ptr_next = strchr(ptr, ',');
    if (NULL == ptr_next)
    {
      /* Look for end of string */
      ptr_next = strchr(ptr, '\0');
      if (NULL == ptr_next)
      {
        /* Fail */
        res = dbeParseFail;
        break;
      }
    }

    /* Work out lengths */
    inlen = ptr_next - ptr;
    outmaxlen = (NULL != max_lengths) ? max_lengths[idx] : single_max_length;

    /* Decode base64 */
    res = db_base64_decode(ptr, inlen, (uint8_t *) data[idx], outmaxlen, &outlen);
    if (dbeSuccess != res)
    {
      break;
    }

    /* Add NULL termination to string */
    if (outlen < outmaxlen)
    {
      data[idx][outlen] = '\0';
    }
    else
    {
      res = dbeParseFail;
      break;
    }

    /* Move to next string after the comma */
    ptr = &ptr_next[1];
  }

  /* Default any others */
  db_interp_uboot_string_array_default(data, idx, num_elems, default_values, single_default, max_lengths, single_max_length);

  /* Clean up */
  if (NULL != buffer.buffer)
    db_buffer_destroy(&buffer);

  return res;

diuuai_fail:
  /* Default everything, and clean up */
  db_interp_uboot_string_array_default(data, 0, num_elems, default_values, single_default, max_lengths, single_max_length);
  if (NULL != buffer.buffer)
    db_buffer_destroy(&buffer);

  return res;
}


/** Create an interpreter for uboot environment variables */
enum eDBError db_create_interp_uboot(struct s_db_interp *db_interp, struct s_db_buffer *db_buffer)
{
  db_interp->interpreter = (void *) db_buffer;

  /* Export functions */
  db_interp->uint8_export = db_interp_uboot_uint8_export;
  db_interp->uint32_export = db_interp_uboot_uint32_export;
  db_interp->int32_export = db_interp_uboot_int32_export;
  db_interp->str_export = db_interp_uboot_str_export;
  db_interp->buf_export = db_interp_uboot_buf_export;
  db_interp->uint8_array_export = db_interp_uboot_uint8_array_export;
  db_interp->uint32_array_export = db_interp_uboot_uint32_array_export;
  db_interp->int32_array_export = db_interp_uboot_int32_array_export;
  db_interp->string_array_export = db_interp_uboot_string_array_export;

  /* Import functions */
  db_interp->uint8_import = db_interp_uboot_uint8_import;
  db_interp->uint32_import = db_interp_uboot_uint32_import;
  db_interp->int32_import = db_interp_uboot_int32_import;
  db_interp->str_import = db_interp_uboot_str_import;
  db_interp->buf_import = db_interp_uboot_buf_import;
  db_interp->uint8_array_import = db_interp_uboot_uint8_array_import;
  db_interp->uint32_array_import = db_interp_uboot_uint32_array_import;
  db_interp->int32_array_import = db_interp_uboot_int32_array_import;
  db_interp->string_array_import = db_interp_uboot_string_array_import;

  return dbeSuccess;
}
