/*
 * db_interp_xml.c
 *
 *  Created on: 12/11/2010
 *      Author: ian
 */
#include <string.h>
#include <stdint.h>
#include <ezxml/ezxml.h>
#include <rindb/db_interp.h>
#include <rindb/db_base64.h>
#include <rindb/db_strncpy.h>
#include <rindb/db_strescape.h>

#include <rindb/db_interp_xml.h>


/** Element name */
#define DB_ELEMENT_ELEM_NAME "elem"


/** Element attribute -> name */
#define DB_ELEMENT_ELEM_ATTRNAME "name"


#define DB_ELEMENT_ARRAY_ELEM_NAME "arrayelem"
#define DB_ELEMENT_ARRAY_ATTRINDEX "index"
#define DB_ELEMENT_ARRAY_ATTRNUMELEMENTS "numelems"


/** Get an element by searching name attribute of all elem elements */
static enum eDBError db_interp_xml_elem_get(ezxml_t xmlsection, const char *elemname, ezxml_t *node)
{
  ezxml_t iter;
  const char *current_elem_attr_name;

  if (NULL == xmlsection || NULL == node)
  {
    return dbeNullPointer;
  }

  /* For each "elem" node under xmlsection */
  for (iter = ezxml_child(xmlsection, DB_ELEMENT_ELEM_NAME); ((ezxml_t) 0 != iter); iter = iter->next)
  {
    /* Check if attribute "name" == "elemname" */
    current_elem_attr_name = ezxml_attr(iter, DB_ELEMENT_ELEM_ATTRNAME);
    if (NULL != current_elem_attr_name && 0 == strcmp(current_elem_attr_name, elemname))
    {
      *node = iter;
      return dbeSuccess;
    }
  }
  return dbeParseFail;
}


/** Save a string to an XML tag, and creates the tag if it doesn't exist. Returns found node in xmlnode if xmlnode not NULL */
static enum eDBError db_interp_xml_set_tag_text(ezxml_t xmlsection, const char *tagname, const char *tagtext, ezxml_t *xmlnode)
{
  ezxml_t xmltag;
  enum eDBError res;

  /* Check pointers */
  if (NULL == xmlsection || NULL == tagname)
  {
    return dbeNullPointer;
  }

  /* Find element attribute name in "elem" tags */
  res = db_interp_xml_elem_get(xmlsection, tagname, &xmltag);
  if (dbeParseFail == res)
  {
    /* Tag doesn't exist, so create it */
    xmltag = ezxml_add_child_d(xmlsection, DB_ELEMENT_ELEM_NAME, 0);
    if (NULL == xmltag)
    {
      return dbeMemAllocFail;
    }

    /* And set name attribute */
    if (NULL == ezxml_set_attr_d(xmltag, DB_ELEMENT_ELEM_ATTRNAME, tagname))
    {
      return dbeMemAllocFail;
    }
  }
  else if (dbeSuccess != res)
  {
    /* Something bad happened */
    return res;
  }

  /* Only set text if tagtext not NULL */
  if (NULL != tagtext)
  {
    /* Set the text for the tag, if different */
    if (NULL != xmltag->txt || 0 != strcmp(xmltag->txt, tagtext))
    {
      res = (NULL == ezxml_set_txt_d(xmltag, tagtext)) ? dbeMemAllocFail : dbeSuccess;
    }
    else
    {
      res = dbeSuccess;
    }
  }
  else
  {
    res = dbeSuccess;
  }

  if (NULL != xmlnode)
  {
    *xmlnode = xmltag;
  }

  return res;
}


/** Create an array root tag, and creates the tag if it doesn't exist */
static enum eDBError db_interp_xml_create_array_root_tag(ezxml_t xmlsection, const char *tagname, size_t num_elements, ezxml_t *xmlarrayroot)
{
  ezxml_t xmltag;
  char str[40];
  enum eDBError res;

  if (NULL == xmlsection || NULL == tagname || NULL == xmlarrayroot)
  {
    return dbeNullPointer;
  }

  /* Get (and create if necessary) array root node */
  res = db_interp_xml_set_tag_text(xmlsection, tagname, NULL, &xmltag);
  if (dbeSuccess != res)
  {
    return res;
  }

  /* Set the attributes for the tag */
  snprintf(str, 40, "%lu", (unsigned long) num_elements);
  res = (NULL == ezxml_set_attr_d(xmltag, DB_ELEMENT_ARRAY_ATTRNUMELEMENTS, str)) ? dbeMemAllocFail : dbeSuccess;
  *xmlarrayroot = xmltag;
  return res;
}


/** Get an array index element if it exists */
static enum eDBError db_interp_xml_array_get_index(ezxml_t xmlarrayroot, size_t index, ezxml_t *node)
{
  ezxml_t iter;
  const char *current_index_str;
  ssize_t current_index;

  if (NULL == xmlarrayroot || NULL == node)
  {
    return dbeNullPointer;
  }

  for (iter = ezxml_child(xmlarrayroot, DB_ELEMENT_ARRAY_ELEM_NAME); ((ezxml_t) 0 != iter); iter = iter->next)
  {
    current_index = -1;
    current_index_str = ezxml_attr(iter, DB_ELEMENT_ARRAY_ATTRINDEX);
    if (NULL != current_index_str)
    {
      current_index = strtoul(current_index_str, (char **) NULL, 10); /* TODO better test this */
    }
    if (current_index == (ssize_t) index)
    {
      *node = iter;
      return dbeSuccess;
    }
  }
  return dbeParseFail;
}


/** Create an array sub element tag, and creates the tag if it doesn't exist */
static enum eDBError db_interp_xml_create_array_element_tag(ezxml_t xmlarrayroot, size_t index, const char *value)
{
  ezxml_t xmlarrayelem;
  char str[40];
  enum eDBError res;

  if (NULL == xmlarrayroot)
  {
    return dbeNullPointer;
  }

  /* See if it already exists */
  res = db_interp_xml_array_get_index(xmlarrayroot, index, &xmlarrayelem);
  if (dbeParseFail == res)
  {
    /* Tag doesn't exist, so create it */
    xmlarrayelem = ezxml_add_child_d(xmlarrayroot, DB_ELEMENT_ARRAY_ELEM_NAME, 0);
    if (NULL == xmlarrayelem)
    {
      return dbeMemAllocFail;
    }

    /* Set the index attribute */
    snprintf(str, 40, "%lu", (unsigned long) index);
    if (NULL == ezxml_set_attr_d(xmlarrayelem, DB_ELEMENT_ARRAY_ATTRINDEX, str))
    {
      return dbeMemAllocFail;
    }
  }
  else if (dbeSuccess != res)
  {
    return res;
  }

  /* Yep, exists, so just change it */
  res = (NULL == ezxml_set_txt_d(xmlarrayelem, value)) ? dbeMemAllocFail : dbeSuccess;
  return res;
}


/** Function to export an unsigned 32bit integer setting to XML */
static enum eDBError db_interp_xml_uint32_export(const char *db_name, const char *name, void *interpreter, uint32_t data)
{
  ezxml_t xmlsection = (ezxml_t) interpreter;
  char str[20];

  /* Set the text for the tag */
  snprintf(str, 20, "%lu", (unsigned long) data);
  return db_interp_xml_set_tag_text(xmlsection, name, str, NULL);
}


/** Function to export a unsigned 8bit integer setting to XML */
static enum eDBError db_interp_xml_uint8_export(const char *db_name, const char *name, void *interpreter, uint8_t data)
{
  return db_interp_xml_uint32_export(db_name, name, interpreter, data);
}


/** Function to export a signed 32bit integer setting to XML */
static enum eDBError db_interp_xml_int32_export(const char *db_name, const char *name, void *interpreter, int32_t data)
{
  ezxml_t xmlsection = (ezxml_t) interpreter;
  char str[20];

  /* Set the text for the tag */
  snprintf(str, 20, "%d", data);
  return db_interp_xml_set_tag_text(xmlsection, name, str, NULL);
}


static enum eDBError db_interp_xml_string_set_value(ezxml_t xmlnode, const char *name, const char *data)
{
  char *str = NULL;
  int strmaxlen;
  enum eDBError res;

  /* Allocate a buffer for the escaped string data (we can't store non-printables in XML, so we escape them like C) */
  strmaxlen = strlen(data) * 4 + 1;
  str = malloc(strmaxlen);
  if (NULL == str)
  {
    res = dbeMemAllocFail;
    goto dixse_done;
  }

  /* Escape the string */
  res = db_string_escape_encode(data, str, strmaxlen);
  if (dbeSuccess != res)
  {
    goto dixse_done;
  }

  /* Set the text for the tag */
  res = db_interp_xml_set_tag_text(xmlnode, name, str, NULL);
  if (dbeSuccess != res)
  {
    goto dixse_done;
  }

  /* Success */
  res = dbeSuccess;

dixse_done:
  if (NULL != str)
    free(str);
  return res;
}


/** Function to export a string (null terminated) setting to XML */
static enum eDBError db_interp_xml_str_export(const char *db_name, const char *name, void *interpreter, const char *data)
{
  ezxml_t xmlsection = (ezxml_t) interpreter;

  return db_interp_xml_string_set_value(xmlsection, name, data);
}


/** Function to export a buffer setting to XML */
static enum eDBError db_interp_xml_buf_export(const char *db_name, const char *name, void *interpreter, const uint8_t *data, size_t datalen)
{
  ezxml_t xmlsection = (ezxml_t) interpreter;
  char *buf = NULL;
  int buflen;
  enum eDBError res;

  /* Allocate a buffer for the base64 encoded data */
  buflen = datalen * 3;
  buf = malloc(buflen);
  if (NULL == buf)
  {
    res = dbeMemAllocFail;
    goto dixbe_done;
  }

  /* Encode data in base64 */
  res = db_base64_encode(data, datalen, buf, buflen);
  if (dbeSuccess != res)
  {
    goto dixbe_done;
  }

  /* Set the text for the tag */
  res = db_interp_xml_set_tag_text(xmlsection, name, buf, NULL);
  if (dbeSuccess != res)
  {
    goto dixbe_done;
  }

  /* Success */
  res = dbeSuccess;

dixbe_done:
  if (NULL != buf)
    free(buf);
  return res;
}


/** Function to export a unsigned 8bit integer array setting to XML */
static enum eDBError db_interp_xml_uint8_array_export(const char *db_name, const char *name, void *interpreter, const uint8_t *data, size_t num_elems)
{
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlarrayroot;
  size_t i;
  enum eDBError res;
  char str[40];

  /* Create array root node */
  res = db_interp_xml_create_array_root_tag(xmldb, name, num_elems, &xmlarrayroot);
  if (dbeSuccess == res)
  {
    /* Create sub nodes for each element in array */
    for (i = 0; i < num_elems; i++)
    {
      snprintf(str, 40, "%u", data[i]);
      res = db_interp_xml_create_array_element_tag(xmlarrayroot, i, str);
      if (dbeSuccess != res)
      {
        break;
      }
    }
  }
  return res;
}


/** Function to export an unsigned 32bit integer array setting to XML */
static enum eDBError db_interp_xml_uint32_array_export(const char *db_name, const char *name, void *interpreter, const uint32_t *data, size_t num_elems)
{
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlarrayroot;
  size_t i;
  enum eDBError res;
  char str[40];

  /* Create array root node */
  res = db_interp_xml_create_array_root_tag(xmldb, name, num_elems, &xmlarrayroot);
  if (dbeSuccess == res)
  {
    /* Create sub nodes for each element in array */
    for (i = 0; i < num_elems; i++)
    {
      snprintf(str, 40, "%lu", (unsigned long) data[i]);
      res = db_interp_xml_create_array_element_tag(xmlarrayroot, i, str);
      if (dbeSuccess != res)
      {
        break;
      }
    }
  }
  return res;
}


/** Function to export a signed 32bit integer array setting to XML */
static enum eDBError db_interp_xml_int32_array_export(const char *db_name, const char *name, void *interpreter, const int32_t *data, size_t num_elems)
{
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlarrayroot;
  size_t i;
  enum eDBError res;
  char str[40];

  /* Create array root node */
  res = db_interp_xml_create_array_root_tag(xmldb, name, num_elems, &xmlarrayroot);
  if (dbeSuccess == res)
  {
    /* Create sub nodes for each element in array */
    for (i = 0; i < num_elems; i++)
    {
      snprintf(str, 40, "%ld", (long) data[i]);
      res = db_interp_xml_create_array_element_tag(xmlarrayroot, i, str);
      if (dbeSuccess != res)
      {
        break;
      }
    }
  }
  return res;
}


/** Function to export a string array setting to XML */
static enum eDBError db_interp_xml_string_array_export(const char *db_name, const char *name, void *interpreter, const char *const *data, size_t num_elems)
{
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlarrayroot;
  size_t i;
  enum eDBError res;
  char *str = NULL;
  size_t strmaxlen;

  /* Create array root node */
  res = db_interp_xml_create_array_root_tag(xmldb, name, num_elems, &xmlarrayroot);
  if (dbeSuccess == res)
  {
    /* Create sub nodes for each element in array */
    for (i = 0; i < num_elems; i++)
    {
      strmaxlen = strlen(data[i]) * 4 + 1;
      str = malloc(strmaxlen);
      if (NULL == str)
      {
        res = dbeMemAllocFail;
      }
      else
      {
        res = db_string_escape_encode(data[i], str, strmaxlen);
        if (dbeSuccess == res)
        {
          res = db_interp_xml_create_array_element_tag(xmlarrayroot, i, str);
        }
        free(str);
        str = NULL;
      }

      if (dbeSuccess != res)
      {
        break;
      }
    } /* for */
  }
  return res;
}


/** Get a string from an XML tag attribute, and set to defaults if it doesn't exist */
static void db_interp_xml_get_tag_attr(ezxml_t xmltag, const char *attrname, char *attrtext, int maxlen)
{
  const char *txt;

  db_safe_strncpy(attrtext, "", maxlen);

  if (NULL != xmltag)
  {
    txt = ezxml_attr(xmltag, attrname);
    if (NULL != txt)
    {
      db_safe_strncpy(attrtext, txt, maxlen);
    }
  }
}


static enum eDBError db_interp_xml_uint32_get_value(ezxml_t xmlnode, uint32_t *data, uint32_t default_value, uint32_t minumum, uint32_t maximum)
{
  unsigned long value;

  if (NULL == xmlnode || NULL == xmlnode->txt)
  {
    *data = default_value;
    return dbeParseFail;
  }

  if (1 != sscanf(xmlnode->txt, "%lu", &value))
  {
    *data = default_value;
    return dbeParseFail;
  }

  if (value < minumum || value > maximum)
  {
    *data = default_value;
    return dbeOutsideLimits;
  }

  /* Success */
  *data = value;
  return dbeSuccess;
}


static enum eDBError db_interp_xml_int32_get_value(ezxml_t xmlnode, int32_t *data, int32_t default_value, int32_t minumum, int32_t maximum)
{
  long value;

  if (NULL == xmlnode || NULL == xmlnode->txt)
  {
    *data = default_value;
    return dbeParseFail;
  }

  if (1 != sscanf(xmlnode->txt, "%ld", &value))
  {
    *data = default_value;
    return dbeParseFail;
  }

  if (value < minumum || value > maximum)
  {
    *data = default_value;
    return dbeOutsideLimits;
  }

  /* Success */
  *data = value;
  return dbeSuccess;
}


static enum eDBError db_interp_xml_string_get_value(ezxml_t xmlnode, char *data, size_t maxlength, const char *default_value)
{
  enum eDBError res;

  /* Check pointers */
  if (NULL == xmlnode ||
      NULL == xmlnode->txt)
  {
    res = dbeNullPointer;
    goto dixsgv_fail;
  }

  /* Decode escaped string */
  res = db_string_escape_decode(xmlnode->txt, data, maxlength);
  if (dbeSuccess != res)
  {
    goto dixsgv_fail;
  }

  /* Success */
  return dbeSuccess;

dixsgv_fail:
  db_safe_strncpy(data, default_value, maxlength);
  return res;
}


/** Function to import an unsigned 32bit integer setting from XML */
static enum eDBError db_interp_xml_uint32_import(const char *db_name, const char *name, void *interpreter, uint32_t *data,  uint32_t lowlim, uint32_t hilim, uint32_t default_value)
{
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlnode = NULL;

  (void) db_interp_xml_elem_get(xmldb, name, &xmlnode);
  return db_interp_xml_uint32_get_value(xmlnode, data, default_value, lowlim, hilim);
}


/** Function to import an unsigned 8bit integer setting from XML */
static enum eDBError db_interp_xml_uint8_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, uint8_t lowlim, uint8_t hilim, uint8_t default_value)
{
  uint32_t intdata;
  enum eDBError result;

  result = db_interp_xml_uint32_import(db_name, name, interpreter, &intdata, lowlim, hilim, default_value);
  *data = (uint8_t) intdata;
  return result;
}


/** Function to import an integer setting from XML */
static enum eDBError db_interp_xml_int32_import(const char *db_name, const char *name, void *interpreter, int32_t *data, int32_t lowlim, int32_t hilim, int32_t default_value)
{
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlnode = NULL;

  (void) db_interp_xml_elem_get(xmldb, name, &xmlnode);
  return db_interp_xml_int32_get_value(xmlnode, data, default_value, lowlim, hilim);
}


/** Function to import a string (null terminated) setting from XML */
static enum eDBError db_interp_xml_str_import(const char *db_name, const char *name, void *interpreter, char *data, size_t maxlen, const char *default_value)
{
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlnode = NULL;

  (void) db_interp_xml_elem_get(xmldb, name, &xmlnode);
  return db_interp_xml_string_get_value(xmlnode, data, maxlen, default_value);
}


/** Function to import a buffer setting from XML */
static enum eDBError db_interp_xml_buf_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, size_t maxlen, const uint8_t *default_value)
{
  enum eDBError res;
  ezxml_t xmlsection = (ezxml_t) interpreter;
  ezxml_t xmltag = NULL;
  size_t length;

  (void) db_interp_xml_elem_get(xmlsection, name, &xmltag);

  /* XML tag not present */
  if (NULL == xmltag ||
      NULL == xmltag->txt )
  {
    /* Fail */
    goto dixbi_default;
  }

  /* Decode base64 data */
  res = db_base64_decode(xmltag->txt, strlen(xmltag->txt), data, maxlen, &length);
  if (dbeSuccess != res)
  {
    /* Fail */
    goto dixbi_default;
  }
  /* TODO check length? */

  /* Success */
  return dbeSuccess;

  /* Set to default */
dixbi_default:
  memcpy(data, default_value, maxlen);
  return dbeOutsideLimits;
}


static void db_interp_xml_uint8_array_default(uint8_t *data, size_t num_elems, const uint8_t *default_values, uint8_t single_default)
{
  size_t i;

  for (i = 0; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
}


/** Function to import a unsigned 8bit integer array setting from XML */
static enum eDBError db_interp_xml_uint8_array_import(const char *db_name, const char *name, void *interpreter, uint8_t *data, size_t num_elems, const uint8_t *lowlims, const uint8_t *hilims, const uint8_t *default_values, uint8_t single_lowlim, uint8_t single_hilim, uint8_t single_default)
{
  enum eDBError res;
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlarrayroot = NULL;
  ezxml_t xmlarraynode;
  char str[20];
  unsigned long read_elems;
  uint32_t u32data;
  size_t i;
  uint8_t min;
  uint8_t max;
  uint8_t def;

  /* Find the array root tag */
  res = db_interp_xml_elem_get(xmldb, name, &xmlarrayroot);
  if (dbeSuccess != res)
  {
    goto dixuai_fail;
  }
  if (NULL == xmlarrayroot)
  {
    res = dbeParseFail;
    goto dixuai_fail;
  }

  /* Get the number of elements */
  db_interp_xml_get_tag_attr(xmlarrayroot, DB_ELEMENT_ARRAY_ATTRNUMELEMENTS, str, 20);
  if (1 != sscanf(str, "%lu", &read_elems))
  {
    res = dbeParseFail;
    goto dixuai_fail;
  }

  if (read_elems > num_elems)
  {
    res = dbeOutsideLimits;
    goto dixuai_fail;
  }

  /* Get each element */
  for (i = 0; i < read_elems; i++)
  {
    /* Get an array index element if it exists */
    res = db_interp_xml_array_get_index(xmlarrayroot, i, &xmlarraynode);
    if (dbeSuccess != res)
    {
      goto dixuai_fail;
    }
    min = (NULL != lowlims) ? lowlims[i] : single_lowlim;
    max = (NULL != hilims) ? hilims[i] : single_hilim;
    def = (NULL != default_values) ? default_values[i] : single_default;
    res = db_interp_xml_uint32_get_value(xmlarraynode, &u32data, def, min, max);
    if (dbeSuccess != res)
    {
      goto dixuai_fail;
    }
    data[i] = u32data;
  }

  /* Rest of elements are defaults */
  for ( ; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
  return res;

dixuai_fail:
  db_interp_xml_uint8_array_default(data, num_elems, default_values, single_default);
  return res;
}


static void db_interp_xml_uint32_array_default(uint32_t *data, size_t num_elems, const uint32_t *default_values, uint32_t single_default)
{
  size_t i;

  for (i = 0; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
}


/** Function to import an unsigned 32bit integer array setting from XML */
static enum eDBError db_interp_xml_uint32_array_import(const char *db_name, const char *name, void *interpreter, uint32_t *data, size_t num_elems, const uint32_t *lowlims, const uint32_t *hilims, const uint32_t *default_values, uint32_t single_lowlim, uint32_t single_hilim, uint32_t single_default)
{
  enum eDBError res;
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlarrayroot = NULL;
  ezxml_t xmlarraynode;
  char str[20];
  unsigned long read_elems;
  size_t i;
  uint32_t min;
  uint32_t max;
  uint32_t def;

  /* Find the array root tag */
  res = db_interp_xml_elem_get(xmldb, name, &xmlarrayroot);
  if (dbeSuccess != res)
  {
    goto dixuai_fail;
  }
  if (NULL == xmlarrayroot)
  {
    res = dbeParseFail;
    goto dixuai_fail;
  }

  /* Get the number of elements */
  db_interp_xml_get_tag_attr(xmlarrayroot, DB_ELEMENT_ARRAY_ATTRNUMELEMENTS, str, 20);
  if (1 != sscanf(str, "%lu", &read_elems))
  {
    res = dbeParseFail;
    goto dixuai_fail;
  }

  if (read_elems > num_elems)
  {
    res = dbeOutsideLimits;
    goto dixuai_fail;
  }

  /* Get each element */
  for (i = 0; i < read_elems; i++)
  {
    /* Get an array index element if it exists */
    res = db_interp_xml_array_get_index(xmlarrayroot, i, &xmlarraynode);
    if (dbeSuccess != res)
    {
      goto dixuai_fail;
    }
    min = (NULL != lowlims) ? lowlims[i] : single_lowlim;
    max = (NULL != hilims) ? hilims[i] : single_hilim;
    def = (NULL != default_values) ? default_values[i] : single_default;
    res = db_interp_xml_uint32_get_value(xmlarraynode, &data[i], def, min, max);
    if (dbeSuccess != res)
    {
      goto dixuai_fail;
    }
  }

  /* Rest of elements are defaults */
  for ( ; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
  return res;

dixuai_fail:
  db_interp_xml_uint32_array_default(data, num_elems, default_values, single_default);
  return res;
}


static void db_interp_xml_int32_array_default(int32_t *data, size_t num_elems, const int32_t *default_values, int32_t single_default)
{
  size_t i;

  for (i = 0; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
}


/** Function to import a signed 32bit integer array setting */
static enum eDBError db_interp_xml_int32_array_import(const char *db_name, const char *name, void *interpreter, int32_t *data, size_t num_elems, const int32_t *lowlims, const int32_t *hilims, const int32_t *default_values, int32_t single_lowlim, int32_t single_hilim, int32_t single_default)
{
  enum eDBError res;
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlarrayroot = NULL;
  ezxml_t xmlarraynode;
  char str[20];
  unsigned long read_elems;
  size_t i;
  int32_t min;
  int32_t max;
  int32_t def;

  /* Find the array root tag */
  res = db_interp_xml_elem_get(xmldb, name, &xmlarrayroot);
  if (dbeSuccess != res)
  {
    goto dixuai_fail;
  }
  if (NULL == xmlarrayroot)
  {
    res = dbeParseFail;
    goto dixuai_fail;
  }

  /* Get the number of elements */
  db_interp_xml_get_tag_attr(xmlarrayroot, DB_ELEMENT_ARRAY_ATTRNUMELEMENTS, str, 20);
  if (1 != sscanf(str, "%lu", &read_elems))
  {
    res = dbeParseFail;
    goto dixuai_fail;
  }

  if (read_elems > num_elems)
  {
    res = dbeOutsideLimits;
    goto dixuai_fail;
  }

  /* Get each element */
  for (i = 0; i < read_elems; i++)
  {
    /* Get an array index element if it exists */
    res = db_interp_xml_array_get_index(xmlarrayroot, i, &xmlarraynode);
    if (dbeSuccess != res)
    {
      goto dixuai_fail;
    }
    min = (NULL != lowlims) ? lowlims[i] : single_lowlim;
    max = (NULL != hilims) ? hilims[i] : single_hilim;
    def = (NULL != default_values) ? default_values[i] : single_default;
    res = db_interp_xml_int32_get_value(xmlarraynode, &data[i], def, min, max);
    if (dbeSuccess != res)
    {
      goto dixuai_fail;
    }
  }

  /* Rest of elements are defaults */
  for ( ; i < num_elems; i++)
  {
    data[i] = (NULL != default_values) ? default_values[i] : single_default;
  }
  return res;

dixuai_fail:
  db_interp_xml_int32_array_default(data, num_elems, default_values, single_default);
  return res;
}


static void db_interp_xml_string_array_default(char *const *data, size_t first_elem, size_t num_elems, const char *const *default_values, const char *single_default, const size_t *max_lengths, size_t single_max_length)
{
  size_t i;
  size_t maxlen;
  const char *def;

  for (i = first_elem; i < num_elems; i++)
  {
    maxlen = (NULL != max_lengths) ? max_lengths[i] : single_max_length;
    def = (NULL != default_values && NULL != default_values[i]) ? default_values[i] : single_default;
    db_safe_strncpy(data[i], def, maxlen);
  }
}


/** Function to import a string array setting */
static enum eDBError db_interp_xml_string_array_import(const char *db_name, const char *name, void *interpreter, char *const *data, size_t num_elems, const size_t *max_lengths, const char *const *default_values, size_t single_max_length, const char *single_default)
{
  enum eDBError res;
  ezxml_t xmldb = (ezxml_t) interpreter;
  ezxml_t xmlarrayroot = NULL;
  ezxml_t xmlarraynode;
  char str[20];
  unsigned long read_elems;
  size_t i;
  const char *def;
  size_t maxlen;

  /* Find the array root tag */
  res = db_interp_xml_elem_get(xmldb, name, &xmlarrayroot);
  if (dbeSuccess != res)
  {
    goto dixuai_fail;
  }
  if (NULL == xmlarrayroot)
  {
    res = dbeParseFail;
    goto dixuai_fail;
  }

  /* Get the number of elements */
  db_interp_xml_get_tag_attr(xmlarrayroot, DB_ELEMENT_ARRAY_ATTRNUMELEMENTS, str, 20);
  if (1 != sscanf(str, "%lu", &read_elems))
  {
    res = dbeParseFail;
    goto dixuai_fail;
  }

  if (read_elems > num_elems)
  {
    res = dbeOutsideLimits;
    goto dixuai_fail;
  }

  /* Get each element */
  for (i = 0; i < read_elems; i++)
  {
    /* Get an array index element if it exists */
    res = db_interp_xml_array_get_index(xmlarrayroot, i, &xmlarraynode);
    if (dbeSuccess != res)
    {
      goto dixuai_fail;
    }
    maxlen = (NULL != max_lengths) ? max_lengths[i] : single_max_length;
    def = (NULL != default_values && NULL != default_values[i]) ? default_values[i] : single_default;
    res = db_interp_xml_string_get_value(xmlarraynode, data[i], maxlen, def);
    if (dbeSuccess != res)
    {
      goto dixuai_fail;
    }
  }

  /* Rest of elements are defaults */
  db_interp_xml_string_array_default(data, i, num_elems, default_values, single_default, max_lengths, single_max_length);
  return res;

dixuai_fail:
  db_interp_xml_string_array_default(data, 0, num_elems, default_values, single_default, max_lengths, single_max_length);
  return res;
}


/** Create an interpreter for XML */
enum eDBError db_create_interp_xml(struct s_db_interp *db_interp, ezxml_t xmlsection)
{
  db_interp->interpreter = (void *) xmlsection;

  /* Export functions */
  db_interp->uint8_export = db_interp_xml_uint8_export;
  db_interp->uint32_export = db_interp_xml_uint32_export;
  db_interp->int32_export = db_interp_xml_int32_export;
  db_interp->str_export = db_interp_xml_str_export;
  db_interp->buf_export = db_interp_xml_buf_export;
  db_interp->uint8_array_export = db_interp_xml_uint8_array_export;
  db_interp->uint32_array_export = db_interp_xml_uint32_array_export;
  db_interp->int32_array_export = db_interp_xml_int32_array_export;
  db_interp->string_array_export = db_interp_xml_string_array_export;

  /* Import functions */
  db_interp->uint8_import = db_interp_xml_uint8_import;
  db_interp->uint32_import = db_interp_xml_uint32_import;
  db_interp->int32_import = db_interp_xml_int32_import;
  db_interp->str_import = db_interp_xml_str_import;
  db_interp->buf_import = db_interp_xml_buf_import;
  db_interp->uint8_array_import = db_interp_xml_uint8_array_import;
  db_interp->uint32_array_import = db_interp_xml_uint32_array_import;
  db_interp->int32_array_import = db_interp_xml_int32_array_import;
  db_interp->string_array_import = db_interp_xml_string_array_import;

  return dbeSuccess;
}
