/*
 * db_strescape.c
 *
 *  Created on: Jun 11, 2012
 *      Author: ian
 */
#include <ctype.h>
#include <rindb/db_error.h>

#include <rindb/db_strescape.h>


/** Hexadecimal characters */
static const char db_hex_string[] = "0123456789abcdef";


/** Is a character a hexadecimal character */
#define db_is_hexchar(ch) (((ch) >= '0' && (ch) <= '9') || ((ch) >= 'a' && (ch) <= 'f') || ((ch) >= 'A' && (ch) <= 'F'))


/** Convert a single hexadecimal character into a character (nibble) */
static char db_hexchar_to_char(char ch)
{
  char res;

  if (ch >= '0' && ch <= '9')
    res = ch - '0';
  else if (ch >= 'a' && ch <= 'f')
    res = ch - 'a' + 0x0A;
  else
    res = ch - 'A' + 0x0A;
  return res;
}


/** See if ch will fit in dst, and add if possible */
#define dst_add_char(ch) \
  if (i >= dstlen) \
  { \
    break; \
  } \
  dst[i++] = ch


/**
 * Escape non-printable characters in a string in C string style. \ is used to escape characters
 * so \\ is a single \, and \xnn is a hex representation of a character. \0 is a special
 * case used to terminate the string
 *
 * @param src is the source string to be escaped. Not NULL
 * @param dst is the destination string that the escaped string will be written to. Not NULL. Must be 4 times longer than src + 1 (null terminator) for worst case.
 * @param dstlen is the maximum length of the destination string. See dst for info on recommended length
 */
enum eDBError db_string_escape_encode(const char *src, char *dst, size_t dstlen)
{
  size_t i;
  enum eDBError res;

  for (i = 0; ('\0' != *src && i < dstlen); src++)
  {
    if ('\\' == *src)
    {
      dst_add_char('\\');
      dst_add_char('\\');
    }
    else if (!isprint(*src))
    {
      dst_add_char('\\');
      dst_add_char('x');
      dst_add_char(db_hex_string[(*src >> 4) & 0x0F]);
      dst_add_char(db_hex_string[(*src & 0x0F)]);
    }
    else
    {
      dst_add_char(*src);
    }
  }

  if (i < dstlen)
  {
    dst[i] = '\0';
    res = dbeSuccess;
  }
  else
  {
    res = dbeOutsideLimits;
  }
  return res;
}


/**
 * Decode an C style escaped string. \ is used to escape characters
 * so \\ is a single \, and \xnn is a hex representation of a character. \0 is a special
 * case used to terminate the string
 *
 * @param src is the source string to be un-escaped. Not NULL
 * @param dst is the destination string that the un-escaped string will be written to. Not NULL. Must be the same length as src + 1 (null terminator) for worst case.
 * @param dstlen is the maximum length of the destination string. See dst for info on recommended length
 */
enum eDBError db_string_escape_decode(const char *src, char *dst, size_t dstlen)
{
  size_t i;
  enum eDBError res;

  for (i = 0, res = dbeSuccess; ('\0' != *src && i < dstlen); i++, src++, dst++)
  {
    if ('\\' == *src)
    {
      /* Escaped character */
      src++;
      if ('\\' == *src)
      {
        *dst = '\\';
      }
      else if ('x' == *src)
      {
        src++;
        if (!db_is_hexchar(*src))
        {
          res = dbeParseFail;
          break;
        }
        *dst = db_hexchar_to_char(*src) << 4;

        src++;
        if (!db_is_hexchar(*src))
        {
          res = dbeParseFail;
          break;
        }
        *dst |= db_hexchar_to_char(*src) & 0x0F;
      }
      else
      {
        /* Illegal escape */
        res = dbeParseFail;
        break;
      }
    }
    else
    {
      /* Normal character */
      *dst = *src;
    }
  }

  if (i < dstlen)
  {
    *dst = '\0';
    /* res is already either dbeSuccess, or dbeParseFail */
  }
  else
  {
    res = dbeOutsideLimits;
  }

  return res;
}
