/*
 * test_db_array_change_size.c
 *
 * Test changing (grow and shrink) array sizes in a database
 *
 *  Created on: 24 Apr. 2019
 *      Author: ian
 */
#include <string.h>
#include <limits.h>
#include <stdbool.h>
#include <cutest/CuTest.h>
#include <rindb/db_buffer.h>
#include <rindb/db_database.h>
#include <rindb/db_database_binbuf.h>

#include <test_db_array_change_size.h>


#define DB_V1_GROW_ARRAY_SIZE 5
#define DB_V2_GROW_ARRAY_SIZE 10


#define DB_V1_SHRINK_ARRAY_SIZE 10
#define DB_V2_SHRINK_ARRAY_SIZE 5


struct s_db_array_size_change_v1
{
  uint32_t grow_array[DB_V1_GROW_ARRAY_SIZE];
  uint8_t shrink_array[DB_V1_SHRINK_ARRAY_SIZE];
};


struct s_db_array_size_change_v2
{
  uint32_t grow_array[DB_V2_GROW_ARRAY_SIZE];
  uint8_t shrink_array[DB_V2_SHRINK_ARRAY_SIZE];
};


static struct s_db_array_size_change_v1 db_array_size_change_v1;
static struct s_db_array_size_change_v2 db_array_size_change_v2;


static const uint32_t db_grow_array_v1_default[DB_V1_GROW_ARRAY_SIZE] = { 1,   2,   3,   4,   5   };
static const uint32_t db_grow_array_v1_minimum[DB_V1_GROW_ARRAY_SIZE] = { 0,   0,   0,   0,   0   };
static const uint32_t db_grow_array_v1_maximum[DB_V1_GROW_ARRAY_SIZE] = { 100, 100, 100, 100, 100 };


static const uint32_t db_grow_array_v2_default[DB_V2_GROW_ARRAY_SIZE] = { 1,   2,   3,   4,   5,   6,   7,   8,   9,   10  };
static const uint32_t db_grow_array_v2_minimum[DB_V2_GROW_ARRAY_SIZE] = { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0   };
static const uint32_t db_grow_array_v2_maximum[DB_V2_GROW_ARRAY_SIZE] = { 100, 100, 100, 100, 100, 100, 100, 100, 100, 100 };


static const uint8_t db_shrink_array_v1_default[DB_V1_SHRINK_ARRAY_SIZE] = { 1,   2,   3,   4,   5,   6,   7,   8,   9,   10  };
static const uint8_t db_shrink_array_v1_minimum[DB_V1_SHRINK_ARRAY_SIZE] = { 0,   0,   0,   0,   0,   0,   0,   0,   0,   0   };
static const uint8_t db_shrink_array_v1_maximum[DB_V1_SHRINK_ARRAY_SIZE] = { 100, 100, 100, 100, 100, 100, 100, 100, 100, 100 };

static const uint8_t db_shrink_array_v2_default[DB_V2_SHRINK_ARRAY_SIZE] = { 1,   2,   3,   4,   5   };
static const uint8_t db_shrink_array_v2_minimum[DB_V2_SHRINK_ARRAY_SIZE] = { 0,   0,   0,   0,   0   };
static const uint8_t db_shrink_array_v2_maximum[DB_V2_SHRINK_ARRAY_SIZE] = { 100, 100, 100, 100, 100 };


#define DB_ARRAY_SIZE_CHANGE_V1 1
#define DB_ARRAY_SIZE_CHANGE_V2 2
#define DB_ARRAY_SIZE_CHANGE_CHECKSUM_SEED 0x1234
#define DB_ARRAY_SIZE_CHANGE_DB_NAME "ArraySizeChangeDB"


static struct s_db_element array_size_change_v1_elements[] =
{
  DB_DECLARE_UINT32_ARRAY_ELEM_INDIVIDUAL("GrowArray",   "GrowArray",   db_array_size_change_v1.grow_array,   DB_V1_GROW_ARRAY_SIZE,   db_grow_array_v1_minimum,   db_grow_array_v1_maximum,   db_grow_array_v1_default,   DB_NO_MINIMUM_VERSION, DB_NO_MAXIMUM_VERSION, 0x0),

  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL( "ShrinkArray", "ShrinkArray", db_array_size_change_v1.shrink_array, DB_V1_SHRINK_ARRAY_SIZE, db_shrink_array_v1_minimum, db_shrink_array_v1_maximum, db_shrink_array_v1_default, DB_NO_MINIMUM_VERSION, DB_NO_MAXIMUM_VERSION, 0x0),
};


static struct s_db_element array_size_change_v2_elements[] =
{
  DB_DECLARE_UINT32_ARRAY_ELEM_INDIVIDUAL("GrowArray",   "GrowArray",   NULL,                                 DB_V1_GROW_ARRAY_SIZE,   db_grow_array_v1_minimum,   db_grow_array_v1_maximum,   db_grow_array_v1_default,   DB_NO_MINIMUM_VERSION,   DB_ARRAY_SIZE_CHANGE_V1, 0x0),
  DB_DECLARE_UINT32_ARRAY_ELEM_INDIVIDUAL("GrowArray",   "GrowArray",   db_array_size_change_v2.grow_array,   DB_V2_GROW_ARRAY_SIZE,   db_grow_array_v2_minimum,   db_grow_array_v2_maximum,   db_grow_array_v2_default,   DB_ARRAY_SIZE_CHANGE_V2, DB_NO_MAXIMUM_VERSION,   0x0),

  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL( "ShrinkArray", "ShrinkArray", NULL,                                 DB_V1_SHRINK_ARRAY_SIZE, db_shrink_array_v1_minimum, db_shrink_array_v1_maximum, db_shrink_array_v1_default, DB_NO_MINIMUM_VERSION,   DB_ARRAY_SIZE_CHANGE_V1, 0x0),
  DB_DECLARE_UINT8_ARRAY_ELEM_INDIVIDUAL( "ShrinkArray", "ShrinkArray", db_array_size_change_v2.shrink_array, DB_V2_SHRINK_ARRAY_SIZE, db_shrink_array_v2_minimum, db_shrink_array_v2_maximum, db_shrink_array_v2_default, DB_ARRAY_SIZE_CHANGE_V2, DB_NO_MAXIMUM_VERSION,   0x0),
};


static struct s_db_database array_size_change_v1_database =
{
  .db_database_version = DB_ARRAY_SIZE_CHANGE_V1,
  .db_database_chksum_seed = DB_ARRAY_SIZE_CHANGE_CHECKSUM_SEED,
  .db_elements = array_size_change_v1_elements,
  .db_num_elements = DB_NUM_ELEMENTS(array_size_change_v1_elements),
  .db_name = DB_ARRAY_SIZE_CHANGE_DB_NAME,
  .db_checksum_buffer = NULL
};


static struct s_db_database array_size_change_v2_database =
{
  .db_database_version = DB_ARRAY_SIZE_CHANGE_V2,
  .db_database_chksum_seed = DB_ARRAY_SIZE_CHANGE_CHECKSUM_SEED,
  .db_elements = array_size_change_v2_elements,
  .db_num_elements = DB_NUM_ELEMENTS(array_size_change_v2_elements),
  .db_name = DB_ARRAY_SIZE_CHANGE_DB_NAME,
  .db_checksum_buffer = NULL
};


/** Test growing and shrinking an array size */
static void test_db_array_change_size(CuTest* tc)
{
  enum eDBError res;
  struct s_db_buffer db_binbuffer;
  struct s_db_database_interp ddi_binbuffer;
  uint32_t version;

  /* Initialise the database (creates previous data for some elements) */
  res = db_database_init(&array_size_change_v1_database);
  CuAssertIntEquals(tc, dbeSuccess, res);
  res = db_database_init(&array_size_change_v2_database);
  CuAssertIntEquals(tc, dbeSuccess, res);

  /* Create database interpreter */
  res = db_buffer_create_malloc(&db_binbuffer, 1024, 1024);
  CuAssertIntEquals(tc, dbeSuccess, res);
  res = db_create_database_interp_binbuffer(&ddi_binbuffer, &db_binbuffer);
  CuAssertIntEquals(tc, dbeSuccess, res);

  /* Write version 1 database to file (grow array with 5 elements, shrink array with 10 elements) */
  res = db_database_write(&array_size_change_v1_database, &ddi_binbuffer, DB_ARRAY_SIZE_CHANGE_V1);
  CuAssertIntEquals(tc, dbeSuccess, res);

  /* It will then read fine back to a version 1 database */
  res = db_database_read(&array_size_change_v1_database, &ddi_binbuffer, &version, false);
  CuAssertIntEquals(tc, dbeSuccess, res);
  CuAssertIntEquals(tc, 1, version);

  /* But we can also read back into a version 2 database */
  res = db_database_read(&array_size_change_v2_database, &ddi_binbuffer, &version, false);
  CuAssertIntEquals(tc, dbeSuccess, res);
  CuAssertIntEquals(tc, 1, version);

  /* Destroy database interpreter */
  res = db_destroy_database_interp_binbuffer(&ddi_binbuffer);
  CuAssertIntEquals(tc, dbeSuccess, res);
  res = db_buffer_destroy(&db_binbuffer);
  CuAssertIntEquals(tc, dbeSuccess, res);

  /* Destroy all elements in the database */
  res = db_database_destroy(&array_size_change_v1_database);
  CuAssertIntEquals(tc, dbeSuccess, res);
  res = db_database_destroy(&array_size_change_v2_database);
  CuAssertIntEquals(tc, dbeSuccess, res);
}


CuSuite *CuGetSuite_DB_ArrayChangeSize(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_db_array_change_size);

  return suite;
}
