/*
 * test_db_base64.c
 *
 *  Created on: 31/08/2011
 *      Author: ian
 */
#include <string.h>
#include <cutest/CuTest.h>
#include <rindb/db_base64.h>

#include "test_db_base64.h"


/** Test the base 64 encoding and decoding of zero length buffer */
static void test_dblib_base64_zerolength(CuTest* tc)
{
  uint8_t raw_buf[256];
  char encoded_buf[256 * 3];
  uint8_t decoded_buf[256];
  enum eDBError res;
  size_t length;

  /* Fill buffer */
  memset(raw_buf, 0, sizeof(raw_buf));

  /* Encode zero length buffer */
  res = db_base64_encode(raw_buf, 0, encoded_buf, 256 * 3);
  CuAssertIntEquals(tc, dbeSuccess, res);
  CuAssertIntEquals(tc, 0, strlen(encoded_buf));

  /* Decode zero length buffer */
  res = db_base64_decode((char *) raw_buf, 0, decoded_buf, 256, &length);
  CuAssertIntEquals(tc, dbeSuccess, res);
  CuAssertIntEquals(tc, 0, length);
}


/** Test the base 64 encoding and decoding of one to four characters buffer */
static void test_dblib_base64_1234chars(CuTest* tc)
{
  #define NUM_B64_TESTS 4
  char encoded_buf[256 * 3];
  uint8_t decoded_buf[256];
  size_t i;
  enum eDBError res;
  size_t length;
  const char *const raw_in_bufs[NUM_B64_TESTS] = { "A", "AB", "ABC", "ABCD" };
  const char *const encoded_answers[NUM_B64_TESTS] = { "QQ==", "QUI=", "QUJD", "QUJDRA==" };

  for (i = 0; i < NUM_B64_TESTS; i++)
  {
    /* Encode raw buffer */
    res = db_base64_encode((uint8_t *) raw_in_bufs[i], strlen(raw_in_bufs[i]), encoded_buf, 256 * 3);
    CuAssertIntEquals(tc, dbeSuccess, res);
    CuAssertStrEquals(tc, encoded_answers[i], encoded_buf);

    /* Decode */
    res = db_base64_decode(encoded_buf, strlen(encoded_buf), decoded_buf, 256, &length);
    CuAssertIntEquals(tc, dbeSuccess, res);
    CuAssertIntEquals(tc, strlen(raw_in_bufs[i]), length);
    CuAssertIntEquals(tc, 0, memcmp(raw_in_bufs[i], decoded_buf, length));
  }
}


/** Test the base 64 decoding of short encoded strings */
static void test_dblib_base64_short_strings(CuTest* tc)
{
  uint8_t decoded_buf[256];
  enum eDBError res;
  size_t length;

  /* Decode 4 character encoded buffer */
  const char encoded_4char[] = "QQ==";
  res = db_base64_decode(encoded_4char, strlen(encoded_4char), decoded_buf, 256, &length);
  CuAssertIntEquals(tc, dbeSuccess, res);
  CuAssertIntEquals(tc, strlen("A"), length);
  CuAssertIntEquals(tc, 0, memcmp("A", decoded_buf, length));

  /* Decode 3 character encoded buffer */
  const char encoded_3char[] = "QQ=";
  res = db_base64_decode(encoded_3char, strlen(encoded_3char), decoded_buf, 256, &length);
  CuAssertIntEquals(tc, dbeParseFail, res);

  /* Decode 2 character encoded buffer */
  const char encoded_2char[] = "QQ";
  res = db_base64_decode(encoded_2char, strlen(encoded_2char), decoded_buf, 256, &length);
  CuAssertIntEquals(tc, dbeParseFail, res);

  /* Decode 1 character encoded buffer */
  const char encoded_1char[] = "Q";
  res = db_base64_decode(encoded_1char, strlen(encoded_1char), decoded_buf, 256, &length);
  CuAssertIntEquals(tc, dbeParseFail, res);

  /* Decode 0 character encoded buffer */
  const char encoded_0char[] = "";
  res = db_base64_decode(encoded_0char, strlen(encoded_0char), decoded_buf, 256, &length);
  CuAssertIntEquals(tc, dbeSuccess, res);
  CuAssertIntEquals(tc, 0, length);
}


/** Test the base 64 encoding and decoding of 0 to FF */
static void test_dblib_base64_0toFF(CuTest* tc)
{
  uint8_t raw_buf[256];
  char encoded_buf[256 * 3];
  size_t enc_buf_len;
  size_t i;
  enum eDBError res;
  size_t length;

  /* Fill buffer */
  for (i = 0; i < 256; i++)
  {
    raw_buf[i] = (uint8_t) i;
  }

  /* Encode buffer */
  res = db_base64_encode(raw_buf, 256, encoded_buf, 256 * 3);
  CuAssertIntEquals(tc, dbeSuccess, res);

  /* Decode buffer */
  enc_buf_len = strlen(encoded_buf);
  res = db_base64_decode(encoded_buf, enc_buf_len, raw_buf, 256, &length);
  CuAssertIntEquals(tc, dbeSuccess, res);
  CuAssertIntEquals(tc, 256, length);

  /* Verify buffer */
  for (i = 0; i < 256; i++)
  {
    CuAssertIntEquals(tc, i, raw_buf[i]);
  }
}


/** Test the base 64 decoding from pre-encoded buffer of of 0 to FF */
//                case 5:All chars from 0 to ff, linesize set to 50:
static void test_dblib_base64_0toFF_preencoded(CuTest* tc)
{
  size_t preenc_buf_len;
  const char preencoded_buffer[] =
    "AAECAwQFBgcICQoLDA0ODxAREhMUFRYXGBkaGxwdHh8gISIj"
    "JCUmJygpKissLS4vMDEyMzQ1Njc4OTo7PD0.P0BBQkNERUZH"
    "SElKS0xNTk9QUVJTVFVWV1hZWltcXV5fYGFiY2RlZmdoaWpr"
    "bG1ub3BxcnN0dXZ3eHl6e3x9fn.AgYKDhIWGh4iJiouMjY6P"
    "kJGSk5SVlpeYmZqbnJ2en6ChoqOkpaanqKmqq6ytrq.wsbKz"
    "tLW2t7i5uru8vb6-wMHCw8TFxsfIycrLzM3Oz9DR0tPU1dbX"
    "2Nna29zd3t-g4eLj5OXm5.jp6uvs7e7v8PHy8-T19vf4.fr7"
    "-P3.-w==";

  uint8_t decoded_buf[512];
  enum eDBError res;
  size_t length;
  size_t i;

  /* Decode buffer */
  preenc_buf_len = strlen(preencoded_buffer);
  res = db_base64_decode(preencoded_buffer, preenc_buf_len, decoded_buf, 512, &length);
  CuAssertIntEquals(tc, dbeSuccess, res);
  CuAssertIntEquals(tc, 256, length);

  /* Verify buffer */
  for (i = 0; i < 256; i++)
  {
    CuAssertIntEquals(tc, i, decoded_buf[i]);
  }
}


CuSuite *CuGetSuite_DB_Base64(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_dblib_base64_zerolength);
  SUITE_ADD_TEST(suite, test_dblib_base64_1234chars);
  SUITE_ADD_TEST(suite, test_dblib_base64_short_strings);
  SUITE_ADD_TEST(suite, test_dblib_base64_0toFF);
  SUITE_ADD_TEST(suite, test_dblib_base64_0toFF_preencoded);

  return suite;
}
