/*
 * dsd_copy.c
 *
 *  Created on: Mar 21, 2012
 *      Author: ian
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>

#include <rinweigh/adcdsd.h>


/**
 * Copy the DSD file to copyfd (no locking).
 *
 * Writing within copyfd occurs from the write position in the file when the function is called.
 * No truncation is performed.
 *
 * @param dsd is the DSD structure
 * @param copyfd is a file descriptor open for writing.
 * @param bytes_written is set to the number of bytes written to copyfd. Can be NULL
 */
enum eAdcDsdError AdcIntDSDFileCopyFD(struct s_adc_dsd *dsd, int copyfd, int32u *bytes_written)
{
  #define COPY_BUF_LEN 1024
  off_t length;
  char *buf = NULL;
  enum eAdcDsdError result;
  ssize_t readlen;
  ssize_t writelen;
  ssize_t pos;

  /* Check pointers */
  if (NULL == dsd)
  {
    return ADC_DSD_ERROR_NULL;
  }

  /* Allocate copy buffer */
  buf = malloc(COPY_BUF_LEN);
  if (NULL == buf)
  {
    result = ADC_DSD_ERROR_MALLOC;
    goto aidfcf_fail;
  }

  /* Find length of DSD file */
  length = lseek(dsd->fd, 0, SEEK_END);
  if (length < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidfcf_fail;
  }

  /* Back to start of DSD file */
  if (lseek(dsd->fd, 0, SEEK_SET) < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidfcf_fail;
  }

  /* Copy DSD from source to destination */
  for (pos = 0; pos < length; pos += readlen)
  {
    /* How many to copy this time */
    readlen = (length - pos > COPY_BUF_LEN) ? COPY_BUF_LEN : length - pos;

    /* Read the bytes */
    if (readlen != read(dsd->fd, buf, readlen))
    {
      result = ADC_DSD_ERROR_FILE;
      goto aidfcf_fail;
    }

    /* Write the bytes */
    writelen = write(copyfd, buf, readlen);
    if (writelen != readlen)
    {
      result = ADC_DSD_ERROR_FILE;
      goto aidfcf_fail;
    }
  }

  if (NULL != bytes_written)
  {
    *bytes_written = pos;
  }

  /* Success */
  result = ADC_DSD_SUCCESS;

aidfcf_fail:
  if (NULL != buf)
  {
    free(buf);
    buf = NULL;
  }
  return result;
}


/**
 * Copy the DSD file to filename (no locking)
 *
 * @param dsd is the DSD structure
 * @param filename is the name of a writeable file to copy the DSD to. File will be overwritten.
 * @param bytes_written is set to the number of bytes written to copyfd. Can be NULL
 */
enum eAdcDsdError AdcIntDSDFileCopyFilename(struct s_adc_dsd *dsd, const char *filename, int32u *byteswritten)
{
  int copyfd = -1;
  enum eAdcDsdError result;

  /* Check pointers */
  if (NULL == dsd || NULL == filename)
  {
    return ADC_DSD_ERROR_NULL;
  }

  /* Open output file */
  copyfd = open(filename, O_CREAT | O_RDWR | O_TRUNC, S_IRUSR | S_IWUSR);
  if (copyfd < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidfcf_fail;
  }
  if (fcntl(copyfd, F_SETFD, FD_CLOEXEC) < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidfcf_fail;
  }

  /* Copy data to file */
  result = AdcIntDSDFileCopyFD(dsd, copyfd, byteswritten);

aidfcf_fail:
  if (copyfd >= 0)
  {
    close(copyfd);
    copyfd = -1;
  }
  return result;
}
