/*
 * dsd_fileindex.c
 *
 *  Created on: Mar 14, 2012
 *      Author: ian
 */
#if defined(WIN32)
  #include <winsock.h>
#else
  #include <arpa/inet.h>
#endif
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include "dsd_byteorder.h"
#include <rinweigh/adcdsd.h>

#include "dsd_fileindex.h"


/*
 * Physical record is record indexed from start of file (not taking into account circular buffer)
 * Record is the record indexed from the oldest record (taking into account circular buffer)
 */


/** Convert a physical record number to an offset within the file */
enum eAdcDsdError AdcIntDSDPhysicalRecordNumberToFileOffset(struct s_adc_dsd *dsd, int32u physicalrecord, off_t *filepos)
{
  if (physicalrecord >= dsd->file_index.dsd_maxrecords)
  {
    return ADC_DSD_ERROR_OUTOFRANGE;
  }

  /* Calculate the position in the file */
  *filepos = 2 * sizeof(struct s_adc_dsd_file_index) + dsd->file_index.dsd_recordsize * physicalrecord;
  return ADC_DSD_SUCCESS;
}


/** Write the file indexes back to the DSD file */
enum eAdcDsdError AdcIntDSDWriteFileIndexes(struct s_adc_dsd *dsd)
{
  enum eAdcDsdError result;
  struct s_adc_dsd_file_index fi_netorder;
  int16u checksum;

  /* Setup file index for write */
  fi_netorder = dsd->file_index;

  /*
   * Checksum is special: so it will work irrespective of machine endianness, we
   * first put the file index contents into network byte order, then calculate the
   * checksum, then set the network order version in the header.
   */
  AdcIntDSDFileIndexToNetworkByteOrder(&fi_netorder);
  AdcIntDSDCalculateFileIndexChecksum(&fi_netorder, &checksum);
  fi_netorder.chksum = htons(checksum);

  /* Seek the start of the file */
  if (lseek(dsd->fd, 0, SEEK_SET) < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto adufi_fail;
  }

  /* Write the first file index */
  if (sizeof(struct s_adc_dsd_file_index) != write(dsd->fd, &fi_netorder, sizeof(struct s_adc_dsd_file_index)))
  {
    result = ADC_DSD_ERROR_FILE;
    goto adufi_fail;
  }

  /* Write the second file index */
  if (sizeof(struct s_adc_dsd_file_index) != write(dsd->fd, &fi_netorder, sizeof(struct s_adc_dsd_file_index)))
  {
    result = ADC_DSD_ERROR_FILE;
    goto adufi_fail;
  }

  /* Success */
  result = ADC_DSD_SUCCESS;

adufi_fail:
  return result;
}


/** Read and check a file index from the DSD file (at current file read pointer) into fi */
static enum eAdcDsdError AdcIntDSDReadCheckIndex(struct s_adc_dsd *dsd, off_t file_pos, struct s_adc_dsd_file_index *fi)
{
  int16u checksum;
  struct s_adc_dsd_file_index fi_netorder;

  /* Seek the start of the file index */
  if (lseek(dsd->fd, file_pos, SEEK_SET) < 0)
  {
    return ADC_DSD_ERROR_FILE;
  }

  /* Read the first file index */
  if (sizeof(struct s_adc_dsd_file_index) != read(dsd->fd, &fi_netorder, sizeof(struct s_adc_dsd_file_index)))
  {
    return ADC_DSD_ERROR_FILE;
  }

  /*
   * Checksum is special: so it will work irrespective of machine endianness, we
   * calculate the checksum over the contents in network byte order, then
   * check against the stored checksum (taking into account byte ordering).
   * Then we can swap the structure contents into host byte ordering.
   */
  AdcIntDSDCalculateFileIndexChecksum(&fi_netorder, &checksum);

  /* Back to our byte order for other members */
  *fi = fi_netorder;
  AdcIntDSDFileIndexFromNetworkByteOrder(fi);
  fi->chksum = ntohs(fi->chksum);

  /* Check checksum */
  if (fi->chksum != checksum)
  {
    return ADC_DSD_ERROR_FILEINDEXCHKSUM;
  }

  /* Check the magic number */
  if (0 != memcmp(fi->dsd_magic, ADC_DSD_FILE_INDEX_MAGIC, ADC_DSD_FILE_INDEX_MAGIC_LENGTH))
  {
    return ADC_DSD_ERROR_FILEINDEXCHKSUM;
  }

  /* Check the version number */
  if (ADC_DSD_FILE_INDEX_VERSION_CURRENT != fi->dsd_version)
  {
    return ADC_DSD_ERROR_FILEINDEXCHKSUM;
  }

  /* Success */
  return ADC_DSD_SUCCESS;
}


/** Read the file indexes from the DSD file */
enum eAdcDsdError AdcIntDSDReadFileIndexes(struct s_adc_dsd *dsd)
{
  enum eAdcDsdError interim_result;
  enum eAdcDsdError result;
  struct s_adc_dsd_file_index fi;

  /* Read the first file index */
  interim_result = AdcIntDSDReadCheckIndex(dsd, 0, &fi);
  if (ADC_DSD_SUCCESS != interim_result)
  {
    /* Try second checksum */
    interim_result = AdcIntDSDReadCheckIndex(dsd, sizeof(struct s_adc_dsd_file_index), &fi);
  }
  if (ADC_DSD_SUCCESS != interim_result)
  {
    result = interim_result;
    goto aidrfi_fail;
  }

  /* Success */
  dsd->file_index = fi;
  result = ADC_DSD_SUCCESS;

aidrfi_fail:
  return result;
}


/**
 * Read a file index from the DSD file (no locking).
 *
 * @param dsd is the DSD structure
 * @param which is 0 for the first index, or 1 for the second index
 * @param fi is the returned file index
 */
enum eAdcDsdError AdcIntDSDReadFileIndex(struct s_adc_dsd *dsd, int which, struct s_adc_dsd_file_index *fi)
{
  enum eAdcDsdError result;
  off_t filepos;

  /* Check pointers */
  if (NULL == dsd || NULL == fi)
  {
    return ADC_DSD_ERROR_NULL;
  }

  if (which > 1 || which < 0)
  {
    result = ADC_DSD_ERROR_OUTOFRANGE;
  }
  else
  {
    filepos = which * sizeof(struct s_adc_dsd_file_index);
    result = AdcIntDSDReadCheckIndex(dsd, filepos, fi);
  }
  return result;
}


/** Calculate the (theoretical) size of the DSD file */
off_t AdcIntDSDCalculateFileSize(struct s_adc_dsd *dsd)
{
  off_t filesize;

  /* File size is two copies of the file index + the number of records * the record size */
  filesize = 2 * sizeof(struct s_adc_dsd_file_index) +
             dsd->file_index.dsd_maxrecords * dsd->file_index.dsd_recordsize;
  return filesize;
}


/** Get the number of records currently held in the DSD */
enum eAdcDsdError AdcIntDSDGetNumberRecords(struct s_adc_dsd *dsd, int32u *records)
{
  /* Check pointers */
  if (NULL == dsd || NULL == records)
  {
    return ADC_DSD_ERROR_NULL;
  }

  *records = dsd->file_index.dsd_recordcount;
  return ADC_DSD_SUCCESS;
}


/** Get the number of records currently held in the DSD (no locking) */
enum eAdcDsdError AdcIntDSDGetCapacity(struct s_adc_dsd *dsd, int32u *records)
{
  /* Check pointers */
  if (NULL == dsd || NULL == records)
  {
    return ADC_DSD_ERROR_NULL;
  }

  *records = dsd->file_index.dsd_maxrecords;
  return ADC_DSD_SUCCESS;
}


/** Get the record size for the DSD (no locking) */
enum eAdcDsdError AdcIntDSDGetRecordSize(struct s_adc_dsd *dsd, int32u *recordsize)
{
  /* Check pointers */
  if (NULL == dsd || NULL == recordsize)
  {
    return ADC_DSD_ERROR_NULL;
  }

  *recordsize = dsd->file_index.dsd_recordsize - sizeof(struct s_adc_dsd_record_header);
  return ADC_DSD_SUCCESS;
}


/**
 * Called prior to writing a record to the DSD to verify
 * that there is room, etc to write the record.
 *
 * Paired with @ref AdcIntDSDPostWrite, but no need to call AdcIntDSDPostWrite if something fails
 *
 * @param write_pos is the file position to write the next record to the DSD, only set on ADC_DSD_SUCCESS
 * @return ADC_DSD_SUCCESS if the write can proceed
 */
enum eAdcDsdError AdcIntDSDPreWrite(struct s_adc_dsd *dsd, off_t *write_pos)
{
  int32u newest;
  int32u oldest;
  int32u maxrecords;
  int32u recordcount;
  int32u phys_record;

  newest = dsd->file_index.dsd_newestindex;
  oldest = dsd->file_index.dsd_oldestindex;
  maxrecords = dsd->file_index.dsd_maxrecords;
  recordcount = dsd->file_index.dsd_recordcount;

  if (0 == maxrecords)
  {
    return ADC_DSD_ERROR_FULL;
  }

  if (newest == oldest)
  {
    /* Either full or empty */
    if (recordcount >= maxrecords)
    {
      /* Full */
      return ADC_DSD_ERROR_FULL;
    }
    else
    {
      /* Empty */
      phys_record = newest;
    }
  }
  else
  {
    /* Has room */
    phys_record = newest;
  }

  /* Convert a physical record number to an offset within the file */
  return AdcIntDSDPhysicalRecordNumberToFileOffset(dsd, phys_record, write_pos);
}


/**
 * Called after to writing a record to the DSD to update file
 * indexes
 *
 * Paired with @ref AdcIntDSDPreWrite, never call by itself.
 *
 * @param write_pos is the file position to write the next record to the DSD, only set on ADC_DSD_SUCCESS
 * @return ADC_DSD_SUCCESS if the write can proceed
 */
enum eAdcDsdError AdcIntDSDPostWrite(struct s_adc_dsd *dsd)
{
  /* TODO some checks? */

  /* Update indexes */
  dsd->file_index.dsd_recordcount++;
  dsd->file_index.dsd_newestindex++;
  if (dsd->file_index.dsd_newestindex >= dsd->file_index.dsd_maxrecords)
  {
    dsd->file_index.dsd_newestindex = 0;
  }

  /* Now write the file indexes back to the DSD file */
  return AdcIntDSDWriteFileIndexes(dsd);
}


/** Locate the file position of record */
enum eAdcDsdError AdcIntDSDGetRecordFilePos(struct s_adc_dsd *dsd, int32u record, off_t *filepos)
{
  int32u newest;
  int32u oldest;
  int32u recordcount;
  int32u maxrecords;
  int32u pys_record;

  newest = dsd->file_index.dsd_newestindex;
  oldest = dsd->file_index.dsd_oldestindex;
  recordcount = dsd->file_index.dsd_recordcount;
  maxrecords = dsd->file_index.dsd_maxrecords;

  /* Quick checks */
  if (0 == recordcount)
  {
    /* No records in DSD */
    return ADC_DSD_ERROR_EMPTY;
  }
  if (record >= recordcount || record >= maxrecords)
  {
    /* Requesting record beyond the number of records we have */
    return ADC_DSD_ERROR_OUTOFRANGE;
  }

  if (newest == oldest)
  {
    /* DSD full, because of empty test above */
    /* Work from oldest onward */
    if (record < maxrecords - oldest)
    {
      pys_record = record + oldest;
    }
    else
    {
      /* Wrap */
      pys_record = record - (maxrecords - oldest);
    }
  }
  else if (newest > oldest)
  {
    /* Newest after oldest */
    if (record >= newest - oldest)
    {
      /* Requesting record beyond the number of records we have */
      return ADC_DSD_ERROR_OUTOFRANGE;
    }
    else
    {
      pys_record = record + oldest;
    }
  }
  else
  {
    /* Newest before oldest, file writing has wrapped */
    if (record >= maxrecords - oldest + newest)
    {
      /* Requesting record beyond the number of records we have */
      return ADC_DSD_ERROR_OUTOFRANGE;
    }
    else
    {
      if (record < maxrecords - oldest)
      {
        /* Between oldest and end of file */
        pys_record = oldest + record;
      }
      else
      {
        /* Wrapped */
        pys_record = record - (maxrecords - oldest);
      }
    }
  }

  /* Now calculate the position in the file */
  return AdcIntDSDPhysicalRecordNumberToFileOffset(dsd, pys_record, filepos);
}
