/*
 * dsd_init.c
 *
 *  Created on: Mar 6, 2012
 *      Author: ian
 */
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include "dsd_fileindex.h"
#include <rinweigh/adcmuteximpl.h>

#include <rinweigh/adcdsd.h>


/**
 * Create and initialise the DSD structure, but not the file.
 * Need to call AdcIntDSDOpenExisting() or AdcIntDSDCreateNew() to access the file.
 *
 * @param dsd is a pointer to a pointer to a DSD structure that is set to point to the new structure created by the function. Not NULL
 * @param mutex_implementation is the implementation of the mutex. Can be NULL for no mutex.
 *  */
enum eAdcDsdError AdcDSDInit(struct s_adc_dsd **dsd, AdcMutexImplementer *mutex_implementation)
{
  enum eAdcDsdError result;
  struct s_adc_dsd *idsd = NULL;

  if (NULL == dsd)
  {
    result = ADC_DSD_ERROR_NULL;
    goto adi_fail;
  }

  /* Allocate DSD structure */
  idsd = malloc(sizeof(struct s_adc_dsd));
  if (NULL == idsd)
  {
    result = ADC_DSD_ERROR_MALLOC;
    goto adi_fail;
  }
  *dsd = idsd;

  /* Initialise members */
  memset(idsd, 0, sizeof(struct s_adc_dsd));
  idsd->mutex_implementation = mutex_implementation;
  idsd->filename = NULL;
  idsd->fd = -1;
  idsd->work_buffer = NULL;

  /* Success */
  return ADC_DSD_SUCCESS;

adi_fail:
  if (NULL != idsd)
  {
    if (NULL != idsd->work_buffer)
    {
      free(idsd->work_buffer);
      idsd->work_buffer = NULL;
    }
    if (NULL != idsd->filename)
    {
      free(idsd->filename);
      idsd->filename = NULL;
    }
    if (idsd->fd >= 0)
    {
      close(idsd->fd);
      idsd->fd = -1;
    }
    free(idsd);
    idsd = NULL;
  }
  *dsd = NULL;
  return result;
}


/** Close the current DSD. dsd structure remains. (no locking) */
enum eAdcDsdError AdcIntDSDClose(struct s_adc_dsd *dsd)
{
  if (NULL == dsd)
  {
    return ADC_DSD_ERROR_NULL;
  }

  /* Free up stuff */
  if (NULL != dsd->work_buffer)
  {
    free(dsd->work_buffer);
    dsd->work_buffer = NULL;
  }
  if (dsd->fd >= 0)
  {
    close(dsd->fd);
    dsd->fd = -1;
  }
  if (NULL != dsd->filename)
  {
    free(dsd->filename);
    dsd->filename = NULL;
  }

  /* Trash members */
  memset(&dsd->file_index, 0, sizeof(struct s_adc_dsd_file_index));

  return ADC_DSD_SUCCESS;
}


/** Close and clean up the DSD */
enum eAdcDsdError AdcDSDDestroy(struct s_adc_dsd *dsd)
{
  if (NULL == dsd)
  {
    return ADC_DSD_ERROR_NULL;
  }

  (void) AdcIntDSDClose(dsd);

  free(dsd);
  dsd = NULL;
  return ADC_DSD_SUCCESS;
}


/**
 * Create a new DSD and ready it for use.
 * Will wipe the file if it already exists
 * DSD parameters will be retrieved from the file.
 * (no locking)
 */
enum eAdcDsdError AdcIntDSDCreateNew(struct s_adc_dsd *dsd, const char *filename, int32u recordsize, int32u maxrecords, int16u recordcrcseed, int32u serialnumber)
{
  enum eAdcDsdError result;
  enum eAdcDsdError interim_result;
  off_t filesize;

  if (NULL == dsd || NULL == filename)
  {
    return ADC_DSD_ERROR_NULL;
  }

  /* Take copy of filename */
  dsd->filename = strdup(filename);
  if (NULL == dsd->filename)
  {
    result = ADC_DSD_ERROR_MALLOC;
    goto aidcn_fail;
  }

  /* Create file, truncate if it already exists */
#if defined(WIN32)
  dsd->fd = open(dsd->filename, O_RDWR | O_CREAT | O_TRUNC);
#else
  dsd->fd = open(dsd->filename, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
#endif
  if (dsd->fd < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidcn_fail;
  }
#if !defined(WIN32)
  if (fcntl(dsd->fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidcn_fail;
  }
#endif

  /* Fill out members */
  memcpy(dsd->file_index.dsd_magic, ADC_DSD_FILE_INDEX_MAGIC, ADC_DSD_FILE_INDEX_MAGIC_LENGTH);
  dsd->file_index.dsd_version = ADC_DSD_FILE_INDEX_VERSION_CURRENT;
  dsd->file_index.dsd_recordsize = recordsize + sizeof(struct s_adc_dsd_record_header);
  dsd->file_index.dsd_maxrecords = maxrecords;
  dsd->file_index.dsd_newestindex = 0;
  dsd->file_index.dsd_oldestindex = 0;
  dsd->file_index.dsd_recordcount = 0;
  dsd->file_index.dsd_recordcrcseed = recordcrcseed;
  dsd->file_index.dsd_serialnumber = serialnumber;

  /* Create file indexes */
  interim_result = AdcIntDSDWriteFileIndexes(dsd);
  if (ADC_DSD_SUCCESS != interim_result)
  {
    result = interim_result;
    goto aidcn_fail;
  }

  /* Make file the right size */
  filesize = AdcIntDSDCalculateFileSize(dsd);
  if (ftruncate(dsd->fd, filesize) < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidcn_fail;
  }

  /* Allocate work buffer */
  dsd->work_buffer = malloc(dsd->file_index.dsd_recordsize);
  if (NULL == dsd->work_buffer)
  {
    result = ADC_DSD_ERROR_MALLOC;
    goto aidcn_fail;
  }

  /* Success */
  return ADC_DSD_SUCCESS;

aidcn_fail:
  (void) AdcIntDSDClose(dsd);
  return result;
}


/**
 * Open an existing DSD read for use.
 * Will fail if file doesn't exist.
 * DSD parameters will be retrieved from the file. (no locking)
 */
enum eAdcDsdError AdcIntDSDOpenExisting(struct s_adc_dsd *dsd, const char *filename)
{
  enum eAdcDsdError result;
  enum eAdcDsdError interim_result;

  /* Check pointers */
  if (NULL == dsd || NULL == filename)
  {
    return ADC_DSD_ERROR_NULL;
  }

  /* Take copy of filename */
  dsd->filename = strdup(filename);
  if (NULL == dsd->filename)
  {
    result = ADC_DSD_ERROR_MALLOC;
    goto aidoe_fail;
  }

  /* Open file, don't create if not present */
  dsd->fd = open(dsd->filename, O_RDWR);
  if (dsd->fd < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidoe_fail;
  }
  if (fcntl(dsd->fd, F_SETFD, FD_CLOEXEC) < 0)
  {
    result = ADC_DSD_ERROR_FILE;
    goto aidoe_fail;
  }

  /* Read DSD file index from file */
  interim_result = AdcIntDSDReadFileIndexes(dsd);
  if (ADC_DSD_SUCCESS != interim_result)
  {
    result = interim_result;
    goto aidoe_fail;
  }

  /* Allocate work buffer */
  dsd->work_buffer = malloc(dsd->file_index.dsd_recordsize);
  if (NULL == dsd->work_buffer)
  {
    result = ADC_DSD_ERROR_MALLOC;
    goto aidoe_fail;
  }

  /* Ready to go */
  return ADC_DSD_SUCCESS;

aidoe_fail:
  (void) AdcIntDSDClose(dsd);
  return result;
}


/** Check if the DSD is up and working (file has been opened) */
enum eAdcDsdError AdcIntDSDIsRunning(struct s_adc_dsd *dsd)
{
  if (NULL == dsd)
  {
    return ADC_DSD_ERROR_NULL;
  }
  return (dsd->fd >= 0) ? ADC_DSD_SUCCESS : ADC_DSD_ERROR_FILE;
}
