/*
 * dsd_interface.c
 *
 *  Created on: Mar 7, 2012
 *      Author: ian
 */
#include "dsd_copy.h"
#include "dsd_fileindex.h"
#include "dsd_init.h"
#include "dsd_purge.h"
#include "dsd_read.h"
#include "dsd_verify.h"
#include "dsd_write.h"

#include <rinweigh/adcdsd.h>


#define ADC_DSD_MUTEX_LOCK() \
  enum eAdcDsdError result; \
\
  if (lock) \
  { \
    result = AdcDSDMutexLock(dsd); \
    if (ADC_DSD_SUCCESS != result) \
    { \
      return result; \
    } \
  }


#define ADC_DSD_MUTEX_UNLOCK() \
  if (lock) \
  { \
    AdcDSDMutexUnlock(dsd); \
  } \
  return result


/*
 * DSD file open/close functions
 */


/**
 * Create a new DSD and ready it for use.
 * Will wipe the file if it already exists
 * DSD parameters will be saved to the file.
 * (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param filename is the name of the DSD file to be created
 * @param recordsize is the size in bytes of a DSD record
 * @param maxrecords is the maximum number of records allowed in the DSD
 * @param recordcrcseed is the 16bit CRC seed used to seed the checksum of each record
 * @param serialnumber is the serial number of the device creating the DSD
 */
enum eAdcDsdError AdcDSDCreateNew(struct s_adc_dsd *dsd, bool lock, const char *filename, int32u recordsize, int32u maxrecords, int16u recordcrcseed, int32u serialnumber)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDCreateNew(dsd, filename, recordsize, maxrecords, recordcrcseed, serialnumber);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Open an existing DSD read for use.
 * Will fail if file doesn't exist.
 * DSD parameters will be retrieved from the file. (no locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param filename is the name of the DSD file to be opened
 */
enum eAdcDsdError AdcDSDOpenExisting(struct s_adc_dsd *dsd, bool lock, const char *filename)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDOpenExisting(dsd, filename);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Close the current DSD. dsd structure remains. (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 */
enum eAdcDsdError AdcDSDClose(struct s_adc_dsd *dsd, bool lock)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDClose(dsd);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Check if the DSD is up and working (file has been opened)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 */
enum eAdcDsdError AdcDSDIsRunning(struct s_adc_dsd *dsd, bool lock)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDIsRunning(dsd);
  ADC_DSD_MUTEX_UNLOCK();
}


/*
 * Writing functions
 */


/**
 * Write a record into the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param buffer is a pointer to a buffer for the record to be written
 * @param buflen is the length of the buffer
 */
enum eAdcDsdError AdcDSDWriteRecord(struct s_adc_dsd *dsd, bool lock, const void *buffer, const int16u buflen)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDWriteRecord(dsd, buffer, buflen);
  ADC_DSD_MUTEX_UNLOCK();
}


/*
 * Read functions
 */


/**
 * Read a record from the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param record is the DSD record number to read
 * @param buffer is a pointer to a buffer for the record to be read into
 * @param buflen is the length of the buffer
 */
enum eAdcDsdError AdcDSDReadRecord(struct s_adc_dsd *dsd, bool lock, int32u record, void *buffer, int16u buflen)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDReadRecord(dsd, record, buffer, buflen);
  ADC_DSD_MUTEX_UNLOCK();
}


/*
 * Management functions
 */


/**
 * Get the number of records currently held in the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param records is a pointer to int32u set on success to the number of records in the DSD. Not NULL.
 */
enum eAdcDsdError AdcDSDGetNumberRecords(struct s_adc_dsd *dsd, bool lock, int32u *records)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDGetNumberRecords(dsd, records);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Get the total number of records that will fit in the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param records is a pointer to int32u set on success to the capacity of the DSD. Not NULL.
 */
enum eAdcDsdError AdcDSDGetCapacity(struct s_adc_dsd *dsd, bool lock, int32u *records)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDGetCapacity(dsd, records);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Get the record size for the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param recordsize is a pointer to int32u set on success to the number of bytes of a record. Not NULL.
 */
enum eAdcDsdError AdcDSDGetRecordSize(struct s_adc_dsd *dsd, bool lock, int32u *recordsize)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDGetRecordSize(dsd, recordsize);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Read a file index from the DSD file.
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param which is 0 for the first index, or 1 for the second index
 * @param fi is the returned file index. Not NULL
 */
enum eAdcDsdError AdcDSDReadFileIndex(struct s_adc_dsd *dsd, bool lock, int which, struct s_adc_dsd_file_index *fi)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDReadFileIndex(dsd, which, fi);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Verify the integrity of all records in the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param fix if true, then tries to fix any problems
 * @param record_num is set to the first record found with a problem
 */
enum eAdcDsdError AdcDSDVerifyIntegrity(struct s_adc_dsd *dsd, bool lock, bool fix, size_t *record_num)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDVerifyIntegrity(dsd, fix, record_num);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Make room for more records, by purging num_records from the start of the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param num_records is the number of records to purge from the DSD.
 */
enum eAdcDsdError AdcDSDPurge(struct s_adc_dsd *dsd, bool lock, size_t num_records)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDPurge(dsd, num_records);
  ADC_DSD_MUTEX_UNLOCK();
}


/*
 * Backup operations
 */


/**
 * Copy the DSD file to copyfd (with locking)
 *
 * Writing within copyfd occurs from the write position in the file when the function is called.
 * No truncation is performed.
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param copyfd is a file descriptor open for writing.
 * @param bytes_written is set to the number of bytes written to copyfd. Can be NULL
 */
enum eAdcDsdError AdcDSDFileCopyFD(struct s_adc_dsd *dsd, bool lock, int copyfd, int32u *bytes_written)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDFileCopyFD(dsd, copyfd, bytes_written);
  ADC_DSD_MUTEX_UNLOCK();
}


/**
 * Copy the DSD file to filename (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param filename is the name of a writeable file to copy the DSD to. File will be overwritten.
 * @param bytes_written is set to the number of bytes written to copyfd. Can be NULL
 */
enum eAdcDsdError AdcDSDFileCopyFilename(struct s_adc_dsd *dsd, bool lock, const char *filename, int32u *byteswritten)
{
  ADC_DSD_MUTEX_LOCK();
  result = AdcIntDSDFileCopyFilename(dsd, filename, byteswritten);
  ADC_DSD_MUTEX_UNLOCK();
}
