/*
 * el_runcmd.h
 *
 *  Created on: Sep 17, 2012
 *      Author: ian
 */

#ifndef EL_RUNCMD_H_
#define EL_RUNCMD_H_

#include <unistd.h>
#include <stddef.h>
#include <stdbool.h>


/** Result codes for exec library functions */
enum e_el_result { elr_success = 0, elr_execfail, elr_allocfail, elr_iofail, elr_bufferfull };


/** What to read back from an executed command */
enum e_el_read_pipe
{
  erp_stdout = 0,         /*!< Read only stdout back from the executed command */
  erp_stderr,             /*!< Read only stderr back from the executed command */
  erp_stdout_and_stderr   /*!< Read both stdout and stderr back from the executed command. The order of output between stdout and stderr is not guaranteed. */
};


/** Function pointer used as a callback for reading data from the forked process (ie read from stdout, stderr) */
typedef enum e_el_result (*read_cmd_pipe_fn_t)(void *user_data, const char *data, size_t bytes_read);


/** Function pointer used as a callback for writing data to the forked process (ie write to stdin) */
typedef enum e_el_result (*write_cmd_pipe_fn_t)(void *user_data, char *data, size_t max_data_len, size_t *data_len);


/** Structure for communicating via stdin/stdout/stderr with the forked process */
struct s_runcmd_pipe_config
{
  void *read_cmd_data_stdout;                   /*!< user_data pointer passed when calling read_cmd_pipe_stdout_fn */
  void *read_cmd_data_stderr;                   /*!< user_data pointer passed when calling read_cmd_pipe_stderr_fn */
  void *write_cmd_data_stdin;                   /*!< user_data pointer passed when calling write_cmd_pipe_stdin_fn */
  read_cmd_pipe_fn_t read_cmd_pipe_stdout_fn;   /*!< Function called when data is read from child process's stdout */
  read_cmd_pipe_fn_t read_cmd_pipe_stderr_fn;   /*!< Function called when data is read from child process's stderr */
  write_cmd_pipe_fn_t write_cmd_pipe_stdin_fn;  /*!< Function called to retrieve data to be written to the child process's stdin */
};


/**
 * Run cmd with arguments args, and return cmd_ret_code
 * Depending on the configuration of rcc, data can be sent and received from the child process
 *
 * @param cmd is the command to run, not NULL
 * @param args is the arguments to pass to the cmd (including arg[0] is the command), not NULL
 * @param cmd_ret_code is a pointer to an integer that is set to the result of the command. Can be NULL if !wait_complete, otherwise not NULL.
 * @param rcc is a pre-configured structure that defines how data should be sent to and received from the child process. Can be NULL if this is not required.
 * @param wait_complete if true, then will wait until the child process exits, otherwise returns immediately. Will still delay if rcc is being used to send/receive data.
 * @param search_path if true, then execvp() will be used (will search path for cmd), otherwise execv() will be used (must specify exact location of cmd)
 * @param close_all_fds if true, then all file descriptors except stdin, stdout, stderr will be closed in the child prior to the execv() call.
 *
 * @return one of the @ref enum e_el_result result codes
 */
enum e_el_result el_run_command(const char *cmd, const char *const args[], int *cmd_ret_code,
                                struct s_runcmd_pipe_config *rcc,
                                bool wait_complete,
                                bool search_path,
                                bool close_all_fds);


/** Initialise a s_runcmd_pipe_config structure (to empty, ie no data will be sent or received from the child process) */
void el_init_runcmd_pipe_config(struct s_runcmd_pipe_config *rcc);


/**
 * Convert cmd_ret_code of the child process from @ref el_run_command into the exited flag.
 * If you only want to compile for posix systems, you can use WIFEXITED, et el.
 * This function is for running on both posix and windows systems.
 *
 * @param cmd_ret_code is the value returned from el_run_command for the child process
 *
 * @return whether the child process has exited
 */
bool el_return_code_has_exited(int cmd_ret_code);


/**
 * Convert cmd_ret_code of the child process from @ref el_run_command into the exit code.
 * If you only want to compile for posix systems, you can use WEXITSTATUS, et el.
 * This function is for running on both posix and windows systems.
 *
 * @param cmd_ret_code is the value returned from el_run_command for the child process
 *
 * @return the return code from the child process
 */
unsigned el_convert_return_code(int cmd_ret_code);


#endif /* EL_RUNCMD_H_ */
