/*
 * el_runcmd_win32.c
 *
 * Windows implementation of el_run_command
 *
 *  Created on: Apr 21, 2017
 *      Author: ian
 */
#include <stdbool.h>
#include <stdio.h>
#include <windows.h>

#include <rinexec/el_runcmd.h>

#define WIN_WIFEXITED_SET(status, value) { if (value) status &= ~0x7F; else status |= 0x7F; }
#define WIN_WEXITSTATUS_SET(status, value) { status &= ~0xFF00; status |= ((value << 8) & 0xFF00); }


/** Child process has exited for Windows */
#define WIN_WIFEXITED(status) ((status & 0x7F) == 0)


/** Child process return code for Windows */
#define WIN_WEXITSTATUS(status) ((status & 0xFF00) >> 8)


static char *internal_args_to_string(const char *const args[])
{
  size_t arg_num;
  size_t length;
  char *argstring;

  /* Work out the length of the string */
  for (arg_num = 0, length = 0; NULL != args[arg_num]; arg_num++)
  {
    length += strlen(args[arg_num]) + sizeof(char);
  }

  /* Allocate the string for the space separated arguments*/
  argstring = calloc(length, sizeof(char));

  /* Copy arguments into string with space separator */
  for (arg_num = 0, length = 0; NULL != args[arg_num]; arg_num++)
  {
    strcat(argstring, args[arg_num]);
    strcat(argstring, " ");
  }
  return argstring;
}


/**
 * Run cmd with arguments args, and return cmd_ret_code
 * Depending on the configuration of rcc, data can be sent and received from the child process
 *
 * @param cmd is the command to run, not NULL
 * @param args is the arguments to pass to the cmd (including arg[0] is the command), not NULL
 * @param cmd_ret_code is a pointer to an integer that is set to the result of the command. Can be NULL if !wait_complete, otherwise not NULL.
 * @param rcc is a pre-configured structure that defines how data should be sent to and received from the child process. Can be NULL if this is not required.
 * @param wait_complete if true, then will wait until the child process exits, otherwise returns immediately. Will still delay if rcc is being used to send/receive data.
 * @param search_path if true, then execvp() will be used (will search path for cmd), otherwise execv() will be used (must specify exact location of cmd)
 * @param close_all_fds if true, then all file descriptors except stdin, stdout, stderr will be closed in the child prior to the execv() call.
 *
 * @return one of the @ref enum e_el_result result codes
 */
enum e_el_result el_run_command(const char *cmd, const char *const args[], int *cmd_ret_code,
                                struct s_runcmd_pipe_config *rcc,
                                bool wait_complete,
                                bool search_path,
                                bool close_all_fds)
{
  /* Only connect up pipes if actually required */
  bool need_child_to_parent;
  bool need_parent_to_child;

  enum e_el_result result;
  STARTUPINFO startup_info;
  PROCESS_INFORMATION process_info;
  SECURITY_ATTRIBUTES saAttr;
  HANDLE hChildStd_IN_Rd = NULL;
  HANDLE hChildStd_IN_Wr = NULL;
  HANDLE hChildStd_OUT_Rd = NULL;
  HANDLE hChildStd_OUT_Wr = NULL;
  char *argstring;
  DWORD wait_result;
  DWORD win_exit_code;

  need_child_to_parent = (NULL != rcc && NULL != rcc->read_cmd_pipe_stdout_fn);
  need_parent_to_child = (NULL != rcc && NULL != rcc->write_cmd_pipe_stdin_fn);

  /* Set the bInheritHandle flag so pipe handles are inherited */
  saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
  saAttr.bInheritHandle = true;
  saAttr.lpSecurityDescriptor = NULL;

  if (need_child_to_parent)
  {
    /* Create a pipe for the child process's stdout */
    if (!CreatePipe(&hChildStd_OUT_Rd, &hChildStd_OUT_Wr, &saAttr, 0))
    {
      result = elr_iofail;
      goto erc_fail;
    }

    /* Ensure the read handle to the pipe for stdout is not inherited */
    if (!SetHandleInformation(hChildStd_OUT_Rd, HANDLE_FLAG_INHERIT, 0))
    {
      result = elr_iofail;
      goto erc_fail;
    }
  }

  if (need_parent_to_child)
  {
    /* Create a pipe for the child process's stdin */
    if (!CreatePipe(&hChildStd_IN_Rd, &hChildStd_IN_Wr, &saAttr, 0))
    {
      result = elr_iofail;
      goto erc_fail;
    }

    /* Ensure the write handle to the pipe for STDIN is not inherited */
    if (!SetHandleInformation(hChildStd_IN_Wr, HANDLE_FLAG_INHERIT, 0))
    {
      result = elr_iofail;
      goto erc_fail;
    }
  }

  /* Setup startup info */
  memset(&startup_info, 0, sizeof(startup_info));
  startup_info.cb = sizeof(startup_info);
  startup_info.hStdError = hChildStd_OUT_Wr;
  startup_info.hStdOutput = hChildStd_OUT_Wr;
  startup_info.hStdInput = hChildStd_IN_Rd;
  startup_info.dwFlags |= STARTF_USESTDHANDLES;

  /* Convert unix arg array into windows command line */
  argstring = internal_args_to_string(args);

  /* Run process */
  if (!CreateProcess(cmd,                           /* LPCTSTR lpApplicationName */
                     argstring,                     /* LPTSTR lpCommandLine */
                     NULL,                          /* LPSECURITY_ATTRIBUTE lpProcessAttributes */
                     NULL,                          /* LPSECURITY_ATTRIBUTES lpThreadAttributes */
                     !close_all_fds,                /* BOOL bInheritHandles */
                     0,                             /* DWORD dwCreationFlags */
                     NULL,                          /* LPVOID lpEnvironment */
                     NULL,                          /* LPCTSTR lpCurrentDirectory */
                     &startup_info,                 /* LPSTARTUPINFO lpStartupInfo */
                     &process_info))                /* LPPROCESS_INFORMATION lpProcessInformation */
  {
    result = elr_execfail;
    goto erc_fail;
  }

  /* Wait for completion and get child process result code */
  if (wait_complete)
  {
    *cmd_ret_code = 0;
    wait_result = WaitForSingleObject(process_info.hProcess, INFINITE);
    WIN_WIFEXITED_SET(*cmd_ret_code, (WAIT_OBJECT_0 == wait_result));
    if (!GetExitCodeProcess(process_info.hProcess, &win_exit_code))
    {
      result = elr_execfail;
      goto erc_fail;
    }
    WIN_WEXITSTATUS_SET(*cmd_ret_code, win_exit_code);
  }
  CloseHandle(process_info.hThread);
  CloseHandle(process_info.hProcess);
  result = elr_success;

erc_fail:
  return result;
}


/**
 * Convert cmd_ret_code of the child process from @ref el_run_command into the exited flag.
 * If you only want to compile for posix systems, you can use WIFEXITED, et el.
 * This function is for running on both posix and windows systems.
 *
 * @param cmd_ret_code is the value returned from el_run_command for the child process
 *
 * @return whether the child process has exited
 */
bool el_return_code_has_exited(int cmd_ret_code)
{
  return WIN_WIFEXITED(cmd_ret_code);
}


/**
 * Convert cmd_ret_code of the child process from @ref el_run_command into the exit code.
 * If you only want to compile for posix systems, you can use WEXITSTATUS, et el.
 * This function is for running on both posix and windows systems.
 *
 * @param cmd_ret_code is the value returned from el_run_command for the child process
 *
 * @return the return code from the child process
 */
unsigned el_convert_return_code(int cmd_ret_code)
{
  return WIN_WEXITSTATUS(cmd_ret_code);
}
