/*
 * el_test_pipesize.c
 *
 * There was a bug where incorrect data would be written
 * if the buffer used by the pipe between child and parent was
 * smaller than BUFSIZ used in read_write_cmd_pipes()
 *
 *  Created on: Nov 6, 2017
 *      Author: ian
 */
#define _GNU_SOURCE
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdint.h>
#include <cutest/CuTest.h>
#include <rinexec/el_runcmd.h>
#include <rinexec/el_pipe.h>
#include <rinexec/el_internal.h>

#include "el_test_pipesize.h"


/* This will only work if F_GETPIPE_SZ and F_SETPIPE_SZ are available */
#if defined(F_GETPIPE_SZ) && defined(F_SETPIPE_SZ)


struct s_big_write_data
{
  size_t writebufsize;
  size_t writeposn;
  char *writebuffer;
  char *readbuffer;
  size_t readposn;
  size_t readbufsize;
};


static enum e_el_result big_write_fn(void *user_data, char *data, size_t max_data_len, size_t *data_len)
{
  struct s_big_write_data *bwd = (struct s_big_write_data *) user_data;
  size_t can_write;

  if (max_data_len > bwd->writebufsize - bwd->writeposn)
    can_write = bwd->writebufsize - bwd->writeposn;
  else
    can_write = max_data_len;
  memcpy(data, &bwd->writebuffer[bwd->writeposn], can_write);
  bwd->writeposn += can_write;
  *data_len = can_write;

  /* printf("write: max_data_len %u, write %u, posn %u, size %u\n", (unsigned) max_data_len, (unsigned) can_write, (unsigned) bwd->writeposn, (unsigned) bwd->writebufsize); */
  return elr_success;
}


static enum e_el_result big_read_fn(void *user_data, const char *data, size_t bytes_read)
{
  struct s_big_write_data *bwd = (struct s_big_write_data *) user_data;
  size_t can_read;

  if (bytes_read > bwd->readbufsize - bwd->readposn)
    can_read = bwd->readbufsize - bwd->readposn;
  else
    can_read = bytes_read;
  memcpy(&bwd->readbuffer[bwd->readposn], data, can_read);
  bwd->readposn += can_read;

  /* printf("read: avail %u, read %u, posn %u, size %u\n", (unsigned) bytes_read, (unsigned) can_read, (unsigned) bwd->readposn, (unsigned) bwd->readbufsize); */
  return elr_success;
}


/** Test for bug where incorrect data can be written to child stdin pipe */
static void test_read_write_cmd_pipes(CuTest* tc)
{
  struct s_big_write_data big_write_data;
  struct s_runcmd_pipe_config rcpc;
  int child_stdin_pipe[2];
#define parent_to_child_stdin_read_fd child_stdin_pipe[0]
#define parent_to_child_stdin_write_fd child_stdin_pipe[1]
  size_t i;
  enum e_el_result result;
  bool outfd_isopen;
  int pipebuffersize;
  char msg[100];

  /* Create a pipe */
  CuAssertIntEquals(tc, 0, pipe2(child_stdin_pipe, O_NONBLOCK));

  /* Need to set the pipe size to smaller than BUFSIZ for bug to occur */
  pipebuffersize = fcntl(parent_to_child_stdin_read_fd, F_SETPIPE_SZ, BUFSIZ / 2);
  CuAssertIntEquals_Msg(tc, "This test is only valid if F_SETPIPE_SZ can be set smaller than BUFSIZ", BUFSIZ / 2, pipebuffersize);
  big_write_data.writebufsize = BUFSIZ;
  /* printf("F_GETPIPE_SZ = %d\n", pipebuffersize); */

  /* Now allocate a buffer of random numbers */
  big_write_data.writebuffer = malloc(big_write_data.writebufsize);
  CuAssertPtrNotNull(tc, big_write_data.writebuffer);
  for (i = 0; i < big_write_data.writebufsize; i++)
    big_write_data.writebuffer[i] = rand() % 256;
  big_write_data.writeposn = 0;

  big_write_data.readbufsize = big_write_data.writebufsize;
  big_write_data.readbuffer = malloc(big_write_data.readbufsize);
  CuAssertPtrNotNull(tc, big_write_data.readbuffer);
  big_write_data.readposn = 0;

  /* Setup callback for writing to librinexectest_waitsigwrite's stdin */
  el_init_runcmd_pipe_config(&rcpc);
  el_init_write_cmd_pipe_callback(&rcpc, big_write_fn, &big_write_data);
  el_init_read_cmd_pipe_callback(&rcpc, erp_stdout, big_read_fn, &big_write_data);

  /* Do the pipe read and write using pipe as a fake child process */
  outfd_isopen = true;
  result = read_write_cmd_pipes(parent_to_child_stdin_read_fd, -1, parent_to_child_stdin_write_fd, &outfd_isopen, &rcpc);
  CuAssertIntEquals(tc, elr_success, result);

  /* Check that what ended up in the read buffer is the same as in the write buffer */
  for (i = 0; i < big_write_data.readbufsize; i++)
  {
    snprintf(msg, 100, "read data (0x%02X) doesn't match write data (0x%02X) at offset %u",
             (uint8_t) big_write_data.readbuffer[i], (uint8_t) big_write_data.writebuffer[i], (unsigned) i);
    CuAssertIntEquals_Msg(tc, msg, big_write_data.writebuffer[i], big_write_data.readbuffer[i]);
  }

  /* Clean up */
  close(parent_to_child_stdin_read_fd);
  close(parent_to_child_stdin_write_fd);
  free(big_write_data.writebuffer);
  free(big_write_data.readbuffer);
}


#endif /* #if defined(F_GETPIPE_SZ) && defined(F_SETPIPE_SZ) */


CuSuite *CuGetSuite_runcmd_pipesize_bug(void)
{
  CuSuite *suite = CuSuiteNew();

#if defined(F_GETPIPE_SZ) && defined(F_SETPIPE_SZ)
  SUITE_ADD_TEST(suite, test_read_write_cmd_pipes);
#endif /* #if defined(F_GETPIPE_SZ) && defined(F_SETPIPE_SZ) */

  return suite;
}
