/*
 * el_test_read.c
 *
 *  Created on: Sep 17, 2012
 *      Author: ian
 */
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <cutest/CuTest.h>
#include <rinexec/el_runcmd.h>
#include <rinexec/el_readpipe_buffer.h>
#include <rinexec/el_readpipe_bufalloc.h>
#include <rinexec/el_readpipe_file.h>
#include <rinexec/el_pipe.h>

#include "el_test_read.h"


/** Test running a command and reading back into a buffer */
static void test_readbuffer(CuTest* tc)
{
  #define HELLO_WORLD_STR "hello world"
  #define SLASH_N_STR "\n"
  const char *const echo_cmd[] = { "echo", HELLO_WORLD_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  char buffer[BUFFER_SIZE];
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_buffer rcb;

  /* Fill buffer with non-null data */
  memset(buffer, 'a', BUFFER_SIZE);
  buffer[BUFFER_SIZE - 1] = '\0';

  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_buffer(&rcpc, &rcb, buffer, BUFFER_SIZE, erp_stdout);

  /* Run "echo hello world", and wait for the command to complete */
  result = el_run_command(echo_cmd[0], echo_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_success, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check for captured hello world in the buffer */
  CuAssertIntEquals(tc, strlen(HELLO_WORLD_STR SLASH_N_STR), rcb.data_posn);
  CuAssertTrue(tc, 0 == memcmp(rcb.data, HELLO_WORLD_STR SLASH_N_STR, strlen(HELLO_WORLD_STR SLASH_N_STR)));
}


/** Test running a command and reading back into a string */
static void test_readstring(CuTest* tc)
{
  #define HELLO_WORLD_STR "hello world"
  #define SLASH_N_STR "\n"
  const char *const echo_cmd[] = { "echo", HELLO_WORLD_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  char buffer[BUFFER_SIZE];
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_buffer rcb;

  /* Fill buffer with non-null data */
  memset(buffer, 'a', BUFFER_SIZE);
  buffer[BUFFER_SIZE - 1] = '\0';

  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_string(&rcpc, &rcb, buffer, BUFFER_SIZE, erp_stdout);

  /* Run "echo hello world", and wait for the command to complete */
  result = el_run_command(echo_cmd[0], echo_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_success, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check for captured hello world in the buffer */
  CuAssertIntEquals(tc, strlen(HELLO_WORLD_STR SLASH_N_STR), rcb.data_posn);
  CuAssertTrue(tc, 0 == strcmp(rcb.data, HELLO_WORLD_STR SLASH_N_STR));
}


/** Test running a command and reading back into a buffer that is too short */
static void test_readbuffer_tooshort(CuTest* tc)
{
  #define HELLO_WORLD_STR "hello world"
  #define SLASH_N_STR "\n"
  const char *const echo_cmd[] = { "echo", HELLO_WORLD_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  char buffer[BUFFER_SIZE];
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_buffer rcb;
  int buflen;

  /* Fill buffer with non-null data */
  memset(buffer, 'a', BUFFER_SIZE);
  buffer[BUFFER_SIZE - 1] = '\0';

  buflen = strlen(HELLO_WORLD_STR) / 2;
  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_buffer(&rcpc, &rcb, buffer, buflen, erp_stdout);

  /* Run "echo hello world", and wait for the command to complete */
  result = el_run_command(echo_cmd[0], echo_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_bufferfull, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check for captured 1/2 of "hello world" in the buffer */
  CuAssertIntEquals(tc, buflen, rcb.data_posn);
  CuAssertTrue(tc, 0 == memcmp(rcb.data, HELLO_WORLD_STR SLASH_N_STR, buflen));
  CuAssertTrue(tc, 0 != memcmp(rcb.data, HELLO_WORLD_STR SLASH_N_STR, strlen(HELLO_WORLD_STR SLASH_N_STR)));
}


/** Test running a command and reading back into a string that is too short */
static void test_readstring_tooshort(CuTest* tc)
{
  #define HELLO_WORLD_STR "hello world"
  #define SLASH_N_STR "\n"
  const char *const echo_cmd[] = { "echo", HELLO_WORLD_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  char buffer[BUFFER_SIZE];
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_buffer rcb;
  int buflen;

  /* Fill buffer with non-null data */
  memset(buffer, 'a', BUFFER_SIZE);
  buffer[BUFFER_SIZE - 1] = '\0';

  buflen = strlen(HELLO_WORLD_STR) / 2;
  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_string(&rcpc, &rcb, buffer, buflen, erp_stdout);

  /* Run "echo hello world", and wait for the command to complete */
  result = el_run_command(echo_cmd[0], echo_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_bufferfull, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check for captured hello world in the buffer */
  CuAssertIntEquals(tc, buflen - sizeof(char), rcb.data_posn);
  CuAssertIntEquals(tc, '\0', rcb.data[rcb.data_posn]);
  CuAssertTrue(tc, 0 == strncmp(rcb.data, HELLO_WORLD_STR SLASH_N_STR, buflen - sizeof(char)));
  CuAssertTrue(tc, 0 != strncmp(rcb.data, HELLO_WORLD_STR SLASH_N_STR, strlen(HELLO_WORLD_STR SLASH_N_STR)));
}


/** Test running a command and reading back into a FILE */
static void test_readfile(CuTest* tc)
{
  #define HELLO_WORLD_STR "hello world"
  #define SLASH_N_STR "\n"
  const char *const echo_cmd[] = { "echo", HELLO_WORLD_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  char buffer[BUFFER_SIZE];
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_file rcf;
  FILE *file;
  size_t len;

  /* Create temporary file */
  file = tmpfile();
  CuAssertPtrNotNull(tc, file);

  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_file(&rcpc, &rcf, file, erp_stdout);

  /* Run "echo hello world", and wait for the command to complete */
  result = el_run_command(echo_cmd[0], echo_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_success, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check the result in the file */
  len = strlen(HELLO_WORLD_STR SLASH_N_STR);
  CuAssertIntEquals(tc, 0, fseek(file, 0, SEEK_SET));
  CuAssertIntEquals(tc, len, fread(buffer, sizeof(char), BUFFER_SIZE, file));
  CuAssertTrue(tc, 0 == memcmp(buffer, HELLO_WORLD_STR SLASH_N_STR, len));

  fclose(file);
}


/** Test running a command and reading back into a dynamically allocated buffer */
static void test_readbufalloc(CuTest* tc)
{
  #define HELLO_WORLD_STR "hello world"
  #define SLASH_N_STR "\n"
  const char *const echo_cmd[] = { "echo", HELLO_WORLD_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_bufferalloc rcba;

  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_bufferalloc(&rcpc, &rcba, BUFFER_SIZE, erp_stdout);

  /* Run "echo hello world", and wait for the command to complete */
  result = el_run_command(echo_cmd[0], echo_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_success, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check for captured hello world in the buffer */
  CuAssertIntEquals(tc, strlen(HELLO_WORLD_STR SLASH_N_STR), rcba.data_posn);
  CuAssertIntEquals(tc, BUFFER_SIZE, rcba.data_max_len);
  CuAssertIntEquals(tc, BUFFER_SIZE, rcba.data_alloc_increments);
  CuAssertTrue(tc, 0 == memcmp(rcba.data, HELLO_WORLD_STR SLASH_N_STR, strlen(HELLO_WORLD_STR SLASH_N_STR)));

  el_free_read_cmd_pipe_bufferalloc(&rcba);
}


/** Test running a command and reading back into a dynamically allocated string */
static void test_readstralloc(CuTest* tc)
{
  #define HELLO_WORLD_STR "hello world"
  #define SLASH_N_STR "\n"
  const char *const echo_cmd[] = { "echo", HELLO_WORLD_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_bufferalloc rcba;

  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_stringalloc(&rcpc, &rcba, BUFFER_SIZE, erp_stdout);

  /* Run "echo hello world", and wait for the command to complete */
  result = el_run_command(echo_cmd[0], echo_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_success, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check for captured hello world in the buffer */
  CuAssertIntEquals(tc, strlen(HELLO_WORLD_STR SLASH_N_STR), rcba.data_posn);
  CuAssertIntEquals(tc, BUFFER_SIZE, rcba.data_max_len);
  CuAssertIntEquals(tc, BUFFER_SIZE, rcba.data_alloc_increments);
  CuAssertTrue(tc, 0 == strcmp(rcba.data, HELLO_WORLD_STR SLASH_N_STR));

  el_free_read_cmd_pipe_stringalloc(&rcba);
}


/** Test reading back from a command that writes stdout and stderr into separate buffers */
static void test_read_stdout_stderr_separate(CuTest* tc)
{
  #define HELLO_STDOUT_STR "hello stdout"
  #define HELLO_STDERR_STR "hi stderr"
  #define SLASH_N_STR "\n"
  const char *const stdout_stderr_cmd[] = { "./librinexectest_echo", "stdout", HELLO_STDOUT_STR, "stderr", HELLO_STDERR_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  char buffer_stdout[BUFFER_SIZE];
  char buffer_stderr[BUFFER_SIZE];
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_buffer rcb_stdout;
  struct s_runcmd_buffer rcb_stderr;

  /* Fill buffer with non-null data */
  memset(buffer_stdout, 'a', BUFFER_SIZE);
  buffer_stdout[BUFFER_SIZE - 1] = '\0';
  memset(buffer_stderr, 'b', BUFFER_SIZE);
  buffer_stderr[BUFFER_SIZE - 1] = '\0';

  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_string(&rcpc, &rcb_stdout, buffer_stdout, BUFFER_SIZE, erp_stdout);
  el_init_read_cmd_pipe_string(&rcpc, &rcb_stderr, buffer_stderr, BUFFER_SIZE, erp_stderr);

  /* Run "librinexectest_echo ...", and wait for the command to complete */
  result = el_run_command(stdout_stderr_cmd[0], stdout_stderr_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_success, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check for captured "hello stdout" in the stdout buffer */
  CuAssertIntEquals(tc, strlen(HELLO_STDOUT_STR SLASH_N_STR), rcb_stdout.data_posn);
  CuAssertTrue(tc, 0 == strcmp(rcb_stdout.data, HELLO_STDOUT_STR SLASH_N_STR));

  /* Check for captured "hi stderr" in the stderr buffer */
  CuAssertIntEquals(tc, strlen(HELLO_STDERR_STR SLASH_N_STR), rcb_stderr.data_posn);
  CuAssertTrue(tc, 0 == strcmp(rcb_stderr.data, HELLO_STDERR_STR SLASH_N_STR));
}


/** Test reading back from a command that writes stdout and stderr into a combined buffer */
static void test_read_stdout_stderr_combined(CuTest* tc)
{
  #define HELLO_STDOUT_STR "hello stdout"
  #define HELLO_STDERR_STR "hi stderr"
  #define SLASH_N_STR "\n"
  const char *const stdout_stderr_cmd[] = { "./librinexectest_echo", "stdout", HELLO_STDOUT_STR, "stderr", HELLO_STDERR_STR, NULL };
  enum e_el_result result;
  int ret_code;
  #define BUFFER_SIZE 100
  char buffer[BUFFER_SIZE];
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_buffer rcb;
  bool have_both;

  /* Fill buffer with non-null data */
  memset(buffer, 'a', BUFFER_SIZE);
  buffer[BUFFER_SIZE - 1] = '\0';

  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_string(&rcpc, &rcb, buffer, BUFFER_SIZE, erp_stdout_and_stderr);

  /* Run "librinexectest_echo ...", and wait for the command to complete */
  result = el_run_command(stdout_stderr_cmd[0], stdout_stderr_cmd, &ret_code, &rcpc, true, true, true);
  CuAssertIntEquals(tc, elr_success, result);
  CuAssertTrue(tc, el_return_code_has_exited(ret_code));
  CuAssertIntEquals(tc, 0, el_convert_return_code(ret_code));

  /* Check for captured "hello stdout" and "hi stderr" in the buffer. The order can't be guaranteed, so we need to check both */
  have_both = (strcmp(rcb.data, HELLO_STDOUT_STR SLASH_N_STR HELLO_STDERR_STR SLASH_N_STR) == 0 ||
               strcmp(rcb.data, HELLO_STDERR_STR SLASH_N_STR HELLO_STDOUT_STR SLASH_N_STR) == 0);
  CuAssertTrue(tc, have_both);
}


CuSuite *CuGetSuite_runcmd_read(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_readbuffer);
  SUITE_ADD_TEST(suite, test_readstring);
  SUITE_ADD_TEST(suite, test_readbuffer_tooshort);
  SUITE_ADD_TEST(suite, test_readstring_tooshort);
  SUITE_ADD_TEST(suite, test_readfile);
  SUITE_ADD_TEST(suite, test_readbufalloc);
  SUITE_ADD_TEST(suite, test_readstralloc);
  SUITE_ADD_TEST(suite, test_read_stdout_stderr_separate);
  SUITE_ADD_TEST(suite, test_read_stdout_stderr_combined);

  return suite;
}
