/*
 * rlc_assemble.c
 *
 *  Created on: 1 Jan. 2019
 *      Author: ian
 */
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <rincmd/rc_cli_error.h>
#include <rinlegacycmd/rlc_parameter.h>
#include <rinlegacycmd/rlc_parameter_escape.h>

#include <rinlegacycmd/rlc_assemble.h>


/** Used in @ref rlc_internal_format_message() to check the result of snprintf() and update the buffer offsets */
#define sncheck() \
do \
{ \
  if (snres < 0 || snres >= (int) buf_remain) \
    goto ricts_fail; \
  buf_pos += snres; \
  buf_remain -= snres; \
} \
while (0)


/** Used in @ref rlc_internal_format_message() to check the result of a @ref enum e_rc_escape_result and update the buffer offsets */
#define rercheck() \
do \
{ \
  if (rer_success != reres) \
    goto ricts_fail; \
  buf_pos = strlen(buffer); \
  buf_remain = buffer_len - buf_pos; \
} \
while (0)


/**
 * Convert a rinstrum legacy response into a string suitable for the indicator to send as a response
 *
 * @param buffer is the buffer to output the message to
 * @param buffer_len is the length of buffer
 * @param rlc_cmd is the rinstrum legacy command (ie IDN)
 * @param request_not_response true if the message is a request, false if the message is a response
 * @return e_rc_cli_result result code
 */
static enum e_rc_cli_result rlc_internal_command_to_string(char *buffer, size_t buffer_len,
                                                           const struct s_rlc_command *rlc_cmd, bool request_not_response)
{
  size_t buf_remain;
  size_t buf_pos;
  int snres;
  size_t param;
  enum e_rc_escape_result reres;

  buf_pos = 0;
  buf_remain = buffer_len;
  if (request_not_response)
  {
    snres = snprintf(&buffer[buf_pos], buf_remain, "%.3s%s", rlc_cmd->command, (rlc_cmd->command_not_query) ? "" : "?");
    sncheck();
  }

  /* And convert each of the parameters to a string */
  for (param = 0; param < rlc_cmd->num_parameters; param++)
  {
    switch (rlc_cmd->parameters[param].param_type)
    {
      case rpt_buffer:
        snres = snprintf(&buffer[buf_pos], buf_remain, "\"");
        sncheck();

        reres = rlc_parameter_escape_buffer(&buffer[buf_pos], buf_remain,
                                            rlc_cmd->parameters[param].buffer,
                                            rlc_cmd->parameters[param].buffer_length);
        rercheck();

        snres = snprintf(&buffer[buf_pos], buf_remain, "\"");
        sncheck();
        break;
      case rpt_integer:
        snres = snprintf(&buffer[buf_pos], buf_remain, "%lld",
                         (long long) rlc_cmd->parameters[param].int_value);
        sncheck();
        break;
      case rpt_error:
        snres = snprintf(&buffer[buf_pos], buf_remain, "?");
        sncheck();
        break;
      case rpt_unused:
      default:
        break;
    }
    if (param + 1 < rlc_cmd->num_parameters)
    {
      snres = snprintf(&buffer[buf_pos], buf_remain, ",");
      sncheck();
    }
  }
  return rcr_success;

ricts_fail:
  return rcr_msgassemblefail;
}


/**
 * Convert a rinstrum legacy command or query request into a string suitable for sending to the indicator
 *
 * @param buffer is the buffer to output the message to
 * @param buffer_len is the length of buffer
 * @param rlc_cmd is the rinstrum legacy command (ie IDN)
 * @param request_not_response true if the message is a request, false if the message is a response
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_command_request_to_string(char *buffer, size_t buffer_len,
                                                   const struct s_rlc_command *rlc_cmd)
{
  enum e_rc_cli_result res;

  res = rlc_internal_command_to_string(buffer, buffer_len,
                                       rlc_cmd, true);
  return res;
}


/**
 * Convert a rinstrum legacy response into a string suitable for the indicator to reply with
 *
 * @param buffer is the buffer to output the message to
 * @param buffer_len is the length of buffer
 * @param rlc_cmd is the rinstrum legacy command (ie IDN)
 * @param request_not_response true if the message is a request, false if the message is a response
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_command_response_to_string(char *buffer, size_t buffer_len,
                                                    const struct s_rlc_command *rlc_cmd)
{
  enum e_rc_cli_result res;

  res = rlc_internal_command_to_string(buffer, buffer_len,
                                       rlc_cmd, false);
  return res;
}
