/*
 * rlc_assemble.c
 *
 *  Created on: 1 Jan. 2019
 *      Author: ian
 */
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <rincmd/rc_cli_error.h>
#include <rincmd/rc_escape.h>
#include "rlc_parameter.h"
#include "rlc_parameter_escape.h"

#include "rlc_assemble_va.h"


/** Used in @ref rlc_internal_format_message() to check the result of snprintf() and update the buffer offsets */
#define sncheck() \
do \
{ \
  if (snres < 0 || snres >= (int) buf_remain) \
    goto rifm_fail; \
  buf_pos += snres; \
  buf_remain -= snres; \
} \
while (0)


/** Used in @ref rlc_internal_format_message() to check the result of a @ref enum e_rc_escape_result and update the buffer offsets */
#define rercheck() \
do \
{ \
  if (rer_success != reres) \
    goto rifm_fail; \
  buf_pos = strlen(buffer); \
  buf_remain = buffer_len - buf_pos; \
} \
while (0)


/**
 * Internal function to format a rinstrum legacy command message (request or response)
 *
 * @param buffer is the buffer to output the message to
 * @param buffer_len is the length of buffer
 * @param cmd is the rinstrum legacy command (ie IDN)
 * @param request_not_response true if the message is a request, false if the message is a response
 * @param cmd_not_query true if the message is a command, false if it is a query. Only applies to requests, not responses
 * @param param_format is a string containing format characters to match the variable argument list
 *         "n" is used when a parameter is not required
 *         "i" is a signed integer parameter (passed as int64_t)
 *         "u" is an unsigned integer parameter (passed as uint64_t)
 *         "s" is null terminated string parameter (passed as const char *)
 *         "b" is a buffer parameter (passed as const void *buf and size_t len)
 * @param args is a variable argument list corresponding to the param_format
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_internal_format_message_va(char *buffer, size_t buffer_len,
                                                    const char cmd[RLC_COMMAND_LENGTH], bool request_not_response, bool cmd_not_query,
                                                    const char *param_format, va_list args)
{
  size_t buf_remain;
  size_t buf_pos;
  int snres;
  size_t param;
  int64_t ivalue;
  uint64_t uvalue;
  const char *sptr;
  size_t blen;
  enum e_rc_escape_result reres;

  buf_pos = 0;
  buf_remain = buffer_len;
  if (request_not_response)
  {
    snres = snprintf(&buffer[buf_pos], buf_remain, "%.3s%s", cmd, (cmd_not_query) ? "" : "?");
    sncheck();
  }

  /* And convert each of the parameters to a string */
  for (param = 0; '\0' != param_format[param]; param++)
  {
    switch (param_format[param])
    {
      case 'n':
        break;

      case 'i':
        ivalue = va_arg(args, int64_t);
        snres = snprintf(&buffer[buf_pos], buf_remain, "%lld", (long long) ivalue);
        sncheck();
        break;

      case 'u':
        uvalue = va_arg(args, uint64_t);
        snres = snprintf(&buffer[buf_pos], buf_remain, "%llu", (unsigned long long) uvalue);
        sncheck();
        break;

      case 's':
        snres = snprintf(&buffer[buf_pos], buf_remain, "\"");
        sncheck();

        sptr = va_arg(args, const char *);
        reres = rlc_parameter_escape_string(&buffer[buf_pos], buf_remain, sptr);
        rercheck();

        snres = snprintf(&buffer[buf_pos], buf_remain, "\"");
        sncheck();
        break;

      case 'b':
        snres = snprintf(&buffer[buf_pos], buf_remain, "\"");
        sncheck();

        sptr = va_arg(args, const char *);
        blen = va_arg(args, size_t);
        reres = rlc_parameter_escape_buffer(&buffer[buf_pos], buf_remain, sptr, blen);
        rercheck();

        snres = snprintf(&buffer[buf_pos], buf_remain, "\"");
        sncheck();
        break;
      default:
        goto rifm_fail;
        break;
    }
    if ('\0' != param_format[param + 1])
    {
      snres = snprintf(&buffer[buf_pos], buf_remain, ",");
      sncheck();
    }
  }
  return rcr_success;

rifm_fail:
  return rcr_msgassemblefail;
}


/**
 * Convert a rinstrum legacy command or query request into a string suitable for sending to the indicator
 *
 * @param buffer is the buffer to output the message to
 * @param buffer_len is the length of buffer
 * @param cmd is the rinstrum legacy command (ie IDN)
 * @param cmd_not_query true if the message is a command, false if it is a query
 * @param param_format is a string containing format characters to match the variable argument list
 *         "n" is used when a parameter is not required
 *         "i" is a signed integer parameter (passed as int64_t)
 *         "u" is an unsigned integer parameter (passed as uint64_t)
 *         "s" is null terminated string parameter (passed as const char *)
 *         "b" is a buffer parameter (passed as const void *buf and size_t len)
 * @param args is a variable argument list corresponding to the param_format
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_format_request_va(char *buffer, size_t buffer_len,
                                           const char cmd[RLC_COMMAND_LENGTH], bool cmd_not_query,
                                           const char *param_format, va_list args)
{
  enum e_rc_cli_result res;

  res = rlc_internal_format_message_va(buffer, buffer_len,
                                       cmd, true, cmd_not_query,
                                       param_format, args);
  return res;
}


/**
 * Convert a rinstrum legacy command or query request into a string suitable for sending to the indicator
 *
 * @param buffer is the buffer to output the message to
 * @param buffer_len is the length of buffer
 * @param cmd is the rinstrum legacy command (ie IDN)
 * @param cmd_not_query true if the message is a command, false if it is a query
 * @param param_format is a string containing format characters to match the variable argument list
 *         "n" is used when a parameter is not required
 *         "i" is a signed integer parameter (passed as int64_t)
 *         "u" is an unsigned integer parameter (passed as uint64_t)
 *         "s" is null terminated string parameter (passed as const char *)
 *         "b" is a buffer parameter (passed as const void *buf and size_t len)
 * @param ... are the arguments corresponding to the param_format
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_format_request(char *buffer, size_t buffer_len,
                                        const char cmd[RLC_COMMAND_LENGTH], bool cmd_not_query, const char *param_format, ...)
{
  va_list args;
  enum e_rc_cli_result res;

  va_start(args, param_format);
  res = rlc_format_request_va(buffer, buffer_len,
                              cmd, cmd_not_query,
                              param_format, args);
  va_end(args);
  return res;
}


/**
 * Convert a rinstrum legacy response into a string suitable for the indicator to send as a response
 *
 * @param buffer is the buffer to output the message to
 * @param buffer_len is the length of buffer
 * @param cmd is the rinstrum legacy command (ie IDN)
 * @param param_format is a string containing format characters to match the variable argument list
 *         "n" is used when a parameter is not required
 *         "i" is a signed integer parameter (passed as int64_t)
 *         "u" is an unsigned integer parameter (passed as uint64_t)
 *         "s" is null terminated string parameter (passed as const char *)
 *         "b" is a buffer parameter (passed as const void *buf and size_t len)
 * @param ... are the arguments corresponding to the param_format
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_format_response(char *buffer, size_t buffer_len,
                                         const char cmd[RLC_COMMAND_LENGTH], const char *param_format, ...)
{
  va_list args;
  enum e_rc_cli_result res;

  va_start(args, param_format);
  res = rlc_internal_format_message_va(buffer, buffer_len,
                                       cmd, false, false,
                                       param_format, args);
  va_end(args);
  return res;
}
