/*
 * rlc_cli_parse.c
 *
 *  Created on: 28 Dec. 2018
 *      Author: ian
 */

#include <stdlib.h>
#include <stdint.h>
#include <rincmd/rc_frame.h>
#include <rinlegacycmd/rlc_parse.h>
#include <rinlegacycmd/rlc_assemble.h>
#include <rinlegacycmd/rlc_assemble_va.h>

#include <rinlegacycmd/rlc_cli_parse.h>


/** Create a rinstrum legacy cmd client parser. The framer must not be destroyed before the parser. */
enum e_rc_cli_result rlc_cli_parse_init(struct s_rlc_cli_parser **rcp, struct s_rc_frame *framer)
{
  struct s_rlc_cli_parser *parser;

  if (NULL == rcp || NULL == framer)
    return rcr_paramfail;

  parser = (struct s_rlc_cli_parser *) calloc(1, sizeof(*parser));
  if (NULL == parser)
    return rcr_allocfail;

  parser->framing = framer;

  *rcp = parser;
  return rcr_success;
}


/** Destroy a rinstrum legacy cmd client parser */
enum e_rc_cli_result rlc_cli_parse_destroy(struct s_rlc_cli_parser *rcp)
{
  if (NULL != rcp)
  {
    free(rcp);
    rcp = NULL;
  }
  return rcr_success;
}


/** Assemble a command ready to be sent to the host device from a rlc_command structure */
enum e_rc_cli_result rlc_cli_assemble_command(const struct s_rlc_cli_parser *rcp, const struct s_rlc_command *rlc_cmd,
                                              char *buffer, size_t maxlen)
{
  size_t endpos;
  enum e_rc_cli_result result;

  if (NULL == rcp || NULL == rlc_cmd || NULL == buffer)
    return rcr_paramfail;

  /* Put pre-framing into buffer */
  result = rc_preframe_message(rcp->framing, true, buffer, maxlen, &endpos);
  if (rcr_success != result)
    return result;

  /* Construct a command to send */
  result = rlc_command_request_to_string(&buffer[endpos], maxlen - endpos, rlc_cmd);
  if (rcr_success != result)
    return result;

  /* Append post framing to message */
  return rc_postframe_message(rcp->framing, true, buffer, maxlen);
}


/** Assemble a command ready to be sent to the host device (with variable arguments) */
enum e_rc_cli_result rlc_cli_assemble_command_va(const struct s_rlc_cli_parser *rcp,
                                                 const char cmd[RLC_COMMAND_LENGTH], bool cmd_not_query,
                                                 const char *param_format, va_list args,
                                                 char *buffer, size_t maxlen)
{
  size_t endpos;
  enum e_rc_cli_result result;

  if (NULL == rcp || NULL == cmd || NULL == buffer)
    return rcr_paramfail;

  /* Put pre-framing into buffer */
  result = rc_preframe_message(rcp->framing, true, buffer, maxlen, &endpos);
  if (rcr_success != result)
    return result;

  /* Construct a command to send */
  result = rlc_format_request_va(buffer, maxlen,
                                 cmd, cmd_not_query,
                                 param_format, args);
  if (rcr_success != result)
    return result;

  /* Append post framing to message */
  return rc_postframe_message(rcp->framing, true, buffer, maxlen);
}


/** Parse a response from the host device into rlc_cmd */
enum e_rc_cli_result rlc_cli_parse_response(const struct s_rlc_cli_parser *rcp, const char *response, struct s_rlc_command *rlc_cmd)
{
  enum e_rc_cli_result result;
  size_t startpos;
  size_t endpos;

  /* Find the start and end position of the encapsulated message */
  result = rc_frame_response_check(rcp->framing, false, response, &startpos, &endpos, NULL);
  if (rcr_success != result)
    return result;

  /* Parse the encapsulated message */
  return rlc_parser_buflen(&response[startpos], endpos - startpos, false, rlc_cmd);
}
