/*
 * rlc_cli_sync.c
 *
 *  Created on: 1 Jan. 2019
 *      Author: ian
 */
#define _XOPEN_SOURCE 700
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <rincmd/rc_cli_parse.h>
#include <rincmd/rc_cli_sync.h>
#include <rinlegacycmd/rlc_parameter.h>
#include <rinlegacycmd/rlc_cli_parse.h>

#include <rinlegacycmd/rlc_cli_sync.h>


/**
 * Create an rinstrum legacy command synchronous client
 *
 * @param rsc is set to the created sync client on success
 * @param cli_comms is the already existing comms interface. You are responsible for destroying this once rsc has been destroyed
 * @param use_crc is a flag to indicate the use of CRC framed messages, or normal framing
 */
enum e_rc_cli_result rlc_sync_client_init(struct s_rlc_sync_client **rsc,
                                          struct s_rc_cli_comms *cli_comms,
                                          bool use_crc)
{
  struct s_rlc_sync_client *irsc = NULL;
  struct s_rc_frame *frame = NULL;
  enum e_rc_cli_result result;

  if (NULL == rsc || NULL == cli_comms)
    return rcr_paramfail;

  /* Allocate the sync client structure */
  irsc = (struct s_rlc_sync_client *) calloc(1, sizeof(*irsc));
  if (NULL == irsc)
  {
    result = rcr_allocfail;
    goto rsci_fail;
  }

  /* Setup comms */
  irsc->cli_comms = cli_comms;

  /* Setup framing */
  result = (use_crc) ? rc_frame_init_crc(&frame) : rc_frame_init_normal(&frame);
  if (rcr_success != result)
    goto rsci_fail;

  /* Setup parser */
  result = rlc_cli_parse_init(&irsc->cli_parser, frame);
  if (rcr_success != result)
    goto rsci_fail;

  /* Create buffer */
  irsc->buffer = calloc(RC_CLI_MSG_MAXLEN, sizeof(char));
  if (NULL == irsc->buffer)
    goto rsci_fail;
  irsc->bufferlen = RC_CLI_MSG_MAXLEN;

  /* Success */
  *rsc = irsc;
  return rcr_success;

rsci_fail:
  rlc_sync_client_destroy(irsc);
  irsc = NULL;
  return result;
}


/** Destroy a rinstrum legacy command synchronous client */
enum e_rc_cli_result rlc_sync_client_destroy(struct s_rlc_sync_client *rsc)
{
  if (NULL != rsc)
  {
    if (NULL != rsc->cli_comms)
    {
      /* Up to the caller to destroy this */
      /* rc_cli_comms_destroy(rsc->cli_comms); */
      rsc->cli_comms = NULL;
    }
    if (NULL != rsc->cli_parser)
    {
      if (NULL != rsc->cli_parser->framing)
      {
        rc_frame_destroy(rsc->cli_parser->framing);
        rsc->cli_parser->framing = NULL;
      }
      rlc_cli_parse_destroy(rsc->cli_parser);
      rsc->cli_parser = NULL;
    }
    if (NULL != rsc->buffer)
    {
      free(rsc->buffer);
      rsc->buffer = NULL;
    }
    rsc = NULL;
  }
  return rcr_success;
}


/** Connect to the server */
enum e_rc_cli_result rlc_sync_client_connect(struct s_rlc_sync_client *rsc, const struct timespec *timeout)
{
  return rc_cli_comms_connect_sync(rsc->cli_comms, timeout);
}


/** Disconnect from the server */
enum e_rc_cli_result rlc_sync_client_disconnect(struct s_rlc_sync_client *rsc)
{
  return rc_cli_comms_disconnect(rsc->cli_comms);
}


/**
 * Get the file descriptor
 *
 * @param rsc is the rinstrum legacy command sync client
 * @param readfd is set to the read file descriptor if the comms is connected, < 0 otherwise. Not NULL.
 */
enum e_rc_cli_result rlc_sync_client_get_fd(struct s_rlc_sync_client *rsc, int *readfd)
{
  int connectfd;

  /* Check params */
  if (NULL == rsc || NULL == readfd)
  {
    return rcr_paramfail;
  }

  /* Get file descriptors */
  return rc_cli_comms_getfds(rsc->cli_comms, readfd, &connectfd);
}


/** Read a response to a command. Don't call directly */
enum e_rc_cli_result rlc_sync_client_read_response(struct s_rlc_sync_client *rsc, const struct timespec *timeout)
{
  size_t posn;
  size_t startpos;
  size_t endpos;
  size_t msgend;
  ssize_t bytes_read;
  enum e_rc_cli_result result;
  fd_set readfds;
  int readfd;
  int connectfd;
  int sres;

  for (posn = 0; posn < rsc->bufferlen; )
  {
    FD_ZERO(&readfds);
    result = rc_cli_comms_getfds(rsc->cli_comms, &readfd, &connectfd);
    if (rcr_success != result)
      return result;
    if (readfd >= 0)
      FD_SET(readfd, &readfds);
    sres = pselect(readfd + 1, &readfds, NULL, NULL, timeout, NULL);
    if (0 == sres)
    {
      /* Timeout, keep looping */
      return rcr_timeout;
    }
    else if (sres < 0)
    {
      return rcr_readfail;
    }
    else if (readfd >= 0 && FD_ISSET(readfd, &readfds))
    {
      bytes_read = rsc->bufferlen - posn;
      result = rc_cli_comms_read(rsc->cli_comms, &rsc->buffer[posn], &bytes_read);
      if (rcr_success != result)
        return result;
      posn += bytes_read;
      rsc->buffer[posn] = '\0';

      result = rc_frame_response_check(rsc->cli_parser->framing, false, rsc->buffer, &startpos, &endpos, &msgend);
      if (rcr_framingfail != result)
      {
        /* Valid message or error, log it first */
        rc_cli_comms_log(rsc->cli_comms, true, rsc->buffer, msgend);
        return result;
      }
    }
  }
  return rcr_bufferoverflow;
}


/**
 * Execute a rinstrum legacy comms command, returning the response in rlc_response
 *
 * @param rsc is the rinstrum legacy comms synchronous client
 * @param timeout is the timeout period or NULL for none
 * @param rlc_cmd is the rinstrum legacy command (ie IDN)
 * @param rlc_response is the rinstrum legacy response (ie 1234,"asdf"), or NULL if no response is expected
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_sync_client_command_request(struct s_rlc_sync_client *rsc, const struct timespec *timeout,
                                                     const struct s_rlc_command *rlc_cmd, struct s_rlc_command *rlc_response)
{
  enum e_rc_cli_result result;
  ssize_t writelen;

  result = rlc_cli_assemble_command(rsc->cli_parser, rlc_cmd, rsc->buffer, rsc->bufferlen);
  if (rcr_success != result)
    return result;

  writelen = strlen(rsc->buffer);
  result = rc_cli_comms_write(rsc->cli_comms, rsc->buffer, &writelen);
  if (rcr_success != result)
    return result;

  if (NULL != rlc_response)
  {
    /* Response expected */
    result = rlc_sync_client_read_response(rsc, timeout);
    if (rcr_success != result)
      return result;
    result = rlc_cli_parse_response(rsc->cli_parser, rsc->buffer, rlc_response);
    memcpy(rlc_response->command, rlc_cmd->command, RLC_COMMAND_LENGTH);
    rlc_response->command_not_query = rlc_cmd->command_not_query;
  }
  else
  {
    /* No response expected */
    result = rcr_success;
  }
  return result;
}


/**
 * Execute a rinstrum legacy comms command using variable arguments, returning the response in rlc_response
 *
 * @param rsc is the rinstrum legacy comms synchronous client
 * @param timeout is the timeout period or NULL for none
 * @param rlc_response is the rinstrum legacy response (ie 1234,"asdf"), or NULL if no response is expected
 * @param cmd is the rinstrum legacy command (ie IDN)
 * @param cmd_not_query true if the message is a command, false if it is a query
 * @param param_format is a string containing format characters to match the variable argument list
 *         "n" is used when a parameter is not required
 *         "i" is a signed integer parameter (passed as int64_t)
 *         "u" is an unsigned integer parameter (passed as uint64_t)
 *         "s" is null terminated string parameter (passed as const char *)
 *         "b" is a buffer parameter (passed as const void *buf and size_t len)
 * @param args is a variable argument list corresponding to the param_format
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_sync_client_command_request_va(struct s_rlc_sync_client *rsc, const struct timespec *timeout,
                                                        struct s_rlc_command *rlc_response,
                                                        const char cmd[RLC_COMMAND_LENGTH], bool cmd_not_query,
                                                        const char *param_format, ...)
{
  enum e_rc_cli_result result;
  ssize_t writelen;
  va_list args;

  va_start(args, param_format);
  result = rlc_cli_assemble_command_va(rsc->cli_parser, cmd, cmd_not_query, param_format, args, rsc->buffer, rsc->bufferlen);
  va_end(args);
  if (rcr_success != result)
    return result;

  writelen = strlen(rsc->buffer);
  result = rc_cli_comms_write(rsc->cli_comms, rsc->buffer, &writelen);
  if (rcr_success != result)
    return result;

  if (NULL != rlc_response)
  {
    /* Response expected */
    result = rlc_sync_client_read_response(rsc, timeout);
    if (rcr_success != result)
      return result;

    result = rlc_cli_parse_response(rsc->cli_parser, rsc->buffer, rlc_response);
    memcpy(rlc_response->command, cmd, RLC_COMMAND_LENGTH);
    rlc_response->command_not_query = cmd_not_query;
  }
  else
  {
    /* No response expected */
    result = rcr_success;
  }
  return result;
}
