/*
 * rlc_cli_sync.h
 *
 *  Created on: 1 Jan. 2019
 *      Author: ian
 */

#ifndef RLC_CLI_SYNC_H_
#define RLC_CLI_SYNC_H_

#include <time.h>
#include <stdbool.h>
#include <stdarg.h>
#include <stddef.h>
#include <rincmd/rc_cli_error.h>
#include <rincmd/rc_cli_sync.h>
#include <rinlegacycmd/rlc_parameter.h>


/** Structure for a running a synchronous rinstrum legacy client */
struct s_rlc_sync_client
{
  struct s_rc_cli_comms *cli_comms;     /*!< Communication with the host */
  struct s_rlc_cli_parser *cli_parser;  /*!< Parser for pulling apart and creating messages */
  char *buffer;                         /*!< Buffer for data */
  size_t bufferlen;                     /*!< Length of the data buffer */
};


/**
 * Create an rinstrum legacy command synchronous client
 *
 * @param rsc is set to the created sync client on success
 * @param cli_comms is the already existing comms interface. You are responsible for destroying this once rsc has been destroyed
 * @param use_crc is a flag to indicate the use of CRC framed messages, or normal framing
 */
enum e_rc_cli_result rlc_sync_client_init(struct s_rlc_sync_client **rsc,
                                          struct s_rc_cli_comms *cli_comms,
                                          bool use_crc);


/** Destroy a rinstrum legacy command synchronous client */
enum e_rc_cli_result rlc_sync_client_destroy(struct s_rlc_sync_client *rsc);


/** Connect to the server */
enum e_rc_cli_result rlc_sync_client_connect(struct s_rlc_sync_client *rsc, const struct timespec *timeout);


/** Disconnect from the server */
enum e_rc_cli_result rlc_sync_client_disconnect(struct s_rlc_sync_client *rsc);


/**
 * Get the file descriptor
 *
 * @param rsc is the rinstrum legacy command sync client
 * @param readfd is set to the read file descriptor if the comms is connected, < 0 otherwise. Not NULL.
 */
enum e_rc_cli_result rlc_sync_client_get_fd(struct s_rlc_sync_client *rsc, int *readfd);


/** Read a response to a command. Don't call directly */
enum e_rc_cli_result rlc_sync_client_read_response(struct s_rlc_sync_client *rsc, const struct timespec *timeout);


/**
 * Execute a rinstrum legacy comms command, returning the response in rlc_response
 *
 * @param rsc is the rinstrum legacy comms synchronous client
 * @param timeout is the timeout period or NULL for none
 * @param rlc_cmd is the rinstrum legacy command (ie IDN)
 * @param rlc_response is the rinstrum legacy response (ie 1234,"asdf"), or NULL if no response is expected
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_sync_client_command_request(struct s_rlc_sync_client *rsc, const struct timespec *timeout,
                                                     const struct s_rlc_command *rlc_cmd, struct s_rlc_command *rlc_response);


/**
 * Execute a rinstrum legacy comms command using variable arguments, returning the response in rlc_response
 *
 * @param rsc is the rinstrum legacy comms synchronous client
 * @param timeout is the timeout period or NULL for none
 * @param rlc_response is the rinstrum legacy response (ie 1234,"asdf"), or NULL if no response is expected
 * @param cmd is the rinstrum legacy command (ie IDN)
 * @param cmd_not_query true if the message is a command, false if it is a query
 * @param param_format is a string containing format characters to match the variable argument list
 *         "n" is used when a parameter is not required
 *         "i" is a signed integer parameter (passed as int64_t)
 *         "u" is an unsigned integer parameter (passed as uint64_t)
 *         "s" is null terminated string parameter (passed as const char *)
 *         "b" is a buffer parameter (passed as const void *buf and size_t len)
 * @param ... is the arguments corresponding to the param_format
 * @return e_rc_cli_result result code
 */
enum e_rc_cli_result rlc_sync_client_command_request_va(struct s_rlc_sync_client *rsc, const struct timespec *timeout,
                                                        struct s_rlc_command *rlc_response,
                                                        const char cmd[RLC_COMMAND_LENGTH], bool cmd_not_query,
                                                        const char *param_format, ...);


#endif /* RLC_CLI_SYNC_H_ */
