/*
 * rlc_parameter.c
 *
 *  Created on: 28 Dec. 2018
 *      Author: ian
 */
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <rincmd/rc_cli_error.h>
#include <rincmd/rc_escape.h>
#include <rinlegacycmd/rlc_parameter_escape.h>

#include <rinlegacycmd/rlc_parameter.h>


/** Initialise an existing rinstrum legacy comms parameter with a pre-allocated buffer */
static enum e_rc_cli_result rlc_parameter_init_prealloc(struct s_rlc_parameter *parameter, size_t buffer_maxlen, bool can_reallocate)
{
  parameter->param_type = rpt_unused;
  parameter->can_reallocate = can_reallocate;
  parameter->buffer_length = 0;
  parameter->buffer_maxlength = buffer_maxlen;
  parameter->int_value = 0;
  parameter->buffer = calloc(buffer_maxlen, sizeof(char));
  return (NULL == parameter->buffer) ? rcr_allocfail : rcr_success;
}


/** Destroy the allocated members within parameter, but not parameter itself */
static enum e_rc_cli_result rlc_parameter_free_members(struct s_rlc_parameter *parameter)
{
  if (NULL != parameter->buffer)
  {
    free(parameter->buffer);
    parameter->buffer = NULL;
  }
  parameter->buffer_length = 0;
  parameter->buffer_maxlength = 0;
  return rcr_success;
}


/** Set the parameter as unused */
static enum e_rc_cli_result rlc_parameter_unused_set(struct s_rlc_parameter *parameter)
{
  parameter->param_type = rpt_unused;
  return rcr_success;
}


/** Set the parameter as error */
static enum e_rc_cli_result rlc_parameter_error_set(struct s_rlc_parameter *parameter)
{
  parameter->param_type = rpt_error;
  return rcr_success;
}


/** Set the parameter as an integer */
static enum e_rc_cli_result rlc_parameter_integer_set(struct s_rlc_parameter *parameter, int64_t value)
{
  parameter->param_type = rpt_integer;
  parameter->int_value = value;
  return rcr_success;
}


/** Copy buffer into the parameter */
static enum e_rc_cli_result rlc_parameter_buffer_set(struct s_rlc_parameter *parameter, const char *buffer, size_t buffer_length)
{
  parameter->param_type = rpt_buffer;
  if (buffer_length > parameter->buffer_maxlength || NULL == parameter->buffer)
  {
    if (parameter->can_reallocate)
    {
      parameter->buffer = realloc(parameter->buffer, buffer_length);
      if (NULL == parameter->buffer)
      {
        parameter->buffer_maxlength = 0;
        parameter->buffer_length = 0;
        return rcr_allocfail;
      }
      parameter->buffer_maxlength = buffer_length;
    }
    else
    {
      return rcr_bufferoverflow;
    }
  }
  memcpy(parameter->buffer, buffer, buffer_length);
  parameter->buffer_length = buffer_length;
  return rcr_success;
}


/** Un-escape buffer into the parameter */
static enum e_rc_cli_result rlc_parameter_buffer_unescape_set(struct s_rlc_parameter *parameter, const char *esc_buffer, size_t esc_buffer_length)
{
  size_t buffer_len;
  enum e_rc_escape_result rer;
  enum e_rc_cli_result result;

  parameter->param_type = rpt_buffer;
  buffer_len = parameter->buffer_maxlength;
  rer = rlc_parameter_unescape_buffer(parameter->buffer, &buffer_len, esc_buffer, esc_buffer_length);
  if (rer_buffer_size == rer)
  {
    /* Didn't fit, reallocate if we can, and try again (buffer_len tells us how long it needs to be) */
    if (parameter->can_reallocate)
    {
      parameter->buffer = realloc(parameter->buffer, buffer_len);
      if (NULL == parameter->buffer)
      {
        parameter->buffer_maxlength = 0;
        parameter->buffer_length = 0;
        return rcr_allocfail;
      }
      parameter->buffer_maxlength = buffer_len;

      rer = rlc_parameter_unescape_buffer(parameter->buffer, &buffer_len, esc_buffer, esc_buffer_length);
    }
    else
    {
      return rcr_bufferoverflow;
    }
  }

  /* Handle result of first or second attempt at rlc_parameter_unescape_buffer() */
  switch (rer)
  {
    case rer_success:
      parameter->buffer_length = buffer_len;
      result = rcr_success;
      break;
    case rer_buffer_size:
      parameter->buffer_length = 0;
      result = rcr_bufferoverflow;
      break;
    default:
      parameter->buffer_length = 0;
      result = rcr_parsefail;
      break;
  }
  return result;
}


/** Copy null terminated string into the parameter (as a buffer parameter) */
static enum e_rc_cli_result rlc_parameter_string_set(struct s_rlc_parameter *parameter, const char *string)
{
  return rlc_parameter_buffer_set(parameter, string, strlen(string));
}


/** Create a rinstrum legacy comms command, pre-allocating parameters and parameter strings */
enum e_rc_cli_result rlc_command_create_prealloc(struct s_rlc_command **rlc_cmd,
                                                 const char command[RLC_COMMAND_LENGTH], bool command_not_query,
                                                 size_t max_parameters, size_t param_buf_length, bool can_reallocate)
{
  struct s_rlc_command *irlc_cmd = NULL;
  enum e_rc_cli_result result;
  size_t param;

  irlc_cmd = calloc(1, sizeof(struct s_rlc_command));
  if (NULL == irlc_cmd)
  {
    result = rcr_allocfail;
    goto rccp_fail;
  }

  memcpy(irlc_cmd->command, command, RLC_COMMAND_LENGTH);
  irlc_cmd->command_not_query = command_not_query;
  irlc_cmd->parameters = calloc(max_parameters, sizeof(struct s_rlc_parameter));
  if (NULL == irlc_cmd->parameters)
  {
    result = rcr_allocfail;
    goto rccp_fail;
  }
  irlc_cmd->num_parameters = 0;
  irlc_cmd->max_parameters = max_parameters;
  for (param = 0; param < max_parameters; param++)
  {
    result = rlc_parameter_init_prealloc(&irlc_cmd->parameters[param], param_buf_length, can_reallocate);
    if (rcr_success != result)
    {
      goto rccp_fail;
    }
  }

  *rlc_cmd = irlc_cmd;
  return rcr_success;

rccp_fail:
  rlc_command_destroy(irlc_cmd);
  return result;
}


/** Destroy a rinstrum legacy comms command, freeing parameters and parameter strings */
enum e_rc_cli_result rlc_command_destroy(struct s_rlc_command *rlc_cmd)
{
  size_t param;

  if (NULL != rlc_cmd)
  {
    if (NULL != rlc_cmd->parameters)
    {
      for (param = 0; param < rlc_cmd->max_parameters; param++)
      {
        rlc_parameter_free_members(&rlc_cmd->parameters[param]);
      }
      free(rlc_cmd->parameters);
      rlc_cmd->parameters = NULL;
    }
    rlc_cmd->num_parameters = 0;
    rlc_cmd->max_parameters = 0;
    free(rlc_cmd);
    rlc_cmd = NULL;
  }
  return rcr_success;
}


/** Set the parameter as unused */
enum e_rc_cli_result rlc_command_parameter_unused_set(struct s_rlc_command *rlc_cmd, size_t param)
{
  if (NULL == rlc_cmd || param >= rlc_cmd->max_parameters)
    return rcr_paramfail;

  return rlc_parameter_unused_set(&rlc_cmd->parameters[param]);
}


/** Set the parameter as error */
enum e_rc_cli_result rlc_command_parameter_error_set(struct s_rlc_command *rlc_cmd, size_t param)
{
  if (NULL == rlc_cmd || param >= rlc_cmd->max_parameters)
    return rcr_paramfail;

  return rlc_parameter_error_set(&rlc_cmd->parameters[param]);
}


/** Set the parameter as an integer */
enum e_rc_cli_result rlc_command_parameter_integer_set(struct s_rlc_command *rlc_cmd, size_t param, int64_t value)
{
  if (NULL == rlc_cmd || param >= rlc_cmd->max_parameters)
    return rcr_paramfail;

  return rlc_parameter_integer_set(&rlc_cmd->parameters[param], value);
}


/** Copy buffer into the parameter */
enum e_rc_cli_result rlc_command_parameter_buffer_set(struct s_rlc_command *rlc_cmd, size_t param, const char *buffer, size_t buffer_length)
{
  if (NULL == rlc_cmd || param >= rlc_cmd->max_parameters)
    return rcr_paramfail;

  return rlc_parameter_buffer_set(&rlc_cmd->parameters[param], buffer, buffer_length);
}


/** Un-escape buffer into the parameter */
enum e_rc_cli_result rlc_command_parameter_buffer_unescape_set(struct s_rlc_command *rlc_cmd, size_t param, const char *esc_buffer, size_t esc_buffer_length)
{
  if (NULL == rlc_cmd || param >= rlc_cmd->max_parameters)
    return rcr_paramfail;

  return rlc_parameter_buffer_unescape_set(&rlc_cmd->parameters[param], esc_buffer, esc_buffer_length);
}


/** Copy null terminated string into the parameter (as a buffer parameter) */
enum e_rc_cli_result rlc_command_parameter_string_set(struct s_rlc_command *rlc_cmd, size_t param, const char *string)
{
  if (NULL == rlc_cmd || param >= rlc_cmd->max_parameters)
    return rcr_paramfail;
  return rlc_parameter_string_set(&rlc_cmd->parameters[param], string);
}


/** Get the parameter type for param */
enum e_rc_cli_result rlc_command_parameter_type_get(const struct s_rlc_command *rlc_cmd, size_t param, enum e_rlc_parameter_type *param_type)
{
  if (NULL == rlc_cmd || NULL == param_type || param >= rlc_cmd->num_parameters)
    return rcr_paramfail;

  *param_type = rlc_cmd->parameters[param].param_type;
  return rcr_success;
}


/** Get the integer value for param */
enum e_rc_cli_result rlc_command_parameter_integer_get(const struct s_rlc_command *rlc_cmd, size_t param, int64_t *value)
{
  if (NULL == rlc_cmd || NULL == value || param >= rlc_cmd->num_parameters)
    return rcr_paramfail;

  if (rpt_integer != rlc_cmd->parameters[param].param_type)
    return rcr_paramfail;

  *value = rlc_cmd->parameters[param].int_value;
  return rcr_success;
}


/** Get the integer value for param after checking limits */
enum e_rc_cli_result rlc_command_parameter_integer_check_get(const struct s_rlc_command *rlc_cmd, size_t param,
                                                             int64_t *value, int64_t minimum, int64_t maximum)
{
  enum e_rc_cli_result result;

  result = rlc_command_parameter_integer_get(rlc_cmd, param, value);
  if (rcr_success != result)
    return result;

  return (*value > maximum || *value < minimum) ? rcr_paramfail : rcr_success;
}


/**
 * Get the buffer value and length for param
 *
 * @rlc_cmd is the legacy comms command to get the parameter buffer from
 * @param param is the parameter to retrieve the buffer for
 * @param buffer is set to point to the parameter buffer
 * @param buffer_length is set to the length of the parameter buffer
 */
enum e_rc_cli_result rlc_command_parameter_buffer_get(const struct s_rlc_command *rlc_cmd, size_t param, const char **buffer, size_t *buffer_length)
{
  if (NULL == rlc_cmd || NULL == buffer || NULL == buffer_length || param >= rlc_cmd->num_parameters)
    return rcr_paramfail;

  if (rpt_buffer != rlc_cmd->parameters[param].param_type)
    return rcr_paramfail;

  *buffer = rlc_cmd->parameters[param].buffer;
  *buffer_length = rlc_cmd->parameters[param].buffer_length;
  return rcr_success;
}


/**
 * Take a copy of the parameter buffer into buffer
 *
 * @rlc_cmd is the legacy comms command to get the parameter buffer from
 * @param param is the parameter to retrieve the buffer for
 * @param buffer is a pre-allocated buffer to copy the parameter buffer into
 * @param buffer_length should be set to the maximum length of buffer before calling. On return it is set to the length of data in buffer
 */
enum e_rc_cli_result rlc_command_parameter_buffer_copy(const struct s_rlc_command *rlc_cmd, size_t param, char *buffer, size_t *buffer_length)
{
  if (NULL == rlc_cmd || NULL == buffer || NULL == buffer_length || param >= rlc_cmd->num_parameters)
    return rcr_paramfail;

  if (rpt_buffer != rlc_cmd->parameters[param].param_type)
    return rcr_paramfail;

  if (*buffer_length < rlc_cmd->parameters[param].buffer_length)
    return rcr_bufferoverflow;

  memcpy(buffer, rlc_cmd->parameters[param].buffer, rlc_cmd->parameters[param].buffer_length);
  *buffer_length = rlc_cmd->parameters[param].buffer_length;
  return rcr_success;
}


/**
 * Take a copy of the parameter buffer into null terminated string after checking lengths.
 * Note if the parameter buffer contains '\0' characters, then string will be prematurely terminated
 *
 * @rlc_cmd is the legacy comms command to get the parameter buffer from
 * @param param is the parameter to retrieve the buffer for
 * @param string is a pre-allocated buffer to copy the parameter buffer into
 * @param string_minlength is the minimum length of the string
 * @param string_maxlength is the maximum length of string including the null termination character
 */
enum e_rc_cli_result rlc_command_parameter_string_check_copy(const struct s_rlc_command *rlc_cmd, size_t param, char *string, size_t string_minlength, size_t string_maxlength)
{
  enum e_rc_cli_result result;
  size_t string_length;

  string_length = string_maxlength;
  result = rlc_command_parameter_buffer_copy(rlc_cmd, param, string, &string_length);
  if (rcr_success != result)
    return result;

  if (string_length < string_minlength)
    return rcr_paramfail;

  if (string_length + 1 > string_maxlength)
    return rcr_bufferoverflow;
  string[string_length] = '\0';
  return rcr_success;
}


/**
 * Take a copy of the parameter buffer into null terminated string after checking lengths, and all characters are within the character set.
 * Note if the parameter buffer contains '\0' characters, then string will be prematurely terminated
 *
 * @rlc_cmd is the legacy comms command to get the parameter buffer from
 * @param param is the parameter to retrieve the buffer for
 * @param string is a pre-allocated buffer to copy the parameter buffer into
 * @param string_minlength is the minimum length of the string
 * @param string_maxlength is the maximum length of string including the null termination character
 * @param charset is a null terminated string containing valid characters for the string
 */
enum e_rc_cli_result rlc_command_parameter_string_check_charset_copy(const struct s_rlc_command *rlc_cmd, size_t param,
                                                                     char *string, size_t string_minlength, size_t string_maxlength, const char *charset)
{
  enum e_rc_cli_result result;
  size_t string_length;
  size_t i;

  string_length = string_maxlength;
  result = rlc_command_parameter_buffer_copy(rlc_cmd, param, string, &string_length);
  if (rcr_success != result)
    return result;

  if (string_length < string_minlength)
    return rcr_paramfail;

  if (string_length + 1 > string_maxlength)
    return rcr_bufferoverflow;
  string[string_length] = '\0';

  for (i = 0; i < string_length; i++)
  {
    if (strchr(charset, string[i]) == NULL)
    {
      return rcr_paramfail;
    }
  }
  return rcr_success;
}
