/*
 * rlc_parameter.h
 *
 *  Created on: 28 Dec. 2018
 *      Author: ian
 */

#ifndef RLC_PARAMETER_H_
#define RLC_PARAMETER_H_

#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <rincmd/rc_cli_error.h>


/** Rinstrum legacy commands are 3 characters in length eg IDN */
#define RLC_COMMAND_LENGTH 3


/** The different types of parameters supported by rinstrum legacy commands */
enum e_rlc_parameter_type
{
  rpt_unused,     /*!< This parameter is unused/not set */
  rpt_integer,    /*!< This parameter is an integer */
  rpt_buffer,     /*!< This parameter is a buffer */
  rpt_error       /*!< Special case for the first (and only) parameter of a command response can be ? to indicate error */
};


/** A rinstrum legacy command parameter */
struct s_rlc_parameter
{
  enum e_rlc_parameter_type param_type; /*!< The parameter type */
  bool can_reallocate;                  /*!< Can buffer be re-allocated to fit larger buffer */
  char *buffer;                         /*!< Buffer for buffer parameter */
  size_t buffer_length;                 /*!< Length of data in buffer */
  size_t buffer_maxlength;              /*!< Allocated size of buffer */
  int64_t int_value;                    /*!< Set if this is an integer parameter */
};


/** A rinstrum legacy command */
struct s_rlc_command
{
  char command[RLC_COMMAND_LENGTH];     /*!< The command (eg IDN) */
  bool command_not_query;               /*!< Is this a command or a query */
  struct s_rlc_parameter *parameters;   /*!< The list of parameters */
  size_t num_parameters;                /*!< The number of parameters used in this command */
  size_t max_parameters;                /*!< The number of parameters allocated to this command */
};


/** Create a rinstrum legacy comms command, pre-allocating parameters and parameter strings */
enum e_rc_cli_result rlc_command_create_prealloc(struct s_rlc_command **rlc_cmd,
                                                 const char command[RLC_COMMAND_LENGTH], bool command_not_query,
                                                 size_t num_parameters, size_t param_buf_length, bool can_reallocate);


/** Destroy a rinstrum legacy comms command, freeing parameters and parameter strings */
enum e_rc_cli_result rlc_command_destroy(struct s_rlc_command *rlc_cmd);


/** Set the parameter as unused */
enum e_rc_cli_result rlc_command_parameter_unused_set(struct s_rlc_command *rlc_cmd, size_t param);


/** Set the parameter as error */
enum e_rc_cli_result rlc_command_parameter_error_set(struct s_rlc_command *rlc_cmd, size_t param);


/** Set the parameter as an integer */
enum e_rc_cli_result rlc_command_parameter_integer_set(struct s_rlc_command *rlc_cmd, size_t param, int64_t value);


/** Copy buffer into the parameter */
enum e_rc_cli_result rlc_command_parameter_buffer_set(struct s_rlc_command *rlc_cmd, size_t param, const char *buffer, size_t buffer_length);


/** Un-escape buffer into the parameter */
enum e_rc_cli_result rlc_command_parameter_buffer_unescape_set(struct s_rlc_command *rlc_cmd, size_t param, const char *esc_buffer, size_t esc_buffer_length);


/** Copy null terminated string into the parameter (as a buffer parameter) */
enum e_rc_cli_result rlc_command_parameter_string_set(struct s_rlc_command *rlc_cmd, size_t param, const char *string);


/** Get the parameter type for param */
enum e_rc_cli_result rlc_command_parameter_type_get(const struct s_rlc_command *rlc_cmd, size_t param, enum e_rlc_parameter_type *param_type);


/** Get the integer value for param */
enum e_rc_cli_result rlc_command_parameter_integer_get(const struct s_rlc_command *rlc_cmd, size_t param, int64_t *value);


/** Get the integer value for param after checking limits */
enum e_rc_cli_result rlc_command_parameter_integer_check_get(const struct s_rlc_command *rlc_cmd, size_t param,
                                                             int64_t *value, int64_t minimum, int64_t maximum);


/**
 * Get the buffer value and length for param
 *
 * @rlc_cmd is the legacy comms command to get the parameter buffer from
 * @param param is the parameter to retrieve the buffer for
 * @param buffer is set to point to the parameter buffer
 * @param buffer_length is set to the length of the parameter buffer
 */
enum e_rc_cli_result rlc_command_parameter_buffer_get(const struct s_rlc_command *rlc_cmd, size_t param, const char **buffer, size_t *buffer_length);


/**
 * Take a copy of the parameter buffer into buffer
 *
 * @rlc_cmd is the legacy comms command to get the parameter buffer from
 * @param param is the parameter to retrieve the buffer for
 * @param buffer is a pre-allocated buffer to copy the parameter buffer into
 * @param buffer_length should be set to the maximum length of buffer before calling. On return it is set to the length of data in buffer
 */
enum e_rc_cli_result rlc_command_parameter_buffer_copy(const struct s_rlc_command *rlc_cmd, size_t param, char *buffer, size_t *buffer_length);


/**
 * Take a copy of the parameter buffer into null terminated string after checking lengths.
 * Note if the parameter buffer contains '\0' characters, then string will be prematurely terminated
 *
 * @rlc_cmd is the legacy comms command to get the parameter buffer from
 * @param param is the parameter to retrieve the buffer for
 * @param string is a pre-allocated buffer to copy the parameter buffer into
 * @param string_minlength is the minimum length of the string
 * @param string_maxlength is the maximum length of string including the null termination character
 */
enum e_rc_cli_result rlc_command_parameter_string_check_copy(const struct s_rlc_command *rlc_cmd, size_t param, char *string, size_t string_minlength, size_t string_maxlength);


/**
 * Take a copy of the parameter buffer into null terminated string after checking lengths, and all characters are within the character set.
 * Note if the parameter buffer contains '\0' characters, then string will be prematurely terminated
 *
 * @rlc_cmd is the legacy comms command to get the parameter buffer from
 * @param param is the parameter to retrieve the buffer for
 * @param string is a pre-allocated buffer to copy the parameter buffer into
 * @param string_minlength is the minimum length of the string
 * @param string_maxlength is the maximum length of string including the null termination character
 * @param charset is a null terminated string containing valid characters for the string
 */
enum e_rc_cli_result rlc_command_parameter_string_check_charset_copy(const struct s_rlc_command *rlc_cmd, size_t param,
                                                                     char *string, size_t string_minlength, size_t string_maxlength, const char *charset);


#endif /* RLC_PARAMETER_H_ */
