/*
 * rlc_parameter_escape.c
 *
 *  Created on: Jun 11, 2012
 *      Author: ian
 */
#include <ctype.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <rincmd/rc_escape.h>

#include <rinlegacycmd/rlc_parameter_escape.h>


#define COMMS_ESCAPE_MAX_CHAR 29


static const char comms_escape_chars[COMMS_ESCAPE_MAX_CHAR + 1] = ".ABCDEFGHIJKLMNOPQRSTUVWXYZ+-";


/**
 * Escape a character as per the strange legacy comms character escape rules:
 *  - 0x00-0x1F,",;,0x98..0xFF  -> \nnn
 *  - 0x80..0x97 -> As per comms_escape_chars (\A, etc)
 *  - \ -> \\
 *  - Other printables -> Not escaped
 *
 * @param escaped is a string, that ch is escaped into. Worst case is 5 characters including null terminator
 * @param ch is the character to be escaped
 * @return result of the escaping
 */
enum e_rc_escape_result rlc_parameter_escape_char(char *escaped, size_t escaped_maxlen, const char ch)
{
  enum e_rc_escape_result res;
  int snres;
  const uint8_t uch = ch;

  if (uch < ' ' || '"' == uch || ';' == uch || uch >= 128 + COMMS_ESCAPE_MAX_CHAR)
  {
    /*
     * Escape these as decimal representation \nnn (eg \001:
     *  - Control characters 0x00-0x1F
     *  - " needs to be escaped as this starts and ends a string parameter
     *  - ; needs to be escaped as this terminates a request
     *  - \r & \n (covered by control characters) need to be escaped as this terminates a response
     *  - 152..255 are escaped as the decimal representation, eg \152
     */
    snres = snprintf(escaped, escaped_maxlen, "\\%03u", (unsigned) uch);
  }
  else if ('\\' == uch)
  {
    /* \ is the escape character, so send two to signify a \ */
    snres = snprintf(escaped, escaped_maxlen, "\\\\");
  }
  else if (uch < 128)
  {
    /* Normal printable character */
    snres = snprintf(escaped, escaped_maxlen, "%c", uch);
  }
  else
  {
    /* 128..151 are escaped as \A, etc */
    snres = snprintf(escaped, escaped_maxlen, "\\%c", comms_escape_chars[uch - 128]);
  }
  res = (snres >= (int) escaped_maxlen) ? rer_buffer_size :
        (snres < 0) ? rer_data_conversion : rer_success;
  return res;
}


/**
 * Escape a buffer as per the strange legacy comms character escape rules. Can contain nulls ('\0')
 * See @ref rlc_parameter_escape_char for the rules.
 *
 * @param escaped is a string that buf is escaped into. Worst case is 5 characters per character in buf plus a null terminator
 * @param escaped_maxlen is the maximum length of escaped
 * @param buf is the buffer to be escaped
 * @param buf_len is the length of the buffer to be escaped
 * @return result of the escaping
 */
enum e_rc_escape_result rlc_parameter_escape_buffer(char *escaped, size_t escaped_maxlen, const char *buf, size_t buf_len)
{
  enum e_rc_escape_result res;
  size_t i;
  size_t esc_pos;

  for (i = 0, esc_pos = 0; i < buf_len; i++)
  {
    res = rlc_parameter_escape_char(&escaped[esc_pos], escaped_maxlen - esc_pos, buf[i]);
    if (rer_success != res)
      return res;
    esc_pos = strlen(escaped);
  }
  return rer_success;
}


/**
 * Escape a string as per the strange legacy comms character escape rules. Can't contain nulls, since they terminate the string
 * See @ref rlc_parameter_escape_char for the rules.
 *
 * @param escaped is a string, that str is escaped into. Worst case is 5 characters per character in str plus a null terminator
 * @param escaped_maxlen is the maximum length of escaped
 * @param str is the string to be escaped
 * @return result of the escaping
 */
enum e_rc_escape_result rlc_parameter_escape_string(char *escaped, size_t escaped_maxlen, const char *str)
{
  enum e_rc_escape_result res;
  size_t esc_pos;

  for (esc_pos = 0; '\0' != *str; str++)
  {
    res = rlc_parameter_escape_char(&escaped[esc_pos], escaped_maxlen - esc_pos, *str);
    if (rer_success != res)
      return res;
    esc_pos = strlen(escaped);
  }
  return rer_success;
}


/**
 * Un-escape a string as per the strange legacy comms character escape rules:
 * \nnn -> char value
 * \\ -> \
 * \A, etc -> 129 as per comms_escape_chars
 * \. -> 128
 * \+ -> 155
 * \- -> 156
 * Other printables -> Not escaped
 *
 * @param unescaped is the output un-escaped string (always <= length of escstr, unless unescaped_len gets in the way)
 * @param unescaped_len is a pointer to the maximum length of unescaped before the call.
 *        Set to the actual length of the un-escaped data.
 *        Can be used to determine the buffer size required if the result is rer_buffer_size
 * @param escaped is the escaped buffer to be decoded
 * @param escaped_len is the length of the escaped buffer
 * @return result of the unescaping
 */
enum e_rc_escape_result rlc_parameter_unescape_buffer(char *unescaped, size_t *unescaped_len, const char *escaped, size_t escaped_len)
{
  size_t escpos;
  size_t unescpos, j;
  size_t unescaped_maxlen;
  char ch;
  enum e_rc_escape_result result;

  if (NULL == unescaped || NULL == unescaped_len || NULL == escaped)
    return rer_buffer_size;

  unescaped_maxlen = *unescaped_len;
  for (escpos = 0, unescpos = 0; escpos < escaped_len; escpos++, unescpos++)
  {
    if (escaped[escpos] == '\\')
    {
      /* Escaped character */
      escpos++;
      switch (escaped[escpos])
      {
        case '\\':
          if (unescpos < unescaped_maxlen)
            unescaped[unescpos] = '\\';
          break;
        case '.':
          if (unescpos < unescaped_maxlen)
            unescaped[unescpos] = COMMS_ESCAPED_DOT_UNESCAPED;
          break;
        case '+':
          if (unescpos < unescaped_maxlen)
            unescaped[unescpos] = COMMS_ESCAPED_PLUS_UNESCAPED;
          break;
        case '-':
          if (unescpos < unescaped_maxlen)
            unescaped[unescpos] = COMMS_ESCAPED_MINUS_UNESCAPED;
          break;
        default:
          if (isupper(escaped[escpos]))
          {
            if (unescpos < unescaped_maxlen)
              unescaped[unescpos] = COMMS_ESCAPED_CHAR_TO_UNESCAPED(escaped[escpos]);
          }
          else
          {
            for (ch = 0, j = 0; j < 3; j++, escpos++)
            {
              if (!isdigit(escaped[escpos]))
              {
                result = rer_data_conversion;
                goto rpub_fail;
              }
              ch = ch * 10 + escaped[escpos] - '0';
            }
            if (ch == 0)
              ch = 128;
            if (unescpos < unescaped_maxlen)
              unescaped[unescpos] = ch;
            escpos--;
          }
          break;
      } /* switch */
    }
    else
    {
      /* Non-escaped character */
      if (unescpos < unescaped_maxlen)
        unescaped[unescpos] = escaped[escpos];
    }
  }
  result = (unescpos >= unescaped_maxlen) ? rer_buffer_size : rer_success;
  *unescaped_len = unescpos;

rpub_fail:
  return result;
}


/**
 * Un-escape a string as per the strange legacy comms character escape rules:
 * \nnn -> char value
 * \\ -> \
 * \A, etc -> 129 as per comms_escape_chars
 * \. -> 128
 * \+ -> 155
 * \- -> 156
 * Other printables -> Not escaped
 *
 * @param unescaped is the output un-escaped string (always <= length of escstr, unless max_len gets in the way)
 * @param unescaped_len is a pointer to the maximum length of unescaped before the call. Set to the actual length of the un-escaped data.
 * @param escaped is the escaped string to be decoded
 * @return result of the unescaping
 */
enum e_rc_escape_result rlc_parameter_unescape_string(char *unescaped, size_t *unescaped_len, const char *escaped)
{
  enum e_rc_escape_result result;

  result = rlc_parameter_unescape_buffer(unescaped, unescaped_len, escaped, strlen(escaped));
  return result;
}
