/*
 * rlc_parameter_escape.h
 *
 *  Created on: Jun 11, 2012
 *      Author: ian
 */

#ifndef RLC_PARAMETER_ESCAPE_H_
#define RLC_PARAMETER_ESCAPE_H_

#include <stdlib.h>
#include <rincmd/rc_escape.h>


/** An escaped dot (\.) is unescaped as 128 */
#define COMMS_ESCAPED_DOT_UNESCAPED (128)

/** An escaped plus (\+) is unescaped as 155 */
#define COMMS_ESCAPED_PLUS_UNESCAPED (155)

/** An escaped plus (\-) is unescaped as 156 */
#define COMMS_ESCAPED_MINUS_UNESCAPED (156)

/** An escaped upper case character (\A-\Z) is unescaped by adding 64 */
#define COMMS_ESCAPED_CHAR_TO_UNESCAPED(ch) (129 - 'A' + (ch))


/**
 * Escape a character as per the strange legacy comms character escape rules:
 *  - 0x00-0x1F,",;,0x98..0xFF  -> \nnn
 *  - 0x80..0x97 -> As per comms_escape_chars (\A, etc)
 *  - \ -> \\
 *  - Other printables -> Not escaped
 *
 * @param escaped is a string, that ch is escaped into. Worst case is 5 characters including null terminator
 * @param ch is the character to be escaped
 * @return result of the escaping
 */
enum e_rc_escape_result rlc_parameter_escape_char(char *escaped, size_t escaped_maxlen, const char ch);


/**
 * Escape a buffer as per the strange legacy comms character escape rules. Can contain nulls ('\0')
 * See @ref rlc_parameter_escape_char for the rules.
 *
 * @param escaped is a string that buf is escaped into. Worst case is 5 characters per character in buf plus a null terminator
 * @param escaped_maxlen is the maximum length of escaped
 * @param buf is the buffer to be escaped
 * @param buf_len is the length of the buffer to be escaped
 * @return result of the escaping
 */
enum e_rc_escape_result rlc_parameter_escape_buffer(char *escaped, size_t escaped_maxlen, const char *buf, size_t buf_len);


/**
 * Escape a string as per the strange legacy comms character escape rules. Can't contain nulls, since they terminate the string
 * See @ref rlc_parameter_escape_char for the rules.
 *
 * @param escaped is a string, that str is escaped into. Worst case is 5 characters per character in str plus a null terminator
 * @param escaped_maxlen is the maximum length of escaped
 * @param str is the string to be escaped
 * @return result of the escaping
 */
enum e_rc_escape_result rlc_parameter_escape_string(char *escaped, size_t escaped_maxlen, const char *str);


/**
 * Un-escape a string as per the strange legacy comms character escape rules:
 * \nnn -> char value
 * \\ -> \
 * \A, etc -> 129 as per comms_escape_chars
 * \. -> 128
 * \+ -> 155
 * \- -> 156
 * Other printables -> Not escaped
 *
 * @param unescaped is the output un-escaped string (always <= length of escstr, unless unescaped_len gets in the way)
 * @param unescaped_len is a pointer to the maximum length of unescaped before the call. Set to the actual length of the un-escaped data.
 * @param escaped is the escaped buffer to be decoded
 * @param escaped_len is the length of the escaped buffer
 * @return result of the unescaping
 */
enum e_rc_escape_result rlc_parameter_unescape_buffer(char *unescaped, size_t *unescaped_len, const char *escaped, size_t escaped_len);


/**
 * Un-escape a string as per the strange legacy comms character escape rules:
 * \nnn -> char value
 * \\ -> \
 * \A, etc -> 129 as per comms_escape_chars
 * \. -> 128
 * \+ -> 155
 * \- -> 156
 * Other printables -> Not escaped
 *
 * @param unescaped is the output un-escaped string (always <= length of escstr, unless max_len gets in the way)
 * @param unescaped_len is a pointer to the maximum length of unescaped before the call. Set to the actual length of the un-escaped data.
 * @param escaped is the escaped string to be decoded
 * @return result of the unescaping
 */
enum e_rc_escape_result rlc_parameter_unescape_string(char *unescaped, size_t *unescaped_len, const char *escaped);


#endif /* RLC_PARAMETER_ESCAPE_H_ */
