/*
 * rlc_parse.c
 *
 *  Created on: 2 Jan. 2019
 *      Author: ian
 */
#include <string.h>
#include <stddef.h>
#include <stdbool.h>
#include <ctype.h>
#include <rinlegacycmd/rlc_parameter.h>

#include <rinlegacycmd/rlc_parse.h>


/** Verify the command field (ie IDN). Alpha-numeric characters only */
static enum e_rc_cli_result rlc_verify_command_field(const char command[RLC_COMMAND_LENGTH])
{
  size_t i;

  for (i = 0; i < RLC_COMMAND_LENGTH; i++)
  {
    if (!isalpha(command[i]) && !isdigit(command[i]))
      return rcr_parsefail;
  }
  return rcr_success;
}


/** Parse a buffer parameter out of incoming data buffer into param */
static void rlc_skip_whitespace(const char *buffer, size_t *curpos, size_t buflen)
{
  for ( ; *curpos < buflen && isspace(buffer[*curpos]); (*curpos)++)
    ;
}


/** Parse a buffer parameter out of incoming data buffer into param */
static enum e_rc_cli_result rlc_parse_string_parameter(const char *buffer, size_t *curpos, size_t buflen,
                                                       struct s_rlc_command *rlc_cmd, size_t param)
{
  size_t icurpos = *curpos;
  const char *endpos;
  intptr_t len;
  enum e_rc_cli_result result;

  /* Must start with " */
  if (icurpos >= buflen || '"' != buffer[icurpos])
    return rcr_parsefail;
  icurpos++;

  /* Look for terminating ". Any " within the data must be escaped as \034 */
  endpos = memchr(&buffer[icurpos], '"', buflen - icurpos);
  if (NULL == endpos)
    return rcr_parsefail;

  len = endpos - &buffer[icurpos];
  result = rlc_command_parameter_buffer_unescape_set(rlc_cmd, param, &buffer[icurpos], len);
  if (rcr_success != result)
    return result;
  *curpos = icurpos + len + 1; /* +1 for terminating " */
  return rcr_success;
}


static enum e_rc_cli_result rlc_parse_integer_parameter(const char *buffer, size_t *curpos, size_t buflen,
                                                        struct s_rlc_command *rlc_cmd, size_t param)
{
  bool minus;
  size_t icurpos = *curpos;
  bool found_digit;
  int64_t value;
  enum e_rc_cli_result result;

  /* Leading sign check */
  if (icurpos < buflen && '-' == buffer[icurpos])
  {
    minus = true;
    icurpos++;
  }
  else
  {
    minus = false;
  }

  /* Skip any whitespace */
  rlc_skip_whitespace(buffer, &icurpos, buflen);

  /* Convert digits */
  for (found_digit = false, value = 0; icurpos < buflen && isdigit(buffer[icurpos]); icurpos++)
  {
    found_digit = true;
    value = value * 10 + buffer[icurpos] - '0';
  }
  if (!found_digit)
    return rcr_parsefail;

  if (minus)
    value = -value;

  /* Set the value in the parameter */
  result = rlc_command_parameter_integer_set(rlc_cmd, param, value);
  if (rcr_success != result)
    return result;

  /* Skip any trailing whitespace */
  rlc_skip_whitespace(buffer, &icurpos, buflen);

  *curpos = icurpos;
  return rcr_success;
}


/**
 * Parse rinstrum legacy comms message string from buf into rlc_cmd
 *
 * @param buf is the incoming message buffer
 * @param buflen is the length of the incoming message buffer
 * @param request_not_response true if this is a request we are parsing, false if this a response
 * @param rlc_cmd is the command/parameter structure to parse the data into
 *
 * @return result of the parsing
 */
enum e_rc_cli_result rlc_parser_buflen(const char *buf, size_t buflen,
                                       bool request_not_response,
                                       struct s_rlc_command *rlc_cmd)
{
  size_t param;
  size_t curpos;
  enum e_rc_cli_result result;

  curpos = 0;
  if (request_not_response)
  {
    /* Parsing a request, so look for the 3 letter command */
    if (curpos + RLC_COMMAND_LENGTH > buflen)
      return rcr_parsefail;
    result = rlc_verify_command_field(&buf[curpos]);
    if (rcr_success != result)
      return result;
    memcpy(rlc_cmd->command, &buf[curpos], RLC_COMMAND_LENGTH);
    curpos += RLC_COMMAND_LENGTH;

    /* And the optional query character */
    if (curpos < buflen && '?' == buf[curpos])
    {
      rlc_cmd->command_not_query = false;
      curpos++;
    }
    else
    {
      rlc_cmd->command_not_query = true;
    }
  }
  else
  {
    memset(rlc_cmd->command, '\0', RLC_COMMAND_LENGTH);
    rlc_cmd->command_not_query = false;
  }

  /* Now work through the parameters */
  for (param = 0; curpos < buflen && param < rlc_cmd->max_parameters; param++)
  {
    /* Skip any whitespace */
    rlc_skip_whitespace(buf, &curpos, buflen);

    /* Parse parameters */
    if (curpos < buflen)
    {
      if (',' == buf[curpos])
      {
        /* No value for this parameter */
        result = rlc_command_parameter_unused_set(rlc_cmd, param);
        curpos++;
      }
      else
      {
        if ('?' == buf[curpos])
        {
          /* Error parameter, only first param, and only for response */
          if (0 != param || request_not_response)
            return rcr_parsefail;
          curpos++;
          result = rlc_command_parameter_error_set(rlc_cmd, param);
        }
        else if ('"' == buf[curpos])
        {
          /* String parameter */
          result = rlc_parse_string_parameter(buf, &curpos, buflen, rlc_cmd, param);
        }
        else
        {
          /* Integer parameter */
          result = rlc_parse_integer_parameter(buf, &curpos, buflen, rlc_cmd, param);
        }
        if (rcr_success != result)
          return result;

        /* Skip over comma if it isn't the end of the buffer */
        if (curpos < buflen)
        {
          if (',' == buf[curpos])
            curpos++;
          else
            return rcr_parsefail;
        }
      }
    }
  }

  rlc_cmd->num_parameters = param;
  result = (curpos < buflen && param >= rlc_cmd->max_parameters) ? rcr_listfull : rcr_success;
  return result;
}
