/*
 * rlc_server.c
 *
 *  Created on: 14 Jan. 2019
 *      Author: ian
 */
#include <stdbool.h>
#include <stddef.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <rincmd/rc_cli_error.h>
#include <rinlegacycmd/rlc_assemble.h>
#include <rinlegacycmd/rlc_parameter.h>
#include <rinlegacycmd/rlc_parse.h>
#include <rinlegacycmd/rlc_server_register.h>
#include <rinlegacycmd/rlc_server_reply.h>
#include <rinlegacycmd/rlc_server_sxx.h>

#include <rinlegacycmd/rlc_server_handle.h>


static enum e_rc_cli_result rlc_server_dispatch_command(struct s_rlc_server *server)
{
  size_t i;
  bool match;
  unsigned cmdint;
  enum e_rc_cli_result result;

  server->response->num_parameters = 0;

  /* Convert the parsed command to an integer */
  result = rlc_command_to_int(server->command->command, &cmdint);

  /* Iterate the list of command handlers */
  for (i = 0, match = false; i < server->num_cmd_handlers; i++)
  {
    if (server->unit_selected ||
        (!server->unit_selected && server->cmd_handlers[i].run_unselected))
    {
      if (NULL != server->cmd_handlers[i].cmd_match_fn)
      {
        match = server->cmd_handlers[i].cmd_match_fn(server->cmd_handlers[i].userdata, server->command->command);
      }
      else
      {
        match = (cmdint == server->cmd_handlers[i].command_int);
      }
      if (match)
      {
        server->cmd_handlers[i].cmd_handler_fn(server->cmd_handlers[i].userdata,
                                               server->command, server->response);
        break;
      }
    }
  }

  if (!match)
  {
    /* No command has been handled */
    rlc_server_command_reply(server->response, server->reply_active, RLC_REPLY_ERROR);
  }
  return rcr_success;
}


/** Handle a rinstrum legacy comms server */
enum e_rc_cli_result rlc_server_handler(struct s_rlc_server *server,
                                        const char *input, size_t input_length,
                                        char *output, size_t output_length)
{
  enum e_rc_cli_result result;
  int res;

  /* Parse the command */
  result = rlc_parser_buflen(input, input_length, true, server->command);
  if (rcr_success != result)
    return result;

  /* Attempt to dispatch it */
  result = rlc_server_dispatch_command(server);
  if (rcr_success != result)
    return result;

  /* Send a response if required */
  if (server->response->num_parameters > 0)
  {
    /* Construct a command to send */
    result = rlc_command_response_to_string(output, output_length, server->response);
    if (rcr_success != result)
      return result;
  }
  else
  {
    res = snprintf(output, output_length, "%s", "");
    if (res < 0 || (size_t) res > output_length)
      return rcr_msgassemblefail;
  }
  return rcr_success;
}


/** Create a rinstrum legacy comms server */
enum e_rc_cli_result rlc_server_create(struct s_rlc_server **server, size_t max_cmd_handlers, size_t max_parameters, size_t param_buffer_length)
{
  struct s_rlc_server *iserver = NULL;
  enum e_rc_cli_result result;

  /* Allocate structure */
  iserver = calloc(1, sizeof(struct s_rlc_server));
  if (NULL == iserver)
  {
    result = rcr_allocfail;
    goto rsc_fail;
  }

  /* Allocate command handlers */
  iserver->max_cmd_handlers = max_cmd_handlers;
  iserver->num_cmd_handlers = 0;
  iserver->cmd_handlers = calloc(max_cmd_handlers, sizeof(struct s_rlc_server_cmd_handler));

  /* Create command and response */
  result = rlc_command_create_prealloc(&iserver->command, "", true, max_parameters, param_buffer_length, false);
  if (rcr_success != result)
    goto rsc_fail;
  result = rlc_command_create_prealloc(&iserver->response, "", true, max_parameters, param_buffer_length, false);
  if (rcr_success != result)
    goto rsc_fail;

  iserver->unit_selected = false;
  iserver->reply_active = false;

  /* Register STP and Snn commands */
  result = rlc_server_register_command(iserver, "STP", rlc_server_handle_STP, iserver, true);
  if (rcr_success != result)
    goto rsc_fail;
  result = rlc_server_register_command_matchfn(iserver, rlc_server_match_Snn, rlc_server_handle_Snn, iserver, true);
  if (rcr_success != result)
    goto rsc_fail;

  /* Success */
  *server = iserver;
  result = rcr_success;

rsc_fail:
  if (rcr_success != result)
    (void) rlc_server_destroy(iserver);
  return result;
}


/** Destroy a rinstrum legacy comms server */
enum e_rc_cli_result rlc_server_destroy(struct s_rlc_server *server)
{
  if (NULL != server)
  {
    if (NULL != server->cmd_handlers)
    {
      free(server->cmd_handlers);
      server->cmd_handlers = NULL;
    }
    if (NULL != server->command)
    {
      rlc_command_destroy(server->command);
      server->command = NULL;
    }
    if (NULL != server->command)
    {
      rlc_command_destroy(server->response);
      server->response = NULL;
    }
    free(server);
    server = NULL;
  }
  return rcr_success;
}
