/*
 * rlc_server_simple.c
 *
 * A simple rinstrum legacy comms server implementation listening on a TCP socket
 *
 *  Created on: 21 Jan. 2019
 *      Author: ian
 */
#define _XOPEN_SOURCE 700
#define _BSD_SOURCE
#include <stdbool.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <rinlegacycmd/rlc_server_handle.h>
#include <rinlegacycmd/rlc_server_reply.h>
#include <rinlegacycmd/rlc_server_register.h>

#include <rinlegacycmd/rlc_server_simple.h>


static enum e_rc_cli_result rlc_server_simple_establish_listening_tcp_socket(struct s_rlc_server_simple *server, const struct in_addr *iface_addr)
{
  struct sockaddr_in server_sockaddr;
  enum e_rc_cli_result rcresult;
  int result;
  int state;

  /* Create TCP socket */
  server->listen_sock_fd = socket(AF_INET, SOCK_STREAM, IPPROTO_IP);
  if (server->listen_sock_fd < 0)
  {
    rcresult = rcr_connectfail;
    goto rssels_fail;
  }

  /* Set the reuse address socket option in case an previous run still has it */
  state = 1;
  result = setsockopt(server->listen_sock_fd, SOL_SOCKET, SO_REUSEADDR, &state, sizeof(state));
  if (result < 0)
  {
    rcresult = rcr_connectfail;
    goto rssels_fail;
  }

  /* Setup server sockaddr */
  memset(&server_sockaddr, 0, sizeof(server_sockaddr));
  server_sockaddr.sin_family = AF_INET;                  /* Use TCP/IP */
  server_sockaddr.sin_port = htons(server->tcp_port);    /* Listen on specified port */
  server_sockaddr.sin_addr = *iface_addr;

  /* Bind to port */
  result = bind(server->listen_sock_fd, (struct sockaddr *) &server_sockaddr, sizeof(server_sockaddr));
  if (result < 0)
  {
    rcresult = rcr_connectfail;
    goto rssels_fail;
  }

  /* Put the socket configured above into a listen state. */
  result = listen(server->listen_sock_fd, 3);  /* 3 = number of pending connections allowed */
  if (result < 0)
  {
    rcresult = rcr_connectfail;
    goto rssels_fail;
  }

  /* Success */
  return rcr_success;

rssels_fail:
  if (server->listen_sock_fd >= 0)
  {
    close(server->listen_sock_fd);
    server->listen_sock_fd = -1;
  }
  return rcresult;
}


static enum e_rc_cli_result rlc_server_simple_establish_listening_uds_socket(struct s_rlc_server_simple *server)
{
  struct sockaddr_un server_sockaddr;
  enum e_rc_cli_result rcresult;
  int result;
  struct stat statbuf;

  /* Create unix domain socket */
  server->listen_sock_fd = socket(AF_LOCAL, SOCK_STREAM, 0);
  if (server->listen_sock_fd < 0)
  {
    rcresult = rcr_connectfail;
    goto rssels_fail;
  }

  /* First check that we are happy about using this UDS file, and delete if necessary */
  if (stat(server->interface, &statbuf) < 0)
  {
    if (ENOENT != errno)
    {
      rcresult = rcr_connectfail;
      goto rssels_fail;
    }
  }
  else
  {
    if (!S_ISSOCK(statbuf.st_mode))
    {
      rcresult = rcr_connectfail;
      goto rssels_fail;
    }
    unlink(server->interface);
  }

  /* Setup server sockaddr */
  memset(&server_sockaddr, 0, sizeof(server_sockaddr));
  server_sockaddr.sun_family = AF_LOCAL;                 /* Use unix domain socket */
  snprintf(server_sockaddr.sun_path, sizeof(server_sockaddr.sun_path), "%s", server->interface);

  /* Bind to port */
  result = bind(server->listen_sock_fd, (struct sockaddr *) &server_sockaddr, SUN_LEN(&server_sockaddr));
  if (result < 0)
  {
    rcresult = rcr_connectfail;
    goto rssels_fail;
  }

  /* Put the socket configured above into a listen state. */
  result = listen(server->listen_sock_fd, 3);  /* 3 = number of pending connections allowed */
  if (result < 0)
  {
    rcresult = rcr_connectfail;
    goto rssels_fail;
  }

  /* Success */
  return rcr_success;

rssels_fail:
  if (server->listen_sock_fd >= 0)
  {
    close(server->listen_sock_fd);
    server->listen_sock_fd = -1;
  }
  return rcresult;
}


static void rlc_server_simple_no_unused_clients(struct s_rlc_server_simple *server, int client_sock_fd)
{
  char buffer[100];

  snprintf(buffer, 100, "Maximum connections (%u) exceeded. Closing connection. Sorry.\r\n", (unsigned) server->max_clients);
  send(client_sock_fd, buffer, strlen(buffer), 0);
}


/** Accept a new connection */
static enum e_rc_cli_result rlc_server_simple_accept_connection(struct s_rlc_server_simple *server)
{
  int client_sock_fd = -1;
  size_t i;
  bool found;
  enum e_rc_cli_result result;

  /* Accept incoming connection request */
  client_sock_fd = accept(server->listen_sock_fd, NULL, 0);
  if (client_sock_fd < 0)
  {
    result = rcr_connectfail;
    goto rssac_fail;
  }

  /* Try to find an unused client to assign to this client connection */
  for (i = 0, found = false; i < server->max_clients; i++)
  {
    if (server->clients[i].client_sock_fd < 0)
    {
      server->clients[i].client_sock_fd = client_sock_fd;
      server->clients[i].inbuffer_posn = 0;
      server->clients[i].outbuffer[0] = '\0';
      found = true;
      break;
    }
  }
  if (!found)
  {
    /* Send no connections available message and close socket */
    rlc_server_simple_no_unused_clients(server, client_sock_fd);
    result = rcr_success;
    goto rssac_fail;
  }

  /* Success */
  return rcr_success;

rssac_fail:
  if (client_sock_fd >= 0)
  {
    close(client_sock_fd);
    client_sock_fd = -1;
  }
  return result;
}


/** Revert output buffer to before the message was sent */
#define adopt_framer_as_preferred() \
do \
{ \
  preferred_framer = client->framers[i]; \
  preferred_startpos = *startpos; \
  preferred_endpos = *endpos; \
  preferred_msgend = *msgend; \
} \
while (0)



/**
 * Try each of our framers, looking for the first valid message:
 *
 *              0   12345   67890
 * For message "\x01asdf\x04test;"
 *   ; framer:   start_pos = 0, end_pos = 9
 *   crc framer: start_pos = 0, end_pos = 5
 * We choose the crc framer because start_pos is the same for all, but crc has the lowest end_pos
 *
 *              012345 6
 * For message "test;\r\n"
 *   ; framer:  start_pos = 0, end_pos = 4
 *   \r framer: start_pos = 0, end_pos = 5
 *   \n framer: start_pos = 0, end_pos = 6
 * We choose the ; framer because start_pos is the same for all, but ; has the lowest end_pos
 */
static enum e_rc_cli_result rlc_server_simple_try_framers(struct s_rlc_server_simple_client *client,
                                                          const struct s_rc_frame **framer, const char *inbuf_ptr,
                                                          size_t *startpos, size_t *endpos, size_t *msgend)
{
  size_t i;
  struct s_rc_frame *preferred_framer;
  enum e_rc_cli_result result;
  size_t preferred_startpos = 0;
  size_t preferred_endpos = 0;
  size_t preferred_msgend = 0;
  bool first_match;

  /* Try all framers, then choose the one that is closest to the start of the buffer */
  for (i = 0, first_match = true, preferred_framer = NULL; i < client->num_framers; i++)
  {
    /* Look for correctly framed message */
    result = rc_frame_response_check(client->framers[i], true, inbuf_ptr, startpos, endpos, msgend);
    if (rcr_success == result)
    {
      if (first_match)
      {
        /* First match, adopt this as the preferred framer */
        adopt_framer_as_preferred();
        first_match = false;
      }
      else
      {
        /* Not first match, only adopt this framer as preferred if it starts or ends before the current preferred framer */
        if (*startpos < preferred_startpos)
        {
          adopt_framer_as_preferred();
        }
        else if (*startpos == preferred_startpos)
        {
          if (*endpos < preferred_endpos)
          {
            adopt_framer_as_preferred();
          }
        }
      }
    }
#if 0
    else
    {
      /* No match for this framer */
    }
#endif
  }

  /* Return the preferred framer, or error if we didn't find one */
  if (NULL == preferred_framer)
  {
    *framer = NULL;
    result = rcr_framingfail;
  }
  else
  {
    *framer = preferred_framer;
    *startpos = preferred_startpos;
    *endpos = preferred_endpos;
    *msgend = preferred_msgend;
    result = rcr_success;
  }
  return result;
}


/** Revert output buffer to before the message was sent */
#define revert_outbuf() \
do \
{ \
  outbuf_ptr = outbuf_ptr_save; \
  outbuf_used = outbuf_used_save; \
  *outbuf_ptr = '\0'; \
} \
while (0)


/** Handle incoming data on a connection */
static enum e_rc_cli_result rlc_server_simple_handle_incoming(struct s_rlc_server_simple_client *client)
{
  enum e_rc_cli_result result;
  size_t startpos;
  size_t endpos;
  size_t msgend;
  const char *inbuf_ptr;
  size_t inbuf_consumed;
  char *outbuf_ptr;
  size_t outbuf_used;
  char *outbuf_ptr_save;
  size_t outbuf_used_save;
  size_t outbuf_space;
  size_t preframe_end;
  size_t resplen;
  const struct s_rc_frame *framer = NULL;

  /*
   * If we run out of room in the output buffer, stop parsing, revert to before the pre-framing and try sending it.
   * revert_preframe.
   *
   * TODO If we don't have room in the output buffer, but don't advance the input
   * buffer, then the command handler will be called twice or more. This may or may not be a problem, depending on what the handler does.
   */

  /* Handle any messages we find in the incoming data buffer (there might be more than one) */
  outbuf_used = strlen(client->outbuffer);
  outbuf_ptr = &client->outbuffer[outbuf_used];
  for (msgend = 0, inbuf_consumed = 0, inbuf_ptr = client->inbuffer;
       '\0' != *inbuf_ptr && outbuf_used < client->outbuffer_maxsize;
       inbuf_ptr += msgend, inbuf_consumed += msgend)
  {
    /* Look for the best correctly framed message */
    result = rlc_server_simple_try_framers(client, &framer, inbuf_ptr, &startpos, &endpos, &msgend);
    if (rcr_success == result)
    {
      /* Save position to allow revert on failure */
      outbuf_ptr_save = outbuf_ptr;
      outbuf_used_save = outbuf_used;

      /* Setup output pre-framing, using the selected framer */
      outbuf_space = client->outbuffer_maxsize - outbuf_used;
      result = rc_preframe_message(framer, false, outbuf_ptr, outbuf_space, &preframe_end);
      if (rcr_success != result)
      {
        revert_outbuf();
        break;
      }

      /* Handle the incoming message */
      outbuf_ptr += preframe_end;
      outbuf_space -= preframe_end;
      result = rlc_server_handler(client->server, &inbuf_ptr[startpos], endpos - startpos,
                                  outbuf_ptr, outbuf_space);
      if (rcr_success != result)
      {
        revert_outbuf();
        if (rcr_msgassemblefail == result)
        {
          /* Out of room in the output buffer, so stop processing now to let it send */
          break;
        }
      }

      resplen = strlen(outbuf_ptr);
      if (resplen > 0)
      {
        /* There is some data to reply with, so append the post framing using the selected framer */
        /* TODO this will erroneously send a response if the preframing length > 0 (ie crc) */
        result = rc_postframe_message(framer, false, outbuf_ptr, outbuf_space);
        if (rcr_success != result)
        {
          revert_outbuf();
          break;
        }
      }
      else
      {
        /* No response data, revert pre-framing, but keep parsing */
        revert_outbuf();
      }

      /* Advance to end of current response */
      outbuf_used += strlen(outbuf_ptr);
      outbuf_ptr += outbuf_used;
    }
  }

  /* Move any left over data back to start of buffer */
  if (inbuf_consumed < client->inbuffer_posn)
  {
    if (msgend > 0)
    {
      memmove(client->inbuffer, &client->inbuffer[inbuf_consumed], client->inbuffer_posn - inbuf_consumed + 1);
      client->inbuffer_posn -= inbuf_consumed;
    }
  }
  else
  {
    client->inbuffer_posn = 0;
  }
  return rcr_success;
}


/** Handle incoming data on a connection */
static enum e_rc_cli_result rlc_server_simple_read_connection(struct s_rlc_server_simple_client *client)
{
  ssize_t recv_len;
  size_t to_read;

  /* Too full, scrap any data */
  if (client->inbuffer_posn >= client->inbuffer_maxsize - 1)
    client->inbuffer_posn = 0;

  to_read = client->inbuffer_maxsize - client->inbuffer_posn - 1;
  recv_len = recv(client->client_sock_fd, &client->inbuffer[client->inbuffer_posn], to_read, 0);
  if (recv_len > 0)
  {
    /* Read some data, handle it */
    client->inbuffer_posn += recv_len;
    client->inbuffer[client->inbuffer_posn] = '\0';
    (void) rlc_server_simple_handle_incoming(client);
  }
  else
  {
    /* Something failed, close the socket and remove it from the list */
    close(client->client_sock_fd);
    client->client_sock_fd = -1;
  }
  return rcr_success;
}


/** Write data to a connection */
static enum e_rc_cli_result rlc_server_simple_write_connection(struct s_rlc_server_simple_client *client)
{
  ssize_t written;
  size_t to_write;

  to_write = strlen(client->outbuffer);
  if (to_write > 0)
  {
    written = send(client->client_sock_fd, client->outbuffer, to_write, 0);
    if (written < 0)
    {
      close(client->client_sock_fd);
      client->client_sock_fd = -1;
    }
    else
    {
      /* Didn't write everything, so shuffle what's left back to the start of the buffer */
      if (written < (ssize_t) to_write)
        memmove(client->outbuffer, &client->outbuffer[written], to_write - written);
      else
        client->outbuffer[0] = '\0';
    }
  }
  return rcr_success;
}


/** Run the the rinstrum legacy client simple server */
enum e_rc_cli_result rlc_server_simple_handler(struct s_rlc_server_simple *server, const struct timespec *timeout)
{
  fd_set readfds;
  fd_set writefds;
  enum e_rc_cli_result rcresult;
  size_t i;
  int result;
  int maxfd;

  /* Setup readfds and writefds */
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  FD_SET(server->listen_sock_fd, &readfds);
  maxfd = server->listen_sock_fd;
  for (i = 0; i < server->max_clients; i++)
  {
    if (server->clients[i].client_sock_fd >= 0)
    {
      /* Read */
      FD_SET(server->clients[i].client_sock_fd, &readfds);
      maxfd = (server->clients[i].client_sock_fd > maxfd) ? server->clients[i].client_sock_fd : maxfd;

      /* If data to write */
      if (strlen(server->clients[i].outbuffer) > 0)
      {
        FD_SET(server->clients[i].client_sock_fd, &writefds);
      }
    }
  }

  result = pselect(maxfd + 1, &readfds, &writefds, NULL, timeout, NULL);
  if (result > 0)
  {
    /* Check for new connection */
    if (server->listen_sock_fd >= 0 && FD_ISSET(server->listen_sock_fd, &readfds))
    {
      rlc_server_simple_accept_connection(server);
    }
    for (i = 0; i < server->max_clients; i++)
    {
      if (server->clients[i].client_sock_fd >= 0 && FD_ISSET(server->clients[i].client_sock_fd, &readfds))
      {
        rlc_server_simple_read_connection(&server->clients[i]);
      }
      if (server->clients[i].client_sock_fd >= 0 && FD_ISSET(server->clients[i].client_sock_fd, &writefds))
      {
        rlc_server_simple_write_connection(&server->clients[i]);
      }
    }
    rcresult = rcr_success;
  }
  else if (0 == result)
  {
    /* Timeout, nothing to do */
    rcresult = rcr_timeout;
  }
  else
  {
    /* Error, shutdown */
    rcresult = rcr_failure;
  }
  return rcresult;
}


static enum e_rc_cli_result rlc_server_simple_setup_client(struct s_rlc_server_simple_client *client,
                                                           size_t client_data_buffer_size,
                                                           size_t max_cmd_handlers, size_t max_parameters, size_t param_buffer_length)
{
  enum e_rc_cli_result result;

  client->client_sock_fd = -1;

  /* Allocate incoming data buffer */
  client->inbuffer_posn = 0;
  client->inbuffer_maxsize = client_data_buffer_size;
  client->inbuffer = calloc(client->inbuffer_maxsize, sizeof(char));
  if (NULL == client->inbuffer)
  {
    result = rcr_allocfail;
    goto rsssc_fail;
  }

  /* Allocate outgoing data buffer */
  client->outbuffer_maxsize = client_data_buffer_size;
  client->outbuffer = calloc(client->outbuffer_maxsize, sizeof(char));
  if (NULL == client->outbuffer)
  {
    result = rcr_allocfail;
    goto rsssc_fail;
  }

  /* Create legacy command server */
  result = rlc_server_create(&client->server, max_cmd_handlers, max_parameters, param_buffer_length);
  if (rcr_success != result)
    goto rsssc_fail;

  /* Create the array of pointers to framers */
  client->num_framers = 3;
  client->framers = calloc(3, sizeof(struct s_rc_frame *));
  if (NULL == client->framers)
  {
    result = rcr_allocfail;
    goto rsssc_fail;
  }

  /* Create the framers */
  result = rc_frame_init_normal(&client->framers[0]);
  if (rcr_success != result)
    goto rsssc_fail;
  result = rc_frame_init_normal_cr(&client->framers[1]);
  if (rcr_success != result)
    goto rsssc_fail;
  result = rc_frame_init_normal_lf(&client->framers[2]);
  if (rcr_success != result)
    goto rsssc_fail;

  /* Success */
  return rcr_success;

rsssc_fail:
  return result;
}


/** Clean up a client structure, but don't free it */
static enum e_rc_cli_result rlc_server_simple_client_free_members(struct s_rlc_server_simple_client *client)
{
  size_t i;

  if (NULL != client)
  {
    if (client->client_sock_fd >= 0)
    {
      close(client->client_sock_fd);
      client->client_sock_fd = -1;
    }
    if (NULL != client->inbuffer)
    {
      free(client->inbuffer);
      client->inbuffer = NULL;
    }
    if (NULL != client->outbuffer)
    {
      free(client->outbuffer);
      client->outbuffer = NULL;
    }
    if (NULL != client->server)
    {
      rlc_server_destroy(client->server);
      client->server = NULL;
    }
    if (NULL != client->framers)
    {
      for (i = 0; i < client->num_framers; i++)
      {
        if (NULL != client->framers[i])
        {
          rc_frame_destroy(client->framers[i]);
          client->framers[i] = NULL;
        }
      }
      free(client->framers);
      client->framers = NULL;
    }
  }
  return rcr_success;
}


/**
 * Create and run the simple server
 *
 * @param server is set to point to an allocated server on success
 * @param interface is the interface to bind to: 0.0.0.0 for all interfaces, non-ip for unix domain socket
 * @param tcp_port is the port to listen on for incoming connections
 * @param max_clients is the maximum number of clients to handle simultaneously
 * @param client_data_buffer_size is the size of the incoming and outgoing data buffers. Must be able to accommodate the largest command/response
 * @param max_cmd_handlers is the maximum number of command handlers the server will support
 * @param max_parameters is the maximum number of parameters for a command/response
 * @param param_buffer_length is the maximum length of buffer parameters
 */
enum e_rc_cli_result rlc_server_simple_create(struct s_rlc_server_simple **server, const char *interface, int tcp_port, size_t max_clients,
                                              size_t client_data_buffer_size, size_t max_cmd_handlers, size_t max_parameters, size_t param_buffer_length)
{
  struct s_rlc_server_simple *iserver = NULL;
  size_t i;
  enum e_rc_cli_result rcresult;
  struct in_addr addr;

  /* Allocate licensing structure */
  iserver = calloc(1, sizeof(struct s_rlc_server_simple));
  if (NULL == iserver)
  {
    rcresult = rcr_allocfail;
    goto rssc_fail;
  }

  iserver->interface = strdup(interface);
  if (NULL == iserver->interface)
  {
    rcresult = rcr_allocfail;
    goto rssc_fail;
  }
  iserver->tcp_port = tcp_port;
  iserver->listen_sock_fd = -1;
  iserver->max_clients = max_clients;

  /* Allocate the clients, and initialise */
  iserver->clients = calloc(iserver->max_clients, sizeof(struct s_rlc_server_simple_client));
  if (NULL == iserver->clients)
  {
    rcresult = rcr_allocfail;
    goto rssc_fail;
  }
  for (i = 0; i < iserver->max_clients; i++)
  {
    rcresult = rlc_server_simple_setup_client(&iserver->clients[i], client_data_buffer_size, max_cmd_handlers, max_parameters, param_buffer_length);
    if (rcr_success != rcresult)
      goto rssc_fail;
  }

  /* Create listening socket */
  if (inet_pton(AF_INET, iserver->interface, &addr) == 1)
  {
    rcresult = rlc_server_simple_establish_listening_tcp_socket(iserver, &addr);
    if (rcr_success != rcresult)
      goto rssc_fail;
  }
  else
  {
    rcresult = rlc_server_simple_establish_listening_uds_socket(iserver);
    if (rcr_success != rcresult)
      goto rssc_fail;
  }

  /* Success */
  *server = iserver;
  return rcr_success;

rssc_fail:
  (void) rlc_server_simple_destroy(iserver);
  return rcresult;
}


/** Clean up the rinstrum legacy client simple server */
enum e_rc_cli_result rlc_server_simple_destroy(struct s_rlc_server_simple *server)
{
  size_t i;

  if (NULL != server)
  {
    if (server->listen_sock_fd >= 0)
    {
      close(server->listen_sock_fd);
      server->listen_sock_fd = -1;
    }
    if (NULL != server->interface)
    {
      free(server->interface);
      server->interface = NULL;
    }
    if (NULL != server->clients)
    {
      for (i = 0; i < server->max_clients; i++)
      {
        (void) rlc_server_simple_client_free_members(&server->clients[i]);
      }
      free(server->clients);
      server->clients = NULL;
    }
    free(server);
    server = NULL;
  }
  return rcr_success;
}


/** Register a command handler using a match function for a simple server */
enum e_rc_cli_result rlc_server_simple_register_command_matchfn(struct s_rlc_server_simple *server,
                                                                t_rlc_server_cmd_match_fn cmd_match_fn, t_rlc_server_cmd_handler_fn cmd_handler_fn, void *userdata,
                                                                bool run_unselected)
{
  size_t i;
  enum e_rc_cli_result result;

  for (i = 0; i < server->max_clients; i++)
  {
    result = rlc_server_register_command_matchfn(server->clients[i].server, cmd_match_fn, cmd_handler_fn, userdata, run_unselected);
    if (rcr_success != result)
      return result;
  }
  return rcr_success;
}


/** Register a command handler using a command for a simple server */
enum e_rc_cli_result rlc_server_simple_register_command(struct s_rlc_server_simple *server,
                                                        const char command[RLC_COMMAND_LENGTH], t_rlc_server_cmd_handler_fn cmd_handler_fn, void *userdata,
                                                        bool run_unselected)
{
  size_t i;
  enum e_rc_cli_result result;

  for (i = 0; i < server->max_clients; i++)
  {
    result = rlc_server_register_command(server->clients[i].server, command, cmd_handler_fn, userdata, run_unselected);
    if (rcr_success != result)
      return result;
  }
  return rcr_success;
}
