/*
 * rc_client.c
 *
 *  Created on: Nov 23, 2017
 *      Author: ian
 */
#include <stdio.h>
#include <stdbool.h>
#include <getopt.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <time.h>
#include <rincmd/rc_parser.h>
#include <rincmd/rc_cli_gettime.h>
#include <rincmd/rc_cli_sync.h>
#include <rincmd/rc_cli_async.h>
#include <rincmd/rc_cli_async_stream.h>
#include <rincmd/rc_cli_sync_stream.h>
#include "rlc_client_sync.h"


static bool application_quit = false;


enum e_protocol { prot_tcp, prot_udp, prot_uds, prot_serial };


static bool cmdline_display_help = false;
static const char *cmdline_remote_host = NULL;
static int cmdline_remote_port = 2222;
static unsigned cmdline_localudp_port = 0;
static bool cmdline_usecrc = false;
static bool cmdline_useasync = false;
static bool cmdline_streaming = false;
static enum e_protocol cmdline_protocol = prot_tcp;
static int cmdline_timeout = 5000;
static const char *cmdline_logfile = NULL;
static unsigned cmdline_repeats = 1;
static const char *cmdline_query_command = "IDN";
/* static bool cmdline_bulk = false; */


static void cmdline_print_usage(const char *prog)
{
  printf("Usage: %s [options] <host>\n"
         "Rinstrum legacy comms client\n", prog);

  printf("Options:\n"
         "  -h --help              This help\n"
         "  -p --port <port>       The remote port (default 2222)\n"
         "  -o --protocol <prot>   The protocol to use to communicate\n"
         "                         (tcp, udp, uds, serial) (default tcp)\n"
/*         "  -c --usecrc            Require CRC checksum on messages\n" */
/*         "  -a --async             Use the asynchronous client rather than synchronous\n"
         "                         client\n" */
/*         "  -s --stream            Setup streaming instead of poll/response.\n" */
         "  -t --timeout           Timeout period in milliseconds (default 5000,\n"
         "                         -1 for indefinite timeout)\n"
         "  -l --local-udp-port    Set the local UDP port to bind to when using \"-o udp\",\n"
         "                         or 0 (default) to use the kernel assigned one\n"
         "  -f --logfile <file>    Write all communications out to file\n"
         "  -r --repeat <times>    Number of times to repeat the command (default 1,\n"
         "                         0 for indefinite)\n"
         "  -m --command <cmd>     The command to query (default IDN).\n"
/*         "  -b --bulk              Use the bulk register access (async only)\n" */
         "Parameters:\n"
         "  host                   For tcp and udp: the host to connect to\n"
         "                         For uds: the socket file to connect to\n"
         "                         For serial: the serial port to connect to\n"
         "\n"
        );
}


static int cmdline_parse_options(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "help",           0, 0, 'h' },
    { "port",           1, 0, 'p' },
    { "protocol",       1, 0, 'o' },
/*    { "usecrc",         0, 0, 'c' }, */
/*     { "async",          0, 0, 'a' }, */
/*     { "stream",         0, 0, 's' }, */
    { "timeout",        1, 0, 't' },
    { "local-udp-port", 1, 0, 'l' },
    { "logfile",        1, 0, 'f' },
    { "repeat",         1, 0, 'r' },
    { "command",        1, 0, 'm' },
/*    { "bulk",           0, 0, 'b' }, */
    { NULL,             0, 0, '\0' },
  };
  int c;

  /* Options */
  while (1)
  {
/*    c = getopt_long(argc, argv, "hp:o:cast:l:f:r:m:b", lopts, NULL); */
    c = getopt_long(argc, argv, "hp:o:t:l:f:r:m:", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      /* Help */
      case 'h':
        cmdline_display_help = true;
        return 0;
        break;

      /* Remote TCP port */
      case 'p':
        errno = 0;
        cmdline_remote_port = strtoul(optarg, NULL, 10);
        if (0 != errno)
        {
          fprintf(stderr, "Error: Invalid remote port, try --help\n");
          return -1;
        }
        break;

      /* TCP, UDP, UDS, serial connection */
      case 'o':
        if (strcasecmp(optarg, "tcp") == 0)
          cmdline_protocol = prot_tcp;
        else if (strcasecmp(optarg, "udp") == 0)
          cmdline_protocol = prot_udp;
        else if (strcasecmp(optarg, "uds") == 0)
          cmdline_protocol = prot_uds;
        else if (strcasecmp(optarg, "serial") == 0)
          cmdline_protocol = prot_serial;
        else
        {
          fprintf(stderr, "Error: Invalid protocol, try --help\n");
          return -1;
        }
        break;

#if 0
      /* Use CRC checksums on messages */
      case 'c':
        cmdline_usecrc = true;
        break;

      /* Use asynchronous client */
      case 'a':
        cmdline_useasync = true;
        break;

      /* Use streaming rather than poll/response */
      case 's':
        cmdline_streaming = true;
        break;
#endif

      /* Set a timeout value */
      case 't':
        errno = 0;
        cmdline_timeout = strtol(optarg, NULL, 10);
        if (0 != errno)
        {
          fprintf(stderr, "Error: Invalid timeout period, try --help\n");
          return -1;
        }
        break;

      /* Set the local udp port */
      case 'l':
        errno = 0;
        cmdline_localudp_port = strtoul(optarg, NULL, 10);
        if (0 != errno || cmdline_localudp_port > 65535)
        {
          fprintf(stderr, "Error: Invalid local udp port, try --help\n");
          return -1;
        }
        break;

      /* Log file */
      case 'f':
        cmdline_logfile = optarg;
        break;

      /* Set the number of times to repeat the command */
      case 'r':
        errno = 0;
        cmdline_repeats = strtoul(optarg, NULL, 10);
        if (0 != errno)
        {
          fprintf(stderr, "Error: Invalid number of repeats, try --help\n");
          return -1;
        }
        break;

      /* The command to query */
      case 'm':
        cmdline_query_command = optarg;
        break;

#if 0
      /* Bulk register access */
      case 'b':
        cmdline_bulk = true;
        break;
#endif

      default:
        fprintf(stderr, "Error: Invalid command line option, try --help\n");
        return -1;
        break;
    }
  }

  /* Remote host */
  if (optind >= argc)
  {
    fprintf(stderr, "Error: No host specified, try --help\n");
    return -1;
  }
  cmdline_remote_host = argv[optind];
  optind++;

  return 0;
}


/** Handler for SIGINT (ctrl-c), SIGTERM (killall), SIGHUP (???) */
static void shutdown_signal_handler(int signum)
{
  application_quit = true;
}


/** Install signal handlers for signals we want to catch */
static bool install_signal_handlers(void)
{
  /* Catch SIGINT (ctrl-c) */
  if (SIG_ERR == signal(SIGINT, shutdown_signal_handler))
  {
    fprintf(stderr, "Failed to install SIGINT handler\n");
    return false;
  }

  /* Catch SIGTERM (killall) */
  if (SIG_ERR == signal(SIGTERM, shutdown_signal_handler))
  {
    fprintf(stderr, "Failed to install SIGTERM handler\n");
    return false;
  }

  /* Catch SIGHUP (????) */
  if (SIG_ERR == signal(SIGHUP, shutdown_signal_handler))
  {
    fprintf(stderr, "Failed to install SIGHUP handler\n");
    return false;
  }

#if 0 /* SIGKILL can't be caught */
  /* Catch SIGKILL (kill -9) */
  if (SIG_ERR == signal(SIGKILL, shutdown_signal_handler))
  {
    fprintf(stderr, "Failed to install SIGKILL handler\n");
    return false;
  }
#endif
  return true;
}


enum e_rc_cli_result rc_cli_external_gettime(struct timespec *ts)
{
  if (NULL == ts)
    return rcr_paramfail;
  else
    return (clock_gettime(CLOCK_MONOTONIC, ts) == 0) ? rcr_success : rcr_failure;
}


static bool create_client_comms(FILE **logfile, struct s_rc_cli_comms **rcc)
{
  struct s_rc_cli_comms *ircc = NULL;
  FILE *ilogfile = NULL;

  /* Open log file */
  if (NULL != cmdline_logfile)
  {
    ilogfile = fopen(cmdline_logfile, "a");
    if (NULL == ilogfile)
    {
      fprintf(stderr, "Error: Could not open log file \"%s\" for appending\n", cmdline_logfile);
      goto ccc_fail;
    }
  }

  /* Create the appropriate connection */
  switch (cmdline_protocol)
  {
    case prot_tcp:
      if (rc_comms_init_tcp(&ircc, cmdline_remote_host, cmdline_remote_port, ilogfile) != rcr_success)
        goto ccc_fail;
      break;
    case prot_udp:
      if (rc_comms_init_udp(&ircc, cmdline_remote_host, cmdline_remote_port, cmdline_localudp_port, ilogfile) != rcr_success)
        goto ccc_fail;
      break;
    case prot_uds:
      if (rc_comms_init_uds(&ircc, cmdline_remote_host, ilogfile) != rcr_success)
        goto ccc_fail;
      break;
    case prot_serial:
      if (rc_comms_init_serial(&ircc, cmdline_remote_host, 9600, 'N', 8, 1, ilogfile) != rcr_success)
        goto ccc_fail;
      break;
    default:
      goto ccc_fail;
      break;
  }

  *logfile = ilogfile;
  *rcc = ircc;
  return true;

ccc_fail:
  if (NULL != ilogfile)
    fclose(ilogfile);
  if (NULL != ircc)
    (void) rc_cli_comms_destroy(ircc);
  return false;
}


int main(int argc, char *argv[])
{
  int mresult;
  FILE *logfile = NULL;
  struct s_rc_cli_comms *rcc = NULL;
  enum e_rc_cli_result result;

  /* Do command line */
  if (cmdline_parse_options(argc, argv) < 0)
  {
    mresult = 1;
    goto m_fail;
  }
  if (cmdline_display_help)
  {
    cmdline_print_usage(argv[0]);
    mresult = 0;
    goto m_fail;
  }

  /* Install signal handlers so we can shutdown cleanly */
  if (!install_signal_handlers())
  {
    mresult = 1;
    goto m_fail;
  }

  /* Create comms connection */
  if (!create_client_comms(&logfile, &rcc))
  {
    mresult = 1;
    goto m_fail;
  }

  /* Do async or sync client example */
  if (cmdline_useasync)
  {
    /* result = async_client_example(rcc, cmdline_register, cmdline_usecrc, cmdline_streaming, cmdline_bulk, cmdline_repeats, cmdline_timeout, &application_quit); */
    fprintf(stderr, "Error: Async client not implemented yet\n");
    result = rcr_paramfail;
  }
  else
  {
    result = rlc_sync_client_example(rcc, cmdline_query_command, cmdline_usecrc, cmdline_streaming, cmdline_repeats, cmdline_timeout, &application_quit);
  }
  mresult = (rcr_success != result) ? 1 : 0;

m_fail:
  if (NULL != rcc)
    (void) rc_cli_comms_destroy(rcc);
  if (NULL != logfile)
    fclose(logfile);
  return mresult;
}
