/*
 * rlc_server.c
 *
 *  Created on: 21 Jan. 2019
 *      Author: ian
 */
#include <stdio.h>
#include <getopt.h>
#include <signal.h>
#include <stdbool.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <rincmd/rc_cli_error.h>
#include <rinlegacycmd/rlc_parameter.h>
#include <rinlegacycmd/rlc_server_simple.h>
#include <rinlegacycmd/rlc_server_reply.h>


static bool application_quit = false;


/*enum e_protocol { prot_tcp, prot_udp, prot_uds, prot_serial }; */
enum e_protocol { prot_tcp, prot_uds };


static bool cmdline_display_help = false;
static int cmdline_listen_port = 2222;
static enum e_protocol cmdline_protocol = prot_tcp;
static const char *cmdline_bind_interface = NULL;
/* static int cmdline_timeout = 5000; */
/* static const char *cmdline_logfile = NULL; */


static void cmdline_print_usage(const char *prog)
{
  printf("Usage: %s [options]\n"
         "Rinstrum legacy comms simple server example\n", prog);

  printf("Options:\n"
         "  -h --help              This help\n"
         "  -p --port <port>       The listen TCP port (default 2222)\n"
         "  -o --protocol <prot>   The protocol to use to communicate\n"
         "                         (tcp, uds) (default tcp)\n"
         "  -i --interface <iface> The interface to bind to.\n"
         "                         For UDS this is the socket file.\n"
         "                         (tcp default: 0.0.0.0)\n"
#if 0
/*         "  -c --usecrc            Require CRC checksum on messages\n" */
/*         "  -a --async             Use the asynchronous client rather than synchronous\n"
         "                         client\n" */
/*         "  -s --stream            Setup streaming instead of poll/response.\n" */
         "  -t --timeout           Timeout period in milliseconds (default 5000,\n"
         "                         -1 for indefinite timeout)\n"
         "  -f --logfile <file>    Write all communications out to file\n"
         "Parameters:\n"
         "  host                   For tcp and udp: the host to connect to\n"
         "                         For uds: the socket file to connect to\n"
         "                         For serial: the serial port to connect to\n"
#endif
         "\n"
        );
}


static int cmdline_parse_options(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "help",           0, 0, 'h' },
    { "port",           1, 0, 'p' },
    { "protocol",       1, 0, 'o' },
    { "interface",      1, 0, 'i' },
#if 0
/*    { "usecrc",         0, 0, 'c' }, */
/*     { "async",          0, 0, 'a' }, */
/*     { "stream",         0, 0, 's' }, */
    { "timeout",        1, 0, 't' },
    { "local-udp-port", 1, 0, 'l' },
    { "logfile",        1, 0, 'f' },
#endif
    { NULL,             0, 0, '\0' },
  };
  int c;

  /* Options */
  while (1)
  {
/*    c = getopt_long(argc, argv, "hp:o:cast:l:f:r:m:b", lopts, NULL); */
    c = getopt_long(argc, argv, "hp:o:i:", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      /* Help */
      case 'h':
        cmdline_display_help = true;
        return 0;
        break;

      /* Listen TCP port */
      case 'p':
        errno = 0;
        cmdline_listen_port = strtoul(optarg, NULL, 10);
        if (0 != errno)
        {
          fprintf(stderr, "Error: Invalid remote port, try --help\n");
          return -1;
        }
        break;

      /* TCP, UDP, UDS, serial connection */
      case 'o':
        if (strcasecmp(optarg, "tcp") == 0)
          cmdline_protocol = prot_tcp;
        else if (strcasecmp(optarg, "uds") == 0)
          cmdline_protocol = prot_uds;
#if 0
        else if (strcasecmp(optarg, "udp") == 0)
          cmdline_protocol = prot_udp;
        else if (strcasecmp(optarg, "serial") == 0)
          cmdline_protocol = prot_serial;
#endif
        else
        {
          fprintf(stderr, "Error: Invalid protocol, try --help\n");
          return -1;
        }
        break;

      /* Bind interface */
      case 'i':
        cmdline_bind_interface = optarg;
        break;

#if 0
      /* Set a timeout value */
      case 't':
        errno = 0;
        cmdline_timeout = strtol(optarg, NULL, 10);
        if (0 != errno)
        {
          fprintf(stderr, "Error: Invalid timeout period, try --help\n");
          return -1;
        }
        break;

      /* Log file */
      case 'f':
        cmdline_logfile = optarg;
        break;
#endif

      default:
        fprintf(stderr, "Error: Invalid command line option, try --help\n");
        return -1;
        break;
    }
  }

  switch (cmdline_protocol)
  {
    case prot_tcp:
      if (NULL == cmdline_bind_interface)
        cmdline_bind_interface = "0.0.0.0";
      break;
    case prot_uds:
      if (NULL == cmdline_bind_interface)
      {
        fprintf(stderr, "Error: No socket file specified for unix domain socket. Specify using --interface or try --help\n");
        return -1;
      }
      break;
  }

  return 0;
}


/** Handler for SIGINT (ctrl-c), SIGTERM (killall), SIGHUP (???) */
static void shutdown_signal_handler(int signum)
{
  application_quit = true;
}


/** Install signal handlers for signals we want to catch */
static bool install_signal_handlers(void)
{
  /* Catch SIGINT (ctrl-c) */
  if (SIG_ERR == signal(SIGINT, shutdown_signal_handler))
  {
    fprintf(stderr, "Failed to install SIGINT handler\n");
    return false;
  }

  /* Catch SIGTERM (killall) */
  if (SIG_ERR == signal(SIGTERM, shutdown_signal_handler))
  {
    fprintf(stderr, "Failed to install SIGTERM handler\n");
    return false;
  }

  /* Catch SIGHUP (????) */
  if (SIG_ERR == signal(SIGHUP, shutdown_signal_handler))
  {
    fprintf(stderr, "Failed to install SIGHUP handler\n");
    return false;
  }

#if 0 /* SIGKILL can't be caught */
  /* Catch SIGKILL (kill -9) */
  if (SIG_ERR == signal(SIGKILL, shutdown_signal_handler))
  {
    fprintf(stderr, "Failed to install SIGKILL handler\n");
    return false;
  }
#endif
  return true;
}


#if 0
static bool create_client_comms(FILE **logfile, struct s_rc_cli_comms **rcc)
{
  struct s_rc_cli_comms *ircc = NULL;
  FILE *ilogfile = NULL;

  /* Open log file */
  if (NULL != cmdline_logfile)
  {
    ilogfile = fopen(cmdline_logfile, "a");
    if (NULL == ilogfile)
    {
      fprintf(stderr, "Error: Could not open log file \"%s\" for appending\n", cmdline_logfile);
      goto ccc_fail;
    }
  }

  /* Create the appropriate connection */
  switch (cmdline_protocol)
  {
    case prot_tcp:
      if (rc_comms_init_tcp(&ircc, cmdline_remote_host, cmdline_listen_port, ilogfile) != rcr_success)
        goto ccc_fail;
      break;
    case prot_udp:
      if (rc_comms_init_udp(&ircc, cmdline_remote_host, cmdline_listen_port, cmdline_localudp_port, ilogfile) != rcr_success)
        goto ccc_fail;
      break;
    case prot_uds:
      if (rc_comms_init_uds(&ircc, cmdline_remote_host, ilogfile) != rcr_success)
        goto ccc_fail;
      break;
    case prot_serial:
      if (rc_comms_init_serial(&ircc, cmdline_remote_host, 9600, 'N', 8, 1, ilogfile) != rcr_success)
        goto ccc_fail;
      break;
    default:
      goto ccc_fail;
      break;
  }

  *logfile = ilogfile;
  *rcc = ircc;
  return true;

ccc_fail:
  if (NULL != ilogfile)
    fclose(ilogfile);
  if (NULL != ircc)
    (void) rc_cli_comms_destroy(ircc);
  return false;
}
#endif


/** Handle IDN - identification */
static void simple_server_handler_IDN(void *userdata, const struct s_rlc_command *command, struct s_rlc_command *response)
{
  if (command->command_not_query)
  {
    rlc_server_command_reply(response, true, RLC_REPLY_ERROR);
  }
  else
  {
    /* Userid, serialno, version, model */
    /* C500","3477738","1.3.2","C520 */
    response->num_parameters = 4;
    rlc_command_parameter_string_set(response, 0, "C520");
    rlc_command_parameter_string_set(response, 1, "123456");
    rlc_command_parameter_string_set(response, 2, "1.2.3");
    rlc_command_parameter_string_set(response, 3, "C520");
  }
}


#define MAX_CLIENTS 10
#define CLIENT_DATA_BUFFER_SIZE 1024
#define MAX_CMD_HANDLERS 10
#define MAX_PARAMETERS 10
#define PARAM_BUFFER_LENGTH 100


static void run_simple_server(void)
{
  struct s_rlc_server_simple *server = NULL;
  enum e_rc_cli_result result;
  const struct timespec timeout = { .tv_sec = 1, .tv_nsec = 0 };
  char bind_to[100];

  if (prot_tcp == cmdline_protocol)
  {
    snprintf(bind_to, 100, " TCP %s:%d", cmdline_bind_interface, cmdline_listen_port);
  }
  else
  {
    snprintf(bind_to, 100, "UDS socket \"%s\"", cmdline_bind_interface);
  }

  /* Create the simple server */
  result = rlc_server_simple_create(&server, cmdline_bind_interface, cmdline_listen_port, MAX_CLIENTS, CLIENT_DATA_BUFFER_SIZE, MAX_CMD_HANDLERS, MAX_PARAMETERS, PARAM_BUFFER_LENGTH);
  if (rcr_success != result)
  {
    fprintf(stderr, "Error: Could not create simple server on %s\n", bind_to);
    goto rss_fail;
  }

  /* Register the IDN command */
  result = rlc_server_simple_register_command(server, "IDN", simple_server_handler_IDN, NULL, false);
  if (rcr_success != result)
  {
    fprintf(stderr, "Error: Could not register IDN command with simple server\n");
    goto rss_fail;
  }

  /* Loop while the server runs */
  printf("Rinstrum Legacy Command simple server is running on %s.\n"
         "Accepting IDN, Snn commands\n", bind_to);
  while (!application_quit)
  {
    result = rlc_server_simple_handler(server, &timeout);
    if (rcr_success != result && rcr_timeout != result)
    {
      fprintf(stderr, "Error: Simple server reported error\n");
      goto rss_fail;
    }
  }

rss_fail:
  (void) rlc_server_simple_destroy(server);
}


int main(int argc, char *argv[])
{
  int mresult;
  /* FILE *logfile = NULL; */

  /* Do command line */
  if (cmdline_parse_options(argc, argv) < 0)
  {
    mresult = 1;
    goto m_fail;
  }
  if (cmdline_display_help)
  {
    cmdline_print_usage(argv[0]);
    mresult = 0;
    goto m_fail;
  }

  /* Install signal handlers so we can shutdown cleanly */
  if (!install_signal_handlers())
  {
    mresult = 1;
    goto m_fail;
  }

  run_simple_server();
  mresult = 0;

m_fail:
/*
  if (NULL != logfile)
    fclose(logfile); */
  return mresult;
}
