/*
 * test_rlc_parameter_escape.c
 *
 *  Created on: 28 Dec. 2018
 *      Author: ian
 */
#include <string.h>
#include <cutest/CuTest.h>
#include <rinlegacycmd/rlc_parameter_escape.h>

#include "test_rlc_parameter_escape.h"


/** Test escaping single character */
static void test_rlc_parameter_escape_char(CuTest *tc)
{
  #define ESCAPED_LEN 10
  char escaped[ESCAPED_LEN];
  enum e_rc_escape_result res;

  /* 0x00-0x1F,",;,0x98..0xFF  -> \nnn */
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0x00);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\000", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0x1F);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\031", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, '"');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\034", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, ';');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\059", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0x9D);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\157", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0xFF);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\255", escaped);

  /* 0x80..0x97 -> As per comms_escape_chars (\A, etc) */
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0x80);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\.", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0x81);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\A", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0x9A);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\Z", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0x9B);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\+", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, 0x9C);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\-", escaped);

  /* \ -> \\ */
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, '\\');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\\\", escaped);

  /* Other printables -> Not escaped */
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, ' ');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, " ", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, '~');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "~", escaped);
}


/** Test escaping the special command and parameter terminating characters */
static void test_rlc_parameter_escape_char_terminating_characters(CuTest *tc)
{
  #define ESCAPED_LEN 10
  char escaped[ESCAPED_LEN];
  enum e_rc_escape_result res;

  /* " starts and ends a string parameter (eg CMD"asdf";), and must be escaped as \nnn */
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, '"');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\034", escaped);

  /* ; terminates a request (eg IDN?;), and must be escaped as \nnn */
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, ';');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\059", escaped);

  /* \r\n terminates a response (eg 0\r\n), and must be escaped as \nnn */
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, '\r');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\013", escaped);
  res = rlc_parameter_escape_char(escaped, ESCAPED_LEN, '\n');
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, "\\010", escaped);
}


/** Test escaping buffer */
static void test_rlc_parameter_escape_buffer(CuTest *tc)
{
#define TEST_ESCAPE_BUFFER_LEN 100
char escaped[TEST_ESCAPE_BUFFER_LEN];
enum e_rc_escape_result res;
const char test_buffer[] =     "test string\\"   "\x00"  "\""    "\x01"  "\x80" "\x81";
const char expected_string[] = "test string\\\\" "\\000" "\\034" "\\001" "\\."  "\\A";

/* Test a valid buffer */
res = rlc_parameter_escape_buffer(escaped, TEST_ESCAPE_BUFFER_LEN, test_buffer, sizeof(test_buffer) - 1);
CuAssertIntEquals(tc, rer_success, res);
CuAssertStrEquals(tc, expected_string, escaped);

/* Test not enough room in buffer */
res = rlc_parameter_escape_buffer(escaped, 5, test_buffer, sizeof(test_buffer));
CuAssertIntEquals(tc, rer_buffer_size, res);
}


/** Test escaping string */
static void test_rlc_parameter_escape_string(CuTest *tc)
{
  #define TEST_ESCAPE_STRING_LEN 100
  char escaped[TEST_ESCAPE_STRING_LEN];
  enum e_rc_escape_result res;
  const char test_string[] =     "test string\\"   "\""    "\x01"  "\x80" "\x81";
  const char expected_string[] = "test string\\\\" "\\034" "\\001" "\\."  "\\A";

  /* Test a valid string */
  res = rlc_parameter_escape_string(escaped, TEST_ESCAPE_STRING_LEN, test_string);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertStrEquals(tc, expected_string, escaped);

  /* Test not enough room in buffer */
  res = rlc_parameter_escape_string(escaped, 5, test_string);
  CuAssertIntEquals(tc, rer_buffer_size, res);
}


/** Test unescaping buffer */
static void test_rlc_parameter_unescape_buffer(CuTest *tc)
{
  #define TEST_UNESCAPE_STRING_LEN 100
  char unescaped[TEST_UNESCAPE_STRING_LEN];
  enum e_rc_escape_result res;
  const char escaped_string[] =  "test string\\\\" "\\034" "\\001" "\\."  "\\A"  "\\+"  "\\-";
  const char expected_string[] = "test string\\"   "\""    "\x01"  "\x80" "\x81" "\x96" "\x97";
  size_t unescaped_len;
  const char invalid1[] =  "blah\\12blah";
  const char invalid2[] =  "blah\\ablah";

  /* Test a valid buffer */
  unescaped_len = TEST_UNESCAPE_STRING_LEN;
  res = rlc_parameter_unescape_buffer(unescaped, &unescaped_len, escaped_string, sizeof(escaped_string) - 1);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertIntEquals(tc, strlen(expected_string), unescaped_len);
  CuAssertTrue(tc, memcmp(expected_string, unescaped, strlen(expected_string)) == 0);

  /* Test not enough room in the buffer */
  unescaped_len = 5;
  res = rlc_parameter_unescape_buffer(unescaped, &unescaped_len, escaped_string, sizeof(escaped_string) - 1);
  CuAssertIntEquals(tc, rer_buffer_size, res);

  /* Test a invalid escaping */
  unescaped_len = TEST_UNESCAPE_STRING_LEN;
  res = rlc_parameter_unescape_buffer(unescaped, &unescaped_len, invalid1, sizeof(invalid1) - 1);
  CuAssertIntEquals(tc, rer_data_conversion, res);
  unescaped_len = TEST_UNESCAPE_STRING_LEN;
  res = rlc_parameter_unescape_buffer(unescaped, &unescaped_len, invalid2, sizeof(invalid2) - 1);
  CuAssertIntEquals(tc, rer_data_conversion, res);
}


/** Test unescaping string */
static void test_rlc_parameter_unescape_string(CuTest *tc)
{
  #define TEST_UNESCAPE_STRING_LEN 100
  char unescaped[TEST_UNESCAPE_STRING_LEN];
  enum e_rc_escape_result res;
  const char escaped_string[] =  "test string\\\\" "\\034" "\\001" "\\."  "\\A"  "\\+"  "\\-";
  const char expected_string[] = "test string\\"   "\""    "\x01"  "\x80" "\x81" "\x96" "\x97";
  size_t unescaped_len;

  /* Test a valid string */
  unescaped_len = TEST_UNESCAPE_STRING_LEN;
  res = rlc_parameter_unescape_string(unescaped, &unescaped_len, escaped_string);
  CuAssertIntEquals(tc, rer_success, res);
  CuAssertIntEquals(tc, strlen(expected_string), unescaped_len);
  //CuAssertTrue(tc, memcmp(expected_string, unescaped, strlen(expected_string)) == 0);
}


CuSuite *CuGetSuite_rlc_parameter_escape(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rlc_parameter_escape_char);
  SUITE_ADD_TEST(suite, test_rlc_parameter_escape_char_terminating_characters);
  SUITE_ADD_TEST(suite, test_rlc_parameter_escape_buffer);
  SUITE_ADD_TEST(suite, test_rlc_parameter_escape_string);
  SUITE_ADD_TEST(suite, test_rlc_parameter_unescape_buffer);
  SUITE_ADD_TEST(suite, test_rlc_parameter_unescape_string);

  return suite;
}
