/*
 * test_rlc_parse.c
 *
 *  Created on: 4 Jan. 2019
 *      Author: ian
 */
#include <string.h>
#include <cutest/CuTest.h>
#include <rincmd/rc_cli_error.h>
#include <rinlegacycmd/rlc_parse.h>

#include "test_rlc_parse.h"


/** Test parsing rinstrum legacy comms message string from buf into rlc_cmd */
static void test_rlc_parser_buflen(CuTest *tc)
{
  #define TEST_NUM_PARAMS 3
  enum e_rc_cli_result result;
  struct s_rlc_command *rlc_cmd = NULL;
  const char *request;
  const char *response;

  /* Create command */
  result = rlc_command_create_prealloc(&rlc_cmd, "", false, TEST_NUM_PARAMS, 20, true);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertPtrNotNull(tc, rlc_cmd);

  /* Try command only, no parameters */
  request = "IDN";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("IDN", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 0, rlc_cmd->num_parameters);

  /* Try command with 1 unset parameter */
  request = "IDN,";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("IDN", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 1, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_unused, rlc_cmd->parameters[0].param_type);

  /* Try query only, no parameters */
  request = "MSV?";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("MSV", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, !rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 0, rlc_cmd->num_parameters);

  /* Try query with 1 unset parameter */
  request = "MSV?,";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("MSV", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, !rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 1, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_unused, rlc_cmd->parameters[0].param_type);

  /* Try command with 1 buffer parameter */
  request = "IDN\"asdf\"";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("IDN", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 1, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_buffer, rlc_cmd->parameters[0].param_type);
  CuAssertIntEquals(tc, 4, rlc_cmd->parameters[0].buffer_length);
  CuAssertIntEquals(tc, 0, memcmp("asdf", rlc_cmd->parameters[0].buffer, 4));

  /* Try command with 1 integer parameter */
  request = "IDN12345";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("IDN", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 1, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_integer, rlc_cmd->parameters[0].param_type);
  CuAssertIntEquals(tc, 12345, rlc_cmd->parameters[0].int_value);

  /* Try command with 1 negative integer parameter */
  request = "IDN-12345";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("IDN", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 1, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_integer, rlc_cmd->parameters[0].param_type);
  CuAssertIntEquals(tc, -12345, rlc_cmd->parameters[0].int_value);

  /* Try command with 1 negative integer parameter with spaces */
  request = "IDN - 12345 ";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("IDN", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 1, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_integer, rlc_cmd->parameters[0].param_type);
  CuAssertIntEquals(tc, -12345, rlc_cmd->parameters[0].int_value);

  /* Try command with multiple parameters */
  request = "IDN-12345,,\"asdf\"";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertTrue(tc, memcmp("IDN", rlc_cmd->command, RLC_COMMAND_LENGTH) == 0);
  CuAssertTrue(tc, rlc_cmd->command_not_query);
  CuAssertIntEquals(tc, 3, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_integer, rlc_cmd->parameters[0].param_type);
  CuAssertIntEquals(tc, -12345, rlc_cmd->parameters[0].int_value);
  CuAssertIntEquals(tc, rpt_unused, rlc_cmd->parameters[1].param_type);
  CuAssertIntEquals(tc, rpt_buffer, rlc_cmd->parameters[2].param_type);
  CuAssertIntEquals(tc, 4, rlc_cmd->parameters[2].buffer_length);
  CuAssertIntEquals(tc, 0, memcmp("asdf", rlc_cmd->parameters[2].buffer, 4));

  /* Try response with multiple parameters */
  response = "-12345,,\"asdf\"";
  result = rlc_parser_buflen(response, strlen(response), false, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 3, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_integer, rlc_cmd->parameters[0].param_type);
  CuAssertIntEquals(tc, -12345, rlc_cmd->parameters[0].int_value);
  CuAssertIntEquals(tc, rpt_unused, rlc_cmd->parameters[1].param_type);
  CuAssertIntEquals(tc, rpt_buffer, rlc_cmd->parameters[2].param_type);
  CuAssertIntEquals(tc, 4, rlc_cmd->parameters[2].buffer_length);
  CuAssertIntEquals(tc, 0, memcmp("asdf", rlc_cmd->parameters[2].buffer, 4));

  /* Try response with error parameter */
  response = "?";
  result = rlc_parser_buflen(response, strlen(response), false, rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertIntEquals(tc, 1, rlc_cmd->num_parameters);
  CuAssertIntEquals(tc, rpt_error, rlc_cmd->parameters[0].param_type);

  /* Clean up */
  result = rlc_command_destroy(rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
}


/** Test parsing rinstrum legacy comms message string from malformed buf into rlc_cmd */
static void test_rlc_parser_buflen_malformed(CuTest *tc)
{
  #define TEST_NUM_PARAMS 3
  enum e_rc_cli_result result;
  struct s_rlc_command *rlc_cmd = NULL;
  const char *request;

  /* Create command */
  result = rlc_command_create_prealloc(&rlc_cmd, "", false, TEST_NUM_PARAMS, 20, true);
  CuAssertIntEquals(tc, rcr_success, result);
  CuAssertPtrNotNull(tc, rlc_cmd);

  /* Try short command */
  request = "I";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_parsefail, result);
  request = "ID";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_parsefail, result);

  /* Try bad command characters */
  request = "I N";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_parsefail, result);
  request = "I-N";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_parsefail, result);

  /* Try command with unterminated buffer parameter */
  request = "IDN\"asdf";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_parsefail, result);

  /* Try command with badly terminated buffer parameter */
  request = "IDN\"as\"df";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_parsefail, result);

  /* Try command with bad integer parameter */
  request = "IDN12A45";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_parsefail, result);
  request = "IDNL";
  result = rlc_parser_buflen(request, strlen(request), true, rlc_cmd);
  CuAssertIntEquals(tc, rcr_parsefail, result);

  /* Clean up */
  result = rlc_command_destroy(rlc_cmd);
  CuAssertIntEquals(tc, rcr_success, result);
}


CuSuite *CuGetSuite_rlc_parse(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rlc_parser_buflen);
  SUITE_ADD_TEST(suite, test_rlc_parser_buflen_malformed);

  return suite;
}
