'''
Created on 31/10/2016

@author: IanB
'''
import unittest
import hashlib
import argparse
import sys


# Each license code is 6 characters in length
LICENSECODE_LEN = 6


# Each license code digit has 26 combinations
LICENSECODE_DIGIT_COMBINATIONS = 26


def rl_hash_to_passcode(hash_hexdigest):
    """ License code = hash mod combinations"""
    combinations = LICENSECODE_DIGIT_COMBINATIONS ** LICENSECODE_LEN
    hash_num = long(hash_hexdigest, 16)
    license_code = hash_num % combinations
    return license_code


# Generated with "shuf -i 0-7 -n 8" x RL_NUM_BYTE_MUNGES */
rl_byte_munges = \
[ \
  [ 0, 6, 3, 2, 7, 4, 5, 1 ], \
  [ 1, 4, 5, 2, 0, 6, 3, 7 ], \
  [ 7, 6, 0, 4, 1, 5, 2, 3 ], \
  [ 1, 3, 6, 5, 4, 7, 0, 2 ], \
  [ 4, 3, 1, 7, 0, 6, 5, 2 ], \
  [ 5, 3, 0, 6, 1, 7, 4, 2 ], \
  [ 6, 7, 5, 2, 4, 3, 1, 0 ], \
  [ 7, 2, 4, 3, 0, 5, 6, 1 ], \
  [ 3, 2, 5, 7, 6, 4, 0, 1 ], \
  [ 3, 6, 0, 4, 7, 2, 5, 1 ], \
  [ 3, 1, 0, 5, 2, 7, 4, 6 ], \
  [ 4, 1, 3, 0, 7, 2, 6, 5 ], \
  [ 4, 2, 7, 6, 1, 3, 0, 5 ], \
  [ 5, 0, 3, 6, 4, 7, 1, 2 ], \
  [ 2, 1, 4, 6, 5, 7, 0, 3 ], \
  [ 6, 1, 4, 2, 3, 5, 7, 0 ] \
]


def rl_byte_munge(algorithm, data):
    result = 0
    for i in range(0, 8):
        if (data & (1 << i) != 0):
            shifts = rl_byte_munges[algorithm % len(rl_byte_munges)][i]
            result = result | (1 << shifts)
    return result


def rl_sn_to_byte(serial_number, idx):
    idx_mod = idx % 4
    data = (serial_number >> (idx_mod * 8)) & 0x000000FF
    return data


def rl_pn_to_byte(package_name, idx):
    idx_mod = idx % len(package_name)
    return ord(package_name[idx_mod])


def rl_munge_serial_and_package(package_name, serial_number):
    """ Munge up serial number and package name into a buffer.
        This is the security by obscurity part of our licensing and should be kept secret.
    """

    uint32len = 4

    # Munge package name and serial number into buffer
    if (len(package_name) < uint32len):
        alt_pack = len(package_name) * 2
    else:
        alt_pack = uint32len * 2

    # Alternate package name characters and serial number bytes
    munge_buffer = ""
    pi = 0
    si = 0
    for i in range(0, alt_pack):
        if ((i % 2) == 0):
            data = rl_pn_to_byte(package_name=package_name, idx=pi)
            pi = pi + 1
        else:
            data = rl_sn_to_byte(serial_number=serial_number, idx=si)
            si = si + 1
        munge_buffer = munge_buffer + chr(rl_byte_munge(algorithm=i, data=data))

    # Pack in whatever is left
    for i in range(pi, len(package_name)):
        munge_buffer = munge_buffer + chr(rl_byte_munge(algorithm=len(munge_buffer), data=rl_pn_to_byte(package_name=package_name, idx=i)))
    for i in range(si, uint32len):
        munge_buffer = munge_buffer + chr(rl_byte_munge(algorithm=len(munge_buffer), data=rl_sn_to_byte(serial_number=serial_number, idx=i)))

    return munge_buffer


def rl_generate_license_code(package_name, serial_number):
    """Generate a passcode for given package name and serial number"""
    
    # Munge up package name and serial number 
    munge_buffer = rl_munge_serial_and_package(package_name=package_name, serial_number=serial_number)
    
    # Create sha256 hash of munged package name and serial number
    sha = hashlib.sha256()
    sha.update(munge_buffer)
    license_code = rl_hash_to_passcode(hash_hexdigest=sha.hexdigest())
    return license_code


def check_for_collisions(serial_number):
    """Check for collisions in the package name range L9xx-5xx for the given serial number"""
    
    license_codes = []
    for cp in range(0, 10000):
        d1 = cp % 10
        d2 = (cp % 100) / 10
        d3 = (cp % 1000) / 100
        d4 = (cp % 10000) / 1000
        
        package_name = "L9%1d%1d-5%1d%1d" % (d4, d3, d2, d1)
        license_code = rl_generate_license_code(package_name=package_name, serial_number=serial_number)
        
        already_exists = filter(lambda x: x[1] == license_code, license_codes)
        if (len(already_exists) > 0):
            print "License code collision: package name 1: %s, package name 2: %s, license code: %d" % (already_exists[0][0], package_name, license_code)
        license_codes.append((package_name, license_code))


def license_code_to_string(license_code):
    license_string = ""
    for _ in range(0, LICENSECODE_LEN):
        digit = license_code % LICENSECODE_DIGIT_COMBINATIONS
        license_code = license_code / LICENSECODE_DIGIT_COMBINATIONS
        license_string = chr(digit + ord('A')) + license_string
    return license_string

class TestC500License(unittest.TestCase):

    # Package name, serial number, munged buffer, hash of buffer, hash % (26^6)
    test_list = \
    [ \
      ( "",         0x00000000, "\x00\x00\x00\x00",                                  "df3f619804a92fdb4057192dc43dd748ea778adc52bc498ce80524c014b81119", 0x1258C019), \
      ( "test",     0x00000000, "\xB8\x00\xA5\x00\x79\x00\x3A\x00",                  "ea03a6190fabdcbb90c18eeb6f2dd3122f138938ccc56cfa892a769713d5c556", 0x06ECF056), \
      ( "",         0x12345678, "\xB4\x39\x23\x18",                                  "92fa4ba377e6422cd253d62925d7bb30a7e86897ac22cd5c4c3fbd30615c63c7", 0x0A00F407), \
      ( "L900-500", 3382159,    "\x2C\xB6\xB2\x3E\x41\xAA\x18\x00\xB8\x8D\x84\x84",  "99d4cf6e9ad0fff7c516a5998233a96ff4ef117a921e20e58c84dcc8885feab3", 0x083C6873) # 3382159=0x00339B8F \
    ]

    def test_rl_hash_to_passcode(self):
        """Test converting a hash into a passcode"""
        combinations = LICENSECODE_DIGIT_COMBINATIONS ** LICENSECODE_LEN
        
        self.assertEqual(0, rl_hash_to_passcode(hash_hexdigest="%x" % (combinations + 0)))
        self.assertEqual(1, rl_hash_to_passcode(hash_hexdigest="%x" % (combinations + 1)))
        self.assertEqual(combinations - 1, rl_hash_to_passcode(hash_hexdigest="%x" % (combinations - 1)))

        # Check that rl_hash_to_passcode() matches pre-computed hash modulos
        for test in self.test_list:
            self.assertEqual(test[4], rl_hash_to_passcode(hash_hexdigest=test[3]))

    def test_rl_byte_munge(self):
        """Test byte munging"""

        # Algorithm 0: 0, 6, 3, 2, 7, 4, 5, 1
        self.assertEqual(0x00, rl_byte_munge(algorithm=0, data=0x00))
        self.assertEqual(0x01, rl_byte_munge(algorithm=0, data=0x01))
        self.assertEqual(0x40, rl_byte_munge(algorithm=0, data=0x02))
        self.assertEqual(0x08, rl_byte_munge(algorithm=0, data=0x04))
        self.assertEqual(0x09, rl_byte_munge(algorithm=0, data=0x05))

        # Algorithm 16: 0, 6, 3, 2, 7, 4, 5, 1 (check wrap)
        self.assertEqual(0x00, rl_byte_munge(algorithm=16, data=0x00))
        self.assertEqual(0x01, rl_byte_munge(algorithm=16, data=0x01))
        self.assertEqual(0x40, rl_byte_munge(algorithm=16, data=0x02))
        self.assertEqual(0x08, rl_byte_munge(algorithm=16, data=0x04))
        self.assertEqual(0x09, rl_byte_munge(algorithm=16, data=0x05))

        # Algorithm 10: 3, 1, 0, 5, 2, 7, 4, 6
        self.assertEqual(0x00, rl_byte_munge(algorithm=10, data=0x00))
        self.assertEqual(0x04, rl_byte_munge(algorithm=10, data=0x10))
        self.assertEqual(0x80, rl_byte_munge(algorithm=10, data=0x20))
        self.assertEqual(0x10, rl_byte_munge(algorithm=10, data=0x40))
        self.assertEqual(0x14, rl_byte_munge(algorithm=10, data=0x50))
    
    def test_rl_sn_to_byte(self):
        """Test serial number to byte"""
        
        self.assertEqual(0x78, rl_sn_to_byte(serial_number=0x12345678, idx=0))
        self.assertEqual(0x56, rl_sn_to_byte(serial_number=0x12345678, idx=1))
        self.assertEqual(0x34, rl_sn_to_byte(serial_number=0x12345678, idx=2))
        self.assertEqual(0x12, rl_sn_to_byte(serial_number=0x12345678, idx=3))

        self.assertEqual(0x78, rl_sn_to_byte(serial_number=0x12345678, idx=4))

    def test_rl_pn_to_byte(self):
        """Test package name to byte"""

        self.assertEqual(ord("t"), rl_pn_to_byte(package_name="test", idx=0))
        self.assertEqual(ord("e"), rl_pn_to_byte(package_name="test", idx=1))
        self.assertEqual(ord("s"), rl_pn_to_byte(package_name="test", idx=2))
        self.assertEqual(ord("t"), rl_pn_to_byte(package_name="test", idx=3))

        self.assertEqual(ord("t"), rl_pn_to_byte(package_name="test", idx=4))

    def test_rl_munge_serial_and_package(self):
        """Test munging serial number and package name into a buffer"""

        for test in self.test_list:
            res = rl_munge_serial_and_package(package_name=test[0], serial_number=test[1])
            self.assertEqual(first=test[2], second=res)

    def test_rl_generate_license_code(self):
        """ Test generating a license code from a package name and the serial number"""

        for test in self.test_list:
            code = rl_generate_license_code(package_name=test[0], serial_number=test[1])
            self.assertEqual(first=test[4], second=code)


if __name__ == '__main__':
    
    # Parse command line options
    parser = argparse.ArgumentParser(description='Generate license codes for Rinstrum C500 indicators')
    parser.add_argument('--serialno', action='store', type=int,
                        help='Serial number to generate license for. Requires --package or --collision-test.')
    parser.add_argument('--package', action='store',
                        help='Package name used when generating license. Requires --serialno.')
    parser.add_argument('--unit-test', action='store_true',
                        help='Run unit tests.')
    parser.add_argument('--collision-test', action='store_true',
                        help='Test algorithm for collisions. Parameter is upper limit of passes. Requires --serialno.')
    args = parser.parse_args()

    if (args.collision_test and args.serialno is None):
        print "--collision-test requires --serialno"
        sys.exit(1)
    if (args.package is not None and args.serialno is None):
        print "--package requires --serialno"
        sys.exit(1)
    if (args.serialno is not None and (not args.collision_test) and args.package is None):
        print "--serialno requires either --collision-test or --package"
        sys.exit(1)

    # Run commands
    if (args.unit_test):
        runner = unittest.TextTestRunner()
        itersuite = unittest.TestLoader().loadTestsFromTestCase(TestC500License)
        runner.run(itersuite)
    if (args.collision_test):
        check_for_collisions(serial_number=args.serialno)
    if (args.serialno is not None and args.package is not None):
        license_code = rl_generate_license_code(package_name=args.package, serial_number=args.serialno)
        license_code_string = license_code_to_string(license_code=license_code)
        print "License code for package %s and serial number %d is %d (%s)" % (args.package, args.serialno, license_code, license_code_string)
    sys.exit(0)
