/*
 * rl_licence.h
 *
 * Licence is the noun.
 * License is the verb.
 *
 *  Created on: Oct 19, 2016
 *      Author: ian
 */

#ifndef RL_LICENCE_H_
#define RL_LICENCE_H_

#include <stdint.h>
#include <stdbool.h>
#include <utarray.h>


/** Each C500 licence code is 6 characters in length (cccccc) */
#define LICENCECODE_LEN_C500 6


/** Each C500 licence code digit has 26 combinations (A-Z) */
#define LICENCECODE_DIGIT_COMBINATIONS_C500 "BCDFGHJKLMNPQRSTVWXYZ25789"


/** Each C500 rude licence code digit has 26 combinations (A-Z) */
#define LICENCECODE_DIGIT_COMBINATIONS_C500RUDE "ABCDEFGHIJKLMNOPQRSTUVWXYZ"


/** C500 primary licensable package name regex */
#define RL_PACKAGE_NAME_REGEX_C500_PRIMARY "C5[0-9]{2}-K5[0-9]{2}"


/** C500 add-on licensable package name regex */
#define RL_PACKAGE_NAME_REGEX_C500_ADDON "L9[0-9]{2}-5[0-9]{2}"


/** C500 licensable package name regex. Group 1 is the package name. */
#define RL_PACKAGE_NAME_REGEX_C500 "^(" RL_PACKAGE_NAME_REGEX_C500_PRIMARY "|" RL_PACKAGE_NAME_REGEX_C500_ADDON ") .*$"


/** Each R300/R400 licence code is 6 characters in length (cccccc) */
#define LICENCECODE_LEN_RX00 6


/** Each R300/R400 licence code digit has 10 combinations (0-9) */
#define LICENCECODE_DIGIT_COMBINATIONS_RX00 "0123456789"


/** Results from the rin licence library functions */
enum e_rl_result
{
  rlr_success,             /*!< Function succeeded */
  rlr_paramfail,           /*!< An invalid parameter was passed to the function */
  rlr_allocfail,           /*!< Memory allocation failed */
  rlr_regexfail,           /*!< Problem with regex */
  rlr_execfail,            /*!< Executing external program failed */
  rlr_mathfail,            /*!< Math routine failure */
  rlr_cryptofail,          /*!< Cryptographic routine failure */
  rlr_unknown_package,     /*!< Unknown package name */
  rlr_unknown_algorithm,   /*!< Unknown licensing algorithm */
  rlr_invalid_licence,     /*!< Invalid licence code */
  rlr_parsefail,           /*!< Parsing has failed */
  rlr_already_licensed,    /*!< Package is already licensed */
  rlr_licence_full,        /*!< The list of licence codes is full */
  rlr_licence_not_removed, /*!< No licence codes were removed */

  rlr_unknown_error,       /*!< Unknown error */
  rlr_num                  /*!< Number of result elements */
};


/** Strings to represent results from the rin licence library functions. Must match @ref enum e_rl_result */
#define RL_RESULT_STRS \
{ \
  "success", \
  "invalid parameter", \
  "memory allocation failure", \
  "regular expression failure", \
  "fork/exec failure", \
  "math routine failure", \
  "cryptographic routine failure", \
  "unknown package name", \
  "unknown algorithm", \
  "invalid licence code", \
  "parsing failure", \
  "package already licensed", \
  "licence code list full", \
  "licence code not removed", \
  "unknown error", \
}


/** The licensing algorithm to use */
enum e_rl_algorithm
{
  rla_C500,         /*!< Rin licence algorithm for C500 */
  rla_C500Rude,     /*!< The old rin licence algorithm for C500 that would occasionally generate rude words */
  rla_R400,         /*!< Rin licence algorithm for R400 */
  rla_R300,         /*!< Rin licence algorithm for R300 */

  rla_num
};


/** Strings for each of the licensing algorithms. Must match @ref enum e_rl_algorithm */
#define RL_ALGORITHM_STRS { "C500", "C500RUDE", "R400", "R300" }


/** Strings for each of the licensing algorithms. Must match @ref enum e_rl_algorithm */
extern const char *const rl_algorithm_strs[rla_num];


/** The package type (primary, addon or other) */
enum e_rl_package_type
{
  rlpt_primary,     /*!< The package is the primary package (ie main application like C500-K501) */
  rlpt_addon,       /*!< The package is an addon package (ie lua, eip, etc like L900-501) */
  rlpt_other,       /*!< The package is another package, not covered by licensing (ie WE21-503-M01 filesystem) */
};


/** Forward declaration */
struct s_rl_licence;


/** Function typedef to generate a licence code for given package name and serial number */
typedef enum e_rl_result (*t_rl_generate_licence_code_fn)(struct s_rl_licence *rl_licence, const char *package_name, uint32_t *licence_code);


/** Function typedef to determine the package type (primary or addon package) */
typedef enum e_rl_result (*t_rl_get_package_type_fn)(struct s_rl_licence *rl_licence, const char *package_name, enum e_rl_package_type *package_type);


/** Rin licence operating structure */
struct s_rl_licence
{
  /* Common elements */
  uint32_t serial_number;                       /*!< The unit serial number to license against */
  uint32_t *package_licence_codes;              /*!< List of currently installed licences */
  uint32_t *num_package_licence_codes;          /*!< Number of currently installed licences */
  size_t total_num_package_licence_codes;       /*!< Maximum number of elements in package_licence_codes */
  UT_array *installed_licensable_packages;      /*!< List of installed licensable (ie name matches package_regexs) packages */
  size_t licence_code_length;                   /*!< Number of digits in the licence code */
  const char *licence_code_digit_combinations;  /*!< Digit combinations in a licence code (eg 0-9, A-Z) */
  t_rl_generate_licence_code_fn rl_generate_licence_code_fn;
  t_rl_get_package_type_fn rl_get_package_type_fn;

  /* R300/R400 specific */
  UT_array *known_packages;                     /*!< List of known packages (applies to R300/R400 only) */

  /* C500 specific */
  char *licensable_package_regex;               /*!< Regex of licensable package names */
  char *primary_package_regex;                  /*!< Regex of primary package names */
  char *addon_package_regex;                    /*!< Regex of addon package names */
  char *opkg_executable;                        /*!< Opkg executable */
};


/* From rl_init_*.c: */


/** Allocate and initialise rin licence structure for an R300 */
enum e_rl_result rl_init_r300(struct s_rl_licence **rl_licence,
                              uint32_t serial_number,
                              uint32_t *package_licence_codes,
                              uint32_t *num_package_licence_codes,
                              size_t total_num_package_licence_codes,
                              const char *installed_package);


/** Allocate and initialise rin licence structure for an R400 */
enum e_rl_result rl_init_r400(struct s_rl_licence **rl_licence,
                              uint32_t serial_number,
                              uint32_t *package_licence_codes,
                              uint32_t *num_package_licence_codes,
                              size_t total_num_package_licence_codes,
                              const char *installed_package);


/** Allocate and initialise rin licence structure for a C500 */
enum e_rl_result rl_init_c500(struct s_rl_licence **rl_licence,
                              uint32_t serial_number,
                              uint32_t *package_licence_codes,
                              uint32_t *num_package_licence_codes,
                              size_t total_num_package_licence_codes,
                              const char *opkg_executable);


/**
 * Allocate and initialise rin licence structure for a C500Rude.
 *
 * This is the old C500 algorithm that would generate rude words
 * from time to time
 */
enum e_rl_result rl_init_c500rude(struct s_rl_licence **rl_licence,
                                  uint32_t serial_number,
                                  uint32_t *package_licence_codes,
                                  uint32_t *num_package_licence_codes,
                                  size_t total_num_package_licence_codes,
                                  const char *opkg_executable);


/** Clean up rin licence structure */
enum e_rl_result rl_destroy(struct s_rl_licence *rl_licence);


/** Change the serial number */
enum e_rl_result rl_set_serial_number(struct s_rl_licence *rl_licence, uint32_t new_serial_number);


/* From rl_package.c: */


/** Get the number of known packages */
enum e_rl_result rl_get_num_known_packages(struct s_rl_licence *rl_licence, size_t *num_packages);


/** Get a known package for index */
enum e_rl_result rl_get_known_package(struct s_rl_licence *rl_licence, size_t index, const char **package_name, uint32_t *pass);


/** Get the number of installed packages */
enum e_rl_result rl_get_num_installed_packages(struct s_rl_licence *rl_licence, size_t *num_packages);


/** Get the installed package name for index */
enum e_rl_result rl_get_installed_package(struct s_rl_licence *rl_licence, size_t index, const char **package_name);


/** Check if a package is installed */
enum e_rl_result rl_package_is_installed(struct s_rl_licence *rl_licence, const char *package_name, bool *result);


/* From rl_licensing.c: */


/** Generate a licence code for given package name and serial number */
enum e_rl_result rl_generate_licence_code(struct s_rl_licence *rl_licence, const char *package_name, uint32_t *licence_code);


/** Confirm if a package is licensed */
enum e_rl_result rl_package_is_licensed(struct s_rl_licence *rl_licence, const char *package_name, bool *is_licensed);


/** License an installed package via a licence code number */
enum e_rl_result rl_license_package_uint32(struct s_rl_licence *rl_licence, const char *package_name, uint32_t licence_code);


/** License an installed package via a licence string */
enum e_rl_result rl_license_package_string(struct s_rl_licence *rl_licence, const char *package_name, const char *licence_code_str);


/** Are all installed licensable packages licensed? */
enum e_rl_result rl_all_installed_licensable_licensed(struct s_rl_licence *rl_licence, bool *all_licensed);


/* From rl_remove.c */


/** Unlicense an installed package. Returns rlr_licence_not_removed is no licence code removed */
enum e_rl_result rl_unlicense_package(struct s_rl_licence *rl_licence, const char *package_name);


/** Remove any licence codes that don't have installed packages. Returns rlr_licence_not_removed is no licence code removed */
enum e_rl_result rl_flush_licence_codes_for_uninstalled_packages(struct s_rl_licence *rl_licence);


/* From rl_string.c: */


/**
 * Convert a licence code (0 .. len(rl_licence->licence_code_digit_combinations)^rl_licence->licence_code_length)
 * into a string of length rl_licence->licence_code_length containing characters from rl_licence->licence_code_digit_combinations
 *
 * Examples:
 * R300/R400: 0 -> "000000"
 * C500: 0 -> "BBBBBB", (26^6) - 1 -> "999999"
 *
 * @param rl_licence is a licence code structure
 * @param licence_code is the licence code to convert
 * @param licence_code_str is a string of minimum length rl_licence->licence_code_length + 1 into which the licence code is written. Null terminated.
 * @param licence_code_str_len is the length of licence_code_str
 */
enum e_rl_result rl_licence_code_to_string(struct s_rl_licence *rl_licence, uint32_t licence_code, char *licence_code_str, size_t licence_code_str_len);


/**
 * Convert a string of length rl_licence->licence_code_length containing characters from rl_licence->licence_code_digit_combinations
 * into a licence code (0 .. len(rl_licence->licence_code_digit_combinations)^rl_licence->licence_code_length)
 *
 * Examples:
 * R300/R400: "000000" -> 0
 * C500: "BBBBBB" -> 0, "999999" -> (26^6) - 1
 *
 * @param rl_licence is a licence code structure
 * @param licence_code_str is a string of length rl_licence->licence_code_length + 1 to be converted to a licence code. Null terminated.
 * @param licence_code is the converted licence code
 */
enum e_rl_result rl_string_to_licence_code(struct s_rl_licence *rl_licence, const char *licence_code_str, uint32_t *licence_code);


/* From rl_error.c */


/** Convert a result code into a human readable string */
const char *rl_result_to_str(enum e_rl_result result);


/* From rl_pkgtype.c */


/** Get the package type for package_name */
enum e_rl_result rl_get_package_type(struct s_rl_licence *rl_licence, const char *package_name, enum e_rl_package_type *package_type);


#endif /* RL_LICENCE_H_ */
