/*
 * rl_licensing.c
 *
 *  Created on: Nov 21, 2016
 *      Author: ian
 */
#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>

#include <rinlicence/rl_licence.h>


/** Generate a licence code for given package name and serial number */
enum e_rl_result rl_generate_licence_code(struct s_rl_licence *rl_licence, const char *package_name, uint32_t *licence_code)
{
  if (NULL == rl_licence || NULL == package_name || NULL == licence_code)
    return rlr_paramfail;

  if (NULL == rl_licence->rl_generate_licence_code_fn)
    return rlr_unknown_algorithm;

  return rl_licence->rl_generate_licence_code_fn(rl_licence, package_name, licence_code);
}


/** Internal function to check if a package is licensed, assumes it is installed */
static enum e_rl_result rl_internal_packaged_is_licensed(struct s_rl_licence *rl_licence, const char *package_name, bool *is_licensed)
{
  enum e_rl_result result;
  uint32_t licence_code;
  size_t i;

  *is_licensed = false;

  /* Generate the licence code for this package */
  result = rl_generate_licence_code(rl_licence, package_name, &licence_code);
  if (rlr_success != result)
  {
    return result;
  }

  /* Is this licence code in the list */
  for (i = 0; i < *rl_licence->num_package_licence_codes; i++)
  {
    if (rl_licence->package_licence_codes[i] == licence_code)
    {
      /* Yep, then package is licensed */
      *is_licensed = true;
      break;
    }
  }
  return rlr_success;
}


/** Confirm if a package is licensed */
enum e_rl_result rl_package_is_licensed(struct s_rl_licence *rl_licence, const char *package_name, bool *is_licensed)
{
  enum e_rl_result result;
  bool installed;

  /* Check parameters */
  if (NULL == rl_licence || NULL == package_name || NULL == is_licensed)
    return rlr_paramfail;

  *is_licensed = false;

  /* Package must be installed */
  installed = false;
  result = rl_package_is_installed(rl_licence, package_name, &installed);
  if (rlr_success != result)
  {
    return result;
  }
  if (!installed)
    return rlr_unknown_package;

  /* Check if package_name is licensed */
  return rl_internal_packaged_is_licensed(rl_licence, package_name, is_licensed);
}


/** License an installed package via a licence code number */
enum e_rl_result rl_license_package_uint32(struct s_rl_licence *rl_licence, const char *package_name, uint32_t licence_code)
{
  enum e_rl_result result;
  bool installed;
  size_t i;
  uint32_t generated_licence_code;

  /* Check parameters */
  if (NULL == rl_licence || NULL == package_name)
    return rlr_paramfail;

  /* Package must be installed */
  installed = false;
  result = rl_package_is_installed(rl_licence, package_name, &installed);
  if (rlr_success != result)
  {
    return result;
  }
  if (!installed)
  {
    return rlr_unknown_package;
  }

  /* Package installed, does supplied licence code match the generated one? */
  result = rl_generate_licence_code(rl_licence, package_name, &generated_licence_code);
  if (rlr_success != result)
  {
    return result;
  }
  if (generated_licence_code != licence_code)
  {
    return rlr_invalid_licence;
  }

  /* Is this licence code in the list (already licensed) */
  for (i = 0; i < *rl_licence->num_package_licence_codes; i++)
  {
    if (rl_licence->package_licence_codes[i] == licence_code)
    {
      /* Yep, package is already licensed */
      return rlr_already_licensed;
    }
  }

  /* Not already licensed, is there room for it? */
  if (*rl_licence->num_package_licence_codes >= rl_licence->total_num_package_licence_codes)
  {
    return rlr_licence_full;
  }

  /* Add to end of list */
  rl_licence->package_licence_codes[*rl_licence->num_package_licence_codes] = licence_code;
  (*rl_licence->num_package_licence_codes)++;
  return rlr_success;
}


/** License an installed package via a licence string */
enum e_rl_result rl_license_package_string(struct s_rl_licence *rl_licence, const char *package_name, const char *licence_code_str)
{
  enum e_rl_result result;
  uint32_t licence_code;

  /* Licence code string to number */
  result = rl_string_to_licence_code(rl_licence, licence_code_str, &licence_code);
  if (rlr_success != result)
    return result;

  /* License it */
  return rl_license_package_uint32(rl_licence, package_name, licence_code);
}


/** Are all installed licensable packages licensed? */
enum e_rl_result rl_all_installed_licensable_licensed(struct s_rl_licence *rl_licence, bool *all_licensed)
{
  const char *package_name;
  size_t i;
  size_t num_installed_packages;
  enum e_rl_result result;
  bool is_licensed;

  if (NULL == rl_licence || NULL == all_licensed)
    return rlr_paramfail;

  result = rl_get_num_installed_packages(rl_licence, &num_installed_packages);
  if (rlr_success != result)
    return result;

  for (i = 0; i < num_installed_packages; i++)
  {
    /* Get each package name */
    result = rl_get_installed_package(rl_licence, i, &package_name);
    if (rlr_success != result)
      return result;

    /* Is it licensed? */
    result = rl_internal_packaged_is_licensed(rl_licence, package_name, &is_licensed);
    if (rlr_success != result)
      return result;
    if (!is_licensed)
    {
      *all_licensed = false;
      return rlr_success;
    }
  }

  /* All packages are licensed */
  *all_licensed = true;
  return rlr_success;
}
