/*
 * rl_opkg.c
 *
 *  Created on: Oct 17, 2016
 *      Author: ian
 */
#include <stdio.h>
#include <regex.h>
#include <rinexec/el_runcmd.h>
#include <rinexec/el_readpipe_bufalloc.h>
#include <rinexec/el_writepipe_buffer.h>

#include <rinlicence/rl_licence.h>
#include "rl_opkg.h"


/*
 * opkg-cl list-installed we21*
 */


/** Filter the list of installed packages down to those that match the regex */
static enum e_rl_result rl_opkg_filter_installed_packages(struct s_rl_licence *rl_licence, regex_t *regex, const char *installed_package)
{
  int regex_ret;
  enum e_rl_result result;
  #define REGEX_GROUPS 2
  #define PKG_NAME_GROUP 1
  regmatch_t regex_groups[REGEX_GROUPS];
  size_t so;
  size_t eo;
  size_t len;
  char *package_name;

  /* Parse the shadow password entry against the regex, and pick it apart */
  regex_ret = regexec(regex, installed_package, REGEX_GROUPS, regex_groups, 0);
  if (REG_NOERROR == regex_ret)
  {
    if (regex_groups[PKG_NAME_GROUP].rm_so == -1)
      return rlr_regexfail;

    /* Matches, add this package to the installed_licensable_packages list */
    so = regex_groups[PKG_NAME_GROUP].rm_so;
    eo = regex_groups[PKG_NAME_GROUP].rm_eo;
    len = eo - so;
    package_name = calloc(1, len + 1);
    if (NULL == package_name)
      return rlr_allocfail;
    memcpy(package_name, &installed_package[so], len);
    package_name[len] = '\0';

    utarray_push_back(rl_licence->installed_licensable_packages, &package_name);
    free(package_name);
    result = rlr_success;
  }
  else if (REG_NOMATCH == regex_ret)
  {
    /* Doesn't match, nothing to do */
    result = rlr_success;
  }
  else
  {
    result = rlr_regexfail;
  }
  return result;
}


/** Get a list of installed packages */
enum e_rl_result rl_opkg_get_installed_packages(struct s_rl_licence *rl_licence)
{
  struct s_runcmd_pipe_config rcpc;
  struct s_runcmd_bufferalloc rcba;
  #define NUM_OPKG_ARGS 3
  const char *opkg_cmd[NUM_OPKG_ARGS];
  int ret_code;
  enum e_rl_result result;
  enum e_el_result res;
  char *ptr;
  char *current_line;
  regex_t regex;
  bool regex_inited = false;

  /* Check parameters */
  if (NULL == rl_licence)
    return rlr_paramfail;
  if (NULL == rl_licence->opkg_executable || NULL == rl_licence->licensable_package_regex)
    return rlr_execfail;

  /* Create opkg command */
  opkg_cmd[0] = rl_licence->opkg_executable;
  opkg_cmd[1] = "list-installed";
  opkg_cmd[2] = NULL;

  /* Configure to read back list of installed packages from opkg */
  el_init_runcmd_pipe_config(&rcpc);
  el_init_read_cmd_pipe_stringalloc(&rcpc, &rcba, 1024, erp_stdout);

  /* Run "opkg-cl", wait for the command to complete, and read back the list of installed packages */
  res = el_run_command(opkg_cmd[0], opkg_cmd, &ret_code, &rcpc, true, true, true);
  if (elr_success != res || !el_return_code_has_exited(ret_code) || 0 != el_convert_return_code(ret_code) || NULL == rcba.data || 0 == rcba.data_posn)
  {
    result = rlr_execfail;
    goto rogip_fail;
  }

  /* Compile regex, ready for comparison against installed packages */
  if (regcomp(&regex, rl_licence->licensable_package_regex, REG_EXTENDED) != REG_NOERROR)
  {
    result = rlr_regexfail;
    goto rogip_fail;
  }
  regex_inited = true;

  /* Iterate through opkg list string, turning each line into a null terminated string, and then checking against the regex */
  for (ptr = rcba.data, current_line = rcba.data; '\0' != *ptr; ptr++)
  {
    if ('\n' == *ptr)
    {
      *ptr = '\0';
      result = rl_opkg_filter_installed_packages(rl_licence, &regex, current_line);
      if (rlr_success != result)
      {
        goto rogip_fail;
      }
      current_line = &ptr[1];
    }
  }

  /* Success */
  result = rlr_success;

rogip_fail:
  if (regex_inited)
    regfree(&regex);
  el_free_read_cmd_pipe_stringalloc(&rcba);
  return result;
}


/** Try to match a package name against a regex */
enum e_rl_result rl_opkg_package_match_regex(struct s_rl_licence *rl_licence, const char *regex_str, const char *package_name)
{
  enum e_rl_result result;
  regex_t regex;
  bool regex_inited = false;
  int regex_ret;
  #define PKG_MATCH_REGEX_GROUPS 1
  regmatch_t regex_groups[PKG_MATCH_REGEX_GROUPS];

  /* Check parameters */
  if (NULL == rl_licence || NULL == regex_str || NULL == package_name)
    return rlr_paramfail;

  /* Compile regex, ready for comparison against package name */
  if (regcomp(&regex, regex_str, REG_EXTENDED) != REG_NOERROR)
  {
    result = rlr_regexfail;
    goto rogip_fail;
  }
  regex_inited = true;

  /* Parse the shadow password entry against the regex, and pick it apart */
  regex_ret = regexec(&regex, package_name, PKG_MATCH_REGEX_GROUPS, regex_groups, 0);
  if (REG_NOERROR == regex_ret)
  {
    result = rlr_success;
  }
  else if (REG_NOMATCH == regex_ret)
  {
    /* Doesn't match, nothing to do */
    result = rlr_unknown_package;
  }
  else
  {
    result = rlr_regexfail;
  }

rogip_fail:
  if (regex_inited)
    regfree(&regex);
  return result;
}
