/*
 * rl_string.c
 *
 *  Created on: Nov 1, 2016
 *      Author: ian
 */
#include <string.h>

#include <rinlicence/rl_licence.h>
#include "rl_string.h"


/**
 * Convert a licence code (0 .. len(rl_licence->licence_code_digit_combinations)^rl_licence->licence_code_length)
 * into a string of length rl_licence->licence_code_length containing characters from rl_licence->licence_code_digit_combinations
 *
 * Examples:
 * R300/R400: 0 -> "000000"
 * C500: 0 -> "BBBBBB", (26^6) - 1 -> "999999"
 *
 * @param rl_licence is a licence code structure
 * @param licence_code is the licence code to convert
 * @param licence_code_str is a string of minimum length rl_licence->licence_code_length + 1 into which the licence code is written. Null terminated.
 * @param licence_code_str_len is the length of licence_code_str
 */
enum e_rl_result rl_licence_code_to_string(struct s_rl_licence *rl_licence, uint32_t licence_code, char *licence_code_str, size_t licence_code_str_len)
{
  size_t i;
  uint32_t licence_mod;
  size_t num_combinations;

  if (NULL == rl_licence || NULL == licence_code_str)
    return rlr_paramfail;

  if (licence_code_str_len < rl_licence->licence_code_length + 1)
    return rlr_paramfail;

  num_combinations = strlen(rl_licence->licence_code_digit_combinations);

  for (i = 0; i < rl_licence->licence_code_length; i++)
  {
    licence_mod = licence_code % num_combinations;
    licence_code /= num_combinations;
    licence_code_str[rl_licence->licence_code_length - 1 - i] = rl_licence->licence_code_digit_combinations[licence_mod];
  }

  licence_code_str[rl_licence->licence_code_length] = '\0';
  return rlr_success;
}


/**
 * Convert a string of length rl_licence->licence_code_length containing characters from rl_licence->licence_code_digit_combinations
 * into a licence code (0 .. len(rl_licence->licence_code_digit_combinations)^rl_licence->licence_code_length)
 *
 * Examples:
 * R300/R400: "000000" -> 0
 * C500: "BBBBBB" -> 0, "999999" -> (26^6) - 1
 *
 * @param rl_licence is a licence code structure
 * @param licence_code_str is a string of length rl_licence->licence_code_length + 1 to be converted to a licence code. Null terminated.
 * @param licence_code is the converted licence code
 */
enum e_rl_result rl_string_to_licence_code(struct s_rl_licence *rl_licence, const char *licence_code_str, uint32_t *licence_code)
{
  size_t i;
  const char *ptr;
  ptrdiff_t digit;
  char ch;
  size_t base;

  /* Check bad parameters */
  if (NULL == rl_licence || NULL == licence_code_str || NULL == licence_code)
    return rlr_paramfail;

  base = strlen(rl_licence->licence_code_digit_combinations);

  /* Convert string to code */
  *licence_code = 0;
  for (i = 0; '\0' != licence_code_str[i] && i < rl_licence->licence_code_length; i++)
  {
    ch = licence_code_str[i];
    ptr = strchr(rl_licence->licence_code_digit_combinations, ch);
    if (NULL == ptr)
    {
      /* Character is not in the set of valid characters */
      return rlr_parsefail;
    }
    digit = ptr - rl_licence->licence_code_digit_combinations;

    /* Character to number */
    *licence_code *= base;
    *licence_code += digit;
  }

  return ('\0' == licence_code_str[i]) ? rlr_success : rlr_parsefail;
}
