/*
 * rinlicence.c
 *
 *  Created on: Nov 4, 2016
 *      Author: ian
 */
#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <getopt.h>
#include <stdlib.h>
#include <rinlicence/rl_licence.h>


enum e_command { cmd_list_passes, cmd_list_algorithms, cmd_generate };


static bool cmdline_display_help = false;
static const char *cmdline_package_name = NULL;
static enum e_rl_algorithm cmdline_algorithm = rla_num;
static uint32_t cmdline_serial_number = 0;
static enum e_command cmdline_command = cmd_list_passes;
static bool cmdline_display_numeric = false;


#define TOTAL_NUM_PACKAGE_LICENCE_CODES 1
static uint32_t package_licence_code = 0;
static uint32_t num_package_licence_codes = 0;


static void cmdline_print_usage(const char *prog)
{
  enum e_rl_algorithm algo;

  printf("Usage: %s [options] <command>\n"
         "Rinstrum Licensing\n", prog);

  printf("Options:\n"
         "  -h --help              this help\n"
         "  -p --package <name>    the package name to generate the licence code for\n"
         "  -s --serialno <num>    the serial number to generate the licence code for\n"
         "  -a --algorithm <alg>   the algorithm to use to generate the licence code\n"
         "  -n --numeric           display the licence code as a number rather than string (C500)\n"
         "Commands:\n"
         "  list-passes            list the passes vs package names. requires -a\n"
         "  list-algorithms        list the licensing algorithms supported\n"
         "  generate               generate a licence code. requires -p, -s and -a\n"
         "\n"
        );

  printf("Algorithms: ");
  for (algo = 0; algo < rla_num; algo++)
  {
    printf("%s ", rl_algorithm_strs[algo]);
  }
  printf("\n");
}


static int cmdline_parse_options(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "help",        0, 0, 'h' },
    { "package",     1, 0, 'p' },
    { "serialno",    1, 0, 's' },
    { "algorithm",   1, 0, 'a' },
    { "numeric",     0, 0, 'n' },

    { NULL,          0, 0, '\0' },
  };
  int c;
  size_t i;
  bool serialno_set = false;

  /* Options */
  while (1)
  {
    c = getopt_long(argc, argv, "hp:s:a:n", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      /* Help */
      case 'h':
        cmdline_display_help = true;
        return 0;
        break;

      /* Package name */
      case 'p':
        cmdline_package_name = optarg;
        break;

      /* Serial number */
      case 's':
        errno = 0;
        cmdline_serial_number = strtoul(optarg, NULL, 10);
        if (0 != errno)
        {
          fprintf(stderr, "error: invalid serial number, try --help\n");
          return -1;
        }
        serialno_set = true;
        break;

      /* Algorithm */
      case 'a':
        for (i = 0, cmdline_algorithm = rla_num; i < rla_num; i++)
        {
          if (strcasecmp(optarg, rl_algorithm_strs[i]) == 0)
          {
            cmdline_algorithm = i;
            break;
          }
        }
        if (rla_num == cmdline_algorithm)
        {
          fprintf(stderr, "error: invalid algorithm specified, try --help\n");
          return -1;
        }
        break;

      /* Numeric */
      case 'n':
        cmdline_display_numeric = true;
        break;

      default:
        fprintf(stderr, "error: invalid command line option, try --help\n");
        return -1;
        break;
    }
  }

  /* Command */
  if (optind >= argc)
  {
    fprintf(stderr, "error: no command specified, try --help\n");
    return -1;
  }
  if (strcmp("list-passes", argv[optind]) == 0)
    cmdline_command = cmd_list_passes;
  else if (strcmp("list-algorithms", argv[optind]) == 0)
    cmdline_command = cmd_list_algorithms;
  else if (strcmp("generate", argv[optind]) == 0)
    cmdline_command = cmd_generate;
  else
  {
    fprintf(stderr, "error: invalid command '%s' specified, try --help\n", argv[optind]);
    return -1;
  }
  optind++;

  /* Check parameters */
  if (cmd_list_passes == cmdline_command && rla_num == cmdline_algorithm)
  {
    fprintf(stderr, "error: list-passes requires --algorithm\n");
    return -1;
  }
  if (cmd_generate == cmdline_command &&
      (NULL == cmdline_package_name ||
       rla_num == cmdline_algorithm ||
       !serialno_set))
  {
    fprintf(stderr, "error: generate requires --algorithm, --serialno and --package\n");
    return -1;
  }

  return 0;
}


static enum e_rl_result create_by_algorithm(enum e_rl_algorithm algorithm, struct s_rl_licence **rl_licence)
{
  enum e_rl_result result;

  switch (algorithm)
  {
    case rla_C500:
      result = rl_init_c500(rl_licence, cmdline_serial_number, &package_licence_code, &num_package_licence_codes, TOTAL_NUM_PACKAGE_LICENCE_CODES, NULL);
      break;
    case rla_C500Rude:
      result = rl_init_c500rude(rl_licence, cmdline_serial_number, &package_licence_code, &num_package_licence_codes, TOTAL_NUM_PACKAGE_LICENCE_CODES, NULL);
      break;
    case rla_R300:
      result = rl_init_r300(rl_licence, cmdline_serial_number, &package_licence_code, &num_package_licence_codes, TOTAL_NUM_PACKAGE_LICENCE_CODES, NULL);
      break;
    case rla_R400:
      result = rl_init_r400(rl_licence, cmdline_serial_number, &package_licence_code, &num_package_licence_codes, TOTAL_NUM_PACKAGE_LICENCE_CODES, NULL);
      break;
    default:
      result = rlr_unknown_algorithm;
      break;
  }
  return result;
}


static enum e_rl_result list_passes(struct s_rl_licence *rl_licence)
{
  enum e_rl_result result;
  size_t num_packages;
  size_t i;
  const char *package_name;
  uint32_t pass;

  result = rl_get_num_known_packages(rl_licence, &num_packages);
  if (rlr_success != result)
    return result;

  printf("Pass Package\n");
  for (i = 0; i < num_packages; i++)
  {
    result = rl_get_known_package(rl_licence, i, &package_name, &pass);
    if (rlr_success != result)
      return result;
    printf("%4lu %s\n", (unsigned long) pass, package_name);
  }
  return rlr_success;
}


static enum e_rl_result list_algorithms(void)
{
  enum e_rl_algorithm algo;

  for (algo = 0; algo < rla_num; algo++)
  {
    printf("%s\n", rl_algorithm_strs[algo]);
  }
  return rlr_success;
}


static enum e_rl_result generate_passcode(struct s_rl_licence *rl_licence)
{
  enum e_rl_result result;
  uint32_t licence_code;
  char licence_code_str[100];

  /* Generate licence code */
  result = rl_generate_licence_code(rl_licence, cmdline_package_name, &licence_code);
  if (rlr_success != result)
    return result;

  result = rl_licence_code_to_string(rl_licence, licence_code, licence_code_str, 100);
  if (rlr_success != result)
    return result;

  /* Success */
  if (cmdline_display_numeric)
    printf("%lu\n", (unsigned long) licence_code);
  else
    printf("%s\n", licence_code_str);
  return rlr_success;
}


/** Application entry point */
int main(int argc, char *argv[])
{
  enum e_rl_result result;
  struct s_rl_licence *rl_licence = NULL;
  int mresult;

  /* Do command line */
  if (cmdline_parse_options(argc, argv) < 0)
  {
    return 1;
  }
  if (cmdline_display_help)
  {
    cmdline_print_usage(argv[0]);
    return 0;
  }

  if (cmd_list_algorithms == cmdline_command)
  {
    result = list_algorithms();
  }
  else
  {
    /* Make rinlicence structure */
    result = create_by_algorithm(cmdline_algorithm, &rl_licence);
    if (rlr_success != result)
    {
      goto m_fail;
    }

    /* Dispatch function */
    switch (cmdline_command)
    {
      case cmd_list_passes:
        result = list_passes(rl_licence);
        break;
      case cmd_list_algorithms:
        result = list_algorithms();
        break;
      case cmd_generate:
        result = generate_passcode(rl_licence);
        break;
      default:
        result = rlr_unknown_error;
        break;
    }
  }

m_fail:
  if (rlr_success != result)
  {
    /* Code to string */
    fprintf(stderr, "failure: %s\n", rl_result_to_str(result));
    mresult = 1;
  }
  else
  {
    mresult = 0;
  }
  if (NULL != rl_licence)
  {
    (void) rl_destroy(rl_licence);
    rl_licence = NULL;
  }

  return mresult;
}
