/*
 * rl_test_license.c
 *
 *  Created on: Nov 21, 2016
 *      Author: ian
 */
#include <rinlicence/rl_licence.h>
#include <cutest/CuTest.h>

#include "rl_test_license.h"


static const char opkg_executable_4_licensed[] = "./opkg-4-licensed";


#define NUM_PACKAGE_LICENCE_CODES 2


/** Check licensing a package */
static void test_rinlicence_licensing(CuTest *tc)
{
  enum e_rl_result result;
  struct s_rl_licence *rl_licence;
  uint32_t package_licence_codes[NUM_PACKAGE_LICENCE_CODES];
  uint32_t num_package_licence_codes = 0;
  uint32_t serial_number = 3382549;
  bool is_licensed;

  /* Create a C500 licensing structure */
  result = rl_init_c500(&rl_licence, serial_number, package_licence_codes, &num_package_licence_codes, NUM_PACKAGE_LICENCE_CODES, opkg_executable_4_licensed);
  CuAssertIntEquals(tc, rlr_success, result);

  /*
   * License packages
   */

  /* Check dodgy parameters */
  result = rl_license_package_uint32(NULL, "", 0);
  CuAssertIntEquals(tc, rlr_paramfail, result);
  result = rl_license_package_uint32(rl_licence, NULL, 0);
  CuAssertIntEquals(tc, rlr_paramfail, result);

  /* Try to license a package that is not installed */
  result = rl_license_package_uint32(rl_licence, "L900-510", 179999666);
  CuAssertIntEquals(tc, rlr_unknown_package, result);

  /* Try to license an installed package with a bad licence code */
  result = rl_license_package_uint32(rl_licence, "L900-520", 280668447 + 1);
  CuAssertIntEquals(tc, rlr_invalid_licence, result);

  /* License it */
  result = rl_license_package_uint32(rl_licence, "L900-520", 280668447);
  CuAssertIntEquals(tc, rlr_success, result);
  CuAssertIntEquals(tc, 1, num_package_licence_codes);
  CuAssertIntEquals(tc, 280668447, package_licence_codes[0]);

  /* Try to license it twice */
  result = rl_license_package_uint32(rl_licence, "L900-520", 280668447);
  CuAssertIntEquals(tc, rlr_already_licensed, result);
  CuAssertIntEquals(tc, 1, num_package_licence_codes);

  /* License more to fill array */
  result = rl_license_package_uint32(rl_licence, "L900-500", 138093037);
  CuAssertIntEquals(tc, rlr_success, result);
  CuAssertIntEquals(tc, 2, num_package_licence_codes);
  CuAssertIntEquals(tc, 138093037, package_licence_codes[1]);

  /* Try to license more now that array is full */
  result = rl_license_package_uint32(rl_licence, "L940-500", 33255967);
  CuAssertIntEquals(tc, rlr_licence_full, result);

  /*
   * Check licensing on packages
   */

  /* Check if packages are licensed/not */
  result = rl_package_is_licensed(rl_licence, "L900-520", &is_licensed);
  CuAssertIntEquals(tc, rlr_success, result);
  CuAssertTrue(tc, is_licensed);
  result = rl_package_is_licensed(rl_licence, "L900-500", &is_licensed);
  CuAssertIntEquals(tc, rlr_success, result);
  CuAssertTrue(tc, is_licensed);
  result = rl_package_is_licensed(rl_licence, "L940-500", &is_licensed);
  CuAssertIntEquals(tc, rlr_success, result);
  CuAssertTrue(tc, !is_licensed);

  /* Check dodgy parameters to rl_package_is_licensed() */
  result = rl_package_is_licensed(NULL, "L900-520", &is_licensed);
  CuAssertIntEquals(tc, rlr_paramfail, result);
  result = rl_package_is_licensed(rl_licence, NULL, &is_licensed);
  CuAssertIntEquals(tc, rlr_paramfail, result);
  result = rl_package_is_licensed(rl_licence, "L900-520", NULL);
  CuAssertIntEquals(tc, rlr_paramfail, result);

  /*
   * Clean up
   */

  /* Clean up */
  result = rl_destroy(rl_licence);
  CuAssertIntEquals(tc, rlr_success, result);
}


/** Check all packages licenced */
static void test_rinlicence_all_licensed(CuTest *tc)
{
  enum e_rl_result result;
  struct s_rl_licence *rl_licence;
  #define NUM_PLC_ALL_LICENSED 4
  uint32_t package_licence_codes[NUM_PLC_ALL_LICENSED];
  uint32_t num_package_licence_codes = 0;
  uint32_t serial_number = 3382549;
  bool all_licensed;

  /* Create a C500 licensing structure */
  result = rl_init_c500(&rl_licence, serial_number, package_licence_codes, &num_package_licence_codes, NUM_PLC_ALL_LICENSED, opkg_executable_4_licensed);
  CuAssertIntEquals(tc, rlr_success, result);

  /* Check dodgy parameters to rl_all_installed_licensable_licensed() */
  result = rl_all_installed_licensable_licensed(NULL, &all_licensed);
  CuAssertIntEquals(tc, rlr_paramfail, result);
  result = rl_all_installed_licensable_licensed(rl_licence, NULL);
  CuAssertIntEquals(tc, rlr_paramfail, result);

  /* License 2 of the 4 installed packages */
  result = rl_license_package_uint32(rl_licence, "L900-520", 280668447);
  CuAssertIntEquals(tc, rlr_success, result);
  result = rl_license_package_uint32(rl_licence, "L900-500", 138093037);
  CuAssertIntEquals(tc, rlr_success, result);

  /* Not all licensed */
  result = rl_all_installed_licensable_licensed(rl_licence, &all_licensed);
  CuAssertIntEquals(tc, rlr_success, result);
  CuAssertTrue(tc, !all_licensed);

  /* License 4 of the 4 installed packages */
  result = rl_license_package_uint32(rl_licence, "L940-500", 33255967);
  CuAssertIntEquals(tc, rlr_success, result);
  result = rl_license_package_uint32(rl_licence, "C500-K501", 144953014);
  CuAssertIntEquals(tc, rlr_success, result);

  /* All licensed */
  result = rl_all_installed_licensable_licensed(rl_licence, &all_licensed);
  CuAssertIntEquals(tc, rlr_success, result);
  CuAssertTrue(tc, all_licensed);

  /* Clean up */
  result = rl_destroy(rl_licence);
  CuAssertIntEquals(tc, rlr_success, result);
}


/** Check licensing a package via a string licence code */
static void test_rinlicence_licensing_by_string(CuTest *tc)
{
  enum e_rl_result result;
  struct s_rl_licence *rl_licence;
  uint32_t package_licence_codes[NUM_PACKAGE_LICENCE_CODES];
  uint32_t num_package_licence_codes = 0;
  uint32_t serial_number = 3382549;

  /* Create a C500 licensing structure */
  result = rl_init_c500(&rl_licence, serial_number, package_licence_codes, &num_package_licence_codes, NUM_PACKAGE_LICENCE_CODES, opkg_executable_4_licensed);
  CuAssertIntEquals(tc, rlr_success, result);

  /* Check dodgy parameters */
  result = rl_license_package_string(NULL, "", "");
  CuAssertIntEquals(tc, rlr_paramfail, result);
  result = rl_license_package_string(rl_licence, NULL, "");
  CuAssertIntEquals(tc, rlr_paramfail, result);
  result = rl_license_package_string(rl_licence, "", NULL);
  CuAssertIntEquals(tc, rlr_paramfail, result);

  /* License some packages */
  result = rl_license_package_string(rl_licence, "L900-520", "7VG5BK");
  CuAssertIntEquals(tc, rlr_success, result);
  result = rl_license_package_string(rl_licence, "L900-500", "PVG7VW");
  CuAssertIntEquals(tc, rlr_success, result);
  CuAssertIntEquals(tc, 2, num_package_licence_codes);

  /* Clean up */
  result = rl_destroy(rl_licence);
  CuAssertIntEquals(tc, rlr_success, result);
}


CuSuite *CuGetSuite_rinlicence_licensing(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rinlicence_licensing);
  SUITE_ADD_TEST(suite, test_rinlicence_all_licensed);
  SUITE_ADD_TEST(suite, test_rinlicence_licensing_by_string);

  return suite;
}
