/*
 * rinprog.h
 *
 *  Created on: Aug 11, 2017
 *      Author: ian
 */

#ifndef RINPROG_H_
#define RINPROG_H_

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>


enum e_rp_result
{
  rpr_success,            /*!< Function was successful */
  rpr_paramfail,          /*!< An invalid parameter was passed to the function */
  rpr_iofail,             /*!< An IO operation failed */
  rpr_streamsizefail,     /*!< A stream was of an incorrect size */
  rpr_allocfail,          /*!< A memory allocation has failed */
  rpr_negotiatefail,      /*!< Failed to negotiate new baud rate with host */
  rpr_connectfail,        /*!< Failed to connect to device */
  rpr_checksumfail,       /*!< Image checksum didn't match data */
  rpr_sizefail,           /*!< Image size doesn't match device */
  rpr_flasherasefail,     /*!< Failed to erase flash on the device */
  rpr_flashwritefail,     /*!< Failed to write flash on the device */
  rpr_flashverifyfail,    /*!< Failed to verify flash on the device */
  rpr_unexpectedresponse, /*!< An unexpected response was received from the device */
  rpr_timedout,           /*!< Communication with the device timed out waiting for a response */
  rpr_chkseed_mismatch,   /*!< The device checksum seed doesn't match the checksum seed in the device settings */
  rpr_startaddr_mismatch, /*!< The device start address doesn't match the start address in the device settings */
  rpr_length_mismatch,    /*!< The device length doesn't match the length in the device settings */
  rpr_unknown_error,      /*!< Something weird has happened */

  rpr_num
};


#define RP_RESULT_STRINGS \
{ \
  "success", \
  "invalid function parameter", \
  "io operation failed", \
  "incorrect stream size", \
  "memory allocation failed", \
  "baud rate negotiation failed", \
  "connect to device failed", \
  "image checksum mismatch", \
  "image size mismatch", \
  "device flash erase failure", \
  "device flash write failure", \
  "device flash verify failure", \
  "unexpected response from device",\
  "time out communicating with device", \
  "device checksum seed doesn't match settings", \
  "device start address doesn't match settings", \
  "device length doesn't match settings", \
  "unknown error" \
}


struct s_rp_serial_port
{
  char *name;
  int fd;
};


struct s_rp_device_serial_settings
{
  uint32_t baud_rate;
  uint8_t stop_bits;
  uint8_t data_bits;
  char parity_bits;
};


enum e_rp_flash_verification
{
  rfv_simple = 0,   /*!< No checksum */
  rfv_block,        /*!< Checksum can appear at flash_block_size intervals throughout flash */
  rfv_all           /*!< Checksum always located at the end of flash */
};


struct s_rp_device_settings
{
  struct s_rp_device_serial_settings connect_settings;
  struct s_rp_device_serial_settings program_settings;

  char *boot_connect_string;
  char *user_connect_string;
  char *connect_response;

  uint32_t flash_start;
  uint32_t flash_length;
  uint16_t flash_checksum_seed;
  uint32_t flash_block_size;
  enum e_rp_flash_verification flash_verification;
  bool flash_encrypted;
};


#define RP_DEVICE_ID_NAME_LEN 20


#pragma pack(push, 1)
/** Packed device id structure sent to/from device */
struct s_rp_device_id_packed
{
  uint16_t checksum_seed;
  uint32_t start;
  uint32_t length;
  char name[RP_DEVICE_ID_NAME_LEN];
};
#pragma pack(pop)


/** Unpacked device id structure used within library */
struct s_rp_device_id
{
  uint16_t checksum_seed;
  uint32_t start;
  uint32_t length;
  char name[RP_DEVICE_ID_NAME_LEN + 1];
};


/** Device structure */
struct s_rp_device
{
  struct s_rp_serial_port *serial_port;
  struct s_rp_device_settings *device_settings;
  struct s_rp_device_id device_id;
};


/* From rp_error.c */


/** Convert a rinprog result code into a string */
const char *rp_result_to_string(enum e_rp_result result);


/* From rp_device.c */


enum e_rp_result rp_device_reboot(struct s_rp_device *rp_device);


enum e_rp_result rp_device_identify(struct s_rp_device *rp_device, struct s_rp_device_id *device_id);


enum e_rp_result rp_device_check_running(struct s_rp_device *rp_device);


enum e_rp_result rp_device_erase(struct s_rp_device *rp_device, uint32_t start_addr, uint32_t length);


enum e_rp_result rp_device_program(struct s_rp_device *rp_device, uint32_t start_addr, uint32_t length, FILE *f, uint32_t *fail_addr);


enum e_rp_result rp_device_verify(struct s_rp_device *rp_device, uint32_t start_addr, uint32_t length, FILE *f, uint32_t *fail_addr);


enum e_rp_result rp_device_connect_bootmode(struct s_rp_device *rp_device);


enum e_rp_result rp_device_connect_usermode(struct s_rp_device *rp_device);


/** Verify device id retrieved from device against device settings */
enum e_rp_result rp_verify_device_settings(const struct s_rp_device *rp_device, const struct s_rp_device_id *device_id);


/** Verify file image is suitable for programming into device */
enum e_rp_result rp_verify_image_for_device(const struct s_rp_device *rp_device, FILE *device_file);


enum e_rp_result rp_device_init(struct s_rp_device **rp_device, FILE *device_file, const char *serial_port);


enum e_rp_result rp_device_free(struct s_rp_device *rp_device);


/* From rp_checksum.c */


/** Read an s-record file from s_file_in, and write its binary equivalent to bin_file_out, padding unused space with pad */
enum e_rp_result rp_read_srecord_file(FILE *s_file_in, FILE *bin_file_out, const uint8_t pad);


/* From rp_util.c */


/** Get the length of a file. Changes file pointer, but resets it before returning */
enum e_rp_result rp_get_file_length(FILE *f, uint32_t *length);


/* From rp_devini.c */


/** Look for %xx in string, convert to character, copy all others over unchanged. Call must free returned string. */
char *rp_ini_connectstr_unescape(const char *connect_str);


#endif /* RINPROG_H_ */
