/*
 * rp_device.c
 *
 *  Created on: Aug 16, 2017
 *      Author: ian
 */
#define _XOPEN_SOURCE 700
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <stddef.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <sys/select.h>
#include <rinprog/rinprog.h>
#include "rp_serial.h"
#include "rp_devini.h"
#include "rp_util.h"
#include "rp_checksum.h"
#include "rp_teacrypt.h"

#define RP_COMMAND_CONNECT 0x44

#define RP_COMMAND_PROGRAM 0xB7D6F53C
#define RP_COMMAND_VERIFY 0x53E1A8D7
#define RP_COMMAND_ERASE 0xD768FEFA
#define RP_COMMAND_IDENTIFY 0x63DE3D96
#define RP_COMMAND_RESET 0xC33723C6

#define RP_RESPONSE_ERASE_FAILED 0x25
#define RP_RESPONSE_ERASE_COMPLETE 0x24

#define RP_COMMAND_WRITE_NEXTCHUNK 0x00
#define RP_COMMAND_WRITE_CANCEL 0xFF
#define RP_RESPONSE_WRITE_FAILURE 0x23
#define RP_RESPONSE_WRITE_SUCCESS 0x22
#define RP_RESPONSE_WRITE_REQUESTDATA 0x00

#define RP_BOOTCONNECT_RETRIES 100
#define RP_WRITE_CHUNK_SIZE 128


static const struct timespec rp_device_default_timeout = { .tv_sec = 0, .tv_nsec = 500000000 };

static const struct timespec rp_device_bootconnect_timeout = { .tv_sec = 0, .tv_nsec = 100000000 };
static const struct timespec rp_device_progverify_start_delay = { .tv_sec = 0, .tv_nsec = 100000000 };
static const struct timespec rp_device_postconnect_delay = { .tv_sec = 0, .tv_nsec = 500000000 };

/**
 * This delay permits the new baud rate byte to be transmitted,
 * before re-configuring our serial port.
 *
 * 1 byte (10bits) at 9600 baud ~ 1ms, so 100ms is heaps even if a
 * bootloader used 1200 baud (none to date do).
 */
static const struct timespec rp_device_rebaud_delay = { .tv_sec = 0, .tv_nsec = 100000000 };

static const struct timespec rp_device_erase_timeout = { .tv_sec = 5, .tv_nsec = 0 };
static const struct timespec rp_device_progverify_timeout = { .tv_sec = 5, .tv_nsec = 0 };

static const struct s_rp_tea_key rp_rinstrum_key = { { 0x5F713C19, 0xEEBA917B, 0x64B16401, 0x8BA93F71 } };
static const struct s_rp_tea_data rp_rinstrum_iv = { { 0x5E89CFB5, 0x9A31AE87 } };


/** Unpack the device id structure from the device into the unpacked one for library use */
static void rp_unpack_device_id(const struct s_rp_device_id_packed *packed, struct s_rp_device_id *unpacked)
{
  unpacked->checksum_seed = packed->checksum_seed;
  unpacked->start = packed->start;
  unpacked->length = packed->length;
  memcpy(unpacked->name, packed->name, RP_DEVICE_ID_NAME_LEN);
  unpacked->name[RP_DEVICE_ID_NAME_LEN] = '\0';
}


static uint8_t rp_scale_baud_rate_register(uint8_t brr, uint32_t oldbaud, uint32_t newbaud)
{
  uint32_t new_brr;

  new_brr = oldbaud * 1000 * brr;
  new_brr /= newbaud;
  new_brr /= 1000;
  return (uint8_t) new_brr;
}


static enum e_rp_result rp_device_send(struct s_rp_device *rp_device, const void *data, size_t length)
{
  return (write(rp_device->serial_port->fd, data, length) != (ssize_t) length) ? rpr_iofail : rpr_success;
}


static enum e_rp_result rp_device_receive(struct s_rp_device *rp_device, const struct timespec *timeout, void *data, size_t length)
{
  ssize_t read_len;
  ssize_t total_read;
  char *cdata;
  fd_set rfds;
  int res;

  for (cdata = data, read_len = 0, total_read = 0; total_read < (ssize_t) length; total_read += read_len)
  {
    FD_ZERO(&rfds);
    FD_SET(rp_device->serial_port->fd, &rfds);
    res = pselect(rp_device->serial_port->fd + 1, &rfds, NULL, NULL, timeout, NULL);
    if (res < 0)
    {
      if (errno == ETIMEDOUT)
        return rpr_timedout;
      else
        return rpr_iofail;
    }
    else if (res == 0)
    {
      return rpr_timedout;
    }
    else if (FD_ISSET(rp_device->serial_port->fd, &rfds))
    {
      read_len = read(rp_device->serial_port->fd, &cdata[total_read], length - total_read);
    }
  }
  /* printf("errno: %d, wanted %d, read %d\n", errno, (int) length, (int) read_len); */
  return (total_read != (ssize_t) length) ? rpr_iofail : rpr_success;
}


static enum e_rp_result rp_device_send_expect(struct s_rp_device *rp_device, const struct timespec *timeout,
                                              const void *senddata, size_t sendlen,
                                              void *recvdata, size_t recvlen)
{
  enum e_rp_result result;

  result = rp_device_send(rp_device, senddata, sendlen);
  if (rpr_success != result)
    return result;
  return rp_device_receive(rp_device, timeout, recvdata, recvlen);
}


enum e_rp_result rp_device_reboot(struct s_rp_device *rp_device)
{
  const uint32_t reboot_cmd = htonl(RP_COMMAND_RESET);

  /* Send reboot command */
  return rp_device_send(rp_device, &reboot_cmd, sizeof(reboot_cmd));
}


enum e_rp_result rp_device_identify(struct s_rp_device *rp_device, struct s_rp_device_id *device_id)
{
  const uint32_t id_cmd = htonl(RP_COMMAND_IDENTIFY);
  struct s_rp_device_id_packed packed;
  enum e_rp_result result;

  /* Send id command, expect 30byte packed id structure */
  result = rp_device_send_expect(rp_device, &rp_device_default_timeout, &id_cmd, sizeof(id_cmd), &packed, sizeof(packed));
  if (rpr_success != result)
    return result;

  /* To host endian */
  packed.checksum_seed = ntohs(packed.checksum_seed);
  packed.start = ntohl(packed.start);
  packed.length = ntohl(packed.length);

  /* Unpack structure into result */
  rp_unpack_device_id(&packed, device_id);
  return rpr_success;
}


enum e_rp_result rp_device_check_running(struct s_rp_device *rp_device)
{
  struct s_rp_device_id device_id;

  return rp_device_identify(rp_device, &device_id);
}


enum e_rp_result rp_device_erase(struct s_rp_device *rp_device, uint32_t start_addr, uint32_t length)
{
  const uint32_t erase_cmd[3] = { htonl(RP_COMMAND_ERASE), htonl(start_addr), htonl(length) };
  uint8_t response;
  enum e_rp_result result;

  /* Send erase command, start and length, expect a response byte */
  result = rp_device_send_expect(rp_device, &rp_device_erase_timeout, &erase_cmd, sizeof(erase_cmd), &response, sizeof(response));
  if (rpr_success != result)
    return result;

  switch(response)
  {
    case RP_RESPONSE_ERASE_COMPLETE:
      /* Erase completed successfully, now check that the device is still running */
      result = rp_device_check_running(rp_device);
      break;
    case RP_RESPONSE_ERASE_FAILED:
      result = rpr_flasherasefail;
      break;
    default:
      result = rpr_unexpectedresponse;
      break;
  }
  return result;
}


static enum e_rp_result rp_encrypt_chunk_if_required(const struct s_rp_device *rp_device, uint8_t data_chunk[RP_WRITE_CHUNK_SIZE], struct s_rp_tea_data *tea_iv)
{
  size_t i;
  struct s_rp_tea_data *tea_v;

  if (rp_device->device_settings->flash_encrypted)
  {
    for (i = 0; i < RP_WRITE_CHUNK_SIZE; i += sizeof(uint32_t) * 2)
    {
      tea_v = (struct s_rp_tea_data *) &data_chunk[i];
      tea_v->data[0] = htonl(tea_v->data[0]);
      tea_v->data[1] = htonl(tea_v->data[1]);
      rp_cbc_tea_encrypt(tea_v, tea_iv, &rp_rinstrum_key);
      tea_v->data[0] = ntohl(tea_v->data[0]);
      tea_v->data[1] = ntohl(tea_v->data[1]);
    }
  }
  return rpr_success;
}


static enum e_rp_result rp_device_prog_verify(struct s_rp_device *rp_device, uint32_t start_addr, uint32_t length, FILE *f, bool prog_not_verify, uint32_t *fail_addr)
{
  uint32_t pv_cmd[3];
  enum e_rp_result result;
  enum e_rp_result running_result;
  uint32_t file_length;
  uint32_t posn;
  uint8_t data_chunk[RP_WRITE_CHUNK_SIZE];
  uint8_t response;
  uint8_t cmd;
  const uint32_t end = start_addr + length;
  struct s_rp_tea_data tea_iv;

  if (NULL == rp_device || NULL == f || NULL == fail_addr)
    return rpr_paramfail;
  *fail_addr = 0;

  /* Run some checks */
  result = rp_get_file_length(f, &file_length);
  if (rpr_success != result)
    return result;
  if (end > file_length)
    return rpr_sizefail;
  if ((length % RP_WRITE_CHUNK_SIZE) != 0)
    return rpr_sizefail;

  /* Setup command */
  pv_cmd[0] = htonl((prog_not_verify) ? RP_COMMAND_PROGRAM : RP_COMMAND_VERIFY);
  pv_cmd[1] = htonl(start_addr);
  pv_cmd[2] = htonl(length);

  /* Send program or verify command, start and length */
  result = rp_device_send(rp_device, &pv_cmd, sizeof(pv_cmd));
  if (rpr_success != result)
    return result;

  /* Do a short delay before beginning programming */
  nanosleep(&rp_device_progverify_start_delay, NULL);

  /* Program/verify the data */
  tea_iv = rp_rinstrum_iv;
  if (fseek(f, start_addr, SEEK_SET) != 0)
    return rpr_iofail;
  for (posn = start_addr, result = rpr_unknown_error; posn < end; posn += RP_WRITE_CHUNK_SIZE)
  {
    /* Read a chunk from the file */
    if (fread(data_chunk, RP_WRITE_CHUNK_SIZE, 1, f) != 1)
    {
      *fail_addr = posn;
      result = rpr_iofail;
      return result;
    }

    /* Encrypt the data if required */
    result = rp_encrypt_chunk_if_required(rp_device, data_chunk, &tea_iv);
    if (rpr_success != result)
    {
      *fail_addr = posn;
      return result;
    }

    /* Send a chunk of data, expect a response byte */
    result = rp_device_send_expect(rp_device, &rp_device_progverify_timeout, data_chunk, RP_WRITE_CHUNK_SIZE, &response, sizeof(response));
    if (rpr_success != result)
    {
      *fail_addr = posn;
      return result;
    }

    /* Handle the response */
    switch (response)
    {
      case RP_RESPONSE_WRITE_FAILURE:
        /* Failed, stop programming */
        *fail_addr = posn;
        result = (prog_not_verify) ? rpr_flashwritefail : rpr_flashverifyfail;
        goto rdpv_finish;
        break;
      case RP_RESPONSE_WRITE_REQUESTDATA:
        /* Device is requesting more data, keep looping */
        break;
      default:
        /* Unknown response */
        *fail_addr = posn;
        result = rpr_unexpectedresponse;
        goto rdpv_finish;
        break;
    }

    /* Send the command for the next chunk */
    cmd = RP_COMMAND_WRITE_NEXTCHUNK;
    result = rp_device_send(rp_device, &cmd, sizeof(cmd));
    if (rpr_success != result)
    {
      *fail_addr = posn;
      return result;
    }
  }

  /* If everything has gone well, we can expect a RP_RESPONSE_WRITE_SUCCESS */
  result = rp_device_receive(rp_device, &rp_device_progverify_timeout, &response, sizeof(response));
  if (rpr_success != result)
  {
    *fail_addr = posn;
  }
  else
  {
    if (RP_RESPONSE_WRITE_SUCCESS != response)
    {
      *fail_addr = posn;
      result = rpr_unexpectedresponse;
    }
    else
    {
      result = rpr_success;
    }
  }

rdpv_finish:
  /* Finish up, check we are still running, but return failure of above code */
  running_result = rp_device_check_running(rp_device);
  if (rpr_success == result)
    result = running_result;
  return result;
}


enum e_rp_result rp_device_program(struct s_rp_device *rp_device, uint32_t start_addr, uint32_t length, FILE *f, uint32_t *fail_addr)
{
  return rp_device_prog_verify(rp_device, start_addr, length, f, true, fail_addr);
}


enum e_rp_result rp_device_verify(struct s_rp_device *rp_device, uint32_t start_addr, uint32_t length, FILE *f, uint32_t *fail_addr)
{
  return rp_device_prog_verify(rp_device, start_addr, length, f, false, fail_addr);
}


enum e_rp_result rp_device_negotiate_baud(struct s_rp_device *rp_device)
{
  #define ACK_LENGTH 5
  const char send_ack_data[ACK_LENGTH] = "ABCDE";
  char recv_ack_data[ACK_LENGTH];
  const char connect_data[2] = { RP_COMMAND_CONNECT, ACK_LENGTH };
  uint8_t brr;
  uint8_t newbrr;
  enum e_rp_result result;

  /* Use connect baud, and programming bit settings */
  result = rp_serial_set_config(rp_device->serial_port,
                                rp_device->device_settings->connect_settings.baud_rate,
                                rp_device->device_settings->program_settings.data_bits,
                                rp_device->device_settings->program_settings.parity_bits,
                                rp_device->device_settings->program_settings.stop_bits,
                                false);
  if (rpr_success != result)
    return result;

  /* Send connect command and ack data length, expect baud rate register */
  result = rp_device_send_expect(rp_device, &rp_device_default_timeout, connect_data, sizeof(connect_data), &brr, sizeof(brr));
  if (rpr_success != result)
    return result;

  /* Scale baud rate register, and send */
  newbrr = rp_scale_baud_rate_register(brr,
                                       rp_device->device_settings->connect_settings.baud_rate,
                                       rp_device->device_settings->program_settings.baud_rate);
  result = rp_device_send(rp_device, &newbrr, sizeof(newbrr));
  if (rpr_success != result)
    return result;

  /* Delay while the character is sent, so that we don't re-configure whilst the character is still being transmitted */
  nanosleep(&rp_device_rebaud_delay, NULL);

  /* Now use programming serial settings */
  result = rp_serial_set_config(rp_device->serial_port,
                                rp_device->device_settings->program_settings.baud_rate,
                                rp_device->device_settings->program_settings.data_bits,
                                rp_device->device_settings->program_settings.parity_bits,
                                rp_device->device_settings->program_settings.stop_bits,
                                false);
  if (rpr_success != result)
    return result;

  /* Send the ack data, and expect it back */
  result = rp_device_send_expect(rp_device, &rp_device_default_timeout, send_ack_data, sizeof(send_ack_data), recv_ack_data, sizeof(recv_ack_data));
  if (rpr_success != result)
    return result;

  /* Does it match */
  return (memcmp(send_ack_data, recv_ack_data, sizeof(send_ack_data)) == 0) ? rpr_success : rpr_negotiatefail;
}


/** After the initial connection to the bootloader is made, this is used to establish a higher working baudrate */
static enum e_rp_result rp_device_post_connect(struct s_rp_device *rp_device)
{
  enum e_rp_result result;

  /* Do post connect delay */
  nanosleep(&rp_device_postconnect_delay, NULL);

  /* Negotiate a (potentially) higher baudrate for programming */
  result = rp_device_negotiate_baud(rp_device);
  if (rpr_success != result)
    return result;

  /* Send id command to check device bootloader is running */
  result = rp_device_identify(rp_device, &rp_device->device_id);
  if (rpr_success != result)
    return result;

  /* Connected, success */
  return rpr_success;
}


enum e_rp_result rp_device_connect_bootmode(struct s_rp_device *rp_device)
{
  size_t i;
  enum e_rp_result result;
  char *reply = NULL;
  bool unexpected;

  /* Use connect serial settings */
  result = rp_serial_set_config(rp_device->serial_port,
                                rp_device->device_settings->connect_settings.baud_rate,
                                rp_device->device_settings->connect_settings.data_bits,
                                rp_device->device_settings->connect_settings.parity_bits,
                                rp_device->device_settings->connect_settings.stop_bits,
                                false);
  if (rpr_success != result)
    goto rdcb_fail;

  /* Allocate buffer to hold response */
  reply = malloc(strlen(rp_device->device_settings->connect_response));
  if (NULL == reply)
  {
    result = rpr_allocfail;
    goto rdcb_fail;
  }

  /* Try to connect */
  for (i = 0, unexpected = false; i < RP_BOOTCONNECT_RETRIES; i++)
  {
    /* Send connect string, expect response */
    result = rp_device_send_expect(rp_device,
                                   &rp_device_bootconnect_timeout,
                                   rp_device->device_settings->boot_connect_string,
                                   strlen(rp_device->device_settings->boot_connect_string),
                                   reply,
                                   strlen(rp_device->device_settings->connect_response));
    if (rpr_timedout == result)
      continue;
    else if (rpr_success != result)
      goto rdcb_fail;

    /* Check response */
    if (memcmp(rp_device->device_settings->connect_response, reply, strlen(rp_device->device_settings->connect_response)) == 0)
    {
      /* Do post connection steps */
      result = rp_device_post_connect(rp_device);
      goto rdcb_fail;
    }
    else
    {
      unexpected = true;
    }
  }

  /* Failed to connect after all retries */
  result = (unexpected) ? rpr_unexpectedresponse : rpr_connectfail;

rdcb_fail:
  if (NULL != reply)
    free(reply);
  return result;
}


enum e_rp_result rp_device_connect_usermode(struct s_rp_device *rp_device)
{
  char *reply = NULL;
  enum e_rp_result result;

  /* Use connect serial settings */
  result = rp_serial_set_config(rp_device->serial_port,
                                rp_device->device_settings->connect_settings.baud_rate,
                                rp_device->device_settings->connect_settings.data_bits,
                                rp_device->device_settings->connect_settings.parity_bits,
                                rp_device->device_settings->connect_settings.stop_bits,
                                false);
  if (rpr_success != result)
    goto rdcu_fail;

  /* Allocate buffer to hold response */
  reply = malloc(strlen(rp_device->device_settings->connect_response));
  if (NULL == reply)
  {
    result = rpr_allocfail;
    goto rdcu_fail;
  }

  /* Send connect string, expect response */
  result = rp_device_send_expect(rp_device,
                                 &rp_device_default_timeout,
                                 rp_device->device_settings->user_connect_string,
                                 strlen(rp_device->device_settings->user_connect_string),
                                 reply,
                                 strlen(rp_device->device_settings->connect_response));
  if (rpr_success != result)
    goto rdcu_fail;

  /* Check response */
  if (memcmp(rp_device->device_settings->connect_response, reply, strlen(rp_device->device_settings->connect_response)) != 0)
  {
    result = rpr_connectfail;
    goto rdcu_fail;
  }

  /* Do post connection steps */
  result = rp_device_post_connect(rp_device);
  if (rpr_success != result)
    goto rdcu_fail;

  /* Connected, success */
  result = rpr_success;

rdcu_fail:
  if (NULL != reply)
    free(reply);
  return result;
}


/** Verify device id retrieved from device against device settings */
enum e_rp_result rp_verify_device_settings(const struct s_rp_device *rp_device, const struct s_rp_device_id *device_id)
{
  if (NULL == rp_device || NULL == device_id)
    return rpr_paramfail;

  if (rp_device->device_settings->flash_checksum_seed != device_id->checksum_seed)
    return rpr_chkseed_mismatch;
  if (rp_device->device_settings->flash_start != device_id->start)
    return rpr_startaddr_mismatch;
  if (rp_device->device_settings->flash_length != device_id->length)
    return rpr_length_mismatch;

  return rpr_success;
}


/** Verify file image is suitable for programming into device */
enum e_rp_result rp_verify_image_for_device(const struct s_rp_device *rp_device, FILE *device_file)
{
  enum e_rp_result result;
  uint32_t length;

  if (NULL == rp_device || NULL == device_file)
    return rpr_paramfail;

  /* Check that file will fit in device */
  result = rp_get_file_length(device_file, &length);
  if (rpr_success != result)
    return result;
  if (length > rp_device->device_settings->flash_start + rp_device->device_settings->flash_length)
    return rpr_sizefail;

  /* Verify file against  */
  switch (rp_device->device_settings->flash_verification)
  {
    case rfv_simple:
      /* Above size check is all we have */
      break;
    case rfv_block:
      result = rp_verify_buffer_block_checksum(device_file,
                                               rp_device->device_settings->flash_checksum_seed,
                                               rp_device->device_settings->flash_start,
                                               rp_device->device_settings->flash_block_size);
      break;
    case rfv_all:
      result = rp_verify_buffer_all_checksum(device_file,
                                             rp_device->device_settings->flash_checksum_seed,
                                             rp_device->device_settings->flash_start);
      break;
    default:
      result = rpr_paramfail;
      break;
  }
  return result;
}


enum e_rp_result rp_device_init(struct s_rp_device **rp_device, FILE *device_file, const char *serial_port)
{
  struct s_rp_device *dev = NULL;
  enum e_rp_result result;

  if (NULL == rp_device || NULL == serial_port)
    return rpr_paramfail;

  /* Allocate structure */
  dev = calloc(1, sizeof(*dev));
  if (NULL == dev)
  {
    result = rpr_allocfail;
    goto rdi_fail;
  }

  /* Read in device file, or set defaults */
  if (NULL != device_file)
  {
    result = rp_device_settings_read(device_file, &dev->device_settings);
  }
  else
  {
    result = rp_device_settings_default(&dev->device_settings);
  }
  if (rpr_success != result)
  {
    goto rdi_fail;
  }

  /* Create & open serial port */
  result = rp_serial_init(&dev->serial_port, serial_port);
  if (rpr_success != result)
  {
    goto rdi_fail;
  }
  result = rp_serial_open(dev->serial_port);
  if (rpr_success != result)
  {
    goto rdi_fail;
  }

  *rp_device = dev;
  return rpr_success;

rdi_fail:
  if (NULL != dev)
  {
    if (NULL != dev->device_settings)
    {
      (void) rp_device_settings_free(dev->device_settings);
      dev->device_settings = NULL;
    }
    if (NULL != dev->serial_port)
    {
      (void) rp_serial_shutdown(dev->serial_port);
      dev->serial_port = NULL;
    }
    free(dev);
  }
  return result;
}


enum e_rp_result rp_device_free(struct s_rp_device *rp_device)
{
  if (NULL != rp_device)
  {
    if (NULL != rp_device->device_settings)
    {
      (void) rp_device_settings_free(rp_device->device_settings);
      rp_device->device_settings = NULL;
    }
    if (NULL != rp_device->serial_port)
    {
      (void) rp_serial_shutdown(rp_device->serial_port);
      rp_device->serial_port = NULL;
    }
    free(rp_device);
  }
  return rpr_success;
}
