/*
 * rp_devini.c
 *
 *  Created on: Aug 15, 2017
 *      Author: ian
 */
#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <iniparser.h>

#include "rp_devini.h"


/* Ini file sections */
#define RP_INI_SECTION_COMCONNECT "ComConnect"
#define RP_INI_SECTION_COMPROGRAM "ComProgram"
#define RP_INI_SECTION_FLASH "Flash"

/* Ini file serial settings */
#define RP_INI_SETTING_BAUD "BaudRate"
#define RP_INI_SETTING_BAUD_DEFAULT 9600
#define RP_INI_SETTING_STOPBITS "StopBits"
#define RP_INI_SETTING_STOPBITS_DEFAULT 1
#define RP_INI_SETTING_DATABITS "DataBits"
#define RP_INI_SETTING_DATABITS_DEFAULT 8
#define RP_INI_SETTING_PARITY "Parity"
#define RP_INI_SETTING_PARITY_DEFAULT "None"

/* Ini file connect settings */
#define RP_INI_SETTING_BOOTCONNECT "BootConnect"
#define RP_INI_SETTING_USERCONNECT "UserConnect"
#define RP_INI_SETTING_CONNECTRESP "ConnectResponse"

/* Ini file flash dimension settings */
#define RP_INI_SETTING_FLASHSTART "FlashStart"
#define RP_INI_SETTING_FLASHLENGTH "FlashLength"
#define RP_INI_SETTING_FLASHCHECKSEED "FlashCheckSeed"
#define RP_INI_SETTING_FLASHBLOCKSIZE "FlashBlockSize"
#define RP_INI_SETTING_FLASHVERIFY "FlashVerification"
#define RP_INI_SETTING_FLASHENCRYPT "FlashEncryption"


static char rp_ini_parity_to_char(const char *parity_str)
{
  char parity;

  if (strcmp(parity_str, "None") == 0)
    parity = 'n';
  else if (strcmp(parity_str, "Odd") == 0)
    parity = 'o';
  if (strcmp(parity_str, "Even") == 0)
    parity = 'e';
  else
    parity = 'n';

  return parity;
}


static bool rp_ishexchar(char ch)
{
  return (ch >= '0' && ch <= '9') ||
         (ch >= 'a' && ch <= 'f') ||
         (ch >= 'A' && ch <= 'F');
}


static char rp_hextonibble(char ch)
{
  if (ch >= '0' && ch <= '9')
    return ch - '0';
  else if (ch >= 'a' && ch <= 'f')
    return ch - 'a' + 10;
  else if (ch >= 'A' && ch <= 'F')
    return ch - 'A' + 10;
  return 0;
}


static char rp_hextochar(const char *hexstr)
{
  return ((rp_hextonibble(hexstr[0]) << 4) & 0xF0) | (rp_hextonibble(hexstr[1]) & 0x0F);
}


/** Look for %xx in string, convert to character, copy all others over unchanged. Call must free returned string. */
char *rp_ini_connectstr_unescape(const char *connect_str)
{
  char *unesc_str;
  char *out;
  const char *in;

  if (NULL == connect_str)
    return NULL;

  unesc_str = calloc(1, strlen(connect_str) + 1);
  if (NULL == unesc_str)
    return NULL;

  for (in = connect_str, out = unesc_str; *in != '\0'; )
  {
    switch (*in)
    {
      case '%':
        if (in[1] == '%')
        {
          /* %% -> % */
          *out++ = '%';
          in += 2;
        }
        else if (rp_ishexchar(in[1]) && rp_ishexchar(in[2]))
        {
          /* %xx -> char(xx) */
          *out++ = rp_hextochar(&in[1]);
          in += 3;

        }
        else
        {
          *out++ = *in++;
        }
        break;
      default:
        *out++ = *in++;
        break;
    }

  }
  return unesc_str;
}


enum e_rp_result rp_device_settings_read(FILE *in, struct s_rp_device_settings **settings)
{
  dictionary *ini = NULL;
  enum e_rp_result res;
  struct s_rp_device_settings *devset;

  if (NULL == in || NULL == settings)
    return rpr_paramfail;
  *settings = NULL;

  devset = calloc(1, sizeof(struct s_rp_device_settings));
  if (NULL == devset)
  {
    res = rpr_allocfail;
    goto rdsr_fail;
  }

  ini = iniparser_load_file(in);
  if (NULL == ini)
  {
    res = rpr_iofail;
    goto rdsr_fail;
  }

  /* Get connect settings */
  devset->connect_settings.baud_rate = iniparser_getint(ini, RP_INI_SECTION_COMCONNECT ":" RP_INI_SETTING_BAUD, RP_INI_SETTING_BAUD_DEFAULT);
  devset->connect_settings.stop_bits = iniparser_getint(ini, RP_INI_SECTION_COMCONNECT ":" RP_INI_SETTING_STOPBITS, RP_INI_SETTING_STOPBITS_DEFAULT);
  devset->connect_settings.data_bits = iniparser_getint(ini, RP_INI_SECTION_COMCONNECT ":" RP_INI_SETTING_DATABITS, RP_INI_SETTING_DATABITS_DEFAULT);
  devset->connect_settings.parity_bits = rp_ini_parity_to_char(iniparser_getstring(ini, RP_INI_SECTION_COMCONNECT ":" RP_INI_SETTING_PARITY, RP_INI_SETTING_PARITY_DEFAULT));

  /* Get program settings */
  devset->program_settings.baud_rate = iniparser_getint(ini, RP_INI_SECTION_COMPROGRAM ":" RP_INI_SETTING_BAUD, RP_INI_SETTING_BAUD_DEFAULT);
  devset->program_settings.stop_bits = iniparser_getint(ini, RP_INI_SECTION_COMPROGRAM ":" RP_INI_SETTING_STOPBITS, RP_INI_SETTING_STOPBITS_DEFAULT);
  devset->program_settings.data_bits = iniparser_getint(ini, RP_INI_SECTION_COMPROGRAM ":" RP_INI_SETTING_DATABITS, RP_INI_SETTING_DATABITS_DEFAULT);
  devset->program_settings.parity_bits = rp_ini_parity_to_char(iniparser_getstring(ini, RP_INI_SECTION_COMPROGRAM ":" RP_INI_SETTING_PARITY, RP_INI_SETTING_PARITY_DEFAULT));

  /* Get connect strings */
  devset->boot_connect_string = rp_ini_connectstr_unescape(iniparser_getstring(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_BOOTCONNECT, ""));
  if (NULL == devset->boot_connect_string)
  {
    res = rpr_allocfail;
    goto rdsr_fail;
  }
  devset->user_connect_string = rp_ini_connectstr_unescape(iniparser_getstring(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_USERCONNECT, ""));
  if (NULL == devset->user_connect_string)
  {
    res = rpr_allocfail;
    goto rdsr_fail;
  }
  devset->connect_response = rp_ini_connectstr_unescape(iniparser_getstring(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_CONNECTRESP, ""));
  if (NULL == devset->connect_response)
  {
    res = rpr_allocfail;
    goto rdsr_fail;
  }

  /* Get flash dimensions */
  devset->flash_start = iniparser_getint(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_FLASHSTART, 0);
  devset->flash_length = iniparser_getint(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_FLASHLENGTH, 0);
  devset->flash_checksum_seed = iniparser_getint(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_FLASHCHECKSEED, 0);
  devset->flash_block_size = iniparser_getint(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_FLASHBLOCKSIZE, 0);
  devset->flash_verification = (enum e_rp_flash_verification) iniparser_getint(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_FLASHVERIFY, (int) rfv_simple);
  devset->flash_encrypted = (iniparser_getint(ini, RP_INI_SECTION_FLASH ":" RP_INI_SETTING_FLASHENCRYPT, (int) true) != 0);

  /* Clean up */
  iniparser_freedict(ini);
  ini = NULL;
  *settings = devset;
  return rpr_success;

rdsr_fail:
  if (NULL != ini)
  {
    iniparser_freedict(ini);
    ini = NULL;
  }

  if (NULL != devset)
  {
    if (NULL != devset->boot_connect_string)
    {
      free(devset->boot_connect_string);
      devset->boot_connect_string = NULL;
    }
    if (NULL != devset->user_connect_string)
    {
      free(devset->user_connect_string);
      devset->user_connect_string = NULL;
    }
    if (NULL != devset->connect_response)
    {
      free(devset->connect_response);
      devset->connect_response = NULL;
    }
    free(devset);
    devset = NULL;
  }
  return res;
}


enum e_rp_result rp_device_settings_default(struct s_rp_device_settings **settings)
{
  enum e_rp_result res;
  struct s_rp_device_settings *devset;

  if (NULL == settings)
    return rpr_paramfail;
  *settings = NULL;

  devset = calloc(1, sizeof(struct s_rp_device_settings));
  if (NULL == devset)
  {
    res = rpr_allocfail;
    goto rdsr_fail;
  }

  /* Get connect settings */
  devset->connect_settings.baud_rate = RP_INI_SETTING_BAUD_DEFAULT;
  devset->connect_settings.stop_bits = RP_INI_SETTING_STOPBITS_DEFAULT;
  devset->connect_settings.data_bits = RP_INI_SETTING_DATABITS_DEFAULT;
  devset->connect_settings.parity_bits = rp_ini_parity_to_char(RP_INI_SETTING_PARITY_DEFAULT);

  /* Get program settings */
  devset->program_settings.baud_rate = RP_INI_SETTING_BAUD_DEFAULT;
  devset->program_settings.stop_bits = RP_INI_SETTING_STOPBITS_DEFAULT;
  devset->program_settings.data_bits = RP_INI_SETTING_DATABITS_DEFAULT;
  devset->program_settings.parity_bits = rp_ini_parity_to_char(RP_INI_SETTING_PARITY_DEFAULT);

  /* Get connect strings */
  devset->boot_connect_string = rp_ini_connectstr_unescape("FlAsHRuN");
  if (NULL == devset->boot_connect_string)
  {
    res = rpr_allocfail;
    goto rdsr_fail;
  }
  devset->user_connect_string = rp_ini_connectstr_unescape("FlAsHRuN");
  if (NULL == devset->user_connect_string)
  {
    res = rpr_allocfail;
    goto rdsr_fail;
  }
  devset->connect_response = rp_ini_connectstr_unescape("FlAsHRuN");
  if (NULL == devset->connect_response)
  {
    res = rpr_allocfail;
    goto rdsr_fail;
  }

  /* Get flash dimensions */
  devset->flash_start = 0;
  devset->flash_length = 0;
  devset->flash_checksum_seed = 0;
  devset->flash_block_size = 0;
  devset->flash_verification = rfv_simple;
  devset->flash_encrypted = true;

  /* Clean up */
  *settings = devset;
  return rpr_success;

rdsr_fail:
  if (NULL != devset)
  {
    if (NULL != devset->boot_connect_string)
    {
      free(devset->boot_connect_string);
      devset->boot_connect_string = NULL;
    }
    if (NULL != devset->user_connect_string)
    {
      free(devset->user_connect_string);
      devset->user_connect_string = NULL;
    }
    if (NULL != devset->connect_response)
    {
      free(devset->connect_response);
      devset->connect_response = NULL;
    }
    free(devset);
    devset = NULL;
  }
  return res;
}


enum e_rp_result rp_device_settings_free(struct s_rp_device_settings *settings)
{
  free(settings->boot_connect_string);
  settings->boot_connect_string = NULL;
  free(settings->user_connect_string);
  settings->user_connect_string = NULL;
  free(settings->connect_response);
  settings->connect_response = NULL;

  free(settings);
  settings = NULL;

  return rpr_success;
}
