/*
 * rp_serial.c
 *
 *  Created on: Aug 15, 2017
 *      Author: ian
 */
#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include <dirent.h>
#include <regex.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <ctype.h>

#include "rp_serial.h"


/** Initialise the serial port structure with the port name. Doesn't open port. */
enum e_rp_result rp_serial_init(struct s_rp_serial_port **serial_port, const char *port_name)
{
  enum e_rp_result res;
  struct s_rp_serial_port *sp;

  if (NULL == serial_port || NULL == port_name)
    return rpr_paramfail;

  sp = calloc(1, sizeof(*sp));
  if (NULL == sp)
  {
    res = rpr_allocfail;
    goto si_fail;
  }
  sp->fd = -1;

  sp->name = strdup(port_name);
  if (NULL == sp->name)
  {
    res = rpr_allocfail;
    goto si_fail;
  }

  *serial_port = sp;
  return rpr_success;

si_fail:
  if (NULL != sp)
  {
    if (NULL != sp->name)
    {
      free(sp->name);
      sp->name = NULL;
    }
    free(sp);
  }
  return res;
}


/** Close port if open and clean up allocated memory in struct */
enum e_rp_result rp_serial_shutdown(struct s_rp_serial_port *serial_port)
{
  if (NULL != serial_port)
  {
    rp_serial_close(serial_port);
    if (NULL != serial_port->name)
    {
      free(serial_port->name);
      serial_port->name = NULL;
    }
    free(serial_port);
  }
  return rpr_success;
}


/** Open the serial port */
enum e_rp_result rp_serial_open(struct s_rp_serial_port *serial_port)
{
  int flags;
  enum e_rp_result result;

  /* Use O_NDELAY to open the port without blocking for CTS? or something */
  serial_port->fd = open(serial_port->name, O_RDWR | O_NOCTTY | O_NDELAY);
  if (serial_port->fd < 0)
  {
    result = rpr_iofail;
    goto lso_done;
  }

  /* Now setup blocking reads and writes */
  flags = fcntl(serial_port->fd, F_GETFL);
  if (flags < 0)
  {
    result = rpr_iofail;
    goto lso_done;
  }
  flags &= ~O_NONBLOCK;
  if (fcntl(serial_port->fd, F_SETFL, flags) < 0)
  {
    result = rpr_iofail;
    goto lso_done;
  }

  /* Success */
  result = rpr_success;

lso_done:
  return result;
}


/** Close the serial port */
enum e_rp_result rp_serial_close(struct s_rp_serial_port *serial_port)
{
  if (serial_port->fd >= 0)
  {
    close(serial_port->fd);
    serial_port->fd = -1;
  }
  return rpr_success;
}


struct s_baud_conversion
{
  speed_t linux_baud;
  uint32_t actual_baud;
};

/* #define NUM_BAUD_CONVERSIONS 30 */
#define NUM_BAUD_CONVERSIONS 17
static const struct s_baud_conversion baud_conversions[NUM_BAUD_CONVERSIONS] =
{
  { B50, 50 },
  { B75, 75 },
  { B110, 110 },
  { B134, 134 },
  { B150, 150 },
  { B200, 200 },
  { B300, 300 },
  { B600, 600 },
  { B1200, 1200 },
  { B1800, 1800 },
  { B2400, 2400 },
  { B4800, 4800 },
  { B9600, 9600 },
  { B19200, 19200 },
  { B38400, 38400 },
  { B57600, 57600 },
  { B115200, 115200 },
#if 0 /* These are #if'd out in uCLibc, or at least the version we have. We don't need them so far, so I've just commented them out for now. */
  { B230400, 230400 },
  { B460800, 460800 },
  { B500000, 500000 },
  { B576000, 576000 },
  { B921600, 921600 },
  { B1000000, 1000000 },
  { B1152000, 1152000 },
  { B1500000, 1500000 },
  { B2000000, 2000000 },
  { B2500000, 2500000 },
  { B3000000, 3000000 },
  { B3500000, 3500000 },
  { B4000000, 4000000 }
#endif
};


static speed_t rp_lookup_linux_baud_rate(uint32_t baud)
{
  int i;
  for (i = 0; i < NUM_BAUD_CONVERSIONS; i++)
  {
    if (baud_conversions[i].actual_baud == baud)
    {
      return baud_conversions[i].linux_baud;
    }
  }
  return B9600;
}


/** Set the configuration for the serial port */
enum e_rp_result rp_serial_set_config(struct s_rp_serial_port *serial_port, uint32_t baud_rate,
                                      uint8_t data_bits, char parity, uint8_t stop_bits, bool use_cts)
{
  struct termios options;
  speed_t linuxbaud;

  /* Zero out settings */
  memset(&options, 0, sizeof(options));

  /* Set the baud rates */
  linuxbaud = rp_lookup_linux_baud_rate(baud_rate);
  cfsetispeed(&options, linuxbaud);
  cfsetospeed(&options, linuxbaud);

  /* Set the data bits */
  options.c_cflag &= ~CSIZE;
  if (5 == data_bits)
    options.c_cflag |= CS5;
  else if (6 == data_bits)
    options.c_cflag |= CS6;
  else if (7 == data_bits)
    options.c_cflag |= CS7;
  else /* if (8 == data_bits) */
    options.c_cflag |= CS8;

  /* Set the stop bits */
  if (1 != stop_bits)
    options.c_cflag |= CSTOPB;  /* 2 Stop bits */
  else
    options.c_cflag &= ~CSTOPB; /* 1 Stop bits */

  /* Set the parity bits */
  if ('o' == parity)  /* Odd */
  {
    options.c_cflag |= PARENB;
    options.c_cflag |= PARODD;
  }
  else if ('e' == parity) /* Even */
  {
    options.c_cflag |= PARENB;
    options.c_cflag &= ~PARODD;
  }
  else /* No parity 'n' */
  {
    options.c_cflag &= ~PARENB;
    options.c_cflag &= ~PARODD;
  }

  /* Enable the receiver and set local mode */
  options.c_cflag |= (CLOCAL | CREAD);

  /* Ignore parity ??? */
  options.c_iflag = IGNPAR;
  options.c_oflag = 0;

  options.c_cc[VTIME] = 0; /* No Time delay */
  options.c_cc[VMIN] = 1; /* Blocking read until 1 character arrives */

  /* CTS handshaking */
  if (use_cts)
  {
    options.c_cflag |= CRTSCTS;
  }

  /* Set the new options for the port */
  return (tcsetattr(serial_port->fd, TCSANOW, &options) < 0) ? rpr_iofail : rpr_success;
}
