#!/usr/bin/python3.5
"""
bl_simulator.py

This program simulates a rinstrum bootloader for the purposes of testing rinflash

You can either connect two physical serial ports together with a crossed null modem cable,
or use socat to create virtual serial ports.

Physical serial port example:
Terminal 1:
  python bl_simulator.py --connect-mode=user /dev/ttyUSB0 R320.dev r320.flash
Terminal 2:
  ./rinflash /dev/ttyUSB1 R320.dev program r320.mot

socat example:
Setup:
  sudo apt install socat
Terminal 1:
  socat -d -d pty,raw,echo=0 pty,raw,echo=0
Terminal 2:
  python bl_simulator.py --connect-mode=user /dev/pts/16 R320.dev r320.flash
Terminal 3:
  ./rinflash /dev/pts/17 R320.dev program r320.mot

"""

import argparse
import configparser
import serial
import socket
import struct
import os
import re
import errno
import tea_cbc

RP_COMMAND_CONNECT = 0x44

RP_COMMAND_PROGRAM = 0xB7D6F53C
RP_COMMAND_VERIFY = 0x53E1A8D7
RP_COMMAND_ERASE = 0xD768FEFA
RP_COMMAND_IDENTIFY = 0x63DE3D96
RP_COMMAND_RESET = 0xC33723C6

RP_RESPONSE_ERASE_FAILED = 0x25
RP_RESPONSE_ERASE_COMPLETE = 0x24

RP_COMMAND_WRITE_NEXTCHUNK = 0x00
RP_COMMAND_WRITE_CANCEL = 0xFF
RP_RESPONSE_WRITE_FAILURE = 0x23
RP_RESPONSE_WRITE_SUCCESS = 0x22
RP_RESPONSE_WRITE_REQUESTDATA = 0x00

RP_WRITE_CHUNK_SIZE = 128


class BLException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


def get_file_size(file):
    """Get the file size of file"""
    curposn = file.tell()

    file.seek(0, os.SEEK_END)
    size = file.tell()

    file.seek(curposn, os.SEEK_SET)
    return size

def erase_flash_file(flashfile, length=None):
    """Erase the flash file by writing all 0xFFs"""
    if (length is None):
        size = get_file_size(file=flashfile)
    else:
        size = length

    flashfile.seek(0, os.SEEK_SET)
    for i in range(size):
        flashfile.write('\xFF')

def establish_flash_file(filename, dev_settings):
    """Establish a file to simulate the flash"""

    # This has a race condition
    try:
        f = open(filename, "r+b")
    except IOError as e:
        if (e.errno == errno.ENOENT):
            f = open(filename, "w+b")
        else:
            raise

    f.seek(0, os.SEEK_END)
    size = f.tell()

    start = dev_settings['Flash'].getint('FlashStart', 0)
    length = dev_settings['Flash'].getint('FlashLength', 0)

    if (size == 0):
        print("New flash file, erasing to length %d" % (start + length))
        erase_flash_file(flashfile=f, length=start + length)
    elif (size != start + length):
        raise BLException("Length of flash file %d does not match device %d" % (size, start + length))
    else:
        print("Opened existing flash file of correct size %d" % (start + length))
    return f

def adopt_device_serial_settings(serport, dev_settings, section):
    serport.baudrate = dev_settings[section].getint('BaudRate', 9600)
    serport.bytesize = dev_settings[section].getint('DataBits', serial.EIGHTBITS)
    serport.stopbits = dev_settings[section].getint('StopBits', serial.STOPBITS_ONE)

def baud_rate_to_byte(baud):
    """Convert baud rate to a byte"""
    brbyte = chr(baud / 1200)
    return brbyte

def byte_to_baud_rate(data):
    """Convert baud rate byte into a baud rate"""
    baud = ord(data) * 1200
    return baud

def negotiate_baud_rate(serport):
    """ Negotiate programming baud rate with rinflash """

    print("Negotiating baud rate with rinflash")
    # Read connect byte and ack length
    data = serport.read(2)
    concmd = ord(data[0])
    if (concmd != RP_COMMAND_CONNECT):
        raise BLException("Connect byte was 0x%02x, expected 0x%02x" % (concmd, RP_COMMAND_CONNECT))
    ack_length = ord(data[1])

    # Send baud rate as a byte
    serport.write(baud_rate_to_byte(serport.baudrate))

    # Expect new baud rate back and convert to real baud rate
    newbrbyte = serport.read(1)
    newbaud = byte_to_baud_rate(data=newbrbyte[0])

    # Adopt new programming settings
    serport.baudrate = newbaud

    # Read in the data to acknowledge, and send back
    ack_data = serport.read(ack_length)
    serport.write(ack_data)

def serial_read_uint32(serport):
    """Read a 32 bit number from the serial port"""
    data = serport.read(4)
    uint32 = struct.unpack('!L', data)
    return uint32[0]

def command_identify(serport, dev_settings):
    """Run the identify command"""

    print("Sending device identity to host")
    chkseed = dev_settings['Flash'].getint('FlashCheckSeed', 0)
    start = dev_settings['Flash'].getint('FlashStart', 0)
    length = dev_settings['Flash'].getint('FlashLength', 0)
    name = 'blsim.py'

    data = struct.pack('!HLL20s', chkseed, start, length, name)
    serport.write(data)

def command_erase(serport, dev_settings, flashfile):
    """Run the erase command"""

    print("Erasing flash on device")
    start = serial_read_uint32(serport=serport)
    length = serial_read_uint32(serport=serport)

    if ((start != dev_settings['Flash'].getint('FlashStart', 0)) or
        (length != dev_settings['Flash'].getint('FlashLength', 0))):
        print("Host requested erase of invalid flash area")
        serport.write(chr(RP_RESPONSE_ERASE_FAILED))
    erase_flash_file(flashfile=flashfile)
    serport.write(chr(RP_RESPONSE_ERASE_COMPLETE))

def flash_verify_func(flashfile, data):
    flashdata = flashfile.read(RP_WRITE_CHUNK_SIZE)
    if (flashdata != data):
        raise BLException("Verify failed")

def flash_write_func(flashfile, data):
    flashfile.write(data)

def to_tea_block(data):
    block_list = []
    for i in range(0, len(data), 4):
        block = struct.unpack("!I", data[i:i + 4])[0]
        block_list.append(block)
    return block_list

def from_tea_block(block):
    data = ""
    for b in block:
        d = struct.pack("!I", b)
        data = data + d
    return data

def decrypt_buffer(encdata, cbciv):
    """Decrypt a data buffer received from the host"""

    # TEA based on data blocks of 8 bytes
    assert((len(encdata) % 8) == 0)

    block_key = [ 0x5F713C19, 0xEEBA917B, 0x64B16401, 0x8BA93F71 ]
    teac = tea_cbc.TinyEncryptionAlgorithmWithCBC()

    # Decrypt buffer
    decdata = ""
    for i in range(0, len(encdata), 8):
        buf = encdata[i:i + 8]
        ciphertext = to_tea_block(buf)

        plaintext, cbciv = teac.cbc_decrypt_block(ciphertext, cbciv, block_key)
        decdata = decdata + from_tea_block(plaintext)
    return decdata, cbciv

def command_program(serport, dev_settings, flashfile, write_verify_func, funcstr):
    """Run the write or verify flash command"""

    print("{} flash on device".format(funcstr))
    start = serial_read_uint32(serport=serport)
    length = serial_read_uint32(serport=serport)

    flashstart = dev_settings['Flash'].getint('FlashStart', 0)
    flashlength = dev_settings['Flash'].getint('FlashLength', 0)
    flashencrypted = (dev_settings['Flash'].getint('FlashEncryption', 1) != 0)

    if (start < flashstart or start + length > flashstart + flashlength):
        print("Host requested {} outside flash area".format(funcstr))
        data = serport.read(RP_WRITE_CHUNK_SIZE)
        reply = RP_RESPONSE_WRITE_FAILURE
        serport.write(reply)
        return

    cbciv = [ 0x5E89CFB5, 0x9A31AE87 ]

    for i in range(start, start + length, RP_WRITE_CHUNK_SIZE):
        encdata = serport.read(RP_WRITE_CHUNK_SIZE)
        if (len(encdata) == RP_WRITE_CHUNK_SIZE):
            try:
                if (flashencrypted):
                    decdata, cbciv = decrypt_buffer(encdata=encdata, cbciv=cbciv)
                else:
                    decdata = encdata
                flashfile.seek(i, os.SEEK_SET)
                flashfile.write(decdata)
                reply = RP_RESPONSE_WRITE_REQUESTDATA
            except:
                print("{} to file failed at address 0x{:x}".format(funcstr, i))
                reply = RP_RESPONSE_WRITE_FAILURE
        else:
            print("{} read serial data failed at address 0x{:x}".format(funcstr, i))
            reply = RP_RESPONSE_WRITE_FAILURE
        serport.write(chr(reply))

        cmd = serport.read(1)
        if (cmd != chr(RP_COMMAND_WRITE_NEXTCHUNK)):
            print("{} read serial command failed at address 0x{:x}".format(funcstr, i))
            return

    reply = RP_RESPONSE_WRITE_SUCCESS
    serport.write(chr(reply))

def main_loop(serport, dev_settings, flashfile):
    """Main loop to service commands from host"""

    print("Running main loop to service commands from rinflash")
    quit = False
    while (not quit):
        cmd = serial_read_uint32(serport=serport)
        if (cmd == RP_COMMAND_IDENTIFY):
            command_identify(serport=serport, dev_settings=dev_settings)
        elif (cmd == RP_COMMAND_RESET):
            print("Reset command received, shutting down")
            quit = True
        elif (cmd == RP_COMMAND_ERASE):
            command_erase(serport=serport, dev_settings=dev_settings, flashfile=flashfile)
        elif (cmd == RP_COMMAND_PROGRAM):
            command_program(serport=serport, dev_settings=dev_settings, flashfile=flashfile,
                            write_verify_func=flash_write_func, funcstr="Write")
        elif (cmd == RP_COMMAND_VERIFY):
            command_program(serport=serport, dev_settings=dev_settings, flashfile=flashfile,
                            write_verify_func=flash_verify_func, funcstr="Verify")
        else:
            print("Unknown command received 0x%08X" % (cmd))
    print("Main loop done, exiting")

def serial_wait_for(serport, str):
    """Wait for str to be received on the serial port, discarding all characters before it"""
    resp = ""
    while (resp != str):
        ch = serport.read(1)
        resp = resp + ch
        if (len(resp) > len(str)):
            resp = resp[len(str):]

def unescape_connect_string(str):
    """Unescape the %xx's in the connect string"""
    more = True
    while (more):
        m = re.match('^([^%]*)%([0-9a-fA-F]{2})(.*)$', str)
        if (m is not None):
            str = m.group(1) + chr(int(m.group(2), 16)) + m.group(3)
        else:
            more = False
    return str

def rinflash_connect(serport, dev_settings, connect_tag):
    """Wait for connect string from rinflash"""

    cs = dev_settings['Flash'][connect_tag].encode('ascii', 'ignore')
    cr = dev_settings['Flash']['ConnectResponse'].encode('ascii', 'ignore')
    uconnectstr = unescape_connect_string(cs)
    uconnectresp = unescape_connect_string(cr)

    print("Waiting for {} string '{}'".format(connect_tag, cs))
    adopt_device_serial_settings(serport=serport, dev_settings=dev_settings, section='ComConnect')
    serial_wait_for(serport=serport, str=uconnectstr)
    serport.write(uconnectresp)

def boot_connect(serport, dev_settings):
    """Wait for boot mode connection string from rinflash"""
    rinflash_connect(serport=serport, dev_settings=dev_settings, connect_tag='BootConnect')

def user_connect(serport, dev_settings):
    """Wait for user mode connection string from rinflash"""
    rinflash_connect(serport=serport, dev_settings=dev_settings, connect_tag='UserConnect')

def import_device_file(filename):
    """Import device file using ini parser"""
    device = configparser.RawConfigParser()
    device.read(filename)
    return device

def main():
    parser = argparse.ArgumentParser(description='Simulate a rinstrum bootloader')
    parser.add_argument('port', action='store',
                        help='Serial port to listen on')
    parser.add_argument('device', action='store',
                        help='Device file to emulate')
    parser.add_argument('flashfile', action='store',
                        help='File to simulate the flash')
    parser.add_argument('--connect-mode', action='store', default='user', choices=['user', 'boot'],
                        help='The method to connect')
    parser.add_argument('--with-socat', action='store_true',
                        help='When using socat, /dev/pts/n doesn''t like the serial rts/dtr configuration, this works around the issue')
    args = parser.parse_args()

    dev_settings = import_device_file(filename=args.device)
    serialport = serial.Serial(port=args.port,
                               baudrate=9600,
                               bytesize=serial.EIGHTBITS,
                               parity=serial.PARITY_NONE,
                               stopbits=serial.STOPBITS_ONE,
                               timeout=None,
                               xonxoff=False,
                               rtscts=True if args.with_socat else False,
                               write_timeout=None,
                               dsrdtr=True if args.with_socat else False,
                               inter_byte_timeout=None)
    flashfile = establish_flash_file(filename=args.flashfile, dev_settings=dev_settings)
    # Do the connect sequence
    if (args.connect_mode == 'user'):
        user_connect(serport=serialport, dev_settings=dev_settings)
    else:
        boot_connect(serport=serialport, dev_settings=dev_settings)
    negotiate_baud_rate(serport=serialport)

    # Run the main loop
    main_loop(serport=serialport, dev_settings=dev_settings, flashfile=flashfile)

if __name__ == '__main__':
    main()
