/*
 * rinflash.c
 *
 *  Created on: Aug 21, 2017
 *      Author: ian
 */
#include <stdio.h>
#include <stdbool.h>
#include <getopt.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <rinprog/rinprog.h>
#include "config.h"


enum e_flash_cmd { fc_program, fc_erase, fc_write, fc_verify, fc_identify };


struct s_flash_command
{
  enum e_flash_cmd command;
  const char *filename;
  FILE *file;
};


static bool cmdline_display_help = false;
static bool cmdline_display_version = false;
static const char *cmdline_serial_port = NULL;
static const char *cmdline_device_file = NULL;
static bool cmdline_connect_usermode = true;
static size_t cmdline_num_commands = 0;
static struct s_flash_command cmdline_flash_commands[100];


static void cmdline_print_version(void)
{
  printf("Rinstrum Device Flash Programming\n"
         "Part of " PACKAGE_STRING ". Maintainer: " PACKAGE_BUGREPORT "\n");
}


static void cmdline_print_usage(const char *prog)
{
  printf("Usage: %s [options] <serial> <device> <command> [<command>] ...\n"
         "Rinstrum Device Flash Programming\n", prog);

  printf("Options:\n"
         "  -h --help           This help\n"
         "  -v --version        Print software version\n"
         "  -c --connect <mode> Connect to the device in either \"boot\" or \"user\" mode.\n"
         "                      Default \"user\"\n"
         "Arguments:\n"
         "  serial              The serial port to communicate with the device\n"
         "  device              The device file with the programming parameters\n"
         "Commands:\n"
         "  program <file>      Erase, program and verify the flash on the device with\n"
         "                      file. Doesn't operate with other commands\n"
         "  erase               Erase the flash on the device\n"
         "  write <file>        Program the flash on the device with file\n"
         "  verify <file>       Verify the flash on the device against file\n"
         "  identify            Print the device identity information\n"
         "Notes:\n"
         "  Program supports Motorola S-Record files only\n"
         "\n"
        );
}


static int cmdline_parse_options(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "help",        0, 0, 'h' },
    { "version",     0, 0, 'v' },
    { "connect",     1, 0, 'c' },
    { NULL,          0, 0, '\0' },
  };
  int c;

  /* Options */
  while (1)
  {
    c = getopt_long(argc, argv, "hvc:", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      /* Help */
      case 'h':
        cmdline_display_help = true;
        return 0;
        break;

      /* Version */
      case 'v':
        cmdline_display_version = true;
        return 0;
        break;

      /* Connect mode */
      case 'c':
        if (strcmp(optarg, "boot") == 0)
          cmdline_connect_usermode = false;
        else if (strcmp(optarg, "user") == 0)
          cmdline_connect_usermode = true;
        else
        {
          fprintf(stderr, "Error: Invalid --connect mode %s specified, try --help\n", optarg);
          return -1;
        }
        break;

      default:
        fprintf(stderr, "error: invalid command line option, try --help\n");
        return -1;
        break;
    }
  }

  /* Serial port */
  if (optind >= argc)
  {
    fprintf(stderr, "error: no serial port specified, try --help\n");
    return -1;
  }
  cmdline_serial_port = argv[optind++];

  /* Device */
  if (optind >= argc)
  {
    fprintf(stderr, "error: no device specified, try --help\n");
    return -1;
  }
  cmdline_device_file = argv[optind++];

  /* Commands */
  while (optind < argc)
  {
    if (strcmp("program", argv[optind]) == 0 ||
        strcmp("write", argv[optind]) == 0 ||
        strcmp("verify", argv[optind]) == 0)
    {
      if (strcmp("program", argv[optind]) == 0)
        cmdline_flash_commands[cmdline_num_commands].command = fc_program;
      else if (strcmp("write", argv[optind]) == 0)
        cmdline_flash_commands[cmdline_num_commands].command = fc_write;
      else /* if (strcmp("verify", argv[optind]) == 0) */
        cmdline_flash_commands[cmdline_num_commands].command = fc_verify;

      if (optind + 1 >= argc)
      {
        fprintf(stderr, "error: no file specified for command %s, try --help\n", argv[optind]);
        return -1;
      }
      cmdline_flash_commands[cmdline_num_commands].filename = argv[optind + 1];
      optind++;
    }
    else if (strcmp("erase", argv[optind]) == 0)
    {
      cmdline_flash_commands[cmdline_num_commands].command = fc_erase;
      cmdline_flash_commands[cmdline_num_commands].filename = NULL;
    }
    else if (strcmp("identify", argv[optind]) == 0)
    {
      cmdline_flash_commands[cmdline_num_commands].command = fc_identify;
      cmdline_flash_commands[cmdline_num_commands].filename = NULL;
    }
    else
    {
      fprintf(stderr, "error: unknown command %s, try --help\n", argv[optind]);
      return -1;
    }

    cmdline_num_commands++;
    optind++;
  }
  if (cmdline_num_commands <= 0)
  {
    fprintf(stderr, "error: no commands specified, try --help\n");
    return -1;
  }

  return 0;
}


static enum e_rp_result import_srecord_file(struct s_flash_command *cmd)
{
  FILE *in_srec = NULL;
  FILE *tempbin = NULL;
  enum e_rp_result result;

  cmd->file = NULL;

  /* Open input s-record file */
  in_srec = fopen(cmd->filename, "rb");
  if (NULL == in_srec)
  {
    result = rpr_iofail;
    goto isf_fail;
  }

  /* Create temporary file for binary version */
  tempbin = tmpfile();
  if (NULL == tempbin)
  {
    result = rpr_iofail;
    goto isf_fail;
  }

  /* Read s-record file into binary file */
  result = rp_read_srecord_file(in_srec, tempbin, 0xFF);
  if (rpr_success != result)
    goto isf_fail;

  /* Success */
  cmd->file = tempbin;
  result = rpr_success;

isf_fail:
  if (rpr_success != result && NULL != tempbin)
  {
    fclose(tempbin);
  }
  if (NULL != in_srec)
  {
    fclose(in_srec);
  }
  return result;
}


/** Make a command sequence from the commands passed (ie expand "program"), and open any necessary files */
static enum e_rp_result make_command_sequence(struct s_flash_command **cmd_list, size_t *cmd_list_size)
{
  size_t i;
  size_t j;
  struct s_flash_command *cmds;
  struct s_flash_command *prog_cmd;
  size_t num_cmds;
  enum e_rp_result result;

  /* Can only have "program" command by itself on command line */
  for (i = 0, prog_cmd = NULL; i < cmdline_num_commands; i++)
  {
    if (cmdline_flash_commands[i].command == fc_program)
    {
      prog_cmd = &cmdline_flash_commands[i];
      break;
    }
  }
  if (cmdline_num_commands != 1 && NULL != prog_cmd)
  {
    fprintf(stderr, "Error: Command \"program\" doesn't operate with other commands, try --help\n");
    return rpr_paramfail;
  }

  if (NULL != prog_cmd)
  {
    /* Program command, allocate a list of commands for erase, write, verify */
    cmds = calloc(3, sizeof(*cmds));
    if (NULL == cmds)
      return rpr_allocfail;

    /* Expand "program" into "erase", "write", "verify" */
    cmds[0].command = fc_erase;
    cmds[0].filename = NULL;
    cmds[0].file = NULL;

    cmds[1].command = fc_write;
    cmds[1].filename = prog_cmd->filename;
    cmds[1].file = NULL;

    cmds[2].command = fc_verify;
    cmds[2].filename = prog_cmd->filename;
    cmds[2].file = NULL;

    num_cmds = 3;
  }
  else
  {
    /* Not program command, copy over existing commands */
    cmds = calloc(cmdline_num_commands, sizeof(*cmds));
    if (NULL == cmds)
      return rpr_allocfail;

    for (i = 0; i < cmdline_num_commands; i++)
    {
      cmds[i].command = cmdline_flash_commands[i].command;
      cmds[i].filename = cmdline_flash_commands[i].filename;
      cmds[i].file = NULL;
    }
    num_cmds = cmdline_num_commands;
  }

  /* Open files as required */
  for (i = 0; i < num_cmds; i++)
  {
    if (NULL != cmds[i].filename)
    {
      /* Has this file already been opened? */
      for (j = 0; j < i; j++)
      {
        if (NULL != cmds[j].filename && strcmp(cmds[i].filename, cmds[j].filename) == 0)
        {
          /* Yep, so use it's existing file pointer */
          cmds[i].file = cmds[j].file;
          break;
        }
      }
      if (NULL == cmds[i].file)
      {
        /* Need to open it */
        result = import_srecord_file(&cmds[i]);
        if (rpr_success != result)
        {
          fprintf(stderr, "Error: Could not open file %s required for step %u\n", cmds[i].filename, (unsigned) i);
          return result;
        }
      }
    }
  }

  *cmd_list = cmds;
  *cmd_list_size = num_cmds;
  return rpr_success;
}


/** Make a command sequence from the commands passed (ie expand "program"), and open any necessary files */
static enum e_rp_result free_command_sequence(struct s_flash_command *cmd_list, size_t cmd_list_size)
{
  size_t i;
  size_t j;

  for (i = 0; i < cmd_list_size; i++)
  {
    if (NULL != cmd_list[i].file)
    {
      /* NULL any duplicate file pointers, so we don't fclose() twice */
      for (j = i + 1; j < cmd_list_size; j++)
      {
        if (cmd_list[j].file == cmd_list[i].file)
        {
          cmd_list[j].file = NULL;
        }
      }
      fclose(cmd_list[i].file);
      cmd_list[i].file = NULL;
    }
  }
  free(cmd_list);
  return rpr_success;
}


/** Write or verify a file */
static enum e_rp_result write_verify(struct s_rp_device *rp_device, const struct s_flash_command *cmd)
{
  enum e_rp_result result;
  uint32_t fail_addr = 0;
  uint32_t start;
  uint32_t length;

  /* Check the file is ok for programming */
  result = rp_verify_image_for_device(rp_device, cmd->file);
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Verifying image %s for device failed with %s\n", cmd->filename, rp_result_to_string(result));
    return result;
  }

  /* Work out start and length */
  start = rp_device->device_settings->flash_start;
  (void) rp_get_file_length(cmd->file, &length);
  length -= start;

  if (fc_write == cmd->command)
  {
    /* Program or verify the file */
    printf("Programming device flash from 0x%08lX to 0x%08lX\n",
           (unsigned long) start,
           (unsigned long) start + length);
    result = rp_device_program(rp_device, start, length, cmd->file, &fail_addr);
    if (rpr_success != result)
    {
      fprintf(stderr, "Error: Programming device failed with %s at address 0x%08lx\n", rp_result_to_string(result), (unsigned long) fail_addr);
      return result;
    }
  }
  else
  {
    printf("Verifying device flash from 0x%08lX to 0x%08lX\n",
           (unsigned long) start,
           (unsigned long) start + length);
    result = rp_device_verify(rp_device, start, length, cmd->file, &fail_addr);
    if (rpr_success != result)
    {
      fprintf(stderr, "Error: Verifying device failed with %s at address 0x%08lx\n", rp_result_to_string(result), (unsigned long) fail_addr);
      return result;
    }

  }
  return rpr_success;
}


static enum e_rp_result identify_device(struct s_rp_device *rp_device)
{
  struct s_rp_device_id device_id;

  enum e_rp_result result;

  result = rp_device_identify(rp_device, &device_id);
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Failed to retrieve bootloader identity with %s\n", rp_result_to_string(result));
    return result;
  }

  printf("Bootloader Device ID:\n");
  printf("  Start:         0x%lX (%lu)\n", (unsigned long) rp_device->device_id.start, (unsigned long) rp_device->device_id.start);
  printf("  Length:        0x%lX (%lu)\n", (unsigned long) rp_device->device_id.length, (unsigned long) rp_device->device_id.length);
  printf("  Checksum Seed: 0x%0X (%u)\n", (unsigned) rp_device->device_id.checksum_seed, (unsigned) rp_device->device_id.checksum_seed);
  printf("  ID String:     %s\n", rp_device->device_id.name);
  return rpr_success;
}


/** Run each of the commands in the command sequence */
static enum e_rp_result run_command_sequence(struct s_rp_device *rp_device, const struct s_flash_command *cmds, size_t num_cmds)
{
  size_t i;
  enum e_rp_result result;

  for (i = 0; i < num_cmds; i++)
  {
    switch (cmds[i].command) {
      case fc_erase:
        printf("Erasing device flash from 0x%08lX to 0x%08lX\n",
               (unsigned long) rp_device->device_settings->flash_start,
               (unsigned long) rp_device->device_settings->flash_start + rp_device->device_settings->flash_length);
        result = rp_device_erase(rp_device, rp_device->device_settings->flash_start, rp_device->device_settings->flash_length);
        if (rpr_success != result)
        {
          fprintf(stderr, "Error: Erasing device failed with %s\n", rp_result_to_string(result));
          goto rcs_fail;
        }
        break;
      case fc_identify:
        result = identify_device(rp_device);
        if (rpr_success != result)
          goto rcs_fail;
        break;
      case fc_write:
      case fc_verify:
        result = write_verify(rp_device, &cmds[i]);
        if (rpr_success != result)
          goto rcs_fail;
        break;
      default:
        fprintf(stderr, "Error: Unknown command\n");
        result = rpr_paramfail;
        goto rcs_fail;
        break;
    }
  }

rcs_fail:
  return result;
}


/** Application entry point */
int main(int argc, char *argv[])
{
  FILE *device_file = NULL;
  struct s_rp_device *rp_device = NULL;
  enum e_rp_result result;
  struct s_flash_command *cmds = NULL;
  size_t num_cmds = 0;

  /* Do command line */
  if (cmdline_parse_options(argc, argv) < 0)
  {
    return (int) rpr_paramfail;
  }
  if (cmdline_display_help)
  {
    cmdline_print_usage(argv[0]);
    return (int) rpr_success;
  }
  if (cmdline_display_version)
  {
    cmdline_print_version();
    return (int) rpr_success;
  }

  /* Read device settings from file */
  device_file = fopen(cmdline_device_file, "r");
  if (NULL == device_file)
  {
    fprintf(stderr, "error: could not open device file %s\n", cmdline_device_file);
    result = (int) rpr_iofail;
    goto m_fail;
  }

  /* Initialise device structure */
  result = rp_device_init(&rp_device, device_file, cmdline_serial_port);
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Device creation failed with %s\n", rp_result_to_string(result));
    goto m_fail;
  }

  /* Make a command sequence from the commands passed */
  result = make_command_sequence(&cmds, &num_cmds);
  if (rpr_success != result)
  {
    goto m_fail;
  }

  /* Connect to device */
  if (cmdline_connect_usermode)
  {
    printf("Connecting to device via user mode\n");
    result = rp_device_connect_usermode(rp_device);
  }
  else
  {
    printf("Connecting to device via boot mode. Please reset device, or power on\n");
    result = rp_device_connect_bootmode(rp_device);
  }
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Connect to device failed with %s\n", rp_result_to_string(result));
    goto m_fail;
  }

  /* Check that the device settings match the device */
  result = rp_verify_device_settings(rp_device, &rp_device->device_id);
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Verifying device settings failed with %s\n", rp_result_to_string(result));
    goto m_fail;
  }

  /* Run the programming sequence */
  result = run_command_sequence(rp_device, cmds, num_cmds);
  if (rpr_success != result)
  {
      goto m_fail;
  }

  /* Reboot the device */
  printf("Rebooting the device\n");
  result = rp_device_reboot(rp_device);
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Rebooting device failed with %s\n", rp_result_to_string(result));
    goto m_fail;
  }

  /* Success */
  printf("Success\n");
  result = rpr_success;

m_fail:
  if (NULL != device_file)
  {
    (void) fclose(device_file);
  }
  if (NULL != rp_device)
  {
    (void) rp_device_free(rp_device);
  }
  if (NULL != cmds)
  {
    (void) free_command_sequence(cmds, num_cmds);
  }
  return (int) result;
}
