/*
 * rinflashiden.c
 *
 *  Created on: Aug 21, 2017
 *      Author: ian
 */
#include <stdio.h>
#include <stdbool.h>
#include <getopt.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include <regex.h>
#include <ctype.h>
#include <rinprog/rinprog.h>
#include "config.h"

static bool cmdline_display_help = false;
static bool cmdline_display_version = false;
static const char *cmdline_serial_port = NULL;
static const char *cmdline_serial_connect_settings = "9600n81";
static const char *cmdline_serial_program_settings = "9600n81";
static const char *cmdline_connect_string = "FlAsHRuN";
static const char *cmdline_reply_string = "FlAsHRuN";
static bool cmdline_connect_usermode = true;


static void cmdline_print_version(void)
{
  printf("Rinstrum Device Flash Bootloader Identifier\n"
         "Part of " PACKAGE_STRING ". Maintainer: " PACKAGE_BUGREPORT "\n");
}


static void cmdline_print_usage(const char *prog)
{
  printf("Usage: %s [options] <serial>\n"
         "Rinstrum Device Flash Bootloader Identifier\n", prog);

  printf("Options:\n"
         "  -h --help           This help\n"
         "  -v --version        Print software version\n"
         "  -c --connect <mode> Connect to the device in either \"boot\" or \"user\" mode.\n"
         "                      Default \"user\"\n"
         "  -s --serial-connect <mode>\n"
         "                      Initial serial connection using these parameters.\n"
         "                      Default \"%s\".\n"
         "  -p --serial-program <mode>\n"
         "                      Programming serial connection using these parameters.\n"
         "                      Default \"%s\".\n"
         "  -o --connect-str <str>\n"
         "                      Attempt to connect to the device with this string.\n"
         "                      Default \"%s\".\n"
         "  -r --reply-str <str>\n"
         "                      Expect the device to reply with this string.\n"
         "                      Default \"%s\".\n"
         "Arguments:\n"
         "  serial              The serial port to communicate with the device\n"
         "\n",
         cmdline_serial_connect_settings,
         cmdline_serial_program_settings,
         cmdline_connect_string,
         cmdline_reply_string
        );
}


static int cmdline_parse_options(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "help",           0, 0, 'h' },
    { "version",        0, 0, 'v' },
    { "connect",        1, 0, 'c' },
    { "serial-connect", 1, 0, 's' },
    { "serial-program", 1, 0, 'p' },
    { "connect-str",    1, 0, 'o' },
    { "reply-str",      1, 0, 'r' },
    { NULL,             0, 0, '\0' },
  };
  int c;

  /* Options */
  while (1)
  {
    c = getopt_long(argc, argv, "hvc:s:p:o:r:", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      /* Help */
      case 'h':
        cmdline_display_help = true;
        return 0;
        break;

      /* Version */
      case 'v':
        cmdline_display_version = true;
        return 0;
        break;

      /* Connect mode */
      case 'c':
        if (strcmp(optarg, "boot") == 0)
          cmdline_connect_usermode = false;
        else if (strcmp(optarg, "user") == 0)
          cmdline_connect_usermode = true;
        else
        {
          fprintf(stderr, "Error: Invalid --connect mode %s specified, try --help\n", optarg);
          return -1;
        }
        break;

      /* Serial port connect settings */
      case 's':
        cmdline_serial_connect_settings = optarg;
        break;

      /* Serial port program settings */
      case 'p':
        cmdline_serial_program_settings = optarg;
        break;

      /* Connect string */
      case 'o':
        cmdline_connect_string = optarg;
        break;

      /* Reply string */
      case 'r':
        cmdline_reply_string = optarg;
        break;

      default:
        fprintf(stderr, "error: invalid command line option, try --help\n");
        return -1;
        break;
    }
  }

  /* Serial port */
  if (optind >= argc)
  {
    fprintf(stderr, "error: no serial port specified, try --help\n");
    return -1;
  }
  cmdline_serial_port = argv[optind++];

  return 0;
}


static enum e_rp_result identify_device(struct s_rp_device *rp_device)
{
  struct s_rp_device_id device_id;

  enum e_rp_result result;

  result = rp_device_identify(rp_device, &device_id);
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Failed to retrieve bootloader identity with %s\n", rp_result_to_string(result));
    return result;
  }

  printf("Bootloader Device ID:\n");
  printf("  Start:         0x%lX (%lu)\n", (unsigned long) rp_device->device_id.start, (unsigned long) rp_device->device_id.start);
  printf("  Length:        0x%lX (%lu)\n", (unsigned long) rp_device->device_id.length, (unsigned long) rp_device->device_id.length);
  printf("  Checksum Seed: 0x%0X (%u)\n", (unsigned) rp_device->device_id.checksum_seed, (unsigned) rp_device->device_id.checksum_seed);
  printf("  ID String:     %s\n", rp_device->device_id.name);
  return rpr_success;
}


static enum e_rp_result parse_serial_config(const char *config, uint32_t *baud_rate,
                                uint8_t *data_bits, char *parity_bits, uint8_t *stop_bits)
{
  #define REGEX_GROUPS 5
  #define CFG_STR_LEN 100
  regex_t regex;
  regmatch_t regex_groups[REGEX_GROUPS];
  size_t group;
  char str[CFG_STR_LEN];
  size_t so;
  size_t eo;
  size_t len;
  enum e_rp_result result;

  /* 9600n81 */
  const char serconfig_regexstr[] = "([0-9]+)([noeNOE])([5-8])([1-2])";
  if (regcomp(&regex, serconfig_regexstr, REG_EXTENDED))
  {
    fprintf(stderr, "parse_serial_config: Could not compile regex string \"%s\"\n", config);
    return rpr_unknown_error;
  }

  /* Parse the serial configuration string against the regex, and pick it apart */
  if (regexec(&regex, config, REGEX_GROUPS, regex_groups, 0) == 0)
  {
    result = rpr_success;
    for (group = 0; group < REGEX_GROUPS; group++)
    {
      if (regex_groups[group].rm_so == -1)
      {
        fprintf(stderr, "parse_serial_config: Not enough groups.\n");
        result = rpr_paramfail;
        goto psc_done;
      }
      so = regex_groups[group].rm_so;
      eo = regex_groups[group].rm_eo;
      len = eo - so;
      if (len >= CFG_STR_LEN)
      {
        fprintf(stderr, "parse_serial_config: Group %lu too long.\n", (unsigned long) group);
        result = rpr_paramfail;
        goto psc_done;
      }
      memcpy(str, &config[so], len);
      str[len] = '\0';

      switch (group)
      {
        case 0:
          break;
        case 1:
          *baud_rate = atol(str);
          break;
        case 2:
          *parity_bits = tolower(str[0]);
          break;
        case 3:
          *data_bits = atoi(str);
          break;
        case 4:
          *stop_bits = atoi(str);
          break;
      }
    }
  }
  else
  {
    fprintf(stderr, "parse_serial_config: Could not execute regex.\n");
    result = rpr_unknown_error;
  }

psc_done:
  regfree(&regex);
  return result;
}


/** Application entry point */
int main(int argc, char *argv[])
{
  struct s_rp_device *rp_device = NULL;
  enum e_rp_result result;

  /* Do command line */
  if (cmdline_parse_options(argc, argv) < 0)
  {
    return (int) rpr_paramfail;
  }
  if (cmdline_display_help)
  {
    cmdline_print_usage(argv[0]);
    return (int) rpr_success;
  }
  if (cmdline_display_version)
  {
    cmdline_print_version();
    return (int) rpr_success;
  }

  /* Initialise device structure with default values */
  result = rp_device_init(&rp_device, NULL, cmdline_serial_port);
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Device creation failed with %s\n", rp_result_to_string(result));
    goto m_fail;
  }

  /* Fill in values from the command line */
  free(rp_device->device_settings->boot_connect_string);
  rp_device->device_settings->boot_connect_string = rp_ini_connectstr_unescape(cmdline_connect_string);
  free(rp_device->device_settings->user_connect_string);
  rp_device->device_settings->user_connect_string = rp_ini_connectstr_unescape(cmdline_connect_string);
  free(rp_device->device_settings->connect_response);
  rp_device->device_settings->connect_response = rp_ini_connectstr_unescape(cmdline_reply_string);
  result = parse_serial_config(cmdline_serial_connect_settings,
                               &rp_device->device_settings->connect_settings.baud_rate,
                               &rp_device->device_settings->connect_settings.data_bits,
                               &rp_device->device_settings->connect_settings.parity_bits,
                               &rp_device->device_settings->connect_settings.stop_bits);
  if (rpr_success != result)
    goto m_fail;
  result = parse_serial_config(cmdline_serial_program_settings,
                               &rp_device->device_settings->program_settings.baud_rate,
                               &rp_device->device_settings->program_settings.data_bits,
                               &rp_device->device_settings->program_settings.parity_bits,
                               &rp_device->device_settings->program_settings.stop_bits);
  if (rpr_success != result)
    goto m_fail;

  /* Connect to device */
  if (cmdline_connect_usermode)
  {
    printf("Connecting to device via user mode\n");
    result = rp_device_connect_usermode(rp_device);
  }
  else
  {
    printf("Connecting to device via boot mode. Please reset device, or power on\n");
    result = rp_device_connect_bootmode(rp_device);
  }
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Connect to device failed with %s\n", rp_result_to_string(result));
    goto m_fail;
  }

  /* Print bootloader information */
  result = identify_device(rp_device);
  if (rpr_success != result)
    goto m_fail;

  /* Reboot the device */
  printf("Rebooting the device\n");
  result = rp_device_reboot(rp_device);
  if (rpr_success != result)
  {
    fprintf(stderr, "Error: Rebooting device failed with %s\n", rp_result_to_string(result));
    goto m_fail;
  }

  /* Success */
  printf("Success\n");
  result = rpr_success;

m_fail:
  if (NULL != rp_device)
  {
    (void) rp_device_free(rp_device);
  }
  return (int) result;
}
