/*
 * rp_test_teacrypt.c
 *
 * See tea_cbc.py and tea.py for generation/verification of values
 *
 *  Created on: Aug 30, 2017
 *      Author: ian
 */
#include <string.h>
#include <stdio.h>
#include <rinprog/rp_teacrypt.h>
#include <cutest/CuTest.h>

#include "rp_test_teacrypt.h"


static const struct s_rp_tea_data plaintext = { { 0x8289B424, 0x9C57D08C } };
static const struct s_rp_tea_data expected_ciphertext = { { 0xe2bcb973, 0x7a82aba4 } };
static const struct s_rp_tea_key key = { { 0x763e5ade, 0xa0c17849, 0xf1b8a106, 0xb760f872 } };

static const struct s_rp_tea_data cbc_iv = { { 0x04EDE2EE, 0x0B8C6D21 } };
static const struct s_rp_tea_data cbc_expected_ciphertext1 = { { 0x864F8443, 0x76C1D5C8 } };
static const struct s_rp_tea_data cbc_expected_ciphertext2 = { { 0x69323C4B, 0xF413DB14 } };


/** Verify TEA encryption/decryption */
static void test_rinprog_tea_crypt_decrypt(CuTest* tc)
{
  struct s_rp_tea_data work_data;

  work_data = plaintext;
  rp_tea_encrypt(&work_data, &key);
  CuAssertIntEquals(tc, expected_ciphertext.data[0], work_data.data[0]);
  CuAssertIntEquals(tc, expected_ciphertext.data[1], work_data.data[1]);

  /* printf("%08x %08x\n", work_data.data[0], work_data.data[1]); */

  rp_tea_decrypt(&work_data, &key);
  CuAssertIntEquals(tc, plaintext.data[0], work_data.data[0]);
  CuAssertIntEquals(tc, plaintext.data[1], work_data.data[1]);

  /* printf("%08x %08x\n", work_data.data[0], work_data.data[1]); */
}


/** Verify TEA encryption/decryption with cipher block chaining */
static void test_rinprog_cbc_tea_crypt_decrypt(CuTest* tc)
{
  struct s_rp_tea_data encrypted_data1;
  struct s_rp_tea_data encrypted_data2;
  struct s_rp_tea_data decrypted_data1;
  struct s_rp_tea_data decrypted_data2;
  struct s_rp_tea_data work_iv1;
  struct s_rp_tea_data work_iv2;

  /* With CBC, encrypting the same data one after the other should give different ciphertext */

  /* Encrypt first round */
  encrypted_data1 = plaintext;
  work_iv1 = cbc_iv;
  rp_cbc_tea_encrypt(&encrypted_data1, &work_iv1, &key);
  CuAssertIntEquals(tc, cbc_expected_ciphertext1.data[0], encrypted_data1.data[0]);
  CuAssertIntEquals(tc, cbc_expected_ciphertext1.data[1], encrypted_data1.data[1]);
  CuAssertIntEquals(tc, cbc_expected_ciphertext1.data[0], work_iv1.data[0]);
  CuAssertIntEquals(tc, cbc_expected_ciphertext1.data[1], work_iv1.data[1]);

  /* Encrypt second round */
  encrypted_data2 = plaintext;
  work_iv2 = work_iv1;
  rp_cbc_tea_encrypt(&encrypted_data2, &work_iv2, &key);
  CuAssertIntEquals(tc, cbc_expected_ciphertext2.data[0], encrypted_data2.data[0]);
  CuAssertIntEquals(tc, cbc_expected_ciphertext2.data[1], encrypted_data2.data[1]);
  CuAssertIntEquals(tc, cbc_expected_ciphertext2.data[0], work_iv2.data[0]);
  CuAssertIntEquals(tc, cbc_expected_ciphertext2.data[1], work_iv2.data[1]);

  /* CBC means ciphertext of two blocks of the same plaintext will be different */
  CuAssertTrue(tc, cbc_expected_ciphertext1.data[0] != cbc_expected_ciphertext2.data[0]);
  CuAssertTrue(tc, cbc_expected_ciphertext1.data[1] != cbc_expected_ciphertext2.data[1]);

  /* Decrypt first round */
  decrypted_data1 = encrypted_data1;
  work_iv1 = cbc_iv;
  rp_cbc_tea_decrypt(&decrypted_data1, &work_iv1, &key);
  CuAssertIntEquals(tc, plaintext.data[0], decrypted_data1.data[0]);
  CuAssertIntEquals(tc, plaintext.data[1], decrypted_data1.data[1]);
  CuAssertIntEquals(tc, encrypted_data1.data[0], work_iv1.data[0]);
  CuAssertIntEquals(tc, encrypted_data1.data[1], work_iv1.data[1]);

  /* Decrypt second round */
  decrypted_data2 = encrypted_data2;
  work_iv2 = work_iv1;
  rp_cbc_tea_decrypt(&decrypted_data2, &work_iv2, &key);
  CuAssertIntEquals(tc, plaintext.data[0], decrypted_data2.data[0]);
  CuAssertIntEquals(tc, plaintext.data[1], decrypted_data2.data[1]);
  CuAssertIntEquals(tc, encrypted_data2.data[0], work_iv2.data[0]);
  CuAssertIntEquals(tc, encrypted_data2.data[1], work_iv2.data[1]);
}


CuSuite *CuGetSuite_rinprog_teacrypt(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_rinprog_tea_crypt_decrypt);
  SUITE_ADD_TEST(suite, test_rinprog_cbc_tea_crypt_decrypt);

  return suite;
}
