import unittest
import struct
import tea

class TinyEncryptionAlgorithmWithCBC(tea.TinyEncryptionAlgorithm):
    """Extend TEA with cipher block chaining """

    def cbc_encrypt_block(self, block_data, block_iv, block_key):
        """Encrypt a block with TEA and chain with CBC"""
        iblk_data = list(block_data)
        iblk_iv = list(block_iv)

        iblk_data[0] ^= iblk_iv[0]
        iblk_data[1] ^= iblk_iv[1]
        ciphertext = self.encrypt_block(iblk_data, block_key)
        iblk_iv = list(ciphertext)
        return ciphertext, iblk_iv

    def cbc_decrypt_block(self, block_data, block_iv, block_key):
        iblk_data = list(block_data)
        iblk_temp_data = list(block_data)
        plaintext = self.decrypt_block(iblk_data, block_key)
        plaintext[0] ^= block_iv[0]
        plaintext[1] ^= block_iv[1]
        iblk_iv = list(iblk_temp_data)
        return plaintext, iblk_iv

class TestTEABase(unittest.TestCase):
    """Base class for testing TEA and TEA CBC"""

    def setUp(self):
        self.plaintext = "\x82\x89\xb4\x24\x9c\x57\xd0\x8c"
        self.key = "\x76\x3e\x5a\xde\xa0\xc1\x78\x49\xf1\xb8\xa1\x06\xb7\x60\xf8\x72"

        self.block_plaintext = self.to_block(self.plaintext)
        self.assertEqual(first=len(self.block_plaintext), second=2, msg="Plaintext must contain 2 blocks")
        #self.dump_text("Plaintext in: ", self.plaintext)

        self.block_key = self.to_block(self.key)
        self.assertEqual(first=len(self.block_key), second=4, msg="Key must contain 4 blocks")
        #self.dump_text("Key: ", block_key)

    def to_block(self, data):
        """Convert arbitrary data into blocks of uint32 for TEA"""
        block_list = []
        for i in range(0, len(data), 4):
            block = struct.unpack("!I", data[i:i + 4])[0]
            block_list.append(block)
        return block_list

    def dump_text(self, title, data):
        """Dump a list of blocks"""
        str = title
        for num in data:
            str = str + "0x{:08X} ".format(num)
        print(str)

class TestTEA(TestTEABase):
    """Tests for the TEA"""

    def setUp(self):
        TestTEABase.setUp(self)

        self.block_expected_ciphertext = [ 0xe2bcb973, 0x7a82aba4 ]

    def test_tea(self):
        teac = tea.TinyEncryptionAlgorithm()

        plaintext = list(self.block_plaintext)
        ciphertext = teac.encrypt_block(plaintext, self.block_key)
        #self.dump_text("Ciphertext: ", ciphertext)
        self.assertEqual(first=ciphertext[0], second=self.block_expected_ciphertext[0], msg="TEA encrypt block 0")
        self.assertEqual(first=ciphertext[1], second=self.block_expected_ciphertext[1], msg="TEA encrypt block 1")

        plaintext_dec = teac.decrypt_block(ciphertext, self.block_key)
        #self.dump_text("Decrypted plaintext: ", plaintext)
        self.assertEqual(first=plaintext_dec[0], second=self.block_plaintext[0], msg="TEA decryption block 0")
        self.assertEqual(first=plaintext_dec[1], second=self.block_plaintext[1], msg="TEA decryption block 1")

class TestTEA_CBC(TestTEABase):
    """Tests for TEA with CBC"""

    def setUp(self):
        TestTEABase.setUp(self)
        self.cbciv = "\x04\xed\xe2\xee\x0b\x8c\x6d\x21"

        self.block_cbciv = self.to_block(self.cbciv)
        self.assertEqual(first=len(self.block_plaintext), second=2, msg="CBC IV must contain 2 blocks")

        self.block_expected_ciphertext1 = [ 0x864F8443, 0x76C1D5C8 ]
        self.block_expected_ciphertext2 = [ 0x69323C4B, 0xF413DB14 ]

    def test_tea_cbc(self):
        teac = TinyEncryptionAlgorithmWithCBC()

        #self.dump_text("Plaintext in: ", self.block_plaintext)
        #self.dump_text("Block IV: ", self.block_cbciv)
        #self.dump_text("Key: ", self.block_key)

        ciphertext1, cbciv1 = teac.cbc_encrypt_block(self.block_plaintext, self.block_cbciv, self.block_key)
        #self.dump_text("Ciphertext1: ", ciphertext1)
        #self.dump_text("CBC IV1: ", cbciv1)
        self.assertEqual(first=ciphertext1[0], second=self.block_expected_ciphertext1[0], msg="TEA CBC first round encrypt block 0")
        self.assertEqual(first=ciphertext1[1], second=self.block_expected_ciphertext1[1], msg="TEA CBC first round encrypt block 1")

        ciphertext2, cbciv2 = teac.cbc_encrypt_block(self.block_plaintext, cbciv1, self.block_key)
        #self.dump_text("Ciphertext2: ", ciphertext2)
        #self.dump_text("CBC IV2: ", cbciv2)
        self.assertEqual(first=ciphertext2[0], second=self.block_expected_ciphertext2[0], msg="TEA CBC second round encrypt block 0")
        self.assertEqual(first=ciphertext2[1], second=self.block_expected_ciphertext2[1], msg="TEA CBC second round encrypt block 1")

        plaintext1, cbciv1 = teac.cbc_decrypt_block(ciphertext1, self.block_cbciv, self.block_key)
        #self.dump_text("Plaintext1: ", plaintext1)
        #self.dump_text("CBC IV1: ", cbciv1)
        self.assertEqual(first=plaintext1[0], second=self.block_plaintext[0], msg="TEA CBC first round decrypt block 0")
        self.assertEqual(first=plaintext1[1], second=self.block_plaintext[1], msg="TEA CBC first round decrypt block 1")

        plaintext2, cbciv2 = teac.cbc_decrypt_block(ciphertext2, cbciv1, self.block_key)
        #self.dump_text("Plaintext2: ", plaintext2)
        #self.dump_text("CBC IV2: ", cbciv2)
        self.assertEqual(first=plaintext2[0], second=self.block_plaintext[0], msg="TEA CBC second round decrypt block 0")
        self.assertEqual(first=plaintext2[1], second=self.block_plaintext[1], msg="TEA CBC second round decrypt block 1")

if __name__ == '__main__':
    unittest.main()
