﻿/*

MODULE NAME :   ADC1216.C DRIVER

FILE NAME :     adc1216.C

AUTHOR(S):      KEN BRAY

DESCRIPTION:    This module drives the adc hardware and provides an interface
                through function calls to higher level software.  All operations
                are performed through functions.  No global variables exist.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rinhw/adc_ads1216.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/rstdio.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcreadings.h>
#include <rinweigh/adcdebug.h>
#include <rinweigh/adctemperature.h>
#include <rinweigh/adcaverage.h>
#include <rinweigh/adcads1216.h>


/*ADC register Constants*/
/* #define DEFAULT_SPAN    (int32s)0x8111BB */ /*the default value of span to use */
#define GAIN128_TOP     (int32s)8192000  /*Adc top out for 128 gain (3.2mV/V) */
#define GAIN64_TOP      (int32s)12800000 /*Adc top out for 64 gain (5mV/V) */


#define BOCS_OFF        (int8u)0x00    /* Burn Out Current Source Off*/
#define PGA             (int8u)0x07    /* B2-B0 Gain select 1,2,4,8,16,32,64,128*/
#define PGA64           (int8u)0x06    /* PGA setting for gain of 64 */
//#define PGA_4_ZMUX      (int8u)0x03    /* Gain value for Z-mux cycle */
#define ACR_VAL         (PGA64)


/*General Constants*/
#if 0
#define SPAN_CAL        15 /* Number of readings discarded after span cal */
#define OFFSET_CAL      (SPAN_CAL + 1)
#define USER_AVE_BUFFER_SIZE 30 /* Maximum size of buffer for user average */
#define NOISE_AVE_BUFFER_SIZE 20/* Maximum size of buffer for noise red averg*/
#define NOISE_FLOOR_LIMIT  (int32s)300/* 700 nV (5 / 128 / 2^24 * 300) V*/
                                               /*   Ref / Gain / Res * Val   */
#endif


static bool AdcIntHwGetData(AdcSuperStructure *ass, int32s *data, int32u timeout, bool *data_wait_failed)
{

  /* Some checks (null pointer or device closed) */
  if (NULL == ass->adcHwInfo.adc_rdg_dev ||
      0 == (ass->adcHwInfo.adc_rdg_dev->status & O_INITIALISED) ||
      0 != (ass->adcHwInfo.adc_rdg_dev->status & O_OPEN))
  {
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
    *data_wait_failed = true;
    return false;
  }

  /* Wait for data to become available (DRDY) */
  *data_wait_failed = false;
  if (rioctl(ass->adcHwInfo.adc_rdg_dev, IOCTL_ADC24BIT_WAIT_FOR_DATA_READY, (unsigned long) timeout) < 0)
  {
    /*
     * Timeout waiting for data. A timeout is not a hardware error,
     * so we don't set the ADC_ERR_HARDWARE flag here
     */
    return false;
  }

  /*
   * Only read if not skipping readings, because if we try to
   * read from the device while it's in the middle of a
   * self zero calibration, then it's possible to confuse
   * it, and it will return rubbish until it is re-initialised
   */
  if (0 == ass->adcHwInfo.skip)
  {
    /* Seek start of device */
    if (rlseek(ass->adcHwInfo.adc_rdg_dev, 0L, SEEK_SET) < 0)
    {
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
      return false;
    }

    /* Read from device */
    if (sizeof(*data) != rread(ass->adcHwInfo.adc_rdg_dev, data, sizeof(*data), timeout))
    {
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
      return false;
    }
  }
  else
  {
    *data = 0;
  }

  /*
   * Success, even if skipped reading,
   * because we test skip again from the caller later on
   * to determine if we should use this reading
   */
  return true;
}


static bool AdcIntHwResetAdc(struct t_io_dev *adc_rdg_dev)
{
  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_FORCE_RESET) >= 0);
}


static bool AdcIntHwSetZeroRegister(struct t_io_dev *adc_rdg_dev, int32s zero_value)
{
  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_SET_ZERO_REGISTER, zero_value) >= 0);
}


#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
static bool AdcIntHwGetZeroRegister(struct t_io_dev *adc_rdg_dev, int32s *zero_value)
{
  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_GET_ZERO_REGISTER, zero_value) >= 0);
}
#endif


static bool AdcIntHwSetSpanRegister(struct t_io_dev *adc_rdg_dev, int32s span_value)
{
  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_SET_SPAN_REGISTER, span_value) >= 0);
}


#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_BASIC) || ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
static bool AdcIntHwGetSpanRegister(struct t_io_dev *adc_rdg_dev, int32s *span_value)
{
  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_GET_SPAN_REGISTER, span_value) >= 0);
}
#endif


static bool AdcIntHwSetPGAGain(struct t_io_dev *adc_rdg_dev, int8u gain)
{
  int ioctlgain = 1 << gain;

  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_SET_PGA_GAIN, ioctlgain) >= 0);
}


static bool AdcIntHwSetBurnoutCurrentSource(struct t_io_dev *adc_rdg_dev, bool on)
{
  int value = (on) ? IOCTL_ADC24BIT_BURNOUT_ON : IOCTL_ADC24BIT_BURNOUT_OFF;

  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_SET_BURNOUTCURRENTSOURCE, value) >= 0);
}


static bool AdcIntHwSetDecimationRatio(struct t_io_dev *adc_rdg_dev, int sync_freq)
{
  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_SET_DECIMATION_RATIO, sync_freq) >= 0);
}


static bool AdcIntHwSetFourWireMode(struct t_io_dev *adc_rdg_dev, bool on)
{
  int value = (on) ? IOCTL_ADC24BIT_LCWIRE_4WIRE : IOCTL_ADC24BIT_LCWIRE_6WIRE;

  if (NULL == adc_rdg_dev ||
      0 == (adc_rdg_dev->status & O_INITIALISED) ||
      0 != (adc_rdg_dev->status & O_OPEN))
  {
    return false;
  }

  /* This is not always implemented, so just ignore the result */
  (void) rioctl(adc_rdg_dev, IOCTL_ADC24BIT_SET_LCWIRE_MODE, value);
  return true;
}


#if 0
static bool AdcIntHwPerformSelfZeroCalibration(struct t_io_dev *adc_rdg_dev)
{
  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_PERFORM_SELF_ZERO_CALIBRATION) >= 0);
}
#endif


static bool AdcIntHwGetAdcRunStatus(struct t_io_dev *adc_rdg_dev)
{
  int responding = 0;

  return (NULL != adc_rdg_dev &&
          0 != (adc_rdg_dev->status & O_INITIALISED) &&
          0 == (adc_rdg_dev->status & O_OPEN) &&
          rioctl(adc_rdg_dev, IOCTL_ADC24BIT_IS_RESPONDING, &responding) >= 0 &&
          0 != responding);
}


static int8u AdcIntHwGetACRVal(AdcSuperStructure *ass)
{
  ass->adcHwInfo.adcMax = GAIN64_TOP;
  ass->adcHwInfo.readingScale = (PGA - PGA64);

  return ACR_VAL;
}


static void AdcIntSetupHardware(AdcSuperStructure *ass)
{
  int8u b;
  int32u self_zero_delay;
#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_BASIC)
  int32s check_zero;
  int32s check_span;

  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntSetupHardware called\n");
#endif

  if (0 != ass->adcHwInfo.skip)
  {
    /*
     * If the ADC is in the process of doing a self
     * zero calibration, then we can't talk to it,
     * otherwise it give strange readings.
     *
     * So we need to wait until it's finished,
     * since we don't have any kind of work queue
     * to schedule the re-init for after the
     * self zero has completed. This shouldn't be
     * a problem since re-init is only called
     * when changing settings, and not during
     * normal weighing.
     */
    if (0 == ass->adcHwInfo.adc_config_vars.prev_sync_freq)
    {
      /* Hasn't been set yet, go with worst case */
      ass->adcHwInfo.adc_config_vars.prev_sync_freq = AdcSyncFrequencies[ADC_SYNC_FREQ_10_HZ];
    }

    /* Calculate delay */
    self_zero_delay = ass->adcHwInfo.skip * 10 * OS_TICKS_PER_SEC;
    self_zero_delay /= ass->adcHwInfo.adc_config_vars.prev_sync_freq;
    self_zero_delay++;
    OSTimeDly(self_zero_delay);
  }
  ass->adcHwInfo.adc_config_vars.prev_sync_freq = ass->adcHwInfo.adc_config_vars.sync_freq;

  /* Reset the ADC */
  if (!AdcIntHwResetAdc(ass->adcHwInfo.adc_rdg_dev))
  {
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
  }

  b = AdcIntHwGetACRVal(ass);

  if (!AdcIntHwSetZeroRegister(ass->adcHwInfo.adc_rdg_dev, ass->adcHwInfo.adc_config_vars.zero_reg))
  {
    ass->adcMiscellaneous.adcErrors |=  ADC_ERR_HARDWARE;
  }
  if (!AdcIntHwSetSpanRegister(ass->adcHwInfo.adc_rdg_dev, ass->adcHwInfo.adc_config_vars.span_reg))
  {
    ass->adcMiscellaneous.adcErrors |=  ADC_ERR_HARDWARE;
  }
  if (!AdcIntHwSetPGAGain(ass->adcHwInfo.adc_rdg_dev, b))
  {
    ass->adcMiscellaneous.adcErrors |=  ADC_ERR_HARDWARE;
  }
  if (!AdcIntHwSetDecimationRatio(ass->adcHwInfo.adc_rdg_dev, ass->adcHwInfo.adc_config_vars.sync_freq))
  {
    ass->adcMiscellaneous.adcErrors |=  ADC_ERR_HARDWARE;
  }
  if (!AdcIntHwSetFourWireMode(ass->adcHwInfo.adc_rdg_dev, ass->adcHwInfo.adc_config_vars.four_wire_mode))
  {
    ass->adcMiscellaneous.adcErrors |=  ADC_ERR_HARDWARE;
  }

#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "SetupHardware: Set ZeroReg:%ld, SpanReg:%ld, Gain:%u, Freq:%u, 4Wire:%u\n",
                   (long) ass->adcHwInfo.adc_config_vars.zero_reg, (long) ass->adcHwInfo.adc_config_vars.span_reg, b,
                   ass->adcHwInfo.adc_config_vars.sync_freq, ass->adcHwInfo.adc_config_vars.four_wire_mode);

  (void) AdcIntHwGetZeroRegister(ass->adcHwInfo.adc_rdg_dev, &check_zero);
  (void) AdcIntHwGetSpanRegister(ass->adcHwInfo.adc_rdg_dev, &check_span);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "SetupHardware: Check ZeroReg:%ld, SpanReg:%ld\n",
                   (long) check_zero, (long) check_span);
#endif
}


static bool AdcIntHwGetAdcReading(AdcSuperStructure *ass, int32s *rdg, bool *data_wait_failed)
{
  int32s reading;
  int32u timeout;

  /* Work out the timeout, rounding up */
  timeout = (OS_TICKS_PER_SEC * 10 + ass->adcHwInfo.adc_config_vars.sync_freq / 2) / ass->adcHwInfo.adc_config_vars.sync_freq;
  if (0 == timeout)
  {
    timeout = 1;
  }
  if (timeout > UINT16_MAX)
  {
    timeout = UINT16_MAX;
  }

  /* Give it some leeway, so the interrupt happens first */
  timeout++;

  /* Get the data from the adc */
  if (!AdcIntHwGetData(ass, &reading, timeout, data_wait_failed))
  {
#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_BASIC)
    if (!*data_wait_failed)
    {
      ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwGetAdcReading: AdcGetData() timed out\n");
    }
#endif
    *rdg = 0;
    return false;
  }

  /* Scale */
  reading <<= ass->adcHwInfo.readingScale;

  /* Keep in limits */
  if (reading > ass->adcHwInfo.adcMax)
  {
    reading = ass->adcHwInfo.adcMax;
  }
  else if (reading < -ass->adcHwInfo.adcMax)
  {
    reading = -ass->adcHwInfo.adcMax;
  }

  /* Return the adc data reading */
  *rdg = reading;
  return true;
}


#if 0
/*
 * This is the old version of this function that takes the
 * result of the self zero and adds other compensation
 * before applying to the ADC. Since we don't use self
 * zero any more, this function is no longer used, and
 * a simplified version is included below.
 */
static void AdcIntHwTempCalFinishZero(AdcSuperStructure *ass)
{
  int32s muxed_zero_reg;
  int32s avged_muxed_zero_reg;
  int32s zero_reg;
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  int32s check_zero;
#endif

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: End Zero Stage\n");
#endif

  /* Get current zero register value (result of zero mux) */
  if (!AdcIntHwGetZeroRegister(ass->adcHwInfo.adc_rdg_dev, &muxed_zero_reg))
  {
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: Read zero register failed\n");
#endif
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
    return;
  }
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: Zero register %ld\n", (long) muxed_zero_reg);
#endif

  /* Before doing zero mux averaging, check for obviously flawed data (0 or > 0.1mV/V). 0 was seen in testing a couple of times (due to some bug at some frequencies). High values occur when the loadcell is unplugged. */
  if (0 == muxed_zero_reg || muxed_zero_reg > 256000 || muxed_zero_reg < -256000)
  {
    /* If there are any problems, just use the previous zero value */
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: Zero register after mux discarded: %ld\n", (long) muxed_zero_reg);
#endif
    avged_muxed_zero_reg = ass->adcTempCal.zeroCal;
  }
  else
  {
    /* Average the zero mux value */
    avged_muxed_zero_reg = AdcIntAverageReading(&ass->adcZeroMuxAveraging, muxed_zero_reg);
  }
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: Zero register after mux: %ld, averaged: %ld\n", (long) muxed_zero_reg, (long) avged_muxed_zero_reg);
#endif

  /* Write averaged zero mux reading to the cal structure */
  ass->adcTempCal.zeroCal = 0;

  /* Calculate the new zero offset (mux value + zero offset + zero drift) */
  zero_reg = AdcIntTempComp_CalculateZero(ass);
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: Add factory mV/V offset + zero drift gives new zero register value of %ld\n",
                   zero_reg);
#endif

  /* Write zero calibration into the ADC, and keep a copy for reference */
  ass->adcHwInfo.adc_config_vars.zero_reg = zero_reg;
  if (!AdcIntHwSetZeroRegister(ass->adcHwInfo.adc_rdg_dev, zero_reg))
  {
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
  }
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  (void) AdcIntHwGetZeroRegister(ass->adcHwInfo.adc_rdg_dev, &check_zero);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: Check ZeroReg:%ld\n", (long) check_zero);
#endif

  if (ass->adcTempCal.start > 0)
  {
    ass->adcTempCal.start--;
  }
}
#endif


static void AdcIntHwTempCalFinishZero(AdcSuperStructure *ass)
{
  int32s zero_reg;
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  int32s check_zero;
#endif

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: End Zero Stage\n");
#endif

  /* Write averaged zero mux reading to the cal structure. Since we don't do self zero any more, this is always zero. */
  ass->adcTempCal.zeroCal = 0;

  /* Calculate the new zero offset (mux value + zero offset + zero drift) */
  zero_reg = AdcIntTempComp_CalculateZero(ass);
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: Add factory mV/V offset + zero drift gives new zero register value of %ld\n",
                   zero_reg);
#endif

  /* Write zero calibration into the ADC, and keep a copy for reference */
  ass->adcHwInfo.adc_config_vars.zero_reg = zero_reg;
  if (!AdcIntHwSetZeroRegister(ass->adcHwInfo.adc_rdg_dev, zero_reg))
  {
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
  }
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  (void) AdcIntHwGetZeroRegister(ass->adcHwInfo.adc_rdg_dev, &check_zero);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwTempCalFinishZero: Check ZeroReg:%ld\n", (long) check_zero);
#endif

  if (ass->adcTempCal.start > 0)
  {
    ass->adcTempCal.start--;
  }
}


/** Calculate a new zero value for the current temperature */
int32s AdcIntTempComp_CalculateZero(const AdcSuperStructure *ass)
{
  int32s temp_delta;
  int32s zero_delta;
  int32s zero;
  int32s zero_offset;
  int32s zero_drift;

  /*
   * Calculate difference in temperature between current temperature,
   * and temperature at which the zero calibration (run on startup) was performed
   */
  temp_delta = (int32s)(ass->adcTempCal.currentTemp);
  temp_delta -= (int32s)(ass->adcFactoryDatabase.calTemp);

  /* Calculate necessary change in zero for this temperature difference */
  zero_delta = temp_delta;
  if (temp_delta >= 0)
    zero_drift = ass->adcFactoryDatabase.zeroDriftPos[ass->adcUserDatabase.syncFreqIdx];
  else
    zero_drift = ass->adcFactoryDatabase.zeroDriftNeg[ass->adcUserDatabase.syncFreqIdx];
  zero_delta *= zero_drift;

  /*
   * Calculate new zero value,
   * includes offset for mV/V cal (adcFactoryDatabase.zeroOffset)
   * and startup temperature zero offset (tempcal->zeroCal)
   */
  zero = zero_delta;
  zero /= ADC_ZERO_DRIFT_SCALING;
  zero += ass->adcTempCal.zeroCal;
  zero_offset = ass->adcFactoryDatabase.zeroOffset[ass->adcUserDatabase.syncFreqIdx];
  zero += zero_offset;

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntTempComp_CalculateZero: Temperature: %lu, Cal Temperature: %lu, Temperature delta: %ld\n"
                              "AdcIntTempComp_CalculateZero: zero delta: %ld, zero drift: %ld, zero cal: %ld\n"
                              "AdcIntTempComp_CalculateZero: zero offset: %ld, zero:%ld\n",
                   (unsigned long) ass->adcTempCal.currentTemp, (unsigned long) ass->adcFactoryDatabase.calTemp, (long) temp_delta,
                   (long) zero_delta, (long) zero_drift, (long) ass->adcTempCal.zeroCal,
                   (long) zero_offset, (long) zero);
#endif
  return zero;
}


/** Calculate a new span value for the current temperature */
int32s AdcIntTempComp_CalculateSpan(const AdcSuperStructure *ass)
{
  int32s temp_delta;
  int32s span_delta;
  int32s span;
  int32s span_drift;
  int32s span_cal;

  /*
   * Calculate difference in temperature between current temperature,
   * and temperature at which the factory calibration was performed
   */
  temp_delta = (int32s)(ass->adcTempCal.currentTemp);
  temp_delta -= (int32s)(ass->adcFactoryDatabase.calTemp);

  /* Calculate necessary change in span for this temperature difference */
  span_delta = temp_delta;
  if (temp_delta >= 0)
    span_drift = ass->adcFactoryDatabase.spanDriftPos[ass->adcUserDatabase.syncFreqIdx];
  else
    span_drift = ass->adcFactoryDatabase.spanDriftNeg[ass->adcUserDatabase.syncFreqIdx];
  span_delta *= span_drift;

  /* Calculate new span value */
  span = span_delta;
  span /= ADC_SPAN_DRIFT_SCALING;
  span_cal = ass->adcFactoryDatabase.spanCal[ass->adcUserDatabase.syncFreqIdx];
  span += span_cal;

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntTempComp_CalculateSpan: Temperature: %lu, Cal Temperature: %lu, Temperature delta: %ld\n"
                              "AdcIntTempComp_CalculateSpan: span delta: %ld, span drift: %ld, span: %ld, span cal @ freq %u: %ld\n",
                   (unsigned long) ass->adcTempCal.currentTemp, (unsigned long) ass->adcFactoryDatabase.calTemp, (long) temp_delta,
                   (long) span_delta, (long) span_drift, (long) span,
                   AdcSyncFrequencies[ass->adcUserDatabase.syncFreqIdx],
                   (long) span_cal);
#endif
  return span;
}


#if 0

/*
 * This is the original temperature compensation routine that
 * used self zero. Testing showed that better results were
 * achieved by just storing a zero offset calibration.
 *
 * If we ever add a analog switch to the front of the analog
 * section like the 5000, we'd need to re-introduce some
 * code like this to perform the zero error cancellation
 * (except we'd be using system zero, rather than self zero)
 */


/**
 * Run the temperature compensation
 *
 * Returns the number of DRDY cycles to skip before any further ADC operations can be performed
 */
static int AdcIntHwDoTempCompensation(AdcSuperStructure *ass)
{
  int32s zero;
  int drdy_skip_cycles = 0;
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  int32s check_span;
#endif

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal called, stage %u, start %u\n", ass->adcTempCal.stage, ass->adcTempCal.start);
#endif

  switch (ass->adcTempCal.stage)
  {
    case ADC_TEMP_CAL_STAGE_END_ZERO:
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
      ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: End Zero Stage\n");
#endif
      AdcIntHwTempCalFinishZero(ass);
      break;

    default:
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
      ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Default Stage\n");
#endif

      /* After 5seconds of samples, kick off another temperature calibration */
      if (ass->adcTempCal.stage > ass->adcTempCal.adcTempCalStagePeriod)
      {
        if (((!ass->adcDataWorking.grossRdg.info.status.bit.motion) ||
             (ass->adcTempCal.stage >= ass->adcTempCal.adcTempCalStagePerMax)) &&
            ((ass->adcRuntimeDatabase.tempCalOn) || (ass->adcTempCal.start)))
        {
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
          ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Calculating new temperature compensated span\n");
#endif
          /* Calculate and write new span */
          ass->adcTempCal.spanReg = AdcIntTempComp_CalculateSpan(ass);
          if (!AdcIntHwSetSpanRegister(ass->adcHwInfo.adc_rdg_dev, ass->adcTempCal.spanReg))
          {
            ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
          }
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
          (void) AdcIntHwGetSpanRegister(ass->adcHwInfo.adc_rdg_dev, &check_span);
          ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Check SpanReg:%ld\n", (long) check_span);
#endif

          /* Start zero calibration */
          if ((ass->adcUserDatabase.tempCalZero) || (ass->adcTempCal.start > 0))
          {
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
            ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Start zero self calibration\n");
#endif
            /* Start self zero calibration */
            if (!AdcIntHwPerformSelfZeroCalibration(ass->adcHwInfo.adc_rdg_dev))
            {
              ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
            }

            ass->adcTempCal.stage = 0;
            drdy_skip_cycles = 7 + 1;
          }
          else
          {
            /* Calculate and write new zero */
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
            ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Calculating zero compensation manually\n");
#endif
            zero = AdcIntTempComp_CalculateZero(ass);
            ass->adcHwInfo.adc_config_vars.zero_reg = zero;
            if (!AdcIntHwSetZeroRegister(ass->adcHwInfo.adc_rdg_dev, zero))
            {
              ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
            }
            ass->adcTempCal.stage = ADC_TEMP_CAL_STAGE_FINISHED - 1;
          }
        }
        else
        {
          // ass->adcTempCal.stage = ADC_TEMP_CAL_STAGE_FINISHED-1;
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
          ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Do nothing 1\n");
#endif
        }
      } /* if (ass->adcTempCal.stage > ass->adcTempCal.adcTempCalStagePeriod) */
      else
      {
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
        ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Do nothing 2\n");
#endif
      }
  } /* switch */

  /* Next stage */
  ass->adcTempCal.stage++;

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal finished\n");
#endif
  return drdy_skip_cycles;
}
#endif


/**
 * Run the temperature compensation
 *
 * Returns the number of DRDY cycles to skip before any further ADC operations can be performed
 *
 * This version doesn't do self zero, so there are no lost readings. We simply calculate
 * new zero and span values using the stored compensation values and current temperature
 * and set them in the ADC.
 *
 * This timing/state machine of this algorithm could be re-factored so it makes it easier
 * to understand exactly what is happening.
 *
 */
static int AdcIntHwDoTempCompensation(AdcSuperStructure *ass)
{
  int drdy_skip_cycles = 0;
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
  int32s check_span;
#endif

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal called, stage %u, start %u\n", ass->adcTempCal.stage, ass->adcTempCal.start);
#endif

  switch (ass->adcTempCal.stage)
  {
    case ADC_TEMP_CAL_STAGE_END_ZERO:
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
      ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: End Zero Stage\n");
#endif
      /* Nothing to do here */
      break;

    default:
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
      ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Default Stage\n");
#endif

      /* After 5seconds of samples, kick off another temperature calibration */
      if (ass->adcTempCal.stage > ass->adcTempCal.adcTempCalStagePeriod)
      {
        if (((!ass->adcDataWorking.grossRdg.info.status.bit.motion) ||
             (ass->adcTempCal.stage >= ass->adcTempCal.adcTempCalStagePerMax)) &&
            ((ass->adcRuntimeDatabase.tempCalOn) || (ass->adcTempCal.start)))
        {
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
          ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Calculating new temperature compensated span\n");
#endif
          /* Calculate and write new span */
          ass->adcTempCal.spanReg = AdcIntTempComp_CalculateSpan(ass);
          if (!AdcIntHwSetSpanRegister(ass->adcHwInfo.adc_rdg_dev, ass->adcTempCal.spanReg))
          {
            ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
          }
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_BASIC)
          (void) AdcIntHwGetSpanRegister(ass->adcHwInfo.adc_rdg_dev, &check_span);
          ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Check SpanReg:%ld\n", (long) check_span);
#endif

          /* Calculate and write new zero */
          AdcIntHwTempCalFinishZero(ass);
          ass->adcTempCal.stage = 0;
        }
        else
        {
          // ass->adcTempCal.stage = ADC_TEMP_CAL_STAGE_FINISHED-1;
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
          ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Do nothing 1\n");
#endif
        }
      } /* if (ass->adcTempCal.stage > ass->adcTempCal.adcTempCalStagePeriod) */
      else
      {
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
        ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal: Do nothing 2\n");
#endif
      }
      break;
  } /* switch */

  /* Next stage */
  ass->adcTempCal.stage++;

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "DoTempCal finished\n");
#endif
  return drdy_skip_cycles;
}


/**
 * Get a reading from the ADC hardware and run the temperature compensation
 *
 * The read will block until the ADC signals data is ready, hence this function
 * controls the rate at which the ADC thread runs.
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param rdg is a pointer to a reading which is set to the new reading value, not NULL
 * @param data_wait_failed is a pointer to a bool that is set if there was a read problem,
 *        and you should implement a delay elsewhere, to prevent the ADC thread going to 100% CPU, not NULL
 * @return true if new data acquired (not a skipped reading)
 */
bool AdcIntHwRead(AdcSuperStructure *ass, int32s *rdg, bool *data_wait_failed)
{
  int32s newrdg;
  bool new_data;

#if (ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL) || ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_FULL))
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll called\n");
#endif

  /* This will block until DRDY line signals new data (part of rread for dev_adc24bit) */
  if (!AdcIntHwGetAdcReading(ass, &newrdg, data_wait_failed))
  {
#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll: Failed to get new reading\n");
#endif
    return false;
  }
  *data_wait_failed = false;
  
  /* This is required because data at these frequencies is particularly low and we are limited in calibration range */
  if ((ass->adcHwInfo.adc_config_vars.sync_freq == AdcSyncFrequencies[ADC_SYNC_FREQ_10_HZ]) ||
      (ass->adcHwInfo.adc_config_vars.sync_freq == AdcSyncFrequencies[ADC_SYNC_FREQ_20_HZ]))
  {
    newrdg = newrdg * 26 / 25;
  }
  
#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll: Got new reading: %ld\n", (long) newrdg);
#endif

  /* Work out if we want to use the new reading, or the last good one */
  if (ass->adcTempCal.stage >= ADC_TEMP_CAL_STAGE_FINISHED)
  {
    /* if not temp cal'ing */
    ass->adcHwInfo.lastRdg = newrdg;

#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll: No temp cal, returning new data:%ld\n", (long) ass->adcHwInfo.lastRdg);
#endif
  }
#if ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL)
  else
  {
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll: Temp cal in progress, using last valid data:%ld\n", (long) ass->adcHwInfo.lastRdg);
  }
#endif

  /* Perform temperature calibration, if not skipping readings */
  if (0 == ass->adcHwInfo.skip)
  {
#if (ADC_DEBUG(ADC_TEMP_COMP_DEBUG_LVL, ADC_DEBUG_FULL) || ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_FULL))
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll: Calling DoTempCompensation()\n");
#endif
    ass->adcHwInfo.skip = AdcIntHwDoTempCompensation(ass);
  }

  /* Check that ADC is working, if not temperature calibration in progress */
  if (0 == ass->adcHwInfo.skip && ass->adcTempCal.stage > ADC_TEMP_CAL_STAGE_FINISHED)
  {
    if (!AdcIntHwGetAdcRunStatus(ass->adcHwInfo.adc_rdg_dev))
    {
      /* Not working, set error flag, re-init hardware */
#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_BASIC)
      ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll: ADC not working, re-initialise\n");
#endif
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
      AdcIntSetupHardware(ass);
    }
  }

  /* Return data */
  *rdg = ass->adcHwInfo.lastRdg;

  /* Decrement skip counter to zero */
  if (ass->adcHwInfo.skip > 0)
  {
#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_FULL)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll: Skip (%d) > 0, so skipping reading\n", ass->adcHwInfo.skip);
#endif
    ass->adcHwInfo.skip--;
    new_data = false;
  }
  else
  {
#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_FULL)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll: Skip (%d) == 0, so valid reading\n", ass->adcHwInfo.skip);
#endif
    new_data = true;
  }

#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwPoll finished\n");
#endif
  return new_data;
}


/** Reconfigure the ADC hardware with the configuration passed in cfg, can be called multiple times */
void AdcIntHwReconfigure(AdcSuperStructure *ass, int32s zero, int32s span, int16u sync_freq, bool four_wire)
{
  int32s rdg;
  bool data_wait_failed;

#if ADC_DEBUG(ADC_HARDWARE_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntHwReconfigure called: zero: %ld, span: %ld, sync: %u, 4wire: %u\n",
                   (long) zero, (long) span, sync_freq, four_wire);
#endif

  /* Setup temperature calibration */
  ass->adcTempCal.adcTempCalStagePeriod = (((5 * sync_freq) / 10) + ADC_TEMP_CAL_STAGE_FINISHED);  //5s
  ass->adcTempCal.adcTempCalStagePerMax = (((60 * sync_freq) / 10) + ADC_TEMP_CAL_STAGE_FINISHED); //60s
  ass->adcTempCal.stage = ass->adcTempCal.adcTempCalStagePerMax;

  /* Setup ADC hardware */
  ass->adcHwInfo.adc_config_vars.span_reg = span;
  ass->adcHwInfo.adc_config_vars.zero_reg = zero;
  ass->adcHwInfo.adc_config_vars.four_wire_mode = four_wire;
  ass->adcHwInfo.adc_config_vars.sync_freq = sync_freq;
  /* ass->adcHwInfo.adc_config_vars.prev_sync_freq; We don't set this here, it get set in AdcIntSetupHardware() */
  AdcIntSetupHardware(ass);

  /*
   * Test if LC connected.  This is only required in 4-wire mode.
   * 6-wire mode can use sense lines to detect a disconnect.
   */
  if (four_wire)
  {
    // turn the burn out current source ON
    if (!AdcIntHwSetBurnoutCurrentSource(ass->adcHwInfo.adc_rdg_dev, true))
    {
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
    }

    // give the current source time to take effect (100ms)
    OSTimeDly(OS_TICKS_PER_SEC * 100 / 1000);

    // if the output out of range
    (void) AdcIntHwGetAdcReading(ass, &rdg, &data_wait_failed);
    if (rdg >= ass->adcHwInfo.adcMax)
    {
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_ADC_MAXED;
    }
    else
    {
      ass->adcMiscellaneous.adcErrors &= ~ADC_ERR_ADC_MAXED;
    }

    // turn the burn out current source OFF
    if (!AdcIntHwSetBurnoutCurrentSource(ass->adcHwInfo.adc_rdg_dev, false))
    {
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
    }

    // give input time to recover (100ms)
    OSTimeDly(OS_TICKS_PER_SEC * 100 / 1000);
  }
}


/** Initialise the ADC hardware, call only once */
bool AdcIntHwInit(AdcHwInfo *hwinfo, struct t_io_dev *adcdev)
{
  bool result = false;

  /* Open ADC device if initialised, and not already open */
  if (NULL != adcdev &&
      0 != (adcdev->status & O_INITIALISED) &&
      0 != (adcdev->status & O_OPEN))
  {
    result = (ropen(adcdev, 0) >= 0);
  }

  /* Keep this device for all ADC operations */
  memset(hwinfo, 0, sizeof(AdcHwInfo));
  hwinfo->adc_rdg_dev = adcdev;
  hwinfo->adcMax = 0;
  hwinfo->lastRdg = 0;
  hwinfo->readingScale = 0;
  hwinfo->skip = 0;
  hwinfo->adc_config_vars.span_reg = 0;
  hwinfo->adc_config_vars.zero_reg = 0;
  hwinfo->adc_config_vars.four_wire_mode = 0;
  hwinfo->adc_config_vars.sync_freq = 0;
  hwinfo->adc_config_vars.prev_sync_freq = 0;

  return result;
}


/** Shutdown ADC hardware */
void AdcIntHwDestroy(AdcHwInfo *hwinfo)
{
  if (NULL != hwinfo->adc_rdg_dev &&
      0 != (hwinfo->adc_rdg_dev->status & O_INITIALISED) &&
      0 == (hwinfo->adc_rdg_dev->status & O_OPEN))
  {
    rclose(hwinfo->adc_rdg_dev);
  }
  hwinfo->adc_rdg_dev = NULL;
}
