#include <assert.h>
#include <string.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcreadings.h>
#include <rinweigh/adcdebug.h>

#include <rinweigh/adcaverage.h>


#if ADC_DEBUG(ADC_AVERAGE_DUMP_DEBUG_LVL, ADC_DEBUG_BASIC)
/** Dump the contents of the ADC averaging structure */
void AdcIntDebugDumpAverage(const AdcAverageData *ad)
{
  ADC_DEBUG_PRINTF(LOG_DEBUG, "ADC Averaging Data:\n");
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avMaxAv:      %u\n", ad->avMaxAv);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avPos:        %u\n", ad->avPos);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avCnt:        %u\n", ad->avCnt);
  /* ADC_DEBUG_PRINTF(LOG_DEBUG, "  avData:\n"); */
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avTotal:      %lld\n", ad->avTotal);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avNumRdgs:    %u\n", ad->avNumRdgs);

  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avPreAvTotal: %ld\n", (long) ad->preAvTotal);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avPreAvCnt:   %u\n", ad->preAvCnt);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avPreAv:      %u\n", ad->preAv);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avPreAvOn2:   %u\n", ad->preAvOn2);

  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avFull:       %u\n", ad->avFull);
  ADC_DEBUG_PRINTF(LOG_DEBUG, "  avValid:      %u\n", ad->avValid);
}
#endif /* #if ADC_DEBUG() */


/**
 * Initialise averaging structure with the number of readings to average
 *
 * @param ad is the averaging structure
 * @param numRdg is the number of readings to average, if greater than @ref ADC_AVERAGE_MAX_RDGS, then pre-averaging is employed to make up the extra averaging
 */
void AdcIntInitAverage(AdcAverageData *ad, int32u numRdg)
{
  int16u temp;

#if ADC_DEBUG(ADC_AVERAGE_INIT_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "InitAverageA called, numRdg=%u\n", numRdg);
#endif

  ad->avNumRdgs = (numRdg > 0) ? numRdg : 1;

  // allow for averaging more rdgs than buffer size allows
  ad->preAv = 1;
  temp = ad->avNumRdgs;
  while (temp > ADC_AVERAGE_MAX_RDGS)
  {
    ad->preAv++;
    temp -= ADC_AVERAGE_MAX_RDGS;
  }
  numRdg = ad->avNumRdgs / ad->preAv;

  ad->preAvOn2 = (int8u)((ad->preAv + 1) / 2);

  ad->avMaxAv = numRdg;

  AdcIntAverageClearData(ad);

#if ADC_DEBUG(ADC_AVERAGE_INIT_DEBUG_LVL, ADC_DEBUG_FULL)
  AdcIntDebugDumpAverage(ad);
#endif
}


/** Run the averaging process, taking new data rdg */
int32s AdcIntAverageReading(AdcAverageData *ad, int32s rdg)
{
  int64_t temp;

  // pre-averaging
  ad->preAvCnt++;
  ad->preAvTotal += rdg;
  if (ad->preAvCnt >= ad->preAv) 
  {
    if (ad->preAv > 1)
    {
      assert( ad->preAvCnt != 0 );
      rdg = (ad->preAvTotal + ad->preAvOn2) / ad->preAvCnt;
    }
    ad->avValid = TRUE;
    ad->preAvCnt = 0;
    ad->preAvTotal = 0;

    // averaging
    if (ad->avMaxAv == 1) 
    {
      ad->avTotal = rdg;
      ad->avCnt = 1;
    }
    else
    {
      ad->avTotal = ad->avTotal + rdg;
      if (ad->avCnt == ad->avMaxAv) ad->avTotal = ad->avTotal - ad->avData[ad->avPos];
      else ad->avCnt++;
      ad->avData[ad->avPos] = rdg;
      if (++ad->avPos >= ad->avMaxAv) 
      {  
        ad->avPos = 0;
        ad->avFull = TRUE;
      }
    }
  }

  // calculate new averaged data
  temp = ad->avTotal;
  if (ad->avMaxAv != 1)
  {
    /*
     * We have a special case when the filter has been re-initialised by AdcIntInitAverage,
     * and there is no data (avCnt == 0). This could lead to a divide by zero, so in
     * this case, we just return zero.
     */
    if (ad->avCnt == 0)
    {
      temp = 0;
    }
    else
    {
      temp /= ad->avCnt;
    }
  }
  return (int32s) temp;
}


/** Return true if average buffer is full (to configured limit) */
bool AdcIntAverageIsFull(AdcAverageData *ad)
{
  return ad->avFull;
}


/** Clear averaging buffer without changing averaging configuration */
void AdcIntAverageClearData(AdcAverageData *ad)
{
  ad->avPos = 0;
  ad->avCnt = 0;
  ad->avTotal = 0;
  ad->avValid = false;
  ad->avFull = false;
  ad->preAvTotal = 0;
  ad->preAvCnt = 0;

  /* Zero out data */
  memset(ad->avData, 0, sizeof(ad->avData));
}
