#include <assert.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcreadings.h>
#include <rinweigh/adcdlong.h>

#include <rinweigh/adcinterface.h>
#include <rinweigh/adcconvert.h>


/**
 * Convert internal ADC counts into final (displayable) numbers,
 * taking into account the resolution of the specified range,
 * and whether an extra digit of resolution is required.
 *
 * @param rdg is the reading in ADC counts
 * @param resolution is the current count by (1, 2, 5, 10, etc)
 * @param hiRes is a flag to indicate that the final number should have an extra digit of resolution
 * @return the rdg converted to final (displayable) numbers
 */
static int32s AdcIntCountsToFinalWithRes(int32s rdg, int32s resolution, bool hiRes)
{
  int32s div;

  /* Check for divide by zero */
  assert(resolution != 0);

  /* Work out divisor */
  div = 16 * resolution;

  /* Take account of hi resolution */
  if (hiRes)
  {
    rdg *= 10;
  }

  /* Rounding */
  if (rdg < 0)
  {
    rdg -= (div >> 1);
  }
  else
  {
    rdg += (div >> 1);
  }

  /* Do conversion */
  rdg /= div;
  rdg *= resolution;

  return rdg;
}


/**
 * Convert internal ADC counts into final (displayable) numbers,
 * taking into account the resolution of the specified range,
 * and whether an extra digit of resolution is required.
 *
 * @param ass is the ADC data super structure
 * @param rdg is the reading in ADC counts
 * @param range is the current range (0, 1, 2)
 * @param hiRes is a flag to indicate that the final number should have an extra digit of resolution
 * @param unit is the library units to set (secondary, tertiary etc)
 * @return the rdg converted to final (displayable) numbers
 */
int32s AdcIntCountsToFinal(const AdcSuperStructure *ass, int32s rdg, int8u range, bool hiRes, enum eAdcUnitSwitch unit)
{
  int32s resolution;
  int32s conv;

  /* Get conversion factor and resolution */
  switch (unit)
  {
    case ADC_UNIT_SWITCH_SECONDARY:
      conv = ass->adcInternalData.altConv;
      resolution = ass->adcInternalData.altRes[range];
      break;

    case ADC_UNIT_SWITCH_TERTIARY:
      conv = ass->adcInternalData.tertConv;
      resolution = ass->adcInternalData.tertRes[range];
      break;

    case ADC_UNIT_SWITCH_PRIMARY:
    default:
      conv = 0;
      resolution = (int32s)AdcResolutions[ass->adcUserDatabase.resolution[range]];
      break;
  }

  /* Do alternate units conversion if required */
  if (conv != 0)
    rdg = AdcMultDiv(rdg, conv, 1000000L);

  /* Check for divide by zero */
  assert(resolution != 0);

  /* Do conversion */
  return AdcIntCountsToFinalWithRes(rdg, resolution, hiRes);
}


/**
 * Convert final (displayable) numbers into internal ADC counts,
 * taking into account whether an extra digit of resolution is present.
 *
 * @param ass is the ADC data super structure
 * @param rdg is the final (displayable) number
 * @param hiRes is a flag to indicate that rdg has an extra digit of resolution
 * @param unit is the library units to set (secondary, tertiary etc)
 * @return the rdg converted to internal ADC counts
 */
int32s AdcIntFinalToCounts(const AdcSuperStructure *ass, int32s rdg, bool hiRes, enum eAdcUnitSwitch unit)
{
  int32s result;
  int32s conv;

  switch(unit)
  {
    case ADC_UNIT_SWITCH_SECONDARY:
      conv = ass->adcInternalData.altConv;
      break;

    case ADC_UNIT_SWITCH_TERTIARY:
      conv = ass->adcInternalData.tertConv;
      break;

    case ADC_UNIT_SWITCH_PRIMARY:
    default:
      conv = 0;
      break;
  }

  /* Do conversion */
  if (conv == 0)
    result = rdg * 16;
  else
    result = AdcMultDiv(rdg, 16000000L, conv);

  /* Factor in hi-res */
  if (hiRes)
  {
    result /= 10;
  }

  return result;
}


/**
 * Convert a final (displayable) number with (possibly) extra resolution
 * into a final (displayable) number with no extra resolution
 *
 * @param ass is the ADC data super structure
 * @param rdg is the final number with (possibly) extra resolution
 * @param range is the current range of the final number
 * @param hiRes is a flag to indicate that rdg has extra resolution
 * @return rdg with any extra resolution removed
 */
int32s AdcIntHiResToNormal(AdcSuperStructure *ass, int32s rdg, int8u range, bool hiRes)
{
  if (hiRes)
  {
    rdg = AdcIntFinalToCounts(ass, rdg, true, ADC_UNIT_SWITCH_PRIMARY);
    rdg = AdcIntCountsToFinal(ass, rdg, range, false, ADC_UNIT_SWITCH_PRIMARY);
  }
  return rdg;
}

