/*
 * adcdatafiller.c
 *
 *  Created on: Mar 31, 2012
 *      Author: ian
 */
#include <string.h>
#include <stdlib.h>
#include <rinweigh/adcmotion.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcreadings.h>
#include <rinweigh/adcdebug.h>
#include <rinweigh/adcconvert.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adclin.h>

#include <rinweigh/adcdatafiller.h>


/** Work out whether the alternate units conversion factors are valid */
bool AdcIntAlternateUnitsValid(const AdcInternalData *iad, enum eAdcUnitSwitch unit)
{
  int i;
  bool zerores;
  int32s res, conv;

  /* All resolutions must have non-zero resolutions */
  for (i = 0, zerores = false; i < ADC_MAX_NUM_INTERVAL; i++)
  {
    switch(unit)
    {
      case ADC_UNIT_SWITCH_SECONDARY:
        res = iad->altRes[i];
        break;

      case ADC_UNIT_SWITCH_TERTIARY:
        res = iad->tertRes[i];
        break;

      default:
        res = 0;
        break;
    }

    if (0 == res)
    {
      zerores = true;
    }
  }

  switch(unit)
  {
    case ADC_UNIT_SWITCH_SECONDARY:
      conv = iad->altConv;
      break;

    case ADC_UNIT_SWITCH_TERTIARY:
      conv = iad->tertConv;
      break;

    default:
      conv = 0;
      break;
  }

  /* And the conversion factor must be non-zero */
  return (0 != conv && !zerores);
}


/** Fill out heldrdg from rdg, and set appropriate status bits */
void AdcIntFillOutHeldReading(AdcRdg *heldrdg, const AdcRdg *rdg)
{
  *heldrdg = *rdg;
  heldrdg->info.status.bit.held = 1;
  heldrdg->info.status.bit.motion = 0;
}


/** Fill out AdcUnitDp structure */
void AdcIntFillOutUnitDp(const AdcSuperStructure *ass, AdcUnitDp *unitdp, enum eAdcUnitSwitch unit)
{
  unitdp->bit.dp = (ass->adcRuntimeDatabase.hiRes) ? (ass->adcUserDatabase.dpPos + 1) : ass->adcUserDatabase.dpPos;
  switch(unit)
  {
    case ADC_UNIT_SWITCH_SECONDARY:
      unitdp->bit.units = ass->adcInternalData.altUnits;
      unitdp->bit.dp += ass->adcInternalData.altDpInc;
      break;

    case ADC_UNIT_SWITCH_TERTIARY:
      unitdp->bit.units = ass->adcInternalData.tertUnits;
      unitdp->bit.dp += ass->adcInternalData.tertDpInc;
      break;

    case ADC_UNIT_SWITCH_PRIMARY:
    default:
      unitdp->bit.units = ass->adcUserDatabase.units;
      break;
  }
}


/** Fill out a tare reading */
void AdcIntFillOutTareReading(const AdcSuperStructure *ass, AdcRdg *tarerdg, int32s cntTare, bool tare_not_presettare, enum eAdcUnitSwitch unit)
{
  uint8_t range;

  /* Fill out unit dp */
  AdcIntFillOutUnitDp(ass, &tarerdg->info.unitdp, unit);

  /* Save the range before clearing status */
  range = tarerdg->info.status.bit.range;

  /* Fill out status */
  memset(&tarerdg->info.status, 0, sizeof(tarerdg->info.status));
  tarerdg->info.status.bit.tare = tare_not_presettare;
  tarerdg->info.status.bit.preTare = !tare_not_presettare;
  tarerdg->info.status.bit.hiRes = ass->adcRuntimeDatabase.hiRes;
  if (0 == ass->adcInternalData.rangeMax)
  {
    tarerdg->info.status.bit.multiRange = 0;
    tarerdg->info.status.bit.range = 0;
  }
  else
  {
    tarerdg->info.status.bit.multiRange = 1;
    tarerdg->info.status.bit.range = range;
  }

  /* Fill out reading */
  if (0 == cntTare)
    tarerdg->rdg = 0;
  else
    tarerdg->rdg = AdcIntCountsToFinal(ass, cntTare, tarerdg->info.status.bit.range, tarerdg->info.status.bit.hiRes, unit);
}


static void AdcIntFillOutAdcInfo(const AdcSuperStructure *ass,
                                 AdcInfo *info,
                                 int32s cntGross,
                                 int32s cntNet,
                                 int32s cntSysRdg,
                                 int32s cntOload,
                                 int8u previousRange)
{
  int32s absRdg;
  int8u rangeG, rangePT, rangeT, rangeN;

  /* Default all 0s */
  memset(info, 0, sizeof(AdcInfo));

  /*
   * Fill out unitdp & id
   */
  info->id = 0;
  AdcIntFillOutUnitDp(ass, &info->unitdp, ADC_UNIT_SWITCH_PRIMARY);

  /*
   * Status other than range
   */

  /* Motion */
  info->status.bit.motion = AdcIntMotionDetectionGetMotion(&ass->adcMotionDetection);

  /* Zero band */
  if ((cntGross >= -ass->adcInternalData.cntZeroBandMax) && (cntGross <= ass->adcInternalData.cntZeroBandMax))
    info->status.bit.zeroBand = 1;
  else
    info->status.bit.zeroBand = 0;

  /* Zero range */
  if ((cntSysRdg >= ass->adcInternalData.cntZeroRangeMin) && (cntSysRdg <= ass->adcInternalData.cntZeroRangeMax))
    info->status.bit.zeroRange = 1;
  else
    info->status.bit.zeroRange = 0;

  /* Center of zero */
  if (ADC_APPROVAL_MODE_NTEP == ass->adcInternalData.approvalMode && ass->adcRuntimeDatabase.netMode)
    absRdg = labs(cntNet);
  else
    absRdg = labs(cntGross);
  if (absRdg <= ass->adcInternalData.cntCOZ)
    info->status.bit.coz = 1;
  else
    info->status.bit.coz = 0;

  /* Overload */
  if (cntGross > cntOload)
    info->status.bit.oload = 1;
  else
    info->status.bit.oload = 0;

  /* Underload */
  if (cntGross < ass->adcInternalData.cntUload)
    info->status.bit.uload = 1;
  else
    info->status.bit.uload = 0;

  /* Tare and preset tare */
  info->status.bit.tare = 0;
  info->status.bit.preTare = 0;

  /* Held */
  info->status.bit.held = 0;

  /* Gross or net */
  info->status.bit.grossOrNet = 1;
  info->status.bit.gross = 1;

  /* Hi-res */
  info->status.bit.hiRes = ass->adcRuntimeDatabase.hiRes;

  /* Grad position */
  info->status.bit.gradPos = 0;

  /*
   * Range
   */

  /* Calculate current range of Preset Tare, Tare and Gross weights given no history */
  rangePT = AdcIntGetRangeForCounts(ass, ass->adcRuntimeDatabase.cntPTare);
  rangeT = AdcIntGetRangeForCounts(ass, ass->adcRuntimeDatabase.cntTare);
  rangeG = AdcIntGetRangeForCounts(ass, cntGross);
  rangeN = AdcIntGetRangeForCounts(ass, cntNet);

  /* Single range */
  if (0 == ass->adcInternalData.rangeMax)
  {
    info->status.bit.range = 0;
    info->status.bit.multiRange = 0;
    return;
  }

  /* Multi-interval */
  info->status.bit.multiRange = 1;
  if (!ass->adcInternalData.rangeMulti)
  {
    if (ass->adcRuntimeDatabase.netMode)
      info->status.bit.range = rangeN;
    else
      info->status.bit.range = rangeG;
    return;
  }

  /* Multi-range */
  info->status.bit.range = previousRange;
  if ((info->status.bit.zeroBand) && (!info->status.bit.motion))
    info->status.bit.range = rangePT;
  if (rangeG > info->status.bit.range)
    info->status.bit.range = rangeG;
  if (rangePT > info->status.bit.range)
    info->status.bit.range = rangePT;
  if (rangeT > info->status.bit.range)
    info->status.bit.range = rangeT;
}


/** Calculate whether the scale is overloaded, when the additive tare is applied */
static bool AdcIntCalculateAdditiveTareOverload(const AdcSuperStructure *ass)
{
  int32s cntAnyTare;
  int32s cntAddTare;
  bool overloaded;

  cntAnyTare = ass->adcRuntimeDatabase.cntPTare + ass->adcRuntimeDatabase.cntTare;
  cntAddTare = AdcIntFinalToCounts(ass, ass->adcUserDatabase.addTare, false, ADC_UNIT_SWITCH_PRIMARY);

  if (cntAnyTare >= 0)
  {
    /* Positive tare, in net mode, overload is pushed out by the tare value, upto a maximum of the additive tare value. */
    if (cntAnyTare <= cntAddTare)
    {
      /* Push out overload limit by tare value */
      overloaded = (ass->adcInternalData.cntGross > ass->adcInternalData.cntOload + cntAnyTare);
    }
    else
    {
      /* Push out overload limit by upper limit (additive tare value) */
      overloaded = (ass->adcInternalData.cntGross > ass->adcInternalData.cntOload + cntAddTare);
    }
  }
  else
  {
    /* Negative tare, in net mode, overload is reduced out by the tare value, upto a maximum of the additive tare value. */
    cntAnyTare = -cntAnyTare;
    if (cntAnyTare <= cntAddTare)
    {
      /* Reduce overload limit by tare value */
      overloaded = (ass->adcInternalData.cntGross > ass->adcInternalData.cntOload - cntAnyTare);
    }
    else
    {
      /* Push out overload limit by upper limit (additive tare value) */
      overloaded = (ass->adcInternalData.cntGross > ass->adcInternalData.cntOload - cntAddTare);
    }
  }
  return overloaded;
}


/** Calculate whether the scale is underloaded, when the additive tare is applied */
static bool AdcIntCalculateAdditiveTareUnderload(const AdcSuperStructure *ass)
{
  int32s cntAnyTare;
  int32s cntAddTare;
  bool underloaded;

  cntAnyTare = ass->adcRuntimeDatabase.cntPTare + ass->adcRuntimeDatabase.cntTare;
  cntAddTare = AdcIntFinalToCounts(ass, ass->adcUserDatabase.addTare, false, ADC_UNIT_SWITCH_PRIMARY);

  if (cntAnyTare < 0)
  {
    /* Negative tare, in net mode, underload is pushed out by the tare value, upto a maximum of the additive tare value. */
    cntAnyTare = -cntAnyTare;
    if (cntAnyTare <= cntAddTare)
    {
      /* Push out underload limit by tare value */
      underloaded = (ass->adcInternalData.cntGross < ass->adcInternalData.cntUload - cntAnyTare);
    }
    else
    {
      /* Push out overload limit by upper limit (additive tare value) */
      underloaded = (ass->adcInternalData.cntGross < ass->adcInternalData.cntUload - cntAddTare);
    }
  }
  else
  {
    /* Positive tare, in net mode, underload is reduced out by the tare value, upto a maximum of the additive tare value. */
    if (cntAnyTare <= cntAddTare)
    {
      /* Reduce overload limit by tare value */
      underloaded = (ass->adcInternalData.cntGross < ass->adcInternalData.cntUload + cntAnyTare);
    }
    else
    {
      /* Push out overload limit by upper limit (additive tare value) */
      underloaded = (ass->adcInternalData.cntGross < ass->adcInternalData.cntUload + cntAddTare);
    }
  }
  return underloaded;
}


/** Calculate grad position */
static int8s AdcIntCalculateGradPos(const AdcSuperStructure *ass, int32s rdg, int32s cntRdg, int8u range)
{
  int8s gradpos;
  int8u resolution;

  /* Work out resolution for range */
  resolution = AdcResolutions[ass->adcUserDatabase.resolution[range]];

  /* Calculate grad position */
  gradpos = 0x0F & ((cntRdg - (rdg * 16)) / resolution);
  return gradpos;
}


/**
 * Fill out the ADC data structure
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param ad is a pointer to the ADC data structure to be filled out, not NULL
 * @param ai is a pointer to an ADC info structure to be filled out, not NULL
 * @param rawRdg is the raw ADC reading (post averaging)
 * @param cntSysRdg is the rawRdg in counts minus the initial zero. Calculate with @ref AdcIntCalculateCounts
 * @param cntGross is the gross weight in counts. Calculate with @ref AdcIntCalculateCounts
 * @param cntNet is the net weight in counts. Calculate with @ref AdcIntCalculateCounts
 * @param cntOload is the overload weight. (In NTEP this changes with range)
 *
 */
void AdcIntFillOutAdcData(const AdcSuperStructure *ass,
                          AdcData *ad,
                          AdcInfo *ai,
                          int32s rawRdg,
                          int32s cntSysRdg,
                          int32s cntGross,
                          int32s cntNet,
                          int32s cntOload,
                          int8u previousRange)
{
  int32s rdg, rdgB;

#if ADC_DEBUG(ADC_CALCULATION_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntFillOutAdcData called\n");
#endif

  /* Fill out raw rdg */
  ad->rawRdg = rawRdg;

  /* Fill out mvvAbsRdg */
  if (ad->rawRdg >= 0)
    ad->mvvAbsRdg = (ad->rawRdg + 128) / 256;
  else
    ad->mvvAbsRdg = (ad->rawRdg - 128) / 256;

  // update status information
#if ADC_DEBUG(ADC_CALCULATION_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntFillOutAdcData: update status\n");
#endif
  AdcIntFillOutAdcInfo(ass,
                       ai,
                       cntGross,
                       cntNet,
                       cntSysRdg,
                       cntOload,
                       previousRange);

  /* Do displayed preset tare, user tare and combined */
  ad->presetTare.info = ad->userTare.info = *ai;
  AdcIntFillOutTareReading(ass, &ad->presetTare, ass->adcRuntimeDatabase.cntPTare, false, ADC_UNIT_SWITCH_PRIMARY);
  AdcIntFillOutTareReading(ass, &ad->userTare, ass->adcRuntimeDatabase.cntTare, true, ADC_UNIT_SWITCH_PRIMARY);
  if (0 != ad->presetTare.rdg)
    ad->anyTare = ad->presetTare;
  else
    ad->anyTare = ad->userTare;

  // calculate final weights
  if (ass->adcRuntimeDatabase.netMode && (0 == ass->adcRuntimeDatabase.cntPTare))
  {
    // calculate final weights
    ad->netRdg.rdg = AdcIntCountsToFinal(ass, cntNet, ai->status.bit.range, ai->status.bit.hiRes, ADC_UNIT_SWITCH_PRIMARY);
    ad->grossRdg.rdg = ad->netRdg.rdg + ad->anyTare.rdg;
    ad->grossRdg.info = ad->netRdg.info = *ai;
    ad->netRdg.info.status.bit.gross = 0;
    ad->netRdg.info.status.bit.oload = AdcIntCalculateAdditiveTareOverload(ass);
    ad->netRdg.info.status.bit.uload = AdcIntCalculateAdditiveTareUnderload(ass);
    if ((ass->adcInternalData.rangeMax) && (!ass->adcInternalData.rangeMulti))
      ad->netRdg.info.status.bit.range = AdcIntGetRangeForCounts(ass, cntNet);
  }
  else
  {
    // calculate final weights
    ad->grossRdg.rdg = AdcIntCountsToFinal(ass, cntGross, ai->status.bit.range, ai->status.bit.hiRes, ADC_UNIT_SWITCH_PRIMARY);
    ad->netRdg.rdg = ad->grossRdg.rdg - ad->anyTare.rdg;
    ad->grossRdg.info = ad->netRdg.info = *ai;
    ad->netRdg.info.status.bit.gross = 0;
    ad->netRdg.info.status.bit.oload = AdcIntCalculateAdditiveTareOverload(ass);
    ad->netRdg.info.status.bit.uload = AdcIntCalculateAdditiveTareUnderload(ass);
    if ((ass->adcInternalData.rangeMax) && (!ass->adcInternalData.rangeMulti))
      ad->netRdg.info.status.bit.range = AdcIntGetRangeForCounts(ass, cntNet);
  }

  /* Do displayed reading */
  if (ass->adcRuntimeDatabase.netMode)
  {
    ad->dispRdg = ad->netRdg;
    ad->notDispRdg = ad->grossRdg;
  }
  else
  {
    ad->dispRdg = ad->grossRdg;
    ad->notDispRdg = ad->netRdg;
  }

  /* Only calculate pieces if configured, prevent divide by zero */
  memset(&ad->dispPieceRdg, 0, sizeof(ad->dispPieceRdg));
  ad->dispPieceRdg.info.unitdp.bit.dp = 0;
  ad->dispPieceRdg.info.unitdp.bit.units = ADC_UNITS_PIECES;
  if (0 == ass->adcInternalData.sampleNum || 0 == ass->adcInternalData.cntSampleWgt)
  {
    ad->dispPieceRdg.rdg = 0;
  }
  else
  {
    /* Calculate pieces (with rounding) */
    if (ass->adcRuntimeDatabase.netMode)
    {
      ad->dispPieceRdg.rdg = AdcMultDiv(ass->adcInternalData.cntNet*16, ass->adcInternalData.sampleNum, ass->adcInternalData.cntSampleWgt);
    }
    else
    {
      ad->dispPieceRdg.rdg = AdcMultDiv(ass->adcInternalData.cntGross*16, ass->adcInternalData.sampleNum, ass->adcInternalData.cntSampleWgt);
    }
  }
  if (ad->dispPieceRdg.rdg > 0)
    ad->dispPieceRdg.rdg += 8;
  else
    ad->dispPieceRdg.rdg -= 8;
  ad->dispPieceRdg.rdg /= 16;

  /* Calculate alternative units status */
  ad->altGrossRdg.info = ad->grossRdg.info;
  ad->altNetRdg.info = ad->netRdg.info;
  ad->altGrossRdg.info.unitdp.bit.units = ad->altNetRdg.info.unitdp.bit.units = ass->adcInternalData.altUnits;
  ad->altGrossRdg.info.status.bit.gradPos = ad->altNetRdg.info.status.bit.gradPos = 0;
  ad->altGrossRdg.info.unitdp.bit.dp += ass->adcInternalData.altDpInc;
  ad->altNetRdg.info.unitdp.bit.dp += ass->adcInternalData.altDpInc;
  ad->altAnyTare.info = ad->altGrossRdg.info;

  /* Calculate alternative units readings */
  if (AdcIntAlternateUnitsValid(&ass->adcInternalData, ADC_UNIT_SWITCH_SECONDARY))
  {
    /* This can only happen if the alternate conversion units have been configured */
    ad->altGrossRdg.rdg = AdcIntCountsToFinal(ass,
                                              cntGross,
                                              ad->altGrossRdg.info.status.bit.range,
                                              ass->adcRuntimeDatabase.hiRes,
                                              ADC_UNIT_SWITCH_SECONDARY);
    if (0 != ass->adcRuntimeDatabase.cntPTare)
    {
      AdcIntFillOutTareReading(ass, &ad->altAnyTare, ass->adcRuntimeDatabase.cntPTare, false, ADC_UNIT_SWITCH_SECONDARY);
    }
    else
    {
      AdcIntFillOutTareReading(ass, &ad->altAnyTare, ass->adcRuntimeDatabase.cntTare, true, ADC_UNIT_SWITCH_SECONDARY);
    }
    ad->altNetRdg.rdg = ad->altGrossRdg.rdg - ad->altAnyTare.rdg;
  }
  else
  {
    /* Not valid, so all zero */
    ad->altGrossRdg.rdg = 0;
    ad->altNetRdg.rdg = 0;
    AdcIntFillOutTareReading(ass, &ad->altAnyTare, 0, true, ADC_UNIT_SWITCH_SECONDARY);
  }

  /* Do alternative units displayed reading */
  if (ass->adcRuntimeDatabase.netMode)
    ad->altDispRdg = ad->altNetRdg;
  else
    ad->altDispRdg = ad->altGrossRdg;

  /* Calculate tertiary units status */
  ad->tertGrossRdg.info = ad->grossRdg.info;
  ad->tertNetRdg.info = ad->netRdg.info;
  ad->tertGrossRdg.info.unitdp.bit.units = ad->tertNetRdg.info.unitdp.bit.units = ass->adcInternalData.tertUnits;
  ad->tertGrossRdg.info.status.bit.gradPos = ad->tertNetRdg.info.status.bit.gradPos = 0;
  ad->tertGrossRdg.info.unitdp.bit.dp += ass->adcInternalData.tertDpInc;
  ad->tertNetRdg.info.unitdp.bit.dp += ass->adcInternalData.tertDpInc;
  ad->tertAnyTare.info = ad->tertGrossRdg.info;

  /* Calculate tertiary units */
  if (AdcIntAlternateUnitsValid(&ass->adcInternalData, ADC_UNIT_SWITCH_TERTIARY))
  {
    /* This can only happen if the tertiary conversion units have been configured */
    ad->tertGrossRdg.rdg = AdcIntCountsToFinal(ass,
                                               cntGross,
                                               ad->altGrossRdg.info.status.bit.range,
                                               ass->adcRuntimeDatabase.hiRes,
                                               ADC_UNIT_SWITCH_TERTIARY);
    if (0 != ass->adcRuntimeDatabase.cntPTare)
    {
      AdcIntFillOutTareReading(ass, &ad->tertAnyTare, ass->adcRuntimeDatabase.cntPTare, false, ADC_UNIT_SWITCH_TERTIARY);
    }
    else
    {
      AdcIntFillOutTareReading(ass, &ad->tertAnyTare, ass->adcRuntimeDatabase.cntTare, true, ADC_UNIT_SWITCH_TERTIARY);
    }
    ad->tertNetRdg.rdg = ad->tertGrossRdg.rdg - ad->tertAnyTare.rdg;
  }
  else
  {
    /* Not valid, so all zero */
    ad->tertGrossRdg.rdg = 0;
    ad->tertNetRdg.rdg = 0;
    AdcIntFillOutTareReading(ass, &ad->tertAnyTare, 0, true, ADC_UNIT_SWITCH_TERTIARY);
  }

  /* Do tertiary units displayed reading */
  if (ass->adcRuntimeDatabase.netMode)
    ad->tertDispRdg = ad->tertNetRdg;
  else
    ad->tertDispRdg = ad->tertGrossRdg;

  /* Set display mode for sample */
  ad->dispMode = ass->adcRuntimeDatabase.dispMode;

  /* Calculate position in grad, can only be done after we have gross, cntGross and range */
  ad->grossRdg.info.status.bit.gradPos = AdcIntCalculateGradPos(ass, ad->grossRdg.rdg, cntGross, ai->status.bit.range);
  ad->netRdg.info.status.bit.gradPos = AdcIntCalculateGradPos(ass, ad->netRdg.rdg, cntNet, ai->status.bit.range);

  /* Error, reading id, and data valid */
  ad->error = (ass->adcMiscellaneous.adcErrors & ADC_ERR_MASK);
  ad->rdgId++;
  ad->dataValid = true;

  /* Calculate peak values */
  if (ad->grossRdg.rdg > 0)
    rdg = ad->grossRdg.rdg;
  else
    rdg = -ad->grossRdg.rdg;
  if (ad->peakAbsGrossRdg.rdg > 0)
    rdgB = ad->peakAbsGrossRdg.rdg;
  else
    rdgB = -ad->peakAbsGrossRdg.rdg;
  if (rdg > rdgB)
  {
    AdcIntFillOutHeldReading(&ad->peakAbsGrossRdg, &ad->grossRdg);
  }
  if (ad->peakMaxDispRdg.rdg < ad->dispRdg.rdg)
  {
    AdcIntFillOutHeldReading(&ad->peakMaxDispRdg, &ad->dispRdg);
  }
  if (ad->peakMinDispRdg.rdg > ad->dispRdg.rdg)
  {
    AdcIntFillOutHeldReading(&ad->peakMinDispRdg, &ad->dispRdg);
  }

#if ADC_DEBUG(ADC_CALCULATION_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcIntFillOutAdcData: finished successfully\n");
#endif
}


/**
 * Calculate various counts from raw reading
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param rawRdg is the raw ADC reading (post averaging)
 * @param cntSysRdg is the rawRdg in counts minus the initial zero
 * @param cntGross is the gross weight in counts
 * @param cntNet is the net weight in counts
 */
void AdcIntCalculateCounts(const AdcSuperStructure *ass, int32s rawRdg,
                           int32s *cntSysRdg, int32s *cntGross, int32s *cntNet)
{
  int32s cntRdg;
  int32s cntRdgWZero;

  /* Convert to counts */
  cntRdg = AdcMultDiv(ass->adcUserDatabase.spanCnt, rawRdg - ass->adcUserDatabase.zeroRaw, ass->adcUserDatabase.spanRaw);

  /* Calculate readings in counts */
  *cntSysRdg = cntRdg - ass->adcInternalData.cntInitZero;
  cntRdgWZero = *cntSysRdg - ass->adcRuntimeDatabase.cntZero;
  *cntGross = cntRdgWZero + AdcIntLinearisationCorrection(&ass->adcLinearisation, cntRdgWZero);
  *cntNet = *cntGross - ass->adcRuntimeDatabase.cntTare - ass->adcRuntimeDatabase.cntPTare;
}


/** Set the held bit in all AdcInfo structures within AdcData */
void AdcIntAdcDataSetHeld(AdcData *adcData, bool held)
{
  #define SET_HELD(rdg) (adcData->rdg.info.status.bit.held = held)

  SET_HELD(dispRdg);
  SET_HELD(notDispRdg);
  SET_HELD(grossRdg);
  SET_HELD(netRdg);
  SET_HELD(altGrossRdg);
  SET_HELD(altNetRdg);
  SET_HELD(altDispRdg);
  SET_HELD(tertGrossRdg);
  SET_HELD(tertNetRdg);
  SET_HELD(tertDispRdg);

  #undef SET_HELD
}

