/**
 * @file adcdatatypes.h
 * @brief Datatypes for the R400 ADC Library
 *
 * Company: Rinstrum Australia
 * Copyright: This code is the property of Rinstrum. Unauthorised copying or use is
 * an offence under international and national law.
 *
 * Overview:
 * This is the interface header for the ADC library written for the R4xx.
 * The ADC library talks to the ADC chip directly and performs all processing
 * to obtain OIML approved weight values.
 *
 * Data structures and defines required to interface to the ADC library can be
 * found in this header file.  Together with adcInterface.h, this file provides
 * a fairly high level interface to the adc library.
 *
 * This header file is broken into x sections:
 *  - AdcRdg structure: 8 byte reading with units, dp and OIML status
 *  - AdcData structure: external view of all volatile adc data
 *  - AdcTraceableWeight: result of an operation to get a traceable weight
 *  - InternAdcData: data used internally by the adc library which is exported for
 *    debug purposes.
 *  - Adc database structures: The adc database is broken into 3 sections
 *  - AdcFactoryDatabase: resides in BNVFAC RAM section
 *  - AdcUserDatabase: resides in BNVUSER RAM section
 *  - AdcRuntimeDatabase: resides in BNVRT RAM section
 */
#ifndef ADC_DATATYPES_DOT_H
#define ADC_DATATYPES_DOT_H 1

#include <stddef.h>
#include <stdint.h>
#include <stdarg.h>
#include <time.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rtc.h>
#include <rinweigh/adcmuteximpl.h>
#include <rinweigh/adcliveimpl.h>


/* Forward declarations */
struct sAdcSuperStructure;
struct sAdcFactoryDatabase;
struct sAdcUserDatabase;
struct sAdcRuntimeDatabase;


/** Structure containing all ADC information */
typedef struct sAdcSuperStructure AdcSuperStructure;


/** The ADC factory database. This is filled in by the Rinstrum oven factory calibration process, and stored in flash. */
typedef struct sAdcFactoryDatabase AdcFactoryDatabase;


/** ADC user calibration database. The first 5 members of this structure must match @ref AdcUserCalSection. */
typedef struct sAdcUserDatabase AdcUserDatabase;


/** The ADC runtime database. Values in this structure change frequently (such as tare values). */
typedef struct sAdcRuntimeDatabase AdcRuntimeDatabase;


///////////////////////////////////////////////////////////////////////////////////////
// AdcRdg: An 8 byte structure holding a weight including all OIML information
///////////////////////////////////////////////////////////////////////////////////////

/** Type containing ADC reading status */
typedef union {
  int32u dword;                 /*!< Status represented as a double word */
  struct {
    int8s gradPos   :4;         /*!< The position within a graduation? */
    int8u grossOrNet:1;         /*!< If set, the reading is either gross or net */
    int8u hiRes     :1;         /*!< If set, the reading has extra resolution (ie *10) */
    int8u range     :2;         /*!< The range of the reading (0=range1, 1=range 2, etc */

    int8u held      :1;         /*!< If set, the reading is held */
    int8u zeroRange :1;         /*!< If set, the reading is within the zero range (trade: zero key operates in this range) */
    int8u zeroBand  :1;         /*!< If set, the reading is within the zero band (non-trade: band around zero used for printing/setpoints, etc) */
    int8u motion    :1;         /*!< If set, the reading is unstable */
    int8u coz       :1;         /*!< If set, the reading is with center of zero */
    int8u gross     :1;         /*!< If set, the reading is a gross value, net otherwise */
    int8u uload     :1;         /*!< If set, the reading is overloaded */
    int8u oload     :1;         /*!< If set, the reading is underloaded */

    int8u tare      :1;         /*!< If set, the reading is a tare weight */
    int8u preTare   :1;         /*!< If set, the reading is a preset tare weight */
    int8u multiRange:1;         /*!< If set, the reading is either multi range or multi interval (ie scale is not single range) */

  } bit;                        /*!< Status represented as individual bits */
} AdcStatus;


/** Type containing ADC reading units and decimal point position */
typedef union {
  int8u byte;                   /*!< Units and DP represented as a byte */
  struct {
    int8u units :4;             /*!< The units of the reading (eg g, kg, lb) */
    int8u dp    :4;             /*!< The decimal point position of the reading */
  } bit;                        /*!< Units and DP represented as a individual bits */
} AdcUnitDp;


/** Type containing ADC reading status, units and decimal point and a reading ID */
typedef struct {
  int8u id;                     /*!< Reading identifier */
  AdcUnitDp unitdp;             /*!< Reading units and decimal point */
  AdcStatus status;             /*!< Reading status */
} AdcInfo;


/** Type containing a reading and associated information */
typedef struct {
  int32s rdg;                   /*!< Reading value */
  AdcInfo info;                 /*!< Reading information (status, units, etc) */
} AdcRdg;


///////////////////////////////////////////////////////////////////////////////////////
// AdcData: A complete snapshot of adc data updated at the sync frequency
///////////////////////////////////////////////////////////////////////////////////////

/** Error type */
typedef int32u AdcErrors;

// AdcData.error
#define ADC_ERR_TEMPERATURE 0x00000010            /*!< Error flag indicating the ADC temperature reading errors. See @ref AdcData::error. */
#define ADC_ERR_FACTORY_DB  0x00000400            /*!< Error flag indicating the ADC factory database/calibration is corrupt. See @ref AdcData::error. */
#define ADC_ERR_USER_DB     0x00000200            /*!< Error flag indicating the ADC user database/calibration is corrupt. See @ref AdcData::error. */
#define ADC_ERR_ADC_MAXED   0x00002000            /*!< Error flag indicating the ADC input is maxxed out. See @ref AdcData::error. */
#define ADC_ERR_RUNTIME_DB  0x00004000            /*!< Error flag indicating the ADC runtime database is corrupt. See @ref AdcData::error. */
#define ADC_ERR_HARDWARE    0x00010000            /*!< Error flag indicating the ADC hardware is faulty. See @ref AdcData::error. */

#define ADC_ERR_RESET_MASK  0x00002000            /*!< Inverted bitmask for all which ADC errors can be reset in software. See @ref AdcData::error. */

/** Bitmask for all ADC errors. See @ref AdcData::error. */
#define ADC_ERR_MASK        (ADC_ERR_FACTORY_DB | \
                             ADC_ERR_USER_DB | \
                             ADC_ERR_ADC_MAXED | ADC_ERR_RUNTIME_DB | \
                             ADC_ERR_TEMPERATURE | ADC_ERR_HARDWARE)

#define APP_ERR_PWR         0x0003                /*!< Error mask for indicating the application is reporting bad input power. See @ref AdcData::error. */
#define APP_ERR_SENSE       0x000C                /*!< Error mask for indicating the application is reporting the sense lines are out of range. See @ref AdcData::error. */
#define APP_ERR_MASK        0x000F                /*!< Bitmask for all application errors. See @ref AdcData::error. */

#define ADC_ERRORS           (ADC_ERR_MASK|APP_ERR_MASK)       /*!< Bitmask for all application and ADC errors. See @ref AdcData::error. */
#define ADC_TEMPORARY_ERRORS (ADC_ERR_ADC_MAXED|APP_ERR_MASK)  /*!< Bitmask for all temporary errors (ones beyond the control of software). See @ref AdcData::error. */


/** Reading display mode */
enum eAdcDisplayMode { ADC_DISPLAY_MODE_WEIGHT = 0,  /*!< Normal weight display mode. See @ref AdcData::dispMode. */
                       ADC_DISPLAY_MODE_PIECES,      /*!< Piece counting display mode. See @ref AdcData::dispMode. */
                       ADC_DISPLAY_MODE_ALT_W,       /*!< Alternate units display mode. See @ref AdcData::dispMode. */
                       ADC_DISPLAY_MODE_TERT_W,      /*!< Tertiary units display mode. See @ref AdcData::dispMode. */
                       ADC_NUM_DISPLAY_MODE          /*!< The number of display modes. Used for putting limits on display mode. See @ref AdcData::dispMode. */
                     };


/** Structure containing all current weight information */
typedef struct
{
  bool dataValid;               /*!< Set if the data in the structure is valid */
  int32s rawRdg;                /*!< The raw reading in millivolts per volt (divide this number by (10000 * 256) to get actual mV/V) */
  int32s mvvAbsRdg;             /*!< The millivolt per volt absolute reading (divide this number by 10000 to get actual mV/V) */
  AdcRdg dispRdg;               /*!< The displayed gross or net rdg in calibrated units */
  AdcRdg notDispRdg;            /*!< The non-displayed weight (ie gross, if net is currently displayed */
  AdcErrors error;              /*!< Any errors, refer to the ADC_ERR_* flags */
  int32u rdgId;                 /*!< Reading identifier, incremented every new reading */
  AdcRdg grossRdg;              /*!< The weight as a gross reading */
  AdcRdg netRdg;                /*!< The weight as a net reading */
  AdcRdg peakAbsGrossRdg;       /*!< The peak absolute gross reading, used in peak hold mode */
  AdcRdg peakMaxDispRdg;        /*!< The peak maximum displayed reading, used in peak hold mode */
  AdcRdg peakMinDispRdg;        /*!< The peak minimum displayed reading, used in peak hold mode */
  AdcRdg presetTare;            /*!< The current preset tare value */
  AdcRdg userTare;              /*!< The current user tare value */
  AdcRdg anyTare;               /*!< The sum of the user and preset tares */
  int8u  dispMode;              /*!< How the data should be displayed. Refer to the @enum eAdcDisplayMode. */
  AdcRdg dispPieceRdg;          /*!< The displayed number of pieces, when in piece counting mode */
  AdcRdg altGrossRdg;           /*!< The alternate weight units gross reading */
  AdcRdg altNetRdg;             /*!< The alternate weight units net reading */
  AdcRdg altDispRdg;            /*!< The alternate weight units displayed reading */
  AdcRdg altAnyTare;            /*!< Sum of any tare(s) present for the alternate weight units */
  AdcRdg tertGrossRdg;          /*!< The tertiary weight units gross reading */
  AdcRdg tertNetRdg;            /*!< The tertiary weight units net reading */
  AdcRdg tertDispRdg;           /*!< The tertiary weight units displayed reading */
  AdcRdg tertAnyTare;           /*!< Sum of any tare(s) present for the tertiary weight units */
} AdcData;


///////////////////////////////////////////////////////////////////////////////////////
// AdcTraceableWeight: Datastructure holding a traceable weight
///////////////////////////////////////////////////////////////////////////////////////

/** Data structure holding a traceable weight */
typedef struct
{
  int32u printId;               /*!< The unique identifier for the traceable weight */
  AdcRdg rdg;                   /*!< The traceable weight reading */
  AdcRdg altRdg;                /*!< The traceable weight reading in alternate units */
  int32s pieces;                /*!< The traceable weight reading as pieces */
  int32s tareWgt;               /*!< The tare weight set for the traceable weight reading  */
  bool   pTare;                 /*!< Flag if the tare is a preset tare */
  bool   valid;                 /*!< Flag if the traceable weight is valid */
  t_datetime dt;                /*!< The timestamp of the traceable weight */
  AdcRdg tertRdg;               /*!< The traceable weight reading in tertiary units */
} AdcTraceableWeight;


/*
 * ADC Internal Datastructures
 */


/*
 * ADC Hardware
 */


/** Data structure for variable configuration settings of adc */
typedef struct
{
  int32s span_reg;                    /*!< Current span register value */
  int32s zero_reg;                    /*!< Current zero register value */
  int8u four_wire_mode;               /*!< ==0 for 6 wire, 4 wire otherwise */
  int16u sync_freq;                   /*!< Sync frequency in 0.1Hz (ie 100=10Hz) */
  int16u prev_sync_freq;              /*!< Previous sync frequency, in 0.1Hz (ie 100=10Hz) */
} AdcConfigVars;


/** Adc hardware data structure */
typedef struct
{
  int8u skip;                         /*!< > 0 to skip next reading, to prevent readings while ADC self zero in progress */
  int8u readingScale;                 /*!< Gain multipliers */
  struct t_io_dev *adc_rdg_dev;       /*!< The ADC device to take readings from */
  int32s adcMax;                      /*!< ADC top out value, depends on gain of ADC */
  AdcConfigVars adc_config_vars;      /*!< This struct holds initialisation variables */
  int32s lastRdg;                     /*!< Last adc reading acquired */
} AdcHwInfo;


/*
 * ADC averaging
 */


/** Maximum length of the averaging buffer */
#define ADC_AVERAGE_MAX_RDGS 500

/** Length of zero mux buffer */
#define ADC_ZERO_MUX_AVERAGE_RDGS 6 // 30 seconds at 5s mux intervals


/** Averaging structure - handles pre-averaging and averaging */
typedef struct
{
  int16u avMaxAv;                         /*!< Actual buffer length used (<= ADC_AVERAGE_MAX_RDGS) */
  int16u avPos;                           /*!< Position of latest data in filter buffer */
  int16u avCnt;                           /*!< Number of entries in the filter */
  int32s avData[ADC_AVERAGE_MAX_RDGS];    /*!< Buffer of readings */
  int64_t avTotal;                        /*!< Total of the readings in the buffer */

  int16u avNumRdgs;                       /*!< Stores total number of readings to average (pre-averaging + averaging) */

  /* Pre-averaging: used where the buffer is not long enough */
  int32s preAvTotal;                      /*!< Total for pre-averaging */
  int8u preAvCnt;                         /*!< Number of readings that have been pre-averaged */
  int8u preAv;                            /*!< Number of pre-average readings required before feeding data into the averaging */
  int8u preAvOn2;                         /*!< Rounded up (preAv / 2), pre-calculated to avoid extra divisions in @ref AdcIntAverageReading */

  bool avFull;                            /*!< Set to true if the average buffer is full */
  bool avValid;                           /*!< Set to true if average data is valid */
} AdcAverageData;


/*
 * ADC linearisation
 */


/**
 * The R400 keeps a table of 257 points of correction.
 * Span and Zero have a correction of 0.
 */
#define ADC_LINEAR_POINTS 257


/** Adc linearisation structure */
typedef struct
{
  int16s linTable[ADC_LINEAR_POINTS];
  int8u fsShr;   // num bits needed to shift fs into a 7 bit unsigned number
} AdcLinearisation;


/*
 * ADC motion detection
 */


#define ADC_MOTION_MAX_DELTA_RDGS 10


typedef struct
{
  int32s motLastDelta;
  int8u motOnCnt;
  int8u motDeltaWindow;
  int32s motDeltaData[ADC_MOTION_MAX_DELTA_RDGS];
  int32s motAbsDelta;
  int32s motDeltaTotal;
  int32s motDeltaSubtotal;
  int8u motSubNum;
  int8u motSubCnt;
  int8u motDeltaPos;
  int8u motDeltaCnt;

  int32s motRawDelta;
  int8u motWindow;

  bool motion_enabled;
} AdcMotionDetection;


/*
 * ADC temperature compensation
 */


///////////////////////////////////////////////////////////////////////////////////////
// AdcTempCal: Datastructure holding information about the temperature calibration
///////////////////////////////////////////////////////////////////////////////////////


/** Function typedef of a callback from the ADC library to your application to get the current temperature */
typedef bool (*t_adc_get_temperature_fn)(AdcSuperStructure *ass, int32u *temperature);


/** Data structure holding information about the temperature calibration */
typedef struct
{
  int16u stage;                               /*!< The current stage of the ADC temperature compensation cycle. See the ADC_TEMP_CAL_STAGE_* defines. */
  int8u  start;                               /*!< >0 when performing the initial zero self calibration */
  int32s zeroCal;                             /*!< The calculated zero compensation for the temperature */
  int32s spanReg;                             /*!< The calculated span compensation for the temperature */
  int32u currentTemp;                         /*!< The current temperature. This data must be written regularly for temperature compensation to function. */

  int16u adcTempCalStagePeriod;               /*!< Was #define ADC_TEMP_CAL_STAGE_PERIOD ((5*ADC_SYNC_FREQ)+ADC_TEMP_CAL_STAGE_FINISHED) == 5s when sync freq was fixed */
  int16u adcTempCalStagePerMax;               /*!< Was #define ADC_TEMP_CAL_STAGE_PER_MAX ((60*ADC_SYNC_FREQ)+ADC_TEMP_CAL_STAGE_FINISHED) == 60s when sync freq was fixed */

  t_adc_get_temperature_fn adc_get_temperature_fn; /*!< Function pointer to get current temperature */
} AdcTempCal;


/*
 * ADC zero tracking
 */


/** Zero tracking structure */
typedef struct
{
  int32u skipSamples;
  int32u sampleCnt;
  int32s maxDelta;
  bool opBlocked;
} AdcZeroTracking;


/*
 * ADC operation
 */


/** ADC operation structure, keeps state of current operation */
typedef struct
{
  int8u lastCommand;       /*!< Read-only current/last command */
  int8u status;            /*!< Read-only status of current/last command */
  int8u tickStatus;        /*!< Do not use */
  int8u timeoutSec;        /*!< Read-only timeout of current/last command */
  int8u cmdCnt;            /*!< Read-only - incremented for each new command */
  bool canceling;          /*!< Do not use */
  bool block;              /*!< Operation should block until complete */
  int16u timeoutSync;      /*!< Read-only timeout of current/last command */
  int16u counter;          /*!< Do not use */
  int32s paramL1;          /*!< User parameter 1, will vary depending on operation */
  int32s paramL2;          /*!< User parameter 2, will vary depending on operation */
  bool immediate;          /*!< Operation should occur immediately, without waiting for no motion. Generally only works in non-trade mode */

  OS_EVENT *opSem;         /*!< Semaphore used to signal operation completion */
  int32u averageLen;       /*!< Something to do with changing the averaging during calibration */
} AdcOperation;


/*
 * ADC miscellaneous data
 */


/** Function typedef of a callback from the ADC library to your application to get the current date and time */
typedef bool (*t_adc_get_datetime_fn)(AdcSuperStructure *ass, t_datetime *datetime);


/**
 * Function typedef for defaulting ADC databases
 *
 * @param ass is the ADC super structure, provided for your reference. Don't default the databases included in this.
 * @param afd is a pointer to a factory database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param aud is a pointer to a user database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param ard is a pointer to a runtime database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param preserve_cal_counters means the calibration counters should be preserved and incremented over the default.
 * @param cal_only means only the calibration settings should be defaulted (this applies to aud only) (settings to be defaulted are zeroRaw, spanRaw, spanCnt, linearisation).
 */
typedef void (*AdcDatabaseDefaultImplementerFn)(AdcSuperStructure *ass,
                                                AdcFactoryDatabase *afd,
                                                AdcUserDatabase *aud,
                                                AdcRuntimeDatabase *ard,
                                                bool preserve_cal_counters,
                                                bool cal_only);


/**
 * Function typedef for saving ADC databases
 *
 * Note the ass mutex will be locked when this is called
 *
 * @param ass is the ADC super structure, provided for your reference. Don't save the databases included in this.
 * @param afd is a pointer to a factory database to be saved. Can be NULL in which case shouldn't be saved.
 * @param aud is a pointer to a user database to be saved. Can be NULL in which case shouldn't be saved.
 * @param ard is a pointer to a runtime database to be saved. Can be NULL in which case shouldn't be saved.
 *
 * @return result of the save
 */
typedef bool (*AdcDatabaseSaveImplementerFn)(AdcSuperStructure *ass,
                                             const AdcFactoryDatabase *afd,
                                             const AdcUserDatabase *aud,
                                             const AdcRuntimeDatabase *ard);


/** Different locations at which the application callback function will be called */
enum eAdcCallbackFunctionLocation
{
  acflInitStart,
  acflInitEnd,
  acflReInitStart,
  acflReInitEnd,
  acflUpdateStart,
  acflUpdateEnd,
  acflOperationBeforeStart,
  acflOperationBeforeEnd,
  acflOperationStart,
  acflOperationEnd,
  acflPrepareStart,
  acflPrepareEnd
};


/**
 * Function typedef for calling back to your application when specific events occur
 *
 * Note the ass mutex will be locked when this is called
 *
 * @param ass is the ADC super structure
 * @param callbackLocation is the location at which you are being called back
 */
typedef void (*AdcAppCallbackImplementerFn)(AdcSuperStructure *ass, enum eAdcCallbackFunctionLocation callbackLocation);


/**
 * Function typedef for checking for ADC user database changes that affect calibration and configuration counters
 *
 * Note the ass mutex will be locked when this is called
 *
 * @param ass is the ADC super structure, provided for your reference. Don't check the databases included in this. Not NULL
 * @param changed_aud is a pointer to a user database that may have been changed. Not NULL.
 * @param reference_aud is a pointer to a reference copy of the user database, that can be compared against to see if settings in changed_aud have changed. Not NULL
 * @param calibration_changed should be set to true if one or more calibration settings have changed. Not NULL
 * @param configuration_changed should be set to true if one or more configuration settings have changed. Not NULL
 */
typedef void (*AdcDatabaseCalCountCheckImplementerFn)(AdcSuperStructure *ass, const AdcUserDatabase *changed_aud, const AdcUserDatabase *reference_aud,
                                                      bool *calibration_changed, bool *configuration_changed);


/**
 * Function typedef for calling back to your application for debug printfs
 *
 * @param priority is one of the LOG_xxx constants from syslog.h
 * @param format is a printf format string
 * @param args is an variable argument list to match format
 */
typedef void (*AdcDebugVPrintfImplementerFn)(int priority, const char *format, va_list args);


/** Miscellaneous ADC data, with no other sensible home */
typedef struct
{
  t_adc_get_datetime_fn adc_get_datetime_fn;                                   /*!< Function callback to your application for getting the current date and time */
  AdcDatabaseDefaultImplementerFn adcDatabaseDefaultImplementerFn;             /*!< Function callback to your application to default databases */
  AdcDatabaseSaveImplementerFn adcDatabaseSaveImplementerFn;                   /*!< Function callback to your application to save databases */
  AdcAppCallbackImplementerFn adcAppCallbackImplementerFn;                     /*!< Function callback to your application at various places in the library */
  AdcDatabaseCalCountCheckImplementerFn adcDatabaseCalCountCheckImplementerFn; /*!< Function callback to your application for checking for ADC user database changes that affect calibration and configuration counters */

  AdcInfo adcInfo;
  int32s jitterAvMaxStep;
  int16u jitterAvLength;

  AdcErrors adcErrors;

  bool updateBlocked;
  bool overloaded;
  int32s cntOverloadCntLevel;

  int16u adcRdgFreq; /*!< Data rate for application data in 0.1Hz ie 125 = 12.5Hz */
  int32s lastRdg;
} AdcMiscellaneous;


/*
 * ADC Databases
 */


// constants required for database structure
#define ADC_MAX_NUM_LIN_PTS  10               /*!< The maximum number of linearisation points. Refer to @ref AdcUserDatabase::linPts and @ref AdcUserDatabase::linCor */
#define ADC_MAX_NUM_INTERVAL 3                /*!< The maximum number of intervals. See @ref AdcUserDatabase::fullScale and @ref AdcUserDatabase::resolution */

// constants for use with AdcUserDatabase.average
#define ADC_AV_1_SEC  100                     /*!< Number of readings to average for 1 second of averaging. See @ref AdcUserDatabase::average. */
#define ADC_AV_5_SECS (5*ADC_AV_1_SEC)        /*!< Number of readings to average for 5 seconds of averaging. See @ref AdcUserDatabase::average. */


/*
 * Used with AdcUserDatabase.scaleType
 */


/** The scale types available to the ADC library */
enum eAdcScaleType { ADC_SCALE_TYPE_SINGLE = 0,       /*!< A single range scale. See @ref AdcUserDatabase::scaleType */
                     ADC_SCALE_TYPE_DUAL_INTERVAL,    /*!< A dual interval scale. See @ref AdcUserDatabase::scaleType */
                     ADC_SCALE_TYPE_DUAL_RANGE,       /*!< A dual range scale. See @ref AdcUserDatabase::scaleType */
                     ADC_SCALE_TYPE_TRIPLE_INTERVAL,  /*!< A triple interval scale. See @ref AdcUserDatabase::scaleType */
                     ADC_SCALE_TYPE_TRIPLE_RANGE,     /*!< A triple range scale. See @ref AdcUserDatabase::scaleType */
                     ADC_NUM_SCALE_TYPE,              /*!< Number of scale type settings. See @ref AdcUserDatabase::scaleType */
                   };


/*
 * Used with AdcUserDatabase.jitter
 */


/** The jitter settings available to the ADC library */
enum eAdcJitter { ADC_JITTER_OFF = 0,   /*!< No jitter filtering.  See @ref AdcUserDatabase::jitter */
                  ADC_JITTER_FINE,      /*!< Fine jitter filtering. See @ref AdcUserDatabase::jitter */
                  ADC_JITTER_COARSE,    /*!< Coarse jitter filtering. See @ref AdcUserDatabase::jitter */
                  ADC_NUM_JITTER        /*!< Number of jitter settings */
                };

/*
 * Constants used with AdcUserDatabase.resolution
 */


/** The resolutions available to the ADC library */
enum eAdcResolutions { ADC_RESOLUTION_ONE = 0,
                       ADC_RESOLUTION_TWO,
                       ADC_RESOLUTION_FIVE,
                       ADC_RESOLUTION_TEN,
                       ADC_RESOLUTION_TWENTY,
                       ADC_RESOLUTION_FIFTY,
                       ADC_RESOLUTION_ONEHUNDRED,
                       ADC_RESOLUTION_TWOHUNDRED,
                       ADC_NUM_RESOLUTION
                     };

/** Resolution constants, must match @ref enum eAdcResolutions */
#define ADC_RESOLUTIONS        { 1, 2, 5, 10, 20, 50, 100, 200 }

/** Resolution lengths, must match @ref enum eAdcResolutions */
#define ADC_RESOLUTION_LENGTHS { 1, 1, 1, 2,  2,  2,  3,   3}

/** Resolution string constants, must match @ref enum eAdcResolutions */
#define ADC_RESOLUTION_STRS    { "1", "2", "5", "10", "20", "50", "100", "200" }


extern const int8u AdcResolutions[ADC_NUM_RESOLUTION];          /*!< The possible ADC library resolutions (count-bys). Array is indexed by @ref AdcUserDatabase::resolution. */
extern const int8u AdcResolutionLengths[ADC_NUM_RESOLUTION];    /*!< The number of digits in each of the resolutions (ie res=1 has reslen=1, res=100 has reslen=3). Array is indexed by @ref AdcUserDatabase::resolution. */


/*!< Units available to the ADC library. There can only be 16, since @ref AdcUnitDp only allows 4 bits. */
enum eAdcUnits { ADC_UNITS_NONE = 0,            /*!< No units */
                 ADC_UNITS_GRAMS,               /*!< Grams (g) */
                 ADC_UNITS_KILOGRAMS,           /*!< Kilograms (kg) */
                 ADC_UNITS_TONNES,              /*!< Tonnes/tons (t) */
                 ADC_UNITS_POUNDS,              /*!< Pounds (lb) */
                 ADC_UNITS_OUNCES,              /*!< Ounces (oz) */
                 ADC_UNITS_NEWTONS,             /*!< Newtons (N) */
                 ADC_UNITS_PIECES,              /*!< Pieces (p) */
                 ADC_UNITS_USER0,               /*!< User defined units 0 */
                 ADC_UNITS_POUNDS_OUNCES,       /*!< Pounds ounces (lb:oz) */
                 ADC_UNITS_USER2,               /*!< User defined units 2 */
                 ADC_UNITS_USER3,               /*!< User defined units 3 */
                 ADC_UNITS_USER4,               /*!< User defined units 4 */
                 ADC_UNITS_USER5,               /*!< User defined units 5 */
                 ADC_UNITS_USER6,               /*!< User defined units 6 */
                 ADC_UNITS_USER7,               /*!< User defined units 7 */
                 ADC_NUM_UNITS                  /*!< Number of units options */
               };


/** ADC units constants, must match @ref enum eAdcUnits */
#define ADC_UNITS_STRS        { "",    "g",   "kg",  "t",   "lb",  "oz",  "N",   "p",   "u0",  "lb oz", "u2",  "u3",  "u4",  "u5",  "u6",  "u7" }
#define ADC_UNITS_SPACED_STRS { "   ", "  g", " kg", "  t", " lb", " oz", "  N", "  p", " u0", "lb oz", " u2", " u3", " u4", " u5", " u6", " u7" }


/** Unit switching available to the ADC library */
enum eAdcUnitSwitch { ADC_UNIT_SWITCH_PRIMARY = 0,  /*!< Primary units */
                      ADC_UNIT_SWITCH_SECONDARY,    /*!< Secondary units */
                      ADC_UNIT_SWITCH_TERTIARY,     /*!< Tertiary units */
                      ADC_NUM_UNIT_SWITCH           /*!< Number of unit switch options */
                    };


/*
 * Constants for use with AdcUserDatabase.motion
 */


/** The motion settings available to the ADC library (divisions of movement in time period) */
enum eAdcMotionDetection { ADC_MOTION_DETECT_OFF = 0,
                           ADC_MOTION_DETECT_HALFD_1S,
                           ADC_MOTION_DETECT_1D_1S,
                           ADC_MOTION_DETECT_2D_1S,
                           ADC_MOTION_DETECT_5D_1S,
                           ADC_MOTION_DETECT_HALFD_HALFS,
                           ADC_MOTION_DETECT_1D_HALFS,
                           ADC_MOTION_DETECT_2D_HALFS,
                           ADC_MOTION_DETECT_5D_HALFS,
                           ADC_MOTION_DETECT_HALFD_FIFTHS,
                           ADC_MOTION_DETECT_1D_FIFTHS,
                           ADC_MOTION_DETECT_2D_FIFTHS,
                           ADC_MOTION_DETECT_5D_FIFTHS,
                           ADC_MOTION_DETECT_3D_1S,
                           ADC_MOTION_DETECT_3D_HALFS,
                           ADC_MOTION_DETECT_3D_FIFTHS,
                           ADC_NUM_MOTION_DETECT
                         };

/** Motion delta constants (divisions), must match @ref enum eAdcMotionDetection */
#define ADC_MOTION_DELTAS { 0, 10, 20, 40, 100, 10, 20, 40, 100, 10, 20, 40, 100, 60, 60, 60 }

/** Motion time window divisors (SampFreq/Divisor), must match @ref enum eAdcMotionDetection */
#define ADC_MOTION_WINDOW_DIVISORS { 0, 1,  1,  1,  1,   2,  2,  2,  2,   5,  5,  5,  5,  1,  2,  5 }

extern const int16u AdcMotionDelta[ADC_NUM_MOTION_DETECT];       /*!< The number of divisions of disturbance required to indicate motion. Array indexed by @ref AdcUserDatabase::motion. */
extern const int8u AdcMotionWindowInv[ADC_NUM_MOTION_DETECT];    /*!< SampleFreq / AdcMotionWindowInv = the time in seconds that the disturbance must occur in to indicate motion. Array indexed by @ref AdcUserDatabase::motion. */


/*
 * Constants for use with AdcUserDatabase.zeroTracking
 */


/** The zero tracking settings available to the ADC library (divisions of zero adjustment in time period) */
enum eAdcZeroTracking { ADC_ZERO_TRACKING_OFF = 0,
                        ADC_ZERO_TRACKING_HALFD_1S,
                        ADC_ZERO_TRACKING_1D_1S,
                        ADC_ZERO_TRACKING_2D_1S,
                        ADC_ZERO_TRACKING_5D_1S,
                        ADC_ZERO_TRACKING_HALFD_HALFS,
                        ADC_ZERO_TRACKING_1D_HALFS,
                        ADC_ZERO_TRACKING_2D_HALFS,
                        ADC_ZERO_TRACKING_5D_HALFS,
                        ADC_ZERO_TRACKING_HALFD_FIFTHS,
                        ADC_ZERO_TRACKING_1D_FIFTHS,
                        ADC_ZERO_TRACKING_2D_FIFTHS,
                        ADC_ZERO_TRACKING_5D_FIFTHS,
                        ADC_ZERO_TRACKING_3D_1s,
                        ADC_ZERO_TRACKING_3D_HALFS,
                        ADC_ZERO_TRACKING_3D_FIFTHS,
                        ADC_NUM_ZERO_TRACKING
                      };

/** Zero tracking delta constants in counts (divisions * 16 - ie 8 = 1/2d), must match @ref enum eAdcZeroTracking */
#define ADC_ZERO_TRACKING_DELTAS { 0,  8, 16, 32, 80,  8, 16, 32, 80,  8, 16, 32, 80,  48, 48, 48 }

/** Zero tracking time window divisors (SampFreq/Divisor), must match @ref enum eAdcZeroTracking */
#define ADC_ZERO_TRACKING_WINDOW_DIVISORS { 0,  1, 1, 1, 1,  2, 2, 2, 2,  5, 5, 5, 5,  1, 2, 5 }

extern const int16u AdcZeroTrackingDelta[ADC_NUM_ZERO_TRACKING];       /*!< The maximum amount (in counts=divisions * 16) of zero cancellation applied by the zero tracking in any zero tracking window period. Array indexed by @ref AdcUserDatabase::zeroTracking. */
extern const int8u AdcZeroTrackingWindowInv[ADC_NUM_ZERO_TRACKING];    /*!< SampleFreq / AdcZeroTrackingWindowInv = the time in seconds over which a maximum of AdcZeroTrackingDelta zero cancellation can be applied. Array indexed by @ref AdcUserDatabase::zeroTracking. */


/*
 * Constants for use with AdcUserDatabase.rangeOfZero
 */


/** The zero range settings available to the ADC library */
enum eAdcZeroRanges { ADC_ZERO_RANGE_NEG2_TO_PLUS2 = 0,
                      ADC_ZERO_RANGE_NEG1_TO_PLUS3,
                      ADC_ZERO_RANGE_NEG10_TO_PLUS10,
                      ADC_ZERO_RANGE_NEG20_TO_PLUS20,
                      ADC_ZERO_RANGE_NEG100_TO_PLUS100,
                      ADC_NUM_ZERO_RANGE
                    };


/** Zero range positive constants, must match @ref enum eAdcZeroRanges */
#define ADC_ZERO_RANGES_PLUS { 2,  3,  10, 20, 100 }

/** Zero range positive constants, must match @ref enum eAdcZeroRanges */
#define ADC_ZERO_RANGES_NEGATIVE { -2, -1, -10, -20, -100 }

extern const int8s AdcZeroRangePos[ADC_NUM_ZERO_RANGE];        /*!< ADC range of zero setting for positive readings. Array indexed by @ref AdcUserDatabase::rangeOfZero */
extern const int8s AdcZeroRangeNeg[ADC_NUM_ZERO_RANGE];        /*!< ADC range of zero setting for negative readings. Array indexed by @ref AdcUserDatabase::rangeOfZero */


/*
 * Constants for use with AdcUserDatabase.syncFreqIdx
 */


/** The sync frequencies available to the ADC library */
enum eAdcSyncFrequency { ADC_SYNC_FREQ_10_HZ = 0,
                         ADC_SYNC_FREQ_12POINT5_HZ,
                         ADC_SYNC_FREQ_15_HZ,
                         ADC_SYNC_FREQ_20_HZ,
                         ADC_SYNC_FREQ_25_HZ,
                         ADC_SYNC_FREQ_30_HZ,
                         ADC_SYNC_FREQ_50_HZ,
                         ADC_SYNC_FREQ_60_HZ,
                         ADC_SYNC_FREQ_100_HZ,
                         ADC_NUM_SYNC_FREQ
                       };


/** Sync frequency constants, must match @ref enum eAdcSyncFrequency */
#define ADC_SYNC_FREQUENCIES { 100, 125, 150, 200, 250, 300, 500, 600, 1000 }

extern const int16u AdcSyncFrequencies[ADC_NUM_SYNC_FREQ];      /*!< ADC sync frequencies. Array indexed by @ref AdcUserDatabase::syncFreqIdx */


/*
 * Constants for use with AdcUserDatabase.approvalMode and InternAdcData.approvalMode
 */

/** The approval modes available to the ADC library */
enum eAdcApprovalMode { ADC_APPROVAL_MODE_NONE = 0,       /*!< No approval mode (ie industrial) */
                        ADC_APPROVAL_MODE_OIML,           /*!< OIML trade approval mode (European) */
                        ADC_APPROVAL_MODE_NTEP,           /*!< NTEP trade approval mode (USA) */
                        ADC_APPROVAL_MODE_NSC,            /*!< NSC trade approval mode (Australian) */
                        ADC_NUM_APPROVAL_MODE             /*!< Number of approval modes */
                      };


// constants for use with calFlag
#define ADC_CAL_FLAG_OK 0xA5                  /*!< Value to indicate that the calibration is valid. Refer to @ref AdcUserDatabase::calFlag */


#define ADC_RAW_MULT_UNITY 1000000            /*!< Used with tilt compensation */


/* Database versioning */
#define ADC_DB_USER_CHECKSUM_VERSION 1        /*!< ADC user database checksum version - bump this when you change @ref AdcUserDatabase */
#define ADC_DB_RUNTIME_CHECKSUM_VERSION 1     /*!< ADC runtime database checksum version - bump this when you change @ref AdcRuntimeDatabase */

/* Database checksum seeds */

/** Checksum seed for the factory database, currently hardcoded to 0xFFFF */
#define ADC_DB_FACTORY_CHECKSUM_SEED (0xFFFF)

/** Checksum seed for the user database, consisting of the size of the database + version */
#define ADC_DB_USER_CHECKSUM_SEED (sizeof(AdcUserDatabase) + ADC_DB_USER_CHECKSUM_VERSION)

/** Checksum seed for the runtime database, consisting of the size of the database + version */
#define ADC_DB_RUNTIME_CHECKSUM_SEED (sizeof(AdcRuntimeDatabase) + ADC_DB_RUNTIME_CHECKSUM_VERSION)

/** Define to calculate a database checksum. Relies on db containing a member "chksum" */
#define DB_CHECKSUM(dbtype, db, seed) \
  AdcCalcCRC16((const int8u *) &(db), offsetof(dbtype, chksum), seed)

/** Define to calculate and set the checksum within a database. Relies on db containing a member "chksum" */
#define DB_SET_CHECKSUM(dbtype, db, seed) \
  (db).chksum = DB_CHECKSUM(dbtype, db, seed)

/** Define to return true if the checksum on a database is good. Relies on db containing a member "chksum" */
#define DB_CHECKSUM_GOOD(dbtype, db, seed) \
  (DB_CHECKSUM(dbtype, db, seed) == (db).chksum)


/**
 * This number is used to generate the span temperature compensation. Because we use
 * integer maths, we need to apply some gain, so bits aren't lost.
 *
 * - the x8 is to shift the numbers up a bit so we don't loose too many bits
 * - the x16 is because we accumulate 16 temperature readings to give more precision
 *
 * The maths is:
 *   Span Register = sAdcFactoryDatabase.spanCal + sAdcFactoryDatabase.spanDrift * ADC_SPAN_DRIFT_SCALING / (current_temperature - calTemp)
 */
#define ADC_SPAN_DRIFT_SCALING (8 * 16)


/**
 * This number is used to generate the zero temperature compensation. Because we use
 * integer maths, we need to apply some gain, so bits aren't lost.
 *
 * - the x64 is to shift the numbers up a bit so we don't loose too many bits
 * - the x16 is because we accumulate 16 temperature readings to give more precision
 *
 * The maths is:
 *   Zero Register = sAdcFactoryDatabase.zeroOffset + sAdcFactoryDatabase.spanDrift * ADC_ZERO_DRIFT_SCALING / (current_temperature - calTemp)
 */
#define ADC_ZERO_DRIFT_SCALING (64 * 16)


/**
 * The ADC factory database. This is filled in by the Rinstrum oven factory calibration process, and stored in flash.
 *
 * We want 2 byte alignment here, because that is what M68k does,
 * and when we compile for x86_64 which normally does 4 byte alignment the checksum
 * calculations don't work.
 */
#pragma pack(push)
#pragma pack(2)
struct sAdcFactoryDatabase
{
  int32s zeroOffset[ADC_NUM_SYNC_FREQ];       /*!< Fixed zero offset, one for each sync frequency */
  int32s zeroDriftPos[ADC_NUM_SYNC_FREQ];     /*!< Zero drift with temperature, one for each sync frequency */
  int32s zeroDriftNeg[ADC_NUM_SYNC_FREQ];     /*!< Zero drift with temperature, one for each sync frequency */
  int32s spanCal[ADC_NUM_SYNC_FREQ];          /*!< Span calibration, one for each sync frequency */
  int32s spanDriftPos[ADC_NUM_SYNC_FREQ];     /*!< Span calibration drift with temperature, one for each sync frequency */
  int32s spanDriftNeg[ADC_NUM_SYNC_FREQ];     /*!< Span calibration drift with temperature, one for each sync frequency */
  int32u calTemp;                             /*!< Raw temperature at calibration */
  int8u  calFlag;                             /*!< Set to 0xA5 if factory calibration valid */

  int32u chksum;                              /*!< Checksum will be checked before database use */
};
#pragma pack(pop)


#define ADC_UINT4_MAX ((1 << 4) - 1)
#define ADC_INT24_MIN (-(1 << 23))
#define ADC_INT24_MAX ((1 << 23) - 1)
#define ADC_BOOLEAN_MIN (0)
#define ADC_BOOLEAN_MAX (1)

/* Factory database default values */
#define ADC_DB_FACT_ZEROOFFSET_MINIMUM (ADC_INT24_MIN)
#define ADC_DB_FACT_ZEROOFFSET_MAXIMUM (ADC_INT24_MAX)
#define ADC_DB_FACT_ZEROOFFSET_DEFAULT (0)

#define ADC_DB_FACT_ZERODRIFT_MINIMUM (ADC_INT24_MIN)
#define ADC_DB_FACT_ZERODRIFT_MAXIMUM (ADC_INT24_MAX)
#define ADC_DB_FACT_ZERODRIFT_DEFAULT (0)

#define ADC_DB_FACT_SPANCAL_MINIMUM (ADC_INT24_MIN)
#define ADC_DB_FACT_SPANCAL_MAXIMUM (ADC_INT24_MAX)
#define ADC_DB_FACT_SPANCAL_DEFAULT0 (8059200L)
#define ADC_DB_FACT_SPANCAL_DEFAULT1 (8181100L)
#define ADC_DB_FACT_SPANCAL_DEFAULT2 (7068500L)
#define ADC_DB_FACT_SPANCAL_DEFAULT3 (8064100L)
#define ADC_DB_FACT_SPANCAL_DEFAULT4 (8185200L)
#define ADC_DB_FACT_SPANCAL_DEFAULT5 (7072000L)
#define ADC_DB_FACT_SPANCAL_DEFAULT6 (8185100L)
#define ADC_DB_FACT_SPANCAL_DEFAULT7 (7071900L)
#define ADC_DB_FACT_SPANCAL_DEFAULT8 (8185200L)

#define ADC_DB_FACT_SPANDRIFT_MINIMUM (ADC_INT24_MIN)
#define ADC_DB_FACT_SPANDRIFT_MAXIMUM (ADC_INT24_MAX)
#define ADC_DB_FACT_SPANDRIFT_DEFAULT (0)

#define ADC_DB_FACT_CALTEMP_MINIMUM (0)
#define ADC_DB_FACT_CALTEMP_MAXIMUM (UINT16_MAX)
#define ADC_DB_FACT_CALTEMP_DEFAULT (600)

#define ADC_DB_FACT_CALFLAG_MINIMUM (0)
#define ADC_DB_FACT_CALFLAG_MAXIMUM (UINT8_MAX)
#define ADC_DB_FACT_CALFLAG_DEFAULT (0)

#define ADC_DB_FACT_CHKSUM_MINIMUM (0)
#define ADC_DB_FACT_CHKSUM_MAXIMUM (UINT16_MAX)
#define ADC_DB_FACT_CHKSUM_DEFAULT (0)


/** Convert deci-microvolts (10^-7) per volt into ADC library raw units (raw: 1mV/V = 256 * 10000) */
#define ADC_DUVV_TO_RAW(duvv) (duvv * 256)


/** Convert ADC library raw units into deci-microvolts (10^-7) per volt (raw: 1mV/V = 256 * 10000) */
#define ADC_RAW_TO_DUVV(raw) (raw / 256)


/** Convert ADC library raw units into deci-microvolts (10^-7) per volt with rounding (raw: 1mV/V = 256 * 10000) */
#define ADC_RAW_TO_DUVV_ROUNDED(raw) \
  ((raw >= 0) ? ((raw + 128) / 256) : ((raw - 128) / 256))


/** Convert microvolts per volt into ADC library raw units (raw: 1mV/V = 256 * 10000) */
#define ADC_UVV_TO_RAW(uvv) (ADC_DUVV_TO_RAW(uvv * 10))


/** Convert ADC library raw units into microvolts per volt (raw: 1mV/V = 256 * 10000) */
#define ADC_RAW_TO_UVV(raw) (ADC_RAW_TO_DUVV(raw / 10))


/** Convert millivolts per volt into ADC library raw units (raw: 1mV/V = 256 * 10000) */
#define ADC_MVV_TO_RAW(mvv) (ADC_UVV_TO_RAW(mvv * 1000))


/** Convert ADC library raw units into millivolts per volt (raw: 1mV/V = 256 * 10000) */
#define ADC_RAW_TO_MVV(raw) (ADC_RAW_TO_UVV(raw / 1000))


/** ADC user calibration database */
struct sAdcUserDatabase
{
  int32s zeroRaw;                             /*!< Zero calibration value in raw ADC readings */
  int32s spanRaw;                             /*!< Span calibration value in raw ADC readings */
  int32s spanCnt;                             /*!< ADC counts at the raw span calibration point (@ref calWeight * 16) */
  int8u  linPts[ADC_MAX_NUM_LIN_PTS+1];       /*!< Position of the linearisation points */
  int32s linCor[ADC_MAX_NUM_LIN_PTS+1];       /*!< Correction required at each of the linearisation points */

  int8u  scaleType;                           /*!< The type of scale as defined by the ADC_SCALE_TYPE_* defines */
  int8u  syncFreqIdx;                         /*!< The index into the sync frequency array. See @ref enum eAdcSyncFrequency */
  int8u  jitter;                              /*!< Jitter setting (secondary averaging), see the ADC_JITTER_* defines */
  int8u  dpPos;                               /*!< The position of the decimal point (0 for none) */
  int32s fullScale[ADC_MAX_NUM_INTERVAL];     /*!< The full scale reading for each of the ranges/intervals */
  int8u  resolution[ADC_MAX_NUM_INTERVAL];    /*!< The resolution/count-by for each of the ranges/intervals */
  int8u  units;                               /*!< Display unit annunciators (eg Kg, Lb) */
  int32s calWeight;                           /*!< Calibration span weight (@ref spanCnt / 16) */
  int32s addTare;                             /*!< Additive tare??? */
  int32s zeroBand;                            /*!< Zero band for the scale */
  int8u  motion;                              /*!< Motion setting. See @ref AdcMotionDelta & @ref AdcMotionWindowInv */
  int8u  zeroTracking;                        /*!< Zero tracking setting. See @ref AdcZeroTrackingFreq */
  int32u average;                             /*!< Reading averaging: average_time = average * 0.01 [s] */
  int8u  rangeOfZero;                         /*!< Zero tracking setting. See @ref AdcZeroRangePos & @ref AdcZeroRangeNeg */
  int8u  approvalMode;                        /*!< The operating mode of the ADC Library. See the ADC_MODE_* defines */
  int8u  zeroOnStart;                         /*!< Boolean setting to force a zero on startup */
  int8u  wire4Mode;                           /*!< Boolean setting for 4 or 6 wire loadcell operating mode */
  int8u  tempCalZero;                         /*!< Boolean setting to turn off/on continuous temperature compensation at zero? */
  int8u  hiRes;                               /*!< Boolean setting to force an extra digit of resolution */

  int8u  qaEnable;                            /*!< Boolean flag to indicate quality assurance is enabled. */
  int8u  qaDay;                               /*!< Combined with qaMonth and qaYear gives a date on which to generate an warning that quality assurance is required. */
  int8u  qaMonth;                             /*!< Combined with qaDay and qaYear gives a date on which to generate an warning that quality assurance is required. */
  int32u qaYear;                              /*!< Combined with qaDay and qaMonth gives a date on which to generate an warning that quality assurance is required. */

  int32u calCntNTEPCalibration;               /*!< Incremented every time NTEP calibration is changed */
  int32u calCntNTEPSettings;                  /*!< Incremented every time NTEP settings are changed */
  int32u calCntOIML;                          /*!< Incremented every time OIML calibration is changed */
  int8u  calFlag;                             /*!< Set to 0xA5 if calibration valid */

  int32u  chksum;                             /*!< The checksum of the ADC user database. It will be checked before database use */
};


/* User database default values */
#define ADC_DB_USER_ZERORAW_MINIMUM (ADC_MVV_TO_RAW(-3))  /*!< -3mV/V */
#define ADC_DB_USER_ZERORAW_MAXIMUM (ADC_MVV_TO_RAW(3))   /*!< 3mV/V */
#define ADC_DB_USER_ZERORAW_DEFAULT (ADC_MVV_TO_RAW(0))   /*!< 0mV/V */

#define ADC_DB_USER_SPANRAW_MINIMUM (ADC_MVV_TO_RAW(-5))  /*!< -5.0 mV/V */
#define ADC_DB_USER_SPANRAW_MAXIMUM (ADC_MVV_TO_RAW(5))   /*!< 5.0 mV/V */
#define ADC_DB_USER_SPANRAW_DEFAULT (ADC_UVV_TO_RAW(800)) /*!< 0.8mV/V */

#define ADC_DB_USER_SPANCNT_MINIMUM (INT32_MIN)
#define ADC_DB_USER_SPANCNT_MAXIMUM (INT32_MAX)
#define ADC_DB_USER_SPANCNT_DEFAULT (ADC_DB_USER_CALWEIGHT_DEFAULT * 16)

#define ADC_DB_USER_LINPTS_MINIMUM (0)
#define ADC_DB_USER_LINPTS_MAXIMUM (UINT8_MAX)
#define ADC_DB_USER_LINPTS_DEFAULT (128)

#define ADC_DB_USER_LINPTSLAST_MINIMUM (ADC_DB_USER_LINPTS_MINIMUM)
#define ADC_DB_USER_LINPTSLAST_MAXIMUM (ADC_DB_USER_LINPTS_MAXIMUM)
#define ADC_DB_USER_LINPTSLAST_DEFAULT (221)

#define ADC_DB_USER_LINCOR_MINIMUM (-2000)
#define ADC_DB_USER_LINCOR_MAXIMUM (2000)
#define ADC_DB_USER_LINCOR_DEFAULT (0)

#define ADC_DB_USER_SCALETYPE_MINIMUM (0)
#define ADC_DB_USER_SCALETYPE_MAXIMUM (ADC_NUM_SCALE_TYPE - 1)
#define ADC_DB_USER_SCALETYPE_DEFAULT (ADC_SCALE_TYPE_SINGLE)

#define ADC_DB_USER_SYNCFREQIDX_MINIMUM (0)
#define ADC_DB_USER_SYNCFREQIDX_MAXIMUM (ADC_NUM_SYNC_FREQ - 1)
#define ADC_DB_USER_SYNCFREQIDX_DEFAULT (ADC_SYNC_FREQ_50_HZ)

#define ADC_DB_USER_JITTER_MINIMUM (0)
#define ADC_DB_USER_JITTER_MAXIMUM (ADC_NUM_JITTER - 1)
#define ADC_DB_USER_JITTER_DEFAULT (ADC_JITTER_FINE)

#define ADC_DB_USER_DPPOS_MINIMUM (0)
#define ADC_DB_USER_DPPOS_MAXIMUM (ADC_UINT4_MAX)                /*!< AdcUnitDp struct can only store 4 bits of decimal point position */
#define ADC_DB_USER_DPPOS_DEFAULT (0)

#define ADC_DB_USER_FULLSCALE0_MINIMUM (INT32_MIN)
#define ADC_DB_USER_FULLSCALE0_MAXIMUM (INT32_MAX)
#define ADC_DB_USER_FULLSCALE0_DEFAULT (3000)

#define ADC_DB_USER_FULLSCALE1_MINIMUM (INT32_MIN)
#define ADC_DB_USER_FULLSCALE1_MAXIMUM (INT32_MAX)
#define ADC_DB_USER_FULLSCALE1_DEFAULT (6000)

#define ADC_DB_USER_FULLSCALE2_MINIMUM (INT32_MIN)
#define ADC_DB_USER_FULLSCALE2_MAXIMUM (INT32_MAX)
#define ADC_DB_USER_FULLSCALE2_DEFAULT (10000)

#define ADC_DB_USER_RESOLUTION0_MINIMUM (0)
#define ADC_DB_USER_RESOLUTION0_MAXIMUM (ADC_NUM_RESOLUTION - 1)
#define ADC_DB_USER_RESOLUTION0_DEFAULT (ADC_RESOLUTION_ONE)

#define ADC_DB_USER_RESOLUTION1_MINIMUM (0)
#define ADC_DB_USER_RESOLUTION1_MAXIMUM (ADC_NUM_RESOLUTION - 1)
#define ADC_DB_USER_RESOLUTION1_DEFAULT (ADC_RESOLUTION_TWO)

#define ADC_DB_USER_RESOLUTION2_MINIMUM (0)
#define ADC_DB_USER_RESOLUTION2_MAXIMUM (ADC_NUM_RESOLUTION - 1)
#define ADC_DB_USER_RESOLUTION2_DEFAULT (ADC_RESOLUTION_FIVE)

#define ADC_DB_USER_UNITS_MINIMUM (0)
#define ADC_DB_USER_UNITS_MAXIMUM (ADC_NUM_UNITS)
#define ADC_DB_USER_UNITS_DEFAULT (ADC_UNITS_KILOGRAMS)

#define ADC_DB_USER_CALWEIGHT_MINIMUM (INT32_MIN)
#define ADC_DB_USER_CALWEIGHT_MAXIMUM (INT32_MAX)
#define ADC_DB_USER_CALWEIGHT_DEFAULT (3000)

#define ADC_DB_USER_ADDTARE_MINIMUM (INT32_MIN)
#define ADC_DB_USER_ADDTARE_MAXIMUM (INT32_MAX)
#define ADC_DB_USER_ADDTARE_DEFAULT (0)

#define ADC_DB_USER_ZEROBAND_MINIMUM (INT32_MIN)
#define ADC_DB_USER_ZEROBAND_MAXIMUM (INT32_MAX)
#define ADC_DB_USER_ZEROBAND_DEFAULT (0)

#define ADC_DB_USER_MOTION_MINIMUM (0)
#define ADC_DB_USER_MOTION_MAXIMUM (ADC_NUM_MOTION_DETECT - 1)
#define ADC_DB_USER_MOTION_DEFAULT (ADC_MOTION_DETECT_HALFD_1S)

#define ADC_DB_USER_ZEROTRACKING_MINIMUM (0)
#define ADC_DB_USER_ZEROTRACKING_MAXIMUM (ADC_NUM_ZERO_TRACKING - 1)
#define ADC_DB_USER_ZEROTRACKING_DEFAULT (ADC_ZERO_TRACKING_OFF)

#define ADC_DB_USER_AVERAGE_MINIMUM (0)
#define ADC_DB_USER_AVERAGE_MAXIMUM (INT16_MAX)
#define ADC_DB_USER_AVERAGE_DEFAULT (ADC_AV_1_SEC)

#define ADC_DB_USER_RANGEOFZERO_MINIMUM (0)
#define ADC_DB_USER_RANGEOFZERO_MAXIMUM (ADC_NUM_ZERO_RANGE - 1)
#define ADC_DB_USER_RANGEOFZERO_DEFAULT (ADC_ZERO_RANGE_NEG2_TO_PLUS2)

#define ADC_DB_USER_APPROVALMODE_MINIMUM (0)
#define ADC_DB_USER_APPROVALMODE_MAXIMUM (ADC_NUM_APPROVAL_MODE - 1)
#define ADC_DB_USER_APPROVALMODE_DEFAULT (ADC_APPROVAL_MODE_NONE)

#define ADC_DB_USER_ZEROONSTART_MINIMUM (ADC_BOOLEAN_MIN)
#define ADC_DB_USER_ZEROONSTART_MAXIMUM (ADC_BOOLEAN_MAX)
#define ADC_DB_USER_ZEROONSTART_DEFAULT (0)

#define ADC_DB_USER_WIRE4MODE_MINIMUM (ADC_BOOLEAN_MIN)
#define ADC_DB_USER_WIRE4MODE_MAXIMUM (ADC_BOOLEAN_MAX)
#define ADC_DB_USER_WIRE4MODE_DEFAULT (0)

#define ADC_DB_USER_TEMPCALZERO_MINIMUM (ADC_BOOLEAN_MIN)
#define ADC_DB_USER_TEMPCALZERO_MAXIMUM (ADC_BOOLEAN_MAX)
#define ADC_DB_USER_TEMPCALZERO_DEFAULT (1)

#define ADC_DB_USER_HIRES_MINIMUM (ADC_BOOLEAN_MIN)
#define ADC_DB_USER_HIRES_MAXIMUM (ADC_BOOLEAN_MAX)
#define ADC_DB_USER_HIRES_DEFAULT (0)

#define ADC_DB_USER_QAENABLE_MINIMUM (ADC_BOOLEAN_MIN)
#define ADC_DB_USER_QAENABLE_MAXIMUM (ADC_BOOLEAN_MAX)
#define ADC_DB_USER_QAENABLE_DEFAULT (0)

#define ADC_DB_USER_QADAY_MINIMUM (1)
#define ADC_DB_USER_QADAY_MAXIMUM (31)
#define ADC_DB_USER_QADAY_DEFAULT (1)

#define ADC_DB_USER_QAMONTH_MINIMUM (1)
#define ADC_DB_USER_QAMONTH_MAXIMUM (12)
#define ADC_DB_USER_QAMONTH_DEFAULT (1)

#define ADC_DB_USER_QAYEAR_MINIMUM (2000)
#define ADC_DB_USER_QAYEAR_MAXIMUM (3000)
#define ADC_DB_USER_QAYEAR_DEFAULT (2010)

#define ADC_DB_USER_CALCNTNTEPCALIBRATION_MINIMUM (0)
#define ADC_DB_USER_CALCNTNTEPCALIBRATION_MAXIMUM (UINT32_MAX)
#define ADC_DB_USER_CALCNTNTEPCALIBRATION_DEFAULT (0)

#define ADC_DB_USER_CALCNTNTEPSETTINGS_MINIMUM (0)
#define ADC_DB_USER_CALCNTNTEPSETTINGS_MAXIMUM (UINT32_MAX)
#define ADC_DB_USER_CALCNTNTEPSETTINGS_DEFAULT (0)

#define ADC_DB_USER_CALCNTOIML_MINIMUM (0)
#define ADC_DB_USER_CALCNTOIML_MAXIMUM (UINT32_MAX)
#define ADC_DB_USER_CALCNTOIML_DEFAULT (0)

#define ADC_DB_USER_CALFLAG_MINIMUM (0)
#define ADC_DB_USER_CALFLAG_MAXIMUM (UINT8_MAX)
#define ADC_DB_USER_CALFLAG_DEFAULT (0)

#define ADC_DB_USER_CHKSUM_MINIMUM (0)
#define ADC_DB_USER_CHKSUM_MAXIMUM (UINT16_MAX)
#define ADC_DB_USER_CHKSUM_DEFAULT (0)


/** The ADC runtime database. Values in this structure change frequently (such as tare values). */
struct sAdcRuntimeDatabase
{
  int32s    cntZero;                          /*!< ADC counts at user zero */
  int32s    cntTare;                          /*!< ADC counts for the user tare */
  int32s    cntPTare;                         /*!< ADC counts for the preset tare */
  int8u     hiRes;                            /*!< True if x10 mode active */
  int8u     netMode;                          /*!< Flag to indicate if the scale is in net mode */
  int32u    printId;                          /*!< Current unique traceable weight id */
  int8u     dispMode;                         /*!< Does not affect adc operation.  Same constants as @ref AdcData::dispMode. */
  int8u     tempCalOn;                        /*!< Temperature compensation is enabled?? */
  int32u    overloadCnt;                      /*!< Over or underload count */

  int32u    chksum;                           /*!< Checksum will be checked before database use */
};


#define ADC_DB_RT_CNTZERO_MINIMUM (INT32_MIN)
#define ADC_DB_RT_CNTZERO_MAXIMUM (INT32_MAX)
#define ADC_DB_RT_CNTZERO_DEFAULT (0)

#define ADC_DB_RT_CNTTARE_MINIMUM (INT32_MIN)
#define ADC_DB_RT_CNTTARE_MAXIMUM (INT32_MAX)
#define ADC_DB_RT_CNTTARE_DEFAULT (0)

#define ADC_DB_RT_CNTPTARE_MINIMUM (INT32_MIN)
#define ADC_DB_RT_CNTPTARE_MAXIMUM (INT32_MAX)
#define ADC_DB_RT_CNTPTARE_DEFAULT (0)

#define ADC_DB_RT_HIRES_MINIMUM (ADC_BOOLEAN_MIN)
#define ADC_DB_RT_HIRES_MAXIMUM (ADC_BOOLEAN_MAX)
#define ADC_DB_RT_HIRES_DEFAULT (0)

#define ADC_DB_RT_NETMODE_MINIMUM (ADC_BOOLEAN_MIN)
#define ADC_DB_RT_NETMODE_MAXIMUM (ADC_BOOLEAN_MAX)
#define ADC_DB_RT_NETMODE_DEFAULT (0)

#define ADC_DB_RT_PRINTID_MINIMUM (0)
#define ADC_DB_RT_PRINTID_MAXIMUM (UINT32_MAX)
#define ADC_DB_RT_PRINTID_DEFAULT (0)

#define ADC_DB_RT_DISPMODE_MINIMUM (0)
#define ADC_DB_RT_DISPMODE_MAXIMUM (ADC_NUM_DISPLAY_MODE - 1)
#define ADC_DB_RT_DISPMODE_DEFAULT (ADC_DISPLAY_MODE_WEIGHT)

#define ADC_DB_RT_TEMPCALON_MINIMUM (ADC_BOOLEAN_MIN)
#define ADC_DB_RT_TEMPCALON_MAXIMUM (ADC_BOOLEAN_MAX)
#define ADC_DB_RT_TEMPCALON_DEFAULT (0)

#define ADC_DB_RT_OVERLOADCNT_MINIMUM (0)
#define ADC_DB_RT_OVERLOADCNT_MAXIMUM (UINT32_MAX)
#define ADC_DB_RT_OVERLOADCNT_DEFAULT (0)

#define ADC_DB_RT_CHKSUM_MINIMUM (0)
#define ADC_DB_RT_CHKSUM_MAXIMUM (UINT16_MAX)
#define ADC_DB_RT_CHKSUM_DEFAULT (0)


///////////////////////////////////////////////////////////////////////////////////////
// InternAdcData: Used by the adc library
///////////////////////////////////////////////////////////////////////////////////////

/** Internal ADC data, not for public consumption */
typedef struct {
  int32s cntLast;                             /*!< ADC counts from the last reading */
  int32s cntInitZero;                         /*!< ADC counts for the initial zeroing */
  int32s cntSysRdg;                           /*!< ADC counts for the system reading (ie before zeroing, etc has been applied) */
  int32s cntGross;                            /*!< ADC counts for the gross reading */
  int32s cntNet;                              /*!< ADC counts for the net reading */

  int32s cntCOZ;                              /*!< The width of the center of zero band in ADC counts */
  int32s cntOload;                            /*!< The ADC counts at which overload occurs */
  int32s cntUload;                            /*!< The ADC counts at which underload occurs */
  int32s cntZeroRangeMin;                     /*!< The lower limit in ADC counts of the range in which a user zero is allowable */
  int32s cntZeroRangeMax;                     /*!< The upper limit in ADC counts of the range in which a user zero is allowable */
  int32s cntZeroBandMax;                      /*!< The upper limit in ADC counts of the band in which to indicate zero band */

  //int8u  rangeCurrent;
  int8u  rangeMax;                            /*!< The maximum number of ranges of the current scale calibration */
  int32s rangeLimits[ADC_MAX_NUM_INTERVAL];   /*!< The limit for each of the ranges in ADC counts */
  bool   rangeMulti;                          /*!< Flag indicating the scale has multiple ranges (but not multiple intervals) */

  int8u  approvalMode;                        /*!< The operating mode of the ADC Library. See the ADC_MODE_* defines */
  bool   userBlocked;                         /*!< Block user operations? / readings while calibrating */
  int8u  serviceSignals;                      /*!< No idea, not used */

  int8u  altUnits;                            /*!< The alternate units to display weight in */
  int8s  altDpInc;                            /*!< The increment in decimal places between calibrated decimal point and the alternate units decimal point */
  int8u  altResLen[ADC_MAX_NUM_INTERVAL];     /*!< The number of digits in each of the alternate weight units resolution (ie resolution=100, has reslen=3) */
  int32s altRes[ADC_MAX_NUM_INTERVAL];        /*!< The alternate weight units resolution/countby for each interval/range */
  int32s altConv;                             /*!< The conversion factor between calibrated weight and the alternate units */

  int32s sampleNum;                           /*!< The number of samples when in piece counting mode */
  int32s cntSampleWgt;                        /*!< The weight of the the number of samples */

  int32s rawAdd;                              /*!< Raw ADC readings addition, used in tilt compensation */
  int32s rawMult;                             /*!< Raw ADC readings multiplier, used in tilt compensation */

  struct timespec tsReading;                  /*!< Monotonic time when the reading was taken */

  int8u  tertUnits;                           /*!< The tertiary units to display weight in */
  int8s  tertDpInc;                           /*!< The increment in decimal places between calibrated decimal point and the tertiary units decimal point */
  int8u  tertResLen[ADC_MAX_NUM_INTERVAL];    /*!< The number of digits in each of the tertiary weight units resolution (ie resolution=100, has reslen=3) */
  int32s tertRes[ADC_MAX_NUM_INTERVAL];       /*!< The tertiary weight units resolution/countby for each interval/range */
  int32s tertConv;                            /*!< The conversion factor between calibrated weight and the tertiarye units */
} AdcInternalData;


/** Contains all data for the ADC */
struct sAdcSuperStructure
{
  /* Locks */
  AdcMutexImplementer *adcMutex;

  /* Databases */
  AdcFactoryDatabase adcFactoryDatabase;
  AdcUserDatabase adcUserDatabase;
  AdcRuntimeDatabase adcRuntimeDatabase;
  AdcRuntimeDatabase adcRuntimeDatabaseDiff;  /*!< Used to diff against adcRuntimeDatabase, so we know when to save it */

  /* Adc produced data */
  AdcData adcDataWorking;
  AdcTraceableWeight adcTraceableWeight;      /*!< This structure holds the last successfully requested traceable weight and may be accessed by the application */
  AdcInternalData adcInternalData;
  AdcData adcHeldData;                        /*!< This is an entire copy of the working data taken when a manual hold is applied */

  /* Adc library internals */
  AdcAverageData adcAveraging;
  AdcAverageData adcJitterAveraging;
  AdcAverageData adcZeroMuxAveraging;
  AdcHwInfo adcHwInfo;
  AdcLinearisation adcLinearisation;
  AdcMotionDetection adcMotionDetection;
  AdcZeroTracking adcZeroTracking;
  AdcTempCal adcTempCal;
  AdcOperation adcOperation;
  AdcMiscellaneous adcMiscellaneous;

  /* Live weigh */
  AdcLiveWeighImplementer *adcLiveWeigh;
};


#endif /* #ifndef ADC_DATATYPES_DOT_H */

