/*
 * AdcDbUser.c
 *
 *  Created on: 31/03/2011
 *      Author: ian
 */
#include <stdlib.h>
#include <string.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rstdio.h>
#include <rinweigh/adcoperations.h>
#include <rinweigh/adcreadings.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcmotion.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adccalcountcheck_std.h>

#include <rinweigh/adcdbuser.h>


/** Default values for the ADC user database */
static const AdcUserDatabase s_C_DefaultAdcUserDB =
{
  .zeroRaw = ADC_DB_USER_ZERORAW_DEFAULT,
  .spanRaw = ADC_DB_USER_SPANRAW_DEFAULT,
  .spanCnt = ADC_DB_USER_SPANCNT_DEFAULT,
  .linPts =
  {                                           /* int8u  linPts[ADC_MAX_NUM_LIN_PTS+1]; */
    ADC_DB_USER_LINPTS_DEFAULT, ADC_DB_USER_LINPTS_DEFAULT, ADC_DB_USER_LINPTS_DEFAULT, ADC_DB_USER_LINPTS_DEFAULT, ADC_DB_USER_LINPTS_DEFAULT,
    ADC_DB_USER_LINPTS_DEFAULT, ADC_DB_USER_LINPTS_DEFAULT, ADC_DB_USER_LINPTS_DEFAULT, ADC_DB_USER_LINPTS_DEFAULT, ADC_DB_USER_LINPTS_DEFAULT,
    ADC_DB_USER_LINPTSLAST_DEFAULT
  },
  .linCor =
  {                                           /* int32s linCor[ADC_MAX_NUM_LIN_PTS+1]; */
    ADC_DB_USER_LINCOR_DEFAULT, ADC_DB_USER_LINCOR_DEFAULT, ADC_DB_USER_LINCOR_DEFAULT, ADC_DB_USER_LINCOR_DEFAULT, ADC_DB_USER_LINCOR_DEFAULT,
    ADC_DB_USER_LINCOR_DEFAULT, ADC_DB_USER_LINCOR_DEFAULT, ADC_DB_USER_LINCOR_DEFAULT, ADC_DB_USER_LINCOR_DEFAULT, ADC_DB_USER_LINCOR_DEFAULT,
    ADC_DB_USER_LINCOR_DEFAULT
  },

  .scaleType = ADC_DB_USER_SCALETYPE_DEFAULT,
  .syncFreqIdx = ADC_DB_USER_SYNCFREQIDX_DEFAULT,
  .jitter = ADC_DB_USER_JITTER_DEFAULT,
  .dpPos = ADC_DB_USER_DPPOS_DEFAULT,
  .fullScale = { ADC_DB_USER_FULLSCALE0_DEFAULT, ADC_DB_USER_FULLSCALE1_DEFAULT, ADC_DB_USER_FULLSCALE2_DEFAULT },
  .resolution = { ADC_DB_USER_RESOLUTION0_DEFAULT, ADC_DB_USER_RESOLUTION1_DEFAULT, ADC_DB_USER_RESOLUTION2_DEFAULT },
  .units = ADC_DB_USER_UNITS_DEFAULT,
  .calWeight = ADC_DB_USER_CALWEIGHT_DEFAULT,
  .addTare = ADC_DB_USER_ADDTARE_DEFAULT,
  .zeroBand = ADC_DB_USER_ZEROBAND_DEFAULT,
  .motion = ADC_DB_USER_MOTION_DEFAULT,
  .zeroTracking = ADC_DB_USER_ZEROTRACKING_DEFAULT,
  .average = ADC_DB_USER_AVERAGE_DEFAULT,
  .rangeOfZero = ADC_DB_USER_RANGEOFZERO_DEFAULT,
  .approvalMode = ADC_DB_USER_APPROVALMODE_DEFAULT,
  .zeroOnStart = ADC_DB_USER_ZEROONSTART_DEFAULT,
  .wire4Mode = ADC_DB_USER_WIRE4MODE_DEFAULT,
  .tempCalZero = ADC_DB_USER_TEMPCALZERO_DEFAULT,
  .hiRes = ADC_DB_USER_HIRES_DEFAULT,

  .qaEnable = ADC_DB_USER_QAENABLE_DEFAULT,
  .qaDay = ADC_DB_USER_QADAY_DEFAULT,
  .qaMonth = ADC_DB_USER_QAMONTH_DEFAULT,
  .qaYear = ADC_DB_USER_QAYEAR_DEFAULT,

  .calCntNTEPCalibration = ADC_DB_USER_CALCNTNTEPCALIBRATION_DEFAULT,
  .calCntNTEPSettings = ADC_DB_USER_CALCNTNTEPSETTINGS_DEFAULT,
  .calCntOIML = ADC_DB_USER_CALCNTOIML_DEFAULT,
  .calFlag = ADC_DB_USER_CALFLAG_DEFAULT,

  .chksum = ADC_DB_USER_CHKSUM_DEFAULT
};


/**
 * Set a user database to defaults
 *
 * @param aud is the user database to be defaulted
 * @param preserve_cal_counters causes the calibration counters to preserved (and incremented)
 * @param cal_only causes only the calibration settings to be defaulted (zero, span, span count, linearisation points)
 */
void AdcIntUserDatabaseDefault(AdcUserDatabase *aud, bool preserve_cal_counters, bool cal_only)
{
  int32u calCntNTEPCalibration;
  int32u calCntNTEPSettings;
  int32u calCntOIML;
  int8u i;

  /* Save current cal counters */
  calCntNTEPCalibration = aud->calCntNTEPCalibration;
  calCntNTEPSettings = aud->calCntNTEPSettings;
  calCntOIML = aud->calCntOIML;

  if (cal_only)
  {
    /* Default only the calibration settings */
    aud->zeroRaw = ADC_DB_USER_ZERORAW_DEFAULT;
    aud->spanRaw = ADC_DB_USER_SPANRAW_DEFAULT;
    aud->spanCnt = ADC_DB_USER_SPANCNT_DEFAULT;
    aud->calWeight = ADC_DB_USER_CALWEIGHT_DEFAULT;
    for (i = 0; i < ADC_MAX_NUM_LIN_PTS; i++)
    {
      aud->linPts[i] = ADC_DB_USER_LINPTS_DEFAULT;
      aud->linCor[i] = ADC_DB_USER_LINCOR_DEFAULT;
    }
    aud->linPts[ADC_MAX_NUM_LIN_PTS] = ADC_DB_USER_LINPTSLAST_DEFAULT;
    aud->linCor[ADC_MAX_NUM_LIN_PTS] = ADC_DB_USER_LINCOR_DEFAULT;
  }
  else
  {
    /* Default the entire database */
    memset(aud, 0, sizeof(AdcUserDatabase));
    *aud = s_C_DefaultAdcUserDB;
  }

  /* Restore cal counters */
  if (preserve_cal_counters)
  {
    aud->calCntNTEPCalibration = calCntNTEPCalibration;
    aud->calCntNTEPSettings = calCntNTEPSettings;
    aud->calCntOIML = calCntOIML;

    /* And increment */
    aud->calCntNTEPCalibration++;
    if (!cal_only)
      aud->calCntNTEPSettings++;
    aud->calCntOIML++;
  }
}


/** Check the user database checksum */
bool AdcIntUserDatabaseCheckChecksum(const AdcUserDatabase *aud)
{
  return DB_CHECKSUM_GOOD(AdcUserDatabase, *aud, ADC_DB_USER_CHECKSUM_SEED);
}


/** Set the user database checksum */
void AdcIntUserDatabaseSetChecksum(AdcUserDatabase *aud)
{
  DB_SET_CHECKSUM(AdcUserDatabase, *aud, ADC_DB_USER_CHECKSUM_SEED);
}


/** Increment the calibration counters if trade relevant settings have changed */
void AdcIntUserDatabaseIncrementCalCountersOnChanges(AdcSuperStructure *ass, AdcUserDatabase *changed_db, const AdcUserDatabase *ref_db)
{
  bool bCalInc = false;
  bool bConfigInc = false;

  /* Check for ADC user database changes that affect calibration and configuration counters */
  if (NULL != ass->adcMiscellaneous.adcDatabaseCalCountCheckImplementerFn)
  {
    ass->adcMiscellaneous.adcDatabaseCalCountCheckImplementerFn(ass, changed_db, ref_db, &bCalInc, &bConfigInc);
  }

  /* Increment the counters as required */
  if (bCalInc)
  {
    changed_db->calCntNTEPCalibration++;
  }
  if (bConfigInc)
  {
    changed_db->calCntNTEPSettings++;
  }
  if (bCalInc || bConfigInc)
  {
    changed_db->calCntOIML++;
  }
}


/** Compare user databases, return true if different */
bool AdcIntUserDatabaseDiff(const AdcUserDatabase *aud1, const AdcUserDatabase *aud2)
{
  int i;
  bool different;

  /* Check linearisation */
  for (i = 0, different = false; i < ADC_MAX_NUM_LIN_PTS + 1 && !different; i++)
  {
    if (aud1->linPts[i] != aud2->linPts[i] ||
        aud1->linCor[i] != aud2->linCor[i])
    {
      different = true;
    }
  }

  /* Check full scale and resolution */
  for (i = 0; i < ADC_MAX_NUM_INTERVAL && !different; i++)
  {
    if (aud1->fullScale[i] != aud2->fullScale[i] ||
        aud1->resolution[i] != aud2->resolution[i])
    {
      different = true;
    }
  }

  /* Check the rest */
  if (!different)
  {
    different = (aud1->zeroRaw != aud2->zeroRaw ||
                 aud1->spanRaw != aud2->spanRaw ||
                 aud1->spanCnt != aud2->spanCnt ||
                 aud1->scaleType != aud2->scaleType ||
                 aud1->syncFreqIdx != aud2->syncFreqIdx ||
                 aud1->jitter != aud2->jitter ||
                 aud1->dpPos != aud2->dpPos ||
                 aud1->units != aud2->units ||
                 aud1->calWeight != aud2->calWeight ||
                 aud1->addTare != aud2->addTare ||
                 aud1->zeroBand != aud2->zeroBand ||
                 aud1->motion != aud2->motion ||
                 aud1->zeroTracking != aud2->zeroTracking ||
                 aud1->average != aud2->average ||
                 aud1->rangeOfZero != aud2->rangeOfZero ||
                 aud1->approvalMode != aud2->approvalMode ||
                 aud1->zeroOnStart != aud2->zeroOnStart ||
                 aud1->wire4Mode != aud2->wire4Mode ||
                 aud1->tempCalZero != aud2->tempCalZero ||
                 aud1->hiRes != aud2->hiRes ||
                 aud1->qaEnable != aud2->qaEnable ||
                 aud1->qaDay != aud2->qaDay ||
                 aud1->qaMonth != aud2->qaMonth ||
                 aud1->qaYear != aud2->qaYear ||
                 aud1->calCntNTEPCalibration != aud2->calCntNTEPCalibration ||
                 aud1->calCntNTEPSettings != aud2->calCntNTEPSettings ||
                 aud1->calCntOIML != aud2->calCntOIML ||
                 aud1->calFlag != aud2->calFlag);
  }
  return different;
}
