/*
 * adcdsd.h
 *
 *  Created on: Mar 6, 2012
 *      Author: ian
 */

#ifndef ADCDSD_H_
#define ADCDSD_H_

#include <stdint.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcmuteximpl.h>


/** Length of the file index magic string */
#define ADC_DSD_FILE_INDEX_MAGIC_LENGTH (8)


/** A magic string at the start of @ref struct s_adc_dsd_file_index to identify DSD files. Not NULL terminated. */
#define ADC_DSD_FILE_INDEX_MAGIC "RinstDSD"


/** Checksum seed for header CRC */
#define ADC_DSD_FILE_HEADER_CHECKSUM_SEED (0x1234)

#define ADC_DSD_FILE_INDEX_VERSION_UNKNOWN (0)
#define ADC_DSD_FILE_INDEX_VERSION_1 (1)


#define ADC_DSD_FILE_INDEX_VERSION_CURRENT ADC_DSD_FILE_INDEX_VERSION_1


#pragma pack(push)
#pragma pack(1)

/** DSD file index */
struct s_adc_dsd_file_index
{
  char dsd_magic[ADC_DSD_FILE_INDEX_MAGIC_LENGTH];      /*!< Magic string at the start of a DSD file, used to identify DSD files */
  int16u dsd_version;               /*!< Version of the DSD file. When things change, this needs to change. */
  int32u dsd_recordsize;            /*!< Size of each record in the DSD */
  int32u dsd_maxrecords;            /*!< Maximum number of records in the DSD */
  int32u dsd_newestindex;           /*!< The position of the newest record in the DSD */
  int32u dsd_oldestindex;           /*!< The position of the oldest record in the DSD */
  int32u dsd_recordcount;           /*!< The number of records currently stored in the DSD (so we don't lose a record in the circular buffer implementation) */
  int16u dsd_recordcrcseed;         /*!< The seed for the CRC checksum on each record */
  int32u dsd_serialnumber;          /*!< The serial number of the indicator that created the DSD file */

  int16u chksum;                    /*!< The CRC16 checksum of all other members of @ref struct s_adc_dsd_file_index */
};


/** DSD record header. Each record has one of these */
struct s_adc_dsd_record_header
{
  int16u crc;
};

#pragma pack(pop)


/** Information about the DSD */
struct s_adc_dsd
{
  AdcMutexImplementer *mutex_implementation;  /*!< DSD structure protection mutex */
  char *filename;                             /*!< Filename of the DSD. Is NULL if no DSD opened */
  int fd;                                     /*!< File descriptor of the DSD file. Is < 0 if no DSD opened */
  int8u *work_buffer;                         /*!< Internal work buffer of size max_recordsize */
  struct s_adc_dsd_file_index file_index;     /*!< Information about the DSD file */
};


/** DSD errors */
enum eAdcDsdError
{
  ADC_DSD_SUCCESS = 0,                /*!< Operation completed successfully */
  ADC_DSD_ERROR_FILE,                 /*!< File operation error */
  ADC_DSD_ERROR_MALLOC,               /*!< Memory allocation error */
  ADC_DSD_ERROR_NULL,                 /*!< NULL pointer passed to function */
  ADC_DSD_ERROR_RECORDSIZE,           /*!< Record size problem (ie trying to write a record that is too large) */
  ADC_DSD_ERROR_RECORDCHKSUM,         /*!< The record checksum doesn't match the one calculated from the data */
  ADC_DSD_ERROR_MUTEX,                /*!< Problem with the DSD mutex */
  ADC_DSD_ERROR_EMPTY,                /*!< No records in DSD */
  ADC_DSD_ERROR_FULL,                 /*!< No room to add more records to DSD */
  ADC_DSD_ERROR_OUTOFRANGE,           /*!< Trying to retrieve a record beyond number of records in the DSD */
  ADC_DSD_ERROR_FILEINDEXCHKSUM,      /*!< The file index is corrupt */
  ADC_DSD_ERROR_MISC                  /*!< Something is really amiss */
};


/*
 * DSD structure creation/destruction functions
 */


/**
 * Create and initialise the DSD structure, but not the file.
 * Need to call AdcIntDSDOpenExisting() or AdcIntDSDCreateNew() to access the file.
 *
 * @param dsd is a pointer to a pointer to a DSD structure that is set to point to the new structure created by the function. Not NULL
 * @param mutex_implementation is the implementation of the mutex. Can be NULL for no mutex.
 *  */
enum eAdcDsdError AdcDSDInit(struct s_adc_dsd **dsd, AdcMutexImplementer *mutex_implementation);


/** Close and clean up the DSD structure */
enum eAdcDsdError AdcDSDDestroy(struct s_adc_dsd *dsd);


/*
 * DSD file open/close functions
 */


/**
 * Create a new DSD and ready it for use.
 * Will wipe the file if it already exists
 * DSD parameters will be saved to the file.
 * (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param filename is the name of the DSD file to be created
 * @param recordsize is the size in bytes of a DSD record
 * @param maxrecords is the maximum number of records allowed in the DSD
 * @param recordcrcseed is the 16bit CRC seed used to seed the checksum of each record
 * @param serialnumber is the serial number of the device creating the DSD
 */
enum eAdcDsdError AdcDSDCreateNew(struct s_adc_dsd *dsd, bool lock, const char *filename, int32u recordsize, int32u maxrecords, int16u recordcrcseed, int32u serialnumber);


/**
 * Open an existing DSD read for use.
 * Will fail if file doesn't exist.
 * DSD parameters will be retrieved from the file. (no locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param filename is the name of the DSD file to be opened
 */
enum eAdcDsdError AdcDSDOpenExisting(struct s_adc_dsd *dsd, bool lock, const char *filename);


/**
 * Close the current DSD. dsd structure remains. (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 */
enum eAdcDsdError AdcDSDClose(struct s_adc_dsd *dsd, bool lock);


/**
 * Check if the DSD is up and working (file has been opened)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 */
enum eAdcDsdError AdcDSDIsRunning(struct s_adc_dsd *dsd, bool lock);


/*
 * Lock/unlock functions
 */


/** Lock the DSD structure */
enum eAdcDsdError AdcDSDMutexLock(struct s_adc_dsd *dsd);


/** Release the DSD structure */
enum eAdcDsdError AdcDSDMutexUnlock(struct s_adc_dsd *dsd);


/*
 * Writing functions
 */


/**
 * Write a record into the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param buffer is a pointer to a buffer for the record to be written
 * @param buflen is the length of the buffer
 */
enum eAdcDsdError AdcDSDWriteRecord(struct s_adc_dsd *dsd, bool lock, const void *buffer, const int16u buflen);


/*
 * Read functions
 */


/**
 * Read a record from the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param record is the DSD record number to read
 * @param buffer is a pointer to a buffer for the record to be read into
 * @param buflen is the length of the buffer
 */
enum eAdcDsdError AdcDSDReadRecord(struct s_adc_dsd *dsd, bool lock, int32u record, void *buffer, int16u buflen);


/*
 * Management functions
 */


/**
 * Get the number of records currently held in the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param records is a pointer to int32u set on success to the number of records in the DSD. Not NULL.
 */
enum eAdcDsdError AdcDSDGetNumberRecords(struct s_adc_dsd *dsd, bool lock, int32u *records);


/**
 * Get the total number of records that will fit in the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param records is a pointer to int32u set on success to the capacity of the DSD. Not NULL.
 */
enum eAdcDsdError AdcDSDGetCapacity(struct s_adc_dsd *dsd, bool lock, int32u *records);


/**
 * Get the record size for the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param recordsize is a pointer to int32u set on success to the number of bytes of a record. Not NULL.
 */
enum eAdcDsdError AdcDSDGetRecordSize(struct s_adc_dsd *dsd, bool lock, int32u *recordsize);


/**
 * Verify the integrity of all records in the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param fix if true, then tries to fix any problems
 * @param record_num is set to the first record found with a problem
 */
enum eAdcDsdError AdcDSDVerifyIntegrity(struct s_adc_dsd *dsd, bool lock, bool fix, size_t *record_num);


/**
 * Make room for more records, by purging num_records from the start of the DSD (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param num_records is the number of records to purge from the DSD.
 */
enum eAdcDsdError AdcDSDPurge(struct s_adc_dsd *dsd, bool lock, size_t num_records);


/**
 * Read a file index from the DSD file.
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param which is 0 for the first index, or 1 for the second index
 * @param fi is the returned file index. Not NULL
 */
enum eAdcDsdError AdcDSDReadFileIndex(struct s_adc_dsd *dsd, bool lock, int which, struct s_adc_dsd_file_index *fi);


/**
 * Calculate the checksum on a file index taking into account
 * the following:
 *
 * Checksum is special: so it will work irrespective of machine endianness, we
 * first put the file index contents into network byte order, then calculate the
 * checksum, then set the network order version in the header.
 *
 * The checksum returned will be in host order, it's up to you to htons() it
 */
void AdcDSDCalculateFileIndexChecksum(const struct s_adc_dsd_file_index *fi, int16u *checksum);


/** Define to calculate a database checksum. Relies on db containing a member "chksum" */
#define ADC_DSD_CHECKSUM(dbtype, db, seed) DB_CHECKSUM(dbtype, db, seed)


/*
 * Backup operations
 */


/**
 * Copy the DSD file to copyfd (with locking)
 *
 * Writing within copyfd occurs from the write position in the file when the function is called.
 * No truncation is performed.
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param copyfd is a file descriptor open for writing.
 * @param bytes_written is set to the number of bytes written to copyfd. Can be NULL
 */
enum eAdcDsdError AdcDSDFileCopyFD(struct s_adc_dsd *dsd, bool lock, int copyfd, int32u *bytes_written);


/**
 * Copy the DSD file to filename (with locking)
 *
 * @param dsd is the DSD structure
 * @param lock allows the structure to be locked prior to performing the operation
 * @param filename is the name of a writeable file to copy the DSD to. File will be overwritten.
 * @param bytes_written is set to the number of bytes written to copyfd. Can be NULL
 */
enum eAdcDsdError AdcDSDFileCopyFilename(struct s_adc_dsd *dsd, bool lock, const char *filename, int32u *byteswritten);


#endif /* ADCDSD_H_ */
