#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/stdrinstrum.h>
#include <rinhw/rstdio.h>
#include <rinhw/hw_vers.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adczerotrack.h>
#include <rinweigh/adcoperations.h>
#include <rinweigh/adcreadings.h>
#include <rinweigh/adcads1216.h>
#include <rinweigh/adclin.h>
#include <rinweigh/adcdebug.h>
#include <rinweigh/adctemperature.h>
#include <rinweigh/adcmiscellaneous.h>
#include <rinweigh/adcmutex.h>
#include <rinweigh/adcconvert.h>
#include <rinweigh/adcdbdefault.h>
#include <rinweigh/adcdbfactory.h>
#include <rinweigh/adcdbuser.h>
#include <rinweigh/adcdbruntime.h>
#include <rinweigh/adcdbsave.h>

#include <rinweigh/adcinterface.h>


/**
 * This procedure re-initialises all aspects of the ADC operation, without locking the structure.
 * It can be called at any time after AdcInit() has been called
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param afd is a pointer to a factory database to be copied to the internal one. If NULL, no change is made to the internal one
 * @param aud is a pointer to a user database to be copied to the internal one. If NULL, no change is made to the internal one
 * @param ard is a pointer to a runtime database to be copied to the internal one. If NULL, no change is made to the internal one
 */
static void AdcIntReInit(AdcSuperStructure *ass, const AdcFactoryDatabase *afd, const AdcUserDatabase *aud, const AdcRuntimeDatabase *ard)
{

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflReInitStart);

  ass->adcMiscellaneous.adcErrors = 0;

  /* Check checksum on passed factory database */
  if (NULL != afd)
  {
    ass->adcFactoryDatabase = *afd;
    if (!AdcIntFactoryDatabaseCheckChecksum(&ass->adcFactoryDatabase))
    {
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_FACTORY_DB;
    }
  }

  /* Check checksum on passed user database */
  if (NULL != aud)
  {
    ass->adcUserDatabase = *aud;
    if (!AdcIntUserDatabaseCheckChecksum(&ass->adcUserDatabase))
    {
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_USER_DB;
    }
  }

  /* Check checksum on passed runtime database */
  if (NULL != ard)
  {
    ass->adcRuntimeDatabase = *ard;
    if (!AdcIntRuntimeDatabaseCheckChecksum(&ass->adcRuntimeDatabase))
    {
      ass->adcMiscellaneous.adcErrors |= ADC_ERR_RUNTIME_DB;
    }
  }

  /* Cal flag not set generates an error */
  if (ass->adcFactoryDatabase.calFlag != ADC_CAL_FLAG_OK)
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_FACTORY_DB;
  if (ass->adcUserDatabase.calFlag != ADC_CAL_FLAG_OK)
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_USER_DB;

  AdcIntReInitReadings(ass);

  AdcIntReInitOperations(ass);

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflReInitEnd);
}


/**
 * This procedure re-initialises all aspects of the ADC operation after locking the structure.
 * It can be called at any time after AdcInit() has been called
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param afd is a pointer to a factory database to be copied to the internal one. If NULL, no change is made to the internal one
 * @param aud is a pointer to a user database to be copied to the internal one. If NULL, no change is made to the internal one
 * @param ard is a pointer to a runtime database to be copied to the internal one. If NULL, no change is made to the internal one
 */
void AdcReInit(AdcSuperStructure *ass, bool lock, const AdcFactoryDatabase *afd, const AdcUserDatabase *aud, const AdcRuntimeDatabase *ard)
{
  ADC_INT_IFACE_LOCK();
  AdcIntReInit(ass, afd, aud, ard);
  ADC_INT_IFACE_UNLOCK();
}


/**
 * You should call AdcPrepare() before any calls to the ADC library.
 * This will setup various fixed aspects of ADC option (such as
 * mutex, ADC device, get temperature, get date/time, DB defaulter,
 * DB saver, App callback, DB calcount check and live weigh implementer)
 *
 * This allows for calls to AdcDefaultDatabases() before AdcInit()
 * is called. This may be required because the application databases
 * need to be present for the call to AdcInit(), but they may not
 * be valid, and have to be defaulted via AdcDefaultDatabases().
 *
 * Do not call more than once (per ass).
 *
 * The sequence of startup operations must be:
 *   AdcPrepare()
 *   Application load databases, and call AdcDefaultDatabases() if necessary
 *   AdcInit()
 *   Application start ADC thread to call AdcUpdate()
 *
 * @param ass is a pointer to an un-initialised Adc Super Structure. Not NULL.
 * @param adcdev is a running 24bit ADC device to get readings from. Not NULL.
 * @param adc_get_temp is a function pointer used by the ADC library to get the current temperature. Not NULL.
 * @param adc_get_datetime is a function pointer used by the ADC to get the current date and time. Not NULL.
 * @param adcDatabaseDefaultImplementerFn is a function pointer used by the ADC library to default some or all of the databases. Not NULL.
 * @param adcDatabaseSaveImplementerFn is a function pointer used by the ADC library to save some or all of the databases. Can be NULL in which case databases won't be saved by the library.
 * @param adcAppCallbackImplementerFn is a function pointer used for calling back to your application when specific events occur. Can be NULL in which case it won't be called
 * @param adcDatabaseCalCountCheckImplementerFn is a function pointer used for calling back to your application for checking for ADC user database changes that affect calibration and configuration counters. Not NULL
 * @param mutex_implementation is the implementation of the exclusive access mutex. Can be NULL if no locking is required.
 * @param liveweigh_implementation is the implementation of the live weigh algorithm. Can be NULL if not required.
 * @param adcDebugVPrintfImplementerFn is a function pointer used by the ADC library for calling back to your application for debug printfs. Can be NULL in which case no debug printfs will be made.
 *
 * @return true if initialised correctly
 */
bool AdcPrepare(AdcSuperStructure *ass, struct t_io_dev *adcdev,
                t_adc_get_temperature_fn adc_get_temp,
                t_adc_get_datetime_fn adc_get_datetime,
                AdcDatabaseDefaultImplementerFn adcDatabaseDefaultImplementerFn,
                AdcDatabaseSaveImplementerFn adcDatabaseSaveImplementerFn,
                AdcAppCallbackImplementerFn adcAppCallbackImplementerFn,
                AdcDatabaseCalCountCheckImplementerFn adcDatabaseCalCountCheckImplementerFn,
                AdcMutexImplementer *mutex_implementation,
                AdcLiveWeighImplementer *liveweigh_implementation,
                AdcDebugVPrintfImplementerFn adcDebugVPrintfImplementerFn
               )
{
  bool hw_init_result;

  /* Check parameters */
  if (NULL == ass ||
      NULL == adcdev ||
      NULL == adc_get_temp ||
      NULL == adc_get_datetime ||
      NULL == adcDatabaseDefaultImplementerFn ||
      NULL == adcDatabaseCalCountCheckImplementerFn)
  {
    return false;
  }

  /* Setup debug */
  AdcIntDebugSetVPrintfImplementerFn(adcDebugVPrintfImplementerFn);

  /* Setup mutex */
  ass->adcMutex = mutex_implementation;

  /* Lock the structure */
  AdcMutexLock(ass);

  /* Call back to application */
  if (NULL != adcAppCallbackImplementerFn)
  {
    adcAppCallbackImplementerFn(ass, acflPrepareStart);
  }

  /* Adc get temperature function, implemented in the user application */
  AdcIntTemperatureInit(ass, adc_get_temp);

  /* Setup ADC hardware */
  hw_init_result = AdcIntHwInit(&ass->adcHwInfo, adcdev);

  /* Setup live weigh */
  ass->adcLiveWeigh = liveweigh_implementation;

  /* Initialise miscellaneous, including get date & time function implemented in the user application */
  AdcIntInitMiscellaneous(&ass->adcMiscellaneous,
                          adc_get_datetime,
                          adcDatabaseDefaultImplementerFn,
                          adcDatabaseSaveImplementerFn,
                          adcAppCallbackImplementerFn,
                          adcDatabaseCalCountCheckImplementerFn);

  if (!hw_init_result)
  {
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_HARDWARE;
  }

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflPrepareEnd);

  AdcMutexUnlock(ass);
  return true;
}


/**
 * This procedure initialises all aspects of the ADC operation.
 * It needs to be called after AdcPrepare(), but before any other adc procedure or data is used.
 * Do not call more than once (per ass).
 *
 * Postconditions: Adc library is ready for operation
 *
 * @param ass is a pointer to an un-initialised Adc Super Structure. Not NULL.
 * @param afd is the factory database to be copied into internal data. Can be NULL, in which case internal factory database is defaulted.
 * @param aud is the user database to be copied into internal data. Can be NULL, in which case internal user database is defaulted.
 * @param ard is the runtime database to be copied to into internal one. Can be NULL, in which case internal runtime database is defaulted.
 *
 * @return true if initialised correctly
 */
bool AdcInit(AdcSuperStructure *ass,
             const AdcFactoryDatabase *afd,
             const AdcUserDatabase *aud,
             const AdcRuntimeDatabase *ard
            )
{
  AdcErrors err_save;

  /* Check parameters */
  if (NULL == ass)
  {
    return false;
  }

  /* Lock the structure */
  AdcMutexLock(ass);

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflInitStart);

  /* Get the current temperature */
  AdcIntUpdateTemperature(ass);

  /* Initialise operations */
  AdcIntInitOperations(ass);

  /* Initialise databases if NULL pointers passed */
  if (NULL == afd)
  {
    AdcIntDefaultDatabases(ass, &ass->adcFactoryDatabase, NULL, NULL, false, false);
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_FACTORY_DB;
  }
  if (NULL == aud)
  {
    AdcIntDefaultDatabases(ass, NULL, &ass->adcUserDatabase, NULL, false, false);
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_USER_DB;
  }
  if (NULL == ard)
  {
    AdcIntDefaultDatabases(ass, NULL, NULL, &ass->adcRuntimeDatabase, false, false);
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_RUNTIME_DB;
  }

  /* Setup all aspects of ADC library */
  err_save = ass->adcMiscellaneous.adcErrors;
  AdcIntReInit(ass, afd, aud, ard);
  ass->adcMiscellaneous.adcErrors |= err_save;

  /* Setup diff runtime database */
  ass->adcRuntimeDatabaseDiff = ass->adcRuntimeDatabase;

  /* Do zero on startup */
  if (!ass->adcUserDatabase.zeroOnStart)
  {
    /* Remove initial zero if none is used */
    ass->adcInternalData.cntInitZero = 0;
  }
  else
  {
    AdcInitialZero(ass, false, false);
  }

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflInitEnd);

  AdcMutexUnlock(ass);

  return true;
}


/** Clean up after ADC library */
void AdcDestroy(AdcSuperStructure *ass)
{
  AdcMutexLock(ass);

  /* Clean up operations */
  AdcIntDestroyOperations(ass);

  /* Clean up ADC hardware */
  AdcIntHwDestroy(&ass->adcHwInfo);

  /* Set pointers to NULL */
  ass->adcTempCal.adc_get_temperature_fn = NULL;
  ass->adcMiscellaneous.adcDatabaseDefaultImplementerFn = NULL;
  ass->adcMiscellaneous.adcDatabaseSaveImplementerFn = NULL;
  ass->adcMiscellaneous.adc_get_datetime_fn = NULL;
  ass->adcLiveWeigh = NULL;

  AdcMutexUnlock(ass);

  ass->adcMutex = NULL;

  /* Clean up mutex, liveweigh done outside ADC library */
}


/**
 * This procedure must be called in it's own thread. It will rate limit to
 * the sync frequency.
 *
 * It obtains a new adc value, performs all calculations to produce trade
 * weights with status information, and processes user operations such as
 * semi-automatic zero.
 *
 * Preconditions: InitAdcOperation has been called
 * Postconditions: If returns TRUE, the weight data structure is updated
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param data_wait_failed is a pointer to a bool that is set if there was a read problem,
 *        and you should implement a delay elsewhere, to prevent the ADC thread going to 100% CPU, not NULL
 * @return TRUE if new weight data is available, else FALSE
 */
bool AdcUpdate(AdcSuperStructure *ass, bool lock, int32u appErrs, bool *data_wait_failed)
{
  bool result = false;

#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdate called\n");
#endif

  /* Lock everything while working */
  ADC_INT_IFACE_LOCK();

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflUpdateStart);

  ass->adcMiscellaneous.adcErrors &= ~APP_ERR_MASK;
  ass->adcMiscellaneous.adcErrors |= (appErrs & APP_ERR_MASK); // must AND to avoid loops

  AdcIntOperationTickBefore(ass);

  if (AdcIntUpdateReadings(ass, data_wait_failed))
  {
    AdcIntZeroTrackingDo(ass);

    AdcIntOperationTick(ass);

    /* Save runtime database on change */
    if (AdcIntRuntimeDatabaseDiff(&ass->adcRuntimeDatabase, &ass->adcRuntimeDatabaseDiff))
    {
      AdcIntSaveDatabases(ass, NULL, NULL, &ass->adcRuntimeDatabase, NULL);
      ass->adcRuntimeDatabaseDiff = ass->adcRuntimeDatabase;
    }

    if (!ass->adcInternalData.userBlocked)
    {
      AdcIntUpdateTemperature(ass);
      result = true;
    }
  }

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflUpdateEnd);

  /* Unlock */
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Convert internal ADC counts into final (displayable) numbers,
 * taking into account the resolution of the specified range,
 * and whether an extra digit of resolution is required.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the reading in ADC counts
 * @param range is the current range (0, 1, 2)
 * @param hiRes is a flag to indicate that the final number should have an extra digit of resolution
 * @return the rdg converted to final (displayable) numbers
 */
int32s AdcCountsToFinal(AdcSuperStructure *ass, bool lock, int32s rdg, int8u range, bool hiRes)
{
  int32s result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntCountsToFinal(ass, rdg, range, hiRes, ADC_UNIT_SWITCH_PRIMARY);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Convert final (displayable) numbers into internal ADC counts,
 * taking into account whether an extra digit of resolution is present.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the final (displayable) number
 * @param hiRes is a flag to indicate that rdg has an extra digit of resolution
 * @return the rdg converted to internal ADC counts
 */
int32s AdcFinalToCounts(AdcSuperStructure *ass, bool lock, int32s rdg, bool hiRes)
{
  int32s result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntFinalToCounts(ass, rdg, hiRes, ADC_UNIT_SWITCH_PRIMARY);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Convert a final (displayable) number with (possibly) extra resolution
 * into a final (displayable) number with no extra resolution
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the final number with (possibly) extra resolution
 * @param range is the current range of the final number
 * @param hiRes is a flag to indicate that rdg has extra resolution
 * @return rdg with any extra resolution removed
 */
int32s AdcHiResToNormal(AdcSuperStructure *ass, bool lock, int32s rdg, int8u range, bool hiRes)
{
  int32s result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntHiResToNormal(ass, rdg, range, hiRes);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Convert internal ADC counts into final alternative units (displayable) numbers,
 * taking into account the resolution of the specified range,
 * and whether an extra digit of resolution is required.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the reading in ADC counts
 * @param range is the current range (0, 1, 2)
 * @param hiRes is a flag to indicate that the final number should have an extra digit of resolution
 * @return the rdg converted to final alternative units (displayable) numbers
 */
int32s AdcCountsToFinalAlt(AdcSuperStructure *ass, bool lock, int32s rdg, int8u range, bool hiRes)
{
  int32s result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntCountsToFinal(ass, rdg, range, hiRes, ADC_UNIT_SWITCH_SECONDARY);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Convert final alternative units (displayable) numbers into internal ADC counts,
 * taking into account whether an extra digit of resolution is present.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the final alternative units (displayable) number
 * @param hiRes is a flag to indicate that rdg has an extra digit of resolution
 * @return the rdg converted to internal ADC counts
 */
int32s AdcFinalToCountsAlt(AdcSuperStructure *ass, bool lock, int32s rdg, bool hiRes)
{
  int32s result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntFinalToCounts(ass, rdg, hiRes, ADC_UNIT_SWITCH_SECONDARY);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Convert internal ADC counts into final tertiary units (displayable) numbers,
 * taking into account the resolution of the specified range,
 * and whether an extra digit of resolution is required.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the reading in ADC counts
 * @param range is the current range (0, 1, 2)
 * @param hiRes is a flag to indicate that the final number should have an extra digit of resolution
 * @return the rdg converted to final tertiary units (displayable) numbers
 */
int32s AdcCountsToFinalTert(AdcSuperStructure *ass, bool lock, int32s rdg, int8u range, bool hiRes)
{
  int32s result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntCountsToFinal(ass, rdg, range, hiRes, ADC_UNIT_SWITCH_TERTIARY);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Convert final tertiary units (displayable) numbers into internal ADC counts,
 * taking into account whether an extra digit of resolution is present.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the final tertiary units (displayable) number
 * @param hiRes is a flag to indicate that rdg has an extra digit of resolution
 * @return the rdg converted to internal ADC counts
 */
int32s AdcFinalToCountsTert(AdcSuperStructure *ass, bool lock, int32s rdg, bool hiRes)
{
  int32s result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntFinalToCounts(ass, rdg, hiRes, ADC_UNIT_SWITCH_TERTIARY);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Returns the current command identifier (a number that is incremented each time a command is started)
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
int8u AdcGetCurrentCommandCnt(AdcSuperStructure *ass, bool lock)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = ass->adcOperation.cmdCnt;
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Returns an identifier of the current command in progress (one of the ADC_OP_CMD_* values).
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
int8u AdcGetCurrentCommand(AdcSuperStructure *ass, bool lock)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = ass->adcOperation.lastCommand;
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Return current operation status. This is one of the ADC_OP_STATUS_* codes.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
int8u AdcGetCurrentOpStatus(AdcSuperStructure *ass, bool lock)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = ass->adcOperation.status;
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a semi-automatic zero
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_BAND if outside zero range, ADC_OP_STATUS_SUCCESS if zero succeeds
 */
int8u AdcSemiautoZero(AdcSuperStructure *ass, bool lock, bool block, bool test, bool immediate)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  if (test)
  {
    result = AdcIntSemiautoZeroOperationIsValid(ass, immediate);
  }
  else
  {
    result = AdcIntOperationStart(ass, ADC_OP_CMD_SEMIAUTO_ZERO, 10, block, 0, 0, immediate);
  }
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Clear a previous semi-automatic zero (only in non-trade mode)
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcClearSemiautoZero(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CLEAR_SEMIAUTO_ZERO, 10, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a semi-automatic tare
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_BAND if outside range of tare, ADC_OP_STATUS_SUCCESS if tare succeeds
 */
int8u AdcSemiautoTare(AdcSuperStructure *ass, bool lock, bool block, bool test, bool immediate)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  if (test)
  {
    result = AdcIntSemiautoTareOperationIsValid(ass, immediate);
  }
  else
  {
    result = AdcIntOperationStart(ass, ADC_OP_CMD_SEMIAUTO_TARE, 10, block, 0, 0, immediate);
  }
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a Gross/Net toggle
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcGrossNetToggle(AdcSuperStructure *ass, bool lock, bool block)
{
  ADC_INT_IFACE_LOCK();
  if (ass->adcRuntimeDatabase.netMode ||
      ass->adcRuntimeDatabase.cntTare != 0 ||
      ass->adcRuntimeDatabase.cntPTare != 0)
  {
    ass->adcRuntimeDatabase.netMode = !ass->adcRuntimeDatabase.netMode;
  }
  ADC_INT_IFACE_UNLOCK();

  return ADC_OP_STATUS_SUCCESS;
}


/**
 * Switch the scale to gross
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcToGross(AdcSuperStructure *ass, bool lock, bool block)
{
  ADC_INT_IFACE_LOCK();
  ass->adcRuntimeDatabase.netMode = FALSE;
  ADC_INT_IFACE_UNLOCK();
  return ADC_OP_STATUS_SUCCESS;
}


/**
 * Switch the scale to net
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcToNet(AdcSuperStructure *ass, bool lock, bool block)
{
  ADC_INT_IFACE_LOCK();
  if (ass->adcRuntimeDatabase.cntTare != 0 || ass->adcRuntimeDatabase.cntPTare != 0)
  {
    ass->adcRuntimeDatabase.netMode = TRUE;
  }
  ADC_INT_IFACE_UNLOCK();
  return ADC_OP_STATUS_SUCCESS;
}


/**
 * Perform a user zero calibration on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcZeroCalibration(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CAL_ZERO, ADC_OP_TIMEOUT_NEVER, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a user span calibration on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param calRdg is the reading value for the span point
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcSpanCalibration(AdcSuperStructure *ass, bool lock, bool block, int32s calRdg)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CAL_SPAN, ADC_OP_TIMEOUT_NEVER, block, calRdg, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a user direct (milli-volt/volt) zero calibration on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param mVV is the millivolt/volt reading of the zero point
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDirectZeroCalibration(AdcSuperStructure *ass, bool lock, bool block, int32s mVV)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CAL_ZERO_DIRECT, ADC_OP_TIMEOUT_NEVER, block, mVV, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a user direct (milli-volt/volt) span calibration on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param calRdg is the reading value for the span point
 * @param mVV is the millivolt/volt reading of the span point
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDirectSpanCalibration(AdcSuperStructure *ass, bool lock, bool block, int32s calRdg, int32s mVV)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CAL_SPAN_DIRECT, ADC_OP_TIMEOUT_NEVER, block, calRdg, mVV, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a factory zero calibration on the scale. This function is not implemented as part of the library
 * as it performed in the factory calibration code, and should never be performed in normal operation.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcFactoryZero(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_FACTORY_ZERO, ADC_OP_TIMEOUT_NEVER, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a factory span calibration on the scale. This function is not implemented as part of the library
 * as it performed in the factory calibration code, and should never be performed in normal operation.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcFactorySpan(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_FACTORY_SPAN, ADC_OP_TIMEOUT_NEVER, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform a user preset tare on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param pt is the preset tare value
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcSetPTare(AdcSuperStructure *ass, bool lock, bool block, bool test, int32s pt)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  if (test)
  {
    result = AdcIntPresetTareOperationIsValid(ass, pt);
  }
  else
  {
    result = AdcIntOperationStart(ass, ADC_OP_CMD_SET_PRESET_TARE, ADC_OP_TIMEOUT_NEVER, block, pt, 0, false);
  }
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Add a linearisation point at point pnt, with a desired output at this point of calRdg
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param pnt is the point at which the linearisation should be applied
 * @param calRdg is the desired reading at point pnt
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcAddLinPnt(AdcSuperStructure *ass, bool lock, bool block, int32s pnt, int32s calRdg)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CAL_LIN_ADD, ADC_OP_TIMEOUT_NEVER, block, pnt, calRdg, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Remove the linearisation point at pnt
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param pnt is the linearisation point to remove
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcClearLinPnt(AdcSuperStructure *ass, bool lock, bool block, int32s pnt)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CAL_LIN_CLEAR, ADC_OP_TIMEOUT_NEVER, block, pnt, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Clear all linearisation points.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcClearAllLinPnt(AdcSuperStructure *ass, bool lock, bool block)
{
  ADC_INT_IFACE_LOCK()
  AdcIntClearAllDbLinear(ass);
  ADC_INT_IFACE_UNLOCK();
  return ADC_OP_STATUS_SUCCESS;
}


/**
 * Check for no motion on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcWaitForMotion(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_WAIT_FOR_MOTION, 10, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Get a traceable reading from the ADC library
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present
 *         ADC_OP_STATUS_FAILED_PRNID_MAX if print id beyond maximum,
 *         ADC_OP_STATUS_FAILED_HI_RES if NTEP && running in hi-res,
 *         ADC_OP_STATUS_FAILED_BAND if overloaded or underloaded, or trade scale and within the zero band,
 *         ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_SUCCESS if operation succeeds
 */
int8u AdcGetTraceableRdg(AdcSuperStructure *ass, bool lock, bool block, bool test, bool immediate)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  if (test)
  {
    result = AdcIntGetTraceableRdgOperationIsValid(ass, immediate);
  }
  else
  {
    ass->adcTraceableWeight.valid = FALSE;
    result = AdcIntOperationStart(ass, ADC_OP_CMD_TRACEABLE_RDG, 10, block, 0, 0, immediate);
  }
  ADC_INT_IFACE_UNLOCK();

  return result;
}


/**
 * Get a non-traceable reading from the ADC library
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcGetNonTraceableRdg(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  ass->adcTraceableWeight.valid = FALSE;
  result = AdcIntOperationStart(ass, ADC_OP_CMD_NON_TRACEABLE_RDG, 10, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Clear the peak hold reading on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param clearMinMax indicates that the minimum and maximum should be cleared as well
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcClearPeakRdg(AdcSuperStructure *ass, bool lock, bool block, bool clearMinMax)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CLEAR_PEAK, ADC_OP_TIMEOUT_NEVER, block, (int32s) clearMinMax, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Perform an initial zero on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcInitialZero(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_INIT_ZERO, 10, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Put the ADC library into counting mode
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param sampleSize is the number of samples
 * @param cntWeight is the weight of the samples in ADC counts rather than displayed weight (cntWeight = weight (as displayed) * 16)
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcSetCounting(AdcSuperStructure *ass, bool lock, bool block, bool test, int32s sampleSize, int32s cntWeight)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  if (test)
  {
    result = AdcIntSetCountingOperationIsValid(ass, sampleSize, cntWeight);
  }
  else
  {
    result = AdcIntOperationStart(ass, ADC_OP_CMD_SET_COUNTING, ADC_OP_TIMEOUT_NEVER, block, sampleSize, cntWeight, false);
  }
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Switch the ADC library to alternate units.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param conv is the conversion factor
 * @param units are the units to display the alternate weight in
 * @param unit is the library units to set (secondary, tertiary etc)
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcSetAltUnits(AdcSuperStructure *ass, bool lock, bool block, int32s conv, int8u units, enum eAdcUnitSwitch unit)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  switch(unit)
  {
    case ADC_UNIT_SWITCH_SECONDARY:
      result = AdcIntOperationStart(ass, ADC_OP_CMD_SET_ALT_WEIGHT, ADC_OP_TIMEOUT_NEVER, block, conv, (int32s) units, false);
      break;

    case ADC_UNIT_SWITCH_TERTIARY:
      result = AdcIntOperationStart(ass, ADC_OP_CMD_SET_TERT_WEIGHT, ADC_OP_TIMEOUT_NEVER, block, conv, (int32s) units, false);
      break;

    default:
      result = ADC_OP_STATUS_FAILED_PARAM;
      break;
  }
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Set the display mode of the ADC library
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param dispMode is the desired display mode. See @ref AdcData::dispMode.
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcSetDisplayMode(AdcSuperStructure *ass, bool lock, bool block, int8u dispMode)
{
  ADC_INT_IFACE_LOCK();
  ass->adcRuntimeDatabase.dispMode = dispMode;
  ADC_INT_IFACE_UNLOCK();
  return ADC_OP_STATUS_SUCCESS;
}


/**
 * Put the ADC library into live stock weighing mode
 *
 * @param ass is the ADC data super structure
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
static int8u AdcIntLivestockWeigh(AdcSuperStructure *ass, bool block, bool test)
{
  int8u result;

  if (test)
  {
    result = AdcIntLivestockWeighOperationIsValid(ass);
  }
  else
  {
    ass->adcHeldData.dataValid = false;
    result = AdcIntOperationStart(ass, ADC_OP_CMD_DO_LIVE, ADC_OP_TIMEOUT_NEVER, block, 0, 0, false);
  }
  return result;
}


/**
 * Put the ADC library into live stock weighing mode
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcLivestockWeigh(AdcSuperStructure *ass, bool lock, bool block, bool test)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntLivestockWeigh(ass, block, test);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Do a complete default calibration (with all settings reverted to factory defaults).
 *
 * Does not re-initialise the ADC library, so it's up to you to call @ref AdcReInit for the changes to take effect.
 * Does not save the settings, so it's up to you to do this
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDoDefaultCal(AdcSuperStructure *ass, bool lock, bool block)
{
  ADC_INT_IFACE_LOCK();

  AdcIntDefaultDatabases(ass, NULL, &ass->adcUserDatabase, &ass->adcRuntimeDatabase, true, false);
  ass->adcUserDatabase.calFlag = ADC_CAL_FLAG_OK;

  ADC_INT_IFACE_UNLOCK();
  return ADC_OP_STATUS_SUCCESS;
}


/**
 * Do a default calibration (but non-calibration settings remain unchanged)
 *
 * Does not re-initialise the ADC library, so it's up to you to call @ref AdcReInit for the changes to take effect.
 * Does not save the settings, so it's up to you to do this
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDoDefaultCalOnly(AdcSuperStructure *ass, bool lock, bool block)
{
  ADC_INT_IFACE_LOCK();

  AdcIntDefaultDatabases(ass, NULL, &ass->adcUserDatabase, &ass->adcRuntimeDatabase, true, true);
  ass->adcUserDatabase.calFlag = ADC_CAL_FLAG_OK;

  ADC_INT_IFACE_UNLOCK();
  return ADC_OP_STATUS_SUCCESS;
}


/**
 * Enable or disable ADC zero tracking
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param blockZT is a flag to indicate if zero tracking should be enabled or disabled
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDoBlockZeroTracking(AdcSuperStructure *ass, bool lock, bool block, bool blockZT)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_BLOCK_ZT, ADC_OP_TIMEOUT_NEVER, block, (int32s) blockZT, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * AdcDoTemperatureCal is not added to the operations queue and will return
 * immediately with ADC_OP_STATUS_IN_PROGRESS or ADC_OP_STATUS_FAILED_BUSY.
 * ADC_OP_STATUS_FAILED_BUSY will occur if a cal is already in progress and
 * cannot be interrupted.
 * There is no way to check progress but the cal will be finished within
 * 12 * ADC_SYNC_PERIOD_MS
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param forceZeroMux forces a zero mux
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDoTemperatureCal(AdcSuperStructure *ass, bool lock, bool forceZeroMux)
{
  int8u result;

  ADC_INT_IFACE_LOCK();

  if (ass->adcTempCal.stage >= ADC_TEMP_CAL_STAGE_FINISHED)
  {
    ass->adcTempCal.stage = ass->adcTempCal.adcTempCalStagePerMax;
    if (forceZeroMux)
      ass->adcTempCal.start++;
    result = ADC_OP_STATUS_IN_PROGRESS;
  }
  else
  {
    result = ADC_OP_STATUS_FAILED_BUSY;
  }

  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * AdcStartZeroCalCycle will return immediately. It starts a series of zero muxes
 * and span calibrations. This is the same as the muxes that occur on instrument
 * startup. This process takes about 31s. Valid readings are available during
 * this time but the zero will be changing slightly.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
void AdcStartZeroCalCycle(AdcSuperStructure *ass, bool lock)
{
  ADC_INT_IFACE_LOCK();
  AdcIntStartZeroCalCycle(&ass->adcTempCal);
  ADC_INT_IFACE_UNLOCK();
}


/**
 * AdcZeroCalCycleFinished returns true if the cal cycle has finished. This works
 * with startup muxes as well as those started by @ref AdcStartZeroCalCycle or
 * @ref AdcDoTemperatureCal (TRUE). Note that muxes will continue to occur while
 * there are ADC errors. In this case, AdcZeroCalCycleFinished will always be
 * false. Therefore, some sort of timeout failure should be used.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @return true if cal cycle has finished
 */
bool AdcZeroCalCycleFinished(AdcSuperStructure *ass, bool lock)
{
  bool result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntZeroCalCycleFinished(&ass->adcTempCal);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Clears the overload counter. Returns immediately.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
void AdcClearOverloadCounter(AdcSuperStructure *ass, bool lock)
{
  ADC_INT_IFACE_LOCK();
  ass->adcRuntimeDatabase.overloadCnt = 0;
  ADC_INT_IFACE_UNLOCK();
}


/**
 * Gets the overload counter. Returns immediately.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @return the overload counter value
 */
int32u AdcGetOverloadCounter(AdcSuperStructure *ass, bool lock)
{
  int32u result;

  ADC_INT_IFACE_LOCK();
  result = ass->adcRuntimeDatabase.overloadCnt;
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Bumps the calibration counters. Returns immediately without saving.
 * This allows for settings stored in the application database to affect the
 * calibration counter.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
void AdcBumpCalCounters(AdcSuperStructure *ass, bool lock)
{
  ADC_INT_IFACE_LOCK();
  ass->adcUserDatabase.calCntNTEPCalibration++;
  ass->adcUserDatabase.calCntNTEPSettings++;
  ass->adcUserDatabase.calCntOIML++;
  ADC_INT_IFACE_UNLOCK();
}


/**
 * Sets compensation values: compensatedRawValue = (rawValue + rawAdd) * rawMult
 * and returns immediately.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rawAdd: set in raw counts
 * @param rawMult: set as 1.000000 ignoring dp
 */
void AdcSetCompValues(AdcSuperStructure *ass, bool lock, int32s rawAdd, int32s rawMult)
{
  ADC_INT_IFACE_LOCK();
  ass->adcInternalData.rawAdd = rawAdd;
  ass->adcInternalData.rawMult = rawMult;
  ADC_INT_IFACE_UNLOCK();
}


/**
 * Calculate which range a number of counts would appear in
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param cnt is the counts to determine which range it is in
 * @return range for cnt
 */
int8u AdcGetRangeForCounts(AdcSuperStructure *ass, bool lock, int32s cnt)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntGetRangeForCounts(ass, cnt);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Calculate which range a final weight would appear in
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the reading to determine which range it is in
 * @param hiRes is if the reading is in hiRes
 * @return range for rdg
 */
int8u AdcGetRangeForFinalWeight(AdcSuperStructure *ass, bool lock, int32s rdg, bool hiRes)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntGetRangeForCounts(ass, AdcIntFinalToCounts(ass, rdg, hiRes, ADC_UNIT_SWITCH_PRIMARY));
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Get linearisation point at a weight
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the weight to get the linearisation point for
 * @return the linearisation point at rdg
 */
int8u AdcGetLinPt(AdcSuperStructure *ass, bool lock, int32s rdg)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntLinearisationGetLinPt(&ass->adcLinearisation, rdg);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Get the weight at a linearisation point
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param pnt is the point to return the weight for
 * @return the weight at linearisation point pnt
 */
int32s AdcGetLinWgt(AdcSuperStructure *ass, bool lock, int8u pnt)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntLinearisationGetLinWgt(&ass->adcLinearisation, pnt);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Apply a manual hold to the current weight (only in non-trade mode)
 *
 * @param ass is the ADC super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcManualHold(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_MANUAL_HOLD, 10, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Clear the manual hold on the current weight (only in non-trade mode)
 *
 * @param ass is the ADC super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcClearManualHold(AdcSuperStructure *ass, bool lock, bool block)
{
  int8u result;

  ADC_INT_IFACE_LOCK();
  result = AdcIntOperationStart(ass, ADC_OP_CMD_CLEAR_MANUAL_HOLD, 10, block, 0, 0, false);
  ADC_INT_IFACE_UNLOCK();
  return result;
}


/**
 * Wait for a complete new set of readings to pass through the average buffer
 *
 * @param ass is the ADC super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcWaitForAverage(AdcSuperStructure *ass, bool lock, bool block)
{
  int32s samples;
  int8u result;

  ADC_INT_IFACE_LOCK();
  samples = AdcIntCalculateAverageNumSamples(ass);
  result = AdcIntOperationStart(ass, ADC_OP_CMD_WAIT_FOR_AVERAGE, ADC_OP_TIMEOUT_NEVER, block, samples, 0, false);
  ADC_INT_IFACE_UNLOCK();

  return result;
}
