/**
 * @file adcinterface.h
 * @brief Interface functions for the R400 ADC Library.
 *
 * Company: Rinstrum Australia
 * Copyright: This code is the property of Rinstrum.  Unauthorised copying or use is
 * an offence under international and national law.
 *
 * Overview:
 * This is the interface header for the ADC library written for the R4xx.
 * The ADC library talks to the ADC chip directly and performs all processing
 * to obtain OIML approved weight values.
 *
 * Everything required to interface to the ADC library can be found in this
 * header file together with adcDataTypes.h.  This header file also contains
 * information on how to use the library.  You need to include this file plus
 * @ref adcDataTypes.h to use this library.
 *
 * This header file is broken into the following sections:
 * - version information
 * - database export
 * - procedures which need to be called to ensure the correct running of the ADC library
 * - type, constant and variable information made available to other code
 * - procedures which can be called by other code to modify the behavior of
 *   the ADC library.
 */
#ifndef ADC_INTERFACE_DOT_H
#define ADC_INTERFACE_DOT_H 1

#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcdatatypes.h>


///////////////////////////////////////////////////////////////////////////////////////
// Adc library defines
///////////////////////////////////////////////////////////////////////////////////////


/** Lower limit of ADC data rate in 0.1Hz */
#define ADC_SYNC_FREQ_LOW_LIM  (10 * 10)

/** Upper limit of ADC data rate in 0.1Hz */
#define ADC_SYNC_FREQ_HIGH_LIM (250 * 10)


///////////////////////////////////////////////////////////////////////////////////////
// Procedures which need to be called to ensure the correct running of the ADC library
///////////////////////////////////////////////////////////////////////////////////////


/**
 * Lock the ADC super structure
 *
 * @param ass is the ADC data super structure
 * @return true on successfully locked
 */
bool AdcMutexLock(AdcSuperStructure *ass);


/**
 * Unlock the ADC super structure
 *
 * @param ass is the ADC data super structure
 * @return true on successfully unlocked
 */
bool AdcMutexUnlock(AdcSuperStructure *ass);


/**
 * This procedure re-initialises all aspects of the ADC operation after locking the structure.
 * It can be called at any time after AdcInit() has been called
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param afd is a pointer to a factory database to be copied to the internal one. If NULL, no change is made to the internal one
 * @param aud is a pointer to a user database to be copied to the internal one. If NULL, no change is made to the internal one
 * @param ard is a pointer to a runtime database to be copied to the internal one. If NULL, no change is made to the internal one
 */
void AdcReInit(AdcSuperStructure *ass, bool lock, const AdcFactoryDatabase *afd, const AdcUserDatabase *aud, const AdcRuntimeDatabase *ard);


/**
 * You should call AdcPrepare() before any calls to the ADC library.
 * This will setup various fixed aspects of ADC option (such as
 * mutex, ADC device, get temperature, get date/time, DB defaulter,
 * DB saver, App callback, DB calcount check and live weigh implementer)
 *
 * This allows for calls to AdcDefaultDatabases() before AdcInit()
 * is called. This may be required because the application databases
 * need to be present for the call to AdcInit(), but they may not
 * be valid, and have to be defaulted via AdcDefaultDatabases().
 *
 * Do not call more than once (per ass).
 *
 * The sequence of startup operations must be:
 *   AdcPrepare()
 *   Application load databases, and call AdcDefaultDatabases() if necessary
 *   AdcInit()
 *   Application start ADC thread to call AdcUpdate()
 *
 * @param ass is a pointer to an un-initialised Adc Super Structure. Not NULL.
 * @param adcdev is a running 24bit ADC device to get readings from. Not NULL.
 * @param adc_get_temp is a function pointer used by the ADC library to get the current temperature. Not NULL.
 * @param adc_get_datetime is a function pointer used by the ADC to get the current date and time. Not NULL.
 * @param adcDatabaseDefaultImplementerFn is a function pointer used by the ADC library to default some or all of the databases. Not NULL.
 * @param adcDatabaseSaveImplementerFn is a function pointer used by the ADC library to save some or all of the databases. Can be NULL in which case databases won't be saved by the library.
 * @param adcAppCallbackImplementerFn is a function pointer used for calling back to your application when specific events occur. Can be NULL in which case it won't be called
 * @param adcDatabaseCalCountCheckImplementerFn is a function pointer used for calling back to your application for checking for ADC user database changes that affect calibration and configuration counters. Not NULL
 * @param mutex_implementation is the implementation of the exclusive access mutex. Can be NULL if no locking is required.
 * @param liveweigh_implementation is the implementation of the live weigh algorithm. Can be NULL if not required.
 * @param adcDebugVPrintfImplementerFn is a function pointer used by the ADC library for calling back to your application for debug printfs. Can be NULL in which case no debug printfs will be made.
 *
 * @return true if initialised correctly
 */
bool AdcPrepare(AdcSuperStructure *ass, struct t_io_dev *adcdev,
                t_adc_get_temperature_fn adc_get_temp,
                t_adc_get_datetime_fn adc_get_datetime,
                AdcDatabaseDefaultImplementerFn adcDatabaseDefaultImplementerFn,
                AdcDatabaseSaveImplementerFn adcDatabaseSaveImplementerFn,
                AdcAppCallbackImplementerFn adcAppCallbackImplementerFn,
                AdcDatabaseCalCountCheckImplementerFn adcDatabaseCalCountCheckImplementerFn,
                AdcMutexImplementer *mutex_implementation,
                AdcLiveWeighImplementer *liveweigh_implementation,
                AdcDebugVPrintfImplementerFn adcDebugVPrintfImplementerFn
               );


/**
 * This procedure initialises all aspects of the ADC operation.
 * It needs to be called after AdcPrepare(), but before any other adc procedure or data is used.
 * Do not call more than once (per ass).
 *
 * Postconditions: Adc library is ready for operation
 *
 * @param ass is a pointer to an un-initialised Adc Super Structure. Not NULL.
 * @param afd is the factory database to be copied into internal data. Can be NULL, in which case internal factory database is defaulted.
 * @param aud is the user database to be copied into internal data. Can be NULL, in which case internal user database is defaulted.
 * @param ard is the runtime database to be copied to into internal one. Can be NULL, in which case internal runtime database is defaulted.
 *
 * @return true if initialised correctly
 */
bool AdcInit(AdcSuperStructure *ass,
             const AdcFactoryDatabase *afd,
             const AdcUserDatabase *aud,
             const AdcRuntimeDatabase *ard
            );


/** Clean up after ADC library */
void AdcDestroy(AdcSuperStructure *ass);


/**
 * This procedure must be called in it's own thread. It will rate limit to
 * the sync frequency.
 *
 * It obtains a new adc value, performs all calculations to produce trade
 * weights with status information, and processes user operations such as
 * semi-automatic zero.
 *
 * Preconditions: InitAdcOperation has been called
 * Postconditions: If returns true, the weight data structure is updated
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param data_wait_failed is a pointer to a bool that is set if there was a read problem,
 *        and you should implement a delay elsewhere, to prevent the ADC thread going to 100% CPU, not NULL
 * @return true if new weight data is available, else false
 */
bool AdcUpdate(AdcSuperStructure *ass, bool lock, int32u appErrs, bool *data_wait_failed);


/**
 * Saving of ADC databases with locking
 *
 * @param ass is the ADC super structure.
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param afd is a pointer to a factory database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param aud is a pointer to a user database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param ard is a pointer to a runtime database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param aud_reference is a pointer to a user database used as a reference to determine whether to increment the calibration counters. Can be NULL only if aud is NULL
 *
 * @return result of the save
 */
bool AdcSaveDatabases(AdcSuperStructure *ass,
                      bool lock,
                      AdcFactoryDatabase *afd,
                      AdcUserDatabase *aud,
                      AdcRuntimeDatabase *ard,
                      const AdcUserDatabase *aud_reference);


///////////////////////////////////////////////////////////////////////////////////////
// Adc Weight Conversions
///////////////////////////////////////////////////////////////////////////////////////


/**
 * Convert internal ADC counts into final (displayable) numbers,
 * taking into account the resolution of the specified range,
 * and whether an extra digit of resolution is required.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the reading in ADC counts
 * @param range is the current range (0, 1, 2)
 * @param hiRes is a flag to indicate that the final number should have an extra digit of resolution
 * @return the rdg converted to final (displayable) numbers
 */
int32s AdcCountsToFinal(AdcSuperStructure *ass, bool lock, int32s rdg, int8u range, bool hiRes);


/**
 * Convert final (displayable) numbers into internal ADC counts,
 * taking into account whether an extra digit of resolution is present.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the final (displayable) number
 * @param hiRes is a flag to indicate that rdg has an extra digit of resolution
 * @return the rdg converted to internal ADC counts
 */
int32s AdcFinalToCounts(AdcSuperStructure *ass, bool lock, int32s rdg, bool hiRes);


/**
 * Convert a final (displayable) number with (possibly) extra resolution
 * into a final (displayable) number with no extra resolution
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the final number with (possibly) extra resolution
 * @param range is the current range of the final number
 * @param hiRes is a flag to indicate that rdg has extra resolution
 * @return rdg with any extra resolution removed
 */
int32s AdcHiResToNormal(AdcSuperStructure *ass, bool lock, int32s rdg, int8u range, bool hiRes);


/**
 * Convert internal ADC counts into final alternative units (displayable) numbers,
 * taking into account the resolution of the specified range,
 * and whether an extra digit of resolution is required.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the reading in ADC counts
 * @param range is the current range (0, 1, 2)
 * @param hiRes is a flag to indicate that the final number should have an extra digit of resolution
 * @return the rdg converted to final alternative units (displayable) numbers
 */
int32s AdcCountsToFinalAlt(AdcSuperStructure *ass, bool lock, int32s rdg, int8u range, bool hiRes);


/**
 * Convert final alternative units (displayable) numbers into internal ADC counts,
 * taking into account whether an extra digit of resolution is present.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the final alternative units (displayable) number
 * @param hiRes is a flag to indicate that rdg has an extra digit of resolution
 * @return the rdg converted to internal ADC counts
 */
int32s AdcFinalToCountsAlt(AdcSuperStructure *ass, bool lock, int32s rdg, bool hiRes);


/**
 * Convert internal ADC counts into final tertiary units (displayable) numbers,
 * taking into account the resolution of the specified range,
 * and whether an extra digit of resolution is required.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the reading in ADC counts
 * @param range is the current range (0, 1, 2)
 * @param hiRes is a flag to indicate that the final number should have an extra digit of resolution
 * @return the rdg converted to final tertiary units (displayable) numbers
 */
int32s AdcCountsToFinalTert(AdcSuperStructure *ass, bool lock, int32s rdg, int8u range, bool hiRes);


/**
 * Convert final tertiary units (displayable) numbers into internal ADC counts,
 * taking into account whether an extra digit of resolution is present.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the final tertiary units (displayable) number
 * @param hiRes is a flag to indicate that rdg has an extra digit of resolution
 * @return the rdg converted to internal ADC counts
 */
int32s AdcFinalToCountsTert(AdcSuperStructure *ass, bool lock, int32s rdg, bool hiRes);


/**
 * Perform a multiply followed by and divide on base
 *
 * @param base is the number to be multiplied and then divided
 * @param mult is the multiplier
 * @param div is the divisor
 * @return is the the result of base * mult / div
 */
int32s AdcMultDiv(int32s base, int32s mult, int32s div);


///////////////////////////////////////////////////////////////////////////////////////
// Adc Operations
///////////////////////////////////////////////////////////////////////////////////////
// More precise control can be accessed via "adcOperations.h".


/**
 * Cancel the current ADC operation
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
void AdcOperationCancel(AdcSuperStructure *ass, bool lock);


/**
 * Returns the current command identifier (a number that is incremented each time a command is started)
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
int8u AdcGetCurrentCommandCnt(AdcSuperStructure *ass, bool lock);


/**
 * Returns an identifier of the current command in progress (one of the ADC_OP_CMD_* values).
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
int8u AdcGetCurrentCommand(AdcSuperStructure *ass, bool lock);


#define ADC_OP_CMD_NONE               0         /*!< No command is currently in progress */
#define ADC_OP_CMD_SEMIAUTO_ZERO      1         /*!< The semi-automatic user zero command */
#define ADC_OP_CMD_SEMIAUTO_TARE      2         /*!< The semi-automatic user tare command */
#define ADC_OP_CMD_SET_PRESET_TARE    3         /*!< The user pre-set tare command */
#define ADC_OP_CMD_CAL_ZERO           4         /*!< The user zero calibration command */
#define ADC_OP_CMD_CAL_SPAN           5         /*!< The user span calibration command */
#define ADC_OP_CMD_SET_COUNTING       6         /*!< The set counting mode command */
#define ADC_OP_CMD_GROSSNET_TOGGLE    7         /*!< The gross/net toggle command */
#define ADC_OP_CMD_GROSSNET_GROSS     8         /*!< Set scale to gross command */
#define ADC_OP_CMD_GROSSNET_NET       9         /*!< Set scale to net command */
#define ADC_OP_CMD_CAL_ZERO_DIRECT    10        /*!< The user direct (milli-volt/volt) zero calibration command */
#define ADC_OP_CMD_CAL_SPAN_DIRECT    11        /*!< The user direct (milli-volt/volt) span calibration command */
#define ADC_OP_CMD_FACTORY_ZERO       12        /*!< The factory zero calibration command. Not implemented as it is performed by the factory calibration code. */
#define ADC_OP_CMD_FACTORY_SPAN       13        /*!< The factory span calibration command. Not implemented as it is performed by the factory calibration code. */
#define ADC_OP_CMD_CAL_LIN_ADD        14        /*!< The add linearisation point command */
#define ADC_OP_CMD_CAL_LIN_CLEAR      15        /*!< The clear linearisation point command */
#define ADC_OP_CMD_WAIT_FOR_MOTION    16        /*!< The wait/check for no motion command */
#define ADC_OP_CMD_TRACEABLE_RDG      17        /*!< Get a traceable reading. @ref adcTraceableWeight is filled out if the command is successful. */
#define ADC_OP_CMD_CLEAR_PEAK         18        /*!< The clear peak hold command */
#define ADC_OP_CMD_INIT_ZERO          19        /*!< The perform initial zero command */
#define ADC_OP_CMD_DO_LIVE            20        /*!< Command to switch to live weighing */
#define ADC_OP_CMD_SET_DISPLAY_MODE   21        /*!< Command to set the display mode */
#define ADC_OP_CMD_SET_ALT_WEIGHT     22        /*!< Command to set the alternate units and conversion factor */
#define ADC_OP_CMD_BLOCK_ZT           23        /*!< Command to enable or disable zero tracking */
#define ADC_OP_CMD_NON_TRACEABLE_RDG  24        /*!< Get a non-traceable reading, this is used to get a unique print ID only */
#define ADC_OP_CMD_CLEAR_SEMIAUTO_ZERO 25       /*!< Clear semi-auto zero (only in non trade mode) */
#define ADC_OP_CMD_MANUAL_HOLD        26        /*!< Hold the current weight (only in non trade mode) */
#define ADC_OP_CMD_CLEAR_MANUAL_HOLD  27        /*!< Clear the hold on the current weight (only in non trade mode) */
#define ADC_OP_CMD_WAIT_FOR_AVERAGE   28        /*!< Wait for a complete new set of readings to pass through the average buffer */
#define ADC_OP_CMD_SET_TERT_WEIGHT    29        /*!< Command to set the tertiary units and conversion factor */


/*
 * All procedures below, return 1 of the codes in the ADC_OP_STATUS series.  Most
 * are self explanatory.  AdcGetCurrentOpStatus returns the status of the current
 * or previous operation.  When the parameter "block" is true, the procedure will
 * only return when the operation has been completed.  When "block" is false, the
 * procedure returns immediately with a ADC_OP_STATUS_FAILED_BUSY,
 * ADC_OP_STATUS_FAILED_NOT_IMP'lemented or ADC_OP_STATUS_IN_PROGRESS result.
 */


/**
 * Return current operation status. This is one of the ADC_OP_STATUS_* codes.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
int8u AdcGetCurrentOpStatus(AdcSuperStructure *ass, bool lock);


#define ADC_OP_STATUS_SUCCESS         0         /*!< The ADC operation was successful */
#define ADC_OP_STATUS_CANCELED        1         /*!< The ADC operation was canceled */
#define ADC_OP_STATUS_IN_PROGRESS     2         /*!< The ADC operation is currently still in progress */
#define ADC_OP_STATUS_FAILED_ERROR    3         /*!< The ADC operation failed because of an error */
#define ADC_OP_STATUS_FAILED_OL_UL    4         /*!< The ADC operation failed because the scale was overloaded or underloaded */
#define ADC_OP_STATUS_FAILED_BUSY     5         /*!< The ADC operation failed because the library was busy */
#define ADC_OP_STATUS_FAILED_MOTION   6         /*!< The ADC operation failed because of motion on the scale */
#define ADC_OP_STATUS_FAILED_BAND     7         /*!< The ADC operation failed because the scale is outside the zeroband or tareband, etc */
#define ADC_OP_STATUS_FAILED_RES      8         /*!< The ADC operation failed because of not enough resolution */
#define ADC_OP_STATUS_FAILED_NOT_IMP  9         /*!< The ADC operation failed because the command is not implemented */
#define ADC_OP_STATUS_FAILED_PNT_DUPL 10        /*!< The ADC operation failed because the linearisation point already exists */
#define ADC_OP_STATUS_FAILED_HI_RES   11        /*!< The ADC operation failed because the scale is in high resolution mode */
#define ADC_OP_STATUS_FAILED_PRNID_MAX 12       /*!< The ADC operation failed because the print id has reached it's maximum */
#define ADC_OP_STATUS_FAILED_DATE_TIME 13       /*!< The ADC operation failed because the current date/time could not be determined */
#define ADC_OP_STATUS_FAILED_PARAM     14       /*!< The ADC operation failed because a parameter passed was invalid */
#define ADC_OP_STATUS_FAILED_TRADE     15       /*!< The ADC operation failed because this operation is not valid in trade mode */
#define ADC_OP_STATUS_NUM              16


/**
 * Perform a semi-automatic zero
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_BAND if outside zero range, ADC_OP_STATUS_SUCCESS if zero succeeds
 */
int8u AdcSemiautoZero(AdcSuperStructure *ass, bool lock, bool block, bool test, bool immediate);


/**
 * Clear a previous semi-automatic zero (only in non-trade mode)
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcClearSemiautoZero(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Perform a semi-automatic tare
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_BAND if outside range of tare, ADC_OP_STATUS_SUCCESS if tare succeeds
 */
int8u AdcSemiautoTare(AdcSuperStructure *ass, bool lock, bool block, bool test, bool immediate);


/**
 * Perform a Gross/Net toggle
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcGrossNetToggle(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Switch the scale to gross
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcToGross(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Switch the scale to net
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcToNet(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Perform a user zero calibration on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcZeroCalibration(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Perform a user span calibration on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param calRdg is the reading value for the span point
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcSpanCalibration(AdcSuperStructure *ass, bool lock, bool block, int32s calRdg);


/**
 * Perform a user direct (milli-volt/volt) zero calibration on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param mVV is the millivolt/volt reading of the zero point
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDirectZeroCalibration(AdcSuperStructure *ass, bool lock, bool block, int32s mVV);


/**
 * Perform a user direct (milli-volt/volt) span calibration on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param calRdg is the reading value for the span point
 * @param mVV is the millivolt/volt reading of the span point
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDirectSpanCalibration(AdcSuperStructure *ass, bool lock, bool block, int32s calRdg, int32s mVV);


/**
 * Perform a factory zero calibration on the scale. This function is not implemented as part of the library
 * as it performed in the factory calibration code, and should never be performed in normal operation.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcFactoryZero(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Perform a factory span calibration on the scale. This function is not implemented as part of the library
 * as it performed in the factory calibration code, and should never be performed in normal operation.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcFactorySpan(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Perform a user preset tare on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param pt is the preset tare value
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcSetPTare(AdcSuperStructure *ass, bool lock, bool block, bool test, int32s pt);


/**
 * Add a linearisation point at point pnt, with a desired output at this point of calRdg
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param pnt is the point at which the linearisation should be applied
 * @param calRdg is the desired reading at point pnt
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcAddLinPnt(AdcSuperStructure *ass, bool lock, bool block, int32s pnt, int32s calRdg);


/**
 * Remove the linearisation point at pnt
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param pnt is the linearisation point to remove
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcClearLinPnt(AdcSuperStructure *ass, bool lock, bool block, int32s pnt);


/**
 * Clear all linearisation points.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcClearAllLinPnt(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Check for no motion on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcWaitForMotion(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Get a traceable reading from the ADC library
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present
 *         ADC_OP_STATUS_FAILED_PRNID_MAX if print id beyond maximum,
 *         ADC_OP_STATUS_FAILED_HI_RES if NTEP && running in hi-res,
 *         ADC_OP_STATUS_FAILED_BAND if overloaded or underloaded, or trade scale and within the zero band,
 *         ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_SUCCESS if operation succeeds
 */
int8u AdcGetTraceableRdg(AdcSuperStructure *ass, bool lock, bool block, bool test, bool immediate);


/**
 * Get a non-traceable reading from the ADC library
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcGetNonTraceableRdg(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Clear the peak hold reading on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param clearMinMax indicates that the minimum and maximum should be cleared as well
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcClearPeakRdg(AdcSuperStructure *ass, bool lock, bool block, bool clearMinMax);


/**
 * Perform an initial zero on the scale
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcInitialZero(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Put the ADC library into counting mode
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 * @param sampleSize is the number of samples
 * @param cntWeight is the weight of the samples in ADC counts rather than displayed weight (cntWeight = weight (as displayed) * 16)
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcSetCounting(AdcSuperStructure *ass, bool lock, bool block, bool test, int32s sampleSize, int32s cntWeight);


/**
 * Switch the ADC library to alternate units.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param conv is the conversion factor
 * @param units are the units to display the alternate weight in
 * @param unit is the library units to set (secondary, tertiary etc)
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcSetAltUnits(AdcSuperStructure *ass, bool lock, bool block, int32s conv, int8u units, enum eAdcUnitSwitch unit);


/**
 * Set the display mode of the ADC library
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param dispMode is the desired display mode. See @ref AdcData::dispMode.
 * @return one of the ADC_OP_STATUS_* codes.
 */
int8u AdcSetDisplayMode(AdcSuperStructure *ass, bool lock, bool block, int8u dispMode);


/**
 * Put the ADC library into live stock weighing mode
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param test if set, will check if the operation would succeed, but not execute it, otherwise, the operation is performed
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcLivestockWeigh(AdcSuperStructure *ass, bool lock, bool block, bool test);


/**
 * Do a complete default calibration (with all settings reverted to factory defaults).
 *
 * Does not re-initialise the ADC library, so it's up to you to call @ref AdcReInit for the changes to take effect.
 * Does not save the settings, so it's up to you to do this
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDoDefaultCal(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Do a default calibration (but non-calibration settings remain unchanged)
 *
 * Does not re-initialise the ADC library, so it's up to you to call @ref AdcReInit for the changes to take effect.
 * Does not save the settings, so it's up to you to do this
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDoDefaultCalOnly(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Defaulting of databases with locking
 *
 * @param ass is the ADC super structure.
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param afd is a pointer to a factory database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param aud is a pointer to a user database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param ard is a pointer to a runtime database to be defaulted. Can be NULL in which case shouldn't be defaulted.
 * @param preserve_cal_counters means the calibration counters should be preserved and incremented over the default.
 * @param cal_only means only the calibration settings should be defaulted (this applies to aud only) (settings to be defaulted are zeroRaw, spanRaw, spanCnt, linearisation).
 */
void AdcDefaultDatabases(AdcSuperStructure *ass,
                         bool lock,
                         AdcFactoryDatabase *afd,
                         AdcUserDatabase *aud,
                         AdcRuntimeDatabase *ard,
                         bool preserve_cal_counters,
                         bool cal_only);


/**
 * Enable or disable ADC zero tracking
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @param blockZT is a flag to indicate if zero tracking should be enabled or disabled
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDoBlockZeroTracking(AdcSuperStructure *ass, bool lock, bool block, bool blockZT);


/**
 * AdcDoTemperatureCal is not added to the operations queue and will return
 * immediately with ADC_OP_STATUS_IN_PROGRESS or ADC_OP_STATUS_FAILED_BUSY.
 * ADC_OP_STATUS_FAILED_BUSY will occur if a cal is already in progress and
 * cannot be interrupted.
 * There is no way to check progress but the cal will be finished within
 * 12 * ADC_SYNC_PERIOD_MS
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param forceZeroMux forces a zero mux
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcDoTemperatureCal(AdcSuperStructure *ass, bool lock, bool forceZeroMux);


/**
 * AdcStartZeroCalCycle will return immediately. It starts a series of zero muxes
 * and span calibrations. This is the same as the muxes that occur on instrument
 * startup. This process takes about 31s. Valid readings are available during
 * this time but the zero will be changing slightly.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
void AdcStartZeroCalCycle(AdcSuperStructure *ass, bool lock);


/**
 * AdcZeroCalCycleFinished returns true if the cal cycle has finished. This works
 * with startup muxes as well as those started by @ref AdcStartZeroCalCycle or
 * @ref AdcDoTemperatureCal (true). Note that muxes will continue to occur while
 * there are ADC errors. In this case, AdcZeroCalCycleFinished will always be
 * false. Therefore, some sort of timeout failure should be used.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @return true if cal cycle has finished
 */
bool AdcZeroCalCycleFinished(AdcSuperStructure *ass, bool lock);


/**
 * Clears the overload counter. Returns immediately.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
void AdcClearOverloadCounter(AdcSuperStructure *ass, bool lock);


/**
 * Gets the overload counter. Returns immediately.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @return the overload counter value
 */
int32u AdcGetOverloadCounter(AdcSuperStructure *ass, bool lock);


/**
 * Bumps the calibration counters. Returns immediately without saving.
 * This allows for settings stored in the application database to affect the
 * calibration counter.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
void AdcBumpCalCounters(AdcSuperStructure *ass, bool lock);


/**
 * Sets compensation values: compensatedRawValue = (rawValue + rawAdd) * rawMult
 * and returns immediately.
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rawAdd: set in raw counts
 * @param rawMult: set as 1.000000 ignoring dp
 */
void AdcSetCompValues(AdcSuperStructure *ass, bool lock, int32s rawAdd, int32s rawMult);


/**
 * Calculate which range a number of counts would appear in
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param cnt is the counts to determine which range it is in
 * @return range for cnt
 */
int8u AdcGetRangeForCounts(AdcSuperStructure *ass, bool lock, int32s cnt);


/**
 * Calculate which range a final weight would appear in
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the reading to determine which range it is in
 * @param hiRes is if the reading is in hiRes
 * @return range for rdg
 */
int8u AdcGetRangeForFinalWeight(AdcSuperStructure *ass, bool lock, int32s rdg, bool hiRes);


/**
 * Get linearisation point at a weight
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param rdg is the weight to get the linearisation point for
 * @return the linearisation point at rdg
 */
int8u AdcGetLinPt(AdcSuperStructure *ass, bool lock, int32s rdg);


/**
 * Get the weight at a linearisation point
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param pnt is the point to return the weight for
 * @return the weight at linearisation point pnt
 */
int32s AdcGetLinWgt(AdcSuperStructure *ass, bool lock, int8u pnt);


/**
 * Apply a manual hold to the current weight (only in non-trade mode)
 *
 * @param ass is the ADC super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcManualHold(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Clear the manual hold on the current weight (only in non-trade mode)
 *
 * @param ass is the ADC super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcClearManualHold(AdcSuperStructure *ass, bool lock, bool block);


/**
 * Wait for a complete new set of readings to pass through the average buffer
 *
 * @param ass is the ADC super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 * @param block is a flag to indicate if the function should block until the operation is complete
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcWaitForAverage(AdcSuperStructure *ass, bool lock, bool block);


#endif /* #ifndef ADC_INTERFACE_DOT_H */
