/*
 * adcliveimpl_improved.c
 *
 * Improved live weigh algorithm, stolen from R300 & R400-A413
 *
 *  Created on: Jun 25, 2012
 *      Author: ian
 */
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdebug.h>
#include <rinweigh/adcconvert.h>

#include <rinweigh/adcliveimpl_improved.h>


/** Clear the livestock averaging variables */
void LiveStockClear(AdcLiveImproved *ali)
{
  ali->liveRestartOk = false;
  ali->livePass = INT32S_MAX;
}


/** Clear the primary averaging */
static void AdcIntLiveImprovedPrimaryAvClear(AdcLiveImproved *ali)
{
  ali->livePrimaryAvMax = INT32S_MIN;
  ali->livePrimaryAvMin = INT32S_MAX;
  ali->livePrimaryAvTotal = 0;
  ali->livePrimaryAvCurrentRdg = 0;
}


/** Run the wait for target weight to be reached state */
static enum eAdcLiveWeighResult AdcIntLiveImproved_RunWaitTarget(AdcLiveImproved *ali, int32s rdg)
{
  if (!ali->configured)
  {
    ali->liveStockState = LIVE_STATE_OFF;
    return alwrDisabled;
  }

#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "AdcIntLiveImproved_RunWaitTarget: rdg: %ld, target: %ld\n", (long) rdg, (long) ali->target);
#endif

  /* Wait for weight to reach target before doing hold off delay */
  if (rdg > ali->target)
  {
    ali->liveStockState = LIVE_STATE_HOLDOFF;
    ali->holdOffSampleCounter = ali->holdOffMillisecondsPreset * ali->sampleFrequency / 10 / 1000;
#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "AdcIntLiveImproved_RunWaitTarget: Target reached, starting holdoff for %lu samples\n", (unsigned long) ali->holdOffSampleCounter);
#endif
  }
  return alwrRunning;
}


/** Run the hold off state */
static enum eAdcLiveWeighResult AdcIntLiveImproved_RunHoldoff(AdcLiveImproved *ali, int32s rdg)
{
  /* Bail if weight drops below target */
  if (rdg <= ali->target)
  {
    ali->liveStockState = LIVE_STATE_WAIT_TARGET;
#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "AdcIntLiveImproved_RunHoldoff: Weight dropped below target\n");
#endif
    return alwrRunning;
  }

  /* If holdoff has expired then move to collect state */
  if (ali->holdOffSampleCounter > 0)
  {
    ali->holdOffSampleCounter--;
  }
  else
  {
    LiveStockClear(ali);
    ali->liveStockState = LIVE_STATE_COLLECT;
    AdcIntLiveImprovedPrimaryAvClear(ali);
    ali->secondaryAvCurrentRdg = 0;
#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "AdcIntLiveImproved_RunWaitTarget: Hold off complete, starting data collection\n");
#endif
  }
  return alwrRunning;
}


/** sampleTol equals number of largest samples to remove */
static int32s LiveProcess(AdcLiveImproved *ali, size_t sampleTol)
{
  int32s largest;
  int64_t total;
  int64_t average;
  size_t av_ctr;
  size_t av_posn;
  size_t av_cur;
  size_t av_discard;

  /* Remove sampleTol readings biggest deltas */
  for (av_discard = 0; av_discard < sampleTol; av_discard++)
  {
    /* Find largest average with flag == 0 */
    for (largest = 0, av_posn = 0, av_cur = 0; av_cur < ali->secondaryAvNumRdgs; av_cur++)
    {
      if (0 == ali->secondaryAverages[av_cur].flag && ali->secondaryAverages[av_cur].delta >= largest)
      {
        largest = ali->secondaryAverages[av_cur].delta;
        av_posn = av_cur;
      }
    }

    /* Mark largest average as bad */
    ali->secondaryAverages[av_posn].flag = 1;
  }

  /* Average those that are left */
  for (total = 0, av_ctr = 0, av_cur = 0; av_cur < ali->secondaryAvNumRdgs; av_cur++)
  {
    if (0 == ali->secondaryAverages[av_cur].flag)
    {
      total += ali->secondaryAverages[av_cur].average;
      av_ctr++;
    }
  }

  if (0 != total && 0 != av_ctr)
  {
    /* Work out average */
    average = total / av_ctr;
  }
  else
  {
    average = 0;
  }

#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "LiveProcess: Average %lu of %lu secondary readings, total %" PRIu64 ", average %" PRIu64"\n",
                   (unsigned long) av_ctr, (unsigned long) ali->secondaryAvNumRdgs,
                   total, average);
#endif

  return (int32s) average;
}


static void LiveStockCalc(AdcLiveImproved *ali, int32s *result_rdg)
{
  size_t av_ctr;
  int32s average;

  /* Work out the average, after discarding ali->tolPreset farthest away from the average */
  average = LiveProcess(ali, ali->secondaryAvNumDiscard);

  /* Re-calculate the deltas (difference between each valid reading and the average) */
  for (av_ctr = 0; av_ctr < ali->secondaryAvNumRdgs; av_ctr++)
  {
    if (0 == ali->secondaryAverages[av_ctr].flag)
    {
      if (ali->secondaryAverages[av_ctr].average >= average)
      {
        ali->secondaryAverages[av_ctr].delta = ali->secondaryAverages[av_ctr].average - average;
      }
      else
      {
        ali->secondaryAverages[av_ctr].delta = average - ali->secondaryAverages[av_ctr].average;
      }
    }
  }

  /* And work out the average again, after discarding ali->tolPreset farthest away from the average */
  average = LiveProcess(ali, ali->secondaryAvNumDiscard);

  /* If enabled, allow re-triggering when the weight changes by ali->retrigPercentPreset */
  if (0 != ali->retrigPercentPreset)
  {
    /* Re-trigger */
    ali->livePass = AdcMultDiv(average, (int32s) ali->retrigPercentPreset, 100L);
    ali->liveRestartOk = true;

    ali->maxAverage = average + ali->livePass;
    ali->minAverage = average - ali->livePass;
  }

  /* Return result reading */
  *result_rdg = average;

  /* Restart if configured to */
  if (ali->liveRestartOk)
  {
    ali->liveStockState = LIVE_STATE_COLLECT;
  }
  else
  {
    ali->liveStockState = LIVE_STATE_COMPLETE;
  }
}


/** Run the collect data state */
static enum eAdcLiveWeighResult AdcIntLiveImproved_RunCollect(AdcLiveImproved *ali, int32s rdg, int32s *result_rdg)
{
  enum eAdcLiveWeighResult result = alwrRunning;

  /* Bail if weight drops below target */
  if (rdg <= ali->target)
  {
    ali->liveStockState = LIVE_STATE_WAIT_TARGET;
    return alwrRunning;
  }

  /* Do primary average of ADC readings */

  /* Add weight to instantaneous average and update min and max */
  ali->livePrimaryAvTotal += rdg;
  ali->livePrimaryAvCurrentRdg++;

  /* Keep max and min */
  if (rdg > ali->livePrimaryAvMax)
  {
    ali->livePrimaryAvMax = rdg;
  }
  if (rdg < ali->livePrimaryAvMin)
  {
    ali->livePrimaryAvMin = rdg;
  }

  /* Every X samples, keep a copy of the primary average */
  if (ali->livePrimaryAvCurrentRdg >= ali->primaryAvNumRdgs)
  {
    ali->secondaryAverages[ali->secondaryAvCurrentRdg].average = ali->livePrimaryAvTotal / ali->livePrimaryAvCurrentRdg;
    ali->secondaryAverages[ali->secondaryAvCurrentRdg].delta = ali->livePrimaryAvMax - ali->livePrimaryAvMin;
    ali->secondaryAverages[ali->secondaryAvCurrentRdg].flag = 0;

    if (ali->liveRestartOk)
    {
      /* Restart if delta is outside limits set by retrig tol */
      if (ali->secondaryAverages[ali->secondaryAvCurrentRdg].average > ali->maxAverage ||
          ali->secondaryAverages[ali->secondaryAvCurrentRdg].average < ali->minAverage)
      {
        ali->liveStockState = LIVE_STATE_WAIT_TARGET;
      }
    }
    else
    {
      /* On to next secondary average */
      ali->secondaryAvCurrentRdg++;
      if (ali->secondaryAvCurrentRdg >= ali->secondaryAvNumRdgs)
      {
        /* All sample windows full, trigger a weight reading calculation */
        LiveStockCalc(ali, result_rdg);
        ali->secondaryAvCurrentRdg = 0;
        result = alwrComplete;
      }
    }

    /* Initialise primary average */
    AdcIntLiveImprovedPrimaryAvClear(ali);
  }
  return result;
}


/** Run the complete state */
static enum eAdcLiveWeighResult AdcIntLiveImproved_RunComplete(AdcLiveImproved *ali, int32s rdg)
{
  /* Restart livestock weighing if weight drops below target */
  if (rdg <= ali->target)
  {
    ali->liveStockState = LIVE_STATE_WAIT_TARGET;
  }
  return alwrRunning;
}


static enum eAdcLiveWeighResult AdcIntRunLiveImproved(AdcLiveImproved *ali, int32s rdg, int32s *result_rdg)
{
  enum eAdcLiveWeighResult res;

  if (!ali->configured || NULL == ali->secondaryAverages)
  {
    return alwrFailed;
  }

#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
  if (ali->prevLiveStockState != ali->liveStockState)
  {
    ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "AdcIntRunLiveImproved: Live weigh state changed from %u to %u\n", ali->prevLiveStockState, ali->liveStockState);
    ali->prevLiveStockState = ali->liveStockState;
  }
#endif

  /* Run the livestock state machine */
  switch (ali->liveStockState)
  {
    case LIVE_STATE_WAIT_TARGET:
      res = AdcIntLiveImproved_RunWaitTarget(ali, rdg);
      break;

    case LIVE_STATE_HOLDOFF:
      res = AdcIntLiveImproved_RunHoldoff(ali, rdg);
      break;

    case LIVE_STATE_COLLECT:
      res = AdcIntLiveImproved_RunCollect(ali, rdg, result_rdg);
      break;

    case LIVE_STATE_COMPLETE:
      res = AdcIntLiveImproved_RunComplete(ali, rdg);
      break;

    case LIVE_STATE_OFF:
    default:
       res = alwrDisabled;
       break;
  }
  return res;
}


/** Initiate improved live weighing */
static enum eAdcLiveWeighResult AdcIntInitiateLiveImproved(AdcLiveImproved *ali)
{
  enum eAdcLiveWeighResult res;

  if (!ali->configured || NULL == ali->secondaryAverages)
  {
    res = alwrFailed;
  }
  else
  {
    ali->liveStockState = LIVE_STATE_WAIT_TARGET;
    LiveStockClear(ali);
    res = alwrRunning;
  }
  return res;
}


/** Disable improved live weighing */
static enum eAdcLiveWeighResult AdcIntDisableLiveImproved(AdcLiveImproved *ali)
{
  ali->liveStockState = LIVE_STATE_OFF;
  return alwrDisabled;
}


static enum eAdcLiveWeighResult AdcLiveWeighImprovedRun(AdcLiveWeighImplementer *alwi, int32s rdg, int32s *resultrdg)
{
  AdcLiveImproved *ali;

  if (NULL == alwi || NULL == alwi->implementer_data)
  {
    return alwrNotImplemented;
  }
  ali = (AdcLiveImproved *) alwi->implementer_data;

  return AdcIntRunLiveImproved(ali, rdg, resultrdg);
}


static enum eAdcLiveWeighResult AdcLiveWeighImprovedInitiate(AdcLiveWeighImplementer *alwi)
{
  AdcLiveImproved *ali;

#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "AdcLiveWeighImprovedInitiate: called\n");
#endif

  if (NULL == alwi || NULL == alwi->implementer_data)
  {
    return alwrNotImplemented;
  }
  ali = (AdcLiveImproved *) alwi->implementer_data;

  return AdcIntInitiateLiveImproved(ali);
}


static enum eAdcLiveWeighResult AdcLiveWeighImprovedDisable(AdcLiveWeighImplementer *alwi)
{
  AdcLiveImproved *ali;

#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "AdcLiveWeighImprovedDisable: called\n");
#endif

  if (NULL == alwi || NULL == alwi->implementer_data)
  {
    return alwrNotImplemented;
  }
  ali = (AdcLiveImproved *) alwi->implementer_data;

  return AdcIntDisableLiveImproved(ali);
}


static enum eAdcLiveWeighResult AdcLiveWeighImprovedRunning(AdcLiveWeighImplementer *alwi)
{
  AdcLiveImproved *ali;
  enum eAdcLiveWeighResult result;

  if (NULL == alwi || NULL == alwi->implementer_data)
  {
    return alwrNotImplemented;
  }
  ali = (AdcLiveImproved *) alwi->implementer_data;

  if (LIVE_STATE_OFF == ali->liveStockState)
  {
    /* Not running, and hasn't completed */
    result = alwrDisabled;
  }
  else if (LIVE_STATE_COMPLETE != ali->liveStockState)
  {
    /* Running, not completed */
    result = alwrRunning;
  }
  else
  {
    /* Completed */
    result = alwrComplete;
  }
  return result;
}


/** Destroy a improved live weigh implementer */
void AdcLiveWeighImprovedDestroy(AdcLiveWeighImplementer *alwi)
{
  AdcLiveImproved *ali;

  if (NULL != alwi && NULL != alwi->implementer_data)
  {
    ali = (AdcLiveImproved *) alwi->implementer_data;
    alwi->implementer_data = NULL;

    if (NULL != ali->secondaryAverages)
    {
      free(ali->secondaryAverages);
    }
    memset(ali, 0, sizeof(*ali));
    free(ali);
  }
}


/**
 * Create an improved live weigh implementer.
 *
 * Once created, the live weigh implementer will need to be configured.
 *
 * @return pointer to initialised AdcLiveWeighImplementer structure
 */
AdcLiveWeighImplementer *AdcLiveWeighImplementerImprovedCreate(void)
{
  AdcLiveWeighImplementer *alwi = NULL;
  AdcLiveImproved *ali = NULL;

#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(ADC_DEBUG_DEV, "AdcLiveWeighImplementerImprovedCreate: called\n");
#endif

  /* Make structure */
  alwi = malloc(sizeof(AdcLiveWeighImplementer));
  if (NULL == alwi)
  {
    goto alwisc_fail;
  }
  memset(alwi, 0, sizeof(*alwi));

  /* Make and initialise improved live weigh */
  ali = malloc(sizeof(AdcLiveImproved));
  if (NULL == ali)
  {
    goto alwisc_fail;
  }
  memset(ali, 0, sizeof(*ali));
  ali->configured = false;
  ali->liveStockState = LIVE_STATE_OFF;
  ali->secondaryAverages = NULL;

#if ADC_DEBUG(ADC_LIVE_IMPROVED_DEBUG_LVL, ADC_DEBUG_BASIC)
  ali->prevLiveStockState = ali->liveStockState;
#endif

  /* Fill out members of alwi */
  alwi->implementer_data = ali;
  alwi->AdcLiveWeighRun = AdcLiveWeighImprovedRun;
  alwi->AdcLiveWeighInitiate = AdcLiveWeighImprovedInitiate;
  alwi->AdcLiveWeighDisable = AdcLiveWeighImprovedDisable;
  alwi->AdcLiveWeighRunning = AdcLiveWeighImprovedRunning;
  alwi->AdcLiveWeighDestroy = AdcLiveWeighImprovedDestroy;

  /* Success */
  return alwi;

alwisc_fail:
  if (NULL != alwi)
  {
    free(alwi);
  }
  if (NULL != ali)
  {
    free(ali);
  }
  return NULL;
}


static int32s AdcIntFinalToMvv(AdcSuperStructure *ass, int32s target)
{
  int32s targetcnt;
  int32s targetmvv;

  /* Convert to counts */
  targetcnt = AdcIntFinalToCounts(ass, target, false, ADC_UNIT_SWITCH_PRIMARY);

  /* Convert to millivolts per volt */
  targetmvv = AdcMultDiv(ass->adcUserDatabase.spanRaw, targetcnt, ass->adcUserDatabase.spanCnt) + ass->adcUserDatabase.zeroRaw;

  return targetmvv;
}


/**
 * Configure an improved live weigh implementer
 *
 * This live weigh works by running a 2 pass algorithm that discards secondaryAvNumDiscard of the out lying secondary readings,
 * and then averaging the remaining readings.
 *
 * The primary average is simply a sum/num average that feeds into the secondary averaging.
 *
 * The averaging part of the the live weigh will take (primaryAvNumRdgs * secondaryAvNumRdgs) / (sampleFrequency / 10) seconds to complete
 *
 * The sequence is:
 *  - wait for weight to go over target
 *  - delay for holdOffPresetMilliseconds
 *  - gather readings
 *  - calculate average
 *
 * @param ass is the ADC super structure running this live weigh. ass mutex should be locked before calling this function
 * @param alwi is an initialised improved live weigh implementer
 * @param sampleFrequency is the sample frequency that the live weigh is called at. In Hz * 10 (ie 100 = 10Hz)
 * @param holdOffPresetMilliseconds is the number of milliseconds to wait before collecting samples after weight has gone above target
 * @param primaryAvNumRdgs is the number of primary readings to average @ ADC sample frequency
 * @param secondaryAvNumRdgs is the number of secondary readings to average. An array is dynamically allocated containing secondaryAvNumRdgs elements
 * @param target is the level that the weight must go above to trigger live weighing. In gross final weight.
 * @param secondaryAvNumDiscard is the number of secondary readings to discard. Because of the 2 pass system, twice this number are discarded. Must be < secondaryAvNumRdgs / 2
 * @param retrigPercentPreset (0-100) When the weight changes by this percentage of the last average, weighing will be re-triggered
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcLiveWeighImplementerImprovedConfigure(AdcSuperStructure *ass,
                                               AdcLiveWeighImplementer *alwi,
                                               int32u sampleFrequency,
                                               int32u holdOffPresetMilliseconds,
                                               size_t primaryAvNumRdgs,
                                               size_t secondaryAvNumRdgs,
                                               int32s target,
                                               size_t secondaryAvNumDiscard,
                                               int8u retrigPercentPreset)
{
  AdcLiveImproved *ali;

  /* Check parameters */
  if (NULL == ass || NULL == alwi || NULL == alwi->implementer_data)
  {
    return ADC_OP_STATUS_FAILED_PARAM;
  }
  ali = (AdcLiveImproved *) alwi->implementer_data;
  ali->configured = false;

  if (0 == sampleFrequency || 0 == primaryAvNumRdgs || 0 == secondaryAvNumRdgs)
  {
    return ADC_OP_STATUS_FAILED_PARAM;
  }

  /* secondaryAvNumDiscard needs to be less than half secondaryAvNumRdgs, otherwise all readings will be discarded in the averaging process */
  if (secondaryAvNumDiscard >= secondaryAvNumRdgs / 2)
  {
    return ADC_OP_STATUS_FAILED_PARAM;
  }

  /* Configure the live weigh */
  ali->sampleFrequency = sampleFrequency;
  ali->holdOffMillisecondsPreset = holdOffPresetMilliseconds;
  ali->primaryAvNumRdgs = primaryAvNumRdgs;
  ali->target = AdcIntFinalToMvv(ass, target);
  ali->secondaryAvNumDiscard = secondaryAvNumDiscard;
  ali->retrigPercentPreset = retrigPercentPreset;
  (void) AdcIntDisableLiveImproved(ali);

  if (secondaryAvNumRdgs != ali->secondaryAvNumRdgs || NULL == ali->secondaryAverages)
  {
    /* Changed to number of secondary readings or buffer not allocated */

    /* Free existing average buffer */
    if (NULL != ali->secondaryAverages)
    {
      free(ali->secondaryAverages);
      ali->secondaryAverages = NULL;
    }

    /* Allocate new average buffer */
    ali->secondaryAverages = malloc(sizeof(AdcLiveImprovedAverage) * secondaryAvNumRdgs);
    if (NULL == ali->secondaryAverages)
    {
      ali->secondaryAvNumRdgs = 0;
    }
    else
    {
      ali->secondaryAvNumRdgs = secondaryAvNumRdgs;
      ali->configured = true;
    }
  }
  else
  {
    /* Number of secondary readings unchanged, and buffer already allocated */
    ali->configured = true;
  }

  return (ali->configured) ?  ADC_OP_STATUS_SUCCESS : ADC_OP_STATUS_FAILED_ERROR;
}
