/*
 * adcliveimpl_improved.h
 *
 * Improved live weigh algorithm, stolen from R300 & R400-A413
 *
 *  Created on: Jun 25, 2012
 *      Author: ian
 */

#ifndef ADCLIVEIMPL_IMPROVED_H_
#define ADCLIVEIMPL_IMPROVED_H_


#include <stdint.h>
#include "adcdatatypes.h"
#include "adcliveimpl.h"


enum eLiveImprovedState
{
   LIVE_STATE_OFF = 0,
   LIVE_STATE_WAIT_TARGET,
   LIVE_STATE_HOLDOFF,
   LIVE_STATE_COLLECT,
   LIVE_STATE_COMPLETE
};


typedef struct
{
  int32s average;
  int32s delta;
  int8u  flag;
} AdcLiveImprovedAverage;


/*
 * From wiki:
 *
 * http://wiki.rinstrumau.local/wiki/index.php?title=R420_Upgrade_for_livestock_averaging_and_external_RFID_Support
 *
 * This is implemented in the ADC library:
 * DELAY: this is the time to wait before collecting samples after the weight has gone above the target (see below). Will be skipped if set to 0. The clock annunciator will be shown while waiting for this time to elapse.
 * SAMPLE: This is the length of time to collect samples for to hold a weight.
 * TARGET: this is the minimum weight on the scale that is required before it assumes an animal is present.
 * TOL: This sets how many averaged samples to discard.
 *
 * You need to implement this in your application:
 * RETRIG: This is the change in held weight in % before the animal weight is recalculated. Retriggering will not occur if this is set to 0.
 * PRT.OUT: This sets the printout to use when a weight is held, no printing will occur if it is set to none or it is set to a record print.
 * TOTAL: This sets whether to add to totals or not, this will only occur if a printout is used. The previous add to totals will be removed if the weight is retriggered.
 * INPUT: If set, this input must be present when the weight goes above the target before the delay starts. Will be skipped if set to NONE. The clock annunciator will flash while waiting for the input.
 *
 */


/** Struct to implement improved live weighing */
typedef struct
{
  enum eLiveImprovedState liveStockState;       /*!< Current live stock state */
  enum eLiveImprovedState prevLiveStockState;   /*!< Previous live stock state, used for debugging */

  /* Primary average */
  int32s livePrimaryAvMax;                    /*!< Maximum reading during a primaryAvNumRdgs accumulation of readings */
  int32s livePrimaryAvMin;                    /*!< Minimum reading during a primaryAvNumRdgs accumulation of readings */
  int64_t livePrimaryAvTotal;                 /*!< primaryAvNumRdgs reading are accumulated in here */
  size_t livePrimaryAvCurrentRdg;             /*!< Counts up to primaryAvNumRdgs as readings are accumulated */

  /* Secondary average, once livePrimaryAvCurrentRdg reaches primaryAvNumRdgs, livePrimaryAvTotal is averaged into the next secondaryAverages */
  AdcLiveImprovedAverage *secondaryAverages;  /*!< Array with secondaryAvNumRdgs elements */
  size_t secondaryAvCurrentRdg;               /*!< Counts up to secondaryAvNumRdgs as secondary readings are accumulated */

  int32s livePass;
  int8u  liveRestartOk;
  int32s maxAverage, minAverage;
  int32u holdOffSampleCounter;        /*!< Counter used to delay for holdOffMillisecondsPreset. Count in samples at the sync frequency */

  /* These are user settings */
  bool configured;                    /*!< Flag set after live weigh has been configured. Won't run until configured */
  int32u sampleFrequency;             /*!< The sample frequency that the live weigh is called at. In Hz * 10 (ie 100 = 10Hz) */
  int32u holdOffMillisecondsPreset;   /*!< The number of milliseconds to wait before collecting samples after weight has gone above target */
  size_t primaryAvNumRdgs;            /*!< Number of primary readings to average @ ADC sample frequency */
  size_t secondaryAvNumRdgs;          /*!< The number of secondary live averages */
  int32s target;                      /*!< Weight must go above this to trigger live weighing */
  size_t secondaryAvNumDiscard;       /*!< The number of secondary readings to discard. Because of the 2 pass system, twice this number are discarded. Must be < secondaryAvNumRdgs / 2 */
  int8u retrigPercentPreset;          /*!< (0-100) When the weight changes by this percentage of the last average, weighing will be re-triggered */
} AdcLiveImproved;


/**
 * Create an improved live weigh implementer.
 *
 * Once created, the live weigh implementer will need to be configured.
 *
 * @return pointer to initialised AdcLiveWeighImplementer structure
 */
AdcLiveWeighImplementer *AdcLiveWeighImplementerImprovedCreate(void);


/**
 * Configure an improved live weigh implementer
 *
 * This live weigh works by running a 2 pass algorithm that discards secondaryAvNumDiscard of the out lying secondary readings,
 * and then averaging the remaining readings.
 *
 * The primary average is simply a sum/num average that feeds into the secondary averaging.
 *
 * The averaging part of the the live weigh will take (primaryAvNumRdgs * secondaryAvNumRdgs) / (sampleFrequency / 10) seconds to complete
 *
 * The sequence is:
 *  - wait for weight to go over target
 *  - delay for holdOffPresetMilliseconds
 *  - gather readings
 *  - calculate average
 *
 * @param ass is the ADC super structure running this live weigh. ass mutex should be locked before calling this function
 * @param alwi is an initialised improved live weigh implementer
 * @param sampleFrequency is the sample frequency that the live weigh is called at. In Hz * 10 (ie 100 = 10Hz)
 * @param holdOffPresetMilliseconds is the number of milliseconds to wait before collecting samples after weight has gone above target
 * @param primaryAvNumRdgs is the number of primary readings to average @ ADC sample frequency
 * @param secondaryAvNumRdgs is the number of secondary readings to average. An array is dynamically allocated containing secondaryAvNumRdgs elements
 * @param target is the level that the weight must go above to trigger live weighing. In gross final weight.
 * @param secondaryAvNumDiscard is the number of secondary readings to discard. Because of the 2 pass system, twice this number are discarded. Must be < secondaryAvNumRdgs / 2
 * @param retrigPercentPreset (0-100) When the weight changes by this percentage of the last average, weighing will be re-triggered
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcLiveWeighImplementerImprovedConfigure(AdcSuperStructure *ass,
                                               AdcLiveWeighImplementer *alwi,
                                               int32u sampleFrequency,
                                               int32u holdOffPresetMilliseconds,
                                               size_t primaryAvNumRdgs,
                                               size_t secondaryAvNumRdgs,
                                               int32s target,
                                               size_t secondaryAvNumDiscard,
                                               int8u retrigPercentPreset);


#endif /* ADCLIVEIMPL_IMPROVED_H_ */
