/*
 * adclivesimple.c
 *
 * Simple live weighing simply averages a specified number of
 * ADC post-averaged readings (after waiting for the main averaging buffer to refill completely)
 */
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcinterface.h>

#include <rinweigh/adcliveimpl_simple.h>


/**
 * Run the simple live weigh
 * Needs to be called at the sample frequency
 *
 * @param als is the simple live weigh structure
 * @param rdg is the reading to be averaged
 * @param resultrdg is a pointer to a value set with the live weigh result on success
 *
 * @return alwrComplete if a live weight has been obtained
 */
static enum eAdcLiveWeighResult AdcIntRunLiveSimple(AdcLiveSimple *als, int32s rdg, int32s *resultrdg)
{
  enum eAdcLiveWeighResult result;

  if (!als->live_enabled || 0 == als->live_toaverage || 0 == als->live_main_average_size)
  {
    /* Live weigh disabled */
    result = alwrDisabled;
  }
  else if (als->live_averaged >= als->live_toaverage)
  {
    /* Live weigh has completed (previously) */
    result = alwrComplete;
  }
  else
  {
    result = alwrRunning;

    /* Run the live weigh */
    if (als->live_main_average_samples++ >= als->live_main_average_size)
    {
      /* Complete new set of readings in the ADC buffer */
      als->live_total += rdg;
      als->live_averaged++;
      if (als->live_averaged >= als->live_toaverage)
      {
        /* Finished, calculate result */
        als->live_result = als->live_total /  als->live_toaverage;
        *resultrdg = als->live_result;
        result = alwrComplete;
      }
    }
  }
  return result;
}


/** Initiate simple live weighing */
static enum eAdcLiveWeighResult AdcIntInitiateLiveSimple(AdcLiveSimple *als)
{
  if (0 == als->live_main_average_size || 0 == als->live_toaverage)
  {
    return alwrFailed;
  }

  als->live_total = 0;
  als->live_averaged = 0;
  als->live_result = 0;
  als->live_main_average_samples = 0;
  als->live_enabled = true;

  return alwrRunning;
}


/** Disable simple live weighing */
static enum eAdcLiveWeighResult AdcIntDisableLiveSimple(AdcLiveSimple *als)
{
  als->live_enabled = false;
  return alwrDisabled;
}


static enum eAdcLiveWeighResult AdcLiveWeighSimpleRun(AdcLiveWeighImplementer *alwi, int32s rdg, int32s *resultrdg)
{
  AdcLiveSimple *als;

  if (NULL == alwi || NULL == alwi->implementer_data)
  {
    return alwrNotImplemented;
  }
  als = (AdcLiveSimple *) alwi->implementer_data;

  return AdcIntRunLiveSimple(als, rdg, resultrdg);
}


static enum eAdcLiveWeighResult AdcLiveWeighSimpleInitiate(AdcLiveWeighImplementer *alwi)
{
  AdcLiveSimple *als;

  if (NULL == alwi || NULL == alwi->implementer_data)
  {
    return alwrNotImplemented;
  }
  als = (AdcLiveSimple *) alwi->implementer_data;

  return AdcIntInitiateLiveSimple(als);
}


static enum eAdcLiveWeighResult AdcLiveWeighSimpleDisable(AdcLiveWeighImplementer *alwi)
{
  AdcLiveSimple *als;

  if (NULL == alwi || NULL == alwi->implementer_data)
  {
    return alwrNotImplemented;
  }
  als = (AdcLiveSimple *) alwi->implementer_data;

  return AdcIntDisableLiveSimple(als);
}


static enum eAdcLiveWeighResult AdcLiveWeighSimpleRunning(AdcLiveWeighImplementer *alwi)
{
  AdcLiveSimple *als;
  enum eAdcLiveWeighResult result;

  if (NULL == alwi || NULL == alwi->implementer_data)
  {
    return alwrNotImplemented;
  }
  als = (AdcLiveSimple *) alwi->implementer_data;

  if (!als->live_enabled || 0 == als->live_toaverage || 0 == als->live_main_average_size)
  {
    /* Not running, and hasn't completed */
    result = alwrDisabled;
  }
  else if (als->live_averaged < als->live_toaverage)
  {
    /* Running, not completed */
    result = alwrRunning;
  }
  else
  {
    /* Completed */
    result = alwrComplete;
  }
  return result;
}


/** Destroy a simple live weigh implementer */
static void AdcLiveWeighSimpleDestroy(AdcLiveWeighImplementer *alwi)
{
  AdcLiveSimple *als = NULL;

  if (NULL != alwi && NULL != alwi->implementer_data)
  {
    als = (AdcLiveSimple *) alwi->implementer_data;
    alwi->implementer_data = NULL;
    memset(als, 0, sizeof(*als));
    free(als);
  }
}


/** Create a simple live weigh implementer */
AdcLiveWeighImplementer *AdcLiveWeighImplementerSimpleCreate(void)
{
  AdcLiveWeighImplementer *alwi = NULL;
  AdcLiveSimple *als = NULL;

  /* Make structure */
  alwi = malloc(sizeof(AdcLiveWeighImplementer));
  if (NULL == alwi)
  {
    goto alwisc_fail;
  }

  /* Make and initialise simple live weigh */
  als = malloc(sizeof(AdcLiveSimple));
  if (NULL == als)
  {
    goto alwisc_fail;
  }
  (void) AdcIntDisableLiveSimple(als);

  /* Fill out members */
  alwi->implementer_data = als;
  alwi->AdcLiveWeighRun = AdcLiveWeighSimpleRun;
  alwi->AdcLiveWeighInitiate = AdcLiveWeighSimpleInitiate;
  alwi->AdcLiveWeighDisable = AdcLiveWeighSimpleDisable;
  alwi->AdcLiveWeighRunning = AdcLiveWeighSimpleRunning;
  alwi->AdcLiveWeighDestroy = AdcLiveWeighSimpleDestroy;

  /* Success */
  return alwi;

alwisc_fail:
  if (NULL != alwi)
  {
    free(alwi);
  }
  if (NULL != als)
  {
    free(als);
  }
  return NULL;
}



/**
 * Configure a simple live weigh implementer
 *
 * This live weigh works by averaging a large number of samples
 *
 * @param ass is the ADC super structure running this live weigh. ass mutex should be locked before calling this function
 * @param alwi is an initialised improved live weigh implementer
 * @param averagingMilliseconds is the number of milliseconds to average data for
 *
 * @return one of the ADC_OP_STATUS_* codes
 */
int8u AdcLiveWeighImplementerSimpleConfigure(AdcSuperStructure *ass,
                                             AdcLiveWeighImplementer *alwi,
                                             int32u averagingMilliseconds)
{
  AdcLiveSimple *ali;
  int32u main_avg_samples;
  int32u samples;

  /* Check parameters */
  if (NULL == ass || NULL == alwi || NULL == alwi->implementer_data)
  {
    return ADC_OP_STATUS_FAILED_PARAM;
  }
  ali = (AdcLiveSimple *) alwi->implementer_data;

  if (0 == averagingMilliseconds)
  {
    return ADC_OP_STATUS_FAILED_PARAM;
  }

  /*
   * Work out how how many samples to wait for a fresh set of readings through the main (system) averaging
   *
   * samples_per_main_average = (db_avg[*0.01s] * sync_freq[*0.1Hz]) / (100 * 10)
   */
  main_avg_samples = ass->adcUserDatabase.average * ass->adcMiscellaneous.adcRdgFreq;
  main_avg_samples /= ADC_AV_1_SEC * 10;

  /*
   * Now work out the number of main average samples to average to get averagingMilliseconds worth of data
   */
  samples = averagingMilliseconds * ass->adcMiscellaneous.adcRdgFreq / 10 / 1000;

  /* Configure the live weigh */
  (void) AdcIntDisableLiveSimple(ali);
  ali->live_toaverage = samples;
  ali->live_main_average_size = main_avg_samples;

  return ADC_OP_STATUS_SUCCESS;
}
