/*
 * adcmuteximpl_pthreads_timed.c
 *
 *  Created on: 31 Jul. 2018
 *      Author: ian
 */


#include <pthread.h>
#include <stdlib.h>
#include <rinhw/timemath.h>
#include "adcmuteximpl.h"
#include "adcdebug.h"

#include "adcmuteximpl_pthreads_timed.h"


struct s_adc_mutex_pthreads_timed
{
  pthread_mutex_t mutex;      /*!< The mutex */
  struct timespec unlocktime; /*!< The amount of time that the mutex has been unlocked for */
  struct timespec locktime;   /*!< The amount of time that the mutex has been locked for */

  struct timespec lockstart;  /*!< The time when the lock was achieved */
  struct timespec lockend;    /*!< The time when the lock was released */
};


/** Lock the pthreads mutex implementer and update the times */
static bool AdcMutexImplementerPthreadsTimedLock(AdcMutexImplementer *ami)
{
  struct s_adc_mutex_pthreads_timed *ampt = (struct s_adc_mutex_pthreads_timed *) ami->implementer_data;
  struct timespec diff;
  bool result;

  result = (0 == pthread_mutex_lock(&ampt->mutex));
  clock_gettime(CLOCK_MONOTONIC, &ampt->lockstart);

  /* Accumulate unlocked time */
  diff = ampt->lockstart;
  SubTimespec(&diff, &ampt->lockend);
  AddTimespec(&ampt->unlocktime, &diff);
  return result;
}


/** Unlock the pthreads mutex implementer */
static bool AdcMutexImplementerPthreadsTimedUnlock(AdcMutexImplementer *ami)
{
  struct s_adc_mutex_pthreads_timed *ampt = (struct s_adc_mutex_pthreads_timed *) ami->implementer_data;
  struct timespec diff;
  struct timespec total;
  bool result;

  result = (0 == pthread_mutex_unlock(&ampt->mutex));
  clock_gettime(CLOCK_MONOTONIC, &ampt->lockend);

  /* Accumulate locked time */
  diff = ampt->lockend;
  SubTimespec(&diff, &ampt->lockstart);
  AddTimespec(&ampt->locktime, &diff);

  total = ampt->unlocktime;
  AddTimespec(&total, &ampt->locktime);
  if (total.tv_sec > 60)
  {
    ADC_DEBUG_PRINTF(LOG_INFO, "AdcMutexImplementerPthreadsTimedUnlock: adc mutex was locked for %ld.%09lu seconds of the past %ld.%09lu seconds\n",
                     (long) ampt->locktime.tv_sec, (unsigned long) ampt->locktime.tv_nsec,
                     (long) total.tv_sec, (unsigned long) total.tv_nsec);
    ampt->locktime.tv_sec = 0;
    ampt->locktime.tv_nsec = 0;
    ampt->unlocktime.tv_sec = 0;
    ampt->unlocktime.tv_nsec = 0;
  }
  return result;
}


/**
 * Create a timed mutex implementer using pthreads
 *
 * This mutex is for debugging purposes. It calculates the
 * time spent locked vs unlocked and prints the result
 * every 60 seconds via ADC_DEBUG_PRINTF().
 *
 * It shouldn't be used in production code, as it will
 * have a performance impact with the additional calculations
 * and debug prints.
 */
AdcMutexImplementer *AdcMutexImplementerPthreadsTimedCreate(void)
{
  AdcMutexImplementer *ami = NULL;
  struct s_adc_mutex_pthreads_timed *ampt = NULL;

  /* Make structure */
  ami = malloc(sizeof(AdcMutexImplementer));
  if (NULL == ami)
  {
    goto amifct_fail;
  }

  /* Make and initialise mutex */
  ampt = calloc(1, sizeof(struct s_adc_mutex_pthreads_timed));
  if (NULL == ampt)
  {
    goto amifct_fail;
  }
  if (0 != pthread_mutex_init(&ampt->mutex, NULL))
  {
    goto amifct_fail;
  }
  clock_gettime(CLOCK_MONOTONIC, &ampt->lockend);
  ampt->lockstart = ampt->lockend;
  ampt->locktime.tv_sec = 0;
  ampt->locktime.tv_nsec = 0;
  ampt->unlocktime.tv_sec = 0;
  ampt->unlocktime.tv_nsec = 0;

  /* Fill out members */
  ami->implementer_data = ampt;
  ami->mutex_lock_fn = AdcMutexImplementerPthreadsTimedLock;
  ami->mutex_unlock_fn = AdcMutexImplementerPthreadsTimedUnlock;

  /* Success */
  return ami;

amifct_fail:
  if (NULL != ami)
  {
    free(ami);
  }
  if (NULL != ampt)
  {
    free(ampt);
  }
  return NULL;
}


/** Destroy a timed mutex implementer using pthreads */
bool AdcMutexImplementerPthreadsTimedDestroy(AdcMutexImplementer *ami)
{
  bool result = false;
  struct s_adc_mutex_pthreads_timed *ampt = NULL;

  if (NULL != ami)
  {
    if (NULL != ami->implementer_data)
    {
      ampt = (struct s_adc_mutex_pthreads_timed *) ami->implementer_data;
      result = (0 == pthread_mutex_destroy(&ampt->mutex));
      free(ami->implementer_data);
    }
    free(ami);
    ami = NULL;
  }
  return result;
}
