#include <stdlib.h>
#include <string.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/rio.h>
#include <rinhw/rtc.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcreadings.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcmotion.h>
#include <rinweigh/adcaverage.h>
#include <rinweigh/adczerotrack.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcads1216.h>
#include <rinweigh/adclin.h>
#include <rinweigh/adcconvert.h>
#include <rinweigh/adcmiscellaneous.h>
#include <rinweigh/adcdatafiller.h>
#include <rinweigh/adcliveimpl_int.h>
#include <rinweigh/adcmutex.h>

#include <rinweigh/adcoperations.h>


// prototypes
static void AdcOpSemiAutoZero(AdcSuperStructure *ass);
static void AdcOpClearSemiAutoZero(AdcSuperStructure *ass);
static void AdcOpSemiAutoTare(AdcSuperStructure *ass);
static void AdcOpCalZero(AdcSuperStructure *ass);
static void AdcOpCalSpan(AdcSuperStructure *ass);
static void AdcOpCalFacZero(AdcSuperStructure *ass);
static void AdcOpCalFacSpan(AdcSuperStructure *ass);
static void AdcOpGrossNet(AdcSuperStructure *ass, int8s opt);
static void AdcOpDirectZero(AdcSuperStructure *ass);
static void AdcOpDirectSpan(AdcSuperStructure *ass);
static void AdcOpPresetTare(AdcSuperStructure *ass);
static void AdcOpAddLin(AdcSuperStructure *ass);
static void AdcOpClearLin(AdcSuperStructure *ass);
static void AdcOpWaitForMotion(AdcSuperStructure *ass);
static void AdcOpGetTraceableRdg(AdcSuperStructure *ass);
static void AdcOpGetNonTraceableRdg(AdcSuperStructure *ass);
static void AdcOpClearPeak(AdcSuperStructure *ass);
static void AdcOpInitialZero(AdcSuperStructure *ass);
static void AdcOpSetCounting(AdcSuperStructure *ass);
static void AdcOpDoLive(AdcSuperStructure *ass);
static void AdcOpSetDisplayMode(AdcSuperStructure *ass);
static void AdcOpSetAlt(AdcSuperStructure *ass);
static void AdcOpBlockZT(AdcSuperStructure *ass);
static void AdcOpManualHold(AdcSuperStructure *ass);
static void AdcOpClearManualHold(AdcSuperStructure *ass);
static void AdcOpWaitForAverage(AdcSuperStructure *ass);
static void AdcOpSetTert(AdcSuperStructure *ass);

static bool AdcIntCheckCancelOK(AdcSuperStructure *ass);


/** Init operations (call only once) */
void AdcIntInitOperations(AdcSuperStructure *ass)
{
  memset(&ass->adcOperation, 0, sizeof(AdcOperation));
  ass->adcOperation.opSem = OSSemCreate(0);
}


/** Destroy operations */
void AdcIntDestroyOperations(AdcSuperStructure *ass)
{
  int8u err;

  if (NULL != ass->adcOperation.opSem)
  {
    OSSemDel(ass->adcOperation.opSem, OS_DEL_ALWAYS, &err);
  }
}


/** Re-initialise operations, can be called multiple times */
void AdcIntReInitOperations(AdcSuperStructure *ass)
{
  ass->adcOperation.lastCommand = ADC_OP_CMD_NONE;
  ass->adcOperation.status      = ADC_OP_STATUS_SUCCESS;
  ass->adcOperation.timeoutSec  = ADC_OP_TIMEOUT_IMMEDIATE;
  ass->adcOperation.timeoutSync = 0;
  ass->adcOperation.canceling   = FALSE;
  ass->adcOperation.cmdCnt      = 0;
  ass->adcOperation.immediate   = FALSE;

  ass->adcOperation.averageLen = ass->adcUserDatabase.average;
  ass->adcTraceableWeight.valid = FALSE;
}


/**
 * Start an ADC operation.
 *
 * @param ass is the ADC super structure
 * @param command is the command to execute. See ADC_OP_CMD_* constants
 * @param timeout is the timeout
 * @param block is whether the operation should block until complete/failure
 * @param pL1 is user parameter 1, command specific
 * @param pL2 is user parameter 2, command specific
 * @param immediate means the operation should complete immediately, without waiting for no motion. Generally only works in non-trade mode
 */
int8u AdcIntOperationStart(AdcSuperStructure *ass, int8u command, int8u timeout, int8u block, int32s pL1, int32s pL2, bool immediate)
{
  int8u semres;
  int8u result;

  /* Check that something else isn't already running, and block if need be */
  if (ass->adcOperation.status == ADC_OP_STATUS_IN_PROGRESS)
  {
    if (block)
    {
      while (ass->adcOperation.status == ADC_OP_STATUS_IN_PROGRESS)
      {
        AdcMutexUnlock(ass);
        OSTimeDly(OS_TICKS_PER_SEC / 10);
        AdcMutexLock(ass);
      }
    }
    else
    {
      result = ADC_OP_STATUS_FAILED_BUSY;
      goto aios_fail;
    }
  }

  /* Setup for operation */
  ass->adcOperation.status      = ADC_OP_STATUS_IN_PROGRESS;
  ass->adcOperation.tickStatus  = ADC_OP_STATUS_CANCELED;
  ass->adcOperation.lastCommand = command;
  ass->adcOperation.timeoutSec  = timeout;
  if (timeout == 0)
    ass->adcOperation.timeoutSync = 1;
  else
  {
    ass->adcOperation.timeoutSync = timeout;
    ass->adcOperation.timeoutSync *= (ass->adcMiscellaneous.adcRdgFreq / 10);
  }

  ass->adcOperation.canceling   = FALSE;
  ass->adcOperation.block       = block;
  ass->adcOperation.counter     = 0;
  ass->adcOperation.paramL1     = pL1;
  ass->adcOperation.paramL2     = pL2;
  ass->adcOperation.immediate   = immediate;
  ass->adcOperation.cmdCnt++;

  /* Wait for operation to complete if need be */
  if (block)
  {
    AdcMutexUnlock(ass);
    OSSemPend(ass->adcOperation.opSem, 0, &semres);
    AdcMutexLock(ass);
    result = ass->adcOperation.status;
  }
  else
  {
    result = ADC_OP_STATUS_IN_PROGRESS;
  }

aios_fail:
  return result;
}


/**
 * Cancel the current ADC operation
 *
 * @param ass is the ADC data super structure
 * @param lock specifies if the function should lock ass before performing it's function. This allows the caller to have locked the structure prior to the call
 */
void AdcOperationCancel(AdcSuperStructure *ass, bool lock)
{
  ADC_INT_IFACE_LOCK();
  if (ass->adcOperation.status == ADC_OP_STATUS_IN_PROGRESS)
    ass->adcOperation.canceling = TRUE;
  ADC_INT_IFACE_UNLOCK();
}


/**
 * Must be called at the sync frequency to perform operations
 *
 * Note: any function which requires more than a single iteration must be done here
 */
void AdcIntOperationTick(AdcSuperStructure *ass)
{
  int8u cmd;

  // handle if nothing to do
  if (ass->adcOperation.status != ADC_OP_STATUS_IN_PROGRESS)
  {
    return;
  }

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflOperationStart);

  // handle cancel
  if ((ass->adcOperation.canceling) && (AdcIntCheckCancelOK(ass)))
  {
    ass->adcOperation.status = ADC_OP_STATUS_CANCELED;
    ass->adcInternalData.userBlocked = FALSE;
    if (ass->adcUserDatabase.average != ass->adcOperation.averageLen)
    {
      ass->adcOperation.averageLen = ass->adcUserDatabase.average;
      AdcIntInitAverage(&ass->adcAveraging, ass->adcOperation.averageLen);
    }
    if (ass->adcOperation.block)
    {
      OSSemPost(ass->adcOperation.opSem);
    }
    goto aiot_done;
  }

  // handle timeout
  if ((ass->adcOperation.timeoutSync == ass->adcOperation.counter) && (AdcIntCheckCancelOK(ass)))
  {
    ass->adcOperation.status = ass->adcOperation.tickStatus;
    ass->adcInternalData.userBlocked = FALSE;
    if (ass->adcUserDatabase.average != ass->adcOperation.averageLen)
    {
      ass->adcOperation.averageLen = ass->adcUserDatabase.average;
      AdcIntInitAverage(&ass->adcAveraging, ass->adcOperation.averageLen);
    }
    if (ass->adcOperation.block)
    {
      OSSemPost(ass->adcOperation.opSem);
    }
    goto aiot_done;
  }
  ass->adcOperation.counter++;
  cmd = ass->adcOperation.lastCommand;

  switch (cmd)
  {
    case ADC_OP_CMD_SEMIAUTO_ZERO:
      AdcOpSemiAutoZero(ass);
      break;
    case ADC_OP_CMD_SEMIAUTO_TARE:
      AdcOpSemiAutoTare(ass);
      break;
    case ADC_OP_CMD_CAL_ZERO:
      AdcOpCalZero(ass);
      break;
    case ADC_OP_CMD_CAL_SPAN:
      AdcOpCalSpan(ass);
      break;
    case ADC_OP_CMD_CAL_LIN_ADD:
      AdcOpAddLin(ass);
      break;
    case ADC_OP_CMD_FACTORY_ZERO:
      AdcOpCalFacZero(ass);
      break;
    case ADC_OP_CMD_FACTORY_SPAN:
      AdcOpCalFacSpan(ass);
      break;
    case ADC_OP_CMD_WAIT_FOR_MOTION:
      AdcOpWaitForMotion(ass);
      break;
    case ADC_OP_CMD_TRACEABLE_RDG:
      AdcOpGetTraceableRdg(ass);
      break;
    case ADC_OP_CMD_NON_TRACEABLE_RDG:
      AdcOpGetNonTraceableRdg(ass);
      break;
    case ADC_OP_CMD_INIT_ZERO:
      AdcOpInitialZero(ass);
      break;
    case ADC_OP_CMD_DO_LIVE:
      AdcOpDoLive(ass);
      break;
    case ADC_OP_CMD_CLEAR_SEMIAUTO_ZERO:
      AdcOpClearSemiAutoZero(ass);
      break;
    case ADC_OP_CMD_MANUAL_HOLD:
      AdcOpManualHold(ass);
      break;
    case ADC_OP_CMD_CLEAR_MANUAL_HOLD:
      AdcOpClearManualHold(ass);
      break;
    case ADC_OP_CMD_WAIT_FOR_AVERAGE:
      AdcOpWaitForAverage(ass);
      break;
    case ADC_OP_CMD_SET_PRESET_TARE:
    case ADC_OP_CMD_CAL_LIN_CLEAR:
    case ADC_OP_CMD_SET_COUNTING:
    case ADC_OP_CMD_SET_ALT_WEIGHT:
    case ADC_OP_CMD_CLEAR_PEAK:
    case ADC_OP_CMD_GROSSNET_TOGGLE:
    case ADC_OP_CMD_GROSSNET_GROSS:
    case ADC_OP_CMD_GROSSNET_NET:
    case ADC_OP_CMD_CAL_ZERO_DIRECT:
    case ADC_OP_CMD_CAL_SPAN_DIRECT:
    case ADC_OP_CMD_SET_DISPLAY_MODE:
    case ADC_OP_CMD_BLOCK_ZT:
    case ADC_OP_CMD_SET_TERT_WEIGHT:
      break;

    default:
      ass->adcOperation.status = ADC_OP_STATUS_FAILED_NOT_IMP;
      break;
  }
  if ((ass->adcOperation.status != ADC_OP_STATUS_IN_PROGRESS) && (ass->adcOperation.block))
    OSSemPost(ass->adcOperation.opSem);

aiot_done:
  /* Call back to application */
  AdcAppCallbackLocation(ass, acflOperationEnd);
}


/**
 * Should be called before UpdateReadings. Can be called faster than sync frequency.
 *
 * Note: anything which is processed here must succeed or fail in a single cycle.
 */
void AdcIntOperationTickBefore(AdcSuperStructure *ass)
{
  // handle if nothing to do
  if (ass->adcOperation.status != ADC_OP_STATUS_IN_PROGRESS)
  {
    return;
  }

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflOperationBeforeStart);

  switch (ass->adcOperation.lastCommand)
  {
    case ADC_OP_CMD_SET_PRESET_TARE:
      AdcOpPresetTare(ass);
      break;
    case ADC_OP_CMD_CAL_LIN_CLEAR:
      AdcOpClearLin(ass);
      break;
    case ADC_OP_CMD_GROSSNET_TOGGLE:
      AdcOpGrossNet(ass, -1);
      break;
    case ADC_OP_CMD_GROSSNET_GROSS:
      AdcOpGrossNet(ass, 0);
      break;
    case ADC_OP_CMD_GROSSNET_NET:
      AdcOpGrossNet(ass, 1);
      break;
    case ADC_OP_CMD_CAL_ZERO_DIRECT:
      AdcOpDirectZero(ass);
      break;
    case ADC_OP_CMD_CAL_SPAN_DIRECT:
      AdcOpDirectSpan(ass);
      break;
    case ADC_OP_CMD_CLEAR_PEAK:
      AdcOpClearPeak(ass);
      break;
    case ADC_OP_CMD_SET_COUNTING:
      AdcOpSetCounting(ass);
      break;
    case ADC_OP_CMD_SET_ALT_WEIGHT:
      AdcOpSetAlt(ass);
      break;
    case ADC_OP_CMD_SET_DISPLAY_MODE:
      AdcOpSetDisplayMode(ass);
      break;
    case ADC_OP_CMD_BLOCK_ZT:
      AdcOpBlockZT(ass);
      break;
    case ADC_OP_CMD_SET_TERT_WEIGHT:
      AdcOpSetTert(ass);
      break;
    case ADC_OP_CMD_SEMIAUTO_ZERO:
    case ADC_OP_CMD_SEMIAUTO_TARE:
    case ADC_OP_CMD_CAL_ZERO:
    case ADC_OP_CMD_CAL_SPAN:
    case ADC_OP_CMD_CAL_LIN_ADD:
    case ADC_OP_CMD_FACTORY_ZERO:
    case ADC_OP_CMD_FACTORY_SPAN:
    case ADC_OP_CMD_WAIT_FOR_MOTION:
    case ADC_OP_CMD_TRACEABLE_RDG:
    case ADC_OP_CMD_NON_TRACEABLE_RDG:
    case ADC_OP_CMD_INIT_ZERO:
    case ADC_OP_CMD_DO_LIVE:
    case ADC_OP_CMD_CLEAR_SEMIAUTO_ZERO:
    case ADC_OP_CMD_MANUAL_HOLD:
    case ADC_OP_CMD_CLEAR_MANUAL_HOLD:
      break;
  }
  if ((ass->adcOperation.status != ADC_OP_STATUS_IN_PROGRESS) && (ass->adcOperation.block))
    OSSemPost(ass->adcOperation.opSem);

  /* Call back to application */
  AdcAppCallbackLocation(ass, acflOperationBeforeEnd);
}


static bool AdcIntCheckCancelOK(AdcSuperStructure *ass)
{
  switch (ass->adcOperation.lastCommand)
  {
    case ADC_OP_CMD_SEMIAUTO_ZERO:
    case ADC_OP_CMD_SEMIAUTO_TARE:
    case ADC_OP_CMD_INIT_ZERO:
    case ADC_OP_CMD_WAIT_FOR_MOTION:
      return(TRUE);
    default :
      return(FALSE);
  }
}


/** Check whether we need to wait for motion */
static int8u AdcIntOpCheckWaitForMotion(const AdcSuperStructure *ass, bool immediate)
{
  int8u result;

  if (immediate)
  {
    /* Immediate is only available in non-trade mode, in which case motion is ignored, and will fail otherwise */
    if (ADC_APPROVAL_MODE_NONE != ass->adcUserDatabase.approvalMode)
      result = ADC_OP_STATUS_FAILED_TRADE;
    else
      result = ADC_OP_STATUS_SUCCESS;
  }
  else
  {
    /* Check motion */
    if (ass->adcDataWorking.grossRdg.info.status.bit.motion)
      result = ADC_OP_STATUS_FAILED_MOTION;
    else
      result = ADC_OP_STATUS_SUCCESS;
  }
  return result;
}


/**
 * Tests if a semi auto zero operation would succeed, but does not perform semi auto zero.
 *
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_BAND if outside zero range, ADC_OP_STATUS_SUCCESS if zero would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntSemiautoZeroOperationIsValid(const AdcSuperStructure *ass, bool immediate)
{
  int8u result;

  if (ass->adcDataWorking.error)
  {
    result = ADC_OP_STATUS_FAILED_ERROR;
  }
  else if (ass->adcInternalData.cntSysRdg < ass->adcInternalData.cntZeroRangeMin ||
           ass->adcInternalData.cntSysRdg > ass->adcInternalData.cntZeroRangeMax)
  {
    /* Outside zero range */
    result = ADC_OP_STATUS_FAILED_BAND;
  }
  else
  {
    /* Check for motion */
    result = AdcIntOpCheckWaitForMotion(ass, immediate);
  }
  return result;
}


static void AdcOpSemiAutoZero(AdcSuperStructure *ass)
{
  int8u result;

  /* Can semi-auto zero operation proceed? */
  result = AdcIntSemiautoZeroOperationIsValid(ass, ass->adcOperation.immediate);
  if (ADC_OP_STATUS_SUCCESS != result)
  {
    /* No something preventing it */
    if (ADC_OP_STATUS_FAILED_MOTION == result)
    {
      /* Can wait until no motion */
      ass->adcOperation.tickStatus = result;
    }
    else
    {
      /* Just fail straight away */
      ass->adcOperation.status = result;
    }
  }
  else
  {
    /* Yep, do the semi-auto zero */
    ass->adcRuntimeDatabase.cntZero = ass->adcInternalData.cntSysRdg;
    ass->adcRuntimeDatabase.cntTare = 0;
    ass->adcRuntimeDatabase.netMode = FALSE;
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
}


/** Clear an existing semi-auto zero */
static void AdcOpClearSemiAutoZero(AdcSuperStructure *ass)
{
  if (ADC_APPROVAL_MODE_NONE != ass->adcUserDatabase.approvalMode)
  {
    /* Not available in trade modes */
    ass->adcOperation.status = ADC_OP_STATUS_FAILED_TRADE;
  }
  else
  {
    /* Clear semi-automatic zero */
    ass->adcRuntimeDatabase.cntZero = 0;
    ass->adcRuntimeDatabase.cntTare = 0;
    ass->adcRuntimeDatabase.cntPTare = 0;
    ass->adcRuntimeDatabase.netMode = FALSE;
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
}


/** Return if the current gross weight is within the valid tare range (ie a AdcSemiautoTare() would succeed here) */
static bool AdcIntInValidRangeOfTare(const AdcSuperStructure *ass)
{
  bool result = true;

  /* Is tare permitted at this gross weight? */
  if ((ass->adcInternalData.cntGross < ass->adcInternalData.cntUload) ||      /* Gross below underload */
      (ass->adcInternalData.cntGross > ass->adcInternalData.cntOload) ||      /* Gross over overload */
      (ass->adcInternalData.approvalMode != ADC_APPROVAL_MODE_NONE && ass->adcDataWorking.grossRdg.rdg < 0) ||  /* Trade mode and reading less than zero */
      (ass->adcInternalData.approvalMode == ADC_APPROVAL_MODE_NTEP && ass->adcInternalData.rangeMax &&             /* NTEP polava */
       !ass->adcInternalData.rangeMulti && ass->adcDataWorking.grossRdg.rdg > ass->adcUserDatabase.fullScale[0]))
  {
    result = false;
  }
  return result;
}


/**
 * Tests if a tare operation would succeed, but does not perform tare.
 *
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_BAND if outside range of tare, ADC_OP_STATUS_SUCCESS if tare would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntSemiautoTareOperationIsValid(const AdcSuperStructure *ass, bool immediate)
{
  int8u result;

  if (ass->adcDataWorking.error)
  {
    result = ADC_OP_STATUS_FAILED_ERROR;
  }
  else if (!AdcIntInValidRangeOfTare(ass))
  {
    result = ADC_OP_STATUS_FAILED_BAND;
  }
  else
  {
    result = AdcIntOpCheckWaitForMotion(ass, immediate);
  }
  return result;
}


static void AdcOpSemiAutoTare(AdcSuperStructure *ass)
{
  int8u result;

  /* Can semi-auto tare operation proceed? */
  result = AdcIntSemiautoTareOperationIsValid(ass, ass->adcOperation.immediate);
  if (ADC_OP_STATUS_SUCCESS != result)
  {
    /* No something preventing it */
    if (ADC_OP_STATUS_FAILED_MOTION == result)
    {
      /* Can wait until no motion */
      ass->adcOperation.tickStatus = result;
    }
    else
    {
      /* Just fail straight away */
      ass->adcOperation.status = result;
    }
  }
  else
  {
    /* Yep, do the semi-auto tare */

    /* Apply tare, and clear preset tare */
    ass->adcRuntimeDatabase.cntTare = ass->adcInternalData.cntGross;
    ass->adcRuntimeDatabase.cntPTare = 0;

    /* In trade mode, tare within zero band clears tare value and switches to gross */
    if ((ass->adcInternalData.approvalMode != ADC_APPROVAL_MODE_NONE) &&
        (ass->adcRuntimeDatabase.cntTare >= -ass->adcInternalData.cntZeroBandMax) && (ass->adcRuntimeDatabase.cntTare <= ass->adcInternalData.cntZeroBandMax))
    {
      ass->adcRuntimeDatabase.cntTare = 0;
      ass->adcRuntimeDatabase.netMode = FALSE;
    }
    else
    {
      ass->adcRuntimeDatabase.netMode = TRUE;
    }
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
}


/** Initialise ADC for large averaging for calibration */
static void AdcIntSetCalAveraging(AdcSuperStructure *ass)
{
  if (ass->adcUserDatabase.average > ADC_AV_5_SECS)
    ass->adcOperation.averageLen = ass->adcUserDatabase.average;
  else
    ass->adcOperation.averageLen = ADC_AV_5_SECS;
  ass->adcOperation.averageLen = (ass->adcOperation.averageLen * ass->adcMiscellaneous.adcRdgFreq / 10) / ADC_AV_1_SEC;
  AdcIntInitAverage(&ass->adcAveraging, ass->adcOperation.averageLen);
  ass->adcOperation.averageLen++;    // first reading will be on 2nd count
}


static void AdcOpCalZero(AdcSuperStructure *ass)
{
  switch (ass->adcOperation.counter)
  {
    case 1:
      if ((ass->adcDataWorking.rawRdg > ADC_DB_USER_ZERORAW_MAXIMUM) ||
          (ass->adcDataWorking.rawRdg < ADC_DB_USER_ZERORAW_MINIMUM))
      {
        ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
        return;
      }

      //block readings to user
      ass->adcInternalData.userBlocked = TRUE;

      //initialise adc for large averaging
      AdcIntSetCalAveraging(ass);
      break;

    default:
      if (ass->adcOperation.counter == ass->adcOperation.averageLen)
      {
        ass->adcOperation.averageLen = ass->adcUserDatabase.average;

        // get zero reading, check it
        if ((ass->adcDataWorking.rawRdg > ADC_DB_USER_ZERORAW_MAXIMUM) ||
            (ass->adcDataWorking.rawRdg < ADC_DB_USER_ZERORAW_MINIMUM))
        {
          ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
        }
        else
        {
          ass->adcUserDatabase.zeroRaw = ass->adcDataWorking.rawRdg;
          ass->adcUserDatabase.calFlag = ADC_CAL_FLAG_OK;
          // ClearAllDbLin(); removed 15/03/2007 for PM.
          ass->adcInternalData.cntInitZero = 0;
          ass->adcRuntimeDatabase.cntZero = 0;
          ass->adcRuntimeDatabase.cntTare = 0;
          ass->adcRuntimeDatabase.netMode = FALSE;

          ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
        }
        AdcIntReInitReadings(ass);
      }
      break;
  }
}

static void AdcOpCalSpan(AdcSuperStructure *ass)
{
  int32s grad;

  switch (ass->adcOperation.counter)
  {
    case 1:
      grad = ass->adcOperation.paramL1 / AdcResolutions[ass->adcUserDatabase.resolution[0]];
      if ((ass->adcDataWorking.rawRdg > ADC_DB_USER_SPANRAW_MAXIMUM) ||
          (ass->adcDataWorking.rawRdg < ADC_DB_USER_SPANRAW_MINIMUM) ||
          (grad < ADC_SPAN_DIVS_MIN))                                     // span cal min: 100 grad0
      {
        ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
        return;
      }

      grad = (ass->adcDataWorking.rawRdg - ass->adcUserDatabase.zeroRaw) / grad;
      if ((grad < ADC_MVV_PER_DIV_MIN) && (grad > -ADC_MVV_PER_DIV_MIN))  // grad min: 0.025uV
      {
        ass->adcOperation.status = ADC_OP_STATUS_FAILED_RES;
        return;
      }

      //block readings to user
      ass->adcInternalData.userBlocked = TRUE;

      //initialise adc for large averaging
      AdcIntSetCalAveraging(ass);
      break;

    default:
      if (ass->adcOperation.counter == ass->adcOperation.averageLen)
      {
        ass->adcOperation.averageLen = ass->adcUserDatabase.average;

        // get zero reading, check it
        if ((ass->adcDataWorking.rawRdg > ADC_DB_USER_SPANRAW_MAXIMUM) ||
            (ass->adcDataWorking.rawRdg < ADC_DB_USER_SPANRAW_MINIMUM))
        {
          ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
        }
        else
        {
          grad = ass->adcOperation.paramL1 / AdcResolutions[ass->adcUserDatabase.resolution[0]];
          grad = (ass->adcDataWorking.rawRdg - ass->adcUserDatabase.zeroRaw) / grad;
          if ((grad < ADC_MVV_PER_DIV_MIN) && (grad > -ADC_MVV_PER_DIV_MIN))  // grad min: 0.025uV
          {
            ass->adcOperation.status = ADC_OP_STATUS_FAILED_RES;
          }
          else
          {
            ass->adcUserDatabase.spanRaw = ass->adcDataWorking.rawRdg - ass->adcUserDatabase.zeroRaw;
            ass->adcUserDatabase.spanCnt = ass->adcOperation.paramL1*16;
            ass->adcUserDatabase.calWeight = ass->adcOperation.paramL1;
            ass->adcUserDatabase.calFlag = ADC_CAL_FLAG_OK;
            // ClearAllDbLin(); removed 15/03/2007 for PM.
            ass->adcUserDatabase.linCor[ADC_MAX_NUM_LIN_PTS] = 0;
            ass->adcUserDatabase.linPts[ADC_MAX_NUM_LIN_PTS] = AdcIntLinearisationGetLinPt(&ass->adcLinearisation, ass->adcUserDatabase.spanCnt);
            ass->adcRuntimeDatabase.cntZero = 0;
            ass->adcRuntimeDatabase.cntTare = 0;
            ass->adcRuntimeDatabase.netMode = FALSE;

            ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
          }
        }
        AdcIntReInitReadings(ass);
      }
      break;
  }
}

static void AdcOpGrossNet(AdcSuperStructure *ass, int8s opt)
{
  switch (opt)
  {
    case -1:
      if (ass->adcRuntimeDatabase.netMode ||
          ass->adcRuntimeDatabase.cntTare != 0 ||
          ass->adcRuntimeDatabase.cntPTare != 0)
      {
        ass->adcRuntimeDatabase.netMode = !ass->adcRuntimeDatabase.netMode;
      }
      break;
    case 0:
      ass->adcRuntimeDatabase.netMode = FALSE;
      break;
    case 1:
      if (ass->adcRuntimeDatabase.cntTare != 0 ||
          ass->adcRuntimeDatabase.cntPTare != 0)
      {
        ass->adcRuntimeDatabase.netMode = TRUE;
      }
      break;
  }
  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
}

static void AdcOpDirectZero(AdcSuperStructure *ass)
{
  ass->adcOperation.paramL1 *= 256;
  if ((ass->adcOperation.paramL1 > ADC_DB_USER_ZERORAW_MAXIMUM) || (ass->adcOperation.paramL1 < ADC_DB_USER_ZERORAW_MINIMUM))
   {
    ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
    return;
  }

  ass->adcUserDatabase.zeroRaw = ass->adcOperation.paramL1;
  ass->adcUserDatabase.calFlag = ADC_CAL_FLAG_OK;
  // ClearAllDbLin(); removed 15/03/2007 for PM.
  ass->adcInternalData.cntInitZero = 0;
  ass->adcRuntimeDatabase.cntZero = 0;
  ass->adcRuntimeDatabase.cntTare = 0;
  ass->adcRuntimeDatabase.netMode = FALSE;

  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  AdcIntReInitReadings(ass);
}

static void AdcOpDirectSpan(AdcSuperStructure *ass)
{
  int32s grad;

  grad = ass->adcOperation.paramL1 / AdcResolutions[ass->adcUserDatabase.resolution[0]];
  if (ass->adcOperation.paramL1 > ADC_DB_USER_SPANRAW_MAXIMUM ||
      ass->adcOperation.paramL1 < ADC_DB_USER_SPANRAW_MINIMUM ||
      grad < ADC_SPAN_DIVS_MIN)
  {
    ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
    return;
  }

  ass->adcOperation.paramL2 *= 256;
  grad = ass->adcOperation.paramL2 / grad;
  if ((grad < ADC_MVV_PER_DIV_MIN) && (grad > -ADC_MVV_PER_DIV_MIN))  // grad min: 0.025uV
  {
    ass->adcOperation.status = ADC_OP_STATUS_FAILED_RES;
    return;
  }

  ass->adcUserDatabase.spanRaw = ass->adcOperation.paramL2;
  ass->adcUserDatabase.spanCnt = ass->adcOperation.paramL1*16;
  ass->adcUserDatabase.calWeight = ass->adcOperation.paramL1;
  ass->adcUserDatabase.calFlag = ADC_CAL_FLAG_OK;
  // ClearAllDbLin(); removed 15/03/2007 for PM.
  ass->adcUserDatabase.linCor[ADC_MAX_NUM_LIN_PTS] = 0;
  ass->adcUserDatabase.linPts[ADC_MAX_NUM_LIN_PTS] = AdcIntLinearisationGetLinPt(&ass->adcLinearisation, ass->adcUserDatabase.spanCnt);
  ass->adcRuntimeDatabase.cntZero = 0;
  ass->adcRuntimeDatabase.cntTare = 0;
  ass->adcRuntimeDatabase.netMode = FALSE;

  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  AdcIntReInitReadings(ass);
}


/**
 * Tests if a preset tare operation would succeed, but does not perform preset tare.
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_BAND if outside range of tare,
 *         ADC_OP_STATUS_SUCCESS if tare would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntPresetTareOperationIsValid(const AdcSuperStructure *ass, int32s presetTare)
{
  int8u result;
  int32s cntPresetTare;
  int32s highlim;
  int32s lowlim;

  if (ass->adcDataWorking.error)
  {
    result = ADC_OP_STATUS_FAILED_ERROR;
  }
  else
  {
    /* Convert PT to counts */
    cntPresetTare = 16 * presetTare;

    /* Calculate high limit */
    if (ass->adcInternalData.rangeMulti)
      highlim = ass->adcInternalData.rangeLimits[ass->adcInternalData.rangeMax];
    else
      highlim = ass->adcInternalData.rangeLimits[0];

    /* Calculate low limit */
    if (ass->adcInternalData.approvalMode)
      lowlim = 0;
    else
      lowlim = -highlim;

    /* Check limits */
    if (cntPresetTare > highlim || cntPresetTare < lowlim)
    {
      result = ADC_OP_STATUS_FAILED_BAND;
    }
    else
    {
      result = ADC_OP_STATUS_SUCCESS;
    }
  }
  return result;
}


static void AdcOpPresetTare(AdcSuperStructure *ass)
{
  int32s presetTare;
  int32s cntPresetTare;
  int8u result;

  /* Convert PT to counts */
  presetTare = ass->adcOperation.paramL1;
  cntPresetTare = 16 * presetTare;

  /* Can the operation proceed */
  result = AdcIntPresetTareOperationIsValid(ass, presetTare);
  if (ADC_OP_STATUS_SUCCESS != result)
  {
    /* Failed */
    ass->adcOperation.status = ass->adcOperation.tickStatus = result;
  }
  else
  {
    /* Apply the preset tare */
    ass->adcRuntimeDatabase.cntTare = 0;
    ass->adcRuntimeDatabase.cntPTare = cntPresetTare;
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
}


static void AdcOpClearLin(AdcSuperStructure *ass)
{
  AdcIntClearDbLinear(ass, (int8u) ass->adcOperation.paramL1);
  ass->adcRuntimeDatabase.cntTare = 0;
  ass->adcRuntimeDatabase.netMode = FALSE;

  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  AdcIntReInitReadings(ass);
}

static void AdcOpAddLin(AdcSuperStructure *ass)
{
  int32s rdg, errRdg;
  int8u i, newPnt;

  switch (ass->adcOperation.counter)
  {
    case 1:

      // check value
      rdg = ass->adcInternalData.cntSysRdg - ass->adcRuntimeDatabase.cntZero;
      errRdg = (ass->adcOperation.paramL2 * 16) - rdg;
      if ((errRdg > ADC_DB_USER_LINCOR_MAXIMUM) || (errRdg < ADC_DB_USER_LINCOR_MINIMUM))
      {
        ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
        return;
      }

      //block readings to user
      ass->adcInternalData.userBlocked = TRUE;

      //initialise adc for large averaging
      AdcIntSetCalAveraging(ass);
      break;

    default:
      if (ass->adcOperation.counter == ass->adcOperation.averageLen)
      {
        ass->adcOperation.averageLen = ass->adcUserDatabase.average;

        rdg = ass->adcInternalData.cntSysRdg - ass->adcRuntimeDatabase.cntZero;
        errRdg = (ass->adcOperation.paramL2 * 16) - rdg;
        if ((errRdg > ADC_DB_USER_LINCOR_MAXIMUM) || (errRdg < ADC_DB_USER_LINCOR_MINIMUM))
        {
          ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
        }
        else
        {
          newPnt = AdcIntLinearisationGetLinPt(&ass->adcLinearisation, rdg);

          // check point is not same
          for (i=0;i<=ADC_MAX_NUM_LIN_PTS;i++)
          {
            if ((i != ass->adcOperation.paramL1) && (newPnt == ass->adcUserDatabase.linPts[i])) break;
          }
          if (i <= ADC_MAX_NUM_LIN_PTS)
          {
            ass->adcOperation.status = ADC_OP_STATUS_FAILED_PNT_DUPL;
          }
          else
          {
            ass->adcUserDatabase.linPts[ass->adcOperation.paramL1] = newPnt;
            ass->adcUserDatabase.linCor[ass->adcOperation.paramL1] = (int16s) errRdg;
            ass->adcRuntimeDatabase.cntTare = 0;
            ass->adcRuntimeDatabase.netMode = FALSE;

            ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
          }
        }
        AdcIntReInitReadings(ass);
      }
      break;
  }
}


static void AdcOpWaitForMotion(AdcSuperStructure *ass)
{
  if (ass->adcDataWorking.grossRdg.info.status.bit.motion)
  {
    ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_MOTION;
  }
  else
  {
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
}


static void AdcIntPopulateTraceableWeight(AdcSuperStructure *ass, const t_datetime *dt, bool valid)
{
  ass->adcTraceableWeight.printId = ass->adcRuntimeDatabase.printId;
  ass->adcTraceableWeight.rdg = ass->adcDataWorking.dispRdg;
  if (ass->adcDataWorking.dispRdg.info.status.bit.gross)
    ass->adcTraceableWeight.altRdg = ass->adcDataWorking.altGrossRdg;
  else
    ass->adcTraceableWeight.altRdg = ass->adcDataWorking.altNetRdg;
  ass->adcTraceableWeight.tertRdg = ass->adcDataWorking.tertDispRdg;
  ass->adcTraceableWeight.pieces = ass->adcDataWorking.dispPieceRdg.rdg;
  ass->adcTraceableWeight.pTare = (ass->adcRuntimeDatabase.cntPTare != 0);
  if (ass->adcTraceableWeight.pTare)
  {
    ass->adcTraceableWeight.tareWgt = AdcIntCountsToFinal(ass,
                                                          ass->adcRuntimeDatabase.cntPTare,
                                                          ass->adcDataWorking.grossRdg.info.status.bit.range,
                                                          ass->adcDataWorking.dispRdg.info.status.bit.hiRes,
                                                          ADC_UNIT_SWITCH_PRIMARY);
  }
  else
  {
    ass->adcTraceableWeight.tareWgt = AdcIntCountsToFinal(ass,
                                                          ass->adcRuntimeDatabase.cntTare,
                                                          ass->adcDataWorking.grossRdg.info.status.bit.range,
                                                          ass->adcDataWorking.dispRdg.info.status.bit.hiRes,
                                                          ADC_UNIT_SWITCH_PRIMARY);
  }
  ass->adcTraceableWeight.dt = *dt;
  ass->adcTraceableWeight.valid = valid;
}


/**
 * Tests if a get traceable reading operation would succeed, but does not perform the get traceable reading operation.
 *
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present
 *         ADC_OP_STATUS_FAILED_PRNID_MAX if print id beyond maximum,
 *         ADC_OP_STATUS_FAILED_HI_RES if NTEP && running in hi-res,
 *         ADC_OP_STATUS_FAILED_BAND if overloaded or underloaded, or trade scale and within the zero band,
 *         ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_SUCCESS if operation would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntGetTraceableRdgOperationIsValid(const AdcSuperStructure *ass, bool immediate)
{
  int8u result;

  if (ass->adcDataWorking.error)
  {
    result = ADC_OP_STATUS_FAILED_ERROR;
  }
  else if (ADC_APPROVAL_MODE_NTEP == ass->adcInternalData.approvalMode && ass->adcRuntimeDatabase.hiRes)
  {
    result = ADC_OP_STATUS_FAILED_HI_RES;
  }
  else if (ass->adcDataWorking.grossRdg.info.status.bit.oload || ass->adcDataWorking.grossRdg.info.status.bit.uload)
  {
    result = ADC_OP_STATUS_FAILED_BAND;
  }
  else if (ADC_APPROVAL_MODE_NONE != ass->adcInternalData.approvalMode &&
           ass->adcDataWorking.dispRdg.rdg <= ass->adcUserDatabase.zeroBand &&
           ass->adcDataWorking.dispRdg.rdg >= -ass->adcUserDatabase.zeroBand)
  {
    result = ADC_OP_STATUS_FAILED_BAND;
  }
  else if (ass->adcRuntimeDatabase.printId >= INT32U_MAX)
  {
    result = ADC_OP_STATUS_FAILED_PRNID_MAX;
  }
  else
  {
    result = AdcIntOpCheckWaitForMotion(ass, immediate);
  }

  return result;
}


static void AdcOpGetTraceableRdg(AdcSuperStructure *ass)
{
  int8u result;
  t_datetime dt;

  result = AdcIntGetTraceableRdgOperationIsValid(ass, ass->adcOperation.immediate);
  switch (result)
  {
    case ADC_OP_STATUS_FAILED_MOTION:
      ass->adcOperation.tickStatus = result;
      break;

    case ADC_OP_STATUS_SUCCESS:
      if (!AdcIntGetDateTime(ass, &dt))
      {
        ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_DATE_TIME;
      }
      else
      {
        ass->adcRuntimeDatabase.printId++;
        AdcIntPopulateTraceableWeight(ass, &dt, true);
        ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
      }
      break;

    default:
      ass->adcOperation.status = ass->adcOperation.tickStatus = result;
      break;
  }
}


static void AdcOpGetNonTraceableRdg(AdcSuperStructure *ass)
{
  t_datetime dt;

  if (ass->adcRuntimeDatabase.printId >= INT32U_MAX)
  {
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_PRNID_MAX;
  }
  else if (!AdcIntGetDateTime(ass, &dt))
  {
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_DATE_TIME;
  }
  else
  {
    ass->adcRuntimeDatabase.printId++;
    AdcIntPopulateTraceableWeight(ass, &dt, false);
    (void) AdcIntGetDateTime(ass, &ass->adcTraceableWeight.dt);
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
}


static void AdcOpCalFacZero(AdcSuperStructure *ass)
{
  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_NOT_IMP;
}


static void AdcOpCalFacSpan(AdcSuperStructure *ass)
{
  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_NOT_IMP;
}


static void AdcOpClearPeak(AdcSuperStructure *ass)
{
  ass->adcDataWorking.peakAbsGrossRdg.rdg = 0;
  if (ass->adcOperation.paramL1)
  {
    ass->adcDataWorking.peakMaxDispRdg.rdg  = INT32S_MIN;
    ass->adcDataWorking.peakMinDispRdg.rdg  = INT32S_MAX;
  }

  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
}


static void AdcOpInitialZero(AdcSuperStructure *ass)
{
  int32s rdg;

  // wait a couple of cycles for the adc readings to settle
  if (ass->adcOperation.counter < ass->adcMiscellaneous.adcRdgFreq / 10)
  {
    ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_MOTION;
    return;
  }

  if (ass->adcDataWorking.grossRdg.info.status.bit.motion)
  {
    ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_MOTION;
    return;
  }

  if (ass->adcDataWorking.error)
  {
    ass->adcOperation.tickStatus = ass->adcOperation.status = ADC_OP_STATUS_FAILED_ERROR;
    return;
  }

  if (ass->adcInternalData.approvalMode == ADC_APPROVAL_MODE_NTEP)
  {
    if ((ass->adcInternalData.cntLast < ass->adcInternalData.cntZeroRangeMin) ||
        (ass->adcInternalData.cntLast > ass->adcInternalData.cntZeroRangeMax))
    {
      ass->adcOperation.tickStatus = ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
      return;
    }
  }
  else
  {
    rdg = ass->adcInternalData.rangeLimits[ass->adcInternalData.rangeMax] / 10;
    if ((ass->adcInternalData.cntLast < -rdg) ||
        (ass->adcInternalData.cntLast > rdg))
    {
      ass->adcOperation.tickStatus = ass->adcOperation.status = ADC_OP_STATUS_FAILED_BAND;
      return;
    }
  }

  if (ass->adcInternalData.approvalMode == ADC_APPROVAL_MODE_NTEP)
  {
    ass->adcInternalData.cntInitZero = 0;
    ass->adcRuntimeDatabase.cntZero = ass->adcInternalData.cntLast;
  }
  else
  {
    ass->adcInternalData.cntInitZero = ass->adcInternalData.cntLast;
    ass->adcRuntimeDatabase.cntZero = 0;
  }
  ass->adcRuntimeDatabase.cntTare = 0;
  ass->adcRuntimeDatabase.cntPTare = 0;
  ass->adcRuntimeDatabase.netMode = FALSE;
  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
}


/**
 * Tests if a set counting operation would succeed, but does not perform the operation.
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present,
 *         ADC_OP_STATUS_FAILED_BAND if overloaded or underloaded,
 *         ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_PARAM if parameters (sampleNum or cntSampleWgt) are out of range,
 *
 *         ADC_OP_STATUS_SUCCESS if operation would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntSetCountingOperationIsValid(const AdcSuperStructure *ass, int32s sampleNum, int32s cntSampleWgt)
{
  int8u result;

  if (ass->adcDataWorking.error)
  {
    result = ADC_OP_STATUS_FAILED_ERROR;
  }
  else if (ass->adcDataWorking.grossRdg.info.status.bit.oload || ass->adcDataWorking.grossRdg.info.status.bit.uload)
  {
    result = ADC_OP_STATUS_FAILED_BAND;
  }
  else if (ass->adcDataWorking.dispRdg.info.status.bit.motion)
  {
    result = ADC_OP_STATUS_FAILED_MOTION;
  }
  else if (0 == sampleNum || 0 == cntSampleWgt)
  {
    result = ADC_OP_STATUS_FAILED_PARAM;
  }
  else
  {
    result = ADC_OP_STATUS_SUCCESS;
  }
  return result;
}


static void AdcOpSetCounting(AdcSuperStructure *ass)
{
  int32s sampleNum;
  int32s cntSampleWgt;
  int8u result;

  sampleNum = ass->adcOperation.paramL1;
  cntSampleWgt = ass->adcOperation.paramL2;

  result = AdcIntSetCountingOperationIsValid(ass, sampleNum, cntSampleWgt);
  if (ADC_OP_STATUS_SUCCESS == result)
  {
    ass->adcInternalData.sampleNum = sampleNum;
    ass->adcInternalData.cntSampleWgt = cntSampleWgt;
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
  else
  {
    ass->adcOperation.status = ass->adcOperation.tickStatus = result;
  }
}


static void AdcOpSetAlt(AdcSuperStructure *ass)
{
  ass->adcInternalData.altConv = (ass->adcOperation.paramL1) ? ass->adcOperation.paramL1 : 1;
  ass->adcInternalData.altUnits = (int8u) ass->adcOperation.paramL2;
  AdcIntGenerateAltRes(ass, ADC_UNIT_SWITCH_SECONDARY);
  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
}


static void AdcOpSetTert(AdcSuperStructure *ass)
{
  ass->adcInternalData.tertConv = (ass->adcOperation.paramL1) ? ass->adcOperation.paramL1 : 1;
  ass->adcInternalData.tertUnits = (int8u) ass->adcOperation.paramL2;
  AdcIntGenerateAltRes(ass, ADC_UNIT_SWITCH_TERTIARY);
  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
}


/**
 * Tests if a livestock weigh operation would succeed, but does not perform the operation.
 *
 * @return ADC_OP_STATUS_FAILED_TRADE if attempted in trade mode,
 *         ADC_OP_STATUS_FAILED_NOT_IMP if no live weigh implementer has been configured,
 *         ADC_OP_STATUS_SUCCESS if operation would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntLivestockWeighOperationIsValid(const AdcSuperStructure *ass)
{
  int8u result;

  if (ADC_APPROVAL_MODE_NONE != ass->adcUserDatabase.approvalMode)
  {
    /* Not available in trade modes */
    result = ADC_OP_STATUS_FAILED_TRADE;
  }
  else if (NULL == ass->adcLiveWeigh)
  {
    result = ADC_OP_STATUS_FAILED_NOT_IMP;
  }
  else
  {
    result = ADC_OP_STATUS_SUCCESS;
  }
  return result;
}


static void AdcOpDoLive(AdcSuperStructure *ass)
{
  int8u result;
  enum eAdcLiveWeighResult lwresult;

  result = AdcIntLivestockWeighOperationIsValid(ass);
  if (ADC_OP_STATUS_SUCCESS != result)
  {
    ass->adcOperation.status = ass->adcOperation.tickStatus = result;
  }
  else
  {
    lwresult = AdcIntLiveWeighRun(ass);
    switch (lwresult)
    {
      case alwrNotImplemented:
        ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_NOT_IMP;
        break;
      case alwrComplete:
        ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
        (void) AdcIntLiveWeighDisable(ass);
        break;
      case alwrDisabled:
        /* Kick it off */
        AdcIntLiveWeighInitiate(ass);
        break;
      case alwrRunning:
        /* Wait for it to complete */
        break;
      case alwrFailed:
      default:
        ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_FAILED_ERROR;
        break;
    }
  }
}


static void AdcOpSetDisplayMode(AdcSuperStructure *ass)
{
  ass->adcRuntimeDatabase.dispMode = (int8u) ass->adcOperation.paramL1;
  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
}


static void AdcOpBlockZT(AdcSuperStructure *ass)
{
  AdcIntZeroTrackingBlock(&ass->adcZeroTracking, (int8u) ass->adcOperation.paramL1);
  ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
}


static void AdcOpManualHold(AdcSuperStructure *ass)
{
  if (ADC_APPROVAL_MODE_NONE != ass->adcUserDatabase.approvalMode)
  {
    /* Not available in trade modes */
    ass->adcOperation.status = ADC_OP_STATUS_FAILED_TRADE;
  }
  else
  {
    ass->adcHeldData = ass->adcDataWorking;
    AdcIntAdcDataSetHeld(&ass->adcHeldData, true);
    ass->adcHeldData.dataValid = true;
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
}


static void AdcOpClearManualHold(AdcSuperStructure *ass)
{
  if (ADC_APPROVAL_MODE_NONE != ass->adcUserDatabase.approvalMode)
  {
    /* Not available in trade modes */
    ass->adcOperation.status = ADC_OP_STATUS_FAILED_TRADE;
  }
  else
  {
    AdcIntAdcDataSetHeld(&ass->adcHeldData, false);
    ass->adcHeldData.dataValid = false;
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
}


static void AdcOpWaitForAverage(AdcSuperStructure *ass)
{
  /* paramL1 is the number of samples still to wait */
  if (ass->adcOperation.paramL1 <= 0)
  {
    ass->adcOperation.status = ass->adcOperation.tickStatus = ADC_OP_STATUS_SUCCESS;
  }
  else
  {
    ass->adcOperation.paramL1--;
  }
}
