#ifndef ADC_OPERATIONS_DOT_H
#define ADC_OPERATIONS_DOT_H 1
                 
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcdatatypes.h>

// adc status   : see adcInterface.h
// adc commands : see adcInterface.h

// adc timeout
#define ADC_OP_TIMEOUT_IMMEDIATE      0
#define ADC_OP_TIMEOUT_NEVER          255

// calibration limits
#define ADC_SPAN_DIVS_MIN     100L                  // need at least 100 divisions of weight when performing span cal
#define ADC_MVV_PER_DIV_MIN   (((1<<8)*5)/100L)     // 0.005uV/V per DIV

/** Init operations (call only once) */
void AdcIntInitOperations(AdcSuperStructure *ass);


/** Destroy operations */
void AdcIntDestroyOperations(AdcSuperStructure *ass);


/** Re-initialise operations, can be called multiple times */
void AdcIntReInitOperations(AdcSuperStructure *ass);


/**
 * Start an ADC operation.
 *
 * @param ass is the ADC super structure
 * @param command is the command to execute. See ADC_OP_CMD_* constants
 * @param timeout is the timeout
 * @param block is whether the operation should block until complete/failure
 * @param pL1 is user parameter 1, command specific
 * @param pL2 is user parameter 2, command specific
 * @param immediate means the operation should complete immediately, without waiting for no motion. Generally only works in non-trade mode
 */
int8u AdcIntOperationStart(AdcSuperStructure *ass, int8u command, int8u timeout, int8u block, int32s pL1, int32s pL2, bool immediate);


/**
 * Must be called at the sync frequency to perform operations
 *
 * Note: any function which requires more than a single iteration must be done here
 */
void AdcIntOperationTick(AdcSuperStructure *ass);


/**
 * Should be called before UpdateReadings. Can be called faster than sync frequency.
 *
 * Note: anything which is processed here must succeed or fail in a single cycle.
 */
void AdcIntOperationTickBefore(AdcSuperStructure *ass);


/**
 * Tests if a semi auto zero operation would succeed, but does not perform semi auto zero.
 *
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_BAND if outside zero range, ADC_OP_STATUS_SUCCESS if zero would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntSemiautoZeroOperationIsValid(const AdcSuperStructure *ass, bool immediate);


/**
 * Tests if a tare operation would succeed, but does not perform tare.
 *
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_BAND if outside range of tare, ADC_OP_STATUS_SUCCESS if tare would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntSemiautoTareOperationIsValid(const AdcSuperStructure *ass, bool immediate);


/**
 * Tests if a preset tare operation would succeed, but does not perform preset tare.
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present, ADC_OP_STATUS_FAILED_BAND if outside range of tare,
 *         ADC_OP_STATUS_SUCCESS if tare would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntPresetTareOperationIsValid(const AdcSuperStructure *ass, int32s presetTare);


/**
 * Tests if a get traceable reading operation would succeed, but does not perform the get traceable reading operation.
 *
 * @param immediate if set will perform the operation without waiting for no motion. Only available in non-trade mode
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present
 *         ADC_OP_STATUS_FAILED_PRNID_MAX if print id beyond maximum,
 *         ADC_OP_STATUS_FAILED_HI_RES if NTEP && running in hi-res,
 *         ADC_OP_STATUS_FAILED_BAND if overloaded or underloaded, or trade scale and within the zero band,
 *         ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_SUCCESS if operation would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntGetTraceableRdgOperationIsValid(const AdcSuperStructure *ass, bool immediate);


/**
 * Tests if a livestock weigh operation would succeed, but does not perform the operation.
 *
 * @return ADC_OP_STATUS_FAILED_TRADE if attempted in trade mode,
 *         ADC_OP_STATUS_FAILED_NOT_IMP if no live weigh implementer has been configured,
 *         ADC_OP_STATUS_SUCCESS if operation would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntLivestockWeighOperationIsValid(const AdcSuperStructure *ass);


/**
 * Tests if a set counting operation would succeed, but does not perform the operation.
 *
 * @return ADC_OP_STATUS_FAILED_ERROR if errors present,
 *         ADC_OP_STATUS_FAILED_BAND if overloaded or underloaded,
 *         ADC_OP_STATUS_FAILED_MOTION if motion present,
 *         ADC_OP_STATUS_FAILED_PARAM if parameters (sampleNum or cntSampleWgt) are out of range,
 *
 *         ADC_OP_STATUS_SUCCESS if operation would succeed
 *
 * Note: Does not lock AdcSuperStructure
 */
int8u AdcIntSetCountingOperationIsValid(const AdcSuperStructure *ass, int32s sampleNum, int32s cntSampleWgt);


#endif /* #ifndef ADC_OPERATIONS_DOT_H */
