/*
 * adcpacked.c
 *
 * Deals with struct packing and endianness
 * for traceable weights so they take up a minimum of space
 * and can be written to a DSD in a consistent fashion
 */
#if defined(WIN32)
  #include <winsock.h>
#else
  #include <arpa/inet.h>
#endif

#include "adcpacked.h"


/** Pack and convert an adc status to network endian (suitable for writing to a DSD file) */
void AdcStatusPackToNetwork(const AdcStatus *as, AdcStatusPacked *asp)
{
  asp->multiRange = as->bit.multiRange;
  asp->preTare = as->bit.preTare;
  asp->tare = as->bit.tare;

  asp->oload = as->bit.oload;
  asp->uload = as->bit.uload;
  asp->gross = as->bit.gross;
  asp->coz = as->bit.coz;
  asp->motion = as->bit.motion;
  asp->zeroBand = as->bit.zeroBand;
  asp->zeroRange = as->bit.zeroRange;
  asp->held = as->bit.held;

  asp->range = as->bit.range;
  asp->hiRes = as->bit.hiRes;
  asp->grossOrNet = as->bit.grossOrNet;
  asp->gradPos = as->bit.gradPos;
}


/** Unpack and convert an adc status from network endian (suitable for reading from a DSD file) */
void AdcStatusUnpackFromNetwork(const AdcStatusPacked *asp, AdcStatus *as)
{
  as->bit.multiRange = asp->multiRange;
  as->bit.preTare = asp->preTare;
  as->bit.tare = asp->tare;

  as->bit.oload = asp->oload;
  as->bit.uload = asp->uload;
  as->bit.gross = asp->gross;
  as->bit.coz = asp->coz;
  as->bit.motion = asp->motion;
  as->bit.zeroBand = asp->zeroBand;
  as->bit.zeroRange = asp->zeroRange;
  as->bit.held = asp->held;

  as->bit.range = asp->range;
  as->bit.hiRes = asp->hiRes;
  as->bit.grossOrNet = asp->grossOrNet;
  as->bit.gradPos = asp->gradPos;
}


/** Pack and convert an adc reading to network endian (suitable for writing to a DSD file) */
void AdcRdgPackToNetwork(const AdcRdg *rdg, AdcRdgPacked *rdgp)
{
  rdgp->rdg = htonl(rdg->rdg);
  rdgp->info.id = rdg->info.id;
  rdgp->info.unitdp.units = rdg->info.unitdp.bit.units;
  rdgp->info.unitdp.dp = rdg->info.unitdp.bit.dp;
  AdcStatusPackToNetwork(&rdg->info.status, &rdgp->info.status);
}


/** Unpack and convert an adc reading from network endian (suitable for reading from a DSD file) */
void AdcRdgUnpackFromNetwork(const AdcRdgPacked *rdgp, AdcRdg *rdg)
{
  rdg->rdg = ntohl(rdgp->rdg);
  rdg->info.id = rdgp->info.id;
  rdg->info.unitdp.bit.units = rdgp->info.unitdp.units;
  rdg->info.unitdp.bit.dp = rdgp->info.unitdp.dp;
  AdcStatusUnpackFromNetwork(&rdgp->info.status, &rdg->info.status);
}


/** Pack and convert a traceable weight to network endian (suitable for writing to a DSD file) */
void AdcTraceablePackToNetwork(const AdcTraceableWeight *atw, AdcTraceableWeightPacked *atwp)
{
  atwp->printId = htonl(atw->printId);
  AdcRdgPackToNetwork(&atw->rdg, &atwp->rdg);
  AdcRdgPackToNetwork(&atw->altRdg, &atwp->altRdg);
  AdcRdgPackToNetwork(&atw->tertRdg, &atwp->tertRdg);
  atwp->pieces = htonl(atw->pieces);
  atwp->tareWgt = htonl(atw->tareWgt);
  atwp->pTare = atw->pTare;
  atwp->valid = atw->valid;
  atwp->dt.second = atw->dt.second;
  atwp->dt.minute = atw->dt.minute;
  atwp->dt.hour = atw->dt.hour;
  atwp->dt.day = atw->dt.day;
  atwp->dt.date = atw->dt.date;
  atwp->dt.month = atw->dt.month;
  atwp->dt.year = htons(atw->dt.year);
}


/** Pack and convert a traceable weight from network endian (suitable for reading from a DSD file) */
void AdcTraceableUnpackFromNetwork(const AdcTraceableWeightPacked *atwp, AdcTraceableWeight *atw)
{
  atw->printId = ntohl(atwp->printId);
  AdcRdgUnpackFromNetwork(&atwp->rdg, &atw->rdg);
  AdcRdgUnpackFromNetwork(&atwp->altRdg, &atw->altRdg);
  AdcRdgUnpackFromNetwork(&atwp->tertRdg, &atw->tertRdg);
  atw->pieces = ntohl(atwp->pieces);
  atw->tareWgt = ntohl(atwp->tareWgt);
  atw->pTare = atwp->pTare;
  atw->valid = atwp->valid;
  atw->dt.second = atwp->dt.second;
  atw->dt.minute = atwp->dt.minute;
  atw->dt.hour = atwp->dt.hour;
  atw->dt.day = atwp->dt.day;
  atw->dt.date = atwp->dt.date;
  atw->dt.month = atwp->dt.month;
  atw->dt.year = ntohs(atwp->dt.year);
}
