/*
 * adcpacked.h
 *
 *  Created on: Mar 17, 2012
 *      Author: ian
 */

#ifndef ADCPACKED_H_
#define ADCPACKED_H_

#include <rinhw/stdrinstrum.h>
#include <rinhw/rinendian.h>
#include <rinweigh/adcdatatypes.h>


/** Byte sized boolean */
typedef int8u int8u_bool;


#pragma pack(push)
#pragma pack(1)


#if (BYTE_ORDER == BIG_ENDIAN)
  /** Packed type containing ADC reading status */
  typedef struct
  {
    /* Byte 0 */
    int8s gradPos   :4;         /*!< The position within a graduation? */
    int8u grossOrNet:1;         /*!< If set, the reading is either gross or net */
    int8u hiRes     :1;         /*!< If set, the reading has extra resolution (ie *10) */
    int8u range     :2;         /*!< The range of the reading (0=range1, 1=range 2, etc */

    /* Byte 1 */
    int8u held      :1;         /*!< If set, the reading is held */
    int8u zeroRange :1;         /*!< If set, the reading is within the zero range (trade: zero key operates in this range) */
    int8u zeroBand  :1;         /*!< If set, the reading is within the zero band (non-trade: band around zero used for printing/setpoints, etc) */
    int8u motion    :1;         /*!< If set, the reading is unstable */
    int8u coz       :1;         /*!< If set, the reading is with center of zero */
    int8u gross     :1;         /*!< If set, the reading is a gross value, net otherwise */
    int8u uload     :1;         /*!< If set, the reading is overloaded */
    int8u oload     :1;         /*!< If set, the reading is underloaded */

    /* Byte 2 */
    int8u tare      :1;         /*!< If set, the reading is a tare weight */
    int8u preTare   :1;         /*!< If set, the reading is a preset tare weight */
    int8u multiRange:1;         /*!< If set, the reading is either multi range or multi interval (ie scale is not single range) */
    int8u unused1   :5;

    /* Byte 3 */
    int8u unused2   :8;
  } AdcStatusPacked;

#elif (BYTE_ORDER == LITTLE_ENDIAN)
  /** Packed type containing ADC reading status */
  typedef struct
  {
    /* Byte 3 */
    int8u range     :2;         /*!< The range of the reading (0=range1, 1=range 2, etc */
    int8u hiRes     :1;         /*!< If set, the reading has extra resolution (ie *10) */
    int8u grossOrNet:1;         /*!< If set, the reading is either gross or net */
    int8s gradPos   :4;         /*!< The position within a graduation? */

    /* Byte 2 */
    int8u oload     :1;         /*!< If set, the reading is underloaded */
    int8u uload     :1;         /*!< If set, the reading is overloaded */
    int8u gross     :1;         /*!< If set, the reading is a gross value, net otherwise */
    int8u coz       :1;         /*!< If set, the reading is with center of zero */
    int8u motion    :1;         /*!< If set, the reading is unstable */
    int8u zeroBand  :1;         /*!< If set, the reading is within the zero band (non-trade: band around zero used for printing/setpoints, etc) */
    int8u zeroRange :1;         /*!< If set, the reading is within the zero range (trade: zero key operates in this range) */
    int8u held      :1;         /*!< If set, the reading is held */

    /* Byte 1 */
    int8u unused1   :5;
    int8u multiRange:1;         /*!< If set, the reading is either multi range or multi interval (ie scale is not single range) */
    int8u preTare   :1;         /*!< If set, the reading is a preset tare weight */
    int8u tare      :1;         /*!< If set, the reading is a tare weight */

    /* Byte 0 */
    int8u unused2   :8;
  } AdcStatusPacked;

#endif


#if (BYTE_ORDER == BIG_ENDIAN)
  /** Packed type containing ADC reading units and decimal point position */
  typedef struct
  {
    int8u units :4;             /*!< The units of the reading (eg g, kg, lb) */
    int8u dp    :4;             /*!< The decimal point position of the reading */
  } AdcUnitDpPacked;
#elif (BYTE_ORDER == LITTLE_ENDIAN)
  /** Packed type containing ADC reading units and decimal point position */
  typedef struct
  {
    int8u dp    :4;             /*!< The decimal point position of the reading */
    int8u units :4;             /*!< The units of the reading (eg g, kg, lb) */
  } AdcUnitDpPacked;
#endif


/** Packed type containing ADC reading status, units and decimal point and a reading ID */
typedef struct
{
  int8u id;                     /*!< Reading identifier */
  AdcUnitDpPacked unitdp;       /*!< Reading units and decimal point */
  AdcStatusPacked status;       /*!< Reading status */
} AdcInfoPacked;


/** Packed type containing a reading and associated information */
typedef struct
{
  int32s rdg;                   /*!< Reading value */
  AdcInfoPacked info;           /*!< Reading information (status, units, etc) */
} AdcRdgPacked;


/** Packed date time structure */
typedef struct
{
  int8u second;           /*!< Seconds 0..59 */
  int8u minute;           /*!< Minutes 0..59 */
  int8u hour;             /*!< Hours 0..23 */
  int8u day;              /*!< Day of week 1..7 */
  int8u date;             /*!< Day of month 1..31 */
  int8u month;            /*!< Month 1..12 */
  int16u year;            /*!< Year 2000+ */
} AdcDateTimePacked;


/** Packed traceable weight structure */
typedef struct
{
  int32u printId;               /*!< The unique identifier for the traceable weight */
  AdcRdgPacked rdg;             /*!< The traceable weight reading */
  AdcRdgPacked altRdg;          /*!< The traceable weight reading in alternate units */
  int32s pieces;                /*!< The traceable weight reading as pieces */
  int32s tareWgt;               /*!< The tare weight set for the traceable weight reading  */
  int8u_bool pTare;             /*!< Flag if the tare is a preset tare */
  int8u_bool valid;             /*!< Flag if the traceable weight is valid */
  AdcDateTimePacked dt;         /*!< The timestamp of the traceable weight */
  AdcRdgPacked tertRdg;         /*!< The traceable weight reading in tertiary units */
} AdcTraceableWeightPacked;

#pragma pack(pop)


/** Pack and convert an adc status to network endian (suitable for writing to a DSD file) */
void AdcStatusPackToNetwork(const AdcStatus *as, AdcStatusPacked *asp);


/** Unpack and convert an adc status from network endian (suitable for reading from a DSD file) */
void AdcStatusUnpackFromNetwork(const AdcStatusPacked *asp, AdcStatus *as);


/** Pack and convert an adc reading to network endian (suitable for writing to a DSD file) */
void AdcRdgPackToNetwork(const AdcRdg *rdg, AdcRdgPacked *rdgp);


/** Unpack and convert an adc reading from network endian (suitable for reading from a DSD file) */
void AdcRdgUnpackFromNetwork(const AdcRdgPacked *rdgp, AdcRdg *rdg);


/** Pack and convert a traceable weight to network endian (suitable for writing to a DSD file) */
void AdcTraceablePackToNetwork(const AdcTraceableWeight *atw, AdcTraceableWeightPacked *atwp);


/** Pack and convert a traceable weight to network endian (suitable for writing to a DSD file) */
void AdcTraceableUnpackFromNetwork(const AdcTraceableWeightPacked *atwp, AdcTraceableWeight *atw);


#endif /* ADCPACKED_H_ */
