#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <rinhw/rstdio.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcdatabase.h>
#include <rinweigh/adcaverage.h>
#include <rinweigh/adcmotion.h>
#include <rinweigh/adclin.h>
#include <rinweigh/adcconvert.h>
#include <rinweigh/adcads1216.h>
#include <rinweigh/adczerotrack.h>
#include <rinweigh/adcoperations.h>
#include <rinweigh/adcdlong.h>
#include <rinweigh/adcdebug.h>
#include <rinweigh/adctemperature.h>
#include <rinweigh/adcdatafiller.h>

#include <rinweigh/adcreadings.h>


/** Calculate which range a number of counts would appear in */
int8u AdcIntGetRangeForCounts(const AdcSuperStructure *ass, int32s cnt)
{
  int8u i;

  cnt = labs(cnt);
  for (i = 0; i < ass->adcInternalData.rangeMax; i++)
  {
    if (cnt <= ass->adcInternalData.rangeLimits[i])
      break;
  }
  return i;
}


/** Calculate number of samples @ the current sampling frequency for required averaging (in 0.01s increments), as configured in ass->adcUserDatabase.average */
int32u AdcIntCalculateAverageNumSamples(const AdcSuperStructure *ass)
{
  int32u avg_num_rdg;

  /* Number of readings to average = (db_avg[*0.01s] * sync_freq[*0.1Hz]) / (100 * 10) */
  avg_num_rdg = ass->adcUserDatabase.average * ass->adcMiscellaneous.adcRdgFreq;
  avg_num_rdg /= ADC_AV_1_SEC * 10;
  return avg_num_rdg;
}


/** Initialise the readings engine. Can be called more than once. */
void AdcIntReInitReadings(AdcSuperStructure *ass)
{
  int8u i;
  int32s templ;
  int16u motion_delta;
  int8u motion_window;
  int8u motion_on_samples;
  int16u sync_freq;
  int16u old_freq;
  int32u avg_num_rdg;
  int32s zero;
  int32s span;
  AdcRdg peak;
  int32u zt_window;
  int32u zt_delta;

#if ADC_DEBUG(ADC_INITIALISATION_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcInitReadings called\n");
#endif

  ass->adcMiscellaneous.updateBlocked = TRUE;

  ass->adcRuntimeDatabase.tempCalOn = TRUE;

  (void) AdcIntUpdateTemperature(ass);

  /*
   * Setup ADC chip
   */

  /* Zero and span cal */
  zero = AdcIntTempComp_CalculateZero(ass);
  span = AdcIntTempComp_CalculateSpan(ass);

  /* Initialise zero mux averaging */
  AdcIntInitAverage(&ass->adcZeroMuxAveraging, ADC_ZERO_MUX_AVERAGE_RDGS);

  /* Setup sync frequency */
  sync_freq = (ass->adcUserDatabase.syncFreqIdx >= ADC_NUM_SYNC_FREQ) ?
               AdcSyncFrequencies[ADC_DB_USER_SYNCFREQIDX_DEFAULT] :
               AdcSyncFrequencies[ass->adcUserDatabase.syncFreqIdx];
  if (sync_freq < ADC_SYNC_FREQ_LOW_LIM || sync_freq > ADC_SYNC_FREQ_HIGH_LIM)
  {
    sync_freq = AdcSyncFrequencies[ADC_DB_USER_SYNCFREQIDX_DEFAULT];
  }

  old_freq = ass->adcMiscellaneous.adcRdgFreq;
  ass->adcMiscellaneous.adcRdgFreq = sync_freq;
  AdcIntHwReconfigure(ass, zero, span, sync_freq, ass->adcUserDatabase.wire4Mode);

  /* Change of sync frequency means an initial zero */
  if (old_freq != sync_freq)
  {
#if ADC_DEBUG(ADC_INITIALISATION_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcInitReadings: Sync freq change from %u to %u, force initial zero cal.\n",
                     old_freq, sync_freq);
#endif
    AdcIntStartZeroCalCycle(&ass->adcTempCal);
  }

  /*
   * Initialise main averaging:
   * Number of readings to average = (db_avg[*0.01s] * sync_freq[*0.1Hz]) / (100 * 10)
   */
  avg_num_rdg = AdcIntCalculateAverageNumSamples(ass);
#if ADC_DEBUG(ADC_AVERAGE_INIT_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcInitReadings: Averaging=%u(*0.01s), gives %lu samples @ freq %u(0.1Hz)\n",
                   ass->adcUserDatabase.average, avg_num_rdg, ass->adcMiscellaneous.adcRdgFreq);
#endif
  AdcIntInitAverage(&ass->adcAveraging, avg_num_rdg);

  // initialise jitter averaging
  templ = (((int32s) ass->adcUserDatabase.average) * (30 * ass->adcMiscellaneous.adcRdgFreq / 10) / ADC_AV_1_SEC); // calc length of secondary average.  Post average in WE2110 is 30x longer than standard average (although this seems extremely long).
  if (templ > ADC_AVERAGE_MAX_RDGS)
    ass->adcMiscellaneous.jitterAvLength = ADC_AVERAGE_MAX_RDGS;  // just limit it to the maximum buffer length
  else
    ass->adcMiscellaneous.jitterAvLength = templ;
  AdcIntInitAverage(&ass->adcJitterAveraging, ass->adcMiscellaneous.jitterAvLength);

  /* Prevent divide by zero */
  assert(ass->adcUserDatabase.spanCnt != 0);
  templ = 16 * ass->adcUserDatabase.spanRaw / ass->adcUserDatabase.spanCnt; // calculate 1 division in raw
  if (ass->adcUserDatabase.jitter == ADC_JITTER_COARSE)
    ass->adcMiscellaneous.jitterAvMaxStep = 5 * templ; // 5 grads - this seems quite large (1.5d might be better?)
  else
    ass->adcMiscellaneous.jitterAvMaxStep = 2 * templ; // 2 grads - this seems quite large (0.5d might be better?)

  // Initialise multiple ranging
  switch (ass->adcUserDatabase.scaleType)
  {
    case ADC_SCALE_TYPE_DUAL_INTERVAL :
      ass->adcInternalData.rangeMax = 1;
      ass->adcInternalData.rangeMulti = FALSE;
      break;
    case ADC_SCALE_TYPE_DUAL_RANGE :
      ass->adcInternalData.rangeMax = 1;
      ass->adcInternalData.rangeMulti = TRUE;
      break;
    case ADC_SCALE_TYPE_TRIPLE_INTERVAL :
      ass->adcInternalData.rangeMax = 2;
      ass->adcInternalData.rangeMulti = FALSE;
      break;
    case ADC_SCALE_TYPE_TRIPLE_RANGE :
      ass->adcInternalData.rangeMax = 2;
      ass->adcInternalData.rangeMulti = TRUE;
      break;
    default : //single range
      ass->adcInternalData.rangeMax = 0;
      ass->adcInternalData.rangeMulti = FALSE;
      break;
  }

  for (i = 0; i <= ass->adcInternalData.rangeMax; i++)
    ass->adcInternalData.rangeLimits[i] = ass->adcUserDatabase.fullScale[i] * 16;

  /* Initialise motion, if enabled */
  if (ADC_MOTION_DETECT_OFF != ass->adcUserDatabase.motion)
  {

    /*
     * Work out number of samples for motion on.
     * Note: motion on time = max(average time, motion window)
     */
    if (ass->adcUserDatabase.average > (ADC_AV_1_SEC * INT8U_MAX / (ass->adcMiscellaneous.adcRdgFreq / 10)))
    {
      motion_on_samples = INT8U_MAX;
    }
    else
    {

      /* Prevent divide by zero */
      assert( AdcMotionWindowInv[ass->adcUserDatabase.motion] != 0 );

      /* Work out motion on time as a number of ADC samples */
      motion_on_samples = (int8u) (ass->adcMiscellaneous.adcRdgFreq / 10 / AdcMotionWindowInv[ass->adcUserDatabase.motion]);
      if (motion_on_samples < (ass->adcUserDatabase.average * ass->adcMiscellaneous.adcRdgFreq / 10) / ADC_AV_1_SEC)
      {
        motion_on_samples = (int8u) ((ass->adcUserDatabase.average * ass->adcMiscellaneous.adcRdgFreq / 10) / ADC_AV_1_SEC);
      }
    }

    /* Work out motion delta (movement required) */
    motion_delta = AdcResolutions[ass->adcUserDatabase.resolution[0]] * AdcMotionDelta[ass->adcUserDatabase.motion];

    /* Work out motion window (time (samples) required) */
    assert( AdcMotionWindowInv[ass->adcUserDatabase.motion] != 0 );
    motion_window = ass->adcMiscellaneous.adcRdgFreq / 10 / AdcMotionWindowInv[ass->adcUserDatabase.motion];

    /* Initialise motion detection */
    AdcIntMotionDetectionInit(&ass->adcMotionDetection, motion_delta, motion_window, motion_on_samples);
  }
  else
  {
    /* Motion detection disabled */
    AdcIntMotionDetectionDisable(&ass->adcMotionDetection);
  }

  // initialise linearity
  AdcIntLinearisationInit(&ass->adcUserDatabase, &ass->adcLinearisation, ass->adcInternalData.rangeLimits[ass->adcInternalData.rangeMax]);

  // initialise compensation data
  ass->adcInternalData.rawAdd = 0;
  ass->adcInternalData.rawMult = ADC_RAW_MULT_UNITY;

  // get trade mode
  ass->adcInternalData.approvalMode = ass->adcUserDatabase.approvalMode;

  // zero range
  ass->adcInternalData.cntZeroRangeMin = ass->adcUserDatabase.fullScale[ass->adcInternalData.rangeMax] * 16L *
    AdcZeroRangeNeg[ass->adcUserDatabase.rangeOfZero] / 100L;
  ass->adcInternalData.cntZeroRangeMax = ass->adcUserDatabase.fullScale[ass->adcInternalData.rangeMax] * 16L *
    AdcZeroRangePos[ass->adcUserDatabase.rangeOfZero] / 100L;

  // zero band
  ass->adcInternalData.cntZeroBandMax = (ass->adcUserDatabase.zeroBand * 16L) + 7;

  // centre of zero
  ass->adcInternalData.cntCOZ = 4L * AdcResolutions[ass->adcUserDatabase.resolution[0]];  // +-0.25 divisions

  // overload and underload
  switch (ass->adcInternalData.approvalMode)
  {
    case ADC_APPROVAL_MODE_NTEP:
      ass->adcInternalData.cntOload = ass->adcUserDatabase.fullScale[ass->adcInternalData.rangeMax] * 16L * 105L / 100L; // 105% fs
      ass->adcInternalData.cntUload = ass->adcInternalData.cntZeroRangeMin;
      break;
    case ADC_APPROVAL_MODE_NONE:
      ass->adcInternalData.cntOload = ass->adcUserDatabase.fullScale[ass->adcInternalData.rangeMax] * 16L * 105L / 100L; // 105% fs
      ass->adcInternalData.cntUload = -ass->adcInternalData.cntOload;
      break;
    default: // ADC_MODE_OIML
      ass->adcInternalData.cntOload = (ass->adcUserDatabase.fullScale[ass->adcInternalData.rangeMax] +
                          (9L * AdcResolutions[ass->adcUserDatabase.resolution[ass->adcInternalData.rangeMax]])) * 16L;
      ass->adcInternalData.cntUload = -320L * AdcResolutions[ass->adcUserDatabase.resolution[0]];  // -20 divisions
      break;
  }

  /* Overload counter trigger point - 135% of fullscale + additive tare */
  ass->adcMiscellaneous.cntOverloadCntLevel = (AdcIntFinalToCounts(ass, ass->adcUserDatabase.fullScale[ass->adcInternalData.rangeMax], false, ADC_UNIT_SWITCH_PRIMARY) * 135L) / 100L +
                                              AdcIntFinalToCounts(ass, ass->adcUserDatabase.addTare, false, ADC_UNIT_SWITCH_PRIMARY);

  // setup default reading information
  ass->adcRuntimeDatabase.hiRes = ass->adcUserDatabase.hiRes;
  ass->adcMiscellaneous.adcInfo.id = 0;
  AdcIntFillOutUnitDp(ass, &ass->adcMiscellaneous.adcInfo.unitdp, ADC_UNIT_SWITCH_PRIMARY);

  /* Fill out status */
  memset(&ass->adcMiscellaneous.adcInfo.status, 0, sizeof(ass->adcMiscellaneous.adcInfo.status));
  ass->adcMiscellaneous.adcInfo.status.bit.gradPos  = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.grossOrNet = 1;
  ass->adcMiscellaneous.adcInfo.status.bit.hiRes    = ass->adcRuntimeDatabase.hiRes;
  ass->adcMiscellaneous.adcInfo.status.bit.range    = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.held     = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.zeroRange = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.zeroBand = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.motion   = 1;
  ass->adcMiscellaneous.adcInfo.status.bit.coz      = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.gross    = 1;
  ass->adcMiscellaneous.adcInfo.status.bit.uload    = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.oload    = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.tare     = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.preTare  = 0;
  ass->adcMiscellaneous.adcInfo.status.bit.multiRange = 0;

  ass->adcDataWorking.rdgId        = 0;
  AdcIntFillOutTareReading(ass, &ass->adcDataWorking.presetTare, 0, false, ADC_UNIT_SWITCH_PRIMARY);

  /* Initialise peak hold values */
  memset(&peak, 0, sizeof(peak));
  peak.rdg = 0;
  peak.info = ass->adcMiscellaneous.adcInfo;
  peak.info.status.bit.gross = 1;
  peak.info.status.bit.grossOrNet = 1;
  AdcIntFillOutHeldReading(&ass->adcDataWorking.peakAbsGrossRdg, &peak);

  memset(&peak, 0, sizeof(peak));
  peak.rdg = INT32S_MIN;
  peak.info = ass->adcMiscellaneous.adcInfo;
  AdcIntFillOutHeldReading(&ass->adcDataWorking.peakMaxDispRdg, &peak);

  memset(&peak, 0, sizeof(peak));
  peak.rdg = INT32S_MAX;
  peak.info = ass->adcMiscellaneous.adcInfo;
  AdcIntFillOutHeldReading(&ass->adcDataWorking.peakMinDispRdg, &peak);

  /* Initialise zero tracking */
  if (ADC_ZERO_TRACKING_OFF != ass->adcUserDatabase.zeroTracking)
  {
    /* Zero tracking enabled */

    /* Divide by zero bad */
    assert(0 != AdcZeroTrackingWindowInv[ass->adcUserDatabase.zeroTracking]);

    zt_delta = AdcZeroTrackingDelta[ass->adcUserDatabase.zeroTracking] * AdcResolutions[ass->adcUserDatabase.resolution[0]];
    zt_window = ass->adcMiscellaneous.adcRdgFreq / 10 / AdcZeroTrackingWindowInv[ass->adcUserDatabase.zeroTracking];

#if ADC_DEBUG(ADC_ZERO_TRACKING_INIT_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcInitReadings: Zero track setting = %u, Zero track delta = %lu/16 divs = %lu counts @ resolution %u\n",
                     (unsigned) ass->adcUserDatabase.zeroTracking,
                     (unsigned long) AdcZeroTrackingDelta[ass->adcUserDatabase.zeroTracking],
                     (unsigned long) zt_delta,
                     (unsigned) AdcResolutions[ass->adcUserDatabase.resolution[0]]);
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcInitReadings: Zero track window = %u/sec = %lu samples @ freq %u(0.1Hz)\n",
                     (unsigned) AdcZeroTrackingWindowInv[ass->adcUserDatabase.zeroTracking],
                     (unsigned long) zt_window,
                     (unsigned) ass->adcMiscellaneous.adcRdgFreq);
#endif
    AdcIntZeroTrackingInit(&ass->adcZeroTracking, zt_delta, zt_window, false);
  }
  else
  {
    /* Zero tracking disabled */
#if ADC_DEBUG(ADC_ZERO_TRACKING_INIT_DEBUG_LVL, ADC_DEBUG_BASIC)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcInitReadings: Zero track disabled\n");
#endif
    AdcIntZeroTrackingInit(&ass->adcZeroTracking, 0, 0, true);
  }

  /*
   * Held data
   */
  memset(&ass->adcHeldData, 0, sizeof(AdcData));
  ass->adcHeldData.dataValid = false;

  ass->adcInternalData.userBlocked = FALSE;
  ass->adcMiscellaneous.updateBlocked = FALSE;

#if ADC_DEBUG(ADC_INITIALISATION_DEBUG_LVL, ADC_DEBUG_BASIC)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcInitReadings finished\n");
#endif
}


/**
 * Update readings. Will rate limit to the sync frequency. Does the hard work.
 *
 * @param ass is a pointer to the ADC super structure containing all information on the ADC, not NULL
 * @param data_wait_failed is a pointer to a bool that is set if there was a read problem,
 *        and you should implement a delay elsewhere, to prevent the ADC thread going to 100% CPU, not NULL
 * @return true if new data acquired (not a skipped reading, or blocked)
 */
bool AdcIntUpdateReadings(AdcSuperStructure *ass, bool *data_wait_failed)
{
  int32s rawRdg;             /* Number straight off ADC */
  int32s rawAvgRdg;          /* Averaged numbers off ADC */
  int32s rawAvgTiltRdg;      /* Averaged and tilt compensated numbers off ADC */
  int32s rawAvgTiltRdgDelta; /* Difference between last time and this time */
  int32s rawAvgTiltJitRdg;   /* Averaged, tilt compensated, jitter corrected numbers off ADC */
  int32s cntRdg;             /* rawAvgTiltJitRdg converted to counts */
  int32s cntZeroedRdg;       /* Reading in counts with initial zero and user zero subtracted */
  int8u previousRange;

#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings called\n");
#endif

  /* Don't run if updates blocked TODO need to delay here, or do post AdcIntHwPoll call */
  if (ass->adcMiscellaneous.updateBlocked)
  {
#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings: updates blocked, returning\n");
#endif
    *data_wait_failed = true;
    return FALSE;
  }

  /* Get ADC reading from hardware */
#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings: get reading from ADC\n");
#endif
  if (!AdcIntHwRead(ass, &rawRdg, data_wait_failed))
  {
#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
    ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings: AdcIntHwPoll() failed\n");
#endif
    /* Because everything below won't run, propagate the errors into the adcWorkingData */
    ass->adcDataWorking.error = (ass->adcMiscellaneous.adcErrors & ADC_ERR_MASK);
    return FALSE;
  }

#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings: got reading %ld, now averaging\n", (long) rawRdg);
#endif

  /* Get time when the reading was taken */
  if (clock_gettime(CLOCK_MONOTONIC, &ass->adcInternalData.tsReading) != 0)
  {
    memset(&ass->adcInternalData.tsReading, 0, sizeof(struct timespec));
  }

  /* Run main averaging */
  rawAvgRdg = AdcIntAverageReading(&ass->adcAveraging, rawRdg);
#if ADC_DEBUG(ADC_AVERAGE_READING_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings: rdg preav: %ld, postav: %ld\n", (long) rawRdg, (long) rawAvgRdg);
  AdcIntDebugDumpAverage(&ass->adcAveraging);
#endif

#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings: averaging complete, rdg %ld\n", (long) rawAvgRdg);
#endif

  /* Work out if ADC maxed on non-averaged values */
  if (rawRdg >= ass->adcHwInfo.adcMax || rawRdg <= -ass->adcHwInfo.adcMax)
    ass->adcMiscellaneous.adcErrors |= ADC_ERR_ADC_MAXED;
  else if (!ass->adcUserDatabase.wire4Mode)
    ass->adcMiscellaneous.adcErrors &= (int16u) ~ADC_ERR_ADC_MAXED;

  /* Do tilt compensation */
  rawAvgTiltRdg = rawAvgRdg;
  if (!ass->adcInternalData.userBlocked)  // do not do tilt compensation while calibrating
  {
    rawAvgTiltRdg += ass->adcInternalData.rawAdd;
    if (ass->adcInternalData.rawMult != ADC_RAW_MULT_UNITY)
      rawAvgTiltRdg = AdcMultDiv(rawAvgTiltRdg, ADC_RAW_MULT_UNITY, ass->adcInternalData.rawMult);
  }

  /* Apply jitter averaging */
  rawAvgTiltRdgDelta = ass->adcMiscellaneous.lastRdg - rawAvgTiltRdg;
  if (AdcIntMotionDetectionGetMotion(&ass->adcMotionDetection) ||
      rawAvgTiltRdgDelta > ass->adcMiscellaneous.jitterAvMaxStep ||
      rawAvgTiltRdgDelta < -ass->adcMiscellaneous.jitterAvMaxStep)
  {
    AdcIntInitAverage(&ass->adcJitterAveraging, ass->adcMiscellaneous.jitterAvLength);
  }
  ass->adcMiscellaneous.lastRdg = rawAvgTiltRdg;
  if (ass->adcUserDatabase.jitter != ADC_JITTER_OFF)
  {
    rawAvgTiltJitRdg = AdcIntAverageReading(&ass->adcJitterAveraging, rawAvgTiltRdg);
#if ADC_DEBUG(ADC_AVERAGE_JITTER_DEBUG_LVL, ADC_DEBUG_FULL)
    AdcIntDebugDumpAverage(&ass->adcJitterAveraging);
#endif
  }
  else
  {
    rawAvgTiltJitRdg = rawAvgTiltRdg;
  }

  /* Convert to counts */
  cntRdg = AdcMultDiv(ass->adcUserDatabase.spanCnt, rawAvgTiltJitRdg - ass->adcUserDatabase.zeroRaw, ass->adcUserDatabase.spanRaw);

  /* Calculate motion */
#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings: update motion\n");
#endif
  AdcIntMotionDetectionUpdate(&ass->adcMotionDetection, cntRdg - ass->adcInternalData.cntLast);
  ass->adcInternalData.cntLast = cntRdg;

  /* Calculate readings in counts */
  ass->adcInternalData.cntSysRdg = cntRdg - ass->adcInternalData.cntInitZero;
  cntZeroedRdg = ass->adcInternalData.cntSysRdg - ass->adcRuntimeDatabase.cntZero;
  ass->adcInternalData.cntGross = cntZeroedRdg + AdcIntLinearisationCorrection(&ass->adcLinearisation, cntZeroedRdg);
  ass->adcInternalData.cntNet = ass->adcInternalData.cntGross - ass->adcRuntimeDatabase.cntTare - ass->adcRuntimeDatabase.cntPTare;

  /* If NTEP then recalc overload limit */
  if (ADC_APPROVAL_MODE_NTEP == ass->adcInternalData.approvalMode)
  {
    ass->adcInternalData.cntOload = (ass->adcUserDatabase.fullScale[ass->adcInternalData.rangeMax] * 16L * 105L / 100L) - ass->adcRuntimeDatabase.cntZero;
  }

  /* Fill out AdcData structure */
  previousRange = ass->adcMiscellaneous.adcInfo.status.bit.range;
  AdcIntFillOutAdcData(ass,
                       &ass->adcDataWorking,
                       &ass->adcMiscellaneous.adcInfo,
                       rawAvgTiltJitRdg,
                       ass->adcInternalData.cntSysRdg,
                       ass->adcInternalData.cntGross,
                       ass->adcInternalData.cntNet,
                       ass->adcInternalData.cntOload,
                       previousRange);

  /* If multi-range, clear tare and zero on return to zero */
  if (0 != ass->adcInternalData.rangeMax && ass->adcInternalData.rangeMulti)
  {
    if (0 != previousRange &&
        0 == ass->adcMiscellaneous.adcInfo.status.bit.range &&
        ass->adcMiscellaneous.adcInfo.status.bit.zeroRange)
    {
      ass->adcRuntimeDatabase.cntZero = ass->adcInternalData.cntSysRdg;
      ass->adcRuntimeDatabase.netMode = FALSE;
      ass->adcRuntimeDatabase.cntTare = 0;
    }
  }

  /* Initiate a zero self calibration if there are temperature errors */
  if (0 != (ass->adcMiscellaneous.adcErrors & ADC_TEMPORARY_ERRORS))
  {
    AdcIntStartZeroCalCycle(&ass->adcTempCal);
  }

  /* Do overload counter */
  if (ass->adcInternalData.cntSysRdg > ass->adcMiscellaneous.cntOverloadCntLevel ||
      ass->adcInternalData.cntSysRdg < -ass->adcMiscellaneous.cntOverloadCntLevel)
  {
    if (!ass->adcMiscellaneous.overloaded)
    {
      ass->adcMiscellaneous.overloaded = TRUE;
      if (ass->adcRuntimeDatabase.overloadCnt < 30000) ass->adcRuntimeDatabase.overloadCnt++;
    }
  }
  else
  {
    ass->adcMiscellaneous.overloaded = FALSE;
  }

#if ADC_DEBUG(ADC_UPDATE_DEBUG_LVL, ADC_DEBUG_FULL)
  ADC_DEBUG_PRINTF(LOG_DEBUG, "AdcUpdateReadings: finished successfully\n");
#endif
  return TRUE;
}


/** Generate the alternate units resolutions */
void AdcIntGenerateAltRes(AdcSuperStructure *ass, enum eAdcUnitSwitch unit)
{
  int8u i;
  int8s cnt;
  int32s *res[ADC_MAX_NUM_INTERVAL];
  int32s *conv;
  int8s *dpInc;
  int8u *resLen[ADC_MAX_NUM_INTERVAL];

  switch(unit)
  {
    case ADC_UNIT_SWITCH_SECONDARY:
      for (i = 0; i < ADC_MAX_NUM_INTERVAL; i++)
      {
        res[i] = &ass->adcInternalData.altRes[i];
        resLen[i] = &ass->adcInternalData.altResLen[i];
      }
      conv = &ass->adcInternalData.altConv;
      dpInc = &ass->adcInternalData.altDpInc;
      break;

    case ADC_UNIT_SWITCH_TERTIARY:
      for (i = 0; i < ADC_MAX_NUM_INTERVAL; i++)
      {
        res[i] = &ass->adcInternalData.tertRes[i];
        resLen[i] = &ass->adcInternalData.tertResLen[i];
      }
      conv = &ass->adcInternalData.tertConv;
      dpInc = &ass->adcInternalData.tertDpInc;
      break;

    default:
      return;
      break;
  }

  // Move dp if required. Aim is to keep the same number of fixed zeros.
  *dpInc = 0;

  if (ass->adcInternalData.approvalMode != ADC_APPROVAL_MODE_OIML)
  {

    // get 1000 * min resolution
    *res[0] = AdcMultDiv((int32s) AdcResolutions[ass->adcUserDatabase.resolution[0]],
      *conv, 1000L);

    if (*res[0])
    {
      // get length of alternate resolution
      cnt = 1;
      while (*res[0] > 8700)
      {
        *res[0] /= 10;
        cnt++;
      }
      while (*res[0] <= 870)
      {
        *res[0] *= 10;
        cnt--;
      }

      // compare to length of primary resolution
      cnt -= AdcResolutionLengths[ass->adcUserDatabase.resolution[0]];

      // move decimal point right or left
      while (cnt > 0 && (*dpInc + ass->adcUserDatabase.dpPos) > 0)
      {
        *dpInc -= 1;
        cnt--;
        *conv /= 10;
      }
      while (cnt < 0 && (*dpInc + ass->adcUserDatabase.dpPos) < 5)
      {
        *dpInc += 1;
        cnt++;
        *conv *= 10;
      }
    }
  }

  for (i = 0; i < ADC_MAX_NUM_INTERVAL; i++)
  {
    cnt = 0;
    *res[i] = AdcMultDiv((int32s) AdcResolutions[ass->adcUserDatabase.resolution[i]],
      *conv, 10000L); // 100 * new resolution

    while (*res[i] >= 1000)
    {
      *res[i] /= 10;
      cnt++;
    }

    if (ass->adcInternalData.approvalMode != ADC_APPROVAL_MODE_OIML)
    {
      if (*res[i] > 575) *res[i] = 10;
      else if (*res[i] > 230) *res[i] = 5;
      else if (*res[i] > 115) *res[i] = 2;
      else *res[i] = 1;
    }
    else
    {
      if (*res[i] > 500) *res[i] = 10;
      else if (*res[i] > 200) *res[i] = 5;
      else if (*res[i] > 100) *res[i] = 2;
      else *res[i] = 1;
    }

    *resLen[i] = (int8u) (cnt + 1);
    if (*res[i] == 10) *resLen[i] += 1;

    while (cnt > 0)
    {
      *res[i] *= 10;
      cnt--;
    }
  }
}

