#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcreadings.h>

#include <rinweigh/adczerotrack.h>


/**
 * Initialise zero tracking
 *
 * maxDelta is the maximum number of counts (weight * 16) to be canceled in skipSamples period. If 0, zero tracking will be disabled
 * skipSamples is the number of samples (at ADC reading frequency) to skip before applying up to maxDelta counts of zero tracking
 * opBlocked is whether zero tracking should be disabled
 */
void AdcIntZeroTrackingInit(AdcZeroTracking *zt, int32u maxDelta, int32u skipSamples, bool opBlocked)
{
  zt->skipSamples = skipSamples;
  zt->sampleCnt = 0;
  zt->maxDelta = maxDelta;
  zt->opBlocked = opBlocked;
}


/** Disable or enable zero tracking */
void AdcIntZeroTrackingBlock(AdcZeroTracking *zt, bool block)
{
  zt->opBlocked = block;
}


/** Run the zero tracking, based on settings in ass->adcZeroTracking */
bool AdcIntZeroTrackingDo(AdcSuperStructure *ass)
{
  int32s rdg;
  bool minus;

  /* No zero tracking if blocked, or maxDelta == 0 */
  if (ass->adcZeroTracking.opBlocked || 0 == ass->adcZeroTracking.maxDelta)
    return false;

  /* No zero tracking during scale motion */
  if (ass->adcDataWorking.grossRdg.info.status.bit.motion)
    return false;

  /* No zero tracking in NTEP when in net mode */
  if ((ass->adcInternalData.approvalMode == ADC_APPROVAL_MODE_NTEP) && (!ass->adcDataWorking.dispRdg.info.status.bit.gross))
    return false;

  /* Rate limiting - only apply zero tracking every skipSamples (at ADC reading frequency) */
  ass->adcZeroTracking.sampleCnt++;
  if (ass->adcZeroTracking.sampleCnt <= ass->adcZeroTracking.skipSamples)
    return false;
  ass->adcZeroTracking.sampleCnt = 0;

  /* Get reading in counts to work with (gross only in NTEP, whatever is displayed otherwise) */
  if ((ass->adcDataWorking.dispRdg.info.status.bit.gross) || (ass->adcInternalData.approvalMode == ADC_APPROVAL_MODE_NTEP))
    rdg = ass->adcInternalData.cntGross;
  else
    rdg = ass->adcInternalData.cntNet;

  /* Keep track of sign */
  if (rdg < 0) 
  {
    rdg = -rdg;
    minus = true;
  }
  else
  {
    minus = false;
  }

  /* If we are outside the zero band (plus or minus, then don't do zero tracking */
  if (rdg > ass->adcInternalData.cntZeroBandMax)
    return false;

  /* Apply zero tracking of upto maxDelta counts */
  if (rdg > ass->adcZeroTracking.maxDelta)
    rdg = ass->adcZeroTracking.maxDelta;
  if (minus)
    rdg = -rdg;

  /* Keep within in absolute limits */
  rdg += ass->adcRuntimeDatabase.cntZero;
  if (rdg > ass->adcInternalData.cntZeroRangeMax)
    rdg = ass->adcInternalData.cntZeroRangeMax;
  if (rdg < ass->adcInternalData.cntZeroRangeMin)
    rdg = ass->adcInternalData.cntZeroRangeMin;
  ass->adcRuntimeDatabase.cntZero = rdg;

  return true;
}
