# Generate the checksum for the adc library
#
# Replacement for rincrcccitt16, implemented in python so it can be
# distributed with librinhw, rather than requiring an external executable
import argparse
import datetime
import os

def calculate_ccitt_crc16(data, seed):
    crc = seed
    for b in data:
        bint = ord(b)
        crc = crc ^ ((bint & 0xFF) << 8)
        for i in range(0, 8):
            if (crc & 0x8000):
                crc = (crc << 1) ^ 0x1021
            else:
                crc = (crc << 1)
    return (crc & 0xFFFF)

def generate_checksum(infile, seed):
    readsize = 1024 * 10
    eof = False
    crc = seed
    inflength = 0
    while (not eof):
        data = infile.read(readsize)
        inflength = inflength + len(data)
        crc = calculate_ccitt_crc16(data, crc)
        eof = (data == "")
    return (crc, inflength)

def write_header_file(oufile, crc, infilename, outfilename, file_def, checksum_def, exename, seed):
    time_str = datetime.datetime.now().isoformat()

    outfile.write("/**\n")
    outfile.write(" * @file %s\n" % (outfilename))
    outfile.write(" *\n")
    outfile.write(" * CCITT CRC 16 Checksum Header\n")
    outfile.write(" *\n")
    outfile.write(" * FILE AUTOMATICALLY GENERATED BY %s. DO NOT EDIT!\n" % (exename))
    outfile.write(" *\n")
    outfile.write(" * This header file generated from \"%s\" on %s\n" % (infilename, time_str))
    outfile.write(" */\n")
    outfile.write("#ifndef %s\n" % (file_def.upper()))
    outfile.write("#define %s 1\n" % (file_def.upper()))
    outfile.write("\n\n")
    outfile.write("/** CCITT CRC 16 Checksum for \"%s\" with seed of 0x%04X */\n" % (infilename, seed))
    outfile.write("#define %s 0x%04X\n" % (checksum_def.upper(), crc))
    outfile.write("\n")
    outfile.write("#endif /* #ifndef %s */\n" % (file_def.upper()))

def generate_file_definition(filename):
    """Generate FILE_DOT_H for filename == 'file.h' """
    fdef =  os.path.basename(filename)
    fdef = fdef.replace(" ", "_")
    fdef = fdef.replace("-", "_")
    fdef = fdef.replace(".", "_DOT_")
    return fdef

def generate_file_checksum_definition(filename):
    """Generate FILE_CCITT_CRC16_CHECKSUM for filename == 'file.x' """
    fdef =  os.path.basename(filename)
    fdef, _ = os.path.splitext(fdef)
    fdef = fdef.replace(" ", "_")
    fdef = fdef.replace("-", "_")
    fdef = fdef.replace(".", "_")

    fdef = fdef + "_CCITT_CRC16_CHECKSUM"
    return fdef

def auto_int(x):
    return int(x, 0)

if __name__ == '__main__':
    # Parse command line options
    parser = argparse.ArgumentParser(description='File CCITT-CRC16 Checksum To C Header File Generator by Ian Burns <ian.burns@rinstrum.com>')
    parser.add_argument('seed', action='store', type=auto_int,
                        help='16bit hexadecimal checksum seed (as hex: 0xNNNN, or decimal).')
    parser.add_argument('inputfile', action='store',
                        help='The file to generate the checksum over.')
    parser.add_argument('outputfile', action='store',
                        help='The c header file to generate which contains the define of the checksum.')
    parser.add_argument('--define', action='store', default=None,
                        help='The name of the #define within the c header file for the checksum.')
    args = parser.parse_args()

    # Open the files
    infile = open(args.inputfile, 'r')
    outfile = open(args.outputfile, 'w')

    # Generate the checksum
    crc, inflength = generate_checksum(infile=infile, seed=args.seed)

    # Generate some strings
    file_def = generate_file_definition(filename=args.outputfile)
    if (args.define is not None):
        chk_def = filename=args.define
    else:
        chk_def = generate_file_checksum_definition(filename=args.inputfile)

    # Write out to header file
    write_header_file(oufile=outfile, crc=crc, infilename=args.inputfile,
                      outfilename=args.outputfile, file_def=file_def, checksum_def=chk_def,
                      exename=parser.prog, seed=args.seed)

    # Print some nice info
    print "CCITT-CRC16 Checksum: 0x%04X over %u bytes of file '%s' using seed 0x%04X." % (crc, inflength, args.inputfile, args.seed)
    print "C Header file '%s' created" % (args.outputfile)

    outfile.close()
    infile.close()

