/*
 * adcdbdump.c
 *
 *  Created on: 05/05/2011
 *      Author: ian
 */
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <rinweigh/adcinterface.h>
#include "print_settings.h"


static int dump_factory = 0;
static int dump_user = 0;
static int dump_runtime = 0;
static int display_help = 0;
static int dump_copy = 0;
static const char *filename = NULL;

static AdcFactoryDatabase adc_factory_database;
static AdcUserDatabase adc_user_database;
static AdcRuntimeDatabase adc_runtime_database;


static void print_usage(const char *prog)
{
  printf("Usage: %s [-furch] dbfile\n", prog);
  printf("  -f --factory  print the factory database\n"
         "  -u --user     print the user database\n"
         "  -r --runtime  print the user runtime database\n"
         "  -c --copy     specify which copy of the database to read (0,1)\n"
         "  -h --help     this help\n"
        );
  printf("Dumps the contents of a ADC Library Database in file dbfile\n");
}


int parse_opts(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "factory",  0, 0, 'f' },
    { "user",     0, 0, 'u' },
    { "runtime",  0, 0, 'r' },
    { "copy",     1, 0, 'c' },
    { "help",     0, 0, 'h' },
    { NULL, 0, 0, 0 },
  };
  int c;

  while (1)
  {
    c = getopt_long(argc, argv, "furc:h", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      case 'f':
        dump_factory = 1;
        break;
      case 'u':
        dump_user = 1;
        break;
      case 'r':
        dump_runtime = 1;
        break;
      case 'c':
        dump_copy = atoi(optarg);
        if (dump_copy < 0 || dump_copy > 1)
        {
          fprintf(stderr, "ERROR: ADC database copy out of range (0..1)\n");
          return -1;
        }
        break;
      case 'h':
        display_help = 1;
        return 0;
        break;
      default:
        fprintf(stderr, "ERROR: Unknown option. Try --help.\n");
        return -1;
        break;
    }
  }

  if (optind >= argc)
  {
    fprintf(stderr, "ERROR: No database file specified\n");
    return -1;
  }
  filename = argv[optind];

  return 0;
}


struct s_db_offset
{
  int factory_offset;
  int user_offset;
  int runtime_offset;
};


static const struct s_db_offset db_offsets[2] =
{
#if 0 /* TODO fix database */
  { START_ADC_DB_FACT1_FRAM, START_ADC_DB_USER1_FRAM, START_ADC_DB_RT1_FRAM },
  { START_ADC_DB_FACT2_FRAM, START_ADC_DB_USER2_FRAM, START_ADC_DB_RT2_FRAM },
#endif
};


int dbread(FILE *dbfile, const struct s_db_offset *dboffs, AdcFactoryDatabase *fdb, AdcUserDatabase *udb, AdcRuntimeDatabase *rtdb)
{

  /* Read factory database */
  if (fseek(dbfile, dboffs->factory_offset, SEEK_SET) < 0)
  {
    fprintf(stderr, "ERROR: Could not seek start of factory database\n");
    return -1;
  }
  if (1 != fread(fdb, sizeof(AdcFactoryDatabase), 1, dbfile))
  {
    fprintf(stderr, "ERROR: Could not read factory database\n");
    return -1;
  }

  /* Read user database */
  if (fseek(dbfile, dboffs->user_offset, SEEK_SET) < 0)
  {
    fprintf(stderr, "ERROR: Could not seek start of user database\n");
    return -1;
  }
  if (1 != fread(udb, sizeof(AdcUserDatabase), 1, dbfile))
  {
    fprintf(stderr, "ERROR: Could not read user database\n");
    return -1;
  }

  /* Read runtime database */
  if (fseek(dbfile, dboffs->runtime_offset, SEEK_SET) < 0)
  {
    fprintf(stderr, "ERROR: Could not seek start of runtime database\n");
    return -1;
  }
  if (1 != fread(rtdb, sizeof(AdcRuntimeDatabase), 1, dbfile))
  {
    fprintf(stderr, "ERROR: Could not read runtime database\n");
    return -1;
  }

  return 0;
}


int main(int argc, char *argv[])
{
  int result = -1;
  FILE *dbfile;

  /* Parse command line*/
  if (parse_opts(argc, argv) < 0)
  {
    goto m_done;
  }

  /* Print help if requested */
  if (display_help)
  {
    print_usage(argv[0]);
    result = 0;
    goto m_done;
  }

  /* Open database file */
  dbfile = fopen(filename, "r");
  if (NULL == dbfile)
  {
    fprintf(stderr, "ERROR: Could not open %s\n", filename);
    goto m_done;
  }

  if (dbread(dbfile, &db_offsets[dump_copy], &adc_factory_database, &adc_user_database, &adc_runtime_database) < 0)
  {
    goto m_close;
  }

  /* Dump factory database */
  if (dump_factory)
  {
    print_factory_settings(stdout, &adc_factory_database);
    print_factory_checksum(stdout, &adc_factory_database);
  }

  /* Dump user database */
  if (dump_user)
  {
    print_build_settings(stdout, &adc_user_database);
    print_option_settings(stdout, &adc_user_database);
    print_other_settings(stdout, &adc_user_database);
    print_cal_settings(stdout, &adc_user_database);
    print_user_checksum(stdout, &adc_user_database);
  }

  /* Dump user runtime database */
  if (dump_runtime)
  {

  }

  /* Success */
  result = 0;

m_close:
  fclose(dbfile);

m_done:
  return result;
}
