/*
 * adctemp_test.c
 *
 * Program to capture the temperature reading from both
 * the WE2111 temperature sensor (connected via the connector board Atmel)
 * and the ADS1216 internal temperature sensor.
 *
 * For testing purposes to determine the difference during warmup
 *
 * Much of this is taken from the TI application note SBAA073A:
 * Measuring Temperature with the ADS1216
 *
 * Make sure you have pins 1-2 & 3-4 of the loadcell connector
 * shorted so that Vref+ and Vref- on the ADC are set to the
 * correct levels
 *
 */
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>
#include <time.h>
#include <rinhw/hwlib_init.h>
#include <rinhw/we21hw.h>
#include <rinblock/rinblock.h>
#include <rinblock/rb_we21_adc.h>
#include <rinblock/rb_startup.h>
#include <rinhw/iic_local_linux.h>
#include <rinhw/sio_local_console.h>


/* ADC Crystal Time/Frequency Dependent Constants */

/** Default operating frequency in 0.1Hz (ie 125 = 12.5Hz) */
#define DEFAULT_FREQ 500

/** ADC oscillator frequency */
#define ADS1216_FOSC 4915200
#define ADS1216_FMOD_SPEED128 (ADS1216_FOSC / 128)
#define ADS1216_FMOD_SPEED256 (ADS1216_FOSC / 256)

/** Time for one oscillation in nanoseconds */
#define ADS1216_TOSC ( 1000000000 / ADS1216_FOSC )

/** Convert a number of TOSCs into nanoseconds */
#define ADS1216_TOSCS_TO_NANOSEC(x) ( x * ADS1216_TOSC )

/** Convert a number of TOSCs into nanoseconds */
#define ADS1216_TOSCS_TO_MICROSEC(x) ( ( ADS1216_TOSCS_TO_NANOSEC( x ) ) / 1000 )

/** Serial clock maximum frequency */
/* #define ADS1216_SCLK_T1_FREQ_HZ ( ADS1216_FOSC / 4 ) */
#define ADS1216_SCLK_T1_FREQ_HZ ( 350000 )

/** Register read last SCK Din to first SCK Dout time t6 in microseconds (see datasheet) */
#define ADS1216_RREG_T6_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 50 ) )

/** Read data last SCK Din to first SCK Dout time t6 in microseconds (see datasheet) */
#define ADS1216_RDATA_T6_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 50 ) )

/** Reset sequence time t11 in microseconds (see datasheet) */
#define ADS1216_RESET_T11_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 16 ) )

/** Read register time t11 in microseconds (see datasheet) */
#define ADS1216_RREG_T11_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 4 ) )

/** Write register time t11 in microseconds (see datasheet) */
#define ADS1216_WREG_T11_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 4 ) )

/** Read RAM time t11 in microseconds (see datasheet) */
#define ADS1216_RRAM_T11_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 4 ) )

/** Write RAM time t11 in microseconds (see datasheet) */
#define ADS1216_WRAM_T11_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 4 ) )

/** Reset sequence time t12 in microseconds (see datasheet) */
#define ADS1216_RESET_T12_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 400 ) )

/** Reset sequence time t13 in microseconds (see datasheet) */
#define ADS1216_RESET_T13_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 5 ) )

/** Reset sequence time t14 in microseconds (see datasheet) */
#define ADS1216_RESET_T14_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 650 ) )

/** Reset sequence time t15 in microseconds (see datasheet) */
#define ADS1216_RESET_T15_MICROSEC ( ADS1216_TOSCS_TO_MICROSEC( 1150 ) )


/* ADC SPI Commands */
#define ADS1216_CMD_RDATA              0x01    /* Reads the latest ADC value */
#define ADS1216_CMD_RDATAC             0x03    /* Read data continuously */
#define ADS1216_CMD_STOPC              0x0F    /* Stop read data continuously */
#define ADS1216_CMD_RREG               0x10    /* Read register in ADC */
#define ADS1216_CMD_RRAM               0x20    /* Read RAM bank */
#define ADS1216_CMD_CREG               0x40    /* Copy registers to RAM bank 'aaa' */
#define ADS1216_CMD_CREGA              0x48    /* Copy registers to all RAM banks */
#define ADS1216_CMD_WREG               0x50    /* Write register in ADC */
#define ADS1216_CMD_WRAM               0x60    /* Write RAM bank */
#define ADS1216_CMD_CRAM               0xC0    /* Copy RAM bank 'aaa' to register */
#define ADS1216_CMD_CSRAMX             0xD0    /* Calculate RAM bank 'aaa' checksum */
#define ADS1216_CMD_CSARAMX            0xD8    /* Calculate all RAM bank checksums */
#define ADS1216_CMD_CSREG              0xDF    /* Calculate register checksum */
#define ADS1216_CMD_CSRAM              0xE0    /* Calculate RAM bank 'aaa' checksum */
#define ADS1216_CMD_CSARAM             0xE8    /* Calculate all RAM bank checksums */
#define ADS1216_CMD_SELFCAL            0xF0    /* Self Calibrate Offset and Gain */
#define ADS1216_CMD_SELFOCAL           0xF1    /* Self calibrate offset */
#define ADS1216_CMD_SELFGCAL           0xF2    /* Self calibrate gain */
#define ADS1216_CMD_SYSOCAL            0xF3    /* System calibrate offset */
#define ADS1216_CMD_SYSGCAL            0xF4    /* System calibrate gain */
#define ADS1216_CMD_DYSNC              0xFC    /* Resets the conversion process */
#define ADS1216_CMD_SLEEP              0xFD    /* Put into sleep mode */
#define ADS1216_CMD_RESET              0xFE    /* Reset to power-up values */


/* ADC register Indexing */
#define ADS1216_REG_SETUP              0x00    /* Setup Register */
#define ADS1216_REG_MUX                0x01    /* Multiplex Control Register */
#define ADS1216_REG_ACR                0x02    /* Analog Control Register */
#define ADS1216_REG_IDAC1              0x03    /* Current DAC 1 */
#define ADS1216_REG_IDAC2              0x04    /* Current DAC 2 */
#define ADS1216_REG_ODAC               0x05    /* Offset DAC Setting */
#define ADS1216_REG_DIO                0x06    /* Digital Port Data Register */
#define ADS1216_REG_DIR                0x07    /* Direction Control Register */
#define ADS1216_REG_DEC0               0x08    /* Decimation Register (lower 8 bits) */
#define ADS1216_REG_MDEC1              0x09    /* Mode and Decimation Register */
#define ADS1216_REG_OCR0               0x0A    /* Offset Calibration Coefficient LSB */
#define ADS1216_REG_OCR1               0x0B    /* Offset Calibration Coefficient Middle Byte */
#define ADS1216_REG_OCR2               0x0C    /* Offset Calibration Coefficient MSB */
#define ADS1216_REG_FSR0               0x0D    /* Full-Scale Register LSB */
#define ADS1216_REG_FSR1               0x0E    /* Full-Scale Register Middle Byte */
#define ADS1216_REG_FSR2               0x0F    /* Full-Scale Register MSB */


/* ADS1216 Setup register constants */
#define REG_SETUP_BIT_ORDER_MSBFIRST   0x00    /* B0, 0=MSB first, 1=LSB first*/
#define REG_SETUP_BIT_ORDER_LSBFIRST   0x01    /* B0, 0=MSB first, 1=LSB first*/
#define REG_SETUP_BUF_EN_OFF           0x00    /* B1, 0=Buffer Off, 1=Buffer On */
#define REG_SETUP_BUF_EN_ON            0x02    /* B1, 0=Buffer Off, 1=Buffer On */
#define REG_SETUP_REF_HI_1V25          0x00    /* B2, 0=Int ref 1.25V 1=Int Reg 2.5V */
#define REG_SETUP_REF_HI_2V50          0x04    /* B2, 0=Int ref 1.25V 1=Int Reg 2.5V */
#define REG_SETUP_REF_EN_OFF           0x00    /* B3, 0=Int Ref off, 1=Int ref on */
#define REG_SETUP_REF_EN_ON            0x08    /* B3, 0=Int Ref off, 1=Int ref on */
#define REG_SETUP_SPEED_MASK           0x10    /* B4, Mask for speed bit */
#define REG_SETUP_SPEED_DIV128         0x00    /* B4, 0=Osc/128, 1=Osc/256 */
#define REG_SETUP_SPEED_DIV256         0x10    /* B4, 0=Osc/128, 1=Osc/256 */
#define REG_SETUP_ID_MASK              0xE0    /* B7-B5 Mask for the three factory programmed ID bits */
#define REG_SETUP_ID_VALUE1            0x80    /* B7-B5 Expected value for the three factory programmed ID bits */
#define REG_SETUP_ID_VALUE2            0x40    /* B7-B5 Expected value for the three factory programmed ID bits */


/* ADS1216 Multiplex register constants */
#define REG_MUX_NSEL_MASK              0x0F
#define REG_MUX_PSEL_MASK              0xF0
#define REG_MUX_CH_AIN0                0x00
#define REG_MUX_CH_AIN1                0x01
#define REG_MUX_CH_AIN2                0x02
#define REG_MUX_CH_AIN3                0x03
#define REG_MUX_CH_AIN4                0x04
#define REG_MUX_CH_AIN5                0x05
#define REG_MUX_CH_AIN6                0x06
#define REG_MUX_CH_AIN7                0x07
#define REG_MUX_CH_AINCOM              0x08
#define REG_MUX_CH_DIODE               0x0F
#define REG_MUX_NSEL_AIN(x)            ( x & REG_MUX_NSEL_MASK ) /* Set negative channel */
#define REG_MUX_PSEL_AIN(x)            ( ( x << 4 ) & REG_MUX_PSEL_MASK ) /* Set positive channel */


/* Analog control register constants */
#define REG_ACR_PGA_MASK               0x07
#define REG_ACR_PGA1                   0x00
#define REG_ACR_PGA2                   0x01
#define REG_ACR_PGA4                   0x02
#define REG_ACR_PGA8                   0x03
#define REG_ACR_PGA16                  0x04
#define REG_ACR_PGA32                  0x05
#define REG_ACR_PGA64                  0x06
#define REG_ACR_PGA128                 0x07
#define REG_ACR_PGA(x)                 ( x & REG_ACR_PGA_MASK ) /* Set PGA */
#define REG_ACR_IDAC1_RANGE_MASK       0x18
#define REG_ACR_IDAC2_RANGE_MASK       0x60
#define REG_ACR_IDAC_OFF               0x00
#define REG_ACR_IDAC_RANGE1            0x01
#define REG_ACR_IDAC_RANGE2            0x02
#define REG_ACR_IDAC_RANGE3            0x03
#define REG_ACR_IDAC1_RANGE(x)         ( ( x << 3 ) & REG_ACR_IDAC1_RANGE_MASK ) /* Set IDAC1 Range */
#define REG_ACR_IDAC2_RANGE(x)         ( ( x << 5 ) & REG_ACR_IDAC2_RANGE_MASK ) /* Set IDAC2 Range */
#define REG_ACR_BOCS_MASK              0x80
#define REG_ACR_BOCS_OFF               0x00   /* Burn Out Current Source Off */
#define REG_ACR_BOCS_ON                0x80   /* Burn Out Current Source On */


/* ADS1216 Mode and decimation register constants */
#define REG_MDEC_DEC_MASK              0x07
#define REG_MDEC_DEC(x)                ( x & REG_MDEC_DEC_MASK )
#define REG_MDEC_RESERVED_MASK         0x08
#define REG_MDEC_SMODE_MASK            0x30
#define REG_MDEC_SMODE_AUTO            0x00
#define REG_MDEC_SMODE_FAST            0x01
#define REG_MDEC_SMODE_SINC2           0x02
#define REG_MDEC_SMODE_SINC3           0x03
#define REG_MDEC_SMODE(x)              ( ( x << 4 ) & REG_MDEC_SMODE_MASK ) /* Set SMODE Range */
#define REG_MDEC_UB_BIPOLAR            0x00
#define REG_MDEC_UB_UNIPOLAR           0x40
#define REG_MDEC_DRDY_MASK             0x80    /* Data Ready Flag - low when data available */


/* Setup register */
#define REG_SETUP_SPEED_DEFAULT REG_SETUP_SPEED_DIV128
#define SETUP_VALUE ( REG_SETUP_BIT_ORDER_MSBFIRST | REG_SETUP_BUF_EN_ON | \
                      REG_SETUP_REF_HI_2V50 | REG_SETUP_REF_EN_ON | \
                      REG_SETUP_SPEED_DIV128 )


/* Multiplex Register: Select AINCOM */
#define MUX_VALUE_AINCOM ( REG_MUX_NSEL_AIN( REG_MUX_CH_AINCOM ) | REG_MUX_PSEL_AIN( REG_MUX_CH_AINCOM ) )


/* Multiplex Register: Select temperature diode */
#define MUX_VALUE_DIODE ( REG_MUX_NSEL_AIN( REG_MUX_CH_DIODE ) | REG_MUX_PSEL_AIN( REG_MUX_CH_DIODE ) )


/* Analog Control Register: */
#define ACR_VALUE_WITH_PGA(x) ( REG_ACR_PGA( x ) | \
                                REG_ACR_IDAC1_RANGE( REG_ACR_IDAC_OFF ) | \
                                REG_ACR_IDAC2_RANGE( REG_ACR_IDAC_OFF ) | \
                                REG_ACR_BOCS_OFF )
#define ACR_VALUE_DEFAULT ( ACR_VALUE_WITH_PGA( REG_ACR_PGA1 ) )


/* Mode and decimation register */
#define MDEC_VALUE_WITH_DEC(x) ( REG_MDEC_DEC(x) | \
                                 REG_MDEC_SMODE( REG_MDEC_SMODE_AUTO ) | \
                                 REG_MDEC_UB_BIPOLAR )


/* The default value of offset calibration register value to use */
#define OCR_VALUE_DEFAULT 0


/* The default value of full scale register value to use */
#define FSR_VALUE_DEFAULT 0x008111BB


/* Command line options */
static bool cmdline_showhelp = false;
static const char *cmdline_rb_i2c_device = NULL;
static const char *cmdline_ads1216_spi_device = NULL;
static const char *cmdline_rin_hw_device = NULL;
static unsigned long cmdline_diode_microamps = 5;
static unsigned long cmdline_sample_frequency = 10;
static bool cmdline_dumpreadings = false;
static unsigned long cmdline_logtime = 0;


/* Rinstrum devices */
static struct t_io_dev dev_iic0;
static struct t_io_dev dev_adc;

static bool application_quit = false;


static bool RinBlockAssignDevices(void)
{
  int i;
  bool adcdone = false;

  for (i = 0; i < MAX_RINBLOCKS; i++)
  {
    /* ADC Device */
    if (RB_IFACE_TYPE_WE21_CONN == rinblocks[i].tag.interface_type && !adcdone)
    {
      /* Replace dev_adc with the rinblock one */
      if (RIO_SUCCESS == RinBlockCreateWE21AdcDevice(&dev_adc, &rinblocks[i]))
      {
        /* printf("Created ADC device for WE2111 rinblock %d ADC\r\n", i); */
        adcdone = true;
      }
    }
  }
  return adcdone;
}


static int hwlib_local_start(const char *i2c_device, const char *rinhw_device)
{
  /* Start the hardware library */
  hwlib_init();

  /* Initialise WE2111 hardware */
  if (rio_init_we21hw(rinhw_device) < 0)
  {
    fprintf(stderr, "ERROR: WE2111 hardware device fail\n");
    return -1;
  }

  /* Create the devices */
  if (rio_create_device_iic_local(&dev_iic0, i2c_device, we21hw_set_slot_select, NULL) < 0)
  {
    fprintf(stderr, "ERROR: I2C device creation failed\n");
    return -1;
  }

  RinBlockInit(&dev_iic0);
  RinBlockEnumerate();
  RinBlockStart();
  if (!RinBlockAssignDevices())
  {
    fprintf(stderr, "ERROR: Rinblock ADC device creation failed\n");
    return -1;
  }

  return 0;
}


static void hwlib_stop(void)
{
  /* Destroy devices */
  rdestroy_io(&dev_adc);
  rdestroy_io(&dev_iic0);

  /* Destroy the WE2111 hardware */
  rio_destroy_we21hw();

  /* Shutdown the hardware library */
  hwlib_destroy();
}


static bool rinblock_get_temperature(uint32_t *adc_temp)
{
  bool result = false;
  uint16_t temp16;

  if (ropen(&dev_adc, 0) >= 0)
  {
    if (rlseek(&dev_adc, sizeof(uint16_t) * ADC_CH_TEMP, SEEK_SET) == sizeof(uint16_t) * ADC_CH_TEMP)
    {
      if (rread(&dev_adc, &temp16, sizeof(uint16_t), 0) == sizeof(uint16_t))
      {
        *adc_temp = temp16;
        result = true;
      }
    }
    rclose(&dev_adc);
  }
  return result;
}


static int adc24bit_linux_transfer(int fd, const int8u *cmd_data, int8u cmd_len, int8u *ret_data, int8u ret_len)
{
  #define NUM_SPI_TRANSFERS 2
  struct spi_ioc_transfer transfer[NUM_SPI_TRANSFERS];
  int result;

  /* Setup write, need different cast for 32/64bit */
#if __LP64__
  transfer[0].tx_buf = (uint64_t) cmd_data;
#else
  transfer[0].tx_buf = (uint32_t) cmd_data;
#endif
  transfer[0].rx_buf = 0;
  transfer[0].len = cmd_len;
  transfer[0].speed_hz = ADS1216_SCLK_T1_FREQ_HZ;
  transfer[0].delay_usecs = ADS1216_RREG_T6_MICROSEC;
  transfer[0].bits_per_word = 8;
  transfer[0].cs_change = 0;
  transfer[0].pad = 0;

  /* Test to see if ADC will return data */
  if (ret_len > 0 && NULL != ret_data)
  {
    /* Setup read, need different cast for 32/64bit */
    transfer[1].tx_buf = 0;
#if __LP64__
    transfer[1].rx_buf = (uint64_t) ret_data;
#else
    transfer[1].rx_buf = (uint32_t) ret_data;
#endif
    transfer[1].len = ret_len;
    transfer[1].speed_hz = ADS1216_SCLK_T1_FREQ_HZ;
    transfer[1].delay_usecs = ADS1216_RREG_T11_MICROSEC;
    transfer[1].bits_per_word = 8;
    transfer[1].cs_change = 0;
    transfer[1].pad = 0;

    /*
     * There are two alternate calls to ioctl here because
     * when compiling for M68K, passing in a variable to
     * SPI_IOC_MESSAGE will cause the link to fail with
     * undefined reference to __invalid_size_argument_for_IOC.
     * Passing in a compile time constant works...
     *
     * This is a bug in the (older) M68k uclibc headers, or
     * maybe the kernel headers are old, or maybe both.
     * The problem exists in <asm-generic/ioctl.h> where
     * _IOC_TYPECHECK is incorrectly defined. This has been
     * fixed in newer kernel headers.
     */
    result = ioctl(fd, SPI_IOC_MESSAGE(NUM_SPI_TRANSFERS), transfer);
  }
  else
  {
    result = ioctl(fd, SPI_IOC_MESSAGE(1), transfer);
  }
  return (result < 0) ? RIO_ERROR : RIO_SUCCESS;
}


/* Execute a command on the ADC */
static int adc24bit_execute_command(int fd, const int8u *cmd_data, int8u cmd_len, int8u *ret_data, int8u ret_len)
{
  return adc24bit_linux_transfer(fd, cmd_data, cmd_len, ret_data, ret_len);
}


/** Get a reading from the ADC */
static int adc24bit_get_reading(int fd, int32s *data)
{
  int8u aui8Buffer[sizeof(int32u)];
  int8u i;
  int32u ui32Data;

  /* Send the RDATA command, and read back 24bits of ADC reading */
  aui8Buffer[0] = ADS1216_CMD_RDATA;
  if (adc24bit_execute_command(fd, aui8Buffer, 1, aui8Buffer, 3) < 0)
  {
    return RIO_ERROR;
  }

  /* Transfer data from the buffer into the long in the order that the ADC spits it out (MSB first) */
  for (i = 0, ui32Data = 0; i < 3; i++)
  {
    ui32Data |= ((int32u) (aui8Buffer[2 - i])) << (i * 8);
  }

#if 0
  fprintf(stderr, "adc24bit_get_reading: data bytes are 0:0x%02x, 1:0x%02x, 2:0x%02x\n", aui8Buffer[0], aui8Buffer[1], aui8Buffer[2]);
  fprintf(stderr, "adc24bit_get_reading: assembled result: 0x%lx\n", ui32Data);
#endif

  /* Sign extend from 24 bits to 32bits */
  if (0x00800000 == (ui32Data & 0x00800000))
  {
    ui32Data |= 0xFF000000;
  }
#if 0
  fprintf(stderr, "adc24bit_get_reading: sign extended result: %ld\n", (int32s) ui32Data);
#endif
  *data = (int32s) ui32Data;
  return RIO_SUCCESS;
}


/** Write a 1, 2 or 3 byte register on the ADC */
static int adc24bit_write_register(int fd, int8u reg, int32u data, int8u len)
{
  int8u aui8Buffer[2 + sizeof(data)];
  int8u i;

  /* Biggest register on the ADC is 24 bits */
  if (len > 3)
  {
    len = 3;
  }

  /* Setup write command with register to read, and setup length byte of number of registers to read */
  aui8Buffer[0] = (int8u) (reg | ADS1216_CMD_WREG);
  aui8Buffer[1] = (int8u) (len - 1);

  /* Put data into buffer in the order that the ADC expects (low order first - yes this is different order to RDATA command) */
  for (i = 0; i < len; i++)
  {
    aui8Buffer[i + 2] = (int8u) (data & 0x000000FF);
    data >>= 8;
  }

  /* Write the register write command and data to the ADC */
  return adc24bit_execute_command(fd, aui8Buffer, (int8u) ( 2 + len ), NULL, 0);
}


/** Read a 1, 2 or 3 byte register on the ADC */
static int adc24bit_read_register(int fd, int8u reg, int32u *pdata, int8u len)
{
  int8u aui8CmdBuffer[2];
  int8u aui8DataBuffer[sizeof(*pdata)];
  int8u i;
  int32u ui32Data;

  /* Biggest register on the ADC is 24 bits */
  if (len > 3)
  {
    len = 3;
  }

  /* Setup read command with register to read, and setup length byte of number of registers to read */
  aui8CmdBuffer[0] = (int8u) (reg | ADS1216_CMD_RREG);
  aui8CmdBuffer[1] = (int8u) (len - 1);

  /* Read the register */
  if (adc24bit_execute_command(fd, aui8CmdBuffer, 2, aui8DataBuffer, len) < 0)
  {
    return RIO_ERROR;
  }

  /* Transfer data from the buffer into the long in the order that the ADC spits it out (low order first - yes this is different order to RDATA command) */
  ui32Data = 0;
  for (i = 0; i < len; i++)
  {
    ui32Data |= ((int32u) (aui8DataBuffer[i])) << (i * 8);
  }
  *pdata = ui32Data;
  return RIO_SUCCESS;
}


static int adc24bit_read_register8(int fd, int8u ui8Reg, int8u *pui8Data)
{
  int res;
  int32u ui32Data;

  res = adc24bit_read_register(fd, ui8Reg, &ui32Data, sizeof(*pui8Data));
  *pui8Data = (int8u) ui32Data;
  return res;
}


static int adc24bit_write_register8(int fd, int8u ui8Reg, int8u ui8Data)
{
  return adc24bit_write_register(fd, ui8Reg, (int32u) ui8Data, sizeof(int8u));
}


static int adc24bit_read_register24(int fd, int8u ui8Reg, int32u *pui32Data)
{
  return adc24bit_read_register(fd, ui8Reg, pui32Data, sizeof(*pui32Data) - 1);
}


#if 0
static int adc24bit_write_register24(int fd, int8u ui8Reg, int32u ui32Data)
{
  return adc24bit_write_register(fd, ui8Reg, ui32Data, sizeof(ui32Data) - 1);
}
#endif


/** Set the decimation ratio of the ADC, frequency in 0.1Hz (125 = 12.5Hz) */
static int adc24bit_set_decimation_ratio(int fd, int frequency)
{
  int32u decimation;
  int8u decimation_lo;
  int8u mdec;
  int8u speed;
  int8u setup;

  /* First work out decimation value using SPEED_DIV128 */
  speed = REG_SETUP_SPEED_DIV128;
  decimation = ADS1216_FMOD_SPEED128 * 10 / frequency;
  if (decimation > 2047)
  {
    /* Can't do that slow using DIV128, but what about DIV256? */
    decimation = ADS1216_FMOD_SPEED256 * 10 / frequency;
    speed = REG_SETUP_SPEED_DIV256;
  }

#if 0
  fprintf(stderr, "adc24bit_set_decimation_ratio: @Fosc %ld, SPEED %d, freq %d (0.1Hz) gives dec ratio of %d\n",
               ADS1216_FOSC, (speed == REG_SETUP_SPEED_DIV256) ? 1 : 0, frequency, decimation);
#endif

  /* Make sure it's in range */
  if (decimation > 2047 || decimation < 20)
  {
    fprintf(stderr, "adc24bit_set_decimation_ratio: decimation ratio out of range\n");
    return -1;
  }

  /* Set the SPEED bit in the setup register */
  adc24bit_read_register8(fd, ADS1216_REG_SETUP, &setup);
  setup &= ~REG_SETUP_SPEED_MASK;
  setup |= speed;
  adc24bit_write_register8(fd, ADS1216_REG_SETUP, setup);

  /* Split into two parts for DEC and MDEC registers */
  decimation_lo = (int8u) (decimation & 0x00FF);
  mdec = (int8u) MDEC_VALUE_WITH_DEC((int8u) (int16u) ((decimation >> 8) & 0x00FF));

  /* Write the two registers */
  adc24bit_write_register8(fd, ADS1216_REG_DEC0, decimation_lo);
  adc24bit_write_register8(fd, ADS1216_REG_MDEC1, mdec);
#if 0
  fprintf(stderr, "adc24bit_set_decimation_ratio: decimation ratio set\n");
#endif
  return 0;
}


static int adc24bit_set_idac1_current(int fd, uint32_t micro_amps)
{
  /*
   * Idac = (Vref * 2^(range -1) * dac_code) / (8 * Rdac)
   *
   * Re-arranged:
   * dac_code = (Idac * 8 * RDac) / (Vref * 2^(range -1))
   *
   */
  uint32_t rdac;
  uint32_t vref;
  uint32_t dac_code;
  uint32_t range;
  uint8_t acr;

  rdac = 150000;      /* Ohms */
  vref = 2500 * 1000; /* 2.5V as uV */

  for (range = 1; range < 4; range++)
  {
    dac_code = micro_amps * 8 * rdac;
    dac_code /= vref;
    dac_code >>= (range - 1);
    if (dac_code > 0 && dac_code < 256)
      break;
  }

#if 0
  printf("adc24bit_set_idac1_current: for current of %lu uA, range = %lu, dac_code = %lu\n",
         (unsigned long) micro_amps, (unsigned long) range, (unsigned long) dac_code);
#endif

  if (0 == dac_code || dac_code > 255 || range > REG_ACR_IDAC_RANGE3)
  {
    fprintf(stderr, "adc24bit_set_idac1_current: requested current (%lu uA) is out of range (range %lu, dac_code %lu)\n",
            (unsigned long) micro_amps, (unsigned long) range, (unsigned long) dac_code);
    return -1;
  }

  /* Read/modify/write ACR to set range */
  adc24bit_read_register8(fd, ADS1216_REG_ACR, &acr);
  acr &= REG_ACR_IDAC1_RANGE_MASK;
  acr |= REG_ACR_IDAC1_RANGE((uint8_t) range);
  adc24bit_write_register8(fd, ADS1216_REG_ACR, acr);

  /* Write IDAC */
  adc24bit_write_register8(fd, ADS1216_REG_IDAC1, (uint8_t) dac_code);

  return 0;
}


#if 0
static void adc24bit_dump_registers(int fd)
{
  uint8_t data;
  uint32_t uldata;

  printf("adc24bit_dump_registers: registers:\n");

  adc24bit_read_register8(fd, ADS1216_REG_SETUP, &data);
  printf("SETUP: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_MUX, &data);
  printf("MUX: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_ACR, &data);
  printf("ACR: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_IDAC1, &data);
  printf("IDAC1: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_IDAC2, &data);
  printf("IDAC2: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_ODAC, &data);
  printf("ODAC: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_DIR, &data);
  printf("DIR: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_DIO, &data);
  printf("DIO: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_DEC0, &data);
  printf("DEC0: 0x%02x\n", data);
  adc24bit_read_register8(fd, ADS1216_REG_MDEC1, &data);
  printf("MDEC: 0x%02x\n", data);
  adc24bit_read_register24(fd, ADS1216_REG_OCR0, &uldata);
  printf("OCR: 0x%06lx\n", (unsigned long) uldata);
  adc24bit_read_register24(fd, ADS1216_REG_FSR0, &uldata);
  printf("FSR: 0x%06lx\n", (unsigned long) uldata);
}
#endif


/** Set the decimation ratio of the ADC */
static int adc24bit_perform_self_calibration(int fd)
{
  const int8u aui8CmdBuffer[1] = { ADS1216_CMD_SELFCAL };

  /* Execute the self zero calibration command */
  adc24bit_execute_command(fd, aui8CmdBuffer, 1, NULL, 0);

  return 14;
}



static int adc24bit_init_hardware(int fd, uint32_t diode_microamps, uint32_t frequency)
{
  /* Configure registers with defaults */
  adc24bit_write_register8(fd, ADS1216_REG_SETUP, SETUP_VALUE);
  adc24bit_write_register8(fd, ADS1216_REG_MUX, MUX_VALUE_AINCOM);
  adc24bit_write_register8(fd, ADS1216_REG_ACR, (int8u) ACR_VALUE_DEFAULT);
  adc24bit_set_decimation_ratio(fd, frequency * 10); /* Sets DEC0 and MDEC1 */

  /* Do self calibration, and delay until finished (any access before this time has expired will cause problems) */
  adc24bit_perform_self_calibration(fd);
  OSTimeDly((14 + 1) * OS_TICKS_PER_SEC / frequency);

#if 0
  printf("Post self cal registers:\n");
  adc24bit_dump_registers(fd);
#endif

  /* Setup for temperature measurement */
  adc24bit_write_register8(fd, ADS1216_REG_MUX, MUX_VALUE_DIODE);
  adc24bit_set_idac1_current(fd, diode_microamps);

#if 0
  printf("Post temperature setup:\n");
  adc24bit_dump_registers(fd);
#endif

  return RIO_SUCCESS;
}


/** Initialise linux device driver */
static int adc24bit_init_linux_spi_device(const char *devicename)
{
  int fd = -1;
  int result;
  uint8_t mode;
  uint8_t bits;
  uint32_t speed;

  fd = open(devicename, O_RDWR);
  if (fd < 0)
  {
    fprintf(stderr, "adc24bit_init_linux_device: failed opening %s, errno:%d, %s\n",
            devicename, errno, strerror(errno));
    goto ailsd_fail;
  }

  /* Set SPI mode */
  mode = 0; /* ( ~SPI_LOOP & ~SPI_CS_HIGH & ~SPI_LSB_FIRST & ~SPI_CPHA & ~SPI_CPOL ); */
  mode |= (SPI_3WIRE | SPI_MODE_3);
  result = ioctl(fd, SPI_IOC_WR_MODE, &mode);
  if (result < 0)
  {
    fprintf(stderr, "adc24bit_init_linux_device: can't set SPI write mode, errno:%d, %s\n",
            errno, strerror(errno));
    goto ailsd_fail;
  }

  /* Set bits per word */
  bits = 8;
  result = ioctl(fd, SPI_IOC_WR_BITS_PER_WORD, &bits);
  if (result < 0)
  {
    fprintf(stderr, "adc24bit_init_linux_device: can't set SPI bits per word, errno:%d, %s\n",
            errno, strerror(errno));
    goto ailsd_fail;
  }

  /* Set max speed in Hz */
  speed = 500000; /* TODO Set from constant */
  result = ioctl(fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed);
  if (result < 0)
  {
    fprintf(stderr, "adc24bit_init_linux_device: can't set SPI speed, errno:%d, %s\n",
            errno, strerror(errno));
    goto ailsd_fail;
  }

  /* All done */
  return fd;

ailsd_fail:
  if (fd >= 0)
    close(fd);
  return -1;
}


/** Initialise ADC hardware */
static int adc24bit_local_init(const char *devicename, uint32_t diode_microamps, uint32_t frequency)
{
  int fd;
  int result;

  /* Setup SPI */
  fd = adc24bit_init_linux_spi_device(devicename);
  if (fd < 0)
  {
    return fd;
  }

  /* Setup ADS1216 */
  result = adc24bit_init_hardware(fd, diode_microamps, frequency);
  if (result < 0)
  {
    return result;
  }

  return fd;
}


static void print_usage(void)
{
  printf("Usage: adctemp_test [-furch] i2cdev rinhwdev spidev\n"
         "  -f --freq     set the sample frequency (1..100Hz, default 10)\n"
         "  -i --current  set the ADS1216 temperature diode current (3..30uA, default 5)\n"
         "  -d --dump <n> dump readings until n seconds have passed, or ctrl-c (0=indefinite)\n"
         "  -h --help     this help\n"
         "  i2cdev        is the Linux I2C device of the rinblock bus\n"
         "  rinhwdev      is the Rinstrum hardware device (usually /dev/we21hw)\n"
         "  spidev        is the Linux SPI device connected to the ADS1216\n"
         "Control the ADS1216 temperature sensing diode\n");
}


static int parse_opts(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "frequency",  1, 0, 'f' },
    { "current",    1, 0, 'i' },
    { "dump",       1, 0, 'd' },
    { "help",       0, 0, 'h' },
    { NULL,         0, 0, '\0' },
  };
  int c;

  while (1)
  {
    c = getopt_long(argc, argv, "f:i:d:h", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      case 'f':
        if (1 != sscanf(optarg, "%lu", &cmdline_sample_frequency))
        {
          fprintf(stderr, "Error parsing sample frequency. Try --help.\n");
          return -1;
        }
        if (cmdline_sample_frequency < 1 || cmdline_sample_frequency > 100)
        {
          fprintf(stderr, "ADS1216 sample frequency out of range (1..100). Try --help\n");
          return -1;
        }
        break;
      case 'i':
        if (1 != sscanf(optarg, "%lu", &cmdline_diode_microamps))
        {
          fprintf(stderr, "Error parsing diode current. Try --help.\n");
          return -1;
        }
        if (cmdline_diode_microamps < 3 || cmdline_diode_microamps > 30)
        {
          fprintf(stderr, "ADS1216 temperature diode current out of range (3..30). Try --help\n");
          return -1;
        }
        break;
      case 'd':
        cmdline_dumpreadings = true;
        if (1 != sscanf(optarg, "%lu", &cmdline_logtime))
        {
          fprintf(stderr, "Error parsing log time. Try --help.\n");
          return -1;
        }
        break;
      case 'h':
        cmdline_showhelp = true;
        break;
      default:
        fprintf(stderr, "ERROR: Unknown option. Try --help.\n");
        return -1;
        break;
    }
  }

  if (optind >= argc && !cmdline_showhelp)
  {
    fprintf(stderr, "ERROR: No I2C device specified. Try --help.\n");
    return -1;
  }
  cmdline_rb_i2c_device = argv[optind];
  optind++;

  if (optind >= argc && !cmdline_showhelp)
  {
    fprintf(stderr, "ERROR: No Rinstrum hardware device specified. Try --help.\n");
    return -1;
  }
  cmdline_rin_hw_device = argv[optind];
  optind++;

  if (optind >= argc && !cmdline_showhelp)
  {
    fprintf(stderr, "ERROR: No SPI device specified. Try --help.\n");
    return -1;
  }
  cmdline_ads1216_spi_device = argv[optind];
  optind++;

  return 0;
}


static int wait_for_reading(int adc_spi_fd, int32_t *reading)
{
  /* Wait for ADC to give a reading */
  if (we21hw_adc_wait_data(OS_TICKS_PER_SEC) < 0)
  {
    fprintf(stderr, "Wait for ADC data failed\n");
    return -1;
  }

  /* Get data from ADC */
  if (adc24bit_get_reading(adc_spi_fd, reading) < 0)
  {
    fprintf(stderr, "Reading from ADC failed\n");
    return -1;
  }

  return 0;
}


static int set_current_wait_for_sinc(int fd, uint32_t current, int32_t *reading)
{
  int32_t dummy;
  int i;

  /* Set 3uA through diode */
  if (adc24bit_set_idac1_current(fd, current) < 0)
  {
    return -1;
  }

  /* Wait for sinc filter to clear */
  for (i = 0; i < 3; i++)
  {
    if (wait_for_reading(fd, &dummy) < 0)
    {
      return -1;
    }
  }

  /* Take a reading */
  if (wait_for_reading(fd, reading) < 0)
  {
    return -1;
  }

  return 0;
}


static double reading_to_voltage(uint32_t reading, uint32_t fsr)
{
  double ratio;
  double vex;
  double voltage;

  /* Vref+/- Resistor ratio for 7.4v excitation */
#if 0
  ratio = 10 / (15 + 10 + 4.7);
  vex = 7.34;
#endif

  /* Vref+/- Resistor ratio for 5v excitation */
  ratio = 10 / (4.99 + 10 + 4.99);
  vex = 4.99;

  voltage = reading;
  voltage *= (vex * ratio);
  voltage /= fsr;

  return voltage;
}


static double adc_calculate_temperature(int adc_spi_fd, uint32_t fsr_value, double *temperature)
{
  int32_t reading1;
  int32_t reading2;
  double voltage1;
  double voltage2;
  double alpha;

  /* Set 3uA through diode and wait for reading */
  if (set_current_wait_for_sinc(adc_spi_fd, 3, &reading1) < 0)
  {
    return -1;
  }

  /* Set 11uA through diode and wait for reading */
  if (set_current_wait_for_sinc(adc_spi_fd, 11, &reading2) < 0)
  {
    return -1;
  }

  voltage1 = reading_to_voltage(reading1, fsr_value);
  voltage2 = reading_to_voltage(reading2, fsr_value);

#if 0
  printf("rdg1:%ld, v1:%0.2f, rdg2:%ld, v2:%0.2f\n", (long) reading1, voltage1, (long) reading2, voltage2);
#endif

  alpha = 7000; /* Kelvin / volt, from TI app note SBAA073A */
  *temperature = (voltage2 - voltage1) * alpha;
  return 0;
}


static void test_adc_temperature2(int adc_spi_fd, unsigned long logtime)
{
  uint32_t rb_adc_temp = 0;
  uint32_t fsr_value;
  double temperature = 0.0;
  time_t ttime;
  struct tm tmtime;
  char timestr[50];
  time_t starttime;

  adc24bit_read_register24(adc_spi_fd, ADS1216_REG_FSR0, &fsr_value);

  if (0 == logtime)
    printf("Logging ADS1216 temperature diode readings indefinitely (ctrl-c to stop)\n");
  else
    printf("Logging ADS1216 temperature diode readings for %lu seconds or ctrl-c to stop\n", logtime);

  printf("Time,ADS1216 temperature,LM335 temperature\n");
  for (starttime = time(NULL); !application_quit && (0 == logtime || time(NULL) < (time_t) (logtime + starttime)); )
  {
    /* Get temperature from ADS1216 */
    if (adc_calculate_temperature(adc_spi_fd, fsr_value, &temperature) < 0)
    {
      return;
    }

    /* Get temperature from rinblock ADC */
    rinblock_get_temperature(&rb_adc_temp);

    /* Get current time */
    ttime = time(NULL);
    localtime_r(&ttime, &tmtime);
    strftime(timestr, 50, "%F %T", &tmtime);

    printf("%s,%0.2f,%lu\n", timestr, temperature, (unsigned long) rb_adc_temp);
  }
}


#if 0
static void test_adc_temperature(int adc_spi_fd)
{
  uint32_t rb_adc_temp;
  int32_t reading;
  uint32_t fsr_value;
  double voltage;

  adc24bit_read_register24(adc_spi_fd, ADS1216_REG_FSR0, &fsr_value);

  while (!application_quit)
  {

    /* Wait for ADC to give a reading */
    if (we21hw_adc_wait_data(OS_TICKS_PER_SEC) < 0)
    {
      fprintf(stderr, "Wait for ADC data failed\n");
      return;
    }

    /* Get data from ADC */
    if (adc24bit_get_reading(adc_spi_fd, &reading) < 0)
    {
      fprintf(stderr, "Reading from ADC failed\n");
      return;
    }

    voltage = reading;
    voltage /= fsr_value;
    voltage *= 2.49;

    /* Get temperature from rinblock ADC */
    rinblock_get_temperature(&rb_adc_temp);

    printf("ADS1216 reading:%8ld, %0.2f, Rinblock ADC reading:%8lu\n", (long) reading, voltage, (unsigned long) rb_adc_temp);
  }
}
#endif


static void sig_handler(int signo)
{
  switch (signo)
  {
    case SIGINT:
    case SIGTERM:
    case SIGHUP:
      application_quit = true;
      break;
    default:
      break;
  }
}


int main(int argc, char *argv[])
{
  int adc_spi_fd = -1;
  int result;
  bool hw_initd = false;

  if (parse_opts(argc, argv) < 0)
  {
    result = 1;
    goto m_fail;
  }

  if (cmdline_showhelp)
  {
    print_usage();
    result = 0;
    goto m_fail;
  }

  if (signal(SIGINT, sig_handler) == SIG_ERR ||
      signal(SIGTERM, sig_handler) == SIG_ERR ||
      signal(SIGHUP, sig_handler) == SIG_ERR)
  {
    fprintf(stderr, "Installing signal handlers failed\n");
    result = 2;
    goto m_fail;
  }

  /* Setup rin hardware (i2c, rinblocks) */
  if (hwlib_local_start(cmdline_rb_i2c_device, cmdline_rin_hw_device) < 0)
  {
    result = 2;
    goto m_fail;
  }
  hw_initd = true;

  /* Setup ADS12126 */
  adc_spi_fd = adc24bit_local_init(cmdline_ads1216_spi_device, cmdline_diode_microamps, cmdline_sample_frequency);

  if (cmdline_dumpreadings)
  {
    test_adc_temperature2(adc_spi_fd, cmdline_logtime);
  }

  result = 0;

m_fail:
  if (!cmdline_showhelp)
    printf("Shutting down\n");
  if (adc_spi_fd >= 0)
    close(adc_spi_fd);
  if (hw_initd)
    hwlib_stop();
  return result;
}
