/*
 * adcthread.c
 *
 *  Created on: 08/03/2011
 *      Author: ian
 */
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <rinhw/ucos_ii.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcdatabase.h>
#include <rinhw/adc_ads1216.h>
#include <rinhw/adc_up.h>
#include <rinweigh/adcmuteximpl_pthreads.h>
#include <rinweigh/adcliveimpl_simple.h>
#include <rinweigh/adcdbdefimpl_std.h>
#include <rinweigh/adccalcountcheck_std.h>
#include <rinhw/rinthread.h>
#include "devices.h"
#include "app_db_factory.h"
#include "app_db_user.h"
#include "app_db_runtime.h"
#include "app_db.h"

#include "adcthread.h"


/** Used to force the ADC thread to quit */
static bool adc_thread_quit = false;

/** Used to wait until the ADC thread is up and running */
static bool adc_thread_running = false;

/** The ADC thread */
static pthread_t adc_pthread;

/** Structure containing all ADC data */
AdcSuperStructure adcSuperStructure;


static bool AdcUpdateTemperature(AdcSuperStructure *ass, int32u *adc_temp)
{
  bool result = false;
  int16u temp16;

  if (ropen(&dev_adc, 0) >= 0)
  {
    if (rlseek(&dev_adc, sizeof(int16u) * ADC_CH_TEMP, SEEK_SET) == sizeof(int16u) * ADC_CH_TEMP)
    {
      if (rread(&dev_adc, &temp16, sizeof(int16u), 0) == sizeof(int16u))
      {
        *adc_temp = temp16;
        result = true;
      }
    }
    rclose(&dev_adc);
  }
  return result;
}


static bool AdcGetCurrentDateTime(AdcSuperStructure *ass, t_datetime *dt)
{
  bool result = false;

  if (ropen(&dev_rtc_test, 0) >= 0)
  {
    if (sizeof(t_datetime) == rread(&dev_rtc_test, dt, sizeof(t_datetime), 0))
    {
      result = true;
    }
    rclose(&dev_rtc_test);
  }
  return result;
}


static void AppAdcDebugVPrintf(int priority, const char *format, va_list args)
{
  (void) vfprintf(stdout, format, args);
}


static void *adc_thread(void *data)
{
  int16u applicErrors = 0;
  bool data_wait_failed;
  int16u ticks;
  int16u freq;
  AdcMutexImplementer *mutex = NULL;
  AdcLiveWeighImplementer *liveweigh = (AdcLiveWeighImplementer *) data;
  bool init_success = false;

#if 0
#if WE_DEBUG(WE_ADC_THREAD_DEBUG_LVL, WE_DEBUG_FULL)
  struct s_loop_timing lt;

  loop_timing_init(&lt, 1);
#endif
#endif

  /* Initialise the ADC library */
  printf("adc_thread: started\n");
  mutex = AdcMutexImplementerPthreadsCreate();
  if (NULL == mutex)
  {
    goto at_fail;
  }

  if (!AdcPrepare(&adcSuperStructure,
                  &dev_adc24bit,
                  AdcUpdateTemperature,
                  AdcGetCurrentDateTime,
                  AdcStandardDatabaseDefaultImplementer,
                  app_save_databases,
                  NULL,
                  AdcStandardDatabaseCalCountCheckImplementer,
                  mutex,
                  liveweigh,
                  AppAdcDebugVPrintf))
  {
    goto at_fail;
  }

  if (!AdcInit(&adcSuperStructure,
               &db_app_factory.adc_factory_database,
               &db_app_user.adc_user_database,
               &db_app_runtime.adc_runtime_database))
  {
    goto at_fail;
  }
  init_success = true;
  adc_thread_running = true;

  printf("adc_thread: initialisation complete\n");

  /* Thread loop */
  while (!adc_thread_quit)
  {
    /* Run ADC hander */
    (void) AdcUpdate(&adcSuperStructure, true, applicErrors, &data_wait_failed);
    if (data_wait_failed)
    {
      /*
       * Something didn't work, delay so we don't use 100% CPU
       * AdcSyncFrequencies: 100 == 10Hz
       */
      freq = (adcSuperStructure.adcUserDatabase.syncFreqIdx < ADC_NUM_SYNC_FREQ) ?
             AdcSyncFrequencies[adcSuperStructure.adcUserDatabase.syncFreqIdx] : AdcSyncFrequencies[ADC_SYNC_FREQ_50_HZ];
      ticks = (OS_TICKS_PER_SEC * 10) / freq;
      if (0 == ticks)
        ticks = 1;
      OSTimeDly(ticks);
    }

#if 0
    /* Debug readings per second */
#if WE_DEBUG(WE_ADC_THREAD_DEBUG_LVL, WE_DEBUG_FULL)
    loop_timing_check(&lt, "adc_thread");
#endif
#endif
  }

  /* Clean up after ADC library */
at_fail:
  if (init_success)
  {
    AdcDestroy(&adcSuperStructure);
  }
  AdcMutexImplementerPthreadsDestroy(mutex);

  printf("adc_thread: finished\n");
  rin_thread_exit(NULL);
}


/** Start the ADC thread */
int adc_thread_init(AdcLiveWeighImplementer *liveweigh)
{
  int res;
  uint32_t start_time;

  printf("adc_thread_init: called\n");
  adc_thread_running = false;

  res = rin_thread_create(&adc_pthread, adc_thread, liveweigh, true);

  /* Wait until the ADC thread starts */
  if (0 == res)
  {
    start_time = OSTimeGet();
    while (!adc_thread_running && OSTimeGet() < start_time + OS_TICKS_PER_SEC)
    {
      OSTimeDly(OS_TICKS_PER_SEC/10);
    }
  }

  printf("adc_thread_init: finished, res %d, thread running %d\n", res, adc_thread_running);
  return (0 == res && adc_thread_running) ? RIO_SUCCESS : RIO_ERROR;
}


/** Shut down the ADC thread */
void adc_thread_destroy(void)
{
  printf("adc_thread_destroy: called\n");

  /* Notify thread to quit */
  adc_thread_quit = true;

  /* Wait for thread to quit */
  pthread_join(adc_pthread, NULL);

  printf("adc_thread_destroy: finished\n");
}
