/*
 * dsdtest_add.c
 *
 *  Created on: Mar 7, 2012
 *      Author: ian
 */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <cutest/CuTest.h>
#include <rinweigh/adcdsd.h>
#include "dsdtest.h"

#include "dsdtest_add.h"

#define NUM_ADD_RECORDS 10


/** Checks on an empty DSD, empty reads, etc */
static void test_dsd_empty(CuTest *tc)
{
  int32u numrecords;
  enum eAdcDsdError res;
  int8u buffer[DSD_TEST_RECORD_SIZE];

  /* Get number of records, should be zero */
  res = AdcDSDGetNumberRecords(test_dsd, true, &numrecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, 0, numrecords);

  /* Try to read first record */
  res = AdcDSDReadRecord(test_dsd, true, 0, buffer, DSD_TEST_RECORD_SIZE);
  CuAssertIntEquals(tc, ADC_DSD_ERROR_EMPTY, res);

  /* Or any other */
  res = AdcDSDReadRecord(test_dsd, true, 3, buffer, DSD_TEST_RECORD_SIZE);
  CuAssertIntEquals(tc, ADC_DSD_ERROR_EMPTY, res);
}


/** Add records to the DSD */
static void test_dsd_add(CuTest *tc)
{
  enum eAdcDsdError res;
  int8u buffer[DSD_TEST_RECORD_SIZE];
  int recctr;
  int32u beforerecords;
  int32u afterrecords;

  /* Get before number of records */
  res = AdcDSDGetNumberRecords(test_dsd, true, &beforerecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Add the records */
  for (recctr = 0; recctr < NUM_ADD_RECORDS; recctr++)
  {
    memset(buffer, 'a' + recctr, DSD_TEST_RECORD_SIZE);
    res = AdcDSDWriteRecord(test_dsd, true, buffer, DSD_TEST_RECORD_SIZE);
    CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  }

  /* Get after number of records, and check how many where added */
  res = AdcDSDGetNumberRecords(test_dsd, true, &afterrecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, NUM_ADD_RECORDS, afterrecords - beforerecords);
}


/** Check whether the add records to the DSD worked */
static void test_dsd_add_check(CuTest *tc)
{
  int8u buffer[DSD_TEST_RECORD_SIZE];
  int32u recctr;
  int i;
  int32u numrecords;
  enum eAdcDsdError res;

  CuAssertPtrNotNull(tc, test_dsd);

  /* How many records */
  res = AdcDSDGetNumberRecords(test_dsd, true, &numrecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Check each record */
  for (recctr = 0; recctr < numrecords; recctr++)
  {
    res = AdcDSDReadRecord(test_dsd, true, recctr, buffer, DSD_TEST_RECORD_SIZE);
    CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

    /* Verify data */
    for (i = 0; i < DSD_TEST_RECORD_SIZE; i++)
    {
      CuAssertIntEquals(tc, 'a' + recctr, buffer[i]);
    }
  }

  /* Now try to read past last record */
  res = AdcDSDReadRecord(test_dsd, true, numrecords, buffer, DSD_TEST_RECORD_SIZE);
  CuAssertIntEquals(tc, ADC_DSD_ERROR_OUTOFRANGE, res);

  /* And more records than maximum */
  res = AdcDSDReadRecord(test_dsd, true, DSD_TEST_MAX_RECORDS + 1, buffer, DSD_TEST_RECORD_SIZE);
  CuAssertIntEquals(tc, ADC_DSD_ERROR_OUTOFRANGE, res);
}


/** Try to add a too big record to the DSD */
static void test_dsd_add_too_big(CuTest *tc)
{
  enum eAdcDsdError res;
  int8u buffer[DSD_TEST_RECORD_SIZE + 1];
  int32u beforerecords;
  int32u afterrecords;

  /* Get before number of records */
  res = AdcDSDGetNumberRecords(test_dsd, true, &beforerecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  memset(buffer, 'a', DSD_TEST_RECORD_SIZE + 1);
  res = AdcDSDWriteRecord(test_dsd, true, buffer, DSD_TEST_RECORD_SIZE + 1);
  CuAssertIntEquals(tc, ADC_DSD_ERROR_RECORDSIZE, res);

  /* Get after number of records, should be the same */
  res = AdcDSDGetNumberRecords(test_dsd, true, &afterrecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, beforerecords, afterrecords);
}


static int get_scaled_rand(int max)
{
  int num;
  unsigned long long ullnum;

  num = rand();
  ullnum = num;
  ullnum *= max;
  ullnum /= RAND_MAX;

  return (int) ullnum;
}


/** Fill the DSD (with random data, but fixed size records) */
void test_dsd_fill(CuTest *tc)
{
  enum eAdcDsdError res;
  int8u buffer[DSD_TEST_RECORD_SIZE];
  int ctr;
  int rec_size;
  int rec_data;
  int32u beforerecords;
  int32u afterrecords;
  int32u capacity;

  /* printf("Fill Start time: %lu\n", (unsigned long) time(NULL)); */

  /* Get before number of records */
  res = AdcDSDGetNumberRecords(test_dsd, true, &beforerecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  for (ctr = 0, res = ADC_DSD_SUCCESS; ADC_DSD_SUCCESS == res; ctr++)
  {
    rec_size = DSD_TEST_RECORD_SIZE;
    rec_data = get_scaled_rand('z' - ' ') + ' ';
    memset(buffer, rec_data, rec_size);
    res = AdcDSDWriteRecord(test_dsd, true, buffer, rec_size);
  }
  CuAssertIntEquals(tc, ADC_DSD_ERROR_FULL, res);

  /* Now check record numbers */
  res = AdcDSDGetNumberRecords(test_dsd, true, &afterrecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, beforerecords + ctr - 1, afterrecords);
  CuAssertIntEquals(tc, DSD_TEST_MAX_RECORDS, afterrecords);

  /* Check capacity */
  res = AdcDSDGetCapacity(test_dsd, true, &capacity);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, DSD_TEST_MAX_RECORDS, capacity);

  /* printf("Fill End time: %lu\n", (unsigned long) time(NULL)); */
}

#if 0
/** Verify the integrity of the DSD */
static void test_dsd_integrity(CuTest *tc)
{
  enum eAdcDsdError res;
  size_t record_num;

  printf("Integrity Start time: %lu\n", (unsigned long) time(NULL));

  /* DSD is fine at this point */
  res = AdcDSDVerifyIntegrity(test_dsd, false, &record_num);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  printf("Integrity End time: %lu\n", (unsigned long) time(NULL));

  printf("found %d records\n", record_num);

  /* Now trash it */

}
#endif


/** Tests for adding records to DSD */
CuSuite *CuGetSuite_DSD_Add(void)
{
  CuSuite *suite = CuSuiteNew();

  /* DSD empty at this point, check empty reads, etc */
  SUITE_ADD_TEST(suite, test_dsd_empty);

  /* Add records, then check them */
  SUITE_ADD_TEST(suite, test_dsd_add);
  SUITE_ADD_TEST(suite, test_dsd_add_check);

  /* Add a record that is too big */
  SUITE_ADD_TEST(suite, test_dsd_add_too_big);

  /* Fill the DSD entirely */
  SUITE_ADD_TEST(suite, test_dsd_fill);
#if 0
  /* Test the integrity of the DSD */
  SUITE_ADD_TEST(suite, test_dsd_integrity);
#endif

  return suite;
}
