/*
 * dsdtest_circular.c
 *
 * Testing the circular buffer works
 *
 *  Created on: Mar 16, 2012
 *      Author: ian
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <cutest/CuTest.h>
#include <rinweigh/adcdsd.h>
#include "dsdtest.h"

#include "dsdtest_circular.h"


/** Circular buffer test DSD */
static struct s_adc_dsd *circular_dsd = NULL;


/** Create a DSD structure for circular tests */
static void test_dsd_circular_create(CuTest *tc)
{
  enum eAdcDsdError res;

  /* Create DSD structure */
  res = AdcDSDInit(&circular_dsd, NULL);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Create DSD, file should be created */
  res = AdcDSDCreateNew(circular_dsd, true, DSD_TEST_FILE_CIRCULAR, DSD_TEST_RECORD_SIZE, DSD_TEST_MAX_RECORDS, DSD_TEST_RECORD_CHECKSUM_SEED, DSD_TEST_SERIAL_NUMBER);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
}


/** Destroy DSD structure used for circular tests */
static void test_dsd_circular_destroy(CuTest *tc)
{
  enum eAdcDsdError res;

  res = AdcDSDDestroy(circular_dsd);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
}


/** Add record to DSD */
static enum eAdcDsdError add_record(struct s_adc_dsd *dsd, char buf_fill_char)
{
  int8u buffer[DSD_TEST_RECORD_SIZE];

  memset(buffer, buf_fill_char, DSD_TEST_RECORD_SIZE);
  return AdcDSDWriteRecord(dsd, true, buffer, DSD_TEST_RECORD_SIZE);
}


/** Test the circular buffer where oldest record is before the newest record */
static void test_dsd_circular_oldest_before_newest(CuTest* tc)
{
  enum eAdcDsdError res;
  int32u capacity;
  int32u numrecords;
  int32u i;
  int32u recordsleft;

  /* Confirm indexes at start of physical records */
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_recordcount);

  /* Confirm functions, purge should fail, and indexes unaffected */
  res = AdcDSDGetNumberRecords(circular_dsd, true, &numrecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, 0, numrecords);
  res = AdcDSDPurge(circular_dsd, true, 1);
  CuAssertIntEquals(tc, ADC_DSD_ERROR_EMPTY, res);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_recordcount);

  /* How many to fill? */
  res = AdcDSDGetCapacity(circular_dsd, true, &capacity);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Add records till one short of full, watching newest advance */
  for (i = 0; i < capacity - 1; i++)
  {
    res = add_record(circular_dsd, 'a');
    CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

    /* Confirm indexes */
    CuAssertIntEquals(tc, i + 1, circular_dsd->file_index.dsd_newestindex);
    CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_oldestindex);
    CuAssertIntEquals(tc, i + 1, circular_dsd->file_index.dsd_recordcount);

    /* Confirm functions */
    res = AdcDSDGetNumberRecords(circular_dsd, true, &numrecords);
    CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
    CuAssertIntEquals(tc, i, numrecords - 1);
  }

  /* Now purge records, watching oldest advance, until empty */
  for (i = 0; i < capacity - 1; i++)
  {
    res = AdcDSDPurge(circular_dsd, true, 1);
    CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

    /* Confirm indexes: records left = capacity - 1 not filled - previous number removed (i) - 1 removed this iteration */
    recordsleft = capacity - 1 - i - 1;
    CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_newestindex);
    CuAssertIntEquals(tc, i + 1, circular_dsd->file_index.dsd_oldestindex);
    CuAssertIntEquals(tc, recordsleft, circular_dsd->file_index.dsd_recordcount);

    /* Confirm functions */
    res = AdcDSDGetNumberRecords(circular_dsd, true, &numrecords);
    CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
    CuAssertIntEquals(tc, recordsleft, numrecords);
  }

  /* Now no records left, with both indexes sitting at last physical record */
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_recordcount);
  res = AdcDSDGetNumberRecords(circular_dsd, true, &numrecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, 0, numrecords);

#if 0
  /* Now add one more, will be full, newest index will have wrapped */
  res = add_record(circular_dsd, 'a');
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, capacity, circular_dsd->file_index.dsd_recordcount);

  /* One more add will fail, and no change of indexes */
  res = add_record(circular_dsd, 'a');
  CuAssertIntEquals(tc, ADC_DSD_ERROR_FULL, res);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, capacity, circular_dsd->file_index.dsd_recordcount);
#endif
}


/** Test the circular buffer where the newest wraps from end of physical records to start of physical records */
static void test_dsd_circular_newest_wrap(CuTest* tc)
{
  enum eAdcDsdError res;
  int32u capacity;
  int32u numrecords;

  /* DSD capacity */
  res = AdcDSDGetCapacity(circular_dsd, true, &capacity);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Now no records left, with both indexes sitting at last physical record */
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_recordcount);

  /* Add a record, newest will wrap */
  res = add_record(circular_dsd, 'a');
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Confirm indexes */
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, 1, circular_dsd->file_index.dsd_recordcount);

  /* Confirm functions */
  res = AdcDSDGetNumberRecords(circular_dsd, true, &numrecords);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, 1, numrecords);
}


/** Test the circular buffer where oldest record is after the newest record */
static void test_dsd_circular_oldest_after_newest(CuTest* tc)
{
  enum eAdcDsdError res;
  int32u capacity;
  int32u numrecords;
  int32u i;

  /* DSD capacity */
  res = AdcDSDGetCapacity(circular_dsd, true, &capacity);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Now with 1 record, newest at start of physical records, oldest at last physical record */
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, 1, circular_dsd->file_index.dsd_recordcount);

  /* Add records until full */
  for (i = 1; i < capacity; i++)
  {
    res = add_record(circular_dsd, 'a');
    CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

    /* Confirm indexes */
    CuAssertIntEquals(tc, i, circular_dsd->file_index.dsd_newestindex);
    CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_oldestindex);
    CuAssertIntEquals(tc, i + 1, circular_dsd->file_index.dsd_recordcount);

    /* Confirm functions */
    res = AdcDSDGetNumberRecords(circular_dsd, true, &numrecords);
    CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
    CuAssertIntEquals(tc, i + 1, numrecords);
  }

  /* Another add will fail, because it's full, indexes unchanged */
  res = add_record(circular_dsd, 'a');
  CuAssertIntEquals(tc, ADC_DSD_ERROR_FULL, res);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, capacity, circular_dsd->file_index.dsd_recordcount);
}


/** Test the circular buffer where the oldest wraps from end of physical records to start of physical records */
static void test_dsd_circular_oldest_wrap(CuTest* tc)
{
  enum eAdcDsdError res;
  int32u capacity;
  int32u recordsleft;

  /* DSD capacity */
  res = AdcDSDGetCapacity(circular_dsd, true, &capacity);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Now full, with both indexes sitting at last physical record */
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, capacity, circular_dsd->file_index.dsd_recordcount);

  /* Purge 5 records, newest will wrap and advance 4 from the first physical record */
  recordsleft = capacity - 5;
  res = AdcDSDPurge(circular_dsd, true, 5);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, 4, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, recordsleft, circular_dsd->file_index.dsd_recordcount);

  /* Now try to purge too many, indexes unchanged */
  res = AdcDSDPurge(circular_dsd, true, recordsleft + 1);
  CuAssertIntEquals(tc, ADC_DSD_ERROR_OUTOFRANGE, res);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, 4, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, recordsleft, circular_dsd->file_index.dsd_recordcount);

  /* Now purge to empty file */
  res = AdcDSDPurge(circular_dsd, true, recordsleft);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_newestindex);
  CuAssertIntEquals(tc, capacity - 1, circular_dsd->file_index.dsd_oldestindex);
  CuAssertIntEquals(tc, 0, circular_dsd->file_index.dsd_recordcount);
}


/** Test the operation of the DSD record circular buffer */
CuSuite *CuGetSuite_DSD_CircularBuffer(void)
{
  CuSuite *suite = CuSuiteNew();

  /* Create a brand new DSD, so we know we are starting from the first physical record in the file */
  SUITE_ADD_TEST(suite, test_dsd_circular_create);

  /* Add our tests, order is important */
  SUITE_ADD_TEST(suite, test_dsd_circular_oldest_before_newest);
  SUITE_ADD_TEST(suite, test_dsd_circular_newest_wrap);
  SUITE_ADD_TEST(suite, test_dsd_circular_oldest_after_newest);
  SUITE_ADD_TEST(suite, test_dsd_circular_oldest_wrap);

  /* Must be last */
  SUITE_ADD_TEST(suite, test_dsd_circular_destroy);

  return suite;
}
