/*
 * dsdtest_init.c
 *
 *  Created on: Mar 7, 2012
 *      Author: ian
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <cutest/CuTest.h>
#include <rinweigh/adcdsd.h>
#include "dsdtest.h"

#include "dsdtest_init.h"


/** Create a DSD structure, and run tests */
static void dsd_init_and_test(CuTest* tc, struct s_adc_dsd **dsd)
{
  enum eAdcDsdError res;

  /* Create DSD structure */
  res = AdcDSDInit(dsd, NULL);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertPtrNotNull(tc, *dsd);
  CuAssertTrue(tc, (ADC_DSD_SUCCESS != AdcDSDIsRunning(*dsd, true)));
}


/** Test creating DSD structure */
static void test_dsd_init(CuTest* tc)
{
  struct s_adc_dsd *dummy_dsd = NULL;
  enum eAdcDsdError res;

  /* Create DSD structure */
  dsd_init_and_test(tc, &dummy_dsd);

  /* Destroy it */
  res = AdcDSDDestroy(dummy_dsd);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
}


/** Test creating DSD file if none exists */
static void test_dsd_create_nofile(CuTest* tc)
{
  enum eAdcDsdError res;
  struct stat statbuf;
  struct s_adc_dsd *dummy_dsd = NULL;

  /* Delete if existing */
  unlink(DSD_TEST_FILE_CREATE);

  /* Create DSD structure */
  dsd_init_and_test(tc, &dummy_dsd);

  /* Create DSD, file should be created */
  res = AdcDSDCreateNew(dummy_dsd, true, DSD_TEST_FILE_CREATE, DSD_TEST_RECORD_SIZE, DSD_TEST_MAX_RECORDS, DSD_TEST_RECORD_CHECKSUM_SEED, DSD_TEST_SERIAL_NUMBER);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  CuAssertIntEquals(tc, 0, stat(DSD_TEST_FILE_CREATE, &statbuf));
  CuAssertTrue(tc, S_ISREG(statbuf.st_mode));

  /* Close DSD file */
  res = AdcDSDClose(dummy_dsd, true);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Has it cleaned up after itself */
  CuAssertPtrEquals(tc, NULL, dummy_dsd->filename);
  CuAssertPtrEquals(tc, NULL, dummy_dsd->work_buffer);
  CuAssertTrue(tc, (dummy_dsd->fd < 0));
  CuAssertIntEquals(tc, 0, dummy_dsd->file_index.dsd_version);
  CuAssertIntEquals(tc, 0, dummy_dsd->file_index.dsd_maxrecords);
  CuAssertIntEquals(tc, 0, dummy_dsd->file_index.dsd_recordsize);
  CuAssertIntEquals(tc, 0, dummy_dsd->file_index.dsd_recordcrcseed);
  CuAssertIntEquals(tc, 0, dummy_dsd->file_index.dsd_recordcrcseed);
  CuAssertIntEquals(tc, 0, dummy_dsd->file_index.dsd_serialnumber);

  /* Now file exists, so try again with existing file */
  res = AdcDSDOpenExisting(dummy_dsd, true, DSD_TEST_FILE_CREATE);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);

  /* Check parameters read from file */
  CuAssertIntEquals(tc, 0, memcmp(dummy_dsd->file_index.dsd_magic, ADC_DSD_FILE_INDEX_MAGIC, ADC_DSD_FILE_INDEX_MAGIC_LENGTH));
  CuAssertIntEquals(tc, ADC_DSD_FILE_INDEX_VERSION_1, dummy_dsd->file_index.dsd_version);
  CuAssertIntEquals(tc, DSD_TEST_MAX_RECORDS, dummy_dsd->file_index.dsd_maxrecords);
  CuAssertIntEquals(tc, DSD_TEST_RECORD_SIZE + sizeof(struct s_adc_dsd_record_header), dummy_dsd->file_index.dsd_recordsize);
  CuAssertIntEquals(tc, DSD_TEST_RECORD_CHECKSUM_SEED, dummy_dsd->file_index.dsd_recordcrcseed);
  CuAssertIntEquals(tc, DSD_TEST_SERIAL_NUMBER, dummy_dsd->file_index.dsd_serialnumber);

  /* Clean up */
  res = AdcDSDDestroy(dummy_dsd);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
  unlink(DSD_TEST_FILE_CREATE);
}


/** Initialise and create a DSD structure & DSD file - this is the one the rest of the tests use */
static void test_dsd_create_for_tests(CuTest* tc)
{
  enum eAdcDsdError res;

  /* Delete if existing, so we start from a known place */
  unlink(DSD_TEST_FILE);

  /* Create dsd structure */
  dsd_init_and_test(tc, &test_dsd);

  /* Now create file */
  res = AdcDSDCreateNew(test_dsd, true, DSD_TEST_FILE, DSD_TEST_RECORD_SIZE, DSD_TEST_MAX_RECORDS, DSD_TEST_RECORD_CHECKSUM_SEED, DSD_TEST_SERIAL_NUMBER);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
}


/** Destroy the DSD that was used for tests */
static void test_dsd_destroy_for_tests(CuTest* tc)
{
  enum eAdcDsdError res;

  res = AdcDSDDestroy(test_dsd);
  CuAssertIntEquals(tc, ADC_DSD_SUCCESS, res);
}


/** Test initialisation of DSD */
CuSuite *CuGetSuite_DSD_Init(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_dsd_init);
  SUITE_ADD_TEST(suite, test_dsd_create_nofile);
  SUITE_ADD_TEST(suite, test_dsd_create_for_tests);

  return suite;
}


/** Test destruction of DSD */
CuSuite *CuGetSuite_DSD_Destroy(void)
{
  CuSuite *suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_dsd_destroy_for_tests);

  return suite;
}
