/*
 * menu_liveweigh.c
 *
 *  Created on: Jun 26, 2012
 *      Author: ian
 */

#include <stdio.h>
#include <rinhw/stdrinstrum.h>
#include <rinweigh/adcinterface.h>
#include <rinweigh/adcliveimpl.h>
#include <rinweigh/adcliveimpl_simple.h>
#include <rinweigh/adcliveimpl_improved.h>
#include "getchar.h"
#include "adcthread.h"
#include "print_rdg.h"
#include "cmdline.h"
#include "menu.h"

#include "menu_liveweigh.h"


/** Create the appropriate live weigh implementation */
AdcLiveWeighImplementer *create_live_weigh_implementation(int liveweigh_algorithm)
{
  AdcLiveWeighImplementer *liveweigh = NULL;

  switch (liveweigh_algorithm)
  {
    default:
    case 0:
      liveweigh = NULL;
      break;
    case 1:
      liveweigh = AdcLiveWeighImplementerSimpleCreate();
      break;
    case 2:
      liveweigh = AdcLiveWeighImplementerImprovedCreate();
      break;
  }
  return liveweigh;
}


static void configure_simple_liveweigh(AdcLiveWeighImplementer *liveweigh)
{
  int8u result;
  int32u averagingMilliseconds;
  int i;

  printf("\nEnter averaging time (milliseconds) (0-20000): ");
  if (!get_int(&i, 0, 20000))
    goto csl_fail;
  averagingMilliseconds = i;

  printf("Configuring Simple Live Weigh with:\n"
         "  averagingMilliseconds: %lu\n"
         "Live weigh will take %lu milliseconds\n",
         (unsigned long) averagingMilliseconds,
         (unsigned long) averagingMilliseconds
        );
  AdcMutexLock(&adcSuperStructure);
  result = AdcLiveWeighImplementerSimpleConfigure(&adcSuperStructure, liveweigh, averagingMilliseconds);
  AdcMutexUnlock(&adcSuperStructure);
  print_adc_operation_error(stdout, result);
  return;

csl_fail:
  printf("\nLiveweigh configuration aborted\n");
}


static void default_configure_simple_liveweigh(AdcLiveWeighImplementer *liveweigh)
{
  int8u result;
  int32u averagingMilliseconds;

  averagingMilliseconds = 5000;

  printf("Configuring Simple Live Weigh with:\n"
         "  averagingMilliseconds: %lu\n"
         "Live weigh will take %lu milliseconds\n",
         (unsigned long) averagingMilliseconds,
         (unsigned long) averagingMilliseconds
        );

  AdcMutexLock(&adcSuperStructure);
  result = AdcLiveWeighImplementerSimpleConfigure(&adcSuperStructure, liveweigh, averagingMilliseconds);
  AdcMutexUnlock(&adcSuperStructure);
  print_adc_operation_error(stdout, result);
}


static void configure_improved_liveweigh(AdcLiveWeighImplementer *liveweigh)
{
  int32u sampleFrequency;             /*!< The sample frequency that the live weigh is called at. In Hz * 10 (ie 100 = 10Hz) */
  int32u holdOffPresetMilliseconds;   /*!< The number of milliseconds to wait before collecting samples after weight has gone above target */
  size_t primaryAvNumRdgs;            /*!< Number of primary readings to average @ ADC sample frequency */
  size_t secondaryAvNumRdgs;          /*!< The number of secondary live averages */
  int32s target;                      /*!< Weight must go above this to trigger live weighing */
  int8u tolPreset;
  int8u retrigPercentPreset;          /*!< (0-100) When the weight changes by this percentage of the last average, weighing will be re-triggered */
  int i;
  int8u result;

  sampleFrequency = AdcSyncFrequencies[adcSuperStructure.adcUserDatabase.syncFreqIdx];
  printf("\nSync frequency is %lu\n", (unsigned long) sampleFrequency / 10);

  printf("\nEnter hold off (milliseconds) (0-20000): ");
  if (!get_int(&i, 0, 20000))
    goto cil_fail;
  holdOffPresetMilliseconds = i;

  printf("\nLive weigh will take (primary_av_samples * secondary_av_samples) / sync_freq");

  printf("\nEnter number of primary average samples (1-100): ");
  if (!get_int(&i, 1, 100))
    goto cil_fail;
  primaryAvNumRdgs = i;

  printf("\nEnter number of secondary average samples (1-100) (sync freq is good): ");
  if (!get_int(&i, 1, 100))
    goto cil_fail;
  secondaryAvNumRdgs = i;

  printf("\nEnter target (gross weight must be greater than this) (0-999999): ");
  if (!get_int(&i, 0, 999999))
    goto cil_fail;
  target = i;

  printf("\nEnter number of biggest readings to discard (0 - #secondary av rdgs / 2 - 1 (%lu)): ", (unsigned long) secondaryAvNumRdgs / 2 - 1);
  if (!get_int(&i, 0, secondaryAvNumRdgs / 2 - 1))
    goto cil_fail;
  tolPreset = i;

  printf("\nEnter re-trigger percentage (0-100): ");
  if (!get_int(&i, 0, 100))
    goto cil_fail;
  retrigPercentPreset = i;

  AdcMutexLock(&adcSuperStructure);
  result = AdcLiveWeighImplementerImprovedConfigure(&adcSuperStructure,
                                                  liveweigh,
                                                  sampleFrequency,
                                                  holdOffPresetMilliseconds,
                                                  primaryAvNumRdgs,
                                                  secondaryAvNumRdgs,
                                                  target,
                                                  tolPreset,
                                                  retrigPercentPreset);
  AdcMutexUnlock(&adcSuperStructure);
  print_adc_operation_error(stdout, result);
  return;

cil_fail:
  printf("\nliveweigh configuration aborted\n");
}


static void default_configure_improved_liveweigh(AdcLiveWeighImplementer *liveweigh)
{
  int32u sampleFrequency;             /*!< The sample frequency that the live weigh is called at. In Hz * 10 (ie 100 = 10Hz) */
  int32u holdOffPresetMilliseconds;   /*!< The number of milliseconds to wait before collecting samples after weight has gone above target */
  size_t primaryAvNumRdgs;            /*!< Number of primary readings to average @ ADC sample frequency */
  size_t secondaryAvNumRdgs;          /*!< The number of secondary live averages */
  int32s target;                      /*!< Weight must go above this to trigger live weighing */
  size_t secondaryAvNumDiscard;
  int8u retrigPercentPreset;          /*!< (0-100) When the weight changes by this percentage of the last average, weighing will be re-triggered */
  int i;
  int8u result;

  sampleFrequency = AdcSyncFrequencies[adcSuperStructure.adcUserDatabase.syncFreqIdx];
  holdOffPresetMilliseconds = 2000;
  primaryAvNumRdgs = 2;
  secondaryAvNumRdgs = sampleFrequency / 10;
  secondaryAvNumDiscard = secondaryAvNumRdgs / 10;

  printf("\nEnter target (gross weight must be greater than this) (0-999999): ");
  if (!get_int(&i, 0, 999999))
    goto dcil_fail;
  target = i;

  retrigPercentPreset = 0;

  printf("Configuring Improved Live Weigh with:\n"
         "  sampleFrequency: %lu\n"
         "  holdOffPresetMilliseconds: %lu\n"
         "  primaryAvNumRdgs: %lu\n"
         "  secondaryAvNumRdgs: %lu\n"
         "  target: %ld\n"
         "  secondaryAvNumDiscard: %lu\n"
         "  retrigPercentPreset: %lu\n"
         "Live weigh will take %lu milliseconds\n",
         (unsigned long) sampleFrequency / 10,
         (unsigned long) holdOffPresetMilliseconds,
         (unsigned long) primaryAvNumRdgs,
         (unsigned long) secondaryAvNumRdgs,
         (long) target,
         (unsigned long) secondaryAvNumDiscard,
         (unsigned long) retrigPercentPreset,
         (unsigned long) holdOffPresetMilliseconds + (primaryAvNumRdgs * secondaryAvNumRdgs * 1000) / (sampleFrequency / 10)
        );

  printf("Live weigh will take (primary_av_samples * secondary_av_samples) / sync_freq\n");

  AdcMutexLock(&adcSuperStructure);
  result = AdcLiveWeighImplementerImprovedConfigure(&adcSuperStructure,
                                                  liveweigh,
                                                  sampleFrequency,
                                                  holdOffPresetMilliseconds,
                                                  primaryAvNumRdgs,
                                                  secondaryAvNumRdgs,
                                                  target,
                                                  secondaryAvNumDiscard,
                                                  retrigPercentPreset);
  AdcMutexUnlock(&adcSuperStructure);
  print_adc_operation_error(stdout, result);
  return;

dcil_fail:
  printf("\nLiveweigh configuration aborted\n");
}


/** Run a live weigh */
static void initiate_live_weigh(void)
{
  int32u start_time;
  int32u end_time;

  printf("Waiting for live weigh to complete\n");
  start_time = OSTimeGet();
  print_adc_operation_error(stdout, AdcLivestockWeigh(&adcSuperStructure, true, true, false));
  end_time = OSTimeGet();
  AdcMutexLock(&adcSuperStructure);
  if (ADC_OP_STATUS_SUCCESS == AdcGetCurrentOpStatus(&adcSuperStructure, false))
  {
    print_adc_data(&adcSuperStructure.adcHeldData);
  }
  AdcMutexUnlock(&adcSuperStructure);
  printf("Operation took %lu milliseconds\n", (unsigned long) (end_time - start_time) * 1000 / OS_TICKS_PER_SEC);
}


/** Run the live weigh menu */
void adc_liveweigh_menu(AdcLiveWeighImplementer *liveweigh)
{
  char ch;
  bool quit;

  for (quit = false; !quit; )
  {
    printf("\nADC Library Live Weigh\n");

    if (1 == cmdline_liveweigh_algorithm)
      printf("Configured with simple live weigh algorithm\n");
    else if (2 == cmdline_liveweigh_algorithm)
      printf("Configured with improved live weigh algorithm\n");
    else
      printf("No live weigh algorithm configured\n");

    printf("\n\n"
           " c - Configure live weigh\n"
           " d - Default configuration of live weigh\n"
           " i - Initiate a live weigh\n"
           " q - Quit\n"
          );

    get_stdin_char_block(&ch);
    switch (ch)
    {
      /* Configure the live weigh */
      case 'c':
        if (1 == cmdline_liveweigh_algorithm)
          configure_simple_liveweigh(liveweigh);
        else if (2 == cmdline_liveweigh_algorithm)
          configure_improved_liveweigh(liveweigh);
        else
          printf("No live weigh configured\n");
        break;

      /* Default configuration of live weigh */
      case 'd':
        if (1 == cmdline_liveweigh_algorithm)
          default_configure_simple_liveweigh(liveweigh);
        else if (2 == cmdline_liveweigh_algorithm)
          default_configure_improved_liveweigh(liveweigh);
        else
          printf("No live weigh configured\n");
        break;

      /* Initiate a live weigh */
      case 'i':
        initiate_live_weigh();
        break;

      /* Quit */
      case 'q':
        quit = true;
        break;
    }
  }
}
