/*
 * print_rdg.c
 *
 *  Created on: 29/03/2011
 *      Author: ian
 */
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <rinweigh/adcdatatypes.h>
#include <rinweigh/adcinterface.h>

#include "print_rdg.h"


struct s_adc_errors
{
  int16u mask;
  const char *description;
};


#define ERRSTR(s) s, #s
#define TOSTR(s) #s

static const struct s_adc_errors adc_errors[] =
{
    { ERRSTR(ADC_ERR_FACTORY_DB) },
    { ERRSTR(ADC_ERR_USER_DB) },
    { ERRSTR(ADC_ERR_ADC_MAXED) },
    { ERRSTR(ADC_ERR_RUNTIME_DB) },
    { 0, NULL }
};


static const char *const adc_operation_errors[] =
{
    TOSTR(ADC_OP_STATUS_SUCCESS),
    TOSTR(ADC_OP_STATUS_CANCELED),
    TOSTR(ADC_OP_STATUS_IN_PROGRESS),
    TOSTR(ADC_OP_STATUS_FAILED_ERROR),
    TOSTR(ADC_OP_STATUS_FAILED_OL_UL),
    TOSTR(ADC_OP_STATUS_FAILED_BUSY),
    TOSTR(ADC_OP_STATUS_FAILED_MOTION),
    TOSTR(ADC_OP_STATUS_FAILED_BAND),
    TOSTR(ADC_OP_STATUS_FAILED_RES),
    TOSTR(ADC_OP_STATUS_FAILED_NOT_IMP),
    TOSTR(ADC_OP_STATUS_FAILED_PNT_DUPL),
    TOSTR(ADC_OP_STATUS_FAILED_HI_RES),
    TOSTR(ADC_OP_STATUS_FAILED_PRNID_MAX),
    TOSTR(ADC_OP_STATUS_FAILED_DATE_TIME),
    TOSTR(ADC_OP_STATUS_FAILED_PARAM),
    TOSTR(ADC_OP_STATUS_FAILED_TRADE),
};


void print_adc_errors(FILE *f, int16u errorbits)
{
  int i;

  if (0 != errorbits)
  {
    fprintf(f, "Errors: 0x%04x: ", errorbits);
    for (i = 0; adc_errors[i].description != NULL; i++)
    {
      if (errorbits & adc_errors[i].mask)
      {
        fprintf(f, "%s ", adc_errors[i].description);
      }
    }
    fprintf(f, "\n");
  }
  else
  {
    fprintf(f, "No Errors: 0x%04x\n", errorbits);
  }
}


void print_adc_operation_error(FILE *f, int8u opresult)
{
  if (opresult < ADC_OP_STATUS_NUM)
  {
    fprintf(f, "Result: %s\n", adc_operation_errors[opresult]);
  }
  else
  {
    fprintf(f, "Result: (unknown) %d\n", opresult);
  }
}


void print_adc_rdg_status(FILE *f, const AdcStatus *stat)
{
  fprintf(f, "GradPos:%d", stat->bit.gradPos);
  fprintf(f, ",GorN:%d", stat->bit.grossOrNet);
  if (stat->bit.hiRes)
    fprintf(f, ",HiRes");
  if (stat->bit.multiRange)
    fprintf(f, ",Range:%d", stat->bit.range + 1);
  else
    fprintf(f, ",SingleRange");
  if (stat->bit.held)
    fprintf(f, ",Held");
  if (stat->bit.zeroRange)
    fprintf(f, ",ZeroRange");
  if (stat->bit.zeroBand)
    fprintf(f, ",ZeroBand");
  if (stat->bit.motion)
    fprintf(f, ",Motion");
  if (stat->bit.coz)
    fprintf(f, ",CenterOfZero");
  if (stat->bit.grossOrNet)
    fprintf(f, ",%s", (stat->bit.gross) ? "Gross" : "Net");
  if (stat->bit.uload)
    fprintf(f, ",Underload");
  if (stat->bit.oload)
    fprintf(f, ",Overload");
  if (stat->bit.tare)
    fprintf(f, ",Tare");
  if (stat->bit.preTare)
    fprintf(f, ",PresetTare");
}


void print_adc_weight_reading(FILE *f, const AdcRdg *rdg)
{
  int32s whole;
  int32s fraction;
  int decimal_divisor;
  int dp;

  for (dp = 0, decimal_divisor = 1; dp < rdg->info.unitdp.bit.dp; dp++)
  {
    decimal_divisor *= 10;
  }
  whole = rdg->rdg / decimal_divisor;
  fraction = rdg->rdg % decimal_divisor;
  if (fraction < 0)
  {
    fraction = -fraction;
  }

  if (rdg->info.unitdp.bit.dp > 0)
  {
    fprintf(f, "%d.%0*d", (int) whole, rdg->info.unitdp.bit.dp, (int) fraction);
  }
  else
  {
    fprintf(f, "%d", (int) whole);
  }
}


const char AlibiUnitsStr[ADC_NUM_UNITS][3] =
{
  "  ",
  "g ",
  "kg",
  "t ",
  "lb",
  "oz",
  "N ",
  "P ",
  "U0",
  "U1",
  "U2",
  "U3",
  "U4",
  "U5",
  "U6",
  "U7"
};


void print_adc_weight_units(FILE *f, const AdcUnitDp *unitdp)
{
  fprintf(f, "%s", AlibiUnitsStr[unitdp->bit.units]);
}


/** Of the form "Rdg: 18.20 kg" */
void print_adc_simple_reading(FILE *f, const AdcRdg *rdg)
{
  fprintf(f, "Rdg: ");
  print_adc_weight_reading(f, rdg);
  fprintf(f, " ");
  print_adc_weight_units(f, &rdg->info.unitdp);
  fprintf(f, "\n");
}


/** Of the form "Id: 0, Rdg: 18.20 kg GradPos:6,GorN:1,Range:1,Gross" */
void print_adc_full_reading(FILE *f, const AdcRdg *rdg)
{
  fprintf(f, "Id: %d, Rdg: ", rdg->info.id);
  print_adc_weight_reading(f, rdg);
  fprintf(f, " ");
  print_adc_weight_units(f, &rdg->info.unitdp);
  fprintf(f, " ");
  print_adc_rdg_status(f, &rdg->info.status);
  fprintf(f, "\n");
}


void print_adc_data(const AdcData *d)
{
  printf("Data Valid: %u\n", d->dataValid);                                                 /* Is data in structure valid */
  printf("Raw Reading: %ld\n", (long) d->rawRdg);                                           /* The raw reading */
  printf("Milli-volt Per Volt Abs Reading: %ld\n", (long) d->mvvAbsRdg);                    /* The millivolt per volt absolute reading */

  printf("Display Reading: "); print_adc_full_reading(stdout, &d->dispRdg);                 /* The displayed gross or net rdg in calibrated units */
  printf("Non-Display Reading: "); print_adc_full_reading(stdout, &d->notDispRdg);          /* The non-displayed weight (ie gross, if net is currently displayed */

  print_adc_errors(stdout, d->error);                                                       /* Any errors, refer to the ADC_ERR_* flags */
  printf("Reading Id: %lu\n", (unsigned long) d->rdgId);                                    /* Reading identifier, incremented every new reading */
  printf("Gross Reading: "); print_adc_full_reading(stdout, &d->grossRdg);                  /* The weight as a gross reading */
  printf("Net Reading: "); print_adc_full_reading(stdout, &d->netRdg);                      /* The weight as a net reading */

  printf("Peak Abs. Gross Reading: "); print_adc_full_reading(stdout, &d->peakAbsGrossRdg); /* The peak absolute gross reading, used in peak hold mode */
  printf("Peak Max. Disp. Reading: "); print_adc_full_reading(stdout, &d->peakMaxDispRdg);  /* The peak maximum displayed reading, used in peak hold mode */
  printf("Peak Min. Disp. Reading: "); print_adc_full_reading(stdout, &d->peakMinDispRdg);  /* The peak minimum displayed reading, used in peak hold mode */

  printf("Preset Tare: "); print_adc_full_reading(stdout, &d->presetTare);                  /* The current preset tare value */
  printf("User Tare: "); print_adc_full_reading(stdout, &d->userTare);                      /* The current user tare value */
  printf("Any Tare: "); print_adc_full_reading(stdout, &d->anyTare);                        /* The sum of the user and preset tares */

  printf("Display Mode: %u\n", d->dispMode);                                                /* How the data should be displayed */

  printf("Pieces: "); print_adc_full_reading(stdout, &d->dispPieceRdg);                     /* The displayed number of pieces, when in peice counting mode */

  printf("Alt. Units Gross Reading: "); print_adc_full_reading(stdout, &d->altGrossRdg);    /* The alternate weight units gross reading */
  printf("Alt. Units Net Reading: "); print_adc_full_reading(stdout, &d->altNetRdg);        /* The alternate weight units net reading */
  printf("Alt. Units Display Reading: "); print_adc_full_reading(stdout, &d->altDispRdg);   /* The alternate weight units displayed reading */
  printf("Alt. Units Any Tare: "); print_adc_full_reading(stdout, &d->altAnyTare);          /* Sum of any tare(s) present for the alternate weight units */
}
