/*
 * Program to get network configuration for an interface
 *
 * Invocation is "getnetconfig <iface>" - iface is "ethX"
 * Prints the following to stdout: <ip> <netmask> <broadcast> <gateway> <nameserver> <mac> <hostname>
 */
#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <getopt.h>
#include "lsnc_internal.h"
#include <setnetconfig.h>

#define PROGRAM_NAME "getnetconfig"

static const char *cmdline_interface = NULL;
static bool cmdline_help = false;
static bool cmdline_version = false;
static bool cmdline_detailed = false;
static const char *cmdline_method = SNC_METHOD_LIB_CONFIG;
static const char *cmdline_configfile = SNC_CONFIG_FILE;


static void print_usage(void)
{
  printf("Usage:\n"
         "  " PROGRAM_NAME " [options] <iface>\n"
         "\nOptions:\n"
         "  -v --version          Print the software version\n"
         "  -h --help             Print this help\n"
         "  -c --config <file>    Load the library configuration from this file\n"
         "                        Default " SNC_CONFIG_FILE "\n"
         "  -m --method <method>  Use <method> to gather information\n"
         "                        Methods are:\n"
         "                          libconf - Use the method set in config file (default)\n"
         "                          auto - Try each of the available modes until success\n"
         "                          rcconf - Get the network config from rc.conf\n"
#if defined(HAVE_LIBNM)
         "                          nm - Get the network config from Network Manager\n"
#endif
         "  -d --detailed         Print detailed configuration\n"
         "  iface                 The network interface to get the configuration for.\n"
         "                        Eg eth0\n"
         "\n"
         "Program to retrieve the currently configured network configuration\n"
         "Will print the following to stdout:\n"
         "  <ip> <netmask> <broadcast> <gateway> <nameserver> <mac> <hostname>\n"
         "Note that <mac> is not stored in the config file, it is simply returned\n"
         "as 00:00:00:00:00:00 for consistent output with getnetcurrent\n"
         );
}


static void print_version(void)
{
  printf(PROGRAM_NAME " " PACKAGE_VERSION "\n");
}


static enum e_snc_result parse_opts(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "version",    0, 0, 'v' },
    { "help",       0, 0, 'h' },
    { "config",     1, 0, 'c' },
    { "method",     1, 0, 'm' },
    { "detailed",   0, 0, 'd' },
    { NULL,         0, 0, '\0' },
  };
  int c;

  while (1)
  {
    c = getopt_long(argc, argv, "vhc:m:d", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      case 'h':
        cmdline_help = true;
        break;

      case 'v':
        cmdline_version = true;
        break;

      case 'c':
        cmdline_configfile = optarg;
        break;

      case 'm':
        cmdline_method = optarg;
        break;

      case 'd':
        cmdline_detailed = true;
        break;

      default:
        fprintf(stderr, PROGRAM_NAME ": Bad command line option. Try --help.\n");
        return sncr_invalid_commandline;
        break;
    }
  }

  if (optind >= argc && !cmdline_help && !cmdline_version)
  {
    fprintf(stderr, PROGRAM_NAME ": No network interface specified. Try --help.\n");
    return sncr_invalid_commandline;
  }
  cmdline_interface = argv[optind];
  optind++;

  return sncr_success;
}


/** getnetconfig entry point */
#if ENABLE_MULTI_CALL
int getnetconfig_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
  struct s_snc_configuration *config = NULL;
  struct s_snc_net_config current_net_config;
  enum e_snc_result result;

  /* Check command line parameters */
  result = parse_opts(argc, argv);
  if (sncr_success != result)
    goto m_done;

  /* Help and version */
  if (cmdline_help)
  {
    print_usage();
    result = sncr_success;
    goto m_done;
  }
  if (cmdline_version)
  {
    print_version();
    result = sncr_success;
    goto m_done;
  }

  /* Try to load the library configuration */
  result = snc_config_init(&config);
  if (sncr_success != result)
    goto m_done;
  (void) snc_config_load(cmdline_configfile, config);

  /* Get the network configuration */
  result = snc_get_net_config(config, cmdline_method, cmdline_interface, &current_net_config);
  if (sncr_success != result)
    goto m_done;

  /* Print it out */
  if (cmdline_detailed)
    snc_print_net_config(stdout, false, &current_net_config);
  else
    snc_print_legacy_net_config_line(stdout, &current_net_config, true);

  /* Success */
  result = sncr_success;

m_done:
  SNC_DESTROY_AND_NULLIFY(snc_config_destroy, config);
  if (sncr_success != result)
    fprintf(stderr, PROGRAM_NAME ": failed with result %s\n", snc_result_to_string(result));
  return result;
}
