/*
 * lsnc_config.c
 *
 * Store or load the libsetnetconfig configuration from a file, usually /etc/libsetnetconfig.conf
 *
 *  Created on: 8 Jun. 2018
 *      Author: ian
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "lsnc_rcconf.h"
#include "lsnc_internal.h"

#include <setnetconfig.h>


#define SNC_CONFIG_NUM_PARAMS 9


static const char *const snc_config_varnames[SNC_CONFIG_NUM_PARAMS] =    { SNC_CONFIG_VARNAME_RESOLV_FILE, SNC_CONFIG_VARNAME_ROUTE_CMD, SNC_CONFIG_VARNAME_IFCONFIG_CMD, SNC_CONFIG_VARNAME_IP_CMD, SNC_CONFIG_VARNAME_RCCONF_FILE, SNC_CONFIG_VARNAME_CONFIG_METHOD, SNC_CONFIG_VARNAME_CURRENT_METHOD, SNC_CONFIG_VARNAME_NET_RECONFIGURE_CMD, SNC_CONFIG_VARNAME_HOSTNAME_RECONFIGURE_CMD };
static const char *const snc_config_vardefaults[SNC_CONFIG_NUM_PARAMS] = { SNC_CONFIG_RESOLV_FILE_DEFAULT, SNC_CONFIG_ROUTE_CMD_DEFAULT, SNC_CONFIG_IFCONFIG_CMD_DEFAULT, SNC_CONFIG_IP_CMD_DEFAULT, SNC_CONFIG_RCCONF_FILE_DEFAULT, SNC_CONFIG_CONFIG_METHOD_DEFAULT, SNC_CONFIG_CURRENT_METHOD_DEFAULT, SNC_CONFIG_NET_RECONFIGURE_CMD_DEFAULT, SNC_CONFIG_HOSTNAME_RECONFIGURE_CMD_DEFAULT };


/** Create a setnetconfig library configuration set to defaults */
enum e_snc_result snc_config_init(struct s_snc_configuration **config)
{
  struct s_snc_configuration *cfg;

  cfg = calloc(1, sizeof(*cfg));
  if (NULL == cfg)
    goto sci_fail;

  cfg->resolv_file = strdup(SNC_CONFIG_RESOLV_FILE_DEFAULT);
  if (NULL == cfg->resolv_file)
    goto sci_fail;
  cfg->route_cmd = strdup(SNC_CONFIG_ROUTE_CMD_DEFAULT);
  if (NULL == cfg->route_cmd)
    goto sci_fail;
  cfg->ifconfig_cmd = strdup(SNC_CONFIG_IFCONFIG_CMD_DEFAULT);
  if (NULL == cfg->ifconfig_cmd)
    goto sci_fail;
  cfg->ip_cmd = strdup(SNC_CONFIG_IP_CMD_DEFAULT);
  if (NULL == cfg->ip_cmd)
    goto sci_fail;
  cfg->rcconf_file = strdup(SNC_CONFIG_RCCONF_FILE_DEFAULT);
  if (NULL == cfg->rcconf_file)
    goto sci_fail;
  cfg->config_method = strdup(SNC_CONFIG_CONFIG_METHOD_DEFAULT);
  if (NULL == cfg->config_method)
    goto sci_fail;
  cfg->current_method = strdup(SNC_CONFIG_CURRENT_METHOD_DEFAULT);
  if (NULL == cfg->current_method)
    goto sci_fail;
  cfg->net_reconfigure_cmd = strdup(SNC_CONFIG_NET_RECONFIGURE_CMD_DEFAULT);
  if (NULL == cfg->net_reconfigure_cmd)
    goto sci_fail;
  cfg->hostname_reconfigure_cmd = strdup(SNC_CONFIG_HOSTNAME_RECONFIGURE_CMD_DEFAULT);
  if (NULL == cfg->hostname_reconfigure_cmd)
    goto sci_fail;

  /* Success */
  *config = cfg;
  return sncr_success;

sci_fail:
  SNC_DESTROY_AND_NULLIFY(snc_config_destroy, cfg);
  *config = NULL;
  return sncr_allocerror;
}


/** Destroy a setnetconfig library configuration */
enum e_snc_result snc_config_destroy(struct s_snc_configuration *config)
{
  if (NULL != config)
  {
    SNC_FREE_AND_NULLIFY(config->resolv_file);
    SNC_FREE_AND_NULLIFY(config->route_cmd);
    SNC_FREE_AND_NULLIFY(config->ifconfig_cmd);
    SNC_FREE_AND_NULLIFY(config->ip_cmd);
    SNC_FREE_AND_NULLIFY(config->rcconf_file);
    SNC_FREE_AND_NULLIFY(config->config_method);
    SNC_FREE_AND_NULLIFY(config->current_method);
    SNC_FREE_AND_NULLIFY(config->net_reconfigure_cmd);
    SNC_FREE_AND_NULLIFY(config->hostname_reconfigure_cmd);
    SNC_FREE_AND_NULLIFY(config);
  }
  return sncr_success;
}


/** Load the library configuration from a file */
enum e_snc_result snc_config_load(const char *config_path, struct s_snc_configuration *config)
{
  char *line = NULL;
  FILE *config_file = NULL;
  enum e_snc_result result;
  char * *const cfgvarptrs[SNC_CONFIG_NUM_PARAMS] = { &config->resolv_file, &config->route_cmd,
                                                      &config->ifconfig_cmd, &config->ip_cmd,
                                                      &config->rcconf_file, &config->config_method,
                                                      &config->current_method, &config->net_reconfigure_cmd,
                                                      &config->hostname_reconfigure_cmd };
  size_t param;
  char *value = NULL;

  line = malloc(SNC_PARSE_LINE_LENGTH);
  if (NULL == line)
  {
    result = sncr_allocerror;
    goto scl_fail;
  }

  value = malloc(PATH_MAX);
  if (NULL == value)
  {
    result = sncr_allocerror;
    goto scl_fail;
  }

  /* Open library config file for reading only */
  config_file = fopen(config_path, "r");
  if (NULL == config_file)
  {
    result = sncr_fileerror;
    goto scl_fail;
  }

  /* Now get each of the parameters for the interface index */
  for (param = 0; param < SNC_CONFIG_NUM_PARAMS; param++)
  {
    /* Free existing value */
    if (NULL != *cfgvarptrs[param])
    {
      free(*cfgvarptrs[param]);
      *cfgvarptrs[param] = NULL;
    }

    /* Find value in config file */
    result = snc_rcconf_find_variable(config_file, line, snc_config_varnames[param], value, PATH_MAX, NULL);
    switch (result)
    {
      case sncr_success:
        *cfgvarptrs[param] = strdup(value);
        if (NULL == *cfgvarptrs[param])
        {
          result = sncr_allocerror;
          goto scl_fail;
        }
        break;
      case sncr_notfound:
        *cfgvarptrs[param] = strdup(snc_config_vardefaults[param]);
        if (NULL == *cfgvarptrs[param])
        {
          result = sncr_allocerror;
          goto scl_fail;
        }
        break;
      default:
        goto scl_fail;
        break;
    }
  }

  /* Success */
  result = sncr_success;

scl_fail:
  SNC_FCLOSE_AND_NULLIFY(config_file);
  SNC_FREE_AND_NULLIFY(value);
  SNC_FREE_AND_NULLIFY(line);
  return result;
}


/** Save the library configuration from a file */
enum e_snc_result snc_config_save(const char *config_path, const struct s_snc_configuration *config)
{
  FILE *config_file = NULL;
  enum e_snc_result result;
  const char *const cfgvarvalue[SNC_CONFIG_NUM_PARAMS] = { config->resolv_file, config->route_cmd,
                                                           config->ifconfig_cmd, config->ip_cmd,
                                                           config->rcconf_file, config->config_method,
                                                           config->current_method, config->net_reconfigure_cmd,
                                                           config->hostname_reconfigure_cmd };
  size_t param;
  const char *value;

  /* Open library config file for writing */
  config_file = fopen(config_path, "w");
  if (NULL == config_file)
  {
    result = sncr_fileerror;
    goto scs_fail;
  }

  /* Now write each of the settings to the file */
  for (param = 0; param < SNC_CONFIG_NUM_PARAMS; param++)
  {
    value = (NULL != cfgvarvalue[param]) ? cfgvarvalue[param] : snc_config_vardefaults[param];
    if (fprintf(config_file, "%s=\"%s\"\n", snc_config_varnames[param], value) < 0)
    {
      result = sncr_fileerror;
      goto scs_fail;
    }
  }
  result = sncr_success;

scs_fail:
  SNC_FCLOSE_AND_NULLIFY(config_file);
  return result;
}
