/*
 * lsnc_gateway.c
 *
 *  Created on: 1 Jun. 2018
 *      Author: ian
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <net/if.h>
#include <linux/rtnetlink.h>
#include <arpa/inet.h>
#include "lsnc_internal.h"

#include <setnetconfig.h>


struct s_snc_route_info
{
  struct in_addr dst_addr;
  struct in_addr src_addr;
  struct in_addr def_gateway;
  int ifindex;
  /* char ifName[IF_NAMESIZE]; */
};


static enum e_snc_result snc_read_nl_sock(int sock_fd, char *buffer, size_t buffer_size, unsigned msg_sequence, int pid, int *total_read)
{
  struct nlmsghdr *nl_hdr;
  int read_len;
  int msg_len;

  msg_len = 0;
  do
  {
    /* Receive response from the kernel */
    read_len = recv(sock_fd, buffer, buffer_size - msg_len, 0);
    if (read_len < 0)
      return sncr_syscallerror;

    /* Check if the header is valid */
    nl_hdr = (struct nlmsghdr *) buffer;
    if (NLMSG_OK(nl_hdr, (size_t) read_len) == 0 || nl_hdr->nlmsg_type == NLMSG_ERROR)
      return sncr_parseerror;

    /* Check if the its the last message */
    if (nl_hdr->nlmsg_type == NLMSG_DONE)
    {
      break;
    }
    else
    {
      /* Else move the pointer to buffer appropriately */
      buffer += read_len;
      msg_len += read_len;
    }

    /* Check if its a multi part message */
    if ((nl_hdr->nlmsg_flags & NLM_F_MULTI) == 0)
    {
      /* Return if its not */
      break;
    }
  }
  while (nl_hdr->nlmsg_seq != msg_sequence || nl_hdr->nlmsg_pid != (unsigned) pid);

  *total_read = msg_len;
  return sncr_success;
}


/* Parse the route information returned and see if it is the default one */
static enum e_snc_result snc_find_default_route(const struct nlmsghdr *nl_hdr, int our_iface_idx, struct in_addr *default_gateway)
{
  const struct rtmsg *rt_msg;
  const struct rtattr *rt_attr;
  int rt_len;
  struct s_snc_route_info route_info;

  rt_msg = (const struct rtmsg *) NLMSG_DATA(nl_hdr);

  /* If the route is not for AF_INET or does not belong to main routing table then return. */
  if (rt_msg->rtm_family != AF_INET || rt_msg->rtm_table != RT_TABLE_MAIN)
    return sncr_notfound;

  /* Get the rtattr field */
  rt_attr = (const struct rtattr *) RTM_RTA(rt_msg);
  rt_len = RTM_PAYLOAD(nl_hdr);

  memset(&route_info, 0, sizeof(route_info));
  for ( ; RTA_OK(rt_attr, rt_len); rt_attr = RTA_NEXT(rt_attr, rt_len))
  {
    switch(rt_attr->rta_type)
    {
      case RTA_OIF:
        route_info.ifindex = *(int *) RTA_DATA(rt_attr);
        break;

      case RTA_GATEWAY:
        memcpy(&route_info.def_gateway, RTA_DATA(rt_attr), sizeof(route_info.def_gateway));
        break;

      case RTA_PREFSRC:
        memcpy(&route_info.src_addr, RTA_DATA(rt_attr), sizeof(route_info.src_addr));
        break;

      case RTA_DST:
        memcpy(&route_info.dst_addr, RTA_DATA(rt_attr), sizeof(route_info.dst_addr));
        break;
    }
  }

  /* Is this the default gateway */
  if (our_iface_idx == route_info.ifindex && route_info.dst_addr.s_addr == htonl(0))
  {
    *default_gateway = route_info.def_gateway;
    return sncr_success;
  }
  else
  {
    return sncr_notfound;
  }
}


/** Get the current IPv4 default gateway using the PF_NETLINK socket interface */
enum e_snc_result snc_ipv4_gateway_get_current(const char *eth_interface, struct in_addr *default_gateway)
{
  enum e_snc_result result;
  struct nlmsghdr *nl_msg;
  char *message_buffer = NULL;
  const int mypid = getpid();
  int sock_fd = -1;
  unsigned msg_sequence = 0;
  int len;
  int our_iface_idx;

  /* Get the interface index of our eth_interface */
  our_iface_idx = if_nametoindex(eth_interface);
  if (0 == our_iface_idx)
    return sncr_paramerror;

  /* Create socket */
  sock_fd = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE);
  if (sock_fd < 0)
  {
    result = sncr_syscallerror;
    goto siggc_fail;
  }

  /* Make a large buffer, and initialise to 0 */
  message_buffer = calloc(SNC_PACKET_FILTER_BUFFER_SIZE, sizeof(char));
  if (NULL == message_buffer)
  {
    result = sncr_allocerror;
    goto siggc_fail;
  }

  /* Point the header and the msg structure pointers into the buffer */
  nl_msg = (struct nlmsghdr *) message_buffer;

  /* Fill in the nlmsg header*/
  nl_msg->nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg)); /* Length of message */
  nl_msg->nlmsg_type = RTM_GETROUTE;                      /* Get the routes from kernel routing table */
  nl_msg->nlmsg_flags = NLM_F_DUMP | NLM_F_REQUEST;       /* The message is a request for dump */
  nl_msg->nlmsg_seq = msg_sequence++;                     /* Sequence of the message packet */
  nl_msg->nlmsg_pid = mypid;                              /* PID of process sending the request */

  /* Send the request */
  if (send(sock_fd, nl_msg, nl_msg->nlmsg_len, 0) < 0)
  {
    result = sncr_syscallerror;
    goto siggc_fail;
  }

  /* Read the response */
  len = 0;
  result = snc_read_nl_sock(sock_fd, message_buffer, SNC_PACKET_FILTER_BUFFER_SIZE, msg_sequence, mypid, &len);
  if (sncr_success != result)
    goto siggc_fail;

  /* Parse and print the response */
  default_gateway->s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  for (result = sncr_notfound; NLMSG_OK(nl_msg, len); nl_msg = NLMSG_NEXT(nl_msg, len))
  {
    result = snc_find_default_route(nl_msg, our_iface_idx, default_gateway);
    switch (result)
    {
      case sncr_success:
        break;
      case sncr_notfound:
        continue;
        break;
      default:
        goto siggc_fail;
        break;
    }
  }

siggc_fail:
  SNC_FREE_AND_NULLIFY(message_buffer);
  SNC_CLOSE_AND_INVALIDATE(sock_fd);
  return result;
}
