/*
 * lsnc_hostname.c
 *
 *  Created on: 5 Jun. 2018
 *      Author: ian
 */
#include <stddef.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <setnetconfig.h>


/** Check hostname contains sensible characters, as per RFC952, RFC1123, and is of sensible length, as per RFC 2181 */
static enum e_snc_result snc_internal_hostname_validate(const char *hostname, bool stop_at_dot)
{
  enum e_snc_result result;
  char ch;
  const char *ptr;
  size_t length;

  /* Check valid characters */
  for (ptr = hostname, length = 0; '\0' != *ptr && (!stop_at_dot || '.' != *ptr); ptr++, length++)
  {
    ch = *ptr;
    if (!((ch >= 'a' && ch <= 'z') ||
          (ch >= 'A' && ch <= 'Z') ||
          (ch >= '0' && ch <= '9') ||
          ('-' == ch)))
    {
      result = sncr_parseerror;
      goto vh_fail;
    }
  }

  /* Check length */
  if (0 == length || length > SNC_HOST_NAME_MAX)
  {
    result = sncr_parseerror;
    goto vh_fail;
  }

  result = sncr_success;

vh_fail:
  return result;
}


/** Check hostname contains sensible characters, as per RFC952, RFC1123, and is of sensible length, as per RFC 2181 */
enum e_snc_result snc_hostname_validate(const char *hostname)
{
  return snc_internal_hostname_validate(hostname, false);
}


/** Check domainname contains sensible characters, as per RFC952, RFC1123, and is of sensible length, as per RFC 2181 */
enum e_snc_result snc_domainname_validate(const char *domainname)
{
  enum e_snc_result result;
  const char *ptr;
  const char *dotptr;
  size_t length;

  /* Check the length and characters each section separated by a . */
  for (ptr = domainname, result = sncr_success;
       NULL != ptr && sncr_success == result;
       dotptr = strchr(ptr, '.'), ptr = (NULL == dotptr) ? dotptr : &dotptr[1])
  {
    result = snc_internal_hostname_validate(ptr, true);
  }

  /* Check length */
  length = strlen(domainname);
  return (length > SNC_DOMAIN_NAME_MAX) ? sncr_parseerror : (length == 0) ? sncr_success : result;
}


/** Get the current system hostname via a system call (gethostname) */
enum e_snc_result snc_get_hostname_syscall(char *hostname, size_t hostname_length)
{
  enum e_snc_result result;

  /* Get the hostname from the system */
  if (gethostname(hostname, hostname_length) < 0)
  {
    result = (ENAMETOOLONG == errno) ? sncr_string_too_long : sncr_syscallerror;
  }
  else
  {
    result = sncr_success;
  }
  return result;
}
