/*ALI
 * lsnc_ipv4_validate.c
 *
 *  Created on: 19 Jul. 2018
 *      Author: ian
 */
#include <stdbool.h>
#include <stdint.h>

#include <setnetconfig.h>


/*
 * Class A: valid address range
 *
 * Top byte:
 *        0000 | 0000  0   (0x80)
 *         to
 *        0111 | 1111  127 (0x7F)
 * Value: 0xxx | xxxx
 * Mask:  1000 | 0000      (0x80)
 */
#define SNC_IPV4_CLASS_A_MASK  0x80000000
#define SNC_IPV4_CLASS_A_VALUE 0x00000000


/*
 * Loopback (127) is a subset of class A
 *
 * Top byte:
 *        0111 | 1111  127 (0x7F)
 * Value: 0111 | 1111
 * Mask:  1111 | 1111      (0xFF)
 */
#define SNC_IPV4_LOOPBACK_MASK  0xFF000000
#define SNC_IPV4_LOOPBACK_VALUE 0x7F000000


/*
 * Class B: valid address range
 *
 * Top byte:
 *        1000 | 0000  128 (0x80)
 *         to
 *        1011 | 1111  191 (0xBF)
 * Value: 10xx | xxxx
 * Mask:  1100 | 0000      (0xC0)
 */
#define SNC_IPV4_CLASS_B_MASK  0xC0000000
#define SNC_IPV4_CLASS_B_VALUE 0x80000000


/*
 * Class C: valid address range
 *
 * Top byte:
 *        1100 | 0000  192 (0xC0)
 *         to
 *        1101 | 1111  223 (0xDF)
 * Value: 110x | xxxx
 * Mask:  1110 | 0000      (0xE0)
 */
#define SNC_IPV4_CLASS_C_MASK  0xE0000000
#define SNC_IPV4_CLASS_C_VALUE 0xC0000000


/*
 * Class D: multicast range
 *
 * Top byte:
 *        1110 | 0000  224 (0xE0)
 *         to
 *        1110 | 1111  239 (0xEF)
 * Value: 1110 | xxxx
 * Mask:  1111 | 0000      (0xF0)
 */
#define SNC_IPV4_CLASS_D_MASK  0xF0000000
#define SNC_IPV4_CLASS_D_VALUE 0xE0000000


/*
 * Class E: reserved range
 *
 * Top byte:
 *        1111 | 0000  240 (0xF0)
 *         to
 *        1111 | 1111  255 (0xFF)
 * Value: 1111 | xxxx
 * Mask:  1111 | 0000      (0xF0)
 */
#define SNC_IPV4_CLASS_E_MASK  0xF0000000
#define SNC_IPV4_CLASS_E_VALUE 0xF0000000


/**
 * Subnet /31 is a special case since bit 0 would normally be broadcast or subnet 0, but instead are valid IP addresses
 *
 * For example:
 *   NM: 255.255.255.254
 *   IP: 192.168.100.0 is a valid IP address on this subnet
 *   IP: 192.168.100.1 is a valid IP address on this subnet
 */
#define SNC_IPV4_SUBNET_31 0xFFFFFFFE


#define SNC_IPV4_SUBNET_ZERO 0x00000000
#define SNC_IPV3_SUBNET_ALLONES 0xFFFFFFFF


/** Is ip a subnet broadcast address */
#define SNC_IP_IS_BROADCAST(ip, nm) \
  (SNC_IPV4_SUBNET_31 != nm &&  /* Except for /31, */ \
   ((ip & ~nm) == ~nm ||        /* all non-netmask bits set is subnet broadcast */ \
    (ip & ~nm) == 0))           /* and all non-netmask bits clear is subnet 0 (oldskool broadcast) */


/**
 * Validate that ipv4_netmask (in network byte order) is a valid netmask
 * Conditions:
 *  - 1s are contiguous from right to left (RFC4632 Section 3.1)
 *  - Excludes 255.255.255.255 (broadcast)
 *  - Excludes 0.0.0.0 (network)
 */
enum e_snc_result snc_ipv4_netmask_validate(const struct in_addr *ipv4_netmask)
{
  size_t i;
  uint32_t nm;

  if (NULL == ipv4_netmask)
    return sncr_paramerror;

  nm = ntohl(ipv4_netmask->s_addr);

  /* Work right to left, looking for the first 0 */
  for (i = 0; i < 32; i++)
  {
    if ((nm & (1 << (32 - 1 - i))) == 0)
      break;
  }

  /* Found first 0, the rest must be 0s */
  for ( ; i < 32; i++)
  {
    if ((nm & (1 << (32 - 1 - i))) != 0)
      return sncr_parseerror;
  }

  return sncr_success;
}


/**
 * Validate that ipv4_address (in network byte order) is suitable as a destination address of a single host (not broadcast/multicast)
 * (permits off subnet)
 * Excludes:
 *  - Class D (multicast range)
 *  - Class E (reserved)
 *  - Subnet broadcast (ie 192.168.1.255) if ipv4_source and ipv4_source_netmask are supplied
 *  - Subnet 0 (ie 192.168.1.0), except for /31 (see RFC3021) - this was a historic broadcast address if ipv4_source and ipv4_source_netmask are supplied
 *
 * @param ipv4_destination is the destination address to validate in network byte order. Not NULL
 * @param ipv4_source is an optional source address, used to verify that the destination is not a subnet broadcast address in network byte order. Can be NULL
 * @param ipv4_source_netmask is an optional netmask of ipv4_source, used to verify that the destination is not a subnet broadcast address in network byte order. Can be NULL
 *
 * @return sncr_success if ipv4_destination is a valid address
 */
enum e_snc_result snc_ipv4_single_destination_address_validate(const struct in_addr *ipv4_destination, const struct in_addr *ipv4_source, const struct in_addr *ipv4_source_netmask)
{
  bool invalid;
  uint32_t dst;
  uint32_t src;
  uint32_t src_nm;

  if (NULL == ipv4_destination)
    return sncr_paramerror;

  dst = ntohl(ipv4_destination->s_addr);

  invalid = ((dst & SNC_IPV4_CLASS_D_MASK) == SNC_IPV4_CLASS_D_VALUE ||    /* Destination can't be class D */
             (dst & SNC_IPV4_CLASS_E_MASK) == SNC_IPV4_CLASS_E_VALUE);     /* Destination can't be class E */

  if (!invalid && NULL != ipv4_source && NULL != ipv4_source_netmask)
  {
    src = ntohl(ipv4_source->s_addr);
    src_nm = ntohl(ipv4_source_netmask->s_addr);
    if ((src & src_nm) == (dst & src_nm))
    {
      /* On the same subnet */
      invalid = SNC_IP_IS_BROADCAST(dst, src_nm);
    }
  }

  return (invalid) ? sncr_parseerror : sncr_success;
}


/**
 * Validate that ipv4_address (in network byte order) is suitable for setting as a host address
 * Excludes:
 *  - Loopback
 *  - Class D (multicast range)
 *  - Class E (reserved)
 *  - Subnet broadcast (ie 192.168.1.255)
 *  - Subnet 0 (ie 192.168.1.0), except for /31 (see RFC3021) - this was a historic broadcast address
 *
 * @param ipv4_address is a host ip address to verify in network byte order. Not NULL
 * @param ipv4_netmask is the netmask for the ip address in network byte order. Not NULL
 *
 * @return sncr_success if ipv4_address is a valid host address
 */
enum e_snc_result snc_ipv4_host_address_validate(const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask)
{
  bool invalid;
  uint32_t nm;
  uint32_t ip;

  if (NULL == ipv4_address || NULL == ipv4_netmask)
    return sncr_paramerror;

  ip = ntohl(ipv4_address->s_addr);
  nm = ntohl(ipv4_netmask->s_addr);

  invalid = (snc_ipv4_netmask_validate(ipv4_netmask) != sncr_success ||   /* Valid netmask */
             (ip & SNC_IPV4_LOOPBACK_MASK) == SNC_IPV4_LOOPBACK_VALUE ||  /* Can't be loopback */
             (ip & SNC_IPV4_CLASS_D_MASK) == SNC_IPV4_CLASS_D_VALUE ||    /* Can't be class D */
             (ip & SNC_IPV4_CLASS_E_MASK) == SNC_IPV4_CLASS_E_VALUE ||    /* Can't be class E */
             SNC_IP_IS_BROADCAST(ip, nm));                                /* Can't be subnet broadcast or subnet 0, except on /31 */

  return (invalid) ? sncr_parseerror : sncr_success;
}


/**
 * Validate that ipv4_address (in network byte order) is suitable for setting as a loopback address
 * Excludes:
 *  - Loopback subnet broadcast (eg ip 127.255.255.255 nm 255.255.255.0)
 *  - Loopback subnet 0 (eg ip 127.255.255.0 nm 255.255.255.0) - this was a historic broadcast address
 *  - Netmask outside of 255.0.0.0 .. 255.255.255.255
 *
 * @param ipv4_address is a loopback ip address to verify in network byte order. Not NULL
 * @param ipv4_netmask is the netmask for the loopback ip address in network byte order. Not NULL
 *
 * @return sncr_success if ipv4_address is a valid loopback address
 */
enum e_snc_result snc_ipv4_loopback_address_validate(const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask)
{
  bool invalid;
  uint32_t nm;
  uint32_t ip;

  if (NULL == ipv4_address || NULL == ipv4_netmask)
    return sncr_paramerror;

  ip = ntohl(ipv4_address->s_addr);
  nm = ntohl(ipv4_netmask->s_addr);

  invalid = (snc_ipv4_netmask_validate(ipv4_netmask) != sncr_success ||   /* Valid netmask */
             (ip & SNC_IPV4_LOOPBACK_MASK) != SNC_IPV4_LOOPBACK_VALUE ||  /* Must be loopback */
             SNC_IP_IS_BROADCAST(ip, nm) ||                               /* Can't be subnet broadcast or subnet 0, except on /31 */
             nm < SNC_IPV4_LOOPBACK_MASK);                                /* Netmask must be 255.0.0.0 .. 255.255.255.255 */

  return (invalid) ? sncr_parseerror : sncr_success;
}


/**
 * Validate that ipv4_gateway (in network byte order) is valid gateway address
 *
 * Must be a valid host address (@ref snc_ipv4_host_address_validate()) and must be on the same subnet as ipv4_address
 *
 * @param ipv4_gateway is the gateway address to verify in network byte order. Not NULL
 * @param ipv4_address is the ip address used to verify the gateway in network byte order. Not NULL
 * @param ipv4_netmask is the netmask used to verify the gateway in network byte order. Not NULL
 *
 * @return sncr_success if ipv4_gateway is a valid gateway address
 */
enum e_snc_result snc_ipv4_gateway_address_validate(const struct in_addr *ipv4_gateway, const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask)
{
  bool invalid;
  uint32_t nm;
  uint32_t ip;
  uint32_t gw;

  if (NULL == ipv4_gateway || NULL == ipv4_address || NULL == ipv4_netmask)
    return sncr_paramerror;

  gw = ntohl(ipv4_gateway->s_addr);
  ip = ntohl(ipv4_address->s_addr);
  nm = ntohl(ipv4_netmask->s_addr);

  invalid = (snc_ipv4_host_address_validate(ipv4_gateway, ipv4_netmask) != sncr_success || /* Gateway is a valid host address */
             (gw & nm) != (ip & nm));                                                      /* Gateway must be on the same network as our IP address */

  return (invalid) ? sncr_parseerror : sncr_success;
}


/**
 * Validate that ipv4_broadcast is valid broadcast address (in network byte order) for ipv4_address and ipv4_netmask
 *
 * @param ipv4_broadcast is the broadcast address to verify in network byte order. Not NULL
 * @param ipv4_address is the ip address used to verify the broadcast in network byte order. Not NULL
 * @param ipv4_netmask is the netmask used to verify the broadcast in network byte order. Not NULL
 *
 * @return sncr_success if ipv4_broadcast is a valid broadcast address
 */
enum e_snc_result snc_ipv4_broadcast_validate(const struct in_addr *ipv4_broadcast, const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask)
{
  bool invalid;
  struct in_addr ipv4_expected_broadcast;

  if (NULL == ipv4_broadcast || NULL == ipv4_address || NULL == ipv4_netmask)
    return sncr_paramerror;

  snc_ipv4_broadcast_from_ip_netmask(ipv4_address, ipv4_netmask, &ipv4_expected_broadcast);
  invalid = ipv4_expected_broadcast.s_addr != ipv4_broadcast->s_addr;

  return (invalid) ? sncr_parseerror : sncr_success;
}


/**
 * Validate all fields within a struct s_snc_net_config
 *
 * @param net_config is the configuration to verify
 * @param valid_mask is set to a bit mask of SNC_VALIDATE_* constants
 * @param check_dhcp_fields if set will check DHCP supplied fields (ip/nm/bc/gw/dns/sd) even DHCP is enabled. Generally use false.
 *
 * @return sncr_success when all fields are valid
 */
enum e_snc_result snc_net_config_validate(const struct s_snc_net_config *net_config, uint32_t *valid_mask, bool check_dhcp_fields)
{
  size_t i;
  uint32_t v;

  if (NULL == net_config)
    return sncr_paramerror;

  v = 0;

  if (!net_config->use_dhcp || (check_dhcp_fields && net_config->use_dhcp))
  {
    /* Validate netmask and IP address */
    if (snc_ipv4_netmask_validate(&net_config->ip_netmask) != sncr_success)
      v |= SNC_VALIDATE_IPV4_NETMASK_INVALID;
    if (snc_ipv4_host_address_validate(&net_config->ip_address, &net_config->ip_netmask) != sncr_success)
      v |= SNC_VALIDATE_IPV4_ADDRESS_INVALID;
    if (snc_ipv4_broadcast_validate(&net_config->ip_broadcast, &net_config->ip_address, &net_config->ip_netmask) != sncr_success)
      v |= SNC_VALIDATE_IPV4_BROADCAST_INVALID;

    /* Gateway doesn't have to be set */
    if (SNC_IPV4_UNSET_INADDR_NETWORK != net_config->ip_gateway.s_addr &&
        snc_ipv4_gateway_address_validate(&net_config->ip_gateway, &net_config->ip_address, &net_config->ip_netmask) != sncr_success)
      v |= SNC_VALIDATE_IPV4_GATEWAY_INVALID;

    /* DNS servers don't have to be set */
    for (i = 0; i < SNC_NUM_DNS_SERVERS; i++)
    {
      if (SNC_IPV4_UNSET_INADDR_NETWORK != net_config->ip_dns[i].s_addr &&
          snc_ipv4_single_destination_address_validate(&net_config->ip_dns[i], &net_config->ip_address, &net_config->ip_netmask) != sncr_success)
        v |= SNC_VALIDATE_IPV4_DNS0_INVALID << i;
    }

    if (snc_domainname_validate(net_config->search_domain) != sncr_success)
      v |= SNC_VALIDATE_SEARCHDOMAIN_INVALID;
  }

  if (snc_hostname_validate(net_config->hostname) != sncr_success)
    v |= SNC_VALIDATE_HOSTNAME_INVALID;
  if (snc_domainname_validate(net_config->domain) != sncr_success)
    v |= SNC_VALIDATE_DOMAIN_INVALID;

  if (NULL != valid_mask)
    *valid_mask = v;
  return (SNC_VALIDATE_VALID == v) ? sncr_success : sncr_parseerror;
}
