/*
 * lsnc_nameservers.c
 *
 *  Created on: 1 Jun. 2018
 *      Author: ian
 */
#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#ifdef HAVE_RESOLV_H
#include <resolv.h>
#endif
#include "lsnc_parse_addr.h"
#include "lsnc_internal.h"

#include "lsnc_nameservers.h"
#include <setnetconfig.h>


#if defined(HAVE_RESOLV_H) && defined(HAVE_LIBRESOLV)

/** Get the current IPv4 nameservers and our domain from libresolv. Function is only available if libresolv is available */
enum e_snc_result snc_ipv4_nameservers_domain_get_current_libresolv(struct s_snc_net_config *net_config)
{
  size_t i;
  ssize_t j;
  struct __res_state resolver_state;

  /* Use the resolver to get our DNS servers */
  if (res_ninit(&resolver_state) < 0)
  {
    return sncr_resolverror;
  }

  /* Iterate the name server list within the resolver_state */
  for (i = 0, j = 0; i < SNC_NUM_DNS_SERVERS && j < resolver_state.nscount; j++)
  {
    if (resolver_state.nsaddr_list[j].sin_family == AF_INET)
    {
      net_config->ip_dns[i] = resolver_state.nsaddr_list[j].sin_addr;
      i++;
    }
  }

  /* Rest are set to 0 */
  for ( ; i < SNC_NUM_DNS_SERVERS; i++)
    net_config->ip_dns[i].s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;

  /* Get domain */
  if (strcmp(resolver_state.defdname, "") != 0)
  {
    snprintf(net_config->domain, SNC_DOMAIN_NAME_MAX, "%s", resolver_state.defdname);
  }
  else
  {
    snprintf(net_config->domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
  }

  /* Get search domain (we only support 1) */
  snprintf(net_config->search_domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
  for (i = 0; i < MAXDNSRCH && NULL != resolver_state.dnsrch[i]; i++)
  {
    if (strcmp(resolver_state.dnsrch[i], "") != 0)
    {
      snprintf(net_config->search_domain, SNC_DOMAIN_NAME_MAX, "%s", resolver_state.dnsrch[i]);
      break;
    }
  }

  res_nclose(&resolver_state);

  return sncr_success;
}


/** Get the current IPv4 nameservers and domain, using libresolv if available, otherwise falling back to parsing /etc/resolv.conf */
enum e_snc_result snc_ipv4_nameservers_domain_get_current(const struct s_snc_configuration *config, struct s_snc_net_config *net_config)
{
  return snc_ipv4_nameservers_domain_get_current_libresolv(net_config);
}


#else /* #if defined(HAVE_RESOLV_H) && defined(HAVE_LIBRESOLV) */


/** Get the current IPv4 nameservers and domain, using libresolv if available, otherwise falling back to parsing /etc/resolv.conf */
enum e_snc_result snc_ipv4_nameservers_domain_get_current(const struct s_snc_configuration *config, struct s_snc_net_config *net_config)
{
  return snc_ipv4_nameservers_domain_get_current_resolvconf(config->resolv_file, net_config);
}


#endif /* #else -> #if defined(HAVE_RESOLV_H) && defined(HAVE_LIBRESOLV) */


#define STRINGIFY(X) INDIRECT(X)
#define INDIRECT(X) #X


/** Get the current IPv4 nameservers and domain by parsing the results from /etc/resolv.conf */
enum e_snc_result snc_ipv4_nameservers_domain_get_current_resolvconf(const char *resolvconf_path, struct s_snc_net_config *net_config)
{
  FILE *ns = NULL;
  size_t i;
  struct in_addr ipv4_conv;
  char *line;
  enum e_snc_result result;

  line = malloc(SNC_PARSE_LINE_LENGTH);
  if (NULL == line)
  {
    result = sncr_allocerror;
    goto singc_fail;
  }

  /* Open resolv.conf file */
  ns = fopen(resolvconf_path, "r");
  if (NULL == ns)
  {
    result = sncr_fileerror;
    goto singc_fail;
  }

  /* Look for 0 or more "nameserver" lines, the "domain" and "search" lines */
  snprintf(net_config->search_domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
  snprintf(net_config->domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
  for (i = 0; ; )
  {
    /* Read a line from the file */
    if (NULL == fgets(line, SNC_PARSE_LINE_LENGTH, ns))
    {
      if (feof(ns))
      {
        break;
      }
      else
      {
        result = sncr_fileerror;
        goto singc_fail;
      }
    }

    /* Look for nameserver line */
    if (i < SNC_NUM_DNS_SERVERS)
    {
      if (snc_parse_ipv4_from_line("nameserver", true, line, &ipv4_conv) == sncr_success)
      {
        net_config->ip_dns[i] = ipv4_conv;
        i++;
      }
    }

    /* Look for "domain" line which contains our domain */
    if (snc_parse_string_from_line("domain", true, line, net_config->domain, SNC_DOMAIN_NAME_MAX) == sncr_success)
    {
      /* Do nothing */
    }

    /* Look for "search" line which contains our domain */
    if (snc_parse_string_from_line("search", true, line, net_config->search_domain, SNC_DOMAIN_NAME_MAX) == sncr_success)
    {
      /* Do nothing */
    }
  }

  /* Rest are set to 0 */
  for ( ; i < SNC_NUM_DNS_SERVERS; i++)
    net_config->ip_dns[i].s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;

  result = sncr_success;

singc_fail:
  SNC_FREE_AND_NULLIFY(line);
  SNC_FCLOSE_AND_NULLIFY(ns);
  return result;
}
