/*
 * lsnc_netcurrent_ifconfig.c
 *
 * Get the current network settings by parsing the output from ip, ifconfig, route and /etc/resolv.conf
 *
 *  Created on: 31 May 2018
 *      Author: ian
 *
 *
# Parse output from ifconfig - looks like:
# eth0      Link encap:Ethernet  HWaddr 00:E0:0C:BC:E5:61
#           inet addr:172.17.10.26  Bcast:172.17.255.255  Mask:255.255.0.0
#           UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
#           RX packets:2819 errors:0 dropped:0 overruns:0 frame:0
#           TX packets:385 errors:0 dropped:0 overruns:0 carrier:0
#           collisions:0 txqueuelen:1000
#           RX bytes:223883 (218.6 KiB)  TX bytes:35257 (34.4 KiB)
#
# If DHCP, and no cable attached
# eth0      Link encap:Ethernet  HWaddr 00:E0:0C:BC:E5:60
#           UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
#           RX packets:0 errors:0 dropped:0 overruns:0 frame:0
#           TX packets:1 errors:0 dropped:0 overruns:0 carrier:0
#           collisions:0 txqueuelen:1000
#           RX bytes:0 (0.0 B)  TX bytes:322 (322.0 B)
#
# Newer versions of ifconfig - look like:
# ens33: flags=4163<UP,BROADCAST,RUNNING,MULTICAST>  mtu 1500
#         inet 172.17.1.99  netmask 255.255.255.0  broadcast 172.17.1.255
#         inet6 fe80::f13d:10a5:4902:3754  prefixlen 64  scopeid 0x20<link>
#         ether 00:0c:29:48:a4:fa  txqueuelen 1000  (Ethernet)
#         RX packets 4173901  bytes 896717545 (896.7 MB)
#         RX errors 0  dropped 0  overruns 0  frame 0
#         TX packets 1864697  bytes 254280496 (254.2 MB)
#         TX errors 0  dropped 0 overruns 0  carrier 0  collisions 0

# lo: flags=73<UP,LOOPBACK,RUNNING>  mtu 65536
#         inet 127.0.0.1  netmask 255.0.0.0
#         inet6 ::1  prefixlen 128  scopeid 0x10<host>
#         loop  txqueuelen 1000  (Local Loopback)
#         RX packets 57867  bytes 4419973 (4.4 MB)
#         RX errors 0  dropped 0  overruns 0  frame 0
#         TX packets 57867  bytes 4419973 (4.4 MB)
#         TX errors 0  dropped 0 overruns 0  carrier 0  collisions 0
#
#
# Output from "ip address show dev ens33"
# 2: ens33: <BROADCAST,MULTICAST,UP,LOWER_UP> mtu 1500 qdisc fq_codel state UP group default qlen 1000
#     link/ether 00:0c:29:48:a4:fa brd ff:ff:ff:ff:ff:ff
#     inet 172.17.1.60/24 brd 172.17.1.255 scope global dynamic noprefixroute ens33
#        valid_lft 3135sec preferred_lft 3135sec
#     inet6 fe80::f13d:10a5:4902:3754/64 scope link noprefixroute
#        valid_lft forever preferred_lft forever
#
#
# Output from "ip route show default dev ens33"
# default via 172.17.1.1 proto dhcp metric 100
#
# Output from "ip route show default eth0", if there is no default route/gw:
# <empty>
#
#
# Parse output from route -n - looks like:
#  Kernel IP routing table
#  Destination     Gateway         Genmask         Flags Metric Ref    Use Iface
#  172.17.0.0      0.0.0.0         255.255.0.0     U     0      0        0 eth0
#  127.0.0.0       0.0.0.0         255.0.0.0       U     0      0        0 lo
#  0.0.0.0         172.17.1.1      0.0.0.0         UG    0      0        0 eth0
#
# Output from route -n, if there is no default route/gw:
#  Kernel IP routing table
#  Destination     Gateway         Genmask         Flags Metric Ref    Use Iface
#  172.17.0.0      0.0.0.0         255.255.0.0     U     0      0        0 eth0
#  127.0.0.0       0.0.0.0         255.0.0.0       U     0      0        0 lo
#
#
# Contents of /etc/resolv.conf:
#  Generated by NetworkManager
#  domain rinstrumint.local
#  search rinstrumint.local
#  nameserver 172.17.1.5
 *
 *
 */
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <unistd.h>
#include "lsnc_nameservers.h"
#include "lsnc_parse_addr.h"
#include "lsnc_internal.h"
#include "lsnc_dhcp.h"

#include <setnetconfig.h>


/** Parse gateway address from the output of route -n */
static enum e_snc_result snc_get_net_current_route(const char *route_path, const char *eth_interface, struct s_snc_net_config *net_config)
{
  enum e_snc_result result;
  char *line = NULL;
  char route_cmd[SNC_EXEC_COMMAND_LENGTH];
  FILE *route_output = NULL;
  int res;
  char scaniface[SNC_IF_NAMESIZE + 1];
  char ipstr[SNC_INET_ADDRSTRLEN + 1];

  line = malloc(SNC_PARSE_LINE_LENGTH);
  if (NULL == line)
  {
    result = sncr_allocerror;
    goto sgncg_fail;
  }

  /* Run route */
  snprintf(route_cmd, SNC_EXEC_COMMAND_LENGTH, "%s -n", route_path);
  route_output = popen(route_cmd, "r");
  if (NULL == route_output)
  {
    result = sncr_fileerror;
    goto sgncg_fail;
  }

  /* Look for the various fields we need */
  for ( ; ; )
  {
    /* Read a line from the route output */
    if (NULL == fgets(line, SNC_PARSE_LINE_LENGTH, route_output))
    {
      if (feof(route_output))
      {
        break;
      }
      else
      {
        result = sncr_fileerror;
        goto sgncg_fail;
      }
    }

    /* Scan a line */
    /*                  dest     gw   genmask flg met ref use iface */
    res = sscanf(line, "0.0.0.0 %" SNC_STRINGIFY(SNC_INET_ADDRSTRLEN) "s 0.0.0.0 %*s %*s %*s %*s %" SNC_STRINGIFY(SNC_IF_NAMESIZE) "s", ipstr, scaniface);
    if (2 == res)
    {
      /* Match on this line, so compare interface */
      if (0 == strcmp(eth_interface, scaniface))
      {
        /* Found default gw for our interface, convert IPv4 string to IPv4 address */
        if (inet_pton(AF_INET, ipstr, &net_config->ip_gateway) == 1)
        {
          result = sncr_success;
          goto sgncg_fail;
        }
      }
    }
  }

  /* Default gateway was not found */
  net_config->ip_gateway.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  result = sncr_notfound;

sgncg_fail:
  SNC_PCLOSE_AND_NULLIFY(route_output);
  SNC_FREE_AND_NULLIFY(line);
  return result;
}


/** Parse ip, broadcast, netmask and mac address from the output of ifconfig */
static enum e_snc_result snc_get_net_current_ifconfig_internal(const char *ifconfig_path, const char *eth_interface, struct s_snc_net_config *net_config)
{
  enum e_snc_result result;
  char *line = NULL;
  char ifconfig_cmd[SNC_EXEC_COMMAND_LENGTH];
  FILE *ifconfig_output = NULL;
  bool have_ip;
  bool have_bcast;
  bool have_mask;
  bool have_mac;

  line = malloc(SNC_PARSE_LINE_LENGTH);
  if (NULL == line)
  {
    result = sncr_allocerror;
    goto sgnci_fail;
  }

  /* Run ifconfig <iface> */
  snprintf(ifconfig_cmd, SNC_EXEC_COMMAND_LENGTH, "%s %s", ifconfig_path, eth_interface);
  ifconfig_output = popen(ifconfig_cmd, "r");
  if (NULL == ifconfig_output)
  {
    result = sncr_fileerror;
    goto sgnci_fail;
  }

  /* Look for the various fields we need */
  for (have_ip = false, have_bcast = false, have_mask = false, have_mac = false; ; )
  {
    /* Read a line from the ifconfig output */
    if (NULL == fgets(line, SNC_PARSE_LINE_LENGTH, ifconfig_output))
    {
      if (feof(ifconfig_output))
      {
        break;
      }
      else
      {
        result = sncr_fileerror;
        goto sgnci_fail;
      }
    }

    /* Look for IP address */
    if (!have_ip &&
        (snc_parse_ipv4_from_line("inet addr:", false, line, &net_config->ip_address) == sncr_success ||
         snc_parse_ipv4_from_line("inet", false, line, &net_config->ip_address) == sncr_success))
    {
      have_ip = true;
    }

    /* Look for broadcast address */
    if (!have_bcast &&
        (snc_parse_ipv4_from_line("Bcast:", false, line, &net_config->ip_broadcast) == sncr_success ||
         snc_parse_ipv4_from_line("broadcast", false, line, &net_config->ip_broadcast) == sncr_success))
    {
      have_bcast = true;
    }

    /* Look for netmask address */
    if (!have_mask &&
        (snc_parse_ipv4_from_line("Mask:", false, line, &net_config->ip_netmask) == sncr_success ||
         snc_parse_ipv4_from_line("netmask", false, line, &net_config->ip_netmask) == sncr_success))
    {
      have_mask = true;
    }

    /* Look for MAC address */
    if (!have_mac &&
        (snc_parse_mac_from_line("HWaddr", false, line, &net_config->mac_address) == sncr_success ||
         snc_parse_mac_from_line("ether", false, line, &net_config->mac_address) == sncr_success))
    {
      have_mac = true;
    }
  }
  result = (have_ip && have_bcast && have_mask && have_mac) ? sncr_success : sncr_parseerror;

sgnci_fail:
  SNC_PCLOSE_AND_NULLIFY(ifconfig_output);
  SNC_FREE_AND_NULLIFY(line);

  return result;
}


/** Get the current network settings from parsing the output from various processes (ifconfig, route) */
enum e_snc_result snc_get_net_current_ifconfig(const struct s_snc_configuration *config, const char *eth_interface, struct s_snc_net_config *net_current)
{
  enum e_snc_result result;

  /* Set some initial values */
  result = snc_net_config_struct_init(net_current);
  if (sncr_success != result)
    return result;

  /* Get the DHCP setting from our configuration */
  result = snc_get_net_current_dhcp(config, SNC_METHOD_IFCONFIG, eth_interface, &net_current->use_dhcp);
  if (sncr_success != result)
    return result;

  /* Get ip, netmask, broadcast and mac from "ifconfig" */
  result = snc_get_net_current_ifconfig_internal(config->ifconfig_cmd, eth_interface, net_current);
  if (sncr_success != result)
    return result;

  /* Get default gateway from "route" */
  result = snc_get_net_current_route(config->route_cmd, eth_interface, net_current);
  if (sncr_success != result && sncr_notfound != result)
    return result;

  /* Parse nameservers and domain from resolv.conf */
  result = snc_ipv4_nameservers_domain_get_current_resolvconf(config->resolv_file, net_current);
  if (sncr_success != result)
    return result;

  /* Get the hostname from the system */
  result = snc_get_hostname_syscall(net_current->hostname, SNC_HOST_NAME_MAX);
  if (sncr_success != result)
    return result;

  return sncr_success;
}


/** Parse ip, broadcast, netmask and mac address from the output of the "ip address" command */
static enum e_snc_result snc_get_net_current_ip_address_command(const char *ip_path, const char *eth_interface, struct s_snc_net_config *net_config)
{
  enum e_snc_result result;
  char *line = NULL;
  char ip_cmd[SNC_EXEC_COMMAND_LENGTH];
  FILE *ip_output = NULL;
  bool have_ip;
  bool have_bcast;
  bool have_mask;
  bool have_mac;

  line = malloc(SNC_PARSE_LINE_LENGTH);
  if (NULL == line)
  {
    result = sncr_allocerror;
    goto sgnci_fail;
  }

  /* Run ip */
  snprintf(ip_cmd, SNC_EXEC_COMMAND_LENGTH, "%s address show dev %s", ip_path, eth_interface);
  ip_output = popen(ip_cmd, "r");
  if (NULL == ip_output)
  {
    result = sncr_fileerror;
    goto sgnci_fail;
  }

  /* Look for the various fields we need */
  for (have_ip = false, have_bcast = false, have_mask = false, have_mac = false; ; )
  {
    /* Read a line from the ifconfig output */
    if (NULL == fgets(line, SNC_PARSE_LINE_LENGTH, ip_output))
    {
      if (feof(ip_output))
      {
        break;
      }
      else
      {
        result = sncr_fileerror;
        goto sgnci_fail;
      }
    }

    /* Look for IP address and netmask: inet 172.17.1.60/24 */
    if (!have_ip &&
        (snc_parse_ipv4_cidr_from_line("inet", false, line, &net_config->ip_address, &net_config->ip_netmask) == sncr_success))
    {
      have_ip = true;
      have_mask = true;
    }

    /* Look for broadcast address: brd 172.17.1.255 */
    if (!have_bcast &&
        (snc_parse_ipv4_from_line("brd", false, line, &net_config->ip_broadcast) == sncr_success))
    {
      have_bcast = true;
    }

    /* Look for MAC address: link/ether 00:0c:29:48:a4:fa */
    if (!have_mac &&
        (snc_parse_mac_from_line("link/ether", false, line, &net_config->mac_address) == sncr_success))
    {
      have_mac = true;
    }
  }
  result = (have_ip && have_bcast && have_mask && have_mac) ? sncr_success : sncr_parseerror;

sgnci_fail:
  SNC_PCLOSE_AND_NULLIFY(ip_output);
  SNC_FREE_AND_NULLIFY(line);

  return result;
}


/**
 * Parse default gateway address from the output of the "ip route" command
 *
 * @return sncr_success if gateway was found, sncr_notfound if not found (gateway will be set to SNC_IPV4_UNSET_INADDR_NETWORK), or or other error
 */
static enum e_snc_result snc_get_net_current_ip_route_command(const char *ip_path, const char *eth_interface, struct s_snc_net_config *net_config)
{
  enum e_snc_result result;
  char *line = NULL;
  char ip_cmd[SNC_EXEC_COMMAND_LENGTH];
  FILE *ip_output = NULL;
  bool have_gateway;

  line = malloc(SNC_PARSE_LINE_LENGTH);
  if (NULL == line)
  {
    result = sncr_allocerror;
    goto sgnci_fail;
  }

  /* Run ip */
  snprintf(ip_cmd, SNC_EXEC_COMMAND_LENGTH, "%s route show default dev %s", ip_path, eth_interface);
  ip_output = popen(ip_cmd, "r");
  if (NULL == ip_output)
  {
    result = sncr_fileerror;
    goto sgnci_fail;
  }

  /* Look for the default gateway field we need */
  net_config->ip_gateway.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  for (have_gateway = false; ; )
  {
    /* Read a line from the ifconfig output */
    if (NULL == fgets(line, SNC_PARSE_LINE_LENGTH, ip_output))
    {
      if (feof(ip_output))
      {
        break;
      }
      else
      {
        result = sncr_fileerror;
        goto sgnci_fail;
      }
    }

    /* Look for gateway address: default via 172.17.1.1 proto dhcp metric 100 */
    if (!have_gateway &&
        (snc_parse_ipv4_from_line("default via",  false, line, &net_config->ip_gateway) == sncr_success))
    {
      have_gateway = true;
    }
  }
  result = (have_gateway) ? sncr_success : sncr_notfound;

sgnci_fail:
  SNC_PCLOSE_AND_NULLIFY(ip_output);
  SNC_FREE_AND_NULLIFY(line);

  return result;
}


/** Get the current network settings from parsing the output from the ip command */
enum e_snc_result snc_get_net_current_ip_command(const struct s_snc_configuration *config, const char *eth_interface, struct s_snc_net_config *net_current)
{
  enum e_snc_result result;

  /* Set some initial values */
  result = snc_net_config_struct_init(net_current);
  if (sncr_success != result)
    return result;

  /* Get the DHCP setting from our configuration */
  result = snc_get_net_current_dhcp(config, SNC_METHOD_IPCMD, eth_interface, &net_current->use_dhcp);
  if (sncr_success != result)
    return result;

  /* Get ip, netmask, broadcast and mac from "ip address" */
  result = snc_get_net_current_ip_address_command(config->ip_cmd, eth_interface, net_current);
  if (sncr_success != result)
    return result;

  /* Get default gateway from "ip route" */
  result = snc_get_net_current_ip_route_command(config->ip_cmd, eth_interface, net_current);
  if (sncr_success != result && sncr_notfound != result)
    return result;

  /* Parse nameservers and domain from resolv.conf */
  result = snc_ipv4_nameservers_domain_get_current_resolvconf(config->resolv_file, net_current);
  if (sncr_success != result)
    return result;

  /* Get the hostname from the system */
  result = snc_get_hostname_syscall(net_current->hostname, SNC_HOST_NAME_MAX);
  if (sncr_success != result)
    return result;

  return sncr_success;
}
