/*
 * lsnc_netcurrent.c
 *
 * Get the current network settings programatically, using the socket interface
 *
 *  Created on: 31 May 2018
 *      Author: ian
 */
#include "config.h"
#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#include <netinet/if_ether.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#ifdef HAVE_IFADDRS_H
#include <ifaddrs.h>
#endif
#include "lsnc_nameservers.h"
#include "lsnc_gateway.h"
#include "lsnc_internal.h"
#include "lsnc_dhcp.h"

#include <setnetconfig.h>


static enum e_snc_result snc_get_net_current_nameservers_domain_hostname(const struct s_snc_configuration *config, struct s_snc_net_config *net_config)
{
  enum e_snc_result result;

  /* Get our nameservers */
  result = snc_ipv4_nameservers_domain_get_current(config, net_config);
  if (sncr_success != result)
    return result;

  /* Get the hostname from the system */
  result = snc_get_hostname_syscall(net_config->hostname, SNC_HOST_NAME_MAX);
  if (sncr_success != result)
    return result;

  return sncr_success;
}


/** Run an ifreq IPv4 IOCTL */
static enum e_snc_result snc_get_net_current_do_ipv4_ifreq_ioctl(int sockfd, int ifreq_request, struct ifreq *ifr, const char *eth_interface)
{
  /* Fill out request for IPv4 */
  memset(ifr, 0, sizeof(*ifr));
  ifr->ifr_addr.sa_family = AF_INET;
  snprintf(ifr->ifr_name, IFNAMSIZ, "%s", eth_interface);

  return (ioctl(sockfd, ifreq_request, ifr) < 0) ? sncr_syscallerror : sncr_success;
}


/** Get the MAC address via the SIOCGIFHWADDR ioctl */
static enum e_snc_result snc_get_net_current_mac_ioctl(int sockfd, const char *eth_interface, struct sockaddr *mac_address)
{
  struct ifreq ifr;
  enum e_snc_result result;

  result = snc_get_net_current_do_ipv4_ifreq_ioctl(sockfd, SIOCGIFHWADDR, &ifr, eth_interface);
  if (sncr_success != result)
    return result;
  mac_address->sa_family = AF_PACKET;
  memcpy(mac_address->sa_data, ifr.ifr_hwaddr.sa_data, ETH_ALEN);
  memset(&mac_address->sa_data[ETH_ALEN], 0, sizeof(mac_address->sa_data) - ETH_ALEN);
  return sncr_success;
}


#ifdef HAVE_IFADDRS_H
/** Use getifaddrs() for ip, netmask, broadcast if it is available */
enum e_snc_result snc_get_net_current_getifaddrs(const struct s_snc_configuration *config, const char *eth_interface, struct s_snc_net_config *net_current)
{
  struct ifaddrs *ifaddr = NULL;
  struct ifaddrs *ifa;
  enum e_snc_result result;
  int sockfd = -1;

  /* Set some initial values */
  result = snc_net_config_struct_init(net_current);
  if (sncr_success != result)
    return result;

  /* Get the DHCP setting from our configuration */
  result = snc_get_net_current_dhcp(config, SNC_METHOD_GETIFADDRS, eth_interface, &net_current->use_dhcp);
  if (sncr_success != result)
    return result;

  if (getifaddrs(&ifaddr) < 0)
  {
    result = sncr_syscallerror;
    goto sgncg_fail;
  }

  for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next)
  {
    if (ifa->ifa_addr == NULL)
      continue;

    /* s=getnameinfo(ifa->ifa_addr, sizeof(struct sockaddr_in), host, NI_MAXHOST, NULL, 0, NI_NUMERICHOST); */

    if (strcmp(ifa->ifa_name, eth_interface) == 0)
    {
      if (ifa->ifa_addr->sa_family == AF_INET)
      {
        net_current->ip_address = ((struct sockaddr_in *) ifa->ifa_addr)->sin_addr;
        net_current->ip_netmask = ((struct sockaddr_in *) ifa->ifa_netmask)->sin_addr;
        if ((ifa->ifa_flags & IFF_BROADCAST) != 0)
          net_current->ip_broadcast = ((struct sockaddr_in *) ifa->ifa_broadaddr)->sin_addr;
/*      else if ((ifa->ifa_flags & IFF_POINTOPOINT) != 0)
          net_config->ip_broadcast = ((struct sockaddr_in *) ifa->ifa_dstaddr)->sin_addr; */
        else
          net_current->ip_broadcast.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
      }
#if 0
      if (ifa->ifa_addr->sa_family == AF_LINK)
      {
        /* BSD only */
      }
#endif
    }
  }

  /* Now get our nameservers and hostname */
  result = snc_get_net_current_nameservers_domain_hostname(config, net_current);
  if (sncr_success != result)
    goto sgncg_fail;

  /* Get the current IPv4 default gateway using the PF_NETLINK socket interface */
  result = snc_ipv4_gateway_get_current(eth_interface, &net_current->ip_gateway);
  if (sncr_success != result && sncr_notfound != result)
    goto sgncg_fail;

  /* MAC address is only available via ioctl() */
  sockfd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sockfd < 0)
  {
    result = sncr_syscallerror;
    goto sgncg_fail;
  }
  result = snc_get_net_current_mac_ioctl(sockfd, eth_interface, &net_current->mac_address);

sgncg_fail:
  SNC_DESTROY_AND_NULLIFY(freeifaddrs, ifaddr);
  SNC_CLOSE_AND_INVALIDATE(sockfd);
  return result;
}
#endif /* #ifdef HAVE_IFADDRS_H */


/** Use ioctl for ip, netmask, broadcast, mac address if it is available */
enum e_snc_result snc_get_net_current_ioctl(const struct s_snc_configuration *config, const char *eth_interface, struct s_snc_net_config *net_current)
{
  int sockfd;
  struct ifreq ifr;
  enum e_snc_result result;

  /* Set some initial values */
  result = snc_net_config_struct_init(net_current);
  if (sncr_success != result)
    return result;

  /* Get the DHCP setting from our configuration */
  result = snc_get_net_current_dhcp(config, SNC_METHOD_IOCTL, eth_interface, &net_current->use_dhcp);
  if (sncr_success != result)
    return result;

  sockfd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sockfd < 0)
  {
    result = sncr_syscallerror;
    goto sgnci_fail;
  }

  /* IP address */
  result = snc_get_net_current_do_ipv4_ifreq_ioctl(sockfd, SIOCGIFADDR, &ifr, eth_interface);
  if (sncr_success != result)
    goto sgnci_fail;
  net_current->ip_address = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr;

  /* Netmask */
  result = snc_get_net_current_do_ipv4_ifreq_ioctl(sockfd, SIOCGIFNETMASK, &ifr, eth_interface);
  if (sncr_success != result)
    goto sgnci_fail;
  net_current->ip_netmask = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr;

  /* Broadcast address */
  result = snc_get_net_current_do_ipv4_ifreq_ioctl(sockfd, SIOCGIFBRDADDR, &ifr, eth_interface);
  if (sncr_success != result)
    goto sgnci_fail;
  net_current->ip_broadcast = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr;

  /* MAC address */
  result = snc_get_net_current_mac_ioctl(sockfd, eth_interface, &net_current->mac_address);
  if (sncr_success != result)
    goto sgnci_fail;

  /* Now get our nameservers and hostname */
  result = snc_get_net_current_nameservers_domain_hostname(config, net_current);
  if (sncr_success != result)
    goto sgnci_fail;

  /* Get the current IPv4 default gateway using the PF_NETLINK socket interface */
  result = snc_ipv4_gateway_get_current(eth_interface, &net_current->ip_gateway);
  if (sncr_success != result && sncr_notfound != result)
    goto sgnci_fail;

  /* Success */
  result = sncr_success;

sgnci_fail:
  SNC_CLOSE_AND_INVALIDATE(sockfd);

  return result;
}
