/*
 * lsnc_parse_ip.c
 *
 *  Created on: 1 Jun. 2018
 *      Author: ian
 */
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <setnetconfig.h>

#include <lsnc_parse_addr.h>


/** Skip over any ' ', '\t'. Return sncr_notfound if no whitespace was found, sncr_success if it was */
enum e_snc_result snc_skip_whitespace(const char **line)
{
  const char *ptr;
  enum e_snc_result result;

  for (ptr = *line; ' ' == *ptr || '\t' == *ptr; ptr++)
    ;
  result = (*line == ptr) ? sncr_notfound : sncr_success;
  *line = ptr;
  return result;
}


/** Convert an IPv4 prefix to a netmask in network byte order */
enum e_snc_result snc_ipv4_prefix_to_netmask(unsigned prefix, struct in_addr *ipv4_netmask)
{
  unsigned i;
  in_addr_t bit_mask;

  if (prefix > 32 || NULL == ipv4_netmask)
    return sncr_paramerror;

  for (i = 0, bit_mask = 0; i < prefix; i++)
  {
    bit_mask >>= 1;
    bit_mask |= 0x80000000;
  }
  ipv4_netmask->s_addr = htonl(bit_mask);
  return sncr_success;
}


/** Convert a netmask in network byte order to an IPv4 prefix */
enum e_snc_result snc_ipv4_netmask_to_prefix(const struct in_addr *ipv4_netmask, unsigned *prefix)
{
  unsigned pfix;
  in_addr_t bit_mask;

  if (NULL == ipv4_netmask || NULL == prefix)
    return sncr_paramerror;

  bit_mask = ntohl(ipv4_netmask->s_addr);

  for (pfix = 0; (bit_mask & 0x80000000) != 0; )
  {
    bit_mask <<= 1;
    pfix++;
  }
  *prefix = pfix;
  return sncr_success;
}


/**
 * Look for the optional prefix in a line of text
 *
 * @param prefix is the prefix to search for. If NULL then parsing the string begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the string from
 */
static enum e_snc_result snc_skip_prefix(const char *prefix, bool prefix_at_line_start, const char **line)
{
  const char *posn;

  /* Find prefix in line, then skip over it */
  if (NULL != prefix)
  {
    posn = strstr(*line, prefix);
    if (NULL == posn)
      return sncr_notfound;
    if (prefix_at_line_start && posn != *line)
      return sncr_notfound;
    posn += strlen(prefix);
  }
  else
  {
    posn = *line;
  }
  *line = posn;
  return sncr_success;
}


/**
 * Parse a non-whitespace string from a line of text, optionally prefixed by prefix
 *
 * @param prefix is the prefix to search for, before parsing the string address. If NULL then parsing the string begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the string from
 * @param str is the string to copy the parsed string into
 * @param strmaxlen is the maximum length of str
 */
enum e_snc_result snc_parse_string_from_line(const char *prefix, bool prefix_at_line_start, const char *line, char *str, size_t strmaxlen)
{
  enum e_snc_result result;
  const char *posn;
  char fmtstr[20];

  /* Find prefix in line, then skip over it */
  posn = line;
  result = snc_skip_prefix(prefix, prefix_at_line_start, &posn);
  if (sncr_success != result)
    return result;

  /* sscanf string from string: " %<strmaxlen>s" */
  snprintf(fmtstr, 20, " %%%us", (unsigned) strmaxlen);
  return (sscanf(posn, fmtstr, str) == 1) ? sncr_success : sncr_parseerror;
}


/**
 * Parse an IPv4 address from a line of text, optionally prefixed by prefix
 *
 * @param prefix is the prefix to search for, before parsing the IPv4 address. If NULL then parsing the IPv4 begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the IP address from
 * @param ipv4_addr is set to the parsed IP address
 */
enum e_snc_result snc_parse_ipv4_from_line(const char *prefix, bool prefix_at_line_start, const char *line, struct in_addr *ipv4_addr)
{
  enum e_snc_result result;
  const char *posn;
  char ipstr[SNC_INET_ADDRSTRLEN + 1];

  /* Find prefix in line, then skip over it */
  posn = line;
  result = snc_skip_prefix(prefix, prefix_at_line_start, &posn);
  if (sncr_success != result)
    return result;

  /* sscanf ip address from string: " %16s" */
  if (sscanf(posn, " %" SNC_STRINGIFY(SNC_INET_ADDRSTRLEN) "s", ipstr) != 1)
    return sncr_parseerror;

  /* Try to convert the result into an IPv4 address */
  return (inet_pton(AF_INET, ipstr, ipv4_addr) == 1) ? sncr_success : sncr_parseerror;
}


/**
 * Parse an IPv4 CIDR (192.168.1.100/24) address from a line of text, prefixed by prefix
 *
 * @param prefix is the prefix to search for, before parsing the IPv4 address. If NULL then parsing the IPv4 begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the IP address from
 * @param ipv4_addr is set to the parsed IP address
 * @param ipv4_mask is set to the parsed IP mask
 */
enum e_snc_result snc_parse_ipv4_cidr_from_line(const char *prefix, bool prefix_at_line_start, const char *line, struct in_addr *ipv4_addr, struct in_addr *ipv4_mask)
{
  const char *posn;
  char ipstr[SNC_INET_ADDRSTRLEN + 1];
  unsigned bit_length;
  const char *sep_posn;
  intptr_t len;
  enum e_snc_result result;

  /* Find prefix in line, then skip over it */
  posn = line;
  result = snc_skip_prefix(prefix, prefix_at_line_start, &posn);
  if (sncr_success != result)
    return result;

  /* Skip any whitespace */
  (void) snc_skip_whitespace(&posn);

  /* Look for ip/bit length separator */
  sep_posn = strchr(posn, '/');
  if (NULL == sep_posn)
    return sncr_notfound;

  /* Copy ip address part to ipstr, for conversion later */
  len = sep_posn - posn;
  if (len > SNC_INET_ADDRSTRLEN - 1)
    return sncr_string_too_long;
  snprintf(ipstr, SNC_INET_ADDRSTRLEN, "%.*s", (int) len, posn);

  /* Convert bit length to integer, then to subnet mask */
  if (sscanf(&sep_posn[1], "%u", &bit_length) != 1)
    return sncr_parseerror;
  result = snc_ipv4_prefix_to_netmask(bit_length, ipv4_mask);
  if (sncr_success != result)
    return result;

  /* Try to convert ipstr into an IPv4 address */
  return (inet_pton(AF_INET, ipstr, ipv4_addr) == 1) ? sncr_success : sncr_parseerror;
}


/**
 * Parse a MAC address from a line of text, optionally prefixed by prefix
 *
 * @param prefix is the prefix to search for, before parsing the mac address. If NULL then parsing the mac begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the mac address from
 * @param mac_address is set to the parsed mac address
 */
enum e_snc_result snc_parse_mac_from_line(const char *prefix, bool prefix_at_line_start, const char *line, struct sockaddr *mac_address)
{
  #define ETH_ADDRSTRLEN (SNC_ETH_ALEN * 3)
  char macstr[ETH_ADDRSTRLEN + 1];
  int res;
  unsigned mac[SNC_ETH_ALEN];
  const char *posn;
  enum e_snc_result result;

  /* Find prefix in line, then skip over it */
  posn = line;
  result = snc_skip_prefix(prefix, prefix_at_line_start, &posn);
  if (sncr_success != result)
    return result;

  /* sscanf mac address from string: " %18s" */
  if (sscanf(posn, " %18s", macstr) != 1)
    return sncr_parseerror;

  /* Convert to MAC address */
  res = sscanf(macstr, "%2x:%2x:%2x:%2x:%2x:%2x",
               &mac[0], &mac[1], &mac[2],
               &mac[3], &mac[4], &mac[5]);
  if (res != SNC_ETH_ALEN)
    return sncr_parseerror;

  /* Copy results over */
  return snc_set_mac_address(mac, mac_address);
}
