/*
 * lsnc_parse_ip.h
 *
 *  Created on: 1 Jun. 2018
 *      Author: ian
 */

#ifndef LSNC_PARSE_ADDR_H_
#define LSNC_PARSE_ADDR_H_

#include <stddef.h>
#include <stdbool.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <setnetconfig.h>


#define SNC_STRINGIFY(x) SNC_INDIRECT(x)
#define SNC_INDIRECT(x) #x


/** Skip over any ' ', '\t'. Return sncr_notfound if no whitespace was found, sncr_success if it was */
enum e_snc_result snc_skip_whitespace(const char **line);


/** Convert an IPv4 prefix to a netmask in network byte order */
enum e_snc_result snc_ipv4_prefix_to_netmask(unsigned prefix, struct in_addr *ipv4_netmask);


/** Convert a netmask in network byte order to an IPv4 prefix */
enum e_snc_result snc_ipv4_netmask_to_prefix(const struct in_addr *ipv4_netmask, unsigned *prefix);


/**
 * Parse a non-whitespace string from a line of text, optionally prefixed by prefix
 *
 * @param prefix is the prefix to search for, before parsing the string address. If NULL then parsing the string begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the string from
 * @param str is the string to copy the parsed string into
 * @param strmaxlen is the maximum length of str
 */
enum e_snc_result snc_parse_string_from_line(const char *prefix, bool prefix_at_line_start, const char *line, char *str, size_t strmaxlen);


/**
 * Parse an IPv4 address from a line of text, optionally prefixed by prefix
 *
 * @param prefix is the prefix to search for, before parsing the IPv4 address. If NULL then parsing the IPv4 begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the IP address from
 * @param ipv4_addr is set to the parsed IP address
 */
enum e_snc_result snc_parse_ipv4_from_line(const char *prefix, bool prefix_at_line_start, const char *line, struct in_addr *ipv4_addr);


/**
 * Parse an IPv4 CIDR (192.168.1.100/24) address from a line of text, prefixed by prefix
 *
 * @param prefix is the prefix to search for, before parsing the IPv4 address. If NULL then parsing the IPv4 begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the IP address from
 * @param ipv4_addr is set to the parsed IP address
 * @param ipv4_mask is set to the parsed IP mask
 */
enum e_snc_result snc_parse_ipv4_cidr_from_line(const char *prefix, bool prefix_at_line_start, const char *line, struct in_addr *ipv4_addr, struct in_addr *ipv4_mask);


/**
 * Parse a MAC address from a line of text, optionally prefixed by prefix
 *
 * @param prefix is the prefix to search for, before parsing the mac address. If NULL then parsing the mac begins at start of line
 * @param prefix_at_line_start means the prefix must start at the beginning of line
 * @param line is the line to parse the mac address from
 * @param mac_address is set to the parsed mac address
 */
enum e_snc_result snc_parse_mac_from_line(const char *prefix, bool prefix_at_line_start, const char *line, struct sockaddr *mac_address);


#endif /* LSNC_PARSE_ADDR_H_ */
