/*
 * lsnc_print_net_config.c
 *
 *  Created on: 5 Jun. 2018
 *      Author: ian
 */
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include "lsnc_internal.h"

#include <setnetconfig.h>


enum e_snc_print_format
{
  spf_ifcfg = 0,
  spf_dns,
  spf_resolv,

  spf_num
};


static const char *const snc_multi_line_formats[spf_num] =
{
  "Use DHCP:    %s\n"
  "IP Address:  %s\n"
  "Netmask:     %s\n"
  "Broadcast:   %s\n"
  "Gateway:     %s\n",

  "DNS %u:       %s\n",

  "Search:      %s\n"
  "Hostname:    %s\n"
  "Domain:      %s\n"
  "MAC Address: %02X:%02X:%02X:%02X:%02X:%02X\n"
};


static const char *const snc_single_line_formats[spf_num] =
{
  "dhcp: %s ip: %s nm: %s bc: %s gw: %s ",

  "ns%u: %s ",

  "sd: \"%s\" hn: \"%s\" dn: \"%s\" mac: %02X:%02X:%02X:%02X:%02X:%02X"
};


/**
 * Convert the network config to a printable string for debugging purposes
 *
 * @param net_config is the network configuration to print
 * @str is the string to print into
 * @str_len is a pointer to the length of str, and is set to the length of the printout by this function
 */
static enum e_snc_result snc_internal_net_config_to_str(const struct s_snc_net_config *net_config, const char *const formats[spf_num], char *str, size_t *str_len)
{
  char ip[SNC_INET_ADDRSTRLEN];
  char nm[SNC_INET_ADDRSTRLEN];
  char bc[SNC_INET_ADDRSTRLEN];
  char gw[SNC_INET_ADDRSTRLEN];
  size_t i;
  int len;
  int res;
  size_t str_max;
  size_t room_left;
  char *ptr;

  str_max = *str_len;
  len = 0;
  ptr = str;
  room_left = str_max;
  res = snprintf(ptr, room_left,
                 formats[spf_ifcfg],
                 (net_config->use_dhcp) ? "yes" : "no",
                 inet_ntop(AF_INET, &net_config->ip_address, ip, SNC_INET_ADDRSTRLEN),
                 inet_ntop(AF_INET, &net_config->ip_netmask, nm, SNC_INET_ADDRSTRLEN),
                 inet_ntop(AF_INET, &net_config->ip_broadcast, bc, SNC_INET_ADDRSTRLEN),
                 inet_ntop(AF_INET, &net_config->ip_gateway, gw, SNC_INET_ADDRSTRLEN));
  if (res < 0)
    return sncr_syscallerror;
  len += res;

  for (i = 0; i < SNC_NUM_DNS_SERVERS; i++)
  {
    ptr = (NULL == str) ? NULL : &str[len];
    room_left = (0 == str_max) ? 0 : str_max - len;
    res = snprintf(ptr, room_left,
                   formats[spf_dns],
                   (unsigned) i, inet_ntop(AF_INET, &net_config->ip_dns[i], ip, SNC_INET_ADDRSTRLEN));
    if (res < 0)
      return sncr_syscallerror;
    len += res;
  }

  ptr = (NULL == str) ? NULL : &str[len];
  room_left = (0 == str_max) ? 0 : str_max - len;
  res = snprintf(ptr, room_left,
                 formats[spf_resolv],
                 net_config->search_domain,
                 net_config->hostname,
                 net_config->domain,
                 (unsigned) (unsigned char) net_config->mac_address.sa_data[0],
                 (unsigned) (unsigned char) net_config->mac_address.sa_data[1],
                 (unsigned) (unsigned char) net_config->mac_address.sa_data[2],
                 (unsigned) (unsigned char) net_config->mac_address.sa_data[3],
                 (unsigned) (unsigned char) net_config->mac_address.sa_data[4],
                 (unsigned) (unsigned char) net_config->mac_address.sa_data[5]);
  if (res < 0)
    return sncr_syscallerror;
  len += res;

  /* Success */
  *str_len = len;
  return sncr_success;
}


/**
 * Convert the network config to a printable string for debugging purposes
 *
 * @param net_config is the network configuration to print
 * @param single_not_multiline is whether to convert the net_config into a single line or multiple line string
 * @param str is the string to print into
 * @param str_len is the maximum length of str
 */
enum e_snc_result snc_net_config_to_str(const struct s_snc_net_config *net_config, bool single_not_multiline, char *str, size_t str_len)
{
  const char *const *const formats = (single_not_multiline) ? snc_single_line_formats : snc_multi_line_formats;

  if (NULL == net_config)
    return sncr_paramerror;

  return snc_internal_net_config_to_str(net_config, formats, str, &str_len);
}


/**
 * Allocate a suitably sized string and convert the network config to a printable string for debugging purposes
 *
 * Caller is responsible for freeing the returned string
 *
 * @param net_config is the network configuration to print
 * @param single_not_multiline is whether to convert the net_config into a single line or multiple line string
 * @param str is a pointer to a pointer to a string that gets set to the newly allocated string containing the network configuration
 */
enum e_snc_result snc_net_config_to_allocstr(const struct s_snc_net_config *net_config, bool single_not_multiline, char **str)
{
  enum e_snc_result result;
  size_t str_len;
  const char *const *const formats = (single_not_multiline) ? snc_single_line_formats : snc_multi_line_formats;

  if (NULL == net_config || NULL == str)
    return sncr_paramerror;

  *str = NULL;

  /* How long would it be */
  str_len = 0;
  result = snc_internal_net_config_to_str(net_config, formats, NULL, &str_len);
  if (sncr_success != result)
    goto sncta_fail;

  /* Allocate enough string */
  str_len++;
  *str = calloc(str_len, sizeof(char));
  if (NULL == *str)
  {
    result = sncr_allocerror;
    goto sncta_fail;
  }

  /* Try again and actually convert into buffer this time */
  result = snc_internal_net_config_to_str(net_config, formats, *str, &str_len);
  if (sncr_success != result)
    goto sncta_fail;

  /* Success */
  return sncr_success;

sncta_fail:
  SNC_FREE_AND_NULLIFY(*str);
  return result;
}


/**
 * Print out the network config for debugging purposes
 *
 * @param to is the FILE to write the network config to
 * @param single_not_multiline is whether to convert the net_config into a single line or multiple line string
 * @param net_config is the network configuration to print
 */
enum e_snc_result snc_print_net_config(FILE *to, bool single_not_multiline, const struct s_snc_net_config *net_config)
{
  char *str = NULL;
  enum e_snc_result result;

  if (NULL == to || NULL == net_config)
    return sncr_paramerror;

  /* Allocate string and convert net_config to string */
  result = snc_net_config_to_allocstr(net_config, single_not_multiline, &str);
  if (sncr_success != result)
    goto spnc_fail;

  /* Print it */
  result = (fprintf(to, "%s", str) < 0) ? sncr_syscallerror : sncr_success;

  /* Clean up */
spnc_fail:
  SNC_FREE_AND_NULLIFY(str);
  return result;
}


/**
 * Print out the network config in a single line (for legacy getnetcurrent and legacy getnetconfig)
 *
 * @param to is the FILE to write the network config to
 * @param net_config is the network configuration to print
 * @param ip_as_dhcp if set causes the IP address field to be "dhcp" if the network configuration is set to use dhcp
 */
enum e_snc_result snc_print_legacy_net_config_line(FILE *to, const struct s_snc_net_config *net_config, bool ip_as_dhcp)
{
  char ipstr[SNC_INET_ADDRSTRLEN];

  if (NULL == to || NULL == net_config)
    return sncr_paramerror;

  if (ip_as_dhcp && net_config->use_dhcp)
  {
    if (fprintf(to, "%s ", SNC_RCCONF_VARVALUE_IPADDR_DHCP) < 0)
      return sncr_fileerror;
  }
  else
  {
    if (fprintf(to, "%s ", inet_ntop(AF_INET, &net_config->ip_address, ipstr, SNC_INET_ADDRSTRLEN)) < 0)
      return sncr_fileerror;
  }

  if (fprintf(to, "%s ", inet_ntop(AF_INET, &net_config->ip_netmask, ipstr, SNC_INET_ADDRSTRLEN)) < 0)
    return sncr_fileerror;
  if (fprintf(to, "%s ", inet_ntop(AF_INET, &net_config->ip_broadcast, ipstr, SNC_INET_ADDRSTRLEN)) < 0)
    return sncr_fileerror;
  if (fprintf(to, "%s ", inet_ntop(AF_INET, &net_config->ip_gateway, ipstr, SNC_INET_ADDRSTRLEN)) < 0)
    return sncr_fileerror;
  if (fprintf(to, "%s ", inet_ntop(AF_INET, &net_config->ip_dns[0], ipstr, SNC_INET_ADDRSTRLEN)) < 0)
    return sncr_fileerror;

  if (fprintf(to, "%02X:%02X:%02X:%02X:%02X:%02X ",
              (unsigned) (unsigned char) net_config->mac_address.sa_data[0],
              (unsigned) (unsigned char) net_config->mac_address.sa_data[1],
              (unsigned) (unsigned char) net_config->mac_address.sa_data[2],
              (unsigned) (unsigned char) net_config->mac_address.sa_data[3],
              (unsigned) (unsigned char) net_config->mac_address.sa_data[4],
              (unsigned) (unsigned char) net_config->mac_address.sa_data[5]) < 0)
    return sncr_fileerror;

  if (fprintf(to, "%s", net_config->hostname) < 0)
    return sncr_fileerror;

  if (fprintf(to, "\n") < 0)
    return sncr_fileerror;

  return sncr_success;
}
