/*
 * Simple parsing of variables and values from the rc.conf file
 *
 * This doesn't support anything fancy like multiple lines '\', or escaping in quoted strings "\",
 * or mixed quoting and unquoted "asdf"fd
 */

#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include "lsnc_parse_addr.h"
#include "lsnc_internal.h"

#include <setnetconfig.h>
#include "lsnc_rcconf.h"


/** An rc.conf network configuration variable */
enum e_snc_rcconf_net_config_variable
{
  NETCFG_VAR_IFACE = 0,
  NETCFG_VAR_IPADDR,
  NETCFG_VAR_NETMASK,
  NETCFG_VAR_BROADCAST,
  NETCFG_VAR_GATEWAY,
  NETCFG_VAR_NAMESERVER,
  NETCFG_VAR_NAMESERVER2,
  NETCFG_VAR_NAMESERVER3,
  NETCFG_VAR_SEARCHDOMAIN,
  NETCFG_VAR_HOSTNAME,
  NETCFG_VAR_DOMAINNAME,
  NUM_NETCFG_VARS
};


/** An rc.conf variable NAME=VALUE */
struct s_snc_rcconf_variable
{
  char *variable_name;      /*!< The variable name */
  char *variable_value;     /*!< The variable value */
  bool write_me;            /*!< Should this be written to the replacement rc.conf file */
  bool strip_me;            /*!< Should this be stripped from the replacement rc.conf file */
};


/*
 * rc.conf typically looks like:
 *
all_services="blah1 blah2"
all_services_r="blah2 blah1"

cfg_services="blah1 blah2"
cfg_services_r="blah2 blah1"

upg_kill_services="blah1 blah2"

export NTP_SERVER=""
export MODLIST=""
export RAMDIRS="/tmp /var"
export TMPFS="ramfs"
export TMPFS_SIZE="512k"
export READONLY_FS=""
export INETD_ARGS=""
export BOA_ARGS="-c /etc"
export SMBD_ARGS=""
export NMBD_ARGS=""
export DHCP_ARG=""
export DEPLOYMENT_STYLE="JFFS2"
export SYSLOGD_OPTIONS=""
export KLOGD_OPTIONS=""
export SYSCFG_DHCPC_CMD="udhcpc --syslog -x hostname:$HOSTNAME --interface "
# net interface 0
export SYSCFG_IFACE0=y
export INTERFACE0="eth0"
export IPADDR0="dhcp"
export HOSTNAME="c520-3477741"
*/


static const char snc_export_str[] = "export";


/**
 * Copy data from an unquoted line
 *
 * eg:
 * ASDF=1234
 */
static enum e_snc_result snc_rcconf_copy_value_unquoted(const char **line, char *data, size_t datalen)
{
  const char *ptr;
  enum e_snc_result result;
  size_t count;

  if (NULL != data)
  {
    /* Copy over to data */
    for (ptr = *line, count = 0;
         count < datalen && '\0' != *ptr && ' ' != *ptr && '\t' != *ptr && ';' != *ptr && '\n' != *ptr;
         data++, ptr++, count++)
    {
      *data = *ptr;
    }
    *data = '\0';
    result = (count < datalen) ? sncr_success : sncr_string_too_long;
  }
  else
  {
    /* Don't copy, just parse */
    for (ptr = *line;
         '\0' != *ptr && ' ' != *ptr && '\t' != *ptr && ';' != *ptr && '\n' != *ptr;
         ptr++)
      ;
    result = sncr_success;
  }
  *line = ptr;
  return result;
}


/**
 * Copy data from a quoted line
 *
 * eg:
 * ASDF='1234'
 * ASDF="1234"
 */
static enum e_snc_result snc_rcconf_copy_value_quoted(const char **line, char quote_char, char *data, size_t datalen)
{
  const char *ptr;
  enum e_snc_result result;
  size_t count;

  if (NULL != data)
  {
    /* Copy over to data */
    for (ptr = *line, count = 0;
         count < datalen && '\0' != *ptr && quote_char != *ptr && '\n' != *ptr;
         data++, ptr++, count++)
    {
      *data = *ptr;
    }
    *data = '\0';
    result = (count < datalen) ? sncr_success : sncr_string_too_long;
  }
  else
  {
    /* Don't copy, just parse */
    for (ptr = *line;
         '\0' != *ptr && quote_char != *ptr && '\n' != *ptr;
         ptr++)
      ;
    result = sncr_success;
  }
  *line = ptr;
  return result;
}


/**
 * Get the value of an exported variable from a line
 *
 * @param line is the line to parse
 * @param varname is the variable name to look for in the line
 * @param data is where the variable value is copied if not NULL (if NULL datalen is not used)
 * @param datalen is the maximum length of data
 * @param is_export if not NULL, is set if this variable is exported ie "export VARNAME=VARVALUE"
 * @return one of the @ref enum e_snc_result results
 */
enum e_snc_result snc_rcconf_get_variable_value_from_line(const char *line, const char *varname, char *data, size_t datalen, bool *is_export)
{
  const char *lineptr;
  const char *exportptr;
  char quote_char;
  enum e_snc_result result = sncr_notfound;
  bool is_export_int = false;

  lineptr = line;

  /* Skip any preceding whitespace */
  (void) snc_skip_whitespace(&lineptr);

  /* Is it an exported variable with whitespace following */
  if (lineptr == strstr(lineptr, snc_export_str))
  {
    exportptr = lineptr;

    /* Skip over "export"*/
    lineptr += strlen(snc_export_str);

    /*
     * Look for whitespace following "export"
     * If no whitespace following, then this is a variable name beginning with "export",
     * so back up to the start of export
     */
    if (snc_skip_whitespace(&lineptr) == sncr_notfound)
    {
      lineptr = exportptr;
    }
    else
    {
      is_export_int = true;
    }
  }

  /* Is it the variable name? */
  if (lineptr == strstr(lineptr, varname))
  {
    /* Skip over variable name */
    lineptr += strlen(varname);

    /* There cannot be any whitespace between variable name and '=' */

    /* Check for equals */
    if ('=' == *lineptr)
    {
      /* Skip over '=' */
      lineptr++;

      /* There cannot be any whitespace between '=' and the variable value */

      /* Skip over (optional) quoting '"' or '\'' */
      if ('"' == *lineptr || '\'' == *lineptr)
      {
        quote_char = *lineptr;
        lineptr++;
        result = snc_rcconf_copy_value_quoted(&lineptr, quote_char, data, datalen);
      }
      else
      {
        result = snc_rcconf_copy_value_unquoted(&lineptr, data, datalen);
      }
    } /* match '=' */
  } /* match variable name */

  if (NULL != is_export)
    *is_export = is_export_int;

  return result;
}


/** Strip the existing config from the file */
static enum e_snc_result snc_rcconf_strip_existing_config(FILE *rcconf, FILE *tmpfile, char *line, struct s_snc_rcconf_variable **vars, size_t numvars)
{
  enum e_snc_result result;
  bool dont_write;
  size_t i;

  /* Back to start of rc.conf */
  if (fseek(rcconf, 0, SEEK_SET) < 0)
  {
    result = sncr_fileerror;
    goto srsec_fail;
  }

  /*
   * Read each line from the config file, and save straight to the
   * temporary file if it's not one of the network configuration
   * variables we want to change
   */
  for ( ; ; )
  {
    /* Read a line */
    if (NULL == fgets(line, SNC_PARSE_LINE_LENGTH, rcconf))
    {
      if (feof(rcconf))
      {
        /* Success */
        result = sncr_success;
        break;
      }
      else /* if (ferror(cfgfile)) */
      {
        /* Fail */
        result = sncr_fileerror;
        break;
      }
    }

    /*
     * Only save to temporary file if it isn't one of the network
     * interface configuration variables we want to change
     */
    for (i = 0, dont_write = false; i < numvars; i++)
    {
      if (vars[i]->strip_me && snc_rcconf_get_variable_value_from_line(line, vars[i]->variable_name, NULL, 0, NULL) == sncr_success)
      {
        dont_write = true;
        break;
      }
    }

    if (!dont_write)
    {
      /* Write line unchanged out to temporary file */
      if (fputs(line, tmpfile) == EOF)
      {
        /* Fail */
        result = sncr_fileerror;
        break;
      }
    }
  } /* for */

srsec_fail:
  return result;
}


/**
 * Find an variable (varname) in rcconf file, and extract its value into data
 *
 * @param rcconf is a config file open for reading
 * @param line is a pre-allocated line of length SNC_PARSE_LINE_LENGTH for reading each line into
 * @param varname is the variable name to look for
 * @param data is the buffer into which the variable value should be set
 * @param datalen is the maximum length of data
 * @param is_export if not NULL, is set if this variable is exported ie "export VARNAME=VARVALUE"
 */
enum e_snc_result snc_rcconf_find_variable(FILE *rcconf, char *line, const char *varname, char *data, size_t datalen, bool *is_export)
{
  enum e_snc_result result;

  /* Back to start of file */
  if (fseek(rcconf, 0, SEEK_SET) < 0)
    return sncr_fileerror;

  /* Look for pattern */
  for ( ; ; )
  {
    /* Read a line */
    if (NULL == fgets(line, SNC_PARSE_LINE_LENGTH, rcconf))
    {
      if (feof(rcconf))
      {
        /* End of file, and we haven't found it yet */
        result = sncr_notfound;
        break;
      }
      else /* if (ferror(cfgfile)) */
      {
        /* Fail */
        result = sncr_fileerror;
        break;
      }
    }

    /* Parse line to see if it matches our variable name */
    result = snc_rcconf_get_variable_value_from_line(line, varname, data, datalen, is_export);
    if (sncr_success == result)
      break;
  }

  return result;
}


/** Destroy a rcconf variable */
static enum e_snc_result snc_rcconf_variable_destroy(struct s_snc_rcconf_variable *var);


/** Initialise a rcconf variable */
static enum e_snc_result snc_rcconf_variable_init(struct s_snc_rcconf_variable **var, const char *name, const char *value, bool write_me, bool strip_me)
{
  struct s_snc_rcconf_variable *ivar;

  ivar = calloc(1, sizeof(*ivar));
  if (NULL == ivar)
    goto srvi_fail;

  ivar->variable_name = strdup(name);
  if (NULL == ivar->variable_name)
    goto srvi_fail;
  ivar->variable_value = strdup(value);
  if (NULL == ivar->variable_value)
    goto srvi_fail;
  ivar->write_me = write_me;
  ivar->strip_me = strip_me;

  /* Success */
  *var = ivar;
  return sncr_success;

srvi_fail:
  SNC_DESTROY_AND_NULLIFY(snc_rcconf_variable_destroy, ivar);
  *var = NULL;
  return sncr_allocerror;
}


/** Destroy a rcconf variable */
static enum e_snc_result snc_rcconf_variable_destroy(struct s_snc_rcconf_variable *var)
{
  if (NULL != var)
  {
    SNC_FREE_AND_NULLIFY(var->variable_name);
    SNC_FREE_AND_NULLIFY(var->variable_value);
    SNC_FREE_AND_NULLIFY(var);
  }
  return sncr_success;
}


/**
 * Search for matching interface in rc.conf.
 *
 * ie find the line: export INTERFACEx="<iface>", and return the interface index x
 */
static enum e_snc_result snc_rcconf_find_iface_index(FILE *cfgfile, char *line, const char *iface, int *iface_idx)
{
  int i;
  bool found;
  char iface_str[SNC_IF_NAMESIZE];
  char varname[SNC_RCCONF_VARNAME_LENGTH];

  for (i = 0, found = false; i < SNC_RCCONF_MAX_INTERFACES && !found; i++)
  {
    snprintf(varname, SNC_RCCONF_VARNAME_LENGTH, "%s%d", SNC_RCCONF_VARNAME_INTERFACE, i);
    if (snc_rcconf_find_variable(cfgfile, line, varname, iface_str, SNC_IF_NAMESIZE, NULL) == sncr_success &&
        strcmp(iface, iface_str) == 0)
    {
      *iface_idx = i;
      found = true;
    }
  }

  return (found) ? sncr_success : sncr_notfound;
}


/** Get the network configuration from rc.conf */
static enum e_snc_result snc_get_net_config_rcconf_internal(FILE *rcconf, char *line, const char *eth_interface, struct s_snc_net_config *net_config, int *interface_index)
{
  enum e_snc_result result;
  int iface_idx;
  char varname[SNC_RCCONF_VARNAME_LENGTH];
  #define NUM_IFPARAMS 7
  const char *const ifvarfmts[NUM_IFPARAMS] = { SNC_RCCONF_VARNAME_IPADDR "%d", SNC_RCCONF_VARNAME_NETMASK "%d", SNC_RCCONF_VARNAME_BROADCAST "%d", SNC_RCCONF_VARNAME_GATEWAY "%d", SNC_RCCONF_VARNAME_NAMESERVER "%d", SNC_RCCONF_VARNAME_NAMESERVER "%d_2", SNC_RCCONF_VARNAME_NAMESERVER "%d_3" };
  struct in_addr *ifaddrs[NUM_IFPARAMS] =     { &net_config->ip_address,        &net_config->ip_netmask,         &net_config->ip_broadcast,         &net_config->ip_gateway,         &net_config->ip_dns[0],             &net_config->ip_dns[1],               &net_config->ip_dns[2] };
  size_t param;
  char ipstr[SNC_INET_ADDRSTRLEN + 1];

  /* Set everything to defaults */
  result = snc_net_config_struct_init(net_config);
  if (sncr_success != result)
    return result;

  if (NULL != eth_interface)
  {
    /* Search for interface */
    result = snc_rcconf_find_iface_index(rcconf, line, eth_interface, &iface_idx);
    if (sncr_success != result)
      goto sgncr_fail;
    if (NULL != interface_index)
      *interface_index = iface_idx;

    /* Now get each of the parameters for the interface index */
    net_config->use_dhcp = false;
    for (param = 0; param < NUM_IFPARAMS; param++)
    {
      snprintf(varname, SNC_RCCONF_VARNAME_LENGTH, ifvarfmts[param], iface_idx);
      result = snc_rcconf_find_variable(rcconf, line, varname, ipstr, SNC_INET_ADDRSTRLEN, NULL);
      switch (result)
      {
        case sncr_success:
          if (ifaddrs[param] == &net_config->ip_address &&
              strcmp(ipstr, SNC_RCCONF_VARVALUE_IPADDR_DHCP) == 0)
          {
            net_config->use_dhcp = true;
            goto sgncr_unset_ip;
          }
          else if (strcmp(ipstr, "") == 0)
          {
            goto sgncr_unset_ip;
          }
          break;
        case sncr_notfound:
sgncr_unset_ip:
          snprintf(ipstr, SNC_INET_ADDRSTRLEN, "%s", SNC_IPV4_UNSET_STR);
          break;
        default:
          goto sgncr_fail;
          break;
      }

      if (inet_pton(AF_INET, ipstr, ifaddrs[param]) != 1)
      {
        result = sncr_parseerror;
        goto sgncr_fail;
      }
    } /* for */

    /* Search domain */
    snprintf(varname, SNC_RCCONF_VARNAME_LENGTH, SNC_RCCONF_VARNAME_SEARCHDOMAIN "%d", iface_idx);
    if (snc_rcconf_find_variable(rcconf, line, varname, net_config->search_domain, SNC_DOMAIN_NAME_MAX, NULL) != sncr_success ||
        strcmp(net_config->search_domain, "") == 0 ||
        snc_domainname_validate(net_config->hostname) != sncr_success)
    {
      snprintf(net_config->search_domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
    }
  }
  else
  {
    /* No interface specified, set all parameters to defaults (done by snc_net_config_struct_init() above) */
  }

  /* Hostname */
  if (snc_rcconf_find_variable(rcconf, line, SNC_RCCONF_VARNAME_HOSTNAME, net_config->hostname, SNC_HOST_NAME_MAX, NULL) != sncr_success ||
      strcmp(net_config->hostname, "") == 0 ||
      snc_hostname_validate(net_config->hostname) != sncr_success)
  {
    snprintf(net_config->hostname, SNC_HOST_NAME_MAX, "%s", SNC_HOSTNAME_DEFAULT);
  }

  /* Domain name */
  if (snc_rcconf_find_variable(rcconf, line, SNC_RCCONF_VARNAME_DOMAINNAME, net_config->domain, SNC_DOMAIN_NAME_MAX, NULL) != sncr_success ||
      strcmp(net_config->domain, "") == 0 ||
      snc_domainname_validate(net_config->domain) != sncr_success)
  {
    snprintf(net_config->domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
  }

  /* MAC address. Set this to 0, as it's not a configuration parameter */
  result = snc_zero_mac_address(&net_config->mac_address);
  if (sncr_success != result)
    goto sgncr_fail;

  /* Success */
  result = sncr_success;

sgncr_fail:
  return result;
}


/** Allocate a parsing line and open the rc.conf file */
static enum e_snc_result snc_rcconf_alloc_and_open(const char *rcconf_path, char **line, FILE **rcconf)
{
  enum e_snc_result result;

  *line = malloc(SNC_PARSE_LINE_LENGTH);
  if (NULL == *line)
  {
    result = sncr_allocerror;
    goto sraao_fail;
  }

  /* Open rc.conf config file for reading only */
  *rcconf = fopen(rcconf_path, "r");
  if (NULL == *rcconf)
  {
    result = sncr_fileerror;
    goto sraao_fail;
  }

  /* Success */
  return sncr_success;

sraao_fail:
  SNC_FCLOSE_AND_NULLIFY(*rcconf);
  SNC_FREE_AND_NULLIFY(*line);
  return result;
}


/** Get the network configuration from rc.conf */
enum e_snc_result snc_get_net_config_rcconf(const char *rcconf_path, const char *eth_interface, struct s_snc_net_config *net_config)
{
  char *line = NULL;
  FILE *rcconf = NULL;
  enum e_snc_result result;

  /* Allocate parsing line and open rc.conf file */
  result = snc_rcconf_alloc_and_open(rcconf_path, &line, &rcconf);
  if (sncr_success != result)
    goto sgncr_fail;

  /* Parse the config from rc.conf */
  result = snc_get_net_config_rcconf_internal(rcconf, line, eth_interface, net_config, NULL);

sgncr_fail:
  SNC_FCLOSE_AND_NULLIFY(rcconf);
  SNC_FREE_AND_NULLIFY(line);
  return result;
}


/** Get the network configuration DHCP setting from rc.conf */
enum e_snc_result snc_get_net_config_dhcp_rcconf(const char *rcconf_path, const char *eth_interface, bool *use_dhcp)
{
  char *line = NULL;
  FILE *rcconf = NULL;
  enum e_snc_result result;
  int iface_idx;
  char varname[SNC_RCCONF_VARNAME_LENGTH];
  char ipstr[SNC_INET_ADDRSTRLEN + 1];

  if (NULL == rcconf_path || NULL == eth_interface || NULL == use_dhcp)
    return sncr_paramerror;

  /* Allocate parsing line and open rc.conf file */
  result = snc_rcconf_alloc_and_open(rcconf_path, &line, &rcconf);
  if (sncr_success != result)
    goto sgncdr_fail;

  /* Search for interface */
  result = snc_rcconf_find_iface_index(rcconf, line, eth_interface, &iface_idx);
  if (sncr_success != result)
    goto sgncdr_fail;

  /* Now get IP address field for the interface index */
  snprintf(varname, SNC_RCCONF_VARNAME_LENGTH, "%s%d", SNC_RCCONF_VARNAME_IPADDR, iface_idx);
  result = snc_rcconf_find_variable(rcconf, line, varname, ipstr, SNC_INET_ADDRSTRLEN, NULL);
  switch (result)
  {
    case sncr_success:
      *use_dhcp = (strcmp(ipstr, SNC_RCCONF_VARVALUE_IPADDR_DHCP) == 0);
      break;
    case sncr_notfound:
      *use_dhcp = false;
      result = sncr_success;
      break;
    default:
      goto sgncdr_fail;
      break;
  }

sgncdr_fail:
  SNC_FCLOSE_AND_NULLIFY(rcconf);
  SNC_FREE_AND_NULLIFY(line);
  return result;
}


/** Generate a temporary file to save the altered rc.conf into */
static enum e_snc_result snc_rcconf_generate_tmpfile(const char *rcconf_path, char **temp_filename, int *temp_fd, FILE **temp_file)
{
  enum e_snc_result result;
  size_t tempfilename_len;
  const char tempfilepostfix[] = ".tmp.XXXXXX";

  *temp_filename = NULL;
  *temp_fd = -1;
  *temp_file = NULL;

  /* Generate temporary file name - "/etc/rc.d/rc.conf.tmp.XXXXXX" */
  tempfilename_len = strlen(rcconf_path) + strlen(tempfilepostfix) + sizeof(char);
  *temp_filename = malloc(tempfilename_len);
  if (NULL == *temp_filename)
  {
    result = sncr_allocerror;
    goto srgt_fail;
  }
  snprintf(*temp_filename, tempfilename_len, "%s%s", rcconf_path, tempfilepostfix);

  /* Make a temporary file to play with */
  *temp_fd = mkstemp(*temp_filename);
  if (*temp_fd < 0)
  {
    result = sncr_fileerror;
    goto srgt_fail;
  }
  *temp_file = fdopen(*temp_fd, "r+");
  if (NULL == *temp_file)
  {
    result = sncr_fileerror;
    goto srgt_fail;
  }

  return sncr_success;

srgt_fail:
  if (*temp_fd >= 0)
  {
    unlink(*temp_filename);
    SNC_CLOSE_AND_INVALIDATE(*temp_fd);
  }
  SNC_FCLOSE_AND_NULLIFY(*temp_file);
  SNC_FREE_AND_NULLIFY(*temp_filename);
  return result;
}


static enum e_snc_result snc_rcconf_set_config_iface_var(struct s_snc_rcconf_variable **config_var,
                                                         const char *fmtstr, const char *varname, int iface_idx,
                                                         const char *varvalue, bool write_me, bool strip_me)
{
  char var_name[SNC_RCCONF_VARNAME_LENGTH];

  snprintf(var_name, SNC_RCCONF_VARNAME_LENGTH, fmtstr, varname, iface_idx);
  return snc_rcconf_variable_init(config_var, var_name, varvalue, write_me, strip_me);
}


static enum e_snc_result snc_rcconf_set_config_iface_var_inet(struct s_snc_rcconf_variable **config_var,
                                                              const char *fmtstr, const char *varname, int iface_idx,
                                                              const struct in_addr *varvalue, bool write_me, bool strip_me,
                                                              bool empty_if_unset)
{
  char ip_str[SNC_INET_ADDRSTRLEN];

  if (empty_if_unset && SNC_IPV4_UNSET_INADDR_NETWORK == varvalue->s_addr)
    snprintf(ip_str, SNC_INET_ADDRSTRLEN, "%s", "");
  else
    inet_ntop(AF_INET, varvalue, ip_str, SNC_INET_ADDRSTRLEN);
  return snc_rcconf_set_config_iface_var(config_var, fmtstr, varname, iface_idx, ip_str, write_me, strip_me);
}


/** Setup the rc.conf config variables ready to write to the new rc.conf */
static enum e_snc_result snc_rcconf_setup_config_vars(struct s_snc_rcconf_variable *config_vars[NUM_NETCFG_VARS],
                                                      const struct s_snc_net_config *net_config,
                                                      int iface_idx,
                                                      bool changed_hostname, bool changed_ifacecfg, bool changed_domain)
{
  const char *var_value;
  bool strip_me;
  bool write_me;
  enum e_snc_result result;

  /* Setup rc.conf variables */

  /* SYSCFG_IFACE - strip if interface config has changed, write if interface config has changed */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg;
  var_value = "y";
  result = snc_rcconf_set_config_iface_var(&config_vars[NETCFG_VAR_IFACE], "%s%d", SNC_RCCONF_VARNAME_SYSCFG_IFACE, iface_idx,
                                           var_value, write_me, strip_me);
  if (sncr_success != result)
    goto scscv_fail;

  /* IP address - strip if interface config has changed, write if interface config has changed */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg;
  if (net_config->use_dhcp)
  {
    var_value = SNC_RCCONF_VARVALUE_IPADDR_DHCP;
    result = snc_rcconf_set_config_iface_var(&config_vars[NETCFG_VAR_IPADDR], "%s%d", SNC_RCCONF_VARNAME_IPADDR, iface_idx,
                                             var_value, write_me, strip_me);
  }
  else
  {
    result = snc_rcconf_set_config_iface_var_inet(&config_vars[NETCFG_VAR_IPADDR], "%s%d", SNC_RCCONF_VARNAME_IPADDR, iface_idx,
                                                  &net_config->ip_address, write_me, strip_me, false);
  }
  if (sncr_success != result)
    goto scscv_fail;

  /* Netmask - strip if interface config has changed, write if interface config has changed and static IP */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg && !net_config->use_dhcp;
  result = snc_rcconf_set_config_iface_var_inet(&config_vars[NETCFG_VAR_NETMASK], "%s%d", SNC_RCCONF_VARNAME_NETMASK, iface_idx,
                                                &net_config->ip_netmask, write_me, strip_me, false);
  if (sncr_success != result)
    goto scscv_fail;

  /* Broadcast - strip if interface config has changed, write if interface config has changed and static IP */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg && !net_config->use_dhcp;
  result = snc_rcconf_set_config_iface_var_inet(&config_vars[NETCFG_VAR_BROADCAST], "%s%d", SNC_RCCONF_VARNAME_BROADCAST, iface_idx,
                                                &net_config->ip_broadcast, write_me, strip_me, false);
  if (sncr_success != result)
    goto scscv_fail;

  /* Gateway - strip if interface config has changed, write if interface config has changed and static IP */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg && !net_config->use_dhcp;
  result = snc_rcconf_set_config_iface_var_inet(&config_vars[NETCFG_VAR_GATEWAY], "%s%d", SNC_RCCONF_VARNAME_GATEWAY, iface_idx,
                                                &net_config->ip_gateway, write_me, strip_me, true);
  if (sncr_success != result)
    goto scscv_fail;

  /* Nameserver 1 - strip if interface config has changed, write if interface config has changed and static IP */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg && !net_config->use_dhcp;
  result = snc_rcconf_set_config_iface_var_inet(&config_vars[NETCFG_VAR_NAMESERVER], "%s%d", SNC_RCCONF_VARNAME_NAMESERVER, iface_idx,
                                                &net_config->ip_dns[0], write_me, strip_me, true);
  if (sncr_success != result)
    goto scscv_fail;

  /* Nameserver 2 - strip if interface config has changed, write if interface config has changed and static IP */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg && !net_config->use_dhcp;
  result = snc_rcconf_set_config_iface_var_inet(&config_vars[NETCFG_VAR_NAMESERVER2], "%s%d_2", SNC_RCCONF_VARNAME_NAMESERVER, iface_idx,
                                                &net_config->ip_dns[1], write_me, strip_me, true);
  if (sncr_success != result)
    goto scscv_fail;

  /* Nameserver 3 - strip if interface config has changed, write if interface config has changed and static IP */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg && !net_config->use_dhcp;
  result = snc_rcconf_set_config_iface_var_inet(&config_vars[NETCFG_VAR_NAMESERVER3], "%s%d_3", SNC_RCCONF_VARNAME_NAMESERVER, iface_idx,
                                                &net_config->ip_dns[2], write_me, strip_me, true);
  if (sncr_success != result)
    goto scscv_fail;

  /* Search domain - strip if interface config has changed, write if interface config has changed and static IP */
  strip_me = changed_ifacecfg;
  write_me = changed_ifacecfg && !net_config->use_dhcp;
  var_value = net_config->search_domain;
  result = snc_rcconf_set_config_iface_var(&config_vars[NETCFG_VAR_SEARCHDOMAIN], "%s%d", SNC_RCCONF_VARNAME_SEARCHDOMAIN, iface_idx,
                                           var_value, write_me, strip_me);
  if (sncr_success != result)
    goto scscv_fail;

  /* Hostname - strip if hostname changed, write if hostname has changed */
  strip_me = changed_hostname;
  write_me = changed_hostname;
  var_value = net_config->hostname;
  result = snc_rcconf_variable_init(&config_vars[NETCFG_VAR_HOSTNAME], SNC_RCCONF_VARNAME_HOSTNAME, var_value, write_me, strip_me);
  if (sncr_success != result)
    goto scscv_fail;

  /* Domain name - strip if domain name changed, write if domain name has changed */
  strip_me = changed_domain;
  write_me = changed_domain;
  var_value = net_config->domain;
  result = snc_rcconf_variable_init(&config_vars[NETCFG_VAR_DOMAINNAME], SNC_RCCONF_VARNAME_DOMAINNAME, var_value, write_me, strip_me);
  if (sncr_success != result)
    goto scscv_fail;

  /* Success */
  result = sncr_success;

scscv_fail:
  return result;
}


/** Restart network and/or hostname if configuration has changed */
static enum e_snc_result snc_rc_conf_hostname_network_reconfigure(const struct s_snc_configuration *config,
                                                                  bool changed_hostname, bool changed_ifacecfg, bool changed_domain)
{
  int ret_code;
  enum e_snc_result result = sncr_success;

  /* Hostname */
  if (changed_hostname &&
      NULL != config->hostname_reconfigure_cmd &&
      strcmp(config->hostname_reconfigure_cmd, "") != 0)
  {
    ret_code = system(config->hostname_reconfigure_cmd);
    result = (WIFEXITED(ret_code) && WEXITSTATUS(ret_code) == 0) ? sncr_success : sncr_execerror;
  }

  /* Network configuration */
  if ((changed_ifacecfg || changed_domain) &&
      NULL != config->net_reconfigure_cmd &&
      strcmp(config->net_reconfigure_cmd, "") != 0)
  {
    ret_code = system(config->net_reconfigure_cmd);
    result = (WIFEXITED(ret_code) && WEXITSTATUS(ret_code) == 0) ? sncr_success : sncr_execerror;
  }
  return result;
}


/**
 * Set the network configuration in rc.conf
 *
 * @param config is the library configuration used for the path to rc.conf and the hostname and network reconfigure scripts
 * @param eth_interface is the interface to set the configuration for. Can be NULL if setting hostname or domain only
 * @param net_config is the network configuration values to set
 * @param set_ifacecfg is a pointer to a bool that should be set to true if you want the interface configuration to be set.
 *          The function will set/clear this depending on if the interface configuration has actually changed.
 * @param set_hostname is a pointer to a bool that should be set to true if you want the hostname to be set.
 *          The function will set/clear this depending on if the hostname has actually changed.
 * @param set_domain is a pointer to a bool that should be set to true if you want the domain to be set.
 *          The function will set/clear this depending on if the domain has actually changed.
 */
enum e_snc_result snc_set_net_config_rcconf(const struct s_snc_configuration *config,
                                            const char *eth_interface,
                                            const struct s_snc_net_config *net_config,
                                            bool *set_ifacecfg, bool *set_hostname, bool *set_domain)
{
  FILE *rcconf = NULL;
  enum e_snc_result result;
  struct s_snc_net_config current_net_config;
  bool hostname_differs = false;
  bool domain_differs = false;
  bool ifacecfg_differs = false;
  bool should_set_hostname = false;
  bool should_set_domain = false;
  bool should_set_ifacecfg = false;
  int iface_idx;
  size_t i;
  struct s_snc_rcconf_variable *config_vars[NUM_NETCFG_VARS];
  char *line = NULL;
  char *temp_filename = NULL;
  int temp_fd = -1;
  FILE *temp_file = NULL;

  if (NULL == config || NULL == net_config || NULL == set_ifacecfg || NULL == set_hostname || NULL == set_domain)
    return sncr_paramerror;
  if (*set_ifacecfg && NULL == eth_interface)
    return sncr_paramerror;

  for (i = 0; i < NUM_NETCFG_VARS; i++)
    config_vars[i] = NULL;

  line = malloc(SNC_PARSE_LINE_LENGTH);
  if (NULL == line)
  {
    result = sncr_allocerror;
    goto ssncr_fail;
  }

  /* Open rc.conf config file for reading only (we write the entire config to a temporary file) */
  rcconf = fopen(config->rcconf_file, "r");
  if (NULL == rcconf)
  {
    result = sncr_fileerror;
    goto ssncr_fail;
  }

  /* Get the current configuration settings */
  result = snc_get_net_config_rcconf_internal(rcconf, line, eth_interface, &current_net_config, &iface_idx);
  if (sncr_success != result)
    goto ssncr_fail;

  /* Work out what's changed, and what we should set */
  result = snc_net_config_diff(net_config, &current_net_config, &ifacecfg_differs, &hostname_differs, &domain_differs, NULL);
  should_set_hostname = hostname_differs && *set_hostname;
  should_set_domain = domain_differs && *set_domain;
  should_set_ifacecfg = ifacecfg_differs && *set_ifacecfg;

  /* If anything has changed and should be set, then strip the variables from the configuration file */
  if (should_set_hostname || should_set_domain || should_set_ifacecfg)
  {
    /* Setup the rc.conf config variables ready to write to the new rc.conf */
    result = snc_rcconf_setup_config_vars(config_vars, net_config, iface_idx, should_set_hostname, should_set_ifacecfg, should_set_domain);
    if (sncr_success != result)
      goto ssncr_fail;

    /* Create temp file */
    result = snc_rcconf_generate_tmpfile(config->rcconf_file, &temp_filename, &temp_fd, &temp_file);
    if (sncr_success != result)
      goto ssncr_fail;

    /* Strip network configuration from config file (into tmp_file) */
    result = snc_rcconf_strip_existing_config(rcconf, temp_file, line, config_vars, NUM_NETCFG_VARS);
    if (sncr_success != result)
      goto ssncr_fail;

    /* Write new interface settings to temp_file */
    for (i = 0; i < NUM_NETCFG_VARS; i++)
    {
      if (config_vars[i]->write_me)
      {
        if (fprintf(temp_file, "%s %s=\"%s\"\n",
                    snc_export_str,
                    config_vars[i]->variable_name,
                    config_vars[i]->variable_value) < 0)
        {
          result = sncr_fileerror;
          goto ssncr_fail;
        }
      }
    }

    /* Now copy the contents of the temporary file to the configuration file */

    /* Make sure all changes are saved to disk */
    if (0 != fflush(temp_file) || 0 != fsync(fileno(temp_file)))
    {
      result = sncr_fileerror;
      goto ssncr_fail;
    }
    SNC_FCLOSE_AND_NULLIFY(temp_file);
    SNC_CLOSE_AND_INVALIDATE(temp_fd);
    SNC_FCLOSE_AND_NULLIFY(rcconf);

    /* Now change the permissions on the file (files from mkstemp() are 600, we want 644) */
    if (chmod(temp_filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) < 0)
    {
      result = sncr_fileerror;
      goto ssncr_fail;
    }

    /* Move from old file to new file */
    if (rename(temp_filename, config->rcconf_file) < 0)
    {
      result = sncr_fileerror;
      goto ssncr_fail;
    }
  } /* Something has changed */

  /* Let caller know what was actually different and hence has been set */
  *set_hostname = should_set_hostname;
  *set_domain = should_set_domain;
  *set_ifacecfg = should_set_ifacecfg;

  /* Success */
  result = sncr_success;

ssncr_fail:
  SNC_FCLOSE_AND_NULLIFY(temp_file);
  SNC_CLOSE_AND_INVALIDATE(temp_fd);
  if (NULL != temp_filename)
  {
    unlink(temp_filename);
    SNC_FREE_AND_NULLIFY(temp_filename);
  }
  SNC_FREE_AND_NULLIFY(line);
  for (i = 0; i < NUM_NETCFG_VARS; i++)
  {
    SNC_DESTROY_AND_NULLIFY(snc_rcconf_variable_destroy, config_vars[i]);
  }

  /* Run the reconfigure scripts */
  if (sncr_success == result)
    result = snc_rc_conf_hostname_network_reconfigure(config, should_set_hostname, should_set_ifacecfg, should_set_domain);

  return result;
}
