/*
 * lsnc_struct.c
 *
 *  Created on: 14 Jun. 2018
 *      Author: ian
 */
#include <stddef.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <setnetconfig.h>


/** Initialise the members of existing @ref struct s_snc_net_config to sensible values */
enum e_snc_result snc_net_config_struct_init(struct s_snc_net_config *net_config)
{
  size_t i;

  if (NULL == net_config)
    return sncr_paramerror;

  memset(net_config, 0, sizeof(*net_config));
  net_config->use_dhcp = false;
  net_config->ip_address.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  net_config->ip_netmask.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  net_config->ip_broadcast.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  net_config->ip_gateway.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;

  for (i = 0; i < SNC_NUM_DNS_SERVERS; i++)
    net_config->ip_dns[i].s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;

  snprintf(net_config->search_domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
  snc_zero_mac_address(&net_config->mac_address);

  snprintf(net_config->hostname, SNC_HOST_NAME_MAX, "%s", SNC_HOSTNAME_DEFAULT);
  snprintf(net_config->domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);

  return sncr_success;
}


/**
 * Allocate a net config structure and initialise the members of @ref struct s_snc_net_config to sensible values
 *
 * Caller is responsible for freeing
 */
enum e_snc_result snc_net_config_struct_allocate(struct s_snc_net_config **net_config)
{
  if (NULL == net_config)
    return sncr_paramerror;

  *net_config = calloc(1, sizeof(struct s_snc_net_config));
  if (NULL == *net_config)
    return sncr_allocerror;

  return snc_net_config_struct_init(*net_config);
}


/** Function to initialise the members of out_mac to the MAC address contained in inmac */
enum e_snc_result snc_set_mac_address(const unsigned in_mac[SNC_ETH_ALEN], struct sockaddr *out_mac)
{
  size_t i;

  if (NULL == in_mac || NULL == out_mac)
    return sncr_paramerror;

  out_mac->sa_family = AF_PACKET;
  for (i = 0; i < SNC_ETH_ALEN; i++)
    out_mac->sa_data[i] = (char) in_mac[i];
  for ( ; i < sizeof(out_mac->sa_data); i++)
    out_mac->sa_data[i] = '\0';
  return sncr_success;
}


/** Function to initialise the members of mac to 0 */
enum e_snc_result snc_zero_mac_address(struct sockaddr *mac)
{
  const unsigned zero_mac[SNC_ETH_ALEN] = { 0, 0, 0, 0, 0, 0 };

  return snc_set_mac_address(zero_mac, mac);
}


/** Make the broadcast address from an IP address and netmask */
enum e_snc_result snc_ipv4_broadcast_from_ip_netmask(const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask, struct in_addr *ipv4_broadcast)
{
  if (NULL == ipv4_address || NULL == ipv4_netmask || NULL == ipv4_broadcast)
    return sncr_paramerror;

  ipv4_broadcast->s_addr = (ipv4_address->s_addr & ipv4_netmask->s_addr) | ~ipv4_netmask->s_addr;
  return sncr_success;
}


/** Compare two network configurations */
enum e_snc_result snc_net_config_diff(const struct s_snc_net_config *nc1, const struct s_snc_net_config *nc2,
                                      bool *ifcfg_differs, bool *hostname_differs, bool *domain_differs, bool *mac_differs)
{
  bool changed;
  size_t i;

  if (NULL == nc1 || NULL == nc2)
    return sncr_paramerror;

  if (NULL != ifcfg_differs)
  {
    if (nc1->use_dhcp != nc2->use_dhcp)
    {
      /* DHCP setting is not the same */
      *ifcfg_differs = true;
    }
    else
    {
      if (nc1->use_dhcp)
      {
        /* Both are using DHCP */
        *ifcfg_differs = false;
      }
      else
      {
        /* Both are using static settings, compare everything */
        changed = (nc1->ip_address.s_addr != nc2->ip_address.s_addr ||
                   nc1->ip_netmask.s_addr != nc2->ip_netmask.s_addr ||
                   nc1->ip_broadcast.s_addr != nc2->ip_broadcast.s_addr ||
                   nc1->ip_gateway.s_addr != nc2->ip_gateway.s_addr);
        for (i = 0; i < SNC_NUM_DNS_SERVERS; i++)
        {
          if (nc1->ip_dns[i].s_addr != nc2->ip_dns[i].s_addr)
            changed = true;
        }
        if (strcmp(nc1->search_domain, nc2->search_domain) != 0)
          changed = true;
        *ifcfg_differs = changed;
      }
    }
  }

  if (NULL != hostname_differs)
  {
    *hostname_differs = (strcmp(nc1->hostname, nc2->hostname) != 0);
  }
  if (NULL != domain_differs)
  {
    *domain_differs = (strcmp(nc1->domain, nc2->domain) != 0);
  }
  if (NULL != mac_differs)
  {
    *mac_differs = (memcmp(nc1->mac_address.sa_data, nc2->mac_address.sa_data, SNC_ETH_ALEN) != 0);
  }
  return sncr_success;
}
