/*
 * Program to set the network configuration for an interface
 *
 * Invocation is "setnetconfig <iface> <ip> <netmask> <broadcast> <gateway> <nameserver>"
 */
#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <getopt.h>
#include <sys/stat.h>
#include "lsnc_parse_addr.h"
#include "lsnc_internal.h"
#include "snc_validate.h"
#include <setnetconfig.h>


#define PROGRAM_NAME "setnetconfig"

static bool cmdline_help = false;
static bool cmdline_version = false;
static bool cmdline_quiet = false;
static const char *cmdline_interface = NULL;
static char *cmdline_ipaddr = NULL;
static char *cmdline_netmask = NULL;
static char *cmdline_broadcast = NULL;
static char *cmdline_gateway = NULL;
static char *cmdline_nameserver = NULL;
static char *cmdline_hostname = NULL;
static char *cmdline_domain = NULL;
static const char *cmdline_method = SNC_METHOD_LIB_CONFIG;
static const char *cmdline_configfile = SNC_CONFIG_FILE;


static void print_usage(void)
{
  printf("Usage:\n"
         "  " PROGRAM_NAME " [options] <iface> <ip> <netmask> <bcast> <gateway> <nameserver>\n"
         "  " PROGRAM_NAME " [options] <iface> dhcp\n"
         "  " PROGRAM_NAME " [options] --hostname <hostname>\n"
         "  " PROGRAM_NAME " [options] --domain <domain>\n"
         "\nOptions:\n"
         "  -v --version          Print the software version\n"
         "  -h --help             Print this help\n"
         "  -q --quiet            Don't print info messages\n"
         "  -c --config <file>    Load the library configuration from this file\n"
         "                        Default " SNC_CONFIG_FILE "\n"
         "  -o --hostname <name>  Set the hostname to <name>\n"
         "  -i --domain <domain>  Set the domain to <domain>\n"
         "  -m --method <method>  Use <method> when setting configuration\n"
         "                        Methods are:\n"
         "                          libconf - Use the method set in config file (default)\n"
         "                          auto - Try each of the available modes until success\n"
         "                          rcconf - Set network config in rc.conf\n"
#if defined(HAVE_LIBNM)
         "                          nm - Set network config via Network Manager\n"
#endif
         "  iface                 The network interface to configure. Eg eth0\n"
         "  ip                    The new ip address or \"dhcp\"\n"
         "  netmask               The new network mask\n"
         "  bcast                 The new network broadcast address\n"
         "  gateway               The new network gateway\n"
         "  nameserver            The name server to use for DNS resolution\n"
         "\n"
         "Program to set the network configuration\n"
         "Network addresses should be in dot-decimal notation (n.n.n.n, 192.168.0.1)\n"
         );
}


static void print_version(void)
{
  printf(PROGRAM_NAME " " PACKAGE_VERSION "\n");
}


static enum e_snc_result parse_opts(int argc, char *argv[])
{
  static const struct option lopts[] =
  {
    { "version",                 0, 0, 'v' },
    { "help",                    0, 0, 'h' },
    { "quiet",                   0, 0, 'q' },
    { "config",                  1, 0, 'c' },
    { "hostname",                1, 0, 'o' },
    { "domain",                  1, 0, 'i' },
    { "method",                  1, 0, 'm' },
    { NULL,                      0, 0, '\0' },
  };
  int c;
  bool isdhcp;
  bool hostname_specified;
  bool domain_specified;

  while (1)
  {
    c = getopt_long(argc, argv, "vhqc:o:i:m:", lopts, NULL);

    if (c == -1)
      break;

    switch (c)
    {
      case 'h':
        cmdline_help = true;
        break;

      case 'v':
        cmdline_version = true;
        break;

      case 'q':
        cmdline_quiet = true;
        break;

      case 'c':
        cmdline_configfile = optarg;
        break;

      case 'o':
        cmdline_hostname = optarg;
        break;

      case 'i':
        cmdline_domain = optarg;
        break;

      case 'm':
        cmdline_method = optarg;
        break;

      default:
        fprintf(stderr, PROGRAM_NAME ": Bad command line option. Try --help.\n");
        return sncr_invalid_commandline;
        break;
    }
  }

  hostname_specified = (NULL != cmdline_hostname);
  domain_specified = (NULL != cmdline_domain);
  if (!cmdline_help && !cmdline_version)
  {
    if (optind >= argc && (hostname_specified || domain_specified))
    {
      return sncr_success;
    }

    if (optind >= argc)
    {
      fprintf(stderr, PROGRAM_NAME ": No network interface specified. Try --help.\n");
      return sncr_invalid_commandline;
    }
    cmdline_interface = argv[optind];
    optind++;

    if (optind >= argc)
    {
      fprintf(stderr, PROGRAM_NAME ": No ip address specified. Try --help.\n");
      return sncr_invalid_commandline;
    }
    cmdline_ipaddr = argv[optind];
    optind++;

    isdhcp = (0 == strcmp(SNC_RCCONF_VARVALUE_IPADDR_DHCP, cmdline_ipaddr));
    if (!isdhcp)
    {
      if (optind >= argc)
      {
        fprintf(stderr, PROGRAM_NAME ": No network mask specified. Try --help.\n");
        return sncr_invalid_commandline;
      }
      cmdline_netmask = argv[optind];
      optind++;

      if (optind >= argc)
      {
        fprintf(stderr, PROGRAM_NAME ": No network broadcast address specified. Try --help.\n");
        return sncr_invalid_commandline;
      }
      cmdline_broadcast = argv[optind];
      optind++;

      if (optind >= argc)
      {
        fprintf(stderr, PROGRAM_NAME ": No network gateway specified. Try --help.\n");
        return sncr_invalid_commandline;
      }
      cmdline_gateway = argv[optind];
      optind++;

      if (optind >= argc)
      {
        fprintf(stderr, PROGRAM_NAME ": No name server specified. Try --help.\n");
        return sncr_invalid_commandline;
      }
      cmdline_nameserver = argv[optind];
      optind++;
    }
  }

  return sncr_success;
}


static enum e_snc_result verify_ip_parameters(const char *ipaddr, const char *netmask,
                                              const char *broadcast, const char *gateway,
                                              const char *nameserver, const char *hostname,
                                              const char *domain,
                                              struct s_snc_net_config *net_config,
                                              bool *set_ifacecfg, bool *set_hostname, bool *set_domain)
{
  enum e_snc_result result;

  *set_ifacecfg = false;
  *set_hostname = false;
  *set_domain = false;

  /* Initialise everything */
  snc_net_config_struct_init(net_config);

  if (NULL != ipaddr)
  {
    if (strcmp(SNC_RCCONF_VARVALUE_IPADDR_DHCP, ipaddr) == 0)
    {
      net_config->use_dhcp = true;
    }
    else if (snc_parse_ipv4_from_line(NULL, false, ipaddr, &net_config->ip_address) == sncr_success)
    {
      net_config->use_dhcp = false;
    }
    else
    {
      fprintf(stderr, PROGRAM_NAME ": IP address parameter should be \"" SNC_RCCONF_VARVALUE_IPADDR_DHCP "\", or of the form nnn.nnn.nnn.nnn (eg 192.168.0.1)\n");
      result = sncr_parseerror;
      goto vip_fail;
    }
    *set_ifacecfg = true;
  }

  if (NULL != netmask && snc_parse_ipv4_from_line(NULL, false, netmask, &net_config->ip_netmask) != sncr_success)
  {
    fprintf(stderr, PROGRAM_NAME ": Netmask parameter should be of the form nnn.nnn.nnn.nnn (eg 255.255.255.0)\n");
    result = sncr_parseerror;
    goto vip_fail;
  }

  if (NULL != broadcast && snc_parse_ipv4_from_line(NULL, false, broadcast, &net_config->ip_broadcast) != sncr_success)
  {
    fprintf(stderr, PROGRAM_NAME ": Broadcast parameter should be of the form nnn.nnn.nnn.nnn (eg 192.168.0.255)\n");
    result = sncr_parseerror;
    goto vip_fail;
  }

  if (NULL != gateway && snc_parse_ipv4_from_line(NULL, false, gateway, &net_config->ip_gateway) != sncr_success)
  {
    fprintf(stderr, PROGRAM_NAME ": Gateway parameter should be of the form nnn.nnn.nnn.nnn (eg 192.168.0.1)\n");
    result = sncr_parseerror;
    goto vip_fail;
  }

  if (NULL != nameserver && snc_parse_ipv4_from_line(NULL, false, nameserver, &net_config->ip_dns[0]) != sncr_success)
  {
    fprintf(stderr, PROGRAM_NAME ": Nameserver parameter should be of the form nnn.nnn.nnn.nnn (eg 192.168.0.1)\n");
    result = sncr_parseerror;
    goto vip_fail;
  }

  /* Verify hostname if specified */
  if (NULL != hostname)
  {
    result = snc_hostname_validate(hostname);
    if (sncr_success != result)
    {
      fprintf(stderr, PROGRAM_NAME ": Hostname should be 1..%d characters in length from the set of [a-zA-Z0-9\\-]\n", SNC_HOST_NAME_MAX);
      goto vip_fail;
    }
    snprintf(net_config->hostname, SNC_HOST_NAME_MAX, "%s", hostname);
    *set_hostname = true;
  }

  /* Verify domain name if specified */
  if (NULL != domain)
  {
    result = snc_domainname_validate(domain);
    if (sncr_success != result)
    {
      fprintf(stderr, PROGRAM_NAME ": Domain name should be 1..%d characters in length from the set of [a-zA-Z0-9\\-.]\n", SNC_DOMAIN_NAME_MAX);
      goto vip_fail;
    }
    snprintf(net_config->domain, SNC_DOMAIN_NAME_MAX, "%s", domain);
    *set_domain = true;
  }

  result = sncr_success;

vip_fail:
  return result;
}


/** setnetconfig entry point */
#if ENABLE_MULTI_CALL
int setnetconfig_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
  struct s_snc_configuration *config = NULL;
  struct s_snc_net_config new_net_config;
  enum e_snc_result result;
  bool set_ifacecfg = false;
  bool set_hostname = false;
  bool set_domain = false;

  /* No exec for everyone, no write for group and other */
  umask(S_IXUSR | S_IXGRP | S_IXOTH | S_IWGRP | S_IWOTH);

  /* Check command line parameters */
  result = parse_opts(argc, argv);
  if (sncr_success != result)
    goto m_done;

  /* Help and version */
  if (cmdline_help)
  {
    print_usage();
    result = sncr_success;
    goto m_done;
  }
  if (cmdline_version)
  {
    print_version();
    result = sncr_success;
    goto m_done;
  }

  /* Try to load the library configuration */
  result = snc_config_init(&config);
  if (sncr_success != result)
    goto m_done;
  (void) snc_config_load(cmdline_configfile, config);

  /* Verify IP parameters */
  result = verify_ip_parameters(cmdline_ipaddr, cmdline_netmask,
                                cmdline_broadcast, cmdline_gateway,
                                cmdline_nameserver, cmdline_hostname,
                                cmdline_domain,
                                &new_net_config,
                                &set_ifacecfg, &set_hostname, &set_domain);
  if (sncr_success != result)
    goto m_done;

  /* Validate the configuration */
  result = validate_and_print_errors(PROGRAM_NAME, &new_net_config, set_ifacecfg, set_hostname, set_domain);
  if (sncr_success != result)
    goto m_done;

  /* Set the new network configuration */
  result = snc_set_net_config(config, cmdline_method, cmdline_interface, &new_net_config,
                              &set_ifacecfg, &set_hostname, &set_domain);
  if (sncr_success != result)
    goto m_done;

  /* Print out what happened */
  if (!cmdline_quiet)
  {
    if (set_hostname)
      fprintf(stdout, PROGRAM_NAME ": Changed the hostname setting\n");
    if (set_domain)
      fprintf(stdout, PROGRAM_NAME ": Changed the domain setting\n");
    if (set_ifacecfg)
    {
      if (new_net_config.use_dhcp)
        fprintf(stdout, PROGRAM_NAME ": Changed to DHCP IP address settings\n");
      else
        fprintf(stdout, PROGRAM_NAME ": Changed static IP address settings\n");
    }

    if (!set_hostname && !set_domain && !set_ifacecfg)
      fprintf(stdout, PROGRAM_NAME ": No change to network settings\n");
  }

  /* Success */
  result = sncr_success;

m_done:
  SNC_DESTROY_AND_NULLIFY(snc_config_destroy, config);
  if (sncr_success != result)
    fprintf(stderr, PROGRAM_NAME ": failed with result %s\n", snc_result_to_string(result));
  return result;
}
