/*
 * setnetconfig.h
 *
 *  Created on: 31 May 2018
 *      Author: ian
 */

#ifndef SETNETCONFIG_H_
#define SETNETCONFIG_H_

#include <stdbool.h>
#include <stdio.h>
#include <stddef.h>
#include <bits/local_lim.h>
#include <netinet/in.h>
#include <net/if.h>
#include <sys/socket.h>
#include <netinet/if_ether.h>


/** Result code returned by libsetnetconfig functions */
enum e_snc_result
{
  sncr_success = 0,             /*!< The function has completed successfully */
  sncr_paramerror,              /*!< There was a problem with one of the parameters passed to the function */
  sncr_fileerror,               /*!< There was a problem with file access (open, read, write, etc) */
  sncr_parseerror,              /*!< There was a problem parsing data */
  sncr_allocerror,              /*!< There was a problem allocating memory */
  sncr_resolverror,             /*!< There was a problem with libresolv */
  sncr_execerror,               /*!< There was a problem with executing a command */
  sncr_syscallerror,            /*!< There was a problem with a system call */
  sncr_notfound,                /*!< An expected item was not found (ie when parsing a file) */
  sncr_invalid_commandline,     /*!< A command line argument was invalid */
  sncr_string_too_long,         /*!< A string was too long */

  sncr_unknownerror,            /*!< Unknown error */
  sncr_num                      /*!< Number of results */
};


/** Result strings, must match @ref enum e_snc_result */
#define SNC_RESULT_STRINGS \
{ \
  "success",                  /*!< sncr_success */ \
  "parameter error",          /*!< sncr_paramerror */ \
  "file i/o error",           /*!< sncr_fileerror */ \
  "parsing error",            /*!< sncr_parseerror */ \
  "memory allocation error",  /*!< sncr_allocerror */ \
  "libresolv error",          /*!< sncr_resolverror */ \
  "execute error",            /*!< sncr_execerror */ \
  "system call error",        /*!< sncr_syscallerror */ \
  "not found",                /*!< sncr_notfound */ \
  "invalid command line",     /*!< sncr_invalid_commandline */ \
  "string too long",          /*!< sncr_string_too_long */ \
\
  "unknown error",            /*!< sncr_unknownerror */ \
}


/** The default library configuration file */
#define SNC_CONFIG_FILE "/etc/libsetnetconfig.conf"


/** Support 3 DNS servers, because that is what libresolv supports */
#define SNC_NUM_DNS_SERVERS 3


/** Maximum length of an Ethernet interface name (eg eth0) */
#define SNC_IF_NAMESIZE IF_NAMESIZE


/** Maximum length of an IPv4 address represented as a string "nnn.nnn.nnn.nnn" */
#define SNC_INET_ADDRSTRLEN INET_ADDRSTRLEN


/** Length of a MAC address */
#define SNC_ETH_ALEN ETH_ALEN


/** Maximum length of a hostname as per RFC2181 Section 11 Name syntax */
#define SNC_HOST_NAME_MAX HOST_NAME_MAX


/** Maximum length of a domain name as per RFC2181 Section 11 Name syntax */
#define SNC_DOMAIN_NAME_MAX 255


/** The IPv4 address to use if a value has not been set */
#define SNC_IPV4_UNSET_STR "0.0.0.0"
#define SNC_IPV4_UNSET_INADDR_NETWORK SNC_IPV4_OCTETS_TO_INADDR_NETWORK(0, 0, 0, 0)


/** The default hostname */
#define SNC_HOSTNAME_DEFAULT "unknown"


/** The default domain/search domain */
#define SNC_DOMAINNAME_DEFAULT ""


/** Libsetnetconfig should set and retrieve settings from rc.conf */
#define SNC_METHOD_RCCONF "rcconf"

/** Libsetnetconfig should set and retrieve settings from Network Manager */
#define SNC_METHOD_NETWORK_MANAGER "nm"

/** Libsetnetconfig should retrieve current active configuration using ioctl() function calls */
#define SNC_METHOD_IOCTL "ioctl"

/** Libsetnetconfig should retrieve current active configuration using getifaddrs() function calls */
#define SNC_METHOD_GETIFADDRS "getifaddrs"

/** Libsetnetconfig should retrieve current active configuration by parsing the output of ifconfig */
#define SNC_METHOD_IFCONFIG "ifconfig"

/** Libsetnetconfig should retrieve current active configuration by parsing the output of the ip command */
#define SNC_METHOD_IPCMD "ip"

/** Libsetnetconfig should retrieve settings using the first successful method */
#define SNC_METHOD_AUTO "auto"

/** Libsetnetconfig should set and retrieve network settings using the method set in the library config file /etc/libsetnetconfig.conf */
#define SNC_METHOD_LIB_CONFIG "libconf"


/* Helper functions */


/** Convert 4 IP octets into an in_addr_t in host endian */
#define SNC_IPV4_OCTETS_TO_INADDR_HOST(a, b, c, d) ((a & 0xFF) << 24 | (b & 0xFF) << 16 | (c & 0xFF) << 8 | (d & 0xFF))


/** Convert 4 IP octets into an in_addr_t in network endian */
#define SNC_IPV4_OCTETS_TO_INADDR_NETWORK(a, b, c, d) (htonl(SNC_IPV4_OCTETS_TO_INADDR_HOST(a, b, c, d)))


/** The network configuration structure. All IP addresses are in network endian */
struct s_snc_net_config
{
  /* These settings are per interface */
  bool use_dhcp;                                  /*!< Is this interface configured by DHCP */
  struct in_addr ip_address;                      /*!< The assigned IP address in network endian */
  struct in_addr ip_netmask;                      /*!< The assigned netmask in network endian */
  struct in_addr ip_broadcast;                    /*!< The assigned broadcast address in network endian */
  struct in_addr ip_gateway;                      /*!< The assigned default gateway address in network endian */
  struct in_addr ip_dns[SNC_NUM_DNS_SERVERS];     /*!< Any assigned DNS server addresses in network endian */
  struct sockaddr mac_address;                    /*!< The MAC address of the interface (read only) */
  char search_domain[SNC_DOMAIN_NAME_MAX + 1];    /*!< The search domain */

  /* These are for the host */
  char hostname[SNC_HOST_NAME_MAX + 1];           /*!< The device hostname */
  char domain[SNC_DOMAIN_NAME_MAX + 1];           /*!< The device domain */
};


/*
 * Library configuration
 */


/** Settings for libsetnetconfig */
struct s_snc_configuration
{
  char *resolv_file;              /*!< The DNS resolver file, usually /etc/resolv.conf */
  char *route_cmd;                /*!< The path to the route binary, usually /sbin/route */
  char *ifconfig_cmd;             /*!< The path to the ifconfig binary, usually /sbin/ifconfig */
  char *ip_cmd;                   /*!< The path to the ip binary, usually /sbin/ip */
  char *rcconf_file;              /*!< The location of the rc.conf file, usually /etc/rc.d/rc.conf */
  char *config_method;            /*!< The method by which to retrieve/set network config (see SNC_METHOD_* constants) */
  char *current_method;           /*!< The method by which to retrieve/set currently active network config (see SNC_METHOD_* constants) */
  char *net_reconfigure_cmd;      /*!< The command to run to reconfigure the current network settings after changes to network settings in rc.conf */
  char *hostname_reconfigure_cmd; /*!< The command to run to reconfigure the current hostname after changes to hostname in rc.conf */
};


/** Create a setnetconfig library configuration set to defaults */
enum e_snc_result snc_config_init(struct s_snc_configuration **config);


/** Destroy a setnetconfig library configuration */
enum e_snc_result snc_config_destroy(struct s_snc_configuration *config);


/** Load the library configuration from a file */
enum e_snc_result snc_config_load(const char *config_path, struct s_snc_configuration *config);


/** Save the library configuration from a file */
enum e_snc_result snc_config_save(const char *config_path, const struct s_snc_configuration *config);


/*
 * Validation functions
 */


/** Check hostname contains sensible characters, as per RFC952, RFC1123, and is of sensible length, as per RFC 2181 */
enum e_snc_result snc_hostname_validate(const char *hostname);


/** Check domainname contains sensible characters, as per RFC952, RFC1123, and is of sensible length, as per RFC 2181 */
enum e_snc_result snc_domainname_validate(const char *domainname);


/**
 * Validate that ipv4_netmask (in network byte order) is a valid netmask
 * Conditions:
 *  - 1s are contiguous from right to left (RFC4632 Section 3.1)
 *  - Excludes 255.255.255.255 (broadcast)
 *  - Excludes 0.0.0.0 (network)
 */
enum e_snc_result snc_ipv4_netmask_validate(const struct in_addr *ipv4_netmask);


/**
 * Validate that ipv4_address (in network byte order) is suitable as a destination address of a single host (not broadcast/multicast)
 * (permits off subnet)
 * Excludes:
 *  - Class D (multicast range)
 *  - Class E (reserved)
 *  - Subnet broadcast (ie 192.168.1.255) if ipv4_source and ipv4_source_netmask are supplied
 *  - Subnet 0 (ie 192.168.1.0), except for /31 (see RFC3021) - this was a historic broadcast address if ipv4_source and ipv4_source_netmask are supplied
 *
 * @param ipv4_destination is the destination address to validate in network byte order. Not NULL
 * @param ipv4_source is an optional source address, used to verify that the destination is not a subnet broadcast address in network byte order. Can be NULL
 * @param ipv4_source_netmask is an optional netmask of ipv4_source, used to verify that the destination is not a subnet broadcast address in network byte order. Can be NULL
 *
 * @return sncr_success if ipv4_destination is a valid address
 */
enum e_snc_result snc_ipv4_single_destination_address_validate(const struct in_addr *ipv4_destination, const struct in_addr *ipv4_source, const struct in_addr *ipv4_source_netmask);


/**
 * Validate that ipv4_address (in network byte order) is suitable for setting as a host address
 * Excludes:
 *  - Loopback
 *  - Class D (multicast range)
 *  - Class E (reserved)
 *  - Subnet broadcast (ie 192.168.1.255)
 *  - Subnet 0 (ie 192.168.1.0), except for /31 (see RFC3021) - this was a historic broadcast address
 *
 * @param ipv4_address is a host ip address to verify in network byte order. Not NULL
 * @param ipv4_netmask is the netmask for the ip address in network byte order. Not NULL
 *
 * @return sncr_success if ipv4_address is a valid host address
 */
enum e_snc_result snc_ipv4_host_address_validate(const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask);


/**
 * Validate that ipv4_address (in network byte order) is suitable for setting as a loopback address
 * Excludes:
 *  - Loopback subnet broadcast (eg ip 127.255.255.255 nm 255.255.255.0)
 *  - Loopback subnet 0 (eg ip 127.255.255.0 nm 255.255.255.0) - this was a historic broadcast address
 *  - Netmask outside of 255.0.0.0 .. 255.255.255.255
 *
 * @param ipv4_address is a loopback ip address to verify in network byte order. Not NULL
 * @param ipv4_netmask is the netmask for the loopback ip address in network byte order. Not NULL
 *
 * @return sncr_success if ipv4_address is a valid loopback address
 */
enum e_snc_result snc_ipv4_loopback_address_validate(const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask);


/**
 * Validate that ipv4_gateway (in network byte order) is valid gateway address
 *
 * Must be a valid host address (@ref snc_ipv4_host_address_validate()) and must be on the same subnet as ipv4_address
 *
 * @param ipv4_gateway is the gateway address to verify in network byte order. Not NULL
 * @param ipv4_address is the ip address used to verify the gateway in network byte order. Not NULL
 * @param ipv4_netmask is the netmask used to verify the gateway in network byte order. Not NULL
 *
 * @return sncr_success if ipv4_gateway is a valid gateway address
 */
enum e_snc_result snc_ipv4_gateway_address_validate(const struct in_addr *ipv4_gateway, const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask);


/**
 * Validate that ipv4_broadcast is valid broadcast address (in network byte order) for ipv4_address and ipv4_netmask
 *
 * @param ipv4_broadcast is the broadcast address to verify in network byte order. Not NULL
 * @param ipv4_address is the ip address used to verify the broadcast in network byte order. Not NULL
 * @param ipv4_netmask is the netmask used to verify the broadcast in network byte order. Not NULL
 *
 * @return sncr_success if ipv4_broadcast is a valid broadcast address
 */
enum e_snc_result snc_ipv4_broadcast_validate(const struct in_addr *ipv4_broadcast, const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask);


#define SNC_VALIDATE_VALID 0x0
#define SNC_VALIDATE_IPV4_ADDRESS_INVALID 0x1
#define SNC_VALIDATE_IPV4_NETMASK_INVALID 0x2
#define SNC_VALIDATE_IPV4_BROADCAST_INVALID 0x4
#define SNC_VALIDATE_IPV4_GATEWAY_INVALID 0x8
#define SNC_VALIDATE_IPV4_DNS0_INVALID 0x10
#define SNC_VALIDATE_IPV4_DNS1_INVALID 0x20
#define SNC_VALIDATE_IPV4_DNS2_INVALID 0x40
#define SNC_VALIDATE_SEARCHDOMAIN_INVALID 0x80
#define SNC_VALIDATE_HOSTNAME_INVALID 0x100
#define SNC_VALIDATE_DOMAIN_INVALID 0x200


/**
 * Validate all fields within a struct s_snc_net_config
 *
 * @param net_config is the configuration to verify
 * @param valid_mask is set to a bit mask of SNC_VALIDATE_* constants
 * @param check_dhcp_fields if set will check DHCP supplied fields (ip/nm/bc/gw/dns/sd) even DHCP is enabled. Generally use false.
 *
 * @return sncr_success when all fields are valid
 */
enum e_snc_result snc_net_config_validate(const struct s_snc_net_config *net_config, uint32_t *valid_mask, bool check_dhcp_fields);


/*
 * Host name functions
 */


/** Get the current system hostname via a system call (gethostname) */
enum e_snc_result snc_get_hostname_syscall(char *hostname, size_t hostname_length);


/*
 * Get the current network settings by various methods
 */


/** Get the current network settings from parsing the output from various processes (ifconfig, route) */
enum e_snc_result snc_get_net_current_ifconfig(const struct s_snc_configuration *config, const char *eth_interface, struct s_snc_net_config *net_current);


/** Get the current network settings from parsing the output from the ip command */
enum e_snc_result snc_get_net_current_ip_command(const struct s_snc_configuration *config, const char *eth_interface, struct s_snc_net_config *net_current);


/** Use getifaddrs() for ip, netmask, broadcast if it is available */
enum e_snc_result snc_get_net_current_getifaddrs(const struct s_snc_configuration *config, const char *eth_interface, struct s_snc_net_config *net_current);


/** Use ioctl for ip, netmask, broadcast, mac address if it is available */
enum e_snc_result snc_get_net_current_ioctl(const struct s_snc_configuration *config, const char *eth_interface, struct s_snc_net_config *net_current);


/** Get the current network settings from network manager via libnm */
enum e_snc_result snc_get_net_current_network_manager(const char *eth_interface, struct s_snc_net_config *net_current);


/** Get the currently active network configuration using specified method (see SNC_METHOD_* constants) */
enum e_snc_result snc_get_net_current(const struct s_snc_configuration *config,
                                      const char *method, const char *eth_interface,
                                      struct s_snc_net_config *net_current);


/*
 * Get/set the network configuration settings by various methods
 */


/** Get the network configuration from rc.conf */
enum e_snc_result snc_get_net_config_rcconf(const char *rcconf_path, const char *eth_interface, struct s_snc_net_config *net_config);


/**
 * Set the network configuration in rc.conf
 *
 * @param config is the library configuration used for the path to rc.conf and the hostname and network reconfigure scripts
 * @param eth_interface is the interface to set the configuration for. Can be NULL if setting hostname or domain only
 * @param net_config is the network configuration values to set
 * @param set_ifacecfg is a pointer to a bool that should be set to true if you want the interface configuration to be set.
 *          The function will set/clear this depending on if the interface configuration has actually changed.
 * @param set_hostname is a pointer to a bool that should be set to true if you want the hostname to be set.
 *          The function will set/clear this depending on if the hostname has actually changed.
 * @param set_domain is a pointer to a bool that should be set to true if you want the domain to be set.
 *          The function will set/clear this depending on if the domain has actually changed.
 */
enum e_snc_result snc_set_net_config_rcconf(const struct s_snc_configuration *config,
                                            const char *eth_interface,
                                            const struct s_snc_net_config *net_config,
                                            bool *set_ifacecfg, bool *set_hostname, bool *set_domain);


/** Get the network configuration from network manager via libnm */
enum e_snc_result snc_get_net_config_network_manager(const char *eth_interface, struct s_snc_net_config *net_config);


/**
 * Set the network configuration in network manager via libnm
 *
 * @param eth_interface is the interface to set the configuration for. Can be NULL if setting hostname or domain only
 * @param net_config is the network configuration values to set
 * @param set_ifacecfg is a pointer to a bool that should be set to true if you want the interface configuration to be set.
 *          The function will set/clear this depending on if the interface configuration has actually changed.
 * @param set_hostname is a pointer to a bool that should be set to true if you want the hostname to be set.
 *          The function will set/clear this depending on if the hostname has actually changed.
 * @param set_domain is a pointer to a bool that should be set to true if you want the domain to be set.
 *          The function will set/clear this depending on if the domain has actually changed.
 */
enum e_snc_result snc_set_net_config_network_manager(const char *eth_interface,
                                                     const struct s_snc_net_config *net_config,
                                                     bool *set_ifacecfg, bool *set_hostname, bool *set_domain);


/** Get the network configuration by the specified method */
enum e_snc_result snc_get_net_config(const struct s_snc_configuration *config,
                                     const char *method, const char *eth_interface,
                                     struct s_snc_net_config *net_config);


/**
 * Set the network configuration by the specified method
 *
 * @param config is the library configuration structure
 * @param method is one of the SNC_METHOD_* strings
 * @param eth_interface is the interface to set the configuration for. Can be NULL if setting hostname or domain only
 * @param net_config is the network configuration values to set
 * @param set_ifacecfg is a pointer to a bool that should be set to true if you want the interface configuration to be set.
 *          The function will set/clear this depending on if the interface configuration has actually changed.
 * @param set_hostname is a pointer to a bool that should be set to true if you want the hostname to be set.
 *          The function will set/clear this depending on if the hostname has actually changed.
 * @param set_domain is a pointer to a bool that should be set to true if you want the domain to be set.
 *          The function will set/clear this depending on if the domain has actually changed.
 */
enum e_snc_result snc_set_net_config(const struct s_snc_configuration *config,
                                     const char *method, const char *eth_interface,
                                     const struct s_snc_net_config *net_config,
                                     bool *set_ifacecfg, bool *set_hostname, bool *set_domain);


/*
 * Print configuration
 */

/**
 * Convert the network config to a printable string for debugging purposes
 *
 * @param net_config is the network configuration to print
 * @param single_not_multiline is whether to convert the net_config into a single line or multiple line string
 * @param str is the string to print into
 * @param str_len is the maximum length of str
 */
enum e_snc_result snc_net_config_to_str(const struct s_snc_net_config *net_config, bool single_not_multiline, char *str, size_t str_len);


/**
 * Allocate a suitably sized string and convert the network config to a printable string for debugging purposes
 *
 * Caller is responsible for freeing the returned string
 *
 * @param net_config is the network configuration to print
 * @param single_not_multiline is whether to convert the net_config into a single line or multiple line string
 * @param str is a pointer to a pointer to a string that gets set to the newly allocated string containing the network configuration
 */
enum e_snc_result snc_net_config_to_allocstr(const struct s_snc_net_config *net_config, bool single_not_multiline, char **str);


/**
 * Print out the network config for debugging purposes
 *
 * @param to is the FILE to write the network config to
 * @param single_not_multiline is whether to convert the net_config into a single line or multiple line string
 * @param net_config is the network configuration to print
 */
enum e_snc_result snc_print_net_config(FILE *to, bool single_not_multiline, const struct s_snc_net_config *net_config);


/**
 * Print out the network config in a single line (for legacy getnetcurrent and legacy getnetconfig)
 *
 * @param to is the FILE to write the network config to
 * @param net_config is the network configuration to print
 * @param ip_as_dhcp if set causes the IP address field to be "dhcp" if the network configuration is set to use dhcp
 */
enum e_snc_result snc_print_legacy_net_config_line(FILE *to, const struct s_snc_net_config *net_config, bool ip_as_dhcp);


/*
 * Helpful utilities
 */


/** Initialise the members of existing @ref struct s_snc_net_config to sensible values */
enum e_snc_result snc_net_config_struct_init(struct s_snc_net_config *net_config);


/**
 * Allocate a net config structure and initialise the members of @ref struct s_snc_net_config to sensible values
 *
 * Caller is responsible for freeing
 */
enum e_snc_result snc_net_config_struct_allocate(struct s_snc_net_config **net_config);


/** Function to initialise the members of out_mac to the MAC address contained in inmac */
enum e_snc_result snc_set_mac_address(const unsigned in_mac[SNC_ETH_ALEN], struct sockaddr *out_mac);


/** Function to initialise the members of mac to 0 */
enum e_snc_result snc_zero_mac_address(struct sockaddr *mac);


/** Make the broadcast address from an IP address and netmask */
enum e_snc_result snc_ipv4_broadcast_from_ip_netmask(const struct in_addr *ipv4_address, const struct in_addr *ipv4_netmask, struct in_addr *ipv4_broadcast);


/** Compare two network configurations */
enum e_snc_result snc_net_config_diff(const struct s_snc_net_config *nc1, const struct s_snc_net_config *nc2,
                                      bool *ifcfg_differs, bool *hostname_differs, bool *domain_differs, bool *mac_differs);


/** Convert an @ref enum e_snc_result into a human readable string */
const char *snc_result_to_string(enum e_snc_result result);


#endif /* SETNETCONFIG_H_ */
