/*
 * setnetconfigx.c
 *
 * Program to set the network configuration for an interface in a more advanced fashion than the legacy setnetconfig
 *
 * Invocation is "setnetconfigx <iface> ip <ip> nm <netmask> bc <broadcast> gw <gateway> ns1 <nameserver>"
 */
#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <getopt.h>
#include <sys/stat.h>
#include "lsnc_parse_addr.h"
#include "lsnc_internal.h"
#include "snc_validate.h"
#include <setnetconfig.h>


#define PROGRAM_NAME "setnetconfigx"

static bool cmdline_help = false;
static bool cmdline_version = false;
static bool cmdline_quiet = false;
static const char *cmdline_method = SNC_METHOD_LIB_CONFIG;
static const char *cmdline_configfile = SNC_CONFIG_FILE;


static void print_usage(void)
{
  printf("Usage:\n"
         "  " PROGRAM_NAME " [options] if <iface> ip <ip> nm <netmask> bc <bcast> \\\n"
         "                gw <gateway> ns1 <nameserver1> ns2 <nameserver2> \\\n"
         "                ns3 <nameserver3> sd <search domain>\n"
         "  " PROGRAM_NAME " [options] if <iface> ip dhcp\n"
         "  " PROGRAM_NAME " [options] hn <hostname>\n"
         "  " PROGRAM_NAME " [options] dn <domain>\n"
         "\nOptions:\n"
         "  -v --version          Print the software version\n"
         "  -h --help             Print this help\n"
         "  -q --quiet            Don't print info messages\n"
         "  -c --config <file>    Load the library configuration from this file\n"
         "                        Default " SNC_CONFIG_FILE "\n"
         "  -m --method <method>  Use <method> when setting configuration\n"
         "                        Methods are:\n"
         "                          libconf - Use the method set in config file (default)\n"
         "                          auto - Try each of the available modes until success\n"
         "                          rcconf - Set network config in rc.conf\n"
#if defined(HAVE_LIBNM)
         "                          nm - Set network config via Network Manager\n"
#endif
         "  if                    The network interface to configure. Eg eth0\n"
         "  ip                    The new ip address or \"dhcp\"\n"
         "  nm                    The new network mask\n"
         "  bc                    The new network broadcast address (optional)\n"
         "  gw                    The new network gateway\n"
         "  ns1                   The name first server to use for DNS resolution\n"
         "                        (optional)\n"
         "  ns2                   The name second server to use for DNS resolution\n"
         "                        (optional)\n"
         "  ns3                   The name third server to use for DNS resolution\n"
         "                        (optional)\n"
         "  sd                    The search domain (optional)\n"
         "  hn                    The hostname (\"if\" parameter not required)\n"
         "  dn                    The domain name (\"if\" parameter not required)\n"
         "\n"
         "Program to set the network configuration\n"
         "Network addresses should be in dot-decimal notation (n.n.n.n, 192.168.0.1)\n"
         );
}


static void print_version(void)
{
  printf(PROGRAM_NAME " " PACKAGE_VERSION "\n");
}


typedef enum e_snc_result (*t_str_verify_fn)(const char *str);


struct s_ext_arg
{
  char *name;
  struct in_addr *ipv4_addr;
  char *str;
  size_t str_len;
  char *longname;
  char *example;
  t_str_verify_fn str_verify;
  bool is_set;
  bool for_iface;
  bool iface_required;
};


enum e_ext_args
{
  ea_interface,
  ea_ip_address,
  ea_netmask,
  ea_broadcast,
  ea_gateway,
  ea_nameserver1,
  ea_nameserver2,
  ea_nameserver3,
  ea_search_domain,
  ea_domain_name,
  ea_hostname,

  ea_num
};


static enum e_snc_result parse_extended_arguments(int argc, char *argv[],
                                                  char *interface, struct s_snc_net_config *net_config,
                                                  bool *set_ifacecfg, bool *set_hostname, bool *set_domain)
{
  size_t arg;
  bool found_arg;
  bool any_iface_set;

  struct s_ext_arg ext_args[ea_num] =
  {
    { .name="if",  .ipv4_addr=NULL,                      .str=interface,                 .str_len=SNC_IF_NAMESIZE,     .longname="Interface",     .example="eth0",               .str_verify=NULL,                  .is_set=false, .for_iface=true,  .iface_required=true },
    { .name="ip",  .ipv4_addr=&net_config->ip_address,   .str=NULL,                                                    .longname="IP address",    .example="192.168.0.5",                                           .is_set=false, .for_iface=true,  .iface_required=true },
    { .name="nm",  .ipv4_addr=&net_config->ip_netmask,   .str=NULL,                                                    .longname="Netmask",       .example="255.255.255.0",                                         .is_set=false, .for_iface=true,  .iface_required=true },
    { .name="bc",  .ipv4_addr=&net_config->ip_broadcast, .str=NULL,                                                    .longname="Broadcast",     .example="192.168.0.255",                                         .is_set=false, .for_iface=true,  .iface_required=false },
    { .name="gw",  .ipv4_addr=&net_config->ip_gateway,   .str=NULL,                                                    .longname="Gateway",       .example="192.168.0.1",                                           .is_set=false, .for_iface=true,  .iface_required=false },
    { .name="ns1", .ipv4_addr=&net_config->ip_dns[0],    .str=NULL,                                                    .longname="Name server 1", .example="192.168.0.1",                                           .is_set=false, .for_iface=true,  .iface_required=false },
    { .name="ns2", .ipv4_addr=&net_config->ip_dns[1],    .str=NULL,                                                    .longname="Name server 2", .example="192.168.0.2",                                           .is_set=false, .for_iface=true,  .iface_required=false },
    { .name="ns3", .ipv4_addr=&net_config->ip_dns[2],    .str=NULL,                                                    .longname="Name server 3", .example="192.168.0.3",                                           .is_set=false, .for_iface=true,  .iface_required=false },
    { .name="sd", . ipv4_addr=NULL,                      .str=net_config->search_domain, .str_len=SNC_DOMAIN_NAME_MAX, .longname="Search domain", .example="your.search.domain",                                    .is_set=false, .for_iface=true,  .iface_required=false },
    { .name="dn", . ipv4_addr=NULL,                      .str=net_config->domain,        .str_len=SNC_DOMAIN_NAME_MAX, .longname="Domain name",   .example="your.domain",        .str_verify=snc_domainname_validate, .is_set=false, .for_iface=false, .iface_required=true },
    { .name="hn", . ipv4_addr=NULL,                      .str=net_config->hostname,      .str_len=SNC_HOST_NAME_MAX,   .longname="Host name",     .example="hostname",           .str_verify=snc_hostname_validate,   .is_set=false, .for_iface=false, .iface_required=true },
  };

  /* For each argument, see if it is in the extended arguments list and parse */
  for ( ; optind < argc; optind++)
  {
    for (arg = 0, found_arg = false; arg < ea_num; arg++)
    {
      if (strcmp(ext_args[arg].name, argv[optind]) == 0)
      {
        /* Found it, make sure it has a value */
        found_arg = true;
        optind++;
        if (optind >= argc)
        {
          fprintf(stderr, PROGRAM_NAME ": No value specified for parameter %s. Try --help.\n", ext_args[arg].name);
          return sncr_invalid_commandline;
        }
        else if (NULL != ext_args[arg].ipv4_addr)
        {
          if (snc_parse_ipv4_from_line(NULL, false, argv[optind], ext_args[arg].ipv4_addr) != sncr_success)
          {
            fprintf(stderr, PROGRAM_NAME ": %s parameter should be of the form nnn.nnn.nnn.nnn (eg %s)\n", ext_args[arg].longname, ext_args[arg].example);
            return sncr_invalid_commandline;
          }
          ext_args[arg].is_set = true;
        }
        else if (NULL != ext_args[arg].str)
        {
          if (NULL != ext_args[arg].str_verify && ext_args[arg].str_verify(argv[optind]) != sncr_success)
          {
            fprintf(stderr, PROGRAM_NAME ": %s parameter is invalid (eg %s)\n", ext_args[arg].longname, ext_args[arg].example);
            return sncr_invalid_commandline;
          }
          snprintf(ext_args[arg].str, ext_args[arg].str_len, "%s", argv[optind]);
          ext_args[arg].is_set = true;
        }
        break;
      }
    } /* for arg */
    if (!found_arg)
    {
      fprintf(stderr, PROGRAM_NAME ": %s is not a valid parameter\n", argv[optind]);
      return sncr_invalid_commandline;
    }
  } /* for optind */

  /* Make up the broadcast address if not specified */
  if (ext_args[ea_netmask].is_set && !ext_args[ea_broadcast].is_set)
    (void) snc_ipv4_broadcast_from_ip_netmask(&net_config->ip_address, &net_config->ip_netmask, &net_config->ip_broadcast);

  /* If any of the interface arguments are set, then make sure all the required ones have been set */
  for (arg = 0, any_iface_set = false; arg < ea_num; arg++)
  {
    if (ext_args[arg].is_set && ext_args[arg].for_iface)
    {
      any_iface_set = true;
      break;
    }
  }
  if (any_iface_set)
  {
    for (arg = 0; arg < ea_num; arg++)
    {
      if (ext_args[arg].for_iface && ext_args[arg].iface_required && !ext_args[arg].is_set)
      {
        fprintf(stderr, PROGRAM_NAME ": %s has been specified, but interface (if) has not been specified\n", ext_args[arg].longname);
        return sncr_invalid_commandline;
      }
    }
  }

  /* What fields have been set */
  *set_hostname = ext_args[ea_hostname].is_set;
  *set_domain = ext_args[ea_domain_name].is_set;
  *set_ifacecfg = any_iface_set;

  /* Do we have anything to do */
  if (!*set_hostname && !*set_domain && !*set_ifacecfg)
  {
    fprintf(stderr, PROGRAM_NAME ": Nothing to do. Try --help.\n");
    return sncr_invalid_commandline;
  }

  return sncr_success;
}


static enum e_snc_result parse_opts(int argc, char *argv[],
                                    char *interface, struct s_snc_net_config *net_config,
                                    bool *set_ifacecfg, bool *set_hostname, bool *set_domain)
{
  static const struct option lopts[] =
  {
    { "version",                 0, 0, 'v' },
    { "help",                    0, 0, 'h' },
    { "quiet",                   0, 0, 'q' },
    { "config",                  1, 0, 'c' },
    { "method",                  1, 0, 'm' },
    { NULL,                      0, 0, '\0' },
  };
  int c;

  while (1)
  {
    c = getopt_long(argc, argv, "vhqc:m:", lopts, NULL);
    if (c == -1)
      break;

    switch (c)
    {
      case 'h':
        cmdline_help = true;
        break;

      case 'v':
        cmdline_version = true;
        break;

      case 'q':
        cmdline_quiet = true;
        break;

      case 'c':
        cmdline_configfile = optarg;
        break;

      case 'm':
        cmdline_method = optarg;
        break;

      default:
        fprintf(stderr, PROGRAM_NAME ": Bad command line option. Try --help.\n");
        return sncr_invalid_commandline;
        break;
    }
  }

  if (cmdline_help || cmdline_version)
    return sncr_success;

  return parse_extended_arguments(argc, argv, interface, net_config, set_ifacecfg, set_hostname, set_domain);


#if 0
  hostname_specified = (NULL != cmdline_hostname);
  domain_specified = (NULL != cmdline_domain);
  if (!cmdline_help && !cmdline_version)
  {
    if (optind >= argc && (hostname_specified || domain_specified))
    {
      return sncr_success;
    }

    if (optind >= argc)
    {
      fprintf(stderr, PROGRAM_NAME ": No network interface specified. Try --help.\n");
      return sncr_invalid_commandline;
    }
    cmdline_interface = argv[optind];
    optind++;

    if (optind >= argc)
    {
      fprintf(stderr, PROGRAM_NAME ": No ip address specified. Try --help.\n");
      return sncr_invalid_commandline;
    }
    cmdline_ipaddr = argv[optind];
    optind++;

    isdhcp = (0 == strcmp(SNC_RCCONF_VARVALUE_IPADDR_DHCP, cmdline_ipaddr));
    if (!isdhcp)
    {
      if (optind >= argc)
      {
        fprintf(stderr, PROGRAM_NAME ": No network mask specified. Try --help.\n");
        return sncr_invalid_commandline;
      }
      cmdline_netmask = argv[optind];
      optind++;

      if (optind >= argc)
      {
        fprintf(stderr, PROGRAM_NAME ": No network broadcast address specified. Try --help.\n");
        return sncr_invalid_commandline;
      }
      cmdline_broadcast = argv[optind];
      optind++;

      if (optind >= argc)
      {
        fprintf(stderr, PROGRAM_NAME ": No network gateway specified. Try --help.\n");
        return sncr_invalid_commandline;
      }
      cmdline_gateway = argv[optind];
      optind++;

      if (optind >= argc)
      {
        fprintf(stderr, PROGRAM_NAME ": No name server specified. Try --help.\n");
        return sncr_invalid_commandline;
      }
      cmdline_nameserver = argv[optind];
      optind++;
    }
  }
#endif

  return sncr_success;
}


/** setnetconfigx entry point */
#if ENABLE_MULTI_CALL
int setnetconfigx_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
  struct s_snc_configuration *config = NULL;
  struct s_snc_net_config new_net_config;
  enum e_snc_result result;
  bool set_ifacecfg = false;
  bool set_hostname = false;
  bool set_domain = false;
  char eth_interface[SNC_IF_NAMESIZE];
  const char *eth_interface_ptr;

  /* No exec for everyone, no write for group and other */
  umask(S_IXUSR | S_IXGRP | S_IXOTH | S_IWGRP | S_IWOTH);

  /* Initialise everything */
  snc_net_config_struct_init(&new_net_config);

  /* Check command line parameters */
  result = parse_opts(argc, argv, eth_interface, &new_net_config, &set_ifacecfg, &set_hostname, &set_domain);
  if (sncr_success != result)
    goto m_done;
  eth_interface_ptr = (set_ifacecfg) ? eth_interface : NULL;

  /* Help and version */
  if (cmdline_help)
  {
    print_usage();
    result = sncr_success;
    goto m_done;
  }
  if (cmdline_version)
  {
    print_version();
    result = sncr_success;
    goto m_done;
  }

  /* Try to load the library configuration */
  result = snc_config_init(&config);
  if (sncr_success != result)
    goto m_done;
  (void) snc_config_load(cmdline_configfile, config);

  /* Validate the configuration */
  result = validate_and_print_errors(PROGRAM_NAME, &new_net_config, set_ifacecfg, set_hostname, set_domain);
  if (sncr_success != result)
    goto m_done;

  /* Set the new network configuration */
  result = snc_set_net_config(config, cmdline_method, eth_interface_ptr, &new_net_config,
                              &set_ifacecfg, &set_hostname, &set_domain);
  if (sncr_success != result)
    goto m_done;

  /* Print out what happened */
  if (!cmdline_quiet)
  {
    if (set_hostname)
      fprintf(stdout, PROGRAM_NAME ": Changed the hostname setting\n");
    if (set_domain)
      fprintf(stdout, PROGRAM_NAME ": Changed the domain setting\n");
    if (set_ifacecfg)
    {
      if (new_net_config.use_dhcp)
        fprintf(stdout, PROGRAM_NAME ": Changed to DHCP IP address settings\n");
      else
        fprintf(stdout, PROGRAM_NAME ": Changed static IP address settings\n");
    }

    if (!set_hostname && !set_domain && !set_ifacecfg)
      fprintf(stdout, PROGRAM_NAME ": No change to network settings\n");
  }

  /* Success */
  result = sncr_success;

m_done:
  SNC_DESTROY_AND_NULLIFY(snc_config_destroy, config);
  if (sncr_success != result)
    fprintf(stderr, PROGRAM_NAME ": failed with result %s\n", snc_result_to_string(result));
  return result;
}
