/*
 * lsnc_test_hostname.c
 *
 *  Created on: 28 Jun. 2018
 *      Author: ian
 */
#include <CuTest.h>
#include <setnetconfig.h>

#include "lsnc_test_hostname.h"


/** Is ch allowed in a hostname */
static bool is_bad_hostname_char(unsigned char ch)
{
  return (ch < '-' ||
          (ch > '-' && ch < '0') ||
          (ch > '9' && ch < 'A') ||
          (ch > 'Z' && ch < 'a') ||
          ch > 'z');
}


/** Test snc_hostname_validate() */
static void test_snc_hostname_validate(CuTest* tc)
{
  enum e_snc_result result;
  enum e_snc_result expected;
  char str[SNC_HOST_NAME_MAX + 2];
  unsigned char ch;

  /* Valid hostname */
  result = snc_hostname_validate("validhostname");
  CuAssertIntEquals(tc, sncr_success, result);

  /* Minimum */
  result = snc_hostname_validate("a");
  CuAssertIntEquals(tc, sncr_success, result);

  /* Too short */
  result = snc_hostname_validate("");
  CuAssertIntEquals(tc, sncr_parseerror, result);

  /* Maximum */
  snprintf(str, SNC_HOST_NAME_MAX + 2, "%0*d", SNC_HOST_NAME_MAX, 0);
  result = snc_hostname_validate(str);
  CuAssertIntEquals(tc, sncr_success, result);

  /* Too long */
  snprintf(str, SNC_HOST_NAME_MAX + 2, "%0*d", SNC_HOST_NAME_MAX + 1, 0);
  result = snc_hostname_validate(str);
  CuAssertIntEquals(tc, sncr_parseerror, result);

  /* Bad characters */
  for (ch = 0; ch < 255; ch++)
  {
    expected = (is_bad_hostname_char(ch)) ? sncr_parseerror : sncr_success;
    snprintf(str, SNC_HOST_NAME_MAX + 2, "%c", (char) ch);
    result = snc_hostname_validate(str);
    CuAssertIntEquals(tc, expected, result);
  }
}


/** Test snc_domainname_validate() */
static void test_snc_domainname_validate(CuTest* tc)
{
  enum e_snc_result result;
  char str[SNC_DOMAIN_NAME_MAX + 2];
  size_t i;

  /* Valid domain names */
  result = snc_domainname_validate("valid.domain");
  CuAssertIntEquals(tc, sncr_success, result);
  result = snc_domainname_validate("validdomain");
  CuAssertIntEquals(tc, sncr_success, result);
  result = snc_domainname_validate("longer.valid.domain");
  CuAssertIntEquals(tc, sncr_success, result);

  /* We allow an empty domain name */
  result = snc_domainname_validate("");
  CuAssertIntEquals(tc, sncr_success, result);

  /* Single character */
  result = snc_domainname_validate("a");
  CuAssertIntEquals(tc, sncr_success, result);

  /* Any one label can be max SNC_HOST_NAME_MAX characters */
  snprintf(str, SNC_DOMAIN_NAME_MAX + 2, "%0*d.%0*d", SNC_HOST_NAME_MAX, 0, SNC_HOST_NAME_MAX, 0);
  result = snc_domainname_validate(str);
  CuAssertIntEquals(tc, sncr_success, result);
  snprintf(str, SNC_DOMAIN_NAME_MAX + 2, "%0*d.%0*d", SNC_HOST_NAME_MAX, 0, SNC_HOST_NAME_MAX + 1, 0);
  result = snc_domainname_validate(str);
  CuAssertIntEquals(tc, sncr_parseerror, result);

  /* Maximum overall length */
  for (i = 0; i < SNC_DOMAIN_NAME_MAX; )
  {
    i += snprintf(&str[i], SNC_DOMAIN_NAME_MAX + 1 - i, "%0*d.", SNC_HOST_NAME_MAX, 0);
  }
  result = snc_domainname_validate(str);
  CuAssertIntEquals(tc, sncr_success, result);

  /* Maximum overall length */
  for (i = 0; i < SNC_DOMAIN_NAME_MAX + 1; )
  {
    i += snprintf(&str[i], SNC_DOMAIN_NAME_MAX + 2 - i, "%0*d.", SNC_HOST_NAME_MAX, 0);
  }
  result = snc_domainname_validate(str);
  CuAssertIntEquals(tc, sncr_parseerror, result);
}


/** Test the hostname functions of libsetnetconfig */
CuSuite *CuGetSuite_lsnc_hostname(void)
{
  CuSuite* suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_snc_hostname_validate);
  SUITE_ADD_TEST(suite, test_snc_domainname_validate);

  return suite;
}
