/*
 * lsnc_test_parse_addr.c
 *
 *  Created on: 29 Jun. 2018
 *      Author: ian
 */
#include <string.h>
#include <CuTest.h>
#include <setnetconfig.h>
#include <lsnc_parse_addr.h>

#include "lsnc_test_parse_addr.h"


/** Convert 4 IP octets into an 32bit number */
#define IP_TO_ULONG(a, b, c, d) (a << 24 | b << 16 | c << 8 | d)


/** Test snc_skip_whitespace() */
static void test_snc_skip_whitespace(CuTest* tc)
{
  enum e_snc_result result;
  const char *line;

  /* Test a string with some whitespace */
  line = "  \tasdf";
  result = snc_skip_whitespace(&line);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertStrEquals(tc, "asdf", line);

  /* Test a string with no whitespace */
  line = "fdsa";
  result = snc_skip_whitespace(&line);
  CuAssertIntEquals(tc, sncr_notfound, result);
  CuAssertStrEquals(tc, "fdsa", line);

  /* Test a string with only whitespace */
  line = "  ";
  result = snc_skip_whitespace(&line);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertStrEquals(tc, "", line);

  /* Test an empty string */
  line = "";
  result = snc_skip_whitespace(&line);
  CuAssertIntEquals(tc, sncr_notfound, result);
  CuAssertStrEquals(tc, "", line);
}


/** Test snc_ipv4_prefix_to_netmask() */
static void test_snc_ipv4_prefix_to_netmask(CuTest* tc)
{
  enum e_snc_result result;
  struct in_addr ipv4_netmask;

  result = snc_ipv4_prefix_to_netmask(0, &ipv4_netmask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(0), ipv4_netmask.s_addr);

  result = snc_ipv4_prefix_to_netmask(8, &ipv4_netmask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(0xFF000000), ipv4_netmask.s_addr);

  result = snc_ipv4_prefix_to_netmask(16, &ipv4_netmask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(0xFFFF0000), ipv4_netmask.s_addr);

  result = snc_ipv4_prefix_to_netmask(24, &ipv4_netmask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(0xFFFFFF00), ipv4_netmask.s_addr);

  result = snc_ipv4_prefix_to_netmask(32, &ipv4_netmask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(0xFFFFFFFF), ipv4_netmask.s_addr);

  result = snc_ipv4_prefix_to_netmask(33, &ipv4_netmask);
  CuAssertIntEquals(tc, sncr_paramerror, result);
  result = snc_ipv4_prefix_to_netmask(0, NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);
}


/** Test snc_ipv4_netmask_to_prefix() */
static void test_snc_ipv4_netmask_to_prefix(CuTest* tc)
{
  enum e_snc_result result;
  struct in_addr ipv4_netmask;
  unsigned prefix;

  ipv4_netmask.s_addr = htonl(0);
  result = snc_ipv4_netmask_to_prefix(&ipv4_netmask, &prefix);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, 0, prefix);

  ipv4_netmask.s_addr = htonl(0xFF000000);
  result = snc_ipv4_netmask_to_prefix(&ipv4_netmask, &prefix);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, 8, prefix);

  ipv4_netmask.s_addr = htonl(0xFFFF0000);
  result = snc_ipv4_netmask_to_prefix(&ipv4_netmask, &prefix);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, 16, prefix);

  ipv4_netmask.s_addr = htonl(0xFFFFFF00);
  result = snc_ipv4_netmask_to_prefix(&ipv4_netmask, &prefix);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, 24, prefix);

  ipv4_netmask.s_addr = htonl(0xFFFFFFFF);
  result = snc_ipv4_netmask_to_prefix(&ipv4_netmask, &prefix);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, 32, prefix);

  result = snc_ipv4_netmask_to_prefix(NULL, &prefix);
  CuAssertIntEquals(tc, sncr_paramerror, result);
  result = snc_ipv4_netmask_to_prefix(&ipv4_netmask, NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);
}


/** Test snc_parse_string_from_line() */
static void test_snc_parse_string_from_line(CuTest* tc)
{
  enum e_snc_result result;
  char str[20];

  result = snc_parse_string_from_line(NULL, false, "testdata asdf", str, 20);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertStrEquals(tc, "testdata", str);

  result = snc_parse_string_from_line("prefix", false, "prefix testdata asdf", str, 20);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertStrEquals(tc, "testdata", str);

  result = snc_parse_string_from_line("prefix", false, "blah testdata asdf", str, 20);
  CuAssertIntEquals(tc, sncr_notfound, result);

  result = snc_parse_string_from_line("prefix", false, "asdf prefix testdata asdf", str, 20);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertStrEquals(tc, "testdata", str);

  result = snc_parse_string_from_line("prefix", true, "prefix testdata asdf", str, 20);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertStrEquals(tc, "testdata", str);

  result = snc_parse_string_from_line("asdf prefix", true, "prefix testdata asdf", str, 20);
  CuAssertIntEquals(tc, sncr_notfound, result);
}


/** Test snc_parse_ipv4_from_line() */
static void test_snc_parse_ipv4_from_line(CuTest* tc)
{
  enum e_snc_result result;
  struct in_addr ipv4_addr;

  result = snc_parse_ipv4_from_line(NULL, false, "172.17.1.1 asdf", &ipv4_addr);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(172, 17, 1, 1)), ipv4_addr.s_addr);

  result = snc_parse_ipv4_from_line("prefix", false, "prefix 172.17.1.2 asdf", &ipv4_addr);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(172, 17, 1, 2)), ipv4_addr.s_addr);

  result = snc_parse_ipv4_from_line("prefix", false, "blah 172.17.1.3 asdf", &ipv4_addr);
  CuAssertIntEquals(tc, sncr_notfound, result);

  result = snc_parse_ipv4_from_line("prefix", false, "asdf prefix 172.17.1.4 asdf", &ipv4_addr);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(172, 17, 1, 4)), ipv4_addr.s_addr);

  result = snc_parse_ipv4_from_line("prefix", true, "prefix 172.17.1.5 asdf", &ipv4_addr);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(172, 17, 1, 5)), ipv4_addr.s_addr);

  result = snc_parse_ipv4_from_line("asdf prefix", true, "prefix 172.17.1.6 asdf", &ipv4_addr);
  CuAssertIntEquals(tc, sncr_notfound, result);
}


/** Test snc_parse_ipv4_cidr_from_line() */
static void test_snc_parse_ipv4_cidr_from_line(CuTest* tc)
{
  enum e_snc_result result;
  struct in_addr ipv4_addr;
  struct in_addr ipv4_mask;

  result = snc_parse_ipv4_cidr_from_line(NULL, false, "172.17.1.1/24 asdf", &ipv4_addr, &ipv4_mask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(172, 17, 1, 1)), ipv4_addr.s_addr);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(255, 255, 255, 0)), ipv4_mask.s_addr);

  result = snc_parse_ipv4_cidr_from_line("prefix", false, "prefix 172.17.1.2/16 asdf", &ipv4_addr, &ipv4_mask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(172, 17, 1, 2)), ipv4_addr.s_addr);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(255, 255, 0, 0)), ipv4_mask.s_addr);

  result = snc_parse_ipv4_cidr_from_line("prefix", false, "blah 172.17.1.3/24 asdf", &ipv4_addr, &ipv4_mask);
  CuAssertIntEquals(tc, sncr_notfound, result);

  result = snc_parse_ipv4_cidr_from_line("prefix", false, "asdf prefix 172.17.1.4/8 asdf", &ipv4_addr, &ipv4_mask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(172, 17, 1, 4)), ipv4_addr.s_addr);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(255, 0, 0, 0)), ipv4_mask.s_addr);

  result = snc_parse_ipv4_cidr_from_line("prefix", true, "prefix 172.17.1.5/24 asdf", &ipv4_addr, &ipv4_mask);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(172, 17, 1, 5)), ipv4_addr.s_addr);
  CuAssertIntEquals(tc, htonl(IP_TO_ULONG(255, 255, 255, 0)), ipv4_mask.s_addr);

  result = snc_parse_ipv4_cidr_from_line("asdf prefix", true, "prefix 172.17.1.6/8 asdf", &ipv4_addr, &ipv4_mask);
  CuAssertIntEquals(tc, sncr_notfound, result);
}


/** Test snc_parse_mac_from_line() */
static void test_snc_parse_mac_from_line(CuTest* tc)
{
  enum e_snc_result result;
  struct sockaddr mac_address;

  result = snc_parse_mac_from_line(NULL, false, "00:11:22:33:44:55 asdf", &mac_address);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, AF_PACKET, mac_address.sa_family);
  CuAssertIntEquals(tc, 0, memcmp("\x00\x11\x22\x33\x44\x55", mac_address.sa_data, SNC_ETH_ALEN));

  result = snc_parse_mac_from_line("prefix", false, "prefix 00:11:22:33:44:55 asdf", &mac_address);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, AF_PACKET, mac_address.sa_family);
  CuAssertIntEquals(tc, 0, memcmp("\x00\x11\x22\x33\x44\x55", mac_address.sa_data, SNC_ETH_ALEN));

  result = snc_parse_mac_from_line("prefix", false, "blah 00:11:22:33:44:55 asdf", &mac_address);
  CuAssertIntEquals(tc, sncr_notfound, result);

  result = snc_parse_mac_from_line("prefix", false, "asdf prefix 00:11:22:33:44:55 asdf", &mac_address);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, AF_PACKET, mac_address.sa_family);
  CuAssertIntEquals(tc, 0, memcmp("\x00\x11\x22\x33\x44\x55", mac_address.sa_data, SNC_ETH_ALEN));

  result = snc_parse_mac_from_line("prefix", true, "prefix 00:11:22:33:44:55 asdf", &mac_address);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, AF_PACKET, mac_address.sa_family);
  CuAssertIntEquals(tc, 0, memcmp("\x00\x11\x22\x33\x44\x55", mac_address.sa_data, SNC_ETH_ALEN));

  result = snc_parse_mac_from_line("asdf prefix", true, "prefix 00:11:22:33:44:55 asdf", &mac_address);
  CuAssertIntEquals(tc, sncr_notfound, result);

  result = snc_parse_mac_from_line(NULL, false, "00:11:22:33:44", &mac_address);
  CuAssertIntEquals(tc, sncr_parseerror, result);

  result = snc_parse_mac_from_line(NULL, false, "00:11:22:33:44:GZ", &mac_address);
  CuAssertIntEquals(tc, sncr_parseerror, result);
}


/** Test the parse address functions of libsetnetconfig */
CuSuite *CuGetSuite_lsnc_parse_addr(void)
{
  CuSuite* suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_snc_skip_whitespace);
  SUITE_ADD_TEST(suite, test_snc_ipv4_prefix_to_netmask);
  SUITE_ADD_TEST(suite, test_snc_ipv4_netmask_to_prefix);
  SUITE_ADD_TEST(suite, test_snc_parse_string_from_line);
  SUITE_ADD_TEST(suite, test_snc_parse_ipv4_from_line);
  SUITE_ADD_TEST(suite, test_snc_parse_ipv4_cidr_from_line);
  SUITE_ADD_TEST(suite, test_snc_parse_mac_from_line);

  return suite;
}
