/*
 * lsnc_test_print_net_config.c
 *
 *  Created on: 5 Jul. 2018
 *      Author: ian
 */
#include <stdlib.h>
#include <stdio.h>
#include <CuTest.h>
#include <setnetconfig.h>
#include <lsnc_internal.h>

#include "lsnc_test_print_net_config.h"


static const char expected_single_line[] = \
  "dhcp: yes ip: 172.17.1.108 nm: 255.255.255.0 "
  "bc: 172.17.1.255 gw: 172.17.1.1 "
  "ns0: 172.17.1.2 ns1: 172.17.1.3 ns2: 0.0.0.0 "
  "sd: \"test.search.domain\" hn: \"test-hostname\" dn: \"test.domain\" mac: BA:DC:0D:E0:00:01";

static const char expected_multi_line[] = \
  "Use DHCP:    yes\n"
  "IP Address:  172.17.1.108\n"
  "Netmask:     255.255.255.0\n"
  "Broadcast:   172.17.1.255\n"
  "Gateway:     172.17.1.1\n"
  "DNS 0:       172.17.1.2\n"
  "DNS 1:       172.17.1.3\n"
  "DNS 2:       0.0.0.0\n"
  "Search:      test.search.domain\n"
  "Hostname:    test-hostname\n"
  "Domain:      test.domain\n"
  "MAC Address: BA:DC:0D:E0:00:01\n";

static const char expected_legacy_line_static[] = \
"172.17.1.108 255.255.255.0 172.17.1.255 172.17.1.1 172.17.1.2 BA:DC:0D:E0:00:01 test-hostname\n";

static const char expected_legacy_line_dhcp[] = \
"dhcp 255.255.255.0 172.17.1.255 172.17.1.1 172.17.1.2 BA:DC:0D:E0:00:01 test-hostname\n";


static void setup_net_config(struct s_snc_net_config *nc)
{
  snc_net_config_struct_init(nc);
  const unsigned mac[SNC_ETH_ALEN] = { 0xBA, 0xDC, 0x0D, 0xE0, 0x00, 0x01 };

  nc->use_dhcp = true;
  nc->ip_address.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(172, 17, 1, 108);
  nc->ip_netmask.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(255, 255, 255, 0);
  nc->ip_broadcast.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(172, 17, 1, 255);
  nc->ip_gateway.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(172, 17, 1, 1);
  nc->ip_dns[0].s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(172, 17, 1, 2);
  nc->ip_dns[1].s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(172, 17, 1, 3);
  nc->ip_dns[2].s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(0, 0, 0, 0);
  snprintf(nc->search_domain, SNC_DOMAIN_NAME_MAX, "test.search.domain");

  snprintf(nc->hostname, SNC_HOST_NAME_MAX, "test-hostname");
  snprintf(nc->domain, SNC_DOMAIN_NAME_MAX, "test.domain");

  snc_set_mac_address(mac, &nc->mac_address);
}


/** Test snc_net_config_to_str() */
static void test_snc_net_config_to_str(CuTest* tc)
{
  enum e_snc_result result;
  struct s_snc_net_config net_config;
  char str[500];

  setup_net_config(&net_config);

  result = snc_net_config_to_str(NULL, false, str, 500);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_net_config_to_str(&net_config, true, str, 500);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertStrEquals(tc, expected_single_line, str);

  result = snc_net_config_to_str(&net_config, false, str, 500);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertStrEquals(tc, expected_multi_line, str);
}


/** Test snc_net_config_to_allocstr() */
static void test_snc_net_config_to_allocstr(CuTest* tc)
{
  enum e_snc_result result;
  struct s_snc_net_config net_config;
  char *str = NULL;

  setup_net_config(&net_config);

  result = snc_net_config_to_allocstr(NULL, false, &str);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_net_config_to_allocstr(&net_config, false, NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_net_config_to_allocstr(&net_config, true, &str);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertPtrNotNull(tc, str);
  CuAssertStrEquals(tc, expected_single_line, str);
  SNC_FREE_AND_NULLIFY(str);

  result = snc_net_config_to_allocstr(&net_config, false, &str);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertPtrNotNull(tc, str);
  CuAssertStrEquals(tc, expected_multi_line, str);
  SNC_FREE_AND_NULLIFY(str);
}


/** Check snc_print_net_config() */
static void check_snc_print_net_config(CuTest* tc, bool single_not_multiline)
{
  enum e_snc_result result;
  struct s_snc_net_config net_config;
  FILE *file;
  char str[500];
  ssize_t readlen;
  const char *expected = (single_not_multiline) ? expected_single_line : expected_multi_line;

  file = tmpfile();
  CuAssertPtrNotNull(tc, file);

  setup_net_config(&net_config);

  result = snc_print_net_config(NULL, single_not_multiline, &net_config);
  CuAssertIntEquals(tc, sncr_paramerror, result);
  result = snc_print_net_config(file, single_not_multiline, NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_print_net_config(file, single_not_multiline, &net_config);
  CuAssertIntEquals(tc, sncr_success, result);
  fseek(file, 0, SEEK_SET);
  readlen = fread(str, 1, 500, file);
  CuAssertTrue(tc, readlen > 0);
  str[readlen] = '\0';

  CuAssertStrEquals(tc, expected, str);
  fclose(file);
}


/** Test snc_print_net_config() */
static void test_snc_print_net_config(CuTest* tc)
{
  check_snc_print_net_config(tc, false);
  check_snc_print_net_config(tc, true);
}


/** Check snc_print_legacy_net_config_line() */
static void check_snc_print_legacy_net_config_line(CuTest* tc, bool dhcp_not_static)
{
  enum e_snc_result result;
  struct s_snc_net_config net_config;
  FILE *file;
  char str[500];
  ssize_t readlen;
  const char *expected = (dhcp_not_static) ? expected_legacy_line_dhcp : expected_legacy_line_static;

  file = tmpfile();
  CuAssertPtrNotNull(tc, file);

  setup_net_config(&net_config);

  result = snc_print_legacy_net_config_line(NULL, &net_config, dhcp_not_static);
  CuAssertIntEquals(tc, sncr_paramerror, result);
  result = snc_print_legacy_net_config_line(file, NULL, dhcp_not_static);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_print_legacy_net_config_line(file, &net_config, dhcp_not_static);
  CuAssertIntEquals(tc, sncr_success, result);
  fseek(file, 0, SEEK_SET);
  readlen = fread(str, 1, 500, file);
  CuAssertTrue(tc, readlen > 0);
  str[readlen] = '\0';

  CuAssertStrEquals(tc, expected, str);
  fclose(file);
}


/** Test snc_print_legacy_net_config_line() */
static void test_snc_print_legacy_net_config_line(CuTest* tc)
{
  check_snc_print_legacy_net_config_line(tc, false);
  check_snc_print_legacy_net_config_line(tc, true);
}


/** Test the print net config functions of libsetnetconfig */
CuSuite *CuGetSuite_lsnc_print_net_config(void)
{
  CuSuite* suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_snc_net_config_to_str);
  SUITE_ADD_TEST(suite, test_snc_net_config_to_allocstr);
  SUITE_ADD_TEST(suite, test_snc_print_net_config);
  SUITE_ADD_TEST(suite, test_snc_print_legacy_net_config_line);

  return suite;
}
