/*
 * lsnc_test_util.c
 *
 *  Created on: 4 Jul. 2018
 *      Author: ian
 */
#include <stdlib.h>
#include <CuTest.h>
#include <setnetconfig.h>
#include "lsnc_internal.h"

#include "lsnc_test_util.h"


/** Test snc_net_config_struct_init() */
static void check_net_config_struct_init(CuTest *tc, const struct s_snc_net_config *net_config)
{
  size_t i;

  CuAssertIntEquals(tc, false, net_config->use_dhcp);
  CuAssertIntEquals(tc, SNC_IPV4_UNSET_INADDR_NETWORK, net_config->ip_address.s_addr);
  CuAssertIntEquals(tc, SNC_IPV4_UNSET_INADDR_NETWORK, net_config->ip_netmask.s_addr);
  CuAssertIntEquals(tc, SNC_IPV4_UNSET_INADDR_NETWORK, net_config->ip_broadcast.s_addr);
  CuAssertIntEquals(tc, SNC_IPV4_UNSET_INADDR_NETWORK, net_config->ip_gateway.s_addr);
  for (i = 0; i < SNC_NUM_DNS_SERVERS; i++)
    CuAssertIntEquals(tc, SNC_IPV4_UNSET_INADDR_NETWORK, net_config->ip_dns[i].s_addr);

  CuAssertStrEquals(tc, SNC_DOMAINNAME_DEFAULT, net_config->search_domain);
  CuAssertStrEquals(tc, SNC_DOMAINNAME_DEFAULT, net_config->domain);
  CuAssertStrEquals(tc, SNC_HOSTNAME_DEFAULT, net_config->hostname);
}


/** Test snc_net_config_struct_init() */
static void test_snc_net_config_struct_init(CuTest *tc)
{
  enum e_snc_result result;
  struct s_snc_net_config net_config;

  result = snc_net_config_struct_init(NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_net_config_struct_init(&net_config);
  CuAssertIntEquals(tc, sncr_success, result);
  check_net_config_struct_init(tc, &net_config);
}


/** Test snc_net_config_struct_allocate() */
static void test_snc_net_config_struct_allocate(CuTest* tc)
{
  enum e_snc_result result;
  struct s_snc_net_config *net_config = NULL;

  result = snc_net_config_struct_allocate(NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_net_config_struct_allocate(&net_config);
  CuAssertIntEquals(tc, sncr_success, result);
  check_net_config_struct_init(tc, net_config);

  SNC_FREE_AND_NULLIFY(net_config);
}


/** Test snc_set_mac_address() */
static void test_snc_set_mac_address(CuTest* tc)
{
  const unsigned in_mac[SNC_ETH_ALEN] = { 0x11, 0x22, 0x33, 0x33, 0x55, 0x66 };
  struct sockaddr out_mac;
  enum e_snc_result result;
  size_t i;

  result = snc_set_mac_address(NULL, &out_mac);
  CuAssertIntEquals(tc, sncr_paramerror, result);
  result = snc_set_mac_address(in_mac, NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_set_mac_address(in_mac, &out_mac);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, AF_PACKET, out_mac.sa_family);
  for (i = 0; i < SNC_ETH_ALEN; i++)
  {
    CuAssertIntEquals(tc, in_mac[i], out_mac.sa_data[i]);
  }
  for ( ; i < sizeof(out_mac.sa_data); i++)
  {
    CuAssertIntEquals(tc, 0, out_mac.sa_data[i]);
  }
}


/** Test snc_zero_mac_address() */
static void test_snc_zero_mac_address(CuTest* tc)
{
  struct sockaddr out_mac;
  enum e_snc_result result;
  size_t i;

  result = snc_zero_mac_address(NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  result = snc_zero_mac_address(&out_mac);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, AF_PACKET, out_mac.sa_family);
  for (i = 0; i < sizeof(out_mac.sa_data); i++)
  {
    CuAssertIntEquals(tc, 0, out_mac.sa_data[i]);
  }
}


/** Test snc_ipv4_broadcast_from_ip_netmask() */
static void test_snc_ipv4_broadcast_from_ip_netmask(CuTest* tc)
{
  struct in_addr ip;
  struct in_addr nm;
  struct in_addr bc;
  enum e_snc_result result;

  result = snc_ipv4_broadcast_from_ip_netmask(NULL, &nm, &bc);
  CuAssertIntEquals(tc, sncr_paramerror, result);
  result = snc_ipv4_broadcast_from_ip_netmask(&ip, NULL, &bc);
  CuAssertIntEquals(tc, sncr_paramerror, result);
  result = snc_ipv4_broadcast_from_ip_netmask(&ip, &nm, NULL);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  ip.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(192, 168, 0, 1);
  nm.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(255, 255, 255, 0);
  result = snc_ipv4_broadcast_from_ip_netmask(&ip, &nm, &bc);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, SNC_IPV4_OCTETS_TO_INADDR_NETWORK(192, 168, 0, 255), bc.s_addr);
}


/** Test snc_net_config_diff() */
static void test_snc_net_config_diff(CuTest* tc)
{
  struct s_snc_net_config nc1;
  struct s_snc_net_config nc2;
  bool ifcfg_differs;
  bool hostname_differs;
  bool domain_differs;
  bool mac_differs;
  enum e_snc_result result;

  result = snc_net_config_diff(&nc1, NULL, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_paramerror, result);
  result = snc_net_config_diff(NULL, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_paramerror, result);

  (void) snc_net_config_struct_init(&nc1);
  (void) snc_net_config_struct_init(&nc2);
  result = snc_net_config_diff(&nc1, &nc2, NULL, NULL, NULL, NULL);
  CuAssertIntEquals(tc, sncr_success, result);

  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, false, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  nc1.use_dhcp = true;
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, true, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  nc1.use_dhcp = false;
  nc1.ip_address.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(192, 168, 0, 1);
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, true, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  nc1.ip_address.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  nc1.ip_netmask.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(192, 168, 0, 1);
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, true, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  nc1.ip_netmask.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  nc1.ip_broadcast.s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(192, 168, 0, 1);
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, true, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  nc1.ip_broadcast.s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  nc1.ip_dns[0].s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(192, 168, 0, 1);
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, true, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  nc1.ip_dns[0].s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  nc1.ip_dns[1].s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(192, 168, 0, 1);
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, true, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  nc1.ip_dns[1].s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  nc1.ip_dns[2].s_addr = SNC_IPV4_OCTETS_TO_INADDR_NETWORK(192, 168, 0, 1);
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, true, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  nc1.ip_dns[2].s_addr = SNC_IPV4_UNSET_INADDR_NETWORK;
  snprintf(nc1.search_domain, SNC_DOMAIN_NAME_MAX, "test.search.domain");
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, true, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  snprintf(nc1.search_domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
  snprintf(nc1.hostname, SNC_HOST_NAME_MAX, "test-hostname");
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, false, ifcfg_differs);
  CuAssertIntEquals(tc, true, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  snprintf(nc1.hostname, SNC_HOST_NAME_MAX, "%s", SNC_HOSTNAME_DEFAULT);
  snprintf(nc1.domain, SNC_DOMAIN_NAME_MAX, "test.domain");
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, false, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, true, domain_differs);
  CuAssertIntEquals(tc, false, mac_differs);

  snprintf(nc1.domain, SNC_DOMAIN_NAME_MAX, "%s", SNC_DOMAINNAME_DEFAULT);
  nc1.mac_address.sa_data[0] = 0x11;
  result = snc_net_config_diff(&nc1, &nc2, &ifcfg_differs, &hostname_differs, &domain_differs, &mac_differs);
  CuAssertIntEquals(tc, sncr_success, result);
  CuAssertIntEquals(tc, false, ifcfg_differs);
  CuAssertIntEquals(tc, false, hostname_differs);
  CuAssertIntEquals(tc, false, domain_differs);
  CuAssertIntEquals(tc, true, mac_differs);
}


/** Test the utility functions of libsetnetconfig */
CuSuite *CuGetSuite_lsnc_util(void)
{
  CuSuite* suite = CuSuiteNew();

  SUITE_ADD_TEST(suite, test_snc_net_config_struct_init);
  SUITE_ADD_TEST(suite, test_snc_net_config_struct_allocate);
  SUITE_ADD_TEST(suite, test_snc_set_mac_address);
  SUITE_ADD_TEST(suite, test_snc_zero_mac_address);
  SUITE_ADD_TEST(suite, test_snc_ipv4_broadcast_from_ip_netmask);
  SUITE_ADD_TEST(suite, test_snc_net_config_diff);

  return suite;
}
