#!/bin/bash
# Testing for getnetcurrent, getnetconfig and setnetconfig

#set -e
set -x

# This interface needs to be present and configured for these tests to work (getnetcurrent)
IFACE=ens33

# This is the interface used in rc.conf
RCIFACE=eth0

# Executables to test
GNCU=../src/getnetcurrent
GNC=../src/getnetconfig
SNC=../src/setnetconfig
SNCX=../src/setnetconfigx

# Methods
GNCU_METHODS="libconf auto ioctl getifaddrs ifconfig ip nm"
GNC_METHODS="libconf auto rcconf"

# Example config file
LSNCCONF=libsetnetconfig.conf
LSNCCONFTMP=libsetnetconfig.conf.temp1
LSNCCONFTMP2=libsetnetconfig.conf.temp2
RCCONF=rc.conf
RCCONFTMP=`sed -n -r 's/^snc_rcconf="?(.*?)"?$/\1/p' $LSNCCONF`
cp $LSNCCONF $LSNCCONFTMP
cp $LSNCCONF $LSNCCONFTMP2
cp $RCCONF $RCCONFTMP

# Make sure it's clean to begin with
rm -f nrresult.txt hrresult.txt

# Return codes
sncr_success=0              # The function has completed successfully
sncr_paramerror=1           # There was a problem with one of the parameters passed to the function
sncr_fileerror=2            # There was a problem with file access (open, read, write, etc)
sncr_parseerror=3           # There was a problem parsing data
sncr_allocerror=4           # There was a problem allocating memory
sncr_resolverror=5          # There was a problem with libresolv
sncr_execerror=6            # There was a problem with executing a command
sncr_syscallerror=7         # There was a problem with a system call
sncr_notfound=8             # An expected item was not found (ie when parsing a file)
sncr_invalid_commandline=9  # A command line argument was invalid
sncr_string_too_long=10     # A string was too long


# Run common tests (command without parameters, command --help, command --version)
commontests() {
  CMD=$1

  $CMD
  if [ "$?" != "$sncr_invalid_commandline" ]; then echo "$CMD: Unexpected return code to no arguments"; exit 1; fi

  $CMD --help
  if [ "$?" != "$sncr_success" ]; then echo "$CMD: Unexpected return code to help"; exit 1; fi

  $CMD --version
  if [ "$?" != "$sncr_success" ]; then echo "$CMD: Unexpected return code to version"; exit 1; fi
}

# Set the library configuration ($LSNCCONFTMP) network reconfigure cmd and hostname reconfigure cmd
set_libconf_net_host_reconf() {
  net_reconf=$1
  hn_reconf=$2

  sed -r \
    -e "s/^(snc_net_reconfigure_cmd=)\"?[^\"]*\"?$/\1\"$1\"/" \
    -e "s/^(snc_hostname_reconfigure_cmd=)\"?[^\"]*\"?$/\1\"$2\"/" \
    $LSNCCONF > $LSNCCONFTMP
  if [ "$?" != "0" ]; then echo "Setting library conf network reconfigure and hostname reconfigure cmd failed"; exit 1; fi

#    -e 's/^(snc_net_reconfigure_cmd=)"?[^"]*"?$/\1".\/nrscript testparamnr"/' \
#    -e 's/^(snc_hostname_reconfigure_cmd=)"?[^"]*"?$/\1".\/hrscript testparamhr"/' \
}

get_detailed_value()
{
  resp=$1
  varname=$2
  value=`echo "$resp" | grep "$2:" | sed -n -r "s/^$2:\s*(\S*)\s*/\1/p"`
  echo $value
}


# Check the detailed response from getnetconfig/getnetcurrent:
# Use DHCP:    yes
# IP Address:  0.0.0.0
# Netmask:     0.0.0.0
# Broadcast:   0.0.0.0
# Gateway:     0.0.0.0
# DNS 0:       0.0.0.0
# DNS 1:       0.0.0.0
# DNS 2:       0.0.0.0
# Search:
# Hostname:    nobbin1804
# Domain:      rinstrumau.local
# MAC Address: 00:00:00:00:00:00
gnc_check_detailed() {
  cmd=$GNC
  cfg=$1

  ex_use_dhcp=$2
  ex_ip=$3
  ex_nm=$4
  ex_bc=$5
  ex_gw=$6
  ex_ns0=$7
  ex_ns1=$8
  ex_ns2=$9
  ex_sd=${10}
  ex_hn=${11}
  ex_dn=${12}
  ex_mac=${13}

  detailed=`$cmd --config $cfg -m rcconf --detailed $RCIFACE`
  if [ "$?" != "$sncr_success" ]; then echo "$cmd: Failed"; exit 1; fi

  ac_use_dhcp=`get_detailed_value "$detailed" "Use DHCP"`
  ac_ip=`get_detailed_value "$detailed" "IP Address"`
  ac_nm=`get_detailed_value "$detailed" "Netmask"`
  ac_bc=`get_detailed_value "$detailed" "Broadcast"`
  ac_gw=`get_detailed_value "$detailed" "Gateway"`
  ac_ns0=`get_detailed_value "$detailed" "DNS 0"`
  ac_ns1=`get_detailed_value "$detailed" "DNS 1"`
  ac_ns2=`get_detailed_value "$detailed" "DNS 2"`
  ac_sd=`get_detailed_value "$detailed" "Search"`
  ac_hn=`get_detailed_value "$detailed" "Hostname"`
  ac_dn=`get_detailed_value "$detailed" "Domain"`
  ac_mac=`get_detailed_value "$detailed" "MAC Address"`

  if [ -n "$ex_use_dhcp" -a "$ex_use_dhcp" != "$ac_use_dhcp" ]; then echo "$cmd: Use DHCP actual didn't match expected"; exit 1; fi
  if [ -n "$ex_ip"       -a "$ex_ip"       != "$ac_ip"       ]; then echo "$cmd: IP Address actual didn't match expected"; exit 1; fi
  if [ -n "$ex_nm"       -a "$ex_nm"       != "$ac_nm"       ]; then echo "$cmd: Netmask actual didn't match expected"; exit 1; fi
  if [ -n "$ex_bc"       -a "$ex_bc"       != "$ac_bc"       ]; then echo "$cmd: Broadcast actual didn't match expected"; exit 1; fi
  if [ -n "$ex_gw"       -a "$ex_gw"       != "$ac_gw"       ]; then echo "$cmd: Gateway actual didn't match expected"; exit 1; fi
  if [ -n "$ex_ns0"      -a "$ex_ns0"      != "$ac_ns0"      ]; then echo "$cmd: Nameserver 0 actual didn't match expected"; exit 1; fi
  if [ -n "$ex_ns1"      -a "$ex_ns1"      != "$ac_ns1"      ]; then echo "$cmd: Nameserver 1 actual didn't match expected"; exit 1; fi
  if [ -n "$ex_ns2"      -a "$ex_ns2"      != "$ac_ns2"      ]; then echo "$cmd: Nameserver 2 actual didn't match expected"; exit 1; fi
  if [ -n "$ex_sd"       -a "$ex_sd"       != "$ac_sd"       ]; then echo "$cmd: Search domain actual didn't match expected"; exit 1; fi
  if [ -n "$ex_hn"       -a "$ex_hn"       != "$ac_hn"       ]; then echo "$cmd: Hostname actual didn't match expected"; exit 1; fi
  if [ -n "$ex_dn"       -a "$ex_dn"       != "$ac_dn"       ]; then echo "$cmd: Domain name actual didn't match expected"; exit 1; fi
  if [ -n "$ex_mac"      -a "$ex_mac"      != "$ac_mac"      ]; then echo "$cmd: MAC Address actual didn't match expected"; exit 1; fi
}


#
# Test getnetcurrent
#
commontests $GNCU

# Default method (auto)
$GNCU --config $LSNCCONF $IFACE
if [ "$?" != "$sncr_success" ]; then echo "$GNCU: Unexpected return code to valid interface"; exit 1; fi
for m in $GNCU_METHODS
do
  $GNCU --config $LSNCCONF -m $m $IFACE
  if [ "$?" != "$sncr_success" ]; then echo "$GNCU: Unexpected return code to valid interface using method $m"; exit 1; fi
done

#
# Test getnetconfig (rcconf)
#
commontests $GNC

$GNC --config $LSNCCONF -m nm $IFACE
if [ "$?" != "$sncr_success" ]; then echo "$GNC: Get net config failed"; exit 1; fi
for m in $GNC_METHODS
do
  $GNC --config $LSNCCONF -m $m $RCIFACE
  if [ "$?" != "$sncr_success" ]; then echo "$GNC: Unexpected return code to valid interface using method $m"; exit 1; fi
done

#
# Test setnetconfig (rcconf only, since we don't want to play with the Network Manager configuration and break our network)
#
commontests $SNC

# Set static
expect="172.17.1.20 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2 00:00:00:00:00:00 host-name"
$SNC --config $LSNCCONF -m rcconf $RCIFACE 172.17.1.20 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2
if [ "$?" != "$sncr_success" ]; then echo "$SNC: Set net config static failed"; exit 1; fi
cfg=`$GNC --config $LSNCCONF -m rcconf $RCIFACE`
if [ "$?" != "$sncr_success" ]; then echo "$GNC: Get net config static failed"; exit 1; fi
if [ "$cfg" != "$expect" ]; then echo "$GNC: static: Expected didn't match actual"; exit 1; fi

# Set DHCP
expect="dhcp 0.0.0.0 0.0.0.0 0.0.0.0 0.0.0.0 00:00:00:00:00:00 host-name"
$SNC --config $LSNCCONF -m rcconf $RCIFACE dhcp
if [ "$?" != "$sncr_success" ]; then echo "$SNC: Set net config dhcp failed"; exit 1; fi
cfg=`$GNC --config $LSNCCONF -m rcconf $RCIFACE`
if [ "$?" != "$sncr_success" ]; then echo "$GNC: Get net config dhcp failed"; exit 1; fi
if [ "$cfg" != "$expect" ]; then echo "$GNC: dhcp: Expected didn't match actual"; exit 1; fi

# Check running network reconfigure and hostname reconfigure commands
sed -r \
  -e 's/^(snc_net_reconfigure_cmd=)"?[^"]*"?$/\1".\/nrscript testparamnr"/' \
  -e 's/^(snc_hostname_reconfigure_cmd=)"?[^"]*"?$/\1".\/hrscript testparamhr"/' \
  $LSNCCONF > $LSNCCONFTMP
if [ "$?" != "$sncr_success" ]; then echo "$SNC: Set net config dhcp failed"; exit 1; fi

# Run network reconfigure script
cp $RCCONF $RCCONFTMP
rm -f nrresult.txt hrresult.txt
expect="dhcp 0.0.0.0 0.0.0.0 0.0.0.0 0.0.0.0 00:00:00:00:00:00 host-name"
$SNC --config $LSNCCONFTMP -m rcconf $RCIFACE dhcp
if [ "$?" != "$sncr_success" ]; then echo "$SNC: Set net config dhcp failed"; exit 1; fi
nrres="`cat nrresult.txt`"
if [ "$nrres" != "testparamnr" ]; then echo "$GNC: dhcp: Net restart script didn't work as expected"; exit 1; fi
if [ -f hrresult.txt ]; then echo "$SNC: Set net config ran hostname reconfigure script"; exit 1; fi

# Check bad parameters
$SNC --config $LSNCCONFTMP -m rcconf $RCIFACE 172.17.1.20 255.255.0.0 172.17.255.255 172.17.1.1
if [ "$?" != "$sncr_invalid_commandline" ]; then echo "$SNC: Missing 1 parameter failed"; exit 1; fi
$SNC --config $LSNCCONFTMP -m rcconf $RCIFACE 172.17.1.20
if [ "$?" != "$sncr_invalid_commandline" ]; then echo "$SNC: Missing 4 parameters failed"; exit 1; fi
$SNC --config $LSNCCONFTMP -m rcconf $RCIFACE abc.def.ghi.jkl 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2
if [ "$?" != "$sncr_parseerror" ]; then echo "$SNC: Invalid IP address failed"; exit 1; fi
$SNC --config $LSNCCONFTMP -m rcconf $RCIFACE 172.17.1 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2
if [ "$?" != "$sncr_parseerror" ]; then echo "$SNC: Missing IP octet failed"; exit 1; fi

# Check setting hostname
cp $RCCONF $RCCONFTMP
rm -f nrresult.txt hrresult.txt
expect="192.168.0.1 255.255.255.0 192.168.0.255 192.168.0.2 192.168.0.3 00:00:00:00:00:00 new-host-name"
$SNC --config $LSNCCONFTMP -m rcconf --hostname new-host-name
if [ "$?" != "$sncr_success" ]; then echo "$SNC: Setting hostname failed"; exit 1; fi
cfg=`$GNC --config $LSNCCONFTMP -m rcconf $RCIFACE`
if [ "$?" != "$sncr_success" ]; then echo "$GNC: Get hostname failed"; exit 1; fi
if [ "$cfg" != "$expect" ]; then echo "$GNC: hostname: Expected didn't match actual"; exit 1; fi
if [ -f nrresult.txt ]; then echo "$SNC: --hostname ran network reconfigure script"; exit 1; fi
hrres="`cat hrresult.txt`"
if [ "$hrres" != "testparamhr" ]; then echo "$SNC: --hostname: Hostname restart script didn't work as expected"; exit 1; fi

# Check setting hostname and dhcp
cp $RCCONF $RCCONFTMP
rm -f nrresult.txt hrresult.txt
expect="dhcp 0.0.0.0 0.0.0.0 0.0.0.0 0.0.0.0 00:00:00:00:00:00 new-host-name"
$SNC --config $LSNCCONFTMP -m rcconf --hostname new-host-name $RCIFACE dhcp
if [ "$?" != "$sncr_success" ]; then echo "$SNC: Setting hostname failed"; exit 1; fi
cfg=`$GNC --config $LSNCCONFTMP -m rcconf $RCIFACE`
if [ "$?" != "$sncr_success" ]; then echo "$GNC: Get hostname failed"; exit 1; fi
if [ "$cfg" != "$expect" ]; then echo "$GNC: hostname: Expected didn't match actual"; exit 1; fi
nrres="`cat nrresult.txt`"
if [ "$nrres" != "testparamnr" ]; then echo "$SNC: --hostname & dhcp: Network restart script didn't work as expected"; exit 1; fi
hrres="`cat hrresult.txt`"
if [ "$hrres" != "testparamhr" ]; then echo "$SNC: --hostname & dhcp: Hostname restart script didn't work as expected"; exit 1; fi

# Check setting hostname and static
cp $RCCONF $RCCONFTMP
rm -f nrresult.txt hrresult.txt
expect="172.17.1.20 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2 00:00:00:00:00:00 new-host-name"
$SNC --config $LSNCCONFTMP -m rcconf --hostname new-host-name $RCIFACE 172.17.1.20 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2
if [ "$?" != "$sncr_success" ]; then echo "$SNC: Setting hostname failed"; exit 1; fi
cfg=`$GNC --config $LSNCCONFTMP -m rcconf $RCIFACE`
if [ "$?" != "$sncr_success" ]; then echo "$GNC: Get hostname failed"; exit 1; fi
if [ "$cfg" != "$expect" ]; then echo "$GNC: hostname: Expected didn't match actual"; exit 1; fi
nrres="`cat nrresult.txt`"
if [ "$nrres" != "testparamnr" ]; then echo "$SNC: --hostname & static: Network restart script didn't work as expected"; exit 1; fi
hrres="`cat hrresult.txt`"
if [ "$hrres" != "testparamhr" ]; then echo "$SNC: --hostname & static: Hostname restart script didn't work as expected"; exit 1; fi

# Check that running setnetconfig --help & setnetconfig --version don't run the network reconfigure script
rm -f nrresult.txt
$SNC --config $LSNCCONFTMP -m rcconf --help
if [ -f nrresult.txt ]; then echo "$SNC: --help ran network reconfigure script"; exit 1; fi
rm -f nrresult.txt
$SNC --config $LSNCCONFTMP -m rcconf --version
if [ -f nrresult.txt ]; then echo "$SNC: --version ran network reconfigure script"; exit 1; fi

# Check permissions
cp $RCCONF $RCCONFTMP
umask 0
$SNC --config $LSNCCONF -m rcconf $RCIFACE 172.17.1.20 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2
PERMS="`stat --format='%a' $RCCONFTMP`"
if [ "644" != "$PERMS" ]; then echo "Incorrect permissions on $RCCONFTMP"; exit 1; fi
cp $RCCONF $RCCONFTMP
umask 0777
$SNC --config $LSNCCONF -m rcconf $RCIFACE 172.17.1.20 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2
PERMS="`stat --format='%a' $RCCONFTMP`"
if [ "644" != "$PERMS" ]; then echo "Incorrect permissions on $RCCONFTMP"; exit 1; fi

# Check problem where NETMASK0="" & so on, getnetconfig would return blanks
sed -r -e 's/export IPADDR0=.*/export IPADDR0=""/' \
       -e 's/export NETMASK0=.*/export NETMASK0=""/' \
       -e 's/export BROADCAST0=.*/export BROADCAST0=""/' \
       -e 's/export GATEWAY0=.*/export GATEWAY0=""/' \
       -e 's/export NAMESERVER0=.*/export NAMESERVER0=""/' $RCCONF > $RCCONFTMP
expect="0.0.0.0 0.0.0.0 0.0.0.0 0.0.0.0 0.0.0.0 00:00:00:00:00:00 host-name"
cfg=`$GNC --config $LSNCCONF -m rcconf $RCIFACE`
if [ "$?" != "$sncr_success" ]; then echo "$GNC: Check blank rcconf variable failed"; exit 1; fi
if [ "$cfg" != "$expect" ]; then echo "$GNC: Check blank rcconf variable didn't match"; exit 1; fi

# Check problem where setnetconfig would set NETMASK0="", etc
cp $RCCONF $RCCONFTMP
$SNC --config $LSNCCONF -m rcconf $RCIFACE dhcp
$SNC --config $LSNCCONF -m rcconf --hostname new-host-name $RCIFACE dhcp
grep -q -E -e 'export NETMASK0=""' -e 'export BROADCAST0=""' -e 'export GATEWAY0=""' -e 'export NAMESERVER0=""' $RCCONFTMP
if [ "$?" == "0" ]; then echo "$SNC: Setting NETMASK0=\"\", etc is a failure"; exit 1; fi

# Check setting hostname with invalid characters and invalid length
cp $RCCONF $RCCONFTMP
$SNC --config $LSNCCONF -m rcconf --hostname ""
if [ "$?" != "$sncr_parseerror" ]; then echo "$SNC: Setting empty hostname is a failure"; exit 1; fi
$SNC --config $LSNCCONF -m rcconf --hostname 01234567890123456789012345678901234567890123456789012345678901234
if [ "$?" != "$sncr_parseerror" ]; then echo "$SNC: Setting 65 characters hostname is a failure"; exit 1; fi
$SNC --config $LSNCCONF -m rcconf --hostname 'asdf!^'
if [ "$?" != "$sncr_parseerror" ]; then echo "$SNC: Setting hostname invalid characters !# is a failure"; exit 1; fi

# getnetcurrent - Test valid and no default gateway from route and ip commands
sed -r \
  -e 's/^(snc_routecmd=)"?[^"]*"?$/\1".\/fake_route_normal"/' \
  -e 's/^(snc_ipcmd=)"?[^"]*"?$/\1".\/fake_ip_normal"/' \
  -e 's/^(snc_ifconfigcmd=)"?[^"]*"?$/\1".\/fake_ifconfig_normal"/' \
  $LSNCCONF > $LSNCCONFTMP
if [ "$?" != "$sncr_success" ]; then echo "$GNCU: Test valid default gateway failed"; exit 1; fi
sed -r \
  -e 's/^(snc_routecmd=)"?[^"]*"?$/\1".\/fake_route_nogw"/' \
  -e 's/^(snc_ipcmd=)"?[^"]*"?$/\1".\/fake_ip_nogw"/' \
  -e 's/^(snc_ifconfigcmd=)"?[^"]*"?$/\1".\/fake_ifconfig_normal"/' \
  $LSNCCONF > $LSNCCONFTMP2
if [ "$?" != "$sncr_success" ]; then echo "$GNCU: Test no default gateway failed"; exit 1; fi

for m in ifconfig ip
do
  expect="Gateway:     172.17.1.1"
  cfg=`$GNCU --config $LSNCCONFTMP -m $m $RCIFACE --detailed | grep "Gateway:"`
  if [ "$?" != "$sncr_success" ]; then echo "$GNCU: Test valid default gateway via $m failed"; exit 1; fi
  if [ "$cfg" != "$expect" ]; then echo "$GNCU: Test valid default gateway via $m didn't match"; exit 1; fi

  expect="Gateway:     0.0.0.0"
  cfg=`$GNCU --config $LSNCCONFTMP2 -m $m $RCIFACE --detailed | grep "Gateway:"`
  if [ "$?" != "$sncr_success" ]; then echo "$GNCU: Test no default gateway via $m failed"; exit 1; fi
  if [ "$cfg" != "$expect" ]; then echo "$GNCU: Test no default gateway via $m didn't match"; exit 1; fi
done


#
# Test setnetconfigx (rcconf only, since we don't want to play with the Network Manager configuration and break our network)
#
commontests $SNCX

# Check running network reconfigure and hostname reconfigure commands
set_libconf_net_host_reconf ".\/nrscript testparamnr" ".\/hrscript testparamhr"

echo "getnetconfig: Checking against unmodified rc.conf"
cp $RCCONF $RCCONFTMP
#                            dhcp ip          nm            bc            gw          ns0         ns1     ns2     sd            hn        dn          mac
gnc_check_detailed $LSNCCONF no   192.168.0.1 255.255.255.0 192.168.0.255 192.168.0.2 192.168.0.3 0.0.0.0 0.0.0.0 search.domain host-name domain.name 00:00:00:00:00:00

echo "setnetconfigx: Setting static ip"
$SNCX --config $LSNCCONF -m rcconf if $RCIFACE ip 172.17.1.20 nm 255.255.0.0 bc 172.17.255.255 gw 172.17.1.1 ns1 172.17.1.2
if [ "$?" != "$sncr_success" ]; then echo "$SNCX: Set net config static failed"; exit 1; fi

#                            dhcp ip          nm          bc             gw         ns0        ns1     ns2     sd hn        dn          mac
gnc_check_detailed $LSNCCONF no   172.17.1.20 255.255.0.0 172.17.255.255 172.17.1.1 172.17.1.2 0.0.0.0 0.0.0.0 "" host-name domain.name 00:00:00:00:00:00

echo "setnetconfigx: Setting static ip (without optional parameters)"
cp $RCCONF $RCCONFTMP
$SNCX --config $LSNCCONF -m rcconf if $RCIFACE ip 172.17.1.20 nm 255.255.0.0
if [ "$?" != "$sncr_success" ]; then echo "$SNCX: Set net config static no optional parameters"; exit 1; fi

#                            dhcp ip          nm          bc             gw      ns0     ns1     ns2     sd hn        dn          mac
gnc_check_detailed $LSNCCONF no   172.17.1.20 255.255.0.0 172.17.255.255 0.0.0.0 0.0.0.0 0.0.0.0 0.0.0.0 "" host-name domain.name 00:00:00:00:00:00

echo "setnetconfigx: Setting static ip (without required parameters)"
cp $RCCONF $RCCONFTMP
$SNCX --config $LSNCCONF -m rcconf if $RCIFACE ip 172.17.1.20
if [ "$?" != "$sncr_invalid_commandline" ]; then echo "$SNCX: Set net config static without required parameter netmask"; exit 1; fi
$SNCX --config $LSNCCONF -m rcconf if $RCIFACE nm 255.255.255.0
if [ "$?" != "$sncr_invalid_commandline" ]; then echo "$SNCX: Set net config static without required parameter ip address"; exit 1; fi
$SNCX --config $LSNCCONF -m rcconf ip 172.17.1.20 nm 255.255.255.0
if [ "$?" != "$sncr_invalid_commandline" ]; then echo "$SNCX: Set net config static without required parameter interface"; exit 1; fi
$SNCX --config $LSNCCONF -m rcconf ns0 172.17.1.2
if [ "$?" != "$sncr_invalid_commandline" ]; then echo "$SNCX: Set net config static without required parameter interface"; exit 1; fi

echo "setnetconfigx: Check setting hostname"
cp $RCCONF $RCCONFTMP
rm -f nrresult.txt hrresult.txt
$SNCX --config $LSNCCONFTMP -m rcconf hn new-host-name
if [ "$?" != "$sncr_success" ]; then echo "$SNCX: Setting hostname failed"; exit 1; fi

#                            dhcp ip          nm            bc            gw          ns0         ns1     ns2     sd            hn            dn          mac
gnc_check_detailed $LSNCCONF no   192.168.0.1 255.255.255.0 192.168.0.255 192.168.0.2 192.168.0.3 0.0.0.0 0.0.0.0 search.domain new-host-name domain.name 00:00:00:00:00:00
if [ -f nrresult.txt ]; then echo "$SNCX: hn ran network reconfigure script"; exit 1; fi
hrres="`cat hrresult.txt`"
if [ "$hrres" != "testparamhr" ]; then echo "$SNCX hn: Hostname restart script didn't work as expected"; exit 1; fi

echo "setnetconfigx: Check setting domain name"
cp $RCCONF $RCCONFTMP
rm -f nrresult.txt hrresult.txt
$SNCX --config $LSNCCONFTMP -m rcconf dn new.domain.name
if [ "$?" != "$sncr_success" ]; then echo "$SNCX: Setting domain name failed"; exit 1; fi

#                            dhcp ip          nm            bc            gw          ns0         ns1     ns2     sd            hn        dn              mac
gnc_check_detailed $LSNCCONF no   192.168.0.1 255.255.255.0 192.168.0.255 192.168.0.2 192.168.0.3 0.0.0.0 0.0.0.0 search.domain host-name new.domain.name 00:00:00:00:00:00
if [ -f hrresult.txt ]; then echo "$SNCX: dn ran hostname reconfigure script"; exit 1; fi
nrres="`cat nrresult.txt`"
if [ "$nrres" != "testparamnr" ]; then echo "$SNCX dn: Network restart script didn't work as expected"; exit 1; fi


# Clean up
rm -f nrresult.txt hrresult.txt
rm -f $RCCONFTMP $LSNCCONFTMP $LSNCCONFTMP2

echo "All tests passed successfully"
exit 0
