#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <setjmp.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <rincmd/rc_parser.h>
#include <rincmd/rc_public.h>
#include <rincmd/rc_registers.h>
#include "ipinfo.h"
#include "commonstrings.h"
#include "db_remote.h"
#include "db_clock.h"
#include "db_serial.h"
#include "db_option.h"
#include "app_db_user.h"
#include "apperror.h"
#include "we_config.h"


#define APP_DB_U_ADCZERO_DUVV_MIN ADC_RAW_TO_DUVV(APP_DB_U_ADCZERORAW_MIN)
#define APP_DB_U_ADCZERO_DUVV_MAX ADC_RAW_TO_DUVV(APP_DB_U_ADCZERORAW_MAX)
#define APP_DB_U_ADCSPAN_DUVV_MIN ADC_RAW_TO_DUVV(APP_DB_U_ADCSPANRAW_MIN)
#define APP_DB_U_ADCSPAN_DUVV_MAX ADC_RAW_TO_DUVV(APP_DB_U_ADCSPANRAW_MAX)

#define APP_DB_U_ADCZERO_MVV_MIN ADC_RAW_TO_MVV((float) APP_DB_U_ADCZERORAW_MIN)
#define APP_DB_U_ADCZERO_MVV_MAX ADC_RAW_TO_MVV((float) APP_DB_U_ADCZERORAW_MAX)
#define APP_DB_U_ADCSPAN_MVV_MIN ADC_RAW_TO_MVV((float) APP_DB_U_ADCSPANRAW_MIN)
#define APP_DB_U_ADCSPAN_MVV_MAX ADC_RAW_TO_MVV((float) APP_DB_U_ADCSPANRAW_MAX)


/* Setting string lists */

/** Off/on setting strings */
const char *const off_on_strings[] = OFF_ON_STRS;


/** Serial port type setting (auto out, net, etc) */
const char *const ser_type_setting[NUM_DB_SERIAL_TYPES] = DB_SERIAL_TYPE_STRINGS;


const char *const baud_setting[NUM_DB_BAUD_SETTINGS] = DB_BAUD_STRINGS;


/** Remote function keys functions string list. Matches @ref enum eRemoteKeyFn */
const char *const remote_keyfn_strings[NUM_REMOTE_KEY_FUNCTIONS] = REMOTE_KEYFN_OPT_STRS;


/** Front function keys function characters */
const char *const frontfunc_keyfn_strings[NUM_FRONTFUNC] = FRONTFUNC_KEYFN_OPT_STRS;


/** Time format options */
const char *const time_format_settings[NUM_TIME_FMTS] = TIME_FORMAT_STRS;


/** Date format options */
const char *const date_format_settings[NUM_DATE_FMTS] = DATE_FORMAT_STRS;


/** Setpoint active name strings */
const char *const setpoint_type_strings[SETP_TYPE_NUM] = SETPOINT_ACTIVE_STRS;


/** Setpoint feeder options */
const char *const setp_feeder_strings[SETP_FEEDER_NUM] = SETP_FEEDER_OPTION_STRS;


/** Automatic output formats */
const char *const auto_setting[AUTO_OUT_FMT_NUM] = AUTO_OUT_FMT_STRS;


/** Printer interlock setting strings */
const char *const print_interlock_setting[NUM_PRINT_INTERLOCK] = PRINT_INTERLOCK_STRS;


/** Units strings used in the menus */
const char *const build_units_menus[NUM_UNITS] = BUILD_UNITS_MENU_STRS;


/** Alternate units strings used in the menus */
const char *const alt_units_menus[NUM_ALT_UNITS] = ALT_UNITS_MENU_STRS;


/** WE2111 approval mode strings */
const char *const we_approval_mode_strings[WE_APPROVAL_MODE_NUM] = WE_APPROVAL_MODE_STRS;


/** Setpoint timing options */
const char *const setp_timing_strings[SETP_TIMING_NUM] = SETP_TIMING_OPTION_STRS;


/** Setpoint reset options */
const char *const setp_reset_strings[SETP_RESET_NUM] = SETP_RESET_OPTION_STRS;


/** Create a latex comment */
#define LATEX_COMMENT(format, args...) \
do { \
  printf("%% " format "\n", ##args); \
} while (0)


#define LATEX_NEWCMD "\\newcommand{\\%s}"
#define LATEX_NEWCMD_PARAMS "\\newcommand{\\%s}[1]"
#define LATEX_CVAR "cvar"
#define LATEX_CSTRLEN "cstrlen"
#define LATEX_CLIST "clist"


/** Declare a C define as a latex integer */
#define DECLARE_LATEX_INT(cdef) \
do { \
  printf(LATEX_NEWCMD "{%lld}\n", cname_to_texname(#cdef, LATEX_CVAR), (long long) cdef); \
} while (0)


/**
 * Declare a C define of a maximum string length as a latex integer, leaving off the null termination character
 *
 * For example:
 *   #define APP_DB_U_SERID_MAXLEN 16
 *   DECLARE_LATEX_CSTR_LENGTH(APP_DB_U_SERID_MAXLEN)
 * gives
 *   \newcommand{\cvarAppDbUSeridMaxlen}{15}
 */
#define DECLARE_LATEX_CSTR_LENGTH(cdef) \
do { \
  printf(LATEX_NEWCMD "{%lld}\n", cname_to_texname(#cdef, LATEX_CSTRLEN), (long long) cdef - 1); \
} while (0)


/** Declare a C define as a latex float */
#define DECLARE_LATEX_FLOAT(cdef, prec) \
do { \
  printf(LATEX_NEWCMD "{%.*f}\n", cname_to_texname(#cdef, LATEX_CVAR), prec, cdef); \
} while (0)


/** Declare a C define as a latex string */
#define DECLARE_LATEX_STR(cdef) \
do { \
  printf(LATEX_NEWCMD "{%s}\n", cname_to_texname(#cdef, LATEX_CVAR), cdef); \
} while (0)


#define DECLARE_LATEX_IP_NET_ENDIAN(cdef) \
do { \
  char ip_str[INET_ADDRSTRLEN]; \
  struct in_addr ip; \
  ip.s_addr = cdef; \
  printf(LATEX_NEWCMD "{%s}\n", cname_to_texname(#cdef, LATEX_CVAR), inet_ntop(AF_INET, &ip, ip_str, INET_ADDRSTRLEN)); \
} while (0)


#define DECLARE_LATEX_IP_HOST_ENDIAN(cdef) \
do { \
  char ip_str[INET_ADDRSTRLEN]; \
  struct in_addr ip; \
  ip.s_addr = htonl(cdef); \
  printf(LATEX_NEWCMD "{%s}\n", cname_to_texname(#cdef, LATEX_CVAR), inet_ntop(AF_INET, &ip, ip_str, INET_ADDRSTRLEN)); \
} while (0)


/** Create an latex list */
#define DECLARE_LATEX_LIST(texlisttype, cdef, cstrlist, cstrlistlen, defaultelem, with_descriptions, offset) \
do { \
  if (with_descriptions) \
    printf(LATEX_NEWCMD_PARAMS, cname_to_texname(#cdef, LATEX_CLIST)); \
  else \
    printf(LATEX_NEWCMD, cname_to_texname(#cdef, LATEX_CLIST)); \
  printf("{%s}\n", cstrlist_to_tex_list(texlisttype, cstrlist, cstrlistlen, defaultelem, with_descriptions, offset)); \
} while (0)


/** Create a latex list of an OFF/ON setting */
#define DECLARE_LATEX_OFF_ON(list_type, cdef, defbool) \
  DECLARE_LATEX_LIST(list_type, cdef, off_on_strings, 2, (defbool) ? 1 : 0, false, 0)

/** Create an simple latex list */
#define DECLARE_LATEX_LIST_SIMPLE(list_type, cstrlist, cstrlistlen, defaultelem, offset) \
  DECLARE_LATEX_LIST(list_type, cstrlist, cstrlist, cstrlistlen, defaultelem, false, offset)

/** Create an simple latex list with specified name cdef */
#define DECLARE_LATEX_LIST_NAMED_SIMPLE(list_type, cdef, cstrlist, cstrlistlen, defaultelem, offset) \
  DECLARE_LATEX_LIST(list_type, cdef, cstrlist, cstrlistlen, defaultelem, false, offset)

/** Create an latex list with descriptions */
#define DECLARE_LATEX_LIST_DESCRIPTION(list_type, cstrlist, cstrlistlen, defaultelem, offset) \
  DECLARE_LATEX_LIST(list_type, cstrlist, cstrlist, cstrlistlen, defaultelem, true, offset)

/** Create an latex list with descriptions with specified name cdef */
#define DECLARE_LATEX_LIST_NAMED_DESCRIPTION(list_type, cdef, cstrlist, cstrlistlen, defaultelem, offset) \
  DECLARE_LATEX_LIST(list_type, cdef, cstrlist, cstrlistlen, defaultelem, true, offset)


/** Latex can't have digits in a command name, so translate 0->Zero, 1->One, and so on */
static const char *const digits[10] =
{
  "Zero", "One", "Two", "Three", "Four", "Five", "Six", "Seven", "Eight", "Nine"
};


#define EPIC_BUFFER_SIZE (10 * 1024)
static char *epic_buffer = NULL;
static size_t epic_buffer_pos = 0;
static jmp_buf epic_buffer_fail;


/** Clear the epic buffer */
#define EPIC_BUFFER_RESET() \
do { \
  epic_buffer[0] = '\0'; \
  epic_buffer_pos = 0; \
} while (0)


/** Print into the epic buffer */
#define EPIC_BUFFER_PRINTF(format, args...) \
do { \
  size_t bufleft = EPIC_BUFFER_SIZE - epic_buffer_pos; \
  int snres = snprintf(&epic_buffer[epic_buffer_pos], bufleft, format, ##args); \
  if (snres < 0 || snres >= bufleft) \
    longjmp(epic_buffer_fail, 1); \
  epic_buffer_pos += snres; \
} while (0)


/** Convert a C variable name into a latex name (ie DEFAULT_ETH_IPADDRESS -> cvarDefaultEthIpaddress) */
static const char *cname_to_texname(const char *cname, const char *prefix)
{
  const char *inptr;
  bool firstletter;

  EPIC_BUFFER_RESET();
  EPIC_BUFFER_PRINTF("%s", prefix);

  for (inptr = cname, firstletter = true; '\0' != *inptr; inptr++)
  {
    switch (*inptr)
    {
      case '_':
        firstletter = true;
        break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        EPIC_BUFFER_PRINTF("%s", digits[*inptr - '0']);
        firstletter = false;
        break;
      default:
        EPIC_BUFFER_PRINTF("%c", (firstletter) ? toupper(*inptr) : tolower(*inptr));
        firstletter = false;
        break;
    }
  }
  return epic_buffer;
}


#define TEX_EASYLIST_NO_DEFAULT (-1)
#define NO_OFFSET (0)

enum e_tex_list_type
{
  tlt_list_itemize,       /*!< Regular \begin{itemize} bulleted list */
  tlt_list_enumerate,     /*!< Regular \begin{enumerate} numbered list */
  tlt_easylist_itemize,   /*!< Easylist \begin{easylist}[itemize] bulleted list */
  tlt_easylist_enumerate, /*!< Easylist \begin{easylist}[enumerate] numbered list */
  tlt_comms_list,         /*!< Used in a lot of comms command documentation: no header, footer, and idx = value\newline eg 0 = Value1\newline, 1 = Value2\newline */
};


/** Start a latex list */
static void tex_list_start(enum e_tex_list_type list_type)
{
  switch (list_type)
  {
    case tlt_list_itemize:
      EPIC_BUFFER_PRINTF("\\begin{itemize}\n");
      break;

    case tlt_list_enumerate:
      EPIC_BUFFER_PRINTF("\\begin{enumerate}\n");
      break;

    case tlt_easylist_itemize:
      EPIC_BUFFER_PRINTF("\\begin{easylist}[itemize]\n");
      break;

    case tlt_easylist_enumerate:
      EPIC_BUFFER_PRINTF("\\begin{easylist}[enumerate]\n");
      break;

    case tlt_comms_list:
    default:
      break;
  }
}


/** End a latex list */
static void tex_list_end(enum e_tex_list_type list_type)
{
  switch (list_type)
  {
    case tlt_list_itemize:
      EPIC_BUFFER_PRINTF("\\end{itemize}");
      break;

    case tlt_list_enumerate:
      EPIC_BUFFER_PRINTF("\\end{enumerate}");
      break;

    case tlt_easylist_itemize:
      EPIC_BUFFER_PRINTF("\\end{easylist}");
      break;

    case tlt_easylist_enumerate:
      EPIC_BUFFER_PRINTF("\\end{easylist}");
      break;

    case tlt_comms_list:
    default:
      break;
  }
}


/** Convert a c string list into a tex list/easylist, marking the default item with "(default)" */
static const char *cstrlist_to_tex_list(enum e_tex_list_type list_type, const char *const list[], size_t nelms, ssize_t default_elem, bool with_descriptions, unsigned offset)
{
  #define EASYLISTLEN 4096
  size_t i;
  const char *default_str;
  const char *item_marker = (list_type == tlt_list_itemize || list_type == tlt_list_enumerate) ? "\\item" : "&";
  const char *newline;

  EPIC_BUFFER_RESET();
  if (with_descriptions)
    EPIC_BUFFER_PRINTF("\\setsepchar{,}\n\\readlist*\\arg{#1}\n");

  tex_list_start(list_type);
  for (i = 0; i < nelms; i++)
  {
    default_str = (i == default_elem) ? " (default)" : "";
    if (tlt_comms_list != list_type)
    {
      if (with_descriptions)
        EPIC_BUFFER_PRINTF("%s %s: \\arg[%u]%s\n", item_marker, list[i], (unsigned) i + 1, default_str);
      else
        EPIC_BUFFER_PRINTF("%s %s%s\n", item_marker, list[i], default_str);
    }
    else
    {
      newline = (i < nelms - 1) ? "\\newline" : "";
      if (with_descriptions)
        EPIC_BUFFER_PRINTF("%u = \\arg[%u] (%s)%s%s\n", (unsigned) i + offset, (unsigned) i + 1, list[i], default_str, newline);
      else
        EPIC_BUFFER_PRINTF("%u = %s%s%s\n", (unsigned) i + offset, list[i], default_str, newline);
    }
  }
  tex_list_end(list_type);
  return epic_buffer;
}


int main()
{
  int result = 0;

  /* Allocate big buffer */
  epic_buffer = calloc(EPIC_BUFFER_SIZE, sizeof(char));
  if (NULL == epic_buffer)
  {
    fprintf(stderr, "Failed to allocate epic buffer\n");
    result = 1;
    goto mfail;
  }

  /* Failure point */
  if (setjmp(epic_buffer_fail) != 0)
  {
    fprintf(stderr, "Epic buffer full/error\n");
    result = 1;
    goto mfail;
  }

  printf("%% defines.tex\n"
         "%%\n"
         "%% File automatically generated from defines_tex.c\n"
         "%% DO NOT EDIT\n"
         "%%\n");

  LATEX_COMMENT("");
  LATEX_COMMENT("Rincmd message addresses");
  DECLARE_LATEX_INT(RC_MESG_SILENT);
  DECLARE_LATEX_INT(RC_MESG_REPLY);
  DECLARE_LATEX_INT(RC_MESG_ERROR);
  DECLARE_LATEX_INT(RC_MESG_RESPONSE);

  LATEX_COMMENT("Rincmd message commands");
  DECLARE_LATEX_INT(RC_MESG_READ_TYPE);
  DECLARE_LATEX_INT(RC_MESG_READ_RANGE_MIN);
  DECLARE_LATEX_INT(RC_MESG_READ_RANGE_MAX);
  DECLARE_LATEX_INT(RC_MESG_READ_LITERAL);
  DECLARE_LATEX_INT(RC_MESG_READ_PERMISSION);
  DECLARE_LATEX_INT(RC_MESG_EXECUTE);
  DECLARE_LATEX_INT(RC_MESG_READ_HEXADEC);
  DECLARE_LATEX_INT(RC_MESG_WRITE_HEXADEC);
  DECLARE_LATEX_INT(RC_MESG_READ_DECIMAL);
  DECLARE_LATEX_INT(RC_MESG_WRITE_DECIMAL);

  LATEX_COMMENT("Rincmd register types");
  DECLARE_LATEX_INT(RC_REG_TYPE_INT8);
  DECLARE_LATEX_INT(RC_REG_TYPE_UINT8);
  DECLARE_LATEX_INT(RC_REG_TYPE_INT16);
  DECLARE_LATEX_INT(RC_REG_TYPE_UINT16);
  DECLARE_LATEX_INT(RC_REG_TYPE_INT32);
  DECLARE_LATEX_INT(RC_REG_TYPE_UINT32);
  DECLARE_LATEX_INT(RC_REG_TYPE_STRING);
  /* DECLARE_LATEX_INT(RC_REG_TYPE_OPTION); */
  /* DECLARE_LATEX_INT(RC_REG_TYPE_MENU); */
  DECLARE_LATEX_INT(RC_REG_TYPE_WEIGHT);
  /* DECLARE_LATEX_INT(RC_REG_TYPE_BLOB); */
  DECLARE_LATEX_INT(RC_REG_TYPE_EXECUTE);
  /* DECLARE_LATEX_INT(RC_REG_TYPE_BITFIELD); */
  DECLARE_LATEX_INT(RC_REG_TYPE_STREAM);
  /* DECLARE_LATEX_INT(RC_REG_TYPE_EXECUTE_STR); */
  /* DECLARE_LATEX_INT(RC_REG_TYPE_MENU_END); */
  /* DECLARE_LATEX_INT(RC_REG_TYPE_STRING_ARRAY); */
  /* DECLARE_LATEX_INT(RC_REG_TYPE_OPTION16); */
  DECLARE_LATEX_INT(RC_REG_TYPE_IP);
  DECLARE_LATEX_INT(RC_REG_TYPE_REGISTER);
  DECLARE_LATEX_INT(RC_REG_TYPE_BINBUFFER);

  LATEX_COMMENT("Rincmd register permissions");
  DECLARE_LATEX_INT(RC_PERM_RD_N);
  DECLARE_LATEX_INT(RC_PERM_RD_S);
  DECLARE_LATEX_INT(RC_PERM_RD_F);
  DECLARE_LATEX_INT(RC_PERM_RD_A);
  DECLARE_LATEX_INT(RC_PERM_WR_N);
  DECLARE_LATEX_INT(RC_PERM_WR_S);
  DECLARE_LATEX_INT(RC_PERM_WR_F);
  DECLARE_LATEX_INT(RC_PERM_WR_A);

  LATEX_COMMENT("Rincmd major error codes");
  DECLARE_LATEX_INT(RC_HNDL_ERR_NONE);
  DECLARE_LATEX_INT(RC_HNDL_ERR_PARSE);
  DECLARE_LATEX_INT(RC_HNDL_ERR_CRC);
  DECLARE_LATEX_INT(RC_HNDL_ERR_REG);
  DECLARE_LATEX_INT(RC_HNDL_ERR_READ);
  DECLARE_LATEX_INT(RC_HNDL_ERR_WRITE);
  DECLARE_LATEX_INT(RC_HNDL_ERR_EXEC);
  DECLARE_LATEX_INT(RC_HNDL_ERR_NOTIMP);

  LATEX_COMMENT("Rincmd success (RC_HNDL_ERR_NONE) minor errors");
  DECLARE_LATEX_INT(RC_NONE_ERR_NONE);

  LATEX_COMMENT("Rincmd minor parse errors");
  DECLARE_LATEX_INT(RC_PARSE_ERR_NONE);
  DECLARE_LATEX_INT(RC_PARSE_ERR_ADDR);
  DECLARE_LATEX_INT(RC_PARSE_ERR_CMD);
  DECLARE_LATEX_INT(RC_PARSE_ERR_REG);
  DECLARE_LATEX_INT(RC_PARSE_ERR_CHAN);
  DECLARE_LATEX_INT(RC_PARSE_ERR_DELIM);
  DECLARE_LATEX_INT(RC_PARSE_ERR_DATA);
  DECLARE_LATEX_INT(RC_PARSE_ERR_FRAMING);

  DECLARE_LATEX_INT(RC_SIZE_CHAN);
  DECLARE_LATEX_INT(RC_SIZE_BUFFER);

  LATEX_COMMENT("Rincmd CRC minor errors");
  DECLARE_LATEX_INT(RC_CRC_ERR_NONE);
  DECLARE_LATEX_INT(RC_CRC_ERR_PARSE);
  DECLARE_LATEX_INT(RC_CRC_ERR_BAD);

  LATEX_COMMENT("Rincmd register minor errors");
  DECLARE_LATEX_INT(RC_REG_ERR_NONE);

  LATEX_COMMENT("Rincmd minor read errors");
  DECLARE_LATEX_INT(RC_READ_ERR_NONE);
  DECLARE_LATEX_INT(RC_READ_ERR_PERMISSION);
  DECLARE_LATEX_INT(RC_READ_ERR_UNKNOWNTYPE);
  DECLARE_LATEX_INT(RC_READ_ERR_NOTYPDAT);
  DECLARE_LATEX_INT(RC_READ_ERR_COMMAND);
  DECLARE_LATEX_INT(RC_READ_ERR_BADDATA);
  DECLARE_LATEX_INT(RC_READ_ERR_REGNULL);
  DECLARE_LATEX_INT(RC_READ_ERR_BUFSIZE);
  DECLARE_LATEX_INT(RC_READ_ERR_SNPRINTF);

  LATEX_COMMENT("Rincmd minor write errors");
  DECLARE_LATEX_INT(RC_WRITE_ERR_NONE);
  DECLARE_LATEX_INT(RC_WRITE_ERR_PERMISSION);
  DECLARE_LATEX_INT(RC_WRITE_ERR_UNKNOWNTYPE);
  DECLARE_LATEX_INT(RC_WRITE_ERR_NOTYPDAT);
  DECLARE_LATEX_INT(RC_WRITE_ERR_COMMAND);
  DECLARE_LATEX_INT(RC_WRITE_ERR_PARSE);
  DECLARE_LATEX_INT(RC_WRITE_ERR_RANGE_LOW);
  DECLARE_LATEX_INT(RC_WRITE_ERR_RANGE_HIGH);
  DECLARE_LATEX_INT(RC_WRITE_ERR_LENGTH_LOW);
  DECLARE_LATEX_INT(RC_WRITE_ERR_LENGTH_HIGH);
  DECLARE_LATEX_INT(RC_WRITE_ERR_REGNULL);
  DECLARE_LATEX_INT(RC_WRITE_ERR_BUFSIZE);
  DECLARE_LATEX_INT(RC_WRITE_ERR_SNPRINTF);
  DECLARE_LATEX_INT(RC_WRITE_ERR_STRTOLL);
  DECLARE_LATEX_INT(RC_WRITE_ERR_TYPE_LEN);

  LATEX_COMMENT("Rincmd minor exec errors");
  DECLARE_LATEX_INT(RC_EXEC_ERR_NONE);
  DECLARE_LATEX_INT(RC_EXEC_ERR_PERMISSION);
  DECLARE_LATEX_INT(RC_EXEC_ERR_UNKNOWNTYPE);
  DECLARE_LATEX_INT(RC_EXEC_ERR_TYPDAT);
  DECLARE_LATEX_INT(RC_EXEC_ERR_COMMAND);
  DECLARE_LATEX_INT(RC_EXEC_ERR_REGNULL);
  DECLARE_LATEX_INT(RC_EXEC_ERR_WRITEBACK);

  LATEX_COMMENT("");
  LATEX_COMMENT("Ethernet settings");
  DECLARE_LATEX_OFF_ON(tlt_list_itemize, eth_use_dhcp, DEFAULT_ETH_USEDHCP);
  DECLARE_LATEX_INT(DEFAULT_ETH_USEDHCP);
  DECLARE_LATEX_IP_NET_ENDIAN(DEFAULT_ETH_IPADDRESS);
  DECLARE_LATEX_IP_NET_ENDIAN(DEFAULT_ETH_NETMASK);
  DECLARE_LATEX_IP_NET_ENDIAN(DEFAULT_ETH_GATEWAY);
  DECLARE_LATEX_IP_NET_ENDIAN(DEFAULT_ETH_DNS1);
  DECLARE_LATEX_IP_NET_ENDIAN(DEFAULT_ETH_DNS2);
  DECLARE_LATEX_IP_NET_ENDIAN(DEFAULT_ETH_DNS3);
  DECLARE_LATEX_STR(DEFAULT_ETH_DOMAIN);
  DECLARE_LATEX_STR(DEFAULT_ETH_SEARCH_DOMAIN);
  DECLARE_LATEX_INT(SNC_HOST_NAME_MAX);
  DECLARE_LATEX_INT(SNC_DOMAIN_NAME_MAX);

  LATEX_COMMENT("");
  LATEX_COMMENT("Serial port settings");
  DECLARE_LATEX_LIST_NAMED_DESCRIPTION(tlt_list_itemize, ser_type_setting_serial, ser_type_setting, NUM_DB_SERIAL_TYPES, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_SIMPLE(tlt_list_itemize, baud_setting, NUM_DB_BAUD_SETTINGS, DB_SERIAL_DEFAULT_BAUD_INDEX, NO_OFFSET);

  LATEX_COMMENT("Serial TCP settings");
  DECLARE_LATEX_INT(DB_SERIAL_TCPPORT_LOW_LIMIT);
  DECLARE_LATEX_INT(DB_SERIAL_TCPPORT_HIGH_LIMIT);
  DECLARE_LATEX_INT(DB_SERIAL_DEFAULT_TCPI1_PORT);
  DECLARE_LATEX_INT(DB_SERIAL_DEFAULT_TCPI2_PORT);
  DECLARE_LATEX_INT(DB_SERIAL_TCPTIMEOUT_LOW_LIMIT);
  DECLARE_LATEX_INT(DB_SERIAL_TCPTIMEOUT_HIGH_LIMIT);
  DECLARE_LATEX_INT(DB_SERIAL_DEFAULT_TCPTIMEOUT);

  LATEX_COMMENT("Serial UDP settings");
  DECLARE_LATEX_IP_HOST_ENDIAN(DB_SERIAL_DEFAULT_UDPO_DEST);
  DECLARE_LATEX_INT(DB_SERIAL_UDP_PORT_LOW_LIMIT);
  DECLARE_LATEX_INT(DB_SERIAL_UDP_PORT_HIGH_LIMIT);
  DECLARE_LATEX_INT(DB_SERIAL_DEFAULT_UDPO_PORT);
  DECLARE_LATEX_INT(DB_SERIAL_DEFAULT_UDPI_PORT);
  DECLARE_LATEX_LIST_NAMED_DESCRIPTION(tlt_list_itemize, ser_type_setting_udp, ser_type_setting, NUM_DB_SERIAL_TYPES_BASIC, DB_SERIAL_DEFAULT_TYPE_UDPO1, NO_OFFSET);

  LATEX_COMMENT("Serial network settings (NET.OPT)");
  DECLARE_LATEX_INT(APP_DB_U_SERADDRESS_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SERADDRESS_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SERADDRESS_DEFAULT);
  DECLARE_LATEX_CSTR_LENGTH(APP_DB_U_SERID_MAXLEN);
  DECLARE_LATEX_STR(APP_DB_U_SERID_DEFAULT);

  LATEX_COMMENT("Serial auto output settings (AUT.OPT)");
  DECLARE_LATEX_INT(AUTO_OUT_FMT_NUM);
  DECLARE_LATEX_LIST_DESCRIPTION(tlt_list_itemize, auto_setting, AUTO_OUT_FMT_NUM, APP_DB_U_SERAOFTYPE_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, auto_setting_comms, auto_setting, AUTO_OUT_FMT_NUM, TEX_EASYLIST_NO_DEFAULT, 1);
  DECLARE_LATEX_INT(APP_DB_U_SERSTARTCHAR_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SERSTARTCHAR_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SERSTARTCHAR_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_SERENDCHAR1_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SERENDCHAR1_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SERENDCHAR1_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_SERENDCHAR2_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SERENDCHAR2_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SERENDCHAR2_DEFAULT);

  LATEX_COMMENT("Serial printer settings (PRN.OPT)");
  DECLARE_LATEX_LIST_DESCRIPTION(tlt_list_itemize, print_interlock_setting, NUM_PRINT_INTERLOCK, APP_DB_U_PRINTINTERLOCK_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_DESCRIPTION(tlt_comms_list, print_interlock_setting_comms, print_interlock_setting, NUM_PRINT_INTERLOCK, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_INT(APP_DB_U_PRINTINTERLOCK_DEFAULT);

  LATEX_COMMENT("");
  LATEX_COMMENT("Special menu settings");
  DECLARE_LATEX_OFF_ON(tlt_list_itemize, buzzer_enable, APP_DB_U_BUZENABLE_DEFAULT != 0);
  DECLARE_LATEX_INT(APP_DB_U_BRIGHTNESS_MIN);
  DECLARE_LATEX_INT(APP_DB_U_BRIGHTNESS_MAX);
  DECLARE_LATEX_INT(APP_DB_U_BRIGHTNESS_DEFAULT);

  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_list_itemize, alt_sec_units_menus, alt_units_menus, APP_DB_U_UNITS_SEC_MAX + 1, APP_DB_U_UNITS_SEC_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, alt_sec_units_menus_comms, alt_units_menus, APP_DB_U_UNITS_SEC_MAX + 1, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_list_itemize, alt_ter_units_menus, alt_units_menus, APP_DB_U_UNITS_TER_MAX + 1, APP_DB_U_UNITS_TER_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, alt_ter_units_menus_comms, alt_units_menus, APP_DB_U_UNITS_TER_MAX + 1, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_INT(APP_DB_U_UNITS_SEC_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_UNITS_TER_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_UNITS_CONV_MIN);
  DECLARE_LATEX_INT(APP_DB_U_UNITS_CONV_MAX);
  DECLARE_LATEX_INT(APP_DB_U_UNITS_CONV_DEFAULT);
  DECLARE_LATEX_OFF_ON(tlt_list_itemize, show_lboz, APP_DB_U_SHOW_LBOZ_DEFAULT != 0);
  DECLARE_LATEX_OFF_ON(tlt_comms_list, show_lboz_comms, APP_DB_U_SHOW_LBOZ_DEFAULT != 0);
  DECLARE_LATEX_INT(APP_DB_U_SHOW_LBOZ_DEFAULT);

  LATEX_COMMENT("");
  LATEX_COMMENT("Front and remote function key settings");
  DECLARE_LATEX_LIST_SIMPLE(tlt_list_itemize, remote_keyfn_strings, NUM_REMOTE_KEY_FUNCTIONS, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_SIMPLE(tlt_list_itemize, frontfunc_keyfn_strings, NUM_FRONTFUNC, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_DESCRIPTION(tlt_comms_list, remote_keyfn_strings_comms, remote_keyfn_strings, NUM_REMOTE_KEY_FUNCTIONS, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_DESCRIPTION(tlt_comms_list, frontfunc_keyfn_strings_comms, frontfunc_keyfn_strings, NUM_FRONTFUNC, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);

  LATEX_COMMENT("");
  LATEX_COMMENT("Build menu settings");
  DECLARE_LATEX_LIST_SIMPLE(tlt_list_itemize, build_units_menus, NUM_UNITS, APP_DB_U_ADCUNITS_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, build_units_menus_comms, build_units_menus, NUM_UNITS, APP_DB_U_ADCUNITS_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_INT(APP_DB_U_ADCUNITS_DEFAULT);

  LATEX_COMMENT("");
  LATEX_COMMENT("Option menu settings");
  DECLARE_LATEX_LIST_DESCRIPTION(tlt_list_itemize, we_approval_mode_strings, WE_APPROVAL_MODE_NUM, WE_APPROVAL_MODE_INDUST, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, we_approval_mode_strings_comms, we_approval_mode_strings, WE_APPROVAL_MODE_NUM, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);


  LATEX_COMMENT("");
  LATEX_COMMENT("Calibration settings");
  DECLARE_LATEX_INT(APP_DB_U_ADCZERO_DUVV_MIN);
  DECLARE_LATEX_INT(APP_DB_U_ADCZERO_DUVV_MAX);
  DECLARE_LATEX_INT(APP_DB_U_ADCSPAN_DUVV_MIN);
  DECLARE_LATEX_INT(APP_DB_U_ADCSPAN_DUVV_MAX);
  DECLARE_LATEX_FLOAT(APP_DB_U_ADCZERO_MVV_MIN, 1);
  DECLARE_LATEX_FLOAT(APP_DB_U_ADCZERO_MVV_MAX, 1);
  DECLARE_LATEX_FLOAT(APP_DB_U_ADCSPAN_MVV_MIN, 1);
  DECLARE_LATEX_FLOAT(APP_DB_U_ADCSPAN_MVV_MAX, 1);
  DECLARE_LATEX_INT(ADC_MAX_NUM_LIN_PTS);

  LATEX_COMMENT("");
  LATEX_COMMENT("Clock menu settings");
  DECLARE_LATEX_INT(APP_DB_U_TIME_FMT_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_DATE_FMT_DEFAULT);
  DECLARE_LATEX_LIST_SIMPLE(tlt_list_itemize, time_format_settings, NUM_TIME_FMTS, APP_DB_U_TIME_FMT_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_SIMPLE(tlt_list_itemize, date_format_settings, NUM_DATE_FMTS, APP_DB_U_DATE_FMT_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, time_format_settings_comms, time_format_settings, NUM_TIME_FMTS, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, date_format_settings_comms, date_format_settings, NUM_DATE_FMTS, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);

  LATEX_COMMENT("");
  LATEX_COMMENT("Setpoint specific settings");
  DECLARE_LATEX_LIST_DESCRIPTION(tlt_list_itemize, setpoint_type_strings, SETP_TYPE_NUM, APP_DB_U_SETPACTIVE_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, setpoint_type_strings_comms, setpoint_type_strings, SETP_TYPE_NUM, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_INT(APP_DB_U_SETPACTIVE_DEFAULT);
  DECLARE_LATEX_LIST_DESCRIPTION(tlt_list_itemize, setp_timing_strings, SETP_TIMING_NUM, APP_DB_U_SETPTIMING_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, setp_timing_strings_comms, setp_timing_strings, SETP_TIMING_NUM, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_INT(APP_DB_U_SETPTIMING_DEFAULT);
  DECLARE_LATEX_LIST_DESCRIPTION(tlt_list_itemize, setp_reset_strings, SETP_RESET_NUM, APP_DB_U_SETPRESET_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, setp_reset_strings_comms, setp_reset_strings, SETP_RESET_NUM, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_INT(APP_DB_U_SETPRESET_DEFAULT);

  LATEX_COMMENT("Setpoint general settings");
  DECLARE_LATEX_LIST_DESCRIPTION(tlt_list_itemize, setp_feeder_strings, SETP_FEEDER_NUM, APP_DB_U_SETPFEEDER_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_LIST_NAMED_SIMPLE(tlt_comms_list, setp_feeder_strings_comms, setp_feeder_strings, SETP_FEEDER_NUM, TEX_EASYLIST_NO_DEFAULT, NO_OFFSET);
  DECLARE_LATEX_INT(APP_DB_U_SETPFEEDER_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGON_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGON_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGON_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGOFF_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGOFF_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGOFF_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGSET_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGSET_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SETPJOGSET_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_SETPDELAYCHECK_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SETPDELAYCHECK_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SETPDELAYCHECK_DEFAULT);
  DECLARE_LATEX_INT(APP_DB_U_SETPMAXSET_MIN);
  DECLARE_LATEX_INT(APP_DB_U_SETPMAXSET_MAX);
  DECLARE_LATEX_INT(APP_DB_U_SETPMAXSET_DEFAULT);

  LATEX_COMMENT("");
  LATEX_COMMENT("Hardcoded configuration settings (we_config.h)");
  DECLARE_LATEX_INT(WECFG_NET_MAX_TCP_CONNECTIONS);

  LATEX_COMMENT("");
  LATEX_COMMENT("Errors");
  DECLARE_LATEX_INT(APP_ERR_NONE);
  DECLARE_LATEX_INT(APP_ERR_LOW_12V);
  DECLARE_LATEX_INT(APP_ERR_HIGH_12V);
  DECLARE_LATEX_INT(APP_ERR_SENSE_POS);
  DECLARE_LATEX_INT(APP_ERR_SENSE_NEG);
  DECLARE_LATEX_INT(ADC_ERR_TEMPERATURE);
  DECLARE_LATEX_INT(APP_ERR_MODULE);
  DECLARE_LATEX_INT(APP_ERR_TILT);
  DECLARE_LATEX_INT(APP_ERR_RESOLUTION);
  DECLARE_LATEX_INT(APP_ERR_RTC);
  DECLARE_LATEX_INT(ADC_ERR_USER_DB);
  DECLARE_LATEX_INT(ADC_ERR_FACTORY_DB);
  DECLARE_LATEX_INT(APP_ERR_LOW_EXCITATION);
  DECLARE_LATEX_INT(APP_ERR_HIGH_EXCITATION);
  DECLARE_LATEX_INT(ADC_ERR_ADC_MAXED);
  DECLARE_LATEX_INT(ADC_ERR_RUNTIME_DB);
  DECLARE_LATEX_INT(APP_ERR_ROM);
  DECLARE_LATEX_INT(ADC_ERR_HARDWARE);
  DECLARE_LATEX_INT(APP_ERR_UNTESTED);

  /* Clean up */
mfail:
  if (NULL != epic_buffer)
  {
    free(epic_buffer);
    epic_buffer = NULL;
  }
  return result;
}
